!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines the basic variable types
!> \note
!>      Data type definitions; tested on:
!>          - IBM AIX xlf90
!>          - SGI IRIX  f90
!>          - CRAY T3E  f90
!>          - DEC ALPHA f90
!>          - NAG_F90
!>          - SUN
!>          - HITACHI
!> \par History
!>      Adapted for CP2K by JGH
!> \author Matthias Krack
! **************************************************************************************************
MODULE kinds

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: sp, dp, print_kind_info, dp_size, sp_size, int_size
   PUBLIC :: int_1, int_4, int_8, int_1_size, int_2_size, int_4_size, int_8_size
   PUBLIC :: real_4, real_8, real_4_size, real_8_size
   PUBLIC :: default_string_length, default_path_length, max_line_length

   INTEGER, PARAMETER :: sp = SELECTED_REAL_KIND(6, 30)
   INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(14, 200)
   ! we rely on this (libraries) but do not check this
   INTEGER, PARAMETER :: dp_size = 8, &
                         int_size = BIT_SIZE(0)/8, &
                         sp_size = 4

   INTEGER, PARAMETER :: real_4 = SELECTED_REAL_KIND(6, 30)
   INTEGER, PARAMETER :: real_8 = SELECTED_REAL_KIND(14, 200)
   INTEGER, PARAMETER :: real_4_size = 4
   INTEGER, PARAMETER :: real_8_size = 8

   INTEGER, PARAMETER :: int_1 = SELECTED_INT_KIND(2)
   INTEGER, PARAMETER :: int_1_size = BIT_SIZE(INT(0, int_1))/8

   INTEGER, PARAMETER :: int_2 = SELECTED_INT_KIND(4)
   INTEGER, PARAMETER :: int_2_size = BIT_SIZE(INT(0, int_2))/8

   INTEGER, PARAMETER :: int_4 = SELECTED_INT_KIND(5)
   INTEGER, PARAMETER :: int_4_size = BIT_SIZE(INT(0, int_4))/8

   INTEGER, PARAMETER :: int_8 = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: int_8_size = BIT_SIZE(INT(0, int_8))/8

   INTEGER, PARAMETER :: default_string_length = 80
   INTEGER, PARAMETER :: default_path_length = 1024
   INTEGER, PARAMETER :: max_line_length = 2*default_path_length
   CHARACTER(LEN=1), PARAMETER, PUBLIC :: default_blank_character(2) = (/" ", CHAR(9)/)

CONTAINS

! **************************************************************************************************
!> \brief Print informations about the used data types.
!> \param iw ...
!> \par History
!>      Adapted by JGH for Cp2k
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE print_kind_info(iw)

      INTEGER, INTENT(IN)                                :: iw

      WRITE (iw, '( /, T2, A )') 'DATA TYPE INFORMATION:'

      WRITE (iw, '( /,T2,A,T79,A,2(/,T2,A,T75,I6),3(/,T2,A,T67,E14.8) )') &
         'REAL: Data type name:', 'dp', '      Kind value:', KIND(0.0_dp), &
         '      Precision:', PRECISION(0.0_dp), &
         '      Smallest non-negligible quantity relative to 1:', &
         EPSILON(0.0_dp), &
         '      Smallest positive number:', TINY(0.0_dp), &
         '      Largest representable number:', HUGE(0.0_dp)
      WRITE (iw, '( /,T2,A,T79,A,2(/,T2,A,T75,I6),3(/,T2,A,T67,E14.8) )') &
         '      Data type name:', 'sp', '      Kind value:', KIND(0.0_sp), &
         '      Precision:', PRECISION(0.0_sp), &
         '      Smallest non-negligible quantity relative to 1:', &
         EPSILON(0.0_sp), &
         '      Smallest positive number:', TINY(0.0_sp), &
         '      Largest representable number:', HUGE(0.0_sp)
      WRITE (iw, '( /,T2,A,T72,A,4(/,T2,A,T61,I20) )') &
         'INTEGER: Data type name:', '(default)', '         Kind value:', &
         KIND(0), &
         '         Bit size:', BIT_SIZE(0), &
         '         Largest representable number:', HUGE(0)
      WRITE (iw, '( /,T2,A,T72,A,/,T2,A,T75,I6,/ )') &
         'LOGICAL: Data type name:', '(default)', &
         '         Kind value:', KIND(.TRUE.)
      WRITE (iw, '( /,T2,A,T72,A,/,T2,A,T75,I6,/ )') &
         'CHARACTER: Data type name:', '(default)', &
         '           Kind value:', KIND('C')

   END SUBROUTINE print_kind_info

END MODULE kinds

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE machine_internal

! use POSIX per default
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Implemenation of machine interface based on Fortran 2003 and POSIX
!> \author Ole Schuett
! *****************************************************************************
  USE kinds,                           ONLY: dp, int_8, default_path_length,&
                                             default_string_length
  USE ISO_C_BINDING, ONLY: C_INT, C_NULL_CHAR, C_CHAR, C_PTR, C_NULL_PTR, C_ASSOCIATED, C_F_POINTER

  IMPLICIT NONE
  PRIVATE

  PUBLIC :: m_flush, m_memory, &
            m_hostnm, m_getcwd, m_getlog, m_getuid, m_getpid, m_getarg, &
            m_iargc, m_abort, m_chdir, m_mov, &
            m_memory_details, m_procrun

  INTEGER(KIND=int_8), PUBLIC, SAVE :: m_memory_max=0

CONTAINS

! *****************************************************************************
! can be used to get a nice core
! *****************************************************************************
  SUBROUTINE m_abort()
    INTERFACE
      SUBROUTINE abort() BIND(C,name="abort")
      END SUBROUTINE
    END INTERFACE

    CALL abort()
  END SUBROUTINE m_abort


! *****************************************************************************
! the number of arguments of the fortran program
! *****************************************************************************
  FUNCTION m_iargc() RESULT (ic)
    INTEGER                                  :: ic

    ic = COMMAND_ARGUMENT_COUNT ()
  END FUNCTION m_iargc


! *****************************************************************************
! flush a given unit
! *****************************************************************************
  SUBROUTINE m_flush(lunit)
    INTEGER, INTENT(IN)                      :: lunit

    FLUSH(lunit)
  END SUBROUTINE m_flush


! *****************************************************************************
! returns if a process is running on the local machine
! 1 if yes and 0 if not
! *****************************************************************************
  FUNCTION m_procrun(pid) RESULT (run_on)
    INTEGER, INTENT(IN)       ::   pid
    INTEGER                   ::   run_on
    INTEGER                   ::   istat

    INTERFACE
      FUNCTION kill(pid, sig) BIND(C,name="kill") RESULT(errno)
        IMPORT
        INTEGER(KIND=C_INT),VALUE                :: pid, sig
        INTEGER(KIND=C_INT)                      :: errno
      END FUNCTION
    END INTERFACE

    ! If sig is 0, then no signal is sent, but error checking is still
    ! performed; this can be used to check for the existence of a process
    ! ID or process group ID.

    istat = kill(pid=pid, sig=0)
    IF(istat == 0) THEN
       run_on = 1 ! no error, process exists
    ELSE
       run_on = 0 ! error, process probably does not exist
    ENDIF

  END FUNCTION m_procrun


! *****************************************************************************
! returns the total amount of memory [bytes] in use, if known, zero otherwise
! *****************************************************************************
  SUBROUTINE m_memory(mem)

      INTEGER(KIND=int_8), OPTIONAL, INTENT(OUT)         :: mem
      INTEGER(KIND=int_8)                      :: m1,m2,m3,mem_local

      !
      ! __NO_STATM_ACCESS can be used to disable the stuff, if getpagesize
      ! lead to linking errors or /proc/self/statm can not be opened
      !
      CHARACTER(LEN=80) :: DATA
      INTEGER :: iostat,i

      ! the size of a page, might not be available everywhere
!      INTERFACE
!       FUNCTION getpagesize() BIND(C,name="getpagesize") RESULT(RES)
!         IMPORT
!         INTEGER(C_INT) :: RES
!       END FUNCTION
!      END INTERFACE

      !
      ! reading from statm
      !
      mem_local=-1
      DATA=""
      OPEN(121245,FILE="/proc/self/statm", &
      ACTION="READ",STATUS="OLD",ACCESS="STREAM")
      DO I=1,80
         READ(121245,END=999) DATA(I:I)
      ENDDO
999   CLOSE(121245)
      DATA(I:80)=""
      ! m1 = total
      ! m2 = resident
      ! m3 = shared
      READ(DATA,*,IOSTAT=iostat) m1,m2,m3
      IF (iostat.NE.0) THEN
         mem_local=0
      ELSE
         mem_local=m2
!         mem_local=mem_local*getpagesize()
      ENDIF

      m_memory_max=MAX(mem_local,m_memory_max)
      IF (PRESENT(mem)) mem=mem_local

  END SUBROUTINE m_memory

! *****************************************************************************
! *** get more detailed memory info, all units are bytes.
! *** the only 'useful' option is MemLikelyFree which is an estimate of remaining memory
! *** assumed to give info like /proc/meminfo while MeMLikelyFree is the amount of
! *** memory we're likely to be able to allocate, but not necessarily in one chunk
! *** zero means not available
! *****************************************************************************
  SUBROUTINE m_memory_details(MemTotal,MemFree,Buffers,Cached,Slab,SReclaimable,MemLikelyFree)

     INTEGER(kind=int_8), OPTIONAL :: MemTotal,MemFree,Buffers,Cached,Slab,SReclaimable,MemLikelyFree

     INTEGER, PARAMETER :: Nbuffer=10000
     CHARACTER(LEN=Nbuffer) :: meminfo


     INTEGER :: i

     MemTotal=0
     MemFree=0
     Buffers=0
     Cached=0
     Slab=0
     SReclaimable=0
     MemLikelyFree=0
     meminfo=""

     OPEN(UNIT=8123,file="/proc/meminfo", &
     ACCESS="STREAM",ERR=901)
     i=0
     DO
       i=i+1
       IF (i>Nbuffer) EXIT
       READ(8123,END=900,ERR=900) meminfo(i:i)
     ENDDO
 900 CONTINUE
     meminfo(i:Nbuffer)=""
 901 CONTINUE
     CLOSE(8123,ERR=902)
 902 CONTINUE
     MemTotal=get_field_value_in_bytes('MemTotal:')
     MemFree=get_field_value_in_bytes('MemFree:')
     Buffers=get_field_value_in_bytes('Buffers:')
     Cached=get_field_value_in_bytes('Cached:')
     Slab=get_field_value_in_bytes('Slab:')
     SReclaimable=get_field_value_in_bytes('SReclaimable:')
     ! opinions here vary but this might work
     MemLikelyFree=MemFree+Buffers+Cached+SReclaimable


  CONTAINS
        INTEGER(int_8) FUNCTION get_field_value_in_bytes(field)
           CHARACTER(LEN=*) :: field
           INTEGER :: start
           INTEGER(KIND=int_8) :: value
           get_field_value_in_bytes=0
           start=INDEX(meminfo,field)
           IF (start.NE.0) THEN
              start=start+LEN_TRIM(field)
              IF (start.LT.Nbuffer) THEN
                 READ(meminfo(start:),*,ERR=999,END=999) value
                 ! XXXXXXX convert from Kb to bytes XXXXXXXX
                 get_field_value_in_bytes=value*1024
 999             CONTINUE
              ENDIF
           ENDIF
        END FUNCTION
  END SUBROUTINE m_memory_details


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_mov(source,TARGET)

    CHARACTER(LEN=*), INTENT(IN)             :: source, TARGET

    INTEGER                                  :: istat

    INTERFACE
      FUNCTION unlink(path) BIND(C,name="unlink") RESULT(errno)
        IMPORT
        CHARACTER(KIND=C_CHAR), DIMENSION(*)     :: path
        INTEGER(KIND=C_INT)                      :: errno
      END FUNCTION
    END INTERFACE

    INTERFACE
      FUNCTION rename(src, dest) BIND(C,name="rename") RESULT(errno)
        IMPORT
        CHARACTER(KIND=C_CHAR), DIMENSION(*)     :: src, dest
        INTEGER(KIND=C_INT)                      :: errno
      END FUNCTION
    END INTERFACE

    IF (TARGET==source) THEN
       WRITE(*,*) "Warning: m_mov ",TRIM(TARGET)," equals ", TRIM(source)
       RETURN
    ENDIF

    ! first remove target (needed on windows / mingw)
    istat = unlink(TRIM(TARGET)//c_null_char)
    ! ignore istat of unlink

    ! now move
    istat = rename(TRIM(source)//c_null_char, TRIM(TARGET)//c_null_char)
    IF (istat .NE. 0) THEN
      WRITE(*,*) "Trying to move "//TRIM(source)//" to "//TRIM(TARGET)//"."
      WRITE(*,*) "rename returned status: ",istat
      WRITE(*,*) "Problem moving file"
      CALL m_abort()
    ENDIF
  END SUBROUTINE m_mov


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_hostnm(hname)
    CHARACTER(len=*), INTENT(OUT)            :: hname

    INTEGER                                  :: istat, i
    CHARACTER(len=default_path_length)       :: buf

    INTERFACE
      FUNCTION  gethostname(buf, buflen) BIND(C,name="gethostname") RESULT(errno)
        IMPORT
        CHARACTER(KIND=C_CHAR), DIMENSION(*)     :: buf
        INTEGER(KIND=C_INT), VALUE               :: buflen
        INTEGER(KIND=C_INT)                      :: errno
      END FUNCTION
    END INTERFACE

    istat = gethostname(buf, LEN(buf))
    IF(istat /= 0) THEN
       WRITE (*,*) "m_hostnm failed"
       CALL m_abort()
    ENDIF
    i = INDEX(buf, c_null_char) -1
    hname = buf(1:i)
  END SUBROUTINE m_hostnm


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_getcwd(curdir)
    CHARACTER(len=*), INTENT(OUT)            :: curdir
    TYPE(C_PTR)                              :: stat
    INTEGER                                  :: i
    CHARACTER(len=default_path_length), TARGET  :: tmp

    INTERFACE
      FUNCTION  getcwd(buf, buflen) BIND(C,name="getcwd") RESULT(stat)
        IMPORT
        CHARACTER(KIND=C_CHAR), DIMENSION(*)     :: buf
        INTEGER(KIND=C_INT), VALUE               :: buflen
        TYPE(C_PTR)                              :: stat
      END FUNCTION
    END INTERFACE

    stat = getcwd(tmp, LEN(tmp))
    IF(.NOT. C_ASSOCIATED(stat)) THEN
       WRITE (*,*) "m_getcwd failed"
       CALL m_abort()
    ENDIF
    i = INDEX(tmp, c_null_char) -1
    curdir = tmp(1:i)
  END SUBROUTINE m_getcwd


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_chdir(dir,ierror)
    CHARACTER(len=*), INTENT(IN)             :: dir
    INTEGER, INTENT(OUT)                     :: ierror

    INTERFACE
      FUNCTION chdir(path) BIND(C,name="chdir") RESULT(errno)
        IMPORT
        CHARACTER(KIND=C_CHAR), DIMENSION(*)     :: path
        INTEGER(KIND=C_INT)                      :: errno
      END FUNCTION
    END INTERFACE

    ierror = chdir(TRIM(dir)//c_null_char)
  END SUBROUTINE m_chdir


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_getlog(user)
    CHARACTER(len=*), INTENT(OUT)            :: user
    INTEGER :: status

    ! on a posix system LOGNAME should be defined
    CALL get_environment_variable("LOGNAME", value=user, status=status)
    ! nope, check alternative
    IF (status/=0) &
       CALL get_environment_variable("USER", value=user, status=status)
    ! fall back
    IF (status/=0) &
       user="root ;-)"

  END SUBROUTINE m_getlog


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_getuid(uid)
   INTEGER, INTENT(OUT)                     :: uid

   INTERFACE
     FUNCTION getuid() BIND(C,name="getuid") RESULT(uid)
       IMPORT
       INTEGER(KIND=C_INT)              :: uid
     END FUNCTION
   END INTERFACE

   uid = getuid()
  END SUBROUTINE m_getuid


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_getpid(pid)
   INTEGER, INTENT(OUT)                     :: pid

   INTERFACE
     FUNCTION getpid() BIND(C,name="getpid") RESULT(pid)
       IMPORT
       INTEGER(KIND=C_INT)              :: pid
     END FUNCTION
   END INTERFACE

   pid = getpid()
  END SUBROUTINE m_getpid


! *****************************************************************************
! *****************************************************************************
  SUBROUTINE m_getarg(i,arg)
    INTEGER, INTENT(IN)                      :: i
    CHARACTER(len=*), INTENT(OUT)            :: arg
    CHARACTER(len=1024)                      :: tmp
    INTEGER                                  :: istat

    CALL GET_COMMAND_ARGUMENT(i, tmp, status=istat)
    IF(istat /= 0) THEN
       WRITE (*,*) "m_getarg failed"
       CALL m_abort()
    ENDIF
    arg = TRIM(tmp)
  END SUBROUTINE m_getarg


END MODULE machine_internal
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (05.07.2001) : added G95 interface
!>      - m_flush added (12.06.2002,MK)
!>      - Missing print_memory added (24.09.2002,MK)
!> \author APSI & JGH
! **************************************************************************************************
MODULE machine
   USE ISO_FORTRAN_ENV,                 ONLY: input_unit,&
                                              output_unit
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE machine_internal,                ONLY: &
        m_abort, m_chdir, m_flush_internal=>m_flush, m_getarg, m_getcwd, m_getlog, m_getpid, &
        m_hostnm, m_iargc, m_memory, m_memory_details, m_memory_max, m_mov, m_procrun

 !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads, OMP_GET_WTIME

  IMPLICIT NONE

  ! Except for some error handling code, all code should
  ! get a unit number from the print keys or from the logger, in order
  ! to guarantee correct output behavior,
  ! for example in farming or path integral runs
  ! default_input_unit should never be used
  ! but we need to know what it is, as we should not try to open it for output
  INTEGER, PUBLIC, PARAMETER                   :: default_output_unit = output_unit, &
                                                  default_input_unit  = input_unit

  PRIVATE

  PUBLIC :: m_walltime, m_datum, m_flush, m_flush_internal,&
            m_hostnm, m_getcwd, m_getlog, m_getpid, m_getarg, m_procrun,&
            m_memory, m_iargc, m_abort, m_chdir, m_mov, m_memory_details,&
            m_energy, m_memory_max, m_cpuinfo

  ! should only be set according to the state in &GLOBAL
  LOGICAL, SAVE, PUBLIC :: flush_should_flush=.FALSE.

CONTAINS


! **************************************************************************************************
!> \brief flushes units if the &GLOBAL flag is set accordingly
!> \param lunit ...
!> \par History
!>      10.2008 created [Joost VandeVondele]
!> \note
!>      flushing might degrade performance significantly (30% and more)
! **************************************************************************************************
SUBROUTINE m_flush(lunit)
      INTEGER, INTENT(IN)                                :: lunit

   IF (flush_should_flush) CALL m_flush_internal(lunit)
END SUBROUTINE
! **************************************************************************************************
!> \brief returns time from a real-time clock, protected against rolling
!>      early/easily
!> \retval wt ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
!> \note
!>      same implementation for all machines.
!>      might still roll, if not called multiple times per count_max/count_rate
! **************************************************************************************************
FUNCTION m_walltime() RESULT (wt)
      REAL(KIND=dp)                                      :: wt

      INTEGER(KIND=int_8)                                :: count
      INTEGER(KIND=int_8), SAVE                          :: count_max, count_rate, cycles = -1, &
                                                            last_count

    !$ IF (.FALSE.) THEN
! count lies in [0,count_max] and increases monotonically

    IF (cycles == -1) THEN ! get parameters of system_clock and initialise
        CALL SYSTEM_CLOCK(count_rate=count_rate,count_max=count_max)
        cycles = 0
        last_count = 0
    ENDIF

    CALL SYSTEM_CLOCK(count=count)

    ! protect against non-standard cases where time might be non-monotonous,
    ! but it is unlikely that the clock cycled (e.g. underlying system clock adjustments)
    ! i.e. if count is smaller than last_count by only a small fraction of count_max,
    ! we use last_count instead
    ! if count is smaller, we assume that the clock cycled.
    IF (count<last_count) THEN
       IF ( last_count-count < count_max / 100 ) THEN
          count=last_count
       ELSE
          cycles=cycles+1
       ENDIF
    ENDIF

    ! keep track of our history
    last_count=count

    wt = ( REAL(count,KIND=dp)+REAL(cycles,KIND=dp)*(1.0_dp+REAL(count_max,KIND=dp)) ) &
         / REAL(count_rate,KIND=dp)
    !$ ELSE
    !$    wt = OMP_GET_WTIME ()
    !$ ENDIF
END FUNCTION m_walltime

! **************************************************************************************************
!> \brief reads /proc/cpuinfo if it exists (i.e. Linux) to return relevant info
!> \param model_name as obtained from the 'model name' field, UNKNOWN otherwise
! **************************************************************************************************
SUBROUTINE m_cpuinfo(model_name)
      CHARACTER(LEN=default_string_length)               :: model_name

      INTEGER, PARAMETER                                 :: bufferlen = 2048 

      CHARACTER(LEN=bufferlen)                           :: buffer
      INTEGER                                            :: i, icol, iline, imod, stat

    model_name="UNKNOWN"
    buffer=""
    OPEN(121245,FILE="/proc/cpuinfo",ACTION="READ", &
    STATUS="OLD",ACCESS="STREAM",IOSTAT=stat)
    IF (stat==0) THEN
        DO i=1,bufferlen
           READ(121245,END=999) buffer(I:I)
        ENDDO
999     CLOSE(121245)
        imod=INDEX(buffer,"model name")
        IF (imod>0) THEN
           icol=imod-1+INDEX(buffer(imod:),":")
           iline=icol-1+INDEX(buffer(icol:),NEW_LINE('A'))
           IF (iline==icol-1) iline=bufferlen+1
           model_name=buffer(icol+1:iline-1)
        ENDIF
    ENDIF
END SUBROUTINE m_cpuinfo

! **************************************************************************************************
!> \brief returns the energy used since some time in the past.
!>        The precise meaning depends on the infrastructure is available.
!>        In the cray_pm_energy case, this is the energy used by the node in kJ.
!> \retval wt ...
!> \par History
!>      09.2013 created [Joost VandeVondele, Ole Schuett]
! **************************************************************************************************
FUNCTION m_energy() RESULT (wt)
    REAL(KIND=dp)                            :: wt

   wt = 0.0 ! fallback default

END FUNCTION m_energy



! **************************************************************************************************
!> \brief returns a datum in human readable format using a standard Fortran routine
!> \param cal_date ...
!> \par History
!>      10.2009 created [Joost VandeVondele]
! **************************************************************************************************
SUBROUTINE m_datum(cal_date)
      CHARACTER(len=*), INTENT(OUT)                      :: cal_date

      CHARACTER(len=10)                                  :: time
      CHARACTER(len=8)                                   :: date

    CALL DATE_AND_TIME(date=date, time=time)
    cal_date=date(1:4)//"-"//date(5:6)//"-"//date(7:8)//" "//time(1:2)//":"//time(3:4)//":"//time(5:10)

END SUBROUTINE m_datum

END MODULE machine
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Central dispatch for basic hooks
!> \author Ole Schuett
! **************************************************************************************************
MODULE base_hooks
   USE kinds,                           ONLY: default_string_length
   USE machine,                         ONLY: default_output_unit,&
                                              m_abort,&
                                              m_flush

   IMPLICIT NONE
   PRIVATE

   !API
   PUBLIC :: cp_abort, cp_warn, timeset, timestop
   !API
   PUBLIC :: cp_abort_hook, cp_warn_hook, timeset_hook, timestop_hook
   !API
   PUBLIC :: cp__a, cp__b, cp__w, cp__l

   ! this interface (with subroutines in it) musst to be defined right before
   ! the regular subroutines/functions - otherwise prettify.py will screw up.
   INTERFACE
      SUBROUTINE cp_abort_interface(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      END SUBROUTINE cp_abort_interface

      SUBROUTINE cp_warn_interface(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      END SUBROUTINE cp_warn_interface

      SUBROUTINE timeset_interface(routineN, handle)
      CHARACTER(LEN=*), INTENT(IN)                       :: routineN
      INTEGER, INTENT(OUT)                               :: handle

      END SUBROUTINE timeset_interface

      SUBROUTINE timestop_interface(handle)
      INTEGER, INTENT(IN)                                :: handle

      END SUBROUTINE timestop_interface
   END INTERFACE

   PROCEDURE(cp_abort_interface), POINTER :: cp_abort_hook => Null()
   PROCEDURE(cp_warn_interface), POINTER :: cp_warn_hook => Null()
   PROCEDURE(timeset_interface), POINTER :: timeset_hook => Null()
   PROCEDURE(timestop_interface), POINTER :: timestop_hook => Null()

CONTAINS

! **************************************************************************************************
!> \brief Terminate the program
!> \param location ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_abort(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      IF (ASSOCIATED(cp_abort_hook)) THEN
         CALL cp_abort_hook(location, message)
      ELSE
         WRITE (default_output_unit, *) "ABORT in "//TRIM(location)//" "//TRIM(message)
         CALL m_flush(default_output_unit)
         CALL m_abort()
      ENDIF
      ! compiler hint
      STOP "Never return from here"
   END SUBROUTINE cp_abort

! **************************************************************************************************
!> \brief Issue a warning
!> \param location ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_warn(location, message)
      CHARACTER(len=*), INTENT(in)                       :: location, message

      IF (ASSOCIATED(cp_warn_hook)) THEN
         CALL cp_warn_hook(location, message)
      ELSE
         WRITE (default_output_unit, *) "WARNING in "//TRIM(location)//" "//TRIM(message)
         CALL m_flush(default_output_unit)
      ENDIF
   END SUBROUTINE cp_warn

! **************************************************************************************************
!> \brief Start timer
!> \param routineN ...
!> \param handle ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE timeset(routineN, handle)
      CHARACTER(LEN=*), INTENT(IN)                       :: routineN
      INTEGER, INTENT(OUT)                               :: handle

      IF (ASSOCIATED(timeset_hook)) THEN
         CALL timeset_hook(routineN, handle)
      ELSE
         handle = -1
      ENDIF
   END SUBROUTINE timeset

! **************************************************************************************************
!> \brief Stop timer
!> \param handle ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE timestop(handle)
      INTEGER, INTENT(IN)                                :: handle

      IF (ASSOCIATED(timestop_hook)) THEN
         CALL timestop_hook(handle)
      ELSE
         IF (handle /= -1) &
            CALL cp_abort(cp__l("base_hooks.F", 124), "Got wrong handle")
      ENDIF
   END SUBROUTINE timestop

! **************************************************************************************************
!> \brief CPASSERT handler
!> \param filename ...
!> \param lineNr ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp__a(filename, lineNr)
      CHARACTER(len=*), INTENT(in)                       :: filename
      INTEGER, INTENT(in)                                :: lineNr

      CALL cp_abort(location=cp__l(filename, lineNr), message="CPASSERT failed")
      ! compiler hint
      STOP "Never return from here"
   END SUBROUTINE cp__a

! **************************************************************************************************
!> \brief CPABORT handler
!> \param filename ...
!> \param lineNr ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp__b(filename, lineNr, message)
      CHARACTER(len=*), INTENT(in)                       :: filename
      INTEGER, INTENT(in)                                :: lineNr
      CHARACTER(len=*), INTENT(in)                       :: message

      CALL cp_abort(location=cp__l(filename, lineNr), message=message)
      ! compiler hint
      STOP "Never return from here"
   END SUBROUTINE cp__b

! **************************************************************************************************
!> \brief CPWARN handler
!> \param filename ...
!> \param lineNr ...
!> \param message ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp__w(filename, lineNr, message)
      CHARACTER(len=*), INTENT(in)                       :: filename
      INTEGER, INTENT(in)                                :: lineNr
      CHARACTER(len=*), INTENT(in)                       :: message

      CALL cp_warn(location=cp__l(filename, lineNr), message=message)
   END SUBROUTINE cp__w

! **************************************************************************************************
!> \brief Helper routine to assemble __LOCATION__
!> \param filename ...
!> \param lineNr ...
!> \retval location ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION cp__l(filename, lineNr) RESULT(location)
      CHARACTER(len=*), INTENT(in)                       :: filename
      INTEGER, INTENT(in)                                :: lineNr
      CHARACTER(len=default_string_length)               :: location

      CHARACTER(len=15)                                  :: lineNr_str

      WRITE (lineNr_str, FMT='(I10)') lineNr
      location = TRIM(filename)//":"//TRIM(ADJUSTL(lineNr_str))

   END FUNCTION cp__l

END MODULE base_hooks
!
!  Imported from Arjen Markus' flibs project
!  http://flibs.sourceforge.net/
!
! globmatch.f90 --
!     Match strings according to (simplified) glob patterns
!
!     The pattern matching is limited to literals, * and ?
!     (character classes are not supported). A backslash escapes
!     any character.
!
! **************************************************************************************************
MODULE glob_matching

   IMPLICIT NONE

   CHARACTER(len=1), PARAMETER, PRIVATE :: backslash = '\\'
   CHARACTER(len=1), PARAMETER, PRIVATE :: star = '*'
   CHARACTER(len=1), PARAMETER, PRIVATE :: question = '?'

   PRIVATE
   PUBLIC :: string_match

CONTAINS

! string_match --
!     Tries to match the given string with the pattern
! Arguments:
!     string     String to be examined
!     pattern    Glob pattern to be used for the matching
! Result:
!     .true. if the entire string matches the pattern, .false.
!     otherwise
! Note:
!     Trailing blanks are ignored
!
! **************************************************************************************************
!> \brief ...
!> \param string ...
!> \param pattern ...
!> \retval match ...
! **************************************************************************************************
   RECURSIVE FUNCTION string_match(string, pattern) RESULT(match)
      CHARACTER(len=*), INTENT(in)                       :: string, pattern
      LOGICAL                                            :: match

      CHARACTER(len=LEN(pattern))                        :: literal
      INTEGER                                            :: k, ll, method, p, ptrim, start, strim

      match = .FALSE.
      method = 0
      ptrim = LEN_TRIM(pattern)
      strim = LEN_TRIM(string)
      p = 1
      ll = 0
      start = 1

      !
      ! Split off a piece of the pattern
      !
      DO WHILE (p <= ptrim)
         SELECT CASE (pattern (p:p))
         CASE (star)
            IF (ll .NE. 0) EXIT
            method = 1
         CASE (question)
            IF (ll .NE. 0) EXIT
            method = 2
            start = start+1
         CASE (backslash)
            p = p+1
            ll = ll+1
            literal(ll:ll) = pattern(p:p)
         CASE default
            ll = ll+1
            literal(ll:ll) = pattern(p:p)
         END SELECT

         p = p+1
      ENDDO

      !
      ! Now look for the literal string (if any!)
      !
      IF (method == 0) THEN
         !
         ! We are at the end of the pattern, and of the string?
         !
         IF (strim == 0 .AND. ptrim == 0) THEN
            match = .TRUE.
         ELSE
            !
            ! The string matches a literal part?
            !
            IF (ll > 0) THEN
               IF (string(start:MIN(strim, start+ll-1)) == literal(1:ll)) THEN
                  start = start+ll
                  match = string_match(string(start:), pattern(p:))
               ENDIF
            ENDIF
         ENDIF
      ENDIF

      IF (method == 1) THEN
         !
         ! Scan the whole of the remaining string ...
         !
         IF (ll == 0) THEN
            match = .TRUE.
         ELSE
            DO WHILE (start <= strim)
               k = INDEX(string(start:), literal(1:ll))
               IF (k > 0) THEN
                  start = start+k+ll-1
                  match = string_match(string(start:), pattern(p:))
                  IF (match) THEN
                     EXIT
                  ENDIF
               ENDIF

               start = start+1
            ENDDO
         ENDIF
      ENDIF

      IF (method == 2 .AND. ll > 0) THEN
         !
         ! Scan the whole of the remaining string ...
         !
         IF (string(start:MIN(strim, start+ll-1)) == literal(1:ll)) THEN
            match = string_match(string(start+ll:), pattern(p:))
         ENDIF
      ENDIF
      RETURN
   END FUNCTION string_match

END MODULE glob_matching
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for string manipulations
!> \par History
!>      Adapted compress and uppercase for use in CP2K (JGH)
!>      string_to_integer and integer_to_string added (06.02.2001, MK)
!>      Cleaned (04.01.2004,MK)
!> \author MK & JGH
! **************************************************************************************************
MODULE string_utilities

   USE glob_matching,                   ONLY: pattern_match=>string_match
   USE kinds,                           ONLY: default_blank_character

   IMPLICIT NONE

   PRIVATE

   CHARACTER(LEN=1), PARAMETER :: backslash = '\\'
   CHARACTER(LEN=1), PARAMETER :: star = '*'
   CHARACTER(LEN=1), PARAMETER :: question = '?'
   CHARACTER(LEN=1), PARAMETER :: newline = ACHAR(10)

   INTEGER, PARAMETER :: maxlen_entity_name = 10
   CHARACTER(LEN=maxlen_entity_name), DIMENSION(252), PARAMETER :: html_entity_table = &
                 (/"&quot;    ", "&#34;     ", "&amp;     ", "&#38;#38; ", "&apos;    ", "&#39;     ", "&lt;      ", "&#38;#60; ", &
                   "&gt;      ", "&#62;     ", "&nbsp;    ", "&#160;    ", "&copy;    ", "&#169;    ", "&deg;     ", "&#176;    ", &
                   "&plusmn;  ", "&#177;    ", "&sup2;    ", "&#178;    ", "&sup3;    ", "&#179;    ", "&middot;  ", "&#183;    ", &
                   "&sup1;    ", "&#185;    ", "&Auml;    ", "&#196;    ", "&Aring;   ", "&#197;    ", "&Ccedil;  ", "&#199;    ", &
                   "&Egrave;  ", "&#200;    ", "&Eacute;  ", "&#201;    ", "&Ecirc;   ", "&#202;    ", "&Ograve;  ", "&#210;    ", &
                   "&Oacute;  ", "&#211;    ", "&Ocirc;   ", "&#212;    ", "&Ouml;    ", "&#214;    ", "&times;   ", "&#215;    ", &
                   "&Uuml;    ", "&#220;    ", "&agrave;  ", "&#224;    ", "&aacute;  ", "&#225;    ", "&acirc;   ", "&#226;    ", &
                   "&auml;    ", "&#228;    ", "&aring;   ", "&#229;    ", "&ccedil;  ", "&#231;    ", "&egrave;  ", "&#232;    ", &
                   "&eacute;  ", "&#233;    ", "&ecirc;   ", "&#234;    ", "&ograve;  ", "&#242;    ", "&oacute;  ", "&#243;    ", &
                   "&ocirc;   ", "&#244;    ", "&ouml;    ", "&#246;    ", "&ugrave;  ", "&#249;    ", "&uacute;  ", "&#250;    ", &
                   "&ucirc;   ", "&#251;    ", "&uuml;    ", "&#252;    ", "&Gamma;   ", "&#915;    ", "&Delta;   ", "&#916;    ", &
                   "&Theta;   ", "&#920;    ", "&Lambda;  ", "&#923;    ", "&Xi;      ", "&#926;    ", "&Pi;      ", "&#928;    ", &
                   "&Sigma;   ", "&#931;    ", "&Phi;     ", "&#934;    ", "&Psi;     ", "&#936;    ", "&Omega;   ", "&#937;    ", &
                   "&alpha;   ", "&#945;    ", "&beta;    ", "&#946;    ", "&gamma;   ", "&#947;    ", "&delta;   ", "&#948;    ", &
                   "&epsilon; ", "&#949;    ", "&zeta;    ", "&#950;    ", "&eta;     ", "&#951;    ", "&theta;   ", "&#952;    ", &
                   "&iota;    ", "&#953;    ", "&kappa;   ", "&#954;    ", "&lambda;  ", "&#955;    ", "&mu;      ", "&#956;    ", &
                   "&nu;      ", "&#957;    ", "&xi;      ", "&#958;    ", "&omicron; ", "&#959;    ", "&pi;      ", "&#960;    ", &
                   "&rho;     ", "&#961;    ", "&sigmaf;  ", "&#962;    ", "&sigma;   ", "&#963;    ", "&tau;     ", "&#964;    ", &
                   "&upsilon; ", "&#965;    ", "&phi;     ", "&#966;    ", "&chi;     ", "&#967;    ", "&psi;     ", "&#968;    ", &
                   "&omega;   ", "&#969;    ", "&thetasym;", "&#977;    ", "&ndash;   ", "&#8211;   ", "&mdash;   ", "&#8212;   ", &
                   "&dagger;  ", "&#8224;   ", "&Dagger;  ", "&#8225;   ", "&bull;    ", "&#8230;   ", "&hellip;  ", "&#8230;   ", &
                   "&permil;  ", "&#8240;   ", "&larr;    ", "&#8592;   ", "&uarr;    ", "&#8593;   ", "&rarr;    ", "&#8594;   ", &
                   "&darr;    ", "&#8595;   ", "&harr;    ", "&#8596;   ", "&lArr;    ", "&#8656;   ", "&uArr;    ", "&#8657;   ", &
                   "&rArr;    ", "&#8658;   ", "&dArr;    ", "&#8659;   ", "&hArr;    ", "&#8660;   ", "&forall;  ", "&#8704;   ", &
                   "&part;    ", "&#8706;   ", "&nabla;   ", "&#8711;   ", "&isin;    ", "&#8712;   ", "&notin;   ", "&#8713;   ", &
                   "&ni;      ", "&#8715;   ", "&prop;    ", "&#8733;   ", "&infin;   ", "&#8734;   ", "&ang;     ", "&#8736;   ", &
                   "&and;     ", "&#8743;   ", "&or;      ", "&#8744;   ", "&cap;     ", "&#8745;   ", "&cup;     ", "&#8746;   ", &
                   "&sim;     ", "&#8764;   ", "&cong;    ", "&#8773;   ", "&asymp;   ", "&#8776;   ", "&ne;      ", "&#8800;   ", &
                   "&equiv;   ", "&#8801;   ", "&le;      ", "&#8804;   ", "&ge;      ", "&#8805;   ", "&sub;     ", "&#8834;   ", &
                   "&sup;     ", "&#8835;   ", "&nsub;    ", "&#8836;   ", "&sube;    ", "&#8838;   ", "&supe;    ", "&#8839;   ", &
                   "&oplus;   ", "&#8853;   ", "&otimes;  ", "&#8855;   ", "&perp;    ", "&#8869;   ", "&sdot;    ", "&#8901;   ", &
                                                                     "&lang;    ", "&#9001;   ", "&rang;    ", "&#9002;   "/)

   INTEGER, PARAMETER :: maxlen_tag_name = 10
   CHARACTER(LEN=maxlen_tag_name), DIMENSION(38), PARAMETER :: html_tag_table = &
                 (/"a         ", "b         ", "big       ", "blockquote", "br/       ", "code      ", "dd        ", "del       ", &
                   "div       ", "dl        ", "dt        ", "em        ", "h1        ", "h2        ", "h3        ", "h4        ", &
                   "h5        ", "h6        ", "hr        ", "i         ", "ins       ", "li        ", "ol        ", "p         ", &
                   "span      ", "sub       ", "sup       ", "table     ", "tbody     ", "td        ", "tfoot     ", "th        ", &
                                                "thead     ", "tr        ", "tt        ", "u         ", "ul        ", "pre       "/)

   PUBLIC :: ascii_to_string, &
             compress, &
             integer_to_string, &
             is_whitespace, &
             remove_word, &
             s2a, a2s, &
             str_comp, &
             string_to_ascii, &
             substitute_special_xml_tokens, &
             typo_match, &
             uppercase, &
             write_html_tables, &
             xstring

   PUBLIC :: html_entity_table, &
             newline

   INTERFACE s2a
      MODULE PROCEDURE s2a_1, s2a_2, s2a_3, s2a_4, s2a_5, s2a_6, s2a_7, s2a_8, s2a_9, &
         s2a_10, s2a_11, s2a_12, s2a_13, s2a_14, s2a_15, s2a_16, s2a_17, s2a_18, s2a_19, &
         s2a_20, s2a_21, s2a_22, s2a_23, s2a_24, s2a_25, s2a_26, s2a_27, s2a_28, s2a_29, &
         s2a_30, s2a_31, s2a_32, s2a_33, s2a_34, s2a_35, s2a_36, s2a_37, s2a_38, s2a_39 ! should be clear how to add more
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief returns a non-zero positive value if typo_string equals string apart from a few typos.
!>     It is case sensitive, apart from typos.
!> \param string ...
!> \param typo_string ...
!> \retval match ...
!> \par History
!>      02.2006 created [Joost VandeVondele]
!> \note
!>     could maybe be made a bit smarter (levenstein distance ?)
! **************************************************************************************************
   FUNCTION typo_match(string, typo_string) RESULT(match)
      CHARACTER(LEN=*), INTENT(IN)                       :: string, typo_string
      INTEGER                                            :: match

      CHARACTER(LEN=1)                                   :: kind
      CHARACTER(LEN=LEN(string))                         :: tmp2
      CHARACTER(LEN=LEN(typo_string))                    :: tmp
      INTEGER                                            :: i, j

      match = 0
      IF (LEN_TRIM(typo_string) .LE. 4) THEN
         kind = question
      ELSE
         kind = star
      ENDIF
      DO i = 1, LEN_TRIM(typo_string)
         DO j = i, LEN_TRIM(typo_string)
            tmp = typo_string
            tmp(i:i) = kind
            tmp(j:j) = kind
            IF (i == j .AND. LEN_TRIM(tmp) > 2) tmp(i:i) = star
            IF (pattern_match(string=string, pattern=tmp)) match = match+1
         ENDDO
      ENDDO
      IF (LEN_TRIM(string) .LE. 4) THEN
         kind = question
      ELSE
         kind = star
      ENDIF
      DO i = 1, LEN_TRIM(string)
         DO j = i, LEN_TRIM(string)
            tmp2 = string
            tmp2(i:i) = kind
            tmp2(j:j) = kind
            IF (i == j .AND. LEN_TRIM(tmp2) > 2) tmp2(i:i) = star
            IF (pattern_match(string=typo_string, pattern=tmp2)) match = match+1
         ENDDO
      ENDDO
      IF (match > 0) THEN
         ! bonus points for small differences in length
         IF (ABS(LEN_TRIM(string)-LEN_TRIM(typo_string)) < 3) match = match+2
      ENDIF

   END FUNCTION typo_match

! **************************************************************************************************
!> \brief Converts a character-array into a string
!> \param array ...
!> \retval string ...
!> \par History
!>      10.2013 created [Ole Schuett]
! **************************************************************************************************
   PURE FUNCTION a2s(array) RESULT(string)
      CHARACTER, DIMENSION(:), INTENT(IN)                :: array
      CHARACTER(LEN=SIZE(array))                         :: string

      INTEGER                                            :: i

      DO i = 1, SIZE(array)
         string(i:i) = array(i)
      ENDDO
   END FUNCTION a2s

! **************************************************************************************************
!> \brief converts a bunch of strings of different length to an array of
!>        strings of the same length
!> \param s1 ...
!> \retval a ...
!> \par History
!>      11.2004 created [Joost VandeVondele ]
!> \note
!>     can be used instead of the illegal (/"12","1234"/) generating
!>     s2a("12","1234").EQ.(/"12  ","1234"/)
! **************************************************************************************************
   PURE FUNCTION s2a_1(s1) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1
      CHARACTER(LEN=1000), DIMENSION(1)                  :: a

      a(1) = s1
   END FUNCTION s2a_1
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_2(s1, s2) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2
      CHARACTER(LEN=1000), DIMENSION(2)                  :: a

      a(1) = s1; a(2) = s2
   END FUNCTION s2a_2
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_3(s1, s2, s3) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3
      CHARACTER(LEN=1000), DIMENSION(3)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3
   END FUNCTION s2a_3
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_4(s1, s2, s3, s4) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4
      CHARACTER(LEN=1000), DIMENSION(4)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4
   END FUNCTION s2a_4
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_5(s1, s2, s3, s4, s5) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5
      CHARACTER(LEN=1000), DIMENSION(5)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5
   END FUNCTION s2a_5
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_6(s1, s2, s3, s4, s5, s6) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6
      CHARACTER(LEN=1000), DIMENSION(6)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6
   END FUNCTION s2a_6
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_7(s1, s2, s3, s4, s5, s6, s7) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7
      CHARACTER(LEN=1000), DIMENSION(7)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
   END FUNCTION s2a_7
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_8(s1, s2, s3, s4, s5, s6, s7, s8) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8
      CHARACTER(LEN=1000), DIMENSION(8)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8
   END FUNCTION s2a_8
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_9(s1, s2, s3, s4, s5, s6, s7, s8, s9) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9
      CHARACTER(LEN=1000), DIMENSION(9)                  :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9
   END FUNCTION s2a_9
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_10(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10
      CHARACTER(LEN=1000), DIMENSION(10)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10
   END FUNCTION s2a_10
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_11(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11
      CHARACTER(LEN=1000), DIMENSION(11)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11
   END FUNCTION s2a_11
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_12(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12
      CHARACTER(LEN=1000), DIMENSION(12)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12
   END FUNCTION s2a_12
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_13(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13
      CHARACTER(LEN=1000), DIMENSION(13)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13
   END FUNCTION s2a_13
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_14(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14
      CHARACTER(LEN=1000), DIMENSION(14)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
   END FUNCTION s2a_14
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_15(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15
      CHARACTER(LEN=1000), DIMENSION(15)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15
   END FUNCTION s2a_15
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_16(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16
      CHARACTER(LEN=1000), DIMENSION(16)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16
   END FUNCTION s2a_16
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_17(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17
      CHARACTER(LEN=1000), DIMENSION(17)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17
   END FUNCTION s2a_17
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_18(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, s18
      CHARACTER(LEN=1000), DIMENSION(18)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18
   END FUNCTION s2a_18
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_19(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19
      CHARACTER(LEN=1000), DIMENSION(19)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19
   END FUNCTION s2a_19
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_20(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20
      CHARACTER(LEN=1000), DIMENSION(20)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
   END FUNCTION s2a_20
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_21(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21
      CHARACTER(LEN=1000), DIMENSION(21)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21
   END FUNCTION s2a_21
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_22(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22
      CHARACTER(LEN=1000), DIMENSION(22)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22
   END FUNCTION s2a_22
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_23(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23
      CHARACTER(LEN=1000), DIMENSION(23)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23
   END FUNCTION s2a_23
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_24(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24
      CHARACTER(LEN=1000), DIMENSION(24)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24
   END FUNCTION s2a_24
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_25(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, s25
      CHARACTER(LEN=1000), DIMENSION(25)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25
   END FUNCTION s2a_25
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_26(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26
      CHARACTER(LEN=1000), DIMENSION(26)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
   END FUNCTION s2a_26
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_27(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27
      CHARACTER(LEN=1000), DIMENSION(27)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27
   END FUNCTION s2a_27
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_28(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28
      CHARACTER(LEN=1000), DIMENSION(28)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28
   END FUNCTION s2a_28
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_29(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29
      CHARACTER(LEN=1000), DIMENSION(29)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29
   END FUNCTION s2a_29
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_30(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30
      CHARACTER(LEN=1000), DIMENSION(30)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30
   END FUNCTION s2a_30
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_31(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30, s31
      CHARACTER(LEN=1000), DIMENSION(31)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31
   END FUNCTION s2a_31
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_32(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30, s31, s32
      CHARACTER(LEN=1000), DIMENSION(32)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
   END FUNCTION s2a_32
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_33(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30, s31, &
                                                            s32, s33
      CHARACTER(LEN=1000), DIMENSION(33)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33
   END FUNCTION s2a_33
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_34(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30, s31, &
                                                            s32, s33, s34
      CHARACTER(LEN=1000), DIMENSION(34)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34
   END FUNCTION s2a_34
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \param s35 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_35(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34, s35) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN)                       :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, &
                                                            s11, s12, s13, s14, s15, s16, s17, &
                                                            s18, s19, s20, s21, s22, s23, s24, &
                                                            s25, s26, s27, s28, s29, s30, s31, &
                                                            s32, s33, s34, s35
      CHARACTER(LEN=1000), DIMENSION(35)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34; a(35) = s35
   END FUNCTION s2a_35
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \param s35 ...
!> \param s36 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_36(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34, s35, s36) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN) :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, &
         s15, s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
         s33, s34, s35, s36
      CHARACTER(LEN=1000), DIMENSION(36)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34; a(35) = s35; a(36) = s36
   END FUNCTION s2a_36
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \param s35 ...
!> \param s36 ...
!> \param s37 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_37(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34, s35, s36, s37) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN) :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, &
         s15, s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
         s33, s34, s35, s36, s37
      CHARACTER(LEN=1000), DIMENSION(37)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34; a(35) = s35; a(36) = s36; a(37) = s37
   END FUNCTION s2a_37
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \param s35 ...
!> \param s36 ...
!> \param s37 ...
!> \param s38 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_38(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34, s35, s36, s37, s38) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN) :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, &
         s15, s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
         s33, s34, s35, s36, s37, s38
      CHARACTER(LEN=1000), DIMENSION(38)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34; a(35) = s35; a(36) = s36; a(37) = s37; a(38) = s38
   END FUNCTION s2a_38
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \param s3 ...
!> \param s4 ...
!> \param s5 ...
!> \param s6 ...
!> \param s7 ...
!> \param s8 ...
!> \param s9 ...
!> \param s10 ...
!> \param s11 ...
!> \param s12 ...
!> \param s13 ...
!> \param s14 ...
!> \param s15 ...
!> \param s16 ...
!> \param s17 ...
!> \param s18 ...
!> \param s19 ...
!> \param s20 ...
!> \param s21 ...
!> \param s22 ...
!> \param s23 ...
!> \param s24 ...
!> \param s25 ...
!> \param s26 ...
!> \param s27 ...
!> \param s28 ...
!> \param s29 ...
!> \param s30 ...
!> \param s31 ...
!> \param s32 ...
!> \param s33 ...
!> \param s34 ...
!> \param s35 ...
!> \param s36 ...
!> \param s37 ...
!> \param s38 ...
!> \param s39 ...
!> \retval a ...
! **************************************************************************************************
   PURE FUNCTION s2a_39(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, &
                        s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
                        s33, s34, s35, s36, s37, s38, s39) RESULT(a)
      CHARACTER(LEN=*), INTENT(IN) :: s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, &
         s15, s16, s17, s18, s19, s20, s21, s22, s23, s24, s25, s26, s27, s28, s29, s30, s31, s32, &
         s33, s34, s35, s36, s37, s38, s39
      CHARACTER(LEN=1000), DIMENSION(39)                 :: a

      a(1) = s1; a(2) = s2; a(3) = s3; a(4) = s4; a(5) = s5; a(6) = s6; a(7) = s7
      a(8) = s8; a(9) = s9; a(10) = s10; a(11) = s11; a(12) = s12; a(13) = s13; a(14) = s14
      a(15) = s15; a(16) = s16; a(17) = s17; a(18) = s18; a(19) = s19; a(20) = s20
      a(21) = s21; a(22) = s22; a(23) = s23; a(24) = s24; a(25) = s25; a(26) = s26
      a(27) = s27; a(28) = s28; a(29) = s29; a(30) = s30; a(31) = s31; a(32) = s32
      a(33) = s33; a(34) = s34; a(35) = s35; a(36) = s36; a(37) = s37; a(38) = s38
      a(39) = s39
   END FUNCTION s2a_39
! **************************************************************************************************
!> \brief Convert a sequence of integer numbers (ASCII code) to a string.
!>         Blanks are inserted for invalid ASCII code numbers.
!> \param nascii ...
!> \param string ...
!> \date    19.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE ascii_to_string(nascii, string)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: nascii
      CHARACTER(LEN=*), INTENT(OUT)                      :: string

      INTEGER                                            :: i

      string = ""

      DO i = 1, MIN(LEN(string), SIZE(nascii))
         IF ((nascii(i) >= 0) .AND. (nascii(i) <= 127)) THEN
            string(i:i) = CHAR(nascii(i))
         ELSE
            string(i:i) = " "
         END IF
      END DO

   END SUBROUTINE ascii_to_string

! **************************************************************************************************
!> \brief   Eliminate multiple space characters in a string.
!>          If full is .TRUE., then all spaces are eliminated.
!> \param string ...
!> \param full ...
!> \date    23.06.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE compress(string, full)

      CHARACTER(LEN=*), INTENT(INOUT)                    :: string
      LOGICAL, INTENT(IN), OPTIONAL                      :: full

      CHARACTER                                          :: tmp
      INTEGER                                            :: i, z
      LOGICAL                                            :: remove_all

      IF (PRESENT(full)) THEN
         remove_all = full
      ELSE
         remove_all = .FALSE.
      END IF

      z = 1

      DO i = 1, LEN_TRIM(string)
         IF ((z == 1) .OR. remove_all) THEN
            IF (string(i:i) /= " ") THEN
               tmp = string(i:i)
               string(z:z) = tmp
               z = z+1
            END IF
         ELSE
            IF ((string(i:i) /= " ") .OR. (string(z-1:z-1) /= " ")) THEN
               tmp = string(i:i)
               string(z:z) = tmp
               z = z+1
            END IF
         END IF
      END DO

      string(z:) = ""

   END SUBROUTINE compress

! **************************************************************************************************
!> \brief   Converts an integer number to a string.
!>          The WRITE statement will return an error message, if the number of
!>          digits of the integer number is larger the than the length of the
!>          supplied string.
!> \param inumber ...
!> \param string ...
!> \date    05.01.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE integer_to_string(inumber, string)

      INTEGER, INTENT(IN)                                :: inumber
      CHARACTER(LEN=*), INTENT(OUT)                      :: string

      WRITE (UNIT=string, FMT='(I0)') inumber
   END SUBROUTINE integer_to_string

! **************************************************************************************************
!> \brief   Convert a string to sequence of integer numbers.
!> \param string ...
!> \param nascii ...
!> \date    19.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE string_to_ascii(string, nascii)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: nascii

      INTEGER                                            :: i

      nascii(:) = 0

      DO i = 1, MIN(LEN(string), SIZE(nascii))
         nascii(i) = ICHAR(string(i:i))
      END DO

   END SUBROUTINE string_to_ascii

! **************************************************************************************************
!> \brief   remove a word from a string (words are separated by white spaces)
!> \param string ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE remove_word(string)
      CHARACTER(LEN=*), INTENT(INOUT)                    :: string

      INTEGER                                            :: i

      i = 1
      ! possibly clean white spaces
      DO WHILE (string(i:i) == " ")
         i = i+1
      END DO
      ! now remove the word
      DO WHILE (string(i:i) /= " ")
         i = i+1
      END DO
      string = string(i:)

   END SUBROUTINE remove_word

! **************************************************************************************************
!> \brief  Substitute special XML tokens like "<" or ">" in inp_string.
!>         Optionally convert also all lowercase characters to uppercase, if
!>         ltu is true.
!> \param inp_string ...
!> \param ltu ...
!> \retval out_string ...
!> \date    10.03.2005
!> \par History
!>      - Enable the use of HTML entity names (06.03.13,MK)
!> \author  Matthias Krack (MK)
!> \version 1.1
! **************************************************************************************************
   FUNCTION substitute_special_xml_tokens(inp_string, ltu) RESULT(out_string)

      CHARACTER(LEN=*), INTENT(IN)                       :: inp_string
      LOGICAL, INTENT(IN), OPTIONAL                      :: ltu
      CHARACTER(LEN=2*LEN(inp_string))                   :: out_string

      CHARACTER(LEN=LEN(inp_string))                     :: string
      CHARACTER(LEN=maxlen_entity_name)                  :: entity_name
      CHARACTER(LEN=maxlen_tag_name)                     :: tag_name
      INTEGER                                            :: i, ientry, ilen, j, k

      string = inp_string
      out_string = ""

      IF (PRESENT(ltu)) THEN
         IF (ltu) CALL uppercase(string)
      END IF

      i = 0
      j = 1
      string_loop: DO
         i = i+1
         IF (i > LEN_TRIM(string)) EXIT string_loop
         IF (string(i:i) == "<") THEN
            ! Detect valid HTML tags and keep them
            ientry = 0
            ilen = INDEX(string(i:), ">")
            IF ((ilen > 2) .AND. (ilen <= maxlen_tag_name+3)) THEN
               IF (string(i+1:i+1) == "/") THEN
                  tag_name(1:) = string(i+2:i+ilen-2)
               ELSE
                  tag_name(1:) = string(i+1:i+ilen-2)
               END IF
               CALL lowercase(tag_name)
               tag_loop: DO k = 1, SIZE(html_tag_table)
                  IF (tag_name == html_tag_table(k)) THEN
                     ientry = k
                     EXIT tag_loop
                  END IF
               END DO tag_loop
               IF (ientry > 0) THEN ! HTML tag found in table
                  IF (string(i+1:i+1) == "/") THEN
                     out_string(j:j+ilen+7) = "&#60;/"//TRIM(tag_name)//"&#62;"
                  ELSE
                     out_string(j:j+ilen+7) = "&#60;"//TRIM(tag_name)//"&#62;"
                  END IF
                  i = i+ilen-1
                  j = j+ilen+8
               END IF
            END IF
            ! HTML tag not found in table
            IF (ientry == 0) THEN
               out_string(j:j+4) = "&#60;"
               j = j+5
            END IF
         ELSE IF (string(i:i) == ">") THEN
            out_string(j:j+4) = "&#62;"
            j = j+5
         ELSE IF (string(i:i) == "&") THEN
            ! Substitute HTML entity names by the corresponding entity number
            ientry = 0
            ilen = INDEX(string(i:), ";")
            IF ((ilen > 2) .AND. (ilen <= maxlen_entity_name)) THEN
               entity_name(1:) = string(i:i+ilen-1)
               entity_loop: DO k = 1, SIZE(html_entity_table), 2
                  IF (entity_name == html_entity_table(k)) THEN
                     ientry = k+1
                     EXIT entity_loop
                  END IF
               END DO entity_loop
               i = i+ilen-1
               IF (ientry > 0) THEN ! HTML entity found in table
                  ilen = LEN_TRIM(html_entity_table(ientry))
                  out_string(j:j+ilen-1) = TRIM(ADJUSTL(html_entity_table(ientry)))
                  j = j+ilen
               END IF
            END IF
            IF (ientry == 0) THEN
               out_string(j:j+4) = "&#38;"
               j = j+5
            END IF
         ELSE
            out_string(j:j) = string(i:i)
            j = j+1
         END IF
      END DO string_loop

   END FUNCTION substitute_special_xml_tokens

! **************************************************************************************************
!> \brief   Write the HTML entity and HTML tag table to unit output_unit
!> \param output_unit ...
!> \date    14.08.2013
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_html_tables(output_unit)

      INTEGER, INTENT(IN)                                :: output_unit

      INTEGER                                            :: i

      WRITE (UNIT=output_unit, FMT="(T2,A)") "<CP2K_HTML>"

      DO i = 1, SIZE(html_tag_table)
         WRITE (UNIT=output_unit, FMT="(T3,A)") &
            "<TAG>", &
            " <NAME>&#60;"//TRIM(html_tag_table(i))//"&#62;</NAME>", &
            "</TAG>"
      END DO
      DO i = 1, SIZE(html_entity_table), 2
         WRITE (UNIT=output_unit, FMT="(T3,A)") &
            "<ENTITY>", &
            " <NAME>&#38;"//TRIM(html_entity_table(i) (2:))//"</NAME>", &
            " <CODE>&#38;"//TRIM(html_entity_table(i+1) (2:))//"</CODE>", &
            "</ENTITY>"
      END DO

      WRITE (UNIT=output_unit, FMT="(T2,A)") "</CP2K_HTML>"

   END SUBROUTINE write_html_tables

! **************************************************************************************************
!> \brief   Convert all upper case characters in a string to lower case.
!> \param string ...
!> \date    14.08.2013
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE lowercase(string)
      CHARACTER(LEN=*), INTENT(INOUT)                    :: string

      INTEGER                                            :: i, iascii

      DO i = 1, LEN_TRIM(string)
         iascii = ICHAR(string(i:i))
         IF ((iascii >= 65) .AND. (iascii <= 90)) THEN
            string(i:i) = CHAR(iascii+32)
         END IF
      END DO

   END SUBROUTINE lowercase

! **************************************************************************************************
!> \brief   Convert all lower case characters in a string to upper case.
!> \param string ...
!> \date    22.06.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE uppercase(string)
      CHARACTER(LEN=*), INTENT(INOUT)                    :: string

      INTEGER                                            :: i, iascii

      DO i = 1, LEN_TRIM(string)
         iascii = ICHAR(string(i:i))
         IF ((iascii >= 97) .AND. (iascii <= 122)) THEN
            string(i:i) = CHAR(iascii-32)
         END IF
      END DO

   END SUBROUTINE uppercase

   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param string ...
!> \param ia ...
!> \param ib ...
! **************************************************************************************************
   SUBROUTINE xstring(string, ia, ib)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      INTEGER, INTENT(OUT)                               :: ia, ib

      ia = 1
      ib = LEN_TRIM(string)
      IF (ib > 0) THEN
         DO WHILE (string(ia:ia) == ' ')
            ia = ia+1
         END DO
      END IF

   END SUBROUTINE xstring

   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param str1 ...
!> \param str2 ...
!> \retval equal ...
! **************************************************************************************************
   FUNCTION str_comp(str1, str2) RESULT(equal)

      CHARACTER(LEN=*), INTENT(IN)                       :: str1, str2
      LOGICAL                                            :: equal

      INTEGER                                            :: i1, i2, j1, j2

      i1 = 0
      i2 = 0
      j1 = 0
      j2 = 0
      CALL xstring(str1, i1, i2)
      CALL xstring(str2, j1, j2)
      equal = (str1(i1:i2) == str2(j1:j2))
   END FUNCTION str_comp

! **************************************************************************************************
!> \brief returns .true. if the character passed is a whitespace char.
!> \param testchar ...
!> \retval resval ...
!> \par History
!>      02.2008 created, AK
! **************************************************************************************************
   FUNCTION is_whitespace(testchar) RESULT(resval)
      CHARACTER(LEN=1), INTENT(IN)                       :: testchar
      LOGICAL                                            :: resval

      resval = .FALSE.
      IF (ANY(default_blank_character == testchar)) resval = .TRUE.
   END FUNCTION is_whitespace

END MODULE string_utilities
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief type to store parallelization informations (at the moment assumes 1d
!>      position and uses mpi).
!>      type def in this separate module to avoid circular ref.
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_para_types

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_para_types'

   PUBLIC :: cp_para_env_type, cp_para_env_p_type, cp_para_cart_type

! **************************************************************************************************
!> \brief stores all the informations relevant to an mpi environement
!> \param mepos rank of the actual processor
!> \param group id of the actual mpi communicator (I know that it is confusing,
!>        but this is not the mpi group)
!> \param num_pe number of processors in the communicator
!> \param source rank of a special processor (for example the one for i-o,
!>        or the master
!> \param owns_group if it owns the group (and thus should free it when
!>        this object is deallocated)
!> \param ref_count the reference count, when it is zero this object gets
!>        deallocated
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_para_env_type
      LOGICAL :: owns_group, ionode
      INTEGER :: mepos, source, num_pe, ref_count
      INTEGER :: group
   END TYPE cp_para_env_type

! **************************************************************************************************
!> \brief represent a pointer to a para env (to build arrays)
!> \param para_env the pointer to the para_env
!> \par History
!>      07.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_para_env_p_type
      TYPE(cp_para_env_type), POINTER :: para_env
   END TYPE cp_para_env_p_type

! **************************************************************************************************
!> \brief represent a multidimensional parallel environment
!> \param mepos the position of the actual processor
!> \param group id of the actual group (context, communicator)
!> \param num_pe number of processors in the group in each dimension
!> \param source id of a special processor (for example the one for i-o,
!>        or the master
!> \param owns_group if it owns the group (and thus should free it when
!>        this object is deallocated)
!> \param ref_count the reference count, when it is zero this object gets
!>        deallocated
!> \note
!>      not yet implemented for mpi
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_para_cart_type
      LOGICAL :: owns_group
      INTEGER :: ndims, rank, ntask
      INTEGER, DIMENSION(:), POINTER :: mepos, source, num_pe
      LOGICAL, DIMENSION(:), POINTER :: periodic
      INTEGER :: group, ref_count
   END TYPE cp_para_cart_type

END MODULE cp_para_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of routines to handle the iteration info
! **************************************************************************************************
MODULE cp_iter_types
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! iteration_info
   PUBLIC :: cp_iteration_info_type, &
             cp_iteration_info_create, &
             cp_iteration_info_retain, &
             cp_iteration_info_release, &
             cp_iteration_info_copy_iter

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_iter_types'
   LOGICAL, PRIVATE, PARAMETER          :: debug_this_module = .FALSE.
   INTEGER, SAVE, PRIVATE               :: last_it_info_id = 0

   ! When adding a new iteration level PLEASE update the following list with the proper name!
   CHARACTER(LEN=default_path_length), PARAMETER, PUBLIC, DIMENSION(16) :: each_possible_labels = (/ &
                                                                           "__ROOT__          ", &
                                                                           "JUST_ENERGY       ", &
                                                                           "POWELL_OPT        ", &
                                                                           "QS_SCF            ", &
                                                                           "XAS_SCF           ", &
                                                                           "MD                ", &
                                                                           "METADYNAMICS      ", &
                                                                           "GEO_OPT           ", &
                                                                           "ROT_OPT           ", &
                                                                           "CELL_OPT          ", &
                                                                           "BAND              ", &
                                                                           "EP_LIN_SOLVER     ", &
                                                                           "SPLINE_FIND_COEFFS", &
                                                                           "REPLICA_EVAL      ", &
                                                                           "BSSE              ", &
                                                                           "SHELL_OPT         "/)

   CHARACTER(LEN=default_path_length), PARAMETER, PUBLIC, DIMENSION(16) ::  each_desc_labels = (/ &
                                                "Iteration level for __ROOT__ (fictitious iteration level)                      ", &
                                                "Iteration level for an ENERGY/ENERGY_FORCE calculation.                        ", &
                                                "Iteration level for POWELL based optimization steps.                           ", &
                                                "Iteration level for the SCF steps.                                             ", &
                                                "Iteration level for the X-Ray Absorption Spectroscopy (XAS) SCF steps.         ", &
                                                "Iteration level for the MD steps.                                              ", &
                                                "Iteration level for the METADYNAMICS steps (number of hills added).            ", &
                                                "Iteration level for the Geometry optimization steps.                           ", &
                                                "Iteration level for the Rotational optimization steps in the Dimer calculation.", &
                                                "Iteration level for the Cell optimization steps.                               ", &
                                                "Iteration level for the BAND calculation steps                                 ", &
                                                "Iteration level for the Energy Perturbation (EP) linear solver                 ", &
                                                "Iteration level for the solution of the coefficients of the splines            ", &
                                                "Iteration level for the evaluation of the Replica Environment                  ", &
                                                "Iteration level for the Basis Set Superposition Error (BSSE) calculation       ", &
                                                 "Iteration level for the Shell-Core distances optimization steps                "/)

! **************************************************************************************************
!> \brief contains the information about the current state of the program
!>      to be able to decide if output is necessary
!> \author fawzi
! **************************************************************************************************
   TYPE cp_iteration_info_type
      INTEGER                              :: ref_count, id_nr
      INTEGER                              :: print_level, n_rlevel
      INTEGER, DIMENSION(:), POINTER       :: iteration
      LOGICAL, DIMENSION(:), POINTER       :: last_iter
      CHARACTER(len=default_string_length) :: project_name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER           :: level_name
   END TYPE cp_iteration_info_type

CONTAINS

! **************************************************************************************************
!> \brief creates an output info object
!> \param iteration_info the object to create
!> \param project_name name of the project, used to create the filenames
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_iteration_info_create(iteration_info, project_name)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      CHARACTER(len=*), INTENT(in)                       :: project_name

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iteration_info_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: stat

      ALLOCATE (iteration_info, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/cp_iter_types.F",99,routineP//" could not allocate iteration_info")

      last_it_info_id = last_it_info_id+1
      iteration_info%id_nr = last_it_info_id
      iteration_info%ref_count = 1
      iteration_info%print_level = 2
      iteration_info%n_rlevel = 1
      iteration_info%project_name = project_name
      NULLIFY (iteration_info%iteration)
      NULLIFY (iteration_info%level_name)
      NULLIFY (iteration_info%last_iter)
      ALLOCATE (iteration_info%iteration(iteration_info%n_rlevel), stat=stat)
      IF (stat /= 0) THEN
         CALL cp__b("common/cp_iter_types.F",112,routineP//" iteration_info%iteration allocation")
      END IF
      ALLOCATE (iteration_info%level_name(iteration_info%n_rlevel), stat=stat)
      IF (stat /= 0) THEN
         CALL cp__b("common/cp_iter_types.F",116,routineP//" iteration_info%level_name allocation")
      END IF
      ALLOCATE (iteration_info%last_iter(iteration_info%n_rlevel), stat=stat)
      IF (stat /= 0) THEN
         CALL cp__b("common/cp_iter_types.F",120,routineP//" iteration_info%last_iter allocation")
      END IF
      iteration_info%iteration(iteration_info%n_rlevel) = 1
      iteration_info%level_name(iteration_info%n_rlevel) = "__ROOT__"
      iteration_info%last_iter(iteration_info%n_rlevel) = .FALSE.

   END SUBROUTINE cp_iteration_info_create

! **************************************************************************************************
!> \brief retains the iteration_info (see doc/ReferenceCounting.html)
!> \param iteration_info the iteration_info to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_iteration_info_retain(iteration_info)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iteration_info_retain', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(iteration_info)) THEN
         CALL cp__b("common/cp_iter_types.F",140,routineP//" iteration_info not associated")
      END IF
      IF (iteration_info%ref_count <= 0) THEN
         CALL cp__b("common/cp_iter_types.F",143,routineP//" iteration_info%ref_counf<=0")
      END IF
      iteration_info%ref_count = iteration_info%ref_count+1
   END SUBROUTINE cp_iteration_info_retain

! **************************************************************************************************
!> \brief releases the iteration_info (see doc/ReferenceCounting.html)
!> \param iteration_info the iteration_info to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_iteration_info_release(iteration_info)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iteration_info_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iteration_info)) THEN
         IF (iteration_info%ref_count <= 0) THEN
            CALL cp__b("common/cp_iter_types.F",161,routineP//" iteration_info%ref_counf<=0")
         END IF
         iteration_info%ref_count = iteration_info%ref_count-1
         IF (iteration_info%ref_count == 0) THEN
            IF (ASSOCIATED(iteration_info%iteration)) THEN
               DEALLOCATE (iteration_info%iteration)
            END IF
            IF (ASSOCIATED(iteration_info%last_iter)) THEN
               DEALLOCATE (iteration_info%last_iter)
            END IF
            IF (ASSOCIATED(iteration_info%level_name)) THEN
               DEALLOCATE (iteration_info%level_name)
            END IF
            DEALLOCATE (iteration_info)
         END IF
      END IF
   END SUBROUTINE cp_iteration_info_release

! **************************************************************************************************
!> \brief Copies iterations info of an iteration info into another iteration info
!> \param iteration_info_in the iteration_info to be copied
!> \param iteration_info_out the iteration_info results of the copy
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE cp_iteration_info_copy_iter(iteration_info_in, iteration_info_out)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info_in, iteration_info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iteration_info_copy_iter', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, stat

      IF (ASSOCIATED(iteration_info_in)) THEN
         IF (iteration_info_in%ref_count <= 0) THEN
            CALL cp__b("common/cp_iter_types.F",195,routineP//" iteration_info_in%ref_counf<=0")
         END IF

         iteration_info_out%n_rlevel = iteration_info_in%n_rlevel

         DEALLOCATE (iteration_info_out%iteration)
         i = SIZE(iteration_info_in%iteration)
         ALLOCATE (iteration_info_out%iteration(i), stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/cp_iter_types.F",204,routineP//" could not allocate iteration_info%iteration")
         iteration_info_out%iteration = iteration_info_in%iteration

         DEALLOCATE (iteration_info_out%last_iter)
         i = SIZE(iteration_info_in%last_iter)
         ALLOCATE (iteration_info_out%last_iter(i), stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/cp_iter_types.F",211,routineP//" could not allocate iteration_info%last_iter")
         iteration_info_out%last_iter = iteration_info_in%last_iter

         DEALLOCATE (iteration_info_out%level_name)
         i = SIZE(iteration_info_in%level_name)
         ALLOCATE (iteration_info_out%level_name(i), stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/cp_iter_types.F",218,routineP//" could not allocate iteration_info%level_name")
         iteration_info_out%level_name = iteration_info_in%level_name

      ELSE
         CALL cp__b("common/cp_iter_types.F",222,routineP//" iteration_info_in not associated!")
      END IF
   END SUBROUTINE cp_iteration_info_copy_iter

END MODULE cp_iter_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines to open and close files. Tracking of preconnections.
!> \par History
!>      - Creation CP2K_WORKSHOP 1.0 TEAM
!>      - Revised (18.02.2011,MK)
!>      - Enhanced error checking (22.02.2011,MK)
!> \author Matthias Krack (MK)
! **************************************************************************************************
MODULE cp_files

   USE kinds,                           ONLY: default_path_length
   USE machine,                         ONLY: default_input_unit,&
                                              default_output_unit,&
                                              m_getcwd
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: close_file, &
             init_preconnection_list, &
             open_file, &
             get_unit_number, &
             file_exists, &
             get_data_dir

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_files'

   INTEGER, PARAMETER :: max_preconnections = 10, &
                         max_unit_number = 999

   TYPE preconnection_type
      PRIVATE
      CHARACTER(LEN=default_path_length) :: file_name
      INTEGER                            :: unit_number
   END TYPE preconnection_type

   TYPE(preconnection_type), DIMENSION(max_preconnections) :: preconnected

CONTAINS

! **************************************************************************************************
!> \brief Add an entry to the list of preconnected units
!> \param file_name ...
!> \param unit_number ...
!> \par History
!>      - Creation (22.02.2011,MK)
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE assign_preconnection(file_name, unit_number)

      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      INTEGER, INTENT(IN)                                :: unit_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'assign_preconnection', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, islot, nc

      IF ((unit_number < 1) .OR. (unit_number > max_unit_number)) THEN
         CALL cp__b("common/cp_files.F",67,"An invalid logical unit number was specified.")
      END IF

      IF (LEN_TRIM(file_name) == 0) THEN
         CALL cp__b("common/cp_files.F",71,"No valid file name was specified")
      END IF

      nc = SIZE(preconnected)

      ! Check if a preconnection already exists
      DO ic = 1, nc
         IF (TRIM(preconnected(ic)%file_name) == TRIM(file_name)) THEN
            ! Return if the entry already exists
            IF (preconnected(ic)%unit_number == unit_number) THEN
               RETURN
            ELSE
               CALL print_preconnection_list()
               CALL cp_abort(cp__l("common/cp_files.F",84), &
                             "Attempt to connect the already connected file <"// &
                             TRIM(ADJUSTL(file_name))//"> to another unit")
            END IF
         END IF
      END DO

      ! Seach for an unused entry
      islot = -1
      DO ic = 1, nc
         IF (preconnected(ic)%unit_number == -1) THEN
            islot = ic
            EXIT
         END IF
      END DO

      IF (islot == -1) THEN
         CALL print_preconnection_list()
         CALL cp__b("common/cp_files.F",102,"No free slot found in the list of preconnected units")
      END IF

      preconnected(islot)%file_name = TRIM(ADJUSTL(file_name))
      preconnected(islot)%unit_number = unit_number

   END SUBROUTINE assign_preconnection

! **************************************************************************************************
!> \brief Close an open file given by its logical unit number.
!>        Optionally, keep the file and unit preconnected.
!> \param unit_number ...
!> \param file_status ...
!> \param keep_preconnection ...
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE close_file(unit_number, file_status, keep_preconnection)

      INTEGER, INTENT(IN)                                :: unit_number
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: file_status
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_preconnection

      CHARACTER(LEN=*), PARAMETER :: routineN = 'close_file', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2*default_path_length)               :: message
      CHARACTER(LEN=6)                                   :: status_string
      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: istat
      LOGICAL                                            :: exists, is_open, keep_file_connection

      keep_file_connection = .FALSE.
      IF (PRESENT(keep_preconnection)) keep_file_connection = keep_preconnection

      INQUIRE (UNIT=unit_number, EXIST=exists, OPENED=is_open, IOSTAT=istat)

      IF (istat /= 0) THEN
         WRITE (UNIT=message, FMT="(A,I0,A,I0,A)") &
            "An error occurred inquiring the unit with the number ", unit_number, &
            " (IOSTAT = ", istat, ")"
         CALL cp__b("common/cp_files.F",141,TRIM(message))
      ELSE IF (.NOT. exists) THEN
         WRITE (UNIT=message, FMT="(A,I0,A)") &
            "The specified unit number ", unit_number, &
            " cannot be closed, because it does not exist."
         CALL cp__b("common/cp_files.F",146,TRIM(message))
      END IF

      ! Close the specified file

      IF (is_open) THEN
         ! Refuse to close any preconnected system unit
         IF (unit_number == default_input_unit) THEN
            WRITE (UNIT=message, FMT="(A,I0)") &
               "Attempt to close the default input unit number ", unit_number
            CALL cp__b("common/cp_files.F",156,TRIM(message))
         END IF
         IF (unit_number == default_output_unit) THEN
            WRITE (UNIT=message, FMT="(A,I0)") &
               "Attempt to close the default output unit number ", unit_number
            CALL cp__b("common/cp_files.F",161,TRIM(message))
         END IF
         ! Define status after closing the file
         IF (PRESENT(file_status)) THEN
            status_string = TRIM(ADJUSTL(file_status))
         ELSE
            status_string = "KEEP"
         END IF
         ! Optionally, keep this unit preconnected
         INQUIRE (UNIT=unit_number, NAME=file_name, IOSTAT=istat)
         IF (istat /= 0) THEN
            WRITE (UNIT=message, FMT="(A,I0,A,I0,A)") &
               "An error occurred inquiring the unit with the number ", unit_number, &
               " (IOSTAT = ", istat, ")"
            CALL cp__b("common/cp_files.F",175,TRIM(message))
         END IF
         ! Manage preconnections
         IF (keep_file_connection) THEN
            CALL assign_preconnection(file_name, unit_number)
         ELSE
            CALL delete_preconnection(file_name, unit_number)
            CLOSE (UNIT=unit_number, IOSTAT=istat, STATUS=TRIM(status_string))
            IF (istat /= 0) THEN
               WRITE (UNIT=message, FMT="(A,I0,A,I0,A)") &
                  "An error occurred closing the file with the logical unit number ", &
                  unit_number, " (IOSTAT = ", istat, ")"
               CALL cp__b("common/cp_files.F",187,TRIM(message))
            END IF
         END IF
      END IF

   END SUBROUTINE close_file

! **************************************************************************************************
!> \brief Remove an entry from the list of preconnected units
!> \param file_name ...
!> \param unit_number ...
!> \par History
!>      - Creation (22.02.2011,MK)
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE delete_preconnection(file_name, unit_number)

      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      INTEGER                                            :: unit_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'delete_preconnection', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, nc

      nc = SIZE(preconnected)

      ! Seach for preconnection entry and delete it when found
      DO ic = 1, nc
         IF (TRIM(preconnected(ic)%file_name) == TRIM(file_name)) THEN
            IF (preconnected(ic)%unit_number == unit_number) THEN
               preconnected(ic)%file_name = ""
               preconnected(ic)%unit_number = -1
               EXIT
            ELSE
               CALL print_preconnection_list()
               CALL cp_abort(cp__l("common/cp_files.F",223), &
                             "Attempt to disconnect the file <"// &
                             TRIM(ADJUSTL(file_name))// &
                             "> from an unlisted unit")
            END IF
         END IF
      END DO

   END SUBROUTINE delete_preconnection

! **************************************************************************************************
!> \brief Returns the first logical unit that is not preconnected
!> \param file_name ...
!> \retval unit_number ...
!> \author Matthias Krack (MK)
!> \note
!>       -1 if no free unit exists
! **************************************************************************************************
   FUNCTION get_unit_number(file_name) RESULT(unit_number)

      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: file_name
      INTEGER                                            :: unit_number

      INTEGER                                            :: ic, istat, nc
      LOGICAL                                            :: exists, is_open

      IF (PRESENT(file_name)) THEN
         nc = SIZE(preconnected)
         ! Check for preconnected units
         DO ic = 3, nc ! Exclude the preconnected system units (< 3)
            IF (TRIM(preconnected(ic)%file_name) == TRIM(file_name)) THEN
               unit_number = preconnected(ic)%unit_number
               RETURN
            END IF
         END DO
      END IF

      ! Get a new unit number
      DO unit_number = 1, max_unit_number
         IF (ANY(unit_number == preconnected(:)%unit_number)) CYCLE
         INQUIRE (UNIT=unit_number, EXIST=exists, OPENED=is_open, IOSTAT=istat)
         IF (exists .AND. (.NOT. is_open) .AND. (istat == 0)) RETURN
      END DO

      unit_number = -1

   END FUNCTION get_unit_number

! **************************************************************************************************
!> \brief Allocate and initialise the list of preconnected units
!> \par History
!>      - Creation (22.02.2011,MK)
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE init_preconnection_list()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_preconnection_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, nc

      nc = SIZE(preconnected)

      DO ic = 1, nc
         preconnected(ic)%file_name = ""
         preconnected(ic)%unit_number = -1
      END DO

      ! Define reserved unit numbers
      preconnected(1)%file_name = "stdin"
      preconnected(1)%unit_number = default_input_unit
      preconnected(2)%file_name = "stdout"
      preconnected(2)%unit_number = default_output_unit

   END SUBROUTINE init_preconnection_list

! **************************************************************************************************
!> \brief Opens the requested file using a free unit number
!> \param file_name ...
!> \param file_status ...
!> \param file_form ...
!> \param file_action ...
!> \param file_position ...
!> \param file_pad ...
!> \param unit_number ...
!> \param debug ...
!> \param skip_get_unit_number ...
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE open_file(file_name, file_status, file_form, file_action, &
                        file_position, file_pad, unit_number, debug, &
                        skip_get_unit_number)

      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: file_status, file_form, file_action, &
                                                            file_position, file_pad
      INTEGER, INTENT(INOUT)                             :: unit_number
      INTEGER, INTENT(IN), OPTIONAL                      :: debug
      LOGICAL, INTENT(IN), OPTIONAL                      :: skip_get_unit_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'open_file', routineP = moduleN//':'//routineN

      CHARACTER(LEN=11) :: access_string, action_string, current_action, current_form, &
         form_string, pad_string, position_string, status_string
      CHARACTER(LEN=2*default_path_length)               :: message
      CHARACTER(LEN=default_path_length)                 :: cwd, iomsgstr, real_file_name
      INTEGER                                            :: debug_unit, istat
      LOGICAL                                            :: exists, get_a_new_unit, is_open

      access_string = "SEQUENTIAL"

      IF (PRESENT(file_status)) THEN
         status_string = TRIM(file_status)
      ELSE
         status_string = "OLD"
      END IF

      IF (PRESENT(file_form)) THEN
         form_string = TRIM(file_form)
      ELSE
         form_string = "FORMATTED"
      END IF

      IF (PRESENT(file_pad)) THEN
         pad_string = file_pad
         IF (form_string == "UNFORMATTED") THEN
            WRITE (UNIT=message, FMT="(A)") &
               "The PAD specifier is not allowed for an UNFORMATTED file"
            CALL cp__b("common/cp_files.F",351,TRIM(message))
         END IF
      ELSE
         pad_string = "YES"
      END IF

      IF (PRESENT(file_action)) THEN
         action_string = TRIM(file_action)
      ELSE
         action_string = "READ"
      END IF

      IF (PRESENT(file_position)) THEN
         position_string = TRIM(file_position)
      ELSE
         position_string = "REWIND"
      END IF

      IF (PRESENT(debug)) THEN
         debug_unit = debug
      ELSE
         debug_unit = 0 ! use default_output_unit for debugging
      END IF

      real_file_name = file_name
      IF (status_string == "OLD") &
         real_file_name = discover_file(file_name)

      ! Check the specified input file name
      INQUIRE (FILE=TRIM(real_file_name), EXIST=exists, OPENED=is_open, IOSTAT=istat)

      IF (istat /= 0) THEN
         WRITE (UNIT=message, FMT="(A,I0,A)") &
            "An error occurred inquiring the file <"//TRIM(real_file_name)// &
            "> (IOSTAT = ", istat, ")"
         CALL cp__b("common/cp_files.F",386,TRIM(message))
      ELSE IF (status_string == "OLD") THEN
         IF (.NOT. exists) THEN
            WRITE (UNIT=message, FMT="(A)") &
               "The specified OLD file <"//TRIM(ADJUSTL(real_file_name))// &
               "> cannot be opened. It does not exist. "// &
               "Data directory path: "//TRIM(get_data_dir())
            CALL cp__b("common/cp_files.F",393,TRIM(message))
         END IF
      END IF

      ! Open the specified input file
      IF (is_open) THEN
         INQUIRE (FILE=TRIM(real_file_name), NUMBER=unit_number, &
                  ACTION=current_action, FORM=current_form)
         IF (TRIM(position_string) == "REWIND") REWIND (UNIT=unit_number)
         IF (TRIM(status_string) == "NEW") THEN
            CALL cp_abort(cp__l("common/cp_files.F",403), &
                          "Attempt to re-open the existing OLD file <"// &
                          TRIM(real_file_name)//"> with status attribute NEW.")
         END IF
         IF (TRIM(current_form) /= TRIM(form_string)) THEN
            CALL cp_abort(cp__l("common/cp_files.F",408), &
                          "Attempt to re-open the existing "// &
                          TRIM(current_form)//" file <"//TRIM(real_file_name)// &
                          "> as "//TRIM(form_string)//" file.")
         END IF
         IF (TRIM(current_action) /= TRIM(action_string)) THEN
            CALL cp_abort(cp__l("common/cp_files.F",414), &
                          "Attempt to re-open the existing file <"// &
                          TRIM(real_file_name)//"> with the modified ACTION attribute "// &
                          TRIM(action_string)//". The current ACTION attribute is "// &
                          TRIM(current_action)//".")
         END IF
      ELSE
         ! Find an unused unit number
         get_a_new_unit = .TRUE.
         IF (PRESENT(skip_get_unit_number)) THEN
            IF (skip_get_unit_number) get_a_new_unit = .FALSE.
         END IF
         IF (get_a_new_unit) unit_number = get_unit_number(TRIM(real_file_name))
         IF (unit_number < 1) THEN
            WRITE (UNIT=message, FMT="(A)") &
               "Cannot open the file <"//TRIM(ADJUSTL(real_file_name))// &
               ">, because no unused logical unit number could be obtained."
            CALL cp__b("common/cp_files.F",431,TRIM(message))
         END IF
         IF (TRIM(form_string) == "FORMATTED") THEN
            OPEN (UNIT=unit_number, &
                  FILE=TRIM(real_file_name), &
                  STATUS=TRIM(status_string), &
                  ACCESS=TRIM(access_string), &
                  FORM=TRIM(form_string), &
                  POSITION=TRIM(position_string), &
                  ACTION=TRIM(action_string), &
                  PAD=TRIM(pad_string), &
                  IOMSG=iomsgstr, &
                  IOSTAT=istat)
         ELSE
            OPEN (UNIT=unit_number, &
                  FILE=TRIM(real_file_name), &
                  STATUS=TRIM(status_string), &
                  ACCESS=TRIM(access_string), &
                  FORM=TRIM(form_string), &
                  POSITION=TRIM(position_string), &
                  ACTION=TRIM(action_string), &
                  IOMSG=iomsgstr, &
                  IOSTAT=istat)
         END IF
         IF (istat /= 0) THEN
            CALL m_getcwd(cwd)
            WRITE (UNIT=message, FMT="(A,I0,A,I0,A)") &
               "An error occurred opening the file '"//TRIM(ADJUSTL(real_file_name))// &
               "' (UNIT = ", unit_number, ", IOSTAT = ", istat, "). "//TRIM(iomsgstr)//". "// &
               "Current working directory: "//TRIM(cwd)

            CALL cp__b("common/cp_files.F",462,TRIM(message))
         END IF
      END IF

      IF (debug_unit > 0) THEN
         INQUIRE (FILE=TRIM(real_file_name), OPENED=is_open, NUMBER=unit_number, &
                  POSITION=position_string, NAME=message, ACCESS=access_string, &
                  FORM=form_string, ACTION=action_string)
         WRITE (UNIT=debug_unit, FMT="(T2,A)") "BEGIN DEBUG "//TRIM(ADJUSTL(routineN))
         WRITE (UNIT=debug_unit, FMT="(T3,A,I0)") "NUMBER  : ", unit_number
         WRITE (UNIT=debug_unit, FMT="(T3,A,L1)") "OPENED  : ", is_open
         WRITE (UNIT=debug_unit, FMT="(T3,A)") "NAME    : "//TRIM(ADJUSTL(message))
         WRITE (UNIT=debug_unit, FMT="(T3,A)") "POSITION: "//TRIM(ADJUSTL(position_string))
         WRITE (UNIT=debug_unit, FMT="(T3,A)") "ACCESS  : "//TRIM(ADJUSTL(access_string))
         WRITE (UNIT=debug_unit, FMT="(T3,A)") "FORM    : "//TRIM(ADJUSTL(form_string))
         WRITE (UNIT=debug_unit, FMT="(T3,A)") "ACTION  : "//TRIM(ADJUSTL(action_string))
         WRITE (UNIT=debug_unit, FMT="(T2,A)") "END DEBUG "//TRIM(ADJUSTL(routineN))
         CALL print_preconnection_list(debug_unit)
      END IF

   END SUBROUTINE open_file

! **************************************************************************************************
!> \brief Checks if file exists, considering also the file discovery mechanism.
!> \param file_name ...
!> \retval exist
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION file_exists(file_name) RESULT(exist)
      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      LOGICAL                                            :: exist

      CHARACTER(LEN=default_path_length)                 :: real_file_name

      real_file_name = discover_file(file_name)
      INQUIRE (FILE=TRIM(real_file_name), exist=exist)
   END FUNCTION file_exists

! **************************************************************************************************
!> \brief Checks various locations for a file name.
!> \param file_name ...
!> \retval real_file_name ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION discover_file(file_name) RESULT(real_file_name)
      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      CHARACTER(LEN=default_path_length)                 :: real_file_name

      CHARACTER(LEN=default_path_length)                 :: candidate, data_dir
      INTEGER                                            :: stat
      LOGICAL                                            :: exists

      real_file_name = TRIM(file_name)

      ! first try file-name directly
      INQUIRE (file=TRIM(file_name), exist=exists, iostat=stat)
      IF (stat == 0 .AND. exists) THEN
         real_file_name = file_name
         RETURN
      ENDIF

      ! then try the data-dir
      data_dir = get_data_dir()
      IF (LEN_TRIM(data_dir) > 0) THEN
         candidate = join_paths(data_dir, file_name)
         INQUIRE (file=TRIM(candidate), exist=exists, iostat=stat)
         IF (stat == 0 .AND. exists) THEN
            real_file_name = candidate
            RETURN
         ENDIF
      ENDIF

      ! fall back in case file not found
      real_file_name = file_name

   END FUNCTION discover_file

! **************************************************************************************************
!> \brief Returns path of data directory if set, otherwise an empty string
!> \retval data_dir_path ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION get_data_dir() RESULT(data_dir_path)
      CHARACTER(LEN=default_path_length)                 :: data_dir_path

      INTEGER                                            :: stat

      CALL GET_ENVIRONMENT_VARIABLE("CP2K_DATA_DIR", data_dir_path, status=stat)
      IF (stat == 0) RETURN

      data_dir_path = "/data/vjoost/clean/cp2k/cp2k/data"

   END FUNCTION get_data_dir

! **************************************************************************************************
!> \brief Joins two file-paths, inserting '/' as needed.
!> \param path1 ...
!> \param path2 ...
!> \retval joined_path ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION join_paths(path1, path2) RESULT(joined_path)
      CHARACTER(LEN=*), INTENT(IN)                       :: path1, path2
      CHARACTER(LEN=default_path_length)                 :: joined_path

      INTEGER                                            :: n

      n = LEN_TRIM(path1)
      IF (path2(1:1) == '/') THEN
         joined_path = path2
      ELSE IF (n == 0 .OR. path1(n:n) == '/') THEN
         joined_path = TRIM(path1)//path2
      ELSE
         joined_path = TRIM(path1)//'/'//path2
      ENDIF
   END FUNCTION join_paths

! **************************************************************************************************
!> \brief Print the list of preconnected units
!> \param output_unit which unit to print to (optional)
!> \par History
!>      - Creation (22.02.2011,MK)
!> \author Matthias Krack (MK)
! **************************************************************************************************
   SUBROUTINE print_preconnection_list(output_unit)
      INTEGER, INTENT(IN), OPTIONAL                      :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_preconnection_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, nc, unit

      IF (PRESENT(output_unit)) THEN
         unit = output_unit
      ELSE
         unit = default_output_unit
      ENDIF

      nc = SIZE(preconnected)

      IF (output_unit > 0) THEN

         WRITE (UNIT=output_unit, FMT="(A,/,A)") &
            " LIST OF PRECONNECTED LOGICAL UNITS", &
            "  Slot   Unit number   File name"
         DO ic = 1, nc
            IF (preconnected(ic)%unit_number > 0) THEN
               WRITE (UNIT=output_unit, FMT="(I6,3X,I6,8X,A)") &
                  ic, preconnected(ic)%unit_number, &
                  TRIM(ADJUSTL(preconnected(ic)%file_name))
            ELSE
               WRITE (UNIT=output_unit, FMT="(I6,17X,A)") &
                  ic, "UNUSED"
            END IF
         END DO
      END IF
   END SUBROUTINE print_preconnection_list

END MODULE cp_files
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Types used by timings.F and timings_report.F
!>        The types in this module are used within dict or list, which are
!>          in turn used in timer_env_type from timings_type.F
!>        Due to the fortran restriction on cicular module-dependencies these
!>        "inner-types" had to go into a separate module.
!> \par History
!>      12.2012 Created [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE timings_base_type

   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8

   IMPLICIT NONE
   PRIVATE

   TYPE routine_stat_type
      INTEGER       :: routine_id
      CHARACTER(len=default_string_length) :: routineN
      REAL(kind=dp) :: excl_walltime_accu
      REAL(kind=dp) :: incl_walltime_accu
      REAL(kind=dp) :: excl_energy_accu
      REAL(kind=dp) :: incl_energy_accu
      INTEGER       :: active_calls
      INTEGER       :: total_calls
      INTEGER       :: stackdepth_accu
      LOGICAL       :: trace
   END TYPE routine_stat_type

   TYPE call_stat_type
      INTEGER       :: total_calls
      REAL(kind=dp) :: incl_walltime_accu
      REAL(kind=dp) :: incl_energy_accu
   END TYPE call_stat_type

   TYPE callstack_entry_type
      INTEGER       :: routine_id
      REAL(kind=dp) :: walltime_start
      REAL(kind=dp) :: energy_start
   END TYPE callstack_entry_type

   TYPE routine_report_type
      CHARACTER(LEN=default_string_length) :: routineN
      REAL(KIND=dp)                        :: max_icost = 0.0_dp
      REAL(KIND=dp)                        :: sum_icost = 0.0_dp
      REAL(KIND=dp)                        :: max_ecost = 0.0_dp
      REAL(KIND=dp)                        :: sum_ecost = 0.0_dp
      INTEGER(kind=int_8)                  :: max_total_calls = 0
      INTEGER(kind=int_8)                  :: sum_total_calls = 0
      INTEGER(kind=int_8)                  :: sum_stackdepth = 0
   END TYPE routine_report_type

   PUBLIC :: routine_stat_type, call_stat_type, callstack_entry_type, routine_report_type

END MODULE timings_base_type

! **************************************************************************************************

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief An array-based list which grows on demand.
!>        When the internal array is full, a new array of twice the size will be
!>        allocated and the items are copied over.
!>
!>        This list can also be used as a stack.
!>        Have look at list_push(), list_pop() and list_peek().
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify list__valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE list_routinestat

   USE timings_base_type,               ONLY: routine_stat_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'list_routinestat'

   !this is an internal type
   TYPE private_item_type
      PRIVATE
      TYPE(routine_stat_type), POINTER                   :: value
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a list-instance
   TYPE list_routinestat_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: arr => Null()
      INTEGER                                       :: size = -1
   END TYPE list_routinestat_type

   PUBLIC :: list_routinestat_type
   PUBLIC :: list_routinestat_init
   PUBLIC :: list_routinestat_push
   PUBLIC :: list_routinestat_pop
   PUBLIC :: list_routinestat_peek
   PUBLIC :: list_routinestat_insert
   PUBLIC :: list_routinestat_set
   PUBLIC :: list_routinestat_get
   PUBLIC :: list_routinestat_del
   PUBLIC :: list_routinestat_clear
   PUBLIC :: list_routinestat_size
   PUBLIC :: list_routinestat_destroy
   PUBLIC :: list_routinestat_isready

CONTAINS

! **************************************************************************************************
!> \brief Test if the given list has been initialized.
!> \param list ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinestat_isready(list) RESULT(res)
      TYPE(list_routinestat_type), INTENT(in)            :: list
      LOGICAL                                            :: res

      res = ASSOCIATED(list%arr)
   END FUNCTION list_routinestat_isready

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given list.
!>        This has to be called before any of the other routines.
!>        For deallocation call list_routinestat_destroy.
!> \param list ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_init(list, initial_capacity)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_, stat

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 0) &
         CALL cp__b("common/list_routinestat.F",104,"list_routinestat_create: initial_capacity < 0")

      IF (ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",107,"list_routinestat_create: list is already initialized.")

      ALLOCATE (list%arr(initial_capacity_), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinestat.F",111,"list_routinestat_init: allocation failed")

      list%size = 0
   END SUBROUTINE list_routinestat_init

! **************************************************************************************************
!> \brief Deallocated the internal data-structures of the given list.
!>        Caution: If the stored values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_destroy(list)
      TYPE(list_routinestat_type), INTENT(inout)         :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",131,"list_routinestat_destroy: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      DEALLOCATE (list%arr)
      list%size = -1
   END SUBROUTINE list_routinestat_destroy

! **************************************************************************************************
!> \brief Assings the given value to the given position in the list.
!>        Thereby, the former value at that position gets overwritten.
!>        If the position is out of bounds, the program stops.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1.
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_set(list, value, pos)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      TYPE(routine_stat_type), INTENT(in), POINTER       :: value
      INTEGER, INTENT(in)                                :: pos

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",157,"list_routinestat_set: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinestat.F",159,"list_routinestat_set: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinestat.F",161,"list_routinestat_set: pos > size")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_routinestat_set

! **************************************************************************************************
!> \brief Appends the given value at the end of the list.
!> \param list ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_push(list, value)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      TYPE(routine_stat_type), INTENT(in), POINTER       :: value

      INTEGER                                            :: stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",180,"list_routinestat_push: list is not initialized.")
      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      ALLOCATE (list%arr(list%size)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinestat.F",187,"list_routinestat_push: allocation failed")
      list%arr(list%size)%p%value => value
   END SUBROUTINE list_routinestat_push

! **************************************************************************************************
!> \brief Inserts the given value at the givenn position within the list.
!>        Values which lay behind the insertion-position move one position up.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+2 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_insert(list, value, pos)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      TYPE(routine_stat_type), INTENT(in), POINTER       :: value
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i, stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",209,"list_routinestat_insert: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinestat.F",211,"list_routinestat_insert: pos < 1")
      IF (pos > list%size+1) &
         CALL cp__b("common/list_routinestat.F",213,"list_routinestat_insert: pos > size+1")

      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      DO i = list%size, pos+1, -1
         list%arr(i)%p => list%arr(i-1)%p
      END DO

      ALLOCATE (list%arr(pos)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinestat.F",225,"list_routinestat_insert: allocation failed.")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_routinestat_insert

! **************************************************************************************************
!> \brief Returns the last element in the list.
!>    Is equivalent to: list_routinestat_get(list, list_routinestat_size(list))
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinestat_peek(list) RESULT(value)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      TYPE(routine_stat_type), POINTER                   :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",243,"list_routinestat_peek: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_routinestat.F",245,"list_routinestat_peek: list is empty.")

      value => list%arr(list%size)%p%value
   END FUNCTION list_routinestat_peek

! **************************************************************************************************
!> \brief Returns the last element in the list and removes it.
!>        Is equivialent to:
!>        value = list_routinestat_get(list, list_routinestat_size(list))
!>            call list_routinestat_del(list, list_routinestat_size(list))
!>
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinestat_pop(list) RESULT(value)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      TYPE(routine_stat_type), POINTER                   :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",267,"list_routinestat_pop: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_routinestat.F",269,"list_routinestat_pop: list is empty.")

      value => list%arr(list%size)%p%value
      DEALLOCATE (list%arr(list%size)%p)
      list%size = list%size-1
   END FUNCTION list_routinestat_pop

! **************************************************************************************************
!> \brief Removes all values from the list. The list itself is not deallocated.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_clear(list)
      TYPE(list_routinestat_type), INTENT(inout)         :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",289,"list_routinestat_clear: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      list%size = 0
   END SUBROUTINE list_routinestat_clear

!
! **************************************************************************************************
!> \brief Returns the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinestat_get(list, pos) RESULT(value)
      TYPE(list_routinestat_type), INTENT(in)            :: list
      INTEGER, INTENT(in)                                :: pos
      TYPE(routine_stat_type), POINTER                   :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",313,"list_routinestat_get: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinestat.F",315,"list_routinestat_get: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinestat.F",317,"list_routinestat_get: pos > size")

      value => list%arr(pos)%p%value

   END FUNCTION list_routinestat_get

! **************************************************************************************************
!> \brief Removes the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinestat_del(list, pos)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",338,"list_routinestat_del: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinestat.F",340,"list_routinestat_det: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinestat.F",342,"list_routinestat_det: pos > size")

      DEALLOCATE (list%arr(pos)%p)
      DO i = pos, list%size-1
         list%arr(i)%p => list%arr(i+1)%p
      END DO

      list%size = list%size-1

   END SUBROUTINE list_routinestat_del

! **************************************************************************************************
!> \brief Returns the current size of the list.
!> \param list ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinestat_size(list) RESULT(size)
      TYPE(list_routinestat_type), INTENT(in)            :: list
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinestat.F",366,"list_routinestat_size: list is not initialized.")

      size = list%size
   END FUNCTION list_routinestat_size

! **************************************************************************************************
!> \brief Internal routine for changing the size of the internal array.
!> \param list ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE change_capacity(list, new_capacity)
      TYPE(list_routinestat_type), INTENT(inout)         :: list
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, stat
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_arr

      new_cap = new_capacity
      IF (new_cap < 0) &
         CALL cp__b("common/list_routinestat.F",388,"list_routinestat_change_capacity: new_capacity < 0")
      IF (new_cap < list%size) &
         CALL cp__b("common/list_routinestat.F",390,"list_routinestat_change_capacity: new_capacity < size")
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(list%arr) == HUGE(i)) &
            CALL cp__b("common/list_routinestat.F",393,"list_routinestat_change_capacity: list has reached integer limit.")
         new_cap = HUGE(i) ! grow as far as possible
      END IF

      old_arr => list%arr
      ALLOCATE (list%arr(new_cap), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinestat.F",400,"list_routinestat_change_capacity: allocation failed")

      DO i = 1, list%size
         ALLOCATE (list%arr(i)%p, stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/list_routinestat.F",405,"list_routinestat_change_capacity: allocation failed")
         list%arr(i)%p%value => old_arr(i)%p%value
         DEALLOCATE (old_arr(i)%p)
      END DO
      DEALLOCATE (old_arr)

   END SUBROUTINE change_capacity

END MODULE list_routinestat
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A dictionary (also known as hashtable or hashmap).
!>        Internally the dictionary uses an array to holds its data.
!>        If this array reaches a load-factor of 75%, a new array with twice the
!>        size will be allocated and the items are then copied over.
!>        This ensures that the dictionary will perform operations in O(1).
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify dict__keytype___valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! ***************************************************************************************************


MODULE dict_str_i4
   USE kinds,                           ONLY: default_string_length,&
                                              int_4,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   !this is an internal type
   !Calculating hashes might be expensive, therefore they are stored
   !for use during change_capacity().
   TYPE private_item_type
      PRIVATE
      CHARACTER(LEN=default_string_length)                            :: key
      INTEGER(kind=int_4)                          :: value
      INTEGER(KIND=int_8)                       :: hash
      TYPE(private_item_type), POINTER          :: next => Null()
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a dictionary-instance
   TYPE dict_str_i4_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER      :: buckets => Null()
      INTEGER                                               :: size = -1
   END TYPE dict_str_i4_type

   ! this is a public type, its returned by dict_items()
   TYPE dict_str_i4_item_type
      CHARACTER(LEN=default_string_length)        :: key
      INTEGER(kind=int_4)      :: value
   END TYPE dict_str_i4_item_type

   PUBLIC :: dict_str_i4_type
   PUBLIC :: dict_str_i4_item_type
   PUBLIC :: dict_str_i4_init
   PUBLIC :: dict_str_i4_isready
   PUBLIC :: dict_str_i4_set
   PUBLIC :: dict_str_i4_get
   PUBLIC :: dict_str_i4_del
   PUBLIC :: dict_str_i4_size
   PUBLIC :: dict_str_i4_haskey
   PUBLIC :: dict_str_i4_items
   PUBLIC :: dict_str_i4_update
   PUBLIC :: dict_str_i4_destroy

CONTAINS

! **************************************************************************************************
! This is joaat_hash from string_table.F
!
!> \brief generates the hash of a given string
!> \param key a string of any length
!> \retval hash ...
!> \par History
!>       09.2006 created [Joost VandeVondele]
!>       12.2012 copied and adopted [ole]
!> \note
!>       http://en.wikipedia.org/wiki/Hash_table
!>       http://www.burtleburtle.net/bob/hash/doobs.html
! **************************************************************************************************
   FUNCTION hash_str(key) RESULT(hash)
      CHARACTER(LEN=*), INTENT(IN)                       :: key
      INTEGER(KIND=int_8)                                :: hash

      INTEGER(KIND=int_8), PARAMETER                     :: b32 = 2_int_8**32-1_int_8

      INTEGER                                            :: i

      hash = 0_int_8
      DO i = 1, LEN(key)
         hash = IAND(hash+ICHAR(key(i:i)), b32)
         hash = IAND(hash+IAND(ISHFT(hash, 10), b32), b32)
         hash = IAND(IEOR(hash, IAND(ISHFT(hash, -6), b32)), b32)
      ENDDO
      hash = IAND(hash+IAND(ISHFT(hash, 3), b32), b32)
      hash = IAND(IEOR(hash, IAND(ISHFT(hash, -11), b32)), b32)
      hash = IAND(hash+IAND(ISHFT(hash, 15), b32), b32)
   END FUNCTION hash_str

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given dictionary.
!> \param dict ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_str_i4_init(dict, initial_capacity)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 1) &
         CALL cp__b("common/dict_str_i4.F",128,"dict_str_i4_init: initial_capacity < 1")

      IF (ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",131,"dict_str_i4_init: dictionary is already initialized.")

      ALLOCATE (dict%buckets(initial_capacity_))
      dict%size = 0

   END SUBROUTINE dict_str_i4_init

! **************************************************************************************************
!> \brief Test if the given dictionary has been initialized.
!> \param dict ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_str_i4_isready(dict) RESULT(res)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      LOGICAL                                            :: res

      res = ASSOCIATED(dict%buckets)
   END FUNCTION dict_str_i4_isready

! **************************************************************************************************
!> \brief Deallocated the internal data-structures if the given dictionary.
!>        Caution: If the stored keys or values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param dict ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_str_i4_destroy(dict)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict

      INTEGER                                            :: i
      TYPE(private_item_type), POINTER                   :: item, prev_item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",169,"dict_str_i4_destroy: dictionary is not initialized.")

      DO i = 1, SIZE(dict%buckets)
         item => dict%buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            prev_item => item
            item => item%next
            DEALLOCATE (prev_item)
         END DO
      END DO

      DEALLOCATE (dict%buckets)
      dict%size = -1

   END SUBROUTINE dict_str_i4_destroy

! **************************************************************************************************
!> \brief Stores, and possibly overwrites, a given value under a given key.
!> \param dict ...
!> \param key ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_str_i4_set(dict, key, value)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      CHARACTER(LEN=default_string_length), INTENT(in)   :: key
      INTEGER(kind=int_4), INTENT(in)                    :: value

      INTEGER(KIND=int_8)                                :: hash

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",202,"dict_str_i4_set: dictionary is not initialized.")

      hash = hash_str(key)
      CALL set_hashed(dict, key, value, hash)
   END SUBROUTINE dict_str_i4_set

! **************************************************************************************************
!> \brief Common code used internally by dict_set() and change_capacity().
!> \param dict ...
!> \param key ...
!> \param value ...
!> \param hash ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE set_hashed(dict, key, value, hash)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      CHARACTER(LEN=default_string_length), INTENT(in)   :: key
      INTEGER(kind=int_4), INTENT(in)                    :: value
      INTEGER(KIND=int_8), INTENT(in)                    :: hash

      INTEGER(KIND=int_8)                                :: idx
      TYPE(private_item_type), POINTER                   :: item, new_item

      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      ! if allready in dict just update its value
      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF ((item%key== key)) THEN
               item%value = value
               RETURN
            END IF
         END IF
         item => item%next
      END DO

      ! check load-factor
      IF (4*dict%size > 3*SIZE(dict%buckets)) THEN ! load-factor > 75%
         CALL change_capacity(dict, 2*SIZE(dict%buckets)) !double capacity
         idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1
      END IF

      ! create a new item
      ALLOCATE (new_item)
      new_item%hash = hash
      new_item%key = key
      new_item%value = value
      new_item%next => dict%buckets(idx)%p
      dict%buckets(idx)%p => new_item
      dict%size = dict%size+1

   END SUBROUTINE set_hashed

! **************************************************************************************************
!> \brief Internal routine for changing the dictionary's capacity.
!> \param dict ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE change_capacity(dict, new_capacity)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, old_size
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_buckets
      TYPE(private_item_type), POINTER                   :: item, prev_item

      new_cap = new_capacity
      ! pre checks
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(dict%buckets) == HUGE(i)) RETURN ! reached maximum - stay there.
         new_cap = HUGE(i) ! grow as far as possible
      END IF
      IF (new_cap < 1) &
         CALL cp__b("common/dict_str_i4.F",281,"dict_str_i4_change_capacity: new_capacity < 1.")
      IF (4*dict%size > 3*new_cap) &
         CALL cp__b("common/dict_str_i4.F",283,"dict_str_i4_change_capacity: new_capacity too small.")

      old_size = dict%size
      old_buckets => dict%buckets
      ALLOCATE (dict%buckets(new_capacity))
      dict%size = 0
      DO i = 1, SIZE(old_buckets)
         item => old_buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            CALL set_hashed(dict, item%key, item%value, item%hash)
            prev_item => item
            item => item%next
            DEALLOCATE (prev_item)
         END DO
      END DO

      DEALLOCATE (old_buckets)

      IF (old_size /= dict%size) &
         CALL cp__b("common/dict_str_i4.F",302,"dict_str_i4_change_capacity: assertion failed")
   END SUBROUTINE change_capacity

! **************************************************************************************************
!> \brief Gets a value for a given key from the dictionary.
!>        If the key is not found the default_value will be returned.
!>        If the key is not found and default_value was not provided the program stops.
!> \param dict ...
!> \param key ...
!> \param default_value ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_str_i4_get(dict, key, default_value) RESULT(value)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      CHARACTER(LEN=default_string_length)               :: key
      INTEGER(kind=int_4), INTENT(in), OPTIONAL          :: default_value
      INTEGER(kind=int_4)                                :: value

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",327,"dict_str_i4_get: dictionary is not initialized.")

      hash = hash_str(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF ((item%key== key)) THEN
               value = item%value
               RETURN
            END IF
         END IF
         item => item%next
      END DO

      IF (PRESENT(default_value)) THEN
         value = default_value
         RETURN
      END IF

      CALL cp__b("common/dict_str_i4.F",348,"dict_str_i4_get: Key not found in dictionary.")
   END FUNCTION dict_str_i4_get

! **************************************************************************************************
!> \brief Remove the value for a given key from the dictionary.
!>        If the key is not found the proigram stops.
!> \param dict ...
!> \param key ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_str_i4_del(dict, key)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      CHARACTER(LEN=default_string_length)               :: key

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item, prev_item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",368,"dict_str_i4_del: dictionary is not initialized.")

      hash = hash_str(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      prev_item => Null()
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF ((item%key== key)) THEN
               IF (ASSOCIATED(prev_item)) THEN
                  prev_item%next => item%next
               ELSE
                  dict%buckets(idx)%p => item%next
               END IF
               DEALLOCATE (item)
               dict%size = dict%size-1
               RETURN
            END IF
         END IF
         prev_item => item
         item => item%next
      END DO

      CALL cp__b("common/dict_str_i4.F",392,"dict_str_i4_del: Key not found in dictionary.")
   END SUBROUTINE dict_str_i4_del

! **************************************************************************************************
!> \brief Returns the number of key/value-items currently stored in the dictionary.
!> \param dict ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_str_i4_size(dict) RESULT(size)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",408,"dict_str_i4_size: dictionary is not initialized.")

      size = dict%size
   END FUNCTION dict_str_i4_size

! **************************************************************************************************
!> \brief Checks whether a given key is currently stored in the dictionary.
!> \param dict ...
!> \param key ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_str_i4_haskey(dict, key) RESULT(res)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      CHARACTER(LEN=default_string_length)               :: key
      LOGICAL                                            :: res

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",431,"dict_str_i4_haskey: dictionary is not initialized.")

      res = .FALSE.
      IF (dict%size == 0) RETURN

      hash = hash_str(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF ((item%key== key)) THEN
               res = .TRUE.
               RETURN
            END IF
         END IF
         item => item%next
      END DO

   END FUNCTION dict_str_i4_haskey

! **************************************************************************************************
!> \brief Returns a pointer to an array of all key/value-items stored in the dictionary.
!>        Caution: The caller is responsible for deallocating targeted array after usage.
!> \param dict ...
!> \retval items ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_str_i4_items(dict) RESULT(items)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict
      TYPE(dict_str_i4_item_type), DIMENSION(:), POINTER :: items

      INTEGER                                            :: i, j
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",469,"dict_str_i4_items: dictionary is not initialized.")

      ALLOCATE (items(dict%size))
      j = 1
      DO i = 1, SIZE(dict%buckets)
         item => dict%buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            items(j)%key = item%key
            items(j)%value = item%value
            j = j+1
            item => item%next
         END DO
      END DO

      IF (j /= dict%size+1) &
         CALL cp__b("common/dict_str_i4.F",484,"dict_str_i4_items: assertion failed!")
   END FUNCTION dict_str_i4_items

! **************************************************************************************************
!> \brief Copies all key/values-items from one dictionary to another.
!>        Afterards dict will contain all items from the from_dict and
!>        additionaly all its previous items, which were not overwritten.
!>        The two dictionaries have to be of the same type.
!> \param dict destination of items
!> \param from_dict source of items - will not be change
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_str_i4_update(dict, from_dict)
      TYPE(dict_str_i4_type), INTENT(inout)              :: dict, from_dict

      INTEGER                                            :: i
      TYPE(dict_str_i4_item_type), DIMENSION(:), POINTER :: from_items

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",505,"dict_str_i4_update: dictionary is not initialized.")
      IF (.NOT. ASSOCIATED(from_dict%buckets)) &
         CALL cp__b("common/dict_str_i4.F",507,"dict_str_i4_update: from_dict is not initialized.")

      from_items => dict_str_i4_items(from_dict)
      DO i = 1, SIZE(from_items)
         CALL dict_str_i4_set(dict, &
                              from_items(i)%key, from_items(i)%value)
      END DO
      DEALLOCATE (from_items)
   END SUBROUTINE dict_str_i4_update

END MODULE dict_str_i4
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief An array-based list which grows on demand.
!>        When the internal array is full, a new array of twice the size will be
!>        allocated and the items are copied over.
!>
!>        This list can also be used as a stack.
!>        Have look at list_push(), list_pop() and list_peek().
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify list__valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE list_callstackentry

   USE timings_base_type,               ONLY: callstack_entry_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'list_callstackentry'

   !this is an internal type
   TYPE private_item_type
      PRIVATE
      TYPE(callstack_entry_type)                   :: value
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a list-instance
   TYPE list_callstackentry_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: arr => Null()
      INTEGER                                       :: size = -1
   END TYPE list_callstackentry_type

   PUBLIC :: list_callstackentry_type
   PUBLIC :: list_callstackentry_init
   PUBLIC :: list_callstackentry_push
   PUBLIC :: list_callstackentry_pop
   PUBLIC :: list_callstackentry_peek
   PUBLIC :: list_callstackentry_insert
   PUBLIC :: list_callstackentry_set
   PUBLIC :: list_callstackentry_get
   PUBLIC :: list_callstackentry_del
   PUBLIC :: list_callstackentry_clear
   PUBLIC :: list_callstackentry_size
   PUBLIC :: list_callstackentry_destroy
   PUBLIC :: list_callstackentry_isready

CONTAINS

! **************************************************************************************************
!> \brief Test if the given list has been initialized.
!> \param list ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_callstackentry_isready(list) RESULT(res)
      TYPE(list_callstackentry_type), INTENT(in)         :: list
      LOGICAL                                            :: res

      res = ASSOCIATED(list%arr)
   END FUNCTION list_callstackentry_isready

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given list.
!>        This has to be called before any of the other routines.
!>        For deallocation call list_callstackentry_destroy.
!> \param list ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_init(list, initial_capacity)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_, stat

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 0) &
         CALL cp__b("common/list_callstackentry.F",104,"list_callstackentry_create: initial_capacity < 0")

      IF (ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",107,"list_callstackentry_create: list is already initialized.")

      ALLOCATE (list%arr(initial_capacity_), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_callstackentry.F",111,"list_callstackentry_init: allocation failed")

      list%size = 0
   END SUBROUTINE list_callstackentry_init

! **************************************************************************************************
!> \brief Deallocated the internal data-structures of the given list.
!>        Caution: If the stored values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_destroy(list)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",131,"list_callstackentry_destroy: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      DEALLOCATE (list%arr)
      list%size = -1
   END SUBROUTINE list_callstackentry_destroy

! **************************************************************************************************
!> \brief Assings the given value to the given position in the list.
!>        Thereby, the former value at that position gets overwritten.
!>        If the position is out of bounds, the program stops.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1.
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_set(list, value, pos)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      TYPE(callstack_entry_type), INTENT(in)             :: value
      INTEGER, INTENT(in)                                :: pos

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",157,"list_callstackentry_set: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_callstackentry.F",159,"list_callstackentry_set: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_callstackentry.F",161,"list_callstackentry_set: pos > size")
      list%arr(pos)%p%value = value
   END SUBROUTINE list_callstackentry_set

! **************************************************************************************************
!> \brief Appends the given value at the end of the list.
!> \param list ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_push(list, value)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      TYPE(callstack_entry_type), INTENT(in)             :: value

      INTEGER                                            :: stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",180,"list_callstackentry_push: list is not initialized.")
      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      ALLOCATE (list%arr(list%size)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_callstackentry.F",187,"list_callstackentry_push: allocation failed")
      list%arr(list%size)%p%value = value
   END SUBROUTINE list_callstackentry_push

! **************************************************************************************************
!> \brief Inserts the given value at the givenn position within the list.
!>        Values which lay behind the insertion-position move one position up.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+2 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_insert(list, value, pos)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      TYPE(callstack_entry_type), INTENT(in)             :: value
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i, stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",209,"list_callstackentry_insert: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_callstackentry.F",211,"list_callstackentry_insert: pos < 1")
      IF (pos > list%size+1) &
         CALL cp__b("common/list_callstackentry.F",213,"list_callstackentry_insert: pos > size+1")

      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      DO i = list%size, pos+1, -1
         list%arr(i)%p => list%arr(i-1)%p
      END DO

      ALLOCATE (list%arr(pos)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_callstackentry.F",225,"list_callstackentry_insert: allocation failed.")
      list%arr(pos)%p%value = value
   END SUBROUTINE list_callstackentry_insert

! **************************************************************************************************
!> \brief Returns the last element in the list.
!>    Is equivalent to: list_callstackentry_get(list, list_callstackentry_size(list))
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_callstackentry_peek(list) RESULT(value)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      TYPE(callstack_entry_type)                         :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",243,"list_callstackentry_peek: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_callstackentry.F",245,"list_callstackentry_peek: list is empty.")

      value = list%arr(list%size)%p%value
   END FUNCTION list_callstackentry_peek

! **************************************************************************************************
!> \brief Returns the last element in the list and removes it.
!>        Is equivialent to:
!>        value = list_callstackentry_get(list, list_callstackentry_size(list))
!>            call list_callstackentry_del(list, list_callstackentry_size(list))
!>
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_callstackentry_pop(list) RESULT(value)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      TYPE(callstack_entry_type)                         :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",267,"list_callstackentry_pop: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_callstackentry.F",269,"list_callstackentry_pop: list is empty.")

      value = list%arr(list%size)%p%value
      DEALLOCATE (list%arr(list%size)%p)
      list%size = list%size-1
   END FUNCTION list_callstackentry_pop

! **************************************************************************************************
!> \brief Removes all values from the list. The list itself is not deallocated.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_clear(list)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",289,"list_callstackentry_clear: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      list%size = 0
   END SUBROUTINE list_callstackentry_clear

!
! **************************************************************************************************
!> \brief Returns the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_callstackentry_get(list, pos) RESULT(value)
      TYPE(list_callstackentry_type), INTENT(in)         :: list
      INTEGER, INTENT(in)                                :: pos
      TYPE(callstack_entry_type)                         :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",313,"list_callstackentry_get: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_callstackentry.F",315,"list_callstackentry_get: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_callstackentry.F",317,"list_callstackentry_get: pos > size")

      value = list%arr(pos)%p%value

   END FUNCTION list_callstackentry_get

! **************************************************************************************************
!> \brief Removes the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_callstackentry_del(list, pos)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",338,"list_callstackentry_del: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_callstackentry.F",340,"list_callstackentry_det: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_callstackentry.F",342,"list_callstackentry_det: pos > size")

      DEALLOCATE (list%arr(pos)%p)
      DO i = pos, list%size-1
         list%arr(i)%p => list%arr(i+1)%p
      END DO

      list%size = list%size-1

   END SUBROUTINE list_callstackentry_del

! **************************************************************************************************
!> \brief Returns the current size of the list.
!> \param list ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_callstackentry_size(list) RESULT(size)
      TYPE(list_callstackentry_type), INTENT(in)         :: list
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_callstackentry.F",366,"list_callstackentry_size: list is not initialized.")

      size = list%size
   END FUNCTION list_callstackentry_size

! **************************************************************************************************
!> \brief Internal routine for changing the size of the internal array.
!> \param list ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE change_capacity(list, new_capacity)
      TYPE(list_callstackentry_type), INTENT(inout)      :: list
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, stat
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_arr

      new_cap = new_capacity
      IF (new_cap < 0) &
         CALL cp__b("common/list_callstackentry.F",388,"list_callstackentry_change_capacity: new_capacity < 0")
      IF (new_cap < list%size) &
         CALL cp__b("common/list_callstackentry.F",390,"list_callstackentry_change_capacity: new_capacity < size")
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(list%arr) == HUGE(i)) &
            CALL cp__b("common/list_callstackentry.F",393,"list_callstackentry_change_capacity: list has reached integer limit.")
         new_cap = HUGE(i) ! grow as far as possible
      END IF

      old_arr => list%arr
      ALLOCATE (list%arr(new_cap), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_callstackentry.F",400,"list_callstackentry_change_capacity: allocation failed")

      DO i = 1, list%size
         ALLOCATE (list%arr(i)%p, stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/list_callstackentry.F",405,"list_callstackentry_change_capacity: allocation failed")
         list%arr(i)%p%value = old_arr(i)%p%value
         DEALLOCATE (old_arr(i)%p)
      END DO
      DEALLOCATE (old_arr)

   END SUBROUTINE change_capacity

END MODULE list_callstackentry
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A dictionary (also known as hashtable or hashmap).
!>        Internally the dictionary uses an array to holds its data.
!>        If this array reaches a load-factor of 75%, a new array with twice the
!>        size will be allocated and the items are then copied over.
!>        This ensures that the dictionary will perform operations in O(1).
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify dict__keytype___valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! ***************************************************************************************************


MODULE dict_i4tuple_callstat
   USE kinds,                           ONLY: int_4,&
                                              int_8
   USE timings_base_type,               ONLY: call_stat_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   !this is an internal type
   !Calculating hashes might be expensive, therefore they are stored
   !for use during change_capacity().
   TYPE private_item_type
      PRIVATE
      INTEGER(kind=int_4), DIMENSION(2)                            :: key
      TYPE(call_stat_type), POINTER                          :: value
      INTEGER(KIND=int_8)                       :: hash
      TYPE(private_item_type), POINTER          :: next => Null()
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a dictionary-instance
   TYPE dict_i4tuple_callstat_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER      :: buckets => Null()
      INTEGER                                               :: size = -1
   END TYPE dict_i4tuple_callstat_type

   ! this is a public type, its returned by dict_items()
   TYPE dict_i4tuple_callstat_item_type
      INTEGER(kind=int_4), DIMENSION(2)        :: key
      TYPE(call_stat_type), POINTER      :: value
   END TYPE dict_i4tuple_callstat_item_type

   PUBLIC :: dict_i4tuple_callstat_type
   PUBLIC :: dict_i4tuple_callstat_item_type
   PUBLIC :: dict_i4tuple_callstat_init
   PUBLIC :: dict_i4tuple_callstat_isready
   PUBLIC :: dict_i4tuple_callstat_set
   PUBLIC :: dict_i4tuple_callstat_get
   PUBLIC :: dict_i4tuple_callstat_del
   PUBLIC :: dict_i4tuple_callstat_size
   PUBLIC :: dict_i4tuple_callstat_haskey
   PUBLIC :: dict_i4tuple_callstat_items
   PUBLIC :: dict_i4tuple_callstat_update
   PUBLIC :: dict_i4tuple_callstat_destroy

CONTAINS

! **************************************************************************************************
! This is joaat_hash from string_table.F
!
!> \brief generates the hash of a given string
!> \param key a string of any length
!> \retval hash ...
!> \par History
!>       09.2006 created [Joost VandeVondele]
!>       12.2012 copied and adopted [ole]
!> \note
!>       http://en.wikipedia.org/wiki/Hash_table
!>       http://www.burtleburtle.net/bob/hash/doobs.html
! **************************************************************************************************
   FUNCTION hash_str(key) RESULT(hash)
      CHARACTER(LEN=*), INTENT(IN)                       :: key
      INTEGER(KIND=int_8)                                :: hash

      INTEGER(KIND=int_8), PARAMETER                     :: b32 = 2_int_8**32-1_int_8

      INTEGER                                            :: i

      hash = 0_int_8
      DO i = 1, LEN(key)
         hash = IAND(hash+ICHAR(key(i:i)), b32)
         hash = IAND(hash+IAND(ISHFT(hash, 10), b32), b32)
         hash = IAND(IEOR(hash, IAND(ISHFT(hash, -6), b32)), b32)
      ENDDO
      hash = IAND(hash+IAND(ISHFT(hash, 3), b32), b32)
      hash = IAND(IEOR(hash, IAND(ISHFT(hash, -11), b32)), b32)
      hash = IAND(hash+IAND(ISHFT(hash, 15), b32), b32)
   END FUNCTION hash_str

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given dictionary.
!> \param dict ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_i4tuple_callstat_init(dict, initial_capacity)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 1) &
         CALL cp__b("common/dict_i4tuple_callstat.F",129,"dict_i4tuple_callstat_init: initial_capacity < 1")

      IF (ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",132,"dict_i4tuple_callstat_init: dictionary is already initialized.")

      ALLOCATE (dict%buckets(initial_capacity_))
      dict%size = 0

   END SUBROUTINE dict_i4tuple_callstat_init

! **************************************************************************************************
!> \brief Test if the given dictionary has been initialized.
!> \param dict ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_i4tuple_callstat_isready(dict) RESULT(res)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      LOGICAL                                            :: res

      res = ASSOCIATED(dict%buckets)
   END FUNCTION dict_i4tuple_callstat_isready

! **************************************************************************************************
!> \brief Deallocated the internal data-structures if the given dictionary.
!>        Caution: If the stored keys or values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param dict ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_i4tuple_callstat_destroy(dict)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict

      INTEGER                                            :: i
      TYPE(private_item_type), POINTER                   :: item, prev_item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",170,"dict_i4tuple_callstat_destroy: dictionary is not initialized.")

      DO i = 1, SIZE(dict%buckets)
         item => dict%buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            prev_item => item
            item => item%next
            DEALLOCATE (prev_item)
         END DO
      END DO

      DEALLOCATE (dict%buckets)
      dict%size = -1

   END SUBROUTINE dict_i4tuple_callstat_destroy

! **************************************************************************************************
!> \brief Stores, and possibly overwrites, a given value under a given key.
!> \param dict ...
!> \param key ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_i4tuple_callstat_set(dict, key, value)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER(kind=int_4), DIMENSION(2), INTENT(in)      :: key
      TYPE(call_stat_type), INTENT(in), POINTER          :: value

      INTEGER(KIND=int_8)                                :: hash

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",203,"dict_i4tuple_callstat_set: dictionary is not initialized.")

      hash = SUM(key)
      CALL set_hashed(dict, key, value, hash)
   END SUBROUTINE dict_i4tuple_callstat_set

! **************************************************************************************************
!> \brief Common code used internally by dict_set() and change_capacity().
!> \param dict ...
!> \param key ...
!> \param value ...
!> \param hash ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE set_hashed(dict, key, value, hash)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER(kind=int_4), DIMENSION(2), INTENT(in)      :: key
      TYPE(call_stat_type), INTENT(in), POINTER          :: value
      INTEGER(KIND=int_8), INTENT(in)                    :: hash

      INTEGER(KIND=int_8)                                :: idx
      TYPE(private_item_type), POINTER                   :: item, new_item

      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      ! if allready in dict just update its value
      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF (ALL(item%key== key)) THEN
               item%value => value
               RETURN
            END IF
         END IF
         item => item%next
      END DO

      ! check load-factor
      IF (4*dict%size > 3*SIZE(dict%buckets)) THEN ! load-factor > 75%
         CALL change_capacity(dict, 2*SIZE(dict%buckets)) !double capacity
         idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1
      END IF

      ! create a new item
      ALLOCATE (new_item)
      new_item%hash = hash
      new_item%key = key
      new_item%value => value
      new_item%next => dict%buckets(idx)%p
      dict%buckets(idx)%p => new_item
      dict%size = dict%size+1

   END SUBROUTINE set_hashed

! **************************************************************************************************
!> \brief Internal routine for changing the dictionary's capacity.
!> \param dict ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE change_capacity(dict, new_capacity)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, old_size
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_buckets
      TYPE(private_item_type), POINTER                   :: item, prev_item

      new_cap = new_capacity
      ! pre checks
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(dict%buckets) == HUGE(i)) RETURN ! reached maximum - stay there.
         new_cap = HUGE(i) ! grow as far as possible
      END IF
      IF (new_cap < 1) &
         CALL cp__b("common/dict_i4tuple_callstat.F",282,"dict_i4tuple_callstat_change_capacity: new_capacity < 1.")
      IF (4*dict%size > 3*new_cap) &
         CALL cp__b("common/dict_i4tuple_callstat.F",284,"dict_i4tuple_callstat_change_capacity: new_capacity too small.")

      old_size = dict%size
      old_buckets => dict%buckets
      ALLOCATE (dict%buckets(new_capacity))
      dict%size = 0
      DO i = 1, SIZE(old_buckets)
         item => old_buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            CALL set_hashed(dict, item%key, item%value, item%hash)
            prev_item => item
            item => item%next
            DEALLOCATE (prev_item)
         END DO
      END DO

      DEALLOCATE (old_buckets)

      IF (old_size /= dict%size) &
         CALL cp__b("common/dict_i4tuple_callstat.F",303,"dict_i4tuple_callstat_change_capacity: assertion failed")
   END SUBROUTINE change_capacity

! **************************************************************************************************
!> \brief Gets a value for a given key from the dictionary.
!>        If the key is not found the default_value will be returned.
!>        If the key is not found and default_value was not provided the program stops.
!> \param dict ...
!> \param key ...
!> \param default_value ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_i4tuple_callstat_get(dict, key, default_value) RESULT(value)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER(kind=int_4), DIMENSION(2)                  :: key
      TYPE(call_stat_type), INTENT(in), OPTIONAL, &
         POINTER                                         :: default_value
      TYPE(call_stat_type), POINTER                      :: value

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",329,"dict_i4tuple_callstat_get: dictionary is not initialized.")

      hash = SUM(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF (ALL(item%key== key)) THEN
               value => item%value
               RETURN
            END IF
         END IF
         item => item%next
      END DO

      IF (PRESENT(default_value)) THEN
         value => default_value
         RETURN
      END IF

      CALL cp__b("common/dict_i4tuple_callstat.F",350,"dict_i4tuple_callstat_get: Key not found in dictionary.")
   END FUNCTION dict_i4tuple_callstat_get

! **************************************************************************************************
!> \brief Remove the value for a given key from the dictionary.
!>        If the key is not found the proigram stops.
!> \param dict ...
!> \param key ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_i4tuple_callstat_del(dict, key)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER(kind=int_4), DIMENSION(2)                  :: key

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item, prev_item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",370,"dict_i4tuple_callstat_del: dictionary is not initialized.")

      hash = SUM(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      prev_item => Null()
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF (ALL(item%key== key)) THEN
               IF (ASSOCIATED(prev_item)) THEN
                  prev_item%next => item%next
               ELSE
                  dict%buckets(idx)%p => item%next
               END IF
               DEALLOCATE (item)
               dict%size = dict%size-1
               RETURN
            END IF
         END IF
         prev_item => item
         item => item%next
      END DO

      CALL cp__b("common/dict_i4tuple_callstat.F",394,"dict_i4tuple_callstat_del: Key not found in dictionary.")
   END SUBROUTINE dict_i4tuple_callstat_del

! **************************************************************************************************
!> \brief Returns the number of key/value-items currently stored in the dictionary.
!> \param dict ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_i4tuple_callstat_size(dict) RESULT(size)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",410,"dict_i4tuple_callstat_size: dictionary is not initialized.")

      size = dict%size
   END FUNCTION dict_i4tuple_callstat_size

! **************************************************************************************************
!> \brief Checks whether a given key is currently stored in the dictionary.
!> \param dict ...
!> \param key ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_i4tuple_callstat_haskey(dict, key) RESULT(res)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      INTEGER(kind=int_4), DIMENSION(2)                  :: key
      LOGICAL                                            :: res

      INTEGER(KIND=int_8)                                :: hash, idx
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",433,"dict_i4tuple_callstat_haskey: dictionary is not initialized.")

      res = .FALSE.
      IF (dict%size == 0) RETURN

      hash = SUM(key)
      idx = MOD(hash, INT(SIZE(dict%buckets), KIND=int_8))+1

      item => dict%buckets(idx)%p
      DO WHILE (ASSOCIATED(item))
         IF (item%hash == hash) THEN
            IF (ALL(item%key== key)) THEN
               res = .TRUE.
               RETURN
            END IF
         END IF
         item => item%next
      END DO

   END FUNCTION dict_i4tuple_callstat_haskey

! **************************************************************************************************
!> \brief Returns a pointer to an array of all key/value-items stored in the dictionary.
!>        Caution: The caller is responsible for deallocating targeted array after usage.
!> \param dict ...
!> \retval items ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dict_i4tuple_callstat_items(dict) RESULT(items)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict
      TYPE(dict_i4tuple_callstat_item_type), &
         DIMENSION(:), POINTER                           :: items

      INTEGER                                            :: i, j
      TYPE(private_item_type), POINTER                   :: item

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",472,"dict_i4tuple_callstat_items: dictionary is not initialized.")

      ALLOCATE (items(dict%size))
      j = 1
      DO i = 1, SIZE(dict%buckets)
         item => dict%buckets(i)%p
         DO WHILE (ASSOCIATED(item))
            items(j)%key = item%key
            items(j)%value => item%value
            j = j+1
            item => item%next
         END DO
      END DO

      IF (j /= dict%size+1) &
         CALL cp__b("common/dict_i4tuple_callstat.F",487,"dict_i4tuple_callstat_items: assertion failed!")
   END FUNCTION dict_i4tuple_callstat_items

! **************************************************************************************************
!> \brief Copies all key/values-items from one dictionary to another.
!>        Afterards dict will contain all items from the from_dict and
!>        additionaly all its previous items, which were not overwritten.
!>        The two dictionaries have to be of the same type.
!> \param dict destination of items
!> \param from_dict source of items - will not be change
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dict_i4tuple_callstat_update(dict, from_dict)
      TYPE(dict_i4tuple_callstat_type), INTENT(inout)    :: dict, from_dict

      INTEGER                                            :: i
      TYPE(dict_i4tuple_callstat_item_type), &
         DIMENSION(:), POINTER                           :: from_items

      IF (.NOT. ASSOCIATED(dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",509,"dict_i4tuple_callstat_update: dictionary is not initialized.")
      IF (.NOT. ASSOCIATED(from_dict%buckets)) &
         CALL cp__b("common/dict_i4tuple_callstat.F",511,"dict_i4tuple_callstat_update: from_dict is not initialized.")

      from_items => dict_i4tuple_callstat_items(from_dict)
      DO i = 1, SIZE(from_items)
         CALL dict_i4tuple_callstat_set(dict, &
                                        from_items(i)%key, from_items(i)%value)
      END DO
      DEALLOCATE (from_items)
   END SUBROUTINE dict_i4tuple_callstat_update

END MODULE dict_i4tuple_callstat
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types used by timings.F and timings_report.F
!>        Due to the fortran restriction on cicular module-dependencies
!>        the types, which are used through dict or list had to go
!>        into the separate module timings_base_type.F
!> \par History
!>      12.2012 Created [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE timings_types
   USE dict_i4tuple_callstat,           ONLY: dict_i4tuple_callstat_type
   USE dict_str_i4,                     ONLY: dict_str_i4_type
   USE list_callstackentry,             ONLY: list_callstackentry_type
   USE list_routinestat,                ONLY: list_routinestat_type

   IMPLICIT NONE
   PRIVATE

   TYPE timer_env_type
      INTEGER                                          :: ref_count
      TYPE(dict_str_i4_type)                           :: routine_names
      TYPE(list_routinestat_type)                      :: routine_stats
      TYPE(list_callstackentry_type)                   :: callstack
      TYPE(dict_i4tuple_callstat_type)                 :: callgraph
      INTEGER                                          :: trace_max
      INTEGER                                          :: trace_unit
      CHARACTER(len=13)                                :: trace_str
      LOGICAL                                          :: trace_all
   END TYPE timer_env_type

   PUBLIC :: timer_env_type

END MODULE timings_types

! **************************************************************************************************

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief An array-based list which grows on demand.
!>        When the internal array is full, a new array of twice the size will be
!>        allocated and the items are copied over.
!>
!>        This list can also be used as a stack.
!>        Have look at list_push(), list_pop() and list_peek().
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify list__valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE list_timerenv

   USE timings_types,                   ONLY: timer_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'list_timerenv'

   !this is an internal type
   TYPE private_item_type
      PRIVATE
      TYPE(timer_env_type), POINTER                   :: value
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a list-instance
   TYPE list_timerenv_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: arr => Null()
      INTEGER                                       :: size = -1
   END TYPE list_timerenv_type

   PUBLIC :: list_timerenv_type
   PUBLIC :: list_timerenv_init
   PUBLIC :: list_timerenv_push
   PUBLIC :: list_timerenv_pop
   PUBLIC :: list_timerenv_peek
   PUBLIC :: list_timerenv_insert
   PUBLIC :: list_timerenv_set
   PUBLIC :: list_timerenv_get
   PUBLIC :: list_timerenv_del
   PUBLIC :: list_timerenv_clear
   PUBLIC :: list_timerenv_size
   PUBLIC :: list_timerenv_destroy
   PUBLIC :: list_timerenv_isready

CONTAINS

! **************************************************************************************************
!> \brief Test if the given list has been initialized.
!> \param list ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_timerenv_isready(list) RESULT(res)
      TYPE(list_timerenv_type), INTENT(in)               :: list
      LOGICAL                                            :: res

      res = ASSOCIATED(list%arr)
   END FUNCTION list_timerenv_isready

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given list.
!>        This has to be called before any of the other routines.
!>        For deallocation call list_timerenv_destroy.
!> \param list ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_init(list, initial_capacity)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_, stat

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 0) &
         CALL cp__b("common/list_timerenv.F",104,"list_timerenv_create: initial_capacity < 0")

      IF (ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",107,"list_timerenv_create: list is already initialized.")

      ALLOCATE (list%arr(initial_capacity_), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_timerenv.F",111,"list_timerenv_init: allocation failed")

      list%size = 0
   END SUBROUTINE list_timerenv_init

! **************************************************************************************************
!> \brief Deallocated the internal data-structures of the given list.
!>        Caution: If the stored values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_destroy(list)
      TYPE(list_timerenv_type), INTENT(inout)            :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",131,"list_timerenv_destroy: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      DEALLOCATE (list%arr)
      list%size = -1
   END SUBROUTINE list_timerenv_destroy

! **************************************************************************************************
!> \brief Assings the given value to the given position in the list.
!>        Thereby, the former value at that position gets overwritten.
!>        If the position is out of bounds, the program stops.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1.
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_set(list, value, pos)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      TYPE(timer_env_type), INTENT(in), POINTER          :: value
      INTEGER, INTENT(in)                                :: pos

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",157,"list_timerenv_set: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_timerenv.F",159,"list_timerenv_set: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_timerenv.F",161,"list_timerenv_set: pos > size")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_timerenv_set

! **************************************************************************************************
!> \brief Appends the given value at the end of the list.
!> \param list ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_push(list, value)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      TYPE(timer_env_type), INTENT(in), POINTER          :: value

      INTEGER                                            :: stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",180,"list_timerenv_push: list is not initialized.")
      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      ALLOCATE (list%arr(list%size)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_timerenv.F",187,"list_timerenv_push: allocation failed")
      list%arr(list%size)%p%value => value
   END SUBROUTINE list_timerenv_push

! **************************************************************************************************
!> \brief Inserts the given value at the givenn position within the list.
!>        Values which lay behind the insertion-position move one position up.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+2 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_insert(list, value, pos)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      TYPE(timer_env_type), INTENT(in), POINTER          :: value
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i, stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",209,"list_timerenv_insert: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_timerenv.F",211,"list_timerenv_insert: pos < 1")
      IF (pos > list%size+1) &
         CALL cp__b("common/list_timerenv.F",213,"list_timerenv_insert: pos > size+1")

      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      DO i = list%size, pos+1, -1
         list%arr(i)%p => list%arr(i-1)%p
      END DO

      ALLOCATE (list%arr(pos)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_timerenv.F",225,"list_timerenv_insert: allocation failed.")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_timerenv_insert

! **************************************************************************************************
!> \brief Returns the last element in the list.
!>    Is equivalent to: list_timerenv_get(list, list_timerenv_size(list))
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_timerenv_peek(list) RESULT(value)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      TYPE(timer_env_type), POINTER                      :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",243,"list_timerenv_peek: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_timerenv.F",245,"list_timerenv_peek: list is empty.")

      value => list%arr(list%size)%p%value
   END FUNCTION list_timerenv_peek

! **************************************************************************************************
!> \brief Returns the last element in the list and removes it.
!>        Is equivialent to:
!>        value = list_timerenv_get(list, list_timerenv_size(list))
!>            call list_timerenv_del(list, list_timerenv_size(list))
!>
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_timerenv_pop(list) RESULT(value)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      TYPE(timer_env_type), POINTER                      :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",267,"list_timerenv_pop: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_timerenv.F",269,"list_timerenv_pop: list is empty.")

      value => list%arr(list%size)%p%value
      DEALLOCATE (list%arr(list%size)%p)
      list%size = list%size-1
   END FUNCTION list_timerenv_pop

! **************************************************************************************************
!> \brief Removes all values from the list. The list itself is not deallocated.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_clear(list)
      TYPE(list_timerenv_type), INTENT(inout)            :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",289,"list_timerenv_clear: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      list%size = 0
   END SUBROUTINE list_timerenv_clear

!
! **************************************************************************************************
!> \brief Returns the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_timerenv_get(list, pos) RESULT(value)
      TYPE(list_timerenv_type), INTENT(in)               :: list
      INTEGER, INTENT(in)                                :: pos
      TYPE(timer_env_type), POINTER                      :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",313,"list_timerenv_get: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_timerenv.F",315,"list_timerenv_get: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_timerenv.F",317,"list_timerenv_get: pos > size")

      value => list%arr(pos)%p%value

   END FUNCTION list_timerenv_get

! **************************************************************************************************
!> \brief Removes the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_timerenv_del(list, pos)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",338,"list_timerenv_del: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_timerenv.F",340,"list_timerenv_det: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_timerenv.F",342,"list_timerenv_det: pos > size")

      DEALLOCATE (list%arr(pos)%p)
      DO i = pos, list%size-1
         list%arr(i)%p => list%arr(i+1)%p
      END DO

      list%size = list%size-1

   END SUBROUTINE list_timerenv_del

! **************************************************************************************************
!> \brief Returns the current size of the list.
!> \param list ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_timerenv_size(list) RESULT(size)
      TYPE(list_timerenv_type), INTENT(in)               :: list
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_timerenv.F",366,"list_timerenv_size: list is not initialized.")

      size = list%size
   END FUNCTION list_timerenv_size

! **************************************************************************************************
!> \brief Internal routine for changing the size of the internal array.
!> \param list ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE change_capacity(list, new_capacity)
      TYPE(list_timerenv_type), INTENT(inout)            :: list
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, stat
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_arr

      new_cap = new_capacity
      IF (new_cap < 0) &
         CALL cp__b("common/list_timerenv.F",388,"list_timerenv_change_capacity: new_capacity < 0")
      IF (new_cap < list%size) &
         CALL cp__b("common/list_timerenv.F",390,"list_timerenv_change_capacity: new_capacity < size")
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(list%arr) == HUGE(i)) &
            CALL cp__b("common/list_timerenv.F",393,"list_timerenv_change_capacity: list has reached integer limit.")
         new_cap = HUGE(i) ! grow as far as possible
      END IF

      old_arr => list%arr
      ALLOCATE (list%arr(new_cap), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_timerenv.F",400,"list_timerenv_change_capacity: allocation failed")

      DO i = 1, list%size
         ALLOCATE (list%arr(i)%p, stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/list_timerenv.F",405,"list_timerenv_change_capacity: allocation failed")
         list%arr(i)%p%value => old_arr(i)%p%value
         DEALLOCATE (old_arr(i)%p)
      END DO
      DEALLOCATE (old_arr)

   END SUBROUTINE change_capacity

END MODULE list_timerenv
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief An array-based list which grows on demand.
!>        When the internal array is full, a new array of twice the size will be
!>        allocated and the items are copied over.
!>
!>        This list can also be used as a stack.
!>        Have look at list_push(), list_pop() and list_peek().
!> \note
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify list__valuetype_.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE list_routinereport

   USE timings_base_type,               ONLY: routine_report_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'list_routinereport'

   !this is an internal type
   TYPE private_item_type
      PRIVATE
      TYPE(routine_report_type), POINTER                   :: value
   END TYPE private_item_type

   !this is an internal type
   TYPE private_item_p_type
      PRIVATE
      TYPE(private_item_type), POINTER :: p => Null()
   END TYPE private_item_p_type

   ! this is the public type, which holds a list-instance
   TYPE list_routinereport_type
      PRIVATE
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: arr => Null()
      INTEGER                                       :: size = -1
   END TYPE list_routinereport_type

   PUBLIC :: list_routinereport_type
   PUBLIC :: list_routinereport_init
   PUBLIC :: list_routinereport_push
   PUBLIC :: list_routinereport_pop
   PUBLIC :: list_routinereport_peek
   PUBLIC :: list_routinereport_insert
   PUBLIC :: list_routinereport_set
   PUBLIC :: list_routinereport_get
   PUBLIC :: list_routinereport_del
   PUBLIC :: list_routinereport_clear
   PUBLIC :: list_routinereport_size
   PUBLIC :: list_routinereport_destroy
   PUBLIC :: list_routinereport_isready

CONTAINS

! **************************************************************************************************
!> \brief Test if the given list has been initialized.
!> \param list ...
!> \retval res ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinereport_isready(list) RESULT(res)
      TYPE(list_routinereport_type), INTENT(in)          :: list
      LOGICAL                                            :: res

      res = ASSOCIATED(list%arr)
   END FUNCTION list_routinereport_isready

! **************************************************************************************************
!> \brief Allocates the internal data-structures of the given list.
!>        This has to be called before any of the other routines.
!>        For deallocation call list_routinereport_destroy.
!> \param list ...
!> \param initial_capacity The initial size of the internal array (default=11).
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_init(list, initial_capacity)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      INTEGER, INTENT(in), OPTIONAL                      :: initial_capacity

      INTEGER                                            :: initial_capacity_, stat

      initial_capacity_ = 11
      IF (PRESENT(initial_capacity)) initial_capacity_ = initial_capacity

      IF (initial_capacity_ < 0) &
         CALL cp__b("common/list_routinereport.F",104,"list_routinereport_create: initial_capacity < 0")

      IF (ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",107,"list_routinereport_create: list is already initialized.")

      ALLOCATE (list%arr(initial_capacity_), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinereport.F",111,"list_routinereport_init: allocation failed")

      list%size = 0
   END SUBROUTINE list_routinereport_init

! **************************************************************************************************
!> \brief Deallocated the internal data-structures of the given list.
!>        Caution: If the stored values are pointers, their targets will
!>                 not get deallocated by this routine.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_destroy(list)
      TYPE(list_routinereport_type), INTENT(inout)       :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",131,"list_routinereport_destroy: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      DEALLOCATE (list%arr)
      list%size = -1
   END SUBROUTINE list_routinereport_destroy

! **************************************************************************************************
!> \brief Assings the given value to the given position in the list.
!>        Thereby, the former value at that position gets overwritten.
!>        If the position is out of bounds, the program stops.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1.
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_set(list, value, pos)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      TYPE(routine_report_type), INTENT(in), POINTER     :: value
      INTEGER, INTENT(in)                                :: pos

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",157,"list_routinereport_set: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinereport.F",159,"list_routinereport_set: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinereport.F",161,"list_routinereport_set: pos > size")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_routinereport_set

! **************************************************************************************************
!> \brief Appends the given value at the end of the list.
!> \param list ...
!> \param value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_push(list, value)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      TYPE(routine_report_type), INTENT(in), POINTER     :: value

      INTEGER                                            :: stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",180,"list_routinereport_push: list is not initialized.")
      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      ALLOCATE (list%arr(list%size)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinereport.F",187,"list_routinereport_push: allocation failed")
      list%arr(list%size)%p%value => value
   END SUBROUTINE list_routinereport_push

! **************************************************************************************************
!> \brief Inserts the given value at the givenn position within the list.
!>        Values which lay behind the insertion-position move one position up.
!> \param list ...
!> \param value ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+2 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_insert(list, value, pos)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      TYPE(routine_report_type), INTENT(in), POINTER     :: value
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i, stat

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",209,"list_routinereport_insert: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinereport.F",211,"list_routinereport_insert: pos < 1")
      IF (pos > list%size+1) &
         CALL cp__b("common/list_routinereport.F",213,"list_routinereport_insert: pos > size+1")

      IF (list%size == SIZE(list%arr)) &
         CALL change_capacity(list, 2*SIZE(list%arr)+1)

      list%size = list%size+1
      DO i = list%size, pos+1, -1
         list%arr(i)%p => list%arr(i-1)%p
      END DO

      ALLOCATE (list%arr(pos)%p, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinereport.F",225,"list_routinereport_insert: allocation failed.")
      list%arr(pos)%p%value => value
   END SUBROUTINE list_routinereport_insert

! **************************************************************************************************
!> \brief Returns the last element in the list.
!>    Is equivalent to: list_routinereport_get(list, list_routinereport_size(list))
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinereport_peek(list) RESULT(value)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      TYPE(routine_report_type), POINTER                 :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",243,"list_routinereport_peek: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_routinereport.F",245,"list_routinereport_peek: list is empty.")

      value => list%arr(list%size)%p%value
   END FUNCTION list_routinereport_peek

! **************************************************************************************************
!> \brief Returns the last element in the list and removes it.
!>        Is equivialent to:
!>        value = list_routinereport_get(list, list_routinereport_size(list))
!>            call list_routinereport_del(list, list_routinereport_size(list))
!>
!> \param list ...
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinereport_pop(list) RESULT(value)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      TYPE(routine_report_type), POINTER                 :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",267,"list_routinereport_pop: list is not initialized.")
      IF (list%size < 1) &
         CALL cp__b("common/list_routinereport.F",269,"list_routinereport_pop: list is empty.")

      value => list%arr(list%size)%p%value
      DEALLOCATE (list%arr(list%size)%p)
      list%size = list%size-1
   END FUNCTION list_routinereport_pop

! **************************************************************************************************
!> \brief Removes all values from the list. The list itself is not deallocated.
!> \param list ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_clear(list)
      TYPE(list_routinereport_type), INTENT(inout)       :: list

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",289,"list_routinereport_clear: list is not initialized.")

      DO i = 1, list%size
         DEALLOCATE (list%arr(i)%p)
      END DO
      list%size = 0
   END SUBROUTINE list_routinereport_clear

!
! **************************************************************************************************
!> \brief Returns the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \retval value ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinereport_get(list, pos) RESULT(value)
      TYPE(list_routinereport_type), INTENT(in)          :: list
      INTEGER, INTENT(in)                                :: pos
      TYPE(routine_report_type), POINTER                 :: value

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",313,"list_routinereport_get: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinereport.F",315,"list_routinereport_get: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinereport.F",317,"list_routinereport_get: pos > size")

      value => list%arr(pos)%p%value

   END FUNCTION list_routinereport_get

! **************************************************************************************************
!> \brief Removes the value at the given position from the list.
!> \param list ...
!> \param pos Position in the list - musst fulfill 0 < pos < list_size+1 .
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE list_routinereport_del(list, pos)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      INTEGER, INTENT(in)                                :: pos

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",338,"list_routinereport_del: list is not initialized.")
      IF (pos < 1) &
         CALL cp__b("common/list_routinereport.F",340,"list_routinereport_det: pos < 1")
      IF (pos > list%size) &
         CALL cp__b("common/list_routinereport.F",342,"list_routinereport_det: pos > size")

      DEALLOCATE (list%arr(pos)%p)
      DO i = pos, list%size-1
         list%arr(i)%p => list%arr(i+1)%p
      END DO

      list%size = list%size-1

   END SUBROUTINE list_routinereport_del

! **************************************************************************************************
!> \brief Returns the current size of the list.
!> \param list ...
!> \retval size ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION list_routinereport_size(list) RESULT(size)
      TYPE(list_routinereport_type), INTENT(in)          :: list
      INTEGER                                            :: size

      IF (.NOT. ASSOCIATED(list%arr)) &
         CALL cp__b("common/list_routinereport.F",366,"list_routinereport_size: list is not initialized.")

      size = list%size
   END FUNCTION list_routinereport_size

! **************************************************************************************************
!> \brief Internal routine for changing the size of the internal array.
!> \param list ...
!> \param new_capacity ...
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE change_capacity(list, new_capacity)
      TYPE(list_routinereport_type), INTENT(inout)       :: list
      INTEGER, INTENT(in)                                :: new_capacity

      INTEGER                                            :: i, new_cap, stat
      TYPE(private_item_p_type), DIMENSION(:), POINTER   :: old_arr

      new_cap = new_capacity
      IF (new_cap < 0) &
         CALL cp__b("common/list_routinereport.F",388,"list_routinereport_change_capacity: new_capacity < 0")
      IF (new_cap < list%size) &
         CALL cp__b("common/list_routinereport.F",390,"list_routinereport_change_capacity: new_capacity < size")
      IF (new_cap > HUGE(i)) THEN
         IF (SIZE(list%arr) == HUGE(i)) &
            CALL cp__b("common/list_routinereport.F",393,"list_routinereport_change_capacity: list has reached integer limit.")
         new_cap = HUGE(i) ! grow as far as possible
      END IF

      old_arr => list%arr
      ALLOCATE (list%arr(new_cap), stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/list_routinereport.F",400,"list_routinereport_change_capacity: allocation failed")

      DO i = 1, list%size
         ALLOCATE (list%arr(i)%p, stat=stat)
         IF (stat /= 0) &
            CALL cp__b("common/list_routinereport.F",405,"list_routinereport_change_capacity: allocation failed")
         list%arr(i)%p%value => old_arr(i)%p%value
         DEALLOCATE (old_arr(i)%p)
      END DO
      DEALLOCATE (old_arr)

   END SUBROUTINE change_capacity

END MODULE list_routinereport
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A common interface for all instances of list__valuetype_.template .
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE list
   USE list_callstackentry,             ONLY: &
        list_callstackentry_clear, list_callstackentry_del, list_callstackentry_destroy, &
        list_callstackentry_get, list_callstackentry_init, list_callstackentry_insert, &
        list_callstackentry_isready, list_callstackentry_peek, list_callstackentry_pop, &
        list_callstackentry_push, list_callstackentry_set, list_callstackentry_size
   USE list_routinereport,              ONLY: &
        list_routinereport_clear, list_routinereport_del, list_routinereport_destroy, &
        list_routinereport_get, list_routinereport_init, list_routinereport_isready, &
        list_routinereport_peek, list_routinereport_pop, list_routinereport_push, &
        list_routinereport_set, list_routinereport_size
   USE list_routinestat,                ONLY: &
        list_routinestat_clear, list_routinestat_del, list_routinestat_destroy, &
        list_routinestat_get, list_routinestat_init, list_routinestat_insert, &
        list_routinestat_isready, list_routinestat_peek, list_routinestat_pop, &
        list_routinestat_push, list_routinestat_set, list_routinestat_size
   USE list_timerenv,                   ONLY: &
        list_timerenv_clear, list_timerenv_del, list_timerenv_destroy, list_timerenv_get, &
        list_timerenv_init, list_timerenv_insert, list_timerenv_isready, list_timerenv_peek, &
        list_timerenv_pop, list_timerenv_push, list_timerenv_set, list_timerenv_size

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'list'

   PUBLIC :: list_init, list_isready, list_push, list_pop, list_peek, &
             list_get, list_size, list_destroy

   INTERFACE list_init
      MODULE PROCEDURE list_timerenv_init, list_routinestat_init, &
         list_callstackentry_init, list_routinereport_init
   END INTERFACE

   INTERFACE list_isready
      MODULE PROCEDURE list_timerenv_isready, &
         list_routinestat_isready, list_callstackentry_isready, &
         list_routinereport_isready
   END INTERFACE

   INTERFACE list_push
      MODULE PROCEDURE list_timerenv_push, list_routinestat_push, &
         list_callstackentry_push, list_routinereport_push
   END INTERFACE

   INTERFACE list_pop
      MODULE PROCEDURE list_timerenv_pop, list_routinestat_pop, &
         list_callstackentry_pop, list_routinereport_pop
   END INTERFACE

   INTERFACE list_peek
      MODULE PROCEDURE list_timerenv_peek, list_routinestat_peek, &
         list_callstackentry_peek, list_routinereport_peek
   END INTERFACE

   INTERFACE list_insert
      MODULE PROCEDURE list_timerenv_insert, &
         list_routinestat_insert, list_callstackentry_insert
   END INTERFACE

   INTERFACE list_set
      MODULE PROCEDURE list_timerenv_set, list_routinestat_set, &
         list_callstackentry_set, list_routinereport_set
   END INTERFACE

   INTERFACE list_get
      MODULE PROCEDURE list_timerenv_get, list_routinestat_get, &
         list_callstackentry_get, list_routinereport_get
   END INTERFACE

   INTERFACE list_del
      MODULE PROCEDURE list_timerenv_del, list_routinestat_del, &
         list_callstackentry_del, list_routinereport_del
   END INTERFACE

   INTERFACE list_clear
      MODULE PROCEDURE list_timerenv_clear, list_routinestat_clear, &
         list_callstackentry_clear, list_routinereport_clear
   END INTERFACE

   INTERFACE list_size
      MODULE PROCEDURE list_timerenv_size, list_routinestat_size, &
         list_callstackentry_size, list_routinereport_size
   END INTERFACE

   INTERFACE list_destroy
      MODULE PROCEDURE list_timerenv_destroy, &
         list_routinestat_destroy, list_callstackentry_destroy, &
         list_routinereport_destroy
   END INTERFACE

END MODULE list
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A common interface for all instances of dict__keytype___valuetype_.template .
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************

MODULE dict
   USE dict_i4tuple_callstat,           ONLY: &
        dict_i4tuple_callstat_del, dict_i4tuple_callstat_destroy, dict_i4tuple_callstat_get, &
        dict_i4tuple_callstat_haskey, dict_i4tuple_callstat_init, dict_i4tuple_callstat_isready, &
        dict_i4tuple_callstat_items, dict_i4tuple_callstat_set, dict_i4tuple_callstat_size, &
        dict_i4tuple_callstat_update
   USE dict_str_i4,                     ONLY: &
        dict_str_i4_del, dict_str_i4_destroy, dict_str_i4_get, dict_str_i4_haskey, &
        dict_str_i4_init, dict_str_i4_isready, dict_str_i4_items, dict_str_i4_set, &
        dict_str_i4_size, dict_str_i4_update

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dict'

   PUBLIC :: dict_init, dict_items, dict_haskey, &
             dict_set, dict_get, dict_size, dict_destroy

   INTERFACE dict_init
      MODULE PROCEDURE dict_str_i4_init, dict_i4tuple_callstat_init
   END INTERFACE

   INTERFACE dict_isready
      MODULE PROCEDURE dict_str_i4_isready, dict_i4tuple_callstat_isready
   END INTERFACE

   INTERFACE dict_haskey
      MODULE PROCEDURE dict_str_i4_haskey, dict_i4tuple_callstat_haskey
   END INTERFACE

   INTERFACE dict_set
      MODULE PROCEDURE dict_str_i4_set, dict_i4tuple_callstat_set
   END INTERFACE

   INTERFACE dict_get
      MODULE PROCEDURE dict_str_i4_get, dict_i4tuple_callstat_get
   END INTERFACE

   INTERFACE dict_del
      MODULE PROCEDURE dict_str_i4_del, dict_i4tuple_callstat_del
   END INTERFACE

   INTERFACE dict_items
      MODULE PROCEDURE dict_str_i4_items, dict_i4tuple_callstat_items
   END INTERFACE

   INTERFACE dict_size
      MODULE PROCEDURE dict_str_i4_size, dict_i4tuple_callstat_size
   END INTERFACE

   INTERFACE dict_update
      MODULE PROCEDURE dict_str_i4_update, dict_i4tuple_callstat_update
   END INTERFACE

   INTERFACE dict_destroy
      MODULE PROCEDURE dict_str_i4_destroy, dict_i4tuple_callstat_destroy
   END INTERFACE

END MODULE dict
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  routines for profiling cuda
!> \par History
!>      05.2013 created
!> \author Ole Schuett
! **************************************************************************************************
MODULE cuda_profiling
   USE ISO_C_BINDING,                   ONLY: C_CHAR,&
                                              C_INT,&
                                              C_NULL_CHAR,&
                                              C_SIZE_T
   USE kinds,                           ONLY: default_string_length,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC  :: cuda_nvtx_init, cuda_nvtx_range_push, cuda_nvtx_range_pop, cuda_mem_info


CONTAINS

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cuda_nvtx_init()
   END SUBROUTINE cuda_nvtx_init

! **************************************************************************************************
!> \brief ...
!> \param routineN ...
! **************************************************************************************************
   SUBROUTINE cuda_nvtx_range_push(routineN)
      CHARACTER(LEN=*), INTENT(IN)             :: routineN
      CALL cp_abort(cp__l("common/cuda_profiling.F",101), "cuda_nvtx_range_push: "// &
                    "__CUDA_PROFILING not compiled in, but called with:"//TRIM(routineN))
   END SUBROUTINE cuda_nvtx_range_push

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cuda_nvtx_range_pop()
      CALL cp__b("common/cuda_profiling.F",114,"cuda_nvtx_range_push: __CUDA_PROFILING not compiled in.")
   END SUBROUTINE cuda_nvtx_range_pop

! **************************************************************************************************
!> \brief ...
!> \param free ...
!> \param total ...
! **************************************************************************************************
   SUBROUTINE cuda_mem_info(free, total)
      INTEGER(KIND=int_8), INTENT(OUT)         :: free, total
      free = 0
      total = 0
   END SUBROUTINE cuda_mem_info

END MODULE cuda_profiling
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Timing routines for accounting
!> \par History
!>      02.2004 made a stacked version (of stacks...) [Joost VandeVondele]
!>      11.2004 storable timer_envs (for f77 interface) [fawzi]
!>      10.2005 binary search to speed up lookup in timeset [fawzi]
!>      12.2012 Complete rewrite based on dictionaries. [ole]
!> \author JGH
! **************************************************************************************************
MODULE timings
   USE base_hooks,                      ONLY: timeset_hook,&
                                              timestop_hook
   USE cuda_profiling,                  ONLY: cuda_mem_info,&
                                              cuda_nvtx_range_pop,&
                                              cuda_nvtx_range_push
   USE dict,                            ONLY: dict_destroy,&
                                              dict_get,&
                                              dict_init,&
                                              dict_items,&
                                              dict_set,&
                                              dict_size
   USE dict_i4tuple_callstat,           ONLY: dict_i4tuple_callstat_item_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE list,                            ONLY: list_destroy,&
                                              list_get,&
                                              list_init,&
                                              list_isready,&
                                              list_peek,&
                                              list_pop,&
                                              list_push,&
                                              list_size
   USE list_timerenv,                   ONLY: list_timerenv_type
   USE machine,                         ONLY: m_energy,&
                                              m_flush,&
                                              m_memory,&
                                              m_walltime
   USE timings_base_type,               ONLY: call_stat_type,&
                                              callstack_entry_type,&
                                              routine_stat_type
   USE timings_types,                   ONLY: timer_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: print_stack, timings_register_hooks

   ! these routines are currently only used by environment.F and f77_interface.F
   PUBLIC :: add_timer_env, rm_timer_env, get_timer_env
   PUBLIC :: timer_env_retain, timer_env_release
   PUBLIC :: timings_setup_tracing

   ! global variables
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'timings'
   TYPE(list_timerenv_type), SAVE, PRIVATE                  :: timers_stack

   !API (via pointer assignment to hook, PR67982, not meant to be called directly)
   PUBLIC :: timeset_handler, timestop_handler

CONTAINS

! **************************************************************************************************
!> \brief Registers handlers with base_hooks.F
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE timings_register_hooks()
      timeset_hook => timeset_handler
      timestop_hook => timestop_handler
   END SUBROUTINE timings_register_hooks

! **************************************************************************************************
!> \brief adds the given timer_env to the top of the stack
!> \param timer_env ...
!> \par History
!>      02.2004 created [Joost VandeVondele]
!> \note
!>      for each init_timer_env there should be the symmetric call to
!>      rm_timer_env
! **************************************************************************************************
   SUBROUTINE add_timer_env(timer_env)
      TYPE(timer_env_type), OPTIONAL, POINTER            :: timer_env

      TYPE(timer_env_type), POINTER                      :: timer_env_

      IF (PRESENT(timer_env)) timer_env_ => timer_env
      IF (.NOT. PRESENT(timer_env)) CALL timer_env_create(timer_env_)
      IF (.NOT. ASSOCIATED(timer_env_)) &
         CALL cp__b("common/timings.F",95,"add_timer_env: not associated")

      CALL timer_env_retain(timer_env_)
      IF (.NOT. list_isready(timers_stack)) CALL list_init(timers_stack)
      CALL list_push(timers_stack, timer_env_)
   END SUBROUTINE add_timer_env

! **************************************************************************************************
!> \brief creates a new timer env
!> \param timer_env ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE timer_env_create(timer_env)
      TYPE(timer_env_type), POINTER                      :: timer_env

      INTEGER                                            :: stat

      ALLOCATE (timer_env, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/timings.F",114,"timer_env_create: allocation failed")
      timer_env%ref_count = 0
      timer_env%trace_max = -1 ! tracing disabled by default
      timer_env%trace_all = .FALSE.
      CALL dict_init(timer_env%routine_names)
      CALL dict_init(timer_env%callgraph)
      CALL list_init(timer_env%routine_stats)
      CALL list_init(timer_env%callstack)
   END SUBROUTINE timer_env_create

! **************************************************************************************************
!> \brief removes the current timer env from the stack
!> \par History
!>      02.2004 created [Joost VandeVondele]
!> \note
!>      for each rm_timer_env there should have been the symmetric call to
!>      add_timer_env
! **************************************************************************************************
   SUBROUTINE rm_timer_env()
      TYPE(timer_env_type), POINTER                      :: timer_env

      timer_env => list_pop(timers_stack)
      CALL timer_env_release(timer_env)
      IF (list_size(timers_stack) == 0) CALL list_destroy(timers_stack)
   END SUBROUTINE rm_timer_env

! **************************************************************************************************
!> \brief returns the current timer env from the stack
!> \retval timer_env ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION get_timer_env() RESULT(timer_env)
      TYPE(timer_env_type), POINTER                      :: timer_env

      timer_env => list_peek(timers_stack)
   END FUNCTION get_timer_env

! **************************************************************************************************
!> \brief retains the given timer env
!> \param timer_env the timer env to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE timer_env_retain(timer_env)
      TYPE(timer_env_type), POINTER                      :: timer_env

      IF (.NOT. ASSOCIATED(timer_env)) &
         CALL cp__b("common/timings.F",160,"timer_env_retain: not associated")
      IF (timer_env%ref_count < 0) &
         CALL cp__b("common/timings.F",162,"timer_env_retain: negativ ref_count")
      timer_env%ref_count = timer_env%ref_count+1
   END SUBROUTINE timer_env_retain

! **************************************************************************************************
!> \brief releases the given timer env
!> \param timer_env the timer env to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE timer_env_release(timer_env)
      TYPE(timer_env_type), POINTER                      :: timer_env

      INTEGER                                            :: i
      TYPE(dict_i4tuple_callstat_item_type), &
         DIMENSION(:), POINTER                           :: ct_items
      TYPE(routine_stat_type), POINTER                   :: r_stat

      IF (.NOT. ASSOCIATED(timer_env)) &
         CALL cp__b("common/timings.F",180,"timer_env_release: not associated")
      IF (timer_env%ref_count < 0) &
         CALL cp__b("common/timings.F",182,"timer_env_release: negativ ref_count")
      timer_env%ref_count = timer_env%ref_count-1
      IF (timer_env%ref_count > 0) RETURN

      ! No more references left - let's tear down this timer_env...

      DO i = 1, list_size(timer_env%routine_stats)
         r_stat => list_get(timer_env%routine_stats, i)
         DEALLOCATE (r_stat)
      END DO

      ct_items => dict_items(timer_env%callgraph)
      DO i = 1, SIZE(ct_items)
         DEALLOCATE (ct_items(i)%value)
      END DO
      DEALLOCATE (ct_items)

      CALL dict_destroy(timer_env%routine_names)
      CALL dict_destroy(timer_env%callgraph)
      CALL list_destroy(timer_env%callstack)
      CALL list_destroy(timer_env%routine_stats)
      DEALLOCATE (timer_env)
   END SUBROUTINE timer_env_release

! **************************************************************************************************
!> \brief Start timer
!> \param routineN ...
!> \param handle ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE timeset_handler(routineN, handle)
      CHARACTER(LEN=*), INTENT(IN)                       :: routineN
      INTEGER, INTENT(OUT)                               :: handle

      CHARACTER(LEN=400)                                 :: line, mystring
      CHARACTER(LEN=60)                                  :: sformat
      CHARACTER(LEN=default_string_length)               :: routine_name_dsl
      INTEGER                                            :: routine_id, stack_size
      INTEGER(KIND=int_8)                                :: cpumem, gpumem_free, gpumem_total
      TYPE(callstack_entry_type)                         :: cs_entry
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

!$OMP MASTER
!$OMP CRITICAL(time_setstop)

      cs_entry%walltime_start = m_walltime()
      cs_entry%energy_start = m_energy()
      timer_env => list_peek(timers_stack)

      IF (LEN_TRIM(routineN) > default_string_length) THEN
         CALL cp__b("common/timings.F",235,'timings_timeset: routineN too long: "'//TRIM(routineN)//"'")
      END IF

      routine_name_dsl = routineN ! converte to default_string_length

      routine_id = routine_name2id(routine_name_dsl)
      ! update routine r_stats
      r_stat => list_get(timer_env%routine_stats, routine_id)
      stack_size = list_size(timer_env%callstack)
      r_stat%total_calls = r_stat%total_calls+1
      r_stat%active_calls = r_stat%active_calls+1
      r_stat%stackdepth_accu = r_stat%stackdepth_accu+stack_size+1

      ! add routine to callstack
      cs_entry%routine_id = routine_id
      CALL list_push(timer_env%callstack, cs_entry)

      !..if debug mode echo the subroutine name
      IF ((timer_env%trace_all .OR. r_stat%trace) .AND. &
          (r_stat%total_calls < timer_env%trace_max)) THEN
         WRITE (sformat, *) "(A,A,", MAX(1, 3*stack_size-4), "X,I4,1X,I6,1X,A,A)"
         WRITE (mystring, sformat) timer_env%trace_str, ">>", stack_size+1, &
            r_stat%total_calls, TRIM(r_stat%routineN), "       start"
         CALL cuda_mem_info(gpumem_free, gpumem_total)
         CALL m_memory(cpumem)
         WRITE (line, '(A,A,I0,A,A,I0,A)') TRIM(mystring), &
            " Hostmem: ", (cpumem+1024*1024-1)/(1024*1024), " MB", &
            " GPUmem: ", (gpumem_total-gpumem_free)/(1024*1024), " MB"
         WRITE (timer_env%trace_unit, *) TRIM(line)
         CALL m_flush(timer_env%trace_unit)
      END IF

      handle = routine_id


!$OMP END CRITICAL(time_setstop)
!$OMP END MASTER

   END SUBROUTINE timeset_handler

! **************************************************************************************************
!> \brief End timer
!> \param handle ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE timestop_handler(handle)
      INTEGER, INTENT(in)                                :: handle

      CHARACTER(LEN=400)                                 :: line, mystring
      CHARACTER(LEN=60)                                  :: sformat
      INTEGER                                            :: routine_id, stack_size
      INTEGER(KIND=int_8)                                :: cpumem, gpumem_free, gpumem_total
      INTEGER, DIMENSION(2)                              :: routine_tuple
      REAL(KIND=dp)                                      :: en_elapsed, en_now, wt_elapsed, wt_now
      TYPE(call_stat_type), POINTER                      :: c_stat
      TYPE(callstack_entry_type)                         :: cs_entry, prev_cs_entry
      TYPE(routine_stat_type), POINTER                   :: prev_stat, r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      routine_id = handle

!$OMP MASTER
!$OMP CRITICAL(time_setstop)


      wt_now = m_walltime()
      en_now = m_energy()
      timer_env => list_peek(timers_stack)
      cs_entry = list_pop(timer_env%callstack)
      r_stat => list_get(timer_env%routine_stats, cs_entry%routine_id)

      IF (handle /= cs_entry%routine_id) THEN
         PRINT*, "list_size(timer_env%callstack) ", list_size(timer_env%callstack), &
            " handle ", handle, " list_size(timers_stack) ", list_size(timers_stack)
         CALL cp__b("common/timings.F",317,'mismatched timestop '//TRIM(r_stat%routineN)//' in routine timestop')
      END IF

      ! add the elapsed time for this timeset/timestop to the time accumulator
      wt_elapsed = wt_now-cs_entry%walltime_start
      en_elapsed = en_now-cs_entry%energy_start
      r_stat%active_calls = r_stat%active_calls-1

      ! if we're the last instance in the stack, we do the accounting of the total time
      IF (r_stat%active_calls == 0) THEN
         r_stat%incl_walltime_accu = r_stat%incl_walltime_accu+wt_elapsed
         r_stat%incl_energy_accu = r_stat%incl_energy_accu+en_elapsed
      ENDIF

      ! exclusive time we always sum, since children will correct this time with their total time
      r_stat%excl_walltime_accu = r_stat%excl_walltime_accu+wt_elapsed
      r_stat%excl_energy_accu = r_stat%excl_energy_accu+en_elapsed

      stack_size = list_size(timer_env%callstack)
      IF (stack_size > 0) THEN
         prev_cs_entry = list_peek(timer_env%callstack)
         prev_stat => list_get(timer_env%routine_stats, prev_cs_entry%routine_id)
         ! we fixup the clock of the caller
         prev_stat%excl_walltime_accu = prev_stat%excl_walltime_accu-wt_elapsed
         prev_stat%excl_energy_accu = prev_stat%excl_energy_accu-en_elapsed

         !update callgraph
         routine_tuple = (/prev_cs_entry%routine_id, routine_id/)
         c_stat => dict_get(timer_env%callgraph, routine_tuple, default_value=Null(c_stat))
         IF (.NOT. ASSOCIATED(c_stat)) THEN
            ALLOCATE (c_stat)
            c_stat%total_calls = 0
            c_stat%incl_walltime_accu = 0.0_dp
            c_stat%incl_energy_accu = 0.0_dp
            CALL dict_set(timer_env%callgraph, routine_tuple, c_stat)
         END IF
         c_stat%total_calls = c_stat%total_calls+1
         c_stat%incl_walltime_accu = c_stat%incl_walltime_accu+wt_elapsed
         c_stat%incl_energy_accu = c_stat%incl_energy_accu+en_elapsed
      ENDIF

      !..if debug mode echo the subroutine name
      IF ((timer_env%trace_all .OR. r_stat%trace) .AND. &
          (r_stat%total_calls < timer_env%trace_max)) THEN
         WRITE (sformat, *) "(A,A,", MAX(1, 3*stack_size-4), "X,I4,1X,I6,1X,A,F12.3)"
         WRITE (mystring, sformat) timer_env%trace_str, "<<", stack_size+1, &
            r_stat%total_calls, TRIM(r_stat%routineN), wt_elapsed
         CALL cuda_mem_info(gpumem_free, gpumem_total)
         CALL m_memory(cpumem)
         WRITE (line, '(A,A,I0,A,A,I0,A)') TRIM(mystring), &
            " Hostmem: ", (cpumem+1024*1024-1)/(1024*1024), " MB", &
            " GPUmem: ", (gpumem_total-gpumem_free)/(1024*1024), " MB"
         WRITE (timer_env%trace_unit, *) TRIM(line)
         CALL m_flush(timer_env%trace_unit)
      ENDIF

!$OMP END CRITICAL(time_setstop)
!$OMP END MASTER

   END SUBROUTINE timestop_handler

! **************************************************************************************************
!> \brief Set routine tracer
!> \param trace_max  maximum number of calls reported per routine.
!>           Setting this to zero disables tracing.
!> \param unit_nr output unit used for printing the trace-messages
!> \param trace_str short info-string which is printed along with every message
!> \param routine_names List of routine-names.
!>                     If provided only these routines will be traced.
!>                     If not present all routines will traced.
!> \par History
!>       12.2012  added ability to trace only certain routines [ole]
!> \author JGH
! **************************************************************************************************
   SUBROUTINE timings_setup_tracing(trace_max, unit_nr, trace_str, routine_names)
      INTEGER, INTENT(IN)                                :: trace_max, unit_nr
      CHARACTER(len=13), INTENT(IN)                      :: trace_str
      CHARACTER(len=default_string_length), &
         DIMENSION(:), INTENT(IN), OPTIONAL              :: routine_names

      INTEGER                                            :: i, routine_id
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      timer_env => list_peek(timers_stack)
      timer_env%trace_max = trace_max
      timer_env%trace_unit = unit_nr
      timer_env%trace_str = trace_str
      timer_env%trace_all = .TRUE.
      IF (.NOT. PRESENT(routine_names)) RETURN

      ! setup routine-specific tracing
      timer_env%trace_all = .FALSE.
      DO i = 1, SIZE(routine_names)
         routine_id = routine_name2id(routine_names(i))
         r_stat => list_get(timer_env%routine_stats, routine_id)
         r_stat%trace = .TRUE.
      END DO

   END SUBROUTINE timings_setup_tracing

! **************************************************************************************************
!> \brief Print current routine stack
!> \param unit_nr ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE print_stack(unit_nr)
      INTEGER, INTENT(IN)                                :: unit_nr

      INTEGER                                            :: i
      TYPE(callstack_entry_type)                         :: cs_entry
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      timer_env => list_peek(timers_stack)
      WRITE (unit_nr, '(/,A,/)') " ===== Routine Calling Stack ===== "
      DO i = list_size(timer_env%callstack), 1, -1
         cs_entry = list_get(timer_env%callstack, i)
         r_stat => list_get(timer_env%routine_stats, cs_entry%routine_id)
         WRITE (unit_nr, '(T10,I4,1X,A)') i, TRIM(r_stat%routineN)
      END DO
      CALL m_flush(unit_nr)

   END SUBROUTINE print_stack

! **************************************************************************************************
!> \brief Internal routine used by timestet and timings_setup_tracing.
!>        If no routine with given name is found in timer_env%routine_names
!>        then a new entiry is created.
!> \param routineN ...
!> \retval routine_id ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION routine_name2id(routineN) RESULT(routine_id)
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: routineN
      INTEGER                                            :: routine_id

      INTEGER                                            :: stat
      TYPE(routine_stat_type), POINTER                   :: r_stat
      TYPE(timer_env_type), POINTER                      :: timer_env

      timer_env => list_peek(timers_stack)
      routine_id = dict_get(timer_env%routine_names, routineN, default_value=-1)

      IF (routine_id /= -1) RETURN ! found an id - let's return it
      ! routine not found - let's create it

      ! enforce space free timer names, to make the output of trace/timings of a fixed number fields
      IF (INDEX(routineN(1:LEN_TRIM(routineN)), ' ') /= 0) THEN
         CALL cp__b("common/timings.F",468,"timings_name2id: routineN contains spaces: "//routineN)
      END IF

      ! register routine_name_dsl with new routine_id
      routine_id = dict_size(timer_env%routine_names)+1
      CALL dict_set(timer_env%routine_names, routineN, routine_id)

      ALLOCATE (r_stat, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/timings.F",477,"timings_name2id: allocation failed")
      r_stat%routine_id = routine_id
      r_stat%routineN = routineN
      r_stat%active_calls = 0
      r_stat%excl_walltime_accu = 0.0_dp
      r_stat%incl_walltime_accu = 0.0_dp
      r_stat%excl_energy_accu = 0.0_dp
      r_stat%incl_energy_accu = 0.0_dp
      r_stat%total_calls = 0
      r_stat%stackdepth_accu = 0
      r_stat%trace = .FALSE.
      CALL list_push(timer_env%routine_stats, r_stat)
      IF (list_size(timer_env%routine_stats) /= dict_size(timer_env%routine_names)) &
         CALL cp__b("common/timings.F",490,"timings_name2id: assertion failed")
   END FUNCTION routine_name2id

END MODULE timings

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface to the message passing library MPI
!> \par History
!>      JGH (02-Jan-2001): New error handling
!>                         Performance tools
!>      JGH (14-Jan-2001): New routines mp_comm_compare, mp_cart_coords,
!>                                      mp_rank_compare, mp_alltoall
!>      JGH (06-Feb-2001): New routines mp_comm_free
!>      JGH (22-Mar-2001): New routines mp_comm_dup
!>      fawzi (04-NOV-2004): storable performance info (for f77 interface)
!>      Wrapper routine for mpi_gatherv added (22.12.2005,MK)
!>      JGH (13-Feb-2006): Flexibel precision
!>      JGH (15-Feb-2006): single precision mp_alltoall
!> \author JGH
! **************************************************************************************************
MODULE message_passing
  USE ISO_C_BINDING,                   ONLY: C_F_POINTER,&
                                             C_PTR
  USE kinds,                           ONLY: &
       dp, int_4, int_4_size, int_8, int_8_size, real_4, real_4_size, real_8, &
       real_8_size
  USE machine,                         ONLY: m_abort,&
                                             m_walltime

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler



   IMPLICIT NONE
   PRIVATE

   ! parameters that might be needed

   LOGICAL, PARAMETER :: cp2k_is_parallel = .FALSE.
   INTEGER, PARAMETER, PUBLIC :: mp_any_tag = -1
   INTEGER, PARAMETER, PUBLIC :: mp_any_source = -2
   INTEGER, PARAMETER, PUBLIC :: mp_comm_null = -3
   INTEGER, PARAMETER, PUBLIC :: mp_request_null = -4
   INTEGER, PARAMETER, PUBLIC :: mp_win_null = -5
   INTEGER, PARAMETER, PUBLIC :: mp_status_size = -6

   INTEGER, PARAMETER, PUBLIC :: file_offset = int_8
   INTEGER, PARAMETER, PUBLIC :: file_amode_create = 1
   INTEGER, PARAMETER, PUBLIC :: file_amode_rdonly = 2
   INTEGER, PARAMETER, PUBLIC :: file_amode_wronly = 4
   INTEGER, PARAMETER, PUBLIC :: file_amode_rdwr = 8
   INTEGER, PARAMETER, PUBLIC :: file_amode_excl = 64

   ! we need to fix this to a given number (crossing fingers)
   ! so that the serial code using Fortran stream IO and the MPI have the same sizes.
   INTEGER, PARAMETER, PUBLIC :: mpi_character_size = 1
   INTEGER, PARAMETER, PUBLIC :: mpi_integer_size = 4

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'message_passing'


   ! init and error
   PUBLIC :: mp_world_init, mp_world_finalize
   PUBLIC :: mp_abort

   ! performance gathering
   PUBLIC :: mp_perf_env_type
   PUBLIC :: mp_perf_env_retain, mp_perf_env_release
   PUBLIC :: add_mp_perf_env, rm_mp_perf_env, get_mp_perf_env, describe_mp_perf_env

   ! informational / generation of sub comms
   PUBLIC :: mp_environ, mp_comm_compare, mp_cart_coords, mp_rank_compare
   PUBLIC :: mp_cart_create, mp_dims_create, mp_cart_rank, mp_cart_sub, mp_comm_free
   PUBLIC :: mp_comm_dup, mp_comm_split, mp_comm_split_direct
   PUBLIC :: cp2k_is_parallel
   PUBLIC :: mp_probe

   ! message passing
   PUBLIC :: mp_bcast, mp_sum, mp_sum_partial, mp_max, mp_maxloc, mp_minloc, mp_min, mp_sync, mp_isum
   PUBLIC :: mp_gather, mp_scatter, mp_alltoall, mp_sendrecv, mp_allgather, mp_iallgather
   PUBLIC :: mp_isend, mp_irecv, mp_ibcast
   PUBLIC :: mp_shift, mp_isendrecv, mp_wait, mp_waitall, mp_waitany, mp_testany
   PUBLIC :: mp_testall, mp_iscatter
   PUBLIC :: mp_gatherv, mp_igatherv
   PUBLIC :: mp_send, mp_recv

   ! Memory management
   PUBLIC :: mp_allocate, mp_deallocate

   ! MPI re-ordering
   PUBLIC :: mp_reordering

   ! default communicators
   PUBLIC :: MPI_COMM_SELF, MPI_COMM_WORLD, MPI_COMM_NULL

   ! I/O
   PUBLIC :: mp_file_open, mp_file_close
   PUBLIC :: mp_file_write_at
   PUBLIC :: mp_file_write_at_all, mp_file_read_at_all
   PUBLIC :: mp_file_get_size

   ! some 'advanced types' currently only used for dbcsr
   PUBLIC :: mp_type_descriptor_type
   PUBLIC :: mp_type_make, mp_type_free
   PUBLIC :: mp_type_size

   ! some benchmarking code
   PUBLIC :: mpi_perf_test

   ! one-sided communication
   PUBLIC :: mp_win_create, mp_win_free, mp_win_lock_all, &
             mp_win_unlock_all, mp_rget

   ! vector types
   PUBLIC :: mp_type_indexed_make_i, mp_type_indexed_make_l, &
             mp_type_indexed_make_r, mp_type_indexed_make_d, &
             mp_type_indexed_make_c, mp_type_indexed_make_z

   INTEGER, PARAMETER :: MPI_COMM_SELF = 0, MPI_COMM_WORLD = 0, MPI_COMM_NULL = -1

   ! assumed to be private

! Interface declarations for non-data-oriented subroutines.

   INTERFACE mp_environ
      MODULE PROCEDURE mp_environ_l, mp_environ_c, mp_environ_c2
   END INTERFACE

   INTERFACE mp_waitall
      MODULE PROCEDURE mp_waitall_1, mp_waitall_2
   END INTERFACE

   INTERFACE mp_testall
      MODULE PROCEDURE mp_testall_t2
   END INTERFACE

   INTERFACE mp_testany
      MODULE PROCEDURE mp_testany_1, mp_testany_2
   END INTERFACE

   INTERFACE mp_type_free
      MODULE PROCEDURE mp_type_free_m, mp_type_free_v
   END INTERFACE

   !
   ! interfaces to deal easily with scalars / vectors / matrice / ...
   ! of the different types (integers, doubles, logicals, characters)
   !
   INTERFACE mp_minloc
      MODULE PROCEDURE mp_minloc_iv, &
         mp_minloc_lv, &
         mp_minloc_rv, &
         mp_minloc_dv
   END INTERFACE

   INTERFACE mp_maxloc
      MODULE PROCEDURE mp_maxloc_iv, &
         mp_maxloc_lv, &
         mp_maxloc_rv, &
         mp_maxloc_dv
   END INTERFACE

   INTERFACE mp_shift
      MODULE PROCEDURE mp_shift_im, mp_shift_i, &
         mp_shift_lm, mp_shift_l, &
         mp_shift_rm, mp_shift_r, &
         mp_shift_dm, mp_shift_d, &
         mp_shift_cm, mp_shift_c, &
         mp_shift_zm, mp_shift_z
   END INTERFACE

   INTERFACE mp_bcast
      MODULE PROCEDURE mp_bcast_i, mp_bcast_iv, mp_bcast_im, mp_bcast_i3, &
         mp_bcast_l, mp_bcast_lv, mp_bcast_lm, mp_bcast_l3, &
         mp_bcast_r, mp_bcast_rv, mp_bcast_rm, mp_bcast_r3, &
         mp_bcast_d, mp_bcast_dv, mp_bcast_dm, mp_bcast_d3, &
         mp_bcast_c, mp_bcast_cv, mp_bcast_cm, mp_bcast_c3, &
         mp_bcast_z, mp_bcast_zv, mp_bcast_zm, mp_bcast_z3
      MODULE PROCEDURE mp_bcast_b, mp_bcast_bv
      MODULE PROCEDURE mp_bcast_av, mp_bcast_am
   END INTERFACE

   INTERFACE mp_ibcast
      MODULE PROCEDURE mp_ibcast_i, mp_ibcast_iv, &
         mp_ibcast_l, mp_ibcast_lv, &
         mp_ibcast_r, mp_ibcast_rv, &
         mp_ibcast_d, mp_ibcast_dv, &
         mp_ibcast_c, mp_ibcast_cv, &
         mp_ibcast_z, mp_ibcast_zv
   END INTERFACE

   INTERFACE mp_sum
      MODULE PROCEDURE mp_sum_i, mp_sum_iv, mp_sum_im, mp_sum_im3, mp_sum_im4, &
         mp_sum_l, mp_sum_lv, mp_sum_lm, mp_sum_lm3, mp_sum_lm4, &
         mp_sum_r, mp_sum_rv, mp_sum_rm, mp_sum_rm3, mp_sum_rm4, &
         mp_sum_d, mp_sum_dv, mp_sum_dm, mp_sum_dm3, mp_sum_dm4, &
         mp_sum_c, mp_sum_cv, mp_sum_cm, mp_sum_cm3, mp_sum_cm4, &
         mp_sum_z, mp_sum_zv, mp_sum_zm, mp_sum_zm3, mp_sum_zm4, &
         mp_sum_root_iv, mp_sum_root_im, &
         mp_sum_root_lv, mp_sum_root_lm, &
         mp_sum_root_rv, mp_sum_root_rm, &
         mp_sum_root_dv, mp_sum_root_dm, &
         mp_sum_root_cv, mp_sum_root_cm, &
         mp_sum_root_zv, mp_sum_root_zm
      MODULE PROCEDURE mp_sum_b, mp_sum_bv
   END INTERFACE

   INTERFACE mp_isum
      MODULE PROCEDURE mp_isum_iv, &
         mp_isum_lv, &
         mp_isum_rv, &
         mp_isum_dv, &
         mp_isum_cv, &
         mp_isum_zv
   END INTERFACE

   INTERFACE mp_sum_partial
      MODULE PROCEDURE mp_sum_partial_im, &
         mp_sum_partial_lm, &
         mp_sum_partial_rm, &
         mp_sum_partial_dm, &
         mp_sum_partial_cm, &
         mp_sum_partial_zm
   END INTERFACE

   INTERFACE mp_max
      MODULE PROCEDURE mp_max_i, mp_max_iv, &
         mp_max_l, mp_max_lv, &
         mp_max_r, mp_max_rv, &
         mp_max_d, mp_max_dv, &
         mp_max_c, mp_max_cv, &
         mp_max_z, mp_max_zv
   END INTERFACE

   INTERFACE mp_min
      MODULE PROCEDURE mp_min_i, mp_min_iv, &
         mp_min_l, mp_min_lv, &
         mp_min_r, mp_min_rv, &
         mp_min_d, mp_min_dv, &
         mp_min_c, mp_min_cv, &
         mp_min_z, mp_min_zv
   END INTERFACE

   INTERFACE mp_gather
      MODULE PROCEDURE mp_gather_i, mp_gather_iv, mp_gather_im, &
         mp_gather_l, mp_gather_lv, mp_gather_lm, &
         mp_gather_r, mp_gather_rv, mp_gather_rm, &
         mp_gather_d, mp_gather_dv, mp_gather_dm, &
         mp_gather_c, mp_gather_cv, mp_gather_cm, &
         mp_gather_z, mp_gather_zv, mp_gather_zm
   END INTERFACE

   INTERFACE mp_gatherv
      MODULE PROCEDURE mp_gatherv_iv, &
         mp_gatherv_lv, &
         mp_gatherv_rv, &
         mp_gatherv_dv, &
         mp_gatherv_cv, &
         mp_gatherv_zv
   END INTERFACE

   INTERFACE mp_igatherv
      MODULE PROCEDURE mp_igatherv_iv, &
         mp_igatherv_lv, &
         mp_igatherv_rv, &
         mp_igatherv_dv, &
         mp_igatherv_cv, &
         mp_igatherv_zv
   END INTERFACE

!> todo: move allgatherv to a separate declaration
   INTERFACE mp_allgather
      MODULE PROCEDURE &
         mp_allgather_i, &
         mp_allgather_i12, mp_allgather_i23, mp_allgather_i34, &
         mp_allgather_i22, &
         mp_allgather_l, &
         mp_allgather_l12, mp_allgather_l23, mp_allgather_l34, &
         mp_allgather_l22, &
         mp_allgather_r, &
         mp_allgather_r12, mp_allgather_r23, mp_allgather_r34, &
         mp_allgather_r22, &
         mp_allgather_d, &
         mp_allgather_d12, mp_allgather_d23, mp_allgather_d34, &
         mp_allgather_d22, &
         mp_allgather_c, &
         mp_allgather_c12, mp_allgather_c23, mp_allgather_c34, &
         mp_allgather_c22, &
         mp_allgather_z, &
         mp_allgather_z12, mp_allgather_z23, mp_allgather_z34, &
         mp_allgather_z22, &
         mp_allgatherv_iv, &
         mp_allgatherv_lv, &
         mp_allgatherv_rv, &
         mp_allgatherv_dv, &
         mp_allgatherv_cv, &
         mp_allgatherv_zv
   END INTERFACE

   INTERFACE mp_iallgather
      MODULE PROCEDURE &
         mp_iallgather_i, mp_iallgather_l, &
         mp_iallgather_r, mp_iallgather_d, &
         mp_iallgather_c, mp_iallgather_z, &
         mp_iallgather_i22, mp_iallgather_l22, &
         mp_iallgather_r22, mp_iallgather_d22, &
         mp_iallgather_c22, mp_iallgather_z22, &
         mp_iallgather_i33, mp_iallgather_l33, &
         mp_iallgather_r33, mp_iallgather_d33, &
         mp_iallgather_c33, mp_iallgather_z33, &
         mp_iallgatherv_iv, &
         mp_iallgatherv_lv, &
         mp_iallgatherv_rv, &
         mp_iallgatherv_dv, &
         mp_iallgatherv_cv, &
         mp_iallgatherv_zv
   END INTERFACE

   INTERFACE mp_scatter
      MODULE PROCEDURE mp_scatter_iv, &
         mp_scatter_lv, &
         mp_scatter_rv, &
         mp_scatter_dv, &
         mp_scatter_cv, &
         mp_scatter_zv
   END INTERFACE

   INTERFACE mp_iscatter
      MODULE PROCEDURE mp_iscatter_i, &
         mp_iscatter_l, &
         mp_iscatter_r, &
         mp_iscatter_d, &
         mp_iscatter_c, &
         mp_iscatter_z, &
         mp_iscatter_iv, &
         mp_iscatter_lv, &
         mp_iscatter_rv, &
         mp_iscatter_dv, &
         mp_iscatter_cv, &
         mp_iscatter_zv, &
         mp_iscatterv_iv, &
         mp_iscatterv_lv, &
         mp_iscatterv_rv, &
         mp_iscatterv_dv, &
         mp_iscatterv_cv, &
         mp_iscatterv_zv
   END INTERFACE

   INTERFACE mp_sum_scatter
      MODULE PROCEDURE mp_sum_scatter_iv, &
         mp_sum_scatter_lv, &
         mp_sum_scatter_rv, &
         mp_sum_scatter_dv, &
         mp_sum_scatter_cv, &
         mp_sum_scatter_zv
   END INTERFACE

   INTERFACE mp_alltoall
      MODULE PROCEDURE mp_alltoall_i, mp_alltoall_i22, mp_alltoall_i33, &
         mp_alltoall_i44, mp_alltoall_i45, mp_alltoall_i34, &
         mp_alltoall_i11v, mp_alltoall_i22v, mp_alltoall_i54, &
         mp_alltoall_l, mp_alltoall_l22, mp_alltoall_l33, &
         mp_alltoall_l44, mp_alltoall_l45, mp_alltoall_l34, &
         mp_alltoall_l11v, mp_alltoall_l22v, mp_alltoall_l54, &
         mp_alltoall_r, mp_alltoall_r22, mp_alltoall_r33, &
         mp_alltoall_r44, mp_alltoall_r45, mp_alltoall_r34, &
         mp_alltoall_r11v, mp_alltoall_r22v, mp_alltoall_r54, &
         mp_alltoall_d, mp_alltoall_d22, mp_alltoall_d33, &
         mp_alltoall_d44, mp_alltoall_d45, mp_alltoall_d34, &
         mp_alltoall_d11v, mp_alltoall_d22v, mp_alltoall_d54, &
         mp_alltoall_c, mp_alltoall_c22, mp_alltoall_c33, &
         mp_alltoall_c44, mp_alltoall_c45, mp_alltoall_c34, &
         mp_alltoall_c11v, mp_alltoall_c22v, mp_alltoall_c54, &
         mp_alltoall_z, mp_alltoall_z22, mp_alltoall_z33, &
         mp_alltoall_z44, mp_alltoall_z45, mp_alltoall_z34, &
         mp_alltoall_z11v, mp_alltoall_z22v, mp_alltoall_z54
   END INTERFACE

   INTERFACE mp_send
      MODULE PROCEDURE mp_send_i, mp_send_iv, &
         mp_send_l, mp_send_lv, &
         mp_send_r, mp_send_rv, &
         mp_send_d, mp_send_dv, &
         mp_send_c, mp_send_cv, &
         mp_send_z, mp_send_zv
   END INTERFACE

   INTERFACE mp_recv
      MODULE PROCEDURE mp_recv_i, mp_recv_iv, &
         mp_recv_l, mp_recv_lv, &
         mp_recv_r, mp_recv_rv, &
         mp_recv_d, mp_recv_dv, &
         mp_recv_c, mp_recv_cv, &
         mp_recv_z, mp_recv_zv
   END INTERFACE

   INTERFACE mp_sendrecv
      MODULE PROCEDURE mp_sendrecv_iv, mp_sendrecv_im2, mp_sendrecv_im3, mp_sendrecv_im4, &
         mp_sendrecv_lv, mp_sendrecv_lm2, mp_sendrecv_lm3, mp_sendrecv_lm4, &
         mp_sendrecv_rv, mp_sendrecv_rm2, mp_sendrecv_rm3, mp_sendrecv_rm4, &
         mp_sendrecv_dv, mp_sendrecv_dm2, mp_sendrecv_dm3, mp_sendrecv_dm4, &
         mp_sendrecv_cv, mp_sendrecv_cm2, mp_sendrecv_cm3, mp_sendrecv_cm4, &
         mp_sendrecv_zv, mp_sendrecv_zm2, mp_sendrecv_zm3, mp_sendrecv_zm4
   END INTERFACE

   INTERFACE mp_isendrecv
      MODULE PROCEDURE mp_isendrecv_i, mp_isendrecv_iv, &
         mp_isendrecv_l, mp_isendrecv_lv, &
         mp_isendrecv_r, mp_isendrecv_rv, &
         mp_isendrecv_d, mp_isendrecv_dv, &
         mp_isendrecv_c, mp_isendrecv_cv, &
         mp_isendrecv_z, mp_isendrecv_zv
   END INTERFACE

   INTERFACE mp_isend
      MODULE PROCEDURE mp_isend_iv, mp_isend_im2, mp_isend_im3, &
         mp_isend_lv, mp_isend_lm2, mp_isend_lm3, &
         mp_isend_rv, mp_isend_rm2, mp_isend_rm3, &
         mp_isend_dv, mp_isend_dm2, mp_isend_dm3, &
         mp_isend_cv, mp_isend_cm2, mp_isend_cm3, &
         mp_isend_zv, mp_isend_zm2, mp_isend_zm3
      MODULE PROCEDURE mp_isend_custom
   END INTERFACE

   INTERFACE mp_irecv
      MODULE PROCEDURE mp_irecv_iv, mp_irecv_im2, mp_irecv_im3, &
         mp_irecv_lv, mp_irecv_lm2, mp_irecv_lm3, &
         mp_irecv_rv, mp_irecv_rm2, mp_irecv_rm3, &
         mp_irecv_dv, mp_irecv_dm2, mp_irecv_dm3, &
         mp_irecv_cv, mp_irecv_cm2, mp_irecv_cm3, &
         mp_irecv_zv, mp_irecv_zm2, mp_irecv_zm3
      MODULE PROCEDURE mp_irecv_custom
   END INTERFACE

   INTERFACE mp_win_create
      MODULE PROCEDURE mp_win_create_iv, &
         mp_win_create_lv, &
         mp_win_create_rv, &
         mp_win_create_dv, &
         mp_win_create_cv, &
         mp_win_create_zv
   END INTERFACE

   INTERFACE mp_rget
      MODULE PROCEDURE mp_rget_iv, &
         mp_rget_lv, &
         mp_rget_rv, &
         mp_rget_dv, &
         mp_rget_cv, &
         mp_rget_zv
   END INTERFACE

   INTERFACE mp_allocate
      MODULE PROCEDURE mp_allocate_i, &
         mp_allocate_l, &
         mp_allocate_r, &
         mp_allocate_d, &
         mp_allocate_c, &
         mp_allocate_z
   END INTERFACE

   INTERFACE mp_deallocate
      MODULE PROCEDURE mp_deallocate_i, &
         mp_deallocate_l, &
         mp_deallocate_r, &
         mp_deallocate_d, &
         mp_deallocate_c, &
         mp_deallocate_z
   END INTERFACE

   INTERFACE mp_type_make
      MODULE PROCEDURE mp_type_make_struct
      MODULE PROCEDURE mp_type_make_i, mp_type_make_l, &
         mp_type_make_r, mp_type_make_d, &
         mp_type_make_c, mp_type_make_z
   END INTERFACE

   INTERFACE mp_file_write_at
      MODULE PROCEDURE mp_file_write_at_ch, mp_file_write_at_chv, &
         mp_file_write_at_i, mp_file_write_at_iv, &
         mp_file_write_at_r, mp_file_write_at_rv, &
         mp_file_write_at_d, mp_file_write_at_dv, &
         mp_file_write_at_c, mp_file_write_at_cv, &
         mp_file_write_at_z, mp_file_write_at_zv, &
         mp_file_write_at_l, mp_file_write_at_lv
   END INTERFACE

   INTERFACE mp_file_write_at_all
      MODULE PROCEDURE mp_file_write_at_all_ch, mp_file_write_at_all_chv, &
         mp_file_write_at_all_i, mp_file_write_at_all_iv, &
         mp_file_write_at_all_l, mp_file_write_at_all_lv, &
         mp_file_write_at_all_r, mp_file_write_at_all_rv, &
         mp_file_write_at_all_d, mp_file_write_at_all_dv, &
         mp_file_write_at_all_c, mp_file_write_at_all_cv, &
         mp_file_write_at_all_z, mp_file_write_at_all_zv
   END INTERFACE

   INTERFACE mp_file_read_at
      MODULE PROCEDURE mp_file_read_at_ch, mp_file_read_at_chv, &
         mp_file_read_at_i, mp_file_read_at_iv, &
         mp_file_read_at_r, mp_file_read_at_rv, &
         mp_file_read_at_d, mp_file_read_at_dv, &
         mp_file_read_at_c, mp_file_read_at_cv, &
         mp_file_read_at_z, mp_file_read_at_zv, &
         mp_file_read_at_l, mp_file_read_at_lv
   END INTERFACE

   INTERFACE mp_file_read_at_all
      MODULE PROCEDURE mp_file_read_at_all_ch, mp_file_read_at_all_chv, &
         mp_file_read_at_all_i, mp_file_read_at_all_iv, &
         mp_file_read_at_all_l, mp_file_read_at_all_lv, &
         mp_file_read_at_all_r, mp_file_read_at_all_rv, &
         mp_file_read_at_all_d, mp_file_read_at_all_dv, &
         mp_file_read_at_all_c, mp_file_read_at_all_cv, &
         mp_file_read_at_all_z, mp_file_read_at_all_zv
   END INTERFACE

   INTERFACE mp_alloc_mem
      MODULE PROCEDURE mp_alloc_mem_i, mp_alloc_mem_l, &
         mp_alloc_mem_d, mp_alloc_mem_z, &
         mp_alloc_mem_r, mp_alloc_mem_c
   END INTERFACE

   INTERFACE mp_free_mem
      MODULE PROCEDURE mp_free_mem_i, mp_free_mem_l, &
         mp_free_mem_d, mp_free_mem_z, &
         mp_free_mem_r, mp_free_mem_c
   END INTERFACE

! Type declarations
   TYPE mp_indexing_meta_type
      INTEGER, DIMENSION(:), POINTER :: index, chunks
   END TYPE mp_indexing_meta_type

   TYPE mp_type_descriptor_type
      INTEGER :: type_handle
      INTEGER :: length
      INTEGER(kind=int_4), DIMENSION(:), POINTER :: data_i
      INTEGER(kind=int_8), DIMENSION(:), POINTER :: data_l
      REAL(kind=real_4), DIMENSION(:), POINTER :: data_r
      REAL(kind=real_8), DIMENSION(:), POINTER :: data_d
      COMPLEX(kind=real_4), DIMENSION(:), POINTER :: data_c
      COMPLEX(kind=real_8), DIMENSION(:), POINTER :: data_z
      TYPE(mp_type_descriptor_type), DIMENSION(:), POINTER :: subtype
      INTEGER :: vector_descriptor(2)
      LOGICAL :: has_indexing
      TYPE(mp_indexing_meta_type) :: index_descriptor
   END TYPE mp_type_descriptor_type

   ! type internally used to store message passing performance indicators
! **************************************************************************************************
   TYPE mp_perf_type
      CHARACTER(LEN=20) :: name
      INTEGER :: count
      REAL(KIND=dp) :: msg_size
      REAL(KIND=dp) :: time
   END TYPE mp_perf_type

   INTEGER, PARAMETER :: MAX_PERF = 21

! **************************************************************************************************
   TYPE mp_perf_env_type
      !private
      INTEGER :: ref_count, id_nr
      TYPE(mp_perf_type), DIMENSION(MAX_PERF) :: mp_perfs
   END TYPE mp_perf_env_type

! **************************************************************************************************
   TYPE mp_perf_env_p_type
      TYPE(mp_perf_env_type), POINTER         :: mp_perf_env
   END TYPE mp_perf_env_p_type

   ! introduce a stack of mp_perfs, first index is the stack pointer, for convience is replacing
   INTEGER, PARAMETER :: max_stack_size = 10
   INTEGER            :: stack_pointer = 0
   ! target attribute needed as a hack around ifc 7.1 bug
   TYPE(mp_perf_env_p_type), DIMENSION(max_stack_size), TARGET :: mp_perf_stack

   CHARACTER(LEN=20), PARAMETER :: sname(MAX_PERF) = &
                                   (/"MP_Group            ", "MP_Bcast            ", "MP_Allreduce        ", &
                                     "MP_Gather           ", "MP_Sync             ", "MP_Alltoall         ", &
                                     "MP_SendRecv         ", "MP_ISendRecv        ", "MP_Wait             ", &
                                     "MP_comm_split       ", "MP_ISend            ", "MP_IRecv            ", &
                                     "MP_Send             ", "MP_Recv             ", "MP_Memory           ", &
                                     "MP_Put              ", "MP_Get              ", "MP_Fence            ", &
                                     "MP_Win_Lock         ", "MP_Win_Create       ", "MP_Win_Free         "/)

   ! we make some assumptions on the length of INTEGERS, REALS and LOGICALS
   INTEGER, PARAMETER :: intlen = BIT_SIZE(0)/8
   INTEGER, PARAMETER :: reallen = 8
   INTEGER, PARAMETER :: loglen = BIT_SIZE(0)/8
   INTEGER, PARAMETER :: charlen = 1
   INTEGER, SAVE, PRIVATE :: last_mp_perf_env_id = 0

   LOGICAL, PUBLIC, SAVE :: mp_collect_timings = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief initializes the system default communicator
!> \param mp_comm [output] : handle of the default communicator
!> \par History
!>      2.2004 created [Joost VandeVondele ]
!> \note
!>      should only be called once
! **************************************************************************************************
   SUBROUTINE mp_world_init(mp_comm)
      INTEGER, INTENT(OUT)                     :: mp_comm
      mp_comm = 0
      CALL add_mp_perf_env()
   END SUBROUTINE mp_world_init

! **************************************************************************************************
!> \brief re-create the system default communicator with a different MPI
!>        rank order
!> \param mp_comm [output] : handle of the default communicator
!> \param mp_new_comm ...
!> \param ranks_order ...
!> \par History
!>      1.2012 created [ Christiane Pousa ]
!> \note
!>      should only be called once, st very begining of CP2K run
! **************************************************************************************************
   SUBROUTINE mp_reordering(mp_comm, mp_new_comm, ranks_order)
      INTEGER, INTENT(IN)                      :: mp_comm
      INTEGER, INTENT(out)                     :: mp_new_comm
      INTEGER, DIMENSION(:), POINTER           :: ranks_order

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_reordering', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      CALL mp_timeset(routineN, handle)
      ierr = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ranks_order))==-1) EXIT ;  END DO ; ENDIF
      mp_new_comm = mp_comm
      CALL mp_timestop(handle)
   END SUBROUTINE mp_reordering

! **************************************************************************************************
!> \brief finalizes the system default communicator
!> \par History
!>      2.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mp_world_finalize()

      CALL rm_mp_perf_env()

   END SUBROUTINE mp_world_finalize

! all the following routines should work for a given communicator, not MPI_WORLD

! **************************************************************************************************
!> \brief start and stop the performance indicators
!>      for every call to start there has to be (exactly) one call to stop
!> \param perf_env ...
!> \par History
!>      2.2004 created [Joost VandeVondele]
!> \note
!>      can be used to measure performance of a sub-part of a program.
!>      timings measured here will not show up in the outer start/stops
!>      Doesn't need a fresh communicator
! **************************************************************************************************
   SUBROUTINE add_mp_perf_env(perf_env)
      TYPE(mp_perf_env_type), OPTIONAL, POINTER          :: perf_env

      stack_pointer = stack_pointer+1
      IF (stack_pointer > max_stack_size) THEN
         CALL cp__b("mpiwrap/message_passing.F",815,"stack_pointer too large : message_passing @ add_mp_perf_env")
      ENDIF
      NULLIFY (mp_perf_stack(stack_pointer)%mp_perf_env)
      IF (PRESENT(perf_env)) THEN
         mp_perf_stack(stack_pointer)%mp_perf_env => perf_env
         IF (ASSOCIATED(perf_env)) CALL mp_perf_env_retain(perf_env)
      END IF
      IF (.NOT. ASSOCIATED(mp_perf_stack(stack_pointer)%mp_perf_env)) THEN
         CALL mp_perf_env_create(mp_perf_stack(stack_pointer)%mp_perf_env)
      END IF
   END SUBROUTINE add_mp_perf_env

! **************************************************************************************************
!> \brief ...
!> \param perf_env ...
! **************************************************************************************************
   SUBROUTINE mp_perf_env_create(perf_env)
      TYPE(mp_perf_env_type), OPTIONAL, POINTER          :: perf_env

      INTEGER                                            :: i, stat

      NULLIFY (perf_env)
      ALLOCATE (perf_env, stat=stat)
      IF (stat /= 0) THEN
         CALL cp__b("mpiwrap/message_passing.F",839,"allocation failed in mp_perf_env_create")
      ENDIF
      last_mp_perf_env_id = last_mp_perf_env_id+1
      perf_env%id_nr = last_mp_perf_env_id
      perf_env%ref_count = 1
      DO i = 1, MAX_PERF
         perf_env%mp_perfs(i)%name = sname(i)
         perf_env%mp_perfs(i)%count = 0
         perf_env%mp_perfs(i)%msg_size = 0.0_dp
         perf_env%mp_perfs(i)%time = 0.0_dp
      END DO

   END SUBROUTINE mp_perf_env_create

! **************************************************************************************************
!> \brief ...
!> \param perf_env ...
! **************************************************************************************************
   SUBROUTINE mp_perf_env_release(perf_env)
      TYPE(mp_perf_env_type), POINTER                    :: perf_env

      IF (ASSOCIATED(perf_env)) THEN
         IF (perf_env%ref_count < 1) THEN
            CALL cp__b("mpiwrap/message_passing.F",862,"invalid ref_count: message_passing @ mp_perf_env_release")
         END IF
         perf_env%ref_count = perf_env%ref_count-1
         IF (perf_env%ref_count == 0) THEN
            DEALLOCATE (perf_env)
         END IF
      END IF
      NULLIFY (perf_env)
   END SUBROUTINE mp_perf_env_release

! **************************************************************************************************
!> \brief ...
!> \param perf_env ...
! **************************************************************************************************
   SUBROUTINE mp_perf_env_retain(perf_env)
      TYPE(mp_perf_env_type), POINTER                    :: perf_env

      IF (.NOT. ASSOCIATED(perf_env)) THEN
         CALL cp__b("mpiwrap/message_passing.F",880,"unassociated perf_env: message_passing @ mp_perf_env_retain")
      END IF
      IF (perf_env%ref_count < 1) THEN
         CALL cp__b("mpiwrap/message_passing.F",883,"invalid ref_count: message_passing @ mp_perf_env_retain")
      END IF
      perf_env%ref_count = perf_env%ref_count+1
   END SUBROUTINE mp_perf_env_retain

!.. reports the performance counters for the MPI run
! **************************************************************************************************
!> \brief ...
!> \param perf_env ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE mp_perf_env_describe(perf_env, iw)
      TYPE(mp_perf_env_type), POINTER          :: perf_env
      INTEGER, INTENT(IN)                      :: iw


      IF (.NOT. ASSOCIATED(perf_env)) THEN
         CALL cp__b("mpiwrap/message_passing.F",904,"unassociated perf_env : message_passing @ mp_perf_env_describe")
      ENDIF
      IF (perf_env%ref_count < 1) THEN
         CALL cp__b("mpiwrap/message_passing.F",907,"invalid perf_env%ref_count : message_passing @ mp_perf_env_describe")
      ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iw))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE mp_perf_env_describe

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE rm_mp_perf_env()
      IF (stack_pointer < 1) THEN
         CALL cp__b("mpiwrap/message_passing.F",951,"no perf_env in the stack : message_passing @ rm_mp_perf_env")
      ENDIF
      CALL mp_perf_env_release(mp_perf_stack(stack_pointer)%mp_perf_env)
      stack_pointer = stack_pointer-1
   END SUBROUTINE rm_mp_perf_env

! **************************************************************************************************
!> \brief ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION get_mp_perf_env() RESULT(res)
      TYPE(mp_perf_env_type), POINTER                    :: res

      IF (stack_pointer < 1) THEN
         CALL cp__b("mpiwrap/message_passing.F",965,"no perf_env in the stack : message_passing @ get_mp_perf_env")
      ENDIF
      res => mp_perf_stack(stack_pointer)%mp_perf_env
   END FUNCTION get_mp_perf_env

! **************************************************************************************************
!> \brief ...
!> \param scr ...
! **************************************************************************************************
   SUBROUTINE describe_mp_perf_env(scr)
      INTEGER, INTENT(in)                                :: scr

      TYPE(mp_perf_env_type), POINTER                    :: perf_env

      perf_env => get_mp_perf_env()
      CALL mp_perf_env_describe(perf_env, scr)
   END SUBROUTINE describe_mp_perf_env

! **************************************************************************************************
!> \brief adds the performance informations of one call
!> \param perf_id ...
!> \param count ...
!> \param time ...
!> \param msg_size ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE add_perf(perf_id, count, time, msg_size)
      INTEGER, INTENT(in)                      :: perf_id
      INTEGER, INTENT(in), OPTIONAL            :: count
      REAL(KIND=dp), INTENT(in), OPTIONAL      :: time
      INTEGER, INTENT(in), OPTIONAL            :: msg_size

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(perf_id))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(time))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg_size))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE add_perf

! **************************************************************************************************
!> \brief globally stops all tasks
!>       this is intended to be low level, most of CP2K should call cp_abort()
! **************************************************************************************************
   SUBROUTINE mp_abort()
      INTEGER                                            :: ierr

      ierr = 0

      CALL m_abort()
      ! this routine never returns and levels with non-zero exit code
      STOP 1
   END SUBROUTINE mp_abort

! **************************************************************************************************
!> \brief stops *after an mpi error* translating the error code
!> \param ierr an error code * returned by an mpi call *
!> \param prg_code ...
!> \note
!>       this function is private to message_passing.F
! **************************************************************************************************
   SUBROUTINE mp_stop(ierr, prg_code)
      INTEGER, INTENT(IN)                      :: ierr
      CHARACTER(LEN=*)                         :: prg_code

      CHARACTER(LEN=512)                       :: full_error

      WRITE (full_error, '(A,I0,A)') ' MPI error (!?) ', ierr, ' in '//TRIM(prg_code)

      CALL cp__b("mpiwrap/message_passing.F",1065,full_error)

   END SUBROUTINE mp_stop

! **************************************************************************************************
!> \brief synchronizes with a barrier a given group of mpi tasks
!> \param group mpi communicator
! **************************************************************************************************
   SUBROUTINE mp_sync(group)
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_sync', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_sync

! **************************************************************************************************
!> \brief returns number of tasks and task id for a given mpi group
!>       simple and cartesian version.. recursive needed in case of failing mpi_comm_rank.
!> \param numtask ...
!> \param taskid ...
!> \param groupid mpi communicator
!> \note
!>         ..mp_world_setup is gone, use mp_environ instead (i.e. give a groupid explicitly)
! **************************************************************************************************
   RECURSIVE SUBROUTINE mp_environ_l(numtask, taskid, groupid)

      INTEGER, INTENT(OUT)                               :: numtask, taskid
      INTEGER, INTENT(IN)                                :: groupid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_l', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      numtask = 1
      taskid = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(groupid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_environ_l

! **************************************************************************************************
!> \brief ...
!> \param numtask ...
!> \param dims ...
!> \param task_coor ...
!> \param groupid ...
! **************************************************************************************************
   SUBROUTINE mp_environ_c(numtask, dims, task_coor, groupid)

      INTEGER, INTENT(OUT)                     :: numtask, dims(2), &
                                                  task_coor(2)
      INTEGER, INTENT(IN)                      :: groupid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_c', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)
      numtask = 1
      task_coor = 0
      dims = 1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(groupid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_environ_c

! **************************************************************************************************
!> \brief ...
!> \param comm ...
!> \param ndims ...
!> \param dims ...
!> \param task_coor ...
!> \param periods ...
! **************************************************************************************************
   SUBROUTINE mp_environ_c2(comm, ndims, dims, task_coor, periods)

      INTEGER, INTENT(IN)                                :: comm, ndims
      INTEGER, INTENT(OUT)                               :: dims(ndims), task_coor(ndims)
      LOGICAL, INTENT(out)                               :: periods(ndims)

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_c2', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      task_coor = 0
      dims = 1
      periods = .FALSE.
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_environ_c2

!..mp_cart_create
! **************************************************************************************************
!> \brief ...
!> \param comm_old ...
!> \param ndims ...
!> \param dims ...
!> \param pos ...
!> \param comm_cart ...
! **************************************************************************************************
   SUBROUTINE mp_cart_create(comm_old, ndims, dims, pos, comm_cart)

      INTEGER, INTENT(IN)                      :: comm_old, ndims
      INTEGER, INTENT(INOUT)                   :: dims(:)
      INTEGER, INTENT(OUT)                     :: pos(:), comm_cart

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_create', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr, nodes

      ierr = 0
      CALL mp_timeset(routineN, handle)

      nodes = 0
      pos(1:ndims) = -1
      comm_cart = comm_old
      pos(1:ndims) = 0
      dims = 1
      comm_cart = 0
      CALL mp_timestop(handle)

   END SUBROUTINE mp_cart_create

!..mp_cart_coords
! **************************************************************************************************
!> \brief ...
!> \param comm ...
!> \param rank ...
!> \param coords ...
! **************************************************************************************************
   SUBROUTINE mp_cart_coords(comm, rank, coords)

      INTEGER, INTENT(IN)                                :: comm, rank
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: coords

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_coords', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, m

      ierr = 0
      CALL mp_timeset(routineN, handle)

      m = SIZE(coords)
      coords = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rank))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_cart_coords

!..mp_comm_compare
! **************************************************************************************************
!> \brief ...
!> \param comm1 ...
!> \param comm2 ...
!> \param res ...
! **************************************************************************************************
   SUBROUTINE mp_comm_compare(comm1, comm2, res)

      INTEGER, INTENT(IN)                                :: comm1, comm2
      INTEGER, INTENT(OUT)                               :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_compare', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, iout

      ierr = 0
      CALL mp_timeset(routineN, handle)

      iout = 0
      res = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm2))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_comm_compare

!..mp_cart_sub
! **************************************************************************************************
!> \brief ...
!> \param comm ...
!> \param rdim ...
!> \param sub_comm ...
! **************************************************************************************************
   SUBROUTINE mp_cart_sub(comm, rdim, sub_comm)

      INTEGER, INTENT(IN)                                :: comm
      LOGICAL, DIMENSION(:), INTENT(IN)                  :: rdim
      INTEGER, INTENT(OUT)                               :: sub_comm

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_sub', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      sub_comm = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdim))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_cart_sub

!..mp_comm_free
! **************************************************************************************************
!> \brief ...
!> \param comm ...
! **************************************************************************************************
   SUBROUTINE mp_comm_free(comm)

      INTEGER, INTENT(INOUT)                             :: comm

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_free', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_comm_free

!..mp_comm_dup
! **************************************************************************************************
!> \brief ...
!> \param comm1 ...
!> \param comm2 ...
! **************************************************************************************************
   SUBROUTINE mp_comm_dup(comm1, comm2)

      INTEGER, INTENT(IN)                                :: comm1
      INTEGER, INTENT(OUT)                               :: comm2

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_dup', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      comm2 = comm1
      CALL mp_timestop(handle)

   END SUBROUTINE mp_comm_dup

!..mp_rank_compare
! **************************************************************************************************
!> \brief ...
!> \param comm1 ...
!> \param comm2 ...
!> \param rank ...
! **************************************************************************************************
   SUBROUTINE mp_rank_compare(comm1, comm2, rank)

      INTEGER, INTENT(IN)                      :: comm1, comm2
      INTEGER, DIMENSION(:), INTENT(OUT)       :: rank

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_rank_compare', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      rank = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm2))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_rank_compare

!..mp_dims_create
! **************************************************************************************************
!> \brief ...
!> \param nodes ...
!> \param dims ...
! **************************************************************************************************
   SUBROUTINE mp_dims_create(nodes, dims)

      INTEGER, INTENT(IN)                                :: nodes
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: dims

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_dims_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, ndim

      ierr = 0
      CALL mp_timeset(routineN, handle)

      ndim = SIZE(dims)
      dims = 1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nodes))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_dims_create

!..mp_cart_rank
! **************************************************************************************************
!> \brief ...
!> \param group ...
!> \param pos ...
!> \param rank ...
! **************************************************************************************************
   SUBROUTINE mp_cart_rank(group, pos, rank)
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(:), INTENT(IN)                  :: pos
      INTEGER, INTENT(OUT)                               :: rank

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_rank', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      rank = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pos))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_cart_rank

! **************************************************************************************************
!> \brief waits for completion of the given request
!> \param request ...
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
!> \note
!>      see isendrecv
! **************************************************************************************************
   SUBROUTINE mp_wait(request)
      INTEGER, INTENT(inout)                             :: request

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_wait', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_wait

! **************************************************************************************************
!> \brief waits for completion of the given requests
!> \param requests ...
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
!> \note
!>      see isendrecv
! **************************************************************************************************
   SUBROUTINE mp_waitall_1(requests)
      INTEGER, DIMENSION(:), INTENT(inout)     :: requests

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_waitall_1', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requests))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_waitall_1

! **************************************************************************************************
!> \brief waits for completion of the given requests
!> \param requests ...
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
! **************************************************************************************************
   SUBROUTINE mp_waitall_2(requests)
      INTEGER, DIMENSION(:, :), INTENT(inout)  :: requests

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_waitall_2', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requests))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_waitall_2

! **************************************************************************************************
!> \brief wrapper needed to deal with interfaces as present in openmpi 1.8.1
!>        the issue is with the rank or requests
!> \param count ...
!> \param array_of_requests ...
!> \param array_of_statuses ...
!> \param ierr ...
!> \author Joost VandeVondele
! **************************************************************************************************

! **************************************************************************************************
!> \brief waits for completion of any of the given requests
!> \param requests ...
!> \param completed ...
!> \par History
!>      09.2008 created
!> \author Iain Bethune (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE mp_waitany(requests, completed)
      INTEGER, DIMENSION(:), INTENT(inout)     :: requests
      INTEGER, INTENT(out)                     :: completed

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_waitany', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requests))==-1) EXIT ;  END DO ; ENDIF
      completed = 1
      CALL mp_timestop(handle)
   END SUBROUTINE mp_waitany

! **************************************************************************************************
!> \brief Tests for completion of the given requests.
!> \brief We use mpi_test so that we can use a single status.
!> \param requests ...
!> \retval flag ...
!> \par History
!>      24.2016 created
!> \author Alfio Lazzaro
! **************************************************************************************************
   FUNCTION mp_testall_t2(requests) RESULT(flag)
      INTEGER, DIMENSION(2), INTENT(inout)  :: requests
      LOGICAL                               :: flag

      INTEGER                               :: ierr


      ierr = 0
      flag = .TRUE.

      requests = mp_request_null
   END FUNCTION mp_testall_t2

! **************************************************************************************************
!> \brief tests for completion of the given requests
!> \param requests ...
!> \param completed ...
!> \param flag ...
!> \par History
!>      08.2011 created
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE mp_testany_1(requests, completed, flag)
      INTEGER, DIMENSION(:), INTENT(inout)  :: requests
      INTEGER, INTENT(out), OPTIONAL           :: completed
      LOGICAL, INTENT(out), OPTIONAL           :: flag

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_testany_1', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: ierr

      ierr = 0

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requests))==-1) EXIT ;  END DO ; ENDIF
      IF (PRESENT(completed)) completed = 1
      IF (PRESENT(flag)) flag = .TRUE.
   END SUBROUTINE mp_testany_1

! **************************************************************************************************
!> \brief tests for completion of the given requests
!> \param requests ...
!> \param completed ...
!> \param flag ...
!> \par History
!>      08.2011 created
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE mp_testany_2(requests, completed, flag)
      INTEGER, DIMENSION(:, :), INTENT(inout)   :: requests
      INTEGER, INTENT(out), OPTIONAL           :: completed
      LOGICAL, INTENT(out), OPTIONAL           :: flag

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_testany_2', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: ierr

      ierr = 0

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requests))==-1) EXIT ;  END DO ; ENDIF
      IF (PRESENT(completed)) completed = 1
      IF (PRESENT(flag)) flag = .TRUE.
   END SUBROUTINE mp_testany_2

! **************************************************************************************************
!> \brief wrapper needed to deal with interfaces as present in openmpi 1.8.1
!>        the issue is with the rank or requests
!> \param count ...
!> \param array_of_requests ...
!> \param index ...
!> \param flag ...
!> \param status ...
!> \param ierr ...
!> \author Joost VandeVondele
! **************************************************************************************************

! **************************************************************************************************
!> \brief the direct way to split a communicator each color is a sub_comm,
!>        the rank order is accoring to the order in the orig comm
!> \param comm ...
!> \param sub_comm ...
!> \param color ...
!> \param key ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE mp_comm_split_direct(comm, sub_comm, color, key)
      INTEGER, INTENT(in)                                :: comm
      INTEGER, INTENT(OUT)                               :: sub_comm
      INTEGER, INTENT(in)                                :: color
      INTEGER, INTENT(in), OPTIONAL                      :: key

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_split_direct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, my_key

      ierr = 0
      CALL mp_timeset(routineN, handle)

      my_key = 0
      CALL mp_comm_dup(comm, sub_comm)
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(color))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(key))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_comm_split_direct
! **************************************************************************************************
!> \brief splits the given communicator in group in subgroups trying to organize
!>      them in a way that the communication within each subgroup is
!>      efficent (but not necessarily the comunication between subgroups)
!> \param comm the mpi communicator that you want to split
!> \param sub_comm the communicator for the subgroup (created, needs to be freed later)
!> \param ngroups actual number of groups
!> \param group_distribution input  : allocated with array with the nprocs entries (0 .. nprocs-1)
!> \param subgroup_min_size the minimum size of the subgroup
!> \param n_subgroups the number of subgroups wanted
!> \param group_partition n_subgroups sized array containing the number of cpus wanted per group.
!>                         should match the total number of cpus (only used if present and associated) (0..ngroups-1)
!> \param stride create groups using a stride (default=1) through the ranks of the comm to be split.
!> \par History
!>      10.2003 created [fawzi]
!>      02.2004 modified [Joost VandeVondele]
!> \author Fawzi Mohamed
!> \note
!>      at least one of subgroup_min_size and n_subgroups is needed,
!>      the other default to the value needed to use most processors.
!>      if less cpus are present than needed for subgroup min size, n_subgroups,
!>      just one comm is created that contains all cpus
! **************************************************************************************************
   SUBROUTINE mp_comm_split(comm, sub_comm, ngroups, group_distribution, &
                            subgroup_min_size, n_subgroups, group_partition, stride)
      INTEGER, INTENT(in)                      :: comm
      INTEGER, INTENT(out)                     :: sub_comm, ngroups
      INTEGER, DIMENSION(:), POINTER           :: group_distribution
      INTEGER, INTENT(in), OPTIONAL            :: subgroup_min_size, n_subgroups
      INTEGER, DIMENSION(:), OPTIONAL, POINTER :: group_partition
      INTEGER, OPTIONAL                        :: stride

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_split', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr, mepos, nnodes

      ierr = 0
      CALL mp_timeset(routineN, handle)

      ! actual number of groups

      IF (.NOT. PRESENT(subgroup_min_size) .AND. .NOT. PRESENT(n_subgroups)) THEN
         CALL cp__b("mpiwrap/message_passing.F",1957,routineP//" missing arguments")
      ENDIF
      IF (PRESENT(subgroup_min_size) .AND. PRESENT(n_subgroups)) THEN
         CALL cp__b("mpiwrap/message_passing.F",1960,routineP//" too many arguments")
      ENDIF

      CALL mp_environ(nnodes, mepos, comm)

      IF (.NOT. ASSOCIATED(group_distribution)) THEN
         CALL cp__b("mpiwrap/message_passing.F",1966,routineP//" group_distribution not associated")
      ENDIF
      IF (LBOUND(group_distribution, 1) .NE. 0 .OR. &
          UBOUND(group_distribution, 1) .NE. nnodes-1) THEN
         CALL cp__b("mpiwrap/message_passing.F",1970,routineP//" group_distribution wrong bounds")
      ENDIF

      CALL mp_comm_dup(comm, sub_comm)
      group_distribution(0) = 0
      ngroups = 1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stride))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group_partition))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)

   END SUBROUTINE mp_comm_split

! **************************************************************************************************
!> \brief probes for an incomming message with any tag
!> \param[inout] source the source of the possible incomming message,
!>        if MP_ANY_SOURCE it is a blocking one and return value is the source
!>        of the next incomming message
!>        if source is a different value it is a non-blocking probe retuning
!>        MP_ANY_SOURCE if there is no incomming message
!> \param[in] comm the communicator
!> \param[out] tag the tag of the incomming message
!> \author Mandes
! **************************************************************************************************
   SUBROUTINE mp_probe(source, comm, tag)
      INTEGER                                  :: source
      INTEGER, INTENT(IN)                      :: comm
      INTEGER, INTENT(OUT)                     :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_probe', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

!   ---------------------------------------------------------------------------

      CALL mp_timeset(routineN, handle)

      ierr = 0
      tag = -1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_probe

! **************************************************************************************************
! Here come the data routines with none of the standard data types.
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param msg ...
!> \param source ...
!> \param gid ...
! **************************************************************************************************
   SUBROUTINE mp_bcast_b(msg, source, gid)
      LOGICAL                                            :: msg
      INTEGER                                            :: source, gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_b', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, msglen

      ierr = 0
      CALL mp_timeset(routineN, handle)

      msglen = 1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_bcast_b

! **************************************************************************************************
!> \brief ...
!> \param msg ...
!> \param source ...
!> \param gid ...
! **************************************************************************************************
   SUBROUTINE mp_bcast_bv(msg, source, gid)
      LOGICAL                                            :: msg(:)
      INTEGER                                            :: source, gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_bv', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, msglen

      ierr = 0
      CALL mp_timeset(routineN, handle)

      msglen = SIZE(msg)
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_bcast_bv

! **************************************************************************************************
!> \brief ...
!> \param msg ...
!> \param source ...
!> \param gid ...
! **************************************************************************************************
   SUBROUTINE mp_bcast_av(msg, source, gid)
      CHARACTER(LEN=*)                         :: msg
      INTEGER                                  :: source, gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_av', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_bcast_av

! **************************************************************************************************
!> \brief ...
!> \param msg ...
!> \param source ...
!> \param gid ...
! **************************************************************************************************
   SUBROUTINE mp_bcast_am(msg, source, gid)
      CHARACTER(LEN=*)                         :: msg(:)
      INTEGER                                  :: source, gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_am', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_bcast_am

! **************************************************************************************************
!> \brief Finds the location of the minimal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MINLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_minloc_dv(msg, gid)
      REAL(kind=real_8), INTENT(INOUT)         :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_minloc_dv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("d" .EQ. "l" .AND. real_8 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2315,"Minimal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_minloc_dv

! **************************************************************************************************
!> \brief Finds the location of the minimal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MINLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_minloc_iv(msg, gid)
      INTEGER(KIND=int_4), INTENT(INOUT)       :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_minloc_iv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("i" .EQ. "l" .AND. int_4 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2363,"Minimal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_minloc_iv

! **************************************************************************************************
!> \brief Finds the location of the minimal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MINLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_minloc_lv(msg, gid)
      INTEGER(KIND=int_8), INTENT(INOUT)       :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_minloc_lv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("l" .EQ. "l" .AND. int_8 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2409,"Minimal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_minloc_lv

! **************************************************************************************************
!> \brief Finds the location of the minimal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MINLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_minloc_rv(msg, gid)
      REAL(kind=real_4), INTENT(INOUT)         :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_minloc_rv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("r" .EQ. "l" .AND. real_4 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2455,"Minimal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_minloc_rv

! **************************************************************************************************
!> \brief Finds the location of the maximal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MAXLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_maxloc_dv(msg, gid)
      REAL(kind=real_8), INTENT(INOUT)         :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_maxloc_dv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("d" .EQ. "l" .AND. real_8 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2501,"Maximal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_maxloc_dv

! **************************************************************************************************
!> \brief Finds the location of the maximal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MAXLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_maxloc_iv(msg, gid)
      INTEGER(KIND=int_4), INTENT(INOUT)       :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_maxloc_iv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("i" .EQ. "l" .AND. int_4 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2547,"Maximal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_maxloc_iv

! **************************************************************************************************
!> \brief Finds the location of the maximal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MAXLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_maxloc_lv(msg, gid)
      INTEGER(KIND=int_8), INTENT(INOUT)       :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_maxloc_lv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("l" .EQ. "l" .AND. int_8 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2593,"Maximal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_maxloc_lv

! **************************************************************************************************
!> \brief Finds the location of the maximal element in a vector.
!> \param[in,out] msg         Find location of maximum element among these
!>                            data (input).
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce with the MPI_MAXLOC reduction function identifier
!> \par Invalid data types
!>      This routine is invalid for (int_8) data!
! **************************************************************************************************
   SUBROUTINE mp_maxloc_rv(msg, gid)
      REAL(kind=real_4), INTENT(INOUT)         :: msg(:)
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_maxloc_rv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr

      ierr = 0
      IF ("r" .EQ. "l" .AND. real_4 .EQ. int_8) THEN
         CALL cp__b("mpiwrap/message_passing.F",2639,"Maximal location not available with long integers @ "//routineN)
      ENDIF
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_maxloc_rv

! **************************************************************************************************
!> \brief Logical OR reduction
!> \param[in,out] msg         Datum to perform inclusive disjunction (input)
!>                            and resultant inclusive disjunction (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! **************************************************************************************************
   SUBROUTINE mp_sum_b(msg, gid)
      LOGICAL, INTENT(INOUT)                   :: msg
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_b', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr, msglen

      CALL mp_timeset(routineN, handle)
      ierr = 0
      msglen = 1
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_sum_b

! **************************************************************************************************
!> \brief Logical OR reduction
!> \param[in,out] msg         Datum to perform inclusive disjunction (input)
!>                            and resultant inclusive disjunction (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! **************************************************************************************************
   SUBROUTINE mp_sum_bv(msg, gid)
      LOGICAL, DIMENSION(:), INTENT(INOUT)     :: msg
      INTEGER, INTENT(IN)                      :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_bv', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ierr, msglen

      CALL mp_timeset(routineN, handle)
      ierr = 0
      msglen = SIZE(msg)
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_sum_bv

! **************************************************************************************************
!> \brief Opens a file
!> \param[in] groupid    message passing environment identifier
!> \param[out] fh        file handle (file storage unit)
!> \param[in] filepath   path to the file
!> \param amode_status   access mode
!> \param info ...
!> \par MPI-I/O mapping  mpi_file_open
!> \par STREAM-I/O mapping  OPEN
!>
!> \param[in](optinal) info   info object
!> \par History
!>      11.2012 created [Hossein Bani-Hashemian]
! **************************************************************************************************
   SUBROUTINE mp_file_open(groupid, fh, filepath, amode_status, info)
      INTEGER, INTENT(IN)                      :: groupid
      INTEGER, INTENT(OUT)                     :: fh
      CHARACTER(len=*), INTENT(IN)             :: filepath
      INTEGER, INTENT(IN)                      :: amode_status
      INTEGER, INTENT(IN), OPTIONAL            :: info

      INTEGER                                  :: ierr, istat
      CHARACTER(LEN=10)                        :: fstatus
      LOGICAL                                  :: exists, is_open

      ierr = 0
      istat = 0
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(groupid))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(info))==-1) EXIT ;  END DO ; ENDIF
      IF ((amode_status .EQ. file_amode_create) .OR. &
          (amode_status .EQ. file_amode_create+file_amode_wronly) .OR. &
          (amode_status .EQ. file_amode_create+file_amode_wronly+file_amode_excl)) THEN
         fstatus = "UNKNOWN"
      ELSE
         fstatus = "OLD"
      END IF
      ! Get a new unit number
      DO fh = 1, 999
         INQUIRE (UNIT=fh, EXIST=exists, OPENED=is_open, IOSTAT=istat)
         IF (exists .AND. (.NOT. is_open) .AND. (istat == 0)) EXIT
      END DO
      OPEN (UNIT=fh, FILE=filepath, STATUS=fstatus, ACCESS="STREAM")
   END SUBROUTINE mp_file_open

! **************************************************************************************************
!> \brief Closes a file
!> \param[in] fh   file handle (file storage unit)
!> \par MPI-I/O mapping   mpi_file_close
!> \par STREAM-I/O mapping   CLOSE
!>
!> \par History
!>      11.2012 created [Hossein Bani-Hashemian]
! **************************************************************************************************
   SUBROUTINE mp_file_close(fh)
      INTEGER, INTENT(INOUT)                             :: fh

      INTEGER                                            :: ierr

      ierr = 0
      CLOSE (fh)
   END SUBROUTINE mp_file_close

! **************************************************************************************************
!> \brief Returns the file size
!> \param[in] fh file handle (file storage unit)
!> \param[out] file_size  the file size
!> \par MPI-I/O mapping   mpi_file_get_size
!> \par STREAM-I/O mapping   INQUIRE
!>
!> \par History
!>      12.2012 created [Hossein Bani-Hashemian]
! **************************************************************************************************
   SUBROUTINE mp_file_get_size(fh, file_size)
      INTEGER, INTENT(IN)                                :: fh
      INTEGER(kind=file_offset), INTENT(OUT)             :: file_size

      INTEGER                                            :: ierr

      ierr = 0
      INQUIRE (UNIT=fh, SIZE=file_size)
   END SUBROUTINE mp_file_get_size

! **************************************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! **************************************************************************************************
   SUBROUTINE mp_file_write_at_chv(fh, offset, msg, msglen)
      CHARACTER, INTENT(IN)                      :: msg(:)
      INTEGER, INTENT(IN)                        :: fh
      INTEGER, INTENT(IN), OPTIONAL              :: msglen
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_chv', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msglen))==-1) EXIT ;  END DO ; ENDIF
      WRITE (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_write_at_chv

! **************************************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! **************************************************************************************************
   SUBROUTINE mp_file_write_at_ch(fh, offset, msg)
      CHARACTER(LEN=*), INTENT(IN)               :: msg
      INTEGER, INTENT(IN)                        :: fh
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_ch', &
                                     routineP = moduleN//':'//routineN

      WRITE (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_write_at_ch

! **************************************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! **************************************************************************************************
   SUBROUTINE mp_file_write_at_all_chv(fh, offset, msg, msglen)
      CHARACTER, INTENT(IN)                      :: msg(:)
      INTEGER, INTENT(IN)                        :: fh
      INTEGER, INTENT(IN), OPTIONAL              :: msglen
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_chv', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msglen))==-1) EXIT ;  END DO ; ENDIF
      WRITE (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_write_at_all_chv

! **************************************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! **************************************************************************************************
   SUBROUTINE mp_file_write_at_all_ch(fh, offset, msg)
      CHARACTER(LEN=*), INTENT(IN)               :: msg
      INTEGER, INTENT(IN)                        :: fh
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_ch', &
                                     routineP = moduleN//':'//routineN

      WRITE (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_write_at_all_ch

! **************************************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! **************************************************************************************************
   SUBROUTINE mp_file_read_at_chv(fh, offset, msg, msglen)
      CHARACTER, INTENT(OUT)                     :: msg(:)
      INTEGER, INTENT(IN)                        :: fh
      INTEGER, INTENT(IN), OPTIONAL              :: msglen
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_chv', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msglen))==-1) EXIT ;  END DO ; ENDIF
      READ (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_read_at_chv

! **************************************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! **************************************************************************************************
   SUBROUTINE mp_file_read_at_ch(fh, offset, msg)
      CHARACTER(LEN=*), INTENT(OUT)              :: msg
      INTEGER, INTENT(IN)                        :: fh
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_ch', &
                                     routineP = moduleN//':'//routineN

      READ (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_read_at_ch

! **************************************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! **************************************************************************************************
   SUBROUTINE mp_file_read_at_all_chv(fh, offset, msg, msglen)
      CHARACTER, INTENT(OUT)                     :: msg(:)
      INTEGER, INTENT(IN)                        :: fh
      INTEGER, INTENT(IN), OPTIONAL              :: msglen
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_chv', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msglen))==-1) EXIT ;  END DO ; ENDIF
      READ (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_read_at_all_chv

! **************************************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! **************************************************************************************************
   SUBROUTINE mp_file_read_at_all_ch(fh, offset, msg)
      CHARACTER(LEN=*), INTENT(OUT)              :: msg
      INTEGER, INTENT(IN)                        :: fh
      INTEGER(kind=file_offset), INTENT(IN)      :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_ch', &
                                     routineP = moduleN//':'//routineN

      READ (UNIT=fh, POS=offset+1) msg
   END SUBROUTINE mp_file_read_at_all_ch

! **************************************************************************************************
!> \brief Returns the size of a data type in bytes
!> \param[in] type_descriptor  data type
!> \param[out] type_size       size of the data type
!> \par MPI mapping
!>      mpi_type_size
!>
! **************************************************************************************************
   SUBROUTINE mp_type_size(type_descriptor, type_size)
      TYPE(mp_type_descriptor_type), INTENT(IN)          :: type_descriptor
      INTEGER, INTENT(OUT)                               :: type_size

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_size', routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr

      ierr = 0
      SELECT CASE (type_descriptor%type_handle)
      CASE (1)
         type_size = real_4_size
      CASE (3)
         type_size = real_8_size
      CASE (5)
         type_size = 2*real_4_size
      CASE (7)
         type_size = 2*real_8_size
      END SELECT
   END SUBROUTINE mp_type_size

! **************************************************************************************************
!> \brief ...
!> \param subtypes ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! **************************************************************************************************
   FUNCTION mp_type_make_struct(subtypes, &
                                vector_descriptor, index_descriptor) &
      RESULT(type_descriptor)
      TYPE(mp_type_descriptor_type), &
         DIMENSION(:), INTENT(IN)               :: subtypes
      INTEGER, DIMENSION(2), INTENT(IN), &
         OPTIONAL                               :: vector_descriptor
      TYPE(mp_indexing_meta_type), &
         INTENT(IN), OPTIONAL                   :: index_descriptor
      TYPE(mp_type_descriptor_type)            :: type_descriptor

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_struct', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: i, ierr, n
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: lengths, old_types

      ierr = 0
      n = SIZE(subtypes)
      !type_descriptor%mpi_type_handle = MPI_DATATYPE_NULL
      type_descriptor%length = 1
      type_descriptor%vector_descriptor(1:2) = 1
      type_descriptor%has_indexing = .FALSE.
      ALLOCATE (type_descriptor%subtype(n))
      type_descriptor%subtype(:) = subtypes(:)
      ALLOCATE (lengths(n), old_types(n))
      DO i = 1, SIZE(subtypes)
         old_types(i) = subtypes(i)%type_handle
         lengths(i) = subtypes(i)%length
      ENDDO
      IF (PRESENT(vector_descriptor) .OR. PRESENT(index_descriptor)) THEN
         CALL cp__b("mpiwrap/message_passing.F",3166,routineN//" Vectors and indices NYI")
      ENDIF
   END FUNCTION mp_type_make_struct

! **************************************************************************************************
!> \brief ...
!> \param type_descriptor ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE mp_type_free_m(type_descriptor)
      TYPE(mp_type_descriptor_type), INTENT(inout)       :: type_descriptor

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_free_m', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ierr

      CALL mp_timeset(routineN, handle)
      ierr = 0

      ! If the subtype is associated, then it's a user-defined data type.

      IF (ASSOCIATED(type_descriptor%subtype)) THEN
         DO i = 1, SIZE(type_descriptor%subtype)
            CALL mp_type_free_m(type_descriptor%subtype(i))
         ENDDO
         DEALLOCATE (type_descriptor%subtype)
      ENDIF

      CALL mp_timestop(handle)

   END SUBROUTINE mp_type_free_m

! **************************************************************************************************
!> \brief ...
!> \param type_descriptors ...
! **************************************************************************************************
   SUBROUTINE mp_type_free_v(type_descriptors)
      TYPE(mp_type_descriptor_type), DIMENSION(:), &
         INTENT(inout)                                   :: type_descriptors

      INTEGER                                            :: i

      DO i = 1, SIZE(type_descriptors)
         CALL mp_type_free(type_descriptors(i))
      ENDDO

   END SUBROUTINE mp_type_free_v

! **************************************************************************************************
!> \brief Non-blocking send of custom type
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
! **************************************************************************************************
   SUBROUTINE mp_isend_custom(msgin, dest, comm, request, tag)
      TYPE(mp_type_descriptor_type), INTENT(IN)          :: msgin
      INTEGER, INTENT(IN)                                :: dest, comm
      INTEGER, INTENT(out)                               :: request
      INTEGER, INTENT(in), OPTIONAL                      :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_custom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr, my_tag

      ierr = 0
      my_tag = 0

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
      ierr = 1
      request = 0
      CALL mp_stop(ierr, "mp_isend called in non parallel case")
   END SUBROUTINE mp_isend_custom

! **************************************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
! **************************************************************************************************
   SUBROUTINE mp_irecv_custom(msgout, source, comm, request, tag)
      TYPE(mp_type_descriptor_type), INTENT(INOUT)       :: msgout
      INTEGER, INTENT(IN)                                :: source, comm
      INTEGER, INTENT(out)                               :: request
      INTEGER, INTENT(in), OPTIONAL                      :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_custom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr, my_tag

      ierr = 0
      my_tag = 0

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
      ierr = 1
      request = 0
      CALL cp__b("mpiwrap/message_passing.F",3294,"mp_irecv called in non parallel case")
   END SUBROUTINE mp_irecv_custom

! **************************************************************************************************
!> \brief Window free
!> \param win ...
! **************************************************************************************************
   SUBROUTINE mp_win_free(win)
      INTEGER, INTENT(INOUT)                             :: win

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_free', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
      win = mp_win_null
      CALL mp_timestop(handle)
   END SUBROUTINE mp_win_free

! **************************************************************************************************
!> \brief Window lock
!> \param win ...
! **************************************************************************************************
   SUBROUTINE mp_win_lock_all(win)
      INTEGER, INTENT(INOUT)                             :: win

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_lock_all', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_win_lock_all

! **************************************************************************************************
!> \brief Window lock
!> \param win ...
! **************************************************************************************************
   SUBROUTINE mp_win_unlock_all(win)
      INTEGER, INTENT(INOUT)                             :: win

      CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_unlock_all', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr

      ierr = 0
      CALL mp_timeset(routineN, handle)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
      CALL mp_timestop(handle)
   END SUBROUTINE mp_win_unlock_all

! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_im(msg, group, displ_in)

    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_im


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_i(msg, group, displ_in)

    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_i

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_i11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER(KIND=int_4), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    INTEGER(KIND=int_4), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_i11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_i22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER(KIND=int_4), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    INTEGER(KIND=int_4), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_i ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER(KIND=int_4), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
! *****************************************************************************
  SUBROUTINE mp_alltoall_i22 ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:, :), INTENT(IN)     :: sb
    INTEGER(KIND=int_4), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
! *****************************************************************************
  SUBROUTINE mp_alltoall_i33 ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:, :, :), INTENT(IN)  :: sb
    INTEGER(KIND=int_4), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
! *****************************************************************************
  SUBROUTINE mp_alltoall_i44 ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_i45 ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_4), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_i34 ( sb, rb, count, group )

    INTEGER(KIND=int_4), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_i
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_i54 ( sb, rb, count, group )

    INTEGER(KIND=int_4), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    INTEGER(KIND=int_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_i54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_i(msg,dest,tag,gid)
    INTEGER(KIND=int_4)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_i

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_i
! *****************************************************************************
  SUBROUTINE mp_send_iv(msg,dest,tag,gid)
    INTEGER(KIND=int_4)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_iv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_i(msg,source,tag,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_i

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_i
! *****************************************************************************
  SUBROUTINE mp_recv_iv(msg,source,tag,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_iv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_i(msg,source,gid)
    INTEGER(KIND=int_4)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_i

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_i(msg,source,gid,request)
    INTEGER(KIND=int_4)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_i

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_i1
! *****************************************************************************
  SUBROUTINE mp_bcast_iv(msg,source,gid)
    INTEGER(KIND=int_4)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_iv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_i1
! *****************************************************************************
  SUBROUTINE mp_ibcast_iv(msg,source,gid,request)
    INTEGER(KIND=int_4)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_iv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_i1
! *****************************************************************************
  SUBROUTINE mp_bcast_im(msg,source,gid)
    INTEGER(KIND=int_4)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_im

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_i1
! *****************************************************************************
  SUBROUTINE mp_bcast_i3(msg,source,gid)
    INTEGER(KIND=int_4)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_i3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_i3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_i(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_i

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_i
! *****************************************************************************
  SUBROUTINE mp_sum_iv(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_iv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_i
! *****************************************************************************
  SUBROUTINE mp_isum_iv(msg,gid,request)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_iv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_i
! *****************************************************************************
  SUBROUTINE mp_sum_im(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_im

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_i
! *****************************************************************************
  SUBROUTINE mp_sum_im3(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_im3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_im3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_i
! *****************************************************************************
  SUBROUTINE mp_sum_im4(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_im4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_im4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_iv(msg,root,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_iv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_iv
! *****************************************************************************
  SUBROUTINE mp_sum_root_im(msg,root,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_im

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_im(msg,res,gid)
    INTEGER(KIND=int_4), INTENT(IN)         :: msg( :, : )
    INTEGER(KIND=int_4), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_im'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_im

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_i(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_i

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_i
! *****************************************************************************
  SUBROUTINE mp_max_iv(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_iv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_i(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_i

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_i
! *****************************************************************************
  SUBROUTINE mp_min_iv(msg,gid)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_iv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_iv(msg_scatter,msg,root,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_iv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_i(msg_scatter,msg,root,gid,request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_i

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_iv(msg_scatter,msg,root,gid,request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_iv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_iv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    INTEGER(KIND=int_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_iv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_i(msg,msg_gather,root,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_i

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_i
! *****************************************************************************
  SUBROUTINE mp_gather_iv(msg,msg_gather,root,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg( : )
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_iv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_i
! *****************************************************************************
  SUBROUTINE mp_gather_im(msg,msg_gather,root,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg( :, : )
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_im

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_iv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    INTEGER(KIND=int_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    INTEGER(KIND=int_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_iv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_iv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    INTEGER(KIND=int_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    INTEGER(KIND=int_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_iv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_i(msgout,msgin,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_i

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_i(msgout,msgin,gid,request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_i

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_i12(msgout, msgin,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_i12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_i12
! *****************************************************************************
  SUBROUTINE mp_allgather_i23(msgout, msgin,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:,:)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_i23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_i12
! *****************************************************************************
  SUBROUTINE mp_allgather_i34(msgout, msgin,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:,:, :)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_i34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_i12
! *****************************************************************************
  SUBROUTINE mp_allgather_i22(msgout, msgin,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:, :)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_i22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_i12
! *****************************************************************************
  SUBROUTINE mp_iallgather_i22(msgout, msgin, gid, request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:, :)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_i22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_i22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_i12
! *****************************************************************************
  SUBROUTINE mp_iallgather_i33(msgout, msgin, gid, request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout(:, :, :)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_i33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_i33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_iv(msgout,msgin,rcount,rdispl,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_iv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_iv(msgout,msgin,rcount,rdispl,gid,request)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_iv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_iv(msgout,msgin,rcount,gid)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_iv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_iv(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_iv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_iv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_im2(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_im2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_im2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_iv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_im3(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_im3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_im3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_iv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_im4(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_im4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_im4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_i(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER(KIND=int_4)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_i

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_iv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_iv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_iv(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_iv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_iv
!> \note see mp_isend_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_im2(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_im2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_im2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_iv
!> \note see mp_isend_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_im3(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_im3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_im3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_iv(msgout,source,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_iv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_iv
!> \note see mp_irecv_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_im2(msgout,source,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_im2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_im2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_iv
!> \note see mp_irecv_iv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_im3(msgout,source,comm,request,tag)
    INTEGER(KIND=int_4), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_im3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_im3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_iv(base,comm,win)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_iv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_iv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    INTEGER(KIND=int_4), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_iv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_i(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_i', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 17
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_i

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_i(DATA, len, stat)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_i', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_i

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_i(DATA, stat)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_i', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_i

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_iv(fh, offset, msg, msglen)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_iv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_iv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_i(fh, offset, msg)
    INTEGER(KIND=int_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_i', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_i

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_iv(fh, offset, msg, msglen)
    INTEGER(KIND=int_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_iv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_iv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_i(fh, offset, msg)
    INTEGER(KIND=int_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_i', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_i

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_iv(fh, offset, msg, msglen)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_iv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_iv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_i(fh, offset, msg)
    INTEGER(KIND=int_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_i', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_i

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_iv(fh, offset, msg, msglen)
    INTEGER(KIND=int_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_iv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_iv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_i(fh, offset, msg)
    INTEGER(KIND=int_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_i', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_i

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_i (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_i', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 17
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_i => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_i

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_i(DATA, len, stat)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_i

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_i(DATA, stat)
    INTEGER(KIND=int_4), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_i
! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_lm(msg, group, displ_in)

    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_lm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_lm


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_l(msg, group, displ_in)

    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_l

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_l11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER(KIND=int_8), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    INTEGER(KIND=int_8), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_l11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_l22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER(KIND=int_8), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    INTEGER(KIND=int_8), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_l ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER(KIND=int_8), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
! *****************************************************************************
  SUBROUTINE mp_alltoall_l22 ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:, :), INTENT(IN)     :: sb
    INTEGER(KIND=int_8), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
! *****************************************************************************
  SUBROUTINE mp_alltoall_l33 ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:, :, :), INTENT(IN)  :: sb
    INTEGER(KIND=int_8), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
! *****************************************************************************
  SUBROUTINE mp_alltoall_l44 ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_l45 ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_8), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_l34 ( sb, rb, count, group )

    INTEGER(KIND=int_8), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    INTEGER(KIND=int_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_l
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_l54 ( sb, rb, count, group )

    INTEGER(KIND=int_8), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    INTEGER(KIND=int_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_l54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_l54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_l(msg,dest,tag,gid)
    INTEGER(KIND=int_8)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_l

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_l
! *****************************************************************************
  SUBROUTINE mp_send_lv(msg,dest,tag,gid)
    INTEGER(KIND=int_8)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_lv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_l(msg,source,tag,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_l

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_l
! *****************************************************************************
  SUBROUTINE mp_recv_lv(msg,source,tag,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_lv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_l(msg,source,gid)
    INTEGER(KIND=int_8)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_l

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_l(msg,source,gid,request)
    INTEGER(KIND=int_8)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_l

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_l1
! *****************************************************************************
  SUBROUTINE mp_bcast_lv(msg,source,gid)
    INTEGER(KIND=int_8)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_lv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_l1
! *****************************************************************************
  SUBROUTINE mp_ibcast_lv(msg,source,gid,request)
    INTEGER(KIND=int_8)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_lv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_l1
! *****************************************************************************
  SUBROUTINE mp_bcast_lm(msg,source,gid)
    INTEGER(KIND=int_8)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_lm

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_l1
! *****************************************************************************
  SUBROUTINE mp_bcast_l3(msg,source,gid)
    INTEGER(KIND=int_8)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_l3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_l3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_l(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_l

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_l
! *****************************************************************************
  SUBROUTINE mp_sum_lv(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_lv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_l
! *****************************************************************************
  SUBROUTINE mp_isum_lv(msg,gid,request)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_lv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_l
! *****************************************************************************
  SUBROUTINE mp_sum_lm(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_lm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_lm

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_l
! *****************************************************************************
  SUBROUTINE mp_sum_lm3(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_lm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_lm3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_l
! *****************************************************************************
  SUBROUTINE mp_sum_lm4(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_lm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_lm4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_lv(msg,root,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_lv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_lv
! *****************************************************************************
  SUBROUTINE mp_sum_root_lm(msg,root,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_lm

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_lm(msg,res,gid)
    INTEGER(KIND=int_8), INTENT(IN)         :: msg( :, : )
    INTEGER(KIND=int_8), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_lm'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_lm

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_l(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_l

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_l
! *****************************************************************************
  SUBROUTINE mp_max_lv(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_lv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_l(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_l

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_l
! *****************************************************************************
  SUBROUTINE mp_min_lv(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_lv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_lv(msg_scatter,msg,root,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_lv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_l(msg_scatter,msg,root,gid,request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_l

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_lv(msg_scatter,msg,root,gid,request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_lv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_lv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    INTEGER(KIND=int_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_lv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_l(msg,msg_gather,root,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_l

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_l
! *****************************************************************************
  SUBROUTINE mp_gather_lv(msg,msg_gather,root,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg( : )
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_lv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_l
! *****************************************************************************
  SUBROUTINE mp_gather_lm(msg,msg_gather,root,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg( :, : )
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_lm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_lm

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_lv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    INTEGER(KIND=int_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    INTEGER(KIND=int_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_lv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_lv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    INTEGER(KIND=int_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    INTEGER(KIND=int_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_lv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_l(msgout,msgin,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_l

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_l(msgout,msgin,gid,request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_l

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_l12(msgout, msgin,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_l12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_l12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_l12
! *****************************************************************************
  SUBROUTINE mp_allgather_l23(msgout, msgin,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:,:)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_l23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_l23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_l12
! *****************************************************************************
  SUBROUTINE mp_allgather_l34(msgout, msgin,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:,:, :)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_l34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_l34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_l12
! *****************************************************************************
  SUBROUTINE mp_allgather_l22(msgout, msgin,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:, :)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_l22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_l22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_l12
! *****************************************************************************
  SUBROUTINE mp_iallgather_l22(msgout, msgin, gid, request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:, :)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_l22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_l22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_l12
! *****************************************************************************
  SUBROUTINE mp_iallgather_l33(msgout, msgin, gid, request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout(:, :, :)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_l33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_l33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_lv(msgout,msgin,rcount,rdispl,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_lv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_lv(msgout,msgin,rcount,rdispl,gid,request)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_lv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_lv(msgout,msgin,rcount,gid)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgout( : )
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_lv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_lv(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_lv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_lv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_lm2(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_lm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_lm2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_lv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_lm3(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_lm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_lm3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_lv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_lm4(msgin,dest,msgout,source,comm)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_lm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_lm4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_l(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER(KIND=int_8)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_l

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_lv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_lv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_lv(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_lv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_lv
!> \note see mp_isend_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_lm2(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_lm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_lm2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_lv
!> \note see mp_isend_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_lm3(msgin,dest,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_lm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_lm3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_lv(msgout,source,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_lv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_lv
!> \note see mp_irecv_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_lm2(msgout,source,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_lm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_lm2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_lv
!> \note see mp_irecv_lv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_lm3(msgout,source,comm,request,tag)
    INTEGER(KIND=int_8), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_lm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_lm3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_lv(base,comm,win)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_lv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_lv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    INTEGER(KIND=int_8), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_lv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_l(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_l', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 19
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_l

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_l(DATA, len, stat)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_l', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_l

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_l(DATA, stat)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_l', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_l

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_lv(fh, offset, msg, msglen)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_lv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_lv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_l(fh, offset, msg)
    INTEGER(KIND=int_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_l', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_l

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_lv(fh, offset, msg, msglen)
    INTEGER(KIND=int_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_lv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_lv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_l(fh, offset, msg)
    INTEGER(KIND=int_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_l', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_l

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_lv(fh, offset, msg, msglen)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_lv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_lv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_l(fh, offset, msg)
    INTEGER(KIND=int_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_l', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_l

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_lv(fh, offset, msg, msglen)
    INTEGER(KIND=int_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_lv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_lv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_l(fh, offset, msg)
    INTEGER(KIND=int_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_l', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_l

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_l (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_l', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 19
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_l => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_l

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_l(DATA, len, stat)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_l

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_l(DATA, stat)
    INTEGER(KIND=int_8), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_l
! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_rm(msg, group, displ_in)

    REAL(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_rm


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_r(msg, group, displ_in)

    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_r

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_r11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_4), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_r11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_r22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    REAL(kind=real_4), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_4), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_r ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: sb
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
! *****************************************************************************
  SUBROUTINE mp_alltoall_r22 ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:, :), INTENT(IN)     :: sb
    REAL(kind=real_4), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
! *****************************************************************************
  SUBROUTINE mp_alltoall_r33 ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:, :, :), INTENT(IN)  :: sb
    REAL(kind=real_4), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
! *****************************************************************************
  SUBROUTINE mp_alltoall_r44 ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_r45 ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_4), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_r34 ( sb, rb, count, group )

    REAL(kind=real_4), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_r
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_r54 ( sb, rb, count, group )

    REAL(kind=real_4), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    REAL(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_r54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_r(msg,dest,tag,gid)
    REAL(kind=real_4)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_r

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_r
! *****************************************************************************
  SUBROUTINE mp_send_rv(msg,dest,tag,gid)
    REAL(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_rv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_r(msg,source,tag,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_r

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_r
! *****************************************************************************
  SUBROUTINE mp_recv_rv(msg,source,tag,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_rv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_r(msg,source,gid)
    REAL(kind=real_4)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_r

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_r(msg,source,gid,request)
    REAL(kind=real_4)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_r

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_r1
! *****************************************************************************
  SUBROUTINE mp_bcast_rv(msg,source,gid)
    REAL(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_rv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_r1
! *****************************************************************************
  SUBROUTINE mp_ibcast_rv(msg,source,gid,request)
    REAL(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_rv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_r1
! *****************************************************************************
  SUBROUTINE mp_bcast_rm(msg,source,gid)
    REAL(kind=real_4)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_rm

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_r1
! *****************************************************************************
  SUBROUTINE mp_bcast_r3(msg,source,gid)
    REAL(kind=real_4)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_r3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_r3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_r(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_r

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_r
! *****************************************************************************
  SUBROUTINE mp_sum_rv(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_rv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_r
! *****************************************************************************
  SUBROUTINE mp_isum_rv(msg,gid,request)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_rv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_r
! *****************************************************************************
  SUBROUTINE mp_sum_rm(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_rm

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_r
! *****************************************************************************
  SUBROUTINE mp_sum_rm3(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_rm3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_r
! *****************************************************************************
  SUBROUTINE mp_sum_rm4(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_rm4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_rv(msg,root,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_rv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_rv
! *****************************************************************************
  SUBROUTINE mp_sum_root_rm(msg,root,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_rm

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_rm(msg,res,gid)
    REAL(kind=real_4), INTENT(IN)         :: msg( :, : )
    REAL(kind=real_4), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_rm'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_rm

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_r(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_r

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_r
! *****************************************************************************
  SUBROUTINE mp_max_rv(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_rv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_r(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_r

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_r
! *****************************************************************************
  SUBROUTINE mp_min_rv(msg,gid)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_rv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_rv(msg_scatter,msg,root,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_4), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_rv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_r(msg_scatter,msg,root,gid,request)
    REAL(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_r

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_rv(msg_scatter,msg,root,gid,request)
    REAL(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_rv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_rv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    REAL(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    REAL(kind=real_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_rv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_r(msg,msg_gather,root,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msg
    REAL(kind=real_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_r

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_r
! *****************************************************************************
  SUBROUTINE mp_gather_rv(msg,msg_gather,root,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msg( : )
    REAL(kind=real_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_rv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_r
! *****************************************************************************
  SUBROUTINE mp_gather_rm(msg,msg_gather,root,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msg( :, : )
    REAL(kind=real_4), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_rm

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_rv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_rv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_rv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_rv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_r(msgout,msgin,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout
    REAL(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_r

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_r(msgout,msgin,gid,request)
    REAL(kind=real_4), INTENT(IN)                      :: msgout
    REAL(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_r

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_r12(msgout, msgin,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_r12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_r12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_r12
! *****************************************************************************
  SUBROUTINE mp_allgather_r23(msgout, msgin,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:,:)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_r23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_r23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_r12
! *****************************************************************************
  SUBROUTINE mp_allgather_r34(msgout, msgin,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:,:, :)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_r34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_r34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_r12
! *****************************************************************************
  SUBROUTINE mp_allgather_r22(msgout, msgin,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:, :)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_r22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_r22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_r12
! *****************************************************************************
  SUBROUTINE mp_iallgather_r22(msgout, msgin, gid, request)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:, :)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_r22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_r22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_r12
! *****************************************************************************
  SUBROUTINE mp_iallgather_r33(msgout, msgin, gid, request)
    REAL(kind=real_4), INTENT(IN)                      :: msgout(:, :, :)
    REAL(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_r33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_r33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_rv(msgout,msgin,rcount,rdispl,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_rv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_rv(msgout,msgin,rcount,rdispl,gid,request)
    REAL(kind=real_4), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_rv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_rv(msgout,msgin,rcount,gid)
    REAL(kind=real_4), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_rv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_rv(msgin,dest,msgout,source,comm)
    REAL(kind=real_4), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_rv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_rv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_rm2(msgin,dest,msgout,source,comm)
    REAL(kind=real_4), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_rm2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_rv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_rm3(msgin,dest,msgout,source,comm)
    REAL(kind=real_4), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_rm3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_rv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_rm4(msgin,dest,msgout,source,comm)
    REAL(kind=real_4), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_rm4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_r(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(kind=real_4)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_r

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_rv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(kind=real_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_rv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_rv(msgin,dest,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_rv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_rv
!> \note see mp_isend_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_rm2(msgin,dest,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_rm2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_rv
!> \note see mp_isend_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_rm3(msgin,dest,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_rm3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_rv(msgout,source,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_rv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_rv
!> \note see mp_irecv_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_rm2(msgout,source,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_rm2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_rv
!> \note see mp_irecv_rv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_rm3(msgout,source,comm,request,tag)
    REAL(kind=real_4), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_rm3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_rv(base,comm,win)
    REAL(kind=real_4), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_rv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_rv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    REAL(kind=real_4), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    REAL(kind=real_4), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_rv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_r(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_r', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 1
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_r

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_r(DATA, len, stat)
    REAL(kind=real_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_r', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_r

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_r(DATA, stat)
    REAL(kind=real_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_r', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_r

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_rv(fh, offset, msg, msglen)
    REAL(kind=real_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_rv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_rv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_r(fh, offset, msg)
    REAL(kind=real_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_r', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_r

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_rv(fh, offset, msg, msglen)
    REAL(kind=real_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_rv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_rv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_r(fh, offset, msg)
    REAL(kind=real_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_r', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_r

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_rv(fh, offset, msg, msglen)
    REAL(kind=real_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_rv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_rv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_r(fh, offset, msg)
    REAL(kind=real_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_r', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_r

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_rv(fh, offset, msg, msglen)
    REAL(kind=real_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_rv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_rv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_r(fh, offset, msg)
    REAL(kind=real_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_r', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_r

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_r (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    REAL(kind=real_4), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_r', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 1
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_r => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_r

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_r(DATA, len, stat)
    REAL(kind=real_4), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_r

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_r(DATA, stat)
    REAL(kind=real_4), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_r
! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_dm(msg, group, displ_in)

    REAL(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_dm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_dm


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_d(msg, group, displ_in)

    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_d

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_d11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_8), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_d11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_d22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    REAL(kind=real_8), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_8), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_d ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: sb
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
! *****************************************************************************
  SUBROUTINE mp_alltoall_d22 ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:, :), INTENT(IN)     :: sb
    REAL(kind=real_8), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
! *****************************************************************************
  SUBROUTINE mp_alltoall_d33 ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:, :, :), INTENT(IN)  :: sb
    REAL(kind=real_8), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
! *****************************************************************************
  SUBROUTINE mp_alltoall_d44 ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_d45 ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_8), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_d34 ( sb, rb, count, group )

    REAL(kind=real_8), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_d
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_d54 ( sb, rb, count, group )

    REAL(kind=real_8), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    REAL(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_d54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_d54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_d(msg,dest,tag,gid)
    REAL(kind=real_8)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_d

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_d
! *****************************************************************************
  SUBROUTINE mp_send_dv(msg,dest,tag,gid)
    REAL(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_dv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_d(msg,source,tag,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_d

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_d
! *****************************************************************************
  SUBROUTINE mp_recv_dv(msg,source,tag,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_dv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_d(msg,source,gid)
    REAL(kind=real_8)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_d

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_d(msg,source,gid,request)
    REAL(kind=real_8)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_d

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_d1
! *****************************************************************************
  SUBROUTINE mp_bcast_dv(msg,source,gid)
    REAL(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_dv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_d1
! *****************************************************************************
  SUBROUTINE mp_ibcast_dv(msg,source,gid,request)
    REAL(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_dv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_d1
! *****************************************************************************
  SUBROUTINE mp_bcast_dm(msg,source,gid)
    REAL(kind=real_8)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_dm

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_d1
! *****************************************************************************
  SUBROUTINE mp_bcast_d3(msg,source,gid)
    REAL(kind=real_8)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_d3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_d3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_d(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_d

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_d
! *****************************************************************************
  SUBROUTINE mp_sum_dv(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_dv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_d
! *****************************************************************************
  SUBROUTINE mp_isum_dv(msg,gid,request)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_dv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_d
! *****************************************************************************
  SUBROUTINE mp_sum_dm(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_dm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_dm

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_d
! *****************************************************************************
  SUBROUTINE mp_sum_dm3(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_dm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_dm3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_d
! *****************************************************************************
  SUBROUTINE mp_sum_dm4(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_dm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_dm4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_dv(msg,root,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_dv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_dv
! *****************************************************************************
  SUBROUTINE mp_sum_root_dm(msg,root,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_dm

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_dm(msg,res,gid)
    REAL(kind=real_8), INTENT(IN)         :: msg( :, : )
    REAL(kind=real_8), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_dm'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_dm

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_d(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_d

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_d
! *****************************************************************************
  SUBROUTINE mp_max_dv(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_dv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_d(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_d

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_d
! *****************************************************************************
  SUBROUTINE mp_min_dv(msg,gid)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_dv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_dv(msg_scatter,msg,root,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_8), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_dv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_d(msg_scatter,msg,root,gid,request)
    REAL(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_d

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_dv(msg_scatter,msg,root,gid,request)
    REAL(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_dv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_dv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    REAL(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    REAL(kind=real_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_dv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_d(msg,msg_gather,root,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msg
    REAL(kind=real_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_d

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_d
! *****************************************************************************
  SUBROUTINE mp_gather_dv(msg,msg_gather,root,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msg( : )
    REAL(kind=real_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_dv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_d
! *****************************************************************************
  SUBROUTINE mp_gather_dm(msg,msg_gather,root,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msg( :, : )
    REAL(kind=real_8), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_dm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_dm

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_dv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_dv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_dv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_dv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_d(msgout,msgin,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout
    REAL(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_d

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_d(msgout,msgin,gid,request)
    REAL(kind=real_8), INTENT(IN)                      :: msgout
    REAL(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_d

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_d12(msgout, msgin,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_d12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_d12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_d12
! *****************************************************************************
  SUBROUTINE mp_allgather_d23(msgout, msgin,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:,:)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_d23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_d23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_d12
! *****************************************************************************
  SUBROUTINE mp_allgather_d34(msgout, msgin,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:,:, :)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_d34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_d34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_d12
! *****************************************************************************
  SUBROUTINE mp_allgather_d22(msgout, msgin,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:, :)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_d22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_d22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_d12
! *****************************************************************************
  SUBROUTINE mp_iallgather_d22(msgout, msgin, gid, request)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:, :)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_d22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_d22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_d12
! *****************************************************************************
  SUBROUTINE mp_iallgather_d33(msgout, msgin, gid, request)
    REAL(kind=real_8), INTENT(IN)                      :: msgout(:, :, :)
    REAL(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_d33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_d33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_dv(msgout,msgin,rcount,rdispl,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_dv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_dv(msgout,msgin,rcount,rdispl,gid,request)
    REAL(kind=real_8), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_dv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_dv(msgout,msgin,rcount,gid)
    REAL(kind=real_8), INTENT(IN)                      :: msgout( : )
    REAL(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_dv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_dv(msgin,dest,msgout,source,comm)
    REAL(kind=real_8), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_dv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_dv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_dm2(msgin,dest,msgout,source,comm)
    REAL(kind=real_8), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_dm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_dm2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_dv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_dm3(msgin,dest,msgout,source,comm)
    REAL(kind=real_8), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_dm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_dm3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_dv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_dm4(msgin,dest,msgout,source,comm)
    REAL(kind=real_8), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_dm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_dm4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_d(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(kind=real_8)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_d

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_dv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(kind=real_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(kind=real_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_dv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_dv(msgin,dest,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_dv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_dv
!> \note see mp_isend_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_dm2(msgin,dest,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_dm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_dm2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_dv
!> \note see mp_isend_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_dm3(msgin,dest,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_dm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_dm3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_dv(msgout,source,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_dv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_dv
!> \note see mp_irecv_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_dm2(msgout,source,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_dm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_dm2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_dv
!> \note see mp_irecv_dv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_dm3(msgout,source,comm,request,tag)
    REAL(kind=real_8), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_dm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_dm3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_dv(base,comm,win)
    REAL(kind=real_8), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_dv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_dv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    REAL(kind=real_8), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    REAL(kind=real_8), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_dv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_dv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_d(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_d', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 3
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_d

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_d(DATA, len, stat)
    REAL(kind=real_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_d', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_d

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_d(DATA, stat)
    REAL(kind=real_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_d', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_d

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_dv(fh, offset, msg, msglen)
    REAL(kind=real_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_dv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_dv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_d(fh, offset, msg)
    REAL(kind=real_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_d', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_d

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_dv(fh, offset, msg, msglen)
    REAL(kind=real_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_dv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_dv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_d(fh, offset, msg)
    REAL(kind=real_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_d', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_d

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_dv(fh, offset, msg, msglen)
    REAL(kind=real_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_dv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_dv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_d(fh, offset, msg)
    REAL(kind=real_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_d', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_d

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_dv(fh, offset, msg, msglen)
    REAL(kind=real_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_dv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_dv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_d(fh, offset, msg)
    REAL(kind=real_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_d', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_d

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_d (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    REAL(kind=real_8), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_d', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 3
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_d => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_d

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_d(DATA, len, stat)
    REAL(kind=real_8), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_d

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_d(DATA, stat)
    REAL(kind=real_8), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_d
! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_cm(msg, group, displ_in)

    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_cm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_cm


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_c(msg, group, displ_in)

    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_c

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_c11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_c11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_c22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(kind=real_4), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_4), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_c ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: sb
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
! *****************************************************************************
  SUBROUTINE mp_alltoall_c22 ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(IN)     :: sb
    COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
! *****************************************************************************
  SUBROUTINE mp_alltoall_c33 ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:, :, :), INTENT(IN)  :: sb
    COMPLEX(kind=real_4), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
! *****************************************************************************
  SUBROUTINE mp_alltoall_c44 ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_c45 ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_4), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_c34 ( sb, rb, count, group )

    COMPLEX(kind=real_4), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_c
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_c54 ( sb, rb, count, group )

    COMPLEX(kind=real_4), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    COMPLEX(kind=real_4), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_c54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_c(msg,dest,tag,gid)
    COMPLEX(kind=real_4)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_c

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_c
! *****************************************************************************
  SUBROUTINE mp_send_cv(msg,dest,tag,gid)
    COMPLEX(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_cv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_c(msg,source,tag,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_c

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_c
! *****************************************************************************
  SUBROUTINE mp_recv_cv(msg,source,tag,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_cv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_c(msg,source,gid)
    COMPLEX(kind=real_4)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_c

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_c(msg,source,gid,request)
    COMPLEX(kind=real_4)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_c

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_c1
! *****************************************************************************
  SUBROUTINE mp_bcast_cv(msg,source,gid)
    COMPLEX(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_cv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_c1
! *****************************************************************************
  SUBROUTINE mp_ibcast_cv(msg,source,gid,request)
    COMPLEX(kind=real_4)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_cv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_c1
! *****************************************************************************
  SUBROUTINE mp_bcast_cm(msg,source,gid)
    COMPLEX(kind=real_4)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_cm

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_c1
! *****************************************************************************
  SUBROUTINE mp_bcast_c3(msg,source,gid)
    COMPLEX(kind=real_4)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_c3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_c3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_c(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_c

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_c
! *****************************************************************************
  SUBROUTINE mp_sum_cv(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_cv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_c
! *****************************************************************************
  SUBROUTINE mp_isum_cv(msg,gid,request)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_cv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_c
! *****************************************************************************
  SUBROUTINE mp_sum_cm(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_cm

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_c
! *****************************************************************************
  SUBROUTINE mp_sum_cm3(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_cm3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_c
! *****************************************************************************
  SUBROUTINE mp_sum_cm4(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_cm4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_cv(msg,root,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_cv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_cv
! *****************************************************************************
  SUBROUTINE mp_sum_root_cm(msg,root,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_cm

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_cm(msg,res,gid)
    COMPLEX(kind=real_4), INTENT(IN)         :: msg( :, : )
    COMPLEX(kind=real_4), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_cm'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_cm

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_c(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_c

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_c
! *****************************************************************************
  SUBROUTINE mp_max_cv(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_cv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_c(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_c

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_c
! *****************************************************************************
  SUBROUTINE mp_min_cv(msg,gid)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_cv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_cv(msg_scatter,msg,root,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_cv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_c(msg_scatter,msg,root,gid,request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_c

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_cv(msg_scatter,msg,root,gid,request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_cv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_cv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_cv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_c(msg,msg_gather,root,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_c

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_c
! *****************************************************************************
  SUBROUTINE mp_gather_cv(msg,msg_gather,root,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg( : )
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_cv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_c
! *****************************************************************************
  SUBROUTINE mp_gather_cm(msg,msg_gather,root,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg( :, : )
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_cm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_cm

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_cv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_cv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_cv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: sendbuf
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_cv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_c(msgout,msgin,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_c

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_c(msgout,msgin,gid,request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_c

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_c12(msgout, msgin,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_c12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_c12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_c12
! *****************************************************************************
  SUBROUTINE mp_allgather_c23(msgout, msgin,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:,:)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_c23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_c23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_c12
! *****************************************************************************
  SUBROUTINE mp_allgather_c34(msgout, msgin,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:,:, :)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_c34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_c34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_c12
! *****************************************************************************
  SUBROUTINE mp_allgather_c22(msgout, msgin,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:, :)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_c22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_c22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_c12
! *****************************************************************************
  SUBROUTINE mp_iallgather_c22(msgout, msgin, gid, request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:, :)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_c22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_c22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_c12
! *****************************************************************************
  SUBROUTINE mp_iallgather_c33(msgout, msgin, gid, request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout(:, :, :)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_c33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_c33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_cv(msgout,msgin,rcount,rdispl,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_cv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_cv(msgout,msgin,rcount,rdispl,gid,request)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_cv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_cv(msgout,msgin,rcount,gid)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_cv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_cv(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_cv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_cv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_cm2(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_cm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_cm2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_cv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_cm3(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_cm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_cm3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_cv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_cm4(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_cm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_cm4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_c(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    COMPLEX(kind=real_4)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_c

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_cv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_cv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_cv(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_cv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_cv
!> \note see mp_isend_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_cm2(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_cm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_cm2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_cv
!> \note see mp_isend_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_cm3(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_cm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_cm3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_cv(msgout,source,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_cv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_cv
!> \note see mp_irecv_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_cm2(msgout,source,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_cm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_cm2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_cv
!> \note see mp_irecv_cv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_cm3(msgout,source,comm,request,tag)
    COMPLEX(kind=real_4), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_cm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_cm3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_cv(base,comm,win)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_cv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_cv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    COMPLEX(kind=real_4), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_cv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_c(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_c', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 5
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_c

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_c(DATA, len, stat)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_c', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_c

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_c(DATA, stat)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_c', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_c

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_cv(fh, offset, msg, msglen)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_cv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_cv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_c(fh, offset, msg)
    COMPLEX(kind=real_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_c', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_c

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_cv(fh, offset, msg, msglen)
    COMPLEX(kind=real_4), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_cv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_cv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_c(fh, offset, msg)
    COMPLEX(kind=real_4), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_c', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_c

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_cv(fh, offset, msg, msglen)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_cv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_cv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_c(fh, offset, msg)
    COMPLEX(kind=real_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_c', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_c

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_cv(fh, offset, msg, msglen)
    COMPLEX(kind=real_4), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_cv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_cv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_c(fh, offset, msg)
    COMPLEX(kind=real_4), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_c', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_c

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_c (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_c', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 5
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_c => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_c

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_c(DATA, len, stat)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_c

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_c(DATA, stat)
    COMPLEX(kind=real_4), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_c
! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Rank-2 data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_zm(msg, group, displ_in)

    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_zm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_zm


! *****************************************************************************
!> \brief Shift around the data in msg
!> \param[in,out] msg         Data to shift
!> \param[in] group           message passing environment identifier
!> \param[in] displ_in        displacements (?)
!> \par Example
!>      msg will be moved from rank to rank+displ_in (in a circular way)
!> \par Limitations
!>      * displ_in will be 1 by default (others not tested)
!>      * the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_z(msg, group, displ_in)

    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierror

    ierror = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displ_in))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)

  END SUBROUTINE mp_shift_z

! *****************************************************************************
!> \brief All-to-all data exchange, rank-1 data of different sizes
!> \param[in] sb              Data to send
!> \param[in] scount          Data counts for data sent to other processes
!> \param[in] sdispl          Respective data offsets for data sent to process
!> \param[in,out] rb          Buffer into which to receive data
!> \param[in] rcount          Data counts for data received from other
!>                            processes
!> \param[in] rdispl          Respective data offsets for data received from
!>                            other processes
!> \param[in] group           Message passing environment identifier
!> \par MPI mapping
!>      mpi_alltoallv
!> \par Array sizes
!>      The scount, rcount, and the sdispl and rdispl arrays have a
!>      size equal to the number of processes.
!> \par Offsets
!>      Values in sdispl and rdispl start with 0.
! *****************************************************************************
  SUBROUTINE mp_alltoall_z11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER                                  :: i

    CALL mp_timeset(routineN,handle)

    ierr = 0
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    !$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(rcount,rdispl,sdispl,rb,sb)
    DO i=1,rcount(1)
       rb(rdispl(1)+i)=sb(sdispl(1)+i)
    ENDDO
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z11v

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 data of different sizes
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
!> \par MPI mapping
!>      mpi_alltoallv
!> \note see mp_alltoall_z11v
! *****************************************************************************
  SUBROUTINE mp_alltoall_z22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(kind=real_8), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_8), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z22v

! *****************************************************************************
!> \brief All-to-all data exchange, rank 1 arrays, equal sizes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
!> \param[in] group           Message passing environment identifier
!> \par Index meaning
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \par Sizes of ranks
!>      All processes have the same data size.
!> \par MPI mapping
!>      mpi_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_z ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: sb
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z

! *****************************************************************************
!> \brief All-to-all data exchange, rank-2 arrays, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
! *****************************************************************************
  SUBROUTINE mp_alltoall_z22 ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(IN)     :: sb
    COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z22

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data with equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
! *****************************************************************************
  SUBROUTINE mp_alltoall_z33 ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:, :, :), INTENT(IN)  :: sb
    COMPLEX(kind=real_8), DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z33

! *****************************************************************************
!> \brief All-to-all data exchange, rank 4 data, equal sizes
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
! *****************************************************************************
  SUBROUTINE mp_alltoall_z44 ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z44', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb=sb
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z44

! *****************************************************************************
!> \brief All-to-all data exchange, rank-4 data to rank-5 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_z45 ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_8), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z45

! *****************************************************************************
!> \brief All-to-all data exchange, rank-3 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_z34 ( sb, rb, count, group )

    COMPLEX(kind=real_8), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    COMPLEX(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z34

! *****************************************************************************
!> \brief All-to-all data exchange, rank-5 data to rank-4 data
!> \param sb ...
!> \param rb ...
!> \param count ...
!> \param group ...
!> \note see mp_alltoall_z
!> \note User must ensure size consistency.
! *****************************************************************************
  SUBROUTINE mp_alltoall_z54 ( sb, rb, count, group )

    COMPLEX(kind=real_8), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    COMPLEX(kind=real_8), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(count))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(group))==-1) EXIT ;  END DO ; ENDIF
    rb = RESHAPE(sb, SHAPE(rb))
    CALL mp_timestop(handle)

  END SUBROUTINE mp_alltoall_z54

! *****************************************************************************
!> \brief Send one datum to another process
!> \param[in] msg             Dum to send
!> \param[in] dest            Destination process
!> \param[in] tag             Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_send_z(msg,dest,tag,gid)
    COMPLEX(kind=real_8)                                  :: msg
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",594,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_z

! *****************************************************************************
!> \brief Send rank-1 data to another process
!> \param[in] msg             Rank-1 data to send
!> \param dest ...
!> \param tag ...
!> \param gid ...
!> \note see mp_send_z
! *****************************************************************************
  SUBROUTINE mp_send_zv(msg,dest,tag,gid)
    COMPLEX(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",632,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_send_zv

! *****************************************************************************
!> \brief Receive one datum from another process
!> \param[in,out] msg         Place received data into this variable
!> \param[in,out] source      Process to receieve from
!> \param[in,out] tag         Transfer identifier
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_send
! *****************************************************************************
  SUBROUTINE mp_recv_z(msg,source,tag,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",679,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_z

! *****************************************************************************
!> \brief Receive rank-1 data from another process
!> \param[in,out] msg         Place receieved data into this rank-1 array
!> \param source ...
!> \param tag ...
!> \param gid ...
!> \note see mp_recv_z
! *****************************************************************************
  SUBROUTINE mp_recv_zv(msg,source,tag,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    ! only defined in parallel
    CALL cp__b("mpiwrap/message_passing.F",725,"not in parallel mode")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_recv_zv

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_z(msg,source,gid)
    COMPLEX(kind=real_8)                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_z

! *****************************************************************************
!> \brief Broadcasts a datum to all processes.
!> \param[in] msg             Datum to broadcast
!> \param[in] source          Processes which broadcasts
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_bcast
! *****************************************************************************
  SUBROUTINE mp_ibcast_z(msg,source,gid,request)
    COMPLEX(kind=real_8)                                  :: msg
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_z

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_z1
! *****************************************************************************
  SUBROUTINE mp_bcast_zv(msg,source,gid)
    COMPLEX(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_zv

! *****************************************************************************
!> \brief Broadcasts rank-1 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_z1
! *****************************************************************************
  SUBROUTINE mp_ibcast_zv(msg,source,gid,request)
    COMPLEX(kind=real_8)                                  :: msg( : )
    INTEGER                                  :: source, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_ibcast_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_ibcast_zv

! *****************************************************************************
!> \brief Broadcasts rank-2 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_z1
! *****************************************************************************
  SUBROUTINE mp_bcast_zm(msg,source,gid)
    COMPLEX(kind=real_8)                                  :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_zm

! *****************************************************************************
!> \brief Broadcasts rank-3 data to all processes
!> \param[in] msg             Data to broadcast
!> \param source ...
!> \param gid ...
!> \note see mp_bcast_z1
! *****************************************************************************
  SUBROUTINE mp_bcast_z3(msg,source,gid)
    COMPLEX(kind=real_8)                                  :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_z3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_bcast_z3

! *****************************************************************************
!> \brief Sums a datum from all processes with result left on all processes.
!> \param[in,out] msg         Datum to sum (input) and result (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_sum_z(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_z

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_z
! *****************************************************************************
  SUBROUTINE mp_sum_zv(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_zv

! *****************************************************************************
!> \brief Element-wise sum of a rank-1 array on all processes.
!> \param[in,out] msg         Vector to sum and result
!> \param gid ...
!> \note see mp_sum_z
! *****************************************************************************
  SUBROUTINE mp_isum_zv(msg,gid,request)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isum_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isum_zv

! *****************************************************************************
!> \brief Element-wise sum of a rank-2 array on all processes.
!> \param[in] msg             Matrix to sum and result
!> \param gid ...
!> \note see mp_sum_z
! *****************************************************************************
  SUBROUTINE mp_sum_zm(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_zm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_zm

! *****************************************************************************
!> \brief Element-wise sum of a rank-3 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_z
! *****************************************************************************
  SUBROUTINE mp_sum_zm3(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_zm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen
    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_zm3

! *****************************************************************************
!> \brief Element-wise sum of a rank-4 array on all processes.
!> \param[in] msg             Array to sum and result
!> \param gid ...
!> \note see mp_sum_z
! *****************************************************************************
  SUBROUTINE mp_sum_zm4(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( :, :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_zm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, &
                                                msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_zm4

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Vector to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_reduce
! *****************************************************************************
  SUBROUTINE mp_sum_root_zv(msg,root,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_zv

! *****************************************************************************
!> \brief Element-wise sum of data from all processes with result left only on
!>        one.
!> \param[in,out] msg         Matrix to sum (input) and (only on process root)
!>                            result (output)
!> \param root ...
!> \param gid ...
!> \note see mp_sum_root_zv
! *****************************************************************************
  SUBROUTINE mp_sum_root_zm(msg,root,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_root_zm

! *****************************************************************************
!> \brief Partial sum of data from all processes with result on each process.
!> \param[in] msg          Matrix to sum (input)
!> \param[out] res         Matrix containing result (output)
!> \param[in] gid          Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_partial_zm(msg,res,gid)
    COMPLEX(kind=real_8), INTENT(IN)         :: msg( :, : )
    COMPLEX(kind=real_8), INTENT(OUT)        :: res( :, : )
    INTEGER, INTENT(IN)         :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_partial_zm'   &
                                 , routineP = moduleN//':'//routineN

    INTEGER                     :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    res = msg
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_partial_zm

! *****************************************************************************
!> \brief Finds the maximum of a datum with the result left on all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_max_z(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_z

! *****************************************************************************
!> \brief Finds the element-wise maximum of a vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find maximum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \note see mp_max_z
! *****************************************************************************
  SUBROUTINE mp_max_zv(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_max_zv

! *****************************************************************************
!> \brief Finds the minimum of a datum with the result left on all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_allreduce
! *****************************************************************************
  SUBROUTINE mp_min_z(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msg))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_z

! *****************************************************************************
!> \brief Finds the element-wise minimum of vector with the result left on
!>        all processes.
!> \param[in,out] msg         Find minimum among these data (input) and
!>                            maximum (output)
!> \param gid ...
!> \par MPI mapping
!>      mpi_allreduce
!> \note see mp_min_z
! *****************************************************************************
  SUBROUTINE mp_min_zv(msg,gid)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    CALL mp_timestop(handle)
  END SUBROUTINE mp_min_zv


! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[out] msg            Received data
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_scatter_zv(msg_scatter,msg,root,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    CALL mp_timestop(handle)
  END SUBROUTINE mp_scatter_zv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_z(msg_scatter,msg,root,gid,request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter(1)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_z

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatter_zv(msg_scatter,msg,root,gid,request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatter_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg = msg_scatter
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatter_zv

! *****************************************************************************
!> \brief Scatters data from one processes to all others
!> \param[in] msg_scatter     Data to scatter (for root process)
!> \param[in] root            Process which scatters data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_scatter
! *****************************************************************************
  SUBROUTINE mp_iscatterv_zv(msg_scatter,sendcounts,displs,msg,recvcount,root,gid,request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(IN)                      :: sendcounts(:), displs(:)
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: msg(:)
    INTEGER, INTENT(IN)                      :: recvcount, root, gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iscatterv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(displs))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg(1:recvcount) = msg_scatter(1+displs(1):1+displs(1)+sendcounts(1))
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iscatterv_zv

! *****************************************************************************
!> \brief Gathers a datum from all processes to one
!> \param[in] msg             Datum to send to root
!> \param[out] msg_gather     Received data (on root)
!> \param[in] root            Process which gathers the data
!> \param[in] gid             Message passing environment identifier
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gather_z(msg,msg_gather,root,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = 1
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather(1) = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_z

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_z
! *****************************************************************************
  SUBROUTINE mp_gather_zv(msg,msg_gather,root,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg( : )
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_zv

! *****************************************************************************
!> \brief Gathers data from all processes to one
!> \param[in] msg             Datum to send to root
!> \param msg_gather ...
!> \param root ...
!> \param gid ...
!> \par Data length
!>      All data (msg) is equal-sized
!> \par MPI mapping
!>      mpi_gather
!> \note see mp_gather_z
! *****************************************************************************
  SUBROUTINE mp_gather_zm(msg,msg_gather,root,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg( :, : )
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg_gather( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_zm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
    CALL mp_timeset(routineN,handle)

    msglen = SIZE(msg)
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msg_gather = msg
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gather_zm

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_gatherv_zv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):) = sendbuf
    CALL mp_timestop(handle)
  END SUBROUTINE mp_gatherv_zv

! *****************************************************************************
!> \brief Gathers data from all processes to one.
!> \param[in] sendbuf         Data to send to root
!> \param[out] recvbuf        Received data (on root)
!> \param[in] recvcounts      Sizes of data received from processes
!> \param[in] displs          Offsets of data received from processes
!> \param[in] root            Process which gathers the data
!> \param[in] comm            Message passing environment identifier
!> \par Data length
!>      Data can have different lengths
!> \par Offsets
!>      Offsets start at 0
!> \par MPI mapping
!>      mpi_gather
! *****************************************************************************
  SUBROUTINE mp_igatherv_zv(sendbuf,sendcount,recvbuf,recvcounts,displs,root,comm,request)
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: sendbuf
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: sendcount, root, comm
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_igatherv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(sendcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(recvcounts))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(root))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    recvbuf(1+displs(1):1+displs(1)+recvcounts(1)) = sendbuf(1:sendcount)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_igatherv_zv


! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_z(msgout,msgin,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_z

! *****************************************************************************
!> \brief Gathers a datum from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Datum to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgather_z(msgout,msgin,gid,request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_z

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      All processes send equal-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgather_z12(msgout, msgin,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_z12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,1) = msgout(:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_z12

! *****************************************************************************
!> \brief Gathers matrix data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_z12
! *****************************************************************************
  SUBROUTINE mp_allgather_z23(msgout, msgin,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:,:)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_z23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,1) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_z23

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_z12
! *****************************************************************************
  SUBROUTINE mp_allgather_z34(msgout, msgin,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:,:, :)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_z34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:,1) = msgout(:,:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_z34


! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \note see mp_allgather_z12
! *****************************************************************************
  SUBROUTINE mp_allgather_z22(msgout, msgin,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:, :)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_z22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgather_z22

! *****************************************************************************
!> \brief Gathers rank-2 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-2 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_z12
! *****************************************************************************
  SUBROUTINE mp_iallgather_z22(msgout, msgin, gid, request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:, :)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_z22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:) = msgout(:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_z22

! *****************************************************************************
!> \brief Gathers rank-3 data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-3 data to send
!> \param msgin ...
!> \param gid ...
!> \param request ...
!> \note see mp_allgather_z12
! *****************************************************************************
  SUBROUTINE mp_iallgather_z33(msgout, msgin, gid, request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout(:, :, :)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid
    INTEGER, INTENT(OUT)                     :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgather_z33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin(:,:,:) = msgout(:,:,:)
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgather_z33

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_allgatherv_zv(msgout,msgin,rcount,rdispl,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allgatherv_zv

! *****************************************************************************
!> \brief Gathers vector data from all processes and all processes receive the
!>        same data
!> \param[in] msgout          Rank-1 data to send
!> \param[out] msgin          Received data
!> \param[in] rcount          Size of sent data for every process
!> \param[in] rdispl          Offset of sent data for every process
!> \param[in] gid             Message passing environment identifier
!> \par Data size
!>      Processes can send different-sized data
!> \par Ranks
!>      The last rank counts the processes
!> \par Offsets
!>      Offsets are from 0
!> \par MPI mapping
!>      mpi_allgather
! *****************************************************************************
  SUBROUTINE mp_iallgatherv_zv(msgout,msgin,rcount,rdispl,gid,request)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid
    INTEGER, INTENT(INOUT)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_iallgatherv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rdispl))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    request = mp_request_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_iallgatherv_zv

! *****************************************************************************
!> \brief Sums a vector and partitions the result among processes
!> \param[in] msgout          Data to sum
!> \param[out] msgin          Received portion of summed data
!> \param[in] rcount          Partition sizes of the summed data for
!>                            every process
!> \param[in] gid             Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_zv(msgout,msgin,rcount,gid)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgout( : )
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(rcount))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gid))==-1) EXIT ;  END DO ; ENDIF
    msgin = msgout
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sum_scatter_zv

! *****************************************************************************
!> \brief Sends and receives vector data
!> \param[in] msgin           Data to send
!> \param[in] dest            Process to send data to
!> \param[out] msgout         Received data
!> \param[in] source          Process from which to receive
!> \param[in] comm            Message passing environment identifier
! *****************************************************************************
  SUBROUTINE mp_sendrecv_zv(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_zv

! *****************************************************************************
!> \brief Sends and receives matrix data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_zv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_zm2(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_zm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_zm2

! *****************************************************************************
!> \brief Sends and receives rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_zv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_zm3(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_zm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_zm3

! *****************************************************************************
!> \brief Sends and receives rank-4 data
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv_zv
! *****************************************************************************
  SUBROUTINE mp_sendrecv_zm4(msgin,dest,msgout,source,comm)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msgin( :, :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msgout( :, :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_zm4', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_sendrecv_zm4

! *****************************************************************************
!> \brief Non-blocking send and receieve of a scalar
!> \param[in] msgin           Scalar data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      02.2005 created [Alfio Lazzaro]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_z(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    COMPLEX(kind=real_8)                                  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8)                                  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_z

! *****************************************************************************
!> \brief Non-blocking send and receieve of a vector
!> \param[in] msgin           Vector data to send
!> \param[in] dest            Which process to send to
!> \param[out] msgout         Receive data into this pointer
!> \param[in] source          Process to receive from
!> \param[in] comm            Message passing environment identifier
!> \param[out] send_request   Request handle for the send
!> \param[out] recv_request   Request handle for the receive
!> \param[in] tag             (optional) tag to differentiate requests
!> \par Implementation
!>      Calls mpi_isend and mpi_irecv.
!> \par History
!>      11.2004 created [Joost VandeVondele]
!> \note
!>      The arguments must be pointers to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isendrecv_zv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    send_request=0
    recv_request=0
    msgout = msgin
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isendrecv_zv

! *****************************************************************************
!> \brief Non-blocking send of vector data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!> \note see mp_isendrecv_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_zv(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_zv

! *****************************************************************************
!> \brief Non-blocking send of matrix data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_zv
!> \note see mp_isend_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_zm2(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_zm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_zm2

! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune]
!>     (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_zv
!> \note see mp_isend_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_isend_zm3(msgin,dest,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_zm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgin))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dest))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    ierr=1
    request=0
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
    CALL mp_timestop(handle)
  END SUBROUTINE mp_isend_zm3

! *****************************************************************************
!> \brief Non-blocking receive of vector data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      08.2003 created [f&j]
!>      2009-11-25 [UB] Made type-generic for templates
!> \note see mp_isendrecv_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_zv(msgout,source,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    CALL cp__b("mpiwrap/message_passing.F",2964,"mp_irecv called in non parallel case")
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_zv

! *****************************************************************************
!> \brief Non-blocking receive of matrix data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_zv
!> \note see mp_irecv_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_zm2(msgout,source,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER        :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_zm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3035,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_zm2


! *****************************************************************************
!> \brief Non-blocking send of rank-3 data
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      2009-11-25 [UB] Made type-generic for templates
!> \author fawzi
!> \note see mp_isendrecv_zv
!> \note see mp_irecv_zv
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_irecv_zm3(msgout,source,comm,request,tag)
    COMPLEX(kind=real_8), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_zm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(msgout))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(request))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tag))==-1) EXIT ;  END DO ; ENDIF
    request=0
    CALL cp__b("mpiwrap/message_passing.F",3105,"mp_irecv called in non parallel case")
    CALL mp_timestop(handle)
  END SUBROUTINE mp_irecv_zm3

! *****************************************************************************
!> \brief Window initialization function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries.
! *****************************************************************************
  SUBROUTINE mp_win_create_zv(base,comm,win)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: base
    INTEGER, INTENT(IN)            :: comm
    INTEGER, INTENT(INOUT)         :: win

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(base))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
    win = mp_win_null
    CALL mp_timestop(handle)
  END SUBROUTINE mp_win_create_zv

! *****************************************************************************
!> \brief Single-sided get function for vector data
!> \param base ...
!> \param comm ...
!> \param win ...
!> \par History
!>      02.2015 created [Alfio Lazzaro]
!> \note
!>      The argument must be a pointer to be sure that we do not get
!>      temporaries. They must point to contiguous memory.
! *****************************************************************************
  SUBROUTINE mp_rget_zv(base,source,win,win_data,myproc,disp,request,&
       origin_datatype, target_datatype)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER                      :: base
    INTEGER, INTENT(IN)                                 :: source, win
    COMPLEX(kind=real_8), DIMENSION(:), POINTER                      :: win_data
    INTEGER, INTENT(IN), OPTIONAL                       :: myproc, disp
    INTEGER, INTENT(OUT)                                :: request
    TYPE(mp_type_descriptor_type), INTENT(IN), OPTIONAL :: origin_datatype, target_datatype

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rget_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(source))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(win))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(myproc))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(origin_datatype))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(target_datatype))==-1) EXIT ;  END DO ; ENDIF

    request = mp_request_null
    !
    IF (PRESENT(disp)) THEN
       base(:) = win_data(disp+1:disp+SIZE(base))
    ELSE
       base(:) = win_data(:SIZE(base))
    ENDIF

    CALL mp_timestop(handle)
  END SUBROUTINE mp_rget_zv

! *****************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \retval type_descriptor ...
! ***************************************************************************
  FUNCTION mp_type_indexed_make_z(count,lengths,displs) &
       RESULT(type_descriptor)
    INTEGER, INTENT(IN)                      :: count
    INTEGER, DIMENSION(1:count), INTENT(IN), TARGET  :: lengths, displs
    TYPE(mp_type_descriptor_type)            :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_indexed_make_z', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr, handle

    ierr = 0
    CALL mp_timeset(routineN,handle)

    type_descriptor%type_handle = 7
    type_descriptor%length = count
    NULLIFY(type_descriptor%subtype)
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .TRUE.
    type_descriptor%index_descriptor%index => lengths
    type_descriptor%index_descriptor%chunks => displs

    CALL mp_timestop(handle)

  END FUNCTION mp_type_indexed_make_z

! *****************************************************************************
!> \brief Allocates special parallel memory
!> \param[in]  DATA      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_allocate_z(DATA, len, stat)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_z', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    ALLOCATE(DATA(len), stat=ierr)
    IF (ierr/=0 .AND. .NOT. PRESENT(stat)) &
       CALL mp_stop(ierr, "ALLOCATE @ "//routineN)
   IF(PRESENT(stat)) stat = ierr
    CALL mp_timestop(handle)
  END SUBROUTINE mp_allocate_z

! *****************************************************************************
!> \brief Deallocates special parallel memory
!> \param[in] DATA         pointer to special memory to deallocate
!> \param stat ...
!> \author UB
! *****************************************************************************
  SUBROUTINE mp_deallocate_z(DATA, stat)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_z', &
         routineP = moduleN//':'//routineN

    INTEGER                             :: ierr, handle

    CALL mp_timeset(routineN,handle)

    ierr = 0
    DEALLOCATE(DATA)
    IF(PRESENT(stat)) stat = 0
    CALL mp_timestop(handle)
  END SUBROUTINE mp_deallocate_z

! *****************************************************************************
!> \brief (parallel) Blocking individual file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[in] msg    data to be writen to the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at
!> \par STREAM-I/O mapping   WRITE
!> \param[in](optional) msglen number of the elements of data
! *****************************************************************************
  SUBROUTINE mp_file_write_at_zv(fh, offset, msg, msglen)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_zv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_zv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_z(fh, offset, msg)
    COMPLEX(kind=real_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_z', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_z

! *****************************************************************************
!> \brief (parallel) Blocking collective file write using explicit offsets
!>        (serial) Unformatted stream write
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_write_at_all
!> \par STREAM-I/O mapping   WRITE
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_zv(fh, offset, msg, msglen)
    COMPLEX(kind=real_8), INTENT(IN)                      :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_zv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    WRITE(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_write_at_all_zv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_write_at_all_z(fh, offset, msg)
    COMPLEX(kind=real_8), INTENT(IN)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_write_at_all_z', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    WRITE(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_write_at_all_z

! *****************************************************************************
!> \brief (parallel) Blocking individual file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param[in] fh     file handle (file storage unit)
!> \param[in] offset file offset (position)
!> \param[out] msg   data to be read from the file
!> \param msglen ...
!> \par MPI-I/O mapping   mpi_file_read_at
!> \par STREAM-I/O mapping   READ
!> \param[in](optional) msglen  number of elements of data
! *****************************************************************************
  SUBROUTINE mp_file_read_at_zv(fh, offset, msg, msglen)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER                                    :: msg_len
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_zv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_zv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_z(fh, offset, msg)
    COMPLEX(kind=real_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset


    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_z', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_z

! *****************************************************************************
!> \brief (parallel) Blocking collective file read using explicit offsets
!>        (serial) Unformatted stream read
!> \param fh ...
!> \param offset ...
!> \param msg ...
!> \param msglen ...
!> \par MPI-I/O mapping    mpi_file_read_at_all
!> \par STREAM-I/O mapping   READ
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_zv(fh, offset, msg, msglen)
    COMPLEX(kind=real_8), INTENT(OUT)                     :: msg(:)
    INTEGER, INTENT(IN)                        :: fh
    INTEGER, INTENT(IN), OPTIONAL              :: msglen
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_zv', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr, msg_len

    ierr = 0
    msg_len = SIZE(msg)
    IF (PRESENT(msglen)) msg_len = msglen
    READ(UNIT=fh, POS=offset+1) msg(1:msg_len)
  END SUBROUTINE mp_file_read_at_all_zv

! *****************************************************************************
!> \brief ...
!> \param fh ...
!> \param offset ...
!> \param msg ...
! *****************************************************************************
  SUBROUTINE mp_file_read_at_all_z(fh, offset, msg)
    COMPLEX(kind=real_8), INTENT(OUT)               :: msg
    INTEGER, INTENT(IN)                        :: fh
    INTEGER(kind=file_offset), INTENT(IN)      :: offset

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_file_read_at_all_z', &
                                   routineP = moduleN//':'//routineN

    INTEGER                                    :: ierr

    ierr = 0
    READ(UNIT=fh, POS=offset+1) msg
  END SUBROUTINE mp_file_read_at_all_z

! *****************************************************************************
!> \brief ...
!> \param ptr ...
!> \param vector_descriptor ...
!> \param index_descriptor ...
!> \retval type_descriptor ...
! *****************************************************************************
  FUNCTION mp_type_make_z (ptr,&
       vector_descriptor, index_descriptor) &
       RESULT (type_descriptor)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER                    :: ptr
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL       :: vector_descriptor
    TYPE(mp_indexing_meta_type), INTENT(IN), OPTIONAL :: index_descriptor
    TYPE(mp_type_descriptor_type)                     :: type_descriptor

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_type_make_z', &
         routineP = moduleN//':'//routineN

    INTEGER :: ierr

    ierr = 0
    NULLIFY (type_descriptor%subtype)
    type_descriptor%length = SIZE (ptr)
    type_descriptor%type_handle = 7
    type_descriptor%vector_descriptor(1:2) = 1
    type_descriptor%has_indexing = .FALSE.
    type_descriptor%data_z => ptr
    IF (PRESENT (vector_descriptor) .OR. PRESENT (index_descriptor)) THEN
       CALL cp__b("mpiwrap/message_passing.F",3665,routineN//": Vectors and indices NYI")
    ENDIF
  END FUNCTION mp_type_make_z

! *****************************************************************************
!> \brief Allocates an array, using MPI_ALLOC_MEM ... this is hackish
!>        as the Fortran version returns an integer, which we take to be a C_PTR
!> \param DATA           data array to allocate
!> \param[in] len        length (in data elements) of data array allocation
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_alloc_mem_z(DATA, len, stat)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: DATA
    INTEGER, INTENT(IN)                      :: len
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     INTEGER                                 :: length, mystat
     length = MAX(len,1)
     IF (PRESENT (stat)) THEN
        ALLOCATE(DATA(length), stat=mystat)
        stat = mystat ! show to convention checker that stat is used
     ELSE
        ALLOCATE(DATA(length))
     ENDIF
   END SUBROUTINE mp_alloc_mem_z

! *****************************************************************************
!> \brief Deallocates am array, ... this is hackish
!>        as the Fortran version takes an integer, which we hope to get by reference
!> \param DATA           data array to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
   SUBROUTINE mp_free_mem_z(DATA, stat)
    COMPLEX(kind=real_8), DIMENSION(:), &
      POINTER                                :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat

     DEALLOCATE(DATA)
     IF (PRESENT (stat)) stat=0
   END SUBROUTINE mp_free_mem_z

! **************************************************************************************************
!> \brief Tests the MPI library
!> \param comm the relevant, initialized communicator
!> \param npow number of sizes to test, 10**1 .. 10**npow
!> \param output_unit where to direct output
!> \par History
!>      JGH  6-Feb-2001 : Test and performance code
!> \author JGH  1-JAN-2001
!> \note
!>      quickly adapted benchmark code, will only work on an even number of CPUs.
! **************************************************************************************************
   SUBROUTINE mpi_perf_test(comm, npow, output_unit)

      INTEGER, INTENT(IN)                      :: comm, npow, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mpi_perf_test', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(npow))==-1) EXIT ;  END DO ; ENDIF
      IF (output_unit > 0) WRITE (output_unit, *) "No MPI tests for a serial program"
   END SUBROUTINE mpi_perf_test

! **************************************************************************************************
!> \brief Starts a timer region
!> \param routineN ...
!> \param handle ...
! **************************************************************************************************
   SUBROUTINE mp_timeset(routineN, handle)
      CHARACTER(len=*), INTENT(IN)                       :: routineN
      INTEGER, INTENT(OUT)                               :: handle

      IF (mp_collect_timings) &
         CALL timeset(routineN, handle)
   END SUBROUTINE mp_timeset

! **************************************************************************************************
!> \brief Ends a timer region
!> \param handle ...
! **************************************************************************************************
   SUBROUTINE mp_timestop(handle)
      INTEGER, INTENT(IN)                                :: handle

      IF (mp_collect_timings) &
         CALL timestop(handle)
   END SUBROUTINE mp_timestop

END MODULE message_passing
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief various routines to log and control the output.
!>      The idea is that decisions about where to log should not be done in
!>      the code that generates the log, but should be globally changeable
!>      a central place.
!>      So some care has been taken to have enough information about the
!>      place from where the log comes so that in the future intelligent and
!>      flexible decisions can be taken by the logger, without having to change
!>      other code.
!> \note
!>      contains also routines to convert to a string.
!>      in my idea they should have been with variable length,
!>      (i.e. they should have returned a trim(adjustl(actual_result)))
!>      As a logger should be robust, at the moment I have given up.
!>
!>      At the moment logging and output refer to the same object
!>      (cp_logger_type)
!>      as these are actually different it might be better to separate them
!>      (they have already separate routines in a separate module
!>      @see cp_output_handling).
!>
!>      some practices (use of print *, no cp_error_type,
!>      manual retain release of some objects) are dictated by the need to
!>      have minimal dependency
!> \par History
!>      08.2002 major update: retain, release, printkeys, para_env,
!>              local logging [fawzi]
!>      02.2004 made a stack of default loggers [Joost VandeVondele]
!> \par
!>      @see cp_error_handling
!> \author Fawzi Mohamed
!>      @version 12.2001
! **************************************************************************************************
MODULE cp_log_handling
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_iter_types,                   ONLY: cp_iteration_info_create,&
                                              cp_iteration_info_release,&
                                              cp_iteration_info_retain,&
                                              cp_iteration_info_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: default_output_unit,&
                                              m_getpid,&
                                              m_hostnm
   USE message_passing,                 ONLY: mp_comm_free
   USE string_utilities,                ONLY: compress
   USE timings,                         ONLY: print_stack
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   !API types
   PUBLIC :: cp_logger_type, cp_logger_p_type
   !API parameter vars
   PUBLIC :: cp_note_level, cp_warning_level, cp_failure_level, cp_fatal_level
   !API default loggers
   PUBLIC :: cp_get_default_logger, cp_add_default_logger, cp_rm_default_logger, &
             cp_default_logger_stack_size
   !API logger routines
   PUBLIC :: cp_logger_create, cp_logger_retain, cp_logger_release, &
             cp_logger_would_log, cp_logger_set, cp_logger_get_default_unit_nr, &
             cp_logger_get_default_io_unit, cp_logger_get_unit_nr, &
             cp_logger_set_log_level, cp_logger_generate_filename, &
             cp_to_string

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_log_handling'
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   !! level of an error
   INTEGER, PARAMETER          :: cp_fatal_level = 3
   !! level of a failure
   INTEGER, PARAMETER          :: cp_failure_level = 2
   !! level of a warning
   INTEGER, PARAMETER          :: cp_warning_level = 1
   !! level of a note
   INTEGER, PARAMETER          :: cp_note_level = 0

   !! a generic function to trasform different types to strings
   INTERFACE cp_to_string
      MODULE PROCEDURE cp_int_to_string, cp_real_dp_to_string, cp_logical_to_string
   END INTERFACE

! **************************************************************************************************
!> \brief type of a logger, at the moment it contains just a print level
!>      starting at which level it should be logged
!>      (0 note, 1 warning, 2 failure, 3 fatal)
!>      it could be expanded with the ability to focus on one or more
!>      module/object/thread/processor
!> \param id_nr unique number to identify the logger
!> \param ref_count reference count (see cp2k/doc/ReferenceCounting.html)
!> \param print_level the level starting at which something gets printed
!> \param default_local_unit_nr default unit for local logging (-1 if not
!>        yet initialized). Local logging guarantee to each task its own
!>        file.
!> \param default_global_unit_nr default unit for global logging
!>        (-1 if not yet initialized). This unit is valid only on the
!>        processor with %para_env%mepos==%para_env%source.
!> \param para_env the parallel environement for the output.
!>        this might be a super environment of your computation environment
!>        i.e. be very careful not to do global operations like broadcast
!>        with a subset of its processors (use your computation environment
!>        instead).
!> \param close_local_unit_on_dealloc if the local unit should be closed
!>        when this logger is deallocated
!> \param close_global_unit_on_dealloc whether the global unit should be
!>        closed when this logger is deallocated
!> \param suffix a short string that is used as suffix in all the filenames
!>        created by this logger. Can be used to guarantee the unicity of
!>        generated filename
!> \param local_filename the root of the name of the file used for local
!>        logging (can be different from the name of the file corresponding
!>        to default_local_unit_nr, only the one used if the unit needs to
!>        be opened)
!> \param global_filename the root of the name of the file used for
!>        global logging (can be different from the name of the file
!>        corresponding to default_global_unit_nr, only the one used if
!>        the unit needs to be opened)
!> \param print_keys print keys that tell what should be logged/outputted
!> \note
!>      This should be private, but as the output functions have been
!>      moved to another module and there is no "friend" keyword, they
!>      are public.
!>      DO NOT USE THE INTERNAL COMPONENTS DIRECTLY!!!
!> \par History
!>      04.2002 revised [fawzi]
!>      08.2002 major update: retain, release, printkeys, para_env,
!>              local logging [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_logger_type
      INTEGER :: id_nr, ref_count
      INTEGER :: print_level
      INTEGER :: default_local_unit_nr
      INTEGER :: default_global_unit_nr
      LOGICAL :: close_local_unit_on_dealloc, close_global_unit_on_dealloc
      CHARACTER(len=default_string_length)  :: suffix
      CHARACTER(len=default_path_length)    :: local_filename, global_filename
      TYPE(cp_para_env_type), POINTER       :: para_env
      TYPE(cp_iteration_info_type), POINTER :: iter_info
   END TYPE cp_logger_type

   TYPE cp_logger_p_type
      TYPE(cp_logger_type), POINTER :: p => Null()
   END TYPE cp_logger_p_type

! **************************************************************************************************
   TYPE default_logger_stack_type
      TYPE(cp_logger_type), POINTER :: cp_default_logger => Null()
   END TYPE default_logger_stack_type

   INTEGER, PRIVATE            :: stack_pointer = 0
   INTEGER, PARAMETER, PRIVATE :: max_stack_pointer = 10
   TYPE(default_logger_stack_type), SAVE, DIMENSION(max_stack_pointer) ::  default_logger_stack

   INTEGER, SAVE, PRIVATE :: last_logger_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_default_logger_stack_size() RESULT(res)
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_default_logger_stack_size', &
         routineP = moduleN//':'//routineN

      res = stack_pointer
   END FUNCTION cp_default_logger_stack_size

! **************************************************************************************************
!> \brief adds a default logger.
!>      MUST be called before logging occours
!> \param logger ...
!> \author Fawzi Mohamed
!> \note
!>      increments a stack of default loggers the latest one will be
!>      available within the program
! **************************************************************************************************
   SUBROUTINE cp_add_default_logger(logger)
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_add_default_logger', &
         routineP = moduleN//':'//routineN

      IF (stack_pointer+1 > max_stack_pointer) THEN
         CALL cp_abort(cp__l("common/cp_log_handling.F",198), routineP// &
                       "too many default loggers, increase max_stack_pointer in "//moduleN)
      ENDIF

      stack_pointer = stack_pointer+1
      NULLIFY (default_logger_stack(stack_pointer)%cp_default_logger)

      default_logger_stack(stack_pointer)%cp_default_logger => logger
      CALL cp_logger_retain(logger)

   END SUBROUTINE cp_add_default_logger

! **************************************************************************************************
!> \brief the cousin of cp_add_default_logger, decrements the stack, so that
!>      the default logger is what it has
!>      been
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_rm_default_logger()
      IF (stack_pointer-1 < 0) THEN
         CALL cp_abort(cp__l("common/cp_log_handling.F",218), moduleN//":cp_rm_default_logger"// &
                       "can not destroy default logger "//moduleN)
      ENDIF

      CALL cp_logger_release(default_logger_stack(stack_pointer)%cp_default_logger)
      NULLIFY (default_logger_stack(stack_pointer)%cp_default_logger)
      stack_pointer = stack_pointer-1

   END SUBROUTINE cp_rm_default_logger

! **************************************************************************************************
!> \brief returns the default logger
!> \retval res ...
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      initializes the default loggers if necessary
! **************************************************************************************************
   FUNCTION cp_get_default_logger() RESULT(res)
      TYPE(cp_logger_type), POINTER                      :: res

      IF (.NOT. stack_pointer > 0) THEN
         CALL cp_abort(cp__l("common/cp_log_handling.F",241), "cp_log_handling:cp_get_default_logger"// &
                       "default logger not yet initialized (CALL cp_init_default_logger)")
      END IF
      res => default_logger_stack(stack_pointer)%cp_default_logger
      IF (.NOT. ASSOCIATED(res)) THEN
         CALL cp_abort(cp__l("common/cp_log_handling.F",246), "cp_log_handling:cp_get_default_logger"// &
                       "default logger is null (released too much ?)")
      END IF
   END FUNCTION cp_get_default_logger

! ================== log ==================

! **************************************************************************************************
!> \brief initializes a logger
!> \param logger the logger to initialize
!> \param para_env the parallel environment (this is most likely the global
!>        parallel environment
!> \param print_level the level starting with which something is written
!>        (defaults to cp_note_level)
!> \param default_global_unit_nr the default unit_nr for output
!>        (if not given, and no file is given defaults to the standard output)
!> \param default_local_unit_nr the default unit number for local (i.e. task)
!>        output. If not given defaults to a out.taskid file created upon
!> \param global_filename a new file to open (can be given instread of the
!>        global_unit_nr)
!> \param local_filename a new file to open (with suffix and para_env%mepos
!>        appended). Can be given instread of the default_local_unit_nr).
!>        the file is created only upon the first local logging request
!> \param close_global_unit_on_dealloc if the unit should be closed when the
!>        logger is deallocated (defaults to true if a local_filename is given,
!>        to false otherwise)
!> \param iter_info ...
!> \param close_local_unit_on_dealloc if the unit should be closed when the
!>        logger is deallocated (defaults to true)
!> \param suffix the suffix that should be added to all the generated filenames
!> \param template_logger a logger from where to take the unspecified things
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      the handling of *_filename, default_*_unit_nr, close_*_unit_on_dealloc
!>      tries to take the right decision with different inputs, and thus is a
!>      little complex.
! **************************************************************************************************
   SUBROUTINE cp_logger_create(logger, para_env, print_level, &
                               default_global_unit_nr, default_local_unit_nr, global_filename, &
                               local_filename, close_global_unit_on_dealloc, iter_info, &
                               close_local_unit_on_dealloc, suffix, template_logger)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      INTEGER, INTENT(in), OPTIONAL                      :: print_level, default_global_unit_nr, &
                                                            default_local_unit_nr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: global_filename, local_filename
      LOGICAL, INTENT(in), OPTIONAL                      :: close_global_unit_on_dealloc
      TYPE(cp_iteration_info_type), OPTIONAL, POINTER    :: iter_info
      LOGICAL, INTENT(in), OPTIONAL                      :: close_local_unit_on_dealloc
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: suffix
      TYPE(cp_logger_type), OPTIONAL, POINTER            :: template_logger

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: stat

      ALLOCATE (logger, stat=stat)
      IF (stat /= 0) &
         CALL cp__b("common/cp_log_handling.F",307,routineP//" could not ALLOCATE a logger")

      NULLIFY (logger%para_env)
      NULLIFY (logger%iter_info)
      logger%ref_count = 1
      last_logger_id_nr = last_logger_id_nr+1
      logger%id_nr = last_logger_id_nr

      IF (PRESENT(template_logger)) THEN
         IF (template_logger%ref_count < 1) &
            CALL cp__b("common/cp_log_handling.F",317,routineP//" template_logger%ref_count<1")
         logger%print_level = template_logger%print_level
         logger%default_global_unit_nr = template_logger%default_global_unit_nr
         logger%close_local_unit_on_dealloc = template_logger%close_local_unit_on_dealloc
         IF (logger%close_local_unit_on_dealloc) THEN
            logger%default_local_unit_nr = -1
         ELSE
            logger%default_local_unit_nr = template_logger%default_local_unit_nr
         END IF
         logger%close_global_unit_on_dealloc = template_logger%close_global_unit_on_dealloc
         IF (logger%close_global_unit_on_dealloc) THEN
            logger%default_global_unit_nr = -1
         ELSE
            logger%default_global_unit_nr = template_logger%default_global_unit_nr
         END IF
         logger%local_filename = template_logger%local_filename
         logger%global_filename = template_logger%global_filename
         logger%para_env => template_logger%para_env
         logger%suffix = template_logger%suffix
         logger%iter_info => template_logger%iter_info
      ELSE
         ! create a file if nothing is specified, one can also get the unit from the default logger
         ! which should have something reasonable as the argument is required in that case
         logger%default_global_unit_nr = -1
         logger%close_global_unit_on_dealloc = .TRUE.
         logger%local_filename = "localLog"
         logger%global_filename = "mainLog"
         logger%print_level = cp_note_level
         ! generate a file for default local logger
         ! except the ionode that should write to the default global logger
         logger%default_local_unit_nr = -1
         logger%close_local_unit_on_dealloc = .TRUE.
         logger%suffix = ""
      END IF
      IF (PRESENT(para_env)) logger%para_env => para_env
      IF (.NOT. ASSOCIATED(logger%para_env)) &
         CALL cp__b("common/cp_log_handling.F",353,routineP//" para env not associated")
      IF (logger%para_env%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",355,routineP//" para_env%ref_count<1")
      logger%para_env%ref_count = logger%para_env%ref_count+1

      IF (PRESENT(print_level)) logger%print_level = print_level

      IF (PRESENT(default_global_unit_nr)) &
         logger%default_global_unit_nr = default_global_unit_nr
      IF (PRESENT(global_filename)) THEN
         logger%global_filename = global_filename
         logger%close_global_unit_on_dealloc = .TRUE.
         logger%default_global_unit_nr = -1
      END IF
      IF (PRESENT(close_global_unit_on_dealloc)) THEN
         logger%close_global_unit_on_dealloc = close_global_unit_on_dealloc
         IF (PRESENT(default_global_unit_nr) .AND. PRESENT(global_filename) .AND. &
             (.NOT. close_global_unit_on_dealloc)) THEN
            logger%default_global_unit_nr = default_global_unit_nr
         END IF
      END IF

      IF (PRESENT(default_local_unit_nr)) &
         logger%default_local_unit_nr = default_local_unit_nr
      IF (PRESENT(local_filename)) THEN
         logger%local_filename = local_filename
         logger%close_local_unit_on_dealloc = .TRUE.
         logger%default_local_unit_nr = -1
      END IF
      IF (PRESENT(suffix)) logger%suffix = suffix

      IF (PRESENT(close_local_unit_on_dealloc)) THEN
         logger%close_local_unit_on_dealloc = close_local_unit_on_dealloc
         IF (PRESENT(default_local_unit_nr) .AND. PRESENT(local_filename) .AND. &
             (.NOT. close_local_unit_on_dealloc)) THEN
            logger%default_local_unit_nr = default_local_unit_nr
         END IF
      END IF

      IF (logger%default_local_unit_nr == -1) THEN
         IF (logger%para_env%mepos == logger%para_env%source) THEN
            logger%default_local_unit_nr = logger%default_global_unit_nr
            logger%close_local_unit_on_dealloc = .FALSE.
         ENDIF
      ENDIF
      IF (PRESENT(iter_info)) logger%iter_info => iter_info
      IF (ASSOCIATED(logger%iter_info)) THEN
         CALL cp_iteration_info_retain(logger%iter_info)
      ELSE
         CALL cp_iteration_info_create(logger%iter_info, "")
      END IF
   END SUBROUTINE cp_logger_create

! **************************************************************************************************
!> \brief retains the given logger (to be called to keep a shared copy of
!>      the logger)
!> \param logger the logger to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_logger_retain(logger)
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_retain', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(logger)) &
         CALL cp__b("common/cp_log_handling.F",421,routineP//" logger not associated")
      IF (logger%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",423,routineP//" logger%ref_count<1")
      logger%ref_count = logger%ref_count+1
   END SUBROUTINE cp_logger_retain

! **************************************************************************************************
!> \brief releases this logger
!> \param logger the logger to release
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_logger_release(logger)
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(logger)) THEN
         IF (logger%ref_count < 1) &
            CALL cp__b("common/cp_log_handling.F",442,routineP//" logger%ref_count<1")
         logger%ref_count = logger%ref_count-1
         IF (logger%ref_count == 0) THEN
            IF (logger%close_global_unit_on_dealloc .AND. &
                logger%default_global_unit_nr >= 0) THEN
               CALL close_file(logger%default_global_unit_nr)
               logger%close_global_unit_on_dealloc = .FALSE.
               logger%default_global_unit_nr = -1
            END IF
            IF (logger%close_local_unit_on_dealloc .AND. &
                logger%default_local_unit_nr >= 0) THEN
               CALL close_file(logger%default_local_unit_nr)
               logger%close_local_unit_on_dealloc = .FALSE.
               logger%default_local_unit_nr = -1
            END IF
            CALL my_cp_para_env_release(logger%para_env)
            CALL cp_iteration_info_release(logger%iter_info)
            DEALLOCATE (logger)
         END IF
      END IF
      NULLIFY (logger)
   END SUBROUTINE cp_logger_release

! **************************************************************************************************
!> \brief this function can be called to check if the logger would log
!>      a message with the given level from the given source
!>      you should use this function if you do direct logging
!>      (without using cp_logger_log), or if you want to know if the generation
!>      of some costly log info is necessary
!> \param logger the logger you want to log in
!> \param level describes the of the message: cp_fatal_level(3),
!>     cp_failure_level(2), cp_warning_level(1), cp_note_level(0).
!> \retval res ...
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_logger_would_log(logger, level) RESULT(res)
      TYPE(cp_logger_type), POINTER                      :: logger
      INTEGER, INTENT(in)                                :: level
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_would_log', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: lggr

      lggr => logger
      IF (.NOT. ASSOCIATED(lggr)) lggr => cp_get_default_logger()
      IF (lggr%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",492,routineP//" logger%ref_count<1")

      res = level >= lggr%print_level
   END FUNCTION cp_logger_would_log

! **************************************************************************************************
!> \brief returns the unit nr for the requested kind of log.
!> \param logger the logger you want to log in
!> \param local if true returns a local logger (one per task), otherwise
!>     returns a global logger (only the process with para_env%mepos==
!>     para_env%source should write to the global logger). Defaults to
!>     false
!> \retval res ...
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_logger_get_unit_nr(logger, local) RESULT(res)
      TYPE(cp_logger_type), POINTER                      :: logger
      LOGICAL, INTENT(in), OPTIONAL                      :: local
      INTEGER                                            :: res

      res = cp_logger_get_default_unit_nr(logger, local=local)
   END FUNCTION cp_logger_get_unit_nr

! **************************************************************************************************
!> \brief returns the unit nr for the ionode (-1 on all other processors)
!>        skips as well checks if the procs calling this function is not the ionode
!> \param logger the logger you want to log in
!> \retval res ...
!> \par History
!>      12.2009 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   FUNCTION cp_logger_get_default_io_unit(logger) RESULT(res)
      TYPE(cp_logger_type), OPTIONAL, POINTER            :: logger
      INTEGER                                            :: res

      TYPE(cp_logger_type), POINTER                      :: local_logger

      IF (PRESENT(logger)) THEN
         local_logger => logger
      ELSE
         local_logger => cp_get_default_logger()
      ENDIF

      res = cp_logger_get_default_unit_nr(local_logger, local=.FALSE., skip_not_ionode=.TRUE.)
   END FUNCTION cp_logger_get_default_io_unit

! *************************** cp_logger_type settings ***************************

! **************************************************************************************************
!> \brief changes the logging level. Log messages with a level less than the one
!>      given wo not be printed.
!> \param logger the logger to change
!> \param level the new logging level for the logger
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_logger_set_log_level(logger, level)
      TYPE(cp_logger_type), POINTER                      :: logger
      INTEGER, INTENT(in)                                :: level

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_set_log_level', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(logger)) &
         CALL cp__b("common/cp_log_handling.F",560,routineP//" logger not associated")
      IF (logger%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",562,routineP//" logger%ref_count<1")
      logger%print_level = level
   END SUBROUTINE cp_logger_set_log_level

! **************************************************************************************************
!> \brief asks the default unit number of the given logger.
!>      try to use cp_logger_get_unit_nr
!> \param logger the logger you want info from
!> \param local if you want the local unit nr (defaults to false)
!> \param skip_not_ionode ...
!> \retval res ...
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   RECURSIVE FUNCTION cp_logger_get_default_unit_nr(logger, local, skip_not_ionode) RESULT(res)
      TYPE(cp_logger_type), OPTIONAL, POINTER            :: logger
      LOGICAL, INTENT(in), OPTIONAL                      :: local, skip_not_ionode
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_get_default_unit_nr', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: filename, host_name
      INTEGER                                            :: iostat, pid
      LOGICAL                                            :: loc, skip
      TYPE(cp_logger_type), POINTER                      :: lggr

      loc = .TRUE.
      skip = .FALSE.
      IF (PRESENT(logger)) THEN
         lggr => logger
      ELSE
         NULLIFY (lggr)
      ENDIF
      IF (.NOT. ASSOCIATED(lggr)) lggr => cp_get_default_logger()
      IF (lggr%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",599,routineP//" logger%ref_count<1")

      IF (PRESENT(local)) loc = local
      IF (PRESENT(skip_not_ionode)) skip = skip_not_ionode
      IF (.NOT. loc) THEN
         IF (lggr%default_global_unit_nr <= 0) THEN
            IF (lggr%para_env%mepos == lggr%para_env%source) THEN
               CALL cp_logger_generate_filename(lggr, filename, lggr%global_filename, &
                                                ".out", local=.FALSE.)
               CALL open_file(TRIM(filename), file_status="unknown", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=lggr%default_global_unit_nr)
            ELSE IF (.NOT. skip) THEN
               lggr%default_global_unit_nr = cp_logger_get_default_unit_nr(lggr, .TRUE.)
               lggr%close_global_unit_on_dealloc = .FALSE.
            ELSE
               lggr%default_global_unit_nr = -1
               lggr%close_global_unit_on_dealloc = .FALSE.
            END IF
         END IF
         IF ((lggr%para_env%mepos /= lggr%para_env%source) .AND. (.NOT. skip)) THEN
            WRITE (UNIT=lggr%default_global_unit_nr, FMT='(/,T2,A)', IOSTAT=iostat) &
               ' *** WARNING non ionode asked for global logger ***'
            IF (iostat /= 0) THEN
               CALL m_getpid(pid)
               CALL m_hostnm(host_name)
               PRINT*, " *** Error trying to WRITE to the local logger ***"
               PRINT*, " *** MPI_id           = ", lggr%para_env%mepos
               PRINT*, " *** MPI_Communicator = ", lggr%para_env%group
               PRINT*, " *** PID              = ", pid
               PRINT*, " *** Hostname         = "//TRIM(host_name)
               CALL print_stack(default_output_unit)
            ELSE
               CALL print_stack(lggr%default_global_unit_nr)
            END IF
         END IF
         res = lggr%default_global_unit_nr
      ELSE
         IF (lggr%default_local_unit_nr <= 0) THEN
            CALL cp_logger_generate_filename(lggr, filename, lggr%local_filename, &
                                             ".out", local=.TRUE.)
            CALL open_file(TRIM(filename), file_status="unknown", &
                           file_action="WRITE", &
                           file_position="APPEND", &
                           unit_number=lggr%default_local_unit_nr)
            WRITE (UNIT=lggr%default_local_unit_nr, FMT='(/,T2,A,I0,A,I0,A)', IOSTAT=iostat) &
               '*** Local logger file of MPI task ', lggr%para_env%mepos, &
               ' in communicator ', lggr%para_env%group, ' ***'
            IF (iostat == 0) THEN
               CALL m_getpid(pid)
               CALL m_hostnm(host_name)
               WRITE (UNIT=lggr%default_local_unit_nr, FMT='(T2,A,I0)', IOSTAT=iostat) &
                  '*** PID      = ', pid, &
                  '*** Hostname = '//host_name
               CALL print_stack(lggr%default_local_unit_nr)
            END IF
            IF (iostat /= 0) THEN
               CALL m_getpid(pid)
               CALL m_hostnm(host_name)
               PRINT*, " *** Error trying to WRITE to the local logger ***"
               PRINT*, " *** MPI_id           = ", lggr%para_env%mepos
               PRINT*, " *** MPI_Communicator = ", lggr%para_env%group
               PRINT*, " *** PID              = ", pid
               PRINT*, " *** Hostname         = "//TRIM(host_name)
               CALL print_stack(default_output_unit)
            END IF

         END IF
         res = lggr%default_local_unit_nr
      END IF
   END FUNCTION cp_logger_get_default_unit_nr

! **************************************************************************************************
!> \brief releases the para object (to be called when you don't want anymore
!>      the shared copy of this object)
!> \param para_env the new group
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      local private version to avoid circular dependencies
!>      (see cp_para_env:cp_para_env_release)!
! **************************************************************************************************
   SUBROUTINE my_cp_para_env_release(para_env)
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'my_cp_para_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(para_env)) THEN
         IF (para_env%ref_count < 1) THEN
            CALL cp__b("common/cp_log_handling.F",690,routineP//" para_env%ref_count<1")
         END IF
         para_env%ref_count = para_env%ref_count-1
         IF (para_env%ref_count < 1) THEN
            IF (para_env%owns_group) THEN
               CALL mp_comm_free(para_env%group)
            END IF
            DEALLOCATE (para_env)
         END IF
      END IF
      NULLIFY (para_env)
   END SUBROUTINE my_cp_para_env_release

! **************************************************************************************************
!> \brief generates a unique filename (ie adding eventual suffixes and
!>      process ids)
!> \param logger ...
!> \param res the resulting string
!> \param root the start of filename
!> \param postfix the end of the name
!> \param local if the name should be local to this task (defaults to false)
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      this should be a function returning a variable length string.
!>      All spaces are moved to the end of the string.
!>      Not fully optimized: result must be a littel longer than the
!>      resulting compressed filename
! **************************************************************************************************
   SUBROUTINE cp_logger_generate_filename(logger, res, root, postfix, &
                                          local)
      TYPE(cp_logger_type), POINTER                      :: logger
      CHARACTER(len=*), INTENT(inout)                    :: res
      CHARACTER(len=*), INTENT(in)                       :: root, postfix
      LOGICAL, INTENT(in), OPTIONAL                      :: local

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_generate_filename', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: loc
      TYPE(cp_logger_type), POINTER                      :: lggr

      loc = .FALSE.
      res = ' '
      lggr => logger

      IF (.NOT. ASSOCIATED(lggr)) lggr => cp_get_default_logger()
      IF (lggr%ref_count < 1) &
         CALL cp__b("common/cp_log_handling.F",739,routineP//" logger%ref_count<1")
      IF (PRESENT(local)) loc = local
      IF (loc) THEN
         res = TRIM(root)//TRIM(lggr%suffix)//'_p'// &
               cp_to_string(lggr%para_env%mepos)//postfix
      ELSE
         res = TRIM(root)//TRIM(lggr%suffix)//postfix
      END IF
      CALL compress(res, full=.TRUE.)
   END SUBROUTINE cp_logger_generate_filename

! **************************************************************************************************
!> \brief sets various attributes of the given logger
!> \param logger the logger you want to change
!> \param local_filename the root of the name of the file used for local
!>        logging
!> \param global_filename the root of the name of the file used for
!>        global logging
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_logger_set(logger, local_filename, global_filename)
      TYPE(cp_logger_type), POINTER                      :: logger
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: local_filename, global_filename

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_logger_set', routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(logger)) &
         CALL cp__b("common/cp_log_handling.F",766,routineP//" unassociated logger")
      IF (PRESENT(local_filename)) logger%local_filename = local_filename
      IF (PRESENT(global_filename)) logger%global_filename = global_filename
   END SUBROUTINE cp_logger_set

! **************************************************************************************************
!> \brief converts an int to a string
!>      (should be a variable length string, but that does not work with
!>      all the compilers)
!> \param i the integer to convert
!> \retval res ...
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_int_to_string(i) RESULT(res)
      INTEGER, INTENT(in)                                :: i
      CHARACTER(len=6)                                   :: res

      CHARACTER(len=6)                                   :: t_res
      INTEGER                                            :: iostat
      REAL(KIND=dp)                                      :: tmp_r

      iostat = 0
      IF (i > 999999 .OR. i < -99999) THEN
         tmp_r = i
         WRITE (t_res, fmt='(es6.1)', iostat=iostat) tmp_r
      ELSE
         WRITE (t_res, fmt='(i6)', iostat=iostat) i
      END IF
      res = t_res
      IF (iostat /= 0) THEN
         PRINT*, "cp_int_to_string ioerror", iostat
         CALL print_stack(cp_logger_get_default_unit_nr())
      END IF
   END FUNCTION cp_int_to_string

! **************************************************************************************************
!> \brief convert a double precision real in a string
!>      (should be a variable length string, but that does not work with
!>      all the compilers)
!> \param val the number to convert
!> \retval res ...
!> \par History
!>      4.2002 revised [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_real_dp_to_string(val) RESULT(res)
      REAL(KIND=dp), INTENT(in)                          :: val
      CHARACTER(len=11)                                  :: res

      WRITE (res, '(es11.4)') val
   END FUNCTION cp_real_dp_to_string

! **************************************************************************************************
!> \brief convert a logical in a string ('T' or 'F')
!> \param val the number to convert
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_logical_to_string(val) RESULT(res)
      LOGICAL, INTENT(in)                                :: val
      CHARACTER(len=1)                                   :: res

      IF (val) THEN
         res = 'T'
      ELSE
         res = 'F'
      END IF
   END FUNCTION cp_logical_to_string

END MODULE cp_log_handling
