!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Self-consistent continuum solvation (SCCS) model implementation
!> \author  Matthias Krack (MK)
!> \version 1.0
!> \par Literature:
!>      - J.-L. Fattebert and F. Gygi,
!>        Density functional theory for efficient ab initio molecular dynamics
!>        simulations in solution, J. Comput. Chem. 23, 662-666 (2002)
!>      - O. Andreussi, I. Dabo, and N. Marzari,
!>        Revised self-consistent continuum solvation in electronic-structure
!>        calculations, J. Chem. Phys. 136, 064102-20 (2012)
!> \par History:
!>      - Creation (10.10.2013,MK)
!>      - Derivatives using finite differences (26.11.2013,MK)
!>      - Cube file dump of the dielectric function (19.12.2013,MK)
!>      - Cube file dump of the polarisation potential (20.12.2013,MK)
!>      - Calculation of volume and surface of the cavity (21.12.2013,MK)
!>      - Functional derivative of the cavitation energy (28.12.2013,MK)
! **************************************************************************************************

MODULE qs_sccs

   USE cp_control_types,                ONLY: dft_control_type,&
                                              sccs_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr,&
                                              low_print_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cp_realspace_grid_init,          ONLY: init_input_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_constants,                 ONLY: sccs_andreussi,&
                                              sccs_derivative_cd3,&
                                              sccs_derivative_cd5,&
                                              sccs_derivative_cd7,&
                                              sccs_derivative_fft,&
                                              sccs_fattebert_gygi
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sum
   USE particle_list_types,             ONLY: particle_list_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_type,&
                                              realspace_grid_input_type,&
                                              realspace_grid_type,&
                                              rs_grid_create,&
                                              rs_grid_create_descriptor,&
                                              rs_grid_release,&
                                              rs_grid_release_descriptor
   USE rs_methods,                      ONLY: derive_fdm_cd3,&
                                              derive_fdm_cd5,&
                                              derive_fdm_cd7
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_sccs'

   PUBLIC :: sccs

CONTAINS

! **************************************************************************************************
!> \brief   Self-consistent continuum solvation (SCCS) model implementation
!> \param qs_env ...
!> \param rho_tot_gspace ...
!> \param v_hartree_gspace ...
!> \param v_sccs ...
!> \param h_stress ...
!> \par History:
!>      - Creation (10.10.2013,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE sccs(qs_env, rho_tot_gspace, v_hartree_gspace, v_sccs, h_stress)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_type), POINTER                             :: rho_tot_gspace, v_hartree_gspace, v_sccs
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT), &
         OPTIONAL                                        :: h_stress

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sccs', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: tol = 1.0E-12_dp

      CHARACTER(LEN=4*default_string_length)             :: message
      CHARACTER(LEN=default_string_length)               :: cube_path, filename, my_pos_cube, &
                                                            print_path
      INTEGER                                            :: cube_unit, di, dj, handle, i, ispin, &
                                                            iter, j, k, nspin, output_unit, &
                                                            print_level
      INTEGER(KIND=int_8)                                :: ngpts
      INTEGER, DIMENSION(3)                              :: lb, ub
      LOGICAL                                            :: append_cube, calculate_stress_tensor, &
                                                            should_output
      REAL(KIND=dp) :: cavity_surface, cavity_volume, cell_volume, dphi2, dvol, eps0, f, &
         norm_drho2, polarisation_charge, rho_delta, rho_delta_avg, rho_delta_max, rho_iter_new, &
         tot_rho_elec, tot_rho_solute
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: cp_subsys
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(3)                      :: dln_eps_elec, dphi_tot, drho_elec
      TYPE(pw_p_type), DIMENSION(3, 3)                   :: d2rho_elec
      TYPE(pw_p_type), DIMENSION(5)                      :: work_r3d
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_p_type), POINTER                           :: rho_r_sccs
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER :: deps_elec, eps_elec, ln_eps_elec, norm_drho_elec, phi_pol, &
         phi_solute, rho_elec, rho_iter_old, rho_solute, rho_tot, rho_tot_zero, theta
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(sccs_control_type), POINTER                   :: sccs_control
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)

      NULLIFY (auxbas_pw_pool)
      NULLIFY (cp_subsys)
      NULLIFY (deps_elec)
      NULLIFY (dft_control)
      NULLIFY (energy)
      NULLIFY (eps_elec)
      NULLIFY (input)
      NULLIFY (ln_eps_elec)
      NULLIFY (logger)
      NULLIFY (norm_drho_elec)
      NULLIFY (para_env)
      NULLIFY (particles)
      NULLIFY (phi_pol)
      NULLIFY (phi_solute)
      NULLIFY (poisson_env)
      NULLIFY (pw_env)
      NULLIFY (pw_pools)
      NULLIFY (rho)
      NULLIFY (rho_elec)
      NULLIFY (rho_iter_old)
      NULLIFY (rho_solute)
      NULLIFY (rho_tot)
      NULLIFY (rho_tot_zero)
      NULLIFY (sccs_control)
      NULLIFY (scf_env)
      NULLIFY (theta)

      ! Load data from Quickstep environment
      CALL get_qs_env(qs_env=qs_env, &
                      cp_subsys=cp_subsys, &
                      dft_control=dft_control, &
                      energy=energy, &
                      input=input, &
                      para_env=para_env, &
                      pw_env=pw_env, &
                      rho=rho, &
                      scf_env=scf_env)
      CALL cp_subsys_get(cp_subsys, particles=particles)

      sccs_control => dft_control%sccs_control

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_sccs.F",212)
      IF(.NOT.(ASSOCIATED(rho_tot_gspace)))CALL cp__a("qs_sccs.F",213)
      IF(.NOT.(ASSOCIATED(v_hartree_gspace)))CALL cp__a("qs_sccs.F",214)
      IF(.NOT.(ASSOCIATED(v_sccs)))CALL cp__a("qs_sccs.F",215)

      IF (PRESENT(h_stress)) THEN
         calculate_stress_tensor = .TRUE.
         h_stress(:, :) = 0.0_dp
         CALL cp__b("qs_sccs.F",220,"No stress tensor is implemented for SCCS models yet")
      ELSE
         calculate_stress_tensor = .FALSE.
      END IF

      ! Get access to the PW grid pool
      CALL pw_env_get(pw_env, &
                      auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, &
                      poisson_env=poisson_env)

      CALL pw_zero(v_sccs)

      ! Calculate no SCCS contribution, if the requested SCF convergence threshold is not reached yet
      IF (.NOT. sccs_control%sccs_activated) THEN
         IF (sccs_control%eps_scf > 0.0_dp) THEN
            IF ((scf_env%iter_delta > sccs_control%eps_scf) .OR. &
                ((qs_env%scf_env%outer_scf%iter_count == 0) .AND. &
                 (qs_env%scf_env%iter_count <= 1))) THEN
               CALL pw_poisson_solve(poisson_env=poisson_env, &
                                     density=rho_tot_gspace, &
                                     ehartree=energy%hartree, &
                                     vhartree=v_hartree_gspace)
               energy%sccs_pol = 0.0_dp
               energy%sccs_cav = 0.0_dp
               energy%sccs_dis = 0.0_dp
               energy%sccs_rep = 0.0_dp
               CALL timestop(handle)
               RETURN
            END IF
         END IF
         sccs_control%sccs_activated = .TRUE.
      END IF

      nspin = dft_control%nspins

      ! Manage print output control
      logger => cp_get_default_logger()
      print_level = logger%iter_info%print_level
      print_path = "DFT%PRINT%SCCS"
      should_output = (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                        TRIM(print_path)), cp_p_file))
      output_unit = cp_print_key_unit_nr(logger, input, TRIM(print_path), &
                                         extension=".sccs", &
                                         ignore_should_output=should_output, &
                                         log_filename=.FALSE.)

      ! Get work storage for the 3d grids in r-space
      DO i = 1, SIZE(work_r3d)
         NULLIFY (work_r3d(i)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                work_r3d(i)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
      END DO

      ! Assign total electronic density in r-space
      rho_elec => work_r3d(1)%pw

      ! Retrieve grid parameters
      ngpts = rho_elec%pw_grid%ngpts
      dvol = rho_elec%pw_grid%dvol
      cell_volume = rho_elec%pw_grid%vol
      abc(1:3) = REAL(rho_elec%pw_grid%npts(1:3), KIND=dp)*rho_elec%pw_grid%dr(1:3)
      lb(1:3) = rho_elec%pw_grid%bounds_local(1, 1:3)
      ub(1:3) = rho_elec%pw_grid%bounds_local(2, 1:3)

      ! Get rho
      CALL qs_rho_get(rho_struct=rho, &
                      rho_r=rho_r, &
                      rho_r_sccs=rho_r_sccs)

      ! Retrieve the last rho_iter from the previous SCCS cycle if available
      IF(.NOT.(ASSOCIATED(rho_r_sccs)))CALL cp__a("qs_sccs.F",293)
      rho_iter_old => rho_r_sccs%pw

      ! Retrieve the total electronic density in r-space
      CALL pw_copy(rho_r(1)%pw, rho_elec)
      DO ispin = 2, nspin
         CALL pw_axpy(rho_r(ispin)%pw, rho_elec)
      END DO
      tot_rho_elec = accurate_sum(rho_elec%cr3d)*dvol
      CALL mp_sum(tot_rho_elec, para_env%group)

      ! Calculate the dielectric (smoothed) function of rho_elec in r-space
      eps_elec => work_r3d(2)%pw
      deps_elec => work_r3d(3)%pw
      eps0 = sccs_control%epsilon_solvent
      SELECT CASE (sccs_control%method_id)
      CASE (sccs_andreussi)
         CALL andreussi(rho_elec, eps_elec, deps_elec, eps0, sccs_control%rho_max, &
                        sccs_control%rho_min)
      CASE (sccs_fattebert_gygi)
         CALL fattebert_gygi(rho_elec, eps_elec, deps_elec, eps0, sccs_control%beta, &
                             sccs_control%rho_zero)
      CASE DEFAULT
         CALL cp__b("qs_sccs.F",316,"Invalid method specified for SCCS model")
      END SELECT

      ! Optional output of the dielectric function in cube file format
      filename = "DIELECTRIC_FUNCTION"
      cube_path = TRIM(print_path)//"%"//TRIM(filename)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, TRIM(cube_path)), &
                cp_p_file)) THEN
         append_cube = section_get_lval(input, TRIM(cube_path)//"%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) my_pos_cube = "APPEND"
         cube_unit = cp_print_key_unit_nr(logger, input, TRIM(cube_path), &
                                          extension=".cube", middle_name=TRIM(filename), &
                                          file_position=my_pos_cube, log_filename=.FALSE.)
         IF (output_unit > 0) THEN
            INQUIRE (UNIT=cube_unit, NAME=filename)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
               "The dielectric function is written in cube file format to the file:", TRIM(filename)
         END IF
         CALL cp_pw_to_cube(eps_elec, cube_unit, TRIM(filename), particles=particles, &
                            stride=section_get_ivals(input, TRIM(cube_path)//"%STRIDE"))
         CALL cp_print_key_finished_output(cube_unit, logger, input, TRIM(cube_path))
      END IF

      ! Calculate the (quantum) volume and surface of the solute cavity
      cavity_surface = 0.0_dp
      cavity_volume = 0.0_dp

      IF (should_output .AND. (ABS(eps0-1.0_dp) > tol)) THEN

         ! Initialise the switching function theta
         theta => work_r3d(4)%pw
         CALL pw_zero(theta)

         ! Calculate the (quantum) volume of the solute cavity
         f = 1.0_dp/(eps0-1.0_dp)
!$OMP       PARALLEL DO DEFAULT(NONE) &
!$OMP                   PRIVATE(i,j,k) &
!$OMP                   SHARED(eps0,eps_elec,f,lb,theta,ub)
         DO k = lb(3), ub(3)
            DO j = lb(2), ub(2)
               DO i = lb(1), ub(1)
                  theta%cr3d(i, j, k) = f*(eps0-eps_elec%cr3d(i, j, k))
               END DO
            END DO
         END DO
!$OMP       END PARALLEL DO
         cavity_volume = accurate_sum(theta%cr3d)*dvol
         CALL mp_sum(cavity_volume, para_env%group)

         ! Calculate the derivative of the electronic density in r-space
         ! TODO: Could be retrieved from the QS environment
         DO i = 1, 3
            NULLIFY (drho_elec(i)%pw)
            CALL pw_pool_create_pw(auxbas_pw_pool, &
                                   drho_elec(i)%pw, &
                                   use_data=REALDATA3D, &
                                   in_space=REALSPACE)
         END DO
         CALL derive(rho_elec, drho_elec, sccs_derivative_fft, pw_env, input, para_env)

         ! Calculate the norm of the gradient of the electronic density in r-space
         norm_drho_elec => work_r3d(5)%pw
!$OMP       PARALLEL DO DEFAULT(NONE) &
!$OMP                   PRIVATE(i,j,k) &
!$OMP                   SHARED(drho_elec,lb,norm_drho_elec,ub)
         DO k = lb(3), ub(3)
            DO j = lb(2), ub(2)
               DO i = lb(1), ub(1)
                  norm_drho_elec%cr3d(i, j, k) = SQRT(drho_elec(1)%pw%cr3d(i, j, k)* &
                                                      drho_elec(1)%pw%cr3d(i, j, k)+ &
                                                      drho_elec(2)%pw%cr3d(i, j, k)* &
                                                      drho_elec(2)%pw%cr3d(i, j, k)+ &
                                                      drho_elec(3)%pw%cr3d(i, j, k)* &
                                                      drho_elec(3)%pw%cr3d(i, j, k))
               END DO
            END DO
         END DO
!$OMP       END PARALLEL DO

         ! Optional output of the norm of the density gradient in cube file format
         filename = "DENSITY_GRADIENT"
         cube_path = TRIM(print_path)//"%"//TRIM(filename)
         IF (BTEST(cp_print_key_should_output(logger%iter_info, input, TRIM(cube_path)), &
                   cp_p_file)) THEN
            append_cube = section_get_lval(input, TRIM(cube_path)//"%APPEND")
            my_pos_cube = "REWIND"
            IF (append_cube) my_pos_cube = "APPEND"
            cube_unit = cp_print_key_unit_nr(logger, input, TRIM(cube_path), &
                                             extension=".cube", middle_name=TRIM(filename), &
                                             file_position=my_pos_cube, log_filename=.FALSE.)
            IF (output_unit > 0) THEN
               INQUIRE (UNIT=cube_unit, NAME=filename)
               WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                  "The norm of the density gradient is written in cube file format to the file:", TRIM(filename)
            END IF
            CALL cp_pw_to_cube(norm_drho_elec, cube_unit, TRIM(filename), particles=particles, &
                               stride=section_get_ivals(input, TRIM(cube_path)//"%STRIDE"))
            CALL cp_print_key_finished_output(cube_unit, logger, input, TRIM(cube_path))
         END IF

         ! Calculate the (quantum) surface of the solute cavity
         SELECT CASE (sccs_control%method_id)
         CASE (sccs_andreussi)
            CALL surface_andreussi(rho_elec, norm_drho_elec, theta, eps0, &
                                   sccs_control%rho_max, sccs_control%rho_min, &
                                   sccs_control%delta_rho)
         CASE (sccs_fattebert_gygi)
            CALL surface_fattebert_gygi(rho_elec, norm_drho_elec, theta, eps0, &
                                        sccs_control%beta, sccs_control%rho_zero, &
                                        sccs_control%delta_rho)
         CASE DEFAULT
            CALL cp__b("qs_sccs.F",428,"Invalid method specified for SCCS model")
         END SELECT
         cavity_surface = accurate_sum(theta%cr3d)*dvol
         CALL mp_sum(cavity_surface, para_env%group)

         IF (sccs_control%gamma_solvent > 0.0_dp) THEN

            CALL cp_warn(cp__l("qs_sccs.F",435), &
                         "Experimental feature requested: The contribution from the "// &
                         "cavitation energy is added as an extra potential term to the "// &
                         "Kohn-Sham potential")

            ! Calculate the second derivatives of the electronic density in r-space
            DO di = 1, 3
               DO dj = 1, 3
                  NULLIFY (d2rho_elec(dj, di)%pw)
                  CALL pw_pool_create_pw(auxbas_pw_pool, &
                                         d2rho_elec(dj, di)%pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
               END DO
               CALL derive(drho_elec(di)%pw, d2rho_elec(:, di), sccs_derivative_fft, pw_env, &
                           input, para_env)
            END DO

            ! Calculate the contribution of the cavitation energy to the Kohn-Sham potential
!$OMP         PARALLEL DO DEFAULT(NONE) &
!$OMP                     PRIVATE(di,dj,i,j,k,norm_drho2) &
!$OMP                     SHARED(d2rho_elec,drho_elec,lb,norm_drho_elec,sccs_control) &
!$OMP                     SHARED(theta,ub,v_sccs)
            DO k = lb(3), ub(3)
               DO j = lb(2), ub(2)
                  DO i = lb(1), ub(1)
                     norm_drho2 = norm_drho_elec%cr3d(i, j, k)*norm_drho_elec%cr3d(i, j, k)
                     DO di = 1, 3
                        DO dj = 1, 3
                           v_sccs%cr3d(i, j, k) = sccs_control%gamma_solvent*theta%cr3d(i, j, k)* &
                                                  (drho_elec(di)%pw%cr3d(i, j, k)* &
                                                   drho_elec(dj)%pw%cr3d(i, j, k)* &
                                                   d2rho_elec(di, dj)%pw%cr3d(i, j, k)/norm_drho2- &
                                                   d2rho_elec(di, di)%pw%cr3d(i, j, k))/norm_drho2
                        END DO
                     END DO
                  END DO
               END DO
            END DO
!$OMP         END PARALLEL DO
            CALL pw_scale(v_sccs, dvol)
            DO di = 1, 3
               DO dj = 1, 3
                  CALL pw_pool_give_back_pw(auxbas_pw_pool, d2rho_elec(dj, di)%pw)
               END DO
            END DO
         END IF

         ! Release storage
         NULLIFY (theta)
         DO i = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, drho_elec(i)%pw)
         END DO

      END IF

      ! Retrieve the total charge density (core + elec) of the solute in r-space
      rho_solute => work_r3d(4)%pw
      CALL pw_zero(rho_solute)
      CALL pw_transfer(rho_tot_gspace, rho_solute)
      tot_rho_solute = accurate_sum(rho_solute%cr3d)*dvol
      CALL mp_sum(tot_rho_solute, para_env%group)

      ! Reassign work storage to rho_tot_zero, because rho_elec is no longer needed
      rho_tot_zero => rho_elec
      NULLIFY (rho_elec)

      ! Build the initial (rho_iter = 0) total charge density (solute plus polarisation) in r-space
      ! eps_elec <- ln(eps_elec)
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(i,j,k) &
!$OMP                 SHARED(eps_elec,lb,message,output_unit,para_env,ub) &
!$OMP                 SHARED(rho_solute,rho_tot_zero)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               IF (eps_elec%cr3d(i, j, k) < 1.0E-12_dp) THEN
                  WRITE (UNIT=message, FMT="(A,ES12.3,A,3(I0,A))") &
                     "SCCS| Invalid dielectric function value ", eps_elec%cr3d(i, j, k), &
                     " encountered at grid point (", i, ",", j, ",", k, ")"
                  CALL cp__b("qs_sccs.F",515,message)
               END IF
               rho_tot_zero%cr3d(i, j, k) = rho_solute%cr3d(i, j, k)/eps_elec%cr3d(i, j, k)
               eps_elec%cr3d(i, j, k) = LOG(eps_elec%cr3d(i, j, k))
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      ! Reassign pointers due to new content
      ln_eps_elec => eps_elec
      NULLIFY (eps_elec)

      ! Build the derivative of ln_eps_elec
      DO i = 1, 3
         NULLIFY (dln_eps_elec(i)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                dln_eps_elec(i)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_zero(dln_eps_elec(i)%pw)
      END DO
      CALL derive(ln_eps_elec, dln_eps_elec, sccs_control%derivative_method, pw_env, input, para_env)

      ! Print header for the SCCS cycle
      IF (should_output .AND. (output_unit > 0)) THEN
         IF (print_level > low_print_level) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
               "SCCS| Total electronic charge density ", -tot_rho_elec, &
               "SCCS| Total charge density (solute)   ", -tot_rho_solute
            WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.3)") &
               "SCCS| Surface of the solute cavity [bohr^2]", cavity_surface, &
               "SCCS|                              [angstrom^2]", &
               cp_unit_from_cp2k(cavity_surface, "angstrom^2"), &
               "SCCS| Volume of the solute cavity  [bohr^3]", cavity_volume, &
               "SCCS|                              [angstrom^3]", &
               cp_unit_from_cp2k(cavity_volume, "angstrom^3"), &
               "SCCS| Volume of the cell           [bohr^3]", cell_volume, &
               "SCCS|                              [angstrom^3]", &
               cp_unit_from_cp2k(cell_volume, "angstrom^3")
            WRITE (UNIT=output_unit, FMT="(T3,A)") &
               "SCCS|", &
               "SCCS|   Step    Average residual    Maximum residual"
         END IF
      END IF

      ! Get storage for the derivative of the total potential (field) in r-space
      DO i = 1, 3
         NULLIFY (dphi_tot(i)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                dphi_tot(i)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
      END DO

      ! Reassign work storage to rho_tot, because ln_eps_elec is no longer needed
      rho_tot => ln_eps_elec
      NULLIFY (ln_eps_elec)

      ! Initialise the total electronic density in r-space rho_tot with rho_tot_zero + rho_iter_zero
      CALL pw_copy(rho_tot_zero, rho_tot)
      CALL pw_axpy(rho_iter_old, rho_tot)

      ! Main SCCS iteration loop
      iter = 0

      iter_loop: DO

         ! Increment iteration counter
         iter = iter+1

         ! Check if the requested maximum number of SCCS iterations is reached
         IF (iter > sccs_control%max_iter) THEN
            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T3,A,/,T3,A,I0,A)") &
                  "SCCS| Maximum number of SCCS iterations reached", &
                  "SCCS| Iteration cycle did not converge in ", sccs_control%max_iter, " steps"
            END IF
            EXIT iter_loop
         END IF

         ! Calculate derivative of the current total potential in r-space
         CALL pw_poisson_solve(poisson_env=poisson_env, &
                               density=rho_tot, &
                               dvhartree=dphi_tot)

         ! Update total charge density (solute plus polarisation) in r-space
         ! based on the iterated polarisation charge density
         f = 0.5_dp/twopi
         rho_delta_avg = 0.0_dp
         rho_delta_max = 0.0_dp
!$OMP       PARALLEL DO DEFAULT(NONE) &
!$OMP                   PRIVATE(i,j,k,rho_delta,rho_iter_new) &
!$OMP                   SHARED(dln_eps_elec,dphi_tot,f,lb,rho_iter_old,ub) &
!$OMP                   SHARED(rho_tot,rho_tot_zero,sccs_control) &
!$OMP                   REDUCTION(+:rho_delta_avg) &
!$OMP                   REDUCTION(MAX:rho_delta_max)
         DO k = lb(3), ub(3)
            DO j = lb(2), ub(2)
               DO i = lb(1), ub(1)
                  rho_iter_new = (dln_eps_elec(1)%pw%cr3d(i, j, k)*dphi_tot(1)%pw%cr3d(i, j, k)+ &
                                  dln_eps_elec(2)%pw%cr3d(i, j, k)*dphi_tot(2)%pw%cr3d(i, j, k)+ &
                                  dln_eps_elec(3)%pw%cr3d(i, j, k)*dphi_tot(3)%pw%cr3d(i, j, k))*f
                  rho_iter_new = rho_iter_old%cr3d(i, j, k)+ &
                                 sccs_control%mixing*(rho_iter_new-rho_iter_old%cr3d(i, j, k))
                  rho_delta = ABS(rho_iter_new-rho_iter_old%cr3d(i, j, k))
                  rho_delta_max = MAX(rho_delta, rho_delta_max)
                  rho_delta_avg = rho_delta_avg+rho_delta
                  rho_tot%cr3d(i, j, k) = rho_tot_zero%cr3d(i, j, k)+rho_iter_new
                  rho_iter_old%cr3d(i, j, k) = rho_iter_new
               END DO
            END DO
         END DO
!$OMP       END PARALLEL DO

         CALL mp_sum(rho_delta_avg, para_env%group)
         rho_delta_avg = rho_delta_avg/REAL(ngpts, KIND=dp)
         CALL mp_max(rho_delta_max, para_env%group)

         IF (should_output .AND. (output_unit > 0)) THEN
            IF (print_level > low_print_level) THEN
               IF ((ABS(rho_delta_avg) < 1.0E-8_dp) .OR. &
                   (ABS(rho_delta_avg) >= 1.0E5_dp)) THEN
                  WRITE (UNIT=output_unit, FMT="(T3,A,I6,4X,ES16.4,4X,ES16.4)") &
                     "SCCS| ", iter, rho_delta_avg, rho_delta_max
               ELSE
                  WRITE (UNIT=output_unit, FMT="(T3,A,I6,4X,F16.8,4X,F16.8)") &
                     "SCCS| ", iter, rho_delta_avg, rho_delta_max
               END IF
            END IF
         END IF

         ! Check if the SCCS iteration cycle is converged to the requested tolerance
         IF (rho_delta_max <= sccs_control%eps_sccs) THEN
            IF (should_output .AND. (output_unit > 0)) THEN
               WRITE (UNIT=output_unit, FMT="(T3,A,I0,A)") &
                  "SCCS| Iteration cycle converged in ", iter, " steps"
            END IF
            EXIT iter_loop
         END IF

      END DO iter_loop

      ! Calculate the total Hartree energy, potential, and its derivatives of
      ! the solute and the implicit solvent
      CALL pw_transfer(rho_tot, rho_tot_gspace)
      IF (calculate_stress_tensor) THEN
         CALL pw_poisson_solve(poisson_env=poisson_env, &
                               density=rho_tot_gspace, &
                               ehartree=energy%sccs_hartree, &
                               vhartree=v_hartree_gspace, &
                               dvhartree=dphi_tot, &
                               h_stress=h_stress)
      ELSE
         CALL pw_poisson_solve(poisson_env=poisson_env, &
                               density=rho_tot_gspace, &
                               ehartree=energy%sccs_hartree, &
                               vhartree=v_hartree_gspace, &
                               dvhartree=dphi_tot)
      END IF
      phi_pol => work_r3d(5)%pw
      CALL pw_transfer(v_hartree_gspace, phi_pol)

      ! Calculate the Hartree energy and potential of the solute only
      phi_solute => rho_tot_zero
      CALL pw_zero(phi_solute)
      NULLIFY (rho_tot_zero)
      CALL pw_poisson_solve(poisson_env=poisson_env, &
                            density=rho_solute, &
                            ehartree=energy%hartree, &
                            vhartree=phi_solute)

      ! Calculate the polarisation potential
      ! phi_pol = phi_tot - phi_solute
      CALL pw_axpy(phi_solute, phi_pol, alpha=-1.0_dp)

      ! Calculate the polarisation charge
      ! rho_pol = rho_tot - rho_solute
      CALL pw_axpy(rho_solute, rho_tot, alpha=-1.0_dp)
      polarisation_charge = accurate_sum(rho_tot%cr3d)*dvol
      CALL mp_sum(polarisation_charge, para_env%group)

      filename = "POLARISATION_POTENTIAL"
      cube_path = TRIM(print_path)//"%"//TRIM(filename)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, TRIM(cube_path)), &
                cp_p_file)) THEN
         append_cube = section_get_lval(input, TRIM(cube_path)//"%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) my_pos_cube = "APPEND"
         cube_unit = cp_print_key_unit_nr(logger, input, TRIM(cube_path), &
                                          extension=".cube", middle_name=TRIM(filename), &
                                          file_position=my_pos_cube, log_filename=.FALSE.)
         IF (output_unit > 0) THEN
            INQUIRE (UNIT=cube_unit, NAME=filename)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
               "The SCCS polarisation potential is written in cube file format to the file:", TRIM(filename)
         END IF
         CALL cp_pw_to_cube(phi_pol, cube_unit, TRIM(filename), particles=particles, &
                            stride=section_get_ivals(input, TRIM(cube_path)//"%STRIDE"))
         CALL cp_print_key_finished_output(cube_unit, logger, input, TRIM(cube_path))
      END IF

      ! Calculate SCCS polarisation energy
      energy%sccs_pol = 0.5_dp*pw_integral_ab(rho_solute, phi_pol)

      ! Calculate the Makov-Payne energy correction for charged systems with PBC
      ! Madelung energy of a simple cubic lattice of point charges immersed in neutralising jellium
      energy%sccs_mpc = -0.5_dp*2.8373_dp*tot_rho_solute**2/MINVAL(abc(1:3))

      ! Calculate additional solvation terms
      energy%sccs_cav = sccs_control%gamma_solvent*cavity_surface
      energy%sccs_dis = sccs_control%beta_solvent*cavity_volume
      energy%sccs_rep = sccs_control%alpha_solvent*cavity_surface

      IF (should_output .AND. (output_unit > 0)) THEN
         WRITE (UNIT=output_unit, FMT="(T3,A)") &
            "SCCS|"
         WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.12)") &
            "SCCS| Polarisation charge", polarisation_charge
         WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.12)") &
            "SCCS| Hartree energy of the solute only [Hartree]", energy%hartree, &
            "SCCS| Hartree energy of solute and solvent [Hartree]", energy%sccs_hartree
         WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.12,/,T3,A,T61,F20.3)") &
            "SCCS| Polarisation energy    [Hartree]", energy%sccs_pol, &
            "SCCS|                        [kcal/mol]", &
            cp_unit_from_cp2k(energy%sccs_pol, "kcalmol"), &
            "SCCS| Makov-Payne correction [Hartree]", energy%sccs_mpc, &
            "SCCS|                        [kcal/mol]", &
            cp_unit_from_cp2k(energy%sccs_mpc, "kcalmol"), &
            "SCCS| Cavitation energy      [Hartree]", energy%sccs_cav, &
            "SCCS|                        [kcal/mol]", &
            cp_unit_from_cp2k(energy%sccs_cav, "kcalmol"), &
            "SCCS| Dispersion free energy [Hartree]", energy%sccs_dis, &
            "SCCS|                        [kcal/mol]", &
            cp_unit_from_cp2k(energy%sccs_dis, "kcalmol"), &
            "SCCS| Repulsion free energy  [Hartree]", energy%sccs_rep, &
            "SCCS|                        [kcal/mol]", &
            cp_unit_from_cp2k(energy%sccs_rep, "kcalmol")
      END IF

      ! Calculate SCCS contribution to the Kohn-Sham potential
      f = -0.25_dp*dvol/twopi
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(dphi2,i,j,k) &
!$OMP                 SHARED(calculate_stress_tensor,f,deps_elec,dphi_tot) &
!$OMP                 SHARED(dvol,lb,ub,v_sccs)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               dphi2 = dphi_tot(1)%pw%cr3d(i, j, k)*dphi_tot(1)%pw%cr3d(i, j, k)+ &
                       dphi_tot(2)%pw%cr3d(i, j, k)*dphi_tot(2)%pw%cr3d(i, j, k)+ &
                       dphi_tot(3)%pw%cr3d(i, j, k)*dphi_tot(3)%pw%cr3d(i, j, k)
               v_sccs%cr3d(i, j, k) = v_sccs%cr3d(i, j, k)+f*deps_elec%cr3d(i, j, k)*dphi2
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      ! Release work storage
      NULLIFY (phi_solute)
      NULLIFY (rho_tot)
      NULLIFY (deps_elec)
      NULLIFY (rho_solute)
      DO i = 1, SIZE(work_r3d)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, work_r3d(i)%pw)
      END DO
      DO i = 1, 3
         CALL pw_pool_give_back_pw(auxbas_pw_pool, dln_eps_elec(i)%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, dphi_tot(i)%pw)
      END DO

      ! Release the SCCS printout environment
      CALL cp_print_key_finished_output(output_unit, logger, input, TRIM(print_path), &
                                        ignore_should_output=should_output)

      CALL timestop(handle)

   END SUBROUTINE sccs

! **************************************************************************************************
!> \brief      Calculate the smoothed dielectric function of Andreussi et al.
!> \param rho_elec ...
!> \param eps_elec ...
!> \param deps_elec ...
!> \param epsilon_solvent ...
!> \param rho_max ...
!> \param rho_min ...
!> \par History:
!>      - Creation (16.10.2013,MK)
!>      - Finite difference of isosurfaces implemented (21.12.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.1
! **************************************************************************************************
   SUBROUTINE andreussi(rho_elec, eps_elec, deps_elec, epsilon_solvent, rho_max, &
                        rho_min)

      TYPE(pw_type), POINTER                             :: rho_elec, eps_elec, deps_elec
      REAL(KIND=dp), INTENT(IN)                          :: epsilon_solvent, rho_max, rho_min

      CHARACTER(LEN=*), PARAMETER :: routineN = 'andreussi', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: tol = 1.0E-12_dp

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: diff, dq, dt, f, ln_rho_max, ln_rho_min, &
                                                            q, rho, t, x, y

      CALL timeset(routineN, handle)

      f = LOG(epsilon_solvent)/twopi
      diff = rho_max-rho_min
      IF (diff > tol) THEN
         ln_rho_max = LOG(rho_max)
         ln_rho_min = LOG(rho_min)
         q = twopi/(ln_rho_max-ln_rho_min)
         dq = -f*q
      END IF

      lb(1:3) = rho_elec%pw_grid%bounds_local(1, 1:3)
      ub(1:3) = rho_elec%pw_grid%bounds_local(2, 1:3)

      ! Calculate the dielectric function and its derivative
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(dt,i,j,k,rho,t,x,y) &
!$OMP                 SHARED(deps_elec,diff,dq,eps_elec,epsilon_solvent,f,lb,ub) &
!$OMP                 SHARED(ln_rho_max,rho_elec,q,rho_max,rho_min)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               rho = rho_elec%cr3d(i, j, k)
               IF (rho < rho_min) THEN
                  eps_elec%cr3d(i, j, k) = epsilon_solvent
                  deps_elec%cr3d(i, j, k) = 0.0_dp
               ELSE IF (rho <= rho_max) THEN
                  IF (diff > tol) THEN
                     x = LOG(rho)
                     y = q*(ln_rho_max-x)
                     t = f*(y-SIN(y))
                     eps_elec%cr3d(i, j, k) = EXP(t)
                     dt = dq*(1.0_dp-COS(y))
                     deps_elec%cr3d(i, j, k) = eps_elec%cr3d(i, j, k)*dt/rho
                  ELSE
                     eps_elec%cr3d(i, j, k) = 1.0_dp
                     deps_elec%cr3d(i, j, k) = 0.0_dp
                  END IF
               ELSE
                  eps_elec%cr3d(i, j, k) = 1.0_dp
                  deps_elec%cr3d(i, j, k) = 0.0_dp
               END IF
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE andreussi

! **************************************************************************************************
!> \brief      Calculate the smoothed dielectric function of Fattebert and Gygi
!> \param rho_elec ...
!> \param eps_elec ...
!> \param deps_elec ...
!> \param epsilon_solvent ...
!> \param beta ...
!> \param rho_zero ...
!> \par History:
!>      - Creation (15.10.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE fattebert_gygi(rho_elec, eps_elec, deps_elec, epsilon_solvent, beta, &
                             rho_zero)

      TYPE(pw_type), POINTER                             :: rho_elec, eps_elec, deps_elec
      REAL(KIND=dp), INTENT(IN)                          :: epsilon_solvent, beta, rho_zero

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fattebert_gygi', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: tol = 1.0E-12_dp

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: df, f, p, q, rho, s, t, twobeta

      CALL timeset(routineN, handle)

      df = (1.0_dp-epsilon_solvent)/rho_zero
      f = 0.5_dp*(epsilon_solvent-1.0_dp)
      q = 1.0_dp/rho_zero
      twobeta = 2.0_dp*beta

      lb(1:3) = rho_elec%pw_grid%bounds_local(1, 1:3)
      ub(1:3) = rho_elec%pw_grid%bounds_local(2, 1:3)

      ! Calculate the smoothed dielectric function and its derivative
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(i,j,k,p,rho,s,t) &
!$OMP                 SHARED(df,deps_elec,eps_elec,epsilon_solvent,f,lb,ub) &
!$OMP                 SHARED(q,rho_elec,twobeta)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               rho = rho_elec%cr3d(i, j, k)
               IF (rho < tol) THEN
                  eps_elec%cr3d(i, j, k) = epsilon_solvent
                  deps_elec%cr3d(i, j, k) = 0.0_dp
               ELSE
                  s = rho*q
                  p = s**twobeta
                  t = 1.0_dp/(1.0_dp+p)
                  eps_elec%cr3d(i, j, k) = 1.0_dp+f*(1.0_dp+(1.0_dp-p)*t)
                  deps_elec%cr3d(i, j, k) = df*twobeta*t*t*p/s
               END IF
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE fattebert_gygi

! **************************************************************************************************
!> \brief      Build the numerical derivative of a function on realspace grid
!> \param f ...
!> \param df ...
!> \param method ...
!> \param pw_env ...
!> \param input ...
!> \param para_env ...
!> \par History:
!>      - Creation (15.11.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE derive(f, df, method, pw_env, input, para_env)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(INOUT)       :: df
      INTEGER, INTENT(IN)                                :: method
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive', routineP = moduleN//':'//routineN

      INTEGER                                            :: border_points, handle, i
      INTEGER, DIMENSION(3)                              :: lb, n, ub
      TYPE(pw_p_type), DIMENSION(2)                      :: work_g1d
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_input_type)                    :: input_settings
      TYPE(realspace_grid_type), POINTER                 :: rs_grid
      TYPE(section_vals_type), POINTER                   :: rs_grid_section

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(f)))CALL cp__a("qs_sccs.F",972)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_sccs.F",973)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("qs_sccs.F",974)

      ! Perform method specific setup
      SELECT CASE (method)
      CASE (sccs_derivative_cd3, sccs_derivative_cd5, sccs_derivative_cd7)
         NULLIFY (rs_desc)
         NULLIFY (rs_grid)
         rs_grid_section => section_vals_get_subs_vals(input, "DFT%MGRID%RS_GRID")
         SELECT CASE (method)
         CASE (sccs_derivative_cd3)
            border_points = 1
         CASE (sccs_derivative_cd5)
            border_points = 2
         CASE (sccs_derivative_cd7)
            border_points = 3
         END SELECT
         CALL init_input_type(input_settings, 2*border_points+1, rs_grid_section, &
                              1, (/-1, -1, -1/))
         CALL rs_grid_create_descriptor(rs_desc, f%pw_grid, input_settings, &
                                        border_points=border_points)
         CALL rs_grid_create(rs_grid, rs_desc)
!MK    CALL rs_grid_print(rs_grid,6)
      CASE (sccs_derivative_fft)
         lb(1:3) = f%pw_grid%bounds_local(1, 1:3)
         ub(1:3) = f%pw_grid%bounds_local(2, 1:3)
         NULLIFY (auxbas_pw_pool)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         ! Get work storage for the 1d grids in g-space (derivative calculation)
         DO i = 1, SIZE(work_g1d)
            NULLIFY (work_g1d(i)%pw)
            CALL pw_pool_create_pw(auxbas_pw_pool, &
                                   work_g1d(i)%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
         END DO
      END SELECT

      ! Calculate the derivatives
      SELECT CASE (method)
      CASE (sccs_derivative_cd3)
         CALL derive_fdm_cd3(f, df, rs_grid)
      CASE (sccs_derivative_cd5)
         CALL derive_fdm_cd5(f, df, rs_grid)
      CASE (sccs_derivative_cd7)
         CALL derive_fdm_cd7(f, df, rs_grid)
      CASE (sccs_derivative_fft)
         ! FFT
         CALL pw_transfer(f, work_g1d(1)%pw)
         DO i = 1, 3
            n(:) = 0
            n(i) = 1
            CALL pw_copy(work_g1d(1)%pw, work_g1d(2)%pw)
            CALL pw_derive(work_g1d(2)%pw, n(:))
            CALL pw_transfer(work_g1d(2)%pw, df(i)%pw)
         END DO
      CASE DEFAULT
         CALL cp__b("qs_sccs.F",1030,"Invalid derivative method for SCCS specified")
      END SELECT

      ! Perform method specific cleanup
      SELECT CASE (method)
      CASE (sccs_derivative_cd3, sccs_derivative_cd5, sccs_derivative_cd7)
         CALL rs_grid_release(rs_grid)
         CALL rs_grid_release_descriptor(rs_desc)
      CASE (sccs_derivative_fft)
         DO i = 1, SIZE(work_g1d)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, work_g1d(i)%pw)
         END DO
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE derive

! **************************************************************************************************
!> \brief      Calculate the finite difference between two isosurfaces of the
!>             electronic density. The smoothed dielectric function of
!>             Andreussi et al. is used as switching function eventually
!>             defining the quantum volume and surface of the cavity.
!> \param rho_elec ...
!> \param norm_drho_elec ...
!> \param dtheta ...
!> \param epsilon_solvent ...
!> \param rho_max ...
!> \param rho_min ...
!> \param delta_rho ...
!> \par History:
!>      - Creation (21.12.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE surface_andreussi(rho_elec, norm_drho_elec, dtheta, &
                                epsilon_solvent, rho_max, rho_min, delta_rho)

      TYPE(pw_type), POINTER                             :: rho_elec, norm_drho_elec, dtheta
      REAL(KIND=dp), INTENT(IN)                          :: epsilon_solvent, rho_max, rho_min, &
                                                            delta_rho

      CHARACTER(LEN=*), PARAMETER :: routineN = 'surface_andreussi', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: tol = 1.0E-12_dp

      INTEGER                                            :: handle, i, j, k, l
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: diff, e, eps_elec, f, ln_rho_max, &
                                                            ln_rho_min, q, rho, t, x, y
      REAL(KIND=dp), DIMENSION(2)                        :: theta

      CALL timeset(routineN, handle)

      e = epsilon_solvent-1.0_dp
      f = LOG(epsilon_solvent)/twopi
      diff = rho_max-rho_min
      IF (diff > tol) THEN
         ln_rho_max = LOG(rho_max)
         ln_rho_min = LOG(rho_min)
         q = twopi/(ln_rho_max-ln_rho_min)
      END IF

      lb(1:3) = rho_elec%pw_grid%bounds_local(1, 1:3)
      ub(1:3) = rho_elec%pw_grid%bounds_local(2, 1:3)

      ! Calculate finite difference between two isosurfaces
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(eps_elec,i,j,k,l,rho,t,theta,x,y) &
!$OMP                 SHARED(delta_rho,diff,dtheta,e,epsilon_solvent,f,lb) &
!$OMP                 SHARED(ln_rho_max,norm_drho_elec,rho_elec,q,rho_max,rho_min,ub)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               DO l = 1, 2
                  rho = rho_elec%cr3d(i, j, k)+(REAL(l, KIND=dp)-1.5_dp)*delta_rho
                  IF (rho < rho_min) THEN
                     eps_elec = epsilon_solvent
                  ELSE IF (rho <= rho_max) THEN
                     IF (diff > tol) THEN
                        x = LOG(rho)
                        y = q*(ln_rho_max-x)
                        t = f*(y-SIN(y))
                        eps_elec = EXP(t)
                     ELSE
                        eps_elec = 1.0_dp
                     END IF
                  ELSE
                     eps_elec = 1.0_dp
                  END IF
                  theta(l) = (epsilon_solvent-eps_elec)/e
               END DO
               dtheta%cr3d(i, j, k) = (theta(2)-theta(1))*norm_drho_elec%cr3d(i, j, k)/delta_rho
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE surface_andreussi

! **************************************************************************************************
!> \brief      Calculate the finite difference between two isosurfaces of the
!>             the electronic density. The smoothed dielectric function of
!>             Fattebert and Gygi is used as switching function eventually
!>             defining the quantum volume and surface of the cavity.
!> \param rho_elec ...
!> \param norm_drho_elec ...
!> \param dtheta ...
!> \param epsilon_solvent ...
!> \param beta ...
!> \param rho_zero ...
!> \param delta_rho ...
!> \par History:
!>      - Creation (21.12.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE surface_fattebert_gygi(rho_elec, norm_drho_elec, dtheta, &
                                     epsilon_solvent, beta, rho_zero, delta_rho)

      TYPE(pw_type), POINTER                             :: rho_elec, norm_drho_elec, dtheta
      REAL(KIND=dp), INTENT(IN)                          :: epsilon_solvent, beta, rho_zero, &
                                                            delta_rho

      CHARACTER(LEN=*), PARAMETER :: routineN = 'surface_fattebert_gygi', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: tol = 1.0E-12_dp

      INTEGER                                            :: handle, i, j, k, l
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: e, eps_elec, f, p, q, rho, s, t, twobeta
      REAL(KIND=dp), DIMENSION(2)                        :: theta

      CALL timeset(routineN, handle)

      e = epsilon_solvent-1.0_dp
      f = 0.5_dp*e
      q = 1.0_dp/rho_zero
      twobeta = 2.0_dp*beta

      lb(1:3) = rho_elec%pw_grid%bounds_local(1, 1:3)
      ub(1:3) = rho_elec%pw_grid%bounds_local(2, 1:3)

      ! Calculate finite difference between two isosurfaces
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(eps_elec,i,j,k,l,p,rho,s,t,theta) &
!$OMP                 SHARED(delta_rho,dtheta,e,epsilon_solvent,f,lb) &
!$OMP                 SHARED(norm_drho_elec,q,rho_elec,twobeta,ub)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               DO l = 1, 2
                  rho = rho_elec%cr3d(i, j, k)+(REAL(l, KIND=dp)-1.5_dp)*delta_rho
                  IF (rho < tol) THEN
                     eps_elec = epsilon_solvent
                  ELSE
                     s = rho*q
                     p = s**twobeta
                     t = 1.0_dp/(1.0_dp+p)
                     eps_elec = 1.0_dp+f*(1.0_dp+(1.0_dp-p)*t)
                  END IF
                  theta(l) = (epsilon_solvent-eps_elec)/e
               END DO
               dtheta%cr3d(i, j, k) = (theta(2)-theta(1))*norm_drho_elec%cr3d(i, j, k)/delta_rho
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE surface_fattebert_gygi

END MODULE qs_sccs
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the energy contribution and the mo_derivative of
!>        a static periodic electric field
!> \par History
!>      none
!> \author fschiff (06.2010)
! **************************************************************************************************
MODULE qs_efield_berry
   USE ai_moments,                      ONLY: cossin
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_add_fm,&
                                              cp_cfm_solve
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_p_type,&
                                              cp_cfm_release,&
                                              cp_cfm_set_all
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_moments,                      ONLY: build_berry_moment_matrix
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_period_efield_types,          ONLY: efield_berry_type,&
                                              init_efield_matrices,&
                                              set_efield_matrices
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_efield_berry'

   ! *** Public subroutines ***

   PUBLIC :: qs_efield_berry_phase

! **************************************************************************************************

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param just_energy ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE qs_efield_berry_phase(qs_env, just_energy, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_efield_berry_phase', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: s_mstruct_changed
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      NULLIFY (dft_control)
      CALL get_qs_env(qs_env, s_mstruct_changed=s_mstruct_changed, &
                      dft_control=dft_control)

      IF (dft_control%apply_period_efield) THEN
         IF (s_mstruct_changed) CALL qs_efield_integrals(qs_env)
         IF (dft_control%period_efield%displacement_field) THEN
            CALL qs_dispfield_derivatives(qs_env, just_energy, calculate_forces)
         ELSE
            CALL qs_efield_derivatives(qs_env, just_energy, calculate_forces)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_efield_berry_phase

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE qs_efield_integrals(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_efield_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(dp), DIMENSION(3)                             :: kvec
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: cosmat, matrix_s, sinmat
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(efield_berry_type), POINTER                   :: efield

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_efield_berry.F",154)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control)
      NULLIFY (matrix_s)
      CALL get_qs_env(qs_env=qs_env, efield=efield, cell=cell, matrix_s=matrix_s)
      CALL init_efield_matrices(efield)
      ALLOCATE (cosmat(3), sinmat(3))
      DO i = 1, 3
         ALLOCATE (cosmat(i)%matrix, sinmat(i)%matrix)
         CALL cp_dbcsr_init(cosmat(i)%matrix)
         CALL cp_dbcsr_init(sinmat(i)%matrix)

         CALL cp_dbcsr_copy(cosmat(i)%matrix, matrix_s(1)%matrix, 'COS MAT')
         CALL cp_dbcsr_copy(sinmat(i)%matrix, matrix_s(1)%matrix, 'SIN MAT')
         CALL cp_dbcsr_set(cosmat(i)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(sinmat(i)%matrix, 0.0_dp)

         kvec(:) = twopi*cell%h_inv(i, :)
         CALL build_berry_moment_matrix(qs_env, cosmat(i)%matrix, sinmat(i)%matrix, kvec)
      END DO
      CALL set_efield_matrices(efield=efield, cosmat=cosmat, sinmat=sinmat)
      CALL set_qs_env(qs_env=qs_env, efield=efield)
      CALL timestop(handle)

   END SUBROUTINE qs_efield_integrals

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param just_energy ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE qs_efield_derivatives(qs_env, just_energy, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_efield_derivatives', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp), &
                                                            zone = (0.0_dp, 1.0_dp)

      COMPLEX(dp)                                        :: zdet, zdeta, zi(3)
      INTEGER :: atom_a, atom_b, handle, i, ia, iatom, icol, idir, ikind, inode, irow, iset, &
         ispin, j, jatom, jkind, jset, ldab, ldsa, ldsb, lsab, n1, n2, nao, natom, ncoa, ncob, &
         nkind, nmo, nseta, nsetb, sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, uniform, use_virial
      REAL(dp)                                           :: charge, ci(3), cqi(3), dab, dd, &
                                                            ener_field, f0, fab, fieldpol(3), &
                                                            focc, fpolvec(3), hmat(3, 3), occ, &
                                                            qi(3), ti(3)
      REAL(dp), DIMENSION(3)                             :: forcea, forceb, kvec, ra, rab, rb, ria
      REAL(dp), DIMENSION(:, :), POINTER                 :: cosab, iblock, rblock, sinab, work
      REAL(dp), DIMENSION(:, :, :), POINTER              :: dcosab, dsinab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(block_p_type), DIMENSION(3, 2)                :: dcost, dsint
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: eigrmat, inv_mat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: tempmat
      TYPE(cp_dbcsr_type), POINTER                       :: cosmat, mo_coeff_b, sinmat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_coeff_tmp, mo_derivs_tmp
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: inv_work, op_fm_set, opvec
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(efield_berry_type), POINTER                   :: efield
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, cell, particle_set)
      CALL get_qs_env(qs_env, dft_control=dft_control, cell=cell, &
                      particle_set=particle_set, virial=virial)
      NULLIFY (qs_kind_set, efield, para_env, sab_orb)
      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      efield=efield, energy=energy, para_env=para_env, sab_orb=sab_orb)

      ! calculate stress only if forces requested also
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      use_virial = use_virial .AND. calculate_forces
      ! disable stress calculation
      IF (use_virial) THEN
         CALL cp__b("qs_efield_berry.F",256,"Stress tensor for periodic E-field not implemented")
      END IF

      fieldpol = dft_control%period_efield%polarisation
      fieldpol = fieldpol/SQRT(DOT_PRODUCT(fieldpol, fieldpol))
      fieldpol = -fieldpol*dft_control%period_efield%strength
      hmat = cell%hmat(:, :)/twopi
      DO idir = 1, 3
         fpolvec(idir) = fieldpol(1)*hmat(1, idir)+fieldpol(2)*hmat(2, idir)+fieldpol(3)*hmat(3, idir)
      END DO

      ! nuclear contribution
      natom = SIZE(particle_set)
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, force=force)
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF
      zi(:) = CMPLX(1._dp, 0._dp, dp)
      DO ia = 1, natom
         CALL get_atomic_kind(particle_set(ia)%atomic_kind, kind_number=ikind)
         CALL get_qs_kind(qs_kind_set(ikind), core_charge=charge)
         ria = particle_set(ia)%r
         ria = pbc(ria, cell)
         DO idir = 1, 3
            kvec(:) = twopi*cell%h_inv(idir, :)
            dd = SUM(kvec(:)*ria(:))
            zdeta = CMPLX(COS(dd), SIN(dd), KIND=dp)**charge
            zi(idir) = zi(idir)*zdeta
         END DO
         IF (calculate_forces) THEN
            IF (para_env%mepos == 0) THEN
               iatom = atom_of_kind(ia)
               forcea(:) = fieldpol(:)*charge
               force(ikind)%efield(:, iatom) = force(ikind)%efield(:, iatom)+forcea(:)
            END IF
         END IF
         IF (use_virial) THEN
            IF (para_env%mepos == 0) &
               CALL virial_pair_force(virial%pv_virial, 1.0_dp, forcea, ria)
         END IF
      END DO
      qi = AIMAG(LOG(zi))

      ! check uniform occupation
      NULLIFY (mos)
      CALL get_qs_env(qs_env=qs_env, mos=mos)
      DO ispin = 1, dft_control%nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=occ, uniform_occupation=uniform)
         IF (.NOT. uniform) THEN
            CALL cp__b("qs_efield_berry.F",306,"Berry phase moments for non uniform MOs' occupation numbers not implemented")
         END IF
      END DO

      NULLIFY (mo_derivs)
      CALL get_qs_env(qs_env=qs_env, mo_derivs=mo_derivs)
      ! initialize all work matrices needed
      ALLOCATE (op_fm_set(2, dft_control%nspins))
      ALLOCATE (opvec(2, dft_control%nspins))
      ALLOCATE (eigrmat(dft_control%nspins))
      ALLOCATE (inv_mat(dft_control%nspins))
      ALLOCATE (inv_work(2, dft_control%nspins))
      ALLOCATE (mo_derivs_tmp(SIZE(mo_derivs)))
      ALLOCATE (mo_coeff_tmp(SIZE(mo_derivs)))

      ! Allocate temp matrices for the wavefunction derivatives
      DO ispin = 1, dft_control%nspins
         NULLIFY (tmp_fm_struct, mo_coeff)
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmo, &
                                  ncol_global=nmo, para_env=para_env, context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(mo_derivs_tmp(ispin)%matrix, mo_coeff%matrix_struct)
         CALL cp_fm_create(mo_coeff_tmp(ispin)%matrix, mo_coeff%matrix_struct)
         CALL copy_dbcsr_to_fm(mo_derivs(ispin)%matrix, mo_derivs_tmp(ispin)%matrix)
         DO i = 1, SIZE(op_fm_set, 1)
            CALL cp_fm_create(opvec(i, ispin)%matrix, mo_coeff%matrix_struct)
            NULLIFY (op_fm_set(i, ispin)%matrix)
            CALL cp_fm_create(op_fm_set(i, ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_create(inv_work(i, ispin)%matrix, op_fm_set(i, ispin)%matrix%matrix_struct)
         END DO
         CALL cp_cfm_create(eigrmat(ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
         CALL cp_cfm_create(inv_mat(ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      END DO
      ! temp matrices for force calculation
      IF (calculate_forces) THEN
         NULLIFY (matrix_s)
         CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s)
         ALLOCATE (tempmat(2, dft_control%nspins))
         DO ispin = 1, dft_control%nspins
            ALLOCATE (tempmat(1, ispin)%matrix, tempmat(2, ispin)%matrix)
            CALL cp_dbcsr_init(tempmat(1, ispin)%matrix)
            CALL cp_dbcsr_init(tempmat(2, ispin)%matrix)
            CALL cp_dbcsr_copy(tempmat(1, ispin)%matrix, matrix_s(1)%matrix, 'TEMPMAT')
            CALL cp_dbcsr_copy(tempmat(2, ispin)%matrix, matrix_s(1)%matrix, 'TEMPMAT')
            CALL cp_dbcsr_set(tempmat(1, ispin)%matrix, 0.0_dp)
            CALL cp_dbcsr_set(tempmat(2, ispin)%matrix, 0.0_dp)
         END DO
         ! integration
         CALL get_qs_kind_set(qs_kind_set, maxco=ldab, maxsgf=lsab)
         ALLOCATE (cosab(ldab, ldab), sinab(ldab, ldab), work(ldab, ldab))
         ALLOCATE (dcosab(ldab, ldab, 3), dsinab(ldab, ldab, 3))
         lsab = MAX(ldab, lsab)
         DO i = 1, 3
            ALLOCATE (dcost(i, 1)%block(lsab, lsab), dsint(i, 1)%block(lsab, lsab))
            ALLOCATE (dcost(i, 2)%block(lsab, lsab), dsint(i, 2)%block(lsab, lsab))
         END DO
      END IF

      !Start the MO derivative calculation
      !loop over all cell vectors
      DO idir = 1, 3
         ci(idir) = 0.0_dp
         zi(idir) = zero
         IF (ABS(fpolvec(idir)) .GT. 1.0E-12_dp) THEN
            cosmat => efield%cosmat(idir)%matrix
            sinmat => efield%sinmat(idir)%matrix
            !evaluate the expression needed for the derivative (S_berry * C  and [C^T S_berry C]^-1)
            !first step S_berry * C  and C^T S_berry C
            DO ispin = 1, dft_control%nspins ! spin
               IF (mos(ispin)%mo_set%use_mo_coeff_b) THEN
                  CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, mo_coeff_b=mo_coeff_b, nmo=nmo)
                  CALL copy_dbcsr_to_fm(mo_coeff_b, mo_coeff_tmp(ispin)%matrix)
               ELSE
                  CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, mo_coeff=mo_coeff_tmp(ispin)%matrix, nmo=nmo)
               END IF
               CALL cp_dbcsr_sm_fm_multiply(cosmat, mo_coeff_tmp(ispin)%matrix, opvec(1, ispin)%matrix, ncol=nmo)
               CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff_tmp(ispin)%matrix, opvec(1, ispin)%matrix, 0.0_dp, &
                            op_fm_set(1, ispin)%matrix)
               CALL cp_dbcsr_sm_fm_multiply(sinmat, mo_coeff_tmp(ispin)%matrix, opvec(2, ispin)%matrix, ncol=nmo)
               CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff_tmp(ispin)%matrix, opvec(2, ispin)%matrix, 0.0_dp, &
                            op_fm_set(2, ispin)%matrix)
            ENDDO
            !second step invert C^T S_berry C
            zdet = one
            DO ispin = 1, dft_control%nspins
               CALL cp_cfm_add_fm(zero, eigrmat(ispin)%matrix, one, op_fm_set(1, ispin)%matrix)
               CALL cp_cfm_add_fm(one, eigrmat(ispin)%matrix, -zone, op_fm_set(2, ispin)%matrix)
               CALL cp_cfm_set_all(inv_mat(ispin)%matrix, zero, one)
               CALL cp_cfm_solve(eigrmat(ispin)%matrix, inv_mat(ispin)%matrix, zdeta)
               zdet = zdet*zdeta
            END DO
            zi(idir) = zdet**occ
            ci(idir) = AIMAG(LOG(zdet**occ))

            IF (.NOT. just_energy) THEN
               !compute the orbital derivative
               focc = fpolvec(idir)
               DO ispin = 1, dft_control%nspins
                  inv_work(1, ispin)%matrix%local_data(:, :) = REAL(inv_mat(ispin)%matrix%local_data(:, :), dp)
                  inv_work(2, ispin)%matrix%local_data(:, :) = AIMAG(inv_mat(ispin)%matrix%local_data(:, :))
                  CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, focc, opvec(1, ispin)%matrix, inv_work(2, ispin)%matrix, &
                               1.0_dp, mo_derivs_tmp(ispin)%matrix)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, -focc, opvec(2, ispin)%matrix, inv_work(1, ispin)%matrix, &
                               1.0_dp, mo_derivs_tmp(ispin)%matrix)
               END DO
            END IF

            !compute nuclear forces
            IF (calculate_forces) THEN
               nkind = SIZE(qs_kind_set)
               natom = SIZE(particle_set)
               kvec(:) = twopi*cell%h_inv(idir, :)

               ! calculate: C [C^T S_berry C]^(-1) C^T
               ! Store this matrix in DBCSR form (only S overlap blocks)
               DO ispin = 1, dft_control%nspins
                  CALL cp_dbcsr_set(tempmat(1, ispin)%matrix, 0.0_dp)
                  CALL cp_dbcsr_set(tempmat(2, ispin)%matrix, 0.0_dp)
                  CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mo_coeff_tmp(ispin)%matrix, inv_work(1, ispin)%matrix, 0.0_dp, &
                               opvec(1, ispin)%matrix)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mo_coeff_tmp(ispin)%matrix, inv_work(2, ispin)%matrix, 0.0_dp, &
                               opvec(2, ispin)%matrix)
                  CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=tempmat(1, ispin)%matrix, &
                                             matrix_v=opvec(1, ispin)%matrix, matrix_g=mo_coeff_tmp(ispin)%matrix, ncol=nmo)
                  CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=tempmat(2, ispin)%matrix, &
                                             matrix_v=opvec(2, ispin)%matrix, matrix_g=mo_coeff_tmp(ispin)%matrix, ncol=nmo)
               END DO

               ! Calculation of derivative integrals (da|eikr|b) and (a|eikr|db)
               ALLOCATE (basis_set_list(nkind))
               DO ikind = 1, nkind
                  qs_kind => qs_kind_set(ikind)
                  CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
                  IF (ASSOCIATED(basis_set_a)) THEN
                     basis_set_list(ikind)%gto_basis_set => basis_set_a
                  ELSE
                     NULLIFY (basis_set_list(ikind)%gto_basis_set)
                  END IF
               END DO
               !
               CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
               DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
                  CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                         iatom=iatom, jatom=jatom, r=rab)
                  basis_set_a => basis_set_list(ikind)%gto_basis_set
                  IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
                  basis_set_b => basis_set_list(jkind)%gto_basis_set
                  IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
                  ! basis ikind
                  first_sgfa => basis_set_a%first_sgf
                  la_max => basis_set_a%lmax
                  la_min => basis_set_a%lmin
                  npgfa => basis_set_a%npgf
                  nseta = basis_set_a%nset
                  nsgfa => basis_set_a%nsgf_set
                  rpgfa => basis_set_a%pgf_radius
                  set_radius_a => basis_set_a%set_radius
                  sphi_a => basis_set_a%sphi
                  zeta => basis_set_a%zet
                  ! basis jkind
                  first_sgfb => basis_set_b%first_sgf
                  lb_max => basis_set_b%lmax
                  lb_min => basis_set_b%lmin
                  npgfb => basis_set_b%npgf
                  nsetb = basis_set_b%nset
                  nsgfb => basis_set_b%nsgf_set
                  rpgfb => basis_set_b%pgf_radius
                  set_radius_b => basis_set_b%set_radius
                  sphi_b => basis_set_b%sphi
                  zetb => basis_set_b%zet

                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)

                  ldsa = SIZE(sphi_a, 1)
                  ldsb = SIZE(sphi_b, 1)
                  ra(:) = pbc(particle_set(iatom)%r(:), cell)
                  rb(:) = ra+rab
                  dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))

                  IF (iatom <= jatom) THEN
                     irow = iatom
                     icol = jatom
                  ELSE
                     irow = jatom
                     icol = iatom
                  END IF

                  IF (iatom == jatom .AND. dab < 1.e-10_dp) THEN
                     fab = 1.0_dp*occ
                  ELSE
                     fab = 2.0_dp*occ
                  END IF

                  DO i = 1, 3
                     dcost(i, 1)%block = 0.0_dp
                     dsint(i, 1)%block = 0.0_dp
                     dcost(i, 2)%block = 0.0_dp
                     dsint(i, 2)%block = 0.0_dp
                  END DO

                  DO iset = 1, nseta
                     ncoa = npgfa(iset)*ncoset(la_max(iset))
                     sgfa = first_sgfa(1, iset)
                     DO jset = 1, nsetb
                        IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE
                        ncob = npgfb(jset)*ncoset(lb_max(jset))
                        sgfb = first_sgfb(1, jset)
                        ! Calculate the primitive integrals (da|b)
                        CALL cossin(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                    lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                    ra, rb, kvec, cosab, sinab, dcosab, dsinab)
                        DO i = 1, 3
                           CALL contract_all(dcost(i, 1)%block, dsint(i, 1)%block, &
                                             ncoa, nsgfa(iset), sgfa, sphi_a, ldsa, &
                                             ncob, nsgfb(jset), sgfb, sphi_b, ldsb, &
                                             dcosab(:, :, i), dsinab(:, :, i), ldab, work, ldab)
                        END DO
                        ! Calculate the primitive integrals (a|db)
                        CALL cossin(lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                    la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                    rb, ra, kvec, cosab, sinab, dcosab, dsinab)
                        DO i = 1, 3
                           dcosab(1:ncoa, 1:ncob, i) = TRANSPOSE(dcosab(1:ncob, 1:ncoa, i))
                           dsinab(1:ncoa, 1:ncob, i) = TRANSPOSE(dsinab(1:ncob, 1:ncoa, i))
                           CALL contract_all(dcost(i, 2)%block, dsint(i, 2)%block, &
                                             ncoa, nsgfa(iset), sgfa, sphi_a, ldsa, &
                                             ncob, nsgfb(jset), sgfb, sphi_b, ldsb, &
                                             dcosab(:, :, i), dsinab(:, :, i), ldab, work, ldab)
                        END DO
                     END DO
                  END DO
                  forcea = 0.0_dp
                  forceb = 0.0_dp
                  DO ispin = 1, dft_control%nspins
                     NULLIFY (rblock, iblock)
                     CALL cp_dbcsr_get_block_p(matrix=tempmat(1, ispin)%matrix, &
                                               row=irow, col=icol, BLOCK=rblock, found=found)
                     IF(.NOT.(found))CALL cp__a("qs_efield_berry.F",547)
                     CALL cp_dbcsr_get_block_p(matrix=tempmat(2, ispin)%matrix, &
                                               row=irow, col=icol, BLOCK=iblock, found=found)
                     IF(.NOT.(found))CALL cp__a("qs_efield_berry.F",550)
                     n1 = SIZE(rblock, 1)
                     n2 = SIZE(rblock, 2)
                     IF(.NOT.(SIZE(iblock, 1) == n1))CALL cp__a("qs_efield_berry.F",553)
                     IF(.NOT.(SIZE(iblock, 2) == n2))CALL cp__a("qs_efield_berry.F",554)
                     IF(.NOT.(lsab >= n1))CALL cp__a("qs_efield_berry.F",555)
                     IF(.NOT.(lsab >= n2))CALL cp__a("qs_efield_berry.F",556)
                     IF (iatom <= jatom) THEN
                        DO i = 1, 3
                           forcea(i) = forcea(i)+SUM(rblock(1:n1, 1:n2)*dsint(i, 1)%block(1:n1, 1:n2)) &
                                       -SUM(iblock(1:n1, 1:n2)*dcost(i, 1)%block(1:n1, 1:n2))
                           forceb(i) = forceb(i)+SUM(rblock(1:n1, 1:n2)*dsint(i, 2)%block(1:n1, 1:n2)) &
                                       -SUM(iblock(1:n1, 1:n2)*dcost(i, 2)%block(1:n1, 1:n2))
                        END DO
                     ELSE
                        DO i = 1, 3
                           forcea(i) = forcea(i)+SUM(TRANSPOSE(rblock(1:n1, 1:n2))*dsint(i, 1)%block(1:n2, 1:n1)) &
                                       -SUM(TRANSPOSE(iblock(1:n1, 1:n2))*dcost(i, 1)%block(1:n2, 1:n1))
                           forceb(i) = forceb(i)+SUM(TRANSPOSE(rblock(1:n1, 1:n2))*dsint(i, 2)%block(1:n2, 1:n1)) &
                                       -SUM(TRANSPOSE(iblock(1:n1, 1:n2))*dcost(i, 2)%block(1:n2, 1:n1))
                        END DO
                     END IF
                  END DO
                  force(ikind)%efield(1:3, atom_a) = force(ikind)%efield(1:3, atom_a)-fab*fpolvec(idir)*forcea(1:3)
                  force(jkind)%efield(1:3, atom_b) = force(jkind)%efield(1:3, atom_b)-fab*fpolvec(idir)*forceb(1:3)
                  IF (use_virial) THEN
                     f0 = -fab*fpolvec(idir)
                     CALL virial_pair_force(virial%pv_virial, f0, forcea, ra)
                     CALL virial_pair_force(virial%pv_virial, f0, forceb, rb)
                  END IF

               END DO
               CALL neighbor_list_iterator_release(nl_iterator)
               DEALLOCATE (basis_set_list)

            END IF
         END IF
      END DO

      ! Energy
      ener_field = 0.0_dp
      ti = 0.0_dp
      DO idir = 1, 3
         ! make sure the total normalized polarization is within [-1:1]
         cqi(idir) = qi(idir)+ci(idir)
         IF (cqi(idir) > pi) cqi(idir) = cqi(idir)-twopi
         IF (cqi(idir) < -pi) cqi(idir) = cqi(idir)+twopi
         ! now check for log branch
         IF (ABS(efield%polarisation(idir)-cqi(idir)) > pi) THEN
            ti(idir) = (efield%polarisation(idir)-cqi(idir))/pi
            DO i = 1, 10
               cqi(idir) = cqi(idir)+SIGN(1.0_dp, ti(idir))*twopi
               IF (ABS(efield%polarisation(idir)-cqi(idir)) < pi) EXIT
            END DO
         END IF
         ener_field = ener_field+fpolvec(idir)*cqi(idir)
      END DO

      ! update the references
      IF (calculate_forces) THEN
         ! check for smoothness of energy surface
         IF (ABS(efield%field_energy-ener_field) > pi*ABS(SUM(fpolvec))) THEN
            CALL cp__w("qs_efield_berry.F",612,"Large change of e-field energy detected. Correct for non-smooth energy surface")
         END IF
         efield%field_energy = ener_field
         efield%polarisation(:) = cqi(:)
      END IF
      energy%efield = ener_field

      IF (.NOT. just_energy) THEN
         ! Add the result to mo_derivativs
         DO ispin = 1, dft_control%nspins
            CALL copy_fm_to_dbcsr(mo_derivs_tmp(ispin)%matrix, mo_derivs(ispin)%matrix)
         END DO
         IF (use_virial) THEN
            ti = 0.0_dp
            DO i = 1, 3
               DO j = 1, 3
                  ti(j) = ti(j)+hmat(j, i)*cqi(i)
               END DO
            END DO
            DO i = 1, 3
               DO j = 1, 3
                  virial%pv_virial(i, j) = virial%pv_virial(i, j)-fieldpol(i)*ti(j)
               END DO
            END DO
         END IF
      END IF

      DO ispin = 1, dft_control%nspins
         CALL cp_cfm_release(eigrmat(ispin)%matrix)
         CALL cp_cfm_release(inv_mat(ispin)%matrix)
         CALL cp_fm_release(mo_derivs_tmp(ispin)%matrix)
         CALL cp_fm_release(mo_coeff_tmp(ispin)%matrix)
         DO i = 1, SIZE(op_fm_set, 1)
            CALL cp_fm_release(opvec(i, ispin)%matrix)
            CALL cp_fm_release(op_fm_set(i, ispin)%matrix)
            CALL cp_fm_release(inv_work(i, ispin)%matrix)
         END DO
      END DO
      DEALLOCATE (inv_mat, inv_work, op_fm_set, opvec, eigrmat)
      DEALLOCATE (mo_coeff_tmp, mo_derivs_tmp)

      IF (calculate_forces) THEN
         DO ikind = 1, SIZE(atomic_kind_set)
            CALL mp_sum(force(ikind)%efield, para_env%group)
         END DO
         DEALLOCATE (atom_of_kind)
         DEALLOCATE (cosab, sinab, work, dcosab, dsinab)
         DO i = 1, 3
            DEALLOCATE (dcost(i, 1)%block, dsint(i, 1)%block)
            DEALLOCATE (dcost(i, 2)%block, dsint(i, 2)%block)
         END DO
         CALL cp_dbcsr_deallocate_matrix_set(tempmat)
      END IF
      CALL timestop(handle)

   END SUBROUTINE qs_efield_derivatives

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param just_energy ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE qs_dispfield_derivatives(qs_env, just_energy, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_dispfield_derivatives', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp), &
                                                            zone = (0.0_dp, 1.0_dp)

      COMPLEX(dp)                                        :: zdet, zdeta, zi(3)
      INTEGER :: handle, i, ia, iatom, icol, idir, ikind, inode, irow, iset, ispin, jatom, jkind, &
         jset, ldab, ldsa, ldsb, lsab, n1, n2, nao, natom, ncoa, ncob, nkind, nmo, nseta, nsetb, &
         sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, uniform, use_virial
      REAL(dp)                                           :: charge, ci(3), cqi(3), dab, dd, di(3), &
                                                            ener_field, fab, fieldpol(3), focc, &
                                                            hmat(3, 3), occ, omega, qi(3), &
                                                            rlog(3), zlog(3)
      REAL(dp), DIMENSION(3)                             :: forcea, forceb, kvec, ra, rab, rb, ria
      REAL(dp), DIMENSION(:, :), POINTER                 :: cosab, iblock, rblock, sinab, work
      REAL(dp), DIMENSION(:, :, :), POINTER              :: dcosab, dsinab, force_tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(block_p_type), DIMENSION(3, 2)                :: dcost, dsint
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: eigrmat, inv_mat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: tempmat
      TYPE(cp_dbcsr_type), POINTER                       :: cosmat, mo_coeff_b, sinmat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_coeff_tmp
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: inv_work, mo_derivs_tmp, op_fm_set, opvec
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(efield_berry_type), POINTER                   :: efield
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, cell, particle_set)
      CALL get_qs_env(qs_env, dft_control=dft_control, cell=cell, &
                      particle_set=particle_set, virial=virial)
      NULLIFY (qs_kind_set, efield, para_env, sab_orb)
      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      efield=efield, energy=energy, para_env=para_env, sab_orb=sab_orb)

      ! calculate stress only if forces requested also
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      use_virial = use_virial .AND. calculate_forces
      ! disable stress calculation
      IF (use_virial) THEN
         CALL cp__b("qs_efield_berry.F",745,"Stress tensor for periodic D-field not implemented")
      END IF

      fieldpol = dft_control%period_efield%polarisation
      fieldpol = fieldpol/SQRT(DOT_PRODUCT(fieldpol, fieldpol))
      fieldpol = fieldpol*dft_control%period_efield%strength

      omega = cell%deth
      hmat = cell%hmat(:, :)/(twopi*omega)

      ! nuclear contribution to polarization
      natom = SIZE(particle_set)
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, force=force)
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
         ALLOCATE (force_tmp(natom, 3, 3))
         force_tmp = 0.0_dp
      END IF
      zi(:) = CMPLX(1._dp, 0._dp, dp)
      DO ia = 1, natom
         CALL get_atomic_kind(particle_set(ia)%atomic_kind, kind_number=ikind)
         CALL get_qs_kind(qs_kind_set(ikind), core_charge=charge)
         ria = particle_set(ia)%r
         ria = pbc(ria, cell)
         DO idir = 1, 3
            kvec(:) = twopi*cell%h_inv(idir, :)
            dd = SUM(kvec(:)*ria(:))
            zdeta = CMPLX(COS(dd), SIN(dd), KIND=dp)**charge
            zi(idir) = zi(idir)*zdeta
         END DO
         IF (calculate_forces) THEN
            IF (para_env%mepos == 0) THEN
               DO i = 1, 3
                  force_tmp(ia, i, i) = force_tmp(ia, i, i)+charge/omega
               END DO
            END IF
         END IF
      END DO
      rlog = AIMAG(LOG(zi))

      ! check uniform occupation
      NULLIFY (mos)
      CALL get_qs_env(qs_env=qs_env, mos=mos)
      DO ispin = 1, dft_control%nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=occ, uniform_occupation=uniform)
         IF (.NOT. uniform) THEN
            CALL cp__b("qs_efield_berry.F",792,"Berry phase moments for non uniform MOs' occupation numbers not implemented")
         END IF
      END DO

      ! initialize all work matrices needed
      NULLIFY (mo_derivs)
      CALL get_qs_env(qs_env=qs_env, mo_derivs=mo_derivs)
      ALLOCATE (op_fm_set(2, dft_control%nspins))
      ALLOCATE (opvec(2, dft_control%nspins))
      ALLOCATE (eigrmat(dft_control%nspins))
      ALLOCATE (inv_mat(dft_control%nspins))
      ALLOCATE (inv_work(2, dft_control%nspins))
      ALLOCATE (mo_derivs_tmp(3, SIZE(mo_derivs)))
      ALLOCATE (mo_coeff_tmp(SIZE(mo_derivs)))

      ! Allocate temp matrices for the wavefunction derivatives
      DO ispin = 1, dft_control%nspins
         NULLIFY (tmp_fm_struct, mo_coeff)
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmo, &
                                  ncol_global=nmo, para_env=para_env, context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(mo_coeff_tmp(ispin)%matrix, mo_coeff%matrix_struct)
         DO i = 1, 3
            CALL cp_fm_create(mo_derivs_tmp(i, ispin)%matrix, mo_coeff%matrix_struct)
            CALL cp_fm_set_all(matrix=mo_derivs_tmp(i, ispin)%matrix, alpha=0.0_dp)
         END DO
         DO i = 1, SIZE(op_fm_set, 1)
            CALL cp_fm_create(opvec(i, ispin)%matrix, mo_coeff%matrix_struct)
            NULLIFY (op_fm_set(i, ispin)%matrix)
            CALL cp_fm_create(op_fm_set(i, ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_create(inv_work(i, ispin)%matrix, op_fm_set(i, ispin)%matrix%matrix_struct)
         END DO
         CALL cp_cfm_create(eigrmat(ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
         CALL cp_cfm_create(inv_mat(ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      END DO
      ! temp matrices for force calculation
      IF (calculate_forces) THEN
         NULLIFY (matrix_s)
         CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s)
         ALLOCATE (tempmat(2, dft_control%nspins))
         DO ispin = 1, dft_control%nspins
            ALLOCATE (tempmat(1, ispin)%matrix, tempmat(2, ispin)%matrix)
            CALL cp_dbcsr_init(tempmat(1, ispin)%matrix)
            CALL cp_dbcsr_init(tempmat(2, ispin)%matrix)
            CALL cp_dbcsr_copy(tempmat(1, ispin)%matrix, matrix_s(1)%matrix, 'TEMPMAT')
            CALL cp_dbcsr_copy(tempmat(2, ispin)%matrix, matrix_s(1)%matrix, 'TEMPMAT')
            CALL cp_dbcsr_set(tempmat(1, ispin)%matrix, 0.0_dp)
            CALL cp_dbcsr_set(tempmat(2, ispin)%matrix, 0.0_dp)
         END DO
         ! integration
         CALL get_qs_kind_set(qs_kind_set, maxco=ldab, maxsgf=lsab)
         ALLOCATE (cosab(ldab, ldab), sinab(ldab, ldab), work(ldab, ldab))
         ALLOCATE (dcosab(ldab, ldab, 3), dsinab(ldab, ldab, 3))
         lsab = MAX(lsab, ldab)
         DO i = 1, 3
            ALLOCATE (dcost(i, 1)%block(lsab, lsab), dsint(i, 1)%block(lsab, lsab))
            ALLOCATE (dcost(i, 2)%block(lsab, lsab), dsint(i, 2)%block(lsab, lsab))
         END DO
      END IF

      !Start the MO derivative calculation
      !loop over all cell vectors
      DO idir = 1, 3
         zi(idir) = zero
         cosmat => efield%cosmat(idir)%matrix
         sinmat => efield%sinmat(idir)%matrix
         !evaluate the expression needed for the derivative (S_berry * C  and [C^T S_berry C]^-1)
         !first step S_berry * C  and C^T S_berry C
         DO ispin = 1, dft_control%nspins ! spin
            IF (mos(ispin)%mo_set%use_mo_coeff_b) THEN
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, mo_coeff_b=mo_coeff_b, nmo=nmo)
               CALL copy_dbcsr_to_fm(mo_coeff_b, mo_coeff_tmp(ispin)%matrix)
            ELSE
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, mo_coeff=mo_coeff_tmp(ispin)%matrix, nmo=nmo)
            END IF
            CALL cp_dbcsr_sm_fm_multiply(cosmat, mo_coeff_tmp(ispin)%matrix, opvec(1, ispin)%matrix, ncol=nmo)
            CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff_tmp(ispin)%matrix, opvec(1, ispin)%matrix, 0.0_dp, &
                         op_fm_set(1, ispin)%matrix)
            CALL cp_dbcsr_sm_fm_multiply(sinmat, mo_coeff_tmp(ispin)%matrix, opvec(2, ispin)%matrix, ncol=nmo)
            CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff_tmp(ispin)%matrix, opvec(2, ispin)%matrix, 0.0_dp, &
                         op_fm_set(2, ispin)%matrix)
         ENDDO
         !second step invert C^T S_berry C
         zdet = one
         DO ispin = 1, dft_control%nspins
            CALL cp_cfm_add_fm(zero, eigrmat(ispin)%matrix, one, op_fm_set(1, ispin)%matrix)
            CALL cp_cfm_add_fm(one, eigrmat(ispin)%matrix, -zone, op_fm_set(2, ispin)%matrix)
            CALL cp_cfm_set_all(inv_mat(ispin)%matrix, zero, one)
            CALL cp_cfm_solve(eigrmat(ispin)%matrix, inv_mat(ispin)%matrix, zdeta)
            zdet = zdet*zdeta
         END DO
         zi(idir) = zdet**occ
         zlog(idir) = AIMAG(LOG(zi(idir)))

         IF (.NOT. just_energy) THEN
            !compute the orbital derivative
            DO ispin = 1, dft_control%nspins
               inv_work(1, ispin)%matrix%local_data(:, :) = REAL(inv_mat(ispin)%matrix%local_data(:, :), dp)
               inv_work(2, ispin)%matrix%local_data(:, :) = AIMAG(inv_mat(ispin)%matrix%local_data(:, :))
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
               DO i = 1, 3
                  focc = hmat(idir, i)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, focc, opvec(1, ispin)%matrix, inv_work(2, ispin)%matrix, &
                               1.0_dp, mo_derivs_tmp(idir, ispin)%matrix)
                  CALL cp_gemm("N", "N", nao, nmo, nmo, -focc, opvec(2, ispin)%matrix, inv_work(1, ispin)%matrix, &
                               1.0_dp, mo_derivs_tmp(idir, ispin)%matrix)
               END DO
            END DO
         END IF

         !compute nuclear forces
         IF (calculate_forces) THEN
            nkind = SIZE(qs_kind_set)
            natom = SIZE(particle_set)
            kvec(:) = twopi*cell%h_inv(idir, :)

            ! calculate: C [C^T S_berry C]^(-1) C^T
            ! Store this matrix in DBCSR form (only S overlap blocks)
            DO ispin = 1, dft_control%nspins
               CALL cp_dbcsr_set(tempmat(1, ispin)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(tempmat(2, ispin)%matrix, 0.0_dp)
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
               CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mo_coeff_tmp(ispin)%matrix, inv_work(1, ispin)%matrix, 0.0_dp, &
                            opvec(1, ispin)%matrix)
               CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mo_coeff_tmp(ispin)%matrix, inv_work(2, ispin)%matrix, 0.0_dp, &
                            opvec(2, ispin)%matrix)
               CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=tempmat(1, ispin)%matrix, &
                                          matrix_v=opvec(1, ispin)%matrix, matrix_g=mo_coeff_tmp(ispin)%matrix, ncol=nmo)
               CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=tempmat(2, ispin)%matrix, &
                                          matrix_v=opvec(2, ispin)%matrix, matrix_g=mo_coeff_tmp(ispin)%matrix, ncol=nmo)
            END DO

            ! Calculation of derivative integrals (da|eikr|b) and (a|eikr|db)
            ALLOCATE (basis_set_list(nkind))
            DO ikind = 1, nkind
               qs_kind => qs_kind_set(ikind)
               CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
               IF (ASSOCIATED(basis_set_a)) THEN
                  basis_set_list(ikind)%gto_basis_set => basis_set_a
               ELSE
                  NULLIFY (basis_set_list(ikind)%gto_basis_set)
               END IF
            END DO
            !
            CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                      iatom=iatom, jatom=jatom, r=rab)
               basis_set_a => basis_set_list(ikind)%gto_basis_set
               IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
               basis_set_b => basis_set_list(jkind)%gto_basis_set
               IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
               ! basis ikind
               first_sgfa => basis_set_a%first_sgf
               la_max => basis_set_a%lmax
               la_min => basis_set_a%lmin
               npgfa => basis_set_a%npgf
               nseta = basis_set_a%nset
               nsgfa => basis_set_a%nsgf_set
               rpgfa => basis_set_a%pgf_radius
               set_radius_a => basis_set_a%set_radius
               sphi_a => basis_set_a%sphi
               zeta => basis_set_a%zet
               ! basis jkind
               first_sgfb => basis_set_b%first_sgf
               lb_max => basis_set_b%lmax
               lb_min => basis_set_b%lmin
               npgfb => basis_set_b%npgf
               nsetb = basis_set_b%nset
               nsgfb => basis_set_b%nsgf_set
               rpgfb => basis_set_b%pgf_radius
               set_radius_b => basis_set_b%set_radius
               sphi_b => basis_set_b%sphi
               zetb => basis_set_b%zet

               ldsa = SIZE(sphi_a, 1)
               ldsb = SIZE(sphi_b, 1)
               ra(:) = pbc(particle_set(iatom)%r(:), cell)
               rb(:) = ra+rab
               dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))

               IF (iatom <= jatom) THEN
                  irow = iatom
                  icol = jatom
               ELSE
                  irow = jatom
                  icol = iatom
               END IF

               IF (iatom == jatom .AND. dab < 1.e-10_dp) THEN
                  fab = 1.0_dp*occ
               ELSE
                  fab = 2.0_dp*occ
               END IF

               DO i = 1, 3
                  dcost(i, 1)%block = 0.0_dp
                  dsint(i, 1)%block = 0.0_dp
                  dcost(i, 2)%block = 0.0_dp
                  dsint(i, 2)%block = 0.0_dp
               END DO

               DO iset = 1, nseta
                  ncoa = npgfa(iset)*ncoset(la_max(iset))
                  sgfa = first_sgfa(1, iset)
                  DO jset = 1, nsetb
                     IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE
                     ncob = npgfb(jset)*ncoset(lb_max(jset))
                     sgfb = first_sgfb(1, jset)
                     ! Calculate the primitive integrals (da|b)
                     CALL cossin(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                 lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                 ra, rb, kvec, cosab, sinab, dcosab, dsinab)
                     DO i = 1, 3
                        CALL contract_all(dcost(i, 1)%block, dsint(i, 1)%block, &
                                          ncoa, nsgfa(iset), sgfa, sphi_a, ldsa, &
                                          ncob, nsgfb(jset), sgfb, sphi_b, ldsb, &
                                          dcosab(:, :, i), dsinab(:, :, i), ldab, work, ldab)
                     END DO
                     ! Calculate the primitive integrals (a|db)
                     CALL cossin(lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                                 la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                                 rb, ra, kvec, cosab, sinab, dcosab, dsinab)
                     DO i = 1, 3
                        dcosab(1:ncoa, 1:ncob, i) = TRANSPOSE(dcosab(1:ncob, 1:ncoa, i))
                        dsinab(1:ncoa, 1:ncob, i) = TRANSPOSE(dsinab(1:ncob, 1:ncoa, i))
                        CALL contract_all(dcost(i, 2)%block, dsint(i, 2)%block, &
                                          ncoa, nsgfa(iset), sgfa, sphi_a, ldsa, &
                                          ncob, nsgfb(jset), sgfb, sphi_b, ldsb, &
                                          dcosab(:, :, i), dsinab(:, :, i), ldab, work, ldab)
                     END DO
                  END DO
               END DO
               forcea = 0.0_dp
               forceb = 0.0_dp
               DO ispin = 1, dft_control%nspins
                  NULLIFY (rblock, iblock)
                  CALL cp_dbcsr_get_block_p(matrix=tempmat(1, ispin)%matrix, &
                                            row=irow, col=icol, BLOCK=rblock, found=found)
                  IF(.NOT.(found))CALL cp__a("qs_efield_berry.F",1032)
                  CALL cp_dbcsr_get_block_p(matrix=tempmat(2, ispin)%matrix, &
                                            row=irow, col=icol, BLOCK=iblock, found=found)
                  IF(.NOT.(found))CALL cp__a("qs_efield_berry.F",1035)
                  n1 = SIZE(rblock, 1)
                  n2 = SIZE(rblock, 2)
                  IF(.NOT.(SIZE(iblock, 1) == n1))CALL cp__a("qs_efield_berry.F",1038)
                  IF(.NOT.(SIZE(iblock, 2) == n2))CALL cp__a("qs_efield_berry.F",1039)
                  IF(.NOT.(lsab >= n1))CALL cp__a("qs_efield_berry.F",1040)
                  IF(.NOT.(lsab >= n2))CALL cp__a("qs_efield_berry.F",1041)
                  IF (iatom <= jatom) THEN
                     DO i = 1, 3
                        forcea(i) = forcea(i)+SUM(rblock(1:n1, 1:n2)*dsint(i, 1)%block(1:n1, 1:n2)) &
                                    -SUM(iblock(1:n1, 1:n2)*dcost(i, 1)%block(1:n1, 1:n2))
                        forceb(i) = forceb(i)+SUM(rblock(1:n1, 1:n2)*dsint(i, 2)%block(1:n1, 1:n2)) &
                                    -SUM(iblock(1:n1, 1:n2)*dcost(i, 2)%block(1:n1, 1:n2))
                     END DO
                  ELSE
                     DO i = 1, 3
                        forcea(i) = forcea(i)+SUM(TRANSPOSE(rblock(1:n1, 1:n2))*dsint(i, 1)%block(1:n2, 1:n1)) &
                                    -SUM(TRANSPOSE(iblock(1:n1, 1:n2))*dcost(i, 1)%block(1:n2, 1:n1))
                        forceb(i) = forceb(i)+SUM(TRANSPOSE(rblock(1:n1, 1:n2))*dsint(i, 2)%block(1:n2, 1:n1)) &
                                    -SUM(TRANSPOSE(iblock(1:n1, 1:n2))*dcost(i, 2)%block(1:n2, 1:n1))
                     END DO
                  END IF
               END DO
               DO i = 1, 3
                  force_tmp(iatom, :, i) = force_tmp(iatom, :, i)-fab*hmat(i, idir)*forcea(:)
                  force_tmp(jatom, :, i) = force_tmp(jatom, :, i)-fab*hmat(i, idir)*forceb(:)
               END DO
            END DO
            CALL neighbor_list_iterator_release(nl_iterator)
            DEALLOCATE (basis_set_list)
         END IF
      END DO

      ! make sure the total normalized polarization is within [-1:1]
      DO idir = 1, 3
         cqi(idir) = rlog(idir)+zlog(idir)
         IF (cqi(idir) > pi) cqi(idir) = cqi(idir)-twopi
         IF (cqi(idir) < -pi) cqi(idir) = cqi(idir)+twopi
         ! now check for log branch
         IF (ABS(efield%polarisation(idir)-cqi(idir)) > pi) THEN
            di(idir) = (efield%polarisation(idir)-cqi(idir))/pi
            DO i = 1, 10
               cqi(idir) = cqi(idir)+SIGN(1.0_dp, di(idir))*twopi
               IF (ABS(efield%polarisation(idir)-cqi(idir)) < pi) EXIT
            END DO
         END IF
      END DO
      DO idir = 1, 3
         qi(idir) = 0.0_dp
         ci(idir) = 0.0_dp
         DO i = 1, 3
            ci(idir) = ci(idir)+hmat(idir, i)*cqi(i)
         END DO
      END DO

      ! update the references
      IF (calculate_forces) THEN
         ener_field = SUM(ci)
         ! check for smoothness of energy surface
         IF (ABS(efield%field_energy-ener_field) > pi*ABS(SUM(hmat))) THEN
            CALL cp__w("qs_efield_berry.F",1095,"Large change of e-field energy detected. Correct for non-smooth energy surface")
         END IF
         efield%field_energy = ener_field
         efield%polarisation(:) = cqi(:)
      END IF

      ! Energy
      ener_field = 0.0_dp
      DO i = 1, 3
         ener_field = ener_field+(fieldpol(i)-2._dp*twopi*ci(i))**2
      END DO
      energy%efield = 0.25_dp*omega/twopi*ener_field

      IF (.NOT. just_energy) THEN
         DO i = 1, 3
            di(i) = -omega*(fieldpol(i)-2._dp*twopi*ci(i))
         END DO
         ! Add the result to mo_derivativs
         DO ispin = 1, dft_control%nspins
            CALL copy_dbcsr_to_fm(mo_derivs(ispin)%matrix, mo_coeff_tmp(ispin)%matrix)
            DO idir = 1, 3
               CALL cp_fm_scale_and_add(1.0_dp, mo_coeff_tmp(ispin)%matrix, di(idir), &
                                        mo_derivs_tmp(idir, ispin)%matrix)
            END DO
         END DO
         DO ispin = 1, dft_control%nspins
            CALL copy_fm_to_dbcsr(mo_coeff_tmp(ispin)%matrix, mo_derivs(ispin)%matrix)
         END DO
      END IF

      IF (calculate_forces) THEN
         DO i = 1, 3
            DO ia = 1, natom
               CALL get_atomic_kind(particle_set(ia)%atomic_kind, kind_number=ikind)
               iatom = atom_of_kind(ia)
               force(ikind)%efield(1:3, iatom) = force(ikind)%efield(1:3, iatom)+di(i)*force_tmp(ia, 1:3, i)
            END DO
         END DO
      END IF

      DO ispin = 1, dft_control%nspins
         CALL cp_cfm_release(eigrmat(ispin)%matrix)
         CALL cp_cfm_release(inv_mat(ispin)%matrix)
         CALL cp_fm_release(mo_coeff_tmp(ispin)%matrix)
         DO i = 1, 3
            CALL cp_fm_release(mo_derivs_tmp(i, ispin)%matrix)
         END DO
         DO i = 1, SIZE(op_fm_set, 1)
            CALL cp_fm_release(opvec(i, ispin)%matrix)
            CALL cp_fm_release(op_fm_set(i, ispin)%matrix)
            CALL cp_fm_release(inv_work(i, ispin)%matrix)
         END DO
      END DO
      DEALLOCATE (inv_mat, inv_work, op_fm_set, opvec, eigrmat)
      DEALLOCATE (mo_coeff_tmp, mo_derivs_tmp)

      IF (calculate_forces) THEN
         DO ikind = 1, SIZE(atomic_kind_set)
            CALL mp_sum(force(ikind)%efield, para_env%group)
         END DO
         DEALLOCATE (atom_of_kind)
         DEALLOCATE (force_tmp)
         DEALLOCATE (cosab, sinab, work, dcosab, dsinab)
         DO i = 1, 3
            DEALLOCATE (dcost(i, 1)%block, dsint(i, 1)%block)
            DEALLOCATE (dcost(i, 2)%block, dsint(i, 2)%block)
         END DO
         CALL cp_dbcsr_deallocate_matrix_set(tempmat)
      END IF
      CALL timestop(handle)

   END SUBROUTINE qs_dispfield_derivatives

! **************************************************************************************************
!> \brief ...
!> \param cos_block ...
!> \param sin_block ...
!> \param ncoa ...
!> \param nsgfa ...
!> \param sgfa ...
!> \param sphi_a ...
!> \param ldsa ...
!> \param ncob ...
!> \param nsgfb ...
!> \param sgfb ...
!> \param sphi_b ...
!> \param ldsb ...
!> \param cosab ...
!> \param sinab ...
!> \param ldab ...
!> \param work ...
!> \param ldwork ...
! **************************************************************************************************
   SUBROUTINE contract_all(cos_block, sin_block, &
                           ncoa, nsgfa, sgfa, sphi_a, ldsa, &
                           ncob, nsgfb, sgfb, sphi_b, ldsb, &
                           cosab, sinab, ldab, work, ldwork)

      REAL(dp), DIMENSION(:, :), POINTER                 :: cos_block, sin_block
      INTEGER, INTENT(IN)                                :: ncoa, nsgfa, sgfa
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: sphi_a
      INTEGER, INTENT(IN)                                :: ldsa, ncob, nsgfb, sgfb
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: sphi_b
      INTEGER, INTENT(IN)                                :: ldsb
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: cosab, sinab
      INTEGER, INTENT(IN)                                :: ldab
      REAL(dp), DIMENSION(:, :)                          :: work
      INTEGER, INTENT(IN)                                :: ldwork

! Calculate cosine

      CALL dgemm("N", "N", ncoa, nsgfb, ncob, 1.0_dp, cosab(1, 1), ldab, &
                 sphi_b(1, sgfb), ldsb, 0.0_dp, work(1, 1), ldwork)

      CALL dgemm("T", "N", nsgfa, nsgfb, ncoa, 1.0_dp, sphi_a(1, sgfa), ldsa, &
                 work(1, 1), ldwork, 1.0_dp, cos_block(sgfa, sgfb), SIZE(cos_block, 1))

      ! Calculate sine
      CALL dgemm("N", "N", ncoa, nsgfb, ncob, 1.0_dp, sinab(1, 1), ldab, &
                 sphi_b(1, sgfb), ldsb, 0.0_dp, work(1, 1), ldwork)

      CALL dgemm("T", "N", nsgfa, nsgfb, ncoa, 1.0_dp, sphi_a(1, sgfa), ldsa, &
                 work(1, 1), ldwork, 1.0_dp, sin_block(sgfa, sgfb), SIZE(sin_block, 1))

   END SUBROUTINE contract_all

END MODULE qs_efield_berry
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief all routins needed for a nonperiodic  electric field
! **************************************************************************************************

MODULE efield_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type,&
                                              efield_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: constant_env,&
                                              custom_env,&
                                              gaussian_env,&
                                              ramp_env
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'efield_utils'

! *** Public subroutines ***

   PUBLIC :: efield_potential, &
             calculate_ecore_efield

CONTAINS

! **************************************************************************************************
!> \brief computes the time dependend potential on the grid
!> \param qs_env ...
!> \param v_efield_rspace ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE efield_potential(qs_env, v_efield_rspace)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type)                                    :: v_efield_rspace

      CHARACTER(len=*), PARAMETER :: routineN = 'efield_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo_global, bo_local
      REAL(kind=dp)                                      :: dvol, efield_ener, field(3)
      REAL(kind=dp), DIMENSION(3)                        :: dr, grid_p
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (dft_control, para_env, rho_r)
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      energy=energy, &
                      rho=rho, &
                      dft_control=dft_control, &
                      para_env=para_env)

      CALL qs_rho_get(rho, rho_r=rho_r)

      v_efield_rspace%pw%cr3d = 0.0_dp

      bo_local = v_efield_rspace%pw%pw_grid%bounds_local
      bo_global = v_efield_rspace%pw%pw_grid%bounds

      dvol = v_efield_rspace%pw%pw_grid%dvol
      dr = v_efield_rspace%pw%pw_grid%dr

      CALL make_field(dft_control, field, qs_env%sim_step, qs_env%sim_time)

      DO k = bo_local(1, 3), bo_local(2, 3)
         DO j = bo_local(1, 2), bo_local(2, 2)
            DO i = bo_local(1, 1), bo_local(2, 1)
               grid_p(1) = (i-bo_global(1, 1))*dr(1)
               grid_p(2) = (j-bo_global(1, 2))*dr(2)
               grid_p(3) = (k-bo_global(1, 3))*dr(3)
               v_efield_rspace%pw%cr3d(i, j, k) = v_efield_rspace%pw%cr3d(i, j, k)+DOT_PRODUCT(field(:), grid_p(:))
            END DO
         END DO
      END DO
      efield_ener = 0.0_dp
      DO i = 1, dft_control%nspins
         efield_ener = efield_ener+accurate_sum(v_efield_rspace%pw%cr3d*rho_r(i)%pw%cr3d)*dvol
      END DO
      CALL mp_sum(efield_ener, para_env%group)
      energy%efield = efield_ener
      CALL timestop(handle)

   END SUBROUTINE efield_potential

! **************************************************************************************************
!> \brief computes the amplitude of the efield within a given envelop
!> \param dft_control ...
!> \param field ...
!> \param sim_step ...
!> \param sim_time ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE make_field(dft_control, field, sim_step, sim_time)
      TYPE(dft_control_type)                             :: dft_control
      REAL(dp)                                           :: field(3)
      INTEGER                                            :: sim_step
      REAL(KIND=dp)                                      :: sim_time

      INTEGER                                            :: i, lower, nfield, upper
      REAL(dp)                                           :: c, env, nu, pol(3), strength
      REAL(KIND=dp)                                      :: dt
      TYPE(efield_type), POINTER                         :: efield

      c = 137.03599962875_dp
      field = 0._dp
      nfield = SIZE(dft_control%efield_fields)
      DO i = 1, nfield
         efield => dft_control%efield_fields(i)%efield
         IF (.NOT. efield%envelop_id == custom_env) nu = c/(efield%wavelength) !in case of a custom efield we do not need nu
         strength = SQRT(efield%strength/(3.50944_dp*10.0_dp**16))
         IF (DOT_PRODUCT(efield%polarisation, efield%polarisation) == 0) THEN
            pol(:) = 1.0_dp/3.0_dp
         ELSE
            pol(:) = efield%polarisation(:)/(SQRT(DOT_PRODUCT(efield%polarisation, efield%polarisation)))
         END IF
         IF (efield%envelop_id == constant_env) THEN
            IF (sim_step .GE. efield%envelop_i_vars(1) .AND. &
                (sim_step .LE. efield%envelop_i_vars(2) .OR. efield%envelop_i_vars(2) .LT. 0)) THEN
               field = field+strength*COS(sim_time*nu*2.0_dp*pi+ &
                                          efield%phase_offset*pi)*pol(:)
            END IF
         ELSE IF (efield%envelop_id == ramp_env) THEN
            IF (sim_step .GE. efield%envelop_i_vars(1) .AND. sim_step .LE. efield%envelop_i_vars(2)) &
               strength = strength*(sim_step-efield%envelop_i_vars(1))/(efield%envelop_i_vars(2)-efield%envelop_i_vars(1))
            IF (sim_step .GE. efield%envelop_i_vars(3) .AND. sim_step .LE. efield%envelop_i_vars(4)) &
               strength = strength*(efield%envelop_i_vars(4)-sim_step)/(efield%envelop_i_vars(4)-efield%envelop_i_vars(3))
            IF (sim_step .GT. efield%envelop_i_vars(4) .AND. efield%envelop_i_vars(4) .GT. 0) strength = 0.0_dp
            IF (sim_step .LE. efield%envelop_i_vars(1)) strength = 0.0_dp
            field = field+strength*COS(sim_time*nu*2.0_dp*pi+ &
                                       efield%phase_offset*pi)*pol(:)
         ELSE IF (efield%envelop_id == gaussian_env) THEN
            env = EXP(-0.5_dp*((sim_time-efield%envelop_r_vars(1))/efield%envelop_r_vars(2))**2.0_dp)
            field = field+strength*env*COS(sim_time*nu*2.0_dp*pi+ &
                                           efield%phase_offset*pi)*pol(:)
         ELSE IF (efield%envelop_id == custom_env) THEN
            dt = efield%envelop_r_vars(1)
            IF (sim_time .LT. (SIZE(efield%envelop_r_vars)-2)*dt) THEN
               !make a linear interpolation between the two next points
               lower = FLOOR(sim_time/dt)
               upper = lower+1
               strength = (efield%envelop_r_vars(lower+2)*(upper*dt-sim_time)+efield%envelop_r_vars(upper+2)*(sim_time-lower*dt))/dt
            ELSE
               strength = 0.0_dp
            ENDIF
            field = field+strength*pol(:)
         END IF
      END DO

   END SUBROUTINE make_field

! **************************************************************************************************
!> \brief computes the force and the energy due to a efield on the cores
!> \param qs_env ...
!> \param calculate_forces ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE calculate_ecore_efield(qs_env, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, OPTIONAL                                  :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_ecore_efield', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ikind, natom, &
                                                            nkind
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: my_force
      REAL(KIND=dp)                                      :: efield_ener, zeff
      REAL(KIND=dp), DIMENSION(3)                        :: field, r
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (dft_control)
      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env, dft_control=dft_control)
      IF (dft_control%apply_efield_field) THEN

         my_force = .FALSE.
         IF (PRESENT(calculate_forces)) my_force = calculate_forces

         CALL get_qs_env(qs_env=qs_env, &
                         atomic_kind_set=atomic_kind_set, &
                         qs_kind_set=qs_kind_set, &
                         energy=energy, &
                         particle_set=particle_set, &
                         cell=cell)
         efield_ener = 0.0_dp
         nkind = SIZE(atomic_kind_set)
         CALL make_field(dft_control, field, qs_env%sim_step, qs_env%sim_time)

         DO ikind = 1, SIZE(atomic_kind_set)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)

            natom = SIZE(list)
            DO iatom = 1, natom

               atom_a = list(iatom)
               r(:) = pbc(particle_set(atom_a)%r(:), cell)
               efield_ener = efield_ener-zeff*DOT_PRODUCT(r, field)
               IF (my_force) THEN
                  CALL get_qs_env(qs_env=qs_env, force=force)
                  force(ikind)%efield(:, iatom) = force(ikind)%efield(:, iatom)-field*zeff
               END IF
            END DO

         END DO
         energy%efield_core = efield_ener
      END IF
      CALL timestop(handle)
   END SUBROUTINE calculate_ecore_efield
END MODULE efield_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Density Derived atomic point charges from a QM calculation
!>      (see Bloechl, J. Chem. Phys. Vol. 103 pp. 7422-7428)
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_ddapc
   USE bibliography,                    ONLY: Blochl1995,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: ddapc_restraint_type,&
                                              dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set
   USE cp_ddapc_forces,                 ONLY: ewald_ddapc_force,&
                                              reset_ch_pulay,&
                                              restraint_functional_force,&
                                              solvation_ddapc_force
   USE cp_ddapc_util,                   ONLY: get_ddapc,&
                                              modify_hartree_pot,&
                                              restraint_functional_potential
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_spin_density
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE pw_methods,                      ONLY: pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_ddapc'

   PUBLIC :: cp_ddapc_apply_CD, & ! Apply Coupling/Decoupling to Periodic Images
             qs_ks_ddapc

CONTAINS

! **************************************************************************************************
!> \brief Set of methods using DDAPC charges
!> \param qs_env ...
!> \param auxbas_pw_pool ...
!> \param rho_tot_gspace ...
!> \param v_hartree_gspace ...
!> \param v_spin_ddapc_rest_r ...
!> \param energy ...
!> \param calculate_forces ...
!> \param ks_matrix ...
!> \param just_energy ...
!> \par History
!>      08.2005 created [tlaino]
!>      08.2008 extended to restraint/constraint DDAPC charges [fschiff]
! **************************************************************************************************
   SUBROUTINE qs_ks_ddapc(qs_env, auxbas_pw_pool, rho_tot_gspace, v_hartree_gspace, &
                          v_spin_ddapc_rest_r, energy, calculate_forces, ks_matrix, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_p_type)                                    :: rho_tot_gspace, v_hartree_gspace, &
                                                            v_spin_ddapc_rest_r
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix
      LOGICAL, INTENT(in)                                :: just_energy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_ks_ddapc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ddapc_size, handle, i, my_id
      LOGICAL                                            :: ddapc_restraint_is_spin, &
                                                            et_coupling_calc, explicit_potential
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type)                                    :: v_spin_ddapc_rest_g
      TYPE(pw_type), POINTER                             :: v_hartree_rspace

      NULLIFY (v_hartree_rspace, dft_control)

      CALL timeset(routineN, handle)
      CALL cite_reference(Blochl1995)
      ! In case decouple periodic images and/or apply restraints to charges
      logger => cp_get_default_logger()
      ddapc_restraint_is_spin = .FALSE.
      et_coupling_calc = .FALSE.
      ddapc_size = 0

      ! no k-points
      IF(.NOT.(SIZE(ks_matrix, 2) == 1))CALL cp__a("cp_ddapc.F",118)

      CALL get_qs_env(qs_env, &
                      v_hartree_rspace=v_hartree_rspace, &
                      dft_control=dft_control)

      IF (dft_control%qs_control%ddapc_restraint) THEN
         ddapc_size = SIZE(dft_control%qs_control%ddapc_restraint_control)
         IF (SIZE(energy%ddapc_restraint) .NE. ddapc_size) THEN
            DEALLOCATE (energy%ddapc_restraint)
            ALLOCATE (energy%ddapc_restraint(ddapc_size))
         END IF

         DO i = 1, SIZE(dft_control%qs_control%ddapc_restraint_control)
            my_id = dft_control%qs_control%ddapc_restraint_control(i)%ddapc_restraint_control%density_type
            IF (my_id == do_spin_density .OR. ddapc_restraint_is_spin) ddapc_restraint_is_spin = .TRUE.
         END DO
         et_coupling_calc = dft_control%qs_control%et_coupling_calc
      ENDIF

      explicit_potential = ddapc_restraint_is_spin .OR. et_coupling_calc
      dft_control%qs_control%ddapc_explicit_potential = explicit_potential
      dft_control%qs_control%ddapc_restraint_is_spin = ddapc_restraint_is_spin
      IF (explicit_potential) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, v_spin_ddapc_rest_g%pw, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         CALL pw_zero(v_spin_ddapc_rest_g%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, v_spin_ddapc_rest_r%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
      ENDIF

      IF (calculate_forces) CALL reset_ch_pulay(qs_env)

      ! Decoupling/Recoupling
      CALL cp_ddapc_apply_CD(qs_env, rho_tot_gspace, energy%hartree, v_hartree_gspace, &
                             calculate_forces, Itype_of_density="FULL DENSITY")
      IF (dft_control%qs_control%ddapc_restraint) THEN
         ! Restraints/Constraints
         DO i = 1, ddapc_size
            NULLIFY (ddapc_restraint_control)
            ddapc_restraint_control => dft_control%qs_control%ddapc_restraint_control(i)%ddapc_restraint_control

            CALL cp_ddapc_apply_RS(qs_env, energy%ddapc_restraint(i), v_hartree_gspace, &
                                   v_spin_ddapc_rest_g, ddapc_restraint_control, calculate_forces)
         END DO
      END IF
      CALL cp_ddapc_apply_RF(qs_env, rho_tot_gspace, energy%hartree, v_hartree_gspace, &
                             calculate_forces, Itype_of_density="FULL DENSITY")

      ! CJM Copying the real-space Hartree potential to KS_ENV
      IF ((.NOT. just_energy) .OR. et_coupling_calc) THEN
         CALL pw_transfer(v_hartree_gspace%pw, v_hartree_rspace)
         CALL pw_scale(v_hartree_rspace, v_hartree_rspace%pw_grid%dvol)
         IF (explicit_potential) THEN
            CALL pw_transfer(v_spin_ddapc_rest_g%pw, v_spin_ddapc_rest_r%pw)
            CALL pw_scale(v_spin_ddapc_rest_r%pw, v_spin_ddapc_rest_r%pw%pw_grid%dvol)
            IF (et_coupling_calc) THEN
               IF (qs_env%et_coupling%keep_matrix) THEN
                  IF (qs_env%et_coupling%first_run) THEN
                     NULLIFY (qs_env%et_coupling%rest_mat(1)%matrix)
                     ALLOCATE (qs_env%et_coupling%rest_mat(1)%matrix)
                     CALL cp_dbcsr_init(qs_env%et_coupling%rest_mat(1)%matrix)
                     CALL cp_dbcsr_copy(qs_env%et_coupling%rest_mat(1)%matrix, ks_matrix(1, 1)%matrix, &
                                        name="ET_RESTRAINT_MATRIX_B")
                     CALL cp_dbcsr_set(qs_env%et_coupling%rest_mat(1)%matrix, 0.0_dp)
                     CALL integrate_v_rspace(v_spin_ddapc_rest_r, &
                                             hmat=qs_env%et_coupling%rest_mat(1), &
                                             qs_env=qs_env, calculate_forces=.FALSE.)
                     qs_env%et_coupling%order_p = &
                        dft_control%qs_control%ddapc_restraint_control(1)%ddapc_restraint_control%ddapc_order_p
                     qs_env%et_coupling%e1 = dft_control%qs_control%ddapc_restraint_control(1)%ddapc_restraint_control%strength
                     qs_env%et_coupling%keep_matrix = .FALSE.
                  ELSE
                     NULLIFY (qs_env%et_coupling%rest_mat(2)%matrix)
                     ALLOCATE (qs_env%et_coupling%rest_mat(2)%matrix)
                     CALL cp_dbcsr_init(qs_env%et_coupling%rest_mat(2)%matrix)
                     CALL cp_dbcsr_copy(qs_env%et_coupling%rest_mat(2)%matrix, ks_matrix(1, 1)%matrix, &
                                        name="ET_RESTRAINT_MATRIX_B")
                     CALL cp_dbcsr_set(qs_env%et_coupling%rest_mat(2)%matrix, 0.0_dp)
                     CALL integrate_v_rspace(v_spin_ddapc_rest_r, &
                                             hmat=qs_env%et_coupling%rest_mat(2), &
                                             qs_env=qs_env, calculate_forces=.FALSE.)
                  END IF
               END IF
            END IF
         ENDIF
      ENDIF

      IF (explicit_potential) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_spin_ddapc_rest_g%pw)
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE qs_ks_ddapc

! **************************************************************************************************
!> \brief Routine to couple/decouple periodic images with the Bloechl scheme
!>
!>      The coupling/decoupling is obtaines evaluating terms E2 and E3 in
!>      J. Chem. Phys. Vol. 103 pp. 7422-7428.. The E2 terms is just a
!>      Ewald summation, and for performance reason I'm writing a specific
!>      driver instead of using and setting-up the environment of the already
!>      available routines
!> \param qs_env ...
!> \param rho_tot_gspace ...
!> \param energy ...
!> \param v_hartree_gspace ...
!> \param calculate_forces ...
!> \param Itype_of_density ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE cp_ddapc_apply_CD(qs_env, rho_tot_gspace, energy, v_hartree_gspace, &
                                calculate_forces, Itype_of_density)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type)                                    :: rho_tot_gspace
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      TYPE(pw_p_type)                                    :: v_hartree_gspace
      LOGICAL, INTENT(IN), OPTIONAL                      :: calculate_forces
      CHARACTER(LEN=*)                                   :: Itype_of_density

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_apply_CD', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: apply_decpl, need_f
      REAL(KINd=dp)                                      :: e_decpl, e_recpl
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges, radii
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dq
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: density_fit_section, force_env_section, &
                                                            multipole_section, poisson_section, &
                                                            qmmm_periodic_section

      CALL timeset(routineN, handle)
      need_f = .FALSE.
      IF (PRESENT(calculate_forces)) need_f = calculate_forces
      logger => cp_get_default_logger()
      apply_decpl = qs_env%cp_ddapc_ewald%do_decoupling .OR. qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl
      IF (apply_decpl) THEN
         ! Initialize
         NULLIFY (multipole_section, &
                  poisson_section, &
                  force_env_section, &
                  particle_set, &
                  qmmm_periodic_section, &
                  density_fit_section, &
                  charges, &
                  radii, &
                  dq, &
                  cell, &
                  super_cell)

         CALL get_qs_env(qs_env=qs_env, &
                         input=force_env_section, &
                         particle_set=particle_set, &
                         cell=cell, &
                         super_cell=super_cell)
         IF(.NOT.(ASSOCIATED(qs_env%cp_ddapc_ewald)))CALL cp__a("cp_ddapc.F",279)
         poisson_section => section_vals_get_subs_vals(force_env_section, "DFT%POISSON")

         density_fit_section => section_vals_get_subs_vals(force_env_section, "DFT%DENSITY_FITTING")

         IF (qs_env%cp_ddapc_ewald%do_decoupling) THEN
            multipole_section => section_vals_get_subs_vals(poisson_section, "MULTIPOLE")
         END IF
         IF (qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl) THEN
            qmmm_periodic_section => section_vals_get_subs_vals(force_env_section, "QMMM%PERIODIC")
            multipole_section => section_vals_get_subs_vals(qmmm_periodic_section, "MULTIPOLE")
         END IF
         ! Start the real calculation
         iw = cp_print_key_unit_nr(logger, multipole_section, "PROGRAM_RUN_INFO", &
                                   extension=".fitChargeLog")
         ! First we evaluate the charges at the corresponding SCF STEP
         IF (need_f) THEN
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           qout1=charges, &
                           out_radii=radii, &
                           dq_out=dq, &
                           ext_rho_tot_g=rho_tot_gspace%pw, &
                           Itype_of_density=Itype_of_density)
         ELSE
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           qout1=charges, &
                           out_radii=radii, &
                           ext_rho_tot_g=rho_tot_gspace%pw, &
                           Itype_of_density=Itype_of_density)
         END IF
         ! Evaluate the Ewald contribution to the decoupling/coupling E2 and E3
         IF (iw > 0) THEN
            e_decpl = 0.5_dp*DOT_PRODUCT(charges, MATMUL(qs_env%cp_ddapc_env%Md, charges))
            WRITE (iw, FMT="(T3,A,T60,F20.10)") "Decoupling Energy: ", e_decpl
         END IF
         IF (qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl .AND. (iw > 0)) THEN
            e_recpl = 0.5_dp*DOT_PRODUCT(charges, MATMUL(qs_env%cp_ddapc_env%Mr, charges))
            WRITE (iw, FMT="(T3,A,T60,F20.10)") "Recoupling Energy: ", e_recpl
         END IF
         CALL modify_hartree_pot(v_hartree_gspace, &
                                 density_fit_section, &
                                 particle_set, &
                                 qs_env%cp_ddapc_env%Mt, &
                                 qs_env%cp_ddapc_env%AmI, &
                                 radii, &
                                 charges)
         ! Modify the Hartree potential due to the decoupling/recoupling
         energy = 0.5_dp*pw_integral_ab(rho_tot_gspace%pw, v_hartree_gspace%pw)
         IF (need_f) THEN
            CALL ewald_ddapc_force(qs_env, qs_env%cp_ddapc_ewald%coeff_qm, &
                                   .FALSE., 1.0_dp, multipole_section, cell, particle_set, &
                                   radii, dq, charges)
            IF (qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl) THEN
               CALL ewald_ddapc_force(qs_env, qs_env%cp_ddapc_ewald%coeff_mm, &
                                      .TRUE., -1.0_dp, multipole_section, super_cell, particle_set, &
                                      radii, dq, charges)
            END IF
         END IF
         ! Clean the allocated arrays
         DEALLOCATE (charges)
         DEALLOCATE (radii)
         IF (ASSOCIATED(dq)) THEN
            DEALLOCATE (dq)
         END IF
         CALL cp_print_key_finished_output(iw, logger, multipole_section, &
                                           "PROGRAM_RUN_INFO")
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_ddapc_apply_CD

! **************************************************************************************************
!> \brief Routine to apply RESTRAINT/CONSTRAINTS to the density
!>      with the Bloechl scheme
!> \param qs_env ...
!> \param energy_res ...
!> \param v_hartree_gspace ...
!> \param v_spin_ddapc_rest_g ...
!> \param ddapc_restraint_control ...
!> \param calculate_forces ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE cp_ddapc_apply_RS(qs_env, energy_res, v_hartree_gspace, &
                                v_spin_ddapc_rest_g, ddapc_restraint_control, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: energy_res
      TYPE(pw_p_type)                                    :: v_hartree_gspace, v_spin_ddapc_rest_g
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      LOGICAL, INTENT(IN), OPTIONAL                      :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_apply_RS', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, my_id
      LOGICAL                                            :: apply_restrain, need_f
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges, radii
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dq
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: density_fit_section, force_env_section, &
                                                            restraint_section

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, restraint_section, force_env_section, particle_set, &
               charges, radii, dq, cell, density_fit_section, super_cell)
      need_f = .FALSE.

      CALL get_qs_env(qs_env=qs_env, &
                      input=force_env_section, &
                      particle_set=particle_set, &
                      cell=cell, &
                      super_cell=super_cell, &
                      dft_control=dft_control)

      IF (PRESENT(calculate_forces)) need_f = calculate_forces
      apply_restrain = dft_control%qs_control%ddapc_restraint
      logger => cp_get_default_logger()
      IF (apply_restrain) THEN
         ! Initialize
         density_fit_section => section_vals_get_subs_vals(force_env_section, "DFT%DENSITY_FITTING")
         restraint_section => section_vals_get_subs_vals(force_env_section, "DFT%QS%DDAPC_RESTRAINT")
         iw = cp_print_key_unit_nr(logger, restraint_section, "PROGRAM_RUN_INFO", &
                                   extension=".fitChargeLog")
         ! First we evaluate the charges at the corresponding SCF STEP
         my_id = ddapc_restraint_control%density_type
         IF (need_f) THEN
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           density_type=my_id, &
                           qout1=charges, &
                           out_radii=radii, &
                           dq_out=dq)
         ELSE
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           density_type=my_id, &
                           qout1=charges, &
                           out_radii=radii)
         END IF

         ! Modify the Hartree potential due to the restrain or the v_spin_ddapc_rest_g
         IF ((my_id == do_spin_density) .OR. dft_control%qs_control%et_coupling_calc) THEN
            CALL restraint_functional_potential(v_spin_ddapc_rest_g, density_fit_section, &
                                                particle_set, qs_env%cp_ddapc_env%AmI, radii, charges, &
                                                ddapc_restraint_control, energy_res)
         ELSE
            CALL restraint_functional_potential(v_hartree_gspace, density_fit_section, &
                                                particle_set, qs_env%cp_ddapc_env%AmI, radii, charges, &
                                                ddapc_restraint_control, energy_res)
         ENDIF

         IF (need_f) THEN
            CALL restraint_functional_force(qs_env, &
                                            ddapc_restraint_control, &
                                            dq, &
                                            charges, &
                                            SIZE(radii), &
                                            particle_set)
         END IF
         ! Clean the allocated arrays
         DEALLOCATE (charges)
         DEALLOCATE (radii)
         IF (ASSOCIATED(dq)) THEN
            DEALLOCATE (dq)
         END IF
         CALL cp_print_key_finished_output(iw, logger, restraint_section, &
                                           "PROGRAM_RUN_INFO")
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_ddapc_apply_RS

! **************************************************************************************************
!> \brief Routine to apply a reaction field during SCF (SCRF) with the Bloechl scheme
!> \param qs_env ...
!> \param rho_tot_gspace ...
!> \param energy ...
!> \param v_hartree_gspace ...
!> \param calculate_forces ...
!> \param Itype_of_density ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE cp_ddapc_apply_RF(qs_env, rho_tot_gspace, energy, &
                                v_hartree_gspace, calculate_forces, Itype_of_density)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type)                                    :: rho_tot_gspace
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      TYPE(pw_p_type)                                    :: v_hartree_gspace
      LOGICAL, INTENT(IN), OPTIONAL                      :: calculate_forces
      CHARACTER(LEN=*)                                   :: Itype_of_density

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_apply_RF', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: apply_solvation, need_f
      REAL(KINd=dp)                                      :: e_recpl
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges, radii
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dq
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: density_fit_section, force_env_section, &
                                                            solvation_section

      CALL timeset(routineN, handle)
      need_f = .FALSE.
      IF (PRESENT(calculate_forces)) need_f = calculate_forces
      logger => cp_get_default_logger()
      apply_solvation = qs_env%cp_ddapc_ewald%do_solvation
      IF (apply_solvation) THEN
         ! Initialize
         NULLIFY (force_env_section, particle_set, charges, &
                  radii, dq, cell, super_cell)

         CALL get_qs_env(qs_env=qs_env, &
                         input=force_env_section, &
                         particle_set=particle_set, &
                         cell=cell, &
                         super_cell=super_cell)

         solvation_section => section_vals_get_subs_vals(force_env_section, "DFT%SCRF")
         ! Start the real calculation
         iw = cp_print_key_unit_nr(logger, solvation_section, "PROGRAM_RUN_INFO", &
                                   extension=".fitChargeLog")
         density_fit_section => section_vals_get_subs_vals(force_env_section, "DFT%DENSITY_FITTING")
         ! First we evaluate the charges at the corresponding SCF STEP
         IF (need_f) THEN
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           qout1=charges, &
                           out_radii=radii, &
                           dq_out=dq, &
                           ext_rho_tot_g=rho_tot_gspace%pw, &
                           Itype_of_density=Itype_of_density)
         ELSE
            CALL get_ddapc(qs_env, &
                           need_f, &
                           density_fit_section, &
                           qout1=charges, &
                           out_radii=radii, &
                           ext_rho_tot_g=rho_tot_gspace%pw, &
                           Itype_of_density=Itype_of_density)
         END IF
         ! Evaluate the Ewald contribution to the decoupling/coupling E2 and E3
         IF (iw > 0) THEN
            e_recpl = 0.5_dp*DOT_PRODUCT(charges, MATMUL(qs_env%cp_ddapc_env%Ms, charges))
            WRITE (iw, FMT="(T3,A,T60,F20.10)") "Solvation  Energy: ", e_recpl
         END IF
         CALL modify_hartree_pot(v_hartree_gspace, &
                                 density_fit_section, &
                                 particle_set, &
                                 qs_env%cp_ddapc_env%Ms, &
                                 qs_env%cp_ddapc_env%AmI, &
                                 radii, &
                                 charges)
         ! Modify the Hartree potential due to the reaction field
         energy = 0.5_dp*pw_integral_ab(rho_tot_gspace%pw, v_hartree_gspace%pw)
         IF (need_f) THEN
            CALL solvation_ddapc_force(qs_env, solvation_section, particle_set, &
                                       radii, dq, charges)
         END IF
         ! Clean the allocated arrays
         DEALLOCATE (charges)
         DEALLOCATE (radii)
         IF (ASSOCIATED(dq)) THEN
            DEALLOCATE (dq)
         END IF
         CALL cp_print_key_finished_output(iw, logger, solvation_section, &
                                           "PROGRAM_RUN_INFO")
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_ddapc_apply_RF

END MODULE cp_ddapc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that build the Kohn-Sham matrix (i.e calculate the coulomb
!>      and xc parts
!> \par History
!>      05.2002 moved from qs_scf (see there the history) [fawzi]
!>      JGH [30.08.02] multi-grid arrays independent from density and potential
!>      10.2002 introduced pools, uses updated rho as input,
!>              removed most temporary variables, renamed may vars,
!>              began conversion to LSD [fawzi]
!>      10.2004 moved calculate_w_matrix here [Joost VandeVondele]
!>              introduced energy derivative wrt MOs [Joost VandeVondele]
!> \author Fawzi Mohamed
! **************************************************************************************************

MODULE qs_ks_utils
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: becke_restraint_type,&
                                              dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release_p, cp_dbcsr_scale, &
        cp_dbcsr_scale_by_vector, cp_dbcsr_set, cp_dbcsr_trace, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_ddapc,                        ONLY: cp_ddapc_apply_CD
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_ppl_grid,&
                                              sic_ad,&
                                              sic_eo,&
                                              sic_list_all,&
                                              sic_list_unpaired,&
                                              sic_mauri_spz,&
                                              sic_mauri_us,&
                                              sic_none
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grid_types,                   ONLY: PW_MODE_DISTRIBUTED
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_implicit,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind_set,&
                                              qs_kind_type
   USE qs_ks_qmmm_methods,              ONLY: qmmm_modify_hartree_pot
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE virial_types,                    ONLY: virial_type
   USE xc,                              ONLY: xc_exc_calc,&
                                              xc_vxc_pw_create1
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ks_utils'

   PUBLIC :: low_spin_roks, sic_explicit_orbitals, calc_v_sic_rspace, print_densities, &
             print_detailed_energy, compute_matrix_vxc, sum_up_and_integrate, &
             calculate_zmp_potential

CONTAINS

! **************************************************************************************************
!> \brief do ROKS calculations yielding low spin states
!> \param energy ...
!> \param qs_env ...
!> \param dft_control ...
!> \param just_energy ...
!> \param calculate_forces ...
!> \param auxbas_pw_pool ...
! **************************************************************************************************
   SUBROUTINE low_spin_roks(energy, qs_env, dft_control, just_energy, &
                            calculate_forces, auxbas_pw_pool)

      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CHARACTER(*), PARAMETER :: routineN = 'low_spin_roks', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, iterm, k, k_alpha, &
                                                            k_beta, n_rep, Nelectron, Nspin, Nterms
      INTEGER, DIMENSION(:), POINTER                     :: ivec
      INTEGER, DIMENSION(:, :, :), POINTER               :: occupations
      LOGICAL                                            :: compute_virial, in_range, &
                                                            uniform_occupation
      REAL(KIND=dp)                                      :: exc, total_rho
      REAL(KIND=dp), DIMENSION(3, 3)                     :: virial_xc_tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: energy_scaling, rvec, scaling
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_h, matrix_p, mo_derivs, rho_ao
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_deriv, fm_deriv, fm_scaled, &
                                                            mo_coeff
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: work_v_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r, tau, vxc, vxc_tau
      TYPE(pw_pool_type), POINTER                        :: xc_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, low_spin_roks_section, xc_section
      TYPE(virial_type), POINTER                         :: virial

      IF (.NOT. dft_control%low_spin_roks) RETURN
      NULLIFY (ks_env, rho_ao)

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      mo_derivs=mo_derivs, &
                      mos=mo_array, &
                      rho=rho, &
                      pw_env=pw_env, &
                      input=input, &
                      cell=cell, &
                      virial=virial)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      compute_virial = virial%pv_calculate .AND. (.NOT. virial%pv_numer)
      xc_section => section_vals_get_subs_vals(input, "DFT%XC")

      ! some assumptions need to be checked
      ! we have two spins
      IF(.NOT.(SIZE(mo_array, 1) == 2))CALL cp__a("qs_ks_utils.F",187)
      Nspin = 2
      ! we want uniform occupations
      CALL get_mo_set(mo_set=mo_array(1)%mo_set, uniform_occupation=uniform_occupation)
      IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_utils.F",191)
      CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff_b=mo_coeff, uniform_occupation=uniform_occupation)
      IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_utils.F",193)

      NULLIFY (dbcsr_deriv)
      CALL cp_dbcsr_init_p(dbcsr_deriv)
      CALL cp_dbcsr_copy(dbcsr_deriv, mo_derivs(1)%matrix)
      CALL cp_dbcsr_set(dbcsr_deriv, 0.0_dp)

      ! basic info
      CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff_b=mo_coeff)
      CALL cp_dbcsr_get_info(mo_coeff, nfullcols_total=k_alpha)
      CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff_b=mo_coeff)
      CALL cp_dbcsr_get_info(mo_coeff, nfullcols_total=k_beta)

      ! read the input
      low_spin_roks_section => section_vals_get_subs_vals(input, "DFT%LOW_SPIN_ROKS")

      CALL section_vals_val_get(low_spin_roks_section, "ENERGY_SCALING", r_vals=rvec)
      Nterms = SIZE(rvec)
      ALLOCATE (energy_scaling(Nterms))
      energy_scaling = rvec !? just wondering, should this add up to 1, in which case we should cpp?

      CALL section_vals_val_get(low_spin_roks_section, "SPIN_CONFIGURATION", n_rep_val=n_rep)
      IF(.NOT.(n_rep == Nterms))CALL cp__a("qs_ks_utils.F",215)
      CALL section_vals_val_get(low_spin_roks_section, "SPIN_CONFIGURATION", i_rep_val=1, i_vals=ivec)
      Nelectron = SIZE(ivec)
      IF(.NOT.(Nelectron == k_alpha-k_beta))CALL cp__a("qs_ks_utils.F",218)
      ALLOCATE (occupations(2, Nelectron, Nterms))
      occupations = 0
      DO iterm = 1, Nterms
         CALL section_vals_val_get(low_spin_roks_section, "SPIN_CONFIGURATION", i_rep_val=iterm, i_vals=ivec)
         IF(.NOT.(Nelectron == SIZE(ivec)))CALL cp__a("qs_ks_utils.F",223)
         in_range = ALL(ivec >= 1) .AND. ALL(ivec <= 2)
         IF(.NOT.(in_range))CALL cp__a("qs_ks_utils.F",225)
         DO k = 1, Nelectron
            occupations(ivec(k), k, iterm) = 1
         ENDDO
      ENDDO

      ! set up general data structures
      ! density matrices, kohn-sham matrices

      NULLIFY (matrix_p)
      CALL cp_dbcsr_allocate_matrix_set(matrix_p, Nspin)
      DO ispin = 1, Nspin
         ALLOCATE (matrix_p(ispin)%matrix)
         CALL cp_dbcsr_init(matrix_p(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_p(ispin)%matrix, rho_ao(1)%matrix, &
                            name="density matrix low spin roks")
         CALL cp_dbcsr_set(matrix_p(ispin)%matrix, 0.0_dp)
      ENDDO

      NULLIFY (matrix_h)
      CALL cp_dbcsr_allocate_matrix_set(matrix_h, Nspin)
      DO ispin = 1, Nspin
         ALLOCATE (matrix_h(ispin)%matrix)
         CALL cp_dbcsr_init(matrix_h(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_h(ispin)%matrix, rho_ao(1)%matrix, &
                            name="KS matrix low spin roks")
         CALL cp_dbcsr_set(matrix_h(ispin)%matrix, 0.0_dp)
      ENDDO

      ! grids in real and g space for rho and vxc
      ! tau functionals are not supported
      NULLIFY (tau, vxc_tau, vxc)
      CALL pw_env_get(pw_env, xc_pw_pool=xc_pw_pool)

      ALLOCATE (rho_r(Nspin))
      ALLOCATE (rho_g(Nspin))
      DO ispin = 1, Nspin
         CALL pw_pool_create_pw(auxbas_pw_pool, rho_r(ispin)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, rho_g(ispin)%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      ENDDO
      CALL pw_pool_create_pw(auxbas_pw_pool, work_v_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      ! get mo matrices needed to construct the density matrices
      ! we will base all on the alpha spin matrix, obviously possible in ROKS
      CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff_b=mo_coeff)
      NULLIFY (fm_scaled, fm_deriv)
      CALL cp_dbcsr_init_p(fm_scaled)
      CALL cp_dbcsr_init_p(fm_deriv)
      CALL cp_dbcsr_copy(fm_scaled, mo_coeff)
      CALL cp_dbcsr_copy(fm_deriv, mo_coeff)

      ALLOCATE (scaling(k_alpha))

      ! for each term, add it with the given scaling factor to the energy, and compute the required derivatives
      DO iterm = 1, Nterms

         DO ispin = 1, Nspin
            ! compute the proper density matrices with the required occupations
            CALL cp_dbcsr_set(matrix_p(ispin)%matrix, 0.0_dp)
            scaling = 1.0_dp
            scaling(k_alpha-Nelectron+1:k_alpha) = occupations(ispin, :, iterm)
            CALL cp_dbcsr_copy(fm_scaled, mo_coeff)
            CALL cp_dbcsr_scale_by_vector(fm_scaled, scaling, side='right')
            CALL cp_dbcsr_multiply('n', 't', 1.0_dp, mo_coeff, fm_scaled, &
                                   0.0_dp, matrix_p(ispin)%matrix, retain_sparsity=.TRUE.)
            ! compute the densities on the grid
            CALL calculate_rho_elec(matrix_p=matrix_p(ispin)%matrix, &
                                    rho=rho_r(ispin), rho_gspace=rho_g(ispin), total_rho=total_rho, &
                                    ks_env=ks_env)
         ENDDO

         ! compute the exchange energies / potential if needed
         IF (just_energy) THEN
            exc = xc_exc_calc(rho_r=rho_r, rho_g=rho_g, tau=tau, xc_section=xc_section, &
                              pw_pool=xc_pw_pool)
         ELSE
            IF(.NOT.(.NOT. compute_virial))CALL cp__a("qs_ks_utils.F",307)
            CALL xc_vxc_pw_create1(vxc_rho=vxc, rho_r=rho_r, &
                                   rho_g=rho_g, tau=tau, vxc_tau=vxc_tau, exc=exc, xc_section=xc_section, &
                                   pw_pool=xc_pw_pool, compute_virial=.FALSE., virial_xc=virial_xc_tmp)
         END IF

         energy%exc = energy%exc+energy_scaling(iterm)*exc

         ! add the corresponding derivatives to the MO derivatives
         IF (.NOT. just_energy) THEN
            ! get the potential in matrix form
            DO ispin = 1, Nspin
               ! use a work_v_rspace
               work_v_rspace%pw%cr3d = (energy_scaling(iterm)*vxc(ispin)%pw%pw_grid%dvol)* &
                                       vxc(ispin)%pw%cr3d
               ! zero first ?!
               CALL cp_dbcsr_set(matrix_h(ispin)%matrix, 0.0_dp)
               CALL integrate_v_rspace(v_rspace=work_v_rspace, pmat=matrix_p(ispin), hmat=matrix_h(ispin), &
                                       qs_env=qs_env, calculate_forces=calculate_forces)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, vxc(ispin)%pw)
            ENDDO
            DEALLOCATE (vxc)

            ! add this to the mo_derivs, again based on the alpha mo_coeff
            DO ispin = 1, Nspin
               CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_h(ispin)%matrix, mo_coeff, &
                                      0.0_dp, dbcsr_deriv, last_column=k_alpha)

               scaling = 1.0_dp
               scaling(k_alpha-Nelectron+1:k_alpha) = occupations(ispin, :, iterm)
               CALL cp_dbcsr_scale_by_vector(dbcsr_deriv, scaling, side='right')
               CALL cp_dbcsr_add(mo_derivs(1)%matrix, dbcsr_deriv, 1.0_dp, 1.0_dp)
            ENDDO

         ENDIF

      ENDDO

      ! release allocated memory
      DO ispin = 1, Nspin
         CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r(ispin)%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_g(ispin)%pw)
      ENDDO
      DEALLOCATE (rho_r, rho_g)
      CALL cp_dbcsr_deallocate_matrix_set(matrix_p)
      CALL cp_dbcsr_deallocate_matrix_set(matrix_h)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, work_v_rspace%pw)

      CALL cp_dbcsr_release_p(fm_deriv)
      CALL cp_dbcsr_release_p(fm_scaled)

      DEALLOCATE (occupations)
      DEALLOCATE (energy_scaling)
      DEALLOCATE (scaling)

      CALL cp_dbcsr_release_p(dbcsr_deriv)

      CALL timestop(handle)

   END SUBROUTINE low_spin_roks
! **************************************************************************************************
!> \brief do sic calculations on explicit orbitals
!> \param energy ...
!> \param qs_env ...
!> \param dft_control ...
!> \param poisson_env ...
!> \param just_energy ...
!> \param calculate_forces ...
!> \param auxbas_pw_pool ...
! **************************************************************************************************
   SUBROUTINE sic_explicit_orbitals(energy, qs_env, dft_control, poisson_env, just_energy, &
                                    calculate_forces, auxbas_pw_pool)

      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CHARACTER(*), PARAMETER :: routineN = 'sic_explicit_orbitals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, Iorb, k_alpha, k_beta, Norb
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: sic_orbital_list
      LOGICAL                                            :: compute_virial, uniform_occupation
      REAL(KIND=dp)                                      :: ener, exc, total_rho
      REAL(KIND=dp), DIMENSION(3, 3)                     :: virial_xc_tmp
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type)                              :: orb_density_matrix_p, orb_h_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mo_derivs, rho_ao, tmp_dbcsr
      TYPE(cp_dbcsr_type), POINTER                       :: orb_density_matrix, orb_h
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_derivs_local
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: matrix_hv, matrix_v, mo_coeff
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: orb_rho_g, orb_rho_r, tmp_g, tmp_r, &
                                                            work_v_gspace, work_v_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r, tau, vxc, vxc_tau
      TYPE(pw_pool_type), POINTER                        :: xc_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, xc_section
      TYPE(virial_type), POINTER                         :: virial

      IF (dft_control%sic_method_id .NE. sic_eo) RETURN

      CALL timeset(routineN, handle)

      NULLIFY (tau, vxc_tau, mo_derivs, ks_env, rho_ao)

      ! generate the lists of orbitals that need sic treatment
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      mo_derivs=mo_derivs, &
                      mos=mo_array, &
                      rho=rho, &
                      pw_env=pw_env, &
                      input=input, &
                      cell=cell, &
                      virial=virial)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      compute_virial = virial%pv_calculate .AND. (.NOT. virial%pv_numer)
      xc_section => section_vals_get_subs_vals(input, "DFT%XC")

      DO i = 1, SIZE(mo_array) !fm->dbcsr
         IF (mo_array(i)%mo_set%use_mo_coeff_b) THEN !fm->dbcsr
            CALL copy_dbcsr_to_fm(mo_array(i)%mo_set%mo_coeff_b, &
                                  mo_array(i)%mo_set%mo_coeff) !fm->dbcsr
         ENDIF !fm->dbcsr
      ENDDO !fm->dbcsr

      CALL pw_env_get(pw_env, xc_pw_pool=xc_pw_pool)

      ! we have two spins
      IF(.NOT.(SIZE(mo_array, 1) == 2))CALL cp__a("qs_ks_utils.F",446)
      ! we want uniform occupations
      CALL get_mo_set(mo_set=mo_array(1)%mo_set, uniform_occupation=uniform_occupation)
      IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_utils.F",449)
      CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff=mo_coeff, uniform_occupation=uniform_occupation)
      IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_utils.F",451)

      NULLIFY (tmp_dbcsr)
      CALL cp_dbcsr_allocate_matrix_set(tmp_dbcsr, SIZE(mo_derivs, 1))
      DO i = 1, SIZE(mo_derivs, 1) !fm->dbcsr
         !
         NULLIFY (tmp_dbcsr(i)%matrix)
         CALL cp_dbcsr_init_p(tmp_dbcsr(i)%matrix)
         CALL cp_dbcsr_copy(tmp_dbcsr(i)%matrix, mo_derivs(i)%matrix)
         CALL cp_dbcsr_set(tmp_dbcsr(i)%matrix, 0.0_dp)
      ENDDO !fm->dbcsr

      k_alpha = 0; k_beta = 0
      SELECT CASE (dft_control%sic_list_id)
      CASE (sic_list_all)

         CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=k_alpha)

         IF (SIZE(mo_array, 1) > 1) THEN
            CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_get_info(mo_coeff, ncol_global=k_beta)
         ENDIF

         Norb = k_alpha+k_beta
         ALLOCATE (sic_orbital_list(3, Norb))

         iorb = 0
         DO i = 1, k_alpha
            iorb = iorb+1
            sic_orbital_list(1, iorb) = 1
            sic_orbital_list(2, iorb) = i
            sic_orbital_list(3, iorb) = 1
         ENDDO
         DO i = 1, k_beta
            iorb = iorb+1
            sic_orbital_list(1, iorb) = 2
            sic_orbital_list(2, iorb) = i
            IF (SIZE(mo_derivs, 1) == 1) THEN
               sic_orbital_list(3, iorb) = 1
            ELSE
               sic_orbital_list(3, iorb) = 2
            ENDIF
         ENDDO

      CASE (sic_list_unpaired)
         ! we have two spins
         IF(.NOT.(SIZE(mo_array, 1) == 2))CALL cp__a("qs_ks_utils.F",498)
         ! we have them restricted
         IF(.NOT.(SIZE(mo_derivs, 1) == 1))CALL cp__a("qs_ks_utils.F",500)
         IF(.NOT.(dft_control%restricted))CALL cp__a("qs_ks_utils.F",501)

         CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=k_alpha)

         CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=k_beta)

         Norb = k_alpha-k_beta
         ALLOCATE (sic_orbital_list(3, Norb))

         iorb = 0
         DO i = k_beta+1, k_alpha
            iorb = iorb+1
            sic_orbital_list(1, iorb) = 1
            sic_orbital_list(2, iorb) = i
            ! we are guaranteed to be restricted
            sic_orbital_list(3, iorb) = 1
         ENDDO

      CASE DEFAULT
         CALL cp__b("qs_ks_utils.F",522,"")
      END SELECT

      ! data needed for each of the orbs
      CALL pw_pool_create_pw(auxbas_pw_pool, orb_rho_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, tmp_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, orb_rho_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, tmp_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, work_v_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, work_v_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      ALLOCATE (orb_density_matrix)
      CALL cp_dbcsr_init(orb_density_matrix)
      CALL cp_dbcsr_copy(orb_density_matrix, rho_ao(1)%matrix, &
                         name="orb_density_matrix")
      CALL cp_dbcsr_set(orb_density_matrix, 0.0_dp)
      orb_density_matrix_p%matrix => orb_density_matrix

      ALLOCATE (orb_h)
      CALL cp_dbcsr_init(orb_h)
      CALL cp_dbcsr_copy(orb_h, rho_ao(1)%matrix, &
                         name="orb_density_matrix")
      CALL cp_dbcsr_set(orb_h, 0.0_dp)
      orb_h_p%matrix => orb_h

      CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff=mo_coeff)

      CALL cp_fm_struct_create(fm_struct_tmp, ncol_global=1, &
                               template_fmstruct=mo_coeff%matrix_struct)
      CALL cp_fm_create(matrix_v, fm_struct_tmp, name="matrix_v")
      CALL cp_fm_create(matrix_hv, fm_struct_tmp, name="matrix_hv")
      CALL cp_fm_struct_release(fm_struct_tmp)

      ALLOCATE (mo_derivs_local(SIZE(mo_array, 1)))
      DO I = 1, SIZE(mo_array, 1)
         CALL get_mo_set(mo_set=mo_array(i)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_create(mo_derivs_local(I)%matrix, mo_coeff%matrix_struct)
      ENDDO

      ALLOCATE (rho_r(2))
      rho_r(1)%pw => orb_rho_r%pw
      rho_r(2)%pw => tmp_r%pw
      CALL pw_zero(tmp_r%pw)

      ALLOCATE (rho_g(2))
      rho_g(1)%pw => orb_rho_g%pw
      rho_g(2)%pw => tmp_g%pw
      CALL pw_zero(tmp_g%pw)

      NULLIFY (vxc)
      ! ALLOCATE(vxc(2),stat=stat)
      ! CPPostcondition(stat==0,cp_failure_level,routineP,failure)
      ! CALL pw_pool_create_pw(xc_pw_pool,vxc(1)%pw,&
      !         in_space=REALSPACE, use_data=REALDATA3D)
      ! CALL pw_pool_create_pw(xc_pw_pool,vxc(2)%pw,&
      !         in_space=REALSPACE, use_data=REALDATA3D)

      ! now apply to SIC correction to each selected orbital
      DO iorb = 1, Norb
         ! extract the proper orbital from the mo_coeff
         CALL get_mo_set(mo_set=mo_array(sic_orbital_list(1, iorb))%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_to_fm(mo_coeff, matrix_v, 1, sic_orbital_list(2, iorb), 1)

         ! construct the density matrix and the corresponding density
         CALL cp_dbcsr_set(orb_density_matrix, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(orb_density_matrix, matrix_v=matrix_v, ncol=1, &
                                    alpha=1.0_dp)

         CALL calculate_rho_elec(matrix_p=orb_density_matrix, &
                                 rho=orb_rho_r, rho_gspace=orb_rho_g, total_rho=total_rho, &
                                 ks_env=ks_env)

         ! write(*,*) 'Orbital ',sic_orbital_list(1,iorb),sic_orbital_list(2,iorb)
         ! write(*,*) 'Total orbital rho= ',total_rho

         ! compute the energy functional for this orbital and its derivative

         CALL pw_poisson_solve(poisson_env, orb_rho_g%pw, ener, work_v_gspace%pw)
         energy%hartree = energy%hartree-dft_control%sic_scaling_a*ener
         IF (.NOT. just_energy) THEN
            CALL pw_transfer(work_v_gspace%pw, work_v_rspace%pw)
            CALL pw_scale(work_v_rspace%pw, -dft_control%sic_scaling_a*work_v_rspace%pw%pw_grid%dvol)
            CALL cp_dbcsr_set(orb_h, 0.0_dp)
         ENDIF

         IF (just_energy) THEN
            exc = xc_exc_calc(rho_r=rho_r, rho_g=rho_g, tau=tau, xc_section=xc_section, &
                              pw_pool=xc_pw_pool)
         ELSE
            IF(.NOT.(.NOT. compute_virial))CALL cp__a("qs_ks_utils.F",623)
            CALL xc_vxc_pw_create1(vxc_rho=vxc, rho_r=rho_r, &
                                   rho_g=rho_g, tau=tau, vxc_tau=vxc_tau, exc=exc, xc_section=xc_section, &
                                   pw_pool=xc_pw_pool, compute_virial=compute_virial, virial_xc=virial_xc_tmp)
            ! add to the existing work_v_rspace
            work_v_rspace%pw%cr3d = work_v_rspace%pw%cr3d- &
                                    dft_control%sic_scaling_b*vxc(1)%pw%pw_grid%dvol*vxc(1)%pw%cr3d
         END IF
         energy%exc = energy%exc-dft_control%sic_scaling_b*exc

         IF (.NOT. just_energy) THEN
            ! note, orb_h (which is being pointed to with orb_h_p) is zeroed above
            CALL integrate_v_rspace(v_rspace=work_v_rspace, pmat=orb_density_matrix_p, hmat=orb_h_p, &
                                    qs_env=qs_env, calculate_forces=calculate_forces)

            ! add this to the mo_derivs
            CALL cp_dbcsr_sm_fm_multiply(orb_h, matrix_v, matrix_hv, 1)
            ! silly trick, copy to an array of the right size and add to mo_derivs
            CALL cp_fm_set_all(mo_derivs_local(sic_orbital_list(3, iorb))%matrix, 0.0_dp)
            CALL cp_fm_to_fm(matrix_hv, mo_derivs_local(sic_orbital_list(3, iorb))%matrix, 1, 1, sic_orbital_list(2, iorb))
            CALL copy_fm_to_dbcsr(mo_derivs_local(sic_orbital_list(3, iorb))%matrix, &
                                  tmp_dbcsr(sic_orbital_list(3, iorb))%matrix)
            CALL cp_dbcsr_add(mo_derivs(sic_orbital_list(3, iorb))%matrix, &
                              tmp_dbcsr(sic_orbital_list(3, iorb))%matrix, 1.0_dp, 1.0_dp)
            !
            ! need to deallocate vxc
            CALL pw_pool_give_back_pw(xc_pw_pool, vxc(1)%pw)
            CALL pw_pool_give_back_pw(xc_pw_pool, vxc(2)%pw)
            DEALLOCATE (vxc)

         ENDIF

      ENDDO

      CALL pw_pool_give_back_pw(auxbas_pw_pool, orb_rho_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, tmp_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, orb_rho_g%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, tmp_g%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, work_v_gspace%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, work_v_rspace%pw)

      CALL cp_dbcsr_deallocate_matrix(orb_density_matrix)
      CALL cp_dbcsr_deallocate_matrix(orb_h)
      CALL cp_fm_release(matrix_v)
      CALL cp_fm_release(matrix_hv)
      DO I = 1, SIZE(mo_derivs_local, 1)
         CALL cp_fm_release(mo_derivs_local(I)%matrix)
      ENDDO
      DEALLOCATE (mo_derivs_local)
      DEALLOCATE (rho_r)
      DEALLOCATE (rho_g)

      CALL cp_dbcsr_deallocate_matrix_set(tmp_dbcsr) !fm->dbcsr

      CALL timestop(handle)

   END SUBROUTINE sic_explicit_orbitals

! **************************************************************************************************
!> \brief do sic calculations on the spin density
!> \param v_sic_rspace ...
!> \param energy ...
!> \param qs_env ...
!> \param dft_control ...
!> \param rho ...
!> \param poisson_env ...
!> \param just_energy ...
!> \param calculate_forces ...
!> \param auxbas_pw_pool ...
! **************************************************************************************************
   SUBROUTINE calc_v_sic_rspace(v_sic_rspace, energy, &
                                qs_env, dft_control, rho, poisson_env, just_energy, &
                                calculate_forces, auxbas_pw_pool)

      TYPE(pw_p_type), INTENT(INOUT)                     :: v_sic_rspace
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      LOGICAL, INTENT(IN)                                :: just_energy, calculate_forces
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CHARACTER(*), PARAMETER :: routineN = 'calc_v_sic_rspace', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nelec, nelec_a, nelec_b, nforce
      REAL(kind=dp)                                      :: ener, full_scaling, scaling
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: store_forces
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(pw_p_type)                                    :: work_rho, work_v
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force

      NULLIFY (mo_array, rho_g)

      IF (dft_control%sic_method_id == sic_none) RETURN
      IF (dft_control%sic_method_id == sic_eo) RETURN

      IF (dft_control%qs_control%gapw) &
         CALL cp__b("qs_ks_utils.F",722,"sic and GAPW not yet compatible")

      ! OK, right now we like two spins to do sic, could be relaxed for AD
      IF(.NOT.(dft_control%nspins == 2))CALL cp__a("qs_ks_utils.F",725)

      CALL pw_pool_create_pw(auxbas_pw_pool, work_rho%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, work_v%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL qs_rho_get(rho, rho_g=rho_g)

      ! Hartree sic corrections
      SELECT CASE (dft_control%sic_method_id)
      CASE (sic_mauri_us, sic_mauri_spz)
         CALL pw_copy(rho_g(1)%pw, work_rho%pw)
         CALL pw_axpy(rho_g(2)%pw, work_rho%pw, alpha=-1._dp)
         CALL pw_poisson_solve(poisson_env, work_rho%pw, ener, work_v%pw)
      CASE (sic_ad)
         ! find out how many elecs we have
         CALL get_qs_env(qs_env, mos=mo_array)
         CALL get_mo_set(mo_set=mo_array(1)%mo_set, nelectron=nelec_a)
         CALL get_mo_set(mo_set=mo_array(2)%mo_set, nelectron=nelec_b)
         nelec = nelec_a+nelec_b
         CALL pw_copy(rho_g(1)%pw, work_rho%pw)
         CALL pw_axpy(rho_g(2)%pw, work_rho%pw)
         scaling = 1.0_dp/REAL(nelec, KIND=dp)
         CALL pw_scale(work_rho%pw, scaling)
         CALL pw_poisson_solve(poisson_env, work_rho%pw, ener, work_v%pw)
      CASE DEFAULT
         CALL cp__b("qs_ks_utils.F",754,"Unknown sic method id")
      END SELECT

      ! Correct for  DDAP charges (if any)
      ! storing whatever force might be there from previous decoupling
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)
         nforce = 0
         DO i = 1, SIZE(force)
            nforce = nforce+SIZE(force(i)%ch_pulay, 2)
         ENDDO
         ALLOCATE (store_forces(3, nforce))
         nforce = 0
         DO i = 1, SIZE(force)
            store_forces(1:3, nforce+1:nforce+SIZE(force(i)%ch_pulay, 2)) = force(i)%ch_pulay(:, :)
            force(i)%ch_pulay(:, :) = 0.0_dp
            nforce = nforce+SIZE(force(i)%ch_pulay, 2)
         ENDDO
      ENDIF

      CALL cp_ddapc_apply_CD(qs_env, &
                             work_rho, &
                             ener, &
                             v_hartree_gspace=work_v, &
                             calculate_forces=calculate_forces, &
                             Itype_of_density="SPIN")

      SELECT CASE (dft_control%sic_method_id)
      CASE (sic_mauri_us, sic_mauri_spz)
         full_scaling = -dft_control%sic_scaling_a
      CASE (sic_ad)
         full_scaling = -dft_control%sic_scaling_a*nelec
      CASE DEFAULT
         CALL cp__b("qs_ks_utils.F",787,"Unknown sic method id")
      END SELECT
      energy%hartree = energy%hartree+full_scaling*ener

      ! add scaled forces, restoring the old
      IF (calculate_forces) THEN
         nforce = 0
         DO i = 1, SIZE(force)
            force(i)%ch_pulay(:, :) = force(i)%ch_pulay(:, :)*full_scaling+ &
                                      store_forces(1:3, nforce+1:nforce+SIZE(force(i)%ch_pulay, 2))
            nforce = nforce+SIZE(force(i)%ch_pulay, 2)
         ENDDO
      ENDIF

      IF (.NOT. just_energy) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, v_sic_rspace%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_transfer(work_v%pw, v_sic_rspace%pw)
         ! also take into account the scaling (in addition to the volume element)
         CALL pw_scale(v_sic_rspace%pw, &
                       dft_control%sic_scaling_a*v_sic_rspace%pw%pw_grid%dvol)
      ENDIF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, work_rho%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, work_v%pw)

   END SUBROUTINE calc_v_sic_rspace

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rho ...
! **************************************************************************************************
   SUBROUTINE print_densities(qs_env, rho)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho

      INTEGER                                            :: img, ispin, n_electrons, output_unit
      REAL(dp)                                           :: tot1_h, tot1_s, tot_rho_r, trace, &
                                                            trace_tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r_arr
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: input, scf_section

      NULLIFY (qs_charges, qs_kind_set, cell, input, logger, scf_section, matrix_s, &
               dft_control, tot_rho_r_arr, rho_ao)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, qs_charges=qs_charges, &
                      input=input, &
                      matrix_s_kp=matrix_s, &
                      dft_control=dft_control)

      CALL get_qs_kind_set(qs_kind_set, nelectron=n_electrons)

      scf_section => section_vals_get_subs_vals(input, "DFT%SCF")
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%TOTAL_DENSITIES", &
                                         extension=".scfLog")

      CALL qs_rho_get(rho, tot_rho_r=tot_rho_r_arr, rho_ao_kp=rho_ao)
      n_electrons = n_electrons-dft_control%charge
      tot_rho_r = accurate_sum(tot_rho_r_arr)

      trace = 0
      IF (BTEST(cp_print_key_should_output(logger%iter_info, scf_section, "PRINT%TOTAL_DENSITIES"), cp_p_file)) THEN
         DO ispin = 1, dft_control%nspins
            DO img = 1, dft_control%nimages
               CALL cp_dbcsr_trace(rho_ao(ispin, img)%matrix, matrix_s(1, img)%matrix, trace_tmp)
               trace = trace+trace_tmp
            END DO
         END DO
      ENDIF

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T3,A,T41,F20.10)") "Trace(PS):", trace
         WRITE (UNIT=output_unit, FMT="((T3,A,T41,2F20.10))") &
            "Electronic density on regular grids: ", &
            tot_rho_r, &
            tot_rho_r+ &
            REAL(n_electrons, dp), &
            "Core density on regular grids:", &
            qs_charges%total_rho_core_rspace, &
            qs_charges%total_rho_core_rspace-REAL(n_electrons+dft_control%charge, dp)
      END IF
      IF (dft_control%qs_control%gapw) THEN
         tot1_h = qs_charges%total_rho1_hard(1)
         tot1_s = qs_charges%total_rho1_soft(1)
         DO ispin = 2, dft_control%nspins
            tot1_h = tot1_h+qs_charges%total_rho1_hard(ispin)
            tot1_s = tot1_s+qs_charges%total_rho1_soft(ispin)
         END DO
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="((T3,A,T41,2F20.10))") &
               "Hard and soft densities (Lebedev):", &
               tot1_h, tot1_s
            WRITE (UNIT=output_unit, FMT="(T3,A,T41,F20.10)") &
               "Total Rho_soft + Rho1_hard - Rho1_soft (r-space): ", &
               tot_rho_r+tot1_h-tot1_s, &
               "Total charge density (r-space):      ", &
               tot_rho_r+tot1_h-tot1_s &
               +qs_charges%total_rho_core_rspace, &
               "Total Rho_soft + Rho0_soft (g-space):", &
               qs_charges%total_rho_gspace
         END IF
         qs_charges%background = tot_rho_r+tot1_h-tot1_s+ &
                                 qs_charges%total_rho_core_rspace
      ELSE IF (dft_control%qs_control%gapw_xc) THEN
         tot1_h = qs_charges%total_rho1_hard(1)
         tot1_s = qs_charges%total_rho1_soft(1)
         DO ispin = 2, dft_control%nspins
            tot1_h = tot1_h+qs_charges%total_rho1_hard(ispin)
            tot1_s = tot1_s+qs_charges%total_rho1_soft(ispin)
         END DO
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T41,2F20.10))") &
               "Hard and soft densities (Lebedev):", &
               tot1_h, tot1_s
            WRITE (UNIT=output_unit, FMT="(T3,A,T41,F20.10)") &
               "Total Rho_soft + Rho1_hard - Rho1_soft (r-space): ", &
               accurate_sum(tot_rho_r_arr)+tot1_h-tot1_s
         END IF
         qs_charges%background = tot_rho_r+ &
                                 qs_charges%total_rho_core_rspace
      ELSE
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T41,F20.10)") &
               "Total charge density on r-space grids:     ", &
               tot_rho_r+ &
               qs_charges%total_rho_core_rspace, &
               "Total charge density g-space grids:     ", &
               qs_charges%total_rho_gspace
         END IF
         qs_charges%background = tot_rho_r+ &
                                 qs_charges%total_rho_core_rspace
      END IF
      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="()")
      qs_charges%background = qs_charges%background/cell%deth

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%TOTAL_DENSITIES")

   END SUBROUTINE print_densities

! **************************************************************************************************
!> \brief Print detailed energies
!>
!> \param qs_env ...
!> \param dft_control ...
!> \param input ...
!> \param energy ...
!> \param mulliken_order_p ...
!> \par History
!>    refactoring 04.03.2011 [MI]
!> \author
! **************************************************************************************************
   SUBROUTINE print_detailed_energy(qs_env, dft_control, input, energy, mulliken_order_p)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(qs_energy_type), POINTER                      :: energy
      REAL(KIND=dp), INTENT(IN)                          :: mulliken_order_p

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_detailed_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bc, n, output_unit, psolver
      REAL(KIND=dp)                                      :: ddapc_order_p, implicit_ps_ehartree, &
                                                            s2_order_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_env_type), POINTER                         :: pw_env

      logger => cp_get_default_logger()

      NULLIFY (pw_env)
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      psolver = pw_env%poisson_env%parameters%solver

      output_unit = cp_print_key_unit_nr(logger, input, "DFT%SCF%PRINT%DETAILED_ENERGY", &
                                         extension=".scfLog")
      IF (output_unit > 0) THEN
         IF (dft_control%do_admm) THEN
            WRITE (UNIT=output_unit, FMT="((T3,A,T60,F20.10))") &
               "Wfn fit exchange-correlation energy:           ", energy%exc_aux_fit
         END IF
         IF (dft_control%do_admm) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
               "Core Hamiltonian energy:                       ", energy%core, &
               "Hartree energy:                                ", energy%hartree, &
               "Exchange-correlation energy:                   ", energy%exc+energy%exc_aux_fit
         ELSE
!ZMP to print some variables at each step
            IF (dft_control%apply_external_density) THEN
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                  "DOING ZMP CALCULATION FROM EXTERNAL DENSITY    "
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                  "Core Hamiltonian energy:                       ", energy%core, &
                  "Hartree energy:                                ", energy%hartree
            ELSE IF (dft_control%apply_external_vxc) THEN
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                  "DOING ZMP READING EXTERNAL VXC                 "
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                  "Core Hamiltonian energy:                       ", energy%core, &
                  "Hartree energy:                                ", energy%hartree
            ELSE
               IF (psolver .EQ. pw_poisson_implicit) THEN
                  implicit_ps_ehartree = pw_env%poisson_env%implicit_env%ehartree
                  bc = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
                  SELECT CASE (bc)
                  CASE (MIXED_PERIODIC_BC, MIXED_BC)
                     WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                        "Core Hamiltonian energy:                       ", energy%core, &
                        "Hartree energy:                                ", implicit_ps_ehartree, &
                        "Electric enthalpy:                             ", energy%hartree, &
                        "Exchange-correlation energy:                   ", energy%exc
                  CASE (PERIODIC_BC, NEUMANN_BC)
                     WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                        "Core Hamiltonian energy:                       ", energy%core, &
                        "Hartree energy:                                ", energy%hartree, &
                        "Exchange-correlation energy:                   ", energy%exc
                  END SELECT
               ELSE
                  WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
                     "Core Hamiltonian energy:                       ", energy%core, &
                     "Hartree energy:                                ", energy%hartree, &
                     "Exchange-correlation energy:                   ", energy%exc
               END IF
            END IF
         END IF

         IF (dft_control%apply_external_density) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
               "Integral of the (density * v_xc):              ", energy%exc
         END IF

         IF (energy%e_hartree /= 0.0_dp) &
            WRITE (UNIT=output_unit, FMT="(T3,A,T61,F20.10)") &
            "Coulomb (electron-electron) energy:            ", energy%e_hartree
         IF (energy%dispersion /= 0.0_dp) &
            WRITE (UNIT=output_unit, FMT="(T3,A,T61,F20.10)") &
            "Dispersion energy:                             ", energy%dispersion
         IF (dft_control%qs_control%gapw) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
               "GAPW| Exc from hard and soft atomic rho1:      ", energy%exc1, &
               "GAPW| local Eh = 1 center integrals:           ", energy%hartree_1c
         END IF
         IF (dft_control%qs_control%gapw_xc) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T61,F20.10))") &
               "GAPW| Exc from hard and soft atomic rho1:      ", energy%exc1
         END IF
         IF (dft_control%dft_plus_u) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T61,F20.10)") &
               "DFT+U energy:", energy%dft_plus_u
         END IF
         IF (qs_env%qmmm) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T61,F20.10)") &
               "QM/MM Electrostatic energy:                    ", energy%qmmm_el
            IF (qs_env%qmmm_env_qm%image_charge) THEN
               WRITE (UNIT=output_unit, FMT="(T3,A,T61,F20.10)") &
                  "QM/MM image charge energy:                ", energy%image_charge
            ENDIF
         END IF
         IF (dft_control%qs_control%mulliken_restraint) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T41,2F20.10)") &
               "Mulliken restraint (order_p,energy) : ", mulliken_order_p, energy%mulliken
         ENDIF
         IF (dft_control%qs_control%ddapc_restraint) THEN
            DO n = 1, SIZE(dft_control%qs_control%ddapc_restraint_control)
               ddapc_order_p = &
                  dft_control%qs_control%ddapc_restraint_control(n)%ddapc_restraint_control%ddapc_order_p
               WRITE (UNIT=output_unit, FMT="(T3,A,T41,2F20.10)") &
                  "DDAPC restraint (order_p,energy) : ", ddapc_order_p, energy%ddapc_restraint(n)
            END DO
         ENDIF
         IF (dft_control%qs_control%s2_restraint) THEN
            s2_order_p = dft_control%qs_control%s2_restraint_control%s2_order_p
            WRITE (UNIT=output_unit, FMT="(T3,A,T41,2F20.10)") &
               "S2 restraint (order_p,energy) : ", s2_order_p, energy%s2_restraint
         ENDIF

      END IF ! output_unit
      CALL cp_print_key_finished_output(output_unit, logger, input, &
                                        "DFT%SCF%PRINT%DETAILED_ENERGY")

   END SUBROUTINE print_detailed_energy

! **************************************************************************************************
!> \brief compute matrix_vxc, defined via the potential created by qs_vxc_create
!>        ignores things like tau functional, gapw, sic, ...
!>         so only OK for GGA & GPW right now
!> \param qs_env ...
!> \param v_rspace ...
!> \param matrix_vxc ...
!> \par History
!>    created 23.10.2012 [Joost VandeVondele]
!> \author
! **************************************************************************************************
   SUBROUTINE compute_matrix_vxc(qs_env, v_rspace, matrix_vxc)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: v_rspace
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_vxc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_matrix_vxc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      LOGICAL                                            :: gapw
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      ! create the matrix using matrix_ks as a template
      IF (ASSOCIATED(matrix_vxc)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(matrix_vxc)
      ENDIF
      CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
      ALLOCATE (matrix_vxc(SIZE(matrix_ks)))
      DO ispin = 1, SIZE(matrix_ks)
         NULLIFY (matrix_vxc(ispin)%matrix)
         CALL cp_dbcsr_init_p(matrix_vxc(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_vxc(ispin)%matrix, matrix_ks(ispin)%matrix, &
                            name="Matrix VXC of spin "//cp_to_string(ispin))
         CALL cp_dbcsr_set(matrix_vxc(ispin)%matrix, 0.0_dp)
      ENDDO

      ! and integrate
      CALL get_qs_env(qs_env, dft_control=dft_control)
      gapw = dft_control%qs_control%gapw .OR. dft_control%qs_control%gapw_xc
      DO ispin = 1, SIZE(matrix_ks)
         CALL integrate_v_rspace(v_rspace=v_rspace(ispin), &
                                 hmat=matrix_vxc(ispin), &
                                 qs_env=qs_env, &
                                 calculate_forces=.FALSE., &
                                 gapw=gapw)
         ! scale by the volume element... should really become part of integrate_v_rspace
         CALL cp_dbcsr_scale(matrix_vxc(ispin)%matrix, v_rspace(ispin)%pw%pw_grid%dvol)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE compute_matrix_vxc

! **************************************************************************************************
!> \brief Sum up all potentials defined  on the grid and integrate
!>
!> \param qs_env ...
!> \param ks_matrix ...
!> \param rho ...
!> \param my_rho ...
!> \param vppl_rspace ...
!> \param v_rspace_new ...
!> \param v_rspace_new_aux_fit ...
!> \param v_tau_rspace ...
!> \param v_tau_rspace_aux_fit ...
!> \param v_efield_rspace ...
!> \param v_sic_rspace ...
!> \param v_spin_ddapc_rest_r ...
!> \param v_sccs_rspace ...
!> \param becke ...
!> \param calculate_forces ...
!> \par History
!>      - refactoring 04.03.2011 [MI]
!>      - SCCS implementation (16.10.2013,MK)
!> \author
! **************************************************************************************************
   SUBROUTINE sum_up_and_integrate(qs_env, ks_matrix, rho, my_rho, &
                                   vppl_rspace, v_rspace_new, &
                                   v_rspace_new_aux_fit, v_tau_rspace, &
                                   v_tau_rspace_aux_fit, v_efield_rspace, &
                                   v_sic_rspace, v_spin_ddapc_rest_r, &
                                   v_sccs_rspace, becke, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: my_rho
      TYPE(pw_p_type), POINTER                           :: vppl_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: v_rspace_new, v_rspace_new_aux_fit, &
                                                            v_tau_rspace, v_tau_rspace_aux_fit
      TYPE(pw_p_type)                                    :: v_efield_rspace, v_sic_rspace, &
                                                            v_spin_ddapc_rest_r, v_sccs_rspace
      TYPE(becke_restraint_type), POINTER                :: becke
      LOGICAL, INTENT(in)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sum_up_and_integrate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: do_ppl, gapw, gapw_xc
      REAL(dp)                                           :: dvol
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ksmat, matrix_ks_aux_fit, &
                                                            matrix_ks_aux_fit_dft, rho_ao, &
                                                            rho_ao_aux
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: v_rspace
      TYPE(pw_p_type), POINTER                           :: vee
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho_aux_fit

      CALL timeset(routineN, handle)

      NULLIFY (auxbas_pw_pool, dft_control, pw_env, matrix_ks_aux_fit, &
               v_rspace%pw, rho_aux_fit, vee, rho_ao, rho_ao_kp, rho_ao_aux, &
               ksmat, matrix_ks_aux_fit_dft)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      pw_env=pw_env, &
                      matrix_ks_aux_fit=matrix_ks_aux_fit, &
                      matrix_ks_aux_fit_dft=matrix_ks_aux_fit_dft, &
                      v_hartree_rspace=v_rspace%pw, &
                      rho_aux_fit=rho_aux_fit, &
                      vee=vee)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
      CALL qs_rho_get(rho_aux_fit, rho_ao=rho_ao_aux)
      CALL pw_env_get(pw_env, poisson_env=poisson_env, auxbas_pw_pool=auxbas_pw_pool)
      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      do_ppl = dft_control%qs_control%do_ppl_method == do_ppl_grid

      nspins = dft_control%nspins

      ! sum up potentials and integrate
      IF (ASSOCIATED(v_rspace_new)) THEN
         DO ispin = 1, nspins
            IF (gapw_xc) THEN
               ! SIC not implemented (or at least not tested)
               IF(.NOT.(dft_control%sic_method_id == sic_none))CALL cp__a("qs_ks_utils.F",1227)
               !Only the xc potential, because it has to be integrated with the soft basis
               v_rspace_new(ispin)%pw%cr3d = &
                  v_rspace_new(ispin)%pw%pw_grid%dvol* &
                  v_rspace_new(ispin)%pw%cr3d

               ! add the xc  part due to v_rspace soft
               rho_ao => rho_ao_kp(ispin, :)
               ksmat => ks_matrix(ispin, :)
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat_kp=rho_ao, hmat_kp=ksmat, &
                                       qs_env=qs_env, &
                                       calculate_forces=calculate_forces, &
                                       gapw=gapw_xc)

               ! Now the Hartree potential to be integrated with the full basis
               v_rspace_new(ispin)%pw%cr3d = v_rspace%pw%cr3d
            ELSE
               ! Add v_hartree + v_xc = v_rspace_new
               v_rspace_new(ispin)%pw%cr3d = &
                  v_rspace_new(ispin)%pw%pw_grid%dvol* &
                  v_rspace_new(ispin)%pw%cr3d+v_rspace%pw%cr3d
            END IF ! gapw_xc
            IF (dft_control%qs_control%ddapc_explicit_potential) THEN
               IF (dft_control%qs_control%ddapc_restraint_is_spin) THEN
                  IF (ispin == 1) THEN
                     v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d &
                                                   +v_spin_ddapc_rest_r%pw%cr3d
                  ELSE
                     v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d &
                                                   -v_spin_ddapc_rest_r%pw%cr3d
                  ENDIF
               ELSE
                  v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d &
                                                +v_spin_ddapc_rest_r%pw%cr3d
               END IF
            END IF
            IF (dft_control%qs_control%becke_restraint) THEN
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d &
                                             +becke%becke_pot%pw%cr3d*dft_control%qs_control%becke_control%strength
            END IF
            ! The efield contribution
            IF (dft_control%apply_efield_field) THEN
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             v_efield_rspace%pw%cr3d
            END IF
            ! functional derivative of the Hartree energy wrt the density in the presence of dielectric
            ! (vhartree + v_eps); v_eps is nonzero only if the dielectric constant is defind as a function
            ! of the charge density
            IF (poisson_env%parameters%solver .EQ. pw_poisson_implicit) THEN
               dvol = poisson_env%implicit_env%v_eps%pw_grid%dvol
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             poisson_env%implicit_env%v_eps%cr3d*dvol
            END IF
            ! Add SCCS contribution
            IF (dft_control%do_sccs) THEN
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             v_sccs_rspace%pw%cr3d
            END IF
            ! External electrostatic potential
            IF (dft_control%apply_external_potential) THEN
               CALL qmmm_modify_hartree_pot(v_hartree=v_rspace_new(ispin), &
                                            v_qmmm=vee, scale=-1.0_dp)
            END IF
            IF (do_ppl) THEN
               IF(.NOT.(.NOT. gapw))CALL cp__a("qs_ks_utils.F",1292)
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             vppl_rspace%pw%cr3d*vppl_rspace%pw%pw_grid%dvol
            END IF
            ! the electrostatic sic contribution
            SELECT CASE (dft_control%sic_method_id)
            CASE (sic_none)
               !
            CASE (sic_mauri_us, sic_mauri_spz)
               IF (ispin == 1) THEN
                  v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d- &
                                                v_sic_rspace%pw%cr3d
               ELSE
                  v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                                v_sic_rspace%pw%cr3d
               ENDIF
            CASE (sic_ad)
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d-v_sic_rspace%pw%cr3d
            CASE (sic_eo)
               ! NOTHING TO BE DONE
            END SELECT
            rho_ao => my_rho(ispin, :)
            ksmat => ks_matrix(ispin, :)
            CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                    pmat_kp=rho_ao, hmat_kp=ksmat, &
                                    qs_env=qs_env, &
                                    calculate_forces=calculate_forces, &
                                    gapw=gapw)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new(ispin)%pw)
         END DO ! ispin

         SELECT CASE (dft_control%sic_method_id)
         CASE (sic_none)
         CASE (sic_mauri_us, sic_mauri_spz, sic_ad)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, v_sic_rspace%pw)
         END SELECT
         DEALLOCATE (v_rspace_new)

      ELSE
         ! not implemented (or at least not tested)
         IF(.NOT.(dft_control%sic_method_id == sic_none))CALL cp__a("qs_ks_utils.F",1332)
         IF(.NOT.(.NOT. dft_control%qs_control%ddapc_restraint_is_spin))CALL cp__a("qs_ks_utils.F",1333)
         DO ispin = 1, nspins
            ! the efield contribution
            IF (dft_control%apply_efield_field) THEN
               v_rspace%pw%cr3d = v_rspace%pw%cr3d+v_efield_rspace%pw%cr3d
            END IF
            ! extra contribution attributed to the dependecy of the dielectric constant to the charge density
            IF (poisson_env%parameters%solver .EQ. pw_poisson_implicit) THEN
               dvol = poisson_env%implicit_env%v_eps%pw_grid%dvol
               v_rspace%pw%cr3d = v_rspace%pw%cr3d+poisson_env%implicit_env%v_eps%cr3d*dvol
            END IF
            ! Add SCCS contribution
            IF (dft_control%do_sccs) THEN
               v_rspace%pw%cr3d = v_rspace%pw%cr3d+v_sccs_rspace%pw%cr3d
            END IF
            rho_ao => my_rho(ispin, :)
            ksmat => ks_matrix(ispin, :)
            CALL integrate_v_rspace(v_rspace=v_rspace, &
                                    pmat_kp=rho_ao, &
                                    hmat_kp=ksmat, &
                                    qs_env=qs_env, &
                                    calculate_forces=calculate_forces, &
                                    gapw=gapw)
         END DO
      END IF ! ASSOCIATED(v_rspace_new)

      IF (ASSOCIATED(v_tau_rspace)) THEN
         DO ispin = 1, nspins
            v_tau_rspace(ispin)%pw%cr3d = &
               v_tau_rspace(ispin)%pw%pw_grid%dvol* &
               v_tau_rspace(ispin)%pw%cr3d

            rho_ao => rho_ao_kp(ispin, :)
            ksmat => ks_matrix(ispin, :)
            CALL integrate_v_rspace(v_rspace=v_tau_rspace(ispin), &
                                    pmat_kp=rho_ao, hmat_kp=ksmat, &
                                    qs_env=qs_env, &
                                    calculate_forces=calculate_forces, compute_tau=.TRUE., &
                                    gapw=gapw)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, v_tau_rspace(ispin)%pw)
         END DO
         DEALLOCATE (v_tau_rspace)
      ENDIF

      ! Add contributions from ADMM if requested
      IF (dft_control%do_admm) THEN
         IF (ASSOCIATED(v_rspace_new_aux_fit)) THEN
            DO ispin = 1, nspins
               ! Calculate the xc potential
               v_rspace_new_aux_fit(ispin)%pw%cr3d = &
                  v_rspace_new_aux_fit(ispin)%pw%pw_grid%dvol* &
                  v_rspace_new_aux_fit(ispin)%pw%cr3d

               ! set matrix_ks_aux_fit_dft = matrix_ks_aux_fit(k_HF)
               CALL cp_dbcsr_copy(matrix_ks_aux_fit_dft(ispin)%matrix, matrix_ks_aux_fit(ispin)%matrix, &
                                  name="DFT exch. part of matrix_ks_aux_fit")

               ! Add potential to ks_matrix aux_fit

               CALL integrate_v_rspace(v_rspace=v_rspace_new_aux_fit(ispin), &
                                       pmat=rho_ao_aux(ispin), &
                                       hmat=matrix_ks_aux_fit(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=calculate_forces, &
                                       force_adm=.TRUE., &
                                       ispin=ispin, &
                                       gapw=gapw_xc, &
                                       basis_type="AUX_FIT")

               ! matrix_ks_aux_fit_dft(x_DFT)=matrix_ks_aux_fit_dft(old,k_HF)-matrix_ks_aux_fit(k_HF-x_DFT)
               CALL cp_dbcsr_add(matrix_ks_aux_fit_dft(ispin)%matrix, &
                                 matrix_ks_aux_fit(ispin)%matrix, 1.0_dp, -1.0_dp)

               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new_aux_fit(ispin)%pw)
            END DO
            DEALLOCATE (v_rspace_new_aux_fit)
         END IF
         ! Clean up v_tau_rspace_aux_fit, which is actually not needed
         IF (ASSOCIATED(v_tau_rspace_aux_fit)) THEN
            DO ispin = 1, nspins
               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_tau_rspace_aux_fit(ispin)%pw)
            END DO
            DEALLOCATE (v_tau_rspace_aux_fit)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE sum_up_and_integrate

!**************************************************************************
!> \brief Calculate the ZMP potential and energy as in Zhao, Morrison Parr
!> PRA 50i, 2138 (1994)
!> V_c^\lambda defined as int_rho-rho_0/r-r' or rho-rho_0 times a Lagrange
!> multiplier, plus Fermi-Amaldi potential that should give the V_xc in the
!> limit \lambda --> \infty
!>
!> \param qs_env ...
!> \param v_rspace_new ...
!> \param rho ...
!> \param exc ...
!> \author D. Varsano  [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE calculate_zmp_potential(qs_env, v_rspace_new, rho, exc)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: v_rspace_new
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(KIND=dp)                                      :: exc

      CHARACTER(*), PARAMETER :: routineN = 'calculate_zmp_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, my_val, nelectron, nspins, stat
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      LOGICAL                                            :: do_zmp_read, fermi_amaldi
      REAL(KIND=dp)                                      :: factor, lambda, total_rho
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_ext_r
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho_eff_gspace, v_xc_gspace, v_xc_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_ext_g, rho_g, rho_r
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(section_vals_type), POINTER                   :: ext_den_section, input

!, v_h_gspace, &

      CALL timeset(routineN, handle)
      NULLIFY (auxbas_pw_pool)
      NULLIFY (pw_env)
      NULLIFY (poisson_env)
      NULLIFY (v_rspace_new)
      NULLIFY (dft_control)
      NULLIFY (rho_r, rho_g, tot_rho_ext_r, rho_ext_g)
      CALL get_qs_env(qs_env=qs_env, &
                      pw_env=pw_env, &
                      ks_env=ks_env, &
                      rho=rho, &
                      input=input, &
                      nelectron_spin=nelectron_spin, &
                      dft_control=dft_control)
      CALL pw_env_get(pw_env=pw_env, &
                      auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)
      CALL qs_rho_get(rho, rho_r=rho_r, rho_g=rho_g)
      nspins = 1
      ALLOCATE (v_rspace_new(nspins), stat=stat)
      IF(.NOT.(stat == 0))CALL cp__a("qs_ks_utils.F",1482)
      CALL pw_pool_create_pw(pool=auxbas_pw_pool, pw=v_rspace_new(1)%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pool=auxbas_pw_pool, pw=v_xc_rspace%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL pw_zero(v_rspace_new(1)%pw)
      do_zmp_read = dft_control%apply_external_vxc
      IF (do_zmp_read) THEN
         CALL pw_copy(qs_env%external_vxc%pw, v_rspace_new(1)%pw)
         exc = 0.0_dp
         exc = accurate_sum(v_rspace_new(1)%pw%cr3d*rho_r(1)%pw%cr3d)* &
               v_rspace_new(1)%pw%pw_grid%dvol
      ELSE
         CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                                pw=rho_eff_gspace%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                                pw=v_xc_gspace%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         CALL pw_zero(rho_eff_gspace%pw)
         CALL pw_zero(v_xc_gspace%pw)
         CALL pw_zero(v_xc_rspace%pw)
         factor = pw_integrate_function(rho_g(1)%pw)
         CALL qs_rho_get(qs_env%rho_external, &
                         rho_g=rho_ext_g, &
                         tot_rho_r=tot_rho_ext_r)
         factor = tot_rho_ext_r(1)/factor

         CALL pw_axpy(rho_g(1)%pw, rho_eff_gspace%pw, alpha=factor)
         CALL pw_axpy(rho_ext_g(1)%pw, rho_eff_gspace%pw, alpha=-1.0_dp)
         total_rho = pw_integrate_function(rho_eff_gspace%pw, isign=1)
         ext_den_section => section_vals_get_subs_vals(input, "DFT%EXTERNAL_DENSITY")
         CALL section_vals_val_get(ext_den_section, "LAMBDA", r_val=lambda)
         CALL section_vals_val_get(ext_den_section, "ZMP_CONSTRAINT", i_val=my_val)
         CALL section_vals_val_get(ext_den_section, "FERMI_AMALDI", l_val=fermi_amaldi)

         CALL pw_scale(rho_eff_gspace%pw, a=lambda)
         nelectron = nelectron_spin(1)
         factor = -1.0_dp/nelectron
         CALL pw_axpy(rho_g(1)%pw, rho_eff_gspace%pw, alpha=factor)

         CALL pw_poisson_solve(poisson_env, rho_eff_gspace%pw, vhartree=v_xc_gspace%pw)
         CALL pw_transfer(v_xc_gspace%pw, v_rspace_new(1)%pw)
         CALL pw_copy(v_rspace_new(1)%pw, v_xc_rspace%pw)

         exc = 0.0_dp
         exc = accurate_sum(v_rspace_new(1)%pw%cr3d*rho_r(1)%pw%cr3d)* &
               v_rspace_new(1)%pw%pw_grid%dvol
         IF (v_rspace_new(1)%pw%pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
            CALL mp_sum(exc, v_rspace_new(1)%pw%pw_grid%para%group)
         END IF

!     IF (v_rspace_new(1)%pw%pw_grid%para%my_pos==0) &
!           WRITE(*,FMT="(T3,A,T61,F20.10)") "ZMP| Integral of (electron density)*(v_xc):    " , exc
!Note that this is not the xc energy but \int(\rho*v_xc)
!Vxc---> v_rspace_new
!Exc---> energy%exc
         CALL pw_pool_give_back_pw(auxbas_pw_pool, &
                                   rho_eff_gspace%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, &
                                   v_xc_gspace%pw)
      ENDIF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_xc_rspace%pw)

      CALL timestop(handle)

   END SUBROUTINE calculate_zmp_potential

END MODULE qs_ks_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates forces for LRIGPW method
!>        lri : local resolution of the identity
!> \par History
!>      created Dorothea Golze [03.2014]
!> \authors Dorothea Golze
! **************************************************************************************************
MODULE lri_forces

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE kinds,                           ONLY: dp
   USE lri_environment_types,           ONLY: &
        allocate_lri_force_components, deallocate_lri_force_components, lri_density_type, &
        lri_environment_type, lri_force_type, lri_int_type, lri_kind_type, lri_list_type, &
        lri_rhoab_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_forces'

   PUBLIC :: calculate_lri_forces

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief calculates the lri forces
!> \param lri_env ...
!> \param lri_density ...
!> \param qs_env ...
!> \param pmatrix density matrix
!> \param atomic_kind_set ...
!> \param use_virial ...
! **************************************************************************************************
   SUBROUTINE calculate_lri_forces(lri_env, lri_density, qs_env, pmatrix, &
                                   atomic_kind_set, use_virial)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      LOGICAL                                            :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_lri_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ikind, ispin, natom, &
                                                            nkind, nspin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: v_dadr, v_dfdr
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: lri_coef
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (atomic_kind, force, lri_coef, v_dadr, v_dfdr, virial)

      IF (ASSOCIATED(lri_env%soo_list)) THEN

         nkind = lri_env%lri_ints%nkind
         nspin = SIZE(pmatrix)
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, natom=natom)
         CALL get_qs_env(qs_env=qs_env, force=force, virial=virial)

         !calculate SUM_i integral(V*fbas_i)*davec/dR
         CALL calculate_v_dadr(lri_env, lri_density, pmatrix, atomic_kind_set, &
                               use_virial, virial)

         DO ispin = 1, nspin

            lri_coef => lri_density%lri_coefs(ispin)%lri_kinds

            DO ikind = 1, nkind
               atomic_kind => atomic_kind_set(ikind)
               CALL get_atomic_kind(atomic_kind=atomic_kind, natom=natom)
               DO iatom = 1, natom
                  v_dadr => lri_coef(ikind)%v_dadr(iatom, :)
                  v_dfdr => lri_coef(ikind)%v_dfdr(iatom, :)

                  force(ikind)%rho_lri_elec(:, iatom) = force(ikind)%rho_lri_elec(:, iatom) &
                                                        +v_dfdr(:)+v_dadr(:)

               END DO
            END DO
         ENDDO

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calculate_lri_forces

! **************************************************************************************************
!> \brief calculates second term of derivative with respect to R, i.e.
!>        SUM_i integral(V * fbas_i)*davec/dR
!> \param lri_env ...
!> \param lri_density ...
!> \param pmatrix ...
!> \param atomic_kind_set ...
!> \param use_virial ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE calculate_v_dadr(lri_env, lri_density, pmatrix, atomic_kind_set, &
                               use_virial, virial)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      LOGICAL, INTENT(IN)                                :: use_virial
      TYPE(virial_type), POINTER                         :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_v_dadr', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, i, iac, iatom, ikind, ilist, ispin, jatom, jkind, &
         jneighbor, k, natom, nba, nbb, nfa, nfb, nkind, nlist, nn, nneighbor, nspin
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: found, trans
      REAL(KIND=dp), DIMENSION(3)                        :: dcharge, dlambda, force_a, force_b, &
                                                            nsdssn, nsdsst, nsdt, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: st, v_dadra, v_dadrb, v_inta, v_intb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: davec, dssn, dsst, dtvec, pbij, sdssn, &
                                                            sdsst, sdt
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: ds
      TYPE(cp_dbcsr_type), POINTER                       :: pmat
      TYPE(lri_force_type), POINTER                      :: lri_force
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: lri_coef
      TYPE(lri_list_type), POINTER                       :: lri_rho
      TYPE(lri_rhoab_type), POINTER                      :: lrho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list

      CALL timeset(routineN, handle)
      NULLIFY (atom_of_kind, lri_coef, lri_force, lrii, lri_rho, lrho, &
               nl_iterator, pbij, pmat, soo_list, v_dadra, v_dadrb, v_inta, &
               v_intb)
      NULLIFY (davec, ds, dssn, dsst, dtvec, sdssn, sdsst, sdt, st)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         nkind = lri_env%lri_ints%nkind
         nspin = SIZE(pmatrix)

         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, natom=natom)
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  atom_of_kind=atom_of_kind)

         DO ispin = 1, nspin

            pmat => pmatrix(ispin)%matrix
            lri_coef => lri_density%lri_coefs(ispin)%lri_kinds
            lri_rho => lri_density%lri_rhos(ispin)%lri_list

            CALL neighbor_list_iterator_create(nl_iterator, soo_list)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, &
                                      jatom=jatom, nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                      r=rab)

               iac = ikind+nkind*(jkind-1)

               IF (.NOT. ASSOCIATED(lri_env%lri_ints%lri_atom(iac)%lri_node)) CYCLE

               lrii => lri_env%lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)
               lrho => lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(jneighbor)

               ! zero contribution for pairs aa or bb and periodc pairs aa' or bb'
               ! calculate forces for periodic pairs aa' and bb' only for virial
               IF (.NOT. lrii%calc_force_pair) CYCLE

               nfa = lrii%nfa
               nfb = lrii%nfb
               nba = lrii%nba
               nbb = lrii%nbb
               nn = nfa+nfb

               CALL allocate_lri_force_components(lri_density%lri_force, nfa, nfb)
               lri_force => lri_density%lri_force
               ds => lri_force%ds
               st => lri_force%st
               dsst => lri_force%dsst
               sdsst => lri_force%sdsst
               dssn => lri_force%dssn
               sdssn => lri_force%sdssn
               sdt => lri_force%sdt
               dtvec => lri_force%dtvec
               davec => lri_force%davec

               ! get the density matrix Pab
               NULLIFY (pbij)
               IF (iatom <= jatom) THEN
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=iatom, col=jatom, block=pbij, found=found)
                  trans = .FALSE.
               ELSE
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=jatom, col=iatom, block=pbij, found=found)
                  trans = .TRUE.
               END IF

               ! compute dtvec/dRa = SUM_ab Pab *d(a,b,x)/dRa
               DO k = 1, 3
                  IF (trans) THEN
                     dcharge(k) = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%dsoo(1:nba, 1:nbb, k))
                     DO i = 1, nfa
                        dtvec(i, k) = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%dabdaint(1:nba, 1:nbb, i, k))
                     END DO
                     DO i = 1, nfb
                        dtvec(nfa+i, k) = SUM(TRANSPOSE(pbij(1:nbb, 1:nba))*lrii%dabbint(1:nba, 1:nbb, i, k))
                     END DO
                  ELSE
                     dcharge(k) = SUM(pbij(1:nba, 1:nbb)*lrii%dsoo(1:nba, 1:nbb, k))
                     DO i = 1, nfa
                        dtvec(i, k) = SUM(pbij(1:nba, 1:nbb)*lrii%dabdaint(1:nba, 1:nbb, i, k))
                     END DO
                     DO i = 1, nfb
                        dtvec(nfa+i, k) = SUM(pbij(1:nba, 1:nbb)*lrii%dabbint(1:nba, 1:nbb, i, k))
                     END DO
                  END IF
               ENDDO

               ! construct derivative of S (overlap) matrix
               ds(1:nfa, 1:nfa, 1:3) = 0._dp
               ds(1:nfa, nfa+1:nn, 1:3) = lrii%dsab(1:nfa, 1:nfb, 1:3)
               DO k = 1, 3
                  ds(nfa+1:nn, 1:nfa, k) = TRANSPOSE(lrii%dsab(1:nfa, 1:nfb, k))
               ENDDO
               ds(nfa+1:nn, nfa+1:nn, 1:3) = 0._dp

               ! n*Sinv*dS*Sinv*t
               DO k = 1, 3
                  st(1:nn) = MATMUL(lrii%sinv(1:nn, 1:nn), lrho%tvec(1:nn))
                  dsst(1:nn, k) = MATMUL(ds(1:nn, 1:nn, k), st(1:nn))
                  sdsst(1:nn, k) = MATMUL(lrii%sinv(1:nn, 1:nn), dsst(1:nn, k))
                  nsdsst(k) = SUM(sdsst(1:nn, k)*lrii%n(1:nn))
               ENDDO

               ! n*Sinv*dS*Sinv*n
               DO k = 1, 3
                  dssn(1:nn, k) = MATMUL(ds(1:nn, 1:nn, k), lrii%sn(1:nn))
                  sdssn(1:nn, k) = MATMUL(lrii%sinv(1:nn, 1:nn), dssn(1:nn, k))
                  nsdssn(k) = SUM(sdssn(1:nn, k)*lrii%n(1:nn))
               ENDDO

               ! Sinv*dtvec/dR_a
               DO k = 1, 3
                  sdt(1:nn, k) = MATMUL(lrii%sinv(1:nn, 1:nn), dtvec(1:nn, k))
                  nsdt(k) = SUM(sdt(1:nn, k)*lrii%n(1:nn))
               ENDDO

               ! dlambda/dRa
               DO k = 1, 3
                  dlambda(k) = (dcharge(k)+nsdsst(k)-nsdt(k))/lrii%nsn &
                               +(lrho%charge-lrho%nst)*nsdssn(k)/(lrii%nsn*lrii%nsn)
               ENDDO

               ! davec/dRa, contains: davec_a/dRa (1..nfa) and davec_b/dRa (nfa+1..nn)
               DO k = 1, 3
                  davec(1:nn, k) = -sdsst(1:nn, k)+sdt(1:nn, k) &
                                   +dlambda(k)*lrii%sn(1:nn)-lrho%lambda*sdssn(1:nn, k)
               ENDDO

               atom_a = atom_of_kind(iatom)
               atom_b = atom_of_kind(jatom)

               v_dadra => lri_coef(ikind)%v_dadr(atom_a, :)
               v_dadrb => lri_coef(jkind)%v_dadr(atom_b, :)
               v_inta => lri_coef(ikind)%v_int(atom_a, :)
               v_intb => lri_coef(jkind)%v_int(atom_b, :)

               ! sum over atom pairs
               DO k = 1, 3
                  force_a(k) = 2.0_dp*SUM(davec(1:nfa, k)*v_inta(1:nfa)) &
                               +2.0_dp*SUM(davec(nfa+1:nn, k)*v_intb(1:nfb))
                  v_dadra(k) = v_dadra(k)+force_a(k)
                  !minus: using that davec/dRa = - davec/dRb
                  force_b(k) = -2.0_dp*SUM(davec(nfa+1:nn, k)*v_intb(1:nfb)) &
                               -2.0_dp*SUM(davec(1:nfa, k)*v_inta(1:nfa))
                  v_dadrb(k) = v_dadrb(k)+force_b(k)
               ENDDO

               ! contribution to virial
               IF (use_virial) THEN
                  !periodic self-pairs aa' contribute only with factor 0.5
                  IF (iatom == jatom) THEN
                     CALL virial_pair_force(virial%pv_virial, 0.5_dp, force_a, rab)
                  ELSE
                     CALL virial_pair_force(virial%pv_virial, 1.0_dp, force_a, rab)
                  ENDIF
               ENDIF

               CALL deallocate_lri_force_components(lri_density%lri_force)

            END DO

            CALL neighbor_list_iterator_release(nl_iterator)

         ENDDO

         DEALLOCATE (atom_of_kind)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calculate_v_dadr

END MODULE lri_forces
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that build the Kohn-Sham matrix (i.e calculate the coulomb
!>        and xc parts
!> \author Fawzi Mohamed
!> \par History
!>      - 05.2002 moved from qs_scf (see there the history) [fawzi]
!>      - JGH [30.08.02] multi-grid arrays independent from density and potential
!>      - 10.2002 introduced pools, uses updated rho as input,
!>                removed most temporary variables, renamed may vars,
!>                began conversion to LSD [fawzi]
!>      - 10.2004 moved calculate_w_matrix here [Joost VandeVondele]
!>                introduced energy derivative wrt MOs [Joost VandeVondele]
!>      - SCCS implementation (16.10.2013,MK)
! **************************************************************************************************
MODULE qs_ks_methods
   USE admm_dm_methods,                 ONLY: admm_dm_calc_rho_aux,&
                                              admm_dm_merge_ks_matrix
   USE admm_methods,                    ONLY: admm_mo_calc_rho_aux,&
                                              admm_mo_merge_derivs,&
                                              admm_mo_merge_ks_matrix
   USE admm_types,                      ONLY: admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: becke_restraint_type,&
                                              dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_distribution, cp_dbcsr_filter, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_name, &
        cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_ddapc,                        ONLY: qs_ks_ddapc
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_symm,&
                                              cp_fm_transpose,&
                                              cp_fm_upper_to_full
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_should_output
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dft_plus_u,                      ONLY: plus_u
   USE efield_utils,                    ONLY: efield_potential
   USE hartree_local_methods,           ONLY: Vh_1c_gg_integrals
   USE hfx_admm_utils,                  ONLY: hfx_admm_init,&
                                              hfx_ks_matrix
   USE input_constants,                 ONLY: do_ppl_grid
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kg_correction,                   ONLY: kg_ekin_subset
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE lri_environment_types,           ONLY: lri_density_type,&
                                              lri_environment_type,&
                                              lri_kind_type
   USE lri_forces,                      ONLY: calculate_lri_forces
   USE lri_ks_methods,                  ONLY: calculate_lri_ks_matrix
   USE mathlib,                         ONLY: abnormal_value
   USE message_passing,                 ONLY: mp_sum
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integral_ab,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_implicit,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qmmm_image_charge,               ONLY: add_image_pot_to_hartree_pot,&
                                              calculate_image_pot,&
                                              integrate_potential_devga_rspace
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_core_energies,                ONLY: calculate_ptrace
   USE qs_dftb_matrices,                ONLY: build_dftb_ks_matrix
   USE qs_efield_berry,                 ONLY: qs_efield_berry_phase
   USE qs_efield_local,                 ONLY: qs_efield_local_operator
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_integrate_potential,          ONLY: integrate_ppl_rspace,&
                                              integrate_rho_nlcc,&
                                              integrate_v_core_rspace,&
                                              integrate_v_rspace_one_center
   USE qs_ks_apply_restraints,          ONLY: qs_ks_becke_restraint,&
                                              qs_ks_mulliken_restraint,&
                                              qs_ks_s2_restraint
   USE qs_ks_atom,                      ONLY: update_ks_atom
   USE qs_ks_qmmm_methods,              ONLY: qmmm_calculate_energy,&
                                              qmmm_modify_hartree_pot
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_ks_utils,                     ONLY: calc_v_sic_rspace,&
                                              calculate_zmp_potential,&
                                              compute_matrix_vxc,&
                                              low_spin_roks,&
                                              print_densities,&
                                              print_detailed_energy,&
                                              sic_explicit_orbitals,&
                                              sum_up_and_integrate
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho0_ggrid,                   ONLY: integrate_vhg0_rspace
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_sccs,                         ONLY: sccs
   USE qs_vxc,                          ONLY: qs_vxc_create
   USE qs_vxc_atom,                     ONLY: calculate_vxc_atom
   USE rtp_admm_methods,                ONLY: rtp_admm_calc_rho_aux,&
                                              rtp_admm_merge_ks_matrix
   USE scptb_ks_matrix,                 ONLY: build_scptb_ks_matrix
   USE se_fock_matrix,                  ONLY: build_se_fock_matrix
   USE surface_dipole,                  ONLY: calc_dipsurf_potential
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTERFACE calculate_w_matrix
      MODULE PROCEDURE calculate_w_matrix_1, &
         calculate_w_matrix_roks

   END INTERFACE

   LOGICAL, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ks_methods'

   PUBLIC :: calc_rho_tot_gspace, qs_ks_update_qs_env, qs_ks_build_kohn_sham_matrix, &
             calculate_w_matrix, calculate_w_matrix_ot, qs_ks_allocate_basics

   TYPE qs_potential_type
      TYPE(pw_p_type), POINTER :: v_efield_rspace, v_hartree_gspace, &
                                  vppl_rspace, v_sic_rspace, v_spin_ddapc_rest_r
      TYPE(pw_p_type), DIMENSION(:), POINTER   :: v_rspace_new, &
                                                  v_rspace_new_aux_fit, &
                                                  v_tau_rspace, &
                                                  v_tau_rspace_aux_fit
   END TYPE qs_potential_type

CONTAINS

! **************************************************************************************************
!> \brief routine where the real calculations are made: the
!>      KS matrix is calculated
!> \param qs_env the qs_env to update
!> \param calculate_forces if true calculate the quantities needed
!>        to calculate the forces. Defaults to false.
!> \param just_energy if true updates the energies but not the
!>        ks matrix. Defaults to false
!> \param print_active ...
!> \param ext_ks_matrix ...
!> \par History
!>      06.2002 moved from qs_scf to qs_ks_methods, use of ks_env
!>              new did_change scheme [fawzi]
!>      10.2002 introduced pools, uses updated rho as input, LSD [fawzi]
!>      10.2004 build_kohn_sham matrix now also computes the derivatives
!>              of the total energy wrt to the MO coefs, if instructed to
!>              do so. This appears useful for orbital dependent functionals
!>              where the KS matrix alone (however this might be defined)
!>               does not contain the info to construct this derivative.
!> \author Matthias Krack
!> \note
!>      make rho, energy and qs_charges optional, defaulting
!>      to qs_env components?
! **************************************************************************************************
   SUBROUTINE qs_ks_build_kohn_sham_matrix(qs_env, calculate_forces, just_energy, &
                                           print_active, ext_ks_matrix)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy
      LOGICAL, INTENT(IN), OPTIONAL                      :: print_active
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ext_ks_matrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_ks_build_kohn_sham_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, img, ispin, nimages, ns, nspins
      LOGICAL :: do_adiabatic_rescaling, do_ddapc, do_hfx, do_ppl, gapw, gapw_xc, &
         hfx_treat_lsd_in_core, just_energy_xc, my_print, use_virial
      REAL(KIND=dp)                                      :: ecore_ppl, edisp, ee_ener, ekin_mol, &
                                                            mulliken_order_p
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_stress
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(becke_restraint_type), POINTER                :: becke
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ksmat, matrix_vxc, mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix, matrix_h, matrix_s, my_rho, &
                                                            rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type) :: rho_tot_gspace, v_efield_rspace, v_hartree_gspace, v_hartree_rspace, &
         v_minus_veps, v_sccs_rspace, v_sic_rspace, v_spin_ddapc_rest_r
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r, v_rspace_new, &
                                                            v_rspace_new_aux_fit, v_tau_rspace, &
                                                            v_tau_rspace_aux_fit
      TYPE(pw_p_type), POINTER                           :: rho0_s_rs, rho_core, rho_nlcc, vee, &
                                                            vppl_rspace
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho, rho_struct, rho_xc
      TYPE(section_vals_type), POINTER                   :: adiabatic_rescaling_section, &
                                                            hfx_sections, input, scf_section, &
                                                            xc_section
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (admm_env, becke, cell, dft_control, logger, &
               mo_derivs, my_rho, rho_struct, para_env, pw_env, virial, &
               vppl_rspace, v_hartree_rspace%pw, adiabatic_rescaling_section, &
               hfx_sections, input, scf_section, xc_section, matrix_h, matrix_s, &
               auxbas_pw_pool, poisson_env, pw_pools, v_rspace_new, &
               v_rspace_new_aux_fit, v_tau_rspace, v_tau_rspace_aux_fit, &
               matrix_vxc, vee, rho_nlcc, ks_env, &
               ks_matrix, rho, energy, rho_xc, rho_r, rho_ao, rho_core)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_ks_methods.F",262)

      logger => cp_get_default_logger()
      my_print = .TRUE.
      IF (PRESENT(print_active)) my_print = print_active

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      matrix_h_kp=matrix_h, &
                      matrix_s_kp=matrix_s, &
                      matrix_ks_kp=ks_matrix, &
                      matrix_vxc=matrix_vxc, &
                      pw_env=pw_env, &
                      cell=cell, &
                      para_env=para_env, &
                      input=input, &
                      virial=virial, &
                      v_hartree_rspace=v_hartree_rspace%pw, &
                      vee=vee, &
                      rho_nlcc=rho_nlcc, &
                      rho=rho, &
                      rho_core=rho_core, &
                      rho_xc=rho_xc, &
                      energy=energy)

      CALL qs_rho_get(rho, rho_r=rho_r, rho_ao_kp=rho_ao)

      IF (PRESENT(ext_ks_matrix)) THEN
         ! remap pointer to allow for non-kpoint external ks matrix
         ! ext_ks_matrix is used in linear response code
         ns = SIZE(ext_ks_matrix)
         ks_matrix(1:ns, 1:1) => ext_ks_matrix(1:ns)
      END IF

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%HF")
      CALL section_vals_get(hfx_sections, explicit=do_hfx)
      IF (do_hfx) THEN
         CALL section_vals_val_get(hfx_sections, "TREAT_LSD_IN_CORE", l_val=hfx_treat_lsd_in_core, &
                                   i_rep_section=1)
      END IF
      adiabatic_rescaling_section => section_vals_get_subs_vals(input, "DFT%XC%ADIABATIC_RESCALING")
      CALL section_vals_get(adiabatic_rescaling_section, explicit=do_adiabatic_rescaling)
      just_energy_xc = just_energy
      IF (do_adiabatic_rescaling) THEN
         !! If we perform adiabatic rescaling, the xc potential has to be scaled by the xc- and
         !! HFX-energy. Thus, let us first calculate the energy
         just_energy_xc = .TRUE.
      END IF

      nimages = dft_control%nimages
      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(matrix_h)))CALL cp__a("qs_ks_methods.F",316)
      IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("qs_ks_methods.F",317)
      IF(.NOT.(ASSOCIATED(rho)))CALL cp__a("qs_ks_methods.F",318)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_ks_methods.F",319)
      IF(.NOT.(SIZE(ks_matrix, 1) > 0))CALL cp__a("qs_ks_methods.F",320)

      ! Setup the possible usage of DDAPC charges
      do_ddapc = dft_control%qs_control%ddapc_restraint .OR. &
                 qs_env%cp_ddapc_ewald%do_decoupling .OR. &
                 qs_env%cp_ddapc_ewald%do_qmmm_periodic_decpl .OR. &
                 qs_env%cp_ddapc_ewald%do_solvation

      ! Check for GAPW method : additional terms for local densities
      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      IF (gapw_xc .AND. gapw) THEN
         CALL cp__b("qs_ks_methods.F",332," GAPW and GAPW_XC are not compatible")
      END IF

      do_ppl = dft_control%qs_control%do_ppl_method == do_ppl_grid
      IF (do_ppl) THEN
         IF(.NOT.(.NOT. gapw))CALL cp__a("qs_ks_methods.F",337)
         CALL get_qs_env(qs_env=qs_env, vppl=vppl_rspace)
      END IF

      IF (gapw_xc) THEN
         IF(.NOT.(ASSOCIATED(rho_xc)))CALL cp__a("qs_ks_methods.F",342)
      END IF

      ! gets the tmp grids
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, poisson_env=poisson_env)

      IF (gapw .AND. (poisson_env%parameters%solver .EQ. pw_poisson_implicit)) THEN
         CALL cp__b("qs_ks_methods.F",350,"The implicit Poisson solver cannot be used in conjunction with GAPW.")
      END IF

      ! ***  Prepare densities for gapw ***
      IF (gapw .OR. gapw_xc) THEN
         CALL prepare_gapw_den(qs_env, do_rho0=(.NOT. gapw_xc))
      ENDIF

      ! Calculate the Hartree potential
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             rho_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      scf_section => section_vals_get_subs_vals(input, "DFT%SCF")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, scf_section, &
                                           "PRINT%DETAILED_ENERGY"), &
                cp_p_file) .AND. &
          (.NOT. gapw) .AND. (.NOT. gapw_xc) .AND. &
          (.NOT. (poisson_env%parameters%solver .EQ. pw_poisson_implicit))) THEN
         CALL pw_zero(rho_tot_gspace%pw)
         CALL calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho, skip_nuclear_density=.TRUE.)
         CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, energy%e_hartree, &
                               v_hartree_gspace%pw)
         CALL pw_zero(rho_tot_gspace%pw)
         CALL pw_zero(v_hartree_gspace%pw)
      END IF

      ! Get the total density in g-space [ions + electrons]
      CALL calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho)

      IF (my_print) THEN
         CALL print_densities(qs_env, rho)
      END IF

      IF (dft_control%do_sccs) THEN
         ! Self-consistent continuum solvation (SCCS) model
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                v_sccs_rspace%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)

         IF (poisson_env%parameters%solver .EQ. pw_poisson_implicit) THEN
            CALL cp__b("qs_ks_methods.F",397,"The implicit Poisson solver cannot be used together with SCCS.")
         END IF

         IF (use_virial .AND. calculate_forces) THEN
            CALL sccs(qs_env, rho_tot_gspace%pw, v_hartree_gspace%pw, v_sccs_rspace%pw, &
                      h_stress=h_stress)
            virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env%num_pe, dp)
         ELSE
            CALL sccs(qs_env, rho_tot_gspace%pw, v_hartree_gspace%pw, v_sccs_rspace%pw)
         END IF
      ELSE
         ! Getting the Hartree energy and Hartree potential.  Also getting the stress tensor
         ! from the Hartree term if needed.  No nuclear force information here
         IF (use_virial .AND. calculate_forces) THEN
            h_stress(:, :) = 0.0_dp
            CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, energy%hartree, &
                                  v_hartree_gspace%pw, h_stress=h_stress, &
                                  rho_core=rho_core)
            virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env%num_pe, dp)
         ELSE
            CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, energy%hartree, &
                                  v_hartree_gspace%pw, rho_core=rho_core)
         END IF
      END IF

      ! In case decouple periodic images and/or apply restraints to charges
      IF (do_ddapc) THEN
         CALL qs_ks_ddapc(qs_env, auxbas_pw_pool, rho_tot_gspace, v_hartree_gspace, &
                          v_spin_ddapc_rest_r, energy, calculate_forces, ks_matrix, &
                          just_energy)
      ELSE
         dft_control%qs_control%ddapc_explicit_potential = .FALSE.
         dft_control%qs_control%ddapc_restraint_is_spin = .FALSE.
         IF (.NOT. just_energy) THEN
            CALL pw_transfer(v_hartree_gspace%pw, v_hartree_rspace%pw)
            CALL pw_scale(v_hartree_rspace%pw, v_hartree_rspace%pw%pw_grid%dvol)
         END IF
      END IF
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)

      IF (dft_control%apply_efield_field) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                v_efield_rspace%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL efield_potential(qs_env, v_efield_rspace)
         CALL pw_scale(v_efield_rspace%pw, v_efield_rspace%pw%pw_grid%dvol)
      END IF

      IF (dft_control%correct_surf_dip) THEN
         CALL calc_dipsurf_potential(qs_env, energy)
         energy%hartree = energy%hartree+energy%surf_dipole
      END IF

      ! SIC
      CALL calc_v_sic_rspace(v_sic_rspace, energy, qs_env, dft_control, rho, poisson_env, &
                             just_energy, calculate_forces, auxbas_pw_pool)

      IF (gapw) CALL Vh_1c_gg_integrals(qs_env, energy%hartree_1c)

      ! Check if becke potential is needed to constrain charges
      CALL qs_ks_becke_restraint(qs_env, auxbas_pw_pool, calculate_forces, matrix_s, becke)

      ! Adds the External Potential if requested
      IF (dft_control%apply_external_potential) THEN
         ! Compute the energy due to the external potential
         ee_ener = 0.0_dp
         DO ispin = 1, nspins
            ee_ener = ee_ener+pw_integral_ab(rho_r(ispin)%pw, vee%pw)
         END DO
         IF (.NOT. just_energy) THEN
            IF (gapw) THEN
               CALL get_qs_env(qs_env=qs_env, &
                               rho0_s_rs=rho0_s_rs)
               IF(.NOT.(ASSOCIATED(rho0_s_rs)))CALL cp__a("qs_ks_methods.F",471)
               ee_ener = ee_ener+pw_integral_ab(rho0_s_rs%pw, vee%pw)
            END IF
         END IF
         ! the sign accounts for the charge of the electrons
         energy%ee = -ee_ener
      END IF

      ! Adds the QM/MM potential
      IF (qs_env%qmmm) THEN
         CALL qmmm_calculate_energy(qs_env=qs_env, &
                                    rho=rho_r, &
                                    v_qmmm=qs_env%ks_qmmm_env%v_qmmm_rspace, &
                                    qmmm_energy=energy%qmmm_el)
         IF (qs_env%qmmm_env_qm%image_charge) THEN
            CALL calculate_image_pot(v_hartree_rspace=v_hartree_rspace, &
                                     rho_hartree_gspace=rho_tot_gspace, &
                                     energy=energy, &
                                     qmmm_env=qs_env%qmmm_env_qm, &
                                     qs_env=qs_env)
            IF (.NOT. just_energy) THEN
               CALL add_image_pot_to_hartree_pot(v_hartree=v_hartree_rspace, &
                                                 v_metal=qs_env%ks_qmmm_env%v_metal_rspace, &
                                                 qs_env=qs_env)
               IF (calculate_forces) THEN
                  CALL integrate_potential_devga_rspace( &
                     potential=v_hartree_rspace, coeff=qs_env%image_coeff, &
                     forces=qs_env%qmmm_env_qm%image_charge_pot%image_forcesMM, &
                     qmmm_env=qs_env%qmmm_env_qm, qs_env=qs_env)
               ENDIF
            ENDIF
            CALL pw_release(qs_env%ks_qmmm_env%v_metal_rspace%pw)
         END IF
         IF (.NOT. just_energy) THEN
            CALL qmmm_modify_hartree_pot(v_hartree=v_hartree_rspace, &
                                         v_qmmm=qs_env%ks_qmmm_env%v_qmmm_rspace, scale=1.0_dp)
         END IF
      END IF
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tot_gspace%pw)

      ! calculate the density matrix for the fitted mo_coeffs
      IF (dft_control%do_admm) THEN
         CALL hfx_admm_init(qs_env)
         IF (dft_control%do_admm_mo) THEN
            IF (qs_env%run_rtp) THEN
               CALL rtp_admm_calc_rho_aux(qs_env)
            ELSE
               CALL admm_mo_calc_rho_aux(qs_env)
            END IF
         ELSEIF (dft_control%do_admm_dm) THEN
            CALL admm_dm_calc_rho_aux(ks_env)
         ENDIF
      ENDIF

      ! only activate stress calculation if
      IF (use_virial .AND. calculate_forces) virial%pv_calculate = .TRUE.

      ! *** calculate the xc potential on the pw density ***
      ! *** associates v_rspace_new if the xc potential needs to be computed.
      ! If we do wavefunction fitting, we need the vxc_potential in the auxiliary basis set
      IF (dft_control%do_admm) THEN
         CALL get_qs_env(qs_env, admm_env=admm_env)
         xc_section => admm_env%xc_section_aux
         IF (gapw_xc) THEN
            CALL get_qs_env(qs_env=qs_env, rho_xc=rho_struct)
         ELSE
            CALL get_qs_env(qs_env=qs_env, rho_aux_fit=rho_struct)
         END IF

         ! here we ignore a possible vdW section in admm_env%xc_section_aux
         CALL qs_vxc_create(ks_env=ks_env, rho_struct=rho_struct, xc_section=xc_section, &
                            vxc_rho=v_rspace_new_aux_fit, vxc_tau=v_tau_rspace_aux_fit, exc=energy%exc_aux_fit, &
                            just_energy=just_energy_xc)

         NULLIFY (rho_struct)

         IF (use_virial .AND. calculate_forces) THEN
            virial%pv_virial = virial%pv_virial-virial%pv_xc
            ! ** virial%pv_xc will be zeroed in the xc routines
         END IF
         xc_section => admm_env%xc_section_primary
      ELSE
         xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      END IF

      IF (gapw_xc) THEN
         CALL get_qs_env(qs_env=qs_env, rho_xc=rho_struct)
      ELSE
         CALL get_qs_env(qs_env=qs_env, rho=rho_struct)
      END IF

      ! zmp
      IF (dft_control%apply_external_density .OR. dft_control%apply_external_vxc) THEN
         energy%exc = 0.
         CALL calculate_zmp_potential(qs_env, v_rspace_new, rho, exc=energy%exc)
      ELSE
         CALL qs_vxc_create(ks_env=ks_env, rho_struct=rho_struct, xc_section=xc_section, &
                            vxc_rho=v_rspace_new, vxc_tau=v_tau_rspace, exc=energy%exc, &
                            edisp=edisp, dispersion_env=qs_env%dispersion_env, &
                            just_energy=just_energy_xc)
         IF (edisp /= 0.0_dp) energy%dispersion = edisp
         IF (qs_env%requires_matrix_vxc) THEN
            CALL compute_matrix_vxc(qs_env=qs_env, v_rspace=v_rspace_new, matrix_vxc=matrix_vxc)
         ENDIF

         IF (gapw .OR. gapw_xc) THEN
            CALL calculate_vxc_atom(qs_env, just_energy_xc)
         ENDIF
      ENDIF

      NULLIFY (rho_struct)
      IF (use_virial .AND. calculate_forces) THEN
         virial%pv_virial = virial%pv_virial-virial%pv_xc
      ENDIF

      ! *** Add Hartree-Fock contribution if required ***
      IF (do_hfx) THEN
         CALL hfx_ks_matrix(qs_env, ks_matrix, rho, energy, calculate_forces, &
                            just_energy, v_rspace_new, v_tau_rspace)
!!    Adiabatic rescaling  only if do_hfx; right?????
      END IF !do_hfx

      IF (do_ppl .AND. calculate_forces) THEN
         IF(.NOT.(.NOT. gapw))CALL cp__a("qs_ks_methods.F",594)
         DO ispin = 1, nspins
            CALL integrate_ppl_rspace(rho_r(ispin), qs_env)
         END DO
      END IF

      IF (ASSOCIATED(rho_nlcc) .AND. calculate_forces) THEN
         IF(.NOT.(.NOT. (gapw .OR. gapw_xc)))CALL cp__a("qs_ks_methods.F",601)
         DO ispin = 1, nspins
            CALL integrate_rho_nlcc(v_rspace_new(ispin), qs_env)
            IF (dft_control%do_admm) CALL integrate_rho_nlcc(v_rspace_new_aux_fit(ispin), qs_env)
         END DO
      ENDIF

      ! calculate KG correction
      IF (dft_control%qs_control%do_kg .AND. just_energy) THEN

         IF(.NOT.(nimages == 1))CALL cp__a("qs_ks_methods.F",611)
         ksmat => ks_matrix(:, 1)
         CALL kg_ekin_subset(qs_env, ksmat, gapw, gapw_xc, ekin_mol, &
                             calculate_forces)

         ! substract kg corr from the total energy
         energy%exc = energy%exc-ekin_mol

      END IF

      ! ***  Single atom contributions ***
      IF (.NOT. just_energy) THEN
         IF (calculate_forces) THEN
            ! Getting nuclear force contribution from the core charge density
            IF ((poisson_env%parameters%solver .EQ. pw_poisson_implicit) .AND. &
                (poisson_env%parameters%dielectric_params%dielec_core_correction)) THEN
               CALL pw_pool_create_pw(auxbas_pw_pool, v_minus_veps%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_copy(v_hartree_rspace%pw, v_minus_veps%pw)
               CALL pw_axpy(poisson_env%implicit_env%v_eps, v_minus_veps%pw, -v_hartree_rspace%pw%pw_grid%dvol)
               CALL integrate_v_core_rspace(v_minus_veps, qs_env)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_minus_veps%pw)
            ELSE
               CALL integrate_v_core_rspace(v_hartree_rspace, qs_env)
            END IF
         END IF

         IF (.NOT. do_hfx) THEN
            ! Initialize the Kohn-Sham matrix with the core Hamiltonian matrix
            ! (sets ks sparsity equal to matrix_h sparsity)
            DO ispin = 1, nspins
               DO img = 1, nimages
                  CALL cp_dbcsr_copy(ks_matrix(ispin, img)%matrix, matrix_h(1, img)%matrix, &
                                     name=cp_dbcsr_name(ks_matrix(ispin, img)%matrix))
               END DO
            END DO
         END IF

         ! sum up potentials and integrate
         ! Pointing my_rho to the density matrix rho_ao
         my_rho => rho_ao
         CALL sum_up_and_integrate(qs_env, ks_matrix, rho, my_rho, vppl_rspace, &
                                   v_rspace_new, v_rspace_new_aux_fit, v_tau_rspace, v_tau_rspace_aux_fit, &
                                   v_efield_rspace, v_sic_rspace, v_spin_ddapc_rest_r, v_sccs_rspace, becke, &
                                   calculate_forces)

         IF (dft_control%qs_control%do_kg) THEN
            IF(.NOT.(nimages == 1))CALL cp__a("qs_ks_methods.F",657)
            ksmat => ks_matrix(:, 1)
            CALL kg_ekin_subset(qs_env, ksmat, gapw, gapw_xc, ekin_mol, &
                                calculate_forces)
            ! substract kg corr from the total energy
            energy%exc = energy%exc-ekin_mol
         END IF

      END IF ! .NOT. just energy

      IF (dft_control%qs_control%ddapc_explicit_potential) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_spin_ddapc_rest_r%pw)
      END IF

      IF (dft_control%apply_efield_field) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_efield_rspace%pw)
      END IF

      IF (calculate_forces .AND. dft_control%qs_control%becke_restraint) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, becke%becke_pot%pw)
      END IF

      IF (dft_control%do_sccs) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_sccs_rspace%pw)
      END IF

      IF (gapw) THEN
         IF (dft_control%apply_external_potential) THEN
            ! Integrals of the Hartree potential with g0_soft
            CALL qmmm_modify_hartree_pot(v_hartree=v_hartree_rspace, &
                                         v_qmmm=vee, scale=-1.0_dp)
         ENDIF
         CALL integrate_vhg0_rspace(qs_env, v_hartree_rspace, calculate_forces)
      END IF

      IF (gapw .OR. gapw_xc) THEN
         ! Single atom contributions in the KS matrix ***
         CALL update_ks_atom(qs_env, ks_matrix, rho_ao, calculate_forces)
      END IF

      !Calculation of Mulliken restraint, if requested
      CALL qs_ks_mulliken_restraint(energy, dft_control, just_energy, para_env, &
                                    ks_matrix, matrix_s, rho, mulliken_order_p)

      ! Add DFT+U contribution, if requested
      IF (dft_control%dft_plus_u) THEN
         IF(.NOT.(nimages == 1))CALL cp__a("qs_ks_methods.F",703)
         IF (just_energy) THEN
            CALL plus_u(qs_env=qs_env)
         ELSE
            ksmat => ks_matrix(:, 1)
            CALL plus_u(qs_env=qs_env, matrix_h=ksmat)
         END IF
      ELSE
         energy%dft_plus_u = 0.0_dp
      END IF

      ! At this point the ks matrix should be up to date, filter it if requested
      DO ispin = 1, nspins
         DO img = 1, nimages
            CALL cp_dbcsr_filter(ks_matrix(ispin, img)%matrix, &
                                 dft_control%qs_control%eps_filter_matrix)
         END DO
      ENDDO

      !** merge the auxiliary KS matrix and the primary one
      IF (dft_control%do_admm_mo) THEN
         IF (qs_env%run_rtp) THEN
            CALL rtp_admm_merge_ks_matrix(qs_env)
         ELSE
            CALL admm_mo_merge_ks_matrix(qs_env)
         ENDIF
      ELSEIF (dft_control%do_admm_dm) THEN
         CALL admm_dm_merge_ks_matrix(ks_env)
      ENDIF

      ! External field (nonperiodic case)
      CALL qs_efield_local_operator(qs_env, just_energy, calculate_forces)

      ! Right now we can compute the orbital derivative here, as it depends currently only on the available
      ! Kohn-Sham matrix. This might change in the future, in which case more pieces might need to be assembled
      ! from this routine, notice that this part of the calculation in not linear scaling
      ! right now this operation is only non-trivial because of occupation numbers and the restricted keyword
      IF (qs_env%requires_mo_derivs .AND. .NOT. just_energy .AND. .NOT. qs_env%run_rtp) THEN
         CALL get_qs_env(qs_env, mo_derivs=mo_derivs)
         IF(.NOT.(nimages == 1))CALL cp__a("qs_ks_methods.F",742)
         ksmat => ks_matrix(:, 1)
         CALL calc_mo_derivatives(qs_env, ksmat, mo_derivs)
      ENDIF

      ! deal with low spin roks
      CALL low_spin_roks(energy, qs_env, dft_control, just_energy, &
                         calculate_forces, auxbas_pw_pool)

      ! deal with sic on explicit orbitals
      CALL sic_explicit_orbitals(energy, qs_env, dft_control, poisson_env, just_energy, &
                                 calculate_forces, auxbas_pw_pool)

      ! Periodic external field
      CALL qs_efield_berry_phase(qs_env, just_energy, calculate_forces)

      ! adds s2_restraint energy and orbital derivatives
      CALL qs_ks_s2_restraint(dft_control, qs_env, matrix_s, &
                              energy, calculate_forces, just_energy)

      IF (do_ppl) THEN
         ! update core energy for grid based local pseudopotential
         ecore_ppl = 0._dp
         DO ispin = 1, nspins
            ecore_ppl = ecore_ppl+ &
                        SUM(vppl_rspace%pw%cr3d*rho_r(ispin)%pw%cr3d)*vppl_rspace%pw%pw_grid%dvol
         END DO
         CALL mp_sum(ecore_ppl, para_env%group)
         energy%core = energy%core+ecore_ppl
      END IF

      ! Sum all energy terms to obtain the total energy
      energy%total = energy%core_overlap+energy%core_self+energy%core+energy%hartree+ &
                     energy%hartree_1c+energy%exc+energy%exc1+energy%ex+ &
                     energy%dispersion+energy%qmmm_el+energy%mulliken+ &
                     SUM(energy%ddapc_restraint)+energy%s2_restraint+ &
                     energy%becke+energy%dft_plus_u+energy%kTS+ &
                     energy%efield+energy%efield_core+energy%ee+ &
                     energy%ee_core+energy%exc_aux_fit+energy%image_charge+ &
                     energy%sccs_pol+energy%sccs_mpc

      IF (abnormal_value(energy%total)) &
         CALL cp__b("qs_ks_methods.F",784,"KS energy is an abnormal value (NaN/Inf).")

      ! Print detailed energy
      IF (my_print) THEN
         CALL print_detailed_energy(qs_env, dft_control, input, energy, mulliken_order_p)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_ks_build_kohn_sham_matrix

! **************************************************************************************************
!> \brief calculation of the LRIGPW KS matrix
!> \param qs_env the qs_env to update
!> \param calculate_forces ...
!> \param just_energy if true updates the energies but not the
!>        ks matrix. Defaults to false
!> \param print_active ...
!> \par History
!>      01.2014 created
!> \author Dorothea Golze
!> \note  might be integrate in qs_ks_build_kohn_sham_matrix later
! **************************************************************************************************
   SUBROUTINE lri_build_ks_matrix(qs_env, calculate_forces, just_energy, print_active)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy
      LOGICAL, INTENT(in), OPTIONAL                      :: print_active

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_build_ks_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: just_energy_xc, my_print, use_virial
      REAL(KIND=dp)                                      :: edisp
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_stress
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_h, matrix_p, mo_derivs
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: lri_v_int
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho_tot_gspace, v_hartree_gspace, &
                                                            v_hartree_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: v_rspace_new, v_tau_rspace
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, xc_section
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, input, mo_derivs, para_env, pw_env, &
               xc_section, ks_env, ks_matrix, rho, energy, auxbas_pw_pool, matrix_h, &
               matrix_p, poisson_env, pw_pools, v_rspace_new, v_tau_rspace, &
               atomic_kind_set, lri_env, lri_density, lri_v_int, v_hartree_rspace%pw, &
               virial)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_ks_methods.F",845)

      my_print = .TRUE.
      IF (PRESENT(print_active)) my_print = print_active

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control, &
                      matrix_h=matrix_h, &
                      lri_env=lri_env, &
                      lri_density=lri_density, &
                      pw_env=pw_env, &
                      para_env=para_env, &
                      input=input, &
                      virial=virial, &
                      v_hartree_rspace=v_hartree_rspace%pw, &
                      matrix_ks=ks_matrix, &
                      rho=rho, &
                      energy=energy)

      nspins = dft_control%nspins
      IF(.NOT.(SIZE(ks_matrix) > 0))CALL cp__a("qs_ks_methods.F",867)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_ks_methods.F",868)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      ! gets the tmp grids
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, poisson_env=poisson_env)

      ! *** calculate the Hartree potential on the LRI pw density ***
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             rho_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      ! get the total density in g-space [ions + electrons]
      CALL calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho)

      IF (my_print) THEN
         CALL print_densities(qs_env, rho)
      END IF

      ! solve Poisson equation to get Hartree energy and potential and the stress
      ! tensor from the Hartee term if needed
      IF (use_virial .AND. calculate_forces) THEN
         h_stress(:, :) = 0.0_dp
         CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, energy%hartree, &
                               v_hartree_gspace%pw, h_stress=h_stress)
         virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env%num_pe, dp)
      ELSE
         CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, energy%hartree, &
                               v_hartree_gspace%pw)
      ENDIF

      IF (.NOT. just_energy) THEN
         CALL pw_transfer(v_hartree_gspace%pw, v_hartree_rspace%pw)
         CALL pw_scale(v_hartree_rspace%pw, v_hartree_rspace%pw%pw_grid%dvol)
      END IF
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tot_gspace%pw)

      ! *** core contribution
      IF (.NOT. just_energy) THEN
         ! Initialize the Kohn-Sham matrix with the core Hamiltonian matrix
         DO ispin = 1, nspins
            CALL cp_dbcsr_copy(ks_matrix(ispin)%matrix, matrix_h(1)%matrix, &
                               name=cp_dbcsr_name(ks_matrix(ispin)%matrix))
         END DO
      END IF

      ! *** exchange-correlation
      IF (use_virial .AND. calculate_forces) virial%pv_calculate = .TRUE.
      xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      just_energy_xc = just_energy

      CALL qs_vxc_create(ks_env=ks_env, rho_struct=rho, xc_section=xc_section, &
                         vxc_rho=v_rspace_new, vxc_tau=v_tau_rspace, exc=energy%exc, &
                         edisp=edisp, dispersion_env=qs_env%dispersion_env, &
                         just_energy=just_energy_xc)
      IF (edisp /= 0.0_dp) energy%dispersion = edisp

      IF (use_virial .AND. calculate_forces) THEN
         virial%pv_virial = virial%pv_virial-virial%pv_xc
      ENDIF

      ! *** sum v_hartree + v_xc
      IF (ASSOCIATED(v_rspace_new)) THEN
         DO ispin = 1, nspins
            v_rspace_new(ispin)%pw%cr3d = &
               v_rspace_new(ispin)%pw%pw_grid%dvol*v_rspace_new(ispin)%pw%cr3d+ &
               v_hartree_rspace%pw%cr3d
         ENDDO
      ENDIF
      ! NO TAU YET

      ! *** do integration and build matrix
      IF (.NOT. just_energy) THEN

         DO ispin = 1, nspins
            lri_v_int => lri_density%lri_coefs(ispin)%lri_kinds
            IF (.NOT. ASSOCIATED(v_rspace_new)) THEN
               CALL integrate_v_rspace_one_center(v_hartree_rspace, qs_env, &
                                                  lri_v_int, calculate_forces)
            ELSE
               CALL integrate_v_rspace_one_center(v_rspace_new(ispin), qs_env, &
                                                  lri_v_int, calculate_forces)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new(ispin)%pw)
            ENDIF
            CALL calculate_lri_ks_matrix(lri_env, lri_v_int, ks_matrix(ispin)%matrix, &
                                         atomic_kind_set)
         ENDDO
         IF (ASSOCIATED(v_rspace_new)) THEN
            DEALLOCATE (v_rspace_new)
         ENDIF

         ! *** LRIGPW forces
         IF (calculate_forces) THEN
            ! Getting nuclear force contribution from the core charge density
            CALL integrate_v_core_rspace(v_hartree_rspace, qs_env)
            ! Getting electronic force contribution from fitted LRI charge density
            CALL qs_rho_get(rho, rho_ao=matrix_p)
            CALL calculate_lri_forces(lri_env, lri_density, qs_env, matrix_p, atomic_kind_set, &
                                      use_virial)
         END IF

      ENDIF

      ! At this point the ks matrix should be up to date, filter it if requested
      DO ispin = 1, nspins
         CALL cp_dbcsr_filter(ks_matrix(ispin)%matrix, &
                              dft_control%qs_control%eps_filter_matrix)
      ENDDO

      ! Right now we can compute the orbital derivative here, as it depends currently only on the available
      ! Kohn-Sham matrix. This might change in the future, in which case more pieces might need to be assembled
      ! from this routine, notice that this part of the calculation in not linear scaling
      ! right now this operation is only non-trivial because of occupation numbers and the restricted keyword
      IF (qs_env%requires_mo_derivs .AND. .NOT. just_energy .AND. .NOT. qs_env%run_rtp) THEN
         CALL get_qs_env(qs_env, mo_derivs=mo_derivs)
         CALL calc_mo_derivatives(qs_env, ks_matrix, mo_derivs)
      ENDIF

      energy%total = energy%core_overlap+energy%core_self+ &
                     energy%core+energy%hartree+ &
                     energy%hartree_1c+ &
                     energy%exc+energy%exc1+energy%ex+energy%dispersion+ &
                     energy%qmmm_el+energy%mulliken+ &
                     SUM(energy%ddapc_restraint)+energy%s2_restraint+ &
                     energy%becke+energy%dft_plus_u+energy%kTS+ &
                     energy%efield+energy%efield_core+ &
                     energy%ee+energy%ee_core+ &
                     energy%exc_aux_fit+ &
                     energy%image_charge

      CALL timestop(handle)

   END SUBROUTINE lri_build_ks_matrix

! **************************************************************************************************
!> \brief ...
!> \param rho_tot_gspace ...
!> \param qs_env ...
!> \param rho ...
!> \param skip_nuclear_density ...
! **************************************************************************************************
   SUBROUTINE calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho, skip_nuclear_density)
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_tot_gspace
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho
      LOGICAL, INTENT(IN), OPTIONAL                      :: skip_nuclear_density

      CHARACTER(*), PARAMETER :: routineN = 'calc_rho_tot_gspace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      LOGICAL                                            :: my_skip
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(pw_p_type), POINTER                           :: rho0_s_gs, rho_core
      TYPE(qs_charges_type), POINTER                     :: qs_charges

      NULLIFY (rho_core, rho0_s_gs, dft_control)
      my_skip = .FALSE.
      IF (PRESENT(skip_nuclear_density)) my_skip = skip_nuclear_density
      CALL get_qs_env(qs_env=qs_env, &
                      rho_core=rho_core, &
                      rho0_s_gs=rho0_s_gs, &
                      qs_charges=qs_charges, &
                      dft_control=dft_control)

      CALL qs_rho_get(rho, rho_g=rho_g)

      IF (.NOT. my_skip) THEN
         IF (dft_control%qs_control%gapw) THEN
            IF (dft_control%qs_control%gapw_control%nopaw_as_gpw) THEN
               IF(.NOT.(ASSOCIATED(rho0_s_gs)))CALL cp__a("qs_ks_methods.F",1046)
               CALL pw_copy(rho0_s_gs%pw, rho_tot_gspace%pw)
               CALL pw_axpy(rho_core%pw, rho_tot_gspace%pw)
               CALL pw_axpy(rho_g(1)%pw, rho_tot_gspace%pw)
            ELSE
               IF(.NOT.(ASSOCIATED(rho0_s_gs)))CALL cp__a("qs_ks_methods.F",1051)
               CALL pw_copy(rho0_s_gs%pw, rho_tot_gspace%pw)
               CALL pw_axpy(rho_g(1)%pw, rho_tot_gspace%pw)
            END IF
         ELSE
            CALL pw_copy(rho_core%pw, rho_tot_gspace%pw)
            CALL pw_axpy(rho_g(1)%pw, rho_tot_gspace%pw)
         END IF
      ELSE
         CALL pw_axpy(rho_g(1)%pw, rho_tot_gspace%pw)
      END IF
      DO ispin = 2, dft_control%nspins
         CALL pw_axpy(rho_g(ispin)%pw, rho_tot_gspace%pw)
      END DO
      IF (.NOT. my_skip) &
         qs_charges%total_rho_gspace = pw_integrate_function(rho_tot_gspace%pw, isign=-1)

   END SUBROUTINE calc_rho_tot_gspace

! **************************************************************************************************
!> \brief compute MO derivatives
!> \param qs_env the qs_env to update
!> \param ks_matrix ...
!> \param mo_derivs ...
!> \par History
!>      01.2014 created, transferred from qs_ks_build_kohn_sham_matrix in
!>      separate subroutine
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE calc_mo_derivatives(qs_env, ks_matrix, mo_derivs)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, mo_derivs

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_mo_derivatives', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      LOGICAL                                            :: uniform_occupation
      REAL(KIND=dp), DIMENSION(:), POINTER               :: occupation_numbers
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_aux_fit
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_b
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_derivs_aux_fit, mo_derivs_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_aux_fit
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array, mos_aux_fit

      NULLIFY (dft_control, mo_array, mo_coeff, mo_coeff_aux_fit, mo_coeff_b, &
               mo_derivs_aux_fit, mo_derivs_tmp, mos_aux_fit, &
               occupation_numbers, matrix_ks_aux_fit)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      mos=mo_array)

      IF (dft_control%do_admm_mo) THEN !fm->dbcsr
         NULLIFY (mo_derivs_tmp) !fm->dbcsr
         ALLOCATE (mo_derivs_tmp(SIZE(mo_derivs)))
         DO ispin = 1, SIZE(mo_derivs) !fm->dbcsr
            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff=mo_coeff) !fm->dbcsr
            NULLIFY (mo_derivs_tmp(ispin)%matrix)
            CALL cp_fm_create(mo_derivs_tmp(ispin)%matrix, mo_coeff%matrix_struct) !fm->dbcsr
         ENDDO !fm->dbcsr
      ENDIF !fm->dbcsr

      DO ispin = 1, SIZE(mo_derivs)

         CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff=mo_coeff, &
                         mo_coeff_b=mo_coeff_b, occupation_numbers=occupation_numbers)
         CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, ks_matrix(ispin)%matrix, mo_coeff_b, &
                                0.0_dp, mo_derivs(ispin)%matrix)

         IF (dft_control%do_admm_mo) THEN
            CALL get_qs_env(qs_env, &
                            mos_aux_fit=mos_aux_fit, &
                            mo_derivs_aux_fit=mo_derivs_aux_fit, &
                            matrix_ks_aux_fit=matrix_ks_aux_fit)
            CALL get_mo_set(mo_set=mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit)

            CALL copy_dbcsr_to_fm(mo_derivs(ispin)%matrix, &
                                  mo_derivs_tmp(ispin)%matrix) !fm->dbcsr
            CALL admm_mo_merge_derivs(ispin, qs_env%admm_env, mo_array(ispin)%mo_set, &
                                      mo_coeff, mo_coeff_aux_fit, mo_derivs_tmp, mo_derivs_aux_fit, &
                                      matrix_ks_aux_fit)
            CALL copy_fm_to_dbcsr(mo_derivs_tmp(ispin)%matrix, mo_derivs(ispin)%matrix) !fm->dbcsr
         END IF

         IF (dft_control%restricted) THEN
            ! only the first mo_set are actual variables, but we still need both
            IF(.NOT.(ispin == 1))CALL cp__a("qs_ks_methods.F",1139)
            IF(.NOT.(SIZE(mo_array) == 2))CALL cp__a("qs_ks_methods.F",1140)
            ! use a temporary array with the same size as the first spin for the second spin

            ! uniform_occupation is needed for this case, otherwise we can no
            ! reconstruct things in ot, since we irreversibly sum
            CALL get_mo_set(mo_set=mo_array(1)%mo_set, uniform_occupation=uniform_occupation)
            IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_methods.F",1146)
            CALL get_mo_set(mo_set=mo_array(2)%mo_set, &
                            uniform_occupation=uniform_occupation)
            IF(.NOT.(uniform_occupation))CALL cp__a("qs_ks_methods.F",1149)

            CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff_b=mo_coeff_b)
            CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, ks_matrix(2)%matrix, mo_coeff_b, &
                                   1.0_dp, mo_derivs(1)%matrix, match_matrix_sizes=.TRUE.)
         ENDIF

      ENDDO

      IF (dft_control%do_admm_mo) THEN !fm->dbcsr
         DO ispin = 1, SIZE(mo_derivs) !fm->dbcsr
            CALL cp_fm_release(mo_derivs_tmp(ispin)%matrix) !fm->dbcsr
         ENDDO !fm->dbcsr
         DEALLOCATE (mo_derivs_tmp) !fm->dbcsr
      ENDIF !fm->dbcsr

   END SUBROUTINE calc_mo_derivatives

! **************************************************************************************************
!> \brief updates the Kohn Sham matrix of the given qs_env (facility method)
!> \param qs_env the qs_env to update
!> \param calculate_forces if true calculate the quantities needed
!>        to calculate the forces. Defaults to false.
!> \param just_energy if true updates the energies but not the
!>        ks matrix. Defaults to false
!> \param print_active ...
!> \par History
!>      4.2002 created [fawzi]
!>      8.2014 kpoints [JGH]
!>     10.2014 refractored [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_ks_update_qs_env(qs_env, calculate_forces, just_energy, &
                                  print_active)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: calculate_forces, just_energy, &
                                                            print_active

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_ks_update_qs_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      LOGICAL                                            :: c_forces, do_rebuild, energy_only, &
                                                            forces_up_to_date, potential_changed, &
                                                            rho_changed, s_mstruct_changed
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (ks_env)
      unit_nr = cp_logger_get_default_io_unit()

      c_forces = .FALSE.
      energy_only = .FALSE.
      IF (PRESENT(just_energy)) energy_only = just_energy
      IF (PRESENT(calculate_forces)) c_forces = calculate_forces

      IF (c_forces) THEN
         CALL timeset(routineN//'_forces', handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_ks_methods.F",1210)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      rho_changed=rho_changed, &
                      s_mstruct_changed=s_mstruct_changed, &
                      potential_changed=potential_changed, &
                      forces_up_to_date=forces_up_to_date)

      do_rebuild = .FALSE.
      do_rebuild = do_rebuild .OR. rho_changed
      do_rebuild = do_rebuild .OR. s_mstruct_changed
      do_rebuild = do_rebuild .OR. potential_changed
      do_rebuild = do_rebuild .OR. (c_forces .AND. .NOT. forces_up_to_date)

      IF (do_rebuild) THEN
         CALL evaluate_core_matrix_traces(qs_env)

         ! the ks matrix will be rebuilt so this is fine now
         CALL set_ks_env(ks_env, potential_changed=.FALSE.)

         CALL rebuild_ks_matrix(qs_env, &
                                calculate_forces=c_forces, &
                                just_energy=energy_only, &
                                print_active=print_active)

         IF (.NOT. energy_only) THEN

            CALL set_ks_env(ks_env, &
                            rho_changed=.FALSE., &
                            s_mstruct_changed=.FALSE., &
                            forces_up_to_date=forces_up_to_date .OR. c_forces)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_ks_update_qs_env

! **************************************************************************************************
!> \brief Calculates the traces of the core matrices and the density matrix.
!> \param qs_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE evaluate_core_matrix_traces(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'evaluate_core_matrix_traces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrixkp_h, matrixkp_t, rho_ao_kp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)
      NULLIFY (energy, rho, dft_control, rho_ao_kp, matrixkp_t, matrixkp_h)

      CALL get_qs_env(qs_env, &
                      rho=rho, &
                      energy=energy, &
                      dft_control=dft_control, &
                      kinetic_kp=matrixkp_t, &
                      matrix_h_kp=matrixkp_h)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)

      CALL calculate_ptrace(matrixkp_h, rho_ao_kp, energy%core, dft_control%nspins)

      ! kinetic energy
      IF (ASSOCIATED(matrixkp_t)) &
         CALL calculate_ptrace(matrixkp_t, rho_ao_kp, energy%kinetic, dft_control%nspins)

      CALL timestop(handle)
   END SUBROUTINE evaluate_core_matrix_traces

! **************************************************************************************************
!> \brief Constructs a new Khon-Sham matrix
!> \param qs_env ...
!> \param calculate_forces ...
!> \param just_energy ...
!> \param print_active ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE rebuild_ks_matrix(qs_env, calculate_forces, just_energy, print_active)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces, just_energy
      LOGICAL, INTENT(IN), OPTIONAL                      :: print_active

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rebuild_ks_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)
      NULLIFY (dft_control)

      CALL get_qs_env(qs_env, dft_control=dft_control)

      IF (dft_control%qs_control%semi_empirical) THEN
         CALL build_se_fock_matrix(qs_env, &
                                   calculate_forces=calculate_forces, &
                                   just_energy=just_energy)

      ELSEIF (dft_control%qs_control%dftb) THEN
         CALL build_dftb_ks_matrix(qs_env, &
                                   calculate_forces=calculate_forces, &
                                   just_energy=just_energy)

      ELSEIF (dft_control%qs_control%scptb) THEN
         CALL build_scptb_ks_matrix(qs_env, &
                                    calculate_forces=calculate_forces, &
                                    just_energy=just_energy)

      ELSEIF (dft_control%qs_control%lrigpw) THEN
         CALL lri_build_ks_matrix(qs_env, &
                                  calculate_forces=calculate_forces, &
                                  just_energy=just_energy, &
                                  print_active=print_active)
      ELSE
         CALL qs_ks_build_kohn_sham_matrix(qs_env, &
                                           calculate_forces=calculate_forces, &
                                           just_energy=just_energy, &
                                           print_active=print_active)
      END IF

      CALL timestop(handle)

   END SUBROUTINE rebuild_ks_matrix

! **************************************************************************************************
!> \brief Calculate the W matrix from the MO eigenvectors, MO eigenvalues,
!>       and the MO occupation numbers. Only works if they are eigenstates
!> \param mo_set type containing the full matrix of the MO and the eigenvalues
!> \param w_matrix sparse matrix
!>        error
!> \par History
!>         Creation (03.03.03,MK)
!>         Modification that computes it as a full block, several times (e.g. 20)
!>               faster at the cost of some additional memory
!> \author MK
! **************************************************************************************************
   SUBROUTINE calculate_w_matrix_1(mo_set, w_matrix)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: w_matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_w_matrix_1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, imo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigocc
      TYPE(cp_fm_type), POINTER                          :: weighted_vectors

      CALL timeset(routineN, handle)
      NULLIFY (weighted_vectors)

      CALL cp_dbcsr_set(w_matrix, 0.0_dp)
      CALL cp_fm_create(weighted_vectors, mo_set%mo_coeff%matrix_struct, "weighted_vectors")
      CALL cp_fm_to_fm(mo_set%mo_coeff, weighted_vectors)

      ! scale every column with the occupation
      ALLOCATE (eigocc(mo_set%homo))

      DO imo = 1, mo_set%homo
         eigocc(imo) = mo_set%eigenvalues(imo)*mo_set%occupation_numbers(imo)
      ENDDO
      CALL cp_fm_column_scale(weighted_vectors, eigocc)
      DEALLOCATE (eigocc)

      CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=w_matrix, &
                                 matrix_v=mo_set%mo_coeff, &
                                 matrix_g=weighted_vectors, &
                                 ncol=mo_set%homo)

      CALL cp_fm_release(weighted_vectors)

      CALL timestop(handle)

   END SUBROUTINE calculate_w_matrix_1

! **************************************************************************************************
!> \brief Calculate the W matrix from the MO coefs, MO derivs
!>        could overwrite the mo_derivs for increased memory efficiency
!> \param mo_set type containing the full matrix of the MO coefs
!>        mo_deriv:
!> \param mo_deriv ...
!> \param w_matrix sparse matrix
!> \param s_matrix sparse matrix for the overlap
!>        error
!> \par History
!>         Creation (JV)
!> \author MK
! **************************************************************************************************
   SUBROUTINE calculate_w_matrix_ot(mo_set, mo_deriv, w_matrix, s_matrix)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: mo_deriv, w_matrix, s_matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_w_matrix_ot', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: check_gradient = .FALSE., &
                                                            do_symm = .FALSE.

      INTEGER                                            :: handle, ncol_block, ncol_global, &
                                                            nrow_block, nrow_global
      REAL(KIND=dp), DIMENSION(:), POINTER               :: occupation_numbers, scaling_factor
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: gradient, h_block, h_block_t, &
                                                            weighted_vectors

      CALL timeset(routineN, handle)
      NULLIFY (weighted_vectors, h_block, fm_struct_tmp)

      CALL cp_fm_get_info(matrix=mo_set%mo_coeff, &
                          ncol_global=ncol_global, &
                          nrow_global=nrow_global, &
                          nrow_block=nrow_block, &
                          ncol_block=ncol_block)

      CALL cp_fm_create(weighted_vectors, mo_set%mo_coeff%matrix_struct, "weighted_vectors")
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol_global, ncol_global=ncol_global, &
                               para_env=mo_set%mo_coeff%matrix_struct%para_env, &
                               context=mo_set%mo_coeff%matrix_struct%context)
      CALL cp_fm_create(h_block, fm_struct_tmp, name="h block")
      IF (do_symm) CALL cp_fm_create(h_block_t, fm_struct_tmp, name="h block t")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL get_mo_set(mo_set=mo_set, occupation_numbers=occupation_numbers)
      ALLOCATE (scaling_factor(SIZE(occupation_numbers)))
      scaling_factor = 2.0_dp*occupation_numbers
      CALL copy_dbcsr_to_fm(mo_deriv, weighted_vectors)
      CALL cp_fm_column_scale(weighted_vectors, scaling_factor)
      DEALLOCATE (scaling_factor)

      ! the convention seems to require the half here, the factor of two is presumably taken care of
      ! internally in qs_core_hamiltonian
      CALL cp_gemm('T', 'N', ncol_global, ncol_global, nrow_global, 0.5_dp, &
                   mo_set%mo_coeff, weighted_vectors, 0.0_dp, h_block)

      IF (do_symm) THEN
         ! at the minimum things are anyway symmetric, but numerically it might not be the case
         ! needs some investigation to find out if using this is better
         CALL cp_fm_transpose(h_block, h_block_t)
         CALL cp_fm_scale_and_add(0.5_dp, h_block, 0.5_dp, h_block_t)
      ENDIF

      ! this could overwrite the mo_derivs to save the weighted_vectors
      CALL cp_gemm('N', 'N', nrow_global, ncol_global, ncol_global, 1.0_dp, &
                   mo_set%mo_coeff, h_block, 0.0_dp, weighted_vectors)

      CALL cp_dbcsr_set(w_matrix, 0.0_dp)
      CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=w_matrix, &
                                 matrix_v=mo_set%mo_coeff, &
                                 matrix_g=weighted_vectors, &
                                 ncol=mo_set%homo)

      IF (check_gradient) THEN
         CALL cp_fm_create(gradient, mo_set%mo_coeff%matrix_struct, "gradient")
         CALL cp_dbcsr_sm_fm_multiply(s_matrix, weighted_vectors, &
                                      gradient, ncol_global)

         ALLOCATE (scaling_factor(SIZE(occupation_numbers)))
         scaling_factor = 2.0_dp*occupation_numbers
         CALL copy_dbcsr_to_fm(mo_deriv, weighted_vectors)
         CALL cp_fm_column_scale(weighted_vectors, scaling_factor)
         DEALLOCATE (scaling_factor)

         WRITE (*, *) " maxabs difference ", MAXVAL(ABS(weighted_vectors%local_data-2.0_dp*gradient%local_data))
         CALL cp_fm_release(gradient)
      ENDIF

      IF (do_symm) CALL cp_fm_release(h_block_t)
      CALL cp_fm_release(weighted_vectors)
      CALL cp_fm_release(h_block)

      CALL timestop(handle)

   END SUBROUTINE calculate_w_matrix_ot

! **************************************************************************************************
!> \brief Calculate the energy-weighted density matrix W if ROKS is active.
!>        The W matrix is returned in matrix_w.
!> \param mo_set ...
!> \param matrix_ks ...
!> \param matrix_p ...
!> \param matrix_w ...
!> \author 04.05.06,MK
! **************************************************************************************************
   SUBROUTINE calculate_w_matrix_roks(mo_set, matrix_ks, matrix_p, matrix_w)
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_ks, matrix_p, matrix_w

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_w_matrix_roks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nao
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: c, ks, p, work
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      NULLIFY (c)
      NULLIFY (context)
      NULLIFY (fm_struct)
      NULLIFY (ks)
      NULLIFY (p)
      NULLIFY (para_env)
      NULLIFY (work)

      CALL get_mo_set(mo_set=mo_set, mo_coeff=c)
      CALL cp_fm_get_info(c, context=context, nrow_global=nao, para_env=para_env)
      CALL cp_fm_struct_create(fm_struct, context=context, nrow_global=nao, &
                               ncol_global=nao, para_env=para_env)
      CALL cp_fm_create(ks, fm_struct, name="Kohn-Sham matrix")
      CALL cp_fm_create(p, fm_struct, name="Density matrix")
      CALL cp_fm_create(work, fm_struct, name="Work matrix")
      CALL cp_fm_struct_release(fm_struct)
      CALL copy_dbcsr_to_fm(matrix_ks, ks)
      CALL copy_dbcsr_to_fm(matrix_p, p)
      CALL cp_fm_upper_to_full(p, work)
      CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, ks, p, 0.0_dp, work)
      CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, p, work, 0.0_dp, ks)
      CALL cp_dbcsr_set(matrix_w, 0.0_dp)
      CALL copy_fm_to_dbcsr(ks, matrix_w, keep_sparsity=.TRUE.)

      CALL cp_fm_release(work)
      CALL cp_fm_release(p)
      CALL cp_fm_release(ks)

      CALL timestop(handle)

   END SUBROUTINE calculate_w_matrix_roks

! **************************************************************************************************
!> \brief Allocate ks_matrix and ks_env if necessary
!> \param qs_env ...
!> \par History
!>    refactoring 04.03.2011 [MI]
!> \author
! **************************************************************************************************

   SUBROUTINE qs_ks_allocate_basics(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_ks_allocate_basics', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: headline
      INTEGER                                            :: ic, ispin, nimg
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_aux_fit, &
                                                            matrix_ks_aux_fit_dft, &
                                                            matrix_ks_aux_fit_hfx, matrix_s_aux_fit
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s_kp, matrixkp_ks
      TYPE(cp_dbcsr_type), POINTER                       :: refmatrix
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_aux_fit, sab_orb
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (dft_control, ks_env, matrix_s_kp, matrix_s_aux_fit, sab_orb, sab_aux_fit, matrixkp_ks)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      matrix_s_kp=matrix_s_kp, &
                      ks_env=ks_env, &
                      sab_orb=sab_orb, &
                      sab_aux_fit=sab_aux_fit, &
                      matrix_ks_kp=matrixkp_ks)

      IF (.NOT. ASSOCIATED(matrixkp_ks)) THEN
         nimg = dft_control%nimages
         CALL cp_dbcsr_allocate_matrix_set(matrixkp_ks, dft_control%nspins, nimg)
         refmatrix => matrix_s_kp(1, 1)%matrix
         DO ispin = 1, dft_control%nspins
            DO ic = 1, nimg
               IF (dft_control%nspins > 1) THEN
                  IF (ispin == 1) THEN
                     headline = "KOHN-SHAM MATRIX FOR ALPHA SPIN"
                  ELSE
                     headline = "KOHN-SHAM MATRIX FOR BETA SPIN"
                  END IF
               ELSE
                  headline = "KOHN-SHAM MATRIX"
               END IF
               ALLOCATE (matrixkp_ks(ispin, ic)%matrix)
               CALL cp_dbcsr_init(matrixkp_ks(ispin, ic)%matrix)
               CALL cp_dbcsr_create(matrix=matrixkp_ks(ispin, ic)%matrix, &
                                    name=TRIM(headline), &
                                    dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                    row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                    col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                    nze=0)
               CALL cp_dbcsr_alloc_block_from_nbl(matrixkp_ks(ispin, ic)%matrix, sab_orb)
               CALL cp_dbcsr_set(matrixkp_ks(ispin, ic)%matrix, 0.0_dp)
            ENDDO
         ENDDO
         CALL set_ks_env(ks_env, matrix_ks_kp=matrixkp_ks)
      END IF

      ! Allocate matrix_ks_aux_fit if requested and put it in the QS environment
      ! Also matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx for ADMMS
      IF (dft_control%do_admm) THEN
         NULLIFY (matrix_ks_aux_fit, matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx)
         CALL get_qs_env(qs_env, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit, &
                         matrix_ks_aux_fit_dft=matrix_ks_aux_fit_dft, &
                         matrix_ks_aux_fit_hfx=matrix_ks_aux_fit_hfx, &
                         matrix_s_aux_fit=matrix_s_aux_fit)
         IF (.NOT. ASSOCIATED(matrix_ks_aux_fit)) THEN
            refmatrix => matrix_s_aux_fit(1)%matrix
            CALL cp_dbcsr_allocate_matrix_set(matrix_ks_aux_fit, dft_control%nspins)
            DO ispin = 1, dft_control%nspins
               ALLOCATE (matrix_ks_aux_fit(ispin)%matrix)
               CALL cp_dbcsr_init(matrix_ks_aux_fit(ispin)%matrix)
               CALL cp_dbcsr_create(matrix=matrix_ks_aux_fit(ispin)%matrix, &
                                    name="KOHN-SHAM_MATRIX for ADMM", &
                                    dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                    row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                    col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                    nze=0)
               CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks_aux_fit(ispin)%matrix, sab_aux_fit)
               CALL cp_dbcsr_set(matrix_ks_aux_fit(ispin)%matrix, 0.0_dp)
            ENDDO
            CALL set_ks_env(ks_env, matrix_ks_aux_fit=matrix_ks_aux_fit)
         END IF
         ! same allocation procedure for matrix_ks_aux_fit_dft
         IF (.NOT. ASSOCIATED(matrix_ks_aux_fit_dft)) THEN
            refmatrix => matrix_s_aux_fit(1)%matrix
            CALL cp_dbcsr_allocate_matrix_set(matrix_ks_aux_fit_dft, dft_control%nspins)
            DO ispin = 1, dft_control%nspins
               ALLOCATE (matrix_ks_aux_fit_dft(ispin)%matrix)
               CALL cp_dbcsr_init(matrix_ks_aux_fit_dft(ispin)%matrix)
               CALL cp_dbcsr_create(matrix=matrix_ks_aux_fit_dft(ispin)%matrix, &
                                    name="AUX. KOHN-SHAM MATRIX FOR ADMM ONLY DFT EXCHANGE", &
                                    dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                    row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                    col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                    nze=0)
               CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks_aux_fit_dft(ispin)%matrix, sab_aux_fit)
               CALL cp_dbcsr_set(matrix_ks_aux_fit_dft(ispin)%matrix, 0.0_dp)
            ENDDO
            CALL set_ks_env(ks_env, matrix_ks_aux_fit_dft=matrix_ks_aux_fit_dft)
         END IF
         ! same allocation procedure for matrix_ks_aux_fit_hfx
         IF (.NOT. ASSOCIATED(matrix_ks_aux_fit_hfx)) THEN
            refmatrix => matrix_s_aux_fit(1)%matrix
            CALL cp_dbcsr_allocate_matrix_set(matrix_ks_aux_fit_hfx, dft_control%nspins)
            DO ispin = 1, dft_control%nspins
               ALLOCATE (matrix_ks_aux_fit_hfx(ispin)%matrix)
               CALL cp_dbcsr_init(matrix_ks_aux_fit_hfx(ispin)%matrix)
               CALL cp_dbcsr_create(matrix=matrix_ks_aux_fit_hfx(ispin)%matrix, &
                                    name="AUX. KOHN-SHAM MATRIX FOR ADMM ONLY HF EXCHANGE", &
                                    dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                    row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                    col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                    nze=0)
               CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks_aux_fit_hfx(ispin)%matrix, sab_aux_fit)
               CALL cp_dbcsr_set(matrix_ks_aux_fit_hfx(ispin)%matrix, 0.0_dp)
            ENDDO
            CALL set_ks_env(ks_env, matrix_ks_aux_fit_hfx=matrix_ks_aux_fit_hfx)
         END IF

      END IF

   END SUBROUTINE qs_ks_allocate_basics

END MODULE qs_ks_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines needed for kpoint calculation
!> \par History
!>       2014.07 created [JGH]
!>       2014.11 unified k-point and gamma-point code [Ole Schuett]
!> \author JGH
! **************************************************************************************************
MODULE kpoint_methods
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_create, cp_dbcsr_deallocate_matrix, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_antisymmetric, &
        dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_create_fm,&
                                              fm_pool_give_back_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_copy_general,&
                                              cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_env,                     ONLY: cp_cart_create,&
                                              cp_para_env_create,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_cart_type,&
                                              cp_para_env_type
   USE fermi_utils,                     ONLY: fermikp,&
                                              fermikp2
   USE input_constants,                 ONLY: smear_fermi_dirac
   USE k290,                            ONLY: csym_type,&
                                              kp_sym_gen,&
                                              release_csym_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_env_create,&
                                              kpoint_env_p_type,&
                                              kpoint_env_type,&
                                              kpoint_type
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_cart_create,&
                                              mp_cart_sub,&
                                              mp_max,&
                                              mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_matrix_pools,                 ONLY: mpools_create,&
                                              mpools_get,&
                                              mpools_rebuild_fm_pools,&
                                              qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: allocate_mo_set,&
                                              get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              get_neighbor_list_set_p,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE scf_control_types,               ONLY: smear_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kpoint_methods'

   PUBLIC :: kpoint_initialize, kpoint_env_initialize, kpoint_initialize_mos
   PUBLIC :: kpoint_init_cell_index, kpoint_set_mo_occupation
   PUBLIC :: kpoint_density_matrices, kpoint_density_transform
   PUBLIC :: rskp_transform

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Generate the kpoints and initialize the kpoint environment
!> \param kpoint       The kpoint environment
!> \param particle_set Particle types and coordinates
!> \param cell         Computational cell information
! **************************************************************************************************
   SUBROUTINE kpoint_initialize(kpoint, particle_set, cell)

      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_initialize', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2), ALLOCATABLE, DIMENSION(:)        :: element
      INTEGER                                            :: handle, i, ik, iounit, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atype
      LOGICAL                                            :: inv, spez, symmo
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: weight, zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coord
      TYPE(csym_type), POINTER                           :: crys_sym

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_methods.F",122)

      SELECT CASE (kpoint%kp_scheme)
      CASE ("NONE")
         ! do nothing
      CASE ("GAMMA")
         kpoint%nkp = 1
         ALLOCATE (kpoint%xkp(3, 1), kpoint%wkp(1))
         kpoint%xkp(1:3, 1) = 0.0_dp
         kpoint%wkp(1) = 1.0_dp
      CASE ("MONKHORST-PACK", "MACDONALD")
         NULLIFY (crys_sym)

         IF (kpoint%full_grid) THEN
            ! we set up a random molecule to avoid any possible symmetry
            natom = 10
            ALLOCATE (coord(3, natom), zeff(natom), weight(natom), atype(natom), element(natom))
            DO i = 1, natom
               atype(i) = i
               coord(1, i) = SIN(i*0.12345_dp)
               coord(2, i) = COS(i*0.23456_dp)
               coord(3, i) = SIN(i*0.34567_dp)
            END DO
         ELSE
            natom = SIZE(particle_set)
            ALLOCATE (coord(3, natom), zeff(natom), weight(natom), atype(natom), element(natom))
            DO i = 1, natom
               CALL get_atomic_kind(atomic_kind=particle_set(i)%atomic_kind, kind_number=atype(i))
               coord(1:3, i) = particle_set(i)%r(1:3)
            END DO
         END IF
         IF (kpoint%verbose) THEN
            iounit = cp_logger_get_default_io_unit()
         ELSE
            iounit = -1
         END IF

         CALL kp_sym_gen(crys_sym, coord, atype, cell%hmat, &
                         nk=kpoint%nkp_grid, &
                         shift=kpoint%kp_shift, &
                         symm=kpoint%symmetry, &
                         delta=kpoint%eps_geo, &
                         iounit=iounit, &
                         xkpoint=kpoint%xkp, &
                         wkpoint=kpoint%wkp)
         kpoint%nkp = SIZE(kpoint%wkp)

         ! inversion symmetry
         inv = (crys_sym%li > 0)
         ! symmorphic group without special translation
         symmo = (crys_sym%isy == 1)

         CALL release_csym_type(crys_sym)
         DEALLOCATE (coord, zeff, weight, atype, element)

      CASE ("GENERAL")
         ! do nothing
      CASE DEFAULT
         IF(.NOT.(.FALSE.))CALL cp__a("kpoint_methods.F",180)
      END SELECT

      ! check for consistency of options
      SELECT CASE (kpoint%kp_scheme)
      CASE ("NONE")
         ! don't use k-point code
      CASE ("GAMMA")
         IF(.NOT.(kpoint%nkp == 1))CALL cp__a("kpoint_methods.F",188)
         IF(.NOT.(SUM(ABS(kpoint%xkp)) <= 1.e-12_dp))CALL cp__a("kpoint_methods.F",189)
         IF(.NOT.(kpoint%wkp(1) == 1.0_dp))CALL cp__a("kpoint_methods.F",190)
         IF(.NOT.(.NOT. kpoint%symmetry))CALL cp__a("kpoint_methods.F",191)
      CASE ("GENERAL")
         IF(.NOT.(.NOT. kpoint%symmetry))CALL cp__a("kpoint_methods.F",193)
         IF(.NOT.(kpoint%nkp >= 1))CALL cp__a("kpoint_methods.F",194)
      CASE ("MONKHORST-PACK", "MACDONALD")
         IF(.NOT.(kpoint%nkp >= 1))CALL cp__a("kpoint_methods.F",196)
      END SELECT
      IF (kpoint%use_real_wfn) THEN
         ! what about inversion symmetry?
         ikloop: DO ik = 1, kpoint%nkp
            DO i = 1, 3
               spez = (kpoint%xkp(i, ik) == 0.0_dp .OR. kpoint%xkp(i, ik) == 0.5_dp)
               IF (.NOT. spez) EXIT ikloop
            END DO
         END DO ikloop
         IF (.NOT. spez .AND. .NOT. (inv .AND. symmo)) THEN
            ! Warning: real wfn might be wrong for this system
            CALL cp_warn(cp__l("kpoint_methods.F",208), &
                         "A calculation using real wavefunctions is requested. "// &
                         "We could not determine if the symmetry of the system allows real wavefunctions. ")

         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE kpoint_initialize

! **************************************************************************************************
!> \brief Initialize the kpoint environment
!> \param kpoint       Kpoint environment
! **************************************************************************************************
   SUBROUTINE kpoint_env_initialize(kpoint)

      TYPE(kpoint_type), POINTER                         :: kpoint

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_env_initialize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm_cart, comm_inter_kp, comm_kp, &
                                                            handle, igr, ik, ikk, ngr, niogrp, &
                                                            nkp, nkp_grp, nkp_loc, npe, unit_nr
      INTEGER, DIMENSION(2)                              :: dims, pos
      LOGICAL, DIMENSION(2)                              :: rdim
      TYPE(cp_para_cart_type), POINTER                   :: cart
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_full, &
                                                            para_env_inter_kp, para_env_kp
      TYPE(kpoint_env_p_type), DIMENSION(:), POINTER     :: kp_env
      TYPE(kpoint_env_type), POINTER                     :: kp

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_methods.F",243)
      para_env => kpoint%para_env
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("kpoint_methods.F",245)
      IF(.NOT.(.NOT. ASSOCIATED(kpoint%kp_env)))CALL cp__a("kpoint_methods.F",246)

      NULLIFY (kp_env)
      nkp = kpoint%nkp
      npe = para_env%num_pe
      IF (npe == 1) THEN
         ! only one process availabe -> owns all kpoints
         ALLOCATE (kp_env(nkp))
         DO ik = 1, nkp
            NULLIFY (kp_env(ik)%kpoint_env)
            CALL kpoint_env_create(kp_env(ik)%kpoint_env)
            kp => kp_env(ik)%kpoint_env
            kp%nkpoint = ik
            kp%wkp = kpoint%wkp(ik)
            kp%xkp(1:3) = kpoint%xkp(1:3, ik)
            kp%is_local = .TRUE.
         END DO
         ALLOCATE (kpoint%kp_dist(2, 1))
         kpoint%kp_dist(1, 1) = 1
         kpoint%kp_dist(2, 1) = nkp
         kpoint%kp_range(1) = 1
         kpoint%kp_range(2) = nkp
         kpoint%kp_env => kp_env
         ! parallel environments
         kpoint%para_env_full => para_env
         CALL cp_para_env_retain(para_env)
         kpoint%para_env_kp => para_env
         CALL cp_para_env_retain(para_env)
         kpoint%para_env_inter_kp => para_env
         CALL cp_para_env_retain(para_env)
         kpoint%iogrp = .TRUE.
         kpoint%nkp_groups = 1
      ELSE
         IF (kpoint%parallel_group_size == -1) THEN
            ! maximum parallelization over kpoints
            ! making sure that the group size divides the npe and the nkp_grp the nkp
            ! in the worst case, there will be no parallelism over kpoints.
            DO igr = npe, 1, -1
               IF (MOD(npe, igr) .NE. 0) CYCLE
               nkp_grp = npe/igr
               IF (MOD(nkp, nkp_grp) .NE. 0) CYCLE
               ngr = igr
            ENDDO
         ELSE IF (kpoint%parallel_group_size == 0) THEN
            ! no parallelization over kpoints
            ngr = npe
         ELSE IF (kpoint%parallel_group_size > 0) THEN
            ngr = MIN(kpoint%parallel_group_size, npe)
         ELSE
            IF(.NOT.(.FALSE.))CALL cp__a("kpoint_methods.F",295)
         END IF
         nkp_grp = npe/ngr
         ! processor dimensions
         dims(1) = ngr
         dims(2) = nkp_grp
         IF(.NOT.(MOD(nkp, nkp_grp) == 0))CALL cp__a("kpoint_methods.F",301)
         nkp_loc = nkp/nkp_grp

         IF ((dims(1)*dims(2) /= npe)) THEN
            CALL cp__b("kpoint_methods.F",305,"Number of processors is not divisible by the kpoint group size.")
         END IF

         NULLIFY (cart, para_env_full, para_env_kp, para_env_inter_kp)
         CALL mp_cart_create(comm_old=para_env%group, ndims=2, dims=dims, pos=pos, &
                             comm_cart=comm_cart)

         CALL cp_cart_create(cart, comm_cart, ndims=2, owns_group=.TRUE.)
         CALL cp_para_env_create(para_env_full, comm_cart, owns_group=.FALSE.)
         rdim(2) = .FALSE.
         rdim(1) = .TRUE.
         CALL mp_cart_sub(comm=comm_cart, rdim=rdim, sub_comm=comm_kp)
         CALL cp_para_env_create(para_env_kp, comm_kp, owns_group=.TRUE.)
         rdim(2) = .TRUE.
         rdim(1) = .FALSE.
         CALL mp_cart_sub(comm=comm_cart, rdim=rdim, sub_comm=comm_inter_kp)
         CALL cp_para_env_create(para_env_inter_kp, comm_inter_kp, &
                                 owns_group=.TRUE.)

         niogrp = 0
         IF (para_env%ionode) niogrp = 1
         CALL mp_sum(niogrp, para_env_kp%group)
         kpoint%iogrp = (niogrp == 1)

         ! parallel groups
         kpoint%cart => cart
         kpoint%para_env_full => para_env_full
         kpoint%para_env_kp => para_env_kp
         kpoint%para_env_inter_kp => para_env_inter_kp

         ! distribution of kpoints
         ALLOCATE (kpoint%kp_dist(2, nkp_grp))
         DO igr = 1, nkp_grp
            kpoint%kp_dist(1:2, igr) = get_limit(nkp, nkp_grp, igr-1)
         END DO
         ! local kpoints
         kpoint%kp_range(1:2) = kpoint%kp_dist(1:2, para_env_inter_kp%mepos+1)
         ALLOCATE (kp_env(nkp_loc))
         DO ik = 1, nkp_loc
            NULLIFY (kp_env(ik)%kpoint_env)
            ikk = kpoint%kp_range(1)+ik-1
            CALL kpoint_env_create(kp_env(ik)%kpoint_env)
            kp => kp_env(ik)%kpoint_env
            kp%nkpoint = ikk
            kp%wkp = kpoint%wkp(ikk)
            kp%xkp(1:3) = kpoint%xkp(1:3, ikk)
            kp%is_local = (ngr == 1)
         END DO

         kpoint%kp_env => kp_env

         unit_nr = cp_logger_get_default_io_unit()
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *)
            WRITE (unit_nr, FMT="(T2,A,T71,I10)") "KPOINTS| Number of kpoint groups ", nkp_grp
            WRITE (unit_nr, FMT="(T2,A,T71,I10)") "KPOINTS| Size of each kpoint group", ngr
            WRITE (unit_nr, FMT="(T2,A,T71,I10)") "KPOINTS| Number of kpoints per group", nkp_loc
         END IF
         kpoint%nkp_groups = nkp_grp

      END IF

      CALL timestop(handle)

   END SUBROUTINE kpoint_env_initialize

! **************************************************************************************************
!> \brief Initialize a set of MOs and density matrix for each kpoint (kpoint group)
!> \param kpoint  Kpoint environment
!> \param mos     Reference MOs (global)
! **************************************************************************************************
   SUBROUTINE kpoint_initialize_mos(kpoint, mos)

      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_initialize_mos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, ik, is, nao, nc, nelectron, &
                                                            nkp_loc, nmo, nspin
      REAL(KIND=dp)                                      :: flexible_electron_count, maxocc, n_el_f
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_ao_fm_pools
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      TYPE(cp_fm_type), POINTER                          :: fmlocal
      TYPE(kpoint_env_type), POINTER                     :: kp
      TYPE(qs_matrix_pools_type), POINTER                :: mpools

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_methods.F",396)

      IF (ASSOCIATED(mos)) THEN

         IF (kpoint%use_real_wfn) THEN
            nc = 1
         ELSE
            nc = 2
         END IF
         nspin = SIZE(mos, 1)
         nkp_loc = kpoint%kp_range(2)-kpoint%kp_range(1)+1
         IF (nkp_loc > 0) THEN
            IF(.NOT.(SIZE(kpoint%kp_env) == nkp_loc))CALL cp__a("kpoint_methods.F",408)
            ! allocate the mo sets, correct number of kpoints (local), real/complex, spin
            DO ik = 1, nkp_loc
               kp => kpoint%kp_env(ik)%kpoint_env
               ALLOCATE (kp%mos(nc, nspin))
               DO is = 1, nspin
                  CALL get_mo_set(mos(is)%mo_set, nao=nao, nmo=nmo, nelectron=nelectron, &
                                  n_el_f=n_el_f, maxocc=maxocc, flexible_electron_count=flexible_electron_count)
                  DO ic = 1, nc
                     NULLIFY (kp%mos(ic, is)%mo_set)
                     CALL allocate_mo_set(kp%mos(ic, is)%mo_set, nao, nmo, nelectron, n_el_f, maxocc, &
                                          flexible_electron_count)
                  END DO
               END DO
            END DO

            ! generate the blacs environment for the kpoint group
            ! we generate a blacs env for each kpoint group in parallel
            ! we assume here that the group para_env_inter_kp will connect
            ! equivalent parts of fm matrices, i.e. no reshuffeling of processors
            NULLIFY (blacs_env)
            CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=kpoint%para_env_kp)
            kpoint%blacs_env => blacs_env

            ! matrix pools for the kpoint group, information on MOs is transfered using
            ! generic mos structure
            NULLIFY (mpools)
            CALL mpools_create(mpools=mpools)
            CALL mpools_rebuild_fm_pools(mpools=mpools, mos=mos, &
                                         blacs_env=blacs_env, para_env=kpoint%para_env_kp)
            kpoint%mpools => mpools

            ! allocate density matrices
            CALL mpools_get(mpools, ao_ao_fm_pools=ao_ao_fm_pools)
            CALL fm_pool_create_fm(ao_ao_fm_pools(1)%pool, fmlocal)
            CALL cp_fm_get_info(fmlocal, matrix_struct=matrix_struct)
            DO ik = 1, nkp_loc
               kp => kpoint%kp_env(ik)%kpoint_env
               ! density matrix
               IF (ASSOCIATED(kp%pmat)) THEN
                  DO is = 1, SIZE(kp%pmat, 2)
                     DO ic = 1, SIZE(kp%pmat, 1)
                        CALL cp_fm_release(kp%pmat(ic, is)%matrix)
                     END DO
                  END DO
                  DEALLOCATE (kp%pmat)
               END IF
               ALLOCATE (kp%pmat(nc, nspin))
               DO is = 1, nspin
                  DO ic = 1, nc
                     CALL cp_fm_create(kp%pmat(ic, is)%matrix, matrix_struct)
                  END DO
               END DO
               ! energy weighted density matrix
               IF (ASSOCIATED(kp%wmat)) THEN
                  DO is = 1, SIZE(kp%wmat, 2)
                     DO ic = 1, SIZE(kp%wmat, 1)
                        CALL cp_fm_release(kp%wmat(ic, is)%matrix)
                     END DO
                  END DO
                  DEALLOCATE (kp%wmat)
               END IF
               ALLOCATE (kp%wmat(nc, nspin))
               DO is = 1, nspin
                  DO ic = 1, nc
                     CALL cp_fm_create(kp%wmat(ic, is)%matrix, matrix_struct)
                  END DO
               END DO
            END DO
            CALL fm_pool_give_back_fm(ao_ao_fm_pools(1)%pool, fmlocal)

         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE kpoint_initialize_mos

! **************************************************************************************************
!> \brief Generates the mapping of cell indices and linear RS index
!>        CELL (0,0,0) is always mapped to index 1
!> \param kpoint    Kpoint environment
!> \param sab_nl    Defining neighbour list
!> \param para_env  Parallel environment
!> \param dft_control ...
! **************************************************************************************************
   SUBROUTINE kpoint_init_cell_index(kpoint, sab_nl, para_env, dft_control)

      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_init_cell_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i1, i2, i3, ic, icount, it, &
                                                            ncount
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: index_to_cell
      INTEGER, DIMENSION(3)                              :: cell, itm
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      INTEGER, DIMENSION(:, :, :), POINTER               :: cti
      LOGICAL                                            :: new
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_methods.F",518)

      ALLOCATE (list(3, 125))
      list = 0
      icount = 1

      CALL neighbor_list_iterator_create(nl_iterator, sab_nl)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, cell=cell)

         new = .TRUE.
         DO ic = 1, icount
            IF (cell(1) == list(1, ic) .AND. cell(2) == list(2, ic) .AND. &
                cell(3) == list(3, ic)) THEN
               new = .FALSE.
               EXIT
            END IF
         END DO
         IF (new) THEN
            icount = icount+1
            IF (icount > SIZE(list, 2)) THEN
               CALL reallocate(list, 1, 3, 1, 2*SIZE(list, 2))
            END IF
            list(1:3, icount) = cell(1:3)
         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      itm(1) = MAXVAL(ABS(list(1, 1:icount)))
      itm(2) = MAXVAL(ABS(list(2, 1:icount)))
      itm(3) = MAXVAL(ABS(list(3, 1:icount)))
      CALL mp_max(itm, para_env%group)
      it = MAXVAL(itm(1:3))
      IF (ASSOCIATED(kpoint%cell_to_index)) THEN
         DEALLOCATE (kpoint%cell_to_index)
      END IF
      ALLOCATE (kpoint%cell_to_index(-itm(1):itm(1), -itm(2):itm(2), -itm(3):itm(3)))
      cti => kpoint%cell_to_index
      cti(:, :, :) = 0
      DO ic = 1, icount
         i1 = list(1, ic)
         i2 = list(2, ic)
         i3 = list(3, ic)
         cti(i1, i2, i3) = ic
      END DO
      CALL mp_sum(cti, para_env%group)
      ncount = 0
      DO i1 = -itm(1), itm(1)
         DO i2 = -itm(2), itm(2)
            DO i3 = -itm(3), itm(3)
               IF (cti(i1, i2, i3) == 0) THEN
                  cti(i1, i2, i3) = 1000000
               ELSE
                  ncount = ncount+1
                  cti(i1, i2, i3) = (ABS(i1)+ABS(i2)+ABS(i3))*1000+ABS(i3)*100+ABS(i2)*10+ABS(i1)
                  cti(i1, i2, i3) = cti(i1, i2, i3)+(i1+i2+i3)
               END IF
            END DO
         END DO
      END DO

      ALLOCATE (index_to_cell(3, ncount))
      DO ic = 1, ncount
         cell = MINLOC(cti)
         i1 = cell(1)-1-itm(1)
         i2 = cell(2)-1-itm(2)
         i3 = cell(3)-1-itm(3)
         cti(i1, i2, i3) = 1000000
         index_to_cell(1, ic) = i1
         index_to_cell(2, ic) = i2
         index_to_cell(3, ic) = i3
      END DO
      cti(:, :, :) = 0
      DO ic = 1, ncount
         i1 = index_to_cell(1, ic)
         i2 = index_to_cell(2, ic)
         i3 = index_to_cell(3, ic)
         cti(i1, i2, i3) = ic
      END DO

      ! keep pointer to this neighborlist
      kpoint%sab_nl => sab_nl

      ! set number of images
      dft_control%nimages = SIZE(index_to_cell, 2)

      DEALLOCATE (list, index_to_cell)

      CALL timestop(handle)
   END SUBROUTINE kpoint_init_cell_index

! **************************************************************************************************
!> \brief Transformation of real space matrices to a kpoint
!> \param rmatrix  Real part of kpoint matrix
!> \param cmatrix  Complex part of kpoint matrix (optional)
!> \param rsmat    Real space matrices
!> \param ispin    Spin index
!> \param xkp      Kpoint coordinates
!> \param cell_to_index   mapping of cell indices to RS index
!> \param sab_nl   Defining neighbor list
! **************************************************************************************************
   SUBROUTINE rskp_transform(rmatrix, cmatrix, rsmat, ispin, &
                             xkp, cell_to_index, sab_nl)

      TYPE(cp_dbcsr_type)                                :: rmatrix
      TYPE(cp_dbcsr_type), OPTIONAL                      :: cmatrix
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rsmat
      INTEGER, INTENT(IN)                                :: ispin
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: xkp
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rskp_transform', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ic, icol, irow, jatom
      INTEGER, DIMENSION(3)                              :: cell
      LOGICAL                                            :: do_symmetric, found, wfn_real_only
      REAL(KIND=dp)                                      :: arg, coskl, fsym, sinkl
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cblock, rblock, rsblock
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)

      wfn_real_only = .TRUE.
      IF (PRESENT(cmatrix)) wfn_real_only = .FALSE.
      CALL get_neighbor_list_set_p(neighbor_list_sets=sab_nl, symmetric=do_symmetric)

      IF (wfn_real_only) THEN
         CALL cp_dbcsr_set(rmatrix, 0.0_dp)
      ELSE
         CALL cp_dbcsr_set(rmatrix, 0.0_dp)
         CALL cp_dbcsr_set(cmatrix, 0.0_dp)
      END IF

      CALL neighbor_list_iterator_create(nl_iterator, sab_nl)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom, cell=cell)

         fsym = 1.0_dp
         irow = iatom
         icol = jatom
         IF (do_symmetric .AND. (iatom > jatom)) THEN
            irow = jatom
            icol = iatom
            fsym = -1.0_dp
         END IF

         arg = REAL(cell(1), dp)*xkp(1)+REAL(cell(2), dp)*xkp(2)+REAL(cell(3), dp)*xkp(3)
         coskl = COS(twopi*arg)
         sinkl = fsym*SIN(twopi*arg)

         ic = cell_to_index(cell(1), cell(2), cell(3))
         IF (.NOT. do_symmetric .AND. ic < 1) ic = cell_to_index(-cell(1), -cell(2), -cell(3))
         IF(.NOT.(ic > 0))CALL cp__a("kpoint_methods.F",674)

         CALL cp_dbcsr_get_block_p(matrix=rsmat(ispin, ic)%matrix, row=irow, col=icol, &
                                   block=rsblock, found=found)
         IF(.NOT.(found))CALL cp__a("kpoint_methods.F",678)

         IF (wfn_real_only) THEN
            CALL cp_dbcsr_get_block_p(matrix=rmatrix, row=irow, col=icol, &
                                      block=rblock, found=found)
            IF(.NOT.(found))CALL cp__a("kpoint_methods.F",683)
            rblock = rblock+coskl*rsblock
         ELSE
            CALL cp_dbcsr_get_block_p(matrix=rmatrix, row=irow, col=icol, &
                                      block=rblock, found=found)
            IF(.NOT.(found))CALL cp__a("kpoint_methods.F",688)
            CALL cp_dbcsr_get_block_p(matrix=cmatrix, row=irow, col=icol, &
                                      block=cblock, found=found)
            IF(.NOT.(found))CALL cp__a("kpoint_methods.F",691)
            rblock = rblock+coskl*rsblock
            cblock = cblock+sinkl*rsblock
         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      CALL timestop(handle)

   END SUBROUTINE rskp_transform

! **************************************************************************************************
!> \brief Given the eigenvalues of all kpoints, calculates the occupation numbers
!> \param kpoint  Kpoint environment
!> \param smear   Smearing information
! **************************************************************************************************
   SUBROUTINE kpoint_set_mo_occupation(kpoint, smear)

      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(smear_type), POINTER                          :: smear

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_set_mo_occupation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, ikpgr, ispin, kplocal, nb, &
                                                            ne_a, ne_b, nelectron, nkp, nmo, nspin
      INTEGER, DIMENSION(2)                              :: kp_range
      REAL(KIND=dp)                                      :: kTS, mu, mus(2), nel
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: weig, wocc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues, occupation, wkp
      TYPE(cp_para_env_type), POINTER                    :: para_env_inter_kp
      TYPE(kpoint_env_type), POINTER                     :: kp
      TYPE(mo_set_type), POINTER                         :: mo_set

! first collect all the eigenvalues

      CALL timeset(routineN, handle)

      CALL get_kpoint_info(kpoint, nkp=nkp)
      kp => kpoint%kp_env(1)%kpoint_env
      nspin = SIZE(kp%mos, 2)
      mo_set => kp%mos(1, 1)%mo_set
      CALL get_mo_set(mo_set, nmo=nmo, nelectron=nelectron)
      ne_a = nelectron
      IF (nspin == 2) THEN
         CALL get_mo_set(kp%mos(1, 2)%mo_set, nmo=nb, nelectron=ne_b)
         IF(.NOT.(nmo == nb))CALL cp__a("kpoint_methods.F",738)
      END IF
      ALLOCATE (weig(nmo, nkp, nspin), wocc(nmo, nkp, nspin))
      weig = 0.0_dp
      wocc = 0.0_dp
      CALL get_kpoint_info(kpoint, kp_range=kp_range)
      kplocal = kp_range(2)-kp_range(1)+1
      DO ikpgr = 1, kplocal
         ik = kp_range(1)+ikpgr-1
         kp => kpoint%kp_env(ikpgr)%kpoint_env
         DO ispin = 1, nspin
            mo_set => kp%mos(1, ispin)%mo_set
            CALL get_mo_set(mo_set, eigenvalues=eigenvalues)
            weig(1:nmo, ik, ispin) = eigenvalues(1:nmo)
         END DO
      END DO
      CALL get_kpoint_info(kpoint, para_env_inter_kp=para_env_inter_kp)
      CALL mp_sum(weig, para_env_inter_kp%group)

      CALL get_kpoint_info(kpoint, wkp=wkp)
      IF (smear%do_smear) THEN
         ! finite electronic temperature
         SELECT CASE (smear%method)
         CASE (smear_fermi_dirac)
            IF (nspin == 1) THEN
               nel = REAL(nelectron, KIND=dp)
               CALL Fermikp(wocc(:, :, 1), mus(1), kTS, weig(:, :, 1), nel, wkp, &
                            smear%electronic_temperature, 2.0_dp)
            ELSE IF (smear%fixed_mag_mom > 0.0_dp) THEN
               CALL cp__b("kpoint_methods.F",767,"kpoints: Not supoorted")
               nel = REAL(ne_a, KIND=dp)
               CALL Fermikp(wocc(:, :, 1), mus(1), kTS, weig(:, :, 1), nel, wkp, &
                            smear%electronic_temperature, 1.0_dp)
               nel = REAL(ne_b, KIND=dp)
               CALL Fermikp(wocc(:, :, 2), mus(2), kTS, weig(:, :, 2), nel, wkp, &
                            smear%electronic_temperature, 1.0_dp)
            ELSE
               nel = REAL(ne_a, KIND=dp)+REAL(ne_b, KIND=dp)
               CALL Fermikp2(wocc(:, :, :), mu, kTS, weig(:, :, :), nel, wkp, &
                             smear%electronic_temperature)
               kTS = kTS/2._dp
               mus(1:2) = mu
            END IF
         CASE DEFAULT
            CALL cp__b("kpoint_methods.F",782,"kpoints: Not supported")
         END SELECT
      ELSE
         ! fixed occupations (2/1)
         IF (nspin == 1) THEN
            nel = REAL(nelectron, KIND=dp)
            CALL Fermikp(wocc(:, :, 1), mus(1), kTS, weig(:, :, 1), nel, wkp, 0.0_dp, 2.0_dp)
         ELSE
            nel = REAL(ne_a, KIND=dp)
            CALL Fermikp(wocc(:, :, 1), mus(1), kTS, weig(:, :, 1), nel, wkp, 0.0_dp, 1.0_dp)
            nel = REAL(ne_b, KIND=dp)
            CALL Fermikp(wocc(:, :, 2), mus(2), kTS, weig(:, :, 2), nel, wkp, 0.0_dp, 1.0_dp)
         END IF
      END IF
      DO ikpgr = 1, kplocal
         ik = kp_range(1)+ikpgr-1
         kp => kpoint%kp_env(ikpgr)%kpoint_env
         DO ispin = 1, nspin
            mo_set => kp%mos(1, ispin)%mo_set
            CALL get_mo_set(mo_set, eigenvalues=eigenvalues, occupation_numbers=occupation)
            eigenvalues(1:nmo) = weig(1:nmo, ik, ispin)
            occupation(1:nmo) = wocc(1:nmo, ik, ispin)
            mo_set%kTS = kTS
            mo_set%mu = mus(ispin)
         END DO
      END DO

      DEALLOCATE (weig, wocc)

      CALL timestop(handle)

   END SUBROUTINE kpoint_set_mo_occupation

! **************************************************************************************************
!> \brief Calculate kpoint density matrices (rho(k), owned by kpoint groups)
!> \param kpoint    kpoint environment
!> \param energy_weighted  calculate energy weighted density matrix
! **************************************************************************************************
   SUBROUTINE kpoint_density_matrices(kpoint, energy_weighted)

      TYPE(kpoint_type), POINTER                         :: kpoint
      LOGICAL, OPTIONAL                                  :: energy_weighted

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_density_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikpgr, ispin, kplocal, nao, nmo, &
                                                            nspin
      INTEGER, DIMENSION(2)                              :: kp_range
      LOGICAL                                            :: wtype
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues, occupation
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      TYPE(cp_fm_type), POINTER                          :: cpmat, fwork, pmat, rpmat
      TYPE(kpoint_env_type), POINTER                     :: kp
      TYPE(mo_set_type), POINTER                         :: mo_set

      CALL timeset(routineN, handle)

      IF (PRESENT(energy_weighted)) THEN
         wtype = energy_weighted
      ELSE
         ! default is normal density matrix
         wtype = .FALSE.
      END IF

      ! work matrix
      mo_set => kpoint%kp_env(1)%kpoint_env%mos(1, 1)%mo_set
      CALL get_mo_set(mo_set, nao=nao, nmo=nmo)
      CALL cp_fm_get_info(mo_set%mo_coeff, matrix_struct=matrix_struct)
      CALL cp_fm_create(fwork, matrix_struct)

      CALL get_kpoint_info(kpoint, kp_range=kp_range)
      kplocal = kp_range(2)-kp_range(1)+1
      DO ikpgr = 1, kplocal
         kp => kpoint%kp_env(ikpgr)%kpoint_env
         nspin = SIZE(kp%mos, 2)
         DO ispin = 1, nspin
            mo_set => kp%mos(1, ispin)%mo_set
            IF (wtype) THEN
               CALL get_mo_set(mo_set, eigenvalues=eigenvalues)
            END IF
            IF (kpoint%use_real_wfn) THEN
               IF (wtype) THEN
                  pmat => kp%wmat(1, ispin)%matrix
               ELSE
                  pmat => kp%pmat(1, ispin)%matrix
               END IF
               CALL get_mo_set(mo_set, occupation_numbers=occupation)
               CALL cp_fm_to_fm(mo_set%mo_coeff, fwork)
               CALL cp_fm_column_scale(fwork, occupation)
               IF (wtype) THEN
                  CALL cp_fm_column_scale(fwork, eigenvalues)
               END IF
               CALL cp_gemm("N", "T", nao, nao, nmo, 1.0_dp, mo_set%mo_coeff, fwork, 0.0_dp, pmat)
            ELSE
               IF (wtype) THEN
                  rpmat => kp%wmat(1, ispin)%matrix
                  cpmat => kp%wmat(2, ispin)%matrix
               ELSE
                  rpmat => kp%pmat(1, ispin)%matrix
                  cpmat => kp%pmat(2, ispin)%matrix
               END IF
               CALL get_mo_set(mo_set, occupation_numbers=occupation)
               CALL cp_fm_to_fm(mo_set%mo_coeff, fwork)
               CALL cp_fm_column_scale(fwork, occupation)
               IF (wtype) THEN
                  CALL cp_fm_column_scale(fwork, eigenvalues)
               END IF
               ! Re(c)*Re(c)
               CALL cp_gemm("N", "T", nao, nao, nmo, 1.0_dp, mo_set%mo_coeff, fwork, 0.0_dp, rpmat)
               mo_set => kp%mos(2, ispin)%mo_set
               ! Im(c)*Re(c)
               CALL cp_gemm("N", "T", nao, nao, nmo, -1.0_dp, mo_set%mo_coeff, fwork, 0.0_dp, cpmat)
               ! Re(c)*Im(c)
               CALL cp_gemm("N", "T", nao, nao, nmo, 1.0_dp, fwork, mo_set%mo_coeff, 1.0_dp, cpmat)
               CALL cp_fm_to_fm(mo_set%mo_coeff, fwork)
               CALL cp_fm_column_scale(fwork, occupation)
               IF (wtype) THEN
                  CALL cp_fm_column_scale(fwork, eigenvalues)
               END IF
               ! Im(c)*Im(c)
               CALL cp_gemm("N", "T", nao, nao, nmo, 1.0_dp, mo_set%mo_coeff, fwork, 1.0_dp, rpmat)
            END IF
         END DO
      END DO

      CALL cp_fm_release(fwork)

      CALL timestop(handle)

   END SUBROUTINE kpoint_density_matrices

! **************************************************************************************************
!> \brief generate real space density matrices in DBCSR format
!> \param kpoint  Kpoint environment
!> \param denmat  Real space (DBCSR) density matrices
!> \param wtype   True = energy weighted density matrix
!>                False = normal density matrix
!> \param tempmat DBCSR matrix to be used as template
!> \param sab_nl ...
!> \param fmwork  FM work matrices (kpoint group)
! **************************************************************************************************
   SUBROUTINE kpoint_density_transform(kpoint, denmat, wtype, tempmat, sab_nl, fmwork)

      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: denmat
      LOGICAL, INTENT(IN)                                :: wtype
      TYPE(cp_dbcsr_type), POINTER                       :: tempmat
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fmwork

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_density_transform', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bcntxt, handle, iatom, ic, icell, icol, &
                                                            ik, ikk, irow, ispin, jatom, nc, nimg, &
                                                            nkp, nspin
      INTEGER, DIMENSION(3)                              :: cell
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: do_symmetric, found, my_kpgrp, real_only
      REAL(KIND=dp)                                      :: arg, coskl, fc, sinkl
      REAL(KIND=dp), DIMENSION(:), POINTER               :: wkp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cblock, dblock, rblock, xkp
      TYPE(cp_dbcsr_type), POINTER                       :: cpmat, rpmat
      TYPE(cp_fm_type), POINTER                          :: fmdummy
      TYPE(cp_para_env_type), POINTER                    :: para_env_inter_kp
      TYPE(kpoint_env_type), POINTER                     :: kp
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)

      ! test for full kpoint grid
      IF (.NOT. kpoint%full_grid) &
         CALL cp_abort(cp__l("kpoint_methods.F",957), "k-points need FULL_GRID currently. "// &
                       "Reduced grids not yet working correctly")

      ! work storage
      ALLOCATE (rpmat, cpmat)
      CALL cp_dbcsr_init(rpmat)
      CALL cp_dbcsr_init(cpmat)
      CALL cp_dbcsr_create(rpmat, template=tempmat, matrix_type=dbcsr_type_symmetric)
      CALL cp_dbcsr_create(cpmat, template=tempmat, matrix_type=dbcsr_type_antisymmetric)
      CALL cp_dbcsr_alloc_block_from_nbl(rpmat, sab_nl)
      CALL cp_dbcsr_alloc_block_from_nbl(cpmat, sab_nl)

      CALL get_kpoint_info(kpoint, nkp=nkp, xkp=xkp, wkp=wkp, &
                           para_env_inter_kp=para_env_inter_kp, &
                           sab_nl=sab_nl, cell_to_index=cell_to_index)
      ! initialize real space density matrices
      kp => kpoint%kp_env(1)%kpoint_env
      nspin = SIZE(kp%mos, 2)
      nc = SIZE(kp%mos, 1)
      nimg = SIZE(denmat, 2)
      real_only = (nc == 1)
      bcntxt = kpoint%blacs_env_all%group
      NULLIFY (fmdummy)
      DO ispin = 1, nspin
         DO ic = 1, nimg
            CALL cp_dbcsr_set(denmat(ispin, ic)%matrix, 0.0_dp)
         END DO
         !
         DO ik = 1, nkp
            my_kpgrp = (ik >= kpoint%kp_range(1) .AND. ik <= kpoint%kp_range(2))
            IF (my_kpgrp) THEN
               ikk = ik-kpoint%kp_range(1)+1
               kp => kpoint%kp_env(ikk)%kpoint_env
            ELSE
               NULLIFY (kp)
            END IF
            ! collect this density matrix on all processors
            IF(.NOT.(SIZE(fmwork) >= nc))CALL cp__a("kpoint_methods.F",994)
            DO ic = 1, nc
               IF (my_kpgrp) THEN
                  ! choose density matrix
                  IF (wtype) THEN
                     CALL cp_fm_copy_general(kp%wmat(ic, ispin)%matrix, fmwork(ic)%matrix, bcntxt)
                  ELSE
                     CALL cp_fm_copy_general(kp%pmat(ic, ispin)%matrix, fmwork(ic)%matrix, bcntxt)
                  END IF
               ELSE
                  CALL cp_fm_copy_general(fmdummy, fmwork(ic)%matrix, bcntxt)
               END IF
            END DO
            ! reduce to dbcsr storage
            IF (real_only) THEN
               CALL copy_fm_to_dbcsr(fmwork(1)%matrix, rpmat, keep_sparsity=.TRUE.)
            ELSE
               CALL copy_fm_to_dbcsr(fmwork(1)%matrix, rpmat, keep_sparsity=.TRUE.)
               ! it seems this copy to a antisymmetric dbcsr matrix changes sign!
               CALL copy_fm_to_dbcsr(fmwork(2)%matrix, cpmat, keep_sparsity=.TRUE.)
            END IF

            ! transformation
            CALL get_neighbor_list_set_p(neighbor_list_sets=sab_nl, symmetric=do_symmetric)
            CALL neighbor_list_iterator_create(nl_iterator, sab_nl)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom, cell=cell)

               ! fc = +/-1 sign change from above, is used to get correct sign due to
               !           a<->b ==> cell = -cell
               irow = iatom
               icol = jatom
               fc = -1.0_dp
               IF (do_symmetric .AND. iatom > jatom) THEN
                  irow = jatom
                  icol = iatom
                  fc = 1.0_dp
               END IF
               icell = cell_to_index(cell(1), cell(2), cell(3))
               arg = REAL(cell(1), dp)*xkp(1, ik)+REAL(cell(2), dp)*xkp(2, ik)+REAL(cell(3), dp)*xkp(3, ik)
               coskl = wkp(ik)*COS(twopi*arg)
               sinkl = wkp(ik)*fc*SIN(twopi*arg)

               CALL cp_dbcsr_get_block_p(matrix=denmat(ispin, icell)%matrix, row=irow, col=icol, &
                                         block=dblock, found=found)
               IF(.NOT.(found))CALL cp__a("kpoint_methods.F",1039)

               IF (real_only) THEN
                  CALL cp_dbcsr_get_block_p(matrix=rpmat, row=irow, col=icol, &
                                            block=rblock, found=found)
                  IF(.NOT.(found))CALL cp__a("kpoint_methods.F",1044)
                  dblock = dblock+coskl*rblock
               ELSE
                  CALL cp_dbcsr_get_block_p(matrix=rpmat, row=irow, col=icol, &
                                            block=rblock, found=found)
                  IF(.NOT.(found))CALL cp__a("kpoint_methods.F",1049)
                  CALL cp_dbcsr_get_block_p(matrix=cpmat, row=irow, col=icol, &
                                            block=cblock, found=found)
                  IF(.NOT.(found))CALL cp__a("kpoint_methods.F",1052)
                  dblock = dblock+coskl*rblock
                  dblock = dblock+sinkl*cblock
               END IF
            END DO
            CALL neighbor_list_iterator_release(nl_iterator)

         END DO
      END DO

      CALL cp_dbcsr_deallocate_matrix(rpmat)
      CALL cp_dbcsr_deallocate_matrix(cpmat)

      CALL timestop(handle)

   END SUBROUTINE kpoint_density_transform

! **************************************************************************************************

END MODULE kpoint_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that contains the algorithms to perform an itrative
!>         diagonalization by the block-Lanczos approach
!> \par History
!>      05.2009 created [MI]
!> \author fawzi
! **************************************************************************************************
MODULE qs_scf_lanczos

   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_qr_factorization,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_transpose,&
                                              cp_fm_triangular_multiply
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_submatrix, cp_fm_p_type, cp_fm_release, cp_fm_set_all, &
        cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type, cp_fm_vectorsnorm
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_scf_types,                    ONLY: krylov_space_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_lanczos'

   PUBLIC :: krylov_space_allocate, lanczos_refinement, lanczos_refinement_2v

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief  allocates matrices and vectros used in the construction of
!>        the krylov space and for the lanczos refinement
!> \param krylov_space ...
!> \param scf_control ...
!> \param mos ...
!> \param
!> \par History
!>      05.2009 created [MI]
! **************************************************************************************************

   SUBROUTINE krylov_space_allocate(krylov_space, scf_control, mos)

      TYPE(krylov_space_type), POINTER                   :: krylov_space
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'krylov_space_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, ispin, max_nmo, nao, nblock, &
                                                            ndim, nk, nmo, nspin
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(krylov_space)))CALL cp__a("qs_scf_lanczos.F",75)

      IF (.NOT. ASSOCIATED(krylov_space%mo_conv)) THEN
         NULLIFY (fm_struct_tmp, mo_coeff)

         krylov_space%nkrylov = scf_control%diagonalization%nkrylov
         krylov_space%nblock = scf_control%diagonalization%nblock_krylov

         nk = krylov_space%nkrylov
         nblock = krylov_space%nblock
         nspin = SIZE(mos, 1)

         ALLOCATE (krylov_space%mo_conv(nspin))
         ALLOCATE (krylov_space%mo_refine(nspin))
         ALLOCATE (krylov_space%chc_mat(nspin))
         ALLOCATE (krylov_space%c_vec(nspin))
         max_nmo = 0
         DO ispin = 1, nspin
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
            CALL cp_fm_create(krylov_space%mo_conv(ispin)%matrix, mo_coeff%matrix_struct)
            CALL cp_fm_create(krylov_space%mo_refine(ispin)%matrix, mo_coeff%matrix_struct)
            NULLIFY (fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmo, ncol_global=nmo, &
                                     para_env=mo_coeff%matrix_struct%para_env, &
                                     context=mo_coeff%matrix_struct%context)
            CALL cp_fm_create(krylov_space%chc_mat(ispin)%matrix, fm_struct_tmp, "chc")
            CALL cp_fm_create(krylov_space%c_vec(ispin)%matrix, fm_struct_tmp, "vec")
            CALL cp_fm_struct_release(fm_struct_tmp)
            max_nmo = MAX(max_nmo, nmo)
         END DO

         !the use of max_nmo might not be ok, in this case allocate nspin matrices
         ALLOCATE (krylov_space%c_eval(max_nmo))

         ALLOCATE (krylov_space%v_mat(nk))

         NULLIFY (fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nblock, &
                                  para_env=mo_coeff%matrix_struct%para_env, &
                                  context=mo_coeff%matrix_struct%context)
         DO ik = 1, nk
            CALL cp_fm_create(krylov_space%v_mat(ik)%matrix, matrix_struct=fm_struct_tmp, &
                              name="v_mat_"//TRIM(ADJUSTL(cp_to_string(ik))))
         END DO
         CALL cp_fm_create(krylov_space%tmp_mat, matrix_struct=fm_struct_tmp, &
                           name="tmp_mat")
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! NOTE: the following matrices are small and could be defined
!           as standard array rather than istributed fm
         NULLIFY (fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nblock, ncol_global=nblock, &
                                  para_env=mo_coeff%matrix_struct%para_env, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(krylov_space%block1_mat, matrix_struct=fm_struct_tmp, &
                           name="a_mat_"//TRIM(ADJUSTL(cp_to_string(ik))))
         CALL cp_fm_create(krylov_space%block2_mat, matrix_struct=fm_struct_tmp, &
                           name="b_mat_"//TRIM(ADJUSTL(cp_to_string(ik))))
         CALL cp_fm_create(krylov_space%block3_mat, matrix_struct=fm_struct_tmp, &
                           name="b2_mat_"//TRIM(ADJUSTL(cp_to_string(ik))))
         CALL cp_fm_struct_release(fm_struct_tmp)

         ndim = nblock*nk
         NULLIFY (fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ndim, ncol_global=ndim, &
                                  para_env=mo_coeff%matrix_struct%para_env, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(krylov_space%block4_mat, matrix_struct=fm_struct_tmp, &
                           name="t_mat")
         CALL cp_fm_create(krylov_space%block5_mat, matrix_struct=fm_struct_tmp, &
                           name="t_vec")
         CALL cp_fm_struct_release(fm_struct_tmp)
         ALLOCATE (krylov_space%t_eval(ndim))

      ELSE
         !Nothing should be done
      END IF

      CALL timestop(handle)

   END SUBROUTINE krylov_space_allocate

! **************************************************************************************************
!> \brief lanczos refinement by blocks of not-converged MOs
!> \param krylov_space ...
!> \param ks ...
!> \param c0 ...
!> \param c1 ...
!> \param eval ...
!> \param nao ...
!> \param eps_iter ...
!> \param ispin ...
!> \param check_moconv_only ...
!> \param
!> \par History
!>      05.2009 created [MI]
! **************************************************************************************************

   SUBROUTINE lanczos_refinement(krylov_space, ks, c0, c1, eval, nao, &
                                 eps_iter, ispin, check_moconv_only)

      TYPE(krylov_space_type), POINTER                   :: krylov_space
      TYPE(cp_fm_type), POINTER                          :: ks, c0, c1
      REAL(dp), DIMENSION(:), POINTER                    :: eval
      INTEGER, INTENT(IN)                                :: nao
      REAL(dp), INTENT(IN)                               :: eps_iter
      INTEGER, INTENT(IN)                                :: ispin
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_moconv_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lanczos_refinement', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rmone = -1.0_dp, rone = 1.0_dp, &
                                                            rzero = 0.0_dp

      INTEGER :: hand1, hand2, hand3, hand4, hand5, handle, ib, ik, imo, imo_low, imo_up, it, jt, &
         nblock, ndim, nmo, nmo_converged, nmo_nc, nmob, num_blocks
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: itaken
      LOGICAL                                            :: my_check_moconv_only
      REAL(dp)                                           :: max_norm, min_norm, vmax
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: q_mat, tblock, tvblock
      REAL(dp), DIMENSION(:), POINTER                    :: c_res, t_eval
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: v_mat
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: a_mat, b2_mat, b_mat, c2_tmp, c3_tmp, &
                                                            c_tmp, chc, evec, hc, t_mat, t_vec

      CALL timeset(routineN, handle)

      NULLIFY (fm_struct_tmp)
      NULLIFY (chc, evec)
      NULLIFY (c_res, t_eval)
      NULLIFY (hc, t_mat, t_vec, c2_tmp)
      NULLIFY (a_mat, b_mat, b2_mat, v_mat)

      nmo = SIZE(eval, 1)
      my_check_moconv_only = .FALSE.
      IF (PRESENT(check_moconv_only)) my_check_moconv_only = check_moconv_only

      chc => krylov_space%chc_mat(ispin)%matrix
      evec => krylov_space%c_vec(ispin)%matrix
      c_res => krylov_space%c_eval
      t_eval => krylov_space%t_eval

      NULLIFY (fm_struct_tmp, c_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo, &
                               para_env=c0%matrix_struct%para_env, &
                               context=c0%matrix_struct%context)
      CALL cp_fm_create(c_tmp, matrix_struct=fm_struct_tmp, &
                        name="c_tmp")
      CALL cp_fm_create(hc, matrix_struct=fm_struct_tmp, &
                        name="hc")
      CALL cp_fm_struct_release(fm_struct_tmp)

      !Compute (C^t)HC
      CALL cp_gemm('N', 'N', nao, nmo, nao, rone, ks, c0, rzero, hc)
      CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, hc, rzero, chc)

      !Diagonalize  (C^t)HC
      CALL timeset(routineN//"diag_chc", hand1)
      CALL choose_eigv_solver(chc, evec, eval)
      CALL timestop(hand1)

      !Rotate the C vectors
      CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, c0, evec, rzero, c1)

      !Check for converged states
      CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, hc, evec, rzero, c_tmp)
      CALL cp_fm_to_fm(c1, c0, nmo, 1, 1)
      CALL cp_fm_column_scale(c1, eval)
      CALL cp_fm_scale_and_add(1.0_dp, c_tmp, rmone, c1)
      CALL cp_fm_vectorsnorm(c_tmp, c_res)

      nmo_converged = 0
      nmo_nc = 0
      max_norm = 0.0_dp
      min_norm = 1.e10_dp
      CALL cp_fm_set_all(c1, rzero)
      DO imo = 1, nmo
         max_norm = MAX(max_norm, c_res(imo))
         min_norm = MIN(min_norm, c_res(imo))
      END DO
      DO imo = 1, nmo
         IF (c_res(imo) <= eps_iter) THEN
            nmo_converged = nmo_converged+1
         ELSE
            nmo_nc = nmo-nmo_converged
            EXIT
         END IF
      END DO

      nblock = krylov_space%nblock
      num_blocks = nmo_nc/nblock

      krylov_space%nmo_nc = nmo_nc
      krylov_space%nmo_conv = nmo_converged
      krylov_space%max_res_norm = max_norm
      krylov_space%min_res_norm = min_norm

      IF (my_check_moconv_only) THEN
         CALL cp_fm_release(c_tmp)
         CALL cp_fm_release(hc)
         CALL timestop(handle)
         RETURN
      ELSE IF (krylov_space%nmo_nc > 0) THEN

         CALL cp_fm_to_fm(c0, c1, nmo_nc, nmo_converged+1, 1)

         nblock = krylov_space%nblock
         IF (MODULO(nmo_nc, nblock) > 0.0_dp) THEN
            num_blocks = nmo_nc/nblock+1
         ELSE
            num_blocks = nmo_nc/nblock
         END IF

         DO ib = 1, num_blocks

            imo_low = (ib-1)*nblock+1
            imo_up = MIN(ib*nblock, nmo_nc)
            nmob = imo_up-imo_low+1
            ndim = krylov_space%nkrylov*nmob

            NULLIFY (fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=ndim, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context)
            CALL cp_fm_create(c2_tmp, matrix_struct=fm_struct_tmp, &
                              name="c2_tmp")
            CALL cp_fm_struct_release(fm_struct_tmp)
            NULLIFY (fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmob, ncol_global=ndim, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context)
            CALL cp_fm_create(c3_tmp, matrix_struct=fm_struct_tmp, &
                              name="c3_tmp")
            CALL cp_fm_struct_release(fm_struct_tmp)

            ! Create local matrix of right size
            IF (nmob /= nblock) THEN
               NULLIFY (a_mat, b_mat, b2_mat, t_mat, t_vec, v_mat)
               NULLIFY (fm_struct_tmp)
               CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmob, ncol_global=nmob, &
                                        para_env=chc%matrix_struct%para_env, &
                                        context=chc%matrix_struct%context)
               CALL cp_fm_create(a_mat, matrix_struct=fm_struct_tmp, &
                                 name="a_mat")
               CALL cp_fm_create(b_mat, matrix_struct=fm_struct_tmp, &
                                 name="b_mat")
               CALL cp_fm_create(b2_mat, matrix_struct=fm_struct_tmp, &
                                 name="b2_mat")
               CALL cp_fm_struct_release(fm_struct_tmp)
               NULLIFY (fm_struct_tmp)
               CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ndim, ncol_global=ndim, &
                                        para_env=chc%matrix_struct%para_env, &
                                        context=chc%matrix_struct%context)
               CALL cp_fm_create(t_mat, matrix_struct=fm_struct_tmp, &
                                 name="t_mat")
               CALL cp_fm_create(t_vec, matrix_struct=fm_struct_tmp, &
                                 name="t_vec")
               CALL cp_fm_struct_release(fm_struct_tmp)
               NULLIFY (fm_struct_tmp)
               CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmob, &
                                        para_env=c0%matrix_struct%para_env, &
                                        context=c0%matrix_struct%context)
               ALLOCATE (v_mat(krylov_space%nkrylov))
               DO ik = 1, krylov_space%nkrylov
                  CALL cp_fm_create(v_mat(ik)%matrix, matrix_struct=fm_struct_tmp, &
                                    name="v_mat")
               END DO
               CALL cp_fm_struct_release(fm_struct_tmp)
            ELSE
               a_mat => krylov_space%block1_mat
               b_mat => krylov_space%block2_mat
               b2_mat => krylov_space%block3_mat
               t_mat => krylov_space%block4_mat
               t_vec => krylov_space%block5_mat
               v_mat => krylov_space%v_mat
            END IF

            ALLOCATE (tblock(nmob, nmob))
            ALLOCATE (tvblock(nmob, ndim))

            CALL timeset(routineN//"_kry_loop", hand2)
            CALL cp_fm_set_all(b_mat, rzero)
            CALL cp_fm_set_all(t_mat, rzero)
            CALL cp_fm_to_fm(c1, v_mat(1)%matrix, nmob, imo_low, 1)

            !Compute A =(V^t)HV
            CALL cp_gemm('N', 'N', nao, nmob, nao, rone, ks, v_mat(1)%matrix, rzero, hc)
            CALL cp_gemm('T', 'N', nmob, nmob, nao, rone, v_mat(1)%matrix, hc, &
                         rzero, a_mat)

            !Compute the residual matrix R for next
            !factorisation
            CALL cp_gemm('N', 'N', nao, nmob, nmob, rone, v_mat(1)%matrix, a_mat, &
                         rzero, c_tmp)
            CALL cp_fm_scale_and_add(rmone, c_tmp, rone, hc)

            ! Build the block tridiagonal matrix
            CALL cp_fm_get_submatrix(a_mat, tblock, 1, 1, nmob, nmob)
            CALL cp_fm_set_submatrix(t_mat, tblock, 1, 1, nmob, nmob)

            DO ik = 2, krylov_space%nkrylov

               ! Call lapack for QR factorization
               CALL cp_fm_set_all(b_mat, rzero)
               CALL cp_fm_to_fm(c_tmp, v_mat(ik)%matrix, nmob, 1, 1)
               CALL cp_fm_qr_factorization(c_tmp, b_mat, nao, nmob, 1, 1)

               CALL cp_fm_triangular_multiply(b_mat, v_mat(ik)%matrix, side="R", invert_tr=.TRUE., &
                                              n_rows=nao, n_cols=nmob)

               !Compute A =(V^t)HV
               CALL cp_gemm('N', 'N', nao, nmob, nao, rone, ks, v_mat(ik)%matrix, rzero, hc)
               CALL cp_gemm('T', 'N', nmob, nmob, nao, rone, v_mat(ik)%matrix, hc, rzero, a_mat)

               !Compute the !residual matrix R !for next !factorisation
               CALL cp_gemm('N', 'N', nao, nmob, nmob, rone, v_mat(ik)%matrix, a_mat, &
                            rzero, c_tmp)
               CALL cp_fm_scale_and_add(rmone, c_tmp, rone, hc)
               CALL cp_fm_to_fm(v_mat(ik-1)%matrix, hc, nmob, 1, 1)
               CALL cp_fm_triangular_multiply(b_mat, hc, side='R', transpose_tr=.TRUE., &
                                              n_rows=nao, n_cols=nmob, alpha=rmone)
               CALL cp_fm_scale_and_add(rone, c_tmp, rone, hc)

               ! Build the block tridiagonal matrix
               it = (ik-2)*nmob+1
               jt = (ik-1)*nmob+1

               CALL cp_fm_get_submatrix(a_mat, tblock, 1, 1, nmob, nmob)
               CALL cp_fm_set_submatrix(t_mat, tblock, jt, jt, nmob, nmob)
               CALL cp_fm_transpose(b_mat, a_mat)
               CALL cp_fm_get_submatrix(a_mat, tblock, 1, 1, nmob, nmob)
               CALL cp_fm_set_submatrix(t_mat, tblock, it, jt, nmob, nmob)

            END DO ! ik
            CALL timestop(hand2)

            DEALLOCATE (tblock)

            CALL timeset(routineN//"_diag_tri", hand3)

            CALL choose_eigv_solver(t_mat, t_vec, t_eval)
            ! Diagonalize the block-tridiagonal matrix
            CALL timestop(hand3)

            CALL timeset(routineN//"_build_cnew", hand4)
!        !Compute the refined vectors
            CALL cp_fm_set_all(c2_tmp, rzero)
            DO ik = 1, krylov_space%nkrylov
               jt = (ik-1)*nmob
               CALL cp_gemm('N', 'N', nao, ndim, nmob, rone, v_mat(ik)%matrix, t_vec, rone, c2_tmp, &
                            b_first_row=(jt+1))
            END DO
            DEALLOCATE (tvblock)

            CALL cp_fm_set_all(c3_tmp, rzero)
            CALL cp_gemm('T', 'N', nmob, ndim, nao, rone, v_mat(1)%matrix, c2_tmp, rzero, c3_tmp)

            !Try to avoid linear dependencies
            ALLOCATE (q_mat(nmob, ndim))
            !get max
            CALL cp_fm_get_submatrix(c3_tmp, q_mat, 1, 1, nmob, ndim)

            ALLOCATE (itaken(ndim))
            itaken = 0
            DO it = 1, nmob
               vmax = 0.0_dp
               !select index ik
               DO jt = 1, ndim
                  IF (itaken(jt) == 0 .AND. ABS(q_mat(it, jt)) > vmax) THEN
                     vmax = ABS(q_mat(it, jt))
                     ik = jt
                  END IF
               END DO
               itaken(ik) = 1

               CALL cp_fm_to_fm(c2_tmp, v_mat(1)%matrix, 1, ik, it)
            END DO
            DEALLOCATE (itaken)
            DEALLOCATE (q_mat)

            !Copy in the converged set to enlarge the converged subspace
            CALL cp_fm_to_fm(v_mat(1)%matrix, c0, nmob, 1, (nmo_converged+imo_low))
            CALL timestop(hand4)

            IF (nmob < nblock) THEN
               CALL cp_fm_release(a_mat)
               CALL cp_fm_release(b_mat)
               CALL cp_fm_release(b2_mat)
               CALL cp_fm_release(t_mat)
               CALL cp_fm_release(t_vec)
               DO ik = 1, krylov_space%nkrylov
                  CALL cp_fm_release(v_mat(ik)%matrix)
               END DO
               DEALLOCATE (v_mat)

            END IF
            CALL cp_fm_release(c2_tmp)
            CALL cp_fm_release(c3_tmp)
         END DO ! ib

         CALL timeset(routineN//"_ortho", hand5)
         CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, c0, rzero, chc)

         CALL cp_fm_cholesky_decompose(chc, nmo)
         CALL cp_fm_triangular_multiply(chc, c0, 'R', invert_tr=.TRUE.)
         CALL timestop(hand5)

         CALL cp_fm_release(c_tmp)
         CALL cp_fm_release(hc)
      ELSE
         CALL cp_fm_release(c_tmp)
         CALL cp_fm_release(hc)
         CALL timestop(handle)
         RETURN
      END IF

      CALL timestop(handle)

   END SUBROUTINE lanczos_refinement

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

! **************************************************************************************************
!> \brief ...
!> \param krylov_space ...
!> \param ks ...
!> \param c0 ...
!> \param c1 ...
!> \param eval ...
!> \param nao ...
!> \param eps_iter ...
!> \param ispin ...
!> \param check_moconv_only ...
! **************************************************************************************************
   SUBROUTINE lanczos_refinement_2v(krylov_space, ks, c0, c1, eval, nao, &
                                    eps_iter, ispin, check_moconv_only)

      TYPE(krylov_space_type), POINTER                   :: krylov_space
      TYPE(cp_fm_type), POINTER                          :: ks, c0, c1
      REAL(dp), DIMENSION(:), POINTER                    :: eval
      INTEGER, INTENT(IN)                                :: nao
      REAL(dp), INTENT(IN)                               :: eps_iter
      INTEGER, INTENT(IN)                                :: ispin
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_moconv_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lanczos_refinement_2v', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rmone = -1.0_dp, rone = 1.0_dp, &
                                                            rzero = 0.0_dp

      INTEGER :: hand1, hand2, hand3, hand4, hand5, hand6, handle, i, ia, ib, ik, imo, imo_low, &
         imo_up, info, it, j, jt, liwork, lwork, nblock, ndim, nmo, nmo_converged, nmo_nc, nmob, &
         num_blocks
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: itaken
      INTEGER, DIMENSION(:), POINTER                     :: iwork
      LOGICAL                                            :: my_check_moconv_only
      REAL(dp)                                           :: max_norm, min_norm, vmax
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: a_block, b_block, q_mat, t_mat
      REAL(dp), DIMENSION(:), POINTER                    :: c_res, t_eval
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a_loc, b_loc
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: v_mat
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: b_mat, c2_tmp, c_tmp, chc, evec, hc, &
                                                            v_tmp

      CALL timeset(routineN, handle)

      NULLIFY (fm_struct_tmp)
      NULLIFY (chc, evec)
      NULLIFY (c_res, t_eval)
      NULLIFY (hc, c_tmp, c2_tmp)
      NULLIFY (v_tmp)
      NULLIFY (b_mat, v_mat)
      NULLIFY (b_loc, a_loc)

      nmo = SIZE(eval, 1)
      my_check_moconv_only = .FALSE.
      IF (PRESENT(check_moconv_only)) my_check_moconv_only = check_moconv_only

      chc => krylov_space%chc_mat(ispin)%matrix
      evec => krylov_space%c_vec(ispin)%matrix
      c_res => krylov_space%c_eval
      t_eval => krylov_space%t_eval

      NULLIFY (fm_struct_tmp, c_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo, &
                               para_env=c0%matrix_struct%para_env, &
                               context=c0%matrix_struct%context)
      CALL cp_fm_create(c_tmp, matrix_struct=fm_struct_tmp, &
                        name="c_tmp")
      CALL cp_fm_create(hc, matrix_struct=fm_struct_tmp, &
                        name="hc")
      CALL cp_fm_struct_release(fm_struct_tmp)

      !Compute (C^t)HC
      CALL cp_gemm('N', 'N', nao, nmo, nao, rone, ks, c0, rzero, hc)
      CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, hc, rzero, chc)

      !Diagonalize  (C^t)HC
      CALL timeset(routineN//"diag_chc", hand1)
      CALL choose_eigv_solver(chc, evec, eval)

      CALL timestop(hand1)

      CALL timeset(routineN//"check_conv", hand6)
      !Rotate the C vectors
      CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, c0, evec, rzero, c1)

      !Check for converged states
      CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, hc, evec, rzero, c_tmp)
      CALL cp_fm_to_fm(c1, c0, nmo, 1, 1)
      CALL cp_fm_column_scale(c1, eval)
      CALL cp_fm_scale_and_add(1.0_dp, c_tmp, rmone, c1)
      CALL cp_fm_vectorsnorm(c_tmp, c_res)

      nmo_converged = 0
      nmo_nc = 0
      max_norm = 0.0_dp
      min_norm = 1.e10_dp
      CALL cp_fm_set_all(c1, rzero)
      DO imo = 1, nmo
         max_norm = MAX(max_norm, c_res(imo))
         min_norm = MIN(min_norm, c_res(imo))
      END DO
      DO imo = 1, nmo
         IF (c_res(imo) <= eps_iter) THEN
            nmo_converged = nmo_converged+1
         ELSE
            nmo_nc = nmo-nmo_converged
            EXIT
         END IF
      END DO
      CALL timestop(hand6)

      CALL cp_fm_release(c_tmp)
      CALL cp_fm_release(hc)

      krylov_space%nmo_nc = nmo_nc
      krylov_space%nmo_conv = nmo_converged
      krylov_space%max_res_norm = max_norm
      krylov_space%min_res_norm = min_norm

      IF (my_check_moconv_only) THEN
         ! Do nothing
      ELSE IF (krylov_space%nmo_nc > 0) THEN

         CALL cp_fm_to_fm(c0, c1, nmo_nc, nmo_converged+1, 1)

         nblock = krylov_space%nblock
         IF (MODULO(nmo_nc, nblock) > 0.0_dp) THEN
            num_blocks = nmo_nc/nblock+1
         ELSE
            num_blocks = nmo_nc/nblock
         END IF

         DO ib = 1, num_blocks

            imo_low = (ib-1)*nblock+1
            imo_up = MIN(ib*nblock, nmo_nc)
            nmob = imo_up-imo_low+1
            ndim = krylov_space%nkrylov*nmob

            ! Allocation
            CALL timeset(routineN//"alloc", hand6)
            ALLOCATE (a_block(nmob, nmob))
            ALLOCATE (b_block(nmob, nmob))
            ALLOCATE (t_mat(ndim, ndim))

            NULLIFY (fm_struct_tmp)
            ! by forcing ncol_block=nmo, the needed part of the matrix is distributed on a subset of processes
            ! this is due to the use of two-dimensional grids of processes
            ! nrow_global is distributed over num_pe(1)
            ! a local_data array is anyway allocated for the processes non included
            ! this should have a minimum size
            ! with ncol_local=1.
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmob, &
                                     ncol_block=nmob, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context, &
                                     force_block=.TRUE.)
            CALL cp_fm_create(c_tmp, matrix_struct=fm_struct_tmp, &
                              name="c_tmp")
            CALL cp_fm_set_all(c_tmp, rzero)
            CALL cp_fm_create(v_tmp, matrix_struct=fm_struct_tmp, &
                              name="v_tmp")
            CALL cp_fm_set_all(v_tmp, rzero)
            CALL cp_fm_struct_release(fm_struct_tmp)
            NULLIFY (fm_struct_tmp)
            ALLOCATE (v_mat(krylov_space%nkrylov))
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmob, &
                                     ncol_block=nmob, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context, &
                                     force_block=.TRUE.)
            DO ik = 1, krylov_space%nkrylov
               CALL cp_fm_create(v_mat(ik)%matrix, matrix_struct=fm_struct_tmp, &
                                 name="v_mat")
            END DO
            CALL cp_fm_create(hc, matrix_struct=fm_struct_tmp, &
                              name="hc")
            CALL cp_fm_struct_release(fm_struct_tmp)
            NULLIFY (fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=ndim, &
                                     ncol_block=ndim, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context, &
                                     force_block=.TRUE.)
            CALL cp_fm_create(c2_tmp, matrix_struct=fm_struct_tmp, &
                              name="c2_tmp")
            CALL cp_fm_struct_release(fm_struct_tmp)

            NULLIFY (fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmob, ncol_global=nmob, &
                                     para_env=c0%matrix_struct%para_env, &
                                     context=c0%matrix_struct%context)
            CALL cp_fm_create(b_mat, matrix_struct=fm_struct_tmp, &
                              name="b_mat")
            CALL cp_fm_struct_release(fm_struct_tmp)
            CALL timestop(hand6)
            !End allocation

            CALL cp_fm_set_all(b_mat, rzero)
            CALL cp_fm_to_fm(c1, v_mat(1)%matrix, nmob, imo_low, 1)

            ! Here starts the construction of krylov space
            CALL timeset(routineN//"_kry_loop", hand2)
            !Compute A =(V^t)HV
            CALL cp_gemm('N', 'N', nao, nmob, nao, rone, ks, v_mat(1)%matrix, rzero, hc)

            a_block = 0.0_dp
            a_loc => v_mat(1)%matrix%local_data
            b_loc => hc%local_data

            IF (SIZE(hc%local_data, 2) == nmob) THEN
               ! this is a work around to avoid problems due to the two dimensional grid of processes
               CALL dgemm('T', 'N', nmob, nmob, SIZE(hc%local_data, 1), 1.0_dp, a_loc(1, 1), &
                          SIZE(hc%local_data, 1), b_loc(1, 1), SIZE(hc%local_data, 1), 0.0_dp, a_block(1, 1), nmob)
            END IF
            CALL mp_sum(a_block, hc%matrix_struct%para_env%group)

            !Compute the residual matrix R for next
            !factorisation
            c_tmp%local_data = 0.0_dp
            IF (SIZE(c_tmp%local_data, 2) == nmob) THEN
               a_loc => v_mat(1)%matrix%local_data
               b_loc => c_tmp%local_data
               CALL dgemm('N', 'N', SIZE(c_tmp%local_data, 1), nmob, nmob, 1.0_dp, a_loc(1, 1), &
                          SIZE(c_tmp%local_data, 1), a_block(1, 1), nmob, 0.0_dp, &
                          b_loc(1, 1), SIZE(c_tmp%local_data, 1))
            END IF
            CALL cp_fm_scale_and_add(rmone, c_tmp, rone, hc)

            ! Build the block tridiagonal matrix
            t_mat = 0.0_dp
            DO i = 1, nmob
               t_mat(1:nmob, i) = a_block(1:nmob, i)
            END DO

            DO ik = 2, krylov_space%nkrylov
               ! Call lapack for QR factorization
               CALL cp_fm_set_all(b_mat, rzero)
               CALL cp_fm_to_fm(c_tmp, v_mat(ik)%matrix, nmob, 1, 1)
               CALL cp_fm_qr_factorization(c_tmp, b_mat, nao, nmob, 1, 1)

               CALL cp_fm_triangular_multiply(b_mat, v_mat(ik)%matrix, side="R", invert_tr=.TRUE., &
                                              n_rows=nao, n_cols=nmob)

               CALL cp_fm_get_submatrix(b_mat, b_block, 1, 1, nmob, nmob)

               !Compute A =(V^t)HV
               CALL cp_gemm('N', 'N', nao, nmob, nao, rone, ks, v_mat(ik)%matrix, rzero, hc)

               a_block = 0.0_dp
               IF (SIZE(hc%local_data, 2) == nmob) THEN
                  a_loc => v_mat(ik)%matrix%local_data
                  b_loc => hc%local_data
                  CALL dgemm('T', 'N', nmob, nmob, SIZE(hc%local_data, 1), 1.0_dp, a_loc(1, 1), &
                             SIZE(hc%local_data, 1), b_loc(1, 1), SIZE(hc%local_data, 1), 0.0_dp, a_block(1, 1), nmob)
               END IF
               CALL mp_sum(a_block, hc%matrix_struct%para_env%group)

               !Compute the residual matrix R for next
               !factorisation
               c_tmp%local_data = 0.0_dp
               IF (SIZE(c_tmp%local_data, 2) == nmob) THEN
                  a_loc => v_mat(ik)%matrix%local_data
                  b_loc => c_tmp%local_data
                  CALL dgemm('N', 'N', SIZE(c_tmp%local_data, 1), nmob, nmob, 1.0_dp, a_loc(1, 1), &
                             SIZE(c_tmp%local_data, 1), a_block(1, 1), nmob, 0.0_dp, &
                             b_loc(1, 1), SIZE(c_tmp%local_data, 1))
               END IF
               CALL cp_fm_scale_and_add(rmone, c_tmp, rone, hc)

               IF (SIZE(c_tmp%local_data, 2) == nmob) THEN
                  a_loc => v_mat(ik-1)%matrix%local_data
                  DO j = 1, nmob
                     DO i = 1, j
                        DO ia = 1, SIZE(c_tmp%local_data, 1)
                           b_loc(ia, i) = b_loc(ia, i)-a_loc(ia, j)*b_block(i, j)
                        END DO
                     END DO
                  END DO
               END IF

               ! Build the block tridiagonal matrix
               it = (ik-2)*nmob
               jt = (ik-1)*nmob
               DO j = 1, nmob
                  t_mat(jt+1:jt+nmob, jt+j) = a_block(1:nmob, j)
                  DO i = 1, nmob
                     t_mat(it+i, jt+j) = b_block(j, i)
                     t_mat(jt+j, it+i) = b_block(j, i)
                  END DO
               END DO
            END DO ! ik
            CALL timestop(hand2)

            CALL timeset(routineN//"_diag_tri", hand3)
            lwork = 1+6*ndim+2*ndim**2+5000
            liwork = 5*ndim+3
            ALLOCATE (work(lwork))
            ALLOCATE (iwork(liwork))

            ! Diagonalize the block-tridiagonal matrix
            CALL dsyevd('V', 'U', ndim, t_mat(1, 1), ndim, t_eval(1), &
                        work(1), lwork, iwork(1), liwork, info)
            DEALLOCATE (work)
            DEALLOCATE (iwork)
            CALL timestop(hand3)

            CALL timeset(routineN//"_build_cnew", hand4)
!        !Compute the refined vectors

            c2_tmp%local_data = 0.0_dp
            ALLOCATE (q_mat(nmob, ndim))
            q_mat = 0.0_dp
            b_loc => c2_tmp%local_data
            DO ik = 1, krylov_space%nkrylov
               CALL cp_fm_to_fm(v_mat(ik)%matrix, v_tmp, nmob, 1, 1)
               IF (SIZE(c2_tmp%local_data, 2) == ndim) THEN
!            a_loc => v_mat(ik)%matrix%local_data
                  a_loc => v_tmp%local_data
                  it = (ik-1)*nmob
                  CALL dgemm('N', 'N', SIZE(c2_tmp%local_data, 1), ndim, nmob, 1.0_dp, a_loc(1, 1), &
                             SIZE(c2_tmp%local_data, 1), t_mat(it+1, 1), ndim, 1.0_dp, &
                             b_loc(1, 1), SIZE(c2_tmp%local_data, 1))
               END IF
            END DO !ik

            !Try to avoid linear dependencies
            CALL cp_fm_to_fm(v_mat(1)%matrix, v_tmp, nmob, 1, 1)
            IF (SIZE(c2_tmp%local_data, 2) == ndim) THEN
!          a_loc => v_mat(1)%matrix%local_data
               a_loc => v_tmp%local_data
               b_loc => c2_tmp%local_data
               CALL dgemm('T', 'N', nmob, ndim, SIZE(v_tmp%local_data, 1), 1.0_dp, a_loc(1, 1), &
                          SIZE(v_tmp%local_data, 1), b_loc(1, 1), SIZE(v_tmp%local_data, 1), &
                          0.0_dp, q_mat(1, 1), nmob)
            END IF
            CALL mp_sum(q_mat, hc%matrix_struct%para_env%group)

            ALLOCATE (itaken(ndim))
            itaken = 0
            DO it = 1, nmob
               vmax = 0.0_dp
               !select index ik
               DO jt = 1, ndim
                  IF (itaken(jt) == 0 .AND. ABS(q_mat(it, jt)) > vmax) THEN
                     vmax = ABS(q_mat(it, jt))
                     ik = jt
                  END IF
               END DO
               itaken(ik) = 1

               CALL cp_fm_to_fm(c2_tmp, v_mat(1)%matrix, 1, ik, it)
            END DO
            DEALLOCATE (itaken)
            DEALLOCATE (q_mat)

            !Copy in the converged set to enlarge the converged subspace
            CALL cp_fm_to_fm(v_mat(1)%matrix, c0, nmob, 1, (nmo_converged+imo_low))
            CALL timestop(hand4)

            CALL cp_fm_release(c2_tmp)
            CALL cp_fm_release(c_tmp)
            CALL cp_fm_release(hc)
            CALL cp_fm_release(v_tmp)
            CALL cp_fm_release(b_mat)

            DEALLOCATE (t_mat)
            DEALLOCATE (a_block)
            DEALLOCATE (b_block)
            DO ik = 1, krylov_space%nkrylov
               CALL cp_fm_release(v_mat(ik)%matrix)
            END DO
            DEALLOCATE (v_mat)

         END DO ! ib

         CALL timeset(routineN//"_ortho", hand5)
         CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, c0, rzero, chc)

         CALL cp_fm_cholesky_decompose(chc, nmo)
         CALL cp_fm_triangular_multiply(chc, c0, 'R', invert_tr=.TRUE.)
         CALL timestop(hand5)
      ELSE
         ! Do nothing
      END IF

      CALL timestop(handle)
   END SUBROUTINE lanczos_refinement_2v

END MODULE qs_scf_lanczos
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qs_gspace_mixing

   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_create, &
        cp_dbcsr_distribution, cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, &
        cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_symmetric
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all,&
                                              invert_matrix
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_max,&
                                              mp_min,&
                                              mp_sum
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_density_mixing_types,         ONLY: broyden_mixing_new_nr,&
                                              broyden_mixing_nr,&
                                              cp_1d_z_p_type,&
                                              gspace_mixing_nr,&
                                              mixing_storage_type,&
                                              multisecant_mixing_nr,&
                                              pulay_mixing_nr
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_methods,                  ONLY: cp_sm_mix
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_gspace_mixing'

   PUBLIC :: gspace_mixing, &
             mixing_allocate, mixing_init, &
             self_consistency_check

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param rho_ao ...
!> \param p_delta ...
!> \param para_env ...
!> \param p_out ...
!> \param delta ...
! **************************************************************************************************
   SUBROUTINE self_consistency_check(rho_ao, p_delta, para_env, p_out, delta)
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao, p_delta
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: p_out
      REAL(KIND=dp), INTENT(INOUT)                       :: delta

      CHARACTER(len=*), PARAMETER :: routineN = 'self_consistency_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, ispin, nimg, nspins
      REAL(KIND=dp)                                      :: tmp
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_q, p_in

      CALL timeset(routineN, handle)

      NULLIFY (matrix_q, p_in)

      IF(.NOT.(ASSOCIATED(p_out)))CALL cp__a("qs_gspace_mixing.F",90)
      NULLIFY (matrix_q, p_in)
      p_in => rho_ao
      matrix_q => p_delta
      nspins = SIZE(p_in, 1)
      nimg = SIZE(p_in, 2)

      ! Compute the difference (p_out - p_in)and check convergence
      delta = 0.0_dp
      DO ispin = 1, nspins
         DO ic = 1, nimg
            CALL cp_dbcsr_set(matrix_q(ispin, ic)%matrix, 0.0_dp)
            CALL cp_sm_mix(m1=p_out(ispin, ic)%matrix, m2=p_in(ispin, ic)%matrix, &
                           p_mix=1.0_dp, delta=tmp, para_env=para_env, &
                           m3=matrix_q(ispin, ic)%matrix)
            delta = MAX(tmp, delta)
         END DO
      END DO
      CALL timestop(handle)

   END SUBROUTINE self_consistency_check

! **************************************************************************************************
!> \brief  Driver for the g-space mixing, calls the proper routine given the
!>requested method
!> \param qs_env ...
!> \param mixing_method ...
!> \param mixing_store ...
!> \param rho ...
!> \param para_env ...
!> \param iter_count ...
!> \par History
!>      05.2009
!>      02.2015 changed input to make g-space mixing available in linear scaling
!>              SCF [Patrick Seewald]
!> \author MI
! **************************************************************************************************
   SUBROUTINE gspace_mixing(qs_env, mixing_method, mixing_store, rho, para_env, iter_count)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: mixing_method
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: iter_count

      CHARACTER(len=*), PARAMETER :: routineN = 'gspace_mixing', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ig, ispin, natom, ng, &
                                                            nspin
      LOGICAL                                            :: gapw
      REAL(dp)                                           :: alpha
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho_tmp
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(mixing_store)))CALL cp__a("qs_gspace_mixing.F",151)
      NULLIFY (auxbas_pw_pool, dft_control, pw_env, rho_atom, rho_g, rho_r, tot_rho_r)

      CALL get_qs_env(qs_env, dft_control=dft_control, pw_env=pw_env)
      CALL qs_rho_get(rho, rho_g=rho_g, rho_r=rho_r, tot_rho_r=tot_rho_r)

      nspin = SIZE(rho_g, 1)
      ng = SIZE(rho_g(1)%pw%pw_grid%gsq)
      IF(.NOT.((ng == SIZE(mixing_store%rhoin(1)%cc))))CALL cp__a("qs_gspace_mixing.F",159)

      alpha = mixing_store%alpha
      gapw = dft_control%qs_control%gapw

      IF (nspin == 2) THEN
         CALL pw_env_get(pw_env=pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, &
                                rho_tmp%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         CALL pw_zero(rho_tmp%pw)
         CALL pw_copy(rho_g(1)%pw, rho_tmp%pw)
         CALL pw_axpy(rho_g(2)%pw, rho_g(1)%pw, 1.0_dp)
         CALL pw_axpy(rho_tmp%pw, rho_g(2)%pw, -1.0_dp)
         CALL pw_scale(rho_g(2)%pw, -1.0_dp)
      END IF

      IF (iter_count+1 <= mixing_store%nskip_mixing) THEN
         ! skip mixing
         DO ispin = 1, nspin
            DO ig = 1, ng
!          mixing_store%rhoin_buffer(1,ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
               mixing_store%rhoin(ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
            END DO
         END DO
         IF (gapw) THEN
            CALL get_qs_env(qs_env=qs_env, &
                            rho_atom_set=rho_atom)
            natom = SIZE(rho_atom)
            DO ispin = 1, nspin
               DO iatom = 1, natom
                  IF (mixing_store%paw(iatom)) THEN
                     mixing_store%cpc_h_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_h(ispin)%r_coef
                     mixing_store%cpc_s_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_s(ispin)%r_coef
                  END IF
               END DO
            END DO
         END IF
         mixing_store%iter_method = "NoMix"
         IF (nspin == 2) THEN
            CALL pw_axpy(rho_g(2)%pw, rho_g(1)%pw, 1.0_dp)
            CALL pw_scale(rho_g(1)%pw, 0.5_dp)
            CALL pw_axpy(rho_g(1)%pw, rho_g(2)%pw, -1.0_dp)
            CALL pw_scale(rho_g(2)%pw, -1.0_dp)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tmp%pw)
         END IF
         CALL timestop(handle)
         RETURN
      END IF

      IF ((iter_count+1-mixing_store%nskip_mixing) <= mixing_store%n_simple_mix) THEN
         CALL gmix_potential_only(qs_env, mixing_store, rho)
         mixing_store%iter_method = "Kerker"
      ELSE

         IF (mixing_method == gspace_mixing_nr) THEN
            CALL gmix_potential_only(qs_env, mixing_store, rho)
            mixing_store%iter_method = "Kerker"
         ELSEIF (mixing_method == pulay_mixing_nr) THEN
            IF(.NOT.(.NOT. gapw))CALL cp__a("qs_gspace_mixing.F",219)
            CALL pulay_mixing(mixing_store, rho, para_env)
            mixing_store%iter_method = "Pulay"
         ELSEIF (mixing_method == broyden_mixing_nr) THEN
            CALL broyden_mixing(qs_env, mixing_store, rho, para_env)
            mixing_store%iter_method = "Broy."
         ELSEIF (mixing_method == broyden_mixing_new_nr) THEN
            IF(.NOT.(.NOT. gapw))CALL cp__a("qs_gspace_mixing.F",226)
            CALL broyden_mixing_new(mixing_store, rho, para_env)
            mixing_store%iter_method = "Broy."
         ELSEIF (mixing_method == multisecant_mixing_nr) THEN
            IF(.NOT.(.NOT. gapw))CALL cp__a("qs_gspace_mixing.F",230)
            CALL multisecant_mixing(mixing_store, rho, para_env)
            mixing_store%iter_method = "MSec."
         END IF
      END IF

      IF (nspin == 2) THEN
         CALL pw_axpy(rho_g(2)%pw, rho_g(1)%pw, 1.0_dp)
         CALL pw_scale(rho_g(1)%pw, 0.5_dp)
         CALL pw_axpy(rho_g(1)%pw, rho_g(2)%pw, -1.0_dp)
         CALL pw_scale(rho_g(2)%pw, -1.0_dp)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tmp%pw)
      END IF

      DO ispin = 1, nspin
         CALL pw_transfer(rho_g(ispin)%pw, rho_r(ispin)%pw)
         tot_rho_r(ispin) = pw_integrate_function(rho_r(ispin)%pw, isign=-1)
!dbg
!  write(*,*) 'rho int 4', ispin , tot_rho_r(ispin)
!dbg
      END DO

      CALL timestop(handle)

   END SUBROUTINE gspace_mixing

! **************************************************************************************************
!> \brief G-space mixing performed via the Kerker damping on the density on the grid
!>        thus affecting only the caluclation of the potential, but not the denisity matrix
!> \param qs_env ...
!> \param mixing_store ...
!> \param rho ...
!> \par History
!>      02.2009
!> \author MI
! **************************************************************************************************

   SUBROUTINE gmix_potential_only(qs_env, mixing_store, rho)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho

      CHARACTER(len=*), PARAMETER :: routineN = 'gmix_potential_only', &
         routineP = moduleN//':'//routineN

      COMPLEX(dp), DIMENSION(:), POINTER                 :: cc_new
      INTEGER                                            :: handle, iatom, ig, ispin, natom, ng, &
                                                            nspin
      LOGICAL                                            :: gapw
      REAL(dp)                                           :: alpha, f_mix
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      IF(.NOT.(ASSOCIATED(mixing_store%rhoin)))CALL cp__a("qs_gspace_mixing.F",285)
      IF(.NOT.(ASSOCIATED(mixing_store%kerker_factor)))CALL cp__a("qs_gspace_mixing.F",286)

      CALL timeset(routineN, handle)
      NULLIFY (cc_new, dft_control, rho_g)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      CALL qs_rho_get(rho, rho_g=rho_g)

      nspin = SIZE(rho_g, 1)
      ng = SIZE(rho_g(1)%pw%pw_grid%gsq)

      gapw = dft_control%qs_control%gapw
      alpha = mixing_store%alpha

      DO ispin = 1, nspin
         cc_new => rho_g(ispin)%pw%cc
         DO ig = 1, mixing_store%ig_max ! ng
            f_mix = mixing_store%alpha*mixing_store%kerker_factor(ig)
            cc_new(ig) = (1.0_dp-f_mix)*mixing_store%rhoin(ispin)%cc(ig)+f_mix*cc_new(ig)
            mixing_store%rhoin(ispin)%cc(ig) = cc_new(ig)
         END DO
         DO ig = mixing_store%ig_max+1, ng
            f_mix = mixing_store%alpha
            cc_new(ig) = (1.0_dp-f_mix)*mixing_store%rhoin(ispin)%cc(ig)+f_mix*cc_new(ig)
            mixing_store%rhoin(ispin)%cc(ig) = cc_new(ig)
         END DO
      END DO

      IF (gapw) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         rho_atom_set=rho_atom)
         natom = SIZE(rho_atom)
         DO ispin = 1, nspin
            DO iatom = 1, natom
               IF (mixing_store%paw(iatom)) THEN
                  rho_atom(iatom)%cpc_h(ispin)%r_coef = alpha*rho_atom(iatom)%cpc_h(ispin)%r_coef+ &
                                                        mixing_store%cpc_h_in(iatom, ispin)%r_coef*(1._dp-alpha)
                  rho_atom(iatom)%cpc_s(ispin)%r_coef = alpha*rho_atom(iatom)%cpc_s(ispin)%r_coef+ &
                                                        mixing_store%cpc_s_in(iatom, ispin)%r_coef*(1._dp-alpha)
                  mixing_store%cpc_h_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_h(ispin)%r_coef
                  mixing_store%cpc_s_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_s(ispin)%r_coef
               END IF
            END DO
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE gmix_potential_only

! **************************************************************************************************
!> \brief Pulay Mixing using as metrics for the residual the Kerer damping factor
!>        The mixing is applied directly on the density in reciprocal space,
!>        therefore it affects the potentials
!>        on the grid but not the density matrix
!> \param mixing_store ...
!> \param rho ...
!> \param para_env ...
!> \par History
!>      03.2009
!> \author MI
! **************************************************************************************************

   SUBROUTINE pulay_mixing(mixing_store, rho, para_env)

      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pulay_mixing', routineP = moduleN//':'//routineN

      COMPLEX(dp), ALLOCATABLE, DIMENSION(:)             :: cc_mix
      INTEGER                                            :: handle, i, ib, ibb, ig, ispin, jb, nb, &
                                                            nb1, nbuffer, ng, nspin
      REAL(dp)                                           :: alpha_kerker, alpha_pulay, beta, f_mix, &
                                                            inv_err, norm, norm_c_inv, vol
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: alpha_c, ev
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: a, b, c, c_inv
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g

      IF(.NOT.(ASSOCIATED(mixing_store%res_buffer)))CALL cp__a("qs_gspace_mixing.F",366)
      IF(.NOT.(ASSOCIATED(mixing_store%rhoin_buffer)))CALL cp__a("qs_gspace_mixing.F",367)
      NULLIFY (rho_g)
      CALL timeset(routineN, handle)

      CALL qs_rho_get(rho, rho_g=rho_g)
      nspin = SIZE(rho_g, 1)
      ng = SIZE(mixing_store%res_buffer(1, 1)%cc)
      vol = rho_g(1)%pw%pw_grid%vol

      alpha_kerker = mixing_store%alpha
      beta = mixing_store%beta
      alpha_pulay = mixing_store%pulay_alpha
      nbuffer = mixing_store%nbuffer

      ib = MODULO(mixing_store%ncall, nbuffer)+1
      mixing_store%ncall = mixing_store%ncall+1
      nb = MIN(mixing_store%ncall, nbuffer)
      ibb = MODULO(mixing_store%ncall, nbuffer)+1

      nb1 = nb+1
      ALLOCATE (a(nb1, nb1))
      a = 0.0_dp
      ALLOCATE (b(nb1, nb1))
      b = 0.0_dp
      ALLOCATE (c(nb, nb))
      c = 0.0_dp
      ALLOCATE (c_inv(nb, nb))
      c_inv = 0.0_dp
      ALLOCATE (alpha_c(nb))
      alpha_c = 0.0_dp
      norm_c_inv = 0.0_dp
      ALLOCATE (ev(nb1))
      ev = 0.0_dp
      ALLOCATE (cc_mix(ng))

      DO ispin = 1, nspin
         mixing_store%res_buffer(ib, ispin)%cc(:) = CMPLX(0._dp, 0._dp, KIND=dp)
         norm = 0.0_dp
         IF (nb == 1) mixing_store%rhoin_buffer(1, ispin)%cc = mixing_store%rhoin(ispin)%cc
         DO ig = 1, ng
            f_mix = mixing_store%kerker_factor(ig)
            !         f_mix = 1.0_dp
            mixing_store%res_buffer(ib, ispin)%cc(ig) = f_mix*(rho_g(ispin)%pw%cc(ig)- &
                                                               mixing_store%rhoin_buffer(ib, ispin)%cc(ig))
         END DO

         DO jb = 1, nb
            mixing_store%pulay_matrix(jb, ib) = 0.0_dp
            DO ig = 1, ng
!           f_mix = 1.0_dp*mixing_store%kerker_factor(ig)
               f_mix = mixing_store%special_metric(ig)
               mixing_store%pulay_matrix(jb, ib) = mixing_store%pulay_matrix(jb, ib)+ &
                                                   f_mix*(REAL(mixing_store%res_buffer(jb, ispin)%cc(ig), dp) &
                                                          *REAL(mixing_store%res_buffer(ib, ispin)%cc(ig), dp)+ &
                                                          AIMAG(mixing_store%res_buffer(jb, ispin)%cc(ig))* &
                                                          AIMAG(mixing_store%res_buffer(ib, ispin)%cc(ig)))
            END DO
            CALL mp_sum(mixing_store%pulay_matrix(jb, ib), para_env%group)
            mixing_store%pulay_matrix(jb, ib) = mixing_store%pulay_matrix(jb, ib)*vol*vol*100.0_dp
            mixing_store%pulay_matrix(ib, jb) = mixing_store%pulay_matrix(jb, ib)
         END DO

         IF (nb == 1) THEN
            DO ig = 1, ng
               f_mix = alpha_kerker*mixing_store%kerker_factor(ig)
               cc_mix(ig) = rho_g(ispin)%pw%cc(ig)- &
                            mixing_store%rhoin_buffer(ib, ispin)%cc(ig)
               rho_g(ispin)%pw%cc(ig) = f_mix*cc_mix(ig)+ &
                                        mixing_store%rhoin_buffer(ib, ispin)%cc(ig)
               mixing_store%rhoin_buffer(ibb, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
            END DO
         ELSE

            b(1:nb, 1:nb) = mixing_store%pulay_matrix(1:nb, 1:nb)
            c(1:nb, 1:nb) = b(1:nb, 1:nb)
            b(nb1, 1:nb) = 1.0_dp
            b(1:nb, nb1) = 1.0_dp
            b(nb1, nb1) = 0.0_dp

            CALL diamat_all(b(1:nb1, 1:nb1), ev(1:nb1))
            CALL invert_matrix(c, c_inv, inv_err)
            alpha_c = 0.0_dp
            DO i = 1, nb
               DO jb = 1, nb
                  alpha_c(i) = alpha_c(i)+c_inv(jb, i)
                  norm_c_inv = norm_c_inv+c_inv(jb, i)
               END DO
            END DO
            alpha_c(1:nb) = alpha_c(1:nb)/norm_c_inv

            a(1:nb1, 1:nb1) = b(1:nb1, 1:nb1)
            DO jb = 1, nb1
               IF (ABS(ev(jb)) < 1.E-5_dp) THEN
                  a(1:nb1, jb) = 0.0_dp
               ELSE
                  a(1:nb1, jb) = a(1:nb1, jb)/ev(jb)
               END IF
            END DO
            ev(1:nb) = MATMUL(a(1:nb, 1:nb1), b(nb1, 1:nb1))

            cc_mix = CMPLX(0._dp, 0._dp, KIND=dp)
            DO jb = 1, nb
               DO ig = 1, ng
                  cc_mix(ig) = cc_mix(ig)+ &
                               alpha_c(jb)*(mixing_store%rhoin_buffer(jb, ispin)%cc(ig)+ &
                                            mixing_store%pulay_beta*mixing_store%res_buffer(jb, ispin)%cc(ig))
               END DO
            END DO
            mixing_store%rhoin_buffer(ibb, ispin)%cc = CMPLX(0._dp, 0._dp, KIND=dp)
            IF (alpha_pulay > 0.0_dp) THEN
               DO ig = 1, ng
                  f_mix = alpha_pulay*mixing_store%kerker_factor(ig)
                  rho_g(ispin)%pw%cc(ig) = f_mix*rho_g(ispin)%pw%cc(ig)+ &
                                           (1.0_dp-f_mix)*cc_mix(ig)
                  mixing_store%rhoin_buffer(ibb, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
               END DO
            ELSE
               DO ig = 1, ng
                  rho_g(ispin)%pw%cc(ig) = cc_mix(ig)
                  mixing_store%rhoin_buffer(ibb, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
               END DO
            END IF

         END IF ! nb==1
      END DO ! ispin

      DEALLOCATE (a)
      DEALLOCATE (b)
      DEALLOCATE (ev)
      DEALLOCATE (cc_mix)
      DEALLOCATE (c, c_inv, alpha_c)

      CALL timestop(handle)

   END SUBROUTINE pulay_mixing

! **************************************************************************************************
!> \brief Broyden Mixing using as metrics for the residual the Kerer damping factor
!>        The mixing is applied directly on the density expansion in reciprocal space
!> \param qs_env ...
!> \param mixing_store ...
!> \param rho ...
!> \param para_env ...
!> \par History
!>      03.2009
!> \author MI
! **************************************************************************************************

   SUBROUTINE broyden_mixing(qs_env, mixing_store, rho, para_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'broyden_mixing', routineP = moduleN//':'//routineN

      COMPLEX(dp)                                        :: cc_mix
      COMPLEX(dp), ALLOCATABLE, DIMENSION(:)             :: res_rho
      INTEGER                                            :: handle, i, iatom, ib, ibb, ig, ispin, j, &
                                                            jb, kb, n1, n2, natom, nb, nbuffer, &
                                                            ng, nspin
      LOGICAL                                            :: gapw
      REAL(dp)                                           :: alpha, beta, dcpc_h_res, dcpc_s_res, &
                                                            delta_norm, f_mix, inv_err, res_norm, &
                                                            valh, vals, w0
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: c, g
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: a, b
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      IF(.NOT.(ASSOCIATED(mixing_store%res_buffer)))CALL cp__a("qs_gspace_mixing.F",539)
      IF(.NOT.(ASSOCIATED(mixing_store%rhoin)))CALL cp__a("qs_gspace_mixing.F",540)
      IF(.NOT.(ASSOCIATED(mixing_store%rhoin_old)))CALL cp__a("qs_gspace_mixing.F",541)
      IF(.NOT.(ASSOCIATED(mixing_store%drho_buffer)))CALL cp__a("qs_gspace_mixing.F",542)
      NULLIFY (dft_control, rho_g)
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      CALL qs_rho_get(rho, rho_g=rho_g)

      nspin = SIZE(rho_g, 1)
      ng = SIZE(mixing_store%res_buffer(1, 1)%cc)

      alpha = mixing_store%alpha
      beta = mixing_store%beta
      w0 = mixing_store%broy_w0
      nbuffer = mixing_store%nbuffer
      gapw = dft_control%qs_control%gapw

      ib = MODULO(mixing_store%ncall, nbuffer)+1
      mixing_store%ncall = mixing_store%ncall+1
!dbg
!    ib = MIN(mixing_store%ncall,nbuffer)
!dbg
      nb = MIN(mixing_store%ncall, nbuffer)
      ibb = MODULO(mixing_store%ncall, nbuffer)+1
!dbg
!    ibb = MIN(mixing_store%ncall+1,nbuffer)
!dbg

      ALLOCATE (a(ib-1, ib-1))
      a = 0.0_dp
      ALLOCATE (b(ib-1, ib-1))
      b = 0.0_dp
      ALLOCATE (c(ib-1))
      c = 0.0_dp
      ALLOCATE (g(ib-1))
      g = 0.0_dp
      ALLOCATE (res_rho(ng))

      IF (gapw) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         rho_atom_set=rho_atom)
         natom = SIZE(rho_atom)
      ELSE
         natom = 0
      END IF

      DO ispin = 1, nspin
         res_rho = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO ig = 1, ng
            res_rho(ig) = rho_g(ispin)%pw%cc(ig)-mixing_store%rhoin(ispin)%cc(ig)
         END DO

         IF (ib == 1) THEN
            DO ig = 1, ng
               mixing_store%last_res(ispin)%cc(ig) = res_rho(ig)
               f_mix = alpha*mixing_store%kerker_factor(ig)
               rho_g(ispin)%pw%cc(ig) = mixing_store%rhoin(ispin)%cc(ig)+f_mix*res_rho(ig)
               mixing_store%rhoin_old(ispin)%cc(ig) = mixing_store%rhoin(ispin)%cc(ig)
               mixing_store%rhoin(ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
            END DO

            IF (gapw) THEN
               DO iatom = 1, natom
                  IF (mixing_store%paw(iatom)) THEN
                     mixing_store%cpc_h_lastres(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_h(ispin)%r_coef- &
                                                                       mixing_store%cpc_h_in(iatom, ispin)%r_coef
                     mixing_store%cpc_s_lastres(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_s(ispin)%r_coef- &
                                                                       mixing_store%cpc_s_in(iatom, ispin)%r_coef

                     rho_atom(iatom)%cpc_h(ispin)%r_coef = alpha*rho_atom(iatom)%cpc_h(ispin)%r_coef+ &
                                                           mixing_store%cpc_h_in(iatom, ispin)%r_coef*(1._dp-alpha)
                     rho_atom(iatom)%cpc_s(ispin)%r_coef = alpha*rho_atom(iatom)%cpc_s(ispin)%r_coef+ &
                                                           mixing_store%cpc_s_in(iatom, ispin)%r_coef*(1._dp-alpha)

                     mixing_store%cpc_h_old(iatom, ispin)%r_coef = mixing_store%cpc_h_in(iatom, ispin)%r_coef
                     mixing_store%cpc_s_old(iatom, ispin)%r_coef = mixing_store%cpc_s_in(iatom, ispin)%r_coef
                     mixing_store%cpc_h_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_h(ispin)%r_coef
                     mixing_store%cpc_s_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_s(ispin)%r_coef
                  END IF
               END DO
            END IF
         ELSE

            delta_norm = 0.0_dp
            res_norm = 0.0_dp
            DO ig = 1, ng
               mixing_store%res_buffer(ib-1, ispin)%cc(ig) = res_rho(ig)-mixing_store%last_res(ispin)%cc(ig)
               mixing_store%last_res(ispin)%cc(ig) = res_rho(ig)
               res_norm = res_norm+ &
                          REAL(res_rho(ig), dp)*REAL(res_rho(ig), dp)+ &
                          AIMAG(res_rho(ig))*AIMAG(res_rho(ig))
               delta_norm = delta_norm+ &
                            REAL(mixing_store%res_buffer(ib-1, ispin)%cc(ig), dp)* &
                            REAL(mixing_store%res_buffer(ib-1, ispin)%cc(ig), dp)+ &
                            AIMAG(mixing_store%res_buffer(ib-1, ispin)%cc(ig))* &
                            AIMAG(mixing_store%res_buffer(ib-1, ispin)%cc(ig))
            END DO
            DO ig = 1, ng
               mixing_store%drho_buffer(ib-1, ispin)%cc(ig) = &
                  mixing_store%rhoin(ispin)%cc(ig)- &
                  mixing_store%rhoin_old(ispin)%cc(ig)
            END DO
            CALL mp_sum(delta_norm, para_env%group)
            delta_norm = SQRT(delta_norm)
            CALL mp_sum(res_norm, para_env%group)
            res_norm = SQRT(res_norm)

            mixing_store%res_buffer(ib-1, ispin)%cc(:) = mixing_store%res_buffer(ib-1, ispin)%cc(:)/delta_norm
            mixing_store%drho_buffer(ib-1, ispin)%cc(:) = mixing_store%drho_buffer(ib-1, ispin)%cc(:)/delta_norm

            IF (gapw) THEN

               DO iatom = 1, natom
                  IF (mixing_store%paw(iatom)) THEN
                     n1 = SIZE(mixing_store%cpc_s_in(iatom, ispin)%r_coef, 1)
                     n2 = SIZE(mixing_store%cpc_s_in(iatom, ispin)%r_coef, 2)
                     DO i = 1, n2
                     DO j = 1, n1
                        mixing_store%dcpc_h_in(ib-1, iatom, ispin)%r_coef(j, i) = &
                           (mixing_store%cpc_h_in(iatom, ispin)%r_coef(j, i)- &
                            mixing_store%cpc_h_old(iatom, ispin)%r_coef(j, i))/delta_norm
                        dcpc_h_res = ((rho_atom(iatom)%cpc_h(ispin)%r_coef(j, i)- &
                                       mixing_store%cpc_h_in(iatom, ispin)%r_coef(j, i))- &
                                      mixing_store%cpc_h_lastres(iatom, ispin)%r_coef(j, i))/delta_norm
                        mixing_store%cpc_h_lastres(iatom, ispin)%r_coef(j, i) = rho_atom(iatom)%cpc_h(ispin)%r_coef(j, i)- &
                                                                                mixing_store%cpc_h_in(iatom, ispin)%r_coef(j, i)

                        mixing_store%dcpc_s_in(ib-1, iatom, ispin)%r_coef(j, i) = &
                           (mixing_store%cpc_s_in(iatom, ispin)%r_coef(j, i)- &
                            mixing_store%cpc_s_old(iatom, ispin)%r_coef(j, i))/delta_norm
                        dcpc_s_res = ((rho_atom(iatom)%cpc_s(ispin)%r_coef(j, i)- &
                                       mixing_store%cpc_s_in(iatom, ispin)%r_coef(j, i))- &
                                      mixing_store%cpc_s_lastres(iatom, ispin)%r_coef(j, i))/delta_norm
                        mixing_store%cpc_s_lastres(iatom, ispin)%r_coef(j, i) = rho_atom(iatom)%cpc_s(ispin)%r_coef(j, i)- &
                                                                                mixing_store%cpc_s_in(iatom, ispin)%r_coef(j, i)

                        mixing_store%dcpc_h_in(ib-1, iatom, ispin)%r_coef(j, i) = &
                           alpha*dcpc_h_res+ &
                           mixing_store%dcpc_h_in(ib-1, iatom, ispin)%r_coef(j, i)
                        mixing_store%dcpc_s_in(ib-1, iatom, ispin)%r_coef(j, i) = &
                           alpha*dcpc_s_res+ &
                           mixing_store%dcpc_s_in(ib-1, iatom, ispin)%r_coef(j, i)
                     END DO
                     END DO
                  END IF
               END DO

            END IF

            a(:, :) = 0.0_dp
            DO ig = 1, ng
               f_mix = alpha*mixing_store%kerker_factor(ig)
               mixing_store%drho_buffer(ib-1, ispin)%cc(ig) = &
                  f_mix*mixing_store%res_buffer(ib-1, ispin)%cc(ig)+ &
                  mixing_store%drho_buffer(ib-1, ispin)%cc(ig)
            END DO
            DO jb = 1, ib-1
               DO kb = jb, ib-1
                  DO ig = 1, mixing_store%ig_max !ng
                     a(kb, jb) = a(kb, jb)+mixing_store%p_metric(ig)*( &
                                 REAL(mixing_store%res_buffer(jb, ispin)%cc(ig), dp)* &
                                 REAL(mixing_store%res_buffer(kb, ispin)%cc(ig), dp)+ &
                                 AIMAG(mixing_store%res_buffer(jb, ispin)%cc(ig))* &
                                 AIMAG(mixing_store%res_buffer(kb, ispin)%cc(ig)))
                  END DO
                  a(jb, kb) = a(kb, jb)
               END DO
            END DO
            CALL mp_sum(a, para_env%group)

            C = 0.0_dp
            DO jb = 1, ib-1
               a(jb, jb) = w0+a(jb, jb)
               DO ig = 1, mixing_store%ig_max !ng
                  c(jb) = c(jb)+mixing_store%p_metric(ig)*( &
                          REAL(mixing_store%res_buffer(jb, ispin)%cc(ig), dp)*REAL(res_rho(ig), dp)+ &
                          AIMAG(mixing_store%res_buffer(jb, ispin)%cc(ig))*AIMAG(res_rho(ig)))
               END DO
            END DO
            CALL mp_sum(c, para_env%group)
            CALL invert_matrix(a, b, inv_err)

            CALL dgemv('T', IB-1, IB-1, 1.0_dp, B, IB-1, C, 1, 0.0_dp, G, 1)

            DO ig = 1, ng
               cc_mix = CMPLX(0.0_dp, 0.0_dp, kind=dp)
               DO jb = 1, ib-1
                  cc_mix = cc_mix-G(jb)*mixing_store%drho_buffer(jb, ispin)%cc(ig)
               END DO
               f_mix = alpha*mixing_store%kerker_factor(ig)
               rho_g(ispin)%pw%cc(ig) = mixing_store%rhoin(ispin)%cc(ig)+ &
                                        f_mix*res_rho(ig)+cc_mix
               mixing_store%rhoin_old(ispin)%cc(ig) = mixing_store%rhoin(ispin)%cc(ig)
               mixing_store%rhoin(ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
            END DO

            IF (gapw) THEN
               DO iatom = 1, natom
                  IF (mixing_store%paw(iatom)) THEN
                     n1 = SIZE(mixing_store%cpc_s_in(iatom, ispin)%r_coef, 1)
                     n2 = SIZE(mixing_store%cpc_s_in(iatom, ispin)%r_coef, 2)
                     DO i = 1, n2
                     DO j = 1, n1
                        valh = 0.0_dp
                        vals = 0.0_dp
                        DO jb = 1, ib-1
                           valh = valh-G(jb)*mixing_store%dcpc_h_in(jb, iatom, ispin)%r_coef(j, i)
                           vals = vals-G(jb)*mixing_store%dcpc_s_in(jb, iatom, ispin)%r_coef(j, i)
                        END DO
                        rho_atom(iatom)%cpc_h(ispin)%r_coef(j, i) = &
                           alpha*rho_atom(iatom)%cpc_h(ispin)%r_coef(j, i)+ &
                           mixing_store%cpc_h_in(iatom, ispin)%r_coef(j, i)*(1._dp-alpha)+valh
                        rho_atom(iatom)%cpc_s(ispin)%r_coef(j, i) = &
                           alpha*rho_atom(iatom)%cpc_s(ispin)%r_coef(j, i)+ &
                           mixing_store%cpc_s_in(iatom, ispin)%r_coef(j, i)*(1._dp-alpha)+vals
                     END DO
                     END DO

                     mixing_store%cpc_h_old(iatom, ispin)%r_coef = mixing_store%cpc_h_in(iatom, ispin)%r_coef
                     mixing_store%cpc_s_old(iatom, ispin)%r_coef = mixing_store%cpc_s_in(iatom, ispin)%r_coef
                     mixing_store%cpc_h_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_h(ispin)%r_coef
                     mixing_store%cpc_s_in(iatom, ispin)%r_coef = rho_atom(iatom)%cpc_s(ispin)%r_coef
                  END IF
               END DO
            END IF

!  MI: this type of propagation makes the convergence worse, but it is not crystal-clear why
!      possibly if is still affected by some bug. Re-check!
!        IF(mixing_store%ncall>=nbuffer) THEN
!         DO jb = 1,nbuffer - 1
!            mixing_store%drho_buffer(jb,ispin)%cc(:) = mixing_store%drho_buffer(jb+1,ispin)%cc(:)
!            mixing_store%res_buffer(jb,ispin)%cc(:) = mixing_store%res_buffer(jb+1,ispin)%cc(:)
!         END DO
!       END IF

         END IF

      END DO

      DEALLOCATE (a, b, c, g, res_rho)

      CALL timestop(handle)

   END SUBROUTINE broyden_mixing

! **************************************************************************************************
!> \brief ...
!> \param mixing_store ...
!> \param rho ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE broyden_mixing_new(mixing_store, rho, para_env)

      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'broyden_mixing_new', &
         routineP = moduleN//':'//routineN

      COMPLEX(dp), ALLOCATABLE, DIMENSION(:)             :: delta_res_p, res_rho, res_rho_p, tmp
      INTEGER                                            :: handle, ib, ibb, ig, info, ispin, jb, &
                                                            kb, kkb, lwork, nb, nbuffer, ng, nspin
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: IWORK
      LOGICAL                                            :: skip_bq
      REAL(dp) :: alpha, beta, delta, delta_p, delta_rhog, delta_rhog_p, f_mix, imp, imp_j, &
         inv_err, norm, norm_ig, rep, rep_j, sqt_uvol, sqt_vol, uvol, vol, wc, wmax
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: aval, work_dgesdd
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: a, a2, au, av, b, bq, fm2, work
      REAL(dp), DIMENSION(:), POINTER                    :: p_metric
      REAL(dp), DIMENSION(:, :), POINTER                 :: fmat, weight
      TYPE(cp_1d_z_p_type), DIMENSION(:), POINTER        :: tmp_z
      TYPE(cp_1d_z_p_type), DIMENSION(:, :), POINTER     :: delta_res, u_vec, z_vec
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g

      IF(.NOT.(ASSOCIATED(mixing_store%rhoin_buffer)))CALL cp__a("qs_gspace_mixing.F",816)

      CALL timeset(routineN, handle)

      NULLIFY (delta_res, u_vec, z_vec)
      NULLIFY (fmat, rho_g)
      CALL qs_rho_get(rho, rho_g=rho_g)

      nspin = SIZE(rho_g, 1)
      ng = SIZE(mixing_store%rhoin_buffer(1, 1)%cc)
      vol = rho_g(1)%pw%pw_grid%vol
      sqt_vol = SQRT(vol)
      uvol = 1.0_dp/vol
      sqt_uvol = SQRT(uvol)
      alpha = mixing_store%alpha
      beta = mixing_store%beta

      wc = mixing_store%wc
      wmax = mixing_store%wmax
      nbuffer = mixing_store%nbuffer

      mixing_store%ncall = mixing_store%ncall+1
      ib = MIN(mixing_store%ncall, nbuffer)
      nb = MIN(mixing_store%ncall, nbuffer)
      ibb = MIN(mixing_store%ncall+1, nbuffer)

      ALLOCATE (res_rho(ng))
      ALLOCATE (res_rho_p(ng))
      IF (ib > 1) THEN
         ALLOCATE (a(ib-1, ib-1))
         a = 0.0_dp
         ALLOCATE (b(ib-1, ib-1))
         b = 0.0_dp
         ALLOCATE (bq(ib-1, ib-1))
         bq = 0.0_dp
         ALLOCATE (tmp(ng))
         ALLOCATE (delta_res_p(ng))
      END IF

      p_metric => mixing_store%p_metric
      weight => mixing_store%weight
      IF(.NOT.(ASSOCIATED(mixing_store%delta_res)))CALL cp__a("qs_gspace_mixing.F",857)
      delta_res => mixing_store%delta_res
      IF(.NOT.(ASSOCIATED(mixing_store%u_vec)))CALL cp__a("qs_gspace_mixing.F",859)
      u_vec => mixing_store%u_vec
      IF(.NOT.(ASSOCIATED(mixing_store%z_vec)))CALL cp__a("qs_gspace_mixing.F",861)
      z_vec => mixing_store%z_vec

      delta_rhog = 0.0_dp
      delta_rhog_p = 0.0_dp

      DO ispin = 1, nspin

         fmat => mixing_store%fmat(:, :, ispin)

         delta = 0.0_dp
         delta_p = 0.0_dp
         ! Residual at this step R_i(G) (rho_out(G)-rho_in(G))
         ! Residual multiplied by the metrics RP_i(G) = (rho_out(G)-rho_in(G)) * P(G)
         ! Delta is the norm of the residual, measures how far we are from convergence
         DO ig = 1, ng
            res_rho(ig) = (rho_g(ispin)%pw%cc(ig)-mixing_store%rhoin_buffer(ib, ispin)%cc(ig))*sqt_uvol
            res_rho_p(ig) = res_rho(ig)*p_metric(ig)
            norm_ig = REAL(res_rho(ig), dp)*REAL(res_rho(ig), dp)+AIMAG(res_rho(ig))*AIMAG(res_rho(ig))
            delta = delta+norm_ig
            delta_p = delta_p+norm_ig*p_metric(ig)
         END DO
         CALL mp_sum(delta, para_env%group)
         delta = SQRT(delta)
         delta = delta
         CALL mp_sum(delta_p, para_env%group)
         delta_p = SQRT(delta_p)
         delta_p = delta_p
         delta_rhog = delta_rhog+delta
         delta_rhog_p = delta_rhog_p+delta_p

         weight(ib, ispin) = 1.0_dp ! wc
         IF (wc < 0.0_dp) weight(ib, ispin) = 0.01_dp*ABS(wc)/(delta_p*delta_p)
         IF (weight(ib, ispin) == 0.0_dp) weight(ib, ispin) = 100.0_dp
         IF (weight(ib, ispin) < 1.0_dp) weight(ib, ispin) = 1.0_dp

         IF (ib == 1) THEN
            ! Simple Kerker damping : linear mixing rho(G) = rho_in(G) - alpha k(G)*(rho_out(G)-rho_in(G))
            DO ig = 1, ng
               f_mix = alpha*mixing_store%kerker_factor(ig)
!          mixing_store%rhoin_buffer(ib,ispin)%cc(ig) = mixing_store%rhoin_buffer(ib,ispin)%cc(ig)*sqt_vol
               rho_g(ispin)%pw%cc(ig) = (mixing_store%rhoin_buffer(ib, ispin)%cc(ig)+f_mix*res_rho(ig))
               mixing_store%rhoin_buffer(ibb, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
!          rho_g(ispin)%pw%cc(ig) =  rho_g(ispin)%pw%cc(ig)*sqt_vol
            END DO
         ELSE
            norm = 0.0_dp
            ! Difference of residuals DR_{i-1)} (G) = R_i(G) - R_{i-1}(G)
            DO ig = 1, ng
               delta_res(ib-1, ispin)%cc(ig) = res_rho(ig)-mixing_store%last_res(ispin)%cc(ig)
               delta_res_p(ig) = p_metric(ig)*(res_rho(ig)-mixing_store%last_res(ispin)%cc(ig))
               norm_ig = REAL(delta_res(ib-1, ispin)%cc(ig), dp)*REAL(delta_res_p(ig), dp)+ &
                         AIMAG(delta_res(ib-1, ispin)%cc(ig))*AIMAG(delta_res_p(ig))
               norm = norm+norm_ig
            END DO
            CALL mp_sum(norm, para_env%group)
            norm = 1._dp/SQRT(norm)
            delta_res(ib-1, ispin)%cc(:) = delta_res(ib-1, ispin)%cc(:)*norm
            delta_res_p(:) = delta_res_p(:)*norm

!dbg
!    write(*,*) 'ibroy ', ib, weight(ib,ispin)
            IF (para_env%ionode) WRITE (*, *) 'norm ', norm
!dbg
            ! Vector U_{i-1}(G) =  Drho_{i-1} + k(G)  * DR_{i-1}(G)
            DO ig = 1, ng
               tmp(ig) = (mixing_store%rhoin_buffer(ib, ispin)%cc(ig)- &
                          mixing_store%rhoin_buffer(ib-1, ispin)%cc(ig))*norm
               u_vec(ib-1, ispin)%cc(ig) = (tmp(ig)+ &
                                            mixing_store%kerker_factor(ig)*delta_res(ib-1, ispin)%cc(ig))
            END DO

            DO jb = 1, ib-1
               fmat(jb, ib-1) = 0.0_dp

               DO ig = 1, ng
                  rep_j = REAL(delta_res(jb, ispin)%cc(ig), dp)
                  imp_j = AIMAG(delta_res(jb, ispin)%cc(ig))
                  ! < DR_{j} | DR_{i-1} >
                  rep = REAL(delta_res_p(ig), dp)
                  imp = AIMAG(delta_res_p(ig))
                  fmat(jb, ib-1) = fmat(jb, ib-1)+rep_j*rep+imp_j*imp
               END DO

            END DO
            CALL mp_sum(fmat(1:ib-1, ib-1), para_env%group)

            fmat(ib-1, ib-1) = 1.0_dp

            DO jb = 1, ib-2
               fmat(ib-1, jb) = fmat(jb, ib-1)
            ENDDO
!dbg
            IF (para_env%ionode) THEN
               WRITE (*, *) 'fmat '
               DO jb = 1, ib-1
                  WRITE (*, *) jb, (fmat(jb, kb), kb=1, ib-1)
               END DO
            END IF
!dbg

!dbg
            ALLOCATE (fm2(ib-1, ib-1))
            fm2 = 0.0_dp
            DO kb = 1, ib-1
               DO jb = 1, kb
                  DO ig = 1, ng
                     rep_j = REAL(delta_res(jb, ispin)%cc(ig), dp)
                     imp_j = AIMAG(delta_res(jb, ispin)%cc(ig))
                     rep = REAL(delta_res(kb, ispin)%cc(ig), dp)
                     imp = AIMAG(delta_res(kb, ispin)%cc(ig))
                     fm2(jb, kb) = fm2(jb, kb)+(rep_j*rep+imp_j*imp)*p_metric(ig)
                  END DO ! ig
               END DO
            END DO
            CALL mp_sum(fm2, para_env%group)
            DO jb = 1, ib-1
               fm2(jb, jb) = 1.0_dp
               DO kb = 1, jb-1
                  fmat(jb, kb) = fmat(kb, jb)
               ENDDO
            END DO
            IF (para_env%ionode) THEN
               WRITE (*, *) 'fm2 '
               DO jb = 1, ib-1
                  WRITE (*, *) jb, (fm2(jb, kb), kb=1, ib-1)
               END DO
            END IF
!dbg

            DO jb = 1, ib-1
!          a(jb,jb) = 1.0_dp + weight(jb,ispin)*weight(jb,ispin)*fmat(jb,jb)
               a(jb, jb) = weight(jb, ispin)*weight(jb, ispin)*fmat(jb, jb)
               DO kb = 1, jb-1
                  a(jb, kb) = weight(jb, ispin)*weight(kb, ispin)*fmat(jb, kb)
                  a(kb, jb) = weight(jb, ispin)*weight(kb, ispin)*fmat(kb, jb)
               ENDDO
            END DO
!dbg
            IF (para_env%ionode) THEN
               WRITE (*, *) 'a mat'
               DO kb = 1, ib-1
                  WRITE (*, *) 'a ', kb, (a(kb, jb), jb=1, ib-1)
               END DO
            END IF
!dbg
!dbg
            ALLOCATE (a2(ib-1, ib-1))
            DO jb = 1, ib-1
            DO kb = 1, ib-1
               a2(kb, jb) = weight(jb, ispin)*weight(kb, ispin)*fm2(kb, jb)
            END DO
!          a2(jb,jb) = 1.0_dp + a2(jb,jb)
            END DO

            IF (.TRUE.) THEN
               b = 0.0_dp
               CALL invert_matrix(a, b, inv_err)
!dbg
               IF (para_env%ionode) THEN
                  WRITE (*, *) 'invert version 1'
                  DO kb = 1, ib-1
                     WRITE (*, *) 'b ', kb, (b(kb, jb), jb=1, ib-1)
                  END DO
               END IF
!dbg
            ELSE
               b = 0.0_dp
               ALLOCATE (work(ib-1, ib-1), aval(ib-1), av(ib-1, ib-1), au(ib-1, ib-1))
               work(:, :) = a
               ALLOCATE (iwork(8*(ib-1)), work_dgesdd(1))
               lwork = -1
               CALL DGESDD('S', ib-1, ib-1, work, ib-1, aval, au, ib-1, av, ib-1, work_dgesdd, lwork, iwork, info)
               lwork = INT(work_dgesdd(1))
               DEALLOCATE (work_dgesdd); ALLOCATE (work_dgesdd(lwork))
               CALL DGESDD('S', ib-1, ib-1, work, ib-1, aval, au, ib-1, av, ib-1, work_dgesdd, lwork, iwork, info)
               ! construct the inverse
               DO kb = 1, ib-1
                  ! invert SV
                  IF (aval(kb) < 1.E-6_dp) THEN
                     aval(kb) = 0.0_dp
                  ELSE
                     aval(kb) = 1.0_dp/aval(kb)
                  ENDIF
                  av(kb, :) = av(kb, :)*aval(kb)
               ENDDO
               CALL DGEMM('T', 'T', ib-1, ib-1, ib-1, 1.0_dp, av, ib-1, au, ib-1, 0.0_dp, b, ib-1)
               DEALLOCATE (iwork, aval, au, av, work_dgesdd, work)

!          CALL diamat_all(work(1:ib-1,1:ib-1),aval(1:ib-1))
!
!          avec(1:ib-1,1:ib-1) = work(1:ib-1,1:ib-1)
!          DO jb = 1,ib-1
!            IF(ABS(aval(jb)) < 1.E-5_dp ) THEN
!              avec(1:ib-1,jb) = 0.0_dp
!            ELSE
!              avec(1:ib-1,jb) = work(1:ib-1,jb) / aval(jb)
!            END IF
!          END DO
!          b(1:ib-1,1:ib-1) = MATMUL(avec(1:ib-1,1:ib-1),work(1:ib-1,1:ib-1))

!          DEALLOCATE(work,avec,aval)
            END IF

            bq = 0.0_dp
            ! Broyden second method requires also bq (NYI)
            skip_bq = .TRUE.
            DO jb = 1, ib-1
!          skip_bq = skip_bq .AND. (weight(jb,ispin) > wmax)
               DO kb = 1, ib-2
                  bq(jb, kb) = 0.0_dp
                  DO kkb = 1, ib-1
                     bq(jb, kb) = bq(jb, kb)-weight(jb, ispin)*weight(kkb, ispin)*b(jb, kkb)*fmat(kkb, kb)
                  END DO
               END DO
               bq(jb, jb) = 1.0_dp+bq(jb, jb)
            END DO

            IF (.NOT. skip_bq) THEN
               ! in this case the old z_vec is needed
               ! a temporary array is needed to store the new one
               ALLOCATE (tmp_z(ib-1))
               DO jb = 1, ib-1
                  ALLOCATE (tmp_z(jb)%cc(ng))
               END DO
            END IF
            DO jb = 1, ib-1
               tmp(:) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
               IF (.NOT. skip_bq) THEN
                  ! sum_{kb} bq(jb,kb) * z_vec_{kb,iter-2}
                  ! added only for small weights
                  DO kb = 1, ib-2
                     IF (weight(kb, ispin) >= (10.0_dp*wmax)) CYCLE
                     DO ig = 1, ng
                        tmp(ig) = tmp(ig)+bq(jb, kb)*z_vec(kb, ispin)%cc(ig)
                     END DO
                  END DO ! kb
               END IF

               ! sum_{kb} w(jb)*w(kb)*b(jb,kb) * u_vec_{kb}
               DO kb = 1, ib-1
                  DO ig = 1, ng
                     tmp(ig) = tmp(ig)+weight(kb, ispin)*weight(jb, ispin)*b(jb, kb)*u_vec(kb, ispin)%cc(ig)
                  END DO
               END DO

               ! store the new z_vec(jb)
               IF (skip_bq .OR. (weight(jb, ispin) >= (10._dp*wmax))) THEN
                  z_vec(jb, ispin)%cc(:) = tmp(:)
               ELSE
                  ! temporary array: old z_vec may still be needed
                  tmp_z(jb)%cc(:) = tmp(:)
               END IF
            END DO !jb

            IF (.NOT. skip_bq) THEN
               DO jb = 1, ib-1
                  IF (weight(jb, ispin) < (10._dp*wmax)) z_vec(jb, ispin)%cc(:) = tmp_z(jb)%cc(:)
                  DEALLOCATE (tmp_z(jb)%cc)
               END DO
               DEALLOCATE (tmp_z)
            END IF

            ! Overwrite the density i reciprocal space
            rho_g(ispin)%pw%cc(:) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
            DO jb = 1, ib-1
               norm = 0.0_dp
               DO ig = 1, ng
                  rep_j = REAL(delta_res(jb, ispin)%cc(ig), dp)
                  imp_j = AIMAG(delta_res(jb, ispin)%cc(ig))
                  rep = REAL(res_rho_p(ig), dp)
                  imp = AIMAG(res_rho_p(ig))
                  norm = norm+rep_j*rep+imp_j*imp
               END DO
               CALL mp_sum(norm, para_env%group)
!dbg
               IF (para_env%ionode) WRITE (*, *) 'norm ', norm
!dbg
               ! Subtract |Z_jb)><DR_jb|P|R_{iter}>
               DO ig = 1, ng
                  rho_g(ispin)%pw%cc(ig) = rho_g(ispin)%pw%cc(ig)-norm*z_vec(jb, ispin)%cc(ig)*sqt_vol

               END DO
            END DO

            DO ig = 1, ng
               f_mix = alpha*mixing_store%kerker_factor(ig)
               rho_g(ispin)%pw%cc(ig) = rho_g(ispin)%pw%cc(ig)+ &
                                        mixing_store%rhoin_buffer(ib, ispin)%cc(ig)+f_mix*res_rho(ig)
               mixing_store%rhoin_buffer(ibb, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
!          rho_g(ispin)%pw%cc(ig) = rho_g(ispin)%pw%cc(ig)*sqt_vol
            END DO
         END IF ! ib

         mixing_store%last_res(ispin)%cc(:) = res_rho(:)
         IF (mixing_store%ncall >= nbuffer) THEN
            DO jb = 1, nbuffer-2
               mixing_store%rhoin_buffer(jb, ispin)%cc(:) = mixing_store%rhoin_buffer(jb+1, ispin)%cc(:)
               mixing_store%delta_res(jb, ispin)%cc(:) = mixing_store%delta_res(jb+1, ispin)%cc(:)
               mixing_store%u_vec(jb, ispin)%cc(:) = mixing_store%u_vec(jb+1, ispin)%cc(:)
               mixing_store%z_vec(jb, ispin)%cc(:) = mixing_store%z_vec(jb+1, ispin)%cc(:)
               DO kb = 1, nbuffer-2
                  mixing_store%fmat(kb, jb, ispin) = mixing_store%fmat(kb+1, jb+1, ispin)
               END DO
            END DO
            mixing_store%rhoin_buffer(nbuffer-1, ispin)%cc(:) = &
               mixing_store%rhoin_buffer(nbuffer, ispin)%cc(:)
            mixing_store%weight(nbuffer-1, ispin) = mixing_store%weight(nbuffer, ispin)
         END IF

      END DO ! ispin
      IF (ib > 1) THEN
         DEALLOCATE (a, b, bq)
         DEALLOCATE (delta_res_p, tmp)
      END IF
      DEALLOCATE (res_rho, res_rho_p)

      CALL timestop(handle)

   END SUBROUTINE broyden_mixing_new

! **************************************************************************************************
!> \brief Multisecant scheme to perform charge density Mixing
!>        as preconditioner we use the Kerker damping factor
!>        The mixing is applied directly on the density in reciprocal space,
!>        therefore it affects the potentials
!>        on the grid but not the density matrix
!> \param mixing_store ...
!> \param rho ...
!> \param para_env ...
!> \par History
!>      05.2009
!> \author MI
! **************************************************************************************************
   SUBROUTINE multisecant_mixing(mixing_store, rho, para_env)

      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'multisecant_mixing', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: cmone = (-1.0_dp, 0.0_dp), &
                                                            cone = (1.0_dp, 0.0_dp), &
                                                            czero = (0.0_dp, 0.0_dp)

      COMPLEX(dp)                                        :: saa, yaa
      COMPLEX(dp), ALLOCATABLE, DIMENSION(:)             :: gn_global, pgn, res_matrix_global, &
                                                            tmp_vec, ugn
      COMPLEX(dp), ALLOCATABLE, DIMENSION(:, :)          :: a_matrix, res_matrix, sa, step_matrix, ya
      COMPLEX(dp), DIMENSION(:), POINTER                 :: gn
      INTEGER                                            :: handle, ib, ib_next, ib_prev, ig, &
                                                            ig_global, iig, ispin, jb, kb, nb, &
                                                            nbuffer, ng, ng_global, nspin
      LOGICAL                                            :: use_zgemm, use_zgemm_rev
      REAL(dp) :: alpha, f_mix, gn_norm, gn_norm_old, inv_err, n_low, n_up, pgn_norm, prec, &
         r_step, reg_par, sigma_max, sigma_tilde, step_size
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: norm_res, norm_res_low, norm_res_up
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: b_matrix, binv_matrix
      REAL(dp), DIMENSION(:), POINTER                    :: g2
      REAL(dp), SAVE                                     :: sigma_old = 1.0_dp
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g

      IF(.NOT.(ASSOCIATED(mixing_store)))CALL cp__a("qs_gspace_mixing.F",1224)

      CALL timeset(routineN, handle)

      NULLIFY (gn, rho_g)

      use_zgemm = .FALSE.
      use_zgemm_rev = .TRUE.
!   use_zgemm_rev = .FALSE.

      ! prepare the parameters
      CALL qs_rho_get(rho, rho_g=rho_g)

      nspin = SIZE(rho_g, 1)
      ! not implemented for large grids.
      IF(.NOT.(rho_g(1)%pw%pw_grid%ngpts < HUGE(ng_global)))CALL cp__a("qs_gspace_mixing.F",1239)
      ng_global = INT(rho_g(1)%pw%pw_grid%ngpts)
      ng = SIZE(mixing_store%rhoin_buffer(1, 1)%cc)
      alpha = mixing_store%alpha

      sigma_max = mixing_store%sigma_max
      reg_par = mixing_store%reg_par
      r_step = mixing_store%r_step
      nbuffer = mixing_store%nbuffer

      ! determine the step number, and multisecant iteration
      nb = MIN(mixing_store%ncall, nbuffer-1)
      ib = MODULO(mixing_store%ncall, nbuffer)+1
      IF (mixing_store%ncall > 0) THEN
         ib_prev = MODULO(mixing_store%ncall-1, nbuffer)+1
      ELSE
         ib_prev = 0
      END IF
      mixing_store%ncall = mixing_store%ncall+1
      ib_next = MODULO(mixing_store%ncall, nbuffer)+1

!dbg
      IF (para_env%ionode) WRITE (*, *) mixing_store%ncall, ib, nb, ib_prev, ib_next
!dbg

      ! compute the residual gn and its norm gn_norm
      DO ispin = 1, nspin
         gn => mixing_store%res_buffer(ib, ispin)%cc
         gn_norm = 0.0_dp
         DO ig = 1, ng
            gn(ig) = (rho_g(ispin)%pw%cc(ig)-mixing_store%rhoin_buffer(ib, ispin)%cc(ig))
            gn_norm = gn_norm+ &
                      REAL(gn(ig), dp)*REAL(gn(ig), dp)+AIMAG(gn(ig))*AIMAG(gn(ig))
         END DO
         CALL mp_sum(gn_norm, para_env%group)
         gn_norm = SQRT(gn_norm)
         mixing_store%norm_res_buffer(ib, ispin) = gn_norm
      END DO

      IF (nb == 0) THEN
         !simple mixing
         DO ispin = 1, nspin
            DO ig = 1, ng
               f_mix = alpha*mixing_store%kerker_factor(ig)
               rho_g(ispin)%pw%cc(ig) = mixing_store%rhoin_buffer(1, ispin)%cc(ig)+ &
                                        f_mix*mixing_store%res_buffer(1, ispin)%cc(ig)
               mixing_store%rhoin_buffer(ib_next, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
            END DO
         END DO
         CALL timestop(handle)
         RETURN
      END IF

      ! allocate temporary arrays
      ! step_matrix  S ngxnb
      ALLOCATE (step_matrix(ng, nb))
      ! res_matrix Y  ngxnb
      ALLOCATE (res_matrix(ng, nb))
      ! matrix A  nbxnb
      ALLOCATE (a_matrix(nb, ng_global))
      ! PSI nb vector of norms
      ALLOCATE (norm_res(nb))
      ALLOCATE (norm_res_low(nb))
      ALLOCATE (norm_res_up(nb))

      ! matrix B   nbxnb
      ALLOCATE (b_matrix(nb, nb))
      ! matrix B_inv   nbxnb
      ALLOCATE (binv_matrix(nb, nb))

      ALLOCATE (gn_global(ng_global))
      ALLOCATE (res_matrix_global(ng_global))
      IF (use_zgemm) THEN
         ALLOCATE (sa(ng, ng_global))
         ALLOCATE (ya(ng, ng_global))
      END IF
      IF (use_zgemm_rev) THEN
         ALLOCATE (tmp_vec(nb))
      END IF
      ALLOCATE (pgn(ng))
      ALLOCATE (ugn(ng))

      DO ispin = 1, nspin
         ! generate the global vector with the present residual
         gn => mixing_store%res_buffer(ib, ispin)%cc
         gn_global = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO ig = 1, ng
            ig_global = mixing_store%ig_global_index(ig)
            gn_global(ig_global) = gn(ig)
         END DO
         CALL mp_sum(gn_global, para_env%group)

         ! compute steps (matrix S) and residual differences (matrix Y)
         ! with respect to the present
         ! step and the present residual (use stored rho_in and res_buffer)

         ! These quantities are pre-conditioned by means of Kerker factor multipliccation
         g2 => rho_g(1)%pw%pw_grid%gsq

         DO jb = 1, nb+1
            IF (jb < ib) THEN
               kb = jb
            ELSEIF (jb > ib) THEN
               kb = jb-1
            ELSE
               CYCLE
            END IF
            norm_res(kb) = 0.0_dp
            norm_res_low(kb) = 0.0_dp
            norm_res_up(kb) = 0.0_dp
            DO ig = 1, ng

!          prec = mixing_store%kerker_factor(ig)
               prec = 1.0_dp

               step_matrix(ig, kb) = prec*(mixing_store%rhoin_buffer(jb, ispin)%cc(ig)- &
                                           mixing_store%rhoin_buffer(ib, ispin)%cc(ig))
               res_matrix(ig, kb) = (mixing_store%res_buffer(jb, ispin)%cc(ig)- &
                                     mixing_store%res_buffer(ib, ispin)%cc(ig))
               norm_res(kb) = norm_res(kb)+REAL(res_matrix(ig, kb), dp)*REAL(res_matrix(ig, kb), dp)+ &
                              AIMAG(res_matrix(ig, kb))*AIMAG(res_matrix(ig, kb))
               IF (g2(ig) < 4.0_dp) THEN
                  norm_res_low(kb) = norm_res_low(kb)+ &
                                     REAL(res_matrix(ig, kb), dp)*REAL(res_matrix(ig, kb), dp)+ &
                                     AIMAG(res_matrix(ig, kb))*AIMAG(res_matrix(ig, kb))
               ELSE
                  norm_res_up(kb) = norm_res_up(kb)+ &
                                    REAL(res_matrix(ig, kb), dp)*REAL(res_matrix(ig, kb), dp)+ &
                                    AIMAG(res_matrix(ig, kb))*AIMAG(res_matrix(ig, kb))
               END IF
               res_matrix(ig, kb) = prec*res_matrix(ig, kb)
            END DO
         END DO !jb

         ! normalize each column of S and Y => Snorm Ynorm
         CALL mp_sum(norm_res, para_env%group)
         CALL mp_sum(norm_res_up, para_env%group)
         CALL mp_sum(norm_res_low, para_env%group)
!dbg
         IF (para_env%ionode) WRITE (*, *) 'norm res   ', norm_res(1:nb)
         IF (para_env%ionode) WRITE (*, *) 'norm res  L', norm_res_low(1:nb)
         IF (para_env%ionode) WRITE (*, *) 'norm res  U', norm_res_up(1:nb)
!dbg
         norm_res(1:nb) = 1.0_dp/SQRT(norm_res(1:nb))
         n_low = 0.0_dp
         n_up = 0.0_dp
         DO jb = 1, nb
            n_low = n_low+norm_res_low(jb)/norm_res(jb)
            n_up = n_up+norm_res_up(jb)/norm_res(jb)
         END DO
         DO ig = 1, ng
            IF (g2(ig) > 4.0_dp) THEN
               step_matrix(ig, 1:nb) = step_matrix(ig, 1:nb)*SQRT(n_low/n_up)
               res_matrix(ig, 1:nb) = res_matrix(ig, 1:nb)*SQRT(n_low/n_up)
            END IF
         END DO
         DO kb = 1, nb
            step_matrix(1:ng, kb) = step_matrix(1:ng, kb)*norm_res(kb)
            res_matrix(1:ng, kb) = res_matrix(1:ng, kb)*norm_res(kb)
         END DO

         ! compute A as [(Ynorm^t Ynorm) + (alpha I)]^(-1) Ynorm^t
         ! compute B
         DO jb = 1, nb
            DO kb = 1, nb
               b_matrix(kb, jb) = 0.0_dp
               DO ig = 1, ng
                  ! it is assumed that summing over all G vector gives a real, because
                  !  y(-G,kb) = (y(G,kb))*
                  b_matrix(kb, jb) = b_matrix(kb, jb)+REAL(res_matrix(ig, kb)*res_matrix(ig, jb), dp)
               END DO
            END DO
         END DO

         CALL mp_sum(b_matrix, para_env%group)
         DO jb = 1, nb
            b_matrix(jb, jb) = b_matrix(jb, jb)+reg_par
         END DO
         ! invert B
         CALL invert_matrix(b_matrix, binv_matrix, inv_err)

         ! A = Binv Ynorm^t
         a_matrix = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO kb = 1, nb
            res_matrix_global = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
            DO ig = 1, ng
               ig_global = mixing_store%ig_global_index(ig)
               res_matrix_global(ig_global) = res_matrix(ig, kb)
            END DO
            CALL mp_sum(res_matrix_global, para_env%group)

            DO jb = 1, nb
               DO ig = 1, ng
                  ig_global = mixing_store%ig_global_index(ig)
                  a_matrix(jb, ig_global) = a_matrix(jb, ig_global)+ &
                                            binv_matrix(jb, kb)*res_matrix_global(ig_global)
               END DO
            END DO
         END DO
         CALL mp_sum(a_matrix, para_env%group)

         ! compute the two components of gn that will be used to update rho
         gn => mixing_store%res_buffer(ib, ispin)%cc
         pgn_norm = 0.0_dp

         IF (use_zgemm) THEN

            CALL zgemm("N", "N", ng, ng_global, nb, cmone, step_matrix(1, 1), ng, &
                       a_matrix(1, 1), nb, czero, sa(1, 1), ng)
            CALL zgemm("N", "N", ng, ng_global, nb, cmone, res_matrix(1, 1), ng, &
                       a_matrix(1, 1), nb, czero, ya(1, 1), ng)
            DO ig = 1, ng
               ig_global = mixing_store%ig_global_index(ig)
               ya(ig, ig_global) = ya(ig, ig_global)+CMPLX(1.0_dp, 0.0_dp, KIND=dp)
            END DO

            CALL zgemv("N", ng, ng_global, cone, sa(1, 1), &
                       ng, gn_global(1), 1, czero, pgn(1), 1)
            CALL zgemv("N", ng, ng_global, cone, ya(1, 1), &
                       ng, gn_global(1), 1, czero, ugn(1), 1)

            DO ig = 1, ng
               pgn_norm = pgn_norm+REAL(pgn(ig), dp)*REAL(pgn(ig), dp)+ &
                          AIMAG(pgn(ig))*AIMAG(pgn(ig))
            END DO
            CALL mp_sum(pgn_norm, para_env%group)
         ELSEIF (use_zgemm_rev) THEN

            CALL zgemv("N", nb, ng_global, cone, a_matrix(1, 1), &
                       nb, gn_global(1), 1, czero, tmp_vec(1), 1)

            CALL zgemv("N", ng, nb, cmone, step_matrix(1, 1), ng, &
                       tmp_vec(1), 1, czero, pgn(1), 1)

            CALL zgemv("N", ng, nb, cmone, res_matrix(1, 1), ng, &
                       tmp_vec(1), 1, czero, ugn(1), 1)

            DO ig = 1, ng
               pgn_norm = pgn_norm+REAL(pgn(ig), dp)*REAL(pgn(ig), dp)+ &
                          AIMAG(pgn(ig))*AIMAG(pgn(ig))
               ugn(ig) = ugn(ig)+gn(ig)
            END DO
            CALL mp_sum(pgn_norm, para_env%group)

         ELSE
            DO ig = 1, ng
               pgn(ig) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
               ugn(ig) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
               ig_global = mixing_store%ig_global_index(ig)
               DO iig = 1, ng_global
                  saa = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
                  yaa = CMPLX(0.0_dp, 0.0_dp, KIND=dp)

!            IF(ig_global==iig) ya(ig,ig_global) =  CMPLX(1.0_dp,0.0_dp,KIND=dp)
                  IF (ig_global == iig) yaa = CMPLX(1.0_dp, 0.0_dp, KIND=dp)

                  DO jb = 1, nb
                     saa = saa-step_matrix(ig, jb)*a_matrix(jb, iig)
                     yaa = yaa-res_matrix(ig, jb)*a_matrix(jb, iig)
                  END DO
                  pgn(ig) = pgn(ig)+saa*gn_global(iig)
                  ugn(ig) = ugn(ig)+yaa*gn_global(iig)
               END DO
            END DO
            DO ig = 1, ng
               pgn_norm = pgn_norm+REAL(pgn(ig), dp)*REAL(pgn(ig), dp)+ &
                          AIMAG(pgn(ig))*AIMAG(pgn(ig))
            END DO
            CALL mp_sum(pgn_norm, para_env%group)
         END IF

         gn_norm = mixing_store%norm_res_buffer(ib, ispin)
         gn_norm_old = mixing_store%norm_res_buffer(ib_prev, ispin)
         IF (ib_prev /= 0) THEN
            sigma_tilde = sigma_old*MAX(0.5_dp, MIN(2.0_dp, gn_norm_old/gn_norm))
         ELSE
            sigma_tilde = 0.5_dp
         END IF
         sigma_tilde = 0.1_dp
         ! Step size for the unpredicted component
         step_size = MIN(sigma_tilde, r_step*pgn_norm/gn_norm, sigma_max)
         sigma_old = step_size

         ! update the density
!dbg
         IF (para_env%ionode) WRITE (*, *) 'step ', step_size, pgn_norm, gn_norm, sigma_tilde
         DO ig = 1, ng
            prec = mixing_store%kerker_factor(ig)
            rho_g(ispin)%pw%cc(ig) = mixing_store%rhoin_buffer(ib, ispin)%cc(ig) &
                                     -prec*step_size*ugn(ig)+prec*pgn(ig) ! - 0.1_dp * prec* gn(ig)
            mixing_store%rhoin_buffer(ib_next, ispin)%cc(ig) = rho_g(ispin)%pw%cc(ig)
         END DO

      END DO ! ispin

      ! Deallocate  temporary arrays
      DEALLOCATE (step_matrix, res_matrix)
      DEALLOCATE (norm_res)
      DEALLOCATE (norm_res_up)
      DEALLOCATE (norm_res_low)
      DEALLOCATE (a_matrix, b_matrix, binv_matrix)
      DEALLOCATE (ugn, pgn)
      IF (use_zgemm) THEN
         DEALLOCATE (sa, ya)
      END IF
      IF (use_zgemm_rev) THEN
         DEALLOCATE (tmp_vec)
      END IF
      DEALLOCATE (gn_global, res_matrix_global)

      CALL timestop(handle)

   END SUBROUTINE multisecant_mixing

! **************************************************************************************************
!> \brief  allocation needed when density mixing is used
!> \param qs_env ...
!> \param mixing_method ...
!> \param p_mix_new ...
!> \param p_delta ...
!> \param nspins ...
!> \param mixing_store ...
!> \par History
!>      05.2009 created [MI]
!>      08.2014 kpoints [JGH]
!>      02.2015 changed input to make g-space mixing available in linear scaling
!>              SCF [Patrick Seewald]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE mixing_allocate(qs_env, mixing_method, p_mix_new, p_delta, nspins, mixing_store)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: mixing_method
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: p_mix_new, p_delta
      INTEGER, INTENT(IN)                                :: nspins
      TYPE(mixing_storage_type), POINTER                 :: mixing_store

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mixing_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iat, ic, ispin, natom, &
                                                            nbuffer, nimg
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: refmatrix
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)

      NULLIFY (matrix_s, dft_control, sab_orb, refmatrix, rho_atom)
      CALL get_qs_env(qs_env, &
                      sab_orb=sab_orb, &
                      matrix_s_kp=matrix_s, &
                      dft_control=dft_control)

      refmatrix => matrix_s(1, 1)%matrix
      nimg = dft_control%nimages

!   *** allocate p_mix_new ***
      IF (PRESENT(p_mix_new)) THEN
         IF (.NOT. ASSOCIATED(p_mix_new)) THEN
            CALL cp_dbcsr_allocate_matrix_set(p_mix_new, nspins, nimg)
            DO i = 1, nspins
               DO ic = 1, nimg
                  ALLOCATE (p_mix_new(i, ic)%matrix)
                  CALL cp_dbcsr_init(p_mix_new(i, ic)%matrix)
                  CALL cp_dbcsr_create(matrix=p_mix_new(i, ic)%matrix, &
                                       name="SCF DENSITY", &
                                       dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                       row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                       col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                       nze=0)
                  CALL cp_dbcsr_alloc_block_from_nbl(p_mix_new(i, ic)%matrix, sab_orb)
                  CALL cp_dbcsr_set(p_mix_new(i, ic)%matrix, 0.0_dp)
               ENDDO
            ENDDO
         END IF
      END IF

!   *** allocate p_delta ***
      IF (PRESENT(p_delta)) THEN
         IF (mixing_method >= gspace_mixing_nr) THEN
            IF (.NOT. ASSOCIATED(p_delta)) THEN
               CALL cp_dbcsr_allocate_matrix_set(p_delta, nspins, nimg)
               DO i = 1, nspins
                  DO ic = 1, nimg
                     ALLOCATE (p_delta(i, ic)%matrix)
                     CALL cp_dbcsr_init(p_delta(i, ic)%matrix)
                     CALL cp_dbcsr_create(matrix=p_delta(i, ic)%matrix, &
                                          name="SCF DENSITY", &
                                          dist=cp_dbcsr_distribution(refmatrix), matrix_type=dbcsr_type_symmetric, &
                                          row_blk_size=cp_dbcsr_row_block_sizes(refmatrix), &
                                          col_blk_size=cp_dbcsr_col_block_sizes(refmatrix), &
                                          nze=0)
                     CALL cp_dbcsr_alloc_block_from_nbl(p_delta(i, ic)%matrix, sab_orb)
                     CALL cp_dbcsr_set(p_delta(i, ic)%matrix, 0.0_dp)
                  ENDDO
               ENDDO
            END IF
            IF(.NOT.(ASSOCIATED(mixing_store)))CALL cp__a("qs_gspace_mixing.F",1640)
         END IF
      END IF

!   *** allocate  buffer for gspace mixing ***
      IF (mixing_method >= gspace_mixing_nr) THEN
         nbuffer = mixing_store%nbuffer
         mixing_store%ncall = 0
         IF (.NOT. ASSOCIATED(mixing_store%rhoin)) THEN
            ALLOCATE (mixing_store%rhoin(nspins))
            DO ispin = 1, nspins
               NULLIFY (mixing_store%rhoin(ispin)%cc)
            END DO
         END IF
         IF (dft_control%qs_control%gapw) THEN
            CALL get_qs_env(qs_env=qs_env, &
                            rho_atom_set=rho_atom)
            natom = SIZE(rho_atom)
            IF (.NOT. ASSOCIATED(mixing_store%paw)) THEN
               ALLOCATE (mixing_store%paw(natom))
               mixing_store%paw = .FALSE.
               ALLOCATE (mixing_store%cpc_h_in(natom, nspins))
               ALLOCATE (mixing_store%cpc_s_in(natom, nspins))
               DO ispin = 1, nspins
                  DO iat = 1, natom
                     NULLIFY (mixing_store%cpc_h_in(iat, ispin)%r_coef)
                     NULLIFY (mixing_store%cpc_s_in(iat, ispin)%r_coef)
                  END DO
               END DO
            END IF
         END IF
      END IF

!   *** allocate rhoin_buffer if needed
      IF (mixing_method == pulay_mixing_nr .OR. mixing_method == broyden_mixing_new_nr &
          .OR. mixing_method == multisecant_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%rhoin_buffer)) THEN
            ALLOCATE (mixing_store%rhoin_buffer(nbuffer, nspins))
            DO ispin = 1, nspins
               DO i = 1, nbuffer
                  NULLIFY (mixing_store%rhoin_buffer(i, ispin)%cc)
               END DO
            END DO
         END IF
      END IF

!   *** allocare res_buffer if needed
      IF (mixing_method >= pulay_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%res_buffer)) THEN
            ALLOCATE (mixing_store%res_buffer(nbuffer, nspins))
            DO ispin = 1, nspins
               DO i = 1, nbuffer
                  NULLIFY (mixing_store%res_buffer(i, ispin)%cc)
               END DO
            END DO
         END IF
      END IF

!   *** allocate pulay buffer ***
      IF (mixing_method == pulay_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%pulay_matrix)) THEN
            ALLOCATE (mixing_store%pulay_matrix(nbuffer, nbuffer))
         END IF

      END IF
!   *** allocate broyden buffer ***
      IF (mixing_method == broyden_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%rhoin_old)) THEN
            ALLOCATE (mixing_store%rhoin_old(nspins))
            DO ispin = 1, nspins
               NULLIFY (mixing_store%rhoin_old(ispin)%cc)
            END DO
         END IF
         IF (.NOT. ASSOCIATED(mixing_store%drho_buffer)) THEN
            ALLOCATE (mixing_store%drho_buffer(nbuffer, nspins))
            ALLOCATE (mixing_store%last_res(nspins))
            DO ispin = 1, nspins
               DO i = 1, nbuffer
                  NULLIFY (mixing_store%drho_buffer(i, ispin)%cc)
               END DO
               NULLIFY (mixing_store%last_res(ispin)%cc)
            END DO
         END IF
         IF (dft_control%qs_control%gapw) THEN
            IF (.NOT. ASSOCIATED(mixing_store%cpc_h_old)) THEN
               ALLOCATE (mixing_store%cpc_h_old(natom, nspins))
               ALLOCATE (mixing_store%cpc_s_old(natom, nspins))
               DO ispin = 1, nspins
                  DO iat = 1, natom
                     NULLIFY (mixing_store%cpc_h_old(iat, ispin)%r_coef)
                     NULLIFY (mixing_store%cpc_s_old(iat, ispin)%r_coef)
                  END DO
               END DO
            END IF
            IF (.NOT. ASSOCIATED(mixing_store%dcpc_h_in)) THEN
               ALLOCATE (mixing_store%dcpc_h_in(nbuffer, natom, nspins))
               ALLOCATE (mixing_store%dcpc_s_in(nbuffer, natom, nspins))
               ALLOCATE (mixing_store%cpc_h_lastres(natom, nspins))
               ALLOCATE (mixing_store%cpc_s_lastres(natom, nspins))
               DO ispin = 1, nspins
                  DO iat = 1, natom
                     DO i = 1, nbuffer
                        NULLIFY (mixing_store%dcpc_h_in(i, iat, ispin)%r_coef)
                        NULLIFY (mixing_store%dcpc_s_in(i, iat, ispin)%r_coef)
                     END DO
                     NULLIFY (mixing_store%cpc_h_lastres(iat, ispin)%r_coef)
                     NULLIFY (mixing_store%cpc_s_lastres(iat, ispin)%r_coef)
                  END DO
               END DO
            END IF
         END IF
      END IF
!   *** allocate broyden buffer ***
      IF (mixing_method == broyden_mixing_new_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%u_vec)) THEN
            ALLOCATE (mixing_store%last_res(nspins))
            ALLOCATE (mixing_store%delta_res(nbuffer-1, nspins))
            ALLOCATE (mixing_store%u_vec(nbuffer-1, nspins))
            ALLOCATE (mixing_store%z_vec(nbuffer-1, nspins))
            ALLOCATE (mixing_store%weight(nbuffer, nspins))
            ALLOCATE (mixing_store%fmat(nbuffer-1, nbuffer-1, nspins))
            DO ispin = 1, nspins
            DO i = 1, nbuffer-1
               NULLIFY (mixing_store%delta_res(i, ispin)%cc)
               NULLIFY (mixing_store%u_vec(i, ispin)%cc)
               NULLIFY (mixing_store%z_vec(i, ispin)%cc)
            END DO
            NULLIFY (mixing_store%last_res(ispin)%cc)
            END DO
         END IF
      END IF

!   *** allocate multisecant buffer ***
      IF (mixing_method == multisecant_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%norm_res_buffer)) THEN
            ALLOCATE (mixing_store%norm_res_buffer(nbuffer, nspins))
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE mixing_allocate

! **************************************************************************************************
!> \brief  initialiation needed when gspace mixing is used
!> \param mixing_method ...
!> \param rho ...
!> \param mixing_store ...
!> \param para_env ...
!> \param rho_atom ...
!> \par History
!>      05.2009 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE mixing_init(mixing_method, rho, mixing_store, para_env, rho_atom)
      INTEGER, INTENT(IN)                                :: mixing_method
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'mixing_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iat, ib, ig, ig1, ig_count, &
                                                            iproc, ispin, n1, n2, natom, nbuffer, &
                                                            ng, nspin
      REAL(dp)                                           :: bconst, beta, fdamp, g2max, g2min, kmin
      REAL(dp), DIMENSION(:), POINTER                    :: g2
      REAL(dp), DIMENSION(:, :), POINTER                 :: g_vec
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g

      CALL timeset(routineN, handle)

      NULLIFY (g2, g_vec, rho_g)
      CALL qs_rho_get(rho, rho_g=rho_g)

      nspin = SIZE(rho_g)
      ng = SIZE(rho_g(1)%pw%pw_grid%gsq, 1)
      mixing_store%ig_max = ng
      g2 => rho_g(1)%pw%pw_grid%gsq
      g_vec => rho_g(1)%pw%pw_grid%g

      IF (mixing_store%max_gvec_exp > 0._dp) THEN
         DO ig = 1, ng
            IF (g2(ig) > mixing_store%max_g2) THEN
               mixing_store%ig_max = ig
               EXIT
            END IF
         END DO
      END IF

      IF (.NOT. ASSOCIATED(mixing_store%kerker_factor)) THEN
         ALLOCATE (mixing_store%kerker_factor(ng))
      END IF
      IF (.NOT. ASSOCIATED(mixing_store%special_metric)) THEN
         ALLOCATE (mixing_store%special_metric(ng))
      END IF
      beta = mixing_store%beta
      kmin = 0.1_dp
      mixing_store%kerker_factor = 1.0_dp
      mixing_store%special_metric = 1.0_dp
      ig1 = 1
      IF (rho_g(1)%pw%pw_grid%have_g0) ig1 = 2
      DO ig = ig1, mixing_store%ig_max
         mixing_store%kerker_factor(ig) = MAX(g2(ig)/(g2(ig)+beta*beta), kmin)
         mixing_store%special_metric(ig) = &
            1.0_dp+50.0_dp/8.0_dp*( &
            1.0_dp+COS(g_vec(1, ig))+COS(g_vec(2, ig))+COS(g_vec(3, ig))+ &
            COS(g_vec(1, ig))*COS(g_vec(2, ig))+ &
            COS(g_vec(2, ig))*COS(g_vec(3, ig))+ &
            COS(g_vec(1, ig))*COS(g_vec(3, ig))+ &
            COS(g_vec(1, ig))*COS(g_vec(2, ig))*COS(g_vec(3, ig)))
      END DO

      nbuffer = mixing_store%nbuffer
      DO ispin = 1, nspin
         IF (.NOT. ASSOCIATED(mixing_store%rhoin(ispin)%cc)) THEN
            ALLOCATE (mixing_store%rhoin(ispin)%cc(ng))
         END IF
         mixing_store%rhoin(ispin)%cc = rho_g(ispin)%pw%cc

         IF (ASSOCIATED(mixing_store%rhoin_buffer)) THEN
            IF (.NOT. ASSOCIATED(mixing_store%rhoin_buffer(1, ispin)%cc)) THEN
               DO ib = 1, nbuffer
                  ALLOCATE (mixing_store%rhoin_buffer(ib, ispin)%cc(ng))
               END DO
            END IF
            mixing_store%rhoin_buffer(1, ispin)%cc(1:ng) = &
               rho_g(ispin)%pw%cc(1:ng)
         END IF
         IF (ASSOCIATED(mixing_store%res_buffer)) THEN
            IF (.NOT. ASSOCIATED(mixing_store%res_buffer(1, ispin)%cc)) THEN
               DO ib = 1, nbuffer
                  ALLOCATE (mixing_store%res_buffer(ib, ispin)%cc(ng))
               END DO
            END IF
         END IF
      END DO

      IF (nspin == 2) THEN
         mixing_store%rhoin(1)%cc = rho_g(1)%pw%cc+rho_g(2)%pw%cc
         mixing_store%rhoin(2)%cc = rho_g(1)%pw%cc-rho_g(2)%pw%cc
         IF (ASSOCIATED(mixing_store%rhoin_buffer)) THEN
            mixing_store%rhoin_buffer(1, 1)%cc = rho_g(1)%pw%cc+rho_g(2)%pw%cc
            mixing_store%rhoin_buffer(1, 2)%cc = rho_g(1)%pw%cc-rho_g(2)%pw%cc
         END IF
      END IF

      IF (PRESENT(rho_atom)) THEN
         natom = SIZE(rho_atom)
         DO ispin = 1, nspin
         DO iat = 1, natom
            IF (ASSOCIATED(rho_atom(iat)%cpc_s(ispin)%r_coef)) THEN
               mixing_store%paw(iat) = .TRUE.
               n1 = SIZE(rho_atom(iat)%cpc_s(ispin)%r_coef, 1)
               n2 = SIZE(rho_atom(iat)%cpc_s(ispin)%r_coef, 2)
               IF (ASSOCIATED(mixing_store%cpc_s_in)) THEN
                  IF (.NOT. ASSOCIATED(mixing_store%cpc_s_in(iat, ispin)%r_coef)) THEN
                     ALLOCATE (mixing_store%cpc_s_in(iat, ispin)%r_coef(n1, n2))
                     ALLOCATE (mixing_store%cpc_h_in(iat, ispin)%r_coef(n1, n2))
                  END IF
                  mixing_store%cpc_h_in(iat, ispin)%r_coef = rho_atom(iat)%cpc_h(ispin)%r_coef
                  mixing_store%cpc_s_in(iat, ispin)%r_coef = rho_atom(iat)%cpc_s(ispin)%r_coef
               END IF
            END IF
         END DO
         END DO
      END IF

      IF (mixing_method == gspace_mixing_nr) THEN
      ELSEIF (mixing_method == pulay_mixing_nr) THEN
      ELSEIF (mixing_method == broyden_mixing_nr) THEN
         DO ispin = 1, nspin
            IF (.NOT. ASSOCIATED(mixing_store%rhoin_old(ispin)%cc)) THEN
               ALLOCATE (mixing_store%rhoin_old(ispin)%cc(ng))
            END IF
            IF (.NOT. ASSOCIATED(mixing_store%drho_buffer(1, ispin)%cc)) THEN
               DO ib = 1, nbuffer
                  ALLOCATE (mixing_store%drho_buffer(ib, ispin)%cc(ng))
               END DO
               ALLOCATE (mixing_store%last_res(ispin)%cc(ng))
            END IF
            DO ib = 1, nbuffer
               mixing_store%drho_buffer(ib, ispin)%cc = CMPLX(0.0_dp, 0.0_dp, kind=dp)
            END DO
            mixing_store%last_res(ispin)%cc = CMPLX(0.0_dp, 0.0_dp, kind=dp)
            mixing_store%rhoin_old(ispin)%cc = CMPLX(0.0_dp, 0.0_dp, kind=dp)
         END DO
         IF (PRESENT(rho_atom)) THEN
         DO ispin = 1, nspin
         DO iat = 1, natom
            IF (mixing_store%paw(iat)) THEN
               n1 = SIZE(rho_atom(iat)%cpc_s(ispin)%r_coef, 1)
               n2 = SIZE(rho_atom(iat)%cpc_s(ispin)%r_coef, 2)
               IF (.NOT. ASSOCIATED(mixing_store%cpc_s_old(iat, ispin)%r_coef)) THEN
                  ALLOCATE (mixing_store%cpc_s_old(iat, ispin)%r_coef(n1, n2))
                  ALLOCATE (mixing_store%cpc_h_old(iat, ispin)%r_coef(n1, n2))
               END IF
               mixing_store%cpc_h_old(iat, ispin)%r_coef = 0.0_dp
               mixing_store%cpc_s_old(iat, ispin)%r_coef = 0.0_dp
               IF (.NOT. ASSOCIATED(mixing_store%dcpc_s_in(1, iat, ispin)%r_coef)) THEN
                  DO ib = 1, nbuffer
                     ALLOCATE (mixing_store%dcpc_h_in(ib, iat, ispin)%r_coef(n1, n2))
                     ALLOCATE (mixing_store%dcpc_s_in(ib, iat, ispin)%r_coef(n1, n2))
                  END DO
                  ALLOCATE (mixing_store%cpc_h_lastres(iat, ispin)%r_coef(n1, n2))
                  ALLOCATE (mixing_store%cpc_s_lastres(iat, ispin)%r_coef(n1, n2))
               END IF
               DO ib = 1, nbuffer
                  mixing_store%dcpc_h_in(ib, iat, ispin)%r_coef = 0.0_dp
                  mixing_store%dcpc_s_in(ib, iat, ispin)%r_coef = 0.0_dp
               END DO
               mixing_store%cpc_h_lastres(iat, ispin)%r_coef = 0.0_dp
               mixing_store%cpc_s_lastres(iat, ispin)%r_coef = 0.0_dp
            END IF
         END DO
         END DO
         END IF

         IF (.NOT. ASSOCIATED(mixing_store%p_metric)) THEN
            ALLOCATE (mixing_store%p_metric(ng))
            bconst = mixing_store%bconst
            g2min = 1.E30_dp
            DO ig = 1, ng
               IF (g2(ig) > 1.E-10_dp) g2min = MIN(g2min, g2(ig))
            END DO
            g2max = -1.E30_dp
            DO ig = 1, ng
               g2max = MAX(g2max, g2(ig))
            END DO
            CALL mp_min(g2min, para_env%group)
            CALL mp_max(g2max, para_env%group)
!      fdamp/g2 varies between (bconst-1) and 0
!      i.e. p_metric varies between bconst and 1
!         fdamp = (bconst-1.0_dp)*g2min
            fdamp = (bconst-1.0_dp)*g2min*g2max/(g2max-g2min*bconst)
            DO ig = 1, ng
               mixing_store%p_metric(ig) = (g2(ig)+fdamp)/MAX(g2(ig), 1.E-10_dp)
            END DO
            IF (rho_g(1)%pw%pw_grid%have_g0) mixing_store%p_metric(1) = bconst
         END IF

      ELSEIF (mixing_method == broyden_mixing_new_nr) THEN
         DO ispin = 1, nspin
            IF (.NOT. ASSOCIATED(mixing_store%u_vec(1, ispin)%cc)) THEN
               DO ib = 1, nbuffer-1
                  ALLOCATE (mixing_store%delta_res(ib, ispin)%cc(ng))
                  ALLOCATE (mixing_store%u_vec(ib, ispin)%cc(ng))
                  ALLOCATE (mixing_store%z_vec(ib, ispin)%cc(ng))
               END DO
               ALLOCATE (mixing_store%last_res(ispin)%cc(ng))
            END IF
         END DO
         IF (.NOT. ASSOCIATED(mixing_store%p_metric)) THEN
            ALLOCATE (mixing_store%p_metric(ng))
            bconst = mixing_store%bconst
            g2min = 1.E30_dp
            DO ig = 1, ng
               IF (g2(ig) > 1.E-10_dp) g2min = MIN(g2min, g2(ig))
            END DO
            CALL mp_min(g2min, para_env%group)
            fdamp = (bconst-1.0_dp)*g2min
            DO ig = 1, ng
               mixing_store%p_metric(ig) = (g2(ig)+fdamp)/MAX(g2(ig), 1.E-10_dp)
            END DO
            IF (rho_g(1)%pw%pw_grid%have_g0) mixing_store%p_metric(1) = bconst
!dbg
!        mixing_store%p_metric = 1.0_dp
!dbg
         END IF
      ELSEIF (mixing_method == multisecant_mixing_nr) THEN
         IF (.NOT. ASSOCIATED(mixing_store%ig_global_index)) THEN
            ALLOCATE (mixing_store%ig_global_index(ng))
         END IF
         mixing_store%ig_global_index = 0
         ig_count = 0
         DO iproc = 0, para_env%num_pe-1
            IF (para_env%mepos == iproc) THEN
               DO ig = 1, ng
                  ig_count = ig_count+1
                  mixing_store%ig_global_index(ig) = ig_count
               END DO
            END IF
            CALL mp_bcast(ig_count, iproc, para_env%group)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE mixing_init

END MODULE qs_gspace_mixing
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Different diagonalization schemes that can be used
!>        for the iterative solution of the eigenvalue problem
!> \par History
!>      started from routines previously located in the qs_scf module
!>      05.2009
! **************************************************************************************************
MODULE qs_scf_diagonalization
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_add_fm
   USE cp_cfm_diag,                     ONLY: cp_cfm_geeig,&
                                              cp_cfm_geeig_canon
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_release,&
                                              cp_cfm_to_fm,&
                                              cp_cfm_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix, cp_dbcsr_desymmetrize, cp_dbcsr_distribution, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_type_antisymmetric, dbcsr_type_no_symmetry, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_symm,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_geeig,&
                                              cp_fm_geeig_canon
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_create_fm,&
                                              fm_pool_give_back_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_add_to_element,&
                                              cp_fm_copy_general,&
                                              cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        cholesky_dbcsr, cholesky_inverse, cholesky_off, cholesky_reduce, cholesky_restore, &
        core_guess, general_roks, high_spin_roks, restart_guess
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE kpoint_methods,                  ONLY: kpoint_density_matrices,&
                                              kpoint_density_transform,&
                                              kpoint_set_mo_occupation,&
                                              rskp_transform
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_env_type,&
                                              kpoint_type
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE preconditioner,                  ONLY: prepare_preconditioner,&
                                              restart_preconditioner
   USE qs_density_mixing_types,         ONLY: direct_mixing_nr,&
                                              gspace_mixing_nr
   USE qs_diis,                         ONLY: qs_diis_b_step
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gspace_mixing,                ONLY: gspace_mixing,&
                                              mixing_allocate,&
                                              mixing_init,&
                                              self_consistency_check
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_matrix_pools,                 ONLY: mpools_get,&
                                              qs_matrix_pools_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix,&
                                              calculate_subspace_eigenvalues
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_ot_eigensolver,               ONLY: ot_eigensolver
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_block_davidson,           ONLY: generate_extended_space,&
                                              generate_extended_space_sparse
   USE qs_scf_lanczos,                  ONLY: lanczos_refinement,&
                                              lanczos_refinement_2v
   USE qs_scf_methods,                  ONLY: combine_ks_matrices,&
                                              eigensolver,&
                                              eigensolver_dbcsr,&
                                              eigensolver_simple,&
                                              eigensolver_symm,&
                                              scf_env_density_mixing
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              subspace_env_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_diagonalization'

   PUBLIC :: do_general_diag, do_general_diag_kp, do_roks_diag, &
             do_special_diag, do_ot_diag, do_block_davidson_diag, &
             do_block_krylov_diag, do_scf_diag_subspace, diag_subspace_allocate, general_eigenproblem

CONTAINS

! **************************************************************************************************
!> \brief the inner loop of scf, specific to diagonalization with S matrix
!>       basically, in goes the ks matrix out goes a new p matrix
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************

   SUBROUTINE general_eigenproblem(scf_env, mos, matrix_ks, &
                                   matrix_s, scf_control, scf_section, &
                                   diis_step)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(len=*), PARAMETER :: routineN = 'general_eigenproblem', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspin
      LOGICAL                                            :: do_level_shift, use_jacobi
      REAL(kind=dp)                                      :: diis_error, eps_diis
      TYPE(cp_dbcsr_type), POINTER                       :: ortho_dbcsr
      TYPE(cp_fm_type), POINTER                          :: matrix_u_fm, ortho

      nspin = SIZE(matrix_ks)
      NULLIFY (ortho, ortho_dbcsr)

      DO ispin = 1, nspin
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, &
                               scf_env%scf_work1(ispin)%matrix)
      END DO

      eps_diis = scf_control%eps_diis

      IF (scf_env%iter_count > 1 .AND. .NOT. scf_env%skip_diis) THEN
         CALL qs_diis_b_step(scf_env%scf_diis_buffer, mos, scf_env%scf_work1, &
                             scf_env%scf_work2, scf_env%iter_delta, diis_error, diis_step, &
                             eps_diis, scf_control%nmixing, &
                             s_matrix=matrix_s, &
                             scf_section=scf_section)
      ELSE
         diis_step = .FALSE.
      END IF

      do_level_shift = ((scf_control%level_shift /= 0.0_dp) .AND. &
                        ((scf_control%density_guess == core_guess) .OR. &
                         (scf_env%iter_count > 1)))

      IF ((scf_env%iter_count > 1) .AND. &
          (scf_env%iter_delta < scf_control%diagonalization%eps_jacobi)) THEN
         use_jacobi = .TRUE.
      ELSE
         use_jacobi = .FALSE.
      END IF

      IF (diis_step) THEN
         scf_env%iter_param = diis_error
         IF (use_jacobi) THEN
            scf_env%iter_method = "DIIS/Jacobi"
         ELSE
            scf_env%iter_method = "DIIS/Diag."
         END IF
      ELSE
         IF (scf_env%mixing_method == 0) THEN
            scf_env%iter_method = "NoMix/Diag."
         ELSE IF (scf_env%mixing_method == 1) THEN
            scf_env%iter_param = scf_env%p_mix_alpha
            IF (use_jacobi) THEN
               scf_env%iter_method = "P_Mix/Jacobi"
            ELSE
               scf_env%iter_method = "P_Mix/Diag."
            END IF
         ELSEIF (scf_env%mixing_method > 1) THEN
            scf_env%iter_param = scf_env%mixing_store%alpha
            IF (use_jacobi) THEN
               scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Jacobi"
            ELSE
               scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Diag."
            END IF
         END IF
      END IF

      IF (scf_env%cholesky_method == cholesky_dbcsr) THEN
         ortho_dbcsr => scf_env%ortho_dbcsr
         DO ispin = 1, nspin
            CALL eigensolver_dbcsr(matrix_ks=matrix_ks(ispin)%matrix, matrix_ks_fm=scf_env%scf_work1(ispin)%matrix, &
                                   mo_set=mos(ispin)%mo_set, &
                                   ortho_dbcsr=ortho_dbcsr, &
                                   ksbuf1=scf_env%buf1_dbcsr, ksbuf2=scf_env%buf2_dbcsr, &
                                   work=scf_env%scf_work2)
         END DO

      ELSE IF (scf_env%cholesky_method > cholesky_off) THEN
         DO ispin = 1, nspin
            IF (scf_env%cholesky_method == cholesky_inverse) THEN
               ortho => scf_env%ortho_m1
            ELSE
               ortho => scf_env%ortho
            END IF

            IF (do_level_shift) THEN
               matrix_u_fm => scf_env%ortho
            ELSE
               NULLIFY (matrix_u_fm)
            END IF

            CALL eigensolver(matrix_ks_fm=scf_env%scf_work1(ispin)%matrix, &
                             mo_set=mos(ispin)%mo_set, &
                             ortho=ortho, &
                             work=scf_env%scf_work2, &
                             cholesky_method=scf_env%cholesky_method, &
                             do_level_shift=do_level_shift, &
                             level_shift=scf_control%level_shift, &
                             matrix_u_fm=matrix_u_fm, &
                             use_jacobi=use_jacobi)
         END DO
      ELSE
         ortho => scf_env%ortho

         IF (do_level_shift) THEN
            matrix_u_fm => scf_env%ortho_m1
         ELSE
            NULLIFY (matrix_u_fm)
         END IF

         DO ispin = 1, nspin
            CALL eigensolver_symm(matrix_ks_fm=scf_env%scf_work1(ispin)%matrix, &
                                  mo_set=mos(ispin)%mo_set, &
                                  ortho=ortho, &
                                  work=scf_env%scf_work2, &
                                  do_level_shift=do_level_shift, &
                                  level_shift=scf_control%level_shift, &
                                  matrix_u_fm=matrix_u_fm, &
                                  use_jacobi=use_jacobi, &
                                  jacobi_threshold=scf_control%diagonalization%jacobi_threshold)
         END DO
      END IF

   END SUBROUTINE general_eigenproblem

! **************************************************************************************************
!> \brief ...
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
! **************************************************************************************************
   SUBROUTINE do_general_diag(scf_env, mos, matrix_ks, &
                              matrix_s, scf_control, scf_section, &
                              diis_step)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(len=*), PARAMETER :: routineN = 'do_general_diag', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspin

      nspin = SIZE(matrix_ks)

      CALL general_eigenproblem(scf_env, mos, matrix_ks, &
                                matrix_s, scf_control, scf_section, diis_step)

      CALL set_mo_occupation(mo_array=mos, &
                             smear=scf_control%smear)

      DO ispin = 1, nspin
         CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                       scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

   END SUBROUTINE do_general_diag

! **************************************************************************************************
!> \brief Kpoint diagonalization routine
!>        Transforms matrices to kpoint, distributes kpoint groups, performs
!>        general diagonalization (no storgae of overlap decomposition), stores
!>        MOs, calculates occupation numbers, calculates density matrices
!>        in kpoint representation, transforms density matrices to real space
!> \param scf_env      SCF environment
!> \param matrix_ks    Kohn-sham matrices (RS indices, global)
!> \param matrix_s     Overlap matrices (RS indices, global)
!> \param kpoints      Kpoint environment
!> \param scf_control  SCF control variables
!> \param diis_step    Have we done a DIIS step
!> \par History
!>      08.2014 created [JGH]
! **************************************************************************************************
   SUBROUTINE do_general_diag_kp(scf_env, matrix_ks, matrix_s, kpoints, &
                                 scf_control, diis_step)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks, matrix_s
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(scf_control_type), POINTER                    :: scf_control
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(len=*), PARAMETER :: routineN = 'do_general_diag_kp', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER :: cone = CMPLX(1.0_dp, 0.0_dp, KIND=dp), &
         czero = CMPLX(0.0_dp, 0.0_dp, KIND=dp), ione = CMPLX(0.0_dp, 1.0_dp, KIND=dp)

      INTEGER                                            :: bcntxt, handle, ik, ikp, ikpgr, ispin, &
                                                            kplocal, nkp, nkp_groups, nspin
      INTEGER, DIMENSION(2)                              :: kp_range
      INTEGER, DIMENSION(:, :), POINTER                  :: kp_dist
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: my_kpgrp, use_real_wfn
      REAL(kind=dp), DIMENSION(:), POINTER               :: eigenvalues
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: xkp
      TYPE(cp_cfm_type), POINTER                         :: cksmat, cmos, csmat, cwork
      TYPE(cp_dbcsr_type), POINTER                       :: cmatrix, rmatrix, tmpmat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fmwork
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_ao_fm_pools
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct, mo_struct
      TYPE(cp_fm_type), POINTER                          :: fmdummy, fmlocal, imos, mo_coeff, &
                                                            rksmat, rmos, rsmat
      TYPE(kpoint_env_type), POINTER                     :: kp
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      TYPE(qs_matrix_pools_type), POINTER                :: mpools

      CALL timeset(routineN, handle)

      diis_step = .FALSE.
      nspin = SIZE(matrix_ks, 1)
      NULLIFY (fmdummy)

      IF (scf_env%mixing_method == 0) THEN
         scf_env%iter_method = "NoMix/Diag."
      ELSE IF (scf_env%mixing_method == 1) THEN
         scf_env%iter_param = scf_env%p_mix_alpha
         scf_env%iter_method = "P_Mix/Diag."
      ELSEIF (scf_env%mixing_method > 1) THEN
         scf_env%iter_param = scf_env%mixing_store%alpha
         scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Diag."
      END IF

      NULLIFY (sab_nl)
      CALL get_kpoint_info(kpoints, nkp=nkp, xkp=xkp, use_real_wfn=use_real_wfn, kp_range=kp_range, &
                           nkp_groups=nkp_groups, kp_dist=kp_dist, sab_nl=sab_nl, &
                           cell_to_index=cell_to_index)
      IF(.NOT.(ASSOCIATED(sab_nl)))CALL cp__a("qs_scf_diagonalization.F",388)
      kplocal = kp_range(2)-kp_range(1)+1

      ! allocate some work matrices
      ALLOCATE (rmatrix, cmatrix, tmpmat)
      CALL cp_dbcsr_init(rmatrix)
      CALL cp_dbcsr_init(cmatrix)
      CALL cp_dbcsr_init(tmpmat)
      CALL cp_dbcsr_create(rmatrix, template=matrix_ks(1, 1)%matrix, &
                           matrix_type=dbcsr_type_symmetric)
      CALL cp_dbcsr_create(cmatrix, template=matrix_ks(1, 1)%matrix, &
                           matrix_type=dbcsr_type_antisymmetric)
      CALL cp_dbcsr_create(tmpmat, template=matrix_ks(1, 1)%matrix, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_alloc_block_from_nbl(rmatrix, sab_nl)
      CALL cp_dbcsr_alloc_block_from_nbl(cmatrix, sab_nl)

      fmwork => scf_env%scf_work1

      ! fm pools to be used within a kpoint group
      CALL get_kpoint_info(kpoints, mpools=mpools)
      CALL mpools_get(mpools, ao_ao_fm_pools=ao_ao_fm_pools)

      CALL fm_pool_create_fm(ao_ao_fm_pools(1)%pool, fmlocal)
      CALL cp_fm_get_info(fmlocal, matrix_struct=matrix_struct)
      IF (use_real_wfn) THEN
         CALL cp_fm_create(rksmat, matrix_struct)
         CALL cp_fm_create(rsmat, matrix_struct)
      ELSE
         CALL cp_cfm_create(cksmat, matrix_struct)
         CALL cp_cfm_create(csmat, matrix_struct)
         CALL cp_cfm_create(cwork, matrix_struct)
         kp => kpoints%kp_env(1)%kpoint_env
         CALL get_mo_set(kp%mos(1, 1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, matrix_struct=mo_struct)
         CALL cp_cfm_create(cmos, mo_struct)
      END IF

      ! global blacs context (=groupid)
      bcntxt = kpoints%blacs_env_all%group

      DO ikpgr = 1, kplocal
         DO ispin = 1, nspin
            DO ikp = 1, nkp_groups
               ! number of current kpoint
               ik = kp_dist(1, ikp)+ikpgr-1
               my_kpgrp = (ik >= kpoints%kp_range(1) .AND. ik <= kpoints%kp_range(2))
               IF (use_real_wfn) THEN
                  ! FT of matrices KS and S, then transfer to FM type
                  CALL rskp_transform(rmatrix=rmatrix, rsmat=matrix_ks, ispin=ispin, &
                                      xkp=xkp(1:3, ik), cell_to_index=cell_to_index, sab_nl=sab_nl)
                  CALL cp_dbcsr_desymmetrize(rmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(1)%matrix)
                  ! s matrix is not spin dependent
                  CALL rskp_transform(rmatrix=rmatrix, rsmat=matrix_s, ispin=1, &
                                      xkp=xkp(1:3, ik), cell_to_index=cell_to_index, sab_nl=sab_nl)
                  CALL cp_dbcsr_desymmetrize(rmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(3)%matrix)
               ELSE
                  ! FT of matrices KS and S, then transfer to FM type
                  CALL rskp_transform(rmatrix=rmatrix, cmatrix=cmatrix, rsmat=matrix_ks, ispin=ispin, &
                                      xkp=xkp(1:3, ik), cell_to_index=cell_to_index, sab_nl=sab_nl)
                  CALL cp_dbcsr_desymmetrize(rmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(1)%matrix)
                  CALL cp_dbcsr_desymmetrize(cmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(2)%matrix)
                  ! s matrix is not spin dependent, double the work
                  CALL rskp_transform(rmatrix=rmatrix, cmatrix=cmatrix, rsmat=matrix_s, ispin=1, &
                                      xkp=xkp(1:3, ik), cell_to_index=cell_to_index, sab_nl=sab_nl)
                  CALL cp_dbcsr_desymmetrize(rmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(3)%matrix)
                  CALL cp_dbcsr_desymmetrize(cmatrix, tmpmat)
                  CALL copy_dbcsr_to_fm(tmpmat, fmwork(4)%matrix)
               END IF
               ! transfer to kpoint group
               ! redistribution of matrices, new blacs environment
               ! fmwork -> fmlocal -> rksmat/cksmat
               ! fmwork -> fmlocal -> rsmat/csmat
               IF (use_real_wfn) THEN
                  IF (my_kpgrp) THEN
                     CALL cp_fm_copy_general(fmwork(1)%matrix, rksmat, bcntxt)
                     CALL cp_fm_copy_general(fmwork(3)%matrix, rsmat, bcntxt)
                  ELSE
                     CALL cp_fm_copy_general(fmwork(1)%matrix, fmdummy, bcntxt)
                     CALL cp_fm_copy_general(fmwork(3)%matrix, fmdummy, bcntxt)
                  END IF
               ELSE
                  IF (my_kpgrp) THEN
                     CALL cp_fm_copy_general(fmwork(1)%matrix, fmlocal, bcntxt)
                     CALL cp_cfm_add_fm(czero, cksmat, cone, fmlocal)
                     CALL cp_fm_copy_general(fmwork(2)%matrix, fmlocal, bcntxt)
                     CALL cp_cfm_add_fm(cone, cksmat, ione, fmlocal)
                     CALL cp_fm_copy_general(fmwork(3)%matrix, fmlocal, bcntxt)
                     CALL cp_cfm_add_fm(czero, csmat, cone, fmlocal)
                     CALL cp_fm_copy_general(fmwork(4)%matrix, fmlocal, bcntxt)
                     CALL cp_cfm_add_fm(cone, csmat, ione, fmlocal)
                  ELSE
                     CALL cp_fm_copy_general(fmwork(1)%matrix, fmdummy, bcntxt)
                     CALL cp_fm_copy_general(fmwork(2)%matrix, fmdummy, bcntxt)
                     CALL cp_fm_copy_general(fmwork(3)%matrix, fmdummy, bcntxt)
                     CALL cp_fm_copy_general(fmwork(4)%matrix, fmdummy, bcntxt)
                  END IF
               END IF
            END DO
            ! Each kpoint group has now information on a kpoint to be diagonalized
            ! General eigensolver Hermite or Symmetric
            kp => kpoints%kp_env(ikpgr)%kpoint_env
            IF (use_real_wfn) THEN
               CALL get_mo_set(kp%mos(1, ispin)%mo_set, mo_coeff=mo_coeff, eigenvalues=eigenvalues)
               IF (scf_env%cholesky_method == cholesky_off) THEN
                  CALL cp_fm_geeig_canon(rksmat, rsmat, mo_coeff, eigenvalues, fmlocal, &
                                         scf_control%eps_eigval)
               ELSE
                  CALL cp_fm_geeig(rksmat, rsmat, mo_coeff, eigenvalues, fmlocal)
               END IF
            ELSE
               CALL get_mo_set(kp%mos(1, ispin)%mo_set, mo_coeff=rmos, eigenvalues=eigenvalues)
               CALL get_mo_set(kp%mos(2, ispin)%mo_set, mo_coeff=imos)
               IF (scf_env%cholesky_method == cholesky_off) THEN
                  CALL cp_cfm_geeig_canon(cksmat, csmat, cmos, eigenvalues, cwork, &
                                          scf_control%eps_eigval)
               ELSE
                  CALL cp_cfm_geeig(cksmat, csmat, cmos, eigenvalues, cwork)
               END IF
               ! copy eigenvalues to imag set (keep them in sync)
               kp%mos(2, ispin)%mo_set%eigenvalues = eigenvalues
               ! split real and imaginary part of mos
               CALL cp_cfm_to_fm(cmos, rmos, imos)
            END IF
         END DO
      END DO
      ! MO occupations
      CALL kpoint_set_mo_occupation(kpoints, scf_control%smear)

      ! density matrices
      CALL kpoint_density_matrices(kpoints)
      ! density matrices in real space
      CALL kpoint_density_transform(kpoints, scf_env%p_mix_new, .FALSE., &
                                    matrix_s(1, 1)%matrix, sab_nl, fmwork)

      CALL cp_dbcsr_deallocate_matrix(rmatrix)
      CALL cp_dbcsr_deallocate_matrix(cmatrix)
      CALL cp_dbcsr_deallocate_matrix(tmpmat)

      IF (use_real_wfn) THEN
         CALL cp_fm_release(rksmat)
         CALL cp_fm_release(rsmat)
      ELSE
         CALL cp_cfm_release(cksmat)
         CALL cp_cfm_release(csmat)
         CALL cp_cfm_release(cwork)
         CALL cp_cfm_release(cmos)
      END IF
      CALL fm_pool_give_back_fm(ao_ao_fm_pools(1)%pool, fmlocal)

      CALL timestop(handle)

   END SUBROUTINE do_general_diag_kp

! **************************************************************************************************
!> \brief inner loop within MOS subspace, to refine occupation and density,
!>        before next diagonalization of the Hamiltonian
!> \param qs_env ...
!> \param scf_env ...
!> \param subspace_env ...
!> \param mos ...
!> \param rho ...
!> \param ks_env ...
!> \param scf_section ...
!> \param scf_control ...
!> \par History
!>      09.2009 created [MI]
!> \note  it is assumed that when diagonalization is used, also some mixing procedure is active
! **************************************************************************************************
   SUBROUTINE do_scf_diag_subspace(qs_env, scf_env, subspace_env, mos, rho, &
                                   ks_env, scf_section, scf_control)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(subspace_env_type), POINTER                   :: subspace_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(section_vals_type), POINTER                   :: scf_section
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_scf_diag_subspace', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, i, iloop, ispin, nao, nmo, &
                                                            nspin, output_unit
      LOGICAL                                            :: converged
      REAL(dp)                                           :: ene_diff, ene_old, iter_delta, max_val, &
                                                            sum_band, sum_val, t1, t2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues, mo_occupations
      TYPE(cp_1d_r_p_type), ALLOCATABLE, DIMENSION(:)    :: eval_first, occ_first
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(cp_fm_type), POINTER                          :: c0, chc, evec, mo_coeff, work
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)
      NULLIFY (c0, chc, energy, evec, matrix_ks, mo_coeff, mo_eigenvalues, &
               mo_occupations, work, dft_control, rho_ao, rho_ao_kp)

      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%DIAG_SUB_SCF", &
                                         extension=".scfLog")

      !Extra loop keeping mos unchanged and refining the subspace occupation
      work => scf_env%scf_work2

      nspin = SIZE(mos)
      CALL qs_rho_get(rho, rho_ao=rho_ao, rho_ao_kp=rho_ao_kp)

      ALLOCATE (eval_first(nspin))
      ALLOCATE (occ_first(nspin))
      DO ispin = 1, nspin
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                         nmo=nmo, &
                         eigenvalues=mo_eigenvalues, &
                         occupation_numbers=mo_occupations)
         ALLOCATE (eval_first(ispin)%array(nmo))
         ALLOCATE (occ_first(ispin)%array(nmo))
         eval_first(ispin)%array(1:nmo) = mo_eigenvalues(1:nmo)
         occ_first(ispin)%array(1:nmo) = mo_occupations(1:nmo)
      END DO

      DO ispin = 1, nspin
         ! does not yet handle k-points
         CALL cp_dbcsr_copy(subspace_env%p_matrix_store(ispin)%matrix, rho_ao(ispin)%matrix)
         CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

      subspace_env%p_matrix_mix => scf_env%p_mix_new

      NULLIFY (matrix_ks, energy, para_env, matrix_s)
      CALL get_qs_env(qs_env, &
                      matrix_ks=matrix_ks, &
                      energy=energy, &
                      matrix_s=matrix_s, &
                      para_env=para_env, &
                      dft_control=dft_control)

      ! mixing storage allocation
      IF (subspace_env%mixing_method >= gspace_mixing_nr) THEN
         CALL mixing_allocate(qs_env, subspace_env%mixing_method, scf_env%p_mix_new, &
                              scf_env%p_delta, nspin, subspace_env%mixing_store)
         IF (dft_control%qs_control%gapw) THEN
            CALL get_qs_env(qs_env=qs_env, &
                            rho_atom_set=rho_atom)
            CALL mixing_init(subspace_env%mixing_method, rho, subspace_env%mixing_store, &
                             para_env, rho_atom=rho_atom)
         ELSE
            CALL mixing_init(subspace_env%mixing_method, rho, subspace_env%mixing_store, &
                             para_env)
         END IF
      END IF

      ene_old = 0.0_dp
      ene_diff = 0.0_dp
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/T19,A)") '<<<<<<<<<   SUBSPACE ROTATION    <<<<<<<<<<'
         WRITE (output_unit, "(T4,A,T13,A,T21,A,T38,A,T51,A,T65,A/,T4,A)") &
            "In-step", "Time", "Convergence", "Band ene.", "Total ene.", "Energy diff.", REPEAT("-", 74)
      END IF

      ! recalculate density matrix here

      ! update of density
      CALL qs_rho_update_rho(rho, qs_env=qs_env)

      DO iloop = 1, subspace_env%max_iter
         t1 = m_walltime()
         converged = .FALSE.
         ene_old = energy%total

         CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)
         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., &
                                  just_energy=.FALSE., print_active=.FALSE.)

         max_val = 0.0_dp
         sum_val = 0.0_dp
         sum_band = 0.0_dp
         DO ispin = 1, SIZE(matrix_ks)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                            nao=nao, &
                            nmo=nmo, &
                            eigenvalues=mo_eigenvalues, &
                            occupation_numbers=mo_occupations, &
                            mo_coeff=mo_coeff)

            !compute C'HC
            chc => subspace_env%chc_mat(ispin)%matrix
            evec => subspace_env%c_vec(ispin)%matrix
            c0 => subspace_env%c0(ispin)%matrix
            CALL cp_fm_to_fm(mo_coeff, c0)
            CALL cp_dbcsr_sm_fm_multiply(matrix_ks(ispin)%matrix, c0, work, nmo)
            CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, work, rzero, chc)
            !diagonalize C'HC
            CALL choose_eigv_solver(chc, evec, mo_eigenvalues)

            !rotate the mos by the eigenvectors of C'HC
            CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, c0, evec, rzero, mo_coeff)

            CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, &
                                   smear=scf_control%smear)

            ! does not yet handle k-points
            CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                          subspace_env%p_matrix_mix(ispin, 1)%matrix)

            DO i = 1, nmo
               sum_band = sum_band+mo_eigenvalues(i)*mo_occupations(i)
            END DO

            !check for self consistency
         END DO

         IF (subspace_env%mixing_method == direct_mixing_nr) THEN
            CALL scf_env_density_mixing(subspace_env%p_matrix_mix, &
                                        scf_env%mixing_store, rho_ao_kp, para_env, iter_delta, iloop)
         ELSE
            CALL self_consistency_check(rho_ao_kp, scf_env%p_delta, para_env, &
                                        subspace_env%p_matrix_mix, delta=iter_delta)
         END IF

         DO ispin = 1, nspin
            ! does not yet handle k-points
            CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, subspace_env%p_matrix_mix(ispin, 1)%matrix)
         END DO
         ! update of density
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         ! Mixing in reciprocal space
         IF (subspace_env%mixing_method >= gspace_mixing_nr) THEN
            CALL gspace_mixing(qs_env, scf_env%mixing_method, subspace_env%mixing_store, &
                               rho, para_env, scf_env%iter_count)
         END IF

         ene_diff = energy%total-ene_old
         converged = (ABS(ene_diff) < subspace_env%eps_ene .AND. &
                      iter_delta < subspace_env%eps_adapt*scf_env%iter_delta)
         t2 = m_walltime()
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T4,I5,T11,F8.3,T18,E14.4,T34,F12.5,T46,F16.8,T62,E14.4)") &
               iloop, t2-t1, iter_delta, sum_band, energy%total, ene_diff
            CALL m_flush(output_unit)
         END IF
         IF (converged) THEN
            IF (output_unit > 0) WRITE (output_unit, "(T10,A,I6,A,/)") &
               " Reached convergence in ", iloop, " iterations "
            EXIT
         END IF

      END DO ! iloop

      NULLIFY (subspace_env%p_matrix_mix)
      DO ispin = 1, nspin
         ! does not yet handle k-points
         CALL cp_dbcsr_copy(scf_env%p_mix_new(ispin, 1)%matrix, rho_ao(ispin)%matrix)
         CALL cp_dbcsr_copy(rho_ao(ispin)%matrix, subspace_env%p_matrix_store(ispin)%matrix)

         DEALLOCATE (eval_first(ispin)%array, occ_first(ispin)%array)
      END DO
      DEALLOCATE (eval_first, occ_first)

      CALL timestop(handle)

   END SUBROUTINE do_scf_diag_subspace

! **************************************************************************************************
!> \brief ...
!> \param subspace_env ...
!> \param qs_env ...
!> \param mos ...
! **************************************************************************************************
   SUBROUTINE diag_subspace_allocate(subspace_env, qs_env, mos)

      TYPE(subspace_env_type), POINTER                   :: subspace_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'diag_subspace_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, nmo, nspin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb

      CALL timeset(routineN, handle)

      NULLIFY (sab_orb, matrix_s)
      CALL get_qs_env(qs_env=qs_env, sab_orb=sab_orb, &
                      matrix_s=matrix_s)

      nspin = SIZE(mos)
!   *** allocate p_atrix_store ***
      IF (.NOT. ASSOCIATED(subspace_env%p_matrix_store)) THEN
         CALL cp_dbcsr_allocate_matrix_set(subspace_env%p_matrix_store, nspin)

         DO i = 1, nspin
            ALLOCATE (subspace_env%p_matrix_store(i)%matrix)
            CALL cp_dbcsr_init(subspace_env%p_matrix_store(i)%matrix)
            CALL cp_dbcsr_create(matrix=subspace_env%p_matrix_store(i)%matrix, &
                                 name="DENSITY_STORE", &
                                 dist=cp_dbcsr_distribution(matrix_s(1)%matrix), matrix_type=dbcsr_type_symmetric, &
                                 row_blk_size=cp_dbcsr_row_block_sizes(matrix_s(1)%matrix), &
                                 col_blk_size=cp_dbcsr_col_block_sizes(matrix_s(1)%matrix), &
                                 nze=0)
            CALL cp_dbcsr_alloc_block_from_nbl(subspace_env%p_matrix_store(i)%matrix, &
                                               sab_orb)
            CALL cp_dbcsr_set(subspace_env%p_matrix_store(i)%matrix, 0.0_dp)
         ENDDO

      END IF

      ALLOCATE (subspace_env%chc_mat(nspin))
      ALLOCATE (subspace_env%c_vec(nspin))
      ALLOCATE (subspace_env%c0(nspin))

      DO ispin = 1, nspin
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         CALL cp_fm_create(subspace_env%c0(ispin)%matrix, mo_coeff%matrix_struct)
         NULLIFY (fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmo, ncol_global=nmo, &
                                  para_env=mo_coeff%matrix_struct%para_env, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(subspace_env%chc_mat(ispin)%matrix, fm_struct_tmp, "chc")
         CALL cp_fm_create(subspace_env%c_vec(ispin)%matrix, fm_struct_tmp, "vec")
         CALL cp_fm_struct_release(fm_struct_tmp)
      END DO

      CALL timestop(handle)

   END SUBROUTINE diag_subspace_allocate

! **************************************************************************************************
!> \brief the inner loop of scf, specific to diagonalization without S matrix
!>       basically, in goes the ks matrix out goes a new p matrix
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE do_special_diag(scf_env, mos, matrix_ks, scf_control, &
                              scf_section, diis_step)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      INTEGER                                            :: ispin, nspin
      LOGICAL                                            :: do_level_shift, use_jacobi
      REAL(kind=dp)                                      :: diis_error

      nspin = SIZE(matrix_ks)

      DO ispin = 1, nspin
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, scf_env%scf_work1(ispin)%matrix)
      END DO
      IF (scf_env%iter_count > 1 .AND. .NOT. scf_env%skip_diis) THEN
         CALL qs_diis_b_step(scf_env%scf_diis_buffer, mos, scf_env%scf_work1, &
                             scf_env%scf_work2, scf_env%iter_delta, diis_error, diis_step, &
                             scf_control%eps_diis, scf_control%nmixing, &
                             scf_section=scf_section)
      ELSE
         diis_step = .FALSE.
      END IF

      IF ((scf_env%iter_count > 1) .AND. (scf_env%iter_delta < scf_control%diagonalization%eps_jacobi)) THEN
         use_jacobi = .TRUE.
      ELSE
         use_jacobi = .FALSE.
      END IF

      do_level_shift = ((scf_control%level_shift /= 0.0_dp) .AND. &
                        ((scf_control%density_guess == core_guess) .OR. (scf_env%iter_count > 1)))
      IF (diis_step) THEN
         scf_env%iter_param = diis_error
         IF (use_jacobi) THEN
            scf_env%iter_method = "DIIS/Jacobi"
         ELSE
            scf_env%iter_method = "DIIS/Diag."
         END IF
      ELSE
         IF (scf_env%mixing_method == 1) THEN
            scf_env%iter_param = scf_env%p_mix_alpha
            IF (use_jacobi) THEN
               scf_env%iter_method = "P_Mix/Jacobi"
            ELSE
               scf_env%iter_method = "P_Mix/Diag."
            END IF
         ELSEIF (scf_env%mixing_method > 1) THEN
            scf_env%iter_param = scf_env%mixing_store%alpha
            IF (use_jacobi) THEN
               scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Jacobi"
            ELSE
               scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Diag."
            END IF
         END IF
      END IF
      scf_env%iter_delta = 0.0_dp

      DO ispin = 1, nspin
         CALL eigensolver_simple(matrix_ks=scf_env%scf_work1(ispin)%matrix, &
                                 mo_set=mos(ispin)%mo_set, &
                                 work=scf_env%scf_work2, &
                                 do_level_shift=do_level_shift, &
                                 level_shift=scf_control%level_shift, &
                                 use_jacobi=use_jacobi, &
                                 jacobi_threshold=scf_control%diagonalization%jacobi_threshold)
      END DO

      CALL set_mo_occupation(mo_array=mos, &
                             smear=scf_control%smear)

      DO ispin = 1, nspin
         ! does not yet handle k-points
         CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                       scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

   END SUBROUTINE do_special_diag

! **************************************************************************************************
!> \brief the inner loop of scf, specific to iterative diagonalization using OT
!>        with S matrix; basically, in goes the ks matrix out goes a new p matrix
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
!> \par History
!>      10.2008 created [JGH]
! **************************************************************************************************
   SUBROUTINE do_ot_diag(scf_env, mos, matrix_ks, matrix_s, &
                         scf_control, scf_section, diis_step)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_ot_diag', routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, ispin, nmo, nspin
      REAL(kind=dp)                                      :: diis_error, eps_iter
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      NULLIFY (eigenvalues)

      nspin = SIZE(matrix_ks)

      DO ispin = 1, nspin
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, &
                               scf_env%scf_work1(ispin)%matrix)
      END DO

      IF ((scf_env%iter_count > 1) .AND. (.NOT. scf_env%skip_diis)) THEN
         CALL qs_diis_b_step(scf_env%scf_diis_buffer, mos, scf_env%scf_work1, &
                             scf_env%scf_work2, scf_env%iter_delta, diis_error, diis_step, &
                             scf_control%eps_diis, scf_control%nmixing, &
                             s_matrix=matrix_s, &
                             scf_section=scf_section)
      ELSE
         diis_step = .FALSE.
      END IF

      eps_iter = scf_control%diagonalization%eps_iter
      IF (diis_step) THEN
         scf_env%iter_param = diis_error
         scf_env%iter_method = "DIIS/OTdiag"
         DO ispin = 1, nspin
            CALL copy_fm_to_dbcsr(scf_env%scf_work1(ispin)%matrix, &
                                  matrix_ks(ispin)%matrix, keep_sparsity=.TRUE.)
         END DO
         eps_iter = MAX(eps_iter, scf_control%diagonalization%eps_adapt*diis_error)
      ELSE
         IF (scf_env%mixing_method == 1) THEN
            scf_env%iter_param = scf_env%p_mix_alpha
            scf_env%iter_method = "P_Mix/OTdiag."
         ELSEIF (scf_env%mixing_method > 1) THEN
            scf_env%iter_param = scf_env%mixing_store%alpha
            scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/OTdiag."
         END IF
      END IF

      scf_env%iter_delta = 0.0_dp

      DO ispin = 1, nspin
         CALL get_mo_set(mos(ispin)%mo_set, &
                         mo_coeff=mo_coeff, &
                         eigenvalues=eigenvalues, &
                         nmo=nmo, &
                         homo=homo)
         CALL ot_eigensolver(matrix_h=matrix_ks(ispin)%matrix, &
                             matrix_s=matrix_s(1)%matrix, &
                             matrix_c_fm=mo_coeff, &
                             preconditioner=scf_env%ot_preconditioner(1)%preconditioner, &
                             eps_gradient=eps_iter, &
                             iter_max=scf_control%diagonalization%max_iter, &
                             silent=.TRUE., &
                             ot_settings=scf_control%diagonalization%ot_settings)
         CALL calculate_subspace_eigenvalues(mo_coeff, matrix_ks(ispin)%matrix, &
                                             evals_arg=eigenvalues, &
                                             do_rotation=.TRUE.)
         !MK WRITE(*,*) routinen//' copy_dbcsr_to_fm'
         CALL copy_fm_to_dbcsr(mos(ispin)%mo_set%mo_coeff, &
                               mos(ispin)%mo_set%mo_coeff_b)
         !fm->dbcsr
      END DO

      CALL set_mo_occupation(mo_array=mos, &
                             smear=scf_control%smear)

      DO ispin = 1, nspin
         ! does not yet handle k-points
         CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                       scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

   END SUBROUTINE do_ot_diag

! **************************************************************************************************
!> \brief Solve a set restricted open Kohn-Sham (ROKS) equations based on the
!>         alpha and beta Kohn-Sham matrices from unrestricted Kohn-Sham.
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
!> \param orthogonal_basis ...
!> \par History
!>      04.2006 created [MK]
!>      Revised (01.05.06,MK)
!> \note
!>         this is only a high-spin ROKS.
! **************************************************************************************************
   SUBROUTINE do_roks_diag(scf_env, mos, matrix_ks, matrix_s, &
                           scf_control, scf_section, diis_step, &
                           orthogonal_basis)

      ! Literature: - C. C. J. Roothaan, Rev. Mod. Phys. 32, 179 (1960)
      !             - M. F. Guest and V. R. Saunders, Mol. Phys. 28(3), 819 (1974)
      !             - M. Filatov and S. Shaik, Chem. Phys. Lett. 288, 689 (1998)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step
      LOGICAL, INTENT(IN)                                :: orthogonal_basis

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_roks_diag', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homoa, homob, imo, nalpha, nao, &
                                                            nbeta, nmo
      REAL(KIND=dp)                                      :: diis_error, level_shift_loc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eiga, eigb, occa, occb
      TYPE(cp_fm_type), POINTER                          :: ksa, ksb, mo2ao, moa, mob, ortho, work

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (scf_env%cholesky_method == cholesky_inverse) THEN
         ortho => scf_env%ortho_m1
      ELSE
         ortho => scf_env%ortho
      END IF
      work => scf_env%scf_work2

      ksa => scf_env%scf_work1(1)%matrix
      ksb => scf_env%scf_work1(2)%matrix

      CALL copy_dbcsr_to_fm(matrix_ks(1)%matrix, ksa)
      CALL copy_dbcsr_to_fm(matrix_ks(2)%matrix, ksb)

      ! Get MO information

      CALL get_mo_set(mo_set=mos(1)%mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      nelectron=nalpha, &
                      homo=homoa, &
                      eigenvalues=eiga, &
                      occupation_numbers=occa, &
                      mo_coeff=moa)

      CALL get_mo_set(mo_set=mos(2)%mo_set, &
                      nelectron=nbeta, &
                      homo=homob, &
                      eigenvalues=eigb, &
                      occupation_numbers=occb, &
                      mo_coeff=mob)

      ! Define the amount of level-shifting

      IF ((scf_control%level_shift /= 0.0_dp) .AND. &
          ((scf_control%density_guess == core_guess) .OR. &
           (scf_control%density_guess == restart_guess) .OR. &
           (scf_env%iter_count > 1))) THEN
         level_shift_loc = scf_control%level_shift
      ELSE
         level_shift_loc = 0.0_dp
      END IF

      IF ((scf_env%iter_count > 1) .OR. &
          (scf_control%density_guess == core_guess) .OR. &
          (scf_control%density_guess == restart_guess)) THEN

         ! Transform the spin unrestricted alpha and beta Kohn-Sham matrices
         ! from AO basis to MO basis: K(MO) = C(T)*K(AO)*C

         CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, ksa, moa, 0.0_dp, work)
         CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, moa, work, 0.0_dp, ksa)

         CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, ksb, moa, 0.0_dp, work)
         CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, moa, work, 0.0_dp, ksb)

         ! Combine the spin unrestricted alpha and beta Kohn-Sham matrices
         ! in the MO basis

         IF (scf_control%roks_scheme == general_roks) THEN
            CALL combine_ks_matrices(ksa, ksb, occa, occb, scf_control%roks_f, &
                                     nalpha, nbeta)
         ELSE IF (scf_control%roks_scheme == high_spin_roks) THEN
            CALL combine_ks_matrices(ksa, ksb, occa, occb, scf_control%roks_parameter)
         ELSE
            CALL cp__b("qs_scf_diagonalization.F",1140,"Unknown ROKS scheme requested")
         END IF

         ! Back-transform the restricted open Kohn-Sham matrix from MO basis
         ! to AO basis

         IF (orthogonal_basis) THEN
            ! Q = C
            mo2ao => moa
         ELSE
            ! Q = S*C
            mo2ao => mob
!MK     CALL copy_sm_to_fm(matrix_s(1)%matrix,work)
!MK     CALL cp_fm_symm("L","U",nao,nao,1.0_dp,work,moa,0.0_dp,mo2ao)
            CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, moa, mo2ao, nao)
         END IF

         ! K(AO) = Q*K(MO)*Q(T)

         CALL cp_gemm("N", "T", nao, nao, nao, 1.0_dp, ksa, mo2ao, 0.0_dp, work)
         CALL cp_gemm("N", "N", nao, nao, nao, 1.0_dp, mo2ao, work, 0.0_dp, ksa)

      ELSE

         ! No transformation matrix available, yet. The closed shell part,
         ! i.e. the beta Kohn-Sham matrix in AO basis, is taken.
         ! There might be better choices, anyhow.

         CALL cp_fm_to_fm(ksb, ksa)

      END IF

      ! Update DIIS buffer and possibly perform DIIS extrapolation step

      IF (scf_env%iter_count > 1) THEN
         IF (orthogonal_basis) THEN
            CALL qs_diis_b_step(diis_buffer=scf_env%scf_diis_buffer, &
                                mo_array=mos, &
                                kc=scf_env%scf_work1, &
                                sc=work, &
                                delta=scf_env%iter_delta, &
                                error_max=diis_error, &
                                diis_step=diis_step, &
                                eps_diis=scf_control%eps_diis, &
                                scf_section=scf_section, &
                                roks=.TRUE.)
         ELSE
            CALL qs_diis_b_step(diis_buffer=scf_env%scf_diis_buffer, &
                                mo_array=mos, &
                                kc=scf_env%scf_work1, &
                                sc=work, &
                                delta=scf_env%iter_delta, &
                                error_max=diis_error, &
                                diis_step=diis_step, &
                                eps_diis=scf_control%eps_diis, &
                                scf_section=scf_section, &
                                s_matrix=matrix_s, &
                                roks=.TRUE.)
         END IF
      END IF

      IF (diis_step) THEN
         scf_env%iter_param = diis_error
         scf_env%iter_method = "DIIS/Diag."
      ELSE
         IF (scf_env%mixing_method == 1) THEN
            scf_env%iter_param = scf_env%p_mix_alpha
            scf_env%iter_method = "P_Mix/Diag."
         ELSEIF (scf_env%mixing_method > 1) THEN
            scf_env%iter_param = scf_env%mixing_store%alpha
            scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Diag."
         END IF
      END IF

      scf_env%iter_delta = 0.0_dp

      IF (level_shift_loc /= 0.0_dp) THEN

         ! Transform the current Kohn-Sham matrix from AO to MO basis
         ! for level-shifting using the current MO set

         CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, ksa, moa, 0.0_dp, work)
         CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, moa, work, 0.0_dp, ksa)

         ! Apply level-shifting using 50:50 split of the shift (could be relaxed)

         DO imo = homob+1, homoa
            CALL cp_fm_add_to_element(ksa, imo, imo, 0.5_dp*level_shift_loc)
         END DO
         DO imo = homoa+1, nmo
            CALL cp_fm_add_to_element(ksa, imo, imo, level_shift_loc)
         END DO

      ELSE IF (.NOT. orthogonal_basis) THEN

         ! Transform the current Kohn-Sham matrix to an orthogonal basis
         SELECT CASE (scf_env%cholesky_method)
         CASE (cholesky_reduce)
            CALL cp_fm_cholesky_reduce(ksa, ortho)
         CASE (cholesky_restore)
            CALL cp_fm_upper_to_full(ksa, work)
            CALL cp_fm_cholesky_restore(ksa, nao, ortho, work, &
                                        "SOLVE", pos="RIGHT")
            CALL cp_fm_cholesky_restore(work, nao, ortho, ksa, &
                                        "SOLVE", pos="LEFT", transa="T")
         CASE (cholesky_inverse)
            CALL cp_fm_upper_to_full(ksa, work)
            CALL cp_fm_cholesky_restore(ksa, nao, ortho, work, &
                                        "MULTIPLY", pos="RIGHT")
            CALL cp_fm_cholesky_restore(work, nao, ortho, ksa, &
                                        "MULTIPLY", pos="LEFT", transa="T")
         CASE (cholesky_off)
            CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, ksa, ortho, 0.0_dp, work)
            CALL cp_gemm("N", "N", nao, nao, nao, 1.0_dp, ortho, work, 0.0_dp, ksa)
         END SELECT

      END IF

      ! Diagonalization of the ROKS operator matrix

      CALL choose_eigv_solver(ksa, work, eiga)

      ! Back-transformation of the orthonormal eigenvectors if needed

      IF (level_shift_loc /= 0.0_dp) THEN
         ! Use old MO set for back-transformation if level-shifting was applied
         CALL cp_fm_to_fm(moa, ortho)
         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, ortho, work, 0.0_dp, moa)
      ELSE
         IF (orthogonal_basis) THEN
            CALL cp_fm_to_fm(work, moa)
         ELSE
            SELECT CASE (scf_env%cholesky_method)
            CASE (cholesky_reduce, cholesky_restore)
               CALL cp_fm_cholesky_restore(work, nmo, ortho, moa, "SOLVE")
            CASE (cholesky_inverse)
               CALL cp_fm_cholesky_restore(work, nmo, ortho, moa, "MULTIPLY")
            CASE (cholesky_off)
               CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, ortho, work, 0.0_dp, moa)
            END SELECT
         END IF
      END IF

      ! Correct MO eigenvalues, if level-shifting was applied

      IF (level_shift_loc /= 0.0_dp) THEN
         DO imo = homob+1, homoa
            eiga(imo) = eiga(imo)-0.5_dp*level_shift_loc
         END DO
         DO imo = homoa+1, nmo
            eiga(imo) = eiga(imo)-level_shift_loc
         END DO
      END IF

      ! Update also the beta MO set

      eigb(:) = eiga(:)
      CALL cp_fm_to_fm(moa, mob)

      ! Calculate the new alpha and beta density matrix

      ! does not yet handle k-points
      CALL calculate_density_matrix(mos(1)%mo_set, scf_env%p_mix_new(1, 1)%matrix)
      CALL calculate_density_matrix(mos(2)%mo_set, scf_env%p_mix_new(2, 1)%matrix)

      CALL timestop(handle)

   END SUBROUTINE do_roks_diag

! **************************************************************************************************
!> \brief iterative diagonalization using the block Krylov-space approach
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param scf_control ...
!> \param scf_section ...
!> \param check_moconv_only ...
!> \param
!> \par History
!>      05.2009 created [MI]
! **************************************************************************************************

   SUBROUTINE do_block_krylov_diag(scf_env, mos, matrix_ks, &
                                   scf_control, scf_section, check_moconv_only)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_moconv_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_block_krylov_diag', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, homo, ispin, iter, nao, nmo, &
                                                            output_unit
      LOGICAL                                            :: converged, my_check_moconv_only
      REAL(dp)                                           :: eps_iter, t1, t2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: c0, c1, chc, evec, ks, mo_coeff, ortho, &
                                                            work
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%LANCZOS", &
                                         extension=".scfLog")

      my_check_moconv_only = .FALSE.
      IF (PRESENT(check_moconv_only)) my_check_moconv_only = check_moconv_only

      NULLIFY (mo_coeff, ortho, work, ks)
      NULLIFY (mo_eigenvalues)
      NULLIFY (c0, c1)

      IF (scf_env%cholesky_method == cholesky_inverse) THEN
         ortho => scf_env%ortho_m1
      ELSE
         ortho => scf_env%ortho
      END IF
      work => scf_env%scf_work2

      DO ispin = 1, SIZE(matrix_ks)
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, &
                               scf_env%scf_work1(ispin)%matrix)
      END DO

      IF (scf_env%mixing_method == 1) THEN
         scf_env%iter_param = scf_env%p_mix_alpha
         scf_env%iter_method = "P_Mix/Lanczos"
      ELSE
!        scf_env%iter_param = scf_env%mixing_store%alpha
         scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Lanc."
      END IF

      DO ispin = 1, SIZE(matrix_ks)

         ks => scf_env%scf_work1(ispin)%matrix
         CALL cp_fm_upper_to_full(ks, work)

         CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                         nao=nao, &
                         nmo=nmo, &
                         homo=homo, &
                         eigenvalues=mo_eigenvalues, &
                         mo_coeff=mo_coeff)

         NULLIFY (c0, c1)
         c0 => scf_env%krylov_space%mo_conv(ispin)%matrix
         c1 => scf_env%krylov_space%mo_refine(ispin)%matrix
         SELECT CASE (scf_env%cholesky_method)
         CASE (cholesky_reduce)
            CALL cp_fm_cholesky_reduce(ks, ortho)
            CALL cp_fm_upper_to_full(ks, work)
            CALL cp_fm_cholesky_restore(mo_coeff, nmo, ortho, c0, "MULTIPLY")
         CASE (cholesky_restore)
            CALL cp_fm_cholesky_restore(ks, nao, ortho, work, &
                                        "SOLVE", pos="RIGHT")
            CALL cp_fm_cholesky_restore(work, nao, ortho, ks, &
                                        "SOLVE", pos="LEFT", transa="T")
            CALL cp_fm_cholesky_restore(mo_coeff, nmo, ortho, c0, "MULTIPLY")
         CASE (cholesky_inverse)
            CALL cp_fm_cholesky_restore(ks, nao, ortho, work, &
                                        "MULTIPLY", pos="RIGHT")
            CALL cp_fm_cholesky_restore(work, nao, ortho, ks, &
                                        "MULTIPLY", pos="LEFT", transa="T")
            CALL cp_fm_cholesky_restore(mo_coeff, nmo, ortho, c0, "SOLVE")
         END SELECT

         scf_env%krylov_space%nmo_nc = nmo
         scf_env%krylov_space%nmo_conv = 0

         t1 = m_walltime()
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(/T15,A)") '<<<<<<<<<   LANCZOS REFINEMENT    <<<<<<<<<<'
            WRITE (output_unit, "(T8,A,T15,A,T23,A,T36,A,T49,A,T60,A,/,T8,A)") &
               " Spin ", " Cycle ", &
               " conv. MOS ", " B2MAX ", " B2MIN ", " Time", REPEAT("-", 60)
         END IF
         eps_iter = MAX(scf_env%krylov_space%eps_conv, scf_env%krylov_space%eps_adapt*scf_env%iter_delta)
         iter = 0
         converged = .FALSE.
         !Check convergence of MOS
         IF (my_check_moconv_only) THEN

            CALL lanczos_refinement(scf_env%krylov_space, ks, c0, c1, mo_eigenvalues, &
                                    nao, eps_iter, ispin, check_moconv_only=my_check_moconv_only)
            t2 = m_walltime()
            IF (output_unit > 0) &
               WRITE (output_unit, '(T8,I3,T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
               ispin, iter, scf_env%krylov_space%nmo_conv, &
               scf_env%krylov_space%max_res_norm, scf_env%krylov_space%min_res_norm, t2-t1

            CYCLE
         ELSE
            !Block Lanczos refinement
            DO iter = 1, scf_env%krylov_space%max_iter
               CALL lanczos_refinement_2v(scf_env%krylov_space, ks, c0, c1, mo_eigenvalues, &
                                          nao, eps_iter, ispin)
               t2 = m_walltime()
               IF (output_unit > 0) THEN
                  WRITE (output_unit, '(T8,I3,T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
                     ispin, iter, scf_env%krylov_space%nmo_conv, &
                     scf_env%krylov_space%max_res_norm, scf_env%krylov_space%min_res_norm, t2-t1
               END IF
               t1 = m_walltime()
               IF (scf_env%krylov_space%max_res_norm < eps_iter) THEN
                  converged = .TRUE.
                  IF (output_unit > 0) WRITE (output_unit, *) &
                     " Reached convergence in ", iter, " iterations "
                  EXIT
               END IF
            END DO

            IF (.NOT. converged .AND. output_unit > 0) THEN
               WRITE (output_unit, "(T4, A)") " WARNING Lanczos refinement could "// &
                  "not converge all the mos:"
               WRITE (output_unit, "(T40,A,T70,I10)") " number of not converged mos ", &
                  scf_env%krylov_space%nmo_nc
               WRITE (output_unit, "(T40,A,T70,E10.2)") " max norm of the residual ", &
                  scf_env%krylov_space%max_res_norm

            END IF

            ! For the moment skip the re-orthogonalization
            IF (.FALSE.) THEN
               !Re-orthogonalization
               NULLIFY (chc, evec)
               chc => scf_env%krylov_space%chc_mat(ispin)%matrix
               evec => scf_env%krylov_space%c_vec(ispin)%matrix
               CALL cp_gemm('N', 'N', nao, nmo, nao, rone, ks, c0, rzero, work)
               CALL cp_gemm('T', 'N', nmo, nmo, nao, rone, c0, work, rzero, chc)
               !Diagonalize  (C^t)HC
               CALL choose_eigv_solver(chc, evec, mo_eigenvalues)
               !Rotate the C vectors
               CALL cp_gemm('N', 'N', nao, nmo, nmo, rone, c0, evec, rzero, c1)
               c0 => scf_env%krylov_space%mo_refine(ispin)%matrix
            END IF

            IF (scf_env%cholesky_method == cholesky_inverse) THEN
               CALL cp_fm_cholesky_restore(c0, nmo, ortho, mo_coeff, "MULTIPLY")
            ELSE
               CALL cp_fm_cholesky_restore(c0, nmo, ortho, mo_coeff, "SOLVE")
            END IF

            CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, &
                                   smear=scf_control%smear)

            ! does not yet handle k-points
            CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                          scf_env%p_mix_new(ispin, 1)%matrix)
         END IF
      END DO ! ispin

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T15,A/)") '<<<<<<<<< END LANCZOS REFINEMENT  <<<<<<<<<<'
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%LANCZOS")

      CALL timestop(handle)

   END SUBROUTINE do_block_krylov_diag

! **************************************************************************************************
!> \brief iterative diagonalization using the block davidson space approach
!> \param qs_env ...
!> \param scf_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param scf_control ...
!> \param scf_section ...
!> \param check_moconv_only ...
!> \param
!> \par History
!>      05.2011 created [MI]
! **************************************************************************************************

   SUBROUTINE do_block_davidson_diag(qs_env, scf_env, mos, matrix_ks, matrix_s, &
                                     scf_control, scf_section, check_moconv_only)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_moconv_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_block_davidson_diag', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins, output_unit
      LOGICAL                                            :: do_prec, my_check_moconv_only
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%DAVIDSON", &
                                         extension=".scfLog")

      IF (output_unit > 0) &
         WRITE (output_unit, "(/T15,A)") '<<<<<<<<<  DAVIDSON ITERATIONS   <<<<<<<<<<'

      IF (scf_env%mixing_method == 1) THEN
         scf_env%iter_param = scf_env%p_mix_alpha
         scf_env%iter_method = "P_Mix/Dav."
      ELSE
         scf_env%iter_param = scf_env%mixing_store%alpha
         scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Dav."
      END IF

      my_check_moconv_only = .FALSE.
      IF (PRESENT(check_moconv_only)) my_check_moconv_only = check_moconv_only
      do_prec = .FALSE.
      IF (scf_env%block_davidson_env(1)%prec_type /= 0 .AND. &
          scf_env%iter_count >= scf_env%block_davidson_env(1)%first_prec) THEN
         do_prec = .TRUE.
      END IF

      nspins = SIZE(matrix_ks)

      IF (do_prec .AND. (scf_env%iter_count == scf_env%block_davidson_env(1)%first_prec .OR. &
                         MODULO(scf_env%iter_count, scf_env%block_davidson_env(1)%niter_new_prec) == 0)) THEN
         CALL restart_preconditioner(qs_env, scf_env%ot_preconditioner, &
                                     prec_type=scf_env%block_davidson_env(1)%prec_type, nspins=nspins)
         CALL prepare_preconditioner(qs_env, mos, matrix_ks, matrix_s, scf_env%ot_preconditioner, &
                                     scf_env%block_davidson_env(1)%prec_type, &
                                     scf_env%block_davidson_env(1)%solver_type, &
                                     scf_env%block_davidson_env(1)%energy_gap, nspins, &
                                     convert_to_dbcsr=scf_env%block_davidson_env(1)%use_sparse_mos)
      END IF

      DO ispin = 1, nspins
         IF (scf_env%block_davidson_env(ispin)%use_sparse_mos) THEN
            IF (.NOT. do_prec) THEN
               CALL generate_extended_space_sparse(scf_env%block_davidson_env(ispin), mos(ispin)%mo_set, &
                                                   matrix_ks(ispin)%matrix, matrix_s(1)%matrix, output_unit)
            ELSE
               CALL generate_extended_space_sparse(scf_env%block_davidson_env(ispin), mos(ispin)%mo_set, &
                                                   matrix_ks(ispin)%matrix, matrix_s(1)%matrix, output_unit, &
                                                   scf_env%ot_preconditioner(ispin)%preconditioner)
            END IF

         ELSE
            IF (.NOT. do_prec) THEN
               CALL generate_extended_space(scf_env%block_davidson_env(ispin), mos(ispin)%mo_set, &
                                            matrix_ks(ispin)%matrix, matrix_s(1)%matrix, output_unit)
            ELSE
               CALL generate_extended_space(scf_env%block_davidson_env(ispin), mos(ispin)%mo_set, &
                                            matrix_ks(ispin)%matrix, matrix_s(1)%matrix, output_unit, &
                                            scf_env%ot_preconditioner(ispin)%preconditioner)
            END IF
         END IF
      END DO !ispin

      CALL set_mo_occupation(mo_array=mos, &
                             smear=scf_control%smear)

      DO ispin = 1, nspins
         ! does not yet handle k-points
         CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                       scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T15,A/)") '<<<<<<<<< END DAVIDSON ITERATION  <<<<<<<<<<'
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%DAVIDSON")

      CALL timestop(handle)

   END SUBROUTINE do_block_davidson_diag

END MODULE qs_scf_diagonalization
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods for deltaSCF calculations
! **************************************************************************************************
MODULE qs_mom_methods
   USE bibliography,                    ONLY: Gilbert2008,&
                                              cite_reference
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_maxabsval,&
                                              cp_fm_release,&
                                              cp_fm_type,&
                                              cp_fm_vectorsnorm,&
                                              cp_fm_vectorssum
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE input_constants,                 ONLY: momproj_norm,&
                                              momproj_sum
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_types,                     ONLY: deallocate_mo_set,&
                                              duplicate_mo_set,&
                                              get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type,&
                                              set_mo_set
   USE qs_scf_diagonalization,          ONLY: general_eigenproblem
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE scf_control_types,               ONLY: scf_control_type
   USE string_utilities,                ONLY: integer_to_string
   USE util,                            ONLY: sort,&
                                              sort_unique
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mom_methods'

   PUBLIC  :: do_mom_guess, do_mom_diag
   PRIVATE :: mom_is_unique_orbital_indices, mom_reoccupy_orbitals

CONTAINS

! **************************************************************************************************
!> \brief check that every molecular orbital index appears only once in each
!>        (de-)occupation list supplied by user. Check that all the indices
!>        are positive integers and abort if it is not the case.
!> \param  iarr      list of molecular orbital indices to be checked
!> \retval is_unique .true. if all the elements are unique or the list contains
!>         exactly one 0 element (meaning no excitation)
!> \par History
!>      01.2016 created [Sergey Chulkov]
! **************************************************************************************************
   FUNCTION mom_is_unique_orbital_indices(iarr) RESULT(is_unique)
      INTEGER, DIMENSION(:), POINTER                     :: iarr
      LOGICAL                                            :: is_unique

      CHARACTER(len=*), PARAMETER :: routineN = 'mom_is_unique_orbital_indices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, norbs
      INTEGER, DIMENSION(:), POINTER                     :: tmp_iarr

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(iarr)))CALL cp__a("qs_mom_methods.F",77)
      norbs = SIZE(iarr)

      IF (norbs > 0) THEN
         ALLOCATE (tmp_iarr(norbs))

         tmp_iarr(:) = iarr(:)
         CALL sort_unique(tmp_iarr, is_unique)

         ! Ensure that all orbital indices are positive integers.
         ! A special value '0' means 'disabled keyword',
         ! it must appear once to be interpreted in such a way
         IF (tmp_iarr(1) < 0 .OR. (tmp_iarr(1) == 0 .AND. norbs > 1)) &
            CALL cp__b("qs_mom_methods.F",90,"MOM: all molecular orbital indices must be positive integer numbers")

         DEALLOCATE (tmp_iarr)
      ENDIF

      is_unique = .TRUE.

      CALL timestop(handle)

   END FUNCTION mom_is_unique_orbital_indices

! **************************************************************************************************
!> \brief swap occupation numbers between molecular orbitals
!>        from occupation and de-occupation lists
!> \param mo_set        set of molecular orbitals
!> \param deocc_orb_set list of de-occupied orbital indices
!> \param occ_orb_set   list of newly occupied orbital indices
!> \param spin          spin component of the molecular orbitals;
!>                      to be used for diagnostic messages
!> \par History
!>      01.2016 created [Sergey Chulkov]
! **************************************************************************************************
   SUBROUTINE mom_reoccupy_orbitals(mo_set, deocc_orb_set, occ_orb_set, spin)
      TYPE(mo_set_type), POINTER                         :: mo_set
      INTEGER, DIMENSION(:), POINTER                     :: deocc_orb_set, occ_orb_set
      CHARACTER(len=*), INTENT(in)                       :: spin

      CHARACTER(len=*), PARAMETER :: routineN = 'mom_reoccupy_orbitals', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=10)                                  :: str_iorb, str_norbs
      CHARACTER(len=3)                                   :: str_prefix
      INTEGER                                            :: handle, homo, iorb, lfomo, nao, nmo, &
                                                            norbs
      REAL(kind=dp)                                      :: maxocc
      REAL(kind=dp), DIMENSION(:), POINTER               :: occ_nums

      CALL timeset(routineN, handle)

      ! MOM electron excitation should preserve both the number of electrons and
      ! multiplicity of the electronic system thus ensuring the following constraint :
      ! norbs = SIZE(deocc_orb_set) == SIZE(occ_orb_set)
      norbs = SIZE(deocc_orb_set)

      ! the following assertion should never raise an exception
      IF(.NOT.(SIZE(deocc_orb_set) == SIZE(occ_orb_set)))CALL cp__a("qs_mom_methods.F",135)

      ! MOM does not follow aufbau principle producing non-uniformly occupied orbitals
      CALL set_mo_set(mo_set=mo_set, uniform_occupation=.FALSE.)

      IF (deocc_orb_set(1) /= 0 .AND. occ_orb_set(1) /= 0) THEN
         CALL get_mo_set(mo_set=mo_set, maxocc=maxocc, &
                         nao=nao, nmo=nmo, occupation_numbers=occ_nums)

         IF (deocc_orb_set(norbs) > nao .OR. occ_orb_set(norbs) > nao) THEN
            ! STOP: one of the molecular orbital index exceeds the number of atomic basis functions available
            CALL integer_to_string(nao, str_norbs)

            IF (deocc_orb_set(norbs) >= occ_orb_set(norbs)) THEN
               iorb = deocc_orb_set(norbs)
               str_prefix = 'de-'
            ELSE
               iorb = occ_orb_set(norbs)
               str_prefix = ''
            ENDIF
            CALL integer_to_string(iorb, str_iorb)

            CALL cp_abort(cp__l("qs_mom_methods.F",157), "Unable to "//TRIM(str_prefix)//"occupy "// &
                          TRIM(spin)//" orbital No. "//TRIM(str_iorb)// &
                          " since its index exceeds the number of atomic orbital functions available ("// &
                          TRIM(str_norbs)//"). Please consider using a larger basis set.")
         ENDIF

         IF (deocc_orb_set(norbs) > nmo .OR. occ_orb_set(norbs) > nmo) THEN
            ! STOP: one of the molecular orbital index exceeds the number of constructed molecular orbitals
            IF (deocc_orb_set(norbs) >= occ_orb_set(norbs)) THEN
               iorb = deocc_orb_set(norbs)
            ELSE
               iorb = occ_orb_set(norbs)
            ENDIF

            IF (iorb-nmo > 1) THEN
               CALL integer_to_string(iorb-nmo, str_iorb)
               str_prefix = 's'
            ELSE
               str_iorb = 'an'
               str_prefix = ''
            ENDIF

            CALL integer_to_string(nmo, str_norbs)

            CALL cp_abort(cp__l("qs_mom_methods.F",181), "The number of molecular orbitals ("//TRIM(str_norbs)// &
                          ") is not enough to perform MOM calculation. Please add "// &
                          TRIM(str_iorb)//" extra orbital"//TRIM(str_prefix)// &
                          " using the ADDED_MOS keyword in the SCF section of your input file.")
         ENDIF

         DO iorb = 1, norbs
            ! swap occupation numbers between two adjoint molecular orbitals
            IF (occ_nums(deocc_orb_set(iorb)) <= 0.0_dp) THEN
               CALL integer_to_string(deocc_orb_set(iorb), str_iorb)

               CALL cp_abort(cp__l("qs_mom_methods.F",192), "The "//TRIM(spin)//" orbital No. "// &
                             TRIM(str_iorb)//" is not occupied thus it cannot be deoccupied.")
            ENDIF

            IF (occ_nums(occ_orb_set(iorb)) > 0.0_dp) THEN
               CALL integer_to_string(occ_orb_set(iorb), str_iorb)

               CALL cp_abort(cp__l("qs_mom_methods.F",199), "The "//TRIM(spin)//" orbital No. "// &
                             TRIM(str_iorb)//" is already occupied thus it cannot be reoccupied.")
            ENDIF

            occ_nums(occ_orb_set(iorb)) = occ_nums(deocc_orb_set(iorb))
            occ_nums(deocc_orb_set(iorb)) = 0.0_dp
         ENDDO

         ! locate the lowest non-maxocc occupied orbital
         DO lfomo = 1, nmo
            IF (occ_nums(lfomo) /= maxocc) EXIT
         ENDDO

         ! locate the highest occupied orbital
         DO homo = nmo, 1, -1
            IF (occ_nums(homo) > 0.0_dp) EXIT
         ENDDO

         CALL set_mo_set(mo_set=mo_set, homo=homo, lfomo=lfomo)

      ELSE IF (deocc_orb_set(1) /= 0 .OR. occ_orb_set(1) /= 0) THEN
         CALL cp_abort(cp__l("qs_mom_methods.F",220), &
                       "Incorrect multiplicity of the MOM reference electronic state")
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE mom_reoccupy_orbitals

! **************************************************************************************************
!> \brief initial guess for the maximum overlap method
!> \param nspins      number of spin components
!> \param mos         array of molecular orbitals
!> \param scf_control SCF control variables
!> \param p_rmpv      density matrix to be computed
!> \par History
!>      01.2016 created [Sergey Chulkov]
! **************************************************************************************************
   SUBROUTINE do_mom_guess(nspins, mos, scf_control, p_rmpv)
      INTEGER, INTENT(in)                                :: nspins
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_rmpv

      CHARACTER(len=*), PARAMETER :: routineN = 'do_mom_guess', routineP = moduleN//':'//routineN

      CHARACTER(len=10)                                  :: str_iter
      INTEGER                                            :: handle, ispin, scf_iter
      LOGICAL                                            :: is_mo
      REAL(kind=dp)                                      :: maxa
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      ! we are about to initialise the maximum overlap method,
      ! so cite the relevant reference first
      CALL cite_reference(Gilbert2008)

      ! ensure we do not have duplicated orbital indices
      IF (.NOT. &
          (mom_is_unique_orbital_indices(scf_control%diagonalization%mom_deoccA) .AND. &
           mom_is_unique_orbital_indices(scf_control%diagonalization%mom_deoccB) .AND. &
           mom_is_unique_orbital_indices(scf_control%diagonalization%mom_occA) .AND. &
           mom_is_unique_orbital_indices(scf_control%diagonalization%mom_occB))) &
         CALL cp_abort(cp__l("qs_mom_methods.F",263), &
                       "Duplicate orbital indices were found in the MOM section")

      ! ignore beta orbitals for spin-unpolarized calculations
      IF (nspins == 1 .AND. (scf_control%diagonalization%mom_deoccB(1) /= 0 &
                             .OR. scf_control%diagonalization%mom_occB(1) /= 0)) THEN

         CALL cp_warn(cp__l("qs_mom_methods.F",270), "Maximum overlap method will"// &
                      " ignore beta orbitals since neither UKS nor ROKS calculation is performed")
      ENDIF

      ! compute the change in multiplicity and number of electrons
      IF (SIZE(scf_control%diagonalization%mom_deoccA) /= &
          SIZE(scf_control%diagonalization%mom_occA) .OR. &
          (nspins > 1 .AND. &
           SIZE(scf_control%diagonalization%mom_deoccB) /= &
           SIZE(scf_control%diagonalization%mom_occB))) THEN

         CALL cp_abort(cp__l("qs_mom_methods.F",281), "Incorrect multiplicity of the MOM reference"// &
                       " electronic state or inconsistent number of electrons")
      ENDIF

      is_mo = .FALSE.
      ! by default activate MOM at the second SCF iteration as the
      ! 'old' molecular orbitals are unavailable from the very beginning
      scf_iter = 2
      ! check if the molecular orbitals are actually there
      ! by finding at least one MO coefficient > 0
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_maxabsval(mo_coeff, maxa)
         ! is_mo |= maxa > 0.0_dp
         IF (maxa > 0.0_dp) THEN
            is_mo = .TRUE.
            ! we already have the molecular orbitals (e.g. from a restart file);
            ! activate MOM immediately if the input keyword START_ITER is not given
            scf_iter = 1
            EXIT
         ENDIF
      ENDDO

      ! proceed alpha orbitals
      IF (nspins >= 1) &
         CALL mom_reoccupy_orbitals(mos(1)%mo_set, &
                                    scf_control%diagonalization%mom_deoccA, &
                                    scf_control%diagonalization%mom_occA, 'alpha')

      ! proceed beta orbitals (if any)
      IF (nspins >= 2) &
         CALL mom_reoccupy_orbitals(mos(2)%mo_set, &
                                    scf_control%diagonalization%mom_deoccB, &
                                    scf_control%diagonalization%mom_occB, 'beta')

      ! recompute the density matrix if the molecular orbitals are here;
      ! otherwise do nothing to prevent zeroing out the density matrix
      ! obtained from atomic guess
      IF (is_mo) THEN
         DO ispin = 1, nspins
            CALL calculate_density_matrix(mos(ispin)%mo_set, p_rmpv(ispin)%matrix)
         ENDDO
      ENDIF

      ! adjust the start SCF iteration number if needed
      IF (scf_control%diagonalization%mom_start < scf_iter) THEN
         IF (scf_control%diagonalization%mom_start > 0) THEN
            ! inappropriate iteration number has been provided through the input file;
            ! fix it and issue a warning message
            CALL integer_to_string(scf_iter, str_iter)
            CALL cp_warn(cp__l("qs_mom_methods.F",331), &
                         "The maximum overlap method will be activated at the SCF iteration No. "// &
                         TRIM(str_iter)//" due to the SCF guess method used.")
         ENDIF
         scf_control%diagonalization%mom_start = scf_iter
      ELSE IF (scf_control%diagonalization%mom_start > scf_iter .AND. &
               (scf_control%diagonalization%mom_occA(1) > 0 .OR. scf_control%diagonalization%mom_occB(1) > 0)) THEN
         ! the keyword START_ITER has been provided for an excited state calculation, ignore it
         CALL integer_to_string(scf_iter, str_iter)
         CALL cp_warn(cp__l("qs_mom_methods.F",340), &
                      "The maximum overlap method will be activated at the SCF iteration No. "// &
                      TRIM(str_iter)//" because an excited state calculation has been requested")
         scf_control%diagonalization%mom_start = scf_iter
      ENDIF

      ! MOM is now initialised properly
      scf_control%diagonalization%mom_didguess = .TRUE.

      CALL timestop(handle)

   END SUBROUTINE do_mom_guess

! **************************************************************************************************
!> \brief do an SCF iteration, then compute occupation numbers of the new
!>  molecular orbitals according to their overlap with the previous ones
!> \param scf_env     SCF environment information
!> \param mos         array of molecular orbitals
!> \param matrix_ks   sparse Kohn-Sham matrix
!> \param matrix_s    sparse overlap matrix
!> \param scf_control SCF control variables
!> \param scf_section SCF input section
!> \param diis_step   have we done a DIIS step
!> \par History
!>      07.2014 created [Matt Watkins]
!>      01.2016 release version [Sergey Chulkov]
! **************************************************************************************************
   SUBROUTINE do_mom_diag(scf_env, mos, matrix_ks, &
                          matrix_s, scf_control, scf_section, &
                          diis_step)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(len=*), PARAMETER :: routineN = 'do_mom_diag', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, iproj, ispin, lfomo, nao, &
                                                            nmo, nspins
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: inds
      REAL(kind=dp)                                      :: maxocc
      REAL(kind=dp), DIMENSION(:), POINTER               :: occ_nums, proj, tmp_occ_nums
      TYPE(cp_fm_struct_type), POINTER                   :: mo_mo_fmstruct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_orig, overlap, svec
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_orig

      CALL timeset(routineN, handle)

      IF (.NOT. scf_control%diagonalization%mom_didguess) &
         CALL cp_abort(cp__l("qs_mom_methods.F",391), &
                       "The current implementation of the maximum overlap method is incompatible with the initial SCF guess")

      ! copy old molecular orbitals
      IF (scf_env%iter_count >= scf_control%diagonalization%mom_start) THEN
         ALLOCATE (mos_orig(SIZE(mos)))
         DO ispin = 1, SIZE(mos)
            CALL duplicate_mo_set(mos_orig(ispin)%mo_set, mos(ispin)%mo_set)
         END DO
      END IF

      ! solve the eigenproblem
      CALL general_eigenproblem(scf_env, mos, matrix_ks, matrix_s, scf_control, scf_section, diis_step)

      ! number of spins == dft_control%nspins
      nspins = SIZE(matrix_ks)

      IF (scf_env%iter_count >= scf_control%diagonalization%mom_start) THEN
         DO ispin = 1, nspins

            CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=maxocc, &
                            nao=nao, nmo=nmo, occupation_numbers=occ_nums)

            ! TO DO: sparse-matrix variant; check if use_mo_coeff_b is set, and if yes use mo_coeff_b instead
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL get_mo_set(mo_set=mos_orig(ispin)%mo_set, mo_coeff=mo_coeff_orig)

            NULLIFY (svec, overlap, mo_mo_fmstruct)

            ! NOTE: whereas it is possible to store the result of the following
            ! matrix multiplication into work matrices scf_env%scf_work2(1:nspins),
            ! we create the our own work storage, since these mentioned works matrices
            ! are allocated only if neither OT nor davidson diagonalization is in use.
            CALL cp_fm_create(svec, mo_coeff%matrix_struct)

            ! S * C(new)
            CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, mo_coeff, svec, nmo)

            ! a temporary overlap matrix
            CALL cp_fm_struct_create(mo_mo_fmstruct, nrow_global=nmo, ncol_global=nmo, &
                                     para_env=mo_coeff%matrix_struct%para_env, context=mo_coeff%matrix_struct%context)
            CALL cp_fm_create(overlap, mo_mo_fmstruct)
            CALL cp_fm_struct_release(mo_mo_fmstruct)

            ! C(old occupied)^T * S * C(new)
            CALL cp_fm_column_scale(mo_coeff_orig, occ_nums)
            CALL cp_gemm('T', 'N', nmo, nmo, nao, 1.0_dp, mo_coeff_orig, svec, 0.0_dp, overlap)
            CALL cp_fm_release(svec)

            ALLOCATE (proj(nmo))
            ALLOCATE (inds(nmo))
            ALLOCATE (tmp_occ_nums(nmo))

            ! project the new molecular orbitals into the space of the occupied old ones
            SELECT CASE (scf_control%diagonalization%mom_proj_formula)
            CASE (momproj_sum)
               ! proj_j = abs( \sum_i overlap(i, j) )
               CALL cp_fm_vectorssum(overlap, proj)

               DO iproj = 1, nmo
                  proj(iproj) = ABS(proj(iproj))
               END DO

            CASE (momproj_norm)
               ! proj_j = (\sum_i overlap(i, j)**2) ** 0.5
               CALL cp_fm_vectorsnorm(overlap, proj)

            CASE DEFAULT
               CALL cp__b("qs_mom_methods.F",459,"Unimplemented projection formula")
            END SELECT

            CALL cp_fm_release(overlap)

            tmp_occ_nums(:) = occ_nums(:)
            ! sort occupation numbers in ascending order
            CALL sort(tmp_occ_nums, nmo, inds)
            ! sort overlap projection in ascending order
            CALL sort(proj, nmo, inds)

            ! reorder occupation numbers according to overlap projections
            DO iproj = 1, nmo
               occ_nums(inds(iproj)) = tmp_occ_nums(iproj)
            ENDDO

            DEALLOCATE (tmp_occ_nums)
            DEALLOCATE (inds)
            DEALLOCATE (proj)

            ! locate the lowest non-fully occupied orbital
            DO lfomo = 1, nmo
               IF (occ_nums(lfomo) /= maxocc) EXIT
            ENDDO

            ! locate the highest occupied orbital
            DO homo = nmo, 1, -1
               IF (occ_nums(homo) > 0.0_dp) EXIT
            ENDDO

            CALL set_mo_set(mo_set=mos(ispin)%mo_set, homo=homo, lfomo=lfomo)

         END DO

         DO ispin = 1, SIZE(mos_orig)
            CALL deallocate_mo_set(mos_orig(ispin)%mo_set)
         ENDDO
         DEALLOCATE (mos_orig)
      ENDIF

      ! recompute density matrix
      DO ispin = 1, nspins
         CALL calculate_density_matrix(mos(ispin)%mo_set, scf_env%p_mix_new(ispin, 1)%matrix)
      END DO

      CALL timestop(handle)

   END SUBROUTINE do_mom_diag

END MODULE qs_mom_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition and initialisation of the mo data type.
!> \par History
!>      - adapted to the new QS environment data structure (02.04.2002,MK)
!>      - set_mo_occupation added (17.04.02,MK)
!>      - correct_mo_eigenvalues added (18.04.02,MK)
!>      - calculate_density_matrix moved from qs_scf to here (22.04.02,MK)
!>      - mo_set_p_type added (23.04.02,MK)
!>      - PRIVATE attribute set for TYPE mo_set_type (23.04.02,MK)
!>      - started conversion to LSD (1.2003, Joost VandeVondele)
!>      - Split of from qs_mo_types (07.2014, JGH)
!> \author Matthias Krack (09.05.2001,MK)
! **************************************************************************************************
MODULE qs_mo_io

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_binary_write,&
                                              cp_dbcsr_checksum,&
                                              cp_dbcsr_create,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_p_type,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm,&
                                              cp_fm_write_unformatted
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast
   USE orbital_pointers,                ONLY: indco,&
                                              nco,&
                                              nso
   USE orbital_symbols,                 ONLY: cgf_symbol,&
                                              sgf_symbol
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: evolt
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: mo_set_p_type,&
                                              mo_set_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mo_io'

   PUBLIC :: wfn_restart_file_name, &
             read_mo_set, &
             write_mo_set, &
             write_rt_mos_to_restart, &
             read_rt_mos_from_restart, &
             write_dm_binary_restart

   INTERFACE read_mo_set
      MODULE PROCEDURE read_mo_set_from_restart
   END INTERFACE

   INTERFACE write_mo_set
      MODULE PROCEDURE write_mo_set_to_output_unit, write_mo_set_to_restart
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param particle_set ...
!> \param dft_section ...
!> \param qs_kind_set ...
! **************************************************************************************************
   SUBROUTINE write_mo_set_to_restart(mo_array, particle_set, dft_section, qs_kind_set)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_mo_set_to_restart', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=30), DIMENSION(2), PARAMETER :: &
         keys = (/"SCF%PRINT%RESTART_HISTORY", "SCF%PRINT%RESTART        "/)

      INTEGER                                            :: handle, ikey, ires, ispin
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(1)), cp_p_file) .OR. &
          BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(2)), cp_p_file)) THEN

         IF (mo_array(1)%mo_set%use_mo_coeff_b) THEN
            ! we are using the dbcsr mo_coeff
            ! we copy it to the fm for anycase
            DO ispin = 1, SIZE(mo_array)
               IF (.NOT. ASSOCIATED(mo_array(ispin)%mo_set%mo_coeff_b)) THEN
                  IF(.NOT.(.FALSE.))CALL cp__a("qs_mo_io.F",140)
               ENDIF
               CALL copy_dbcsr_to_fm(mo_array(ispin)%mo_set%mo_coeff_b, &
                                     mo_array(ispin)%mo_set%mo_coeff) !fm->dbcsr
            ENDDO
         ENDIF

         DO ikey = 1, SIZE(keys)

            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 dft_section, keys(ikey)), cp_p_file)) THEN

               ires = cp_print_key_unit_nr(logger, dft_section, keys(ikey), &
                                           extension=".wfn", file_status="REPLACE", file_action="WRITE", &
                                           do_backup=.TRUE., file_form="UNFORMATTED")

               CALL write_mo_set_low(mo_array, particle_set=particle_set, &
                                     qs_kind_set=qs_kind_set, ires=ires)

               CALL cp_print_key_finished_output(ires, logger, dft_section, TRIM(keys(ikey)))
            END IF
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_mo_set_to_restart

! **************************************************************************************************
!> \brief calculates density matrix from mo set and writes the density matrix
!>        into a binary restart file
!> \param mo_array mos
!> \param dft_section dft input section
!> \param tmpl_matrix template dbcsr matrix
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE write_dm_binary_restart(mo_array, dft_section, tmpl_matrix)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: tmpl_matrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_dm_binary_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, project_name
      INTEGER                                            :: handle, ispin, unit_nr
      LOGICAL                                            :: do_dm_restart
      REAL(KIND=dp)                                      :: cs_pos
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_p_tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      project_name = logger%iter_info%project_name
      CALL section_vals_val_get(dft_section, "SCF%PRINT%DM_RESTART_WRITE", l_val=do_dm_restart)
      NULLIFY (matrix_p_tmp)

      IF (do_dm_restart) THEN
         ALLOCATE (matrix_p_tmp)
         DO ispin = 1, SIZE(mo_array)
            CALL cp_dbcsr_init(matrix_p_tmp)
            CALL cp_dbcsr_create(matrix_p_tmp, template=tmpl_matrix(ispin)%matrix, name="DM RESTART")

            IF (.NOT. ASSOCIATED(mo_array(ispin)%mo_set%mo_coeff_b)) CALL cp__b("qs_mo_io.F",210,"mo_coeff_b NOT ASSOCIATED")

            CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, mo_array(ispin)%mo_set%mo_coeff_b)
            CALL calculate_density_matrix(mo_array(ispin)%mo_set, matrix_p_tmp, &
                                          use_dbcsr=.TRUE., retain_sparsity=.FALSE.)

            WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_SCF_DM_SPIN_", ispin, "_RESTART.dm"
            cs_pos = cp_dbcsr_checksum(matrix_p_tmp, pos=.TRUE.)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T2,A,E20.8)') "Writing restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
            ENDIF
            CALL cp_dbcsr_binary_write(matrix_p_tmp, file_name)

            CALL cp_dbcsr_release(matrix_p_tmp)
         ENDDO
         DEALLOCATE (matrix_p_tmp)
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_dm_binary_restart

! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param rt_mos ...
!> \param particle_set ...
!> \param dft_section ...
!> \param qs_kind_set ...
! **************************************************************************************************
   SUBROUTINE write_rt_mos_to_restart(mo_array, rt_mos, particle_set, dft_section, &
                                      qs_kind_set)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: rt_mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_rt_mos_to_restart', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=43), DIMENSION(2), PARAMETER :: keys = (/ &
         "REAL_TIME_PROPAGATION%PRINT%RESTART_HISTORY", &
         "REAL_TIME_PROPAGATION%PRINT%RESTART        "/)

      INTEGER                                            :: handle, ikey, ires
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(1)), cp_p_file) .OR. &
          BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(2)), cp_p_file)) THEN

         DO ikey = 1, SIZE(keys)

            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 dft_section, keys(ikey)), cp_p_file)) THEN

               ires = cp_print_key_unit_nr(logger, dft_section, keys(ikey), &
                                           extension=".rtpwfn", file_status="REPLACE", file_action="WRITE", &
                                           do_backup=.TRUE., file_form="UNFORMATTED")

               CALL write_mo_set_low(mo_array, rt_mos=rt_mos, qs_kind_set=qs_kind_set, &
                                     particle_set=particle_set, ires=ires)
               CALL cp_print_key_finished_output(ires, logger, dft_section, TRIM(keys(ikey)))
            END IF
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_rt_mos_to_restart

! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param ires ...
!> \param rt_mos ...
! **************************************************************************************************
   SUBROUTINE write_mo_set_low(mo_array, qs_kind_set, particle_set, ires, rt_mos)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER                                            :: ires
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rt_mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_mo_set_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ikind, imat, iset, &
                                                            ishell, ispin, lmax, lshell, &
                                                            max_block, nao, natom, nmo, nset, &
                                                            nset_max, nshell_max, nspin
      INTEGER, DIMENSION(:), POINTER                     :: nset_info, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, nshell_info
      INTEGER, DIMENSION(:, :, :), POINTER               :: nso_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      CALL timeset(routineN, handle)
      nspin = SIZE(mo_array)
      nao = mo_array(1)%mo_set%nao

      IF (ires > 0) THEN
         !     *** create some info about the basis set first ***
         natom = SIZE(particle_set, 1)
         nset_max = 0
         nshell_max = 0

         DO iatom = 1, natom
            NULLIFY (orb_basis_set, dftb_parameter)
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             basis_set=orb_basis_set, dftb_parameter=dftb_parameter)
            IF (ASSOCIATED(orb_basis_set)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      nset=nset, &
                                      nshell=nshell, &
                                      l=l)
               nset_max = MAX(nset_max, nset)
               DO iset = 1, nset
                  nshell_max = MAX(nshell_max, nshell(iset))
               END DO
            ELSEIF (ASSOCIATED(dftb_parameter)) THEN
               CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
               nset_max = MAX(nset_max, 1)
               nshell_max = MAX(nshell_max, lmax+1)
            ELSE
               CALL cp__b("qs_mo_io.F",345,"Unknown basis type. ")
            END IF
         END DO

         ALLOCATE (nso_info(nshell_max, nset_max, natom))
         nso_info(:, :, :) = 0

         ALLOCATE (nshell_info(nset_max, natom))
         nshell_info(:, :) = 0

         ALLOCATE (nset_info(natom))
         nset_info(:) = 0

         DO iatom = 1, natom
            NULLIFY (orb_basis_set, dftb_parameter)
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             basis_set=orb_basis_set, dftb_parameter=dftb_parameter)
            IF (ASSOCIATED(orb_basis_set)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      nset=nset, &
                                      nshell=nshell, &
                                      l=l)
               nset_info(iatom) = nset
               DO iset = 1, nset
                  nshell_info(iset, iatom) = nshell(iset)
                  DO ishell = 1, nshell(iset)
                     lshell = l(ishell, iset)
                     nso_info(ishell, iset, iatom) = nso(lshell)
                  END DO
               END DO
            ELSEIF (ASSOCIATED(dftb_parameter)) THEN
               CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
               nset_info(iatom) = 1
               nshell_info(1, iatom) = lmax+1
               DO ishell = 1, lmax+1
                  lshell = ishell-1
                  nso_info(ishell, 1, iatom) = nso(lshell)
               END DO
            ELSE
               CALL cp__b("qs_mo_io.F",385,"Unknown basis set type. ")
            END IF
         END DO

         WRITE (ires) natom, nspin, nao, nset_max, nshell_max
         WRITE (ires) nset_info
         WRITE (ires) nshell_info
         WRITE (ires) nso_info

         DEALLOCATE (nset_info)

         DEALLOCATE (nshell_info)

         DEALLOCATE (nso_info)
      END IF

      ! use the scalapack block size as a default for buffering columns
      CALL cp_fm_get_info(mo_array(1)%mo_set%mo_coeff, ncol_block=max_block)

      DO ispin = 1, nspin
         nmo = mo_array(ispin)%mo_set%nmo
         IF ((ires > 0) .AND. (nmo > 0)) THEN
            WRITE (ires) nmo, &
               mo_array(ispin)%mo_set%homo, &
               mo_array(ispin)%mo_set%lfomo, &
               mo_array(ispin)%mo_set%nelectron
            WRITE (ires) mo_array(ispin)%mo_set%eigenvalues(1:nmo), &
               mo_array(ispin)%mo_set%occupation_numbers(1:nmo)
         END IF
         IF (PRESENT(rt_mos)) THEN
            DO imat = 2*ispin-1, 2*ispin
               CALL cp_fm_write_unformatted(rt_mos(imat)%matrix, ires)
            END DO
         ELSE
            CALL cp_fm_write_unformatted(mo_array(ispin)%mo_set%mo_coeff, ires)
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE write_mo_set_low

! **************************************************************************************************
!> \brief ...
!> \param filename ...
!> \param exist ...
!> \param section ...
!> \param logger ...
!> \param kp ...
!> \param xas ...
!> \param rtp ...
! **************************************************************************************************
   SUBROUTINE wfn_restart_file_name(filename, exist, section, logger, kp, xas, rtp)
      CHARACTER(LEN=default_path_length), INTENT(OUT)    :: filename
      LOGICAL, INTENT(OUT)                               :: exist
      TYPE(section_vals_type), POINTER                   :: section
      TYPE(cp_logger_type), POINTER                      :: logger
      LOGICAL, INTENT(IN), OPTIONAL                      :: kp, xas, rtp

      INTEGER                                            :: n_rep_val
      LOGICAL                                            :: my_kp, my_rtp, my_xas
      TYPE(section_vals_type), POINTER                   :: print_key

      my_kp = .FALSE.
      my_xas = .FALSE.
      my_rtp = .FALSE.
      IF (PRESENT(kp)) my_kp = kp
      IF (PRESENT(xas)) my_xas = xas
      IF (PRESENT(rtp)) my_rtp = rtp

      exist = .FALSE.
      CALL section_vals_val_get(section, "WFN_RESTART_FILE_NAME", n_rep_val=n_rep_val)
      IF (n_rep_val > 0) THEN
         CALL section_vals_val_get(section, "WFN_RESTART_FILE_NAME", c_val=filename)
      ELSE
         IF (my_xas) THEN
            ! try to read from the filename that is generated automatically from the printkey
            print_key => section_vals_get_subs_vals(section, "PRINT%RESTART")
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension="", my_local=.FALSE.)
         ELSE IF (my_rtp) THEN
            ! try to read from the filename that is generated automatically from the printkey
            print_key => section_vals_get_subs_vals(section, "REAL_TIME_PROPAGATION%PRINT%RESTART")
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension=".rtpwfn", my_local=.FALSE.)
         ELSE IF (my_kp) THEN
            ! try to read from the filename that is generated automatically from the printkey
            print_key => section_vals_get_subs_vals(section, "SCF%PRINT%RESTART")
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension=".kp", my_local=.FALSE.)
         ELSE
            ! try to read from the filename that is generated automatically from the printkey
            print_key => section_vals_get_subs_vals(section, "SCF%PRINT%RESTART")
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension=".wfn", my_local=.FALSE.)
         END IF
      ENDIF
      IF (.NOT. my_xas) THEN
         INQUIRE (FILE=filename, exist=exist)
      END IF

   END SUBROUTINE wfn_restart_file_name

! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param para_env ...
!> \param id_nr ...
!> \param multiplicity ...
!> \param dft_section ...
!> \param natom_mismatch ...
! **************************************************************************************************
   SUBROUTINE read_mo_set_from_restart(mo_array, atomic_kind_set, qs_kind_set, particle_set, &
                                       para_env, id_nr, multiplicity, dft_section, natom_mismatch)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: id_nr, multiplicity
      TYPE(section_vals_type), POINTER                   :: dft_section
      LOGICAL, INTENT(OUT), OPTIONAL                     :: natom_mismatch

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_mo_set_from_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: group, handle, ispin, natom, nspin, &
                                                            restart_unit, source
      LOGICAL                                            :: exist
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      nspin = SIZE(mo_array)
      restart_unit = -1

      group = para_env%group
      source = para_env%source

      IF (para_env%ionode) THEN

         natom = SIZE(particle_set, 1)
         CALL wfn_restart_file_name(file_name, exist, dft_section, logger)
         IF (id_nr /= 0) THEN
            ! Is it one of the backup files?
            file_name = TRIM(file_name)//".bak-"//ADJUSTL(cp_to_string(id_nr))
         END IF

         CALL open_file(file_name=file_name, &
                        file_action="READ", &
                        file_form="UNFORMATTED", &
                        file_status="OLD", &
                        unit_number=restart_unit)

      END IF

      CALL read_mos_restart_low(mo_array, para_env=para_env, qs_kind_set=qs_kind_set, &
                                particle_set=particle_set, natom=natom, &
                                rst_unit=restart_unit, multiplicity=multiplicity, natom_mismatch=natom_mismatch)
      IF (PRESENT(natom_mismatch)) THEN
         ! read_mos_restart_low only the io_node returns natom_mismatch, must broadcast it
         CALL mp_bcast(natom_mismatch, source, group)
         IF (natom_mismatch) THEN
            IF (para_env%ionode) CALL close_file(unit_number=restart_unit)
            CALL timestop(handle)
            RETURN
         ENDIF
      ENDIF

      ! Close restart file
      IF (para_env%ionode) CALL close_file(unit_number=restart_unit)

      DO ispin = 1, nspin
         CALL write_mo_set(mo_array(ispin)%mo_set, atomic_kind_set, qs_kind_set, &
                           particle_set, 4, dft_section)
      END DO

      CALL timestop(handle)

   END SUBROUTINE read_mo_set_from_restart
! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param rt_mos ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param para_env ...
!> \param id_nr ...
!> \param multiplicity ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_rt_mos_from_restart(mo_array, rt_mos, atomic_kind_set, qs_kind_set, &
                                       particle_set, para_env, id_nr, multiplicity, dft_section)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: rt_mos
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: id_nr, multiplicity
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_rt_mos_from_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: group, handle, ispin, natom, nspin, &
                                                            restart_unit, source
      LOGICAL                                            :: exist
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      nspin = SIZE(mo_array)
      restart_unit = -1

      group = para_env%group
      source = para_env%source

      IF (para_env%ionode) THEN

         natom = SIZE(particle_set, 1)
         CALL wfn_restart_file_name(file_name, exist, dft_section, logger, rtp=.TRUE.)
         IF (id_nr /= 0) THEN
            ! Is it one of the backup files?
            file_name = TRIM(file_name)//".bak-"//ADJUSTL(cp_to_string(id_nr))
         END IF

         CALL open_file(file_name=file_name, &
                        file_action="READ", &
                        file_form="UNFORMATTED", &
                        file_status="OLD", &
                        unit_number=restart_unit)

      END IF

      CALL read_mos_restart_low(mo_array, rt_mos=rt_mos, para_env=para_env, &
                                particle_set=particle_set, qs_kind_set=qs_kind_set, natom=natom, &
                                rst_unit=restart_unit, multiplicity=multiplicity)

      ! Close restart file
      IF (para_env%ionode) CALL close_file(unit_number=restart_unit)

      DO ispin = 1, nspin
         CALL write_mo_set(mo_array(ispin)%mo_set, atomic_kind_set, qs_kind_set, &
                           particle_set, 4, dft_section)
      END DO

      CALL timestop(handle)

   END SUBROUTINE read_rt_mos_from_restart

! **************************************************************************************************
!> \brief Reading the mos from apreviously defined restart file
!> \param mos ...
!> \param para_env ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param natom ...
!> \param rst_unit ...
!> \param multiplicity ...
!> \param rt_mos ...
!> \param natom_mismatch ...
!> \par History
!>      12.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE read_mos_restart_low(mos, para_env, qs_kind_set, particle_set, natom, rst_unit, &
                                   multiplicity, rt_mos, natom_mismatch)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: natom, rst_unit
      INTEGER, INTENT(in), OPTIONAL                      :: multiplicity
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rt_mos
      LOGICAL, INTENT(OUT), OPTIONAL                     :: natom_mismatch

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_mos_restart_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: group, homo, homo_read, i, iatom, ikind, imat, irow, iset, iset_read, ishell, &
         ishell_read, iso, ispin, lfomo_read, lmax, lshell, my_mult, nao, nao_read, natom_read, &
         nelectron, nelectron_read, nmo, nmo_read, nnshell, nset, nset_max, nshell_max, nspin, &
         nspin_read, offset_read, source
      INTEGER, DIMENSION(:), POINTER                     :: nset_info, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, nshell_info
      INTEGER, DIMENSION(:, :, :), POINTER               :: nso_info, offset_info
      LOGICAL                                            :: minbas, natom_match, use_this
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eig_read, occ_read
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer, vecbuffer_read
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      logger => cp_get_default_logger()

      nspin = SIZE(mos)
      nao = mos(1)%mo_set%nao
      my_mult = 0
      IF (PRESENT(multiplicity)) my_mult = multiplicity
      group = para_env%group
      source = para_env%source

      IF (para_env%ionode) THEN
         READ (rst_unit) natom_read, nspin_read, nao_read, nset_max, nshell_max
         IF (PRESENT(rt_mos)) THEN
            IF (nspin_read /= nspin) THEN
               CALL cp__b("qs_mo_io.F",704,"To change nspin is not possible. ")
            END IF
         ELSE
            ! we should allow for restarting with different spin settings
            IF (nspin_read /= nspin) THEN
               WRITE (cp_logger_get_default_unit_nr(logger), *) &
                  "READ RESTART : WARNING : nspin is not equal "
            END IF
            ! this case needs fixing of homo/lfomo/nelec/occupations ...
            IF (nspin_read > nspin) THEN
               CALL cp__b("qs_mo_io.F",714,"Reducing nspin is not possible. ")
            ENDIF
         END IF

         natom_match = (natom_read == natom)

         IF (natom_match) THEN ! actually do the read read

            ! Let's make it possible to change the basis set
            ALLOCATE (nso_info(nshell_max, nset_max, natom_read))
            ALLOCATE (nshell_info(nset_max, natom_read))
            ALLOCATE (nset_info(natom_read))
            ALLOCATE (offset_info(nshell_max, nset_max, natom_read))

            IF (nao_read /= nao) THEN
               WRITE (cp_logger_get_default_unit_nr(logger), *) &
                  " READ RESTART : WARNING : DIFFERENT # AOs ", nao, nao_read
               IF (PRESENT(rt_mos)) &
                  CALL cp__b("qs_mo_io.F",732,"To change basis is not possible. ")
            END IF

            READ (rst_unit) nset_info
            READ (rst_unit) nshell_info
            READ (rst_unit) nso_info

            i = 1
            DO iatom = 1, natom
               DO iset = 1, nset_info(iatom)
                  DO ishell = 1, nshell_info(iset, iatom)
                     offset_info(ishell, iset, iatom) = i
                     i = i+nso_info(ishell, iset, iatom)
                  END DO
               END DO
            END DO

            ALLOCATE (vecbuffer_read(1, nao_read))

         END IF ! natom_match
      END IF ! ionode

      ! make natom_match and natom_mismatch uniform across all nodes
      CALL mp_bcast(natom_match, source, group)
      IF (PRESENT(natom_mismatch)) natom_mismatch = .NOT. natom_match
      ! handle natom_match false
      IF (.NOT. natom_match) THEN
         IF (PRESENT(natom_mismatch)) THEN
            WRITE (cp_logger_get_default_unit_nr(logger), *) &
               " READ RESTART : WARNING : DIFFERENT natom, returning ", natom, natom_read
            RETURN
         ELSE
            CALL cp__b("qs_mo_io.F",764,"Incorrect number of atoms in restart file. ")
         ENDIF
      ENDIF

      CALL mp_bcast(nspin_read, source, group)

      ALLOCATE (vecbuffer(1, nao))

      DO ispin = 1, nspin

         nmo = mos(ispin)%mo_set%nmo
         homo = mos(ispin)%mo_set%homo
         mos(ispin)%mo_set%eigenvalues(:) = 0.0_dp
         mos(ispin)%mo_set%occupation_numbers(:) = 0.0_dp
         CALL cp_fm_set_all(mos(ispin)%mo_set%mo_coeff, 0.0_dp)

         IF (para_env%ionode .AND. (nmo > 0)) THEN
            READ (rst_unit) nmo_read, homo_read, lfomo_read, nelectron_read
            ALLOCATE (eig_read(nmo_read), occ_read(nmo_read))
            eig_read = 0.0_dp
            occ_read = 0.0_dp

            nmo = MIN(nmo, nmo_read)
            IF (nmo_read < nmo) &
               CALL cp_warn(cp__l("qs_mo_io.F",788), &
                            "The number of MOs on the restart unit is smaller than the number of "// &
                            "the allocated MOs. The MO set will be padded with zeros!")
            IF (nmo_read > nmo) &
               CALL cp_warn(cp__l("qs_mo_io.F",792), &
                            "The number of MOs on the restart unit is greater than the number of "// &
                            "the allocated MOs. The read MO set will be truncated!")

            READ (rst_unit) eig_read(1:nmo_read), occ_read(1:nmo_read)
            mos(ispin)%mo_set%eigenvalues(1:nmo) = eig_read(1:nmo)
            mos(ispin)%mo_set%occupation_numbers(1:nmo) = occ_read(1:nmo)
            DEALLOCATE (eig_read, occ_read)

            mos(ispin)%mo_set%homo = homo_read
            mos(ispin)%mo_set%lfomo = lfomo_read
            IF (homo_read > nmo) THEN
               IF (nelectron_read == mos(ispin)%mo_set%nelectron) THEN
                  CALL cp_warn(cp__l("qs_mo_io.F",805), &
                               "The number of occupied MOs on the restart unit is larger than "// &
                               "the allocated MOs. The read MO set will be truncated and the occupation numbers recalculated!")
                  CALL set_mo_occupation(mo_set=mos(ispin)%mo_set)
               ELSE
                  ! can not make this a warning i.e. homo must be smaller than nmo
                  ! otherwise e.g. set_mo_occupation will go out of bounds
                  CALL cp__b("qs_mo_io.F",812,"Number of occupied MOs on restart unit larger than allocated MOs. ")
               END IF
            END IF
         END IF

         CALL mp_bcast(nmo, source, group)
         CALL mp_bcast(mos(ispin)%mo_set%homo, source, group)
         CALL mp_bcast(mos(ispin)%mo_set%lfomo, source, group)
         CALL mp_bcast(mos(ispin)%mo_set%nelectron, source, group)
         CALL mp_bcast(mos(ispin)%mo_set%eigenvalues, source, group)
         CALL mp_bcast(mos(ispin)%mo_set%occupation_numbers, source, group)
         IF (PRESENT(rt_mos)) THEN
            DO imat = 2*ispin-1, 2*ispin
               DO i = 1, nmo
                  IF (para_env%ionode) THEN
                     READ (rst_unit) vecbuffer
                  ELSE
                     vecbuffer(1, :) = 0.0_dp
                  END IF
                  CALL mp_bcast(vecbuffer, source, group)
                  CALL cp_fm_set_submatrix(rt_mos(imat)%matrix, &
                                           vecbuffer, 1, i, nao, 1, transpose=.TRUE.)
               END DO
            END DO
         ELSE
            DO i = 1, nmo
               IF (para_env%ionode) THEN
                  READ (rst_unit) vecbuffer_read
                  ! now, try to assign the read to the real vector
                  ! in case the basis set changed this involves some guessing
                  irow = 1
                  DO iatom = 1, natom
                     NULLIFY (orb_basis_set, dftb_parameter, l, nshell)
                     CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
                     CALL get_qs_kind(qs_kind_set(ikind), &
                                      basis_set=orb_basis_set, dftb_parameter=dftb_parameter)
                     IF (ASSOCIATED(orb_basis_set)) THEN
                        CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                               nset=nset, &
                                               nshell=nshell, &
                                               l=l)
                        minbas = .FALSE.
                     ELSEIF (ASSOCIATED(dftb_parameter)) THEN
                        CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
                        nset = 1
                        minbas = .TRUE.
                     ELSE
                        CALL cp__b("qs_mo_io.F",859,"Unknown basis set type. ")
                     END IF

                     use_this = .TRUE.
                     iset_read = 1
                     DO iset = 1, nset
                        ishell_read = 1
                        IF (minbas) THEN
                           nnshell = lmax+1
                        ELSE
                           nnshell = nshell(iset)
                        END IF
                        DO ishell = 1, nnshell
                           IF (minbas) THEN
                              lshell = ishell-1
                           ELSE
                              lshell = l(ishell, iset)
                           END IF
                           IF (iset_read > nset_info(iatom)) use_this = .FALSE.
                           IF (use_this) THEN ! avoids out of bound access of the lower line if false
                              IF (nso(lshell) == nso_info(ishell_read, iset_read, iatom)) THEN
                                 offset_read = offset_info(ishell_read, iset_read, iatom)
                                 ishell_read = ishell_read+1
                                 IF (ishell_read > nshell_info(iset, iatom)) THEN
                                    ishell_read = 1
                                    iset_read = iset_read+1
                                 END IF
                              ELSE
                                 use_this = .FALSE.
                              END IF
                           END IF
                           DO iso = 1, nso(lshell)
                              IF (use_this) THEN
                                 IF (offset_read-1+iso .LT. 1 .OR. offset_read-1+iso .GT. nao_read) THEN
                                    vecbuffer(1, irow) = 0.0_dp
                                 ELSE
                                    vecbuffer(1, irow) = vecbuffer_read(1, offset_read-1+iso)
                                 END IF
                              ELSE
                                 vecbuffer(1, irow) = 0.0_dp
                              END IF
                              irow = irow+1
                           END DO
                           use_this = .TRUE.
                        END DO
                     END DO
                  END DO

               ELSE

                  vecbuffer(1, :) = 0.0_dp

               END IF

               CALL mp_bcast(vecbuffer, source, group)
               CALL cp_fm_set_submatrix(mos(ispin)%mo_set%mo_coeff, &
                                        vecbuffer, 1, i, nao, 1, transpose=.TRUE.)
            END DO
         END IF
         ! Skip extra MOs if there any
         IF (para_env%ionode) THEN
            !ignore nmo = 0
            IF (nmo > 0) THEN
               DO i = nmo+1, nmo_read
                  READ (rst_unit) vecbuffer_read
               END DO
            END IF
         END IF

         IF (.NOT. PRESENT(rt_mos)) THEN
            IF (ispin == 1 .AND. nspin_read < nspin) THEN

               mos(ispin+1)%mo_set%homo = mos(ispin)%mo_set%homo
               mos(ispin+1)%mo_set%lfomo = mos(ispin)%mo_set%lfomo
               nelectron = mos(ispin)%mo_set%nelectron
               IF (my_mult .NE. 1) THEN
                  CALL cp_abort(cp__l("qs_mo_io.F",935), &
                                "Restarting an LSD calculation from an LDA wfn only works for multiplicity=1 (singlets).")
               END IF
               IF (mos(ispin+1)%mo_set%nelectron < 0) THEN
                  CALL cp__b("qs_mo_io.F",939,"LSD: too few electrons for this multiplisity. ")
               END IF
               mos(ispin+1)%mo_set%eigenvalues = mos(ispin)%mo_set%eigenvalues
               mos(ispin)%mo_set%occupation_numbers = mos(ispin)%mo_set%occupation_numbers/2.0_dp
               mos(ispin+1)%mo_set%occupation_numbers = mos(ispin)%mo_set%occupation_numbers
               CALL cp_fm_to_fm(mos(ispin)%mo_set%mo_coeff, mos(ispin+1)%mo_set%mo_coeff)
               EXIT
            END IF
         END IF
      END DO ! ispin

      DEALLOCATE (vecbuffer)

      IF (para_env%ionode) THEN
         DEALLOCATE (vecbuffer_read)
         DEALLOCATE (offset_info)
         DEALLOCATE (nso_info)
         DEALLOCATE (nshell_info)
         DEALLOCATE (nset_info)
      END IF

   END SUBROUTINE read_mos_restart_low

! **************************************************************************************************
!> \brief   Write the MO eigenvalues, MO occupation numbers and
!>          MO mo_coeff.
!> \param mo_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param before ...
!> \param dft_section ...
!> \param spin ...
!> \param last ...
!> \date    15.05.2001
!> \par History:
!>       - Optionally print Cartesian MOs (20.04.2005,MK)
!> \par Variables
!>       - after : Number of digits after point.
!>       - before: Number of digits before point.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_mo_set_to_output_unit(mo_set, atomic_kind_set, qs_kind_set, particle_set, &
                                          before, dft_section, spin, last)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: before
      TYPE(section_vals_type), POINTER                   :: dft_section
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: spin
      LOGICAL, INTENT(IN), OPTIONAL                      :: last

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_mo_set_to_output_unit', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=12)                                  :: symbol
      CHARACTER(LEN=12), DIMENSION(:), POINTER           :: bcgf_symbol
      CHARACTER(LEN=16)                                  :: fmtstr5, fmtstr6
      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=2*default_string_length)             :: name
      CHARACTER(LEN=22)                                  :: fmtstr2
      CHARACTER(LEN=25)                                  :: fmtstr1, fmtstr7
      CHARACTER(LEN=27)                                  :: fmtstr4
      CHARACTER(LEN=38)                                  :: fmtstr3
      CHARACTER(LEN=6), DIMENSION(:), POINTER            :: bsgf_symbol
      INTEGER :: after, first_mo, from, iatom, icgf, ico, icol, ikind, imo, irow, iset, isgf, &
         ishell, iso, iw, jcol, last_mo, left, lmax, lshell, natom, ncgf, ncol, ncol_global, &
         nrow_global, nset, nsgf, right, scf_step, to, width
      INTEGER, DIMENSION(:), POINTER                     :: mo_index_range, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l
      LOGICAL                                            :: ionode, my_last, p_cart, p_eval, p_evec, &
                                                            p_occ, should_output
      REAL(KIND=dp)                                      :: gap
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cmatrix, smatrix
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      NULLIFY (bcgf_symbol)
      NULLIFY (bsgf_symbol)
      NULLIFY (logger)
      NULLIFY (mo_index_range)
      NULLIFY (nshell)

      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source
      CALL section_vals_val_get(dft_section, "PRINT%MO%EIGENVALUES", l_val=p_eval)
      CALL section_vals_val_get(dft_section, "PRINT%MO%EIGENVECTORS", l_val=p_evec)
      CALL section_vals_val_get(dft_section, "PRINT%MO%OCCUPATION_NUMBERS", l_val=p_occ)
      CALL section_vals_val_get(dft_section, "PRINT%MO%CARTESIAN", l_val=p_cart)
      CALL section_vals_val_get(dft_section, "PRINT%MO%MO_INDEX_RANGE", i_vals=mo_index_range)
      CALL section_vals_val_get(dft_section, "PRINT%MO%NDIGITS", i_val=after)
      after = MIN(MAX(after, 1), 16)
      should_output = BTEST(cp_print_key_should_output(logger%iter_info, dft_section, &
                                                       "PRINT%MO"), cp_p_file)

      IF ((.NOT. should_output) .OR. (.NOT. (p_eval .OR. p_evec .OR. p_occ))) RETURN

      IF (PRESENT(last)) THEN
         my_last = last
      ELSE
         my_last = .FALSE.
      END IF

      scf_step = logger%iter_info%iteration(logger%iter_info%n_rlevel)-1

      IF (p_evec) THEN
         CALL cp_fm_get_info(mo_set%mo_coeff, &
                             nrow_global=nrow_global, &
                             ncol_global=ncol_global)
         ALLOCATE (smatrix(nrow_global, ncol_global))
         CALL cp_fm_get_submatrix(mo_set%mo_coeff, smatrix)
         IF (.NOT. ionode) THEN
            DEALLOCATE (smatrix)
         END IF
      END IF

      iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%MO", &
                                ignore_should_output=should_output, &
                                extension=".MOLog")

      IF (iw > 0) THEN

         CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
         CALL get_qs_kind_set(qs_kind_set, ncgf=ncgf, nsgf=nsgf)

         ! Definition of the variable formats

         fmtstr1 = "(/,T2,21X,  (  X,I5,  X))"
         fmtstr2 = "(T2,21X,  (1X,F  .  ))"
         fmtstr3 = "(T2,I5,1X,I5,1X,A,1X,A6,  (1X,F  .  ))"

         width = before+after+3
         ncol = INT(56/width)

         right = MAX((after-2), 1)
         left = width-right-5

         WRITE (UNIT=fmtstr1(11:12), FMT="(I2)") ncol
         WRITE (UNIT=fmtstr1(14:15), FMT="(I2)") left
         WRITE (UNIT=fmtstr1(21:22), FMT="(I2)") right

         WRITE (UNIT=fmtstr2(9:10), FMT="(I2)") ncol
         WRITE (UNIT=fmtstr2(16:17), FMT="(I2)") width-1
         WRITE (UNIT=fmtstr2(19:20), FMT="(I2)") after

         WRITE (UNIT=fmtstr3(25:26), FMT="(I2)") ncol
         WRITE (UNIT=fmtstr3(32:33), FMT="(I2)") width-1
         WRITE (UNIT=fmtstr3(35:36), FMT="(I2)") after

         IF (p_evec) THEN

            IF (p_cart) THEN

               IF (my_last) THEN
                  name = "MO EIGENVALUES, MO OCCUPATION NUMBERS, AND "// &
                         "CARTESIAN MO EIGENVECTORS"
               ELSE
                  WRITE (UNIT=name, FMT="(A,I6)") &
                     "MO EIGENVALUES, MO OCCUPATION NUMBERS, AND "// &
                     "CARTESIAN MO EIGENVECTORS AFTER SCF STEP", scf_step
               END IF

               ALLOCATE (cmatrix(ncgf, ncgf))

               cmatrix = 0.0_dp

               ! Transform spherical MOs to Cartesian MOs

               icgf = 1
               isgf = 1
               DO iatom = 1, natom
                  NULLIFY (orb_basis_set, dftb_parameter)
                  CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
                  CALL get_qs_kind(qs_kind_set(ikind), &
                                   basis_set=orb_basis_set, &
                                   dftb_parameter=dftb_parameter)
                  IF (ASSOCIATED(orb_basis_set)) THEN
                     CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                            nset=nset, &
                                            nshell=nshell, &
                                            l=l)
                     DO iset = 1, nset
                        DO ishell = 1, nshell(iset)
                           lshell = l(ishell, iset)
                           CALL dgemm("T", "N", nco(lshell), nsgf, nso(lshell), 1.0_dp, &
                                      orbtramat(lshell)%s2c, nso(lshell), &
                                      smatrix(isgf, 1), nsgf, 0.0_dp, &
                                      cmatrix(icgf, 1), ncgf)
                           icgf = icgf+nco(lshell)
                           isgf = isgf+nso(lshell)
                        END DO
                     END DO
                  ELSE IF (ASSOCIATED(dftb_parameter)) THEN
                     CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
                     DO ishell = 1, lmax+1
                        lshell = ishell-1
                        CALL dgemm("T", "N", nco(lshell), nsgf, nso(lshell), 1.0_dp, &
                                   orbtramat(lshell)%s2c, nso(lshell), &
                                   smatrix(isgf, 1), nsgf, 0.0_dp, &
                                   cmatrix(icgf, 1), ncgf)
                        icgf = icgf+nco(lshell)
                        isgf = isgf+nso(lshell)
                     END DO
                  ELSE
                     CALL cp__b("qs_mo_io.F",1147,"Unknown basis set type. ")
                  END IF
               END DO ! iatom

            ELSE

               IF (my_last) THEN
                  name = "MO EIGENVALUES, MO OCCUPATION NUMBERS, AND "// &
                         "SPHERICAL MO EIGENVECTORS"
               ELSE
                  WRITE (UNIT=name, FMT="(A,I6)") &
                     "MO EIGENVALUES, MO OCCUPATION NUMBERS, AND "// &
                     "SPHERICAL MO EIGENVECTORS AFTER SCF STEP", scf_step
               END IF

            END IF ! p_cart

         ELSE IF (p_occ .OR. p_eval) THEN

            IF (my_last) THEN
               name = "MO EIGENVALUES AND MO OCCUPATION NUMBERS"
            ELSE
               WRITE (UNIT=name, FMT="(A,I6)") &
                  "MO EIGENVALUES AND MO OCCUPATION NUMBERS AFTER "// &
                  "SCF STEP", scf_step
            END IF

         END IF ! p_evec

         CALL compress(name)

         ! Print headline

         IF (PRESENT(spin)) THEN
            WRITE (UNIT=iw, FMT="(/,/,T2,A)") spin//" "//TRIM(name)
         ELSE
            WRITE (UNIT=iw, FMT="(/,/,T2,A)") TRIM(name)
         END IF

         ! Check if only a subset of the MOs has to be printed
         IF ((mo_index_range(1) > 0) .AND. &
             (mo_index_range(2) > 0) .AND. &
             (mo_index_range(2) >= mo_index_range(1))) THEN
            first_mo = MAX(1, mo_index_range(1))
            last_mo = MIN(mo_set%nmo, mo_index_range(2))
         ELSE
            first_mo = 1
            last_mo = mo_set%nmo
         END IF

         IF (p_evec) THEN

            ! Print full MO information

            DO icol = first_mo, last_mo, ncol

               from = icol
               to = MIN((from+ncol-1), last_mo)

               WRITE (UNIT=iw, FMT=fmtstr1) (jcol, jcol=from, to)
               WRITE (UNIT=iw, FMT=fmtstr2) (mo_set%eigenvalues(jcol), jcol=from, to)
               WRITE (UNIT=iw, FMT="(A)") ""

               WRITE (UNIT=iw, FMT=fmtstr2) (mo_set%occupation_numbers(jcol), jcol=from, to)
               WRITE (UNIT=iw, FMT="(A)") ""

               irow = 1

               DO iatom = 1, natom

                  IF (iatom /= 1) WRITE (UNIT=iw, FMT="(A)") ""

                  NULLIFY (orb_basis_set, dftb_parameter)
                  CALL get_atomic_kind(particle_set(iatom)%atomic_kind, &
                                       element_symbol=element_symbol, kind_number=ikind)
                  CALL get_qs_kind(qs_kind_set(ikind), &
                                   basis_set=orb_basis_set, &
                                   dftb_parameter=dftb_parameter)

                  IF (p_cart) THEN

                     IF (ASSOCIATED(orb_basis_set)) THEN
                        CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                               nset=nset, &
                                               nshell=nshell, &
                                               l=l, &
                                               cgf_symbol=bcgf_symbol)

                        icgf = 1
                        DO iset = 1, nset
                           DO ishell = 1, nshell(iset)
                              lshell = l(ishell, iset)
                              DO ico = 1, nco(lshell)
                                 WRITE (UNIT=iw, FMT=fmtstr3) &
                                    irow, iatom, ADJUSTR(element_symbol), bcgf_symbol(icgf), &
                                    (cmatrix(irow, jcol), jcol=from, to)
                                 icgf = icgf+1
                                 irow = irow+1
                              END DO
                           END DO
                        END DO
                     ELSE IF (ASSOCIATED(dftb_parameter)) THEN
                        CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
                        icgf = 1
                        DO ishell = 1, lmax+1
                           lshell = ishell-1
                           DO ico = 1, nco(lshell)
                              symbol = cgf_symbol(1, indco(1:3, icgf))
                              symbol(1:2) = "  "
                              WRITE (UNIT=iw, FMT=fmtstr3) &
                                 irow, iatom, ADJUSTR(element_symbol), symbol, &
                                 (cmatrix(irow, jcol), jcol=from, to)
                              icgf = icgf+1
                              irow = irow+1
                           END DO
                        END DO
                     ELSE
                        CALL cp__b("qs_mo_io.F",1264,"Unknown basis set type. ")
                     END IF

                  ELSE

                     IF (ASSOCIATED(orb_basis_set)) THEN
                        CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                               nset=nset, &
                                               nshell=nshell, &
                                               l=l, &
                                               sgf_symbol=bsgf_symbol)
                        isgf = 1
                        DO iset = 1, nset
                           DO ishell = 1, nshell(iset)
                              lshell = l(ishell, iset)
                              DO iso = 1, nso(lshell)
                                 WRITE (UNIT=iw, FMT=fmtstr3) &
                                    irow, iatom, ADJUSTR(element_symbol), bsgf_symbol(isgf), &
                                    (smatrix(irow, jcol), jcol=from, to)
                                 isgf = isgf+1
                                 irow = irow+1
                              END DO
                           END DO
                        END DO
                     ELSE IF (ASSOCIATED(dftb_parameter)) THEN
                        CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
                        isgf = 1
                        DO ishell = 1, lmax+1
                           lshell = ishell-1
                           DO iso = 1, nso(lshell)
                              symbol = sgf_symbol(1, lshell, -lshell+iso-1)
                              symbol(1:2) = "  "
                              WRITE (UNIT=iw, FMT=fmtstr3) &
                                 irow, iatom, ADJUSTR(element_symbol), symbol, &
                                 (smatrix(irow, jcol), jcol=from, to)
                              isgf = isgf+1
                              irow = irow+1
                           END DO
                        END DO
                     ELSE
                        CALL cp__b("qs_mo_io.F",1304,"Unknown basis type. ")
                     END IF

                  END IF ! p_cart

               END DO ! iatom

            END DO ! icol

            WRITE (UNIT=iw, FMT="(/)")

            ! Release work storage

            DEALLOCATE (smatrix)
            IF (p_cart) THEN
               DEALLOCATE (cmatrix)
            END IF

         ELSE IF (p_occ .OR. p_eval) THEN

            fmtstr4 = "(T2,I9,2X,F28.  ,1X,F24.  )"
            WRITE (UNIT=fmtstr4(15:16), FMT="(I2)") after
            WRITE (UNIT=fmtstr4(25:26), FMT="(I2)") after
            WRITE (UNIT=iw, FMT="(/,A)") &
               "# MO index          MO eigenvalue [a.u.]            MO occupation"
            DO imo = first_mo, last_mo
               WRITE (UNIT=iw, FMT=fmtstr4) &
                  imo, mo_set%eigenvalues(imo), mo_set%occupation_numbers(imo)
            END DO
            fmtstr5 = "(A,T42,F24.  ,/)"
            WRITE (UNIT=fmtstr5(12:13), FMT="(I2)") after
            WRITE (UNIT=iw, FMT=fmtstr5) &
               "# Sum", accurate_sum(mo_set%occupation_numbers(:))

         END IF ! p_evec

         fmtstr6 = "(A,T17,F24.  ,/)"
         WRITE (UNIT=fmtstr6(12:13), FMT="(I2)") after
         WRITE (UNIT=iw, FMT=fmtstr6) "  Fermi energy:", mo_set%mu
         IF ((mo_set%homo > 0)) THEN
            IF ((mo_set%occupation_numbers(mo_set%homo) == mo_set%maxocc) .AND. (last_mo > mo_set%homo)) THEN
               gap = mo_set%eigenvalues(mo_set%homo+1)- &
                     mo_set%eigenvalues(mo_set%homo)
               fmtstr7 = "(A,T17,F24.  ,A,F6.2,A,/)"
               WRITE (UNIT=fmtstr7(12:13), FMT="(I2)") after
               WRITE (UNIT=iw, FMT=fmtstr7) &
                  "  HOMO-LUMO gap:", gap, " = ", gap*evolt, " eV"
            END IF
         END IF

      END IF ! iw

      CALL cp_print_key_finished_output(iw, logger, dft_section, "PRINT%MO", &
                                        ignore_should_output=should_output)

   END SUBROUTINE write_mo_set_to_output_unit

END MODULE qs_mo_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Restart file for k point calculations
!> \par History
!> \author JGH (30.09.2015)
! **************************************************************************************************
MODULE kpoint_io

   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_info,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_read_unformatted,&
                                              cp_fm_type,&
                                              cp_fm_write_unformatted
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE message_passing,                 ONLY: mp_bcast
   USE orbital_pointers,                ONLY: nso
   USE particle_types,                  ONLY: particle_type
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_mo_io,                        ONLY: wfn_restart_file_name
   USE qs_scf_types,                    ONLY: qs_scf_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kpoint_io'

   INTEGER, PARAMETER                   :: version = 1

   PUBLIC :: read_kpoints_restart, &
             write_kpoints_restart

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param denmat ...
!> \param kpoints ...
!> \param scf_env ...
!> \param dft_section ...
!> \param particle_set ...
!> \param qs_kind_set ...
! **************************************************************************************************
   SUBROUTINE write_kpoints_restart(denmat, kpoints, scf_env, dft_section, &
                                    particle_set, qs_kind_set)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: denmat
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_kpoints_restart', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=30), DIMENSION(2), PARAMETER :: &
         keys = (/"SCF%PRINT%RESTART_HISTORY", "SCF%PRINT%RESTART        "/)

      INTEGER                                            :: handle, ic, ikey, ires, ispin, nimages, &
                                                            nspin
      INTEGER, DIMENSION(3)                              :: cell
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      TYPE(cp_fm_type), POINTER                          :: fmat
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(1)), cp_p_file) .OR. &
          BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, keys(2)), cp_p_file)) THEN

         DO ikey = 1, SIZE(keys)

            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 dft_section, keys(ikey)), cp_p_file)) THEN

               ires = cp_print_key_unit_nr(logger, dft_section, keys(ikey), &
                                           extension=".kp", file_status="REPLACE", file_action="WRITE", &
                                           do_backup=.TRUE., file_form="UNFORMATTED")

               CALL write_kpoints_file_header(qs_kind_set, particle_set, ires)

               nspin = SIZE(denmat, 1)
               nimages = SIZE(denmat, 2)
               NULLIFY (cell_to_index)
               IF (nimages > 1) THEN
                  CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
               END IF
               IF(.NOT.(ASSOCIATED(scf_env%scf_work1)))CALL cp__a("kpoint_io.F",120)
               fmat => scf_env%scf_work1(1)%matrix

               DO ispin = 1, nspin
                  IF (ires > 0) WRITE (ires) ispin, nspin, nimages
                  DO ic = 1, nimages
                     IF (nimages > 1) THEN
                        cell = get_cell(ic, cell_to_index)
                     ELSE
                        cell = 0
                     END IF
                     IF (ires > 0) WRITE (ires) ic, cell
                     CALL copy_dbcsr_to_fm(denmat(ispin, ic)%matrix, fmat)
                     CALL cp_fm_write_unformatted(fmat, ires)
                  END DO
               END DO

               CALL cp_print_key_finished_output(ires, logger, dft_section, TRIM(keys(ikey)))

            END IF

         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE write_kpoints_restart

! **************************************************************************************************
!> \brief ...
!> \param ic ...
!> \param cell_to_index ...
!> \retval cell ...
! **************************************************************************************************
   FUNCTION get_cell(ic, cell_to_index) RESULT(cell)
      INTEGER, INTENT(IN)                                :: ic
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      INTEGER, DIMENSION(3)                              :: cell

      INTEGER                                            :: i1, i2, i3

      cell = 0
      ALLCELL: DO i3 = LBOUND(cell_to_index, 3), UBOUND(cell_to_index, 3)
         DO i2 = LBOUND(cell_to_index, 2), UBOUND(cell_to_index, 2)
            DO i1 = LBOUND(cell_to_index, 1), UBOUND(cell_to_index, 1)
               IF (cell_to_index(i1, i2, i3) == ic) THEN
                  cell(1) = i1
                  cell(2) = i2
                  cell(3) = i3
                  EXIT ALLCELL
               END IF
            END DO
         END DO
      END DO ALLCELL

   END FUNCTION get_cell

! **************************************************************************************************
!> \brief ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param ires ...
! **************************************************************************************************
   SUBROUTINE write_kpoints_file_header(qs_kind_set, particle_set, ires)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER                                            :: ires

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_kpoints_file_header', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ikind, iset, ishell, &
                                                            lmax, lshell, nao, natom, nset, &
                                                            nset_max, nsgf, nshell_max
      INTEGER, DIMENSION(:), POINTER                     :: nset_info, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, nshell_info
      INTEGER, DIMENSION(:, :, :), POINTER               :: nso_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      CALL timeset(routineN, handle)

      IF (ires > 0) THEN
         ! create some info about the basis set first
         natom = SIZE(particle_set, 1)
         nset_max = 0
         nshell_max = 0

         DO iatom = 1, natom
            NULLIFY (orb_basis_set, dftb_parameter)
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             basis_set=orb_basis_set, dftb_parameter=dftb_parameter)
            IF (ASSOCIATED(orb_basis_set)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      nset=nset, &
                                      nshell=nshell, &
                                      l=l)
               nset_max = MAX(nset_max, nset)
               DO iset = 1, nset
                  nshell_max = MAX(nshell_max, nshell(iset))
               END DO
            ELSEIF (ASSOCIATED(dftb_parameter)) THEN
               CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
               nset_max = MAX(nset_max, 1)
               nshell_max = MAX(nshell_max, lmax+1)
            ELSE
               CALL cp__b("kpoint_io.F",229,"Unknown basis type.")
            END IF
         END DO

         ALLOCATE (nso_info(nshell_max, nset_max, natom))
         nso_info(:, :, :) = 0

         ALLOCATE (nshell_info(nset_max, natom))
         nshell_info(:, :) = 0

         ALLOCATE (nset_info(natom))
         nset_info(:) = 0

         nao = 0
         DO iatom = 1, natom
            NULLIFY (orb_basis_set, dftb_parameter)
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             basis_set=orb_basis_set, dftb_parameter=dftb_parameter)
            IF (ASSOCIATED(orb_basis_set)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, nsgf=nsgf, nset=nset, nshell=nshell, l=l)
               nao = nao+nsgf
               nset_info(iatom) = nset
               DO iset = 1, nset
                  nshell_info(iset, iatom) = nshell(iset)
                  DO ishell = 1, nshell(iset)
                     lshell = l(ishell, iset)
                     nso_info(ishell, iset, iatom) = nso(lshell)
                  END DO
               END DO
            ELSEIF (ASSOCIATED(dftb_parameter)) THEN
               CALL get_dftb_atom_param(dftb_parameter, lmax=lmax)
               nset_info(iatom) = 1
               nshell_info(1, iatom) = lmax+1
               DO ishell = 1, lmax+1
                  lshell = ishell-1
                  nso_info(ishell, 1, iatom) = nso(lshell)
               END DO
               nao = nao+(lmax+1)**2
            ELSE
               CALL cp__b("kpoint_io.F",269,"Unknown basis set type.")
            END IF
         END DO

         WRITE (ires) version
         WRITE (ires) natom, nao, nset_max, nshell_max
         WRITE (ires) nset_info
         WRITE (ires) nshell_info
         WRITE (ires) nso_info

         DEALLOCATE (nset_info, nshell_info, nso_info)
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_kpoints_file_header

! **************************************************************************************************
!> \brief ...
!> \param denmat ...
!> \param kpoints ...
!> \param fmwork ...
!> \param natom ...
!> \param para_env ...
!> \param id_nr ...
!> \param dft_section ...
!> \param natom_mismatch ...
! **************************************************************************************************
   SUBROUTINE read_kpoints_restart(denmat, kpoints, fmwork, natom, &
                                   para_env, id_nr, dft_section, natom_mismatch)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: denmat
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fmwork
      INTEGER, INTENT(IN)                                :: natom
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: id_nr
      TYPE(section_vals_type), POINTER                   :: dft_section
      LOGICAL, INTENT(OUT)                               :: natom_mismatch

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_kpoints_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: group, handle, restart_unit, source
      LOGICAL                                            :: exist
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      restart_unit = -1

      group = para_env%group
      source = para_env%source

      IF (para_env%ionode) THEN

         CALL wfn_restart_file_name(file_name, exist, dft_section, logger, kp=.TRUE.)
         IF (id_nr /= 0) THEN
            ! Is it one of the backup files?
            file_name = TRIM(file_name)//".bak-"//ADJUSTL(cp_to_string(id_nr))
         END IF

         CALL open_file(file_name=file_name, &
                        file_action="READ", &
                        file_form="UNFORMATTED", &
                        file_status="OLD", &
                        unit_number=restart_unit)

      END IF

      IF(.NOT.(ASSOCIATED(fmwork)))CALL cp__a("kpoint_io.F",341)
      IF(.NOT.(ASSOCIATED(fmwork(1)%matrix)))CALL cp__a("kpoint_io.F",342)

      CALL read_kpoints_restart_low(denmat, kpoints, fmwork(1)%matrix, para_env, &
                                    natom, restart_unit, natom_mismatch)

      ! only the io_node returns natom_mismatch, must broadcast it
      CALL mp_bcast(natom_mismatch, source, group)

      ! Close restart file
      IF (para_env%ionode) CALL close_file(unit_number=restart_unit)

      CALL timestop(handle)

   END SUBROUTINE read_kpoints_restart

! **************************************************************************************************
!> \brief Reading the mos from apreviously defined restart file
!> \param denmat ...
!> \param kpoints ...
!> \param fmat ...
!> \param para_env ...
!> \param natom ...
!> \param rst_unit ...
!> \param natom_mismatch ...
!> \par History
!>      12.2007 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE read_kpoints_restart_low(denmat, kpoints, fmat, para_env, &
                                       natom, rst_unit, natom_mismatch)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: denmat
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(cp_fm_type), POINTER                          :: fmat
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: natom, rst_unit
      LOGICAL, INTENT(OUT)                               :: natom_mismatch

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_kpoints_restart_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: group, ic, image, ispin, nao, nao_read, natom_read, ni, nimages, nset_max, &
         nshell_max, nspin, nspin_read, source, version_read
      INTEGER, DIMENSION(3)                              :: cell
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: natom_match

      group = para_env%group
      source = para_env%source

      IF(.NOT.(ASSOCIATED(denmat)))CALL cp__a("kpoint_io.F",392)
      CALL cp_dbcsr_get_info(denmat(1, 1)%matrix, nfullrows_total=nao)

      nspin = SIZE(denmat, 1)
      nimages = SIZE(denmat, 2)
      NULLIFY (cell_to_index)
      IF (nimages > 1) THEN
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      END IF

      IF (para_env%ionode) THEN
         READ (rst_unit) version_read
         IF (version_read /= version) THEN
            CALL cp_abort(cp__l("kpoint_io.F",405), &
                          " READ RESTART : Version of restart file not supported")
         END IF
         READ (rst_unit) natom_read, nao_read, nset_max, nshell_max
         natom_match = (natom_read == natom)
         IF (natom_match) THEN
            IF (nao_read /= nao) THEN
               CALL cp_abort(cp__l("kpoint_io.F",412), &
                             " READ RESTART : Different number of basis functions")
            END IF
            READ (rst_unit)
            READ (rst_unit)
            READ (rst_unit)
         END IF
      END IF

      ! make natom_match and natom_mismatch uniform across all nodes
      CALL mp_bcast(natom_match, source, group)
      natom_mismatch = .NOT. natom_match
      ! handle natom_match false
      IF (natom_mismatch) THEN
         CALL cp_warn(cp__l("kpoint_io.F",426), &
                      " READ RESTART : WARNING : DIFFERENT natom, returning ")
      ELSE

         DO
            ispin = 0
            IF (para_env%ionode) THEN
               READ (rst_unit) ispin, nspin_read, ni
            END IF
            CALL mp_bcast(ispin, source, group)
            CALL mp_bcast(nspin_read, source, group)
            CALL mp_bcast(ni, source, group)
            IF (nspin_read /= nspin) THEN
               CALL cp_abort(cp__l("kpoint_io.F",439), &
                             " READ RESTART : Different spin polarisation not supported")
            END IF
            DO ic = 1, ni
               IF (para_env%ionode) THEN
                  READ (rst_unit) image, cell
               END IF
               CALL mp_bcast(image, source, group)
               CALL mp_bcast(cell, source, group)
               !
               CALL cp_fm_read_unformatted(fmat, rst_unit)
               !
               IF (nimages > 1) THEN
                  image = get_index(cell, cell_to_index)
               ELSE IF (SUM(ABS(cell(:))) == 0) THEN
                  image = 1
               ELSE
                  image = 0
               END IF
               IF (image >= 1 .AND. image <= nimages) THEN
                  CALL copy_fm_to_dbcsr(fmat, denmat(ispin, image)%matrix)
               END IF
            END DO
            IF (ispin == nspin_read) EXIT
         END DO

      ENDIF

   END SUBROUTINE read_kpoints_restart_low

! **************************************************************************************************
!> \brief ...
!> \param cell ...
!> \param cell_to_index ...
!> \retval index ...
! **************************************************************************************************
   FUNCTION get_index(cell, cell_to_index) RESULT(index)
      INTEGER, DIMENSION(3), INTENT(IN)                  :: cell
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      INTEGER                                            :: index

      IF (CELL(1) >= LBOUND(cell_to_index, 1) .AND. CELL(1) <= UBOUND(cell_to_index, 1) .AND. &
          CELL(2) >= LBOUND(cell_to_index, 2) .AND. CELL(2) <= UBOUND(cell_to_index, 2) .AND. &
          CELL(3) >= LBOUND(cell_to_index, 3) .AND. CELL(3) <= UBOUND(cell_to_index, 3)) THEN

         index = cell_to_index(cell(1), cell(2), cell(3))

      ELSE

         index = 0

      END IF

   END FUNCTION get_index

END MODULE kpoint_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to somehow generate an intial guess
!> \par History
!>       2006.03 Moved here from qs_scf.F [Joost VandeVondele]
! **************************************************************************************************
MODULE qs_initial_guess
   USE atom_kind_orbitals,              ONLY: calculate_atomic_orbitals
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_checksum, cp_dbcsr_copy, cp_dbcsr_filter, cp_dbcsr_get_diag, &
        cp_dbcsr_get_num_blocks, cp_dbcsr_get_occupation, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_nfullrows_total, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_set_diag, cp_dbcsr_trace, &
        cp_dbcsr_type, cp_dbcsr_verify_matrix
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply,&
                                              cp_fm_to_dbcsr_row_template
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_submatrix, cp_fm_init_random, cp_fm_p_type, cp_fm_release, &
        cp_fm_set_all, cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE external_potential_types,        ONLY: all_potential_type,&
                                              gth_potential_type
   USE input_constants,                 ONLY: atomic_guess,&
                                              core_guess,&
                                              history_guess,&
                                              mopac_guess,&
                                              no_guess,&
                                              random_guess,&
                                              restart_guess,&
                                              sparse_guess
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE kpoint_io,                       ONLY: read_kpoints_restart
   USE kpoint_types,                    ONLY: kpoint_type
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_mo_io,                        ONLY: read_mo_set,&
                                              wfn_restart_file_name
   USE qs_mo_methods,                   ONLY: calculate_density_matrix,&
                                              make_basis_lowdin,&
                                              make_basis_simple,&
                                              make_basis_sm
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_restrict
   USE qs_mom_methods,                  ONLY: do_mom_guess
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_methods,                  ONLY: eigensolver,&
                                              eigensolver_simple
   USE qs_scf_types,                    ONLY: block_davidson_diag_method_nr,&
                                              block_krylov_diag_method_nr,&
                                              ot_diag_method_nr,&
                                              qs_scf_env_type
   USE qs_wf_history_methods,           ONLY: wfi_update
   USE scf_control_types,               ONLY: scf_control_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_initial_guess'

   PUBLIC ::  calculate_first_density_matrix, calculate_atomic_block_dm, calculate_mopac_dm
   PUBLIC ::  calculate_atomic_fock_matrix

   TYPE atom_matrix_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: mat
   END TYPE atom_matrix_type

CONTAINS

! **************************************************************************************************
!> \brief can use a variety of methods to come up with an initial
!>      density matrix and optionally an initial wavefunction
!> \param scf_env  SCF environment information
!> \param qs_env   QS environment
!> \par History
!>      03.2006 moved here from qs_scf [Joost VandeVondele]
!>      06.2007 allow to skip the initial guess [jgh]
!>      08.2014 kpoints [JGH]
!> \note
!>      badly needs to be split in subroutines each doing one of the possible
!>      schemes
! **************************************************************************************************
   SUBROUTINE calculate_first_density_matrix(scf_env, qs_env)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_first_density_matrix', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, filename
      INTEGER :: atom_a, blk, density_guess, group, handle, homo, i, iatom, ic, icol, id_nr, &
         ikind, irow, iseed(4), ispin, istart_col, istart_row, j, last_read, n, n_cols, n_rows, &
         nao, natom, natoms, natoms_tmp, nelectron, nmo, nmo_tmp, not_read, nsgf, nspin, nvec, &
         output_unit, qs_env_id, safe_density_guess, size_atomic_kind_set, z
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, kind_of, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, elec_conf, nelec_kind, &
                                                            sort_kind
      LOGICAL                                            :: did_guess, do_kpoints, do_std_diag, &
                                                            exist, has_unit_metric, &
                                                            natom_mismatch, ofgpw
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: buff, buff2
      REAL(dp), DIMENSION(:, :), POINTER                 :: DATA
      REAL(KIND=dp)                                      :: checksum, eps, length, maxocc, occ, &
                                                            scale, trps1, zeff
      REAL(KIND=dp), DIMENSION(0:3)                      :: edftb
      TYPE(atom_matrix_type), DIMENSION(:), POINTER      :: pmat
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: h_core_sparse, matrix_ks, p_rmpv, &
                                                            s_sparse
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h_kp, matrix_ks_kp, matrix_s_kp, &
                                                            rho_ao_kp
      TYPE(cp_dbcsr_type)                                :: mo_dbcsr, mo_tmp_dbcsr
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: work1
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, moa, mob, ortho, sv, work2
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, input, subsys_section

      logger => cp_get_default_logger()
      NULLIFY (atomic_kind, qs_kind, mo_coeff, sv, orb_basis_set, atomic_kind_set, &
               qs_kind_set, particle_set, ortho, work2, work1, mo_array, s_sparse, &
               scf_control, dft_control, p_rmpv, ortho, work2, work1, para_env, &
               s_sparse, scf_control, dft_control, h_core_sparse, matrix_ks, rho)
      NULLIFY (dft_section, input, subsys_section)
      NULLIFY (matrix_s_kp, matrix_h_kp, matrix_ks_kp, rho_ao_kp)
      NULLIFY (moa, mob)
      NULLIFY (atom_list, elec_conf, kpoints)
      edftb = 0.0_dp

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      mos=mo_array, &
                      matrix_s_kp=matrix_s_kp, &
                      matrix_h_kp=matrix_h_kp, &
                      matrix_ks_kp=matrix_ks_kp, &
                      input=input, &
                      scf_control=scf_control, &
                      id_nr=qs_env_id, &
                      dft_control=dft_control, &
                      has_unit_metric=has_unit_metric, &
                      do_kpoints=do_kpoints, &
                      kpoints=kpoints, &
                      rho=rho, &
                      para_env=para_env)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)

      ! just initialize the first image, the other density are set to zero
      DO ispin = 1, dft_control%nspins
         DO ic = 1, SIZE(rho_ao_kp, 2)
            CALL cp_dbcsr_set(rho_ao_kp(ispin, ic)%matrix, 0.0_dp)
         END DO
      END DO
      s_sparse => matrix_s_kp(:, 1)
      h_core_sparse => matrix_h_kp(:, 1)
      matrix_ks => matrix_ks_kp(:, 1)
      p_rmpv => rho_ao_kp(:, 1)

      work1 => scf_env%scf_work1
      work2 => scf_env%scf_work2
      ortho => scf_env%ortho

      dft_section => section_vals_get_subs_vals(input, "DFT")

      nspin = dft_control%nspins
      ofgpw = dft_control%qs_control%ofgpw
      density_guess = scf_control%density_guess
      do_std_diag = .FALSE.

      safe_density_guess = atomic_guess
      IF (dft_control%qs_control%semi_empirical .OR. dft_control%qs_control%dftb .OR. &
          dft_control%qs_control%scptb) THEN
         IF (density_guess == atomic_guess) density_guess = mopac_guess
         ! in case we need to bail to a safe restart type later on
         safe_density_guess = mopac_guess
      END IF

      IF (scf_control%use_ot .AND. &
          (.NOT. ((density_guess == random_guess) .OR. &
                  (density_guess == atomic_guess) .OR. &
                  (density_guess == mopac_guess) .OR. &
                  (density_guess == sparse_guess) .OR. &
                  (((density_guess == restart_guess) .OR. &
                    (density_guess == history_guess)) .AND. &
                   (scf_control%level_shift == 0.0_dp))))) THEN
         CALL cp_abort(cp__l("qs_initial_guess.F",238), &
                       "OT needs GUESS ATOMIC / RANDOM / SPARSE / RESTART / HISTORY RESTART: other options NYI")
      END IF

      ! if a restart was requested, check that the file exists,
      ! if not we fall back to an atomic guess. No kidding, the file name should remain
      ! in sync with read_mo_set
      id_nr = 0
      IF (density_guess == restart_guess) THEN
         ! only check existence on I/O node, otherwise if file exists there but
         ! not on compute nodes, everything goes crazy even though only I/O
         ! node actually reads the file
         IF (do_kpoints) THEN
            IF (para_env%ionode) THEN
               CALL wfn_restart_file_name(file_name, exist, dft_section, logger, kp=.TRUE.)
            END IF
         ELSE
            IF (para_env%ionode) THEN
               CALL wfn_restart_file_name(file_name, exist, dft_section, logger)
            END IF
         ENDIF
         CALL mp_bcast(exist, para_env%source, para_env%group)
         CALL mp_bcast(file_name, para_env%source, para_env%group)
         IF (.NOT. exist) THEN
            CALL cp_warn(cp__l("qs_initial_guess.F",262), &
                         "User requested to restart the wavefunction from the file named: "// &
                         TRIM(file_name)//". This file does not exist. Please check the existence of"// &
                         " the file or change properly the value of the keyword WFN_RESTART_FILE_NAME."// &
                         " Calculation continues using ATOMIC GUESS. ")
            density_guess = safe_density_guess
         END IF
      ELSE IF (density_guess == history_guess) THEN
         IF (do_kpoints) THEN
            CALL cp__b("qs_initial_guess.F",271,"calculate_first_density_matrix: history_guess not implemented for k-points")
         ENDIF
         IF (para_env%ionode) &
            CALL wfn_restart_file_name(file_name, exist, dft_section, logger)
         CALL mp_bcast(exist, para_env%source, para_env%group)
         CALL mp_bcast(file_name, para_env%source, para_env%group)
         nvec = qs_env%wf_history%memory_depth
         not_read = nvec+1
         ! At this level we read the saved backup RESTART files..
         DO i = 1, nvec
            j = i-1
            filename = TRIM(file_name)
            IF (j /= 0) filename = TRIM(file_name)//".bak-"//ADJUSTL(cp_to_string(j))
            IF (para_env%ionode) &
               INQUIRE (FILE=filename, exist=exist)
            CALL mp_bcast(exist, para_env%source, para_env%group)
            IF ((.NOT. exist) .AND. (i < not_read)) THEN
               not_read = i
            END IF
         END DO
         IF (not_read == 1) THEN
            density_guess = restart_guess
            filename = TRIM(file_name)
            IF (para_env%ionode) INQUIRE (FILE=filename, exist=exist)
            CALL mp_bcast(exist, para_env%source, para_env%group)
            IF (.NOT. exist) THEN
               CALL cp_warn(cp__l("qs_initial_guess.F",297), &
                            "User requested to restart the wavefunction from a series of restart files named: "// &
                            TRIM(file_name)//" with extensions (.bak-n). These files do not exist."// &
                            " Even trying to switch to a plain restart wave-function failes because the"// &
                            " file named: "//TRIM(file_name)//" does not exist. Please check the existence of"// &
                            " the file or change properly the value of the keyword WFN_RESTART_FILE_NAME. "// &
                            " Calculation continues using ATOMIC GUESS. ")
               density_guess = safe_density_guess
            END IF
         END IF
         last_read = not_read-1
      END IF

      did_guess = .FALSE.

      IF (density_guess == restart_guess) THEN

         IF (do_kpoints) THEN
            natoms = SIZE(particle_set)
            CALL read_kpoints_restart(rho_ao_kp, kpoints, work1, &
                                      natoms, para_env, id_nr, dft_section, natom_mismatch)
            IF (natom_mismatch) density_guess = safe_density_guess
         ELSE
            CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set)
            CALL read_mo_set(mo_array, atomic_kind_set, qs_kind_set, particle_set, para_env, &
                             id_nr=id_nr, multiplicity=dft_control%multiplicity, dft_section=dft_section, &
                             natom_mismatch=natom_mismatch)

            IF (natom_mismatch) THEN
               density_guess = safe_density_guess
            ELSE
               DO ispin = 1, nspin
                  IF (scf_control%level_shift /= 0.0_dp) THEN
                     CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff=mo_coeff)
                     CALL cp_fm_to_fm(mo_coeff, ortho)
                  END IF

                  ! make all nmo vectors present orthonormal
                  CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                                  mo_coeff=mo_coeff, nmo=nmo, homo=homo)

                  IF (has_unit_metric) THEN
                     CALL make_basis_simple(mo_coeff, nmo)
                  ELSEIF (dft_control%smear) THEN
                     CALL make_basis_lowdin(vmatrix=mo_coeff, ncol=nmo, &
                                            matrix_s=s_sparse(1)%matrix)
                  ELSE
                     ! ortho so that one can restart for different positions (basis sets?)
                     CALL make_basis_sm(mo_coeff, homo, s_sparse(1)%matrix)
                  ENDIF
                  ! only alpha spin is kept for restricted
                  IF (dft_control%restricted) EXIT
               ENDDO
               IF (dft_control%restricted) CALL mo_set_restrict(mo_array)

               IF (.NOT. scf_control%diagonalization%mom) &
                  CALL set_mo_occupation(mo_array, smear=qs_env%scf_control%smear)

               DO ispin = 1, nspin

                  IF (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr) THEN !fm->dbcsr
                     CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                           mo_array(ispin)%mo_set%mo_coeff_b) !fm->dbcsr
                  ENDIF !fm->dbcsr

                  CALL calculate_density_matrix(mo_array(ispin)%mo_set, &
                                                p_rmpv(ispin)%matrix)
               ENDDO
            ENDIF ! natom_mismatch

         END IF

         ! Maximum Overlap Method
         IF (scf_control%diagonalization%mom) THEN
            CALL do_mom_guess(nspin, mo_array, scf_control, p_rmpv)
         ENDIF

         did_guess = .TRUE.
      END IF

      IF (density_guess == history_guess) THEN
         IF (not_read > 1) THEN
            DO i = 1, last_read
               j = last_read-i
               CALL read_mo_set(mo_array, atomic_kind_set, qs_kind_set, particle_set, para_env, &
                                id_nr=j, multiplicity=dft_control%multiplicity, &
                                dft_section=dft_section)

               DO ispin = 1, nspin
                  IF (scf_control%level_shift /= 0.0_dp) THEN
                     CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff=mo_coeff)
                     CALL cp_fm_to_fm(mo_coeff, ortho)
                  END IF

                  ! make all nmo vectors present orthonormal
                  CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo, homo=homo)

                  IF (has_unit_metric) THEN
                     CALL make_basis_simple(mo_coeff, nmo)
                  ELSE
                     ! ortho so that one can restart for different positions (basis sets?)
                     CALL make_basis_sm(mo_coeff, homo, s_sparse(1)%matrix)
                  ENDIF
                  ! only alpha spin is kept for restricted
                  IF (dft_control%restricted) EXIT
               END DO
               IF (dft_control%restricted) CALL mo_set_restrict(mo_array)

               DO ispin = 1, nspin
                  CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                         smear=qs_env%scf_control%smear)
               ENDDO

               DO ispin = 1, nspin
                  IF (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr) THEN !fm->dbcsr
                     CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                           mo_array(ispin)%mo_set%mo_coeff_b) !fm->dbcsr
                  END IF !fm->dbcsr
                  CALL calculate_density_matrix(mo_array(ispin)%mo_set, p_rmpv(ispin)%matrix)
               ENDDO

               ! Write to extrapolation pipeline
               CALL wfi_update(wf_history=qs_env%wf_history, qs_env=qs_env, dt=1.0_dp)
            END DO
         END IF

         did_guess = .TRUE.
      END IF

      IF (density_guess == random_guess) THEN

         DO ispin = 1, nspin
            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                            mo_coeff=mo_coeff, nmo=nmo)
            CALL cp_fm_init_random(mo_coeff, nmo)
            IF (has_unit_metric) THEN
               CALL make_basis_simple(mo_coeff, nmo)
            ELSE
               CALL make_basis_sm(mo_coeff, nmo, s_sparse(1)%matrix)
            ENDIF
            ! only alpha spin is kept for restricted
            IF (dft_control%restricted) EXIT
         ENDDO
         IF (dft_control%restricted) CALL mo_set_restrict(mo_array)

         DO ispin = 1, nspin
            CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                   smear=qs_env%scf_control%smear)
         ENDDO

         DO ispin = 1, nspin

            IF (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr) THEN !fm->dbcsr
               CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                     mo_array(ispin)%mo_set%mo_coeff_b) !fm->dbcsr
            ENDIF !fm->dbcsr

            CALL calculate_density_matrix(mo_array(ispin)%mo_set, p_rmpv(ispin)%matrix)
         ENDDO

         did_guess = .TRUE.
      END IF

      IF (density_guess == core_guess) THEN

         IF (do_kpoints) THEN
            CALL cp__b("qs_initial_guess.F",463,"calculate_first_density_matrix: core_guess not implemented for k-points")
         ENDIF

         ispin = 1
         ! Load core Hamiltonian into work matrix
         CALL copy_dbcsr_to_fm(h_core_sparse(1)%matrix, work1(ispin)%matrix)

         ! Diagonalize the core Hamiltonian matrix and retrieve a first set of
         ! molecular orbitals (MOs)
         IF (has_unit_metric) THEN
            CALL eigensolver_simple(matrix_ks=work1(ispin)%matrix, &
                                    mo_set=mo_array(ispin)%mo_set, &
                                    work=work2, &
                                    do_level_shift=.FALSE., &
                                    level_shift=0.0_dp, &
                                    use_jacobi=.FALSE., jacobi_threshold=0._dp)
         ELSE
            CALL eigensolver(matrix_ks_fm=work1(ispin)%matrix, &
                             mo_set=mo_array(ispin)%mo_set, &
                             ortho=ortho, &
                             work=work2, &
                             cholesky_method=scf_env%cholesky_method, &
                             do_level_shift=.FALSE., &
                             level_shift=0.0_dp, &
                             matrix_u_fm=null(), &
                             use_jacobi=.FALSE.)
         END IF

         ! Open shell case: copy alpha MOs to beta MOs
         IF (nspin == 2) THEN
            CALL get_mo_set(mo_set=mo_array(1)%mo_set, mo_coeff=moa)
            CALL get_mo_set(mo_set=mo_array(2)%mo_set, mo_coeff=mob, nmo=nmo)
            CALL cp_fm_to_fm(moa, mob, nmo)
         END IF

         ! Build an initial density matrix (for each spin in the case of
         ! an open shell calculation) from the first MOs set
         DO ispin = 1, nspin
            CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, smear=scf_control%smear)
            CALL calculate_density_matrix(mo_array(ispin)%mo_set, p_rmpv(ispin)%matrix)
         END DO

         did_guess = .TRUE.
      END IF

      IF (density_guess == atomic_guess) THEN

         subsys_section => section_vals_get_subs_vals(input, "SUBSYS")
         output_unit = cp_print_key_unit_nr(logger, subsys_section, "PRINT%KINDS", extension=".Log")
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T2,A))") &
               "Atomic guess: The first density matrix is obtained in terms of atomic orbitals", &
               "              and electronic configurations assigned to each atomic kind"
         END IF

         DO ispin = 1, nspin

            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, nelectron=nelectron)

            CALL calculate_atomic_block_dm(p_rmpv(ispin)%matrix, s_sparse(1)%matrix, &
                                           particle_set, atomic_kind_set, qs_kind_set, &
                                           ispin, nspin, nelectron, output_unit)

            ! The orbital transformation method (OT) requires not only an
            ! initial density matrix, but also an initial wavefunction (MO set)
            IF (ofgpw .AND. (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr)) THEN
               ! get orbitals later
            ELSE
               IF (ASSOCIATED(scf_env%krylov_space)) do_std_diag = (scf_env%krylov_space%eps_std_diag > 0.0_dp)
               IF (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr .OR. &
                   (scf_env%method == block_krylov_diag_method_nr .AND. .NOT. do_std_diag) &
                   .OR. dft_control%do_admm .OR. scf_env%method == block_davidson_diag_method_nr) THEN
                  IF (dft_control%restricted .AND. (ispin == 2)) THEN
                     CALL mo_set_restrict(mo_array)
                  ELSE
                     CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                                     mo_coeff=mo_coeff, &
                                     nmo=nmo, nao=nao, homo=homo)

                     CALL cp_fm_set_all(mo_coeff, 0.0_dp)
                     CALL cp_fm_init_random(mo_coeff, nmo)

                     CALL cp_fm_create(sv, mo_coeff%matrix_struct, "SV")
                     ! multiply times PS
                     IF (has_unit_metric) THEN
                        CALL cp_fm_to_fm(mo_coeff, sv)
                     ELSE
                        ! PS*C(:,1:nomo)+C(:,nomo+1:nmo) (nomo=NINT(nelectron/maxocc))
                        CALL cp_dbcsr_sm_fm_multiply(s_sparse(1)%matrix, mo_coeff, sv, nmo)
                     END IF
                     CALL cp_dbcsr_sm_fm_multiply(p_rmpv(ispin)%matrix, sv, mo_coeff, homo)

                     CALL cp_fm_release(sv)
                     ! and ortho the result
                     IF (has_unit_metric) THEN
                        CALL make_basis_simple(mo_coeff, nmo)
                     ELSE
                        CALL make_basis_sm(mo_coeff, nmo, s_sparse(1)%matrix)
                     END IF
                  END IF

                  CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                         smear=qs_env%scf_control%smear)

                  CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                        mo_array(ispin)%mo_set%mo_coeff_b) !fm->dbcsr

                  CALL calculate_density_matrix(mo_array(ispin)%mo_set, &
                                                p_rmpv(ispin)%matrix)
               END IF
            END IF

         END DO

         IF (ofgpw .AND. (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr)) THEN
            ! We fit a function to the square root of the density
            CALL qs_rho_update_rho(rho, qs_env)
            IF(.NOT.(1 == 0))CALL cp__a("qs_initial_guess.F",580)
!         CALL cp_fm_create(sv,mo_coeff%matrix_struct,"SV")
!         DO ispin=1,nspin
!           CALL integrate_ppl_rspace(qs%rho%rho_r(ispin),qs_env)
!           CALL cp_cfm_solve(overlap,mos)
!           CALL get_mo_set(mo_set=mo_array(ispin)%mo_set,&
!                           mo_coeff=mo_coeff, nmo=nmo, nao=nao)
!           CALL cp_fm_init_random(mo_coeff,nmo)
!         END DO
!         CALL cp_fm_release(sv)
         END IF

         IF (scf_control%diagonalization%mom) THEN
            CALL do_mom_guess(nspin, mo_array, scf_control, p_rmpv)
         ENDIF

         CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                           "PRINT%KINDS")

         did_guess = .TRUE.
      END IF

      IF (density_guess == sparse_guess) THEN

         IF (ofgpw) CALL cp__b("qs_initial_guess.F",604,"SPARSE_GUESS not implemented for OFGPW")
         IF (.NOT. scf_control%use_ot) CALL cp__b("qs_initial_guess.F",605,"OT needed!")
         IF (do_kpoints) THEN
            CALL cp__b("qs_initial_guess.F",607,"calculate_first_density_matrix: sparse_guess not implemented for k-points")
         ENDIF

         eps = 1.0E-5_dp

         output_unit = cp_logger_get_default_io_unit(logger)
         group = para_env%group
         natoms = SIZE(particle_set)
         ALLOCATE (kind_of(natoms))
         ALLOCATE (first_sgf(natoms), last_sgf(natoms))

         checksum = cp_dbcsr_checksum(s_sparse(1)%matrix)
         i = cp_dbcsr_get_num_blocks(s_sparse(1)%matrix); CALL mp_sum(i, group)
         IF (output_unit > 0) WRITE (output_unit, *) 'S nblks', i, ' checksum', checksum
         CALL cp_dbcsr_filter(s_sparse(1)%matrix, eps)
         checksum = cp_dbcsr_checksum(s_sparse(1)%matrix)
         i = cp_dbcsr_get_num_blocks(s_sparse(1)%matrix); CALL mp_sum(i, group)
         IF (output_unit > 0) WRITE (output_unit, *) 'S nblks', i, ' checksum', checksum

         CALL get_particle_set(particle_set, qs_kind_set, first_sgf=first_sgf, &
                               last_sgf=last_sgf)
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, kind_of=kind_of)

         ALLOCATE (pmat(SIZE(atomic_kind_set)))

         DO ispin = 1, nspin
            scale = 1._dp
            IF (nspin == 2) scale = 0.5_dp
            DO ikind = 1, SIZE(atomic_kind_set)
               atomic_kind => atomic_kind_set(ikind)
               qs_kind => qs_kind_set(ikind)
               NULLIFY (pmat(ikind)%mat)
               CALL calculate_atomic_orbitals(atomic_kind, qs_kind, pmat=pmat(ikind)%mat, ispin=ispin)
               NULLIFY (atomic_kind)
            END DO
            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                            maxocc=maxocc, &
                            nelectron=nelectron)
            !
            CALL cp_dbcsr_iterator_start(iter, p_rmpv(ispin)%matrix)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, irow, icol, DATA, blk)
               ikind = kind_of(irow)
               IF (icol .EQ. irow) DATA(:, :) = pmat(ikind)%mat(:, :)*scale
            ENDDO
            CALL cp_dbcsr_iterator_stop(iter)

            !CALL cp_dbcsr_verify_matrix(p_rmpv(ispin)%matrix)
            checksum = cp_dbcsr_checksum(p_rmpv(ispin)%matrix)
            occ = cp_dbcsr_get_occupation(p_rmpv(ispin)%matrix)
            IF (output_unit > 0) WRITE (output_unit, *) 'P_init occ', occ, ' checksum', checksum
            ! so far p needs to have the same sparsity as S
            !CALL cp_dbcsr_filter(p_rmpv(ispin)%matrix, eps)
            !CALL cp_dbcsr_verify_matrix(p_rmpv(ispin)%matrix)
            checksum = cp_dbcsr_checksum(p_rmpv(ispin)%matrix)
            occ = cp_dbcsr_get_occupation(p_rmpv(ispin)%matrix)
            IF (output_unit > 0) WRITE (output_unit, *) 'P_init occ', occ, ' checksum', checksum

            CALL cp_dbcsr_trace(p_rmpv(ispin)%matrix, s_sparse(1)%matrix, trps1)
            scale = REAL(nelectron, dp)/trps1
            CALL cp_dbcsr_scale(p_rmpv(ispin)%matrix, scale)

            !CALL cp_dbcsr_verify_matrix(p_rmpv(ispin)%matrix)
            checksum = cp_dbcsr_checksum(p_rmpv(ispin)%matrix)
            occ = cp_dbcsr_get_occupation(p_rmpv(ispin)%matrix)
            IF (output_unit > 0) WRITE (output_unit, *) 'P occ', occ, ' checksum', checksum
            !
            ! The orbital transformation method (OT) requires not only an
            ! initial density matrix, but also an initial wavefunction (MO set)
            IF (dft_control%restricted .AND. (ispin == 2)) THEN
               CALL mo_set_restrict(mo_array)
            ELSE
               CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                               mo_coeff=mo_coeff, &
                               nmo=nmo, nao=nao, homo=homo)
               CALL cp_fm_set_all(mo_coeff, 0.0_dp)

               n = MAXVAL(last_sgf-first_sgf)+1
               size_atomic_kind_set = SIZE(atomic_kind_set)

               ALLOCATE (buff(n, n), sort_kind(size_atomic_kind_set), &
                         nelec_kind(size_atomic_kind_set))
               !
               ! sort kind vs nbr electron
               DO ikind = 1, size_atomic_kind_set
                  atomic_kind => atomic_kind_set(ikind)
                  qs_kind => qs_kind_set(ikind)
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       natom=natom, &
                                       atom_list=atom_list, &
                                       z=z)
                  CALL get_qs_kind(qs_kind, nsgf=nsgf, elec_conf=elec_conf, &
                                   basis_set=orb_basis_set, zeff=zeff)
                  nelec_kind(ikind) = SUM(elec_conf)
               ENDDO
               CALL sort(nelec_kind, size_atomic_kind_set, sort_kind)
               !
               ! a -very- naive sparse guess
               nmo_tmp = nmo
               natoms_tmp = natoms
               istart_col = 1
               iseed(1) = 4; iseed(2) = 3; iseed(3) = 2; iseed(4) = 1 ! set the seed for dlarnv
               DO i = 1, size_atomic_kind_set
                  ikind = sort_kind(i)
                  atomic_kind => atomic_kind_set(ikind)
                  CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                       natom=natom, atom_list=atom_list)
                  DO iatom = 1, natom
                     !
                     atom_a = atom_list(iatom)
                     istart_row = first_sgf(atom_a)
                     n_rows = last_sgf(atom_a)-first_sgf(atom_a)+1
                     !
                     ! compute the "potential" nbr of states for this atom
                     n_cols = MAX(INT(REAL(nmo_tmp, dp)/REAL(natoms_tmp, dp)), 1)
                     IF (n_cols .GT. n_rows) n_cols = n_rows
                     !
                     nmo_tmp = nmo_tmp-n_cols
                     natoms_tmp = natoms_tmp-1
                     IF (nmo_tmp .LT. 0 .OR. natoms_tmp .LT. 0) THEN
                        CALL cp__b("qs_initial_guess.F",727,"Wrong1!")
                     END IF
                     DO j = 1, n_cols
                        CALL dlarnv(1, iseed, n_rows, buff(1, j))
                     ENDDO
                     CALL cp_fm_set_submatrix(mo_coeff, buff, istart_row, istart_col, &
                                              n_rows, n_cols)
                     istart_col = istart_col+n_cols
                  ENDDO
               ENDDO

               IF (istart_col .LE. nmo) THEN
                  CALL cp__b("qs_initial_guess.F",739,"Wrong2!")
               END IF

               DEALLOCATE (buff, nelec_kind, sort_kind)

               IF (.FALSE.) THEN
                  ALLOCATE (buff(nao, 1), buff2(nao, 1))
                  DO i = 1, nmo
                     CALL cp_fm_get_submatrix(mo_coeff, buff, 1, i, nao, 1)
                     IF (SUM(buff**2) .LT. 1E-10_dp) THEN
                        WRITE (*, *) 'wrong', i, SUM(buff**2)
                     ENDIF
                     length = SQRT(DOT_PRODUCT(buff(:, 1), buff(:, 1)))
                     buff(:, :) = buff(:, :)/length
                     DO j = i+1, nmo
                        CALL cp_fm_get_submatrix(mo_coeff, buff2, 1, j, nao, 1)
                        length = SQRT(DOT_PRODUCT(buff2(:, 1), buff2(:, 1)))
                        buff2(:, :) = buff2(:, :)/length
                        IF (ABS(DOT_PRODUCT(buff(:, 1), buff2(:, 1))-1.0_dp) .LT. 1E-10_dp) THEN
                           WRITE (*, *) 'wrong2', i, j, DOT_PRODUCT(buff(:, 1), buff2(:, 1))
                           DO ikind = 1, nao
                              IF (ABS(mo_coeff%local_data(ikind, i)) .GT. 1e-10_dp) THEN
                                 WRITE (*, *) 'c1', ikind, mo_coeff%local_data(ikind, i)
                              ENDIF
                              IF (ABS(mo_coeff%local_data(ikind, j)) .GT. 1e-10_dp) THEN
                                 WRITE (*, *) 'c2', ikind, mo_coeff%local_data(ikind, j)
                              ENDIF
                           ENDDO
                           CALL cp__b("qs_initial_guess.F",767,"")
                        ENDIF
                     ENDDO
                  ENDDO
                  DEALLOCATE (buff, buff2)

               ENDIF
               !
               CALL cp_dbcsr_init(mo_dbcsr)
               CALL cp_fm_to_dbcsr_row_template(mo_dbcsr, mo_coeff, s_sparse(1)%matrix)
               !CALL cp_dbcsr_verify_matrix(mo_dbcsr)
               checksum = cp_dbcsr_checksum(mo_dbcsr)

               occ = cp_dbcsr_get_occupation(mo_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'C occ', occ, ' checksum', checksum
               CALL cp_dbcsr_filter(mo_dbcsr, eps)
               !CALL cp_dbcsr_verify_matrix(mo_dbcsr)
               occ = cp_dbcsr_get_occupation(mo_dbcsr)
               checksum = cp_dbcsr_checksum(mo_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'C occ', occ, ' checksum', checksum
               !
               ! multiply times PS
               IF (has_unit_metric) THEN
                  CALL cp__b("qs_initial_guess.F",790,"has_unit_metric will be removed soon")
               END IF
               !
               ! S*C
               CALL cp_dbcsr_init(mo_tmp_dbcsr)
               CALL cp_dbcsr_copy(mo_tmp_dbcsr, mo_dbcsr, name="mo_tmp")
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, s_sparse(1)%matrix, mo_dbcsr, &
                                      0.0_dp, mo_tmp_dbcsr, &
                                      retain_sparsity=.TRUE.)
               !CALL cp_dbcsr_verify_matrix(mo_tmp_dbcsr)
               checksum = cp_dbcsr_checksum(mo_tmp_dbcsr)
               occ = cp_dbcsr_get_occupation(mo_tmp_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'S*C occ', occ, ' checksum', checksum
               CALL cp_dbcsr_filter(mo_tmp_dbcsr, eps)
               !CALL cp_dbcsr_verify_matrix(mo_tmp_dbcsr)
               checksum = cp_dbcsr_checksum(mo_tmp_dbcsr)
               occ = cp_dbcsr_get_occupation(mo_tmp_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'S*C occ', occ, ' checksum', checksum
               !
               ! P*SC
               ! the destroy is needed for the moment to avoid memory leaks !
               ! This one is not needed because _destroy takes care of zeroing.
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, p_rmpv(ispin)%matrix, &
                                      mo_tmp_dbcsr, 0.0_dp, mo_dbcsr)
               IF (.FALSE.) CALL cp_dbcsr_verify_matrix(mo_dbcsr)
               checksum = cp_dbcsr_checksum(mo_dbcsr)
               occ = cp_dbcsr_get_occupation(mo_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'P*SC occ', occ, ' checksum', checksum
               CALL cp_dbcsr_filter(mo_dbcsr, eps)
               !CALL cp_dbcsr_verify_matrix(mo_dbcsr)
               checksum = cp_dbcsr_checksum(mo_dbcsr)
               occ = cp_dbcsr_get_occupation(mo_dbcsr)
               IF (output_unit > 0) WRITE (output_unit, *) 'P*SC occ', occ, ' checksum', checksum
               !
               CALL copy_dbcsr_to_fm(mo_dbcsr, mo_coeff)

               CALL cp_dbcsr_release(mo_dbcsr)
               CALL cp_dbcsr_release(mo_tmp_dbcsr)

               ! and ortho the result
               CALL make_basis_sm(mo_coeff, nmo, s_sparse(1)%matrix)
            END IF

            CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                   smear=qs_env%scf_control%smear)

            CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                  mo_array(ispin)%mo_set%mo_coeff_b) !fm->dbcsr

            CALL calculate_density_matrix(mo_array(ispin)%mo_set, &
                                          p_rmpv(ispin)%matrix)
            DO ikind = 1, SIZE(atomic_kind_set)
               IF (ASSOCIATED(pmat(ikind)%mat)) THEN
                  DEALLOCATE (pmat(ikind)%mat)
               END IF
            END DO
         END DO

         DEALLOCATE (pmat)

         DEALLOCATE (kind_of)

         DEALLOCATE (first_sgf, last_sgf)

         did_guess = .TRUE.
      END IF
      IF (density_guess == mopac_guess) THEN

         DO ispin = 1, nspin

            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, nelectron=nelectron)

            CALL calculate_mopac_dm(p_rmpv(ispin)%matrix, s_sparse(1)%matrix, has_unit_metric, dft_control, &
                                    particle_set, atomic_kind_set, qs_kind_set, nspin, nelectron, para_env)

            ! The orbital transformation method (OT) requires not only an
            ! initial density matrix, but also an initial wavefunction (MO set)
            IF (ASSOCIATED(scf_env%krylov_space)) do_std_diag = (scf_env%krylov_space%eps_std_diag > 0.0_dp)
            IF (scf_control%use_ot .OR. scf_env%method == ot_diag_method_nr .OR. &
                (scf_env%method == block_krylov_diag_method_nr .AND. .NOT. do_std_diag)) THEN
               IF (dft_control%restricted .AND. (ispin == 2)) THEN
                  CALL mo_set_restrict(mo_array)
               ELSE
                  CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                                  mo_coeff=mo_coeff, &
                                  nmo=nmo, homo=homo)
                  CALL cp_fm_init_random(mo_coeff, nmo)
                  CALL cp_fm_create(sv, mo_coeff%matrix_struct, "SV")
                  ! multiply times PS
                  IF (has_unit_metric) THEN
                     CALL cp_fm_to_fm(mo_coeff, sv)
                  ELSE
                     CALL cp_dbcsr_sm_fm_multiply(s_sparse(1)%matrix, mo_coeff, sv, nmo)
                  END IF
                  ! here we could easily multiply with the diag that we actually have replicated already
                  CALL cp_dbcsr_sm_fm_multiply(p_rmpv(ispin)%matrix, sv, mo_coeff, homo)
                  CALL cp_fm_release(sv)
                  ! and ortho the result
                  IF (has_unit_metric) THEN
                     CALL make_basis_simple(mo_coeff, nmo)
                  ELSE
                     CALL make_basis_sm(mo_coeff, nmo, s_sparse(1)%matrix)
                  END IF
               END IF

               CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                      smear=qs_env%scf_control%smear)
               CALL copy_fm_to_dbcsr(mo_array(ispin)%mo_set%mo_coeff, &
                                     mo_array(ispin)%mo_set%mo_coeff_b)

               CALL calculate_density_matrix(mo_array(ispin)%mo_set, &
                                             p_rmpv(ispin)%matrix)
            END IF
         END DO

         did_guess = .TRUE.
      END IF

      IF (density_guess == no_guess) THEN
         did_guess = .TRUE.
      END IF

      IF (.NOT. did_guess) THEN
         CALL cp__b("qs_initial_guess.F",913,"An invalid keyword for the initial density guess was specified")
      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_first_density_matrix

! **************************************************************************************************
!> \brief returns a block diagonal density matrix. Blocks correspond to the atomic densities.
!> \param matrix_p ...
!> \param matrix_s ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param ispin ...
!> \param nspin ...
!> \param nelectron_spin ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE calculate_atomic_block_dm(matrix_p, matrix_s, particle_set, atomic_kind_set, &
                                        qs_kind_set, ispin, nspin, nelectron_spin, output_unit)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p, matrix_s
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: ispin, nspin, nelectron_spin, output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atomic_block_dm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, handle, icol, ikind, irow, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      REAL(dp), DIMENSION(:, :), POINTER                 :: DATA
      REAL(KIND=dp)                                      :: scale, trps1
      TYPE(atom_matrix_type), ALLOCATABLE, DIMENSION(:)  :: pmat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      IF ((output_unit > 0) .AND. (nspin > 1)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,I0)") "Spin ", ispin
      END IF

      CALL cp_dbcsr_set(matrix_p, 0.0_dp)

      natom = SIZE(particle_set)

      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, kind_of=kind_of)

      ALLOCATE (pmat(SIZE(atomic_kind_set)))

      ! precompute the atomic blocks corresponding to spherical atoms
      DO ikind = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(ikind)
         qs_kind => qs_kind_set(ikind)
         NULLIFY (pmat(ikind)%mat)
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "Guess for atomic kind: "//TRIM(atomic_kind%name)
         END IF
         CALL calculate_atomic_orbitals(atomic_kind, qs_kind, iunit=output_unit, pmat=pmat(ikind)%mat, &
                                        ispin=ispin)
      END DO

      scale = 1.0_dp
      IF (nspin == 2) scale = 0.5_dp

      CALL cp_dbcsr_iterator_start(iter, matrix_p)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, irow, icol, DATA, blk)
         ikind = kind_of(irow)
         IF (icol .EQ. irow) DATA(:, :) = pmat(ikind)%mat(:, :)*scale
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_trace(matrix_p, matrix_s, trps1)
      scale = 0.0_dp
      IF (nelectron_spin > 0) & ! could be a ghost-atoms-only simulation
         scale = REAL(nelectron_spin, dp)/trps1
      CALL cp_dbcsr_scale(matrix_p, scale)

      IF (output_unit > 0) THEN
         IF (nspin > 1) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,I1)") &
               "Re-scaling the density matrix to get the right number of electrons for spin ", ispin
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "Re-scaling the density matrix to get the right number of electrons"
         END IF
         WRITE (output_unit, '(T19,A,T44,A,T67,A)') "# Electrons", "Trace(P)", "Scaling factor"
         WRITE (output_unit, '(T20,I10,T40,F12.3,T67,F14.3)') nelectron_spin, trps1, scale
      END IF

      DO ikind = 1, SIZE(atomic_kind_set)
         IF (ASSOCIATED(pmat(ikind)%mat)) THEN
            DEALLOCATE (pmat(ikind)%mat)
         END IF
      END DO

      DEALLOCATE (pmat)

      DEALLOCATE (kind_of)

      CALL timestop(handle)

   END SUBROUTINE calculate_atomic_block_dm

! **************************************************************************************************
!> \brief returns a block diagonal fock matrix.
!> \param matrix_f ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE calculate_atomic_fock_matrix(matrix_f, particle_set, atomic_kind_set, &
                                           qs_kind_set, output_unit)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_f
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atomic_fock_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icol, ikind, irow
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      REAL(dp), DIMENSION(:, :), POINTER                 :: block
      TYPE(atom_matrix_type), ALLOCATABLE, DIMENSION(:)  :: fmat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      ALLOCATE (kind_of(SIZE(particle_set)))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, kind_of=kind_of)
      ALLOCATE (fmat(SIZE(atomic_kind_set)))

      ! precompute the atomic blocks for each atomic-kind
      DO ikind = 1, SIZE(atomic_kind_set)
         atomic_kind => atomic_kind_set(ikind)
         qs_kind => qs_kind_set(ikind)
         NULLIFY (fmat(ikind)%mat)
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Calculating atomic Fock matrix for atomic kind: "//TRIM(atomic_kind%name)

         !Currently only ispin=1 is supported
         CALL calculate_atomic_orbitals(atomic_kind, qs_kind, iunit=output_unit, &
                                        fmat=fmat(ikind)%mat, ispin=1)
      END DO

      ! zero result matrix
      CALL cp_dbcsr_set(matrix_f, 0.0_dp)

      ! copy precomputed blocks onto diagonal of result matrix
      CALL cp_dbcsr_iterator_start(iter, matrix_f)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, irow, icol, block)
         ikind = kind_of(irow)
         IF (icol .EQ. irow) block(:, :) = fmat(ikind)%mat(:, :)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! cleanup
      DO ikind = 1, SIZE(atomic_kind_set)
         DEALLOCATE (fmat(ikind)%mat)
      END DO
      DEALLOCATE (fmat, kind_of)

      CALL timestop(handle)

   END SUBROUTINE calculate_atomic_fock_matrix

! **************************************************************************************************
!> \brief returns a block diagonal density matrix. Blocks correspond to the mopac initial guess.
!> \param matrix_p ...
!> \param matrix_s ...
!> \param has_unit_metric ...
!> \param dft_control ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param nspin ...
!> \param nelectron_spin ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE calculate_mopac_dm(matrix_p, matrix_s, has_unit_metric, &
                                 dft_control, particle_set, atomic_kind_set, qs_kind_set, &
                                 nspin, nelectron_spin, para_env)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p, matrix_s
      LOGICAL                                            :: has_unit_metric
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: nspin, nelectron_spin
      TYPE(cp_para_env_type)                             :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_mopac_dm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, group, handle, iatom, ikind, &
                                                            iset, isgf, isgfa, ishell, la, maxl, &
                                                            maxll, nao, natom, ncount, nset, nsgf, &
                                                            z
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, elec_conf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, l, last_sgfa
      REAL(KIND=dp)                                      :: maxocc, my_sum, nelec, paa, scale, &
                                                            trps1, trps2, yy, zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: econf, pdiag, sdiag
      REAL(KIND=dp), DIMENSION(0:3)                      :: edftb
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      CALL timeset(routineN, handle)

      group = para_env%group
      natom = SIZE(particle_set)
      nao = cp_dbcsr_nfullrows_total(matrix_p)
      IF (nspin == 1) THEN
         maxocc = 2.0_dp
      ELSE
         maxocc = 1.0_dp
      ENDIF

      ALLOCATE (first_sgf(natom))

      CALL get_particle_set(particle_set, qs_kind_set, first_sgf=first_sgf)
      CALL get_qs_kind_set(qs_kind_set, maxlgto=maxl)

      ALLOCATE (econf(0:maxl))

      ALLOCATE (pdiag(nao))
      pdiag(:) = 0.0_dp

      ALLOCATE (sdiag(nao))

      sdiag(:) = 0.0_dp
      IF (has_unit_metric) THEN
         sdiag(:) = 1.0_dp
      ELSE
         CALL cp_dbcsr_get_diag(matrix_s, sdiag)
         CALL mp_sum(sdiag, group)
      END IF

      ncount = 0
      trps1 = 0.0_dp
      trps2 = 0.0_dp
      pdiag(:) = 0.0_dp

      IF (nelectron_spin /= 0) THEN
         DO ikind = 1, SIZE(atomic_kind_set)

            CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, &
                             all_potential=all_potential, &
                             gth_potential=gth_potential)

            IF (dft_control%qs_control%dftb) THEN
               CALL get_dftb_atom_param(qs_kind_set(ikind)%dftb_parameter, &
                                        lmax=maxll, occupation=edftb)
               maxll = MIN(maxll, maxl)
               econf(0:maxl) = edftb(0:maxl)
            ELSEIF (ASSOCIATED(all_potential)) THEN
               CALL get_atomic_kind(atomic_kind_set(ikind), z=z)
               CALL get_qs_kind(qs_kind_set(ikind), nsgf=nsgf, elec_conf=elec_conf, zeff=zeff)
               maxll = MIN(SIZE(elec_conf)-1, maxl)
               econf(:) = 0.0_dp
               econf(0:maxll) = 0.5_dp*maxocc*REAL(elec_conf(0:maxll), dp)
            ELSE IF (ASSOCIATED(gth_potential)) THEN
               CALL get_atomic_kind(atomic_kind_set(ikind), z=z)
               CALL get_qs_kind(qs_kind_set(ikind), nsgf=nsgf, elec_conf=elec_conf, zeff=zeff)
               maxll = MIN(SIZE(elec_conf)-1, maxl)
               econf(:) = 0.0_dp
               econf(0:maxll) = 0.5_dp*maxocc*REAL(elec_conf(0:maxll), dp)
            ELSE
               CYCLE
            END IF

            ! MOPAC TYEP GUESS
            IF (dft_control%qs_control%dftb) THEN
               DO iatom = 1, natom
                  atom_a = atom_list(iatom)
                  isgfa = first_sgf(atom_a)
                  DO la = 0, maxll
                     SELECT CASE (la)
                     CASE (0)
                        pdiag(isgfa) = econf(0)
                     CASE (1)
                        pdiag(isgfa+1) = econf(1)/3._dp
                        pdiag(isgfa+2) = econf(1)/3._dp
                        pdiag(isgfa+3) = econf(1)/3._dp
                     CASE (2)
                        pdiag(isgfa+4) = econf(2)/5._dp
                        pdiag(isgfa+5) = econf(2)/5._dp
                        pdiag(isgfa+6) = econf(2)/5._dp
                        pdiag(isgfa+7) = econf(2)/5._dp
                        pdiag(isgfa+8) = econf(2)/5._dp
                     CASE (3)
                        pdiag(isgfa+9) = econf(3)/7._dp
                        pdiag(isgfa+10) = econf(3)/7._dp
                        pdiag(isgfa+11) = econf(3)/7._dp
                        pdiag(isgfa+12) = econf(3)/7._dp
                        pdiag(isgfa+13) = econf(3)/7._dp
                        pdiag(isgfa+14) = econf(3)/7._dp
                        pdiag(isgfa+15) = econf(3)/7._dp
                     CASE DEFAULT
                        CALL cp__b("qs_initial_guess.F",1228,"")
                     END SELECT
                  END DO
               END DO
            ELSEIF (dft_control%qs_control%semi_empirical) THEN
               yy = REAL(dft_control%charge, KIND=dp)/REAL(nao, KIND=dp)
               DO iatom = 1, natom
                  atom_a = atom_list(iatom)
                  isgfa = first_sgf(atom_a)
                  SELECT CASE (nsgf)
                  CASE (1) ! s-basis
                     pdiag(isgfa) = (zeff-yy)*0.5_dp*maxocc
                  CASE (4) ! sp-basis
                     IF (z == 1) THEN
                        ! special case: hydrogen with sp basis
                        pdiag(isgfa) = (zeff-yy)*0.5_dp*maxocc
                        pdiag(isgfa+1) = 0._dp
                        pdiag(isgfa+2) = 0._dp
                        pdiag(isgfa+3) = 0._dp
                     ELSE
                        pdiag(isgfa) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+1) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+2) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+3) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                     END IF
                  CASE (9) ! spd-basis
                     IF (z < 21 .OR. z > 30 .AND. z < 39 .OR. z > 48 .AND. z < 57) THEN
                        !   Main Group Element:  The "d" shell is formally empty.
                        pdiag(isgfa) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+1) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+2) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+3) = (zeff*0.25_dp-yy)*0.5_dp*maxocc
                        pdiag(isgfa+4) = (-yy)*0.5_dp*maxocc
                        pdiag(isgfa+5) = (-yy)*0.5_dp*maxocc
                        pdiag(isgfa+6) = (-yy)*0.5_dp*maxocc
                        pdiag(isgfa+7) = (-yy)*0.5_dp*maxocc
                        pdiag(isgfa+8) = (-yy)*0.5_dp*maxocc
                     ELSE IF (z < 99) THEN
                        my_sum = zeff-9.0_dp*yy
                        !   First, put 2 electrons in the 's' shell
                        pdiag(isgfa) = (MAX(0.0_dp, MIN(my_sum, 2.0_dp)))*0.5_dp*maxocc
                        my_sum = my_sum-2.0_dp
                        IF (my_sum > 0.0_dp) THEN
                           !   Now put as many electrons as possible into the 'd' shell
                           pdiag(isgfa+4) = (MAX(0.0_dp, MIN(my_sum*0.2_dp, 2.0_dp)))*0.5_dp*maxocc
                           pdiag(isgfa+5) = (MAX(0.0_dp, MIN(my_sum*0.2_dp, 2.0_dp)))*0.5_dp*maxocc
                           pdiag(isgfa+6) = (MAX(0.0_dp, MIN(my_sum*0.2_dp, 2.0_dp)))*0.5_dp*maxocc
                           pdiag(isgfa+7) = (MAX(0.0_dp, MIN(my_sum*0.2_dp, 2.0_dp)))*0.5_dp*maxocc
                           pdiag(isgfa+8) = (MAX(0.0_dp, MIN(my_sum*0.2_dp, 2.0_dp)))*0.5_dp*maxocc
                           my_sum = MAX(0.0_dp, my_sum-10.0_dp)
                           !   Put the remaining electrons in the 'p' shell
                           pdiag(isgfa+1) = (my_sum/3.0_dp)*0.5_dp*maxocc
                           pdiag(isgfa+2) = (my_sum/3.0_dp)*0.5_dp*maxocc
                           pdiag(isgfa+3) = (my_sum/3.0_dp)*0.5_dp*maxocc
                        END IF
                     END IF
                  CASE DEFAULT
                     CALL cp__b("qs_initial_guess.F",1285,"")
                  END SELECT
               END DO
            ELSE
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      nset=nset, &
                                      nshell=nshell, &
                                      l=l, &
                                      first_sgf=first_sgfa, &
                                      last_sgf=last_sgfa)

               DO iset = 1, nset
                  DO ishell = 1, nshell(iset)
                     la = l(ishell, iset)
                     nelec = maxocc*REAL(2*la+1, dp)
                     IF (econf(la) > 0.0_dp) THEN
                        IF (econf(la) >= nelec) THEN
                           paa = maxocc
                           econf(la) = econf(la)-nelec
                        ELSE
                           paa = maxocc*econf(la)/nelec
                           econf(la) = 0.0_dp
                           ncount = ncount+NINT(nelec/maxocc)
                        END IF
                        DO isgfa = first_sgfa(ishell, iset), last_sgfa(ishell, iset)
                           DO iatom = 1, natom
                              atom_a = atom_list(iatom)
                              isgf = first_sgf(atom_a)+isgfa-1
                              pdiag(isgf) = paa
                              IF (paa == maxocc) THEN
                                 trps1 = trps1+paa*sdiag(isgf)
                              ELSE
                                 trps2 = trps2+paa*sdiag(isgf)
                              END IF
                           END DO
                        END DO
                     END IF
                  END DO ! ishell
               END DO ! iset
            END IF
         END DO ! ikind

         IF (trps2 == 0.0_dp) THEN
            DO isgf = 1, nao
               IF (sdiag(isgf) > 0.0_dp) pdiag(isgf) = pdiag(isgf)/sdiag(isgf)
            END DO
         ELSE
            scale = (REAL(nelectron_spin, dp)-trps1)/trps2
            DO isgf = 1, nao
               IF (pdiag(isgf) < maxocc) pdiag(isgf) = scale*pdiag(isgf)
            END DO
         END IF
      END IF

      CALL cp_dbcsr_set_diag(matrix_p, pdiag)

      DEALLOCATE (econf)

      DEALLOCATE (first_sgf)

      DEALLOCATE (pdiag)

      DEALLOCATE (sdiag)

      CALL timestop(handle)

   END SUBROUTINE calculate_mopac_dm

END MODULE qs_initial_guess
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for a linear scaling quickstep SCF run based on the density
!>        matrix, with a focus on the interface between dm_ls_scf and qs
!> \par History
!>       2011.04 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE dm_ls_scf_qs
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_complete_redistribute, &
        cp_dbcsr_copy, cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_distribution_new, cp_dbcsr_finalize, cp_dbcsr_init, &
        cp_dbcsr_multiply, cp_dbcsr_name, cp_dbcsr_nblkrows_total, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_has_col_clusters, dbcsr_distribution_has_row_clusters, &
        dbcsr_distribution_init, dbcsr_distribution_mp, dbcsr_distribution_obj, &
        dbcsr_distribution_release, dbcsr_distribution_row_clusters, dbcsr_distribution_row_dist, &
        dbcsr_type_no_symmetry, dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_symmetric
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE dm_ls_scf_types,                 ONLY: ls_cluster_atomic,&
                                              ls_cluster_molecular,&
                                              ls_mstruct_type,&
                                              ls_scf_env_type
   USE input_constants,                 ONLY: ls_cluster_atomic,&
                                              ls_cluster_molecular,&
                                              ls_cluster_pao
   USE kinds,                           ONLY: dp
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_density_mixing_types,         ONLY: direct_mixing_nr,&
                                              gspace_mixing_nr
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gspace_mixing,                ONLY: gspace_mixing,&
                                              mixing_allocate,&
                                              mixing_init
   USE qs_initial_guess,                ONLY: calculate_atomic_block_dm,&
                                              calculate_mopac_dm
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_scf_qs'

   PUBLIC :: matrix_ls_create, matrix_qs_to_ls, matrix_ls_to_qs, ls_scf_init_qs, &
             ls_scf_dm_to_ks, ls_scf_qs_atomic_guess, write_matrix_to_cube, rho_mixing_ls_init

CONTAINS

! **************************************************************************************************
!> \brief create a matrix for use (and as a template) in ls based on a qs template
!> \param matrix_ls ...
!> \param matrix_qs ...
!> \param ls_mstruct ...
!> \par History
!>       2011.03 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE matrix_ls_create(matrix_ls, matrix_qs, ls_mstruct)
      TYPE(cp_dbcsr_type)                                :: matrix_ls, matrix_qs
      TYPE(ls_mstruct_type), INTENT(IN)                  :: ls_mstruct

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_ls_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle, iatom, imol, jatom, &
                                                            natom, nmol
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_to_cluster, atom_to_cluster_primus, &
                                                            primus_of_mol
      INTEGER, DIMENSION(:), POINTER :: blk_dist, blk_sizes, blk_sizes_aux, clustered_blk_sizes, &
         clustered_col_clusters, clustered_col_dist, clustered_row_clusters, clustered_row_dist
      TYPE(dbcsr_distribution_obj)                       :: block_distribution, &
                                                            clustered_distribution

      CALL timeset(routineN, handle)

      IF (ls_mstruct%single_precision) THEN
         data_type = dbcsr_type_real_4
      ELSE
         data_type = dbcsr_type_real_8
      ENDIF

      ! here we allocate our basic matrices, right now, use QS S as a template
      ! later, we might want to use our own format (e.g. clustered)
      SELECT CASE (ls_mstruct%cluster_type)
      CASE (ls_cluster_atomic)
         CALL cp_dbcsr_init(matrix_ls)
         CALL cp_dbcsr_create(matrix_ls, template=matrix_qs, &
                              data_type=data_type)
      CASE (ls_cluster_molecular)
         ! create format of the clustered matrix
         natom = cp_dbcsr_nblkrows_total(matrix_qs)
         nmol = MAXVAL(ls_mstruct%atom_to_molecule)
         ALLOCATE (atom_to_cluster_primus(natom))
         ALLOCATE (atom_to_cluster(natom))
         ALLOCATE (primus_of_mol(nmol))
         DO iatom = 1, natom
            atom_to_cluster(iatom) = ls_mstruct%atom_to_molecule(iatom)
            ! the first atom of the molecule is the primus
            ! if the number of atoms per molecule is independent of system size, this is not a quadratic loop
            ! it assumes that all atoms of the molecule are consecutive.
            DO jatom = iatom, 1, -1
               IF (ls_mstruct%atom_to_molecule(jatom) == atom_to_cluster(iatom)) THEN
                  atom_to_cluster_primus(iatom) = jatom
               ELSE
                  EXIT
               ENDIF
            ENDDO
            primus_of_mol(atom_to_cluster(iatom)) = atom_to_cluster_primus(iatom)
         ENDDO

         block_distribution = cp_dbcsr_distribution(matrix_qs)
         CALL dbcsr_distribution_init(clustered_distribution)
         ! row
         ALLOCATE (clustered_row_dist(nmol))
         blk_dist => dbcsr_distribution_row_dist(block_distribution)
         DO imol = 1, nmol
            clustered_row_dist(imol) = blk_dist(primus_of_mol(imol))
         ENDDO
         NULLIFY (clustered_row_clusters)
         IF (dbcsr_distribution_has_row_clusters(block_distribution)) THEN
            ALLOCATE (clustered_row_clusters(nmol))
            blk_dist => dbcsr_distribution_row_clusters(block_distribution)
            DO imol = 1, nmol
               clustered_row_clusters(imol) = blk_dist(primus_of_mol(imol))
            ENDDO
         ENDIF

         ! col
         ALLOCATE (clustered_col_dist(nmol))
         blk_dist => dbcsr_distribution_col_dist(block_distribution)
         DO imol = 1, nmol
            clustered_col_dist(imol) = blk_dist(primus_of_mol(imol))
         ENDDO
         NULLIFY (clustered_col_clusters)
         IF (dbcsr_distribution_has_col_clusters(block_distribution)) THEN
            ALLOCATE (clustered_col_clusters(nmol))
            blk_dist => dbcsr_distribution_col_clusters(block_distribution)
            DO imol = 1, nmol
               clustered_col_clusters(imol) = blk_dist(primus_of_mol(imol))
            ENDDO
         ENDIF

         blk_sizes => cp_dbcsr_row_block_sizes(matrix_qs)
         ALLOCATE (clustered_blk_sizes(nmol))
         clustered_blk_sizes = 0
         DO iatom = 1, natom
            clustered_blk_sizes(atom_to_cluster(iatom)) = clustered_blk_sizes(atom_to_cluster(iatom))+ &
                                                          blk_sizes(iatom)
         ENDDO

         ! the parallel distribution
         CALL cp_dbcsr_distribution_new(clustered_distribution, &
                                        dbcsr_distribution_mp(block_distribution), &
                                        clustered_row_dist, clustered_col_dist, &
                                        clustered_row_clusters, clustered_col_clusters, &
                                        reuse_arrays=.TRUE.)

         ! the matrix
         CALL cp_dbcsr_init(matrix_ls)
         CALL cp_dbcsr_create(matrix_ls, cp_dbcsr_name(matrix_qs), &
                              clustered_distribution, &
                              dbcsr_type_symmetric, &
                              clustered_blk_sizes, clustered_blk_sizes, &
                              data_type=data_type)
         DEALLOCATE (clustered_blk_sizes)
         CALL dbcsr_distribution_release(clustered_distribution)
         CALL cp_dbcsr_finalize(matrix_ls)
      CASE (ls_cluster_pao)
         blk_sizes_aux => cp_dbcsr_col_block_sizes(ls_mstruct%matrix_A)
         CALL cp_dbcsr_init(matrix_ls)
         CALL cp_dbcsr_create(matrix_ls, &
                              template=matrix_qs, &
                              data_type=data_type, &
                              row_blk_size=blk_sizes_aux, &
                              col_blk_size=blk_sizes_aux)
      CASE DEFAULT
         CALL cp__b("dm_ls_scf_qs.F",219," BUG cluster option")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE matrix_ls_create

! **************************************************************************************************
!> \brief first link to QS, copy a QS matrix to LS matrix
!>        used to isolate QS style matrices from LS style
!>        will be useful for future features (e.g. precision, symmetry, blocking, ...)
!> \param matrix_ls ...
!> \param matrix_qs ...
!> \param ls_mstruct ...
!> \param covariant ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE matrix_qs_to_ls(matrix_ls, matrix_qs, ls_mstruct, covariant)
      TYPE(cp_dbcsr_type)                                :: matrix_ls, matrix_qs
      TYPE(ls_mstruct_type), INTENT(IN), TARGET          :: ls_mstruct
      LOGICAL                                            :: covariant

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_qs_to_ls', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_qs_nosym, matrix_tmp
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_trafo

      CALL timeset(routineN, handle)

      SELECT CASE (ls_mstruct%cluster_type)
      CASE (ls_cluster_atomic)
         ! takes care of an eventual data_type conversion
         CALL cp_dbcsr_copy(matrix_ls, matrix_qs)
      CASE (ls_cluster_molecular)
         ! desymmetrize the qs matrix
         CALL cp_dbcsr_init(matrix_qs_nosym)
         CALL cp_dbcsr_create(matrix_qs_nosym, template=matrix_qs, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_desymmetrize(matrix_qs, matrix_qs_nosym)

         ! complete_redistribute does not zero blocks that might be present in the
         ! target but not in the source
         CALL cp_dbcsr_set(matrix_ls, 0.0_dp)

         ! perform the magic complete redistribute copy
         IF (ls_mstruct%single_precision) THEN
            CALL cp_dbcsr_init(matrix_tmp)
            CALL cp_dbcsr_create(matrix_tmp, template=matrix_qs_nosym, &
                                 data_type=dbcsr_type_real_4)
            CALL cp_dbcsr_copy(matrix_tmp, matrix_qs_nosym)
            CALL cp_dbcsr_complete_redistribute(matrix_tmp, matrix_ls); 
            CALL cp_dbcsr_release(matrix_tmp)
         ELSE
            CALL cp_dbcsr_complete_redistribute(matrix_qs_nosym, matrix_ls); 
         ENDIF
         CALL cp_dbcsr_release(matrix_qs_nosym)

      CASE (ls_cluster_pao)
         matrix_trafo => ls_mstruct%matrix_A !contra-variant
         IF (covariant) matrix_trafo => ls_mstruct%matrix_B !    co-variant
         CALL cp_dbcsr_init(matrix_tmp)
         CALL cp_dbcsr_create(matrix_tmp, template=matrix_trafo)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_qs, matrix_trafo, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, matrix_trafo, matrix_tmp, &
                                0.0_dp, matrix_ls)
         CALL cp_dbcsr_release(matrix_tmp)
      CASE DEFAULT
         CALL cp__b("dm_ls_scf_qs.F",291,"")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE matrix_qs_to_ls

! **************************************************************************************************
!> \brief second link to QS, copy a LS matrix to QS matrix
!>        used to isolate QS style matrices from LS style
!>        will be useful for future features (e.g. precision, symmetry, blocking, ...)
!> \param matrix_qs ...
!> \param matrix_ls ...
!> \param ls_mstruct ...
!> \param covariant ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE matrix_ls_to_qs(matrix_qs, matrix_ls, ls_mstruct, covariant)
      TYPE(cp_dbcsr_type)                                :: matrix_qs, matrix_ls
      TYPE(ls_mstruct_type), INTENT(IN), TARGET          :: ls_mstruct
      LOGICAL                                            :: covariant

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_ls_to_qs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_tmp, matrix_tmp1, matrix_tmp2
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_trafo

      CALL timeset(routineN, handle)

      IF (ls_mstruct%single_precision) THEN
         CALL cp_dbcsr_init(matrix_tmp)
         CALL cp_dbcsr_create(matrix_tmp, template=matrix_ls, &
                              data_type=dbcsr_type_real_8)
         CALL cp_dbcsr_copy(matrix_tmp, matrix_ls)
      ENDIF

      SELECT CASE (ls_mstruct%cluster_type)
      CASE (ls_cluster_atomic)
         IF (ls_mstruct%single_precision) THEN
            CALL cp_dbcsr_copy_into_existing(matrix_qs, matrix_tmp)
         ELSE
            CALL cp_dbcsr_copy_into_existing(matrix_qs, matrix_ls)
         ENDIF
      CASE (ls_cluster_molecular)
         CALL cp_dbcsr_set(matrix_qs, 0.0_dp)
         IF (ls_mstruct%single_precision) THEN
            CALL cp_dbcsr_complete_redistribute(matrix_tmp, matrix_qs, keep_sparsity=.TRUE.)
         ELSE
            CALL cp_dbcsr_complete_redistribute(matrix_ls, matrix_qs, keep_sparsity=.TRUE.)
         ENDIF
      CASE (ls_cluster_pao)
         matrix_trafo => ls_mstruct%matrix_B !contra-variant
         IF (covariant) matrix_trafo => ls_mstruct%matrix_A !    co-variant
         CALL cp_dbcsr_init(matrix_tmp1)
         CALL cp_dbcsr_create(matrix_tmp1, template=matrix_trafo)
         CALL cp_dbcsr_init(matrix_tmp2)
         CALL cp_dbcsr_create(matrix_tmp2, template=matrix_qs)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_trafo, matrix_ls, &
                                0.0_dp, matrix_tmp1)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp1, matrix_trafo, &
                                0.0_dp, matrix_tmp2)
         !use copy_into_existing to retain sparsity-pattern of S for sure
         CALL cp_dbcsr_copy_into_existing(matrix_qs, matrix_tmp2)
         CALL cp_dbcsr_release(matrix_tmp1)
         CALL cp_dbcsr_release(matrix_tmp2)
      CASE DEFAULT
         CALL cp__b("dm_ls_scf_qs.F",361,"BUG")
      END SELECT

      IF (ls_mstruct%single_precision) THEN
         CALL cp_dbcsr_release(matrix_tmp)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE matrix_ls_to_qs

! **************************************************************************************************
!> \brief further required initialization of QS.
!>        Might be factored-out since this seems common code with the other SCF.
!> \param qs_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_init_qs(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_init_qs', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (sab_orb)
      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! get basic quantities from the qs_env
      CALL get_qs_env(qs_env, dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      ks_env=ks_env, &
                      sab_orb=sab_orb)

      nspin = dft_control%nspins

      ! we might have to create matrix_ks
      IF (.NOT. ASSOCIATED(matrix_ks)) THEN
         CALL cp_dbcsr_allocate_matrix_set(matrix_ks, nspin)
         DO ispin = 1, nspin
            ALLOCATE (matrix_ks(ispin)%matrix)
            CALL cp_dbcsr_init(matrix_ks(ispin)%matrix)
            CALL cp_dbcsr_create(matrix_ks(ispin)%matrix, template=matrix_s(1)%matrix)
            CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks(ispin)%matrix, sab_orb)
            CALL cp_dbcsr_set(matrix_ks(ispin)%matrix, 0.0_dp)
         ENDDO
         CALL set_ks_env(ks_env, matrix_ks=matrix_ks)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_init_qs

! **************************************************************************************************
!> \brief get an atomic initial guess
!> \param qs_env ...
!> \param energy ...
!> \par History
!>       2012.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_qs_atomic_guess(qs_env, energy)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp)                                      :: energy

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_qs_atomic_guess', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      LOGICAL                                            :: has_unit_metric
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: qs_energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)
      NULLIFY (rho, rho_ao)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! get basic quantities from the qs_env
      CALL get_qs_env(qs_env, dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      ks_env=ks_env, &
                      energy=qs_energy, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      has_unit_metric=has_unit_metric, &
                      para_env=para_env, &
                      nelectron_spin=nelectron_spin, &
                      rho=rho)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      nspin = dft_control%nspins

      ! create an initial atomic guess
      DO ispin = 1, nspin
         CALL cp_dbcsr_set(rho_ao(ispin)%matrix, 0.0_dp)
         IF (dft_control%qs_control%dftb .OR. dft_control%qs_control%semi_empirical) THEN
            CALL calculate_mopac_dm(rho_ao(ispin)%matrix, matrix_s(1)%matrix, has_unit_metric, &
                                    dft_control, particle_set, atomic_kind_set, qs_kind_set, &
                                    nspin, nelectron_spin(ispin), &
                                    para_env)
         ELSE
            CALL calculate_atomic_block_dm(rho_ao(ispin)%matrix, matrix_s(1)%matrix, &
                                           particle_set, atomic_kind_set, qs_kind_set, &
                                           ispin, nspin, nelectron_spin(ispin), &
                                           unit_nr)
         ENDIF
      ENDDO

      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.FALSE.)

      energy = qs_energy%total

      CALL timestop(handle)

   END SUBROUTINE ls_scf_qs_atomic_guess

! **************************************************************************************************
!> \brief use the density matrix in ls_scf_env to compute the new energy and KS matrix
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param energy_new ...
!> \param iscf ...
!> \par History
!>       2011.04 created [Joost VandeVondele]
!>       2015.02 added gspace density mixing [Patrick Seewald]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_dm_to_ks(qs_env, ls_scf_env, energy_new, iscf)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      REAL(KIND=dp)                                      :: energy_new
      INTEGER, INTENT(IN)                                :: iscf

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_dm_to_ks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (energy, rho, rho_ao)
      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = ls_scf_env%nspins
      CALL get_qs_env(qs_env, para_env=para_env, energy=energy, rho=rho)
      CALL qs_rho_get(rho, rho_ao=rho_ao)

      ! set the new density matrix
      DO ispin = 1, nspin
         CALL matrix_ls_to_qs(rho_ao(ispin)%matrix, ls_scf_env%matrix_p(ispin), &
                              ls_scf_env%ls_mstruct, covariant=.FALSE.)
      END DO

      ! compute the corresponding KS matrix and new energy, mix density if requested
      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      IF (ls_scf_env%do_rho_mixing) THEN
         IF (ls_scf_env%density_mixing_method .EQ. direct_mixing_nr) &
            CALL cp__b("dm_ls_scf_qs.F",565,"Direct P mixing not implemented in linear scaling SCF. ")
         IF (ls_scf_env%density_mixing_method >= gspace_mixing_nr) THEN
            IF (iscf .GT. MAX(ls_scf_env%mixing_store%nskip_mixing, 1)) THEN
               CALL gspace_mixing(qs_env, ls_scf_env%density_mixing_method, &
                                  ls_scf_env%mixing_store, rho, para_env, &
                                  iscf-1)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, '(A57)') &
                     "*********************************************************"
                  WRITE (unit_nr, '(A13,F5.3,A20,A6,A7,I3)') &
                     " Using ALPHA=", ls_scf_env%mixing_store%alpha, &
                     " to mix rho: method=", ls_scf_env%mixing_store%iter_method, ", iscf=", iscf
                  WRITE (unit_nr, '(A8,F5.3,A6,F5.3,A8)') &
                     " rho_nw=", ls_scf_env%mixing_store%alpha, "*rho + ", &
                     1.0_dp-ls_scf_env%mixing_store%alpha, "*rho_old"
                  WRITE (unit_nr, '(A57)') &
                     "*********************************************************"
               ENDIF
            ENDIF
         ENDIF
      ENDIF

      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., &
                               just_energy=.FALSE., print_active=.TRUE.)
      energy_new = energy%total

      CALL timestop(handle)

   END SUBROUTINE ls_scf_dm_to_ks

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param matrix_p_ls ...
!> \param unit_nr ...
!> \param title ...
!> \param stride ...
! **************************************************************************************************
   SUBROUTINE write_matrix_to_cube(qs_env, ls_scf_env, matrix_p_ls, unit_nr, title, stride)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_p_ls
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(LEN=*), INTENT(IN)                       :: title
      INTEGER, DIMENSION(:), POINTER                     :: stride

      CHARACTER(len=*), PARAMETER :: routineN = 'write_matrix_to_cube', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: tot_rho
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_p_qs
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_subsys_type), POINTER                      :: subsys

      CALL timeset(routineN, handle)

      NULLIFY (ks_env, pw_env, auxbas_pw_pool, pw_pools, particles, subsys, matrix_ks)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      subsys=subsys, &
                      pw_env=pw_env, &
                      matrix_ks=matrix_ks)

      CALL qs_subsys_get(subsys, particles=particles)

      ! convert the density matrix (ls style) to QS style
      ALLOCATE (matrix_p_qs)
      CALL cp_dbcsr_init(matrix_p_qs) !initialization
      CALL cp_dbcsr_copy(matrix_p_qs, matrix_ks(1)%matrix)
      CALL cp_dbcsr_set(matrix_p_qs, 0.0_dp) !zero matrix creation
      CALL matrix_ls_to_qs(matrix_p_qs, matrix_p_ls, ls_scf_env%ls_mstruct, covariant=.FALSE.)

      ! Print total electronic density
      CALL pw_env_get(pw_env=pw_env, &
                      auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools)
      CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                             pw=wf_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_zero(wf_r%pw)
      CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                             pw=wf_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_zero(wf_g%pw)
      CALL calculate_rho_elec(matrix_p=matrix_p_qs, &
                              rho=wf_r, &
                              rho_gspace=wf_g, &
                              total_rho=tot_rho, &
                              ks_env=ks_env)

      ! write this to a cube
      CALL cp_pw_to_cube(wf_r%pw, unit_nr=unit_nr, title=title, &
                         particles=particles, stride=stride)

      !free memory
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)
      CALL cp_dbcsr_release(matrix_p_qs)
      DEALLOCATE (matrix_p_qs)

      CALL timestop(handle)

   END SUBROUTINE write_matrix_to_cube

! **************************************************************************************************
!> \brief Initialize g-space density mixing
!> \param qs_env ...
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE rho_mixing_ls_init(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rho_mixing_ls_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, dft_control=dft_control, &
                      rho=rho)

      CALL mixing_allocate(qs_env, ls_scf_env%density_mixing_method, nspins=ls_scf_env%nspins, &
                           mixing_store=ls_scf_env%mixing_store)
      IF (ls_scf_env%density_mixing_method >= gspace_mixing_nr) THEN
         IF (dft_control%qs_control%gapw) THEN
            CALL get_qs_env(qs_env, rho_atom_set=rho_atom)
            CALL mixing_init(ls_scf_env%density_mixing_method, rho, ls_scf_env%mixing_store, &
                             ls_scf_env%para_env, rho_atom=rho_atom)
         ELSE
            CALL mixing_init(ls_scf_env%density_mixing_method, rho, ls_scf_env%mixing_store, &
                             ls_scf_env%para_env)
         ENDIF
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE rho_mixing_ls_init

END MODULE dm_ls_scf_qs
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Factory routines for potentials used e.g. by pao_param_exp and pao_ml
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_potentials
   USE ai_overlap,                      ONLY: overlap_aab
   USE atom_kind_orbitals,              ONLY: calculate_atomic_orbitals
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: multinomial
   USE orbital_pointers,                ONLY: indco,&
                                              init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_potentials'

   PUBLIC :: pao_build_V_neighbors, pao_build_V_core

CONTAINS

! **************************************************************************************************
!> \brief Construct a potential which mimics the chemical environment created by neighboring atoms
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param iatom ...
!> \param block_V ...
! **************************************************************************************************
   SUBROUTINE pao_build_V_neighbors(particle_set, qs_kind_set, cell, iatom, block_V)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: iatom
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: block_V

      INTEGER                                            :: ikind, jatom, jkind, N, natoms
      REAL(dp), DIMENSION(3)                             :: Ra, Rab, Rb
      REAL(KIND=dp)                                      :: pot_beta, pot_weight
      TYPE(gto_basis_set_type), POINTER                  :: basis_set

      IF(.NOT.(.NOT. ALLOCATED(block_V)))CALL cp__a("pao_potentials.F",57)

      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set)
      N = basis_set%nsgf
      natoms = SIZE(particle_set)

      ALLOCATE (block_V(N, N))
      block_V = 0.0_dp
      DO jatom = 1, natoms
         IF (jatom == iatom) CYCLE
         CALL get_atomic_kind(particle_set(jatom)%atomic_kind, kind_number=jkind)
         CALL get_qs_kind(qs_kind_set(jkind), pao_pot_beta=pot_beta, pao_pot_weight=pot_weight)
         Ra = particle_set(iatom)%r
         Rb = particle_set(jatom)%r
         Rab = pbc(ra, rb, cell)
         CALL calc_gaussian(basis_set, block_V, Rab, lpot=0, beta=pot_beta, weight=pot_weight)
      ENDDO

   END SUBROUTINE pao_build_V_neighbors

! **************************************************************************************************
!> \brief Construct a potential based on the atomic core Hamiltonian
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param iatom ...
!> \param block_V ...
! **************************************************************************************************
   SUBROUTINE pao_build_V_core(particle_set, qs_kind_set, iatom, block_V)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: iatom
      REAL(dp), ALLOCATABLE, DIMENSION(:, :), TARGET     :: block_V

      INTEGER                                            :: ikind, N
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_V_ptr

      IF(.NOT.(.NOT. ALLOCATED(block_V)))CALL cp__a("pao_potentials.F",94)

      NULLIFY (block_V_ptr)
      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL calculate_atomic_orbitals(atomic_kind=particle_set(iatom)%atomic_kind, &
                                     qs_kind=qs_kind_set(ikind), &
                                     iunit=-1, &
                                     fmat=block_V_ptr, &
                                     ispin=1)

      ! turn pointer into allocatable
      N = SIZE(block_V_ptr, 1)
      ALLOCATE (block_V(N, N))
      block_V(:, :) = block_V_ptr
      DEALLOCATE (block_V_ptr)
   END SUBROUTINE pao_build_V_core

! **************************************************************************************************
!> \brief Calculates potential term of the form r**lpot * Exp(-beta*r**2)
!> \param basis_set ...
!> \param block_V potential term that is returned
!> \param Rab ...
!> \param lpot polynomial prefactor, r**lpot
!> \param beta exponent of the Gaussian
!> \param weight ...
! **************************************************************************************************
   SUBROUTINE calc_gaussian(basis_set, block_V, Rab, lpot, beta, weight)
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V
      REAL(dp), DIMENSION(3)                             :: Rab
      INTEGER, INTENT(IN)                                :: lpot
      REAL(dp), INTENT(IN)                               :: beta, weight

      CHARACTER(len=*), PARAMETER                        :: routineN = 'calc_gaussian'
      REAL(dp), PARAMETER                                :: inf_dist = 0.1*HUGE(1.0)

      INTEGER :: handle, ic, iset, ishell, jset, jshell, la1_max, la1_min, la2_max, la2_min, &
         lb_max, lb_min, N, na1, na2, nb, ncfga1, ncfga2, ncfgb, npgfa1, npgfa2, npgfb
      REAL(dp)                                           :: coeff
      REAL(dp), DIMENSION(:), POINTER                    :: rpgfa1, rpgfa2, rpgfb, zeta1, zeta2, zetb
      REAL(dp), DIMENSION(:, :), POINTER                 :: new_block, sab
      REAL(dp), DIMENSION(:, :, :), POINTER              :: saab

      CALL timeset(routineN, handle)

      IF(.NOT.(MOD(lpot, 2) == 0))CALL cp__a("pao_potentials.F",139) ! otherwise it's not rotationally invariant
      CALL init_orbital_pointers(lpot)

      N = SIZE(block_V, 1)
      ALLOCATE (new_block(N, N))
      new_block = 0.0_dp

      ! setup description of potential
      lb_min = lpot
      lb_max = lpot
      ncfgb = ncoset(lb_max)-ncoset(lb_min-1)
      npgfb = 1 ! number of exponents
      nb = npgfb*ncfgb

      ! initialize exponents, using R_vdw from neighbour atom
      ALLOCATE (rpgfb(npgfb), zetb(npgfb))
      rpgfb(1) = inf_dist ! range of functions, we dont want screening
      zetb(1) = beta

      ! loop over all set/shell combination and fill block_V
      DO iset = 1, basis_set%nset
      DO jset = 1, basis_set%nset
      DO ishell = 1, basis_set%nshell(iset)
      DO jshell = 1, basis_set%nshell(jset)

         ! setup iset
         la1_max = basis_set%l(ishell, iset)
         la1_min = basis_set%l(ishell, iset)
         npgfa1 = basis_set%npgf(iset)
         ncfga1 = ncoset(la1_max)-ncoset(la1_min-1)
         na1 = npgfa1*ncfga1
         zeta1 => basis_set%zet(:, iset)
         !rpgfa1  => basis_set%pgf_radius(:,iset) !TODO: reenable screening
         ALLOCATE (rpgfa1(npgfa1))
         rpgfa1(:) = inf_dist ! range of functions, we dont want screening

         ! setup jset
         la2_max = basis_set%l(jshell, jset)
         la2_min = basis_set%l(jshell, jset)
         npgfa2 = basis_set%npgf(jset)
         ncfga2 = ncoset(la2_max)-ncoset(la2_min-1)
         na2 = npgfa2*ncfga2
         zeta2 => basis_set%zet(:, jset)
         !rpgfa2  => basis_set%pgf_radius(:,jset) !TODO: reenable screening
         ALLOCATE (rpgfa2(npgfa2))
         rpgfa2(:) = inf_dist ! range of functions, we dont want screening

         ! calculate integrals
         ALLOCATE (saab(na1, na2, nb))
         saab = 0.0_dp
         CALL overlap_aab(la1_max=la1_max, la1_min=la1_min, npgfa1=npgfa1, rpgfa1=rpgfa1, zeta1=zeta1, &
                          la2_max=la2_max, la2_min=la2_min, npgfa2=npgfa2, rpgfa2=rpgfa2, zeta2=zeta2, &
                          lb_max=lb_max, lb_min=lb_min, npgfb=npgfb, rpgfb=rpgfb, zetb=zetb, &
                          rab=Rab, saab=saab)
         DEALLOCATE (rpgfa1, rpgfa2)

         ! sum potential terms: POW(x**2 + y**2 + z**2, lpot/2)
         ALLOCATE (sab(na1, na2))
         sab = 0.0_dp
         DO ic = 1, ncfgb
            coeff = multinomial(lpot/2, indco(:, ncoset(lpot-1)+ic)/2)
            sab = sab+coeff*saab(:, :, ic)
         ENDDO

         CALL my_contract(sab=sab, block_V=new_block, basis_set=basis_set, &
                          iset=iset, ishell=ishell, jset=jset, jshell=jshell)
         DEALLOCATE (sab, saab)
      ENDDO
      ENDDO
      ENDDO
      ENDDO

      DEALLOCATE (rpgfb, zetb)

      block_V = block_V+weight*new_block
      DEALLOCATE (new_block)

      ! symmetrize
      block_V = 0.5_dp*(block_V+TRANSPOSE(block_V))

      CALL timestop(handle)
   END SUBROUTINE calc_gaussian

! **************************************************************************************************
!> \brief Helper routine, contracts a basis block
!> \param sab ...
!> \param block_V ...
!> \param basis_set ...
!> \param iset ...
!> \param ishell ...
!> \param jset ...
!> \param jshell ...
! **************************************************************************************************
   SUBROUTINE my_contract(sab, block_V, basis_set, iset, ishell, jset, jshell)
      REAL(dp), DIMENSION(:, :), POINTER                 :: sab
      REAL(dp), DIMENSION(:, :), INTENT(OUT), TARGET     :: block_V
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      INTEGER, INTENT(IN)                                :: iset, ishell, jset, jshell

      INTEGER                                            :: a, b, c, d, ipgf, jpgf, l1, l2, n1, n2, &
                                                            nn1, nn2, sgfa1, sgfa2, sgla1, sgla2
      REAL(dp), DIMENSION(:, :), POINTER                 :: S, T1, T2, V

      ! first and last indices of given shell in block_V.
      ! This matrix is in the contracted spherical basis.
      sgfa1 = basis_set%first_sgf(ishell, iset)
      sgla1 = basis_set%last_sgf(ishell, iset)
      sgfa2 = basis_set%first_sgf(jshell, jset)
      sgla2 = basis_set%last_sgf(jshell, jset)

      ! prepare the result matrix
      V => block_V(sgfa1:sgla1, sgfa2:sgla2)

      ! Calculate strides of sphi matrix.
      ! This matrix is in the uncontraced cartesian basis.
      ! It contains all shells of the set.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configurations of *the entire set*. (0->lmax)
      nn1 = ncoset(basis_set%lmax(iset))
      nn2 = ncoset(basis_set%lmax(jset))

      ! Calculate strides of sab matrix
      ! This matrix is also in the uncontraced cartensian basis,
      ! however it contains only a single shell.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configrations of *the given shell*.
      l1 = basis_set%l(ishell, iset)
      l2 = basis_set%l(jshell, jset)
      n1 = ncoset(l1)-ncoset(l1-1)
      n2 = ncoset(l2)-ncoset(l2-1)

      DO ipgf = 1, basis_set%npgf(iset)
      DO jpgf = 1, basis_set%npgf(jset)
         ! prepare first trafo-matrix
         a = (ipgf-1)*nn1+ncoset(l1-1)+1
         T1 => basis_set%sphi(a:a+n1-1, sgfa1:sgla1)

         ! prepare second trafo-matrix
         b = (jpgf-1)*nn2+ncoset(l2-1)+1
         T2 => basis_set%sphi(b:b+n2-1, sgfa2:sgla2)

         ! prepare SAB matrix
         c = (ipgf-1)*n1+1
         d = (jpgf-1)*n2+1
         S => sab(c:c+n1-1, d:d+n2-1)

         ! do the transformation
         V = V+MATMUL(TRANSPOSE(T1), MATMUL(S, T2))
      ENDDO
      ENDDO

   END SUBROUTINE my_contract

END MODULE pao_potentials
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Original matrix exponential parametrization
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_param_exp
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_create, cp_dbcsr_get_block_p, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_reserve_diag_blocks, &
        cp_dbcsr_set, cp_dbcsr_type
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all
   USE pao_potentials,                  ONLY: pao_build_V_neighbors
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_initial_guess,                ONLY: calculate_atomic_fock_matrix
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_param_exp'

   PUBLIC :: pao_param_init_exp, pao_param_finalize_exp, pao_calc_U_exp
   PUBLIC :: pao_param_count_exp, pao_param_initguess_exp

CONTAINS

! **************************************************************************************************
!> \brief Initialize matrix exponential parametrization
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_init_exp(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_init_exp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, handle, i, iatom, N
      LOGICAL                                            :: found
      REAL(dp)                                           :: probe
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: block_V
      REAL(dp), DIMENSION(:), POINTER                    :: H_evals
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_H, block_H0, block_N, block_U0, &
                                                            H_evecs
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_dbcsr_type)                                :: matrix_H0
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      cell=cell, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      ! allocate matrix_U0
      CALL cp_dbcsr_init(pao%matrix_U0)
      CALL cp_dbcsr_create(pao%matrix_U0, template=matrix_s(1)%matrix, matrix_type="N")
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_U0)

      ! allocate matrix_H0
      CALL cp_dbcsr_init(matrix_H0)
      CALL cp_dbcsr_create(matrix_H0, template=matrix_s(1)%matrix)
      CALL cp_dbcsr_reserve_diag_blocks(matrix_H0)

      ! calculate inital atomic fock matrix H0
      ! Can't use matrix_ks from ls_scf_qs_atomic_guess(), it's not rotationally invariant
      ! getting H0 directly from the atomic code
      CALL calculate_atomic_fock_matrix(matrix_H0, &
                                        particle_set, &
                                        atomic_kind_set, &
                                        qs_kind_set, &
                                        output_unit=pao%iw)

      ! diagonalize each block of H0 and store eigenvectors in U0
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_U0)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_U0)
         iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_param_exp.F",101)
         CALL cp_dbcsr_get_block_p(matrix=matrix_H0, row=iatom, col=iatom, block=block_H0, found=found)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_N, row=iatom, col=iatom, block=block_N, found=found)
         IF(.NOT.(ASSOCIATED(block_H0) .AND. ASSOCIATED(block_N)))CALL cp__a("pao_param_exp.F",104)
         N = SIZE(block_U0, 1)

         ! build potential V
         CALL pao_build_V_neighbors(particle_set, qs_kind_set, cell, iatom, block_V)

         ! construct H
         ALLOCATE (block_H(N, N))
         block_H = MATMUL(MATMUL(block_N, block_H0+block_V), block_N) ! transform H0 into orthonormal basis

         ! diagonalize H
         ALLOCATE (H_evecs(N, N), H_evals(N))
         H_evecs = block_H
         CALL diamat_all(H_evecs, H_evals)

         ! ensure H_evals are non-degenerate
         DO i = 1, N-1
            IF (H_evals(i+1)-H_evals(i) < 1e-4_dp) &
               CALL cp__w("pao_param_exp.F",122,"min_gap of H too small")
         ENDDO

         ! initialize with rudimentary sign-convention (allows regtesting rotational invariance)
         DO i = 1, N
            probe = SUM(MATMUL(MATMUL(TRANSPOSE(H_evecs(:, 1:1)), block_H0), H_evecs(:, i:i)))
            block_U0(:, i) = SIGN(1.0_dp, probe)*H_evecs(:, i)
         ENDDO

         DEALLOCATE (block_H, H_evecs, H_evals, block_V)
      END DO
      CALL cp_dbcsr_iterator_stop(iter)
      CALL cp_dbcsr_release(matrix_H0)

      IF (pao%precondition) &
         CALL cp__b("pao_param_exp.F",137,"PAO preconditioning not supported for selected parametrization.")

      CALL timestop(handle)
   END SUBROUTINE pao_param_init_exp

! **************************************************************************************************
!> \brief Finalize exponential parametrization
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_finalize_exp(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CALL cp_dbcsr_release(pao%matrix_U0)

   END SUBROUTINE pao_param_finalize_exp

! **************************************************************************************************
!> \brief Returns the number of parameters for given atomic kind
!> \param qs_env ...
!> \param ikind ...
!> \param nparams ...
! **************************************************************************************************
   SUBROUTINE pao_param_count_exp(qs_env, ikind, nparams)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nparams

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_count_exp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cols, pao_basis_size, pri_basis_size, &
                                                            rows
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set)
      CALL get_qs_kind(qs_kind_set(ikind), &
                       basis_set=basis_set, &
                       pao_basis_size=pao_basis_size)
      pri_basis_size = basis_set%nsgf

      ! we only consider rotations between occupied and virtuals
      rows = pao_basis_size
      cols = pri_basis_size-pao_basis_size
      nparams = rows*cols

   END SUBROUTINE pao_param_count_exp

! **************************************************************************************************
!> \brief Fills matrix_X with an initial guess
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_initguess_exp(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_initguess_exp', &
         routineP = moduleN//':'//routineN

      CALL cp_dbcsr_set(pao%matrix_X, 0.0_dp) ! actual initial guess is matrix_U0

   END SUBROUTINE pao_param_initguess_exp

! **************************************************************************************************
!> \brief Calculate new matrix U and optionally its gradient G
!> \param pao ...
!> \param matrix_M ...
!> \param matrix_G ...
! **************************************************************************************************
   SUBROUTINE pao_calc_U_exp(pao, matrix_M, matrix_G)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(cp_dbcsr_type), OPTIONAL                      :: matrix_M, matrix_G

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_calc_U_exp', routineP = moduleN//':'//routineN

      COMPLEX(dp)                                        :: denom
      COMPLEX(dp), DIMENSION(:), POINTER                 :: evals
      COMPLEX(dp), DIMENSION(:, :), POINTER              :: block_D, evecs
      INTEGER                                            :: acol, arow, handle, i, iatom, j, k, M, &
                                                            N, nparams
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_G, block_G_full, block_M, &
                                                            block_tmp, block_U, block_U0, block_X, &
                                                            block_X_full, block_Y
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)
         IF (arow /= acol) CALL cp__b("pao_param_exp.F",227,"pao_calc_U_exp: encountered off-diagonal block")
         iatom = arow
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_U, row=iatom, col=iatom, block=block_U, found=found)
         IF(.NOT.(ASSOCIATED(block_U)))CALL cp__a("pao_param_exp.F",230)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_U0, row=iatom, col=iatom, block=block_U0, found=found)
         IF(.NOT.(ASSOCIATED(block_U0)))CALL cp__a("pao_param_exp.F",232)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_Y, row=iatom, col=iatom, block=block_Y, found=found)
         IF(.NOT.(ASSOCIATED(block_Y)))CALL cp__a("pao_param_exp.F",234)

         N = SIZE(block_Y, 1)
         M = SIZE(block_Y, 2)
         nparams = SIZE(block_X, 1)

         ! block_X stores only rotations between occupied and virtuals
         ! hence, we first have to build the full anti-symmetric exponent block
         ALLOCATE (block_X_full(N, N))
         block_X_full(:, :) = 0.0_dp
         DO i = 1, nparams
            block_X_full(MOD(i-1, M)+1, M+(i-1)/M+1) = +block_X(i, 1)
            block_X_full(M+(i-1)/M+1, MOD(i-1, M)+1) = -block_X(i, 1)
         ENDDO

         ! diagonalize block_X_full
         ALLOCATE (evals(N), evecs(N, N))
         CALL diag_antisym(block_X_full, evecs, evals)

         ! construct rotation matrix
         block_U(:, :) = 0.0_dp
         DO k = 1, N
            DO i = 1, N
               DO j = 1, N
                  block_U(i, j) = block_U(i, j)+REAL(EXP(evals(k))*evecs(i, k)*CONJG(evecs(j, k)), dp)
               ENDDO
            ENDDO
         ENDDO

         block_U = MATMUL(block_U0, block_U) ! prepend initial guess rotation

         ! TURNING POINT (if calc grad) ------------------------------------------
         IF (PRESENT(matrix_G)) THEN
            IF(.NOT.(PRESENT(matrix_M)))CALL cp__a("pao_param_exp.F",267)

            CALL cp_dbcsr_get_block_p(matrix=pao%matrix_G, row=iatom, col=iatom, block=block_G, found=found)
            IF(.NOT.(ASSOCIATED(block_G)))CALL cp__a("pao_param_exp.F",270)
            CALL cp_dbcsr_get_block_p(matrix=matrix_M, row=iatom, col=iatom, block=block_M, found=found)
            ! don't check ASSOCIATED(block_M), it might have been filtered out.

            ALLOCATE (block_D(N, N), block_tmp(N, N), block_G_full(N, N))
            DO i = 1, N
               DO j = 1, N
                  denom = evals(i)-evals(j)
                  IF (i == j) THEN
                     block_D(i, i) = EXP(evals(i)) ! diagonal elements
                  ELSE IF (ABS(denom) > 1e-10_dp) THEN
                     block_D(i, j) = (EXP(evals(i))-EXP(evals(j)))/denom
                  ELSE
                     block_D(i, j) = 1.0_dp ! limit according to L'Hospital's rule
                  ENDIF
               ENDDO
            ENDDO

            IF (ASSOCIATED(block_M)) THEN
               block_tmp = MATMUL(TRANSPOSE(block_U0), block_M)
            ELSE
               block_tmp = 0.0_dp
            ENDIF
            block_G_full = fold_derivatives(block_tmp, block_D, evecs)

            ! return only gradient for rotations between occupied and virtuals
            DO i = 1, nparams
               block_G(i, 1) = 2.0_dp*block_G_full(MOD(i-1, M)+1, M+(i-1)/M+1)
            ENDDO

            DEALLOCATE (block_D, block_tmp, block_G_full)
         ENDIF

         DEALLOCATE (block_X_full, evals, evecs)

      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL timestop(handle)
   END SUBROUTINE pao_calc_U_exp

! **************************************************************************************************
!> \brief Helper routine, for calculating derivatives
!> \param M ...
!> \param D ...
!> \param R ...
!> \retval G ...
! **************************************************************************************************
   FUNCTION fold_derivatives(M, D, R) RESULT(G)
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: M
      COMPLEX(dp), DIMENSION(:, :), INTENT(IN)           :: D, R
      REAL(dp), DIMENSION(SIZE(M, 1), SIZE(M, 1))        :: G

      CHARACTER(len=*), PARAMETER :: routineN = 'fold_derivatives', &
         routineP = moduleN//':'//routineN

      COMPLEX(dp), DIMENSION(:, :), POINTER              :: F, RF, RM, RMR
      INTEGER                                            :: n
      REAL(dp), DIMENSION(:, :), POINTER                 :: RFR

      n = SIZE(M, 1)

      ALLOCATE (RM(n, n), RMR(n, n), F(n, n), RF(n, n), RFR(n, n))

      RM = MATMUL(TRANSPOSE(CONJG(R)), TRANSPOSE(M))
      RMR = MATMUL(RM, R)
      F = RMR*D !Hadamard product
      RF = MATMUL(R, F)
      RFR = REAL(MATMUL(RF, TRANSPOSE(CONJG(R))))

      ! gradient dE/dX has to be anti-symmetric
      G = 0.5_dp*(TRANSPOSE(RFR)-RFR)

      DEALLOCATE (RM, RMR, F, RF, RFR)
   END FUNCTION fold_derivatives

! **************************************************************************************************
!> \brief Helper routine for diagonalizing anti symmetric matrices
!> \param matrix ...
!> \param evecs ...
!> \param evals ...
! **************************************************************************************************
   SUBROUTINE diag_antisym(matrix, evecs, evals)
      REAL(dp), DIMENSION(:, :)                          :: matrix
      COMPLEX(dp), DIMENSION(:, :)                       :: evecs
      COMPLEX(dp), DIMENSION(:)                          :: evals

      CHARACTER(len=*), PARAMETER :: routineN = 'diag_antisym', routineP = moduleN//':'//routineN

      COMPLEX(dp), DIMENSION(:, :), POINTER              :: matrix_c
      INTEGER                                            :: N
      REAL(dp), DIMENSION(:), POINTER                    :: evals_r

      IF (MAXVAL(ABS(matrix+TRANSPOSE(matrix))) > 1e-14_dp) CALL cp__b("pao_param_exp.F",363,"Expected anti-symmetric matrix")
      N = SIZE(matrix, 1)
      ALLOCATE (matrix_c(N, N), evals_r(N))

      matrix_c = CMPLX(0.0_dp, -matrix, kind=dp)
      CALL zheevd_wrapper(matrix_c, evecs, evals_r)
      evals = CMPLX(0.0_dp, evals_r, kind=dp)

      DEALLOCATE (matrix_c, evals_r)
   END SUBROUTINE diag_antisym

! **************************************************************************************************
!> \brief Helper routine for calling LAPACK zheevd
!> \param matrix ...
!> \param eigenvectors ...
!> \param eigenvalues ...
! **************************************************************************************************
   SUBROUTINE zheevd_wrapper(matrix, eigenvectors, eigenvalues)
      COMPLEX(dp), DIMENSION(:, :)                       :: matrix, eigenvectors
      REAL(dp), DIMENSION(:)                             :: eigenvalues

      CHARACTER(len=*), PARAMETER :: routineN = 'zheevd_wrapper', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: work
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: A
      INTEGER                                            :: handle, info, liwork, lrwork, lwork, n
      INTEGER, DIMENSION(:), POINTER                     :: iwork
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rwork

      CALL timeset(routineN, handle)

      IF (SIZE(matrix, 1) /= SIZE(matrix, 2)) CALL cp__b("pao_param_exp.F",394,"expected square matrix")
      IF (MAXVAL(ABS(matrix-CONJG(TRANSPOSE(matrix)))) > 1e-14_dp) CALL cp__b("pao_param_exp.F",395,"Expect hermitian matrix")

      n = SIZE(matrix, 1)
      ALLOCATE (iwork(1), rwork(1), work(1), A(n, n))

      A(:, :) = matrix ! ZHEEVD will overwrite A
      ! work space query
      lwork = -1
      lrwork = -1
      liwork = -1

      CALL ZHEEVD('V', 'U', n, A(1, 1), n, eigenvalues(1), &
                  work(1), lwork, rwork(1), lrwork, iwork(1), liwork, info)
      lwork = INT(REAL(work(1)))
      lrwork = INT(REAL(rwork(1)))
      liwork = iwork(1)

      DEALLOCATE (iwork, rwork, work)
      ALLOCATE (iwork(liwork))
      iwork(:) = 0
      ALLOCATE (rwork(lrwork))
      rwork(:) = 0.0_dp
      ALLOCATE (work(lwork))
      work(:) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)

      CALL ZHEEVD('V', 'U', n, A(1, 1), n, eigenvalues(1), &
                  work(1), lwork, rwork(1), lrwork, iwork(1), liwork, info)

      eigenvectors = A

      IF (info /= 0) CALL cp__b("pao_param_exp.F",425,"diagonalization failed")

      DEALLOCATE (iwork, rwork, work, A)

      CALL timestop(handle)

   END SUBROUTINE zheevd_wrapper

END MODULE pao_param_exp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Common framework for using eigenvectors of a Fock matrix as PAO basis.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_param_fock
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_create,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_reserve_diag_blocks
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_initial_guess,                ONLY: calculate_atomic_fock_matrix
   USE qs_kind_types,                   ONLY: qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_param_fock'

   PUBLIC :: pao_param_init_fock, pao_param_finalize_fock, pao_calc_U_block_fock

CONTAINS

! **************************************************************************************************
!> \brief Initialize fock parametrization
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_init_fock(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_init_fock', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      ! allocate matrix_H0
      CALL cp_dbcsr_init(pao%matrix_H0)
      CALL cp_dbcsr_create(pao%matrix_H0, template=matrix_s(1)%matrix)
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_H0)

      ! calculate inital atomic fock matrix H0
      ! Can't use matrix_ks from ls_scf_qs_atomic_guess(), it's not rotationally invariant
      ! getting H0 directly from the atomic code
      CALL calculate_atomic_fock_matrix(pao%matrix_H0, &
                                        particle_set, &
                                        atomic_kind_set, &
                                        qs_kind_set, &
                                        output_unit=pao%iw)

      CALL timestop(handle)
   END SUBROUTINE pao_param_init_fock

! **************************************************************************************************
!> \brief Finalize fock parametrization
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_finalize_fock(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_finalize_fock', &
         routineP = moduleN//':'//routineN

      CALL cp_dbcsr_release(pao%matrix_H0)

   END SUBROUTINE pao_param_finalize_fock

! **************************************************************************************************
!> \brief Calculate new matrix U and optinally its gradient G
!> \param pao ...
!> \param iatom ...
!> \param penalty ...
!> \param min_gap ...
!> \param V ...
!> \param U ...
!> \param M1 ...
!> \param G ...
! **************************************************************************************************
   SUBROUTINE pao_calc_U_block_fock(pao, iatom, penalty, min_gap, V, U, M1, G)
      TYPE(pao_env_type), POINTER                        :: pao
      INTEGER                                            :: iatom
      REAL(dp), INTENT(INOUT)                            :: penalty, min_gap
      REAL(dp), DIMENSION(:, :), POINTER                 :: V, U
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: M1, G

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_calc_U_block_fock', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, m, n
      LOGICAL                                            :: found
      REAL(dp)                                           :: alpha, beta, denom, diff, gap
      REAL(dp), DIMENSION(:), POINTER                    :: H_evals
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_N, D1, D2, H, H0, H_evecs, M2, M3, &
                                                            M4, M5, M6, M7, Y

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_block_p(matrix=pao%matrix_H0, row=iatom, col=iatom, block=H0, found=found)
      IF(.NOT.(ASSOCIATED(H0)))CALL cp__a("pao_param_fock.F",126)
      CALL cp_dbcsr_get_block_p(matrix=pao%matrix_N, row=iatom, col=iatom, block=block_N, found=found)
      IF(.NOT.(ASSOCIATED(block_N)))CALL cp__a("pao_param_fock.F",128)
      CALL cp_dbcsr_get_block_p(matrix=pao%matrix_Y, row=iatom, col=iatom, block=Y, found=found)
      IF(.NOT.(ASSOCIATED(Y)))CALL cp__a("pao_param_fock.F",130)

      IF (MAXVAL(ABS(V-TRANSPOSE(V))) > 1e-14_dp) CALL cp__b("pao_param_fock.F",132,"Expect symmetric matrix")
      n = SIZE(Y, 1) ! size of primary basis
      m = SIZE(Y, 2) ! size of pao basis

      ! calculate H in the orthonormal basis
      ALLOCATE (H(n, n))
      H = MATMUL(MATMUL(block_N, H0+V), block_N)

      ! diagonalize H
      ALLOCATE (H_evals(n), H_evecs(n, n))
      H_evecs = H
      CALL diamat_all(H_evecs, H_evals)

      ! the eigenvectors of H become the rotation matrix U
      U = H_evecs

      ! calculate homo-lumo gap (it's useful for detecting numerical issues)
      IF (m < n) THEN ! catch special case n==m
         gap = H_evals(m+1)-H_evals(m)
         min_gap = MIN(min_gap, gap)
      ENDIF

      ! penalty terms: occupied and virtual eigenvalues repel each other
      alpha = pao%penalty_strength
      beta = pao%penalty_dist
      DO i = 1, m
      DO j = m+1, n
         diff = H_evals(i)-H_evals(j)
         penalty = penalty+alpha*EXP(-(diff/beta)**2)
      ENDDO
      ENDDO

      IF (PRESENT(G)) THEN ! TURNING POINT (if calc grad) -------------------------

         IF(.NOT.(PRESENT(M1)))CALL cp__a("pao_param_fock.F",166)

         ! calculate derivatives between eigenvectors of H
         ALLOCATE (D1(n, n), M2(n, n), M3(n, n), M4(n, n))
         DO i = 1, n
         DO j = 1, n
            ! ignore changes among occupied or virtual eigenvectors
            ! They will get filtered out by M2*D1 anyways, however this early
            ! intervention might stabilize numerics in the case of level-crossings.
            IF (i <= m .EQV. j <= m) THEN
               D1(i, j) = 0.0_dp
            ELSE
               denom = H_evals(i)-H_evals(j)
               IF (ABS(denom) > 1e-9_dp) THEN ! avoid division by zero
                  D1(i, j) = 1.0_dp/denom
               ELSE
                  D1(i, j) = SIGN(1e+9_dp, denom)
               ENDIF
            ENDIF
         ENDDO
         ENDDO
         IF (ASSOCIATED(M1)) THEN
            M2 = MATMUL(TRANSPOSE(M1), H_evecs)
         ELSE
            M2 = 0.0_dp
         ENDIF
         M3 = M2*D1 ! Hadamard product
         M4 = MATMUL(MATMUL(H_evecs, M3), TRANSPOSE(H_evecs))

         ! gradient contribution from penalty terms
         ALLOCATE (D2(n, n), M5(n, n))
         D2 = 0.0_dp
         DO i = 1, n
         DO j = 1, n
            IF (i <= m .EQV. j <= m) CYCLE
            diff = H_evals(i)-H_evals(j)
            D2(i, i) = D2(i, i)-2.0_dp*alpha*diff/beta**2*EXP(-(diff/beta)**2)
         ENDDO
         ENDDO
         M5 = MATMUL(MATMUL(H_evecs, D2), TRANSPOSE(H_evecs))

         ! combine M1*dH/dH_evecs with dE_penalty/dH_evecs
         ALLOCATE (M6(n, n))
         M6 = M4+M5

         ! dH / dV, return to non-orthonormal basis
         ALLOCATE (M7(n, n))
         M7 = MATMUL(MATMUL(block_N, M6), block_N)

         ! symmetrize
         G = 0.5_dp*(M7+TRANSPOSE(M7)) ! the final gradient

         DEALLOCATE (D1, D2, M2, M3, M4, M5, M6, M7)
      ENDIF

      DEALLOCATE (H, H_evals, H_evecs)

      CALL timestop(handle)
   END SUBROUTINE pao_calc_U_block_fock

END MODULE pao_param_fock
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Parametrization based on GTH pseudo potentials
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_param_gth
   USE ai_overlap,                      ONLY: overlap_aab
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_arnoldi_extremal, cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_get_block_p, &
        cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_reserve_all_blocks, &
        cp_dbcsr_reserve_diag_blocks, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_distribution_mp, dbcsr_mp_group
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type
   USE iterate_matrix,                  ONLY: matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: gamma1
   USE mathlib,                         ONLY: multinomial
   USE message_passing,                 ONLY: mp_min,&
                                              mp_sum
   USE orbital_pointers,                ONLY: indco,&
                                              init_orbital_pointers,&
                                              ncoset
   USE pao_param_fock,                  ONLY: pao_calc_U_block_fock,&
                                              pao_param_finalize_fock,&
                                              pao_param_init_fock
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pao_param_init_gth, pao_param_finalize_gth, pao_calc_U_gth
   PUBLIC :: pao_param_count_gth, pao_param_initguess_gth

CONTAINS

! **************************************************************************************************
!> \brief Initialize the linear potential parametrization
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_init_gth(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_init_gth'

      INTEGER                                            :: acol, arow, handle, iatom, idx, ikind, &
                                                            iterm, jatom, n, natoms
      INTEGER, DIMENSION(:), POINTER                     :: blk_sizes_pri, col_blk_size, nterms, &
                                                            row_blk_size
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_V_term, vec_V_terms
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      CALL pao_param_init_fock(pao, qs_env)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      particle_set=particle_set, &
                      natom=natoms)

      ALLOCATE (row_blk_size(natoms), col_blk_size(natoms), nterms(natoms))
      DO iatom = 1, natoms
         CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
         CALL pao_param_count_gth(qs_env, ikind, nterms(iatom))
      ENDDO

      ! allocate matrix_V_terms
      blk_sizes_pri => cp_dbcsr_row_block_sizes(matrix_s(1)%matrix)
      col_blk_size = SUM(nterms)
      row_blk_size = blk_sizes_pri**2
      CALL cp_dbcsr_init(pao%matrix_V_terms)
      CALL cp_dbcsr_create(pao%matrix_V_terms, &
                           template=matrix_s(1)%matrix, &
                           matrix_type="N", &
                           row_blk_size=row_blk_size, &
                           col_blk_size=col_blk_size)
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_V_terms)
      CALL cp_dbcsr_set(pao%matrix_V_terms, 0.0_dp)

      ! calculate and store poential terms
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_V_terms)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, vec_V_terms)
         iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",104)
         n = blk_sizes_pri(iatom)
         DO jatom = 1, natoms
            IF (jatom == iatom) CYCLE ! waste some storage to simplify things later
            DO iterm = 1, nterms(jatom)
               idx = SUM(nterms(1:jatom-1))+iterm
               block_V_term(1:n, 1:n) => vec_V_terms(:, idx) ! map column into matrix
               CALL gth_calc_term(qs_env, block_V_term, iatom, jatom, iterm)
            ENDDO
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      IF (pao%precondition) &
         CALL pao_param_gth_preconditioner(pao, qs_env, nterms)

      DEALLOCATE (row_blk_size, col_blk_size, nterms)
      CALL timestop(handle)
   END SUBROUTINE pao_param_init_gth

! **************************************************************************************************
!> \brief Finalize the GTH potential parametrization
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_finalize_gth(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CALL pao_param_finalize_fock(pao)
      CALL cp_dbcsr_release(pao%matrix_V_terms)
      IF (pao%precondition) THEN
         CALL cp_dbcsr_release(pao%matrix_precon)
         CALL cp_dbcsr_release(pao%matrix_precon_inv)
      ENDIF

   END SUBROUTINE pao_param_finalize_gth

! **************************************************************************************************
!> \brief Builds the preconditioner matrix_precon and matrix_precon_inv
!> \param pao ...
!> \param qs_env ...
!> \param nterms ...
! **************************************************************************************************
   SUBROUTINE pao_param_gth_preconditioner(pao, qs_env, nterms)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, DIMENSION(:), POINTER                     :: nterms

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_gth_preconditioner'

      INTEGER                                            :: acol, arow, group, handle, i, iatom, &
                                                            ioffset, j, jatom, joffset, m, n, &
                                                            natoms
      LOGICAL                                            :: arnoldi_converged, converged, found
      REAL(dp)                                           :: eval_max, eval_min
      REAL(dp), DIMENSION(:, :), POINTER                 :: block, block_overlap, block_V_term
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: matrix_gth_overlap
      TYPE(ls_scf_env_type), POINTER                     :: ls_scf_env

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, ls_scf_env=ls_scf_env)
      group = dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(pao%matrix_V_terms)))
      natoms = SIZE(nterms)

      CALL cp_dbcsr_init(matrix_gth_overlap)
      CALL cp_dbcsr_create(matrix_gth_overlap, &
                           template=pao%matrix_V_terms, &
                           matrix_type="N", &
                           row_blk_size=nterms, &
                           col_blk_size=nterms)
      CALL cp_dbcsr_reserve_all_blocks(matrix_gth_overlap)
      CALL cp_dbcsr_set(matrix_gth_overlap, 0.0_dp)

      DO iatom = 1, natoms
      DO jatom = 1, natoms
         ioffset = SUM(nterms(1:iatom-1))
         joffset = SUM(nterms(1:jatom-1))
         n = nterms(iatom)
         m = nterms(jatom)

         ALLOCATE (block(n, m))
         block = 0.0_dp

         CALL cp_dbcsr_iterator_start(iter, pao%matrix_V_terms)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_V_term)
            IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",191)
            DO i = 1, n
            DO j = 1, m
               block(i, j) = block(i, j)+SUM(block_V_term(:, ioffset+i)*block_V_term(:, joffset+j))
            ENDDO
            ENDDO
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         CALL mp_sum(block, group)

         CALL cp_dbcsr_get_block_p(matrix=matrix_gth_overlap, row=iatom, col=jatom, block=block_overlap, found=found)
         IF (ASSOCIATED(block_overlap)) &
            block_overlap = block

         DEALLOCATE (block)
      ENDDO
      ENDDO

      !TODO: good setting for arnoldi?
      CALL cp_dbcsr_arnoldi_extremal(matrix_gth_overlap, eval_max, eval_min, max_iter=100, &
                                     threshold=1e-2_dp, converged=arnoldi_converged)
      IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| GTH-preconditioner converged, min, max, max/min:", &
         arnoldi_converged, eval_min, eval_max, eval_max/eval_min

      CALL cp_dbcsr_init(pao%matrix_precon)
      CALL cp_dbcsr_create(pao%matrix_precon, template=matrix_gth_overlap)
      CALL cp_dbcsr_init(pao%matrix_precon_inv)
      CALL cp_dbcsr_create(pao%matrix_precon_inv, template=matrix_gth_overlap)

      CALL matrix_sqrt_Newton_Schulz(pao%matrix_precon_inv, pao%matrix_precon, matrix_gth_overlap, &
                                     threshold=ls_scf_env%eps_filter, &
                                     order=ls_scf_env%sign_sqrt_order, &
                                     max_iter_lanczos=ls_scf_env%max_iter_lanczos, &
                                     eps_lanczos=ls_scf_env%eps_lanczos, &
                                     converged=converged)
      CALL cp_dbcsr_release(matrix_gth_overlap)

      IF (.NOT. converged) &
         CALL cp__b("pao_param_gth.F",230,"PAO: Sqrt of GTH-preconditioner did not converge.")

      CALL timestop(handle)
   END SUBROUTINE pao_param_gth_preconditioner

! **************************************************************************************************
!> \brief Calculate new matrix U and optinally its gradient G
!> \param pao ...
!> \param penalty ...
!> \param matrix_M1 ...
!> \param matrix_G ...
! **************************************************************************************************
   SUBROUTINE pao_calc_U_gth(pao, penalty, matrix_M1, matrix_G)
      TYPE(pao_env_type), POINTER                        :: pao
      REAL(dp), INTENT(INOUT)                            :: penalty
      TYPE(cp_dbcsr_type), OPTIONAL                      :: matrix_M1, matrix_G

      CHARACTER(len=*), PARAMETER                        :: routineN = 'pao_calc_U_gth'

      INTEGER                                            :: acol, arow, group, handle, iatom, idx, &
                                                            iterm, n, natoms
      INTEGER, DIMENSION(:), POINTER                     :: nterms
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:), POINTER                    :: world_G, world_X
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_G, block_M1, block_M2, block_U, &
                                                            block_V, block_V_term, block_X, &
                                                            vec_V_terms
      REAL(KIND=dp)                                      :: min_gap
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      min_gap = HUGE(dp)
      nterms => cp_dbcsr_row_block_sizes(pao%matrix_X)
      natoms = SIZE(nterms)
      group = dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(pao%matrix_X)))

      ! allocate arrays for world-view
      ALLOCATE (world_X(SUM(nterms)), world_G(SUM(nterms)))
      world_X = 0.0_dp; world_G = 0.0_dp

      ! collect world_X from atomic blocks
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)
         iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",275)
         idx = SUM(nterms(1:iatom-1))
         world_X(idx+1:idx+nterms(iatom)) = block_X(:, 1)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)
      CALL mp_sum(world_X, group) ! sync world view across MPI ranks

      ! loop over atoms
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_U)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_U)
         iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",286)
         n = SIZE(block_U, 1)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_V_terms, row=iatom, col=iatom, block=vec_V_terms, found=found)
         IF(.NOT.(ASSOCIATED(vec_V_terms)))CALL cp__a("pao_param_gth.F",289)

         ! calculate potential V of i'th atom
         ALLOCATE (block_V(n, n))
         block_V = 0.0_dp
         DO iterm = 1, SIZE(world_X)
            block_V_term(1:n, 1:n) => vec_V_terms(:, iterm) ! map column into matrix
            block_V = block_V+world_X(iterm)*block_V_term
         ENDDO

         ! calculate gradient block of i'th atom
         IF (.NOT. PRESENT(matrix_G)) THEN
            CALL pao_calc_U_block_fock(pao, iatom, penalty, min_gap, block_V, block_U)

         ELSE ! TURNING POINT (if calc grad) ------------------------------------
            IF(.NOT.(PRESENT(matrix_M1)))CALL cp__a("pao_param_gth.F",304)
            CALL cp_dbcsr_get_block_p(matrix=matrix_M1, row=iatom, col=iatom, block=block_M1, found=found)
            ALLOCATE (block_M2(n, n))
            CALL pao_calc_U_block_fock(pao, iatom, penalty, min_gap, block_V, block_U, block_M1, block_M2)
            DO iterm = 1, SIZE(world_G)
               block_V_term(1:n, 1:n) => vec_V_terms(:, iterm) ! map column into matrix
               world_G(iterm) = world_G(iterm)+SUM(block_V_term*block_M2)
            ENDDO
            DEALLOCATE (block_M2)
         ENDIF
         DEALLOCATE (block_V)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! distribute world_G across atomic blocks
      IF (PRESENT(matrix_G)) THEN
         CALL mp_sum(world_G, group) ! sync world view across MPI ranks
         CALL cp_dbcsr_iterator_start(iter, matrix_G)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_G)
            iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",324)
            idx = SUM(nterms(1:iatom-1))
            block_G(:, 1) = world_G(idx+1:idx+nterms(iatom))
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)
      ENDIF

      DEALLOCATE (world_X, world_G)

      ! sum penalty energies across ranks
      CALL mp_sum(penalty, group)

      ! print smallest homo-lumo gap encountered by fock-layer
      CALL mp_min(min_gap, group)
      IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| min_gap:", min_gap

      CALL timestop(handle)

   END SUBROUTINE pao_calc_U_gth

! **************************************************************************************************
!> \brief Returns the number of parameters for given atomic kind
!> \param qs_env ...
!> \param ikind ...
!> \param nparams ...
! **************************************************************************************************
   SUBROUTINE pao_param_count_gth(qs_env, ikind, nparams)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nparams

      INTEGER                                            :: max_projector, maxl, ncombis
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set)
      CALL get_qs_kind(qs_kind_set(ikind), pao_pot_maxl=maxl, pao_pot_max_projector=max_projector)

      IF (maxl < 0) &
         CALL cp__b("pao_param_gth.F",362,"GTH parametrization requires non-negative PAO_POTENTIAL%MAXL")

      IF (max_projector < 0) &
         CALL cp__b("pao_param_gth.F",365,"GTH parametrization requires non-negative PAO_POTENTIAL%MAX_PROJECTOR")

      IF (MOD(maxl, 2) /= 0) &
         CALL cp__b("pao_param_gth.F",368,"GTH parametrization requires even-numbered PAO_POTENTIAL%MAXL")

      ncombis = (max_projector+1)*(max_projector+2)/2
      nparams = ncombis*(maxl/2+1)

   END SUBROUTINE pao_param_count_gth

! **************************************************************************************************
!> \brief Fills the given block_V with the requested potential term
!> \param qs_env ...
!> \param block_V ...
!> \param iatom ...
!> \param jatom ...
!> \param kterm ...
! **************************************************************************************************
   SUBROUTINE gth_calc_term(qs_env, block_V, iatom, jatom, kterm)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V
      INTEGER, INTENT(IN)                                :: iatom, jatom, kterm

      INTEGER                                            :: c, jkind, lpot, max_shell, min_shell, &
                                                            pot_max_projector, pot_maxl
      REAL(KIND=dp)                                      :: beta, norm2, pot_beta, R_vdw
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, particle_set=particle_set, qs_kind_set=qs_kind_set)
      CALL get_atomic_kind(particle_set(jatom)%atomic_kind, rvdw=R_vdw, kind_number=jkind)
      CALL get_qs_kind(qs_kind_set(jkind), pao_pot_max_projector=pot_max_projector, pao_pot_maxl=pot_maxl)

      c = 0
      outer: &
         DO lpot = 0, pot_maxl, 2
      DO max_shell = 0, pot_max_projector
      DO min_shell = 0, max_shell
         c = c+1
         IF (c == kterm) EXIT outer
      ENDDO
      ENDDO
      ENDDO outer

      CALL get_qs_kind(qs_kind_set(jkind), pao_pot_beta=pot_beta)
      beta = 0.5_dp/(R_vdw*pot_beta)**2

      CALL calc_gaussian(qs_env, block_V, iatom, jatom, lpot=lpot, beta=beta, &
                         min_shell=min_shell, max_shell=max_shell)

      ! normalize
      norm2 = (2.0_dp*beta)**(-0.5_dp-lpot)*gamma1(lpot)
      block_V = block_V/SQRT(norm2)

   END SUBROUTINE gth_calc_term

! **************************************************************************************************
!> \brief Calculate intial guess for matrix_X
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_initguess_gth(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      INTEGER                                            :: acol, arow
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_X
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)
         IF(.NOT.(arow == acol))CALL cp__a("pao_param_gth.F",435)
         IF(.NOT.(SIZE(block_X, 2) == 1))CALL cp__a("pao_param_gth.F",436)

         ! a simplistic guess, which at least makes the atom visible to others
         block_X = 0.0_dp
         block_X(1, 1) = 0.01_dp
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

   END SUBROUTINE pao_param_initguess_gth

! **************************************************************************************************
!> \brief Calculates potential term of the form r**lpot * Exp(-beta*r**2)
!> \param qs_env ...
!> \param block_V potential term that is returned
!> \param iatom central atom
!> \param jatom remote atom on which the Gaussian is centered
!> \param lpot polynomial prefactor, r**lpot
!> \param beta exponent of the Gaussian
!> \param min_shell ...
!> \param max_shell ...
! **************************************************************************************************
   SUBROUTINE calc_gaussian(qs_env, block_V, iatom, jatom, lpot, beta, min_shell, max_shell)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V
      INTEGER, INTENT(IN)                                :: iatom, jatom, lpot
      REAL(dp), INTENT(IN)                               :: beta
      INTEGER, INTENT(IN)                                :: min_shell, max_shell

      CHARACTER(len=*), PARAMETER                        :: routineN = 'calc_gaussian'
      REAL(dp), PARAMETER                                :: inf_dist = 0.1*HUGE(1.0)

      INTEGER :: handle, ic, ikind, iset, ishell, jset, jshell, la1_max, la1_min, la2_max, &
         la2_min, lb_max, lb_min, na1, na2, natoms, nb, ncfga1, ncfga2, ncfgb, npgfa1, npgfa2, &
         npgfb
      REAL(dp), DIMENSION(:), POINTER                    :: rpgfa1, rpgfa2, zeta1, zeta2
      REAL(dp), DIMENSION(:, :), POINTER                 :: sab
      REAL(KIND=dp)                                      :: coeff
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rb
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: saab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)
      CALL init_orbital_pointers(lpot)

      CALL get_qs_env(qs_env, &
                      natom=natoms, &
                      cell=cell, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set)

      ! setup description of potential
      lb_min = lpot
      lb_max = lpot
      ncfgb = ncoset(lb_max)-ncoset(lb_min-1)
      npgfb = 1 ! number of exponents
      nb = npgfb*ncfgb

      ! initialize exponents, using R_vdw from neighbour atom
      ALLOCATE (rpgfb(npgfb), zetb(npgfb))
      rpgfb(1) = inf_dist ! range of functions, we dont want screening
      zetb(1) = beta

      ! distance vector A-B
      ra = particle_set(iatom)%r
      rb = particle_set(jatom)%r
      rab = pbc(ra, rb, cell)

      ! loop over all set/shell combination and fill block_V
      block_V(:, :) = 0.0_dp
      DO iset = 1, basis_set%nset
      DO jset = 1, basis_set%nset
      DO ishell = 1, basis_set%nshell(iset)
      DO jshell = 1, basis_set%nshell(jset)
         IF (MIN(basis_set%l(ishell, iset), basis_set%l(jshell, jset)) /= min_shell) CYCLE
         IF (MAX(basis_set%l(ishell, iset), basis_set%l(jshell, jset)) /= max_shell) CYCLE

         ! setup iset
         la1_max = basis_set%l(ishell, iset)
         la1_min = basis_set%l(ishell, iset)
         npgfa1 = basis_set%npgf(iset)
         ncfga1 = ncoset(la1_max)-ncoset(la1_min-1)
         na1 = npgfa1*ncfga1
         zeta1 => basis_set%zet(:, iset)
         !rpgfa1  => basis_set%pgf_radius(:,iset) !TODO: reenable screening
         ALLOCATE (rpgfa1(npgfa1))
         rpgfa1(:) = inf_dist ! range of functions, we dont want screening

         ! setup jset
         la2_max = basis_set%l(jshell, jset)
         la2_min = basis_set%l(jshell, jset)
         npgfa2 = basis_set%npgf(jset)
         ncfga2 = ncoset(la2_max)-ncoset(la2_min-1)
         na2 = npgfa2*ncfga2
         zeta2 => basis_set%zet(:, jset)
         !rpgfa2  => basis_set%pgf_radius(:,jset) !TODO: reenable screening
         ALLOCATE (rpgfa2(npgfa2))
         rpgfa2(:) = inf_dist ! range of functions, we dont want screening

         ! calculate integrals
         ALLOCATE (saab(na1, na2, nb))
         saab = 0.0_dp
         CALL overlap_aab(la1_max, la1_min, npgfa1, rpgfa1, zeta1, &
                          la2_max, la2_min, npgfa2, rpgfa2, zeta2, &
                          lb_max, lb_min, npgfb, rpgfb, zetb, &
                          rab=rab, saab=saab)
         DEALLOCATE (rpgfa1, rpgfa2)

         ! sum potential terms: POW(x**2 + y**2 + z**2, lpot/2)
         ALLOCATE (sab(na1, na2))
         sab = 0.0_dp
         DO ic = 1, ncfgb
            coeff = multinomial(lpot/2, indco(:, ncoset(lpot-1)+ic)/2)
            sab = sab+coeff*saab(:, :, ic)
         ENDDO

         CALL my_contract(sab, block_V, basis_set, iset, ishell, jset, jshell)
         DEALLOCATE (sab, saab)
      ENDDO
      ENDDO
      ENDDO
      ENDDO

      DEALLOCATE (rpgfb, zetb)

      ! symmetrize
      block_V = 0.5_dp*(block_V+TRANSPOSE(block_V))

      CALL timestop(handle)
   END SUBROUTINE calc_gaussian

! **************************************************************************************************
!> \brief Helper routine, contracts a basis block
!> \param sab ...
!> \param block_V ...
!> \param basis_set ...
!> \param iset ...
!> \param ishell ...
!> \param jset ...
!> \param jshell ...
! **************************************************************************************************
   SUBROUTINE my_contract(sab, block_V, basis_set, iset, ishell, jset, jshell)
      REAL(dp), DIMENSION(:, :), POINTER                 :: sab
      REAL(dp), DIMENSION(:, :), INTENT(OUT), TARGET     :: block_V
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      INTEGER, INTENT(IN)                                :: iset, ishell, jset, jshell

      INTEGER                                            :: a, b, c, d, ipgf, jpgf, l1, l2, n1, n2, &
                                                            nn1, nn2, sgfa1, sgfa2, sgla1, sgla2
      REAL(dp), DIMENSION(:, :), POINTER                 :: S, T1, T2, V

! first and last indices of given shell in block_V.
! This matrix is in the contracted spherical basis.

      sgfa1 = basis_set%first_sgf(ishell, iset)
      sgla1 = basis_set%last_sgf(ishell, iset)
      sgfa2 = basis_set%first_sgf(jshell, jset)
      sgla2 = basis_set%last_sgf(jshell, jset)

      ! prepare the result matrix
      V => block_V(sgfa1:sgla1, sgfa2:sgla2)

      ! Calculate strides of sphi matrix.
      ! This matrix is in the uncontraced cartesian basis.
      ! It contains all shells of the set.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configurations of *the entire set*. (0->lmax)
      nn1 = ncoset(basis_set%lmax(iset))
      nn2 = ncoset(basis_set%lmax(jset))

      ! Calculate strides of sab matrix
      ! This matrix is also in the uncontraced cartensian basis,
      ! however it contains only a single shell.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configrations of *the given shell*.
      l1 = basis_set%l(ishell, iset)
      l2 = basis_set%l(jshell, jset)
      n1 = ncoset(l1)-ncoset(l1-1)
      n2 = ncoset(l2)-ncoset(l2-1)

      DO ipgf = 1, basis_set%npgf(iset)
      DO jpgf = 1, basis_set%npgf(jset)
         ! prepare first trafo-matrix
         a = (ipgf-1)*nn1+ncoset(l1-1)+1
         T1 => basis_set%sphi(a:a+n1-1, sgfa1:sgla1)

         ! prepare second trafo-matrix
         b = (jpgf-1)*nn2+ncoset(l2-1)+1
         T2 => basis_set%sphi(b:b+n2-1, sgfa2:sgla2)

         ! prepare SAB matrix
         c = (ipgf-1)*n1+1
         d = (jpgf-1)*n2+1
         S => sab(c:c+n1-1, d:d+n2-1)

         ! do the transformation
         V = V+MATMUL(TRANSPOSE(T1), MATMUL(S, T2))
      ENDDO
      ENDDO

   END SUBROUTINE my_contract

END MODULE pao_param_gth
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rotationally invariant parametrization of Fock matrix.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_linpot_rotinv
   USE ai_overlap,                      ONLY: overlap_aab
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE orbital_transformation_matrices, ONLY: init_spherical_harmonics,&
                                              orbtramat
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE spherical_harmonics,             ONLY: y_lm
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_linpot_rotinv'

   PUBLIC :: linpot_rotinv_count_terms, linpot_rotinv_calc_term

CONTAINS

! **************************************************************************************************
!> \brief Count number of terms for given atomic kind
!> \param qs_env ...
!> \param ikind ...
!> \param nterms ...
! **************************************************************************************************
   SUBROUTINE linpot_rotinv_count_terms(qs_env, ikind, nterms)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nterms

      CHARACTER(len=*), PARAMETER :: routineN = 'linpot_rotinv_count_terms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dim_avail, iset, ishell, jneighbor, &
                                                            jset, jshell, li, lj, lpot, natoms, &
                                                            pot_maxl, pot_neighbors
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, &
                      natom=natoms, &
                      qs_kind_set=qs_kind_set)

      CALL get_qs_kind(qs_kind_set(ikind), &
                       pao_pot_maxl=pot_maxl, &
                       pao_pot_neighbors=pot_neighbors, &
                       basis_set=basis_set)

      pot_neighbors = MIN(pot_neighbors, natoms-1)

      nterms = 0
      DO iset = 1, basis_set%nset
      DO jset = 1, iset
         DO ishell = 1, basis_set%nshell(iset)
         DO jshell = 1, ishell
            li = basis_set%l(ishell, iset)
            lj = basis_set%l(jshell, jset)

            !calculate the block's number of dimensions
            !TODO: this seems to be too large in some cases
            dim_avail = (2*li+1)*(2*lj+1)

            inner: &
               DO lpot = 0, pot_maxl
               ! basis function can only detect potentials of lower order
               IF (lpot > MIN(li, lj)) CYCLE
               DO jneighbor = 1, pot_neighbors
                  nterms = nterms+1
                  ! don't generate more terms than available dimensions
                  dim_avail = dim_avail-1
                  IF (dim_avail == 0) EXIT inner
               ENDDO
            ENDDO inner
         ENDDO
         ENDDO
      ENDDO
      ENDDO

   END SUBROUTINE linpot_rotinv_count_terms

! **************************************************************************************************
!> \brief Count number of potential terms
!> \param qs_env ...
!> \param iatom ...
!> \param kterm ...
!> \param block_V ...
! **************************************************************************************************
   SUBROUTINE linpot_rotinv_calc_term(qs_env, iatom, kterm, block_V)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom, kterm
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V

      CHARACTER(len=*), PARAMETER :: routineN = 'linpot_rotinv_calc_term', &
         routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: inf_dist = 0.1*HUGE(1.0)

      INTEGER :: c, dim_avail, i, idx_i, idx_j, ikind, iset, ishell, jatom, jneighbor, jset, &
         jshell, la1_max, la1_min, la2_max, la2_min, lb_max, lb_min, li, lj, lpot, m, na1, na2, &
         natoms, nb, ncfga1, ncfga2, ncfgb, npgfa1, npgfa2, npgfb, nsgfa1, nsgfa2, pot_maxl, &
         pot_neighbors
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: neighbor_order
      REAL(dp)                                           :: coeff
      REAL(dp), DIMENSION(:), POINTER                    :: rpgfa1, rpgfa2, zeta1, zeta2
      REAL(dp), DIMENSION(:, :), POINTER                 :: sab
      REAL(KIND=dp)                                      :: c2s, pot_beta, pot_exp, R_vdw, y
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: neighbor_dist
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rab_dir, rb
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: saab
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, &
                      natom=natoms, &
                      cell=cell, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL get_qs_kind(qs_kind_set(ikind), &
                       pao_pot_maxl=pot_maxl, &
                       pao_pot_neighbors=pot_neighbors, &
                       pao_pot_beta=pot_beta, &
                       basis_set=basis_set)

      ! avoid overflows
      pot_neighbors = MIN(pot_neighbors, natoms-1)
      CALL init_orbital_pointers(2*pot_maxl)
      CALL init_spherical_harmonics(2*pot_maxl, output_unit=-1)

      c = 0 ! counter
      outer: &
         DO iset = 1, basis_set%nset
      DO jset = 1, iset
         DO ishell = 1, basis_set%nshell(iset)
         DO jshell = 1, ishell
            li = basis_set%l(ishell, iset)
            lj = basis_set%l(jshell, jset)
            dim_avail = (2*li+1)*(2*lj+1)

            inner: &
               DO lpot = 0, pot_maxl
               IF (lpot > MIN(li, lj)) CYCLE
               DO jneighbor = 1, pot_neighbors
                  c = c+1
                  IF (c == kterm) EXIT outer
                  dim_avail = dim_avail-1
                  IF (dim_avail == 0) EXIT inner
               ENDDO
            ENDDO inner
         ENDDO
         ENDDO
      ENDDO
      ENDDO outer

      !find neighbors
      !TODO: this is a quadratic algorithm, use a neighbor-list instead
      ALLOCATE (neighbor_dist(natoms), neighbor_order(natoms))
      DO jatom = 1, natoms
         ra = particle_set(iatom)%r
         rb = particle_set(jatom)%r
         rab = pbc(ra, rb, cell)
         neighbor_dist(jatom) = SQRT(SUM(rab*rab))
      ENDDO
      CALL sort(neighbor_dist, natoms, neighbor_order)
      IF(.NOT.(neighbor_order(1) == iatom))CALL cp__a("pao_linpot_rotinv.F",189) !central atom should be closesd to itself

      block_V(:, :) = 0.0_dp
      IF (jneighbor < 1) THEN
         ! figure out where to store result
         idx_i = basis_set%first_sgf(ishell, iset)
         idx_j = basis_set%first_sgf(jshell, jset)
         block_V(idx_i, idx_j) = 1.0_dp
      ELSE
         !map jneighbor to jatom
         jatom = neighbor_order(jneighbor+1)

         ! setup description of potential
         lb_min = lpot
         lb_max = lpot
         ncfgb = ncoset(lb_max)-ncoset(lb_min-1)
         npgfb = 1 ! number of exponents
         nb = npgfb*ncfgb

         ! initialize exponents, using R_vdw from neighbour atom
         ALLOCATE (rpgfb(npgfb), zetb(npgfb))
         rpgfb(1) = inf_dist ! range of functions, we dont want screening
         CALL get_atomic_kind(particle_set(jatom)%atomic_kind, rvdw=R_vdw)
         pot_exp = 0.5_dp/(R_vdw*pot_beta)**2
         !pot_exp = 0.5_dp / pot_beta**2
         zetb(1) = pot_exp !* 1.6_dp**(kpot-1)

         ! distance vector A-B
         ra = particle_set(iatom)%r
         rb = particle_set(jatom)%r
         rab = pbc(ra, rb, cell)

         ! setup iset
         la1_max = basis_set%l(ishell, iset)
         la1_min = basis_set%l(ishell, iset)
         npgfa1 = basis_set%npgf(iset)
         ncfga1 = ncoset(la1_max)-ncoset(la1_min-1)
         na1 = npgfa1*ncfga1
         zeta1 => basis_set%zet(:, iset)
         !rpgfa1  => basis_set%pgf_radius(:,iset) !TODO: reenable screening
         ALLOCATE (rpgfa1(npgfa1))
         rpgfa1(:) = inf_dist ! range of functions, we dont want screening

         ! setup jset
         la2_max = basis_set%l(jshell, jset)
         la2_min = basis_set%l(jshell, jset)
         npgfa2 = basis_set%npgf(jset)
         ncfga2 = ncoset(la2_max)-ncoset(la2_min-1)
         na2 = npgfa2*ncfga2
         zeta2 => basis_set%zet(:, jset)
         !rpgfa2  => basis_set%pgf_radius(:,jset) !TODO: reenable screening
         ALLOCATE (rpgfa2(npgfa2))
         rpgfa2(:) = inf_dist ! range of functions, we dont want screening

         ! allocate storage for integrals in cartesian basis
         ALLOCATE (saab(na1, na2, nb))
         saab = 0.0_dp

         CALL overlap_aab(la1_max, la1_min, npgfa1, rpgfa1, zeta1, &
                          la2_max, la2_min, npgfa2, rpgfa2, zeta2, &
                          lb_max, lb_min, npgfb, rpgfb, zetb, &
                          rab=rab, saab=saab)

         DEALLOCATE (rpgfb, zetb, rpgfa1, rpgfa2)

         ! number of spherical Gaussian functions in the set
         nsgfa1 = 2*basis_set%l(ishell, iset)+1
         nsgfa2 = 2*basis_set%l(jshell, jset)+1

         ! allocate storage for integrals in spherical basis
         ALLOCATE (sab(na1, na2))

         ! map zonal Harmonic in direction of atom_j onto carthesian Harmonics
         rab_dir = rab/SQRT(SUM(rab*rab))
         sab(:, :) = 0.0_dp
         DO i = 1, ncfgb
            coeff = 0.0_dp
            DO m = -lpot, +lpot
               CALL Y_lm(rab_dir, y, lpot, m)
               ! with this normalization the y**2 actually sum to 1.
               y = y*SQRT(fourpi/(2*lpot+1))

               ! Empirics showed that one has to use %slm instead of %c2s...
               ! presumably this get's the norms right.
               c2s = orbtramat(lpot)%slm(lpot+m+1, i)

               coeff = coeff+y*c2s
            ENDDO
            sab = sab+coeff*saab(:, :, i)
         ENDDO

         CALL my_contract(sab, block_V, basis_set, iset, ishell, jset, jshell)
         DEALLOCATE (saab, sab)
      ENDIF

      block_V = block_V+TRANSPOSE(block_V)

   END SUBROUTINE linpot_rotinv_calc_term

! **************************************************************************************************
!> \brief Helper routine, contracts a basis block
!> \param sab ...
!> \param block_V ...
!> \param basis_set ...
!> \param iset ...
!> \param ishell ...
!> \param jset ...
!> \param jshell ...
! **************************************************************************************************
   SUBROUTINE my_contract(sab, block_V, basis_set, iset, ishell, jset, jshell)
      REAL(dp), DIMENSION(:, :), POINTER                 :: sab
      REAL(dp), DIMENSION(:, :), INTENT(OUT), TARGET     :: block_V
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      INTEGER, INTENT(IN)                                :: iset, ishell, jset, jshell

      INTEGER                                            :: a, b, c, d, ipgf, jpgf, l1, l2, n1, n2, &
                                                            nn1, nn2, sgfa1, sgfa2, sgla1, sgla2
      REAL(dp), DIMENSION(:, :), POINTER                 :: S, T1, T2, V

! first and last indices of given shell in block_V.
! This matrix is in the contracted spherical basis.

      sgfa1 = basis_set%first_sgf(ishell, iset)
      sgla1 = basis_set%last_sgf(ishell, iset)
      sgfa2 = basis_set%first_sgf(jshell, jset)
      sgla2 = basis_set%last_sgf(jshell, jset)

      ! prepare the result matrix
      block_V(:, :) = 0.0_dp
      V => block_V(sgfa1:sgla1, sgfa2:sgla2)

      ! Calculate strides of sphi matrix.
      ! This matrix is in the uncontraced cartesian basis.
      ! It contains all shells of the set.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configurations of *the entire set*. (0->lmax)
      nn1 = ncoset(basis_set%lmax(iset))
      nn2 = ncoset(basis_set%lmax(jset))

      ! Calculate strides of sab matrix
      ! This matrix is also in the uncontraced cartensian basis,
      ! however it contains only a single shell.
      ! It's index runs over all primitive gaussians of the set
      ! and then for each gaussian over all configrations of *the given shell*.
      l1 = basis_set%l(ishell, iset)
      l2 = basis_set%l(jshell, jset)
      n1 = ncoset(l1)-ncoset(l1-1)
      n2 = ncoset(l2)-ncoset(l2-1)

      DO ipgf = 1, basis_set%npgf(iset)
      DO jpgf = 1, basis_set%npgf(jset)
         ! prepare first trafo-matrix
         a = (ipgf-1)*nn1+ncoset(l1-1)+1
         T1 => basis_set%sphi(a:a+n1-1, sgfa1:sgla1)

         ! prepare second trafo-matrix
         b = (jpgf-1)*nn2+ncoset(l2-1)+1
         T2 => basis_set%sphi(b:b+n2-1, sgfa2:sgla2)

         ! prepare SAB matrix
         c = (ipgf-1)*n1+1
         d = (jpgf-1)*n2+1
         S => sab(c:c+n1-1, d:d+n2-1)

         ! do the transformation
         V = V+MATMUL(TRANSPOSE(T1), MATMUL(S, T2))
      ENDDO
      ENDDO

   END SUBROUTINE my_contract

END MODULE pao_linpot_rotinv
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Full parametrization of Fock matrix, ie. the identity parametrization.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_linpot_full
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_linpot_full'

   PUBLIC :: linpot_full_count_terms, linpot_full_calc_term

CONTAINS

! **************************************************************************************************
!> \brief Count number of terms for given atomic kind
!> \param qs_env ...
!> \param ikind ...
!> \param nterms ...
! **************************************************************************************************
   SUBROUTINE linpot_full_count_terms(qs_env, ikind, nterms)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nterms

      CHARACTER(len=*), PARAMETER :: routineN = 'linpot_full_count_terms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set)
      CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set)
      n = basis_set%nsgf
      nterms = n+n*(n-1)/2

   END SUBROUTINE linpot_full_count_terms

! **************************************************************************************************
!> \brief Count number of linear terms in parametrization
!> \param kterm ...
!> \param block_V ...
! **************************************************************************************************
   SUBROUTINE linpot_full_calc_term(kterm, block_V)
      INTEGER, INTENT(IN)                                :: kterm
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V

      CHARACTER(len=*), PARAMETER :: routineN = 'linpot_full_calc_term', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, i, j, n

      block_V = 0.0
      n = SIZE(block_V, 1)
      IF (kterm > n+n*(n-1)/2) CALL cp__b("pao_linpot_full.F",70,"kterm out of bounds")

      a = 0
      outer: &
         DO i = 1, SIZE(block_V, 1)
      DO j = 1, SIZE(block_V, 2)
         IF (i >= j) a = a+1
         IF (a == kterm) EXIT outer
      ENDDO
      ENDDO outer

      block_V(i, j) = 1.0_dp
      block_V(j, i) = 1.0_dp
   END SUBROUTINE linpot_full_calc_term

END MODULE pao_linpot_full
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Common framework for a linear parametrization of the potential.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_param_linpot
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_col_block_sizes, cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_frobenius_norm, &
        cp_dbcsr_get_block_p, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_reserve_diag_blocks, cp_dbcsr_type, &
        dbcsr_distribution_mp, dbcsr_mp_group
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_min,&
                                              mp_sum
   USE pao_input,                       ONLY: pao_fock_param,&
                                              pao_rotinv_param
   USE pao_linpot_full,                 ONLY: linpot_full_calc_term,&
                                              linpot_full_count_terms
   USE pao_linpot_rotinv,               ONLY: linpot_rotinv_calc_term,&
                                              linpot_rotinv_count_terms
   USE pao_param_fock,                  ONLY: pao_calc_U_block_fock,&
                                              pao_param_finalize_fock,&
                                              pao_param_init_fock
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pao_param_init_linpot, pao_param_finalize_linpot, pao_calc_U_linpot
   PUBLIC :: pao_param_count_linpot, pao_param_initguess_linpot

CONTAINS

! **************************************************************************************************
!> \brief Initialize the linear potential parametrization
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_init_linpot(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_init_linpot'

      INTEGER                                            :: acol, arow, handle, iatom, ikind, kterm, &
                                                            n, natoms, nterms
      INTEGER, DIMENSION(:), POINTER                     :: blk_sizes_pri, col_blk_size, row_blk_size
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_V_term, block_V_terms
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      CALL pao_param_init_fock(pao, qs_env)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      natom=natoms)

      IF (dft_control%nspins /= 1) CALL cp__b("pao_param_linpot.F",75,"open shell not yet implemented")

      ! figure out number of potential terms
      ALLOCATE (row_blk_size(natoms), col_blk_size(natoms))
      DO iatom = 1, natoms
         CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
         CALL pao_param_count_linpot(pao, qs_env, ikind, nterms)
         col_blk_size(iatom) = nterms
      ENDDO

      ! allocate matrix_V_terms
      blk_sizes_pri => cp_dbcsr_col_block_sizes(matrix_s(1)%matrix)
      row_blk_size = blk_sizes_pri**2
      CALL cp_dbcsr_init(pao%matrix_V_terms)
      CALL cp_dbcsr_create(pao%matrix_V_terms, &
                           template=matrix_s(1)%matrix, &
                           matrix_type="N", &
                           row_blk_size=row_blk_size, &
                           col_blk_size=col_blk_size)
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_V_terms)
      DEALLOCATE (row_blk_size, col_blk_size)

      ! calculate, normalize, and store potential terms as rows of block_V_terms
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_V_terms)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_V_terms)
         iatom = arow
         n = blk_sizes_pri(iatom)
         IF(.NOT.(n*n == SIZE(block_V_terms, 1)))CALL cp__a("pao_param_linpot.F",103)
         nterms = SIZE(block_V_terms, 2)
         DO kterm = 1, nterms
            block_V_term(1:n, 1:n) => block_V_terms(:, kterm) ! map column into matrix
            CALL linpot_calc_term(pao, qs_env, iatom, kterm, block_V_term)
            block_V_term = block_V_term/SQRT(SUM(block_V_term*block_V_term)) ! normalize
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      IF (pao%precondition) &
         CALL cp__b("pao_param_linpot.F",114,"PAO preconditioning not supported for selected parametrization.")

      CALL timestop(handle)
   END SUBROUTINE pao_param_init_linpot

! **************************************************************************************************
!> \brief Finalize the linear potential parametrization
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_finalize_linpot(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CALL pao_param_finalize_fock(pao)
      CALL cp_dbcsr_release(pao%matrix_V_terms)

   END SUBROUTINE pao_param_finalize_linpot

! **************************************************************************************************
!> \brief Returns the number of potential terms for given atomic kind
!> \param pao ...
!> \param qs_env ...
!> \param ikind ...
!> \param nparams ...
! **************************************************************************************************
   SUBROUTINE pao_param_count_linpot(pao, qs_env, ikind, nparams)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nparams

      SELECT CASE (pao%parameterization)
      CASE (pao_fock_param)
         CALL linpot_full_count_terms(qs_env, ikind, nterms=nparams)
      CASE (pao_rotinv_param)
         CALL linpot_rotinv_count_terms(qs_env, ikind, nterms=nparams)
      CASE DEFAULT
         CALL cp__b("pao_param_linpot.F",150,"unkown parameterization")
      END SELECT

   END SUBROUTINE pao_param_count_linpot

! **************************************************************************************************
!> \brief Calculate new matrix U and optinally its gradient G
!> \param pao ...
!> \param penalty ...
!> \param matrix_M ...
!> \param matrix_G ...
! **************************************************************************************************
   SUBROUTINE pao_calc_U_linpot(pao, penalty, matrix_M, matrix_G)
      TYPE(pao_env_type), POINTER                        :: pao
      REAL(dp), INTENT(INOUT)                            :: penalty
      TYPE(cp_dbcsr_type), OPTIONAL                      :: matrix_M, matrix_G

      CHARACTER(len=*), PARAMETER                        :: routineN = 'pao_calc_U_linpot'

      INTEGER                                            :: acol, arow, group, handle, iatom, n
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:), POINTER                    :: vec_M2, vec_V
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_G, block_M1, block_M2, block_U, &
                                                            block_V, block_V_terms, block_X
      REAL(KIND=dp)                                      :: min_gap
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      min_gap = HUGE(dp)
      group = dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(pao%matrix_U)))

      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)
         IF (arow /= acol) CALL cp__b("pao_param_linpot.F",185,"pao_calc_U_linpot: encountered off-diagonal block")
         iatom = arow
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_U, row=iatom, col=iatom, block=block_U, found=found)
         IF(.NOT.(ASSOCIATED(block_U)))CALL cp__a("pao_param_linpot.F",188)
         n = SIZE(block_U, 1)

         ! calculate potential V
         ALLOCATE (vec_V(n*n))
         vec_V(:) = 0.0_dp
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_V_terms, row=iatom, col=iatom, block=block_V_terms, found=found)
         IF (SIZE(block_V_terms) > 0) & ! protect against corner-case of zero pao parameters
            vec_V = MATMUL(block_V_terms, block_X(:, 1))
         block_V(1:n, 1:n) => vec_V(:) ! map vector into matrix

         ! symmetrize
         IF (MAXVAL(ABS(block_V-TRANSPOSE(block_V))/MAX(1.0_dp, MAXVAL(ABS(block_V)))) > 1e-12) &
            CALL cp__b("pao_param_linpot.F",201,"block_V not symmetric")
         block_V = 0.5_dp*(block_V+TRANSPOSE(block_V)) ! symmetrize exactly

         IF (.NOT. PRESENT(matrix_G)) THEN
            CALL pao_calc_U_block_fock(pao, iatom, penalty, min_gap, block_V, block_U)

         ELSE ! TURNING POINT (if calc grad) ------------------------------------
            IF(.NOT.(PRESENT(matrix_M)))CALL cp__a("pao_param_linpot.F",208)
            CALL cp_dbcsr_get_block_p(matrix=pao%matrix_G, row=iatom, col=iatom, block=block_G, found=found)
            IF(.NOT.(ASSOCIATED(block_G)))CALL cp__a("pao_param_linpot.F",210)
            CALL cp_dbcsr_get_block_p(matrix=matrix_M, row=iatom, col=iatom, block=block_M1, found=found)
            ! don't check ASSOCIATED(block_M1), it might have been filtered out.

            ALLOCATE (vec_M2(n*n))
            block_M2(1:n, 1:n) => vec_M2(:) ! map vector into matrix
            CALL pao_calc_U_block_fock(pao, iatom, penalty, min_gap, block_V, block_U, block_M1, block_M2)
            IF (MAXVAL(ABS(block_M2-TRANSPOSE(block_M2))) > 1e-14_dp) CALL cp__b("pao_param_linpot.F",217,"Expect symmetric matrix")
            IF (SIZE(block_V_terms) > 0) & ! protect against corner-case of zero pao parameters
               block_G(:, 1) = MATMUL(vec_M2, block_V_terms)
            DEALLOCATE (vec_M2)
         ENDIF

         DEALLOCATE (vec_V)

      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      ! sum penalty energies across ranks
      CALL mp_sum(penalty, group)

      ! print smallest homo-lumo gap encountered by fock-layer
      CALL mp_min(min_gap, group)
      IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| min_gap:", min_gap

      CALL timestop(handle)

   END SUBROUTINE pao_calc_U_linpot

! **************************************************************************************************
!> \brief Internal routine, calculates terms in potential parametrization
!> \param pao ...
!> \param qs_env ...
!> \param iatom ...
!> \param kterm ...
!> \param block_V ...
! **************************************************************************************************
   SUBROUTINE linpot_calc_term(pao, qs_env, iatom, kterm, block_V)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom, kterm
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: block_V

      SELECT CASE (pao%parameterization)
      CASE (pao_fock_param)
         CALL linpot_full_calc_term(kterm, block_V)
      CASE (pao_rotinv_param)
         CALL linpot_rotinv_calc_term(qs_env, iatom, kterm, block_V)
      CASE DEFAULT
         CALL cp__b("pao_param_linpot.F",259,"unkown parameterization")
      END SELECT

   END SUBROUTINE linpot_calc_term

! **************************************************************************************************
!> \brief Calculate intial guess for matrix_X
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_initguess_linpot(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_initguess_linpot'

      INTEGER                                            :: acol, arow, handle, n
      LOGICAL                                            :: found
      REAL(dp)                                           :: norm
      REAL(dp), DIMENSION(:), POINTER                    :: V_guess_vec
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_H0, block_ks, block_V_terms, &
                                                            block_X, V_guess
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, matrix_ks=matrix_ks)

      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)

         ! retrieve potential terms
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_V_terms, row=arow, col=acol, block=block_V_terms, found=found)
         IF(.NOT.(ASSOCIATED(block_V_terms)))CALL cp__a("pao_param_linpot.F",294)
         IF (SIZE(block_V_terms) == 0) CYCLE ! protect against corner-case of zero pao parameters

         ! retrieve block_ks and block_H0
         CALL cp_dbcsr_get_block_p(matrix=matrix_ks(1)%matrix, row=arow, col=acol, block=block_ks, found=found)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_H0, row=arow, col=acol, block=block_H0, found=found)
         IF(.NOT.(ASSOCIATED(block_H0) .AND. ASSOCIATED(block_ks)))CALL cp__a("pao_param_linpot.F",300)

         ! guess a potential
         n = SIZE(block_ks, 1)
         ALLOCATE (V_guess_vec(n*n))
         V_guess(1:n, 1:n) => V_guess_vec
         V_guess = block_ks-block_H0

         ! project out anything that the potential terms can not represent
         V_guess_vec = MATMUL(MATMUL(block_V_terms, TRANSPOSE(block_V_terms)), V_guess_vec)

         ! find least-square-fit of guessed potential
         CALL dgelsy_wrapper(block_V_terms, V_guess_vec, block_X(:, 1))

         DEALLOCATE (V_guess_vec)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      norm = cp_dbcsr_frobenius_norm(pao%matrix_X)
      IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| Calculated initial guess, norm(X):", norm

      CALL timestop(handle)
   END SUBROUTINE pao_param_initguess_linpot

! **************************************************************************************************
!> \brief Helper routine for calling LAPACK dgelsy
!>  computes the linear least square fit min|| A * X - B ||
!> \param matrix_A possibly non-square matrix
!> \param vec_B ...
!> \param vec_X ...
! **************************************************************************************************
   SUBROUTINE dgelsy_wrapper(matrix_A, vec_B, vec_X)
      REAL(dp), DIMENSION(:, :)                          :: matrix_A
      REAL(dp), DIMENSION(:)                             :: vec_B, vec_X

      CHARACTER(len=*), PARAMETER                        :: routineN = 'dgelsy_wrapper'

      INTEGER                                            :: handle, info, lda, ldb, lwork, m, n, &
                                                            nrhs, rank
      INTEGER, DIMENSION(:), POINTER                     :: jpvt
      REAL(KIND=dp)                                      :: rcon
      REAL(KIND=dp), DIMENSION(:), POINTER               :: B, work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: A

      CALL timeset(routineN, handle)
      NULLIFY (jpvt, work, B)

      IF (SIZE(matrix_A, 1) /= SIZE(vec_B)) CALL cp__b("pao_param_linpot.F",347,"DGELSY: matrix_A and vec_B not compatible")
      IF (SIZE(matrix_A, 2) /= SIZE(vec_X)) CALL cp__b("pao_param_linpot.F",348,"DGELSY: matrix_A and vec_X not compatible")
      m = SIZE(matrix_A, 1)
      n = SIZE(matrix_A, 2)
      lda = m
      ldb = MAX(m, n)
      nrhs = 1 ! number of right hand sides

      rcon = 0.0_dp ! est. condition number

      ! optional permutations of A
      ALLOCATE (jpvt(n))
      jpvt(:) = 0 ! no permutation

      ! make copies of matrix_A and vec_B
      ALLOCATE (A(m, n), B(MAX(m, n)))
      B(1:m) = vec_B(:)
      A(:, :) = matrix_A(:, :)

      ! query required work space
      lwork = -1
      ALLOCATE (work(1))
      CALL DGELSY(m, n, nrhs, A(1, 1), lda, B(1), ldb, jpvt, rcon, rank, work(1), lwork, info)
      IF (info /= 0) CALL cp__b("pao_param_linpot.F",370,"DGELSY work space query failed")

      ! allocate work space
      lwork = INT(REAL(work(1)))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))

      ! perform actual optimization
      CALL DGELSY(m, n, nrhs, A(1, 1), lda, B(1), ldb, jpvt, rcon, rank, work(1), lwork, info)
      IF (info /= 0) CALL cp__b("pao_param_linpot.F",379,"DGELSY optimization failed")

      ! copy results
      vec_X(:) = B(1:n)

      DEALLOCATE (A, B, work, jpvt)
      CALL timestop(handle)

   END SUBROUTINE dgelsy_wrapper

END MODULE pao_param_linpot
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Front-End for any PAO parametrization
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_param
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_get_block_p, &
        cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_multiply, cp_dbcsr_release, cp_dbcsr_type, dbcsr_distribution_mp, dbcsr_mp_group
   USE cp_log_handling,                 ONLY: cp_to_string
   USE dm_ls_scf_types,                 ONLY: ls_mstruct_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_max
   USE pao_input,                       ONLY: pao_exp_param,&
                                              pao_fock_param,&
                                              pao_gth_param,&
                                              pao_rotinv_param
   USE pao_param_exp,                   ONLY: pao_calc_U_exp,&
                                              pao_param_count_exp,&
                                              pao_param_finalize_exp,&
                                              pao_param_init_exp,&
                                              pao_param_initguess_exp
   USE pao_param_gth,                   ONLY: pao_calc_U_gth,&
                                              pao_param_count_gth,&
                                              pao_param_finalize_gth,&
                                              pao_param_init_gth,&
                                              pao_param_initguess_gth
   USE pao_param_linpot,                ONLY: pao_calc_U_linpot,&
                                              pao_param_count_linpot,&
                                              pao_param_finalize_linpot,&
                                              pao_param_init_linpot,&
                                              pao_param_initguess_linpot
   USE pao_types,                       ONLY: pao_env_type
   USE qs_environment_types,            ONLY: qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_param'

   PUBLIC :: pao_update_AB, pao_param_count, pao_param_initial_guess
   PUBLIC :: pao_param_init, pao_param_finalize, pao_calc_U

CONTAINS

! **************************************************************************************************
!> \brief Takes current matrix_X and recalculates derived matrices U, A, and B.
!> \param pao ...
!> \param ls_mstruct ...
!> \param penalty ...
! **************************************************************************************************
   SUBROUTINE pao_update_AB(pao, ls_mstruct, penalty)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(ls_mstruct_type)                              :: ls_mstruct
      REAL(dp), INTENT(INOUT)                            :: penalty

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_update_AB', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_tmp

      CALL timeset(routineN, handle)

      CALL pao_calc_U(pao, penalty=penalty) !update matrix_U = Function of matrix_X

      !update matrix_A
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=pao%matrix_U)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_N_inv, pao%matrix_U, 0.0_dp, &
                             matrix_tmp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp, pao%matrix_Y, &
                             0.0_dp, ls_mstruct%matrix_A)

      ! update matrix_B
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_N, pao%matrix_U, 0.0_dp, &
                             matrix_tmp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp, pao%matrix_Y, &
                             0.0_dp, ls_mstruct%matrix_B)

      CALL cp_dbcsr_release(matrix_tmp)

      CALL timestop(handle)
   END SUBROUTINE pao_update_AB

! **************************************************************************************************
!> \brief Initialize PAO parametrization
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_init(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_init', routineP = moduleN//':'//routineN

      SELECT CASE (pao%parameterization)
      CASE (pao_exp_param)
         CALL pao_param_init_exp(pao, qs_env)
      CASE (pao_fock_param, pao_rotinv_param)
         CALL pao_param_init_linpot(pao, qs_env)
      CASE (pao_gth_param)
         CALL pao_param_init_gth(pao, qs_env)
      CASE DEFAULT
         CALL cp__b("pao_param.F",112,"PAO: unkown parametrization")
      END SELECT

   END SUBROUTINE pao_param_init

! **************************************************************************************************
!> \brief Finalize PAO parametrization
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_param_finalize(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_finalize', &
         routineP = moduleN//':'//routineN

      SELECT CASE (pao%parameterization)
      CASE (pao_exp_param)
         CALL pao_param_finalize_exp(pao)
      CASE (pao_fock_param, pao_rotinv_param)
         CALL pao_param_finalize_linpot(pao)
      CASE (pao_gth_param)
         CALL pao_param_finalize_gth(pao)
      CASE DEFAULT
         CALL cp__b("pao_param.F",135,"PAO: unkown parametrization")
      END SELECT

   END SUBROUTINE pao_param_finalize

! **************************************************************************************************
!> \brief Returns the number of parameters for given atomic kind
!> \param pao ...
!> \param qs_env ...
!> \param ikind ...
!> \param nparams ...
! **************************************************************************************************
   SUBROUTINE pao_param_count(pao, qs_env, ikind, nparams)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      INTEGER, INTENT(OUT)                               :: nparams

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_count', &
         routineP = moduleN//':'//routineN

      SELECT CASE (pao%parameterization)
      CASE (pao_exp_param)
         CALL pao_param_count_exp(qs_env, ikind=ikind, nparams=nparams)
      CASE (pao_fock_param, pao_rotinv_param)
         CALL pao_param_count_linpot(pao, qs_env, ikind=ikind, nparams=nparams)
      CASE (pao_gth_param)
         CALL pao_param_count_gth(qs_env, ikind=ikind, nparams=nparams)
      CASE DEFAULT
         CALL cp__b("pao_param.F",164,"PAO: unkown parametrization")
      END SELECT

   END SUBROUTINE pao_param_count

! **************************************************************************************************
!> \brief Fills matrix_X with an initial guess
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_param_initial_guess(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_param_initial_guess', &
         routineP = moduleN//':'//routineN

      TYPE(cp_dbcsr_type)                                :: matrix_tmp

      SELECT CASE (pao%parameterization)
      CASE (pao_exp_param)
         CALL pao_param_initguess_exp(pao)
      CASE (pao_fock_param, pao_rotinv_param)
         CALL pao_param_initguess_linpot(pao, qs_env)
      CASE (pao_gth_param)
         CALL pao_param_initguess_gth(pao)
      CASE DEFAULT
         CALL cp__b("pao_param.F",191,"PAO: unkown parametrization")
      END SELECT

      IF (pao%precondition) THEN
         !TODO: multiplying a matrix into itself while retaining sparsity seems to be broken
         CALL cp_dbcsr_init(matrix_tmp)
         CALL cp_dbcsr_copy(matrix_tmp, pao%matrix_X)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_precon, matrix_tmp, &
                                0.0_dp, pao%matrix_X, retain_sparsity=.TRUE.)
         CALL cp_dbcsr_release(matrix_tmp)
      ENDIF

   END SUBROUTINE pao_param_initial_guess

! **************************************************************************************************
!> \brief Calculate new matrix U
!> \param pao ...
!> \param matrix_M ...
!> \param matrix_G ...
!> \param penalty ...
! **************************************************************************************************
   SUBROUTINE pao_calc_U(pao, matrix_M, matrix_G, penalty)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(cp_dbcsr_type), OPTIONAL                      :: matrix_M, matrix_G
      REAL(dp), INTENT(INOUT), OPTIONAL                  :: penalty

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_calc_U', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: my_penalty

      CALL timeset(routineN, handle)
      my_penalty = 0.0_dp

      SELECT CASE (pao%parameterization)
      CASE (pao_exp_param)
         CALL pao_calc_U_exp(pao, matrix_M, matrix_G)
      CASE (pao_fock_param, pao_rotinv_param)
         CALL pao_calc_U_linpot(pao, my_penalty, matrix_M, matrix_G)
      CASE (pao_gth_param)
         CALL pao_calc_U_gth(pao, my_penalty, matrix_M, matrix_G)
      CASE DEFAULT
         CALL cp__b("pao_param.F",233,"PAO: unkown parametrization")
      END SELECT

      IF (PRESENT(penalty)) penalty = my_penalty

      CALL pao_assert_unitary(pao, pao%matrix_U, pao%matrix_Y)
      CALL timestop(handle)
   END SUBROUTINE pao_calc_U

! **************************************************************************************************
!> \brief Debugging routine, check unitaryness of U
!> \param pao ...
!> \param matrix_test ...
!> \param matrix_Y ...
! **************************************************************************************************
   SUBROUTINE pao_assert_unitary(pao, matrix_test, matrix_Y)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(cp_dbcsr_type)                                :: matrix_test, matrix_Y

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_assert_unitary', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, group, handle, i, M, N
      LOGICAL                                            :: found
      REAL(dp)                                           :: delta_max
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_test, block_Y, tmp1, tmp2
      TYPE(cp_dbcsr_iterator)                            :: iter

      IF (pao%check_unitary_tol < 0.0_dp) RETURN ! no checking

      CALL timeset(routineN, handle)
      delta_max = 0.0_dp

      CALL cp_dbcsr_iterator_start(iter, matrix_test)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_test)
         CALL cp_dbcsr_get_block_p(matrix=matrix_Y, row=arow, col=acol, block=block_Y, found=found)
         IF(.NOT.(ASSOCIATED(block_Y)))CALL cp__a("pao_param.F",270)
         N = SIZE(block_Y, 1)
         M = SIZE(block_Y, 2)
         ALLOCATE (tmp1(N, M), tmp2(M, M))

         ! we only need the upper left "PAO-corner" to be unitary
         tmp1 = MATMUL(block_test, block_Y)
         tmp2 = MATMUL(TRANSPOSE(tmp1), tmp1)
         DO i = 1, M
            tmp2(i, i) = tmp2(i, i)-1.0_dp
         ENDDO

         delta_max = MAX(delta_max, MAXVAL(ABS(tmp2)))
         DEALLOCATE (tmp1, tmp2)
      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      group = dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(matrix_test)))
      CALL mp_max(delta_max, group)
      IF (pao%iw > 0) WRITE (pao%iw, *) 'PAO| checked unitaryness, max delta:', delta_max
      IF (delta_max > pao%check_unitary_tol) &
         CALL cp__b("pao_param.F",291,"Found bad unitaryness:"//cp_to_string(delta_max))

      CALL timestop(handle)
   END SUBROUTINE pao_assert_unitary

END MODULE pao_param
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief lower level routines for linear scaling SCF
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE dm_ls_scf_methods
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_extremal, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_desymmetrize, cp_dbcsr_filter, cp_dbcsr_finalize, &
        cp_dbcsr_frobenius_norm, cp_dbcsr_get_data_type, cp_dbcsr_get_occupation, cp_dbcsr_init, &
        cp_dbcsr_iterator_start, cp_dbcsr_multiply, cp_dbcsr_put_block, cp_dbcsr_release, &
        cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_trace, cp_dbcsr_type, dbcsr_data_clear_pointer, &
        dbcsr_data_init, dbcsr_data_new, dbcsr_data_obj, dbcsr_data_release, dbcsr_iterator, &
        dbcsr_iterator_blocks_left, dbcsr_iterator_next_block, dbcsr_iterator_stop, &
        dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dm_ls_scf_qs,                    ONLY: matrix_qs_to_ls
   USE dm_ls_scf_types,                 ONLY: ls_cluster_atomic,&
                                              ls_mstruct_type,&
                                              ls_scf_env_type
   USE input_constants,                 ONLY: ls_cluster_atomic,&
                                              ls_s_preconditioner_atomic,&
                                              ls_s_preconditioner_molecular,&
                                              ls_s_preconditioner_none
   USE iterate_matrix,                  ONLY: invert_Hotelling,&
                                              matrix_sign_Newton_Schulz,&
                                              matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathlib,                         ONLY: abnormal_value
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_scf_methods'

   PUBLIC :: ls_scf_init_matrix_S
   PUBLIC :: density_matrix_sign, density_matrix_sign_fixed_mu
   PUBLIC :: compute_matrix_preconditioner, apply_matrix_preconditioner
   PUBLIC :: density_matrix_trs4, density_matrix_tc2, compute_homo_lumo

CONTAINS

! **************************************************************************************************
!> \brief initialize S matrix related properties (sqrt, inverse...)
!>        Might be factored-out since this seems common code with the other SCF.
!> \param matrix_s ...
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_init_matrix_S(matrix_s, ls_scf_env)
      TYPE(cp_dbcsr_type)                                :: matrix_s
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_init_matrix_S', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      REAL(KIND=dp)                                      :: frob_matrix, frob_matrix_base
      TYPE(cp_dbcsr_type)                                :: matrix_tmp1, matrix_tmp2
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! make our own copy of S
      IF (ls_scf_env%has_unit_metric) THEN
         CALL cp_dbcsr_set(ls_scf_env%matrix_s, 0.0_dp)
         CALL cp_dbcsr_add_on_diag(ls_scf_env%matrix_s, 1.0_dp)
      ELSE
         CALL matrix_qs_to_ls(ls_scf_env%matrix_s, matrix_s, ls_scf_env%ls_mstruct, covariant=.TRUE.)
      ENDIF

      CALL cp_dbcsr_filter(ls_scf_env%matrix_s, ls_scf_env%eps_filter)

      ! needs a preconditioner for S
      IF (ls_scf_env%has_s_preconditioner) THEN
         CALL cp_dbcsr_init(ls_scf_env%matrix_bs_sqrt)
         CALL cp_dbcsr_create(ls_scf_env%matrix_bs_sqrt, template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_init(ls_scf_env%matrix_bs_sqrt_inv)
         CALL cp_dbcsr_create(ls_scf_env%matrix_bs_sqrt_inv, template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL compute_matrix_preconditioner(ls_scf_env%matrix_s, &
                                            ls_scf_env%s_preconditioner_type, ls_scf_env%ls_mstruct, &
                                            ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv, &
                                            ls_scf_env%eps_filter, ls_scf_env%sign_sqrt_order, &
                                            ls_scf_env%eps_lanczos, ls_scf_env%max_iter_lanczos)
      ENDIF

      ! precondition S
      IF (ls_scf_env%has_s_preconditioner) THEN
         CALL apply_matrix_preconditioner(ls_scf_env%matrix_s, "forward", &
                                          ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
      ENDIF

      ! compute sqrt(S) and inv(sqrt(S))
      IF (ls_scf_env%use_s_sqrt) THEN
         CALL cp_dbcsr_init(ls_scf_env%matrix_s_sqrt)
         CALL cp_dbcsr_init(ls_scf_env%matrix_s_sqrt_inv)
         CALL cp_dbcsr_create(ls_scf_env%matrix_s_sqrt, template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(ls_scf_env%matrix_s_sqrt_inv, template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)

         CALL matrix_sqrt_Newton_Schulz(ls_scf_env%matrix_s_sqrt, ls_scf_env%matrix_s_sqrt_inv, &
                                        ls_scf_env%matrix_s, ls_scf_env%eps_filter, &
                                        ls_scf_env%sign_sqrt_order, &
                                        ls_scf_env%eps_lanczos, ls_scf_env%max_iter_lanczos)

         IF (ls_scf_env%check_s_inv) THEN
            CALL cp_dbcsr_init(matrix_tmp1)
            CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_init(matrix_tmp2)
            CALL cp_dbcsr_create(matrix_tmp2, template=ls_scf_env%matrix_s, &
                                 matrix_type=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_s, &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)

            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt_inv, &
                                   0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)

            frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
            CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
            frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) "Error for (inv(sqrt(S))*S*inv(sqrt(S))-I)", frob_matrix/frob_matrix_base
            ENDIF

            CALL cp_dbcsr_release(matrix_tmp1)
            CALL cp_dbcsr_release(matrix_tmp2)
         ENDIF
      ENDIF

      ! compute the inverse of S
      IF (ls_scf_env%needs_s_inv) THEN
         CALL cp_dbcsr_init(ls_scf_env%matrix_s_inv)
         CALL cp_dbcsr_create(ls_scf_env%matrix_s_inv, template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         IF (.NOT. ls_scf_env%use_s_sqrt) THEN
            CALL invert_Hotelling(ls_scf_env%matrix_s_inv, ls_scf_env%matrix_s, ls_scf_env%eps_filter)
         ELSE
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_s_sqrt_inv, &
                                   0.0_dp, ls_scf_env%matrix_s_inv, filter_eps=ls_scf_env%eps_filter)
         ENDIF
         IF (ls_scf_env%check_s_inv) THEN
            CALL cp_dbcsr_init(matrix_tmp1)
            CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_inv, ls_scf_env%matrix_s, &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
            frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
            CALL cp_dbcsr_add_on_diag(matrix_tmp1, -1.0_dp)
            frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) "Error for (inv(S)*S-I)", frob_matrix/frob_matrix_base
            ENDIF
            CALL cp_dbcsr_release(matrix_tmp1)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_init_matrix_s

! **************************************************************************************************
!> \brief compute for a block positive definite matrix s (bs)
!>        the sqrt(bs) and inv(sqrt(bs))
!> \param matrix_s ...
!> \param preconditioner_type ...
!> \param ls_mstruct ...
!> \param matrix_bs_sqrt ...
!> \param matrix_bs_sqrt_inv ...
!> \param threshold ...
!> \param order ...
!> \param eps_lanczos ...
!> \param max_iter_lanczos ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE compute_matrix_preconditioner(matrix_s, preconditioner_type, ls_mstruct, &
                                            matrix_bs_sqrt, matrix_bs_sqrt_inv, threshold, order, eps_lanczos, max_iter_lanczos)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_s
      INTEGER                                            :: preconditioner_type
      TYPE(ls_mstruct_type)                              :: ls_mstruct
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_bs_sqrt, matrix_bs_sqrt_inv
      REAL(KIND=dp)                                      :: threshold
      INTEGER                                            :: order
      REAL(KIND=dp)                                      :: eps_lanczos
      INTEGER                                            :: max_iter_lanczos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_matrix_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iblock_col, iblock_row
      LOGICAL                                            :: block_needed, tr
      TYPE(cp_dbcsr_type)                                :: matrix_bs
      TYPE(dbcsr_data_obj)                               :: block_any
      TYPE(dbcsr_iterator)                               :: iter

      CALL timeset(routineN, handle)

      ! first generate a block diagonal copy of s
      CALL cp_dbcsr_init(matrix_bs)
      CALL cp_dbcsr_create(matrix_bs, template=matrix_s)

      CALL dbcsr_data_init(block_any)
      CALL dbcsr_data_new(block_any, cp_dbcsr_get_data_type(matrix_s))

      SELECT CASE (preconditioner_type)
      CASE (ls_s_preconditioner_none)
      CASE (ls_s_preconditioner_atomic, ls_s_preconditioner_molecular)
         CALL cp_dbcsr_iterator_start(iter, matrix_s)
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            !CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, block_s, blk)
            CALL dbcsr_iterator_next_block(iter, iblock_row, iblock_col, block_any, tr)

            ! do we need the block ?
            ! this depends on the preconditioner, but also the matrix clustering method employed
            ! for a clustered matrix, right now, we assume that atomic and molecular preconditioners
            ! are actually the same, and only require that the diagonal blocks (clustered) are present

            block_needed = .FALSE.

            IF (iblock_row == iblock_col) THEN
               block_needed = .TRUE.
            ELSE
               IF (preconditioner_type == ls_s_preconditioner_molecular .AND. &
                   ls_mstruct%cluster_type == ls_cluster_atomic) THEN
                  IF (ls_mstruct%atom_to_molecule(iblock_row) == ls_mstruct%atom_to_molecule(iblock_col)) block_needed = .TRUE.
               ENDIF
            ENDIF

            ! add it
            IF (block_needed) THEN
               CALL cp_dbcsr_put_block(matrix=matrix_bs, &
                                       row=iblock_row, col=iblock_col, block=block_any)

            ENDIF

         ENDDO
         CALL dbcsr_iterator_stop(iter)
      END SELECT

      CALL dbcsr_data_clear_pointer(block_any)
      CALL dbcsr_data_release(block_any)

      CALL cp_dbcsr_finalize(matrix_bs)

      SELECT CASE (preconditioner_type)
      CASE (ls_s_preconditioner_none)
         ! for now make it a simple identity matrix
         CALL cp_dbcsr_copy(matrix_bs_sqrt, matrix_bs)
         CALL cp_dbcsr_set(matrix_bs_sqrt, 0.0_dp)
         CALL cp_dbcsr_add_on_diag(matrix_bs_sqrt, 1.0_dp)

         ! for now make it a simple identity matrix
         CALL cp_dbcsr_copy(matrix_bs_sqrt_inv, matrix_bs)
         CALL cp_dbcsr_set(matrix_bs_sqrt_inv, 0.0_dp)
         CALL cp_dbcsr_add_on_diag(matrix_bs_sqrt_inv, 1.0_dp)
      CASE (ls_s_preconditioner_atomic, ls_s_preconditioner_molecular)
         CALL cp_dbcsr_copy(matrix_bs_sqrt, matrix_bs)
         CALL cp_dbcsr_copy(matrix_bs_sqrt_inv, matrix_bs)
         ! XXXXXXXXXXX
         ! XXXXXXXXXXX the threshold here could be done differently,
         ! XXXXXXXXXXX using eps_filter is reducing accuracy for no good reason, this is cheap
         ! XXXXXXXXXXX
         CALL matrix_sqrt_Newton_Schulz(matrix_bs_sqrt, matrix_bs_sqrt_inv, matrix_bs, &
                                        threshold=MIN(threshold, 1.0E-10_dp), order=order, &
                                        eps_lanczos=eps_lanczos, max_iter_lanczos=max_iter_lanczos)
      END SELECT

      CALL cp_dbcsr_release(matrix_bs)

      CALL timestop(handle)

   END SUBROUTINE compute_matrix_preconditioner

! **************************************************************************************************
!> \brief apply a preconditioner either
!>        forward (precondition)            inv(sqrt(bs)) * A * inv(sqrt(bs))
!>        backward (restore to old form)        sqrt(bs)  * A * sqrt(bs)
!> \param matrix ...
!> \param direction ...
!> \param matrix_bs_sqrt ...
!> \param matrix_bs_sqrt_inv ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE apply_matrix_preconditioner(matrix, direction, matrix_bs_sqrt, matrix_bs_sqrt_inv)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      CHARACTER(LEN=*)                                   :: direction
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_bs_sqrt, matrix_bs_sqrt_inv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_matrix_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_tmp

      CALL timeset(routineN, handle)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix, matrix_type=dbcsr_type_no_symmetry)

      SELECT CASE (direction)
      CASE ("forward")
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix, matrix_bs_sqrt_inv, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_bs_sqrt_inv, matrix_tmp, &
                                0.0_dp, matrix)
      CASE ("backward")
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix, matrix_bs_sqrt, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_bs_sqrt, matrix_tmp, &
                                0.0_dp, matrix)
      CASE DEFAULT
         CALL cp__b("dm_ls_scf_methods.F",344,"")
      END SELECT

      CALL cp_dbcsr_release(matrix_tmp)

      CALL timestop(handle)

   END SUBROUTINE apply_matrix_preconditioner

! **************************************************************************************************
!> \brief compute the density matrix with a trace that is close to nelectron.
!>        take a mu as input, and improve by bisection as needed.
!> \param matrix_p ...
!> \param mu ...
!> \param fixed_mu ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param matrix_s_inv ...
!> \param nelectron ...
!> \param threshold ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE density_matrix_sign(matrix_p, mu, fixed_mu, matrix_ks, matrix_s, matrix_s_inv, nelectron, threshold)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      REAL(KIND=dp), INTENT(INOUT)                       :: mu
      LOGICAL                                            :: fixed_mu
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_ks, matrix_s, matrix_s_inv
      INTEGER, INTENT(IN)                                :: nelectron
      REAL(KIND=dp), INTENT(IN)                          :: threshold

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_matrix_sign', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: initial_increment = 0.01_dp

      INTEGER                                            :: handle, iter, unit_nr
      LOGICAL                                            :: has_mu_high, has_mu_low
      REAL(KIND=dp)                                      :: increment, mu_high, mu_low, trace
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      increment = initial_increment

      has_mu_low = .FALSE.
      has_mu_high = .FALSE.

      ! bisect if both bounds are known, otherwise find the bounds with a linear search
      DO iter = 1, 30
         IF (has_mu_low .AND. has_mu_high) THEN
            mu = (mu_low+mu_high)/2
            IF (ABS(mu_high-mu_low) < threshold) EXIT
         ENDIF

         CALL density_matrix_sign_fixed_mu(matrix_p, trace, mu, matrix_ks, matrix_s, matrix_s_inv, threshold)
         IF (unit_nr > 0) WRITE (unit_nr, '(T2,A,I2,1X,F13.9,1X,F15.9)') &
            "Density matrix:  iter, mu, trace error: ", iter, mu, trace-nelectron

         ! OK, we can skip early if we are as close as possible to the exact result
         ! smaller differences should be considered 'noise'
         IF (ABS(trace-nelectron) < 0.5_dp .OR. fixed_mu) EXIT

         IF (trace < nelectron) THEN
            mu_low = mu
            mu = mu+increment
            has_mu_low = .TRUE.
            increment = increment*2
         ELSE
            mu_high = mu
            mu = mu-increment
            has_mu_high = .TRUE.
            increment = increment*2
         ENDIF
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE density_matrix_sign

! **************************************************************************************************
!> \brief for a fixed mu, compute the corresponding density matrix and its trace
!> \param matrix_p ...
!> \param trace ...
!> \param mu ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param matrix_s_inv ...
!> \param threshold ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE density_matrix_sign_fixed_mu(matrix_p, trace, mu, matrix_ks, matrix_s, matrix_s_inv, threshold)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      REAL(KIND=dp), INTENT(OUT)                         :: trace
      REAL(KIND=dp), INTENT(INOUT)                       :: mu
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_ks, matrix_s, matrix_s_inv
      REAL(KIND=dp), INTENT(IN)                          :: threshold

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_matrix_sign_fixed_mu', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      REAL(KIND=dp)                                      :: frob_matrix
      TYPE(cp_dbcsr_type)                                :: matrix_p_ud, matrix_sign, &
                                                            matrix_sinv_ks, matrix_tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! get inv(S)*H-I*mu
      CALL cp_dbcsr_init(matrix_sinv_ks)
      CALL cp_dbcsr_create(matrix_sinv_ks, template=matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_inv, matrix_ks, &
                             0.0_dp, matrix_sinv_ks, filter_eps=threshold)
      CALL cp_dbcsr_add_on_diag(matrix_sinv_ks, -mu)

      ! compute sign(inv(S)*H-I*mu)
      CALL cp_dbcsr_init(matrix_sign)
      CALL cp_dbcsr_create(matrix_sign, template=matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL matrix_sign_Newton_Schulz(matrix_sign, matrix_sinv_ks, threshold)
      CALL cp_dbcsr_release(matrix_sinv_ks)

      ! now construct the density matrix PS=0.5*(I-sign(inv(S)H-I*mu))
      CALL cp_dbcsr_init(matrix_p_ud)
      CALL cp_dbcsr_create(matrix_p_ud, template=matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(matrix_p_ud, matrix_sign)
      CALL cp_dbcsr_scale(matrix_p_ud, -0.5_dp)
      CALL cp_dbcsr_add_on_diag(matrix_p_ud, 0.5_dp)
      CALL cp_dbcsr_release(matrix_sign)

      ! we now have PS, lets get its trace
      CALL cp_dbcsr_trace(matrix_p_ud, trace)

      ! we can also check it is idempotent PS*PS=PS
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p_ud, matrix_p_ud, &
                             0.0_dp, matrix_tmp, filter_eps=threshold)
      CALL cp_dbcsr_add(matrix_tmp, matrix_p_ud, 1.0_dp, -1.0_dp)
      frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp)
      CALL cp_dbcsr_release(matrix_tmp)
      IF (unit_nr > 0) WRITE (unit_nr, '(T2,A,F20.12)') "Deviation from idempotency: ", frob_matrix

      ! get P=PS*inv(S)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p_ud, matrix_s_inv, &
                             0.0_dp, matrix_p, filter_eps=threshold)
      CALL cp_dbcsr_release(matrix_p_ud)

      CALL timestop(handle)

   END SUBROUTINE density_matrix_sign_fixed_mu

! **************************************************************************************************
!> \brief compute the density matrix using a trace-resetting algorithm
!> \param matrix_p ...
!> \param matrix_ks ...
!> \param matrix_s_sqrt_inv ...
!> \param nelectron ...
!> \param threshold ...
!> \param e_homo ...
!> \param e_lumo ...
!> \param e_mu ...
!> \param dynamic_threshold ...
!> \param matrix_ks_deviation ...
!> \param max_iter_lanczos ...
!> \param eps_lanczos ...
!> \param converged ...
!> \par History
!>       2012.06 created [Florian Thoele]
!> \author Florian Thoele
! **************************************************************************************************
   SUBROUTINE density_matrix_trs4(matrix_p, matrix_ks, matrix_s_sqrt_inv, &
                                  nelectron, threshold, e_homo, e_lumo, e_mu, &
                                  dynamic_threshold, matrix_ks_deviation, &
                                  max_iter_lanczos, eps_lanczos, converged)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_ks, matrix_s_sqrt_inv
      INTEGER, INTENT(IN)                                :: nelectron
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      REAL(KIND=dp), INTENT(INOUT)                       :: e_homo, e_lumo, e_mu
      LOGICAL, INTENT(IN), OPTIONAL                      :: dynamic_threshold
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: matrix_ks_deviation
      INTEGER, INTENT(IN)                                :: max_iter_lanczos
      REAL(KIND=dp), INTENT(IN)                          :: eps_lanczos
      LOGICAL, INTENT(OUT), OPTIONAL                     :: converged

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_matrix_trs4', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: max_iter = 100
      REAL(KIND=dp), PARAMETER                           :: gamma_max = 6.0_dp, gamma_min = 0.0_dp

      INTEGER                                            :: branch, estimated_steps, handle, i, j, &
                                                            unit_nr
      INTEGER(kind=int_8)                                :: flop1, flop2
      LOGICAL                                            :: arnoldi_converged, do_dyn_threshold
      REAL(KIND=dp) :: current_threshold, delta_n, eps_max, eps_min, est_threshold, frob_id, &
         frob_x, gam, homo, lumo, max_eig, max_threshold, maxdev, maxev, min_eig, minev, mmin, mu, &
         mu_a, mu_b, mu_c, mu_fa, mu_fc, occ_matrix, scaled_homo_bound, scaled_lumo_bound, t1, t2, &
         trace_fx, trace_gx, xi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: gamma_values
      TYPE(cp_dbcsr_type)                                :: matrix_k0, matrix_x, matrix_x_nosym, &
                                                            matrix_xidsq, matrix_xsq, tmp_gx
      TYPE(cp_logger_type), POINTER                      :: logger

      IF (nelectron == 0) THEN
         CALL cp_dbcsr_set(matrix_p, 0.0_dp)
         RETURN
      ENDIF

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      do_dyn_threshold = .FALSE.
      IF (PRESENT(dynamic_threshold)) do_dyn_threshold = dynamic_threshold

      IF (PRESENT(converged)) converged = .FALSE.

      ! init X = (eps_n*I - H)/(eps_n - eps_0)  ... H* = S^-1/2*H*S^-1/2
      CALL cp_dbcsr_init(matrix_x)
      CALL cp_dbcsr_create(matrix_x, template=matrix_ks, matrix_type="S")

      ! at some points the non-symmetric version of x is required
      CALL cp_dbcsr_init(matrix_x_nosym)
      CALL cp_dbcsr_create(matrix_x_nosym, template=matrix_ks, matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_sqrt_inv, matrix_ks, &
                             0.0_dp, matrix_x_nosym, filter_eps=threshold)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x_nosym, matrix_s_sqrt_inv, &
                             0.0_dp, matrix_x, filter_eps=threshold)
      CALL cp_dbcsr_desymmetrize(matrix_x, matrix_x_nosym)

      CALL cp_dbcsr_init(matrix_k0)
      CALL cp_dbcsr_create(matrix_k0, template=matrix_ks, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(matrix_k0, matrix_x_nosym)

      ! compute the deviation in the mixed matrix, as seen in the ortho basis
      IF (do_dyn_threshold) THEN
         IF(.NOT.(PRESENT(matrix_ks_deviation)))CALL cp__a("dm_ls_scf_methods.F",606)
         CALL cp_dbcsr_add(matrix_ks_deviation, matrix_x_nosym, -1.0_dp, 1.0_dp)
         CALL cp_dbcsr_arnoldi_extremal(matrix_ks_deviation, maxev, minev, max_iter=max_iter_lanczos, threshold=eps_lanczos, &
                                        converged=arnoldi_converged)
         maxdev = MAX(ABS(maxev), ABS(minev))
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T6,A,1X,L12)') "Lanczos converged:      ", arnoldi_converged
            WRITE (unit_nr, '(T6,A,1X,F12.5)') "change in mixed matrix: ", maxdev
            WRITE (unit_nr, '(T6,A,1X,F12.5)') "HOMO upper bound:       ", e_homo+maxdev
            WRITE (unit_nr, '(T6,A,1X,F12.5)') "LUMO lower bound:       ", e_lumo-maxdev
            WRITE (unit_nr, '(T6,A,1X,L12)') "Predicts a gap ?        ", ((e_lumo-maxdev)-(e_homo+maxdev)) > 0
         ENDIF
         ! save the old mixed matrix
         CALL cp_dbcsr_copy(matrix_ks_deviation, matrix_x_nosym)

      ENDIF

      ! get largest/smallest eigenvalues for scaling
      CALL cp_dbcsr_arnoldi_extremal(matrix_x_nosym, max_eig, min_eig, max_iter=max_iter_lanczos, threshold=eps_lanczos, &
                                     converged=arnoldi_converged)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,1X,2F12.5,1X,A,1X,L1)') "Est. extremal eigenvalues", &
         min_eig, max_eig, " converged: ", arnoldi_converged
      eps_max = max_eig
      eps_min = min_eig

      ! scale KS matrix
      IF (eps_max == eps_min) THEN
         CALL cp_dbcsr_scale(matrix_x, 1.0_dp/eps_max)
      ELSE
         CALL cp_dbcsr_add_on_diag(matrix_x, -eps_max)
         CALL cp_dbcsr_scale(matrix_x, -1.0_dp/(eps_max-eps_min))
      ENDIF

      current_threshold = threshold
      IF (do_dyn_threshold) THEN
         ! scale bounds for HOMO/LUMO
         scaled_homo_bound = (eps_max-(e_homo+maxdev))/(eps_max-eps_min)
         scaled_lumo_bound = (eps_max-(e_lumo-maxdev))/(eps_max-eps_min)
      ENDIF

      CALL cp_dbcsr_init(matrix_xsq)
      CALL cp_dbcsr_create(matrix_xsq, template=matrix_ks, matrix_type="S")

      CALL cp_dbcsr_init(matrix_xidsq)
      CALL cp_dbcsr_create(matrix_xidsq, template=matrix_ks, matrix_type="S")

      CALL cp_dbcsr_init(tmp_gx)
      CALL cp_dbcsr_create(tmp_gx, template=matrix_ks, matrix_type="S")

      ALLOCATE (gamma_values(max_iter))

      DO i = 1, max_iter
         t1 = m_walltime()
         flop1 = 0; flop2 = 0

         ! get X*X
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_x, &
                                0.0_dp, matrix_xsq, &
                                filter_eps=current_threshold, flop=flop1)

         ! intermediate use matrix_xidsq to compute = X*X-X
         CALL cp_dbcsr_copy(matrix_xidsq, matrix_x)
         CALL cp_dbcsr_add(matrix_xidsq, matrix_xsq, -1.0_dp, 1.0_dp)
         frob_id = cp_dbcsr_frobenius_norm(matrix_xidsq)
         frob_x = cp_dbcsr_frobenius_norm(matrix_x)

         ! xidsq = (1-X)*(1-X)
         ! use (1-x)*(1-x) = 1 + x*x - 2*x
         CALL cp_dbcsr_copy(matrix_xidsq, matrix_x)
         CALL cp_dbcsr_add(matrix_xidsq, matrix_xsq, -2.0_dp, 1.0_dp)
         CALL cp_dbcsr_add_on_diag(matrix_xidsq, 1.0_dp)

         ! tmp_gx = 4X-3X*X
         CALL cp_dbcsr_copy(tmp_gx, matrix_x)
         CALL cp_dbcsr_add(tmp_gx, matrix_xsq, 4.0_dp, -3.0_dp)

         ! get gamma
         ! Tr(F) = Tr(XX*tmp_gx) Tr(G) is equivalent
         CALL cp_dbcsr_trace(matrix_xsq, matrix_xidsq, trace_gx)
         CALL cp_dbcsr_trace(matrix_xsq, tmp_gx, trace_fx)

         ! if converged, and gam becomes noisy, fix it to 3, which results in a final McWeeny step.
         ! do this only if the electron count is reasonable.
         ! maybe tune if the current criterion is not good enough
         delta_n = nelectron-trace_fx
         IF ((ABS(frob_id/frob_x) < SQRT(threshold)) .AND. (ABS(delta_n) < 0.5_dp)) THEN
            gam = 3.0_dp
         ELSE IF (ABS(delta_n) < 1e-14_dp) THEN
            gam = 0.0_dp ! rare case of perfect electron count
         ELSE
            ! make sure, we don't divide by zero, as soon as gam is outside the interval gam_min,gam_max, it doesn't matter
            gam = delta_n/MAX(trace_gx, ABS(delta_n)/100)
         ENDIF
         gamma_values(i) = gam

         IF (unit_nr > 0 .AND. .FALSE.) THEN
            WRITE (unit_nr, *) "trace_fx", trace_fx, "trace_gx", trace_gx, "gam", gam, &
               "frob_id", frob_id, "conv", ABS(frob_id/frob_x)
         ENDIF

         IF (do_dyn_threshold) THEN
            ! quantities used for dynamic thresholding, when the estimated gap is larger than zero
            xi = (scaled_homo_bound-scaled_lumo_bound)
            IF (xi > 0.0_dp) THEN
               mmin = 0.5*(scaled_homo_bound+scaled_lumo_bound)
               max_threshold = ABS(1-2*mmin)*xi

               scaled_homo_bound = evaluate_trs4_polynomial(scaled_homo_bound, gamma_values(i:), 1)
               scaled_lumo_bound = evaluate_trs4_polynomial(scaled_lumo_bound, gamma_values(i:), 1)
               estimated_steps = estimate_steps(scaled_homo_bound, scaled_lumo_bound, threshold)

               est_threshold = (threshold/(estimated_steps+i+1))*xi/(1+threshold/(estimated_steps+i+1))
               est_threshold = MIN(max_threshold, est_threshold)
               IF (i > 1) est_threshold = MAX(est_threshold, 0.1_dp*current_threshold)
               current_threshold = est_threshold
            ELSE
               current_threshold = threshold
            ENDIF
         ENDIF

         IF (gam > gamma_max) THEN
            ! Xn+1 = 2X-X*X
            CALL cp_dbcsr_add(matrix_x, matrix_xsq, 2.0_dp, -1.0_dp)
            CALL cp_dbcsr_filter(matrix_x, current_threshold)
            branch = 1
         ELSE IF (gam < gamma_min) THEN
            ! Xn+1 = X*X
            CALL cp_dbcsr_copy(matrix_x, matrix_xsq)
            branch = 2
         ELSE
            ! Xn+1 = F(X) + gam*G(X)
            CALL cp_dbcsr_add(tmp_gx, matrix_xidsq, 1.0_dp, gam)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_xsq, tmp_gx, &
                                   0.0_dp, matrix_x, &
                                   flop=flop2, filter_eps=current_threshold)
            branch = 3
         ENDIF

         occ_matrix = cp_dbcsr_get_occupation(matrix_x)
         t2 = m_walltime()
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, &
                   '(T6,A,I3,1X,F10.8,E12.3,F12.3,F13.3,E12.3)') "TRS4 it ", &
               i, occ_matrix, ABS(trace_gx), t2-t1, &
               (flop1+flop2)/(1.0E6_dp*MAX(t2-t1, 0.001_dp)), current_threshold
            CALL m_flush(unit_nr)
         ENDIF

         IF (abnormal_value(trace_gx)) &
            CALL cp__b("dm_ls_scf_methods.F",755,"trace_gx is an abnormal value (NaN/Inf).")

         ! a branch of 1 or 2 appears to lead to a less accurate electron number count and premature exit
         ! if it turns out this does not exit because we get stuck in branch 1/2 for a reason we need to refine further
         IF (ABS(frob_id/frob_x) < SQRT(threshold) .AND. branch == 3 .AND. (ABS(delta_n) < 0.5_dp)) THEN
            IF (PRESENT(converged)) converged = .TRUE.
            EXIT
         ENDIF

      END DO

      occ_matrix = cp_dbcsr_get_occupation(matrix_x)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,I3,1X,F10.8,E12.3)') 'Final TRS4 iteration  ', i, occ_matrix, ABS(trace_gx)

      ! free some memory
      CALL cp_dbcsr_release(tmp_gx)
      CALL cp_dbcsr_release(matrix_xsq)
      CALL cp_dbcsr_release(matrix_xidsq)

      ! output to matrix_p, P = inv(S)^0.5 X inv(S)^0.5
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_s_sqrt_inv, &
                             0.0_dp, matrix_x_nosym, filter_eps=threshold)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_sqrt_inv, matrix_x_nosym, &
                             0.0_dp, matrix_p, filter_eps=threshold)

      ! calculate the chemical potential by doing a bisection of fk(x0)-0.5, where fk is evaluated using the stored values for gamma
      ! E. Rubensson et al., Chem Phys Lett 432, 2006, 591-594
      mu_a = 0.0_dp; mu_b = 1.0_dp; 
      mu_fa = evaluate_trs4_polynomial(mu_a, gamma_values, i-1)-0.5_dp
      DO j = 1, 40
         mu_c = 0.5*(mu_a+mu_b)
         mu_fc = evaluate_trs4_polynomial(mu_c, gamma_values, i-1)-0.5_dp ! i-1 because in the last iteration, only convergence is checked
         IF (ABS(mu_fc) < 1.0E-6_dp .OR. (mu_b-mu_a)/2 < 1.0E-6_dp) EXIT !TODO: define threshold values

         IF (mu_fc*mu_fa > 0) THEN
            mu_a = mu_c
            mu_fa = mu_fc
         ELSE
            mu_b = mu_c
         ENDIF
      ENDDO
      mu = (eps_min-eps_max)*mu_c+eps_max
      DEALLOCATE (gamma_values)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,F12.5)') 'Chemical potential (mu): ', mu
      ENDIF
      e_mu = mu

      IF (do_dyn_threshold) THEN
         CALL cp_dbcsr_desymmetrize(matrix_x, matrix_x_nosym)
         CALL compute_homo_lumo(matrix_k0, matrix_x_nosym, eps_min, eps_max, &
                                threshold, max_iter_lanczos, eps_lanczos, homo, lumo, unit_nr)
         e_homo = homo
         e_lumo = lumo
      ENDIF

      CALL cp_dbcsr_release(matrix_x)
      CALL cp_dbcsr_release(matrix_x_nosym)
      CALL cp_dbcsr_release(matrix_k0)
      CALL timestop(handle)

   END SUBROUTINE density_matrix_trs4

! **************************************************************************************************
!> \brief compute the density matrix using a non monotonic trace conserving
!>  algorithm based on SIAM DOI. 10.1137/130911585.
!>       2014.04 created [Jonathan Mullin]
!> \param matrix_p ...
!> \param matrix_ks ...
!> \param matrix_s_sqrt_inv ...
!> \param nelectron ...
!> \param threshold ...
!> \param e_homo ...
!> \param e_lumo ...
!> \param non_monotonic ...
!> \param eps_lanczos ...
!> \param max_iter_lanczos ...
!> \author Jonathan Mullin
! **************************************************************************************************
   SUBROUTINE density_matrix_tc2(matrix_p, matrix_ks, matrix_s_sqrt_inv, &
                                 nelectron, threshold, e_homo, e_lumo, &
                                 non_monotonic, eps_lanczos, max_iter_lanczos)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_ks, matrix_s_sqrt_inv
      INTEGER, INTENT(IN)                                :: nelectron
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      REAL(KIND=dp), INTENT(INOUT)                       :: e_homo, e_lumo
      LOGICAL, INTENT(IN), OPTIONAL                      :: non_monotonic
      REAL(KIND=dp), INTENT(IN)                          :: eps_lanczos
      INTEGER, INTENT(IN)                                :: max_iter_lanczos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_matrix_tc2', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: max_iter = 100

      INTEGER                                            :: handle, i, j, k, unit_nr
      INTEGER(kind=int_8)                                :: flop1, flop2
      LOGICAL                                            :: converged, do_non_monotonic
      REAL(KIND=dp)                                      :: beta, betaB, eps_max, eps_min, gama, &
                                                            max_eig, min_eig, occ_matrix, t1, t2, &
                                                            trace_fx, trace_gx
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: alpha, lambda, nu, poly, wu, X, Y
      TYPE(cp_dbcsr_type)                                :: matrix_tmp, matrix_x, matrix_xsq
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      do_non_monotonic = .FALSE.
      IF (PRESENT(non_monotonic)) do_non_monotonic = non_monotonic

      ! init X = (eps_n*I - H)/(eps_n - eps_0)  ... H* = S^-1/2*H*S^-1/2
      CALL cp_dbcsr_init(matrix_x)
      CALL cp_dbcsr_create(matrix_x, template=matrix_ks, matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_sqrt_inv, matrix_ks, &
                             0.0_dp, matrix_x, filter_eps=threshold)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_s_sqrt_inv, &
                             0.0_dp, matrix_x, filter_eps=threshold)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,F12.5)') "HOMO upper bound:       ", e_homo
         WRITE (unit_nr, '(T6,A,1X,F12.5)') "LUMO lower bound:       ", e_lumo
         WRITE (unit_nr, '(T6,A,1X,L12)') "Predicts a gap ?        ", ((e_lumo)-(e_homo)) > 0
      ENDIF

      ! get largest/smallest eigenvalues for scaling
      CALL cp_dbcsr_arnoldi_extremal(matrix_x, max_eig, min_eig, max_iter=max_iter_lanczos, threshold=eps_lanczos, &
                                     converged=converged)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,1X,2F12.5,1X,A,1X,L1)') "Est. extremal eigenvalues", &
         min_eig, max_eig, " converged: ", converged

      eps_max = max_eig
      eps_min = min_eig

      ! scale KS matrix
      CALL cp_dbcsr_scale(matrix_x, -1.0_dp)
      CALL cp_dbcsr_add_on_diag(matrix_x, eps_max)
      CALL cp_dbcsr_scale(matrix_x, 1/(eps_max-eps_min))

      CALL cp_dbcsr_init(matrix_xsq)
      CALL cp_dbcsr_create(matrix_xsq, template=matrix_ks, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(matrix_xsq, matrix_x)

      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_ks, matrix_type=dbcsr_type_no_symmetry)

      ALLOCATE (poly(max_iter))
      ALLOCATE (nu(max_iter))
      ALLOCATE (wu(max_iter))
      ALLOCATE (alpha(max_iter))
      ALLOCATE (X(4))
      ALLOCATE (Y(4))
      ALLOCATE (lambda(4))

! Controls over the non_monotonic bounds, First if low gap, bias slightly
      beta = (eps_max-ABS(e_lumo))/(eps_max-eps_min)
      betaB = (eps_max+ABS(e_homo))/(eps_max-eps_min)

      IF ((beta-betaB) < 0.005_dp) THEN
         beta = beta-0.002_dp
         betaB = betaB+0.002_dp
      ENDIF
! Check if input specifies to use monotonic bounds.
      IF (.NOT. do_non_monotonic) THEN
         beta = 0.0_dp
         betaB = 1.0_dp
      ENDIF
! inital SCF cycle has no reliable estimate of homo/lumo, force monotinic bounds.
      IF (e_homo == 0.0_dp) THEN
         beta = 0.0_dp
         BetaB = 1.0_dp
      ENDIF

      ! init to take true branch first
      trace_fx = nelectron
      trace_gx = 0

      DO i = 1, max_iter
         t1 = m_walltime()
         flop1 = 0; flop2 = 0

         IF (ABS(trace_fx-nelectron) <= ABS(trace_gx-nelectron)) THEN
! Xn+1 = (aX+ (1-a)I)^2
            poly(i) = 1.0_dp
            alpha(i) = 2.0_dp/(2.0_dp-beta)

            CALL cp_dbcsr_scale(matrix_x, alpha(i))
            CALL cp_dbcsr_add_on_diag(matrix_x, 1.0_dp-alpha(i))
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_x, &
                                   0.0_dp, matrix_xsq, &
                                   filter_eps=threshold, flop=flop1)

!save X for control variables
            CALL cp_dbcsr_copy(matrix_tmp, matrix_x)

            CALL cp_dbcsr_copy(matrix_x, matrix_xsq)

            beta = (1.0_dp-alpha(i))+alpha(i)*beta
            beta = beta*beta
            betaB = (1.0_dp-alpha(i))+alpha(i)*betaB
            betaB = betaB*betaB
         ELSE
! Xn+1 = 2aX-a^2*X^2
            poly(i) = 0.0_dp
            alpha(i) = 2.0_dp/(1.0_dp+betaB)

            CALL cp_dbcsr_scale(matrix_x, alpha(i))
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_x, &
                                   0.0_dp, matrix_xsq, &
                                   filter_eps=threshold, flop=flop1)

!save X for control variables
            CALL cp_dbcsr_copy(matrix_tmp, matrix_x)
!
            CALL cp_dbcsr_add(matrix_x, matrix_xsq, 2.0_dp, -1.0_dp)

            beta = alpha(i)*beta
            beta = 2.0_dp*beta-beta*beta
            betaB = alpha(i)*betaB
            betaB = 2.0_dp*betaB-betaB*betaB

         ENDIF
         occ_matrix = cp_dbcsr_get_occupation(matrix_x)
         t2 = m_walltime()
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, &
                   '(T6,A,I3,1X,F10.8,E12.3,F12.3,F13.3,E12.3)') "TC2 it ", &
               i, occ_matrix, t2-t1, &
               (flop1+flop2)/(1.0E6_dp*(t2-t1)), threshold
            CALL m_flush(unit_nr)
         ENDIF

! calculate control terms
         CALL cp_dbcsr_trace(matrix_xsq, trace_fx)

! intermediate use matrix_xsq compute X- X*X , temorarily use trace_gx
         CALL cp_dbcsr_add(matrix_xsq, matrix_tmp, -1.0_dp, 1.0_dp)
         CALL cp_dbcsr_trace(matrix_xsq, trace_gx)
         nu(i) = cp_dbcsr_frobenius_norm(matrix_xsq)
         wu(i) = trace_gx

! intermediate use matrix_xsq to compute = 2X - X*X
         CALL cp_dbcsr_add(matrix_xsq, matrix_tmp, 1.0_dp, 1.0_dp)
         CALL cp_dbcsr_trace(matrix_xsq, trace_gx)
! TC2 has quadratic convergence, using the frobeniums norm as an idempotency deviation test.
         IF (ABS(nu(i)) < (threshold)) EXIT
      END DO

      occ_matrix = cp_dbcsr_get_occupation(matrix_x)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,I3,1X,1F10.8,1X,1F10.8)') 'Final TC2 iteration  ', i, occ_matrix, ABS(nu(i))

      CALL cp_dbcsr_release(matrix_xsq)
      CALL cp_dbcsr_release(matrix_tmp)

      ! output to matrix_p, P = inv(S)^0.5 X inv(S)^0.5
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_x, matrix_s_sqrt_inv, &
                             0.0_dp, matrix_p, filter_eps=threshold)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_sqrt_inv, matrix_p, &
                             0.0_dp, matrix_p, filter_eps=threshold)

      ! ALGO 3 from. SIAM DOI. 10.1137/130911585
      X(1) = 1.0_dp
      X(2) = 1.0_dp
      X(3) = 0.0_dp
      X(4) = 0.0_dp
      gama = 6.0_dp-4.0_dp*(SQRT(2.0_dp))
      gama = gama-gama*gama
      DO WHILE (nu(i) < gama)
         Y(1) = 0.5_dp*(1.0_dp-SQRT(1.0_dp-4.0_dp*nu(i)*nu(i)/wu(i)))
         Y(2) = 0.5_dp*(1.0_dp-SQRT(1.0_dp-4.0_dp*nu(i)))
         Y(3) = 0.5_dp*(1.0_dp+SQRT(1.0_dp-4.0_dp*nu(i)))
         Y(4) = 0.5_dp*(1.0_dp+SQRT(1.0_dp-4.0_dp*nu(i)*nu(i)/wu(i)))
         Y(:) = MIN(1.0_dp, MAX(0.0_dp, Y(:)))
         DO j = i, 1, -1
            IF (poly(j) == 1.0_dp) THEN
               DO k = 1, 4
                  Y(k) = SQRT(Y(k))
                  Y(k) = (Y(k)-1.0_dp+alpha(j))/alpha(j)
               ENDDO ! end K
            ELSE
               DO k = 1, 4
                  Y(k) = 1.0_dp-SQRT(1.0_dp-Y(k))
                  Y(k) = Y(k)/alpha(j)
               ENDDO ! end K
            ENDIF ! end poly
         ENDDO ! end j
         X(1) = MIN(X(1), Y(1))
         X(2) = MIN(X(2), Y(2))
         X(3) = MAX(X(3), Y(3))
         X(4) = MAX(X(4), Y(4))
         i = i-1
      ENDDO ! end i
!   lambda 1,2,3,4 are:: out lumo, in lumo, in homo, out homo
      DO k = 1, 4
         lambda(k) = eps_max-(eps_max-eps_min)*X(k)
      ENDDO ! end k
! END  ALGO 3 from. SIAM DOI. 10.1137/130911585
      e_homo = lambda(4)
      e_lumo = lambda(1)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,3E12.4)') "outer homo/lumo/gap", e_homo, e_lumo, (e_lumo-e_homo)
      IF (unit_nr > 0) WRITE (unit_nr, '(T6,A,3E12.4)') "inner homo/lumo/gap", lambda(3), lambda(2), (lambda(2)-lambda(3))

      DEALLOCATE (poly)
      DEALLOCATE (nu)
      DEALLOCATE (wu)
      DEALLOCATE (alpha)
      DEALLOCATE (X)
      DEALLOCATE (Y)
      DEALLOCATE (lambda)

      CALL cp_dbcsr_release(matrix_x)
      CALL timestop(handle)

   END SUBROUTINE density_matrix_tc2

! **************************************************************************************************
!> \brief compute the homo and lumo given a KS matrix and a density matrix in the orthonormalized basis
!>        and the eps_min and eps_max, min and max eigenvalue of the ks matrix
!> \param matrix_k ...
!> \param matrix_p ...
!> \param eps_min ...
!> \param eps_max ...
!> \param threshold ...
!> \param max_iter_lanczos ...
!> \param eps_lanczos ...
!> \param homo ...
!> \param lumo ...
!> \param unit_nr ...
!> \par History
!>       2012.06 created [Florian Thoele]
!> \author Florian Thoele
! **************************************************************************************************
   SUBROUTINE compute_homo_lumo(matrix_k, matrix_p, eps_min, eps_max, threshold, max_iter_lanczos, eps_lanczos, homo, lumo, unit_nr)
      TYPE(cp_dbcsr_type)                                :: matrix_k, matrix_p
      REAL(KIND=dp)                                      :: eps_min, eps_max, threshold
      INTEGER, INTENT(IN)                                :: max_iter_lanczos
      REAL(KIND=dp), INTENT(IN)                          :: eps_lanczos
      REAL(KIND=dp)                                      :: homo, lumo
      INTEGER                                            :: unit_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_homo_lumo', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: converged
      REAL(KIND=dp)                                      :: max_eig, min_eig, shift1, shift2
      TYPE(cp_dbcsr_type)                                :: tmp1, tmp2, tmp3

! temporary matrices used for HOMO/LUMO calculation

      CALL cp_dbcsr_init(tmp1)
      CALL cp_dbcsr_create(tmp1, template=matrix_k, matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=matrix_k, matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init(tmp3)
      CALL cp_dbcsr_create(tmp3, template=matrix_k, matrix_type=dbcsr_type_no_symmetry)

      shift1 = -eps_min
      shift2 = eps_max

      ! find largest ev of P*(K+shift*1), where shift is the neg. val. of the smallest ev of K
      CALL cp_dbcsr_copy(tmp2, matrix_k)
      CALL cp_dbcsr_add_on_diag(tmp2, shift1)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p, tmp2, &
                             0.0_dp, tmp1, filter_eps=threshold)
      CALL cp_dbcsr_arnoldi_extremal(tmp1, max_eig, min_eig, converged=converged, &
                                     threshold=eps_lanczos, max_iter=max_iter_lanczos)
      homo = max_eig-shift1
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,L12)') "Lanczos converged:      ", converged
      ENDIF

      ! -(1-P)*(K-shift*1) = (1-P)*(shift*1 - K), where shift is the largest ev of K
      CALL cp_dbcsr_copy(tmp3, matrix_p)
      CALL cp_dbcsr_scale(tmp3, -1.0_dp)
      CALL cp_dbcsr_add_on_diag(tmp3, 1.0_dp) !tmp3 = 1-P
      CALL cp_dbcsr_copy(tmp2, matrix_k)
      CALL cp_dbcsr_add_on_diag(tmp2, -shift2)
      CALL cp_dbcsr_multiply("N", "N", -1.0_dp, tmp3, tmp2, &
                             0.0_dp, tmp1, filter_eps=threshold)
      CALL cp_dbcsr_arnoldi_extremal(tmp1, max_eig, min_eig, converged=converged, &
                                     threshold=eps_lanczos, max_iter=max_iter_lanczos)
      lumo = -max_eig+shift2

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,L12)') "Lanczos converged:      ", converged
         WRITE (unit_nr, '(T6,A,1X,3F12.5)') 'HOMO/LUMO/gap', homo, lumo, lumo-homo
      ENDIF
      CALL cp_dbcsr_release(tmp1)
      CALL cp_dbcsr_release(tmp2)
      CALL cp_dbcsr_release(tmp3)

   END SUBROUTINE compute_homo_lumo

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param gamma_values ...
!> \param i ...
!> \retval xr ...
! **************************************************************************************************
   FUNCTION evaluate_trs4_polynomial(x, gamma_values, i) RESULT(xr)
      REAL(KIND=dp)                                      :: x
      REAL(KIND=dp), DIMENSION(:)                        :: gamma_values
      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: xr

      REAL(KIND=dp), PARAMETER                           :: gam_max = 6.0_dp, gam_min = 0.0_dp

      INTEGER                                            :: k

      xr = x
      DO k = 1, i
         IF (gamma_values(k) > gam_max) THEN
            xr = 2*xr-xr**2
         ELSE IF (gamma_values(k) < gam_min) THEN
            xr = xr**2
         ELSE
            xr = (xr*xr)*(4*xr-3*xr*xr)+gamma_values(k)*xr*xr*((1-xr)**2)
         ENDIF
      ENDDO
   END FUNCTION evaluate_trs4_polynomial

! **************************************************************************************************
!> \brief ...
!> \param homo ...
!> \param lumo ...
!> \param threshold ...
!> \retval steps ...
! **************************************************************************************************
   FUNCTION estimate_steps(homo, lumo, threshold) RESULT(steps)
      REAL(KIND=dp)                                      :: homo, lumo, threshold
      INTEGER                                            :: steps

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: h, l, m

      l = lumo
      h = homo

      DO i = 1, 200
         IF (ABS(l) < threshold .AND. ABS(1-h) < threshold) EXIT
         m = 0.5_dp*(h+l)
         IF (m > 0.5_dp) THEN
            h = h**2
            l = l**2
         ELSE
            h = 2*h-h**2
            l = 2*l-l**2
         ENDIF
      ENDDO
      steps = i-1
   END FUNCTION estimate_steps

END MODULE dm_ls_scf_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods used by pao_main.F
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_methods
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_binary_read, cp_dbcsr_checksum, cp_dbcsr_col_block_sizes, &
        cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_frobenius_norm, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, &
        cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_reserve_diag_blocks, cp_dbcsr_row_block_sizes, cp_dbcsr_scale, &
        cp_dbcsr_set, cp_dbcsr_trace, cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dm_ls_scf_methods,               ONLY: density_matrix_trs4,&
                                              ls_scf_init_matrix_S
   USE dm_ls_scf_qs,                    ONLY: matrix_qs_to_ls
   USE dm_ls_scf_types,                 ONLY: ls_mstruct_type,&
                                              ls_scf_env_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE machine,                         ONLY: m_walltime
   USE mathlib,                         ONLY: diamat_all
   USE message_passing,                 ONLY: mp_max
   USE pao_param,                       ONLY: pao_param_count,&
                                              pao_update_AB
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_methods'

   PUBLIC :: pao_print_atom_info
   PUBLIC :: pao_build_orthogonalizer, pao_build_selector, pao_build_matrix_X
   PUBLIC :: pao_test_convergence
   PUBLIC :: pao_calc_energy, pao_check_trace_ps
   PUBLIC :: pao_read_preopt_dm
   PUBLIC :: pao_calc_outer_grad_lnv
   PUBLIC :: pao_check_grad

CONTAINS

! **************************************************************************************************
!> \brief Prints a one line summary for each atom.
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_print_atom_info(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_print_atom_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, natoms
      INTEGER, DIMENSION(:), POINTER                     :: pao_basis, param_cols, param_rows, &
                                                            pri_basis

      pri_basis => cp_dbcsr_row_block_sizes(pao%matrix_Y)
      pao_basis => cp_dbcsr_col_block_sizes(pao%matrix_Y)

      IF(.NOT.(SIZE(pao_basis) == SIZE(pri_basis)))CALL cp__a("pao_methods.F",85)
      natoms = SIZE(pao_basis)

      param_rows => cp_dbcsr_row_block_sizes(pao%matrix_X)
      param_cols => cp_dbcsr_col_block_sizes(pao%matrix_X)
      IF(.NOT.(SIZE(param_rows) == natoms .AND. SIZE(param_cols) == natoms))CALL cp__a("pao_methods.F",90)

      DO iatom = 1, natoms
         IF (pao%iw > 0) WRITE (pao%iw, "(A,I7,T20,A,I3,T45,A,I3,T65,A,I3)") &
            " PAO| atom: ", iatom, &
            " prim_basis: ", pri_basis(iatom), &
            " pao_basis: ", pao_basis(iatom), &
            " pao_params: ", (param_cols(iatom)*param_rows(iatom))
      ENDDO
   END SUBROUTINE pao_print_atom_info

! **************************************************************************************************
!> \brief Constructs matrix_N and its inverse.
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_build_orthogonalizer(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_build_orthogonalizer', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, handle, i, iatom, j, k, N, &
                                                            natoms
      LOGICAL                                            :: found
      REAL(dp)                                           :: v, w
      REAL(dp), DIMENSION(:), POINTER                    :: evals
      REAL(dp), DIMENSION(:, :), POINTER                 :: A, block_N, block_N_inv, block_S
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      natom=natoms)

      CALL cp_dbcsr_init(pao%matrix_N)
      CALL cp_dbcsr_create(pao%matrix_N, template=matrix_s(1)%matrix, name="PAO matrix_N")
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_N)

      CALL cp_dbcsr_init(pao%matrix_N_inv)
      CALL cp_dbcsr_create(pao%matrix_N_inv, template=matrix_s(1)%matrix, name="PAO matrix_N_inv")
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_N_inv)

      CALL cp_dbcsr_iterator_start(iter, pao%matrix_N)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_N)
         IF (arow /= acol) CALL cp__b("pao_methods.F",139,"got non-diagonal block")
         iatom = arow

         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_N_inv, row=arow, col=acol, block=block_N_inv, found=found)
         IF(.NOT.(ASSOCIATED(block_N_inv)))CALL cp__a("pao_methods.F",143)

         CALL cp_dbcsr_get_block_p(matrix=matrix_s(1)%matrix, row=arow, col=acol, block=block_S, found=found)
         IF(.NOT.(ASSOCIATED(block_S)))CALL cp__a("pao_methods.F",146)

         IF (SIZE(block_S, 1) /= SIZE(block_S, 2)) CALL cp__b("pao_methods.F",148,"Expected symmetric block")
         N = SIZE(block_S, 1)
         ALLOCATE (A(N, N), evals(N))

         ! take square root of atomic overlap matrix
         A(:, :) = block_S(:, :)
         CALL diamat_all(A, evals) !afterwards A contains the eigenvectors
         DO k = 1, N
            ! NOTE: To maintain a consistent notation with the Berghold paper,
            ! the "_inv" is swapped: N^{-1}=sqrt(S); N=sqrt(S)^{-1}
            w = 1.0_dp/SQRT(evals(k))
            v = SQRT(evals(k))
            DO i = 1, N
               DO j = 1, N
                  block_N(i, j) = block_N(i, j)+w*A(i, k)*A(j, k)
                  block_N_inv(i, j) = block_N_inv(i, j)+v*A(i, k)*A(j, k)
               ENDDO
            ENDDO
         ENDDO
         DEALLOCATE (A, evals)
      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL timestop(handle)
   END SUBROUTINE pao_build_orthogonalizer

! **************************************************************************************************
!> \brief Build rectangular matrix to converert between primary and PAO basis.
!> \param qs_env ...
!> \param matrix_Y ...
! **************************************************************************************************
   SUBROUTINE pao_build_selector(qs_env, matrix_Y)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_type)                                :: matrix_Y

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_build_selector', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acol, arow, handle, i, ikind, nblocks, &
                                                            pao_basis_size
      INTEGER, DIMENSION(:), POINTER                     :: blk_sizes_aux, blk_sizes_pri
      REAL(dp), DIMENSION(:, :), POINTER                 :: block
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (ks_env, matrix_s)
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      matrix_s=matrix_s, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      blk_sizes_pri => cp_dbcsr_col_block_sizes(matrix_s(1)%matrix)
      nblocks = SIZE(blk_sizes_pri)
      IF (nblocks /= SIZE(particle_set, 1)) CALL cp__b("pao_methods.F",207,"nblocks wrong")

      ALLOCATE (blk_sizes_aux(nblocks))
      DO i = 1, nblocks
         CALL get_atomic_kind(particle_set(i)%atomic_kind, kind_number=ikind)
         CALL get_qs_kind(qs_kind_set(ikind), pao_basis_size=pao_basis_size)
         IF (pao_basis_size < 1) CALL cp__b("pao_methods.F",213,"PAO_BASIS_SIZE too small.")
         blk_sizes_aux(i) = pao_basis_size
      ENDDO

      CALL cp_dbcsr_init(matrix_Y)
      CALL cp_dbcsr_create(matrix_Y, &
                           template=matrix_s(1)%matrix, &
                           matrix_type="N", &
                           row_blk_size=blk_sizes_pri, &
                           col_blk_size=blk_sizes_aux, &
                           name="PAO matrix_Y")
      DEALLOCATE (blk_sizes_aux)

      CALL cp_dbcsr_reserve_diag_blocks(matrix_Y)

      CALL cp_dbcsr_iterator_start(iter, matrix_Y)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block)
         IF (SIZE(block, 2) > SIZE(block, 1)) &
            CALL cp__b("pao_methods.F",232,"PAO basis size exceeds primary basis size.")
         block = 0.0_dp
         DO i = 1, SIZE(block, 2)
            block(i, i) = 1.0_dp
         ENDDO
      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL timestop(handle)
   END SUBROUTINE pao_build_selector

! **************************************************************************************************
!> \brief Creates the matrix_X
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_build_matrix_X(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_build_matrix_X', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ikind, natoms
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL get_qs_env(qs_env, &
                      natom=natoms, &
                      matrix_s=matrix_s, &
                      particle_set=particle_set)

      ALLOCATE (row_blk_size(natoms), col_blk_size(natoms))
      col_blk_size = 1
      DO iatom = 1, natoms
         CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
         CALL pao_param_count(pao, qs_env, ikind, nparams=row_blk_size(iatom))
      ENDDO

      CALL cp_dbcsr_init(pao%matrix_X)
      CALL cp_dbcsr_create(pao%matrix_X, &
                           template=matrix_s(1)%matrix, &
                           matrix_type="N", &
                           row_blk_size=row_blk_size, &
                           col_blk_size=col_blk_size)
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_X)
      CALL cp_dbcsr_set(pao%matrix_X, 0.0_dp)

      DEALLOCATE (row_blk_size, col_blk_size)

   END SUBROUTINE pao_build_matrix_X

! **************************************************************************************************
!> \brief Test whether the PAO optimization has reached convergence
!> \param pao ...
!> \param ls_scf_env ...
!> \param new_energy ...
!> \param is_converged ...
! **************************************************************************************************
   SUBROUTINE pao_test_convergence(pao, ls_scf_env, new_energy, is_converged)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      REAL(KIND=dp), INTENT(IN)                          :: new_energy
      LOGICAL, INTENT(OUT)                               :: is_converged

      REAL(KIND=dp)                                      :: energy_diff, loop_eps, norm_G, now, &
                                                            time_diff

      is_converged = .FALSE.

      ! calculate progress
      energy_diff = new_energy-pao%energy_prev
      pao%energy_prev = new_energy
      now = m_walltime()
      time_diff = now-pao%step_start_time
      pao%step_start_time = now

      IF (pao%precondition) THEN
         norm_G = cp_dbcsr_frobenius_norm(pao%matrix_G_preconed)
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| norm of preconditioned gradient:", norm_G
      ELSE
         norm_G = cp_dbcsr_frobenius_norm(pao%matrix_G)
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| norm of gradient:", norm_G
      ENDIF

      IF (pao%istep > 1) THEN
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| energy improvement:", energy_diff
         ! IF(energy_diff>0.0_dp) CALL cp__w("pao_methods.F",320,"PAO| energy increased")

         ! convergence criterion
         loop_eps = norm_G/ls_scf_env%nelectron_total
         is_converged = loop_eps < pao%eps_pao

         ! print one-liner
         IF (pao%iw > 0) WRITE (pao%iw, '(A,I6,11X,F20.9,1X,E10.3,1X,E10.3,1X,F9.3)') &
            " PAO| step ", &
            pao%istep, &
            new_energy, &
            loop_eps, &
            pao%linesearch%step_size, & !prev step, which let to the current energy
            time_diff
      ENDIF
   END SUBROUTINE pao_test_convergence

! **************************************************************************************************
!> \brief Calculate the pao energy
!> \param pao ...
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param energy ...
! **************************************************************************************************
   SUBROUTINE pao_calc_energy(pao, qs_env, ls_scf_env, energy)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env
      REAL(KIND=dp), INTENT(OUT)                         :: energy

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_calc_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: penalty, trace_PH

      CALL timeset(routineN, handle)

      ! calculate matrix U, which determines the pao basis
      CALL pao_update_AB(pao, ls_scf_env%ls_mstruct, penalty)

      ! calculat S, S_inv, S_sqrt, and S_sqrt_inv in the new pao basis
      CALL pao_rebuild_S(qs_env, ls_scf_env)

      ! calulate the density matrix P in the pao basis
      CALL pao_dm_trs4(qs_env, ls_scf_env)

      ! calculate the energy from the trace(PH) in the pao basis
      energy = 0.0_dp
      DO ispin = 1, ls_scf_env%nspins
         CALL cp_dbcsr_trace(ls_scf_env%matrix_p(ispin), ls_scf_env%matrix_ks(ispin), trace_PH)
         energy = energy+trace_PH
      ENDDO

      ! add penalty term
      energy = energy+penalty

      IF (pao%iw > 0) THEN
         WRITE (pao%iw, *) ""
         WRITE (pao%iw, *) "PAO| energy:", energy, "penalty:", penalty
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE pao_calc_energy

! **************************************************************************************************
!> \brief Ensure that the number of electrons is correct.
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_check_trace_PS(ls_scf_env)
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_check_trace_PS', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: tmp, trace_PS
      TYPE(cp_dbcsr_type)                                :: matrix_S_desym

      CALL timeset(routineN, handle)
      CALL cp_dbcsr_init(matrix_S_desym)
      CALL cp_dbcsr_create(matrix_S_desym, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_desymmetrize(ls_scf_env%matrix_s, matrix_S_desym)

      trace_PS = 0.0_dp
      DO ispin = 1, ls_scf_env%nspins
         CALL cp_dbcsr_trace(ls_scf_env%matrix_p(ispin), matrix_S_desym, tmp)
         trace_PS = trace_PS+tmp
      ENDDO

      CALL cp_dbcsr_release(matrix_S_desym)

      IF (ABS(ls_scf_env%nelectron_total-trace_PS) > 0.5) &
         CALL cp__b("pao_methods.F",412,"Number of electrons wrong. Trace(PS) ="//cp_to_string(trace_PS))

      CALL timestop(handle)
   END SUBROUTINE pao_check_trace_PS

! **************************************************************************************************
!> \brief Read primary density matrix from file.
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_read_preopt_dm(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_read_preopt_dm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: cs_pos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, rho_ao
      TYPE(cp_dbcsr_type)                                :: matrix_tmp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      rho=rho, &
                      energy=energy)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      IF (dft_control%nspins /= 1) CALL cp__b("pao_methods.F",447,"open shell not yet implemented")

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_init(matrix_tmp) ! read into tmp-matrix to preserve sparsity of rho_ao
         CALL cp_dbcsr_binary_read(pao%preopt_dm_file, matrix_new=matrix_tmp, &
                                   distribution=cp_dbcsr_distribution(matrix_s(1)%matrix))
         cs_pos = cp_dbcsr_checksum(matrix_tmp, pos=.TRUE.)
         IF (pao%iw > 0) WRITE (pao%iw, '(T2,A,E20.8)') "Read restart DM "//TRIM(pao%preopt_dm_file)// &
            " with checksum: ", cs_pos
         CALL cp_dbcsr_copy_into_existing(rho_ao(ispin)%matrix, matrix_tmp)
         CALL cp_dbcsr_release(matrix_tmp)
      ENDDO

      ! calculate corresponding ks matrix
      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., &
                               just_energy=.FALSE., print_active=.TRUE.)
      IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| Quickstep energy from restart density:", energy%total

      CALL timestop(handle)

   END SUBROUTINE pao_read_preopt_dm

! **************************************************************************************************
!> \brief Rebuilds S, S_inv, S_sqrt, and S_sqrt_inv in the pao basis
!> \param qs_env ...
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_rebuild_S(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_rebuild_S', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_release(ls_scf_env%matrix_s_inv)
      CALL cp_dbcsr_release(ls_scf_env%matrix_s_sqrt)
      CALL cp_dbcsr_release(ls_scf_env%matrix_s_sqrt_inv)

      CALL get_qs_env(qs_env, matrix_s=matrix_s)
      CALL ls_scf_init_matrix_s(matrix_s(1)%matrix, ls_scf_env)

      CALL timestop(handle)
   END SUBROUTINE pao_rebuild_S

! **************************************************************************************************
!> \brief Calculate density matrix using TRS4 purification
!> \param qs_env ...
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_dm_trs4(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_dm_trs4', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: project_name
      INTEGER                                            :: handle, ispin, nelectron_spin_real, nspin
      LOGICAL                                            :: converged
      REAL(KIND=dp)                                      :: homo_spin, lumo_spin, mu_spin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      project_name = logger%iter_info%project_name
      nspin = ls_scf_env%nspins

      CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
      DO ispin = 1, nspin
         CALL matrix_qs_to_ls(ls_scf_env%matrix_ks(ispin), matrix_ks(ispin)%matrix, &
                              ls_scf_env%ls_mstruct, covariant=.TRUE.)

         nelectron_spin_real = ls_scf_env%nelectron_spin(ispin)
         IF (ls_scf_env%nspins == 1) nelectron_spin_real = nelectron_spin_real/2
         CALL density_matrix_trs4(ls_scf_env%matrix_p(ispin), ls_scf_env%matrix_ks(ispin), &
                                  ls_scf_env%matrix_s_sqrt_inv, &
                                  nelectron_spin_real, ls_scf_env%eps_filter, homo_spin, lumo_spin, mu_spin, &
                                  dynamic_threshold=.FALSE., converged=converged, &
                                  max_iter_lanczos=ls_scf_env%max_iter_lanczos, &
                                  eps_lanczos=ls_scf_env%eps_lanczos)
         IF (.NOT. converged) CALL cp__b("pao_methods.F",533,"TRS4 did not converge")
      ENDDO

      IF (nspin == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(1), 2.0_dp)

      CALL timestop(handle)
   END SUBROUTINE pao_dm_trs4

! **************************************************************************************************
!> \brief Helper routine, calculates partial derivative dE/dU
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param matrix_M ...
! **************************************************************************************************
   SUBROUTINE pao_calc_outer_grad_lnv(qs_env, ls_scf_env, matrix_M)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env
      TYPE(cp_dbcsr_type)                                :: matrix_M

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_calc_outer_grad_lnv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nspin
      REAL(KIND=dp)                                      :: filter_eps
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho_ao
      TYPE(cp_dbcsr_type) :: matrix_HB, matrix_HPS, matrix_M1, matrix_M2, matrix_M3, matrix_NHB, &
         matrix_NHBM2, matrix_NPA, matrix_NPAM1, matrix_NSB, matrix_NSBM3, matrix_PA, matrix_PH, &
         matrix_PHP, matrix_PSP, matrix_SB, matrix_SP
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ls_mstruct_type), POINTER                     :: ls_mstruct
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)

      ls_mstruct => ls_scf_env%ls_mstruct
      pao => ls_scf_env%pao_env

      CALL get_qs_env(qs_env, &
                      rho=rho, &
                      matrix_ks=matrix_ks, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control)
      CALL qs_rho_get(rho, rho_ao=rho_ao)
      nspin = dft_control%nspins
      filter_eps = ls_scf_env%eps_filter

      IF (nspin /= 1) CALL cp__b("pao_methods.F",580,"open shell not yet implemented")
      !TODO: handle openshell case properly

      ! notation according to pao_math_lnv.pdf

      CALL cp_dbcsr_init(matrix_M)
      CALL cp_dbcsr_create(matrix_M, template=matrix_s(1)%matrix, matrix_type="N")

      !---------------------------------------------------------------------------
      ! calculate need products in pao basis
      CALL cp_dbcsr_init(matrix_PH)
      CALL cp_dbcsr_create(matrix_PH, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_p(1), ls_scf_env%matrix_ks(1), &
                             0.0_dp, matrix_PH, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_PHP)
      CALL cp_dbcsr_create(matrix_PHP, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_PH, ls_scf_env%matrix_p(1), &
                             0.0_dp, matrix_PHP, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_SP)
      CALL cp_dbcsr_create(matrix_SP, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s, ls_scf_env%matrix_p(1), &
                             0.0_dp, matrix_SP, filter_eps=filter_eps)

      IF (SIZE(ls_scf_env%matrix_p) == 1) CALL cp_dbcsr_scale(matrix_SP, 0.5_dp)

      CALL cp_dbcsr_init(matrix_HPS)
      CALL cp_dbcsr_create(matrix_HPS, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, ls_scf_env%matrix_ks(1), matrix_SP, &
                             0.0_dp, matrix_HPS, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_PSP)
      CALL cp_dbcsr_create(matrix_PSP, template=ls_scf_env%matrix_s, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_p(1), matrix_SP, &
                             0.0_dp, matrix_PSP, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! M1 = dE_lnv / dP_pao
      CALL cp_dbcsr_init(matrix_M1)
      CALL cp_dbcsr_create(matrix_M1, template=ls_scf_env%matrix_s, matrix_type="N")

      CALL cp_dbcsr_multiply("N", "T", 3.0_dp, ls_scf_env%matrix_ks(1), matrix_SP, &
                             1.0_dp, matrix_M1, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "N", 3.0_dp, matrix_SP, ls_scf_env%matrix_ks(1), &
                             1.0_dp, matrix_M1, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", -2.0_dp, matrix_HPS, matrix_SP, &
                             1.0_dp, matrix_M1, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_SP, matrix_HPS, &
                             1.0_dp, matrix_M1, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", -2.0_dp, matrix_SP, matrix_HPS, &
                             1.0_dp, matrix_M1, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! M2 = dE_lnv / dH
      CALL cp_dbcsr_init(matrix_M2)
      CALL cp_dbcsr_create(matrix_M2, template=ls_scf_env%matrix_s, matrix_type="N")

      CALL cp_dbcsr_add(matrix_M2, matrix_PSP, 1.0_dp, 3.0_dp)

      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_PSP, matrix_SP, &
                             1.0_dp, matrix_M2, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! M3 = dE_lnv / dS
      CALL cp_dbcsr_init(matrix_M3)
      CALL cp_dbcsr_create(matrix_M3, template=ls_scf_env%matrix_s, matrix_type="N")

      CALL cp_dbcsr_add(matrix_M3, matrix_PHP, 1.0_dp, 3.0_dp)

      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_PHP, matrix_SP, &
                             1.0_dp, matrix_M3, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", -2.0_dp, matrix_PSP, matrix_PH, &
                             1.0_dp, matrix_M3, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! combine M1 with matrices from primary basis
      CALL cp_dbcsr_init(matrix_PA)
      CALL cp_dbcsr_create(matrix_PA, template=ls_mstruct%matrix_A, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, rho_ao(1)%matrix, ls_mstruct%matrix_A, &
                             0.0_dp, matrix_PA, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_NPA)
      CALL cp_dbcsr_create(matrix_NPA, template=ls_mstruct%matrix_A, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_N_inv, matrix_PA, &
                             0.0_dp, matrix_NPA, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_NPAM1)
      CALL cp_dbcsr_create(matrix_NPAM1, template=ls_mstruct%matrix_A, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_NPA, matrix_M1, &
                             0.0_dp, matrix_NPAM1, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_NPAM1, pao%matrix_Y, &
                             1.0_dp, matrix_M, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! combine M2 with matrices from primary basis
      CALL cp_dbcsr_init(matrix_HB)
      CALL cp_dbcsr_create(matrix_HB, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_ks(1)%matrix, ls_mstruct%matrix_B, &
                             0.0_dp, matrix_HB, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_NHB)
      CALL cp_dbcsr_create(matrix_NHB, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_N, matrix_HB, &
                             0.0_dp, matrix_NHB, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_NHBM2)
      CALL cp_dbcsr_create(matrix_NHBM2, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_NHB, matrix_M2, &
                             0.0_dp, matrix_NHBM2, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_NHBM2, pao%matrix_Y, &
                             1.0_dp, matrix_M, filter_eps=filter_eps)

      !---------------------------------------------------------------------------
      ! combine M3 with matrices from primary basis
      CALL cp_dbcsr_init(matrix_SB)
      CALL cp_dbcsr_create(matrix_SB, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s(1)%matrix, ls_mstruct%matrix_B, &
                             0.0_dp, matrix_SB, filter_eps=filter_eps)

      IF (SIZE(ls_scf_env%matrix_p) == 1) CALL cp_dbcsr_scale(matrix_SB, 0.5_dp)

      CALL cp_dbcsr_init(matrix_NSB)
      CALL cp_dbcsr_create(matrix_NSB, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_N, matrix_SB, &
                             0.0_dp, matrix_NSB, filter_eps=filter_eps)

      CALL cp_dbcsr_init(matrix_NSBM3)
      CALL cp_dbcsr_create(matrix_NSBM3, template=ls_mstruct%matrix_B, matrix_type="N")
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_NSB, matrix_M3, &
                             0.0_dp, matrix_NSBM3, filter_eps=filter_eps)

      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_NSBM3, pao%matrix_Y, &
                             1.0_dp, matrix_M, filter_eps=filter_eps)

      IF (SIZE(ls_scf_env%matrix_p) == 1) CALL cp_dbcsr_scale(matrix_M, 2.0_dp)

      !---------------------------------------------------------------------------
      ! cleanup: TODO release matrices as early as possible
      CALL cp_dbcsr_release(matrix_PH)
      CALL cp_dbcsr_release(matrix_PHP)
      CALL cp_dbcsr_release(matrix_SP)
      CALL cp_dbcsr_release(matrix_HPS)
      CALL cp_dbcsr_release(matrix_PSP)
      CALL cp_dbcsr_release(matrix_M1)
      CALL cp_dbcsr_release(matrix_M2)
      CALL cp_dbcsr_release(matrix_M3)
      CALL cp_dbcsr_release(matrix_PA)
      CALL cp_dbcsr_release(matrix_NPA)
      CALL cp_dbcsr_release(matrix_NPAM1)
      CALL cp_dbcsr_release(matrix_HB)
      CALL cp_dbcsr_release(matrix_NHB)
      CALL cp_dbcsr_release(matrix_NHBM2)
      CALL cp_dbcsr_release(matrix_SB)
      CALL cp_dbcsr_release(matrix_NSB)
      CALL cp_dbcsr_release(matrix_NSBM3)

      CALL timestop(handle)
   END SUBROUTINE pao_calc_outer_grad_lnv

! **************************************************************************************************
!> \brief Debugging routine for checking the analytic gradient.
!> \param pao ...
!> \param qs_env ...
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_check_grad(pao, qs_env, ls_scf_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_check_grad', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, j, natoms
      INTEGER, DIMENSION(:), POINTER                     :: blk_sizes_col, blk_sizes_row
      LOGICAL                                            :: found
      REAL(dp)                                           :: delta, delta_max, eps, Gij_num
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_G, block_X
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(ls_mstruct_type), POINTER                     :: ls_mstruct

      IF (pao%check_grad_tol < 0.0_dp) RETURN ! no checking

      CALL timeset(routineN, handle)

      ls_mstruct => ls_scf_env%ls_mstruct

      CALL get_qs_env(qs_env, para_env=para_env, natom=natoms)

      eps = pao%num_grad_eps
      delta_max = 0.0_dp

      blk_sizes_col => cp_dbcsr_col_block_sizes(pao%matrix_X)
      blk_sizes_row => cp_dbcsr_row_block_sizes(pao%matrix_X)

      ! can not use an iterator here, because other DBCSR routines are called within loop.
      DO iatom = 1, natoms
         IF (pao%iw > 0) WRITE (pao%iw, *) 'PAO| checking gradient of atom ', iatom
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_X, row=iatom, col=iatom, block=block_X, found=found)

         IF (ASSOCIATED(block_X)) THEN !only one node actually has the block
            CALL cp_dbcsr_get_block_p(matrix=pao%matrix_G, row=iatom, col=iatom, block=block_G, found=found)
            IF(.NOT.(ASSOCIATED(block_G)))CALL cp__a("pao_methods.F",789)
         ENDIF

         DO i = 1, blk_sizes_row(iatom)
            DO j = 1, blk_sizes_col(iatom)
               SELECT CASE (pao%num_grad_order)
               CASE (2) ! calculate derivative to 2th order
                  Gij_num = -eval_point(block_X, i, j, -eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num+eval_point(block_X, i, j, +eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num/(2.0_dp*eps)

               CASE (4) ! calculate derivative to 4th order
                  Gij_num = eval_point(block_X, i, j, -2_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num-8_dp*eval_point(block_X, i, j, -1_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num+8_dp*eval_point(block_X, i, j, +1_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num-eval_point(block_X, i, j, +2_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num/(12.0_dp*eps)

               CASE (6) ! calculate derivative to 6th order
                  Gij_num = -1_dp*eval_point(block_X, i, j, -3_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num+9_dp*eval_point(block_X, i, j, -2_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num-45_dp*eval_point(block_X, i, j, -1_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num+45_dp*eval_point(block_X, i, j, +1_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num-9_dp*eval_point(block_X, i, j, +2_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num+1_dp*eval_point(block_X, i, j, +3_dp*eps, pao, ls_scf_env, qs_env)
                  Gij_num = Gij_num/(60.0_dp*eps)

               CASE DEFAULT
                  CALL cp__b("pao_methods.F",817,"Unsupported numerical derivative order: "//cp_to_string(pao%num_grad_order))
               END SELECT

               IF (ASSOCIATED(block_X)) THEN
                  delta = ABS(Gij_num-block_G(i, j))
                  delta_max = MAX(delta_max, delta)
                  !WRITE (*,*) "gradient check", iatom, i, j, Gij_num, block_G(i,j), delta
               ENDIF
            ENDDO
         ENDDO
      END DO

      CALL mp_max(delta_max, para_env%group)
      IF (pao%iw > 0) WRITE (pao%iw, *) 'PAO| checked gradient, max delta:', delta_max
      IF (delta_max > pao%check_grad_tol) CALL cp_abort(cp__l("pao_methods.F",831), &
                                                        "Analytic and numeric gradients differ too much:"//cp_to_string(delta_max))

      CALL timestop(handle)
   END SUBROUTINE pao_check_grad

! **************************************************************************************************
!> \brief Helper routine for pao_check_grad()
!> \param block_X ...
!> \param i ...
!> \param j ...
!> \param eps ...
!> \param pao ...
!> \param ls_scf_env ...
!> \param qs_env ...
!> \retval energy ...
! **************************************************************************************************
   FUNCTION eval_point(block_X, i, j, eps, pao, ls_scf_env, qs_env) RESULT(energy)
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_X
      INTEGER, INTENT(IN)                                :: i, j
      REAL(dp), INTENT(IN)                               :: eps
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp)                                           :: energy

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_point', routineP = moduleN//':'//routineN

      REAL(dp)                                           :: old_Xij

      IF (ASSOCIATED(block_X)) THEN
         old_Xij = block_X(i, j) ! backup old block_X
         block_X(i, j) = block_X(i, j)+eps ! add pertubation
      ENDIF

      ! calculate energy
      CALL pao_calc_energy(pao, qs_env, ls_scf_env, energy)

      ! restore old block_X
      IF (ASSOCIATED(block_X)) THEN
         block_X(i, j) = old_Xij
      ENDIF

   END FUNCTION eval_point

END MODULE pao_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gaussian Process implementation
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_ml_gaussprocess
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: invmat
   USE pao_types,                       ONLY: pao_env_type,&
                                              training_matrix_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_ml_gaussprocess'

   PUBLIC ::pao_ml_gp_train, pao_ml_gp_predict

CONTAINS

! **************************************************************************************************
!> \brief Builds the covariance matrix
!> \param pao ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE pao_ml_gp_train(pao, training_matrices)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(training_matrix_type), DIMENSION(:), TARGET   :: training_matrices

      INTEGER                                            :: i, ikind, info, j, npoints
      REAL(dp), DIMENSION(:), POINTER                    :: idescr, jdescr
      TYPE(training_matrix_type), POINTER                :: training_matrix

      ! TODO this could be parallelized over ranks
      DO ikind = 1, SIZE(training_matrices)
         training_matrix => training_matrices(ikind)
         npoints = SIZE(training_matrix%inputs, 1) ! number of points
         IF(.NOT.(SIZE(training_matrix%outputs, 1) == npoints))CALL cp__a("pao_ml_gaussprocess.F",44)
         IF(.NOT.(npoints > 0))CALL cp__a("pao_ml_gaussprocess.F",45)

         !TODO proper output
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO|ML| Building covariance matrix for kind: ", &
            TRIM(training_matrix%kindname), " from ", npoints, "training points."

         ! build co-variance matrix A
         ALLOCATE (training_matrix%GP(npoints, npoints))
         DO i = 1, npoints
         DO j = 1, npoints
            idescr => training_matrix%inputs(i, :)
            jdescr => training_matrix%inputs(j, :)
            training_matrix%GP(i, j) = kernel(pao%gp_scale, idescr, jdescr)
         ENDDO
         ENDDO

         ! add noise of training data
         DO i = 1, npoints
            training_matrix%GP(i, i) = training_matrix%GP(i, i)+pao%gp_noise_var**2
         ENDDO

         ! invert covariance matrix
         CALL invmat(training_matrix%GP, info) !TODO use upper triangular matrix and invmat_symm
         IF(.NOT.(info == 0))CALL cp__a("pao_ml_gaussprocess.F",68)

         ! TODO print min eigenvalue to estimate reduandency of training data and estimate gp_noise?
      ENDDO

   END SUBROUTINE pao_ml_gp_train

! **************************************************************************************************
!> \brief Uses covariance matrix to make prediction
!> \param pao ...
!> \param training_matrix ...
!> \param descriptor ...
!> \param output ...
!> \param variance ...
! **************************************************************************************************
   SUBROUTINE pao_ml_gp_predict(pao, training_matrix, descriptor, output, variance)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(training_matrix_type), INTENT(IN)             :: training_matrix
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: descriptor
      REAL(dp), DIMENSION(:), INTENT(OUT)                :: output
      REAL(dp), INTENT(OUT)                              :: variance

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_ml_gp_predict', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, npoints
      REAL(dp)                                           :: cov
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: covs, weights

      npoints = SIZE(training_matrix%outputs, 1)

      ! calculate weights
      ALLOCATE (weights(npoints), covs(npoints))
      weights = 0.0_dp
      covs = 0.0_dp
      DO i = 1, npoints
         cov = kernel(pao%gp_scale, descriptor, training_matrix%inputs(i, :))
         covs(i) = cov
         weights(:) = weights(:)+cov*training_matrix%GP(i, :)
      ENDDO

      ! calculate variance
      variance = kernel(pao%gp_scale, descriptor, descriptor)
      DO i = 1, npoints
         cov = kernel(pao%gp_scale, descriptor, training_matrix%inputs(i, :))
         variance = variance-cov*weights(i)
      ENDDO

      ! build output
      output = 0.0_dp
      DO i = 1, npoints
         output(:) = output+weights(i)*training_matrix%outputs(i, :)
      ENDDO

   END SUBROUTINE pao_ml_gp_predict

! **************************************************************************************************
!> \brief Gaussian kernel used to measure covariance between two descriptors.
!> \param scale ...
!> \param descr1 ...
!> \param descr2 ...
!> \retval cov ...
! **************************************************************************************************
   FUNCTION kernel(scale, descr1, descr2) RESULT(dist)
      REAL(dp), INTENT(IN)                               :: scale
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: descr1, descr2
      REAL(dp)                                           :: dist

      REAL(dp)                                           :: fdist2

      fdist2 = SUM(((descr1-descr2)/scale)**2)
      dist = EXP(-fdist2/2.0_dp)
   END FUNCTION kernel

END MODULE pao_ml_gaussprocess
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for encoding and decoding the output presented to the learning algorithm
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_ml_codec
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE kinds,                           ONLY: dp
   USE pao_potentials,                  ONLY: pao_build_V_core,&
                                              pao_build_V_neighbors
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_ml_codec'

   PUBLIC :: pao_ml_encode_xblock, pao_ml_decode_xblock

CONTAINS

! **************************************************************************************************
!> \brief Uses U0 to encode the sign of rotation angles in a contiguous fashion.
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param iatom ...
!> \param u0block ...
!> \param xblock ...
!> \param output ...
! **************************************************************************************************
   SUBROUTINE pao_ml_encode_xblock(particle_set, qs_kind_set, cell, iatom, u0block, xblock, output)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: iatom
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: u0block, xblock
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: output

      INTEGER, PARAMETER                                 :: L = 3

      INTEGER                                            :: a, b, i, ikind, M, N, nparams
      REAL(dp)                                           :: Vab_core, Vab_neighbors
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: V_core, V_neighbors

      IF(.NOT.(.NOT. ALLOCATED(output)))CALL cp__a("pao_ml_codec.F",55)

      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL get_qs_kind(qs_kind_set(ikind), pao_basis_size=M)
      N = SIZE(u0block, 1) ! primary basis-size

      ! build potentials
      CALL pao_build_V_neighbors(particle_set, qs_kind_set, cell, iatom, V_neighbors)
      CALL pao_build_V_core(particle_set, qs_kind_set, iatom, V_core)

      nparams = SIZE(xblock)
      ALLOCATE (output(L*nparams))

      DO i = 1, nparams
         a = MOD(i-1, M)+1
         b = M+(i-1)/M+1

         Vab_neighbors = SUM(MATMUL(MATMUL(TRANSPOSE(u0block(:, a:a)), V_neighbors), u0block(:, b:b)))
         Vab_core = SUM(MATMUL(MATMUL(TRANSPOSE(u0block(:, a:a)), V_core), u0block(:, b:b)))

         output(L*(i-1)+1) = ABS(xblock(i, 1))
         output(L*(i-1)+2) = SIGN(1.0_dp, xblock(i, 1))*Vab_core
         output(L*(i-1)+3) = SIGN(1.0_dp, xblock(i, 1))*Vab_neighbors
      ENDDO

      DEALLOCATE (V_neighbors, V_core)

   END SUBROUTINE pao_ml_encode_xblock

! **************************************************************************************************
!> \brief Uses U0 to deencode the sign of rotation angles
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param iatom ...
!> \param u0block ...
!> \param output ...
!> \param xblock ...
! **************************************************************************************************
   SUBROUTINE pao_ml_decode_xblock(particle_set, qs_kind_set, cell, iatom, u0block, output, xblock)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: iatom
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: u0block
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: output
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: xblock

      INTEGER, PARAMETER                                 :: L = 3

      INTEGER                                            :: a, b, i, ikind, M, N, nparams
      REAL(dp)                                           :: sign_quenched, sign_vote, Vab_core, &
                                                            Vab_neighbors
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: V_core, V_neighbors

      CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
      CALL get_qs_kind(qs_kind_set(ikind), pao_basis_size=M)
      N = SIZE(u0block, 1) ! primary basis-size

      ! build potentials
      CALL pao_build_V_neighbors(particle_set, qs_kind_set, cell, iatom, V_neighbors)
      CALL pao_build_V_core(particle_set, qs_kind_set, iatom, V_core)

      nparams = SIZE(xblock)

      xblock(:, :) = 0.0_dp
      DO i = 1, nparams
         a = MOD(i-1, M)+1
         b = M+(i-1)/M+1

         Vab_neighbors = SUM(MATMUL(MATMUL(TRANSPOSE(u0block(:, a:a)), V_neighbors), u0block(:, b:b)))
         Vab_core = SUM(MATMUL(MATMUL(TRANSPOSE(u0block(:, a:a)), V_core), u0block(:, b:b)))

         sign_vote = Vab_core*output(L*(i-1)+2)+Vab_neighbors*output(L*(i-1)+3)
         sign_quenched = TANH(1e5_dp*sign_vote) !TODO make scaling factor input param
         xblock(i, 1) = sign_quenched*output(L*(i-1)+1)

      ENDDO

      DEALLOCATE (V_neighbors, V_core)
   END SUBROUTINE pao_ml_decode_xblock

END MODULE pao_ml_codec
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Feature vectors for describing chemical environments in a rotationally invariant fashion.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_ml_descriptor
   USE cell_types,                      ONLY: cell_type
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: diamat_all
   USE pao_potentials,                  ONLY: pao_build_V_neighbors
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_ml_descriptor'

   PUBLIC :: pao_ml_calc_descriptor

CONTAINS

! **************************************************************************************************
!> \brief Calulates a descriptor based on the eigenvalues of V_neighbors
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param iatom ...
!> \param descriptor ...
! **************************************************************************************************
   SUBROUTINE pao_ml_calc_descriptor(particle_set, qs_kind_set, cell, iatom, descriptor)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: iatom
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: descriptor

      INTEGER                                            :: N
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: V_neighbors

      IF(.NOT.(.NOT. ALLOCATED(descriptor)))CALL cp__a("pao_ml_descriptor.F",47)

      CALL pao_build_V_neighbors(particle_set, qs_kind_set, cell, iatom, V_neighbors)
      N = SIZE(V_neighbors, 1)
      ALLOCATE (descriptor(N))
      CALL diamat_all(V_neighbors, descriptor)
      DEALLOCATE (V_neighbors)

   END SUBROUTINE pao_ml_calc_descriptor

END MODULE pao_ml_descriptor
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for reading and writing restart files.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_io
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_col_block_sizes,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_row_block_sizes,&
                                              cp_dbcsr_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_max,&
                                              mp_sum
   USE pao_input,                       ONLY: id2str,&
                                              pao_exp_param
   USE pao_param,                       ONLY: pao_param_count
   USE pao_types,                       ONLY: pao_env_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_io'

   PUBLIC :: pao_read_restart, pao_write_restart
   PUBLIC :: pao_read_raw, pao_kinds_ensure_equal
   PUBLIC :: pao_ioblock_type, pao_iokind_type

   ! data types used by pao_read_raw()
   TYPE pao_ioblock_type
      REAL(dp), DIMENSION(:, :), ALLOCATABLE    :: p
   END TYPE pao_ioblock_type

   TYPE pao_iokind_type
      CHARACTER(LEN=default_string_length)     :: name = ""
      INTEGER                                  :: z = -1
      CHARACTER(LEN=default_string_length)     :: prim_basis_name = ""
      INTEGER                                  :: prim_basis_size = -1
      INTEGER                                  :: pao_basis_size = -1
      INTEGER                                  :: pao_pot_maxl = -1
      REAL(dp)                                 :: pao_pot_beta = 0.0_dp
      REAL(dp)                                 :: pao_pot_weight = 0.0_dp
      INTEGER                                  :: pao_pot_max_projector = -1
      INTEGER                                  :: pao_pot_neighbors = -1
      INTEGER                                  :: nparams = -1
   END TYPE pao_iokind_type

   INTEGER, PARAMETER, PRIVATE :: file_format_version = 2

CONTAINS

! **************************************************************************************************
!> \brief Reads restart file
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_read_restart(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_read_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: param
      INTEGER                                            :: iatom, ikind, natoms
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom2kind
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes, row_blk_sizes
      LOGICAL                                            :: found
      REAL(dp)                                           :: diff
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: hmat, positions
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_X, buffer
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pao_ioblock_type), ALLOCATABLE, DIMENSION(:)  :: u0blocks, xblocks
      TYPE(pao_iokind_type), ALLOCATABLE, DIMENSION(:)   :: kinds
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      IF(.NOT.(LEN_TRIM(pao%restart_file) > 0))CALL cp__a("pao_io.F",104)
      IF (pao%iw > 0) WRITE (pao%iw, '(A,A)') " PAO| Reading restart file: ", TRIM(pao%restart_file)

      CALL get_qs_env(qs_env, &
                      para_env=para_env, &
                      natom=natoms, &
                      cell=cell, &
                      particle_set=particle_set)

      ! read and check restart file on first rank only
      IF (para_env%mepos == para_env%source) THEN
         CALL pao_read_raw(pao%restart_file, param, hmat, kinds, atom2kind, positions, xblocks, u0blocks)

         ! check cell
         IF (MAXVAL(ABS(hmat-cell%hmat)) > 1e-10) &
            CALL cp__b("pao_io.F",119,"Restarting from differnt cell")

         ! check parametrization
         IF (TRIM(param) .NE. TRIM(ADJUSTL(id2str(pao%parameterization)))) &
            CALL cp__b("pao_io.F",123,"Restart PAO parametrization does not match")

         ! check kinds
         DO ikind = 1, SIZE(kinds)
            CALL pao_kinds_ensure_equal(pao, qs_env, ikind, kinds(ikind))
         ENDDO

         ! check number of atoms
         IF (SIZE(positions, 1) /= natoms) &
            CALL cp__b("pao_io.F",132,"Number of atoms do not match")

         ! check atom2kind
         DO iatom = 1, natoms
            IF (atom2kind(iatom) /= particle_set(iatom)%atomic_kind%kind_number) &
               CALL cp__b("pao_io.F",137,"Restart atomic kinds do not match.")
         ENDDO

         ! check positions, warning only
         diff = 0.0_dp
         DO iatom = 1, natoms
            diff = MAX(diff, MAXVAL(ABS(positions(iatom, :)-particle_set(iatom)%r)))
         ENDDO
         IF (diff > 1e-10) &
            CALL cp__w("pao_io.F",146,"Restarting from different atom positions")

      ENDIF

      ! scatter xblocks across ranks to fill pao%matrix_X
      ! this could probably be done more efficiently
      row_blk_sizes => cp_dbcsr_row_block_sizes(pao%matrix_X)
      col_blk_sizes => cp_dbcsr_col_block_sizes(pao%matrix_X)
      DO iatom = 1, natoms
         ALLOCATE (buffer(row_blk_sizes(iatom), col_blk_sizes(iatom)))
         IF (para_env%mepos == para_env%source) THEN
            IF(.NOT.(row_blk_sizes(iatom) == SIZE(xblocks(iatom)%p, 1)))CALL cp__a("pao_io.F",157)
            IF(.NOT.(col_blk_sizes(iatom) == SIZE(xblocks(iatom)%p, 2)))CALL cp__a("pao_io.F",158)
            buffer = xblocks(iatom)%p
         ENDIF
         CALL mp_bcast(buffer, para_env%source, para_env%group)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_X, row=iatom, col=iatom, block=block_X, found=found)
         IF (ASSOCIATED(block_X)) &
            block_X = buffer
         DEALLOCATE (buffer)
      ENDDO

      ! ALLOCATABLEs deallocate themselfs

   END SUBROUTINE pao_read_restart

! **************************************************************************************************
!> \brief Reads a restart file into temporary datastructures
!> \param filename ...
!> \param param ...
!> \param hmat ...
!> \param kinds ...
!> \param atom2kind ...
!> \param positions ...
!> \param xblocks ...
!> \param u0blocks ...
! **************************************************************************************************
   SUBROUTINE pao_read_raw(filename, param, hmat, kinds, atom2kind, positions, xblocks, u0blocks)
      CHARACTER(LEN=default_path_length), INTENT(IN)     :: filename
      CHARACTER(LEN=default_string_length), INTENT(OUT)  :: param
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: hmat
      TYPE(pao_iokind_type), ALLOCATABLE, DIMENSION(:)   :: kinds
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom2kind
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: positions
      TYPE(pao_ioblock_type), ALLOCATABLE, DIMENSION(:)  :: xblocks, u0blocks

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_read_raw', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: label, str_in
      INTEGER                                            :: i1, i2, i3, i4, iatom, ikind, n, natoms, &
                                                            nkinds, nparams, u0blocks_read, &
                                                            unit_nr, xblocks_read
      LOGICAL                                            :: check_version
      REAL(dp)                                           :: r1, r2
      REAL(dp), DIMENSION(3)                             :: pos_in
      REAL(dp), DIMENSION(3, 3)                          :: hmat_angstrom

      IF(.NOT.(.NOT. ALLOCATED(hmat)))CALL cp__a("pao_io.F",203)
      IF(.NOT.(.NOT. ALLOCATED(kinds)))CALL cp__a("pao_io.F",204)
      IF(.NOT.(.NOT. ALLOCATED(atom2kind)))CALL cp__a("pao_io.F",205)
      IF(.NOT.(.NOT. ALLOCATED(positions)))CALL cp__a("pao_io.F",206)
      IF(.NOT.(.NOT. ALLOCATED(xblocks)))CALL cp__a("pao_io.F",207)
      IF(.NOT.(.NOT. ALLOCATED(u0blocks)))CALL cp__a("pao_io.F",208)

      natoms = -1
      nkinds = -1
      xblocks_read = 0
      u0blocks_read = 0
      check_version = .FALSE.

      CALL open_file(file_name=filename, file_status="OLD", file_form="FORMATTED", &
                     file_action="READ", unit_number=unit_nr)

      DO WHILE (.TRUE.)
         READ (unit_nr, fmt=*) label
         BACKSPACE (unit_nr)

         IF (TRIM(label) == "Version") THEN
            READ (unit_nr, fmt=*) label, i1
            IF (i1 /= file_format_version) CALL cp__b("pao_io.F",225,"Restart PAO file format version is wrong")
            check_version = .TRUE.

         ELSE IF (TRIM(label) == "Parametrization") THEN
            READ (unit_nr, fmt=*) label, str_in
            param = str_in

         ELSE IF (TRIM(label) == "Cell") THEN
            READ (unit_nr, fmt=*) label, hmat_angstrom
            ALLOCATE (hmat(3, 3))
            hmat(:, :) = hmat_angstrom(:, :)/angstrom

         ELSE IF (TRIM(label) == "Nkinds") THEN
            READ (unit_nr, fmt=*) label, nkinds
            ALLOCATE (kinds(nkinds))

         ELSE IF (TRIM(label) == "Kind") THEN
            READ (unit_nr, fmt=*) label, ikind, str_in, i1
            IF(.NOT.(ALLOCATED(kinds)))CALL cp__a("pao_io.F",243)
            kinds(ikind)%name = str_in
            kinds(ikind)%z = i1

         ELSE IF (TRIM(label) == "PrimBasis") THEN
            READ (unit_nr, fmt=*) label, ikind, i1, str_in
            IF(.NOT.(ALLOCATED(kinds)))CALL cp__a("pao_io.F",249)
            kinds(ikind)%prim_basis_size = i1
            kinds(ikind)%prim_basis_name = str_in

         ELSE IF (TRIM(label) == "PaoBasis") THEN
            READ (unit_nr, fmt=*) label, ikind, i1, i2, i3, i4, r1, r2
            IF(.NOT.(ALLOCATED(kinds)))CALL cp__a("pao_io.F",255)
            kinds(ikind)%pao_basis_size = i1
            kinds(ikind)%pao_pot_maxl = i2
            kinds(ikind)%pao_pot_max_projector = i3
            kinds(ikind)%pao_pot_neighbors = i4
            kinds(ikind)%pao_pot_beta = r1
            kinds(ikind)%pao_pot_weight = r2

         ELSE IF (TRIM(label) == "NParams") THEN
            READ (unit_nr, fmt=*) label, ikind, i1
            IF(.NOT.(ALLOCATED(kinds)))CALL cp__a("pao_io.F",265)
            kinds(ikind)%nparams = i1

         ELSE IF (TRIM(label) == "Natoms") THEN
            READ (unit_nr, fmt=*) label, natoms
            ALLOCATE (positions(natoms, 3), atom2kind(natoms), xblocks(natoms), u0blocks(natoms))
            positions = 0.0_dp; atom2kind = -1

         ELSE IF (TRIM(label) == "Atom") THEN
            READ (unit_nr, fmt=*) label, iatom, str_in, pos_in
            IF(.NOT.(ALLOCATED(kinds)))CALL cp__a("pao_io.F",275)
            DO ikind = 1, nkinds
               IF (TRIM(kinds(ikind)%name) .EQ. TRIM(str_in)) EXIT
            ENDDO
            IF(.NOT.(ALLOCATED(atom2kind) .AND. ALLOCATED(positions)))CALL cp__a("pao_io.F",279)
            atom2kind(iatom) = ikind
            positions(iatom, :) = pos_in/angstrom

         ELSE IF (TRIM(label) == "Xblock") THEN
            READ (unit_nr, fmt=*) label, iatom
            IF(.NOT.(ALLOCATED(kinds) .AND. ALLOCATED(atom2kind)))CALL cp__a("pao_io.F",285)
            ikind = atom2kind(iatom)
            nparams = kinds(ikind)%nparams
            IF(.NOT.(nparams >= 0))CALL cp__a("pao_io.F",288)
            ALLOCATE (xblocks(iatom)%p(nparams, 1))
            BACKSPACE (unit_nr)
            READ (unit_nr, fmt=*) label, iatom, xblocks(iatom)%p
            xblocks_read = xblocks_read+1
            IF(.NOT.(iatom == xblocks_read))CALL cp__a("pao_io.F",293) ! ensure blocks are read in order

         ELSE IF (TRIM(label) == "U0block") THEN
            READ (unit_nr, fmt=*) label, iatom
            IF(.NOT.(ALLOCATED(kinds) .AND. ALLOCATED(atom2kind)))CALL cp__a("pao_io.F",297)
            ikind = atom2kind(iatom)
            N = kinds(ikind)%prim_basis_size
            ALLOCATE (u0blocks(iatom)%p(N, N))
            BACKSPACE (unit_nr)
            READ (unit_nr, fmt=*) label, iatom, u0blocks(iatom)%p
            u0blocks_read = u0blocks_read+1
            IF(.NOT.(iatom == u0blocks_read))CALL cp__a("pao_io.F",304) ! ensure blocks are read in order

         ELSE IF (TRIM(label) == "THE_END") THEN
            EXIT
         ELSE
            !CALL cp__w("pao_io.F",309,"Skipping restart header with label: "//TRIM(label))
            READ (unit_nr, fmt=*) label ! just read again and ignore
         ENDIF
      END DO
      CALL close_file(unit_number=unit_nr)

      IF(.NOT.(check_version))CALL cp__a("pao_io.F",315)
      IF(.NOT.(xblocks_read == natoms))CALL cp__a("pao_io.F",316) ! ensure we read all blocks
      IF(.NOT.(u0blocks_read == 0 .OR. u0blocks_read == natoms))CALL cp__a("pao_io.F",317)
      IF (u0blocks_read == 0) DEALLOCATE (u0blocks)

   END SUBROUTINE pao_read_raw

! **************************************************************************************************
!> \brief Ensure that the kind read from the restart is equal to the kind curretly in use.
!> \param pao ...
!> \param qs_env ...
!> \param ikind ...
!> \param pao_kind ...
! **************************************************************************************************
   SUBROUTINE pao_kinds_ensure_equal(pao, qs_env, ikind, pao_kind)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: ikind
      TYPE(pao_iokind_type), INTENT(IN)                  :: pao_kind

      CHARACTER(LEN=default_string_length)               :: name
      INTEGER                                            :: nparams, pao_basis_size, &
                                                            pao_pot_max_projector, pao_pot_maxl, &
                                                            pao_pot_neighbors, z
      REAL(dp)                                           :: pao_pot_beta, pao_pot_weight
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      IF (ikind > SIZE(atomic_kind_set) .OR. ikind > SIZE(qs_kind_set)) &
         CALL cp__b("pao_io.F",349,"Some kinds are missing.")

      CALL get_atomic_kind(atomic_kind_set(ikind), z=z, name=name)
      CALL get_qs_kind(qs_kind_set(ikind), &
                       basis_set=basis_set, &
                       pao_basis_size=pao_basis_size, &
                       pao_pot_maxl=pao_pot_maxl, &
                       pao_pot_max_projector=pao_pot_max_projector, &
                       pao_pot_neighbors=pao_pot_neighbors, &
                       pao_pot_beta=pao_pot_beta, &
                       pao_pot_weight=pao_pot_weight)
      CALL pao_param_count(pao, qs_env, ikind=ikind, nparams=nparams)

      IF (pao_kind%nparams /= nparams) &
         CALL cp__b("pao_io.F",363,"Number of parameters do not match")
      IF (TRIM(pao_kind%name) .NE. TRIM(name)) &
         CALL cp__b("pao_io.F",365,"Kind names do not match")
      IF (pao_kind%z /= z) &
         CALL cp__b("pao_io.F",367,"Atomic numbers do not match")
      IF (TRIM(pao_kind%prim_basis_name) .NE. TRIM(basis_set%name)) &
         CALL cp__b("pao_io.F",369,"Primary Basis-set name does not match")
      IF (pao_kind%prim_basis_size /= basis_set%nsgf) &
         CALL cp__b("pao_io.F",371,"Primary Basis-set size does not match")
      IF (pao_kind%pao_basis_size /= pao_basis_size) &
         CALL cp__b("pao_io.F",373,"PAO basis size does not match")
      IF (pao_kind%pao_pot_maxl /= pao_pot_maxl) &
         CALL cp__b("pao_io.F",375,"PAO_POT_MAXL does not match")
      IF (pao_kind%pao_pot_max_projector /= pao_pot_max_projector) &
         CALL cp__b("pao_io.F",377,"PAO_POT_MAX_PROJECTOR does not match")
      IF (pao_kind%pao_pot_neighbors /= pao_pot_neighbors) &
         CALL cp__b("pao_io.F",379,"PAO_POT_NEIGHBORS does not match")
      IF (pao_kind%pao_pot_beta /= pao_pot_beta) &
         CALL cp__b("pao_io.F",381,"PAO_POT_BETA does not match")
      IF (pao_kind%pao_pot_weight /= pao_pot_weight) &
         CALL cp__b("pao_io.F",383,"PAO_POT_WEIGHT does not match")

   END SUBROUTINE pao_kinds_ensure_equal

! **************************************************************************************************
!> \brief Writes restart file
!> \param pao ...
!> \param qs_env ...
!> \param energy ...
! **************************************************************************************************
   SUBROUTINE pao_write_restart(pao, qs_env, energy)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp)                                           :: energy

      CHARACTER(len=*), PARAMETER :: printkey_section = 'DFT%LS_SCF%PAO%PRINT%RESTART', &
         routineN = 'pao_write_restart', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_max, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env, input=input, para_env=para_env)

      ! open file
      unit_nr = cp_print_key_unit_nr(logger, &
                                     input, &
                                     printkey_section, &
                                     extension=".pao", &
                                     file_action="WRITE", &
                                     file_position="REWIND", &
                                     file_status="UNKNOWN", &
                                     do_backup=.TRUE.)

      ! although just rank-0 writes the trajectory it requires collective MPI calls
      unit_max = unit_nr
      CALL mp_max(unit_max, para_env%group)
      IF (unit_max > 0) THEN
         IF (pao%iw > 0) WRITE (pao%iw, '(A,A)') " PAO| Writing restart file."
         IF (unit_nr > 0) &
            CALL write_restart_header(pao, qs_env, energy, unit_nr)

         CALL pao_write_diagonal_blocks(para_env, pao%matrix_X, "Xblock", unit_nr)
         IF (pao%parameterization == pao_exp_param) &
            CALL pao_write_diagonal_blocks(para_env, pao%matrix_U0, "U0block", unit_nr)

      ENDIF

      ! close file
      IF (unit_nr > 0) WRITE (unit_nr, '(A)') "THE_END"
      CALL cp_print_key_finished_output(unit_nr, logger, input, printkey_section)

      CALL timestop(handle)
   END SUBROUTINE pao_write_restart

! **************************************************************************************************
!> \brief Write the digonal blocks of given DBCSR matrix into the provided unit_nr
!> \param para_env ...
!> \param matrix ...
!> \param label ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE pao_write_diagonal_blocks(para_env, matrix, label, unit_nr)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_type)                                :: matrix
      CHARACTER(LEN=*), INTENT(IN)                       :: label
      INTEGER, INTENT(IN)                                :: unit_nr

      INTEGER                                            :: iatom, natoms
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes, row_blk_sizes
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:, :), POINTER                 :: local_block, mpi_buffer

      !TODO: this is a serial algorithm
      row_blk_sizes => cp_dbcsr_row_block_sizes(matrix)
      col_blk_sizes => cp_dbcsr_col_block_sizes(matrix)
      IF(.NOT.(SIZE(row_blk_sizes) == SIZE(col_blk_sizes)))CALL cp__a("pao_io.F",463)
      natoms = SIZE(row_blk_sizes)

      DO iatom = 1, natoms
         ALLOCATE (mpi_buffer(row_blk_sizes(iatom), col_blk_sizes(iatom)))
         NULLIFY (local_block)
         CALL cp_dbcsr_get_block_p(matrix=matrix, row=iatom, col=iatom, block=local_block, found=found)
         IF (ASSOCIATED(local_block)) THEN
            IF (SIZE(local_block) > 0) & ! catch corner-case
               mpi_buffer(:, :) = local_block(:, :)
         ELSE
            mpi_buffer(:, :) = 0.0_dp
         ENDIF

         CALL mp_sum(mpi_buffer, para_env%group)
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, fmt="(A,1X,I10,1X)", advance='no') label, iatom
            WRITE (unit_nr, *) mpi_buffer
         ENDIF
         DEALLOCATE (mpi_buffer)
      ENDDO

      ! flush
      IF (unit_nr > 0) FLUSH(unit_nr)

   END SUBROUTINE pao_write_diagonal_blocks

! **************************************************************************************************
!> \brief Writes header of restart file
!> \param pao ...
!> \param qs_env ...
!> \param energy ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE write_restart_header(pao, qs_env, energy, unit_nr)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp)                                           :: energy
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'write_restart_header', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: kindname
      INTEGER                                            :: iatom, ikind, nparams, pao_basis_size, &
                                                            pao_pot_max_projector, pao_pot_maxl, &
                                                            pao_pot_neighbors, z
      REAL(dp)                                           :: pao_pot_beta, pao_pot_weight
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, &
                      cell=cell, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      WRITE (unit_nr, "(A,5X,I0)") "Version", file_format_version
      WRITE (unit_nr, "(A,5X,F20.10)") "Energy", energy
      WRITE (unit_nr, "(A,5X,I0)") "Step", pao%istep
      WRITE (unit_nr, "(A,5X,A10)") "Parametrization", id2str(pao%parameterization)

      ! write kinds
      WRITE (unit_nr, "(A,5X,I0)") "Nkinds", SIZE(atomic_kind_set)
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), name=kindname, z=z)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          pao_basis_size=pao_basis_size, &
                          pao_pot_maxl=pao_pot_maxl, &
                          pao_pot_max_projector=pao_pot_max_projector, &
                          pao_pot_neighbors=pao_pot_neighbors, &
                          pao_pot_beta=pao_pot_beta, &
                          pao_pot_weight=pao_pot_weight, &
                          basis_set=basis_set)
         CALL pao_param_count(pao, qs_env, ikind, nparams)
         WRITE (unit_nr, "(A,5X,I10,1X,A,1X,I3)") "Kind", ikind, TRIM(kindname), z
         WRITE (unit_nr, "(A,5X,I10,1X,I10,1X,A)") "PrimBasis", ikind, basis_set%nsgf, TRIM(basis_set%name)
         WRITE (unit_nr, "(A,5X,I10,1X,I3,1X,I3)", advance='no') "PaoBasis", ikind, pao_basis_size, pao_pot_maxl
         WRITE (unit_nr, "(1X,I3,1X,I3,1X)", advance='no') pao_pot_max_projector, pao_pot_neighbors
         WRITE (unit_nr, "(F20.16,1X,F20.16)") pao_pot_beta, pao_pot_weight
         WRITE (unit_nr, "(A,5X,I10,1X,I3)") "NParams", ikind, nparams
      ENDDO

      ! write cell
      WRITE (unit_nr, fmt="(A,5X)", advance='no') "Cell"
      WRITE (unit_nr, *) cell%hmat*angstrom

      ! write atoms
      WRITE (unit_nr, "(A,5X,I0)") "Natoms", SIZE(particle_set)
      DO iatom = 1, SIZE(particle_set)
         kindname = particle_set(iatom)%atomic_kind%name
         WRITE (unit_nr, fmt="(A,5X,I10,5X,A,1X)", advance='no') "Atom ", iatom, TRIM(kindname)
         WRITE (unit_nr, *) particle_set(iatom)%r*angstrom
      ENDDO

   END SUBROUTINE write_restart_header

END MODULE pao_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Neural Network implementation
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_ml_neuralnet
   USE kinds,                           ONLY: dp
   USE pao_types,                       ONLY: pao_env_type,&
                                              training_matrix_type
   USE parallel_rng_types,              ONLY: create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_ml_neuralnet'

   PUBLIC ::pao_ml_nn_train, pao_ml_nn_predict

   ! TODO turn these into input parameters
   REAL(dp), PARAMETER   :: step_size = 0.001_dp
   INTEGER, PARAMETER    :: nlayers = 5
   REAL(dp), PARAMETER   :: convergence_eps = 1e-7_dp
   INTEGER, PARAMETER    :: max_training_cycles = 50000

CONTAINS

! **************************************************************************************************
!> \brief Uses neural network to make a prediction
!> \param training_matrix ...
!> \param descriptor ...
!> \param output ...
!> \param variance ...
! **************************************************************************************************
   SUBROUTINE pao_ml_nn_predict(training_matrix, descriptor, output, variance)
      TYPE(training_matrix_type), INTENT(IN)             :: training_matrix
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: descriptor
      REAL(dp), DIMENSION(:), INTENT(OUT)                :: output
      REAL(dp), INTENT(OUT)                              :: variance

      CALL nn_eval(training_matrix%NN, input=descriptor, prediction=output)

      variance = 0.0_dp ! Neural Networks don't provide a variance
   END SUBROUTINE pao_ml_nn_predict

! **************************************************************************************************
!> \brief Trains the neural network on given training points
!> \param pao ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE pao_ml_nn_train(pao, training_matrices)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(training_matrix_type), DIMENSION(:), TARGET   :: training_matrices

      INTEGER                                            :: i, icycle, ikind, ilayer, ipoint, j, &
                                                            npoints, width, width_in, width_out
      REAL(dp)                                           :: bak, eps, error, error1, error2, num_grad
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: prediction
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: gradient
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(training_matrix_type), POINTER                :: training_matrix

      NULLIFY (rng_stream)

      ! TODO this could be parallelized over ranks
      DO ikind = 1, SIZE(training_matrices)
         training_matrix => training_matrices(ikind)

         npoints = SIZE(training_matrix%inputs, 1) ! number of points
         IF(.NOT.(SIZE(training_matrix%outputs, 1) == npoints))CALL cp__a("pao_ml_neuralnet.F",78)
         IF(.NOT.(npoints > 0))CALL cp__a("pao_ml_neuralnet.F",79)

         !TODO proper output
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO|ML| Training neural network for kind: ", &
            TRIM(training_matrix%kindname), " from ", npoints, "training points."

         ! determine network width and allocate it
         width_in = SIZE(training_matrix%inputs, 2)
         width_out = SIZE(training_matrix%outputs, 2)
         width = MAX(width_in, width_out)
         ALLOCATE (training_matrix%NN(nlayers, width, width))

         ! initialize network with random numbers from -1.0 ... +1.0
         CALL create_rng_stream(rng_stream, name="pao_nn")
         DO ilayer = 1, nlayers
         DO i = 1, width
         DO j = 1, width
            training_matrix%NN(ilayer, i, j) = -1.0_dp+2.0_dp*next_random_number(rng_stream)
         ENDDO
         ENDDO
         ENDDO
         CALL delete_rng_stream(rng_stream)

         ! train the network using backpropagation
         ALLOCATE (gradient(nlayers, width, width))
         DO icycle = 1, max_training_cycles
            error = 0.0_dp
            gradient = 0.0_dp
            DO ipoint = 1, npoints
               CALL nn_backpropagate(training_matrix%NN, &
                                     input=training_matrix%inputs(ipoint, :), &
                                     TARGET=training_matrix%outputs(ipoint, :), &
                                     gradient=gradient, &
                                     error=error)
            ENDDO
            training_matrix%NN(:, :, :) = training_matrix%NN-step_size*gradient

            IF (pao%iw > 0 .AND. MOD(icycle, 100) == 0) WRITE (pao%iw, *) &
               "PAO|ML| ", TRIM(training_matrix%kindname), &
               " training-cycle:", icycle, "SQRT(error):", SQRT(error), "grad:", SUM(gradient**2)

            IF (SUM(gradient**2) < convergence_eps) EXIT
         ENDDO

         ! numeric gradient for debugging ----------------------------------------------------------
         IF (.FALSE.) THEN
            eps = 1e-4_dp
            ilayer = 1
            ipoint = 1
            error = 0.0_dp
            gradient = 0.0_dp
            CALL nn_backpropagate(training_matrix%NN, &
                                  input=training_matrix%inputs(ipoint, :), &
                                  TARGET=training_matrix%outputs(ipoint, :), &
                                  gradient=gradient, &
                                  error=error)

            ALLOCATE (prediction(width_out))
            DO i = 1, width
            DO j = 1, width
               bak = training_matrix%NN(ilayer, i, j)

               training_matrix%NN(ilayer, i, j) = bak+eps
               CALL nn_eval(training_matrix%NN, &
                            input=training_matrix%inputs(ipoint, :), &
                            prediction=prediction)
               error1 = SUM((training_matrix%outputs(ipoint, :)-prediction)**2)

               training_matrix%NN(ilayer, i, j) = bak-eps
               CALL nn_eval(training_matrix%NN, &
                            input=training_matrix%inputs(ipoint, :), &
                            prediction=prediction)
               error2 = SUM((training_matrix%outputs(ipoint, :)-prediction)**2)

               training_matrix%NN(ilayer, i, j) = bak
               num_grad = (error1-error2)/(2.0_dp*eps)
               IF (pao%iw > 0) WRITE (pao%iw, *) "PAO|ML| Numeric gradient:", i, j, gradient(ilayer, i, j), num_grad

            ENDDO
            ENDDO
            DEALLOCATE (prediction)
         ENDIF
         !------------------------------------------------------------------------------------------

         DEALLOCATE (gradient)

         ! test training points individually
         ALLOCATE (prediction(width_out))
         DO ipoint = 1, npoints
            CALL nn_eval(training_matrix%NN, &
                         input=training_matrix%inputs(ipoint, :), &
                         prediction=prediction)
            error = MAXVAL(ABS(training_matrix%outputs(ipoint, :)-prediction))
            IF (pao%iw > 0) WRITE (pao%iw, *) "PAO|ML| ", TRIM(training_matrix%kindname), &
               "verify training-point:", ipoint, "SQRT(error):", SQRT(error)
         ENDDO
         DEALLOCATE (prediction)

      ENDDO

   END SUBROUTINE pao_ml_nn_train

! **************************************************************************************************
!> \brief Evaluates the neural network for a given input
!> \param A ...
!> \param input ...
!> \param prediction ...
! **************************************************************************************************
   SUBROUTINE nn_eval(A, input, prediction)
      REAL(dp), DIMENSION(:, :, :), INTENT(IN)           :: A
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: input
      REAL(dp), DIMENSION(:), INTENT(OUT)                :: prediction

      INTEGER                                            :: i, ilayer, j, nlayers, width, width_in, &
                                                            width_out
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: forward

      nlayers = SIZE(A, 1)
      width = SIZE(A, 2); IF(.NOT.(SIZE(A, 2) == SIZE(A, 3)))CALL cp__a("pao_ml_neuralnet.F",197)
      width_in = SIZE(input)
      width_out = SIZE(prediction)

      ALLOCATE (forward(0:nlayers, width))

      forward = 0.0_dp
      forward(0, 1:width_in) = input(:)

      DO ilayer = 1, nlayers
      DO i = 1, width
      DO j = 1, width
         forward(ilayer, i) = forward(ilayer, i)+A(ilayer, i, j)*TANH(forward(ilayer-1, j))
      ENDDO
      ENDDO
      ENDDO

      prediction(:) = forward(nlayers, 1:width_out)

   END SUBROUTINE nn_eval

! **************************************************************************************************
!> \brief Uses backpropagation to calculate the gradient for a given training point
!> \param A ...
!> \param input ...
!> \param TARGET ...
!> \param error ...
!> \param gradient ...
! **************************************************************************************************
   SUBROUTINE nn_backpropagate(A, input, TARGET, error, gradient)
      REAL(dp), DIMENSION(:, :, :), INTENT(IN)           :: A
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: input, TARGET
      REAL(dp), INTENT(INOUT)                            :: error
      REAL(dp), DIMENSION(:, :, :), INTENT(INOUT)        :: gradient

      INTEGER                                            :: i, ilayer, j, nlayers, width, width_in, &
                                                            width_out
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: prediction
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: backward, forward

      nlayers = SIZE(A, 1)
      width = SIZE(A, 2); IF(.NOT.(SIZE(A, 2) == SIZE(A, 3)))CALL cp__a("pao_ml_neuralnet.F",238)
      width_in = SIZE(input)
      width_out = SIZE(TARGET)

      ALLOCATE (forward(0:nlayers, width), prediction(width_out), backward(0:nlayers, width))

      forward = 0.0_dp
      forward(0, 1:width_in) = input

      DO ilayer = 1, nlayers
      DO i = 1, width
      DO j = 1, width
         forward(ilayer, i) = forward(ilayer, i)+A(ilayer, i, j)*TANH(forward(ilayer-1, j))
      ENDDO
      ENDDO
      ENDDO

      prediction(:) = forward(nlayers, 1:width_out)

      error = error+SUM((prediction-TARGET)**2)

      ! Turning Point ------------------------------------------------------------------------------
      backward = 0.0_dp
      backward(nlayers, 1:width_out) = prediction-TARGET

      DO ilayer = nlayers, 1, -1
      DO i = 1, width
      DO j = 1, width
         gradient(ilayer, i, j) = gradient(ilayer, i, j)+2.0_dp*backward(ilayer, i)*TANH(forward(ilayer-1, j))
         backward(ilayer-1, j) = backward(ilayer-1, j)+backward(ilayer, i)*A(ilayer, i, j)*(1.0_dp-TANH(forward(ilayer-1, j))**2)
      ENDDO
      ENDDO
      ENDDO

      DEALLOCATE (forward, backward, prediction)
   END SUBROUTINE nn_backpropagate

END MODULE pao_ml_neuralnet
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Main module for PAO Machine Learning
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_ml
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_create,&
                                              cell_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_iterator,&
                                              cp_dbcsr_iterator_blocks_left,&
                                              cp_dbcsr_iterator_next_block,&
                                              cp_dbcsr_iterator_start,&
                                              cp_dbcsr_iterator_stop,&
                                              cp_dbcsr_p_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
   USE pao_input,                       ONLY: id2str,&
                                              pao_exp_param,&
                                              pao_ml_gp,&
                                              pao_ml_nn
   USE pao_io,                          ONLY: pao_ioblock_type,&
                                              pao_iokind_type,&
                                              pao_kinds_ensure_equal,&
                                              pao_read_raw
   USE pao_ml_codec,                    ONLY: pao_ml_decode_xblock,&
                                              pao_ml_encode_xblock
   USE pao_ml_descriptor,               ONLY: pao_ml_calc_descriptor
   USE pao_ml_gaussprocess,             ONLY: pao_ml_gp_predict,&
                                              pao_ml_gp_train
   USE pao_ml_neuralnet,                ONLY: pao_ml_nn_predict,&
                                              pao_ml_nn_train
   USE pao_types,                       ONLY: pao_env_type,&
                                              training_matrix_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_ml'

   PUBLIC :: pao_machine_learning_init

   ! linked list used to group training points by kind
   TYPE training_point_type
      TYPE(training_point_type), POINTER       :: next => Null()
      REAL(dp), DIMENSION(:), ALLOCATABLE      :: input
      REAL(dp), DIMENSION(:), ALLOCATABLE      :: output
   END TYPE training_point_type

   TYPE training_list_type
      CHARACTER(LEN=default_string_length)     :: kindname = ""
      TYPE(training_point_type), POINTER       :: head => Null()
      INTEGER                                  :: npoints = 0
   END TYPE training_list_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes the learning machinery
!> \param pao ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE pao_machine_learning_init(pao, qs_env)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_machine_learning_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: line
      INTEGER                                            :: ikind, istat, unit_nr
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(training_list_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: training_lists
      TYPE(training_matrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: training_matrices

      IF(.NOT.(LEN_TRIM(pao%ml_training_file) > 0))CALL cp__a("pao_ml.F",97)

      IF (pao%iw > 0) WRITE (pao%iw, *) 'PAO|ML| Initializing maschine learning...'

      IF (pao%parameterization /= pao_exp_param) &
         CALL cp__b("pao_ml.F",102,"PAO maschine learning requires EXP parametrization")

      CALL get_qs_env(qs_env, para_env=para_env, atomic_kind_set=atomic_kind_set)

      ! create training-set data-structure
      ALLOCATE (training_lists(SIZE(atomic_kind_set)))
      DO ikind = 1, SIZE(training_lists)
         CALL get_atomic_kind(atomic_kind_set(ikind), name=training_lists(ikind)%kindname)
      ENDDO

      IF (para_env%mepos == para_env%source) THEN
         CALL open_file(file_name=pao%ml_training_file, file_status="OLD", file_form="FORMATTED", &
                        file_action="READ", unit_number=unit_nr)
         DO WHILE (.TRUE.)
            READ (unit_nr, IOSTAT=istat, FMT='(A)') line
            IF (istat /= 0) EXIT ! end of file reached
            line = ADJUSTL(line)
            IF (LEN_TRIM(line) == 0) CYCLE ! skip empty line
            IF (line(1:1) == "!" .OR. line(1:1) == "#") CYCLE ! skipt comment

            CALL add_to_training_list(pao, qs_env, training_lists, filename=line)
         ENDDO

         CALL close_file(unit_number=unit_nr)

         CALL training_list2matrix(training_lists, training_matrices)
         DEALLOCATE (training_lists)
      ENDIF

      CALL bcast_training(para_env, training_matrices)
      CALL pao_ml_train(pao, training_matrices)
      CALL pao_ml_predict(pao, qs_env, training_matrices)

      DEALLOCATE (training_matrices)

   END SUBROUTINE pao_machine_learning_init

! **************************************************************************************************
!> \brief Reads the given file and adds it's training points to linked lists.
!> \param pao ...
!> \param qs_env ...
!> \param training_lists ...
!> \param filename ...
! **************************************************************************************************
   SUBROUTINE add_to_training_list(pao, qs_env, training_lists, filename)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(training_list_type), DIMENSION(:)             :: training_lists
      CHARACTER(LEN=default_path_length)                 :: filename

      CHARACTER(len=*), PARAMETER :: routineN = 'add_to_training_list', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: param
      INTEGER                                            :: iatom, ikind, natoms
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom2kind, kindsmap
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: hmat, positions
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(pao_ioblock_type), ALLOCATABLE, DIMENSION(:)  :: u0blocks, xblocks
      TYPE(pao_iokind_type), ALLOCATABLE, DIMENSION(:)   :: kinds
      TYPE(particle_type), DIMENSION(:), POINTER         :: my_particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(training_point_type), POINTER                 :: new_point

      NULLIFY (new_point, cell)

      IF (pao%iw > 0) WRITE (pao%iw, '(A,A)') " PAO|ML| Reading training frame from file: ", TRIM(filename)

      ! parse training data
      CALL pao_read_raw(filename, param, hmat, kinds, atom2kind, positions, xblocks, u0blocks)

      ! check parametrization
      IF (TRIM(param) .NE. TRIM(ADJUSTL(id2str(pao%parameterization)))) &
         CALL cp__b("pao_ml.F",176,"Restart PAO parametrization does not match")

      ! create cell from read-in h-matrix
      CALL cell_create(cell, hmat)

      ! map read-in kinds onto kinds of this run
      CALL match_kinds(pao, qs_env, kinds, kindsmap)

      ! create a particle_set based on read-in positions and refere to kinds of this run
      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set)
      natoms = SIZE(positions, 1)
      ALLOCATE (my_particle_set(natoms))
      DO iatom = 1, natoms
         ikind = kindsmap(atom2kind(iatom))
         my_particle_set(iatom)%atomic_kind => atomic_kind_set(ikind)
         my_particle_set(iatom)%r = positions(iatom, :)
      ENDDO

      IF (.NOT. ALLOCATED(u0blocks)) &
         CALL cp__b("pao_ml.F",195,"PAO-ML requires U0_blocks e.g. from EXP parametization")

      ! build training points
      DO ikind = 1, SIZE(kindsmap)
         DO iatom = 1, SIZE(xblocks)
            IF (kindsmap(atom2kind(iatom)) == ikind) THEN
               ALLOCATE (new_point)

               CALL pao_ml_calc_descriptor(my_particle_set, &
                                           qs_kind_set, &
                                           cell, &
                                           iatom=iatom, &
                                           descriptor=new_point%input)

               CALL pao_ml_encode_xblock(my_particle_set, &
                                         qs_kind_set, &
                                         cell, &
                                         iatom=iatom, &
                                         u0block=u0blocks(iatom)%p, &
                                         xblock=xblocks(iatom)%p, &
                                         output=new_point%output)

               ! add to linked list
               training_lists(ikind)%npoints = training_lists(ikind)%npoints+1
               new_point%next => training_lists(ikind)%head
               training_lists(ikind)%head => new_point
            ENDIF
         ENDDO
      ENDDO

      DEALLOCATE (cell, my_particle_set)
      ! ALLOCATABLEs deallocate themselfs

   END SUBROUTINE add_to_training_list

! **************************************************************************************************
!> \brief Make read-in kinds on to atomic-kinds of this run
!> \param pao ...
!> \param qs_env ...
!> \param kinds ...
!> \param kindsmap ...
! **************************************************************************************************
   SUBROUTINE match_kinds(pao, qs_env, kinds, kindsmap)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pao_iokind_type), DIMENSION(:)                :: kinds
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kindsmap

      CHARACTER(LEN=default_string_length)               :: name
      INTEGER                                            :: ikind, jkind
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set)

      IF(.NOT.(.NOT. ALLOCATED(kindsmap)))CALL cp__a("pao_ml.F",249)
      ALLOCATE (kindsmap(SIZE(kinds)))
      kindsmap(:) = -1

      DO ikind = 1, SIZE(kinds)
         DO jkind = 1, SIZE(atomic_kind_set)
            CALL get_atomic_kind(atomic_kind_set(jkind), name=name)
            ! match kinds via their name
            IF (TRIM(kinds(ikind)%name) .EQ. TRIM(name)) THEN
               CALL pao_kinds_ensure_equal(pao, qs_env, jkind, kinds(ikind))
               kindsmap(ikind) = jkind
               EXIT
            ENDIF
         ENDDO
      ENDDO

   END SUBROUTINE match_kinds

! **************************************************************************************************
!> \brief Turns the linked lists of training points into matrices
!> \param training_lists ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE training_list2matrix(training_lists, training_matrices)
      TYPE(training_list_type), ALLOCATABLE, &
         DIMENSION(:), TARGET                            :: training_lists
      TYPE(training_matrix_type), ALLOCATABLE, &
         DIMENSION(:), TARGET                            :: training_matrices

      INTEGER                                            :: i, ikind, inp_size, npoints, out_size
      TYPE(training_list_type), POINTER                  :: training_list
      TYPE(training_matrix_type), POINTER                :: training_matrix
      TYPE(training_point_type), POINTER                 :: cur_point, prev_point

      IF(.NOT.(ALLOCATED(training_lists) .AND. .NOT. ALLOCATED(training_matrices)))CALL cp__a("pao_ml.F",283)

      ALLOCATE (training_matrices(SIZE(training_lists)))

      DO ikind = 1, SIZE(training_lists)
         training_list => training_lists(ikind)
         training_matrix => training_matrices(ikind)
         training_matrix%kindname = training_list%kindname ! copy kindname
         npoints = training_list%npoints ! number of points
         IF(.NOT.(npoints > 0))CALL cp__a("pao_ml.F",292)

         inp_size = SIZE(training_list%head%input)
         out_size = SIZE(training_list%head%output)

         ! copy from link-list into arrays
         ALLOCATE (training_matrix%inputs(npoints, inp_size))
         ALLOCATE (training_matrix%outputs(npoints, out_size))

         cur_point => training_list%head
         NULLIFY (training_list%head)
         DO i = 1, npoints
            training_matrix%inputs(i, :) = cur_point%input(:)
            training_matrix%outputs(i, :) = cur_point%output(:)
            prev_point => cur_point
            cur_point => cur_point%next
            DEALLOCATE (prev_point) ! deallocate linked-list
         ENDDO
         training_list%npoints = 0 ! list is now empty
      ENDDO

   END SUBROUTINE training_list2matrix

! **************************************************************************************************
!> \brief Broadcasts the matrices of trainging points across all MPI ranks.
!> \param para_env ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE bcast_training(para_env, training_matrices)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(training_matrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: training_matrices

      INTEGER                                            :: ikind, nkinds, s1, s2

      IF (para_env%mepos == para_env%source) &
         nkinds = SIZE(training_matrices)
      CALL mp_bcast(nkinds, para_env%source, para_env%group)
      IF (para_env%mepos /= para_env%source) &
         ALLOCATE (training_matrices(nkinds))

      DO ikind = 1, nkinds
         CALL mp_bcast(training_matrices(ikind)%kindname, para_env%source, para_env%group)

         IF (para_env%mepos == para_env%source) THEN
            s1 = SIZE(training_matrices(ikind)%outputs, 1)
            s2 = SIZE(training_matrices(ikind)%outputs, 2)
         ENDIF
         CALL mp_bcast(s1, para_env%source, para_env%group)
         CALL mp_bcast(s2, para_env%source, para_env%group)
         IF (para_env%mepos /= para_env%source) &
            ALLOCATE (training_matrices(ikind)%outputs(s1, s2))
         CALL mp_bcast(training_matrices(ikind)%outputs, para_env%source, para_env%group)

         IF (para_env%mepos == para_env%source) THEN
            s1 = SIZE(training_matrices(ikind)%inputs, 1)
            s2 = SIZE(training_matrices(ikind)%inputs, 2)
         ENDIF
         CALL mp_bcast(s1, para_env%source, para_env%group)
         CALL mp_bcast(s2, para_env%source, para_env%group)
         IF (para_env%mepos /= para_env%source) &
            ALLOCATE (training_matrices(ikind)%inputs(s1, s2))
         CALL mp_bcast(training_matrices(ikind)%inputs, para_env%source, para_env%group)
      ENDDO

   END SUBROUTINE bcast_training

! **************************************************************************************************
!> \brief Calls the actuall learning algorthim to traing on the given matrices
!> \param pao ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE pao_ml_train(pao, training_matrices)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(training_matrix_type), DIMENSION(:)           :: training_matrices

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_ml_train', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (pao%ml_method)
      CASE (pao_ml_gp)
         CALL pao_ml_gp_train(pao, training_matrices)
      CASE (pao_ml_nn)
         CALL pao_ml_nn_train(pao, training_matrices)
      CASE DEFAULT
         CALL cp__b("pao_ml.F",380,"PAO: unkown machine learning scheme")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE pao_ml_train

! **************************************************************************************************
!> \brief Fills pao%matrix_X based on machine learning predictions
!> \param pao ...
!> \param qs_env ...
!> \param training_matrices ...
! **************************************************************************************************
   SUBROUTINE pao_ml_predict(pao, qs_env, training_matrices)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(training_matrix_type), DIMENSION(:)           :: training_matrices

      CHARACTER(len=*), PARAMETER                        :: routineN = 'pao_ml_predict'

      INTEGER                                            :: acol, arow, handle, iatom, ikind, natoms
      LOGICAL                                            :: found
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: descriptor, output, variances
      REAL(dp), DIMENSION(:, :), POINTER                 :: block_U0, block_X
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      para_env=para_env, &
                      cell=cell, &
                      matrix_s=matrix_s, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      natom=natoms)

      ! fill matrix_X
      ALLOCATE (variances(natoms))
      variances = 0.0_dp
      CALL cp_dbcsr_iterator_start(iter, pao%matrix_X)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, arow, acol, block_X)
         iatom = arow; IF(.NOT.(arow == acol))CALL cp__a("pao_ml.F",429)
         CALL cp_dbcsr_get_block_p(matrix=pao%matrix_U0, row=iatom, col=iatom, block=block_U0, found=found)
         IF(.NOT.(ASSOCIATED(block_U0)))CALL cp__a("pao_ml.F",431)
         CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)

         ! calculate decriptor
         CALL pao_ml_calc_descriptor(particle_set, &
                                     qs_kind_set, &
                                     cell, &
                                     iatom, &
                                     descriptor)

         ! call actual machine learning for prediction
         CALL pao_ml_predict_low(pao, training_matrices(ikind), descriptor, output, variances(iatom))

         ! decode output into block_X
         CALL pao_ml_decode_xblock(particle_set, &
                                   qs_kind_set, &
                                   cell, &
                                   iatom=iatom, &
                                   u0block=block_U0, &
                                   output=output, &
                                   xblock=block_X)

         DEALLOCATE (output, descriptor)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! print variances
      CALL mp_sum(variances, para_env%group)
      IF (pao%iw > 0) THEN
         DO iatom = 1, natoms
            WRITE (pao%iw, *) "PAO|ML| predicting atom:", iatom, "variance:", variances(iatom)
         ENDDO
         WRITE (pao%iw, *) "PAO|ML| max prediction variance:", MAXVAL(variances)
      ENDIF

      IF (MAXVAL(variances) > pao%ml_tolerance) &
         CALL cp__b("pao_ml.F",467,"Variance of prediction above ML_TOLERANCE.")

      DEALLOCATE (variances)

      CALL timestop(handle)

   END SUBROUTINE pao_ml_predict

! **************************************************************************************************
!> \brief Queries the actuall learning algorthim to make a prediction
!> \param pao ...
!> \param training_matrix ...
!> \param descriptor ...
!> \param output ...
!> \param variance ...
! **************************************************************************************************
   SUBROUTINE pao_ml_predict_low(pao, training_matrix, descriptor, output, variance)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(training_matrix_type), INTENT(IN)             :: training_matrix
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: descriptor
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: output
      REAL(dp), INTENT(OUT)                              :: variance

      INTEGER                                            :: npoints, width_out

      IF(.NOT.(.NOT. ALLOCATED(output)))CALL cp__a("pao_ml.F",492)

      npoints = SIZE(training_matrix%outputs, 1)
      IF(.NOT.(npoints > 0))CALL cp__a("pao_ml.F",495)

      width_out = SIZE(training_matrix%outputs, 2)
      ALLOCATE (output(width_out))

      SELECT CASE (pao%ml_method)
      CASE (pao_ml_gp)
         CALL pao_ml_gp_predict(pao, training_matrix, descriptor, output, variance)
      CASE (pao_ml_nn)
         CALL pao_ml_nn_predict(training_matrix, descriptor, output, variance)
      CASE DEFAULT
         CALL cp__b("pao_ml.F",506,"PAO: unkown machine learning scheme")
      END SELECT

   END SUBROUTINE pao_ml_predict_low

END MODULE pao_ml
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to handle the external control of CP2K
!> \par History
!>      - Moved from MODULE termination to here (18.02.2011,MK)
!>      - add communication control (20.02.2013 Mandes)
!> \author Marcella Iannuzzi (10.03.2005,MI)
! **************************************************************************************************
MODULE cp_external_control

   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE global_types,                    ONLY: global_environment_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_walltime
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_probe
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_external_control'

   PUBLIC :: external_control
   PUBLIC :: set_external_comm

   INTEGER, SAVE :: external_comm = -1
   INTEGER, SAVE :: external_master_id = -1
   INTEGER, SAVE :: scf_energy_message_tag = -1
   INTEGER, SAVE :: exit_tag = -1

CONTAINS

! **************************************************************************************************
!> \brief set the communicator to an external source or destination,
!>        to send messages (e.g. intermediate energies during scf) or
!>        reveive commands (e.g. aborting the calculation)
!> \param comm ...
!> \param in_external_master_id ...
!> \param in_scf_energy_message_tag ...
!> \param in_exit_tag ...
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE set_external_comm(comm, in_external_master_id, &
                                in_scf_energy_message_tag, in_exit_tag)
      INTEGER, INTENT(IN)                                :: comm, in_external_master_id
      INTEGER, INTENT(IN), OPTIONAL                      :: in_scf_energy_message_tag, in_exit_tag

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_external_comm', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(in_external_master_id .GE. 0))CALL cp__a("cp_external_control.F",62)

      external_comm = comm
      external_master_id = in_external_master_id

      IF (PRESENT(in_scf_energy_message_tag)) &
         scf_energy_message_tag = in_scf_energy_message_tag
      IF (PRESENT(in_exit_tag)) THEN
         ! the exit tag should be different from the mp_probe tag default
         IF(.NOT.(in_exit_tag .NE. -1))CALL cp__a("cp_external_control.F",71)
         exit_tag = in_exit_tag
      END IF
   END SUBROUTINE set_external_comm

! **************************************************************************************************
!> \brief External manipulations during a run : when the <PROJECT_NAME>.EXIT_$runtype
!>      command is sent the program stops at the level of $runtype
!>      when a general <PROJECT_NAME>.EXIT command is sent the program is stopped
!>      at all levels (at least those that call this function)
!>      if the file WAIT exists, the program waits here till it disappears
!> \param should_stop ...
!> \param flag ...
!> \param globenv ...
!> \param target_time ...
!> \param start_time ...
!> \author MI (10.03.2005)
! **************************************************************************************************
   SUBROUTINE external_control(should_stop, flag, globenv, target_time, start_time)

      LOGICAL, INTENT(OUT)                               :: should_stop
      CHARACTER(LEN=*), INTENT(IN)                       :: flag
      TYPE(global_environment_type), OPTIONAL, POINTER   :: globenv
      REAL(dp), OPTIONAL                                 :: target_time, start_time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'external_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: exit_fname, exit_fname_level, &
                                                            exit_gname, exit_gname_level
      INTEGER                                            :: handle, i, tag, unit_number
      LOGICAL                                            :: should_wait
      LOGICAL, SAVE                                      :: check_always = .FALSE.
      REAL(KIND=dp)                                      :: my_start_time, my_target_time, t1, t2, &
                                                            time_check
      REAL(KIND=dp), SAVE                                :: t_last_file_check = 0.0_dp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      should_stop = .FALSE.

      exit_gname = "EXIT"
      exit_gname_level = TRIM(exit_gname)//"_"//TRIM(flag)
      exit_fname = TRIM(logger%iter_info%project_name)//"."//TRIM(exit_gname)
      exit_fname_level = TRIM(logger%iter_info%project_name)//"."//TRIM(exit_gname_level)

      ! check for incomming messages and if it is tagged with the exit tag
      IF (exit_tag .NE. -1) THEN
         i = external_master_id
         CALL mp_probe(source=i, comm=external_comm, tag=tag)
         IF (tag .EQ. exit_tag) should_stop = .TRUE.
      END IF

      IF (logger%para_env%source == logger%para_env%mepos) THEN
         ! files will only be checked every 20 seconds, or if the clock wraps/does not exist,
         ! otherwise 64 waters on 64 cores can spend up to 10% of time here, on lustre
         ! however, if should_stop has been true, we should always check
         ! (at each level scf, md, ... the file must be there to guarantee termination)
         t1 = m_walltime()
         IF (t1 > t_last_file_check+20.0_dp .OR. t1 <= t_last_file_check .OR. check_always) THEN

            t_last_file_check = t1
            ! allows for halting execution for a while
            ! this is useful to copy a consistent snapshot of the output
            ! while a simulation is running
            INQUIRE (FILE="/mnt/sdcard/cp2k/work/WAIT", EXIST=should_wait)
            IF (should_wait) THEN
               CALL open_file(file_name="/mnt/sdcard/cp2k/work/WAITING", file_status="UNKNOWN", &
                              file_form="FORMATTED", file_action="WRITE", &
                              unit_number=unit_number)
               WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,/)") &
                  "*** waiting till the file WAIT has been removed ***"
               DO
                  ! sleep a bit (to save the file system)
                  t1 = m_walltime()
                  DO I = 1, 100000000
                     t2 = m_walltime()
                     IF (t2-t1 > 1.0_dp) EXIT
                  ENDDO
                  ! and ask again
                  INQUIRE (FILE="/mnt/sdcard/cp2k/work/WAIT", EXIST=should_wait)
                  IF (.NOT. should_wait) EXIT
               ENDDO
               CALL close_file(unit_number=unit_number, file_status="DELETE")
            ENDIF
            ! EXIT control sequence
            ! Check for <PROJECT_NAME>.EXIT_<FLAG>
            IF (.NOT. should_stop) THEN
               INQUIRE (FILE=exit_fname_level, EXIST=should_stop)
               IF (should_stop) THEN
                  CALL open_file(file_name=exit_fname_level, unit_number=unit_number)
                  CALL close_file(unit_number=unit_number, file_status="DELETE")
                  WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,/)") &
                     "*** "//flag//" run terminated by external request ***"
               END IF
            END IF
            ! Check for <PROJECT_NAME>.EXIT
            IF (.NOT. should_stop) THEN
               INQUIRE (FILE=exit_fname, EXIST=should_stop)
               IF (should_stop) THEN
                  WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,/)") &
                     "*** "//TRIM(flag)//" run terminated by external request ***"
               ENDIF
            END IF
            ! Check for EXIT_<FLAG>
            IF (.NOT. should_stop) THEN
               INQUIRE (FILE=exit_gname_level, EXIST=should_stop)
               IF (should_stop) THEN
                  CALL open_file(file_name=exit_gname_level, unit_number=unit_number)
                  CALL close_file(unit_number=unit_number, file_status="DELETE")
                  WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,/)") &
                     "*** "//flag//" run terminated by external request ***"
               END IF
            END IF
            ! Check for EXIT
            IF (.NOT. should_stop) THEN
               INQUIRE (FILE=exit_gname, EXIST=should_stop)
               IF (should_stop) THEN
                  WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,/)") &
                     "*** "//TRIM(flag)//" run terminated by external request ***"
               ENDIF
            END IF
         ENDIF

         IF (PRESENT(target_time)) THEN
            my_target_time = target_time
            my_start_time = start_time
         ELSEIF (PRESENT(globenv)) THEN
            my_target_time = globenv%cp2k_target_time
            my_start_time = globenv%cp2k_start_time
         ELSE
            ! If none of the two arguments is present abort.. This routine should always check about time.
            CALL cp__b("cp_external_control.F",205,"")
         END IF

         IF ((.NOT. should_stop) .AND. (my_target_time > 0.0_dp)) THEN
            ! Check for execution time
            time_check = m_walltime()-my_start_time
            IF (time_check .GT. my_target_time) THEN
               should_stop = .TRUE.
               WRITE (UNIT=cp_logger_get_default_unit_nr(logger), FMT="(/,T2,A,f12.3,A)") &
                  "*** "//TRIM(flag)//" run terminated - exceeded requested execution time:", &
                  my_target_time, " seconds.", &
                  "*** Execution time now: ", time_check, " seconds."
            END IF
         END IF
      END IF
      CALL mp_bcast(should_stop, logger%para_env%source, logger%para_env%group)

      check_always = should_stop

      CALL timestop(handle)

   END SUBROUTINE external_control

END MODULE cp_external_control

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Optimizers used by pao_main.F
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_optimizer
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_extremal, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_release, &
        cp_dbcsr_reserve_diag_blocks, cp_dbcsr_row_block_sizes, cp_dbcsr_scale, cp_dbcsr_set, &
        cp_dbcsr_trace, cp_dbcsr_type
   USE kinds,                           ONLY: dp
   USE pao_input,                       ONLY: pao_opt_bfgs,&
                                              pao_opt_cg
   USE pao_types,                       ONLY: pao_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pao_opt_init, pao_opt_finalize, pao_opt_new_dir

CONTAINS

! **************************************************************************************************
!> \brief Initialize the optimizer
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_opt_init(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CALL cp_dbcsr_init(pao%matrix_D)
      CALL cp_dbcsr_copy(pao%matrix_D, pao%matrix_G)
      CALL cp_dbcsr_set(pao%matrix_D, 0.0_dp)

      CALL cp_dbcsr_init(pao%matrix_G_prev)
      CALL cp_dbcsr_copy(pao%matrix_G_prev, pao%matrix_D)

      IF (pao%precondition) THEN
         CALL cp_dbcsr_init(pao%matrix_D_preconed)
         CALL cp_dbcsr_copy(pao%matrix_D_preconed, pao%matrix_D)
         CALL cp_dbcsr_init(pao%matrix_G_preconed)
         CALL cp_dbcsr_copy(pao%matrix_G_preconed, pao%matrix_G)
      ENDIF

      IF (pao%optimizer == pao_opt_bfgs) &
         CALL pao_opt_init_bfgs(pao)

   END SUBROUTINE pao_opt_init

! **************************************************************************************************
!> \brief Initialize the BFGS optimizer
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_opt_init_bfgs(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      INTEGER, DIMENSION(:), POINTER                     :: nparams

      nparams => cp_dbcsr_row_block_sizes(pao%matrix_X)

      CALL cp_dbcsr_init(pao%matrix_BFGS)
      CALL cp_dbcsr_create(pao%matrix_BFGS, &
                           template=pao%matrix_X, &
                           row_blk_size=nparams, &
                           col_blk_size=nparams, &
                           name="PAO matrix_BFGS")

      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_BFGS)
      CALL cp_dbcsr_set(pao%matrix_BFGS, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(pao%matrix_BFGS, 1.0_dp)

   END SUBROUTINE pao_opt_init_bfgs

! **************************************************************************************************
!> \brief Finalize the optimizer
!> \param pao ...
! **************************************************************************************************
   SUBROUTINE pao_opt_finalize(pao)
      TYPE(pao_env_type), POINTER                        :: pao

      CALL cp_dbcsr_release(pao%matrix_D)
      CALL cp_dbcsr_release(pao%matrix_G_prev)
      IF (pao%precondition) THEN
         CALL cp_dbcsr_release(pao%matrix_D_preconed)
         CALL cp_dbcsr_release(pao%matrix_G_preconed)
      ENDIF

      IF (pao%optimizer == pao_opt_bfgs) &
         CALL cp_dbcsr_release(pao%matrix_BFGS)

   END SUBROUTINE pao_opt_finalize

! **************************************************************************************************
!> \brief Calculates the new search direction.
!> \param pao ...
!> \param icycle ...
! **************************************************************************************************
   SUBROUTINE pao_opt_new_dir(pao, icycle)
      TYPE(pao_env_type), POINTER                        :: pao
      INTEGER, INTENT(IN)                                :: icycle

      CHARACTER(len=*), PARAMETER                        :: routineN = 'pao_opt_new_dir'

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (pao%precondition) THEN
         ! We can't convert matrix_D for and back every time, the numeric noise would disturb the CG,
         ! hence we keep matrix_D_preconed around.
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_precon, pao%matrix_G, &
                                0.0_dp, pao%matrix_G_preconed, retain_sparsity=.TRUE.)
         CALL pao_opt_new_dir_low(pao, icycle, pao%matrix_G_preconed, pao%matrix_G_prev, pao%matrix_D_preconed)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_precon, pao%matrix_D_preconed, &
                                0.0_dp, pao%matrix_D, retain_sparsity=.TRUE.)
         ! store preconditioned gradient for next iteration
         CALL cp_dbcsr_copy(pao%matrix_G_prev, pao%matrix_G_preconed)
         ! We keep matrix_G_preconed, because its norm is later used for convergence detection.
      ELSE
         CALL pao_opt_new_dir_low(pao, icycle, pao%matrix_G, pao%matrix_G_prev, pao%matrix_D)
         CALL cp_dbcsr_copy(pao%matrix_G_prev, pao%matrix_G) ! store gradient for next iteration
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE pao_opt_new_dir

! **************************************************************************************************
!> \brief Calculates the new search direction.
!> \param pao ...
!> \param icycle ...
!> \param matrix_G ...
!> \param matrix_G_prev ...
!> \param matrix_D ...
! **************************************************************************************************
   SUBROUTINE pao_opt_new_dir_low(pao, icycle, matrix_G, matrix_G_prev, matrix_D)
      TYPE(pao_env_type), POINTER                        :: pao
      INTEGER, INTENT(IN)                                :: icycle
      TYPE(cp_dbcsr_type)                                :: matrix_G, matrix_G_prev, matrix_D

      SELECT CASE (pao%optimizer)
      CASE (pao_opt_cg)
         CALL pao_opt_newdir_cg(pao, icycle, matrix_G, matrix_G_prev, matrix_D)
      CASE (pao_opt_bfgs)
         CALL pao_opt_newdir_bfgs(pao, icycle, matrix_G, matrix_G_prev, matrix_D)
      CASE DEFAULT
         CALL cp__b("pao_optimizer.F",153,"PAO: unknown optimizer")
      END SELECT

   END SUBROUTINE pao_opt_new_dir_low

! **************************************************************************************************
!> \brief Conjugate Gradient algorithm
!> \param pao ...
!> \param icycle ...
!> \param matrix_G ...
!> \param matrix_G_prev ...
!> \param matrix_D ...
! **************************************************************************************************
   SUBROUTINE pao_opt_newdir_cg(pao, icycle, matrix_G, matrix_G_prev, matrix_D)
      TYPE(pao_env_type), POINTER                        :: pao
      INTEGER, INTENT(IN)                                :: icycle
      TYPE(cp_dbcsr_type)                                :: matrix_G, matrix_G_prev, matrix_D

      REAL(KIND=dp)                                      :: beta, change, trace_D, trace_D_Gnew, &
                                                            trace_G_mix, trace_G_new, trace_G_prev

      ! determine CG mixing factor
      IF (icycle <= pao%cg_init_steps) THEN
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| warming up with steepest descent"
         beta = 0.0_dp
      ELSE
         CALL cp_dbcsr_trace(matrix_G, matrix_G, trace_G_new)
         CALL cp_dbcsr_trace(matrix_G_prev, matrix_G_prev, trace_G_prev)
         CALL cp_dbcsr_trace(matrix_G, matrix_G_prev, trace_G_mix)
         CALL cp_dbcsr_trace(matrix_D, matrix_G, trace_D_Gnew)
         CALL cp_dbcsr_trace(matrix_D, matrix_D, trace_D)
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| trace_G_new ", trace_G_new
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| trace_G_prev ", trace_G_prev
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| trace_G_mix ", trace_G_mix
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| trace_D ", trace_D
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| trace_D_Gnew", trace_D_Gnew

         IF (trace_G_prev /= 0.0_dp) THEN
            beta = (trace_G_new-trace_G_mix)/trace_G_prev !Polak-Ribiere
         ENDIF

         IF (beta < 0.0_dp) THEN
            IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| resetting because beta < 0"
            beta = 0.0_dp
         ENDIF

         change = trace_D_Gnew**2/trace_D*trace_G_new
         IF (change > pao%cg_reset_limit) THEN
            IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| resetting because change > CG_RESET_LIMIT"
            beta = 0.0_dp
         ENDIF

      ENDIF

      IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|CG| beta: ", beta

      ! calculate new CG direction matrix_D
      CALL cp_dbcsr_add(matrix_D, matrix_G, beta, -1.0_dp)

   END SUBROUTINE pao_opt_newdir_cg

! **************************************************************************************************
!> \brief Broyden-Fletcher-Goldfarb-Shanno algorithm
!> \param pao ...
!> \param icycle ...
!> \param matrix_G ...
!> \param matrix_G_prev ...
!> \param matrix_D ...
! **************************************************************************************************
   SUBROUTINE pao_opt_newdir_bfgs(pao, icycle, matrix_G, matrix_G_prev, matrix_D)
      TYPE(pao_env_type), POINTER                        :: pao
      INTEGER, INTENT(IN)                                :: icycle
      TYPE(cp_dbcsr_type)                                :: matrix_G, matrix_G_prev, matrix_D

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_opt_newdir_bfgs'

      INTEGER                                            :: handle
      LOGICAL                                            :: arnoldi_converged
      REAL(dp)                                           :: eval_max, eval_min, theta, trace_ry, &
                                                            trace_sy, trace_yHy, trace_yy
      TYPE(cp_dbcsr_type)                                :: matrix_Hy, matrix_Hyr, matrix_r, &
                                                            matrix_rr, matrix_ryH, matrix_ryHyr, &
                                                            matrix_s, matrix_y, matrix_yr

      CALL timeset(routineN, handle)

      !TODO add filtering?

      ! Notation according to the book from Nocedal and Wright, see chapter 6.
      IF (icycle > 1) THEN
         ! y = G - G_prev
         CALL cp_dbcsr_init(matrix_y)
         CALL cp_dbcsr_copy(matrix_y, matrix_G)
         CALL cp_dbcsr_add(matrix_y, matrix_G_prev, 1.0_dp, -1.0_dp) ! dG

         ! s = X - X_prev
         CALL cp_dbcsr_init(matrix_s)
         CALL cp_dbcsr_copy(matrix_s, matrix_D)
         CALL cp_dbcsr_scale(matrix_s, pao%linesearch%step_size) ! dX

         ! sy = MATMUL(TRANPOSE(s), y)
         CALL cp_dbcsr_trace(matrix_s, matrix_y, trace_sy)

         ! heuristic initialization
         IF (icycle == 2) THEN
            CALL cp_dbcsr_trace(matrix_Y, matrix_Y, trace_yy)
            CALL cp_dbcsr_scale(pao%matrix_BFGS, trace_sy/trace_yy)
            IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|BFGS| Initializing with:", trace_sy/trace_yy
         ENDIF

         ! Hy = MATMUL(H, y)
         CALL cp_dbcsr_init(matrix_Hy)
         CALL cp_dbcsr_create(matrix_Hy, template=matrix_G, matrix_type="N")
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_BFGS, matrix_y, 0.0_dp, matrix_Hy)

         ! yHy = MATMUL(TRANPOSE(y), Hy)
         CALL cp_dbcsr_trace(matrix_y, matrix_Hy, trace_yHy)

         ! Use damped BFGS algorithm to ensure H remains positive definite.
         ! See chapter 18 in Nocedal and Wright's book for details.
         ! The formulas were adopted to inverse Hessian algorithm.
         IF (trace_sy < 0.2_dp*trace_yHy) THEN
            theta = 0.8_dp*trace_yHy/(trace_yHy-trace_sy)
            IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|BFGS| Dampening theta:", theta
         ELSE
            theta = 1.0
         ENDIF

         ! r = theta*s + (1-theta)*Hy
         CALL cp_dbcsr_init(matrix_r)
         CALL cp_dbcsr_copy(matrix_r, matrix_s)
         CALL cp_dbcsr_add(matrix_r, matrix_Hy, theta, (1.0_dp-theta))

         ! use t instead of y to update B matrix
         CALL cp_dbcsr_trace(matrix_r, matrix_y, trace_ry)
         IF(.NOT.(trace_RY > 0.0_dp))CALL cp__a("pao_optimizer.F",288)

         ! yr = MATMUL(y, TRANSPOSE(r))
         CALL cp_dbcsr_init(matrix_yr)
         CALL cp_dbcsr_create(matrix_yr, template=pao%matrix_BFGS, matrix_type="N")
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_y, matrix_r, 0.0_dp, matrix_yr)

         ! Hyr = MATMUL(H, yr)
         CALL cp_dbcsr_init(matrix_Hyr)
         CALL cp_dbcsr_create(matrix_Hyr, template=pao%matrix_BFGS, matrix_type="N")
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pao%matrix_BFGS, matrix_yr, 0.0_dp, matrix_Hyr)

         ! ryH = MATMUL(TRANSPOSE(yr), H)
         CALL cp_dbcsr_init(matrix_ryH)
         CALL cp_dbcsr_create(matrix_ryH, template=pao%matrix_BFGS, matrix_type="N")
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, matrix_yr, pao%matrix_BFGS, 0.0_dp, matrix_ryH)

         ! ryHry = MATMUL(ryH,yr)
         CALL cp_dbcsr_init(matrix_ryHyr)
         CALL cp_dbcsr_create(matrix_ryHyr, template=pao%matrix_BFGS, matrix_type="N")
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_ryH, matrix_yr, 0.0_dp, matrix_ryHyr)

         ! rr = MATMUL(r,TRANSPOSE(r))
         CALL cp_dbcsr_init(matrix_rr)
         CALL cp_dbcsr_create(matrix_rr, template=pao%matrix_BFGS, matrix_type="N")
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_r, matrix_r, 0.0_dp, matrix_rr)

         ! H = H - Hyr/ry - ryH/ry + ryHyr/(ry**2) + rr/ry
         CALL cp_dbcsr_add(pao%matrix_BFGS, matrix_HYR, 1.0_dp, -1.0_dp/trace_ry)
         CALL cp_dbcsr_add(pao%matrix_BFGS, matrix_ryH, 1.0_dp, -1.0_dp/trace_ry)
         CALL cp_dbcsr_add(pao%matrix_BFGS, matrix_ryHyr, 1.0_dp, +1.0_dp/(trace_ry**2))
         CALL cp_dbcsr_add(pao%matrix_BFGS, matrix_rr, 1.0_dp, +1.0_dp/trace_ry)

         ! clean up
         CALL cp_dbcsr_release(matrix_y)
         CALL cp_dbcsr_release(matrix_s)
         CALL cp_dbcsr_release(matrix_r)
         CALL cp_dbcsr_release(matrix_Hy)
         CALL cp_dbcsr_release(matrix_yr)
         CALL cp_dbcsr_release(matrix_Hyr)
         CALL cp_dbcsr_release(matrix_ryH)
         CALL cp_dbcsr_release(matrix_ryHyr)
         CALL cp_dbcsr_release(matrix_rr)
      ENDIF

      ! approximate condition of Hessian
      !TODO: good setting for arnoldi?
      CALL cp_dbcsr_arnoldi_extremal(pao%matrix_BFGS, eval_max, eval_min, max_iter=100, &
                                     threshold=1e-2_dp, converged=arnoldi_converged)
      IF (arnoldi_converged) THEN
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|BFGS| evals of inv. Hessian: min, max, max/min", &
            eval_min, eval_max, eval_max/eval_min
      ELSE
         IF (pao%iw_opt > 0) WRITE (pao%iw_opt, *) "PAO|BFGS| arnoldi of inv. Hessian did not converged."
      ENDIF

      ! calculate new direction
      ! d = MATMUL(H, -g)
      CALL cp_dbcsr_multiply("N", "N", -1.0_dp, pao%matrix_BFGS, matrix_G, &
                             0.0_dp, matrix_D, retain_sparsity=.TRUE.)

      CALL timestop(handle)
   END SUBROUTINE pao_opt_newdir_bfgs

END MODULE pao_optimizer
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Main module for the PAO method
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_main
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_frobenius_norm, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_reserve_diag_blocks, cp_dbcsr_set, &
        cp_dbcsr_type
   USE cp_external_control,             ONLY: external_control
   USE dm_ls_scf_qs,                    ONLY: ls_scf_qs_atomic_guess
   USE dm_ls_scf_types,                 ONLY: ls_mstruct_type,&
                                              ls_scf_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE linesearch,                      ONLY: linesearch_finalize,&
                                              linesearch_init,&
                                              linesearch_reset,&
                                              linesearch_step
   USE machine,                         ONLY: m_walltime
   USE pao_input,                       ONLY: parse_pao_section
   USE pao_io,                          ONLY: pao_read_restart,&
                                              pao_write_restart
   USE pao_methods,                     ONLY: &
        pao_build_matrix_X, pao_build_orthogonalizer, pao_build_selector, pao_calc_energy, &
        pao_calc_outer_grad_lnv, pao_check_grad, pao_check_trace_ps, pao_print_atom_info, &
        pao_read_preopt_dm, pao_test_convergence
   USE pao_ml,                          ONLY: pao_machine_learning_init
   USE pao_optimizer,                   ONLY: pao_opt_finalize,&
                                              pao_opt_init,&
                                              pao_opt_new_dir
   USE pao_param,                       ONLY: pao_calc_U,&
                                              pao_param_finalize,&
                                              pao_param_init,&
                                              pao_param_initial_guess,&
                                              pao_update_AB
   USE pao_types,                       ONLY: pao_env_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_main'

   PUBLIC :: pao_init, pao_update, pao_finalize

CONTAINS

! **************************************************************************************************
!> \brief Initialize the PAO environment
!> \param qs_env ...
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_init(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: penalty
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(ls_mstruct_type), POINTER                     :: ls_mstruct
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(section_vals_type), POINTER                   :: input, section

      IF (.NOT. ls_scf_env%do_pao) RETURN

      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env=qs_env, &
                      matrix_s=matrix_s, &
                      input=input)

      pao => ls_scf_env%pao_env
      ls_mstruct => ls_scf_env%ls_mstruct

      CALL parse_pao_section(pao, input)

      ! get initial P and KS matrices in primary basis
      IF (LEN_TRIM(pao%preopt_dm_file) > 0) THEN
         CALL pao_read_preopt_dm(pao, qs_env)
      ELSE
         CALL ls_scf_qs_atomic_guess(qs_env, ls_scf_env%energy_init)
         IF (pao%iw > 0) WRITE (pao%iw, '(A,F20.9)') " PAO| Energy from initial atomic guess:", ls_scf_env%energy_init
         pao%need_initial_scf = .TRUE.
      ENDIF

      CALL pao_build_orthogonalizer(pao, qs_env)
      CALL pao_build_selector(qs_env, pao%matrix_Y)
      CALL pao_param_init(pao, qs_env)

      IF (pao%istep > 0) THEN ! was there a previous MD step?
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| reusing matrix_X from last MD step"
      ELSE
         CALL pao_build_matrix_X(pao, qs_env)
         IF (LEN_TRIM(pao%restart_file) > 0) THEN
            CALL pao_read_restart(pao, qs_env)
         ELSE IF (LEN_TRIM(pao%ml_training_file) > 0) THEN
            CALL pao_machine_learning_init(pao, qs_env)
         ELSE
            CALL pao_param_initial_guess(pao, qs_env)
         ENDIF
      ENDIF

      CALL pao_print_atom_info(pao)

      ! init counters
      pao%step_start_time = m_walltime()
      pao%istep = 0

      CALL cp_dbcsr_init(pao%matrix_X_orig)

      CALL cp_dbcsr_init(pao%matrix_G)
      CALL cp_dbcsr_copy(pao%matrix_G, pao%matrix_X)
      CALL cp_dbcsr_set(pao%matrix_G, 0.0_dp)

      CALL cp_dbcsr_init(pao%matrix_U)
      CALL cp_dbcsr_create(pao%matrix_U, template=matrix_s(1)%matrix, matrix_type="N", name="PAO matrix_U")
      CALL cp_dbcsr_reserve_diag_blocks(pao%matrix_U)

      CALL cp_dbcsr_init(ls_mstruct%matrix_A)
      CALL cp_dbcsr_create(ls_mstruct%matrix_A, template=pao%matrix_Y)

      CALL cp_dbcsr_init(ls_mstruct%matrix_B)
      CALL cp_dbcsr_create(ls_mstruct%matrix_B, template=pao%matrix_Y)

      CALL pao_update_AB(pao, ls_mstruct, penalty)
      IF (pao%iw > 0) WRITE (pao%iw, '(A,F20.9)') " PAO| Energy from initial penalty:", penalty

      section => section_vals_get_subs_vals(input, "DFT%LS_SCF%PAO%LINE_SEARCH")
      CALL linesearch_init(pao%linesearch, section, "PAO|")

      CALL timestop(handle)
   END SUBROUTINE pao_init

! **************************************************************************************************
!> \brief Called after the SCF optimization, updates the PAO basis.
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param pao_is_done ...
! **************************************************************************************************
   SUBROUTINE pao_update(qs_env, ls_scf_env, pao_is_done)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env
      LOGICAL, INTENT(OUT)                               :: pao_is_done

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_update', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icycle
      LOGICAL                                            :: cycle_converged, should_stop
      REAL(KIND=dp)                                      :: energy, norm_X
      TYPE(cp_dbcsr_type)                                :: matrix_M, matrix_X_start
      TYPE(ls_mstruct_type), POINTER                     :: ls_mstruct
      TYPE(pao_env_type), POINTER                        :: pao

      IF (.NOT. ls_scf_env%do_pao) THEN
         pao_is_done = .TRUE.
         RETURN
      ENDIF

      ls_mstruct => ls_scf_env%ls_mstruct
      pao => ls_scf_env%pao_env

      IF (pao%need_initial_scf) THEN
         pao_is_done = .FALSE.
         pao%need_initial_scf = .FALSE.
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| Performing initial SCF optimization."
         RETURN
      ENDIF

      CALL timeset(routineN, handle)

      ! only perform mixing if the pao%matrix_X is non-zero
      ! useful if started from a reference DM.
      norm_X = cp_dbcsr_frobenius_norm(pao%matrix_X)
      IF (pao%mixing /= 1.0_dp .AND. norm_X /= 0.0_dp) THEN
         CALL cp_dbcsr_init(matrix_X_start)
         CALL cp_dbcsr_copy(matrix_X_start, pao%matrix_X)
      ENDIF

      cycle_converged = .FALSE.
      icycle = 0
      CALL linesearch_reset(pao%linesearch)
      CALL pao_opt_init(pao)

      DO WHILE (.TRUE.)
         pao%istep = pao%istep+1

         IF (pao%iw > 0) WRITE (pao%iw, "(A,I9,A)") " PAO| ======================= Iteration: ", &
            pao%istep, " ============================="

         ! calc energy and check trace_PS
         CALL pao_calc_energy(pao, qs_env, ls_scf_env, energy)
         CALL pao_check_trace_PS(ls_scf_env)

         IF (pao%linesearch%starts) THEN
            icycle = icycle+1
            ! calc new gradient
            CALL pao_calc_outer_grad_lnv(qs_env, ls_scf_env, matrix_M)
            CALL pao_calc_U(pao, matrix_M, pao%matrix_G)
            CALL cp_dbcsr_release(matrix_M)
            CALL pao_check_grad(pao, qs_env, ls_scf_env)

            ! calculate new direction for line-search
            CALL pao_opt_new_dir(pao, icycle)

            !backup X
            CALL cp_dbcsr_copy(pao%matrix_X_orig, pao%matrix_X)

            ! print info and convergence test
            CALL pao_test_convergence(pao, ls_scf_env, energy, cycle_converged)
            IF (cycle_converged) THEN
               pao_is_done = icycle < 3
               IF (pao_is_done .AND. pao%iw > 0) WRITE (pao%iw, *) "PAO| converged after ", pao%istep, " steps :-)"
               EXIT
            ENDIF

            ! if we have reached the maximum number of cycles exit in order
            ! to restart with a fresh hamiltonian
            IF (icycle >= pao%max_cycles) THEN
               IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| CG not yet converged after ", icycle, " cylces."
               pao_is_done = .FALSE.
               EXIT
            ENDIF

            IF (MOD(icycle, pao%write_cycles) == 0) &
               CALL pao_write_restart(pao, qs_env, energy) ! write an intermediate restart file
         ENDIF

         ! check for early abort without convergence?
         CALL external_control(should_stop, "PAO", start_time=qs_env%start_time, target_time=qs_env%target_time)
         IF (should_stop .OR. pao%istep >= pao%max_pao) THEN
            CALL cp__w("pao_main.F",242,"PAO not converged!")
            pao_is_done = .TRUE.
            EXIT
         ENDIF

         ! perform line-search step
         CALL linesearch_step(pao%linesearch, &
                              energy=energy, &
                              slope=cp_dbcsr_frobenius_norm(pao%matrix_G)**2)
         IF (pao%linesearch%step_size < 1e-10_dp) CALL cp__b("pao_main.F",251,"PAO gradient is wrong.")
         CALL cp_dbcsr_copy(pao%matrix_X, pao%matrix_X_orig) !restore X
         CALL cp_dbcsr_add(pao%matrix_X, pao%matrix_D, 1.0_dp, pao%linesearch%step_size)
      ENDDO

      ! perform mixing of matrix_X
      IF (pao%mixing /= 1.0_dp .AND. norm_X /= 0.0_dp) THEN
         CALL cp_dbcsr_add(pao%matrix_X, matrix_X_start, pao%mixing, 1.0_dp-pao%mixing)
         CALL cp_dbcsr_release(matrix_X_start)
         IF (pao%iw > 0) WRITE (pao%iw, *) "PAO| Recalculating energy after mixing."
         CALL pao_calc_energy(pao, qs_env, ls_scf_env, energy)
      ENDIF

      CALL pao_write_restart(pao, qs_env, energy)
      CALL pao_opt_finalize(pao)

      CALL timestop(handle)
   END SUBROUTINE pao_update

! **************************************************************************************************
!> \brief Finalize the PAO environment
!> \param ls_scf_env ...
! **************************************************************************************************
   SUBROUTINE pao_finalize(ls_scf_env)
      TYPE(ls_scf_env_type), TARGET                      :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pao_finalize', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(ls_mstruct_type), POINTER                     :: ls_mstruct
      TYPE(pao_env_type), POINTER                        :: pao

      IF (.NOT. ls_scf_env%do_pao) RETURN

      CALL timeset(routineN, handle)
      pao => ls_scf_env%pao_env
      ls_mstruct => ls_scf_env%ls_mstruct

      CALL pao_param_finalize(pao)

      ! We keep pao%matrix_X for next scf-run, e.g. during MD or GEO-OPT
      CALL cp_dbcsr_release(pao%matrix_X_orig)
      CALL cp_dbcsr_release(pao%matrix_G)
      CALL cp_dbcsr_release(pao%matrix_U)
      CALL cp_dbcsr_release(pao%matrix_Y)
      CALL cp_dbcsr_release(pao%matrix_N)
      CALL cp_dbcsr_release(pao%matrix_N_inv)

      CALL cp_dbcsr_release(ls_mstruct%matrix_A)
      CALL cp_dbcsr_release(ls_mstruct%matrix_B)

      CALL linesearch_finalize(pao%linesearch)
      !TODO: should finish printkey
      CALL timestop(handle)
   END SUBROUTINE pao_finalize

END MODULE pao_main
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief density matrix optimization using exponential transformations
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dm_ls_scf_curvy
   USE bibliography,                    ONLY: Shao2003,&
                                              cite_reference
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_filter, cp_dbcsr_frobenius_norm, &
        cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_norm, cp_dbcsr_release, cp_dbcsr_scale, &
        cp_dbcsr_set, cp_dbcsr_trace, cp_dbcsr_transposed, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dm_ls_scf_types,                 ONLY: ls_scf_curvy_type,&
                                              ls_scf_env_type
   USE input_constants,                 ONLY: ls_scf_line_search_3point,&
                                              ls_scf_line_search_3point_2d
   USE iterate_matrix,                  ONLY: purify_mcweeny
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: ifac
   USE mathlib,                         ONLY: invmat
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_scf_curvy'

   PUBLIC :: dm_ls_curvy_optimization, deallocate_curvy_data

CONTAINS

! **************************************************************************************************
!> \brief driver routine for Head-Gordon curvy step approach
!> \param ls_scf_env ...
!> \param energy ...
!> \param check_conv ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE dm_ls_curvy_optimization(ls_scf_env, energy, check_conv)
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      REAL(KIND=dp)                                      :: energy
      LOGICAL                                            :: check_conv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dm_ls_curvy_optimization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, lsstep

      CALL timeset(routineN, handle)

      CALL cite_reference(Shao2003)

! Upon first call initialize all matrices needed curing optimization
! In addtion transform P into orthonormal basis. Will be scaled by 0.5 in closed shell case
! Only to be done once as it will be stored and reused afterwards
! TRS4 might give nondiempotent P therefore McWeeny purification is applied on initial P

      IF (.NOT. ALLOCATED(ls_scf_env%curvy_data%matrix_dp)) THEN
         CALL init_curvy(ls_scf_env%curvy_data, ls_scf_env%matrix_s, ls_scf_env%nspins)
         ls_scf_env%curvy_data%line_search_step = 1

         IF (ls_scf_env%curvy_data%line_search_type == ls_scf_line_search_3point_2d) THEN
            DO i = 1, ls_scf_env%nspins
               CALL cp_dbcsr_copy(ls_scf_env%curvy_data%matrix_psave(i, 1), &
                                  ls_scf_env%matrix_p(i))
            END DO
         END IF
         IF (ls_scf_env%nspins == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(1), 0.5_dp)
         CALL transform_matrix_orth(ls_scf_env%matrix_p, ls_scf_env%matrix_s_sqrt, &
                                    ls_scf_env%eps_filter)
         CALL purify_mcweeny(ls_scf_env%matrix_p, ls_scf_env%eps_filter, 3)
         DO i = 1, ls_scf_env%nspins
            CALL cp_dbcsr_copy(ls_scf_env%curvy_data%matrix_p(i), ls_scf_env%matrix_p(i))
         END DO
      END IF

      lsstep = ls_scf_env%curvy_data%line_search_step

! If new search direction has to be computed transform H into the orthnormal basis

      IF (ls_scf_env%curvy_data%line_search_step == 1) &
         CALL transform_matrix_orth(ls_scf_env%matrix_ks, ls_scf_env%matrix_s_sqrt_inv, &
                                    ls_scf_env%eps_filter)

! Set the energies for the line search and make sure to give the correct energy back to scf_main
      ls_scf_env%curvy_data%energies(lsstep) = energy
      IF (lsstep .NE. 1) energy = ls_scf_env%curvy_data%energies(1)

! start the optimization by calling the driver routine or simply combine saved P(2D line search)
      IF (lsstep .LE. 2) THEN
         CALL optimization_step(ls_scf_env%curvy_data, ls_scf_env)
      ELSE IF (lsstep == ls_scf_env%curvy_data%line_search_type) THEN
! line_search type has the value appropriate to the number of energy calculations needed
         CALL optimization_step(ls_scf_env%curvy_data, ls_scf_env)
      ELSE
         CALL new_p_from_save(ls_scf_env%matrix_p, ls_scf_env%curvy_data%matrix_psave, lsstep, &
                              ls_scf_env%curvy_data%double_step_size)
         ls_scf_env%curvy_data%line_search_step = ls_scf_env%curvy_data%line_search_step+1
         CALL timestop(handle)
         RETURN
      END IF
      lsstep = ls_scf_env%curvy_data%line_search_step

! transform new density matrix back into nonorthonormal basis (again scaling might apply)

      CALL transform_matrix_orth(ls_scf_env%matrix_p, ls_scf_env%matrix_s_sqrt_inv, &
                                 ls_scf_env%eps_filter)
      IF (ls_scf_env%nspins == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(1), 2.0_dp)

! P-matrices only need to be stored in case of 2D line search
      IF (lsstep .LE. 3 .AND. ls_scf_env%curvy_data%line_search_type == ls_scf_line_search_3point_2d) THEN
         DO i = 1, ls_scf_env%nspins
            CALL cp_dbcsr_copy(ls_scf_env%curvy_data%matrix_psave(i, lsstep), &
                               ls_scf_env%matrix_p(i))
         END DO
      END IF
      check_conv = lsstep == 1

      CALL timestop(handle)

   END SUBROUTINE dm_ls_curvy_optimization

! **************************************************************************************************
!> \brief low level routine for Head-Gordons curvy step approach
!>        computes gradients, performs a cg and line search,
!>        and evaluates the BCH series to obtain the new P matrix
!> \param curvy_data ...
!> \param ls_scf_env ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE optimization_step(curvy_data, ls_scf_env)
      TYPE(ls_scf_curvy_type)                            :: curvy_data
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'optimization_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: filter, step_size(2)

! Upon first line search step compute new search direction and apply CG if required

      CALL timeset(routineN, handle)

      IF (curvy_data%line_search_step == 1) THEN
         curvy_data%step_size = MAXVAL(curvy_data%step_size)
         curvy_data%step_size = MIN(MAX(0.10_dp, 0.5_dp*ABS(curvy_data%step_size(1))), 0.5_dp)
! Dynamic eps_filter for newton steps
         filter = MAX(ls_scf_env%eps_filter*curvy_data%min_filter, &
                      ls_scf_env%eps_filter*curvy_data%filter_factor)
         CALL compute_direction_newton(curvy_data%matrix_p, ls_scf_env%matrix_ks, &
                                       curvy_data%matrix_dp, filter, curvy_data%fix_shift, curvy_data%shift, &
                                       curvy_data%cg_numer, curvy_data%cg_denom, curvy_data%min_shift)
         curvy_data%filter_factor = curvy_data%scale_filter*curvy_data%filter_factor
         step_size = curvy_data%step_size
         curvy_data%BCH_saved = 0
      ELSE IF (curvy_data%line_search_step == 2) THEN
         step_size = curvy_data%step_size
         IF (curvy_data%energies(1)-curvy_data%energies(2) .GT. 0.0_dp) THEN
            curvy_data%step_size = curvy_data%step_size*2.0_dp
            curvy_data%double_step_size = .TRUE.
         ELSE
            curvy_data%step_size = curvy_data%step_size*0.5_dp
            curvy_data%double_step_size = .FALSE.
         END IF
         step_size = curvy_data%step_size
      ELSE IF (curvy_data%line_search_step == ls_scf_line_search_3point_2d) THEN
         CALL line_search_2d(curvy_data%energies, curvy_data%step_size)
         step_size = curvy_data%step_size
      ELSE IF (curvy_data%line_search_step == ls_scf_line_search_3point) THEN
         CALL line_search_3pnt(curvy_data%energies, curvy_data%step_size)
         step_size = curvy_data%step_size
      END IF

      CALL update_p_exp(curvy_data%matrix_p, ls_scf_env%matrix_p, curvy_data%matrix_dp, &
                        curvy_data%matrix_BCH, ls_scf_env%eps_filter, step_size, curvy_data%BCH_saved, &
                        curvy_data%n_bch_hist)

! line_search type has the value appropriate to the numeber of energy calculations needed
      curvy_data%line_search_step = MOD(curvy_data%line_search_step, curvy_data%line_search_type)+1
      IF (curvy_data%line_search_step == 1) THEN
         DO ispin = 1, SIZE(curvy_data%matrix_p)
            CALL cp_dbcsr_copy(curvy_data%matrix_p(ispin), ls_scf_env%matrix_p(ispin))
         END DO
      END IF
      CALL timestop(handle)

   END SUBROUTINE optimization_step

! **************************************************************************************************
!> \brief Perform a 6pnt-2D line search for spin polarized calculations.
!>        Fit a 2D parabolic function to 6 points
!> \param energies ...
!> \param step_size ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE line_search_2d(energies, step_size)
      REAL(KIND=dp)                                      :: energies(6), step_size(2)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'line_search_2d', routineP = moduleN//':'//routineN

      INTEGER                                            :: info, unit_nr
      REAL(KIND=dp)                                      :: e_pred, param(6), s1, s1sq, s2, s2sq, &
                                                            sys_lin_eq(6, 6), tmp_e, v1, v2
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      IF (energies(1)-energies(2) .LT. 0._dp) THEN
         tmp_e = energies(2); energies(2) = energies(3); energies(3) = tmp_e
         step_size = step_size*2.0_dp
      END IF
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF
      s1 = 0.5_dp*step_size(1); s2 = step_size(1); s1sq = s1**2; s2sq = s2**2
      sys_lin_eq = 0.0_dp; sys_lin_eq(:, 6) = 1.0_dp
      sys_lin_eq(2, 1) = s1sq; sys_lin_eq(2, 2) = s1sq; sys_lin_eq(2, 3) = s1sq; sys_lin_eq(2, 4) = s1; sys_lin_eq(2, 5) = s1
      sys_lin_eq(3, 1) = s2sq; sys_lin_eq(3, 2) = s2sq; sys_lin_eq(3, 3) = s2sq; sys_lin_eq(3, 4) = s2; sys_lin_eq(3, 5) = s2
      sys_lin_eq(4, 3) = s1sq; sys_lin_eq(4, 5) = s1
      sys_lin_eq(5, 1) = s1sq; sys_lin_eq(5, 4) = s1
      sys_lin_eq(6, 3) = s2sq; sys_lin_eq(6, 5) = s2

      CALL invmat(sys_lin_eq, info)
      param = MATMUL(sys_lin_eq, energies)
      v1 = (param(2)*param(4))/(2.0_dp*param(1))-param(5)
      v2 = -(param(2)**2)/(2.0_dp*param(1))+2.0_dp*param(3)
      step_size(2) = v1/v2
      step_size(1) = (-param(2)*step_size(2)-param(4))/(2.0_dp*param(1))
      IF (step_size(1) .LT. 0.0_dp) step_size(1) = 1.0_dp
      IF (step_size(2) .LT. 0.0_dp) step_size(2) = 1.0_dp
!    step_size(1)=MIN(step_size(1),2.0_dp)
!    step_size(2)=MIN(step_size(2),2.0_dp)
      e_pred = param(1)*step_size(1)**2+param(2)*step_size(1)*step_size(2)+ &
               param(3)*step_size(2)**2+param(4)*step_size(1)+param(5)*step_size(2)+param(6)
      IF (unit_nr .GT. 0) WRITE (unit_nr, "(t3,a,F10.5,F10.5,A,F20.9)") &
         " Line Search: Step Size", step_size, " Predicted energy", e_pred
      e_pred = param(1)*s1**2+param(2)*s2*s1*0.0_dp+ &
               param(3)*s1**2*0.0_dp+param(4)*s1+param(5)*s1*0.0_dp+param(6)

   END SUBROUTINE line_search_2d

! **************************************************************************************************
!> \brief Perform a 3pnt line search
!> \param energies ...
!> \param step_size ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE line_search_3pnt(energies, step_size)
      REAL(KIND=dp)                                      :: energies(3), step_size(2)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'line_search_3pnt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr
      REAL(KIND=dp)                                      :: a, b, c, e_pred, min_val, step1, tmp, &
                                                            tmp_e
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      IF (energies(1)-energies(2) .LT. 0._dp) THEN
         tmp_e = energies(2); energies(2) = energies(3); energies(3) = tmp_e
         step_size = step_size*2.0_dp
      END IF
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF
      step1 = 0.5_dp*step_size(1)
      c = energies(1)
      a = (energies(3)+c-2.0_dp*energies(2))/(2.0_dp*step1**2)
      b = (energies(2)-c-a*step1**2)/step1
      IF (a .LT. 1.0E-12_dp) a = -1.0E-12_dp
      min_val = -b/(2.0_dp*a)
      e_pred = a*min_val**2+b*min_val+c
      tmp = step_size(1)
      IF (e_pred .LT. energies(1) .AND. e_pred .LT. energies(2)) THEN
         step_size = MAX(-1.0_dp, &
                         MIN(min_val, 10_dp*step_size))
      ELSE
         step_size = 1.0_dp
      END IF
      e_pred = a*(step_size(1))**2+b*(step_size(1))+c
      IF (unit_nr .GT. 0) THEN
         WRITE (unit_nr, "(t3,a,f16.8,a,F20.9)") "Line Search: Step Size", step_size(1), " Predicted energy", e_pred
         CALL m_flush(unit_nr)
      ENDIF
   END SUBROUTINE line_search_3pnt

! **************************************************************************************************
!> \brief Get a new search direction. Iterate to obtain a Newton like step
!>        Refine with a CG update of the search direction
!> \param matrix_p ...
!> \param matrix_ks ...
!> \param matrix_dp ...
!> \param eps_filter ...
!> \param fix_shift ...
!> \param curvy_shift ...
!> \param cg_numer ...
!> \param cg_denom ...
!> \param min_shift ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE compute_direction_newton(matrix_p, matrix_ks, matrix_dp, eps_filter, fix_shift, &
                                       curvy_shift, cg_numer, cg_denom, min_shift)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix_p, matrix_ks, matrix_dp
      REAL(KIND=dp)                                      :: eps_filter
      LOGICAL                                            :: fix_shift(2)
      REAL(KIND=dp)                                      :: curvy_shift(2), cg_numer(2), &
                                                            cg_denom(2), min_shift

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_direction_newton', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, ncyc, nspin, unit_nr
      LOGICAL                                            :: at_limit
      REAL(KIND=dp)                                      :: beta, conv_val, maxel, old_conv, shift
      TYPE(cp_dbcsr_type)                                :: matrix_Ax, matrix_b, matrix_cg, &
                                                            matrix_dp_old, matrix_PKs, matrix_res, &
                                                            matrix_tmp, matrix_tmp1
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()

      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF
      CALL timeset(routineN, handle)
      nspin = SIZE(matrix_p)

      CALL cp_dbcsr_init(matrix_PKs)
      CALL cp_dbcsr_create(matrix_PKs, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_Ax)
      CALL cp_dbcsr_create(matrix_Ax, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp1)
      CALL cp_dbcsr_create(matrix_tmp1, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_res)
      CALL cp_dbcsr_create(matrix_res, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_cg)
      CALL cp_dbcsr_create(matrix_cg, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_b)
      CALL cp_dbcsr_create(matrix_b, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_dp_old)
      CALL cp_dbcsr_create(matrix_dp_old, template=matrix_dp(1), matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, nspin
         CALL cp_dbcsr_copy(matrix_dp_old, matrix_dp(ispin))

! Precompute some matrices to save work during iterations
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p(ispin), matrix_ks(ispin), &
                                0.0_dp, matrix_PKs, filter_eps=eps_filter)
         CALL cp_dbcsr_transposed(matrix_b, matrix_PKs)
         CALL cp_dbcsr_copy(matrix_cg, matrix_b)

! Starting CG with guess 0-matrix gives -2*gradient=[Ks*P-(Ks*P)T] for cg_matrix in second step
         CALL cp_dbcsr_add(matrix_cg, matrix_PKs, 2.0_dp, -2.0_dp)

! Residual matrix in first step=cg matrix. Keep Pks for later use in CG!
         CALL cp_dbcsr_copy(matrix_res, matrix_cg)

! Precompute -FP-[FP]T which will be used throughout the CG iterations
         CALL cp_dbcsr_add(matrix_b, matrix_PKs, -1.0_dp, -1.0_dp)

! Setup some values to check convergence and safty checks for eigenvalue shifting
         CALL cp_dbcsr_norm(matrix_res, which_norm=2, norm_scalar=old_conv)
         old_conv = cp_dbcsr_frobenius_norm(matrix_res)
         shift = MIN(10.0_dp, MAX(min_shift, 0.05_dp*old_conv))
         conv_val = MAX(0.010_dp*old_conv, 100.0_dp*eps_filter)
         old_conv = 100.0_dp
         IF (fix_shift(ispin)) THEN
            shift = MAX(min_shift, MIN(10.0_dp, MAX(shift, curvy_shift(ispin)-0.5_dp*curvy_shift(ispin))))
            curvy_shift(ispin) = shift
         END IF

! Begin the real optimization loop
         CALL cp_dbcsr_set(matrix_dp(ispin), 0.0_dp)
         ncyc = 10
         DO i = 1, ncyc

! One step to compute: -FPD-DPF-DFP-PFD (not obvious but symmetry allows for some tricks)
            CALL commutator_symm(matrix_b, matrix_cg, matrix_Ax, eps_filter, 1.0_dp)

! Compute the missing bits 2*(FDP+PDF) (again use symmetry to compute as a commutator)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_cg, matrix_p(ispin), &
                                   0.0_dp, matrix_tmp, filter_eps=eps_filter)
            CALL commutator_symm(matrix_ks(ispin), matrix_tmp, matrix_tmp1, eps_filter, 2.0_dp)
            CALL cp_dbcsr_add(matrix_Ax, matrix_tmp1, 1.0_dp, 1.0_dp)

! Apply the shift and hope it's enough to stabilize the CG iterations
            CALL cp_dbcsr_add(matrix_Ax, matrix_cg, 1.0_dp, shift)

            CALL compute_cg_matrices(matrix_Ax, matrix_res, matrix_cg, matrix_dp(ispin), &
                                     matrix_tmp, eps_filter, at_limit)
            CALL cp_dbcsr_filter(matrix_cg, eps_filter)

! check for convergence of the newton step
            maxel = cp_dbcsr_frobenius_norm(matrix_res)
            IF (unit_nr .GT. 0) THEN
               WRITE (unit_nr, "(T3,A,F12.6)") "Convergence of Newton iteration ", maxel
               CALL m_flush(unit_nr)
            ENDIF
            at_limit = at_limit .OR. (old_conv/maxel .LT. 1.01_dp)
            old_conv = maxel
            IF (i == ncyc .AND. maxel/conv_val .GT. 5.0_dp) THEN
               fix_shift(ispin) = .TRUE.
               curvy_shift(ispin) = 4.0_dp*shift
            END IF
            IF (maxel .LT. conv_val .OR. at_limit) EXIT
         END DO

! Refine the Newton like search direction with a preconditioned cg update
         CALL cp_dbcsr_transposed(matrix_b, matrix_PKs)
         !compute b= -2*KsP+2*PKs=-(2*gradient)
         CALL cp_dbcsr_copy(matrix_cg, matrix_b)
         CALL cp_dbcsr_add(matrix_cg, matrix_PKs, 1.0_dp, -1.0_dp)
         cg_denom(ispin) = cg_numer(ispin)
         CALL cp_dbcsr_trace(matrix_cg, matrix_dp(ispin), cg_numer(ispin))
         beta = cg_numer(ispin)/MAX(cg_denom(ispin), 1.0E-6_dp)
         IF (beta .LT. 1.0_dp) THEN
            beta = MAX(0.0_dp, beta)
            CALL cp_dbcsr_add(matrix_dp(ispin), matrix_dp_old, 1.0_dp, beta)
         END IF
         IF (unit_nr .GT. 0) WRITE (unit_nr, "(A)") " "
      END DO

      CALL cp_dbcsr_release(matrix_PKs)
      CALL cp_dbcsr_release(matrix_dp_old)
      CALL cp_dbcsr_release(matrix_b)
      CALL cp_dbcsr_release(matrix_Ax)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_tmp1)
      CALL cp_dbcsr_release(matrix_b)
      CALL cp_dbcsr_release(matrix_res)
      CALL cp_dbcsr_release(matrix_cg)

      IF (unit_nr .GT. 0) CALL m_flush(unit_nr)
      CALL timestop(handle)
   END SUBROUTINE compute_direction_newton

! **************************************************************************************************
!> \brief compute the optimal step size of the current cycle and update the
!>        matrices needed to solve the system of linear equations
!> \param Ax ...
!> \param res ...
!> \param cg ...
!> \param deltp ...
!> \param tmp ...
!> \param eps_filter ...
!> \param at_limit ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE compute_cg_matrices(Ax, res, cg, deltp, tmp, eps_filter, at_limit)
      TYPE(cp_dbcsr_type)                                :: Ax, res, cg, deltp, tmp
      REAL(KIND=dp)                                      :: eps_filter
      LOGICAL                                            :: at_limit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_cg_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info
      REAL(KIND=dp)                                      :: alpha, beta, devi(3), fac, fac1, &
                                                            lin_eq(3, 3), new_norm, norm_cA, &
                                                            norm_rr, vec(3)

      at_limit = .FALSE.
      CALL cp_dbcsr_trace(res, res, norm_rr)
      CALL cp_dbcsr_trace(cg, Ax, norm_cA)
      lin_eq = 0.0_dp
      fac = norm_rr/norm_cA
      fac1 = fac
! Use a 3point line serach and a fit to a quadratic function to determine optimal step size
      DO i = 1, 3
         CALL cp_dbcsr_copy(tmp, res)
         CALL cp_dbcsr_add(tmp, Ax, 1.0_dp, -fac)
         devi(i) = cp_dbcsr_frobenius_norm(tmp)
         lin_eq(i, :) = (/fac**2, fac, 1.0_dp/)
         fac = fac1+fac1*((-1)**i)*0.5_dp
      END DO
      CALL invmat(lin_eq, info)
      vec = MATMUL(lin_eq, devi)
      alpha = -vec(2)/(2.0_dp*vec(1))
      fac = SQRT(norm_rr/(norm_cA*alpha))
!scale the previous matrices to match the step size
      CALL cp_dbcsr_scale(Ax, fac)
      CALL cp_dbcsr_scale(cg, fac)
      norm_cA = norm_cA*fac**2

! USe CG to get the new matrices
      alpha = norm_rr/norm_cA
      CALL cp_dbcsr_add(res, Ax, 1.0_dp, -alpha)
      CALL cp_dbcsr_trace(res, res, new_norm)
      IF (norm_rr .LT. eps_filter*0.001_dp .OR. new_norm .LT. eps_filter*0.001_dp) THEN
         beta = 0.0_dp
         at_limit = .TRUE.
      ELSE
         beta = new_norm/norm_rr
         CALL cp_dbcsr_add(deltp, cg, 1.0_dp, alpha)
      END IF
      beta = new_norm/norm_rr
      CALL cp_dbcsr_add(cg, res, beta, 1.0_dp)

   END SUBROUTINE compute_cg_matrices

! **************************************************************************************************
!> \brief Only for 2D line serach. Use saved P-components to construct new
!>        test density matrix. Takes care as well, whether step_size
!>        increased or decreased during 2nd step and combines matrices accordingly
!> \param matrix_p ...
!> \param matrix_psave ...
!> \param lsstep ...
!> \param DOUBLE ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE new_p_from_save(matrix_p, matrix_psave, lsstep, DOUBLE)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix_p
      TYPE(cp_dbcsr_type), DIMENSION(:, :)               :: matrix_psave
      INTEGER                                            :: lsstep
      LOGICAL                                            :: DOUBLE

      SELECT CASE (lsstep)
      CASE (3)
         CALL cp_dbcsr_copy(matrix_p(1), matrix_psave(1, 1))
         IF (DOUBLE) THEN
            CALL cp_dbcsr_copy(matrix_p(2), matrix_psave(2, 2))
         ELSE
            CALL cp_dbcsr_copy(matrix_p(2), matrix_psave(2, 3))
         END IF
      CASE (4)
         IF (DOUBLE) THEN
            CALL cp_dbcsr_copy(matrix_p(1), matrix_psave(1, 2))
         ELSE
            CALL cp_dbcsr_copy(matrix_p(1), matrix_psave(1, 3))
         END IF
         CALL cp_dbcsr_copy(matrix_p(2), matrix_psave(2, 1))
      CASE (5)
         CALL cp_dbcsr_copy(matrix_p(1), matrix_psave(1, 1))
         IF (DOUBLE) THEN
            CALL cp_dbcsr_copy(matrix_p(2), matrix_psave(2, 3))
         ELSE
            CALL cp_dbcsr_copy(matrix_p(2), matrix_psave(2, 2))
         END IF
      END SELECT

   END SUBROUTINE new_p_from_save

! **************************************************************************************************
!> \brief computes a commutator exploiting symmetry RES=k*[A,B]=k*[AB-(AB)T]
!> \param a ...
!> \param b ...
!> \param res ...
!> \param eps_filter   filtering threshold for sparse matrices
!> \param prefac      prefactor k in above equation
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE commutator_symm(a, b, res, eps_filter, prefac)
      TYPE(cp_dbcsr_type)                                :: a, b, res
      REAL(KIND=dp)                                      :: eps_filter, prefac

      CHARACTER(LEN=*), PARAMETER :: routineN = 'commutator_symm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: work

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(work)
      CALL cp_dbcsr_create(work, template=a, matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_multiply("N", "N", prefac, a, b, 0.0_dp, res, filter_eps=eps_filter)
      CALL cp_dbcsr_transposed(work, res)
      CALL cp_dbcsr_add(res, work, 1.0_dp, -1.0_dp)

      CALL cp_dbcsr_release(work)

      CALL timestop(handle)
   END SUBROUTINE commutator_symm

! **************************************************************************************************
!> \brief Use the BCH update to get the new idempotent P
!>        Numerics don't allow for perfect idempotency, therefore a mc weeny
!>        step is used to make sure we stay close to the idempotent surface
!> \param matrix_p_in ...
!> \param matrix_p_out ...
!> \param matrix_dp ...
!> \param matrix_BCH ...
!> \param threshold ...
!> \param step_size ...
!> \param BCH_saved ...
!> \param n_bch_hist ...
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE update_p_exp(matrix_p_in, matrix_p_out, matrix_dp, matrix_BCH, threshold, step_size, &
                           BCH_saved, n_bch_hist)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix_p_in, matrix_p_out, matrix_dp
      TYPE(cp_dbcsr_type), DIMENSION(:, :)               :: matrix_BCH
      REAL(KIND=dp)                                      :: threshold, step_size(2)
      INTEGER                                            :: BCH_saved(2), n_bch_hist

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_p_exp', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, nsave, nspin, unit_nr
      LOGICAL                                            :: save_BCH
      REAL(KIND=dp)                                      :: frob_norm, step_fac
      TYPE(cp_dbcsr_type)                                :: matrix, matrix_tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL cp_dbcsr_init(matrix)
      CALL cp_dbcsr_create(matrix, template=matrix_p_in(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_p_in(1), matrix_type=dbcsr_type_no_symmetry)
      nspin = SIZE(matrix_p_in)

      DO ispin = 1, nspin
         step_fac = 1.0_dp
         frob_norm = 1.0_dp
         nsave = 0

         CALL cp_dbcsr_copy(matrix_tmp, matrix_p_in(ispin))
         CALL cp_dbcsr_copy(matrix_p_out(ispin), matrix_p_in(ispin))
! If a BCH history is used make good use of it and do a few steps as a copy and scale update of P
! else BCH_saved will be 0 and loop is skipped
         DO i = 1, BCH_saved(ispin)
            step_fac = step_fac*step_size(ispin)
            CALL cp_dbcsr_copy(matrix_tmp, matrix_p_out(ispin))
            CALL cp_dbcsr_add(matrix_p_out(ispin), matrix_BCH(ispin, i), 1.0_dp, ifac(i)*step_fac)
            CALL cp_dbcsr_add(matrix_tmp, matrix_p_out(ispin), 1.0_dp, -1.0_dp)
            frob_norm = cp_dbcsr_frobenius_norm(matrix_tmp)
            IF (unit_nr .GT. 0) WRITE (unit_nr, "(t3,a,i3,a,f16.8)") "BCH: step", i, " Norm of P_old-Pnew:", frob_norm
            IF (frob_norm .LT. threshold) EXIT
         END DO
         IF (frob_norm .LT. threshold) CYCLE

! If the copy and scale isn't enough compute a few more BCH steps. 20 seems high but except of the first step it will never be close
         save_BCH = BCH_saved(ispin) == 0 .AND. n_bch_hist .GT. 0
         DO i = BCH_saved(ispin)+1, 20
            step_fac = step_fac*step_size(ispin)
            !allow for a bit of matrix magic here by exploiting matrix and matrix_tmp
            !matrix_tmp is alway the previous order of the BCH series
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp, matrix_dp(ispin), &
                                   0.0_dp, matrix, filter_eps=threshold)

            !(anti)symmetry allows to sum the transposed instead of the full commutator, matrix becomes the latest result

            CALL cp_dbcsr_transposed(matrix_tmp, matrix)
            CALL cp_dbcsr_add(matrix, matrix_tmp, 1.0_dp, 1.0_dp)

            !Finally, add the new BCH order to P, but store the previous one for a convergence check
            CALL cp_dbcsr_copy(matrix_tmp, matrix_p_out(ispin))
            CALL cp_dbcsr_add(matrix_p_out(ispin), matrix, 1.0_dp, ifac(i)*step_fac)
            IF (save_BCH .AND. i .LE. n_bch_hist) THEN
               CALL cp_dbcsr_copy(matrix_BCH(ispin, i), matrix)
               nsave = i
            END IF

            CALL cp_dbcsr_add(matrix_tmp, matrix_p_out(ispin), 1.0_dp, -1.0_dp)

            !Stop the BCH-series if two successive P's differ by less the threshold
            frob_norm = cp_dbcsr_frobenius_norm(matrix_tmp)
            IF (unit_nr .GT. 0) WRITE (unit_nr, "(t3,a,i3,a,f16.8)") "BCH: step", i, " Norm of P_old-Pnew:", frob_norm
            IF (frob_norm .LT. threshold) EXIT

            !Copy the latest BCH-matrix on matrix tmp, so we can cycle with all matrices in place
            CALL cp_dbcsr_copy(matrix_tmp, matrix)
            CALL cp_dbcsr_filter(matrix_tmp, threshold)
         END DO
         BCH_saved(ispin) = nsave
         IF (unit_nr .GT. 0) WRITE (unit_nr, "(A)") " "
      END DO

      CALL purify_mcweeny(matrix_p_out, threshold, 1)
      IF (unit_nr .GT. 0) CALL m_flush(unit_nr)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix)
      CALL timestop(handle)
   END SUBROUTINE update_p_exp

! **************************************************************************************************
!> \brief performs a tranformation of a matrix back to/into orthonormal basis
!>        in case of P a scaling of 0.5 has to be applied for closed shell case
!> \param matrix       matrix to be transformed
!> \param matrix_trafo transformation matrix
!> \param eps_filter   filtering threshold for sparse matrices
!> \par History
!>       2012.05 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE transform_matrix_orth(matrix, matrix_trafo, eps_filter)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix
      TYPE(cp_dbcsr_type)                                :: matrix_trafo
      REAL(KIND=dp)                                      :: eps_filter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transform_matrix_orth', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      TYPE(cp_dbcsr_type)                                :: matrix_tmp, matrix_work

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(matrix_work)
      CALL cp_dbcsr_create(matrix_work, template=matrix(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix(1), matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, SIZE(matrix)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix(ispin), matrix_trafo, &
                                0.0_dp, matrix_work, filter_eps=eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_trafo, matrix_work, &
                                0.0_dp, matrix_tmp, filter_eps=eps_filter)
         ! symmetrize results (this is again needed to make sure everything is stable)
         CALL cp_dbcsr_transposed(matrix_work, matrix_tmp)
         CALL cp_dbcsr_add(matrix_tmp, matrix_work, 0.5_dp, 0.5_dp)
         CALL cp_dbcsr_copy(matrix(ispin), matrix_tmp)
      END DO

      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_work)
      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param curvy_data ...
! **************************************************************************************************
   SUBROUTINE deallocate_curvy_data(curvy_data)
      TYPE(ls_scf_curvy_type)                            :: curvy_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_curvy_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      CALL release_dbcsr_array(curvy_data%matrix_dp)
      CALL release_dbcsr_array(curvy_data%matrix_p)

      IF (ALLOCATED(curvy_data%matrix_psave)) THEN
         DO i = 1, SIZE(curvy_data%matrix_psave, 1)
            DO j = 1, 3
               CALL cp_dbcsr_release(curvy_data%matrix_psave(i, j))
            END DO
         END DO
         DEALLOCATE (curvy_data%matrix_psave)
      END IF
      IF (ALLOCATED(curvy_data%matrix_BCH)) THEN
         DO i = 1, SIZE(curvy_data%matrix_BCH, 1)
            DO j = 1, 7
               CALL cp_dbcsr_release(curvy_data%matrix_BCH(i, j))
            END DO
         END DO
         DEALLOCATE (curvy_data%matrix_BCH)
      END IF
   END SUBROUTINE deallocate_curvy_data

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE release_dbcsr_array(matrix)
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: matrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_dbcsr_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ALLOCATED(matrix)) THEN
         DO i = 1, SIZE(matrix)
            CALL cp_dbcsr_release(matrix(i))
         END DO
         DEALLOCATE (matrix)
      END IF
   END SUBROUTINE release_dbcsr_array

! **************************************************************************************************
!> \brief ...
!> \param curvy_data ...
!> \param matrix_s ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE init_curvy(curvy_data, matrix_s, nspins)
      TYPE(ls_scf_curvy_type)                            :: curvy_data
      TYPE(cp_dbcsr_type)                                :: matrix_s
      INTEGER                                            :: nspins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_curvy', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, j

      ALLOCATE (curvy_data%matrix_dp(nspins))
      ALLOCATE (curvy_data%matrix_p(nspins))
      DO ispin = 1, nspins
         CALL cp_dbcsr_init(curvy_data%matrix_dp(ispin))
         CALL cp_dbcsr_create(curvy_data%matrix_dp(ispin), template=matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_set(curvy_data%matrix_dp(ispin), 0.0_dp)
         CALL cp_dbcsr_init(curvy_data%matrix_p(ispin))
         CALL cp_dbcsr_create(curvy_data%matrix_p(ispin), template=matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
         curvy_data%fix_shift = .FALSE.
         curvy_data%double_step_size = .TRUE.
         curvy_data%shift = 1.0_dp
         curvy_data%BCH_saved = 0
         curvy_data%step_size = 0.60_dp
         curvy_data%cg_numer = 0.00_dp
         curvy_data%cg_denom = 0.00_dp
      END DO
      IF (curvy_data%line_search_type == ls_scf_line_search_3point_2d) THEN
         ALLOCATE (curvy_data%matrix_psave(nspins, 3))
         DO ispin = 1, nspins
            DO j = 1, 3
               CALL cp_dbcsr_init(curvy_data%matrix_psave(ispin, j))
               CALL cp_dbcsr_create(curvy_data%matrix_psave(ispin, j), template=matrix_s, &
                                    matrix_type=dbcsr_type_no_symmetry)
            END DO
         END DO
      END IF
      IF (curvy_data%n_bch_hist .GT. 0) THEN
         ALLOCATE (curvy_data%matrix_BCH(nspins, curvy_data%n_bch_hist))
         DO ispin = 1, nspins
            DO j = 1, curvy_data%n_bch_hist
               CALL cp_dbcsr_init(curvy_data%matrix_BCH(ispin, j))
               CALL cp_dbcsr_create(curvy_data%matrix_BCH(ispin, j), template=matrix_s, &
                                    matrix_type=dbcsr_type_no_symmetry)
            END DO
         END DO
      END IF

   END SUBROUTINE init_curvy

END MODULE dm_ls_scf_curvy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The types needed for the calculation of Hirshfeld charges and
!>        related functions
!> \par History
!>      11.2014 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE hirshfeld_types

   USE input_constants,                 ONLY: shape_function_gaussian
   USE kinds,                           ONLY: dp
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hirshfeld_types'

   PUBLIC :: hirshfeld_type
   PUBLIC :: create_hirshfeld_type, release_hirshfeld_type
   PUBLIC :: get_hirshfeld_info, set_hirshfeld_info

! **************************************************************************************************
!> \brief quantities needed for a Hischfeld based partitioning of real space
!> \author JGH
! **************************************************************************************************
   TYPE hirshfeld_type
      LOGICAL                       :: iterative
      INTEGER                       :: shape_function_type
      INTEGER                       :: ref_charge
      TYPE(shape_fn), DIMENSION(:), &
         POINTER                    :: kind_shape_fn
      REAL(KIND=dp), DIMENSION(:), &
         POINTER                    :: charges
      TYPE(pw_p_type), POINTER      :: fnorm
   END TYPE hirshfeld_type

   TYPE shape_fn
      INTEGER                       :: numexp
      REAL(KIND=dp), DIMENSION(:), &
         POINTER                    :: zet
      REAL(KIND=dp), DIMENSION(:), &
         POINTER                    :: coef
   END TYPE shape_fn

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param hirshfeld_env ...
! **************************************************************************************************
   SUBROUTINE create_hirshfeld_type(hirshfeld_env)
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hirshfeld_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(hirshfeld_env)) THEN
         CALL release_hirshfeld_type(hirshfeld_env)
      END IF

      ALLOCATE (hirshfeld_env)

      hirshfeld_env%iterative = .FALSE.
      hirshfeld_env%shape_function_type = shape_function_gaussian
      NULLIFY (hirshfeld_env%kind_shape_fn)
      NULLIFY (hirshfeld_env%charges)
      NULLIFY (hirshfeld_env%fnorm)

   END SUBROUTINE create_hirshfeld_type

! **************************************************************************************************
!> \brief ...
!> \param hirshfeld_env ...
! **************************************************************************************************
   SUBROUTINE release_hirshfeld_type(hirshfeld_env)
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env

      CHARACTER(len=*), PARAMETER :: routineN = 'release_hirshfeld_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind
      TYPE(shape_fn), DIMENSION(:), POINTER              :: kind_shape

      IF (ASSOCIATED(hirshfeld_env)) THEN

         IF (ASSOCIATED(hirshfeld_env%kind_shape_fn)) THEN
            kind_shape => hirshfeld_env%kind_shape_fn
            DO ikind = 1, SIZE(kind_shape)
               IF (ASSOCIATED(hirshfeld_env%kind_shape_fn(ikind)%zet)) THEN
                  DEALLOCATE (kind_shape(ikind)%zet)
               END IF
               IF (ASSOCIATED(hirshfeld_env%kind_shape_fn(ikind)%coef)) THEN
                  DEALLOCATE (kind_shape(ikind)%coef)
               END IF
            END DO
            DEALLOCATE (kind_shape)
         END IF

         IF (ASSOCIATED(hirshfeld_env%charges)) THEN
            DEALLOCATE (hirshfeld_env%charges)
         END IF

         IF (ASSOCIATED(hirshfeld_env%fnorm)) THEN
            CALL pw_release(hirshfeld_env%fnorm%pw)
            DEALLOCATE (hirshfeld_env%fnorm)
         ENDIF

         DEALLOCATE (hirshfeld_env)

      END IF

   END SUBROUTINE release_hirshfeld_type

! **************************************************************************************************
!> \brief ...
!> \param hirshfeld_env ...
!> \param shape_function_type ...
!> \param iterative ...
!> \param ref_charge ...
!> \param fnorm ...
! **************************************************************************************************
   SUBROUTINE get_hirshfeld_info(hirshfeld_env, shape_function_type, iterative, &
                                 ref_charge, fnorm)
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      INTEGER, INTENT(OUT), OPTIONAL                     :: shape_function_type
      LOGICAL, INTENT(OUT), OPTIONAL                     :: iterative
      INTEGER, INTENT(OUT), OPTIONAL                     :: ref_charge
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: fnorm

      CHARACTER(len=*), PARAMETER :: routineN = 'get_hirshfeld_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(hirshfeld_env)))CALL cp__a("hirshfeld_types.F",143)

      IF (PRESENT(shape_function_type)) THEN
         shape_function_type = hirshfeld_env%shape_function_type
      END IF
      IF (PRESENT(iterative)) THEN
         iterative = hirshfeld_env%iterative
      END IF
      IF (PRESENT(ref_charge)) THEN
         ref_charge = hirshfeld_env%ref_charge
      END IF
      IF (PRESENT(fnorm)) THEN
         fnorm => hirshfeld_env%fnorm
      END IF

   END SUBROUTINE get_hirshfeld_info

! **************************************************************************************************
!> \brief ...
!> \param hirshfeld_env ...
!> \param shape_function_type ...
!> \param iterative ...
!> \param ref_charge ...
!> \param fnorm ...
! **************************************************************************************************
   SUBROUTINE set_hirshfeld_info(hirshfeld_env, shape_function_type, iterative, &
                                 ref_charge, fnorm)
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      INTEGER, INTENT(IN), OPTIONAL                      :: shape_function_type
      LOGICAL, INTENT(IN), OPTIONAL                      :: iterative
      INTEGER, INTENT(IN), OPTIONAL                      :: ref_charge
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: fnorm

      CHARACTER(len=*), PARAMETER :: routineN = 'set_hirshfeld_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(hirshfeld_env)))CALL cp__a("hirshfeld_types.F",179)

      IF (PRESENT(shape_function_type)) THEN
         hirshfeld_env%shape_function_type = shape_function_type
      END IF
      IF (PRESENT(iterative)) THEN
         hirshfeld_env%iterative = iterative
      END IF
      IF (PRESENT(ref_charge)) THEN
         hirshfeld_env%ref_charge = ref_charge
      END IF
      IF (PRESENT(fnorm)) THEN
         hirshfeld_env%fnorm => fnorm
      END IF

   END SUBROUTINE set_hirshfeld_info
! **************************************************************************************************

END MODULE hirshfeld_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides interfaces to LAPACK eigenvalue/SVD routines
!> \note
!>      We are using LAPACK interfaces, so please make sure in IBM/AIX you have
!>      the lapack library before essl: "xlf90 ... -llapack -lessl" !!!
!> \par History
!>      JGH (26-5-2001): delay D/S C/Z problem to the lapack library call
!> \author APSI
! **************************************************************************************************
MODULE eigenvalueproblems

   USE kinds,                           ONLY: dp
   USE lapack,                          ONLY: lapack_cgesvd,&
                                              lapack_chpev,&
                                              lapack_sgesvd,&
                                              lapack_ssyev
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: diagonalise

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eigenvalueproblems'

   INTERFACE diagonalise
      MODULE PROCEDURE diagonalise_ssyev
      MODULE PROCEDURE diagonalise_chpev
   END INTERFACE

   INTERFACE singular_values
      MODULE PROCEDURE cp2k_sgesvd
      MODULE PROCEDURE cp2k_cgesvd
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param mysize ...
!> \param storageform ...
!> \param eigenvalues ...
!> \param eigenvectors ...
! **************************************************************************************************
   SUBROUTINE diagonalise_ssyev(matrix, mysize, storageform, eigenvalues, &
                                eigenvectors)

      REAL(KIND=dp), INTENT(IN)                          :: matrix(:, :)
      INTEGER, INTENT(IN)                                :: mysize
      CHARACTER(LEN=*), INTENT(IN)                       :: storageform
      REAL(KIND=dp), INTENT(OUT)                         :: eigenvalues(:), eigenvectors(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'diagonalise_ssyev', &
         routineP = moduleN//':'//routineN
      CHARACTER, PARAMETER                               :: jobz = "V"

      CHARACTER                                          :: uplo
      INTEGER                                            :: info, lda, lwork
      REAL(KIND=dp)                                      :: work(3*mysize-1)

      IF (storageform(1:5) == "Lower" .OR. &
          storageform(1:5) == "LOWER" .OR. &
          storageform(1:5) == "lower") THEN
         uplo = "L"
      ELSE IF (storageform(1:5) == "Upper" .OR. &
               storageform(1:5) == "upper" .OR. &
               storageform(1:5) == "UPPER") THEN
         uplo = "U"
      ELSE
         CALL cp__b("common/eigenvalueproblems.F",77,"Unknown form of storage")
      END IF

      lda = SIZE(matrix, 1)
      lwork = 3*mysize-1

      eigenvectors = matrix

      CALL lapack_ssyev(jobz, uplo, mysize, eigenvectors, lda, eigenvalues, &
                        work, lwork, info)
      IF (info /= 0) THEN
         CALL cp__b("common/eigenvalueproblems.F",88,"Error in diagonalisation")
      END IF

   END SUBROUTINE diagonalise_ssyev

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param mysize ...
!> \param storageform ...
!> \param eigenvalues ...
!> \param eigenvectors ...
! **************************************************************************************************
   SUBROUTINE diagonalise_chpev(matrix, mysize, storageform, eigenvalues, &
                                eigenvectors)

      COMPLEX(KIND=dp), INTENT(INOUT)                    :: matrix(:)
      INTEGER, INTENT(IN)                                :: mysize
      CHARACTER(LEN=*), INTENT(IN)                       :: storageform
      REAL(KIND=dp), INTENT(OUT)                         :: eigenvalues(:)
      COMPLEX(KIND=dp), INTENT(OUT)                      :: eigenvectors(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'diagonalise_chpev', &
         routineP = moduleN//':'//routineN
      CHARACTER, PARAMETER                               :: jobz = "V"

      CHARACTER                                          :: uplo
      INTEGER                                            :: info
      COMPLEX(KIND=dp)                                   :: work(2*mysize-1)
      REAL(KIND=dp)                                      :: rwork(3*mysize-2)

      IF (storageform(1:5) == "Lower" .OR. &
          storageform(1:5) == "LOWER" .OR. &
          storageform(1:5) == "lower") THEN
         uplo = "L"
      ELSE IF (storageform(1:5) == "Upper" .OR. &
               storageform(1:5) == "upper" .OR. &
               storageform(1:5) == "UPPER") THEN
         uplo = "U"
      ELSE
         CALL cp__b("common/eigenvalueproblems.F",128,"Unknown form of storage")
      END IF

      CALL lapack_chpev(jobz, uplo, mysize, matrix, eigenvalues, &
                        eigenvectors, mysize, work, rwork, info)
      IF (info /= 0) THEN
         CALL cp__b("common/eigenvalueproblems.F",134,"Error in diagonalisation")
      END IF

   END SUBROUTINE diagonalise_chpev

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param svalues ...
!> \param mrow ...
!> \param ncol ...
!> \param uvec ...
!> \param vtvec ...
! **************************************************************************************************
   SUBROUTINE cp2k_sgesvd(matrix, svalues, mrow, ncol, uvec, vtvec)

      REAL(KIND=dp), INTENT(IN)                          :: matrix(:, :)
      REAL(KIND=dp), INTENT(OUT)                         :: svalues(:)
      INTEGER, INTENT(IN)                                :: mrow, ncol
      REAL(KIND=dp), INTENT(OUT)                         :: uvec(:, :), vtvec(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_sgesvd', routineP = moduleN//':'//routineN
      CHARACTER, PARAMETER                               :: jobu = "A", jobvt = "A"

      INTEGER                                            :: info, lda, ldu, ldvt, lwork
      REAL(KIND=dp)                                      :: work(25*(mrow+ncol))

      lwork = 25*(mrow+ncol)
      lda = SIZE(matrix, 1)
      ldu = SIZE(uvec, 1)
      ldvt = SIZE(vtvec, 1)

      CALL lapack_sgesvd(jobu, jobvt, mrow, ncol, matrix, lda, svalues, &
                         uvec, ldu, vtvec, ldvt, work, lwork, info)
      IF (info /= 0) THEN
         CALL cp__b("common/eigenvalueproblems.F",169,"Error in singular value decomposition.")
      END IF

   END SUBROUTINE cp2k_sgesvd

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param svalues ...
!> \param mrow ...
!> \param ncol ...
!> \param uvec ...
!> \param vtvec ...
! **************************************************************************************************
   SUBROUTINE cp2k_cgesvd(matrix, svalues, mrow, ncol, uvec, vtvec)

      COMPLEX(KIND=dp), INTENT(IN)                       :: matrix(:, :)
      REAL(KIND=dp), INTENT(OUT)                         :: svalues(:)
      INTEGER, INTENT(IN)                                :: mrow, ncol
      COMPLEX(KIND=dp), INTENT(OUT)                      :: uvec(:, :), vtvec(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_cgesvd', routineP = moduleN//':'//routineN
      CHARACTER, PARAMETER                               :: jobu = "A", jobvt = "A"

      INTEGER                                            :: info, lda, ldu, ldvt, lwork
      COMPLEX(KIND=dp)                                   :: work(25*(mrow+ncol))
      REAL(KIND=dp)                                      :: rwork(25*(mrow+ncol))

      lwork = 25*(mrow+ncol)
      lda = SIZE(matrix, 1)
      ldu = SIZE(uvec, 1)
      ldvt = SIZE(vtvec, 1)

      CALL lapack_cgesvd(jobu, jobvt, mrow, ncol, matrix, lda, svalues, &
                         uvec, ldu, vtvec, ldvt, work, lwork, rwork, info)
      IF (info /= 0) THEN
         CALL cp__b("common/eigenvalueproblems.F",205,"Error in singular value decomposition.")
      END IF

   END SUBROUTINE cp2k_cgesvd

END MODULE eigenvalueproblems

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates electric field gradients
!>      H.M. Petrili, P.E. Blochl, P. Blaha, K. Schwarz, PRB 57, 14690 (1998)
!> \par History
!>      12.2007 Added checksum for interpolation regtest [rdeclerck]
!> \author JGH (03-05-2006)
! **************************************************************************************************
MODULE qs_electric_field_gradient
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE eigenvalueproblems,              ONLY: diagonalise
   USE input_section_types,             ONLY: section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac,&
                                              fourpi
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset
   USE particle_types,                  ONLY: particle_type
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE physcon,                         ONLY: a_bohr,&
                                              e_charge,&
                                              joule
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_dr2,&
                                              pw_integral_ab,&
                                              pw_smoothing,&
                                              pw_structure_factor,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: &
        Eval_Interp_Spl3_pbc, Eval_d_Interp_Spl3_pbc, find_coeffs, pw_spline_do_precond, &
        pw_spline_precond_create, pw_spline_precond_release, pw_spline_precond_set_kind, &
        pw_spline_precond_type, spl3_pbc
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_methods,                   ONLY: calc_rho_tot_gspace
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_type
   USE util,                            ONLY: get_limit,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: qs_efg_calc

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_electric_field_gradient'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE qs_efg_calc(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_efg_calc', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: aint_precond, handle, i, iat, iatom, ij, &
                                                            ikind, j, max_iter, natom, natomkind, &
                                                            nkind, nspins, precond_kind, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: efg_debug, efg_interpolation, gapw, &
                                                            paw_atom, smoothing, success
      REAL(KIND=dp)                                      :: chk_spl, ecut, efg_units, efg_val, &
                                                            ehartree, eps_r, eps_x, f1, f2, sigma
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: efg_diagval, vh0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: efg_pw, efg_tensor
      REAL(KIND=dp), DIMENSION(3)                        :: eigenvalues, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: eigenvectors
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rvals
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: dvr2rs, rho_tot_gspace, &
                                                            structure_factor, v_hartree_gspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: dvr2, dvspl
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(section_vals_type), POINTER                   :: dft_section, input, interp_section

      NULLIFY (atomic_kind_set, qs_kind_set, dft_control, para_env, particle_set, rho, &
               rho_atom_set, input, dft_section, interp_section)

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      chk_spl = 0.0_dp
      efg_units = Joule/a_bohr**2/e_charge*1.e-21_dp
      f1 = SQRT(15._dp/fourpi)
      f2 = SQRT(5._dp/fourpi)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      rho=rho, qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      rho_atom_set=rho_atom_set, pw_env=pw_env, &
                      particle_set=particle_set, para_env=para_env, &
                      input=input)

      dft_section => section_vals_get_subs_vals(input, "DFT")

      efg_interpolation = section_get_lval(section_vals=dft_section, &
                                           keyword_name="PRINT%ELECTRIC_FIELD_GRADIENT%INTERPOLATION")
      efg_debug = section_get_lval(section_vals=dft_section, &
                                   keyword_name="PRINT%ELECTRIC_FIELD_GRADIENT%DEBUG")
      CALL section_vals_val_get(dft_section, &
                                "PRINT%ELECTRIC_FIELD_GRADIENT%GSPACE_SMOOTHING", &
                                r_vals=rvals)
      ecut = rvals(1)
      sigma = rvals(2)
      IF (ecut == 0._dp .AND. sigma <= 0._dp) THEN
         smoothing = .FALSE.
         ecut = 1.e10_dp ! not used, just to have vars defined
         sigma = 1._dp ! not used, just to have vars defined
      ELSEIF (ecut == -1._dp .AND. sigma == -1._dp) THEN
         smoothing = .TRUE.
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, dvr2rs%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         ecut = 2._dp*dvr2rs%pw%pw_grid%cutoff*0.875_dp
         sigma = 2._dp*dvr2rs%pw%pw_grid%cutoff*0.125_dp
         CALL pw_pool_give_back_pw(auxbas_pw_pool, dvr2rs%pw)
      ELSE
         smoothing = .TRUE.
      END IF
      IF(.NOT.(ecut > 0._dp))CALL cp__a("qs_electric_field_gradient.F",170)
      IF(.NOT.(sigma > 0._dp))CALL cp__a("qs_electric_field_gradient.F",171)

      unit_nr = cp_print_key_unit_nr(logger, dft_section, "PRINT%ELECTRIC_FIELD_GRADIENT", &
                                     extension=".efg", log_filename=.FALSE.)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, "(/,A,/)") " ELECTRIC FIELD GRADIENTS [10**21 V/m^2]"
         IF (efg_interpolation) THEN
            WRITE (unit_nr, "(T16,A)") &
               " Smooth potential contribution calculated by spline interpolation"
         ELSE
            WRITE (unit_nr, "(T12,A)") &
               " Smooth potential contribution calculated by plane wave interpolation"
         END IF
         IF (smoothing) THEN
            WRITE (unit_nr, "(T36,A)") &
               " G-Space potential smoothed by Fermi function"
            WRITE (unit_nr, "(T36,A,T71,F10.4)") " Cutoff (eV) ", ecut
            WRITE (unit_nr, "(T36,A,T71,F10.4)") " Width (eV) ", sigma
         END IF
         WRITE (unit_nr, *)
      END IF

      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins

      natom = SIZE(particle_set, 1)
      ALLOCATE (efg_tensor(3, 3, natom))
      efg_tensor = 0._dp
      IF (efg_debug) THEN
         ALLOCATE (efg_pw(3, 3, natom))
         efg_pw = 0._dp
      END IF
      ALLOCATE (efg_diagval(3, natom))
      efg_diagval = 0._dp

      ALLOCATE (vh0(1:natom, -2:2))
      vh0 = 0._dp

      ALLOCATE (dvr2(6))
      ALLOCATE (dvspl(6))

      !prepare calculation
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)
      IF (gapw) CALL prepare_gapw_den(qs_env, do_rho0=.TRUE.)

      !calculate electrostatic potential
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho)

      CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, ehartree, &
                            v_hartree_gspace%pw)

      ! smoothing of potential
      IF (smoothing) CALL pw_smoothing(v_hartree_gspace%pw, ecut, sigma)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tot_gspace%pw)

      DO i = 1, 3
         DO j = 1, i
            ij = (i*(i-1))/2+j
            CALL pw_pool_create_pw(auxbas_pw_pool, dvr2(ij)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL pw_dr2(v_hartree_gspace%pw, dvr2(ij)%pw, i, j)
         END DO
      END DO

      IF (.NOT. efg_interpolation) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, structure_factor%pw, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      ELSE

         interp_section => section_vals_get_subs_vals(dft_section, &
                                                      "PRINT%ELECTRIC_FIELD_GRADIENT%INTERPOLATOR")
         CALL section_vals_val_get(interp_section, "aint_precond", &
                                   i_val=aint_precond)
         CALL section_vals_val_get(interp_section, "precond", i_val=precond_kind)
         CALL section_vals_val_get(interp_section, "max_iter", i_val=max_iter)
         CALL section_vals_val_get(interp_section, "eps_r", r_val=eps_r)
         CALL section_vals_val_get(interp_section, "eps_x", r_val=eps_x)

         CALL pw_pool_create_pw(auxbas_pw_pool, dvr2rs%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         DO i = 1, 6
            CALL pw_pool_create_pw(auxbas_pw_pool, dvspl(i)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_transfer(dvr2(i)%pw, dvr2rs%pw)
            ! calculate spline coefficients
            CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                          pool=auxbas_pw_pool, pbc=.TRUE., transpose=.FALSE.)
            CALL pw_spline_do_precond(precond, dvr2rs%pw, dvspl(i)%pw)
            CALL pw_spline_precond_set_kind(precond, precond_kind)
            success = find_coeffs(values=dvr2rs%pw, coeffs=dvspl(i)%pw, &
                                  linOp=spl3_pbc, preconditioner=precond, pool=auxbas_pw_pool, &
                                  eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
            IF(.NOT.(success))CALL cp__a("qs_electric_field_gradient.F",270)
            CALL pw_spline_precond_release(precond)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dvr2(i)%pw)
         END DO
         CALL pw_pool_give_back_pw(auxbas_pw_pool, dvr2rs%pw)
      END IF

      nkind = SIZE(qs_kind_set)

      DO ikind = 1, nkind
         NULLIFY (atom_list)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natomkind)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom)
         DO iat = 1, natomkind
            iatom = atom_list(iat)
            ra = particle_set(iatom)%r
            IF (efg_interpolation) THEN
               DO i = 1, 3
                  DO j = 1, i
                     ij = (i*(i-1))/2+j
                     efg_val = Eval_Interp_Spl3_pbc(ra, dvspl(ij)%pw)
                     efg_tensor(i, j, iatom) = -efg_val
                     efg_tensor(j, i, iatom) = efg_tensor(i, j, iatom)
                     IF (efg_debug) THEN
                        chk_spl = chk_spl+efg_val+ &
                                  SUM(Eval_d_Interp_Spl3_pbc(ra, dvspl(ij)%pw))
                     END IF
                  END DO
               END DO
            ELSE
               CALL pw_structure_factor(structure_factor%pw, ra)
               DO i = 1, 3
                  DO j = 1, i
                     ij = (i*(i-1))/2+j
                     efg_tensor(i, j, iatom) = -pw_integral_ab(dvr2(ij)%pw, structure_factor%pw)
                     efg_tensor(j, i, iatom) = efg_tensor(i, j, iatom)
                  END DO
               END DO
               efg_tensor(:, :, iatom) = efg_tensor(:, :, iatom)/structure_factor%pw%pw_grid%vol
            END IF
            IF (efg_debug) THEN
               efg_pw(:, :, iatom) = efg_tensor(:, :, iatom)
            END IF
         END DO

         IF (paw_atom) THEN
            CALL vlimit_atom(para_env, vh0, rho_atom_set, qs_kind_set(ikind), &
                             atom_list, natomkind, nspins)
            DO iat = 1, natomkind
               iatom = atom_list(iat)
               efg_tensor(1, 1, iatom) = efg_tensor(1, 1, iatom) &
                                         +f1*(vh0(iatom, 2))-f2*(vh0(iatom, 0))
               efg_tensor(2, 2, iatom) = efg_tensor(2, 2, iatom) &
                                         -f1*(vh0(iatom, 2))-f2*(vh0(iatom, 0))
               efg_tensor(3, 3, iatom) = efg_tensor(3, 3, iatom)+2._dp*f2*(vh0(iatom, 0))
               efg_tensor(1, 2, iatom) = efg_tensor(1, 2, iatom)+f1*(vh0(iatom, -2))
               efg_tensor(2, 1, iatom) = efg_tensor(2, 1, iatom)+f1*(vh0(iatom, -2))
               efg_tensor(1, 3, iatom) = efg_tensor(1, 3, iatom)+f1*(vh0(iatom, 1))
               efg_tensor(3, 1, iatom) = efg_tensor(3, 1, iatom)+f1*(vh0(iatom, 1))
               efg_tensor(2, 3, iatom) = efg_tensor(2, 3, iatom)+f1*(vh0(iatom, -1))
               efg_tensor(3, 2, iatom) = efg_tensor(3, 2, iatom)+f1*(vh0(iatom, -1))
            END DO
         END IF

         DO iat = 1, natomkind
            iatom = atom_list(iat)
            CALL diagonalise(efg_tensor(:, :, iatom), 3, "UPPER", &
                             eigenvalues, eigenvectors)
            CALL efgsort(eigenvalues, efg_diagval(:, iatom))
         END DO
      END DO ! ikind

      efg_tensor(:, :, :) = efg_tensor(:, :, :)*efg_units
      efg_diagval(:, :) = efg_diagval(:, :)*efg_units

      IF (efg_debug) THEN
         efg_pw(:, :, :) = efg_pw(:, :, :)*efg_units
         DO iatom = 1, natom
            IF (unit_nr > 0) THEN
               CALL get_atomic_kind(particle_set(iatom)%atomic_kind, &
                                    element_symbol=element_symbol)
               WRITE (UNIT=unit_nr, FMT="(T2,I5,T8,A,T12,A,T15,6F11.5)") &
                  iatom, element_symbol, "PW", ((efg_pw(i, j, iatom), i=1, j), j=1, 3)
               WRITE (UNIT=unit_nr, FMT="(T12,A,T15,6F11.5)") &
                  "AT", ((efg_tensor(i, j, iatom)-efg_pw(i, j, iatom), i=1, j), j=1, 3)
            END IF
         ENDDO
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT=*)
         END IF
         IF (efg_interpolation) THEN
            IF (unit_nr > 0) THEN
               WRITE (UNIT=unit_nr, FMT="(T2,A,E24.16)") "CheckSum splines =", &
                  chk_spl
               WRITE (UNIT=unit_nr, FMT=*)
            END IF
         END IF
      END IF

      DO iatom = 1, natom
         IF (unit_nr > 0) THEN
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol)
            WRITE (UNIT=unit_nr, FMT="(T2,I5,T8,A,T12,A,3(T39,3F14.7,/))") &
               iatom, element_symbol, "EFG Tensor", (efg_tensor(i, :, iatom), i=1, 3)
            WRITE (UNIT=unit_nr, FMT="(T12,A,T39,3F14.7)") &
               "EFG Tensor eigenvalues", efg_diagval(:, iatom)
            WRITE (UNIT=unit_nr, FMT="(T12,A,T67,F14.7)") "EFG Tensor anisotropy", &
               (efg_diagval(1, iatom)-efg_diagval(2, iatom))/efg_diagval(3, iatom)
            WRITE (UNIT=unit_nr, FMT=*)
         END IF
      ENDDO

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)
      IF (.NOT. efg_interpolation) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, structure_factor%pw)
         DO i = 1, 6
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dvr2(i)%pw)
         END DO
      ELSE
         DO i = 1, 6
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dvspl(i)%pw)
         END DO
      END IF

      DEALLOCATE (efg_tensor)
      IF (efg_debug) THEN
         DEALLOCATE (efg_pw)
      END IF

      DEALLOCATE (vh0)

      DEALLOCATE (dvr2)
      DEALLOCATE (dvspl)

      CALL timestop(handle)

   END SUBROUTINE qs_efg_calc

! **************************************************************************************************
!> \brief ...
!> \param para_env ...
!> \param vlimit ...
!> \param rho_atom_set ...
!> \param qs_kind ...
!> \param atom_list ...
!> \param natom ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE vlimit_atom(para_env, vlimit, rho_atom_set, qs_kind, &
                          atom_list, natom, nspins)

      ! calculate : Limit(r->0) V_hartree(r)/r^2

      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(dp), DIMENSION(:, -2:), INTENT(inout)         :: vlimit
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      INTEGER, DIMENSION(:), INTENT(IN)                  :: atom_list
      INTEGER, INTENT(IN)                                :: natom, nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'vlimit_atom', routineP = moduleN//':'//routineN

      INTEGER :: i, i1, i2, iat, iatom, icg, ipgf1, ipgf2, iset1, iset2, iso, iso1, iso1_first, &
         iso1_last, iso2, iso2_first, iso2_last, l, l_iso, llmax, m1s, m2s, m_iso, max_iso_not0, &
         max_iso_not0_local, max_s_harm, maxl, maxso, mepos, n1s, n2s, nset, num_pe, size1, size2
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, o2nindex
      REAL(dp)                                           :: zet12
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: CPC_sphere
      REAL(dp), DIMENSION(20)                            :: vgg
      REAL(dp), DIMENSION(:, :), POINTER                 :: coeff, zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj

      NULLIFY (orb_basis)
      NULLIFY (harmonics)
      NULLIFY (lmin, lmax, npgf, zet, my_CG, coeff)

      CALL get_qs_kind(qs_kind=qs_kind, basis_set=orb_basis, &
                       paw_proj_set=paw_proj, harmonics=harmonics)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                             maxl=maxl, npgf=npgf, nset=nset, zet=zet, &
                             maxso=maxso)

      CALL get_paw_proj_set(paw_proj_set=paw_proj, o2nindex=o2nindex)

      max_iso_not0 = harmonics%max_iso_not0
      max_s_harm = harmonics%max_s_harm
      llmax = harmonics%llmax

      ! Distribute the atoms of this kind
      num_pe = para_env%num_pe
      mepos = para_env%mepos
      bo = get_limit(natom, num_pe, mepos)

      my_CG => harmonics%my_CG

      ALLOCATE (CPC_sphere(nsoset(maxl), nsoset(maxl)))
      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm))

      m1s = 0
      DO iset1 = 1, nset
         m2s = 0
         DO iset2 = 1, nset
            CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, llmax, cg_list, cg_n_list, max_iso_not0_local)
            IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_electric_field_gradient.F",482)

            n1s = nsoset(lmax(iset1))
            DO ipgf1 = 1, npgf(iset1)
               iso1_first = nsoset(lmin(iset1)-1)+1+n1s*(ipgf1-1)+m1s
               iso1_last = nsoset(lmax(iset1))+n1s*(ipgf1-1)+m1s
               size1 = iso1_last-iso1_first+1
               iso1_first = o2nindex(iso1_first)
               iso1_last = o2nindex(iso1_last)
               i1 = iso1_last-iso1_first+1
               IF(.NOT.(size1 == i1))CALL cp__a("qs_electric_field_gradient.F",492)
               i1 = nsoset(lmin(iset1)-1)+1

               n2s = nsoset(lmax(iset2))
               DO ipgf2 = 1, npgf(iset2)
                  iso2_first = nsoset(lmin(iset2)-1)+1+n2s*(ipgf2-1)+m2s
                  iso2_last = nsoset(lmax(iset2))+n2s*(ipgf2-1)+m2s
                  size2 = iso2_last-iso2_first+1
                  iso2_first = o2nindex(iso2_first)
                  iso2_last = o2nindex(iso2_last)
                  i2 = iso2_last-iso2_first+1
                  IF(.NOT.(size2 == i2))CALL cp__a("qs_electric_field_gradient.F",503)
                  i2 = nsoset(lmin(iset2)-1)+1

                  zet12 = zet(ipgf1, iset1)+zet(ipgf2, iset2)

                  vgg = 0.0_dp

                  DO iso = 1, max_iso_not0_local
                     l_iso = indso(1, iso)
                     IF (l_iso /= 2) CYCLE
                     DO icg = 1, cg_n_list(iso)
                        iso1 = cg_list(1, icg, iso)
                        iso2 = cg_list(2, icg, iso)
                        l = indso(1, iso1)+indso(1, iso2)
                        IF (MOD(l, 2) == 0 .AND. l > 0) THEN
                           vgg(l/2) = fourpi/10._dp*fac(l-2)/zet12**(l/2)
                        END IF
                     END DO
                  END DO

                  DO iat = bo(1), bo(2)
                     iatom = atom_list(iat)

                     CPC_sphere = 0.0_dp
                     DO i = 1, nspins
                        coeff => rho_atom_set(iatom)%cpc_h(i)%r_coef
                        CPC_sphere(i1:i1+size1-1, i2:i2+size2-1) = &
                           CPC_sphere(i1:i1+size1-1, i2:i2+size2-1)+ &
                           coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                        coeff => rho_atom_set(iatom)%cpc_s(i)%r_coef
                        CPC_sphere(i1:i1+size1-1, i2:i2+size2-1) = &
                           CPC_sphere(i1:i1+size1-1, i2:i2+size2-1)- &
                           coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                     ENDDO ! i

                     DO iso = 1, max_iso_not0_local
                        l_iso = indso(1, iso)
                        m_iso = indso(1, iso)
                        IF (l_iso /= 2) CYCLE
                        DO icg = 1, cg_n_list(iso)
                           iso1 = cg_list(1, icg, iso)
                           iso2 = cg_list(2, icg, iso)
                           l = indso(1, iso1)+indso(1, iso2)
                           IF (MOD(l, 2) == 0 .AND. l > 0) THEN
                              vlimit(iatom, m_iso) = vlimit(iatom, m_iso)+ &
                                                     vgg(l/2)*CPC_sphere(iso1, iso2)*my_CG(iso1, iso2, iso)
                           END IF
                        ENDDO ! icg
                     ENDDO ! iso

                  ENDDO ! iat

               ENDDO ! ipgf2
            ENDDO ! ipgf1
            m2s = m2s+maxso
         ENDDO ! iset2
         m1s = m1s+maxso
      ENDDO ! iset1

      CALL mp_sum(vlimit, para_env%group)

      DEALLOCATE (CPC_sphere)
      DEALLOCATE (cg_list, cg_n_list)

   END SUBROUTINE vlimit_atom

! **************************************************************************************************
!> \brief ...
!> \param ein ...
!> \param eout ...
! **************************************************************************************************
   SUBROUTINE efgsort(ein, eout)
      REAL(dp), DIMENSION(3), INTENT(in)                 :: ein
      REAL(dp), DIMENSION(3), INTENT(inout)              :: eout

      INTEGER                                            :: i
      INTEGER, DIMENSION(3)                              :: ind
      REAL(dp), DIMENSION(3)                             :: eab

      eab = ABS(ein)
      CALL sort(eab, 3, ind)
      DO i = 1, 3
         eout(i) = ein(ind(i))
      END DO

   END SUBROUTINE efgsort

END MODULE qs_electric_field_gradient

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the interaction radii for the operator matrix calculation.
!> \par History
!>      Joost VandeVondele : added exp_radius_very_extended
!>      24.09.2002 overloaded init_interaction_radii for KG use (gt)
!>      27.02.2004 integrated KG into QS version of init_interaction_radii (jgh)
!>      07.03.2006 new routine to extend kind interaction radius for semi-empiric (jgh)
!> \author MK (12.07.2000)
! **************************************************************************************************
MODULE qs_interactions

   USE ao_util,                         ONLY: exp_radius
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type,&
                                              set_gto_basis_set
   USE cp_control_types,                ONLY: qs_control_type,&
                                              semi_empirical_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential,&
                                              gth_potential_type,&
                                              set_potential
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type,&
                                              set_paw_proj_set
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_interactions'

! *** Public subroutines ***

   PUBLIC :: init_interaction_radii, init_se_nlradius, init_interaction_radii_orb_basis

   PUBLIC :: write_paw_radii, write_ppnl_radii, write_ppl_radii, write_core_charge_radii, &
             write_pgf_orb_radii

CONTAINS

! **************************************************************************************************
!> \brief   Initialize all the atomic kind radii for a given threshold value.
!> \param qs_control ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \date    24.09.2002
!> \author  GT
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_interaction_radii(qs_control, atomic_kind_set, qs_kind_set)

      TYPE(qs_control_type), INTENT(IN)                  :: qs_control
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_interaction_radii', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iexp_ppl, ikind, ip, &
                                                            iprj_ppnl, j, l, lppl, lppnl, lprj, &
                                                            lprj_ppnl, maxl, maxnset, nexp_lpot, &
                                                            nexp_lsd, nexp_ppl, nkind
      INTEGER, DIMENSION(:), POINTER                     :: nct_lpot, nct_lsd, nprj, nprj_ppnl
      LOGICAL                                            :: llsd, lpot, paw_atom
      REAL(KIND=dp)                                      :: alpha_core_charge, alpha_ppl, &
                                                            ccore_charge, cerf_ppl, &
                                                            core_charge_radius, ppl_radius, &
                                                            ppnl_radius, rcprj, zeta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_lpot, alpha_lsd, alpha_ppnl, &
                                                            cexp_ppl
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cprj_ppnl, cval_lpot, cval_lsd, rzetprj, &
                                                            zet
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: aux_basis_set, aux_fit_basis_set, &
                                                            lri_basis, mao_basis, orb_basis_set, &
                                                            soft_basis
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set

      CALL timeset(routineN, handle)

      NULLIFY (all_potential, gth_potential, orb_basis_set, aux_basis_set, soft_basis, &
               lri_basis, mao_basis, paw_proj_set)
      NULLIFY (nprj_ppnl, nprj)
      NULLIFY (alpha_ppnl, cexp_ppl, cprj_ppnl, zet)

      nkind = SIZE(atomic_kind_set)
      CALL get_qs_kind_set(qs_kind_set, maxnset=maxnset)

      nexp_ppl = 0

      DO ikind = 1, nkind

         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, basis_type="ORB")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_basis_set, basis_type="AUX")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_fit_basis_set, basis_type="AUX_FIT")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis, basis_type="LRI")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=mao_basis, basis_type="MAO")
         CALL get_qs_kind(qs_kind_set(ikind), soft_basis_set=soft_basis)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          paw_proj_set=paw_proj_set, &
                          paw_atom=paw_atom, &
                          all_potential=all_potential, &
                          gth_potential=gth_potential)

         ! Calculate the orbital basis function radii ***
         ! For ALL electron this has to come before the calculation of the
         ! radii used to calculate the 3c lists.
         ! Indeed, there the radii of the GTO primitives are needed
         IF (ASSOCIATED(orb_basis_set)) THEN

            CALL init_interaction_radii_orb_basis(orb_basis_set, qs_control%eps_pgf_orb, &
                                                  qs_control%eps_kg_orb)

         END IF

         IF (ASSOCIATED(all_potential)) THEN

            CALL get_potential(potential=all_potential, &
                               alpha_core_charge=alpha_core_charge, &
                               ccore_charge=ccore_charge)

            ! Calculate the radius of the core charge distribution

            core_charge_radius = exp_radius(0, alpha_core_charge, &
                                            qs_control%eps_core_charge, &
                                            ccore_charge)

            CALL set_potential(potential=all_potential, &
                               core_charge_radius=core_charge_radius)

         ELSE IF (ASSOCIATED(gth_potential)) THEN

            CALL get_potential(potential=gth_potential, &
                               alpha_core_charge=alpha_core_charge, &
                               ccore_charge=ccore_charge, &
                               alpha_ppl=alpha_ppl, &
                               cerf_ppl=cerf_ppl, &
                               nexp_ppl=nexp_ppl, &
                               cexp_ppl=cexp_ppl, &
                               lpot_present=lpot, &
                               lsd_present=llsd, &
                               lppnl=lppnl, &
                               alpha_ppnl=alpha_ppnl, &
                               nprj_ppnl=nprj_ppnl, &
                               cprj_ppnl=cprj_ppnl)

            ! Calculate the radius of the core charge distribution
            core_charge_radius = exp_radius(0, alpha_core_charge, &
                                            qs_control%eps_core_charge, &
                                            ccore_charge)

            ! Calculate the radii of the local part
            ! of the Goedecker pseudopotential (GTH)
            ppl_radius = exp_radius(0, alpha_ppl, qs_control%eps_ppl, cerf_ppl)

            DO iexp_ppl = 1, nexp_ppl
               lppl = 2*(iexp_ppl-1)
               ppl_radius = MAX(ppl_radius, &
                                exp_radius(lppl, alpha_ppl, &
                                           qs_control%eps_ppl, &
                                           cexp_ppl(iexp_ppl)))
            END DO
            IF (lpot) THEN
               ! extended local potential
               CALL get_potential(potential=gth_potential, &
                                  nexp_lpot=nexp_lpot, &
                                  alpha_lpot=alpha_lpot, &
                                  nct_lpot=nct_lpot, &
                                  cval_lpot=cval_lpot)
               DO j = 1, nexp_lpot
                  DO i = 1, nct_lpot(j)
                     lppl = 2*(i-1)
                     ppl_radius = MAX(ppl_radius, &
                                      exp_radius(lppl, alpha_lpot(j), qs_control%eps_ppl, cval_lpot(i, j)))
                  END DO
               END DO
            END IF
            IF (llsd) THEN
               ! lsd local potential
               CALL get_potential(potential=gth_potential, &
                                  nexp_lsd=nexp_lsd, &
                                  alpha_lsd=alpha_lsd, &
                                  nct_lsd=nct_lsd, &
                                  cval_lsd=cval_lsd)
               DO j = 1, nexp_lsd
                  DO i = 1, nct_lsd(j)
                     lppl = 2*(i-1)
                     ppl_radius = MAX(ppl_radius, &
                                      exp_radius(lppl, alpha_lsd(j), qs_control%eps_ppl, cval_lsd(i, j)))
                  END DO
               END DO
            END IF

            ! Calculate the radii of the non-local part
            ! of the Goedecker pseudopotential (GTH)

            ppnl_radius = 0.0_dp

            DO l = 0, lppnl
               DO iprj_ppnl = 1, nprj_ppnl(l)
                  lprj_ppnl = l+2*(iprj_ppnl-1)
                  ppnl_radius = MAX(ppnl_radius, &
                                    exp_radius(lprj_ppnl, alpha_ppnl(l), &
                                               qs_control%eps_pgf_orb, &
                                               cprj_ppnl(iprj_ppnl, l)))
               END DO
            END DO

            CALL set_potential(potential=gth_potential, &
                               core_charge_radius=core_charge_radius, &
                               ppl_radius=ppl_radius, &
                               ppnl_radius=ppnl_radius)

         END IF

         ! Calculate the aux fit orbital basis function radii

         IF (ASSOCIATED(aux_fit_basis_set)) THEN

            CALL init_interaction_radii_orb_basis(aux_fit_basis_set, qs_control%eps_pgf_orb, &
                                                  qs_control%eps_kg_orb)

         END IF

         ! Calculate the aux orbital basis function radii

         IF (ASSOCIATED(aux_basis_set)) THEN

            CALL init_interaction_radii_orb_basis(aux_basis_set, qs_control%eps_pgf_orb)

         END IF

         ! Calculate the MAO orbital basis function radii

         IF (ASSOCIATED(mao_basis)) THEN

            CALL init_interaction_radii_orb_basis(mao_basis, qs_control%eps_pgf_orb)

         END IF

         ! Calculate the soft orbital basis function radii

         IF (ASSOCIATED(soft_basis)) THEN

            IF (paw_atom) THEN
               CALL init_interaction_radii_orb_basis(soft_basis, qs_control%eps_pgf_orb)

            END IF
         END IF

         ! Calculate the lri basis function radii
         IF (ASSOCIATED(lri_basis)) THEN
            CALL init_interaction_radii_orb_basis(lri_basis, qs_control%eps_pgf_orb)
         END IF

         ! Calculate the paw projector basis function radii

         IF (ASSOCIATED(paw_proj_set)) THEN

            CALL get_paw_proj_set(paw_proj_set=paw_proj_set, &
                                  nprj=nprj, &
                                  maxl=maxl, &
                                  zetprj=zet, &
                                  rzetprj=rzetprj)
            rcprj = 0.0_dp
            rzetprj = 0.0_dp
            DO lprj = 0, maxl
               DO ip = 1, nprj(lprj)
                  zeta = zet(ip, lprj)
                  rzetprj(ip, lprj) = MAX(rzetprj(ip, lprj), &
                                          exp_radius(lprj, zeta, &
                                                     qs_control%eps_pgf_orb, 1.0_dp))
                  rcprj = MAX(rcprj, rzetprj(ip, lprj))
               ENDDO
            ENDDO
            CALL set_paw_proj_set(paw_proj_set=paw_proj_set, &
                                  rzetprj=rzetprj, &
                                  rcprj=rcprj)

         END IF
      END DO ! ikind

      CALL timestop(handle)

   END SUBROUTINE init_interaction_radii

! **************************************************************************************************
!> \brief ...
!> \param orb_basis_set ...
!> \param eps_pgf_orb ...
!> \param eps_pgf_short ...
! **************************************************************************************************
   SUBROUTINE init_interaction_radii_orb_basis(orb_basis_set, eps_pgf_orb, eps_pgf_short)

      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      REAL(KIND=dp), INTENT(IN)                          :: eps_pgf_orb
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: eps_pgf_short

      CHARACTER(len=*), PARAMETER :: routineN = 'init_interaction_radii_orb_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, ishell, l, nset
      INTEGER, DIMENSION(:), POINTER                     :: npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: lshell
      REAL(KIND=dp)                                      :: eps_short, gcca, kind_radius, &
                                                            short_radius, zeta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pgf_radius, zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc

      IF (ASSOCIATED(orb_basis_set)) THEN

         IF (PRESENT(eps_pgf_short)) THEN
            eps_short = eps_pgf_short
         ELSE
            eps_short = eps_pgf_orb
         END IF

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                nset=nset, &
                                nshell=nshell, &
                                npgf=npgf, &
                                l=lshell, &
                                zet=zet, &
                                gcc=gcc, &
                                pgf_radius=pgf_radius, &
                                set_radius=set_radius)

         kind_radius = 0.0_dp
         short_radius = 0.0_dp

         DO iset = 1, nset
            set_radius(iset) = 0.0_dp
            DO ipgf = 1, npgf(iset)
               pgf_radius(ipgf, iset) = 0.0_dp
               DO ishell = 1, nshell(iset)
                  l = lshell(ishell, iset)
                  gcca = gcc(ipgf, ishell, iset)
                  zeta = zet(ipgf, iset)
                  pgf_radius(ipgf, iset) = MAX(pgf_radius(ipgf, iset), &
                                               exp_radius(l, zeta, &
                                                          eps_pgf_orb, &
                                                          gcca))
                  short_radius = MAX(short_radius, &
                                     exp_radius(l, zeta, &
                                                eps_short, &
                                                gcca))
               END DO
               set_radius(iset) = MAX(set_radius(iset), pgf_radius(ipgf, iset))
            END DO
            kind_radius = MAX(kind_radius, set_radius(iset))
         END DO

         CALL set_gto_basis_set(gto_basis_set=orb_basis_set, &
                                pgf_radius=pgf_radius, &
                                set_radius=set_radius, &
                                kind_radius=kind_radius, &
                                short_kind_radius=short_radius)

      END IF

   END SUBROUTINE init_interaction_radii_orb_basis

! **************************************************************************************************
!> \brief ...
!> \param se_control ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE init_se_nlradius(se_control, atomic_kind_set, qs_kind_set, &
                               subsys_section)

      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'init_se_nlradius', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind
      REAL(KIND=dp)                                      :: kind_radius
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      NULLIFY (orb_basis_set, qs_kind)

      nkind = SIZE(qs_kind_set)

      DO ikind = 1, nkind

         qs_kind => qs_kind_set(ikind)

         CALL get_qs_kind(qs_kind=qs_kind, basis_set=orb_basis_set)

         IF (ASSOCIATED(orb_basis_set)) THEN

            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                   kind_radius=kind_radius)

            kind_radius = MAX(kind_radius, se_control%cutoff_exc)

            CALL set_gto_basis_set(gto_basis_set=orb_basis_set, &
                                   kind_radius=kind_radius)

         END IF

      END DO

      CALL write_kind_radii(atomic_kind_set, qs_kind_set, subsys_section)

   END SUBROUTINE init_se_nlradius

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE write_kind_radii(atomic_kind_set, qs_kind_set, subsys_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=10)                                  :: bas
      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, nkind, output_unit
      REAL(KIND=dp)                                      :: conv, kind_radius
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      NULLIFY (logger)
      logger => cp_get_default_logger()
      NULLIFY (orb_basis_set)
      bas = "ORBITAL   "

      nkind = SIZE(atomic_kind_set)
!   *** Print the kind radii ***
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/KIND_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T56,A,T63,A,T75,A)") &
            "RADII: "//TRIM(bas)//" BASIS in "//TRIM(unit_str), &
            "Kind", "Label", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
            IF (ASSOCIATED(orb_basis_set)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      kind_radius=kind_radius)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, kind_radius*conv
            ELSE
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T72,A9)") &
                  ikind, name, "no basis"
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/KIND_RADII")

   END SUBROUTINE write_kind_radii

! **************************************************************************************************
!> \brief  Write the radii of the core charge distributions to the output
!>         unit.
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \date    15.09.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_core_charge_radii(atomic_kind_set, qs_kind_set, subsys_section)
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, nkind, output_unit
      REAL(KIND=dp)                                      :: conv, core_charge_radius
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gth_potential_type), POINTER                  :: gth_potential

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/CORE_CHARGE_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         nkind = SIZE(atomic_kind_set)
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T56,A,T63,A,T75,A)") &
            "RADII: CORE CHARGE DISTRIBUTIONS in "// &
            TRIM(unit_str), "Kind", "Label", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             all_potential=all_potential, gth_potential=gth_potential)

            IF (ASSOCIATED(all_potential)) THEN
               CALL get_potential(potential=all_potential, &
                                  core_charge_radius=core_charge_radius)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, core_charge_radius*conv
            ELSE IF (ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, &
                                  core_charge_radius=core_charge_radius)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, core_charge_radius*conv
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/CORE_CHARGE_RADII")
   END SUBROUTINE write_core_charge_radii

! **************************************************************************************************
!> \brief   Write the orbital basis function radii to the output unit.
!> \param basis ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \date    05.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_pgf_orb_radii(basis, atomic_kind_set, qs_kind_set, subsys_section)

      CHARACTER(len=*)                                   :: basis
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_pgf_orb_radii', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: bas
      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, ipgf, iset, nkind, nset, &
                                                            output_unit
      INTEGER, DIMENSION(:), POINTER                     :: npgf
      REAL(KIND=dp)                                      :: conv, kind_radius, short_kind_radius
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pgf_radius
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: aux_basis_set, lri_basis_set, &
                                                            orb_basis_set

      NULLIFY (logger)
      logger => cp_get_default_logger()
      NULLIFY (aux_basis_set, orb_basis_set, lri_basis_set)
      bas = " "
      bas(1:3) = basis(1:3)
      CALL uppercase(bas)
      IF (bas(1:3) == "ORB") THEN
         bas = "ORBITAL   "
      ELSE IF (bas(1:3) == "AUX") THEN
         bas = "AUXILLIARY"
      ELSE IF (bas(1:3) == "LRI") THEN
         bas = "LOCAL RI"
      ELSE
         CALL cp__b("qs_interactions.F",592,"Undefined basis set type")
      END IF

      nkind = SIZE(qs_kind_set)

!   *** Print the kind radii ***
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/KIND_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T46,A,T53,A,T63,A,T71,A)") &
            "RADII: "//TRIM(bas)//" BASIS in "//TRIM(unit_str), &
            "Kind", "Label", "Radius", "OCE Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            IF (bas(1:3) == "ORB") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
               IF (ASSOCIATED(orb_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         kind_radius=kind_radius, &
                                         short_kind_radius=short_kind_radius)
               END IF
            ELSE IF (bas(1:3) == "AUX") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_basis_set, basis_type="AUX")
               IF (ASSOCIATED(aux_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=aux_basis_set, &
                                         kind_radius=kind_radius)
               END IF
            ELSE IF (bas(1:3) == "LOC") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
               IF (ASSOCIATED(lri_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=lri_basis_set, &
                                         kind_radius=kind_radius)
               END IF
            ELSE
               CALL cp__b("qs_interactions.F",628,"Undefined basis set type")
            END IF
            IF (ASSOCIATED(aux_basis_set) .OR. ASSOCIATED(orb_basis_set)) THEN
               WRITE (UNIT=output_unit, FMT="(T45,I5,T53,A5,T57,F12.6,T69,F12.6)") &
                  ikind, name, kind_radius*conv, short_kind_radius*conv
            ELSE
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T72,A9)") &
                  ikind, name, "no basis"
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/KIND_RADII")

!   *** Print the shell set radii ***
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/SET_RADII", extension=".Log")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T51,A,T57,A,T65,A,T75,A)") &
            "RADII: SHELL SETS OF "//TRIM(bas)//" BASIS in "// &
            TRIM(unit_str), "Kind", "Label", "Set", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            IF (bas(1:3) == "ORB") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
               IF (ASSOCIATED(orb_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         nset=nset, &
                                         set_radius=set_radius)
               END IF
            ELSE IF (bas(1:3) == "AUX") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_basis_set, basis_type="AUX")
               IF (ASSOCIATED(aux_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=aux_basis_set, &
                                         nset=nset, &
                                         set_radius=set_radius)
               END IF
            ELSE IF (bas(1:3) == "LOC") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
               IF (ASSOCIATED(lri_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=lri_basis_set, &
                                         nset=nset, &
                                         set_radius=set_radius)
               END IF
            ELSE
               CALL cp__b("qs_interactions.F",673,"Undefined basis set type")
            END IF
            IF (ASSOCIATED(aux_basis_set) .OR. ASSOCIATED(orb_basis_set)) THEN
               WRITE (UNIT=output_unit, FMT="(T50,I5,T57,A5,(T63,I5,T69,F12.6))") &
                  ikind, name, (iset, set_radius(iset)*conv, iset=1, nset)
            ELSE
               WRITE (UNIT=output_unit, FMT="(T50,I5,T58,A5,T73,A8)") &
                  ikind, name, "no basis"
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/SET_RADII")
!   *** Print the primitive Gaussian function radii ***
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/PGF_RADII", extension=".Log")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T51,A,T57,A,T65,A,T75,A)") &
            "RADII: PRIMITIVE GAUSSIANS OF "//TRIM(bas)//" BASIS in "// &
            TRIM(unit_str), "Kind", "Label", "Set", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            IF (bas(1:3) == "ORB") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
               IF (ASSOCIATED(orb_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         nset=nset, &
                                         npgf=npgf, &
                                         pgf_radius=pgf_radius)
               END IF
            ELSE IF (bas(1:3) == "AUX") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_basis_set, basis_type="AUX")
               IF (ASSOCIATED(aux_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=aux_basis_set, &
                                         nset=nset, &
                                         npgf=npgf, &
                                         pgf_radius=pgf_radius)
               END IF
            ELSE IF (bas(1:3) == "LOC") THEN
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
               IF (ASSOCIATED(lri_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=lri_basis_set, &
                                         nset=nset, &
                                         npgf=npgf, &
                                         pgf_radius=pgf_radius)
               END IF
            ELSE
               CALL cp__b("qs_interactions.F",720,"Undefined basis type")
            END IF
            IF (ASSOCIATED(aux_basis_set) .OR. ASSOCIATED(orb_basis_set) .OR. &
                ASSOCIATED(lri_basis_set)) THEN
               DO iset = 1, nset
                  WRITE (UNIT=output_unit, FMT="(T50,I5,T57,A5,T63,I5,(T69,F12.6))") &
                     ikind, name, iset, &
                     (pgf_radius(ipgf, iset)*conv, ipgf=1, npgf(iset))
               END DO
            ELSE
               WRITE (UNIT=output_unit, FMT="(T50,I5,T58,A5,T73,A8)") &
                  ikind, name, "no basis"
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/PGF_RADII")
   END SUBROUTINE write_pgf_orb_radii

! **************************************************************************************************
!> \brief   Write the radii of the exponential functions of the Goedecker
!>           pseudopotential (GTH, local part) to the logical unit number
!>          "output_unit".
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \date    06.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_ppl_radii(atomic_kind_set, qs_kind_set, subsys_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, nkind, output_unit
      REAL(KIND=dp)                                      :: conv, ppl_radius
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gth_potential_type), POINTER                  :: gth_potential

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/GTH_PPL_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         nkind = SIZE(atomic_kind_set)
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T56,A,T63,A,T75,A)") &
            "RADII: LOCAL PART OF GTH/ELP PP in "// &
            TRIM(unit_str), "Kind", "Label", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
            IF (ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, &
                                  ppl_radius=ppl_radius)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, ppl_radius*conv
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/GTH_PPL_RADII")
   END SUBROUTINE write_ppl_radii

! **************************************************************************************************
!> \brief  Write the radii of the projector functions of the Goedecker
!>          pseudopotential (GTH, non-local part) to the logical unit number
!>          "output_unit".
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \date    06.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_ppnl_radii(atomic_kind_set, qs_kind_set, subsys_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, nkind, output_unit
      REAL(KIND=dp)                                      :: conv, ppnl_radius
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gth_potential_type), POINTER                  :: gth_potential

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/GTH_PPNL_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         nkind = SIZE(atomic_kind_set)
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T56,A,T63,A,T75,A)") &
            "RADII: NON-LOCAL PART OF GTH PP in "// &
            TRIM(unit_str), "Kind", "Label", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
            IF (ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, &
                                  ppnl_radius=ppnl_radius)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, ppnl_radius*conv
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/GTH_PPNL_RADII")
   END SUBROUTINE write_ppnl_radii

! **************************************************************************************************
!> \brief   Write the radii of the one center projector
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_paw_radii(atomic_kind_set, qs_kind_set, subsys_section)
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=default_string_length)               :: name, unit_str
      INTEGER                                            :: ikind, nkind, output_unit
      LOGICAL                                            :: paw_atom
      REAL(KIND=dp)                                      :: conv, rcprj
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%RADII/GAPW_PRJ_RADII", extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%RADII%UNIT", c_val=unit_str)
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      IF (output_unit > 0) THEN
         nkind = SIZE(qs_kind_set)
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T56,A,T63,A,T75,A)") &
            "RADII: ONE CENTER PROJECTORS in "// &
            TRIM(unit_str), "Kind", "Label", "Radius"
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), name=name)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             paw_atom=paw_atom, paw_proj_set=paw_proj_set)
            IF (paw_atom .AND. ASSOCIATED(paw_proj_set)) THEN
               CALL get_paw_proj_set(paw_proj_set=paw_proj_set, &
                                     rcprj=rcprj)
               WRITE (UNIT=output_unit, FMT="(T55,I5,T63,A5,T69,F12.6)") &
                  ikind, name, rcprj*conv
            END IF
         END DO
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%RADII/GAPW_PRJ_RADII")
   END SUBROUTINE write_paw_radii

END MODULE qs_interactions
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate MAO's and analyze wavefunctions
!> \par History
!>      03.2016 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE mao_methods
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE basis_set_container_types,       ONLY: add_basis_set_to_container
   USE basis_set_types,                 ONLY: create_primitive_basis_set,&
                                              get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type,&
                                              write_gto_basis_set
   USE bibliography,                    ONLY: Ehrhardt1985,&
                                              Heinzmann1976,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_restore
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_desymmetrize, cp_dbcsr_get_block_diag, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_norm, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_replicate_all, cp_dbcsr_reserve_diag_blocks, cp_dbcsr_trace, &
        cp_dbcsr_type, dbcsr_distribution_obj, dbcsr_norm_frobenius, dbcsr_norm_maxabsnorm, &
        dbcsr_type_no_symmetry, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_plus_fm_fm_t
   USE cp_fm_diag,                      ONLY: cp_fm_geeig
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: mao_basis_ext,&
                                              mao_basis_orb,&
                                              mao_basis_prim
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE iterate_matrix,                  ONLY: invert_Hotelling
   USE kinds,                           ONLY: dp
   USE kpoint_methods,                  ONLY: rskp_transform
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE lapack,                          ONLY: lapack_ssyev,&
                                              lapack_ssygv
   USE mathlib,                         ONLY: invmat_symm
   USE message_passing,                 ONLY: mp_sum
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_interactions,                 ONLY: init_interaction_radii_orb_basis
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_neighbor_lists,               ONLY: setup_neighbor_list
   USE qs_overlap,                      ONLY: build_overlap_matrix_simple
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   TYPE block_type
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE  :: mat
   END TYPE block_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mao_methods'

   PUBLIC ::  mao_analysis

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param input_section ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE mao_analysis(qs_env, input_section, unit_nr)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: input_section
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_analysis', routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: element_symbol, esa, esb, esc
      INTEGER :: fall, handle, ia, iab, iabc, iatom, ib, ic, icol, ikind, im, irow, iset, ishell, &
         ispin, iter, jatom, l, lmax, lshell, m, ma, mao_basis, max_iter, me, na, nab, nabc, &
         natom, nb, nbas, nc, nimages, nkind, nocc, norb, nspin, ssize
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes, mao_blk, mao_blk_sizes, &
                                                            orb_blk, row_blk_sizes
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: analyze_ua, explicit, fo, for, fos, &
                                                            found, neglect_abc, print_basis
      REAL(KIND=dp) :: a1, a2, alpha, an, beta, deltaq, electra(2), elex, eps_ab, eps_abc, &
         eps_filter, eps_fun, eps_grad, eps_pgf_orb, fa1, fa2, fnnew, fnold, fval, grad_norm, &
         senabc, senmax, threshold, total_charge, total_spin, ua_charge(2), xkp(3), zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cmask, vec1, vec2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: occnumA, occnumABC, qab, qmatab, qmatac, &
                                                            qmatbc, raq, sab, selnABC, sinv, &
                                                            smatab, smatac, smatbc, uaq, weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: occnumAB, selnAB
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: block, cmao, diag, qblka, qblkb, qblkc, &
                                                            rblkl, rblku, sblk, sblka, sblkb, sblkc
      TYPE(block_type), ALLOCATABLE, DIMENSION(:)        :: rowblock
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_iterator)                            :: dbcsr_iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mao_coef, mao_dmat, mao_grad, mao_qmat, &
                                                            mao_smat, matrix_q, matrix_smm, &
                                                            matrix_smo
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks, matrix_p, matrix_s
      TYPE(cp_dbcsr_type)                                :: amat, axmat, cgmat, cholmat, crumat, &
                                                            ksmat, pmat, qmat, qmat_diag, rumat, &
                                                            smat, smat_diag, sumat, tmat
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: mao_basis_set_list, orb_basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set, pbasis
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all, sab_orb, smm_list, smo_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho

! only do MAO analysis if explicitely requested

      CALL section_vals_get(input_section, explicit=explicit)
      IF (.NOT. explicit) RETURN

      CALL timeset(routineN, handle)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(/,T2,A)') '!-----------------------------------------------------------------------------!'
         WRITE (UNIT=unit_nr, FMT="(T36,A)") "MAO ANALYSIS"
         WRITE (UNIT=unit_nr, FMT="(T12,A)") "Claus Ehrhardt and Reinhart Ahlrichs, TCA 68:231-245 (1985)"
         WRITE (unit_nr, '(T2,A)') '!-----------------------------------------------------------------------------!'
      END IF
      CALL cite_reference(Heinzmann1976)
      CALL cite_reference(Ehrhardt1985)

      ! input options
      CALL section_vals_val_get(input_section, "REFERENCE_BASIS", i_val=mao_basis)
      CALL section_vals_val_get(input_section, "EPS_FILTER", r_val=eps_filter)
      CALL section_vals_val_get(input_section, "EPS_FUNCTION", r_val=eps_fun)
      CALL section_vals_val_get(input_section, "EPS_GRAD", r_val=eps_grad)
      CALL section_vals_val_get(input_section, "MAX_ITER", i_val=max_iter)
      CALL section_vals_val_get(input_section, "PRINT_BASIS", l_val=print_basis)
      CALL section_vals_val_get(input_section, "NEGLECT_ABC", l_val=neglect_abc)
      CALL section_vals_val_get(input_section, "AB_THRESHOLD", r_val=eps_ab)
      CALL section_vals_val_get(input_section, "ABC_THRESHOLD", r_val=eps_abc)
      CALL section_vals_val_get(input_section, "ANALYZE_UNASSIGNED_CHARGE", l_val=analyze_ua)

      ! k-points?
      CALL get_qs_env(qs_env, dft_control=dft_control)
      nimages = dft_control%nimages
      IF (nimages > 1) THEN
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T2,A)") &
               "K-Points: MAO's determined and analyzed using Gamma-Point only."
         END IF
      END IF

      ! Reference basis set
      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set)
      nkind = SIZE(qs_kind_set)
      ALLOCATE (mao_basis_set_list(nkind), orb_basis_set_list(nkind))
      DO ikind = 1, nkind
         NULLIFY (mao_basis_set_list(ikind)%gto_basis_set)
         NULLIFY (orb_basis_set_list(ikind)%gto_basis_set)
      END DO
      !
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set, basis_type="ORB")
         IF (ASSOCIATED(basis_set)) orb_basis_set_list(ikind)%gto_basis_set => basis_set
      END DO
      !
      SELECT CASE (mao_basis)
      CASE (mao_basis_orb)
         DO ikind = 1, nkind
            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set, basis_type="ORB")
            IF (ASSOCIATED(basis_set)) mao_basis_set_list(ikind)%gto_basis_set => basis_set
         END DO
      CASE (mao_basis_prim)
         DO ikind = 1, nkind
            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set, basis_type="ORB")
            NULLIFY (pbasis)
            IF (ASSOCIATED(basis_set)) THEN
               CALL create_primitive_basis_set(basis_set, pbasis)
               CALL get_qs_env(qs_env, dft_control=dft_control)
               eps_pgf_orb = dft_control%qs_control%eps_pgf_orb
               CALL init_interaction_radii_orb_basis(pbasis, eps_pgf_orb)
               pbasis%kind_radius = basis_set%kind_radius
               mao_basis_set_list(ikind)%gto_basis_set => pbasis
               CALL add_basis_set_to_container(qs_kind%basis_sets, pbasis, "MAO")
            END IF
         END DO
      CASE (mao_basis_ext)
         DO ikind = 1, nkind
            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set, basis_type="MAO")
            IF (ASSOCIATED(basis_set)) THEN
               basis_set%kind_radius = orb_basis_set_list(ikind)%gto_basis_set%kind_radius
               mao_basis_set_list(ikind)%gto_basis_set => basis_set
            END IF
         END DO
      CASE DEFAULT
         CALL cp__b("mao_methods.F",242,"Unknown option for MAO basis")
      END SELECT
      IF (unit_nr > 0) THEN
         DO ikind = 1, nkind
            IF (.NOT. ASSOCIATED(mao_basis_set_list(ikind)%gto_basis_set)) THEN
               WRITE (UNIT=unit_nr, FMT="(T2,A,I4)") &
                  "WARNING: No MAO basis set associated with Kind ", ikind
            ELSE
               nbas = mao_basis_set_list(ikind)%gto_basis_set%nsgf
               WRITE (UNIT=unit_nr, FMT="(T2,A,I4,T56,A,I10)") &
                  "MAO basis set Kind ", ikind, " Number of BSF:", nbas
            END IF
         END DO
      END IF

      IF (print_basis) THEN
         DO ikind = 1, nkind
            basis_set => mao_basis_set_list(ikind)%gto_basis_set
            CALL write_gto_basis_set(basis_set, unit_nr, "MAO REFERENCE BASIS")
         END DO
      END IF

      ! neighbor lists
      NULLIFY (smm_list, smo_list)
      CALL setup_neighbor_list(smm_list, mao_basis_set_list, qs_env=qs_env)
      CALL setup_neighbor_list(smo_list, mao_basis_set_list, orb_basis_set_list, qs_env=qs_env)

      ! overlap matrices
      NULLIFY (matrix_smm, matrix_smo)
      CALL get_qs_env(qs_env, ks_env=ks_env)
      CALL build_overlap_matrix_simple(ks_env, matrix_smm, &
                                       mao_basis_set_list, mao_basis_set_list, smm_list)
      CALL build_overlap_matrix_simple(ks_env, matrix_smo, &
                                       mao_basis_set_list, orb_basis_set_list, smo_list)

      ! get reference density matrix and overlap matrix
      CALL get_qs_env(qs_env, rho=rho, matrix_s_kp=matrix_s)
      CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
      nspin = SIZE(matrix_p, 1)
      DO ispin = 1, nspin
         electra(ispin) = 0.0_dp
         DO im = 1, nimages
            CALL cp_dbcsr_trace(matrix_p(ispin, im)%matrix, matrix_s(1, im)%matrix, elex)
            electra(ispin) = electra(ispin)+elex
         END DO
      END DO

      ! calculate Q=APA(T) matrix, A=(MAO,ORB) overlap
      ! Q matrix
      NULLIFY (matrix_q)
      CALL cp_dbcsr_allocate_matrix_set(matrix_q, nspin)
      DO ispin = 1, nspin
         ALLOCATE (matrix_q(ispin)%matrix)
         CALL cp_dbcsr_init(matrix_q(ispin)%matrix)
         CALL cp_dbcsr_create(matrix_q(ispin)%matrix, template=matrix_smm(1)%matrix)
         CALL cp_dbcsr_alloc_block_from_nbl(matrix_q(ispin)%matrix, smm_list)
      END DO
      ! temp matrix
      CALL cp_dbcsr_init(tmat)
      CALL cp_dbcsr_create(tmat, template=matrix_smo(1)%matrix, matrix_type=dbcsr_type_no_symmetry)
      ! Q=APA(T)
      DO ispin = 1, nspin
         IF (nimages == 1) THEN
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_smo(1)%matrix, matrix_p(ispin, 1)%matrix, &
                                   0.0_dp, tmat, filter_eps=eps_filter)
            CALL cp_dbcsr_multiply("N", "T", 1.0_dp, tmat, matrix_smo(1)%matrix, &
                                   0.0_dp, matrix_q(ispin)%matrix, filter_eps=eps_filter)
         ELSE
            ! k-points
            CALL cp_dbcsr_init(pmat)
            CALL cp_dbcsr_create(pmat, template=matrix_s(1, 1)%matrix)
            CALL cp_dbcsr_init(smat)
            CALL cp_dbcsr_create(smat, template=matrix_s(1, 1)%matrix)
            CALL cp_dbcsr_init(ksmat)
            CALL cp_dbcsr_create(ksmat, template=matrix_s(1, 1)%matrix)
            CALL get_qs_env(qs_env=qs_env, sab_orb=sab_orb)
            CALL cp_dbcsr_alloc_block_from_nbl(pmat, sab_orb)
            CALL cp_dbcsr_alloc_block_from_nbl(smat, sab_orb)
            CALL cp_dbcsr_alloc_block_from_nbl(ksmat, sab_orb)
            CALL get_qs_env(qs_env, kpoints=kpoints)
            NULLIFY (cell_to_index)
            CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
            ! calculate density matrix at gamma point
            xkp = 0.0_dp
            ! transform KS and S matrices to the gamma point
            CALL get_qs_env(qs_env, matrix_ks_kp=matrix_ks)
            CALL rskp_transform(rmatrix=ksmat, rsmat=matrix_ks, ispin=ispin, &
                                xkp=xkp, cell_to_index=cell_to_index, sab_nl=sab_orb)
            CALL rskp_transform(rmatrix=smat, rsmat=matrix_s, ispin=1, &
                                xkp=xkp, cell_to_index=cell_to_index, sab_nl=sab_orb)
            norb = NINT(electra(ispin))
            nocc = MOD(2, nspin)+1
            CALL calculate_p_gamma(pmat, ksmat, smat, kpoints, norb, REAL(nocc, KIND=dp))
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_smo(1)%matrix, pmat, &
                                   0.0_dp, tmat, filter_eps=eps_filter)
            CALL cp_dbcsr_multiply("N", "T", 1.0_dp, tmat, matrix_smo(1)%matrix, &
                                   0.0_dp, matrix_q(ispin)%matrix, filter_eps=eps_filter)
            CALL cp_dbcsr_release(pmat)
            CALL cp_dbcsr_release(smat)
            CALL cp_dbcsr_release(ksmat)
         END IF
      END DO
      ! free temp matrix
      CALL cp_dbcsr_release(tmat)

      ! check for extended basis sets
      fall = 0
      CALL neighbor_list_iterator_create(nl_iterator, smm_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom)
         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         END IF
         CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, 1)%matrix, &
                                   row=irow, col=icol, block=block, found=found)
         IF (.NOT. found) fall = fall+1
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      CALL get_qs_env(qs_env=qs_env, para_env=para_env)
      CALL mp_sum(fall, para_env%group)
      IF (unit_nr > 0 .AND. fall > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(/,T2,A,/,T2,A,/)") &
            "Warning: Extended MAO basis used with original basis filtered density matrix", &
            "Warning: Possible errors can be controled with EPS_PGF_ORB"
      END IF

      ! MAO matrices
      CALL get_qs_env(qs_env, natom=natom)
      CALL get_ks_env(ks_env=ks_env, particle_set=particle_set, dbcsr_dist=dbcsr_dist)
      NULLIFY (mao_coef, mao_grad)
      CALL cp_dbcsr_allocate_matrix_set(mao_coef, nspin)
      CALL cp_dbcsr_allocate_matrix_set(mao_grad, nspin)
      ALLOCATE (row_blk_sizes(natom), col_blk_sizes(natom))
      CALL get_particle_set(particle_set, qs_kind_set, nsgf=row_blk_sizes, &
                            basis=mao_basis_set_list)
      CALL get_particle_set(particle_set, qs_kind_set, nmao=col_blk_sizes)
      ! check if MAOs have been specified
      DO iab = 1, natom
         IF (col_blk_sizes(iab) < 0) &
            CALL cp__b("mao_methods.F",386,"Number of MAOs has to be specified in KIND section for all elements")
      END DO
      DO ispin = 1, nspin
         ! coeficients
         ALLOCATE (mao_coef(ispin)%matrix)
         CALL cp_dbcsr_init(mao_coef(ispin)%matrix)
         CALL cp_dbcsr_create(matrix=mao_coef(ispin)%matrix, &
                              name="MAO_COEF", dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                              row_blk_size=row_blk_sizes, col_blk_size=col_blk_sizes, nze=0)
         CALL cp_dbcsr_reserve_diag_blocks(matrix=mao_coef(ispin)%matrix)
         ! gradients
         ALLOCATE (mao_grad(ispin)%matrix)
         CALL cp_dbcsr_init(mao_grad(ispin)%matrix)
         CALL cp_dbcsr_create(matrix=mao_grad(ispin)%matrix, &
                              name="MAO_GRAD", dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                              row_blk_size=row_blk_sizes, col_blk_size=col_blk_sizes, nze=0)
         CALL cp_dbcsr_reserve_diag_blocks(matrix=mao_grad(ispin)%matrix)
      END DO
      DEALLOCATE (row_blk_sizes, col_blk_sizes)

      ! initialize MAO coeficients from diagonal blocks of the Q matrix
      DO ispin = 1, nspin
         CALL mao_initialization(mao_coef(ispin)%matrix, &
                                 matrix_q(ispin)%matrix, matrix_smm(1)%matrix)
      END DO

      ! optimize MAOs
      alpha = 0.25_dp
      beta = 0.0_dp
      CALL cp_dbcsr_get_info(mao_coef(1)%matrix, col_blk_size=col_blk_sizes, distribution=dbcsr_dist)
      CALL cp_dbcsr_init(tmat)
      CALL cp_dbcsr_create(tmat, name="Binv", dist=dbcsr_dist, matrix_type=dbcsr_type_symmetric, &
                           row_blk_size=col_blk_sizes, col_blk_size=col_blk_sizes, nze=0)
      CALL cp_dbcsr_init(cgmat)
      CALL cp_dbcsr_create(cgmat, template=mao_grad(1)%matrix)
      CALL cp_dbcsr_init(amat)
      CALL cp_dbcsr_create(amat, template=mao_coef(1)%matrix)
      DO ispin = 1, nspin
         CALL mao_function_gradient(mao_coef(ispin)%matrix, fval, mao_grad(ispin)%matrix, &
                                    matrix_q(ispin)%matrix, matrix_smm(1)%matrix, tmat, .FALSE.)
         CALL cp_dbcsr_copy(cgmat, mao_grad(ispin)%matrix)
         CALL cp_dbcsr_norm(mao_grad(ispin)%matrix, dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
         CALL cp_dbcsr_norm(mao_grad(ispin)%matrix, dbcsr_norm_frobenius, norm_scalar=fnold)
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(/,T2,A,T73,A,I2)") "MAO OPTIMIZATION", "Spin =", ispin
            WRITE (UNIT=unit_nr, FMT="(T2,A,T24,A,F11.8,T48,A,F11.8,T69,A,F6.3)") &
               "Initialization", "fval =", fval/electra(ispin), "grad =", grad_norm, "step =", alpha
         END IF
         DO iter = 1, max_iter
            IF (grad_norm < eps_grad) EXIT
            IF ((1.0_dp-fval/electra(ispin)) < eps_fun) EXIT
            CALL cp_dbcsr_add(mao_coef(ispin)%matrix, cgmat, 1.0_dp, alpha)
            CALL mao_orthogonalization(mao_coef(ispin)%matrix, matrix_smm(1)%matrix)
            CALL mao_function_gradient(mao_coef(ispin)%matrix, fval, mao_grad(ispin)%matrix, &
                                       matrix_q(ispin)%matrix, matrix_smm(1)%matrix, tmat, .TRUE.)
            CALL cp_dbcsr_norm(mao_grad(ispin)%matrix, dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
            CALL cp_dbcsr_norm(mao_grad(ispin)%matrix, dbcsr_norm_frobenius, norm_scalar=fnnew)
            IF (unit_nr > 0) THEN
               WRITE (UNIT=unit_nr, FMT="(T2,A,i8,T24,A,F11.8,T48,A,F11.8,T69,A,F6.3)") &
                  "iter=", iter, "fval =", fval/electra(ispin), "grad =", grad_norm, "step =", alpha
            END IF
            beta = fnnew/fnold
            CALL cp_dbcsr_add(cgmat, mao_grad(ispin)%matrix, beta, 1.0_dp)
            fnold = fnnew
            ! line search, update alpha
            CALL cp_dbcsr_copy(amat, mao_coef(ispin)%matrix)
            CALL cp_dbcsr_add(amat, cgmat, 1.0_dp, 0.5_dp*alpha)
            CALL mao_orthogonalization(amat, matrix_smm(1)%matrix)
            CALL mao_function(amat, fa1, matrix_q(ispin)%matrix, matrix_smm(1)%matrix, tmat, .TRUE.)
            CALL cp_dbcsr_copy(amat, mao_coef(ispin)%matrix)
            CALL cp_dbcsr_add(amat, cgmat, 1.0_dp, alpha)
            CALL mao_orthogonalization(amat, matrix_smm(1)%matrix)
            CALL mao_function(amat, fa2, matrix_q(ispin)%matrix, matrix_smm(1)%matrix, tmat, .TRUE.)
            a2 = (4._dp*fa1-fa2-3._dp*fval)/alpha
            a1 = (fa2-fval-a2*alpha)/(alpha*alpha)
            an = -a2/(2._dp*a1)
            an = MIN(an, 2.0_dp*alpha)
            IF (an < 0.05_dp .OR. a1 > 0.0_dp) THEN
               CALL cp_dbcsr_copy(cgmat, mao_grad(ispin)%matrix)
               alpha = 0.25_dp
            ELSE
               alpha = an
            END IF
         END DO
      END DO
      CALL cp_dbcsr_release(tmat)
      CALL cp_dbcsr_release(cgmat)
      CALL cp_dbcsr_release(amat)

      ! Analyze the MAO basis
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, "(/,A)") " Analyze angular momentum character of MAOs "
         WRITE (unit_nr, "(T7,A,T15,A,T20,A,T40,A,T50,A,T60,A,T70,A,T80,A)") &
            "ATOM", "Spin", "MAO", "S", "P", "D", "F", "G"
      END IF
      lmax = 4 ! analyze up to g-functions
      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              element_symbol=element_symbol, kind_number=ikind)
         basis_set => mao_basis_set_list(ikind)%gto_basis_set
         CALL get_qs_kind(qs_kind_set(ikind), mao=na)
         CALL get_gto_basis_set(basis_set, nsgf=ma)
         ALLOCATE (cmask(ma), vec1(ma), vec2(ma), weight(0:lmax, na))
         weight = 0.0_dp
         CALL cp_dbcsr_get_block_p(matrix=matrix_smm(1)%matrix, row=iatom, col=iatom, &
                                   block=block, found=found)
         DO ispin = 1, nspin
            CALL cp_dbcsr_get_block_p(matrix=mao_coef(ispin)%matrix, row=iatom, col=iatom, &
                                      block=cmao, found=found)
            IF (found) THEN
               DO l = 0, lmax
                  cmask = 0.0_dp
                  iab = 0
                  DO iset = 1, basis_set%nset
                     DO ishell = 1, basis_set%nshell(iset)
                        lshell = basis_set%l(ishell, iset)
                        DO m = -lshell, lshell
                           iab = iab+1
                           IF (l == lshell) cmask(iab) = 1.0_dp
                        END DO
                     END DO
                  END DO
                  DO ia = 1, na
                     vec1(1:ma) = cmask*cmao(1:ma, ia)
                     vec2(1:ma) = MATMUL(block, vec1)
                     weight(l, ia) = SUM(vec1(1:ma)*vec2(1:ma))
                  END DO
               END DO
            END IF
            CALL mp_sum(weight, para_env%group)
            IF (unit_nr > 0) THEN
               DO ia = 1, na
                  IF (ispin == 1 .AND. ia == 1) THEN
                     WRITE (unit_nr, "(i6,T9,A2,T17,i2,T20,i3,T31,5F10.4)") &
                        iatom, element_symbol, ispin, ia, weight(0:lmax, ia)
                  ELSE
                     WRITE (unit_nr, "(T17,i2,T20,i3,T31,5F10.4)") ispin, ia, weight(0:lmax, ia)
                  END IF
               END DO
            END IF
         END DO
         DEALLOCATE (cmask, weight, vec1, vec2)
      END DO

      ! Calculate the overlap and density matrix in the new MAO basis
      NULLIFY (mao_dmat, mao_smat, mao_qmat)
      CALL cp_dbcsr_allocate_matrix_set(mao_qmat, nspin)
      CALL cp_dbcsr_allocate_matrix_set(mao_dmat, nspin)
      CALL cp_dbcsr_allocate_matrix_set(mao_smat, nspin)
      CALL cp_dbcsr_get_info(mao_coef(1)%matrix, col_blk_size=col_blk_sizes, distribution=dbcsr_dist)
      DO ispin = 1, nspin
         ALLOCATE (mao_dmat(ispin)%matrix)
         CALL cp_dbcsr_init(mao_dmat(ispin)%matrix)
         CALL cp_dbcsr_create(mao_dmat(ispin)%matrix, name="MAO density", dist=dbcsr_dist, &
                              matrix_type=dbcsr_type_symmetric, row_blk_size=col_blk_sizes, &
                              col_blk_size=col_blk_sizes, nze=0)
         ALLOCATE (mao_smat(ispin)%matrix)
         CALL cp_dbcsr_init(mao_smat(ispin)%matrix)
         CALL cp_dbcsr_create(mao_smat(ispin)%matrix, name="MAO overlap", dist=dbcsr_dist, &
                              matrix_type=dbcsr_type_symmetric, row_blk_size=col_blk_sizes, &
                              col_blk_size=col_blk_sizes, nze=0)
         ALLOCATE (mao_qmat(ispin)%matrix)
         CALL cp_dbcsr_init(mao_qmat(ispin)%matrix)
         CALL cp_dbcsr_create(mao_qmat(ispin)%matrix, name="MAO covar density", dist=dbcsr_dist, &
                              matrix_type=dbcsr_type_symmetric, row_blk_size=col_blk_sizes, &
                              col_blk_size=col_blk_sizes, nze=0)
      END DO
      CALL cp_dbcsr_init(amat)
      CALL cp_dbcsr_create(amat, name="MAO overlap", template=mao_dmat(1)%matrix)
      CALL cp_dbcsr_init(tmat)
      CALL cp_dbcsr_create(tmat, name="MAO Overlap Inverse", template=amat)
      CALL cp_dbcsr_init(qmat)
      CALL cp_dbcsr_create(qmat, name="MAO covar density", template=amat)
      CALL cp_dbcsr_init(cgmat)
      CALL cp_dbcsr_create(cgmat, name="TEMP matrix", template=mao_coef(1)%matrix)
      CALL cp_dbcsr_init(axmat)
      CALL cp_dbcsr_create(axmat, name="TEMP", template=amat, matrix_type=dbcsr_type_no_symmetry)
      DO ispin = 1, nspin
         ! calculate MAO overlap matrix
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_smm(1)%matrix, mao_coef(ispin)%matrix, &
                                0.0_dp, cgmat)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mao_coef(ispin)%matrix, cgmat, 0.0_dp, amat)
         ! calculate inverse of MAO overlap
         threshold = 1.e-8_dp
         CALL invert_Hotelling(tmat, amat, threshold, norm_convergence=1.e-4_dp, silent=.TRUE.)
         CALL cp_dbcsr_copy(mao_smat(ispin)%matrix, amat)
         ! calculate q-matrix q = C*Q*C
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_q(ispin)%matrix, mao_coef(ispin)%matrix, &
                                0.0_dp, cgmat, filter_eps=eps_filter)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mao_coef(ispin)%matrix, cgmat, &
                                0.0_dp, qmat, filter_eps=eps_filter)
         CALL cp_dbcsr_copy(mao_qmat(ispin)%matrix, qmat)
         ! calculate density matrix
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qmat, tmat, 0.0_dp, axmat, filter_eps=eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmat, axmat, 0.0_dp, mao_dmat(ispin)%matrix, &
                                filter_eps=eps_filter)
      END DO
      CALL cp_dbcsr_release(amat)
      CALL cp_dbcsr_release(tmat)
      CALL cp_dbcsr_release(qmat)
      CALL cp_dbcsr_release(cgmat)
      CALL cp_dbcsr_release(axmat)

      ! calculate unassigned charge : n - Tr PS
      DO ispin = 1, nspin
         CALL cp_dbcsr_trace(mao_dmat(ispin)%matrix, mao_smat(ispin)%matrix, ua_charge(ispin))
         ua_charge(ispin) = electra(ispin)-ua_charge(ispin)
      END DO
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *)
         DO ispin = 1, nspin
            WRITE (UNIT=unit_nr, FMT="(T2,A,T32,A,i2,T55,A,F12.8)") &
               "Unassigned charge", "Spin ", ispin, "delta charge =", ua_charge(ispin)
         END DO
      END IF

      ! occupation numbers: single atoms
      ! We use S_A = 1
      ! At the gamma point we use an effective MIC
      CALL get_qs_env(qs_env, natom=natom)
      ALLOCATE (occnumA(natom, nspin))
      occnumA = 0.0_dp
      DO ispin = 1, nspin
         DO iatom = 1, natom
            CALL cp_dbcsr_get_block_p(matrix=mao_qmat(ispin)%matrix, &
                                      row=iatom, col=iatom, block=block, found=found)
            IF (found) THEN
               DO iab = 1, SIZE(block, 1)
                  occnumA(iatom, ispin) = occnumA(iatom, ispin)+block(iab, iab)
               END DO
            END IF
         END DO
      END DO
      CALL mp_sum(occnumA, para_env%group)

      ! occupation numbers: atom pairs
      ALLOCATE (occnumAB(natom, natom, nspin))
      occnumAB = 0.0_dp
      DO ispin = 1, nspin
         CALL cp_dbcsr_init(qmat_diag)
         CALL cp_dbcsr_create(qmat_diag, name="MAO diagonal density", template=mao_dmat(1)%matrix)
         CALL cp_dbcsr_init(smat_diag)
         CALL cp_dbcsr_create(smat_diag, name="MAO diagonal overlap", template=mao_dmat(1)%matrix)
         ! replicate the diagonal blocks of the density and overlap matrices
         CALL cp_dbcsr_get_block_diag(mao_qmat(ispin)%matrix, qmat_diag)
         CALL cp_dbcsr_replicate_all(qmat_diag)
         CALL cp_dbcsr_get_block_diag(mao_smat(ispin)%matrix, smat_diag)
         CALL cp_dbcsr_replicate_all(smat_diag)
         DO ia = 1, natom
            DO ib = ia+1, natom
               iab = 0
               CALL cp_dbcsr_get_block_p(matrix=mao_qmat(ispin)%matrix, &
                                         row=ia, col=ib, block=block, found=found)
               IF (found) iab = 1
               CALL mp_sum(iab, para_env%group)
               IF(.NOT.(iab <= 1))CALL cp__a("mao_methods.F",641)
               IF (iab == 0 .AND. para_env%ionode) THEN
                  ! AB block is not available N_AB = N_A + N_B
                  ! Do this only on the "source" processor
                  occnumAB(ia, ib, ispin) = occnumA(ia, ispin)+occnumA(ib, ispin)
                  occnumAB(ib, ia, ispin) = occnumA(ia, ispin)+occnumA(ib, ispin)
               ELSE IF (found) THEN
                  ! owner of AB block performs calculation
                  na = SIZE(block, 1)
                  nb = SIZE(block, 2)
                  nab = na+nb
                  ALLOCATE (sab(nab, nab), qab(nab, nab), sinv(nab, nab))
                  ! qmat
                  qab(1:na, na+1:nab) = block(1:na, 1:nb)
                  qab(na+1:nab, 1:na) = TRANSPOSE(block(1:na, 1:nb))
                  CALL cp_dbcsr_get_block_p(matrix=qmat_diag, row=ia, col=ia, block=diag, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",657)
                  qab(1:na, 1:na) = diag(1:na, 1:na)
                  CALL cp_dbcsr_get_block_p(matrix=qmat_diag, row=ib, col=ib, block=diag, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",660)
                  qab(na+1:nab, na+1:nab) = diag(1:nb, 1:nb)
                  ! smat
                  CALL cp_dbcsr_get_block_p(matrix=mao_smat(ispin)%matrix, &
                                            row=ia, col=ib, block=block, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",665)
                  sab(1:na, na+1:nab) = block(1:na, 1:nb)
                  sab(na+1:nab, 1:na) = TRANSPOSE(block(1:na, 1:nb))
                  CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ia, col=ia, block=diag, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",669)
                  sab(1:na, 1:na) = diag(1:na, 1:na)
                  CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ib, col=ib, block=diag, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",672)
                  sab(na+1:nab, na+1:nab) = diag(1:nb, 1:nb)
                  ! inv smat
                  sinv(1:nab, 1:nab) = sab(1:nab, 1:nab)
                  CALL invmat_symm(sinv)
                  ! Tr(Q*Sinv)
                  occnumAB(ia, ib, ispin) = SUM(qab*sinv)
                  occnumAB(ib, ia, ispin) = occnumAB(ia, ib, ispin)
                  !
                  DEALLOCATE (sab, qab, sinv)
               END IF
            END DO
         END DO
         CALL cp_dbcsr_release(qmat_diag)
         CALL cp_dbcsr_release(smat_diag)
      END DO
      CALL mp_sum(occnumAB, para_env%group)

      ! calculate shared electron numbers (AB)
      ALLOCATE (selnAB(natom, natom, nspin))
      selnAB = 0.0_dp
      DO ispin = 1, nspin
         DO ia = 1, natom
            DO ib = ia+1, natom
               selnAB(ia, ib, ispin) = occnumA(ia, ispin)+occnumA(ib, ispin)-occnumAB(ia, ib, ispin)
               selnAB(ib, ia, ispin) = selnAB(ia, ib, ispin)
            END DO
         END DO
      END DO

      IF (.NOT. neglect_abc) THEN
         ! calculate N_ABC
         nabc = (natom*(natom-1)*(natom-2))/6
         ALLOCATE (occnumABC(nabc, nspin))
         occnumABC = -1.0_dp
         DO ispin = 1, nspin
            CALL cp_dbcsr_init(qmat_diag)
            CALL cp_dbcsr_create(qmat_diag, name="MAO diagonal density", template=mao_dmat(1)%matrix)
            CALL cp_dbcsr_init(smat_diag)
            CALL cp_dbcsr_create(smat_diag, name="MAO diagonal overlap", template=mao_dmat(1)%matrix)
            ! replicate the diagonal blocks of the density and overlap matrices
            CALL cp_dbcsr_get_block_diag(mao_qmat(ispin)%matrix, qmat_diag)
            CALL cp_dbcsr_replicate_all(qmat_diag)
            CALL cp_dbcsr_get_block_diag(mao_smat(ispin)%matrix, smat_diag)
            CALL cp_dbcsr_replicate_all(smat_diag)
            iabc = 0
            DO ia = 1, natom
               CALL cp_dbcsr_get_block_p(matrix=qmat_diag, row=ia, col=ia, block=qblka, found=fo)
               IF(.NOT.(fo))CALL cp__a("mao_methods.F",720)
               CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ia, col=ia, block=sblka, found=fo)
               IF(.NOT.(fo))CALL cp__a("mao_methods.F",722)
               na = SIZE(qblka, 1)
               DO ib = ia+1, natom
                  ! screen with SEN(AB)
                  IF (selnAB(ia, ib, ispin) < eps_abc) THEN
                     iabc = iabc+(natom-ib)
                     CYCLE
                  END IF
                  CALL cp_dbcsr_get_block_p(matrix=qmat_diag, row=ib, col=ib, block=qblkb, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",731)
                  CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ib, col=ib, block=sblkb, found=fo)
                  IF(.NOT.(fo))CALL cp__a("mao_methods.F",733)
                  nb = SIZE(qblkb, 1)
                  nab = na+nb
                  ALLOCATE (qmatab(na, nb), smatab(na, nb))
                  CALL cp_dbcsr_get_block_p(matrix=mao_qmat(ispin)%matrix, row=ia, col=ib, &
                                            block=block, found=found)
                  qmatab = 0.0_dp
                  IF (found) qmatab(1:na, 1:nb) = block(1:na, 1:nb)
                  CALL mp_sum(qmatab, para_env%group)
                  CALL cp_dbcsr_get_block_p(matrix=mao_smat(ispin)%matrix, row=ia, col=ib, &
                                            block=block, found=found)
                  smatab = 0.0_dp
                  IF (found) smatab(1:na, 1:nb) = block(1:na, 1:nb)
                  CALL mp_sum(smatab, para_env%group)
                  DO ic = ib+1, natom
                     ! screen with SEN(AB)
                     IF ((selnAB(ia, ic, ispin) < eps_abc) .OR. (selnAB(ib, ic, ispin) < eps_abc)) THEN
                        iabc = iabc+1
                        CYCLE
                     END IF
                     CALL cp_dbcsr_get_block_p(matrix=qmat_diag, row=ic, col=ic, block=qblkc, found=fo)
                     IF(.NOT.(fo))CALL cp__a("mao_methods.F",754)
                     CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ic, col=ic, block=sblkc, found=fo)
                     IF(.NOT.(fo))CALL cp__a("mao_methods.F",756)
                     nc = SIZE(qblkc, 1)
                     ALLOCATE (qmatac(na, nc), smatac(na, nc))
                     CALL cp_dbcsr_get_block_p(matrix=mao_qmat(ispin)%matrix, row=ia, col=ic, &
                                               block=block, found=found)
                     qmatac = 0.0_dp
                     IF (found) qmatac(1:na, 1:nc) = block(1:na, 1:nc)
                     CALL mp_sum(qmatac, para_env%group)
                     CALL cp_dbcsr_get_block_p(matrix=mao_smat(ispin)%matrix, row=ia, col=ic, &
                                               block=block, found=found)
                     smatac = 0.0_dp
                     IF (found) smatac(1:na, 1:nc) = block(1:na, 1:nc)
                     CALL mp_sum(smatac, para_env%group)
                     ALLOCATE (qmatbc(nb, nc), smatbc(nb, nc))
                     CALL cp_dbcsr_get_block_p(matrix=mao_qmat(ispin)%matrix, row=ib, col=ic, &
                                               block=block, found=found)
                     qmatbc = 0.0_dp
                     IF (found) qmatbc(1:nb, 1:nc) = block(1:nb, 1:nc)
                     CALL mp_sum(qmatbc, para_env%group)
                     CALL cp_dbcsr_get_block_p(matrix=mao_smat(ispin)%matrix, row=ib, col=ic, &
                                               block=block, found=found)
                     smatbc = 0.0_dp
                     IF (found) smatbc(1:nb, 1:nc) = block(1:nb, 1:nc)
                     CALL mp_sum(smatbc, para_env%group)
                     !
                     nabc = na+nb+nc
                     ALLOCATE (sab(nabc, nabc), sinv(nabc, nabc), qab(nabc, nabc))
                     !
                     qab(1:na, 1:na) = qblka(1:na, 1:na)
                     qab(na+1:nab, na+1:nab) = qblkb(1:nb, 1:nb)
                     qab(nab+1:nabc, nab+1:nabc) = qblkc(1:nc, 1:nc)
                     qab(1:na, na+1:nab) = qmatab(1:na, 1:nb)
                     qab(na+1:nab, 1:na) = TRANSPOSE(qmatab(1:na, 1:nb))
                     qab(1:na, nab+1:nabc) = qmatac(1:na, 1:nc)
                     qab(nab+1:nabc, 1:na) = TRANSPOSE(qmatac(1:na, 1:nc))
                     qab(na+1:nab, nab+1:nabc) = qmatbc(1:nb, 1:nc)
                     qab(nab+1:nabc, na+1:nab) = TRANSPOSE(qmatbc(1:nb, 1:nc))
                     !
                     sab(1:na, 1:na) = sblka(1:na, 1:na)
                     sab(na+1:nab, na+1:nab) = sblkb(1:nb, 1:nb)
                     sab(nab+1:nabc, nab+1:nabc) = sblkc(1:nc, 1:nc)
                     sab(1:na, na+1:nab) = smatab(1:na, 1:nb)
                     sab(na+1:nab, 1:na) = TRANSPOSE(smatab(1:na, 1:nb))
                     sab(1:na, nab+1:nabc) = smatac(1:na, 1:nc)
                     sab(nab+1:nabc, 1:na) = TRANSPOSE(smatac(1:na, 1:nc))
                     sab(na+1:nab, nab+1:nabc) = smatbc(1:nb, 1:nc)
                     sab(nab+1:nabc, na+1:nab) = TRANSPOSE(smatbc(1:nb, 1:nc))
                     ! inv smat
                     sinv(1:nabc, 1:nabc) = sab(1:nabc, 1:nabc)
                     CALL invmat_symm(sinv)
                     ! Tr(Q*Sinv)
                     iabc = iabc+1
                     me = MOD(iabc, para_env%num_pe)
                     IF (me == para_env%mepos) THEN
                        occnumABC(iabc, ispin) = SUM(qab*sinv)
                     ELSE
                        occnumABC(iabc, ispin) = 0.0_dp
                     END IF
                     !
                     DEALLOCATE (sab, sinv, qab)
                     DEALLOCATE (qmatac, smatac)
                     DEALLOCATE (qmatbc, smatbc)
                  END DO
                  DEALLOCATE (qmatab, smatab)
               END DO
            END DO
            CALL cp_dbcsr_release(qmat_diag)
            CALL cp_dbcsr_release(smat_diag)
         END DO
         CALL mp_sum(occnumABC, para_env%group)
      END IF

      IF (.NOT. neglect_abc) THEN
         ! calculate shared electron numbers (ABC)
         nabc = (natom*(natom-1)*(natom-2))/6
         ALLOCATE (selnABC(nabc, nspin))
         selnABC = 0.0_dp
         DO ispin = 1, nspin
            iabc = 0
            DO ia = 1, natom
               DO ib = ia+1, natom
                  DO ic = ib+1, natom
                     iabc = iabc+1
                     IF (occnumABC(iabc, ispin) >= 0.0_dp) THEN
                        selnABC(iabc, ispin) = occnumA(ia, ispin)+occnumA(ib, ispin)+occnumA(ic, ispin)- &
                                               occnumAB(ia, ib, ispin)-occnumAB(ia, ic, ispin)-occnumAB(ib, ic, ispin)+ &
                                               occnumABC(iabc, ispin)
                     END IF
                  END DO
               END DO
            END DO
         END DO
      END IF

      ! calculate atomic charge
      ALLOCATE (raq(natom, nspin))
      raq = 0.0_dp
      DO ispin = 1, nspin
         DO ia = 1, natom
            raq(ia, ispin) = occnumA(ia, ispin)
            DO ib = 1, natom
               raq(ia, ispin) = raq(ia, ispin)-0.5_dp*selnAB(ia, ib, ispin)
            END DO
         END DO
         IF (.NOT. neglect_abc) THEN
            iabc = 0
            DO ia = 1, natom
               DO ib = ia+1, natom
                  DO ic = ib+1, natom
                     iabc = iabc+1
                     raq(ia, ispin) = raq(ia, ispin)+selnABC(iabc, ispin)/3._dp
                     raq(ib, ispin) = raq(ib, ispin)+selnABC(iabc, ispin)/3._dp
                     raq(ic, ispin) = raq(ic, ispin)+selnABC(iabc, ispin)/3._dp
                  END DO
               END DO
            END DO
         END IF
      END DO

      ! calculate unassigned charge (from sum over atomic charges)
      DO ispin = 1, nspin
         deltaq = (electra(ispin)-SUM(raq(1:natom, ispin)))-ua_charge(ispin)
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T2,A,T32,A,i2,T55,A,F12.8)") &
               "Cutoff error on charge", "Spin ", ispin, "error charge =", deltaq
         END IF
      END DO

      ! analyze unassigned charge
      ALLOCATE (uaq(natom, nspin))
      uaq = 0.0_dp
      IF (analyze_ua) THEN
         CALL get_qs_env(qs_env=qs_env, para_env=para_env, blacs_env=blacs_env)
         CALL get_qs_env(qs_env=qs_env, sab_orb=sab_orb, sab_all=sab_all)
         CALL cp_dbcsr_get_info(mao_coef(1)%matrix, row_blk_size=mao_blk_sizes, &
                                col_blk_size=col_blk_sizes, distribution=dbcsr_dist)
         CALL cp_dbcsr_get_info(matrix_s(1, 1)%matrix, row_blk_size=row_blk_sizes)
         CALL cp_dbcsr_init(amat)
         CALL cp_dbcsr_create(amat, name="temp", template=matrix_s(1, 1)%matrix)
         CALL cp_dbcsr_init(tmat)
         CALL cp_dbcsr_create(tmat, name="temp", template=mao_coef(1)%matrix)
         ! replicate diagonal of smm matrix
         CALL cp_dbcsr_init(smat_diag)
         CALL cp_dbcsr_get_block_diag(matrix_smm(1)%matrix, smat_diag)
         CALL cp_dbcsr_replicate_all(smat_diag)

         ALLOCATE (orb_blk(natom), mao_blk(natom))
         DO ia = 1, natom
            orb_blk = row_blk_sizes
            mao_blk = row_blk_sizes
            mao_blk(ia) = col_blk_sizes(ia)
            CALL cp_dbcsr_init(sumat)
            CALL cp_dbcsr_create(sumat, name="Smat", dist=dbcsr_dist, matrix_type=dbcsr_type_symmetric, &
                                 row_blk_size=mao_blk, col_blk_size=mao_blk, nze=0)
            CALL cp_dbcsr_alloc_block_from_nbl(sumat, sab_orb)
            CALL cp_dbcsr_init(cholmat)
            CALL cp_dbcsr_create(cholmat, name="Cholesky matrix", dist=dbcsr_dist, &
                                 matrix_type=dbcsr_type_no_symmetry, row_blk_size=mao_blk, col_blk_size=mao_blk, nze=0)
            CALL cp_dbcsr_init(rumat)
            CALL cp_dbcsr_create(rumat, name="Rmat", dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                                 row_blk_size=orb_blk, col_blk_size=mao_blk, nze=0)
            CALL cp_dbcsr_alloc_block_from_nbl(rumat, sab_orb, .TRUE.)
            CALL cp_dbcsr_init(crumat)
            CALL cp_dbcsr_create(crumat, name="Rmat*Umat", dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                                 row_blk_size=orb_blk, col_blk_size=mao_blk, nze=0)
            ! replicate row and col of smo matrix
            ALLOCATE (rowblock(natom))
            DO ib = 1, natom
               na = mao_blk_sizes(ia)
               nb = row_blk_sizes(ib)
               ALLOCATE (rowblock(ib)%mat(na, nb))
               rowblock(ib)%mat = 0.0_dp
               CALL cp_dbcsr_get_block_p(matrix=matrix_smo(1)%matrix, row=ia, col=ib, &
                                         block=block, found=found)
               IF (found) rowblock(ib)%mat(1:na, 1:nb) = block(1:na, 1:nb)
               CALL mp_sum(rowblock(ib)%mat, para_env%group)
            END DO
            !
            DO ispin = 1, nspin
               CALL cp_dbcsr_copy(tmat, mao_coef(ispin)%matrix)
               CALL cp_dbcsr_replicate_all(tmat)
               CALL cp_dbcsr_iterator_start(dbcsr_iter, matrix_s(1, 1)%matrix)
               DO WHILE (cp_dbcsr_iterator_blocks_left(dbcsr_iter))
                  CALL cp_dbcsr_iterator_next_block(dbcsr_iter, iatom, jatom, block)
                  CALL cp_dbcsr_get_block_p(matrix=sumat, row=iatom, col=jatom, block=sblk, found=fos)
                  IF(.NOT.(fos))CALL cp__a("mao_methods.F",941)
                  CALL cp_dbcsr_get_block_p(matrix=rumat, row=iatom, col=jatom, block=rblku, found=for)
                  IF(.NOT.(for))CALL cp__a("mao_methods.F",943)
                  CALL cp_dbcsr_get_block_p(matrix=rumat, row=jatom, col=iatom, block=rblkl, found=for)
                  IF(.NOT.(for))CALL cp__a("mao_methods.F",945)
                  CALL cp_dbcsr_get_block_p(matrix=tmat, row=ia, col=ia, block=cmao, found=found)
                  IF(.NOT.(found))CALL cp__a("mao_methods.F",947)
                  IF (iatom /= ia .AND. jatom /= ia) THEN
                     ! copy original overlap matrix
                     sblk = block
                     rblku = block
                     rblkl = TRANSPOSE(block)
                  ELSE IF (iatom /= ia) THEN
                     rblkl = TRANSPOSE(block)
                     sblk = MATMUL(TRANSPOSE(rowblock(iatom)%mat), cmao)
                     rblku = sblk
                  ELSE IF (jatom /= ia) THEN
                     rblku = block
                     sblk = MATMUL(TRANSPOSE(cmao), rowblock(jatom)%mat)
                     rblkl = TRANSPOSE(sblk)
                  ELSE
                     CALL cp_dbcsr_get_block_p(matrix=smat_diag, row=ia, col=ia, block=block, found=found)
                     IF(.NOT.(found))CALL cp__a("mao_methods.F",963)
                     sblk = MATMUL(TRANSPOSE(cmao), MATMUL(block, cmao))
                     rblku = MATMUL(TRANSPOSE(rowblock(ia)%mat), cmao)
                  END IF
               END DO
               CALL cp_dbcsr_iterator_stop(dbcsr_iter)
               ! Cholesky decomposition of SUMAT = U'U
               CALL cp_dbcsr_desymmetrize(sumat, cholmat)
               CALL cp_dbcsr_cholesky_decompose(cholmat, para_env=para_env, blacs_env=blacs_env)
               ! T = R*inv(U)
               ssize = SUM(mao_blk)
               CALL cp_dbcsr_cholesky_restore(rumat, ssize, cholmat, crumat, op="SOLVE", pos="RIGHT", &
                                              transa="N", para_env=para_env, blacs_env=blacs_env)
               ! A = T*transpose(T)
               CALL cp_dbcsr_multiply("N", "T", 1.0_dp, crumat, crumat, 0.0_dp, amat, &
                                      filter_eps=eps_filter)
               ! Tr(P*A)
               CALL cp_dbcsr_trace(matrix_p(ispin, 1)%matrix, amat, uaq(ia, ispin))
               uaq(ia, ispin) = uaq(ia, ispin)-electra(ispin)
            END DO
            !
            CALL cp_dbcsr_release(sumat)
            CALL cp_dbcsr_release(cholmat)
            CALL cp_dbcsr_release(rumat)
            CALL cp_dbcsr_release(crumat)
            !
            DO ib = 1, natom
               DEALLOCATE (rowblock(ib)%mat)
            END DO
            DEALLOCATE (rowblock)
         END DO
         CALL cp_dbcsr_release(smat_diag)
         CALL cp_dbcsr_release(amat)
         CALL cp_dbcsr_release(tmat)
         DEALLOCATE (orb_blk, mao_blk)
      END IF
      !
      raq(1:natom, 1:nspin) = raq(1:natom, 1:nspin)-uaq(1:natom, 1:nspin)
      DO ispin = 1, nspin
         deltaq = electra(ispin)-SUM(raq(1:natom, ispin))
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T2,A,T32,A,i2,T55,A,F12.8)") &
               "Charge/Atom redistributed", "Spin ", ispin, "delta charge =", &
               (deltaq+ua_charge(ispin))/REAL(natom, KIND=dp)
         END IF
      END DO

      ! output charges
      IF (unit_nr > 0) THEN
         IF (nspin == 1) THEN
            WRITE (unit_nr, "(/,T2,A,T40,A,T75,A)") "MAO atomic charges ", "Atom", "Charge"
         ELSE
            WRITE (unit_nr, "(/,T2,A,T40,A,T55,A,T70,A)") "MAO atomic charges ", "Atom", "Charge", "Spin Charge"
         END IF
         DO ispin = 1, nspin
            deltaq = electra(ispin)-SUM(raq(1:natom, ispin))
            raq(:, ispin) = raq(:, ispin)+deltaq/REAL(natom, KIND=dp)
         END DO
         total_charge = 0.0_dp
         total_spin = 0.0_dp
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)
            IF (nspin == 1) THEN
               WRITE (unit_nr, "(T30,I6,T42,A2,T69,F12.6)") iatom, element_symbol, zeff-raq(iatom, 1)
               total_charge = total_charge+(zeff-raq(iatom, 1))
            ELSE
               WRITE (unit_nr, "(T30,I6,T42,A2,T48,F12.6,T69,F12.6)") iatom, element_symbol, &
                  zeff-raq(iatom, 1)-raq(iatom, 2), raq(iatom, 1)-raq(iatom, 2)
               total_charge = total_charge+(zeff-raq(iatom, 1)-raq(iatom, 2))
               total_spin = total_spin+(raq(iatom, 1)-raq(iatom, 2))
            END IF
         END DO
         IF (nspin == 1) THEN
            WRITE (unit_nr, "(T2,A,T69,F12.6)") "Total Charge", total_charge
         ELSE
            WRITE (unit_nr, "(T2,A,T49,F12.6,T69,F12.6)") "Total Charge", total_charge, total_spin
         END IF
      END IF

      IF (analyze_ua) THEN
         ! output unassigned charges
         IF (unit_nr > 0) THEN
            IF (nspin == 1) THEN
               WRITE (unit_nr, "(/,T2,A,T40,A,T75,A)") "MAO hypervalent charges ", "Atom", "Charge"
            ELSE
               WRITE (unit_nr, "(/,T2,A,T40,A,T55,A,T70,A)") "MAO hypervalent charges ", "Atom", &
                  "Charge", "Spin Charge"
            END IF
            total_charge = 0.0_dp
            total_spin = 0.0_dp
            DO iatom = 1, natom
               CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                    element_symbol=element_symbol)
               IF (nspin == 1) THEN
                  WRITE (unit_nr, "(T30,I6,T42,A2,T69,F12.6)") iatom, element_symbol, uaq(iatom, 1)
                  total_charge = total_charge+uaq(iatom, 1)
               ELSE
                  WRITE (unit_nr, "(T30,I6,T42,A2,T48,F12.6,T69,F12.6)") iatom, element_symbol, &
                     uaq(iatom, 1)+uaq(iatom, 2), uaq(iatom, 1)-uaq(iatom, 2)
                  total_charge = total_charge+uaq(iatom, 1)+uaq(iatom, 2)
                  total_spin = total_spin+uaq(iatom, 1)-uaq(iatom, 2)
               END IF
            END DO
            IF (nspin == 1) THEN
               WRITE (unit_nr, "(T2,A,T69,F12.6)") "Total Charge", total_charge
            ELSE
               WRITE (unit_nr, "(T2,A,T49,F12.6,T69,F12.6)") "Total Charge", total_charge, total_spin
            END IF
         END IF
      END IF

      ! output shared electron numbers AB
      IF (unit_nr > 0) THEN
         IF (nspin == 1) THEN
            WRITE (unit_nr, "(/,T2,A,T31,A,T40,A,T78,A)") "Shared electron numbers ", "Atom", "Atom", "SEN"
         ELSE
            WRITE (unit_nr, "(/,T2,A,T31,A,T40,A,T51,A,T63,A,T71,A)") "Shared electron numbers ", "Atom", "Atom", &
               "SEN(1)", "SEN(2)", "SEN(total)"
         END IF
         DO ia = 1, natom
            DO ib = ia+1, natom
               CALL get_atomic_kind(atomic_kind=particle_set(ia)%atomic_kind, element_symbol=esa)
               CALL get_atomic_kind(atomic_kind=particle_set(ib)%atomic_kind, element_symbol=esb)
               IF (nspin == 1) THEN
                  IF (selnAB(ia, ib, 1) > eps_ab) THEN
                     WRITE (unit_nr, "(T26,I6,' ',A2,T35,I6,' ',A2,T69,F12.6)") ia, esa, ib, esb, selnAB(ia, ib, 1)
                  END IF
               ELSE
                  IF ((selnAB(ia, ib, 1)+selnAB(ia, ib, 2)) > eps_ab) THEN
                     WRITE (unit_nr, "(T26,I6,' ',A2,T35,I6,' ',A2,T45,3F12.6)") ia, esa, ib, esb, &
                        selnAB(ia, ib, 1), selnAB(ia, ib, 2), (selnAB(ia, ib, 1)+selnAB(ia, ib, 2))
                  END IF
               END IF
            END DO
         END DO
      END IF

      IF (.NOT. neglect_abc) THEN
         ! output shared electron numbers ABC
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, "(/,T2,A,T40,A,T49,A,T58,A,T78,A)") "Shared electron numbers ABC", &
               "Atom", "Atom", "Atom", "SEN"
            senmax = 0.0_dp
            iabc = 0
            DO ia = 1, natom
               DO ib = ia+1, natom
                  DO ic = ib+1, natom
                     iabc = iabc+1
                     senabc = SUM(selnABC(iabc, :))
                     senmax = MAX(senmax, senabc)
                     IF (senabc > eps_abc) THEN
                        CALL get_atomic_kind(atomic_kind=particle_set(ia)%atomic_kind, element_symbol=esa)
                        CALL get_atomic_kind(atomic_kind=particle_set(ib)%atomic_kind, element_symbol=esb)
                        CALL get_atomic_kind(atomic_kind=particle_set(ic)%atomic_kind, element_symbol=esc)
                        WRITE (unit_nr, "(T35,I6,' ',A2,T44,I6,' ',A2,T53,I6,' ',A2,T69,F12.6)") &
                           ia, esa, ib, esb, ic, esc, senabc
                     END IF
                  END DO
               END DO
            END DO
            WRITE (unit_nr, "(T2,A,T69,F12.6)") "Maximum SEN value calculated", senmax
         END IF
      END IF

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(/,T2,A)') &
            '!---------------------------END OF MAO ANALYSIS-------------------------------!'
      END IF

      ! Deallocate temporary arrays
      DEALLOCATE (occnumA, occnumAB, selnAB, raq, uaq)
      IF (.NOT. neglect_abc) THEN
         DEALLOCATE (occnumABC, selnABC)
      END IF

      ! Deallocate the neighbor list structure
      IF (ASSOCIATED(smm_list)) THEN
         DO iab = 1, SIZE(smm_list)
            CALL deallocate_neighbor_list_set(smm_list(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (smm_list)
      END IF
      IF (ASSOCIATED(smo_list)) THEN
         DO iab = 1, SIZE(smo_list)
            CALL deallocate_neighbor_list_set(smo_list(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (smo_list)
      END IF

      DEALLOCATE (mao_basis_set_list, orb_basis_set_list)

      IF (ASSOCIATED(matrix_smm)) CALL cp_dbcsr_deallocate_matrix_set(matrix_smm)
      IF (ASSOCIATED(matrix_smo)) CALL cp_dbcsr_deallocate_matrix_set(matrix_smo)
      IF (ASSOCIATED(matrix_q)) CALL cp_dbcsr_deallocate_matrix_set(matrix_q)

      IF (ASSOCIATED(mao_coef)) CALL cp_dbcsr_deallocate_matrix_set(mao_coef)
      IF (ASSOCIATED(mao_grad)) CALL cp_dbcsr_deallocate_matrix_set(mao_grad)
      IF (ASSOCIATED(mao_dmat)) CALL cp_dbcsr_deallocate_matrix_set(mao_dmat)
      IF (ASSOCIATED(mao_smat)) CALL cp_dbcsr_deallocate_matrix_set(mao_smat)
      IF (ASSOCIATED(mao_qmat)) CALL cp_dbcsr_deallocate_matrix_set(mao_qmat)

      CALL timestop(handle)

   END SUBROUTINE mao_analysis

! **************************************************************************************************
!> \brief ...
!> \param mao_coef ...
!> \param pmat ...
!> \param smat ...
! **************************************************************************************************
   SUBROUTINE mao_initialization(mao_coef, pmat, smat)
      TYPE(cp_dbcsr_type)                                :: mao_coef, pmat, smat

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_initialization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, info, jatom, lwork, m, n
      LOGICAL                                            :: found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: w, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: amat, bmat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cblock, pblock, sblock
      TYPE(cp_dbcsr_iterator)                            :: dbcsr_iter

      CALL cp_dbcsr_iterator_start(dbcsr_iter, mao_coef)
      DO WHILE (cp_dbcsr_iterator_blocks_left(dbcsr_iter))
         CALL cp_dbcsr_iterator_next_block(dbcsr_iter, iatom, jatom, cblock)
         IF(.NOT.(iatom == jatom))CALL cp__a("mao_methods.F",1192)
         m = SIZE(cblock, 2)
         NULLIFY (pblock, sblock)
         CALL cp_dbcsr_get_block_p(matrix=pmat, row=iatom, col=jatom, block=pblock, found=found)
         IF(.NOT.(found))CALL cp__a("mao_methods.F",1196)
         CALL cp_dbcsr_get_block_p(matrix=smat, row=iatom, col=jatom, block=sblock, found=found)
         IF(.NOT.(found))CALL cp__a("mao_methods.F",1198)
         n = SIZE(sblock, 1)
         lwork = MAX(n*n, 100)
         ALLOCATE (amat(n, n), bmat(n, n), w(n), work(lwork))
         amat(1:n, 1:n) = pblock(1:n, 1:n)
         bmat(1:n, 1:n) = sblock(1:n, 1:n)
         info = 0
         CALL lapack_ssygv(1, "V", "U", n, amat, n, bmat, n, w, work, lwork, info)
         IF(.NOT.(info == 0))CALL cp__a("mao_methods.F",1206)
         cblock(1:n, 1:m) = amat(1:n, n:n-m+1:-1)
         DEALLOCATE (amat, bmat, w, work)
      END DO
      CALL cp_dbcsr_iterator_stop(dbcsr_iter)

   END SUBROUTINE mao_initialization

! **************************************************************************************************
!> \brief ...
!> \param mao_coef ...
!> \param fval ...
!> \param qmat ...
!> \param smat ...
!> \param binv ...
!> \param reuse ...
! **************************************************************************************************
   SUBROUTINE mao_function(mao_coef, fval, qmat, smat, binv, reuse)
      TYPE(cp_dbcsr_type)                                :: mao_coef
      REAL(KIND=dp), INTENT(OUT)                         :: fval
      TYPE(cp_dbcsr_type)                                :: qmat, smat, binv
      LOGICAL, INTENT(IN)                                :: reuse

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_function', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: convergence, threshold
      TYPE(cp_dbcsr_type)                                :: bmat, scmat, tmat

      threshold = 1.e-8_dp
      convergence = 1.e-4_dp
      ! temp matrices
      CALL cp_dbcsr_init(scmat)
      CALL cp_dbcsr_create(scmat, template=mao_coef)
      CALL cp_dbcsr_init(bmat)
      CALL cp_dbcsr_create(bmat, template=binv)
      CALL cp_dbcsr_init(tmat)
      CALL cp_dbcsr_create(tmat, template=qmat)
      ! calculate B=C(T)*S*C matrix, S=(MAO,MAO) overlap
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, smat, mao_coef, 0.0_dp, scmat)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mao_coef, scmat, 0.0_dp, bmat)
      ! calculate inverse of B
      CALL invert_Hotelling(binv, bmat, threshold, use_inv_as_guess=reuse, &
                            norm_convergence=convergence, silent=.TRUE.)
      ! calculate Binv*C and T=C(T)*Binv*C
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mao_coef, binv, 0.0_dp, scmat)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, scmat, mao_coef, 0.0_dp, tmat)
      ! function = Tr(Q*T)
      CALL cp_dbcsr_trace(qmat, tmat, fval)
      ! free temp matrices
      CALL cp_dbcsr_release(scmat)
      CALL cp_dbcsr_release(bmat)
      CALL cp_dbcsr_release(tmat)

   END SUBROUTINE mao_function

! **************************************************************************************************
!> \brief ...
!> \param mao_coef ...
!> \param fval ...
!> \param mao_grad ...
!> \param qmat ...
!> \param smat ...
!> \param binv ...
!> \param reuse ...
! **************************************************************************************************
   SUBROUTINE mao_function_gradient(mao_coef, fval, mao_grad, qmat, smat, binv, reuse)
      TYPE(cp_dbcsr_type)                                :: mao_coef
      REAL(KIND=dp), INTENT(OUT)                         :: fval
      TYPE(cp_dbcsr_type)                                :: mao_grad, qmat, smat, binv
      LOGICAL, INTENT(IN)                                :: reuse

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_function_gradient', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: convergence, threshold
      TYPE(cp_dbcsr_type)                                :: bmat, scmat, t2mat, tmat

      threshold = 1.e-8_dp
      convergence = 1.e-4_dp
      ! temp matrices
      CALL cp_dbcsr_init(scmat)
      CALL cp_dbcsr_create(scmat, template=mao_coef)
      CALL cp_dbcsr_init(bmat)
      CALL cp_dbcsr_create(bmat, template=binv)
      CALL cp_dbcsr_init(tmat)
      CALL cp_dbcsr_create(tmat, template=qmat)
      CALL cp_dbcsr_init(t2mat)
      CALL cp_dbcsr_create(t2mat, template=scmat)
      ! calculate B=C(T)*S*C matrix, S=(MAO,MAO) overlap
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, smat, mao_coef, 0.0_dp, scmat)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mao_coef, scmat, 0.0_dp, bmat)
      ! calculate inverse of B
      CALL invert_Hotelling(binv, bmat, threshold, use_inv_as_guess=reuse, &
                            norm_convergence=convergence, silent=.TRUE.)
      ! calculate Binv*C and T=C(T)*Binv*C
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mao_coef, binv, 0.0_dp, scmat)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, scmat, mao_coef, 0.0_dp, tmat)
      ! function = Tr(Q*T)
      CALL cp_dbcsr_trace(qmat, tmat, fval)
      ! Gradient part 1: g = 2*C*Q
      CALL cp_dbcsr_multiply("N", "N", 2.0_dp, qmat, scmat, 0.0_dp, mao_grad, &
                             retain_sparsity=.TRUE.)
      ! Gradient part 2: g = -R*Q*R*S*C; R=Binv*C
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qmat, scmat, 0.0_dp, t2mat)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, scmat, t2mat, 0.0_dp, bmat)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, smat, mao_coef, 0.0_dp, scmat)
      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, scmat, bmat, 1.0_dp, mao_grad, &
                             retain_sparsity=.TRUE.)
      ! free temp matrices
      CALL cp_dbcsr_release(scmat)
      CALL cp_dbcsr_release(bmat)
      CALL cp_dbcsr_release(tmat)
      CALL cp_dbcsr_release(t2mat)

      CALL mao_project_gradient(mao_coef, mao_grad, smat)

   END SUBROUTINE mao_function_gradient

! **************************************************************************************************
!> \brief ...
!> \param mao_coef ...
!> \param smat ...
! **************************************************************************************************
   SUBROUTINE mao_orthogonalization(mao_coef, smat)
      TYPE(cp_dbcsr_type)                                :: mao_coef, smat

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_orthogonalization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, info, jatom, lwork, m, n
      LOGICAL                                            :: found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: w, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: amat, bmat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cblock, sblock
      TYPE(cp_dbcsr_iterator)                            :: dbcsr_iter

      CALL cp_dbcsr_iterator_start(dbcsr_iter, mao_coef)
      DO WHILE (cp_dbcsr_iterator_blocks_left(dbcsr_iter))
         CALL cp_dbcsr_iterator_next_block(dbcsr_iter, iatom, jatom, cblock)
         IF(.NOT.(iatom == jatom))CALL cp__a("mao_methods.F",1345)
         m = SIZE(cblock, 2)
         n = SIZE(cblock, 1)
         NULLIFY (sblock)
         CALL cp_dbcsr_get_block_p(matrix=smat, row=iatom, col=jatom, block=sblock, found=found)
         IF(.NOT.(found))CALL cp__a("mao_methods.F",1350)
         lwork = MAX(n*n, 100)
         ALLOCATE (amat(n, m), bmat(m, m), w(m), work(lwork))
         amat(1:n, 1:m) = MATMUL(sblock(1:n, 1:n), cblock(1:n, 1:m))
         bmat(1:m, 1:m) = MATMUL(TRANSPOSE(cblock(1:n, 1:m)), amat(1:n, 1:m))
         info = 0
         CALL lapack_ssyev("V", "U", m, bmat, m, w, work, lwork, info)
         IF(.NOT.(info == 0))CALL cp__a("mao_methods.F",1357)
         IF(.NOT.(ALL(w > 0.0_dp)))CALL cp__a("mao_methods.F",1358)
         w = 1.0_dp/SQRT(w)
         DO i = 1, m
            amat(1:m, i) = bmat(1:m, i)*w(i)
         END DO
         bmat(1:m, 1:m) = MATMUL(amat(1:m, 1:m), TRANSPOSE(bmat(1:m, 1:m)))
         cblock(1:n, 1:m) = MATMUL(cblock(1:n, 1:m), bmat(1:m, 1:m))
         DEALLOCATE (amat, bmat, w, work)
      END DO
      CALL cp_dbcsr_iterator_stop(dbcsr_iter)

   END SUBROUTINE mao_orthogonalization

! **************************************************************************************************
!> \brief ...
!> \param mao_coef ...
!> \param mao_grad ...
!> \param smat ...
! **************************************************************************************************
   SUBROUTINE mao_project_gradient(mao_coef, mao_grad, smat)
      TYPE(cp_dbcsr_type)                                :: mao_coef, mao_grad, smat

      CHARACTER(len=*), PARAMETER :: routineN = 'mao_project_gradient', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, jatom, m, n
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: fnin, fnout
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: amat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cblock, gblock, sblock
      TYPE(cp_dbcsr_iterator)                            :: dbcsr_iter

      CALL cp_dbcsr_norm(mao_grad, dbcsr_norm_frobenius, norm_scalar=fnin)

      CALL cp_dbcsr_iterator_start(dbcsr_iter, mao_coef)
      DO WHILE (cp_dbcsr_iterator_blocks_left(dbcsr_iter))
         CALL cp_dbcsr_iterator_next_block(dbcsr_iter, iatom, jatom, cblock)
         IF(.NOT.(iatom == jatom))CALL cp__a("mao_methods.F",1395)
         m = SIZE(cblock, 2)
         n = SIZE(cblock, 1)
         NULLIFY (sblock)
         CALL cp_dbcsr_get_block_p(matrix=smat, row=iatom, col=jatom, block=sblock, found=found)
         IF(.NOT.(found))CALL cp__a("mao_methods.F",1400)
         NULLIFY (gblock)
         CALL cp_dbcsr_get_block_p(matrix=mao_grad, row=iatom, col=jatom, block=gblock, found=found)
         IF(.NOT.(found))CALL cp__a("mao_methods.F",1403)
         ALLOCATE (amat(m, m))
         amat(1:m, 1:m) = MATMUL(TRANSPOSE(cblock(1:n, 1:m)), MATMUL(sblock(1:n, 1:n), gblock(1:n, 1:m)))
         gblock(1:n, 1:m) = gblock(1:n, 1:m)-MATMUL(cblock(1:n, 1:m), amat(1:m, 1:m))
         DEALLOCATE (amat)
      END DO
      CALL cp_dbcsr_iterator_stop(dbcsr_iter)

      CALL cp_dbcsr_norm(mao_grad, dbcsr_norm_frobenius, norm_scalar=fnout)

   END SUBROUTINE mao_project_gradient

! **************************************************************************************************
!> \brief Calculate the density matrix at the Gamma point
!> \param pmat ...
!> \param ksmat ...
!> \param smat ...
!> \param kpoints      Kpoint environment
!> \param nmos         Number of occupied orbitals
!> \param occ          Maximum occupation per orbital
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE calculate_p_gamma(pmat, ksmat, smat, kpoints, nmos, occ)

      TYPE(cp_dbcsr_type)                                :: pmat, ksmat, smat
      TYPE(kpoint_type), POINTER                         :: kpoints
      INTEGER, INTENT(IN)                                :: nmos
      REAL(KIND=dp), INTENT(IN)                          :: occ

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_p_gamma', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: norb
      REAL(KIND=dp)                                      :: de
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues
      TYPE(cp_dbcsr_type)                                :: tempmat
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      TYPE(cp_fm_type), POINTER                          :: fmksmat, fmsmat, fmvec, fmwork

      ! FM matrices

      CALL cp_dbcsr_get_info(smat, nfullrows_total=norb)
      CALL cp_fm_struct_create(fmstruct=matrix_struct, context=kpoints%blacs_env_all, &
                               nrow_global=norb, ncol_global=norb)
      CALL cp_fm_create(fmksmat, matrix_struct)
      CALL cp_fm_create(fmsmat, matrix_struct)
      CALL cp_fm_create(fmvec, matrix_struct)
      CALL cp_fm_create(fmwork, matrix_struct)
      ALLOCATE (eigenvalues(norb))

      ! DBCSR matrix
      CALL cp_dbcsr_init(tempmat)
      CALL cp_dbcsr_create(tempmat, template=smat, matrix_type=dbcsr_type_no_symmetry)

      ! transfer to FM
      CALL cp_dbcsr_desymmetrize(smat, tempmat)
      CALL copy_dbcsr_to_fm(tempmat, fmsmat)
      CALL cp_dbcsr_desymmetrize(ksmat, tempmat)
      CALL copy_dbcsr_to_fm(tempmat, fmksmat)

      ! diagonalize
      CALL cp_fm_geeig(fmksmat, fmsmat, fmvec, eigenvalues, fmwork)
      de = eigenvalues(nmos+1)-eigenvalues(nmos)
      IF (de < 0.001_dp) THEN
         CALL cp_warn(cp__l("mao_methods.F",1468), "MAO: No band gap at "// &
                      "Gamma point. MAO analysis not reliable.")
      END IF
      ! density matrix
      CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=pmat, matrix_v=fmvec, ncol=nmos, alpha=occ)

      DEALLOCATE (eigenvalues)
      CALL cp_dbcsr_release(tempmat)
      CALL cp_fm_release(fmksmat)
      CALL cp_fm_release(fmsmat)
      CALL cp_fm_release(fmvec)
      CALL cp_fm_release(fmwork)
      CALL cp_fm_struct_release(matrix_struct)

   END SUBROUTINE calculate_p_gamma

END MODULE mao_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Does all kind of post scf calculations for GPW/GAPW
!> \par History
!>      Started as a copy from the relevant part of qs_scf
!> \author Joost VandeVondele (10.2003)
! **************************************************************************************************
MODULE qs_scf_wfn_mix
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_mo_io,                        ONLY: write_mo_set
   USE qs_mo_methods,                   ONLY: calculate_orthonormality
   USE qs_mo_types,                     ONLY: deallocate_mo_set,&
                                              duplicate_mo_set,&
                                              mo_set_p_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              special_diag_method_nr
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_wfn_mix'
   PUBLIC :: wfn_mix

CONTAINS

! **************************************************************************************************
!> \brief writes a new 'mixed' set of mos to restart file, without touching the current MOs
!> \param mos ...
!> \param particle_set ...
!> \param dft_section ...
!> \param qs_kind_set ...
!> \param unoccupied_orbs ...
!> \param scf_env ...
!> \param matrix_s ...
!> \param output_unit ...
!> \param marked_states ...
! **************************************************************************************************
   SUBROUTINE wfn_mix(mos, particle_set, dft_section, qs_kind_set, &
                      unoccupied_orbs, scf_env, matrix_s, output_unit, marked_states)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: unoccupied_orbs
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      INTEGER                                            :: output_unit
      INTEGER, DIMENSION(:, :, :), OPTIONAL, POINTER     :: marked_states

      CHARACTER(len=*), PARAMETER :: routineN = 'wfn_mix', routineP = moduleN//':'//routineN

      INTEGER :: handle, i_rep, ispin, mark_ind, mark_number, n_rep, orig_mo_index, &
         orig_spin_index, result_mo_index, result_spin_index
      LOGICAL                                            :: explicit, orig_is_virtual, overwrite_mos
      REAL(KIND=dp)                                      :: orig_scale, orthonormality, result_scale
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_vector
      TYPE(cp_fm_type), POINTER                          :: matrix_x, matrix_y
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_new
      TYPE(section_vals_type), POINTER                   :: update_section, wfn_mix_section

      CALL timeset(routineN, handle)
      wfn_mix_section => section_vals_get_subs_vals(dft_section, "PRINT%WFN_MIX")
      CALL section_vals_get(wfn_mix_section, explicit=explicit)

      ! only perform action if explicitly required
      IF (explicit) THEN

         IF (output_unit > 0) THEN
            WRITE (output_unit, '()')
            WRITE (output_unit, '(T2,A)') "Performing wfn mixing"
            WRITE (output_unit, '(T2,A)') "====================="
         ENDIF

         ALLOCATE (mos_new(SIZE(mos)))
         DO ispin = 1, SIZE(mos)
            CALL duplicate_mo_set(mos_new(ispin)%mo_set, mos(ispin)%mo_set)
         ENDDO

         ! a single vector matrix structure
         NULLIFY (fm_struct_vector)
         CALL cp_fm_struct_create(fm_struct_vector, template_fmstruct=mos(1)%mo_set%mo_coeff%matrix_struct, &
                                  ncol_global=1)
         CALL cp_fm_create(matrix_x, fm_struct_vector, name="x")
         CALL cp_fm_create(matrix_y, fm_struct_vector, name="y")
         CALL cp_fm_struct_release(fm_struct_vector)

         update_section => section_vals_get_subs_vals(wfn_mix_section, "UPDATE")
         CALL section_vals_get(update_section, n_repetition=n_rep)
         CALL section_vals_get(update_section, explicit=explicit)
         IF (.NOT. explicit) n_rep = 0

         DO i_rep = 1, n_rep

            CALL section_vals_val_get(update_section, "RESULT_MO_INDEX", i_rep_section=i_rep, i_val=result_mo_index)
            CALL section_vals_val_get(update_section, "RESULT_MARKED_STATE", i_rep_section=i_rep, i_val=mark_number)
            CALL section_vals_val_get(update_section, "RESULT_SPIN_INDEX", i_rep_section=i_rep, i_val=result_spin_index)
            CALL section_vals_val_get(update_section, "RESULT_SCALE", i_rep_section=i_rep, r_val=result_scale)

            mark_ind = 1
            IF (mark_number .GT. 0) result_mo_index = marked_states(mark_number, result_spin_index, mark_ind)

            CALL section_vals_val_get(update_section, "ORIG_MO_INDEX", i_rep_section=i_rep, i_val=orig_mo_index)
            CALL section_vals_val_get(update_section, "ORIG_MARKED_STATE", i_rep_section=i_rep, i_val=mark_number)
            CALL section_vals_val_get(update_section, "ORIG_SPIN_INDEX", i_rep_section=i_rep, i_val=orig_spin_index)
            CALL section_vals_val_get(update_section, "ORIG_SCALE", i_rep_section=i_rep, r_val=orig_scale)
            CALL section_vals_val_get(update_section, "ORIG_IS_VIRTUAL", i_rep_section=i_rep, l_val=orig_is_virtual)

            IF (orig_is_virtual) mark_ind = 2
            IF (mark_number .GT. 0) orig_mo_index = marked_states(mark_number, orig_spin_index, mark_ind)

            CALL section_vals_val_get(wfn_mix_section, "OVERWRITE_MOS", l_val=overwrite_mos)

            ! first get a copy of the proper orig
            IF (.NOT. ORIG_IS_VIRTUAL) THEN
               CALL cp_fm_to_fm(mos(orig_spin_index)%mo_set%mo_coeff, matrix_x, 1, &
                                mos(orig_spin_index)%mo_set%nmo-orig_mo_index+1, 1)
            ELSE
               CALL cp_fm_to_fm(unoccupied_orbs(orig_spin_index)%matrix, matrix_x, 1, orig_mo_index, 1)
            ENDIF

            ! now get a copy of the target
            CALL cp_fm_to_fm(mos_new(result_spin_index)%mo_set%mo_coeff, matrix_y, &
                             1, mos_new(result_spin_index)%mo_set%nmo-result_mo_index+1, 1)

            ! properly mix
            CALL cp_fm_scale_and_add(result_scale, matrix_y, orig_scale, matrix_x)

            ! and copy back in the result mos
            CALL cp_fm_to_fm(matrix_y, mos_new(result_spin_index)%mo_set%mo_coeff, &
                             1, 1, mos_new(result_spin_index)%mo_set%nmo-result_mo_index+1)

         ENDDO

         CALL cp_fm_release(matrix_x)
         CALL cp_fm_release(matrix_y)

         IF (scf_env%method == special_diag_method_nr) THEN
            CALL calculate_orthonormality(orthonormality, mos)
         ELSE
            CALL calculate_orthonormality(orthonormality, mos, matrix_s(1)%matrix)
         END IF

         IF (output_unit > 0) THEN
            WRITE (output_unit, '()')
            WRITE (output_unit, '(T2,A,T61,E20.4)') &
               "Maximum deviation from MO S-orthonormality", orthonormality
            WRITE (output_unit, '(T2,A)') "Writing new MOs to file"
         ENDIF

         ! *** Write WaveFunction restart file ***

         DO ispin = 1, SIZE(mos_new)
            IF (overwrite_mos) THEN
               CALL cp_fm_to_fm(mos_new(ispin)%mo_set%mo_coeff, mos(ispin)%mo_set%mo_coeff)
               IF (mos_new(1)%mo_set%use_mo_coeff_b) &
                  CALL copy_fm_to_dbcsr(mos_new(ispin)%mo_set%mo_coeff, mos_new(ispin)%mo_set%mo_coeff_b)
            END IF
            IF (mos(1)%mo_set%use_mo_coeff_b) &
               CALL copy_fm_to_dbcsr(mos_new(ispin)%mo_set%mo_coeff, mos(ispin)%mo_set%mo_coeff_b)
         END DO
         CALL write_mo_set(mos_new, particle_set, dft_section=dft_section, qs_kind_set=qs_kind_set)

         DO ispin = 1, SIZE(mos_new)
            CALL deallocate_mo_set(mos_new(ispin)%mo_set)
         ENDDO
         DEALLOCATE (mos_new)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE wfn_mix

END MODULE qs_scf_wfn_mix
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for calculating a complex matrix exponential.
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE matrix_exp

   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_add,&
                                              cp_cfm_gemm,&
                                              cp_cfm_solve
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_p_type,&
                                              cp_cfm_release,&
                                              cp_cfm_set_all,&
                                              cp_cfm_to_cfm,&
                                              cp_cfm_type
   USE cp_fm_basic_linalg,              ONLY: cp_complex_fm_gemm,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_solve
   USE cp_fm_struct,                    ONLY: cp_fm_struct_double,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac
   USE message_passing,                 ONLY: mp_comm_free,&
                                              mp_max,&
                                              mp_reordering,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'matrix_exp'

   PUBLIC :: taylor_only_imaginary, &
             taylor_full_complex, &
             exp_pade_full_complex, &
             exp_pade_only_imaginary, &
             get_nsquare_norder, &
             arnoldi, exp_pade_real

CONTAINS

! **************************************************************************************************
!> \brief specialized subroutine for purely imaginary matrix exponentials
!> \param exp_H ...
!> \param im_matrix ...
!> \param nsquare ...
!> \param ntaylor ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE taylor_only_imaginary(exp_H, im_matrix, nsquare, ntaylor)
      TYPE(cp_fm_p_type), DIMENSION(2)                   :: exp_H
      TYPE(cp_fm_type), POINTER                          :: im_matrix
      INTEGER, INTENT(in)                                :: nsquare, ntaylor

      CHARACTER(len=*), PARAMETER :: routineN = 'taylor_only_imaginary', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, ndim, nloop
      REAL(KIND=dp)                                      :: square_fac, Tfac, tmp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data_im
      TYPE(cp_fm_type), POINTER                          :: T1, T2, Tres_im, Tres_re

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(im_matrix, local_data=local_data_im)
      ndim = im_matrix%matrix_struct%nrow_global

      square_fac = 1.0_dp/(2.0_dp**REAL(nsquare, dp))
!    CALL cp_fm_scale(square_fac,im_matrix)
      CALL cp_fm_create(T1, &
                        matrix_struct=im_matrix%matrix_struct, &
                        name="T1")

      CALL cp_fm_create(T2, &
                        matrix_struct=T1%matrix_struct, &
                        name="T2")
      CALL cp_fm_create(Tres_im, &
                        matrix_struct=T1%matrix_struct, &
                        name="T3")
      CALL cp_fm_create(Tres_re, &
                        matrix_struct=T1%matrix_struct, &
                        name="Tres")
      tmp = 1.0_dp

      CALL cp_fm_set_all(Tres_re, zero, one)
      CALL cp_fm_set_all(Tres_im, zero, zero)
      CALL cp_fm_set_all(T1, zero, one)

      Tfac = one
      nloop = CEILING(REAL(ntaylor, dp)/2.0_dp)

      DO i = 1, nloop
         tmp = tmp*(REAL(i, dp)*2.0_dp-1.0_dp)
         CALL cp_gemm("N", "N", ndim, ndim, ndim, square_fac, im_matrix, T1, zero, &
                      !       CALL cp_gemm("N","N",ndim,ndim,ndim,one,im_matrix,T1,zero,&
                      T2)
         Tfac = 1._dp/tmp
         IF (MOD(i, 2) == 0) Tfac = -Tfac
         CALL cp_fm_scale_and_add(one, Tres_im, Tfac, T2)
         tmp = tmp*REAL(i, dp)*2.0_dp
         CALL cp_gemm("N", "N", ndim, ndim, ndim, square_fac, im_matrix, T2, zero, &
                      !       CALL cp_gemm("N","N",ndim,ndim,ndim,one,im_matrix,T2,zero,&
                      T1)
         Tfac = 1._dp/tmp
         IF (MOD(i, 2) == 1) Tfac = -Tfac
         CALL cp_fm_scale_and_add(one, Tres_re, Tfac, T1)

      END DO

      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_complex_fm_gemm("N", "N", ndim, ndim, ndim, one, Tres_re, Tres_im, &
                                    Tres_re, Tres_im, zero, exp_H(1)%matrix, exp_H(2)%matrix)

            CALL cp_fm_to_fm(exp_H(1)%matrix, Tres_re)
            CALL cp_fm_to_fm(exp_H(2)%matrix, Tres_im)
         END DO
      ELSE
         CALL cp_fm_to_fm(Tres_re, exp_H(1)%matrix)
         CALL cp_fm_to_fm(Tres_im, exp_H(2)%matrix)
      END IF

      CALL cp_fm_release(T1)
      CALL cp_fm_release(T2)
      CALL cp_fm_release(Tres_re)
      CALL cp_fm_release(Tres_im)

      CALL timestop(handle)

   END SUBROUTINE taylor_only_imaginary

! **************************************************************************************************
!> \brief subroutine for general complex matrix exponentials
!>        on input a separate cp_fm_type for real and complex part
!>        on output a size 2 cp_fm_p_type, first element is the real part of
!>        the exponential second the imaginary
!> \param exp_H ...
!> \param re_part ...
!> \param im_part ...
!> \param nsquare ...
!> \param ntaylor ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE taylor_full_complex(exp_H, re_part, im_part, nsquare, ntaylor)
      TYPE(cp_fm_p_type), DIMENSION(2)                   :: exp_H
      TYPE(cp_fm_type), POINTER                          :: re_part, im_part
      INTEGER, INTENT(in)                                :: nsquare, ntaylor

      CHARACTER(len=*), PARAMETER :: routineN = 'taylor_full_complex', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp)                                   :: Tfac
      INTEGER                                            :: handle, i, ndim
      REAL(KIND=dp)                                      :: square_fac, tmp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data_im, local_data_re
      TYPE(cp_cfm_type), POINTER                         :: T1, T2, T3, Tres

      CALL timeset(routineN, handle)
      CALL cp_fm_get_info(re_part, local_data=local_data_re)
      CALL cp_fm_get_info(im_part, local_data=local_data_im)
      ndim = re_part%matrix_struct%nrow_global

      CALL cp_cfm_create(T1, &
                         matrix_struct=re_part%matrix_struct, &
                         name="T1")

      square_fac = 2.0_dp**REAL(nsquare, dp)

      T1%local_data = CMPLX(local_data_re/square_fac, local_data_im/square_fac, KIND=dp)

      CALL cp_cfm_create(T2, &
                         matrix_struct=T1%matrix_struct, &
                         name="T2")
      CALL cp_cfm_create(T3, &
                         matrix_struct=T1%matrix_struct, &
                         name="T3")
      CALL cp_cfm_create(Tres, &
                         matrix_struct=T1%matrix_struct, &
                         name="Tres")
      tmp = 1.0_dp
      CALL cp_cfm_set_all(Tres, zero, one)
      CALL cp_cfm_set_all(T2, zero, one)
      Tfac = one

      DO i = 1, ntaylor
         tmp = tmp*REAL(i, dp)
         CALL cp_cfm_gemm("N", "N", ndim, ndim, ndim, one, T1, T2, zero, &
                          T3)
         Tfac = CMPLX(1._dp/tmp, 0.0_dp, kind=dp)
         CALL cp_cfm_add(one, Tres, Tfac, T3)
         CALL cp_cfm_to_cfm(T3, T2)
      END DO

      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_cfm_gemm("N", "N", ndim, ndim, ndim, one, Tres, Tres, zero, &
                             T2)
            CALL cp_cfm_to_cfm(T2, Tres)
         END DO
      END IF

      exp_H(1)%matrix%local_data = REAL(Tres%local_data, KIND=dp)
      exp_H(2)%matrix%local_data = AIMAG(Tres%local_data)

      CALL cp_cfm_release(T1)
      CALL cp_cfm_release(T2)
      CALL cp_cfm_release(T3)
      CALL cp_cfm_release(Tres)
      CALL timestop(handle)

   END SUBROUTINE taylor_full_complex

! **************************************************************************************************
!> \brief optimization function for pade/taylor order and number of squaring steps
!> \param norm ...
!> \param nsquare ...
!> \param norder ...
!> \param eps_exp ...
!> \param method ...
!> \param do_emd ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************
   SUBROUTINE get_nsquare_norder(norm, nsquare, norder, eps_exp, method, do_emd)

      REAL(dp), INTENT(in)                               :: norm
      INTEGER, INTENT(out)                               :: nsquare, norder
      REAL(dp), INTENT(in)                               :: eps_exp
      INTEGER, INTENT(in)                                :: method
      LOGICAL, INTENT(in)                                :: do_emd

      CHARACTER(len=*), PARAMETER :: routineN = 'get_nsquare_norder', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cost, i, iscale, orders(3), p, &
                                                            prev_cost, q
      LOGICAL                                            :: new_scale
      REAL(dp)                                           :: D, eval, myval, N, scaleD, scaleN

      orders(:) = (/12, 12, 12/)
      IF (method == 2) THEN
         DO iscale = 0, 12
            new_scale = .FALSE.
            eval = norm/(2.0_dp**REAL(iscale, dp))
            DO q = 1, 12
               DO p = MAX(1, q-1), q
                  IF (p > q) EXIT
                  D = 1.0_dp
                  N = 1.0_dp
                  DO i = 1, q
                     IF (i .LE. p) scaleN = fac(p+q-i)*fac(p)/(fac(p+q)*fac(i)*fac(p-i))
                     scaleD = (-1.0)**i*fac(p+q-i)*fac(q)/(fac(p+q)*fac(i)*fac(q-i))
                     IF (i .LE. p) N = N+scaleN*eval**i
                     D = D+scaleD*eval**i
                  END DO
                  IF (ABS((EXP(norm)-(N/D)**(2.0_dp**iscale))/MAX(1.0_dp, EXP(norm))) .LE. eps_exp) THEN
                     IF (do_emd) THEN
                        cost = iscale+q
                        prev_cost = orders(1)+orders(2)
                     ELSE
                        cost = iscale+CEILING(REAL(q, dp)/3.0_dp)
                        prev_cost = orders(1)+CEILING(REAL(orders(2), dp)/3.0_dp)
                     END IF
                     IF (cost .LT. prev_cost) THEN
                        orders(:) = (/iscale, q, p/)
                        myval = (N/D)**(2.0_dp**iscale)
                     END IF
                     new_scale = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (new_scale) EXIT
            END DO
            IF (iscale .GE. orders(1)+orders(2) .AND. new_scale) EXIT
         END DO
      ELSE IF (method == 1) THEN
         q = 0
         eval = norm
         DO iscale = 0, 6
            new_scale = .FALSE.
            IF (iscale .GE. 1) eval = norm/(2.0_dp**REAL(iscale, dp))
            DO p = 1, 20
               D = 1.0_dp
               N = 1.0_dp
               DO i = 1, p
                  scaleN = 1.0_dp/fac(i)
                  N = N+scaleN*(eval**REAL(i, dp))
               END DO
               IF (ABS((EXP(norm)-N**(2.0_dp**REAL(iscale, dp)))/MAX(1.0_dp, EXP(norm))) .LE. eps_exp) THEN
                  IF (do_emd) THEN
                     cost = iscale+p
                     prev_cost = orders(1)+orders(2)
                  ELSE
                     cost = iscale+CEILING(REAL(p, dp)/3.0_dp)
                     prev_cost = orders(1)+CEILING(REAL(orders(2), dp)/3.0_dp)
                  END IF
                  IF (cost .LT. prev_cost) THEN
                     orders(:) = (/iscale, p, 0/)
                     myval = (N)**(2.0_dp**iscale)
                  END IF
                  new_scale = .TRUE.
                  EXIT
               END IF
            END DO
            IF (iscale .GE. orders(1)+orders(2) .AND. new_scale) EXIT
         END DO
      END IF

      nsquare = orders(1)
      norder = orders(2)

   END SUBROUTINE get_nsquare_norder

! **************************************************************************************************
!> \brief exponential of a complex matrix,
!>        calculated using pade approximation together with scaling and squaring
!> \param exp_H ...
!> \param re_part ...
!> \param im_part ...
!> \param nsquare ...
!> \param npade ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE exp_pade_full_complex(exp_H, re_part, im_part, nsquare, npade)
      TYPE(cp_fm_p_type), DIMENSION(2)                   :: exp_H
      TYPE(cp_fm_type), POINTER                          :: re_part, im_part
      INTEGER, INTENT(in)                                :: nsquare, npade

      CHARACTER(len=*), PARAMETER :: routineN = 'exp_pade_full_complex', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp)                                   :: scaleD, scaleN
      INTEGER                                            :: handle, i, ldim, ndim, p, q
      REAL(KIND=dp)                                      :: square_fac, tmp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data_im, local_data_re
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: mult_p
      TYPE(cp_cfm_type), POINTER                         :: Dpq, fin_p, Npq, T1, T2, Tres

      p = npade
      q = npade

      CALL timeset(routineN, handle)
      CALL cp_fm_get_info(re_part, local_data=local_data_re, ncol_local=ldim, &
                          nrow_global=ndim)
      CALL cp_fm_get_info(im_part, local_data=local_data_im)

      ALLOCATE (mult_p(2))

      CALL cp_cfm_create(Dpq, &
                         matrix_struct=re_part%matrix_struct, &
                         name="Dpq")

      square_fac = 2.0_dp**REAL(nsquare, dp)

      CALL cp_cfm_create(T1, &
                         matrix_struct=Dpq%matrix_struct, &
                         name="T1")

      CALL cp_cfm_create(T2, &
                         matrix_struct=T1%matrix_struct, &
                         name="T2")
      CALL cp_cfm_create(Npq, &
                         matrix_struct=T1%matrix_struct, &
                         name="Npq")
      CALL cp_cfm_create(Tres, &
                         matrix_struct=T1%matrix_struct, &
                         name="Tres")

      DO i = 1, ldim
         T2%local_data(:, i) = CMPLX(local_data_re(:, i)/square_fac, local_data_im(:, i)/square_fac, KIND=dp)
      END DO
      CALL cp_cfm_to_cfm(T2, T1)
      mult_p(1)%matrix => T2
      mult_p(2)%matrix => Tres
      tmp = 1.0_dp
      CALL cp_cfm_set_all(Npq, zero, one)
      CALL cp_cfm_set_all(Dpq, zero, one)

      CALL cp_cfm_add(one, Npq, one*0.5_dp, T2)
      CALL cp_cfm_add(one, Dpq, -one*0.5_dp, T2)

      IF (npade .GT. 2) THEN
         DO i = 2, npade
            IF (i .LE. p) scaleN = CMPLX(fac(p+q-i)*fac(p)/(fac(p+q)*fac(i)*fac(p-i)), 0.0_dp, kind=dp)
            scaleD = CMPLX((-1.0_dp)**i*fac(p+q-i)*fac(q)/(fac(p+q)*fac(i)*fac(q-i)), 0.0_dp, kind=dp)
            CALL cp_cfm_gemm("N", "N", ndim, ndim, ndim, one, T1, mult_p(MOD(i, 2)+1)%matrix, zero, &
                             mult_p(MOD(i+1, 2)+1)%matrix)
            IF (i .LE. p) CALL cp_cfm_add(one, Npq, scaleN, mult_p(MOD(i+1, 2)+1)%matrix)
            IF (i .LE. q) CALL cp_cfm_add(one, Dpq, scaleD, mult_p(MOD(i+1, 2)+1)%matrix)
         END DO
      END IF

      CALL cp_cfm_solve(Dpq, Npq)

      mult_p(2)%matrix => Npq
      mult_p(1)%matrix => Tres
      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_cfm_gemm("N", "N", ndim, ndim, ndim, one, mult_p(MOD(i, 2)+1)%matrix, mult_p(MOD(i, 2)+1)%matrix, zero, &
                             mult_p(MOD(i+1, 2)+1)%matrix)
            fin_p => mult_p(MOD(i+1, 2)+1)%matrix
         END DO
      ELSE
         fin_p => Npq
      END IF
      DO i = 1, ldim
         exp_H(1)%matrix%local_data(:, i) = REAL(fin_p%local_data(:, i), KIND=dp)
         exp_H(2)%matrix%local_data(:, i) = AIMAG(fin_p%local_data(:, i))
      END DO

      CALL cp_cfm_release(Npq)
      CALL cp_cfm_release(Dpq)
      CALL cp_cfm_release(T1)
      CALL cp_cfm_release(T2)
      CALL cp_cfm_release(Tres)
      DEALLOCATE (mult_p)
      CALL timestop(handle)

   END SUBROUTINE exp_pade_full_complex

! **************************************************************************************************
!> \brief exponential of a complex matrix,
!>        calculated using pade approximation together with scaling and squaring
!> \param exp_H ...
!> \param im_part ...
!> \param nsquare ...
!> \param npade ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE exp_pade_only_imaginary(exp_H, im_part, nsquare, npade)
      TYPE(cp_fm_p_type), DIMENSION(2)                   :: exp_H
      TYPE(cp_fm_type), POINTER                          :: im_part
      INTEGER, INTENT(in)                                :: nsquare, npade

      CHARACTER(len=*), PARAMETER :: routineN = 'exp_pade_only_imaginary', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      COMPLEX(KIND=dp)                                   :: scaleD, scaleN
      INTEGER                                            :: handle, i, j, k, ldim, ndim, p, q
      REAL(KIND=dp)                                      :: my_fac, square_fac
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data_im
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: cmult_p
      TYPE(cp_cfm_type), POINTER                         :: Dpq, fin_p, Npq, T1
      TYPE(cp_fm_type), POINTER                          :: T2, Tres

      CALL timeset(routineN, handle)
      p = npade
      q = npade !p==q seems to be neccesary for the rest of the code

      CALL cp_fm_get_info(im_part, local_data=local_data_im, ncol_local=ldim, nrow_global=ndim)
      square_fac = 1.0_dp/(2.0_dp**REAL(nsquare, dp))
      ALLOCATE (cmult_p(2))

      CALL cp_cfm_create(Dpq, &
                         matrix_struct=im_part%matrix_struct, &
                         name="Dpq")

      CALL cp_cfm_create(Npq, &
                         matrix_struct=Dpq%matrix_struct, &
                         name="Npq")

      CALL cp_cfm_create(T1, &
                         matrix_struct=Dpq%matrix_struct, &
                         name="T1")

      CALL cp_fm_create(T2, &
                        matrix_struct=T1%matrix_struct, &
                        name="T2")

      CALL cp_fm_create(Tres, &
                        matrix_struct=T1%matrix_struct, &
                        name="Tres")

!    DO i=1,ldim
!       local_data_im(:,i)=local_data_im(:,i)/square_fac
!    END DO

      CALL cp_fm_to_fm(im_part, T2)

      CALL cp_cfm_set_all(Npq, zero, one)
      CALL cp_cfm_set_all(Dpq, zero, one)

      DO i = 1, ldim
         Npq%local_data(:, i) = Npq%local_data(:, i)+CMPLX(rzero, 0.5_dp*square_fac*local_data_im(:, i), dp)
         Dpq%local_data(:, i) = Dpq%local_data(:, i)-CMPLX(rzero, 0.5_dp*square_fac*local_data_im(:, i), dp)
      END DO

      IF (npade .GT. 2) THEN
         DO j = 1, FLOOR(npade/2.0_dp)
            i = 2*j
            my_fac = (-rone)**j
            IF (i .LE. p) scaleN = CMPLX(my_fac*fac(p+q-i)*fac(p)/(fac(p+q)*fac(i)*fac(p-i)), 0.0_dp, dp)
            scaleD = CMPLX(my_fac*fac(p+q-i)*fac(q)/(fac(p+q)*fac(i)*fac(q-i)), 0.0_dp, dp)
            CALL cp_gemm("N", "N", ndim, ndim, ndim, square_fac, im_part, T2, rzero, Tres)

            DO k = 1, ldim
               Npq%local_data(:, k) = Npq%local_data(:, k)+scaleN*Tres%local_data(:, k)
               Dpq%local_data(:, k) = Dpq%local_data(:, k)+scaleD*Tres%local_data(:, k)
            END DO

            IF (2*j+1 .LE. q) THEN
               i = 2*j+1
               IF (i .LE. p) scaleN = CMPLX(my_fac*fac(p+q-i)*fac(p)/(fac(p+q)*fac(i)*fac(p-i)), rzero, dp)
               scaleD = CMPLX(-my_fac*fac(p+q-i)*fac(q)/(fac(p+q)*fac(i)*fac(q-i)), rzero, dp)
               CALL cp_gemm("N", "N", ndim, ndim, ndim, square_fac, im_part, Tres, rzero, T2)

               DO k = 1, ldim
                  Npq%local_data(:, k) = Npq%local_data(:, k)+scaleN*CMPLX(rzero, T2%local_data(:, k), dp)
                  Dpq%local_data(:, k) = Dpq%local_data(:, k)+scaleD*CMPLX(rzero, T2%local_data(:, k), dp)
               END DO
            ENDIF
         END DO
      END IF

      CALL cp_cfm_solve(Dpq, Npq)

      cmult_p(2)%matrix => Npq
      cmult_p(1)%matrix => T1
      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_cfm_gemm("N", "N", ndim, ndim, ndim, one, cmult_p(MOD(i, 2)+1)%matrix, cmult_p(MOD(i, 2)+1)%matrix, zero, &
                             cmult_p(MOD(i+1, 2)+1)%matrix)
            fin_p => cmult_p(MOD(i+1, 2)+1)%matrix
         END DO
      ELSE
         fin_p => Npq
      END IF

      DO k = 1, ldim
         exp_H(1)%matrix%local_data(:, k) = REAL(fin_p%local_data(:, k), KIND=dp)
         exp_H(2)%matrix%local_data(:, k) = AIMAG(fin_p%local_data(:, k))
      END DO

      CALL cp_cfm_release(Npq)
      CALL cp_cfm_release(Dpq)
      CALL cp_cfm_release(T1)
      CALL cp_fm_release(T2)
      CALL cp_fm_release(Tres)
      DEALLOCATE (cmult_p)
      CALL timestop(handle)
   END SUBROUTINE exp_pade_only_imaginary

   ! **************************************************************************************************
!> \brief exponential of a real matrix,
!>        calculated using pade approximation together with scaling and squaring
!> \param exp_H ...
!> \param matrix ...
!> \param nsquare ...
!> \param npade ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE exp_pade_real(exp_H, matrix, nsquare, npade)
      TYPE(cp_fm_type)                                   :: exp_H
      TYPE(cp_fm_type), POINTER                          :: matrix
      INTEGER, INTENT(in)                                :: nsquare, npade

      CHARACTER(len=*), PARAMETER :: routineN = 'exp_pade_real', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, j, k, ldim, ndim, p, q
      REAL(KIND=dp)                                      :: my_fac, scaleD, scaleN, square_fac
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mult_p
      TYPE(cp_fm_type), POINTER                          :: Dpq, fin_p, Npq, T1, T2, Tres

      CALL timeset(routineN, handle)
      p = npade
      q = npade !p==q seems to be neccesary for the rest of the code

      CALL cp_fm_get_info(matrix, local_data=local_data, ncol_local=ldim, nrow_global=ndim)
      square_fac = 2.0_dp**REAL(nsquare, dp)
      ALLOCATE (mult_p(2))

      CALL cp_fm_create(Dpq, &
                        matrix_struct=matrix%matrix_struct, &
                        name="Dpq")

      CALL cp_fm_create(Npq, &
                        matrix_struct=Dpq%matrix_struct, &
                        name="Npq")

      CALL cp_fm_create(T1, &
                        matrix_struct=Dpq%matrix_struct, &
                        name="T1")

      CALL cp_fm_create(T2, &
                        matrix_struct=T1%matrix_struct, &
                        name="T2")

      CALL cp_fm_create(Tres, &
                        matrix_struct=T1%matrix_struct, &
                        name="Tres")

      DO i = 1, ldim
         T2%local_data(:, i) = local_data(:, i)/square_fac
      END DO

      CALL cp_fm_to_fm(T2, T1)
      CALL cp_fm_set_all(Npq, zero, one)
      CALL cp_fm_set_all(Dpq, zero, one)

      DO i = 1, ldim
         Npq%local_data(:, i) = Npq%local_data(:, i)+0.5_dp*local_data(:, i)
         Dpq%local_data(:, i) = Dpq%local_data(:, i)-0.5_dp*local_data(:, i)
      END DO

      mult_p(1)%matrix => T2
      mult_p(2)%matrix => Tres

      IF (npade .GE. 2) THEN
         DO j = 2, npade
            my_fac = (-1.0_dp)**j
            scaleN = fac(p+q-j)*fac(p)/(fac(p+q)*fac(j)*fac(p-j))
            scaleD = my_fac*fac(p+q-j)*fac(q)/(fac(p+q)*fac(j)*fac(q-j))
            CALL cp_gemm("N", "N", ndim, ndim, ndim, one, mult_p(MOD(j, 2)+1)%matrix, T1, &
                         zero, mult_p(MOD(j+1, 2)+1)%matrix)

            DO k = 1, ldim
               Npq%local_data(:, k) = Npq%local_data(:, k)+scaleN*mult_p(MOD(j+1, 2)+1)%matrix%local_data(:, k)
               Dpq%local_data(:, k) = Dpq%local_data(:, k)+scaleD*mult_p(MOD(j+1, 2)+1)%matrix%local_data(:, k)
            END DO
         END DO
      END IF

      CALL cp_fm_solve(Dpq, Npq)

      mult_p(2)%matrix => Npq
      mult_p(1)%matrix => T1
      IF (nsquare .GT. 0) THEN
         DO i = 1, nsquare
            CALL cp_gemm("N", "N", ndim, ndim, ndim, one, mult_p(MOD(i, 2)+1)%matrix, mult_p(MOD(i, 2)+1)%matrix, zero, &
                         mult_p(MOD(i+1, 2)+1)%matrix)
            fin_p => mult_p(MOD(i+1, 2)+1)%matrix
         END DO
      ELSE
         fin_p => Npq
      END IF

      DO k = 1, ldim
         exp_H%local_data(:, k) = fin_p%local_data(:, k)
      END DO

      CALL cp_fm_release(Npq)
      CALL cp_fm_release(Dpq)
      CALL cp_fm_release(T1)
      CALL cp_fm_release(T2)
      CALL cp_fm_release(Tres)
      DEALLOCATE (mult_p)
      CALL timestop(handle)

   END SUBROUTINE exp_pade_real

! ***************************************************************************************************
!> \brief exponential of a complex matrix,
!>        calculated using arnoldi subspace method (directly applies to the MOs)
!> \param mos_old ...
!> \param mos_new ...
!> \param eps_exp ...
!> \param Hre ...
!> \param Him ...
!> \param mos_next ...
!> \param narn_old ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE arnoldi(mos_old, mos_new, eps_exp, Hre, Him, mos_next, narn_old)

      TYPE(cp_fm_p_type), DIMENSION(2)                   :: mos_old, mos_new
      REAL(KIND=dp), INTENT(in)                          :: eps_exp
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: Hre
      TYPE(cp_fm_type), POINTER                          :: Him
      TYPE(cp_fm_p_type), DIMENSION(2), OPTIONAL         :: mos_next
      INTEGER, INTENT(inout)                             :: narn_old

      CHARACTER(len=*), PARAMETER :: routineN = 'arnoldi', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: col_group, handle, i, icol_local, idim, &
                                                            info, j, l, mydim, nao, narnoldi, &
                                                            ncol_local, newdim, nmo, npade, &
                                                            pade_step
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipivot
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_procs
      LOGICAL                                            :: convergence, double_col, double_row
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: last_norm, norm1, results
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: D, mat1, mat2, mat3, N
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: H_approx, H_approx_save
      REAL(KIND=dp)                                      :: conv_norm, prefac, scaleD, scaleN
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: V_mats
      TYPE(cp_fm_struct_type), POINTER                   :: mo_struct, newstruct
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)
      para_env => mos_new(1)%matrix%matrix_struct%para_env

      CALL cp_fm_get_info(mos_old(1)%matrix, ncol_local=ncol_local, col_indices=col_indices, &
                          nrow_global=nao, ncol_global=nmo, matrix_struct=mo_struct)
      narnoldi = MIN(18, nao)

      ALLOCATE (results(ncol_local))
      ALLOCATE (norm1(ncol_local))
      ALLOCATE (V_mats(narnoldi+1))
      ALLOCATE (last_norm(ncol_local))
      ALLOCATE (H_approx(narnoldi, narnoldi, ncol_local))
      ALLOCATE (H_approx_save(narnoldi, narnoldi, ncol_local))
      col_procs => mo_struct%context%blacs2mpi(:, mo_struct%context%mepos(2))
      CALL mp_reordering(para_env%group, col_group, col_procs)

      double_col = .TRUE.
      double_row = .FALSE.
      CALL cp_fm_struct_double(newstruct, mo_struct, mo_struct%context, double_col, double_row)
      H_approx_save = rzero

      DO i = 1, narnoldi+1
         CALL cp_fm_create(V_mats(i)%matrix, matrix_struct=newstruct, &
                           name="V_mat"//cp_to_string(i))
      END DO
      CALL cp_fm_get_info(V_mats(1)%matrix, ncol_global=newdim)

      norm1 = 0.0_dp
!$OMP PARALLEL DO PRIVATE(icol_local) DEFAULT(NONE) SHARED(V_mats,norm1,mos_old,ncol_local)
      DO icol_local = 1, ncol_local
         V_mats(1)%matrix%local_data(:, icol_local) = mos_old(1)%matrix%local_data(:, icol_local)
         V_mats(1)%matrix%local_data(:, icol_local+ncol_local) = mos_old(2)%matrix%local_data(:, icol_local)
         norm1(icol_local) = SUM(V_mats(1)%matrix%local_data(:, icol_local)**2) &
                             +SUM(V_mats(1)%matrix%local_data(:, icol_local+ncol_local)**2)
      END DO

      CALL mp_sum(norm1, col_group)
      !!! normalize the mo vectors
      norm1(:) = SQRT(norm1(:))

!$OMP PARALLEL DO PRIVATE(icol_local) DEFAULT(NONE) SHARED(V_mats,norm1,ncol_local)
      DO icol_local = 1, ncol_local
         V_mats(1)%matrix%local_data(:, icol_local) = V_mats(1)%matrix%local_data(:, icol_local)/norm1(icol_local)
         V_mats(1)%matrix%local_data(:, icol_local+ncol_local) = &
            V_mats(1)%matrix%local_data(:, icol_local+ncol_local)/norm1(icol_local)
      END DO

      ! arnoldi subspace procedure to get H_approx
      DO i = 2, narnoldi+1
         !Be careful, imaginary matrix multiplied with complex. Unfortunately requires a swap of arrays afterwards
         CALL cp_gemm("N", "N", nao, newdim, nao, 1.0_dp, Him, V_mats(i-1)%matrix, 0.0_dp, V_mats(i)%matrix)

!$OMP PARALLEL DO PRIVATE(icol_local) DEFAULT(NONE) SHARED(mos_new,V_mats,ncol_local,i)
         DO icol_local = 1, ncol_local
            mos_new(1)%matrix%local_data(:, icol_local) = V_mats(i)%matrix%local_data(:, icol_local)
            V_mats(i)%matrix%local_data(:, icol_local) = -V_mats(i)%matrix%local_data(:, icol_local+ncol_local)
            V_mats(i)%matrix%local_data(:, icol_local+ncol_local) = mos_new(1)%matrix%local_data(:, icol_local)
         END DO

         IF (PRESENT(Hre)) THEN
            CALL cp_gemm("N", "N", nao, newdim, nao, 1.0_dp, Hre, V_mats(i-1)%matrix, 1.0_dp, V_mats(i)%matrix)
         END IF

         DO l = 1, i-1
!$OMP PARALLEL DO DEFAULT(NONE) SHARED(results,V_mats,ncol_local,l,i)
            DO icol_local = 1, ncol_local
               results(icol_local) = SUM(V_mats(l)%matrix%local_data(:, icol_local)*V_mats(i)%matrix%local_data(:, icol_local))+ &
                                     SUM(V_mats(l)%matrix%local_data(:, icol_local+ncol_local)* &
                                         V_mats(i)%matrix%local_data(:, icol_local+ncol_local))
            END DO

            CALL mp_sum(results, col_group)

!$OMP PARALLEL DO DEFAULT(NONE) SHARED(H_approx_save,V_mats,ncol_local,l,i,results)
            DO icol_local = 1, ncol_local
               H_approx_save(l, i-1, icol_local) = results(icol_local)
               V_mats(i)%matrix%local_data(:, icol_local) = V_mats(i)%matrix%local_data(:, icol_local)- &
                                                            results(icol_local)*V_mats(l)%matrix%local_data(:, icol_local)
               V_mats(i)%matrix%local_data(:, icol_local+ncol_local) = &
                  V_mats(i)%matrix%local_data(:, icol_local+ncol_local)- &
                  results(icol_local)*V_mats(l)%matrix%local_data(:, icol_local+ncol_local)
            END DO
         END DO

!$OMP PARALLEL DO DEFAULT(NONE) SHARED(ncol_local,V_mats,results,i)
         DO icol_local = 1, ncol_local
            results(icol_local) = SUM(V_mats(i)%matrix%local_data(:, icol_local)**2)+ &
                                  SUM(V_mats(i)%matrix%local_data(:, icol_local+ncol_local)**2)
         END DO

         CALL mp_sum(results, col_group)

         IF (i .LE. narnoldi) THEN

!$OMP PARALLEL DO DEFAULT(NONE) SHARED(H_approx_save,last_norm,V_mats,ncol_local,i,results)
            DO icol_local = 1, ncol_local
               H_approx_save(i, i-1, icol_local) = SQRT(results(icol_local))
               last_norm(icol_local) = SQRT(results(icol_local))
               V_mats(i)%matrix%local_data(:, icol_local) = V_mats(i)%matrix%local_data(:, icol_local)/SQRT(results(icol_local))
               V_mats(i)%matrix%local_data(:, icol_local+ncol_local) = &
                  V_mats(i)%matrix%local_data(:, icol_local+ncol_local)/SQRT(results(icol_local))
            END DO
         ELSE
!$OMP PARALLEL DO DEFAULT(NONE) SHARED(ncol_local,last_norm,results)
            DO icol_local = 1, ncol_local
               last_norm(icol_local) = SQRT(results(icol_local))
            END DO
         END IF

         H_approx(:, :, :) = H_approx_save

         ! PADE approximation for exp(H_approx), everything is done locally

         convergence = .FALSE.
         IF (i .GE. narn_old) THEN
            npade = 9
            mydim = MIN(i, narnoldi)
            ALLOCATE (ipivot(mydim))
            ALLOCATE (mat1(mydim, mydim))
            ALLOCATE (mat2(mydim, mydim))
            ALLOCATE (mat3(mydim, mydim))
            ALLOCATE (N(mydim, mydim))
            ALLOCATE (D(mydim, mydim))
            DO icol_local = 1, ncol_local
               DO idim = 1, mydim
                  DO j = 1, mydim
                     mat1(idim, j) = H_approx(idim, j, icol_local)/16.0_dp
                     mat3(idim, j) = mat1(idim, j)
                  END DO
               END DO
               N = 0.0_dp
               D = 0.0_dp
               DO idim = 1, mydim
                  N(idim, idim) = rone
                  D(idim, idim) = rone
               END DO
               N(:, :) = N+0.5_dp*mat1
               D(:, :) = D-0.5_dp*mat1
               pade_step = 1
               DO idim = 1, 4
                  pade_step = pade_step+1
                  CALL dgemm("N", 'N', mydim, mydim, mydim, one, mat1(1, 1), &
                             mydim, mat3(1, 1), mydim, zero, mat2(1, 1), mydim)
                  scaleN = REAL(fac(2*npade-pade_step)*fac(npade)/ &
                                (fac(2*npade)*fac(pade_step)*fac(npade-pade_step)), dp)
                  scaleD = REAL((-1.0_dp)**pade_step*fac(2*npade-pade_step)*fac(npade)/ &
                                (fac(2*npade)*fac(pade_step)*fac(npade-pade_step)), dp)
                  N(:, :) = N+scaleN*mat2
                  D(:, :) = D+scaleD*mat2
                  pade_step = pade_step+1
                  CALL dgemm("N", 'N', mydim, mydim, mydim, one, mat2(1, 1), &
                             mydim, mat1(1, 1), mydim, zero, mat3(1, 1), mydim)
                  scaleN = REAL(fac(2*npade-pade_step)*fac(npade)/ &
                                (fac(2*npade)*fac(pade_step)*fac(npade-pade_step)), dp)
                  scaleD = REAL((-1.0_dp)**pade_step*fac(2*npade-pade_step)*fac(npade)/ &
                                (fac(2*npade)*fac(pade_step)*fac(npade-pade_step)), dp)
                  N(:, :) = N+scaleN*mat3
                  D(:, :) = D+scaleD*mat3
               END DO

               CALL dgetrf(mydim, mydim, D(1, 1), mydim, ipivot, info)
               CALL dgetrs("N", mydim, mydim, D(1, 1), mydim, ipivot, N, mydim, info)
               CALL dgemm("N", 'N', mydim, mydim, mydim, one, N(1, 1), mydim, N(1, 1), mydim, zero, mat1(1, 1), mydim)
               CALL dgemm("N", 'N', mydim, mydim, mydim, one, mat1(1, 1), mydim, mat1(1, 1), mydim, zero, N(1, 1), mydim)
               CALL dgemm("N", 'N', mydim, mydim, mydim, one, N(1, 1), mydim, N(1, 1), mydim, zero, mat1(1, 1), mydim)
               CALL dgemm("N", 'N', mydim, mydim, mydim, one, mat1(1, 1), mydim, mat1(1, 1), mydim, zero, N(1, 1), mydim)
               DO idim = 1, mydim
                  DO j = 1, mydim
                     H_approx(idim, j, icol_local) = N(idim, j)
                  END DO
               END DO
            END DO
            ! H_approx is exp(H_approx) right now, calculate new MOs and check for convergence
            conv_norm = 0.0_dp
            results = 0.0_dp
            DO icol_local = 1, ncol_local
               results(icol_local) = last_norm(icol_local)*H_approx(i-1, 1, icol_local)
               conv_norm = MAX(conv_norm, ABS(results(icol_local)))
            END DO

            CALL mp_max(conv_norm, para_env%group)

            IF (conv_norm .LT. eps_exp .OR. i .GT. narnoldi) THEN

               mos_new(1)%matrix%local_data = rzero
               mos_new(2)%matrix%local_data = rzero
               DO icol_local = 1, ncol_local
                  DO idim = 1, mydim
                     prefac = H_approx(idim, 1, icol_local)*norm1(icol_local)
                     mos_new(1)%matrix%local_data(:, icol_local) = mos_new(1)%matrix%local_data(:, icol_local)+ &
                                                                   V_mats(idim)%matrix%local_data(:, icol_local)*prefac
                     mos_new(2)%matrix%local_data(:, icol_local) = mos_new(2)%matrix%local_data(:, icol_local)+ &
                                                                   V_mats(idim)%matrix%local_data(:, icol_local+ncol_local)*prefac
                  END DO
               END DO

               IF (PRESENT(mos_next)) THEN
                  DO icol_local = 1, ncol_local
                     DO idim = 1, mydim
                        DO j = 1, mydim
                           N(idim, j) = H_approx(idim, j, icol_local)
                        END DO
                     END DO
                     CALL dgemm("N", 'N', mydim, mydim, mydim, one, N(1, 1), mydim, N(1, 1), mydim, zero, mat1(1, 1), mydim)
                     DO idim = 1, mydim
                        DO j = 1, mydim
                           H_approx(idim, j, icol_local) = mat1(idim, j)
                        END DO
                     END DO
                  END DO
                  mos_next(1)%matrix%local_data = rzero
                  mos_next(2)%matrix%local_data = rzero
                  DO icol_local = 1, ncol_local
                     DO idim = 1, mydim
                        prefac = H_approx(idim, 1, icol_local)*norm1(icol_local)
                        mos_next(1)%matrix%local_data(:, icol_local) = &
                           mos_next(1)%matrix%local_data(:, icol_local)+ &
                           V_mats(idim)%matrix%local_data(:, icol_local)*prefac
                        mos_next(2)%matrix%local_data(:, icol_local) = &
                           mos_next(2)%matrix%local_data(:, icol_local)+ &
                           V_mats(idim)%matrix%local_data(:, icol_local+ncol_local)*prefac
                     END DO
                  END DO
               END IF
               IF (conv_norm .LT. eps_exp) THEN
                  convergence = .TRUE.
                  narn_old = i-1
               END IF
            END IF

            DEALLOCATE (ipivot)
            DEALLOCATE (mat1)
            DEALLOCATE (mat2)
            DEALLOCATE (mat3)
            DEALLOCATE (N)
            DEALLOCATE (D)
         END IF
         IF (convergence) EXIT

      END DO
      IF (.NOT. convergence) &
         CALL cp__w("matrix_exp.F",967,"ARNOLDI method did not converge")
      !deallocate all work matrices

      DO i = 1, SIZE(V_mats)
         CALL cp_fm_release(V_mats(i)%matrix)
      END DO
      CALL cp_fm_struct_release(newstruct)
      CALL mp_comm_free(col_group)

      DEALLOCATE (V_mats)
      DEALLOCATE (H_approx)
      DEALLOCATE (H_approx_save)
      DEALLOCATE (results)
      DEALLOCATE (norm1)
      DEALLOCATE (last_norm)
      CALL timestop(handle)
   END SUBROUTINE arnoldi

END MODULE matrix_exp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Localization methods such as 2x2 Jacobi rotations
!>                                   Steepest Decents
!>                                   Conjugate Gradient
!> \par History
!>      Initial parallellization of jacobi (JVDV 07.2003)
!>      direct minimization using exponential parametrization (JVDV 09.2003)
!>      crazy rotations go fast (JVDV 10.2003)
!> \author CJM (04.2003)
! **************************************************************************************************
MODULE qs_localization_methods
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_column_scale,&
                                              cp_cfm_gemm,&
                                              cp_cfm_schur_product
   USE cp_cfm_diag,                     ONLY: cp_cfm_heevd
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_get_element,&
                                              cp_cfm_get_info,&
                                              cp_cfm_p_type,&
                                              cp_cfm_release,&
                                              cp_cfm_set_all,&
                                              cp_cfm_to_cfm,&
                                              cp_cfm_type
   USE cp_external_control,             ONLY: external_control
   USE cp_fm_basic_linalg,              ONLY: cp_fm_frobenius_norm,&
                                              cp_fm_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_trace,&
                                              cp_fm_transpose
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_element, cp_fm_get_info, cp_fm_get_submatrix, cp_fm_maxabsrownorm, &
        cp_fm_maxabsval, cp_fm_p_type, cp_fm_release, cp_fm_set_all, cp_fm_set_submatrix, &
        cp_fm_to_fm, cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_logger_get_default_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE matrix_exp,                      ONLY: exp_pade_real,&
                                              get_nsquare_norder
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_bcast,&
                                              mp_max,&
                                              mp_sendrecv,&
                                              mp_sum,&
                                              mp_sync
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PUBLIC :: initialize_weights, crazy_rotations, &
             direct_mini, rotate_orbitals, approx_l1_norm_sd, jacobi_rotations

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_localization_methods'

   PRIVATE

   TYPE set_c_1d_type
      COMPLEX(KIND=dp), POINTER, DIMENSION(:) :: c_array
   END TYPE
   TYPE set_c_2d_type
      COMPLEX(KIND=dp), POINTER, DIMENSION(:, :) :: c_array
   END TYPE

CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param C ...
!> \param iterations ...
!> \param eps ...
!> \param converged ...
!> \param sweeps ...
! **************************************************************************************************
   SUBROUTINE approx_l1_norm_sd(C, iterations, eps, converged, sweeps)
      TYPE(cp_fm_type), POINTER                          :: C
      INTEGER, INTENT(IN)                                :: iterations
      REAL(KIND=dp), INTENT(IN)                          :: eps
      LOGICAL, INTENT(INOUT)                             :: converged
      INTEGER, INTENT(INOUT)                             :: sweeps

      CHARACTER(len=*), PARAMETER :: routineN = 'approx_l1_norm_sd', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: taylor_order = 100
      REAL(KIND=dp), PARAMETER                           :: alpha = 0.1_dp, f2_eps = 0.01_dp

      INTEGER                                            :: handle, i, istep, k, n, ncol_local, &
                                                            nrow_local, output_unit, p
      REAL(KIND=dp)                                      :: expfactor, f2, f2old, gnorm, tnorm
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_k_k
      TYPE(cp_fm_type), POINTER                          :: CTmp, G, Gp1, Gp2, U
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      NULLIFY (CTmp, U, G, Gp1, Gp2, context, para_env, fm_struct_k_k)

      output_unit = cp_logger_get_default_io_unit()

      CALL cp_fm_struct_get(C%matrix_struct, nrow_global=n, ncol_global=k, &
                            nrow_local=nrow_local, ncol_local=ncol_local, &
                            para_env=para_env, context=context)
      CALL cp_fm_struct_create(fm_struct_k_k, para_env=para_env, context=context, &
                               nrow_global=k, ncol_global=k)
      CALL cp_fm_create(CTmp, C%matrix_struct)
      CALL cp_fm_create(U, fm_struct_k_k)
      CALL cp_fm_create(G, fm_struct_k_k)
      CALL cp_fm_create(Gp1, fm_struct_k_k)
      CALL cp_fm_create(Gp2, fm_struct_k_k)
      !
      ! printing
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(1X)')
         WRITE (output_unit, '(2X,A)') '-----------------------------------------------------------------------------'
         WRITE (output_unit, '(A,I5)') '      Nbr iterations =', iterations
         WRITE (output_unit, '(A,E10.2)') '     eps convergence =', eps
         WRITE (output_unit, '(A,I5)') '    Max Taylor order =', taylor_order
         WRITE (output_unit, '(A,E10.2)') '              f2 eps =', f2_eps
         WRITE (output_unit, '(A,E10.2)') '               alpha =', alpha
         WRITE (output_unit, '(A)') '     iteration    approx_l1_norm    g_norm   rel_err'
      ENDIF
      !
      f2old = 0.0_dp
      converged = .FALSE.
      !
      ! Start the steepest descent
      DO istep = 1, iterations
         !
         !-------------------------------------------------------------------
         ! compute f_2
         ! f_2(x)=(x^2+eps)^1/2
         f2 = 0.0_dp
         DO p = 1, ncol_local ! p
            DO i = 1, nrow_local ! i
               f2 = f2+SQRT(C%local_data(i, p)**2+f2_eps)
            ENDDO
         ENDDO
         CALL mp_sum(f2, C%matrix_struct%para_env%group)
         !write(*,*) 'qs_localize: f_2=',f2
         !-------------------------------------------------------------------
         ! compute the derivative of f_2
         ! f_2(x)=(x^2+eps)^1/2
         DO p = 1, ncol_local ! p
            DO i = 1, nrow_local ! i
               CTmp%local_data(i, p) = C%local_data(i, p)/SQRT(C%local_data(i, p)**2+f2_eps)
            ENDDO
         ENDDO
         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, CTmp, C, 0.0_dp, G)
         ! antisymmetrize
         CALL cp_fm_transpose(G, U)
         CALL cp_fm_scale_and_add(-0.5_dp, G, 0.5_dp, U)
         !
         !-------------------------------------------------------------------
         !
         CALL cp_fm_frobenius_norm(G, gnorm)
         !write(*,*) 'qs_localize: norm(G)=',gnorm
         !
         ! rescale for steepest descent
         CALL cp_fm_scale(-alpha, G)
         !
         ! compute unitary transform
         ! zeroth order
         CALL cp_fm_set_all(U, 0.0_dp, 1.0_dp)
         ! first order
         expfactor = 1.0_dp
         CALL cp_fm_scale_and_add(1.0_dp, U, expfactor, G)
         CALL cp_fm_frobenius_norm(G, tnorm)
         !write(*,*) 'Taylor expansion i=',1,' norm(X^i)/i!=',tnorm
         IF (tnorm .GT. 1.0E-10_dp) THEN
            ! other orders
            CALL cp_fm_to_fm(G, Gp1)
            DO i = 2, taylor_order
               ! new power of G
               CALL cp_gemm('N', 'N', k, k, k, 1.0_dp, G, Gp1, 0.0_dp, Gp2)
               CALL cp_fm_to_fm(Gp2, Gp1)
               ! add to the taylor expansion so far
               expfactor = expfactor/REAL(i, KIND=dp)
               CALL cp_fm_scale_and_add(1.0_dp, U, expfactor, Gp1)
               CALL cp_fm_frobenius_norm(Gp1, tnorm)
               !write(*,*) 'Taylor expansion i=',i,' norm(X^i)/i!=',tnorm*expfactor
               IF (tnorm*expfactor .LT. 1.0E-10_dp) EXIT
            ENDDO
         ENDIF
         !
         ! incrementaly rotate the MOs
         CALL cp_gemm('N', 'N', n, k, k, 1.0_dp, C, U, 0.0_dp, CTmp)
         CALL cp_fm_to_fm(CTmp, C)
         !
         ! printing
         IF (output_unit .GT. 0) THEN
            WRITE (output_unit, '(10X,I4,E18.10,2E10.2)') istep, f2, gnorm, ABS((f2-f2old)/f2)
         ENDIF
         !
         ! Are we done?
         sweeps = istep
         !IF(gnorm.LE.grad_thresh.AND.ABS((f2-f2old)/f2).LE.f2_thresh.AND.istep.GT.1) THEN
         IF (ABS((f2-f2old)/f2) .LE. eps .AND. istep .GT. 1) THEN
            converged = .TRUE.
            EXIT
         ENDIF
         f2old = f2
      ENDDO
      !
      ! here we should do one refine step to enforce C'*S*C=1 for any case
      !
      ! Print the final result
      IF (output_unit .GT. 0) WRITE (output_unit, '(A,E16.10)') ' sparseness function f2 = ', f2
      !
      ! sparsity
      !DO i=1,size(thresh,1)
      !   gnorm = 0.0_dp
      !   DO o=1,ncol_local
      !      DO p=1,nrow_local
      !         IF(ABS(C%local_data(p,o)).GT.thresh(i)) THEN
      !            gnorm = gnorm + 1.0_dp
      !         ENDIF
      !      ENDDO
      !   ENDDO
      !   CALL mp_sum(gnorm,C%matrix_struct%para_env%group)
      !   IF(output_unit.GT.0) THEN
      !      WRITE(output_unit,*) 'qs_localize: ratio2=',gnorm / ( REAL(k,KIND=dp)*REAL(n,KIND=dp) ),thresh(i)
      !   ENDIF
      !ENDDO
      !
      ! deallocate
      CALL cp_fm_struct_release(fm_struct_k_k)
      CALL cp_fm_release(CTmp)
      CALL cp_fm_release(U)
      CALL cp_fm_release(G)
      CALL cp_fm_release(Gp1)
      CALL cp_fm_release(Gp2)

      CALL timestop(handle)

   END SUBROUTINE approx_l1_norm_sd
! **************************************************************************************************
!> \brief ...
!> \param cell ...
!> \param weights ...
! **************************************************************************************************
   SUBROUTINE initialize_weights(cell, weights)
      TYPE(cell_type), INTENT(IN)                        :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: weights

      REAL(KIND=dp), DIMENSION(3, 3)                     :: metric

      metric = 0.0_dp
      CALL dgemm('T', 'N', 3, 3, 3, 1._dp, cell%hmat, 3, cell%hmat, 3, 0.0_dp, metric, 3)

      weights(1) = METRIC(1, 1)-METRIC(1, 2)-METRIC(1, 3)
      weights(2) = METRIC(2, 2)-METRIC(1, 2)-METRIC(2, 3)
      weights(3) = METRIC(3, 3)-METRIC(1, 3)-METRIC(2, 3)
      weights(4) = METRIC(1, 2)
      weights(5) = METRIC(1, 3)
      weights(6) = METRIC(2, 3)

   END SUBROUTINE initialize_weights

! **************************************************************************************************
!> \brief wrapper for the jacobi routines, should be removed if jacobi_rot_para
!>        can deal with serial para_envs.
!> \param weights ...
!> \param zij ...
!> \param vectors ...
!> \param para_env ...
!> \param max_iter ...
!> \param eps_localization ...
!> \param sweeps ...
!> \param out_each ...
!> \param target_time ...
!> \param start_time ...
!> \par History
!> \author Joost VandeVondele (02.2010)
! **************************************************************************************************
   SUBROUTINE jacobi_rotations(weights, zij, vectors, para_env, max_iter, eps_localization, &
                               sweeps, out_each, target_time, start_time)

      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      TYPE(cp_fm_type), POINTER                          :: vectors
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: max_iter
      REAL(KIND=dp), INTENT(IN)                          :: eps_localization
      INTEGER                                            :: sweeps
      INTEGER, INTENT(IN)                                :: out_each
      REAL(dp)                                           :: target_time, start_time

      IF (para_env%num_pe == 1) THEN
         CALL jacobi_rotations_serial(weights, zij, vectors, max_iter, eps_localization, sweeps, out_each)
      ELSE
         CALL jacobi_rot_para(weights, zij, vectors, para_env, max_iter, eps_localization, &
                              sweeps, out_each, target_time, start_time)
      ENDIF

   END SUBROUTINE jacobi_rotations

! **************************************************************************************************
!> \brief this routine, private to the module is a serial backup, till we have jacobi_rot_para to work in serial
!>        while the routine below works in parallel, it is too slow to be useful
!> \param weights ...
!> \param zij ...
!> \param vectors ...
!> \param max_iter ...
!> \param eps_localization ...
!> \param sweeps ...
!> \param out_each ...
! **************************************************************************************************
   SUBROUTINE jacobi_rotations_serial(weights, zij, vectors, max_iter, eps_localization, sweeps, out_each)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      TYPE(cp_fm_type), POINTER                          :: vectors
      INTEGER, INTENT(IN)                                :: max_iter
      REAL(KIND=dp), INTENT(IN)                          :: eps_localization
      INTEGER                                            :: sweeps
      INTEGER, INTENT(IN)                                :: out_each

      CHARACTER(len=*), PARAMETER :: routineN = 'jacobi_rotations_serial', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), POINTER                          :: mii(:), mij(:), mjj(:)
      INTEGER                                            :: dim2, handle, idim, istate, jstate, &
                                                            nstate, unit_nr
      REAL(KIND=dp)                                      :: ct, st, t1, t2, theta, tolerance
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: c_zij
      TYPE(cp_cfm_type), POINTER                         :: c_rmat
      TYPE(cp_fm_type), POINTER                          :: rmat

      CALL timeset(routineN, handle)

      dim2 = SIZE(zij, 2)
      NULLIFY (rmat, c_rmat, c_zij)
      ALLOCATE (c_zij(dim2))
      NULLIFY (mii, mij, mjj)
      ALLOCATE (mii(dim2), mij(dim2), mjj(dim2))

      CALL cp_fm_create(rmat, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_set_all(rmat, 0._dp, 1._dp)

      CALL cp_cfm_create(c_rmat, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_set_all(c_rmat, (0._dp, 0._dp), (1._dp, 0._dp))
      DO idim = 1, dim2
         NULLIFY (c_zij(idim)%matrix)
         CALL cp_cfm_create(c_zij(idim)%matrix, zij(1, 1)%matrix%matrix_struct)
         c_zij(idim)%matrix%local_data = CMPLX(zij(1, idim)%matrix%local_data, &
                                               zij(2, idim)%matrix%local_data, dp)
      ENDDO

      CALL cp_fm_get_info(rmat, nrow_global=nstate)
      tolerance = 1.0e10_dp
      sweeps = 0
      unit_nr = -1
      IF (rmat%matrix_struct%para_env%mepos .EQ. rmat%matrix_struct%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr()
         WRITE (unit_nr, '(T4,A )') " Localization by iterative Jacobi rotation"
      END IF
      ! do jacobi sweeps until converged
      DO WHILE (tolerance >= eps_localization .AND. sweeps < max_iter)
         sweeps = sweeps+1
         t1 = m_walltime()
         DO istate = 1, nstate
            DO jstate = istate+1, nstate
               DO idim = 1, dim2
                  CALL cp_cfm_get_element(c_zij(idim)%matrix, istate, istate, mii(idim))
                  CALL cp_cfm_get_element(c_zij(idim)%matrix, istate, jstate, mij(idim))
                  CALL cp_cfm_get_element(c_zij(idim)%matrix, jstate, jstate, mjj(idim))
               END DO
               CALL get_angle(mii, mjj, mij, weights, theta)
               st = SIN(theta)
               ct = COS(theta)
               CALL rotate_zij(istate, jstate, st, ct, c_zij)
               CALL rotate_rmat(istate, jstate, st, ct, c_rmat)
            END DO
         END DO
         CALL check_tolerance(c_zij, weights, tolerance)
         t2 = m_walltime()
         IF (unit_nr > 0 .AND. MODULO(sweeps, out_each) == 0) THEN
            WRITE (unit_nr, '(T4,A,I7,T30,A,E12.4,T60,A,F8.3)') &
               "Iteration:", sweeps, "Tolerance:", tolerance, "Time:", t2-t1
            CALL m_flush(unit_nr)
         ENDIF
      END DO

      DO idim = 1, dim2
         zij(1, idim)%matrix%local_data = REAL(c_zij(idim)%matrix%local_data, dp)
         zij(2, idim)%matrix%local_data = AIMAG(c_zij(idim)%matrix%local_data)
         CALL cp_cfm_release(c_zij(idim)%matrix)
      ENDDO
      DEALLOCATE (c_zij)
      DEALLOCATE (mii, mij, mjj)
      rmat%local_data = REAL(c_rmat%local_data, dp)
      CALL cp_cfm_release(c_rmat)
      CALL rotate_orbitals(rmat, vectors)
      CALL cp_fm_release(rmat)

      CALL timestop(handle)

   END SUBROUTINE jacobi_rotations_serial
! **************************************************************************************************
!> \brief ...
!> \param istate ...
!> \param jstate ...
!> \param st ...
!> \param ct ...
!> \param zij ...
! **************************************************************************************************
   SUBROUTINE rotate_zij(istate, jstate, st, ct, zij)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: istate, jstate
      TYPE(cp_cfm_p_type) :: zij(:)
      REAL(KIND=dp), INTENT(IN) :: st, ct
      ! Locals
      INTEGER :: idim, nstate
      COMPLEX(KIND=dp) :: st_cmplx
      INTEGER :: stride

      st_cmplx = CMPLX(st, 0.0_dp, dp)
      CALL cp_cfm_get_info(zij(1)%matrix, nrow_global=nstate)
      DO idim = 1, SIZE(zij, 1)
         CALL zrot(nstate, zij(idim)%matrix%local_data(1, istate), 1, &
                   zij(idim)%matrix%local_data(1, jstate), 1, ct, st_cmplx)
         stride = SIZE(zij(idim)%matrix%local_data, 1)
         CALL zrot(nstate, zij(idim)%matrix%local_data(istate, 1), stride, &
                   zij(idim)%matrix%local_data(jstate, 1), stride, ct, st_cmplx)
      END DO
   END SUBROUTINE rotate_zij
! **************************************************************************************************
!> \brief ...
!> \param istate ...
!> \param jstate ...
!> \param st ...
!> \param ct ...
!> \param rmat ...
! **************************************************************************************************
   SUBROUTINE rotate_rmat(istate, jstate, st, ct, rmat)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: istate, jstate
      TYPE(cp_cfm_type), POINTER :: rmat
      REAL(KIND=dp), INTENT(IN) :: ct, st
! Locals
      INTEGER :: nstate
      COMPLEX(KIND=dp) :: st_cmplx

      st_cmplx = CMPLX(st, 0.0_dp, dp)
      CALL cp_cfm_get_info(rmat, nrow_global=nstate)
      CALL zrot(nstate, rmat%local_data(1, istate), 1, rmat%local_data(1, jstate), 1, ct, st_cmplx)
   END SUBROUTINE rotate_rmat
! **************************************************************************************************
!> \brief ...
!> \param mii ...
!> \param mjj ...
!> \param mij ...
!> \param weights ...
!> \param theta ...
! **************************************************************************************************
   SUBROUTINE get_angle(mii, mjj, mij, weights, theta)
      COMPLEX(KIND=dp), POINTER                          :: mii(:), mjj(:), mij(:)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      REAL(KIND=dp), INTENT(OUT)                         :: theta

      COMPLEX(KIND=dp)                                   :: z11, z12, z22
      INTEGER                                            :: dim_m, idim
      REAL(KIND=dp)                                      :: a12, b12, d2, ratio

      a12 = 0.0_dp
      b12 = 0.0_dp
      dim_m = SIZE(mii)
      DO idim = 1, dim_m
         z11 = mii(idim)
         z22 = mjj(idim)
         z12 = mij(idim)
         a12 = a12+weights(idim)*REAL(CONJG(z12)*(z11-z22), KIND=dp)
         b12 = b12+weights(idim)*REAL((z12*CONJG(z12)- &
                                       0.25_dp*(z11-z22)*(CONJG(z11)-CONJG(z22))), KIND=dp)
      END DO
      IF (ABS(b12) > 1.e-10_dp) THEN
         ratio = -a12/b12
         theta = 0.25_dp*ATAN(ratio)
      ELSEIF (ABS(b12) < 1.e-10_dp) THEN
         b12 = 0.0_dp
         theta = 0.0_dp
      ELSE
         theta = 0.25_dp*pi
      ENDIF
! Check second derivative info
      d2 = a12*SIN(4._dp*theta)-b12*COS(4._dp*theta)
      IF (d2 <= 0._dp) THEN ! go to the maximum, not the minimum
         IF (theta > 0.0_dp) THEN ! make theta as small as possible
            theta = theta-0.25_dp*pi
         ELSE
            theta = theta+0.25_dp*pi
         ENDIF
      ENDIF
   END SUBROUTINE get_angle
! **************************************************************************************************
!> \brief ...
!> \param zij ...
!> \param weights ...
!> \param tolerance ...
! **************************************************************************************************
   SUBROUTINE check_tolerance(zij, weights, tolerance)
      TYPE(cp_cfm_p_type)                                :: zij(:)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      REAL(KIND=dp), INTENT(OUT)                         :: tolerance

      CHARACTER(len=*), PARAMETER :: routineN = 'check_tolerance', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_fm_type), POINTER                          :: force

      CALL timeset(routineN, handle)

! compute gradient at t=0

      NULLIFY (force)
      CALL cp_fm_create(force, zij(1)%matrix%matrix_struct)
      CALL cp_fm_set_all(force, 0._dp)
      CALL grad_at_0(zij, weights, force)
      CALL cp_fm_maxabsval(force, tolerance)
      CALL cp_fm_release(force)

      CALL timestop(handle)

   END SUBROUTINE check_tolerance
! **************************************************************************************************
!> \brief ...
!> \param rmat ...
!> \param vectors ...
! **************************************************************************************************
   SUBROUTINE rotate_orbitals(rmat, vectors)
      TYPE(cp_fm_type), POINTER                          :: rmat, vectors

      INTEGER                                            :: k, n
      TYPE(cp_fm_type), POINTER                          :: wf

      NULLIFY (wf)
      CALL cp_fm_create(wf, vectors%matrix_struct)
      CALL cp_fm_get_info(vectors, nrow_global=n, ncol_global=k)
      CALL cp_gemm("N", "N", n, k, k, 1.0_dp, vectors, rmat, 0.0_dp, wf)
      CALL cp_fm_to_fm(wf, vectors)
      CALL cp_fm_release(wf)
   END SUBROUTINE rotate_orbitals
! **************************************************************************************************
!> \brief ...
!> \param diag ...
!> \param weights ...
!> \param matrix ...
!> \param ndim ...
! **************************************************************************************************
   SUBROUTINE gradsq_at_0(diag, weights, matrix, ndim)
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: diag
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_type), POINTER                          :: matrix
      INTEGER, INTENT(IN)                                :: ndim

      COMPLEX(KIND=dp)                                   :: zii, zjj
      INTEGER                                            :: idim, istate, jstate, ncol_local, &
                                                            nrow_global, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: gradsq_ij

      CALL cp_fm_get_info(matrix, nrow_local=nrow_local, &
                          ncol_local=ncol_local, nrow_global=nrow_global, &
                          row_indices=row_indices, col_indices=col_indices)

      DO istate = 1, nrow_local
         DO jstate = 1, ncol_local
! get real and imaginary parts
            gradsq_ij = 0.0_dp
            DO idim = 1, ndim
               zii = diag(row_indices(istate), idim)
               zjj = diag(col_indices(jstate), idim)
               gradsq_ij = gradsq_ij+weights(idim)* &
                           4.0_dp*REAL((CONJG(zii)*zii+CONJG(zjj)*zjj), KIND=dp)
            END DO
            matrix%local_data(istate, jstate) = gradsq_ij
         END DO
      END DO
   END SUBROUTINE gradsq_at_0
! **************************************************************************************************
!> \brief ...
!> \param matrix_p ...
!> \param weights ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE grad_at_0(matrix_p, weights, matrix)
      TYPE(cp_cfm_p_type)                                :: matrix_p(:)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_type), POINTER                          :: matrix

      COMPLEX(KIND=dp)                                   :: zii, zij, zjj
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: diag
      INTEGER                                            :: dim_m, idim, istate, jstate, ncol_local, &
                                                            nrow_global, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: grad_ij

      NULLIFY (diag)
      CALL cp_fm_get_info(matrix, nrow_local=nrow_local, &
                          ncol_local=ncol_local, nrow_global=nrow_global, &
                          row_indices=row_indices, col_indices=col_indices)
      dim_m = SIZE(matrix_p, 1)
      ALLOCATE (diag(nrow_global, dim_m))

      DO idim = 1, dim_m
         DO istate = 1, nrow_global
            CALL cp_cfm_get_element(matrix_p(idim)%matrix, istate, istate, diag(istate, idim))
         ENDDO
      ENDDO

      DO istate = 1, nrow_local
         DO jstate = 1, ncol_local
! get real and imaginary parts
            grad_ij = 0.0_dp
            DO idim = 1, dim_m
               zii = diag(row_indices(istate), idim)
               zjj = diag(col_indices(jstate), idim)
               zij = matrix_p(idim)%matrix%local_data(istate, jstate)
               grad_ij = grad_ij+weights(idim)* &
                         REAL(4.0_dp*CONJG(zij)*(zjj-zii), dp)
            END DO
            matrix%local_data(istate, jstate) = grad_ij
         END DO
      END DO
      DEALLOCATE (diag)
   END SUBROUTINE grad_at_0

! return energy and maximum gradient in the current point
! **************************************************************************************************
!> \brief ...
!> \param weights ...
!> \param zij ...
!> \param tolerance ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE check_tolerance_new(weights, zij, tolerance, value)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      REAL(KIND=dp)                                      :: tolerance, value

      COMPLEX(KIND=dp)                                   :: kii, kij, kjj
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: diag
      INTEGER                                            :: idim, istate, jstate, ncol_local, &
                                                            nrow_global, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: grad_ij, ra, rb

      NULLIFY (diag)
      CALL cp_fm_get_info(zij(1, 1)%matrix, nrow_local=nrow_local, &
                          ncol_local=ncol_local, nrow_global=nrow_global, &
                          row_indices=row_indices, col_indices=col_indices)
      ALLOCATE (diag(nrow_global, SIZE(zij, 2)))
      value = 0.0_dp
      DO idim = 1, SIZE(zij, 2)
         DO istate = 1, nrow_global
            CALL cp_fm_get_element(zij(1, idim)%matrix, istate, istate, ra)
            CALL cp_fm_get_element(zij(2, idim)%matrix, istate, istate, rb)
            diag(istate, idim) = CMPLX(ra, rb, dp)
            value = value+weights(idim)-weights(idim)*ABS(diag(istate, idim))**2
         ENDDO
      ENDDO
      tolerance = 0.0_dp
      DO istate = 1, nrow_local
         DO jstate = 1, ncol_local
            grad_ij = 0.0_dp
            DO idim = 1, SIZE(zij, 2)
               kii = diag(row_indices(istate), idim)
               kjj = diag(col_indices(jstate), idim)
               ra = zij(1, idim)%matrix%local_data(istate, jstate)
               rb = zij(2, idim)%matrix%local_data(istate, jstate)
               kij = CMPLX(ra, rb, dp)
               grad_ij = grad_ij+weights(idim)* &
                         REAL(4.0_dp*CONJG(kij)*(kjj-kii), dp)
            END DO
            tolerance = MAX(ABS(grad_ij), tolerance)
         ENDDO
      ENDDO
      CALL mp_max(tolerance, zij(1, 1)%matrix%matrix_struct%para_env%group)

      DEALLOCATE (diag)

   END SUBROUTINE check_tolerance_new

! **************************************************************************************************
!> \brief yet another crazy try, computes the angles needed to rotate the orbitals first
!>        and rotates them all at the same time (hoping for the best of course)
!> \param weights ...
!> \param zij ...
!> \param vectors ...
!> \param max_iter ...
!> \param max_crazy_angle ...
!> \param crazy_scale ...
!> \param crazy_use_diag ...
!> \param eps_localization ...
!> \param iterations ...
!> \param converged ...
! **************************************************************************************************
   SUBROUTINE crazy_rotations(weights, zij, vectors, max_iter, max_crazy_angle, crazy_scale, crazy_use_diag, &
                              eps_localization, iterations, converged)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      TYPE(cp_fm_type), POINTER                          :: vectors
      INTEGER, INTENT(IN)                                :: max_iter
      REAL(KIND=dp), INTENT(IN)                          :: max_crazy_angle
      REAL(KIND=dp)                                      :: crazy_scale
      LOGICAL                                            :: crazy_use_diag
      REAL(KIND=dp), INTENT(IN)                          :: eps_localization
      INTEGER                                            :: iterations
      LOGICAL, INTENT(out), OPTIONAL                     :: converged

      CHARACTER(len=*), PARAMETER :: routineN = 'crazy_rotations', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: cone = (1.0_dp, 0.0_dp), &
                                                            czero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: evals_exp
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: diag_z
      COMPLEX(KIND=dp), POINTER                          :: mii(:), mij(:), mjj(:)
      INTEGER                                            :: dim2, handle, i, icol, idim, irow, &
                                                            method, ncol_global, ncol_local, &
                                                            norder, nrow_global, nrow_local, &
                                                            nsquare, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: do_emd
      REAL(KIND=dp)                                      :: eps_exp, limit_crazy_angle, maxeval, &
                                                            norm, ra, rb, theta, tolerance, value
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals
      TYPE(cp_cfm_type), POINTER                         :: cmat_A, cmat_R, cmat_t1
      TYPE(cp_fm_type), POINTER                          :: mat_R, mat_t, mat_theta, mat_U

      CALL timeset(routineN, handle)
      NULLIFY (row_indices, col_indices)
      NULLIFY (mat_U, mat_t, mat_R)
      NULLIFY (cmat_A, cmat_R, cmat_t1)
      CALL cp_fm_get_info(zij(1, 1)%matrix, nrow_global=nrow_global, &
                          ncol_global=ncol_global, &
                          row_indices=row_indices, col_indices=col_indices, &
                          nrow_local=nrow_local, ncol_local=ncol_local)

      limit_crazy_angle = max_crazy_angle

      NULLIFY (diag_z, evals, evals_exp, mii, mij, mjj)
      dim2 = SIZE(zij, 2)
      ALLOCATE (diag_z(nrow_global, dim2))
      ALLOCATE (evals(nrow_global))
      ALLOCATE (evals_exp(nrow_global))

      CALL cp_cfm_create(cmat_A, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_R, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_t1, zij(1, 1)%matrix%matrix_struct)

      CALL cp_fm_create(mat_U, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(mat_t, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(mat_R, zij(1, 1)%matrix%matrix_struct)

      NULLIFY (mat_theta)
      CALL cp_fm_create(mat_theta, zij(1, 1)%matrix%matrix_struct)

      CALL cp_fm_set_all(mat_R, 0.0_dp, 1.0_dp)
      CALL cp_fm_set_all(mat_t, 0.0_dp)
      ALLOCATE (mii(dim2), mij(dim2), mjj(dim2))
      DO idim = 1, dim2
         CALL cp_fm_scale_and_add(1.0_dp, mat_t, weights(idim), zij(1, idim)%matrix)
         CALL cp_fm_scale_and_add(1.0_dp, mat_t, weights(idim), zij(2, idim)%matrix)
      ENDDO
      CALL cp_fm_syevd(mat_t, mat_U, evals)
      DO idim = 1, dim2
         ! rotate z's
         CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, zij(1, idim)%matrix, mat_U, 0.0_dp, mat_t)
         CALL cp_gemm('T', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_U, mat_t, 0.0_dp, zij(1, idim)%matrix)
         CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, zij(2, idim)%matrix, mat_U, 0.0_dp, mat_t)
         CALL cp_gemm('T', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_U, mat_t, 0.0_dp, zij(2, idim)%matrix)
      ENDDO
      ! collect rotation matrix
      CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_R, mat_U, 0.0_dp, mat_t)
      CALL cp_fm_to_fm(mat_t, mat_R)

      unit_nr = -1
      IF (cmat_A%matrix_struct%para_env%mepos .EQ. cmat_A%matrix_struct%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr()
         WRITE (unit_nr, '(T2,A7,A6,1X,A20,A12,A12,A12)') &
            "CRAZY| ", "Iter", "value    ", "gradient", "Max. eval", "limit"
      ENDIF

      iterations = 0
      tolerance = 1.0_dp

      DO
         iterations = iterations+1
         DO idim = 1, dim2
            DO i = 1, nrow_global
               CALL cp_fm_get_element(zij(1, idim)%matrix, i, i, ra)
               CALL cp_fm_get_element(zij(2, idim)%matrix, i, i, rb)
               diag_z(i, idim) = CMPLX(ra, rb, dp)
            ENDDO
         ENDDO
         DO irow = 1, nrow_local
            DO icol = 1, ncol_local
               DO idim = 1, dim2
                  ra = zij(1, idim)%matrix%local_data(irow, icol)
                  rb = zij(2, idim)%matrix%local_data(irow, icol)
                  mij(idim) = CMPLX(ra, rb, dp)
                  mii(idim) = diag_z(row_indices(irow), idim)
                  mjj(idim) = diag_z(col_indices(icol), idim)
               ENDDO
               IF (row_indices(irow) .NE. col_indices(icol)) THEN
                  CALL get_angle(mii, mjj, mij, weights, theta)
                  theta = crazy_scale*theta
                  IF (theta .GT. limit_crazy_angle) theta = limit_crazy_angle
                  IF (theta .LT. -limit_crazy_angle) theta = -limit_crazy_angle
                  IF (crazy_use_diag) THEN
                     cmat_A%local_data(irow, icol) = -CMPLX(0.0_dp, theta, dp)
                  ELSE
                     mat_theta%local_data(irow, icol) = -theta
                  ENDIF
               ELSE
                  IF (crazy_use_diag) THEN
                     cmat_A%local_data(irow, icol) = czero
                  ELSE
                     mat_theta%local_data(irow, icol) = 0.0_dp
                  ENDIF
               ENDIF
            ENDDO
         ENDDO

         ! construct rotation matrix U based on A using diagonalization
         ! alternatively, exp based on repeated squaring could be faster
         IF (crazy_use_diag) THEN
            CALL cp_cfm_heevd(cmat_A, cmat_R, evals)
            maxeval = MAXVAL(ABS(evals))
            evals_exp(:) = EXP((0.0_dp, -1.0_dp)*evals(:))
            CALL cp_cfm_to_cfm(cmat_R, cmat_t1)
            CALL cp_cfm_column_scale(cmat_t1, evals_exp)
            CALL cp_cfm_gemm('N', 'C', nrow_global, nrow_global, nrow_global, cone, &
                             cmat_t1, cmat_R, czero, cmat_A)
            mat_U%local_data = REAL(cmat_A%local_data, KIND=dp) ! U is a real matrix
         ELSE
            do_emd = .FALSE.
            method = 2
            eps_exp = 1.0_dp*EPSILON(eps_exp)
            CALL cp_fm_maxabsrownorm(mat_theta, norm)
            maxeval = norm ! an upper bound
            CALL get_nsquare_norder(norm, nsquare, norder, eps_exp, method, do_emd)
            CALL exp_pade_real(mat_U, mat_theta, nsquare, norder)
         ENDIF

         DO idim = 1, dim2
            ! rotate z's
            CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, zij(1, idim)%matrix, mat_U, 0.0_dp, mat_t)
            CALL cp_gemm('T', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_U, mat_t, 0.0_dp, zij(1, idim)%matrix)
            CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, zij(2, idim)%matrix, mat_U, 0.0_dp, mat_t)
            CALL cp_gemm('T', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_U, mat_t, 0.0_dp, zij(2, idim)%matrix)
         ENDDO
         ! collect rotation matrix
         CALL cp_gemm('N', 'N', nrow_global, nrow_global, nrow_global, 1.0_dp, mat_R, mat_U, 0.0_dp, mat_t)
         CALL cp_fm_to_fm(mat_t, mat_R)

         CALL check_tolerance_new(weights, zij, tolerance, value)

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T2,A7,I6,1X,G20.15,E12.4,E12.4,E12.4)') &
               "CRAZY| ", iterations, value, tolerance, maxeval, limit_crazy_angle
            CALL m_flush(unit_nr)
         ENDIF
         IF (tolerance .LT. eps_localization .OR. iterations .GE. max_iter) EXIT
      ENDDO

      IF (PRESENT(converged)) converged = (tolerance .LT. eps_localization)

      CALL cp_cfm_release(cmat_A)
      CALL cp_cfm_release(cmat_R)
      CALL cp_cfm_release(cmat_T1)

      CALL cp_fm_release(mat_U)
      CALL cp_fm_release(mat_T)
      CALL cp_fm_release(mat_theta)

      CALL rotate_orbitals(mat_R, vectors)

      CALL cp_fm_release(mat_R)
      DEALLOCATE (evals_exp, evals, diag_z)
      DEALLOCATE (mii, mij, mjj)

      CALL timestop(handle)

   END SUBROUTINE crazy_rotations

! **************************************************************************************************
!> \brief use the exponential parametrization as described in to perform a direct mini
!>        Gerd Berghold et al. PRB 61 (15), pag. 10040 (2000)
!> none of the input is modified for the time being, just finds the rotations
!> that minimizes, and throws it away afterwards :-)
!> apart from being expensive and not cleaned, this works fine
!> useful to try different spread functionals
!> \param weights ...
!> \param zij ...
!> \param vectors ...
!> \param max_iter ...
!> \param eps_localization ...
!> \param iterations ...
! **************************************************************************************************
   SUBROUTINE direct_mini(weights, zij, vectors, max_iter, eps_localization, iterations)
      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      TYPE(cp_fm_type), POINTER                          :: vectors
      INTEGER, INTENT(IN)                                :: max_iter
      REAL(KIND=dp), INTENT(IN)                          :: eps_localization
      INTEGER                                            :: iterations

      CHARACTER(len=*), PARAMETER :: routineN = 'direct_mini', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: cone = (1.0_dp, 0.0_dp), &
                                                            czero = (0.0_dp, 0.0_dp)
      REAL(KIND=dp), PARAMETER                           :: gold_sec = 0.3819_dp

      COMPLEX(KIND=dp)                                   :: lk, ll, tmp
      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: evals_exp
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: diag_z
      INTEGER                                            :: handle, i, icol, idim, irow, &
                                                            line_search_count, line_searches, lsl, &
                                                            lsm, lsr, n, ncol_local, ndim, &
                                                            nrow_local, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: new_direction
      REAL(KIND=dp)                                      :: a, b, beta_pr, c, denom, ds, ds_min, fa, &
                                                            fb, fc, nom, normg, normg_cross, &
                                                            normg_old, npos, omega, tol, val, x0, &
                                                            x1, xa, xb, xc
      REAL(KIND=dp), DIMENSION(150)                      :: energy, grad, pos
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals, fval, fvald
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: c_zij
      TYPE(cp_cfm_type), POINTER                         :: cmat_A, cmat_B, cmat_M, cmat_R, cmat_t1, &
                                                            cmat_t2, cmat_U
      TYPE(cp_fm_type), POINTER                          :: matrix_A, matrix_G, matrix_G_old, &
                                                            matrix_G_search, matrix_H, matrix_R, &
                                                            matrix_T

      NULLIFY (evals, evals_exp, diag_z, fval, fvald, c_zij)
      NULLIFY (matrix_A, matrix_G, matrix_T, matrix_G_search, matrix_G_old)
      NULLIFY (cmat_A, cmat_U, cmat_R, cmat_t1, cmat_t2, cmat_B, cmat_M)

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()

      n = zij(1, 1)%matrix%matrix_struct%nrow_global
      ndim = (SIZE(zij, 2))

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T4,A )') "Localization by direct minimization of the functional; "
         WRITE (output_unit, '(T5,2A13,A20,A20,A10 )') " Line search ", " Iteration ", " Functional ", " Tolerance ", " ds Min "
      END IF

      ALLOCATE (evals(n), evals_exp(n), diag_z(n, ndim), fval(n), fvald(n))
      ALLOCATE (c_zij(ndim))

      ! create the three complex matrices Z
      DO idim = 1, ndim
         NULLIFY (c_zij(idim)%matrix)
         CALL cp_cfm_create(c_zij(idim)%matrix, zij(1, 1)%matrix%matrix_struct)
         c_zij(idim)%matrix%local_data = CMPLX(zij(1, idim)%matrix%local_data, &
                                               zij(2, idim)%matrix%local_data, dp)
      ENDDO

      CALL cp_fm_create(matrix_A, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_G, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_T, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_H, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_G_search, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_G_old, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_create(matrix_R, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_set_all(matrix_R, 0.0_dp, 1.0_dp)

      CALL cp_fm_set_all(matrix_A, 0.0_dp)
!    CALL cp_fm_init_random ( matrix_A )

      CALL cp_cfm_create(cmat_A, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_U, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_R, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_t1, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_t2, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_B, zij(1, 1)%matrix%matrix_struct)
      CALL cp_cfm_create(cmat_M, zij(1, 1)%matrix%matrix_struct)

      CALL cp_cfm_get_info(cmat_B, nrow_local=nrow_local, ncol_local=ncol_local, &
                           row_indices=row_indices, col_indices=col_indices)

      CALL cp_fm_set_all(matrix_G_old, 0.0_dp)
      CALL cp_fm_set_all(matrix_G_search, 0.0_dp)
      normg_old = 1.0E30_dp
      ds_min = 1.0_dp
      new_direction = .TRUE.
      Iterations = 0
      line_searches = 0
      line_search_count = 0
      DO
         iterations = iterations+1
         ! compute U,R,evals given A
         cmat_A%local_data = CMPLX(0.0_dp, matrix_A%local_data, dp) ! cmat_A is hermitian, evals are reals
         CALL cp_cfm_heevd(cmat_A, cmat_R, evals)
         evals_exp(:) = EXP((0.0_dp, -1.0_dp)*evals(:))
         CALL cp_cfm_to_cfm(cmat_R, cmat_t1)
         CALL cp_cfm_column_scale(cmat_t1, evals_exp)
         CALL cp_cfm_gemm('N', 'C', n, n, n, cone, cmat_t1, cmat_R, czero, cmat_U)
         cmat_U%local_data = REAL(cmat_U%local_data, KIND=dp) ! enforce numerics, U is a real matrix

         IF (new_direction .AND. MOD(line_searches, 20) .EQ. 5) THEN ! reset with A .eq. 0
            DO idim = 1, ndim
               CALL cp_cfm_gemm('N', 'N', n, n, n, cone, c_zij(idim)%matrix, cmat_U, czero, cmat_t1)
               CALL cp_cfm_gemm('C', 'N', n, n, n, cone, cmat_U, cmat_t1, czero, c_zij(idim)%matrix)
            ENDDO
            ! collect rotation matrix
            matrix_H%local_data = REAL(cmat_U%local_data, KIND=dp)
            CALL cp_gemm('N', 'N', n, n, n, 1.0_dp, matrix_R, matrix_H, 0.0_dp, matrix_T)
            CALL cp_fm_to_fm(matrix_T, matrix_R)

            CALL cp_cfm_set_all(cmat_U, czero, cone)
            CALL cp_cfm_set_all(cmat_R, czero, cone)
            CALL cp_cfm_set_all(cmat_A, czero)
            CALL cp_fm_set_all(matrix_A, 0.0_dp)
            evals(:) = 0.0_dp
            evals_exp(:) = EXP((0.0_dp, -1.0_dp)*evals(:))
            CALL cp_fm_set_all(matrix_G_old, 0.0_dp)
            CALL cp_fm_set_all(matrix_G_search, 0.0_dp)
            normg_old = 1.0E30_dp
         ENDIF

         ! compute Omega and M
         CALL cp_cfm_set_all(cmat_M, czero)
         omega = 0.0_dp
         DO idim = 1, ndim
            CALL cp_cfm_gemm('N', 'N', n, n, n, cone, c_zij(idim)%matrix, cmat_U, czero, cmat_t1) ! t1=ZU
            CALL cp_cfm_gemm('C', 'N', n, n, n, cone, cmat_U, cmat_t1, czero, cmat_t2) ! t2=(U^T)ZU
            DO i = 1, n
               CALL cp_cfm_get_element(cmat_t2, i, i, diag_z(i, idim))
               SELECT CASE (2) ! allows for selection of different spread functionals
               CASE (1)
                  fval(i) = -weights(idim)*LOG(ABS(diag_z(i, idim))**2)
                  fvald(i) = -weights(idim)/(ABS(diag_z(i, idim))**2)
               CASE (2) ! corresponds to the jacobi setup
                  fval(i) = weights(idim)-weights(idim)*ABS(diag_z(i, idim))**2
                  fvald(i) = -weights(idim)
               END SELECT
               omega = omega+fval(i)
            ENDDO
            DO icol = 1, ncol_local
               DO irow = 1, nrow_local
                  tmp = cmat_t1%local_data(irow, icol)*CONJG(diag_z(col_indices(icol), idim))
                  cmat_M%local_data(irow, icol) = cmat_M%local_data(irow, icol) &
                                                  +4.0_dp*fvald(col_indices(icol))*REAL(tmp, KIND=dp)
               ENDDO
            ENDDO
         ENDDO

         ! compute Hessian diagonal approximation for the preconditioner
         IF (.TRUE.) THEN
            CALL gradsq_at_0(diag_z, weights, matrix_H, ndim)
         ELSE
            CALL cp_fm_set_all(matrix_H, 1.0_dp)
         ENDIF

         ! compute B
         DO icol = 1, ncol_local
            DO irow = 1, nrow_local
               ll = (0.0_dp, -1.0_dp)*evals(row_indices(irow))
               lk = (0.0_dp, -1.0_dp)*evals(col_indices(icol))
               IF (ABS(ll-lk) .LT. 0.5_dp) THEN ! use a series expansion to avoid loss of precision
                  tmp = 1.0_dp
                  cmat_B%local_data(irow, icol) = 0.0_dp
                  DO i = 1, 16
                     cmat_B%local_data(irow, icol) = cmat_B%local_data(irow, icol)+tmp
                     tmp = tmp*(ll-lk)/(i+1)
                  ENDDO
                  cmat_B%local_data(irow, icol) = cmat_B%local_data(irow, icol)*EXP(lk)
               ELSE
                  cmat_B%local_data(irow, icol) = (EXP(lk)-EXP(ll))/(lk-ll)
               ENDIF
            ENDDO
         ENDDO
         ! compute gradient matrix_G

         CALL cp_cfm_gemm('C', 'N', n, n, n, cone, cmat_M, cmat_R, czero, cmat_t1) ! t1=(M^T)(R^T)
         CALL cp_cfm_gemm('C', 'N', n, n, n, cone, cmat_R, cmat_t1, czero, cmat_t2) ! t2=(R)t1
         CALL cp_cfm_schur_product(cmat_t2, cmat_B, cmat_t1)
         CALL cp_cfm_gemm('N', 'C', n, n, n, cone, cmat_t1, cmat_R, czero, cmat_t2)
         CALL cp_cfm_gemm('N', 'N', n, n, n, cone, cmat_R, cmat_t2, czero, cmat_t1)
         matrix_G%local_data = REAL(cmat_t1%local_data, KIND=dp)
         CALL cp_fm_transpose(matrix_G, matrix_T)
         CALL cp_fm_scale_and_add(-1.0_dp, matrix_G, 1.0_dp, matrix_T)
         CALL cp_fm_maxabsval(matrix_G, tol)

         ! from here on, minimizing technology
         IF (new_direction) THEN
            ! energy converged up to machine precision ?
            line_searches = line_searches+1
            ! DO i=1,line_search_count
            !   write(15,*) pos(i),energy(i)
            ! ENDDO
            ! write(15,*) ""
            ! CALL m_flush(15)
            !write(16,*) evals(:)
            !write(17,*) matrix_A%local_data(:,:)
            !write(18,*) matrix_G%local_data(:,:)
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T5,I10,T18,I10,T31,2F20.6,F10.3)') line_searches, Iterations, Omega, tol, ds_min
               CALL m_flush(output_unit)
            ENDIF
            IF (tol < eps_localization .OR. iterations > max_iter) EXIT

            IF (.TRUE.) THEN ! do conjugate gradient CG
               CALL cp_fm_trace(matrix_G, matrix_G_old, normg_cross)
               normg_cross = normg_cross*0.5_dp ! takes into account the fact that A is antisymmetric
               ! apply the preconditioner
               DO icol = 1, ncol_local
                  DO irow = 1, nrow_local
                     matrix_G_old%local_data(irow, icol) = matrix_G%local_data(irow, icol)/matrix_H%local_data(irow, icol)
                  ENDDO
               ENDDO
               CALL cp_fm_trace(matrix_G, matrix_G_old, normg)
               normg = normg*0.5_dp
               beta_pr = (normg-normg_cross)/normg_old
               normg_old = normg
               beta_pr = MAX(beta_pr, 0.0_dp)
               CALL cp_fm_scale_and_add(beta_pr, matrix_G_search, -1.0_dp, matrix_G_old)
               CALL cp_fm_trace(matrix_G_search, matrix_G_old, normg_cross)
               IF (normg_cross .GE. 0) THEN ! back to SD
                  IF (matrix_A%matrix_struct%para_env%mepos .EQ. &
                      matrix_A%matrix_struct%para_env%source) THEN
                     WRITE (cp_logger_get_default_unit_nr(), *) "!"
                  ENDIF
                  beta_pr = 0.0_dp
                  CALL cp_fm_scale_and_add(beta_pr, matrix_G_search, -1.0_dp, matrix_G_old)
               ENDIF
            ELSE ! SD
               CALL cp_fm_scale_and_add(0.0_dp, matrix_G_search, -1.0_dp, matrix_G)
            ENDIF
            ! ds_min=1.0E-4_dp
            line_search_count = 0
         END IF
         line_search_count = line_search_count+1
         energy(line_search_count) = Omega

         ! line search section
         SELECT CASE (3)
         CASE (1) ! two point line search
            SELECT CASE (line_search_count)
            CASE (1)
               pos(1) = 0.0_dp
               pos(2) = ds_min
               CALL cp_fm_trace(matrix_G, matrix_G_search, grad(1))
               grad(1) = grad(1)/2.0_dp
               new_direction = .FALSE.
            CASE (2)
               new_direction = .TRUE.
               x0 = pos(1) ! 0.0_dp
               c = energy(1)
               b = grad(1)
               x1 = pos(2)
               a = (energy(2)-b*x1-c)/(x1**2)
               IF (a .LE. 0.0_dp) a = 1.0E-15_dp
               npos = -b/(2.0_dp*a)
               val = a*npos**2+b*npos+c
               IF (val .LT. energy(1) .AND. val .LE. energy(2)) THEN
                  ! we go to a minimum, but ...
                  ! we take a guard against too large steps
                  pos(3) = MIN(npos, MAXVAL(pos(1:2))*4.0_dp)
               ELSE ! just take an extended step
                  pos(3) = MAXVAL(pos(1:2))*2.0_dp
               ENDIF
            END SELECT
         CASE (2) ! 3 point line search
            SELECT CASE (line_search_count)
            CASE (1)
               new_direction = .FALSE.
               pos(1) = 0.0_dp
               pos(2) = ds_min*0.8_dp
            CASE (2)
               new_direction = .FALSE.
               IF (energy(2) .GT. energy(1)) THEN
                  pos(3) = ds_min*0.7_dp
               ELSE
                  pos(3) = ds_min*1.4_dp
               ENDIF
            CASE (3)
               new_direction = .TRUE.
               xa = pos(1)
               xb = pos(2)
               xc = pos(3)
               fa = energy(1)
               fb = energy(2)
               fc = energy(3)
               nom = (xb-xa)**2*(fb-fc)-(xb-xc)**2*(fb-fa)
               denom = (xb-xa)*(fb-fc)-(xb-xc)*(fb-fa)
               IF (ABS(denom) .LE. 1.0E-18_dp*MAX(ABS(fb-fc), ABS(fb-fa))) THEN
                  npos = xb
               ELSE
                  npos = xb-0.5_dp*nom/denom ! position of the stationary point
               ENDIF
               val = (npos-xa)*(npos-xb)*fc/((xc-xa)*(xc-xb))+ &
                     (npos-xb)*(npos-xc)*fa/((xa-xb)*(xa-xc))+ &
                     (npos-xc)*(npos-xa)*fb/((xb-xc)*(xb-xa))
               IF (val .LT. fa .AND. val .LE. fb .AND. val .LE. fc) THEN ! OK, we go to a minimum
                  ! we take a guard against too large steps
                  pos(4) = MAX(MAXVAL(pos(1:3))*0.01_dp, &
                               MIN(npos, MAXVAL(pos(1:3))*4.0_dp))
               ELSE ! just take an extended step
                  pos(4) = MAXVAL(pos(1:3))*2.0_dp
               ENDIF
            END SELECT
         CASE (3) ! golden section hunt
            new_direction = .FALSE.
            IF (line_search_count .EQ. 1) THEN
               lsl = 1
               lsr = 0
               lsm = 1
               pos(1) = 0.0_dp
               pos(2) = ds_min/gold_sec
            ELSE
               IF (line_search_count .EQ. 150) CALL cp__b("qs_localization_methods.F",1253,"Too many")
               IF (lsr .EQ. 0) THEN
                  IF (energy(line_search_count-1) .LT. energy(line_search_count)) THEN
                     lsr = line_search_count
                     pos(line_search_count+1) = pos(lsm)+(pos(lsr)-pos(lsm))*gold_sec
                  ELSE
                     lsl = lsm
                     lsm = line_search_count
                     pos(line_search_count+1) = pos(line_search_count)/gold_sec
                  ENDIF
               ELSE
                  IF (pos(line_search_count) .LT. pos(lsm)) THEN
                     IF (energy(line_search_count) .LT. energy(lsm)) THEN
                        lsr = lsm
                        lsm = line_search_count
                     ELSE
                        lsl = line_search_count
                     ENDIF
                  ELSE
                     IF (energy(line_search_count) .LT. energy(lsm)) THEN
                        lsl = lsm
                        lsm = line_search_count
                     ELSE
                        lsr = line_search_count
                     ENDIF
                  ENDIF
                  IF (pos(lsr)-pos(lsm) .GT. pos(lsm)-pos(lsl)) THEN
                     pos(line_search_count+1) = pos(lsm)+gold_sec*(pos(lsr)-pos(lsm))
                  ELSE
                     pos(line_search_count+1) = pos(lsl)+gold_sec*(pos(lsm)-pos(lsl))
                  ENDIF
                  IF ((pos(lsr)-pos(lsl)) .LT. 1.0E-3_dp*pos(lsr)) THEN
                     new_direction = .TRUE.
                  ENDIF
               ENDIF ! lsr .eq. 0
            ENDIF ! first step
         END SELECT
         ! now go to the suggested point
         ds_min = pos(line_search_count+1)
         ds = pos(line_search_count+1)-pos(line_search_count)
         CALL cp_fm_scale_and_add(1.0_dp, matrix_A, ds, matrix_G_search)
      ENDDO

      ! collect rotation matrix
      matrix_H%local_data = REAL(cmat_U%local_data, KIND=dp)
      CALL cp_gemm('N', 'N', n, n, n, 1.0_dp, matrix_R, matrix_H, 0.0_dp, matrix_T)
      CALL cp_fm_to_fm(matrix_T, matrix_R)
      CALL rotate_orbitals(matrix_R, vectors)
      CALL cp_fm_release(matrix_R)

      CALL cp_fm_release(matrix_A)
      CALL cp_fm_release(matrix_G)
      CALL cp_fm_release(matrix_H)
      CALL cp_fm_release(matrix_T)
      CALL cp_fm_release(matrix_G_search)
      CALL cp_fm_release(matrix_G_old)
      CALL cp_cfm_release(cmat_A)
      CALL cp_cfm_release(cmat_U)
      CALL cp_cfm_release(cmat_R)
      CALL cp_cfm_release(cmat_t1)
      CALL cp_cfm_release(cmat_t2)
      CALL cp_cfm_release(cmat_B)
      CALL cp_cfm_release(cmat_M)

      DEALLOCATE (evals, evals_exp, fval, fvald)

      DO idim = 1, SIZE(c_zij)
         zij(1, idim)%matrix%local_data = REAL(c_zij(idim)%matrix%local_data, dp)
         zij(2, idim)%matrix%local_data = AIMAG(c_zij(idim)%matrix%local_data)
         CALL cp_cfm_release(c_zij(idim)%matrix)
      ENDDO
      DEALLOCATE (c_zij)
      DEALLOCATE (diag_z)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Parallel algorithm for jacobi rotations
!> \param weights ...
!> \param zij ...
!> \param vectors ...
!> \param para_env ...
!> \param max_iter ...
!> \param eps_localization ...
!> \param sweeps ...
!> \param out_each ...
!> \param target_time ...
!> \param start_time ...
!> \par History
!>      use allgather for improved performance
!> \author MI (11.2009)
! **************************************************************************************************
   SUBROUTINE jacobi_rot_para(weights, zij, vectors, para_env, max_iter, eps_localization, &
                              sweeps, out_each, target_time, start_time)

      REAL(KIND=dp), INTENT(IN)                          :: weights(:)
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: ZIJ(:, :)
      TYPE(cp_fm_type), POINTER                          :: vectors
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: max_iter
      REAL(KIND=dp), INTENT(IN)                          :: eps_localization
      INTEGER                                            :: sweeps
      INTEGER, INTENT(IN)                                :: out_each
      REAL(dp)                                           :: target_time, start_time

      CHARACTER(len=*), PARAMETER :: routineN = 'jacobi_rot_para', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: zi, zj
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: c_array_me, c_array_partner
      COMPLEX(KIND=dp), POINTER                          :: mii(:), mij(:), mjj(:)
      INTEGER :: dim2, handle, i, idim, ii, ik, il1, il2, il_recv, il_recv_partner, ilow1, ilow2, &
         ip, ip_has_i, ip_partner, ip_recv_from, ip_recv_partner, ipair, iperm, istat, istate, &
         iu1, iu2, iup1, iup2, j, jj, jstate, k, kk, n1, n2, nblock, nblock_max, npair, nperm, &
         ns_me, ns_partner, ns_recv_from, ns_recv_partner, nstate, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rcount, rdispl
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: list_pair, ns_bound
      LOGICAL                                            :: should_stop
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: gmat, rmat_loc, rmat_recv, rmat_send, &
                                                            rotmat, z_ij_loc_im, z_ij_loc_re
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: rmat_recv_all
      REAL(KIND=dp)                                      :: ct, func, gmax, grad, ri, rj, st, t1, &
                                                            t2, theta, tolerance, xlow, xstate, &
                                                            xup, zc, zr
      TYPE(cp_fm_type), POINTER                          :: rmat
      TYPE(set_c_1d_type), DIMENSION(:), POINTER         :: zdiag_all, zdiag_me
      TYPE(set_c_2d_type), DIMENSION(:), POINTER         :: cz_ij_loc, xyz_mix, xyz_mix_ns

      CALL timeset(routineN, handle)

      output_unit = cp_logger_get_default_io_unit()

      NULLIFY (rmat, cz_ij_loc, zdiag_all, zdiag_me)
      NULLIFY (xyz_mix, xyz_mix_ns)
      NULLIFY (mii, mij, mjj)

      dim2 = SIZE(zij, 2)
      ALLOCATE (mii(dim2), mij(dim2), mjj(dim2))

      CALL cp_fm_create(rmat, zij(1, 1)%matrix%matrix_struct)
      CALL cp_fm_set_all(rmat, 0._dp, 1._dp)

      CALL cp_fm_get_info(rmat, nrow_global=nstate)

      ALLOCATE (rcount(para_env%num_pe), STAT=istat)
      ALLOCATE (rdispl(para_env%num_pe), STAT=istat)

      tolerance = 1.0e10_dp
      sweeps = 0

      ! number of processor pairs and number of permutations
      npair = (para_env%num_pe+1)/2
      nperm = para_env%num_pe-MOD(para_env%num_pe+1, 2)
      ALLOCATE (list_pair(2, npair))

      ! Distribution of the states (XXXXX safe against more pe than states ??? XXXXX)
      xstate = REAL(nstate, dp)/REAL(para_env%num_pe, dp)
      nblock_max = 0
      ALLOCATE (ns_bound(0:para_env%num_pe-1, 2))
      Xlow = 0.0D0
      Xup = 0.0D0
      DO ip = 1, para_env%num_pe
         xup = xlow+xstate
         ns_bound(ip-1, 1) = NINT(xlow)+1
         ns_bound(ip-1, 2) = NINT(xup)
         IF (NINT(xup) .GT. nstate) THEN
            ns_bound(ip-1, 2) = nstate
         ENDIF
         IF (NINT(xlow) .GT. nstate) THEN
            ns_bound(ip-1, 1) = nstate+1
         ENDIF
         xlow = xup
      ENDDO
      DO ip = 0, para_env%num_pe-1
         nblock = ns_bound(ip, 2)-ns_bound(ip, 1)+1
         nblock_max = MAX(nblock_max, nblock)
      ENDDO

      ! otbtain local part of the matrix (could be made faster, but is likely irrelevant).
      ALLOCATE (z_ij_loc_re(nstate, nblock_max))
      ALLOCATE (z_ij_loc_im(nstate, nblock_max))
      ALLOCATE (cz_ij_loc(dim2))
      DO idim = 1, dim2
         DO ip = 0, para_env%num_pe-1
            nblock = ns_bound(ip, 2)-ns_bound(ip, 1)+1
            CALL cp_fm_get_submatrix(zij(1, idim)%matrix, z_ij_loc_re, 1, ns_bound(ip, 1), nstate, nblock)
            CALL cp_fm_get_submatrix(zij(2, idim)%matrix, z_ij_loc_im, 1, ns_bound(ip, 1), nstate, nblock)
            IF (para_env%mepos == ip) THEN
               ns_me = nblock
               ALLOCATE (cz_ij_loc(idim)%c_array(nstate, ns_me))
               DO i = 1, ns_me
                  DO j = 1, nstate
                     cz_ij_loc(idim)%c_array(j, i) = CMPLX(z_ij_loc_re(j, i), z_ij_loc_im(j, i), dp)
                  END DO
               END DO
            END IF
         END DO ! ip
      END DO
      DEALLOCATE (z_ij_loc_re)
      DEALLOCATE (z_ij_loc_im)

      ! initialize rotation matrix
      ALLOCATE (rotmat(nstate, 2*nblock_max))
      rotmat = 0.0_dp
      DO i = ns_bound(para_env%mepos, 1), ns_bound(para_env%mepos, 2)
         ii = i-ns_bound(para_env%mepos, 1)+1
         rotmat(i, ii) = 1.0_dp
      END DO

      ALLOCATE (xyz_mix(dim2))
      ALLOCATE (xyz_mix_ns(dim2))
      ALLOCATE (zdiag_me(dim2))
      ALLOCATE (zdiag_all(dim2))

      ns_me = ns_bound(para_env%mepos, 2)-ns_bound(para_env%mepos, 1)+1
      IF (ns_me /= 0) THEN
         ALLOCATE (c_array_me(nstate, ns_me, dim2))
         DO idim = 1, dim2
            ALLOCATE (xyz_mix_ns(idim)%c_array(nstate, ns_me))
         END DO
         ALLOCATE (gmat(nstate, ns_me))
      END IF

      DO idim = 1, dim2
         ALLOCATE (zdiag_me(idim)%c_array(nblock_max))
         zdiag_me(idim)%c_array = CMPLX(0.0_dp, 0.0_dp, dp)
         ALLOCATE (zdiag_all(idim)%c_array(para_env%num_pe*nblock_max))
         zdiag_all(idim)%c_array = CMPLX(0.0_dp, 0.0_dp, dp)
      END DO
      ALLOCATE (rmat_recv(nblock_max*2, nblock_max))
      ALLOCATE (rmat_send(nblock_max*2, nblock_max))

      ! buffer for message passing
      ALLOCATE (rmat_recv_all(nblock_max*2, nblock_max, 0:para_env%num_pe-1))

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T4,A )') " Localization by iterative distributed Jacobi rotation"
         WRITE (output_unit, '(T20,A12,T32, A22,T60, A12,A8 )') "Iteration", "Functional", "Tolerance", " Time "
      END IF

      DO sweeps = 1, max_iter+1
         IF (sweeps == max_iter+1) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, *) ' LOCALIZATION! loop did not converge within the maximum number of iterations.'
               WRITE (output_unit, *) '               Present  Max. gradient = ', tolerance
            END IF
            EXIT
         ENDIF
         t1 = m_walltime()

         DO iperm = 1, nperm

            ! fix partners for this permutation, and get the number of states
            CALL eberlein(iperm, para_env, list_pair)
            ip_partner = -1
            ns_partner = 0
            DO ipair = 1, npair
               IF (list_pair(1, ipair) == para_env%mepos) THEN
                  ip_partner = list_pair(2, ipair)
                  EXIT
               ELSE IF (list_pair(2, ipair) == para_env%mepos) THEN
                  ip_partner = list_pair(1, ipair)
                  EXIT
               END IF
            END DO
            IF (ip_partner >= 0) THEN
               ns_partner = ns_bound(ip_partner, 2)-ns_bound(ip_partner, 1)+1
            ELSE
               ns_partner = 0
            END IF

            ! if there is a non-zero block connecting two partners, jacobi-sweep it.
            IF (ns_partner*ns_me /= 0) THEN

               ALLOCATE (rmat_loc(ns_me+ns_partner, ns_me+ns_partner))
               rmat_loc = 0.0_dp
               DO i = 1, ns_me+ns_partner
                  rmat_loc(i, i) = 1.0_dp
               END DO

               ALLOCATE (c_array_partner(nstate, ns_partner, dim2))

               DO idim = 1, dim2
                  ALLOCATE (xyz_mix(idim)%c_array(ns_me+ns_partner, ns_me+ns_partner))
                  DO i = 1, ns_me
                     c_array_me(1:nstate, i, idim) = cz_ij_loc(idim)%c_array(1:nstate, i)
                  END DO
               END DO

               CALL mp_sendrecv(msgin=c_array_me, dest=ip_partner, &
                                msgout=c_array_partner, source=ip_partner, comm=para_env%group)

               n1 = ns_me
               n2 = ns_partner
               ilow1 = ns_bound(para_env%mepos, 1)
               iup1 = ns_bound(para_env%mepos, 1)+n1-1
               ilow2 = ns_bound(ip_partner, 1)
               iup2 = ns_bound(ip_partner, 1)+n2-1
               IF (ns_bound(para_env%mepos, 1) < ns_bound(ip_partner, 1)) THEN
                  il1 = 1
                  iu1 = n1
                  iu1 = n1
                  il2 = 1+n1
                  iu2 = n1+n2
               ELSE
                  il1 = 1+n2
                  iu1 = n1+n2
                  iu1 = n1+n2
                  il2 = 1
                  iu2 = n2
               END IF

               DO idim = 1, dim2
                  DO i = 1, n1
                     xyz_mix(idim)%c_array(il1:iu1, il1+i-1) = c_array_me(ilow1:iup1, i, idim)
                     xyz_mix(idim)%c_array(il2:iu2, il1+i-1) = c_array_me(ilow2:iup2, i, idim)
                  END DO
                  DO i = 1, n2
                     xyz_mix(idim)%c_array(il2:iu2, il2+i-1) = c_array_partner(ilow2:iup2, i, idim)
                     xyz_mix(idim)%c_array(il1:iu1, il2+i-1) = c_array_partner(ilow1:iup1, i, idim)
                  END DO
               END DO

               DO istate = 1, n1+n2
                  DO jstate = istate+1, n1+n2
                     DO idim = 1, dim2
                        mii(idim) = xyz_mix(idim)%c_array(istate, istate)
                        mij(idim) = xyz_mix(idim)%c_array(istate, jstate)
                        mjj(idim) = xyz_mix(idim)%c_array(jstate, jstate)
                     END DO
                     CALL get_angle(mii, mjj, mij, weights, theta)
                     st = SIN(theta)
                     ct = COS(theta)
                     DO idim = 1, dim2
                        DO i = 1, n1+n2
                           zi = ct*xyz_mix(idim)%c_array(i, istate)+st*xyz_mix(idim)%c_array(i, jstate)
                           zj = -st*xyz_mix(idim)%c_array(i, istate)+ct*xyz_mix(idim)%c_array(i, jstate)
                           xyz_mix(idim)%c_array(i, istate) = zi
                           xyz_mix(idim)%c_array(i, jstate) = zj
                        END DO
                        DO i = 1, n1+n2
                           zi = ct*xyz_mix(idim)%c_array(istate, i)+st*xyz_mix(idim)%c_array(jstate, i)
                           zj = -st*xyz_mix(idim)%c_array(istate, i)+ct*xyz_mix(idim)%c_array(jstate, i)
                           xyz_mix(idim)%c_array(istate, i) = zi
                           xyz_mix(idim)%c_array(jstate, i) = zj
                        END DO
                     END DO

                     DO i = 1, n1+n2
                        ri = ct*rmat_loc(i, istate)+st*rmat_loc(i, jstate)
                        rj = ct*rmat_loc(i, jstate)-st*rmat_loc(i, istate)
                        rmat_loc(i, istate) = ri
                        rmat_loc(i, jstate) = rj
                     END DO
                  END DO
               END DO

               k = nblock_max+1
               CALL mp_sendrecv(rotmat(1:nstate, 1:ns_me), ip_partner, &
                                rotmat(1:nstate, k:k+n2-1), ip_partner, para_env%group)

               IF (ilow1 < ilow2) THEN
                  CALL dgemm("N", "N", nstate, n1, n2, 1.0_dp, rotmat(1, k), nstate, rmat_loc(1+n1, 1), n1+n2, 0.0_dp, gmat, nstate)
                  CALL dgemm("N", "N", nstate, n1, n1, 1.0_dp, rotmat(1, 1), nstate, rmat_loc(1, 1), n1+n2, 1.0_dp, gmat, nstate)
               ELSE
                  CALL dgemm("N", "N", nstate, n1, n2, 1.0_dp, rotmat(1, k), nstate, rmat_loc(1, n2+1), n1+n2, 0.0_dp, gmat, nstate)
                  CALL dgemm("N", "N", nstate, n1, n1, 1.0_dp, rotmat(1, 1), nstate, &
                             rmat_loc(n2+1, n2+1), n1+n2, 1.0_dp, gmat, nstate)
               END IF

               CALL dcopy(nstate*n1, gmat(1, 1), 1, rotmat(1, 1), 1)

               DO idim = 1, dim2
                  DO i = 1, n1
                     xyz_mix_ns(idim)%c_array(1:nstate, i) = CMPLX(0.0_dp, 0.0_dp, dp)
                  END DO

                  DO istate = 1, n1
                     DO jstate = 1, nstate
                        DO i = 1, n2
                           xyz_mix_ns(idim)%c_array(jstate, istate) = &
                              xyz_mix_ns(idim)%c_array(jstate, istate)+ &
                              c_array_partner(jstate, i, idim)*rmat_loc(il2+i-1, il1+istate-1)
                        END DO
                     END DO
                  END DO
                  DO istate = 1, n1
                     DO jstate = 1, nstate
                        DO i = 1, n1
                           xyz_mix_ns(idim)%c_array(jstate, istate) = xyz_mix_ns(idim)%c_array(jstate, istate)+ &
                                                                      c_array_me(jstate, i, idim)*rmat_loc(il1+i-1, il1+istate-1)
                        END DO
                     END DO
                  END DO
               END DO ! idim

               DEALLOCATE (c_array_partner)

            ELSE ! save my data
               DO idim = 1, dim2
                  DO i = 1, ns_me
                     xyz_mix_ns(idim)%c_array(1:nstate, i) = cz_ij_loc(idim)%c_array(1:nstate, i)
                  END DO
               END DO
            END IF

            DO idim = 1, dim2
               DO i = 1, ns_me
                  cz_ij_loc(idim)%c_array(1:nstate, i) = CMPLX(0.0_dp, 0.0_dp, dp)
               END DO
            END DO

            IF (ns_partner*ns_me /= 0) THEN
               ! transpose rotation matrix rmat_loc
               DO i = 1, ns_me+ns_partner
                  DO j = i+1, ns_me+ns_partner
                     ri = rmat_loc(i, j)
                     rmat_loc(i, j) = rmat_loc(j, i)
                     rmat_loc(j, i) = ri
                  END DO
               END DO

               ! prepare for distribution
               DO i = 1, n1
                  rmat_send(1:n1, i) = rmat_loc(il1:iu1, il1+i-1)
               END DO
               ik = nblock_max
               DO i = 1, n2
                  rmat_send(ik+1:ik+n1, i) = rmat_loc(il1:iu1, il2+i-1)
               END DO
            ELSE
               rmat_send = 0.0_dp
            END IF

            ! collect data from all tasks (this takes some significant time)
            CALL mp_allgather(rmat_send, rmat_recv_all, para_env%group)

            ! update blocks everywhere
            DO ip = 0, para_env%num_pe-1

               ip_recv_from = MOD(para_env%mepos-IP+para_env%num_pe, para_env%num_pe)
               rmat_recv(:, :) = rmat_recv_all(:, :, ip_recv_from)

               ns_recv_from = ns_bound(ip_recv_from, 2)-ns_bound(ip_recv_from, 1)+1

               IF (ns_me /= 0) THEN
                  IF (ns_recv_from /= 0) THEN
                     !look for the partner of ip_recv_from
                     ip_recv_partner = -1
                     ns_recv_partner = 0
                     DO ipair = 1, npair
                        IF (list_pair(1, ipair) == ip_recv_from) THEN
                           ip_recv_partner = list_pair(2, ipair)
                           EXIT
                        ELSE IF (list_pair(2, ipair) == ip_recv_from) THEN
                           ip_recv_partner = list_pair(1, ipair)
                           EXIT
                        END IF
                     END DO

                     IF (ip_recv_partner >= 0) THEN
                        ns_recv_partner = ns_bound(ip_recv_partner, 2)-ns_bound(ip_recv_partner, 1)+1
                     END IF
                     IF (ns_recv_partner > 0) THEN
                        il1 = ns_bound(para_env%mepos, 1)
                        il_recv = ns_bound(ip_recv_from, 1)
                        il_recv_partner = ns_bound(ip_recv_partner, 1)
                        ik = nblock_max

                        DO idim = 1, dim2
                           DO i = 1, ns_recv_from
                              ii = il_recv+i-1
                              DO j = 1, ns_me
                                 jj = j
                                 DO k = 1, ns_recv_from
                                    kk = il_recv+k-1
                                    cz_ij_loc(idim)%c_array(ii, jj) = cz_ij_loc(idim)%c_array(ii, jj)+ &
                                                                      rmat_recv(i, k)*xyz_mix_ns(idim)%c_array(kk, j)
                                 END DO
                              END DO
                           END DO
                           DO i = 1, ns_recv_from
                              ii = il_recv+i-1
                              DO j = 1, ns_me
                                 jj = j
                                 DO k = 1, ns_recv_partner
                                    kk = il_recv_partner+k-1
                                    cz_ij_loc(idim)%c_array(ii, jj) = cz_ij_loc(idim)%c_array(ii, jj)+ &
                                                                      rmat_recv(ik+i, k)*xyz_mix_ns(idim)%c_array(kk, j)
                                 END DO
                              END DO
                           END DO
                        END DO ! idim
                     ELSE
                        il1 = ns_bound(para_env%mepos, 1)
                        il_recv = ns_bound(ip_recv_from, 1)
                        DO idim = 1, dim2
                           DO j = 1, ns_me
                              jj = j
                              DO i = 1, ns_recv_from
                                 ii = il_recv+i-1
                                 cz_ij_loc(idim)%c_array(ii, jj) = xyz_mix_ns(idim)%c_array(ii, j)
                              END DO
                           END DO
                        END DO ! idim
                     END IF
                  END IF
               END IF ! ns_me
            END DO ! ip

            IF (ns_partner*ns_me /= 0) THEN
               DEALLOCATE (rmat_loc)
               DO idim = 1, dim2
                  DEALLOCATE (xyz_mix(idim)%c_array)
               END DO
            END IF

         END DO ! iperm

         ! calculate the max gradient
         DO idim = 1, dim2
            DO i = ns_bound(para_env%mepos, 1), ns_bound(para_env%mepos, 2)
               ii = i-ns_bound(para_env%mepos, 1)+1
               zdiag_me(idim)%c_array(ii) = cz_ij_loc(idim)%c_array(i, ii)
               zdiag_me(idim)%c_array(ii) = cz_ij_loc(idim)%c_array(i, ii)
            END DO
            rcount(:) = SIZE(zdiag_me(idim)%c_array)
            rdispl(1) = 0
            DO ip = 2, para_env%num_pe
               rdispl(ip) = rdispl(ip-1)+rcount(ip-1)
            ENDDO
            ! collect all the diagonal elements in a replicated 1d array
            CALL mp_allgather(zdiag_me(idim)%c_array, zdiag_all(idim)%c_array, rcount, rdispl, para_env%group)
         END DO

         gmax = 0.0_dp
         DO j = ns_bound(para_env%mepos, 1), ns_bound(para_env%mepos, 2)
            k = j-ns_bound(para_env%mepos, 1)+1
            DO i = 1, j-1
               ! find the location of the diagonal element (i,i)
               DO ip = 0, para_env%num_pe-1
                  IF (i >= ns_bound(ip, 1) .AND. i <= ns_bound(ip, 2)) THEN
                     ip_has_i = ip
                     EXIT
                  END IF
               END DO
               ii = nblock_max*ip_has_i+i-ns_bound(ip_has_i, 1)+1
               ! mepos has the diagonal element (j,j), as well as the off diagonal (i,j)
               jj = nblock_max*para_env%mepos+j-ns_bound(para_env%mepos, 1)+1
               grad = 0.0_dp
               DO idim = 1, dim2
                  zi = zdiag_all(idim)%c_array(ii)
                  zj = zdiag_all(idim)%c_array(jj)
                  grad = grad+weights(idim)*REAL(4.0_dp*CONJG(cz_ij_loc(idim)%c_array(i, k))*(zj-zi), dp)
               END DO
               gmax = MAX(gmax, ABS(grad))
            END DO
         END DO

         CALL mp_max(gmax, para_env%group)
         tolerance = gmax

         func = 0.0_dp
         DO i = ns_bound(para_env%mepos, 1), ns_bound(para_env%mepos, 2)
            k = i-ns_bound(para_env%mepos, 1)+1
            DO idim = 1, dim2
               zr = REAL(cz_ij_loc(idim)%c_array(i, k), dp)
               zc = AIMAG(cz_ij_loc(idim)%c_array(i, k))
               func = func+weights(idim)*(1.0_dp-(zr*zr+zc*zc))/twopi/twopi
            END DO
         END DO
         CALL mp_sum(func, para_env%group)
         t2 = m_walltime()

         IF (output_unit > 0 .AND. MODULO(sweeps, out_each) == 0) THEN
            WRITE (output_unit, '(T20,I12,T35,F20.10,T60,E12.4,F8.3)') sweeps, func, tolerance, t2-t1
            CALL m_flush(output_unit)
         END IF
         IF (tolerance < eps_localization) EXIT
         CALL external_control(should_stop, "LOC", target_time=target_time, start_time=start_time)
         IF (should_stop) EXIT

      END DO ! sweeps

      ! buffer for message passing
      DEALLOCATE (rmat_recv_all)

      DEALLOCATE (rmat_recv)
      DEALLOCATE (rmat_send)
      IF (ns_me > 0) THEN
         DEALLOCATE (c_array_me)
      ENDIF
      DO idim = 1, dim2
         DEALLOCATE (zdiag_me(idim)%c_array)
         DEALLOCATE (zdiag_all(idim)%c_array)
      END DO
      DEALLOCATE (zdiag_me)
      DEALLOCATE (zdiag_all)
      DEALLOCATE (xyz_mix)
      DO idim = 1, dim2
         IF (ns_me /= 0) THEN
            DEALLOCATE (xyz_mix_ns(idim)%c_array)
         ENDIF
      END DO
      DEALLOCATE (xyz_mix_ns)
      IF (ns_me /= 0) THEN
         DEALLOCATE (gmat)
      ENDIF
      DEALLOCATE (mii)
      DEALLOCATE (mij)
      DEALLOCATE (mjj)

      ilow1 = ns_bound(para_env%mepos, 1)
      ns_me = ns_bound(para_env%mepos, 2)-ns_bound(para_env%mepos, 1)+1
      ALLOCATE (z_ij_loc_re(nstate, nblock_max))
      ALLOCATE (z_ij_loc_im(nstate, nblock_max))
      DO idim = 1, dim2
         DO ip = 0, para_env%num_pe-1
            z_ij_loc_re = 0.0_dp
            z_ij_loc_im = 0.0_dp
            nblock = ns_bound(ip, 2)-ns_bound(ip, 1)+1
            IF (ip == para_env%mepos) THEN
               ns_me = nblock
               DO i = 1, ns_me
                  ii = ilow1+i-1
                  DO j = 1, nstate
                     z_ij_loc_re(j, i) = REAL(cz_ij_loc(idim)%c_array(j, i), dp)
                     z_ij_loc_im(j, i) = AIMAG(cz_ij_loc(idim)%c_array(j, i))
                  END DO
               END DO
            END IF
            CALL mp_bcast(z_ij_loc_re, ip, para_env%group)
            CALL mp_bcast(z_ij_loc_im, ip, para_env%group)
            CALL cp_fm_set_submatrix(zij(1, idim)%matrix, z_ij_loc_re, 1, ns_bound(ip, 1), nstate, nblock)
            CALL cp_fm_set_submatrix(zij(2, idim)%matrix, z_ij_loc_im, 1, ns_bound(ip, 1), nstate, nblock)
         END DO ! ip
      END DO

      DO ip = 0, para_env%num_pe-1
         z_ij_loc_re = 0.0_dp
         nblock = ns_bound(ip, 2)-ns_bound(ip, 1)+1
         IF (ip == para_env%mepos) THEN
            ns_me = nblock
            DO i = 1, ns_me
               ii = ilow1+i-1
               DO j = 1, nstate
                  z_ij_loc_re(j, i) = rotmat(j, i)
               END DO
            END DO
         END IF
         CALL mp_bcast(z_ij_loc_re, ip, para_env%group)
         CALL cp_fm_set_submatrix(rmat, z_ij_loc_re, 1, ns_bound(ip, 1), nstate, nblock)
      END DO

      DEALLOCATE (z_ij_loc_re)
      DEALLOCATE (z_ij_loc_im)
      DO idim = 1, dim2
         DEALLOCATE (cz_ij_loc(idim)%c_array)
      END DO
      DEALLOCATE (cz_ij_loc)

      CALL mp_sync(para_env%group)
      CALL rotate_orbitals(rmat, vectors)
      CALL cp_fm_release(rmat)

      DEALLOCATE (rotmat)
      DEALLOCATE (ns_bound, list_pair)

      CALL timestop(handle)

   END SUBROUTINE jacobi_rot_para

! **************************************************************************************************
!> \brief ...
!> \param iperm ...
!> \param para_env ...
!> \param list_pair ...
! **************************************************************************************************
   SUBROUTINE eberlein(iperm, para_env, list_pair)
      INTEGER, INTENT(IN)                                :: iperm
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, DIMENSION(:, :)                           :: list_pair

      CHARACTER(len=*), PARAMETER :: routineN = 'eberlein', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, jj, npair

      npair = (para_env%num_pe+1)/2
      IF (iperm == 1) THEN
!..set up initial ordering
         DO I = 0, para_env%num_pe-1
            II = ((i+1)+1)/2
            JJ = MOD((i+1)+1, 2)+1
            list_pair(JJ, II) = i
         ENDDO
         IF (MOD(para_env%num_pe, 2) == 1) list_pair(2, npair) = -1
      ELSEIF (MOD(iperm, 2) == 0) THEN
!..a type shift
         jj = list_pair(1, npair)
         DO I = npair, 3, -1
            list_pair(1, I) = list_pair(1, I-1)
         ENDDO
         list_pair(1, 2) = list_pair(2, 1)
         list_pair(2, 1) = jj
      ELSE
!..b type shift
         jj = list_pair(2, 1)
         DO I = 1, npair-1
            list_pair(2, I) = list_pair(2, I+1)
         ENDDO
         list_pair(2, npair) = jj
      ENDIF

   END SUBROUTINE eberlein

END MODULE qs_localization_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Some utilities for the construction of
!>      the localization environment
!> \author MI (05-2005)
! **************************************************************************************************
MODULE qs_loc_utils

   USE ai_moments,                      ONLY: contract_cossin,&
                                              cossin
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_element, cp_fm_get_info, cp_fm_get_submatrix, cp_fm_maxabsval, &
        cp_fm_p_type, cp_fm_release, cp_fm_set_all, cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type, &
        cp_fm_write_unformatted
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_constants,                 ONLY: &
        do_loc_crazy, do_loc_direct, do_loc_jacobi, do_loc_l1_norm_sd, do_loc_none, &
        energy_loc_range, op_loc_berry, op_loc_boys, op_loc_pipek, state_loc_all, state_loc_list, &
        state_loc_none, state_loc_range
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_bcast
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_loc_types,                    ONLY: get_qs_loc_env,&
                                              localized_wfn_control_create,&
                                              localized_wfn_control_release,&
                                              localized_wfn_control_type,&
                                              qs_loc_env_new_type,&
                                              set_qs_loc_env
   USE qs_localization_methods,         ONLY: initialize_weights,&
                                              rotate_orbitals
   USE qs_mo_methods,                   ONLY: make_mo_eig
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_scf_types,                    ONLY: ot_method_nr
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_loc_utils'

! *** Public ***
   PUBLIC :: jacobi_rotation_pipek, qs_loc_env_init, loc_write_restart, &
             retain_history, qs_loc_init, &
             set_loc_centers, set_loc_wfn_lists, qs_loc_control_init

CONTAINS

! **************************************************************************************************
!> \brief copy old mos to new ones, allocating as necessary
!> \param mo_loc_history ...
!> \param mo_loc ...
! **************************************************************************************************
   SUBROUTINE retain_history(mo_loc_history, mo_loc)

      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_loc_history, mo_loc

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_history', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ncol_hist, ncol_loc

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(mo_loc_history)) THEN
         ALLOCATE (mo_loc_history(SIZE(mo_loc)))
         DO i = 1, SIZE(mo_loc_history)
            CALL cp_fm_create(mo_loc_history(i)%matrix, mo_loc(i)%matrix%matrix_struct)
         ENDDO
      ENDIF

      DO i = 1, SIZE(mo_loc_history)
         CALL cp_fm_get_info(mo_loc_history(i)%matrix, ncol_global=ncol_hist)
         CALL cp_fm_get_info(mo_loc(i)%matrix, ncol_global=ncol_loc)
         IF(.NOT.(ncol_hist == ncol_loc))CALL cp__a("qs_loc_utils.F",133)
         CALL cp_fm_to_fm(mo_loc(i)%matrix, mo_loc_history(i)%matrix)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE retain_history

! **************************************************************************************************
!> \brief rotate the mo_new, so that the orbitals are as similar
!>        as possible to ones in mo_ref.
!> \param mo_new ...
!> \param mo_ref ...
!> \param matrix_S ...
! **************************************************************************************************
   SUBROUTINE rotate_state_to_ref(mo_new, mo_ref, matrix_S)

      TYPE(cp_fm_type), POINTER                          :: mo_new, mo_ref
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_S

      CHARACTER(len=*), PARAMETER :: routineN = 'rotate_state_to_ref', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ncol, ncol_ref, nrow
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: o1, o2, o3, o4, smo

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(mo_new, nrow_global=nrow, ncol_global=ncol)
      CALL cp_fm_get_info(mo_ref, ncol_global=ncol_ref)
      IF(.NOT.(ncol == ncol_ref))CALL cp__a("qs_loc_utils.F",165)

      NULLIFY (fm_struct_tmp, o1, o2, o3, o4, smo)
      CALL cp_fm_create(smo, mo_ref%matrix_struct)

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ncol, &
                               ncol_global=ncol, para_env=mo_new%matrix_struct%para_env, &
                               context=mo_new%matrix_struct%context)
      CALL cp_fm_create(o1, fm_struct_tmp)
      CALL cp_fm_create(o2, fm_struct_tmp)
      CALL cp_fm_create(o3, fm_struct_tmp)
      CALL cp_fm_create(o4, fm_struct_tmp)
      CALL cp_fm_struct_release(fm_struct_tmp)

      ! o1 = (mo_new)^T matrix_S mo_ref
      CALL cp_dbcsr_sm_fm_multiply(matrix_S, mo_ref, smo, ncol)
      CALL cp_gemm('T', 'N', ncol, ncol, nrow, 1.0_dp, mo_new, smo, 0.0_dp, o1)

      ! o2 = (o1^T o1)
      CALL cp_gemm('T', 'N', ncol, ncol, ncol, 1.0_dp, o1, o1, 0.0_dp, o2)

      ! o2 = (o1^T o1)^-1/2
      ALLOCATE (eigenvalues(ncol))
      CALL choose_eigv_solver(o2, o3, eigenvalues)
      CALL cp_fm_to_fm(o3, o4)
      eigenvalues(:) = 1.0_dp/SQRT(eigenvalues(:))
      CALL cp_fm_column_scale(o4, eigenvalues)
      CALL cp_gemm('N', 'T', ncol, ncol, ncol, 1.0_dp, o3, o4, 0.0_dp, o2)

      ! o3 = o1 (o1^T o1)^-1/2
      CALL cp_gemm('N', 'N', ncol, ncol, ncol, 1.0_dp, o1, o2, 0.0_dp, o3)

      ! mo_new o1 (o1^T o1)^-1/2
      CALL cp_gemm('N', 'N', nrow, ncol, ncol, 1.0_dp, mo_new, o3, 0.0_dp, smo)
      CALL cp_fm_to_fm(smo, mo_new)

      ! XXXXXXX testing
      ! CALL cp_gemm('N','T',ncol,ncol,ncol,1.0_dp,o3,o3,0.0_dp,o1)
      ! WRITE(*,*) o1%local_data
      ! CALL cp_gemm('T','N',ncol,ncol,ncol,1.0_dp,o3,o3,0.0_dp,o1)
      ! WRITE(*,*) o1%local_data

      CALL cp_fm_release(o1)
      CALL cp_fm_release(o2)
      CALL cp_fm_release(o3)
      CALL cp_fm_release(o4)
      CALL cp_fm_release(smo)

      CALL timestop(handle)

   END SUBROUTINE rotate_state_to_ref

! **************************************************************************************************
!> \brief 2by2 rotation for the pipek operator
!>       in this case the z_ij numbers are reals
!> \param zij_fm_set ...
!> \param vectors ...
!> \param sweeps ...
!> \par History
!>       05-2005 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE jacobi_rotation_pipek(zij_fm_set, vectors, sweeps)

      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: zij_fm_set
      TYPE(cp_fm_type), POINTER                          :: vectors
      INTEGER                                            :: sweeps

      CHARACTER(len=*), PARAMETER :: routineN = 'jacobi_rotation_pipek', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, istate, jstate, natom, nstate
      REAL(KIND=dp)                                      :: aij, bij, ct, mii, mij, mjj, ratio, st, &
                                                            theta, tolerance
      TYPE(cp_fm_type), POINTER                          :: rmat

      NULLIFY (rmat)

      CALL cp_fm_create(rmat, zij_fm_set(1, 1)%matrix%matrix_struct)
      CALL cp_fm_set_all(rmat, 0.0_dp, 1.0_dp)

      CALL cp_fm_get_info(rmat, nrow_global=nstate)
      tolerance = 1.0e10_dp
      sweeps = 0
      natom = SIZE(zij_fm_set, 1)
      ! do jacobi sweeps until converged
      DO WHILE (tolerance >= 1.0e-4_dp)
         sweeps = sweeps+1
         DO istate = 1, nstate
            DO jstate = istate+1, nstate
               aij = 0.0_dp
               bij = 0.0_dp
               DO iatom = 1, natom
                  CALL cp_fm_get_element(zij_fm_set(iatom, 1)%matrix, istate, istate, mii)
                  CALL cp_fm_get_element(zij_fm_set(iatom, 1)%matrix, istate, jstate, mij)
                  CALL cp_fm_get_element(zij_fm_set(iatom, 1)%matrix, jstate, jstate, mjj)
                  aij = aij+mij*(mii-mjj)
                  bij = bij+mij*mij-0.25_dp*(mii-mjj)*(mii-mjj)

               END DO
               IF (ABS(bij) > 1.E-10_dp) THEN
                  ratio = -aij/bij
                  theta = 0.25_dp*ATAN(ratio)
               ELSE
                  bij = 0.0_dp
                  theta = 0.0_dp
               END IF
               ! Check max or min
               ! To minimize the spread
               IF (theta > pi*0.5_dp) THEN
                  theta = theta-pi*0.25_dp
               ELSE IF (theta < -pi*0.5_dp) THEN
                  theta = theta+pi*0.25_dp
               END IF

               ct = COS(theta)
               st = SIN(theta)
               CALL rotate_rmat_real(istate, jstate, st, ct, rmat)
               CALL rotate_zij_real(istate, jstate, st, ct, zij_fm_set)
            END DO
         END DO
         CALL check_tolerance_real(zij_fm_set, tolerance)
      END DO

      CALL rotate_orbitals(rmat, vectors)
      CALL cp_fm_release(rmat)

   END SUBROUTINE jacobi_rotation_pipek

! **************************************************************************************************
!> \brief ...
!> \param istate ...
!> \param jstate ...
!> \param st ...
!> \param ct ...
!> \param rmat ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE rotate_rmat_real(istate, jstate, st, ct, rmat)

      INTEGER, INTENT(IN)                      :: istate, jstate
      REAL(dp), INTENT(IN)                     :: st, ct
      TYPE(cp_fm_type), POINTER                :: rmat

      INTEGER                                  :: nstate

      CALL cp_fm_get_info(rmat, nrow_global=nstate)
      CALL drot(nstate, rmat%local_data(1, istate), 1, &
                rmat%local_data(1, jstate), 1, ct, st)

   END SUBROUTINE rotate_rmat_real

! **************************************************************************************************
!> \brief ...
!> \param istate ...
!> \param jstate ...
!> \param st ...
!> \param ct ...
!> \param zij ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE rotate_zij_real(istate, jstate, st, ct, zij)

      INTEGER, INTENT(IN)                      :: istate, jstate
      REAL(dp), INTENT(IN)                     :: st, ct
      TYPE(cp_fm_p_type), DIMENSION(:, :), &
         POINTER                                :: zij

      INTEGER                                  :: iatom, natom, nstate
      INTEGER                                  :: stride

      CALL cp_fm_get_info(zij(1, 1)%matrix, nrow_global=nstate)

      natom = SIZE(zij, 1)
      DO iatom = 1, natom
         CALL drot(nstate, zij(iatom, 1)%matrix%local_data(1, istate), &
                   1, zij(iatom, 1)%matrix%local_data(1, jstate), 1, ct, st)
         stride = SIZE(zij(iatom, 1)%matrix%local_data, 1)
         CALL drot(nstate, zij(iatom, 1)%matrix%local_data(istate, 1), &
                   stride, zij(iatom, 1)%matrix%local_data(jstate, 1), stride, ct, st)
      END DO

   END SUBROUTINE rotate_zij_real

! **************************************************************************************************
!> \brief ...
!> \param zij_fm_set ...
!> \param tolerance ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE check_tolerance_real(zij_fm_set, tolerance)

      TYPE(cp_fm_p_type), DIMENSION(:, :)                :: zij_fm_set
      REAL(dp), INTENT(OUT)                              :: tolerance

      CHARACTER(len=*), PARAMETER :: routineN = 'check_tolerance_real', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, istate, jstate, natom, &
                                                            ncol_local, nrow_global, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(dp)                                           :: grad_ij, zii, zij, zjj
      REAL(dp), DIMENSION(:, :), POINTER                 :: diag
      TYPE(cp_fm_type), POINTER                          :: force

      NULLIFY (force, diag, col_indices, row_indices)
      natom = SIZE(zij_fm_set, 1)
      CALL cp_fm_get_info(zij_fm_set(1, 1)%matrix, nrow_local=nrow_local, &
                          ncol_local=ncol_local, nrow_global=nrow_global, &
                          row_indices=row_indices, col_indices=col_indices)
      ALLOCATE (diag(nrow_global, natom))

      CALL cp_fm_create(force, zij_fm_set(1, 1)%matrix%matrix_struct)
      CALL cp_fm_set_all(force, 0._dp)

      DO iatom = 1, natom
         DO istate = 1, nrow_global
            CALL cp_fm_get_element(zij_fm_set(iatom, 1)%matrix, istate, istate, diag(istate, iatom))
         END DO
      END DO

      DO istate = 1, nrow_local
         DO jstate = 1, ncol_local
            grad_ij = 0.0_dp
            DO iatom = 1, natom
               zii = diag(istate, iatom)
               zjj = diag(jstate, iatom)
               zij = zij_fm_set(iatom, 1)%matrix%local_data(istate, jstate)
               grad_ij = grad_ij+4.0_dp*zij*(zjj-zii)
            END DO
            force%local_data(istate, jstate) = grad_ij
         END DO
      END DO

      DEALLOCATE (diag)

      CALL cp_fm_maxabsval(force, tolerance)

      CALL cp_fm_release(force)

   END SUBROUTINE check_tolerance_real

! **************************************************************************************************
!> \brief allocates the data, and initializes the operators
!> \param qs_loc_env new environment for the localization calculations
!> \param localized_wfn_control variables and directives for the localization
!> \param qs_env the qs_env in which the qs_env lives
!> \param myspin ...
!> \param do_localize ...
!> \param loc_coeff ...
!> \param mo_loc_history ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
!> \note
!>      similar to the old one, but not quite
! **************************************************************************************************
   SUBROUTINE qs_loc_env_init(qs_loc_env, localized_wfn_control, qs_env, myspin, do_localize, &
                              loc_coeff, mo_loc_history)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN), OPTIONAL                      :: myspin
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_localize
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: loc_coeff, mo_loc_history

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_env_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dim_op, handle, i, iatom, imo, imoloc, &
                                                            ispin, j, l_spin, lb, my_nspins, nao, &
                                                            naosub, natoms, nmo, nmosub, nspins, &
                                                            s_spin, ub
      REAL(KIND=dp)                                      :: my_occ, occ_imo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: occupations
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: moloc_coeff
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mat_ptr, mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)

      NULLIFY (mos, matrix_s, moloc_coeff, particle_set, para_env, cell, local_molecules, occupations, mat_ptr)
      IF (PRESENT(do_localize)) qs_loc_env%do_localize = do_localize
      IF (qs_loc_env%do_localize) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s, cell=cell, &
                         local_molecules=local_molecules, particle_set=particle_set, &
                         para_env=para_env, mos=mos)
         nspins = SIZE(mos, 1)
         s_spin = 1
         l_spin = nspins
         my_nspins = nspins
         IF (PRESENT(myspin)) THEN
            s_spin = myspin
            l_spin = myspin
            my_nspins = 1
         END IF
         ALLOCATE (moloc_coeff(my_nspins))
         DO ispin = s_spin, l_spin
            NULLIFY (tmp_fm_struct, mo_coeff)
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
            nmosub = localized_wfn_control%nloc_states(ispin)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                     ncol_global=nmosub, para_env=para_env, context=mo_coeff%matrix_struct%context)
            CALL cp_fm_create(moloc_coeff(ispin)%matrix, tmp_fm_struct)

            CALL cp_fm_get_info(moloc_coeff(ispin)%matrix, nrow_global=naosub, &
                                ncol_global=nmosub)
            IF(.NOT.(nao == naosub))CALL cp__a("qs_loc_utils.F",510)
            IF (localized_wfn_control%do_homo) THEN
               IF(.NOT.(nmo >= nmosub))CALL cp__a("qs_loc_utils.F",512)
            ELSE
               IF(.NOT.(nao-nmo >= nmosub))CALL cp__a("qs_loc_utils.F",514)
            END IF
            CALL cp_fm_set_all(moloc_coeff(ispin)%matrix, 0.0_dp)
            CALL cp_fm_struct_release(tmp_fm_struct)
         END DO ! ispin
         ! Copy the submatrix

         DO ispin = s_spin, l_spin
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            occupation_numbers=occupations, nao=nao, nmo=nmo)
            lb = localized_wfn_control%lu_bound_states(1, ispin)
            ub = localized_wfn_control%lu_bound_states(2, ispin)

            IF (PRESENT(loc_coeff)) THEN
               mat_ptr => loc_coeff(ispin)%matrix
            ELSE
               mat_ptr => mo_coeff
            END IF
            IF (localized_wfn_control%set_of_states == state_loc_list) THEN
               ALLOCATE (vecbuffer(1, nao))
               IF (localized_wfn_control%do_homo) THEN
                  my_occ = occupations(localized_wfn_control%loc_states(1, ispin))
               END IF
               nmosub = SIZE(localized_wfn_control%loc_states, 1)
               IF(.NOT.(nmosub > 0))CALL cp__a("qs_loc_utils.F",538)
               imoloc = 0
               DO i = lb, ub
                  ! Get the index in the subset
                  imoloc = imoloc+1
                  ! Get the index in the full set
                  imo = localized_wfn_control%loc_states(i, ispin)
                  IF (localized_wfn_control%do_homo) THEN
                     occ_imo = occupations(imo)
                     IF (ABS(occ_imo-my_occ) > localized_wfn_control%eps_occ) THEN
                        IF (localized_wfn_control%localization_method /= do_loc_none) &
                           CALL cp_abort(cp__l("qs_loc_utils.F",549), &
                                         "States with different occupations "// &
                                         "cannot be rotated together")
                     END IF
                  END IF
                  ! Take the imo vector from the full set and copy in the imoloc vector of the subset
                  CALL cp_fm_get_submatrix(mat_ptr, vecbuffer, 1, imo, &
                                           nao, 1, transpose=.TRUE.)
                  CALL cp_fm_set_submatrix(moloc_coeff(ispin)%matrix, vecbuffer, 1, imoloc, &
                                           nao, 1, transpose=.TRUE.)
               END DO
               DEALLOCATE (vecbuffer)
            ELSE
               my_occ = occupations(lb)
               occ_imo = occupations(ub)
               IF (ABS(occ_imo-my_occ) > localized_wfn_control%eps_occ) THEN
                  IF (localized_wfn_control%localization_method /= do_loc_none) &
                     CALL cp_abort(cp__l("qs_loc_utils.F",566), &
                                   "States with different occupations "// &
                                   "cannot be rotated together")
               END IF
               nmosub = localized_wfn_control%nloc_states(ispin)
               CALL cp_fm_to_fm(mat_ptr, moloc_coeff(ispin)%matrix, nmosub, lb, 1)
            END IF

            ! we have the mo's to be localized now, see if we can rotate them according to the history
            ! only do that if we have a history of course. The history is filled
            IF (PRESENT(mo_loc_history)) THEN
               IF (localized_wfn_control%use_history .AND. ASSOCIATED(mo_loc_history)) THEN
                  CALL rotate_state_to_ref(moloc_coeff(ispin)%matrix, &
                                           mo_loc_history(ispin)%matrix, matrix_s(1)%matrix)
               ENDIF
            ENDIF

         END DO

         CALL set_qs_loc_env(qs_loc_env=qs_loc_env, cell=cell, local_molecules=local_molecules, &
                             moloc_coeff=moloc_coeff, particle_set=particle_set, para_env=para_env, &
                             localized_wfn_control=localized_wfn_control)

         ! Prepare the operators
         NULLIFY (tmp_fm_struct, mo_coeff)
         nmosub = MAXVAL(localized_wfn_control%nloc_states)
         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmosub, &
                                  ncol_global=nmosub, para_env=para_env, context=mo_coeff%matrix_struct%context)

         IF (localized_wfn_control%operator_type == op_loc_berry) THEN
            IF (qs_loc_env%cell%orthorhombic) THEN
               dim_op = 3
            ELSE
               dim_op = 6
            END IF
            CALL set_qs_loc_env(qs_loc_env=qs_loc_env, dim_op=dim_op)
            ALLOCATE (qs_loc_env%op_sm_set(2, dim_op))
            DO i = 1, dim_op
               DO j = 1, SIZE(qs_loc_env%op_sm_set, 1)
                  NULLIFY (qs_loc_env%op_sm_set(j, i)%matrix)
                  ALLOCATE (qs_loc_env%op_sm_set(j, i)%matrix)
                  CALL cp_dbcsr_init(qs_loc_env%op_sm_set(j, i)%matrix)
                  CALL cp_dbcsr_copy(qs_loc_env%op_sm_set(j, i)%matrix, matrix_s(1)%matrix, &
                                     name="qs_loc_env%op_sm_"//TRIM(ADJUSTL(cp_to_string(j)))//"-"//TRIM(ADJUSTL(cp_to_string(i))))
                  CALL cp_dbcsr_set(qs_loc_env%op_sm_set(j, i)%matrix, 0.0_dp)
               ENDDO
            END DO

         ELSEIF (localized_wfn_control%operator_type == op_loc_pipek) THEN
            natoms = SIZE(qs_loc_env%particle_set, 1)
            ALLOCATE (qs_loc_env%op_fm_set(natoms, 1))
            CALL set_qs_loc_env(qs_loc_env=qs_loc_env, dim_op=natoms)
            DO ispin = 1, SIZE(qs_loc_env%op_fm_set, 2)
               CALL get_mo_set(mos(ispin)%mo_set, nmo=nmo)
               DO iatom = 1, natoms
                  NULLIFY (qs_loc_env%op_fm_set(iatom, ispin)%matrix)

                  CALL cp_fm_create(qs_loc_env%op_fm_set(iatom, ispin)%matrix, tmp_fm_struct)

                  CALL cp_fm_get_info(qs_loc_env%op_fm_set(iatom, ispin)%matrix, nrow_global=nmosub)
                  IF(.NOT.(nmo >= nmosub))CALL cp__a("qs_loc_utils.F",627)
                  CALL cp_fm_set_all(qs_loc_env%op_fm_set(iatom, ispin)%matrix, 0.0_dp)
               END DO ! iatom
            END DO ! ispin
         ELSE
            CALL cp__b("qs_loc_utils.F",632,"Type of operator not implemented")
         ENDIF
         CALL cp_fm_struct_release(tmp_fm_struct)

         IF (localized_wfn_control%operator_type == op_loc_berry) THEN

            CALL initialize_weights(qs_loc_env%cell, qs_loc_env%weights)

            CALL get_berry_operator(qs_loc_env, qs_env)

         ELSEIF (localized_wfn_control%operator_type == op_loc_pipek) THEN

            !!    here we don't have to do anything
            !!    CALL get_pipek_mezey_operator ( qs_loc_env, qs_env )

         ENDIF

         qs_loc_env%molecular_states = .FALSE.
         qs_loc_env%wannier_states = .FALSE.
      END IF
      CALL timestop(handle)

   END SUBROUTINE qs_loc_env_init

! **************************************************************************************************
!> \brief Computes the Berry operator for periodic systems
!>       used to define the spread of the MOS
!>       Here the matrix elements of the type <mu|cos(kr)|nu> and  <mu|sin(kr)|nu>
!>       are computed, where mu and nu are the contracted basis functions.
!>       Namely the Berry operator is exp(ikr)
!>       k is defined somewhere
!>       the pair lists are exploited and sparse matrixes are constructed
!> \param qs_loc_env new environment for the localization calculations
!> \param qs_env the qs_env in which the qs_env lives
!> \par History
!>      04.2005 created [MI]
!> \author MI
!> \note
!>      The intgrals are computed analytically  using the primitives GTO
!>      The contraction is performed block-wise
! **************************************************************************************************
   SUBROUTINE get_berry_operator(qs_loc_env, qs_env)
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_berry_operator', &
         routineP = moduleN//':'//routineN

      INTEGER :: dim_op, handle, i, iatom, icol, ikind, inode, irow, iset, jatom, jkind, jset, &
         last_jatom, ldab, ldsa, ldsb, ldwork, maxl, ncoa, ncob, nkind, nrow, nseta, nsetb, sgfa, &
         sgfb
      INTEGER, DIMENSION(3)                              :: perd0
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b
      REAL(KIND=dp)                                      :: dab, kvec(3), rab2, vector_k(3, 6)
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rb
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cosab, rpgfa, rpgfb, sinab, sphi_a, &
                                                            sphi_b, work, zeta, zetb
      TYPE(block_p_type), DIMENSION(:), POINTER          :: op_cos, op_sin
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_sm_set
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind, qs_kind_set)
      NULLIFY (cell, op_sm_set, particle_set)
      NULLIFY (sab_orb)
      NULLIFY (cosab, sinab, work)
      NULLIFY (la_max, la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb)
      NULLIFY (set_radius_a, set_radius_b, rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, sab_orb=sab_orb)

      nkind = SIZE(qs_kind_set)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=ldwork, maxlgto=maxl)
      ldab = ldwork
      ALLOCATE (cosab(ldab, ldab))
      cosab = 0.0_dp
      ALLOCATE (sinab(ldab, ldab))
      sinab = 0.0_dp
      ALLOCATE (work(ldwork, ldwork))
      work = 0.0_dp

      CALL get_qs_loc_env(qs_loc_env=qs_loc_env, op_sm_set=op_sm_set, &
                          cell=cell, dim_op=dim_op)

      ALLOCATE (op_cos(dim_op))
      ALLOCATE (op_sin(dim_op))
      DO i = 1, dim_op
         NULLIFY (op_cos(i)%block)
         NULLIFY (op_sin(i)%block)
      END DO

      kvec = 0.0_dp
      vector_k = 0.0_dp
      vector_k(:, 1) = twopi*cell%h_inv(1, :)
      vector_k(:, 2) = twopi*cell%h_inv(2, :)
      vector_k(:, 3) = twopi*cell%h_inv(3, :)
      vector_k(:, 4) = twopi*(cell%h_inv(1, :)+cell%h_inv(2, :))
      vector_k(:, 5) = twopi*(cell%h_inv(1, :)+cell%h_inv(3, :))
      vector_k(:, 6) = twopi*(cell%h_inv(2, :)+cell%h_inv(3, :))

      ! This operator can be used only for periodic systems
      ! If an isolated system is used the periodicity is overimposed
      perd0(1:3) = cell%perd(1:3)
      cell%perd(1:3) = 1

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         ldsa = SIZE(sphi_a, 1)
         ldsb = SIZE(sphi_b, 1)
         IF (inode == 1) last_jatom = 0

         rb = rab+ra

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            DO i = 1, dim_op
               NULLIFY (op_cos(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=op_sm_set(1, i)%matrix, &
                                         row=irow, col=icol, block=op_cos(i)%block, found=found)
               NULLIFY (op_sin(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=op_sm_set(2, i)%matrix, &
                                         row=irow, col=icol, block=op_sin(i)%block, found=found)
            END DO
         END IF ! new_atom_b

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         nrow = 0
         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (set_radius_a(iset)+set_radius_b(jset) >= dab) THEN

!           *** Calculate the primitive overlap integrals ***
                  DO i = 1, dim_op

                     kvec(1:3) = vector_k(1:3, i)
                     cosab = 0.0_dp
                     sinab = 0.0_dp
                     CALL cossin(la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), &
                                 la_min(iset), lb_max(jset), npgfb(jset), zetb(:, jset), &
                                 rpgfb(:, jset), lb_min(jset), &
                                 ra, rb, kvec, cosab, sinab)

                     CALL contract_cossin(op_cos(i)%block, op_sin(i)%block, &
                                          iatom, ncoa, nsgfa(iset), sgfa, sphi_a, ldsa, &
                                          jatom, ncob, nsgfb(jset), sgfb, sphi_b, ldsb, &
                                          cosab, sinab, ldab, work, ldwork)
                  END DO

               END IF !  >= dab

            END DO ! jset

            nrow = nrow+ncoa

         END DO ! iset

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! Set back the correct periodicity
      cell%perd(1:3) = perd0(1:3)

      DO i = 1, dim_op
         NULLIFY (op_cos(i)%block)
         NULLIFY (op_sin(i)%block)
      END DO
      DEALLOCATE (op_cos, op_sin)

      DEALLOCATE (cosab, sinab, work, basis_set_list)

      CALL timestop(handle)
   END SUBROUTINE get_berry_operator

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \param section ...
!> \param mo_array ...
!> \param coeff_localized ...
!> \param do_homo ...
!> \param evals ...
! **************************************************************************************************
   SUBROUTINE loc_write_restart(qs_loc_env, section, mo_array, coeff_localized, &
                                do_homo, evals)
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: section
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: coeff_localized
      LOGICAL, INTENT(IN)                                :: do_homo
      TYPE(cp_1d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: evals

      CHARACTER(LEN=*), PARAMETER :: routineN = 'loc_write_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_middle
      INTEGER                                            :: handle, ispin, max_block, nao, nloc, &
                                                            nmo, output_unit, rst_unit
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF (qs_loc_env%do_localize) THEN

         print_key => section_vals_get_subs_vals(section, "LOC_RESTART")
         IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                              section, "LOC_RESTART"), &
                   cp_p_file)) THEN

            ! Open file
            rst_unit = -1

            IF (do_homo) THEN
               my_middle = "LOC_HOMO"
            ELSE
               my_middle = "LOC_LUMO"
            END IF

            rst_unit = cp_print_key_unit_nr(logger, section, "LOC_RESTART", &
                                            extension=".wfn", file_status="REPLACE", file_action="WRITE", &
                                            file_form="UNFORMATTED", middle_name=TRIM(my_middle))

            IF (rst_unit > 0) filename = cp_print_key_generate_filename(logger, print_key, &
                                                                        middle_name=TRIM(my_middle), extension=".wfn", &
                                                                        my_local=.FALSE.)

            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,T2,A, A/)") &
                  "LOCALIZATION| Write restart file for the localized MOS : ", &
                  TRIM(filename)
            END IF

            IF (rst_unit > 0) THEN
               WRITE (rst_unit) qs_loc_env%localized_wfn_control%set_of_states
               WRITE (rst_unit) qs_loc_env%localized_wfn_control%lu_bound_states
               WRITE (rst_unit) qs_loc_env%localized_wfn_control%nloc_states
            END IF

            DO ispin = 1, SIZE(coeff_localized)
               mo_coeff => coeff_localized(ispin)%matrix
               CALL cp_fm_get_info(mo_coeff, nrow_global=nao, ncol_global=nmo, ncol_block=max_block)
               nloc = qs_loc_env%localized_wfn_control%nloc_states(ispin)
               IF (rst_unit > 0) THEN
                  WRITE (rst_unit) qs_loc_env%localized_wfn_control%loc_states(1:nloc, ispin)
                  IF (do_homo) THEN
                     WRITE (rst_unit) nmo, &
                        mo_array(ispin)%mo_set%homo, &
                        mo_array(ispin)%mo_set%lfomo, &
                        mo_array(ispin)%mo_set%nelectron
                     WRITE (rst_unit) mo_array(ispin)%mo_set%eigenvalues(1:nmo), &
                        mo_array(ispin)%mo_set%occupation_numbers(1:nmo)
                  ELSE
                     WRITE (rst_unit) nmo
                     WRITE (rst_unit) evals(ispin)%array(1:nmo)
                  END IF
               END IF

               CALL cp_fm_write_unformatted(mo_coeff, rst_unit)

            END DO

            CALL cp_print_key_finished_output(rst_unit, logger, section, &
                                              "LOC_RESTART")
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE loc_write_restart

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \param mos ...
!> \param mos_localized ...
!> \param section ...
!> \param section2 ...
!> \param para_env ...
!> \param do_homo ...
!> \param restart_found ...
!> \param evals ...
! **************************************************************************************************
   SUBROUTINE loc_read_restart(qs_loc_env, mos, mos_localized, section, section2, para_env, &
                               do_homo, restart_found, evals)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_localized
      TYPE(section_vals_type), POINTER                   :: section, section2
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: do_homo
      LOGICAL, INTENT(INOUT)                             :: restart_found
      TYPE(cp_1d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: evals

      CHARACTER(len=*), PARAMETER :: routineN = 'loc_read_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=25)                                  :: fname_key
      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_middle
      INTEGER :: group, handle, homo_read, i, ispin, lfomo_read, max_nloc, n_rep_val, nao, &
         nelectron_read, nloc, nmo, nmo_read, nspin, output_unit, rst_unit, source
      LOGICAL                                            :: file_exists
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eig_read, occ_read
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      nspin = SIZE(mos_localized)
      nao = mos(1)%mo_set%nao
      rst_unit = -1

      group = para_env%group
      source = para_env%source
      output_unit = cp_print_key_unit_nr(logger, section2, &
                                         "PROGRAM_RUN_INFO", extension=".Log")

      IF (do_homo) THEN
         fname_key = "LOCHOMO_RESTART_FILE_NAME"
      ELSE
         fname_key = "LOCLUMO_RESTART_FILE_NAME"
         IF (.NOT. PRESENT(evals)) &
            CALL cp__b("qs_loc_utils.F",1047,"Missing argument to localize unoccupied states.")
      END IF

      file_exists = .FALSE.
      CALL section_vals_val_get(section, fname_key, n_rep_val=n_rep_val)
      IF (n_rep_val > 0) THEN
         CALL section_vals_val_get(section, fname_key, c_val=filename)
      ELSE

         print_key => section_vals_get_subs_vals(section2, "LOC_RESTART")
         IF (do_homo) THEN
            my_middle = "LOC_HOMO"
         ELSE
            my_middle = "LOC_LUMO"
         END IF
         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   middle_name=TRIM(my_middle), extension=".wfn", &
                                                   my_local=.FALSE.)
      END IF

      IF (para_env%ionode) INQUIRE (FILE=filename, exist=file_exists)

      IF (file_exists) THEN
         IF (para_env%ionode) THEN
            CALL open_file(file_name=filename, &
                           file_action="READ", &
                           file_form="UNFORMATTED", &
                           file_status="OLD", &
                           unit_number=rst_unit)

            READ (rst_unit) qs_loc_env%localized_wfn_control%set_of_states
            READ (rst_unit) qs_loc_env%localized_wfn_control%lu_bound_states
            READ (rst_unit) qs_loc_env%localized_wfn_control%nloc_states
         END IF
      ELSE
         IF (output_unit > 0) WRITE (output_unit, "(/,T10,A)") &
            "Restart file not available filename=<"//TRIM(filename)//'>'
      END IF
      CALL mp_bcast(file_exists, source, group)

      IF (file_exists) THEN
         restart_found = .TRUE.

         CALL mp_bcast(qs_loc_env%localized_wfn_control%set_of_states, source, group)
         CALL mp_bcast(qs_loc_env%localized_wfn_control%lu_bound_states, source, group)
         CALL mp_bcast(qs_loc_env%localized_wfn_control%nloc_states, source, group)

         max_nloc = MAXVAL(qs_loc_env%localized_wfn_control%nloc_states(:))

         ALLOCATE (vecbuffer(1, nao))
         IF (ASSOCIATED(qs_loc_env%localized_wfn_control%loc_states)) THEN
            DEALLOCATE (qs_loc_env%localized_wfn_control%loc_states)
         END IF
         ALLOCATE (qs_loc_env%localized_wfn_control%loc_states(max_nloc, 2))
         qs_loc_env%localized_wfn_control%loc_states = 0

         DO ispin = 1, nspin
            IF (do_homo) THEN
               nmo = mos(ispin)%mo_set%nmo
            ELSE
               nmo = SIZE(evals(ispin)%array, 1)
            END IF
            IF (para_env%ionode .AND. (nmo > 0)) THEN
               nloc = qs_loc_env%localized_wfn_control%nloc_states(ispin)
               READ (rst_unit) qs_loc_env%localized_wfn_control%loc_states(1:nloc, ispin)
               IF (do_homo) THEN
                  READ (rst_unit) nmo_read, homo_read, lfomo_read, nelectron_read
                  ALLOCATE (eig_read(nmo_read), occ_read(nmo_read))
                  eig_read = 0.0_dp
                  occ_read = 0.0_dp
                  READ (rst_unit) eig_read(1:nmo_read), occ_read(1:nmo_read)
               ELSE
                  READ (rst_unit) nmo_read
                  ALLOCATE (eig_read(nmo_read))
                  eig_read = 0.0_dp
                  READ (rst_unit) eig_read(1:nmo_read)
               END IF
               IF (nmo_read < nmo) &
                  CALL cp_warn(cp__l("qs_loc_utils.F",1125), &
                               "The number of MOs on the restart unit is smaller than the number of "// &
                               "the allocated MOs. ")
               IF (nmo_read > nmo) &
                  CALL cp_warn(cp__l("qs_loc_utils.F",1129), &
                               "The number of MOs on the restart unit is greater than the number of "// &
                               "the allocated MOs. The read MO set will be truncated!")

               nmo = MIN(nmo, nmo_read)
               IF (do_homo) THEN
                  mos(ispin)%mo_set%eigenvalues(1:nmo) = eig_read(1:nmo)
                  mos(ispin)%mo_set%occupation_numbers(1:nmo) = occ_read(1:nmo)
                  DEALLOCATE (eig_read, occ_read)
               ELSE
                  evals(ispin)%array(1:nmo) = eig_read(1:nmo)
                  DEALLOCATE (eig_read)
               END IF

            END IF
            IF (do_homo) THEN
               CALL mp_bcast(mos(ispin)%mo_set%eigenvalues, source, group)
               CALL mp_bcast(mos(ispin)%mo_set%occupation_numbers, source, group)
            ELSE
               CALL mp_bcast(evals(ispin)%array, source, group)
            END IF

            DO i = 1, nmo
               IF (para_env%ionode) THEN
                  READ (rst_unit) vecbuffer
               ELSE
                  vecbuffer(1, :) = 0.0_dp
               END IF
               CALL mp_bcast(vecbuffer, source, group)
               CALL cp_fm_set_submatrix(mos_localized(ispin)%matrix, &
                                        vecbuffer, 1, i, nao, 1, transpose=.TRUE.)
            END DO
         END DO

         CALL mp_bcast(qs_loc_env%localized_wfn_control%loc_states, source, group)

         DEALLOCATE (vecbuffer)

      END IF

      ! Close restart file
      IF (para_env%ionode) THEN
         IF (file_exists) CALL close_file(unit_number=rst_unit)
      END IF

      CALL timestop(handle)

   END SUBROUTINE loc_read_restart

! **************************************************************************************************
!> \brief initializes everything needed for localization of the HOMOs
!> \param qs_loc_env ...
!> \param loc_section ...
!> \param do_homo ...
!> \param do_xas ...
!> \param nloc_xas ...
!> \par History
!>      2009 created
! **************************************************************************************************
   SUBROUTINE qs_loc_control_init(qs_loc_env, loc_section, do_homo, do_xas, nloc_xas)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: loc_section
      LOGICAL, INTENT(IN)                                :: do_homo
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_xas
      INTEGER, INTENT(IN), OPTIONAL                      :: nloc_xas

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_control_init', &
         routineP = moduleN//':'//routineN

      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control

      NULLIFY (localized_wfn_control)

      CALL localized_wfn_control_create(localized_wfn_control)
      CALL set_qs_loc_env(qs_loc_env, localized_wfn_control=localized_wfn_control)
      CALL localized_wfn_control_release(localized_wfn_control)
      CALL get_qs_loc_env(qs_loc_env, localized_wfn_control=localized_wfn_control)
      localized_wfn_control%do_homo = do_homo
      CALL read_loc_section(localized_wfn_control, loc_section, &
                            qs_loc_env%do_localize, do_xas, nloc_xas)

   END SUBROUTINE qs_loc_control_init

! **************************************************************************************************
!> \brief initializes everything needed for localization of the molecular orbitals
!> \param qs_env ...
!> \param qs_loc_env ...
!> \param localize_section ...
!> \param mos_localized ...
!> \param do_homo ...
!> \param do_mo_cubes ...
!> \param mo_loc_history ...
!> \param evals ...
! **************************************************************************************************
   SUBROUTINE qs_loc_init(qs_env, qs_loc_env, localize_section, mos_localized, &
                          do_homo, do_mo_cubes, mo_loc_history, evals)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: localize_section
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_localized
      LOGICAL, OPTIONAL                                  :: do_homo, do_mo_cubes
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mo_loc_history
      TYPE(cp_1d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: evals

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, i, ilast_intocc, ilow, &
                                                            ispin, iup, n_mo(2), n_mos(2), nao, &
                                                            nelectron, nmoloc(2), nspin, &
                                                            output_unit
      LOGICAL                                            :: my_do_homo, my_do_mo_cubes, restart_found
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues, occupation
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, mo_derivs
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: loc_print_section

      CALL timeset(routineN, handle)

      NULLIFY (mos, mo_coeff, mo_eigenvalues, occupation, ks_rmpv, mo_derivs, scf_control, para_env)
      CALL get_qs_env(qs_env, &
                      mos=mos, &
                      matrix_ks=ks_rmpv, &
                      mo_derivs=mo_derivs, &
                      scf_control=scf_control, &
                      para_env=para_env)

      loc_print_section => section_vals_get_subs_vals(localize_section, "PRINT")

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      nspin = SIZE(mos)
      IF (PRESENT(do_homo)) THEN
         my_do_homo = do_homo
      ELSE
         my_do_homo = .TRUE.
      END IF
      IF (PRESENT(do_mo_cubes)) THEN
         my_do_mo_cubes = do_mo_cubes
      ELSE
         my_do_mo_cubes = .FALSE.
      END IF
      restart_found = .FALSE.
      IF (qs_loc_env%do_localize) THEN
         ! Some setup for MOs to be localized
         CALL get_qs_loc_env(qs_loc_env, localized_wfn_control=localized_wfn_control)
         IF (localized_wfn_control%loc_restart) THEN
            CALL loc_read_restart(qs_loc_env, mos, mos_localized, localize_section, &
                                  loc_print_section, para_env, my_do_homo, restart_found, evals=evals)
            IF (output_unit > 0) WRITE (output_unit, "(/,T2,A,A)") "LOCALIZATION| ", &
               "   The orbitals to be localized are read from localization restart file."
            nmoloc = localized_wfn_control%nloc_states
         END IF
         IF (.NOT. restart_found) THEN
            nmoloc = 0
            DO ispin = 1, nspin
               CALL get_mo_set(mos(ispin)%mo_set, nmo=n_mo(ispin), nelectron=nelectron, homo=homo, nao=nao, &
                               mo_coeff=mo_coeff, eigenvalues=mo_eigenvalues, occupation_numbers=occupation)
               ! Get eigenstates (only needed if not already calculated before)
               IF ((.NOT. my_do_mo_cubes) &
                   !                  .OR. section_get_ival(dft_section,"PRINT%MO_CUBES%NHOMO")==0)&
                   .AND. my_do_homo .AND. qs_env%scf_env%method == ot_method_nr) THEN
                  CALL make_mo_eig(mos, nspin, ks_rmpv, scf_control, mo_derivs)
               END IF
               IF (localized_wfn_control%set_of_states == state_loc_all .AND. my_do_homo) THEN
                  nmoloc(ispin) = NINT(nelectron/occupation(1))
                  IF (n_mo(ispin) > homo) THEN
                     DO i = nmoloc(ispin), 1, -1
                        IF (occupation(1)-occupation(i) < localized_wfn_control%eps_occ) THEN
                           ilast_intocc = i
                           EXIT
                        END IF
                     END DO
                  ELSE
                     ilast_intocc = nmoloc(ispin)
                  END IF
                  nmoloc(ispin) = ilast_intocc
                  localized_wfn_control%lu_bound_states(1, ispin) = 1
                  localized_wfn_control%lu_bound_states(2, ispin) = ilast_intocc
                  IF (nmoloc(ispin) /= n_mo(ispin)) THEN
                     IF (output_unit > 0) &
                        WRITE (output_unit, "(/,T2,A,I4,A,I6,A,/,T15,A,F12.6,A,F12.6,A)") &
                        "LOCALIZATION| Spin ", ispin, " The first ", &
                        ilast_intocc, " occupied orbitals are localized,", " with energies from ", &
                        mo_eigenvalues(1), " to ", mo_eigenvalues(ilast_intocc), " [a.u.]."
                  END IF
               ELSE IF (localized_wfn_control%set_of_states == energy_loc_range .AND. my_do_homo) THEN
                  ilow = 0
                  iup = 0
                  DO i = 1, n_mo(ispin)
                     IF (mo_eigenvalues(i) >= localized_wfn_control%lu_ene_bound(1)) THEN
                        ilow = i
                        EXIT
                     END IF
                  END DO
                  DO i = n_mo(ispin), 1, -1
                     IF (mo_eigenvalues(i) <= localized_wfn_control%lu_ene_bound(2)) THEN
                        iup = i
                        EXIT
                     END IF
                  END DO
                  localized_wfn_control%lu_bound_states(1, ispin) = ilow
                  localized_wfn_control%lu_bound_states(2, ispin) = iup
                  localized_wfn_control%nloc_states(ispin) = iup-ilow+1
                  nmoloc(ispin) = localized_wfn_control%nloc_states(ispin)
                  IF (occupation(ilow)-occupation(iup) > localized_wfn_control%eps_occ) THEN
                     CALL cp_abort(cp__l("qs_loc_utils.F",1343), &
                                   "The selected energy range includes orbitals with different occupation number. "// &
                                   " The localization procedure cannot be applied.")
                  END IF
                  IF (output_unit > 0) WRITE (output_unit, "(/,T2,A,I4,A,I6,A)") "LOCALIZATION| Spin ", ispin, " : ", &
                     nmoloc(ispin), " orbitals in the selected energy range are localized."
               ELSE IF (localized_wfn_control%set_of_states == state_loc_all .AND. (.NOT. my_do_homo)) THEN
                  nmoloc(ispin) = n_mo(ispin)-homo
                  localized_wfn_control%lu_bound_states(1, ispin) = homo+1
                  localized_wfn_control%lu_bound_states(2, ispin) = n_mo(ispin)
                  IF (output_unit > 0) &
                     WRITE (output_unit, "(/,T2,A,I4,A,I6,A,/,T15,A,F12.6,A,F12.6,A)") &
                     "LOCALIZATION| Spin ", ispin, " The first ", &
                     nmoloc(ispin), " virtual orbitals are localized,", " with energies from ", &
                     mo_eigenvalues(homo+1), " to ", mo_eigenvalues(n_mo(ispin)), " [a.u.]."
               ELSE
                  nmoloc(ispin) = MIN(localized_wfn_control%nloc_states(1), n_mo(ispin))
                  IF (output_unit > 0 .AND. my_do_homo) WRITE (output_unit, "(/,T2,A,I4,A,I6,A)") "LOCALIZATION| Spin ", ispin, &
                     " : ", nmoloc(ispin), " occupied orbitals are localized, as given in the input list."
                  IF (output_unit > 0 .AND. (.NOT. my_do_homo)) WRITE (output_unit, "(/,T2,A,I4,A,I6,A)") "LOCALIZATION| Spin ", &
                     ispin, " : ", nmoloc(ispin), " unoccupied orbitals are localized, as given in the input list."
                  IF (n_mo(ispin) > homo .AND. my_do_homo) THEN
                     ilow = localized_wfn_control%loc_states(1, ispin)
                     DO i = 2, nmoloc(ispin)
                        iup = localized_wfn_control%loc_states(i, ispin)
                        IF (ABS(occupation(ilow)-occupation(iup)) > localized_wfn_control%eps_occ) THEN
                           ! write warning
                           CALL cp_warn(cp__l("qs_loc_utils.F",1370), &
                                        "User requested the calculation of localized wavefunction from a subset of MOs, "// &
                                        "including MOs with different occupations. Check the selected subset, "// &
                                        " the electronic density is not invariant with "// &
                                        "respect to rotations among orbitals with different occupation numbers!")
                        END IF
                     END DO
                  END IF
               ENDIF
            ENDDO ! ispin
            n_mos(:) = nao-n_mo(:)
            IF (my_do_homo) n_mos = n_mo
            CALL set_loc_wfn_lists(localized_wfn_control, nmoloc, n_mos, nspin)
         END IF
         CALL set_loc_centers(localized_wfn_control, nmoloc, nspin)
         IF (my_do_homo) THEN
            CALL qs_loc_env_init(qs_loc_env, localized_wfn_control, qs_env, &
                                 loc_coeff=mos_localized, mo_loc_history=mo_loc_history)
         END IF
      ELSE
         ! Let's inform in case the section is not present in the input
         CALL cp_warn(cp__l("qs_loc_utils.F",1391), &
                      "User requested the calculation of the localized wavefunction but the section "// &
                      "LOCALIZE was not specified. Localization will not be performed!")
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_loc_init

! **************************************************************************************************
!> \brief read the controlparameter from input, using the new input scheme
!> \param localized_wfn_control ...
!> \param loc_section ...
!> \param localize ...
!> \param do_xas ...
!> \param nloc_xas ...
!> \par History
!>      05.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE read_loc_section(localized_wfn_control, loc_section, &
                               localize, do_xas, nloc_xas)

      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(section_vals_type), POINTER                   :: loc_section
      LOGICAL, INTENT(OUT)                               :: localize
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_xas
      INTEGER, INTENT(IN), OPTIONAL                      :: nloc_xas

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_loc_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind, ir, n_list, n_rep, n_state, &
                                                            nline, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: list, loc_list
      LOGICAL                                            :: my_do_xas
      REAL(dp), POINTER                                  :: ene(:)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: loc_print_section

      my_do_xas = .FALSE.
      IF (PRESENT(do_xas)) THEN
         my_do_xas = do_xas
         IF(.NOT.(PRESENT(nloc_xas)))CALL cp__a("qs_loc_utils.F",1433)
      END IF
      IF(.NOT.(ASSOCIATED(loc_section)))CALL cp__a("qs_loc_utils.F",1435)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      CALL section_vals_val_get(loc_section, "_SECTION_PARAMETERS_", l_val=localize)
      IF (localize) THEN
         loc_print_section => section_vals_get_subs_vals(loc_section, "PRINT")
         NULLIFY (list)
         NULLIFY (loc_list)
         localized_wfn_control%lu_bound_states = 0
         localized_wfn_control%lu_ene_bound = 0.0_dp
         localized_wfn_control%nloc_states = 0
         localized_wfn_control%set_of_states = 0
         n_state = 0

         CALL section_vals_val_get(loc_section, "MAX_ITER", &
                                   i_val=localized_wfn_control%max_iter)
         CALL section_vals_val_get(loc_section, "MAX_CRAZY_ANGLE", &
                                   r_val=localized_wfn_control%max_crazy_angle)
         CALL section_vals_val_get(loc_section, "CRAZY_SCALE", &
                                   r_val=localized_wfn_control%crazy_scale)
         CALL section_vals_val_get(loc_section, "EPS_OCCUPATION", &
                                   r_val=localized_wfn_control%eps_occ)
         CALL section_vals_val_get(loc_section, "CRAZY_USE_DIAG", &
                                   l_val=localized_wfn_control%crazy_use_diag)
         CALL section_vals_val_get(loc_section, "OUT_ITER_EACH", &
                                   i_val=localized_wfn_control%out_each)
         CALL section_vals_val_get(loc_section, "EPS_LOCALIZATION", &
                                   r_val=localized_wfn_control%eps_localization)
         CALL section_vals_val_get(loc_section, "MIN_OR_MAX", &
                                   i_val=localized_wfn_control%min_or_max)
         CALL section_vals_val_get(loc_section, "JACOBI_FALLBACK", &
                                   l_val=localized_wfn_control%jacobi_fallback)
         CALL section_vals_val_get(loc_section, "METHOD", &
                                   i_val=localized_wfn_control%localization_method)
         CALL section_vals_val_get(loc_section, "OPERATOR", &
                                   i_val=localized_wfn_control%operator_type)
         CALL section_vals_val_get(loc_section, "RESTART", &
                                   l_val=localized_wfn_control%loc_restart)
         CALL section_vals_val_get(loc_section, "USE_HISTORY", &
                                   l_val=localized_wfn_control%use_history)

         IF (localized_wfn_control%do_homo) THEN
            ! List of States HOMO
            CALL section_vals_val_get(loc_section, "LIST", n_rep_val=n_rep)
            IF (n_rep > 0) THEN
               n_list = 0
               DO ir = 1, n_rep
                  NULLIFY (list)
                  CALL section_vals_val_get(loc_section, "LIST", i_rep_val=ir, i_vals=list)
                  IF (ASSOCIATED(list)) THEN
                     CALL reallocate(loc_list, 1, n_list+SIZE(list))
                     DO i = 1, SIZE(list)
                        loc_list(n_list+i) = list(i)
                     END DO ! i
                     n_list = n_list+SIZE(list)
                  END IF
               END DO ! ir
               IF (n_list /= 0) THEN
                  localized_wfn_control%set_of_states = state_loc_list
                  ALLOCATE (localized_wfn_control%loc_states(n_list, 2))
                  localized_wfn_control%loc_states = 0
                  localized_wfn_control%loc_states(:, 1) = loc_list(:)
                  localized_wfn_control%loc_states(:, 2) = loc_list(:)
                  localized_wfn_control%nloc_states(1) = n_list
                  IF (my_do_xas) THEN
                     localized_wfn_control%nloc_states(2) = 0
                     localized_wfn_control%loc_states(:, 2) = 0
                  END IF
                  DEALLOCATE (loc_list)
               END IF
            END IF

         ELSE
            ! List of States LUMO
            CALL section_vals_val_get(loc_section, "LIST_UNOCCUPIED", n_rep_val=n_rep)
            IF (n_rep > 0) THEN
               n_list = 0
               DO ir = 1, n_rep
                  NULLIFY (list)
                  CALL section_vals_val_get(loc_section, "LIST_UNOCCUPIED", i_rep_val=ir, i_vals=list)
                  IF (ASSOCIATED(list)) THEN
                     CALL reallocate(loc_list, 1, n_list+SIZE(list))
                     DO i = 1, SIZE(list)
                        loc_list(n_list+i) = list(i)
                     END DO ! i
                     n_list = n_list+SIZE(list)
                  END IF
               END DO ! ir
               IF (n_list /= 0) THEN
                  localized_wfn_control%set_of_states = state_loc_list
                  ALLOCATE (localized_wfn_control%loc_states(n_list, 2))
                  localized_wfn_control%loc_states = 0
                  localized_wfn_control%loc_states(:, 1) = loc_list(:)
                  localized_wfn_control%loc_states(:, 2) = loc_list(:)
                  localized_wfn_control%nloc_states(1) = n_list
                  DEALLOCATE (loc_list)
               END IF
            END IF
         END IF

         IF (localized_wfn_control%set_of_states == 0) THEN
            CALL section_vals_val_get(loc_section, "ENERGY_RANGE", r_vals=ene)
            IF (ene(1) /= ene(2)) THEN
               localized_wfn_control%set_of_states = energy_loc_range
               localized_wfn_control%lu_ene_bound(1) = ene(1)
               localized_wfn_control%lu_ene_bound(2) = ene(2)
            END IF
         END IF

         ! All States or XAS specific states
         IF (localized_wfn_control%set_of_states == 0) THEN
            IF (my_do_xas) THEN
               localized_wfn_control%set_of_states = state_loc_range
               localized_wfn_control%nloc_states(1) = nloc_xas
               localized_wfn_control%lu_bound_states(1, 1) = 1
               localized_wfn_control%lu_bound_states(2, 1) = nloc_xas
               localized_wfn_control%nloc_states(2) = 0
               localized_wfn_control%lu_bound_states(1, 2) = 0
               localized_wfn_control%lu_bound_states(2, 2) = 0
            ELSE
               localized_wfn_control%set_of_states = state_loc_all
            END IF
         END IF

         localized_wfn_control%print_centers = &
            BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                             "WANNIER_CENTERS"), cp_p_file)
         localized_wfn_control%print_spreads = &
            BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                             "WANNIER_SPREADS"), cp_p_file)
         localized_wfn_control%print_cubes = &
            BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                             "WANNIER_CUBES"), cp_p_file)

         output_unit = cp_print_key_unit_nr(logger, loc_print_section, "PROGRAM_RUN_INFO", &
                                            extension=".Log")

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "LOCALIZE| The spread relative to a set of orbitals is computed"

            SELECT CASE (localized_wfn_control%set_of_states)
            CASE (state_loc_all)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Orbitals to be localized: All orbitals"
               WRITE (UNIT=output_unit, FMT="(T2,A,/,T12,A,F16.8)") &
                  "LOCALIZE| If fractional occupation, fully occupied MOs are those ", &
                  "within occupation tolerance of ", localized_wfn_control%eps_occ
            CASE (state_loc_range)
               WRITE (UNIT=output_unit, FMT="(T2,A,T65,I8,A,I8)") &
                  "LOCALIZE| Orbitals to be localized: Those with index between ", &
                  localized_wfn_control%lu_bound_states(1, 1), " and ", &
                  localized_wfn_control%lu_bound_states(2, 1)
            CASE (state_loc_list)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Orbitals to be localized: Those with index in the following list"
               nline = localized_wfn_control%nloc_states(1)/10+1
               ind = 0
               DO i = 1, nline
                  IF (ind+10 < localized_wfn_control%nloc_states(1)) THEN
                     WRITE (UNIT=output_unit, FMT="(T8,10I7)") localized_wfn_control%loc_states(ind+1:ind+10, 1)
                     ind = ind+10
                  ELSE
                     WRITE (UNIT=output_unit, FMT="(T8,10I7)") &
                        localized_wfn_control%loc_states(ind+1:localized_wfn_control%nloc_states(1), 1)
                     ind = localized_wfn_control%nloc_states(1)
                  END IF
               END DO
            CASE (energy_loc_range)
               WRITE (UNIT=output_unit, FMT="(T2,A,T65,/,f16.6,A,f16.6,A)") &
                  "LOCALIZE| Orbitals to be localized: Those with energy in the range between ", &
                  localized_wfn_control%lu_ene_bound(1), " and ", localized_wfn_control%lu_ene_bound(2), " a.u."
            CASE DEFAULT
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Orbitals to be localized: None "
            END SELECT

            SELECT CASE (localized_wfn_control%operator_type)
            CASE (op_loc_berry)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Spread defined by the Berry phase operator "
            CASE (op_loc_boys)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Spread defined by the Boys phase operator "
            CASE DEFAULT
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Spread defined by the Pipek phase operator "
            END SELECT

            SELECT CASE (localized_wfn_control%localization_method)
            CASE (do_loc_jacobi)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Optimal unitary transformation generated by Jacobi algorithm"
            CASE (do_loc_crazy)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Optimal unitary transformation generated by Crazy angle algorithm"
               WRITE (UNIT=output_unit, FMT="(T2,A,F16.8)") &
                  "LOCALIZE| maximum angle: ", localized_wfn_control%max_crazy_angle
               WRITE (UNIT=output_unit, FMT="(T2,A,F16.8)") &
                  "LOCALIZE| scaling: ", localized_wfn_control%crazy_scale
               WRITE (UNIT=output_unit, FMT="(T2,A,L1)") &
                  "LOCALIZE| use diag:", localized_wfn_control%crazy_use_diag
            CASE (do_loc_direct)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Optimal unitary transformation generated by direct algorithm"
            CASE (do_loc_l1_norm_sd)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| Optimal unitary transformation generated by "// &
                  " steepest descent algorithm applied on an approximate l1 norm"
            CASE (do_loc_none)
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "LOCALIZE| No unitary transformation is applied"
            END SELECT

         END IF ! process has output_unit

         CALL cp_print_key_finished_output(output_unit, logger, loc_print_section, "PROGRAM_RUN_INFO")

      ELSE
         localized_wfn_control%localization_method = do_loc_none
         localized_wfn_control%localization_method = state_loc_none
         localized_wfn_control%print_centers = .FALSE.
         localized_wfn_control%print_spreads = .FALSE.
         localized_wfn_control%print_cubes = .FALSE.
      END IF

   END SUBROUTINE read_loc_section

! **************************************************************************************************
!> \brief create the center and spread array and the file names for the output
!> \param localized_wfn_control ...
!> \param nmoloc ...
!> \param nspins ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE set_loc_centers(localized_wfn_control, nmoloc, nspins)

      TYPE(localized_wfn_control_type)                   :: localized_wfn_control
      INTEGER, DIMENSION(2), INTENT(IN)                  :: nmoloc
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'set_loc_centers', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin

      DO ispin = 1, nspins
         ALLOCATE (localized_wfn_control%centers_set(ispin)%array(6, nmoloc(ispin)))
         localized_wfn_control%centers_set(ispin)%array = 0.0_dp
      END DO

   END SUBROUTINE set_loc_centers

! **************************************************************************************************
!> \brief create the lists of mos that are taken into account
!> \param localized_wfn_control ...
!> \param nmoloc ...
!> \param nmo ...
!> \param nspins ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE set_loc_wfn_lists(localized_wfn_control, nmoloc, nmo, nspins)

      TYPE(localized_wfn_control_type)                   :: localized_wfn_control
      INTEGER, DIMENSION(2), INTENT(IN)                  :: nmoloc, nmo
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'set_loc_wfn_lists', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ispin, max_iloc, max_nmoloc, state

      CALL timeset(routineN, state)

      localized_wfn_control%nloc_states(1:2) = nmoloc(1:2)
      max_nmoloc = MAX(nmoloc(1), nmoloc(2))

      SELECT CASE (localized_wfn_control%set_of_states)
      CASE (state_loc_list)
         ! List
         IF(.NOT.(ASSOCIATED(localized_wfn_control%loc_states)))CALL cp__a("qs_loc_utils.F",1718)
         DO ispin = 1, nspins
            localized_wfn_control%lu_bound_states(1, ispin) = 1
            localized_wfn_control%lu_bound_states(2, ispin) = nmoloc(ispin)
            IF (nmoloc(ispin) < 1) THEN
               localized_wfn_control%lu_bound_states(1, ispin) = 0
               localized_wfn_control%loc_states(:, ispin) = 0
            END IF
         END DO
      CASE (state_loc_range)
         ! Range
         ALLOCATE (localized_wfn_control%loc_states(max_nmoloc, 2))
         localized_wfn_control%loc_states = 0
         DO ispin = 1, nspins
            localized_wfn_control%lu_bound_states(1, ispin) = &
               localized_wfn_control%lu_bound_states(1, 1)
            localized_wfn_control%lu_bound_states(2, ispin) = &
               localized_wfn_control%lu_bound_states(1, 1)+nmoloc(ispin)-1
            max_iloc = localized_wfn_control%lu_bound_states(2, ispin)
            DO i = 1, nmoloc(ispin)
               localized_wfn_control%loc_states(i, ispin) = localized_wfn_control%lu_bound_states(1, ispin)+i-1
            END DO
            IF(.NOT.(max_iloc <= nmo(ispin)))CALL cp__a("qs_loc_utils.F",1740)
         END DO
      CASE (energy_loc_range)
         ! Energy
         ALLOCATE (localized_wfn_control%loc_states(max_nmoloc, 2))
         localized_wfn_control%loc_states = 0
         DO ispin = 1, nspins
            DO i = 1, nmoloc(ispin)
               localized_wfn_control%loc_states(i, ispin) = localized_wfn_control%lu_bound_states(1, ispin)+i-1
            END DO
         END DO
      CASE (state_loc_all)
         ! All
         ALLOCATE (localized_wfn_control%loc_states(max_nmoloc, 2))
         localized_wfn_control%loc_states = 0

         IF (localized_wfn_control%lu_bound_states(1, 1) == 1) THEN
            DO ispin = 1, nspins
               localized_wfn_control%lu_bound_states(1, ispin) = 1
               localized_wfn_control%lu_bound_states(2, ispin) = nmoloc(ispin)
               IF (nmoloc(ispin) < 1) localized_wfn_control%lu_bound_states(1, ispin) = 0
               DO i = 1, nmoloc(ispin)
                  localized_wfn_control%loc_states(i, ispin) = i
               END DO
            END DO
         ELSE
            DO ispin = 1, nspins
               IF (nmoloc(ispin) < 1) localized_wfn_control%lu_bound_states(1, ispin) = 0
               DO i = 1, nmoloc(ispin)
                  localized_wfn_control%loc_states(i, ispin) = &
                     localized_wfn_control%lu_bound_states(1, ispin)+i-1
               END DO
            END DO
         END IF
      END SELECT

      CALL timestop(state)

   END SUBROUTINE set_loc_wfn_lists

END MODULE qs_loc_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Output Utilities for MOTION_SECTION
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   02.2008
! **************************************************************************************************
MODULE motion_utils

   USE cell_types,                      ONLY: cell_type
   USE cp2k_info,                       ONLY: compile_revision,&
                                              cp2k_version,&
                                              r_datx,&
                                              r_host_name,&
                                              r_user_name
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_constants,                 ONLY: dump_atomic,&
                                              dump_dcd,&
                                              dump_dcd_aligned_cell,&
                                              dump_pdb,&
                                              dump_xmol
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              sp
   USE machine,                         ONLY: m_flush
   USE mathlib,                         ONLY: diamat_all
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: write_trajectory, write_stress_tensor, write_simulation_cell, &
             get_output_format, rot_ana

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'motion_utils'
   REAL(KIND=dp), PARAMETER, PUBLIC     :: thrs_motion = 1.0E4_dp*EPSILON(0.0_dp)

CONTAINS

! **************************************************************************************************
!> \brief Performs an analysis of the principal inertia axis
!>      Getting back the generators of the translating and
!>      rotating frame
!> \param particles ...
!> \param mat ...
!> \param dof ...
!> \param print_section ...
!> \param keep_rotations ...
!> \param mass_weighted ...
!> \param natoms ...
!> \param rot_dof ...
!> \param inertia ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE rot_ana(particles, mat, dof, print_section, keep_rotations, mass_weighted, &
                      natoms, rot_dof, inertia)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: mat
      INTEGER, INTENT(OUT)                               :: dof
      TYPE(section_vals_type), POINTER                   :: print_section
      LOGICAL, INTENT(IN)                                :: keep_rotations, mass_weighted
      INTEGER, INTENT(IN)                                :: natoms
      INTEGER, INTENT(OUT), OPTIONAL                     :: rot_dof
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: inertia(3)

      CHARACTER(len=*), PARAMETER :: routineN = 'rot_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iparticle, iseq, iw, j, k, &
                                                            lrot(3)
      LOGICAL                                            :: present_mat
      REAL(KIND=dp)                                      :: cp(3), Ip(3, 3), Ip_eigval(3), mass, &
                                                            masst, norm, rcom(3), rm(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Rot, Tr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      present_mat = PRESENT(mat)
      IF(.NOT.(ASSOCIATED(particles)))CALL cp__a("motion_utils.F",101)
      IF (present_mat) THEN
         IF(.NOT.(.NOT. ASSOCIATED(mat)))CALL cp__a("motion_utils.F",103)
      END IF
      IF (.NOT. keep_rotations) THEN
         rcom = 0.0_dp
         masst = 0.0_dp
         ! Center of mass
         DO iparticle = 1, natoms
            mass = 1.0_dp
            IF (mass_weighted) mass = particles(iparticle)%atomic_kind%mass
            IF(.NOT.(mass >= 0.0_dp))CALL cp__a("motion_utils.F",112)
            masst = masst+mass
            rcom = particles(iparticle)%r*mass+rcom
         END DO
         IF(.NOT.(masst > 0.0_dp))CALL cp__a("motion_utils.F",116)
         rcom = rcom/masst
         ! Intertia Tensor
         Ip = 0.0_dp
         DO iparticle = 1, natoms
            mass = 1.0_dp
            IF (mass_weighted) mass = particles(iparticle)%atomic_kind%mass
            rm = particles(iparticle)%r-rcom
            Ip(1, 1) = Ip(1, 1)+mass*(rm(2)**2+rm(3)**2)
            Ip(2, 2) = Ip(2, 2)+mass*(rm(1)**2+rm(3)**2)
            Ip(3, 3) = Ip(3, 3)+mass*(rm(1)**2+rm(2)**2)
            Ip(1, 2) = Ip(1, 2)-mass*(rm(1)*rm(2))
            Ip(1, 3) = Ip(1, 3)-mass*(rm(1)*rm(3))
            Ip(2, 3) = Ip(2, 3)-mass*(rm(2)*rm(3))
         END DO
         ! Diagonalize the Inertia Tensor
         CALL diamat_all(Ip, Ip_eigval)
         IF (PRESENT(inertia)) inertia = Ip_eigval
         iw = cp_print_key_unit_nr(logger, print_section, "ROTATIONAL_INFO", extension=".vibLog")
         IF (iw > 0) THEN
            WRITE (iw, '(T2,A,3F12.6)') "ROT| Rotational Analysis Info "
            WRITE (iw, '(T2,A)') "ROT| Principal axes and moments of inertia in atomic units:"
            WRITE (iw, '(T2,A,T27,3(3X,I9,6X))') "ROT|", 1, 2, 3
            WRITE (iw, '(T2,A,T27,3(3X,E15.9))') "ROT| EIGENVALUES", Ip_eigval
            WRITE (iw, '(T2,A,T27,3(3X,F15.9))') "ROT|      X", Ip(1, 1), Ip(1, 2), Ip(1, 3)
            WRITE (iw, '(T2,A,T27,3(3X,F15.9))') "ROT|      Y", Ip(2, 1), Ip(2, 2), Ip(2, 3)
            WRITE (iw, '(T2,A,T27,3(3X,F15.9))') "ROT|      Z", Ip(3, 1), Ip(3, 2), Ip(3, 3)
         END IF
         CALL cp_print_key_finished_output(iw, logger, print_section, "ROTATIONAL_INFO")
         iw = cp_print_key_unit_nr(logger, print_section, "ROTATIONAL_INFO/COORDINATES", extension=".vibLog")
         IF (iw > 0) THEN
            WRITE (iw, '(/,T2,A)') "ROT| Standard Molecule Orientation - ANGSTROM "
            DO iparticle = 1, natoms
               WRITE (iw, '(T2,"ROT|",T20,A,T27,3(3X,F15.9))') &
                  TRIM(particles(iparticle)%atomic_kind%name), &
                  MATMUL(particles(iparticle)%r, Ip)*angstrom
            END DO
         END IF
         CALL cp_print_key_finished_output(iw, logger, print_section, "ROTATIONAL_INFO/COORDINATES")
      END IF
      ! Build up the Translational vectors
      ALLOCATE (Tr(natoms*3, 3))
      Tr = 0.0_dp
      DO k = 1, 3
         iseq = 0
         DO iparticle = 1, natoms
            mass = 1.0_dp
            IF (mass_weighted) mass = SQRT(particles(iparticle)%atomic_kind%mass)
            DO j = 1, 3
               iseq = iseq+1
               IF (j == k) Tr(iseq, k) = mass
            END DO
         END DO
      END DO
      ! Normalize Translations
      DO i = 1, 3
         norm = SQRT(DOT_PRODUCT(Tr(:, i), Tr(:, i)))
         Tr(:, i) = Tr(:, i)/norm
      END DO
      dof = 3
      ! Build up the Rotational vectors
      ALLOCATE (Rot(natoms*3, 3))
      lrot = 0
      IF (.NOT. keep_rotations) THEN
         DO iparticle = 1, natoms
            mass = 1.0_dp
            IF (mass_weighted) mass = SQRT(particles(iparticle)%atomic_kind%mass)
            rm = particles(iparticle)%r-rcom
            cp(1) = rm(1)*Ip(1, 1)+rm(2)*Ip(2, 1)+rm(3)*Ip(3, 1)
            cp(2) = rm(1)*Ip(1, 2)+rm(2)*Ip(2, 2)+rm(3)*Ip(3, 2)
            cp(3) = rm(1)*Ip(1, 3)+rm(2)*Ip(2, 3)+rm(3)*Ip(3, 3)
            ! X Rot
            Rot((iparticle-1)*3+1, 1) = (cp(2)*Ip(1, 3)-Ip(1, 2)*cp(3))*mass
            Rot((iparticle-1)*3+2, 1) = (cp(2)*Ip(2, 3)-Ip(2, 2)*cp(3))*mass
            Rot((iparticle-1)*3+3, 1) = (cp(2)*Ip(3, 3)-Ip(3, 2)*cp(3))*mass
            ! Y Rot
            Rot((iparticle-1)*3+1, 2) = (cp(3)*Ip(1, 1)-Ip(1, 3)*cp(1))*mass
            Rot((iparticle-1)*3+2, 2) = (cp(3)*Ip(2, 1)-Ip(2, 3)*cp(1))*mass
            Rot((iparticle-1)*3+3, 2) = (cp(3)*Ip(3, 1)-Ip(3, 3)*cp(1))*mass
            ! Z Rot
            Rot((iparticle-1)*3+1, 3) = (cp(1)*Ip(1, 2)-Ip(1, 1)*cp(2))*mass
            Rot((iparticle-1)*3+2, 3) = (cp(1)*Ip(2, 2)-Ip(2, 1)*cp(2))*mass
            Rot((iparticle-1)*3+3, 3) = (cp(1)*Ip(3, 2)-Ip(3, 1)*cp(2))*mass
         END DO

         ! Normalize Rotations and count the number of degree of freedom
         lrot = 1
         DO i = 1, 3
            norm = SQRT(DOT_PRODUCT(Rot(:, i), Rot(:, i)))
            IF (norm <= thrs_motion) THEN
               lrot(i) = 0
               CYCLE
            END IF
            Rot(:, i) = Rot(:, i)/norm
            ! Clean Rotational modes for spurious/numerical contamination
            IF (i < 3) THEN
               DO j = 1, i
                  Rot(:, i+1) = Rot(:, i+1)-DOT_PRODUCT(Rot(:, i+1), Rot(:, j))*Rot(:, j)
               END DO
            END IF
         END DO
      END IF
      IF (PRESENT(rot_dof)) rot_dof = COUNT(lrot == 1)
      dof = dof+COUNT(lrot == 1)
      iw = cp_print_key_unit_nr(logger, print_section, "ROTATIONAL_INFO", extension=".vibLog")
      IF (iw > 0) THEN
         WRITE (iw, '(T2,A,I6)') "ROT| Numer of Rotovibrational vectors:", dof
         IF (dof == 5) WRITE (iw, '(T2,A)') "ROT| Linear Molecule detected.."
         IF (dof == 3 .AND. (.NOT. keep_rotations)) WRITE (iw, '(T2,A)') "ROT| Single Atom detected.."
      END IF
      CALL cp_print_key_finished_output(iw, logger, print_section, "ROTATIONAL_INFO")
      IF (present_mat) THEN
         ! Give back the vectors generating the rototranslating Frame
         ALLOCATE (mat(natoms*3, dof))
         iseq = 0
         DO i = 1, 3
            mat(:, i) = Tr(:, i)
            IF (lrot(i) == 1) THEN
               iseq = iseq+1
               mat(:, 3+iseq) = Rot(:, i)
            END IF
         END DO
      END IF
      DEALLOCATE (Tr)
      DEALLOCATE (Rot)
      CALL timestop(handle)
   END SUBROUTINE rot_ana

! **************************************************************************************************
!> \brief   Prints the information controlled by the TRAJECTORY section
!> \param force_env ...
!> \param root_section ...
!> \param it ...
!> \param time ...
!> \param dtime ...
!> \param etot ...
!> \param pk_name ...
!> \param pos ...
!> \param act ...
!> \param middle_name ...
!> \param particles ...
!> \param extended_xmol_title ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_trajectory(force_env, root_section, it, time, dtime, etot, pk_name, &
                               pos, act, middle_name, particles, extended_xmol_title)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: it
      REAL(KIND=dp), INTENT(IN)                          :: time, dtime, etot
      CHARACTER(LEN=*), OPTIONAL                         :: pk_name
      CHARACTER(LEN=default_string_length), OPTIONAL     :: pos, act
      CHARACTER(LEN=*), OPTIONAL                         :: middle_name
      TYPE(particle_list_type), OPTIONAL, POINTER        :: particles
      LOGICAL, INTENT(IN), OPTIONAL                      :: extended_xmol_title

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_trajectory', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=4)                                   :: id_dcd
      CHARACTER(LEN=default_string_length)               :: id_label, id_wpc, my_act, my_ext, &
                                                            my_form, my_middle, my_pk_name, &
                                                            my_pos, remark1, remark2, section_ref, &
                                                            title, unit_str
      INTEGER                                            :: handle, i, ii, iskip, nat, outformat, &
                                                            traj_unit
      INTEGER, POINTER                                   :: force_mixing_indices(:), &
                                                            force_mixing_labels(:)
      LOGICAL                                            :: charge_beta, charge_extended, &
                                                            charge_occup, explicit, &
                                                            my_extended_xmol_title, new_file
      REAL(dp), ALLOCATABLE                              :: fml_array(:)
      REAL(KIND=dp)                                      :: unit_conv
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: my_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: force_env_section, &
                                                            force_mixing_restart_section

      CALL timeset(routineN, handle)

      NULLIFY (logger, cell, subsys, my_particles, particle_set)
      logger => cp_get_default_logger()
      id_label = logger%iter_info%level_name(logger%iter_info%n_rlevel)
      my_pos = "APPEND"
      my_act = "WRITE"
      my_middle = "pos"
      my_pk_name = "TRAJECTORY"
      IF (PRESENT(middle_name)) my_middle = middle_name
      IF (PRESENT(pos)) my_pos = pos
      IF (PRESENT(act)) my_act = act
      IF (PRESENT(pk_name)) my_pk_name = pk_name

      SELECT CASE (TRIM (my_pk_name))
      CASE ("TRAJECTORY", "SHELL_TRAJECTORY", "CORE_TRAJECTORY")
         id_dcd = "CORD"
         id_wpc = "POS"
      CASE ("VELOCITIES", "SHELL_VELOCITIES", "CORE_VELOCITIES")
         id_dcd = "VEL "
         id_wpc = "VEL"
      CASE ("FORCES", "SHELL_FORCES", "CORE_FORCES")
         id_dcd = "FRC "
         id_wpc = "FORCE"
      CASE ("FORCE_MIXING_LABELS")
         id_dcd = "FML "
         id_wpc = "FORCE_MIXING_LABELS"
      CASE DEFAULT
         CALL cp__b("motion_utils.F",327,"")
      END SELECT

      charge_occup = .FALSE.
      charge_beta = .FALSE.
      charge_extended = .FALSE.

      CALL force_env_get(force_env, cell=cell, subsys=subsys)
      IF (PRESENT(particles)) THEN
         IF(.NOT.(ASSOCIATED(particles)))CALL cp__a("motion_utils.F",336)
         my_particles => particles
      ELSE
         CALL cp_subsys_get(subsys=subsys, particles=my_particles)
      END IF
      particle_set => my_particles%els
      nat = my_particles%n_els

      ! Gather units of measure for output (if available)
      IF (TRIM(my_pk_name) /= "FORCE_MIXING_LABELS") THEN
         CALL section_vals_val_get(root_section, "MOTION%PRINT%"//TRIM(my_pk_name)//"%UNIT", &
                                   c_val=unit_str)
         unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      ENDIF

      ! Get the output format
      CALL get_output_format(root_section, "MOTION%PRINT%"//TRIM(my_pk_name), my_form, my_ext)
      traj_unit = cp_print_key_unit_nr(logger, root_section, "MOTION%PRINT%"//TRIM(my_pk_name), &
                                       extension=my_ext, file_position=my_pos, file_action=my_act, &
                                       file_form=my_form, middle_name=TRIM(my_middle), is_new_file=new_file)
      IF (traj_unit > 0) THEN
         CALL section_vals_val_get(root_section, "MOTION%PRINT%"//TRIM(my_pk_name)//"%FORMAT", &
                                   i_val=outformat)
         title = ""
         SELECT CASE (outformat)
         CASE (dump_dcd, dump_dcd_aligned_cell)
            IF (new_file) THEN
               !Lets write the header for the coordinate dcd
               section_ref = "MOTION%PRINT%"//TRIM(my_pk_name)//"%EACH%"//TRIM(id_label)
               iskip = section_get_ival(root_section, TRIM(section_ref))
               WRITE (UNIT=traj_unit) id_dcd, 0, it, iskip, 0, 0, 0, 0, 0, 0, REAL(dtime, KIND=sp), &
                  1, 0, 0, 0, 0, 0, 0, 0, 0, 24
               remark1 = "REMARK "//id_dcd//" DCD file created by "//TRIM(cp2k_version)// &
                         " (revision "//TRIM(compile_revision)//")"
               remark2 = "REMARK "//TRIM(r_user_name)//"@"//TRIM(r_host_name)
               WRITE (UNIT=traj_unit) 2, remark1, remark2
               WRITE (UNIT=traj_unit) nat
               CALL m_flush(traj_unit)
            ENDIF
         CASE (dump_xmol)
            my_extended_xmol_title = .FALSE.
            IF (PRESENT(extended_xmol_title)) my_extended_xmol_title = extended_xmol_title
            ! This information can be digested by Molden
            IF (my_extended_xmol_title) THEN
               WRITE (UNIT=title, FMT="(A,I8,A,F12.3,A,F20.10)") &
                  " i = ", it, ", time = ", time, ", E = ", etot
            ELSE
               WRITE (UNIT=title, FMT="(A,I8,A,F20.10)") " i = ", it, ", E = ", etot
            END IF
         CASE (dump_atomic)
            ! Do nothing
         CASE (dump_pdb)
            IF (id_wpc == "POS") THEN
               CALL section_vals_val_get(root_section, "MOTION%PRINT%TRAJECTORY%CHARGE_OCCUP", &
                                         l_val=charge_occup)
               CALL section_vals_val_get(root_section, "MOTION%PRINT%TRAJECTORY%CHARGE_BETA", &
                                         l_val=charge_beta)
               CALL section_vals_val_get(root_section, "MOTION%PRINT%TRAJECTORY%CHARGE_EXTENDED", &
                                         l_val=charge_extended)
               i = COUNT((/charge_occup, charge_beta, charge_extended/))
               IF (i > 1) &
                  CALL cp__b("motion_utils.F",397,"Either only CHARGE_OCCUP, CHARGE_BETA, or CHARGE_EXTENDED can be selected, ")
            END IF
            IF (new_file) THEN
               ! COLUMNS        DATA TYPE       FIELD          DEFINITION
               !  1 -  6        Record name     "TITLE "
               !  9 - 10        Continuation    continuation   Allows concatenation
               ! 11 - 70        String          title          Title of the experiment
               WRITE (UNIT=traj_unit, FMT="(A6,T11,A)") &
                  "TITLE ", "PDB file created by "//TRIM(cp2k_version)//" (revision "//TRIM(compile_revision)//")", &
                  "AUTHOR", TRIM(r_user_name)//"@"//TRIM(r_host_name)//" "//r_datx(1:19)
            END IF
            my_extended_xmol_title = .FALSE.
            IF (PRESENT(extended_xmol_title)) my_extended_xmol_title = extended_xmol_title
            IF (my_extended_xmol_title) THEN
               WRITE (UNIT=title, FMT="(A,I0,A,F0.3,A,F0.10)") &
                  "Step ", it, ", time = ", time, ", E = ", etot
            ELSE
               WRITE (UNIT=title, FMT="(A,I0,A,F0.10)") &
                  "Step ", it, ", E = ", etot
            END IF
         CASE DEFAULT
            CALL cp__b("motion_utils.F",418,"")
         END SELECT
         IF (TRIM(my_pk_name) == "FORCE_MIXING_LABELS") THEN
            ALLOCATE (fml_array(3*SIZE(particle_set)))
            fml_array = 0.0_dp
            CALL force_env_get(force_env, force_env_section=force_env_section)
            force_mixing_restart_section => section_vals_get_subs_vals(force_env_section, &
                                                                       "QMMM%FORCE_MIXING%RESTART_INFO", &
                                                                       can_return_null=.TRUE.)
            IF (ASSOCIATED(force_mixing_restart_section)) THEN
               CALL section_vals_get(force_mixing_restart_section, explicit=explicit)
               IF (explicit) THEN
                  CALL section_vals_val_get(force_mixing_restart_section, "INDICES", i_vals=force_mixing_indices)
                  CALL section_vals_val_get(force_mixing_restart_section, "LABELS", i_vals=force_mixing_labels)
                  DO i = 1, SIZE(force_mixing_indices)
                     ii = force_mixing_indices(i)
                     IF(.NOT.(ii <= SIZE(particle_set)))CALL cp__a("motion_utils.F",434)
                     fml_array((ii-1)*3+1:(ii-1)*3+3) = force_mixing_labels(i)
                  END DO
               ENDIF
            ENDIF
            CALL write_particle_coordinates(particle_set, traj_unit, outformat, TRIM(id_wpc), TRIM(title), cell, &
                                            array=fml_array)
            DEALLOCATE (fml_array)
         ELSE
            CALL write_particle_coordinates(particle_set, traj_unit, outformat, TRIM(id_wpc), TRIM(title), cell, &
                                            unit_conv=unit_conv, &
                                            charge_occup=charge_occup, &
                                            charge_beta=charge_beta, &
                                            charge_extended=charge_extended)
         END IF
      END IF

      CALL cp_print_key_finished_output(traj_unit, logger, root_section, "MOTION%PRINT%"//TRIM(my_pk_name))

      CALL timestop(handle)

   END SUBROUTINE write_trajectory

! **************************************************************************************************
!> \brief Info on the unit to be opened to dump MD informations
!> \param section ...
!> \param path ...
!> \param my_form ...
!> \param my_ext ...
!> \author Teodoro Laino - University of Zurich - 07.2007
! **************************************************************************************************
   SUBROUTINE get_output_format(section, path, my_form, my_ext)

      TYPE(section_vals_type), POINTER                   :: section
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: path
      CHARACTER(LEN=*), INTENT(OUT)                      :: my_form, my_ext

      CHARACTER(len=*), PARAMETER :: routineN = 'get_output_format', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: output_format

      IF (PRESENT(path)) THEN
         CALL section_vals_val_get(section, TRIM(path)//"%FORMAT", i_val=output_format)
      ELSE
         CALL section_vals_val_get(section, "FORMAT", i_val=output_format)
      END IF

      SELECT CASE (output_format)
      CASE (dump_dcd, dump_dcd_aligned_cell)
         my_form = "UNFORMATTED"
         my_ext = ".dcd"
      CASE (dump_pdb)
         my_form = "FORMATTED"
         my_ext = ".pdb"
      CASE DEFAULT
         my_form = "FORMATTED"
         my_ext = ".xyz"
      END SELECT

   END SUBROUTINE get_output_format

! **************************************************************************************************
!> \brief   Prints the Stress Tensor
!> \param virial ...
!> \param cell ...
!> \param motion_section ...
!> \param itimes ...
!> \param time ...
!> \param pos ...
!> \param act ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_stress_tensor(virial, cell, motion_section, itimes, time, pos, &
                                  act)

      TYPE(virial_type), POINTER                         :: virial
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: motion_section
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: pos, act

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_stress_tensor', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_act, my_pos
      INTEGER                                            :: output_unit
      LOGICAL                                            :: new_file
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_total_bar
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (virial%pv_availability) THEN
         my_pos = "APPEND"
         my_act = "WRITE"
         IF (PRESENT(pos)) my_pos = pos
         IF (PRESENT(act)) my_act = act
         output_unit = cp_print_key_unit_nr(logger, motion_section, "PRINT%STRESS", &
                                            extension=".stress", file_position=my_pos, &
                                            file_action=my_act, file_form="FORMATTED", &
                                            is_new_file=new_file)
      ELSE
         output_unit = 0
      END IF

      IF (output_unit > 0) THEN
         IF (new_file) THEN
            WRITE (UNIT=output_unit, FMT='(A,9(12X,A2," [bar]"),6X,A)') &
               "#   Step   Time [fs]", "xx", "xy", "xz", "yx", "yy", "yz", "zx", "zy", "zz"
         END IF
         pv_total_bar(1, 1) = cp_unit_from_cp2k(virial%pv_total(1, 1)/cell%deth, "bar")
         pv_total_bar(1, 2) = cp_unit_from_cp2k(virial%pv_total(1, 2)/cell%deth, "bar")
         pv_total_bar(1, 3) = cp_unit_from_cp2k(virial%pv_total(1, 3)/cell%deth, "bar")
         pv_total_bar(2, 1) = cp_unit_from_cp2k(virial%pv_total(2, 1)/cell%deth, "bar")
         pv_total_bar(2, 2) = cp_unit_from_cp2k(virial%pv_total(2, 2)/cell%deth, "bar")
         pv_total_bar(2, 3) = cp_unit_from_cp2k(virial%pv_total(2, 3)/cell%deth, "bar")
         pv_total_bar(3, 1) = cp_unit_from_cp2k(virial%pv_total(3, 1)/cell%deth, "bar")
         pv_total_bar(3, 2) = cp_unit_from_cp2k(virial%pv_total(3, 2)/cell%deth, "bar")
         pv_total_bar(3, 3) = cp_unit_from_cp2k(virial%pv_total(3, 3)/cell%deth, "bar")
         WRITE (UNIT=output_unit, FMT='(I8,F12.3,9(1X,F19.10))') itimes, time, &
            pv_total_bar(1, 1), pv_total_bar(1, 2), pv_total_bar(1, 3), &
            pv_total_bar(2, 1), pv_total_bar(2, 2), pv_total_bar(2, 3), &
            pv_total_bar(3, 1), pv_total_bar(3, 2), pv_total_bar(3, 3)
         CALL m_flush(output_unit)
      END IF

      IF (virial%pv_availability) THEN
         CALL cp_print_key_finished_output(output_unit, logger, motion_section, &
                                           "PRINT%STRESS")
      END IF

   END SUBROUTINE write_stress_tensor

! **************************************************************************************************
!> \brief   Prints the Simulation Cell
!> \param cell ...
!> \param motion_section ...
!> \param itimes ...
!> \param time ...
!> \param pos ...
!> \param act ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_simulation_cell(cell, motion_section, itimes, time, pos, act)

      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: motion_section
      INTEGER, INTENT(IN)                                :: itimes
      REAL(KIND=dp), INTENT(IN)                          :: time
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: pos, act

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_simulation_cell', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_act, my_pos
      INTEGER                                            :: output_unit
      LOGICAL                                            :: new_file
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      my_pos = "APPEND"
      my_act = "WRITE"
      IF (PRESENT(pos)) my_pos = pos
      IF (PRESENT(act)) my_act = act

      output_unit = cp_print_key_unit_nr(logger, motion_section, "PRINT%CELL", &
                                         extension=".cell", file_position=my_pos, &
                                         file_action=my_act, file_form="FORMATTED", &
                                         is_new_file=new_file)

      IF (output_unit > 0) THEN
         IF (new_file) THEN
            WRITE (UNIT=output_unit, FMT='(A,9(7X,A2," [Angstrom]"),6X,A)') &
               "#   Step   Time [fs]", "Ax", "Ay", "Az", "Bx", "By", "Bz", "Cx", "Cy", "Cz", &
               "Volume [Angstrom^3]"
         END IF
         WRITE (UNIT=output_unit, FMT="(I8,F12.3,9(1X,F19.10),1X,F24.10)") itimes, time, &
            cell%hmat(1, 1)*angstrom, cell%hmat(2, 1)*angstrom, cell%hmat(3, 1)*angstrom, &
            cell%hmat(1, 2)*angstrom, cell%hmat(2, 2)*angstrom, cell%hmat(3, 2)*angstrom, &
            cell%hmat(1, 3)*angstrom, cell%hmat(2, 3)*angstrom, cell%hmat(3, 3)*angstrom, &
            cell%deth*angstrom*angstrom*angstrom
         CALL m_flush(output_unit)
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, motion_section, &
                                        "PRINT%CELL")

   END SUBROUTINE write_simulation_cell

END MODULE motion_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Driver for the localization that should be general
!>      for all the methods available and all the definition of the
!>      spread functional
!>      Write centers, spread and cubes only if required and for the
!>      selected states
!>      The localized functions are copied in the standard mos array
!>      for the next use
!> \par History
!>      01.2008 Teodoro Laino [tlaino] - University of Zurich
!>        - Merging the two localization codes and updating to new structures
!> \author MI (04.2005)
! **************************************************************************************************
MODULE qs_loc_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              deallocate_atomic_kind_set,&
                                              set_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_create_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_element, cp_fm_get_info, cp_fm_get_submatrix, cp_fm_p_type, &
        cp_fm_release, cp_fm_set_all, cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_iter_string,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_constants,                 ONLY: &
        do_loc_crazy, do_loc_direct, do_loc_jacobi, do_loc_l1_norm_sd, do_loc_none, dump_dcd, &
        dump_dcd_aligned_cell, dump_xmol, op_loc_berry, op_loc_boys, op_loc_pipek, state_loc_list
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_ivals,&
                                              section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              sp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE motion_utils,                    ONLY: get_output_format
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: get_particle_set,&
                                              write_particle_coordinates
   USE particle_types,                  ONLY: allocate_particle_set,&
                                              deallocate_particle_set,&
                                              particle_type
   USE physcon,                         ONLY: angstrom
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_loc_types,                    ONLY: get_qs_loc_env,&
                                              localized_wfn_control_type,&
                                              qs_loc_env_new_type
   USE qs_loc_utils,                    ONLY: jacobi_rotation_pipek
   USE qs_localization_methods,         ONLY: approx_l1_norm_sd,&
                                              crazy_rotations,&
                                              direct_mini,&
                                              jacobi_rotations
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE string_utilities,                ONLY: xstring
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_loc_methods'

! *** Public ***
   PUBLIC ::  qs_loc_driver, qs_print_cubes

CONTAINS

! **************************************************************************************************
!> \brief Calculate and optimize the spread functional as calculated from
!>       the selected mos  and by the definition using the berry phase
!>       as given by silvestrelli et al
!>       If required the centers and the spreads for each mos selected
!>       are calculated from z_ij and printed to file.
!>       The centers files is appended if already exists
!> \param method indicates localization algorithm
!> \param qs_loc_env new environment for the localization calculations
!> \param vectors selected mos to be localized
!> \param op_sm_set sparse matrices containing the integrals of the kind <mi e{iGr} nu>
!> \param zij_fm_set set of full matrix of size nmoloc x nmoloc, will contain the z_ij numbers
!>                    as defined by Silvestrelli et al
!> \param para_env ...
!> \param cell ...
!> \param weights ...
!> \param ispin ...
!> \param print_loc_section ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
!> \note
!>       This definition need the use of complex numbers, therefore the
!>       optimization routines are specific for this case
!>       The file for the centers and the spreads have a xyz format
! **************************************************************************************************
   SUBROUTINE optimize_loc_berry(method, qs_loc_env, vectors, op_sm_set, &
                                 zij_fm_set, para_env, cell, weights, ispin, print_loc_section)

      INTEGER, INTENT(IN)                                :: method
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cp_fm_type), POINTER                          :: vectors
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_sm_set
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: zij_fm_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp), DIMENSION(:)                             :: weights
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(section_vals_type), POINTER                   :: print_loc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'optimize_loc_berry', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, max_iter, nao, nmoloc, &
                                                            out_each, output_unit, sweeps
      LOGICAL                                            :: converged, crazy_use_diag
      REAL(dp)                                           :: crazy_scale, eps_localization, &
                                                            max_crazy_angle, start_time, &
                                                            target_time
      TYPE(cp_fm_type), POINTER                          :: opvec
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, print_loc_section, "PROGRAM_RUN_INFO", &
                                         extension=".locInfo")

      NULLIFY (opvec)
      ! get rows and cols of the input
      CALL cp_fm_get_info(vectors, nrow_global=nao, ncol_global=nmoloc)
      ! replicate the input kind of matrix
      CALL cp_fm_create(opvec, vectors%matrix_struct)

      ! Compute zij here
      DO i = 1, SIZE(zij_fm_set, 2)
         DO j = 1, SIZE(zij_fm_set, 1)
            CALL cp_fm_set_all(zij_fm_set(j, i)%matrix, 0.0_dp)
            CALL cp_dbcsr_sm_fm_multiply(op_sm_set(j, i)%matrix, vectors, opvec, ncol=nmoloc)
            CALL cp_gemm("T", "N", nmoloc, nmoloc, nao, 1.0_dp, vectors, opvec, 0.0_dp, &
                         zij_fm_set(j, i)%matrix)
         ENDDO
      ENDDO

      max_iter = qs_loc_env%localized_wfn_control%max_iter
      max_crazy_angle = qs_loc_env%localized_wfn_control%max_crazy_angle
      crazy_use_diag = qs_loc_env%localized_wfn_control%crazy_use_diag
      crazy_scale = qs_loc_env%localized_wfn_control%crazy_scale
      eps_localization = qs_loc_env%localized_wfn_control%eps_localization
      out_each = qs_loc_env%localized_wfn_control%out_each
      target_time = qs_loc_env%target_time
      start_time = qs_loc_env%start_time
      CALL centers_spreads_berry(qs_loc_env, zij_fm_set, nmoloc, cell, weights, &
                                 ispin, print_loc_section, only_initial_out=.TRUE.)
      SELECT CASE (method)
      CASE (do_loc_jacobi)
         CALL jacobi_rotations(weights, zij_fm_set, vectors, para_env, max_iter=max_iter, &
                               eps_localization=eps_localization, sweeps=sweeps, &
                               out_each=out_each, target_time=target_time, start_time=start_time)
      CASE (do_loc_crazy)
         CALL crazy_rotations(weights, zij_fm_set, vectors, max_iter=max_iter, max_crazy_angle=max_crazy_angle, &
                              crazy_scale=crazy_scale, crazy_use_diag=crazy_use_diag, &
                              eps_localization=eps_localization, iterations=sweeps, converged=converged)
         ! Possibly fallback to jacobi if the crazy rotation fails
         IF (.NOT. converged) THEN
            IF (qs_loc_env%localized_wfn_control%jacobi_fallback) THEN
               IF (output_unit > 0) WRITE (output_unit, "(T4,A,I6,/,T4,A)") &
                  " Crazy Wannier localization not converged after ", sweeps, &
                  " iterations, switching to jacobi rotations"
               CALL jacobi_rotations(weights, zij_fm_set, vectors, para_env, max_iter=max_iter, &
                                     eps_localization=eps_localization, sweeps=sweeps, &
                                     out_each=out_each, target_time=target_time, start_time=start_time)
            ELSE
               IF (output_unit > 0) WRITE (output_unit, "(T4,A,I6,/,T4,A)") &
                  " Crazy Wannier localization not converged after ", sweeps, &
                  " iterations, and jacobi_fallback switched off "
            ENDIF
         END IF
      CASE (do_loc_direct)
         CALL direct_mini(weights, zij_fm_set, vectors, max_iter=max_iter, &
                          eps_localization=eps_localization, iterations=sweeps)
      CASE (do_loc_l1_norm_sd)
         IF (.NOT. cell%orthorhombic) THEN
            CALL cp__b("qs_loc_methods.F",228,"Non-orthorhombic cell with the selected method NYI")
         ELSE
            CALL approx_l1_norm_sd(vectors, max_iter, eps_localization, converged, sweeps)
            ! here we need to set zij for the computation of the centers and spreads
            DO i = 1, SIZE(zij_fm_set, 2)
               DO j = 1, SIZE(zij_fm_set, 1)
                  CALL cp_fm_set_all(zij_fm_set(j, i)%matrix, 0.0_dp)
                  CALL cp_dbcsr_sm_fm_multiply(op_sm_set(j, i)%matrix, vectors, opvec, ncol=nmoloc)
                  CALL cp_gemm("T", "N", nmoloc, nmoloc, nao, 1.0_dp, vectors, opvec, 0.0_dp, &
                               zij_fm_set(j, i)%matrix)
               ENDDO
            ENDDO
         END IF
      CASE (do_loc_none)
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T4,A,I6,A)') " No MOS localization applied "
         ENDIF
      CASE DEFAULT
         CALL cp__b("qs_loc_methods.F",246,"Unknown localization method")
      END SELECT
      IF (output_unit > 0) THEN
         IF (sweeps <= max_iter) WRITE (output_unit, '(T4,A,I3,A,I6,A)') " Localization  for spin ", ispin, &
            " converged in ", sweeps, " iterations"
      ENDIF

      CALL centers_spreads_berry(qs_loc_env, zij_fm_set, nmoloc, cell, weights, &
                                 ispin, print_loc_section)
      CALL cp_fm_release(opvec)
      CALL cp_print_key_finished_output(output_unit, logger, print_loc_section, "PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE optimize_loc_berry

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param method ...
!> \param qs_loc_env ...
!> \param vectors ...
!> \param zij_fm_set ...
!> \param ispin ...
!> \param print_loc_section ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE optimize_loc_pipek(qs_env, method, qs_loc_env, vectors, zij_fm_set, &
                                 ispin, print_loc_section)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: method
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cp_fm_type), POINTER                          :: vectors
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: zij_fm_set
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(section_vals_type), POINTER                   :: print_loc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'optimize_loc_pipek', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, isgf, ldz, nao, natom, &
                                                            ncol, nmoloc, output_unit, sweeps
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf, nsgf
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_ao_fm_pools
      TYPE(cp_fm_type), POINTER                          :: opvec, ov_fm
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, print_loc_section, "PROGRAM_RUN_INFO", &
                                         extension=".locInfo")

      NULLIFY (particle_set)
      ! get rows and cols of the input
      CALL cp_fm_get_info(vectors, nrow_global=nao, ncol_global=nmoloc)
      ! replicate the input kind of matrix
      CALL cp_fm_create(opvec, vectors%matrix_struct)
      CALL cp_fm_set_all(opvec, 0.0_dp)

      CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s, &
                      particle_set=particle_set, qs_kind_set=qs_kind_set)

      natom = SIZE(particle_set, 1)
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))
      ALLOCATE (nsgf(natom))

      !   construction of
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, last_sgf=last_sgf, nsgf=nsgf)

      !   Copy the overlap sparse matrix in a full matrix
      CALL mpools_get(qs_env%mpools, ao_ao_fm_pools=ao_ao_fm_pools)
      CALL fm_pool_create_fm(ao_ao_fm_pools(1)%pool, ov_fm, name=" ")
      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, ov_fm)

      !   Compute zij here
      DO iatom = 1, natom
         CALL cp_fm_set_all(zij_fm_set(iatom, 1)%matrix, 0.0_dp)
         CALL cp_fm_get_info(zij_fm_set(iatom, 1)%matrix, ncol_global=ldz)
         isgf = first_sgf(iatom)
         ncol = nsgf(iatom)
         ! multiply fmxfm, using only part of the ao : Ct x S
         CALL cp_gemm('N', 'N', nao, nmoloc, nao, 1.0_dp, ov_fm, vectors, 0.0_dp, opvec, &
                      a_first_col=1, a_first_row=1, b_first_col=1, b_first_row=1)
         CALL cp_gemm('T', 'N', nmoloc, nmoloc, ncol, 0.5_dp, vectors, opvec, &
                      0.0_dp, zij_fm_set(iatom, 1)%matrix, &
                      a_first_col=1, a_first_row=isgf, b_first_col=1, b_first_row=isgf)

         CALL cp_gemm('N', 'N', nao, nmoloc, ncol, 1.0_dp, ov_fm, vectors, 0.0_dp, opvec, &
                      a_first_col=1, a_first_row=isgf, b_first_col=1, b_first_row=isgf)
         CALL cp_gemm('T', 'N', nmoloc, nmoloc, nao, 0.5_dp, vectors, opvec, &
                      1.0_dp, zij_fm_set(iatom, 1)%matrix, &
                      a_first_col=1, a_first_row=1, b_first_col=1, b_first_row=1)
      END DO ! iatom

      !   And now perform the optimization and rotate the orbitals
      SELECT CASE (method)
      CASE (do_loc_jacobi)
         CALL jacobi_rotation_pipek(zij_fm_set, vectors, sweeps)
      CASE (do_loc_crazy)
         CALL cp__b("qs_loc_methods.F",352,"Crazy and Pipek not implemented.")
      CASE (do_loc_l1_norm_sd)
         CALL cp__b("qs_loc_methods.F",354,"L1 norm and Pipek not implemented.")
      CASE (do_loc_direct)
         CALL cp__b("qs_loc_methods.F",356,"Direct and Pipek not implemented.")
      CASE (do_loc_none)
         IF (output_unit > 0) WRITE (output_unit, '(A,I6,A)') " No MOS localization applied "
      CASE DEFAULT
         CALL cp__b("qs_loc_methods.F",360,"Unknown localization method")
      END SELECT

      IF (output_unit > 0) WRITE (output_unit, '(A,I3,A,I6,A)') " Localization  for spin ", ispin, &
         " converged in ", sweeps, " iterations"

      CALL centers_spreads_pipek(qs_loc_env, zij_fm_set, particle_set, ispin, &
                                 print_loc_section)

      DEALLOCATE (first_sgf, last_sgf, nsgf)

      CALL cp_fm_release(opvec)
      CALL cp_print_key_finished_output(output_unit, logger, print_loc_section, "PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE optimize_loc_pipek

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \param zij ...
!> \param nmoloc ...
!> \param cell ...
!> \param weights ...
!> \param ispin ...
!> \param print_loc_section ...
!> \param only_initial_out ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE centers_spreads_berry(qs_loc_env, zij, nmoloc, cell, weights, ispin, &
                                    print_loc_section, only_initial_out)
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cp_fm_p_type), INTENT(INOUT)                  :: zij(:, :)
      INTEGER, INTENT(IN)                                :: nmoloc
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp), DIMENSION(:)                             :: weights
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(section_vals_type), POINTER                   :: print_loc_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: only_initial_out

      CHARACTER(len=*), PARAMETER :: routineN = 'centers_spreads_berry', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: file_tmp, iter
      COMPLEX(KIND=dp)                                   :: z
      INTEGER                                            :: idir, istate, jdir, nstates, &
                                                            output_unit, unit_out_s
      LOGICAL                                            :: my_only_init
      REAL(dp)                                           :: spread_i, spread_ii, sum_spread_i, &
                                                            sum_spread_ii
      REAL(dp), DIMENSION(3)                             :: c, c2, cpbc
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers
      REAL(KIND=dp)                                      :: imagpart, realpart
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (centers, logger, print_key)
      logger => cp_get_default_logger()
      my_only_init = .FALSE.
      IF (PRESENT(only_initial_out)) my_only_init = only_initial_out

      file_tmp = TRIM(qs_loc_env%tag_mo)//"_spreads_s"//TRIM(ADJUSTL(cp_to_string(ispin)))
      output_unit = cp_print_key_unit_nr(logger, print_loc_section, "PROGRAM_RUN_INFO", &
                                         extension=".locInfo")
      unit_out_s = cp_print_key_unit_nr(logger, print_loc_section, "WANNIER_SPREADS", &
                                        middle_name=file_tmp, extension=".data")
      iter = cp_iter_string(logger%iter_info)
      IF (unit_out_s > 0 .AND. .NOT. my_only_init) WRITE (unit_out_s, '(i6,/,A)') nmoloc, TRIM(iter)

      CALL cp_fm_get_info(zij(1, 1)%matrix, nrow_global=nstates)
      IF(.NOT.(nstates >= nmoloc))CALL cp__a("qs_loc_methods.F",433)

      centers => qs_loc_env%localized_wfn_control%centers_set(ispin)%array
      IF(.NOT.(ASSOCIATED(centers)))CALL cp__a("qs_loc_methods.F",436)
      IF(.NOT.(SIZE(centers, 2) == nmoloc))CALL cp__a("qs_loc_methods.F",437)
      sum_spread_i = 0.0_dp
      sum_spread_ii = 0.0_dp
      DO istate = 1, nmoloc
         c = 0.0_dp
         c2 = 0.0_dp
         spread_i = 0.0_dp
         spread_ii = 0.0_dp
         DO jdir = 1, SIZE(zij, 2)
            CALL cp_fm_get_element(zij(1, jdir)%matrix, istate, istate, realpart)
            CALL cp_fm_get_element(zij(2, jdir)%matrix, istate, istate, imagpart)
            z = CMPLX(realpart, imagpart, dp)
            spread_i = spread_i-weights(jdir)* &
                       LOG(realpart*realpart+imagpart*imagpart)/twopi/twopi
            spread_ii = spread_ii+weights(jdir)* &
                        (1.0_dp-(realpart*realpart+imagpart*imagpart))/twopi/twopi
            IF (jdir < 4) THEN
               DO idir = 1, 3
                  c(idir) = c(idir)+ &
                            (cell%hmat(idir, jdir)/twopi)*AIMAG(LOG(z))
               ENDDO
            END IF
         END DO
         cpbc = pbc(c, cell)
         centers(1:3, istate) = cpbc(1:3)
         centers(4, istate) = spread_i
         centers(5, istate) = spread_ii
         sum_spread_i = sum_spread_i+centers(4, istate)
         sum_spread_ii = sum_spread_ii+centers(5, istate)
         IF (unit_out_s > 0 .AND. .NOT. my_only_init) WRITE (unit_out_s, '(I6,2F16.8)') istate, centers(4:5, istate)
      ENDDO

      ! Print of wannier centers
      print_key => section_vals_get_subs_vals(print_loc_section, "WANNIER_CENTERS")
      IF (.NOT. my_only_init) CALL print_wannier_centers(qs_loc_env, print_key, centers, logger, ispin)

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T4, A, 2x, A26, A26)') " Spread Functional ", "sum_in -w_i ln(|z_in|^2)", &
            "sum_in w_i(1-|z_in|^2)"
         IF (my_only_init) THEN
            WRITE (output_unit, '(T4,A,T38,2F20.10)') " Initial Spread (Berry) : ", sum_spread_i, sum_spread_ii
         ELSE
            WRITE (output_unit, '(T4,A,T38,2F20.10)') " Total Spread (Berry) : ", sum_spread_i, sum_spread_ii
         END IF
      END IF

      IF (unit_out_s > 0 .AND. .NOT. my_only_init) WRITE (unit_out_s, '(A,2F16.10)') " Total ", sum_spread_i, sum_spread_ii

      CALL cp_print_key_finished_output(unit_out_s, logger, print_loc_section, "WANNIER_SPREADS")
      CALL cp_print_key_finished_output(output_unit, logger, print_loc_section, "PROGRAM_RUN_INFO")

   END SUBROUTINE centers_spreads_berry

! **************************************************************************************************
!> \brief define and print the centers and spread
!>       when the pipek operator is used
!> \param qs_loc_env ...
!> \param zij_fm_set matrix elements that define the populations on atoms
!> \param particle_set ...
!> \param ispin spin 1 or 2
!> \param print_loc_section ...
!> \par History
!>      05.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE centers_spreads_pipek(qs_loc_env, zij_fm_set, particle_set, ispin, &
                                    print_loc_section)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: zij_fm_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: ispin
      TYPE(section_vals_type), POINTER                   :: print_loc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'centers_spreads_pipek', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: file_tmp, iter
      INTEGER                                            :: iatom, istate, natom, nstate, unit_out_s
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_state
      REAL(dp)                                           :: r(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: Qii, ziimax
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: diag
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (centers, logger, print_key)
      logger => cp_get_default_logger()

      CALL cp_fm_get_info(zij_fm_set(1, 1)%matrix, nrow_global=nstate)
      natom = SIZE(zij_fm_set, 1)

      centers => qs_loc_env%localized_wfn_control%centers_set(ispin)%array
      IF(.NOT.(ASSOCIATED(centers)))CALL cp__a("qs_loc_methods.F",530)
      IF(.NOT.(SIZE(centers, 2) == nstate))CALL cp__a("qs_loc_methods.F",531)

      file_tmp = TRIM(qs_loc_env%tag_mo)//"_spreads_s"//TRIM(ADJUSTL(cp_to_string(ispin)))
      unit_out_s = cp_print_key_unit_nr(logger, print_loc_section, "WANNIER_SPREADS", &
                                        middle_name=file_tmp, extension=".data", log_filename=.FALSE.)
      iter = cp_iter_string(logger%iter_info)
      IF (unit_out_s > 0) WRITE (unit_out_s, '(i6,/,A)') TRIM(iter)

      ALLOCATE (atom_of_state(nstate))
      atom_of_state = 0

      ALLOCATE (diag(nstate, natom))
      diag = 0.0_dp

      DO iatom = 1, natom
         DO istate = 1, nstate
            CALL cp_fm_get_element(zij_fm_set(iatom, 1)%matrix, istate, istate, diag(istate, iatom))
         END DO
      END DO

      ALLOCATE (Qii(nstate), ziimax(nstate))
      ziimax = 0.0_dp
      Qii = 0.0_dp

      DO iatom = 1, natom
         DO istate = 1, nstate
            Qii(istate) = Qii(istate)+diag(istate, iatom)*diag(istate, iatom)
            IF (ABS(diag(istate, iatom)) > ziimax(istate)) THEN
               ziimax(istate) = ABS(diag(istate, iatom))
               atom_of_state(istate) = iatom
            END IF
         END DO
      END DO

      DO istate = 1, nstate
         iatom = atom_of_state(istate)
         r(1:3) = particle_set(iatom)%r(1:3)
         centers(1:3, istate) = r(1:3)
         centers(4, istate) = 1.0_dp/Qii(istate)
         IF (unit_out_s > 0) WRITE (unit_out_s, '(I6,F16.8)') istate, angstrom*centers(4, istate)
      END DO

      ! Print the wannier centers
      print_key => section_vals_get_subs_vals(print_loc_section, "WANNIER_CENTERS")
      CALL print_wannier_centers(qs_loc_env, print_key, centers, logger, ispin)

      CALL cp_print_key_finished_output(unit_out_s, logger, print_loc_section, "WANNIER_SPREADS")

      DEALLOCATE (Qii, ziimax, atom_of_state, diag)

   END SUBROUTINE centers_spreads_pipek

! **************************************************************************************************
!> \brief set up the calculation of localized orbitals
!> \param qs_env ...
!> \param qs_loc_env ...
!> \param print_loc_section ...
!> \param myspin ...
!> \param ext_mo_coeff ...
!> \par History
!>      04.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE qs_loc_driver(qs_env, qs_loc_env, print_loc_section, myspin, &
                            ext_mo_coeff)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: print_loc_section
      INTEGER, INTENT(IN), OPTIONAL                      :: myspin
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: ext_mo_coeff

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_driver', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_pos
      INTEGER                                            :: dim_op, handle, i, imo, imoloc, ir, &
                                                            ispin, istate, j, jstate, l_spin, lb, &
                                                            loc_method, n_rep, nao, ncubes, nmo, &
                                                            nmosub, s_spin, ub
      INTEGER, DIMENSION(:), POINTER                     :: bounds, list, list_cubes
      LOGICAL                                            :: append_cube, list_cubes_setup
      LOGICAL, SAVE                                      :: first_time = .TRUE.
      REAL(dp), DIMENSION(6)                             :: weights
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: centers, vecbuffer
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_sm_set
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: moloc_coeff
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: op_fm_set
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)
      NULLIFY (para_env, mos, dft_control, list_cubes)
      NULLIFY (cell, localized_wfn_control, moloc_coeff, op_sm_set, op_fm_set)
      qs_loc_env%first_time = first_time
      qs_loc_env%target_time = qs_env%target_time
      qs_loc_env%start_time = qs_env%start_time

      CALL get_qs_loc_env(qs_loc_env=qs_loc_env, &
                          localized_wfn_control=localized_wfn_control, &
                          moloc_coeff=moloc_coeff, op_sm_set=op_sm_set, op_fm_set=op_fm_set, cell=cell, &
                          weights=weights, dim_op=dim_op)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      para_env=para_env, mos=mos)

      s_spin = 1
      l_spin = dft_control%nspins
      IF (PRESENT(myspin)) THEN
         s_spin = myspin
         l_spin = myspin
      END IF

      DO ispin = s_spin, l_spin
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
         loc_method = localized_wfn_control%localization_method
         SELECT CASE (localized_wfn_control%operator_type)
         CASE (op_loc_berry)
            ! Here we allocate op_fm_set with the RIGHT size for uks
            NULLIFY (tmp_fm_struct, mo_coeff)
            nmosub = localized_wfn_control%nloc_states(ispin)
            IF (PRESENT(ext_mo_coeff)) THEN
               CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmosub, &
                                        ncol_global=nmosub, para_env=para_env, context=ext_mo_coeff%matrix_struct%context)
            ELSE
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
               CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmosub, &
                                        ncol_global=nmosub, para_env=para_env, context=mo_coeff%matrix_struct%context)
            END IF
            !
            ALLOCATE (op_fm_set(2, dim_op))
            DO i = 1, dim_op
               DO j = 1, SIZE(op_fm_set, 1)
                  NULLIFY (op_fm_set(j, i)%matrix)
                  CALL cp_fm_create(op_fm_set(j, i)%matrix, tmp_fm_struct)
                  CALL cp_fm_get_info(op_fm_set(j, i)%matrix, nrow_global=nmosub)
                  IF(.NOT.(nmo >= nmosub))CALL cp__a("qs_loc_methods.F",672)
                  CALL cp_fm_set_all(op_fm_set(j, i)%matrix, 0.0_dp)
               END DO
            END DO
            CALL cp_fm_struct_release(tmp_fm_struct)

            CALL optimize_loc_berry(loc_method, qs_loc_env, moloc_coeff(ispin)%matrix, op_sm_set, &
                                    op_fm_set, para_env, cell, weights, ispin, print_loc_section)

            ! Here we dealloctate op_fm_set
            IF (ASSOCIATED(op_fm_set)) THEN
               DO i = 1, dim_op
                  DO j = 1, SIZE(op_fm_set, 1)
                     CALL cp_fm_release(op_fm_set(j, i)%matrix)
                  END DO
               END DO
               DEALLOCATE (op_fm_set)
            END IF

         CASE (op_loc_boys)
            CALL cp__b("qs_loc_methods.F",692,"Boys localization not implemented")

         CASE (op_loc_pipek)
            CALL optimize_loc_pipek(qs_env, loc_method, qs_loc_env, moloc_coeff(ispin)%matrix, &
                                    op_fm_set, ispin, print_loc_section)

         END SELECT

         ! give back the localized orbitals
         IF (.NOT. PRESENT(ext_mo_coeff)) THEN
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         END IF

         lb = localized_wfn_control%lu_bound_states(1, ispin)
         ub = localized_wfn_control%lu_bound_states(2, ispin)

         IF (localized_wfn_control%set_of_states == state_loc_list) THEN
            ALLOCATE (vecbuffer(1, nao))
            nmosub = SIZE(localized_wfn_control%loc_states, 1)
            imoloc = 0
            DO i = lb, ub
               ! Get the index in the subset
               imoloc = imoloc+1
               ! Get the index in the full set
               imo = localized_wfn_control%loc_states(i, ispin)

               CALL cp_fm_get_submatrix(moloc_coeff(ispin)%matrix, vecbuffer, 1, imoloc, &
                                        nao, 1, transpose=.TRUE.)
               IF (PRESENT(ext_mo_coeff)) THEN
                  CALL cp_fm_set_submatrix(ext_mo_coeff, vecbuffer, 1, imo, &
                                           nao, 1, transpose=.TRUE.)
               ELSE
                  CALL cp_fm_set_submatrix(mo_coeff, vecbuffer, 1, imo, &
                                           nao, 1, transpose=.TRUE.)
               END IF
            END DO
            DEALLOCATE (vecbuffer)

         ELSE
            nmosub = localized_wfn_control%nloc_states(ispin)
            IF (PRESENT(ext_mo_coeff)) THEN
               CALL cp_fm_to_fm(moloc_coeff(ispin)%matrix, ext_mo_coeff, nmosub, 1, lb)
            ELSE
               CALL cp_fm_to_fm(moloc_coeff(ispin)%matrix, mo_coeff, nmosub, 1, lb)
            END IF
         END IF

         ! Write cube files if required
         IF (localized_wfn_control%print_cubes) THEN
            list_cubes_setup = .FALSE.
            NULLIFY (bounds, list, list_cubes)

            ! Provides boundaries of MOs
            CALL section_vals_val_get(print_loc_section, "WANNIER_CUBES%CUBES_LU_BOUNDS", &
                                      i_vals=bounds)
            ncubes = bounds(2)-bounds(1)+1
            IF (ncubes > 0) THEN
               list_cubes_setup = .TRUE.
               ALLOCATE (list_cubes(ncubes))
               DO ir = 1, ncubes
                  list_cubes(ir) = bounds(1)+(ir-1)
               END DO
            END IF

            ! Provides the list of MOs
            CALL section_vals_val_get(print_loc_section, "WANNIER_CUBES%CUBES_LIST", &
                                      n_rep_val=n_rep)
            IF (.NOT. list_cubes_setup) THEN
               ncubes = 0
               DO ir = 1, n_rep
                  CALL section_vals_val_get(print_loc_section, "WANNIER_CUBES%CUBES_LIST", &
                                            i_rep_val=ir, i_vals=list)
                  IF (ASSOCIATED(list)) THEN
                     CALL reallocate(list_cubes, 1, ncubes+SIZE(list))
                     DO i = 1, SIZE(list)
                        list_cubes(i+ncubes) = list(i)
                     END DO
                     ncubes = ncubes+SIZE(list)
                  END IF
               END DO
               IF (ncubes > 0) list_cubes_setup = .TRUE.
            END IF

            ! Full list of Mos
            IF (.NOT. list_cubes_setup) THEN
               list_cubes_setup = .TRUE.
               ncubes = localized_wfn_control%nloc_states(1)
               IF (ncubes > 0) THEN
                  ALLOCATE (list_cubes(ncubes))
               END IF
               DO i = 1, ncubes
                  list_cubes(i) = i
               END DO
            END IF

            ncubes = SIZE(list_cubes)
            ncubes = MIN(ncubes, nmo)
            ALLOCATE (centers(6, ncubes))
            DO i = 1, ncubes
               istate = list_cubes(i)
               DO j = 1, localized_wfn_control%nloc_states(ispin)
                  jstate = localized_wfn_control%loc_states(j, ispin)
                  IF (istate == jstate) THEN
                     centers(1:6, i) = localized_wfn_control%centers_set(ispin)%array(1:6, j)
                     EXIT
                  ENDIF
               END DO
            END DO ! ncubes

            ! Real call for dumping the cube files
            print_key => section_vals_get_subs_vals(print_loc_section, "WANNIER_CUBES")
            append_cube = section_get_lval(print_loc_section, "WANNIER_CUBES%APPEND")
            my_pos = "REWIND"
            IF (append_cube) THEN
               my_pos = "APPEND"
            END IF

            CALL qs_print_cubes(qs_env, moloc_coeff(ispin)%matrix, ncubes, list_cubes, centers, &
                                print_key, "loc"//TRIM(ADJUSTL(qs_loc_env%tag_mo)), &
                                ispin=ispin, file_position=my_pos)

            DEALLOCATE (centers)
            DEALLOCATE (list_cubes)
         END IF
      END DO ! ispin
      first_time = .FALSE.
      CALL timestop(handle)
   END SUBROUTINE qs_loc_driver

! **************************************************************************************************
!> \brief write the cube files for a set of selected states
!> \param qs_env ...
!> \param mo_coeff set mos from which the states to be printed are extracted
!> \param nstates number of states to be printed
!> \param state_list list of the indexes of the states to be printed
!> \param centers centers and spread, all=0 if they hva not been calculated
!> \param print_key ...
!> \param root initial part of the cube file names
!> \param ispin ...
!> \param idir ...
!> \param state0 ...
!> \param file_position ...
!> \par History
!>      08.2005 created [MI]
!> \author MI
!> \note
!>      This routine shoul not be in this module
! **************************************************************************************************
   SUBROUTINE qs_print_cubes(qs_env, mo_coeff, nstates, state_list, centers, &
                             print_key, root, ispin, idir, state0, file_position)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER, INTENT(IN)                                :: nstates
      INTEGER, DIMENSION(:), POINTER                     :: state_list
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers
      TYPE(section_vals_type), POINTER                   :: print_key
      CHARACTER(LEN=*)                                   :: root
      INTEGER, INTENT(IN), OPTIONAL                      :: ispin, idir
      INTEGER, OPTIONAL                                  :: state0
      CHARACTER(LEN=default_string_length), OPTIONAL     :: file_position

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_print_cubes', routineP = moduleN//':'//routineN
      CHARACTER, DIMENSION(3), PARAMETER                 :: labels = (/'x', 'y', 'z'/)

      CHARACTER                                          :: label
      CHARACTER(LEN=default_path_length)                 :: file_tmp, filename, my_pos
      CHARACTER(LEN=default_string_length)               :: title
      INTEGER                                            :: handle, ia, ie, istate, ivector, &
                                                            my_ispin, my_state0, unit_out_c
      LOGICAL                                            :: add_idir, add_spin
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_subsys_type), POINTER                      :: subsys

      CALL timeset(routineN, handle)
      NULLIFY (auxbas_pw_pool, pw_env, logger)
      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, subsys=subsys)
      CALL qs_subsys_get(subsys, particles=particles)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, wf_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      my_state0 = 0
      IF (PRESENT(state0)) my_state0 = state0

      add_spin = .FALSE.
      my_ispin = 1
      IF (PRESENT(ispin)) THEN
         add_spin = .TRUE.
         my_ispin = ispin
      END IF
      add_idir = .FALSE.
      IF (PRESENT(idir)) THEN
         add_idir = .TRUE.
         label = labels(idir)
      END IF

      my_pos = "REWIND"
      IF (PRESENT(file_position)) THEN
         my_pos = file_position
      END IF

      DO istate = 1, nstates
         ivector = state_list(istate)-my_state0
         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, cell=cell, &
                         dft_control=dft_control, particle_set=particle_set, pw_env=pw_env)

         CALL calculate_wavefunction(mo_coeff, ivector, wf_r, wf_g, atomic_kind_set, &
                                     qs_kind_set, cell, dft_control, particle_set, pw_env)

         ! Formatting the middle part of the name
         ivector = state_list(istate)
         CALL xstring(root, ia, ie)
         IF (add_idir) THEN
            filename = root(ia:ie)//"_"//label//"_w"//TRIM(ADJUSTL(cp_to_string(ivector)))
         ELSE
            filename = root(ia:ie)//"_w"//TRIM(ADJUSTL(cp_to_string(ivector)))
         END IF
         IF (add_spin) THEN
            file_tmp = filename
            CALL xstring(file_tmp, ia, ie)
            filename = file_tmp(ia:ie)//"_s"//TRIM(ADJUSTL(cp_to_string(ispin)))
         END IF

         ! Using the print_key tools to open the file with the proper name
         unit_out_c = cp_print_key_unit_nr(logger, print_key, "", middle_name=filename, &
                                           extension=".cube", file_position=my_pos, log_filename=.FALSE.)
         IF (SIZE(centers, 1) == 6) THEN
            WRITE (title, '(A7,I5.5,A2,I1.1,A1,6F10.4)') "WFN ", ivector, "_s", my_ispin, " ", &
               centers(1:3, istate)*angstrom, centers(4:6, istate)*angstrom
         ELSE
            WRITE (title, '(A7,I5.5,A2,I1.1,A1,3F10.4)') "WFN ", ivector, "_s", my_ispin, " ", &
               centers(1:3, istate)*angstrom
         END IF
         CALL cp_pw_to_cube(wf_r%pw, unit_out_c, title, &
                            particles=particles, &
                            stride=section_get_ivals(print_key, "STRIDE"))
         CALL cp_print_key_finished_output(unit_out_c, logger, print_key, "")
      END DO

      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)
      CALL timestop(handle)
   END SUBROUTINE qs_print_cubes

! **************************************************************************************************
!> \brief Prints wannier centers
!> \param qs_loc_env ...
!> \param print_key ...
!> \param center ...
!> \param logger ...
!> \param ispin ...
! **************************************************************************************************
   SUBROUTINE print_wannier_centers(qs_loc_env, print_key, center, logger, ispin)
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: print_key
      REAL(KIND=dp), INTENT(IN)                          :: center(:, :)
      TYPE(cp_logger_type), POINTER                      :: logger
      INTEGER, INTENT(IN)                                :: ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'print_wannier_centers', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: iter, unit_str
      CHARACTER(LEN=default_string_length)               :: my_ext, my_form
      INTEGER                                            :: iunit, l, nstates
      LOGICAL                                            :: first_time, init_traj
      REAL(KIND=dp)                                      :: unit_conv

      nstates = SIZE(center, 2)
      my_form = "formatted"
      my_ext = ".data"
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key, first_time=first_time) &
                , cp_p_file)) THEN
         ! Find out if we want to print IONS+CENTERS or ONLY CENTERS..
         IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key, "/IONS+CENTERS") &
                   , cp_p_file)) THEN
            CALL get_output_format(print_key, my_form=my_form, my_ext=my_ext)
         END IF
         IF (first_time .OR. (.NOT. qs_loc_env%first_time)) THEN
            iunit = cp_print_key_unit_nr(logger, print_key, "", extension=my_ext, file_form=my_form, &
                                         middle_name=TRIM(qs_loc_env%tag_mo)//"_centers_s"//TRIM(ADJUSTL(cp_to_string(ispin))), &
                                         log_filename=.FALSE., on_file=.TRUE., is_new_file=init_traj)
            IF (iunit > 0) THEN
               ! Gather units of measure for output (if available)
               CALL section_vals_val_get(print_key, "UNIT", c_val=unit_str)
               unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

               IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key, "/IONS+CENTERS"), cp_p_file)) THEN
                  ! Different possible formats
                  CALL print_wannier_traj(qs_loc_env, print_key, center, iunit, init_traj, unit_conv)
               ELSE
                  ! Default print format
                  iter = cp_iter_string(logger%iter_info)
                  WRITE (iunit, '(i6,/,a)') nstates, TRIM(iter)
                  DO l = 1, nstates
                     WRITE (iunit, '(A,4F16.8)') "X", unit_conv*center(1:4, l)
                  END DO
               END IF
            END IF
            CALL cp_print_key_finished_output(iunit, logger, print_key, on_file=.TRUE.)
         END IF
      END IF
   END SUBROUTINE print_wannier_centers

! **************************************************************************************************
!> \brief computes spread and centers
!> \param qs_loc_env ...
!> \param print_key ...
!> \param center ...
!> \param iunit ...
!> \param init_traj ...
!> \param unit_conv ...
! **************************************************************************************************
   SUBROUTINE print_wannier_traj(qs_loc_env, print_key, center, iunit, init_traj, unit_conv)
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: print_key
      REAL(KIND=dp), INTENT(IN)                          :: center(:, :)
      INTEGER, INTENT(IN)                                :: iunit
      LOGICAL, INTENT(IN)                                :: init_traj
      REAL(KIND=dp), INTENT(IN)                          :: unit_conv

      CHARACTER(len=*), PARAMETER :: routineN = 'print_wannier_traj', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: iter, remark1, remark2, title
      INTEGER                                            :: i, iskip, natom, ntot, outformat
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      NULLIFY (particle_set, atomic_kind_set, atomic_kind, logger)
      logger => cp_get_default_logger()
      natom = SIZE(qs_loc_env%particle_set)
      ntot = natom+SIZE(center, 2)
      CALL allocate_particle_set(particle_set, ntot)
      ALLOCATE (atomic_kind_set(1))
      atomic_kind => atomic_kind_set(1)
      CALL set_atomic_kind(atomic_kind=atomic_kind, kind_number=0, &
                           name="X", element_symbol="X", mass=0.0_dp)
      ! Particles
      DO i = 1, natom
         particle_set(i)%atomic_kind => qs_loc_env%particle_set(i)%atomic_kind
         particle_set(i)%r = pbc(qs_loc_env%particle_set(i)%r, qs_loc_env%cell)
      END DO
      ! Wannier Centers
      DO i = natom+1, ntot
         particle_set(i)%atomic_kind => atomic_kind
         particle_set(i)%r = pbc(center(1:3, i-natom), qs_loc_env%cell)
      END DO
      ! Dump the structure
      CALL section_vals_val_get(print_key, "FORMAT", i_val=outformat)

      ! Header file
      SELECT CASE (outformat)
      CASE (dump_dcd, dump_dcd_aligned_cell)
         IF (init_traj) THEN
            !Lets write the header for the coordinate dcd
            ! Note (TL) : even the new DCD format is unfortunately too poor
            !             for our capabilities.. for example here the printing
            !             of the geometry could be nested inside several iteration
            !             levels.. this cannot be exactly reproduce with DCD.
            !             Just as a compromise let's pick-up the value of the MD iteration
            !             level. In any case this is not any sensible information for the standard..
            iskip = section_get_ival(print_key, "EACH%MD")
            WRITE (iunit) "CORD", 0, -1, iskip, &
               0, 0, 0, 0, 0, 0, REAL(0, KIND=sp), 1, 0, 0, 0, 0, 0, 0, 0, 0, 24
            remark1 = "REMARK FILETYPE CORD DCD GENERATED BY CP2K"
            remark2 = "REMARK Support new DCD format with cell information"
            WRITE (iunit) 2, remark1, remark2
            WRITE (iunit) ntot
            CALL m_flush(iunit)
         END IF
      CASE (dump_xmol)
         iter = cp_iter_string(logger%iter_info)
         WRITE (UNIT=title, FMT="(A)") " Particles+Wannier centers. Iteration:"//TRIM(iter)
      CASE DEFAULT
         title = ""
      END SELECT
      CALL write_particle_coordinates(particle_set, iunit, outformat, "POS", title, qs_loc_env%cell, &
                                      unit_conv=unit_conv)
      CALL m_flush(iunit)
      CALL deallocate_particle_set(particle_set)
      CALL deallocate_atomic_kind_set(atomic_kind_set)
   END SUBROUTINE print_wannier_traj

END MODULE qs_loc_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Determine active space Hamiltonian
!> \par History
!>      04.2016 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE qs_active_space_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: &
        allocate_sto_basis_set, create_gto_from_sto_basis, deallocate_gto_basis_set, &
        deallocate_sto_basis_set, gto_basis_set_type, init_orb_basis_set, set_sto_basis_set, &
        srules, sto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_csr_create,&
                                              cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type,&
                                              csr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_element, cp_fm_get_info, cp_fm_p_type, cp_fm_release, &
        cp_fm_set_all, cp_fm_set_element, cp_fm_to_fm, cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: &
        cp_p_file, cp_print_key_finished_output, cp_print_key_should_output, cp_print_key_unit_nr, &
        debug_print_level, high_print_level, low_print_level, medium_print_level, &
        silent_print_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE input_constants,                 ONLY: &
        casci_canonical, eri_method_full_gpw, eri_method_gpw_ht, eri_operator_coulomb, &
        eri_operator_erf, eri_operator_erfc, eri_operator_gaussian, eri_operator_yukawa, &
        mao_projection, wannier_projection
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp,&
                                              int_8
   USE mathconstants,                   ONLY: fourpi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_environ,&
                                              mp_sum,&
                                              mp_sync
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE periodic_table,                  ONLY: ptable
   USE pw_env_methods,                  ONLY: pw_env_create,&
                                              pw_env_rebuild
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_release,&
                                              pw_env_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_integrate_function,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_rebuild,&
                                              pw_poisson_solve
   USE pw_poisson_types,                ONLY: ANALYTIC0D,&
                                              PERIODIC3D,&
                                              greens_fn_type,&
                                              pw_poisson_analytic,&
                                              pw_poisson_periodic,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE qs_active_space_types,           ONLY: active_space_type,&
                                              create_active_space_type,&
                                              eri_type
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_loc_methods,                  ONLY: qs_loc_driver
   USE qs_loc_types,                    ONLY: qs_loc_env_create,&
                                              qs_loc_env_new_type,&
                                              qs_loc_env_release
   USE qs_loc_utils,                    ONLY: qs_loc_control_init,&
                                              qs_loc_env_init,&
                                              qs_loc_init
   USE qs_mo_methods,                   ONLY: calculate_density_matrix,&
                                              calculate_subspace_eigenvalues
   USE qs_mo_types,                     ONLY: allocate_mo_set,&
                                              get_mo_set,&
                                              init_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE task_list_methods,               ONLY: generate_qs_task_list
   USE task_list_types,                 ONLY: allocate_task_list,&
                                              deallocate_task_list,&
                                              task_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_active_space_methods'

   PUBLIC ::  active_space_main

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param input ...
!> \param logger ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE active_space_main(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'active_space_main', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=10)                                  :: cshell, lnam(5)
      INTEGER :: eri_method, eri_operator, eri_print, handle, i, iatom, ishell, isp, ispin, iw, j, &
         jm, m, mselect, n1, n2, nactive, nao, natom, ncol, nel, ninactive, ninactive_alpha, &
         ninactive_beta, nmo, nn1, nn2, nspins, ntot, zval
      INTEGER, DIMENSION(2)                              :: nactive_orb, nepol, ninactive_orb
      INTEGER, DIMENSION(5)                              :: nshell
      INTEGER, DIMENSION(:), POINTER                     :: invals
      LOGICAL                                            :: do_kpoints, ex_operator, ex_perd, &
                                                            explicit, isolated, store_wfn
      REAL(KIND=dp) :: eref, eri_eps_grid, eri_eps_int, eri_gpw_cutoff, eri_op_param, eri_rcut, &
         eri_rel_cutoff, fel, focc, maxocc, nze_percentage
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: eigenvalues
      TYPE(active_space_type), POINTER                   :: active_space_env
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, rho_ao
      TYPE(cp_dbcsr_type), POINTER                       :: denmat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_localized
      TYPE(cp_fm_type), POINTER                          :: fm_ref, fm_target, fmat, mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(csr_type), POINTER                            :: eri_mat
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: pro_basis_set
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env_occ, qs_loc_env_virt
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: as_input, loc_print, loc_section, &
                                                            print_orb

      !--------------------------------------------------------------------------------------------!

      as_input => section_vals_get_subs_vals(input, "DFT%PRINT%ACTIVE_SPACE")
      CALL section_vals_get(as_input, explicit=explicit)
      IF (.NOT. explicit) RETURN
      CALL timeset(routineN, handle)

      iw = cp_logger_get_default_io_unit(logger)

      IF (iw > 0) THEN
         WRITE (iw, '(/,T2,A)') &
            '!-----------------------------------------------------------------------------!'
         WRITE (iw, '(T26,A)') "Generate Active Space Hamiltonian"
         WRITE (iw, '(T27,A)') "Interface to CAS-CI and DMRG-CI"
         WRITE (iw, '(T2,A)') &
            '!-----------------------------------------------------------------------------!'
      END IF

      ! k-points?
      CALL get_qs_env(qs_env, do_kpoints=do_kpoints)
      IF (do_kpoints) THEN
         CALL cp_abort(cp__l("qs_active_space_methods.F",206), "K-points not allowd for Active Space Interface")
      END IF

      NULLIFY (active_space_env)
      CALL create_active_space_type(active_space_env)

      ! input options

      ! isolated (molecular) system?
      CALL section_vals_val_get(as_input, "ISOLATED_SYSTEM", l_val=isolated)
      active_space_env%molecule = isolated
      IF (iw > 0) THEN
         IF (active_space_env%molecule) THEN
            WRITE (iw, '(T3,A)') "System is treated without periodicity"
         END IF
      END IF

      CALL section_vals_val_get(as_input, "ACTIVE_ELECTRONS", i_val=nactive)
      IF (nactive <= 0) CALL cp__b("qs_active_space_methods.F",224,"Specify a positive number of active electrons.")
      CALL get_qs_env(qs_env, nelectron_total=ntot, nelectron_spin=nepol)
      CALL get_qs_env(qs_env, dft_control=dft_control)
      nspins = dft_control%nspins
      CALL section_vals_val_get(as_input, "INACTIVE_ELECTRONS", explicit=explicit)
      CALL section_vals_val_get(as_input, "INACTIVE_ELECTRONS", i_vals=invals)
      IF (nspins > 1) THEN
         IF (.NOT. explicit) THEN
            CALL cp_abort(cp__l("qs_active_space_methods.F",232), "Number of Inactive Electrons has to be specified"// &
                          " in spin polarised case.")
         END IF
         ninactive_alpha = invals(1)
         ninactive_beta = invals(2)
         ninactive = ninactive_alpha+ninactive_beta
      ELSE
         IF (explicit) THEN
            ninactive = invals(1)
         ELSE
            ninactive = ntot-nactive
         END IF
         ninactive_alpha = ninactive
         ninactive_beta = 0
      END IF
      IF(.NOT.(ninactive >= 0))CALL cp__a("qs_active_space_methods.F",247)
      IF(.NOT.(ninactive_alpha >= 0))CALL cp__a("qs_active_space_methods.F",248)
      IF(.NOT.(ninactive_beta >= 0))CALL cp__a("qs_active_space_methods.F",249)
      IF(.NOT.(ntot == ninactive+nactive))CALL cp__a("qs_active_space_methods.F",250)
      IF (nspins > 1) THEN
         IF(.NOT.(nepol(1) >= ninactive_alpha))CALL cp__a("qs_active_space_methods.F",252)
         IF(.NOT.(nepol(2) >= ninactive_beta))CALL cp__a("qs_active_space_methods.F",253)
      END IF

      IF (iw > 0) THEN
         IF (nspins < 2) THEN
            WRITE (iw, '(T4,A,T69,I10)') "Total number of electrons", ntot
            WRITE (iw, '(T4,A,T69,I10)') "Number of active electrons", nactive
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive electrons", ninactive
         ELSE
            WRITE (iw, '(T4,A,T69,I10)') "Total number of electrons", ntot
            WRITE (iw, '(T4,A,T69,I10)') "Number of active electrons", nactive
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive electrons", ninactive
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive electrons (alpha)", ninactive_alpha
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive electrons (beta)", ninactive_beta
         END IF
      END IF
      active_space_env%nactive = nactive
      active_space_env%ninactive = ninactive
      IF (nspins > 1) THEN
         active_space_env%nelectrons(1) = ntot
         active_space_env%nelectrons(2) = 0
      ELSE
         active_space_env%nelectrons(1) = nepol(1)
         active_space_env%nelectrons(2) = nepol(2)
      END IF
      active_space_env%ninspin(1) = ninactive_alpha
      active_space_env%ninspin(2) = ninactive_beta
      active_space_env%multiplicity = dft_control%multiplicity
      active_space_env%nspins = nspins

      ! define the active/inactive space orbitals
      CALL section_vals_val_get(as_input, "ACTIVE_ORBITALS", explicit=explicit)
      CALL section_vals_val_get(as_input, "ACTIVE_ORBITALS", i_vals=invals)
      IF (.NOT. explicit) THEN
         CALL cp_abort(cp__l("qs_active_space_methods.F",287), "Number of Active Orbitals has to be specified.")
      END IF
      IF(.NOT.(SIZE(invals) >= nspins))CALL cp__a("qs_active_space_methods.F",289)
      DO ispin = 1, nspins
         nactive_orb(ispin) = invals(ispin)
      END DO
      ninactive_orb = 0
      CALL section_vals_val_get(as_input, "INACTIVE_ORBITALS", explicit=explicit)
      CALL section_vals_val_get(as_input, "INACTIVE_ORBITALS", i_vals=invals)
      IF (.NOT. explicit) THEN
         IF (nspins == 1) THEN
            ninactive_orb(1) = active_space_env%ninspin(1)/2
         ELSE
            DO ispin = 1, nspins
               ninactive_orb(ispin) = active_space_env%ninspin(ispin)
            END DO
         END IF
      ELSE
         IF(.NOT.(SIZE(invals) >= nspins))CALL cp__a("qs_active_space_methods.F",305)
         DO ispin = 1, nspins
            ninactive_orb(ispin) = invals(ispin)
         END DO
      END IF
      IF (nspins == 1) THEN
         IF(.NOT.(MOD(ninactive, 2) == 0))CALL cp__a("qs_active_space_methods.F",311)
         IF(.NOT.(ninactive_orb(1) >= ninactive/2))CALL cp__a("qs_active_space_methods.F",312)
      ELSE
         IF(.NOT.(ninactive_orb(1) >= ninactive_alpha))CALL cp__a("qs_active_space_methods.F",314)
         IF(.NOT.(ninactive_orb(2) >= ninactive_beta))CALL cp__a("qs_active_space_methods.F",315)
      END IF

      CALL get_qs_env(qs_env, mos=mos)
      IF (.NOT. ASSOCIATED(mos(1)%mo_set)) THEN
         CALL cp_abort(cp__l("qs_active_space_methods.F",320), "A reference MO set has to be available when determining"// &
                       " the active space orbitals.")
      END IF
      CALL get_mo_set(mos(1)%mo_set, mo_coeff=fm_ref, nao=nao)
      maxocc = 2.0_dp
      IF (nspins > 1) maxocc = 1.0_dp
      ALLOCATE (active_space_env%mos_active(nspins))
      ALLOCATE (active_space_env%mos_inactive(nspins))
      DO ispin = 1, nspins
         nmo = nactive_orb(ispin)
         NULLIFY (mo_set)
         CALL allocate_mo_set(mo_set, nao, nmo, 0, 0.0_dp, maxocc, 0.0_dp)
         CALL init_mo_set(mo_set, fm_ref=fm_ref, name="Active Space MO")
         active_space_env%mos_active(ispin)%mo_set => mo_set
         mo_set => active_space_env%mos_inactive(ispin)%mo_set
         nmo = ninactive_orb(ispin)
         nel = active_space_env%ninspin(ispin)
         NULLIFY (mo_set)
         CALL allocate_mo_set(mo_set, nao, nmo, nel, REAL(nel, KIND=dp), maxocc, 0.0_dp)
         CALL init_mo_set(mo_set, fm_ref=fm_ref, name="Inactive Space MO")
         active_space_env%mos_inactive(ispin)%mo_set => mo_set
      END DO

      CALL section_vals_val_get(as_input, "ORBITAL_SELECTION", i_val=mselect)
      IF (iw > 0) THEN
         SELECT CASE (mselect)
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",347,"Unknown orbital selection method")
         CASE (casci_canonical)
            WRITE (iw, '(/,T4,A)') &
               "Active space orbitals selected using energy ordered canonical orbitals"
         CASE (wannier_projection)
            WRITE (iw, '(/,T4,A)') &
               "Active space orbitals selected using projected Wannier orbitals"
         CASE (mao_projection)
            WRITE (iw, '(/,T4,A)') &
               "Active space orbitals selected using modified atomic orbitals (MAO)"
         END SELECT
         IF (nspins < 2) THEN
            WRITE (iw, '(T4,A,T69,I10)') "Number of active orbitals", nactive_orb(1)
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive orbitals", ninactive_orb(1)
         ELSE
            WRITE (iw, '(T4,A,T69,I10)') "Number of active orbitals", SUM(nactive_orb)
            WRITE (iw, '(T4,A,T69,I10)') "Number of active orbitals (alpha)", nactive_orb(1)
            WRITE (iw, '(T4,A,T69,I10)') "Number of active orbitals (beta)", nactive_orb(2)
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive orbitals", SUM(ninactive_orb)
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive orbitals (alpha)", ninactive_orb(1)
            WRITE (iw, '(T4,A,T69,I10)') "Number of inactive orbitals (beta)", ninactive_orb(2)
         END IF
      END IF

      ! get projection spaces
      CALL section_vals_val_get(as_input, "SUBSPACE_ATOM", i_val=iatom, explicit=explicit)
      IF (explicit) THEN
         CALL get_qs_env(qs_env, natom=natom)
         IF (iatom <= 0 .OR. iatom > natom) THEN
            IF (iw > 0) THEN
               WRITE (iw, '(/,T4,A,I3)') "ERROR: SUBSPACE_ATOM number is not valid", iatom
            END IF
            CALL cp__b("qs_active_space_methods.F",379,"Select a valid SUBSPACE_ATOM")
         END IF
      END IF
      CALL section_vals_val_get(as_input, "SUBSPACE_SHELL", c_val=cshell, explicit=explicit)
      nshell = 0
      lnam = ""
      IF (explicit) THEN
         cshell = ADJUSTL(cshell)
         n1 = 1
         DO i = 1, 5
            ishell = i
            IF (cshell(n1:n1) == " ") THEN
               ishell = ishell-1
               EXIT
            END IF
            READ (cshell(n1:), "(I1,A1)") nshell(i), lnam(i)
            n1 = n1+2
         END DO
      END IF

      ! generate orbitals
      SELECT CASE (mselect)
      CASE DEFAULT
         CALL cp__b("qs_active_space_methods.F",402,"Unknown orbital selection method")
      CASE (casci_canonical)
         ! create canonical orbitals
         IF (dft_control%restricted) THEN
            CALL cp__b("qs_active_space_methods.F",406," Unclear how we define MOs in the restricted case ... stopping")
         ELSE
            IF (dft_control%do_admm) THEN
               CALL cp__b("qs_active_space_methods.F",409,"ADMM currently not possible for canonical orbital options")
            END IF
            nmo = 0
            DO ispin = 1, nspins
               CALL get_mo_set(mos(ispin)%mo_set, nmo=m)
               IF (m < ninactive_orb(ispin)+nactive_orb(ispin)) THEN
                  CALL cp__b("qs_active_space_methods.F",415,"Not enough canonical orbitals available.")
               END IF
               nmo = MAX(m, nmo)
            END DO
            ALLOCATE (eigenvalues(nmo, nspins))
            eigenvalues = 100000._dp
            CALL get_qs_env(qs_env, matrix_ks=ks_matrix)
            IF (mos(1)%mo_set%use_mo_coeff_b) THEN
               CALL cp__b("qs_active_space_methods.F",423,"Needs MOs in FM format, missing code for DBCSR")
            ELSE
               DO ispin = 1, nspins
                  CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=fm_ref, nmo=nmo)
                  CALL calculate_subspace_eigenvalues(fm_ref, ks_matrix(ispin)%matrix, &
                                                      evals_arg=eigenvalues(:, ispin), do_rotation=.TRUE.)
                  ! copy inactive orbitals
                  mo_set => active_space_env%mos_inactive(ispin)%mo_set
                  CALL get_mo_set(mo_set, mo_coeff=fm_target)
                  ncol = ninactive_orb(ispin)
                  CALL cp_fm_to_fm(fm_ref, fm_target, ncol, 1, 1)
                  ! copy active orbitals
                  mo_set => active_space_env%mos_active(ispin)%mo_set
                  CALL get_mo_set(mo_set, mo_coeff=fm_target)
                  ncol = nactive_orb(ispin)
                  CALL cp_fm_to_fm(fm_ref, fm_target, ncol, ninactive_orb(ispin)+1, 1)
               END DO
            END IF
            IF (iw > 0) THEN
               DO ispin = 1, nspins
                  WRITE (iw, '(/,T4,A,I3,T65,A)') "Canonical Orbital Selection for spin", ispin, &
                     "[atomic units]"
                  DO i = 1, ninactive_orb(ispin), 4
                     jm = MIN(3, ninactive_orb(ispin)-i)
                     WRITE (iw, '(T3,4(F14.6,A5))') (eigenvalues(i+j, 1), " [I]", j=0, jm)
                  END DO
                  DO i = ninactive_orb(ispin)+1, ninactive_orb(ispin)+nactive_orb(ispin), 4
                     jm = MIN(3, ninactive_orb(ispin)+nactive_orb(ispin)-i)
                     WRITE (iw, '(T3,4(F14.6,A5))') (eigenvalues(i+j, 1), " [A]", j=0, jm)
                  END DO
               END DO
            END IF
            DEALLOCATE (eigenvalues)
         END IF

      CASE (wannier_projection)
         NULLIFY (loc_section, loc_print)
         loc_section => section_vals_get_subs_vals(as_input, "LOCALIZE")
         IF(.NOT.(ASSOCIATED(loc_section)))CALL cp__a("qs_active_space_methods.F",461)
         loc_print => section_vals_get_subs_vals(as_input, "LOCALIZE%PRINT")
         NULLIFY (qs_loc_env_occ, qs_loc_env_virt)
         CALL qs_loc_env_create(qs_loc_env_occ)
         ALLOCATE (mos_localized(nspins))
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
            CALL cp_fm_create(mos_localized(ispin)%matrix, mo_coeff%matrix_struct)
            CALL cp_fm_to_fm(mo_coeff, mos_localized(ispin)%matrix)
         END DO
         ! occupied states
         NULLIFY (qs_loc_env_occ)
         CALL qs_loc_env_create(qs_loc_env_occ)
         CALL qs_loc_control_init(qs_loc_env_occ, loc_section, do_homo=.TRUE.)
         CALL qs_loc_init(qs_env, qs_loc_env_occ, loc_section, &
                          mos_localized=mos_localized, do_homo=.TRUE.)
         DO ispin = 1, nspins
            CALL qs_loc_driver(qs_env, qs_loc_env_occ, loc_print, myspin=ispin, &
                               ext_mo_coeff=mos_localized(ispin)%matrix)
         END DO
         ! virtual states
         NULLIFY (qs_loc_env_virt)
         CALL qs_loc_env_create(qs_loc_env_virt)
         CALL qs_loc_control_init(qs_loc_env_virt, loc_section, do_homo=.FALSE.)
         CALL qs_loc_init(qs_env, qs_loc_env_virt, loc_section, &
                          mos_localized=mos_localized, do_homo=.FALSE.)
         CALL qs_loc_env_init(qs_loc_env_virt, qs_loc_env_virt%localized_wfn_control, qs_env, &
                              loc_coeff=mos_localized)
         DO ispin = 1, nspins
            CALL qs_loc_driver(qs_env, qs_loc_env_virt, loc_print, myspin=ispin, &
                               ext_mo_coeff=mos_localized(ispin)%matrix)
         END DO
         !
         ! get definition of subspace
         ! iatom, ishell, nshell, lnam
         IF(.NOT.(iatom > 0))CALL cp__a("qs_active_space_methods.F",496)
         IF(.NOT.(ishell > 0))CALL cp__a("qs_active_space_methods.F",497)
         CALL get_qs_env(qs_env, particle_set=particle_set)
         atomic_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, z=zval)
         NULLIFY (pro_basis_set)
         CALL create_pro_basis(pro_basis_set, zval, ishell, nshell, lnam)
         !
         CALL deallocate_gto_basis_set(pro_basis_set)
         !
         CALL cp__b("qs_active_space_methods.F",506,"not yet available")
         !
         CALL qs_loc_env_release(qs_loc_env_occ)
         CALL qs_loc_env_release(qs_loc_env_virt)
         CALL cp_fm_vect_dealloc(mos_localized)
         !
         CALL cp__b("qs_active_space_methods.F",512,"not yet available")
         !
      CASE (mao_projection)
         !
         CALL cp__b("qs_active_space_methods.F",516,"not yet available")
         !
      END SELECT

      ! Print orbitals on Cube files
      print_orb => section_vals_get_subs_vals(as_input, "PRINT_ORBITAL_CUBES")
      CALL section_vals_get(print_orb, explicit=explicit)
      IF (explicit) THEN
         !
         CALL print_orbital_cubes(print_orb, qs_env, active_space_env%mos_active)
         !
      END IF

      ! calculate inactive density matrix
      CALL get_qs_env(qs_env, rho=rho)
      CALL qs_rho_get(rho, rho_ao=rho_ao)
      IF(.NOT.(ASSOCIATED(rho_ao)))CALL cp__a("qs_active_space_methods.F",532)
      CALL cp_dbcsr_allocate_matrix_set(active_space_env%pmat_inactive, nspins)
      DO ispin = 1, nspins
         ALLOCATE (denmat)
         CALL cp_dbcsr_init(denmat)
         CALL cp_dbcsr_copy(denmat, rho_ao(ispin)%matrix)
         mo_set => active_space_env%mos_inactive(ispin)%mo_set
         CALL calculate_density_matrix(mo_set, denmat)
         active_space_env%pmat_inactive(ispin)%matrix => denmat
      END DO

      ! generate integrals
      ! make sure that defaults are set correctly (from basic calculation)
      ! make sure that periodicity is consistent
      CALL section_vals_val_get(as_input, "ERI%METHOD", i_val=eri_method)
      active_space_env%eri%method = eri_method
      CALL section_vals_val_get(as_input, "ERI%OPERATOR", i_val=eri_operator, explicit=ex_operator)
      active_space_env%eri%operator = eri_operator
      CALL section_vals_val_get(as_input, "ERI%OPERATOR_PARAMETER", r_val=eri_op_param)
      active_space_env%eri%operator_parameter = eri_op_param
      CALL section_vals_val_get(as_input, "ERI%CUTOFF_RADIUS", r_val=eri_rcut)
      active_space_env%eri%cutoff_radius = eri_rcut
      CALL section_vals_val_get(as_input, "ERI%PERIODICITY", i_vals=invals, explicit=ex_perd)
      CALL section_vals_val_get(as_input, "ERI%EPS_INTEGRAL", r_val=eri_eps_int)
      active_space_env%eri%eps_integral = eri_eps_int
      IF (active_space_env%molecule) THEN
         ! check that we are in a non-periodic setting
         CALL get_qs_env(qs_env, cell=cell)
         IF (SUM(cell%perd) /= 0) THEN
            CALL cp__b("qs_active_space_methods.F",561,"Active space option ISOLATED_SYSTEM requires non-periodic setting")
         END IF
         IF (ex_perd) THEN
            IF (SUM(invals) /= 0) THEN
               CALL cp__b("qs_active_space_methods.F",565,"Active space option ISOLATED_SYSTEM requires non-periodic setting")
            END IF
         END IF
         active_space_env%eri%periodicity(1:3) = 0
      ELSE IF (ex_perd) THEN
         IF (SIZE(invals) == 1) THEN
            active_space_env%eri%periodicity(1:3) = invals(1)
         ELSE
            active_space_env%eri%periodicity(1:3) = invals(1:3)
         END IF
      ELSE
         CALL get_qs_env(qs_env, cell=cell)
         active_space_env%eri%periodicity(1:3) = cell%perd(1:3)
      END IF
      IF (iw > 0) THEN
         WRITE (iw, '(/,T4,A)') "Calculation or Electron Repulsion Intergrals"
         SELECT CASE (eri_method)
         CASE (eri_method_full_gpw)
            WRITE (iw, '(T4,A,T48,A)') "Integration method", "GPW Fourier transform over MOs"
         CASE (eri_method_gpw_ht)
            WRITE (iw, '(T4,A,T43,A)') "Integration method", "Half transformed integrals from GPW"
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",587,"Unknown ERI method")
         END SELECT
         SELECT CASE (eri_operator)
         CASE (eri_operator_coulomb)
            WRITE (iw, '(T4,A,T40,A,T73,A)') "ERI operator", "Coulomb", "<1/R>"
         CASE (eri_operator_yukawa)
            WRITE (iw, '(T4,A,T40,A,T67,A)') "ERI operator", "Yukawa", "<EXP(-a*R)/R>"
            WRITE (iw, '(T4,A,T63,F14.6)') "ERI operator parameter", eri_op_param
         CASE (eri_operator_erf)
            WRITE (iw, '(T4,A,T40,A,T67,A)') "ERI operator", "Error function", "<ERF(a*R)/R>"
            WRITE (iw, '(T4,A,T63,F14.6)') "ERI operator parameter", eri_op_param
         CASE (eri_operator_erfc)
            WRITE (iw, '(T4,A,T40,A,T66,A)') "ERI operator", "Compl. error function", "<ERFC(a*R)/R>"
            WRITE (iw, '(T4,A,T63,F14.6)') "ERI operator parameter", eri_op_param
         CASE (eri_operator_gaussian)
            WRITE (iw, '(T4,A,T40,A,T66,A)') "ERI operator", "Gaussian attenuated", "<EXP(-a*R^2)/R>"
            WRITE (iw, '(T4,A,T63,F14.6)') "ERI operator parameter", eri_op_param
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",605,"Unknown ERI operator")
         END SELECT
         WRITE (iw, '(T4,A,T66,E12.4)') "Accuracy of ERI", eri_eps_int
         WRITE (iw, '(T4,A,T69,3I3)') "Periodicity", active_space_env%eri%periodicity(1:3)
         IF (PRODUCT(active_space_env%eri%periodicity(1:3)) == 0) THEN
            IF (eri_rcut > 0.0_dp) WRITE (iw, '(T4,A,T63,F14.6)') "Periodicity (Cutoff)", eri_rcut
         END IF
         IF (nspins < 2) THEN
            WRITE (iw, '(T4,A,T66,I12)') "Total Number of ERI", (nactive_orb(1)**4)/8
         ELSE
            WRITE (iw, '(T4,A,T66,I12)') "Total Number of ERI (aa|aa)", (nactive_orb(1)**4)/8
            WRITE (iw, '(T4,A,T66,I12)') "Total Number of ERI (bb|bb)", (nactive_orb(2)**4)/8
            WRITE (iw, '(T4,A,T66,I12)') "Total Number of ERI (aa|bb)", &
               (nactive_orb(1)**2)*(nactive_orb(2)**2)/4
         END IF
      END IF

      ! allocate container for integrals (CSR matrix)
      CALL get_qs_env(qs_env, para_env=para_env)
      m = (nspins*(nspins+1))/2
      ALLOCATE (active_space_env%eri%eri(m))
      DO i = 1, m
         ALLOCATE (active_space_env%eri%eri(i)%csr_mat)
         eri_mat => active_space_env%eri%eri(i)%csr_mat
         IF (i == 1) THEN
            n1 = nactive_orb(1)
            n2 = nactive_orb(1)
         ELSEIF (i == 2) THEN
            n1 = nactive_orb(1)
            n2 = nactive_orb(2)
         ELSE
            n1 = nactive_orb(2)
            n2 = nactive_orb(2)
         END IF
         nn1 = (n1*(n1+1))/2
         nn2 = (n2*(n2+1))/2
         CALL cp_csr_create(eri_mat, nn1, nn2, 0_int_8, 0, 0, para_env%group)
      END DO

      SELECT CASE (eri_method)
      CASE (eri_method_full_gpw, eri_method_gpw_ht)
         CALL section_vals_val_get(as_input, "ERI_GPW%EPS_GRID", r_val=eri_eps_grid)
         active_space_env%eri%eri_gpw%eps_grid = eri_eps_grid
         CALL section_vals_val_get(as_input, "ERI_GPW%CUTOFF", r_val=eri_gpw_cutoff)
         active_space_env%eri%eri_gpw%cutoff = eri_gpw_cutoff
         CALL section_vals_val_get(as_input, "ERI_GPW%REL_CUTOFF", r_val=eri_rel_cutoff)
         active_space_env%eri%eri_gpw%rel_cutoff = eri_rel_cutoff
         CALL section_vals_val_get(as_input, "ERI_GPW%PRINT_LEVEL", i_val=eri_print)
         active_space_env%eri%eri_gpw%print_level = eri_print
         CALL section_vals_val_get(as_input, "ERI_GPW%STORE_WFN", l_val=store_wfn)
         active_space_env%eri%eri_gpw%store_wfn = store_wfn
         active_space_env%eri%eri_gpw%redo_poisson = (ex_operator .OR. ex_perd)
         IF (iw > 0) THEN
            WRITE (iw, '(T4,A,T68,F10.4)') "ERI_GPW| Energy cutoff [Ry]", eri_gpw_cutoff
            WRITE (iw, '(T4,A,T68,F10.4)') "ERI_GPW| Relative energy cutoff [Ry]", eri_rel_cutoff
         END IF
         !
         CALL calculate_eri_gpw(active_space_env%mos_active, active_space_env%eri, qs_env, iw)
         !
      CASE DEFAULT
         CALL cp__b("qs_active_space_methods.F",665,"Unknown ERI method")
      END SELECT
      IF (iw > 0) THEN
         DO isp = 1, SIZE(active_space_env%eri%eri)
            eri_mat => active_space_env%eri%eri(isp)%csr_mat
            nze_percentage = 100.0_dp*(REAL(eri_mat%nze_total, KIND=dp) &
                                       /REAL(eri_mat%nrows_total, KIND=dp))/REAL(eri_mat%ncols_total, KIND=dp)
            WRITE (iw, '(T4,A,I2,T30,A,T66,I12)') "ERI_GPW| Spinmatrix:", isp, &
               "Number of  CSR non-zero elements:", eri_mat%nze_total
            WRITE (iw, '(T4,A,I2,T30,A,T66,F12.4)') "ERI_GPW| Spinmatrix:", isp, &
               "Percentage CSR non-zero elements:", nze_percentage
         END DO
      END IF

      ! set the reference active space density matrix
      nspins = active_space_env%nspins
      ALLOCATE (active_space_env%p_ref(nspins))
      DO isp = 1, nspins
         mo_set => active_space_env%mos_active(isp)%mo_set
         CALL get_mo_set(mo_set, mo_coeff=mo_coeff)
         NULLIFY (active_space_env%p_ref(isp)%matrix)
         CALL create_subspace_matrix(mo_coeff, active_space_env%p_ref(isp)%matrix)
      END DO
      SELECT CASE (mselect)
      CASE DEFAULT
         CALL cp__b("qs_active_space_methods.F",690,"Unknown orbital selection method")
      CASE (casci_canonical)
         focc = 2.0_dp
         IF (nspins == 2) focc = 1.0_dp
         DO isp = 1, nspins
            fmat => active_space_env%p_ref(isp)%matrix
            CALL cp_fm_set_all(fmat, alpha=0.0_dp)
            n1 = active_space_env%nelectrons(isp)
            DO i = 1, nactive_orb(isp)
               fel = MIN(focc, REAL(n1, KIND=dp))
               CALL cp_fm_set_element(fmat, i, i, fel)
               n1 = n1-NINT(fel)
               n1 = MAX(n1, 0)
            END DO
         END DO
      CASE (wannier_projection)
         CALL cp__b("qs_active_space_methods.F",706,"NOT IMPLEMENTED")
      CASE (mao_projection)
         CALL cp__b("qs_active_space_methods.F",708,"NOT IMPLEMENTED")
      END SELECT

      ! calculate one-electron operators in the subspace
      CALL calculate_operators(active_space_env%mos_active, qs_env, active_space_env)

      ! Output a FCIDUMP file
      CALL get_qs_env(qs_env, energy=energy)
      eref = energy%total
      CALL fcidump(active_space_env, as_input, eref)

      CALL set_qs_env(qs_env, active_space=active_space_env)

      IF (iw > 0) THEN
         WRITE (iw, '(/,T2,A)') &
            '!-------------------- End of Active Space Interface --------------------------!'
      END IF

      CALL timestop(handle)

   END SUBROUTINE active_space_main

! **************************************************************************************************
!> \brief computes the one-electron operators in the subspace of the provided orbital set
!> \param mos ...
!> \param qs_env ...
!> \param active_space_env ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE calculate_operators(mos, qs_env, active_space_env)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(active_space_type), POINTER                   :: active_space_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_operators', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, is, nmo, nspins
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: h_matrix, ks_matrix
      TYPE(cp_fm_type), POINTER                          :: hmat, ksmat, mo_coeff

      CALL timeset(routineN, handle)

      ! Kohn-Sham / Fock operator
      CALL get_qs_env(qs_env, matrix_ks_kp=ks_matrix)
      IF (SIZE(ks_matrix, 2) > 1) THEN
         CALL cp__b("qs_active_space_methods.F",756,"No k-points allowed at this point")
      END IF
      nspins = active_space_env%nspins
      ALLOCATE (active_space_env%ks_sub(nspins))
      DO is = 1, nspins
         CALL get_mo_set(mo_set=mos(is)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         NULLIFY (ksmat)
         CALL subspace_operator(mo_coeff, ks_matrix(is, 1)%matrix, ksmat)
         active_space_env%ks_sub(is)%matrix => ksmat
      END DO

      ! Core Hamiltonian
      NULLIFY (h_matrix)
      CALL get_qs_env(qs_env=qs_env, matrix_h_kp=h_matrix)
      IF (SIZE(h_matrix, 2) > 1) THEN
         CALL cp__b("qs_active_space_methods.F",771,"No k-points allowed at this point")
      END IF
      ALLOCATE (active_space_env%h_sub(nspins))
      DO is = 1, nspins
         CALL get_mo_set(mo_set=mos(is)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         NULLIFY (hmat)
         CALL subspace_operator(mo_coeff, h_matrix(1, 1)%matrix, hmat)
         active_space_env%h_sub(is)%matrix => hmat
      END DO

      CALL timestop(handle)

   END SUBROUTINE calculate_operators

! **************************************************************************************************
!> \brief computes a one-electron operator in the subspace of the provided orbital set
!> \param orbitals ...
!> \param op_matrix operator matrix in AO basis
!> \param op_sub operator in orbital basis
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE subspace_operator(orbitals, op_matrix, op_sub)

      TYPE(cp_fm_type), POINTER                          :: orbitals
      TYPE(cp_dbcsr_type), POINTER                       :: op_matrix
      TYPE(cp_fm_type), POINTER                          :: op_sub

      CHARACTER(len=*), PARAMETER :: routineN = 'subspace_operator', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ncol, nrow
      TYPE(cp_fm_type), POINTER                          :: vectors

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(op_sub)))CALL cp__a("qs_active_space_methods.F",807)

      CALL cp_fm_get_info(matrix=orbitals, ncol_global=ncol, nrow_global=nrow)

      IF (ncol .GT. 0) THEN

         CALL cp_fm_create(vectors, orbitals%matrix_struct, "vectors")

         CALL create_subspace_matrix(orbitals, op_sub)

         CALL cp_dbcsr_sm_fm_multiply(op_matrix, orbitals, vectors, ncol)

         CALL cp_gemm('T', 'N', ncol, ncol, nrow, 1.0_dp, orbitals, vectors, 0.0_dp, op_sub)

         CALL cp_fm_release(vectors)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE subspace_operator

! **************************************************************************************************
!> \brief creates a matrix of subspace size
!> \param orbitals ...
!> \param op_sub operator in orbital basis
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE create_subspace_matrix(orbitals, op_sub)

      TYPE(cp_fm_type), POINTER                          :: orbitals, op_sub

      CHARACTER(len=*), PARAMETER :: routineN = 'create_subspace_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncol, nrow
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct

      IF(.NOT.(.NOT. ASSOCIATED(op_sub)))CALL cp__a("qs_active_space_methods.F",846)

      CALL cp_fm_get_info(matrix=orbitals, ncol_global=ncol, nrow_global=nrow)

      IF (ncol .GT. 0) THEN

         NULLIFY (fm_struct)
         CALL cp_fm_struct_create(fm_struct, nrow_global=ncol, ncol_global=ncol, &
                                  para_env=orbitals%matrix_struct%para_env, &
                                  context=orbitals%matrix_struct%context)
         CALL cp_fm_create(op_sub, fm_struct, name="Subspace operator")
         CALL cp_fm_struct_release(fm_struct)

      ENDIF

   END SUBROUTINE create_subspace_matrix

! **************************************************************************************************
!> \brief computes a electron repulsion integrals using GPW technology
!> \param mos ...
!> \param eri_env ...
!> \param qs_env ...
!> \param iw ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE calculate_eri_gpw(mos, eri_env, qs_env, iw)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(eri_type)                                     :: eri_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_eri_gpw', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i_multigrid, icount2, intcount, irange(2), isp, isp1, isp2, ispin, iwa1, &
         iwa12, iwa2, iwb1, iwb2, iwbs, iwbt, iwfn, n_multigrid, nmm, nmo, nmo1, nmo2, nspins, nx
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: eri_index
      LOGICAL                                            :: print1, print2, &
                                                            skip_load_balance_distributed
      REAL(KIND=dp)                                      :: cutoff_old, dvol, erint, pair_int, &
                                                            progression_factor, rc, &
                                                            relative_cutoff_old, rsize
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: e_cutoff_old, eri
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_p_type)                                    :: pot_g, rho_g, rho_r, wfn_r
      TYPE(pw_p_type), DIMENSION(:, :), POINTER          :: wfn_a
      TYPE(pw_p_type), POINTER                           :: wfn1, wfn2, wfn3, wfn4
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(task_list_type), POINTER                      :: task_list_sub

      CALL timeset(routineN, handle)

      ! print levels
      SELECT CASE (eri_env%eri_gpw%print_level)
      CASE (silent_print_level)
         print1 = .FALSE.
         print2 = .FALSE.
      CASE (low_print_level)
         print1 = .FALSE.
         print2 = .FALSE.
      CASE (medium_print_level)
         print1 = .TRUE.
         print2 = .FALSE.
      CASE (high_print_level)
         print1 = .TRUE.
         print2 = .TRUE.
      CASE (debug_print_level)
         print1 = .TRUE.
         print2 = .TRUE.
      CASE DEFAULT
         ! do nothing
      END SELECT

      ! This should be done differently! Copied from MP2 code
      CALL get_qs_env(qs_env, dft_control=dft_control)
      progression_factor = dft_control%qs_control%progression_factor
      n_multigrid = SIZE(dft_control%qs_control%e_cutoff)
      ALLOCATE (e_cutoff_old(n_multigrid))
      e_cutoff_old(:) = dft_control%qs_control%e_cutoff
      cutoff_old = dft_control%qs_control%cutoff

      dft_control%qs_control%cutoff = eri_env%eri_gpw%cutoff*0.5_dp
      dft_control%qs_control%e_cutoff(1) = dft_control%qs_control%cutoff
      DO i_multigrid = 2, n_multigrid
         dft_control%qs_control%e_cutoff(i_multigrid) = dft_control%qs_control%e_cutoff(i_multigrid-1) &
                                                        /progression_factor
      END DO

      relative_cutoff_old = dft_control%qs_control%relative_cutoff
      dft_control%qs_control%relative_cutoff = eri_env%eri_gpw%rel_cutoff*0.5_dp

      ! Generate the appropriate  pw_env
      NULLIFY (pw_env_sub)
      CALL pw_env_create(pw_env_sub)
      CALL pw_env_rebuild(pw_env_sub, qs_env)
      CALL pw_env_get(pw_env_sub, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)
      IF (eri_env%eri_gpw%redo_poisson) THEN
         IF (.NOT. (poisson_env%parameters%solver == pw_poisson_analytic .OR. &
                    poisson_env%parameters%solver == pw_poisson_periodic)) THEN
            CALL cp__b("qs_active_space_methods.F",958,"Only use simple analytic Poisson solvers with ERI calculation")
         END IF
         poisson_env%parameters%periodic = eri_env%periodicity
         CALL pw_poisson_rebuild(poisson_env)
         IF (eri_env%cutoff_radius > 0.0_dp) THEN
            poisson_env%green_fft%radius = eri_env%cutoff_radius
         ELSE
            CALL get_qs_env(qs_env, cell=cell)
            rc = cell%hmat(1, 1)
            DO iwa1 = 1, 3
               rc = MIN(rc, 0.5_dp*cell%hmat(iwa1, iwa1))
            END DO
            poisson_env%green_fft%radius = rc
         END IF
         CALL pw_eri_green_create(poisson_env%green_fft, eri_env)
         IF (print1 .AND. iw > 0) THEN
            WRITE (iw, "(T4,'ERI_GPW|',' Redefine Poisson Greens function ')")
         END IF
      END IF

      IF (eri_env%method == eri_method_gpw_ht) THEN
         ! We need a task list
         NULLIFY (task_list_sub)
         skip_load_balance_distributed = dft_control%qs_control%skip_load_balance_distributed
         CALL get_qs_env(qs_env, ks_env=ks_env, sab_orb=sab_orb_sub)
         CALL allocate_task_list(task_list_sub)
         CALL generate_qs_task_list(ks_env, task_list_sub, &
                                    reorder_rs_grid_ranks=.TRUE., soft_valid=.FALSE., &
                                    skip_load_balance_distributed=skip_load_balance_distributed, &
                                    pw_env_external=pw_env_sub, sab_orb_external=sab_orb_sub)
      END IF

      NULLIFY (wfn_r%pw, rho_g%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, wfn_r%pw, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_g%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set, cell=cell, &
                      particle_set=particle_set, atomic_kind_set=atomic_kind_set)

      ! pre-calculate wavefunctions on reals space grid
      nspins = SIZE(mos)
      IF (eri_env%eri_gpw%store_wfn) THEN
         rsize = 0.0_dp
         nmo = 0
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nx)
            nmo = MAX(nmo, nx)
            rsize = REAL(SIZE(wfn_r%pw%cr3d), KIND=dp)*nx
         END DO
         IF (print1 .AND. iw > 0) THEN
            rsize = rsize*8._dp/1000000._dp
            WRITE (iw, "(T4,'ERI_GPW|',' Store active orbitals on real space grid ',T63,F12.3,' MB')") rsize
         END IF
         ALLOCATE (wfn_a(nmo, nspins))
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
            DO iwfn = 1, nmo
               NULLIFY (wfn_a(iwfn, ispin)%pw)
               CALL pw_pool_create_pw(auxbas_pw_pool, wfn_a(iwfn, ispin)%pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL calculate_wavefunction(mo_coeff, iwfn, wfn_a(iwfn, ispin), rho_g, atomic_kind_set, &
                                           qs_kind_set, cell, dft_control, particle_set, pw_env_sub)
               IF (print2 .AND. iw > 0) THEN
                  WRITE (iw, "(T4,'ERI_GPW|',' Orbital stored ',I4,'  Spin ',i1)") iwfn, ispin
               END IF
            END DO
         END DO
      END IF

      ! get some of the grids ready
      NULLIFY (rho_r%pw, pot_g%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_r%pw, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, pot_g%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      ! run the FFT once, to set up buffers and to take into account the memory
      rho_r%pw%cr3d = 0.0D0
      CALL pw_transfer(rho_r%pw, rho_g%pw)
      dvol = rho_r%pw%pw_grid%dvol

      ! calculate the integrals
      intcount = 0
      DO isp1 = 1, nspins
         CALL get_mo_set(mo_set=mos(isp1)%mo_set, nmo=nmo1)
         nmm = (nmo1*(nmo1+1))/2
         irange = get_irange_csr(nmm, eri_env%eri(1)%csr_mat%mp_group)
         DO iwa1 = 1, nmo1
            IF (eri_env%eri_gpw%store_wfn) THEN
               wfn1 => wfn_a(iwa1, isp1)
            ELSE
               CALL cp__b("qs_active_space_methods.F",1047,"")
            END IF
            DO iwa2 = iwa1, nmo1
               iwa12 = get_index_ij(iwa1, iwa2, nmo1)
               IF (iwa12 >= irange(1) .AND. iwa12 <= irange(2)) THEN
                  iwa12 = iwa12-irange(1)+1
               ELSE
                  iwa12 = 0
               END IF
               IF (eri_env%eri_gpw%store_wfn) THEN
                  wfn2 => wfn_a(iwa2, isp1)
               ELSE
                  CALL cp__b("qs_active_space_methods.F",1059,"")
               END IF
               ! calculate charge distribution and potential
               rho_r%pw%cr3d = wfn1%pw%cr3d*wfn2%pw%cr3d
               IF (print2 .AND. iw > 0) THEN
                  erint = pw_integrate_function(rho_r%pw)/dvol
                  WRITE (iw, "(T4,'ERI_GPW| Integral rho_ab ',T32,2I4,' [',I1,']',T58,G20.14)") &
                     iwa1, iwa2, isp1, erint
               END IF
               CALL pw_transfer(rho_r%pw, rho_g%pw)
               CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_int, pot_g%pw)
               ! screening using pair_int
               IF (pair_int < eri_env%eps_integral) CYCLE
               CALL pw_transfer(pot_g%pw, rho_r%pw)
               !
               IF (eri_env%method == eri_method_gpw_ht) THEN
                  CALL cp__b("qs_active_space_methods.F",1075,"Not available")
               ELSEIF (eri_env%method == eri_method_full_gpw) THEN
                  DO isp2 = isp1, nspins
                     CALL get_mo_set(mo_set=mos(isp1)%mo_set, nmo=nmo2)
                     nx = (nmo2*(nmo2+1))/2
                     ALLOCATE (eri(nx), eri_index(nx))
                     icount2 = 0
                     iwbs = 1
                     IF (isp1 == isp2) iwbs = iwa1
                     isp = (isp1-1)*isp2-((isp1-1)*(isp1-2))/2+(isp2-isp1+1)
                     DO iwb1 = iwbs, nmo2
                        IF (eri_env%eri_gpw%store_wfn) THEN
                           wfn3 => wfn_a(iwb1, isp2)
                        ELSE
                           CALL cp__b("qs_active_space_methods.F",1089,"")
                        END IF
                        iwbt = iwb1
                        IF (isp1 == isp2 .AND. iwa1 == iwb1) iwbt = iwa2
                        DO iwb2 = iwbt, nmo2
                           IF (eri_env%eri_gpw%store_wfn) THEN
                              wfn4 => wfn_a(iwb2, isp2)
                           ELSE
                              CALL cp__b("qs_active_space_methods.F",1097,"")
                           END IF
                           wfn_r%pw%cr3d = rho_r%pw%cr3d*wfn3%pw%cr3d*wfn4%pw%cr3d
                           erint = pw_integrate_function(wfn_r%pw)
                           IF (erint > eri_env%eps_integral) THEN
                              intcount = intcount+1
                              IF (print2 .AND. iw > 0) THEN
                                 WRITE (iw, "(T4,'ERI_GPW|',T20,2I4,' [',I1,']',2I4,' [',I1,']',T58,G20.14)") &
                                    iwa1, iwa2, isp1, iwb1, iwb2, isp2, erint
                              END IF
                              icount2 = icount2+1
                              eri(icount2) = erint
                              eri_index(icount2) = get_index_ij(iwb1, iwb2, nmo2)
                           END IF
                        END DO
                     END DO
                     !
                     CALL update_csr_matrix(eri_env%eri(isp)%csr_mat, icount2, eri, eri_index, iwa12)
                     !
                     DEALLOCATE (eri, eri_index)
                  END DO
               ELSE
                  CALL cp__b("qs_active_space_methods.F",1119,"Unknown option")
               END IF
            END DO
         END DO
      END DO

      IF (print1 .AND. iw > 0) THEN
         WRITE (iw, "(T4,'ERI_GPW|',' Number of Integrals stored ',T68,I10)") intcount
      END IF

      IF (eri_env%eri_gpw%store_wfn) THEN
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo)
            DO iwfn = 1, nmo
               CALL pw_release(wfn_a(iwfn, ispin)%pw)
            END DO
         END DO
         DEALLOCATE (wfn_a)
      END IF
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wfn_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_g%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, pot_g%pw)

      IF (eri_env%method == eri_method_gpw_ht) THEN
         CALL deallocate_task_list(task_list_sub)
      END IF
      CALL pw_env_release(pw_env_sub)
      ! restore the initial value of the cutoff
      dft_control%qs_control%e_cutoff = e_cutoff_old
      dft_control%qs_control%cutoff = cutoff_old
      dft_control%qs_control%relative_cutoff = relative_cutoff_old
      DEALLOCATE (e_cutoff_old)

      CALL timestop(handle)

   END SUBROUTINE calculate_eri_gpw

! **************************************************************************************************
!> \brief Sets the Green's function
!> \param green ...
!> \param eri_env ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE pw_eri_green_create(green, eri_env)

      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(eri_type)                                     :: eri_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_eri_green_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ig
      REAL(KIND=dp)                                      :: a, g2, g3d, gg, rlength
      TYPE(pw_grid_type), POINTER                        :: grid
      TYPE(pw_type), POINTER                             :: gf

      ! initialize influence function
      gf => green%influence_fn
      grid => green%influence_fn%pw_grid
      SELECT CASE (green%method)
      CASE (PERIODIC3D)

         SELECT CASE (eri_env%operator)
         CASE (eri_operator_coulomb)
            DO ig = grid%first_gne0, grid%ngpts_cut_local
               g2 = grid%gsq(ig)
               gf%cc(ig) = fourpi/g2
            END DO
            IF (grid%have_g0) gf%cc(1) = 0.0_dp
         CASE (eri_operator_yukawa)
            a = eri_env%operator_parameter**2
            DO ig = grid%first_gne0, grid%ngpts_cut_local
               g2 = grid%gsq(ig)
               gf%cc(ig) = fourpi/(a+g2)
            END DO
            IF (grid%have_g0) gf%cc(1) = fourpi/a
         CASE (eri_operator_erf)
            CALL cp__b("qs_active_space_methods.F",1198,"")
         CASE (eri_operator_erfc)
            CALL cp__b("qs_active_space_methods.F",1200,"")
         CASE (eri_operator_gaussian)
            CALL cp__b("qs_active_space_methods.F",1202,"")
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",1204,"")
         END SELECT

      CASE (ANALYTIC0D)

         SELECT CASE (eri_env%operator)
         CASE (eri_operator_coulomb)
            rlength = green%radius
            DO ig = grid%first_gne0, grid%ngpts_cut_local
               g2 = grid%gsq(ig)
               gg = SQRT(g2)
               g3d = fourpi/g2
               gf%cc(ig) = g3d*(1.0_dp-COS(rlength*gg))
            END DO
            IF (grid%have_g0) gf%cc(1) = 0.5_dp*fourpi*rlength*rlength
         CASE (eri_operator_yukawa)
            CALL cp__b("qs_active_space_methods.F",1220,"")
         CASE (eri_operator_erf)
            CALL cp__b("qs_active_space_methods.F",1222,"")
         CASE (eri_operator_erfc)
            CALL cp__b("qs_active_space_methods.F",1224,"")
         CASE (eri_operator_gaussian)
            CALL cp__b("qs_active_space_methods.F",1226,"")
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",1228,"")
         END SELECT

      CASE DEFAULT
         CALL cp__b("qs_active_space_methods.F",1232,"")
      END SELECT

   END SUBROUTINE pw_eri_green_create

! **************************************************************************************************
!> \brief calculates combined index (ij)
!> \param i ...
!> \param j ...
!> \param n ...
!> \retval ij ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   FUNCTION get_index_ij(i, j, n) RESULT(ij)

      INTEGER                                            :: i, j, n, ij

      IF(.NOT.(i <= j))CALL cp__a("qs_active_space_methods.F",1250)
      IF(.NOT.(i <= n))CALL cp__a("qs_active_space_methods.F",1251)
      IF(.NOT.(j <= n))CALL cp__a("qs_active_space_methods.F",1252)

      ij = (i-1)*n-((i-1)*(i-2))/2+(j-i+1)

      IF(.NOT.(ij <= (n*(n+1))/2))CALL cp__a("qs_active_space_methods.F",1256)

   END FUNCTION get_index_ij
! **************************************************************************************************
!> \brief extracts indeces i and j from combined index ij
!> \param ij ...
!> \param n ...
!> \param i ...
!> \param j ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE get_index_pair(ij, n, i, j)
      INTEGER                                            :: ij, n, i, j

      INTEGER                                            :: m, m0

      m = MAX(ij/n, 1)
      DO i = m, n
         m0 = (i-1)*n-((i-1)*(i-2))/2
         j = ij-m0+i-1
         IF (j <= n) EXIT
      END DO

      IF(.NOT.(i > 0 .AND. i <= n))CALL cp__a("qs_active_space_methods.F",1280)
      IF(.NOT.(j > 0 .AND. j <= n))CALL cp__a("qs_active_space_methods.F",1281)
      IF(.NOT.(i <= j))CALL cp__a("qs_active_space_methods.F",1282)

   END SUBROUTINE get_index_pair
! **************************************************************************************************
!> \brief calculates index range for processor in group mp_group
!> \param nindex ...
!> \param mp_group ...
!> \retval irange ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   FUNCTION get_irange_csr(nindex, mp_group) RESULT(irange)

      INTEGER                                            :: nindex, mp_group
      INTEGER, DIMENSION(2)                              :: irange

      INTEGER                                            :: numtask, taskid
      REAL(KIND=dp)                                      :: rat

      CALL mp_environ(numtask, taskid, mp_group)

      IF (numtask == 1 .AND. taskid == 0) THEN
         irange(1) = 1
         irange(2) = nindex
      ELSEIF (numtask >= nindex) THEN
         IF (taskid >= nindex) THEN
            irange(1) = 1
            irange(2) = 0
         ELSE
            irange(1) = taskid+1
            irange(2) = taskid+1
         END IF
      ELSE
         rat = REAL(nindex, KIND=dp)/REAL(numtask, KIND=dp)
         irange(1) = NINT(rat*taskid)+1
         irange(2) = NINT(rat*taskid+rat)
      END IF

   END FUNCTION get_irange_csr

! **************************************************************************************************
!> \brief Adds data for a new row to the csr matrix
!> \param csr_mat ...
!> \param nnz ...
!> \param rdat ...
!> \param rind ...
!> \param irow ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE update_csr_matrix(csr_mat, nnz, rdat, rind, irow)

      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat
      INTEGER, INTENT(IN)                                :: nnz
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rdat
      INTEGER, DIMENSION(:), INTENT(IN)                  :: rind
      INTEGER, INTENT(IN)                                :: irow

      INTEGER                                            :: k, nrow, nze, nze_new

      IF (irow /= 0) THEN
         IF(.NOT.(irow <= csr_mat%nrows_total))CALL cp__a("qs_active_space_methods.F",1343)
         IF(.NOT.(irow > csr_mat%nrows_local))CALL cp__a("qs_active_space_methods.F",1344)
         nze = csr_mat%nze_local
         nze_new = nze+nnz
         ! values
         CALL reallocate(csr_mat%nzval_local%r_dp, 1, nze_new)
         csr_mat%nzval_local%r_dp(nze+1:nze_new) = rdat(1:nnz)
         ! col indices
         CALL reallocate(csr_mat%colind_local, 1, nze_new)
         csr_mat%colind_local(nze+1:nze_new) = rind(1:nnz)
         ! rows
         nrow = csr_mat%nrows_local
         CALL reallocate(csr_mat%rowptr_local, 1, irow+1)
         csr_mat%rowptr_local(nrow+1:irow) = nze+1
         csr_mat%rowptr_local(irow+1) = nze_new+1
         ! nzerow
         CALL reallocate(csr_mat%nzerow_local, 1, irow)
         DO k = nrow+1, irow
            csr_mat%nzerow_local(k) = csr_mat%rowptr_local(k+1)-csr_mat%rowptr_local(k)
         ENDDO
         csr_mat%nrows_local = irow
         csr_mat%nze_local = csr_mat%nze_local+nnz
      END IF
      csr_mat%nze_total = csr_mat%nze_total+nnz
      csr_mat%has_indices = .TRUE.

   END SUBROUTINE update_csr_matrix

! **************************************************************************************************
!> \brief Computes and prints the active orbitals on Cube Files
!> \param input ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param mos ...
! **************************************************************************************************
   SUBROUTINE print_orbital_cubes(input, qs_env, mos)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mo_set_p_type), DIMENSION(:)                  :: mos

      CHARACTER(len=*), PARAMETER :: routineN = 'print_orbital_cubes', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filebody, filename, title
      INTEGER                                            :: i, imo, isp, nmo, str(3), unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: alist, blist, istride
      LOGICAL                                            :: do_mo, explicit_a, explicit_b
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_subsys_type), POINTER                      :: subsys

      CALL section_vals_val_get(input, "FILENAME", c_val=filebody)
      CALL section_vals_val_get(input, "STRIDE", i_vals=istride)
      IF (SIZE(istride) == 1) THEN
         str(1:3) = istride(1)
      ELSEIF (SIZE(istride) == 3) THEN
         str(1:3) = istride(1:3)
      ELSE
         CALL cp__b("qs_active_space_methods.F",1409,"STRIDE arguments inconsistent")
      END IF
      CALL section_vals_val_get(input, "ALIST", i_vals=alist, explicit=explicit_a)
      CALL section_vals_val_get(input, "BLIST", i_vals=blist, explicit=explicit_b)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      subsys=subsys, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      particle_set=particle_set, &
                      pw_env=pw_env)
      CALL qs_subsys_get(subsys, particles=particles)
      !
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
      CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, wf_g%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      !
      DO isp = 1, SIZE(mos)
         CALL get_mo_set(mo_set=mos(isp)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         DO imo = 1, nmo
            IF (isp == 1 .AND. explicit_a) THEN
               IF (alist(1) == -1) THEN
                  do_mo = .TRUE.
               ELSE
                  do_mo = .FALSE.
                  DO i = 1, SIZE(alist)
                     IF (imo == alist(i)) do_mo = .TRUE.
                  END DO
               END IF
            ELSE IF (isp == 2 .AND. explicit_b) THEN
               IF (blist(1) == -1) THEN
                  do_mo = .TRUE.
               ELSE
                  do_mo = .FALSE.
                  DO i = 1, SIZE(blist)
                     IF (imo == blist(i)) do_mo = .TRUE.
                  END DO
               END IF
            ELSE
               do_mo = .TRUE.
            END IF
            IF (.NOT. do_mo) CYCLE
            CALL calculate_wavefunction(mo_coeff, imo, wf_r, wf_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env)
            IF (para_env%ionode) THEN
               WRITE (filename, '(A,A1,I4.4,A1,I1.1,A)') TRIM(filebody), "_", imo, "_", isp, ".cube"
               CALL open_file(filename, unit_number=unit_nr, file_status="UNKNOWN", file_action="WRITE")
               WRITE (title, *) "Active Orbital ", imo, " spin ", isp
            ELSE
               unit_nr = -1
            END IF
            CALL cp_pw_to_cube(wf_r%pw, unit_nr, title, particles=particles, stride=istride)
            IF (para_env%ionode) THEN
               CALL close_file(unit_nr)
            END IF
         ENDDO
      ENDDO

      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)

   END SUBROUTINE print_orbital_cubes
! **************************************************************************************************
!> \brief Writes a FCIDUMP file
!> \param active_space_env ...
!> \param as_input ...
!> \param eref ...
!> \par History
!>      04.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE fcidump(active_space_env, as_input, eref)

      TYPE(active_space_type), POINTER                   :: active_space_env
      TYPE(section_vals_type), POINTER                   :: as_input
      REAL(KIND=dp), INTENT(IN)                          :: eref

      CHARACTER(len=*), PARAMETER :: routineN = 'fcidump', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, i12, i12l, i2, i3, i34, i34l, i4, &
                                                            irange(2), irptr, isym, iw, ms2, &
                                                            nelec, nindex, norb, nspins
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: colind
      REAL(KIND=dp)                                      :: eeri, erint, esub, mval
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: erival
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: ks_mat, ks_ref, p_mat
      TYPE(cp_fm_type), POINTER                          :: matrix
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(csr_type), POINTER                            :: eri

      logger => cp_get_default_logger()
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           as_input, "FCIDUMP"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, as_input, "FCIDUMP", &
                                   extension=".fcidump", file_status="REPLACE", file_action="WRITE", file_form="FORMATTED")
         !
         nspins = active_space_env%nspins
         IF (nspins == 1) THEN
            ms2 = 0 !active_space_env%multiplicity ?
            nelec = active_space_env%nactive
            CALL get_mo_set(active_space_env%mos_active(1)%mo_set, nmo=norb)
            IF (iw > 0) THEN
               WRITE (iw, "(A,A,I4,A,I4,A,I2,A)") "&FCI", " NORB=", norb, ",NELEC=", nelec, ",MS2=", ms2, ","
               isym = 1
               WRITE (iw, "(A,1000(I1,','))") "  ORBSYM=", (isym, i=1, norb)
               isym = 0
               WRITE (iw, "(A,I1,A)") "  ISYM=", isym, ","
               WRITE (iw, "(A)") " /"
            END IF
            !
            ! Loop over ERI, calculate subspace HF energy and Fock matrix
            !
            ! replicate KS, Core, and P matrices
            ALLOCATE (ks_mat(norb, norb), ks_ref(norb, norb), p_mat(norb, norb))
            ks_mat = 0.0_dp
            ks_ref = 0.0_dp
            p_mat = 0.0_dp
            DO i1 = 1, norb
               DO i2 = i1, norb
                  matrix => active_space_env%p_ref(1)%matrix
                  CALL cp_fm_get_element(matrix, i1, i2, mval)
                  p_mat(i1, i2) = mval
                  p_mat(i2, i1) = mval
                  matrix => active_space_env%ks_sub(1)%matrix
                  CALL cp_fm_get_element(matrix, i1, i2, mval)
                  ks_mat(i1, i2) = mval
                  ks_mat(i2, i1) = mval
               END DO
            END DO
            ! calculate energy
            eeri = 0.0_dp
            eri => active_space_env%eri%eri(1)%csr_mat
            nindex = (norb*(norb+1))/2
            irange = get_irange_csr(nindex, eri%mp_group)
            DO i1 = 1, norb
               DO i2 = i1, norb
                  i12 = get_index_ij(i1, i2, norb)
                  IF (i12 >= irange(1) .AND. i12 <= irange(2)) THEN
                     i12l = i12-irange(1)+1
                     irptr = eri%rowptr_local(i12l)-1
                     DO i34l = 1, eri%nzerow_local(i12l)
                        i34 = eri%colind_local(irptr+i34l)
                        CALL get_index_pair(i34, norb, i3, i4)
                        erint = eri%nzval_local%r_dp(irptr+i34l)
                        ! Coulomb
                        ks_ref(i1, i2) = ks_ref(i1, i2)+erint*p_mat(i3, i4)
                        IF (i3 /= i4) THEN
                           ks_ref(i1, i2) = ks_ref(i1, i2)+erint*p_mat(i3, i4)
                        END IF
                        IF (i12 /= i34) THEN
                           ks_ref(i3, i4) = ks_ref(i3, i4)+erint*p_mat(i1, i2)
                           IF (i1 /= i2) THEN
                              ks_ref(i3, i4) = ks_ref(i3, i4)+erint*p_mat(i1, i2)
                           END IF
                        END IF
                        ! Exchange
                        erint = -0.5_dp*erint
                        ks_ref(i1, i3) = ks_ref(i1, i3)+erint*p_mat(i2, i4)
                        IF (i1 /= i2) THEN
                           ks_ref(i2, i3) = ks_ref(i2, i3)+erint*p_mat(i1, i4)
                        END IF
                        IF (i3 /= i4) THEN
                           ks_ref(i1, i4) = ks_ref(i1, i4)+erint*p_mat(i2, i3)
                        END IF
                        IF (i1 /= i2 .AND. i3 /= i4) THEN
                           ks_ref(i2, i4) = ks_ref(i2, i4)+erint*p_mat(i1, i3)
                        END IF
                     END DO
                  END IF
               END DO
            END DO
            !
            DO i1 = 1, norb
               DO i2 = i1, norb
                  ks_ref(i1, i2) = ks_ref(i2, i1)
               END DO
            END DO
            CALL mp_sum(ks_ref, eri%mp_group)
            !
            eeri = 0.5_dp*SUM(ks_ref*p_mat)
            ks_mat(1:norb, 1:norb) = ks_mat(1:norb, 1:norb)+ks_ref(1:norb, 1:norb)
            esub = eref-SUM(ks_mat(1:norb, 1:norb)*p_mat(1:norb, 1:norb))+eeri
            !
            ! Print integrals: ERI
            eri => active_space_env%eri%eri(1)%csr_mat
            nindex = (norb*(norb+1))/2
            irange = get_irange_csr(nindex, eri%mp_group)
            ALLOCATE (erival(nindex), colind(nindex))
            DO i1 = 1, norb
               DO i2 = i1, norb
                  i12 = get_index_ij(i1, i2, norb)
                  IF (i12 >= irange(1) .AND. i12 <= irange(2)) THEN
                     i12l = i12-irange(1)+1
                     irptr = eri%rowptr_local(i12l)
                     nindex = eri%nzerow_local(i12l)
                     colind(1:nindex) = eri%colind_local(irptr:irptr+nindex-1)
                     erival(1:nindex) = eri%nzval_local%r_dp(irptr:irptr+nindex-1)
                  ELSE
                     erival = 0.0_dp
                     colind = 0
                     nindex = 0
                  END IF
                  CALL mp_sum(nindex, eri%mp_group)
                  CALL mp_sum(colind(1:nindex), eri%mp_group)
                  CALL mp_sum(erival(1:nindex), eri%mp_group)
                  CALL mp_sync(eri%mp_group)
                  IF (iw > 0) THEN
                     DO i34l = 1, nindex
                        i34 = colind(i34l)
                        erint = erival(i34l)
                        CALL get_index_pair(i34, norb, i3, i4)
                        WRITE (iw, "(ES23.16,4I4)") erint, i1, i2, i3, i4
                     END DO
                  END IF
               END DO
            END DO
            DEALLOCATE (erival, colind)
            ! Print integrals: Fij
            IF (iw > 0) THEN
               i3 = 0; i4 = 0
               DO i1 = 1, norb
                  DO i2 = i1, norb
                     WRITE (iw, "(ES23.16,4I4)") ks_mat(i1, i2), i1, i2, i3, i4
                  END DO
               END DO
            END IF
            ! Print energy
            i1 = 0; i2 = 0; i3 = 0; i4 = 0
            IF (iw > 0) WRITE (iw, "(ES23.16,4I4)") esub, i1, i2, i3, i4
            DEALLOCATE (ks_mat, ks_ref, p_mat)
         ELSE
            CALL cp__b("qs_active_space_methods.F",1642,"Option not yet available")
         END IF
         !
         CALL cp_print_key_finished_output(iw, logger, as_input, "FCIDUMP")
      END IF

   END SUBROUTINE fcidump
! **************************************************************************************************
!> \brief Creates a local basis
!> \param pro_basis_set ...
!> \param zval ...
!> \param ishell ...
!> \param nshell ...
!> \param lnam ...
!> \par History
!>      05.2016 created [JGH]
! **************************************************************************************************
   SUBROUTINE create_pro_basis(pro_basis_set, zval, ishell, nshell, lnam)
      TYPE(gto_basis_set_type), POINTER                  :: pro_basis_set
      INTEGER, INTENT(IN)                                :: zval, ishell
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nshell
      CHARACTER(len=*), DIMENSION(:), INTENT(IN)         :: lnam

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pro_basis', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=6), DIMENSION(:), POINTER            :: sym
      INTEGER                                            :: i, l, nj
      INTEGER, DIMENSION(4, 7)                           :: ne
      INTEGER, DIMENSION(:), POINTER                     :: lq, nq
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet
      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set

      IF(.NOT.(.NOT. ASSOCIATED(pro_basis_set)))CALL cp__a("qs_active_space_methods.F",1675)
      NULLIFY (sto_basis_set)

      ! electronic configuration
      ne = 0
      DO l = 1, 4 !lq(1)+1
         nj = 2*(l-1)+1
         DO i = l, 7 ! nq(1)
            ne(l, i) = ptable(zval)%e_conv(l-1)-2*nj*(i-l)
            ne(l, i) = MAX(ne(l, i), 0)
            ne(l, i) = MIN(ne(l, i), 2*nj)
         END DO
      END DO
      ALLOCATE (nq(ishell), lq(ishell), zet(ishell), sym(ishell))
      DO i = 1, ishell
         nq(i) = nshell(i)
         SELECT CASE (lnam (i))
         CASE ('S', 's')
            lq(i) = 0
         CASE ('P', 'p')
            lq(i) = 1
         CASE ('D', 'd')
            lq(i) = 2
         CASE ('F', 'f')
            lq(i) = 3
         CASE DEFAULT
            CALL cp__b("qs_active_space_methods.F",1701,"Wrong l QN")
         END SELECT
         sym(i) = lnam(i)
         zet(i) = srules(zval, ne, nq(1), lq(1))
      END DO
      CALL allocate_sto_basis_set(sto_basis_set)
      CALL set_sto_basis_set(sto_basis_set, nshell=1, nq=nq, lq=lq, zet=zet, symbol=sym)
      CALL create_gto_from_sto_basis(sto_basis_set, pro_basis_set, 6)
      pro_basis_set%norm_type = 2
      CALL init_orb_basis_set(pro_basis_set)
      CALL deallocate_sto_basis_set(sto_basis_set)

   END SUBROUTINE create_pro_basis

END MODULE qs_active_space_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set of routines handling the localization for molecular properties
! **************************************************************************************************
MODULE qs_loc_molecules
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_max,&
                                              mp_minloc,&
                                              mp_sum
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE moments_utils,                   ONLY: get_reference_point
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: debye
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_loc_types,                    ONLY: qs_loc_env_new_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public ***
   PUBLIC :: wfc_to_molecule

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_loc_molecules'

CONTAINS

! **************************************************************************************************
!> \brief maps wfc's to molecules and also prints molecular dipoles
!> \param qs_env the qs_env in which the qs_env lives
!> \param qs_loc_env ...
!> \param loc_print_key ...
!> \param center ...
!> \param molecule_set ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE wfc_to_molecule(qs_env, qs_loc_env, loc_print_key, center, &
                              molecule_set, nspins)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_loc_env_new_type), INTENT(IN)              :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: loc_print_key
      REAL(KIND=dp), INTENT(IN)                          :: center(:, :)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'wfc_to_molecule', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: zeta
      COMPLEX(KIND=dp), DIMENSION(3)                     :: ggamma, zphase
      INTEGER :: akind, counter, first_atom, i, iatom, ikind, imol, imol_now, istate, j, &
         local_location, natom, natom_loc, natom_max, nkind, nmol, nstate, output_unit, reference
      INTEGER, POINTER                                   :: wfc_to_atom_map(:)
      LOGICAL                                            :: do_berry, ghost
      REAL(KIND=dp)                                      :: charge_tot, dipole(3), dr(3), mydist(2), &
                                                            ria(3), theta, zeff, zwfc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: charge_set
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dipole_set
      REAL(KIND=dp), DIMENSION(3)                        :: ci, gvec, rcc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      REAL(KIND=dp), POINTER                             :: distance(:), r(:, :)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      logger => cp_get_default_logger()

      ! Molecular Dipoles availables only for nspin == 1
      IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_key, &
                                           "MOLECULAR_DIPOLES"), cp_p_file)) THEN
         IF (nspins > 1) THEN
            CALL cp_abort(cp__l("qs_loc_molecules.F",104), &
                          "Molecular Dipoles not implemented for SPIN multiplicity "// &
                          "larger than 1!")
         END IF

         ! Setup reference point and some warning..
         reference = section_get_ival(loc_print_key, keyword_name="MOLECULAR_DIPOLES%REFERENCE")
         CALL section_vals_val_get(loc_print_key, "MOLECULAR_DIPOLES%REF_POINT", r_vals=ref_point)
         CALL section_vals_val_get(loc_print_key, "MOLECULAR_DIPOLES%PERIODIC", l_val=do_berry)
      END IF
      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set)
      particle_set => qs_loc_env%particle_set
      para_env => qs_loc_env%para_env
      local_molecules => qs_loc_env%local_molecules
      nstate = SIZE(center, 2)
      ALLOCATE (wfc_to_atom_map(nstate))
      !---------------------------------------------------------------------------
      !---------------------------------------------------------------------------
      nkind = SIZE(local_molecules%n_el)
      natom = 0
      natom_max = 0
      DO ikind = 1, nkind
         nmol = SIZE(local_molecules%list(ikind)%array)
         DO imol = 1, nmol
            i = local_molecules%list(ikind)%array(imol)
            molecule_kind => molecule_set(i)%molecule_kind
            CALL get_molecule_kind(molecule_kind=molecule_kind, natom=natom)
            natom_max = natom_max+natom
            IF (.NOT. ASSOCIATED(molecule_set(i)%lmi)) THEN
               ALLOCATE (molecule_set(i)%lmi)
               NULLIFY (molecule_set(i)%lmi%states)
            ENDIF
            molecule_set(i)%lmi%nstates = 0
            IF (ASSOCIATED(molecule_set(i)%lmi%states)) THEN
               DEALLOCATE (molecule_set(i)%lmi%states)
            END IF
         END DO
      END DO
      natom_loc = natom_max
      natom = natom_max

      CALL mp_max(natom_max, para_env%group)

      ALLOCATE (r(3, natom_max))

      ALLOCATE (distance(natom_max))

      !Zero all the stuff
      r(:, :) = 0.0_dp
      distance(:) = 1.E10_dp

      !---------------------------------------------------------------------------
      !---------------------------------------------------------------------------
      counter = 0
      nkind = SIZE(local_molecules%n_el)
      DO ikind = 1, nkind
         nmol = SIZE(local_molecules%list(ikind)%array)
         DO imol = 1, nmol
            i = local_molecules%list(ikind)%array(imol)
            molecule_kind => molecule_set(i)%molecule_kind
            first_atom = molecule_set(i)%first_atom
            CALL get_molecule_kind(molecule_kind=molecule_kind, natom=natom)

            DO iatom = 1, natom
               counter = counter+1
               r(:, counter) = particle_set(first_atom+iatom-1)%r(:)
            END DO
         END DO
      END DO

      !---------------------------------------------------------------------------
      !---------------------------------------------------------------------------
      DO istate = 1, nstate
         distance(:) = 1.E10_dp
         DO iatom = 1, natom_loc
            dr(1) = r(1, iatom)-center(1, istate)
            dr(2) = r(2, iatom)-center(2, istate)
            dr(3) = r(3, iatom)-center(3, istate)
            ria = pbc(dr, qs_loc_env%cell)
            distance(iatom) = SQRT(DOT_PRODUCT(ria, ria))
         END DO

         !combine distance() from all procs
         local_location = MAX(1, MINLOC(distance, DIM=1))

         mydist(1) = distance(local_location)
         mydist(2) = para_env%mepos

         CALL mp_minloc(mydist, para_env%group)

         IF (mydist(2) == para_env%mepos) THEN
            wfc_to_atom_map(istate) = local_location
         ELSE
            wfc_to_atom_map(istate) = 0
         END IF
      END DO
      !---------------------------------------------------------------------------
      !---------------------------------------------------------------------------
      IF (natom_loc /= 0) THEN
         DO istate = 1, nstate
            iatom = wfc_to_atom_map(istate)
            IF (iatom /= 0) THEN
               counter = 0
               nkind = SIZE(local_molecules%n_el)
               DO ikind = 1, nkind
                  nmol = SIZE(local_molecules%list(ikind)%array)
                  DO imol = 1, nmol
                     imol_now = local_molecules%list(ikind)%array(imol)
                     molecule_kind => molecule_set(imol_now)%molecule_kind
                     CALL get_molecule_kind(molecule_kind=molecule_kind, natom=natom)
                     counter = counter+natom
                     IF (counter >= iatom) EXIT
                  END DO
                  IF (counter >= iatom) EXIT
               END DO
               i = molecule_set(imol_now)%lmi%nstates
               i = i+1
               molecule_set(imol_now)%lmi%nstates = i
               CALL reallocate(molecule_set(imol_now)%lmi%states, 1, i)
               molecule_set(imol_now)%lmi%states(i) = istate
            END IF
         END DO
      END IF

      !---------------------------------------------------------------------------
      ! Figure out dipole of the molecule.
      !---------------------------------------------------------------------------
      IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_key, &
                                           "MOLECULAR_DIPOLES"), cp_p_file)) THEN

         particle_set => qs_loc_env%particle_set
         para_env => qs_loc_env%para_env
         local_molecules => qs_loc_env%local_molecules
         nstate = SIZE(center, 2)
         ALLOCATE (dipole_set(3, SIZE(molecule_set)))
         ALLOCATE (charge_set(SIZE(molecule_set)))
         dipole_set = 0.0_dp
         charge_set = 0.0_dp
         cell => qs_loc_env%cell
         zwfc = 3.0_dp-REAL(nspins, KIND=dp)
         nkind = SIZE(local_molecules%n_el)
         DO ikind = 1, nkind ! loop over different molecules
            nmol = SIZE(local_molecules%list(ikind)%array)
            DO imol = 1, nmol ! all the molecules of the kind
               imol_now = local_molecules%list(ikind)%array(imol) ! index in the global array
               IF (.NOT. ASSOCIATED(molecule_set(imol_now)%lmi%states)) CYCLE
               molecule_kind => molecule_set(imol_now)%molecule_kind
               first_atom = molecule_set(imol_now)%first_atom
               CALL get_molecule_kind(molecule_kind=molecule_kind, natom=natom)

               ! Get reference point for this molecule
               CALL get_reference_point(rcc, qs_env=qs_env, reference=reference, &
                                        ref_point=ref_point, ifirst=first_atom, &
                                        ilast=first_atom+natom-1)

               dipole = 0.0_dp
               IF (do_berry) THEN
                  rcc = pbc(rcc, cell)
                  ! Find out the total charge of the molecule
                  DO iatom = 1, natom
                     i = first_atom+iatom-1
                     atomic_kind => particle_set(i)%atomic_kind
                     CALL get_atomic_kind(atomic_kind, kind_number=akind)
                     CALL get_qs_kind(qs_kind_set(akind), ghost=ghost)
                     IF (.NOT. ghost) THEN
                        CALL get_qs_kind(qs_kind_set(akind), core_charge=zeff)
                        charge_set(imol_now) = charge_set(imol_now)+zeff
                     END IF
                  END DO
                  ! Charges of the wfc involved
                  DO istate = 1, SIZE(molecule_set(imol_now)%lmi%states)
                     charge_set(imol_now) = charge_set(imol_now)-zwfc
                  ENDDO

                  charge_tot = charge_set(imol_now)
                  ria = twopi*MATMUL(cell%h_inv, rcc)
                  zphase = CMPLX(COS(ria), SIN(ria), KIND=dp)**charge_tot
                  ggamma = CMPLX(1.0_dp, 0.0_dp, KIND=dp)

                  ! Nuclear charges
                  DO iatom = 1, natom
                     i = first_atom+iatom-1
                     atomic_kind => particle_set(i)%atomic_kind
                     CALL get_atomic_kind(atomic_kind, kind_number=akind)
                     CALL get_qs_kind(qs_kind_set(akind), ghost=ghost)
                     IF (.NOT. ghost) THEN
                        CALL get_qs_kind(qs_kind_set(akind), core_charge=zeff)
                        ria = pbc(particle_set(i)%r, cell)
                        DO j = 1, 3
                           gvec = twopi*cell%h_inv(j, :)
                           theta = SUM(ria(:)*gvec(:))
                           zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(zeff)
                           ggamma(j) = ggamma(j)*zeta
                        END DO
                     END IF
                  END DO

                  ! Charges of the wfc involved
                  DO istate = 1, SIZE(molecule_set(imol_now)%lmi%states)
                     i = molecule_set(imol_now)%lmi%states(istate)
                     ria = pbc(center(1:3, i), cell)
                     DO j = 1, 3
                        gvec = twopi*cell%h_inv(j, :)
                        theta = SUM(ria(:)*gvec(:))
                        zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(-zwfc)
                        ggamma(j) = ggamma(j)*zeta
                     END DO
                  ENDDO

                  ggamma = ggamma*zphase
                  ci = AIMAG(LOG(ggamma))/twopi
                  dipole = MATMUL(cell%hmat, ci)
               ELSE
                  ! Nuclear charges
                  DO iatom = 1, natom
                     i = first_atom+iatom-1
                     atomic_kind => particle_set(i)%atomic_kind
                     CALL get_atomic_kind(atomic_kind, kind_number=akind)
                     CALL get_qs_kind(qs_kind_set(akind), ghost=ghost)
                     IF (.NOT. ghost) THEN
                        CALL get_qs_kind(qs_kind_set(akind), core_charge=zeff)
                        ria = pbc(particle_set(i)%r, cell)-rcc
                        dipole = dipole+zeff*(ria-rcc)
                        charge_set(imol_now) = charge_set(imol_now)+zeff
                     END IF
                  END DO
                  ! Charges of the wfc involved
                  DO istate = 1, SIZE(molecule_set(imol_now)%lmi%states)
                     i = molecule_set(imol_now)%lmi%states(istate)
                     ria = pbc(center(1:3, i), cell)
                     dipole = dipole-zwfc*(ria-rcc)
                     charge_set(imol_now) = charge_set(imol_now)-zwfc
                  ENDDO
               END IF
               dipole_set(:, imol_now) = dipole ! a.u.
            ENDDO
         ENDDO
         CALL mp_sum(dipole_set, para_env%group)
         CALL mp_sum(charge_set, para_env%group)

         output_unit = cp_print_key_unit_nr(logger, loc_print_key, "MOLECULAR_DIPOLES", &
                                            extension=".MolDip", middle_name="MOLECULAR_DIPOLES")
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT='(A80)') "# molecule nr, charge, dipole vector, dipole (Debye) "
            dipole_set(:, :) = dipole_set(:, :)*debye ! Debye
            DO I = 1, SIZE(dipole_set, 2)
               WRITE (UNIT=output_unit, FMT='(I6,5F12.6)') I, charge_set(I), dipole_set(1:3, I), &
                  SQRT(DOT_PRODUCT(dipole_set(1:3, I), dipole_set(1:3, I)))
            ENDDO
         ENDIF
         CALL cp_print_key_finished_output(output_unit, logger, loc_print_key, &
                                           "MOLECULAR_DIPOLES")
         DEALLOCATE (dipole_set, charge_set)
      END IF
      !---------------------------------------------------------------------------
      ! end of molecular dipole calculation
      !---------------------------------------------------------------------------

      DEALLOCATE (distance)
      DEALLOCATE (r)

      DEALLOCATE (wfc_to_atom_map)

   END SUBROUTINE wfc_to_molecule
   !------------------------------------------------------------------------------

END MODULE qs_loc_molecules

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief provides a table for UFF vdW radii:
!>        Rappe et al. J. Am. Chem. Soc. 114, 10024 (1992)
!> \par History
!>      created 2015/10/28
!> \author Lianheng Tong (lianheng.tong [AT] kcl.ac.uk)
! **************************************************************************************************
MODULE uff_vdw_radii_table

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: get_uff_vdw_radius

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'uff_vdw_radii_table'

   INTEGER, PARAMETER :: NELEMENTS = 118
   REAL(KIND=dp), PARAMETER ::  NAN = -1.0_dp
   ! the table: units are in Angstrom, in the order of z
   REAL(KIND=dp), DIMENSION(NELEMENTS), PARAMETER :: uff_vdw_radius = &
                                                     (/ &
                                                     1.443_dp, & !  1 H
                                                     1.181_dp, & !  2 He
                                                     1.226_dp, & !  3 Li
                                                     1.373_dp, & !  4 Be
                                                     2.042_dp, & !  5 B
                                                     1.926_dp, & !  6 C
                                                     1.830_dp, & !  7 N
                                                     1.750_dp, & !  8 O
                                                     1.682_dp, & !  9 F
                                                     1.621_dp, & ! 10 Ne
                                                     1.492_dp, & ! 11 Na
                                                     1.510_dp, & ! 12 Mg
                                                     2.250_dp, & ! 13 Al
                                                     2.148_dp, & ! 14 Si
                                                     2.074_dp, & ! 15 P
                                                     2.017_dp, & ! 16 S
                                                     1.973_dp, & ! 17 Cl
                                                     1.934_dp, & ! 18 Ar
                                                     1.906_dp, & ! 19 K
                                                     1.700_dp, & ! 20 Ca
                                                     1.647_dp, & ! 21 Sc
                                                     1.587_dp, & ! 22 Ti
                                                     1.572_dp, & ! 23 V
                                                     1.511_dp, & ! 24 Cr
                                                     1.481_dp, & ! 25 Mn
                                                     1.456_dp, & ! 26 Fe
                                                     1.436_dp, & ! 27 Co
                                                     1.417_dp, & ! 28 Ni
                                                     1.748_dp, & ! 29 Cu
                                                     1.382_dp, & ! 30 Zn
                                                     2.191_dp, & ! 31 Ga
                                                     2.140_dp, & ! 32 Ge
                                                     2.115_dp, & ! 33 As
                                                     2.103_dp, & ! 34 Se
                                                     2.094_dp, & ! 35 Br
                                                     2.071_dp, & ! 36 Kr
                                                     2.057_dp, & ! 37 Rb
                                                     1.821_dp, & ! 38 Sr
                                                     1.673_dp, & ! 39 Y
                                                     1.562_dp, & ! 40 Zr
                                                     1.582_dp, & ! 41 Nb
                                                     1.526_dp, & ! 42 Mo
                                                     1.499_dp, & ! 43 Tc
                                                     1.482_dp, & ! 44 Ru
                                                     1.464_dp, & ! 45 Rh
                                                     1.450_dp, & ! 46 Pd
                                                     1.574_dp, & ! 47 Ag
                                                     1.424_dp, & ! 48 Cd
                                                     2.231_dp, & ! 49 In
                                                     2.196_dp, & ! 50 Sn
                                                     2.210_dp, & ! 51 Sb
                                                     2.235_dp, & ! 52 Te
                                                     2.250_dp, & ! 53 I
                                                     2.202_dp, & ! 54 Xe
                                                     2.259_dp, & ! 55 Cs
                                                     1.852_dp, & ! 56 Ba
                                                     1.761_dp, & ! 57 La
                                                     1.778_dp, & ! 58 Ce
                                                     1.803_dp, & ! 59 Pr
                                                     1.788_dp, & ! 60 Nd
                                                     1.774_dp, & ! 61 Pm
                                                     1.760_dp, & ! 62 Sm
                                                     1.747_dp, & ! 63 Eu
                                                     1.684_dp, & ! 64 Gd
                                                     1.725_dp, & ! 65 Tb
                                                     1.714_dp, & ! 66 Dy
                                                     1.705_dp, & ! 67 Ho
                                                     1.695_dp, & ! 68 Er
                                                     1.687_dp, & ! 69 Tm
                                                     1.678_dp, & ! 70 Yb
                                                     1.820_dp, & ! 71 Lu
                                                     1.570_dp, & ! 72 Hf
                                                     1.585_dp, & ! 73 Ta
                                                     1.534_dp, & ! 74 W
                                                     1.477_dp, & ! 75 Re
                                                     1.560_dp, & ! 76 Os
                                                     1.420_dp, & ! 77 Ir
                                                     1.377_dp, & ! 78 Pt
                                                     1.646_dp, & ! 79 Au
                                                     1.352_dp, & ! 80 Hg
                                                     2.174_dp, & ! 81 Tl
                                                     2.149_dp, & ! 82 Pb
                                                     2.185_dp, & ! 83 Bi
                                                     2.354_dp, & ! 84 Po
                                                     2.375_dp, & ! 85 At
                                                     2.382_dp, & ! 86 Rn
                                                     2.450_dp, & ! 87 Fr
                                                     1.838_dp, & ! 88 Ra
                                                     1.739_dp, & ! 89 Ac
                                                     1.698_dp, & ! 90 Th
                                                     1.712_dp, & ! 91 Pa
                                                     1.697_dp, & ! 92 U
                                                     1.712_dp, & ! 93 Np
                                                     1.712_dp, & ! 94 Pu
                                                     1.690_dp, & ! 95 Am
                                                     1.663_dp, & ! 96 Cm
                                                     1.670_dp, & ! 97 Bk
                                                     1.656_dp, & ! 98 Cf
                                                     1.649_dp, & ! 99 Es
                                                     1.643_dp, & !100 Fm
                                                     1.637_dp, & !101 Md
                                                     1.624_dp, & !102 No
                                                     1.618_dp, & !103 Lr
                                                     NAN, & !104 Rf
                                                     NAN, & !105 Db
                                                     NAN, & !106 Sg
                                                     NAN, & !107 Bh
                                                     NAN, & !108 Hs
                                                     NAN, & !109 Mt
                                                     NAN, & !110 Ds
                                                     NAN, & !111 Rg
                                                     NAN, & !112 Cn
                                                     NAN, & !113 Uut
                                                     NAN, & !114 Fl
                                                     NAN, & !115 Uup
                                                     NAN, & !116 Lv
                                                     NAN, & !117 Uus
                                                     NAN & !118 Uuo
                                                     /)

CONTAINS

! **************************************************************************************************
!> \brief get UFF vdW radius for a given element
!> \param z      : on input, atomic number
!> \param radius : on output, UFF vdW radius, 0.0 if element not in table
!> \param found  : on output, T if element is in table, F if not
!> \author Lianheng Tong (lianheng.tong [AT] kcl.ac.uk)
! **************************************************************************************************
   PURE SUBROUTINE get_uff_vdw_radius(z, radius, found)
      INTEGER, INTENT(IN)                                :: z
      REAL(KIND=dp), INTENT(OUT)                         :: radius
      LOGICAL, INTENT(OUT)                               :: found

      CHARACTER(len=*), PARAMETER :: routineN = 'get_uff_vdw_radius', &
         routineP = moduleN//':'//routineN

      IF (uff_vdw_radius(z) == NAN) THEN
         radius = 0.0_dp
         found = .FALSE.
      ELSE
         radius = uff_vdw_radius(z)
         found = .TRUE.
      END IF
   END SUBROUTINE get_uff_vdw_radius

END MODULE uff_vdw_radii_table
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief provides a resp fit for gas phase systems
!> \par History
!>      created
!>      Dorothea Golze [06.2012] (1) extension to periodic systems
!>                               (2) re-structured the code
!> \author Joost VandeVondele (02.2007)
! **************************************************************************************************
MODULE qs_resp
   USE atomic_charges,                  ONLY: print_atomic_charges
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE bibliography,                    ONLY: Campana2009,&
                                              Golze2015,&
                                              Rappe1992,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc,&
                                              use_perd_none,&
                                              use_perd_xyz
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE input_constants,                 ONLY: do_resp_minus_x_dir,&
                                              do_resp_minus_y_dir,&
                                              do_resp_minus_z_dir,&
                                              do_resp_x_dir,&
                                              do_resp_y_dir,&
                                              do_resp_z_dir,&
                                              use_cambridge_vdw_radii,&
                                              use_uff_vdw_radii
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_get_lval,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: pi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_irecv,&
                                              mp_isend,&
                                              mp_sum,&
                                              mp_wait
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE periodic_table,                  ONLY: get_ptable_info
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE qs_collocate_density,            ONLY: calculate_rho_resp_all,&
                                              calculate_rho_resp_single
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE uff_vdw_radii_table,             ONLY: get_uff_vdw_radius
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_resp'

   PUBLIC :: resp_fit

   TYPE resp_type
      LOGICAL                                :: equal_charges, itc, &
                                                molecular_sys, rheavies, &
                                                use_repeat_method
      INTEGER                                :: nres, ncons, &
                                                nrest_sec, ncons_sec, &
                                                npoints, stride(3), my_fit, &
                                                npoints_proc, &
                                                auto_vdw_radii_table
      INTEGER, DIMENSION(:), POINTER         :: atom_surf_list
      INTEGER, DIMENSION(:, :), POINTER       :: fitpoints
      REAL(KIND=dp)                          :: rheavies_strength, &
                                                length, eta, &
                                                sum_vhartree, offset
      REAL(KIND=dp), DIMENSION(3)            :: box_hi, box_low
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rmin_kind, &
                                                rmax_kind
      REAL(KIND=dp), DIMENSION(:), POINTER   :: range_surf
      REAL(KIND=dp), DIMENSION(:), POINTER   :: rhs
      REAL(KIND=dp), DIMENSION(:), POINTER   :: sum_vpot
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: matrix
   END TYPE resp_type

   TYPE resp_p_type
      TYPE(resp_type), POINTER              ::  p_resp
   END TYPE resp_p_type

CONTAINS

! **************************************************************************************************
!> \brief performs resp fit and generates RESP charges
!> \param qs_env the qs environment
! **************************************************************************************************
   SUBROUTINE resp_fit(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'resp_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info, my_per, natom, nvar, &
                                                            output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      LOGICAL                                            :: has_resp
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(section_vals_type), POINTER                   :: cons_section, input, poisson_section, &
                                                            resp_section, rest_section

      CALL timeset(routineN, handle)

      NULLIFY (logger, atomic_kind_set, cell, subsys, particles, particle_set, input, &
               resp_section, cons_section, rest_section, poisson_section, resp_env, rep_sys)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_resp.F",163)

      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, input=input, &
                      subsys=subsys, particle_set=particle_set, cell=cell)
      resp_section => section_vals_get_subs_vals(input, "PROPERTIES%RESP")
      CALL section_vals_get(resp_section, explicit=has_resp)

      IF (has_resp) THEN
         logger => cp_get_default_logger()
         poisson_section => section_vals_get_subs_vals(input, "DFT%POISSON")
         CALL section_vals_val_get(poisson_section, "PERIODIC", i_val=my_per)
         CALL create_resp_type(resp_env, rep_sys)
         !initialize the RESP fitting, get all the keywords
         CALL init_resp(resp_env, rep_sys, subsys, atomic_kind_set, &
                        cell, resp_section, cons_section, rest_section)

         !print info
         CALL print_resp_parameter_info(qs_env, resp_env, rep_sys, my_per)

         CALL qs_subsys_get(subsys, particles=particles)
         natom = particles%n_els
         nvar = natom+resp_env%ncons

         CALL resp_allocate(resp_env, natom, nvar)
         ALLOCATE (ipiv(nvar))
         ipiv = 0

         ! calculate the matrix and the vector rhs
         SELECT CASE (my_per)
         CASE (use_perd_none)
            CALL calc_resp_matrix_nonper(qs_env, resp_env, atomic_kind_set, particles, cell, &
                                         resp_env%matrix, resp_env%rhs, natom)
         CASE (use_perd_xyz)
            CALL cite_reference(Golze2015)
            IF (resp_env%use_repeat_method) CALL cite_reference(Campana2009)
            CALL calc_resp_matrix_periodic(qs_env, resp_env, rep_sys, particles, cell, natom)
         CASE DEFAULT
            CALL cp_abort(cp__l("qs_resp.F",200), &
                          "RESP charges only implemented for nonperiodic systems"// &
                          " or XYZ periodicity!")
         END SELECT

         output_unit = cp_print_key_unit_nr(logger, resp_section, "PRINT%PROGRAM_RUN_INFO", &
                                            extension=".resp")
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T3,A,T69,I12)') "Number of fitting points "// &
               "found: ", resp_env%npoints
            WRITE (output_unit, '()')
         ENDIF

         !adding restraints and constraints
         CALL add_restraints_and_constraints(qs_env, resp_env, rest_section, &
                                             subsys, natom, cons_section, particle_set)

         !solve system for the values of the charges and the lagrangian multipliers
         CALL DGETRF(nvar, nvar, resp_env%matrix, nvar, ipiv, info)
         IF(.NOT.(info == 0))CALL cp__a("qs_resp.F",219)

         CALL DGETRS('N', nvar, 1, resp_env%matrix, nvar, ipiv, resp_env%rhs, nvar, info)
         IF(.NOT.(info == 0))CALL cp__a("qs_resp.F",222)

         IF (resp_env%use_repeat_method) resp_env%offset = resp_env%rhs(natom+1)
         CALL print_resp_charges(qs_env, resp_env, output_unit, natom)
         CALL print_fitting_points(qs_env, resp_env)
         CALL print_pot_from_resp_charges(qs_env, resp_env, particles, natom, output_unit)

         DEALLOCATE (ipiv)
         CALL resp_dealloc(resp_env, rep_sys)
         CALL cp_print_key_finished_output(output_unit, logger, resp_section, &
                                           "PRINT%PROGRAM_RUN_INFO")

      END IF

      CALL timestop(handle)

   END SUBROUTINE resp_fit

! **************************************************************************************************
!> \brief creates the resp_type structure
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
! **************************************************************************************************
   SUBROUTINE create_resp_type(resp_env, rep_sys)
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys

      CHARACTER(len=*), PARAMETER :: routineN = 'create_resp_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(resp_env)) CALL resp_dealloc(resp_env, rep_sys)
      ALLOCATE (resp_env)

      NULLIFY (resp_env%matrix, &
               resp_env%fitpoints, &
               resp_env%rmin_kind, &
               resp_env%rmax_kind, &
               resp_env%rhs, &
               resp_env%sum_vpot)

      resp_env%equal_charges = .FALSE.
      resp_env%itc = .FALSE.
      resp_env%molecular_sys = .FALSE.
      resp_env%rheavies = .FALSE.
      resp_env%use_repeat_method = .FALSE.

      resp_env%box_hi = 0.0_dp
      resp_env%box_low = 0.0_dp

      resp_env%ncons = 0
      resp_env%ncons_sec = 0
      resp_env%nres = 0
      resp_env%nrest_sec = 0
      resp_env%npoints = 0
      resp_env%npoints_proc = 0
      resp_env%auto_vdw_radii_table = use_cambridge_vdw_radii

   END SUBROUTINE create_resp_type

! **************************************************************************************************
!> \brief allocates the resp
!> \param resp_env the resp environment
!> \param natom ...
!> \param nvar ...
! **************************************************************************************************
   SUBROUTINE resp_allocate(resp_env, natom, nvar)
      TYPE(resp_type), POINTER                           :: resp_env
      INTEGER, INTENT(IN)                                :: natom, nvar

      CHARACTER(len=*), PARAMETER :: routineN = 'resp_allocate', routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(resp_env%matrix)) THEN
         ALLOCATE (resp_env%matrix(nvar, nvar))
      ENDIF
      IF (.NOT. ASSOCIATED(resp_env%rhs)) THEN
         ALLOCATE (resp_env%rhs(nvar))
      ENDIF
      IF (.NOT. ASSOCIATED(resp_env%sum_vpot)) THEN
         ALLOCATE (resp_env%sum_vpot(natom))
      ENDIF
      resp_env%matrix = 0.0_dp
      resp_env%rhs = 0.0_dp
      resp_env%sum_vpot = 0.0_dp

   END SUBROUTINE resp_allocate

! **************************************************************************************************
!> \brief deallocates the resp_type structure
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
! **************************************************************************************************
   SUBROUTINE resp_dealloc(resp_env, rep_sys)
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys

      CHARACTER(len=*), PARAMETER :: routineN = 'resp_dealloc', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(resp_env)) THEN
         IF (ASSOCIATED(resp_env%matrix)) THEN
            DEALLOCATE (resp_env%matrix)
         ENDIF
         IF (ASSOCIATED(resp_env%rhs)) THEN
            DEALLOCATE (resp_env%rhs)
         ENDIF
         IF (ASSOCIATED(resp_env%sum_vpot)) THEN
            DEALLOCATE (resp_env%sum_vpot)
         ENDIF
         IF (ASSOCIATED(resp_env%fitpoints)) THEN
            DEALLOCATE (resp_env%fitpoints)
         ENDIF
         IF (ASSOCIATED(resp_env%rmin_kind)) THEN
            DEALLOCATE (resp_env%rmin_kind)
         ENDIF
         IF (ASSOCIATED(resp_env%rmax_kind)) THEN
            DEALLOCATE (resp_env%rmax_kind)
         ENDIF
         DEALLOCATE (resp_env)
      ENDIF
      IF (ASSOCIATED(rep_sys)) THEN
         DO i = 1, SIZE(rep_sys)
            DEALLOCATE (rep_sys(i)%p_resp%atom_surf_list)
            DEALLOCATE (rep_sys(i)%p_resp)
         ENDDO
         DEALLOCATE (rep_sys)
      ENDIF

   END SUBROUTINE resp_dealloc

! **************************************************************************************************
!> \brief intializes the resp fit. Getting the parameters
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
!> \param subsys ...
!> \param atomic_kind_set ...
!> \param cell parameters related to the simulation cell
!> \param resp_section resp section
!> \param cons_section constraints section, part of resp section
!> \param rest_section restraints section, part of resp section
! **************************************************************************************************
   SUBROUTINE init_resp(resp_env, rep_sys, subsys, atomic_kind_set, &
                        cell, resp_section, cons_section, rest_section)

      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: resp_section, cons_section, rest_section

      CHARACTER(len=*), PARAMETER :: routineN = 'init_resp', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, nrep
      INTEGER, DIMENSION(:), POINTER                     :: atom_list_cons, my_stride
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: slab_section, sphere_section

      CALL timeset(routineN, handle)

      NULLIFY (atom_list_cons, my_stride, sphere_section, slab_section)

      ! get the subsections
      sphere_section => section_vals_get_subs_vals(resp_section, "SPHERE_SAMPLING")
      slab_section => section_vals_get_subs_vals(resp_section, "SLAB_SAMPLING")
      cons_section => section_vals_get_subs_vals(resp_section, "CONSTRAINT")
      rest_section => section_vals_get_subs_vals(resp_section, "RESTRAINT")

      ! get the general keywords
      CALL section_vals_val_get(resp_section, "INTEGER_TOTAL_CHARGE", &
                                l_val=resp_env%itc)
      IF (resp_env%itc) resp_env%ncons = resp_env%ncons+1

      CALL section_vals_val_get(resp_section, "RESTRAIN_HEAVIES_TO_ZERO", &
                                l_val=resp_env%rheavies)
      IF (resp_env%rheavies) THEN
         CALL section_vals_val_get(resp_section, "RESTRAIN_HEAVIES_STRENGTH", &
                                   r_val=resp_env%rheavies_strength)
      ENDIF
      CALL section_vals_val_get(resp_section, "STRIDE", i_vals=my_stride)
      IF (SIZE(my_stride) /= 1 .AND. SIZE(my_stride) /= 3) &
         CALL cp_abort(cp__l("qs_resp.F",403), "STRIDE keyword can accept only 1 (the same for X,Y,Z) "// &
                       "or 3 values. Correct your input file.")
      IF (SIZE(my_stride) == 1) THEN
         DO i = 1, 3
            resp_env%stride(i) = my_stride(1)
         END DO
      ELSE
         resp_env%stride = my_stride(1:3)
      END IF
      CALL section_vals_val_get(resp_section, "WIDTH", r_val=resp_env%eta)

      ! get if the user wants to use REPEAT method
      CALL section_vals_val_get(resp_section, "USE_REPEAT_METHOD", &
                                l_val=resp_env%use_repeat_method)
      IF (resp_env%use_repeat_method) THEN
         resp_env%ncons = resp_env%ncons+1
         ! restrain heavies should be off
         resp_env%rheavies = .FALSE.
      END IF

      ! get and set the parameters for molecular (non-surface) systems
      ! this must come after the repeat settings being set
      CALL get_parameter_molecular_sys(resp_env, sphere_section, cell, &
                                       atomic_kind_set)

      ! get the parameter for periodic/surface systems
      CALL section_vals_get(slab_section, explicit=explicit, n_repetition=nrep)
      IF (explicit) THEN
         IF (resp_env%molecular_sys) THEN
            CALL cp_abort(cp__l("qs_resp.F",432), &
                          "You can only use either SPHERE_SAMPLING or SLAB_SAMPLING, but "// &
                          "not both.")
         ENDIF
         ALLOCATE (rep_sys(nrep))
         DO i = 1, nrep
            ALLOCATE (rep_sys(i)%p_resp)
            NULLIFY (rep_sys(i)%p_resp%range_surf, rep_sys(i)%p_resp%atom_surf_list)
            CALL section_vals_val_get(slab_section, "RANGE", r_vals=rep_sys(i)%p_resp%range_surf, &
                                      i_rep_section=i)
            CALL section_vals_val_get(slab_section, "LENGTH", r_val=rep_sys(i)%p_resp%length, &
                                      i_rep_section=i)
            CALL section_vals_val_get(slab_section, "SURF_DIRECTION", &
                                      i_rep_section=i, i_val=rep_sys(i)%p_resp%my_fit)
            IF (ANY(rep_sys(i)%p_resp%range_surf < 0.0_dp)) THEN
               CALL cp__b("qs_resp.F",447,"Numbers in RANGE in SLAB_SAMPLING cannot be negative.")
            END IF
            IF (rep_sys(i)%p_resp%length <= EPSILON(0.0_dp)) THEN
               CALL cp__b("qs_resp.F",450,"Parameter LENGTH in SLAB_SAMPLING has to be larger than zero.")
            END IF
            !list of atoms specifing the surface
            CALL build_atom_list(slab_section, subsys, rep_sys(i)%p_resp%atom_surf_list, rep=i)
         END DO
      END IF

      ! get the parameters for the constraint and restraint sections
      CALL section_vals_get(cons_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_get(cons_section, n_repetition=resp_env%ncons_sec)
         DO i = 1, resp_env%ncons_sec
            CALL section_vals_val_get(cons_section, "EQUAL_CHARGES", &
                                      l_val=resp_env%equal_charges, explicit=explicit)
            IF (.NOT. explicit) CYCLE
            CALL build_atom_list(cons_section, subsys, atom_list_cons, i)
            !instead of using EQUAL_CHARGES the constraint sections could be repeated
            resp_env%ncons = resp_env%ncons+SIZE(atom_list_cons)-2
            DEALLOCATE (atom_list_cons)
         END DO
      END IF
      CALL section_vals_get(rest_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_get(rest_section, n_repetition=resp_env%nrest_sec)
      END IF
      resp_env%ncons = resp_env%ncons+resp_env%ncons_sec
      resp_env%nres = resp_env%nres+resp_env%nrest_sec

      CALL timestop(handle)

   END SUBROUTINE init_resp

! **************************************************************************************************
!> \brief getting the parameters for nonperiodic/non-surface systems
!> \param resp_env the resp environment
!> \param sphere_section input section setting parameters for sampling
!>        fitting in spheres around the atom
!> \param cell parameters related to the simulation cell
!> \param atomic_kind_set ...
! **************************************************************************************************
   SUBROUTINE get_parameter_molecular_sys(resp_env, sphere_section, cell, &
                                          atomic_kind_set)

      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(section_vals_type), POINTER                   :: sphere_section
      TYPE(cell_type), POINTER                           :: cell
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'get_parameter_molecular_sys', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: symbol
      CHARACTER(LEN=default_string_length)               :: missing_rmax, missing_rmin
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: ikind, j, kind_number, n_rmax_missing, &
                                                            n_rmin_missing, nkind, nrep_rmax, &
                                                            nrep_rmin, z
      LOGICAL                                            :: explicit, has_rmax, has_rmin
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: rmax_is_set, rmin_is_set
      REAL(KIND=dp)                                      :: auto_rmax_scale, auto_rmin_scale, rmax, &
                                                            rmin
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      nrep_rmin = 0
      nrep_rmax = 0
      nkind = SIZE(atomic_kind_set)

      has_rmin = .FALSE.
      has_rmax = .FALSE.

      CALL section_vals_get(sphere_section, explicit=explicit)
      IF (explicit) THEN
         resp_env%molecular_sys = .TRUE.
         CALL section_vals_val_get(sphere_section, "AUTO_VDW_RADII_TABLE", &
                                   i_val=resp_env%auto_vdw_radii_table)
         CALL section_vals_val_get(sphere_section, "AUTO_RMIN_SCALE", r_val=auto_rmin_scale)
         CALL section_vals_val_get(sphere_section, "AUTO_RMAX_SCALE", r_val=auto_rmax_scale)
         CALL section_vals_val_get(sphere_section, "RMIN", explicit=has_rmin, r_val=rmin)
         CALL section_vals_val_get(sphere_section, "RMAX", explicit=has_rmax, r_val=rmax)
         CALL section_vals_val_get(sphere_section, "RMIN_KIND", n_rep_val=nrep_rmin)
         CALL section_vals_val_get(sphere_section, "RMAX_KIND", n_rep_val=nrep_rmax)
         ALLOCATE (resp_env%rmin_kind(nkind))
         ALLOCATE (resp_env%rmax_kind(nkind))
         resp_env%rmin_kind = 0.0_dp
         resp_env%rmax_kind = 0.0_dp
         ALLOCATE (rmin_is_set(nkind))
         ALLOCATE (rmax_is_set(nkind))
         rmin_is_set = .FALSE.
         rmax_is_set = .FALSE.
         ! define rmin_kind and rmax_kind to predefined vdW radii
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, &
                                 element_symbol=symbol, &
                                 kind_number=kind_number, &
                                 z=z)
            SELECT CASE (resp_env%auto_vdw_radii_table)
            CASE (use_cambridge_vdw_radii)
               CALL get_ptable_info(symbol, vdw_radius=resp_env%rmin_kind(kind_number))
               rmin_is_set(kind_number) = .TRUE.
            CASE (use_uff_vdw_radii)
               CALL cite_reference(Rappe1992)
               CALL get_uff_vdw_radius(z, radius=resp_env%rmin_kind(kind_number), &
                                       found=rmin_is_set(kind_number))
            CASE DEFAULT
               CALL get_ptable_info(symbol, vdw_radius=resp_env%rmin_kind(kind_number))
               rmin_is_set(kind_number) = .TRUE.
            END SELECT
            IF (rmin_is_set(kind_number)) THEN
               resp_env%rmin_kind(kind_number) = cp_unit_to_cp2k(resp_env%rmin_kind(kind_number), &
                                                                 "angstrom")
               resp_env%rmin_kind(kind_number) = resp_env%rmin_kind(kind_number)*auto_rmin_scale
               ! set RMAX_KIND accourding by scaling RMIN_KIND
               resp_env%rmax_kind(kind_number) = &
                  MAX(resp_env%rmin_kind(kind_number), &
                      resp_env%rmin_kind(kind_number)*auto_rmax_scale)
               rmax_is_set(kind_number) = .TRUE.
            END IF
         END DO
         ! if RMIN or RMAX are present, overwrite the rmin_kind(:) and
         ! rmax_kind(:) to those values
         IF (has_rmin) THEN
            resp_env%rmin_kind = rmin
            rmin_is_set = .TRUE.
         END IF
         IF (has_rmax) THEN
            resp_env%rmax_kind = rmax
            rmax_is_set = .TRUE.
         END IF
         ! if RMIN_KIND's or RMAX_KIND's are present, overwrite the
         ! rmin_kinds(:) or rmax_kind(:) to those values
         DO j = 1, nrep_rmin
            CALL section_vals_val_get(sphere_section, "RMIN_KIND", i_rep_val=j, &
                                      c_vals=tmpstringlist)
            DO ikind = 1, nkind
               atomic_kind => atomic_kind_set(ikind)
               CALL get_atomic_kind(atomic_kind, element_symbol=symbol, kind_number=kind_number)
               IF (TRIM(tmpstringlist(2)) == TRIM(symbol)) THEN
                  READ (tmpstringlist(1), *) resp_env%rmin_kind(kind_number)
                  resp_env%rmin_kind(kind_number) = &
                     cp_unit_to_cp2k(resp_env%rmin_kind(kind_number), &
                                     "angstrom")
                  rmin_is_set(kind_number) = .TRUE.
               END IF
            END DO
         END DO
         DO j = 1, nrep_rmax
            CALL section_vals_val_get(sphere_section, "RMAX_KIND", i_rep_val=j, &
                                      c_vals=tmpstringlist)
            DO ikind = 1, nkind
               atomic_kind => atomic_kind_set(ikind)
               CALL get_atomic_kind(atomic_kind, element_symbol=symbol, kind_number=kind_number)
               IF (TRIM(tmpstringlist(2)) == TRIM(symbol)) THEN
                  READ (tmpstringlist(1), *) resp_env%rmax_kind(kind_number)
                  resp_env%rmax_kind(kind_number) = cp_unit_to_cp2k(resp_env%rmax_kind(kind_number), &
                                                                    "angstrom")
                  rmax_is_set(kind_number) = .TRUE.
               END IF
            END DO
         END DO
         ! check if rmin and rmax are set for each kind
         n_rmin_missing = 0
         n_rmax_missing = 0
         missing_rmin = ""
         missing_rmax = ""
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, &
                                 element_symbol=symbol, &
                                 kind_number=kind_number)
            IF (.NOT. rmin_is_set(kind_number)) THEN
               n_rmin_missing = n_rmin_missing+1
               missing_rmin = TRIM(missing_rmin)//" "//TRIM(symbol)//","
            END IF
            IF (.NOT. rmax_is_set(kind_number)) THEN
               n_rmax_missing = n_rmax_missing+1
               missing_rmax = TRIM(missing_rmax)//" "//TRIM(symbol)//","
            END IF
         END DO
         IF (n_rmin_missing > 0) THEN
            CALL cp_warn(cp__l("qs_resp.F",632), &
                         "RMIN for the following elements are missing: "// &
                         TRIM(missing_rmin)// &
                         " please set these values manually using "// &
                         "RMIN_KIND in SPHERE_SAMPLING section")
         END IF
         IF (n_rmax_missing > 0) THEN
            CALL cp_warn(cp__l("qs_resp.F",639), &
                         "RMAX for the following elements are missing: "// &
                         TRIM(missing_rmax)// &
                         " please set these values manually using "// &
                         "RMAX_KIND in SPHERE_SAMPLING section")
         END IF
         IF (n_rmin_missing > 0 .OR. &
             n_rmax_missing > 0) THEN
            CALL cp__b("qs_resp.F",647,"Insufficient data for RMIN or RMAX")
         END IF

         CALL get_cell(cell=cell, h=hmat)
         resp_env%box_hi = (/hmat(1, 1), hmat(2, 2), hmat(3, 3)/)
         resp_env%box_low = 0.0_dp
         CALL section_vals_val_get(sphere_section, "X_HI", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "X_HI", &
                                                 r_val=resp_env%box_hi(1))
         CALL section_vals_val_get(sphere_section, "X_LOW", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "X_LOW", &
                                                 r_val=resp_env%box_low(1))
         CALL section_vals_val_get(sphere_section, "Y_HI", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "Y_HI", &
                                                 r_val=resp_env%box_hi(2))
         CALL section_vals_val_get(sphere_section, "Y_LOW", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "Y_LOW", &
                                                 r_val=resp_env%box_low(2))
         CALL section_vals_val_get(sphere_section, "Z_HI", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "Z_HI", &
                                                 r_val=resp_env%box_hi(3))
         CALL section_vals_val_get(sphere_section, "Z_LOW", explicit=explicit)
         IF (explicit) CALL section_vals_val_get(sphere_section, "Z_LOW", &
                                                 r_val=resp_env%box_low(3))

         DEALLOCATE (rmin_is_set)
         DEALLOCATE (rmax_is_set)
      END IF

   END SUBROUTINE get_parameter_molecular_sys

! **************************************************************************************************
!> \brief building atom lists for different sections of RESP
!> \param section input section
!> \param subsys ...
!> \param atom_list list of atoms for restraints, constraints and fit point
!>        sampling for slab-like systems
!> \param rep input section can be repeated, this param defines for which
!>        repetition of the input section the atom_list is built
! **************************************************************************************************
   SUBROUTINE build_atom_list(section, subsys, atom_list, rep)

      TYPE(section_vals_type), POINTER                   :: section
      TYPE(qs_subsys_type), POINTER                      :: subsys
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      INTEGER, INTENT(IN), OPTIONAL                      :: rep

      CHARACTER(len=*), PARAMETER :: routineN = 'build_atom_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, i, irep, j, &
                                                            max_index, n_var, num_atom
      INTEGER, DIMENSION(:), POINTER                     :: indexes
      LOGICAL                                            :: index_in_range

      CALL timeset(routineN, handle)

      NULLIFY (indexes)
      irep = 1
      IF (PRESENT(rep)) irep = rep

      CALL section_vals_val_get(section, "ATOM_LIST", i_rep_section=irep, &
                                n_rep_val=n_var)
      num_atom = 0
      DO i = 1, n_var
         CALL section_vals_val_get(section, "ATOM_LIST", i_rep_section=irep, &
                                   i_rep_val=i, i_vals=indexes)
         num_atom = num_atom+SIZE(indexes)
      ENDDO
      ALLOCATE (atom_list(num_atom))
      atom_list = 0
      num_atom = 1
      DO i = 1, n_var
         CALL section_vals_val_get(section, "ATOM_LIST", i_rep_section=irep, &
                                   i_rep_val=i, i_vals=indexes)
         atom_list(num_atom:num_atom+SIZE(indexes)-1) = indexes(:)
         num_atom = num_atom+SIZE(indexes)
      ENDDO
      !check atom list
      num_atom = num_atom-1
      CALL qs_subsys_get(subsys, nparticle=max_index)
      IF(.NOT.(SIZE(atom_list) /= 0))CALL cp__a("qs_resp.F",728)
      index_in_range = (MAXVAL(atom_list) <= max_index) &
                       .AND. (MINVAL(atom_list) > 0)
      IF(.NOT.(index_in_range))CALL cp__a("qs_resp.F",731)
      DO i = 1, num_atom
         DO j = i+1, num_atom
            atom_a = atom_list(i)
            atom_b = atom_list(j)
            IF (atom_a == atom_b) &
               CALL cp__b("qs_resp.F",737,"There are atoms doubled in atom list for RESP.")
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE build_atom_list

! **************************************************************************************************
!> \brief build matrix and vector for nonperiodic RESP fitting
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param atomic_kind_set ...
!> \param particles ...
!> \param cell parameters related to the simulation cell
!> \param matrix coefficient matrix of the linear system of equations
!> \param rhs vector of the linear system of equations
!> \param natom number of atoms
! **************************************************************************************************
   SUBROUTINE calc_resp_matrix_nonper(qs_env, resp_env, atomic_kind_set, particles, &
                                      cell, matrix, rhs, natom)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: matrix
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rhs
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_resp_matrix_nonper', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2, 3), gbo(2, 3), handle, i, ikind, &
                                                            jx, jy, jz, k, kind_number, l, m, &
                                                            nkind, now, np(3), p
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: not_in_range
      REAL(KIND=dp)                                      :: delta, dh(3, 3), dvol, r(3), rmax, rmin, &
                                                            vec(3), vec_pbc(3), vj
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dist
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat, hmat_inv
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_type), POINTER                             :: v_hartree_pw

      CALL timeset(routineN, handle)

      NULLIFY (particle_set, v_hartree_pw)
      delta = 1.0E-13_dp

      CALL get_cell(cell=cell, h=hmat, h_inv=hmat_inv)

      IF (.NOT. cell%orthorhombic) THEN
         CALL cp_abort(cp__l("qs_resp.F",790), &
                       "Nonperiodic solution for RESP charges only"// &
                       " implemented for orthorhombic cells!")
      END IF
      IF (.NOT. resp_env%molecular_sys) THEN
         CALL cp_abort(cp__l("qs_resp.F",795), &
                       "Nonperiodic solution for RESP charges (i.e. nonperiodic"// &
                       " Poisson solver) can only be used with section SPHERE_SAMPLING")
      ENDIF
      IF (resp_env%use_repeat_method) THEN
         CALL cp_abort(cp__l("qs_resp.F",800), &
                       "REPEAT method only reasonable for periodic RESP fitting")
      ENDIF
      CALL get_qs_env(qs_env, particle_set=particle_set, v_hartree_rspace=v_hartree_pw)

      bo = v_hartree_pw%pw_grid%bounds_local
      gbo = v_hartree_pw%pw_grid%bounds
      np = v_hartree_pw%pw_grid%npts
      dh = v_hartree_pw%pw_grid%dh
      dvol = v_hartree_pw%pw_grid%dvol
      nkind = SIZE(atomic_kind_set)

      ALLOCATE (dist(natom))
      ALLOCATE (not_in_range(natom, 2))

      ! store fitting points to calculate the RMS and RRMS later
      IF (.NOT. ASSOCIATED(resp_env%fitpoints)) THEN
         now = 1000
         ALLOCATE (resp_env%fitpoints(3, now))
      ELSE
         now = SIZE(resp_env%fitpoints, 2)
      END IF

      DO jz = bo(1, 3), bo(2, 3)
      DO jy = bo(1, 2), bo(2, 2)
      DO jx = bo(1, 1), bo(2, 1)
         IF (.NOT. (MODULO(jz, resp_env%stride(3)) == 0)) CYCLE
         IF (.NOT. (MODULO(jy, resp_env%stride(2)) == 0)) CYCLE
         IF (.NOT. (MODULO(jx, resp_env%stride(1)) == 0)) CYCLE
         !bounds bo reach from -np/2 to np/2. shift of np/2 so that r(1,1,1)=(0,0,0)
         l = jx-gbo(1, 1)
         k = jy-gbo(1, 2)
         p = jz-gbo(1, 3)
         r(3) = p*dh(3, 3)+k*dh(3, 2)+l*dh(3, 1)
         r(2) = p*dh(2, 3)+k*dh(2, 2)+l*dh(2, 1)
         r(1) = p*dh(1, 3)+k*dh(1, 2)+l*dh(1, 1)
         IF (r(3) < resp_env%box_low(3) .OR. r(3) > resp_env%box_hi(3)) CYCLE
         IF (r(2) < resp_env%box_low(2) .OR. r(2) > resp_env%box_hi(2)) CYCLE
         IF (r(1) < resp_env%box_low(1) .OR. r(1) > resp_env%box_hi(1)) CYCLE
         ! compute distance from the grid point to all atoms
         not_in_range = .FALSE.
         DO i = 1, natom
            vec = r-particles%els(i)%r
            vec_pbc(1) = vec(1)-hmat(1, 1)*ANINT(hmat_inv(1, 1)*vec(1))
            vec_pbc(2) = vec(2)-hmat(2, 2)*ANINT(hmat_inv(2, 2)*vec(2))
            vec_pbc(3) = vec(3)-hmat(3, 3)*ANINT(hmat_inv(3, 3)*vec(3))
            dist(i) = SQRT(SUM(vec_pbc**2))
            CALL get_atomic_kind(atomic_kind=particle_set(i)%atomic_kind, &
                                 kind_number=kind_number)
            DO ikind = 1, nkind
               IF (ikind == kind_number) THEN
                  rmin = resp_env%rmin_kind(ikind)
                  rmax = resp_env%rmax_kind(ikind)
                  EXIT
               ENDIF
            ENDDO
            IF (dist(i) < rmin+delta) not_in_range(i, 1) = .TRUE.
            IF (dist(i) > rmax-delta) not_in_range(i, 2) = .TRUE.
         ENDDO
         ! check if the point is sufficiently close and  far. if OK, we can use
         ! the point for fitting, add/subtract 1.0E-13 to get rid of rounding errors when shifting atoms
         IF (ANY(not_in_range(:, 1)) .OR. ALL(not_in_range(:, 2))) CYCLE
         resp_env%npoints_proc = resp_env%npoints_proc+1
         IF (resp_env%npoints_proc > now) THEN
            now = 2*now
            CALL reallocate(resp_env%fitpoints, 1, 3, 1, now)
         ENDIF
         resp_env%fitpoints(1, resp_env%npoints_proc) = jx
         resp_env%fitpoints(2, resp_env%npoints_proc) = jy
         resp_env%fitpoints(3, resp_env%npoints_proc) = jz
         ! correct for the fact that v_hartree is scaled by dvol, and has the opposite sign
         IF (qs_env%qmmm) THEN
            ! If it's a QM/MM run let's remove the contribution of the MM potential out of the Hartree pot
            vj = -v_hartree_pw%cr3d(jx, jy, jz)/dvol+qs_env%ks_qmmm_env%v_qmmm_rspace%pw%cr3d(jx, jy, jz)
         ELSE
            vj = -v_hartree_pw%cr3d(jx, jy, jz)/dvol
         END IF
         dist(:) = 1.0_dp/dist(:)

         DO i = 1, natom
            DO m = 1, natom
               matrix(m, i) = matrix(m, i)+2.0_dp*dist(i)*dist(m)
            ENDDO
            rhs(i) = rhs(i)+2.0_dp*vj*dist(i)
         ENDDO
      ENDDO
      ENDDO
      ENDDO

      resp_env%npoints = resp_env%npoints_proc
      CALL mp_sum(resp_env%npoints, v_hartree_pw%pw_grid%para%group)
      CALL mp_sum(matrix, v_hartree_pw%pw_grid%para%group)
      CALL mp_sum(rhs, v_hartree_pw%pw_grid%para%group)
      !weighted sum
      matrix = matrix/resp_env%npoints
      rhs = rhs/resp_env%npoints

      DEALLOCATE (dist)
      DEALLOCATE (not_in_range)

      CALL timestop(handle)

   END SUBROUTINE calc_resp_matrix_nonper

! **************************************************************************************************
!> \brief build matrix and vector for periodic RESP fitting
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
!> \param particles ...
!> \param cell parameters related to the simulation cell
!> \param natom number of atoms
! **************************************************************************************************
   SUBROUTINE calc_resp_matrix_periodic(qs_env, resp_env, rep_sys, particles, cell, &
                                        natom)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_resp_matrix_periodic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ip, j, jx, jy, jz
      INTEGER, DIMENSION(3)                              :: periodic
      REAL(KIND=dp)                                      :: normalize_factor
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vpot
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho_ga, va_gspace, va_rspace
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, para_env, auxbas_pw_pool, poisson_env)

      CALL get_cell(cell=cell, periodic=periodic)

      IF (.NOT. ALL(periodic /= 0)) THEN
         CALL cp_abort(cp__l("qs_resp.F",943), &
                       "Periodic solution for RESP (with periodic Poisson solver)"// &
                       " can only be obtained with a cell that has XYZ periodicity")
      ENDIF

      CALL get_qs_env(qs_env, pw_env=pw_env, para_env=para_env)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             rho_ga%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             va_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             va_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      !get fitting points and store them in resp_env%fitpoints
      CALL get_fitting_points(qs_env, resp_env, rep_sys, particles=particles, &
                              cell=cell)
      ALLOCATE (vpot(resp_env%npoints_proc, natom))
      normalize_factor = SQRT((resp_env%eta/pi)**3)

      DO i = 1, natom
         !collocate gaussian for each atom
         CALL pw_zero(rho_ga%pw)
         CALL calculate_rho_resp_single(rho_ga, qs_env, resp_env%eta, i)
         !calculate potential va and store the part needed for fitting in vpot
         CALL pw_zero(va_gspace%pw)
         CALL pw_poisson_solve(poisson_env, rho_ga%pw, vhartree=va_gspace%pw)
         CALL pw_zero(va_rspace%pw)
         CALL pw_transfer(va_gspace%pw, va_rspace%pw)
         CALL pw_scale(va_rspace%pw, normalize_factor)
         DO ip = 1, resp_env%npoints_proc
            jx = resp_env%fitpoints(1, ip)
            jy = resp_env%fitpoints(2, ip)
            jz = resp_env%fitpoints(3, ip)
            vpot(ip, i) = va_rspace%pw%cr3d(jx, jy, jz)
         END DO
      ENDDO

      CALL pw_release(va_gspace%pw)
      CALL pw_release(va_rspace%pw)
      CALL pw_release(rho_ga%pw)

      DO i = 1, natom
         DO j = 1, natom
            ! calculate matrix
            resp_env%matrix(i, j) = resp_env%matrix(i, j)+2.0_dp*SUM(vpot(:, i)*vpot(:, j))
         ENDDO
         ! calculate vector resp_env%rhs
         CALL calculate_rhs(qs_env, resp_env, resp_env%rhs(i), vpot(:, i))
      ENDDO

      CALL mp_sum(resp_env%matrix, para_env%group)
      CALL mp_sum(resp_env%rhs, para_env%group)
      !weighted sum
      resp_env%matrix = resp_env%matrix/resp_env%npoints
      resp_env%rhs = resp_env%rhs/resp_env%npoints

      ! REPEAT stuff
      IF (resp_env%use_repeat_method) THEN
         ! sum over selected points of single Gaussian potential vpot
         DO i = 1, natom
            resp_env%sum_vpot(i) = 2.0_dp*accurate_sum(vpot(:, i))/resp_env%npoints
         ENDDO
         CALL mp_sum(resp_env%sum_vpot, para_env%group)
         CALL mp_sum(resp_env%sum_vhartree, para_env%group)
         resp_env%sum_vhartree = 2.0_dp*resp_env%sum_vhartree/resp_env%npoints
      ENDIF

      DEALLOCATE (vpot)
      CALL timestop(handle)

   END SUBROUTINE calc_resp_matrix_periodic

! **************************************************************************************************
!> \brief get RESP fitting points for the periodic fitting
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
!> \param particles ...
!> \param cell parameters related to the simulation cell
! **************************************************************************************************
   SUBROUTINE get_fitting_points(qs_env, resp_env, rep_sys, particles, cell)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'get_fitting_points', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2, 3), gbo(2, 3), handle, i, iatom, &
                                                            ikind, in_x, in_y, in_z, jx, jy, jz, &
                                                            k, kind_number, l, m, natom, nkind, &
                                                            now, p
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: not_in_range
      REAL(KIND=dp)                                      :: delta, dh(3, 3), r(3), rmax, rmin, &
                                                            vec_pbc(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dist
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_type), POINTER                             :: v_hartree_pw

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, v_hartree_pw, para_env, particle_set)
      delta = 1.0E-13_dp

      CALL get_qs_env(qs_env, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      para_env=para_env, &
                      v_hartree_rspace=v_hartree_pw)

      bo = v_hartree_pw%pw_grid%bounds_local
      gbo = v_hartree_pw%pw_grid%bounds
      dh = v_hartree_pw%pw_grid%dh
      natom = SIZE(particles%els)
      nkind = SIZE(atomic_kind_set)

      IF (.NOT. ASSOCIATED(resp_env%fitpoints)) THEN
         now = 1000
         ALLOCATE (resp_env%fitpoints(3, now))
      ELSE
         now = SIZE(resp_env%fitpoints, 2)
      END IF

      ALLOCATE (dist(natom))
      ALLOCATE (not_in_range(natom, 2))

      !every proc gets another bo, grid is distributed
      DO jz = bo(1, 3), bo(2, 3)
         IF (.NOT. (MODULO(jz, resp_env%stride(3)) == 0)) CYCLE
         DO jy = bo(1, 2), bo(2, 2)
            IF (.NOT. (MODULO(jy, resp_env%stride(2)) == 0)) CYCLE
            DO jx = bo(1, 1), bo(2, 1)
               IF (.NOT. (MODULO(jx, resp_env%stride(1)) == 0)) CYCLE
               !bounds gbo reach from -np/2 to np/2. shift of np/2 so that r(1,1,1)=(0,0,0)
               l = jx-gbo(1, 1)
               k = jy-gbo(1, 2)
               p = jz-gbo(1, 3)
               r(3) = p*dh(3, 3)+k*dh(3, 2)+l*dh(3, 1)
               r(2) = p*dh(2, 3)+k*dh(2, 2)+l*dh(2, 1)
               r(1) = p*dh(1, 3)+k*dh(1, 2)+l*dh(1, 1)
               IF (resp_env%molecular_sys) THEN
                  not_in_range = .FALSE.
                  DO m = 1, natom
                     vec_pbc = pbc(r, particles%els(m)%r, cell)
                     dist(m) = SQRT(SUM(vec_pbc**2))
                     CALL get_atomic_kind(atomic_kind=particle_set(m)%atomic_kind, &
                                          kind_number=kind_number)
                     DO ikind = 1, nkind
                        IF (ikind == kind_number) THEN
                           rmin = resp_env%rmin_kind(ikind)
                           rmax = resp_env%rmax_kind(ikind)
                           EXIT
                        ENDIF
                     ENDDO
                     IF (dist(m) < rmin+delta) not_in_range(m, 1) = .TRUE.
                     IF (dist(m) > rmax-delta) not_in_range(m, 2) = .TRUE.
                  ENDDO
                  IF (ANY(not_in_range(:, 1)) .OR. ALL(not_in_range(:, 2))) CYCLE
               ELSE
                  DO i = 1, SIZE(rep_sys)
                     DO m = 1, SIZE(rep_sys(i)%p_resp%atom_surf_list)
                        in_z = 0
                        in_y = 0
                        in_x = 0
                        iatom = rep_sys(i)%p_resp%atom_surf_list(m)
                        SELECT CASE (rep_sys (i)%p_resp%my_fit)
                        CASE (do_resp_x_dir, do_resp_y_dir, do_resp_z_dir)
                           vec_pbc = pbc(particles%els(iatom)%r, r, cell)
                        CASE (do_resp_minus_x_dir, do_resp_minus_y_dir, do_resp_minus_z_dir)
                           vec_pbc = pbc(r, particles%els(iatom)%r, cell)
                        END SELECT
                        SELECT CASE (rep_sys (i)%p_resp%my_fit)
                           !subtract delta=1.0E-13 to get rid of rounding errors when shifting atoms
                        CASE (do_resp_x_dir, do_resp_minus_x_dir)
                           IF (ABS(vec_pbc(3)) < rep_sys(i)%p_resp%length-delta) in_z = 1
                           IF (ABS(vec_pbc(2)) < rep_sys(i)%p_resp%length-delta) in_y = 1
                           IF (vec_pbc(1) > rep_sys(i)%p_resp%range_surf(1)+delta .AND. &
                               vec_pbc(1) < rep_sys(i)%p_resp%range_surf(2)-delta) in_x = 1
                        CASE (do_resp_y_dir, do_resp_minus_y_dir)
                           IF (ABS(vec_pbc(3)) < rep_sys(i)%p_resp%length-delta) in_z = 1
                           IF (vec_pbc(2) > rep_sys(i)%p_resp%range_surf(1)+delta .AND. &
                               vec_pbc(2) < rep_sys(i)%p_resp%range_surf(2)-delta) in_y = 1
                           IF (ABS(vec_pbc(1)) < rep_sys(i)%p_resp%length-delta) in_x = 1
                        CASE (do_resp_z_dir, do_resp_minus_z_dir)
                           IF (vec_pbc(3) > rep_sys(i)%p_resp%range_surf(1)+delta .AND. &
                               vec_pbc(3) < rep_sys(i)%p_resp%range_surf(2)-delta) in_z = 1
                           IF (ABS(vec_pbc(2)) < rep_sys(i)%p_resp%length-delta) in_y = 1
                           IF (ABS(vec_pbc(1)) < rep_sys(i)%p_resp%length-delta) in_x = 1
                        END SELECT
                        IF (in_z*in_y*in_x == 1) EXIT
                     ENDDO
                     IF (in_z*in_y*in_x == 1) EXIT
                  ENDDO
                  IF (in_z*in_y*in_x == 0) CYCLE
               ENDIF
               resp_env%npoints_proc = resp_env%npoints_proc+1
               IF (resp_env%npoints_proc > now) THEN
                  now = 2*now
                  CALL reallocate(resp_env%fitpoints, 1, 3, 1, now)
               ENDIF
               resp_env%fitpoints(1, resp_env%npoints_proc) = jx
               resp_env%fitpoints(2, resp_env%npoints_proc) = jy
               resp_env%fitpoints(3, resp_env%npoints_proc) = jz
            ENDDO
         ENDDO
      ENDDO

      resp_env%npoints = resp_env%npoints_proc
      CALL mp_sum(resp_env%npoints, para_env%group)

      DEALLOCATE (dist)
      DEALLOCATE (not_in_range)

      CALL timestop(handle)

   END SUBROUTINE get_fitting_points

! **************************************************************************************************
!> \brief calculate vector rhs
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param rhs vector
!> \param vpot single gaussian potential
! **************************************************************************************************
   SUBROUTINE calculate_rhs(qs_env, resp_env, rhs, vpot)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      REAL(KIND=dp), INTENT(INOUT)                       :: rhs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vpot

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rhs', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ip, jx, jy, jz
      REAL(KIND=dp)                                      :: dvol
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vhartree
      TYPE(pw_type), POINTER                             :: v_hartree_pw

      CALL timeset(routineN, handle)

      NULLIFY (v_hartree_pw)
      CALL get_qs_env(qs_env, v_hartree_rspace=v_hartree_pw)
      dvol = v_hartree_pw%pw_grid%dvol
      ALLOCATE (vhartree(resp_env%npoints_proc))
      vhartree = 0.0_dp

      !multiply v_hartree and va_rspace and calculate the vector rhs
      !taking into account that v_hartree has opposite site; remove v_qmmm
      DO ip = 1, resp_env%npoints_proc
         jx = resp_env%fitpoints(1, ip)
         jy = resp_env%fitpoints(2, ip)
         jz = resp_env%fitpoints(3, ip)
         vhartree(ip) = -v_hartree_pw%cr3d(jx, jy, jz)/dvol
         IF (qs_env%qmmm) THEN
            !taking into account that v_qmmm has also opposite sign
            vhartree(ip) = vhartree(ip)+qs_env%ks_qmmm_env%v_qmmm_rspace%pw%cr3d(jx, jy, jz)
         ENDIF
         rhs = rhs+2.0_dp*vhartree(ip)*vpot(ip)
      ENDDO

      IF (resp_env%use_repeat_method) THEN
         resp_env%sum_vhartree = accurate_sum(vhartree)
      ENDIF

      DEALLOCATE (vhartree)

      CALL timestop(handle)

   END SUBROUTINE calculate_rhs

! **************************************************************************************************
!> \brief print the atom coordinates and the coordinates of the fitting points
!>        to an xyz file
!> \param qs_env the qs environment
!> \param resp_env the resp environment
! **************************************************************************************************
   SUBROUTINE print_fitting_points(qs_env, resp_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env

      CHARACTER(len=*), PARAMETER :: routineN = 'print_fitting_points', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: gbo(2, 3), handle, i, iatom, ip, jx, jy, &
                                                            jz, k, l, my_pos, nobjects, &
                                                            output_unit, p, req(6)
      INTEGER, DIMENSION(:), POINTER                     :: tmp_npoints, tmp_size
      INTEGER, DIMENSION(:, :), POINTER                  :: tmp_points
      REAL(KIND=dp)                                      :: conv, dh(3, 3), r(3)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_type), POINTER                             :: v_hartree_pw
      TYPE(section_vals_type), POINTER                   :: input, print_key, resp_section

      CALL timeset(routineN, handle)

      NULLIFY (para_env, input, logger, resp_section, print_key, particle_set, tmp_size, &
               tmp_points, tmp_npoints, v_hartree_pw)

      CALL get_qs_env(qs_env, input=input, para_env=para_env, &
                      particle_set=particle_set, v_hartree_rspace=v_hartree_pw)
      conv = cp_unit_from_cp2k(1.0_dp, "angstrom")
      gbo = v_hartree_pw%pw_grid%bounds
      dh = v_hartree_pw%pw_grid%dh
      nobjects = SIZE(particle_set)+resp_env%npoints

      resp_section => section_vals_get_subs_vals(input, "PROPERTIES%RESP")
      print_key => section_vals_get_subs_vals(resp_section, "PRINT%COORD_FIT_POINTS")
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, resp_section, &
                                         "PRINT%COORD_FIT_POINTS", &
                                         extension=".xyz", &
                                         file_status="REPLACE", &
                                         file_action="WRITE", &
                                         file_form="FORMATTED")

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           resp_section, "PRINT%COORD_FIT_POINTS"), &
                cp_p_file)) THEN
         IF (output_unit > 0) THEN
            filename = cp_print_key_generate_filename(logger, &
                                                      print_key, extension=".xyz", &
                                                      my_local=.FALSE.)
            WRITE (unit=output_unit, FMT="(I12,/)") nobjects
            DO iatom = 1, SIZE(particle_set)
               CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                    element_symbol=element_symbol)
               WRITE (UNIT=output_unit, FMT="(A,1X,3F10.5)") element_symbol, &
                  particle_set(iatom)%r(1:3)*conv
            ENDDO
            !printing points of proc which is doing the output (should be proc 0)
            DO ip = 1, resp_env%npoints_proc
               jx = resp_env%fitpoints(1, ip)
               jy = resp_env%fitpoints(2, ip)
               jz = resp_env%fitpoints(3, ip)
               l = jx-gbo(1, 1)
               k = jy-gbo(1, 2)
               p = jz-gbo(1, 3)
               r(3) = p*dh(3, 3)+k*dh(3, 2)+l*dh(3, 1)
               r(2) = p*dh(2, 3)+k*dh(2, 2)+l*dh(2, 1)
               r(1) = p*dh(1, 3)+k*dh(1, 2)+l*dh(1, 1)
               r(:) = r(:)*conv
               WRITE (UNIT=output_unit, FMT="(A,2X,3F10.5)") "X", r(1), r(2), r(3)
            ENDDO
         ENDIF

         ALLOCATE (tmp_size(1))
         ALLOCATE (tmp_npoints(1))

         !sending data of all other procs to proc which makes the output (proc 0)
         IF (output_unit > 0) THEN
            my_pos = para_env%mepos
            DO i = 1, para_env%num_pe
               IF (my_pos == i-1) CYCLE
               CALL mp_irecv(msgout=tmp_size, source=i-1, comm=para_env%group, &
                             request=req(1))
               CALL mp_wait(req(1))
               ALLOCATE (tmp_points(3, tmp_size(1)))
               CALL mp_irecv(msgout=tmp_points, source=i-1, comm=para_env%group, &
                             request=req(3))
               CALL mp_wait(req(3))
               CALL mp_irecv(msgout=tmp_npoints, source=i-1, comm=para_env%group, &
                             request=req(5))
               CALL mp_wait(req(5))
               DO ip = 1, tmp_npoints(1)
                  jx = tmp_points(1, ip)
                  jy = tmp_points(2, ip)
                  jz = tmp_points(3, ip)
                  l = jx-gbo(1, 1)
                  k = jy-gbo(1, 2)
                  p = jz-gbo(1, 3)
                  r(3) = p*dh(3, 3)+k*dh(3, 2)+l*dh(3, 1)
                  r(2) = p*dh(2, 3)+k*dh(2, 2)+l*dh(2, 1)
                  r(1) = p*dh(1, 3)+k*dh(1, 2)+l*dh(1, 1)
                  r(:) = r(:)*conv
                  WRITE (UNIT=output_unit, FMT="(A,2X,3F10.5)") "X", r(1), r(2), r(3)
               ENDDO
               DEALLOCATE (tmp_points)
            ENDDO
         ELSE
            tmp_size(1) = SIZE(resp_env%fitpoints, 2)
            !para_env%source should be 0
            CALL mp_isend(msgin=tmp_size, dest=para_env%source, comm=para_env%group, &
                          request=req(2))
            CALL mp_wait(req(2))
            CALL mp_isend(msgin=resp_env%fitpoints, dest=para_env%source, comm=para_env%group, &
                          request=req(4))
            CALL mp_wait(req(4))
            tmp_npoints(1) = resp_env%npoints_proc
            CALL mp_isend(msgin=tmp_npoints, dest=para_env%source, comm=para_env%group, &
                          request=req(6))
            CALL mp_wait(req(6))
         ENDIF

         DEALLOCATE (tmp_size)
         DEALLOCATE (tmp_npoints)
      ENDIF

      CALL cp_print_key_finished_output(output_unit, logger, resp_section, &
                                        "PRINT%COORD_FIT_POINTS")

      CALL timestop(handle)

   END SUBROUTINE print_fitting_points

! **************************************************************************************************
!> \brief add restraints and constraints
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param rest_section input section for restraints
!> \param subsys ...
!> \param natom number of atoms
!> \param cons_section input section for constraints
!> \param particle_set ...
! **************************************************************************************************
   SUBROUTINE add_restraints_and_constraints(qs_env, resp_env, rest_section, &
                                             subsys, natom, cons_section, particle_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(section_vals_type), POINTER                   :: rest_section
      TYPE(qs_subsys_type), POINTER                      :: subsys
      INTEGER, INTENT(IN)                                :: natom
      TYPE(section_vals_type), POINTER                   :: cons_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'add_restraints_and_constraints', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, k, m, ncons_v, z
      INTEGER, DIMENSION(:), POINTER                     :: atom_list_cons, atom_list_res
      LOGICAL                                            :: explicit_coeff
      REAL(KIND=dp)                                      :: my_atom_coef(2), strength, TARGET
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atom_coef
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      NULLIFY (atom_coef, atom_list_res, atom_list_cons, dft_control)

      CALL get_qs_env(qs_env, dft_control=dft_control)

      !*** add the restraints
      DO i = 1, resp_env%nrest_sec
         CALL section_vals_val_get(rest_section, "TARGET", i_rep_section=i, r_val=TARGET)
         CALL section_vals_val_get(rest_section, "STRENGTH", i_rep_section=i, r_val=strength)
         CALL build_atom_list(rest_section, subsys, atom_list_res, i)
         CALL section_vals_val_get(rest_section, "ATOM_COEF", i_rep_section=i, explicit=explicit_coeff)
         IF (explicit_coeff) THEN
            CALL section_vals_val_get(rest_section, "ATOM_COEF", i_rep_section=i, r_vals=atom_coef)
            IF(.NOT.(SIZE(atom_list_res) == SIZE(atom_coef)))CALL cp__a("qs_resp.F",1411)
         ENDIF
         DO m = 1, SIZE(atom_list_res)
            IF (explicit_coeff) THEN
               DO k = 1, SIZE(atom_list_res)
                  resp_env%matrix(atom_list_res(m), atom_list_res(k)) = &
                     resp_env%matrix(atom_list_res(m), atom_list_res(k))+ &
                     atom_coef(m)*atom_coef(k)*2.0_dp*strength
               ENDDO
               resp_env%rhs(atom_list_res(m)) = resp_env%rhs(atom_list_res(m))+ &
                                                2.0_dp*TARGET*strength*atom_coef(m)
            ELSE
               resp_env%matrix(atom_list_res(m), atom_list_res(m)) = &
                  resp_env%matrix(atom_list_res(m), atom_list_res(m))+ &
                  2.0_dp*strength
               resp_env%rhs(atom_list_res(m)) = resp_env%rhs(atom_list_res(m))+ &
                                                2.0_dp*TARGET*strength
            ENDIF
         ENDDO
         DEALLOCATE (atom_list_res)
      ENDDO

      ! if heavies are restrained to zero, add these as well
      IF (resp_env%rheavies) THEN
         DO i = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(i)%atomic_kind, z=z)
            IF (z .NE. 1) THEN
               resp_env%matrix(i, i) = resp_env%matrix(i, i)+2.0_dp*resp_env%rheavies_strength
            ENDIF
         ENDDO
      ENDIF

      !*** add the constraints
      ncons_v = 0
      ncons_v = ncons_v+natom

      ! REPEAT charges: treat the offset like a constraint
      IF (resp_env%use_repeat_method) THEN
         ncons_v = ncons_v+1
         resp_env%matrix(1:natom, ncons_v) = resp_env%sum_vpot(1:natom)
         resp_env%matrix(ncons_v, 1:natom) = resp_env%sum_vpot(1:natom)
         resp_env%matrix(ncons_v, ncons_v) = 2.0_dp
         resp_env%rhs(ncons_v) = resp_env%sum_vhartree
      ENDIF

      ! total charge constraint
      IF (resp_env%itc) THEN
         ncons_v = ncons_v+1
         resp_env%matrix(1:natom, ncons_v) = 1.0_dp
         resp_env%matrix(ncons_v, 1:natom) = 1.0_dp
         resp_env%rhs(ncons_v) = dft_control%charge
      ENDIF

      ! explict constraints
      DO i = 1, resp_env%ncons_sec
         CALL build_atom_list(cons_section, subsys, atom_list_cons, i)
         IF (.NOT. resp_env%equal_charges) THEN
            ncons_v = ncons_v+1
            CALL section_vals_val_get(cons_section, "ATOM_COEF", i_rep_section=i, r_vals=atom_coef)
            CALL section_vals_val_get(cons_section, "TARGET", i_rep_section=i, r_val=TARGET)
            IF(.NOT.(SIZE(atom_list_cons) == SIZE(atom_coef)))CALL cp__a("qs_resp.F",1471)
            DO m = 1, SIZE(atom_list_cons)
               resp_env%matrix(atom_list_cons(m), ncons_v) = atom_coef(m)
               resp_env%matrix(ncons_v, atom_list_cons(m)) = atom_coef(m)
            ENDDO
            resp_env%rhs(ncons_v) = TARGET
         ELSE
            my_atom_coef(1) = 1.0_dp
            my_atom_coef(2) = -1.0_dp
            DO k = 2, SIZE(atom_list_cons)
               ncons_v = ncons_v+1
               resp_env%matrix(atom_list_cons(1), ncons_v) = my_atom_coef(1)
               resp_env%matrix(ncons_v, atom_list_cons(1)) = my_atom_coef(1)
               resp_env%matrix(atom_list_cons(k), ncons_v) = my_atom_coef(2)
               resp_env%matrix(ncons_v, atom_list_cons(k)) = my_atom_coef(2)
               resp_env%rhs(ncons_v) = 0.0_dp
            ENDDO
         ENDIF
         DEALLOCATE (atom_list_cons)
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE add_restraints_and_constraints

! **************************************************************************************************
!> \brief print input information
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param rep_sys structure for repeating input sections defining fit points
!> \param my_per ...
! **************************************************************************************************
   SUBROUTINE print_resp_parameter_info(qs_env, resp_env, rep_sys, my_per)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(resp_p_type), DIMENSION(:), POINTER           :: rep_sys
      INTEGER, INTENT(IN)                                :: my_per

      CHARACTER(len=*), PARAMETER :: routineN = 'print_resp_parameter_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: symbol
      INTEGER                                            :: handle, i, ikind, kind_number, nkinds, &
                                                            output_unit
      REAL(KIND=dp)                                      :: conv, eta_conv
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: input, resp_section

      CALL timeset(routineN, handle)
      NULLIFY (logger, input, resp_section)

      CALL get_qs_env(qs_env, &
                      input=input, &
                      atomic_kind_set=atomic_kind_set)
      resp_section => section_vals_get_subs_vals(input, "PROPERTIES%RESP")
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, resp_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".resp")
      nkinds = SIZE(atomic_kind_set)

      conv = cp_unit_from_cp2k(1.0_dp, "angstrom")
      IF (.NOT. my_per == use_perd_none) THEN
         eta_conv = cp_unit_from_cp2k(resp_env%eta, "angstrom", power=-2)
      ENDIF

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(/,1X,A,/)') "STARTING RESP FIT"
         IF (resp_env%use_repeat_method) THEN
            WRITE (output_unit, '(T3,A)') &
               "Fit the variance of the potential (REPEAT method)."
         ENDIF
         IF (.NOT. resp_env%equal_charges) THEN
            WRITE (output_unit, '(T3,A,T75,I6)') "Number of explicit constraints: ", resp_env%ncons_sec
         ELSE
            IF (resp_env%itc) THEN
               WRITE (output_unit, '(T3,A,T75,I6)') "Number of explicit constraints: ", resp_env%ncons-1
            ELSE
               WRITE (output_unit, '(T3,A,T75,I6)') "Number of explicit constraints: ", resp_env%ncons
            ENDIF
         ENDIF
         WRITE (output_unit, '(T3,A,T75,I6)') "Number of explicit restraints: ", resp_env%nrest_sec
         WRITE (output_unit, '(T3,A,T80,A)') "Constrain total charge ", MERGE("T", "F", resp_env%itc)
         WRITE (output_unit, '(T3,A,T80,A)') "Restrain heavy atoms ", MERGE("T", "F", resp_env%rheavies)
         IF (resp_env%rheavies) THEN
            WRITE (output_unit, '(T3,A,T71,F10.6)') "Heavy atom restraint strength: ", &
               resp_env%rheavies_strength
         ENDIF
         WRITE (output_unit, '(T3,A,T66,3I5)') "Stride: ", resp_env%stride
         IF (resp_env%molecular_sys) THEN
            WRITE (output_unit, '(T3,A)') &
               "------------------------------------------------------------------------------"
            WRITE (output_unit, '(T3,A)') "Using sphere sampling"
            WRITE (output_unit, '(T3,A,T46,A,T66,A)') &
               "Element", "RMIN [angstrom]", "RMAX [angstrom]"
            DO ikind = 1, nkinds
               CALL get_atomic_kind(atomic_kind=atomic_kind_set(ikind), &
                                    kind_number=kind_number, &
                                    element_symbol=symbol)
               WRITE (output_unit, '(T3,A,T51,F10.5,T71,F10.5)') &
                  symbol, &
                  resp_env%rmin_kind(kind_number)*conv, &
                  resp_env%rmax_kind(kind_number)*conv
            END DO
            IF (my_per == use_perd_none) THEN
               WRITE (output_unit, '(T3,A,T51,3F10.5)') "Box min [angstrom]: ", resp_env%box_low(1:3)*conv
               WRITE (output_unit, '(T3,A,T51,3F10.5)') "Box max [angstrom]: ", resp_env%box_hi(1:3)*conv
            END IF
            WRITE (output_unit, '(T3,A)') &
               "------------------------------------------------------------------------------"
         ELSE
            WRITE (output_unit, '(T3,A)') &
               "------------------------------------------------------------------------------"
            WRITE (output_unit, '(T3,A)') "Using slab sampling"
            WRITE (output_unit, '(2X,A,F10.5)') "Index of atoms defining the surface: "
            DO i = 1, SIZE(rep_sys)
               IF (i > 1 .AND. ALL(rep_sys(i)%p_resp%atom_surf_list == rep_sys(1)%p_resp%atom_surf_list)) EXIT
               WRITE (output_unit, '(7X,10I6)') rep_sys(i)%p_resp%atom_surf_list
            ENDDO
            DO i = 1, SIZE(rep_sys)
               IF (i > 1 .AND. ALL(rep_sys(i)%p_resp%range_surf == rep_sys(1)%p_resp%range_surf)) EXIT
               WRITE (output_unit, '(T3,A,T61,2F10.5)') &
                  "Range for sampling above the surface [angstrom]:", &
                  rep_sys(i)%p_resp%range_surf(1:2)*conv
            ENDDO
            DO i = 1, SIZE(rep_sys)
               IF (i > 1 .AND. rep_sys(i)%p_resp%length == rep_sys(1)%p_resp%length) EXIT
               WRITE (output_unit, '(T3,A,T71,F10.5)') "Length of sampling box above each"// &
                  " surface atom [angstrom]: ", rep_sys(i)%p_resp%length*conv
            ENDDO
            WRITE (output_unit, '(T3,A)') &
               "------------------------------------------------------------------------------"
         ENDIF
         IF (.NOT. my_per == use_perd_none) THEN
            WRITE (output_unit, '(T3,A,T71,F10.5)') "Width of Gaussian charge"// &
               " distribution [angstrom^-2]: ", eta_conv
         ENDIF
         CALL m_flush(output_unit)
      ENDIF
      CALL cp_print_key_finished_output(output_unit, logger, resp_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE print_resp_parameter_info

! **************************************************************************************************
!> \brief print RESP charges to an extra file or to the normal output file
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param output_runinfo ...
!> \param natom number of atoms
! **************************************************************************************************
   SUBROUTINE print_resp_charges(qs_env, resp_env, output_runinfo, natom)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      INTEGER, INTENT(IN)                                :: output_runinfo, natom

      CHARACTER(len=*), PARAMETER :: routineN = 'print_resp_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: handle, output_file
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: input, print_key, resp_section

      CALL timeset(routineN, handle)

      NULLIFY (particle_set, qs_kind_set, input, logger, resp_section, print_key)

      CALL get_qs_env(qs_env, input=input, particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      resp_section => section_vals_get_subs_vals(input, "PROPERTIES%RESP")
      print_key => section_vals_get_subs_vals(resp_section, &
                                              "PRINT%RESP_CHARGES_TO_FILE")
      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           resp_section, "PRINT%RESP_CHARGES_TO_FILE"), &
                cp_p_file)) THEN
         output_file = cp_print_key_unit_nr(logger, resp_section, &
                                            "PRINT%RESP_CHARGES_TO_FILE", &
                                            extension=".resp", &
                                            file_status="REPLACE", &
                                            file_action="WRITE", &
                                            file_form="FORMATTED")
         IF (output_file > 0) THEN
            filename = cp_print_key_generate_filename(logger, &
                                                      print_key, extension=".resp", &
                                                      my_local=.FALSE.)
            CALL print_atomic_charges(particle_set, qs_kind_set, output_file, title="RESP charges:", &
                                      atomic_charges=resp_env%rhs(1:natom))
            IF (output_runinfo > 0) WRITE (output_runinfo, '(2X,A,/)') "PRINTED RESP CHARGES TO FILE"
         ENDIF

         CALL cp_print_key_finished_output(output_file, logger, resp_section, &
                                           "PRINT%RESP_CHARGES_TO_FILE")
      ELSE
         CALL print_atomic_charges(particle_set, qs_kind_set, output_runinfo, title="RESP charges:", &
                                   atomic_charges=resp_env%rhs(1:natom))
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE print_resp_charges

! **************************************************************************************************
!> \brief print potential generated by RESP charges to file
!> \param qs_env the qs environment
!> \param resp_env the resp environment
!> \param particles ...
!> \param natom number of atoms
!> \param output_runinfo ...
! **************************************************************************************************
   SUBROUTINE print_pot_from_resp_charges(qs_env, resp_env, particles, natom, output_runinfo)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(resp_type), POINTER                           :: resp_env
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: natom, output_runinfo

      CHARACTER(len=*), PARAMETER :: routineN = 'print_pot_from_resp_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: my_pos_cube
      INTEGER                                            :: handle, ip, jx, jy, jz, unit_nr
      LOGICAL                                            :: append_cube
      REAL(KIND=dp)                                      :: dvol, normalize_factor, rms, rrms, &
                                                            sum_diff, sum_hartree, udvol
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: aux_r, rho_resp, v_resp_gspace, &
                                                            v_resp_rspace
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: v_hartree_rspace
      TYPE(section_vals_type), POINTER                   :: input, print_key, resp_section

      CALL timeset(routineN, handle)

      NULLIFY (auxbas_pw_pool, logger, pw_env, poisson_env, input, print_key, &
               para_env, resp_section, v_hartree_rspace)
      CALL get_qs_env(qs_env, &
                      input=input, &
                      para_env=para_env, &
                      pw_env=pw_env, &
                      v_hartree_rspace=v_hartree_rspace)
      normalize_factor = SQRT((resp_env%eta/pi)**3)
      resp_section => section_vals_get_subs_vals(input, "PROPERTIES%RESP")
      print_key => section_vals_get_subs_vals(resp_section, &
                                              "PRINT%V_RESP_CUBE")
      logger => cp_get_default_logger()

      !*** calculate potential generated from RESP charges
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)

      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             rho_resp%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             v_resp_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             v_resp_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      CALL pw_zero(rho_resp%pw)
      CALL calculate_rho_resp_all(rho_resp, resp_env%rhs, natom, &
                                  resp_env%eta, qs_env)
      CALL pw_zero(v_resp_gspace%pw)
      CALL pw_poisson_solve(poisson_env, rho_resp%pw, &
                            vhartree=v_resp_gspace%pw)
      CALL pw_zero(v_resp_rspace%pw)
      CALL pw_transfer(v_resp_gspace%pw, v_resp_rspace%pw)
      dvol = v_resp_rspace%pw%pw_grid%dvol
      CALL pw_scale(v_resp_rspace%pw, dvol)
      CALL pw_scale(v_resp_rspace%pw, -normalize_factor)
      ! REPEAT: correct for offset, take into account that potentials have reverse sign
      ! and are scaled by dvol
      IF (resp_env%use_repeat_method) THEN
         v_resp_rspace%pw%cr3d(:, :, :) = v_resp_rspace%pw%cr3d(:, :, :)-resp_env%offset*dvol
      ENDIF
      CALL pw_release(v_resp_gspace%pw)
      CALL pw_release(rho_resp%pw)

      !***now print the v_resp_rspace%pw to a cube file if requested
      IF (BTEST(cp_print_key_should_output(logger%iter_info, resp_section, &
                                           "PRINT%V_RESP_CUBE"), cp_p_file)) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         append_cube = section_get_lval(resp_section, "PRINT%V_RESP_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         unit_nr = cp_print_key_unit_nr(logger, resp_section, &
                                        "PRINT%V_RESP_CUBE", &
                                        extension=".cube", &
                                        file_position=my_pos_cube)
         udvol = 1.0_dp/dvol
         CALL pw_copy(v_resp_rspace%pw, aux_r%pw)
         CALL pw_scale(aux_r%pw, udvol)
         CALL cp_pw_to_cube(aux_r%pw, unit_nr, "RESP POTENTIAL", particles=particles, &
                            stride=section_get_ivals(resp_section, &
                                                     "PRINT%V_RESP_CUBE%STRIDE"))
         CALL cp_print_key_finished_output(unit_nr, logger, resp_section, &
                                           "PRINT%V_RESP_CUBE")
         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
      ENDIF

      !*** RMS and RRMS
      sum_diff = 0.0_dp
      sum_hartree = 0.0_dp
      rms = 0.0_dp
      rrms = 0.0_dp
      DO ip = 1, resp_env%npoints_proc
         jx = resp_env%fitpoints(1, ip)
         jy = resp_env%fitpoints(2, ip)
         jz = resp_env%fitpoints(3, ip)
         sum_diff = sum_diff+(v_hartree_rspace%cr3d(jx, jy, jz)- &
                              v_resp_rspace%pw%cr3d(jx, jy, jz))**2
         sum_hartree = sum_hartree+v_hartree_rspace%cr3d(jx, jy, jz)**2
      ENDDO
      CALL mp_sum(sum_diff, para_env%group)
      CALL mp_sum(sum_hartree, para_env%group)
      rms = SQRT(sum_diff/resp_env%npoints)
      rrms = SQRT(sum_diff/sum_hartree)
      IF (output_runinfo > 0) THEN
         WRITE (output_runinfo, '(2X,A,T69,ES12.5)') "Root-mean-square (RMS) "// &
            "error of RESP fit:", rms
         WRITE (output_runinfo, '(2X,A,T69,ES12.5,/)') "Relative root-mean-square "// &
            "(RRMS) error of RESP fit:", rrms
      ENDIF

      CALL pw_release(v_resp_rspace%pw)

      CALL timestop(handle)

   END SUBROUTINE print_pot_from_resp_charges

END MODULE qs_resp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation and writing of projected density of  states
!>         The DOS is computed per angular momentum and per kind
!> \par History
!>      -
!> \author Marcella (29.02.2008,MK)
! **************************************************************************************************
MODULE qs_pdos
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_diag,                      ONLY: cp_fm_power
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_init_random,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: nso,&
                                              nsoset
   USE orbital_symbols,                 ONLY: l_sym,&
                                              sgf_symbol
   USE particle_types,                  ONLY: particle_type
   USE preconditioner_types,            ONLY: preconditioner_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_mo_methods,                   ONLY: calculate_subspace_eigenvalues
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_type
   USE qs_ot_eigensolver,               ONLY: ot_eigensolver
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_pdos'

! **************************************************************************************************
   ! *** Public subroutines ***

   PUBLIC :: calculate_projected_dos

   TYPE ldos_type
      INTEGER :: maxl, nlist
      LOGICAL :: separate_components
      INTEGER, DIMENSION(:), POINTER           :: list_index
      REAL(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: pdos_array
   END TYPE ldos_type

   TYPE r_ldos_type
      INTEGER :: nlist, npoints
      INTEGER, DIMENSION(:, :), POINTER :: index_grid_local
      INTEGER, DIMENSION(:), POINTER           :: list_index
      REAL(KIND=dp), DIMENSION(:), POINTER     :: x_range, y_range, z_range
      REAL(KIND=dp), DIMENSION(:), POINTER     :: eval_range
      REAL(KIND=dp), DIMENSION(:), &
         POINTER                                 :: pdos_array
   END TYPE r_ldos_type

   TYPE ldos_p_type
      TYPE(ldos_type), POINTER :: ldos
   END TYPE ldos_p_type

   TYPE r_ldos_p_type
      TYPE(r_ldos_type), POINTER :: ldos
   END TYPE r_ldos_p_type
CONTAINS

! **************************************************************************************************
!> \brief   Compute and write projected density of states
!> \param mo_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param qs_env ...
!> \param dft_section ...
!> \param ispin ...
!> \param xas_mittle ...
!> \date    26.02.2008
!> \par History:
!>       -
!> \par Variables
!>       -
!>       -
!> \author  MI
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE calculate_projected_dos(mo_set, atomic_kind_set, qs_kind_set, particle_set, qs_env, &
                                      dft_section, ispin, xas_mittle)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      INTEGER, INTENT(IN), OPTIONAL                      :: ispin
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: xas_mittle

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_projected_dos', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=16)                                  :: fmtstr2
      CHARACTER(LEN=27)                                  :: fmtstr1
      CHARACTER(LEN=6), ALLOCATABLE, DIMENSION(:, :, :)  :: tmp_str
      CHARACTER(LEN=default_string_length)               :: kind_name, my_act, my_mittle, my_pos, &
                                                            spin(2)
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: ldos_index, r_ldos_index
      INTEGER :: handle, homo, i, iatom, ikind, il, ildos, im, imo, in_x, in_y, in_z, ir, irow, &
         iset, isgf, ishell, iso, iterstep, iw, j, jx, jy, jz, k, lcomponent, lshell, maxl, &
         maxlgto, my_spin, n_dependent, n_r_ldos, n_rep, nao, natom, ncol_global, nkind, nldos, &
         nlumo, nmo, np_tot, npoints, nrow_global, nset, nsgf, nvirt, out_each, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: firstrow
      INTEGER, DIMENSION(:), POINTER                     :: list, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: bo, l
      LOGICAL                                            :: append, do_ldos, do_r_ldos, do_virt, &
                                                            ionode, separate_components, &
                                                            should_output
      LOGICAL, DIMENSION(:, :), POINTER                  :: read_r
      REAL(KIND=dp)                                      :: dh(3, 3), dvol, e_fermi, r(3), r_vec(3), &
                                                            ratom(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues, evals_virt, &
                                                            occupation_numbers
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pdos_array
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: s_matrix
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: matrix_shalf, matrix_shalfc, &
                                                            matrix_work, mo_coeff, mo_virt
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(ldos_p_type), DIMENSION(:), POINTER           :: ldos_p
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(r_ldos_p_type), DIMENSION(:), POINTER         :: r_ldos_p
      TYPE(section_vals_type), POINTER                   :: ldos_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source
      should_output = BTEST(cp_print_key_should_output(logger%iter_info, dft_section, &
                                                       "PRINT%PDOS"), cp_p_file)
      output_unit = cp_logger_get_default_io_unit(logger)

      spin(1) = "ALPHA"
      spin(2) = "BETA"
      IF ((.NOT. should_output)) RETURN

      NULLIFY (context, s_matrix, orb_basis_set, para_env, pdos_array)
      NULLIFY (eigenvalues, fm_struct_tmp, matrix_work, matrix_shalf, mo_coeff, vecbuffer)
      NULLIFY (ldos_section, list, cell, pw_env, auxbas_pw_pool, mo_virt, evals_virt)
      NULLIFY (occupation_numbers, ldos_p, r_ldos_p, dft_control, occupation_numbers)

      CALL timeset(routineN, handle)
      iterstep = logger%iter_info%iteration(logger%iter_info%n_rlevel)

      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T3,A,T61,I10))') &
         " Calculate PDOS at iteration step ", iterstep
      CALL get_qs_env(qs_env=qs_env, &
                      matrix_s=s_matrix)

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, nsgf=nsgf, maxlgto=maxlgto)
      nkind = SIZE(atomic_kind_set)

      CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff, homo=homo, nao=nao, nmo=nmo, &
                      mu=e_fermi)
      CALL cp_fm_get_info(mo_coeff, &
                          context=context, para_env=para_env, &
                          nrow_global=nrow_global, &
                          ncol_global=ncol_global)

      CALL section_vals_val_get(dft_section, "PRINT%PDOS%OUT_EACH_MO", i_val=out_each)
      IF (out_each == -1) out_each = nao+1
      CALL section_vals_val_get(dft_section, "PRINT%PDOS%NLUMO", i_val=nlumo)
      IF (nlumo == -1) nlumo = nao-homo
      do_virt = (nlumo > (nmo-homo))
      nvirt = nlumo-(nmo-homo)
      ! Generate virtual orbitals
      IF (do_virt) THEN
         IF (PRESENT(ispin)) THEN
            my_spin = ispin
         ELSE
            my_spin = 1
         END IF

         CALL generate_virtual_mo(qs_env, mo_set, evals_virt, mo_virt, nvirt, ispin=my_spin)
      ELSE
         NULLIFY (evals_virt, mo_virt)
         nvirt = 0
      END IF

      ! Create S^1/2 : from sparse to full matrix
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=context, &
                               nrow_global=nrow_global, ncol_global=nrow_global)
      CALL cp_fm_create(matrix_shalf, fm_struct_tmp, name="matrix_shalf")
      CALL cp_fm_create(matrix_work, fm_struct_tmp, name="matrix_work")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL copy_dbcsr_to_fm(s_matrix(1)%matrix, matrix_shalf)
      CALL cp_fm_power(matrix_shalf, matrix_work, 0.5_dp, EPSILON(0.0_dp), n_dependent)
      CALL cp_fm_release(matrix_work)

      ! Multiply S^(1/2) time the mOS coefficients to get orthonormalized MOS
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=context, &
                               nrow_global=nrow_global, ncol_global=ncol_global)
      CALL cp_fm_create(matrix_shalfc, fm_struct_tmp, name="matrix_shalfc")
      CALL cp_gemm("N", "N", nrow_global, ncol_global, nrow_global, &
                   1.0_dp, matrix_shalf, mo_coeff, 0.0_dp, matrix_shalfc)
      CALL cp_fm_struct_release(fm_struct_tmp)

      IF (do_virt) THEN
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T3,A,T14,I10,T27,A))') &
            " Compute ", nvirt, " additional unoccupied KS orbitals"
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=context, &
                                  nrow_global=nrow_global, ncol_global=nvirt)
         CALL cp_fm_create(matrix_work, fm_struct_tmp, name="matrix_shalfc")
         CALL cp_gemm("N", "N", nrow_global, nvirt, nrow_global, &
                      1.0_dp, matrix_shalf, mo_virt, 0.0_dp, matrix_work)
         CALL cp_fm_struct_release(fm_struct_tmp)
      END IF

      CALL cp_fm_release(matrix_shalf)
      ! Array to store the PDOS per kind and angular momentum
      do_ldos = .FALSE.
      ldos_section => section_vals_get_subs_vals(dft_section, "PRINT%PDOS%LDOS")

      CALL section_vals_get(ldos_section, n_repetition=nldos)
      IF (nldos > 0) THEN
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T3,A,T61,I10))') &
            " Prepare the list of atoms for LDOS.   Number of lists: ", nldos
         do_ldos = .TRUE.
         ALLOCATE (ldos_p(nldos))
         ALLOCATE (ldos_index(nldos))
         DO ildos = 1, nldos
            WRITE (ldos_index(ildos), '(I0)') ildos
            ALLOCATE (ldos_p(ildos)%ldos)
            NULLIFY (ldos_p(ildos)%ldos%pdos_array)
            NULLIFY (ldos_p(ildos)%ldos%list_index)

            CALL section_vals_val_get(ldos_section, "LIST", i_rep_section=ildos, n_rep_val=n_rep)
            IF (n_rep > 0) THEN
               ldos_p(ildos)%ldos%nlist = 0
               DO ir = 1, n_rep
                  NULLIFY (list)
                  CALL section_vals_val_get(ldos_section, "LIST", i_rep_section=ildos, i_rep_val=ir, &
                                            i_vals=list)
                  IF (ASSOCIATED(list)) THEN
                     CALL reallocate(ldos_p(ildos)%ldos%list_index, 1, ldos_p(ildos)%ldos%nlist+SIZE(list))
                     DO i = 1, SIZE(list)
                        ldos_p(ildos)%ldos%list_index(i+ldos_p(ildos)%ldos%nlist) = list(i)
                     END DO
                     ldos_p(ildos)%ldos%nlist = ldos_p(ildos)%ldos%nlist+SIZE(list)
                  END IF
               END DO
            ELSE
               ! stop, LDOS without list of atoms is not implemented
            END IF

            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='((T10,A,T18,I6,T25,A,T36,I10,A))') &
               " List ", ildos, " contains ", ldos_p(ildos)%ldos%nlist, " atoms"
            CALL section_vals_val_get(ldos_section, "COMPONENTS", i_rep_section=ildos, &
                                      l_val=ldos_p(ildos)%ldos%separate_components)
            IF (ldos_p(ildos)%ldos%separate_components) THEN
               ALLOCATE (ldos_p(ildos)%ldos%pdos_array(nsoset(maxlgto), nmo+nvirt))
            ELSE
               ALLOCATE (ldos_p(ildos)%ldos%pdos_array(0:maxlgto, nmo+nvirt))
            END IF
            ldos_p(ildos)%ldos%pdos_array = 0.0_dp
            ldos_p(ildos)%ldos%maxl = -1

         END DO
      END IF

      do_r_ldos = .FALSE.
      ldos_section => section_vals_get_subs_vals(dft_section, "PRINT%PDOS%R_LDOS")
      CALL section_vals_get(ldos_section, n_repetition=n_r_ldos)
      IF (n_r_ldos > 0) THEN
         do_r_ldos = .TRUE.
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T3,A,T61,I10))') &
            " Prepare the list of points for R_LDOS.   Number of lists: ", n_r_ldos
         ALLOCATE (r_ldos_p(n_r_ldos))
         ALLOCATE (r_ldos_index(n_r_ldos))
         CALL get_qs_env(qs_env=qs_env, &
                         cell=cell, &
                         dft_control=dft_control, &
                         pw_env=pw_env)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                         pw_pools=pw_pools)

         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_g%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         ALLOCATE (read_r(4, n_r_ldos))
         DO ildos = 1, n_r_ldos
            WRITE (r_ldos_index(ildos), '(I0)') ildos
            ALLOCATE (r_ldos_p(ildos)%ldos)
            NULLIFY (r_ldos_p(ildos)%ldos%pdos_array)
            NULLIFY (r_ldos_p(ildos)%ldos%list_index)

            CALL section_vals_val_get(ldos_section, "LIST", i_rep_section=ildos, n_rep_val=n_rep)
            IF (n_rep > 0) THEN
               r_ldos_p(ildos)%ldos%nlist = 0
               DO ir = 1, n_rep
                  NULLIFY (list)
                  CALL section_vals_val_get(ldos_section, "LIST", i_rep_section=ildos, i_rep_val=ir, &
                                            i_vals=list)
                  IF (ASSOCIATED(list)) THEN
                     CALL reallocate(r_ldos_p(ildos)%ldos%list_index, 1, r_ldos_p(ildos)%ldos%nlist+SIZE(list))
                     DO i = 1, SIZE(list)
                        r_ldos_p(ildos)%ldos%list_index(i+r_ldos_p(ildos)%ldos%nlist) = list(i)
                     END DO
                     r_ldos_p(ildos)%ldos%nlist = r_ldos_p(ildos)%ldos%nlist+SIZE(list)
                  END IF
               END DO
            ELSE
               ! stop, LDOS without list of atoms is not implemented
            END IF

            ALLOCATE (r_ldos_p(ildos)%ldos%pdos_array(nmo+nvirt))
            r_ldos_p(ildos)%ldos%pdos_array = 0.0_dp
            read_r(1:3, ildos) = .FALSE.
            CALL section_vals_val_get(ldos_section, "XRANGE", i_rep_section=ildos, explicit=read_r(1, ildos))
            IF (read_r(1, ildos)) THEN
               CALL section_vals_val_get(ldos_section, "XRANGE", i_rep_section=ildos, r_vals= &
                                         r_ldos_p(ildos)%ldos%x_range)
            ELSE
               ALLOCATE (r_ldos_p(ildos)%ldos%x_range(2))
               r_ldos_p(ildos)%ldos%x_range = 0.0_dp
            END IF
            CALL section_vals_val_get(ldos_section, "YRANGE", i_rep_section=ildos, explicit=read_r(2, ildos))
            IF (read_r(2, ildos)) THEN
               CALL section_vals_val_get(ldos_section, "YRANGE", i_rep_section=ildos, r_vals= &
                                         r_ldos_p(ildos)%ldos%y_range)
            ELSE
               ALLOCATE (r_ldos_p(ildos)%ldos%y_range(2))
               r_ldos_p(ildos)%ldos%y_range = 0.0_dp
            END IF
            CALL section_vals_val_get(ldos_section, "ZRANGE", i_rep_section=ildos, explicit=read_r(3, ildos))
            IF (read_r(3, ildos)) THEN
               CALL section_vals_val_get(ldos_section, "ZRANGE", i_rep_section=ildos, r_vals= &
                                         r_ldos_p(ildos)%ldos%z_range)
            ELSE
               ALLOCATE (r_ldos_p(ildos)%ldos%z_range(2))
               r_ldos_p(ildos)%ldos%z_range = 0.0_dp
            END IF

            CALL section_vals_val_get(ldos_section, "ERANGE", i_rep_section=ildos, explicit=read_r(4, ildos))
            IF (read_r(4, ildos)) THEN
               CALL section_vals_val_get(ldos_section, "ERANGE", i_rep_section=ildos, &
                                         r_vals=r_ldos_p(ildos)%ldos%eval_range)
            ELSE
               ALLOCATE (r_ldos_p(ildos)%ldos%eval_range(2))
               r_ldos_p(ildos)%ldos%eval_range(1) = -HUGE(0.0_dp)
               r_ldos_p(ildos)%ldos%eval_range(2) = +HUGE(0.0_dp)
            END IF

            bo => wf_r%pw%pw_grid%bounds_local
            dh = wf_r%pw%pw_grid%dh
            dvol = wf_r%pw%pw_grid%dvol
            np_tot = wf_r%pw%pw_grid%npts(1)*wf_r%pw%pw_grid%npts(2)*wf_r%pw%pw_grid%npts(3)
            ALLOCATE (r_ldos_p(ildos)%ldos%index_grid_local(3, np_tot))

            r_ldos_p(ildos)%ldos%npoints = 0
            DO jz = bo(1, 3), bo(2, 3)
            DO jy = bo(1, 2), bo(2, 2)
            DO jx = bo(1, 1), bo(2, 1)
               !compute the position of the grid point
               i = jx-wf_r%pw%pw_grid%bounds(1, 1)
               j = jy-wf_r%pw%pw_grid%bounds(1, 2)
               k = jz-wf_r%pw%pw_grid%bounds(1, 3)
               r(3) = k*dh(3, 3)+j*dh(3, 2)+i*dh(3, 1)
               r(2) = k*dh(2, 3)+j*dh(2, 2)+i*dh(2, 1)
               r(1) = k*dh(1, 3)+j*dh(1, 2)+i*dh(1, 1)

               DO il = 1, r_ldos_p(ildos)%ldos%nlist
                  iatom = r_ldos_p(ildos)%ldos%list_index(il)
                  ratom = particle_set(iatom)%r
                  r_vec = pbc(ratom, r, cell)
                  IF (cell%orthorhombic) THEN
                     IF (cell%perd(1) == 0) r_vec(1) = MODULO(r_vec(1), cell%hmat(1, 1))
                     IF (cell%perd(2) == 0) r_vec(2) = MODULO(r_vec(2), cell%hmat(2, 2))
                     IF (cell%perd(3) == 0) r_vec(3) = MODULO(r_vec(3), cell%hmat(3, 3))
                  END IF

                  in_x = 0
                  in_y = 0
                  in_z = 0
                  IF (r_ldos_p(ildos)%ldos%x_range(1) /= 0.0_dp) THEN
                     IF (r_vec(1) > r_ldos_p(ildos)%ldos%x_range(1) .AND. &
                         r_vec(1) < r_ldos_p(ildos)%ldos%x_range(2)) THEN
                        in_x = 1
                     END IF
                  ELSE
                     in_x = 1
                  END IF
                  IF (r_ldos_p(ildos)%ldos%y_range(1) /= 0.0_dp) THEN
                     IF (r_vec(2) > r_ldos_p(ildos)%ldos%y_range(1) .AND. &
                         r_vec(2) < r_ldos_p(ildos)%ldos%y_range(2)) THEN
                        in_y = 1
                     END IF
                  ELSE
                     in_y = 1
                  END IF
                  IF (r_ldos_p(ildos)%ldos%z_range(1) /= 0.0_dp) THEN
                     IF (r_vec(3) > r_ldos_p(ildos)%ldos%z_range(1) .AND. &
                         r_vec(3) < r_ldos_p(ildos)%ldos%z_range(2)) THEN
                        in_z = 1
                     END IF
                  ELSE
                     in_z = 1
                  END IF
                  IF (in_x*in_y*in_z > 0) THEN
                     r_ldos_p(ildos)%ldos%npoints = r_ldos_p(ildos)%ldos%npoints+1
                     r_ldos_p(ildos)%ldos%index_grid_local(1, r_ldos_p(ildos)%ldos%npoints) = jx
                     r_ldos_p(ildos)%ldos%index_grid_local(2, r_ldos_p(ildos)%ldos%npoints) = jy
                     r_ldos_p(ildos)%ldos%index_grid_local(3, r_ldos_p(ildos)%ldos%npoints) = jz
                     EXIT
                  END IF
               END DO
            END DO
            END DO
            END DO
            CALL reallocate(r_ldos_p(ildos)%ldos%index_grid_local, 1, 3, 1, r_ldos_p(ildos)%ldos%npoints)
            npoints = r_ldos_p(ildos)%ldos%npoints
            CALL mp_sum(npoints, para_env%group)
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='((T10,A,T18,I6,T25,A,T36,I10,A))') &
               " List ", ildos, " contains ", npoints, " grid points"
         END DO
      END IF

      CALL section_vals_val_get(dft_section, "PRINT%PDOS%COMPONENTS", l_val=separate_components)
      IF (separate_components) THEN
         ALLOCATE (pdos_array(nsoset(maxlgto), nkind, nmo+nvirt))
      ELSE
         ALLOCATE (pdos_array(0:maxlgto, nkind, nmo+nvirt))
      END IF
      IF (do_virt) THEN
         ALLOCATE (eigenvalues(nmo+nvirt))
         eigenvalues(1:nmo) = mo_set%eigenvalues(1:nmo)
         eigenvalues(nmo+1:nmo+nvirt) = evals_virt(1:nvirt)
         ALLOCATE (occupation_numbers(nmo+nvirt))
         occupation_numbers(:) = 0.0_dp
         occupation_numbers(1:nmo) = mo_set%occupation_numbers(1:nmo)
      ELSE
         eigenvalues => mo_set%eigenvalues
         occupation_numbers => mo_set%occupation_numbers
      END IF

      pdos_array = 0.0_dp
      nao = mo_set%nao
      ALLOCATE (vecbuffer(1, nao))
      vecbuffer = 0.0_dp
      ALLOCATE (firstrow(natom))
      firstrow = 0

      !Adjust energy range for r_ldos
      DO ildos = 1, n_r_ldos
         IF (eigenvalues(1) > r_ldos_p(ildos)%ldos%eval_range(1)) &
            r_ldos_p(ildos)%ldos%eval_range(1) = eigenvalues(1)
         IF (eigenvalues(nmo+nvirt) < r_ldos_p(ildos)%ldos%eval_range(2)) &
            r_ldos_p(ildos)%ldos%eval_range(2) = eigenvalues(nmo+nvirt)
      END DO

      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T15,A))') &
         "---- PDOS: start iteration on the KS states --- "

      DO imo = 1, nmo+nvirt

         IF (output_unit > 0 .AND. MOD(imo, out_each) == 0) WRITE (UNIT=output_unit, FMT='((T20,A,I10))') &
            " KS state index : ", imo
         ! Extract the eigenvector from the distributed full matrix
         IF (imo > nmo) THEN
            CALL cp_fm_get_submatrix(matrix_work, vecbuffer, 1, imo-nmo, &
                                     nao, 1, transpose=.TRUE.)
         ELSE
            CALL cp_fm_get_submatrix(matrix_shalfc, vecbuffer, 1, imo, &
                                     nao, 1, transpose=.TRUE.)
         END IF

         ! Calculate the pdos for all the kinds
         irow = 1
         DO iatom = 1, natom
            firstrow(iatom) = irow
            NULLIFY (orb_basis_set)
            CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                   nset=nset, &
                                   nshell=nshell, &
                                   l=l, maxl=maxl)
            IF (separate_components) THEN
               isgf = 1
               DO iset = 1, nset
                  DO ishell = 1, nshell(iset)
                     lshell = l(ishell, iset)
                     DO iso = 1, nso(lshell)
                        lcomponent = nsoset(lshell-1)+iso
                        pdos_array(lcomponent, ikind, imo) = &
                           pdos_array(lcomponent, ikind, imo)+ &
                           vecbuffer(1, irow)*vecbuffer(1, irow)
                        irow = irow+1
                     END DO ! iso
                  END DO ! ishell
               END DO ! iset
            ELSE
               isgf = 1
               DO iset = 1, nset
                  DO ishell = 1, nshell(iset)
                     lshell = l(ishell, iset)
                     DO iso = 1, nso(lshell)
                        pdos_array(lshell, ikind, imo) = &
                           pdos_array(lshell, ikind, imo)+ &
                           vecbuffer(1, irow)*vecbuffer(1, irow)
                        irow = irow+1
                     END DO ! iso
                  END DO ! ishell
               END DO ! iset
            END IF
         END DO ! iatom

         ! Calculate the pdos for all the lists
         DO ildos = 1, nldos
            DO il = 1, ldos_p(ildos)%ldos%nlist
               iatom = ldos_p(ildos)%ldos%list_index(il)

               irow = firstrow(iatom)
               NULLIFY (orb_basis_set)
               CALL get_atomic_kind(particle_set(iatom)%atomic_kind, kind_number=ikind)
               CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)

               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      nset=nset, &
                                      nshell=nshell, &
                                      l=l, maxl=maxl)
               ldos_p(ildos)%ldos%maxl = MAX(ldos_p(ildos)%ldos%maxl, maxl)
               IF (ldos_p(ildos)%ldos%separate_components) THEN
                  isgf = 1
                  DO iset = 1, nset
                     DO ishell = 1, nshell(iset)
                        lshell = l(ishell, iset)
                        DO iso = 1, nso(lshell)
                           lcomponent = nsoset(lshell-1)+iso
                           ldos_p(ildos)%ldos%pdos_array(lcomponent, imo) = &
                              ldos_p(ildos)%ldos%pdos_array(lcomponent, imo)+ &
                              vecbuffer(1, irow)*vecbuffer(1, irow)
                           irow = irow+1
                        END DO ! iso
                     END DO ! ishell
                  END DO ! iset
               ELSE
                  isgf = 1
                  DO iset = 1, nset
                     DO ishell = 1, nshell(iset)
                        lshell = l(ishell, iset)
                        DO iso = 1, nso(lshell)
                           ldos_p(ildos)%ldos%pdos_array(lshell, imo) = &
                              ldos_p(ildos)%ldos%pdos_array(lshell, imo)+ &
                              vecbuffer(1, irow)*vecbuffer(1, irow)
                           irow = irow+1
                        END DO ! iso
                     END DO ! ishell
                  END DO ! iset
               END IF
            END DO !il
         END DO !ildos

         ! Calculate the DOS projected in a given volume in real space
         DO ildos = 1, n_r_ldos
            IF (r_ldos_p(ildos)%ldos%eval_range(1) <= eigenvalues(imo) .AND. &
                r_ldos_p(ildos)%ldos%eval_range(2) >= eigenvalues(imo)) THEN

               IF (imo > nmo) THEN
                  CALL calculate_wavefunction(mo_virt, imo-nmo, &
                                              wf_r, wf_g, atomic_kind_set, qs_kind_set, cell, dft_control, particle_set, &
                                              pw_env)
               ELSE
                  CALL calculate_wavefunction(mo_coeff, imo, &
                                              wf_r, wf_g, atomic_kind_set, qs_kind_set, cell, dft_control, particle_set, &
                                              pw_env)
               END IF
               r_ldos_p(ildos)%ldos%pdos_array(imo) = 0.0_dp
               DO il = 1, r_ldos_p(ildos)%ldos%npoints
                  j = j+1
                  jx = r_ldos_p(ildos)%ldos%index_grid_local(1, il)
                  jy = r_ldos_p(ildos)%ldos%index_grid_local(2, il)
                  jz = r_ldos_p(ildos)%ldos%index_grid_local(3, il)
                  r_ldos_p(ildos)%ldos%pdos_array(imo) = r_ldos_p(ildos)%ldos%pdos_array(imo)+ &
                                                         wf_r%pw%cr3d(jx, jy, jz)*wf_r%pw%cr3d(jx, jy, jz)
               END DO
               r_ldos_p(ildos)%ldos%pdos_array(imo) = r_ldos_p(ildos)%ldos%pdos_array(imo)*dvol
            END IF
         END DO
      END DO ! imo

      CALL cp_fm_release(matrix_shalfc)
      DEALLOCATE (vecbuffer)

      CALL section_vals_val_get(dft_section, "PRINT%PDOS%APPEND", l_val=append)
      IF (append .AND. iterstep > 1) THEN
         my_pos = "APPEND"
      ELSE
         my_pos = "REWIND"
      END IF
      my_act = "WRITE"
      DO ikind = 1, nkind

         NULLIFY (orb_basis_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), name=kind_name)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, maxl=maxl)

         ! basis none has no associated maxl, and no pdos
         IF (maxl < 0) CYCLE

         IF (PRESENT(ispin)) THEN
            IF (PRESENT(xas_mittle)) THEN
               my_mittle = TRIM(xas_mittle)//TRIM(spin(ispin))//"_k"//TRIM(ADJUSTL(cp_to_string(ikind)))
            ELSE
               my_mittle = TRIM(spin(ispin))//"_k"//TRIM(ADJUSTL(cp_to_string(ikind)))
            END IF
            my_spin = ispin
         ELSE
            my_mittle = "k"//TRIM(ADJUSTL(cp_to_string(ikind)))
            my_spin = 1
         END IF

         iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%PDOS", &
                                   extension=".pdos", file_position=my_pos, file_action=my_act, &
                                   file_form="FORMATTED", middle_name=TRIM(my_mittle))
         IF (iw > 0) THEN

            fmtstr1 = "(I8,2X,2F16.6,  (2X,F16.8))"
            fmtstr2 = "(A42,  (10X,A8))"
            IF (separate_components) THEN
               WRITE (UNIT=fmtstr1(15:16), FMT="(I2)") nsoset(maxl)
               WRITE (UNIT=fmtstr2(6:7), FMT="(I2)") nsoset(maxl)
            ELSE
               WRITE (UNIT=fmtstr1(15:16), FMT="(I2)") maxl+1
               WRITE (UNIT=fmtstr2(6:7), FMT="(I2)") maxl+1
            END IF

            WRITE (UNIT=iw, FMT="(A,I0,A,F12.6,A)") &
               "# Projected DOS for atomic kind "//TRIM(kind_name)//" at iteration step i = ", &
               iterstep, ", E(Fermi) = ", e_fermi, " a.u."
            IF (separate_components) THEN
               ALLOCATE (tmp_str(0:0, 0:maxl, -maxl:maxl))
               tmp_str = ""
               DO j = 0, maxl
                  DO i = -j, j
                     tmp_str(0, j, i) = sgf_symbol(0, j, i)
                  END DO
               END DO

               WRITE (UNIT=iw, FMT=fmtstr2) &
                  "#     MO Eigenvalue [a.u.]      Occupation", &
                  ((TRIM(tmp_str(0, il, im)), im=-il, il), il=0, maxl)
               DO imo = 1, nmo+nvirt
                  WRITE (UNIT=iw, FMT=fmtstr1) imo, eigenvalues(imo), occupation_numbers(imo), &
                     (pdos_array(lshell, ikind, imo), lshell=1, nsoset(maxl))
               END DO
               DEALLOCATE (tmp_str)
            ELSE
               WRITE (UNIT=iw, FMT=fmtstr2) &
                  "#     MO Eigenvalue [a.u.]      Occupation", &
                  (TRIM(l_sym(il)), il=0, maxl)
               DO imo = 1, nmo+nvirt
                  WRITE (UNIT=iw, FMT=fmtstr1) imo, eigenvalues(imo), occupation_numbers(imo), &
                     (pdos_array(lshell, ikind, imo), lshell=0, maxl)
               END DO
            END IF
         END IF
         CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                           "PRINT%PDOS")

      END DO ! ikind

      ! write the pdos for the lists, each ona different file,
      ! the filenames are indexed with the list number
      DO ildos = 1, nldos
         ! basis none has no associated maxl, and no pdos
         IF (ldos_p(ildos)%ldos%maxl > 0) THEN

            IF (PRESENT(ispin)) THEN
               IF (PRESENT(xas_mittle)) THEN
                  my_mittle = TRIM(xas_mittle)//TRIM(spin(ispin))//"_list"//TRIM(ldos_index(ildos))
               ELSE
                  my_mittle = TRIM(spin(ispin))//"_list"//TRIM(ldos_index(ildos))
               END IF
               my_spin = ispin
            ELSE
               my_mittle = "list"//TRIM(ldos_index(ildos))
               my_spin = 1
            END IF

            iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%PDOS", &
                                      extension=".pdos", file_position=my_pos, file_action=my_act, &
                                      file_form="FORMATTED", middle_name=TRIM(my_mittle))
            IF (iw > 0) THEN

               fmtstr1 = "(I8,2X,2F16.6,  (2X,F16.8))"
               fmtstr2 = "(A42,  (10X,A8))"
               IF (ldos_p(ildos)%ldos%separate_components) THEN
                  WRITE (UNIT=fmtstr1(15:16), FMT="(I2)") nsoset(ldos_p(ildos)%ldos%maxl)
                  WRITE (UNIT=fmtstr2(6:7), FMT="(I2)") nsoset(ldos_p(ildos)%ldos%maxl)
               ELSE
                  WRITE (UNIT=fmtstr1(15:16), FMT="(I2)") ldos_p(ildos)%ldos%maxl+1
                  WRITE (UNIT=fmtstr2(6:7), FMT="(I2)") ldos_p(ildos)%ldos%maxl+1
               END IF

               WRITE (UNIT=iw, FMT="(A,I0,A,I0,A,I0,A,F12.6,A)") &
                  "# Projected DOS for list ", ildos, " of ", ldos_p(ildos)%ldos%nlist, &
                  " atoms, at iteration step i = ", iterstep, &
                  ", E(Fermi) = ", e_fermi, " a.u."
               IF (ldos_p(ildos)%ldos%separate_components) THEN
                  ALLOCATE (tmp_str(0:0, 0:ldos_p(ildos)%ldos%maxl, -ldos_p(ildos)%ldos%maxl:ldos_p(ildos)%ldos%maxl))
                  tmp_str = ""
                  DO j = 0, ldos_p(ildos)%ldos%maxl
                     DO i = -j, j
                        tmp_str(0, j, i) = sgf_symbol(0, j, i)
                     END DO
                  END DO

                  WRITE (UNIT=iw, FMT=fmtstr2) &
                     "#     MO Eigenvalue [a.u.]      Occupation", &
                     ((TRIM(tmp_str(0, il, im)), im=-il, il), il=0, ldos_p(ildos)%ldos%maxl)
                  DO imo = 1, nmo+nvirt
                     WRITE (UNIT=iw, FMT=fmtstr1) imo, eigenvalues(imo), occupation_numbers(imo), &
                        (ldos_p(ildos)%ldos%pdos_array(lshell, imo), lshell=1, nsoset(ldos_p(ildos)%ldos%maxl))
                  END DO
                  DEALLOCATE (tmp_str)
               ELSE
                  WRITE (UNIT=iw, FMT=fmtstr2) &
                     "#     MO Eigenvalue [a.u.]      Occupation", &
                     (TRIM(l_sym(il)), il=0, ldos_p(ildos)%ldos%maxl)
                  DO imo = 1, nmo+nvirt
                     WRITE (UNIT=iw, FMT=fmtstr1) imo, eigenvalues(imo), occupation_numbers(imo), &
                        (ldos_p(ildos)%ldos%pdos_array(lshell, imo), lshell=0, ldos_p(ildos)%ldos%maxl)
                  END DO
               END IF
            END IF
            CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                              "PRINT%PDOS")
         END IF ! maxl>0
      END DO ! ildos

      ! write the pdos for the lists, each ona different file,
      ! the filenames are indexed with the list number
      DO ildos = 1, n_r_ldos

         npoints = r_ldos_p(ildos)%ldos%npoints
         CALL mp_sum(npoints, para_env%group)
         CALL mp_sum(np_tot, para_env%group)
         CALL mp_sum(r_ldos_p(ildos)%ldos%pdos_array, para_env%group)
         IF (PRESENT(ispin)) THEN
            IF (PRESENT(xas_mittle)) THEN
               my_mittle = TRIM(xas_mittle)//TRIM(spin(ispin))//"_r_list"//TRIM(r_ldos_index(ildos))
            ELSE
               my_mittle = TRIM(spin(ispin))//"_r_list"//TRIM(r_ldos_index(ildos))
            END IF
            my_spin = ispin
         ELSE
            my_mittle = "r_list"//TRIM(r_ldos_index(ildos))
            my_spin = 1
         END IF

         iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%PDOS", &
                                   extension=".pdos", file_position=my_pos, file_action=my_act, &
                                   file_form="FORMATTED", middle_name=TRIM(my_mittle))
         IF (iw > 0) THEN
            fmtstr1 = "(I8,2X,2F16.6,  (2X,F16.8))"
            fmtstr2 = "(A42,  (10X,A8))"

            WRITE (UNIT=iw, FMT="(A,I0,A,F12.6,F12.6,A,F12.6,A)") &
               "# Projected DOS in real space, using ", npoints, &
               " points of the grid, and eval in the range", r_ldos_p(ildos)%ldos%eval_range(1:2), &
               " Hartree, E(Fermi) = ", e_fermi, " a.u."
            WRITE (UNIT=iw, FMT="(A)") &
               "#     MO Eigenvalue [a.u.]      Occupation      LDOS"
            DO imo = 1, nmo+nvirt
               IF (r_ldos_p(ildos)%ldos%eval_range(1) <= eigenvalues(imo) .AND. &
                   r_ldos_p(ildos)%ldos%eval_range(2) >= eigenvalues(imo)) THEN
                  WRITE (UNIT=iw, FMT="(I8,2X,2F16.6,E20.10,E20.10)") imo, eigenvalues(imo), occupation_numbers(imo), &
                     r_ldos_p(ildos)%ldos%pdos_array(imo), r_ldos_p(ildos)%ldos%pdos_array(imo)*np_tot
               END IF
            END DO

         END IF
         CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                           "PRINT%PDOS")
      END DO

      ! deallocate local variables
      DEALLOCATE (pdos_array)
      DEALLOCATE (firstrow)
      IF (do_ldos) THEN
         DO ildos = 1, nldos
            DEALLOCATE (ldos_p(ildos)%ldos%pdos_array)
            DEALLOCATE (ldos_p(ildos)%ldos%list_index)
            DEALLOCATE (ldos_p(ildos)%ldos)
         END DO
         DEALLOCATE (ldos_p)
         DEALLOCATE (ldos_index)
      END IF
      IF (do_r_ldos) THEN
         DO ildos = 1, n_r_ldos
            DEALLOCATE (r_ldos_p(ildos)%ldos%index_grid_local)
            DEALLOCATE (r_ldos_p(ildos)%ldos%pdos_array)
            DEALLOCATE (r_ldos_p(ildos)%ldos%list_index)
            IF (.NOT. read_r(1, ildos)) THEN
               DEALLOCATE (r_ldos_p(ildos)%ldos%x_range)
            END IF
            IF (.NOT. read_r(2, ildos)) THEN
               DEALLOCATE (r_ldos_p(ildos)%ldos%y_range)
            END IF
            IF (.NOT. read_r(3, ildos)) THEN
               DEALLOCATE (r_ldos_p(ildos)%ldos%z_range)
            END IF
            IF (.NOT. read_r(4, ildos)) THEN
               DEALLOCATE (r_ldos_p(ildos)%ldos%eval_range)
            END IF
            DEALLOCATE (r_ldos_p(ildos)%ldos)
         END DO
         DEALLOCATE (read_r)
         DEALLOCATE (r_ldos_p)
         DEALLOCATE (r_ldos_index)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)
      END IF
      IF (do_virt) THEN
         DEALLOCATE (evals_virt)
         CALL cp_fm_release(mo_virt)
         CALL cp_fm_release(matrix_work)
         DEALLOCATE (eigenvalues)
         DEALLOCATE (occupation_numbers)
      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_projected_dos

! **************************************************************************************************
!> \brief   Compute additional virtual states  starting from the available MOS
!> \param qs_env ...
!> \param mo_set ...
!> \param evals_virt ...
!> \param mo_virt ...
!> \param nvirt ...
!> \param ispin ...
!> \date    08.03.2008
!> \par History:
!>       -
!> \par Variables
!>       -
!>       -
!> \author  MI
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE generate_virtual_mo(qs_env, mo_set, evals_virt, mo_virt, &
                                  nvirt, ispin)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mo_set_type), POINTER                         :: mo_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals_virt
      TYPE(cp_fm_type), POINTER                          :: mo_virt
      INTEGER, INTENT(IN)                                :: nvirt, ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_virtual_mo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nmo, nrow_global
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, s_matrix
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(preconditioner_type), POINTER                 :: local_preconditioner
      TYPE(scf_control_type), POINTER                    :: scf_control

      NULLIFY (evals_virt, mo_virt)
      ALLOCATE (evals_virt(nvirt))

      CALL get_qs_env(qs_env, matrix_ks=ks_matrix, matrix_s=s_matrix, &
                      scf_control=scf_control)
      CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff, nmo=nmo)
      CALL cp_fm_get_info(mo_coeff, context=context, para_env=para_env, &
                          nrow_global=nrow_global)

      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=context, &
                               nrow_global=nrow_global, ncol_global=nvirt)
      CALL cp_fm_create(mo_virt, fm_struct_tmp, name="virtual")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_init_random(mo_virt, nvirt)

      NULLIFY (local_preconditioner)

      CALL ot_eigensolver(matrix_h=ks_matrix(ispin)%matrix, matrix_s=s_matrix(1)%matrix, &
                          matrix_c_fm=mo_virt, matrix_orthogonal_space_fm=mo_coeff, &
                          eps_gradient=scf_control%eps_lumos, &
                          preconditioner=local_preconditioner, &
                          iter_max=scf_control%max_iter_lumos, &
                          size_ortho_space=nmo)

      CALL calculate_subspace_eigenvalues(mo_virt, ks_matrix(ispin)%matrix, &
                                          evals_virt)

   END SUBROUTINE generate_virtual_mo

END MODULE qs_pdos

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate Hirshfeld charges and related functions
!> \par History
!>      11.2014 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE hirshfeld_methods
   USE atom_kind_orbitals,              ONLY: calculate_atomic_density
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cube_utils,                      ONLY: cube_info_type
   USE hirshfeld_types,                 ONLY: get_hirshfeld_info,&
                                              hirshfeld_type,&
                                              set_hirshfeld_info
   USE input_constants,                 ONLY: shape_function_density,&
                                              shape_function_gaussian
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE periodic_table,                  ONLY: get_ptable_info
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_integrate_function
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qs_collocate_density,            ONLY: collocate_pgf_product_rspace
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_integrate_potential_low,      ONLY: integrate_pgf_product_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_modify_pab_block,             ONLY: FUNC_AB
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE realspace_grid_types,            ONLY: pw2rs,&
                                              realspace_grid_desc_type,&
                                              realspace_grid_type,&
                                              rs2pw,&
                                              rs_grid_release,&
                                              rs_grid_retain,&
                                              rs_grid_zero,&
                                              rs_pw_transfer
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hirshfeld_methods'

   PUBLIC :: create_shape_function, comp_hirshfeld_charges, &
             comp_hirshfeld_i_charges, write_hirshfeld_charges

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param charges ...
!> \param hirshfeld_env ...
!> \param particle_set ...
!> \param qs_kind_set ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE write_hirshfeld_charges(charges, hirshfeld_env, particle_set, &
                                      qs_kind_set, unit_nr)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: charges
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'write_hirshfeld_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: element_symbol
      INTEGER                                            :: iatom, ikind, natom, nspin
      REAL(KIND=dp)                                      :: refc, tc1, zeff

      natom = SIZE(charges, 1)
      nspin = SIZE(charges, 2)
      WRITE (unit_nr, '(/,T2,A)') '!-----------------------------------------------------------------------------!'
      WRITE (UNIT=unit_nr, FMT="(T28,A)") "Hirshfeld Charges"
      IF (nspin == 1) THEN
         WRITE (UNIT=unit_nr, FMT="(/,T3,A,A)") &
            "#Atom  Element  Kind ", " Ref Charge     Population                    Net charge"
      ELSE
         WRITE (UNIT=unit_nr, FMT="(/,T3,A,A)") &
            "#Atom  Element  Kind ", " Ref Charge     Population       Spin moment  Net charge"
      END IF
      tc1 = 0.0_dp
      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              element_symbol=element_symbol, kind_number=ikind)
         refc = hirshfeld_env%charges(iatom)
         CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)
         IF (nspin == 1) THEN
            WRITE (UNIT=unit_nr, FMT="(i7,T15,A2,T20,i3,T27,F8.3,T42,F8.3,T72,F8.3)") &
               iatom, element_symbol, ikind, refc, charges(iatom, 1), zeff-charges(iatom, 1)
         ELSE
            WRITE (UNIT=unit_nr, FMT="(i7,T15,A2,T20,i3,T27,F8.3,T36,2F8.3,T61,F8.3,T72,F8.3)") &
               iatom, element_symbol, ikind, refc, charges(iatom, 1), charges(iatom, 2), &
               charges(iatom, 1)-charges(iatom, 2), zeff-SUM(charges(iatom, :))
         END IF
         tc1 = tc1+(zeff-SUM(charges(iatom, :)))
      END DO
      WRITE (UNIT=unit_nr, FMT="(/,T3,A,T72,F8.3)") "Total Charge ", tc1
      WRITE (unit_nr, '(T2,A)') '!-----------------------------------------------------------------------------!'

   END SUBROUTINE write_hirshfeld_charges

! **************************************************************************************************
!> \brief ...
!> \param hirshfeld_env ...
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
! **************************************************************************************************
   SUBROUTINE create_shape_function(hirshfeld_env, qs_kind_set, atomic_kind_set)
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shape_function', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: ngto = 8

      CHARACTER(len=2)                                   :: esym
      INTEGER                                            :: ikind, nkind
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: al, rco, zeff
      REAL(KIND=dp), DIMENSION(ngto, 2)                  :: ppdens
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      IF(.NOT.(ASSOCIATED(hirshfeld_env)))CALL cp__a("hirshfeld_methods.F",151)

      nkind = SIZE(qs_kind_set)
      ALLOCATE (hirshfeld_env%kind_shape_fn(nkind))

      SELECT CASE (hirshfeld_env%shape_function_type)
      CASE (shape_function_gaussian)
         DO ikind = 1, nkind
            hirshfeld_env%kind_shape_fn(ikind)%numexp = 1
            ALLOCATE (hirshfeld_env%kind_shape_fn(ikind)%zet(1))
            ALLOCATE (hirshfeld_env%kind_shape_fn(ikind)%coef(1))
            CALL get_qs_kind(qs_kind_set(ikind), element_symbol=esym)
            rco = 2.0_dp
            CALL get_ptable_info(symbol=esym, covalent_radius=rco, found=found)
            rco = MAX(rco, 1.0_dp)
            al = 0.5_dp/rco**2
            hirshfeld_env%kind_shape_fn(ikind)%zet(1) = al
            hirshfeld_env%kind_shape_fn(ikind)%coef(1) = (al/pi)**1.5_dp
         END DO
      CASE (shape_function_density)
         ! calculate atomic density
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            qs_kind => qs_kind_set(ikind)
            CALL calculate_atomic_density(ppdens(:, :), atomic_kind, qs_kind, ngto, &
                                          confine=.FALSE.)
            hirshfeld_env%kind_shape_fn(ikind)%numexp = ngto
            ALLOCATE (hirshfeld_env%kind_shape_fn(ikind)%zet(ngto))
            ALLOCATE (hirshfeld_env%kind_shape_fn(ikind)%coef(ngto))
            hirshfeld_env%kind_shape_fn(ikind)%zet(:) = ppdens(:, 1)
            CALL get_qs_kind(qs_kind, zeff=zeff)
            hirshfeld_env%kind_shape_fn(ikind)%coef(:) = ppdens(:, 2)/zeff
         END DO

      CASE DEFAULT
         CALL cp__b("hirshfeld_methods.F",186,"Unknown shape function")
      END SELECT

   END SUBROUTINE create_shape_function

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param hirshfeld_env ...
!> \param charges ...
! **************************************************************************************************
   SUBROUTINE comp_hirshfeld_charges(qs_env, hirshfeld_env, charges)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'comp_hirshfeld_charges', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: is
      LOGICAL                                            :: rho_r_valid
      REAL(KIND=dp)                                      :: tnfun
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_p_type), POINTER                           :: rhonorm
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (rho_r)
      ! normalization function on grid
      CALL calculate_hirshfeld_normalization(qs_env, hirshfeld_env)
      ! check normalization
      tnfun = pw_integrate_function(hirshfeld_env%fnorm%pw)
      tnfun = ABS(tnfun-SUM(hirshfeld_env%charges))
      !
      ALLOCATE (rhonorm)
      !
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, rho=rho)
      CALL qs_rho_get(rho, rho_r=rho_r, rho_r_valid=rho_r_valid)
      CALL pw_env_get(pw_env=pw_env, auxbas_pw_pool=auxbas_pw_pool)
      CALL pw_pool_create_pw(auxbas_pw_pool, rhonorm%pw, use_data=REALDATA3D)
      ! loop over spins
      DO is = 1, SIZE(rho_r)
         IF (rho_r_valid) THEN
            CALL hfun_scale(rhonorm%pw%cr3d, rho_r(is)%pw%cr3d, &
                            hirshfeld_env%fnorm%pw%cr3d)
         ELSE
            CALL cp__b("hirshfeld_methods.F",233,"We need rho in real space")
         END IF
         CALL hirshfeld_integration(qs_env, hirshfeld_env, rhonorm, charges(:, is))
         charges(:, is) = charges(:, is)*hirshfeld_env%charges(:)
      END DO
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhonorm%pw)
      !
      DEALLOCATE (rhonorm)

   END SUBROUTINE comp_hirshfeld_charges
! **************************************************************************************************
!> \brief Calculate fout = fun1/fun2
!> \param fout ...
!> \param fun1 ...
!> \param fun2 ...
! **************************************************************************************************
   SUBROUTINE hfun_scale(fout, fun1, fun2)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: fout
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: fun1, fun2

      CHARACTER(len=*), PARAMETER :: routineN = 'hfun_scale', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: small = 1.0e-12_dp

      INTEGER                                            :: i1, i2, i3, n1, n2, n3

      n1 = SIZE(fout, 1)
      n2 = SIZE(fout, 2)
      n3 = SIZE(fout, 3)
      IF(.NOT.(n1 == SIZE(fun1, 1)))CALL cp__a("hirshfeld_methods.F",261)
      IF(.NOT.(n2 == SIZE(fun1, 2)))CALL cp__a("hirshfeld_methods.F",262)
      IF(.NOT.(n3 == SIZE(fun1, 3)))CALL cp__a("hirshfeld_methods.F",263)
      IF(.NOT.(n1 == SIZE(fun2, 1)))CALL cp__a("hirshfeld_methods.F",264)
      IF(.NOT.(n2 == SIZE(fun2, 2)))CALL cp__a("hirshfeld_methods.F",265)
      IF(.NOT.(n3 == SIZE(fun2, 3)))CALL cp__a("hirshfeld_methods.F",266)

      DO i3 = 1, n3
         DO i2 = 1, n2
            DO i1 = 1, n1
               IF (fun2(i1, i2, i3) > small) THEN
                  fout(i1, i2, i3) = fun1(i1, i2, i3)/fun2(i1, i2, i3)
               ELSE
                  fout(i1, i2, i3) = 0.0_dp
               END IF
            END DO
         END DO
      END DO

   END SUBROUTINE hfun_scale

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param hirshfeld_env ...
!> \param charges ...
!> \param ounit ...
! **************************************************************************************************
   SUBROUTINE comp_hirshfeld_i_charges(qs_env, hirshfeld_env, charges, ounit)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: charges
      INTEGER, INTENT(IN)                                :: ounit

      CHARACTER(len=*), PARAMETER :: routineN = 'comp_hirshfeld_i_charges', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxloop = 100
      REAL(KIND=dp), PARAMETER                           :: maxres = 1.0e-2_dp

      CHARACTER(len=3)                                   :: yesno
      INTEGER                                            :: iat, iloop, is, natom
      LOGICAL                                            :: rho_r_valid
      REAL(KIND=dp)                                      :: res, tnfun
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_p_type), POINTER                           :: rhonorm
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (rho_r)

      natom = SIZE(charges, 1)

      IF (ounit > 0) WRITE (ounit, "(/,T2,A)") "Hirshfeld charge iterations: Residuals ..."
      !
      ALLOCATE (rhonorm)
      !
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, rho=rho)
      CALL qs_rho_get(rho, rho_r=rho_r, rho_r_valid=rho_r_valid)
      CALL pw_env_get(pw_env=pw_env, auxbas_pw_pool=auxbas_pw_pool)
      CALL pw_pool_create_pw(auxbas_pw_pool, rhonorm%pw, use_data=REALDATA3D)
      !
      DO iloop = 1, maxloop

         ! normalization function on grid
         CALL calculate_hirshfeld_normalization(qs_env, hirshfeld_env)
         ! check normalization
         tnfun = pw_integrate_function(hirshfeld_env%fnorm%pw)
         tnfun = ABS(tnfun-SUM(hirshfeld_env%charges))
         ! loop over spins
         DO is = 1, SIZE(rho_r)
            IF (rho_r_valid) THEN
               CALL hfun_scale(rhonorm%pw%cr3d, rho_r(is)%pw%cr3d, &
                               hirshfeld_env%fnorm%pw%cr3d)
            ELSE
               CALL cp__b("hirshfeld_methods.F",336,"We need rho in real space")
            END IF
            CALL hirshfeld_integration(qs_env, hirshfeld_env, rhonorm, charges(:, is))
            charges(:, is) = charges(:, is)*hirshfeld_env%charges(:)
         END DO
         ! residual
         res = 0.0_dp
         DO iat = 1, natom
            res = res+(SUM(charges(iat, :))-hirshfeld_env%charges(iat))**2
         END DO
         res = SQRT(res/REAL(natom, KIND=dp))
         IF (ounit > 0) THEN
            yesno = "NO "
            IF (MOD(iloop, 10) == 0) yesno = "YES"
            WRITE (ounit, FMT="(F8.3)", ADVANCE=yesno) res
         END IF
         ! update
         DO iat = 1, natom
            hirshfeld_env%charges(iat) = SUM(charges(iat, :))
         END DO
         IF (res < maxres) EXIT

      END DO
      !
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhonorm%pw)
      !
      DEALLOCATE (rhonorm)

   END SUBROUTINE comp_hirshfeld_i_charges

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param hirshfeld_env ...
! **************************************************************************************************
   SUBROUTINE calculate_hirshfeld_normalization(qs_env, hirshfeld_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_hirshfeld_normalization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, iex, ikind, &
                                                            ithread, j, natom, npme, nthread, &
                                                            numexp
      INTEGER(KIND=int_8)                                :: subpatch_pattern
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      REAL(KIND=dp)                                      :: alpha, coef, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cube_info_type)                               :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), POINTER                           :: fnorm
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_desc_type), POINTER            :: auxbas_rs_desc
      TYPE(realspace_grid_type), POINTER                 :: rs_rho

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, cell=cell, &
                      dft_control=dft_control, particle_set=particle_set, pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_desc=auxbas_rs_desc, auxbas_rs_grid=rs_rho, &
                      auxbas_pw_pool=auxbas_pw_pool)
      cube_info = pw_env%cube_info(1)
      ! be careful in parallel nsmax is choosen with multigrid in mind!
      CALL rs_grid_retain(rs_rho)
      CALL rs_grid_zero(rs_rho)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      ALLOCATE (pab(1, 1))
      nthread = 1
      ithread = 0

      DO ikind = 1, SIZE(atomic_kind_set)
         numexp = hirshfeld_env%kind_shape_fn(ikind)%numexp
         IF (numexp <= 0) CYCLE
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         ALLOCATE (cores(natom))

         DO iex = 1, numexp
            alpha = hirshfeld_env%kind_shape_fn(ikind)%zet(iex)
            coef = hirshfeld_env%kind_shape_fn(ikind)%coef(iex)
            npme = 0
            cores = 0
            DO iatom = 1, natom
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               IF (rs_rho%desc%parallel .AND. .NOT. rs_rho%desc%distributed) THEN
                  ! replicated realspace grid, split the atoms up between procs
                  IF (MODULO(iatom, rs_rho%desc%group_size) == rs_rho%desc%my_pos) THEN
                     npme = npme+1
                     cores(npme) = iatom
                  ENDIF
               ELSE
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            END DO
            DO j = 1, npme
               iatom = cores(j)
               atom_a = atom_list(iatom)
               pab(1, 1) = hirshfeld_env%charges(atom_a)*coef
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               subpatch_pattern = 0
               ! la_max==0 so set lmax_global to 0
               CALL collocate_pgf_product_rspace(0, alpha, 0, 0, 0.0_dp, 0, ra, &
                                                 (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, rs_rho, &
                                                 cell, cube_info, eps_rho_rspace, ga_gb_function=FUNC_AB, &
                                                 ithread=ithread, use_subpatch=.TRUE., subpatch_pattern=subpatch_pattern, &
                                                 lmax_global=0)
            END DO
         END DO

         DEALLOCATE (cores)
      END DO
      DEALLOCATE (pab)

      NULLIFY (fnorm)
      CALL get_hirshfeld_info(hirshfeld_env, fnorm=fnorm)
      IF (ASSOCIATED(fnorm)) THEN
         CALL pw_release(fnorm%pw)
         DEALLOCATE (fnorm)
      ENDIF
      ALLOCATE (fnorm)
      CALL pw_pool_create_pw(auxbas_pw_pool, fnorm%pw, use_data=REALDATA3D)
      fnorm%pw%in_space = REALSPACE
      CALL set_hirshfeld_info(hirshfeld_env, fnorm=fnorm)

      CALL rs_pw_transfer(rs_rho, fnorm%pw, rs2pw)
      CALL rs_grid_release(rs_rho)

      CALL timestop(handle)

   END SUBROUTINE calculate_hirshfeld_normalization

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param hirshfeld_env ...
!> \param rfun ...
!> \param fval ...
!> \param fderiv ...
! **************************************************************************************************
   SUBROUTINE hirshfeld_integration(qs_env, hirshfeld_env, rfun, fval, fderiv)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      TYPE(pw_p_type), POINTER                           :: rfun
      REAL(KIND=dp), DIMENSION(:), INTENT(inout)         :: fval
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout), &
         OPTIONAL                                        :: fderiv

      CHARACTER(len=*), PARAMETER :: routineN = 'hirshfeld_integration', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, iex, ikind, &
                                                            ithread, j, natom, npme, nthread, &
                                                            numexp
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cores
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: do_force
      REAL(KIND=dp)                                      :: alpha, coef, dvol, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hab, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(realspace_grid_desc_type), POINTER            :: auxbas_rs_desc
      TYPE(realspace_grid_type), POINTER                 :: rs_v

      CALL timeset(routineN, handle)

      do_force = PRESENT(fderiv)
      fval = 0.0_dp
      dvol = rfun%pw%pw_grid%dvol

      NULLIFY (pw_env, auxbas_rs_desc)
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env=pw_env, auxbas_rs_desc=auxbas_rs_desc, &
                      auxbas_rs_grid=rs_v)
      CALL rs_grid_retain(rs_v)
      CALL rs_pw_transfer(rs_v, rfun%pw, pw2rs)

      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, cell=cell, &
                      dft_control=dft_control, particle_set=particle_set)
      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      nthread = 1
      ithread = 0
      ALLOCATE (hab(1, 1), pab(1, 1))

      DO ikind = 1, SIZE(atomic_kind_set)
         numexp = hirshfeld_env%kind_shape_fn(ikind)%numexp
         IF (numexp <= 0) CYCLE
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         ALLOCATE (cores(natom))

         DO iex = 1, numexp
            alpha = hirshfeld_env%kind_shape_fn(ikind)%zet(iex)
            coef = hirshfeld_env%kind_shape_fn(ikind)%coef(iex)
            npme = 0
            cores = 0
            DO iatom = 1, natom
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               IF (rs_v%desc%parallel .AND. .NOT. rs_v%desc%distributed) THEN
                  ! replicated realspace grid, split the atoms up between procs
                  IF (MODULO(iatom, rs_v%desc%group_size) == rs_v%desc%my_pos) THEN
                     npme = npme+1
                     cores(npme) = iatom
                  ENDIF
               ELSE
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            END DO

            DO j = 1, npme
               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               pab(1, 1) = coef
               hab(1, 1) = 0.0_dp
               force_a(:) = 0.0_dp
               force_b(:) = 0.0_dp
               !
               CALL integrate_pgf_product_rspace(0, alpha, 0, &
                                                 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, &
                                                 rs_v, cell, pw_env%cube_info(1), hab, pab=pab, o1=0, o2=0, &
                                                 eps_gvg_rspace=eps_rho_rspace, calculate_forces=do_force, &
                                                 force_a=force_a, force_b=force_b, use_virial=.FALSE., &
                                                 use_subpatch=.TRUE., subpatch_pattern=0_int_8)
               fval(atom_a) = fval(atom_a)+hab(1, 1)*dvol*coef
               IF (do_force) THEN
                  fderiv(:, atom_a) = fderiv(:, atom_a)+force_a(:)*dvol
               END IF
            END DO

         END DO
         DEALLOCATE (cores)

      END DO

      CALL rs_grid_release(rs_v)
      DEALLOCATE (hab, pab)

      CALL get_qs_env(qs_env=qs_env, para_env=para_env)
      CALL mp_sum(fval, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE hirshfeld_integration

END MODULE hirshfeld_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par Literature
!>      M. Krack, A. Gambirasio, and M. Parrinello,
!>      "Ab-initio x-ray scattering of liquid water",
!>      J. Chem. Phys. 117, 9409 (2002)
!> \author Matthias Krack
!> \date   30.11.2005
! **************************************************************************************************
MODULE xray_diffraction

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE bibliography,                    ONLY: Krack2002,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_gather,&
                                              mp_gatherv
   USE orbital_pointers,                ONLY: nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE physcon,                         ONLY: angstrom
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grids,                        ONLY: get_pw_grid_info
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: collocate_pgf_product_gspace
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xray_diffraction'

   PUBLIC :: calculate_rhotot_elec_gspace, &
             xray_diffraction_spectrum

CONTAINS

! **************************************************************************************************
!> \brief Calculate the coherent X-ray diffraction spectrum using the total
!>        electronic density in reciprocal space (g-space).
!> \param qs_env ...
!> \param unit_number ...
!> \param q_max ...
!> \date   30.11.2005
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE xray_diffraction_spectrum(qs_env, unit_number, q_max)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: unit_number
      REAL(KIND=dp), INTENT(IN)                          :: q_max

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xray_diffraction_spectrum', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 100

      INTEGER                                            :: group, handle, i, ig, ig_shell, ipe, &
                                                            ishell, jg, ng, npe, nshell, &
                                                            nshell_gather, source
      INTEGER(KIND=int_8)                                :: ngpts
      INTEGER, DIMENSION(3)                              :: npts
      INTEGER, DIMENSION(:), POINTER                     :: aux_index, ng_shell, ng_shell_gather, &
                                                            nshell_pe, offset_pe
      REAL(KIND=dp)                                      :: cutoff, f, f2, q, rho_hard, rho_soft, &
                                                            rho_total
      REAL(KIND=dp), DIMENSION(3)                        :: dg, dr
      REAL(KIND=dp), DIMENSION(:), POINTER :: f2sum, f2sum_gather, f4sum, f4sum_gather, fmax, &
         fmax_gather, fmin, fmin_gather, fsum, fsum_gather, gsq, q_shell, q_shell_gather
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rhotot_elec_gspace
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("xray_diffraction.F",120)

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set)
      NULLIFY (aux_index)
      NULLIFY (auxbas_pw_pool)
      NULLIFY (dft_control)
      NULLIFY (f2sum)
      NULLIFY (f2sum_gather)
      NULLIFY (f4sum)
      NULLIFY (f4sum_gather)
      NULLIFY (fmax)
      NULLIFY (fmax_gather)
      NULLIFY (fmin)
      NULLIFY (fmin_gather)
      NULLIFY (fsum)
      NULLIFY (fsum_gather)
      NULLIFY (gsq)
      NULLIFY (ng_shell)
      NULLIFY (ng_shell_gather)
      NULLIFY (nshell_pe)
      NULLIFY (offset_pe)
      NULLIFY (para_env)
      NULLIFY (particle_set)
      NULLIFY (pw_env)
      NULLIFY (q_shell)
      NULLIFY (q_shell_gather)
      NULLIFY (rho)
      NULLIFY (rho_atom_set)

      CALL cite_reference(Krack2002)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      rho=rho, &
                      rho_atom_set=rho_atom_set)

      CALL pw_env_get(pw_env=pw_env, &
                      auxbas_pw_pool=auxbas_pw_pool)

      group = para_env%group
      npe = para_env%num_pe
      source = para_env%source

      ! Plane waves grid to assemble the total electronic density

      CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                             pw=rhotot_elec_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_zero(rhotot_elec_gspace%pw)

      CALL get_pw_grid_info(pw_grid=rhotot_elec_gspace%pw%pw_grid, &
                            dr=dr, &
                            npts=npts, &
                            cutoff=cutoff, &
                            ngpts=ngpts, &
                            gsquare=gsq)

      dg(:) = twopi/(npts(:)*dr(:))

      ! Build the total electronic density in reciprocal space

      CALL calculate_rhotot_elec_gspace(qs_env=qs_env, &
                                        auxbas_pw_pool=auxbas_pw_pool, &
                                        rhotot_elec_gspace=rhotot_elec_gspace, &
                                        q_max=q_max, &
                                        rho_hard=rho_hard, &
                                        rho_soft=rho_soft)

      rho_total = rho_hard+rho_soft

      ! Calculate the coherent X-ray spectrum

      ! Now we have to gather the data from all processes, since each
      ! process has only worked his sub-grid

      ! Scan the g-vector shells

      CALL reallocate(q_shell, 1, nblock)
      CALL reallocate(ng_shell, 1, nblock)

      ng = SIZE(gsq)

      jg = 1
      nshell = 1
      q_shell(1) = SQRT(gsq(1))
      ng_shell(1) = 1

      DO ig = 2, ng
         IF(.NOT.(gsq(ig) >= gsq(jg)))CALL cp__a("xray_diffraction.F",215)
         IF (ABS(gsq(ig)-gsq(jg)) > 1.0E-12_dp) THEN
            nshell = nshell+1
            IF (nshell > SIZE(q_shell)) THEN
               CALL reallocate(q_shell, 1, SIZE(q_shell)+nblock)
               CALL reallocate(ng_shell, 1, SIZE(ng_shell)+nblock)
            END IF
            q = SQRT(gsq(ig))
            IF (q > q_max) THEN
               nshell = nshell-1
               EXIT
            END IF
            q_shell(nshell) = q
            ng_shell(nshell) = 1
            jg = ig
         ELSE
            ng_shell(nshell) = ng_shell(nshell)+1
         END IF
      END DO

      CALL reallocate(q_shell, 1, nshell)
      CALL reallocate(ng_shell, 1, nshell)
      CALL reallocate(fmin, 1, nshell)
      CALL reallocate(fmax, 1, nshell)
      CALL reallocate(fsum, 1, nshell)
      CALL reallocate(f2sum, 1, nshell)
      CALL reallocate(f4sum, 1, nshell)

      ig = 0
      DO ishell = 1, nshell
         fmin(ishell) = HUGE(0.0_dp)
         fmax(ishell) = 0.0_dp
         fsum(ishell) = 0.0_dp
         f2sum(ishell) = 0.0_dp
         f4sum(ishell) = 0.0_dp
         DO ig_shell = 1, ng_shell(ishell)
            f = ABS(rhotot_elec_gspace%pw%cc(ig+ig_shell))
            fmin(ishell) = MIN(fmin(ishell), f)
            fmax(ishell) = MAX(fmax(ishell), f)
            fsum(ishell) = fsum(ishell)+f
            f2 = f*f
            f2sum(ishell) = f2sum(ishell)+f2
            f4sum(ishell) = f4sum(ishell)+f2*f2
         END DO
         ig = ig+ng_shell(ishell)
      END DO

      CALL reallocate(nshell_pe, 0, npe-1)
      CALL reallocate(offset_pe, 0, npe-1)

      ! Root (source) process gathers the number of shell of each process

      CALL mp_gather(nshell, nshell_pe, source, group)

      ! Only the root process which has to print the full spectrum has to
      ! allocate here the receive buffers with their real sizes

      IF (unit_number > 0) THEN
         nshell_gather = SUM(nshell_pe)
         offset_pe(0) = 0
         DO ipe = 1, npe-1
            offset_pe(ipe) = offset_pe(ipe-1)+nshell_pe(ipe-1)
         END DO
      ELSE
         nshell_gather = 1 ! dummy value for the non-root processes
      END IF

      CALL reallocate(q_shell_gather, 1, nshell_gather)
      CALL reallocate(ng_shell_gather, 1, nshell_gather)
      CALL reallocate(fmin_gather, 1, nshell_gather)
      CALL reallocate(fmax_gather, 1, nshell_gather)
      CALL reallocate(fsum_gather, 1, nshell_gather)
      CALL reallocate(f2sum_gather, 1, nshell_gather)
      CALL reallocate(f4sum_gather, 1, nshell_gather)

      CALL mp_gatherv(q_shell, q_shell_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(ng_shell, ng_shell_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(fmax, fmax_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(fmin, fmin_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(fsum, fsum_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(f2sum, f2sum_gather, nshell_pe, offset_pe, source, group)
      CALL mp_gatherv(f4sum, f4sum_gather, nshell_pe, offset_pe, source, group)

      IF (ASSOCIATED(offset_pe)) THEN
         DEALLOCATE (offset_pe)
      END IF

      IF (ASSOCIATED(nshell_pe)) THEN
         DEALLOCATE (nshell_pe)
      END IF

      ! Print X-ray diffraction spectrum (I/O node only)

      IF (unit_number > 0) THEN

         CALL reallocate(aux_index, 1, nshell_gather)

         ! Sort the gathered shells

         CALL sort(q_shell_gather, nshell_gather, aux_index)

         ! Allocate final arrays of sufficient size, i.e. nshell_gather
         ! is always greater or equal the final nshell value

         CALL reallocate(q_shell, 1, nshell_gather)
         CALL reallocate(ng_shell, 1, nshell_gather)
         CALL reallocate(fmin, 1, nshell_gather)
         CALL reallocate(fmax, 1, nshell_gather)
         CALL reallocate(fsum, 1, nshell_gather)
         CALL reallocate(f2sum, 1, nshell_gather)
         CALL reallocate(f4sum, 1, nshell_gather)

         jg = 1
         nshell = 1
         q_shell(1) = q_shell_gather(1)
         i = aux_index(1)
         ng_shell(1) = ng_shell_gather(i)
         fmin(1) = fmin_gather(i)
         fmax(1) = fmax_gather(i)
         fsum(1) = fsum_gather(i)
         f2sum(1) = f2sum_gather(i)
         f4sum(1) = f4sum_gather(i)

         DO ig = 2, nshell_gather
            i = aux_index(ig)
            IF (ABS(q_shell_gather(ig)-q_shell_gather(jg)) > 1.0E-12_dp) THEN
               nshell = nshell+1
               q_shell(nshell) = q_shell_gather(ig)
               ng_shell(nshell) = ng_shell_gather(i)
               fmin(nshell) = fmin_gather(i)
               fmax(nshell) = fmax_gather(i)
               fsum(nshell) = fsum_gather(i)
               f2sum(nshell) = f2sum_gather(i)
               f4sum(nshell) = f4sum_gather(i)
               jg = ig
            ELSE
               ng_shell(nshell) = ng_shell(nshell)+ng_shell_gather(i)
               fmin(nshell) = MIN(fmin(nshell), fmin_gather(i))
               fmax(nshell) = MAX(fmax(nshell), fmax_gather(i))
               fsum(nshell) = fsum(nshell)+fsum_gather(i)
               f2sum(nshell) = f2sum(nshell)+f2sum_gather(i)
               f4sum(nshell) = f4sum(nshell)+f4sum_gather(i)
            END IF
         END DO

         ! The auxiliary index array is no longer needed now

         IF (ASSOCIATED(aux_index)) THEN
            DEALLOCATE (aux_index)
         END IF

         ! Allocate the final arrays for printing with their real size

         CALL reallocate(q_shell, 1, nshell)
         CALL reallocate(ng_shell, 1, nshell)
         CALL reallocate(fmin, 1, nshell)
         CALL reallocate(fmax, 1, nshell)
         CALL reallocate(fsum, 1, nshell)
         CALL reallocate(f2sum, 1, nshell)
         CALL reallocate(f4sum, 1, nshell)

         ! Write the X-ray diffraction spectrum to the specified file

         WRITE (UNIT=unit_number, FMT="(A)") &
            "#", &
            "# Coherent X-ray diffraction spectrum", &
            "#"
         WRITE (UNIT=unit_number, FMT="(A,1X,F20.10)") &
            "# Soft electronic charge (G-space) :", rho_soft, &
            "# Hard electronic charge (G-space) :", rho_hard, &
            "# Total electronic charge (G-space):", rho_total, &
            "# Density cutoff [Rydberg]         :", 2.0_dp*cutoff, &
            "# q(min) [1/Angstrom]              :", q_shell(2)/angstrom, &
            "# q(max) [1/Angstrom]              :", q_shell(nshell)/angstrom, &
            "# q(max) [1/Angstrom] (requested)  :", q_max/angstrom
         WRITE (UNIT=unit_number, FMT="(A,2X,I8)") &
            "# Number of g-vectors (grid points):", ngpts, &
            "# Number of g-vector shells        :", nshell
         WRITE (UNIT=unit_number, FMT="(A,3(1X,I6))") &
            "# Grid size (a,b,c)                :", npts(1:3)
         WRITE (UNIT=unit_number, FMT="(A,3F7.3)") &
            "# dg [1/Angstrom]                  :", dg(1:3)/angstrom, &
            "# dr [Angstrom]                    :", dr(1:3)*angstrom
         WRITE (UNIT=unit_number, FMT="(A)") &
            "#", &
            "# shell  points         q [1/A]      <|F(q)|^2>     Min(|F(q)|) "// &
            "    Max(|F(q)|)      <|F(q)|>^2      <|F(q)|^4>"

         DO ishell = 1, nshell
            WRITE (UNIT=unit_number, FMT="(T2,I6,2X,I6,5(1X,F15.6),1X,ES15.6)") &
               ishell, &
               ng_shell(ishell), &
               q_shell(ishell)/angstrom, &
               f2sum(ishell)/REAL(ng_shell(ishell), KIND=dp), &
               fmin(ishell), &
               fmax(ishell), &
               (fsum(ishell)/REAL(ng_shell(ishell), KIND=dp))**2, &
               f4sum(ishell)/REAL(ng_shell(ishell), KIND=dp)
         END DO

      END IF

      ! Release work storage

      IF (ASSOCIATED(fmin)) THEN
         DEALLOCATE (fmin)
      END IF

      IF (ASSOCIATED(fmax)) THEN
         DEALLOCATE (fmax)
      END IF

      IF (ASSOCIATED(fsum)) THEN
         DEALLOCATE (fsum)
      END IF

      IF (ASSOCIATED(f2sum)) THEN
         DEALLOCATE (f2sum)
      END IF

      IF (ASSOCIATED(f4sum)) THEN
         DEALLOCATE (f4sum)
      END IF

      IF (ASSOCIATED(ng_shell)) THEN
         DEALLOCATE (ng_shell)
      END IF

      IF (ASSOCIATED(q_shell)) THEN
         DEALLOCATE (q_shell)
      END IF

      IF (ASSOCIATED(fmin_gather)) THEN
         DEALLOCATE (fmin_gather)
      END IF

      IF (ASSOCIATED(fmax_gather)) THEN
         DEALLOCATE (fmax_gather)
      END IF

      IF (ASSOCIATED(fsum_gather)) THEN
         DEALLOCATE (fsum_gather)
      END IF

      IF (ASSOCIATED(f2sum_gather)) THEN
         DEALLOCATE (f2sum_gather)
      END IF

      IF (ASSOCIATED(f4sum_gather)) THEN
         DEALLOCATE (f4sum_gather)
      END IF

      IF (ASSOCIATED(ng_shell_gather)) THEN
         DEALLOCATE (ng_shell_gather)
      END IF

      IF (ASSOCIATED(q_shell_gather)) THEN
         DEALLOCATE (q_shell_gather)
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, &
                                rhotot_elec_gspace%pw)

      CALL timestop(handle)

   END SUBROUTINE xray_diffraction_spectrum

! **************************************************************************************************
!> \brief  The total electronic density in reciprocal space (g-space) is
!>         calculated.
!> \param qs_env ...
!> \param auxbas_pw_pool ...
!> \param rhotot_elec_gspace ...
!> \param q_max ...
!> \param rho_hard ...
!> \param rho_soft ...
!> \param fsign ...
!> \date   14.03.2008 (splitted from the routine xray_diffraction_spectrum)
!> \author Matthias Krack
!> \note   This code assumes that the g-vectors are ordered (in gsq and %cc)
! **************************************************************************************************
   SUBROUTINE calculate_rhotot_elec_gspace(qs_env, auxbas_pw_pool, &
                                           rhotot_elec_gspace, q_max, rho_hard, &
                                           rho_soft, fsign)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_p_type), INTENT(INOUT)                     :: rhotot_elec_gspace
      REAL(KIND=dp), INTENT(IN)                          :: q_max
      REAL(KIND=dp), INTENT(OUT)                         :: rho_hard, rho_soft
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: fsign

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_rhotot_elec_gspace', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom, handle, iatom, ico, ico1_pgf, ico1_set, ikind, ipgf, iset, iso, iso1_pgf, &
         iso1_set, ison, ispin, jco, jco1_pgf, jco1_set, jpgf, jset, jso, jso1_pgf, jso1_set, &
         json, la, lb, maxco, maxso, na, natom, nb, ncoa, ncob, ncotot, nkind, nsatbas, nset, &
         nsoa, nsob, nsotot, nspin
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, lmax, lmin, npgf, o2nindex
      LOGICAL                                            :: orthorhombic, paw_atom
      REAL(KIND=dp)                                      :: alpha, eps_rho_gspace, rho_total, scale, &
                                                            volume
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: delta_cpc, pab, work, zet
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(pw_p_type)                                    :: rho_elec_gspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: cpc_h, cpc_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("xray_diffraction.F",534)
      IF(.NOT.(ASSOCIATED(auxbas_pw_pool)))CALL cp__a("xray_diffraction.F",535)

      CALL timeset(routineN, handle)

      NULLIFY (atom_list)
      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind_set)
      NULLIFY (cell)
      NULLIFY (cpc_h)
      NULLIFY (cpc_s)
      NULLIFY (delta_cpc)
      NULLIFY (dft_control)
      NULLIFY (lmax)
      NULLIFY (lmin)
      NULLIFY (npgf)
      NULLIFY (orb_basis_set)
      NULLIFY (pab)
      NULLIFY (particle_set)
      NULLIFY (rho, rho_r)
      NULLIFY (rho_atom)
      NULLIFY (rho_atom_set)
      NULLIFY (work)
      NULLIFY (zet)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      rho=rho, &
                      rho_atom_set=rho_atom_set)

      CALL qs_rho_get(rho, rho_r=rho_r)
      eps_rho_gspace = dft_control%qs_control%eps_rho_gspace
      nkind = SIZE(atomic_kind_set)
      nspin = dft_control%nspins

      ! Load the soft contribution of the electronic density

      CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                             pw=rho_elec_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL pw_zero(rhotot_elec_gspace%pw)

      DO ispin = 1, nspin
         CALL pw_zero(rho_elec_gspace%pw)
         CALL pw_transfer(rho_r(ispin)%pw, rho_elec_gspace%pw, debug=.FALSE.)
         IF (PRESENT(fsign) .AND. (ispin == 2)) THEN
            alpha = fsign
         ELSE
            alpha = 1.0_dp
         END IF
         CALL pw_axpy(rho_elec_gspace%pw, rhotot_elec_gspace%pw, alpha=alpha)
      END DO

      ! Release the auxiliary PW grid for the calculation of the soft
      ! contribution

      CALL pw_pool_give_back_pw(pool=auxbas_pw_pool, &
                                pw=rho_elec_gspace%pw)

      rho_soft = pw_integrate_function(rhotot_elec_gspace%pw, isign=-1)

      CALL get_pw_grid_info(pw_grid=rhotot_elec_gspace%pw%pw_grid, &
                            vol=volume, &
                            orthorhombic=orthorhombic)
      IF(.NOT.(orthorhombic))CALL cp__a("xray_diffraction.F",604)

      CALL pw_scale(rhotot_elec_gspace%pw, volume)

      ! Add the hard contribution of the electronic density

      ! Each process has to loop over all PAW atoms, since the g-space grid
      ! is already distributed over all processes

      DO ikind = 1, nkind

         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, &
                              natom=natom)

         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis_set, &
                          paw_proj_set=paw_proj, &
                          paw_atom=paw_atom)

         IF (.NOT. paw_atom) CYCLE ! no PAW atom: nothing to do

         CALL get_paw_proj_set(paw_proj_set=paw_proj, o2nindex=o2nindex, nsatbas=nsatbas)

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                lmax=lmax, &
                                lmin=lmin, &
                                maxco=maxco, &
                                maxso=maxso, &
                                npgf=npgf, &
                                nset=nset, &
                                zet=zet)

         ncotot = maxco*nset
         nsotot = maxso*nset
         CALL reallocate(delta_cpc, 1, nsatbas, 1, nsatbas)
         CALL reallocate(pab, 1, ncotot, 1, ncotot)
         CALL reallocate(work, 1, maxso, 1, maxco)

         DO iatom = 1, natom

            atom = atom_list(iatom)
            rho_atom => rho_atom_set(atom)

            CALL get_rho_atom(rho_atom=rho_atom, &
                              cpc_h=cpc_h, &
                              cpc_s=cpc_s)

            ra(:) = pbc(particle_set(iatom)%r, cell)

            delta_cpc = 0.0_dp

            DO ispin = 1, nspin
               IF (PRESENT(fsign) .AND. (ispin == 2)) THEN
                  alpha = fsign
               ELSE
                  alpha = 1.0_dp
               END IF
               delta_cpc = delta_cpc+alpha*(cpc_h(ispin)%r_coef-cpc_s(ispin)%r_coef)
            END DO

            scale = 1.0_dp

            DO iset = 1, nset
               ico1_set = (iset-1)*maxco+1
               iso1_set = (iset-1)*maxso+1
               ncoa = ncoset(lmax(iset))
               nsoa = nsoset(lmax(iset))
               DO jset = 1, nset
                  jco1_set = (jset-1)*maxco+1
                  jso1_set = (jset-1)*maxso+1
                  ncob = ncoset(lmax(jset))
                  nsob = nsoset(lmax(jset))
                  DO ipgf = 1, npgf(iset)
                     ico1_pgf = ico1_set+(ipgf-1)*ncoa
                     iso1_pgf = iso1_set+(ipgf-1)*nsoa
                     DO jpgf = 1, npgf(jset)
                        jco1_pgf = jco1_set+(jpgf-1)*ncob
                        jso1_pgf = jso1_set+(jpgf-1)*nsob
                        ico = ico1_pgf+ncoset(lmin(iset)-1)
                        iso = iso1_pgf+nsoset(lmin(iset)-1)

                        ! Transformation spherical to Cartesian

                        DO la = lmin(iset), lmax(iset)
                           jco = jco1_pgf+ncoset(lmin(jset)-1)
                           jso = jso1_pgf+nsoset(lmin(jset)-1)
                           DO lb = lmin(jset), lmax(jset)
                              ison = o2nindex(iso)
                              json = o2nindex(jso)
                              CALL dgemm("N", "N", nso(la), nco(lb), nso(lb), 1.0_dp, &
                                         delta_cpc(ison, json), SIZE(delta_cpc, 1), &
                                         orbtramat(lb)%slm, nso(lb), 0.0_dp, work, &
                                         maxso)
                              CALL dgemm("T", "N", nco(la), nco(lb), nso(la), 1.0_dp, &
                                         orbtramat(la)%slm, nso(la), work, maxso, &
                                         0.0_dp, pab(ico, jco), SIZE(pab, 1))
                              jco = jco+nco(lb)
                              jso = jso+nso(lb)
                           END DO ! next lb
                           ico = ico+nco(la)
                           iso = iso+nso(la)
                        END DO ! next la

                        ! Collocate current product of primitive Cartesian functions

                        na = ico1_pgf-1
                        nb = jco1_pgf-1

                        CALL collocate_pgf_product_gspace( &
                           la_max=lmax(iset), &
                           zeta=zet(ipgf, iset), &
                           la_min=lmin(iset), &
                           lb_max=lmax(jset), &
                           zetb=zet(jpgf, jset), &
                           lb_min=lmin(jset), &
                           ra=ra, &
                           rab=(/0.0_dp, 0.0_dp, 0.0_dp/), &
                           rab2=0.0_dp, &
                           scale=scale, &
                           pab=pab, &
                           na=na, &
                           nb=nb, &
                           eps_rho_gspace=eps_rho_gspace, &
                           gsq_max=q_max*q_max, &
                           pw=rhotot_elec_gspace%pw)

                     END DO ! next primitive Gaussian function "jpgf"
                  END DO ! next primitive Gaussian function "ipgf"
               END DO ! next shell set "jset"
            END DO ! next shell set "iset"
         END DO ! next atom "iatom" of atomic kind "ikind"
      END DO ! next atomic kind "ikind"

      rho_total = pw_integrate_function(rhotot_elec_gspace%pw, isign=-1)/volume

      rho_hard = rho_total-rho_soft

      ! Release work storage

      IF (ASSOCIATED(delta_cpc)) THEN
         DEALLOCATE (delta_cpc)
      END IF

      IF (ASSOCIATED(work)) THEN
         DEALLOCATE (work)
      END IF

      IF (ASSOCIATED(pab)) THEN
         DEALLOCATE (pab)
      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_rhotot_elec_gspace

END MODULE xray_diffraction
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates hyperfine values
!> \par History
!>      created 04-2006 [RD]
!>      adapted 02-2007 [JGH]
!> \author R. Declerck (Reinout.Declerck@UGent.be)
! **************************************************************************************************
MODULE qs_epr_hyp
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE message_passing,                 ONLY: mp_sum,&
                                              mp_sync
   USE particle_types,                  ONLY: particle_type
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: a_bohr,&
                                              a_fine,&
                                              e_charge,&
                                              e_gfactor,&
                                              e_mass,&
                                              h_bar,&
                                              mu_perm
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_dr2_gg,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: qs_epr_hyp_calc

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_epr_hyp'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE qs_epr_hyp_calc(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_epr_hyp_calc', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: bo(2), group, ia, iat, iatom, idir1, &
                                                            idir2, ig, ikind, ir, iso, jatom, &
                                                            mepos, natom, natomkind, nkind, &
                                                            num_pe, output_unit, z
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: lsd, paw_atom
      REAL(dp)                                           :: arg, esum, hard_radius, hypanisotemp, &
                                                            hypfactor, int_radius, rab2, rtemp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: hypiso, hypiso_one
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: hypaniso
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(pw_p_type), POINTER                           :: hypaniso_gspace, rhototspin_elec_gspace
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: rho_rad_h, rho_rad_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(section_vals_type), POINTER                   :: dft_section

      NULLIFY (pw_env, cell, atomic_kind_set, qs_kind_set, auxbas_pw_pool, dft_control, &
               logger, dft_section, para_env, particle_set, rho, rho_atom, &
               rho_atom_set, rhototspin_elec_gspace, hypaniso_gspace, rho_g)

      logger => cp_get_default_logger()
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      output_unit = cp_print_key_unit_nr(logger, dft_section, &
                                         "PRINT%HYPERFINE_COUPLING_TENSOR", &
                                         extension=".eprhyp", log_filename=.FALSE.)
      CALL section_vals_val_get(dft_section, &
                                "PRINT%HYPERFINE_COUPLING_TENSOR%INTERACTION_RADIUS", &
                                r_val=int_radius)
      CALL section_vals_val_get(dft_section, "LSD", l_val=lsd)

      IF (.NOT. lsd) THEN
         ! EPR calculation only for LSD
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "Calculation of EPR hyperfine coupling tensors only for LSD"
         END IF
         NULLIFY (logger, dft_section)
         RETURN
      END IF

      hypfactor = -1.0_dp*mu_perm*e_charge*h_bar*e_gfactor/(2.0_dp*e_mass*a_bohr**3)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, &
                      rho=rho, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      rho_atom_set=rho_atom_set, pw_env=pw_env, &
                      particle_set=particle_set, para_env=para_env)
      group = para_env%group

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A,/,T2,A)") &
            "Calculation of EPR hyperfine coupling tensors", &
            REPEAT("-", 79)
      END IF

      ! allocate hyperfine matrices
      natom = SIZE(particle_set, 1)
      ALLOCATE (hypaniso(3, 3, natom))
      ALLOCATE (hypiso(natom))
      ALLOCATE (hypiso_one(natom))

      ! set the matrices to zero
      hypiso = 0.0_dp
      hypiso_one = 0.0_dp
      hypaniso = 0.0_dp

      nkind = SIZE(atomic_kind_set) ! nkind = number of atom types

      DO ikind = 1, nkind ! loop over atom types
         NULLIFY (atom_list, grid_atom, harmonics)
         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, natom=natomkind, z=z)

         CALL get_qs_kind(qs_kind_set(ikind), harmonics=harmonics, &
                          grid_atom=grid_atom, paw_atom=paw_atom, hard_radius=hard_radius)

         IF (.NOT. paw_atom) CYCLE ! skip the rest and go to next atom type

         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(natomkind, num_pe, mepos)

         DO iat = bo(1), bo(2) ! natomkind = # atoms for ikind
            iatom = atom_list(iat)
            rho_atom => rho_atom_set(iatom)
            NULLIFY (rho_rad_h, rho_rad_s)
            CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=rho_rad_h, &
                              rho_rad_s=rho_rad_s)
            ! Non-relativistic isotropic hyperfine value (hypiso_one)
            DO ia = 1, grid_atom%ng_sphere
               DO iso = 1, harmonics%max_iso_not0
                  hypiso_one(iatom) = hypiso_one(iatom)+ &
                                      (rho_rad_h(1)%r_coef(grid_atom%nr, iso)- &
                                       rho_rad_h(2)%r_coef(grid_atom%nr, iso))* &
                                      harmonics%slm(ia, iso)*grid_atom%wa(ia)/fourpi
               END DO
            END DO
            ! First calculate hard-soft contributions for the own nucleus
            ! + scalar relativistic isotropic hyperfine value (hypiso)
            DO ir = 1, grid_atom%nr
               IF (grid_atom%rad(ir) <= hard_radius) THEN
                  DO ia = 1, grid_atom%ng_sphere
                     hypanisotemp = 0.0_dp
                     DO iso = 1, harmonics%max_iso_not0
                        hypiso(iatom) = hypiso(iatom)+ &
                                        (rho_rad_h(1)%r_coef(ir, iso)-rho_rad_h(2)%r_coef(ir, iso))* &
                                        harmonics%slm(ia, iso)*grid_atom%wr(ir)*grid_atom%wa(ia)* &
                                        2._dp/(REAL(z, KIND=dp)*a_fine**2* &
                                               (1._dp+2._dp*grid_atom%rad(ir)/(REAL(z, KIND=dp)*a_fine**2))**2* &
                                               fourpi*grid_atom%rad(ir)**2)
                        hypanisotemp = hypanisotemp+ &
                                       (rho_rad_h(1)%r_coef(ir, iso)-rho_rad_h(2)%r_coef(ir, iso) &
                                        -(rho_rad_s(1)%r_coef(ir, iso)-rho_rad_s(2)%r_coef(ir, iso)))* &
                                       harmonics%slm(ia, iso)*grid_atom%wr(ir)*grid_atom%wa(ia)/ &
                                       grid_atom%rad(ir)**3
                     END DO ! iso
                     hypaniso(1, 1, iatom) = hypaniso(1, 1, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia)* &
                                              grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia)-1._dp)
                     hypaniso(1, 2, iatom) = hypaniso(1, 2, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia)* &
                                              grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia)-0._dp)
                     hypaniso(1, 3, iatom) = hypaniso(1, 3, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia)* &
                                              grid_atom%cos_pol(ia)-0._dp)
                     hypaniso(2, 2, iatom) = hypaniso(2, 2, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia)* &
                                              grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia)-1._dp)
                     hypaniso(2, 3, iatom) = hypaniso(2, 3, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia)* &
                                              grid_atom%cos_pol(ia)-0._dp)
                     hypaniso(3, 3, iatom) = hypaniso(3, 3, iatom)+hypanisotemp* &
                                             (3._dp*grid_atom%cos_pol(ia)* &
                                              grid_atom%cos_pol(ia)-1._dp)
                  END DO ! ia
               END IF ! hard_radius
            END DO ! ir

            ! Now calculate hard-soft anisotropic contributions for the other nuclei
            DO jatom = 1, natom
               IF (jatom .EQ. iatom) CYCLE ! iatom already done
               rab = pbc(particle_set(iatom)%r, particle_set(jatom)%r, cell)
               rab2 = DOT_PRODUCT(rab, rab)
               IF (SQRT(rab2) <= int_radius) THEN
                  DO ir = 1, grid_atom%nr
                     IF (grid_atom%rad(ir) <= hard_radius) THEN
                        DO ia = 1, grid_atom%ng_sphere
                           hypanisotemp = 0.0_dp
                           rtemp = SQRT(rab2+grid_atom%rad(ir)**2+2.0_dp*grid_atom%rad(ir)* &
                                        (rab(1)*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia)+ &
                                         rab(2)*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia)+ &
                                         rab(3)*grid_atom%cos_pol(ia)))
                           DO iso = 1, harmonics%max_iso_not0
                              hypanisotemp = hypanisotemp+ &
                                             (rho_rad_h(1)%r_coef(ir, iso)-rho_rad_h(2)%r_coef(ir, iso) &
                                              -(rho_rad_s(1)%r_coef(ir, iso)-rho_rad_s(2)%r_coef(ir, iso)))* &
                                             harmonics%slm(ia, iso)*grid_atom%wr(ir)*grid_atom%wa(ia)/ &
                                             rtemp**5
                           END DO ! iso
                           hypaniso(1, 1, jatom) = hypaniso(1, 1, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(1)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia))* &
                                                    (rab(1)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia))-rtemp**2)
                           hypaniso(1, 2, jatom) = hypaniso(1, 2, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(1)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia))* &
                                                    (rab(2)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia))-0._dp)
                           hypaniso(1, 3, jatom) = hypaniso(1, 3, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(1)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%cos_azi(ia))* &
                                                    (rab(3)+grid_atom%rad(ir)*grid_atom%cos_pol(ia))-0._dp)
                           hypaniso(2, 2, jatom) = hypaniso(2, 2, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(2)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia))* &
                                                    (rab(2)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia))-rtemp**2)
                           hypaniso(2, 3, jatom) = hypaniso(2, 3, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(2)+grid_atom%rad(ir)*grid_atom%sin_pol(ia)*grid_atom%sin_azi(ia))* &
                                                    (rab(3)+grid_atom%rad(ir)*grid_atom%cos_pol(ia))-0._dp)
                           hypaniso(3, 3, jatom) = hypaniso(3, 3, jatom)+hypanisotemp* &
                                                   (3._dp*(rab(3)+grid_atom%rad(ir)*grid_atom%cos_pol(ia))* &
                                                    (rab(3)+grid_atom%rad(ir)*grid_atom%cos_pol(ia))-rtemp**2)
                        END DO ! ia
                     END IF ! hard_radius
                  END DO ! ir
               END IF ! rab2
            END DO ! jatom
         END DO ! iat
      END DO ! ikind

      ! Now calculate the soft electronic spin density in reciprocal space (g-space)
      ! Plane waves grid to assemble the soft electronic spin density
      CALL pw_env_get(pw_env=pw_env, &
                      auxbas_pw_pool=auxbas_pw_pool)

      ALLOCATE (rhototspin_elec_gspace)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             rhototspin_elec_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_zero(rhototspin_elec_gspace%pw)

      pw_grid => rhototspin_elec_gspace%pw%pw_grid

      ! Load the contribution of the soft electronic density
      CALL qs_rho_get(rho, rho_g=rho_g)
      IF(.NOT.(SIZE(rho_g) > 1))CALL cp__a("qs_epr_hyp.F",301)
      CALL pw_axpy(rho_g(1)%pw, rhototspin_elec_gspace%pw)
      CALL pw_axpy(rho_g(2)%pw, rhototspin_elec_gspace%pw, alpha=-1._dp)
      ! grid to assemble anisotropic hyperfine terms
      ALLOCATE (hypaniso_gspace)
      CALL pw_pool_create_pw(auxbas_pw_pool, &
                             hypaniso_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      DO idir1 = 1, 3
         DO idir2 = idir1, 3 ! tensor symmetry
            CALL pw_zero(hypaniso_gspace%pw)
            CALL pw_dr2_gg(rhototspin_elec_gspace%pw, hypaniso_gspace%pw, &
                           idir1, idir2)
            DO iatom = 1, natom
               esum = 0.0_dp
               ra(:) = pbc(particle_set(iatom)%r, cell)
               DO ig = 1, SIZE(hypaniso_gspace%pw%cc)
                  arg = DOT_PRODUCT(pw_grid%g(:, ig), ra)
                  esum = esum+COS(arg)*REAL(hypaniso_gspace%pw%cc(ig), dp) &
                         -SIN(arg)*AIMAG(hypaniso_gspace%pw%cc(ig))
               END DO
               ! Actually, we need -1.0 * fourpi * hypaniso_gspace
               esum = esum*fourpi*(-1.0_dp)
               hypaniso(idir1, idir2, iatom) = hypaniso(idir1, idir2, iatom)+esum
            END DO
         END DO ! idir2
      END DO ! idir1

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rhototspin_elec_gspace%pw)
      DEALLOCATE (rhototspin_elec_gspace)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, hypaniso_gspace%pw)
      DEALLOCATE (hypaniso_gspace)

      ! Multiply hyperfine matrices with constant*gyromagnetic ratio's
      ! to have it in units of Mhz.

      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              z=z)
         hypiso(iatom) = hypiso(iatom)* &
                         2.0_dp/3.0_dp*hypfactor*ptable(z)%gyrom_ratio
         hypiso_one(iatom) = hypiso_one(iatom)* &
                             2.0_dp/3.0_dp*hypfactor*ptable(z)%gyrom_ratio
         DO idir1 = 1, 3
            DO idir2 = idir1, 3
               hypaniso(idir1, idir2, iatom) = hypaniso(idir1, idir2, iatom)* &
                                               hypfactor/fourpi*ptable(z)%gyrom_ratio
               IF (idir1 /= idir2) THEN
                  hypaniso(idir2, idir1, iatom) = hypaniso(idir1, idir2, iatom)
               END IF
            END DO
         END DO
      END DO

      ! Global sum
      CALL mp_sync(group)
      CALL mp_sum(hypaniso, group)
      CALL mp_sum(hypiso, group)
      CALL mp_sum(hypiso_one, group)

      ! Print hyperfine matrices
      IF (output_unit > 0) THEN
         DO iatom = 1, natom
            CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                                 element_symbol=element_symbol, z=z)
            WRITE (UNIT=output_unit, FMT="(T1,I5,T7,A,T10,I3,T14,F16.10,T31,A,T60,F20.10)") &
               iatom, element_symbol, ptable(z)%gyrom_ratio_isotope, ptable(z)%gyrom_ratio, &
               "[Mhz/T]  Sca-Rel A_iso [Mhz]", hypiso(iatom)
            WRITE (UNIT=output_unit, FMT="(T31,A,T60,F20.10)") &
               "         Non-Rel A_iso [Mhz]", hypiso_one(iatom)
            WRITE (UNIT=output_unit, FMT="(T4,A,T18,F20.10,1X,F20.10,1X,F20.10)") &
               "             ", hypaniso(1, 1, iatom), hypaniso(1, 2, iatom), hypaniso(1, 3, iatom), &
               "  A_ani [Mhz]", hypaniso(2, 1, iatom), hypaniso(2, 2, iatom), hypaniso(2, 3, iatom), &
               "             ", hypaniso(3, 1, iatom), hypaniso(3, 2, iatom), hypaniso(3, 3, iatom)
         ENDDO
      END IF

      ! Deallocate the remainings ...
      DEALLOCATE (hypiso)
      DEALLOCATE (hypiso_one)
      DEALLOCATE (hypaniso)

   END SUBROUTINE qs_epr_hyp_calc

END MODULE qs_epr_hyp

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Provide various population analyses and print the requested output
!>        information
!>
!> \author  Matthias Krack (MK)
!> \date    09.07.2010
!> \version 1.0
! **************************************************************************************************

MODULE population_analyses
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_deallocate_matrix, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, &
        cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, cp_dbcsr_p_type, cp_dbcsr_set, &
        cp_dbcsr_setname, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix,&
                                              write_fm_with_basis_info
   USE cp_fm_diag,                      ONLY: cp_fm_power
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_diag,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: nso
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'population_analyses'

   PUBLIC :: lowdin_population_analysis, &
             mulliken_population_analysis

CONTAINS

! **************************************************************************************************
!> \brief Perform a Lowdin population analysis based on a symmetric
!>        orthogonalisation of the density matrix using S^(1/2)
!>
!> \param qs_env ...
!> \param output_unit ...
!> \param print_level ...
!> \date    06.07.2010
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE lowdin_population_analysis(qs_env, output_unit, print_level)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: output_unit, print_level

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lowdin_population_analysis', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: headline
      INTEGER                                            :: handle, ispin, ndep, nsgf, nspin
      LOGICAL                                            :: print_gop
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: orbpop
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrixkp_p, matrixkp_s
      TYPE(cp_dbcsr_type), POINTER                       :: sm_p, sm_s
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct
      TYPE(cp_fm_type), POINTER                          :: fm_s_half, fm_work1, fm_work2
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind_set)
      NULLIFY (fmstruct)
      NULLIFY (fm_s_half)
      NULLIFY (fm_work1)
      NULLIFY (fm_work2)
      NULLIFY (matrix_p)
      NULLIFY (matrixkp_p)
      NULLIFY (matrixkp_s)
      NULLIFY (orbpop)
      NULLIFY (particle_set)
      NULLIFY (rho)
      NULLIFY (scf_control)
      NULLIFY (sm_p)
      NULLIFY (sm_s)
      NULLIFY (orbpop)
      NULLIFY (para_env)
      NULLIFY (blacs_env)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      matrix_s_kp=matrixkp_s, &
                      particle_set=particle_set, &
                      rho=rho, &
                      scf_control=scf_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      IF(.NOT.(ASSOCIATED(atomic_kind_set)))CALL cp__a("population_analyses.F",135)
      IF(.NOT.(ASSOCIATED(qs_kind_set)))CALL cp__a("population_analyses.F",136)
      IF(.NOT.(ASSOCIATED(matrixkp_s)))CALL cp__a("population_analyses.F",137)
      IF(.NOT.(ASSOCIATED(particle_set)))CALL cp__a("population_analyses.F",138)
      IF(.NOT.(ASSOCIATED(rho)))CALL cp__a("population_analyses.F",139)
      IF(.NOT.(ASSOCIATED(scf_control)))CALL cp__a("population_analyses.F",140)

      IF (SIZE(matrixkp_s, 2) > 1) THEN

         CALL cp__w("population_analyses.F",144,"Lowdin population analysis not implemented for k-points.")

      ELSE

         sm_s => matrixkp_s(1, 1)%matrix ! Overlap matrix in sparse format
         CALL qs_rho_get(rho, rho_ao_kp=matrixkp_p) ! Density matrices in sparse format

         matrix_p => matrixkp_p(:, 1)
         nspin = SIZE(matrix_p, 1)

         ! Get the total number of contracted spherical Gaussian basis functions
         CALL get_qs_kind_set(qs_kind_set, nsgf=nsgf)

         ! Provide an array to store the orbital populations for each spin
         ALLOCATE (orbpop(nsgf, nspin))
         orbpop(:, :) = 0.0_dp

         ! Write headline
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") "LOWDIN POPULATION ANALYSIS"
         END IF

         ! Provide full size work matrices
         CALL cp_fm_struct_create(fmstruct=fmstruct, &
                                  para_env=para_env, &
                                  context=blacs_env, &
                                  nrow_global=nsgf, &
                                  ncol_global=nsgf)
         CALL cp_fm_create(matrix=fm_s_half, &
                           matrix_struct=fmstruct, &
                           name="S^(1/2) MATRIX")
         CALL cp_fm_create(matrix=fm_work1, &
                           matrix_struct=fmstruct, &
                           name="FULL WORK MATRIX 1")
         headline = "SYMMETRICALLY ORTHOGONALISED DENSITY MATRIX"
         CALL cp_fm_create(matrix=fm_work2, &
                           matrix_struct=fmstruct, &
                           name=TRIM(headline))
         CALL cp_fm_struct_release(fmstruct=fmstruct)

         ! Build full S^(1/2) matrix (computationally expensive)
         CALL copy_dbcsr_to_fm(sm_s, fm_s_half)
         CALL cp_fm_power(fm_s_half, fm_work1, 0.5_dp, scf_control%eps_eigval, ndep)
         IF (ndep /= 0) &
            CALL cp_warn(cp__l("population_analyses.F",188), &
                         "Overlap matrix exhibits linear dependencies. At least some "// &
                         "eigenvalues have been quenched.")

         ! Build Lowdin population matrix for each spin
         DO ispin = 1, nspin
            sm_p => matrix_p(ispin)%matrix ! Density matrix for spin ispin in sparse format
            ! Calculate S^(1/2)*P*S^(1/2) as a full matrix (Lowdin)
            CALL cp_dbcsr_sm_fm_multiply(sm_p, fm_s_half, fm_work1, nsgf)
            CALL cp_gemm(transa="N", &
                         transb="N", &
                         m=nsgf, &
                         n=nsgf, &
                         k=nsgf, &
                         alpha=1.0_dp, &
                         matrix_a=fm_s_half, &
                         matrix_b=fm_work1, &
                         beta=0.0_dp, &
                         matrix_c=fm_work2)
            IF (print_level > 2) THEN
               ! Write the full Lowdin population matrix
               IF (nspin > 1) THEN
                  IF (ispin == 1) THEN
                     fm_work2%name = TRIM(headline)//" FOR ALPHA SPIN"
                  ELSE
                     fm_work2%name = TRIM(headline)//" FOR BETA SPIN"
                  END IF
               END IF
               CALL write_fm_with_basis_info(fm_work2, 4, 6, qs_env, para_env, &
                                             output_unit=output_unit)
            END IF
            CALL cp_fm_get_diag(fm_work2, orbpop(:, ispin))
         END DO ! next spin ispin

         ! Write atomic populations and charges
         IF (output_unit > 0) THEN
            print_gop = (print_level > 1) ! Print also orbital populations
            CALL write_orbpop(orbpop, atomic_kind_set, qs_kind_set, particle_set, output_unit, print_gop)
         END IF

         ! Release local working storage
         CALL cp_fm_release(matrix=fm_s_half)
         CALL cp_fm_release(matrix=fm_work1)
         CALL cp_fm_release(matrix=fm_work2)
         IF (ASSOCIATED(orbpop)) THEN
            DEALLOCATE (orbpop)
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE lowdin_population_analysis

! **************************************************************************************************
!> \brief Perform a Mulliken population analysis
!>
!> \param qs_env ...
!> \param output_unit ...
!> \param print_level ...
!> \date    10.07.2010
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE mulliken_population_analysis(qs_env, output_unit, print_level)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: output_unit, print_level

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mulliken_population_analysis', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: headline
      INTEGER                                            :: blk, handle, iatom, ic, isgf, ispin, &
                                                            jatom, jsgf, natom, nsgf, nspin, sgfa, &
                                                            sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf_atom
      LOGICAL                                            :: found, print_gop
      REAL(KIND=dp)                                      :: ps
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: orbpop, p_block, ps_block, s_block
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: sm_p, sm_ps, sm_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind_set)
      NULLIFY (matrix_p)
      NULLIFY (matrix_s)
      NULLIFY (orbpop)
      NULLIFY (particle_set)
      NULLIFY (ps_block)
      NULLIFY (p_block)
      NULLIFY (rho)
      NULLIFY (sm_p)
      NULLIFY (sm_ps)
      NULLIFY (sm_s)
      NULLIFY (s_block)
      NULLIFY (para_env)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      matrix_s_kp=matrix_s, &
                      particle_set=particle_set, &
                      rho=rho, &
                      para_env=para_env)

      IF(.NOT.(ASSOCIATED(atomic_kind_set)))CALL cp__a("population_analyses.F",302)
      IF(.NOT.(ASSOCIATED(qs_kind_set)))CALL cp__a("population_analyses.F",303)
      IF(.NOT.(ASSOCIATED(particle_set)))CALL cp__a("population_analyses.F",304)
      IF(.NOT.(ASSOCIATED(rho)))CALL cp__a("population_analyses.F",305)
      IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("population_analyses.F",306)

      CALL qs_rho_get(rho, rho_ao_kp=matrix_p) ! Density matrices in sparse format
      nspin = SIZE(matrix_p, 1)

      ! Get the total number of contracted spherical Gaussian basis functions
      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, nsgf=nsgf)
      ALLOCATE (first_sgf_atom(natom))
      first_sgf_atom(:) = 0

      CALL get_particle_set(particle_set, qs_kind_set, first_sgf=first_sgf_atom)

      ! Provide an array to store the orbital populations for each spin
      ALLOCATE (orbpop(nsgf, nspin))
      orbpop(:, :) = 0.0_dp

      ! Write headline
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            '!-----------------------------------------------------------------------------!'
         WRITE (UNIT=output_unit, FMT="(T22,A)") "Mulliken Population Analysis"
      END IF

      ! Create a DBCSR work matrix, if needed
      IF (print_level > 2) THEN
         sm_s => matrix_s(1, 1)%matrix ! Overlap matrix in sparse format
         ALLOCATE (sm_ps)
         headline = "MULLIKEN NET ATOMIC ORBITAL AND OVERLAP POPULATION MATRIX"
         CALL cp_dbcsr_init(sm_ps)
         CALL cp_dbcsr_copy(matrix_b=sm_ps, &
                            matrix_a=sm_s, &
                            name=TRIM(headline))
      END IF

      ! Build Mulliken population matrix for each spin
      DO ispin = 1, nspin
         DO ic = 1, SIZE(matrix_s, 2)
            IF (print_level > 2) THEN
               CALL cp_dbcsr_set(sm_ps, 0.0_dp)
            END IF
            sm_s => matrix_s(1, ic)%matrix ! Overlap matrix in sparse format
            sm_p => matrix_p(ispin, ic)%matrix ! Density matrix for spin ispin in sparse format
            ! Calculate Hadamard product of P and S as sparse matrix (Mulliken)
            ! CALL cp_dbcsr_hadamard_product(sm_p,sm_s,sm_ps)
            CALL cp_dbcsr_iterator_start(iter, sm_s)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, s_block, blk)
               IF (.NOT. (ASSOCIATED(s_block))) CYCLE
               CALL cp_dbcsr_get_block_p(matrix=sm_p, &
                                         row=iatom, &
                                         col=jatom, &
                                         block=p_block, &
                                         found=found)
               IF (print_level > 2) THEN
                  CALL cp_dbcsr_get_block_p(matrix=sm_ps, &
                                            row=iatom, &
                                            col=jatom, &
                                            block=ps_block, &
                                            found=found)
                  IF(.NOT.(ASSOCIATED(ps_block)))CALL cp__a("population_analyses.F",366)
               END IF

               sgfb = first_sgf_atom(jatom)
               DO jsgf = 1, SIZE(s_block, 2)
                  DO isgf = 1, SIZE(s_block, 1)
                     ps = p_block(isgf, jsgf)*s_block(isgf, jsgf)
                     IF (ASSOCIATED(ps_block)) ps_block(isgf, jsgf) = ps_block(isgf, jsgf)+ps
                     orbpop(sgfb, ispin) = orbpop(sgfb, ispin)+ps
                  END DO
                  sgfb = sgfb+1
               END DO
               IF (iatom /= jatom) THEN
                  sgfa = first_sgf_atom(iatom)
                  DO isgf = 1, SIZE(s_block, 1)
                     DO jsgf = 1, SIZE(s_block, 2)
                        ps = p_block(isgf, jsgf)*s_block(isgf, jsgf)
                        orbpop(sgfa, ispin) = orbpop(sgfa, ispin)+ps
                     END DO
                     sgfa = sgfa+1
                  END DO
               END IF
            END DO
            CALL cp_dbcsr_iterator_stop(iter)
         END DO

         IF (print_level > 2) THEN
            ! Write the full Mulliken net AO and overlap population matrix
            IF (nspin > 1) THEN
               IF (ispin == 1) THEN
                  CALL cp_dbcsr_setname(sm_ps, TRIM(headline)//" For Alpha Spin")
               ELSE
                  CALL cp_dbcsr_setname(sm_ps, TRIM(headline)//" For Beta Spin")
               END IF
            END IF
            CALL cp_dbcsr_write_sparse_matrix(sm_ps, 4, 6, qs_env, para_env, &
                                              output_unit=output_unit)
         END IF
      END DO

      CALL mp_sum(orbpop, para_env%group)

      ! Write atomic populations and charges
      IF (output_unit > 0) THEN
         print_gop = (print_level > 1) ! Print also orbital populations
         CALL write_orbpop(orbpop, atomic_kind_set, qs_kind_set, particle_set, output_unit, print_gop)
      END IF

      ! Release local working storage
      IF (ASSOCIATED(sm_ps)) CALL cp_dbcsr_deallocate_matrix(sm_ps)
      IF (ASSOCIATED(orbpop)) THEN
         DEALLOCATE (orbpop)
      END IF
      IF (ALLOCATED(first_sgf_atom)) THEN
         DEALLOCATE (first_sgf_atom)
      END IF

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T2,A)") &
            '!-----------------------------------------------------------------------------!'
      END IF

      CALL timestop(handle)

   END SUBROUTINE mulliken_population_analysis

! **************************************************************************************************
!> \brief Write atomic orbital populations and net atomic charges
!>
!> \param orbpop ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param output_unit ...
!> \param print_orbital_contributions ...
!> \date    07.07.2010
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_orbpop(orbpop, atomic_kind_set, qs_kind_set, particle_set, output_unit, &
                           print_orbital_contributions)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: orbpop
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: print_orbital_contributions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_orbpop', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=6), DIMENSION(:), POINTER            :: sgf_symbol
      INTEGER                                            :: handle, iao, iatom, ikind, iset, isgf, &
                                                            ishell, iso, l, natom, nset, nsgf, &
                                                            nspin
      INTEGER, DIMENSION(:), POINTER                     :: nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: lshell
      REAL(KIND=dp)                                      :: zeff
      REAL(KIND=dp), DIMENSION(3)                        :: sumorbpop, totsumorbpop
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      CALL timeset(routineN, handle)

      NULLIFY (lshell)
      NULLIFY (nshell)
      NULLIFY (orb_basis_set)
      NULLIFY (sgf_symbol)

      IF(.NOT.(ASSOCIATED(orbpop)))CALL cp__a("population_analyses.F",475)
      IF(.NOT.(ASSOCIATED(atomic_kind_set)))CALL cp__a("population_analyses.F",476)
      IF(.NOT.(ASSOCIATED(particle_set)))CALL cp__a("population_analyses.F",477)

      nspin = SIZE(orbpop, 2)

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)
      CALL get_qs_kind_set(qs_kind_set, nsgf=nsgf)

      ! Select and write headline
      IF (nspin == 1) THEN
         IF (print_orbital_contributions) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "# Orbital  AO symbol  Orbital population                            Net charge"
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "#  Atom  Element  Kind  Atomic population                           Net charge"
         END IF
      ELSE
         IF (print_orbital_contributions) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "# Orbital  AO symbol  Orbital population (alpha,beta)  Net charge  Spin moment"
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "#  Atom  Element  Kind  Atomic population (alpha,beta) Net charge  Spin moment"
         END IF
      END IF

      totsumorbpop(:) = 0.0_dp

      iao = 1
      DO iatom = 1, natom
         sumorbpop(:) = 0.0_dp
         NULLIFY (orb_basis_set)
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              element_symbol=element_symbol, &
                              kind_number=ikind)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, zeff=zeff)
         IF (ASSOCIATED(orb_basis_set)) THEN
            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                   nset=nset, &
                                   nshell=nshell, &
                                   l=lshell, &
                                   sgf_symbol=sgf_symbol)
            isgf = 1
            DO iset = 1, nset
               DO ishell = 1, nshell(iset)
                  l = lshell(ishell, iset)
                  DO iso = 1, nso(l)
                     IF (nspin == 1) THEN
                        sumorbpop(1) = sumorbpop(1)+orbpop(iao, 1)
                        IF (print_orbital_contributions) THEN
                           IF (isgf == 1) WRITE (UNIT=output_unit, FMT="(A)") ""
                           WRITE (UNIT=output_unit, &
                                  FMT="(T2,I9,2X,A2,1X,A,T30,F12.6)") &
                              iao, element_symbol, sgf_symbol(isgf), orbpop(iao, 1)
                        END IF
                     ELSE
                        sumorbpop(1:2) = sumorbpop(1:2)+orbpop(iao, 1:2)
                        sumorbpop(3) = sumorbpop(3)+orbpop(iao, 1)-orbpop(iao, 2)
                        IF (print_orbital_contributions) THEN
                           IF (isgf == 1) WRITE (UNIT=output_unit, FMT="(A)") ""
                           WRITE (UNIT=output_unit, &
                                  FMT="(T2,I9,2X,A2,1X,A,T29,2(1X,F12.6),T68,F12.6)") &
                              iao, element_symbol, sgf_symbol(isgf), orbpop(iao, 1:2), &
                              orbpop(iao, 1)-orbpop(iao, 2)
                        END IF
                     END IF
                     isgf = isgf+1
                     iao = iao+1
                  END DO
               END DO
            END DO
            IF (nspin == 1) THEN
               totsumorbpop(1) = totsumorbpop(1)+sumorbpop(1)
               totsumorbpop(3) = totsumorbpop(3)+zeff-sumorbpop(1)
               WRITE (UNIT=output_unit, &
                      FMT="(T2,I7,5X,A2,2X,I6,T30,F12.6,T68,F12.6)") &
                  iatom, element_symbol, ikind, sumorbpop(1), zeff-sumorbpop(1)
            ELSE
               totsumorbpop(1:2) = totsumorbpop(1:2)+sumorbpop(1:2)
               totsumorbpop(3) = totsumorbpop(3)+zeff-sumorbpop(1)-sumorbpop(2)
               WRITE (UNIT=output_unit, &
                      FMT="(T2,I7,5X,A2,2X,I6,T28,4(1X,F12.6))") &
                  iatom, element_symbol, ikind, sumorbpop(1:2), &
                  zeff-sumorbpop(1)-sumorbpop(2), sumorbpop(3)
            END IF
         END IF ! atom has an orbital basis
      END DO ! next atom iatom

      ! Write total sums
      IF (print_orbital_contributions) WRITE (UNIT=output_unit, FMT="(A)") ""
      IF (nspin == 1) THEN
         WRITE (UNIT=output_unit, &
                FMT="(T2,A,T42,F12.6,T68,F12.6,/)") &
            "# Total charge", totsumorbpop(1), totsumorbpop(3)
      ELSE
         WRITE (UNIT=output_unit, &
                FMT="(T2,A,T28,4(1X,F12.6),/)") &
            "# Total charge and spin", totsumorbpop(1:2), totsumorbpop(3), &
            totsumorbpop(1)-totsumorbpop(2)
      END IF

      IF (output_unit > 0) CALL m_flush(output_unit)

      CALL timestop(handle)

   END SUBROUTINE write_orbpop

END MODULE population_analyses
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the calculation of molecular states
!> \author CJM
! **************************************************************************************************
MODULE molecular_states
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE bibliography,                    ONLY: Hunt2003,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_element,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_size
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_maxloc
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'molecular_states'

   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

! *** Public subroutines ***

   PUBLIC :: construct_molecular_states

CONTAINS

! **************************************************************************************************
!> \brief constructs molecular states. mo_localized gets overwritten!
!> \param molecule_set ...
!> \param mo_localized ...
!> \param mo_coeff ...
!> \param mo_eigenvalues ...
!> \param Hks ...
!> \param matrix_S ...
!> \param qs_env ...
!> \param wf_r ...
!> \param wf_g ...
!> \param loc_print_section ...
!> \param particles ...
!> \param tag ...
!> \param marked_states ...
! **************************************************************************************************
   SUBROUTINE construct_molecular_states(molecule_set, mo_localized, &
                                         mo_coeff, mo_eigenvalues, Hks, matrix_S, qs_env, wf_r, wf_g, &
                                         loc_print_section, particles, tag, marked_states)

      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(cp_fm_type), POINTER                          :: mo_localized, mo_coeff
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_dbcsr_type), POINTER                       :: Hks, matrix_S
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type), INTENT(INOUT)                     :: wf_r, wf_g
      TYPE(section_vals_type), POINTER                   :: loc_print_section
      TYPE(particle_list_type), POINTER                  :: particles
      CHARACTER(LEN=*), INTENT(IN)                       :: tag
      INTEGER, DIMENSION(:), POINTER                     :: marked_states

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_molecular_states', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: title
      INTEGER                                            :: handle, i, imol, iproc, isos, k, n_rep, &
                                                            ncol_global, nproc, nrow_global, ns, &
                                                            output_unit, unit_nr, unit_report
      INTEGER, DIMENSION(:), POINTER                     :: ind, mark_list
      INTEGER, DIMENSION(:, :), POINTER                  :: mark_states
      INTEGER, POINTER                                   :: nstates(:), states(:)
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: tmp
      REAL(KIND=dp), ALLOCATABLE                         :: evals(:)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eval_range
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: b, c, d, D_igk, e_vectors, &
                                                            rot_e_vectors, smo, storage
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL cite_reference(Hunt2003)

      CALL timeset(routineN, handle)

      NULLIFY (logger, mark_states, mark_list, para_env)
      logger => cp_get_default_logger()
      !-----------------------------------------------------------------------------
      ! 1.
      !-----------------------------------------------------------------------------
      CALL get_qs_env(qs_env, para_env=para_env)
      nproc = para_env%num_pe
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL section_vals_val_get(loc_print_section, "MOLECULAR_STATES%CUBE_EVAL_RANGE", &
                                explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(loc_print_section, "MOLECULAR_STATES%CUBE_EVAL_RANGE", &
                                   r_vals=eval_range)
      ELSE
         ALLOCATE (eval_range(2))
         eval_range(1) = -HUGE(0.0_dp)
         eval_range(2) = +HUGE(0.0_dp)
      ENDIF
      CALL section_vals_val_get(loc_print_section, "MOLECULAR_STATES%MARK_STATES", &
                                n_rep_val=n_rep)
      IF (n_rep .GT. 0) THEN
         ALLOCATE (mark_states(2, n_rep), STAT=isos)
         IF (isos /= 0) CALL stop_memory(routineN, moduleN, 160, &
                                         "mark_states", int_size*2*n_rep)
         IF (.NOT. ASSOCIATED(marked_states)) THEN
            ALLOCATE (marked_states(n_rep), stat=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 164, &
                                            "mark_states", int_size*n_rep)
         END IF
         DO i = 1, n_rep
            CALL section_vals_val_get(loc_print_section, "MOLECULAR_STATES%MARK_STATES", &
                                      i_rep_val=i, i_vals=mark_list)
            mark_states(:, i) = mark_list(:)
         END DO
      ELSE
         NULLIFY (marked_states)
      END IF

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2.
      !-----------------------------------------------------------------------------
      unit_report = cp_print_key_unit_nr(logger, loc_print_section, "MOLECULAR_STATES", &
                                         extension=".data", middle_name="Molecular_DOS", log_filename=.FALSE.)
      IF (unit_report > 0) THEN
         WRITE (unit_report, *) SIZE(mo_eigenvalues), " number of states "
         DO i = 1, SIZE(mo_eigenvalues)
            WRITE (unit_report, *) mo_eigenvalues(i)
         ENDDO
      ENDIF

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3.
      !-----------------------------------------------------------------------------
      CALL cp_fm_get_info(mo_localized, &
                          ncol_global=ncol_global, &
                          nrow_global=nrow_global)
      NULLIFY (smo)
      CALL cp_fm_create(smo, mo_coeff%matrix_struct)
      CALL cp_dbcsr_sm_fm_multiply(matrix_S, mo_coeff, smo, ncol_global)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4.
      !-----------------------------------------------------------------------------
      ALLOCATE (nstates(2), STAT=isos)
      IF (isos /= 0) CALL stop_memory(routineN, moduleN, 205, &
                                      "nstates", int_size*2)

      CALL cp_fm_create(storage, mo_localized%matrix_struct, name='storage')

      DO imol = 1, SIZE(molecule_set)
         IF (ASSOCIATED(molecule_set(imol)%lmi)) THEN
            nstates(1) = molecule_set(imol)%lmi%nstates
         ELSE
            nstates(1) = 0
         ENDIF
         nstates(2) = para_env%mepos

         CALL mp_maxloc(nstates, para_env%group)

         IF (nstates(1) == 0) CYCLE
         NULLIFY (states)
         ALLOCATE (states(nstates(1)), STAT=isos)
         IF (isos /= 0) CALL stop_memory(routineN, moduleN, 223, &
                                         "states", int_size*nstates(1))
         states(:) = 0

         iproc = nstates(2)
         IF (iproc == para_env%mepos) THEN
            states(:) = molecule_set(imol)%lmi%states(:)
         END IF
         !!BCAST from here root = iproc
         CALL mp_bcast(states, iproc, para_env%group)

         ns = nstates(1)
         ind => states(:)
         ALLOCATE (evals(ns), STAT=isos)
         IF (isos /= 0) CALL stop_memory(routineN, moduleN, 237, &
                                         "evals", dp_size*ns)

         NULLIFY (b, c, d, fm_struct_tmp, e_vectors, rot_e_vectors)

         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nrow_global, &
                                  ncol_global=ns, &
                                  para_env=mo_localized%matrix_struct%para_env, &
                                  context=mo_localized%matrix_struct%context)

         CALL cp_fm_create(b, fm_struct_tmp, name="b")
         CALL cp_fm_create(c, fm_struct_tmp, name="c")
         CALL cp_fm_create(rot_e_vectors, fm_struct_tmp, name="rot_e_vectors")

         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ns, ncol_global=ns, &
                                  para_env=mo_localized%matrix_struct%para_env, &
                                  context=mo_localized%matrix_struct%context)

         CALL cp_fm_create(d, fm_struct_tmp, name="d")
         CALL cp_fm_create(e_vectors, fm_struct_tmp, name="e_vectors")
         CALL cp_fm_struct_release(fm_struct_tmp)

         DO i = 1, ns
            CALL cp_fm_to_fm(mo_localized, b, 1, ind(i), i)
         END DO

         CALL cp_dbcsr_sm_fm_multiply(Hks, b, c, ns)

         CALL cp_gemm('T', 'N', ns, ns, nrow_global, 1.0_dp, &
                      b, c, 0.0_dp, d)

         CALL choose_eigv_solver(d, e_vectors, evals)

         IF (output_unit > 0) WRITE (output_unit, *) ""
         IF (output_unit > 0) WRITE (output_unit, *) "MOLECULE ", imol
         IF (output_unit > 0) WRITE (output_unit, *) "NUMBER OF STATES  ", ns
         IF (output_unit > 0) WRITE (output_unit, *) "EIGENVALUES"
         IF (output_unit > 0) WRITE (output_unit, *) ""
         IF (output_unit > 0) WRITE (output_unit, *) "ENERGY      original MO-index"

         DO k = 1, ns
            IF (ASSOCIATED(mark_states)) THEN
               DO i = 1, n_rep
                  IF (imol == mark_states(1, i) .AND. k == mark_states(2, i)) marked_states(i) = ind(k)
               END DO
            END IF
            IF (output_unit > 0) WRITE (output_unit, *) evals(k), ind(k)
         END DO
         IF (unit_report > 0) THEN
            WRITE (unit_report, *) imol, ns, " imol, number of states"
            DO k = 1, ns
               WRITE (unit_report, *) evals(k)
            ENDDO
         ENDIF

         CALL cp_gemm('N', 'N', nrow_global, ns, ns, 1.0_dp, &
                      b, e_vectors, 0.0_dp, rot_e_vectors)

         DO i = 1, ns
            CALL cp_fm_to_fm(rot_e_vectors, storage, 1, i, ind(i))
         END DO

         IF (.FALSE.) THEN ! this is too much data for large systems
            ! compute Eq. 6 from P. Hunt et al. (CPL 376, p. 68-74)
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=ns, &
                                     ncol_global=ncol_global, &
                                     para_env=mo_localized%matrix_struct%para_env, &
                                     context=mo_localized%matrix_struct%context)
            CALL cp_fm_create(D_igk, fm_struct_tmp)
            CALL cp_fm_struct_release(fm_struct_tmp)
            CALL cp_gemm('T', 'N', ns, ncol_global, nrow_global, 1.0_dp, &
                         rot_e_vectors, smo, 0.0_dp, D_igk)
            DO i = 1, ns
               DO k = 1, ncol_global
                  CALL cp_fm_get_element(D_igk, i, k, tmp)
                  IF (unit_report > 0) THEN
                     WRITE (unit_report, *) tmp**2
                  ENDIF
               ENDDO
            ENDDO
            CALL cp_fm_release(D_igk)
         ENDIF

         IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                              "MOLECULAR_STATES%CUBES"), cp_p_file)) THEN

            CALL get_qs_env(qs_env=qs_env, &
                            atomic_kind_set=atomic_kind_set, &
                            qs_kind_set=qs_kind_set, &
                            cell=cell, &
                            dft_control=dft_control, &
                            particle_set=particle_set, &
                            pw_env=pw_env)

            DO i = 1, ns
               IF (evals(i) < eval_range(1) .OR. evals(i) > eval_range(2)) CYCLE

               CALL calculate_wavefunction(rot_e_vectors, i, wf_r, &
                                           wf_g, atomic_kind_set, qs_kind_set, cell, dft_control, particle_set, &
                                           pw_env)

               WRITE (filename, '(a9,I4.4,a1,I5.5,a4)') "MOLECULE_", imol, "_", i, tag
               WRITE (title, '(A,I0,A,I0,A,F14.6,A,I0)') "Mol. Eigenstate ", i, " of ", ns, " E [a.u.] = ", &
                  evals(i), " Orig. index ", ind(i)
               unit_nr = cp_print_key_unit_nr(logger, loc_print_section, "MOLECULAR_STATES%CUBES", &
                                              extension=".cube", middle_name=TRIM(filename), log_filename=.FALSE.)
               CALL cp_pw_to_cube(wf_r%pw, unit_nr, particles=particles, title=title, &
                                  stride=section_get_ivals(loc_print_section, &
                                                           "MOLECULAR_STATES%CUBES%STRIDE"))
               CALL cp_print_key_finished_output(unit_nr, logger, loc_print_section, &
                                                 "MOLECULAR_STATES%CUBES")
            END DO
         ENDIF

         DEALLOCATE (evals)
         CALL cp_fm_release(b)
         CALL cp_fm_release(c)
         CALL cp_fm_release(d)
         CALL cp_fm_release(e_vectors)
         CALL cp_fm_release(rot_e_vectors)

         DEALLOCATE (states)

      END DO
      CALL cp_fm_release(smo)
      CALL cp_fm_to_fm(storage, mo_localized)
      CALL cp_fm_release(storage)
      IF (ASSOCIATED(mark_states)) THEN
         DEALLOCATE (mark_states)
      END IF
      DEALLOCATE (nstates)
      CALL cp_print_key_finished_output(unit_report, logger, loc_print_section, &
                                        "MOLECULAR_STATES")
      !------------------------------------------------------------------------------

      IF (.NOT. explicit) THEN
         DEALLOCATE (eval_range)
      END IF

      CALL timestop(handle)

   END SUBROUTINE construct_molecular_states

END MODULE molecular_states
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of STM image as post processing of an electronic
!>     structure calculation,
!> \par History
!>      Started as a copy from the code in qs_scf_post
!> \author Joost VandeVondele 7.2008, MI 02.2009
! **************************************************************************************************
MODULE stm_images
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE particle_list_types,             ONLY: particle_list_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'stm_images'
   PUBLIC :: th_stm_image

CONTAINS
! **************************************************************************************************
!> \brief Driver for the calculation of STM image, as post processing of a
!>        ground-state electronic structure calculation.
!> \param qs_env ...
!> \param stm_section ...
!> \param particles ...
!> \param unoccupied_orbs ...
!> \param unoccupied_evals ...
!> \param
!> \par History
!>      02.2009 Created [MI]
!> \author MI
!> \note
!>   The Tersoff-Hamman
!>        approximation is applied, occupied and a sufficient number of
!>        unoccupied eigenstates are needed (depending on the given Bias potential)
!>        and should be computed in advance. Unoccupied states are calculated
!>        before enetering this module when NLUMO =/ 0
! **************************************************************************************************

   SUBROUTINE th_stm_image(qs_env, stm_section, particles, unoccupied_orbs, &
                           unoccupied_evals)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: stm_section
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: unoccupied_orbs
      TYPE(cp_1d_r_p_type), DIMENSION(:), POINTER        :: unoccupied_evals

      CHARACTER(len=*), PARAMETER :: routineN = 'th_stm_image', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, irep, ispin, n_rep, ndim, nmo, &
                                                            nspin, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: nadd_unocc, stm_th_torb
      LOGICAL                                            :: append_cube, use_ref_energy
      REAL(KIND=dp)                                      :: efermi, ref_energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues, mo_occ, stm_biases
      TYPE(cp_1d_r_p_type), ALLOCATABLE, DIMENSION(:)    :: evals, occupation
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_dbcsr_type), POINTER                       :: stm_density_ao
      TYPE(cp_fm_p_type), ALLOCATABLE, DIMENSION(:)      :: mo_arrays
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      NULLIFY (ks_env, mos, rho, rho_ao, pw_env, stm_th_torb, fm_struct_tmp)
      NULLIFY (auxbas_pw_pool, pw_pools, stm_density_ao, mo_coeff)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      mos=mos, &
                      rho=rho, &
                      pw_env=pw_env)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      CALL section_vals_val_get(stm_section, "APPEND", l_val=append_cube)
      CALL section_vals_val_get(stm_section, "BIAS", r_vals=stm_biases)
      CALL section_vals_val_get(stm_section, "REF_ENERGY", r_val=ref_energy, explicit=use_ref_energy)
      CALL section_vals_val_get(stm_section, "TH_TORB", n_rep_val=n_rep)
      IF (n_rep == 0) THEN
         ALLOCATE (stm_th_torb(1))
         stm_th_torb(1) = 0
      ELSE
         ALLOCATE (stm_th_torb(n_rep))
         DO irep = 1, n_rep
            CALL section_vals_val_get(stm_section, "TH_TORB", &
                                      i_rep_val=irep, i_val=stm_th_torb(irep))
         END DO
      END IF

      ALLOCATE (stm_density_ao)
      CALL cp_dbcsr_init(stm_density_ao)
      CALL cp_dbcsr_copy(stm_density_ao, rho_ao(1)%matrix, &
                         name="stm_density_ao")

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools)
      CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, wf_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      nspin = SIZE(mos, 1)
      ALLOCATE (nadd_unocc(nspin))
      nadd_unocc = 0
      IF (ASSOCIATED(unoccupied_orbs)) THEN
         DO ispin = 1, nspin
            nadd_unocc(ispin) = SIZE(unoccupied_evals(ispin)%array)
         END DO
      END IF

      ALLOCATE (mo_arrays(nspin))
      ALLOCATE (evals(nspin))
      ALLOCATE (occupation(nspin))
      DO ispin = 1, nspin
         IF (nadd_unocc(ispin) == 0) THEN
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            eigenvalues=mo_eigenvalues, nmo=nmo, mu=efermi, occupation_numbers=mo_occ)
            mo_arrays(ispin)%matrix => mo_coeff
            evals(ispin)%array => mo_eigenvalues
            occupation(ispin)%array => mo_occ
         ELSE
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            eigenvalues=mo_eigenvalues, nmo=nmo, mu=efermi, occupation_numbers=mo_occ)
            ndim = nmo+nadd_unocc(ispin)
            ALLOCATE (evals(ispin)%array(ndim))
            evals(ispin)%array(1:nmo) = mo_eigenvalues(1:nmo)
            evals(ispin)%array(1+nmo:ndim) = unoccupied_evals(ispin)%array(1:nadd_unocc(ispin))
            ALLOCATE (occupation(ispin)%array(ndim))
            occupation(ispin)%array(1:nmo) = mo_occ(1:nmo)
            occupation(ispin)%array(1+nmo:ndim) = 0.0_dp
            CALL cp_fm_struct_create(fm_struct_tmp, ncol_global=ndim, &
                                     template_fmstruct=mo_coeff%matrix_struct)
            CALL cp_fm_create(mo_arrays(ispin)%matrix, fm_struct_tmp, name="mo_arrays")
            CALL cp_fm_struct_release(fm_struct_tmp)
            CALL cp_fm_to_fm(mo_coeff, mo_arrays(ispin)%matrix, nmo, 1, 1)
            CALL cp_fm_to_fm(unoccupied_orbs(ispin)%matrix, mo_arrays(ispin)%matrix, &
                             nadd_unocc(ispin), 1, nmo+1)
         END IF
      ENDDO
      IF (use_ref_energy) efermi = ref_energy

      CALL stm_cubes(ks_env, stm_section, stm_density_ao, wf_r, wf_g, mo_arrays, evals, &
                     occupation, efermi, stm_biases, stm_th_torb, particles, &
                     output_unit, append_cube)
      DO ispin = 1, nspin
         IF (nadd_unocc(ispin) > 0) THEN
            DEALLOCATE (evals(ispin)%array)
            DEALLOCATE (occupation(ispin)%array)
            CALL cp_fm_release(mo_arrays(ispin)%matrix)
         END IF
      END DO
      DEALLOCATE (mo_arrays)
      DEALLOCATE (evals)
      DEALLOCATE (occupation)

      CALL cp_dbcsr_deallocate_matrix(stm_density_ao)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)

      DEALLOCATE (stm_th_torb)
      DEALLOCATE (nadd_unocc)

      CALL timestop(handle)

   END SUBROUTINE th_stm_image

! **************************************************************************************************
!> \brief computes a simple approximation to the tunneling current for STM
!> \param ks_env ...
!> \param stm_section ...
!> \param stm_density_ao ...
!> \param wf_r ...
!> \param wf_g ...
!> \param mo_arrays ...
!> \param evals ...
!> \param occupation ...
!> \param efermi ...
!> \param stm_biases ...
!> \param stm_th_torb ...
!> \param particles ...
!> \param output_unit ...
!> \param append_cube ...
!> \param
!> \par History
!>      7.2008 Created [Joost VandeVondele]
!>       07.2009 modified MI
!> \author Joost VandeVondele
!> \note
!>      requires the MOs that are passed to be eigenstates, and energy ordered
! **************************************************************************************************
   SUBROUTINE stm_cubes(ks_env, stm_section, stm_density_ao, wf_r, wf_g, mo_arrays, evals, &
                        occupation, efermi, stm_biases, stm_th_torb, particles, &
                        output_unit, append_cube)

      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(section_vals_type), POINTER                   :: stm_section
      TYPE(cp_dbcsr_type), POINTER                       :: stm_density_ao
      TYPE(pw_p_type)                                    :: wf_r, wf_g
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(IN)       :: mo_arrays
      TYPE(cp_1d_r_p_type), DIMENSION(:), INTENT(IN)     :: evals, occupation
      REAL(KIND=dp)                                      :: efermi
      REAL(KIND=dp), DIMENSION(:), POINTER               :: stm_biases
      INTEGER, DIMENSION(:), POINTER                     :: stm_th_torb
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: append_cube

      CHARACTER(LEN=*), DIMENSION(0:9), PARAMETER :: &
         torb_string = (/"  s", " px", " py", " pz", "dxy", "dyz", "dzx", "dx2", "dy2", "dz2"/)
      CHARACTER(len=*), PARAMETER :: routineN = 'stm_cubes', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_pos, oname, title
      INTEGER                                            :: handle, i, ibias, imo, iorb, ispin, &
                                                            istates, nmo, nspin, nstates(2), &
                                                            state_start(2), unit_nr
      REAL(KIND=dp)                                      :: alpha, total_rho
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: occ_tot
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: matrix_v, matrix_vf
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      NULLIFY (fm_struct_tmp)

      nspin = SIZE(mo_arrays)

      IF (output_unit > 0) WRITE (output_unit, '(T2,A)') ""
      IF (output_unit > 0) WRITE (output_unit, '(T2,A,F12.6, A)') "STM : Reference energy ", efermi, " a.u. "
      DO ibias = 1, SIZE(stm_biases)

         IF (output_unit > 0) WRITE (output_unit, '(T2,A)') ""
         IF (output_unit > 0) WRITE (output_unit, '(T2,A,F16.6)') &
            "Preparing for STM image at bias [a.u.] ", stm_biases(ibias)

         istates = 0
         nstates = 0
         state_start = 0
         DO ispin = 1, nspin
            IF (stm_biases(ibias) < 0.0_dp) THEN
               nmo = SIZE(evals(ispin)%array)
               DO imo = 1, nmo
                  IF (evals(ispin)%array(imo) > (efermi+stm_biases(ibias)) .AND. &
                      evals(ispin)%array(imo) <= efermi) THEN
                     IF (nstates(ispin) == 0) state_start(ispin) = imo
                     nstates(ispin) = nstates(ispin)+1
                  END IF
               END DO
               IF ((output_unit > 0) .AND. evals(ispin)%array(1) > efermi+stm_biases(ibias)) &
                  WRITE (output_unit, '(T4,A)') "Warning: EFermi+bias below lowest computed occupied MO"
            ELSE
               nmo = SIZE(evals(ispin)%array)
               DO imo = 1, nmo
                  IF (evals(ispin)%array(imo) <= (efermi+stm_biases(ibias)) .AND. &
                      evals(ispin)%array(imo) > efermi) THEN
                     IF (nstates(ispin) == 0) state_start(ispin) = imo
                     nstates(ispin) = nstates(ispin)+1
                  END IF
               END DO
               IF ((output_unit > 0) .AND. evals(ispin)%array(nmo) < efermi+stm_biases(ibias)) &
                  WRITE (output_unit, '(T4,A)') "Warning: E-Fermi+bias above highest computed unoccupied MO"
            ENDIF
            istates = istates+nstates(ispin)
         ENDDO
         IF ((output_unit > 0)) WRITE (output_unit, '(T4,A,I0,A)') "Using a total of ", istates, " states"
         IF (istates == 0) CYCLE

         CALL cp_fm_struct_create(fm_struct_tmp, ncol_global=istates, &
                                  template_fmstruct=mo_arrays(1)%matrix%matrix_struct)
         CALL cp_fm_create(matrix_v, fm_struct_tmp, name="matrix_v")
         CALL cp_fm_create(matrix_vf, fm_struct_tmp, name="matrix_vf")
         CALL cp_fm_struct_release(fm_struct_tmp)

         ALLOCATE (occ_tot(istates))

         ! we sum both alpha and beta electrons together for this density of states
         istates = 0
         alpha = 1.0_dp
         IF (nspin == 1) alpha = 2.0_dp
         DO ispin = 1, nspin
            CALL cp_fm_to_fm(mo_arrays(ispin)%matrix, matrix_v, nstates(ispin), state_start(ispin), istates+1)
            CALL cp_fm_to_fm(mo_arrays(ispin)%matrix, matrix_vf, nstates(ispin), state_start(ispin), istates+1)
            IF (stm_biases(ibias) < 0.0_dp) THEN
               occ_tot(istates+1:istates+nstates(ispin)) = &
                  occupation(ispin)%array(state_start(ispin):state_start(ispin)-1+nstates(ispin))
            ELSE
               occ_tot(istates+1:istates+nstates(ispin)) = &
                  alpha-occupation(ispin)%array(state_start(ispin):state_start(ispin)-1+nstates(ispin))
            END IF
            istates = istates+nstates(ispin)
         ENDDO

         CALL cp_fm_column_scale(matrix_vf, occ_tot(1:istates))
         alpha = 1.0_dp

         CALL cp_dbcsr_set(stm_density_ao, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(stm_density_ao, matrix_v=matrix_v, matrix_g=matrix_vf, ncol=istates, &
                                    alpha=alpha)

         DO i = 1, SIZE(stm_th_torb)
            iorb = stm_th_torb(i)
            CALL calculate_rho_elec(matrix_p=stm_density_ao, &
                                    rho=wf_r, rho_gspace=wf_g, total_rho=total_rho, &
                                    ks_env=ks_env, der_type=iorb)

            oname = torb_string(iorb)
!         fname = "STM_"//TRIM(torb_string(iorb))
            WRITE (filename, '(a4,I2.2,a1,I5.5)') "STM_d", iorb, "_", ibias
            my_pos = "REWIND"
            IF (append_cube) THEN
               my_pos = "APPEND"
            END IF

            unit_nr = cp_print_key_unit_nr(logger, stm_section, extension=".cube", &
                                           middle_name=TRIM(filename), file_position=my_pos, file_action="WRITE", &
                                           log_filename=.FALSE.)
            WRITE (title, '(A,I0,A,I0,A,F16.8)') "STM cube ", ibias, " wfn deriv. ", iorb, " at bias ", stm_biases(ibias)
            CALL cp_pw_to_cube(wf_r%pw, unit_nr, title, particles=particles, &
                               stride=section_get_ivals(stm_section, "STRIDE"), zero_tails=.TRUE.)

            CALL cp_print_key_finished_output(unit_nr, logger, stm_section)
         END DO

         CALL cp_fm_release(matrix_v)
         CALL cp_fm_release(matrix_vf)
         DEALLOCATE (occ_tot)

      ENDDO

      CALL timestop(handle)

   END SUBROUTINE stm_cubes

END MODULE stm_images
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the calculation of wannier states
!> \author Alin M Elena
! **************************************************************************************************
MODULE wannier_states
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_element,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE wannier_states_types,            ONLY: wannier_centres_type
!!!! this ones are needed to mapping
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'wannier_states'

   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .TRUE.

! *** Public subroutines ***

   PUBLIC :: construct_wannier_states

CONTAINS

! **************************************************************************************************
!> \brief constructs wannier states. mo_localized should not be overwritten!
!> \param mo_localized ...
!> \param Hks ...
!> \param qs_env ...
!> \param loc_print_section ...
!> \param WannierCentres ...
!> \param ns ...
!> \param states ...
! **************************************************************************************************
   SUBROUTINE construct_wannier_states(mo_localized, &
                                       Hks, qs_env, loc_print_section, WannierCentres, ns, states)

      TYPE(cp_fm_type), INTENT(in), POINTER              :: mo_localized
      TYPE(cp_dbcsr_type), POINTER                       :: Hks
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: loc_print_section
      TYPE(wannier_centres_type), INTENT(INOUT)          :: WannierCentres
      INTEGER, INTENT(IN)                                :: ns
      INTEGER, INTENT(IN), POINTER                       :: states(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_wannier_states', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: unit_str
      INTEGER                                            :: handle, i, iproc, ncol_global, nproc, &
                                                            nrow_global, nstates(2), output_unit, &
                                                            unit_mat
      REAL(KIND=dp)                                      :: unit_conv
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: b, c, d
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: print_key

!-----------------------------------------------------------------------
!-----------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (logger, para_env)

      CALL get_qs_env(qs_env, para_env=para_env)
      nproc = para_env%num_pe

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL cp_fm_get_info(mo_localized, &
                          ncol_global=ncol_global, &
                          nrow_global=nrow_global)

      nstates(1) = ns
      nstates(2) = para_env%mepos
      iproc = nstates(2)
      NULLIFY (b, c, d, fm_struct_tmp, print_key)

      print_key => section_vals_get_subs_vals(loc_print_section, "WANNIER_CENTERS")
      CALL section_vals_val_get(print_key, "UNIT", c_val=unit_str)
      unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nrow_global, &
                               ncol_global=1, &
                               para_env=mo_localized%matrix_struct%para_env, &
                               context=mo_localized%matrix_struct%context)

      CALL cp_fm_create(b, fm_struct_tmp, name="b")
      CALL cp_fm_create(c, fm_struct_tmp, name="c")

      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=1, ncol_global=1, &
                               para_env=mo_localized%matrix_struct%para_env, &
                               context=mo_localized%matrix_struct%context)

      CALL cp_fm_create(d, fm_struct_tmp, name="d")
      CALL cp_fm_struct_release(fm_struct_tmp)

      WannierCentres%WannierHamDiag = 0.0_dp
      ! try to print the matrix

      unit_mat = cp_print_key_unit_nr(logger, loc_print_section, &
                                      "WANNIER_STATES", extension=".whks", &
                                      ignore_should_output=.FALSE.)
      IF (unit_mat > 0) THEN
         WRITE (unit_mat, '(a16,1(i0,1x))') "Wannier states: ", ns
         WRITE (unit_mat, '(a16)') "#No x y z energy "
      ENDIF
      DO i = 1, ns
         CALL cp_fm_to_fm(mo_localized, b, 1, states(i), 1)
         CALL cp_dbcsr_sm_fm_multiply(Hks, b, c, 1)
         CALL cp_gemm('T', 'N', 1, 1, nrow_global, 1.0_dp, &
                      b, c, 0.0_dp, d)
         CALL cp_fm_get_element(d, 1, 1, WannierCentres%WannierHamDiag(i))
         !               if (iproc==para_env%mepos) WRITE(unit_mat,'(f16.8,2x)', advance='no')WannierCentres%WannierHamDiag(i)
         IF (unit_mat > 0) WRITE (unit_mat, '(i0,1x,4(f16.8,2x))') states(i), &
            WannierCentres%centres(1:3, states(i))*unit_conv, WannierCentres%WannierHamDiag(states(i))
      END DO

      IF (unit_mat > 0) WRITE (unit_mat, *)
      CALL cp_print_key_finished_output(unit_mat, logger, loc_print_section, &
                                        "WANNIER_STATES")
      IF (output_unit > 0) THEN
         WRITE (output_unit, *) ""
         WRITE (output_unit, *) "NUMBER OF Wannier STATES  ", ns
         WRITE (output_unit, *) "ENERGY      original MO-index"
         DO i = 1, ns
            WRITE (output_unit, '(f16.8,2x,i0)') WannierCentres%WannierHamDiag(i), states(i)
         ENDDO
      ENDIF
      CALL cp_fm_release(b)
      CALL cp_fm_release(c)
      CALL cp_fm_release(d)
      CALL timestop(handle)
   END SUBROUTINE construct_wannier_states

END MODULE wannier_states

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Does all kind of post scf calculations for GPW/GAPW
!> \par History
!>      Started as a copy from the relevant part of qs_scf
!>      Start to adapt for k-points [07.2015, JGH]
!> \author Joost VandeVondele (10.2003)
! **************************************************************************************************
MODULE qs_scf_post_gpw
   USE admm_types,                      ONLY: admm_type
   USE admm_utils,                      ONLY: admm_correct_for_eigenvalues,&
                                              admm_uncorrect_for_eigenvalues
   USE ai_onecenter,                    ONLY: sg_overlap
   USE atom_kind_orbitals,              ONLY: calculate_atomic_density
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_ddapc_util,                   ONLY: get_ddapc
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_init_random,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_iter_string,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              get_results,&
                                              put_results
   USE cp_result_types,                 ONLY: cp_result_type
   USE dct,                             ONLY: pw_shrink
   USE et_coupling_types,               ONLY: set_et_coupling_type
   USE hirshfeld_methods,               ONLY: comp_hirshfeld_charges,&
                                              comp_hirshfeld_i_charges,&
                                              create_shape_function,&
                                              write_hirshfeld_charges
   USE hirshfeld_types,                 ONLY: create_hirshfeld_type,&
                                              hirshfeld_type,&
                                              release_hirshfeld_type,&
                                              set_hirshfeld_info
   USE input_constants,                 ONLY: do_loc_both,&
                                              do_loc_homo,&
                                              do_loc_lumo,&
                                              ot_precond_full_all,&
                                              ref_charge_atomic,&
                                              ref_charge_mulliken
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_ivals,&
                                              section_get_lval,&
                                              section_get_rval,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE lapack,                          ONLY: lapack_sgesv
   USE mao_methods,                     ONLY: mao_analysis
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE message_passing,                 ONLY: mp_sum
   USE molecular_states,                ONLY: construct_molecular_states
   USE molecule_types_new,              ONLY: molecule_type
   USE moments_utils,                   ONLY: get_reference_point
   USE mulliken,                        ONLY: mulliken_charges
   USE orbital_pointers,                ONLY: indso
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom,&
                                              debye,&
                                              evolt
   USE population_analyses,             ONLY: lowdin_population_analysis,&
                                              mulliken_population_analysis
   USE preconditioner_types,            ONLY: preconditioner_type
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grids,                        ONLY: get_pw_grid_info
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_types,                ONLY: pw_poisson_implicit,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_active_space_methods,         ONLY: active_space_main
   USE qs_collocate_density,            ONLY: calculate_rho_elec,&
                                              calculate_wavefunction
   USE qs_core_energies,                ONLY: calculate_ptrace
   USE qs_electric_field_gradient,      ONLY: qs_efg_calc
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_epr_hyp,                      ONLY: qs_epr_hyp_calc
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_loc_methods,                  ONLY: qs_loc_driver
   USE qs_loc_molecules,                ONLY: wfc_to_molecule
   USE qs_loc_types,                    ONLY: qs_loc_env_create,&
                                              qs_loc_env_destroy,&
                                              qs_loc_env_new_type
   USE qs_loc_utils,                    ONLY: loc_write_restart,&
                                              qs_loc_control_init,&
                                              qs_loc_env_init,&
                                              qs_loc_init,&
                                              retain_history
   USE qs_mo_io,                        ONLY: write_dm_binary_restart,&
                                              write_mo_set
   USE qs_mo_methods,                   ONLY: calculate_subspace_eigenvalues,&
                                              make_mo_eig
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_moments,                      ONLY: qs_moment_berry_phase,&
                                              qs_moment_locop
   USE qs_ot_eigensolver,               ONLY: ot_eigensolver
   USE qs_pdos,                         ONLY: calculate_projected_dos
   USE qs_resp,                         ONLY: resp_fit
   USE qs_rho0_types,                   ONLY: get_rho0_mpole,&
                                              mpole_rho_atom,&
                                              rho0_mpole_type
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: ot_method_nr,&
                                              qs_scf_env_type
   USE qs_scf_wfn_mix,                  ONLY: wfn_mix
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE s_square_methods,                ONLY: compute_s_square
   USE scf_control_types,               ONLY: scf_control_type
   USE stm_images,                      ONLY: th_stm_image
   USE wannier_states,                  ONLY: construct_wannier_states
   USE wannier_states_types,            ONLY: wannier_centres_type
   USE xray_diffraction,                ONLY: calculate_rhotot_elec_gspace,&
                                              xray_diffraction_spectrum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_post_gpw'
   PUBLIC :: scf_post_calculation_gpw, &
             qs_scf_post_moments, &
             write_available_results, &
             write_mo_free_results

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief collects possible post - scf calculations and prints info / computes properties.
!> \param dft_section ...
!> \param scf_env the scf_env whose info should be written out
!> \param qs_env the qs_env in which the qs_env lives
!> \param wf_type ...
!> \par History
!>      02.2003 created [fawzi]
!>      10.2004 moved here from qs_scf [Joost VandeVondele]
!>              started splitting out different subroutines
!>      10.2015 added header for wave-function correlated methods [Vladimir Rybkin]
!> \author fawzi
!> \note
!>      this function changes mo_eigenvectors and mo_eigenvalues, depending on the print keys.
!>      In particular, MO_CUBES causes the MOs to be rotated to make them eigenstates of the KS
!>      matrix, and mo_eigenvalues is updated accordingly. This can, for unconverged wavefunctions,
!>      change afterwards slightly the forces (hence small numerical differences between MD
!>      with and without the debug print level). Ideally this should not happen...
! **************************************************************************************************
   SUBROUTINE scf_post_calculation_gpw(dft_section, scf_env, qs_env, wf_type)

      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      CHARACTER(6), OPTIONAL                             :: wf_type

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_post_calculation_gpw', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, ispin, min_lumos, n_rep, &
                                                            nhomo, nlumo, nlumo_stm, nlumo_tddft, &
                                                            nlumos, nmo, output_unit, unit_nr
      INTEGER, DIMENSION(:, :, :), POINTER               :: marked_states
      LOGICAL :: check_write, compute_lumos, do_homo, do_kpoints, do_mo_cubes, do_stm, &
         do_wannier_cubes, has_homo, has_lumo, loc_explicit, loc_print_explicit, my_localized_wfn, &
         p_loc, p_loc_homo, p_loc_lumo
      REAL(dp)                                           :: e_kin
      REAL(KIND=dp)                                      :: gap, homo_lumo(2, 2)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_1d_r_p_type), DIMENSION(:), POINTER        :: occupied_evals, unoccupied_evals, &
                                                            unoccupied_evals_stm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, matrix_s, mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: kinetic_m, rho_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: homo_localized, lumo_localized, lumo_ptr, &
         mo_loc_history, occupied_orbs, unoccupied_orbs, unoccupied_orbs_stm
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env_homo, qs_loc_env_lumo
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: input, loc_print_section, &
                                                            localize_section, print_key, &
                                                            stm_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      ! Print out the type of wavefunction to distinguish between SCF and post-SCF
      IF (PRESENT(wf_type)) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT='(/,(T1,A))') REPEAT("-", 40)
            WRITE (UNIT=output_unit, FMT='(/,(T3,A,T19,A,T25,A))') "Properties from ", wf_type, " density"
            WRITE (UNIT=output_unit, FMT='(/,(T1,A))') REPEAT("-", 40)
         ENDIF
      ENDIF

      ! Writes the data that is already available in qs_env
      CALL write_available_results(qs_env, scf_env)

      my_localized_wfn = .FALSE.
      NULLIFY (admm_env, dft_control, pw_env, auxbas_pw_pool, pw_pools, mos, rho, &
               mo_coeff, ks_rmpv, matrix_s, qs_loc_env_homo, qs_loc_env_lumo, scf_control, &
               unoccupied_orbs, unoccupied_orbs_stm, mo_eigenvalues, unoccupied_evals, &
               unoccupied_evals_stm, molecule_set, mo_derivs, &
               subsys, particles, input, print_key, kinetic_m, marked_states)
      NULLIFY (homo_localized, lumo_localized, lumo_ptr, rho_ao)

      has_homo = .FALSE.
      has_lumo = .FALSE.
      p_loc = .FALSE.
      p_loc_homo = .FALSE.
      p_loc_lumo = .FALSE.

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_post_gpw.F",298)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_post_gpw.F",299)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_gpw.F",300)
      ! Here we start with data that needs a postprocessing...
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      molecule_set=molecule_set, &
                      scf_control=scf_control, &
                      do_kpoints=do_kpoints, &
                      input=input, &
                      subsys=subsys, &
                      rho=rho, &
                      pw_env=pw_env, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)
      CALL qs_subsys_get(subsys, particles=particles)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao)

      !    **** the kinetic energy
      IF (cp_print_key_should_output(logger%iter_info, input, &
                                     "DFT%PRINT%KINETIC_ENERGY") /= 0) THEN
         CALL get_qs_env(qs_env, kinetic_kp=kinetic_m)
         IF(.NOT.(ASSOCIATED(kinetic_m)))CALL cp__a("qs_scf_post_gpw.F",322)
         IF(.NOT.(ASSOCIATED(kinetic_m(1, 1)%matrix)))CALL cp__a("qs_scf_post_gpw.F",323)
         CALL calculate_ptrace(kinetic_m, rho_ao, e_kin, dft_control%nspins)
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%KINETIC_ENERGY", &
                                        extension=".Log")
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T3,A,T55,F25.14)') "Electronic kinetic energy:", e_kin
         ENDIF
         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%KINETIC_ENERGY")
      END IF

      ! Atomic Charges that require further computation
      CALL qs_scf_post_charges(input, logger, qs_env)

      ! Moments of charge distribution
      CALL qs_scf_post_moments(input, logger, qs_env, output_unit)

      ! Determine if we need to computer properties using the localized centers
      localize_section => section_vals_get_subs_vals(dft_section, "LOCALIZE")
      loc_print_section => section_vals_get_subs_vals(localize_section, "PRINT")
      CALL section_vals_get(localize_section, explicit=loc_explicit)
      CALL section_vals_get(loc_print_section, explicit=loc_print_explicit)

      ! Print_keys controlled by localization
      IF (loc_print_explicit) THEN
         print_key => section_vals_get_subs_vals(loc_print_section, "MOLECULAR_DIPOLES")
         p_loc = BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
         print_key => section_vals_get_subs_vals(loc_print_section, "TOTAL_DIPOLE")
         p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
         print_key => section_vals_get_subs_vals(loc_print_section, "WANNIER_CENTERS")
         p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
         print_key => section_vals_get_subs_vals(loc_print_section, "WANNIER_SPREADS")
         p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
         print_key => section_vals_get_subs_vals(loc_print_section, "WANNIER_CUBES")
         p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
         print_key => section_vals_get_subs_vals(loc_print_section, "MOLECULAR_STATES")
         p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)
      ELSE
         p_loc = .FALSE.
      END IF
      IF (loc_explicit) THEN
         p_loc_homo = (section_get_ival(localize_section, "STATES") == do_loc_homo .OR. &
                       section_get_ival(localize_section, "STATES") == do_loc_both) .AND. p_loc
         p_loc_lumo = (section_get_ival(localize_section, "STATES") == do_loc_lumo .OR. &
                       section_get_ival(localize_section, "STATES") == do_loc_both) .AND. p_loc
         CALL section_vals_val_get(localize_section, "LIST_UNOCCUPIED", n_rep_val=n_rep)
      ELSE
         p_loc_homo = .FALSE.
         p_loc_lumo = .FALSE.
         n_rep = 0
      END IF

      IF (n_rep == 0 .AND. p_loc_lumo) THEN
         CALL cp_abort(cp__l("qs_scf_post_gpw.F",376), "No LIST_UNOCCUPIED was specified, "// &
                       "therefore localization of unoccupied states will be skipped!")
         p_loc_lumo = .FALSE.
      END IF
      print_key => section_vals_get_subs_vals(loc_print_section, "WANNIER_STATES")
      p_loc = p_loc .OR. BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)

      ! Control for STM
      stm_section => section_vals_get_subs_vals(input, "DFT%PRINT%STM")
      CALL section_vals_get(stm_section, explicit=do_stm)
      nlumo_stm = 0
      IF (do_stm) nlumo_stm = section_get_ival(stm_section, "NLUMO")

      ! check for CUBES (MOs and WANNIERS)
      do_mo_cubes = BTEST(cp_print_key_should_output(logger%iter_info, dft_section, "PRINT%MO_CUBES") &
                          , cp_p_file)
      IF (loc_print_explicit) THEN
         do_wannier_cubes = BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                                             "WANNIER_CUBES"), cp_p_file)
      ELSE
         do_wannier_cubes = .FALSE.
      END IF
      nlumo = section_get_ival(dft_section, "PRINT%MO_CUBES%NLUMO")
      nhomo = section_get_ival(dft_section, "PRINT%MO_CUBES%NHOMO")
      nlumo_tddft = 0
      IF (dft_control%do_tddfpt_calculation) THEN
         nlumo_tddft = section_get_ival(dft_section, "TDDFPT%NLUMO")
      END IF

      ! Setup the grids needed to compute a wavefunction given a vector..
      IF (((do_mo_cubes .OR. do_wannier_cubes) .AND. (nlumo /= 0 .OR. nhomo /= 0)) .OR. p_loc) THEN
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                         pw_pools=pw_pools)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_g%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      END IF

      ! Makes the MOs eigenstates, computes eigenvalues, write cubes
      IF (do_kpoints) THEN
         IF (do_mo_cubes) THEN
            CALL cp__w("qs_scf_post_gpw.F",420,"Print MO Cubes not implemented for k-point calculations!!")
         END IF
      ELSE
         CALL get_qs_env(qs_env, &
                         mos=mos, &
                         matrix_ks=ks_rmpv)
         IF ((do_mo_cubes .AND. nhomo /= 0) .OR. do_stm .OR. dft_control%do_tddfpt_calculation) THEN
            IF (dft_control%restricted) THEN
               IF (output_unit > 0) WRITE (output_unit, *) &
                  " Unclear how we define MOs in the restricted case ... skipping"
            ELSE
               CALL get_qs_env(qs_env, mo_derivs=mo_derivs)
               IF (dft_control%do_admm) THEN
                  CALL get_qs_env(qs_env, admm_env=admm_env)
                  CALL make_mo_eig(mos, dft_control%nspins, ks_rmpv, scf_control, mo_derivs, admm_env=admm_env)
               ELSE
                  CALL make_mo_eig(mos, dft_control%nspins, ks_rmpv, scf_control, mo_derivs)
               END IF
            END IF
            DO ispin = 1, dft_control%nspins
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, eigenvalues=mo_eigenvalues, homo=homo)
               homo_lumo(ispin, 1) = mo_eigenvalues(homo)
            END DO
            has_homo = .TRUE.
         END IF
         IF (do_mo_cubes .AND. nhomo /= 0) THEN
            DO ispin = 1, dft_control%nspins
               ! Prints the cube files of OCCUPIED ORBITALS
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                               eigenvalues=mo_eigenvalues, homo=homo, nmo=nmo)
               CALL qs_scf_post_occ_cubes(input, dft_section, dft_control, logger, qs_env, &
                                          mo_coeff, wf_g, wf_r, particles, homo, ispin)
            END DO
         ENDIF
      ENDIF

      ! Initialize the localization environment, needed e.g. for wannier functions and molecular states
      ! Gets localization info for the occupied orbs
      !  - Possibly gets wannier functions
      !  - Possibly gets molecular states
      IF (p_loc_homo) THEN
         IF (do_kpoints) THEN
            CALL cp__w("qs_scf_post_gpw.F",462,"Localization not implemented for k-point calculations!!")
         ELSEIF (dft_control%restricted) THEN
            IF (output_unit > 0) WRITE (output_unit, *) &
               " Unclear how we define MOs / localization in the restricted case ... skipping"
         ELSE
            ALLOCATE (occupied_orbs(dft_control%nspins))
            ALLOCATE (occupied_evals(dft_control%nspins))
            ALLOCATE (homo_localized(dft_control%nspins))
            DO ispin = 1, dft_control%nspins
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                               eigenvalues=mo_eigenvalues)
               occupied_orbs(ispin)%matrix => mo_coeff
               occupied_evals(ispin)%array => mo_eigenvalues
               CALL cp_fm_create(homo_localized(ispin)%matrix, occupied_orbs(ispin)%matrix%matrix_struct)
               CALL cp_fm_to_fm(occupied_orbs(ispin)%matrix, homo_localized(ispin)%matrix)
            END DO

            CALL get_qs_env(qs_env, mo_loc_history=mo_loc_history)
            do_homo = .TRUE.

            CALL qs_loc_env_create(qs_loc_env_homo)
            CALL qs_loc_control_init(qs_loc_env_homo, localize_section, do_homo=do_homo)
            CALL qs_loc_init(qs_env, qs_loc_env_homo, localize_section, homo_localized, do_homo, &
                             do_mo_cubes, mo_loc_history=mo_loc_history)
            CALL get_localization_info(qs_env, qs_loc_env_homo, localize_section, homo_localized, &
                                       wf_r, wf_g, particles, occupied_orbs, occupied_evals, marked_states)

            !retain the homo_localized for future use
            IF (qs_loc_env_homo%localized_wfn_control%use_history) THEN
               CALL retain_history(mo_loc_history, homo_localized)
               CALL set_qs_env(qs_env, mo_loc_history=mo_loc_history)
            ENDIF

            !write restart for localization of occupied orbitals
            CALL loc_write_restart(qs_loc_env_homo, loc_print_section, mos, &
                                   homo_localized, do_homo)
            CALL cp_fm_vect_dealloc(homo_localized)
            DEALLOCATE (occupied_orbs)
            DEALLOCATE (occupied_evals)
            ! Print Total Dipole if the localization has been performed
            IF (qs_loc_env_homo%do_localize) THEN
               CALL qs_scf_post_loc_dip(input, dft_control, qs_loc_env_homo, logger, qs_env)
            END IF
         END IF
      ENDIF

      ! Gets the lumos, and eigenvalues for the lumos, and localize them if requested
      IF (do_kpoints) THEN
         ! nothing at the moment, not implemented
         CALL cp__w("qs_scf_post_gpw.F",511,"Localization and MO related output not implemented for k-point calculations!")
      ELSE
         IF (nlumo .GT. -1) THEN
            nlumo = MAX(nlumo, nlumo_tddft)
         END IF
         compute_lumos = (do_mo_cubes .OR. dft_control%do_tddfpt_calculation) .AND. nlumo .NE. 0
         compute_lumos = compute_lumos .OR. p_loc_lumo

         DO ispin = 1, dft_control%nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, homo=homo, nmo=nmo)
            compute_lumos = compute_lumos .AND. homo == nmo
         END DO

         IF (do_mo_cubes .AND. .NOT. compute_lumos) THEN

            nlumo = section_get_ival(dft_section, "PRINT%MO_CUBES%NLUMO")
            DO ispin = 1, dft_control%nspins

               CALL get_mo_set(mo_set=mos(ispin)%mo_set, homo=homo, nmo=nmo, eigenvalues=mo_eigenvalues)
               IF (nlumo > nmo-homo) THEN
                  ! this case not yet implemented
               ELSE
                  IF (nlumo .EQ. -1) THEN
                     nlumo = nmo-homo
                  END IF
                  IF (output_unit > 0) WRITE (output_unit, *) " "
                  IF (output_unit > 0) WRITE (output_unit, *) " Lowest eigenvalues of the unoccupied subspace spin ", ispin
                  IF (output_unit > 0) WRITE (output_unit, *) "---------------------------------------------"
                  IF (output_unit > 0) WRITE (output_unit, '(4(1X,1F16.8))') mo_eigenvalues(homo+1:homo+nlumo)

                  ! Prints the cube files of UNOCCUPIED ORBITALS
                  CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
                  CALL qs_scf_post_unocc_cubes(input, dft_section, dft_control, logger, qs_env, &
                                               mo_coeff, wf_g, wf_r, particles, nlumo, homo, ispin, lumo=homo+1)
               END IF
            END DO

         END IF

         IF (compute_lumos) THEN
            check_write = .TRUE.
            min_lumos = nlumo
            IF (nlumo == 0) check_write = .FALSE.
            IF (p_loc_lumo) THEN
               do_homo = .FALSE.
               CALL qs_loc_env_create(qs_loc_env_lumo)
               CALL qs_loc_control_init(qs_loc_env_lumo, localize_section, do_homo=do_homo)
               min_lumos = MAX(MAXVAL(qs_loc_env_lumo%localized_wfn_control%loc_states(:, :)), nlumo)
            END IF

            ALLOCATE (unoccupied_orbs(dft_control%nspins))
            ALLOCATE (unoccupied_evals(dft_control%nspins))
            CALL make_lumo(qs_env, scf_env, unoccupied_orbs, unoccupied_evals, min_lumos, nlumos)
            lumo_ptr => unoccupied_orbs
            DO ispin = 1, dft_control%nspins
               has_lumo = .TRUE.
               homo_lumo(ispin, 2) = unoccupied_evals(ispin)%array(1)
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, homo=homo)
               IF (check_write) THEN
                  IF (p_loc_lumo .AND. nlumo .NE. -1) nlumos = MIN(nlumo, nlumos)
                  ! Prints the cube files of UNOCCUPIED ORBITALS
                  CALL qs_scf_post_unocc_cubes(input, dft_section, dft_control, logger, qs_env, &
                                               unoccupied_orbs(ispin)%matrix, wf_g, wf_r, particles, nlumos, homo, ispin)
               END IF
            END DO

            ! Save the info for tddfpt calculation
            IF (dft_control%do_tddfpt_calculation) THEN
               ALLOCATE (dft_control%tddfpt_control%lumos_eigenvalues(nlumos, dft_control%nspins))
               DO ispin = 1, dft_control%nspins
                  dft_control%tddfpt_control%lumos_eigenvalues(1:nlumos, ispin) = &
                     unoccupied_evals(ispin)%array(1:nlumos)
               END DO
               dft_control%tddfpt_control%lumos => unoccupied_orbs
            END IF

            IF (p_loc_lumo) THEN
               ALLOCATE (lumo_localized(dft_control%nspins))
               DO ispin = 1, dft_control%nspins
                  CALL cp_fm_create(lumo_localized(ispin)%matrix, unoccupied_orbs(ispin)%matrix%matrix_struct)
                  CALL cp_fm_to_fm(unoccupied_orbs(ispin)%matrix, lumo_localized(ispin)%matrix)
               END DO
               CALL qs_loc_init(qs_env, qs_loc_env_lumo, localize_section, lumo_localized, do_homo, do_mo_cubes, &
                                evals=unoccupied_evals)
               CALL qs_loc_env_init(qs_loc_env_lumo, qs_loc_env_lumo%localized_wfn_control, qs_env, &
                                    loc_coeff=unoccupied_orbs)
               CALL get_localization_info(qs_env, qs_loc_env_lumo, localize_section, &
                                          lumo_localized, wf_r, wf_g, particles, &
                                          unoccupied_orbs, unoccupied_evals, marked_states)
               CALL loc_write_restart(qs_loc_env_lumo, loc_print_section, mos, homo_localized, do_homo, &
                                      evals=unoccupied_evals)
               lumo_ptr => lumo_localized
            END IF
         ENDIF

         IF (has_homo .AND. has_lumo) THEN
            IF (output_unit > 0) WRITE (output_unit, *) " "
            DO ispin = 1, dft_control%nspins
               IF (.NOT. scf_control%smear%do_smear) THEN
                  gap = homo_lumo(ispin, 2)-homo_lumo(ispin, 1)
                  IF (output_unit > 0) WRITE (output_unit, '(T2,A,F12.6)') &
                     "HOMO - LUMO gap [eV] :", gap*evolt
               END IF
            ENDDO
         ENDIF
      ENDIF

      ! Deallocate grids needed to compute wavefunctions
      IF (((do_mo_cubes .OR. do_wannier_cubes) .AND. (nlumo /= 0 .OR. nhomo /= 0)) .OR. p_loc) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_g%pw)
      END IF

      ! Destroy the localization environment
      IF (.NOT. do_kpoints) THEN
         IF (p_loc_homo) CALL qs_loc_env_destroy(qs_loc_env_homo)
         IF (p_loc_lumo) CALL qs_loc_env_destroy(qs_loc_env_lumo)
      END IF

      ! generate a mix of wfns, and write to a restart
      IF (do_kpoints) THEN
         ! nothing at the moment, not implemented
      ELSE
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         CALL wfn_mix(mos, particle_set, dft_section, qs_kind_set, &
                      lumo_ptr, scf_env, matrix_s, output_unit, marked_states)
         IF (p_loc_lumo) CALL cp_fm_vect_dealloc(lumo_localized)
      END IF
      IF (ASSOCIATED(marked_states)) THEN
         DEALLOCATE (marked_states)
      END IF

      ! This is just a deallocation for printing MO_CUBES or TDDFPT
      IF (.NOT. do_kpoints) THEN
         IF (compute_lumos) THEN
            DO ispin = 1, dft_control%nspins
               DEALLOCATE (unoccupied_evals(ispin)%array)
               IF (.NOT. dft_control%do_tddfpt_calculation) &
                  CALL cp_fm_release(unoccupied_orbs(ispin)%matrix)
            ENDDO
            DEALLOCATE (unoccupied_evals)
            IF (.NOT. dft_control%do_tddfpt_calculation) THEN
               DEALLOCATE (unoccupied_orbs)
            END IF
         ENDIF
      ENDIF

      !stm images
      IF (do_stm) THEN
         IF (do_kpoints) THEN
            CALL cp__w("qs_scf_post_gpw.F",661,"STM not implemented for k-point calculations!")
         ELSE
            IF (nlumo_stm > 0) THEN
               ALLOCATE (unoccupied_orbs_stm(dft_control%nspins))
               ALLOCATE (unoccupied_evals_stm(dft_control%nspins))
               CALL make_lumo(qs_env, scf_env, unoccupied_orbs_stm, unoccupied_evals_stm, &
                              nlumo_stm, nlumos)
            END IF

            CALL th_stm_image(qs_env, stm_section, particles, unoccupied_orbs_stm, &
                              unoccupied_evals_stm)

            IF (nlumo_stm > 0) THEN
               DO ispin = 1, dft_control%nspins
                  DEALLOCATE (unoccupied_evals_stm(ispin)%array)
                  CALL cp_fm_release(unoccupied_orbs_stm(ispin)%matrix)
               ENDDO
               DEALLOCATE (unoccupied_evals_stm)
               DEALLOCATE (unoccupied_orbs_stm)
            END IF
         END IF
      END IF

      ! Print coherent X-ray diffraction spectrum
      CALL qs_scf_post_xray(input, dft_section, logger, qs_env, output_unit)

      ! Calculation of Electric Field Gradients
      CALL qs_scf_post_efg(input, logger, qs_env)

      ! Calculation of ET
      CALL qs_scf_post_et(input, qs_env, dft_control)

      ! Calculation of EPR Hyperfine Coupling Tensors
      CALL qs_scf_post_epr(input, logger, qs_env)

      ! Calculation of properties needed for BASIS_MOLOPT optimizations
      CALL qs_scf_post_molopt(input, logger, qs_env)

      ! Calculate ELF
      CALL qs_scf_post_elf(input, logger, qs_env)

      ! Do active space calculation
      CALL active_space_main(input, logger, qs_env)

      CALL timestop(handle)

   END SUBROUTINE scf_post_calculation_gpw

! **************************************************************************************************
!> \brief Performs localization of the orbitals
!> \param qs_env ...
!> \param qs_loc_env ...
!> \param loc_section ...
!> \param mo_local ...
!> \param wf_r ...
!> \param wf_g ...
!> \param particles ...
!> \param coeff ...
!> \param evals ...
!> \param marked_states ...
! **************************************************************************************************
   SUBROUTINE get_localization_info(qs_env, qs_loc_env, loc_section, mo_local, &
                                    wf_r, wf_g, particles, coeff, evals, marked_states)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: loc_section
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_local
      TYPE(pw_p_type)                                    :: wf_r, wf_g
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: coeff
      TYPE(cp_1d_r_p_type), DIMENSION(:), POINTER        :: evals
      INTEGER, DIMENSION(:, :, :), POINTER               :: marked_states

      CHARACTER(len=*), PARAMETER :: routineN = 'get_localization_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, mystate, ns, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: lstates, marked_states_spin
      LOGICAL                                            :: do_homo
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: scenter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(section_vals_type), POINTER                   :: loc_print_section
      TYPE(wannier_centres_type), DIMENSION(:), POINTER  :: wc

      CALL timeset(routineN, handle)
      NULLIFY (mos, ks_rmpv, dft_control, loc_print_section, marked_states_spin, &
               matrix_s, molecule_set, scenter, wc)
      CALL get_qs_env(qs_env, mos=mos, matrix_ks=ks_rmpv, molecule_set=molecule_set, &
                      dft_control=dft_control, matrix_s=matrix_s)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      loc_print_section => section_vals_get_subs_vals(loc_section, "PRINT")
      do_homo = qs_loc_env%localized_wfn_control%do_homo
      IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                           "WANNIER_STATES"), cp_p_file)) THEN
         CALL get_qs_env(qs_env=qs_env, WannierCentres=wc)
         IF (.NOT. ASSOCIATED(wc)) THEN
            ALLOCATE (wc(dft_control%nspins))
         ENDIF
      ENDIF
      DO ispin = 1, dft_control%nspins
         !
         IF (do_homo) THEN
            qs_loc_env%tag_mo = "HOMO"
         ELSE
            qs_loc_env%tag_mo = "LUMO"
         END IF

         IF (qs_loc_env%do_localize) THEN
            ! Do the Real localization..
            IF (output_unit > 0 .AND. do_homo) WRITE (output_unit, "(/,T2,A,I3)") &
               "LOCALIZATION| Computing localization properties "// &
               "for OCCUPIED ORBITALS. Spin:", ispin
            IF (output_unit > 0 .AND. (.NOT. do_homo)) WRITE (output_unit, "(/,T2,A,I3)") &
               "LOCALIZATION| Computing localization properties "// &
               "for UNOCCUPIED ORBITALS. Spin:", ispin

            scenter => qs_loc_env%localized_wfn_control%centers_set(ispin)%array

            CALL qs_loc_driver(qs_env, qs_loc_env, loc_print_section, &
                               myspin=ispin, ext_mo_coeff=mo_local(ispin)%matrix)

            ! maps wfc to molecules, and compute the molecular dipoles if required
            IF ((BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                                  "MOLECULAR_DIPOLES"), cp_p_file) .OR. &
                 BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                                  "MOLECULAR_STATES"), cp_p_file))) THEN
               CALL wfc_to_molecule(qs_env, qs_loc_env, loc_print_section, scenter, &
                                    molecule_set, dft_control%nspins)
            END IF

            ! Compute the wannier states
            IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                                 "WANNIER_STATES"), cp_p_file)) THEN
               ns = SIZE(qs_loc_env%localized_wfn_control%loc_states, 1)
               IF (.NOT. ASSOCIATED(wc(ispin)%centres)) THEN
                  ALLOCATE (wc(ispin)%WannierHamDiag(ns))
                  ALLOCATE (wc(ispin)%centres(3, ns))
               ENDIF

               wc(ispin)%centres(:, :) = scenter(1+(ispin-1)*3:ispin*3, :)
               lstates => qs_loc_env%localized_wfn_control%loc_states(:, ispin)
               CALL construct_wannier_states(mo_local(ispin)%matrix, &
                                             ks_rmpv(ispin)%matrix, qs_env, loc_print_section=loc_print_section, &
                                             WannierCentres=wc(ispin), ns=ns, states=lstates)
            ENDIF
            ! Compute the molecular states
            IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                                 "MOLECULAR_STATES"), cp_p_file)) THEN
               CALL construct_molecular_states( &
                  molecule_set, mo_local(ispin)%matrix, coeff(ispin)%matrix, &
                  evals(ispin)%array, ks_rmpv(ispin)%matrix, matrix_s(1)%matrix, qs_env, wf_r, wf_g, &
                  loc_print_section=loc_print_section, particles=particles, tag=TRIM(qs_loc_env%tag_mo), &
                  marked_states=marked_states_spin)
               IF (ASSOCIATED(marked_states_spin)) THEN
                  IF (.NOT. ASSOCIATED(marked_states)) THEN
                     ALLOCATE (marked_states(SIZE(marked_states_spin), dft_control%nspins, 2))
                  END IF
                  mystate = 1
                  IF (qs_loc_env%tag_mo == "LUMO") mystate = 2
                  marked_states(:, ispin, mystate) = marked_states_spin(:)
                  DEALLOCATE (marked_states_spin)
               END IF
            ENDIF
         END IF
      ENDDO
      IF (BTEST(cp_print_key_should_output(logger%iter_info, loc_print_section, &
                                           "WANNIER_STATES"), cp_p_file)) THEN
         CALL set_qs_env(qs_env=qs_env, WannierCentres=wc)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE get_localization_info

! **************************************************************************************************
!> \brief Gets the lumos, and eigenvalues for the lumos
!> \param qs_env ...
!> \param scf_env ...
!> \param unoccupied_orbs ...
!> \param unoccupied_evals ...
!> \param nlumo ...
!> \param nlumos ...
! **************************************************************************************************
   SUBROUTINE make_lumo(qs_env, scf_env, unoccupied_orbs, unoccupied_evals, nlumo, nlumos)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: unoccupied_orbs
      TYPE(cp_1d_r_p_type), DIMENSION(:), POINTER        :: unoccupied_evals
      INTEGER                                            :: nlumo
      INTEGER, INTENT(OUT)                               :: nlumos

      CHARACTER(len=*), PARAMETER :: routineN = 'make_lumo', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, ispin, n, nao, nmo, &
                                                            output_unit
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(preconditioner_type), POINTER                 :: local_preconditioner
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      NULLIFY (mos, ks_rmpv, scf_control, dft_control, admm_env, para_env, blacs_env)
      CALL get_qs_env(qs_env, &
                      mos=mos, &
                      matrix_ks=ks_rmpv, &
                      scf_control=scf_control, &
                      dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      admm_env=admm_env, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      DO ispin = 1, dft_control%nspins
         NULLIFY (unoccupied_orbs(ispin)%matrix)
         NULLIFY (unoccupied_evals(ispin)%array)
         ! Always write eigenvalues
         IF (output_unit > 0) WRITE (output_unit, *) " "
         IF (output_unit > 0) WRITE (output_unit, *) " Lowest Eigenvalues of the unoccupied subspace spin ", ispin
         IF (output_unit > 0) WRITE (output_unit, FMT='(1X,A)') "-----------------------------------------------------"
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, homo=homo, nao=nao, nmo=nmo)
         CALL cp_fm_get_info(mo_coeff, nrow_global=n)
         nlumos = MAX(1, MIN(nlumo, nao-nmo))
         IF (nlumo == -1) nlumos = nao-nmo
         ALLOCATE (unoccupied_evals(ispin)%array(nlumos))
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=n, ncol_global=nlumos)
         CALL cp_fm_create(unoccupied_orbs(ispin)%matrix, fm_struct_tmp, name="lumos")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_init_random(unoccupied_orbs(ispin)%matrix, nlumos)

         ! the full_all preconditioner makes not much sense for lumos search
         NULLIFY (local_preconditioner)
         IF (ASSOCIATED(scf_env%ot_preconditioner)) THEN
            local_preconditioner => scf_env%ot_preconditioner(1)%preconditioner
            ! this one can for sure not be right (as it has to match a given C0)
            IF (local_preconditioner%in_use == ot_precond_full_all) THEN
               NULLIFY (local_preconditioner)
            ENDIF
         ENDIF

         ! ** If we do ADMM, we add have to modify the kohn-sham matrix
         IF (dft_control%do_admm) THEN
            CALL admm_correct_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
         END IF

         CALL ot_eigensolver(matrix_h=ks_rmpv(ispin)%matrix, matrix_s=matrix_s(1)%matrix, &
                             matrix_c_fm=unoccupied_orbs(ispin)%matrix, &
                             matrix_orthogonal_space_fm=mo_coeff, &
                             eps_gradient=scf_control%eps_lumos, &
                             preconditioner=local_preconditioner, &
                             iter_max=scf_control%max_iter_lumos, &
                             size_ortho_space=nmo)

         CALL calculate_subspace_eigenvalues(unoccupied_orbs(ispin)%matrix, ks_rmpv(ispin)%matrix, &
                                             unoccupied_evals(ispin)%array, scr=output_unit, &
                                             ionode=output_unit > 0)

         ! ** If we do ADMM, we restore the original kohn-sham matrix
         IF (dft_control%do_admm) THEN
            CALL admm_uncorrect_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
         END IF

      END DO

      CALL timestop(handle)

   END SUBROUTINE make_lumo
! **************************************************************************************************
!> \brief Computes and Prints Atomic Charges with several methods
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE qs_scf_post_charges(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_charges', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, print_level, unit_nr
      LOGICAL                                            :: do_kpoints, print_it
      TYPE(section_vals_type), POINTER                   :: density_fit_section, print_key

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, do_kpoints=do_kpoints)

      ! Mulliken charges require no further computation and are printed from write_mo_free_results

      ! Compute the Lowdin charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%LOWDIN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%LOWDIN", extension=".lowdin", &
                                        log_filename=.FALSE.)
         print_level = 1
         CALL section_vals_val_get(print_key, "PRINT_GOP", l_val=print_it)
         IF (print_it) print_level = 2
         CALL section_vals_val_get(print_key, "PRINT_ALL", l_val=print_it)
         IF (print_it) print_level = 3
         IF (do_kpoints) THEN
            CALL cp__w("qs_scf_post_gpw.F",981,"Lowdin charges not implemented for k-point calculations!")
         ELSE
            CALL lowdin_population_analysis(qs_env, unit_nr, print_level)
         END IF
         CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%LOWDIN")
      END IF

      ! Compute the RESP charges
      CALL resp_fit(qs_env)

      ! Compute the Density Derived Atomic Point charges with the Bloechl scheme
      print_key => section_vals_get_subs_vals(input, "PROPERTIES%FIT_CHARGE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, input, "PROPERTIES%FIT_CHARGE", extension=".Fitcharge", &
                                        log_filename=.FALSE.)
         density_fit_section => section_vals_get_subs_vals(input, "DFT%DENSITY_FITTING")
         CALL get_ddapc(qs_env, .FALSE., density_fit_section, iwc=unit_nr)
         CALL cp_print_key_finished_output(unit_nr, logger, input, "PROPERTIES%FIT_CHARGE")
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_charges

! **************************************************************************************************
!> \brief Computes and prints the Dipole (using localized charges)
!> \param input ...
!> \param dft_control ...
!> \param qs_loc_env ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE qs_scf_post_loc_dip(input, dft_control, qs_loc_env, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_loc_dip', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description, descriptionThisDip, iter
      COMPLEX(KIND=dp)                                   :: zeta
      COMPLEX(KIND=dp), DIMENSION(3)                     :: ggamma, zphase
      INTEGER                                            :: handle, i, ikind, ispins, j, n_rep, &
                                                            reference, unit_nr
      LOGICAL                                            :: do_berry, first_time, ghost
      REAL(KIND=dp)                                      :: charge_tot, theta, zeff, zwfc
      REAL(KIND=dp), DIMENSION(3)                        :: ci, dipole, dipole_old, gvec, rcc, ria
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(input, "DFT%LOCALIZE%PRINT%TOTAL_DIPOLE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key, first_time=first_time) &
                , cp_p_file)) THEN
         NULLIFY (cell, particle_set, qs_kind_set, ref_point, results)
         CALL get_qs_env(qs_env=qs_env, &
                         cell=cell, &
                         particle_set=particle_set, &
                         qs_kind_set=qs_kind_set, &
                         results=results)

         reference = section_get_ival(print_key, keyword_name="REFERENCE")
         CALL section_vals_val_get(print_key, "REF_POINT", r_vals=ref_point)
         CALL section_vals_val_get(print_key, "PERIODIC", l_val=do_berry)
         description = '[DIPOLE]'
         descriptionThisDip = '[TOTAL_DIPOLE]'
         CALL get_reference_point(rcc, qs_env=qs_env, reference=reference, ref_point=ref_point)

         dipole = 0.0_dp
         IF (do_berry) THEN
            rcc = pbc(rcc, cell)
            charge_tot = REAL(dft_control%charge, KIND=dp)
            ria = twopi*MATMUL(cell%h_inv, rcc)
            zphase = CMPLX(COS(ria), SIN(ria), KIND=dp)**charge_tot
            ggamma = CMPLX(1.0_dp, 0.0_dp, KIND=dp)

            ! Nuclear charges
            DO i = 1, SIZE(particle_set)
               CALL get_atomic_kind(particle_set(i)%atomic_kind, kind_number=ikind)
               CALL get_qs_kind(qs_kind_set(ikind), ghost=ghost)
               IF (.NOT. ghost) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), core_charge=zeff)
                  ria = pbc(particle_set(i)%r, cell)
                  DO j = 1, 3
                     gvec = twopi*cell%h_inv(j, :)
                     theta = SUM(ria(:)*gvec(:))
                     zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(zeff)
                     ggamma(j) = ggamma(j)*zeta
                  END DO
               END IF
            END DO

            ! Charges of the wfc involved
            ! Warning, this assumes the same occupation for all states
            zwfc = 3.0_dp-REAL(dft_control%nspins, dp)

            DO ispins = 1, dft_control%nspins
               DO i = 1, SIZE(qs_loc_env%localized_wfn_control%centers_set(ispins)%array, 2)
                  ria = pbc(qs_loc_env%localized_wfn_control%centers_set(ispins)%array(1:3, i), cell)
                  DO j = 1, 3
                     gvec = twopi*cell%h_inv(j, :)
                     theta = SUM(ria(:)*gvec(:))
                     zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(-zwfc)
                     ggamma(j) = ggamma(j)*zeta
                  END DO
               ENDDO
            ENDDO
            ggamma = ggamma*zphase
            ci = AIMAG(LOG(ggamma))/twopi
            dipole = MATMUL(cell%hmat, ci)
         ELSE
            ! Charges of the atoms involved
            DO i = 1, SIZE(particle_set)
               CALL get_atomic_kind(particle_set(i)%atomic_kind, kind_number=ikind)
               CALL get_qs_kind(qs_kind_set(ikind), ghost=ghost)
               IF (.NOT. ghost) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), core_charge=zeff)
                  ria = pbc(particle_set(i)%r, cell)
                  dipole = dipole+zeff*(ria-rcc)
               END IF
            END DO

            ! Charges of the wfc involved
            ! Warning, this assumes the same occupation for all states
            zwfc = 3.0_dp-REAL(dft_control%nspins, dp)

            DO ispins = 1, dft_control%nspins
               DO i = 1, SIZE(qs_loc_env%localized_wfn_control%centers_set(ispins)%array, 2)
                  ria = pbc(qs_loc_env%localized_wfn_control%centers_set(ispins)%array(1:3, i), cell)
                  dipole = dipole-zwfc*(ria-rcc)
               ENDDO
            ENDDO
         END IF

         ! Print and possibly store results
         unit_nr = cp_print_key_unit_nr(logger, print_key, extension=".Dipole", &
                                        middle_name="TOTAL_DIPOLE")
         IF (unit_nr > 0) THEN
            IF (first_time) THEN
               WRITE (unit=unit_nr, fmt="(A,T31,A,T88,A,T136,A)") &
                  "# iter_level", "dipole(x,y,z)[atomic units]", &
                  "dipole(x,y,z)[debye]", &
                  "delta_dipole(x,y,z)[atomic units]"
            END IF
            iter = cp_iter_string(logger%iter_info)
            CALL get_results(results, descriptionThisDip, n_rep=n_rep)
            IF (n_rep == 0) THEN
               dipole_old = 0._dp
            ELSE
               CALL get_results(results, descriptionThisDip, dipole_old, nval=n_rep)
            END IF
            IF (do_berry) THEN
               WRITE (unit=unit_nr, fmt="(a,9(es18.8))") &
                  iter(1:15), dipole, dipole*debye, pbc(dipole-dipole_old, cell)
            ELSE
               WRITE (unit=unit_nr, fmt="(a,9(es18.8))") &
                  iter(1:15), dipole, dipole*debye, (dipole-dipole_old)
            END IF
         END IF
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
         CALL cp_results_erase(results, description)
         CALL put_results(results, description, dipole)
         CALL cp_results_erase(results, descriptionThisDip)
         CALL put_results(results, descriptionThisDip, dipole)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_loc_dip

! **************************************************************************************************
!> \brief Computes and prints the Cube Files for MO
!> \param input ...
!> \param dft_section ...
!> \param dft_control ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param mo_coeff ...
!> \param wf_g ...
!> \param wf_r ...
!> \param particles ...
!> \param homo ...
!> \param ispin ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_occ_cubes(input, dft_section, dft_control, logger, qs_env, &
                                    mo_coeff, wf_g, wf_r, particles, homo, ispin)
      TYPE(section_vals_type), POINTER                   :: input, dft_section
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: homo, ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_occ_cubes', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename, my_pos_cube, title
      INTEGER                                            :: handle, ivector, nhomo, unit_nr
      LOGICAL                                            :: append_cube
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, dft_section, "PRINT%MO_CUBES") &
                , cp_p_file) .AND. section_get_lval(dft_section, "PRINT%MO_CUBES%WRITE_CUBE")) THEN
         nhomo = section_get_ival(dft_section, "PRINT%MO_CUBES%NHOMO")
         append_cube = section_get_lval(dft_section, "PRINT%MO_CUBES%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         IF (nhomo == -1) nhomo = homo
         DO ivector = MAX(1, homo-nhomo+1), homo
            CALL get_qs_env(qs_env=qs_env, &
                            atomic_kind_set=atomic_kind_set, &
                            qs_kind_set=qs_kind_set, &
                            cell=cell, &
                            particle_set=particle_set, &
                            pw_env=pw_env)
            CALL calculate_wavefunction(mo_coeff, ivector, wf_r, wf_g, atomic_kind_set, qs_kind_set, &
                                        cell, dft_control, particle_set, pw_env)
            WRITE (filename, '(a4,I5.5,a1,I1.1)') "WFN_", ivector, "_", ispin
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MO_CUBES", extension=".cube", &
                                           middle_name=TRIM(filename), file_position=my_pos_cube, log_filename=.FALSE.)
            WRITE (title, *) "WAVEFUNCTION ", ivector, " spin ", ispin, " i.e. HOMO - ", ivector-homo
            CALL cp_pw_to_cube(wf_r%pw, unit_nr, title, particles=particles, &
                               stride=section_get_ivals(dft_section, "PRINT%MO_CUBES%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%MO_CUBES")
         ENDDO
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_occ_cubes

! **************************************************************************************************
!> \brief Computes and prints the Cube Files for MO
!> \param input ...
!> \param dft_section ...
!> \param dft_control ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param unoccupied_orbs ...
!> \param wf_g ...
!> \param wf_r ...
!> \param particles ...
!> \param nlumos ...
!> \param homo ...
!> \param ispin ...
!> \param lumo ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_unocc_cubes(input, dft_section, dft_control, logger, qs_env, &
                                      unoccupied_orbs, wf_g, wf_r, particles, nlumos, homo, ispin, lumo)
      TYPE(section_vals_type), POINTER                   :: input, dft_section
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_type), POINTER                          :: unoccupied_orbs
      TYPE(pw_p_type)                                    :: wf_g, wf_r
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: nlumos, homo, ispin
      INTEGER, INTENT(IN), OPTIONAL                      :: lumo

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_unocc_cubes', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename, my_pos_cube, title
      INTEGER                                            :: handle, ifirst, index_mo, ivector, &
                                                            unit_nr
      LOGICAL                                            :: append_cube
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, dft_section, "PRINT%MO_CUBES"), cp_p_file) &
          .AND. section_get_lval(dft_section, "PRINT%MO_CUBES%WRITE_CUBE")) THEN
         NULLIFY (qs_kind_set, particle_set, pw_env, cell)
         append_cube = section_get_lval(dft_section, "PRINT%MO_CUBES%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         ifirst = 1
         IF (PRESENT(lumo)) ifirst = lumo
         DO ivector = ifirst, ifirst+nlumos-1
            CALL get_qs_env(qs_env=qs_env, &
                            atomic_kind_set=atomic_kind_set, &
                            qs_kind_set=qs_kind_set, &
                            cell=cell, &
                            particle_set=particle_set, &
                            pw_env=pw_env)
            CALL calculate_wavefunction(unoccupied_orbs, ivector, wf_r, wf_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env)

            IF (ifirst == 1) THEN
               index_mo = homo+ivector
            ELSE
               index_mo = ivector
            END IF
            WRITE (filename, '(a4,I5.5,a1,I1.1)') "WFN_", index_mo, "_", ispin

            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MO_CUBES", extension=".cube", &
                                           middle_name=TRIM(filename), file_position=my_pos_cube, log_filename=.FALSE.)
            WRITE (title, *) "WAVEFUNCTION ", index_mo, " spin ", ispin, " i.e. LUMO + ", ifirst+ivector-2
            CALL cp_pw_to_cube(wf_r%pw, unit_nr, title, particles=particles, &
                               stride=section_get_ivals(dft_section, "PRINT%MO_CUBES%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%MO_CUBES")
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_unocc_cubes

! **************************************************************************************************
!> \brief Computes and prints electric moments
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_moments(input, logger, qs_env, output_unit)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_moments', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: handle, maxmom, reference, unit_nr
      LOGICAL                                            :: do_kpoints, magnetic, periodic
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%MOMENTS")

      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN

         maxmom = section_get_ival(section_vals=input, &
                                   keyword_name="DFT%PRINT%MOMENTS%MAX_MOMENT")
         periodic = section_get_lval(section_vals=input, &
                                     keyword_name="DFT%PRINT%MOMENTS%PERIODIC")
         reference = section_get_ival(section_vals=input, &
                                      keyword_name="DFT%PRINT%MOMENTS%REFERENCE")
         magnetic = section_get_lval(section_vals=input, &
                                     keyword_name="DFT%PRINT%MOMENTS%MAGNETIC")
         NULLIFY (ref_point)
         CALL section_vals_val_get(input, "DFT%PRINT%MOMENTS%REF_POINT", r_vals=ref_point)
         unit_nr = cp_print_key_unit_nr(logger=logger, basis_section=input, &
                                        print_key_path="DFT%PRINT%MOMENTS", extension=".dat", &
                                        middle_name="moments", log_filename=.FALSE.)

         IF (output_unit > 0) THEN
            IF (unit_nr /= output_unit) THEN
               INQUIRE (UNIT=unit_nr, NAME=filename)
               WRITE (UNIT=output_unit, FMT="(/,T2,A,2(/,T3,A),/)") &
                  "MOMENTS", "The electric/magnetic moments are written to file:", &
                  TRIM(filename)
            ELSE
               WRITE (UNIT=output_unit, FMT="(/,T2,A)") "ELECTRIC/MAGNETIC MOMENTS"
            END IF
         END IF

         CALL get_qs_env(qs_env, do_kpoints=do_kpoints)

         IF (do_kpoints) THEN
            CALL cp__w("qs_scf_post_gpw.F",1370,"Moments not implemented for k-point calculations!")
         ELSE
            IF (periodic) THEN
               CALL qs_moment_berry_phase(qs_env, magnetic, maxmom, reference, ref_point, unit_nr)
            ELSE
               CALL qs_moment_locop(qs_env, magnetic, maxmom, reference, ref_point, unit_nr)
            END IF
         END IF

         CALL cp_print_key_finished_output(unit_nr=unit_nr, logger=logger, &
                                           basis_section=input, print_key_path="DFT%PRINT%MOMENTS")
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_moments

! **************************************************************************************************
!> \brief Computes and prints the X-ray diffraction spectrum.
!> \param input ...
!> \param dft_section ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_xray(input, dft_section, logger, qs_env, output_unit)

      TYPE(section_vals_type), POINTER                   :: input, dft_section
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_post_xray', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: handle, unit_nr
      REAL(KIND=dp)                                      :: q_max
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%XRAY_DIFFRACTION_SPECTRUM")

      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         q_max = section_get_rval(section_vals=dft_section, &
                                  keyword_name="PRINT%XRAY_DIFFRACTION_SPECTRUM%Q_MAX")
         unit_nr = cp_print_key_unit_nr(logger=logger, &
                                        basis_section=input, &
                                        print_key_path="DFT%PRINT%XRAY_DIFFRACTION_SPECTRUM", &
                                        extension=".dat", &
                                        middle_name="xrd", &
                                        log_filename=.FALSE.)
         IF (output_unit > 0) THEN
            INQUIRE (UNIT=unit_nr, NAME=filename)
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") &
               "X-RAY DIFFRACTION SPECTRUM"
            IF (unit_nr /= output_unit) THEN
               WRITE (UNIT=output_unit, FMT="(/,T3,A,/,/,T3,A,/)") &
                  "The coherent X-ray diffraction spectrum is written to the file:", &
                  TRIM(filename)
            END IF
         END IF
         CALL xray_diffraction_spectrum(qs_env=qs_env, &
                                        unit_number=unit_nr, &
                                        q_max=q_max)
         CALL cp_print_key_finished_output(unit_nr=unit_nr, &
                                           logger=logger, &
                                           basis_section=input, &
                                           print_key_path="DFT%PRINT%XRAY_DIFFRACTION_SPECTRUM")
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_xray

! **************************************************************************************************
!> \brief Computes and prints Electric Field Gradient
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE qs_scf_post_efg(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_efg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%ELECTRIC_FIELD_GRADIENT")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN
         CALL qs_efg_calc(qs_env=qs_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_efg

! **************************************************************************************************
!> \brief Computes the Electron Transfer Coupling matrix element
!> \param input ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param dft_control ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_et(input, qs_env, dft_control)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_et', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      LOGICAL                                            :: do_et
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: my_mos
      TYPE(section_vals_type), POINTER                   :: et_section

      CALL timeset(routineN, handle)

      do_et = .FALSE.
      et_section => section_vals_get_subs_vals(input, "PROPERTIES%ET_COUPLING")
      CALL section_vals_get(et_section, explicit=do_et)
      IF (do_et) THEN
         IF (qs_env%et_coupling%first_run) THEN
            NULLIFY (my_mos)
            ALLOCATE (my_mos(dft_control%nspins))
            ALLOCATE (qs_env%et_coupling%et_mo_coeff(dft_control%nspins))
            DO ispin = 1, dft_control%nspins
               NULLIFY (my_mos(ispin)%matrix)
               CALL cp_fm_create(matrix=my_mos(ispin)%matrix, &
                                 matrix_struct=qs_env%mos(ispin)%mo_set%mo_coeff%matrix_struct, &
                                 name="FIRST_RUN_COEFF"//TRIM(ADJUSTL(cp_to_string(ispin)))//"MATRIX")
               CALL cp_fm_to_fm(qs_env%mos(ispin)%mo_set%mo_coeff, &
                                my_mos(ispin)%matrix)
            END DO
            CALL set_et_coupling_type(qs_env%et_coupling, et_mo_coeff=my_mos)
            DEALLOCATE (my_mos)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_et

! **************************************************************************************************
!> \brief compute the electron localization function
!>
!> \param input ...
!> \param logger ...
!> \param qs_env ...
!> \par History
!>      2012-07 Created [MI]
! **************************************************************************************************
   SUBROUTINE qs_scf_post_elf(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_elf', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename, my_pos_cube, title
      INTEGER                                            :: handle, ispin, output_unit, unit_nr
      LOGICAL                                            :: append_cube, do_elf, gapw
      REAL(dp)                                           :: rho_cutoff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: elf_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: elf_section

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit(logger)

      do_elf = .FALSE.
      elf_section => section_vals_get_subs_vals(input, "DFT%PRINT%ELF_CUBE")
      CALL section_vals_get(elf_section, explicit=do_elf)
      IF (do_elf) THEN
         NULLIFY (dft_control, pw_env, auxbas_pw_pool, pw_pools, particles, subsys, elf_r)
         CALL get_qs_env(qs_env, dft_control=dft_control, pw_env=pw_env, subsys=subsys)
         CALL qs_subsys_get(subsys, particles=particles)

         gapw = dft_control%qs_control%gapw
         IF (.NOT. gapw) THEN
            ! allocate
            ALLOCATE (elf_r(dft_control%nspins))
            CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                            pw_pools=pw_pools)
            DO ispin = 1, dft_control%nspins
               CALL pw_pool_create_pw(auxbas_pw_pool, elf_r(ispin)%pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_zero(elf_r(ispin)%pw)
            END DO

            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,T15,A,/)") &
                  " ----- ELF is computed on the real space grid -----"
            END IF
            rho_cutoff = section_get_rval(elf_section, "density_cutoff")
            CALL qs_elf_calc(qs_env, elf_r, rho_cutoff)

            ! write ELF into cube file
            append_cube = section_get_lval(elf_section, "APPEND")
            my_pos_cube = "REWIND"
            IF (append_cube) THEN
               my_pos_cube = "APPEND"
            END IF

            DO ispin = 1, dft_control%nspins
               WRITE (filename, '(a5,I1.1)') "ELF_S", ispin
               WRITE (title, *) "ELF spin ", ispin
               unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%ELF_CUBE", extension=".cube", &
                                              middle_name=TRIM(filename), file_position=my_pos_cube, log_filename=.FALSE.)
               IF (output_unit > 0) THEN
                  INQUIRE (UNIT=unit_nr, NAME=filename)
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                     "ELF is written in cube file format to the file:", &
                     TRIM(filename)
               END IF

               CALL cp_pw_to_cube(elf_r(ispin)%pw, unit_nr, title, particles=particles, &
                                  stride=section_get_ivals(elf_section, "STRIDE"))
               CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%ELF_CUBE")

               CALL pw_pool_give_back_pw(auxbas_pw_pool, elf_r(ispin)%pw)
            END DO

            ! deallocate
            DEALLOCATE (elf_r)

         ELSE
            ! not implemented
            CALL cp__w("qs_scf_post_gpw.F",1614,"ELF not implemented for GAPW calculations!!")

         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_elf

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param elf_r ...
!> \param rho_cutoff ...
! **************************************************************************************************
   SUBROUTINE qs_elf_calc(qs_env, elf_r, rho_cutoff)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: elf_r
      REAL(kind=dp), INTENT(IN)                          :: rho_cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_elf_calc', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: ELFCUT = 0.0001_dp, &
                                                            f18 = (1.0_dp/8.0_dp), &
                                                            f23 = (2.0_dp/3.0_dp), &
                                                            f53 = (5.0_dp/3.0_dp)

      INTEGER                                            :: handle, i, idir, ispin, j, k, nspin
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, DIMENSION(3, 3)                           :: nd
      LOGICAL                                            :: deriv_pw, drho_r_valid, tau_r_valid
      REAL(kind=dp)                                      :: cfermi, dum, elf_kernel, norm_drho, &
                                                            rho_53, udvol
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_struct_ao
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: drho_g, drho_r, drho_struct_r, rho_r, &
                                                            rho_struct_r, tau_g, tau_r, &
                                                            tau_struct_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: tmp_g
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho_struct

! for spin restricted systems

      CALL timeset(routineN, handle)

      NULLIFY (rho_struct, rho_r, drho_g, drho_r, tau_r, tau_g, pw_env, auxbas_pw_pool, pw_pools, tmp_g, ks_env)
      NULLIFY (rho_struct_ao, rho_struct_r, tau_struct_r, drho_struct_r)

      CALL get_qs_env(qs_env, ks_env=ks_env, pw_env=pw_env, rho=rho_struct)

      CALL qs_rho_get(rho_struct, &
                      rho_ao_kp=rho_struct_ao, &
                      rho_r=rho_struct_r, &
                      tau_r=tau_struct_r, &
                      drho_r=drho_struct_r, &
                      tau_r_valid=tau_r_valid, &
                      drho_r_valid=drho_r_valid)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools)
      nspin = SIZE(rho_struct_r)
      bo = rho_struct_r(1)%pw%pw_grid%bounds_local
      cfermi = (3.0_dp/10.0_dp)*(pi*pi*3.0_dp)**f23

      ALLOCATE (rho_r(nspin))
      ALLOCATE (tau_r(nspin))
      ALLOCATE (tau_g(nspin))
      ALLOCATE (drho_r(3*nspin))
      ALLOCATE (drho_g(3*nspin))

      DO ispin = 1, nspin
         rho_r(ispin)%pw => rho_struct_r(ispin)%pw
         IF (tau_r_valid) THEN
            tau_r(ispin)%pw => tau_struct_r(ispin)%pw
         ELSE
            CALL pw_pool_create_pw(auxbas_pw_pool, tau_r(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(auxbas_pw_pool, tau_g(ispin)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            rho_ao => rho_struct_ao(ispin, :)
            CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                    rho=tau_r(ispin), &
                                    rho_gspace=tau_g(ispin), &
                                    total_rho=dum, &
                                    ks_env=ks_env, soft_valid=.FALSE., &
                                    compute_tau=.TRUE.)
         END IF

         IF (drho_r_valid) THEN
            DO idir = 1, 3
               drho_r(3*(ispin-1)+idir)%pw => drho_struct_r(3*(ispin-1)+idir)%pw
            END DO
         ELSE
            deriv_pw = .FALSE.
            !        deriv_pw = .TRUE.
            IF (deriv_pw) THEN
               nd = RESHAPE((/1, 0, 0, 0, 1, 0, 0, 0, 1/), (/3, 3/))
               CALL pw_pool_create_pw(auxbas_pw_pool, tmp_g, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               udvol = 1.0_dp/rho_struct_r(ispin)%pw%pw_grid%dvol
               DO idir = 1, 3
                  CALL pw_pool_create_pw(auxbas_pw_pool, drho_r(3*(ispin-1)+idir)%pw, &
                                         use_data=REALDATA3D, in_space=REALSPACE)
                  CALL pw_transfer(rho_struct_r(ispin)%pw, tmp_g)
                  CALL pw_derive(tmp_g, nd(:, idir))
                  CALL pw_transfer(tmp_g, drho_r(3*(ispin-1)+idir)%pw)
!              CALL pw_scale(drho_r(3*(ispin-1)+idir)%pw,udvol)

               END DO

            ELSE
               DO idir = 1, 3
                  CALL pw_pool_create_pw(auxbas_pw_pool, drho_r(3*(ispin-1)+idir)%pw, &
                                         use_data=REALDATA3D, in_space=REALSPACE)
                  CALL pw_pool_create_pw(auxbas_pw_pool, drho_g(3*(ispin-1)+idir)%pw, &
                                         use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                  rho_ao => rho_struct_ao(ispin, :)
                  CALL calculate_rho_elec(matrix_p_kp=rho_ao, &
                                          rho=drho_r(3*(ispin-1)+idir), &
                                          rho_gspace=drho_g(3*(ispin-1)+idir), &
                                          total_rho=dum, &
                                          ks_env=ks_env, soft_valid=.FALSE., &
                                          compute_tau=.FALSE., compute_grad=.TRUE., idir=idir)

               END DO
            END IF
         END IF

         ! Calculate elf_r
!$OMP        PARALLEL DO DEFAULT(NONE) SHARED(bo,elf_r, ispin, drho_r,rho_r, tau_r, cfermi, rho_cutoff)&
!$OMP                    PRIVATE(k,j,i, norm_drho, rho_53, elf_kernel)
         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  norm_drho = drho_r(3*(ispin-1)+1)%pw%cr3d(i, j, k)**2+ &
                              drho_r(3*(ispin-1)+2)%pw%cr3d(i, j, k)**2+ &
                              drho_r(3*(ispin-1)+3)%pw%cr3d(i, j, k)**2
                  norm_drho = norm_drho/MAX(rho_r(ispin)%pw%cr3d(i, j, k), rho_cutoff)
                  rho_53 = cfermi*MAX(rho_r(ispin)%pw%cr3d(i, j, k), rho_cutoff)**f53
                  elf_kernel = (tau_r(ispin)%pw%cr3d(i, j, k)-f18*norm_drho)+2.87E-5_dp
                  elf_kernel = (elf_kernel/rho_53)**2
                  elf_r(ispin)%pw%cr3d(i, j, k) = 1.0_dp/(1.0_dp+elf_kernel)
                  IF (elf_r(ispin)%pw%cr3d(i, j, k) < ELFCUT) elf_r(ispin)%pw%cr3d(i, j, k) = 0.0_dp
               END DO
            END DO
         END DO

         IF (.NOT. tau_r_valid) THEN
            CALL pw_pool_give_back_pw(auxbas_pw_pool, tau_r(ispin)%pw)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, tau_g(ispin)%pw)
         END IF
         IF (.NOT. drho_r_valid) THEN
            IF (deriv_pw) THEN
               CALL pw_pool_give_back_pw(auxbas_pw_pool, tmp_g)
               DO idir = 1, 3
                  CALL pw_pool_give_back_pw(auxbas_pw_pool, drho_r(3*(ispin-1)+idir)%pw)
               END DO
            ELSE
               DO idir = 1, 3
                  CALL pw_pool_give_back_pw(auxbas_pw_pool, drho_r(3*(ispin-1)+idir)%pw)
                  CALL pw_pool_give_back_pw(auxbas_pw_pool, drho_g(3*(ispin-1)+idir)%pw)
               END DO
            END IF
         END IF
      END DO !ispin

      DEALLOCATE (rho_r)
      DEALLOCATE (tau_r)
      DEALLOCATE (tau_g)
      DEALLOCATE (drho_r)
      DEALLOCATE (drho_g)

      CALL timestop(handle)

   END SUBROUTINE qs_elf_calc

! **************************************************************************************************
!> \brief computes the condition number of the overlap matrix and
!>      prints the value of the total energy. This is needed
!>      for BASIS_MOLOPT optimizations
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \par History
!>      2007-07 Created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE qs_scf_post_molopt(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_molopt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nao, unit_nr
      REAL(KIND=dp)                                      :: S_cond_number
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: ao_ao_fmstruct
      TYPE(cp_fm_type), POINTER                          :: fm_s, fm_work, mo_coeff
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%BASIS_MOLOPT_QUANTITIES")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN

         CALL get_qs_env(qs_env, energy=energy, matrix_s=matrix_s, mos=mos)

         ! set up the two needed full matrices, using mo_coeff as a template
         CALL get_mo_set(mo_set=mos(1)%mo_set, mo_coeff=mo_coeff, nao=nao)
         CALL cp_fm_struct_create(fmstruct=ao_ao_fmstruct, &
                                  nrow_global=nao, ncol_global=nao, &
                                  template_fmstruct=mo_coeff%matrix_struct)
         CALL cp_fm_create(fm_s, matrix_struct=ao_ao_fmstruct, &
                           name="fm_s")
         CALL cp_fm_create(fm_work, matrix_struct=ao_ao_fmstruct, &
                           name="fm_work")
         CALL cp_fm_struct_release(ao_ao_fmstruct)
         ALLOCATE (eigenvalues(nao))

         CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, fm_s)
         CALL choose_eigv_solver(fm_s, fm_work, eigenvalues)

         CALL cp_fm_release(fm_s)
         CALL cp_fm_release(fm_work)

         S_cond_number = MAXVAL(ABS(eigenvalues))/MAX(MINVAL(ABS(eigenvalues)), EPSILON(0.0_dp))

         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%BASIS_MOLOPT_QUANTITIES", &
                                        extension=".molopt")

         IF (unit_nr > 0) THEN
            ! please keep this format fixed, needs to be grepable for molopt
            ! optimizations
            WRITE (unit_nr, '(T2,A28,2A25)') "", "Tot. Ener.", "S Cond. Numb."
            WRITE (unit_nr, '(T2,A28,2E25.17)') "BASIS_MOLOPT_QUANTITIES", energy%total, S_cond_number
         ENDIF

         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%BASIS_MOLOPT_QUANTITIES")

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_molopt

! **************************************************************************************************
!> \brief Dumps EPR
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE qs_scf_post_epr(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_epr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)

      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%HYPERFINE_COUPLING_TENSOR")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN
         CALL qs_epr_hyp_calc(qs_env=qs_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_epr

! **************************************************************************************************
!> \brief Interface routine to trigger writing of results available from normal
!>        SCF. Can write MO-dependent and MO free results (needed for call from
!>        the linear scaling code)
!> \param qs_env the qs_env in which the qs_env lives
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE write_available_results(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_available_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ! those properties that require MOs (not suitable density matrix based methods)
      CALL write_mo_dependent_results(qs_env, scf_env)

      ! those that depend only on the density matrix, they should be linear scaling in their implementation
      CALL write_mo_free_results(qs_env)

      CALL timestop(handle)

   END SUBROUTINE write_available_results

! **************************************************************************************************
!> \brief Write QS results available if MO's are present (if switched on through the print_keys)
!>        Writes only MO dependent results. Split is necessary as ls_scf does not
!>        provide MO's
!> \param qs_env the qs_env in which the qs_env lives
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE write_mo_dependent_results(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_mo_dependent_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, ispin, nmo, output_unit
      LOGICAL                                            :: all_equal, do_kpoints
      REAL(KIND=dp)                                      :: maxocc, s_square, s_square_ideal, &
                                                            total_abs_spin_dens
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues, occupation_numbers
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_deriv
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_r
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, input

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, pw_env, auxbas_pw_pool, pw_pools, mo_coeff, &
               mo_coeff_deriv, mo_eigenvalues, mos, atomic_kind_set, qs_kind_set, &
               particle_set, rho, ks_rmpv, matrix_s, scf_control, dft_section, &
               molecule_set, input, particles, subsys, rho_r)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_gpw.F",1981)
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      molecule_set=molecule_set, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      admm_env=admm_env, &
                      scf_control=scf_control, &
                      input=input, &
                      cell=cell, &
                      subsys=subsys)
      CALL qs_subsys_get(subsys, particles=particles)
      CALL get_qs_env(qs_env, rho=rho)
      CALL qs_rho_get(rho, rho_r=rho_r)

      ! kpoints
      CALL get_qs_env(qs_env, do_kpoints=do_kpoints)

      ! *** if the dft_section tells you to do so, write last wavefunction to screen
      dft_section => section_vals_get_subs_vals(input, "DFT")
      IF (.NOT. qs_env%run_rtp) THEN
         IF (.NOT. do_kpoints) THEN
            CALL get_qs_env(qs_env, mos=mos)
            IF (dft_control%nspins == 2) THEN
               CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                                 dft_section, spin="ALPHA", last=.TRUE.)
               CALL write_mo_set(mos(2)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                                 dft_section, spin="BETA", last=.TRUE.)
            ELSE
               CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                                 dft_section, last=.TRUE.)
            END IF
            CALL get_qs_env(qs_env, matrix_ks=ks_rmpv)
            CALL write_dm_binary_restart(mos, dft_section, ks_rmpv)
         END IF

         ! *** at the end of scf print out the projected dos per kind
         IF (BTEST(cp_print_key_should_output(logger%iter_info, dft_section, "PRINT%PDOS") &
                   , cp_p_file)) THEN
            IF (do_kpoints) THEN
               CALL cp__w("qs_scf_post_gpw.F",2022,"Projected density of states not implemented for k-points.")
            ELSE
               CALL get_qs_env(qs_env, &
                               mos=mos, &
                               matrix_ks=ks_rmpv)
               DO ispin = 1, dft_control%nspins

                  ! ** If we do ADMM, we add have to modify the kohn-sham matrix
                  IF (dft_control%do_admm) THEN
                     CALL admm_correct_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
                  END IF
                  IF (PRESENT(scf_env)) THEN
                     IF (scf_env%method == ot_method_nr) THEN
                        CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                                        eigenvalues=mo_eigenvalues)
                        IF (ASSOCIATED(qs_env%mo_derivs)) THEN
                           mo_coeff_deriv => qs_env%mo_derivs(ispin)%matrix
                        ELSE
                           mo_coeff_deriv => NULL()
                        ENDIF

                        CALL calculate_subspace_eigenvalues(mo_coeff, ks_rmpv(ispin)%matrix, mo_eigenvalues, &
                                                            do_rotation=.TRUE., &
                                                            co_rotate_dbcsr=mo_coeff_deriv)
                        CALL set_mo_occupation(mo_set=mos(ispin)%mo_set)

                     END IF
                  END IF
                  IF (dft_control%nspins == 2) THEN
                     CALL calculate_projected_dos(mos(ispin)%mo_set, atomic_kind_set, &
                                                  qs_kind_set, particle_set, qs_env, dft_section, ispin=ispin)
                  ELSE
                     CALL calculate_projected_dos(mos(ispin)%mo_set, atomic_kind_set, &
                                                  qs_kind_set, particle_set, qs_env, dft_section)
                  END IF

                  ! ** If we do ADMM, we add have to modify the kohn-sham matrix
                  IF (dft_control%do_admm) THEN
                     CALL admm_uncorrect_for_eigenvalues(ispin, admm_env, ks_rmpv(ispin)%matrix)
                  END IF

               END DO
            ENDIF
         ENDIF
      END IF

      !   *** Integrated absolute spin density and spin contamination ***
      IF (dft_control%nspins .EQ. 2) THEN
         CALL get_qs_env(qs_env, mos=mos)
         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                         pw_pools=pw_pools)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_copy(rho_r(1)%pw, wf_r%pw)
         CALL pw_axpy(rho_r(2)%pw, wf_r%pw, alpha=-1._dp)
         total_abs_spin_dens = pw_integrate_function(wf_r%pw, oprt="ABS")
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,(T3,A,T61,F20.10))') &
            "Integrated absolute spin density  : ", total_abs_spin_dens
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
         !
         ! XXX Fix Me XXX
         ! should be extended to the case where added MOs are present
         ! should be extended to the k-point case
         !
         IF (do_kpoints) THEN
            CALL cp__w("qs_scf_post_gpw.F",2089,"Spin contamination estimate not implemented for k-points.")
         ELSE
            all_equal = .TRUE.
            DO ispin = 1, dft_control%nspins
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                               occupation_numbers=occupation_numbers, &
                               homo=homo, &
                               nmo=nmo, &
                               maxocc=maxocc)
               IF (nmo > 0) THEN
                  all_equal = all_equal .AND. &
                              (ALL(occupation_numbers(1:homo) == maxocc) .AND. &
                               ALL(occupation_numbers(homo+1:nmo) == 0.0_dp))
               END IF
            END DO
            IF (.NOT. all_equal) THEN
               IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(T3,A)") &
                  "WARNING: S**2 computation does not yet treat fractional occupied orbitals"
            ELSE
               CALL get_qs_env(qs_env=qs_env, &
                               matrix_s=matrix_s, &
                               energy=energy)
               CALL compute_s_square(mos=mos, matrix_s=matrix_s, s_square=s_square, &
                                     s_square_ideal=s_square_ideal)
               IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(T3,A,T51,2F15.6)') &
                  "Ideal and single determinant S**2 : ", s_square_ideal, s_square
               energy%s_square = s_square
            END IF
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE write_mo_dependent_results

! **************************************************************************************************
!> \brief Write QS results always available (if switched on through the print_keys)
!>        Can be called from ls_scf
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE write_mo_free_results(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_mo_free_results', &
         routineP = moduleN//':'//routineN
      CHARACTER(len=1), DIMENSION(3), PARAMETER          :: cdir = (/"x", "y", "z"/)

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_path_length)                 :: filename, my_pos_cube
      CHARACTER(LEN=default_string_length)               :: name
      INTEGER                                            :: after, handle, i, iat, id, ikind, img, &
                                                            iso, ispin, iw, l, nd(3), ngto, niso, &
                                                            nkind, np, nr, output_unit, &
                                                            print_level, unit_nr
      LOGICAL :: append_cube, do_kpoints, omit_headers, print_it, print_total_density, &
         rho_r_valid, write_ks, write_xc, xrd_interface
      REAL(KIND=dp)                                      :: q_max, rho_hard, rho_soft, rho_total, &
                                                            rho_total_rspace, udvol, volume
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: bfun
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: aedens, ccdens, ppdens
      REAL(KIND=dp), DIMENSION(3)                        :: dr
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_rmpv, matrix_vxc, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: aux_g, aux_r, rho_elec_gspace, &
                                                            rho_elec_rspace, wf_r
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_p_type), POINTER                           :: rho0_s_gs, rho_core, vee
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: v_hartree_rspace
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(section_vals_type), POINTER                   :: dft_section, input, print_key, xc_section

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, pw_env, auxbas_pw_pool, pw_pools, &
               atomic_kind_set, qs_kind_set, particle_set, rho, ks_rmpv, rho_ao, rho_r, &
               dft_section, xc_section, input, particles, subsys, matrix_vxc, v_hartree_rspace, vee)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_gpw.F",2184)
      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      cell=cell, &
                      para_env=para_env, &
                      dft_control=dft_control, &
                      input=input, &
                      do_kpoints=do_kpoints, &
                      subsys=subsys)
      dft_section => section_vals_get_subs_vals(input, "DFT")
      CALL qs_subsys_get(subsys, particles=particles)

      ! Print the total density (electronic + core charge)
      CALL get_qs_env(qs_env, rho=rho)
      CALL qs_rho_get(rho, rho_r=rho_r)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%TOT_DENSITY_CUBE"), cp_p_file)) THEN
         NULLIFY (rho_core, rho0_s_gs)
         append_cube = section_get_lval(input, "DFT%PRINT%TOT_DENSITY_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF

         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, rho_core=rho_core, &
                         rho0_s_gs=rho0_s_gs)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                         pw_pools=pw_pools)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         IF (dft_control%qs_control%gapw) THEN
            CALL pw_transfer(rho0_s_gs%pw, wf_r%pw)
            IF (dft_control%qs_control%gapw_control%nopaw_as_gpw) THEN
               CALL pw_axpy(rho_core%pw, wf_r%pw)
            END IF
         ELSE
            CALL pw_transfer(rho_core%pw, wf_r%pw)
         END IF
         DO ispin = 1, dft_control%nspins
            CALL pw_axpy(rho_r(ispin)%pw, wf_r%pw)
         END DO
         filename = "TOTAL_DENSITY"
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%TOT_DENSITY_CUBE", &
                                        extension=".cube", middle_name=TRIM(filename), file_position=my_pos_cube, &
                                        log_filename=.FALSE.)
         CALL cp_pw_to_cube(wf_r%pw, unit_nr, "TOTAL DENSITY", &
                            particles=particles, &
                            stride=section_get_ivals(dft_section, "PRINT%TOT_DENSITY_CUBE%STRIDE"))
         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%TOT_DENSITY_CUBE")
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      END IF

      ! Write cube file with electron density
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%E_DENSITY_CUBE"), cp_p_file)) THEN
         CALL section_vals_val_get(dft_section, &
                                   keyword_name="PRINT%E_DENSITY_CUBE%TOTAL_DENSITY", &
                                   l_val=print_total_density)
         append_cube = section_get_lval(input, "DFT%PRINT%E_DENSITY_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         ! Write the info on core densities for the interface between cp2k and the XRD code
         ! together with the valence density they are used to compute the form factor (Fourier transform)
         xrd_interface = section_get_lval(input, "DFT%PRINT%E_DENSITY_CUBE%XRD_INTERFACE")
         IF (xrd_interface) THEN
            !cube file only contains soft density (GAPW)
            IF (dft_control%qs_control%gapw) print_total_density = .FALSE.

            filename = "ELECTRON_DENSITY"
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                           extension=".xrd", middle_name=TRIM(filename), &
                                           file_position=my_pos_cube, log_filename=.FALSE.)
            ngto = section_get_ival(input, "DFT%PRINT%E_DENSITY_CUBE%NGAUSS")
            IF (output_unit > 0) THEN
               INQUIRE (UNIT=unit_nr, NAME=filename)
               WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                  "The electron density (atomic part) is written to the file:", &
                  TRIM(filename)
            END IF

            xc_section => section_vals_get_subs_vals(input, "DFT%XC")
            nkind = SIZE(atomic_kind_set)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) "Atomic (core) densities"
               WRITE (unit_nr, *) "Unit cell"
               WRITE (unit_nr, FMT="(3F20.12)") cell%hmat(1, 1), cell%hmat(1, 2), cell%hmat(1, 3)
               WRITE (unit_nr, FMT="(3F20.12)") cell%hmat(2, 1), cell%hmat(2, 2), cell%hmat(2, 3)
               WRITE (unit_nr, FMT="(3F20.12)") cell%hmat(3, 1), cell%hmat(3, 2), cell%hmat(3, 3)
               WRITE (unit_nr, *) "Atomic types"
               WRITE (unit_nr, *) nkind
            END IF
            ! calculate atomic density and core density
            ALLOCATE (ppdens(ngto, 2, nkind), aedens(ngto, 2, nkind), ccdens(ngto, 2, nkind))
            DO ikind = 1, nkind
               atomic_kind => atomic_kind_set(ikind)
               qs_kind => qs_kind_set(ikind)
               CALL get_atomic_kind(atomic_kind, name=name, element_symbol=element_symbol)
               CALL calculate_atomic_density(ppdens(:, :, ikind), atomic_kind, qs_kind, ngto, &
                                             iunit=output_unit, confine=.TRUE.)
               CALL calculate_atomic_density(aedens(:, :, ikind), atomic_kind, qs_kind, ngto, &
                                             iunit=output_unit, allelectron=.TRUE., confine=.TRUE.)
               ccdens(:, 1, ikind) = aedens(:, 1, ikind)
               ccdens(:, 2, ikind) = 0._dp
               CALL project_function_a(ccdens(1:ngto, 2, ikind), ccdens(1:ngto, 1, ikind), &
                                       ppdens(1:ngto, 2, ikind), ppdens(1:ngto, 1, ikind), 0)
               ccdens(:, 2, ikind) = aedens(:, 2, ikind)-ccdens(:, 2, ikind)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, FMT="(I6,A10,A20)") ikind, TRIM(element_symbol), TRIM(name)
                  WRITE (unit_nr, FMT="(I6)") ngto
                  WRITE (unit_nr, *) "   Total density"
                  WRITE (unit_nr, FMT="(2G24.12)") (aedens(i, 1, ikind), aedens(i, 2, ikind), i=1, ngto)
                  WRITE (unit_nr, *) "    Core density"
                  WRITE (unit_nr, FMT="(2G24.12)") (ccdens(i, 1, ikind), ccdens(i, 2, ikind), i=1, ngto)
               END IF
               NULLIFY (atomic_kind)
            END DO

            IF (dft_control%qs_control%gapw) THEN
               CALL get_qs_env(qs_env=qs_env, rho_atom_set=rho_atom_set)

               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "Coordinates and GAPW density"
               END IF
               np = particles%n_els
               DO iat = 1, np
                  CALL get_atomic_kind(particles%els(iat)%atomic_kind, kind_number=ikind)
                  CALL get_qs_kind(qs_kind_set(ikind), grid_atom=grid_atom)
                  rho_atom => rho_atom_set(iat)
                  IF (ASSOCIATED(rho_atom%rho_rad_h(1)%r_coef)) THEN
                     nr = SIZE(rho_atom%rho_rad_h(1)%r_coef, 1)
                     niso = SIZE(rho_atom%rho_rad_h(1)%r_coef, 2)
                  ELSE
                     nr = 0
                     niso = 0
                  ENDIF
                  CALL mp_sum(nr, para_env%group)
                  CALL mp_sum(niso, para_env%group)

                  ALLOCATE (bfun(nr, niso))
                  bfun = 0._dp
                  DO ispin = 1, dft_control%nspins
                     IF (ASSOCIATED(rho_atom%rho_rad_h(1)%r_coef)) THEN
                        bfun(:, :) = bfun+rho_atom%rho_rad_h(ispin)%r_coef-rho_atom%rho_rad_s(ispin)%r_coef
                     END IF
                  END DO
                  CALL mp_sum(bfun, para_env%group)
                  ccdens(:, 1, ikind) = ppdens(:, 1, ikind)
                  ccdens(:, 2, ikind) = 0._dp
                  IF (unit_nr > 0) THEN
                     WRITE (unit_nr, '(I10,I5,3f12.6)') iat, ikind, particles%els(iat)%r
                  END IF
                  DO iso = 1, niso
                     l = indso(1, iso)
                     CALL project_function_b(ccdens(:, 2, ikind), ccdens(:, 1, ikind), bfun(:, iso), grid_atom, l)
                     IF (unit_nr > 0) THEN
                        WRITE (unit_nr, FMT="(3I6)") iso, l, ngto
                        WRITE (unit_nr, FMT="(2G24.12)") (ccdens(i, 1, ikind), ccdens(i, 2, ikind), i=1, ngto)
                     END IF
                  END DO
                  DEALLOCATE (bfun)
               END DO
            ELSE
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "Coordinates"
                  np = particles%n_els
                  DO iat = 1, np
                     CALL get_atomic_kind(particles%els(iat)%atomic_kind, kind_number=ikind)
                     WRITE (unit_nr, '(I10,I5,3f12.6)') iat, ikind, particles%els(iat)%r
                  END DO
               END IF
            END IF

            DEALLOCATE (ppdens, aedens, ccdens)

            CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                              "DFT%PRINT%E_DENSITY_CUBE")

         END IF
         IF (dft_control%qs_control%gapw .AND. print_total_density) THEN
            ! total density in g-space not implemented for k-points
            IF(.NOT.(.NOT. do_kpoints))CALL cp__a("qs_scf_post_gpw.F",2370)
            ! Print total electronic density
            CALL get_qs_env(qs_env=qs_env, &
                            pw_env=pw_env)
            CALL pw_env_get(pw_env=pw_env, &
                            auxbas_pw_pool=auxbas_pw_pool, &
                            pw_pools=pw_pools)
            CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                                   pw=rho_elec_rspace%pw, &
                                   use_data=REALDATA3D, &
                                   in_space=REALSPACE)
            CALL pw_zero(rho_elec_rspace%pw)
            CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                                   pw=rho_elec_gspace%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            CALL pw_zero(rho_elec_gspace%pw)
            CALL get_pw_grid_info(pw_grid=rho_elec_gspace%pw%pw_grid, &
                                  dr=dr, &
                                  vol=volume)
            q_max = SQRT(SUM((pi/dr(:))**2))
            CALL calculate_rhotot_elec_gspace(qs_env=qs_env, &
                                              auxbas_pw_pool=auxbas_pw_pool, &
                                              rhotot_elec_gspace=rho_elec_gspace, &
                                              q_max=q_max, &
                                              rho_hard=rho_hard, &
                                              rho_soft=rho_soft)
            rho_total = rho_hard+rho_soft
            CALL get_pw_grid_info(pw_grid=rho_elec_gspace%pw%pw_grid, &
                                  vol=volume)
            CALL pw_transfer(rho_elec_gspace%pw, rho_elec_rspace%pw, debug=.FALSE.)
            rho_total_rspace = pw_integrate_function(rho_elec_rspace%pw, isign=-1)/volume
            filename = "TOTAL_ELECTRON_DENSITY"
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                           extension=".cube", middle_name=TRIM(filename), &
                                           file_position=my_pos_cube, log_filename=.FALSE.)
            IF (output_unit > 0) THEN
               INQUIRE (UNIT=unit_nr, NAME=filename)
               WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                  "The total electron density is written in cube file format to the file:", &
                  TRIM(filename)
               WRITE (UNIT=output_unit, FMT="(/,(T2,A,F20.10))") &
                  "q(max) [1/Angstrom]              :", q_max/angstrom, &
                  "Soft electronic charge (G-space) :", rho_soft, &
                  "Hard electronic charge (G-space) :", rho_hard, &
                  "Total electronic charge (G-space):", rho_total, &
                  "Total electronic charge (R-space):", rho_total_rspace
            END IF
            CALL cp_pw_to_cube(rho_elec_rspace%pw, unit_nr, "TOTAL ELECTRON DENSITY", &
                               particles=particles, &
                               stride=section_get_ivals(dft_section, "PRINT%E_DENSITY_CUBE%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                              "DFT%PRINT%E_DENSITY_CUBE")
            ! Print total spin density for spin-polarized systems
            IF (dft_control%nspins > 1) THEN
               CALL pw_zero(rho_elec_gspace%pw)
               CALL pw_zero(rho_elec_rspace%pw)
               CALL calculate_rhotot_elec_gspace(qs_env=qs_env, &
                                                 auxbas_pw_pool=auxbas_pw_pool, &
                                                 rhotot_elec_gspace=rho_elec_gspace, &
                                                 q_max=q_max, &
                                                 rho_hard=rho_hard, &
                                                 rho_soft=rho_soft, &
                                                 fsign=-1.0_dp)
               rho_total = rho_hard+rho_soft
               CALL pw_transfer(rho_elec_gspace%pw, rho_elec_rspace%pw, debug=.FALSE.)
               rho_total_rspace = pw_integrate_function(rho_elec_rspace%pw, isign=-1)/volume
               filename = "TOTAL_SPIN_DENSITY"
               unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                              extension=".cube", middle_name=TRIM(filename), &
                                              file_position=my_pos_cube, log_filename=.FALSE.)
               IF (output_unit > 0) THEN
                  INQUIRE (UNIT=unit_nr, NAME=filename)
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                     "The total spin density is written in cube file format to the file:", &
                     TRIM(filename)
                  WRITE (UNIT=output_unit, FMT="(/,(T2,A,F20.10))") &
                     "q(max) [1/Angstrom]                    :", q_max/angstrom, &
                     "Soft part of the spin density (G-space):", rho_soft, &
                     "Hard part of the spin density (G-space):", rho_hard, &
                     "Total spin density (G-space)           :", rho_total, &
                     "Total spin density (R-space)           :", rho_total_rspace
               END IF
               CALL cp_pw_to_cube(rho_elec_rspace%pw, unit_nr, "TOTAL SPIN DENSITY", &
                                  particles=particles, &
                                  stride=section_get_ivals(dft_section, "PRINT%E_DENSITY_CUBE%STRIDE"))
            END IF
            CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_elec_gspace%pw)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_elec_rspace%pw)
         ELSE
            IF (dft_control%nspins > 1) THEN
               CALL get_qs_env(qs_env=qs_env, &
                               pw_env=pw_env)
               CALL pw_env_get(pw_env=pw_env, &
                               auxbas_pw_pool=auxbas_pw_pool, &
                               pw_pools=pw_pools)
               CALL pw_pool_create_pw(pool=auxbas_pw_pool, &
                                      pw=rho_elec_rspace%pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_copy(rho_r(1)%pw, rho_elec_rspace%pw)
               CALL pw_axpy(rho_r(2)%pw, rho_elec_rspace%pw)
               filename = "ELECTRON_DENSITY"
               unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                              extension=".cube", middle_name=TRIM(filename), &
                                              file_position=my_pos_cube, log_filename=.FALSE.)
               IF (output_unit > 0) THEN
                  INQUIRE (UNIT=unit_nr, NAME=filename)
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                     "The sum of alpha and beta density is written in cube file format to the file:", &
                     TRIM(filename)
               END IF
               CALL cp_pw_to_cube(rho_elec_rspace%pw, unit_nr, "SUM OF ALPHA AND BETA DENSITY", &
                                  particles=particles, stride=section_get_ivals(dft_section, "PRINT%E_DENSITY_CUBE%STRIDE"))
               CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                                 "DFT%PRINT%E_DENSITY_CUBE")
               CALL pw_copy(rho_r(1)%pw, rho_elec_rspace%pw)
               CALL pw_axpy(rho_r(2)%pw, rho_elec_rspace%pw, alpha=-1.0_dp)
               filename = "SPIN_DENSITY"
               unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                              extension=".cube", middle_name=TRIM(filename), &
                                              file_position=my_pos_cube, log_filename=.FALSE.)
               IF (output_unit > 0) THEN
                  INQUIRE (UNIT=unit_nr, NAME=filename)
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                     "The spin density is written in cube file format to the file:", &
                     TRIM(filename)
               END IF
               CALL cp_pw_to_cube(rho_elec_rspace%pw, unit_nr, "SPIN DENSITY", &
                                  particles=particles, &
                                  stride=section_get_ivals(dft_section, "PRINT%E_DENSITY_CUBE%STRIDE"))
               CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                                 "DFT%PRINT%E_DENSITY_CUBE")
               CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_elec_rspace%pw)
            ELSE
               filename = "ELECTRON_DENSITY"
               unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%E_DENSITY_CUBE", &
                                              extension=".cube", middle_name=TRIM(filename), &
                                              file_position=my_pos_cube, log_filename=.FALSE.)
               IF (output_unit > 0) THEN
                  INQUIRE (UNIT=unit_nr, NAME=filename)
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A)") &
                     "The electron density is written in cube file format to the file:", &
                     TRIM(filename)
               END IF
               CALL cp_pw_to_cube(rho_r(1)%pw, unit_nr, "ELECTRON DENSITY", &
                                  particles=particles, &
                                  stride=section_get_ivals(dft_section, "PRINT%E_DENSITY_CUBE%STRIDE"))
               CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                                 "DFT%PRINT%E_DENSITY_CUBE")
            END IF ! nspins
         END IF ! total density for GAPW
      END IF ! print key

      ! Print the hartree potential
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%V_HARTREE_CUBE"), cp_p_file)) THEN

         CALL get_qs_env(qs_env=qs_env, &
                         pw_env=pw_env, &
                         v_hartree_rspace=v_hartree_rspace)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)

         append_cube = section_get_lval(input, "DFT%PRINT%V_HARTREE_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
         CALL pw_env_get(pw_env)
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%V_HARTREE_CUBE", &
                                        extension=".cube", middle_name="v_hartree", file_position=my_pos_cube)
         udvol = 1.0_dp/v_hartree_rspace%pw_grid%dvol

         CALL pw_copy(v_hartree_rspace, aux_r%pw)
         CALL pw_scale(aux_r%pw, udvol)

         CALL cp_pw_to_cube(aux_r%pw, unit_nr, "HARTREE POTENTIAL", particles=particles, &
                            stride=section_get_ivals(dft_section, &
                                                     "PRINT%V_HARTREE_CUBE%STRIDE"))
         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%V_HARTREE_CUBE")

         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
      ENDIF

      ! Print the external potential
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%EXTERNAL_POTENTIAL_CUBE"), cp_p_file)) THEN
         IF (dft_control%apply_external_potential) THEN
            CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, vee=vee)
            CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, &
                                   use_data=REALDATA3D, &
                                   in_space=REALSPACE)

            append_cube = section_get_lval(input, "DFT%PRINT%EXTERNAL_POTENTIAL_CUBE%APPEND")
            my_pos_cube = "REWIND"
            IF (append_cube) THEN
               my_pos_cube = "APPEND"
            END IF
            CALL pw_env_get(pw_env)
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%EXTERNAL_POTENTIAL_CUBE", &
                                           extension=".cube", middle_name="ext_pot", file_position=my_pos_cube)

            CALL pw_copy(vee%pw, aux_r%pw)

            CALL cp_pw_to_cube(aux_r%pw, unit_nr, "EXTERNAL POTENTIAL", particles=particles, &
                               stride=section_get_ivals(dft_section, &
                                                        "PRINT%EXTERNAL_POTENTIAL_CUBE%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                              "DFT%PRINT%EXTERNAL_POTENTIAL_CUBE")

            CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
         ENDIF
      ENDIF

      ! Print the Electrical Field Components
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%EFIELD_CUBE"), cp_p_file)) THEN

         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_g%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)

         append_cube = section_get_lval(input, "DFT%PRINT%EFIELD_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, &
                         v_hartree_rspace=v_hartree_rspace)
         CALL pw_env_get(pw_env)
         udvol = 1.0_dp/v_hartree_rspace%pw_grid%dvol
         DO id = 1, 3
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%EFIELD_CUBE", &
                                           extension=".cube", middle_name="efield_"//cdir(id), file_position=my_pos_cube)

            CALL pw_transfer(v_hartree_rspace, aux_g%pw)
            nd = 0
            nd(id) = 1
            CALL pw_derive(aux_g%pw, nd)
            CALL pw_transfer(aux_g%pw, aux_r%pw)
            CALL pw_scale(aux_r%pw, udvol)

            CALL cp_pw_to_cube(aux_r%pw, &
                               unit_nr, "ELECTRIC FIELD", particles=particles, &
                               stride=section_get_ivals(dft_section, &
                                                        "PRINT%EFIELD_CUBE%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                              "DFT%PRINT%EFIELD_CUBE")
         END DO

         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_g%pw)
      END IF

      ! Write cube files from the implicit Poisson solver
      CALL qs_scf_post_ps_implicit(input, logger, qs_env)

      CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
      ! Write the density matrices
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%AO_MATRICES/DENSITY"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, input, "DFT%PRINT%AO_MATRICES/DENSITY", &
                                   extension=".Log")
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         CALL qs_rho_get(rho, rho_ao_kp=rho_ao)
         after = MIN(MAX(after, 1), 16)
         DO ispin = 1, dft_control%nspins
            DO img = 1, dft_control%nimages
               CALL cp_dbcsr_write_sparse_matrix(rho_ao(ispin, img)%matrix, 4, after, qs_env, &
                                                 para_env, output_unit=iw, omit_headers=omit_headers)
            END DO
         END DO
         CALL cp_print_key_finished_output(iw, logger, input, &
                                           "DFT%PRINT%AO_MATRICES/DENSITY")
      END IF

      ! Write the Kohn-Sham matrices
      write_ks = BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                  "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX"), cp_p_file)
      write_xc = BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                  "DFT%PRINT%AO_MATRICES/MATRIX_VXC"), cp_p_file)
      ! we need to update stuff before writing, potentially computing the matrix_vxc
      IF (write_ks .OR. write_xc) THEN
         IF (write_xc) qs_env%requires_matrix_vxc = .TRUE.
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., &
                                  just_energy=.FALSE.)
         IF (write_xc) qs_env%requires_matrix_vxc = .FALSE.
      END IF

      ! Write the Kohn-Sham matrices
      IF (write_ks) THEN
         iw = cp_print_key_unit_nr(logger, input, "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX", &
                                   extension=".Log")
         CALL get_qs_env(qs_env=qs_env, matrix_ks_kp=ks_rmpv)
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         DO ispin = 1, dft_control%nspins
            DO img = 1, dft_control%nimages
               CALL cp_dbcsr_write_sparse_matrix(ks_rmpv(ispin, img)%matrix, 4, after, qs_env, &
                                                 para_env, output_unit=iw, omit_headers=omit_headers)
            END DO
         END DO
         CALL cp_print_key_finished_output(iw, logger, input, &
                                           "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX")
      END IF

      ! Write the xc matrix
      IF (write_xc) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_vxc_kp=matrix_vxc)
         IF(.NOT.(ASSOCIATED(matrix_vxc)))CALL cp__a("qs_scf_post_gpw.F",2691)
         iw = cp_print_key_unit_nr(logger, input, "DFT%PRINT%AO_MATRICES/MATRIX_VXC", &
                                   extension=".Log")
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         DO ispin = 1, dft_control%nspins
            DO img = 1, dft_control%nimages
               CALL cp_dbcsr_write_sparse_matrix(matrix_vxc(ispin, img)%matrix, 4, after, qs_env, &
                                                 para_env, output_unit=iw, omit_headers=omit_headers)
            END DO
         END DO
         CALL cp_print_key_finished_output(iw, logger, input, &
                                           "DFT%PRINT%AO_MATRICES/MATRIX_VXC")
      END IF

      ! Compute the Mulliken charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%MULLIKEN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MULLIKEN", extension=".mulliken", &
                                        middle_name="", log_filename=.FALSE.)
         print_level = 1
         CALL section_vals_val_get(print_key, "PRINT_GOP", l_val=print_it)
         IF (print_it) print_level = 2
         CALL section_vals_val_get(print_key, "PRINT_ALL", l_val=print_it)
         IF (print_it) print_level = 3
         CALL mulliken_population_analysis(qs_env, unit_nr, print_level)
         CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%MULLIKEN")
      END IF

      ! Compute the Hirshfeld charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%HIRSHFELD")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         ! we check if real space density is available
         NULLIFY (rho)
         CALL get_qs_env(qs_env=qs_env, rho=rho)
         CALL qs_rho_get(rho, rho_r_valid=rho_r_valid)
         IF (rho_r_valid) THEN
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%HIRSHFELD", extension=".hirshfeld", &
                                           middle_name="", log_filename=.FALSE.)
            CALL hirshfeld_charges(qs_env, print_key, unit_nr)
            CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%HIRSHFELD")
         END IF
      END IF

      ! MAO analysis
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%MAO_ANALYSIS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MAO_ANALYSIS", extension=".mao", &
                                        middle_name="", log_filename=.FALSE.)
         CALL mao_analysis(qs_env, print_key, unit_nr)
         CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%MAO_ANALYSIS")
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_mo_free_results

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param input_section ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE hirshfeld_charges(qs_env, input_section, unit_nr)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: input_section
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'hirshfeld_charges', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ikind, natom, nkind, nspin, &
                                                            refc, shapef
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: do_sc, paw_atom
      REAL(KIND=dp)                                      :: zeff
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p, matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(hirshfeld_type), POINTER                      :: hirshfeld_env
      TYPE(mpole_rho_atom), DIMENSION(:), POINTER        :: mp_rho
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole

      NULLIFY (hirshfeld_env)
      CALL create_hirshfeld_type(hirshfeld_env)
      !
      CALL get_qs_env(qs_env, nkind=nkind, natom=natom)
      ALLOCATE (hirshfeld_env%charges(natom))
      ! input options
      CALL section_vals_val_get(input_section, "SELF_CONSISTENT", l_val=do_sc)
      CALL section_vals_val_get(input_section, "SHAPE_FUNCTION", i_val=shapef)
      CALL section_vals_val_get(input_section, "REFERENCE_CHARGE", i_val=refc)
      CALL set_hirshfeld_info(hirshfeld_env, shape_function_type=shapef, &
                              iterative=do_sc, ref_charge=refc)
      ! shape function
      CALL get_qs_env(qs_env, qs_kind_set=qs_kind_set, atomic_kind_set=atomic_kind_set)
      CALL create_shape_function(hirshfeld_env, qs_kind_set, atomic_kind_set)
      ! reference charges
      CALL get_qs_env(qs_env, rho=rho)
      CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
      nspin = SIZE(matrix_p, 1)
      ALLOCATE (charges(natom, nspin))
      SELECT CASE (refc)
      CASE (ref_charge_atomic)
         DO ikind = 1, nkind
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, atom_list=atom_list)
            DO iat = 1, SIZE(atom_list)
               i = atom_list(iat)
               hirshfeld_env%charges(i) = zeff
            END DO
         END DO
      CASE (ref_charge_mulliken)
         CALL get_qs_env(qs_env, matrix_s_kp=matrix_s, para_env=para_env)
         CALL mulliken_charges(matrix_p, matrix_s, para_env, charges)
         DO iat = 1, natom
            hirshfeld_env%charges(iat) = SUM(charges(iat, :))
         END DO
      CASE DEFAULT
         CALL cp__b("qs_scf_post_gpw.F",2817,"Unknown type of reference charge for Hirshfeld partitioning.")
      END SELECT
      !
      charges = 0.0_dp
      IF (hirshfeld_env%iterative) THEN
         ! Hirshfeld-I charges
         CALL comp_hirshfeld_i_charges(qs_env, hirshfeld_env, charges, unit_nr)
      ELSE
         ! Hirshfeld charges
         CALL comp_hirshfeld_charges(qs_env, hirshfeld_env, charges)
      END IF
      CALL get_qs_env(qs_env, particle_set=particle_set, dft_control=dft_control)
      IF (dft_control%qs_control%gapw) THEN
         ! GAPW: add core charges (rho_hard - rho_soft)
         CALL get_qs_env(qs_env, rho0_mpole=rho0_mpole)
         CALL get_rho0_mpole(rho0_mpole, mp_rho=mp_rho)
         DO iat = 1, natom
            atomic_kind => particle_set(iat)%atomic_kind
            CALL get_atomic_kind(atomic_kind, kind_number=ikind)
            CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom)
            IF (paw_atom) THEN
               charges(iat, 1:nspin) = charges(iat, 1:nspin)+mp_rho(iat)%q0(1:nspin)
            END IF
         END DO
      END IF
      !
      IF (unit_nr > 0) THEN
         CALL write_hirshfeld_charges(charges, hirshfeld_env, particle_set, &
                                      qs_kind_set, unit_nr)
      END IF
      !
      CALL release_hirshfeld_type(hirshfeld_env)
      DEALLOCATE (charges)

   END SUBROUTINE hirshfeld_charges

! **************************************************************************************************
!> \brief ...
!> \param ca ...
!> \param a ...
!> \param cb ...
!> \param b ...
!> \param l ...
! **************************************************************************************************
   SUBROUTINE project_function_a(ca, a, cb, b, l)
      ! project function cb on ca
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: ca
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a, cb, b
      INTEGER, INTENT(IN)                                :: l

      CHARACTER(len=*), PARAMETER :: routineN = 'project_function_a', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: info, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: smat, tmat, v

      n = SIZE(ca)
      ALLOCATE (smat(n, n), tmat(n, n), v(n, 1), ipiv(n))

      CALL sg_overlap(smat, l, a, a)
      CALL sg_overlap(tmat, l, a, b)
      v(:, 1) = MATMUL(tmat, cb)
      CALL lapack_sgesv(n, 1, smat, n, ipiv, v, n, info)
      IF(.NOT.(info == 0))CALL cp__a("qs_scf_post_gpw.F",2881)
      ca(:) = v(:, 1)

      DEALLOCATE (smat, tmat, v, ipiv)

   END SUBROUTINE project_function_a

! **************************************************************************************************
!> \brief ...
!> \param ca ...
!> \param a ...
!> \param bfun ...
!> \param grid_atom ...
!> \param l ...
! **************************************************************************************************
   SUBROUTINE project_function_b(ca, a, bfun, grid_atom, l)
      ! project function f on ca
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: ca
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a, bfun
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      INTEGER, INTENT(IN)                                :: l

      CHARACTER(len=*), PARAMETER :: routineN = 'project_function_b', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, n, nr
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: afun
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: smat, v

      n = SIZE(ca)
      nr = grid_atom%nr
      ALLOCATE (smat(n, n), v(n, 1), ipiv(n), afun(nr))

      CALL sg_overlap(smat, l, a, a)
      DO i = 1, n
         afun(:) = grid_atom%rad(:)**l*EXP(-a(i)*grid_atom%rad2(:))
         v(i, 1) = SUM(afun(:)*bfun(:)*grid_atom%wr(:))
      END DO
      CALL lapack_sgesv(n, 1, smat, n, ipiv, v, n, info)
      IF(.NOT.(info == 0))CALL cp__a("qs_scf_post_gpw.F",2921)
      ca(:) = v(:, 1)

      DEALLOCATE (smat, v, ipiv, afun)

   END SUBROUTINE project_function_b

! **************************************************************************************************
!> \brief Performs printing of cube files related to the implicit Poisson solver
!> \param input input
!> \param logger the logger
!> \param qs_env the qs_env in which the qs_env lives
!> \par History
!>      03.2016 refactored from write_mo_free_results [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE qs_scf_post_ps_implicit(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_ps_implicit', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename, my_pos_cube
      INTEGER                                            :: boundary_condition, handle, i, j, &
                                                            n_cstr, n_tiles, unit_nr
      LOGICAL :: append_cube, do_cstr_charge_cube, do_dielectric_cube, do_dirichlet_bc_cube, &
         has_dirichlet_bc, has_implicit_ps, tile_cubes
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: aux_r
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: dirichlet_tile
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: dft_section

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, auxbas_pw_pool, dft_section, particles)

      dft_section => section_vals_get_subs_vals(input, "DFT")
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, subsys=subsys)
      CALL qs_subsys_get(subsys, particles=particles)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      has_implicit_ps = .FALSE.
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      IF (pw_env%poisson_env%parameters%solver .EQ. pw_poisson_implicit) has_implicit_ps = .TRUE.

      ! Write the dielectric constant into a cube file
      do_dielectric_cube = BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                            "DFT%PRINT%IMPLICIT_PSOLVER%DIELECTRIC_CUBE"), cp_p_file)
      IF (has_implicit_ps .AND. do_dielectric_cube) THEN
         append_cube = section_get_lval(input, "DFT%PRINT%IMPLICIT_PSOLVER%DIELECTRIC_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%IMPLICIT_PSOLVER%DIELECTRIC_CUBE", &
                                        extension=".cube", middle_name="DIELCTRIC_CONSTANT", file_position=my_pos_cube)
         CALL pw_env_get(pw_env, poisson_env=poisson_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, use_data=REALDATA3D, in_space=REALSPACE)

         boundary_condition = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
         SELECT CASE (boundary_condition)
         CASE (MIXED_PERIODIC_BC)
            CALL pw_copy(poisson_env%implicit_env%dielectric%eps, aux_r%pw)
         CASE (MIXED_BC, NEUMANN_BC)
            CALL pw_shrink(pw_env%poisson_env%parameters%ps_implicit_params%neumann_directions, &
                           pw_env%poisson_env%implicit_env%dct_env%dests_shrink, &
                           pw_env%poisson_env%implicit_env%dct_env%srcs_shrink, &
                           pw_env%poisson_env%implicit_env%dct_env%bounds_local_shftd, &
                           poisson_env%implicit_env%dielectric%eps, aux_r%pw)
         END SELECT

         CALL cp_pw_to_cube(aux_r%pw, unit_nr, "DIELECTRIC CONSTANT", particles=particles, &
                            stride=section_get_ivals(dft_section, "PRINT%IMPLICIT_PSOLVER%DIELECTRIC_CUBE%STRIDE"))
         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%IMPLICIT_PSOLVER%DIELECTRIC_CUBE")

         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
      ENDIF

      ! Write Dirichlet constraint charges into a cube file
      do_cstr_charge_cube = BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                             "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_CSTR_CHARGE_CUBE"), cp_p_file)

      has_dirichlet_bc = .FALSE.
      IF (has_implicit_ps) THEN
         boundary_condition = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
         IF (boundary_condition .EQ. MIXED_PERIODIC_BC .OR. boundary_condition .EQ. MIXED_BC) THEN
            has_dirichlet_bc = .TRUE.
         END IF
      END IF

      IF (has_implicit_ps .AND. do_cstr_charge_cube .AND. has_dirichlet_bc) THEN
         append_cube = section_get_lval(input, &
                                        "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_CSTR_CHARGE_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         unit_nr = cp_print_key_unit_nr(logger, input, &
                                        "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_CSTR_CHARGE_CUBE", &
                                        extension=".cube", middle_name="dirichlet_cstr_charge", file_position=my_pos_cube)
         CALL pw_env_get(pw_env, poisson_env=poisson_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, use_data=REALDATA3D, in_space=REALSPACE)

         boundary_condition = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
         SELECT CASE (boundary_condition)
         CASE (MIXED_PERIODIC_BC)
            CALL pw_copy(poisson_env%implicit_env%cstr_charge, aux_r%pw)
         CASE (MIXED_BC)
            CALL pw_shrink(pw_env%poisson_env%parameters%ps_implicit_params%neumann_directions, &
                           pw_env%poisson_env%implicit_env%dct_env%dests_shrink, &
                           pw_env%poisson_env%implicit_env%dct_env%srcs_shrink, &
                           pw_env%poisson_env%implicit_env%dct_env%bounds_local_shftd, &
                           poisson_env%implicit_env%cstr_charge, aux_r%pw)
         END SELECT

         CALL cp_pw_to_cube(aux_r%pw, unit_nr, "DIRICHLET CONSTRAINT CHARGE", particles=particles, &
                            stride=section_get_ivals(dft_section, "PRINT%IMPLICIT_PSOLVER%DIRICHLET_CSTR_CHARGE_CUBE%STRIDE"))
         CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                           "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_CSTR_CHARGE_CUBE")

         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
      ENDIF

      ! Write Dirichlet type constranits into cube files
      do_dirichlet_bc_cube = BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                                              "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE"), cp_p_file)
      has_dirichlet_bc = .FALSE.
      IF (has_implicit_ps) THEN
         boundary_condition = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
         IF (boundary_condition .EQ. MIXED_PERIODIC_BC .OR. boundary_condition .EQ. MIXED_BC) THEN
            has_dirichlet_bc = .TRUE.
         END IF
      END IF

      IF (has_implicit_ps .AND. has_dirichlet_bc .AND. do_dirichlet_bc_cube) THEN
         append_cube = section_get_lval(input, "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE%APPEND")
         my_pos_cube = "REWIND"
         IF (append_cube) THEN
            my_pos_cube = "APPEND"
         END IF
         tile_cubes = section_get_lval(input, "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE%TILE_CUBES")

         CALL pw_env_get(pw_env, poisson_env=poisson_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_create_pw(auxbas_pw_pool, aux_r%pw, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_zero(aux_r%pw)

         IF (tile_cubes) THEN
            ! one cube file per tile
            n_cstr = SIZE(poisson_env%implicit_env%contacts)
            DO j = 1, n_cstr
               n_tiles = poisson_env%implicit_env%contacts(j)%dirichlet_bc%n_tiles
               DO i = 1, n_tiles
                  filename = "dirichlet_cstr_"//TRIM(ADJUSTL(cp_to_string(j)))// &
                             "_tile_"//TRIM(ADJUSTL(cp_to_string(i)))
                  unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE", &
                                                 extension=".cube", middle_name=filename, file_position=my_pos_cube)

                  CALL pw_copy(poisson_env%implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%tile_pw, aux_r%pw)

                  CALL cp_pw_to_cube(aux_r%pw, unit_nr, "DIRICHLET TYPE CONSTRAINT", particles=particles, &
                                     stride=section_get_ivals(dft_section, "PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE%STRIDE"))
                  CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                                    "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE")
               END DO
            END DO
         ELSE
            ! a single cube file
            NULLIFY (dirichlet_tile)
            CALL pw_pool_create_pw(auxbas_pw_pool, dirichlet_tile, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(dirichlet_tile)
            unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE", &
                                           extension=".cube", middle_name="DIRICHLET_CSTR", file_position=my_pos_cube)

            n_cstr = SIZE(poisson_env%implicit_env%contacts)
            DO j = 1, n_cstr
               n_tiles = poisson_env%implicit_env%contacts(j)%dirichlet_bc%n_tiles
               DO i = 1, n_tiles
                  CALL pw_copy(poisson_env%implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%tile_pw, dirichlet_tile)
                  CALL pw_axpy(dirichlet_tile, aux_r%pw)
               END DO
            END DO

            CALL cp_pw_to_cube(aux_r%pw, unit_nr, "DIRICHLET TYPE CONSTRAINT", particles=particles, &
                               stride=section_get_ivals(dft_section, "PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                              "DFT%PRINT%IMPLICIT_PSOLVER%DIRICHLET_BC_CUBE")
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dirichlet_tile)
         END IF

         CALL pw_pool_give_back_pw(auxbas_pw_pool, aux_r%pw)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_post_ps_implicit

END MODULE qs_scf_post_gpw
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods using the PEXSI library to calculate the density matrix and
!>        related quantities using the Kohn-Sham and overlap matrices from the
!>        linear scaling quickstep SCF environment.
!> \par History
!>       2014.11 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************

MODULE pexsi_methods
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_to_csr_screening
   USE cp_dbcsr_interface,              ONLY: &
        cp_convert_csr_to_dbcsr, cp_convert_dbcsr_to_csr, cp_csr_create, cp_csr_create_from_dbcsr, &
        cp_csr_destroy, cp_csr_print_sparsity, cp_dbcsr_allocate_matrix_set, cp_dbcsr_arnoldi_ev, &
        cp_dbcsr_copy, cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_has_symmetry, cp_dbcsr_init, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_setup_arnoldi_data, &
        cp_dbcsr_type, cp_set_arnoldi_initial_vector, csr_eqrow_floor_dist, dbcsr_arnoldi_data, &
        dbcsr_distribution_mp, dbcsr_mp_mynode, dbcsr_type_no_symmetry, dbcsr_type_real_8, &
        deallocate_arnoldi_data, get_selected_ritz_val, get_selected_ritz_vec
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dm_ls_scf_qs,                    ONLY: matrix_ls_to_qs
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8
   USE message_passing,                 ONLY: mp_bcast
   USE pexsi_interface,                 ONLY: cp_pexsi_dft_driver,&
                                              cp_pexsi_get_options,&
                                              cp_pexsi_load_real_symmetric_hs_matrix,&
                                              cp_pexsi_retrieve_real_symmetric_dft_matrix,&
                                              cp_pexsi_set_default_options,&
                                              cp_pexsi_set_options
   USE pexsi_types,                     ONLY: convert_nspin_cp2k_pexsi,&
                                              cp2k_to_pexsi,&
                                              lib_pexsi_env,&
                                              pexsi_to_cp2k
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pexsi_methods'

   LOGICAL, PARAMETER, PRIVATE          :: careful_mod = .FALSE.

   PUBLIC :: density_matrix_pexsi, pexsi_init_read_input, pexsi_to_qs, pexsi_init_scf, pexsi_finalize_scf, &
             pexsi_set_convergence_tolerance

CONTAINS

! **************************************************************************************************
!> \brief Read CP2K input section PEXSI and pass it to the PEXSI environment
!> \param pexsi_section ...
!> \param pexsi_env ...
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE pexsi_init_read_input(pexsi_section, pexsi_env)
      TYPE(section_vals_type), INTENT(IN), POINTER       :: pexsi_section
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pexsi_init_read_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: isInertiaCount_int, maxPEXSIIter, &
                                                            min_ranks_per_pole, npSymbFact, &
                                                            numPole, ordering, verbosity
      LOGICAL                                            :: csr_screening, isInertiaCount
      REAL(KIND=dp) :: gap, muInertiaExpansion, muInertiaTolerance, muMax0, muMin0, &
         muPEXSISafeGuard, numElectronInitialTolerance, numElectronTargetTolerance, temperature

! Note: omitting the following PEXSI options: deltaE (estimated by Arnoldi
! before invoking PEXSI), mu0 (taken from previous SCF step), matrixType
! (not implemented in PEXSI yet), isSymbolicFactorize (not needed because
! of fixed sparsity pattern)

      CALL section_vals_val_get(pexsi_section, "TEMPERATURE", &
                                r_val=temperature)
      CALL section_vals_val_get(pexsi_section, "GAP", &
                                r_val=gap)
      CALL section_vals_val_get(pexsi_section, "NUM_POLE", &
                                i_val=numPole)
      CALL section_vals_val_get(pexsi_section, "IS_INERTIA_COUNT", &
                                l_val=isInertiaCount)
      CALL section_vals_val_get(pexsi_section, "MAX_PEXSI_ITER", &
                                i_val=maxPEXSIIter)
      CALL section_vals_val_get(pexsi_section, "MU_MIN_0", &
                                r_val=muMin0)
      CALL section_vals_val_get(pexsi_section, "MU_MAX_0", &
                                r_val=muMax0)
      CALL section_vals_val_get(pexsi_section, "MU_INERTIA_TOLERANCE", &
                                r_val=muInertiaTolerance)
      CALL section_vals_val_get(pexsi_section, "MU_INERTIA_EXPANSION", &
                                r_val=muInertiaExpansion)
      CALL section_vals_val_get(pexsi_section, "MU_PEXSI_SAFE_GUARD", &
                                r_val=muPEXSISafeGuard)
      CALL section_vals_val_get(pexsi_section, "NUM_ELECTRON_INITIAL_TOLERANCE", &
                                r_val=numElectronInitialTolerance)
      CALL section_vals_val_get(pexsi_section, "NUM_ELECTRON_PEXSI_TOLERANCE", &
                                r_val=numElectronTargetTolerance)
      CALL section_vals_val_get(pexsi_section, "ORDERING", &
                                i_val=ordering)
      CALL section_vals_val_get(pexsi_section, "NP_SYMB_FACT", &
                                i_val=npSymbFact)
      CALL section_vals_val_get(pexsi_section, "VERBOSITY", &
                                i_val=verbosity)
      CALL section_vals_val_get(pexsi_section, "MIN_RANKS_PER_POLE", &
                                i_val=min_ranks_per_pole)
      CALL section_vals_val_get(pexsi_section, "CSR_SCREENING", &
                                l_val=csr_screening)

      isInertiaCount_int = MERGE(1, 0, isInertiaCount) ! is integer in PEXSI

      ! Set default options inside PEXSI
      CALL cp_pexsi_set_default_options(pexsi_env%options)

      ! Pass CP2K input to PEXSI options
      CALL cp_pexsi_set_options(pexsi_env%options, temperature=temperature, gap=gap, &
                                numPole=numPole, isInertiaCount=isInertiaCount_int, maxPEXSIIter=maxPEXSIIter, &
                                muMin0=muMin0, muMax0=muMax0, muInertiaTolerance=muInertiaTolerance, &
                                muInertiaExpansion=muInertiaExpansion, muPEXSISafeGuard=muPEXSISafeGuard, &
                                ordering=ordering, npSymbFact=npSymbFact, verbosity=verbosity)

      pexsi_env%num_ranks_per_pole = min_ranks_per_pole ! not a PEXSI option
      pexsi_env%csr_screening = csr_screening

      IF (numElectronInitialTolerance .LT. numElectronTargetTolerance) &
         numElectronInitialTolerance = numElectronTargetTolerance

      pexsi_env%tol_nel_initial = numElectronInitialTolerance
      pexsi_env%tol_nel_target = numElectronTargetTolerance

   END SUBROUTINE pexsi_init_read_input

! **************************************************************************************************
!> \brief Initializations needed before SCF
!> \param ks_env ...
!> \param pexsi_env ...
!> \param template_matrix DBCSR matrix that defines the block structure and
!>        sparsity pattern of all matrices that are sent to PEXSI
! **************************************************************************************************
   SUBROUTINE pexsi_init_scf(ks_env, pexsi_env, template_matrix)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: template_matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'pexsi_init_scf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! Create template matrices fixing sparsity pattern for PEXSI
      CALL cp_dbcsr_init(pexsi_env%dbcsr_template_matrix_sym)
      CALL cp_dbcsr_init(pexsi_env%dbcsr_template_matrix_nonsym)
      CALL cp_dbcsr_init(pexsi_env%csr_sparsity)

      IF (cp_dbcsr_has_symmetry(template_matrix)) THEN
         CALL cp_dbcsr_copy(pexsi_env%dbcsr_template_matrix_sym, template_matrix, &
                            "symmetric template matrix for CSR conversion")
         CALL cp_dbcsr_desymmetrize(pexsi_env%dbcsr_template_matrix_sym, &
                                    pexsi_env%dbcsr_template_matrix_nonsym)
      ELSE
         CALL cp_dbcsr_copy(pexsi_env%dbcsr_template_matrix_nonsym, template_matrix, &
                            "non-symmetric template matrix for CSR conversion")
         CALL cp_dbcsr_copy(pexsi_env%dbcsr_template_matrix_sym, template_matrix, &
                            "symmetric template matrix for CSR conversion")
      ENDIF

      CALL cp_dbcsr_create(pexsi_env%csr_sparsity, "CSR sparsity", &
                           template=pexsi_env%dbcsr_template_matrix_sym, &
                           data_type=dbcsr_type_real_8)
      CALL cp_dbcsr_copy(pexsi_env%csr_sparsity, pexsi_env%dbcsr_template_matrix_sym)

      CALL cp_dbcsr_to_csr_screening(ks_env, pexsi_env%csr_sparsity)

      IF (.NOT. pexsi_env%csr_screening) CALL cp_dbcsr_set(pexsi_env%csr_sparsity, 1.0)
      CALL cp_csr_create_from_dbcsr(pexsi_env%dbcsr_template_matrix_nonsym, &
                                    pexsi_env%csr_mat_s, &
                                    csr_eqrow_floor_dist, &
                                    csr_sparsity=pexsi_env%csr_sparsity, &
                                    numnodes=pexsi_env%num_ranks_per_pole)

      IF (unit_nr > 0) WRITE (unit_nr, "(/T2,A)") "SPARSITY OF THE OVERLAP MATRIX IN CSR FORMAT"
      CALL cp_csr_print_sparsity(pexsi_env%csr_mat_s, unit_nr)

      CALL cp_convert_dbcsr_to_csr(pexsi_env%dbcsr_template_matrix_nonsym, pexsi_env%csr_mat_s)

      CALL cp_csr_create(pexsi_env%csr_mat_ks, pexsi_env%csr_mat_s)
      CALL cp_csr_create(pexsi_env%csr_mat_p, pexsi_env%csr_mat_s)
      CALL cp_csr_create(pexsi_env%csr_mat_E, pexsi_env%csr_mat_s)
      CALL cp_csr_create(pexsi_env%csr_mat_F, pexsi_env%csr_mat_s)

      DO ispin = 1, pexsi_env%nspin
         ! max_ev_vector only initialised to avoid warning in case max_scf==0
         CALL cp_dbcsr_init(pexsi_env%max_ev_vector(ispin))
         CALL cp_dbcsr_init(pexsi_env%matrix_w(ispin)%matrix)
         CALL cp_dbcsr_create(pexsi_env%matrix_w(ispin)%matrix, "W matrix", &
                              template=template_matrix, matrix_type=dbcsr_type_no_symmetry)
      ENDDO

      CALL cp_pexsi_set_options(pexsi_env%options, numElectronPEXSITolerance=pexsi_env%tol_nel_initial)

      CALL timestop(handle)

   END SUBROUTINE pexsi_init_scf

! **************************************************************************************************
!> \brief Deallocations and post-processing after SCF
!> \param pexsi_env ...
!> \param mu_spin ...
! **************************************************************************************************
   SUBROUTINE pexsi_finalize_scf(pexsi_env, mu_spin)
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: mu_spin

      CHARACTER(len=*), PARAMETER :: routineN = 'pexsi_finalize_scf', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      REAL(KIND=dp)                                      :: kTS_total, mu_total
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      mu_total = SUM(mu_spin(1:pexsi_env%nspin))/REAL(pexsi_env%nspin, KIND=dp)
      kTS_total = SUM(pexsi_env%kTS)
      IF (pexsi_env%nspin .EQ. 1) kTS_total = kTS_total*2.0_dp

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, "(/A,T55,F26.15)") &
            " PEXSI| Electronic entropic energy (a.u.):", kTS_total
         WRITE (unit_nr, "(A,T55,F26.15)") &
            " PEXSI| Chemical potential (a.u.):", mu_total
      ENDIF

      CALL cp_dbcsr_release(pexsi_env%dbcsr_template_matrix_sym)
      CALL cp_dbcsr_release(pexsi_env%dbcsr_template_matrix_nonsym)
      CALL cp_dbcsr_release(pexsi_env%csr_sparsity)
      CALL cp_csr_destroy(pexsi_env%csr_mat_p)
      CALL cp_csr_destroy(pexsi_env%csr_mat_ks)
      CALL cp_csr_destroy(pexsi_env%csr_mat_s)
      CALL cp_csr_destroy(pexsi_env%csr_mat_E)
      CALL cp_csr_destroy(pexsi_env%csr_mat_F)
      DO ispin = 1, pexsi_env%nspin
         CALL cp_dbcsr_release(pexsi_env%max_ev_vector(ispin))
         CALL cp_dbcsr_release(pexsi_env%matrix_w(ispin)%matrix)
      ENDDO
      CALL timestop(handle)
      pexsi_env%tol_nel_initial = pexsi_env%tol_nel_target ! Turn off adaptive threshold for subsequent SCF cycles
   END SUBROUTINE pexsi_finalize_scf

! **************************************************************************************************
!> \brief Calculate density matrix, energy-weighted density matrix and entropic
!>        energy contribution with the DFT driver of the PEXSI library.
!> \param[in,out] pexsi_env     PEXSI environment
!> \param[in,out] matrix_p      density matrix returned by PEXSI
!> \param[in,out] matrix_w      energy-weighted density matrix returned by PEXSI
!> \param[out] kTS              entropic energy contribution returned by PEXSI
!> \param[in] matrix_ks         Kohn-Sham matrix from linear scaling QS environment
!> \param[in] matrix_s          overlap matrix from linear scaling QS environment
!> \param[in] nelectron_exact   exact number of electrons
!> \param[out] mu               chemical potential calculated by PEXSI
!> \param[in] iscf              SCF step
!> \param[in] ispin             Number of spin
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE density_matrix_pexsi(pexsi_env, matrix_p, matrix_w, kTS, matrix_ks, matrix_s, &
                                   nelectron_exact, mu, iscf, ispin)
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      TYPE(cp_dbcsr_p_type), INTENT(INOUT)               :: matrix_w
      REAL(KIND=dp), INTENT(OUT)                         :: kTS
      TYPE(cp_dbcsr_type), INTENT(IN), TARGET            :: matrix_ks, matrix_s
      INTEGER, INTENT(IN)                                :: nelectron_exact
      REAL(KIND=dp), INTENT(OUT)                         :: mu
      INTEGER, INTENT(IN)                                :: iscf, ispin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_matrix_pexsi', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: S_not_identity = 0

      INTEGER :: handle, is_symbolic_factorize, isInertiaCount, isInertiaCount_out, mynode, &
         n_total_inertia_iter, n_total_pexsi_iter, unit_nr
      LOGICAL                                            :: first_call, pexsi_convergence
      REAL(KIND=dp) :: delta_E, energy_H, energy_S, free_energy, mu_max_in, mu_max_out, mu_min_in, &
         mu_min_out, nel_tol, nelectron_diff, nelectron_exact_pexsi, nelectron_out
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: arnoldi_matrices
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dbcsr_arnoldi_data)                           :: my_arnoldi

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      first_call = (iscf .EQ. 1) .AND. (ispin .EQ. 1)

      ! Assert a few things the first time PEXSI is called
      IF (first_call) THEN
         ! Assertion that matrices have the expected symmetry (both should be symmetric if no
         ! S preconditioning and no molecular clustering)
         IF (.NOT. cp_dbcsr_has_symmetry(matrix_ks)) &
            CALL cp__b("pexsi_methods.F",341,"PEXSI interface expects a non-symmetric DBCSR Kohn-Sham matrix")
         IF (.NOT. cp_dbcsr_has_symmetry(matrix_s)) &
            CALL cp__b("pexsi_methods.F",343,"PEXSI interface expects a non-symmetric DBCSR overlap matrix")

         ! Assertion on datatype
         IF ((pexsi_env%csr_mat_s%nzval_local%data_type .NE. dbcsr_type_real_8) &
             .OR. (pexsi_env%csr_mat_ks%nzval_local%data_type .NE. dbcsr_type_real_8)) &
            CALL cp__b("pexsi_methods.F",348,"Complex data type not supported by PEXSI")

         ! Assertion on number of non-zero elements
         !(TODO: update when PEXSI changes to Long Int)
         IF (pexsi_env%csr_mat_s%nze_total .GE. INT(2, kind=int_8)**31) &
            CALL cp__b("pexsi_methods.F",353,"Total number of non-zero elements of CSR matrix is too large to be handled by PEXSI")
      ENDIF

      mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(cp_dbcsr_distribution(matrix_ks)))

      ! Convert DBCSR matrices to PEXSI CSR format. Intermediate step to template matrix
      ! needed in order to retain the initial sparsity pattern that is required for the
      ! conversion to CSR format.
      CALL cp_dbcsr_copy_into_existing(pexsi_env%dbcsr_template_matrix_sym, matrix_s)
      CALL cp_convert_dbcsr_to_csr(pexsi_env%dbcsr_template_matrix_sym, &
                                   pexsi_env%csr_mat_s)

      CALL cp_dbcsr_copy_into_existing(pexsi_env%dbcsr_template_matrix_sym, &
                                       matrix_ks)
      CALL cp_convert_dbcsr_to_csr(pexsi_env%dbcsr_template_matrix_sym, &
                                   pexsi_env%csr_mat_ks)

      ! Get PEXSI input delta_E (upper bound for largest eigenvalue) using Arnoldi
      NULLIFY (arnoldi_matrices)
      CALL cp_dbcsr_allocate_matrix_set(arnoldi_matrices, 2)
      arnoldi_matrices(1)%matrix => matrix_ks
      arnoldi_matrices(2)%matrix => matrix_s
      CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, arnoldi_matrices, max_iter=20, &
                                       threshold=1.0E-2_dp, selection_crit=2, nval_request=1, nrestarts=21, &
                                       generalized_ev=.TRUE., iram=.FALSE.)
      IF (iscf .GT. 1) CALL cp_set_arnoldi_initial_vector(my_arnoldi, &
                                                          pexsi_env%max_ev_vector(ispin))
      CALL cp_dbcsr_arnoldi_ev(arnoldi_matrices, my_arnoldi)
      delta_E = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)
      ! increase delta_E a bit to make sure that it really is an upper bound
      delta_E = delta_E+1.0E-2_dp*ABS(delta_E)
      CALL get_selected_ritz_vec(my_arnoldi, 1, arnoldi_matrices(1)%matrix, &
                                 pexsi_env%max_ev_vector(ispin))
      CALL deallocate_arnoldi_data(my_arnoldi)
      DEALLOCATE (arnoldi_matrices)

      nelectron_exact_pexsi = nelectron_exact

      CALL cp_pexsi_set_options(pexsi_env%options, deltaE=delta_E)

      ! Set PEXSI options appropriately for first SCF iteration
      IF (iscf .EQ. 1) THEN
         ! Get option isInertiaCount to reset it later on and set it to 1 for first SCF iteration
         CALL cp_pexsi_get_options(pexsi_env%options, isInertiaCount=isInertiaCount)
         CALL cp_pexsi_set_options(pexsi_env%options, isInertiaCount=1, &
                                   isSymbolicFactorize=1)
      ENDIF

      ! Write PEXSI options to output
      CALL cp_pexsi_get_options(pexsi_env%options, isInertiaCount=isInertiaCount_out, &
                                isSymbolicFactorize=is_symbolic_factorize, &
                                muMin0=mu_min_in, muMax0=mu_max_in, &
                                NumElectronPEXSITolerance=nel_tol)

!    IF(unit_nr>0) WRITE(unit_nr,'(/A,I4,A,I4)') " PEXSI| SCF", iscf, &
!                                                ", spin component", ispin

      IF (unit_nr > 0) WRITE (unit_nr, '(/A,T41,L20)') " PEXSI| Do inertia counting?", &
         isInertiaCount_out .EQ. 1
      IF (unit_nr > 0) WRITE (unit_nr, '(A,T50,F5.2,T56,F5.2)') &
         " PEXSI| Guess for min mu, max mu", mu_min_in, mu_max_in

      IF (unit_nr > 0) WRITE (unit_nr, '(A,T41,E20.3)') &
         " PEXSI| Tolerance in number of electrons", nel_tol

!    IF(unit_nr>0) WRITE(unit_nr,'(A,T41,L20)') &
!                  " PEXSI|   Do symbolic factorization?", is_symbolic_factorize.EQ.1

      IF (unit_nr > 0) WRITE (unit_nr, '(A,T41,F20.2)') &
         " PEXSI| Arnoldi est. spectral radius", delta_E

      ! Load data into PEXSI
      CALL cp_pexsi_load_real_symmetric_hs_matrix( &
         pexsi_env%plan, &
         pexsi_env%options, &
         pexsi_env%csr_mat_ks%nrows_total, &
         INT(pexsi_env%csr_mat_ks%nze_total, kind=int_4), & ! TODO: update when PEXSI changes to Long Int
         pexsi_env%csr_mat_ks%nze_local, &
         pexsi_env%csr_mat_ks%nrows_local, &
         pexsi_env%csr_mat_ks%rowptr_local, &
         pexsi_env%csr_mat_ks%colind_local, &
         pexsi_env%csr_mat_ks%nzval_local%r_dp, &
         S_not_identity, &
         pexsi_env%csr_mat_s%nzval_local%r_dp)

      ! convert to spin restricted before passing number of electrons to PEXSI
      CALL convert_nspin_cp2k_pexsi(cp2k_to_pexsi, &
                                    numElectron=nelectron_exact_pexsi)

      ! Call DFT driver of PEXSI doing the actual calculation
      CALL cp_pexsi_dft_driver(pexsi_env%plan, pexsi_env%options, &
                               nelectron_exact_pexsi, mu, nelectron_out, mu_min_out, mu_max_out, &
                               n_total_inertia_iter, n_total_pexsi_iter)

      ! Check convergence
      nelectron_diff = nelectron_out-nelectron_exact_pexsi
      pexsi_convergence = ABS(nelectron_diff) .LT. nel_tol

      IF (unit_nr > 0) THEN
         IF (pexsi_convergence) THEN
            WRITE (unit_nr, '(/A)') " PEXSI| Converged"
         ELSE
            WRITE (unit_nr, '(/A)') " PEXSI| PEXSI did not converge!"
         ENDIF

!      WRITE(unit_nr,'(A,T41,F20.10,T61,F20.10)') " PEXSI|   Number of electrons", &
!                      nelectron_out/pexsi_env%nspin, nelectron_diff/pexsi_env%nspin

         WRITE (unit_nr, '(A,T41,F20.6)') " PEXSI|   Chemical potential", mu

         WRITE (unit_nr, '(A,T41,I20)') " PEXSI|   PEXSI iterations", n_total_pexsi_iter
         WRITE (unit_nr, '(A,T41,I20/)') " PEXSI|   Inertia counting iterations", &
            n_total_inertia_iter
      ENDIF

      IF (.NOT. pexsi_convergence) &
         CALL cp__b("pexsi_methods.F",469,"PEXSI did not converge. Consider logPEXSI0 for more information.")

      ! Retrieve results from PEXSI
      IF (mynode < pexsi_env%mp_dims(1)*pexsi_env%mp_dims(2)) THEN
         CALL cp_pexsi_retrieve_real_symmetric_dft_matrix( &
            pexsi_env%plan, &
            pexsi_env%csr_mat_p%nzval_local%r_dp, &
            pexsi_env%csr_mat_E%nzval_local%r_dp, &
            pexsi_env%csr_mat_F%nzval_local%r_dp, &
            energy_H, energy_S, free_energy)
         ! calculate entropic energy contribution -TS = A - U
         kTS = (free_energy-energy_H)
      ENDIF

      ! send kTS to all nodes:
      CALL mp_bcast(kTS, 0, pexsi_env%mp_group)

      ! Convert PEXSI CSR matrices to DBCSR matrices
      CALL cp_convert_csr_to_dbcsr(pexsi_env%dbcsr_template_matrix_nonsym, &
                                   pexsi_env%csr_mat_p)
      CALL cp_dbcsr_copy(matrix_p, pexsi_env%dbcsr_template_matrix_nonsym)
      CALL cp_convert_csr_to_dbcsr(pexsi_env%dbcsr_template_matrix_nonsym, &
                                   pexsi_env%csr_mat_E)
      CALL cp_dbcsr_copy(matrix_w%matrix, pexsi_env%dbcsr_template_matrix_nonsym)

      ! Convert to spin unrestricted
      CALL convert_nspin_cp2k_pexsi(pexsi_to_cp2k, matrix_p=matrix_p, &
                                    matrix_w=matrix_w, kTS=kTS)

      ! Pass resulting mu as initial guess for next SCF to PEXSI
      CALL cp_pexsi_set_options(pexsi_env%options, mu0=mu, muMin0=mu_min_out, &
                                muMax0=mu_max_out)

      ! Reset isInertiaCount according to user input
      IF (iscf .EQ. 1) THEN
         CALL cp_pexsi_set_options(pexsi_env%options, isInertiaCount= &
                                   isInertiaCount)
      ENDIF

      ! Turn off symbolic factorization for subsequent calls
      IF (first_call) THEN
         CALL cp_pexsi_set_options(pexsi_env%options, isSymbolicFactorize=0)
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE density_matrix_pexsi

! **************************************************************************************************
!> \brief Set PEXSI convergence tolerance (numElectronPEXSITolerance), adapted
!>        to the convergence error of the previous SCF step. The tolerance is
!>        calculated using an initial convergence threshold (tol_nel_initial)
!>        and a target convergence threshold (tol_nel_target):
!>        numElectronPEXSITolerance(delta_scf) = alpha*delta_scf+beta
!>        where alpha and beta are chosen such that
!>        numElectronPEXSITolerance(delta_scf_0) = tol_nel_initial
!>        numElectronPEXSITolerance(eps_scf) = tol_nel_target
!>        and delta_scf_0 is the initial SCF convergence error.
!> \param pexsi_env ...
!> \param delta_scf convergence error of previous SCF step
!> \param eps_scf SCF convergence criterion
!> \param initialize whether or not adaptive thresholing should be initialized
!>        with delta_scf as initial convergence error
!> \param check_convergence is set to .FALSE. if convergence in number of electrons
!>        will not be achieved in next SCF step
! **************************************************************************************************
   SUBROUTINE pexsi_set_convergence_tolerance(pexsi_env, delta_scf, eps_scf, initialize, &
                                              check_convergence)
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env
      REAL(KIND=dp), INTENT(IN)                          :: delta_scf, eps_scf
      LOGICAL, INTENT(IN)                                :: initialize
      LOGICAL, INTENT(OUT)                               :: check_convergence

      CHARACTER(len=*), PARAMETER :: routineN = 'pexsi_set_convergence_tolerance', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: tol_nel

      CALL timeset(routineN, handle)

      tol_nel = pexsi_env%tol_nel_initial

      IF (initialize) THEN
         pexsi_env%adaptive_nel_alpha = &
            (pexsi_env%tol_nel_initial-pexsi_env%tol_nel_target)/(ABS(delta_scf)-eps_scf)
         pexsi_env%adaptive_nel_beta = &
            pexsi_env%tol_nel_initial-pexsi_env%adaptive_nel_alpha*ABS(delta_scf)
         pexsi_env%do_adaptive_tol_nel = .TRUE.
      ENDIF
      IF (pexsi_env%do_adaptive_tol_nel) THEN
         tol_nel = pexsi_env%adaptive_nel_alpha*ABS(delta_scf)+pexsi_env%adaptive_nel_beta
         tol_nel = MAX(tol_nel, pexsi_env%tol_nel_target)
         tol_nel = MIN(tol_nel, pexsi_env%tol_nel_initial)
      ENDIF

      check_convergence = (tol_nel .LE. pexsi_env%tol_nel_target)

      CALL cp_pexsi_set_options(pexsi_env%options, numElectronPEXSITolerance=tol_nel)
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Pass energy weighted density matrix and entropic energy contribution
!>        to QS environment
!> \param ls_scf_env ...
!> \param qs_env ...
!> \param kTS ...
!> \param matrix_w ...
!> \par History
!>      12.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE pexsi_to_qs(ls_scf_env, qs_env, kTS, matrix_w)
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      TYPE(qs_environment_type), INTENT(INOUT), POINTER  :: qs_env
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: kTS
      TYPE(cp_dbcsr_p_type), DIMENSION(:), INTENT(IN), &
         OPTIONAL                                        :: matrix_w

      CHARACTER(len=*), PARAMETER :: routineN = 'pexsi_to_qs', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      REAL(KIND=dp)                                      :: kTS_total
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_w_qs
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_energy_type), POINTER                      :: energy

      CALL timeset(routineN, handle)

      NULLIFY (energy)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL get_qs_env(qs_env, energy=energy, matrix_w=matrix_w_qs)

      IF (PRESENT(matrix_w)) THEN
         DO ispin = 1, ls_scf_env%nspins
            CALL matrix_ls_to_qs(matrix_w_qs(ispin)%matrix, matrix_w(ispin)%matrix, &
                                 ls_scf_env%ls_mstruct, covariant=.FALSE.)
            IF (ls_scf_env%nspins .EQ. 1) CALL cp_dbcsr_scale(matrix_w_qs(ispin)%matrix, 2.0_dp)
         ENDDO
      ENDIF

      IF (PRESENT(kTS)) THEN
         kTS_total = SUM(kTS)
         IF (ls_scf_env%nspins .EQ. 1) kTS_total = kTS_total*2.0_dp
         energy%kTS = kTS_total
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE pexsi_to_qs

END MODULE pexsi_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines using linear scaling chebyshev methods
!> \par History
!>       2012.10 created [Jinwoong Cha]
!> \author Jinwoong Cha
! **************************************************************************************************
MODULE dm_ls_chebyshev
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_extremal, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_frobenius_norm, cp_dbcsr_get_info, cp_dbcsr_get_occupation, &
        cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, &
        cp_dbcsr_trace, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE dm_ls_scf_qs,                    ONLY: write_matrix_to_cube
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathconstants,                   ONLY: pi
   USE qs_environment_types,            ONLY: qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_chebyshev'

   PUBLIC :: compute_chebyshev

CONTAINS

! **************************************************************************************************
!> \brief compute chebyshev polynomials up to order n for a given value of x
!> \param value ...
!> \param x ...
!> \param n ...
!> \par History
!>       2012.11 created [Jinwoong Cha]
!> \author Jinwoong Cha
! **************************************************************************************************
   SUBROUTINE chebyshev_poly(value, x, n)
      REAL(KIND=dp), INTENT(OUT)                         :: value
      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: n

!polynomial values
!number of chev polynomials

      value = COS((n-1)*ACOS(x))

   END SUBROUTINE chebyshev_poly

! **************************************************************************************************
!> \brief kernel for chebyshev polynomials expansion (Jackson kernel)
!> \param value ...
!> \param n ...
!> \param nc ...
!> \par History
!>       2012.11 created [Jinwoong Cha]
!> \author Jinwoong Cha
! **************************************************************************************************
   SUBROUTINE kernel(value, n, nc)
      REAL(KIND=dp), INTENT(OUT)                         :: value
      INTEGER, INTENT(IN)                                :: n, nc

!kernel at n
!n-1 order of chebyshev polynomials
!number of total chebyshev polynomials
!Kernel define

      value = 1.0_dp/(nc+1.0_dp)*((nc-(n-1)+1.0_dp)* &
                                  COS(pi*(n-1)/(nc+1.0_dp))+SIN(pi*(n-1)/(nc+1.0_dp))*1.0_dp/TAN(pi/(nc+1.0_dp)))

   END SUBROUTINE kernel

! **************************************************************************************************
!> \brief compute properties based on chebyshev expansion
!> \param qs_env ...
!> \param ls_scf_env ...
!> \par History
!>       2012.10 created [Jinwoong Cha]
!> \author Jinwoong Cha
! **************************************************************************************************
   SUBROUTINE compute_chebyshev(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_chebyshev', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: scale_evals = 1.01_dp

      CHARACTER(LEN=30)                                  :: middle_name
      CHARACTER(LEN=default_string_length)               :: title
      INTEGER                                            :: handle, icheb, igrid, iinte, ispin, &
                                                            iwindow, n_gridpoint_dos, ncheb, &
                                                            ninte, Nrows, nwindow, unit_cube, &
                                                            unit_dos, unit_nr
      LOGICAL                                            :: converged, write_cubes
      REAL(KIND=dp) :: chev_T, chev_T_dos, dummy1, final, frob_matrix, initial, interval_a, &
         interval_b, max_ev, min_ev, occ, orbital_occ, summa, t1, t2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: chev_E, chev_Es_dos, dos, dummy2, ev1, &
                                                            ev2, kernel_g, mu, sev1, sev2, trace_dm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: aitchev_T, E_inte, gdensity, sqrt_vec
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp_r
      TYPE(cp_dbcsr_type)                                :: matrix_dummy1, matrix_F, matrix_tmp1, &
                                                            matrix_tmp2, matrix_tmp3
      TYPE(cp_dbcsr_type), DIMENSION(:), POINTER         :: matrix_dummy2
      TYPE(cp_logger_type), POINTER                      :: logger

      IF (.NOT. ls_scf_env%chebyshev%compute_chebyshev) RETURN

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ncheb = ls_scf_env%chebyshev%n_chebyshev
      ninte = 2*ncheb
      n_gridpoint_dos = ls_scf_env%chebyshev%n_gridpoint_dos

      write_cubes = BTEST(cp_print_key_should_output(logger%iter_info, ls_scf_env%chebyshev%print_key_cube), cp_p_file)
      IF (write_cubes) THEN
         IF (ASSOCIATED(ls_scf_env%chebyshev%min_energy)) DEALLOCATE (ls_scf_env%chebyshev%min_energy)
         CALL section_vals_val_get(ls_scf_env%chebyshev%print_key_cube, "MIN_ENERGY", r_vals=tmp_r)
         ALLOCATE (ls_scf_env%chebyshev%min_energy(SIZE(tmp_r)))
         ls_scf_env%chebyshev%min_energy = tmp_r

         IF (ASSOCIATED(ls_scf_env%chebyshev%max_energy)) DEALLOCATE (ls_scf_env%chebyshev%max_energy)
         CALL section_vals_val_get(ls_scf_env%chebyshev%print_key_cube, "MAX_ENERGY", r_vals=tmp_r)
         ALLOCATE (ls_scf_env%chebyshev%max_energy(SIZE(tmp_r)))
         ls_scf_env%chebyshev%max_energy = tmp_r

         nwindow = SIZE(ls_scf_env%chebyshev%min_energy)
      ELSE
         nwindow = 0
      ENDIF

      ALLOCATE (ev1(1:nwindow))
      ALLOCATE (ev2(1:nwindow))
      ALLOCATE (sev1(1:nwindow))
      ALLOCATE (sev2(1:nwindow))
      ALLOCATE (trace_dm(1:nwindow))
      ALLOCATE (matrix_dummy2(1:nwindow))

      DO iwindow = 1, nwindow
         ev1(iwindow) = ls_scf_env%chebyshev%min_energy(iwindow)
         ev2(iwindow) = ls_scf_env%chebyshev%max_energy(iwindow)
      END DO

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '()')
         WRITE (unit_nr, '(T2,A)') "STARTING CHEBYSHEV CALCULATION"
      ENDIF

      ! create 3 temporary matrices
      CALL cp_dbcsr_init(matrix_tmp1) !initialization
      CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp2)
      CALL cp_dbcsr_create(matrix_tmp2, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp3)
      CALL cp_dbcsr_create(matrix_tmp3, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_F)
      CALL cp_dbcsr_create(matrix_F, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_dummy1)
      CALL cp_dbcsr_create(matrix_dummy1, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)

      DO iwindow = 1, nwindow
         CALL cp_dbcsr_init(matrix_dummy2(iwindow)) ! density matrix
         CALL cp_dbcsr_create(matrix_dummy2(iwindow), template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
      END DO

      DO ispin = 1, SIZE(ls_scf_env%matrix_ks)
         ! create matrix_F=inv(sqrt(S))*H*inv(sqrt(S))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_ks(ispin), &
                                0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt_inv, &
                                0.0_dp, matrix_F, filter_eps=ls_scf_env%eps_filter)

         ! find largest and smallest eigenvalues
         CALL cp_dbcsr_arnoldi_extremal(matrix_F, max_ev, min_ev, converged=converged, max_iter=ls_scf_env%max_iter_lanczos, &
                                        threshold=ls_scf_env%eps_lanczos) !Lanczos algorithm to calculate eigenvalue
         IF (unit_nr > 0) WRITE (unit_nr, '(T2,A,2F16.8,A,L2)') &
            "smallest largest eigenvalue", min_ev, max_ev, " converged ", converged
         IF (nwindow > 0) THEN
            IF (unit_nr > 0) WRITE (unit_nr, '(T2,A,1000F16.8)') "requested interval-min_energy", ev1(:)
            IF (unit_nr > 0) WRITE (unit_nr, '(T2,A,1000F16.8)') "requested interval-max_energy", ev2(:)
         ENDIF
         interval_a = (max_ev-min_ev)*scale_evals/2
         interval_b = (max_ev+min_ev)/2

         sev1(:) = (ev1(:)-interval_b)/interval_a !scaled ev1 vector
         sev2(:) = (ev2(:)-interval_b)/interval_a !scaled ev2 vector

         !chebyshev domain,pi*sqrt(1-x^2) vector construction and chebyshev polynomials for integration (for g(E))
         ALLOCATE (E_inte(1:ninte+1, 1:nwindow))
         ALLOCATE (sqrt_vec(1:ninte+1, 1:nwindow))

         DO iwindow = 1, nwindow
            DO iinte = 1, ninte+1
               E_inte(iinte, iwindow) = sev1(iwindow)+((sev2(iwindow)-sev1(iwindow))/ninte)*(iinte-1)
               sqrt_vec(iinte, iwindow) = pi*SQRT(1.0_dp-E_inte(iinte, iwindow)*E_inte(iinte, iwindow))
            END DO
         END DO

         !integral.. (identical to the coefficient for g(E))

         ALLOCATE (aitchev_T(1:ncheb, 1:nwindow)) !after intergral. =>ainte

         DO iwindow = 1, nwindow
            DO icheb = 1, ncheb
               CALL chebyshev_poly(initial, E_inte(1, iwindow), icheb)
               CALL chebyshev_poly(final, E_inte(1, iwindow), icheb)
               summa = (sev2(iwindow)-sev1(iwindow))/(2.0_dp*ninte)*(initial/sqrt_vec(1, iwindow)+final/sqrt_vec(ninte+1, iwindow))
               DO iinte = 2, ninte
                  CALL chebyshev_poly(chev_T, E_inte(iinte, iwindow), icheb)
                  summa = summa+((sev2(iwindow)-sev1(iwindow))/ninte)*(chev_T/sqrt_vec(iinte, iwindow))
               END DO
               aitchev_T(icheb, iwindow) = summa
               summa = 0
            END DO
         END DO

         ! scale the matrix to get evals in the interval -1,1
         CALL cp_dbcsr_add_on_diag(matrix_F, -interval_b)
         CALL cp_dbcsr_scale(matrix_F, 1/interval_a)

         ! compute chebyshev matrix recursion
         CALL cp_dbcsr_get_info(matrix=matrix_F, nfullrows_total=Nrows) !get information about a matrix
         CALL cp_dbcsr_set(matrix_dummy1, 0.0_dp) !empty matrix creation(for density matrix)

         DO iwindow = 1, nwindow
            CALL cp_dbcsr_set(matrix_dummy2(iwindow), 0.0_dp) !empty matrix creation(for density matrix)
         END DO

         ALLOCATE (mu(1:ncheb))
         ALLOCATE (kernel_g(1:ncheb))
         CALL kernel(kernel_g(1), 1, ncheb)
         CALL kernel(kernel_g(2), 2, ncheb)

         CALL cp_dbcsr_set(matrix_tmp1, 0.0_dp) !matrix creation
         CALL cp_dbcsr_add_on_diag(matrix_tmp1, 1.0_dp) !add a only number to diagonal elements
         CALL cp_dbcsr_trace(matrix_tmp1, trace=mu(1))
         CALL cp_dbcsr_copy(matrix_tmp2, matrix_F) !make matrix_tmp2 = matrix_F
         CALL cp_dbcsr_trace(matrix_tmp2, trace=mu(2))

         DO iwindow = 1, nwindow
            CALL cp_dbcsr_copy(matrix_dummy1, matrix_tmp1)
            CALL cp_dbcsr_copy(matrix_dummy2(iwindow), matrix_tmp2) !matrix_dummy2=
            CALL cp_dbcsr_scale(matrix_dummy1, kernel_g(1)*aitchev_T(1, iwindow)) !first term of chebyshev poly(matrix)
            CALL cp_dbcsr_scale(matrix_dummy2(iwindow), 2.0_dp*kernel_g(2)*aitchev_T(2, iwindow)) !second term of chebyshev poly(matrix)

            CALL cp_dbcsr_add(matrix_dummy2(iwindow), matrix_dummy1, 1.0_dp, 1.0_dp)
         END DO

         DO icheb = 2, ncheb-1
            t1 = m_walltime()
            CALL cp_dbcsr_multiply("N", "N", 2.0_dp, matrix_F, matrix_tmp2, &
                                   -1.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter) !matrix multiplication(Recursion)
            CALL cp_dbcsr_copy(matrix_tmp3, matrix_tmp1)
            CALL cp_dbcsr_copy(matrix_tmp1, matrix_tmp2)
            CALL cp_dbcsr_copy(matrix_tmp2, matrix_tmp3)
            CALL cp_dbcsr_trace(matrix_tmp2, trace=mu(icheb+1)) !icheb+1 th coefficient
            CALL kernel(kernel_g(icheb+1), icheb+1, ncheb)

            DO iwindow = 1, nwindow

               CALL cp_dbcsr_copy(matrix_dummy1, matrix_tmp2)
               CALL cp_dbcsr_scale(matrix_dummy1, 2.0_dp*kernel_g(icheb+1)*aitchev_T(icheb+1, iwindow)) !second term of chebyshev poly(matrix)
               CALL cp_dbcsr_add(matrix_dummy2(iwindow), matrix_dummy1, 1.0_dp, 1.0_dp)
               CALL cp_dbcsr_trace(matrix_dummy2(iwindow), trace=trace_dm(iwindow)) !icheb+1 th coefficient

            END DO

            occ = cp_dbcsr_get_occupation(matrix_tmp1)
            t2 = m_walltime()
            IF (unit_nr > 0 .AND. MOD(icheb, 20) == 0) THEN
               CALL m_flush(unit_nr)
               IF (nwindow > 0) THEN
                  WRITE (unit_nr, '(T2,A,I5,1X,A,1X,F8.3,1X,A,1X,F8.6,1X,A,1X,1000F16.8)') &
                     "Iter.", icheb, "time=", t2-t1, "occ=", occ, "traces=", trace_dm(:)
               ELSE
                  WRITE (unit_nr, '(T2,A,I5,1X,A,1X,F8.3,1X,A,1X,F8.6)') &
                     "Iter.", icheb, "time=", t2-t1, "occ=", occ
               ENDIF
            ENDIF
         ENDDO

         DO iwindow = 1, nwindow
            IF (SIZE(ls_scf_env%matrix_ks) == 1) THEN
               orbital_occ = 2.0_dp
            ELSE
               orbital_occ = 1.0_dp
            ENDIF
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, matrix_dummy2(iwindow), &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", orbital_occ, matrix_tmp1, ls_scf_env%matrix_s_sqrt_inv, &
                                   0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)
            CALL cp_dbcsr_copy(matrix_dummy2(iwindow), matrix_tmp2)

            ! look at the difference with the density matrix from the ls routines
            IF (.FALSE.) THEN
               CALL cp_dbcsr_copy(matrix_tmp1, matrix_tmp2)
               CALL cp_dbcsr_add(matrix_tmp1, ls_scf_env%matrix_p(ispin), 1.0_dp, -1.0_dp) !comparison
               frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
               IF (unit_nr > 0) WRITE (unit_nr, *) "Difference between Chebyshev DM and LS DM", frob_matrix
            ENDIF
         END DO

         write_cubes = BTEST(cp_print_key_should_output(logger%iter_info, &
                                                        ls_scf_env%chebyshev%print_key_cube), cp_p_file)
         IF (write_cubes) THEN
            DO iwindow = 1, nwindow
               WRITE (middle_name, "(A,I0)") "E_DENSITY_WINDOW_", iwindow
               WRITE (title, "(A,1X,F16.8,1X,A,1X,F16.8)") "Energy range : ", ev1(iwindow), "to", ev2(iwindow)
               unit_cube = cp_print_key_unit_nr(logger, ls_scf_env%chebyshev%print_key_cube, &
                                                "", extension=".cube", & !added 01/22/2012
                                                middle_name=TRIM(middle_name), log_filename=.FALSE.)
               CALL write_matrix_to_cube(qs_env, ls_scf_env, matrix_dummy2(iwindow), unit_cube, title, &
                                         section_get_ivals(ls_scf_env%chebyshev%print_key_cube, "STRIDE"))
               CALL cp_print_key_finished_output(unit_cube, logger, ls_scf_env%chebyshev%print_key_cube, "")
            END DO
         ENDIF

      ENDDO

      ! Chebyshev expansion with calculated coefficient
      ! grid construction and rescaling (by J)
      unit_dos = cp_print_key_unit_nr(logger, ls_scf_env%chebyshev%print_key_dos, "", extension=".xy", &
                                      middle_name="DOS", log_filename=.FALSE.)

      IF (unit_dos > 0) THEN
         ALLOCATE (dos(1:n_gridpoint_dos))
         ALLOCATE (gdensity(1:n_gridpoint_dos, 1:nwindow))
         ALLOCATE (chev_E(1:n_gridpoint_dos))
         ALLOCATE (chev_Es_dos(1:n_gridpoint_dos))
         ALLOCATE (dummy2(1:nwindow))
         DO igrid = 1, n_gridpoint_dos
            chev_E(igrid) = min_ev+(igrid-1)*(max_ev-min_ev)/(n_gridpoint_dos-1)
            chev_Es_dos(igrid) = (chev_E(igrid)-interval_b)/interval_a
         END DO
         DO igrid = 1, n_gridpoint_dos
            dummy1 = 0.0_dp !summation of polynomials
            dummy2(:) = 0.0_dp !summation of polynomials
            DO icheb = 2, ncheb
               CALL chebyshev_poly(chev_T_dos, chev_Es_dos(igrid), icheb)
               dummy1 = dummy1+kernel_g(icheb)*mu(icheb)*chev_T_dos
               DO iwindow = 1, nwindow
                  dummy2(iwindow) = dummy2(iwindow)+kernel_g(icheb)*aitchev_T(icheb, iwindow)*chev_T_dos
               END DO
            END DO
            dos(igrid) = 1.0_dp/(interval_a*Nrows* &
                                 (pi*SQRT(1.0_dp-chev_Es_dos(igrid)*chev_Es_dos(igrid))))*(kernel_g(1)*mu(1)+2.0_dp*dummy1)
            DO iwindow = 1, nwindow
               gdensity(igrid, iwindow) = kernel_g(1)*aitchev_T(1, iwindow)+2.0_dp*dummy2(iwindow)
            END DO
            WRITE (unit_dos, '(1000F16.8)') chev_E(igrid), dos(igrid), gdensity(igrid, :)
         END DO
         DEALLOCATE (chev_Es_dos, chev_E, dos, gdensity)
      ENDIF
      CALL cp_print_key_finished_output(unit_dos, logger, ls_scf_env%chebyshev%print_key_dos, "")

      ! free the matrices
      CALL cp_dbcsr_release(matrix_tmp1)
      CALL cp_dbcsr_release(matrix_tmp2)
      CALL cp_dbcsr_release(matrix_tmp3)
      CALL cp_dbcsr_release(matrix_F)
      CALL cp_dbcsr_release(matrix_dummy1)

      DO iwindow = 1, nwindow
         CALL cp_dbcsr_release(matrix_dummy2(iwindow))
      END DO

      DEALLOCATE (ev1, ev2, sev1, sev2, matrix_dummy2)

      !Need deallocation
      DEALLOCATE (mu, kernel_g, aitchev_T, E_inte, sqrt_vec)

      IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') "ENDING CHEBYSHEV CALCULATION"

      CALL timestop(handle)

   END SUBROUTINE compute_chebyshev

END MODULE dm_ls_chebyshev
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for a linear scaling quickstep SCF run based on the density
!>        matrix
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE dm_ls_scf
   USE bibliography,                    ONLY: Kolafa2004,&
                                              Lin2009,&
                                              Lin2013,&
                                              Niklasson2003,&
                                              Niklasson2014,&
                                              Shao2003,&
                                              VandeVondele2012,&
                                              cite_reference
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_arnoldi_extremal, cp_dbcsr_binary_read, cp_dbcsr_binary_write, &
        cp_dbcsr_checksum, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_filter, &
        cp_dbcsr_get_occupation, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_env,                     ONLY: cp_para_env_retain
   USE dm_ls_chebyshev,                 ONLY: compute_chebyshev
   USE dm_ls_scf_curvy,                 ONLY: deallocate_curvy_data,&
                                              dm_ls_curvy_optimization
   USE dm_ls_scf_methods,               ONLY: apply_matrix_preconditioner,&
                                              compute_homo_lumo,&
                                              density_matrix_sign,&
                                              density_matrix_sign_fixed_mu,&
                                              density_matrix_tc2,&
                                              density_matrix_trs4,&
                                              ls_scf_init_matrix_S
   USE dm_ls_scf_qs,                    ONLY: ls_scf_dm_to_ks,&
                                              ls_scf_init_qs,&
                                              ls_scf_qs_atomic_guess,&
                                              matrix_ls_create,&
                                              matrix_ls_to_qs,&
                                              matrix_qs_to_ls,&
                                              rho_mixing_ls_init
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type
   USE input_constants,                 ONLY: &
        ls_cluster_atomic, ls_cluster_molecular, ls_cluster_pao, ls_s_inversion_hotelling, &
        ls_s_inversion_none, ls_s_inversion_sign_sqrt, ls_s_preconditioner_atomic, &
        ls_s_preconditioner_molecular, ls_s_preconditioner_none, ls_scf_ns, ls_scf_pexsi, &
        ls_scf_tc2, ls_scf_trs4, transport_transmission
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE iterate_matrix,                  ONLY: purify_mcweeny
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathlib,                         ONLY: binomial
   USE molecule_types_new,              ONLY: molecule_of_atom,&
                                              molecule_type
   USE pao_main,                        ONLY: pao_finalize,&
                                              pao_init,&
                                              pao_update
   USE particle_types,                  ONLY: particle_type
   USE pexsi_methods,                   ONLY: density_matrix_pexsi,&
                                              pexsi_finalize_scf,&
                                              pexsi_init_read_input,&
                                              pexsi_init_scf,&
                                              pexsi_set_convergence_tolerance,&
                                              pexsi_to_qs
   USE pexsi_types,                     ONLY: lib_pexsi_init
   USE qs_density_mixing_types,         ONLY: create_mixing_section,&
                                              mixing_storage_create
   USE qs_diis,                         ONLY: qs_diis_b_clear_sparse,&
                                              qs_diis_b_create_sparse,&
                                              qs_diis_b_step_4lscf
   USE qs_diis_types,                   ONLY: qs_diis_b_release_sparse,&
                                              qs_diis_buffer_type_sparse
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_scf_post_gpw,                 ONLY: qs_scf_post_moments,&
                                              write_mo_free_results
   USE transport,                       ONLY: external_scf_method,&
                                              transport_initialize
   USE transport_env_types,             ONLY: transport_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_scf'

   PUBLIC :: ls_scf, ls_scf_create

CONTAINS

! **************************************************************************************************
!> \brief perform an linear scaling scf procedure: entry point
!>
!> \param qs_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: pao_is_done
      TYPE(ls_scf_env_type), POINTER                     :: ls_scf_env

      CALL timeset(routineN, handle)
      CALL get_qs_env(qs_env, ls_scf_env=ls_scf_env)
      CALL pao_init(qs_env, ls_scf_env)

      pao_is_done = .FALSE.
      DO WHILE (.NOT. pao_is_done)
         CALL ls_scf_init_scf(qs_env, ls_scf_env)
         CALL pao_update(qs_env, ls_scf_env, pao_is_done)
         CALL ls_scf_main(qs_env, ls_scf_env)
         CALL ls_scf_post(qs_env, ls_scf_env)
      END DO

      CALL pao_finalize(ls_scf_env)

      CALL timestop(handle)

   END SUBROUTINE ls_scf

! **************************************************************************************************
!> \brief Creation and basic initialization of the LS type.
!> \param qs_env ...
!> \par History
!>       2012.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_create(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ls_scf_env_type), POINTER                     :: ls_scf_env
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: input, mixing_section

      CALL timeset(routineN, handle)

      CALL cite_reference(VandeVondele2012)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ALLOCATE (ls_scf_env)

      ! get basic quantities from the qs_env
      CALL get_qs_env(qs_env, nelectron_total=ls_scf_env%nelectron_total, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      molecule_set=molecule_set, &
                      input=input, &
                      has_unit_metric=ls_scf_env%has_unit_metric, &
                      para_env=ls_scf_env%para_env, &
                      do_transport=ls_scf_env%do_transport, &
                      nelectron_spin=ls_scf_env%nelectron_spin)

      ! copy some basic stuff
      ls_scf_env%nspins = dft_control%nspins
      ls_scf_env%natoms = SIZE(particle_set, 1)
      CALL cp_para_env_retain(ls_scf_env%para_env)

      ! initialize block to group to defined molecules
      ALLOCATE (ls_scf_env%ls_mstruct%atom_to_molecule(ls_scf_env%natoms))

      CALL molecule_of_atom(molecule_set, atom_to_mol=ls_scf_env%ls_mstruct%atom_to_molecule)

      ! parse the ls_scf section and set derived quantities
      CALL ls_scf_init_read_write_input(input, ls_scf_env, unit_nr)

      ! set up the buffer for the history of matrices
      ls_scf_env%scf_history%nstore = ls_scf_env%extrapolation_order
      ls_scf_env%scf_history%istore = 0
      ALLOCATE (ls_scf_env%scf_history%matrix(ls_scf_env%nspins, ls_scf_env%scf_history%nstore))

      NULLIFY (ls_scf_env%mixing_store)
      mixing_section => section_vals_get_subs_vals(input, "DFT%LS_SCF%RHO_MIXING")
      CALL mixing_storage_create(ls_scf_env%mixing_store, mixing_section, &
                                 ls_scf_env%density_mixing_method, &
                                 dft_control%qs_control%cutoff)

      ! initialize PEXSI
      IF (ls_scf_env%do_pexsi) THEN
         IF (dft_control%qs_control%eps_filter_matrix .NE. 0.0_dp) &
            CALL cp__b("dm_ls_scf.F",225,"EPS_FILTER_MATRIX must be set to 0 for PEXSI.")
         CALL lib_pexsi_init(ls_scf_env%pexsi, ls_scf_env%para_env%group, ls_scf_env%nspins)
      ENDIF

      ! put the ls_scf_env in qs_env
      CALL set_qs_env(qs_env, ls_scf_env=ls_scf_env)

      CALL timestop(handle)

   END SUBROUTINE ls_scf_create

! **************************************************************************************************
!> \brief initialization needed for scf
!> \param qs_env ...
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_init_scf(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_init_scf', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! get basic quantities from the qs_env
      CALL get_qs_env(qs_env, nelectron_total=ls_scf_env%nelectron_total, &
                      matrix_s=matrix_s, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      molecule_set=molecule_set, &
                      input=input, &
                      has_unit_metric=ls_scf_env%has_unit_metric, &
                      para_env=ls_scf_env%para_env, &
                      nelectron_spin=ls_scf_env%nelectron_spin)

      ! some basic initialization of the QS side of things
      CALL ls_scf_init_qs(qs_env)

      ! create the matrix template for use in the ls procedures
      CALL matrix_ls_create(matrix_ls=ls_scf_env%matrix_s, matrix_qs=matrix_s(1)%matrix, &
                            ls_mstruct=ls_scf_env%ls_mstruct)

      nspin = ls_scf_env%nspins
      ALLOCATE (ls_scf_env%matrix_p(nspin))
      DO ispin = 1, nspin
         CALL cp_dbcsr_init(ls_scf_env%matrix_p(ispin))
         CALL cp_dbcsr_create(ls_scf_env%matrix_p(ispin), template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
      ENDDO

      ALLOCATE (ls_scf_env%matrix_ks(nspin))
      DO ispin = 1, nspin
         CALL cp_dbcsr_init(ls_scf_env%matrix_ks(ispin))
         CALL cp_dbcsr_create(ls_scf_env%matrix_ks(ispin), template=ls_scf_env%matrix_s, &
                              matrix_type=dbcsr_type_no_symmetry)
      ENDDO

      ! set up matrix S, and needed functions of S
      CALL ls_scf_init_matrix_s(matrix_s(1)%matrix, ls_scf_env)

      ! get the initial guess for the SCF
      CALL ls_scf_initial_guess(qs_env, ls_scf_env)

      IF (ls_scf_env%do_rho_mixing) THEN
         CALL rho_mixing_ls_init(qs_env, ls_scf_env)
      ENDIF

      IF (ls_scf_env%do_pexsi) THEN
         CALL pexsi_init_scf(ks_env, ls_scf_env%pexsi, matrix_s(1)%matrix)
      ENDIF

      IF (qs_env%do_transport) THEN
         CALL transport_initialize(ks_env, qs_env%transport_env, matrix_s(1)%matrix)
      END IF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_init_scf

! **************************************************************************************************
!> \brief deal with the scf initial guess
!> \param qs_env ...
!> \param ls_scf_env ...
!> \par History
!>       2012.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_initial_guess(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_initial_guess', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: aspc_guess = 2, atomic_guess = 1, &
                                                            restart_guess = 3

      CHARACTER(LEN=default_path_length)                 :: file_name, project_name
      INTEGER                                            :: handle, iaspc, initial_guess_type, &
                                                            ispin, istore, naspc, unit_nr
      REAL(KIND=dp)                                      :: alpha, cs_pos
      TYPE(cp_dbcsr_type)                                :: matrix_tmp1
      TYPE(cp_logger_type), POINTER                      :: logger

      IF (ls_scf_env%do_pao) RETURN ! pao has its own initial guess

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (unit_nr > 0) WRITE (unit_nr, '()')
      ! if there is no history go for the atomic guess, otherwise extrapolate the dm history
      IF (ls_scf_env%scf_history%istore == 0) THEN
         IF (ls_scf_env%restart_read) THEN
            initial_guess_type = restart_guess
         ELSE
            initial_guess_type = atomic_guess
         ENDIF
      ELSE
         initial_guess_type = aspc_guess
      ENDIF

      ! how to get the initial guess
      SELECT CASE (initial_guess_type)
      CASE (atomic_guess)
         CALL ls_scf_qs_atomic_guess(qs_env, ls_scf_env%energy_init)
      CASE (restart_guess)
         project_name = logger%iter_info%project_name
         DO ispin = 1, SIZE(ls_scf_env%matrix_p)
            WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_", ispin, "_RESTART.dm"
            CALL cp_dbcsr_binary_read(file_name, distribution=cp_dbcsr_distribution(ls_scf_env%matrix_p(1)), &
                                      matrix_new=ls_scf_env%matrix_p(ispin))
            cs_pos = cp_dbcsr_checksum(ls_scf_env%matrix_p(ispin), pos=.TRUE.)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T2,A,E20.8)') "Read restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
            ENDIF
         ENDDO

         ! directly go to computing the corresponding energy and ks matrix
         CALL ls_scf_dm_to_ks(qs_env, ls_scf_env, ls_scf_env%energy_init, iscf=0)
      CASE (aspc_guess)
         CALL cite_reference(Kolafa2004)
         naspc = MIN(ls_scf_env%scf_history%istore, ls_scf_env%scf_history%nstore)
         DO ispin = 1, SIZE(ls_scf_env%matrix_p)
            ! actual extrapolation
            CALL cp_dbcsr_set(ls_scf_env%matrix_p(ispin), 0.0_dp)
            DO iaspc = 1, naspc
               alpha = (-1.0_dp)**(iaspc+1)*REAL(iaspc, KIND=dp)* &
                       binomial(2*naspc, naspc-iaspc)/binomial(2*naspc-2, naspc-1)
               istore = MOD(ls_scf_env%scf_history%istore-iaspc, ls_scf_env%scf_history%nstore)+1
               CALL cp_dbcsr_add(ls_scf_env%matrix_p(ispin), ls_scf_env%scf_history%matrix(ispin, istore), 1.0_dp, alpha)
            ENDDO
         ENDDO
      END SELECT

      ! which cases need getting purified and non-orthogonal ?
      SELECT CASE (initial_guess_type)
      CASE (atomic_guess, restart_guess)
         ! do nothing
      CASE (aspc_guess)
         ! purification can't be done on the pexsi matrix, which is not necessarily idempotent,
         ! and not stored in an ortho basis form
         IF (.NOT. (ls_scf_env%do_pexsi)) THEN
            DO ispin = 1, SIZE(ls_scf_env%matrix_p)
               ! linear combination of P's is not idempotent. A bit of McWeeny is needed to ensure it is again
               IF (SIZE(ls_scf_env%matrix_p) == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(ispin), 0.5_dp)
               CALL purify_mcweeny(ls_scf_env%matrix_p(ispin:ispin), ls_scf_env%eps_filter, 3)
               IF (SIZE(ls_scf_env%matrix_p) == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(ispin), 2.0_dp)

               IF (ls_scf_env%use_s_sqrt) THEN
                  ! need to get P in the non-orthogonal basis if it was stored differently
                  CALL cp_dbcsr_init(matrix_tmp1)
                  CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_p(ispin), &
                                         0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt_inv, &
                                         0.0_dp, ls_scf_env%matrix_p(ispin), &
                                         filter_eps=ls_scf_env%eps_filter)
                  CALL cp_dbcsr_release(matrix_tmp1)

                  IF (ls_scf_env%has_s_preconditioner) THEN
                     CALL apply_matrix_preconditioner(ls_scf_env%matrix_p(ispin), "forward", &
                                                      ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
                  ENDIF
               ENDIF
            ENDDO
         ENDIF

         ! compute corresponding energy and ks matrix
         CALL ls_scf_dm_to_ks(qs_env, ls_scf_env, ls_scf_env%energy_init, iscf=0)
      END SELECT

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T2,A,F20.9)') "Energy with the initial guess:", ls_scf_env%energy_init
         WRITE (unit_nr, '()')
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_initial_guess

! **************************************************************************************************
!> \brief store a history of matrices for later use in ls_scf_initial_guess
!> \param ls_scf_env ...
!> \par History
!>       2012.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_store_result(ls_scf_env)
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_store_result', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, project_name
      INTEGER                                            :: handle, ispin, istore, unit_nr
      REAL(KIND=dp)                                      :: cs_pos
      TYPE(cp_dbcsr_type)                                :: matrix_tmp1
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (ls_scf_env%restart_write) THEN
         DO ispin = 1, SIZE(ls_scf_env%matrix_p)
            project_name = logger%iter_info%project_name
            WRITE (file_name, '(A,I0,A)') TRIM(project_name)//"_LS_DM_SPIN_", ispin, "_RESTART.dm"
            cs_pos = cp_dbcsr_checksum(ls_scf_env%matrix_p(ispin), pos=.TRUE.)
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T2,A,E20.8)') "Writing restart DM "//TRIM(file_name)//" with checksum: ", cs_pos
            ENDIF
            IF (ls_scf_env%do_transport .OR. ls_scf_env%do_pexsi) THEN
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, '(T6,A)') "The restart DM "//TRIM(file_name)//" has the sparsity of S, therefore,"
                  WRITE (unit_nr, '(T6,A)') "not compatible with methods that require a full DM! "
               END IF
            END IF
            CALL cp_dbcsr_binary_write(ls_scf_env%matrix_p(ispin), file_name)
         ENDDO
      END IF

      IF (ls_scf_env%scf_history%nstore > 0) THEN
         ls_scf_env%scf_history%istore = ls_scf_env%scf_history%istore+1
         DO ispin = 1, SIZE(ls_scf_env%matrix_p)
            istore = MOD(ls_scf_env%scf_history%istore-1, ls_scf_env%scf_history%nstore)+1
            IF (ls_scf_env%scf_history%istore <= ls_scf_env%scf_history%nstore) &
               CALL cp_dbcsr_init(ls_scf_env%scf_history%matrix(ispin, istore))
            CALL cp_dbcsr_copy(ls_scf_env%scf_history%matrix(ispin, istore), ls_scf_env%matrix_p(ispin))

            ! if we have the sqrt around, we use it to go to the orthogonal basis
            IF (ls_scf_env%use_s_sqrt) THEN
               ! usualy sqrt(S) * P * sqrt(S) should be available, or could be stored at least,
               ! so that the next multiplications could be saved.
               CALL cp_dbcsr_init(matrix_tmp1)
               CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, &
                                    matrix_type=dbcsr_type_no_symmetry)

               IF (ls_scf_env%has_s_preconditioner) THEN
                  CALL apply_matrix_preconditioner(ls_scf_env%scf_history%matrix(ispin, istore), "backward", &
                                                   ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
               ENDIF
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt, ls_scf_env%scf_history%matrix(ispin, istore), &
                                      0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt, &
                                      0.0_dp, ls_scf_env%scf_history%matrix(ispin, istore), &
                                      filter_eps=ls_scf_env%eps_filter)
               CALL cp_dbcsr_release(matrix_tmp1)
            ENDIF

         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_store_result

! **************************************************************************************************
!> \brief parse the input section, no need to pass it around
!> \param input ...
!> \param ls_scf_env ...
!> \param unit_nr ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_init_read_write_input(input, ls_scf_env, unit_nr)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(ls_scf_env_type), INTENT(INOUT)               :: ls_scf_env
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_init_read_write_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: mu
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: chebyshev_section, curvy_section, &
                                                            ls_scf_section, mixing_section, &
                                                            pao_section, pexsi_section

      CALL timeset(routineN, handle)
      CALL cite_reference(VandeVondele2012)
      ls_scf_section => section_vals_get_subs_vals(input, "DFT%LS_SCF")
      curvy_section => section_vals_get_subs_vals(ls_scf_section, "CURVY_STEPS")

      ! should come from input
      CALL section_vals_val_get(ls_scf_section, "LS_DIIS", l_val=ls_scf_env%ls_diis)
      CALL section_vals_val_get(ls_scf_section, "INI_DIIS", i_val=ls_scf_env%iter_ini_diis)
      CALL section_vals_val_get(ls_scf_section, "MAX_DIIS", i_val=ls_scf_env%max_diis)
      CALL section_vals_val_get(ls_scf_section, "NMIXING", i_val=ls_scf_env%nmixing)
      CALL section_vals_val_get(ls_scf_section, "EPS_DIIS", r_val=ls_scf_env%eps_diis)
      CALL section_vals_val_get(ls_scf_section, "EPS_SCF", r_val=ls_scf_env%eps_scf)
      CALL section_vals_val_get(ls_scf_section, "EPS_FILTER", r_val=ls_scf_env%eps_filter)
      CALL section_vals_val_get(ls_scf_section, "MU", r_val=mu)
      CALL section_vals_val_get(ls_scf_section, "FIXED_MU", l_val=ls_scf_env%fixed_mu)
      ls_scf_env%mu_spin = mu
      CALL section_vals_val_get(ls_scf_section, "MIXING_FRACTION", r_val=ls_scf_env%mixing_fraction)
      CALL section_vals_val_get(ls_scf_section, "MAX_SCF", i_val=ls_scf_env%max_scf)
      CALL section_vals_val_get(ls_scf_section, "S_PRECONDITIONER", i_val=ls_scf_env%s_preconditioner_type)
      CALL section_vals_val_get(ls_scf_section, "MATRIX_CLUSTER_TYPE", i_val=ls_scf_env%ls_mstruct%cluster_type)
      CALL section_vals_val_get(ls_scf_section, "SINGLE_PRECISION_MATRICES", l_val=ls_scf_env%ls_mstruct%single_precision)
      CALL section_vals_val_get(ls_scf_section, "S_INVERSION", i_val=ls_scf_env%s_inversion_type)
      CALL section_vals_val_get(ls_scf_section, "CHECK_S_INV", l_val=ls_scf_env%check_s_inv)
      CALL section_vals_val_get(ls_scf_section, "REPORT_ALL_SPARSITIES", l_val=ls_scf_env%report_all_sparsities)
      CALL section_vals_val_get(ls_scf_section, "PERFORM_MU_SCAN", l_val=ls_scf_env%perform_mu_scan)
      CALL section_vals_val_get(ls_scf_section, "PURIFICATION_METHOD", i_val=ls_scf_env%purification_method)
      CALL section_vals_val_get(ls_scf_section, "DYNAMIC_THRESHOLD", l_val=ls_scf_env%dynamic_threshold)
      CALL section_vals_val_get(ls_scf_section, "NON_MONOTONIC", l_val=ls_scf_env%non_monotonic)
      CALL section_vals_val_get(ls_scf_section, "SIGN_SQRT_ORDER", i_val=ls_scf_env%sign_sqrt_order)
      CALL section_vals_val_get(ls_scf_section, "EXTRAPOLATION_ORDER", i_val=ls_scf_env%extrapolation_order)
      CALL section_vals_val_get(ls_scf_section, "RESTART_READ", l_val=ls_scf_env%restart_read)
      CALL section_vals_val_get(ls_scf_section, "RESTART_WRITE", l_val=ls_scf_env%restart_write)
      CALL section_vals_val_get(ls_scf_section, "EPS_LANCZOS", r_val=ls_scf_env%eps_lanczos)
      CALL section_vals_val_get(ls_scf_section, "MAX_ITER_LANCZOS", i_val=ls_scf_env%max_iter_lanczos)

      CALL section_vals_get(curvy_section, explicit=ls_scf_env%curvy_steps)
      CALL section_vals_val_get(curvy_section, "LINE_SEARCH", i_val=ls_scf_env%curvy_data%line_search_type)
      CALL section_vals_val_get(curvy_section, "N_BCH_HISTORY", i_val=ls_scf_env%curvy_data%n_bch_hist)
      CALL section_vals_val_get(curvy_section, "MIN_HESSIAN_SHIFT", r_val=ls_scf_env%curvy_data%min_shift)
      CALL section_vals_val_get(curvy_section, "FILTER_FACTOR", r_val=ls_scf_env%curvy_data%filter_factor)
      CALL section_vals_val_get(curvy_section, "FILTER_FACTOR_SCALE", r_val=ls_scf_env%curvy_data%scale_filter)
      CALL section_vals_val_get(curvy_section, "MIN_FILTER", r_val=ls_scf_env%curvy_data%min_filter)

      ls_scf_env%extrapolation_order = MAX(0, ls_scf_env%extrapolation_order)

      chebyshev_section => section_vals_get_subs_vals(input, "DFT%LS_SCF%CHEBYSHEV")
      CALL section_vals_get(chebyshev_section, explicit=ls_scf_env%chebyshev%compute_chebyshev)
      IF (ls_scf_env%chebyshev%compute_chebyshev) THEN
         CALL section_vals_val_get(chebyshev_section, "N_CHEBYSHEV", i_val=ls_scf_env%chebyshev%n_chebyshev)
         CALL section_vals_val_get(chebyshev_section, "DOS%N_GRIDPOINTS", i_val=ls_scf_env%chebyshev%n_gridpoint_dos)

         ls_scf_env%chebyshev%print_key_dos => &
            section_vals_get_subs_vals(chebyshev_section, "DOS")
         CALL section_vals_retain(ls_scf_env%chebyshev%print_key_dos)

         ls_scf_env%chebyshev%print_key_cube => &
            section_vals_get_subs_vals(chebyshev_section, "PRINT_SPECIFIC_E_DENSITY_CUBE")
         CALL section_vals_retain(ls_scf_env%chebyshev%print_key_cube)
      ENDIF

      mixing_section => section_vals_get_subs_vals(input, "DFT%LS_SCF%RHO_MIXING")
      CALL section_vals_get(mixing_section, explicit=ls_scf_env%do_rho_mixing)

      CALL section_vals_val_get(mixing_section, "METHOD", i_val=ls_scf_env%density_mixing_method)
      IF (ls_scf_env%ls_diis .AND. ls_scf_env%do_rho_mixing) &
         CALL cp__b("dm_ls_scf.F",625,"LS_DIIS and RHO_MIXING are not compatible.")

      pexsi_section => section_vals_get_subs_vals(input, "DFT%LS_SCF%PEXSI")
      CALL section_vals_get(pexsi_section)

      ls_scf_env%do_pexsi = ls_scf_env%purification_method .EQ. ls_scf_pexsi &
                            .AND. .NOT. ls_scf_env%do_transport
      IF (ls_scf_env%do_pexsi) THEN
         CALL pexsi_init_read_input(pexsi_section, ls_scf_env%pexsi)
         ! Turn off S inversion (not used for PEXSI).
         ! Methods such as purification must thus be avoided... which is OK, as the density matrix computed in pexsi is
         ! a finite temperature one, and thus not idempotent
         ls_scf_env%s_inversion_type = ls_s_inversion_none
         ! PEXSI needs the sparsity pattern of S to be fixed by the upper bound ~ Int[|phi_a||phi_b|],
         ! they can not be filtered based on magnitude, as small elements in S (e.g. symmetry) do not necessarily
         ! correspond to small elements in the density matrix, with non-zero contributions to the total density.
         ! the easiest way to make sure S is untouched is to set eps_filter to zero (which should be inconsequential,
         ! as a run based on pexsi should execute exactly zero multiplications)
         ls_scf_env%eps_filter = 0.0_dp
      ENDIF

      ! Turn off S inversion and set eps_filter to zero for transport
      IF (ls_scf_env%do_transport) THEN
         ls_scf_env%s_inversion_type = ls_s_inversion_none
         ls_scf_env%eps_filter = 0.0_dp
      END IF

      SELECT CASE (ls_scf_env%s_inversion_type)
      CASE (ls_s_inversion_sign_sqrt)
         ls_scf_env%needs_s_inv = .TRUE.
         ls_scf_env%use_s_sqrt = .TRUE.
      CASE (ls_s_inversion_hotelling)
         ls_scf_env%needs_s_inv = .TRUE.
         ls_scf_env%use_s_sqrt = .FALSE.
      CASE (ls_s_inversion_none)
         ls_scf_env%needs_s_inv = .FALSE.
         ls_scf_env%use_s_sqrt = .FALSE.
      CASE DEFAULT
         CALL cp__b("dm_ls_scf.F",663,"")
      END SELECT

      SELECT CASE (ls_scf_env%s_preconditioner_type)
      CASE (ls_s_preconditioner_none)
         ls_scf_env%has_s_preconditioner = .FALSE.
      CASE DEFAULT
         ls_scf_env%has_s_preconditioner = .TRUE.
      END SELECT

      ! verify some requirements for the curvy steps
      IF (ls_scf_env%curvy_steps .AND. ls_scf_env%do_pexsi) &
         CALL cp__b("dm_ls_scf.F",675,"CURVY_STEPS cannot be used together with PEXSI.")
      IF (ls_scf_env%curvy_steps .AND. ls_scf_env%do_transport) &
         CALL cp__b("dm_ls_scf.F",677,"CURVY_STEPS cannot be used together with TRANSPORT.")
      IF (ls_scf_env%curvy_steps .AND. ls_scf_env%has_s_preconditioner) &
         CALL cp__b("dm_ls_scf.F",679,"S Preconditioning not implemented in combination with CURVY_STEPS.")
      IF (ls_scf_env%curvy_steps .AND. .NOT. ls_scf_env%use_s_sqrt) &
         CALL cp__b("dm_ls_scf.F",681,"CURVY_STEPS requires the use of the sqrt inversion.")

      ! an undocumented feature ... allows for just doing the initial guess, no expensive stuff
      IF (ls_scf_env%max_scf < 0) THEN
         ls_scf_env%needs_s_inv = .FALSE.
         ls_scf_env%use_s_sqrt = .FALSE.
         ls_scf_env%has_s_preconditioner = .FALSE.
      ENDIF

      pao_section => section_vals_get_subs_vals(input, "DFT%LS_SCF%PAO")
      CALL section_vals_get(pao_section, explicit=ls_scf_env%do_pao)
      IF (ls_scf_env%do_pao) THEN
         ls_scf_env%ls_mstruct%cluster_type = ls_cluster_pao
      ENDIF

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '()')
         WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 30), " Linear scaling SCF ", REPEAT("-", 29)
         WRITE (unit_nr, '(T2,A,T38,E20.3)') "eps_scf:", ls_scf_env%eps_scf
         WRITE (unit_nr, '(T2,A,T38,E20.3)') "eps_filter:", ls_scf_env%eps_filter
         IF (ls_scf_env%do_rho_mixing) THEN
            IF (ls_scf_env%density_mixing_method > 0) THEN
               NULLIFY (section)
               CALL create_mixing_section(section, ls_scf=.TRUE.)
               keyword => section_get_keyword(section, "METHOD")
               CALL keyword_get(keyword, enum=enum)
               WRITE (unit_nr, "(T2,A,T38,A20)") &
                  "Density mixing in g-space:", ADJUSTR(TRIM(enum_i2c(enum, &
                                                                      ls_scf_env%density_mixing_method)))
               CALL section_release(section)
            END IF
         ELSE
            WRITE (unit_nr, '(T2,A,T38,E20.3)') "mixing_fraction:", ls_scf_env%mixing_fraction
         ENDIF
         WRITE (unit_nr, '(T2,A,T38,I20)') "max_scf:", ls_scf_env%max_scf
         IF (ls_scf_env%ls_diis) THEN
            WRITE (unit_nr, '(T2,A,T38,I20)') "DIIS: max_diis:", ls_scf_env%max_diis
            WRITE (unit_nr, '(T2,A,T38,E20.3)') "DIIS: eps_diis:", ls_scf_env%eps_diis
            WRITE (unit_nr, '(T2,A,T38,I20)') "DIIS: ini_diis:", ls_scf_env%iter_ini_diis
            WRITE (unit_nr, '(T2,A,T38,I20)') "DIIS: nmixing:", ls_scf_env%nmixing
         ENDIF
         WRITE (unit_nr, '(T2,A,T38,L20)') "fixed chemical potential (mu)", ls_scf_env%fixed_mu
         WRITE (unit_nr, '(T2,A,T38,L20)') "has unit metric:", ls_scf_env%has_unit_metric
         WRITE (unit_nr, '(T2,A,T38,L20)') "Computing inv(S):", ls_scf_env%needs_s_inv
         WRITE (unit_nr, '(T2,A,T38,L20)') "Computing sqrt(S):", ls_scf_env%use_s_sqrt
         WRITE (unit_nr, '(T2,A,T38,L20)') "Computing s preconditioner ", ls_scf_env%has_s_preconditioner
         WRITE (unit_nr, '(T2,A,T38,I20)') "sign sqrt order:", ls_scf_env%sign_sqrt_order
         WRITE (unit_nr, '(T2,A,T38,I20)') "Extrapolation order:", ls_scf_env%extrapolation_order
         WRITE (unit_nr, '(T2,A,T38,L20)') "Use single precision matrices", ls_scf_env%ls_mstruct%single_precision

         SELECT CASE (ls_scf_env%s_preconditioner_type)
         CASE (ls_s_preconditioner_none)
            WRITE (unit_nr, '(T2,A,T38,A20)') "S preconditioner type ", "NONE"
         CASE (ls_s_preconditioner_atomic)
            WRITE (unit_nr, '(T2,A,T38,A20)') "S preconditioner type ", "ATOMIC"
         CASE (ls_s_preconditioner_molecular)
            WRITE (unit_nr, '(T2,A,T38,A20)') "S preconditioner type ", "MOLECULAR"
         END SELECT

         IF (ls_scf_env%curvy_steps) THEN
            WRITE (unit_nr, '(T2,A,T38,A30)') "Using curvy steps to optimize the density matrix"
            CALL cite_reference(Shao2003)
         ENDIF

         SELECT CASE (ls_scf_env%purification_method)
         CASE (ls_scf_ns)
            WRITE (unit_nr, '(T2,A,T38,A30)') "Purification method", "Newton-Schulz sign iter"
         CASE (ls_scf_tc2)
            CALL cite_reference(Niklasson2014)
            WRITE (unit_nr, '(T2,A,T38,A30)') "Purification method", "Trace conserving 2nd  order"
         CASE (ls_scf_trs4)
            CALL cite_reference(Niklasson2003)
            WRITE (unit_nr, '(T2,A,T38,A30)') "Purification method", "Trace resetting 4th order"
         CASE (ls_scf_pexsi)
            CALL cite_reference(Lin2009)
            CALL cite_reference(Lin2013)
            WRITE (unit_nr, '(T2,A,T38,A20)') "Purification method", "PEXSI"
         CASE DEFAULT
            CALL cp__b("dm_ls_scf.F",759,"")
         END SELECT

         SELECT CASE (ls_scf_env%ls_mstruct%cluster_type)
         CASE (ls_cluster_atomic)
            WRITE (unit_nr, '(T2,A,T38,A20)') "Cluster type", "ATOMIC"
         CASE (ls_cluster_molecular)
            WRITE (unit_nr, '(T2,A,T38,A20)') "Cluster type", "MOLECULAR"
         CASE (ls_cluster_pao)
            WRITE (unit_nr, '(T2,A,T38,A20)') "Cluster type", "PAO"
         CASE DEFAULT
            CALL cp__b("dm_ls_scf.F",770,"Unkown cluster type")
         END SELECT

         IF (ls_scf_env%chebyshev%compute_chebyshev) THEN
            WRITE (unit_nr, '(T2,A,T38,A20)') "Computing Chebyshev", "TRUE"
            WRITE (unit_nr, '(T2,A,T38,I20)') "N_CHEBYSHEV:", ls_scf_env%chebyshev%n_chebyshev
            WRITE (unit_nr, '(T2,A,T38,I20)') "N_GRIDPOINT_DOS:", ls_scf_env%chebyshev%n_gridpoint_dos
         ELSE
            WRITE (unit_nr, '(T2,A,T38,A20)') "Computing Chebyshev", "FALSE"
         ENDIF

         WRITE (unit_nr, '(T2,A,T38,L20)') "Using PAO", ls_scf_env%do_pao

         WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)
         WRITE (unit_nr, '()')
         CALL m_flush(unit_nr)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ls_scf_init_read_write_input

! **************************************************************************************************
!> \brief Main SCF routine. Can we keep it clean ?
!> \param qs_env ...
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_main(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_main', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iscf, ispin, &
                                                            nelectron_spin_real, nmixing, nspin, &
                                                            unit_nr
      LOGICAL :: check_convergence, diis_step, do_transport, extra_scf, maxscf_reached, &
         scf_converged, should_stop, transm_maxscf_reached, transm_scf_converged
      REAL(KIND=dp)                                      :: energy_diff, energy_new, energy_old, &
                                                            eps_diis, t1, t2
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: matrix_ks_deviation, matrix_mixing_old
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer
      TYPE(transport_env_type), POINTER                  :: transport_env

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = ls_scf_env%nspins

      ! old quantities, useful for mixing
      ALLOCATE (matrix_mixing_old(nspin), matrix_ks_deviation(nspin))
      DO ispin = 1, nspin
         CALL cp_dbcsr_init(matrix_mixing_old(ispin))
         CALL cp_dbcsr_create(matrix_mixing_old(ispin), template=ls_scf_env%matrix_ks(ispin))

         CALL cp_dbcsr_init(matrix_ks_deviation(ispin))
         CALL cp_dbcsr_create(matrix_ks_deviation(ispin), template=ls_scf_env%matrix_ks(ispin))
         CALL cp_dbcsr_set(matrix_ks_deviation(ispin), 0.0_dp)
      ENDDO
      ls_scf_env%homo_spin(:) = 0.0_dp
      ls_scf_env%lumo_spin(:) = 0.0_dp

      transm_scf_converged = .FALSE.
      transm_maxscf_reached = .FALSE.

      energy_old = 0.0_dp
      IF (ls_scf_env%scf_history%istore > 0) energy_old = ls_scf_env%energy_init
      check_convergence = .TRUE.
      iscf = 0
      IF (ls_scf_env%ls_diis) THEN
         diis_step = .FALSE.
         eps_diis = ls_scf_env%eps_diis
         nmixing = ls_scf_env%nmixing
         NULLIFY (diis_buffer)
         IF (.NOT. ASSOCIATED(diis_buffer)) THEN
            CALL qs_diis_b_create_sparse(diis_buffer, &
                                         nbuffer=ls_scf_env%max_diis)
         END IF
         CALL qs_diis_b_clear_sparse(diis_buffer)
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
      END IF

      CALL get_qs_env(qs_env, transport_env=transport_env, do_transport=do_transport)

      ! the real SCF loop
      DO

         ! check on max SCF or timing/exit
         CALL external_control(should_stop, "SCF", start_time=qs_env%start_time, target_time=qs_env%target_time)
         IF (do_transport) THEN
            maxscf_reached = should_stop .OR. iscf >= ls_scf_env%max_scf
            ! one extra scf step for post-processing in transmission calculations
            IF (transport_env%params%method .EQ. transport_transmission) THEN
               IF (transm_maxscf_reached) THEN
                  IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') "SCF not converged! "
                  EXIT
               END IF
               transm_maxscf_reached = maxscf_reached
            ELSE
               IF (maxscf_reached) THEN
                  IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') "SCF not converged! "
                  EXIT
               ENDIF
            END IF
         ELSE
            IF (should_stop .OR. iscf >= ls_scf_env%max_scf) THEN
               IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') "SCF not converged! "
               EXIT
            ENDIF
         END IF

         t1 = m_walltime()
         iscf = iscf+1

         ! first get a copy of the current KS matrix
         CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
         DO ispin = 1, nspin
            CALL matrix_qs_to_ls(ls_scf_env%matrix_ks(ispin), matrix_ks(ispin)%matrix, &
                                 ls_scf_env%ls_mstruct, covariant=.TRUE.)
            IF (ls_scf_env%has_s_preconditioner) THEN
               CALL apply_matrix_preconditioner(ls_scf_env%matrix_ks(ispin), "forward", &
                                                ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
            ENDIF
            CALL cp_dbcsr_filter(ls_scf_env%matrix_ks(ispin), ls_scf_env%eps_filter)
         ENDDO
         ! run curvy steps if required. Needs an idempotent DM (either perification or restart)
         IF ((iscf > 1 .OR. ls_scf_env%scf_history%istore > 0) .AND. ls_scf_env%curvy_steps) THEN
            CALL dm_ls_curvy_optimization(ls_scf_env, energy_old, check_convergence)
         ELSE
            ! turn the KS matrix in a density matrix
            DO ispin = 1, nspin
               IF (ls_scf_env%do_rho_mixing) THEN
                  CALL cp_dbcsr_copy(matrix_mixing_old(ispin), ls_scf_env%matrix_ks(ispin))
               ELSE
                  IF (iscf == 1) THEN
                     ! initialize the mixing matrix with the current state if needed
                     CALL cp_dbcsr_copy(matrix_mixing_old(ispin), ls_scf_env%matrix_ks(ispin))
                  ELSE
                     IF (ls_scf_env%ls_diis) THEN ! ------- IF-DIIS+MIX--- START
                        IF (diis_step .AND. (iscf-1) .GE. ls_scf_env%iter_ini_diis) THEN
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, '(A61)') &
                                 '*************************************************************'
                              WRITE (unit_nr, '(A50,2(I3,A1),L1,A1)') &
                                 " Using DIIS mixed KS:  (iscf,INI_DIIS,DIIS_STEP)=(", &
                                 iscf, ",", ls_scf_env%iter_ini_diis, ",", diis_step, ")"
                              WRITE (unit_nr, '(A52)') &
                                 " KS_nw= DIIS-Linear-Combination-Previous KS matrices"
                              WRITE (unit_nr, '(61A)') &
                                 "*************************************************************"
                           ENDIF
                           CALL cp_dbcsr_copy(matrix_mixing_old(ispin), & ! out
                                              ls_scf_env%matrix_ks(ispin)) ! in
                        ELSE
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, '(A57)') &
                                 "*********************************************************"
                              WRITE (unit_nr, '(A23,F5.3,A25,I3)') &
                                 " Using MIXING_FRACTION=", ls_scf_env%mixing_fraction, &
                                 " to mix KS matrix:  iscf=", iscf
                              WRITE (unit_nr, '(A7,F5.3,A6,F5.3,A7)') &
                                 " KS_nw=", ls_scf_env%mixing_fraction, "*KS + ", &
                                 1.0_dp-ls_scf_env%mixing_fraction, "*KS_old"
                              WRITE (unit_nr, '(A57)') &
                                 "*********************************************************"
                           ENDIF
                           ! perform the mixing of ks matrices
                           CALL cp_dbcsr_add(matrix_mixing_old(ispin), &
                                             ls_scf_env%matrix_ks(ispin), &
                                             1.0_dp-ls_scf_env%mixing_fraction, &
                                             ls_scf_env%mixing_fraction)
                        ENDIF
                     ELSE ! otherwise
                        IF (unit_nr > 0) THEN
                           WRITE (unit_nr, '(A57)') &
                              "*********************************************************"
                           WRITE (unit_nr, '(A23,F5.3,A25,I3)') &
                              " Using MIXING_FRACTION=", ls_scf_env%mixing_fraction, &
                              " to mix KS matrix:  iscf=", iscf
                           WRITE (unit_nr, '(A7,F5.3,A6,F5.3,A7)') &
                              " KS_nw=", ls_scf_env%mixing_fraction, "*KS + ", &
                              1.0_dp-ls_scf_env%mixing_fraction, "*KS_old"
                           WRITE (unit_nr, '(A57)') &
                              "*********************************************************"
                        ENDIF
                        ! perform the mixing of ks matrices
                        CALL cp_dbcsr_add(matrix_mixing_old(ispin), &
                                          ls_scf_env%matrix_ks(ispin), &
                                          1.0_dp-ls_scf_env%mixing_fraction, &
                                          ls_scf_env%mixing_fraction)
                     ENDIF ! ------- IF-DIIS+MIX--- END
                  ENDIF
               ENDIF

               ! compute the density matrix that matches it
               ! we need the proper number of states
               nelectron_spin_real = ls_scf_env%nelectron_spin(ispin)
               IF (ls_scf_env%nspins == 1) nelectron_spin_real = nelectron_spin_real/2

               IF (do_transport) THEN
                  IF (ls_scf_env%has_s_preconditioner) &
                     CALL cp__b("dm_ls_scf.F",983,"NOT YET IMPLEMENTED with S preconditioner. ")
                  IF (ls_scf_env%ls_mstruct%cluster_type .NE. ls_cluster_atomic) &
                     CALL cp__b("dm_ls_scf.F",985,"NOT YET IMPLEMENTED with molecular clustering. ")

                  extra_scf = maxscf_reached .OR. scf_converged
                  ! get the current Kohn-Sham matrix (ks) and return matrix_p evaluated using an external C routine
                  CALL external_scf_method(transport_env, ls_scf_env%matrix_s, matrix_mixing_old(ispin), &
                                           ls_scf_env%matrix_p(ispin), nelectron_spin_real, ls_scf_env%natoms, &
                                           energy_diff, extra_scf)

               ELSE
                  SELECT CASE (ls_scf_env%purification_method)
                  CASE (ls_scf_ns)
                     CALL density_matrix_sign(ls_scf_env%matrix_p(ispin), ls_scf_env%mu_spin(ispin), ls_scf_env%fixed_mu, &
                                              matrix_mixing_old(ispin), ls_scf_env%matrix_s, ls_scf_env%matrix_s_inv, &
                                              nelectron_spin_real, ls_scf_env%eps_filter)
                  CASE (ls_scf_tc2)
                     CALL density_matrix_tc2(ls_scf_env%matrix_p(ispin), matrix_mixing_old(ispin), ls_scf_env%matrix_s_sqrt_inv, &
                                             nelectron_spin_real, ls_scf_env%eps_filter, ls_scf_env%homo_spin(ispin), &
                                             ls_scf_env%lumo_spin(ispin), non_monotonic=ls_scf_env%non_monotonic, &
                                             eps_lanczos=ls_scf_env%eps_lanczos, max_iter_lanczos=ls_scf_env%max_iter_lanczos)
                  CASE (ls_scf_trs4)
                     CALL density_matrix_trs4(ls_scf_env%matrix_p(ispin), matrix_mixing_old(ispin), ls_scf_env%matrix_s_sqrt_inv, &
                                              nelectron_spin_real, ls_scf_env%eps_filter, ls_scf_env%homo_spin(ispin), &
                                              ls_scf_env%lumo_spin(ispin), ls_scf_env%mu_spin(ispin), &
                                              dynamic_threshold=ls_scf_env%dynamic_threshold, &
                                              matrix_ks_deviation=matrix_ks_deviation(ispin), &
                                              eps_lanczos=ls_scf_env%eps_lanczos, max_iter_lanczos=ls_scf_env%max_iter_lanczos)
                  CASE (ls_scf_pexsi)
                     IF (ls_scf_env%has_s_preconditioner) &
                        CALL cp__b("dm_ls_scf.F",1013,"S preconditioning not implemented in combination with the PEXSI library. ")
                     IF (ls_scf_env%ls_mstruct%cluster_type .NE. ls_cluster_atomic) &
                        CALL cp_abort(cp__l("dm_ls_scf.F",1015), &
                                      "Molecular clustering not implemented in combination with the PEXSI library. ")
                     IF (ls_scf_env%ls_mstruct%single_precision) &
                        CALL cp__b("dm_ls_scf.F",1018,"PEXSI library requires double precision datatype. ")
                     CALL density_matrix_pexsi(ls_scf_env%pexsi, ls_scf_env%matrix_p(ispin), ls_scf_env%pexsi%matrix_w(ispin), &
                                               ls_scf_env%pexsi%kTS(ispin), matrix_mixing_old(ispin), ls_scf_env%matrix_s, &
                                               nelectron_spin_real, ls_scf_env%mu_spin(ispin), iscf, ispin)
                  END SELECT
               END IF

               IF (ls_scf_env%has_s_preconditioner) THEN
                  CALL apply_matrix_preconditioner(ls_scf_env%matrix_p(ispin), "forward", &
                                                   ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
               ENDIF
               CALL cp_dbcsr_filter(ls_scf_env%matrix_p(ispin), ls_scf_env%eps_filter)

               IF (ls_scf_env%nspins == 1) CALL cp_dbcsr_scale(ls_scf_env%matrix_p(ispin), 2.0_dp)

            ENDDO
         END IF

         ! compute the corresponding new energy KS matrix and new energy
         CALL ls_scf_dm_to_ks(qs_env, ls_scf_env, energy_new, iscf)

         IF (ls_scf_env%do_pexsi) THEN
            CALL pexsi_to_qs(ls_scf_env, qs_env, kTS=ls_scf_env%pexsi%kTS)
         ENDIF

         ! report current SCF loop
         energy_diff = energy_new-energy_old
         energy_old = energy_new

         t2 = m_walltime()
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *)
            WRITE (unit_nr, '(T2,A,I6,F20.9,F20.9,F12.6)') "SCF", iscf, energy_new, energy_diff, t2-t1
            WRITE (unit_nr, *)
            CALL m_flush(unit_nr)
         ENDIF

         IF (do_transport) THEN
            scf_converged = check_convergence .AND. ABS(energy_diff) < ls_scf_env%eps_scf*ls_scf_env%nelectron_total
            ! one extra scf step for post-processing in transmission calculations
            IF (transport_env%params%method .EQ. transport_transmission) THEN
               IF (transm_scf_converged) EXIT
               transm_scf_converged = scf_converged
            ELSE
               IF (scf_converged) EXIT
            END IF
         ELSE
            ! exit criterium on the energy only for the time being
            IF (check_convergence .AND. ABS(energy_diff) < ls_scf_env%eps_scf*ls_scf_env%nelectron_total) EXIT
         END IF

         IF (ls_scf_env%ls_diis) THEN
! diis_buffer, buffer with 1) Kohn-Sham history matrix,
!                          2) KS error history matrix (f=KPS-SPK),
!                          3) B matrix (for finding DIIS weighting coefficients)
            CALL qs_diis_b_step_4lscf(diis_buffer, qs_env, ls_scf_env, unit_nr, &
                                      iscf, diis_step, eps_diis, nmixing, matrix_s(1)%matrix, &
                                      ls_scf_env%eps_filter)
         ENDIF

         IF (ls_scf_env%do_pexsi) THEN
            CALL pexsi_set_convergence_tolerance(ls_scf_env%pexsi, energy_diff, &
                                                 ls_scf_env%eps_scf*ls_scf_env%nelectron_total, &
                                                 ! initialize in second scf step of first SCF cycle:
                                                 (iscf .EQ. 2) .AND. (ls_scf_env%scf_history%istore .EQ. 0), &
                                                 check_convergence)
         ENDIF

      ENDDO

      ! free storage
      IF (ls_scf_env%ls_diis) THEN
         CALL qs_diis_b_release_sparse(diis_buffer)
      ENDIF
      DO ispin = 1, nspin
         CALL cp_dbcsr_release(matrix_mixing_old(ispin))
         CALL cp_dbcsr_release(matrix_ks_deviation(ispin))
      ENDDO
      DEALLOCATE (matrix_mixing_old, matrix_ks_deviation)

      CALL timestop(handle)

   END SUBROUTINE ls_scf_main

! **************************************************************************************************
!> \brief after SCF we have a density matrix, and the self consistent KS matrix
!>        analyze its properties.
!> \param qs_env ...
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_post(qs_env, ls_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_post', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_w
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! store the matrix for a next scf run
      IF (.NOT. ls_scf_env%do_pao) &
         CALL ls_scf_store_result(ls_scf_env)

      ! write homo and lumo energy (if not already part of the output)
      IF (ls_scf_env%curvy_steps) THEN
         CALL post_scf_homo_lumo(ls_scf_env)
      ENDIF

      ! compute the matrix_w if associated
      CALL get_qs_env(qs_env, matrix_w=matrix_w)
      IF (ASSOCIATED(matrix_w)) THEN
         IF (ls_scf_env%do_pexsi) THEN
            CALL pexsi_to_qs(ls_scf_env, qs_env, matrix_w=ls_scf_env%pexsi%matrix_w)
         ELSE
            CALL calculate_w_matrix(matrix_w, ls_scf_env)
         ENDIF
      ENDIF

      ! compute properties

      IF (ls_scf_env%perform_mu_scan) CALL post_scf_mu_scan(ls_scf_env)

      IF (ls_scf_env%report_all_sparsities) CALL post_scf_sparsities(ls_scf_env)

      CALL write_mo_free_results(qs_env)

      IF (ls_scf_env%chebyshev%compute_chebyshev) CALL compute_chebyshev(qs_env, ls_scf_env)

      IF (.TRUE.) CALL post_scf_experiment()

      CALL qs_scf_post_moments(qs_env%input, logger, qs_env, unit_nr)

      ! clean up used data

      CALL cp_dbcsr_release(ls_scf_env%matrix_s)
      CALL deallocate_curvy_data(ls_scf_env%curvy_data)

      IF (ls_scf_env%has_s_preconditioner) THEN
         CALL cp_dbcsr_release(ls_scf_env%matrix_bs_sqrt)
         CALL cp_dbcsr_release(ls_scf_env%matrix_bs_sqrt_inv)
      ENDIF

      IF (ls_scf_env%needs_s_inv) THEN
         CALL cp_dbcsr_release(ls_scf_env%matrix_s_inv)
      ENDIF

      IF (ls_scf_env%use_s_sqrt) THEN
         CALL cp_dbcsr_release(ls_scf_env%matrix_s_sqrt)
         CALL cp_dbcsr_release(ls_scf_env%matrix_s_sqrt_inv)
      ENDIF

      DO ispin = 1, SIZE(ls_scf_env%matrix_p)
         CALL cp_dbcsr_release(ls_scf_env%matrix_p(ispin))
      ENDDO
      DEALLOCATE (ls_scf_env%matrix_p)

      DO ispin = 1, SIZE(ls_scf_env%matrix_ks)
         CALL cp_dbcsr_release(ls_scf_env%matrix_ks(ispin))
      ENDDO
      DEALLOCATE (ls_scf_env%matrix_ks)

      IF (ls_scf_env%do_pexsi) &
         CALL pexsi_finalize_scf(ls_scf_env%pexsi, ls_scf_env%mu_spin)

      CALL timestop(handle)

   END SUBROUTINE ls_scf_post

! **************************************************************************************************
!> \brief Compute the HOMO LUMO energies post SCF
!> \param ls_scf_env ...
!> \par History
!>       2013.06 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE post_scf_homo_lumo(ls_scf_env)
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'post_scf_homo_lumo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      LOGICAL                                            :: converged
      REAL(KIND=dp)                                      :: eps_max, eps_min, homo, lumo
      TYPE(cp_dbcsr_type)                                :: matrix_k, matrix_p
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') ""

      ! TODO: remove these limitations
      IF(.NOT.(.NOT. ls_scf_env%has_s_preconditioner))CALL cp__a("dm_ls_scf.F",1232)
      IF(.NOT.(ls_scf_env%use_s_sqrt))CALL cp__a("dm_ls_scf.F",1233)

      nspin = ls_scf_env%nspins

      CALL cp_dbcsr_init(matrix_p)
      CALL cp_dbcsr_create(matrix_p, template=ls_scf_env%matrix_p(1), matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init(matrix_k)
      CALL cp_dbcsr_create(matrix_k, template=ls_scf_env%matrix_p(1), matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, nspin
         ! ortho basis ks
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_ks(ispin), &
                                0.0_dp, matrix_k, filter_eps=ls_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_k, ls_scf_env%matrix_s_sqrt_inv, &
                                0.0_dp, matrix_k, filter_eps=ls_scf_env%eps_filter)

         ! extremal eigenvalues ks
         CALL cp_dbcsr_arnoldi_extremal(matrix_k, eps_max, eps_min, max_iter=ls_scf_env%max_iter_lanczos, &
                                        threshold=ls_scf_env%eps_lanczos, converged=converged)

         ! ortho basis p
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt, ls_scf_env%matrix_p(ispin), &
                                0.0_dp, matrix_p, filter_eps=ls_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p, ls_scf_env%matrix_s_sqrt, &
                                0.0_dp, matrix_p, filter_eps=ls_scf_env%eps_filter)
         IF (nspin == 1) CALL cp_dbcsr_scale(matrix_p, 0.5_dp)

         ! go compute homo lumo
         CALL compute_homo_lumo(matrix_k, matrix_p, eps_min, eps_max, ls_scf_env%eps_filter, &
                                ls_scf_env%max_iter_lanczos, ls_scf_env%eps_lanczos, homo, lumo, unit_nr)

      ENDDO

      CALL cp_dbcsr_release(matrix_p)
      CALL cp_dbcsr_release(matrix_k)

      CALL timestop(handle)

   END SUBROUTINE post_scf_homo_lumo

! **************************************************************************************************
!> \brief Compute the density matrix for various values of the chemical potential
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE post_scf_mu_scan(ls_scf_env)
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'post_scf_mu_scan', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, imu, ispin, nelectron_spin_real, &
                                                            nmu, nspin, unit_nr
      REAL(KIND=dp)                                      :: mu, t1, t2, trace
      TYPE(cp_dbcsr_type)                                :: matrix_p
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = ls_scf_env%nspins

      CALL cp_dbcsr_init(matrix_p)
      CALL cp_dbcsr_create(matrix_p, template=ls_scf_env%matrix_p(1))

      nmu = 10
      DO imu = 0, nmu

         t1 = m_walltime()

         mu = -0.4_dp+imu*(0.1_dp+0.4_dp)/nmu

         IF (unit_nr > 0) WRITE (unit_nr, *) "------- starting with mu ", mu

         DO ispin = 1, nspin
            ! we need the proper number of states
            nelectron_spin_real = ls_scf_env%nelectron_spin(ispin)
            IF (ls_scf_env%nspins == 1) nelectron_spin_real = nelectron_spin_real/2

            CALL density_matrix_sign_fixed_mu(matrix_p, trace, mu, &
                                              ls_scf_env%matrix_ks(ispin), ls_scf_env%matrix_s, &
                                              ls_scf_env%matrix_s_inv, ls_scf_env%eps_filter)
         ENDDO

         t2 = m_walltime()

         IF (unit_nr > 0) WRITE (unit_nr, *) " obtained ", mu, trace, t2-t1

      ENDDO

      CALL cp_dbcsr_release(matrix_p)

      CALL timestop(handle)

   END SUBROUTINE post_scf_mu_scan

! **************************************************************************************************
!> \brief Report on the sparsities of various interesting matrices.
!>
!> \param ls_scf_env ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE post_scf_sparsities(ls_scf_env)
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'post_scf_sparsities', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: title
      INTEGER                                            :: handle, ispin, nspin, unit_nr
      TYPE(cp_dbcsr_type)                                :: matrix_tmp1, matrix_tmp2
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = ls_scf_env%nspins

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '()')
         WRITE (unit_nr, '(T2,A,E17.3)') "Sparsity reports for eps_filter: ", ls_scf_env%eps_filter
         WRITE (unit_nr, '()')
      ENDIF

      CALL report_matrix_sparsity(ls_scf_env%matrix_s, unit_nr, "overlap matrix (S)", &
                                  ls_scf_env%eps_filter)

      DO ispin = 1, nspin
         WRITE (title, '(A,I3)') "Kohn-Sham matrix (H) for spin ", ispin
         CALL report_matrix_sparsity(ls_scf_env%matrix_ks(ispin), unit_nr, title, &
                                     ls_scf_env%eps_filter)
      ENDDO

      CALL cp_dbcsr_init(matrix_tmp1)
      CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp2)
      CALL cp_dbcsr_create(matrix_tmp2, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, nspin
         WRITE (title, '(A,I3)') "Density matrix (P) for spin ", ispin
         CALL report_matrix_sparsity(ls_scf_env%matrix_p(ispin), unit_nr, title, &
                                     ls_scf_env%eps_filter)

         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s, ls_scf_env%matrix_p(ispin), &
                                0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)

         WRITE (title, '(A,I3,A)') "S * P(", ispin, ")"
         CALL report_matrix_sparsity(matrix_tmp1, unit_nr, title, ls_scf_env%eps_filter)

         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s, &
                                0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)
         WRITE (title, '(A,I3,A)') "S * P(", ispin, ") * S"
         CALL report_matrix_sparsity(matrix_tmp2, unit_nr, title, ls_scf_env%eps_filter)
      ENDDO

      IF (ls_scf_env%needs_s_inv) THEN
         CALL report_matrix_sparsity(ls_scf_env%matrix_s_inv, unit_nr, "inv(S)", &
                                     ls_scf_env%eps_filter)
         DO ispin = 1, nspin
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_inv, ls_scf_env%matrix_ks(ispin), &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)

            WRITE (title, '(A,I3,A)') "inv(S) * H(", ispin, ")"
            CALL report_matrix_sparsity(matrix_tmp1, unit_nr, title, ls_scf_env%eps_filter)
         ENDDO
      ENDIF

      IF (ls_scf_env%use_s_sqrt) THEN

         CALL report_matrix_sparsity(ls_scf_env%matrix_s_sqrt, unit_nr, "sqrt(S)", &
                                     ls_scf_env%eps_filter)
         CALL report_matrix_sparsity(ls_scf_env%matrix_s_sqrt_inv, unit_nr, "inv(sqrt(S))", &
                                     ls_scf_env%eps_filter)

         DO ispin = 1, nspin
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt_inv, ls_scf_env%matrix_ks(ispin), &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt_inv, &
                                   0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)
            WRITE (title, '(A,I3,A)') "inv(sqrt(S)) * H(", ispin, ") * inv(sqrt(S))"
            CALL report_matrix_sparsity(matrix_tmp2, unit_nr, title, ls_scf_env%eps_filter)
         ENDDO

         DO ispin = 1, nspin
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ls_scf_env%matrix_s_sqrt, ls_scf_env%matrix_p(ispin), &
                                   0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_s_sqrt, &
                                   0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)
            WRITE (title, '(A,I3,A)') "sqrt(S) * P(", ispin, ") * sqrt(S)"
            CALL report_matrix_sparsity(matrix_tmp2, unit_nr, title, ls_scf_env%eps_filter)
         ENDDO

      ENDIF

      CALL cp_dbcsr_release(matrix_tmp1)
      CALL cp_dbcsr_release(matrix_tmp2)

      CALL timestop(handle)

   END SUBROUTINE post_scf_sparsities

! **************************************************************************************************
!> \brief Helper routine to report on the sparsity of a single matrix,
!>        for several filtering values
!> \param matrix ...
!> \param unit_nr ...
!> \param title ...
!> \param eps ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE report_matrix_sparsity(matrix, unit_nr, title, eps)
      TYPE(cp_dbcsr_type)                                :: matrix
      INTEGER                                            :: unit_nr
      CHARACTER(LEN=*)                                   :: title
      REAL(KIND=dp)                                      :: eps

      CHARACTER(len=*), PARAMETER :: routineN = 'report_matrix_sparsity', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: eps_local, occ
      TYPE(cp_dbcsr_type)                                :: matrix_tmp

      CALL timeset(routineN, handle)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix, name=TRIM(title))
      CALL cp_dbcsr_copy(matrix_tmp, matrix, name=TRIM(title))

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T2,A)') "Sparsity for : "//TRIM(title)
      ENDIF

      eps_local = eps
      DO
         IF (eps_local > 1.1_dp) EXIT
         CALL cp_dbcsr_filter(matrix_tmp, eps_local)
         occ = cp_dbcsr_get_occupation(matrix_tmp)
         IF (unit_nr > 0) WRITE (unit_nr, '(T2,F16.12,A3,F16.12)') eps_local, " : ", occ
         eps_local = eps_local*10
      ENDDO

      CALL cp_dbcsr_release(matrix_tmp)

      CALL timestop(handle)

   END SUBROUTINE report_matrix_sparsity

! **************************************************************************************************
!> \brief Compute matrix_w as needed for the forces
!> \param matrix_w ...
!> \param ls_scf_env ...
!> \par History
!>       2010.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE calculate_w_matrix(matrix_w, ls_scf_env)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_w
      TYPE(ls_scf_env_type)                              :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_w_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: scaling
      TYPE(cp_dbcsr_type)                                :: matrix_tmp1, matrix_tmp2, matrix_tmp3

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(matrix_tmp1)
      CALL cp_dbcsr_create(matrix_tmp1, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp2)
      CALL cp_dbcsr_create(matrix_tmp2, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp3)
      CALL cp_dbcsr_create(matrix_tmp3, template=ls_scf_env%matrix_s, matrix_type=dbcsr_type_no_symmetry)

      IF (ls_scf_env%nspins == 1) THEN
         scaling = 0.5_dp
      ELSE
         scaling = 1.0_dp
      ENDIF

      DO ispin = 1, ls_scf_env%nspins

         CALL cp_dbcsr_copy(matrix_tmp3, ls_scf_env%matrix_ks(ispin))
         IF (ls_scf_env%has_s_preconditioner) THEN
            CALL apply_matrix_preconditioner(matrix_tmp3, "backward", &
                                             ls_scf_env%matrix_bs_sqrt, ls_scf_env%matrix_bs_sqrt_inv)
         ENDIF
         CALL cp_dbcsr_filter(matrix_tmp3, ls_scf_env%eps_filter)

         CALL cp_dbcsr_multiply("N", "N", scaling, ls_scf_env%matrix_p(ispin), matrix_tmp3, &
                                0.0_dp, matrix_tmp1, filter_eps=ls_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, ls_scf_env%matrix_p(ispin), &
                                0.0_dp, matrix_tmp2, filter_eps=ls_scf_env%eps_filter)
         CALL matrix_ls_to_qs(matrix_w(ispin)%matrix, matrix_tmp2, ls_scf_env%ls_mstruct, covariant=.FALSE.)
      ENDDO

      CALL cp_dbcsr_release(matrix_tmp1)
      CALL cp_dbcsr_release(matrix_tmp2)
      CALL cp_dbcsr_release(matrix_tmp3)

      CALL timestop(handle)

   END SUBROUTINE calculate_w_matrix

! **************************************************************************************************
!> \brief a place for quick experiments
!> \par History
!>       2010.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE post_scf_experiment()

      CHARACTER(len=*), PARAMETER :: routineN = 'post_scf_experiment', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE post_scf_experiment

END MODULE dm_ls_scf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types needed for MP2 calculations
!> \par History
!>       2011.05 created [Mauro Del Ben]
!> \author MDB
! **************************************************************************************************
MODULE mp2_setup
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE eri_mme_types,                   ONLY: eri_mme_init_read_input
   USE input_constants,                 ONLY: do_eri_mme
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE mp2_types,                       ONLY: mp2_method_direct,&
                                              mp2_method_gpw,&
                                              mp2_method_none,&
                                              mp2_ri_optimize_basis,&
                                              mp2_type,&
                                              ri_mp2_laplace,&
                                              ri_mp2_method_gpw,&
                                              ri_rpa_method_gpw
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_setup'

   PUBLIC :: read_mp2_section

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param input ...
!> \param mp2_env ...
! **************************************************************************************************
   SUBROUTINE read_mp2_section(input, mp2_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(mp2_type), INTENT(INOUT)                      :: mp2_env

      CHARACTER(len=*), PARAMETER :: routineN = 'read_mp2_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ival, unit_nr
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: eri_mme_section, mp2_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      mp2_section => section_vals_get_subs_vals(input, "DFT%XC%WF_CORRELATION")

      ! should come from input
      CALL section_vals_val_get(mp2_section, "METHOD", i_val=mp2_env%method)
      CALL section_vals_val_get(mp2_section, "MEMORY", r_val=mp2_env%mp2_memory)
      CALL section_vals_val_get(mp2_section, "SCALE_S", r_val=mp2_env%scale_S)
      CALL section_vals_val_get(mp2_section, "SCALE_T", r_val=mp2_env%scale_T)
      CALL section_vals_val_get(mp2_section, "GROUP_SIZE", i_val=mp2_env%mp2_num_proc)
      CALL section_vals_val_get(mp2_section, "DIRECT_CANONICAL%BIG_SEND", l_val=mp2_env%direct_canonical%big_send)
      CALL section_vals_val_get(mp2_section, "ROW_BLOCK", i_val=mp2_env%block_size_row)
      CALL section_vals_val_get(mp2_section, "COL_BLOCK", i_val=mp2_env%block_size_col)
      CALL section_vals_val_get(mp2_section, "RI_METRIC", i_val=mp2_env%ri_metric)
      CALL section_vals_val_get(mp2_section, "CALC_COND_NUM", l_val=mp2_env%calc_PQ_cond_num)

      CALL section_vals_val_get(mp2_section, "INTERACTION_POTENTIAL%POTENTIAL_TYPE", &
                                i_val=mp2_env%potential_parameter%potential_type)
      CALL section_vals_val_get(mp2_section, "INTERACTION_POTENTIAL%TRUNCATION_RADIUS", &
                                r_val=mp2_env%potential_parameter%truncation_radius)
      CALL section_vals_val_get(mp2_section, "INTERACTION_POTENTIAL%TShPSC_DATA", &
                                c_val=mp2_env%potential_parameter%filename)

      CALL section_vals_val_get(mp2_section, "ERI_METHOD", &
                                i_val=mp2_env%eri_method)

      CALL section_vals_val_get(mp2_section, "WFC_GPW%EPS_FILTER", &
                                r_val=mp2_env%mp2_gpw%eps_filter)
      CALL section_vals_val_get(mp2_section, "WFC_GPW%EPS_GRID", &
                                r_val=mp2_env%mp2_gpw%eps_grid)
      CALL section_vals_val_get(mp2_section, "WFC_GPW%CUTOFF", &
                                r_val=mp2_env%mp2_gpw%cutoff)
      CALL section_vals_val_get(mp2_section, "WFC_GPW%REL_CUTOFF", &
                                r_val=mp2_env%mp2_gpw%relative_cutoff)
      CALL section_vals_val_get(mp2_section, "WFC_GPW%MULTIPOLE_TWO_CENT_INT", &
                                l_val=mp2_env%mp2_gpw%do_mult_2c)
      CALL section_vals_val_get(mp2_section, "WFC_GPW%PRINT_LEVEL", &
                                i_val=mp2_env%mp2_gpw%print_level)

      CALL section_vals_val_get(mp2_section, "RI_RPA%RPA_NUM_QUAD_POINTS", &
                                i_val=mp2_env%ri_rpa%rpa_num_quad_points)
      CALL section_vals_val_get(mp2_section, "RI_RPA%SIZE_FREQ_INTEG_GROUP", &
                                i_val=mp2_env%ri_rpa%rpa_integ_group_size)
      CALL section_vals_val_get(mp2_section, "RI_RPA%MM_STYLE", &
                                i_val=mp2_env%ri_rpa%mm_style)
      CALL section_vals_val_get(mp2_section, "RI_RPA%MINIMAX_QUADRATURE", &
                                l_val=mp2_env%ri_rpa%minimax_quad)
      CALL section_vals_val_get(mp2_section, "RI_RPA%ADMM", &
                                l_val=mp2_env%ri_rpa%do_admm)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0", &
                                l_val=mp2_env%ri_rpa%do_ri_g0w0)
      CALL section_vals_val_get(mp2_section, "RI_RPA%IM_TIME", &
                                l_val=mp2_env%ri_rpa%do_im_time)

      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CORR_MOS_OCC", &
                                i_val=mp2_env%ri_g0w0%corr_mos_occ)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CORR_MOS_VIRT", &
                                i_val=mp2_env%ri_g0w0%corr_mos_virt)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%SCALING", &
                                r_val=mp2_env%ri_g0w0%scaling)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%NUMB_POLES", &
                                i_val=mp2_env%ri_g0w0%num_poles)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%OMEGA_MAX_FIT", &
                                r_val=mp2_env%ri_g0w0%omega_max_fit)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%STOP_CRIT", &
                                r_val=mp2_env%ri_g0w0%stop_crit)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%MAX_ITER_FIT", &
                                i_val=mp2_env%ri_g0w0%max_iter_fit)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CHECK_FIT", &
                                l_val=mp2_env%ri_g0w0%check_fit)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CROSSING_SEARCH", &
                                i_val=mp2_env%ri_g0w0%crossing_search)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%FERMI_LEVEL_OFFSET", &
                                r_val=mp2_env%ri_g0w0%fermi_level_offset)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%EV_SC_ITER", &
                                i_val=mp2_env%ri_g0w0%iter_ev_sc)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%EPS_EV_SC_ITER", &
                                r_val=mp2_env%ri_g0w0%eps_ev_sc_iter)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%EV_SC_GW_REMOVE_NEG_VIRT_ENERGIES", &
                                l_val=mp2_env%ri_g0w0%remove_neg_virt_energies)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%HF_LIKE_EV_START", &
                                l_val=mp2_env%ri_g0w0%hf_like_ev_start)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%PRINT_GW_DETAILS", &
                                l_val=mp2_env%ri_g0w0%print_gw_details)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%RI_SIGMA_X", &
                                l_val=mp2_env%ri_g0w0%do_ri_Sigma_x)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%RI_METRIC", &
                                i_val=mp2_env%ri_g0w0%ri_metric)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%MIX_EXCHANGE", &
                                l_val=mp2_env%ri_g0w0%mix_exchange)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%FRACTION_EXX", &
                                r_val=mp2_env%ri_g0w0%frac_exx)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CONTOUR_DEF_START", &
                                i_val=mp2_env%ri_g0w0%contour_def_start)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CONTOUR_DEF_END", &
                                i_val=mp2_env%ri_g0w0%contour_def_end)
      CALL section_vals_val_get(mp2_section, "RI_RPA%RI_G0W0%CONTOUR_DEF_OFFSET", &
                                r_val=mp2_env%ri_g0w0%contour_def_offset)

      CALL section_vals_val_get(mp2_section, "RI_LAPLACE%QUADRATURE_POINTS", &
                                i_val=mp2_env%ri_laplace%n_quadrature)
      CALL section_vals_val_get(mp2_section, "RI_LAPLACE%SIZE_INTEG_GROUP", &
                                i_val=mp2_env%ri_laplace%integ_group_size)

      CALL section_vals_val_get(mp2_section, "RI_MP2%BLOCK_SIZE", &
                                i_val=mp2_env%ri_mp2%block_size)
      CALL section_vals_val_get(mp2_section, "RI_MP2%EPS_CANONICAL", &
                                r_val=mp2_env%ri_mp2%eps_canonical)
      CALL section_vals_val_get(mp2_section, "RI_MP2%FREE_HFX_BUFFER", &
                                l_val=mp2_env%ri_mp2%free_hfx_buffer)

      CALL section_vals_val_get(mp2_section, "CPHF%MAX_ITER", &
                                i_val=mp2_env%ri_grad%cphf_max_num_iter)
      CALL section_vals_val_get(mp2_section, "CPHF%EPS_CONV", &
                                r_val=mp2_env%ri_grad%cphf_eps_conv)

      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%DELTA_I_REL", &
                                r_val=mp2_env%ri_opt_param%DI_rel)
      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%DELTA_RI", &
                                r_val=mp2_env%ri_opt_param%DRI)
      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%EPS_DERIV", &
                                r_val=mp2_env%ri_opt_param%eps_step)
      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%MAX_ITER", &
                                i_val=mp2_env%ri_opt_param%max_num_iter)
      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%BASIS_SIZE", &
                                i_val=mp2_env%ri_opt_param%basis_quality)
      NULLIFY (tmplist)
      CALL section_vals_val_get(mp2_section, "OPT_RI_BASIS%NUM_FUNC", &
                                i_vals=tmplist)
      IF (tmplist(1) > 0) THEN
         ALLOCATE (mp2_env%ri_opt_param%RI_nset_per_l(0:SIZE(tmplist)-1))
         mp2_env%ri_opt_param%RI_nset_per_l = 0
         DO ival = 1, SIZE(tmplist)
            mp2_env%ri_opt_param%RI_nset_per_l(ival-1) = tmplist(ival)
         END DO
      END IF

      NULLIFY (mp2_env%eri_mme_param)
      IF (mp2_env%eri_method .EQ. do_eri_mme) THEN
         eri_mme_section => section_vals_get_subs_vals(mp2_section, "ERI_MME")
         ALLOCATE (mp2_env%eri_mme_param)
         CALL eri_mme_init_read_input(eri_mme_section, mp2_env%eri_mme_param)
      ENDIF

      ! Set some parameters in case of P screening
      mp2_env%not_last_hfx = .TRUE.
      mp2_env%p_screen = .TRUE.

      ! print some info about the MP2 parameters
      unit_nr = cp_print_key_unit_nr(logger, mp2_section, "MP2_INFO", &
                                     extension=".mp2Log")
      IF ((mp2_env%method .NE. mp2_method_none) .AND. unit_nr > 0) THEN
         WRITE (unit_nr, '(T2,A)') ""
         SELECT CASE (mp2_env%method)
         CASE (mp2_method_direct)
            WRITE (unit_nr, '(T2,A)') "MP2| using direct canonical MP2"
         CASE (mp2_method_gpw)
            WRITE (unit_nr, '(T2,A)') "MP2| using MP2 GPW style"
         CASE (ri_mp2_method_gpw)
            WRITE (unit_nr, '(T2,A)') "MP2| using RI-MP2-GPW"
         CASE (ri_rpa_method_gpw)
            WRITE (unit_nr, '(T2,A)') "RI-RPA| using GPW style"
         CASE (ri_mp2_laplace)
            WRITE (unit_nr, '(T2,A)') "RI-SOS-Laplace-MP2| using GPW style"
         CASE (mp2_ri_optimize_basis)
            WRITE (unit_nr, '(T2,A)') "MP2| Optimize RI auxiliary basis"
         CASE DEFAULT
            CALL cp__b("mp2_setup.F",227,"")
         END SELECT
         WRITE (unit_nr, '(T2,A)') ""
      ENDIF
      CALL cp_print_key_finished_output(unit_nr, logger, mp2_section, &
                                        "MP2_INFO")

      CALL timestop(handle)

   END SUBROUTINE read_mp2_section

END MODULE mp2_setup

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
!> \author APSI & CJM & JGH
! **************************************************************************************************
MODULE header
   USE cp2k_info,                       ONLY: r_cwd,&
                                              r_datx,&
                                              r_host_name,&
                                              r_pid,&
                                              r_user_name
   USE kinds,                           ONLY: default_path_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: cp2k_header, cp2k_footer, vib_header, scptb_header
   PUBLIC :: fist_header, se_header, dftb_header, qs_header, tddfpt_header
   PUBLIC :: qmmm_header, atom_header, atom_footer, band_header
   PUBLIC :: tmc_header, tmc_ana_header

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param wdir ...
! **************************************************************************************************
   SUBROUTINE cp2k_header(iw, wdir)
      INTEGER, INTENT(IN)                                :: iw
      CHARACTER(LEN=default_path_length), OPTIONAL       :: wdir

      CHARACTER(LEN=default_path_length)                 :: cwd
      INTEGER                                            :: l

      IF (iw < 0) RETURN
      cwd = r_cwd
      IF (PRESENT(wdir)) cwd = wdir
      WRITE (UNIT=iw, &
             FMT="(/,A,T55,A26,/,A,T51,A30,/,A,T51,A30,/,A,T71,I10)") &
         "  **** **** ******  **  PROGRAM STARTED AT ", ADJUSTR(r_datx), &
         " ***** ** ***  *** **   PROGRAM STARTED ON ", ADJUSTR(r_host_name(1:30)), &
         " **    ****   ******    PROGRAM STARTED BY ", ADJUSTR(r_user_name(1:30)), &
         " ***** **    ** ** **   PROGRAM PROCESS ID ", r_pid
      WRITE (UNIT=iw, FMT="(A,T44,A37)") &
         "  **** **  *******  **  PROGRAM STARTED IN ", ADJUSTR(cwd(1:37))
      DO l = 38, LEN_TRIM(cwd), 37
         WRITE (UNIT=iw, FMT="(T44,A)") cwd(l:MIN(LEN_TRIM(cwd), l+36))
      END DO

   END SUBROUTINE cp2k_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param wdir ...
! **************************************************************************************************
   SUBROUTINE cp2k_footer(iw, wdir)
      INTEGER, INTENT(IN)                                :: iw
      CHARACTER(LEN=default_path_length), OPTIONAL       :: wdir

      CHARACTER(LEN=default_path_length)                 :: cwd
      INTEGER                                            :: l

      IF (iw < 0) RETURN
      cwd = r_cwd
      IF (PRESENT(wdir)) cwd = wdir
      WRITE (UNIT=iw, &
             FMT="(A,T55,A26,/,A,T51,A30,/,A,T51,A30,/,A,T71,I10)") &
         "  **** **** ******  **  PROGRAM ENDED AT   ", ADJUSTR(r_datx), &
         " ***** ** ***  *** **   PROGRAM RAN ON     ", ADJUSTR(r_host_name(1:30)), &
         " **    ****   ******    PROGRAM RAN BY     ", ADJUSTR(r_user_name(1:30)), &
         " ***** **    ** ** **   PROGRAM PROCESS ID ", r_pid
      WRITE (UNIT=iw, FMT="(A,T44,A37)") &
         "  **** **  *******  **  PROGRAM STOPPED IN ", ADJUSTR(cwd(1:37))
      DO l = 38, LEN_TRIM(cwd), 37
         WRITE (UNIT=iw, FMT="(T44,A)") cwd(l:MIN(LEN_TRIM(cwd), l+36))
      END DO
   END SUBROUTINE cp2k_footer

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE qmmm_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '(/,(T2,A79))') &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**             #######  ##     ## ##     ## ##     ##                        **", &
         "**            ##     ## ###   ### ###   ### ###   ###                        **", &
         "**            ##     ## #### #### #### #### #### ####                        **", &
         "**            ##     ## ## ### ## ## ### ## ## ### ##                        **", &
         "**            ##  ## ## ##     ## ##     ## ##     ##                        **", &
         "**            ##    ##  ##     ## ##     ## ##     ##  T.Laino and F.Mohamed **", &
         "**             ##### ## ##     ## ##     ## ##     ##  2005 - 2015           **", &
         "**                                                                           **", &
         "**                                                     Calculation Started.. **", &
         "*******************************************************************************", &
         "*******************************************************************************"

   END SUBROUTINE qmmm_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE tddfpt_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '(/,(T2,A79))') &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**   ######## #######   #######   ######## ########                          **", &
         "**      ##    ##     #  ##     #  ##          ##                             **", &
         "**      ##    ##     ## ##     ## ##          ##                             **", &
         "**      ##    ##     ## ##     ## ######      ##                             **", &
         "**      ##    ##     ## ##     ## ##          ##                             **", &
         "**      ##    ##     #  ##     #  ##          ##    T.Chassaing and J.Hutter **", &
         "**      ##    #######   #######   ##          ##    2005                     **", &
         "**                                                                           **", &
         "**                                                     Calculation Started.. **", &
         "*******************************************************************************", &
         "*******************************************************************************"

   END SUBROUTINE tddfpt_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE fist_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw > 0) THEN
         WRITE (iw, '( / )')
         WRITE (iw, '( 14(20x,a,/) )') &
            '  ************* ******** ***********   ', &
            ' ************* **********************  ', &
            ' ****     **** ****      ***********   ', &
            ' ******** ****  *******     ****       ', &
            ' *******  ****   *******   ****        ', &
            ' ****     ****       ****  ****        ', &
            ' ****     **** **********  ****        ', &
            '  ****     **** ********    ****       ', &
            ' FRONTIERS IN SIMULATION TECHNOLOGY    ', &
            '                                       ', &
            '    C.J. Mundy, S. Balasubramanian,    ', &
            ' Ken Bagchi, J. Hutter, Ari Seitsonen  ', &
            '  IFW Kuo, T. Laino, J. VandeVondele   ', &
            '             Version 1.0               ', &
            '                                       '
      END IF

   END SUBROUTINE fist_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE dftb_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '( / )')
      WRITE (iw, '( 11(12x,a,/) )') &
         '  #####   #####        # ######  ####### ####### ######   ', &
         ' #     # #     #      #  #     # #          #    #     #  ', &
         ' #     # #           #   #     # #          #    #     #  ', &
         ' #     #  #####     #    #     # #####      #    ######   ', &
         ' #   # #       #   #     #     # #          #    #     #  ', &
         ' #    #  #     #  #      #     # #          #    #     #  ', &
         '  #### #  #####  #       ######  #          #    ######   ', &
         '                                                          ', &
         '                        Version 0.2                       ', &
         '                   J. Hutter and T. Heine                 ', &
         '                                                          '
   END SUBROUTINE dftb_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE scptb_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '( / )')
      WRITE (iw, '( 11(12x,a,/) )') &
         ' #######################################################  ', &
         ' #######################################################  ', &
         ' #######################################################  ', &
         ' #######################################################  ', &
         ' #######################################################  ', &
         ' #######################################################  ', &
         ' #######################################################  ', &
         '                                                          ', &
         '                        Version 0.01                      ', &
         '         Juerg Hutter, Yannick Misteli, Ralph Koitz       ', &
         '                                                          '
   END SUBROUTINE scptb_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE se_header(iw)
      INTEGER, INTENT(in)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '( / )')
      WRITE (iw, '( 12(12x,a,/) )') &
         '  #####   #####        # #     # ######  ######   #####  ', &
         ' #     # #     #      #  ##    # #     # #     # #     # ', &
         ' #     # #           #   # #   # #     # #     # #     # ', &
         ' #     #  #####     #    #  #  # #     # #     # #     # ', &
         ' #   # #       #   #     #   # # #     # #     # #     # ', &
         ' #    #  #     #  #      #    ## #     # #     # #     # ', &
         '  #### #  #####  #       #     # ######  ######   #####  ', &
         '                                                         ', &
         '                  T. Laino and J. Hutter                 ', &
         '                                                         ', &
         '                       2008 - 2015                       '

   END SUBROUTINE se_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE qs_header(iw)
      INTEGER, INTENT(IN)                                :: iw

      IF (iw < 0) RETURN
      WRITE (UNIT=iw, FMT="(/,(T2,A79))") &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**     #####                         ##              ##                      **", &
         "**    ##   ##            ##          ##              ##                      **", &
         "**   ##     ##                       ##            ######                    **", &
         "**   ##     ##  ##   ##  ##   #####  ##  ##   ####   ##    #####    #####    **", &
         "**   ##     ##  ##   ##  ##  ##      ## ##   ##      ##   ##   ##  ##   ##   **", &
         "**   ##  ## ##  ##   ##  ##  ##      ####     ###    ##   ######   ######    **", &
         "**    ##  ###   ##   ##  ##  ##      ## ##      ##   ##   ##       ##        **", &
         "**     #######   #####   ##   #####  ##  ##  ####    ##    #####   ##        **", &
         "**           ##                                                    ##        **", &
         "**                                                                           **", &
         "**                                                ... make the atoms dance   **", &
         "**                                                                           **", &
         "**            Copyright (C) by CP2K developers group (2000 - 2016)           **", &
         "**                                                                           **", &
         "*******************************************************************************"
   END SUBROUTINE qs_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_header(iw)
      INTEGER, INTENT(IN)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '( / )')
      WRITE (iw, '( 14(25x,a,/) )') &
         '  ****  ******  ****   ****   ', &
         ' **  ** ****** **  ** ******  ', &
         ' ******   **   **  ** **  **  ', &
         ' **  **   **    ****  **  **  ', &
         '                              ', &
         '    University of Zurich      ', &
         '        2009 - 2015           ', &
         '                              ', &
         '        Version 0.0           ', &
         '                                          '
   END SUBROUTINE atom_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_footer(iw)
      INTEGER, INTENT(IN)                                :: iw

      IF (iw < 0) RETURN
      WRITE (iw, '( / )')
      WRITE (iw, '( 14(25x,a,/) )') &
         '    NORMAL TERMINATION OF     ', &
         '                              ', &
         '  ****  ******  ****   ****   ', &
         ' **  ** ****** **  ** ******  ', &
         ' ******   **   **  ** **  **  ', &
         ' **  **   **    ****  **  **  '

   END SUBROUTINE atom_footer

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param ni ...
!> \param nr ...
!> \param np ...
! **************************************************************************************************
   SUBROUTINE band_header(iw, ni, nr, np)
      INTEGER, INTENT(IN)                                :: iw, ni, nr, np

      IF (iw < 0) RETURN
      WRITE (UNIT=iw, FMT="(/,(T2,A79))") &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**  ########     ###    ##    ## ########    Nudged Elastic Band Method      **", &
         "**  ##     ##   ## ##   ###   ## ##     ##   String Method                   **", &
         "**  ##     ##  ##   ##  ####  ## ##     ##                                   **"
      WRITE (UNIT=iw, FMT="(T2,A,T73,I5,1X,A)") &
         "**  ########  ##     ## ## ## ## ##     ##   Number of Images :", ni, "**", &
         "**  ##     ## ######### ##  #### ##     ##   Number of Replicas:", nr, "**", &
         "**  ##     ## ##     ## ##   ### ##     ##   Number of Procs/Rep:", np, "**"
      WRITE (UNIT=iw, FMT="(T2,A79)") &
         "**  ########  ##     ## ##    ## ########                                    **", &
         "**                                           T. Laino  2009 - 2015           **", &
         "*******************************************************************************"
      WRITE (UNIT=iw, FMT="(T2,A79,/)") &
         "*******************************************************************************"

   END SUBROUTINE band_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param nr ...
!> \param np ...
! **************************************************************************************************
   SUBROUTINE vib_header(iw, nr, np)
      INTEGER, INTENT(IN)                                :: iw, nr, np

      IF (iw < 0) RETURN
      WRITE (UNIT=iw, FMT="(/,(T2,A79))") &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**  #  #   #    #                  #     #                      ##           **", &
         "**  #  #        ###   # ##   ###  ####         ##   # #    ###   #           **", &
         "**  #  #   #    #  #  ##    #  #   #     #    #  #  ## #  #  #   #           **", &
         "**   ##    #    #  #  #     # ##   #     #    #  #  #  #  # ##   #           **", &
         "**   ##    #    ###   #      # #    ##   #     ##   #  #   # #  ###          **", &
         "**                                                                           **", &
         "**   ##               ##                 #                                   **", &
         "**  #  #  # #    ###   #    #  #   ###         ###                           **", &
         "**  #  #  ## #  #  #   #    #  #  ##     #    ##                             **"
      WRITE (UNIT=iw, FMT="(T2,A,T73,I5,1X,A)") &
         "**  ####  #  #  # ##   #     ###    ##   #      ##  N. Replicas:", nr, "**", &
         "**  #  #  #  #   # #  ###      #  ###    #    ###   N. Procs/Rep:", np, "**"
      WRITE (UNIT=iw, FMT="(T2,A79)") &
         "**                           ##                                              **", &
         "**                                       T. Laino  and F. Schiffmann         **", &
         "**                                       2008 - 2015                         **", &
         "*******************************************************************************"
      WRITE (UNIT=iw, FMT="(T2,A79,/)") &
         "*******************************************************************************"

   END SUBROUTINE vib_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE tmc_header(iw)
      INTEGER, INTENT(IN)                                :: iw

      IF (iw < 0) RETURN
      WRITE (UNIT=iw, FMT="(/,(T2,A79))") &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**                      ###########  ###     ###    ######                   **", &
         "**                          ###      ####   ####   ###                       **", &
         "**                          ###      ### # # ###  ###                        **", &
         "**                          ###      ###  #  ###   ###                       **", &
         "**                          ###      ###     ###    ######                   **", &
         "**  TreeMonteCarlo                                        Mandes Schoenherr  **", &
         "**                                                        2009 - 2015        **", &
         "*******************************************************************************", &
         "*******************************************************************************"
   END SUBROUTINE tmc_header

! **************************************************************************************************
!> \brief ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE tmc_ana_header(iw)
      INTEGER, INTENT(IN)                                :: iw

      IF (iw < 0) RETURN
      WRITE (UNIT=iw, FMT="(/,(T2,A79))") &
         "*******************************************************************************", &
         "*******************************************************************************", &
         "**                                                                           **", &
         "**  ###########  ###     ###    ######              ##    ###   ##    ##     **", &
         "**      ###      ####   ####   ###                 ####   ####  ##   ####    **", &
         "**      ###      ### # # ###  ###                 ##  ##  ## ## ##  ##  ##   **", &
         "**      ###      ###  #  ###   ###                ######  ##  ####  ######   **", &
         "**      ###      ###     ###    ######            ##  ##  ##   ###  ##  ##   **", &
         "**  Trajectory analyzer for TreeMonteCarlo ensembles      Mandes Schoenherr  **", &
         "**                                                        2013 - 2015        **", &
         "*******************************************************************************", &
         "*******************************************************************************"
   END SUBROUTINE tmc_ana_header
END MODULE header

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - Merged with the Quickstep MODULE method_specification (17.01.2002,MK)
!>      - USE statements cleaned, added
!>        (25.09.2002,MK)
!>      - Added more LSD structure (01.2003,Joost VandeVondele)
!>      - New molecule data types introduced (Sep. 2003,MK)
!>      - Cleaning; getting rid of pnode (02.10.2003,MK)
!>      - Sub-system setup added (08.10.2003,MK)
!> \author MK (18.05.2000)
! **************************************************************************************************
MODULE qs_environment
   USE atom_kind_orbitals,              ONLY: calculate_atomic_relkin
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE bibliography,                    ONLY: Iannuzzi2006,&
                                              Iannuzzi2007,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_release,&
                                              dft_control_type,&
                                              dftb_control_type,&
                                              gapw_control_type,&
                                              qs_control_type,&
                                              scptb_control_type,&
                                              semi_empirical_control_type
   USE cp_control_utils,                ONLY: read_becke_section,&
                                              read_ddapc_section,&
                                              read_dft_control,&
                                              read_mgrid_section,&
                                              read_qs_section,&
                                              read_tddfpt_control,&
                                              write_dft_control,&
                                              write_qs_control
   USE cp_ddapc_types,                  ONLY: cp_ddapc_ewald_create
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE cp_symmetry,                     ONLY: write_symmetry
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_type
   USE distribution_methods,            ONLY: distribute_molecules_1d
   USE dm_ls_scf,                       ONLY: ls_scf_create
   USE et_coupling_types,               ONLY: et_coupling_create
   USE ewald_environment_types,         ONLY: ewald_env_create,&
                                              ewald_env_get,&
                                              ewald_env_release,&
                                              ewald_env_set,&
                                              ewald_environment_type,&
                                              read_ewald_section
   USE ewald_pw_methods,                ONLY: ewald_pw_grid_update
   USE ewald_pw_types,                  ONLY: ewald_pw_create,&
                                              ewald_pw_release,&
                                              ewald_pw_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_create,&
                                              fist_nonbond_env_type
   USE gamma,                           ONLY: init_md_ftable
   USE global_types,                    ONLY: global_environment_type
   USE hartree_local_methods,           ONLY: init_coulomb_local
   USE header,                          ONLY: dftb_header,&
                                              qs_header,&
                                              scptb_header,&
                                              se_header
   USE hfx_types,                       ONLY: hfx_create
   USE input_constants,                 ONLY: &
        dispersion_d3, do_et_becke, do_et_ddapc, do_method_am1, do_method_dftb, do_method_gapw, &
        do_method_gapw_xc, do_method_gpw, do_method_lrigpw, do_method_mndo, do_method_mndod, &
        do_method_ofgpw, do_method_pdg, do_method_pm3, do_method_pm6, do_method_pnnl, &
        do_method_rm1, do_method_scptb, do_qmmm_gauss, do_qmmm_swave, general_roks, rel_none, &
        rel_trans_atom, vdw_pairpot_dftd3, xc_vdw_fun_none, xc_vdw_fun_nonloc, xc_vdw_fun_pairpot
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kg_environment,                  ONLY: kg_env_create
   USE kinds,                           ONLY: dp
   USE kpoint_methods,                  ONLY: kpoint_env_initialize,&
                                              kpoint_initialize,&
                                              kpoint_initialize_mos
   USE kpoint_types,                    ONLY: kpoint_type,&
                                              read_kpoint_section,&
                                              write_kpoint_info
   USE lri_environment_init,            ONLY: lri_env_init
   USE machine,                         ONLY: m_flush
   USE molecule_kind_types,             ONLY: molecule_kind_type,&
                                              write_molecule_kind_set
   USE molecule_types_new,              ONLY: molecule_type
   USE mp2_setup,                       ONLY: read_mp2_section
   USE mp2_types,                       ONLY: mp2_env_create
   USE multipole_types,                 ONLY: do_multipole_none
   USE orbital_pointers,                ONLY: init_orbital_pointers
   USE orbital_transformation_matrices, ONLY: init_spherical_harmonics
   USE particle_methods,                ONLY: write_particle_distances,&
                                              write_qs_particle_coordinates,&
                                              write_structure_data
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type
   USE qs_dftb_parameters,              ONLY: qs_dftb_param_init
   USE qs_dftb_types,                   ONLY: qs_dftb_pairpot_type
   USE qs_dispersion_nonloc,            ONLY: qs_dispersion_nonloc_init
   USE qs_dispersion_pairpot,           ONLY: qs_dispersion_pairpot_init
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE qs_dispersion_utils,             ONLY: qs_dispersion_env_set,&
                                              qs_write_dispersion
   USE qs_energy_types,                 ONLY: allocate_qs_energy,&
                                              qs_energy_type
   USE qs_environment_methods,          ONLY: qs_env_setup
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_interactions,                 ONLY: init_interaction_radii,&
                                              init_se_nlradius,&
                                              write_core_charge_radii,&
                                              write_paw_radii,&
                                              write_pgf_orb_radii,&
                                              write_ppl_radii,&
                                              write_ppnl_radii
   USE qs_kind_types,                   ONLY: check_qs_kind_set,&
                                              get_qs_kind_set,&
                                              init_gapw_basis_set,&
                                              init_qs_kind_set,&
                                              qs_kind_type,&
                                              set_qs_kind,&
                                              write_gto_basis_sets,&
                                              write_qs_kind_set
   USE qs_ks_types,                     ONLY: qs_ks_env_create,&
                                              qs_ks_env_type,&
                                              qs_ks_release,&
                                              set_ks_env
   USE qs_mo_types,                     ONLY: allocate_mo_set,&
                                              mo_set_p_type
   USE qs_rho0_ggrid,                   ONLY: rho0_s_grid_create
   USE qs_rho0_methods,                 ONLY: init_rho0
   USE qs_rho0_types,                   ONLY: rho0_mpole_type
   USE qs_rho_atom_methods,             ONLY: init_rho_atom
   USE qs_subsys_methods,               ONLY: qs_subsys_create
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_release,&
                                              qs_subsys_set,&
                                              qs_subsys_type
   USE qs_wf_history_methods,           ONLY: wfi_create,&
                                              wfi_create_for_kp
   USE qs_wf_history_types,             ONLY: qs_wf_history_type,&
                                              wfi_release
   USE rel_control_types,               ONLY: rel_c_create,&
                                              rel_c_read_parameters,&
                                              rel_c_release,&
                                              rel_control_type
   USE scf_control_types,               ONLY: scf_c_create,&
                                              scf_c_read_parameters,&
                                              scf_c_release,&
                                              scf_c_write_parameters,&
                                              scf_control_type
   USE scptb_utils,                     ONLY: scptb_parameter_init
   USE semi_empirical_expns3_methods,   ONLY: semi_empirical_expns3_setup
   USE semi_empirical_int_arrays,       ONLY: init_se_intd_array
   USE semi_empirical_mpole_methods,    ONLY: nddo_mpole_setup
   USE semi_empirical_mpole_types,      ONLY: nddo_mpole_type
   USE semi_empirical_store_int_types,  ONLY: semi_empirical_si_create,&
                                              semi_empirical_si_type
   USE semi_empirical_types,            ONLY: se_taper_create,&
                                              se_taper_type
   USE semi_empirical_utils,            ONLY: se_cutoff_compatible
   USE transport,                       ONLY: transport_env_create
!  USE xas_control,                     ONLY: write_xas_control
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_environment'

   ! *** Public subroutines ***
   PUBLIC :: qs_init

CONTAINS

! **************************************************************************************************
!> \brief Read the input and the database files for the setup of the
!>      QUICKSTEP environment.
!> \param qs_env ...
!> \param para_env ...
!> \param globenv ...
!> \param root_section ...
!> \param cp_subsys ...
!> \param cell ...
!> \param cell_ref ...
!> \param qmmm ...
!> \param qmmm_env_qm ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \author Creation (22.05.2000,MK)
! **************************************************************************************************
   SUBROUTINE qs_init(qs_env, para_env, globenv, root_section, cp_subsys, cell, cell_ref, qmmm, &
                      qmmm_env_qm, force_env_section, subsys_section, use_motion_section)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      LOGICAL, INTENT(IN), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_qm_type), OPTIONAL, POINTER          :: qmmm_env_qm
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, method_id, natom, nkind
      LOGICAL                                            :: do_et, do_exx, do_hfx, do_kpoints, &
                                                            mp2_present, my_qmmm, qmmm_decoupl, &
                                                            use_ref_cell
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rtmat
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: my_cell, my_cell_ref
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      TYPE(rel_control_type), POINTER                    :: rel_control
      TYPE(section_vals_type), POINTER                   :: dft_section, et_coupling_section, &
                                                            hfx_section, kpoint_section, &
                                                            mp2_section, transport_section

      NULLIFY (my_cell, my_cell_ref, atomic_kind_set, particle_set, &
               qs_kind_set, kpoint_section, dft_section, &
               subsys, ks_env, dft_control, blacs_env)

      CALL set_qs_env(qs_env, input=force_env_section)
      IF (.NOT. ASSOCIATED(subsys_section)) THEN
         subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      END IF

      ! QMMM
      my_qmmm = .FALSE.
      IF (PRESENT(qmmm)) my_qmmm = qmmm
      qmmm_decoupl = .FALSE.
      IF (PRESENT(qmmm_env_qm)) THEN
         IF (qmmm_env_qm%qmmm_coupl_type == do_qmmm_gauss .OR. &
             qmmm_env_qm%qmmm_coupl_type == do_qmmm_swave) THEN
            ! For GAUSS/SWAVE methods there could be a DDAPC decoupling requested
            qmmm_decoupl = my_qmmm .AND. qmmm_env_qm%periodic .AND. qmmm_env_qm%multipole
         END IF
         qs_env%qmmm_env_qm => qmmm_env_qm
      END IF
      CALL set_qs_env(qs_env=qs_env, qmmm=my_qmmm)

      ! Possibly initialize arrays for SE
      CALL section_vals_val_get(force_env_section, "DFT%QS%METHOD", i_val=method_id)
      SELECT CASE (method_id)
      CASE (do_method_rm1, do_method_am1, do_method_mndo, do_method_pdg, &
            do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
         CALL init_se_intd_array()
      CASE DEFAULT
         ! Do nothing
      END SELECT

      CALL qs_subsys_create(subsys, para_env, root_section=root_section, &
                            force_env_section=force_env_section, &
                            subsys_section=subsys_section, &
                            use_motion_section=use_motion_section, &
                            cp_subsys=cp_subsys, cell=cell, cell_ref=cell_ref)

      CALL qs_ks_env_create(ks_env)
      CALL set_ks_env(ks_env, subsys=subsys)
      CALL set_qs_env(qs_env, ks_env=ks_env)

      CALL qs_subsys_get(subsys, &
                         cell=my_cell, &
                         cell_ref=my_cell_ref, &
                         use_ref_cell=use_ref_cell, &
                         atomic_kind_set=atomic_kind_set, &
                         qs_kind_set=qs_kind_set, &
                         particle_set=particle_set)

      CALL set_ks_env(ks_env, para_env=para_env)
      CALL cp_blacs_env_create(blacs_env, para_env, globenv%blacs_grid_layout, &
                               globenv%blacs_repeatable)
      CALL set_ks_env(ks_env, blacs_env=blacs_env)
      CALL cp_blacs_env_release(blacs_env)

      !   *** Setup the grids for the G-space Interpolation if any
      CALL cp_ddapc_ewald_create(qs_env%cp_ddapc_ewald, qmmm_decoupl, my_cell, &
                                 force_env_section, subsys_section, para_env)

      ! kpoints
      CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
      kpoint_section => section_vals_get_subs_vals(qs_env%input, "DFT%KPOINTS")
      CALL read_kpoint_section(kpoints, kpoint_section)
      CALL kpoint_initialize(kpoints, particle_set, my_cell)
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      CALL write_kpoint_info(kpoints, dft_section)

      CALL qs_init_subsys(qs_env, para_env, subsys, my_cell, my_cell_ref, use_ref_cell, &
                          root_section, subsys_section)

      ! more kpoint stuff
      kpoints%para_env => para_env
      CALL cp_para_env_retain(para_env)
      CALL get_qs_env(qs_env=qs_env, blacs_env=blacs_env)
      kpoints%blacs_env_all => blacs_env
      CALL cp_blacs_env_retain(blacs_env)
      CALL get_qs_env(qs_env=qs_env, do_kpoints=do_kpoints)
      IF (do_kpoints) THEN
         CALL kpoint_env_initialize(kpoints)
         CALL kpoint_initialize_mos(kpoints, qs_env%mos)
         CALL get_qs_env(qs_env=qs_env, wf_history=wf_history)
         CALL wfi_create_for_kp(wf_history)
      END IF

      do_hfx = .FALSE.
      hfx_section => section_vals_get_subs_vals(qs_env%input, "DFT%XC%HF")
      CALL section_vals_get(hfx_section, explicit=do_hfx)
      CALL get_qs_env(qs_env, dft_control=dft_control)
      IF (do_hfx) THEN
         ! Retrieve particle_set and atomic_kind_set (needed for both kinds of initialization)
         natom = SIZE(particle_set)
         CALL hfx_create(qs_env%x_data, para_env, hfx_section, natom, atomic_kind_set, &
                         qs_kind_set, dft_control, my_cell)
      END IF

      mp2_section => section_vals_get_subs_vals(qs_env%input, "DFT%XC%WF_CORRELATION")
      CALL section_vals_get(mp2_section, explicit=mp2_present)
      IF (mp2_present) THEN
         CALL mp2_env_create(qs_env%mp2_env)
         CALL read_mp2_section(qs_env%input, qs_env%mp2_env)
         ! create the EXX section if necessary
         do_exx = .FALSE.
         hfx_section => section_vals_get_subs_vals(qs_env%input, "DFT%XC%WF_CORRELATION%RI_RPA%HF")
         CALL section_vals_get(hfx_section, explicit=do_exx)
         IF (do_exx) THEN
            ! Retrieve particle_set and atomic_kind_set (needed for both kinds of initialization)
            natom = SIZE(particle_set)
            CALL hfx_create(qs_env%mp2_env%ri_rpa%x_data, para_env, hfx_section, natom, atomic_kind_set, &
                            qs_kind_set, dft_control, my_cell, do_exx=.TRUE.)
         END IF
      END IF

      IF (dft_control%qs_control%do_kg) THEN
         CALL cite_reference(Iannuzzi2006)
         CALL kg_env_create(qs_env%kg_env, qs_env%input)
      END IF

      et_coupling_section => section_vals_get_subs_vals(qs_env%input, &
                                                        "PROPERTIES%ET_COUPLING")
      CALL section_vals_get(et_coupling_section, explicit=do_et)
      IF (do_et) CALL et_coupling_create(qs_env%et_coupling)

      ! lri env
      IF (method_id == do_method_lrigpw .OR. dft_control%qs_control%lri_optbas) THEN
         CALL lri_env_init(qs_env, force_env_section, qs_kind_set)
      END IF

      transport_section => section_vals_get_subs_vals(qs_env%input, "DFT%TRANSPORT")
      CALL section_vals_get(transport_section, explicit=qs_env%do_transport)
      IF (qs_env%do_transport) THEN
         CALL transport_env_create(qs_env)
      END IF

      IF (dft_control%qs_control%do_ls_scf) THEN
         CALL ls_scf_create(qs_env)
      ENDIF

      ! see if we have atomic relativistic corrections
      CALL get_qs_env(qs_env, rel_control=rel_control)
      IF (rel_control%rel_method /= rel_none) THEN
         IF (rel_control%rel_transformation == rel_trans_atom) THEN
            nkind = SIZE(atomic_kind_set)
            DO ikind = 1, nkind
               NULLIFY (rtmat)
               CALL calculate_atomic_relkin(atomic_kind_set(ikind), qs_kind_set(ikind), rel_control, rtmat)
               IF (ASSOCIATED(rtmat)) CALL set_qs_kind(qs_kind_set(ikind), reltmat=rtmat)
            END DO
         END IF
      END IF

      CALL qs_subsys_release(subsys)
      CALL qs_ks_release(ks_env)
   END SUBROUTINE qs_init

! **************************************************************************************************
!> \brief Initialize the qs environment (subsys)
!> \param qs_env ...
!> \param para_env ...
!> \param subsys ...
!> \param cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param root_section ...
!> \param subsys_section ...
!> \author Creation (22.05.2000,MK)
! **************************************************************************************************
   SUBROUTINE qs_init_subsys(qs_env, para_env, subsys, cell, cell_ref, use_ref_cell, &
                             root_section, subsys_section)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      LOGICAL, INTENT(in)                                :: use_ref_cell
      TYPE(section_vals_type), POINTER                   :: root_section, subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_init_subsys', routineP = moduleN//':'//routineN

      INTEGER :: handle, ispin, iw, lmax_sphere, maxl, maxlgto, maxlgto_lri, maxlppl, maxlppnl, &
         method_id, multiplicity, my_ival, n_ao, n_ao_aux_fit, n_mo_add, natom, nelectron, &
         output_unit
      INTEGER, DIMENSION(2)                              :: n_mo, nelectron_spin
      LOGICAL                                            :: all_potential_present, do_kpoints, &
                                                            has_unit_metric, was_present
      REAL(dp)                                           :: ewald_rcut, maxocc, verlet_skin
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(dftb_control_type), POINTER                   :: dftb_control
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: se_nonbond_env
      TYPE(gapw_control_type), POINTER                   :: gapw_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(nddo_mpole_type), POINTER                     :: se_nddo_mpole
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_control_type), POINTER                     :: qs_control
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         POINTER                                         :: dftb_potential
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      TYPE(rel_control_type), POINTER                    :: rel_control
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(scptb_control_type), POINTER                  :: scptb_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(section_vals_type), POINTER :: dft_section, et_becke_section, et_coupling_section, &
         et_ddapc_section, ewald_section, nl_section, poisson_section, pp_section, print_section, &
         qs_section, se_section, xc_section
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_si_type), POINTER              :: se_store_int_env

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      was_present = .FALSE.

      ! Initialise the Quickstep environment
      NULLIFY (mos, se_taper, mos_aux_fit)
      NULLIFY (dft_control)
      NULLIFY (energy)
      NULLIFY (force)
      NULLIFY (local_molecules)
      NULLIFY (local_particles)
      NULLIFY (scf_control)
      NULLIFY (dft_section)
      NULLIFY (et_coupling_section)
      NULLIFY (ks_env)
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      qs_section => section_vals_get_subs_vals(dft_section, "QS")
      et_coupling_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%ET_COUPLING")

      CALL qs_subsys_get(subsys, particle_set=particle_set, &
                         qs_kind_set=qs_kind_set, &
                         atomic_kind_set=atomic_kind_set, &
                         molecule_set=molecule_set, &
                         molecule_kind_set=molecule_kind_set)

      !   *** Read the input section with the DFT control parameters ***
      CALL read_dft_control(dft_control, dft_section)

      IF (dft_control%do_tddfpt_calculation) THEN
         CALL read_tddfpt_control(dft_control%tddfpt_control, &
                                  dft_section)
      END IF

      !   *** Print the Quickstep program banner (copyright and version number) ***
      iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%PROGRAM_BANNER", extension=".Log")
      CALL section_vals_val_get(qs_section, "METHOD", i_val=method_id)
      SELECT CASE (method_id)
      CASE DEFAULT
         CALL qs_header(iw)
      CASE (do_method_rm1, do_method_am1, do_method_mndo, do_method_pdg, &
            do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
         CALL se_header(iw)
      CASE (do_method_dftb)
         CALL dftb_header(iw)
      CASE (do_method_scptb)
         CALL scptb_header(iw)
      END SELECT

      CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                        "PRINT%PROGRAM_BANNER")

      !   *** Read the input section with the Quickstep control parameters ***
      CALL read_qs_section(dft_control%qs_control, qs_section)

      !   *******  check if any kind of electron transfer calculation has to be performed
      CALL section_vals_val_get(et_coupling_section, "TYPE_OF_CONSTRAINT", i_val=my_ival)
      dft_control%qs_control%et_coupling_calc = .FALSE.
      IF (my_ival == do_et_ddapc) THEN
         et_ddapc_section => section_vals_get_subs_vals(et_coupling_section, "DDAPC_RESTRAINT_A")
         dft_control%qs_control%et_coupling_calc = .TRUE.
         dft_control%qs_control%ddapc_restraint = .TRUE.
         CALL read_ddapc_section(dft_control%qs_control, ddapc_restraint_section=et_ddapc_section)
      ENDIF

      IF (my_ival == do_et_becke) THEN
         dft_control%qs_control%becke_restraint = .TRUE.
         dft_control%qs_control%et_coupling_calc = .TRUE.
         et_becke_section => section_vals_get_subs_vals(et_coupling_section, "BECKE_RESTRAINT_A")
         CALL read_becke_section(dft_control%qs_control, et_becke_section)
      END IF
      CALL read_mgrid_section(dft_control%qs_control, dft_section)

      !   Create relativistic control section
      CALL rel_c_create(rel_control)
      CALL rel_c_read_parameters(rel_control, dft_section)
      CALL set_qs_env(qs_env, rel_control=rel_control)
      CALL rel_c_release(rel_control)

      !   *** Read DFTB parameter files ***
      IF (dft_control%qs_control%method == "DFTB") THEN
         NULLIFY (ewald_env, ewald_pw, dftb_potential)
         dftb_control => dft_control%qs_control%dftb_control
         CALL qs_dftb_param_init(atomic_kind_set, qs_kind_set, dftb_control, dftb_potential, &
                                 subsys_section=subsys_section, para_env=para_env)
         CALL set_qs_env(qs_env, dftb_potential=dftb_potential)
         ! check for Ewald
         IF (dftb_control%do_ewald) THEN
            CALL ewald_env_create(ewald_env, para_env)
            poisson_section => section_vals_get_subs_vals(dft_section, "POISSON")
            CALL ewald_env_set(ewald_env, poisson_section=poisson_section)
            ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD")
            print_section => section_vals_get_subs_vals(qs_env%input, "PRINT%GRID_INFORMATION")
            CALL read_ewald_section(ewald_env, ewald_section)
            CALL ewald_pw_create(ewald_pw, ewald_env, cell, cell_ref, print_section=print_section)
            CALL set_qs_env(qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
            CALL ewald_env_release(ewald_env)
            CALL ewald_pw_release(ewald_pw)
         END IF
      ELSE IF (dft_control%qs_control%method == "SCPTB") THEN
         scptb_control => dft_control%qs_control%scptb_control
         print_section => section_vals_get_subs_vals(subsys_section, "PRINT")
         CALL scptb_parameter_init(atomic_kind_set, qs_kind_set, scptb_control, print_section, para_env)
         ! check for Ewald
         IF (scptb_control%do_ewald) THEN
            CALL ewald_env_create(ewald_env, para_env)
            poisson_section => section_vals_get_subs_vals(dft_section, "POISSON")
            CALL ewald_env_set(ewald_env, poisson_section=poisson_section)
            ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD")
            print_section => section_vals_get_subs_vals(qs_env%input, "PRINT%GRID_INFORMATION")
            CALL read_ewald_section(ewald_env, ewald_section)
            CALL ewald_pw_create(ewald_pw, ewald_env, cell, cell_ref, print_section=print_section)
            CALL set_qs_env(qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
            CALL ewald_env_release(ewald_env)
            CALL ewald_pw_release(ewald_pw)
         END IF
      END IF

      ! DFT+U
      CALL get_qs_kind_set(qs_kind_set, dft_plus_u_atom_present=dft_control%dft_plus_u)

      !   *** Check basis and fill in missing parts ***
      CALL check_qs_kind_set(qs_kind_set, dft_control, subsys_section=subsys_section)

      !   *** Check that no all-electron potential is present if GPW or GAPW_XC
      CALL get_qs_kind_set(qs_kind_set, all_potential_present=all_potential_present)
      IF ((dft_control%qs_control%method == "GPW") .OR. &
          (dft_control%qs_control%method == "GAPW_XC") .OR. &
          (dft_control%qs_control%method == "OFGPW")) THEN
         IF (all_potential_present) THEN
            CALL cp__b("qs_environment.F",601,"all-el calculations with GPW, GAPW_XC, and OFGPW are not implemented ")
         END IF
      END IF

      !   *** Initialize the spherical harmonics and ***
      !   *** the orbital transformation matrices    ***
      CALL get_qs_kind_set(qs_kind_set, maxlgto=maxlgto, maxlppl=maxlppl, maxlppnl=maxlppnl)

      lmax_sphere = dft_control%qs_control%gapw_control%lmax_sphere
      IF (lmax_sphere .LT. 0) THEN
         lmax_sphere = 2*maxlgto
         dft_control%qs_control%gapw_control%lmax_sphere = lmax_sphere
      END IF
      IF (dft_control%qs_control%method == "LRIGPW" .OR. dft_control%qs_control%lri_optbas) THEN
         CALL get_qs_kind_set(qs_kind_set, maxlgto=maxlgto_lri, basis_type="LRI")
         !take maxlgto from lri basis if larger (usually)
         maxlgto = MAX(maxlgto, maxlgto_lri)
      END IF
      maxl = MAX(2*maxlgto, maxlppl, maxlppnl, lmax_sphere)+1

      CALL init_orbital_pointers(maxl)

      output_unit = cp_print_key_unit_nr(logger, root_section, "GLOBAL%PRINT/SPHERICAL_HARMONICS", &
                                         extension=".Log")
      CALL init_spherical_harmonics(maxl, output_unit)
      CALL cp_print_key_finished_output(output_unit, logger, root_section, &
                                        "GLOBAL%PRINT/SPHERICAL_HARMONICS")

      !   *** Initialise the qs_kind_set ***
      CALL init_qs_kind_set(qs_kind_set)

      !   *** Initialise GAPW soft basis and projectors
      IF (dft_control%qs_control%method == "GAPW" .OR. &
          dft_control%qs_control%method == "GAPW_XC") THEN
         qs_control => dft_control%qs_control
         gapw_control => dft_control%qs_control%gapw_control
         CALL init_gapw_basis_set(qs_kind_set, qs_control, qs_env%input)
      ENDIF

      !   *** Initialize the pretabulation for the calculation of the   ***
      !   *** incomplete Gamma function F_n(t) after McMurchie-Davidson ***
      CALL get_qs_kind_set(qs_kind_set, maxlgto=maxlgto)
      maxl = MAX(3*maxlgto+1, 0)
      CALL init_md_ftable(maxl)

      !   *** Initialize the atomic interaction radii ***
      CALL init_interaction_radii(dft_control%qs_control, atomic_kind_set, qs_kind_set)

      CALL write_pgf_orb_radii("orb", atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_pgf_orb_radii("aux", atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_pgf_orb_radii("lri", atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_core_charge_radii(atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_ppl_radii(atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_ppnl_radii(atomic_kind_set, qs_kind_set, subsys_section)
      CALL write_paw_radii(atomic_kind_set, qs_kind_set, subsys_section)

      !   *** Distribute molecules and atoms using the new data structures ***
      CALL distribute_molecules_1d(atomic_kind_set=atomic_kind_set, &
                                   particle_set=particle_set, &
                                   local_particles=local_particles, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   local_molecules=local_molecules, &
                                   force_env_section=qs_env%input)

      !   *** SCF parameters ***
      CALL scf_c_create(scf_control)
      CALL scf_c_read_parameters(scf_control, dft_section)

      !   *** Allocate the data structure for Quickstep energies ***
      CALL allocate_qs_energy(energy)

      ! check for orthogoanl basis
      has_unit_metric = .FALSE.
      IF (dft_control%qs_control%semi_empirical) THEN
         IF (dft_control%qs_control%se_control%orthogonal_basis) has_unit_metric = .TRUE.
      END IF
      IF (dft_control%qs_control%dftb) THEN
         IF (dft_control%qs_control%dftb_control%orthogonal_basis) has_unit_metric = .TRUE.
      END IF
      CALL set_qs_env(qs_env, has_unit_metric=has_unit_metric)

      !   *** Activate the interpolation ***
      CALL wfi_create(wf_history, &
                      interpolation_method_nr= &
                      dft_control%qs_control%wf_interpolation_method_nr, &
                      extrapolation_order=dft_control%qs_control%wf_extrapolation_order, &
                      has_unit_metric=has_unit_metric)

      !   *** Set the current Quickstep environment ***
      CALL set_qs_env(qs_env=qs_env, &
                      scf_control=scf_control, &
                      wf_history=wf_history)

      CALL qs_subsys_set(subsys, &
                         cell_ref=cell_ref, &
                         use_ref_cell=use_ref_cell, &
                         energy=energy, &
                         force=force)

      CALL get_qs_env(qs_env, ks_env=ks_env)
      CALL set_ks_env(ks_env, dft_control=dft_control)

      CALL qs_subsys_set(subsys, local_molecules_new=local_molecules, &
                         local_particles=local_particles, cell=cell)

      CALL distribution_1d_release(local_particles)
      CALL distribution_1d_release(local_molecules)
      CALL scf_c_release(scf_control)
      CALL wfi_release(wf_history)
      CALL dft_control_release(dft_control)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control, &
                      scf_control=scf_control)

      ! decide what conditions need mo_derivs
      ! right now, this only appears to be OT
      IF (dft_control%qs_control%do_ls_scf .OR. &
          dft_control%qs_control%do_almo_scf) THEN
         CALL set_qs_env(qs_env=qs_env, requires_mo_derivs=.FALSE.)
      ELSE
         IF (scf_control%use_ot) THEN
            CALL set_qs_env(qs_env=qs_env, requires_mo_derivs=.TRUE.)
         ELSE
            CALL set_qs_env(qs_env=qs_env, requires_mo_derivs=.FALSE.)
         ENDIF
      ENDIF

      ! XXXXXXX this is backwards XXXXXXXX
      dft_control%smear = scf_control%smear%do_smear

      !   Initialize the GAPW local densities and potentials
      IF (dft_control%qs_control%method_id == do_method_gapw .OR. &
          dft_control%qs_control%method_id == do_method_gapw_xc) THEN
         !     *** Allocate and initialize the set of atomic densities ***
         CALL init_rho_atom(qs_env, gapw_control)
         IF (dft_control%qs_control%method_id /= do_method_gapw_xc) THEN
            CALL get_qs_env(qs_env=qs_env, natom=natom)
            !       *** Allocate and initialize the compensation density rho0 ***
            CALL init_rho0(qs_env, gapw_control)
            !       *** Allocate and Initialize the local coulomb term ***
            CALL init_coulomb_local(qs_env%hartree_local, natom)
         END IF
      ELSE IF (dft_control%qs_control%method_id == do_method_lrigpw) THEN
         ! allocate local ri environment
      ELSE IF (dft_control%qs_control%semi_empirical) THEN
         NULLIFY (se_store_int_env, se_nddo_mpole, se_nonbond_env)
         natom = SIZE(particle_set)
         se_section => section_vals_get_subs_vals(qs_section, "SE")
         se_control => dft_control%qs_control%se_control

         ! Make the cutoff radii choice a bit smarter
         CALL se_cutoff_compatible(se_control, se_section, cell, output_unit)

         SELECT CASE (dft_control%qs_control%method_id)
         CASE DEFAULT
         CASE (do_method_rm1, do_method_am1, do_method_mndo, do_method_pm3, &
               do_method_pm6, do_method_mndod, do_method_pnnl)
            ! Neighbor lists have to be MAX(interaction range, orbital range)
            ! set new kind radius
            CALL init_se_nlradius(se_control, atomic_kind_set, qs_kind_set, subsys_section)
         END SELECT
         ! Initialize to zero the max multipole to treat in the EWALD scheme..
         se_control%max_multipole = do_multipole_none
         ! check for Ewald
         IF (se_control%do_ewald .OR. se_control%do_ewald_gks) THEN
            CALL ewald_env_create(ewald_env, para_env)
            poisson_section => section_vals_get_subs_vals(dft_section, "POISSON")
            CALL ewald_env_set(ewald_env, poisson_section=poisson_section)
            ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD")
            print_section => section_vals_get_subs_vals(qs_env%input, &
                                                        "PRINT%GRID_INFORMATION")
            CALL read_ewald_section(ewald_env, ewald_section)
            ! Create ewald grids
            CALL ewald_pw_create(ewald_pw, ewald_env, cell, cell_ref, &
                                 print_section=print_section)
            ! Initialize ewald grids
            CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)
            ! Setup the nonbond environment (real space part of Ewald)
            CALL ewald_env_get(ewald_env, rcut=ewald_rcut)
            ! Setup the maximum level of multipoles to be treated in the periodic SE scheme
            IF (se_control%do_ewald) THEN
               CALL ewald_env_get(ewald_env, max_multipole=se_control%max_multipole)
            ENDIF
            CALL section_vals_val_get(se_section, "NEIGHBOR_LISTS%VERLET_SKIN", &
                                      r_val=verlet_skin)
            CALL fist_nonbond_env_create(se_nonbond_env, atomic_kind_set, &
                                         do_nonbonded=.TRUE., verlet_skin=verlet_skin, ewald_rcut=ewald_rcut, &
                                         ei_scale14=0.0_dp, vdw_scale14=0.0_dp, shift_cutoff=.FALSE.)
            ! Create and Setup NDDO multipole environment
            CALL nddo_mpole_setup(se_nddo_mpole, natom)
            CALL set_qs_env(qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw, &
                            se_nonbond_env=se_nonbond_env, se_nddo_mpole=se_nddo_mpole)
            CALL ewald_env_release(ewald_env)
            CALL ewald_pw_release(ewald_pw)
            ! Handle the residual integral part 1/R^3
            CALL semi_empirical_expns3_setup(qs_kind_set, se_control, &
                                             dft_control%qs_control%method_id)
         END IF
         ! Taper function
         CALL se_taper_create(se_taper, se_control%integral_screening, se_control%do_ewald, &
                              se_control%taper_cou, se_control%range_cou, &
                              se_control%taper_exc, se_control%range_exc, &
                              se_control%taper_scr, se_control%range_scr, &
                              se_control%taper_lrc, se_control%range_lrc)
         CALL set_qs_env(qs_env, se_taper=se_taper)
         ! Store integral environment
         CALL semi_empirical_si_create(se_store_int_env, se_section)
         CALL set_qs_env(qs_env, se_store_int_env=se_store_int_env)
      ENDIF

      !   Initialize possible dispersion parameters
      IF (dft_control%qs_control%method_id == do_method_gpw .OR. &
          dft_control%qs_control%method_id == do_method_gapw .OR. &
          dft_control%qs_control%method_id == do_method_gapw_xc .OR. &
          dft_control%qs_control%method_id == do_method_lrigpw .OR. &
          dft_control%qs_control%method_id == do_method_ofgpw) THEN
         ALLOCATE (dispersion_env)
         NULLIFY (xc_section)
         xc_section => section_vals_get_subs_vals(dft_section, "XC")
         CALL qs_dispersion_env_set(dispersion_env, xc_section)
         IF (dispersion_env%type == xc_vdw_fun_pairpot) THEN
            NULLIFY (pp_section)
            pp_section => section_vals_get_subs_vals(xc_section, "VDW_POTENTIAL%PAIR_POTENTIAL")
            CALL qs_dispersion_pairpot_init(atomic_kind_set, qs_kind_set, dispersion_env, pp_section, para_env)
         ELSE IF (dispersion_env%type == xc_vdw_fun_nonloc) THEN
            NULLIFY (nl_section)
            nl_section => section_vals_get_subs_vals(xc_section, "VDW_POTENTIAL%NON_LOCAL")
            CALL qs_dispersion_nonloc_init(dispersion_env, para_env)
         END IF
         CALL set_qs_env(qs_env, dispersion_env=dispersion_env)
      ELSE IF (dft_control%qs_control%method_id == do_method_scptb) THEN
         ALLOCATE (dispersion_env)
         ! set general defaults
         dispersion_env%doabc = .FALSE.
         dispersion_env%c9cnst = .FALSE.
         dispersion_env%lrc = .FALSE.
         dispersion_env%verbose = .FALSE.
         NULLIFY (dispersion_env%c6ab, dispersion_env%maxci, dispersion_env%r0ab, dispersion_env%rcov, &
                  dispersion_env%r2r4, dispersion_env%cn, dispersion_env%cnkind, dispersion_env%cnlist)
         NULLIFY (dispersion_env%q_mesh, dispersion_env%kernel, dispersion_env%d2phi_dk2, &
                  dispersion_env%d2y_dx2, dispersion_env%dftd_section)
         NULLIFY (dispersion_env%sab_vdw, dispersion_env%sab_cn)
         IF (scptb_control%dispersion) THEN
            dispersion_env%type = xc_vdw_fun_pairpot
            dispersion_env%pp_type = vdw_pairpot_dftd3
            dispersion_env%eps_cn = scptb_control%epscn
            dispersion_env%s6 = scptb_control%sd3(1)
            dispersion_env%sr6 = scptb_control%sd3(2)
            dispersion_env%s8 = scptb_control%sd3(3)
            dispersion_env%rc_disp = scptb_control%rcdisp
            dispersion_env%exp_pre = 0._dp
            dispersion_env%scaling = 0._dp
            dispersion_env%parameter_file_name = scptb_control%dispersion_parameter_file
            CALL qs_dispersion_pairpot_init(atomic_kind_set, qs_kind_set, dispersion_env, para_env=para_env)
         ELSE
            dispersion_env%type = xc_vdw_fun_none
         END IF
         CALL set_qs_env(qs_env, dispersion_env=dispersion_env)
      ELSE IF (dft_control%qs_control%method_id == do_method_dftb) THEN
         ALLOCATE (dispersion_env)
         ! set general defaults
         dispersion_env%doabc = .FALSE.
         dispersion_env%c9cnst = .FALSE.
         dispersion_env%lrc = .FALSE.
         dispersion_env%verbose = .FALSE.
         NULLIFY (dispersion_env%c6ab, dispersion_env%maxci, dispersion_env%r0ab, dispersion_env%rcov, &
                  dispersion_env%r2r4, dispersion_env%cn, dispersion_env%cnkind, dispersion_env%cnlist)
         NULLIFY (dispersion_env%q_mesh, dispersion_env%kernel, dispersion_env%d2phi_dk2, &
                  dispersion_env%d2y_dx2, dispersion_env%dftd_section)
         NULLIFY (dispersion_env%sab_vdw, dispersion_env%sab_cn)
         IF (dftb_control%dispersion .AND. dftb_control%dispersion_type == dispersion_d3) THEN
            dispersion_env%type = xc_vdw_fun_pairpot
            dispersion_env%pp_type = vdw_pairpot_dftd3
            dispersion_env%eps_cn = dftb_control%epscn
            dispersion_env%s6 = dftb_control%sd3(1)
            dispersion_env%sr6 = dftb_control%sd3(2)
            dispersion_env%s8 = dftb_control%sd3(3)
            dispersion_env%rc_disp = dftb_control%rcdisp
            dispersion_env%exp_pre = 0._dp
            dispersion_env%scaling = 0._dp
            dispersion_env%parameter_file_name = dftb_control%dispersion_parameter_file
            CALL qs_dispersion_pairpot_init(atomic_kind_set, qs_kind_set, dispersion_env, para_env=para_env)
         ELSE
            dispersion_env%type = xc_vdw_fun_none
         END IF
         CALL set_qs_env(qs_env, dispersion_env=dispersion_env)
      ELSE IF (dft_control%qs_control%semi_empirical) THEN
         ALLOCATE (dispersion_env)
         ! set general defaults
         dispersion_env%doabc = .FALSE.
         dispersion_env%c9cnst = .FALSE.
         dispersion_env%lrc = .FALSE.
         dispersion_env%verbose = .FALSE.
         NULLIFY (dispersion_env%c6ab, dispersion_env%maxci, dispersion_env%r0ab, dispersion_env%rcov, &
                  dispersion_env%r2r4, dispersion_env%cn, dispersion_env%cnkind, dispersion_env%cnlist)
         NULLIFY (dispersion_env%q_mesh, dispersion_env%kernel, dispersion_env%d2phi_dk2, &
                  dispersion_env%d2y_dx2, dispersion_env%dftd_section)
         NULLIFY (dispersion_env%sab_vdw, dispersion_env%sab_cn)
         IF (se_control%dispersion) THEN
            dispersion_env%type = xc_vdw_fun_pairpot
            dispersion_env%pp_type = vdw_pairpot_dftd3
            dispersion_env%eps_cn = se_control%epscn
            dispersion_env%s6 = se_control%sd3(1)
            dispersion_env%sr6 = se_control%sd3(2)
            dispersion_env%s8 = se_control%sd3(3)
            dispersion_env%rc_disp = se_control%rcdisp
            dispersion_env%exp_pre = 0._dp
            dispersion_env%scaling = 0._dp
            dispersion_env%parameter_file_name = se_control%dispersion_parameter_file
            CALL qs_dispersion_pairpot_init(atomic_kind_set, qs_kind_set, dispersion_env, para_env=para_env)
         ELSE
            dispersion_env%type = xc_vdw_fun_none
         END IF
         CALL set_qs_env(qs_env, dispersion_env=dispersion_env)
      END IF

      !   *** Allocate the MO data types ***
      CALL get_qs_kind_set(qs_kind_set, nsgf=n_ao, nelectron=nelectron)

      ! the total number of electrons
      nelectron = nelectron-dft_control%charge

      IF (dft_control%multiplicity == 0) THEN
         IF (MODULO(nelectron, 2) == 0) THEN
            dft_control%multiplicity = 1
         ELSE
            dft_control%multiplicity = 2
         END IF
      END IF

      multiplicity = dft_control%multiplicity

      IF ((dft_control%nspins < 1) .OR. (dft_control%nspins > 2)) THEN
         CALL cp__b("qs_environment.F",937,"nspins should be 1 or 2 for the time being ...")
      END IF

      IF ((MODULO(nelectron, 2) /= 0) .AND. (dft_control%nspins == 1)) THEN
         IF (.NOT. dft_control%qs_control%ofgpw .AND. .NOT. dft_control%smear) THEN
            CALL cp__b("qs_environment.F",942,"Use the LSD option for an odd number of electrons")
         END IF
      END IF

      ! The transition potential method to calculate XAS needs LSD
      IF (dft_control%do_xas_calculation) THEN
         IF (dft_control%nspins == 1) THEN
            CALL cp__b("qs_environment.F",949,"Use the LSD option for XAS with transition potential")
         END IF
      END IF

      !   assigning the number of states per spin initial version, not yet very
      !   general. Should work for an even number of electrons and a single
      !   additional electron this set of options that requires full matrices,
      !   however, makes things a bit ugly right now.... we try to make a
      !   distinction between the number of electrons per spin and the number of
      !   MOs per spin this should allow the use of fractional occupations later
      !   on
      IF (dft_control%qs_control%ofgpw) THEN

         IF (dft_control%nspins == 1) THEN
            maxocc = nelectron
            nelectron_spin(1) = nelectron
            nelectron_spin(2) = 0
            n_mo(1) = 1
            n_mo(2) = 0
         ELSE
            IF (MODULO(nelectron+multiplicity-1, 2) /= 0) THEN
               CALL cp__b("qs_environment.F",970,"LSD: try to use a different multiplicity")
            END IF
            nelectron_spin(1) = (nelectron+multiplicity-1)/2
            nelectron_spin(2) = (nelectron-multiplicity+1)/2
            IF (nelectron_spin(1) < 0) THEN
               CALL cp__b("qs_environment.F",975,"LSD: too few electrons for this multiplicity")
            END IF
            maxocc = MAXVAL(nelectron_spin)
            n_mo(1) = MIN(nelectron_spin(1), 1)
            n_mo(2) = MIN(nelectron_spin(2), 1)
         END IF

      ELSE

         IF (dft_control%nspins == 1) THEN
            maxocc = 2.0_dp
            nelectron_spin(1) = nelectron
            nelectron_spin(2) = 0
            IF (MODULO(nelectron, 2) == 0) THEN
               n_mo(1) = nelectron/2
            ELSE
               n_mo(1) = INT(nelectron/2._dp)+1
            END IF
            n_mo(2) = 0
         ELSE
            maxocc = 1.0_dp

            ! The simplist spin distribution is written here. Special cases will
            ! need additional user input
            IF (MODULO(nelectron+multiplicity-1, 2) /= 0) THEN
               CALL cp__b("qs_environment.F",1000,"LSD: try to use a different multiplicity")
            END IF

            nelectron_spin(1) = (nelectron+multiplicity-1)/2
            nelectron_spin(2) = (nelectron-multiplicity+1)/2

            IF (nelectron_spin(2) < 0) THEN
               CALL cp__b("qs_environment.F",1007,"LSD: too few electrons for this multiplicity")
            END IF

            n_mo(1) = nelectron_spin(1)
            n_mo(2) = nelectron_spin(2)

         END IF

      END IF

      ! store the number of electrons once an for all
      CALL qs_subsys_set(subsys, &
                         nelectron_total=nelectron, &
                         nelectron_spin=nelectron_spin)

      ! Check and set number of added (unoccupied) MOs
      IF (dft_control%nspins == 2) THEN
         IF (scf_control%added_mos(2) > 0) THEN
            n_mo_add = scf_control%added_mos(2)
         ELSE
            n_mo_add = scf_control%added_mos(1)
         END IF
         IF (n_mo_add > n_ao-n_mo(2)) &
            CALL cp__w("qs_environment.F",1030,"More added MOs requested for beta spin than available.")
         scf_control%added_mos(2) = MIN(n_mo_add, n_ao-n_mo(2))
         n_mo(2) = n_mo(2)+scf_control%added_mos(2)
      END IF

      ! proceed alpha orbitals after the beta orbitals; this is essential to avoid
      ! reduction in the number of available unoccupied molecular orbitals.
      ! E.g. n_ao = 10, nelectrons = 10, multiplicity = 3 implies n_mo(1) = 6, n_mo(2) = 4;
      ! added_mos(1:2) = (6,undef) should increase the number of molecular orbitals as
      ! n_mo(1) = min(n_ao, n_mo(1) + added_mos(1)) = 10, n_mo(2) = 10.
      ! However, if we try to proceed alpha orbitals first, this leads us n_mo(1:2) = (10,8)
      ! due to the following assignment instruction above:
      !   IF (scf_control%added_mos(2) > 0) THEN ... ELSE; n_mo_add = scf_control%added_mos(1); END IF
      IF (scf_control%added_mos(1) > n_ao-n_mo(1)) &
         CALL cp_warn(cp__l("qs_environment.F",1044), &
                      "More added MOs requested than available. "// &
                      "The full set of unoccupied MOs will be used.")
      scf_control%added_mos(1) = MIN(scf_control%added_mos(1), n_ao-n_mo(1))
      n_mo(1) = n_mo(1)+scf_control%added_mos(1)

      IF (dft_control%nspins == 2) THEN
         IF (n_mo(2) > n_mo(1)) &
            CALL cp_warn(cp__l("qs_environment.F",1052), &
                         "More beta than alpha MOs requested. "// &
                         "The number of beta MOs will be reduced to the number alpha MOs.")
         n_mo(2) = MIN(n_mo(1), n_mo(2))
         IF(.NOT.(n_mo(1) >= nelectron_spin(1)))CALL cp__a("qs_environment.F",1056)
         IF(.NOT.(n_mo(2) >= nelectron_spin(2)))CALL cp__a("qs_environment.F",1057)
      END IF

      ! kpoints
      CALL get_qs_env(qs_env=qs_env, do_kpoints=do_kpoints)
      IF (do_kpoints .AND. dft_control%nspins == 2) THEN
         ! we need equal number of calculated states
         IF (n_mo(2) /= n_mo(1)) &
            CALL cp_warn(cp__l("qs_environment.F",1065), &
                         "Kpoints: Different number of MOs requested. "// &
                         "The number of beta MOs will be set to the number alpha MOs.")
         n_mo(2) = n_mo(1)
         IF(.NOT.(n_mo(1) >= nelectron_spin(1)))CALL cp__a("qs_environment.F",1069)
         IF(.NOT.(n_mo(2) >= nelectron_spin(2)))CALL cp__a("qs_environment.F",1070)
      END IF

      ! Compatibility checks for smearing
      IF (scf_control%smear%do_smear) THEN
         IF (scf_control%added_mos(1) == 0) THEN
            CALL cp__b("qs_environment.F",1076,"Extra MOs (ADDED_MOS) are required for smearing")
         END IF
      END IF

      !   *** Some options require that all MOs are computed ... ***
      IF (BTEST(cp_print_key_should_output(logger%iter_info, dft_section, &
                                           "PRINT%MO/CARTESIAN"), &
                cp_p_file) .OR. &
          (scf_control%level_shift /= 0.0_dp) .OR. &
          (scf_control%diagonalization%eps_jacobi /= 0.0_dp) .OR. &
          (dft_control%roks .AND. (.NOT. scf_control%use_ot))) THEN
         n_mo(:) = n_ao
      END IF

      ! Compatibility checks for ROKS
      IF (dft_control%roks .AND. (.NOT. scf_control%use_ot)) THEN
         IF (scf_control%roks_scheme == general_roks) &
            CALL cp__w("qs_environment.F",1093,"General ROKS scheme is not yet tested!")

         IF (scf_control%smear%do_smear) THEN
            CALL cp_abort(cp__l("qs_environment.F",1096), &
                          "The options ROKS and SMEAR are not compatible. "// &
                          "Try UKS instead of ROKS")
         END IF
      END IF

      ! in principle the restricted calculation could be performed
      ! using just one set of MOs and special casing most of the code
      ! right now we'll just take care of what is effectively an additional constraint
      ! at as few places as possible, just duplicating the beta orbitals
      IF (dft_control%restricted .AND. (output_unit > 0)) THEN
         ! it is really not yet tested till the end ! Joost
         WRITE (output_unit, *) ""
         WRITE (output_unit, *) " **************************************"
         WRITE (output_unit, *) " restricted calculation cutting corners"
         WRITE (output_unit, *) " experimental feature, check code      "
         WRITE (output_unit, *) " **************************************"
      ENDIF

      ! no point in allocating these things here ?
      IF (dft_control%qs_control%do_ls_scf .OR. &
          dft_control%qs_control%do_almo_scf) THEN
         NULLIFY (mos)
      ELSE
         ALLOCATE (mos(dft_control%nspins))
         DO ispin = 1, dft_control%nspins
            NULLIFY (mos(ispin)%mo_set)
            CALL allocate_mo_set(mo_set=mos(ispin)%mo_set, &
                                 nao=n_ao, &
                                 nmo=n_mo(ispin), &
                                 nelectron=nelectron_spin(ispin), &
                                 n_el_f=REAL(nelectron_spin(ispin), dp), &
                                 maxocc=maxocc, &
                                 flexible_electron_count=dft_control%relax_multiplicity)
         END DO
      END IF

      CALL set_qs_env(qs_env, mos=mos)

      ! If we use auxiliary density matrix methods , set mo_set_aux_fit
      IF (dft_control%do_admm) THEN
         CALL get_qs_kind_set(qs_kind_set, nelectron=nelectron, nsgf=n_ao_aux_fit, &
                              basis_type="AUX_FIT")
         ALLOCATE (mos_aux_fit(dft_control%nspins))

         DO ispin = 1, dft_control%nspins
            NULLIFY (mos_aux_fit(ispin)%mo_set)
            CALL allocate_mo_set(mo_set=mos_aux_fit(ispin)%mo_set, &
                                 nao=n_ao_aux_fit, &
                                 nmo=n_mo(ispin), &
                                 nelectron=nelectron_spin(ispin), &
                                 n_el_f=REAL(nelectron_spin(ispin), dp), &
                                 maxocc=maxocc, &
                                 flexible_electron_count=dft_control%relax_multiplicity)
         END DO
         CALL set_qs_env(qs_env, mos_aux_fit=mos_aux_fit)
      END IF

      ! Print the DFT control parameters
      CALL write_dft_control(dft_control, dft_section)

      ! Print the vdW control parameters
      IF (dft_control%qs_control%method_id == do_method_gpw .OR. &
          dft_control%qs_control%method_id == do_method_gapw .OR. &
          dft_control%qs_control%method_id == do_method_gapw_xc .OR. &
          dft_control%qs_control%method_id == do_method_lrigpw .OR. &
          dft_control%qs_control%method_id == do_method_scptb .OR. &
          dft_control%qs_control%method_id == do_method_dftb .OR. &
          dft_control%qs_control%method_id == do_method_ofgpw) THEN
         CALL get_qs_env(qs_env, dispersion_env=dispersion_env)
         CALL qs_write_dispersion(qs_env, dispersion_env)
      END IF

      ! Print the Quickstep control parameters
      CALL write_qs_control(dft_control%qs_control, dft_section)

      ! Print XES/XAS control parameters
      IF (dft_control%do_xas_calculation) THEN
         CALL cite_reference(Iannuzzi2007)
!       CALL write_xas_control(dft_control%xas_control,dft_section)
      END IF

      ! Print the unnormalized basis set information (input data)
      CALL write_gto_basis_sets(qs_kind_set, subsys_section)

      ! Print the atomic kind set
      CALL write_qs_kind_set(qs_kind_set, subsys_section)

      ! Print the molecule kind set
      CALL write_molecule_kind_set(molecule_kind_set, subsys_section)

      ! Print the total number of kinds, atoms, basis functions etc.
      CALL write_total_numbers(qs_kind_set, particle_set, qs_env%input)

      ! Print the atomic coordinates
      CALL write_qs_particle_coordinates(particle_set, qs_kind_set, subsys_section, label="QUICKSTEP")

      ! Print the interatomic distances
      CALL write_particle_distances(particle_set, cell, subsys_section)

      ! Print the requested structure data
      CALL write_structure_data(particle_set, cell, subsys_section)

      ! Print symmetry information
      CALL write_symmetry(particle_set, cell, subsys_section)

      ! Print the SCF parameters
      IF ((.NOT. dft_control%qs_control%do_ls_scf) .AND. &
          (.NOT. dft_control%qs_control%do_almo_scf)) THEN
         CALL scf_c_write_parameters(scf_control, dft_section)
      ENDIF

      ! Sets up pw_env, qs_charges, mpools ...
      CALL qs_env_setup(qs_env, cp_logger_get_default_io_unit(logger))

      ! Allocate and Initialie rho0 soft on the global grid
      IF (dft_control%qs_control%method == "GAPW") THEN
         CALL get_qs_env(qs_env=qs_env, rho0_mpole=rho0_mpole)
         CALL rho0_s_grid_create(qs_env, rho0_mpole)
      END IF

      IF (output_unit > 0) CALL m_flush(output_unit)
      CALL timestop(handle)

   END SUBROUTINE qs_init_subsys

! **************************************************************************************************
!> \brief Write the total number of kinds, atoms, etc. to the logical unit
!>      number lunit.
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param force_env_section ...
!> \author Creation (06.10.2000)
! **************************************************************************************************
   SUBROUTINE write_total_numbers(qs_kind_set, particle_set, force_env_section)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_total_numbers', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: maxlgto, maxlppl, maxlppnl, natom, ncgf, &
                                                            nkind, npgf, nset, nsgf, nshell, &
                                                            output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, force_env_section, "PRINT%TOTAL_NUMBERS", &
                                         extension=".Log")

      IF (output_unit > 0) THEN
         natom = SIZE(particle_set)
         nkind = SIZE(qs_kind_set)

         CALL get_qs_kind_set(qs_kind_set, &
                              maxlgto=maxlgto, &
                              ncgf=ncgf, &
                              npgf=npgf, &
                              nset=nset, &
                              nsgf=nsgf, &
                              nshell=nshell, &
                              maxlppl=maxlppl, &
                              maxlppnl=maxlppnl)

         WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") &
            "TOTAL NUMBERS AND MAXIMUM NUMBERS"

         IF (nset+npgf+ncgf > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,(T30,A,T71,I10))") &
               "Total number of", &
               "- Atomic kinds:                  ", nkind, &
               "- Atoms:                         ", natom, &
               "- Shell sets:                    ", nset, &
               "- Shells:                        ", nshell, &
               "- Primitive Cartesian functions: ", npgf, &
               "- Cartesian basis functions:     ", ncgf, &
               "- Spherical basis functions:     ", nsgf
         ELSE IF (nshell+nsgf > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,(T30,A,T71,I10))") &
               "Total number of", &
               "- Atomic kinds:                  ", nkind, &
               "- Atoms:                         ", natom, &
               "- Shells:                        ", nshell, &
               "- Spherical basis functions:     ", nsgf
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T3,A,(T30,A,T71,I10))") &
               "Total number of", &
               "- Atomic kinds:                  ", nkind, &
               "- Atoms:                         ", natom
         END IF

         IF ((maxlppl > -1) .AND. (maxlppnl > -1)) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,(T30,A,T75,I6))") &
               "Maximum angular momentum of the", &
               "- Orbital basis functions:                   ", maxlgto, &
               "- Local part of the GTH pseudopotential:     ", maxlppl, &
               "- Non-local part of the GTH pseudopotential: ", maxlppnl
         ELSEIF (maxlppl > -1) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,(T30,A,T75,I6))") &
               "Maximum angular momentum of the", &
               "- Orbital basis functions:                   ", maxlgto, &
               "- Local part of the GTH pseudopotential:     ", maxlppl
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T3,A,T75,I6)") &
               "Maximum angular momentum of the orbital basis functions: ", maxlgto
         END IF

      END IF
      CALL cp_print_key_finished_output(output_unit, logger, force_env_section, &
                                        "PRINT%TOTAL_NUMBERS")

   END SUBROUTINE write_total_numbers

END MODULE qs_environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A simple hash table of integer keys, using hash function:
!>          H(k) = (k*p) mod n + 1
!>        where:
!>          k = key
!>          p = a prime number >= n
!>          n = size of the hash table
!>         And collision resolvation is done by open addressing with linear
!>         probing.
!>
!>         The table consists of an array of (key,val) pairs, and
!>         there are no intermediate buckets. For every new entry (k,v):
!>         We first look up slot H(k), and if it already contains an entry,
!>         then move to the next empty slot using a predefined linear probing
!>         sequence (e.g. iterate from slots H(k) to n, and then 1 to H(k)-1).
!>         When we look up, we use the same probing sequence.
! **************************************************************************************************
MODULE qs_fb_hash_table_types

   USE kinds,                           ONLY: int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! public types
   PUBLIC :: fb_hash_table_obj

! public methods
   PUBLIC :: fb_hash_table_add, &
             fb_hash_table_create, &
             fb_hash_table_get, &
             fb_hash_table_has_data, &
             fb_hash_table_nullify, &
             fb_hash_table_release

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_hash_table_types'
   INTEGER, PRIVATE, SAVE :: last_fb_hash_table_id = 0

! key value indicating an empty slot
   INTEGER(KIND=int_8), PARAMETER, PRIVATE :: EMPTY_KEY = -1_int_8
! Parameters related to automatic resizing of the hash_table:
! Resize by EXPAND_FACTOR if total no. slots / no. of filled slots < ENLARGE_RATIO
   INTEGER, PARAMETER, PRIVATE :: ENLARGE_RATIO = 1
   INTEGER, PARAMETER, PRIVATE :: REDUCE_RATIO = 3
   INTEGER, PARAMETER, PRIVATE :: EXPAND_FACTOR = 2
   INTEGER, PARAMETER, PRIVATE :: SHRINK_FACTOR = 2

! **************************************************************************************************
!> \brief hash table entry data type
!> \param key       : key of the entry
!> \param val       : value of the entry
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_hash_table_element
      INTEGER(KIND=int_8) :: key
      INTEGER :: val
   END TYPE fb_hash_table_element

! **************************************************************************************************
!> \brief data defining a hash table using open addressing for collision
!>        resolvation. Uses simple entry structure to be memory efficient
!>        as well as small overhead
!> \param table     : hash table data area
!> \param nelements : number of non-empty slots in table
!> \param nmax      : max number of slots in table
!> \param prime     : prime number used in the hash function
!> \param id_nr     : unique id for the object
!> \param ref_count : reference counter for the object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_hash_table_data
      INTEGER :: id_nr, ref_count
      TYPE(fb_hash_table_element), DIMENSION(:), POINTER :: table => NULL()
      INTEGER :: nelements
      INTEGER :: nmax
      INTEGER :: prime
   END TYPE fb_hash_table_data

! **************************************************************************************************
!> \brief the object container which allows for the creation of an array
!>        of pointers to fb_hash_table objects
!> \param obj : pointer to the fb_hash_table object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_hash_table_obj
      TYPE(fb_hash_table_data), POINTER, PRIVATE :: obj => NULL()
   END TYPE fb_hash_table_obj

CONTAINS

! **************************************************************************************************
!> \brief Add element to a hash table, auto resize if necessary
!> \param hash_table : the fb_hash_table object
!> \param key        : key of the element
!> \param val        : value of the element
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   RECURSIVE SUBROUTINE fb_hash_table_add(hash_table, key, val)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER(KIND=int_8), INTENT(IN)                    :: key
      INTEGER, INTENT(IN)                                :: val

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: islot
      LOGICAL                                            :: check_ok

      check_ok = fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",117)
      ! check hash table size, if too small rehash in a larger table
      IF (hash_table%obj%nelements*ENLARGE_RATIO .GE. &
          hash_table%obj%nmax) THEN
         CALL fb_hash_table_rehash(hash_table=hash_table, &
                                   nmax=hash_table%obj%nmax*EXPAND_FACTOR)
      END IF
      ! find the right slot for the given key
      islot = fb_hash_table_linear_probe(hash_table, key)
      IF(.NOT.(islot > 0))CALL cp__a("qs_fb_hash_table_types.F",126)
      ! we are adding a new entry only if islot points to an empty slot,
      ! otherwise just change the val of the existing entry
      IF (hash_table%obj%table(islot)%key == EMPTY_KEY) THEN
         hash_table%obj%nelements = hash_table%obj%nelements+1
         hash_table%obj%table(islot)%key = key
      END IF
      hash_table%obj%table(islot)%val = val
   END SUBROUTINE fb_hash_table_add

! **************************************************************************************************
!> \brief Associates one fb_hash_table object to another
!> \param a : the fb_hash_table object to be associated
!> \param b : the fb_hash_table object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_associate(a, b)
      TYPE(fb_hash_table_obj), INTENT(OUT)               :: a
      TYPE(fb_hash_table_obj), INTENT(IN)                :: b

      a%obj => b%obj
   END SUBROUTINE fb_hash_table_associate

! **************************************************************************************************
!> \brief Creates and initialises an empty fb_hash_table object
!> \param hash_table : the fb_hash_table object, its content must be NULL
!>                     and cannot be UNDEFINED
!> \param nmax       : total size of the table, optional. If absent default
!>                     size is 1.
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_create(hash_table, nmax)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER, INTENT(IN), OPTIONAL                      :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_nmax
      LOGICAL                                            :: check_ok

      check_ok = .NOT. fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",168)
      ALLOCATE (hash_table%obj)
      NULLIFY (hash_table%obj%table)
      hash_table%obj%nmax = 0
      hash_table%obj%nelements = 0
      hash_table%obj%prime = 2
      my_nmax = 1
      IF (PRESENT(nmax)) my_nmax = nmax
      CALL fb_hash_table_init(hash_table=hash_table, &
                              nmax=my_nmax)
      ! book keeping stuff
      hash_table%obj%ref_count = 1
      hash_table%obj%id_nr = last_fb_hash_table_id+1
      last_fb_hash_table_id = hash_table%obj%id_nr
   END SUBROUTINE fb_hash_table_create

! **************************************************************************************************
!> \brief Retrieve value from a key from a hash table
!> \param hash_table : the fb_hash_table object
!> \param key        : input key
!> \param val        : output value, equals to 0 if key not found
!> \param found      : .TRUE. if key is found, .FALSE. otherwise
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_get(hash_table, key, val, found)
      TYPE(fb_hash_table_obj), INTENT(IN)                :: hash_table
      INTEGER(KIND=int_8), INTENT(IN)                    :: key
      INTEGER, INTENT(OUT)                               :: val
      LOGICAL, INTENT(OUT)                               :: found

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: islot
      LOGICAL                                            :: check_ok

      check_ok = fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",205)
      found = .FALSE.
      val = 0
      islot = fb_hash_table_linear_probe(hash_table, key)
      IF (islot > 0) THEN
         IF (hash_table%obj%table(islot)%key == key) THEN
            val = hash_table%obj%table(islot)%val
            found = .TRUE.
         END IF
      END IF
   END SUBROUTINE fb_hash_table_get

! **************************************************************************************************
!> \brief check if the object has data associated to it
!> \param hash_table : the fb_hash_table object in question
!> \retval res       : true if hash_table%obj is associated, false otherwise
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_hash_table_has_data(hash_table) RESULT(res)
      TYPE(fb_hash_table_obj), INTENT(IN)                :: hash_table
      LOGICAL                                            :: res

      res = ASSOCIATED(hash_table%obj)
   END FUNCTION fb_hash_table_has_data

! **************************************************************************************************
!> \brief Initialises a fb_hash_table object
!> \param hash_table : the fb_hash_table object, its content must be NULL
!>                     and cannot be UNDEFINED
!> \param nmax       : new size of the table, optional. If absent use the
!>                     old size
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_init(hash_table, nmax)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER, INTENT(IN), OPTIONAL                      :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, my_nmax, power
      LOGICAL                                            :: check_ok

      check_ok = fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",249)
      my_nmax = hash_table%obj%nmax
      IF (PRESENT(nmax)) my_nmax = nmax
      ! table length should always be power of 2. Find the least
      ! power that is greater or equal to my_nmax
      power = 0
      DO WHILE (2**power .LT. my_nmax)
         power = power+1
      END DO
      my_nmax = 2**power
      IF (ASSOCIATED(hash_table%obj%table)) THEN
         IF (SIZE(hash_table%obj%table) .NE. my_nmax) THEN
            DEALLOCATE (hash_table%obj%table)
            ALLOCATE (hash_table%obj%table(my_nmax))
         END IF
      ELSE
         ALLOCATE (hash_table%obj%table(my_nmax))
      END IF
      hash_table%obj%nmax = my_nmax
      hash_table%obj%prime = fb_hash_table_matching_prime(my_nmax)
      ! initiate element to be "empty"
      DO ii = 1, hash_table%obj%nmax
         hash_table%obj%table(ii)%key = EMPTY_KEY
         hash_table%obj%table(ii)%val = 0
      END DO
      hash_table%obj%nelements = 0
   END SUBROUTINE fb_hash_table_init

! **************************************************************************************************
!> \brief Nullifies a fb_hash_table object
!> \param hash_table : the fb_hash_table object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE SUBROUTINE fb_hash_table_nullify(hash_table)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table

      NULLIFY (hash_table%obj)
   END SUBROUTINE fb_hash_table_nullify

! **************************************************************************************************
!> \brief Rehash table. If nmax is present, then also change the table size
!>        to MAX(nmax, number_of_non_empty_elements).
!> \param hash_table      : the fb_hash_table object
!> \param nmax [OPTIONAL] : maximum size of the rehashed table
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_rehash(hash_table, nmax)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER, INTENT(IN), OPTIONAL                      :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_rehash', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, my_nmax
      TYPE(fb_hash_table_element), ALLOCATABLE, &
         DIMENSION(:)                                    :: tmp_table

      IF (.NOT. fb_hash_table_has_data(hash_table)) THEN
         CALL fb_hash_table_create(hash_table, nmax)
         RETURN
      END IF
      IF (PRESENT(nmax)) THEN
         my_nmax = MAX(nmax, hash_table%obj%nelements)
      ELSE
         my_nmax = hash_table%obj%nmax
      END IF
      ALLOCATE (tmp_table(hash_table%obj%nmax))
      tmp_table(:) = hash_table%obj%table(:)
      CALL fb_hash_table_release(hash_table)
      CALL fb_hash_table_create(hash_table=hash_table, &
                                nmax=my_nmax)
      DO ii = 1, SIZE(tmp_table)
         IF (tmp_table(ii)%key .NE. EMPTY_KEY) THEN
            CALL fb_hash_table_add(hash_table=hash_table, &
                                   key=tmp_table(ii)%key, &
                                   val=tmp_table(ii)%val)
         END IF
      END DO
      DEALLOCATE (tmp_table)
   END SUBROUTINE fb_hash_table_rehash

! **************************************************************************************************
!> \brief releases given object
!> \param hash_table : the fb_hash_table object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_release(hash_table)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_release', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      IF (ASSOCIATED(hash_table%obj)) THEN
         check_ok = hash_table%obj%ref_count > 0
         IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",345)
         hash_table%obj%ref_count = hash_table%obj%ref_count-1
         IF (hash_table%obj%ref_count == 0) THEN
            hash_table%obj%ref_count = 1
            IF (ASSOCIATED(hash_table%obj%table)) THEN
               DEALLOCATE (hash_table%obj%table)
            END IF
            hash_table%obj%ref_count = 0
            DEALLOCATE (hash_table%obj)
         END IF
      ELSE
         NULLIFY (hash_table%obj)
      END IF
   END SUBROUTINE fb_hash_table_release

! **************************************************************************************************
!> \brief Remove element from a table, automatic resize if necessary
!> \param hash_table : the fb_hash_table object
!> \param key        : key of the element to be removed
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_remove(hash_table, key)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER(KIND=int_8), INTENT(IN)                    :: key

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_remove', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: islot
      LOGICAL                                            :: check_ok

      check_ok = fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",377)
      islot = fb_hash_table_linear_probe(hash_table, key)
      ! we are only removing an entry if the key is found
      IF (islot > 0) THEN
         IF (hash_table%obj%table(islot)%key == key) THEN
            hash_table%obj%table(islot)%key = EMPTY_KEY
            hash_table%obj%nelements = hash_table%obj%nelements-1
            ! must rehash after setting a filled slot to empty, otherwise the
            ! table will not work. Automatic resize if required
            IF (hash_table%obj%nelements*REDUCE_RATIO .LT. &
                hash_table%obj%nmax) THEN
               CALL fb_hash_table_rehash(hash_table=hash_table, &
                                         nmax=hash_table%obj%nmax/SHRINK_FACTOR)
            ELSE
               CALL fb_hash_table_rehash(hash_table=hash_table)
            END IF
         END IF
      END IF
   END SUBROUTINE fb_hash_table_remove

! **************************************************************************************************
!> \brief retains given object
!> \param hash_table : the fb_hash_table object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_retain(hash_table)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_retain', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      check_ok = ASSOCIATED(hash_table%obj)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",411)
      check_ok = hash_table%obj%ref_count > 0
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",413)
      hash_table%obj%ref_count = hash_table%obj%ref_count+1
   END SUBROUTINE fb_hash_table_retain

! **************************************************************************************************
!> \brief outputs the current information about the table
!> \param hash_table : the fb_hash_table object in question
!> \param nelements  : number of non-empty slots in the table
!> \param nmax       : maximum number of slots in the table
!> \param prime      : the prime used in the hash function
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_hash_table_status(hash_table, nelements, nmax, prime)
      TYPE(fb_hash_table_obj), INTENT(INOUT)             :: hash_table
      INTEGER, INTENT(OUT), OPTIONAL                     :: nelements, nmax, prime

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_hash_table_status', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      check_ok = fb_hash_table_has_data(hash_table)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_hash_table_types.F",435)
      IF (PRESENT(nelements)) nelements = hash_table%obj%nelements
      IF (PRESENT(nmax)) nmax = hash_table%obj%nmax
      IF (PRESENT(prime)) prime = hash_table%obj%prime
   END SUBROUTINE fb_hash_table_status

! **************************************************************************************************
!> \brief Find a prime number equal or larger than ii
!> \param ii   : input integer
!> \retval res : the prime number
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_hash_table_matching_prime(ii) RESULT(res)
      INTEGER, INTENT(IN)                                :: ii
      INTEGER                                            :: res

      res = ii
      DO WHILE (is_positive_prime(res))
         res = res+1
      END DO
   END FUNCTION fb_hash_table_matching_prime

! **************************************************************************************************
!> \brief Check if a number is a positive prime
!> \param num  : number to check
!> \retval res : returns TRUE if num is a positive prime, FALSE otherwise
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION is_positive_prime(num) &
      RESULT(res)
      INTEGER, INTENT(IN)                                :: num
      LOGICAL                                            :: res

      INTEGER                                            :: ii

      IF (num .LE. 3) THEN
         res = num .GE. 2
         RETURN
      END IF
      IF (MOD(num, 2) == 0 .OR. MOD(num, 3) == 0) THEN
         res = .FALSE.
         RETURN
      END IF
      ! all primes > 3 are of the form 6*kk +/- 1, kk=1,2,3...
      ! (although not all 6*kk +/- 1 is a prime);
      ! and we only have to check factors less than and equal to SQRT(num)
      ii = 5
      DO WHILE (ii*ii .LE. num)
         IF (MOD(num, ii) == 0 .OR. MOD(num, ii+2) == 0) THEN
            res = .FALSE.
            RETURN
         END IF
         ii = ii+6
      END DO
      res = .TRUE.
   END FUNCTION is_positive_prime

! **************************************************************************************************
!> \brief Linear probing algorithm for the hash table
!> \param hash_table : the fb_hash_table object
!> \param key        : key to locate
!> \retval islot     : slot location in the table correspond to key, 0 if key
!>                     not found
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_hash_table_linear_probe(hash_table, key) &
      RESULT(islot)
      TYPE(fb_hash_table_obj), INTENT(IN)                :: hash_table
      INTEGER(KIND=int_8), INTENT(IN)                    :: key
      INTEGER                                            :: islot

      INTEGER                                            :: guess

! first guess is mapped by the hash_function

      guess = fb_hash_table_hash_function(hash_table, key)
      ! then search for key and stop at first empty slot from guess to
      ! nmax.  using the same linear probe for adding and retreiving
      ! makes all non-empty keys being put before the first empty slot.
      DO islot = guess, hash_table%obj%nmax
         IF ((hash_table%obj%table(islot)%key == key) .OR. &
             (hash_table%obj%table(islot)%key == EMPTY_KEY)) RETURN
      END DO
      ! if unsuccessful, search from 1 to guess
      DO islot = 1, guess-1
         IF ((hash_table%obj%table(islot)%key == key) .OR. &
             (hash_table%obj%table(islot)%key == EMPTY_KEY)) RETURN
      END DO
      ! if not found and table is full set islot to 0
      islot = 0
   END FUNCTION fb_hash_table_linear_probe

! **************************************************************************************************
!> \brief Hash function
!> \param hash_table : the fb_hash_table object
!> \param key        : key to locate
!> \retval islot     : slot location in the table correspond to key, 0 if key
!>                     not found
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_hash_table_hash_function(hash_table, key) RESULT(hash)
      TYPE(fb_hash_table_obj), INTENT(IN)                :: hash_table
      INTEGER(KIND=int_8), INTENT(IN)                    :: key
      INTEGER                                            :: hash

      INTEGER(KIND=int_8)                                :: hash_8, nmax_8, prime_8

      nmax_8 = INT(hash_table%obj%nmax, int_8)
      prime_8 = INT(hash_table%obj%prime, int_8)
      ! IAND with nmax-1 is equivalent to MOD nmax if nmax is alway a power of 2.
      hash_8 = IAND(key*prime_8, nmax_8-1)+1_int_8
      hash = INT(hash_8)
   END FUNCTION fb_hash_table_hash_function

END MODULE qs_fb_hash_table_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_buffer_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! public types
   PUBLIC :: fb_buffer_i_obj, &
             fb_buffer_d_obj

! public methods
   PUBLIC :: fb_buffer_add, &
             fb_buffer_associate, &
             fb_buffer_create, &
             fb_buffer_calc_disps, &
             fb_buffer_calc_sizes, &
             fb_buffer_get, &
             fb_buffer_has_data, &
             fb_buffer_release, &
             fb_buffer_retain, &
             fb_buffer_nullify, &
             fb_buffer_replace

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_buffer_types'
   INTEGER, PRIVATE, SAVE :: last_fb_buffer_i_id = 0
   INTEGER, PRIVATE, SAVE :: last_fb_buffer_d_id = 0

! **********************************************************************
!> \brief data for the fb_buffer object (integer)
!> \param n : number of data slices in the buffer
!> \param disps : displacement in data array of each slice, it contains
!>                one more element at the end recording the total
!>                size of the current data, which is the same as the
!>                displacement for the new data to be added
!> \param data_1d : where all of the slices are stored
!> \param id_nr : unique id of this object
!> \param ref_count : reference counter of this object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_buffer_i_data
      INTEGER :: id_nr, ref_count
      INTEGER :: n
      INTEGER, DIMENSION(:), POINTER :: disps
      INTEGER, DIMENSION(:), POINTER :: data_1d
   END TYPE fb_buffer_i_data

! **********************************************************************
!> \brief object/pointer wrapper for fb_buffer object
!> \param obj : pointer to fb_buffer data
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_buffer_i_obj
      TYPE(fb_buffer_i_data), POINTER, PRIVATE :: obj => NULL()
   END TYPE fb_buffer_i_obj

! **********************************************************************
!> \brief data for the fb_buffer object (real, double)
!> \param n : number of data slices in the buffer
!> \param disps : displacement in data array of each slice, it contains
!>                one more element at the end recording the total
!>                size of the current data, which is the same as the
!>                displacement for the new data to be added
!> \param data_1d : where all of the slices are stored
!> \param id_nr : unique id of this object
!> \param ref_count : reference counter of this object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_buffer_d_data
      INTEGER :: id_nr, ref_count
      INTEGER :: n
      INTEGER, DIMENSION(:), POINTER :: disps
      REAL(KIND=dp), DIMENSION(:), POINTER :: data_1d
   END TYPE fb_buffer_d_data

! **********************************************************************
!> \brief object/pointer wrapper for fb_buffer object
!> \param obj : pointer to fb_buffer data
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_buffer_d_obj
      TYPE(fb_buffer_d_data), POINTER, PRIVATE :: obj => NULL()
   END TYPE fb_buffer_d_obj

! method overload interfaces
   INTERFACE fb_buffer_add
      MODULE PROCEDURE fb_buffer_i_add
      MODULE PROCEDURE fb_buffer_d_add
   END INTERFACE fb_buffer_add

   INTERFACE fb_buffer_associate
      MODULE PROCEDURE fb_buffer_i_associate
      MODULE PROCEDURE fb_buffer_d_associate
   END INTERFACE fb_buffer_associate

   INTERFACE fb_buffer_create
      MODULE PROCEDURE fb_buffer_i_create
      MODULE PROCEDURE fb_buffer_d_create
   END INTERFACE fb_buffer_create

   INTERFACE fb_buffer_calc_disps
      MODULE PROCEDURE fb_buffer_i_calc_disps
      MODULE PROCEDURE fb_buffer_d_calc_disps
   END INTERFACE fb_buffer_calc_disps

   INTERFACE fb_buffer_calc_sizes
      MODULE PROCEDURE fb_buffer_i_calc_sizes
      MODULE PROCEDURE fb_buffer_d_calc_sizes
   END INTERFACE fb_buffer_calc_sizes

   INTERFACE fb_buffer_get
      MODULE PROCEDURE fb_buffer_i_get
      MODULE PROCEDURE fb_buffer_d_get
   END INTERFACE fb_buffer_get

   INTERFACE fb_buffer_has_data
      MODULE PROCEDURE fb_buffer_i_has_data
      MODULE PROCEDURE fb_buffer_d_has_data
   END INTERFACE fb_buffer_has_data

   INTERFACE fb_buffer_release
      MODULE PROCEDURE fb_buffer_i_release
      MODULE PROCEDURE fb_buffer_d_release
   END INTERFACE fb_buffer_release

   INTERFACE fb_buffer_retain
      MODULE PROCEDURE fb_buffer_i_retain
      MODULE PROCEDURE fb_buffer_d_retain
   END INTERFACE fb_buffer_retain

   INTERFACE fb_buffer_nullify
      MODULE PROCEDURE fb_buffer_i_nullify
      MODULE PROCEDURE fb_buffer_d_nullify
   END INTERFACE fb_buffer_nullify

   INTERFACE fb_buffer_replace
      MODULE PROCEDURE fb_buffer_i_replace
      MODULE PROCEDURE fb_buffer_d_replace
   END INTERFACE fb_buffer_replace

CONTAINS

! INTEGER VERSION

! **************************************************************************************************
!> \brief retains the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_retain(buffer)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(buffer%obj)))CALL cp__a("qs_fb_buffer_types.F",163)
      buffer%obj%ref_count = buffer%obj%ref_count+1
   END SUBROUTINE fb_buffer_i_retain

! **************************************************************************************************
!> \brief releases the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_release(buffer)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(buffer%obj)) THEN
         IF(.NOT.(buffer%obj%ref_count > 0))CALL cp__a("qs_fb_buffer_types.F",179)
         buffer%obj%ref_count = buffer%obj%ref_count-1
         IF (buffer%obj%ref_count == 0) THEN
            buffer%obj%ref_count = 1
            IF (ASSOCIATED(buffer%obj%data_1d)) THEN
               DEALLOCATE (buffer%obj%data_1d)
            END IF
            IF (ASSOCIATED(buffer%obj%disps)) THEN
               DEALLOCATE (buffer%obj%disps)
            END IF
            buffer%obj%ref_count = 0
            DEALLOCATE (buffer%obj)
         END IF
      ELSE
         NULLIFY (buffer%obj)
      END IF
   END SUBROUTINE fb_buffer_i_release

! **************************************************************************************************
!> \brief nullify the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_nullify(buffer)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (buffer%obj)
   END SUBROUTINE fb_buffer_i_nullify

! **************************************************************************************************
!> \brief associate object a to object b
!> \param a : object to associate
!> \param b : object target
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_associate(a, b)
      TYPE(fb_buffer_i_obj), INTENT(OUT)                 :: a
      TYPE(fb_buffer_i_obj), INTENT(IN)                  :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
      CALL fb_buffer_retain(a)
   END SUBROUTINE fb_buffer_i_associate

! **************************************************************************************************
!> \brief check if an object as associated data
!> \param buffer : fb_buffer object
!> \retval res : .TRUE. if buffer has associated data
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_buffer_i_has_data(buffer) RESULT(res)
      TYPE(fb_buffer_i_obj), INTENT(IN)                  :: buffer
      LOGICAL                                            :: res

      res = ASSOCIATED(buffer%obj)
   END FUNCTION fb_buffer_i_has_data

! **************************************************************************************************
!> \brief creates a fb_buffer object
!> \param buffer : fb_buffer object
!> \param max_size : requested total size of the data array
!> \param nslices : total number of slices for the data
!> \param data_1d : the data to be copied to the buffer
!> \param sizes : the size of the slices in the buffer
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_create(buffer, &
                                 max_size, &
                                 nslices, &
                                 data_1d, &
                                 sizes)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer
      INTEGER, INTENT(IN), OPTIONAL                      :: max_size, nslices
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: data_1d, sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_max_size, my_ndata, my_nslices
      LOGICAL                                            :: check_ok

! check optional input

      IF (PRESENT(data_1d)) THEN
         IF(.NOT.(PRESENT(sizes)))CALL cp__a("qs_fb_buffer_types.F",268)
      END IF

      IF(.NOT.(.NOT. ASSOCIATED(buffer%obj)))CALL cp__a("qs_fb_buffer_types.F",271)
      ALLOCATE (buffer%obj)
      ! work out the size of the data array and number of slices
      my_max_size = 0
      my_nslices = 0
      my_ndata = 0
      NULLIFY (buffer%obj%data_1d, &
               buffer%obj%disps)
      ! work out sizes
      IF (PRESENT(max_size)) my_max_size = max_size
      IF (PRESENT(nslices)) my_nslices = nslices
      IF (PRESENT(sizes)) THEN
         my_nslices = MIN(my_nslices, SIZE(sizes))
         my_ndata = SUM(sizes(1:my_nslices))
         my_max_size = MAX(my_max_size, my_ndata)
      END IF
      ! allocate the arrays
      ALLOCATE (buffer%obj%data_1d(my_max_size))
      ALLOCATE (buffer%obj%disps(my_nslices))
      buffer%obj%data_1d = 0
      buffer%obj%disps = 0
      ! set n for buffer before calc disps
      buffer%obj%n = my_nslices
      ! compute disps from sizes if required
      IF (PRESENT(sizes)) THEN
         CALL fb_buffer_calc_disps(buffer, sizes)
      END IF
      ! copy data
      IF (PRESENT(data_1d)) THEN
         check_ok = SIZE(data_1d) .GE. my_max_size .AND. &
                    PRESENT(sizes)
         IF(.NOT.(check_ok))CALL cp__a("qs_fb_buffer_types.F",302)
         buffer%obj%data_1d(1:my_ndata) = data_1d(1:my_ndata)
      END IF
      ! obj meta data update
      buffer%obj%ref_count = 1
      buffer%obj%id_nr = last_fb_buffer_i_id+1
      last_fb_buffer_i_id = buffer%obj%id_nr
   END SUBROUTINE fb_buffer_i_create

! **************************************************************************************************
!> \brief add some data into the buffer
!> \param buffer : fb_buffer object
!> \param data_1d : data to be copied into the object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_add(buffer, data_1d)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer
      INTEGER, DIMENSION(:), INTENT(IN)                  :: data_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: new_data_size, new_n, this_size
      INTEGER, DIMENSION(:), POINTER                     :: new_data, new_disps

      NULLIFY (new_disps, new_data)

      this_size = SIZE(data_1d)
      new_n = buffer%obj%n+1
      new_data_size = buffer%obj%disps(new_n)+this_size
      ! resize when needed
      IF (SIZE(buffer%obj%disps) .LT. new_n+1) THEN
         ALLOCATE (new_disps(new_n*2))
         new_disps = 0
         new_disps(1:buffer%obj%n+1) = buffer%obj%disps(1:buffer%obj%n+1)
         DEALLOCATE (buffer%obj%disps)
         buffer%obj%disps => new_disps
      END IF
      IF (SIZE(buffer%obj%data_1d) .LT. new_data_size) THEN
         ALLOCATE (new_data(new_data_size*2))
         new_data = 0
         new_data(1:buffer%obj%disps(new_n)) = &
            buffer%obj%data_1d(1:buffer%obj%disps(new_n))
         DEALLOCATE (buffer%obj%data_1d)
         buffer%obj%data_1d => new_data
      END IF
      ! append to the buffer
      buffer%obj%disps(new_n+1) = new_data_size
      buffer%obj%data_1d(buffer%obj%disps(new_n)+1:new_data_size) = &
         data_1d(1:this_size)
      buffer%obj%n = new_n
   END SUBROUTINE fb_buffer_i_add

! **************************************************************************************************
!> \brief compute the displacements of each slice in a data buffer from
!>        a given list of sizes of each slice
!> \param buffer : fb_buffer object
!> \param sizes  : list of sizes of each slice on input
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_calc_disps(buffer, sizes)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer
      INTEGER, DIMENSION(:), INTENT(IN)                  :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_calc_disps', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF(.NOT.(SIZE(sizes) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",371)
      buffer%obj%disps(1) = 0
      DO ii = 2, buffer%obj%n+1
         buffer%obj%disps(ii) = buffer%obj%disps(ii-1)+sizes(ii-1)
      END DO
   END SUBROUTINE fb_buffer_i_calc_disps

! **************************************************************************************************
!> \brief compute the sizes of each slice
!> \param buffer : fb_buffer object
!> \param sizes  : list of sizes of each slice on output
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_calc_sizes(buffer, sizes)
      TYPE(fb_buffer_i_obj), INTENT(IN)                  :: buffer
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_calc_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF(.NOT.(SIZE(sizes) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",393)
      DO ii = 1, buffer%obj%n
         sizes(ii) = buffer%obj%disps(ii+1)-buffer%obj%disps(ii)
      END DO
   END SUBROUTINE fb_buffer_i_calc_sizes

! **************************************************************************************************
!> \brief get data from the fb_buffer object
!> \param buffer  : fb_buffer object
!> \param i_slice : see data_1d, data_2d
!> \param n     : outputs number of slices in data array
!> \param data_size : outputs the total size of stored data
!> \param sizes : outputs sizes of the slices in data array
!> \param disps : outputs displacements in the data array for each slice
!> \param data_1d  : if i_slice is present:
!>                      returns pointer to the section of data array corresponding
!>                      to i_slice-th slice
!>                   else:
!>                      return pointer to the entire non-empty part of the data array
!> \param data_2d : similar to data_1d, but with the 1D data array reshaped to 2D
!>                  works only with i_slice present
!> \param data_2d_ld : leading dimension for data_2d for slice i_slice
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_get(buffer, &
                              i_slice, &
                              n, &
                              data_size, &
                              sizes, &
                              disps, &
                              data_1d, &
                              data_2d, &
                              data_2d_ld)
      TYPE(fb_buffer_i_obj), INTENT(IN)                  :: buffer
      INTEGER, INTENT(IN), OPTIONAL                      :: i_slice
      INTEGER, INTENT(OUT), OPTIONAL                     :: n, data_size
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: sizes, disps
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: data_1d
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: data_2d
      INTEGER, INTENT(IN), OPTIONAL                      :: data_2d_ld

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncols, slice_size

      IF (PRESENT(n)) n = buffer%obj%n
      IF (PRESENT(data_size)) data_size = buffer%obj%disps(buffer%obj%n+1)
      IF (PRESENT(sizes)) THEN
         CALL fb_buffer_calc_sizes(buffer, sizes)
      END IF
      IF (PRESENT(disps)) THEN
         IF(.NOT.(SIZE(disps) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",445)
         disps(1:buffer%obj%n) = buffer%obj%disps(1:buffer%obj%n)
      END IF
      IF (PRESENT(data_1d)) THEN
         IF (PRESENT(i_slice)) THEN
            IF(.NOT.(i_slice .LE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",450)
            data_1d => buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                                          buffer%obj%disps(i_slice+1))
         ELSE
            data_1d => buffer%obj%data_1d(1:buffer%obj%disps(buffer%obj%n+1))
         END IF
      END IF
      IF (PRESENT(data_2d)) THEN
         IF(.NOT.(PRESENT(data_2d_ld)))CALL cp__a("qs_fb_buffer_types.F",458)
         IF(.NOT.(PRESENT(i_slice)))CALL cp__a("qs_fb_buffer_types.F",459)
         ! cannot, or rather, it is inefficient to use reshape here, as
         ! a) reshape does not return a targeted array, so cannot
         ! associate pointer unless copied to a targeted array. b) in
         ! F2003 standard, pointers should rank remap automatically by
         ! association to a rank 1 array
         slice_size = buffer%obj%disps(i_slice+1)-buffer%obj%disps(i_slice)
         ncols = slice_size/data_2d_ld
         IF(.NOT.(slice_size == data_2d_ld*ncols))CALL cp__a("qs_fb_buffer_types.F",467)
         data_2d(1:data_2d_ld, 1:ncols) => &
            buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                               buffer%obj%disps(i_slice+1))
      END IF
   END SUBROUTINE fb_buffer_i_get

! **************************************************************************************************
!> \brief replace a slice of the buffer, the replace data size must be
!>        identical to the original slice size
!> \param buffer  : fb_buffer object
!> \param i_slice : the slice index in the buffer
!> \param data_1d : the data to replace the slice
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_i_replace(buffer, i_slice, data_1d)
      TYPE(fb_buffer_i_obj), INTENT(INOUT)               :: buffer
      INTEGER, INTENT(IN)                                :: i_slice
      INTEGER, DIMENSION(:), INTENT(IN)                  :: data_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_i_replace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: slice_size

      slice_size = buffer%obj%disps(i_slice+1)-buffer%obj%disps(i_slice)
      IF(.NOT.(SIZE(data_1d) == slice_size))CALL cp__a("qs_fb_buffer_types.F",493)
      buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                         buffer%obj%disps(i_slice+1)) = data_1d
   END SUBROUTINE fb_buffer_i_replace

! DOUBLE PRECISION VERSION

! **************************************************************************************************
!> \brief retains the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_retain(buffer)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(buffer%obj)))CALL cp__a("qs_fb_buffer_types.F",511)
      buffer%obj%ref_count = buffer%obj%ref_count+1
   END SUBROUTINE fb_buffer_d_retain

! **************************************************************************************************
!> \brief releases the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_release(buffer)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(buffer%obj)) THEN
         IF(.NOT.(buffer%obj%ref_count > 0))CALL cp__a("qs_fb_buffer_types.F",527)
         buffer%obj%ref_count = buffer%obj%ref_count-1
         IF (buffer%obj%ref_count == 0) THEN
            buffer%obj%ref_count = 1
            IF (ASSOCIATED(buffer%obj%data_1d)) THEN
               DEALLOCATE (buffer%obj%data_1d)
            END IF
            IF (ASSOCIATED(buffer%obj%disps)) THEN
               DEALLOCATE (buffer%obj%disps)
            END IF
            buffer%obj%ref_count = 0
            DEALLOCATE (buffer%obj)
         END IF
      ELSE
         NULLIFY (buffer%obj)
      END IF
   END SUBROUTINE fb_buffer_d_release

! **************************************************************************************************
!> \brief nullify the given fb_buffer
!> \param buffer : the fb_bffer object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_nullify(buffer)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (buffer%obj)
   END SUBROUTINE fb_buffer_d_nullify

! **************************************************************************************************
!> \brief associate object a to object b
!> \param a : object to associate
!> \param b : object target
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_associate(a, b)
      TYPE(fb_buffer_d_obj), INTENT(OUT)                 :: a
      TYPE(fb_buffer_d_obj), INTENT(IN)                  :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
      CALL fb_buffer_retain(a)
   END SUBROUTINE fb_buffer_d_associate

! **************************************************************************************************
!> \brief check if an object as associated data
!> \param buffer : fb_buffer object
!> \retval res : .TRUE. if buffer has associated data
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_buffer_d_has_data(buffer) RESULT(res)
      TYPE(fb_buffer_d_obj), INTENT(IN)                  :: buffer
      LOGICAL                                            :: res

      res = ASSOCIATED(buffer%obj)
   END FUNCTION fb_buffer_d_has_data

! **************************************************************************************************
!> \brief creates a fb_buffer object
!> \param buffer : fb_buffer object
!> \param max_size : requested total size of the data array
!> \param nslices : total number of slices for the data
!> \param data_1d : the data to be copied to the buffer
!> \param sizes : the size of the slices in the buffer
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_create(buffer, &
                                 max_size, &
                                 nslices, &
                                 data_1d, &
                                 sizes)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer
      INTEGER, INTENT(IN), OPTIONAL                      :: max_size, nslices
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: data_1d
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_max_size, my_ndata, my_nslices
      LOGICAL                                            :: check_ok

! check optional input

      IF (PRESENT(data_1d)) THEN
         IF(.NOT.(PRESENT(sizes)))CALL cp__a("qs_fb_buffer_types.F",617)
      END IF

      IF(.NOT.(.NOT. ASSOCIATED(buffer%obj)))CALL cp__a("qs_fb_buffer_types.F",620)
      ALLOCATE (buffer%obj)
      ! work out the size of the data array and number of slices
      my_max_size = 0
      my_nslices = 0
      my_ndata = 0
      NULLIFY (buffer%obj%data_1d, &
               buffer%obj%disps)
      ! work out sizes
      IF (PRESENT(max_size)) my_max_size = max_size
      IF (PRESENT(nslices)) my_nslices = nslices
      IF (PRESENT(sizes)) THEN
         my_nslices = MIN(my_nslices, SIZE(sizes))
         my_ndata = SUM(sizes(1:my_nslices))
         my_max_size = MAX(my_max_size, my_ndata)
      END IF
      ! allocate the arrays
      ALLOCATE (buffer%obj%data_1d(my_max_size))
      ALLOCATE (buffer%obj%disps(my_nslices+1))
      buffer%obj%data_1d = 0
      buffer%obj%disps = 0
      ! set n for buffer before calc disps
      buffer%obj%n = my_nslices
      ! compute disps from sizes if required
      IF (PRESENT(sizes)) THEN
         CALL fb_buffer_calc_disps(buffer, sizes)
      END IF
      ! copy data
      IF (PRESENT(data_1d)) THEN
         check_ok = SIZE(data_1d) .GE. my_max_size .AND. &
                    PRESENT(sizes)
         IF(.NOT.(check_ok))CALL cp__a("qs_fb_buffer_types.F",651)
         buffer%obj%data_1d(1:my_ndata) = data_1d(1:my_ndata)
      END IF
      ! obj meta data update
      buffer%obj%ref_count = 1
      buffer%obj%id_nr = last_fb_buffer_d_id+1
      last_fb_buffer_d_id = buffer%obj%id_nr
   END SUBROUTINE fb_buffer_d_create

! **************************************************************************************************
!> \brief add some data into the buffer
!> \param buffer : fb_buffer object
!> \param data_1d : data to be copied into the object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_add(buffer, data_1d)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: data_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: new_data_size, new_n, this_size
      INTEGER, DIMENSION(:), POINTER                     :: new_disps
      REAL(KIND=dp), DIMENSION(:), POINTER               :: new_data

      NULLIFY (new_disps, new_data)

      this_size = SIZE(data_1d)
      new_n = buffer%obj%n+1
      new_data_size = buffer%obj%disps(new_n)+this_size
      ! resize when needed
      IF (SIZE(buffer%obj%disps) .LT. new_n+1) THEN
         ALLOCATE (new_disps(new_n*2))
         new_disps = 0
         new_disps(1:buffer%obj%n+1) = buffer%obj%disps(1:buffer%obj%n+1)
         DEALLOCATE (buffer%obj%disps)
         buffer%obj%disps => new_disps
      END IF
      IF (SIZE(buffer%obj%data_1d) .LT. new_data_size) THEN
         ALLOCATE (new_data(new_data_size*2))
         new_data = 0.0_dp
         new_data(1:buffer%obj%disps(new_n)) = &
            buffer%obj%data_1d(1:buffer%obj%disps(new_n))
         DEALLOCATE (buffer%obj%data_1d)
         buffer%obj%data_1d => new_data
      END IF
      ! append to the buffer
      buffer%obj%disps(new_n+1) = new_data_size
      buffer%obj%data_1d(buffer%obj%disps(new_n)+1:new_data_size) = &
         data_1d(1:this_size)
      buffer%obj%n = new_n
   END SUBROUTINE fb_buffer_d_add

! **************************************************************************************************
!> \brief compute the displacements of each slice in a data buffer from
!>        a given list of sizes of each slice
!> \param buffer : fb_buffer object
!> \param sizes  : list of sizes of each slice on input
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_calc_disps(buffer, sizes)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer
      INTEGER, DIMENSION(:), INTENT(IN)                  :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_calc_disps', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF(.NOT.(SIZE(sizes) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",721)
      buffer%obj%disps(1) = 0
      DO ii = 2, buffer%obj%n+1
         buffer%obj%disps(ii) = buffer%obj%disps(ii-1)+sizes(ii-1)
      END DO
   END SUBROUTINE fb_buffer_d_calc_disps

! **************************************************************************************************
!> \brief compute the sizes of each slice
!> \param buffer : fb_buffer object
!> \param sizes  : list of sizes of each slice on output
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_calc_sizes(buffer, sizes)
      TYPE(fb_buffer_d_obj), INTENT(IN)                  :: buffer
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_calc_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF(.NOT.(SIZE(sizes) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",743)
      DO ii = 1, buffer%obj%n
         sizes(ii) = buffer%obj%disps(ii+1)-buffer%obj%disps(ii)
      END DO
   END SUBROUTINE fb_buffer_d_calc_sizes

! **************************************************************************************************
!> \brief get data from the fb_buffer object
!> \param buffer  : fb_buffer object
!> \param i_slice : see data_1d, data_2d
!> \param n     : outputs number of slices in data array
!> \param data_size : outputs the total size of stored data
!> \param sizes : outputs sizes of the slices in data array
!> \param disps : outputs displacements in the data array for each slice
!> \param data_1d  : if i_slice is present:
!>                      returns pointer to the section of data array corresponding
!>                      to i_slice-th slice
!>                   else:
!>                      return pointer to the entire non-empty part of the data array
!> \param data_2d : similar to data_1d, but with the 1D data array reshaped to 2D
!>                  works only with i_slice present
!> \param data_2d_ld : leading dimension for data_2d for slice i_slice
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_get(buffer, &
                              i_slice, &
                              n, &
                              data_size, &
                              sizes, &
                              disps, &
                              data_1d, &
                              data_2d, &
                              data_2d_ld)
      TYPE(fb_buffer_d_obj), INTENT(IN)                  :: buffer
      INTEGER, INTENT(IN), OPTIONAL                      :: i_slice
      INTEGER, INTENT(OUT), OPTIONAL                     :: n, data_size
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: sizes, disps
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: data_1d
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: data_2d
      INTEGER, INTENT(IN), OPTIONAL                      :: data_2d_ld

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncols, slice_size

      IF (PRESENT(n)) n = buffer%obj%n
      IF (PRESENT(data_size)) data_size = buffer%obj%disps(buffer%obj%n+1)
      IF (PRESENT(sizes)) THEN
         CALL fb_buffer_calc_sizes(buffer, sizes)
      END IF
      IF (PRESENT(disps)) THEN
         IF(.NOT.(SIZE(disps) .GE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",795)
         disps(1:buffer%obj%n) = buffer%obj%disps(1:buffer%obj%n)
      END IF
      IF (PRESENT(data_1d)) THEN
         IF (PRESENT(i_slice)) THEN
            IF(.NOT.(i_slice .LE. buffer%obj%n))CALL cp__a("qs_fb_buffer_types.F",800)
            data_1d => buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                                          buffer%obj%disps(i_slice+1))
         ELSE
            data_1d => buffer%obj%data_1d(1:buffer%obj%disps(buffer%obj%n+1))
         END IF
      END IF
      IF (PRESENT(data_2d)) THEN
         IF(.NOT.(PRESENT(data_2d_ld)))CALL cp__a("qs_fb_buffer_types.F",808)
         IF(.NOT.(PRESENT(i_slice)))CALL cp__a("qs_fb_buffer_types.F",809)
         ! cannot, or rather, it is inefficient to use reshape here, as
         ! a) reshape does not return a targeted array, so cannot
         ! associate pointer unless copied to a targeted array. b) in
         ! F2003 standard, pointers should rank remap automatically by
         ! association to a rank 1 array
         slice_size = buffer%obj%disps(i_slice+1)-buffer%obj%disps(i_slice)
         ncols = slice_size/data_2d_ld
         IF(.NOT.(slice_size == data_2d_ld*ncols))CALL cp__a("qs_fb_buffer_types.F",817)
         data_2d(1:data_2d_ld, 1:ncols) => &
            buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                               buffer%obj%disps(i_slice+1))
      END IF
   END SUBROUTINE fb_buffer_d_get

! **************************************************************************************************
!> \brief replace a slice of the buffer, the replace data size must be
!>        identical to the original slice size
!> \param buffer  : fb_buffer object
!> \param i_slice : the slice index in the buffer
!> \param data_1d : the data to replace the slice
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_buffer_d_replace(buffer, i_slice, data_1d)
      TYPE(fb_buffer_d_obj), INTENT(INOUT)               :: buffer
      INTEGER, INTENT(IN)                                :: i_slice
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: data_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_buffer_d_replace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: slice_size

      slice_size = buffer%obj%disps(i_slice+1)-buffer%obj%disps(i_slice)
      IF(.NOT.(SIZE(data_1d) == slice_size))CALL cp__a("qs_fb_buffer_types.F",843)
      buffer%obj%data_1d(buffer%obj%disps(i_slice)+1: &
                         buffer%obj%disps(i_slice+1)) = data_1d
   END SUBROUTINE fb_buffer_d_replace

END MODULE qs_fb_buffer_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_matrix_data_types

   USE kinds,                           ONLY: dp,&
                                              int_8
   USE qs_fb_buffer_types,              ONLY: fb_buffer_add,&
                                              fb_buffer_create,&
                                              fb_buffer_d_obj,&
                                              fb_buffer_get,&
                                              fb_buffer_has_data,&
                                              fb_buffer_nullify,&
                                              fb_buffer_release,&
                                              fb_buffer_replace
   USE qs_fb_hash_table_types,          ONLY: fb_hash_table_add,&
                                              fb_hash_table_create,&
                                              fb_hash_table_get,&
                                              fb_hash_table_has_data,&
                                              fb_hash_table_nullify,&
                                              fb_hash_table_obj,&
                                              fb_hash_table_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! public types
   PUBLIC :: fb_matrix_data_obj

   ! public methods
   PUBLIC :: fb_matrix_data_add, &
             fb_matrix_data_create, &
             fb_matrix_data_get, &
             fb_matrix_data_has_data, &
             fb_matrix_data_nullify, &
             fb_matrix_data_status, &
             fb_matrix_data_release

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_matrix_data_types'
   INTEGER, PRIVATE, SAVE :: last_fb_matrix_data_id = 0

   ! Parameters related to automatic resizing of matrix_data:
   INTEGER, PARAMETER, PRIVATE :: EXPAND_FACTOR = 2

! **************************************************************************************************
!> \brief data type for storing a list of matrix blocks
!> \param nmax      : maximum number of blocks can be stored
!> \param nblks     : number of blocks currently stored
!> \param nencode   : integer used to encode global block coordinate (row, col)
!>                    into a single combined integer
!> \param ind       : hash table maping the global combined index of the blocks
!>                    to the location in the data area
!> \param blks      : data area, well the matrix elements are actuaally stored
!> \param lds       : leading dimensions of each block
!> \param id_nr     : unique id for the object
!> \param ref_count : reference counter for the object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_matrix_data_data
      INTEGER :: id_nr, ref_count
      INTEGER :: nmax
      INTEGER :: nblks
      INTEGER :: nencode
      TYPE(fb_hash_table_obj) :: ind
      TYPE(fb_buffer_d_obj) :: blks
      INTEGER, DIMENSION(:), POINTER :: lds => NULL()
   END TYPE fb_matrix_data_data

! **************************************************************************************************
!> \brief the object container which allows for the creation of an array
!>        of pointers to fb_matrix_data objects
!> \param obj : pointer to the fb_matrix_data object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_matrix_data_obj
      TYPE(fb_matrix_data_data), POINTER, PRIVATE :: obj => NULL()
   END TYPE fb_matrix_data_obj

CONTAINS

! **************************************************************************************************
!> \brief Add a matrix block to a fb_matrix_data object
!> \param matrix_data : the fb_matrix_data object
!> \param row         : block row index of the matrix block
!> \param col         : block col index of the matrix block
!> \param blk         : the matrix block to add
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_add(matrix_data, row, col, blk)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data
      INTEGER, INTENT(IN)                                :: row, col
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: blk

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: existing_ii, ii, ncols, nrows, old_nblks
      INTEGER(KIND=int_8)                                :: pair_ind
      INTEGER, DIMENSION(:), POINTER                     :: new_lds
      LOGICAL                                            :: check_ok, found

      check_ok = fb_matrix_data_has_data(matrix_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",107)
      NULLIFY (new_lds)
      nrows = SIZE(blk, 1)
      ncols = SIZE(blk, 2)
      ! first check if the block already exists in matrix_data
      pair_ind = fb_matrix_data_encode_pair(row, col, matrix_data%obj%nencode)
      CALL fb_hash_table_get(matrix_data%obj%ind, pair_ind, existing_ii, found)
      IF (found) THEN
         CALL fb_buffer_replace(matrix_data%obj%blks, existing_ii, RESHAPE(blk, (/nrows*ncols/)))
      ELSE
         old_nblks = matrix_data%obj%nblks
         matrix_data%obj%nblks = old_nblks+1
         ii = matrix_data%obj%nblks
         ! resize lds if necessary
         IF (SIZE(matrix_data%obj%lds) .LT. ii) THEN
            ALLOCATE (new_lds(ii*EXPAND_FACTOR))
            new_lds = 0
            new_lds(1:old_nblks) = matrix_data%obj%lds(1:old_nblks)
            DEALLOCATE (matrix_data%obj%lds)
            matrix_data%obj%lds => new_lds
         END IF
         ! add data block
         matrix_data%obj%lds(ii) = nrows
         CALL fb_buffer_add(matrix_data%obj%blks, RESHAPE(blk, (/nrows*ncols/)))
         ! record blk index in the index table
         CALL fb_hash_table_add(matrix_data%obj%ind, pair_ind, ii)
      END IF
   END SUBROUTINE fb_matrix_data_add

! **************************************************************************************************
!> \brief Associates one fb_matrix_data object to another
!> \param a : the fb_matrix_data object to be associated
!> \param b : the fb_matrix_data object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_associate(a, b)
      TYPE(fb_matrix_data_obj), INTENT(OUT)              :: a
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: b

      a%obj => b%obj
   END SUBROUTINE fb_matrix_data_associate

! **************************************************************************************************
!> \brief Creates and initialises an empty fb_matrix_data object of a given size
!> \param matrix_data : the fb_matrix_data object, its content must be NULL
!>                      and cannot be UNDEFINED
!> \param nmax        : max number of matrix blks can be stored
!> \param nencode ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_create(matrix_data, nmax, nencode)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data
      INTEGER, INTENT(IN)                                :: nmax, nencode

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_create', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      check_ok = .NOT. fb_matrix_data_has_data(matrix_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",167)
      ALLOCATE (matrix_data%obj)
      CALL fb_hash_table_nullify(matrix_data%obj%ind)
      CALL fb_buffer_nullify(matrix_data%obj%blks)
      NULLIFY (matrix_data%obj%lds)
      matrix_data%obj%nmax = 0
      matrix_data%obj%nblks = 0
      matrix_data%obj%nencode = nencode
      CALL fb_matrix_data_init(matrix_data=matrix_data, &
                               nmax=nmax, &
                               nencode=nencode)
      ! book keeping stuff
      matrix_data%obj%ref_count = 1
      matrix_data%obj%id_nr = last_fb_matrix_data_id+1
      last_fb_matrix_data_id = matrix_data%obj%id_nr
   END SUBROUTINE fb_matrix_data_create

! **************************************************************************************************
!> \brief retrieve a matrix block from a matrix_data object
!> \param matrix_data : the fb_matrix_data object
!> \param row         : row index
!> \param col         : col index
!> \param blk_p       : pointer to the block in the fb_matrix_data object
!> \param found       : if the requested block exists in the fb_matrix_data
!>                      object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_get(matrix_data, row, col, blk_p, found)
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: matrix_data
      INTEGER, INTENT(IN)                                :: row, col
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: blk_p
      LOGICAL, INTENT(OUT)                               :: found

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ind_in_blks
      INTEGER(KIND=int_8)                                :: pair_ind
      LOGICAL                                            :: check_ok

      check_ok = fb_matrix_data_has_data(matrix_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",208)
      pair_ind = fb_matrix_data_encode_pair(row, col, matrix_data%obj%nencode)
      CALL fb_hash_table_get(matrix_data%obj%ind, pair_ind, ind_in_blks, found)
      IF (found) THEN
         CALL fb_buffer_get(buffer=matrix_data%obj%blks, &
                            i_slice=ind_in_blks, &
                            data_2d=blk_p, &
                            data_2d_ld=matrix_data%obj%lds(ind_in_blks))
      ELSE
         NULLIFY (blk_p)
      END IF
   END SUBROUTINE fb_matrix_data_get

! **************************************************************************************************
!> \brief check if the object has data associated to it
!> \param matrix_data : the fb_matrix_data object in question
!> \retval res        : true if matrix_data%obj is associated, false otherwise
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_matrix_data_has_data(matrix_data) RESULT(res)
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: matrix_data
      LOGICAL                                            :: res

      res = ASSOCIATED(matrix_data%obj)
   END FUNCTION fb_matrix_data_has_data

! **************************************************************************************************
!> \brief Initialises a fb_matrix_data object of a given size
!> \param matrix_data : the fb_matrix_data object, its content must be NULL
!>                      and cannot be UNDEFINED
!> \param nmax        : max number of matrix blocks can be stored, default is
!>                      to use the existing number of blocks in matrix_data
!> \param nencode     : integer used to incode (row, col) to a single combined
!>                      index
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_init(matrix_data, nmax, nencode)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data
      INTEGER, INTENT(IN), OPTIONAL                      :: nmax, nencode

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_nmax
      LOGICAL                                            :: check_ok

      check_ok = fb_matrix_data_has_data(matrix_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",255)
      my_nmax = matrix_data%obj%nmax
      IF (PRESENT(nmax)) my_nmax = nmax
      my_nmax = MAX(my_nmax, 1)
      IF (fb_hash_table_has_data(matrix_data%obj%ind)) THEN
         CALL fb_hash_table_release(matrix_data%obj%ind)
      END IF
      CALL fb_hash_table_create(matrix_data%obj%ind, my_nmax)
      IF (fb_buffer_has_data(matrix_data%obj%blks)) THEN
         CALL fb_buffer_release(matrix_data%obj%blks)
      END IF
      CALL fb_buffer_create(buffer=matrix_data%obj%blks)
      IF (ASSOCIATED(matrix_data%obj%lds)) THEN
         DEALLOCATE (matrix_data%obj%lds)
      END IF
      ALLOCATE (matrix_data%obj%lds(0))
      matrix_data%obj%nblks = 0
      IF (PRESENT(nencode)) matrix_data%obj%nencode = nencode
   END SUBROUTINE fb_matrix_data_init

! **************************************************************************************************
!> \brief Nullifies a fb_matrix_data object
!> \param matrix_data : the fb_matrix_data object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE SUBROUTINE fb_matrix_data_nullify(matrix_data)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data

      NULLIFY (matrix_data%obj)
   END SUBROUTINE fb_matrix_data_nullify

! **************************************************************************************************
!> \brief releases given object
!> \param matrix_data : the fb_matrix_data object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_release(matrix_data)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_release', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      IF (ASSOCIATED(matrix_data%obj)) THEN
         check_ok = matrix_data%obj%ref_count > 0
         IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",301)
         matrix_data%obj%ref_count = matrix_data%obj%ref_count-1
         IF (matrix_data%obj%ref_count == 0) THEN
            matrix_data%obj%ref_count = 1
            IF (fb_hash_table_has_data(matrix_data%obj%ind)) THEN
               CALL fb_hash_table_release(matrix_data%obj%ind)
            END IF
            IF (fb_buffer_has_data(matrix_data%obj%blks)) THEN
               CALL fb_buffer_release(matrix_data%obj%blks)
            END IF
            IF (ASSOCIATED(matrix_data%obj%lds)) THEN
               DEALLOCATE (matrix_data%obj%lds)
            END IF
            matrix_data%obj%ref_count = 0
            DEALLOCATE (matrix_data%obj)
         END IF
      ELSE
         NULLIFY (matrix_data%obj)
      END IF
   END SUBROUTINE fb_matrix_data_release

! **************************************************************************************************
!> \brief retains given object
!> \param matrix_data : the fb_matrix_data object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_retain(matrix_data)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_retain', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      check_ok = ASSOCIATED(matrix_data%obj)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",336)
      check_ok = matrix_data%obj%ref_count > 0
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",338)
      matrix_data%obj%ref_count = matrix_data%obj%ref_count+1
   END SUBROUTINE fb_matrix_data_retain

! **************************************************************************************************
!> \brief outputs the current information about fb_matrix_data object
!> \param matrix_data : the fb_matrix_data object
!> \param nmax        : outputs fb_matrix_data%obj%nmax
!> \param nblks       : outputs fb_matrix_data%obj%nblks
!> \param nencode     : outputs fb_matrix_data%obj%nencode
!> \param blk_sizes   : blk_sizes(ii,jj) gives size of jj-th dim of the
!>                      ii-th block stored
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_matrix_data_status(matrix_data, nmax, nblks, nencode, blk_sizes)
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_data
      INTEGER, INTENT(OUT), OPTIONAL                     :: nmax, nblks, nencode
      INTEGER, DIMENSION(:, :), INTENT(OUT), OPTIONAL    :: blk_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_matrix_data_status', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: buffer_sizes
      LOGICAL                                            :: check_ok

      check_ok = fb_matrix_data_has_data(matrix_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",365)
      IF (PRESENT(nmax)) nmax = matrix_data%obj%nmax
      IF (PRESENT(nblks)) nblks = matrix_data%obj%nblks
      IF (PRESENT(nencode)) nencode = matrix_data%obj%nencode
      IF (PRESENT(blk_sizes)) THEN
         check_ok = (SIZE(blk_sizes, 1) .GE. matrix_data%obj%nblks .AND. &
                     SIZE(blk_sizes, 2) .GE. 2)
         IF(.NOT.(check_ok))CALL cp__a("qs_fb_matrix_data_types.F",372)
         blk_sizes(:, :) = 0
         ALLOCATE (buffer_sizes(matrix_data%obj%nblks))
         CALL fb_buffer_get(buffer=matrix_data%obj%blks, &
                            sizes=buffer_sizes)
         DO ii = 1, matrix_data%obj%nblks
            blk_sizes(ii, 1) = matrix_data%obj%lds(ii)
            blk_sizes(ii, 2) = buffer_sizes(ii)/matrix_data%obj%lds(ii)
         END DO
         DEALLOCATE (buffer_sizes)
      END IF
   END SUBROUTINE fb_matrix_data_status

! **************************************************************************************************
!> \brief Encodes (row, col) index pair into a single combined index
!> \param row     : row index (assume to start counting from 1)
!> \param col     : col index (assume to start counting from 1)
!> \param nencode : integer used for encoding
!> \retval pair_ind : the returned value
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_matrix_data_encode_pair(row, col, nencode) &
      RESULT(pair_ind)
      INTEGER, INTENT(IN)                                :: row, col, nencode
      INTEGER(KIND=int_8)                                :: pair_ind

      INTEGER(KIND=int_8)                                :: col_8, nencode_8, row_8

      row_8 = INT(row, int_8)
      col_8 = INT(col, int_8)
      nencode_8 = INT(nencode, int_8)
      pair_ind = (row_8-1_int_8)*nencode_8+(col_8-1_int_8)+1
   END FUNCTION fb_matrix_data_encode_pair

END MODULE qs_fb_matrix_data_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_com_tasks_types

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_col_block_sizes,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_put_block,&
                                              cp_dbcsr_row_block_sizes,&
                                              cp_dbcsr_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_alltoall
   USE qs_fb_matrix_data_types,         ONLY: fb_matrix_data_add,&
                                              fb_matrix_data_get,&
                                              fb_matrix_data_has_data,&
                                              fb_matrix_data_obj
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! public parameters:
   PUBLIC :: TASK_N_RECORDS, &
             TASK_DEST, &
             TASK_SRC, &
             TASK_PAIR, &
             TASK_COST

! public types
   PUBLIC :: fb_com_tasks_obj, &
             fb_com_atom_pairs_obj

! public methods
   PUBLIC :: fb_com_tasks_release, &
             fb_com_tasks_nullify, &
             fb_com_tasks_create, &
             fb_com_tasks_get, &
             fb_com_tasks_set, &
             fb_com_tasks_transpose_dest_src, &
             fb_com_tasks_build_atom_pairs, &
             fb_com_tasks_encode_pair, &
             fb_com_tasks_decode_pair, &
             fb_com_atom_pairs_release, &
             fb_com_atom_pairs_nullify, &
             fb_com_atom_pairs_has_data, &
             fb_com_atom_pairs_create, &
             fb_com_atom_pairs_init, &
             fb_com_atom_pairs_get, &
             fb_com_atom_pairs_decode, &
             fb_com_atom_pairs_calc_buffer_sizes, &
             fb_com_atom_pairs_gather_blks, &
             fb_com_atom_pairs_distribute_blks

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_com_tasks_types'
   INTEGER, PRIVATE, SAVE :: last_fb_com_tasks_id = 0
   INTEGER, PRIVATE, SAVE :: last_fb_com_atom_pairs_id = 0

! **********************************************************************
! explaination on format of task lists (same for tasks_recv and tasks_send):
! tasks_recv has dimension (4, ntasks_recv), and stores information on
! the block to be copied or transfered
! - tasks_recv(TASK_DEST,itask) = destination MPI rank of itask-th task
! - tasks_recv(TASK_SRC,itask) = source MPI rank of itask-th task
! - tasks_recv(TASK_PAIR,itask) = compressed pair indices of the block of itask-th task
! - tasks_recv(TASK_COST,itask) = the cost of itask-th task
!
! number of record slots in each task in the task lists
   INTEGER, PARAMETER :: TASK_N_RECORDS = 4
! the indices for the records (1:TASK_DIM) in a task
   INTEGER, PARAMETER :: TASK_DEST = 1, &
                         TASK_SRC = 2, &
                         TASK_PAIR = 3, &
                         TASK_COST = 4
! **********************************************************************

! **********************************************************************
!> \brief data content for communication tasks used for send and receive
!>        matrix blocks
!> \param id_nr     : unique id for the object
!> \param ref_count : reference count on the object
!> \param tasks     : the list of communication tasks, which is
!>                    represented by a 2D array, first dim stores
!>                    info for the communication: src and desc procs
!>                    and the atomic pair indexing the matrix block
!>                    to be communicated, etc.
!> \param task_dim  : the size of the first dimension of tasks
!> \param ntasks    : total number of local tasks
!> \param nencode   : the total number of atoms used for encoding
!>                    the block coordinates (iatom, jatom)
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_com_tasks_data
      INTEGER :: id_nr, ref_count
      ! use pure integer arrays to facilitate easier MPI coms
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER :: tasks
      INTEGER :: task_dim
      INTEGER :: ntasks
      INTEGER :: nencode
   END TYPE fb_com_tasks_data

!**********************************************************************
!> \brief defines a fb_com_tasks object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
!**********************************************************************
   TYPE fb_com_tasks_obj
      TYPE(fb_com_tasks_data), POINTER, PRIVATE :: obj
   END TYPE fb_com_tasks_obj

! **********************************************************************
!> \brief data content for the list of block coordinates with the
!>        associated src/dest proc id for communication. These will be
!>        generated from the fb_com_tasks object
!> \param id_nr         : unique id for the object
!> \param ref_count     : reference count on the object
!> \param pairs         : the list of communication tasks, which is
!>                        represented by a 2D array, first dim stores
!>                        info for the communication: src and desc procs
!>                        and the atomic pair indexing the matrix block
!>                        to be communicated, etc.
!> \param npairs        : number of blks to be communicated in the atom
!>                        pair list
!> \param natoms_encode : the total number of atoms used for encoding
!>                        the proc + block coordinates (pe, iatom, jatom)
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_com_atom_pairs_data
      INTEGER :: id_nr, ref_count
      INTEGER(KIND=int_8), DIMENSION(:), POINTER :: pairs
      INTEGER :: npairs
      INTEGER :: natoms_encode
   END TYPE fb_com_atom_pairs_data

! **********************************************************************
!> \brief defines a fb_com_atom_pairs object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_com_atom_pairs_obj
      TYPE(fb_com_atom_pairs_data), POINTER, PRIVATE :: obj
   END TYPE fb_com_atom_pairs_obj

CONTAINS

! **********************************************************************
!> \brief Retains an fb_com_tasks object
!> \param com_tasks the fb_com_tasks object, its content must not be
!>                   NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_retain(com_tasks)
      TYPE(fb_com_tasks_obj), INTENT(IN)                 :: com_tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(com_tasks%obj)))CALL cp__a("qs_fb_com_tasks_types.F",163)
      IF(.NOT.(com_tasks%obj%ref_count > 0))CALL cp__a("qs_fb_com_tasks_types.F",164)
      com_tasks%obj%ref_count = com_tasks%obj%ref_count+1
   END SUBROUTINE fb_com_tasks_retain

! **********************************************************************
!> \brief Retains an fb_com_atom_pairs object
!> \param atom_pairs the fb_com_atom_pairs object, its content must not be
!>                   NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_retain(atom_pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom_pairs%obj)))CALL cp__a("qs_fb_com_tasks_types.F",180)
      IF(.NOT.(atom_pairs%obj%ref_count > 0))CALL cp__a("qs_fb_com_tasks_types.F",181)
      atom_pairs%obj%ref_count = atom_pairs%obj%ref_count+1
   END SUBROUTINE fb_com_atom_pairs_retain

! **********************************************************************
!> \brief Releases an fb_com_tasks object
!> \param com_tasks the fb_com_tasks object, its content must not be
!>                   UNDEFINED, and the subroutine does nothing if the
!>                   content points to NULL
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_release(com_tasks)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: com_tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(com_tasks%obj)) THEN
         IF(.NOT.(com_tasks%obj%ref_count > 0))CALL cp__a("qs_fb_com_tasks_types.F",199)
         com_tasks%obj%ref_count = com_tasks%obj%ref_count-1
         IF (com_tasks%obj%ref_count == 0) THEN
            com_tasks%obj%ref_count = 1
            IF (ASSOCIATED(com_tasks%obj%tasks)) THEN
               DEALLOCATE (com_tasks%obj%tasks)
            END IF
            com_tasks%obj%ref_count = 0
            DEALLOCATE (com_tasks%obj)
         END IF
      ELSE
         NULLIFY (com_tasks%obj)
      END IF
   END SUBROUTINE fb_com_tasks_release

! **********************************************************************
!> \brief Releases an fb_com_atom_pairs object
!> \param atom_pairs the fb_com_atom_pairs object, its content must not
!>                    be UNDEFINED, and the subroutine does nothing if
!>                    the content points to NULL
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_release(atom_pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(atom_pairs%obj)) THEN
         IF(.NOT.(atom_pairs%obj%ref_count > 0))CALL cp__a("qs_fb_com_tasks_types.F",228)
         atom_pairs%obj%ref_count = atom_pairs%obj%ref_count-1
         IF (atom_pairs%obj%ref_count == 0) THEN
            atom_pairs%obj%ref_count = 1
            IF (ASSOCIATED(atom_pairs%obj%pairs)) THEN
               DEALLOCATE (atom_pairs%obj%pairs)
            END IF
            atom_pairs%obj%ref_count = 0
            DEALLOCATE (atom_pairs%obj)
         END IF
      ELSE
         NULLIFY (atom_pairs%obj)
      END IF
   END SUBROUTINE fb_com_atom_pairs_release

! **********************************************************************
!> \brief Nullifies a fb_com_tasks object, note that it does not release
!>        the original object. This procedure is used to nullify the
!>        pointer contained in the object which is used to associate to
!>        the actual object content
!> \param com_tasks the com_tasks object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_nullify(com_tasks)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: com_tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (com_tasks%obj)
   END SUBROUTINE fb_com_tasks_nullify

! **********************************************************************
!> \brief Nullifies a fb_com_atom_pairs object, note that it does not
!>        release the original object. This procedure is used to nullify
!>        the pointer contained in the object which is used to associate
!>        to the actual object content
!> \param atom_pairs the fb_com_atom_pairs object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_nullify(atom_pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (atom_pairs%obj)
   END SUBROUTINE fb_com_atom_pairs_nullify

! **********************************************************************
!> \brief Associates one fb_com_tasks object to another
!> \param a the fb_com_tasks object to be associated
!> \param b the fb_com_tasks object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_associate(a, b)
      TYPE(fb_com_tasks_obj), INTENT(OUT)                :: a
      TYPE(fb_com_tasks_obj), INTENT(IN)                 :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_com_tasks_associate

! **********************************************************************
!> \brief Associates one fb_com_atom_pairs object to another
!> \param a the fb_com_atom_pairs object to be associated
!> \param b the fb_com_atom_pairs object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_associate(a, b)
      TYPE(fb_com_atom_pairs_obj), INTENT(OUT)           :: a
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_com_atom_pairs_associate

! **********************************************************************
!> \brief Checks if a fb_com_tasks object is associated with an actual
!>        data content or not
!> \param com_tasks the fb_com_tasks object
!> \retval res ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_com_tasks_has_data(com_tasks) RESULT(res)
      TYPE(fb_com_tasks_obj), INTENT(IN)                 :: com_tasks
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(com_tasks%obj)
   END FUNCTION fb_com_tasks_has_data

! **********************************************************************
!> \brief Checks if a fb_com_atom_pairs object is associated with an actual
!>        data content or not
!> \param atom_pairs the fb_com_atom_pairs object
!> \retval res ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_com_atom_pairs_has_data(atom_pairs) RESULT(res)
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(atom_pairs%obj)
   END FUNCTION fb_com_atom_pairs_has_data

! **********************************************************************
!> \brief Creates and initialises an empty fb_com_tasks object
!> \param com_tasks the fb_com_tasks object, its content must be NULL
!>                   and cannot be UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_create(com_tasks)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: com_tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(com_tasks%obj)))CALL cp__a("qs_fb_com_tasks_types.F",355)
      ALLOCATE (com_tasks%obj)
      com_tasks%obj%task_dim = TASK_N_RECORDS
      com_tasks%obj%ntasks = 0
      com_tasks%obj%nencode = 0
      NULLIFY (com_tasks%obj%tasks)
      com_tasks%obj%ref_count = 1
      com_tasks%obj%id_nr = last_fb_com_tasks_id+1
      last_fb_com_tasks_id = com_tasks%obj%id_nr
   END SUBROUTINE fb_com_tasks_create

! **********************************************************************
!> \brief Creates and initialises an empty fb_com_atom_pairs object
!> \param atom_pairs the fb_com_atom_pairs object, its content must be
!>                    NULL and cannot be UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_create(atom_pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(atom_pairs%obj)))CALL cp__a("qs_fb_com_tasks_types.F",378)
      ALLOCATE (atom_pairs%obj)
      atom_pairs%obj%npairs = 0
      atom_pairs%obj%natoms_encode = 0
      NULLIFY (atom_pairs%obj%pairs)
      atom_pairs%obj%ref_count = 1
      atom_pairs%obj%id_nr = last_fb_com_atom_pairs_id+1
      last_fb_com_atom_pairs_id = atom_pairs%obj%id_nr
   END SUBROUTINE fb_com_atom_pairs_create

! **********************************************************************
!> \brief Initialises an fb_com_tasks object, and makes it empty
!> \param com_tasks the fb_com_tasks object, its content must not be
!>                   NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_init(com_tasks)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: com_tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_init', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(com_tasks%obj)))CALL cp__a("qs_fb_com_tasks_types.F",400)
      IF (ASSOCIATED(com_tasks%obj%tasks)) THEN
         DEALLOCATE (com_tasks%obj%tasks)
      END IF
      com_tasks%obj%task_dim = TASK_N_RECORDS
      com_tasks%obj%ntasks = 0
      com_tasks%obj%nencode = 0
   END SUBROUTINE fb_com_tasks_init

! **********************************************************************
!> \brief Initialises an fb_com_atom_pairs object, and makes it empty
!> \param atom_pairs the fb_com_atom_pairs object, its content must not
!>                    be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_init(atom_pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_init', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom_pairs%obj)))CALL cp__a("qs_fb_com_tasks_types.F",421)
      IF (ASSOCIATED(atom_pairs%obj%pairs)) THEN
         DEALLOCATE (atom_pairs%obj%pairs)
      END IF
      atom_pairs%obj%npairs = 0
      atom_pairs%obj%natoms_encode = 0
   END SUBROUTINE fb_com_atom_pairs_init

! **********************************************************************
!> \brief Gets attributes from a fb_com_tasks object, one should only
!>        access the data content in a fb_com_tasks object outside this
!>        module via this procedure.
!> \param com_tasks the fb_com_tasks object, its content must not be
!>                   NULL or UNDEFINED
!> \param task_dim [OPTIONAL]: if present, outputs com_tasks%obj%task_dim
!> \param ntasks [OPTIONAL]: if present, outputs com_tasks%obj%ntasks
!> \param nencode [OPTIONAL]: if present, outputs com_tasks%obj%nencode
!> \param tasks [OPTIONAL]: if present, outputs pointer com_tasks%obj%tasks
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_get(com_tasks, &
                               task_dim, &
                               ntasks, &
                               nencode, &
                               tasks)
      TYPE(fb_com_tasks_obj), INTENT(IN)                 :: com_tasks
      INTEGER, INTENT(OUT), OPTIONAL                     :: task_dim, ntasks, nencode
      INTEGER(KIND=int_8), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(com_tasks%obj)))CALL cp__a("qs_fb_com_tasks_types.F",454)
      IF (PRESENT(task_dim)) task_dim = com_tasks%obj%task_dim
      IF (PRESENT(ntasks)) ntasks = com_tasks%obj%ntasks
      IF (PRESENT(nencode)) nencode = com_tasks%obj%nencode
      IF (PRESENT(tasks)) tasks => com_tasks%obj%tasks
   END SUBROUTINE fb_com_tasks_get

! **********************************************************************
!> \brief Gets attributes from a fb_com_atom_pairs object, one should
!>        only access the data content in a fb_com_atom_pairs object
!>        outside this module via this procedure.
!> \param atom_pairs the fb_com_atom_pairs object, its content must not
!>                    be NULL or UNDEFINED
!> \param npairs [OPTIONAL]: if present, outputs atom_pairs%obj%npairs
!> \param natoms_encode [OPTIONAL]: if present, outputs atom_pairs%obj%natoms_encode
!> \param pairs [OPTIONAL]: if present, outputs pointer atom_pairs%obj%pairs
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_get(atom_pairs, &
                                    npairs, &
                                    natoms_encode, &
                                    pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs
      INTEGER, INTENT(OUT), OPTIONAL                     :: npairs, natoms_encode
      INTEGER(KIND=int_8), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom_pairs%obj)))CALL cp__a("qs_fb_com_tasks_types.F",484)
      IF (PRESENT(npairs)) npairs = atom_pairs%obj%npairs
      IF (PRESENT(natoms_encode)) natoms_encode = atom_pairs%obj%natoms_encode
      IF (PRESENT(pairs)) pairs => atom_pairs%obj%pairs
   END SUBROUTINE fb_com_atom_pairs_get

! **********************************************************************
!> \brief Sets attributes in a fb_com_tasks object, one should only
!>        access the data content in a fb_com_tasks object outside this
!>        module via this procedure.
!> \param com_tasks the fb_com_tasks object, its content must not be
!>                   NULL or UNDEFINED
!> \param task_dim [OPTIONAL]: if present, sets com_tasks%obj%task_dim
!> \param ntasks [OPTIONAL]: if present, sets com_tasks%obj%ntasks
!> \param nencode [OPTIONAL]: if present, sets com_tasks%obj%nencode
!> \param tasks [OPTIONAL]: if present, associates pointer com_tasks%obj%tasks
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_set(com_tasks, &
                               task_dim, &
                               ntasks, &
                               nencode, &
                               tasks)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: com_tasks
      INTEGER, INTENT(IN), OPTIONAL                      :: task_dim, ntasks, nencode
      INTEGER(KIND=int_8), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: tasks

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_tasks_set', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(com_tasks%obj)))CALL cp__a("qs_fb_com_tasks_types.F",515)
      IF (PRESENT(task_dim)) com_tasks%obj%task_dim = task_dim
      IF (PRESENT(ntasks)) com_tasks%obj%ntasks = ntasks
      IF (PRESENT(nencode)) com_tasks%obj%nencode = nencode
      IF (PRESENT(tasks)) THEN
         IF (ASSOCIATED(com_tasks%obj%tasks)) THEN
            DEALLOCATE (com_tasks%obj%tasks)
         END IF
         com_tasks%obj%tasks => tasks
      END IF
   END SUBROUTINE fb_com_tasks_set

! **********************************************************************
!> \brief Sets attributes in a fb_com_atom_pairs object, one should only
!>        access the data content in a fb_com_atom_pairs object outside
!>        this module via this procedure.
!> \param atom_pairs the fb_com_atom_pairs object, its content must not
!>                    be NULL or UNDEFINED
!> \param npairs [OPTIONAL]: if present, sets atom_pairs%obj%npairs
!> \param natoms_encode [OPTIONAL]: if present, sets atom_pairs%obj%natoms_encode
!> \param pairs [OPTIONAL]: if present, associates pointer atom_pairs%obj%pairs
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_set(atom_pairs, &
                                    npairs, &
                                    natoms_encode, &
                                    pairs)
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs
      INTEGER, INTENT(IN), OPTIONAL                      :: npairs, natoms_encode
      INTEGER(KIND=int_8), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: pairs

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_com_atom_pairs_set', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom_pairs%obj)))CALL cp__a("qs_fb_com_tasks_types.F",550)
      IF (PRESENT(npairs)) atom_pairs%obj%npairs = npairs
      IF (PRESENT(natoms_encode)) atom_pairs%obj%natoms_encode = natoms_encode
      IF (PRESENT(pairs)) THEN
         IF (ASSOCIATED(atom_pairs%obj%pairs)) THEN
            DEALLOCATE (atom_pairs%obj%pairs)
         END IF
         atom_pairs%obj%pairs => pairs
      END IF
   END SUBROUTINE fb_com_atom_pairs_set

! **********************************************************************
!> \brief Start from a local set of tasks that has desc/src process equal
!>        to the local MPI rank, communicate with other processes so
!>        that a new local set of tasks is constructed with src/desc
!>        process equal to the local MPI rank
!> \param tasks_dest_is_me the local com_task object with all tasks
!>                          having the desc process id equal to my_id
!> \param direction direction of operation:
!>                   ">" means from tasks_dest_is_me construct tasks_src_is_me
!>                   "<" means from tasks_src_is_me construct tasks_dest_is_me
!> \param tasks_src_is_me the local com_task object with all tasks
!>                          having the src process id equal to my_id
!> \param para_env CP2K parallel environment object that stores MPI related
!>                  information of the current run
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_transpose_dest_src(tasks_dest_is_me, &
                                              direction, &
                                              tasks_src_is_me, &
                                              para_env)
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: tasks_dest_is_me
      CHARACTER, INTENT(IN)                              :: direction
      TYPE(fb_com_tasks_obj), INTENT(INOUT)              :: tasks_src_is_me
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_tasks_transpose_dest_src', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ii, ind, ipe, itask, jj, &
                                                            nencode, ntasks_in, ntasks_out, rank, &
                                                            rank_pos, task_dim
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_in, tasks_out
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: recv_buf, recv_disps, recv_sizes, &
                                                            send_buf, send_disps, send_sizes

      CALL timeset(routineN, handle)

      NULLIFY (tasks_in, tasks_out)

      IF (direction == "<") THEN
         CALL fb_com_tasks_get(com_tasks=tasks_src_is_me, &
                               task_dim=task_dim, &
                               ntasks=ntasks_in, &
                               tasks=tasks_in, &
                               nencode=nencode)
         rank_pos = TASK_DEST
      ELSE
         CALL fb_com_tasks_get(com_tasks=tasks_dest_is_me, &
                               task_dim=task_dim, &
                               ntasks=ntasks_in, &
                               tasks=tasks_in, &
                               nencode=nencode)
         rank_pos = TASK_SRC
      END IF

      ! allocate local arrays
      ALLOCATE (send_sizes(para_env%num_pe))
      ALLOCATE (send_disps(para_env%num_pe))
      ALLOCATE (send_buf(para_env%num_pe))

      ALLOCATE (recv_sizes(para_env%num_pe))
      ALLOCATE (recv_disps(para_env%num_pe))
      ALLOCATE (recv_buf(para_env%num_pe))

      ! first count how many local recv/send tasks need to be sent to
      ! other processes, and share this information with the other
      ! processes.  using send_buf as a temporary array for counting
      send_buf = 0
      ! looping over local task list
      DO itask = 1, ntasks_in
         rank = INT(tasks_in(rank_pos, itask))+1
         send_buf(rank) = send_buf(rank)+1
      END DO

      CALL mp_alltoall(send_buf, recv_buf, 1, para_env%group)

      ! now that we know how many recv/send tasks to send, pack the
      ! tasks, and send them around, so that the recv/send tasks are
      ! sent to the correct src/dest processes, and these then are
      ! collected into the send/recv tasks list on each of the src/dest
      ! processes

      send_sizes = 0
      send_disps = 0
      recv_sizes = 0
      recv_disps = 0

      ! work out the sizes of send and recv buffers and allocate them
      send_sizes(1) = send_buf(1)*task_dim
      recv_sizes(1) = recv_buf(1)*task_dim
      DO ipe = 2, para_env%num_pe
         send_sizes(ipe) = send_buf(ipe)*task_dim
         send_disps(ipe) = send_disps(ipe-1)+send_sizes(ipe-1)
         recv_sizes(ipe) = recv_buf(ipe)*task_dim
         recv_disps(ipe) = recv_disps(ipe-1)+recv_sizes(ipe-1)
      END DO

      ! reallocate send and recv buffers to the correct sizes for
      ! transferring the actual tasks
      DEALLOCATE (send_buf)
      DEALLOCATE (recv_buf)
      ALLOCATE (send_buf(SUM(send_sizes)))
      ALLOCATE (recv_buf(SUM(recv_sizes)))

      ! now that the send buffer is of correct size, do packing
      ! send_buf and recv_buf may be zero sized
      IF (SIZE(send_buf) > 0) send_buf = 0
      IF (SIZE(recv_buf) > 0) recv_buf = 0
      send_sizes = 0
      DO itask = 1, ntasks_in
         rank = INT(tasks_in(rank_pos, itask))+1
         DO ii = 1, task_dim
            ind = send_disps(rank)+send_sizes(rank)+ii
            send_buf(ind) = INT(tasks_in(ii, itask))
         END DO
         send_sizes(rank) = send_sizes(rank)+task_dim
      END DO
      ! do communication
      CALL mp_alltoall(send_buf, send_sizes, send_disps, &
                       recv_buf, recv_sizes, recv_disps, &
                       para_env%group)

      ! deallocate send buffers
      DEALLOCATE (send_buf)
      DEALLOCATE (send_sizes)
      DEALLOCATE (send_disps)

      ! allocate the output task list
      ntasks_out = SUM(recv_sizes)/task_dim
      ! this will not be deallocated in this subroutine
      ALLOCATE (tasks_out(task_dim, ntasks_out))

      ! do unpacking
      itask = 0
      DO ipe = 1, para_env%num_pe
         DO ii = 0, recv_sizes(ipe)/task_dim-1
            itask = itask+1
            DO jj = 1, task_dim
               ind = recv_disps(ipe)+ii*task_dim+jj
               tasks_out(jj, itask) = recv_buf(ind)
            END DO
         END DO
      END DO

      ! set output tasks
      IF (direction == "<") THEN
         CALL fb_com_tasks_set(com_tasks=tasks_dest_is_me, &
                               task_dim=task_dim, &
                               ntasks=ntasks_out, &
                               tasks=tasks_out, &
                               nencode=nencode)
      ELSE
         CALL fb_com_tasks_set(com_tasks=tasks_src_is_me, &
                               task_dim=task_dim, &
                               ntasks=ntasks_out, &
                               tasks=tasks_out, &
                               nencode=nencode)
      END IF

      ! deallocate recv buffers
      DEALLOCATE (recv_buf)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (recv_disps)

      CALL timestop(handle)

   END SUBROUTINE fb_com_tasks_transpose_dest_src

! **********************************************************************
!> \brief Generate send or receive atom_pair lists from a com_tasks
!>        object. atom_pair list is used as a condensed index for the
!>        local/remote matrix blocks to be sent/received.
!> \param com_tasks the com_tasks object
!> \param atom_pairs fb_com_atom_pairs_obj containing  list of encoded
!>                    atomic pair indices and the dest/src proc id for
!>                    the matrix block to be sent/received.
!> \param natoms_encode the total number of atoms the atomic pair indices
!>                       corresponds to, and it is used for encode the
!>                       atom_pairs values
!> \param send_or_recv whether the atom_pair to be generated is for
!>                      the local matrix blocks to be sent or the
!>                      remote matrix blocks to be received for this MPI
!>                      process
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_build_atom_pairs(com_tasks, &
                                            atom_pairs, &
                                            natoms_encode, &
                                            send_or_recv)
      TYPE(fb_com_tasks_obj), INTENT(IN)                 :: com_tasks
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs
      INTEGER, INTENT(IN)                                :: natoms_encode
      CHARACTER(len=*), INTENT(IN)                       :: send_or_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_tasks_build_atom_pairs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ii, itask, jatom, npairs, &
                                                            ntasks, rank, rank_pos
      INTEGER(KIND=int_8)                                :: pair
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: tmp_index
      LOGICAL                                            :: check_ok

      CALL timeset(routineN, handle)

      NULLIFY (pairs, tasks)

      check_ok = fb_com_atom_pairs_has_data(atom_pairs)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",771)

      ! initialise atom_pairs
      CALL fb_com_atom_pairs_init(atom_pairs)

      IF (TRIM(send_or_recv) == "send") THEN
         rank_pos = TASK_DEST
      ELSE
         rank_pos = TASK_SRC
      END IF

      CALL fb_com_tasks_get(com_tasks=com_tasks, &
                            ntasks=ntasks, &
                            tasks=tasks)

      ALLOCATE (pairs(ntasks))
      ! we can have cases where ntasks == 0
      IF (SIZE(pairs) > 0) pairs = 0
      npairs = ntasks

      DO itask = 1, ntasks
         pair = tasks(TASK_PAIR, itask)
         CALL fb_com_tasks_decode_pair(pair, iatom, jatom, natoms_encode)
         rank = INT(tasks(rank_pos, itask))
         CALL fb_com_atom_pairs_encode(pairs(itask), &
                                       rank, iatom, jatom, natoms_encode)
      END DO

      ! sort atom_pairs so that the pairs are ordered process blocks and
      ! that possible duplicates may be found (we don't want to send or
      ! receive same information to the same destination or source more
      ! than once)
      IF (npairs > 0) THEN
         ALLOCATE (tmp_index(npairs))
         ! only sort the actual pairs recorded in the send list
         CALL sort(pairs, npairs, tmp_index)
         DEALLOCATE (tmp_index)
      END IF

      ! remove duplicates
      IF (npairs > 1) THEN
         npairs = 1
         ! first atom pair must be allowed
         DO ii = 2, ntasks
            IF (pairs(ii) > pairs(ii-1)) THEN
               npairs = npairs+1
               pairs(npairs) = pairs(ii)
            END IF
         END DO
         ! reallocate the pairs list
         CALL reallocate(pairs, 1, npairs)
      END IF

      CALL fb_com_atom_pairs_set(atom_pairs=atom_pairs, &
                                 pairs=pairs, &
                                 npairs=npairs, &
                                 natoms_encode=natoms_encode)

      CALL timestop(handle)

   END SUBROUTINE fb_com_tasks_build_atom_pairs

! **********************************************************************
!> \brief Encodes (iatom, jatom) pair index of a block into a single
!>        integer
!> \param ind encoded integer
!> \param iatom the first index of the (iatom, jatom) block index
!> \param jatom the second index of the (iatom, jatom) block index
!> \param natoms the total number of atoms iatom and jatom indexes
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_encode_pair(ind, iatom, jatom, natoms)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: ind
      INTEGER, INTENT(IN)                                :: iatom, jatom, natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_tasks_encode_pair', &
         routineP = moduleN//':'//routineN

      INTEGER(KIND=int_8)                                :: iatom8, jatom8, natoms8

      natoms8 = INT(natoms, int_8)
      iatom8 = INT(iatom, int_8)
      jatom8 = INT(jatom, int_8)

      ind = (iatom8-1_int_8)*natoms8+(jatom8-1_int_8)
   END SUBROUTINE fb_com_tasks_encode_pair

! **********************************************************************
!> \brief Dncodes a single integer into (iatom, jatom) pair index of
!>        a block into a single
!> \param ind encoded integer
!> \param iatom the first index of the (iatom, jatom) block index
!> \param jatom the second index of the (iatom, jatom) block index
!> \param natoms the total number of atoms iatom and jatom indexes
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_tasks_decode_pair(ind, iatom, jatom, natoms)
      INTEGER(KIND=int_8), INTENT(IN)                    :: ind
      INTEGER, INTENT(OUT)                               :: iatom, jatom
      INTEGER, INTENT(IN)                                :: natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_tasks_decode_pair', &
         routineP = moduleN//':'//routineN

      INTEGER(KIND=int_8)                                :: iatom8, jatom8, natoms8

      natoms8 = INT(natoms, int_8)
      iatom8 = ind/natoms8+1_int_8
      jatom8 = MOD(ind, natoms8)+1_int_8
      iatom = INT(iatom8, int_4)
      jatom = INT(jatom8, int_4)
   END SUBROUTINE fb_com_tasks_decode_pair

! **********************************************************************
!> \brief Encodes (rank, iatom, jatom) index of a communication task---to
!>         send/receive a block to/from a process---into a single integer
!> \param ind encoded integer
!> \param pe the rank of the process the block to be send to or receive
!>            from
!> \param iatom the first index of the (iatom, jatom) block index
!> \param jatom the second index of the (iatom, jatom) block index
!> \param natoms the total number of atoms iatom and jatom indexes
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_encode(ind, pe, iatom, jatom, natoms)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: ind
      INTEGER, INTENT(IN)                                :: pe, iatom, jatom, natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_atom_pairs_encode', &
         routineP = moduleN//':'//routineN

      INTEGER(KIND=int_8)                                :: natoms8, pair

! pe must start count from 0 (i.e same as MPI convension)

      natoms8 = INT(natoms, int_8)
      CALL fb_com_tasks_encode_pair(pair, iatom, jatom, natoms)
      ind = INT(pe, int_8)*natoms8*natoms8+pair
   END SUBROUTINE fb_com_atom_pairs_encode

! **********************************************************************
!> \brief Decodes a single integer into the (rank, iatom, jatom) index
!>        of a communication task to send/receive a block to/from a
!>        process
!> \param ind    : encoded integer
!> \param pe     : the rank of the process the block to be send to or receive
!>            from
!> \param iatom  : the first index of the (iatom, jatom) block index
!> \param jatom  : the second index of the (iatom, jatom) block index
!> \param natoms : the total number of atoms iatom and jatom indexes
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_decode(ind, pe, iatom, jatom, natoms)
      INTEGER(KIND=int_8), INTENT(IN)                    :: ind
      INTEGER, INTENT(OUT)                               :: pe, iatom, jatom
      INTEGER, INTENT(IN)                                :: natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_atom_pairs_decode', &
         routineP = moduleN//':'//routineN

      INTEGER(KIND=int_8)                                :: natoms8, pair

! pe start count from 0 (i.e same as MPI convension)

      natoms8 = INT(natoms, int_8)
      pe = INT(ind/(natoms8*natoms8), int_4)
      pair = MOD(ind, natoms8*natoms8)
      CALL fb_com_tasks_decode_pair(pair, iatom, jatom, natoms)
   END SUBROUTINE fb_com_atom_pairs_decode

! **********************************************************************
!> \brief Calculate the MPI send or recv buffer sizes according to the
!>        communication pairs (atom_pairs) and DBCSR matrix data.
!>        Each atom_pair corresponds to one DBCSR matrix block that
!>        needs to be sent or recerived.
!> \param atom_pairs : the communication pair object for either sending
!>                     or receiving
!> \param nprocs : total number of MPI processes in communicator
!> \param row_blk_sizes : row_blk_sizes(iblkrow) = number of element rows
!>                        in each block in the iblkrow-th block row of
!>                        the DBCSR matrix
!> \param col_blk_sizes : col_blk_sizes(iblkcol) = number of element cols
!>                        in each block in the iblkcol-th block col of
!>                        the DBCSR matrix
!> \param sendrecv_sizes : size required for the send of recv buffer
!>                         for each dest/src process
!> \param sendrecv_disps : sendrecv_disps(ipe) + 1 = starting location
!>                         in send/recv buffer for data destined for
!>                         process ipe
!> \param sendrecv_pair_counts : sendrecv_pair_counts(ipe) = number of
!>                               pairs (blocks) to be sent to or recv
!>                               from process ipe
!> \param sendrecv_pair_disps send_recv_pair_disps(ipe) + 1 = start
!>                               location in atom_pairs array for
!>                               all the pairs to be sent to or recv
!>                               from process ipe
!> \param row_map : optional blk row map for the DBCSR blocks
!> \param col_map : optional blk col map for the DBCSR blocks
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_calc_buffer_sizes(atom_pairs, &
                                                  nprocs, &
                                                  row_blk_sizes, &
                                                  col_blk_sizes, &
                                                  sendrecv_sizes, &
                                                  sendrecv_disps, &
                                                  sendrecv_pair_counts, &
                                                  sendrecv_pair_disps, &
                                                  row_map, &
                                                  col_map)
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs
      INTEGER, INTENT(IN)                                :: nprocs
      INTEGER, DIMENSION(:), INTENT(IN)                  :: row_blk_sizes, col_blk_sizes
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sendrecv_sizes, sendrecv_disps, &
                                                            sendrecv_pair_counts, &
                                                            sendrecv_pair_disps
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: row_map, col_map

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_atom_pairs_calc_buffer_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ipair, ipe, jatom, natoms_encode, &
                                                            ncols_blk, npairs, nrows_blk, pe
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs
      LOGICAL                                            :: check_ok

      NULLIFY (pairs)

      check_ok = SIZE(sendrecv_sizes) == nprocs .AND. &
                 SIZE(sendrecv_disps) == nprocs .AND. &
                 SIZE(sendrecv_pair_counts) == nprocs .AND. &
                 SIZE(sendrecv_pair_disps) == nprocs
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1003)

      check_ok = fb_com_atom_pairs_has_data(atom_pairs)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1006)

      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs, &
                                 pairs=pairs, &
                                 npairs=npairs, &
                                 natoms_encode=natoms_encode)

      sendrecv_sizes = 0
      sendrecv_pair_counts = 0
      DO ipair = 1, npairs
         ! decode processor and (iatom, jatom) information
         CALL fb_com_atom_pairs_decode(pairs(ipair), &
                                       pe, iatom, jatom, natoms_encode)
         pe = pe+1 ! we need proc to count from 1
         IF (PRESENT(row_map)) iatom = row_map(iatom)
         IF (PRESENT(col_map)) jatom = row_map(jatom)
         nrows_blk = row_blk_sizes(iatom)
         ncols_blk = col_blk_sizes(jatom)
         sendrecv_sizes(pe) = sendrecv_sizes(pe)+nrows_blk*ncols_blk
         sendrecv_pair_counts(pe) = sendrecv_pair_counts(pe)+1
      END DO
      ! calculate displacements of the data of each destibation pe in
      ! send buffer and in the list of pairs to be sent
      sendrecv_disps = 0
      sendrecv_pair_disps = 0
      DO ipe = 2, nprocs
         sendrecv_disps(ipe) = sendrecv_disps(ipe-1)+sendrecv_sizes(ipe-1)
         sendrecv_pair_disps(ipe) = sendrecv_pair_disps(ipe-1)+sendrecv_pair_counts(ipe-1)
      END DO

   END SUBROUTINE fb_com_atom_pairs_calc_buffer_sizes

! ****************************************************************************
!> \brief Given send and recv fb_com_atom_pair object, gather all the
!>        relevant DBCSR matrix blocks together, and add them to
!>        a fb_matrix_data object for storage
!> \param dbcsr_mat : the DBCSR matrix where the matrix blocks will be
!>                    obtained from
!> \param atom_pairs_send : prescription on exactly which DBCSR blocks
!>                          are to be sent to where
!> \param atom_pairs_recv : prescription on exactly which DBCSR blocks
!>                          are to be received from where
!> \param para_env        : CP2K parallel environment
!> \param matrix_storage  : the fb_matrix_data object to store the
!>                          received DBCSR matrix blocks
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_gather_blks(dbcsr_mat, &
                                            atom_pairs_send, &
                                            atom_pairs_recv, &
                                            para_env, &
                                            matrix_storage)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs_send, atom_pairs_recv
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: matrix_storage

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_atom_pairs_gather_blks', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iatom, ii, ind, ipair, ipe, jatom, jj, ncols_blk, ncols_blk_max, &
         npairs_recv, npairs_send, nrows_blk, nrows_blk_max, numprocs, pe, recv_encode, send_encode
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs_recv, pairs_send
      INTEGER, ALLOCATABLE, DIMENSION(:) :: recv_disps, recv_pair_count, recv_pair_disps, &
         recv_sizes, send_disps, send_pair_count, send_pair_disps, send_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_block_size_data, row_block_size_data
      LOGICAL                                            :: check_ok, found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf, send_buf
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat_block

      CALL timeset(routineN, handle)

      NULLIFY (pairs_send, pairs_recv, mat_block, &
               row_block_size_data, col_block_size_data)

      check_ok = fb_com_atom_pairs_has_data(atom_pairs_send)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1082)
      check_ok = fb_com_atom_pairs_has_data(atom_pairs_send)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1084)
      check_ok = fb_matrix_data_has_data(matrix_storage)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1086)

      ! get com pair informations
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_send, &
                                 pairs=pairs_send, &
                                 npairs=npairs_send, &
                                 natoms_encode=send_encode)
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_recv, &
                                 pairs=pairs_recv, &
                                 npairs=npairs_recv, &
                                 natoms_encode=recv_encode)
      ! get para_env info
      numprocs = para_env%num_pe

      ! get dbcsr row and col block sizes
      row_block_size_data => cp_dbcsr_row_block_sizes(dbcsr_mat)
      col_block_size_data => cp_dbcsr_col_block_sizes(dbcsr_mat)

      ! allocate temporary arrays for send
      ALLOCATE (send_sizes(numprocs))
      ALLOCATE (send_disps(numprocs))
      ALLOCATE (send_pair_count(numprocs))
      ALLOCATE (send_pair_disps(numprocs))

      ! setup send buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_send, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               send_sizes, &
                                               send_disps, &
                                               send_pair_count, &
                                               send_pair_disps)

      ! allocate send buffer
      ALLOCATE (send_buf(SUM(send_sizes)))

      ! allocate temporary arrays for recv
      ALLOCATE (recv_sizes(numprocs))
      ALLOCATE (recv_disps(numprocs))
      ALLOCATE (recv_pair_count(numprocs))
      ALLOCATE (recv_pair_disps(numprocs))

      ! setup recv buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_recv, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               recv_sizes, &
                                               recv_disps, &
                                               recv_pair_count, &
                                               recv_pair_disps)

      ! allocate recv buffer
      ALLOCATE (recv_buf(SUM(recv_sizes)))

      ! do packing
      DO ipe = 1, numprocs
         ! need to reuse send_sizes as an accumulative displacement, so recalculate
         send_sizes(ipe) = 0
         DO ipair = 1, send_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_send(send_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, send_encode)
            nrows_blk = row_block_size_data(iatom)
            ncols_blk = col_block_size_data(jatom)
            CALL cp_dbcsr_get_block_p(matrix=dbcsr_mat, &
                                      row=iatom, col=jatom, block=mat_block, &
                                      found=found)
            IF (.NOT. found) THEN
               CALL cp__b("qs_fb_com_tasks_types.F",1155,"Matrix block not found")
            ELSE
               ! we have found the matrix block
               DO jj = 1, ncols_blk
                  DO ii = 1, nrows_blk
                     ! column major format in blocks
                     ind = send_disps(ipe)+send_sizes(ipe)+ii+(jj-1)*nrows_blk
                     send_buf(ind) = mat_block(ii, jj)
                  END DO ! ii
               END DO ! jj
               send_sizes(ipe) = send_sizes(ipe)+nrows_blk*ncols_blk
            END IF
         END DO ! ipair
      END DO ! ipe

      ! do communication
      CALL mp_alltoall(send_buf, send_sizes, send_disps, &
                       recv_buf, recv_sizes, recv_disps, &
                       para_env%group)

      ! cleanup temporary arrays no longer needed
      DEALLOCATE (send_buf)
      DEALLOCATE (send_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (send_pair_count)
      DEALLOCATE (send_pair_disps)

      ! unpack into matrix_data object
      NULLIFY (mat_block)
      nrows_blk_max = MAXVAL(row_block_size_data)
      ncols_blk_max = MAXVAL(col_block_size_data)
      ALLOCATE (mat_block(nrows_blk_max, ncols_blk_max))
      DO ipe = 1, numprocs
         recv_sizes(ipe) = 0
         DO ipair = 1, recv_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_recv(recv_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, recv_encode)
            nrows_blk = row_block_size_data(iatom)
            ncols_blk = col_block_size_data(jatom)
            ! ALLOCATE(mat_block(nrows_blk,ncols_blk), STAT=stat)
            ! CPPostcondition(stat==0, cp_failure_level, routineP,failure)
            mat_block(:, :) = 0.0_dp
            DO jj = 1, ncols_blk
               DO ii = 1, nrows_blk
                  ! column major format in blocks
                  ind = recv_disps(ipe)+recv_sizes(ipe)+ii+(jj-1)*nrows_blk
                  mat_block(ii, jj) = recv_buf(ind)
               END DO ! ii
            END DO ! jj
            CALL fb_matrix_data_add(matrix_storage, &
                                    iatom, jatom, &
                                    mat_block(1:nrows_blk, 1:ncols_blk))
            recv_sizes(ipe) = recv_sizes(ipe)+nrows_blk*ncols_blk
            ! DEALLOCATE(mat_block, STAT=stat)
            ! CPPostcondition(stat==0, cp_failure_level, routineP,failure)
         END DO ! ipair
      END DO ! ipe
      DEALLOCATE (mat_block)

      ! cleanup rest of the temporary arrays
      DEALLOCATE (recv_buf)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (recv_disps)
      DEALLOCATE (recv_pair_count)
      DEALLOCATE (recv_pair_disps)

      CALL timestop(handle)

   END SUBROUTINE fb_com_atom_pairs_gather_blks

! ****************************************************************************
!> \brief Given send and recv fb_com_atom_pair object, distribute the matrix
!>        blocks stored in a fb_matrix_data object to a compatable DBCSR
!>        matrix. It is assumed in this subroutine that the sizes of each
!>        block stored in fb_matrix_data object is consistent with the
!>        pre-defined block sizes in the DBCSR matrix.
!> \param matrix_storage  : the fb_matrix_data object
!> \param atom_pairs_send : prescription on exactly which DBCSR blocks
!>                          are to be sent to where
!> \param atom_pairs_recv : prescription on exactly which DBCSR blocks
!>                          are to be received from where
!> \param para_env        : CP2K parallel environment
!> \param dbcsr_mat : the DBCSR matrix where the matrix blocks will be
!>                    distributed to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_com_atom_pairs_distribute_blks(matrix_storage, &
                                                atom_pairs_send, &
                                                atom_pairs_recv, &
                                                para_env, &
                                                dbcsr_mat)
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: matrix_storage
      TYPE(fb_com_atom_pairs_obj), INTENT(IN)            :: atom_pairs_send, atom_pairs_recv
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_com_atom_pairs_distribute_blks', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iatom, ii, ind, ipair, ipe, jatom, jj, ncols_blk, npairs_recv, &
         npairs_send, nrows_blk, numprocs, pe, recv_encode, send_encode
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs_recv, pairs_send
      INTEGER, ALLOCATABLE, DIMENSION(:) :: recv_disps, recv_pair_count, recv_pair_disps, &
         recv_sizes, send_disps, send_pair_count, send_pair_disps, send_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_block_size_data, row_block_size_data
      LOGICAL                                            :: check_ok, found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf, send_buf
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat_block

      CALL timeset(routineN, handle)

      NULLIFY (pairs_send, pairs_recv, mat_block, &
               row_block_size_data, col_block_size_data)

      check_ok = fb_com_atom_pairs_has_data(atom_pairs_send)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1270)
      check_ok = fb_com_atom_pairs_has_data(atom_pairs_send)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1272)
      check_ok = fb_matrix_data_has_data(matrix_storage)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_com_tasks_types.F",1274)

      ! get com pair informations
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_send, &
                                 pairs=pairs_send, &
                                 npairs=npairs_send, &
                                 natoms_encode=send_encode)
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_recv, &
                                 pairs=pairs_recv, &
                                 npairs=npairs_recv, &
                                 natoms_encode=recv_encode)
      ! get para_env info
      numprocs = para_env%num_pe

      ! get dbcsr row and col block sizes
      row_block_size_data => cp_dbcsr_row_block_sizes(dbcsr_mat)
      col_block_size_data => cp_dbcsr_col_block_sizes(dbcsr_mat)

      ! allocate temporary arrays for send
      ALLOCATE (send_sizes(numprocs))
      ALLOCATE (send_disps(numprocs))
      ALLOCATE (send_pair_count(numprocs))
      ALLOCATE (send_pair_disps(numprocs))

      ! setup send buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_send, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               send_sizes, &
                                               send_disps, &
                                               send_pair_count, &
                                               send_pair_disps)

      ! allocate send buffer
      ALLOCATE (send_buf(SUM(send_sizes)))

      ! allocate temporary arrays for recv
      ALLOCATE (recv_sizes(numprocs))
      ALLOCATE (recv_disps(numprocs))
      ALLOCATE (recv_pair_count(numprocs))
      ALLOCATE (recv_pair_disps(numprocs))

      ! setup recv buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_recv, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               recv_sizes, &
                                               recv_disps, &
                                               recv_pair_count, &
                                               recv_pair_disps)

      ! allocate recv buffer
      ALLOCATE (recv_buf(SUM(recv_sizes)))

      ! do packing
      DO ipe = 1, numprocs
         ! need to reuse send_sizes as an accumulative displacement, so recalculate
         send_sizes(ipe) = 0
         DO ipair = 1, send_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_send(send_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, send_encode)
            CALL fb_matrix_data_get(matrix_storage, &
                                    iatom, jatom, &
                                    mat_block, found)
            IF (.NOT. found) THEN
               CALL cp__b("qs_fb_com_tasks_types.F",1341,"Matrix block not found")
            ELSE
               nrows_blk = row_block_size_data(iatom)
               ncols_blk = col_block_size_data(jatom)
               DO jj = 1, ncols_blk
                  DO ii = 1, nrows_blk
                     ! column major format in blocks
                     ind = send_disps(ipe)+send_sizes(ipe)+ii+(jj-1)*nrows_blk
                     send_buf(ind) = mat_block(ii, jj)
                  END DO ! ii
               END DO ! jj
               send_sizes(ipe) = send_sizes(ipe)+nrows_blk*ncols_blk
            END IF
         END DO ! ipair
      END DO ! ipe

      ! do communication
      CALL mp_alltoall(send_buf, send_sizes, send_disps, &
                       recv_buf, recv_sizes, recv_disps, &
                       para_env%group)

      ! cleanup temporary arrays no longer needed
      DEALLOCATE (send_buf)
      DEALLOCATE (send_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (send_pair_count)
      DEALLOCATE (send_pair_disps)

      ! unpack into DBCSR matrix
      DO ipe = 1, numprocs
         recv_sizes(ipe) = 0
         DO ipair = 1, recv_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_recv(recv_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, recv_encode)
            nrows_blk = row_block_size_data(iatom)
            ncols_blk = col_block_size_data(jatom)
            ind = recv_disps(ipe)+recv_sizes(ipe)
            CALL cp_dbcsr_put_block(dbcsr_mat, &
                                    iatom, jatom, &
                                    recv_buf((ind+1):(ind+nrows_blk*ncols_blk)))
            recv_sizes(ipe) = recv_sizes(ipe)+nrows_blk*ncols_blk
         END DO ! ipair
      END DO ! ipe

      ! cleanup rest of the temporary arrays
      DEALLOCATE (recv_buf)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (recv_disps)
      DEALLOCATE (recv_pair_count)
      DEALLOCATE (recv_pair_disps)

      ! dbcsr matrix is not finalised in this subroutine

      CALL timestop(handle)

   END SUBROUTINE fb_com_atom_pairs_distribute_blks

END MODULE qs_fb_com_tasks_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_atomic_matrix_methods

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_col_block_sizes,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_get_info,&
                                              cp_dbcsr_get_stored_coordinates,&
                                              cp_dbcsr_row_block_sizes,&
                                              cp_dbcsr_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_alltoall
   USE qs_fb_atomic_halo_types,         ONLY: fb_atomic_halo_atom_global2halo,&
                                              fb_atomic_halo_get,&
                                              fb_atomic_halo_has_data,&
                                              fb_atomic_halo_list_get,&
                                              fb_atomic_halo_list_obj,&
                                              fb_atomic_halo_obj
   USE qs_fb_com_tasks_types,           ONLY: &
        TASK_COST, TASK_DEST, TASK_N_RECORDS, TASK_PAIR, TASK_SRC, &
        fb_com_atom_pairs_calc_buffer_sizes, fb_com_atom_pairs_create, fb_com_atom_pairs_decode, &
        fb_com_atom_pairs_get, fb_com_atom_pairs_has_data, fb_com_atom_pairs_init, &
        fb_com_atom_pairs_nullify, fb_com_atom_pairs_obj, fb_com_atom_pairs_release, &
        fb_com_tasks_build_atom_pairs, fb_com_tasks_create, fb_com_tasks_decode_pair, &
        fb_com_tasks_encode_pair, fb_com_tasks_get, fb_com_tasks_nullify, fb_com_tasks_obj, &
        fb_com_tasks_release, fb_com_tasks_set, fb_com_tasks_transpose_dest_src
   USE qs_fb_matrix_data_types,         ONLY: fb_matrix_data_get,&
                                              fb_matrix_data_has_data,&
                                              fb_matrix_data_obj
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_atomic_matrix_methods'

   PUBLIC :: fb_atmatrix_calc_size, &
             fb_atmatrix_construct, &
             fb_atmatrix_construct_2, &
             fb_atmatrix_generate_com_pairs_2

CONTAINS

! **********************************************************************
!> \brief Calculates the atomic matrix size from a given DBCSR matrix
!>        and atomic halo. It also calculates the first row (col) or the
!>        row (col) atomic blocks in the atomic matrix
!> \param dbcsr_mat : pointer to the DBCSR matrix the atomic matrix is
!>                    to be constructed from
!> \param atomic_halo : the atomic halo used for defining the atomic
!>                      matrix from the DBCSR matrix
!> \param nrows : outputs total number of rows in the atomic matrix
!> \param ncols : outputs total number of cols in the atomic matrix
!> \param blk_row_start : first row in each atomic blk row in the
!>                        atomic matrix
!> \param blk_col_start : first col in each atomic blk col in the
!>                        atomic matrix
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atmatrix_calc_size(dbcsr_mat, &
                                    atomic_halo, &
                                    nrows, &
                                    ncols, &
                                    blk_row_start, &
                                    blk_col_start)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      INTEGER, INTENT(OUT)                               :: nrows, ncols
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: blk_row_start, blk_col_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_atmatrix_calc_size', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, natoms_in_halo
      INTEGER, DIMENSION(:), POINTER                     :: col_block_size_data, halo_atoms, &
                                                            row_block_size_data
      LOGICAL                                            :: check_ok

      NULLIFY (halo_atoms, row_block_size_data, col_block_size_data)

      row_block_size_data => cp_dbcsr_row_block_sizes(dbcsr_mat)
      col_block_size_data => cp_dbcsr_col_block_sizes(dbcsr_mat)
      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)
      check_ok = SIZE(blk_row_start) .GE. (natoms_in_halo+1)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_atomic_matrix_methods.F",93)
      check_ok = SIZE(blk_col_start) .GE. (natoms_in_halo+1)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_atomic_matrix_methods.F",95)
      blk_row_start = 0
      blk_col_start = 0
      nrows = 0
      ncols = 0
      DO ii = 1, natoms_in_halo
         blk_row_start(ii) = nrows+1
         blk_col_start(ii) = ncols+1
         nrows = nrows+row_block_size_data(halo_atoms(ii))
         ncols = ncols+col_block_size_data(halo_atoms(ii))
      END DO
      blk_row_start(natoms_in_halo+1) = nrows+1
      blk_col_start(natoms_in_halo+1) = ncols+1
   END SUBROUTINE fb_atmatrix_calc_size

! ****************************************************************************
!> \brief Constructs atomic matrix for filter basis method from a given
!>        DBCSR matrix and a set of atomic send and recv pairs
!>        corresponding to the matrix blocks that needs to be included
!>        in the atomic matrix. This version is for when we do MPI
!>        communications at every step, for each atomic matrix.
!> \param dbcsr_mat : the DBCSR matrix the atomic matrix is to be
!>                    constructed from
!> \param atomic_halo : the atomic halo conrresponding to this atomic
!>                      matrix
!> \param para_env : cp2k parallel environment
!> \param atomic_matrix : the atomic matrix to be constructed, it should
!>                        have already been allocated prior entering
!>                        this subroutine
!> \param blk_row_start : first row in each atomic blk row in the
!>                        atomic matrix
!> \param blk_col_start : first col in each atomic blk col in the
!>                        atomic matrix
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atmatrix_construct(dbcsr_mat, &
                                    atomic_halo, &
                                    para_env, &
                                    atomic_matrix, &
                                    blk_row_start, &
                                    blk_col_start)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: atomic_matrix
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_row_start, blk_col_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_atmatrix_construct', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iatom, iatom_in_halo, ii, ind, ipair, ipe, jatom, jatom_in_halo, jj, &
         ncols_blk, npairs_recv, npairs_send, nrows_blk, numprocs, pe, recv_encode, send_encode
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs_recv, pairs_send
      INTEGER, ALLOCATABLE, DIMENSION(:) :: recv_disps, recv_pair_count, recv_pair_disps, &
         recv_sizes, send_disps, send_pair_count, send_pair_disps, send_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_block_size_data, row_block_size_data
      LOGICAL                                            :: found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf, send_buf
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat_block
      TYPE(fb_com_atom_pairs_obj)                        :: atom_pairs_recv, atom_pairs_send

      CALL timeset(routineN, handle)

      NULLIFY (pairs_send, pairs_recv, mat_block, &
               row_block_size_data, col_block_size_data)
      CALL fb_com_atom_pairs_nullify(atom_pairs_send)
      CALL fb_com_atom_pairs_nullify(atom_pairs_recv)

      ! initialise atomic matrix
      IF (SIZE(atomic_matrix, 1) > 0 .AND. SIZE(atomic_matrix, 2) > 0) THEN
         atomic_matrix = 0.0_dp
      END IF

      ! generate send and receiv atomic pairs
      CALL fb_com_atom_pairs_create(atom_pairs_send)
      CALL fb_com_atom_pairs_create(atom_pairs_recv)
      CALL fb_atmatrix_generate_com_pairs(dbcsr_mat, &
                                          atomic_halo, &
                                          para_env, &
                                          atom_pairs_send, &
                                          atom_pairs_recv)

      ! get com pair informations
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_send, &
                                 pairs=pairs_send, &
                                 npairs=npairs_send, &
                                 natoms_encode=send_encode)
      CALL fb_com_atom_pairs_get(atom_pairs=atom_pairs_recv, &
                                 pairs=pairs_recv, &
                                 npairs=npairs_recv, &
                                 natoms_encode=recv_encode)

      ! get para_env info
      numprocs = para_env%num_pe

      ! get dbcsr row and col block sizes
      row_block_size_data => cp_dbcsr_row_block_sizes(dbcsr_mat)
      col_block_size_data => cp_dbcsr_col_block_sizes(dbcsr_mat)

      ! allocate temporary arrays for send
      ALLOCATE (send_sizes(numprocs))
      ALLOCATE (send_disps(numprocs))
      ALLOCATE (send_pair_count(numprocs))
      ALLOCATE (send_pair_disps(numprocs))

      ! setup send buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_send, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               send_sizes, &
                                               send_disps, &
                                               send_pair_count, &
                                               send_pair_disps)
      ! allocate send buffer
      ALLOCATE (send_buf(SUM(send_sizes)))

      ! allocate temporary arrays for recv
      ALLOCATE (recv_sizes(numprocs))
      ALLOCATE (recv_disps(numprocs))
      ALLOCATE (recv_pair_count(numprocs))
      ALLOCATE (recv_pair_disps(numprocs))

      ! setup recv buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(atom_pairs_recv, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               recv_sizes, &
                                               recv_disps, &
                                               recv_pair_count, &
                                               recv_pair_disps)
      ! allocate recv buffer
      ALLOCATE (recv_buf(SUM(recv_sizes)))
      ! do packing
      DO ipe = 1, numprocs
         ! need to reuse send_sizes as an accumulative displacement, so recalculate
         send_sizes(ipe) = 0
         DO ipair = 1, send_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_send(send_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, send_encode)
            nrows_blk = row_block_size_data(iatom)
            ncols_blk = col_block_size_data(jatom)
            CALL cp_dbcsr_get_block_p(matrix=dbcsr_mat, &
                                      row=iatom, col=jatom, block=mat_block, &
                                      found=found)
            IF (.NOT. found) THEN
               CALL cp__b("qs_fb_atomic_matrix_methods.F",242,"Matrix block not found")
            ELSE
               ! we have found the matrix block
               DO jj = 1, ncols_blk
                  DO ii = 1, nrows_blk
                     ! column major format in blocks
                     ind = send_disps(ipe)+send_sizes(ipe)+ii+(jj-1)*nrows_blk
                     send_buf(ind) = mat_block(ii, jj)
                  END DO ! ii
               END DO ! jj
               send_sizes(ipe) = send_sizes(ipe)+nrows_blk*ncols_blk
            END IF
         END DO ! ipair
      END DO ! ipe

      ! do communication
      CALL mp_alltoall(send_buf, send_sizes, send_disps, &
                       recv_buf, recv_sizes, recv_disps, &
                       para_env%group)

      ! cleanup temporary arrays no longer needed
      DEALLOCATE (send_buf)
      DEALLOCATE (send_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (send_pair_count)
      DEALLOCATE (send_pair_disps)

      ! do unpacking
      DO ipe = 1, numprocs
         recv_sizes(ipe) = 0
         DO ipair = 1, recv_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_recv(recv_pair_disps(ipe)+ipair), &
                                          pe, iatom, jatom, recv_encode)
            ! nrows_blk = last_row(iatom) - first_row(iatom) + 1
            ! ncols_blk = last_col(jatom) - first_col(jatom) + 1
            nrows_blk = row_block_size_data(iatom)
            ncols_blk = col_block_size_data(jatom)
            ! get the corresponding atom indices in halo
            ! the atoms from the recv_pairs should be in the atomic_halo, because
            ! the recv_pairs are the matrix blocks requested by the local proc for
            ! this particular atomic_halo
            CALL fb_atomic_halo_atom_global2halo(atomic_halo, &
                                                 iatom, iatom_in_halo, &
                                                 found)
            IF(.NOT.(found))CALL cp__a("qs_fb_atomic_matrix_methods.F",286)
            CALL fb_atomic_halo_atom_global2halo(atomic_halo, &
                                                 jatom, jatom_in_halo, &
                                                 found)
            IF(.NOT.(found))CALL cp__a("qs_fb_atomic_matrix_methods.F",290)
            ! put block into the full conventional matrix
            DO jj = 1, ncols_blk
               DO ii = 1, nrows_blk
                  ! column major format in blocks
                  ind = recv_disps(ipe)+recv_sizes(ipe)+ii+(jj-1)*nrows_blk
                  atomic_matrix(blk_row_start(iatom_in_halo)+ii-1, &
                                blk_col_start(jatom_in_halo)+jj-1) = recv_buf(ind)

               END DO ! ii
            END DO ! jj
            recv_sizes(ipe) = recv_sizes(ipe)+nrows_blk*ncols_blk
         END DO ! ipair
      END DO ! ipe

      ! the constructed matrix is upper triangular, fill it up to full
      DO ii = 2, SIZE(atomic_matrix, 1)
         DO jj = 1, ii-1
            atomic_matrix(ii, jj) = atomic_matrix(jj, ii)
         END DO
      END DO

      ! cleanup rest of the temporary arrays
      DEALLOCATE (recv_buf)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (recv_disps)
      DEALLOCATE (recv_pair_count)
      DEALLOCATE (recv_pair_disps)
      CALL fb_com_atom_pairs_release(atom_pairs_send)
      CALL fb_com_atom_pairs_release(atom_pairs_recv)

      CALL timestop(handle)

   END SUBROUTINE fb_atmatrix_construct

! ****************************************************************************
!> \brief Constructs atomic matrix for filter basis method from a given
!>        DBCSR matrix and a set of atomic send and recv pairs
!>        corresponding to the matrix blocks that needs to be included
!>        in the atomic matrix. This version is for when we do MPI
!>        communications collectively in one go at the beginning.
!> \param matrix_storage : data storing the relevant DBCSR matrix blocks
!>                         needed for constructing the atomic matrix
!> \param atomic_halo : the atomic halo conrresponding to this atomic
!>                      matrix
!> \param atomic_matrix : the atomic matrix to be constructed, it should
!>                        have already been allocated prior entering
!>                        this subroutine
!> \param blk_row_start : first row in each atomic blk row in the
!>                        atomic matrix
!> \param blk_col_start : first col in each atomic blk col in the
!>                        atomic matrix
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atmatrix_construct_2(matrix_storage, &
                                      atomic_halo, &
                                      atomic_matrix, &
                                      blk_row_start, &
                                      blk_col_start)
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: matrix_storage
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: atomic_matrix
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_row_start, blk_col_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_atmatrix_construct_2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, iatom_global, icol, ii, &
                                                            irow, jatom, jatom_global, jj, &
                                                            natoms_in_halo
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      LOGICAL                                            :: check_ok, found
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: blk_p

      CALL timeset(routineN, handle)

      check_ok = fb_matrix_data_has_data(matrix_storage)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_atomic_matrix_methods.F",367)
      check_ok = fb_atomic_halo_has_data(atomic_halo)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_atomic_matrix_methods.F",369)

      NULLIFY (halo_atoms, blk_p)

      ! initialise atomic matrix
      IF (SIZE(atomic_matrix, 1) > 0 .AND. SIZE(atomic_matrix, 2) > 0) THEN
         atomic_matrix = 0.0_dp
      END IF

      ! get atomic halo information
      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)

      ! construct atomic matrix using data from matrix_storage
      DO iatom = 1, natoms_in_halo
         iatom_global = halo_atoms(iatom)
         DO jatom = 1, natoms_in_halo
            jatom_global = halo_atoms(jatom)
            ! atomic matrices are symmetric, fill only the top
            ! triangular part
            IF (jatom_global .GE. iatom_global) THEN
               CALL fb_matrix_data_get(matrix_storage, &
                                       iatom_global, &
                                       jatom_global, &
                                       blk_p, &
                                       found)
               ! copy data to atomic_matrix if found
               IF (found) THEN
                  DO jj = 1, SIZE(blk_p, 2)
                     icol = blk_col_start(jatom)+jj-1
                     DO ii = 1, SIZE(blk_p, 1)
                        irow = blk_row_start(iatom)+ii-1
                        atomic_matrix(irow, icol) = blk_p(ii, jj)
                     END DO ! ii
                  END DO ! jj
               END IF
            END IF
         END DO ! jatom
      END DO ! iatom

      ! the constructed matrix is upper triangular, fill it up to full
      DO ii = 2, SIZE(atomic_matrix, 1)
         DO jj = 1, ii-1
            atomic_matrix(ii, jj) = atomic_matrix(jj, ii)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE fb_atmatrix_construct_2

! ****************************************************************************
!> \brief generate list of blocks (atom pairs) of a DBCSR matrix to be
!>        sent and recived in order to construct an atomic matrix
!>        corresponding to a given atomic halo. This version is for the case
!>        when we do MPI communications at each step, for each atomic matrix.
!> \param dbcsr_mat : The DBCSR matrix the atom blocks come from
!> \param atomic_halo : the atomic halo used to construct the atomic
!>                      matrix
!> \param para_env : cp2k parallel environment
!> \param atom_pairs_send : list of atom blocks from local DBCSR matrix
!>                          data to be sent
!> \param atom_pairs_recv : list of atom blocks from remote DBCSR matrix
!>                          data to be recveived
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atmatrix_generate_com_pairs(dbcsr_mat, &
                                             atomic_halo, &
                                             para_env, &
                                             atom_pairs_send, &
                                             atom_pairs_recv)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs_send, atom_pairs_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_atmatrix_generate_com_pairs', &
         routineP = moduleN//':'//routineN

      INTEGER :: counter, dest, handle, iatom, iatom_global, itask, jatom, jatom_global, &
         natoms_in_halo, nblkrows_total, nencode, ntasks_recv, ntasks_send, src
      INTEGER(KIND=int_8)                                :: pair
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_recv, tasks_send
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      LOGICAL                                            :: found
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat_block
      TYPE(fb_com_tasks_obj)                             :: com_tasks_recv, com_tasks_send

      CALL timeset(routineN, handle)

      NULLIFY (halo_atoms, tasks_send, tasks_recv)
      CALL fb_com_tasks_nullify(com_tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_recv)

      ! initialise atom_pairs_send and atom_pairs_receive
      IF (fb_com_atom_pairs_has_data(atom_pairs_send)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_send)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_send)
      END IF
      IF (fb_com_atom_pairs_has_data(atom_pairs_recv)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_recv)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_recv)
      END IF

      ! get atomic halo information
      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)

      ! get the total number of atoms, we can obtain this directly
      ! from the global block row dimension of the dbcsr matrix
      CALL cp_dbcsr_get_info(matrix=dbcsr_mat, &
                             nblkrows_total=nblkrows_total)

      ! destination proc is always the local processor
      dest = para_env%mepos ! my MPI rank

      ! generate recv task list (tasks_recv)

      ! a recv task corresponds to the copying or transfering of a
      ! matrix block in the part of the DBCSR matrix owned by the src
      ! proc to this proc in order to construct the atomic matrix
      ! corresponding to the given atomic halo. As an upper-bound, the
      ! number of matrix blocks requred do not exceed natoms_in_halo**2
      ntasks_recv = natoms_in_halo*natoms_in_halo

      ALLOCATE (tasks_recv(TASK_N_RECORDS, ntasks_recv))

      ! now that tasks_recv has been allocated, generate the tasks
      itask = 1
      DO iatom = 1, natoms_in_halo
         iatom_global = halo_atoms(iatom)
         DO jatom = 1, natoms_in_halo
            jatom_global = halo_atoms(jatom)
            ! atomic matrix is symmetric, and only upper triangular part
            ! is stored in DBCSR matrix
            IF (jatom_global .GE. iatom_global) THEN
               ! find the source proc that supposed to own the block
               ! (iatom_global, jatom_global)
               CALL cp_dbcsr_get_stored_coordinates(dbcsr_mat, &
                                                    iatom_global, &
                                                    jatom_global, &
                                                    processor=src)
               ! we must encode the global atom indices rather the halo
               ! atomic indices in each task, because halo atomic
               ! indices are local to each halo, and each processor is
               ! working on a different halo local to them. So one
               ! processor would not have the information about the halo
               ! on another processor, rendering the halo atomic indices
               ! rather useless outside the local processor.
               tasks_recv(TASK_DEST, itask) = dest
               tasks_recv(TASK_SRC, itask) = src

               CALL fb_com_tasks_encode_pair(tasks_recv(TASK_PAIR, itask), &
                                             iatom_global, jatom_global, &
                                             nblkrows_total)
               ! calculation of cost not implemented at the moment
               tasks_recv(TASK_COST, itask) = 0
               itask = itask+1
            END IF
         END DO ! jatom
      END DO ! iatom

      ! get the actual number of tasks
      ntasks_recv = itask-1

      ! create tasks
      CALL fb_com_tasks_create(com_tasks_recv)
      CALL fb_com_tasks_create(com_tasks_send)

      CALL fb_com_tasks_set(com_tasks=com_tasks_recv, &
                            task_dim=TASK_N_RECORDS, &
                            ntasks=ntasks_recv, &
                            nencode=nblkrows_total, &
                            tasks=tasks_recv)

      ! genearte the send task list (tasks_send) from the recv task list
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, ">", com_tasks_send, &
                                           para_env)

      CALL fb_com_tasks_get(com_tasks=com_tasks_send, &
                            ntasks=ntasks_send, &
                            tasks=tasks_send, &
                            nencode=nencode)

      ! because the atomic_halos and the neighbor_list_set used to
      ! generate the sparse structure of the DBCSR matrix do not
      ! necessarily have to coincide, we must check of the blocks in
      ! tasks_send (these should be local to the processor) do indeed
      ! exist in the DBCSR matrix, if not, then we need to prune these
      ! out of the task list

      counter = 0
      DO itask = 1, ntasks_send
         pair = tasks_send(TASK_PAIR, itask)
         CALL fb_com_tasks_decode_pair(pair, iatom_global, jatom_global, nencode)
         ! check if block exists in DBCSR matrix
         CALL cp_dbcsr_get_block_p(matrix=dbcsr_mat, &
                                   row=iatom_global, col=jatom_global, block=mat_block, &
                                   found=found)
         IF (found) THEN
            counter = counter+1
            ! we can do this here, because essencially we are inspecting
            ! the send tasks one by one, and then omit ones which the
            ! block is not found in the DBCSR matrix. itask is always
            ! .GE. counter
            tasks_send(1:TASK_N_RECORDS, counter) = tasks_send(1:TASK_N_RECORDS, itask)
         END IF
      END DO
      ! the new send task list should have size counter. counter
      ! .LE. the old ntasks_send, thus the task list does not really
      ! need to be reallocated (as it is just a temporary array), and
      ! the useful data will cutoff at counter, and the rest of the
      ! array will just be garbage
      ntasks_send = counter

      ! tasks_send is set through the pointer already
      CALL fb_com_tasks_set(com_tasks=com_tasks_send, &
                            ntasks=ntasks_send)

      ! now, re-distribute the new send tasks list to other processors
      ! to build the updated recv tasks list
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, "<", com_tasks_send, &
                                           para_env)

      ! task lists are now complete, now construct the atom_pairs_send
      ! and atom_pairs_recv from the tasks lists
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_send, &
                                         atom_pairs=atom_pairs_send, &
                                         natoms_encode=nencode, &
                                         send_or_recv="send")
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_recv, &
                                         atom_pairs=atom_pairs_recv, &
                                         natoms_encode=nencode, &
                                         send_or_recv="recv")

      ! cleanup
      CALL fb_com_tasks_release(com_tasks_recv)
      CALL fb_com_tasks_release(com_tasks_send)

      CALL timestop(handle)

   END SUBROUTINE fb_atmatrix_generate_com_pairs

! ****************************************************************************
!> \brief generate list of blocks (atom pairs) of a DBCSR matrix to be
!>        sent and recived in order to construct all local atomic matrices
!>        corresponding to the atomic halos. This version is for the case
!>        when we do MPI communications collectively in one go at the
!>        beginning.
!> \param dbcsr_mat : The DBCSR matrix the atom blocks come from
!> \param atomic_halos : the list of all atomic halos local to the process
!> \param para_env : cp2k parallel environment
!> \param atom_pairs_send : list of atom blocks from local DBCSR matrix
!>                          data to be sent
!> \param atom_pairs_recv : list of atom blocks from remote DBCSR matrix
!>                          data to be recveived
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atmatrix_generate_com_pairs_2(dbcsr_mat, &
                                               atomic_halos, &
                                               para_env, &
                                               atom_pairs_send, &
                                               atom_pairs_recv)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs_send, atom_pairs_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_atmatrix_generate_com_pairs_2', &
         routineP = moduleN//':'//routineN

      INTEGER :: counter, dest, handle, iatom, iatom_global, ihalo, itask, jatom, jatom_global, &
         natoms_in_halo, nblkrows_total, nencode, nhalos, ntasks_recv, ntasks_send, src
      INTEGER(KIND=int_8)                                :: pair
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_recv, tasks_send
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      LOGICAL                                            :: found
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mat_block
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos
      TYPE(fb_com_tasks_obj)                             :: com_tasks_recv, com_tasks_send

      CALL timeset(routineN, handle)

      NULLIFY (halo_atoms, tasks_send, tasks_recv)
      CALL fb_com_tasks_nullify(com_tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_recv)

      ! initialise atom_pairs_send and atom_pairs_receive
      IF (fb_com_atom_pairs_has_data(atom_pairs_send)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_send)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_send)
      END IF
      IF (fb_com_atom_pairs_has_data(atom_pairs_recv)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_recv)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_recv)
      END IF

      ! get atomic halo list information
      CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                   nhalos=nhalos, &
                                   halos=halos)
      ! get the total number of atoms, we can obtain this directly
      ! from the global block row dimension of the dbcsr matrix
      CALL cp_dbcsr_get_info(matrix=dbcsr_mat, &
                             nblkrows_total=nblkrows_total)

      ! estimate the maximum number of blocks to be received
      ntasks_recv = 0
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 natoms=natoms_in_halo)
         ntasks_recv = ntasks_recv+natoms_in_halo*natoms_in_halo
      END DO
      ALLOCATE (tasks_recv(TASK_N_RECORDS, ntasks_recv))

      ! now that tasks_recv has been allocated, generate the tasks

      ! destination proc is always the local process
      dest = para_env%mepos
      itask = 1
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 natoms=natoms_in_halo, &
                                 halo_atoms=halo_atoms)
         DO iatom = 1, natoms_in_halo
            iatom_global = halo_atoms(iatom)
            DO jatom = 1, natoms_in_halo
               jatom_global = halo_atoms(jatom)
               ! atomic matrices are always symmetric, treat it as such.
               ! so only deal with upper triangular parts
               IF (jatom_global .GE. iatom_global) THEN
                  ! find the source proc that supposed to own the block
                  ! (iatom_global, jatom_global)
                  CALL cp_dbcsr_get_stored_coordinates(dbcsr_mat, &
                                                       iatom_global, &
                                                       jatom_global, &
                                                       processor=src)
                  ! we must encode the global atom indices rather the halo
                  ! atomic indices in each task, because halo atomic indices
                  ! are local to each halo, and each processor is working on a
                  ! different halo local to them. So one processor would not
                  ! have the information about the halo on another processor,
                  ! rendering the halo atomic indices rather useless outside
                  ! the local processor.
                  tasks_recv(TASK_DEST, itask) = dest
                  tasks_recv(TASK_SRC, itask) = src
                  CALL fb_com_tasks_encode_pair(tasks_recv(TASK_PAIR, itask), &
                                                iatom_global, jatom_global, &
                                                nblkrows_total)
                  ! calculation of cost not implemented at the moment
                  tasks_recv(TASK_COST, itask) = 0
                  itask = itask+1
               END IF
            END DO ! jatom
         END DO ! iatom
      END DO ! ihalo

      ! set the actual number of tasks obtained
      ntasks_recv = itask-1

      ! create tasks
      CALL fb_com_tasks_create(com_tasks_recv)
      CALL fb_com_tasks_create(com_tasks_send)

      CALL fb_com_tasks_set(com_tasks=com_tasks_recv, &
                            task_dim=TASK_N_RECORDS, &
                            ntasks=ntasks_recv, &
                            nencode=nblkrows_total, &
                            tasks=tasks_recv)

      ! genearte the send task list (tasks_send) from the recv task list
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, ">", com_tasks_send, &
                                           para_env)

      CALL fb_com_tasks_get(com_tasks=com_tasks_send, &
                            ntasks=ntasks_send, &
                            tasks=tasks_send, &
                            nencode=nencode)

      ! because the atomic_halos and the neighbor_list_set used to
      ! generate the sparse structure of the DBCSR matrix do not
      ! necessarily have to coincide, we must check of the blocks in
      ! tasks_send (these should be local to the processor) do indeed
      ! exist in the DBCSR matrix, if not, then we need to prune these
      ! out of the task list

      counter = 0
      DO itask = 1, ntasks_send
         pair = tasks_send(TASK_PAIR, itask)
         CALL fb_com_tasks_decode_pair(pair, iatom_global, jatom_global, nencode)
         ! check if block exists in DBCSR matrix
         CALL cp_dbcsr_get_block_p(matrix=dbcsr_mat, row=iatom_global, &
                                   col=jatom_global, block=mat_block, &
                                   found=found)
         IF (found) THEN
            counter = counter+1
            ! we can do this here, because essencially we are inspecting
            ! the send tasks one by one, and then omit ones which the
            ! block is not found in the DBCSR matrix. itask is always
            ! .GE. counter
            tasks_send(1:TASK_N_RECORDS, counter) = tasks_send(1:TASK_N_RECORDS, itask)
         END IF
      END DO
      ! the new send task list should have size counter. counter
      ! .LE. the old ntasks_send, thus the task list does not really
      ! need to be reallocated (as it is just a temporary array), and
      ! the useful data will cutoff at counter, and the rest of the
      ! array will just be garbage
      ntasks_send = counter

      ! tasks_send is set through the pointer already
      CALL fb_com_tasks_set(com_tasks=com_tasks_send, &
                            ntasks=ntasks_send)

      ! now, re-distribute the new send tasks list to other processors
      ! to build the updated recv tasks list
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, "<", com_tasks_send, &
                                           para_env)

      ! task lists are now complete, now construct the atom_pairs_send
      ! and atom_pairs_recv from the tasks lists
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_send, &
                                         atom_pairs=atom_pairs_send, &
                                         natoms_encode=nencode, &
                                         send_or_recv="send")
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_recv, &
                                         atom_pairs=atom_pairs_recv, &
                                         natoms_encode=nencode, &
                                         send_or_recv="recv")

      ! cleanup
      CALL fb_com_tasks_release(com_tasks_recv)
      CALL fb_com_tasks_release(com_tasks_send)

      CALL timestop(handle)

   END SUBROUTINE fb_atmatrix_generate_com_pairs_2

END MODULE qs_fb_atomic_matrix_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_filter_matrix_methods

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_create, cp_dbcsr_finalize, cp_dbcsr_get_info, cp_dbcsr_get_stored_coordinates, &
        cp_dbcsr_init, cp_dbcsr_put_block, cp_dbcsr_row_block_sizes, cp_dbcsr_type, &
        dbcsr_distribution_obj, dbcsr_type_no_symmetry
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE fermi_utils,                     ONLY: Fermi,&
                                              FermiFixed
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_alltoall
   USE particle_types,                  ONLY: particle_type
   USE qs_fb_atomic_halo_types,         ONLY: fb_atomic_halo_create,&
                                              fb_atomic_halo_get,&
                                              fb_atomic_halo_list_get,&
                                              fb_atomic_halo_list_obj,&
                                              fb_atomic_halo_nullify,&
                                              fb_atomic_halo_obj,&
                                              fb_atomic_halo_release,&
                                              fb_atomic_halo_set
   USE qs_fb_atomic_matrix_methods,     ONLY: fb_atmatrix_calc_size,&
                                              fb_atmatrix_construct,&
                                              fb_atmatrix_construct_2,&
                                              fb_atmatrix_generate_com_pairs_2
   USE qs_fb_com_tasks_types,           ONLY: &
        TASK_COST, TASK_DEST, TASK_N_RECORDS, TASK_PAIR, TASK_SRC, &
        fb_com_atom_pairs_calc_buffer_sizes, fb_com_atom_pairs_create, fb_com_atom_pairs_decode, &
        fb_com_atom_pairs_distribute_blks, fb_com_atom_pairs_gather_blks, fb_com_atom_pairs_get, &
        fb_com_atom_pairs_has_data, fb_com_atom_pairs_init, fb_com_atom_pairs_nullify, &
        fb_com_atom_pairs_obj, fb_com_atom_pairs_release, fb_com_tasks_build_atom_pairs, &
        fb_com_tasks_create, fb_com_tasks_encode_pair, fb_com_tasks_nullify, fb_com_tasks_obj, &
        fb_com_tasks_release, fb_com_tasks_set, fb_com_tasks_transpose_dest_src
   USE qs_fb_matrix_data_types,         ONLY: fb_matrix_data_add,&
                                              fb_matrix_data_create,&
                                              fb_matrix_data_has_data,&
                                              fb_matrix_data_nullify,&
                                              fb_matrix_data_obj,&
                                              fb_matrix_data_release
   USE qs_fb_trial_fns_types,           ONLY: fb_trial_fns_get,&
                                              fb_trial_fns_obj
   USE string_utilities,                ONLY: compress,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_filter_matrix_methods'

   PUBLIC :: fb_fltrmat_build, &
             fb_fltrmat_build_2

CONTAINS

! **************************************************************************************************
!> \brief Build the filter matrix, with MPI communications happening at each
!>        step. Less efficient on communication, but more efficient on
!>        memory usage (compared to fb_fltrmat_build_2)
!> \param H_mat : DBCSR system KS matrix
!> \param S_mat : DBCSR system overlap matrix
!> \param atomic_halos : list of all local atomic halos, each halo gives
!>                       one atomic matrix and contributes to one blk
!>                       col to the filter matrix
!> \param trial_fns : the trial functions to be used to shrink the
!>                     size of the new "filtered" basis
!> \param para_env : cp2k parallel environment
!> \param particle_set : set of all particles in the system
!> \param fermi_level : the fermi level used for defining the filter
!>                      function, which is a Fermi-Dirac distribution
!>                      function
!> \param filter_temp : the filter temperature used for defining the
!>                      filter function
!> \param name        : name given to the filter matrix
!> \param filter_mat  : DBCSR format filter matrix
!> \param tolerance   : anything less than tolerance is treated as zero
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_build(H_mat, &
                               S_mat, &
                               atomic_halos, &
                               trial_fns, &
                               para_env, &
                               particle_set, &
                               fermi_level, &
                               filter_temp, &
                               name, &
                               filter_mat, &
                               tolerance)
      TYPE(cp_dbcsr_type), POINTER                       :: H_mat, S_mat
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: fermi_level, filter_temp
      CHARACTER(LEN=*), INTENT(IN)                       :: name
      TYPE(cp_dbcsr_type), POINTER                       :: filter_mat
      REAL(KIND=dp), INTENT(IN)                          :: tolerance

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_build', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=32)                                  :: symmetry_string
      CHARACTER(LEN=default_string_length)               :: name_string
      INTEGER                                            :: handle, iblkcol, ihalo, ikind, &
                                                            max_nhalos, nblkcols_total, nhalos
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, dummy_halo_atoms, ntfns, &
                                                            row_blk_size
      LOGICAL                                            :: send_data_only
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(dbcsr_distribution_obj)                       :: dbcsr_dist
      TYPE(fb_atomic_halo_obj)                           :: dummy_atomic_halo
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos

      CALL timeset(routineN, handle)

      NULLIFY (halos, atomic_kind, ntfns, dummy_halo_atoms, row_blk_size, col_blk_size)
      CALL fb_atomic_halo_nullify(dummy_atomic_halo)

      ! filter_mat must be of a dissassociated status (i.e. brand new)
      IF(.NOT.(.NOT. ASSOCIATED(filter_mat)))CALL cp__a("qs_fb_filter_matrix_methods.F",130)

      ! get trial function information
      CALL fb_trial_fns_get(trial_fns=trial_fns, &
                            nfunctions=ntfns)

      ! calculate the row_blk_size and col_blk_size arrays for
      ! constructing the filter matrix in DBCSR format
      ! row_blk_size for the filter matrix is the same as H or S
      CALL cp_dbcsr_get_info(H_mat, &
                             nblkcols_total=nblkcols_total, &
                             row_blk_size=row_blk_size, &
                             distribution=dbcsr_dist)
      ALLOCATE (col_blk_size(nblkcols_total))
      col_blk_size = 0
      DO iblkcol = 1, nblkcols_total
         atomic_kind => particle_set(iblkcol)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              kind_number=ikind)
         col_blk_size(iblkcol) = ntfns(ikind)
      END DO
      ! DO NOT deallocate cbs if gift=.TRUE. as col_blk_sizes will only point to cbs
      name_string = name
      CALL compress(name_string)
      CALL uppercase(name_string)
      ! the filter matrix is non-square and is always non-symmetric
      symmetry_string = dbcsr_type_no_symmetry
      ! create empty filter matrix
      ALLOCATE (filter_mat)
      CALL cp_dbcsr_init(filter_mat)
      CALL cp_dbcsr_create(matrix=filter_mat, &
                           name=name_string, &
                           dist=dbcsr_dist, &
                           matrix_type=symmetry_string, &
                           row_blk_size=row_blk_size, &
                           col_blk_size=col_blk_size, &
                           nze=0)
      DEALLOCATE (col_blk_size)

      CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                   nhalos=nhalos, &
                                   max_nhalos=max_nhalos, &
                                   halos=halos)

      ! create dummy empty atomic halo
      CALL fb_atomic_halo_create(dummy_atomic_halo)
      ALLOCATE (dummy_halo_atoms(0))
      CALL fb_atomic_halo_set(atomic_halo=dummy_atomic_halo, &
                              owner_atom=0, &
                              owner_id_in_halo=0, &
                              natoms=0, &
                              halo_atoms=dummy_halo_atoms, &
                              nelectrons=0, &
                              sorted=.TRUE.)

      send_data_only = .FALSE.

      DO ihalo = 1, max_nhalos
         IF (ihalo > nhalos) THEN
            send_data_only = .TRUE.
         END IF
         ! construct the filter matrix block by block
         IF (send_data_only) THEN
            CALL fb_fltrmat_add_blkcol(H_mat, &
                                       S_mat, &
                                       dummy_atomic_halo, &
                                       trial_fns, &
                                       para_env, &
                                       particle_set, &
                                       fermi_level, &
                                       filter_temp, &
                                       filter_mat, &
                                       tolerance)
         ELSE
            CALL fb_fltrmat_add_blkcol(H_mat, &
                                       S_mat, &
                                       halos(ihalo), &
                                       trial_fns, &
                                       para_env, &
                                       particle_set, &
                                       fermi_level, &
                                       filter_temp, &
                                       filter_mat, &
                                       tolerance)
         END IF ! send_data_only
      END DO

      ! finalise the filter matrix
      CALL cp_dbcsr_finalize(filter_mat)

      ! cleanup
      CALL fb_atomic_halo_release(dummy_atomic_halo)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_build

! **************************************************************************************************
!> \brief Build the filter matrix, with MPI communications grouped together.
!>        More effcient on communication, less efficient on memory (compared
!>        to fb_fltrmat_build)
!> \param H_mat : DBCSR system KS matrix
!> \param S_mat : DBCSR system overlap matrix
!> \param atomic_halos : list of all local atomic halos, each halo gives
!>                       one atomic matrix and contributes to one blk
!>                       col to the filter matrix
!> \param trial_fns : the trial functions to be used to shrink the
!>                     size of the new "filtered" basis
!> \param para_env : cp2k parallel environment
!> \param particle_set : set of all particles in the system
!> \param fermi_level : the fermi level used for defining the filter
!>                      function, which is a Fermi-Dirac distribution
!>                      function
!> \param filter_temp : the filter temperature used for defining the
!>                      filter function
!> \param name        : name given to the filter matrix
!> \param filter_mat  : DBCSR format filter matrix
!> \param tolerance   : anything less than tolerance is treated as zero
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_build_2(H_mat, &
                                 S_mat, &
                                 atomic_halos, &
                                 trial_fns, &
                                 para_env, &
                                 particle_set, &
                                 fermi_level, &
                                 filter_temp, &
                                 name, &
                                 filter_mat, &
                                 tolerance)
      TYPE(cp_dbcsr_type), POINTER                       :: H_mat, S_mat
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: fermi_level, filter_temp
      CHARACTER(LEN=*), INTENT(IN)                       :: name
      TYPE(cp_dbcsr_type), POINTER                       :: filter_mat
      REAL(KIND=dp), INTENT(IN)                          :: tolerance

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_build_2', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_string
      INTEGER                                            :: handle, iblkcol, ihalo, ikind, &
                                                            natoms_global, natoms_in_halo, &
                                                            nblkcols_total, nblks_recv, nhalos, &
                                                            nmax
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, ntfns, row_blk_size
      LOGICAL                                            :: check_ok
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(dbcsr_distribution_obj)                       :: dbcsr_dist
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos
      TYPE(fb_com_atom_pairs_obj)                        :: atmatrix_blks_recv, atmatrix_blks_send, &
                                                            filter_mat_blks_recv, &
                                                            filter_mat_blks_send
      TYPE(fb_matrix_data_obj)                           :: filter_mat_data, H_mat_data, S_mat_data

      CALL timeset(routineN, handle)

      NULLIFY (halos, atomic_kind, row_blk_size, col_blk_size, ntfns)

      ! filter_mat must be of a dissassociated status (i.e. brand new)
      check_ok = .NOT. ASSOCIATED(filter_mat)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_filter_matrix_methods.F",295)

      ! get total number of atoms
      natoms_global = SIZE(particle_set)

      ! get trial function information
      CALL fb_trial_fns_get(trial_fns=trial_fns, &
                            nfunctions=ntfns)

      ! calculate the row_blk_size and col_blk_size arrays for
      ! constructing the filter matrix in DBCSR format
      ! row_blk_size for the filter matrix is the same as H or S
      CALL cp_dbcsr_get_info(H_mat, &
                             nblkcols_total=nblkcols_total, &
                             row_blk_size=row_blk_size, &
                             distribution=dbcsr_dist)
      ALLOCATE (col_blk_size(nblkcols_total))
      col_blk_size = 0
      DO iblkcol = 1, nblkcols_total
         atomic_kind => particle_set(iblkcol)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              kind_number=ikind)
         col_blk_size(iblkcol) = ntfns(ikind)
      END DO
      ! DO NOT deallocate cbs if gift=.TRUE. as col_blk_sizes will only point to cbs
      name_string = name
      CALL compress(name_string)
      CALL uppercase(name_string)
      ! create empty filter matrix (it is always non-symmetric as it is non-square)
      ALLOCATE (filter_mat)
      CALL cp_dbcsr_init(filter_mat)
      CALL cp_dbcsr_create(matrix=filter_mat, &
                           name=name_string, &
                           dist=dbcsr_dist, &
                           matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=row_blk_size, &
                           col_blk_size=col_blk_size, &
                           nze=0)
      DEALLOCATE (col_blk_size)

      ! get all the blocks required for constructing atomic matrics, and
      ! store it in a fb_matrix_data object
      CALL fb_matrix_data_nullify(H_mat_data)
      CALL fb_matrix_data_nullify(S_mat_data)
      CALL fb_com_atom_pairs_nullify(atmatrix_blks_send)
      CALL fb_com_atom_pairs_nullify(atmatrix_blks_recv)
      CALL fb_com_atom_pairs_create(atmatrix_blks_send)
      CALL fb_com_atom_pairs_create(atmatrix_blks_recv)
      ! H matrix
      CALL fb_atmatrix_generate_com_pairs_2(H_mat, &
                                            atomic_halos, &
                                            para_env, &
                                            atmatrix_blks_send, &
                                            atmatrix_blks_recv)
      CALL fb_com_atom_pairs_get(atom_pairs=atmatrix_blks_recv, &
                                 npairs=nblks_recv)
      CALL fb_matrix_data_create(H_mat_data, &
                                 nblks_recv, &
                                 natoms_global)
      CALL fb_com_atom_pairs_gather_blks(H_mat, &
                                         atmatrix_blks_send, &
                                         atmatrix_blks_recv, &
                                         para_env, &
                                         H_mat_data)
      ! S matrix
      CALL fb_atmatrix_generate_com_pairs_2(S_mat, &
                                            atomic_halos, &
                                            para_env, &
                                            atmatrix_blks_send, &
                                            atmatrix_blks_recv)
      CALL fb_com_atom_pairs_get(atom_pairs=atmatrix_blks_recv, &
                                 npairs=nblks_recv)
      CALL fb_matrix_data_create(S_mat_data, &
                                 nblks_recv, &
                                 natoms_global)
      CALL fb_com_atom_pairs_gather_blks(S_mat, &
                                         atmatrix_blks_send, &
                                         atmatrix_blks_recv, &
                                         para_env, &
                                         S_mat_data)
      ! cleanup
      CALL fb_com_atom_pairs_release(atmatrix_blks_send)
      CALL fb_com_atom_pairs_release(atmatrix_blks_recv)

      ! make filter matrix blocks one by one and store in an
      ! matrix_data_obj
      CALL fb_matrix_data_nullify(filter_mat_data)
      CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                   nhalos=nhalos, &
                                   halos=halos)
      nmax = 0
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 natoms=natoms_in_halo)
         nmax = nmax+natoms_in_halo
      END DO
      CALL fb_matrix_data_create(filter_mat_data, &
                                 nmax, &
                                 natoms_global)
      DO ihalo = 1, nhalos
         CALL fb_fltrmat_add_blkcol_2(H_mat, &
                                      S_mat, &
                                      H_mat_data, &
                                      S_mat_data, &
                                      halos(ihalo), &
                                      trial_fns, &
                                      particle_set, &
                                      fermi_level, &
                                      filter_temp, &
                                      filter_mat_data, &
                                      tolerance)
      END DO
      ! clean up
      CALL fb_matrix_data_release(H_mat_data)
      CALL fb_matrix_data_release(S_mat_data)

      ! distribute the relevant blocks from the matrix_data_obj to DBCSR
      ! filter matrix
      CALL fb_com_atom_pairs_nullify(filter_mat_blks_send)
      CALL fb_com_atom_pairs_nullify(filter_mat_blks_recv)
      CALL fb_com_atom_pairs_create(filter_mat_blks_send)
      CALL fb_com_atom_pairs_create(filter_mat_blks_recv)
      CALL fb_fltrmat_generate_com_pairs_2(filter_mat, &
                                           atomic_halos, &
                                           para_env, &
                                           filter_mat_blks_send, &
                                           filter_mat_blks_recv)
      CALL fb_com_atom_pairs_distribute_blks(filter_mat_data, &
                                             filter_mat_blks_send, &
                                             filter_mat_blks_recv, &
                                             para_env, &
                                             filter_mat)
      ! cleanup
      CALL fb_com_atom_pairs_release(filter_mat_blks_send)
      CALL fb_com_atom_pairs_release(filter_mat_blks_recv)
      CALL fb_matrix_data_release(filter_mat_data)

      ! finalise matrix
      CALL cp_dbcsr_finalize(filter_mat)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_build_2

! **************************************************************************************************
!> \brief Add a computed blocks in one column to the filter matrix. This
!>        version is used by fb_fltrmat_build, for the case where MPI
!>        communications are done at each step
!>        It does not finalise the filter matrix
!> \param H_mat : DBCSR system KS matrix
!> \param S_mat : DBCSR system overlap matrix
!> \param atomic_halo :  the halo that contributes to the blk
!>                       col of the filter matrix
!> \param trial_fns ...
!> \param para_env : cp2k parallel environment
!> \param particle_set : set of all particles in the system
!> \param fermi_level : the fermi level used for defining the filter
!>                      function, which is a Fermi-Dirac distribution
!>                      function
!> \param filter_temp : the filter temperature used for defining the
!>                      filter function
!> \param filter_mat  : DBCSR format filter matrix
!> \param tolerance   : anything smaller than tolerance is treated as zero
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_add_blkcol(H_mat, &
                                    S_mat, &
                                    atomic_halo, &
                                    trial_fns, &
                                    para_env, &
                                    particle_set, &
                                    fermi_level, &
                                    filter_temp, &
                                    filter_mat, &
                                    tolerance)
      TYPE(cp_dbcsr_type), POINTER                       :: H_mat, S_mat
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: fermi_level, filter_temp
      TYPE(cp_dbcsr_type), POINTER                       :: filter_mat
      REAL(KIND=dp), INTENT(IN)                          :: tolerance

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_add_blkcol', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, handle_mpi, iatom_global, iatom_in_halo, ind, ipair, ipe, itrial, &
         jatom_global, jatom_in_halo, jkind, natoms_global, natoms_in_halo, ncols_atmatrix, &
         ncols_blk, nrows_atmatrix, nrows_blk, numprocs, pe, recv_encode, send_encode, stat
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: pairs_recv, pairs_send
      INTEGER, ALLOCATABLE, DIMENSION(:) :: atomic_H_blk_col_start, atomic_H_blk_row_start, &
         atomic_S_blk_col_start, atomic_S_blk_row_start, col_block_size_data, ind_in_halo, &
         recv_disps, recv_pair_count, recv_pair_disps, recv_sizes, send_disps, send_pair_count, &
         send_pair_disps, send_sizes
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms, ntfns, row_block_size_data
      INTEGER, DIMENSION(:, :), POINTER                  :: tfns
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_buf, send_buf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: atomic_filter_mat, atomic_H, atomic_S
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fb_com_atom_pairs_obj)                        :: com_pairs_recv, com_pairs_send

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind, halo_atoms, ntfns, pairs_send, pairs_recv, &
               row_block_size_data, tfns)
      CALL fb_com_atom_pairs_nullify(com_pairs_send)
      CALL fb_com_atom_pairs_nullify(com_pairs_recv)

      ! ----------------------------------------------------------------------
      ! Get communication buffers ready
      ! ----------------------------------------------------------------------

      ! generate send and recv atom pairs
      CALL fb_com_atom_pairs_create(com_pairs_send)
      CALL fb_com_atom_pairs_create(com_pairs_recv)
      CALL fb_fltrmat_generate_com_pairs(filter_mat, &
                                         atomic_halo, &
                                         para_env, &
                                         com_pairs_send, &
                                         com_pairs_recv)
      CALL fb_com_atom_pairs_get(atom_pairs=com_pairs_send, &
                                 natoms_encode=send_encode, &
                                 pairs=pairs_send)
      CALL fb_com_atom_pairs_get(atom_pairs=com_pairs_recv, &
                                 natoms_encode=recv_encode, &
                                 pairs=pairs_recv)

      ! get para_env info
      numprocs = para_env%num_pe
      ! me = para_env%mepos + 1   ! my process id, starting counting from 1

      ! obtain trail function information
      CALL fb_trial_fns_get(trial_fns=trial_fns, &
                            nfunctions=ntfns, &
                            functions=tfns)

      ! obtain row and col block size data for filter matrix
      row_block_size_data => cp_dbcsr_row_block_sizes(H_mat)
      natoms_global = SIZE(particle_set)
      ALLOCATE (col_block_size_data(natoms_global))
      DO jatom_global = 1, natoms_global
         atomic_kind => particle_set(jatom_global)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, kind_number=jkind)
         col_block_size_data(jatom_global) = ntfns(jkind)
      END DO

      ! allocate temporary arrays for send
      ALLOCATE (send_sizes(numprocs))
      ALLOCATE (send_disps(numprocs))
      ALLOCATE (send_pair_count(numprocs))
      ALLOCATE (send_pair_disps(numprocs))
      ! setup send buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(com_pairs_send, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               send_sizes, &
                                               send_disps, &
                                               send_pair_count, &
                                               send_pair_disps)
      ! allocate send buffer
      ALLOCATE (send_buf(SUM(send_sizes)))

      ! allocate temporary array for recv
      ALLOCATE (recv_sizes(numprocs))
      ALLOCATE (recv_disps(numprocs))
      ALLOCATE (recv_pair_count(numprocs))
      ALLOCATE (recv_pair_disps(numprocs))
      ! setup recv buffer sizes
      CALL fb_com_atom_pairs_calc_buffer_sizes(com_pairs_recv, &
                                               numprocs, &
                                               row_block_size_data, &
                                               col_block_size_data, &
                                               recv_sizes, &
                                               recv_disps, &
                                               recv_pair_count, &
                                               recv_pair_disps)
      ! allocate recv buffer
      ALLOCATE (recv_buf(SUM(recv_sizes)))

      ! ----------------------------------------------------------------------
      ! Construct atomic filter matrix for this atomic_halo
      ! ----------------------------------------------------------------------

      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)

      ! construct atomic matrix for H for atomic_halo
      ALLOCATE (atomic_H_blk_row_start(natoms_in_halo+1), &
                atomic_H_blk_col_start(natoms_in_halo+1), &
                STAT=stat)
      IF(.NOT.(stat == 0))CALL cp__a("qs_fb_filter_matrix_methods.F",588)
      CALL fb_atmatrix_calc_size(H_mat, &
                                 atomic_halo, &
                                 nrows_atmatrix, &
                                 ncols_atmatrix, &
                                 atomic_H_blk_row_start, &
                                 atomic_H_blk_col_start)

      ALLOCATE (atomic_H(nrows_atmatrix, ncols_atmatrix))
      CALL fb_atmatrix_construct(H_mat, &
                                 atomic_halo, &
                                 para_env, &
                                 atomic_H, &
                                 atomic_H_blk_row_start, &
                                 atomic_H_blk_col_start)

      ! construct atomic matrix for S for atomic_halo
      ALLOCATE (atomic_S_blk_row_start(natoms_in_halo+1), &
                atomic_S_blk_col_start(natoms_in_halo+1), &
                STAT=stat)
      IF(.NOT.(stat == 0))CALL cp__a("qs_fb_filter_matrix_methods.F",608)
      CALL fb_atmatrix_calc_size(S_mat, &
                                 atomic_halo, &
                                 nrows_atmatrix, &
                                 ncols_atmatrix, &
                                 atomic_S_blk_row_start, &
                                 atomic_S_blk_col_start)
      ALLOCATE (atomic_S(nrows_atmatrix, ncols_atmatrix))
      CALL fb_atmatrix_construct(S_mat, &
                                 atomic_halo, &
                                 para_env, &
                                 atomic_S, &
                                 atomic_S_blk_row_start, &
                                 atomic_S_blk_col_start)

      ! construct the atomic filter matrix
      ALLOCATE (atomic_filter_mat(nrows_atmatrix, ncols_atmatrix))
      ! calculate atomic filter matrix only if it is non-zero sized
      IF (nrows_atmatrix > 0 .AND. ncols_atmatrix > 0) THEN
         CALL fb_fltrmat_build_atomic_fltrmat(atomic_H, &
                                              atomic_S, &
                                              fermi_level, &
                                              filter_temp, &
                                              atomic_filter_mat, &
                                              tolerance)
      END IF

      ! ----------------------------------------------------------------------
      ! Construct filter matrix blocks and add to the correct locations
      ! in send_buffer
      ! ----------------------------------------------------------------------

      ! preconstruct iatom_global to iatom_in_halo map
      ALLOCATE (ind_in_halo(natoms_global))
      ind_in_halo = 0
      DO iatom_in_halo = 1, natoms_in_halo
         iatom_global = halo_atoms(iatom_in_halo)
         ind_in_halo(iatom_global) = iatom_in_halo
      END DO

      ! initialise send buffer
      IF (SIZE(send_buf) > 0) send_buf = 0.0_dp
      ! assign values
      DO ipe = 1, numprocs
         send_sizes(ipe) = 0
         DO ipair = 1, send_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_send(send_pair_disps(ipe)+ipair), &
                                          pe, iatom_global, jatom_global, &
                                          send_encode)
            iatom_in_halo = ind_in_halo(iatom_global)
            IF(.NOT.(iatom_in_halo > 0))CALL cp__a("qs_fb_filter_matrix_methods.F",658)
            jatom_in_halo = ind_in_halo(jatom_global)
            IF(.NOT.(jatom_in_halo > 0))CALL cp__a("qs_fb_filter_matrix_methods.F",660)
            atomic_kind => particle_set(jatom_global)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 kind_number=jkind)
            nrows_blk = row_block_size_data(iatom_global)
            ncols_blk = ntfns(jkind)

            ! do it column-wise one trial function at a time
            DO itrial = 1, ntfns(jkind)
               ind = send_disps(ipe)+send_sizes(ipe)+(itrial-1)*nrows_blk
               CALL dgemv("N", &
                          nrows_blk, &
                          ncols_atmatrix, &
                          1.0_dp, &
                          atomic_filter_mat( &
                          atomic_H_blk_row_start(iatom_in_halo): &
                          atomic_H_blk_row_start(iatom_in_halo+1)-1, &
                          1:ncols_atmatrix &
                          ), &
                          nrows_blk, &
                          atomic_S( &
                          1:nrows_atmatrix, &
                          atomic_S_blk_col_start(jatom_in_halo)+ &
                          tfns(itrial, jkind)-1 &
                          ), &
                          1, &
                          0.0_dp, &
                          send_buf(ind+1:ind+nrows_blk), &
                          1)
            END DO ! itrial
            send_sizes(ipe) = send_sizes(ipe)+nrows_blk*ncols_blk
         END DO ! ipair
      END DO ! ipe

      DEALLOCATE (atomic_H)
      DEALLOCATE (atomic_H_blk_row_start)
      DEALLOCATE (atomic_S)
      DEALLOCATE (atomic_S_blk_row_start)
      DEALLOCATE (atomic_filter_mat)
      DEALLOCATE (ind_in_halo)

      ! ----------------------------------------------------------------------
      ! Do communication
      ! ----------------------------------------------------------------------

      CALL timeset("fb_fltrmat_add_blkcol_mpi", handle_mpi)

      CALL mp_alltoall(send_buf, send_sizes, send_disps, &
                       recv_buf, recv_sizes, recv_disps, &
                       para_env%group)

      CALL timestop(handle_mpi)

      DEALLOCATE (send_buf)
      DEALLOCATE (send_sizes)
      DEALLOCATE (send_disps)
      DEALLOCATE (send_pair_count)
      DEALLOCATE (send_pair_disps)

      ! ----------------------------------------------------------------------
      ! Unpack the recv buffer and add the blocks to correct parts of
      ! the DBCSR filter matrix
      ! ----------------------------------------------------------------------

      DO ipe = 1, numprocs
         recv_sizes(ipe) = 0
         DO ipair = 1, recv_pair_count(ipe)
            CALL fb_com_atom_pairs_decode(pairs_recv(recv_pair_disps(ipe)+ipair), &
                                          pe, iatom_global, jatom_global, &
                                          recv_encode)
            nrows_blk = row_block_size_data(iatom_global)
            ncols_blk = col_block_size_data(jatom_global)
            ind = recv_disps(ipe)+recv_sizes(ipe)
            CALL cp_dbcsr_put_block(filter_mat, &
                                    iatom_global, jatom_global, &
                                    recv_buf((ind+1):(ind+nrows_blk*ncols_blk)))
            recv_sizes(ipe) = recv_sizes(ipe)+nrows_blk*ncols_blk
         END DO ! ipair
      END DO ! ipe

      ! cleanup rest of the temporary arrays
      DEALLOCATE (recv_buf)
      DEALLOCATE (recv_sizes)
      DEALLOCATE (recv_pair_count)
      DEALLOCATE (recv_pair_disps)

      CALL fb_com_atom_pairs_release(com_pairs_send)
      CALL fb_com_atom_pairs_release(com_pairs_recv)

      ! cannot finalise the matrix until all blocks has been added

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_add_blkcol

! **************************************************************************************************
!> \brief Computed blocks in one filter matrix column. This version is used by
!>        fb_fltrmat_build_2, where MPI communication is done collectively
!> \param H_mat : DBCSR system KS matrix
!> \param S_mat : DBCSR system overlap matrix
!> \param H_mat_data  :  local storage of the relevant H_mat matrix blocks
!> \param S_mat_data  :  local storage of the relevant S_mat matrix blocks
!> \param atomic_halo :  the halo that contributes to the blk
!>                       col of the filter matrix
!> \param trial_fns   :  trial functions data
!> \param particle_set : set of all particles in the system
!> \param fermi_level : the fermi level used for defining the filter
!>                      function, which is a Fermi-Dirac distribution
!>                      function
!> \param filter_temp : the filter temperature used for defining the
!>                      filter function
!> \param filter_mat_data : local storage for the the computed filter matrix
!>                          blocks
!> \param tolerance : anything less than this is regarded as zero
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_add_blkcol_2(H_mat, &
                                      S_mat, &
                                      H_mat_data, &
                                      S_mat_data, &
                                      atomic_halo, &
                                      trial_fns, &
                                      particle_set, &
                                      fermi_level, &
                                      filter_temp, &
                                      filter_mat_data, &
                                      tolerance)
      TYPE(cp_dbcsr_type), POINTER                       :: H_mat, S_mat
      TYPE(fb_matrix_data_obj), INTENT(IN)               :: H_mat_data, S_mat_data
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: fermi_level, filter_temp
      TYPE(fb_matrix_data_obj), INTENT(INOUT)            :: filter_mat_data
      REAL(KIND=dp), INTENT(IN)                          :: tolerance

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_add_blkcol_2', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iatom_global, iatom_in_halo, itrial, jatom_global, jatom_in_halo, jkind, &
         natoms_global, natoms_in_halo, ncols_atmatrix, ncols_blk, ncols_blk_max, nrows_atmatrix, &
         nrows_blk, nrows_blk_max, stat
      INTEGER, ALLOCATABLE, DIMENSION(:) :: atomic_H_blk_col_start, atomic_H_blk_row_start, &
         atomic_S_blk_col_start, atomic_S_blk_row_start, col_block_size_data
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms, ntfns, row_block_size_data
      INTEGER, DIMENSION(:, :), POINTER                  :: tfns
      LOGICAL                                            :: check_ok
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: atomic_filter_mat, atomic_H, atomic_S, &
                                                            mat_blk
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind, halo_atoms, ntfns, row_block_size_data, tfns)

      check_ok = fb_matrix_data_has_data(H_mat_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_filter_matrix_methods.F",816)
      check_ok = fb_matrix_data_has_data(S_mat_data)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_filter_matrix_methods.F",818)

      ! obtain trial function information
      CALL fb_trial_fns_get(trial_fns=trial_fns, &
                            nfunctions=ntfns, &
                            functions=tfns)

      ! obtain row and col block size data for filter matrix
      row_block_size_data => cp_dbcsr_row_block_sizes(H_mat)
      natoms_global = SIZE(particle_set)
      ALLOCATE (col_block_size_data(natoms_global))
      DO jatom_global = 1, natoms_global
         atomic_kind => particle_set(jatom_global)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, kind_number=jkind)
         col_block_size_data(jatom_global) = ntfns(jkind)
      END DO

      ! ----------------------------------------------------------------------
      ! Construct atomic filter matrix for this atomic_halo
      ! ----------------------------------------------------------------------

      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)

      ! construct atomic matrix for H for atomic_halo
      ALLOCATE (atomic_H_blk_row_start(natoms_in_halo+1), &
                atomic_H_blk_col_start(natoms_in_halo+1), &
                STAT=stat)
      IF(.NOT.(stat == 0))CALL cp__a("qs_fb_filter_matrix_methods.F",847)
      CALL fb_atmatrix_calc_size(H_mat, &
                                 atomic_halo, &
                                 nrows_atmatrix, &
                                 ncols_atmatrix, &
                                 atomic_H_blk_row_start, &
                                 atomic_H_blk_col_start)
      ALLOCATE (atomic_H(nrows_atmatrix, ncols_atmatrix))
      CALL fb_atmatrix_construct_2(H_mat_data, &
                                   atomic_halo, &
                                   atomic_H, &
                                   atomic_H_blk_row_start, &
                                   atomic_H_blk_col_start)

      ! construct atomic matrix for S for atomic_halo
      ALLOCATE (atomic_S_blk_row_start(natoms_in_halo+1), &
                atomic_S_blk_col_start(natoms_in_halo+1), &
                STAT=stat)
      IF(.NOT.(stat == 0))CALL cp__a("qs_fb_filter_matrix_methods.F",865)
      CALL fb_atmatrix_calc_size(S_mat, &
                                 atomic_halo, &
                                 nrows_atmatrix, &
                                 ncols_atmatrix, &
                                 atomic_S_blk_row_start, &
                                 atomic_S_blk_col_start)
      ALLOCATE (atomic_S(nrows_atmatrix, ncols_atmatrix))
      CALL fb_atmatrix_construct_2(S_mat_data, &
                                   atomic_halo, &
                                   atomic_S, &
                                   atomic_S_blk_row_start, &
                                   atomic_S_blk_col_start)

      ! construct the atomic filter matrix
      ALLOCATE (atomic_filter_mat(nrows_atmatrix, ncols_atmatrix))
      ! calculate atomic filter matrix only if it is non-zero sized
      IF (nrows_atmatrix > 0 .AND. ncols_atmatrix > 0) THEN
         CALL fb_fltrmat_build_atomic_fltrmat(atomic_H, &
                                              atomic_S, &
                                              fermi_level, &
                                              filter_temp, &
                                              atomic_filter_mat, &
                                              tolerance)
      END IF

      ! ----------------------------------------------------------------------
      ! Construct filter matrix block and add to filter_mat_data
      ! ----------------------------------------------------------------------

      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              owner_atom=jatom_global, &
                              owner_id_in_halo=jatom_in_halo)
      nrows_blk_max = MAXVAL(row_block_size_data)
      ncols_blk_max = MAXVAL(ntfns)
      ALLOCATE (mat_blk(nrows_blk_max, ncols_blk_max))
      mat_blk(:, :) = 0.0_dp
      DO iatom_in_halo = 1, natoms_in_halo
         iatom_global = halo_atoms(iatom_in_halo)
         atomic_kind => particle_set(jatom_global)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              kind_number=jkind)
         nrows_blk = row_block_size_data(iatom_global)
         ncols_blk = ntfns(jkind)

         ! ALLOCATE(mat_blk(nrows_blk,ncols_blk) STAT=stat)
         ! CPPostcondition(stat==0, cp_failure_level, routineP,failure)

         ! do it column-wise one trial function at a time
         DO itrial = 1, ntfns(jkind)
            CALL dgemv("N", &
                       nrows_blk, &
                       ncols_atmatrix, &
                       1.0_dp, &
                       atomic_filter_mat( &
                       atomic_H_blk_row_start(iatom_in_halo): &
                       atomic_H_blk_row_start(iatom_in_halo+1)-1, &
                       1:ncols_atmatrix &
                       ), &
                       nrows_blk, &
                       atomic_S( &
                       1:nrows_atmatrix, &
                       atomic_S_blk_col_start(jatom_in_halo)+ &
                       tfns(itrial, jkind)-1 &
                       ), &
                       1, &
                       0.0_dp, &
                       mat_blk( &
                       1:nrows_blk, &
                       itrial), &
                       1)
         END DO ! itrial
         CALL fb_matrix_data_add(filter_mat_data, &
                                 iatom_global, &
                                 jatom_global, &
                                 mat_blk(1:nrows_blk, 1:ncols_blk))
         ! DEALLOCATE(mat_blk, STAT=stat)
         ! CPPostcondition(stat==0, cp_failure_level, routineP,failure)
      END DO ! iatom_in_halo
      DEALLOCATE (mat_blk)

      ! clean up
      DEALLOCATE (atomic_H)
      DEALLOCATE (atomic_H_blk_row_start)
      DEALLOCATE (atomic_S)
      DEALLOCATE (atomic_S_blk_row_start)
      DEALLOCATE (atomic_filter_mat)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_add_blkcol_2

! **************************************************************************************************
!> \brief generate the list of blocks (atom pairs) to be sent and received
!>        in order to construct the filter matrix for each atomic halo.
!>        This version is for use with fb_fltrmat_build, where MPI
!>        communications are done at each step
!> \param filter_mat : DBCSR formated filter matrix
!> \param atomic_halo :  the halo that contributes to a blk
!>                       col of the filter matrix
!> \param para_env : cp2k parallel environment
!> \param atom_pairs_send : list of blocks to be sent
!> \param atom_pairs_recv : list of blocks to be received
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_generate_com_pairs(filter_mat, &
                                            atomic_halo, &
                                            para_env, &
                                            atom_pairs_send, &
                                            atom_pairs_recv)
      TYPE(cp_dbcsr_type), POINTER                       :: filter_mat
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs_send, atom_pairs_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_generate_com_pairs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dest, handle, iatom_global, &
                                                            iatom_in_halo, itask, jatom_global, &
                                                            natoms_in_halo, nblkrows_total, &
                                                            ntasks_send, src
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_send
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      TYPE(fb_com_tasks_obj)                             :: com_tasks_recv, com_tasks_send

      CALL timeset(routineN, handle)

      NULLIFY (tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_recv)

      ! initialise atom_pairs_send and atom_pairs_recv
      IF (fb_com_atom_pairs_has_data(atom_pairs_send)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_send)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_send)
      END IF
      IF (fb_com_atom_pairs_has_data(atom_pairs_recv)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_recv)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_recv)
      END IF

      ! source is always the local processor
      src = para_env%mepos

      ! The total number of filter matrix blocks each processor is going
      ! to construct equals to the total number of halo atoms in all of
      ! the atomic halos local to the processor. The number of send
      ! tasks will not exceed this. We do one halo (col) at a time, and
      ! each call of this subroutine will only work on one filter matrix
      ! col corresponding to atomic_halo.

      ! The col atom block index for each filter matrix block are the
      ! owner atom of each halo. The row atom block index for each
      ! filter matrix block corresponding to each col are the halo atoms
      ! of the corresponding halos. Filter matrix is non-symmetric: it
      ! is non-square, because the blocks themselves are non-sqaure

      CALL fb_atomic_halo_get(atomic_halo=atomic_halo, &
                              owner_atom=jatom_global, &
                              natoms=natoms_in_halo, &
                              halo_atoms=halo_atoms)
      ntasks_send = natoms_in_halo

      ! allocate send tasks
      ALLOCATE (tasks_send(TASK_N_RECORDS, ntasks_send))

      ! Get the total number of atoms, this can be obtained from the
      ! total number of block rows in the DBCSR filter matrix.  We
      ! assumes that before calling this subroutine, the filter_mat has
      ! already been created and initialised: i.e. using
      ! cp_dbcsr_create_new. Even if the matrix is at the moment empty,
      ! the attribute nblkrows_total is already assigned from the dbcsr
      ! distribution data
      CALL cp_dbcsr_get_info(filter_mat, &
                             nblkrows_total=nblkrows_total)

      ! construct send tasks
      itask = 1
      DO iatom_in_halo = 1, natoms_in_halo
         iatom_global = halo_atoms(iatom_in_halo)
         ! find where the constructed block of filter matrix belongs to
         CALL cp_dbcsr_get_stored_coordinates(filter_mat, &
                                              iatom_global, &
                                              jatom_global, &
                                              processor=dest)
         ! create the send tasks
         tasks_send(TASK_DEST, itask) = dest
         tasks_send(TASK_SRC, itask) = src
         CALL fb_com_tasks_encode_pair(tasks_send(TASK_PAIR, itask), &
                                       iatom_global, jatom_global, &
                                       nblkrows_total)
         ! calculation of cost not implemented at the moment
         tasks_send(TASK_COST, itask) = 0
         itask = itask+1
      END DO ! iatom_in_halo

      CALL fb_com_tasks_create(com_tasks_recv)
      CALL fb_com_tasks_create(com_tasks_send)

      CALL fb_com_tasks_set(com_tasks=com_tasks_send, &
                            task_dim=TASK_N_RECORDS, &
                            ntasks=ntasks_send, &
                            nencode=nblkrows_total, &
                            tasks=tasks_send)

      ! generate the recv task list (tasks_recv) from the send task list
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, "<", com_tasks_send, &
                                           para_env)

      ! task lists are now complete, now construct the atom_pairs_send
      ! and atom_pairs_recv from the tasks lists
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_send, &
                                         atom_pairs=atom_pairs_send, &
                                         natoms_encode=nblkrows_total, &
                                         send_or_recv="send")
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_recv, &
                                         atom_pairs=atom_pairs_recv, &
                                         natoms_encode=nblkrows_total, &
                                         send_or_recv="recv")

      ! cleanup
      CALL fb_com_tasks_release(com_tasks_recv)
      CALL fb_com_tasks_release(com_tasks_send)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_generate_com_pairs

! **************************************************************************************************
!> \brief generate the list of blocks (atom pairs) to be sent and received
!>        in order to construct the filter matrix for each atomic halo.
!>        This vesion is for use with fb_fltrmat_build_2, where MPI
!>        communications are done collectively.
!> \param filter_mat  : DBCSR formated filter matrix
!> \param atomic_halos : set of all local atomic halos contributing to the
!>                       filter matrix
!> \param para_env : cp2k parallel environment
!> \param atom_pairs_send : list of blocks to be sent
!> \param atom_pairs_recv : list of blocks to be received
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_generate_com_pairs_2(filter_mat, &
                                              atomic_halos, &
                                              para_env, &
                                              atom_pairs_send, &
                                              atom_pairs_recv)
      TYPE(cp_dbcsr_type), POINTER                       :: filter_mat
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_com_atom_pairs_obj), INTENT(INOUT)         :: atom_pairs_send, atom_pairs_recv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_generate_com_pairs_2', &
         routineP = moduleN//':'//routineN

      INTEGER :: dest, handle, iatom_global, iatom_in_halo, iatom_stored, ihalo, itask, &
         jatom_global, jatom_stored, natoms_in_halo, nblkrows_total, nhalos, ntasks_send, src
      INTEGER(KIND=int_8), DIMENSION(:, :), POINTER      :: tasks_send
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      LOGICAL                                            :: transpose
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos
      TYPE(fb_com_tasks_obj)                             :: com_tasks_recv, com_tasks_send

      CALL timeset(routineN, handle)

      NULLIFY (tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_send)
      CALL fb_com_tasks_nullify(com_tasks_recv)

      ! initialise atom_pairs_send and atom_pairs_recv
      IF (fb_com_atom_pairs_has_data(atom_pairs_send)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_send)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_send)
      END IF
      IF (fb_com_atom_pairs_has_data(atom_pairs_recv)) THEN
         CALL fb_com_atom_pairs_init(atom_pairs_recv)
      ELSE
         CALL fb_com_atom_pairs_create(atom_pairs_recv)
      END IF

      ! source is always the local processor
      src = para_env%mepos

      ! The col atom block index for each filter matrix block are the
      ! owner atom of each halo. The row atom block index for each
      ! filter matrix block corresponding to each col are the halo atoms
      ! of the corresponding halos. Filter matrix is non-symmetric: it
      ! is non-square, because the blocks themselves are non-sqaure

      CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                   nhalos=nhalos, &
                                   halos=halos)

      ! estimate the maximum number of blocks (i.e. atom paris) to send
      ntasks_send = 0
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 natoms=natoms_in_halo)
         ntasks_send = ntasks_send+natoms_in_halo
      END DO ! ihalo

      ! allocate send tasks
      ALLOCATE (tasks_send(TASK_N_RECORDS, ntasks_send))

      ! Get the total number of atoms. This can be obtained from the
      ! total number of block rows in the DBCSR filter matrix.  We
      ! assumes that before calling this subroutine, the filter_mat has
      ! already been created and initialised: i.e. using
      ! cp_dbcsr_create_new. Even if the matrix is at the moment empty,
      ! the attribute nblkrows_total is already assigned from the dbcsr
      ! distribution data
      CALL cp_dbcsr_get_info(filter_mat, &
                             nblkrows_total=nblkrows_total)

      ! construct send tasks
      itask = 1
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 owner_atom=jatom_global, &
                                 natoms=natoms_in_halo, &
                                 halo_atoms=halo_atoms)
         DO iatom_in_halo = 1, natoms_in_halo
            iatom_global = halo_atoms(iatom_in_halo)
            iatom_stored = iatom_global
            jatom_stored = jatom_global
            transpose = .FALSE.
            ! find where the constructed block of filter matrix belongs to
            CALL cp_dbcsr_get_stored_coordinates(filter_mat, &
                                                 iatom_stored, &
                                                 jatom_stored, &
                                                 processor=dest)
            ! create the send tasks
            tasks_send(TASK_DEST, itask) = dest
            tasks_send(TASK_SRC, itask) = src
            CALL fb_com_tasks_encode_pair(tasks_send(TASK_PAIR, itask), &
                                          iatom_global, jatom_global, &
                                          nblkrows_total)
            ! calculation of cost not implemented at the moment
            tasks_send(TASK_COST, itask) = 0
            itask = itask+1
         END DO ! iatom_in_halo
      END DO ! ihalo

      ! get the actual number of tasks
      ntasks_send = itask-1

      CALL fb_com_tasks_create(com_tasks_send)
      CALL fb_com_tasks_set(com_tasks=com_tasks_send, &
                            task_dim=TASK_N_RECORDS, &
                            ntasks=ntasks_send, &
                            nencode=nblkrows_total, &
                            tasks=tasks_send)

      ! generate the recv task list (tasks_recv) from the send task list
      CALL fb_com_tasks_create(com_tasks_recv)
      CALL fb_com_tasks_transpose_dest_src(com_tasks_recv, "<", com_tasks_send, &
                                           para_env)

      ! task lists are now complete, now construct the atom_pairs_send
      ! and atom_pairs_recv from the tasks lists
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_send, &
                                         atom_pairs=atom_pairs_send, &
                                         natoms_encode=nblkrows_total, &
                                         send_or_recv="send")
      CALL fb_com_tasks_build_atom_pairs(com_tasks=com_tasks_recv, &
                                         atom_pairs=atom_pairs_recv, &
                                         natoms_encode=nblkrows_total, &
                                         send_or_recv="recv")

      ! cleanup
      CALL fb_com_tasks_release(com_tasks_recv)
      CALL fb_com_tasks_release(com_tasks_send)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_generate_com_pairs_2

! **************************************************************************************************
!> \brief Build the atomic filter matrix for each atomic halo
!> \param atomic_H : atomic KS matrix
!> \param atomic_S : atomic overlap matrix
!> \param fermi_level : fermi level used to construct the Fermi-Dirac
!>                      filter function
!> \param filter_temp : temperature used to construct the Fermi-Dirac
!>                      filter function
!> \param atomic_filter_mat : the atomic filter matrix
!> \param tolerance : anything smaller than tolerance is treated as zero
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_build_atomic_fltrmat(atomic_H, &
                                              atomic_S, &
                                              fermi_level, &
                                              filter_temp, &
                                              atomic_filter_mat, &
                                              tolerance)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: atomic_H, atomic_S
      REAL(KIND=dp), INTENT(IN)                          :: fermi_level, filter_temp
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: atomic_filter_mat
      REAL(KIND=dp), INTENT(IN)                          :: tolerance

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_fltrmat_build_atomic_fltrmat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle_dgemm, handle_dsygv, ii, &
                                                            info, jj, mat_dim, work_array_size
      LOGICAL                                            :: check_ok
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, filter_function, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: atomic_S_copy, eigenvectors, &
                                                            filtered_eigenvectors

      CALL timeset(routineN, handle)

      ! This subroutine assumes atomic_filter_mat is not zero size, in
      ! other words, it really has to be constructed, instead of just
      ! being a dummy

      check_ok = SIZE(atomic_filter_mat, 1) > 0 .AND. &
                 SIZE(atomic_filter_mat, 2) > 0
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_filter_matrix_methods.F",1286)

      ! initialise
      atomic_filter_mat = 0.0_dp
      mat_dim = SIZE(atomic_H, 1)

      ! diagonalise using LAPACK
      ALLOCATE (eigenvalues(mat_dim))
      ! get optimal work array size
      ALLOCATE (work(1))
      ! dsygv will overwrite part of atomic_H and atomic_S, thus need to copy them
      ALLOCATE (atomic_S_copy(SIZE(atomic_S, 1), SIZE(atomic_S, 2)))
      atomic_S_copy(:, :) = atomic_S(:, :)
      ALLOCATE (eigenvectors(SIZE(atomic_H, 1), SIZE(atomic_H, 2)))
      eigenvectors(:, :) = atomic_H(:, :)

      CALL timeset("fb_atomic_filter_dsygv", handle_dsygv)

      info = 0
      CALL dsygv(1, &
                 'V', &
                 'U', &
                 mat_dim, &
                 eigenvectors, &
                 mat_dim, &
                 atomic_S_copy, &
                 mat_dim, &
                 eigenvalues, &
                 work, &
                 -1, &
                 info)
      work_array_size = NINT(work(1))
      ! now allocate work array
      DEALLOCATE (work)
      ALLOCATE (work(work_array_size))
      work = 0.0_dp
      ! do calculation
      atomic_S_copy(:, :) = atomic_S(:, :)
      eigenvectors(:, :) = atomic_H(:, :)
      info = 0
      CALL dsygv(1, &
                 'V', &
                 'U', &
                 mat_dim, &
                 eigenvectors, &
                 mat_dim, &
                 atomic_S_copy, &
                 mat_dim, &
                 eigenvalues, &
                 work, &
                 work_array_size, &
                 info)
      ! check if diagonalisation is successful
      IF (info .NE. 0) THEN
         WRITE (*, *) "DSYGV ERROR MESSAGE: ", info
         CALL cp__b("qs_fb_filter_matrix_methods.F",1341,"DSYGV failed")
      END IF

      CALL timestop(handle_dsygv)

      DEALLOCATE (work)
      DEALLOCATE (atomic_S_copy)

      ! first get the filter function
      ALLOCATE (filter_function(mat_dim))
      filter_function = 0.0_dp
      CALL fb_fltrmat_fermi_dirac_mu(filter_function, &
                                     eigenvalues, &
                                     filter_temp, &
                                     fermi_level)
      DEALLOCATE (eigenvalues)

      ! atomic_H has the eigenvectors, construct the version of it
      ! filtered through the filter function
      ALLOCATE (filtered_eigenvectors(mat_dim, mat_dim))
      DO jj = 1, mat_dim
         DO ii = 1, mat_dim
            filtered_eigenvectors(ii, jj) = &
               filter_function(jj)*eigenvectors(ii, jj)
         END DO ! ii
      END DO ! jj

      DEALLOCATE (filter_function)

      CALL timeset("fb_atomic_filter_dgemm", handle_dgemm)

      ! construct atomic filter matrix
      CALL dgemm("N", &
                 "T", &
                 mat_dim, &
                 mat_dim, &
                 mat_dim, &
                 1.0_dp, &
                 filtered_eigenvectors, &
                 mat_dim, &
                 eigenvectors, &
                 mat_dim, &
                 0.0_dp, &
                 atomic_filter_mat, &
                 mat_dim)

      CALL timestop(handle_dgemm)

      ! remove small negative terms due to numerical error, the filter
      ! matrix must not be negative definite
      DO jj = 1, SIZE(atomic_filter_mat, 2)
         DO ii = 1, SIZE(atomic_filter_mat, 1)
            IF (ABS(atomic_filter_mat(ii, jj)) < tolerance) THEN
               atomic_filter_mat(ii, jj) = 0.0_dp
            END IF
         END DO
      END DO

      DEALLOCATE (filtered_eigenvectors)
      DEALLOCATE (eigenvectors)

      CALL timestop(handle)

   END SUBROUTINE fb_fltrmat_build_atomic_fltrmat

! **************************************************************************************************
!> \brief get values of Fermi-Dirac distribution based on a given fermi
!>        level at a given set of energy eigenvalues
!> \param f : the Fermi-Dirac distribution function values
!> \param eigenvals : set of energy eigenvalues
!> \param T : temperature
!> \param mu : the fermi level
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_fermi_dirac_mu(f, eigenvals, T, mu)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: f
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: eigenvals
      REAL(KIND=dp), INTENT(IN)                          :: T, mu

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_fltrmat_fermi_dirac_mu', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: kTS, ne

! we want fermi function max at 1, so maxocc = 1 here

      CALL Fermi(f, ne, kTS, eigenvals, mu, T, 1.0_dp)
   END SUBROUTINE fb_fltrmat_fermi_dirac_mu

! **************************************************************************************************
!> \brief get values of Fermi-Dirac distribution based on a given electron
!>        number at a given set of energy eigenvales
!> \param f : the Fermi-Dirac distribution function values
!> \param eigenvals : set of energy eigenvalues
!> \param T : temperature
!> \param ne : number of electrons
!> \param maxocc : maximum occupancy per orbital
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_fltrmat_fermi_dirac_ne(f, eigenvals, T, ne, maxocc)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: f
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: eigenvals
      REAL(KIND=dp), INTENT(IN)                          :: T, ne, maxocc

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_fltrmat_fermi_dirac_ne', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: kTS, mu

! mu is the calculated fermi level
! kTS is the calculated entropic contribution to the energy i.e. -TS
! kTS = kT*[f ln f + (1-f) ln (1-f)]

      CALL FermiFixed(f, mu, kTS, eigenvals, ne, T, maxocc)
   END SUBROUTINE fb_fltrmat_fermi_dirac_ne

END MODULE qs_fb_filter_matrix_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_env_methods

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_distribution, cp_dbcsr_finalize, &
        cp_dbcsr_get_info, cp_dbcsr_get_matrix_type, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_nfullcols_total, &
        cp_dbcsr_nfullrows_total, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_reserve_blocks, &
        cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_basic_linalg,              ONLY: cp_fm_gemm,&
                                              cp_fm_symm,&
                                              cp_fm_triangular_invert,&
                                              cp_fm_triangular_multiply,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_power
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_constants,                 ONLY: cholesky_dbcsr,&
                                              cholesky_inverse,&
                                              cholesky_off,&
                                              cholesky_reduce,&
                                              cholesky_restore
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_max
   USE orbital_pointers,                ONLY: nco,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_diis,                         ONLY: qs_diis_b_step
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_fb_atomic_halo_types,         ONLY: &
        fb_atomic_halo_build_halo_atoms, fb_atomic_halo_cost, fb_atomic_halo_create, &
        fb_atomic_halo_list_create, fb_atomic_halo_list_nullify, fb_atomic_halo_list_obj, &
        fb_atomic_halo_list_release, fb_atomic_halo_list_set, fb_atomic_halo_list_write_info, &
        fb_atomic_halo_nelectrons_estimate_Z, fb_atomic_halo_nullify, fb_atomic_halo_obj, &
        fb_atomic_halo_set, fb_atomic_halo_sort, fb_build_pair_radii
   USE qs_fb_env_types,                 ONLY: fb_env_get,&
                                              fb_env_has_data,&
                                              fb_env_obj,&
                                              fb_env_set
   USE qs_fb_filter_matrix_methods,     ONLY: fb_fltrmat_build,&
                                              fb_fltrmat_build_2
   USE qs_fb_trial_fns_types,           ONLY: fb_trial_fns_create,&
                                              fb_trial_fns_nullify,&
                                              fb_trial_fns_obj,&
                                              fb_trial_fns_release,&
                                              fb_trial_fns_set
   USE qs_integral_utils,               ONLY: basis_set_list_setup
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_matrix_pools,                 ONLY: mpools_create,&
                                              mpools_rebuild_fm_pools,&
                                              mpools_release,&
                                              qs_matrix_pools_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: allocate_mo_set,&
                                              deallocate_mo_set,&
                                              get_mo_set,&
                                              init_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type,&
                                              set_mo_set
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE scf_control_types,               ONLY: scf_control_type
   USE string_utilities,                ONLY: compress,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_env_methods'

   PUBLIC :: fb_env_do_diag, &
             fb_env_read_input, &
             fb_env_build_rcut_auto, &
             fb_env_build_atomic_halos, &
             fb_env_write_info

CONTAINS

! **************************************************************************************************
!> \brief Do filtered matrix method diagonalisation
!> \param fb_env : the filter matrix environment
!> \param qs_env : quickstep environment
!> \param matrix_ks : DBCSR system (unfiltered) input KS matrix
!> \param matrix_s  : DBCSR system (unfiltered) input overlap matrix
!> \param scf_section : SCF input section
!> \param diis_step : whether we are doing a DIIS step
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_do_diag(fb_env, &
                             qs_env, &
                             matrix_ks, &
                             matrix_s, &
                             scf_section, &
                             diis_step)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(INOUT)                             :: diis_step

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fb_env_do_diag', routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: spin_string
      CHARACTER(len=default_string_length)               :: name
      INTEGER :: filtered_nfullrowsORcols_total, handle, homo_filtered, ispin, lfomo_filtered, &
         my_nmo, ndep, nelectron, nmo, nmo_filtered, nspin, original_nfullrowsORcols_total
      INTEGER, DIMENSION(:), POINTER                     :: filtered_rowORcol_block_sizes, &
                                                            original_rowORcol_block_sizes
      LOGICAL                                            :: collective_com
      REAL(kind=dp) :: diis_error, eps_default, eps_diis, eps_eigval, fermi_level, filter_temp, &
         flexible_electron_count, KTS_filtered, maxocc, mu_filtered
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues, eigenvalues_filtered, occ, &
                                                            occ_filtered
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_filter
      TYPE(cp_dbcsr_type)                                :: matrix_filtered_ks, matrix_filtered_s, &
                                                            matrix_tmp
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_filtered_p
      TYPE(cp_fm_struct_type), POINTER                   :: filter_fm_struct, fm_struct
      TYPE(cp_fm_type), POINTER :: fm_matrix_filter, fm_matrix_filtered_ks, fm_matrix_filtered_s, &
         fm_matrix_ortho, fm_matrix_work, mo_coeff, mo_coeff_filtered
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_atomic_halo_list_obj)                      :: atomic_halos
      TYPE(fb_trial_fns_obj)                             :: trial_fns
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_filtered
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_matrix_pools_type), POINTER                :: my_mpools
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control

! TYPE(neighbor_list_set_p_type), DIMENSION(:), POINTER :: sab_orb

      CALL timeset(routineN, handle)

      NULLIFY (scf_env, scf_control, para_env, blacs_env, particle_set)
      NULLIFY (eigenvalues, eigenvalues_filtered, occ, occ_filtered)
      NULLIFY (mos, mos_filtered)
      NULLIFY (my_mpools)
      NULLIFY (matrix_filter, matrix_filtered_p)
      NULLIFY (fm_struct, filter_fm_struct)
      NULLIFY (fm_matrix_filter, fm_matrix_filtered_s, &
               fm_matrix_filtered_ks, fm_matrix_work, &
               fm_matrix_ortho, mo_coeff_filtered, mo_coeff)
      ! NULLIFY(sab_orb)
      CALL fb_atomic_halo_list_nullify(atomic_halos)
      CALL fb_trial_fns_nullify(trial_fns)
      NULLIFY (original_rowORcol_block_sizes, filtered_rowORcol_block_sizes)

      ! get qs_env information
      CALL get_qs_env(qs_env=qs_env, &
                      scf_env=scf_env, &
                      scf_control=scf_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env, &
                      particle_set=particle_set, &
                      mos=mos)

      nspin = SIZE(matrix_ks)

      ! ----------------------------------------------------------------------
      ! DIIS step - based on non-filtered matrices and MOs
      ! ----------------------------------------------------------------------

      DO ispin = 1, nspin
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, &
                               scf_env%scf_work1(ispin)%matrix)
      END DO

      eps_diis = scf_control%eps_diis
      eps_eigval = EPSILON(0.0_dp)

      IF (scf_env%iter_count > 1 .AND. .NOT. scf_env%skip_diis) THEN
         CALL qs_diis_b_step(scf_env%scf_diis_buffer, mos, scf_env%scf_work1, &
                             scf_env%scf_work2, scf_env%iter_delta, &
                             diis_error, diis_step, eps_diis, scf_control%nmixing, &
                             s_matrix=matrix_s, scf_section=scf_section)
      ELSE
         diis_step = .FALSE.
      END IF

      IF (diis_step) THEN
         scf_env%iter_param = diis_error
         scf_env%iter_method = "DIIS/Filter"
      ELSE
         IF (scf_env%mixing_method == 0) THEN
            scf_env%iter_method = "NoMix/Filter"
         ELSE IF (scf_env%mixing_method == 1) THEN
            scf_env%iter_param = scf_env%p_mix_alpha
            scf_env%iter_method = "P_Mix/Filter"
         ELSE IF (scf_env%mixing_method > 1) THEN
            scf_env%iter_param = scf_env%mixing_store%alpha
            scf_env%iter_method = TRIM(scf_env%mixing_store%iter_method)//"/Filter"
         END IF
      END IF

      ! ----------------------------------------------------------------------
      ! Construct Filter Matrix
      ! ----------------------------------------------------------------------

      CALL fb_env_get(fb_env=fb_env, &
                      filter_temperature=filter_temp, &
                      atomic_halos=atomic_halos, &
                      eps_default=eps_default)

      ! construct trial functions
      CALL get_mo_set(mo_set=mos(1)%mo_set, maxocc=maxocc)
      CALL fb_env_build_trial_fns_auto(fb_env, qs_env, maxocc)
      CALL fb_env_get(fb_env=fb_env, &
                      trial_fns=trial_fns)

      ! allocate filter matrix (matrix_filter(ispin)%matrix are
      ! nullified by cp_dbcsr_allocate_matrix_set)
      CALL cp_dbcsr_allocate_matrix_set(matrix_filter, nspin)
      DO ispin = 1, nspin
         ! get system-wide fermi energy and occupancy, we use this to
         ! define the filter function used for the filter matrix
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                         mu=fermi_level, &
                         maxocc=maxocc)
         ! get filter matrix name
         WRITE (spin_string, FMT="(I1)") ispin
         name = TRIM("FILTER MATRIX SPIN "//spin_string)
         CALL compress(name)
         CALL uppercase(name)
         ! calculate filter matrix (matrix_s(1) is the overlap, the rest
         ! in the array are its derivatives)
         CALL fb_env_get(fb_env=fb_env, &
                         collective_com=collective_com)
         IF (collective_com) THEN
            CALL fb_fltrmat_build_2(H_mat=matrix_ks(ispin)%matrix, &
                                    S_mat=matrix_s(1)%matrix, &
                                    atomic_halos=atomic_halos, &
                                    trial_fns=trial_fns, &
                                    para_env=para_env, &
                                    particle_set=particle_set, &
                                    fermi_level=fermi_level, &
                                    filter_temp=filter_temp, &
                                    name=name, &
                                    filter_mat=matrix_filter(ispin)%matrix, &
                                    tolerance=eps_default)
         ELSE
            CALL fb_fltrmat_build(H_mat=matrix_ks(ispin)%matrix, &
                                  S_mat=matrix_s(1)%matrix, &
                                  atomic_halos=atomic_halos, &
                                  trial_fns=trial_fns, &
                                  para_env=para_env, &
                                  particle_set=particle_set, &
                                  fermi_level=fermi_level, &
                                  filter_temp=filter_temp, &
                                  name=name, &
                                  filter_mat=matrix_filter(ispin)%matrix, &
                                  tolerance=eps_default)
         END IF
      END DO ! ispin

      ! ----------------------------------------------------------------------
      ! Do Filtered Diagonalisation
      ! ----------------------------------------------------------------------

      ! Obtain matrix dimensions. KS and S matrices are symmetric, so
      ! row_block_sizes and col_block_sizes should be identical. The
      ! same applies to the filtered block sizes. Note that filter
      ! matrix will have row_block_sizes equal to that of the original,
      ! and col_block_sizes equal to that of the filtered.  We assume
      ! also that the matrix dimensions are identical for both spin
      ! channels.
      original_rowORcol_block_sizes => cp_dbcsr_row_block_sizes(matrix_ks(1)%matrix)
      filtered_rowORcol_block_sizes => cp_dbcsr_col_block_sizes(matrix_filter(1)%matrix)
      original_nfullrowsORcols_total = cp_dbcsr_nfullrows_total(matrix_ks(1)%matrix)
      filtered_nfullrowsORcols_total = cp_dbcsr_nfullcols_total(matrix_filter(1)%matrix)

      ! filter diagonalisation works on a smaller basis set, and thus
      ! requires a new mo_set (molecular orbitals | eigenvectors) and
      ! the corresponding matrix pools for the eigenvector coefficients
      ALLOCATE (mos_filtered(nspin))
      DO ispin = 1, nspin
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                         maxocc=maxocc, &
                         nelectron=nelectron, &
                         flexible_electron_count=flexible_electron_count)
         NULLIFY (mos_filtered(ispin)%mo_set)
         CALL allocate_mo_set(mo_set=mos_filtered(ispin)%mo_set, &
                              nao=filtered_nfullrowsORcols_total, &
                              nmo=filtered_nfullrowsORcols_total, &
                              nelectron=nelectron, &
                              n_el_f=REAL(nelectron, dp), &
                              maxocc=maxocc, &
                              flexible_electron_count=flexible_electron_count)
      END DO ! ispin
      CALL mpools_create(mpools=my_mpools)
      CALL mpools_rebuild_fm_pools(mpools=my_mpools, &
                                   mos=mos_filtered, &
                                   blacs_env=blacs_env, &
                                   para_env=para_env)

      ! create DBCSR filtered KS matrix, this is reused for each spin
      ! channel
      CALL cp_dbcsr_init(matrix_filtered_ks)
      ! both row_blk_size and col_blk_size should be that of
      ! col_blk_size of the filter matrix
      CALL cp_dbcsr_create(matrix=matrix_filtered_ks, &
                           name=TRIM("FILTERED_KS_MATRIX"), &
                           dist=cp_dbcsr_distribution(matrix_ks(1)%matrix), &
                           matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=filtered_rowORcol_block_sizes, &
                           col_blk_size=filtered_rowORcol_block_sizes, &
                           nze=0)
      CALL cp_dbcsr_finalize(matrix_filtered_ks)

      ! create DBCSR filtered S (overlap) matrix. Note that
      ! matrix_s(1)%matrix is the orginal overlap matrix---the rest in
      ! the array are derivatives, and it should not depend on
      ! spin. HOWEVER, since the filter matrix is constructed from KS
      ! matrix, and does depend on spin, the filtered S also becomes
      ! spin dependent. Nevertheless this matrix is reused for each spin
      ! channel
      CALL cp_dbcsr_init(matrix_filtered_s)
      ! both row_blk_size and col_blk_size should be that of
      ! col_blk_size of the filter matrix
      CALL cp_dbcsr_create(matrix=matrix_filtered_s, &
                           name=TRIM("FILTERED_S_MATRIX"), &
                           dist=cp_dbcsr_distribution(matrix_s(1)%matrix), &
                           matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=filtered_rowORcol_block_sizes, &
                           col_blk_size=filtered_rowORcol_block_sizes, &
                           nze=0)
      CALL cp_dbcsr_finalize(matrix_filtered_s)

      ! create temporary matrix for constructing filtered KS and S
      CALL cp_dbcsr_init(matrix_tmp)
      ! the temporary matrix won't be square
      CALL cp_dbcsr_create(matrix=matrix_tmp, &
                           name=TRIM("TEMPORARY_MATRIX"), &
                           dist=cp_dbcsr_distribution(matrix_s(1)%matrix), &
                           matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=original_rowORcol_block_sizes, &
                           col_blk_size=filtered_rowORcol_block_sizes, &
                           nze=0)
      CALL cp_dbcsr_finalize(matrix_tmp)

      ! create fm format matrices used for diagonalisation
      CALL cp_fm_struct_create(fmstruct=fm_struct, &
                               para_env=para_env, &
                               context=blacs_env, &
                               nrow_global=filtered_nfullrowsORcols_total, &
                               ncol_global=filtered_nfullrowsORcols_total)
      ! both fm_matrix_filtered_s and fm_matrix_filtered_ks are reused
      ! for each spin channel
      CALL cp_fm_create(fm_matrix_filtered_s, &
                        fm_struct, &
                        name="FM_MATRIX_FILTERED_S")
      CALL cp_fm_create(fm_matrix_filtered_ks, &
                        fm_struct, &
                        name="FM_MATRIX_FILTERED_KS")
      ! creaate work matrix
      CALL cp_fm_create(fm_matrix_work, fm_struct, name="FM_MATRIX_WORK")
      CALL cp_fm_create(fm_matrix_ortho, fm_struct, name="FM_MATRIX_ORTHO")
      ! all fm matrices are created, so can release fm_struct
      CALL cp_fm_struct_release(fm_struct)

      ! construct filtered KS, S matrix and diagonalise
      DO ispin = 1, nspin

         ! construct filtered KS matrix
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                matrix_ks(ispin)%matrix, matrix_filter(ispin)%matrix, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                matrix_filter(ispin)%matrix, matrix_tmp, &
                                0.0_dp, matrix_filtered_ks)
         ! construct filtered S_matrix
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                matrix_s(1)%matrix, matrix_filter(ispin)%matrix, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                matrix_filter(ispin)%matrix, matrix_tmp, &
                                0.0_dp, matrix_filtered_s)

         ! now that we have the filtered KS and S matrices for this spin
         ! channel, perform ordinary diagonalisation

         ! convert DBCSR matrices to fm format
         CALL copy_dbcsr_to_fm(matrix_filtered_s, fm_matrix_filtered_s)
         CALL copy_dbcsr_to_fm(matrix_filtered_ks, fm_matrix_filtered_ks)

         ! setup matrix pools for the molecular orbitals
         CALL init_mo_set(mos_filtered(ispin)%mo_set, &
                          fm_pool=my_mpools%ao_mo_fm_pools(ispin)%pool, &
                          name="FILTERED_MOS")

         ! now diagonalise
         CALL fb_env_eigensolver(fm_matrix_filtered_ks, &
                                 fm_matrix_filtered_s, &
                                 mos_filtered(ispin)%mo_set, &
                                 fm_matrix_ortho, &
                                 fm_matrix_work, &
                                 eps_eigval, &
                                 ndep, &
                                 scf_env%cholesky_method)
      END DO ! ispin

      ! release temporary matrices
      CALL cp_dbcsr_release(matrix_filtered_s)
      CALL cp_dbcsr_release(matrix_filtered_ks)
      CALL cp_fm_release(fm_matrix_filtered_s)
      CALL cp_fm_release(fm_matrix_filtered_ks)
      CALL cp_fm_release(fm_matrix_work)
      CALL cp_fm_release(fm_matrix_ortho)

      ! ----------------------------------------------------------------------
      ! Construct New Density Matrix
      ! ----------------------------------------------------------------------

      ! calculate filtered molecular orbital occupation numbers and fermi
      ! level etc
      CALL set_mo_occupation(mo_array=mos_filtered, &
                             smear=scf_control%smear)

      ! get the filtered density matrix and then convert back to the
      ! full basis version in scf_env ready to be used outside this
      ! subroutine
      ALLOCATE (matrix_filtered_p)
      CALL cp_dbcsr_init(matrix_filtered_p)
      ! the filtered density matrix should have the same sparse
      ! structure as the original density matrix, we must copy the
      ! sparse structure here, since construction of the density matrix
      ! preserves its sparse form, and therefore matrix_filtered_p must
      ! have its blocks allocated here now. We assume the original
      ! density matrix scf_env%p_mix_new has the same sparse structure
      ! in both spin channels.
      CALL cp_dbcsr_create(matrix=matrix_filtered_p, &
                           name=TRIM("FILTERED_MATRIX_P"), &
                           dist=cp_dbcsr_distribution(scf_env%p_mix_new(1, 1)%matrix), &
                           matrix_type=cp_dbcsr_get_matrix_type(scf_env%p_mix_new(1, 1)%matrix), &
                           row_blk_size=filtered_rowORcol_block_sizes, &
                           col_blk_size=filtered_rowORcol_block_sizes, &
                           nze=0)
      CALL cp_dbcsr_finalize(matrix_filtered_p)
      CALL fb_dbcsr_copy_sparse_struct(matrix_filtered_p, &
                                       scf_env%p_mix_new(1, 1)%matrix)
      ! old implementation, using sab_orb to allocate the blocks in matrix_filtered_p
      ! CALL get_qs_env(qs_env=qs_env, sab_orb=sab_orb)
      ! CALL cp_dbcsr_alloc_block_from_nbl(matrix_filtered_p, sab_orb)
      CALL cp_dbcsr_set(matrix_filtered_p, 0.0_dp)

      DO ispin = 1, nspin
         ! calculate matrix_filtered_p
         CALL calculate_density_matrix(mos_filtered(ispin)%mo_set, &
                                       matrix_filtered_p)
         ! convert back to full basis p
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                matrix_filter(ispin)%matrix, matrix_filtered_p, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                matrix_tmp, matrix_filter(ispin)%matrix, &
                                0.0_dp, scf_env%p_mix_new(ispin, 1)%matrix, &
                                retain_sparsity=.TRUE.)
         ! note that we want to retain the sparse structure of
         ! scf_env%p_mix_new
      END DO ! ispin

      ! release temporary matrices
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_filtered_p)
      DEALLOCATE (matrix_filtered_p)

      ! ----------------------------------------------------------------------
      ! Update MOs
      ! ----------------------------------------------------------------------

      ! we still need to convert mos_filtered back to the full basis
      ! version (mos) for this, we need to update mo_coeff (and/or
      ! mo_coeff_b --- the DBCSR version, if used) of mos

      ! note also that mo_eigenvalues cannot be fully updated, given
      ! that the eigenvalues are computed in a smaller basis, and thus
      ! do not give the full spectron. Printing of molecular states
      ! (molecular DOS) at each SCF step is therefore not recommended
      ! when using this method. The idea is that if one wants a full
      ! molecular DOS, then one should perform a full diagonalisation
      ! without the filters once the SCF has been achieved.

      ! NOTE: from reading the source code, it appears that mo_coeff_b
      ! is actually never used by default (DOUBLE CHECK?!). Even
      ! subroutine eigensolver_dbcsr updates mo_coeff, and not
      ! mo_coeff_b.

      ! create FM format filter matrix
      CALL cp_fm_struct_create(fmstruct=filter_fm_struct, &
                               para_env=para_env, &
                               context=blacs_env, &
                               nrow_global=original_nfullrowsORcols_total, &
                               ncol_global=filtered_nfullrowsORcols_total)
      CALL cp_fm_create(fm_matrix_filter, &
                        filter_fm_struct, &
                        name="FM_MATRIX_FILTER")
      CALL cp_fm_struct_release(filter_fm_struct)

      DO ispin = 1, nspin
         ! now the full basis mo_set should only contain the reduced
         ! number of eigenvectors and eigenvalues
         CALL get_mo_set(mo_set=mos_filtered(ispin)%mo_set, &
                         homo=homo_filtered, &
                         lfomo=lfomo_filtered, &
                         nmo=nmo_filtered, &
                         eigenvalues=eigenvalues_filtered, &
                         occupation_numbers=occ_filtered, &
                         mo_coeff=mo_coeff_filtered, &
                         kTS=kTS_filtered, &
                         mu=mu_filtered)
         ! first set all the relevent scalars
         CALL set_mo_set(mo_set=mos(ispin)%mo_set, &
                         homo=homo_filtered, &
                         lfomo=lfomo_filtered, &
                         kTS=kTS_filtered, &
                         mu=mu_filtered)
         ! now set the arrays and fm_matrices
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                         nmo=nmo, &
                         occupation_numbers=occ, &
                         eigenvalues=eigenvalues, &
                         mo_coeff=mo_coeff)
         ! number of mos in original mo_set may sometimes be less than
         ! nmo_filtered, so we must make sure we do not go out of bounds
         my_nmo = MIN(nmo, nmo_filtered)
         eigenvalues(:) = 0.0_dp
         eigenvalues(1:my_nmo) = eigenvalues_filtered(1:my_nmo)
         occ(:) = 0.0_dp
         occ(1:my_nmo) = occ_filtered(1:my_nmo)
         ! convert mo_coeff_filtered back to original basis
         CALL cp_fm_set_all(matrix=mo_coeff, alpha=0.0_dp)
         CALL copy_dbcsr_to_fm(matrix_filter(ispin)%matrix, fm_matrix_filter)
         CALL cp_fm_gemm("N", "N", &
                         original_nfullrowsORcols_total, &
                         my_nmo, &
                         filtered_nfullrowsORcols_total, &
                         1.0_dp, fm_matrix_filter, mo_coeff_filtered, &
                         0.0_dp, mo_coeff)

      END DO ! ispin

      ! release temporary matrices
      CALL cp_fm_release(fm_matrix_filter)

      ! ----------------------------------------------------------------------
      ! Final Clean Up
      ! ----------------------------------------------------------------------

      CALL mpools_release(mpools=my_mpools)
      DO ispin = 1, nspin
         CALL deallocate_mo_set(mo_set=mos_filtered(ispin)%mo_set)
      END DO
      DEALLOCATE (mos_filtered)
      CALL cp_dbcsr_deallocate_matrix_set(matrix_filter)

      CALL timestop(handle)

   END SUBROUTINE fb_env_do_diag

! **************************************************************************************************
!> \brief The main parallel eigensolver engine for filter matrix diagonalisation
!> \param fm_KS : the BLACS distributed Kohn-Sham matrix, input only
!> \param fm_S  : the BLACS distributed overlap matrix, input only
!> \param mo_set : upon output contains the molecular orbitals (eigenvectors)
!>                 and eigenvalues
!> \param fm_ortho : one of the work matrices, on output, the BLACS distributed
!>                   matrix for orthogalising the eigen problem. E.g. if using
!>                   Cholesky inversse, then the upper triangle part contains
!>                   the inverse of Cholesky U; if not using Cholesky, then it
!>                   contains the S^-1/2.
!> \param fm_work : work matrix used by eigen solver
!> \param eps_eigval : used for quenching the small numbers when computing S^-1/2
!>                     any values less than eps_eigval is truncated to zero.
!> \param ndep : if the overlap is not positive definite, then ndep > 0,
!>               and equals to the number of linear dependent basis functions
!>               in the filtered basis set
!> \param method : method for solving generalised eigenvalue problem
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_eigensolver(fm_KS, fm_S, mo_set, fm_ortho, &
                                 fm_work, eps_eigval, ndep, method)
      TYPE(cp_fm_type), POINTER                          :: fm_KS, fm_S
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_fm_type), POINTER                          :: fm_ortho, fm_work
      REAL(KIND=dp), INTENT(IN)                          :: eps_eigval
      INTEGER, INTENT(OUT)                               :: ndep
      INTEGER, INTENT(IN)                                :: method

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_eigensolver', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=8)                                   :: ndep_string
      INTEGER                                            :: handle, info, my_method, nao, nmo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      CALL get_mo_set(mo_set=mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      eigenvalues=mo_eigenvalues, &
                      mo_coeff=mo_coeff)
      my_method = method
      ndep = 0

      ! first, obtain orthogonalisation (ortho) matrix
      IF (my_method .NE. cholesky_off) THEN
         CALL cp_fm_to_fm(fm_S, fm_ortho)
         CALL cp_fm_cholesky_decompose(fm_ortho, info_out=info)
         IF (info .NE. 0) THEN
            CALL cp_warn(cp__l("qs_fb_env_methods.F",653), &
                         "Unable to perform Cholesky decomposition on the overlap "// &
                         "matrix. The new filtered basis may not be linearly "// &
                         "independent set. Revert to using inverse square-root "// &
                         "of the overlap. To avoid this warning, you can try"// &
                         "to use a higher filter termperature.")
            my_method = cholesky_off
         ELSE
            SELECT CASE (my_method)
            CASE (cholesky_dbcsr)
               CALL cp_abort(cp__l("qs_fb_env_methods.F",663), &
                             "filter matrix method with CHOLESKY_DBCSR is not yet implemented")
            CASE (cholesky_reduce)
               CALL cp_fm_cholesky_reduce(fm_KS, fm_ortho)
               CALL choose_eigv_solver(fm_KS, fm_work, mo_eigenvalues)
               CALL cp_fm_cholesky_restore(fm_work, nmo, fm_ortho, mo_coeff, "SOLVE")
            CASE (cholesky_restore)
               CALL cp_fm_upper_to_full(fm_KS, fm_work)
               CALL cp_fm_cholesky_restore(fm_KS, nao, fm_ortho, fm_work, "SOLVE", &
                                           pos="RIGHT")
               CALL cp_fm_cholesky_restore(fm_work, nao, fm_ortho, fm_KS, "SOLVE", &
                                           pos="LEFT", transa="T")
               CALL choose_eigv_solver(fm_KS, fm_work, mo_eigenvalues)
               CALL cp_fm_cholesky_restore(fm_work, nmo, fm_ortho, mo_coeff, "SOLVE")
            CASE (cholesky_inverse)
               CALL cp_fm_triangular_invert(fm_ortho)
               CALL cp_fm_upper_to_full(fm_KS, fm_work)
               CALL cp_fm_triangular_multiply(fm_ortho, &
                                              fm_KS, &
                                              side="R", &
                                              transpose_tr=.FALSE., &
                                              invert_tr=.FALSE., &
                                              uplo_tr="U", &
                                              n_rows=nao, &
                                              n_cols=nao, &
                                              alpha=1.0_dp)
               CALL cp_fm_triangular_multiply(fm_ortho, &
                                              fm_KS, &
                                              side="L", &
                                              transpose_tr=.TRUE., &
                                              invert_tr=.FALSE., &
                                              uplo_tr="U", &
                                              n_rows=nao, &
                                              n_cols=nao, &
                                              alpha=1.0_dp)
               CALL choose_eigv_solver(fm_KS, fm_work, mo_eigenvalues)
               CALL cp_fm_triangular_multiply(fm_ortho, &
                                              fm_work, &
                                              side="L", &
                                              transpose_tr=.FALSE., &
                                              invert_tr=.FALSE., &
                                              uplo_tr="U", &
                                              n_rows=nao, &
                                              n_cols=nmo, &
                                              alpha=1.0_dp)
               CALL cp_fm_to_fm(fm_work, mo_coeff, nmo, 1, 1)
            END SELECT
         END IF
      END IF
      IF (my_method == cholesky_off) THEN
         ! calculating ortho as S^-1/2 using diagonalisation of S, and
         ! solve accordingly
         CALL cp_fm_to_fm(fm_S, fm_ortho)
         CALL cp_fm_power(fm_ortho, fm_work, -0.5_dp, &
                          eps_eigval, ndep)
         IF (ndep > 0) THEN
            WRITE (ndep_string, FMT="(I8)") ndep
            CALL cp_warn(cp__l("qs_fb_env_methods.F",720), &
                         "Number of linearly dependent filtered orbitals: "//ndep_string)
         END IF
         ! solve eigen equatoin using S^-1/2
         CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, fm_KS, fm_ortho, &
                         0.0_dp, fm_work)
         CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, fm_ortho, &
                      fm_work, 0.0_dp, fm_KS)
         CALL choose_eigv_solver(fm_KS, fm_work, mo_eigenvalues)
         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, fm_ortho, &
                      fm_work, 0.0_dp, mo_coeff)
      END IF

      CALL timestop(handle)

   END SUBROUTINE fb_env_eigensolver

! **************************************************************************************************
!> \brief Read input sections for filter matrix method
!> \param fb_env : the filter matrix environment
!> \param scf_section : SCF input section
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_read_input(fb_env, scf_section)

      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_read_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: l_val
      REAL(KIND=dp)                                      :: r_val
      TYPE(section_vals_type), POINTER                   :: fb_section

      CALL timeset(routineN, handle)

      NULLIFY (fb_section)
      fb_section => section_vals_get_subs_vals(scf_section, &
                                               "DIAGONALIZATION%FILTER_MATRIX")
      ! filter_temperature
      CALL section_vals_val_get(fb_section, "FILTER_TEMPERATURE", &
                                r_val=r_val)
      CALL fb_env_set(fb_env=fb_env, &
                      filter_temperature=r_val)
      ! auto_cutoff_scale
      CALL section_vals_val_get(fb_section, "AUTO_CUTOFF_SCALE", &
                                r_val=r_val)
      CALL fb_env_set(fb_env=fb_env, &
                      auto_cutoff_scale=r_val)
      ! communication model
      CALL section_vals_val_get(fb_section, "COLLECTIVE_COMMUNICATION", &
                                l_val=l_val)
      CALL fb_env_set(fb_env=fb_env, &
                      collective_com=l_val)
      ! eps_default
      CALL section_vals_val_get(fb_section, "EPS_FB", &
                                r_val=r_val)
      CALL fb_env_set(fb_env=fb_env, &
                      eps_default=r_val)

      CALL timestop(handle)

   END SUBROUTINE fb_env_read_input

! **************************************************************************************************
!> \brief Automatically generate the cutoff radii of atoms used for
!>        constructing the atomic halos, based on basis set cutoff
!>        ranges for each kind
!> \param fb_env : the filter matrix environment
!> \param qs_env : quickstep environment
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_build_rcut_auto(fb_env, qs_env)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_build_rcut_auto', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, nkinds
      REAL(KIND=dp)                                      :: auto_cutoff_scale, kind_radius
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rcut
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      NULLIFY (rcut, qs_kind_set, dft_control)

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control)
      CALL fb_env_get(fb_env=fb_env, &
                      auto_cutoff_scale=auto_cutoff_scale)

      nkinds = SIZE(qs_kind_set)
      ALLOCATE (rcut(nkinds))

      ! reading from the other parts of the code, it seemed that
      ! aux_fit_basis_set is only used when do_admm is TRUE. This can be
      ! seen from the calls to generate_qs_task_list subroutine in
      ! qs_create_task_list, found in qs_environment_methods.F:
      ! basis_type is only set as input parameter for do_admm
      ! calculations, and if not set, the task list is generated using
      ! the default basis_set="ORB".
      ALLOCATE (basis_set_list(nkinds))
      IF (dft_control%do_admm) THEN
         CALL basis_set_list_setup(basis_set_list, "AUX_FIT", qs_kind_set)
      ELSE
         CALL basis_set_list_setup(basis_set_list, "ORB", qs_kind_set)
      END IF

      DO ikind = 1, nkinds
         basis_set => basis_set_list(ikind)%gto_basis_set
         CALL get_gto_basis_set(gto_basis_set=basis_set, kind_radius=kind_radius)
         rcut(ikind) = kind_radius*auto_cutoff_scale
      END DO

      CALL fb_env_set(fb_env=fb_env, &
                      rcut=rcut)

      ! cleanup
      DEALLOCATE (basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE fb_env_build_rcut_auto

! **************************************************************************************************
!> \brief Builds an fb_atomic_halo_list object using information
!>        from fb_env
!> \param fb_env the fb_env object
!> \param qs_env : quickstep environment (need this to access particle)
!>                 positions and their kinds as well as which particles
!>                 are local to this process
!> \param scf_section : SCF input section, for printing output
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_build_atomic_halos(fb_env, qs_env, scf_section)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_build_atomic_halos', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iatom, ihalo, max_natoms_local, natoms_global, natoms_local, nelectrons, &
         nhalo_atoms, nkinds_global, owner_id_in_halo
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms, local_atoms
      REAL(KIND=dp)                                      :: cost
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pair_radii
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rcut
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_atomic_halo_list_obj)                      :: atomic_halos
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      IF(.NOT.(fb_env_has_data(fb_env)))CALL cp__a("qs_fb_env_methods.F",885)

      NULLIFY (cell, halos, halo_atoms, rcut, particle_set, para_env, &
               qs_kind_set, local_atoms)
      CALL fb_atomic_halo_list_nullify(atomic_halos)

      ! get relevant data from fb_env
      CALL fb_env_get(fb_env=fb_env, &
                      rcut=rcut, &
                      local_atoms=local_atoms, &
                      nlocal_atoms=natoms_local)

      ! create atomic_halos
      CALL fb_atomic_halo_list_create(atomic_halos)

      ! get the number of atoms and kinds:
      CALL get_qs_env(qs_env=qs_env, &
                      natom=natoms_global, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      nkind=nkinds_global, &
                      para_env=para_env, &
                      cell=cell)

      ! get the maximum number of local atoms across the procs.
      max_natoms_local = natoms_local
      CALL mp_max(max_natoms_local, para_env%group)

      ! create the halos, one for each local atom
      ALLOCATE (halos(natoms_local))
      DO ihalo = 1, natoms_local
         CALL fb_atomic_halo_nullify(halos(ihalo))
         CALL fb_atomic_halo_create(halos(ihalo))
      END DO
      CALL fb_atomic_halo_list_set(atomic_halos=atomic_halos, &
                                   nhalos=natoms_local, &
                                   max_nhalos=max_natoms_local)
      ! build halos
      ALLOCATE (pair_radii(nkinds_global, nkinds_global))
      CALL fb_build_pair_radii(rcut, nkinds_global, pair_radii)
      ihalo = 0
      DO iatom = 1, natoms_local
         ihalo = ihalo+1
         CALL fb_atomic_halo_build_halo_atoms(local_atoms(iatom), &
                                              particle_set, &
                                              cell, &
                                              pair_radii, &
                                              halo_atoms, &
                                              nhalo_atoms, &
                                              owner_id_in_halo)
         CALL fb_atomic_halo_set(atomic_halo=halos(ihalo), &
                                 owner_atom=local_atoms(iatom), &
                                 owner_id_in_halo=owner_id_in_halo, &
                                 natoms=nhalo_atoms, &
                                 halo_atoms=halo_atoms)
         ! prepare halo_atoms for another halo, do not deallocate, as
         ! original data is being pointed at by the atomic halo data
         ! structure
         NULLIFY (halo_atoms)
         ! calculate the number of electrons in each halo
         nelectrons = fb_atomic_halo_nelectrons_estimate_Z(halos(ihalo), &
                                                           particle_set)
         ! calculate cost
         cost = fb_atomic_halo_cost(halos(ihalo), particle_set, qs_kind_set)
         CALL fb_atomic_halo_set(atomic_halo=halos(ihalo), &
                                 nelectrons=nelectrons, &
                                 cost=cost)
         ! sort atomic halo
         CALL fb_atomic_halo_sort(halos(ihalo))
      END DO ! iatom
      DEALLOCATE (pair_radii)

      ! finalise
      CALL fb_atomic_halo_list_set(atomic_halos=atomic_halos, &
                                   halos=halos)
      CALL fb_env_set(fb_env=fb_env, &
                      atomic_halos=atomic_halos)
      CALL fb_atomic_halo_list_release(atomic_halos)

      ! print info
      CALL fb_atomic_halo_list_write_info(atomic_halos, &
                                          para_env, &
                                          scf_section)

      CALL timestop(handle)

   END SUBROUTINE fb_env_build_atomic_halos

! **************************************************************************************************
!> \brief Automatically construct the trial functiosn used for generating
!>        the filter matrix. It tries to use the single zeta subset from
!>        the system GTO basis set as the trial functions
!> \param fb_env : the filter matrix environment
!> \param qs_env : quickstep environment
!> \param maxocc : maximum occupancy for an orbital
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_build_trial_fns_auto(fb_env, qs_env, maxocc)

      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp), INTENT(IN)                          :: maxocc

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_build_trial_fns_auto', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, handle, icgf, ico, ikind, iset, &
                                                            ishell, itrial, lshell, max_n_trial, &
                                                            nkinds, nset, old_lshell
      INTEGER, DIMENSION(:), POINTER                     :: lmax, nfunctions, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: functions
      REAL(KIND=dp)                                      :: zeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(fb_trial_fns_obj)                             :: trial_fns
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      IF(.NOT.(fb_env_has_data(fb_env)))CALL cp__a("qs_fb_env_methods.F",1005)
      NULLIFY (nfunctions, functions, basis_set, basis_set_list, qs_kind_set, dft_control)
      CALL fb_trial_fns_nullify(trial_fns)

      ! create a new trial_fn object
      CALL fb_trial_fns_create(trial_fns)

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control)

      nkinds = SIZE(qs_kind_set)

      ! reading from the other parts of the code, it seemed that
      ! aux_fit_basis_set is only used when do_admm is TRUE. This can be
      ! seen from the calls to generate_qs_task_list subroutine in
      ! qs_create_task_list, found in qs_environment_methods.F:
      ! basis_type is only set as input parameter for do_admm
      ! calculations, and if not set, the task list is generated using
      ! the default basis_set="ORB".
      ALLOCATE (basis_set_list(nkinds))
      IF (dft_control%do_admm) THEN
         CALL basis_set_list_setup(basis_set_list, "AUX_FIT", qs_kind_set)
      ELSE
         CALL basis_set_list_setup(basis_set_list, "ORB", qs_kind_set)
      END IF

      ALLOCATE (nfunctions(nkinds))
      nfunctions = 0

      DO ikind = 1, nkinds
         ! "gto = gaussian type orbital"
         basis_set => basis_set_list(ikind)%gto_basis_set
         CALL get_gto_basis_set(gto_basis_set=basis_set, &
                                nset=nset, &
                                lmax=lmax, &
                                nshell=nshell)
         CALL get_qs_kind(qs_kind=qs_kind_set(ikind), &
                          zeff=zeff)

         bset1: DO iset = 1, nset
!          old_lshell = lmax(iset)
            old_lshell = -1
            DO ishell = 1, nshell(iset)
               lshell = basis_set%l(ishell, iset)
               counter = 0
               ! loop over orbitals within the same l
               DO ico = ncoset(lshell-1)+1, ncoset(lshell)
                  counter = counter+1
                  ! only include the first zeta orbitals
                  IF ((lshell .GT. old_lshell) .AND. (counter .LE. nco(lshell))) THEN
                     nfunctions(ikind) = nfunctions(ikind)+1
                  END IF
               END DO
               ! we have got enough trial functions when we have enough
               ! basis functions to accomodate the number of electrons,
               ! AND that that we have included all the first zeta
               ! orbitals of an angular momentum quantum number l
               IF (((lshell .GT. old_lshell) .OR. (lshell .EQ. lmax(iset))) .AND. &
                   (maxocc*REAL(nfunctions(ikind), dp) .GE. zeff)) THEN
                  EXIT bset1
               END IF
               old_lshell = lshell
            END DO
         END DO bset1
      END DO ! ikind

      ! now that we have the number of trial functions get the trial
      ! functions
      max_n_trial = MAXVAL(nfunctions)
      ALLOCATE (functions(max_n_trial, nkinds))
      functions(:, :) = 0
      ! redo the loops to get the trial function indices within the basis set
      DO ikind = 1, nkinds
         ! "gto = gaussian type orbital"
         basis_set => basis_set_list(ikind)%gto_basis_set
         CALL get_gto_basis_set(gto_basis_set=basis_set, &
                                nset=nset, &
                                lmax=lmax, &
                                nshell=nshell)
         CALL get_qs_kind(qs_kind=qs_kind_set(ikind), &
                          zeff=zeff)
         icgf = 0
         itrial = 0
         bset2: DO iset = 1, nset
            old_lshell = -1
            DO ishell = 1, nshell(iset)
               lshell = basis_set%l(ishell, iset)
               counter = 0
               ! loop over orbitals within the same l
               DO ico = ncoset(lshell-1)+1, ncoset(lshell)
                  icgf = icgf+1
                  counter = counter+1
                  ! only include the first zeta orbitals
                  IF ((lshell .GT. old_lshell) .AND. (counter .LE. nco(lshell))) THEN
                     itrial = itrial+1
                     functions(itrial, ikind) = icgf
                  END IF
               END DO
               ! we have got enough trial functions when we have more
               ! basis functions than the number of electrons (obtained
               ! from atomic z), AND that that we have included all the
               ! first zeta orbitals of an angular momentum quantum
               ! number l
               IF (((lshell .GT. old_lshell) .OR. (lshell .EQ. lmax(iset))) .AND. &
                   (maxocc*REAL(itrial, dp) .GE. zeff)) THEN
                  EXIT bset2
               END IF
               old_lshell = lshell
            END DO
         END DO bset2
      END DO ! ikind

      ! set trial_functions
      CALL fb_trial_fns_set(trial_fns=trial_fns, &
                            nfunctions=nfunctions, &
                            functions=functions)
      ! set fb_env
      CALL fb_env_set(fb_env=fb_env, &
                      trial_fns=trial_fns)
      CALL fb_trial_fns_release(trial_fns)

      ! cleanup
      DEALLOCATE (basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE fb_env_build_trial_fns_auto

! **************************************************************************************************
!> \brief Copy the sparse structure of a DBCSR matrix to another, this
!>        means the other matrix will have the same number of blocks
!>        and their corresponding logical locations allocated, although
!>        the blocks does not have to be the same size as the original
!> \param matrix_out : DBCSR matrix whose blocks are to be allocated
!> \param matrix_in  : DBCSR matrix with exising sparse structure that
!>                     is to be copied
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_dbcsr_copy_sparse_struct(matrix_out, matrix_in)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_out
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_in

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_dbcsr_copy_sparse_struct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, iblk, jatom, nblkcols_total, &
                                                            nblkrows_total, nblks
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cols, rows
      REAL(dp), DIMENSION(:, :), POINTER                 :: mat_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL cp_dbcsr_get_info(matrix=matrix_in, &
                             nblkrows_total=nblkrows_total, &
                             nblkcols_total=nblkcols_total)

      nblks = nblkrows_total*nblkcols_total
      ALLOCATE (rows(nblks))
      ALLOCATE (cols(nblks))
      rows(:) = 0
      cols(:) = 0
      iblk = 0
      nblks = 0
      CALL cp_dbcsr_iterator_start(iter, matrix_in)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, mat_block, iblk)
         rows(iblk) = iatom
         cols(iblk) = jatom
         nblks = nblks+1
      END DO
      CALL cp_dbcsr_iterator_stop(iter)
      CALL cp_dbcsr_reserve_blocks(matrix_out, rows(1:nblks), cols(1:nblks))
      CALL cp_dbcsr_finalize(matrix_out)

      ! cleanup
      DEALLOCATE (rows)
      DEALLOCATE (cols)

   END SUBROUTINE fb_dbcsr_copy_sparse_struct

! **************************************************************************************************
!> \brief Write out parameters used for the filter matrix method to
!>        output
!> \param fb_env : the filter matrix environment
!> \param qs_env : quickstep environment
!> \param scf_section : SCF input section
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_write_info(fb_env, qs_env, scf_section)
      TYPE(fb_env_obj), INTENT(IN)                       :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_write_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: handle, ikind, nkinds, unit_nr
      LOGICAL                                            :: collective_com
      REAL(KIND=dp)                                      :: auto_cutoff_scale, filter_temperature
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rcut
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (rcut, atomic_kind_set, logger)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set)
      CALL fb_env_get(fb_env=fb_env, &
                      filter_temperature=filter_temperature, &
                      auto_cutoff_scale=auto_cutoff_scale, &
                      rcut=rcut, &
                      collective_com=collective_com)

      nkinds = SIZE(atomic_kind_set)

      logger => cp_get_default_logger()
      unit_nr = cp_print_key_unit_nr(logger, scf_section, &
                                     "PRINT%FILTER_MATRIX", &
                                     extension="")
      IF (unit_nr > 0) THEN
         IF (collective_com) THEN
            WRITE (UNIT=unit_nr, FMT="(/,A,T71,A)") &
               " FILTER_MAT_DIAG| MPI communication method:", &
               "Collective"
         ELSE
            WRITE (UNIT=unit_nr, FMT="(/,A,T71,A)") &
               " FILTER_MAT_DIAG| MPI communication method:", &
               "At each step"
         END IF
         WRITE (UNIT=unit_nr, FMT="(A,T71,g10.4)") &
            " FILTER_MAT_DIAG| Filter temperature [K]:", &
            cp_unit_from_cp2k(filter_temperature, "K")
         WRITE (UNIT=unit_nr, FMT="(A,T71,f10.4)") &
            " FILTER_MAT_DIAG| Filter temperature [a.u.]:", &
            filter_temperature
         WRITE (UNIT=unit_nr, FMT="(A,T71,f10.4)") &
            " FILTER_MAT_DIAG| Auto atomic cutoff radius scale:", &
            auto_cutoff_scale
         WRITE (UNIT=unit_nr, FMT="(A)") &
            " FILTER_MAT_DIAG| atomic cutoff radii [a.u.]"
         DO ikind = 1, nkinds
            CALL get_atomic_kind(atomic_kind=atomic_kind_set(ikind), &
                                 element_symbol=element_symbol)
            WRITE (UNIT=unit_nr, FMT="(A,A,T71,f10.4)") &
               " FILTER_MAT_DIAG|   ", element_symbol, rcut(ikind)
         END DO ! ikind
      END IF
      CALL cp_print_key_finished_output(unit_nr, logger, scf_section, &
                                        "PRINT%FILTER_MATRIX")

      CALL timestop(handle)

   END SUBROUTINE fb_env_write_info

END MODULE qs_fb_env_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basic functionality for using ot in the scf routines.
!> \par History
!>      01.2003 : Joost VandeVondele : adapted for LSD
!> \author Joost VandeVondele (25.08.2002)
! **************************************************************************************************
MODULE qs_ot_scf
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_get_diag, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_scale_by_vector, &
        cp_dbcsr_set, cp_dbcsr_set_diag, cp_dbcsr_trace, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr,&
                                              cp_dbcsr_m_by_n_from_row_template
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_restrict
   USE qs_ot,                           ONLY: qs_ot_get_orbitals,&
                                              qs_ot_get_orbitals_ref,&
                                              qs_ot_get_p
   USE qs_ot_minimizer,                 ONLY: ot_mini
   USE qs_ot_types,                     ONLY: ot_readwrite_input,&
                                              qs_ot_allocate,&
                                              qs_ot_destroy,&
                                              qs_ot_init,&
                                              qs_ot_settings_init,&
                                              qs_ot_type
   USE scf_control_types,               ONLY: smear_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ot_scf'
   ! *** Public subroutines ***

   PUBLIC :: ot_scf_init
   PUBLIC :: ot_scf_mini
   PUBLIC :: ot_scf_destroy
   PUBLIC :: ot_scf_read_input

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param scf_section ...
! **************************************************************************************************
   SUBROUTINE ot_scf_read_input(qs_ot_env, scf_section)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_scf_read_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, output_unit
      LOGICAL                                            :: explicit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: ot_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".log")

      ! decide default settings
      CALL qs_ot_settings_init(qs_ot_env(1)%settings)

      ! use ot input new style
      ot_section => section_vals_get_subs_vals(scf_section, "OT")
      CALL section_vals_get(ot_section, explicit=explicit)

      CALL ot_readwrite_input(qs_ot_env(1)%settings, ot_section, output_unit)

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      ! copy the ot settings type so it is identical
      nspin = SIZE(qs_ot_env)
      DO ispin = 2, nspin
         qs_ot_env(ispin)%settings = qs_ot_env(1)%settings
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE ot_scf_read_input
! **************************************************************************************************
   !
   ! performs the actual minimisation, needs only limited info
   ! updated for restricted calculations
   ! matrix_dedc is the derivative of the energy with respect to the orbitals (except for a factor 2*fi)
   ! a null pointer for matrix_s implies that matrix_s is the unit matrix
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param matrix_dedc ...
!> \param smear ...
!> \param matrix_s ...
!> \param energy ...
!> \param energy_only ...
!> \param delta ...
!> \param qs_ot_env ...
!> \param input ...
! **************************************************************************************************
   SUBROUTINE ot_scf_mini(mo_array, matrix_dedc, smear, matrix_s, energy, &
                          energy_only, delta, qs_ot_env, input)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_dedc
      TYPE(smear_type), POINTER                          :: smear
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s
      REAL(KIND=dp)                                      :: energy
      LOGICAL, INTENT(INOUT)                             :: energy_only
      REAL(KIND=dp)                                      :: delta
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env
      TYPE(section_vals_type), POINTER                   :: input

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_scf_mini', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, k, n, nspin, output_unit
      REAL(KIND=dp)                                      :: ener_nondiag, trace
      TYPE(cp_1d_r_p_type), ALLOCATABLE, DIMENSION(:)    :: expectation_values, occupation_numbers, &
                                                            scaling_factor
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_dedc_scaled
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      nspin = SIZE(mo_array)

      ALLOCATE (occupation_numbers(nspin))
      ALLOCATE (scaling_factor(nspin))

      IF (qs_ot_env(1)%settings%do_ener) THEN
         ALLOCATE (expectation_values(nspin))
      ENDIF

      DO ispin = 1, nspin
         CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, occupation_numbers=occupation_numbers(ispin)%array)
         ALLOCATE (scaling_factor(ispin)%array(SIZE(occupation_numbers(ispin)%array)))
         scaling_factor(ispin)%array = 2.0_dp*occupation_numbers(ispin)%array
         IF (qs_ot_env(1)%settings%do_ener) THEN
            ALLOCATE (expectation_values(ispin)%array(SIZE(occupation_numbers(ispin)%array)))
         ENDIF
      ENDDO

      ! optimizing orbital energies somehow implies non-equivalent orbitals
      IF (qs_ot_env(1)%settings%do_ener) THEN
         IF(.NOT.(qs_ot_env(1)%settings%do_rotation))CALL cp__a("qs_ot_scf.F",172)
      ENDIF
      ! add_nondiag_energy requires do_ener
      IF (qs_ot_env(1)%settings%add_nondiag_energy) THEN
         IF(.NOT.(qs_ot_env(1)%settings%do_ener))CALL cp__a("qs_ot_scf.F",176)
      ENDIF

      ! get a rotational force
      IF (.NOT. energy_only) THEN
         IF (qs_ot_env(1)%settings%do_rotation) THEN
            DO ispin = 1, SIZE(qs_ot_env)
               CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff_b=mo_coeff)
               CALL cp_dbcsr_get_info(mo_coeff, nfullrows_total=n, nfullcols_total=k)
               CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, mo_coeff, matrix_dedc(ispin)%matrix, &
                                      0.0_dp, qs_ot_env(ispin)%rot_mat_chc)
               CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_buf1, qs_ot_env(ispin)%rot_mat_chc)

               CALL cp_dbcsr_scale_by_vector(qs_ot_env(ispin)%matrix_buf1, alpha=scaling_factor(ispin)%array, side='right')
               ! create the derivative of the energy wrt to rot_mat_u
               CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, qs_ot_env(ispin)%rot_mat_u, qs_ot_env(ispin)%matrix_buf1, &
                                      0.0_dp, qs_ot_env(ispin)%rot_mat_dedu)
            ENDDO

            ! here we construct the derivative of the free energy with respect to the evals
            ! (note that this requires the diagonal elements of chc)
            ! the mo occupations should in principle remain unaltered
            IF (qs_ot_env(1)%settings%do_ener) THEN
               DO ispin = 1, SIZE(mo_array)
                  CALL cp_dbcsr_get_diag(qs_ot_env(ispin)%rot_mat_chc, expectation_values(ispin)%array)
                  qs_ot_env(ispin)%ener_gx = expectation_values(ispin)%array
                  CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                         smear=smear, eval_deriv=qs_ot_env(ispin)%ener_gx)
               ENDDO
            ENDIF

            ! chc only needs to be stored in u independent form if we require add_nondiag_energy,
            ! which will use it in non-selfconsistent form for e.g. the linesearch
            ! transform C^T H C -> U C^T H C U ^ T
            IF (qs_ot_env(1)%settings%add_nondiag_energy) THEN
               DO ispin = 1, SIZE(qs_ot_env)
                  CALL cp_dbcsr_get_info(qs_ot_env(ispin)%rot_mat_u, nfullcols_total=k)
                  CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, qs_ot_env(ispin)%rot_mat_u, qs_ot_env(ispin)%rot_mat_chc, &
                                         0.0_dp, qs_ot_env(ispin)%matrix_buf1)
                  CALL cp_dbcsr_multiply('N', 'T', 1.0_dp, qs_ot_env(ispin)%matrix_buf1, qs_ot_env(ispin)%rot_mat_u, &
                                         0.0_dp, qs_ot_env(ispin)%rot_mat_chc)
               ENDDO
            ENDIF
         ENDIF
      ENDIF

      ! evaluate non-diagonal energy contribution
      ener_nondiag = 0.0_dp
      IF (qs_ot_env(1)%settings%add_nondiag_energy) THEN
         DO ispin = 1, SIZE(qs_ot_env)
            ! transform \tilde H to the current basis of C (assuming non-selfconsistent H)
            CALL cp_dbcsr_get_info(qs_ot_env(ispin)%rot_mat_u, nfullcols_total=k)
            CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, qs_ot_env(ispin)%rot_mat_u, qs_ot_env(ispin)%rot_mat_chc, &
                                   0.0_dp, qs_ot_env(ispin)%matrix_buf1)
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, qs_ot_env(ispin)%matrix_buf1, qs_ot_env(ispin)%rot_mat_u, &
                                   0.0_dp, qs_ot_env(ispin)%matrix_buf2)

            ! subtract the current ener_x from the diagonal
            CALL cp_dbcsr_get_diag(qs_ot_env(ispin)%matrix_buf2, expectation_values(ispin)%array)
            expectation_values(ispin)%array = expectation_values(ispin)%array-qs_ot_env(ispin)%ener_x
            CALL cp_dbcsr_set_diag(qs_ot_env(ispin)%matrix_buf2, expectation_values(ispin)%array)

            ! get nondiag energy trace (D^T D)
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_buf2, qs_ot_env(ispin)%matrix_buf2, trace)
            ener_nondiag = ener_nondiag+0.5_dp*qs_ot_env(1)%settings%nondiag_energy_strength*trace

            ! get gradient (again ignoring dependencies of H)
            IF (.NOT. energy_only) THEN
               ! first for the ener_x (-2*(diag(C^T H C)-ener_x))
               qs_ot_env(ispin)%ener_gx = qs_ot_env(ispin)%ener_gx- &
                                          qs_ot_env(1)%settings%nondiag_energy_strength*expectation_values(ispin)%array

               ! next for the rot_mat_u derivative (2 * k * \tilde H U D)
               CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, qs_ot_env(ispin)%rot_mat_chc, qs_ot_env(ispin)%rot_mat_u, &
                                      0.0_dp, qs_ot_env(ispin)%matrix_buf1)
               CALL cp_dbcsr_multiply('N', 'N', 2.0_dp*qs_ot_env(1)%settings%nondiag_energy_strength, &
                                      qs_ot_env(ispin)%matrix_buf1, qs_ot_env(ispin)%matrix_buf2, &
                                      1.0_dp, qs_ot_env(ispin)%rot_mat_dedu)
            ENDIF
         ENDDO
      ENDIF

      ! this is kind of a hack so far (costly memory wise), we locally recreate the scaled matrix_hc, and
      ! use it in the following, eventually, as occupations numbers get more integrated, it should become possible
      ! to remove this.
      ALLOCATE (matrix_dedc_scaled(SIZE(matrix_dedc)))
      DO ispin = 1, SIZE(matrix_dedc)
         ALLOCATE (matrix_dedc_scaled(ispin)%matrix)
         CALL cp_dbcsr_init(matrix_dedc_scaled(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_dedc_scaled(ispin)%matrix, matrix_dedc(ispin)%matrix)

         ! as a preconditioner, one might want to scale only with a constant, not with f(i)
         ! for the convergence criterium, maybe take it back out
         IF (qs_ot_env(1)%settings%occupation_preconditioner) THEN
            scaling_factor(ispin)%array = 2.0_dp
         ENDIF
         CALL cp_dbcsr_scale_by_vector(matrix_dedc_scaled(ispin)%matrix, alpha=scaling_factor(ispin)%array, side='right')
      ENDDO

      ! notice we use qs_ot_env(1) for driving all output and the minimization in case of LSD
      qs_ot_env(1)%etotal = energy+ener_nondiag

      output_unit = cp_print_key_unit_nr(logger, input, "DFT%SCF%PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")
      CALL ot_mini(qs_ot_env, matrix_dedc_scaled, output_unit)

      delta = qs_ot_env(1)%delta
      energy_only = qs_ot_env(1)%energy_only

      ! generate the orbitals using the new matrix_x
      DO ispin = 1, SIZE(qs_ot_env)
         CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff_b=mo_coeff)
         CALL cp_dbcsr_get_info(mo_coeff, nfullrows_total=n, nfullcols_total=k)
         SELECT CASE (qs_ot_env (1)%settings%ot_algorithm)
         CASE ("TOD")
            IF (ASSOCIATED(matrix_s)) THEN
               CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, qs_ot_env(ispin)%matrix_x, &
                                      0.0_dp, qs_ot_env(ispin)%matrix_sx)
            ELSE
               CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_sx, qs_ot_env(ispin)%matrix_x)
            ENDIF
            CALL qs_ot_get_p(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_sx, qs_ot_env(ispin))
            CALL qs_ot_get_orbitals(mo_coeff, qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin))
         CASE ("REF")
            CALL qs_ot_get_orbitals_ref(mo_coeff, matrix_s, qs_ot_env(ispin)%matrix_x, &
                                        qs_ot_env(ispin)%matrix_sx, qs_ot_env(ispin)%matrix_gx_old, &
                                        qs_ot_env(ispin)%matrix_dx, qs_ot_env(ispin), qs_ot_env(1), &
                                        output_unit)
         CASE DEFAULT
            CALL cp__b("qs_ot_scf.F",305,"Algorithm not yet implemented")
         END SELECT
      ENDDO

      IF (qs_ot_env(1)%restricted) THEN
         CALL mo_set_restrict(mo_array, convert_dbcsr=.TRUE.)
      ENDIF
      !
      ! obtain the new set of OT eigenvalues and set the occupations accordingly
      !
      IF (qs_ot_env(1)%settings%do_ener) THEN
         DO ispin = 1, SIZE(mo_array)
            mo_array(ispin)%mo_set%eigenvalues = qs_ot_env(ispin)%ener_x
            CALL set_mo_occupation(mo_set=mo_array(ispin)%mo_set, &
                                   smear=smear)
         ENDDO
      ENDIF

      ! cleanup
      CALL cp_print_key_finished_output(output_unit, logger, input, &
                                        "DFT%SCF%PRINT%PROGRAM_RUN_INFO")
      DO ispin = 1, SIZE(scaling_factor)
         DEALLOCATE (scaling_factor(ispin)%array)
      ENDDO
      DEALLOCATE (scaling_factor)
      IF (qs_ot_env(1)%settings%do_ener) THEN
         DO ispin = 1, SIZE(expectation_values)
            DEALLOCATE (expectation_values(ispin)%array)
         ENDDO
         DEALLOCATE (expectation_values)
      ENDIF
      DEALLOCATE (occupation_numbers)
      DO ispin = 1, SIZE(matrix_dedc_scaled)
         CALL cp_dbcsr_release(matrix_dedc_scaled(ispin)%matrix)
         DEALLOCATE (matrix_dedc_scaled(ispin)%matrix)
      ENDDO
      DEALLOCATE (matrix_dedc_scaled)

      CALL timestop(handle)

   END SUBROUTINE ot_scf_mini
   !
   ! initialises qs_ot_env so that mo_coeff is the current point
   ! and that the mimizization can be started.
   !
! **************************************************************************************************
!> \brief ...
!> \param mo_array ...
!> \param matrix_s ...
!> \param qs_ot_env ...
!> \param matrix_ks ...
!> \param broyden_adaptive_sigma ...
! **************************************************************************************************
   SUBROUTINE ot_scf_init(mo_array, matrix_s, qs_ot_env, matrix_ks, broyden_adaptive_sigma)

      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_ks
      REAL(KIND=dp)                                      :: broyden_adaptive_sigma

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_scf_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, k, n, nspin
      LOGICAL                                            :: is_equal
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff
      TYPE(cp_fm_type), POINTER                          :: mo_coeff_fm

      CALL timeset(routineN, handle)

      DO ispin = 1, SIZE(mo_array)
         IF (.NOT. ASSOCIATED(mo_array(ispin)%mo_set%mo_coeff_b)) THEN
            CALL cp__b("qs_ot_scf.F",377,"Shouldn't get there")
            ! we do ot then copy fm to dbcsr
            ! allocate that somewhere else ! fm -> dbcsr
            CALL cp_dbcsr_init_p(mo_array(ispin)%mo_set%mo_coeff_b)
            CALL cp_dbcsr_m_by_n_from_row_template(mo_array(ispin)%mo_set%mo_coeff_b, template=matrix_ks, &
                                                   n=mo_array(ispin)%mo_set%nmo, &
                                                   sym=dbcsr_type_no_symmetry)
         ENDIF
      ENDDO

      ! *** set a history for broyden
      DO ispin = 1, SIZE(qs_ot_env)
         qs_ot_env(ispin)%broyden_adaptive_sigma = broyden_adaptive_sigma
      ENDDO

      ! **** SCP
      ! **** SCP
      ! adapted for work with the restricted keyword
      nspin = SIZE(qs_ot_env)

      DO ispin = 1, nspin

         CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, mo_coeff_b=mo_coeff, mo_coeff=mo_coeff_fm)
         CALL copy_fm_to_dbcsr(mo_coeff_fm, mo_coeff) !fm -> dbcsr

         CALL cp_dbcsr_get_info(mo_coeff, nfullrows_total=n, nfullcols_total=k)

         ! allocate
         CALL qs_ot_allocate(qs_ot_env(ispin), matrix_ks, mo_coeff_fm%matrix_struct)

         ! set c0,sc0
         CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_c0, mo_coeff)
         IF (ASSOCIATED(matrix_s)) THEN
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, qs_ot_env(ispin)%matrix_c0, &
                                   0.0_dp, qs_ot_env(ispin)%matrix_sc0)
         ELSE
            CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_sc0, qs_ot_env(ispin)%matrix_c0)
         ENDIF

         ! init
         CALL qs_ot_init(qs_ot_env(ispin))

         ! set x
         CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_x, 0.0_dp)
         CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_sx, 0.0_dp)

         IF (qs_ot_env(ispin)%settings%do_rotation) THEN
            CALL cp_dbcsr_set(qs_ot_env(ispin)%rot_mat_x, 0.0_dp)
         ENDIF

         IF (qs_ot_env(ispin)%settings%do_ener) THEN
            is_equal = SIZE(qs_ot_env(ispin)%ener_x) == SIZE(mo_array(ispin)%mo_set%eigenvalues)
            IF(.NOT.(is_equal))CALL cp__a("qs_ot_scf.F",429)
            qs_ot_env(ispin)%ener_x = mo_array(ispin)%mo_set%eigenvalues
         ENDIF

         SELECT CASE (qs_ot_env (1)%settings%ot_algorithm)
         CASE ("TOD")
            ! get c
            CALL qs_ot_get_p(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_sx, qs_ot_env(ispin))
         CASE ("REF")
            CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_c0)
            CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_sx, qs_ot_env(ispin)%matrix_sc0)
         CASE DEFAULT
            CALL cp__b("qs_ot_scf.F",441,"Algorithm not yet implemented")
         END SELECT

      END DO
      CALL timestop(handle)
   END SUBROUTINE ot_scf_init

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE ot_scf_destroy(qs_ot_env)

      TYPE(qs_ot_type)                                   :: qs_ot_env

      CALL qs_ot_destroy(qs_ot_env)

   END SUBROUTINE ot_scf_destroy

END MODULE qs_ot_scf

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_scf_output
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: evolt,&
                                              kcalmol
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE pw_env_types,                    ONLY: pw_env_type
   USE pw_poisson_types,                ONLY: pw_poisson_implicit
   USE qmmm_image_charge,               ONLY: print_image_coefficients
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_mo_io,                        ONLY: write_mo_set
   USE qs_mo_methods,                   ONLY: calculate_magnitude,&
                                              calculate_orthonormality
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              special_diag_method_nr
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_output'

   PUBLIC :: qs_scf_loop_info, &
             qs_scf_print_summary, &
             qs_scf_loop_print, &
             qs_scf_outer_loop_info, &
             qs_scf_initial_info, &
             qs_scf_write_mos

CONTAINS

! **************************************************************************************************
!> \brief writes a summary of information after scf
!> \param output_unit ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_print_summary(output_unit, qs_env)
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_print_summary', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nelectron_total
      LOGICAL                                            :: gapw, gapw_xc, qmmm
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      NULLIFY (rho, energy, dft_control, scf_env, qs_charges)
      CALL get_qs_env(qs_env=qs_env, rho=rho, energy=energy, dft_control=dft_control, &
                      scf_env=scf_env, qs_charges=qs_charges)

      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      qmmm = qs_env%qmmm
      nelectron_total = scf_env%nelectron

      CALL qs_scf_print_scf_summary(output_unit, rho, qs_charges, energy, nelectron_total, &
                                    dft_control, qmmm, qs_env, gapw, gapw_xc)

   END SUBROUTINE qs_scf_print_summary

! **************************************************************************************************
!> \brief writes basic information at the beginning of an scf run
!> \param output_unit ...
!> \param mos ...
!> \param dft_control ...
! **************************************************************************************************
   SUBROUTINE qs_scf_initial_info(output_unit, mos, dft_control)
      INTEGER                                            :: output_unit
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_initial_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, ispin, nao, nelectron_spin, nmo

! print occupation numbers

      IF (output_unit > 0) THEN
         DO ispin = 1, dft_control%nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, &
                            homo=homo, &
                            nelectron=nelectron_spin, &
                            nao=nao, &
                            nmo=nmo)
            IF (dft_control%nspins > 1) THEN
               WRITE (UNIT=output_unit, FMT="(/,T2,A,I2)") "Spin", ispin
            END IF
            WRITE (UNIT=output_unit, FMT="(/,(T2,A,T71,I10))") &
               "Number of electrons:", nelectron_spin, &
               "Number of occupied orbitals:", homo, &
               "Number of molecular orbitals:", nmo
         END DO
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T71,I10)") &
            "Number of orbital functions:", nao
      END IF

   END SUBROUTINE qs_scf_initial_info

! **************************************************************************************************
!> \brief writes out the mos in the scf loop if needed
!> \param mos ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE qs_scf_write_mos(mos, atomic_kind_set, qs_kind_set, particle_set, dft_section)
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_write_mos', &
         routineP = moduleN//':'//routineN

      IF (SIZE(mos) > 1) THEN
         CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, spin="ALPHA", last=.FALSE.)
         CALL write_mo_set(mos(2)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, spin="BETA", last=.FALSE.)
      ELSE
         CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, last=.FALSE.)
      END IF

   END SUBROUTINE qs_scf_write_mos
! **************************************************************************************************
!> \brief writes basic information obtained in a scf outer loop step
!> \param output_unit ...
!> \param scf_control ...
!> \param scf_env ...
!> \param energy ...
!> \param total_steps ...
!> \param should_stop ...
!> \param outer_loop_converged ...
! **************************************************************************************************

   SUBROUTINE qs_scf_outer_loop_info(output_unit, scf_control, scf_env, &
                                     energy, total_steps, should_stop, outer_loop_converged)
      INTEGER                                            :: output_unit
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_energy_type), POINTER                      :: energy
      INTEGER                                            :: total_steps
      LOGICAL, INTENT(IN)                                :: should_stop, outer_loop_converged

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_outer_loop_info', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: outer_loop_eps

      outer_loop_eps = SQRT(SUM(scf_env%outer_scf%gradient(:, scf_env%outer_scf%iter_count)**2))/ &
                       SIZE(scf_env%outer_scf%gradient, 1)
      IF (output_unit > 0) WRITE (output_unit, '(/,T3,A,I4,A,E10.2,A,F22.10)') &
         "outer SCF iter = ", scf_env%outer_scf%iter_count, &
         " RMS gradient = ", outer_loop_eps, " energy =", energy%total

      IF (outer_loop_converged) THEN
         IF (output_unit > 0) WRITE (output_unit, '(T3,A,I4,A,I4,A,/)') &
            "outer SCF loop converged in", scf_env%outer_scf%iter_count, &
            " iterations or ", total_steps, " steps"
      END IF

      IF (scf_env%outer_scf%iter_count > scf_control%outer_scf%max_scf &
          .OR. should_stop) THEN
         IF (output_unit > 0) WRITE (output_unit, '(T3,A,I4,A,I4,A,/)') &
            "outer SCF loop FAILED to converge after ", &
            scf_env%outer_scf%iter_count, " iterations or ", total_steps, " steps"
      END IF

   END SUBROUTINE qs_scf_outer_loop_info

! **************************************************************************************************
!> \brief writes basic information obtained in a scf step
!> \param scf_env ...
!> \param output_unit ...
!> \param just_energy ...
!> \param t1 ...
!> \param t2 ...
!> \param energy ...
! **************************************************************************************************

   SUBROUTINE qs_scf_loop_info(scf_env, output_unit, just_energy, t1, t2, energy)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      INTEGER                                            :: output_unit
      LOGICAL                                            :: just_energy
      REAL(KIND=dp)                                      :: t1, t2
      TYPE(qs_energy_type), POINTER                      :: energy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_loop_info', &
         routineP = moduleN//':'//routineN

      IF ((output_unit > 0) .AND. scf_env%print_iter_line) THEN
         IF (just_energy) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(T2,I5,1X,A,T20,E8.2,1X,F6.1,16X,F20.10)") &
               scf_env%iter_count, TRIM(scf_env%iter_method), scf_env%iter_param, &
               t2-t1, energy%total
         ELSE
            IF ((ABS(scf_env%iter_delta) < 1.0E-8_dp) .OR. &
                (ABS(scf_env%iter_delta) >= 1.0E5_dp)) THEN
               WRITE (UNIT=output_unit, &
                      FMT="(T2,I5,1X,A,T20,E8.2,1X,F6.1,1X,ES14.4,1X,F20.10,1X,ES9.2)") &
                  scf_env%iter_count, TRIM(scf_env%iter_method), scf_env%iter_param, &
                  t2-t1, scf_env%iter_delta, energy%total, energy%total-energy%tot_old
            ELSE
               WRITE (UNIT=output_unit, &
                      FMT="(T2,I5,1X,A,T20,E8.2,1X,F6.1,1X,F14.8,1X,F20.10,1X,ES9.2)") &
                  scf_env%iter_count, TRIM(scf_env%iter_method), scf_env%iter_param, &
                  t2-t1, scf_env%iter_delta, energy%total, energy%total-energy%tot_old
            END IF
         END IF
      END IF

   END SUBROUTINE qs_scf_loop_info

! **************************************************************************************************
!> \brief writes rather detailed summary of densities and energies
!>      after the SCF
!> \param output_unit ...
!> \param rho ...
!> \param qs_charges ...
!> \param energy ...
!> \param nelectron_total ...
!> \param dft_control ...
!> \param qmmm ...
!> \param qs_env ...
!> \param gapw ...
!> \param gapw_xc ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE qs_scf_print_scf_summary(output_unit, rho, qs_charges, energy, nelectron_total, &
                                       dft_control, qmmm, qs_env, gapw, gapw_xc)
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_energy_type), POINTER                      :: energy
      INTEGER, INTENT(IN)                                :: nelectron_total
      TYPE(dft_control_type), POINTER                    :: dft_control
      LOGICAL, INTENT(IN)                                :: qmmm
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: gapw, gapw_xc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_print_scf_summary', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bc, handle, ispin, psolver
      REAL(kind=dp)                                      :: implicit_ps_ehartree, tot1_h, tot1_s
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r
      TYPE(pw_env_type), POINTER                         :: pw_env

      NULLIFY (tot_rho_r, pw_env)
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      psolver = pw_env%poisson_env%parameters%solver

      IF (output_unit > 0) THEN
         CALL qs_rho_get(rho, tot_rho_r=tot_rho_r)
         IF (.NOT. (dft_control%qs_control%semi_empirical .OR. dft_control%qs_control%dftb .OR. &
                    dft_control%qs_control%scptb)) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T41,2F20.10))") &
               "Electronic density on regular grids: ", &
               accurate_sum(tot_rho_r), &
               accurate_sum(tot_rho_r)+nelectron_total, &
               "Core density on regular grids:", &
               qs_charges%total_rho_core_rspace, &
               qs_charges%total_rho_core_rspace-REAL(nelectron_total+dft_control%charge, dp)
            IF (gapw) THEN
               tot1_h = qs_charges%total_rho1_hard(1)
               tot1_s = qs_charges%total_rho1_soft(1)
               DO ispin = 2, dft_control%nspins
                  tot1_h = tot1_h+qs_charges%total_rho1_hard(ispin)
                  tot1_s = tot1_s+qs_charges%total_rho1_soft(ispin)
               END DO
               WRITE (UNIT=output_unit, FMT="((T3,A,T41,2F20.10))") &
                  "Hard and soft densities (Lebedev):", &
                  tot1_h, tot1_s
               WRITE (UNIT=output_unit, FMT="(T3,A,T41,F20.10)") &
                  "Total Rho_soft + Rho1_hard - Rho1_soft (r-space): ", &
                  accurate_sum(tot_rho_r)+tot1_h-tot1_s, &
                  "Total charge density (r-space):      ", &
                  accurate_sum(tot_rho_r)+tot1_h-tot1_s &
                  +qs_charges%total_rho_core_rspace, &
                  "Total Rho_soft + Rho0_soft (g-space):", &
                  qs_charges%total_rho_gspace
            ELSE
               WRITE (UNIT=output_unit, FMT="(T3,A,T41,F20.10)") &
                  "Total charge density on r-space grids:     ", &
                  accurate_sum(tot_rho_r)+ &
                  qs_charges%total_rho_core_rspace, &
                  "Total charge density g-space grids:     ", &
                  qs_charges%total_rho_gspace
            END IF
         END IF
         IF (dft_control%qs_control%semi_empirical) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Core-core repulsion energy [eV]:               ", energy%core_overlap*evolt, &
               "Core Hamiltonian energy [eV]:                  ", energy%core*evolt, &
               "Two-electron integral energy [eV]:             ", energy%hartree*evolt, &
               "Electronic energy [eV]:                        ", &
               (energy%core+0.5_dp*energy%hartree)*evolt
            IF (energy%dispersion /= 0.0_dp) &
               WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "Dispersion energy [eV]:                     ", energy%dispersion*evolt
         ELSEIF (dft_control%qs_control%dftb) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Core Hamiltonian energy:                       ", energy%core, &
               "Repulsive potential energy:                    ", energy%repulsive, &
               "Electronic energy:                             ", energy%hartree, &
               "Dispersion energy:                             ", energy%dispersion
            IF (energy%dftb3 /= 0.0_dp) &
               WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "DFTB3 3rd order energy:                     ", energy%dftb3
         ELSEIF (dft_control%qs_control%scptb) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Repulsive pair potential energy:               ", energy%repulsive, &
               "Zeroth order Hamiltonian energy:               ", energy%core, &
               "Kinetic energy:                                ", energy%kinetic, &
               "Charge fluctuation energy:                     ", energy%hartree, &
               "London dispersion energy:                      ", energy%dispersion
         ELSE
            IF (dft_control%do_admm) THEN
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                  "Overlap energy of the core charge distribution:", energy%core_overlap, &
                  "Self energy of the core charge distribution:   ", energy%core_self, &
                  "Core Hamiltonian energy:                       ", energy%core, &
                  "Hartree energy:                                ", energy%hartree, &
                  "Exchange-correlation energy:                   ", energy%exc+energy%exc_aux_fit
            ELSE
               IF (psolver .EQ. pw_poisson_implicit) THEN
                  implicit_ps_ehartree = pw_env%poisson_env%implicit_env%ehartree
                  bc = pw_env%poisson_env%parameters%ps_implicit_params%boundary_condition
                  SELECT CASE (bc)
                  CASE (MIXED_PERIODIC_BC, MIXED_BC)
                     WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                        "Overlap energy of the core charge distribution:", energy%core_overlap, &
                        "Self energy of the core charge distribution:   ", energy%core_self, &
                        "Core Hamiltonian energy:                       ", energy%core, &
                        "Hartree energy:                                ", implicit_ps_ehartree, &
                        "Electric enthalpy:                             ", energy%hartree, &
                        "Exchange-correlation energy:                   ", energy%exc
                  CASE (PERIODIC_BC, NEUMANN_BC)
                     WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                        "Overlap energy of the core charge distribution:", energy%core_overlap, &
                        "Self energy of the core charge distribution:   ", energy%core_self, &
                        "Core Hamiltonian energy:                       ", energy%core, &
                        "Hartree energy:                                ", energy%hartree, &
                        "Exchange-correlation energy:                   ", energy%exc
                  END SELECT
               ELSE
                  WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                     "Overlap energy of the core charge distribution:", energy%core_overlap, &
                     "Self energy of the core charge distribution:   ", energy%core_self, &
                     "Core Hamiltonian energy:                       ", energy%core, &
                     "Hartree energy:                                ", energy%hartree, &
                     "Exchange-correlation energy:                   ", energy%exc
               END IF
            END IF
            IF (energy%e_hartree /= 0.0_dp) &
               WRITE (UNIT=output_unit, FMT="(T3,A,/,T3,A,T56,F25.14)") &
               "Coulomb Electron-Electron Interaction Energy ", &
               "- Already included in the total Hartree term ", energy%e_hartree
            IF (energy%ex /= 0.0_dp) &
               WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "Hartree-Fock Exchange energy:                  ", energy%ex
            IF (energy%dispersion /= 0.0_dp) &
               WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "Dispersion energy:                             ", energy%dispersion
            IF (gapw) THEN
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                  "GAPW| Exc from hard and soft atomic rho1:      ", energy%exc1, &
                  "GAPW| local Eh = 1 center integrals:           ", energy%hartree_1c
            END IF
            IF (gapw_xc) THEN
               WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
                  "GAPW_XC| Exc from hard and soft atomic rho1:      ", energy%exc1
            END IF
         END IF
         IF (dft_control%smear) THEN
            WRITE (UNIT=output_unit, FMT="((T3,A,T56,F25.14))") &
               "Electronic entropic energy:", energy%kTS
            WRITE (UNIT=output_unit, FMT="((T3,A,T56,F25.14))") &
               "Fermi energy:", energy%efermi
         END IF
         IF (dft_control%dft_plus_u) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "DFT+U energy:", energy%dft_plus_u
         END IF
         IF (dft_control%do_sccs) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,T56,F25.14)") &
               "SCCS| Hartree energy of solute and solvent [Hartree]", energy%sccs_hartree
            WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14,/,T3,A,T61,F20.3)") &
               "SCCS| Polarisation energy                  [Hartree]", energy%sccs_pol, &
               "SCCS|                                      [kcal/mol]", &
               cp_unit_from_cp2k(energy%sccs_pol, "kcalmol")
         END IF
         IF (qmmm) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "QM/MM Electrostatic energy:                    ", energy%qmmm_el
            IF (qs_env%qmmm_env_qm%image_charge) THEN
               WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
                  "QM/MM image charge energy:                ", energy%image_charge
            ENDIF
         END IF
         IF (dft_control%qs_control%mulliken_restraint) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T56,F25.14)") &
               "Mulliken restraint energy: ", energy%mulliken
         END IF
         IF (dft_control%qs_control%semi_empirical) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Total energy [eV]:                             ", energy%total*evolt
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Atomic reference energy [eV]:                  ", energy%core_self*evolt, &
               "Heat of formation [kcal/mol]:                  ", &
               (energy%total+energy%core_self)*kcalmol
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T56,F25.14))") &
               "Total energy:                                  ", energy%total
         END IF
         IF (qmmm) THEN
            IF (qs_env%qmmm_env_qm%image_charge) THEN
               CALL print_image_coefficients(qs_env%image_coeff, qs_env)
            ENDIF
         ENDIF
         CALL m_flush(output_unit)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_print_scf_summary

! **************************************************************************************************
!> \brief collects the 'heavy duty' printing tasks out of the SCF loop
!> \param qs_env ...
!> \param scf_env ...
!> \param para_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE qs_scf_loop_print(qs_env, scf_env, para_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_loop_print', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: after, handle, ic, ispin, iw
      LOGICAL                                            :: do_kpoints, omit_headers
      REAL(KIND=dp)                                      :: mo_mag_max, mo_mag_min, orthonormality
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks, matrix_p, matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: dft_section, input, scf_section

      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env, input=input, dft_control=dft_control, &
                      do_kpoints=do_kpoints)

      dft_section => section_vals_get_subs_vals(input, "DFT")
      scf_section => section_vals_get_subs_vals(dft_section, "SCF")

      CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
      DO ispin = 1, dft_control%nspins

         IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                              dft_section, "PRINT%AO_MATRICES/DENSITY"), cp_p_file)) THEN
            CALL get_qs_env(qs_env, rho=rho)
            CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
            iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%AO_MATRICES/DENSITY", &
                                      extension=".Log")
            CALL section_vals_val_get(dft_section, "PRINT%AO_MATRICES%NDIGITS", i_val=after)
            after = MIN(MAX(after, 1), 16)
            DO ic = 1, SIZE(matrix_p, 2)
               CALL cp_dbcsr_write_sparse_matrix(matrix_p(ispin, ic)%matrix, 4, after, qs_env, para_env, &
                                                 output_unit=iw, omit_headers=omit_headers)
            END DO
            CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                              "PRINT%AO_MATRICES/DENSITY")
         END IF

         IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                              dft_section, "PRINT%AO_MATRICES/KOHN_SHAM_MATRIX"), cp_p_file)) THEN
            iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%AO_MATRICES/KOHN_SHAM_MATRIX", &
                                      extension=".Log")
            CALL section_vals_val_get(dft_section, "PRINT%AO_MATRICES%NDIGITS", i_val=after)
            after = MIN(MAX(after, 1), 16)
            CALL get_qs_env(qs_env=qs_env, matrix_ks_kp=matrix_ks)
            DO ic = 1, SIZE(matrix_ks, 2)
               IF (dft_control%qs_control%semi_empirical) THEN
                  CALL cp_dbcsr_write_sparse_matrix(matrix_ks(ispin, ic)%matrix, 4, after, qs_env, para_env, &
                                                    scale=evolt, output_unit=iw, omit_headers=omit_headers)
               ELSE
                  CALL cp_dbcsr_write_sparse_matrix(matrix_ks(ispin, ic)%matrix, 4, after, qs_env, para_env, &
                                                    output_unit=iw, omit_headers=omit_headers)
               END IF
            END DO
            CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                              "PRINT%AO_MATRICES/KOHN_SHAM_MATRIX")
         END IF

      ENDDO

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           scf_section, "PRINT%MO_ORTHONORMALITY"), cp_p_file)) THEN
         IF (do_kpoints) THEN
            iw = cp_print_key_unit_nr(logger, scf_section, "PRINT%MO_ORTHONORMALITY", &
                                      extension=".scfLog")
            IF (iw > 0) THEN
               WRITE (iw, '(T8,A)') &
                  " K-points: Maximum deviation from MO S-orthonormality not determined"
            ENDIF
            CALL cp_print_key_finished_output(iw, logger, scf_section, &
                                              "PRINT%MO_ORTHONORMALITY")
         ELSE
            CALL get_qs_env(qs_env, mos=mos)
            IF (scf_env%method == special_diag_method_nr) THEN
               CALL calculate_orthonormality(orthonormality, mos)
            ELSE
               CALL get_qs_env(qs_env=qs_env, matrix_s_kp=matrix_s)
               CALL calculate_orthonormality(orthonormality, mos, matrix_s(1, 1)%matrix)
            END IF
            iw = cp_print_key_unit_nr(logger, scf_section, "PRINT%MO_ORTHONORMALITY", &
                                      extension=".scfLog")
            IF (iw > 0) THEN
               WRITE (iw, '(T8,A,T61,E20.4)') &
                  " Maximum deviation from MO S-orthonormality", orthonormality
            ENDIF
            CALL cp_print_key_finished_output(iw, logger, scf_section, &
                                              "PRINT%MO_ORTHONORMALITY")
         END IF
      ENDIF
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           scf_section, "PRINT%MO_MAGNITUDE"), cp_p_file)) THEN
         IF (do_kpoints) THEN
            iw = cp_print_key_unit_nr(logger, scf_section, "PRINT%MO_MAGNITUDE", &
                                      extension=".scfLog")
            IF (iw > 0) THEN
               WRITE (iw, '(T8,A)') &
                  " K-points: Minimum/Maximum MO magnitude not determined"
            ENDIF
            CALL cp_print_key_finished_output(iw, logger, scf_section, &
                                              "PRINT%MO_MAGNITUDE")
         ELSE
            CALL get_qs_env(qs_env, mos=mos)
            CALL calculate_magnitude(mos, mo_mag_min, mo_mag_max)
            iw = cp_print_key_unit_nr(logger, scf_section, "PRINT%MO_MAGNITUDE", &
                                      extension=".scfLog")
            IF (iw > 0) THEN
               WRITE (iw, '(T8,A,T41,2E20.4)') &
                  " Minimum/Maximum MO magnitude ", mo_mag_min, mo_mag_max
            ENDIF
            CALL cp_print_key_finished_output(iw, logger, scf_section, &
                                              "PRINT%MO_MAGNITUDE")
         END IF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_loop_print

END MODULE qs_scf_output
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Utility routines for qs_scf
! **************************************************************************************************
MODULE qs_scf_loop_utils
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_name,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: kpoint_type
   USE qs_density_mixing_types,         ONLY: broyden_mixing_new_nr,&
                                              broyden_mixing_nr,&
                                              direct_mixing_nr,&
                                              gspace_mixing_nr,&
                                              multisecant_mixing_nr,&
                                              no_mixing_nr,&
                                              pulay_mixing_nr
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_fb_env_methods,               ONLY: fb_env_do_diag
   USE qs_gspace_mixing,                ONLY: gspace_mixing,&
                                              self_consistency_check
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE qs_mom_methods,                  ONLY: do_mom_diag
   USE qs_ot_scf,                       ONLY: ot_scf_destroy,&
                                              ot_scf_mini
   USE qs_outer_scf,                    ONLY: outer_loop_gradient
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_diagonalization,          ONLY: do_block_davidson_diag,&
                                              do_block_krylov_diag,&
                                              do_general_diag,&
                                              do_general_diag_kp,&
                                              do_ot_diag,&
                                              do_roks_diag,&
                                              do_scf_diag_subspace,&
                                              do_special_diag
   USE qs_scf_methods,                  ONLY: scf_env_density_mixing
   USE qs_scf_output,                   ONLY: qs_scf_print_summary
   USE qs_scf_types,                    ONLY: block_davidson_diag_method_nr,&
                                              block_krylov_diag_method_nr,&
                                              filter_matrix_diag_method_nr,&
                                              general_diag_method_nr,&
                                              ot_diag_method_nr,&
                                              ot_method_nr,&
                                              qs_scf_env_type,&
                                              special_diag_method_nr
   USE scf_control_types,               ONLY: scf_control_type,&
                                              smear_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_loop_utils'

   PUBLIC :: qs_scf_set_loop_flags, &
             qs_scf_new_mos, qs_scf_new_mos_kp, &
             qs_scf_density_mixing, qs_scf_check_inner_exit, &
             qs_scf_check_outer_exit, qs_scf_inner_finalize, qs_scf_rho_update

CONTAINS

! **************************************************************************************************
!> \brief computes properties for a given hamiltonian using the current wfn
!> \param scf_env ...
!> \param diis_step ...
!> \param energy_only ...
!> \param just_energy ...
!> \param exit_inner_loop ...
! **************************************************************************************************
   SUBROUTINE qs_scf_set_loop_flags(scf_env, diis_step, &
                                    energy_only, just_energy, exit_inner_loop)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      LOGICAL                                            :: diis_step, energy_only, just_energy, &
                                                            exit_inner_loop

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_set_loop_flags', &
         routineP = moduleN//':'//routineN

! Some flags needed to be set at the beginning of the loop

      diis_step = .FALSE.
      energy_only = .FALSE.
      just_energy = .FALSE.

      ! SCF loop, optimisation of the wfn coefficients
      ! qs_env%rho%rho_r and qs_env%rho%rho_g should be up to date here

      scf_env%iter_count = 0
      exit_inner_loop = .FALSE.

   END SUBROUTINE qs_scf_set_loop_flags

! **************************************************************************************************
!> \brief takes known energy and derivatives and produces new wfns
!>        and or density matrix
!> \param qs_env ...
!> \param scf_env ...
!> \param scf_control ...
!> \param scf_section ...
!> \param diis_step ...
!> \param energy_only ...
! **************************************************************************************************
   SUBROUTINE qs_scf_new_mos(qs_env, scf_env, scf_control, scf_section, diis_step, &
                             energy_only)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL                                            :: diis_step, energy_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_new_mos', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      LOGICAL                                            :: has_unit_metric, skip_diag_sub
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)

      NULLIFY (energy, ks_env, matrix_ks, matrix_s, rho, mos, dft_control)

      CALL get_qs_env(qs_env=qs_env, &
                      matrix_s=matrix_s, energy=energy, &
                      ks_env=ks_env, &
                      matrix_ks=matrix_ks, rho=rho, mos=mos, &
                      dft_control=dft_control, &
                      has_unit_metric=has_unit_metric)
      scf_env%iter_param = 0.0_dp

      SELECT CASE (scf_env%method)
      CASE DEFAULT
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",155), &
                       "unknown scf method: "// &
                       cp_to_string(scf_env%method))

         ! *************************************************************************
         ! Filter matrix diagonalisation: ugly implementation at this point of time
         ! *************************************************************************
      CASE (filter_matrix_diag_method_nr)

         CALL fb_env_do_diag(scf_env%filter_matrix_env, qs_env, &
                             matrix_ks, matrix_s, scf_section, diis_step)

         ! Diagonlization in non orthonormal case
      CASE (general_diag_method_nr)
         IF (dft_control%roks) THEN
            CALL do_roks_diag(scf_env, mos, matrix_ks, matrix_s, &
                              scf_control, scf_section, diis_step, &
                              has_unit_metric)
         ELSE
            IF (scf_control%diagonalization%mom) THEN
               CALL do_mom_diag(scf_env, mos, matrix_ks, &
                                matrix_s, scf_control, scf_section, &
                                diis_step)
            ELSE
               CALL do_general_diag(scf_env, mos, matrix_ks, &
                                    matrix_s, scf_control, scf_section, &
                                    diis_step)
            END IF
            IF (scf_control%do_diag_sub) THEN
               skip_diag_sub = (scf_env%subspace_env%eps_diag_sub > 0.0_dp) .AND. &
                               (scf_env%iter_count == 1 .OR. scf_env%iter_delta > scf_env%subspace_env%eps_diag_sub)
               IF (.NOT. skip_diag_sub) THEN
                  CALL do_scf_diag_subspace(qs_env, scf_env, scf_env%subspace_env, mos, rho, &
                                            ks_env, scf_section, scf_control)
               END IF
            END IF
         END IF
         ! Diagonlization in orthonormal case
      CASE (special_diag_method_nr)
         IF (dft_control%roks) THEN
            CALL do_roks_diag(scf_env, mos, matrix_ks, matrix_s, &
                              scf_control, scf_section, diis_step, &
                              has_unit_metric)
         ELSE
            CALL do_special_diag(scf_env, mos, matrix_ks, &
                                 scf_control, scf_section, &
                                 diis_step)
         END IF
         ! OT diagonlization
      CASE (ot_diag_method_nr)
         CALL do_ot_diag(scf_env, mos, matrix_ks, matrix_s, &
                         scf_control, scf_section, diis_step)
         ! Block Krylov diagonlization
      CASE (block_krylov_diag_method_nr)
         IF ((scf_env%krylov_space%eps_std_diag > 0.0_dp) .AND. &
             (scf_env%iter_count == 1 .OR. scf_env%iter_delta > scf_env%krylov_space%eps_std_diag)) THEN
            IF (scf_env%krylov_space%always_check_conv) THEN
               CALL do_block_krylov_diag(scf_env, mos, matrix_ks, &
                                         scf_control, scf_section, check_moconv_only=.TRUE.)
            END IF
            CALL do_general_diag(scf_env, mos, matrix_ks, &
                                 matrix_s, scf_control, scf_section, diis_step)
         ELSE
            CALL do_block_krylov_diag(scf_env, mos, matrix_ks, &
                                      scf_control, scf_section)
         END IF
         IF (scf_control%do_diag_sub) THEN
            skip_diag_sub = (scf_env%subspace_env%eps_diag_sub > 0.0_dp) .AND. &
                            (scf_env%iter_count == 1 .OR. scf_env%iter_delta > scf_env%subspace_env%eps_diag_sub)
            IF (.NOT. skip_diag_sub) THEN
               CALL do_scf_diag_subspace(qs_env, scf_env, scf_env%subspace_env, mos, rho, &
                                         ks_env, scf_section, scf_control)
            END IF
         END IF
         ! Block Davidson diagonlization
      CASE (block_davidson_diag_method_nr)
         CALL do_block_davidson_diag(qs_env, scf_env, mos, matrix_ks, matrix_s, scf_control, &
                                     scf_section, .FALSE.)
         ! OT without diagonlization. Needs special treatment for SCP runs
      CASE (ot_method_nr)
         CALL qs_scf_loop_do_ot(qs_env, scf_env, scf_control%smear, mos, rho, &
                                qs_env%mo_derivs, energy%total, &
                                matrix_s, energy_only=energy_only, has_unit_metric=has_unit_metric)
      END SELECT

      energy%kTS = 0.0_dp
      energy%efermi = 0.0_dp
      CALL get_qs_env(qs_env, mos=mos)
      DO ispin = 1, SIZE(mos)
         energy%kTS = energy%kTS+mos(ispin)%mo_set%kTS
         energy%efermi = energy%efermi+mos(ispin)%mo_set%mu
      ENDDO
      energy%efermi = energy%efermi/REAL(SIZE(mos), KIND=dp)

      CALL timestop(handle)

   END SUBROUTINE qs_scf_new_mos

! **************************************************************************************************
!> \brief Updates MOs and density matrix using diagonalization
!>        Kpoint code
!> \param qs_env ...
!> \param scf_env ...
!> \param scf_control ...
!> \param diis_step ...
! **************************************************************************************************
   SUBROUTINE qs_scf_new_mos_kp(qs_env, scf_env, scf_control, diis_step)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      LOGICAL                                            :: diis_step

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_new_mos_kp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      LOGICAL                                            :: has_unit_metric
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks, matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(mo_set_p_type), DIMENSION(:, :), POINTER      :: mos
      TYPE(qs_energy_type), POINTER                      :: energy

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, kpoints, matrix_ks, matrix_s)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, kpoints=kpoints)
      scf_env%iter_param = 0.0_dp

      IF (dft_control%roks) &
         CALL cp__b("qs_scf_loop_utils.F",286,"KP code: ROKS method not available: ")

      SELECT CASE (scf_env%method)
      CASE DEFAULT
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",290), &
                       "KP code: Unknown scf method: "// &
                       cp_to_string(scf_env%method))
      CASE (general_diag_method_nr)
         ! Diagonlization in non orthonormal case
         CALL get_qs_env(qs_env, matrix_ks_kp=matrix_ks, matrix_s_kp=matrix_s)
         CALL do_general_diag_kp(scf_env, matrix_ks, matrix_s, kpoints, &
                                 scf_control, diis_step)
      CASE (special_diag_method_nr)
         CALL get_qs_env(qs_env=qs_env, has_unit_metric=has_unit_metric)
         IF(.NOT.(has_unit_metric))CALL cp__a("qs_scf_loop_utils.F",300)
         ! Diagonlization in orthonormal case
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",302), &
                       "KP code: Scf method not available: "// &
                       cp_to_string(scf_env%method))
      CASE (ot_diag_method_nr, &
            block_krylov_diag_method_nr, &
            block_davidson_diag_method_nr, &
            ot_method_nr)
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",309), &
                       "KP code: Scf method not available: "// &
                       cp_to_string(scf_env%method))
      END SELECT

      CALL get_qs_env(qs_env=qs_env, energy=energy)
      energy%kTS = 0.0_dp
      energy%efermi = 0.0_dp
      mos => kpoints%kp_env(1)%kpoint_env%mos
      DO ispin = 1, SIZE(mos, 2)
         energy%kTS = energy%kTS+mos(1, ispin)%mo_set%kTS
         energy%efermi = energy%efermi+mos(1, ispin)%mo_set%mu
      ENDDO
      energy%efermi = energy%efermi/REAL(SIZE(mos, 2), KIND=dp)

      CALL timestop(handle)

   END SUBROUTINE qs_scf_new_mos_kp

! **************************************************************************************************
!> \brief the inner loop of scf, specific to using to the orbital transformation method
!>       basically, in goes the ks matrix out goes a new p matrix
!> \param qs_env ...
!> \param scf_env ...
!> \param smear ...
!> \param mos ...
!> \param rho ...
!> \param mo_derivs ...
!> \param total_energy ...
!> \param matrix_s ...
!> \param energy_only ...
!> \param has_unit_metric ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
!>      2013    moved from qs_scf [Florian Schiffmann]
! **************************************************************************************************
   SUBROUTINE qs_scf_loop_do_ot(qs_env, scf_env, smear, mos, rho, mo_derivs, total_energy, &
                                matrix_s, energy_only, has_unit_metric)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(smear_type), POINTER                          :: smear
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mo_derivs
      REAL(KIND=dp), INTENT(IN)                          :: total_energy
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      LOGICAL, INTENT(INOUT)                             :: energy_only
      LOGICAL, INTENT(IN)                                :: has_unit_metric

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_loop_do_ot', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_dbcsr_type), POINTER                       :: orthogonality_metric

      CALL timeset(routineN, handle)
      NULLIFY (rho_ao)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      IF (has_unit_metric) THEN
         NULLIFY (orthogonality_metric)
      ELSE
         orthogonality_metric => matrix_s(1)%matrix
      END IF

      ! in case of LSD the first spin qs_ot_env will drive the minimization
      ! in the case of a restricted calculation, it will make sure the spin orbitals are equal

      CALL ot_scf_mini(mos, mo_derivs, smear, orthogonality_metric, &
                       total_energy, energy_only, scf_env%iter_delta, &
                       scf_env%qs_ot_env, qs_env%input)

      DO ispin = 1, SIZE(mos)
         CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, &
                                smear=smear)
      ENDDO

      DO ispin = 1, SIZE(mos)
         CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                       rho_ao(ispin)%matrix, &
                                       use_dbcsr=.TRUE.)
      END DO

      scf_env%iter_method = scf_env%qs_ot_env(1)%OT_METHOD_FULL
      scf_env%iter_param = scf_env%qs_ot_env(1)%ds_min
      qs_env%broyden_adaptive_sigma = scf_env%qs_ot_env(1)%broyden_adaptive_sigma

      CALL timestop(handle)

   END SUBROUTINE qs_scf_loop_do_ot

! **************************************************************************************************
!> \brief Performs the requested density mixing if any needed
!> \param scf_env   Holds SCF environment information
!> \param rho       All data for the electron density
!> \param para_env  Parallel environment
!> \param diis_step Did we do a DIIS step?
! **************************************************************************************************
   SUBROUTINE qs_scf_density_mixing(scf_env, rho, para_env, diis_step)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL                                            :: diis_step

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_density_mixing', &
         routineP = moduleN//':'//routineN

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp

      NULLIFY (rho_ao_kp)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)

      SELECT CASE (scf_env%mixing_method)
      CASE (direct_mixing_nr)
         CALL scf_env_density_mixing(scf_env%p_mix_new, &
                                     scf_env%mixing_store, rho_ao_kp, para_env, scf_env%iter_delta, scf_env%iter_count, &
                                     diis=diis_step)
      CASE (gspace_mixing_nr, pulay_mixing_nr, broyden_mixing_nr, &
            broyden_mixing_new_nr, multisecant_mixing_nr)
         ! Compute the difference p_out-p_in
         CALL self_consistency_check(rho_ao_kp, scf_env%p_delta, para_env, scf_env%p_mix_new, &
                                     delta=scf_env%iter_delta)
      CASE (no_mixing_nr)
      CASE DEFAULT
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",437), &
                       "unknown scf mixing method: "// &
                       cp_to_string(scf_env%mixing_method))
      END SELECT

   END SUBROUTINE qs_scf_density_mixing

! **************************************************************************************************
!> \brief checks whether exit conditions for outer loop are satisfied
!> \param qs_env ...
!> \param scf_env ...
!> \param scf_control ...
!> \param should_stop ...
!> \param outer_loop_converged ...
!> \param exit_outer_loop ...
! **************************************************************************************************
   SUBROUTINE qs_scf_check_outer_exit(qs_env, scf_env, scf_control, should_stop, &
                                      outer_loop_converged, exit_outer_loop)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      LOGICAL                                            :: should_stop, outer_loop_converged, &
                                                            exit_outer_loop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_check_outer_exit', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: outer_loop_eps

      outer_loop_converged = .TRUE.
      IF (scf_control%outer_scf%have_scf) THEN
         ! We have an outer SCF loop...
         scf_env%outer_scf%iter_count = scf_env%outer_scf%iter_count+1
         outer_loop_converged = .FALSE.

         CALL outer_loop_gradient(qs_env, scf_env)
         outer_loop_eps = SQRT(SUM(scf_env%outer_scf%gradient(:, scf_env%outer_scf%iter_count)**2))/ &
                          SIZE(scf_env%outer_scf%gradient, 1)
         IF (outer_loop_eps < scf_control%outer_scf%eps_scf) outer_loop_converged = .TRUE.
      END IF

      exit_outer_loop = should_stop .OR. outer_loop_converged .OR. &
                        scf_env%outer_scf%iter_count > scf_control%outer_scf%max_scf

   END SUBROUTINE qs_scf_check_outer_exit

! **************************************************************************************************
!> \brief checks whether exit conditions for inner loop are satisfied
!> \param qs_env ...
!> \param scf_env ...
!> \param scf_control ...
!> \param should_stop ...
!> \param exit_inner_loop ...
!> \param inner_loop_converged ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_scf_check_inner_exit(qs_env, scf_env, scf_control, should_stop, &
                                      exit_inner_loop, inner_loop_converged, output_unit)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      LOGICAL                                            :: should_stop, exit_inner_loop, &
                                                            inner_loop_converged
      INTEGER                                            :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_check_inner_exit', &
         routineP = moduleN//':'//routineN

      inner_loop_converged = .FALSE.
      exit_inner_loop = .FALSE.

      CALL external_control(should_stop, "SCF", target_time=qs_env%target_time, &
                            start_time=qs_env%start_time)
      IF (scf_env%iter_delta < scf_control%eps_scf) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,I5,A/)") &
               "*** SCF run converged in ", scf_env%iter_count, " steps ***"
         END IF
         inner_loop_converged = .TRUE.
         exit_inner_loop = .TRUE.
      ELSE IF (should_stop .OR. scf_env%iter_count >= scf_control%max_scf) THEN
         inner_loop_converged = .FALSE.
         exit_inner_loop = .TRUE.
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,I5,A/)") &
               "Leaving inner SCF loop after reaching ", scf_env%iter_count, " steps."
         END IF
      END IF

   END SUBROUTINE qs_scf_check_inner_exit

! **************************************************************************************************
!> \brief undoing density mixing. Important upon convergence
!> \param scf_env ...
!> \param rho ...
!> \param dft_control ...
!> \param para_env ...
!> \param diis_step ...
! **************************************************************************************************
   SUBROUTINE qs_scf_undo_mixing(scf_env, rho, dft_control, para_env, diis_step)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL                                            :: diis_step

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_undo_mixing', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, ispin, nc
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp

      NULLIFY (rho_ao_kp)

      IF (scf_env%mixing_method > 0) THEN
         CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)
         nc = SIZE(scf_env%p_mix_new, 2)
         SELECT CASE (scf_env%mixing_method)
         CASE (direct_mixing_nr)
            CALL scf_env_density_mixing(scf_env%p_mix_new, scf_env%mixing_store, &
                                        rho_ao_kp, para_env, scf_env%iter_delta, &
                                        scf_env%iter_count, diis=diis_step, &
                                        invert=.TRUE.)
            DO ic = 1, nc
               DO ispin = 1, dft_control%nspins
                  CALL cp_dbcsr_copy(rho_ao_kp(ispin, ic)%matrix, scf_env%p_mix_new(ispin, ic)%matrix, &
                                     name=TRIM(cp_dbcsr_name(rho_ao_kp(ispin, ic)%matrix)))
               END DO
            END DO
         CASE (gspace_mixing_nr, pulay_mixing_nr, broyden_mixing_nr, &
               broyden_mixing_new_nr, multisecant_mixing_nr)
            DO ic = 1, nc
               DO ispin = 1, dft_control%nspins
                  CALL cp_dbcsr_copy(rho_ao_kp(ispin, ic)%matrix, scf_env%p_mix_new(ispin, ic)%matrix, &
                                     name=TRIM(cp_dbcsr_name(rho_ao_kp(ispin, ic)%matrix)))
               END DO
            END DO
         END SELECT
      ENDIF
   END SUBROUTINE qs_scf_undo_mixing

! **************************************************************************************************
!> \brief Performes the updates rho (takes care of mixing as well)
!> \param rho ...
!> \param qs_env ...
!> \param scf_env ...
!> \param ks_env ...
!> \param mix_rho ...
! **************************************************************************************************
   SUBROUTINE qs_scf_rho_update(rho, qs_env, scf_env, ks_env, mix_rho)
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      LOGICAL, INTENT(IN)                                :: mix_rho

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_rho_update', &
         routineP = moduleN//':'//routineN

      TYPE(cp_para_env_type), POINTER                    :: para_env

      NULLIFY (para_env)
      CALL get_qs_env(qs_env, para_env=para_env)
      ! ** update qs_env%rho
      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      ! ** Density mixing through density matrix or on the reciprocal space grid (exclusive)
      IF (mix_rho) THEN
         CALL gspace_mixing(qs_env, scf_env%mixing_method, scf_env%mixing_store, rho, &
                            para_env, scf_env%iter_count)

      END IF
      CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)

   END SUBROUTINE qs_scf_rho_update

! **************************************************************************************************
!> \brief Performes the necessary steps before leaving innner scf loop
!> \param scf_env ...
!> \param qs_env ...
!> \param diis_step ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_scf_inner_finalize(scf_env, qs_env, diis_step, output_unit)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL                                            :: diis_step
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_inner_finalize', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: do_kpoints
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (energy, rho, dft_control, ks_env)

      CALL get_qs_env(qs_env=qs_env, energy=energy, ks_env=ks_env, &
                      rho=rho, dft_control=dft_control, para_env=para_env, &
                      do_kpoints=do_kpoints)

      CALL cleanup_scf_loop(scf_env)

      ! now, print out energies and charges corresponding to the obtained wfn
      ! (this actually is not 100% consistent at this point)!
      CALL qs_scf_print_summary(output_unit, qs_env)

      CALL qs_scf_undo_mixing(scf_env, rho, dft_control, para_env, diis_step)

      !   *** update rspace rho since the mo changed
      !   *** this might not always be needed (i.e. no post calculation / no forces )
      !   *** but guarantees that rho and wfn are consistent at this point
      CALL qs_scf_rho_update(rho, qs_env, scf_env, ks_env, mix_rho=.FALSE.)

   END SUBROUTINE qs_scf_inner_finalize

! **************************************************************************************************
!> \brief perform cleanup operations at the end of an scf loop
!> \param scf_env ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE cleanup_scf_loop(scf_env)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cleanup_scf_loop', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_loop_utils.F",672)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_loop_utils.F",673)
! *** method dependent cleanup
      SELECT CASE (scf_env%method)
      CASE (ot_method_nr)
         DO ispin = 1, SIZE(scf_env%qs_ot_env)
            CALL ot_scf_destroy(scf_env%qs_ot_env(ispin))
         ENDDO
         DEALLOCATE (scf_env%qs_ot_env)
      CASE (ot_diag_method_nr)
         !
      CASE (general_diag_method_nr)
         !
      CASE (special_diag_method_nr)
         !
      CASE (block_krylov_diag_method_nr, block_davidson_diag_method_nr)
         !
      CASE (filter_matrix_diag_method_nr)
         !
      CASE DEFAULT
         CALL cp_abort(cp__l("qs_scf_loop_utils.F",692), &
                       "unknown scf method method:"// &
                       cp_to_string(scf_env%method))
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cleanup_scf_loop

END MODULE qs_scf_loop_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_distribution_methods

   USE cell_types,                      ONLY: cell_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_get_info, cp_dbcsr_nblkcols_total, cp_dbcsr_p_type, cp_dbcsr_type, &
        dbcsr_distribution_mp, dbcsr_distribution_obj, dbcsr_mp_get_process, dbcsr_mp_npcols, &
        dbcsr_mp_nprows, dbcsr_mp_obj
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_fb_atomic_halo_types,         ONLY: &
        fb_atomic_halo_build_halo_atoms, fb_atomic_halo_cost, fb_atomic_halo_create, &
        fb_atomic_halo_init, fb_atomic_halo_nullify, fb_atomic_halo_obj, fb_atomic_halo_release, &
        fb_atomic_halo_set, fb_build_pair_radii
   USE qs_fb_env_types,                 ONLY: fb_env_get,&
                                              fb_env_obj,&
                                              fb_env_set
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: fb_distribution_build

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_distribution_methods'

! **************************************************************************************************
!> \brief derived type containing cost data used for process distribution
!> \param id               : global atomic index
!> \param cost             : computational cost for the atomic matrix associated
!>                           to this atom
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_distribution_element
      INTEGER :: id
      REAL(KIND=dp) :: cost
   END TYPE fb_distribution_element

! **************************************************************************************************
!> \brief derived type containing the list of atoms currently allocated to a
!>        processor
!> \param list             : list of atoms and their associated costs
!> \param cost             : total cost of the list
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_distribution_list
      TYPE(fb_distribution_element), DIMENSION(:), POINTER :: list => NULL()
      INTEGER :: nelements
      REAL(KIND=dp) :: cost
   END TYPE fb_distribution_list

! **************************************************************************************************
!> \brief In filter matrix algorithm, each atomic matrix contributes to a
!>        column in the filter matrix, which is stored in DBCSR format.
!>        When distributing the atoms (and hence the atomic matrics) to the
!>        processors, we want the processors to have atoms that would be
!>        correspond to the block columns in the DBCSR format local to them.
!>        This derived type stores this information. For each atom, it
!>        corresponds to a DBCSR block column, and the list of processors
!>        in the 2D processor grid responsible for this column will be the
!>        preferred processors for this atom.
!> \param list             : list of preferred processors for an atom
!>                           note that here the processors are indexed from
!>                           1, i.e. = MPI_RANK+1
!> \param nprocs           : number of processors in the list
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_preferred_procs_list
      INTEGER, DIMENSION(:), POINTER :: list => NULL()
      INTEGER :: nprocs
   END TYPE fb_preferred_procs_list

! Parameters related to automatic resizing of the hash_table:
! Resize by EXPAND_FACTOR if total no. slots / no. of filled slots < ENLARGE_RATIO
   INTEGER, PARAMETER, PRIVATE :: ENLARGE_RATIO = 1
   INTEGER, PARAMETER, PRIVATE :: REDUCE_RATIO = 3
   INTEGER, PARAMETER, PRIVATE :: EXPAND_FACTOR = 2
   INTEGER, PARAMETER, PRIVATE :: SHRINK_FACTOR = 2

   INTERFACE fb_distribution_remove
      MODULE PROCEDURE fb_distribution_remove_ind, &
         fb_distribution_remove_el
   END INTERFACE fb_distribution_remove

   INTERFACE fb_distribution_move
      MODULE PROCEDURE fb_distribution_move_ind, &
         fb_distribution_move_el
   END INTERFACE fb_distribution_move

CONTAINS

! **************************************************************************************************
!> \brief Build local atoms associated to filter matrix algorithm for each
!>        MPI process, trying to balance the load for calculating the
!>        filter matrix
!> \param fb_env : the filter matrix environment
!> \param qs_env : quickstep environment
!> \param scf_section : SCF input section
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_build(fb_env, qs_env, scf_section)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_build', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i_common_set, iatom, ii, ipe, lb, lowest_cost_ind, my_pe, n_common_sets, &
         natoms, nhalo_atoms, nkinds, nprocs, owner_id_in_halo, pref_pe, ub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: common_set_ids, local_atoms_all, &
                                                            local_atoms_sizes, local_atoms_starts, &
                                                            pe, pos_in_preferred_list
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms, local_atoms
      LOGICAL                                            :: acceptable_move, move_happened
      REAL(KIND=dp)                                      :: average_cost
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cost_per_atom, cost_per_proc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pair_radii
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rcut
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_ks
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fb_atomic_halo_obj)                           :: atomic_halo
      TYPE(fb_distribution_element)                      :: element
      TYPE(fb_distribution_list), ALLOCATABLE, &
         DIMENSION(:)                                    :: dist
      TYPE(fb_preferred_procs_list), ALLOCATABLE, &
         DIMENSION(:)                                    :: preferred_procs_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)

      NULLIFY (mat_ks, rcut, cell, para_env, particle_set, qs_kind_set, &
               halo_atoms, local_atoms)
      CALL fb_atomic_halo_nullify(atomic_halo)

      ! obtain relevant data from fb_env, qs_env
      CALL fb_env_get(fb_env=fb_env, &
                      rcut=rcut)
      CALL get_qs_env(qs_env=qs_env, &
                      natom=natoms, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      nkind=nkinds, &
                      cell=cell, &
                      para_env=para_env, &
                      matrix_ks=mat_ks)
      nprocs = para_env%num_pe
      my_pe = para_env%mepos+1 ! counting from 1

      ! for each global atom, build atomic halo and get the associated cost
      ALLOCATE (pair_radii(nkinds, nkinds))
      CALL fb_build_pair_radii(rcut, nkinds, pair_radii)
      CALL fb_atomic_halo_create(atomic_halo)
      ALLOCATE (cost_per_atom(natoms))
      DO iatom = 1, natoms
         CALL fb_atomic_halo_init(atomic_halo)
         CALL fb_atomic_halo_build_halo_atoms(iatom, &
                                              particle_set, &
                                              cell, &
                                              pair_radii, &
                                              halo_atoms, &
                                              nhalo_atoms, &
                                              owner_id_in_halo)
         CALL fb_atomic_halo_set(atomic_halo=atomic_halo, &
                                 owner_atom=iatom, &
                                 natoms=nhalo_atoms, &
                                 halo_atoms=halo_atoms)
         NULLIFY (halo_atoms)
         cost_per_atom(iatom) = fb_atomic_halo_cost(atomic_halo, particle_set, qs_kind_set)
      END DO
      DEALLOCATE (pair_radii)
      CALL fb_atomic_halo_release(atomic_halo)

      ! build the preferred_procs_set according to DBCSR mat H
      ALLOCATE (preferred_procs_set(natoms))
      ALLOCATE (common_set_ids(natoms))
      CALL fb_build_preferred_procs(mat_ks(1)%matrix, &
                                    natoms, &
                                    preferred_procs_set, &
                                    common_set_ids, &
                                    n_common_sets)

      ! for each atomic halo, construct distribution_element, and assign
      ! the element to a processors using preferred_procs_set in a
      ! round-robin manner
      ALLOCATE (dist(nprocs))
      DO ipe = 1, nprocs
         CALL fb_distribution_init(dist=dist(ipe))
      END DO
      ALLOCATE (pos_in_preferred_list(n_common_sets))
      pos_in_preferred_list(:) = 0
      DO iatom = 1, natoms
         element%id = iatom
         element%cost = cost_per_atom(iatom)
         i_common_set = common_set_ids(iatom)
         pos_in_preferred_list(i_common_set) = &
            MOD(pos_in_preferred_list(i_common_set), &
                preferred_procs_set(iatom)%nprocs)+1
         ipe = preferred_procs_set(iatom)%list(pos_in_preferred_list(i_common_set))
         CALL fb_distribution_add(dist(ipe), element)
      END DO

      DEALLOCATE (pos_in_preferred_list)
      DEALLOCATE (common_set_ids)
      DEALLOCATE (cost_per_atom)

      ! sort processors according to the overall cost of their assigned
      ! corresponding distribution
      ALLOCATE (cost_per_proc(nprocs))
      DO ipe = 1, nprocs
         cost_per_proc(ipe) = dist(ipe)%cost
      END DO
      ALLOCATE (pe(nprocs))
      CALL sort(cost_per_proc, nprocs, pe)
      ! now that cost_per_proc is sorted, ipe's no longer give mpi
      ! ranks, the correct one to use should be pe(ipe)

      ! work out the ideal average cost per proc if work load is evenly
      ! distributed
      average_cost = SUM(cost_per_proc)/REAL(nprocs, dp)

      DEALLOCATE (cost_per_proc)

      ! loop over the processors, starting with the highest cost, move
      ! atoms one by one:
      !   1. FIRST to the next processor in the preferred list that has
      !      cost below average. IF no such proc is found, THEN
      !   2. to the next procesor in the overall list that has cost
      !      below average.
      ! repeat until the cost on this processor is less than or equal
      ! to the average cost
      lowest_cost_ind = 1
      DO ipe = nprocs, 1, -1
         redistribute: DO WHILE (dist(pe(ipe))%cost .GT. average_cost)
            iatom = dist(pe(ipe))%list(lowest_cost_ind)%id
            move_happened = .FALSE.
            ! first try to move to a preferred process
            preferred: DO ii = 1, preferred_procs_set(iatom)%nprocs
               pref_pe = preferred_procs_set(iatom)%list(ii)
               acceptable_move = &
                  fb_distribution_acceptable_move(dist(pe(ipe)), &
                                                  dist(pe(ipe))%list(lowest_cost_ind), &
                                                  dist(pref_pe), &
                                                  average_cost)
               IF ((pref_pe .NE. pe(ipe)) .AND. acceptable_move) THEN
                  CALL fb_distribution_move(dist(pe(ipe)), &
                                            lowest_cost_ind, &
                                            dist(pref_pe))
                  move_happened = .TRUE.
                  EXIT preferred
               END IF
            END DO preferred
            ! if no preferred process is avaliable, move to a proc in
            ! the sorted list that has cost less than average.  remember
            ! that some of the proc may have already taken redistributed
            ! atoms, and thus may beome unavaliable (full)
            IF (.NOT. move_happened) THEN
               ! searching from the proc with the least initial cost
               next_in_line: DO ii = 1, nprocs
                  acceptable_move = &
                     fb_distribution_acceptable_move(dist(pe(ipe)), &
                                                     dist(pe(ipe))%list(lowest_cost_ind), &
                                                     dist(pe(ii)), &
                                                     average_cost)
                  IF ((pe(ii) .NE. pe(ipe)) .AND. acceptable_move) THEN
                     CALL fb_distribution_move(dist(pe(ipe)), &
                                               lowest_cost_ind, &
                                               dist(pe(ii)))
                     move_happened = .TRUE.
                     EXIT next_in_line
                  END IF
               END DO next_in_line
            END IF
            ! if the atom cannot be moved, then this means it is too
            ! costly for all other processes to accept. When this
            ! happens we must stop the redistribution process for this
            ! processor---as all other of its atoms will be even more
            ! costly
            IF (.NOT. move_happened) THEN
               EXIT redistribute
            END IF
         END DO redistribute ! while
      END DO ! ipe

      DEALLOCATE (pe)
      DO ii = 1, SIZE(preferred_procs_set)
         CALL fb_preferred_procs_list_release(preferred_procs_set(ii))
      END DO
      DEALLOCATE (preferred_procs_set)

      ! generate local atoms from dist
      ALLOCATE (local_atoms_all(natoms))
      ALLOCATE (local_atoms_starts(nprocs))
      ALLOCATE (local_atoms_sizes(nprocs))
      CALL fb_distribution_to_local_atoms(dist, &
                                          local_atoms_all, &
                                          local_atoms_starts, &
                                          local_atoms_sizes)
      ALLOCATE (local_atoms(local_atoms_sizes(my_pe)))
      lb = local_atoms_starts(my_pe)
      ub = local_atoms_starts(my_pe)+local_atoms_sizes(my_pe)-1
      local_atoms(1:local_atoms_sizes(my_pe)) = local_atoms_all(lb:ub)
      CALL fb_env_set(fb_env=fb_env, &
                      local_atoms=local_atoms, &
                      nlocal_atoms=local_atoms_sizes(my_pe))

      ! write out info
      CALL fb_distribution_write_info(dist, scf_section)

      DEALLOCATE (local_atoms_all)
      DEALLOCATE (local_atoms_starts)
      DEALLOCATE (local_atoms_sizes)
      DO ipe = 1, SIZE(dist)
         CALL fb_distribution_release(dist(ipe))
      END DO
      DEALLOCATE (dist)

      CALL timestop(handle)

   END SUBROUTINE fb_distribution_build

! **************************************************************************************************
!> \brief Checks if moving an element from one distribution to another is
!>        allowed in mind of load balancing.
!> \param dist_from : the source distribution
!> \param element   : the element in source distribution considered for the
!>                    move
!> \param dist_to   : the destination distribution
!> \param threshold ...
!> \retval res : TRUE or FALSE
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_distribution_acceptable_move(dist_from, &
                                                 element, &
                                                 dist_to, &
                                                 threshold) &
      RESULT(acceptable)
      TYPE(fb_distribution_list), INTENT(IN)             :: dist_from
      TYPE(fb_distribution_element), INTENT(IN)          :: element
      TYPE(fb_distribution_list), INTENT(IN)             :: dist_to
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      LOGICAL                                            :: acceptable

      acceptable = (dist_to%cost+element%cost .LT. dist_from%cost) .AND. &
                   (dist_to%cost .LT. threshold)
   END FUNCTION fb_distribution_acceptable_move

! **************************************************************************************************
!> \brief Write out information on the load distribution on processors
!> \param dist_set    : set of distributions for the processors
!> \param scf_section : SCF input section
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_write_info(dist_set, scf_section)
      TYPE(fb_distribution_list), DIMENSION(:), &
         INTENT(IN)                                      :: dist_set
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_write_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, max_natoms, min_natoms, natoms, &
                                                            nprocs, unit_nr
      REAL(KIND=dp)                                      :: ave_cost, ave_natoms, max_cost, &
                                                            min_cost, total_cost
      TYPE(cp_logger_type), POINTER                      :: logger

      nprocs = SIZE(dist_set)
      natoms = 0
      total_cost = 0.0_dp
      DO ii = 1, nprocs
         natoms = natoms+dist_set(ii)%nelements
         total_cost = total_cost+dist_set(ii)%cost
      END DO
      ave_natoms = REAL(natoms, dp)/REAL(nprocs, dp)
      ave_cost = total_cost/REAL(nprocs, dp)
      max_natoms = 0
      max_cost = 0._dp
      DO ii = 1, nprocs
         max_natoms = MAX(max_natoms, dist_set(ii)%nelements)
         max_cost = MAX(max_cost, dist_set(ii)%cost)
      END DO
      min_natoms = natoms
      min_cost = total_cost
      DO ii = 1, nprocs
         min_natoms = MIN(min_natoms, dist_set(ii)%nelements)
         min_cost = MIN(min_cost, dist_set(ii)%cost)
      END DO

      logger => cp_get_default_logger()
      unit_nr = cp_print_key_unit_nr(logger, scf_section, &
                                     "PRINT%FILTER_MATRIX", &
                                     extension="")

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(/,A,I6,A)") &
            " FILTER_MAT_DIAG| Load distribution across ", nprocs, " processors:"
         WRITE (UNIT=unit_nr, &
                FMT="(A,T40,A,T55,A,T70,A,T85,A)") &
            " FILTER_MAT_DIAG| ", "Total", "Average", "Max", "Min"
         WRITE (UNIT=unit_nr, &
                FMT="(A,T40,I12,T55,F12.1,T70,I12,T85,I10)") &
            " FILTER_MAT_DIAG|   Atomic Matrices", &
            natoms, ave_natoms, max_natoms, min_natoms
         WRITE (UNIT=unit_nr, &
                FMT="(A,T40,D12.7,T55,D12.7,T70,D12.7,T85,D12.7)") &
            " FILTER_MAT_DIAG|   Cost*", &
            total_cost, ave_cost, max_cost, min_cost
         WRITE (UNIT=unit_nr, FMT="(A)") &
            " FILTER_MAT_DIAG| (* cost is calculated as sum of cube of atomic matrix sizes)"
      END IF
      CALL cp_print_key_finished_output(unit_nr, logger, scf_section, &
                                        "PRINT%FILTER_MATRIX")
   END SUBROUTINE fb_distribution_write_info

! **************************************************************************************************
!> \brief Build the preferred list of processors for atoms
!> \param dbcsr_mat   : the reference DBCSR matrix, from which the local block
!>                      cols and the processor maps are obtained
!> \param natoms      : total number of atoms globally
!> \param preferred_procs_set : set of preferred procs list for each atom
!> \param common_set_ids : atoms (block cols) local to the same processor grid
!>                         col will have the same preferred list. This list
!>                         maps each atom to their corresponding group
!> \param n_common_sets  : number of unique preferred lists (groups)
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_build_preferred_procs(dbcsr_mat, &
                                       natoms, &
                                       preferred_procs_set, &
                                       common_set_ids, &
                                       n_common_sets)
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_mat
      INTEGER, INTENT(IN)                                :: natoms
      TYPE(fb_preferred_procs_list), DIMENSION(:), &
         INTENT(INOUT)                                   :: preferred_procs_set
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: common_set_ids
      INTEGER, INTENT(OUT)                               :: n_common_sets

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_build_preferred_procs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icol, nprows, pcol, prow
      INTEGER, DIMENSION(:), POINTER                     :: col_dist
      LOGICAL                                            :: check_ok
      TYPE(dbcsr_distribution_obj)                       :: dbcsr_dist
      TYPE(dbcsr_mp_obj)                                 :: dbcsr_mp

      check_ok = natoms .LE. cp_dbcsr_nblkcols_total(dbcsr_mat)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",467)
      check_ok = SIZE(preferred_procs_set) .GE. natoms
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",469)
      check_ok = SIZE(common_set_ids) .GE. natoms
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",471)

      CALL cp_dbcsr_get_info(dbcsr_mat, distribution=dbcsr_dist, proc_col_dist=col_dist)
      dbcsr_mp = dbcsr_distribution_mp(dbcsr_dist)
      nprows = dbcsr_mp_nprows(dbcsr_mp)

      DO icol = 1, natoms
         IF (ASSOCIATED(preferred_procs_set(icol)%list)) THEN
            DEALLOCATE (preferred_procs_set(icol)%list)
         END IF
         ALLOCATE (preferred_procs_set(icol)%list(nprows))
         pcol = col_dist(icol)
         ! dbcsr prow and pcol counts from 0
         DO prow = 0, nprows-1
            ! here, we count processes from 1, so +1 from mpirank
            preferred_procs_set(icol)%list(prow+1) = &
               dbcsr_mp_get_process(dbcsr_mp, prow, pcol)+1
         END DO
         preferred_procs_set(icol)%nprocs = nprows
      END DO

      common_set_ids(:) = 0
      common_set_ids(1:natoms) = col_dist(1:natoms)+1
      n_common_sets = dbcsr_mp_npcols(dbcsr_mp)

   END SUBROUTINE fb_build_preferred_procs

! **************************************************************************************************
!> \brief Release a preferred_procs_list
!> \param preferred_procs_list  : the preferred procs list in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_preferred_procs_list_release(preferred_procs_list)
      TYPE(fb_preferred_procs_list), INTENT(INOUT)       :: preferred_procs_list

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_preferred_procs_list_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(preferred_procs_list%list)) THEN
         DEALLOCATE (preferred_procs_list%list)
      END IF
   END SUBROUTINE fb_preferred_procs_list_release

! **************************************************************************************************
!> \brief Convert distribution data to 1D array containing information of
!>        which atoms are distributed to which processor
!> \param dist_set    : set of distributions for the processors
!> \param local_atoms : continuous array of atoms arranged in order
!>                      corresponding their allocated processors
!> \param local_atoms_starts : starting position in local_atoms array for
!>                             each processor
!> \param local_atoms_sizes  : number of atoms local to each processor
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_to_local_atoms(dist_set, &
                                             local_atoms, &
                                             local_atoms_starts, &
                                             local_atoms_sizes)
      TYPE(fb_distribution_list), DIMENSION(:), &
         INTENT(IN)                                      :: dist_set
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: local_atoms, local_atoms_starts, &
                                                            local_atoms_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_to_local_atoms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ipe, n_procs, pos
      LOGICAL                                            :: check_ok

      n_procs = SIZE(dist_set)

      check_ok = SIZE(local_atoms_starts) .GE. n_procs
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",543)
      check_ok = SIZE(local_atoms_sizes) .GE. n_procs
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",545)

      local_atoms(:) = 0
      local_atoms_starts(:) = 0
      local_atoms_sizes(:) = 0

      pos = 1
      DO ipe = 1, n_procs
         local_atoms_starts(ipe) = pos
         DO iatom = 1, dist_set(ipe)%nelements
            local_atoms(pos) = dist_set(ipe)%list(iatom)%id
            pos = pos+1
            local_atoms_sizes(ipe) = local_atoms_sizes(ipe)+1
         END DO
      END DO
   END SUBROUTINE fb_distribution_to_local_atoms

! **************************************************************************************************
!> \brief Initialise a distribution
!> \param dist        : the distribution in question
!> \param nmax        : [OPTIONAL] size of the list array to be allocated
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_init(dist, nmax)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist
      INTEGER, INTENT(IN), OPTIONAL                      :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, my_nmax

      my_nmax = 0
      IF (PRESENT(nmax)) my_nmax = nmax
      IF (ASSOCIATED(dist%list)) THEN
         DEALLOCATE (dist%list)
      END IF
      NULLIFY (dist%list)
      IF (my_nmax .GT. 0) THEN
         ALLOCATE (dist%list(my_nmax))
         DO ii = 1, SIZE(dist%list)
            dist%list(ii)%id = 0
            dist%list(ii)%cost = 0.0_dp
         END DO
      END IF
      dist%nelements = 0
      dist%cost = 0.0_dp
   END SUBROUTINE fb_distribution_init

! **************************************************************************************************
!> \brief Resize the list array in a distribution
!> \param dist        : The distribution in question
!> \param nmax        : new size of the list array
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_resize(dist, nmax)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist
      INTEGER, INTENT(IN)                                :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_resize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, my_nmax
      TYPE(fb_distribution_element), DIMENSION(:), &
         POINTER                                         :: new_list

      IF (.NOT. ASSOCIATED(dist%list)) THEN
         my_nmax = MAX(nmax, 1)
         ALLOCATE (dist%list(my_nmax))
      ELSE
         my_nmax = MAX(nmax, dist%nelements)
         ALLOCATE (new_list(my_nmax))
         DO ii = 1, SIZE(new_list)
            new_list(ii)%id = 0
            new_list(ii)%cost = 0.0_dp
         END DO
         DO ii = 1, dist%nelements
            new_list(ii) = dist%list(ii)
         END DO
         DEALLOCATE (dist%list)
         dist%list => new_list
      END IF
   END SUBROUTINE fb_distribution_resize

! **************************************************************************************************
!> \brief Add an atom (element) to a distribution
!> \param dist        : the distribution in question
!> \param element     : the element to be added
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_add(dist, element)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist
      TYPE(fb_distribution_element), INTENT(IN)          :: element

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, new_nelements, pos

      new_nelements = dist%nelements+1

      ! resize list if necessary
      IF (.NOT. ASSOCIATED(dist%list)) THEN
         CALL fb_distribution_resize(dist, new_nelements)
      ELSE IF (new_nelements*ENLARGE_RATIO .GT. SIZE(dist%list)) THEN
         CALL fb_distribution_resize(dist, SIZE(dist%list)*EXPAND_FACTOR)
      END IF
      ! assuming the list of elements is always sorted with respect to cost
      ! slot the new element into the appropriate spot
      IF (new_nelements == 1) THEN
         dist%list(1) = element
      ELSE
         pos = fb_distribution_find_slot(dist, element)
         DO ii = dist%nelements, pos, -1
            dist%list(ii+1) = dist%list(ii)
         END DO
         dist%list(pos) = element
      END IF
      dist%nelements = new_nelements
      dist%cost = dist%cost+element%cost
   END SUBROUTINE fb_distribution_add

! **************************************************************************************************
!> \brief Find the correct slot in the list array to add a new element, so that
!>        the list will always be ordered with respect to cost
!> \param dist        : the distribution in question
!> \param element     : element to be added
!> \retval pos        : the correct position to add the new element
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE FUNCTION fb_distribution_find_slot(dist, element) RESULT(pos)
      TYPE(fb_distribution_list), INTENT(IN)             :: dist
      TYPE(fb_distribution_element), INTENT(IN)          :: element
      INTEGER                                            :: pos

      INTEGER                                            :: lower, middle, N, upper

      N = dist%nelements
      IF (element%cost .LT. dist%list(1)%cost) THEN
         pos = 1
         RETURN
      END IF
      IF (element%cost .GE. dist%list(N)%cost) THEN
         pos = N+1
         RETURN
      END IF
      lower = 1
      upper = N
      DO WHILE ((upper-lower) .GT. 1)
         middle = (lower+upper)/2
         IF (element%cost .LT. dist%list(middle)%cost) THEN
            upper = middle
         ELSE
            lower = middle
         END IF
      END DO
      pos = upper
   END FUNCTION fb_distribution_find_slot

! **************************************************************************************************
!> \brief Remove the pos-th element from a distribution
!> \param dist        : the distribution in question
!> \param pos         : index of the element in the list array
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_remove_ind(dist, pos)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist
      INTEGER, INTENT(IN)                                :: pos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_remove_ind', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      LOGICAL                                            :: check_ok

      check_ok = pos .GT. 0
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",721)
      IF (pos .LE. dist%nelements) THEN
         dist%cost = dist%cost-dist%list(pos)%cost
         DO ii = pos, dist%nelements-1
            dist%list(ii) = dist%list(ii+1)
         END DO
         dist%list(dist%nelements)%id = 0
         dist%list(dist%nelements)%cost = 0.0_dp
         dist%nelements = dist%nelements-1
         ! auto resize if required
         IF (dist%nelements*REDUCE_RATIO .LT. SIZE(dist%list)) THEN
            CALL fb_distribution_resize(dist, dist%nelements/SHRINK_FACTOR)
         END IF
      END IF
   END SUBROUTINE fb_distribution_remove_ind

! **************************************************************************************************
!> \brief Remove a given element from a distribution
!> \param dist        : the distribution in question
!> \param element     : the element in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_remove_el(dist, element)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist
      TYPE(fb_distribution_element), INTENT(IN)          :: element

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_remove_el', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii, pos

      pos = dist%nelements+1
      DO ii = 1, dist%nelements
         IF (element%id == dist%list(ii)%id) THEN
            pos = ii
            EXIT
         END IF
      END DO
      CALL fb_distribution_remove_ind(dist, pos)
   END SUBROUTINE fb_distribution_remove_el

! **************************************************************************************************
!> \brief Move the pos-th element from a distribution to another
!> \param dist_from   : the source distribution
!> \param pos         : index of the element in the source distribution
!> \param dist_to     : the destination distribution
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_move_ind(dist_from, pos, dist_to)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist_from
      INTEGER, INTENT(IN)                                :: pos
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist_to

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_move_ind', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok
      TYPE(fb_distribution_element)                      :: element

      check_ok = ASSOCIATED(dist_from%list)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",781)
      check_ok = pos .LE. dist_from%nelements
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",783)
      element = dist_from%list(pos)
      CALL fb_distribution_add(dist_to, element)
      CALL fb_distribution_remove(dist_from, pos)
   END SUBROUTINE fb_distribution_move_ind

! **************************************************************************************************
!> \brief Move a given element from a distribution to another
!> \param dist_from   : the source distribution
!> \param element     : the element in question
!> \param dist_to     : the destination distribution
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_move_el(dist_from, element, dist_to)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist_from
      TYPE(fb_distribution_element), INTENT(IN)          :: element
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist_to

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_move_el', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check_ok

      check_ok = ASSOCIATED(dist_from%list)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_distribution_methods.F",807)
      CALL fb_distribution_add(dist_to, element)
      CALL fb_distribution_remove(dist_from, element)
   END SUBROUTINE fb_distribution_move_el

! **************************************************************************************************
!> \brief Release a distribution
!> \param dist  : the distribution in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_distribution_release(dist)
      TYPE(fb_distribution_list), INTENT(INOUT)          :: dist

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_distribution_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dist%list)) THEN
         DEALLOCATE (dist%list)
      END IF
   END SUBROUTINE fb_distribution_release

END MODULE qs_fb_distribution_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief sets the environment for optimization of exponents and contraction
!>        coefficients of the lri auxiliary
!>        lri : local resolution of the identity
!> \par History
!>      created Dorothea Golze [12.2014]
!> \authors Dorothea Golze
! **************************************************************************************************
MODULE lri_optimize_ri_basis_types

   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_optimize_ri_basis_types'
   PUBLIC :: lri_opt_type
   PUBLIC :: create_lri_opt, deallocate_lri_opt, get_original_gcc, &
             orthonormalize_gcc

! **************************************************************************************************

   TYPE lri_gcc_p_type
      ! gcc without normalization factor
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: gcc_orig
   END TYPE lri_gcc_p_type

   TYPE lri_subset_type
      ! amount of l quantum numbers per set
      INTEGER                                            :: nl
      ! number of contraction per l quantum number for a given set
      INTEGER, DIMENSION(:), POINTER                     :: ncont_l
   END TYPE lri_subset_type

   TYPE lri_opt_type
      LOGICAL                                            :: opt_exps
      LOGICAL                                            :: opt_coeffs
      LOGICAL                                            :: use_condition_number
      LOGICAL                                            :: use_geometric_seq
      LOGICAL                                            :: use_constraints
      INTEGER                                            :: nexp
      INTEGER                                            :: ncoeff
      REAL(KIND=dp)                                      :: cond_weight
      REAL(KIND=dp)                                      :: scale_exp
      REAL(KIND=dp)                                      :: fermi_exp
      REAL(KIND=dp)                                      :: rho_diff
      ! array holding the variables that are optimized
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x
      ! intial exponents
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet_init
      ! holds the original contraction coeff of the lri basis
      TYPE(lri_gcc_p_type), DIMENSION(:), POINTER        :: ri_gcc_orig
      TYPE(lri_subset_type), DIMENSION(:), POINTER      :: subset
   END TYPE lri_opt_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief creates lri_opt
!> \param lri_opt optimization environment
! **************************************************************************************************
   SUBROUTINE create_lri_opt(lri_opt)

      TYPE(lri_opt_type), POINTER                        :: lri_opt

      CHARACTER(len=*), PARAMETER :: routineN = 'create_lri_opt', routineP = moduleN//':'//routineN

      ALLOCATE (lri_opt)

      NULLIFY (lri_opt%ri_gcc_orig)
      NULLIFY (lri_opt%subset)
      NULLIFY (lri_opt%x)
      NULLIFY (lri_opt%zet_init)

      lri_opt%opt_exps = .FALSE.
      lri_opt%opt_coeffs = .FALSE.
      lri_opt%use_condition_number = .FALSE.
      lri_opt%use_geometric_seq = .FALSE.
      lri_opt%use_constraints = .FALSE.

      lri_opt%nexp = 0
      lri_opt%ncoeff = 0

   END SUBROUTINE create_lri_opt

! **************************************************************************************************
!> \brief deallocates lri_opt
!> \param lri_opt optimization environment
! **************************************************************************************************
   SUBROUTINE deallocate_lri_opt(lri_opt)

      TYPE(lri_opt_type), POINTER                        :: lri_opt

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_opt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(lri_opt)) THEN
         IF (ASSOCIATED(lri_opt%subset)) THEN
            DO i = 1, SIZE(lri_opt%subset)
               DEALLOCATE (lri_opt%subset(i)%ncont_l)
            ENDDO
            DEALLOCATE (lri_opt%subset)
         ENDIF
         IF (ASSOCIATED(lri_opt%x)) THEN
            DEALLOCATE (lri_opt%x)
         ENDIF
         IF (ASSOCIATED(lri_opt%zet_init)) THEN
            DEALLOCATE (lri_opt%zet_init)
         ENDIF
         IF (ASSOCIATED(lri_opt%ri_gcc_orig)) THEN
            DO i = 1, SIZE(lri_opt%ri_gcc_orig)
               DEALLOCATE (lri_opt%ri_gcc_orig(i)%gcc_orig)
            ENDDO
            DEALLOCATE (lri_opt%ri_gcc_orig)
         ENDIF
         DEALLOCATE (lri_opt)
      ENDIF
   END SUBROUTINE deallocate_lri_opt

! **************************************************************************************************
!> \brief primitive Cartesian Gaussian functions are normalized. The normalization
!>        factor is included in the Gaussian contraction coefficients.
!>        Division by this factor to get the original gcc.
!> \param gcc_orig original contraction coefficient
!> \param gto_basis_set gaussian type basis set
!> \param lri_opt optimization environment
! **************************************************************************************************
   SUBROUTINE get_original_gcc(gcc_orig, gto_basis_set, lri_opt)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc_orig
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      TYPE(lri_opt_type), POINTER                        :: lri_opt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_original_gcc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: il, ipgf, iset, ishell, l, maxpgf, &
                                                            maxshell, nl, nset
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, ncont_l
      REAL(KIND=dp)                                      :: expzet, gcca, prefac, zeta

      maxpgf = SIZE(gto_basis_set%gcc, 1)
      maxshell = SIZE(gto_basis_set%gcc, 2)
      nset = SIZE(gto_basis_set%gcc, 3)

      ALLOCATE (gcc_orig(maxpgf, maxshell, nset))
      gcc_orig = 0.0_dp

      DO iset = 1, gto_basis_set%nset
         DO ishell = 1, gto_basis_set%nshell(iset)
            l = gto_basis_set%l(ishell, iset)
            expzet = 0.25_dp*REAL(2*l+3, dp)
            prefac = 2.0_dp**l*(2.0_dp/pi)**0.75_dp
            DO ipgf = 1, gto_basis_set%npgf(iset)
               gcca = gto_basis_set%gcc(ipgf, ishell, iset)
               zeta = gto_basis_set%zet(ipgf, iset)
               gcc_orig(ipgf, ishell, iset) = gcca/(prefac*zeta**expzet)
            END DO
         END DO
      END DO

      IF (lri_opt%opt_coeffs) THEN
         ! **** get number of contractions per quantum number
         CALL get_gto_basis_set(gto_basis_set=gto_basis_set, &
                                lmax=lmax, lmin=lmin)
         ALLOCATE (lri_opt%subset(nset))
         DO iset = 1, gto_basis_set%nset
            nl = lmax(iset)-lmin(iset)+1
            lri_opt%subset(iset)%nl = nl
            il = 1
            ALLOCATE (lri_opt%subset(iset)%ncont_l(nl))
            ncont_l => lri_opt%subset(iset)%ncont_l
            ncont_l = 1
            DO ishell = 2, gto_basis_set%nshell(iset)
               l = gto_basis_set%l(ishell, iset)
               IF (l == gto_basis_set%l(ishell-1, iset)) THEN
                  ncont_l(il) = ncont_l(il)+1
               ELSE
                  il = il+1
                  ncont_l(il) = 1
               ENDIF
            ENDDO
         END DO
      ENDIF

   END SUBROUTINE get_original_gcc

! **************************************************************************************************
!> \brief orthonormalize contraction coefficients using Gram-Schmidt
!> \param gcc contraction coefficient
!> \param gto_basis_set gaussian type basis set
!> \param lri_opt optimization environment
! **************************************************************************************************
   SUBROUTINE orthonormalize_gcc(gcc, gto_basis_set, lri_opt)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      TYPE(lri_opt_type), POINTER                        :: lri_opt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'orthonormalize_gcc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: il, iset, ishell, ishell1, ishell2, &
                                                            istart, nset
      INTEGER, DIMENSION(:), POINTER                     :: nshell
      REAL(KIND=dp)                                      :: gs_scale

      CALL get_gto_basis_set(gto_basis_set=gto_basis_set, nset=nset, nshell=nshell)

      DO iset = 1, nset
         istart = 1
         DO il = 1, lri_opt%subset(iset)%nl
            DO ishell1 = istart, istart+lri_opt%subset(iset)%ncont_l(il)-2
               DO ishell2 = ishell1+1, istart+lri_opt%subset(iset)%ncont_l(il)-1
                  gs_scale = DOT_PRODUCT(gcc(:, ishell2, iset), gcc(:, ishell1, iset))/ &
                             DOT_PRODUCT(gcc(:, ishell1, iset), gcc(:, ishell1, iset))
                  gcc(:, ishell2, iset) = gcc(:, ishell2, iset)- &
                                          gs_scale*gcc(:, ishell1, iset)
               ENDDO
            ENDDO
            istart = istart+lri_opt%subset(iset)%ncont_l(il)
         ENDDO

         DO ishell = 1, gto_basis_set%nshell(iset)
            gcc(:, ishell, iset) = gcc(:, ishell, iset)/ &
                                   SQRT(DOT_PRODUCT(gcc(:, ishell, iset), gcc(:, ishell, iset)))
         END DO
      ENDDO

   END SUBROUTINE orthonormalize_gcc

END MODULE lri_optimize_ri_basis_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Optimizes exponents and contraction coefficients of the lri auxiliary
!>        basis sets using the UOBYQA minimizer
!>        lri : local resolution of the identity
!> \par History
!>      created Dorothea Golze [05.2014]
!> \authors Dorothea Golze
! **************************************************************************************************
MODULE lri_optimize_ri_basis

   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type,&
                                              init_orb_basis_set
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_lri_opt_all,&
                                              do_lri_opt_coeff,&
                                              do_lri_opt_exps
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE lri_environment_init,            ONLY: lri_basis_init
   USE lri_environment_methods,         ONLY: calculate_avec,&
                                              calculate_lri_integrals,&
                                              calculate_lri_overlap_aabb
   USE lri_environment_types,           ONLY: lri_density_type,&
                                              lri_environment_type,&
                                              lri_int_rho_type,&
                                              lri_int_type,&
                                              lri_list_type,&
                                              lri_rhoab_type
   USE lri_optimize_ri_basis_types,     ONLY: create_lri_opt,&
                                              deallocate_lri_opt,&
                                              get_original_gcc,&
                                              lri_opt_type,&
                                              orthonormalize_gcc
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE powell,                          ONLY: opt_state_type,&
                                              powell_optimize
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_optimize_ri_basis'

   PUBLIC :: optimize_lri_basis

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief optimizes the lri basis set
!> \param qs_env qs environment
! **************************************************************************************************
   SUBROUTINE optimize_lri_basis(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'optimize_lri_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iunit, nkind
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(opt_state_type)                               :: opt_state
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(section_vals_type), POINTER                   :: dft_section, input, lri_optbas_section

      NULLIFY (atomic_kind_set, dft_section, lri_density, lri_env, &
               lri_opt, lri_optbas_section, rho_struct)
      NULLIFY (input, logger, para_env)

      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, input=input, &
                      lri_env=lri_env, lri_density=lri_density, nkind=nkind, &
                      para_env=para_env, rho=rho_struct)

      ! density matrix
      CALL qs_rho_get(rho_struct, rho_ao=pmatrix)

      logger => cp_get_default_logger()
      dft_section => section_vals_get_subs_vals(input, "DFT")
      lri_optbas_section => section_vals_get_subs_vals(input, &
                                                       "DFT%QS%OPTIMIZE_LRI_BASIS")
      iunit = cp_print_key_unit_nr(logger, input, "PRINT%PROGRAM_RUN_INFO", &
                                   extension=".opt")

      IF (iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Start optimization procedure")')
      ENDIF

      ! *** initialization
      CALL create_lri_opt(lri_opt)
      CALL init_optimization(lri_env, lri_opt, lri_optbas_section, &
                             opt_state, lri_opt%x, lri_opt%zet_init, nkind, iunit)

      CALL calculate_lri_overlap_aabb(lri_env, qs_env)

      ! *** ======================= START optimization =====================
      opt_state%state = 0
      DO
         IF (opt_state%state == 2) THEN
            CALL calc_lri_integrals_get_objective(lri_env, lri_density, qs_env, &
                                                  lri_opt, opt_state, pmatrix, para_env, &
                                                  nkind)
         ENDIF

         IF (opt_state%state == -1) EXIT

         CALL powell_optimize(opt_state%nvar, lri_opt%x, opt_state)
         CALL update_exponents(lri_env, lri_opt, lri_opt%x, lri_opt%zet_init, nkind)
         CALL print_optimization_update(opt_state, lri_opt, iunit)
      ENDDO
      ! *** ======================= END optimization =======================

      ! *** get final optimized parameters
      opt_state%state = 8
      CALL powell_optimize(opt_state%nvar, lri_opt%x, opt_state)
      CALL update_exponents(lri_env, lri_opt, lri_opt%x, lri_opt%zet_init, nkind)

      CALL write_optimized_lri_basis(lri_env, dft_section, nkind, lri_opt, &
                                     atomic_kind_set)

      IF (iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Number of function evaluations",T71,I10)') opt_state%nf
         WRITE (iunit, '(" POWELL| Final value of function",T61,F20.10)') opt_state%fopt
         WRITE (iunit, '(/," Printed optimized lri basis set to file")')
      ENDIF

      CALL cp_print_key_finished_output(iunit, logger, input, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL deallocate_lri_opt(lri_opt)

   END SUBROUTINE optimize_lri_basis

! **************************************************************************************************
!> \brief initialize optimization parameter
!> \param lri_env lri environment
!> \param lri_opt optimization environment
!> \param lri_optbas_section ...
!> \param opt_state state of the optimizer
!> \param x parameters to be optimized, i.e. exponents and contraction coeffs
!>        of the lri basis set
!> \param zet_init initial values of the exponents
!> \param nkind number of atom kinds
!> \param iunit output unit
! **************************************************************************************************
   SUBROUTINE init_optimization(lri_env, lri_opt, lri_optbas_section, opt_state, &
                                x, zet_init, nkind, iunit)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(section_vals_type), POINTER                   :: lri_optbas_section
      TYPE(opt_state_type)                               :: opt_state
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x, zet_init
      INTEGER, INTENT(IN)                                :: nkind, iunit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_optimization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, iset, ishell, n, nset
      INTEGER, DIMENSION(:), POINTER                     :: npgf, nshell
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc_orig
      TYPE(gto_basis_set_type), POINTER                  :: fbas

      NULLIFY (fbas, gcc_orig, npgf, nshell, zet)

      ALLOCATE (lri_opt%ri_gcc_orig(nkind))

      ! *** get parameters
      CALL get_optimization_parameter(lri_opt, lri_optbas_section, &
                                      opt_state)

      opt_state%nvar = 0
      opt_state%nf = 0
      opt_state%iprint = 1
      opt_state%unit = iunit

      ! *** init exponents
      IF (lri_opt%opt_exps) THEN
         n = 0
         DO ikind = 1, nkind
            fbas => lri_env%ri_basis(ikind)%gto_basis_set
            CALL get_gto_basis_set(gto_basis_set=fbas, &
                                   npgf=npgf, nset=nset, zet=zet)
            DO iset = 1, nset
               IF (lri_opt%use_geometric_seq .AND. npgf(iset) > 2) THEN
                  opt_state%nvar = opt_state%nvar+2
                  CALL reallocate(x, 1, opt_state%nvar)
                  x(n+1) = MAXVAL(zet(1:npgf(iset), iset))
                  x(n+2) = MINVAL(zet(1:npgf(iset), iset))
                  n = n+2
               ELSE
                  opt_state%nvar = opt_state%nvar+npgf(iset)
                  CALL reallocate(x, 1, opt_state%nvar)
                  x(n+1:n+npgf(iset)) = zet(1:npgf(iset), iset)
                  n = n+npgf(iset)
               ENDIF
               lri_opt%nexp = lri_opt%nexp+npgf(iset)
            ENDDO
         ENDDO

         ! *** constraints on exponents
         IF (lri_opt%use_constraints) THEN
            ALLOCATE (zet_init(SIZE(x)))
            zet_init(:) = x
         ELSE
            x(:) = SQRT(x)
         ENDIF
      ENDIF

      ! *** get the original gcc without normalization factor
      DO ikind = 1, nkind
         fbas => lri_env%ri_basis(ikind)%gto_basis_set
         CALL get_original_gcc(lri_opt%ri_gcc_orig(ikind)%gcc_orig, fbas, &
                               lri_opt)
      ENDDO

      ! *** init coefficients
      IF (lri_opt%opt_coeffs) THEN
         DO ikind = 1, nkind
            fbas => lri_env%ri_basis(ikind)%gto_basis_set
            gcc_orig => lri_opt%ri_gcc_orig(ikind)%gcc_orig
            CALL get_gto_basis_set(gto_basis_set=fbas, &
                                   npgf=npgf, nset=nset, nshell=nshell, zet=zet)
            ! *** Gram Schmidt orthonormalization
            CALL orthonormalize_gcc(gcc_orig, fbas, lri_opt)
            n = opt_state%nvar
            DO iset = 1, nset
               DO ishell = 1, nshell(iset)
                  opt_state%nvar = opt_state%nvar+npgf(iset)
                  CALL reallocate(x, 1, opt_state%nvar)
                  x(n+1:n+npgf(iset)) = gcc_orig(1:npgf(iset), ishell, iset)
                  lri_opt%ncoeff = lri_opt%ncoeff+npgf(iset)
                  n = n+npgf(iset)
               ENDDO
            ENDDO
         ENDDO
      ENDIF

      IF (iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Accuracy",T69,ES12.5)') opt_state%rhoend
         WRITE (iunit, '(" POWELL| Initial step size",T69,ES12.5)') opt_state%rhobeg
         WRITE (iunit, '(" POWELL| Maximum number of evaluations",T71,I10)') &
            opt_state%maxfun
         WRITE (iunit, '(" POWELL| Total number of parameters",T71,I10)') &
            opt_state%nvar
      ENDIF

   END SUBROUTINE init_optimization

! **************************************************************************************************
!> \brief read input for optimization
!> \param lri_opt optimization environmnet
!> \param lri_optbas_section ...
!> \param opt_state state of the optimizer
! **************************************************************************************************
   SUBROUTINE get_optimization_parameter(lri_opt, lri_optbas_section, &
                                         opt_state)

      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(section_vals_type), POINTER                   :: lri_optbas_section
      TYPE(opt_state_type)                               :: opt_state

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_optimization_parameter', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: degree_freedom
      TYPE(section_vals_type), POINTER                   :: constrain_exp_section

      NULLIFY (constrain_exp_section)

      ! *** parameter for POWELL optimizer
      CALL section_vals_val_get(lri_optbas_section, "ACCURACY", &
                                r_val=opt_state%rhoend)
      CALL section_vals_val_get(lri_optbas_section, "STEP_SIZE", &
                                r_val=opt_state%rhobeg)
      CALL section_vals_val_get(lri_optbas_section, "MAX_FUN", &
                                i_val=opt_state%maxfun)

      ! *** parameters which are optimized, i.e. exps or coeff or both
      CALL section_vals_val_get(lri_optbas_section, "DEGREES_OF_FREEDOM", &
                                i_val=degree_freedom)

      SELECT CASE (degree_freedom)
      CASE (do_lri_opt_all)
         lri_opt%opt_coeffs = .TRUE.
         lri_opt%opt_exps = .TRUE.
      CASE (do_lri_opt_coeff)
         lri_opt%opt_coeffs = .TRUE.
      CASE (do_lri_opt_exps)
         lri_opt%opt_exps = .TRUE.
      CASE DEFAULT
         CALL cp__b("lri_optimize_ri_basis.F",334,"No initialization available?????")
      END SELECT

      ! *** restraint
      CALL section_vals_val_get(lri_optbas_section, "USE_CONDITION_NUMBER", &
                                l_val=lri_opt%use_condition_number)
      CALL section_vals_val_get(lri_optbas_section, "CONDITION_WEIGHT", &
                                r_val=lri_opt%cond_weight)
      CALL section_vals_val_get(lri_optbas_section, "GEOMETRIC_SEQUENCE", &
                                l_val=lri_opt%use_geometric_seq)

      ! *** get constraint info
      constrain_exp_section => section_vals_get_subs_vals(lri_optbas_section, &
                                                          "CONSTRAIN_EXPONENTS")
      CALL section_vals_get(constrain_exp_section, explicit=lri_opt%use_constraints)

      IF (lri_opt%use_constraints) THEN
         CALL section_vals_val_get(constrain_exp_section, "SCALE", &
                                   r_val=lri_opt%scale_exp)
         CALL section_vals_val_get(constrain_exp_section, "FERMI_EXP", &
                                   r_val=lri_opt%fermi_exp)
      ENDIF

   END SUBROUTINE get_optimization_parameter

! **************************************************************************************************
!> \brief update exponents after optimization step
!> \param lri_env lri environment
!> \param lri_opt optimization environment
!> \param x optimization parameters
!> \param zet_init initial values of the exponents
!> \param nkind number of atomic kinds
! **************************************************************************************************
   SUBROUTINE update_exponents(lri_env, lri_opt, x, zet_init, nkind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x, zet_init
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_exponents', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, iset, ishell, n, nset, nvar_exp
      INTEGER, DIMENSION(:), POINTER                     :: npgf, nshell
      REAL(KIND=dp)                                      :: zet_max, zet_min
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet, zet_trans
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc_orig
      TYPE(gto_basis_set_type), POINTER                  :: fbas

      NULLIFY (fbas, gcc_orig, npgf, nshell, zet_trans, zet)

      ! nvar_exp: number of exponents that are variables
      nvar_exp = SIZE(x)-lri_opt%ncoeff
      ALLOCATE (zet_trans(nvar_exp))

      ! *** update exponents
      IF (lri_opt%opt_exps) THEN
         IF (lri_opt%use_constraints) THEN
            zet => x(1:nvar_exp)
            CALL transfer_exp(lri_opt, zet, zet_init, zet_trans, nvar_exp)
         ELSE
            zet_trans(:) = x(1:nvar_exp)**2.0_dp
         ENDIF
         n = 0
         DO ikind = 1, nkind
            fbas => lri_env%ri_basis(ikind)%gto_basis_set
            CALL get_gto_basis_set(gto_basis_set=fbas, npgf=npgf, nset=nset)
            DO iset = 1, nset
               IF (lri_opt%use_geometric_seq .AND. npgf(iset) > 2) THEN
                  zet_max = MAXVAL(zet_trans(n+1:n+2))
                  zet_min = MINVAL(zet_trans(n+1:n+2))
                  zet => fbas%zet(1:npgf(iset), iset)
                  CALL geometric_progression(zet, zet_max, zet_min, npgf(iset))
                  n = n+2
               ELSE
                  fbas%zet(1:npgf(iset), iset) = zet_trans(n+1:n+npgf(iset))
                  n = n+npgf(iset)
               ENDIF
            ENDDO
         ENDDO
      ENDIF

      ! *** update coefficients
      IF (lri_opt%opt_coeffs) THEN
         n = nvar_exp
         DO ikind = 1, nkind
            fbas => lri_env%ri_basis(ikind)%gto_basis_set
            gcc_orig => lri_opt%ri_gcc_orig(ikind)%gcc_orig
            CALL get_gto_basis_set(gto_basis_set=fbas, &
                                   nshell=nshell, npgf=npgf, nset=nset)
            DO iset = 1, nset
               DO ishell = 1, nshell(iset)
                  gcc_orig(1:npgf(iset), ishell, iset) = x(n+1:n+npgf(iset))
                  n = n+npgf(iset)
               ENDDO
            ENDDO
            ! *** Gram Schmidt orthonormalization
            CALL orthonormalize_gcc(gcc_orig, fbas, lri_opt)
         ENDDO
      ENDIF

      DEALLOCATE (zet_trans)
   END SUBROUTINE update_exponents

! **************************************************************************************************
!> \brief employ Fermi constraint, transfer exponents
!> \param lri_opt optimization environment
!> \param zet untransferred exponents
!> \param zet_init intial value of the eponents
!> \param zet_trans transferred exponents
!> \param nvar number of optimized exponents
! **************************************************************************************************
   SUBROUTINE transfer_exp(lri_opt, zet, zet_init, zet_trans, nvar)

      TYPE(lri_opt_type), POINTER                        :: lri_opt
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet, zet_init, zet_trans
      INTEGER, INTENT(IN)                                :: nvar

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transfer_exp', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: a
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet_max, zet_min

      ALLOCATE (zet_max(nvar), zet_min(nvar))

      zet_min(:) = zet_init(:)*(1.0_dp-lri_opt%scale_exp)
      zet_max(:) = zet_init(:)*(1.0_dp+lri_opt%scale_exp)

      a = lri_opt%fermi_exp

      zet_trans = zet_min+(zet_max-zet_min)/(1+EXP(-a*(zet-zet_init)))

      DEALLOCATE (zet_max, zet_min)

   END SUBROUTINE transfer_exp

! **************************************************************************************************
!> \brief complete geometric sequence
!> \param zet all exponents of the set
!> \param zet_max maximal exponent of the set
!> \param zet_min minimal exponent of the set
!> \param nexp number of exponents of the set
! **************************************************************************************************
   SUBROUTINE geometric_progression(zet, zet_max, zet_min, nexp)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet
      REAL(KIND=dp), INTENT(IN)                          :: zet_max, zet_min
      INTEGER, INTENT(IN)                                :: nexp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'geometric_progression', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      REAL(KIND=dp)                                      :: q

      n = nexp-1

      q = (zet_min/zet_max)**(1._dp/REAL(n, dp))

      DO i = 1, nexp
         zet(i) = zet_max*q**(i-1)
      ENDDO

   END SUBROUTINE geometric_progression

! **************************************************************************************************
!> \brief calculates the lri integrals and coefficients with the new exponents
!>        of the lri basis sets and calculates the objective function
!> \param lri_env lri environment
!> \param lri_density ...
!> \param qs_env ...
!> \param lri_opt optimization enviroment
!> \param opt_state state of the optimizer
!> \param pmatrix density matrix
!> \param para_env ...
!> \param nkind number of atomic kinds
! **************************************************************************************************
   SUBROUTINE calc_lri_integrals_get_objective(lri_env, lri_density, qs_env, &
                                               lri_opt, opt_state, pmatrix, para_env, &
                                               nkind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(opt_state_type)                               :: opt_state
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_lri_integrals_get_objective', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nset
      INTEGER, DIMENSION(:), POINTER                     :: npgf
      TYPE(gto_basis_set_type), POINTER                  :: fbas

      NULLIFY (fbas, npgf)

      !*** build new transformation matrices sphi with new exponents
      DO ikind = 1, nkind
         fbas => lri_env%ri_basis(ikind)%gto_basis_set
         CALL get_gto_basis_set(gto_basis_set=fbas, npgf=npgf, nset=nset)
         !build new sphi
         fbas%gcc = lri_opt%ri_gcc_orig(ikind)%gcc_orig
         CALL init_orb_basis_set(fbas)
      ENDDO
      CALL lri_basis_init(lri_env)
      CALL calculate_lri_integrals(lri_env, qs_env, calculate_forces=.FALSE.)
      CALL calculate_avec(lri_env, lri_density, qs_env, pmatrix)
      IF (lri_opt%use_condition_number) THEN
         CALL get_condition_number_of_overlap(lri_env)
      ENDIF
      CALL calculate_objective(lri_env, lri_density, lri_opt, pmatrix, para_env, &
                               opt_state%f)

   END SUBROUTINE calc_lri_integrals_get_objective

! **************************************************************************************************
!> \brief calculates the objective function defined as integral of the square
!>        of rhoexact - rhofit, i.e. integral[(rhoexact-rhofit)**2]
!>        rhoexact is the exact pair density and rhofit the lri pair density
!> \param lri_env lri environment
!> \param lri_density ...
!> \param lri_opt optimization environment
!> \param pmatrix density matrix
!> \param para_env ...
!> \param fobj objective function
! **************************************************************************************************
   SUBROUTINE calculate_objective(lri_env, lri_density, lri_opt, pmatrix, para_env, &
                                  fobj)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: pmatrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(OUT)                         :: fobj

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_objective', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, iac, iatom, ikind, ilist, isgfa, ispin, jatom, jkind, jneighbor, jsgfa, &
         ksgfb, lsgfb, nba, nbb, nfa, nfb, nkind, nlist, nn, nneighbor, nspin
      LOGICAL                                            :: found, trans
      REAL(KIND=dp)                                      :: obj_ab, rhoexact_sq, rhofit_sq, rhomix
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pbij
      TYPE(cp_dbcsr_type), POINTER                       :: pmat
      TYPE(lri_int_rho_type), POINTER                    :: lriir
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(lri_list_type), POINTER                       :: lri_rho
      TYPE(lri_rhoab_type), POINTER                      :: lrho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list

      CALL timeset(routineN, handle)
      NULLIFY (lrii, lriir, lri_rho, lrho, nl_iterator, pmat, soo_list)

      IF (ASSOCIATED(lri_env%soo_list)) THEN
         soo_list => lri_env%soo_list

         nkind = lri_env%lri_ints%nkind
         nspin = SIZE(pmatrix)
         fobj = 0._dp
         lri_opt%rho_diff = 0._dp

         DO ispin = 1, nspin

            pmat => pmatrix(ispin)%matrix
            lri_rho => lri_density%lri_rhos(ispin)%lri_list

            CALL neighbor_list_iterator_create(nl_iterator, soo_list)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, &
                                      jatom=jatom, nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor)

               iac = ikind+nkind*(jkind-1)

               IF (.NOT. ASSOCIATED(lri_env%lri_ints%lri_atom(iac)%lri_node)) CYCLE

               lrii => lri_env%lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)
               lriir => lri_env%lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho(jneighbor)
               lrho => lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(jneighbor)
               nfa = lrii%nfa
               nfb = lrii%nfb
               nba = lrii%nba
               nbb = lrii%nbb
               nn = nfa+nfb

               rhoexact_sq = 0._dp
               rhomix = 0._dp
               rhofit_sq = 0._dp
               obj_ab = 0._dp

               NULLIFY (pbij)
               IF (iatom <= jatom) THEN
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=iatom, col=jatom, block=pbij, found=found)
                  trans = .FALSE.
               ELSE
                  CALL cp_dbcsr_get_block_p(matrix=pmat, row=jatom, col=iatom, block=pbij, found=found)
                  trans = .TRUE.
               END IF
               IF(.NOT.(found))CALL cp__a("lri_optimize_ri_basis.F",639)

               ! *** calculate integral of the square of exact density rhoexact_sq
               IF (trans) THEN
                  DO isgfa = 1, nba
                     DO jsgfa = 1, nba
                        DO ksgfb = 1, nbb
                           DO lsgfb = 1, nbb
                              rhoexact_sq = rhoexact_sq+pbij(ksgfb, isgfa)*pbij(lsgfb, jsgfa) &
                                            *lriir%soaabb(isgfa, jsgfa, ksgfb, lsgfb)
                           END DO
                        END DO
                     ENDDO
                  END DO
               ELSE
                  DO isgfa = 1, nba
                     DO jsgfa = 1, nba
                        DO ksgfb = 1, nbb
                           DO lsgfb = 1, nbb
                              rhoexact_sq = rhoexact_sq+pbij(isgfa, ksgfb)*pbij(jsgfa, lsgfb) &
                                            *lriir%soaabb(isgfa, jsgfa, ksgfb, lsgfb)
                           END DO
                        END DO
                     ENDDO
                  END DO
               ENDIF

               ! *** calculate integral of the square of the fitted density rhofit_sq
               DO isgfa = 1, nfa
                  DO jsgfa = 1, nfa
                     rhofit_sq = rhofit_sq+lrho%avec(isgfa)*lrho%avec(jsgfa) &
                                 *lri_env%bas_prop(ikind)%ri_ovlp(isgfa, jsgfa)
                  ENDDO
               ENDDO
               IF (iatom /= jatom) THEN
                  DO ksgfb = 1, nfb
                     DO lsgfb = 1, nfb
                        rhofit_sq = rhofit_sq+lrho%avec(nfa+ksgfb)*lrho%avec(nfa+lsgfb) &
                                    *lri_env%bas_prop(jkind)%ri_ovlp(ksgfb, lsgfb)
                     ENDDO
                  ENDDO
                  DO isgfa = 1, nfa
                     DO ksgfb = 1, nfb
                        rhofit_sq = rhofit_sq+2._dp*lrho%avec(isgfa)*lrho%avec(nfa+ksgfb) &
                                    *lrii%sab(isgfa, ksgfb)
                     ENDDO
                  ENDDO
               ENDIF

               ! *** and integral of the product of exact and fitted density rhomix
               IF (iatom == jatom) THEN
                  rhomix = SUM(lrho%avec(1:nfa)*lrho%tvec(1:nfa))
               ELSE
                  rhomix = SUM(lrho%avec(1:nn)*lrho%tvec(1:nn))
               ENDIF

               ! *** calculate contribution to the objective function for pair ab
               ! *** taking density matrix symmetry in account, double-count for off-diagonal blocks
               IF (iatom == jatom) THEN
                  obj_ab = rhoexact_sq-2._dp*rhomix+rhofit_sq
               ELSE
                  obj_ab = 2.0_dp*(rhoexact_sq-2._dp*rhomix+rhofit_sq)
               ENDIF

               IF (lri_opt%use_condition_number) THEN
                  fobj = fobj+obj_ab+lri_opt%cond_weight*LOG(lrii%cond_num)
                  lri_opt%rho_diff = lri_opt%rho_diff+obj_ab
               ELSE
                  fobj = fobj+obj_ab
               ENDIF

            ENDDO

            CALL neighbor_list_iterator_release(nl_iterator)

         ENDDO
         CALL mp_sum(fobj, para_env%group)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calculate_objective

! **************************************************************************************************
!> \brief get condition number of overlap matrix
!> \param lri_env lri environment
! **************************************************************************************************
   SUBROUTINE get_condition_number_of_overlap(lri_env)

      TYPE(lri_environment_type), POINTER                :: lri_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_condition_number_of_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iac, iatom, ikind, ilist, info, &
                                                            jatom, jkind, jneighbor, lwork, nfa, &
                                                            nfb, nkind, nlist, nn, nneighbor
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: diag, off_diag, tau
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: smat
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: soo_list

      CALL timeset(routineN, handle)
      NULLIFY (lrii, nl_iterator, smat, soo_list)

      soo_list => lri_env%soo_list

      nkind = lri_env%lri_ints%nkind
      CALL neighbor_list_iterator_create(nl_iterator, soo_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, &
                                jatom=jatom, nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor)

         iac = ikind+nkind*(jkind-1)
         IF (.NOT. ASSOCIATED(lri_env%lri_ints%lri_atom(iac)%lri_node)) CYCLE
         lrii => lri_env%lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

         nfa = lrii%nfa
         nfb = lrii%nfb
         nn = nfa+nfb

         ! build the overlap matrix
         IF (iatom == jatom) THEN
            ALLOCATE (smat(nfa, nfa))
         ELSE
            ALLOCATE (smat(nn, nn))
         ENDIF
         smat(1:nfa, 1:nfa) = lri_env%bas_prop(ikind)%ri_ovlp(1:nfa, 1:nfa)
         IF (iatom /= jatom) THEN
            nn = nfa+nfb
            smat(1:nfa, nfa+1:nn) = lrii%sab(1:nfa, 1:nfb)
            smat(nfa+1:nn, 1:nfa) = TRANSPOSE(lrii%sab(1:nfa, 1:nfb))
            smat(nfa+1:nn, nfa+1:nn) = lri_env%bas_prop(jkind)%ri_ovlp(1:nfb, 1:nfb)
         ENDIF

         IF (iatom == jatom) nn = nfa
         ALLOCATE (diag(nn), off_diag(nn-1), tau(nn-1), work(1))
         diag = 0.0_dp
         off_diag = 0.0_dp
         tau = 0.0_dp
         work = 0.0_dp
         lwork = -1
         ! get lwork
         CALL DSYTRD('U', nn, smat, nn, diag, off_diag, tau, work, lwork, info)
         lwork = INT(work(1))
         CALL reallocate(work, 1, lwork)
         ! get the eigenvalues
         CALL DSYTRD('U', nn, smat, nn, diag, off_diag, tau, work, lwork, info)
         CALL DSTERF(nn, diag, off_diag, info)

         lrii%cond_num = MAXVAL(ABS(diag))/MINVAL(ABS(diag))

         DEALLOCATE (diag, off_diag, smat, tau, work)
      END DO

      CALL neighbor_list_iterator_release(nl_iterator)

      CALL timestop(handle)

   END SUBROUTINE get_condition_number_of_overlap

! **************************************************************************************************
!> \brief print recent information on optimization
!> \param opt_state state of the optimizer
!> \param lri_opt optimization environment
!> \param iunit ...
! **************************************************************************************************
   SUBROUTINE print_optimization_update(opt_state, lri_opt, iunit)

      TYPE(opt_state_type)                               :: opt_state
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      INTEGER, INTENT(IN)                                :: iunit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_optimization_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n10

      n10 = MAX(opt_state%maxfun/100, 1)

      IF (opt_state%nf == 2 .AND. opt_state%state == 2 .AND. iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Initial value of function",T61,F20.10)') opt_state%f
      END IF
      IF (MOD(opt_state%nf, n10) == 0 .AND. opt_state%nf > 1 .AND. iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Reached",i4,"% of maximal function calls",T61,F20.10)') &
            INT(REAL(opt_state%nf, dp)/REAL(opt_state%maxfun, dp)*100._dp), opt_state%fopt
      ENDIF
      IF (lri_opt%use_condition_number) THEN
         IF (MOD(opt_state%nf, n10) == 0 .AND. opt_state%nf > 1 .AND. iunit > 0) THEN
            WRITE (iunit, '(" POWELL| Recent value of function without condition nr.",T61,F20.10)') &
               lri_opt%rho_diff
         ENDIF
      ENDIF

   END SUBROUTINE print_optimization_update

! **************************************************************************************************
!> \brief write optimized LRI basis set to file
!> \param lri_env ...
!> \param dft_section ...
!> \param nkind ...
!> \param lri_opt ...
!> \param atomic_kind_set ...
! **************************************************************************************************
   SUBROUTINE write_optimized_lri_basis(lri_env, dft_section, nkind, lri_opt, &
                                        atomic_kind_set)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      INTEGER, INTENT(IN)                                :: nkind
      TYPE(lri_opt_type), POINTER                        :: lri_opt
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_optimized_lri_basis', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER                                            :: cc_l, ikind, ipgf, iset, ishell, nset, &
                                                            output_file
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc_orig
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: fbas
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (fbas, gcc_orig, l, lmax, lmin, logger, npgf, nshell, print_key, zet)

      !*** do the printing
      print_key => section_vals_get_subs_vals(dft_section, &
                                              "PRINT%OPTIMIZE_LRI_BASIS")
      logger => cp_get_default_logger()
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           dft_section, "PRINT%OPTIMIZE_LRI_BASIS"), &
                cp_p_file)) THEN
         output_file = cp_print_key_unit_nr(logger, dft_section, &
                                            "PRINT%OPTIMIZE_LRI_BASIS", &
                                            extension=".opt", &
                                            file_status="REPLACE", &
                                            file_action="WRITE", &
                                            file_form="FORMATTED")

         IF (output_file > 0) THEN

            filename = cp_print_key_generate_filename(logger, &
                                                      print_key, extension=".opt", &
                                                      my_local=.TRUE.)

            DO ikind = 1, nkind
               fbas => lri_env%ri_basis(ikind)%gto_basis_set
               gcc_orig => lri_opt%ri_gcc_orig(ikind)%gcc_orig
               CALL get_gto_basis_set(gto_basis_set=fbas, &
                                      l=l, lmax=lmax, lmin=lmin, &
                                      npgf=npgf, nshell=nshell, &
                                      nset=nset, zet=zet)
               WRITE (output_file, '(T1,A2,T5,A)') TRIM(atomic_kind_set(ikind)%name), &
                  TRIM(fbas%name)
               WRITE (output_file, '(T1,I4)') nset
               DO iset = 1, nset
                  WRITE (output_file, '(4(1X,I0))', advance='no') 2, lmin(iset), &
                     lmax(iset), npgf(iset)
                  cc_l = 1
                  DO ishell = 1, nshell(iset)
                     IF (ishell /= nshell(iset)) THEN
                        IF (l(ishell, iset) == l(ishell+1, iset)) THEN
                           cc_l = cc_l+1
                        ELSE
                           WRITE (output_file, '(1X,I0)', advance='no') cc_l
                           cc_l = 1
                        ENDIF
                     ELSE
                        WRITE (output_file, '(1X,I0)') cc_l
                     ENDIF
                  ENDDO
                  DO ipgf = 1, npgf(iset)
                     WRITE (output_file, '(F18.12)', advance='no') zet(ipgf, iset)
                     DO ishell = 1, nshell(iset)
                        IF (ishell == nshell(iset)) THEN
                           WRITE (output_file, '(T5,F18.12)') gcc_orig(ipgf, ishell, iset)
                        ELSE
                           WRITE (output_file, '(T5,F18.12)', advance='no') gcc_orig(ipgf, ishell, iset)
                        ENDIF
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

         ENDIF

         CALL cp_print_key_finished_output(output_file, logger, dft_section, &
                                           "PRINT%OPTIMIZE_LRI_BASIS")
      ENDIF

   END SUBROUTINE write_optimized_lri_basis

END MODULE lri_optimize_ri_basis
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize the XAS orbitals for specific core excitations
!>       Either the GS orbitals are used as initial guess, or the
!>       xas mos are read from a previous calculation.
!>       In the latter case, the core-hole potetial should be the same.
!> \note
!>       The restart with the same core-hole potential should be checked
!>       and a wrong restart should stop the program
!> \par History
!>      created 09.2006
!> \author MI (09.2006)
! **************************************************************************************************
MODULE xas_restart

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_type,&
                                              cp_fm_write_unformatted
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gspace_mixing,                ONLY: mixing_init
   USE qs_ks_types,                     ONLY: qs_ks_did_change
   USE qs_mo_io,                        ONLY: wfn_restart_file_name
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              set_mo_set
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE scf_control_types,               ONLY: scf_control_type
   USE string_utilities,                ONLY: xstring
   USE xas_env_types,                   ONLY: get_xas_env,&
                                              set_xas_env,&
                                              xas_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xas_restart'

! *** Public subroutines ***

   PUBLIC ::  xas_read_restart, xas_write_restart, xas_initialize_rho, find_excited_core_orbital

CONTAINS

! **************************************************************************************************
!> \brief Set up for reading the restart
!>      corresponing to the excitation of iatom
!>      If the corresponding restart file does not exist
!>      the GS orbitals are used as initial guess
!> \param xas_env ...
!> \param xas_section input section for XAS calculations
!>      qs_env:
!> \param qs_env ...
!> \param xas_method ...
!> \param iatom index of the absorbing atom
!> \param estate index of the core-hole orbital
!>      error:
!> \par History
!>      09.2006 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE xas_read_restart(xas_env, xas_section, qs_env, xas_method, iatom, estate)

      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(section_vals_type), POINTER                   :: xas_section
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: xas_method, iatom
      INTEGER, INTENT(OUT)                               :: estate

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xas_read_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      INTEGER :: group, handle, i, ia, ie, ispin, nao, nao_read, nelectron, nexc_atoms, &
         nexc_atoms_read, nexc_search, nexc_search_read, nmo, nmo_read, output_unit, rst_unit, &
         source, xas_estate, xas_estate_read, xas_method_read
      LOGICAL                                            :: file_exists
      REAL(dp)                                           :: occ_estate, occ_estate_read, &
                                                            xas_nelectron, xas_nelectron_read
      REAL(dp), DIMENSION(:), POINTER                    :: eigenvalues, occupation_numbers
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eig_read, occ_read
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CALL timeset(routineN, handle)

      file_exists = .FALSE.
      rst_unit = -1

      NULLIFY (eigenvalues, matrix_s, mos, occupation_numbers, vecbuffer)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      output_unit = cp_print_key_unit_nr(logger, xas_section, &
                                         "PRINT%PROGRAM_RUN_INFO", extension=".Log")

      CALL get_qs_env(qs_env=qs_env, para_env=para_env)
      group = para_env%group
      source = para_env%source

      IF (para_env%ionode) THEN
         CALL wfn_restart_file_name(filename, file_exists, xas_section, logger, &
                                    xas=.TRUE.)

         CALL xstring(filename, ia, ie)
         filename = filename(ia:ie)//'-at'// &
                    TRIM(ADJUSTL(cp_to_string(iatom)))//'.rst'

         INQUIRE (FILE=filename, EXIST=file_exists)
         ! open file
         IF (file_exists) THEN

            CALL open_file(file_name=TRIM(filename), &
                           file_action="READ", &
                           file_form="UNFORMATTED", &
                           file_position="REWIND", &
                           file_status="OLD", &
                           unit_number=rst_unit)

            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,T20,A,I5,/)") &
               "Read restart file for atom ", iatom

         ELSE IF (.NOT. file_exists) THEN
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,T10,A,I5,A,/)") &
               "Restart file for atom ", iatom, &
               " not available. Initialization done with GS orbitals"
         END IF
      END IF
      CALL mp_bcast(file_exists, source, group)

      CALL get_xas_env(xas_env=xas_env, occ_estate=occ_estate, xas_estate=xas_estate, &
                       xas_nelectron=xas_nelectron, nexc_search=nexc_search, nexc_atoms=nexc_atoms)

      IF (file_exists) THEN
         CALL get_qs_env(qs_env=qs_env, mos=mos, matrix_s=matrix_s)

         IF (rst_unit > 0) THEN
            READ (rst_unit) xas_method_read
            READ (rst_unit) nexc_search_read, nexc_atoms_read, occ_estate_read, xas_nelectron_read
            READ (rst_unit) xas_estate_read

            IF (xas_method_read /= xas_method) &
               CALL cp__b("xas_restart.F",180," READ XAS RESTART: restart with different XAS method is not possible. ")
            IF (nexc_atoms_read /= nexc_atoms) &
               CALL cp_abort(cp__l("xas_restart.F",182), &
                             " READ XAS RESTART: restart with different excited atoms "// &
                             " is not possible. Start instead a new XAS run with the new set of atoms. ")
         ENDIF

         CALL mp_bcast(xas_estate_read, source, group)
         CALL set_xas_env(xas_env=xas_env, xas_estate=xas_estate_read)
         estate = xas_estate_read

         CALL get_mo_set(mo_set=mos(1)%mo_set, nao=nao)
         ALLOCATE (vecbuffer(1, nao))

         DO ispin = 1, SIZE(mos)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo, eigenvalues=eigenvalues, &
                            occupation_numbers=occupation_numbers, mo_coeff=mo_coeff, nelectron=nelectron)
            eigenvalues = 0.0_dp
            occupation_numbers = 0.0_dp
            CALL cp_fm_set_all(mo_coeff, 0.0_dp)
            IF (para_env%ionode) THEN
               READ (rst_unit) nao_read, nmo_read
               IF (nao /= nao_read) &
                  CALL cp__b("xas_restart.F",203,"To change basis is not possible. ")
               ALLOCATE (eig_read(nmo_read), occ_read(nmo_read))
               eig_read = 0.0_dp
               occ_read = 0.0_dp
               nmo = MIN(nmo, nmo_read)
               READ (rst_unit) eig_read(1:nmo_read), occ_read(1:nmo_read)
               eigenvalues(1:nmo) = eig_read(1:nmo)
               occupation_numbers(1:nmo) = occ_read(1:nmo)
               IF (nmo_read > nmo) THEN
                  IF (occupation_numbers(nmo) >= EPSILON(0.0_dp)) &
                     CALL cp_warn(cp__l("xas_restart.F",213), &
                                  "The number of occupied MOs on the restart unit is larger than "// &
                                  "the allocated MOs.")

               END IF
               DEALLOCATE (eig_read, occ_read)
            ENDIF
            CALL mp_bcast(eigenvalues, source, group)
            CALL mp_bcast(occupation_numbers, source, group)

            DO i = 1, nmo
               IF (para_env%ionode) THEN
                  READ (rst_unit) vecbuffer
               ELSE
                  vecbuffer(1, :) = 0.0_dp
               END IF
               CALL mp_bcast(vecbuffer, source, group)
               CALL cp_fm_set_submatrix(mo_coeff, &
                                        vecbuffer, 1, i, nao, 1, transpose=.TRUE.)
            END DO
            ! Skip extra MOs if there any
            IF (para_env%ionode) THEN
               DO i = nmo+1, nmo_read
                  READ (rst_unit) vecbuffer
               END DO
            END IF

         END DO ! ispin

         DEALLOCATE (vecbuffer)

!      nspin = SIZE(mos,1)
!      DO ispin = 1,nspin
!      ! ortho so that one can restart for different positions (basis sets?)
!         NULLIFY(mo_coeff)
!         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff,homo=homo)
!         CALL make_basis_sm(mo_coeff,homo,matrix_s(1)%matrix)
!      END DO
      END IF !file_exist

      IF (para_env%ionode) THEN
         IF (file_exists) CALL close_file(unit_number=rst_unit)
      END IF

      CALL timestop(handle)

   END SUBROUTINE xas_read_restart

! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
!> \param xas_section ...
!> \param qs_env ...
!> \param xas_method ...
!> \param iatom ...
! **************************************************************************************************
   SUBROUTINE xas_write_restart(xas_env, xas_section, qs_env, xas_method, iatom)

      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(section_vals_type), POINTER                   :: xas_section
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: xas_method, iatom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xas_write_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_middle
      INTEGER                                            :: handle, ispin, nao, nexc_atoms, &
                                                            nexc_search, nmo, output_unit, &
                                                            rst_unit, xas_estate
      REAL(dp)                                           :: occ_estate, xas_nelectron
      REAL(dp), DIMENSION(:), POINTER                    :: eigenvalues, occupation_numbers
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: print_key

      CALL timeset(routineN, handle)
      NULLIFY (mos, logger, print_key)
      logger => cp_get_default_logger()

      CALL get_xas_env(xas_env=xas_env, occ_estate=occ_estate, xas_estate=xas_estate, &
                       xas_nelectron=xas_nelectron, nexc_search=nexc_search, nexc_atoms=nexc_atoms)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           xas_section, "PRINT%RESTART", used_print_key=print_key), &
                cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, xas_section, &
                                            "PRINT%PROGRAM_RUN_INFO", extension=".Log")

         CALL get_qs_env(qs_env=qs_env, mos=mos)

         ! Open file
         rst_unit = -1
         my_middle = 'at'//TRIM(ADJUSTL(cp_to_string(iatom)))
         rst_unit = cp_print_key_unit_nr(logger, xas_section, "PRINT%RESTART", &
                                         extension=".rst", file_status="REPLACE", file_action="WRITE", &
                                         file_form="UNFORMATTED", middle_name=TRIM(my_middle))

         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   middle_name=TRIM(my_middle), extension=".rst", &
                                                   my_local=.FALSE.)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,I5,A,A,/)") &
               "Xas orbitals  for the absorbing atom ", iatom, &
               " are written in ", TRIM(filename)

         END IF

         ! Write mos
         IF (rst_unit > 0) THEN
            WRITE (rst_unit) xas_method
            WRITE (rst_unit) nexc_search, nexc_atoms, occ_estate, xas_nelectron
            WRITE (rst_unit) xas_estate
         ENDIF
         DO ispin = 1, SIZE(mos)
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo, &
                            eigenvalues=eigenvalues, occupation_numbers=occupation_numbers)
            IF ((rst_unit > 0)) THEN
               WRITE (rst_unit) nao, nmo
               WRITE (rst_unit) eigenvalues(1:nmo), &
                  occupation_numbers(1:nmo)
            END IF
            CALL cp_fm_write_unformatted(mo_coeff, rst_unit)
         END DO

! Close file
         CALL cp_print_key_finished_output(rst_unit, logger, xas_section, &
                                           "PRINT%RESTART")
      END IF
      CALL timestop(handle)

   END SUBROUTINE xas_write_restart

!****f* xas_restart/xas_initialize_rho [1.0] *

! **************************************************************************************************
!> \brief Once the mos and the occupation numbers are initialized
!>      the electronic density of the excited state can be calclated
!> \param qs_env ...
!> \param scf_env ...
!> \param scf_control ...
!> \par History
!>      09-2006 MI created
!> \author MI
! **************************************************************************************************
   SUBROUTINE xas_initialize_rho(qs_env, scf_env, scf_control)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xas_initialize_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nelectron
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom
      TYPE(xas_environment_type), POINTER                :: xas_env

      CALL timeset(routineN, handle)

      NULLIFY (mos, rho, xas_env, para_env, rho_ao)

      CALL get_qs_env(qs_env, &
                      mos=mos, &
                      rho=rho, &
                      xas_env=xas_env, &
                      para_env=para_env)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      DO ispin = 1, SIZE(mos)
         IF (ispin == 1) THEN
            IF (xas_env%homo_occ == 0) THEN
               CALL get_mo_set(mos(ispin)%mo_set, nelectron=nelectron)
               nelectron = nelectron-1
               CALL set_mo_set(mos(ispin)%mo_set, nelectron=nelectron)
            END IF
            CALL set_mo_occupation(mo_set=qs_env%mos(ispin)%mo_set, smear=scf_control%smear, &
                                   xas_env=xas_env)
         ELSE
            CALL set_mo_occupation(mo_set=qs_env%mos(ispin)%mo_set, smear=scf_control%smear)
         END IF
         CALL calculate_density_matrix(mo_set=mos(ispin)%mo_set, &
                                       density_matrix=rho_ao(ispin)%matrix)
      END DO

      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)

      IF (scf_env%mixing_method > 1) THEN
         CALL get_qs_env(qs_env=qs_env, rho_atom_set=rho_atom)
         CALL mixing_init(scf_env%mixing_method, rho, scf_env%mixing_store, &
                          para_env, rho_atom=rho_atom)
      END IF

      CALL timestop(handle)

   END SUBROUTINE xas_initialize_rho

! **************************************************************************************************
!> \brief Find the index of the core orbital that has been excited by XAS
!> \param xas_env ...
!> \param mos ...
!> \param matrix_s ...
!> \par History
!>      03-2010 MI created
!> \author MI
! **************************************************************************************************

   SUBROUTINE find_excited_core_orbital(xas_env, mos, matrix_s)

      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_excited_core_orbital', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ic_max, ir_max, m, n, nao, &
                                                            nexc_search, nmo, xas_estate
      INTEGER, DIMENSION(:), POINTER                     :: col_indices
      REAL(dp)                                           :: a_max, b_max, ip_energy, occ_estate
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues, occupation_numbers
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer, vecbuffer2
      TYPE(cp_fm_type), POINTER                          :: excvec_coeff, excvec_overlap, fm_work, &
                                                            mo_coeff

      NULLIFY (excvec_coeff, excvec_overlap, fm_work, mo_coeff)
      ! Some elements from the xas_env
      CALL get_xas_env(xas_env=xas_env, excvec_coeff=excvec_coeff, &
                       excvec_overlap=excvec_overlap, fm_work=fm_work, nexc_search=nexc_search, &
                       xas_estate=xas_estate, occ_estate=occ_estate)
      IF(.NOT.(ASSOCIATED(excvec_overlap)))CALL cp__a("xas_restart.F",452)

      CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo, &
                      eigenvalues=eigenvalues, occupation_numbers=occupation_numbers)
      ALLOCATE (vecbuffer(1, nao))
      vecbuffer = 0.0_dp
      ALLOCATE (vecbuffer2(1, nexc_search))
      vecbuffer2 = 0.0_dp

      ! ** use the maximum overlap criterion to find the index of the excited orbital
      CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, mo_coeff, fm_work, ncol=nmo)
      CALL cp_gemm("T", "N", 1, xas_env%nexc_search, nao, 1.0_dp, excvec_coeff, &
                   fm_work, 0.0_dp, excvec_overlap, b_first_col=1)
      CALL cp_fm_get_info(matrix=excvec_overlap, col_indices=col_indices, &
                          nrow_global=m, ncol_global=n)
      CALL cp_fm_get_submatrix(excvec_overlap, vecbuffer2, 1, 1, &
                               1, nexc_search, transpose=.FALSE.)

      b_max = 0.0_dp
      ic_max = xas_estate
      DO i = 1, nexc_search
         a_max = ABS(vecbuffer2(1, i))
         IF (a_max > b_max) THEN
            ic_max = i
            b_max = a_max
         ENDIF
      END DO

      IF (ic_max /= xas_estate) THEN
         ir_max = xas_estate
         xas_estate = ic_max
         occupation_numbers(xas_estate) = occ_estate
         occupation_numbers(ir_max) = 1.0_dp
      END IF

      ! Ionization Potential
      iP_energy = eigenvalues(xas_estate)
      CALL set_xas_env(xas_env=xas_env, xas_estate=xas_estate, ip_energy=ip_energy)

      CALL cp_fm_get_submatrix(mo_coeff, vecbuffer, 1, xas_estate, &
                               nao, 1, transpose=.TRUE.)
      CALL cp_fm_set_submatrix(excvec_coeff, vecbuffer, 1, 1, &
                               nao, 1, transpose=.TRUE.)

      DEALLOCATE (vecbuffer, vecbuffer2)

   END SUBROUTINE find_excited_core_orbital

END MODULE xas_restart
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Does all kind of post scf calculations for semi-empirical
!> \par History
!>      Started printing preliminary stuff for MO_CUBES and MO requires some
!>      more work to complete all other functionalities
!> \author Teodoro Laino (07.2008)
! **************************************************************************************************
MODULE qs_scf_post_se
!
   USE ai_moments,                      ONLY: moment
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_sum
   USE moments_utils,                   ONLY: get_reference_point
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: debye
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change
   USE qs_mo_io,                        ONLY: write_mo_set
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_post_se'
   PUBLIC :: scf_post_calculation_se

CONTAINS

! **************************************************************************************************
!> \brief collects possible post - scf calculations and prints info / computes properties.
!>        specific for Semi-empirical calculations
!> \param qs_env the qs_env in which the qs_env lives
!> \par History
!>      07.2008 created [tlaino] - Splitted from qs_scf_post (general)
!> \author tlaino
!> \note
!>      this function changes mo_eigenvectors and mo_eigenvalues, depending on the print keys.
!>      In particular, MO_CUBES causes the MOs to be rotated to make them eigenstates of the KS
!>      matrix, and mo_eigenvalues is updated accordingly. This can, for unconverged wavefunctions,
!>      change afterwards slightly the forces (hence small numerical differences between MD
!>      with and without the debug print level). Ideally this should not happen...
! **************************************************************************************************
   SUBROUTINE scf_post_calculation_se(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_post_calculation_se', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      LOGICAL                                            :: explicit, my_localized_wfn
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: input, print_key, wfn_mix_section

      CALL timeset(routineN, handle)

      ! Writes the data that is already available in qs_env
      CALL write_available_results(qs_env)

      my_localized_wfn = .FALSE.
      NULLIFY (dft_control, mos, rho, &
               subsys, particles, input, print_key, para_env)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_se.F",117)
      ! Here we start with data that needs a postprocessing...
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      rho=rho, &
                      input=input, &
                      subsys=subsys, &
                      para_env=para_env)
      CALL qs_subsys_get(subsys, particles=particles)

      ! Compute Atomic Charges
      CALL qs_scf_post_charges(input, logger, qs_env, rho, para_env)

      ! Moments of charge distribution
      CALL qs_scf_post_moments(input, logger, qs_env)

      ! MO_CUBES
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%MO_CUBES")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",138,"Printing of MO cube files not implemented for Semi-Empirical method.")
      END IF

      ! STM
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%STM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",145,"STM not implemented for Semi-Empirical method.")
      END IF

      ! DFT+U
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%PLUS_U")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",152,"DFT+U not available for Semi-Empirical method.")
      END IF

      ! Kinetic Energy
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%KINETIC_ENERGY")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",159,"Kinetic energy not available for Semi-Empirical method.")
      END IF

      ! Wavefunction mixing
      wfn_mix_section => section_vals_get_subs_vals(input, "DFT%PRINT%WFN_MIX")
      CALL section_vals_get(wfn_mix_section, explicit=explicit)
      IF (explicit .AND. .NOT. qs_env%run_rtp) THEN
         CALL cp__w("qs_scf_post_se.F",166,"Wavefunction mixing not implemented for Semi-Empirical  method.")
      END IF

      ! Print coherent X-ray diffraction spectrum
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%XRAY_DIFFRACTION_SPECTRUM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",173,"XRAY_DIFFRACTION_SPECTRUM  not implemented for Semi-Empirical calculations!!")
      END IF

      ! Calculation of Electric Field Gradients
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%ELECTRIC_FIELD_GRADIENT")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",180,"ELECTRIC_FIELD_GRADIENT not implemented for Semi-Empirical calculations!!")
      END IF

      ! Calculation of EPR Hyperfine Coupling Tensors
      print_key => section_vals_get_subs_vals(section_vals=input, &
                                              subsection_name="DFT%PRINT%HYPERFINE_COUPLING_TENSOR")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",188,"HYPERFINE_COUPLING_TENSOR  not implemented for Semi-Empirical calculations!!")
      END IF

      CALL timestop(handle)

   END SUBROUTINE scf_post_calculation_se

! **************************************************************************************************
!> \brief Computes and prints electric dipole moments
!>        We use the approximation for NDDO from
!>        Pople and Beveridge, Approximate Molecular Orbital Theory,
!>        Mc Graw Hill 1970
!>        mu = \sum_A [ Q_A * R_a + Tr(P_A*D_A) ]
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE qs_scf_post_moments(input, logger, qs_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_moments', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description, dipole_type
      COMPLEX(KIND=dp)                                   :: dzeta, zeta
      COMPLEX(KIND=dp), DIMENSION(3)                     :: dggamma, dzphase, ggamma, zphase
      INTEGER                                            :: i, iat, iatom, ikind, ix, j, nat, natom, &
                                                            natorb, nkind, nspin, reference, &
                                                            unit_nr
      LOGICAL                                            :: do_berry, found
      REAL(KIND=dp) :: charge_tot, ci(3), dci(3), dipole(3), dipole_deriv(3), drcc(3), dria(3), &
         dtheta, gvec(3), q, rcc(3), ria(3), tcharge(2), theta, tmp(3), via(3), zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ncharge
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: mom
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: print_key
      TYPE(semi_empirical_type), POINTER                 :: se_kind

      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%MOMENTS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         ! Dipole Moments
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MOMENTS", &
                                        extension=".data", middle_name="se_dipole", log_filename=.FALSE.)

         ! Reference point
         reference = section_get_ival(print_key, keyword_name="REFERENCE")
         NULLIFY (ref_point)
         description = '[DIPOLE]'
         CALL section_vals_val_get(print_key, "REF_POINT", r_vals=ref_point)
         CALL section_vals_val_get(print_key, "PERIODIC", l_val=do_berry)
         CALL get_reference_point(rcc, drcc, qs_env=qs_env, reference=reference, &
                                  ref_point=ref_point)
         !
         NULLIFY (particle_set)
         CALL get_qs_env(qs_env=qs_env, &
                         rho=rho, &
                         cell=cell, &
                         atomic_kind_set=atomic_kind_set, &
                         natom=natom, &
                         qs_kind_set=qs_kind_set, &
                         particle_set=particle_set, &
                         dft_control=dft_control)

         CALL qs_rho_get(rho, rho_ao=matrix_p)
         nspin = SIZE(matrix_p)
         nkind = SIZE(atomic_kind_set)
         ! net charges
         ALLOCATE (ncharge(natom))
         ncharge = 0.0_dp
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind)
            CALL get_se_param(se_kind, zeff=zeff, natorb=natorb)
            DO iatom = 1, nat
               iat = atomic_kind_set(ikind)%atom_list(iatom)
               tcharge = 0.0_dp
               DO i = 1, nspin
                  CALL cp_dbcsr_get_block_p(matrix=matrix_p(i)%matrix, row=iat, col=iat, &
                                            block=pblock, found=found)
                  IF (found) THEN
                     DO j = 1, natorb
                        tcharge(i) = tcharge(i)+pblock(j, j)
                     END DO
                  END IF
               END DO
               ncharge(iat) = zeff-SUM(tcharge)
            END DO
         END DO
         ! Contributions from net atomic charges
         ! Dipole deriv will be the derivative of the Dipole(dM/dt=\sum e_j v_j)
         dipole_deriv = 0.0_dp
         dipole = 0.0_dp
         IF (do_berry) THEN
            dipole_type = "[BERRY PHASE]"
            rcc = pbc(rcc, cell)
            charge_tot = 0._dp
            charge_tot = SUM(ncharge)
            ria = twopi*MATMUL(cell%h_inv, rcc)
            zphase = CMPLX(COS(ria), SIN(ria), dp)**charge_tot

            dria = twopi*MATMUL(cell%h_inv, drcc)
            dzphase = charge_tot*CMPLX(-SIN(ria), COS(ria), dp)**(charge_tot-1.0_dp)*dria

            ggamma = CMPLX(1.0_dp, 0.0_dp, KIND=dp)
            dggamma = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
            DO ikind = 1, SIZE(atomic_kind_set)
               CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
               DO i = 1, nat
                  iat = atomic_kind_set(ikind)%atom_list(i)
                  ria = particle_set(iat)%r(:)
                  ria = pbc(ria, cell)
                  via = particle_set(iat)%v(:)
                  q = ncharge(iat)
                  DO j = 1, 3
                     gvec = twopi*cell%h_inv(j, :)
                     theta = SUM(ria(:)*gvec(:))
                     dtheta = SUM(via(:)*gvec(:))
                     zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(-q)
                     dzeta = -q*CMPLX(-SIN(theta), COS(theta), KIND=dp)**(-q-1.0_dp)*dtheta
                     dggamma(j) = dggamma(j)*zeta+ggamma(j)*dzeta
                     ggamma(j) = ggamma(j)*zeta
                  END DO
               ENDDO
            END DO
            dggamma = dggamma*zphase+ggamma*dzphase
            ggamma = ggamma*zphase
            IF (ALL(REAL(ggamma, KIND=dp) /= 0.0_dp)) THEN
               tmp = AIMAG(ggamma)/REAL(ggamma, KIND=dp)
               ci = ATAN(tmp)
               dci = (1.0_dp/(1.0_dp+tmp**2))* &
                     (AIMAG(dggamma)*REAL(ggamma, KIND=dp)-AIMAG(ggamma)* &
                      REAL(dggamma, KIND=dp))/(REAL(ggamma, KIND=dp))**2
               dipole = MATMUL(cell%hmat, ci)/twopi
               dipole_deriv = MATMUL(cell%hmat, dci)/twopi
            END IF
         ELSE
            dipole_type = "[Non Periodic]"
            DO i = 1, natom
               ! no pbc(particle_set(i)%r(:),cell) so that the total dipole is the sum of the molecular dipoles
               ria = particle_set(i)%r(:)
               q = ncharge(i)
               dipole = dipole-q*(ria-rcc)
               dipole_deriv(:) = dipole_deriv(:)-q*(particle_set(i)%v(:)-drcc)
            END DO
         END IF
         ! Contributions from atomic polarization
         ! No contribution to dipole derivatives
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set)
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind)
            CALL get_se_param(se_kind, natorb=natorb)
            ALLOCATE (mom(natorb, natorb, 3))
            mom = 0.0_dp
            CALL atomic_moments(mom, basis_set)
            DO iatom = 1, nat
               iat = atomic_kind_set(ikind)%atom_list(iatom)
               DO i = 1, nspin
                  CALL cp_dbcsr_get_block_p(matrix=matrix_p(i)%matrix, row=iat, col=iat, &
                                            block=pblock, found=found)
                  IF (found) THEN
                     IF(.NOT.(natorb == SIZE(pblock, 1)))CALL cp__a("qs_scf_post_se.F",360)
                     ix = coset(1, 0, 0)-1
                     dipole(1) = dipole(1)+SUM(pblock*mom(:, :, ix))
                     ix = coset(0, 1, 0)-1
                     dipole(2) = dipole(2)+SUM(pblock*mom(:, :, ix))
                     ix = coset(0, 0, 1)-1
                     dipole(3) = dipole(3)+SUM(pblock*mom(:, :, ix))
                  END IF
               END DO
            END DO
            DEALLOCATE (mom)
         END DO
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//"(A.U.)|", dipole
            WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//"(Debye)|", dipole*debye
            WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//" DERIVATIVE(A.U.)|", dipole_deriv
         END IF
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
      END IF

   END SUBROUTINE qs_scf_post_moments

! **************************************************************************************************
!> \brief Computes the dipole integrals for an atom (a|x|b), a,b on atom A
!> \param mom ...
!> \param basis_set ...
! **************************************************************************************************
   SUBROUTINE atomic_moments(mom, basis_set)
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: mom
      TYPE(gto_basis_set_type), POINTER                  :: basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'atomic_moments', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iset, jset, ncoa, ncob, nm, nset, &
                                                            sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgf, nsgf
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: mab
      REAL(KIND=dp), DIMENSION(3)                        :: rac, rbc
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgf, sphi, zet

      rac = 0.0_dp
      rbc = 0.0_dp

      first_sgf => basis_set%first_sgf
      la_max => basis_set%lmax
      la_min => basis_set%lmin
      npgf => basis_set%npgf
      nset = basis_set%nset
      nsgf => basis_set%nsgf_set
      rpgf => basis_set%pgf_radius
      sphi => basis_set%sphi
      zet => basis_set%zet

      nm = 0
      DO iset = 1, nset
         ncoa = npgf(iset)*ncoset(la_max(iset))
         nm = MAX(nm, ncoa)
      END DO
      ALLOCATE (mab(nm, nm, 4), work(nm, nm))

      DO iset = 1, nset
         ncoa = npgf(iset)*ncoset(la_max(iset))
         sgfa = first_sgf(1, iset)
         DO jset = 1, nset
            ncob = npgf(jset)*ncoset(la_max(jset))
            sgfb = first_sgf(1, jset)
            !*** Calculate the primitive integrals ***
            CALL moment(la_max(iset), npgf(iset), zet(:, iset), rpgf(:, iset), la_min(iset), &
                        la_max(jset), npgf(jset), zet(:, jset), rpgf(:, jset), 1, rac, rbc, mab)
            !*** Contraction step ***
            DO i = 1, 3
               CALL dgemm("N", "N", ncoa, nsgf(jset), ncob, 1.0_dp, mab(1, 1, i), SIZE(mab, 1), &
                          sphi(1, sgfb), SIZE(sphi, 1), 0.0_dp, work(1, 1), SIZE(work, 1))
               CALL dgemm("T", "N", nsgf(iset), nsgf(jset), ncoa, 1.0_dp, sphi(1, sgfa), SIZE(sphi, 1), &
                          work(1, 1), SIZE(work, 1), 1.0_dp, mom(sgfa, sgfb, i), SIZE(mom, 1))
            END DO
         END DO
      END DO
      DEALLOCATE (mab, work)

   END SUBROUTINE atomic_moments
! **************************************************************************************************
!> \brief Computes and Prints Atomic Charges with several methods
!> \param input ...
!> \param logger ...
!> \param qs_env the qs_env in which the qs_env lives
!> \param rho ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_post_charges(input, logger, qs_env, rho, &
                                  para_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_post_charges', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: ana
      CHARACTER(LEN=default_string_length)               :: aname
      INTEGER                                            :: i, iat, iatom, ikind, j, nat, natom, &
                                                            natorb, nkind, nspin, unit_nr
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mcharge
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: charges
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: print_key
      TYPE(semi_empirical_type), POINTER                 :: se_kind

      NULLIFY (particle_set)
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      natom=natom, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      dft_control=dft_control)

      ! Compute the mulliken charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%MULLIKEN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, input, "DFT%PRINT%MULLIKEN", extension=".mulliken", &
                                        middle_name="", log_filename=.FALSE.)
         CALL qs_rho_get(rho, rho_ao=matrix_p)
         nspin = SIZE(matrix_p)
         ALLOCATE (charges(natom, nspin), mcharge(natom))
         charges = 0.0_dp
         mcharge = 0.0_dp
         ! calculate atomic charges
         nkind = SIZE(atomic_kind_set)
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind)
            CALL get_se_param(se_kind, zeff=zeff, natorb=natorb)
            DO iatom = 1, nat
               iat = atomic_kind_set(ikind)%atom_list(iatom)
               DO i = 1, nspin
                  CALL cp_dbcsr_get_block_p(matrix=matrix_p(i)%matrix, row=iat, col=iat, &
                                            block=pblock, found=found)
                  IF (found) THEN
                     DO j = 1, natorb
                        charges(iat, i) = charges(iat, i)+pblock(j, j)
                     END DO
                  END IF
               END DO
               mcharge(iat) = zeff-SUM(charges(iat, 1:nspin))
            END DO
         END DO
         !
         CALL mp_sum(charges, para_env%group)
         CALL mp_sum(mcharge, para_env%group)
         !
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(/,/,T2,A)") "POPULATION ANALYSIS"
            IF (nspin == 1) THEN
               WRITE (UNIT=unit_nr, FMT="(/,T2,A,T70,A)") &
                  " # Atom   Element   Kind        Atomic population", " Net charge"
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
                  CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind)
                  CALL get_se_param(se_kind, name=aname)
                  ana = ADJUSTR(TRIM(ADJUSTL(aname)))
                  DO iatom = 1, nat
                     iat = atomic_kind_set(ikind)%atom_list(iatom)
                     WRITE (UNIT=unit_nr, &
                            FMT="(T2,I7,6X,A2,3X,I6,T39,F12.6,T69,F12.6)") &
                        iat, ana, ikind, charges(iat, 1), mcharge(iat)
                  END DO
               END DO
               WRITE (UNIT=unit_nr, &
                      FMT="(T2,A,T39,F12.6,T69,F12.6,/)") &
                  "# Total charge", SUM(charges(:, 1)), SUM(mcharge(:))
            ELSE
               WRITE (UNIT=unit_nr, FMT="(/,T2,A)") &
                  "# Atom  Element  Kind  Atomic population (alpha,beta)   Net charge  Spin moment"
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
                  CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind)
                  CALL get_se_param(se_kind, name=aname)
                  ana = ADJUSTR(TRIM(ADJUSTL(aname)))
                  DO iatom = 1, nat
                     iat = atomic_kind_set(ikind)%atom_list(iatom)
                     WRITE (UNIT=unit_nr, &
                            FMT="(T2,I6,5X,A2,2X,I6,T29,4(1X,F12.6))") &
                        iat, ana, ikind, charges(iat, 1:2), mcharge(iat), charges(iat, 1)-charges(iat, 2)
                  END DO
               END DO
               WRITE (UNIT=unit_nr, &
                      FMT="(T2,A,T29,4(1X,F12.6),/)") &
                  "# Total charge and spin", SUM(charges(:, 1)), SUM(charges(:, 2)), SUM(mcharge(:))
            END IF
         END IF

         CALL cp_print_key_finished_output(unit_nr, logger, input, "DFT%PRINT%MULLIKEN")

         DEALLOCATE (charges, mcharge)
      END IF

      ! Compute the Lowdin charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%LOWDIN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",570,"Lowdin charges not available for semi-empirical calculations!")
      END IF

      ! Hirshfeld charges
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%HIRSHFELD")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",576,"Hirshfeld charges not available for semi-empirical calculations!")
      END IF

      ! MAO
      print_key => section_vals_get_subs_vals(input, "DFT%PRINT%MAO_ANALYSIS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",582,"MAO analysis not available for semi-empirical calculations!")
      END IF

   END SUBROUTINE qs_scf_post_charges

! **************************************************************************************************
!> \brief Write QS results always available (if switched on through the print_keys)
!> \param qs_env the qs_env in which the qs_env lives
! **************************************************************************************************
   SUBROUTINE write_available_results(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_available_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: after, handle, ispin, iw, output_unit
      LOGICAL                                            :: omit_headers
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_rmpv, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: dft_section, input

      CALL timeset(routineN, handle)
      NULLIFY (cell, dft_control, mos, atomic_kind_set, particle_set, rho, &
               ks_rmpv, dft_section, input, &
               particles, subsys, para_env, rho_ao)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_se.F",620)
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      rho=rho, &
                      matrix_ks=ks_rmpv, &
                      input=input, &
                      cell=cell, &
                      subsys=subsys, &
                      para_env=para_env)
      CALL qs_subsys_get(subsys, particles=particles)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      ! *** if the dft_section tells you to do so, write last wavefunction to screen
      dft_section => section_vals_get_subs_vals(input, "DFT")
      IF (dft_control%nspins == 2) THEN
         CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, spin="ALPHA", last=.TRUE.)
         CALL write_mo_set(mos(2)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, spin="BETA", last=.TRUE.)
      ELSE
         CALL write_mo_set(mos(1)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                           dft_section, last=.TRUE.)
      END IF

      ! *** at the end of scf print out the projected dos per kind
      IF (BTEST(cp_print_key_should_output(logger%iter_info, dft_section, "PRINT%PDOS") &
                , cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",651,"PDOS not implemented for Semi-Empirical calculations!!")
      ENDIF

      ! Print the total density (electronic + core charge)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%TOT_DENSITY_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",657,"TOT_DENSITY_CUBE  not implemented for Semi-Empirical calculations!!")
      END IF

      ! Write cube file with electron density
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%E_DENSITY_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",663,"E_DENSITY_CUBE not implemented for Semi-Empirical calculations!!")
      END IF ! print key

      ! Write cube file with EFIELD
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%EFIELD_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",669,"EFIELD_CUBE not implemented for Semi-Empirical calculations!!")
      END IF ! print key

      ! Write cube file with ELF
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%ELF_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",675,"ELF function not implemented for Semi-Empirical calculations!!")
      END IF ! print key

      ! Print the hartree potential
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%V_HARTREE_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",681,"V_HARTREE_CUBE not implemented for Semi-Empirical calculations!!")
      ENDIF

      ! Print the XC potential
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%V_XC_CUBE"), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_se.F",687,"V_XC_CUBE not available for Semi-Empirical calculations!!")
      ENDIF

      ! Write the density matrix
      CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%AO_MATRICES/DENSITY"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, input, "DFT%PRINT%AO_MATRICES/DENSITY", &
                                   extension=".Log")
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_write_sparse_matrix(rho_ao(ispin)%matrix, 4, after, qs_env, &
                                              para_env, output_unit=iw, omit_headers=omit_headers)
         END DO
         CALL cp_print_key_finished_output(iw, logger, input, &
                                           "DFT%PRINT%AO_MATRICES/DENSITY")
      END IF

      ! The Kohn-Sham matrix itself
      IF (BTEST(cp_print_key_should_output(logger%iter_info, input, &
                                           "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX"), cp_p_file)) THEN
         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.FALSE.)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
         iw = cp_print_key_unit_nr(logger, input, "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX", &
                                   extension=".Log")
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL cp_dbcsr_write_sparse_matrix(ks_rmpv(1)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_print_key_finished_output(iw, logger, input, &
                                           "DFT%PRINT%AO_MATRICES/KOHN_SHAM_MATRIX")
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_available_results

END MODULE qs_scf_post_se
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Does all kind of post scf calculations for DFTB
!> \par History
!>      Started as a copy from the GPW file
!> \author JHU (03.2013)
! **************************************************************************************************
MODULE qs_scf_post_dftb
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_to_fm_submat,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: twopi
   USE moments_utils,                   ONLY: get_reference_point
   USE mulliken,                        ONLY: mulliken_charges
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: debye
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE qs_scf_wfn_mix,                  ONLY: wfn_mix
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_post_dftb'
   PUBLIC :: scf_post_calculation_dftb

CONTAINS

! **************************************************************************************************
!> \brief collects possible post - scf calculations and prints info / computes properties.
!> \param dft_section ...
!> \param scf_env ...
!> \param qs_env ...
!> \par History
!>      03.2013 copy of scf_post_gpw
!> \author JHU
!> \note
! **************************************************************************************************
   SUBROUTINE scf_post_calculation_dftb(dft_section, scf_env, qs_env)

      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_post_calculation_dftb', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: ana
      CHARACTER(LEN=default_string_length)               :: aname
      INTEGER                                            :: handle, iat, iatom, ikind, nat, natom, &
                                                            nkind, nspins, unit_nr
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mcharge
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p, matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_kind
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: moments_section, print_key, &
                                                            print_section, wfn_mix_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_dftb.F",118)
      NULLIFY (dft_control, rho, para_env, matrix_s, matrix_p)
      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, rho=rho, natom=natom, para_env=para_env, &
                      matrix_s_kp=matrix_s)
      nspins = dft_control%nspins
      CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
      ! Mulliken charges
      ALLOCATE (charges(natom, nspins), mcharge(natom))
      !
      CALL mulliken_charges(matrix_p, matrix_s, para_env, charges)
      !
      nkind = SIZE(atomic_kind_set)
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
         CALL get_dftb_atom_param(dftb_kind, zeff=zeff)
         DO iatom = 1, nat
            iat = atomic_kind_set(ikind)%atom_list(iatom)
            mcharge(iat) = zeff-SUM(charges(iat, 1:nspins))
         END DO
      END DO

      print_section => section_vals_get_subs_vals(dft_section, "PRINT")

      ! Mulliken
      print_key => section_vals_get_subs_vals(print_section, "MULLIKEN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, print_section, "MULLIKEN", extension=".mulliken", &
                                        middle_name="", log_filename=.FALSE.)
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(/,/,T2,A)") "MULLIKEN POPULATION ANALYSIS"
            IF (nspins == 1) THEN
               WRITE (UNIT=unit_nr, FMT="(/,T2,A,T70,A)") &
                  " # Atom   Element   Kind        Atomic population", " Net charge"
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
                  CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
                  CALL get_dftb_atom_param(dftb_kind, name=aname)
                  ana = ADJUSTR(TRIM(ADJUSTL(aname)))
                  DO iatom = 1, nat
                     iat = atomic_kind_set(ikind)%atom_list(iatom)
                     WRITE (UNIT=unit_nr, &
                            FMT="(T2,I7,6X,A2,3X,I6,T39,F12.6,T69,F12.6)") &
                        iat, ana, ikind, charges(iat, 1), mcharge(iat)
                  END DO
               END DO
               WRITE (UNIT=unit_nr, &
                      FMT="(T2,A,T39,F12.6,T69,F12.6,/)") &
                  "# Total charge", SUM(charges(:, 1)), SUM(mcharge(:))
            ELSE
               WRITE (UNIT=unit_nr, FMT="(/,T2,A)") &
                  "# Atom  Element  Kind  Atomic population (alpha,beta)   Net charge  Spin moment"
               DO ikind = 1, nkind
                  CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
                  CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
                  CALL get_dftb_atom_param(dftb_kind, name=aname)
                  ana = ADJUSTR(TRIM(ADJUSTL(aname)))
                  DO iatom = 1, nat
                     iat = atomic_kind_set(ikind)%atom_list(iatom)
                     WRITE (UNIT=unit_nr, &
                            FMT="(T2,I6,5X,A2,2X,I6,T29,4(1X,F12.6))") &
                        iat, ana, ikind, charges(iat, 1:2), mcharge(iat), charges(iat, 1)-charges(iat, 2)
                  END DO
               END DO
               WRITE (UNIT=unit_nr, &
                      FMT="(T2,A,T29,4(1X,F12.6),/)") &
                  "# Total charge and spin", SUM(charges(:, 1)), SUM(charges(:, 2)), SUM(mcharge(:))
            END IF
            CALL m_flush(unit_nr)
         END IF
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
      END IF

      ! Lowdin
      print_key => section_vals_get_subs_vals(print_section, "LOWDIN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",195,"Lowdin population analysis not implemented for DFTB method.")
      END IF

      ! Hirshfeld
      print_key => section_vals_get_subs_vals(print_section, "HIRSHFELD")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",201,"Hirshfeld charges not available for DFTB method.")
      END IF

      ! MAO
      print_key => section_vals_get_subs_vals(print_section, "MAO_ANALYSIS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",207,"MAO analysis not available for DFTB method.")
      END IF

      ! Dipole Moments
      print_key => section_vals_get_subs_vals(print_section, "MOMENTS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, print_section, "MOMENTS", &
                                        extension=".data", middle_name="dftb_dipole", log_filename=.FALSE.)
         moments_section => section_vals_get_subs_vals(print_section, "MOMENTS")
         CALL dftb_dipole(qs_env, moments_section, unit_nr, mcharge)
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
      END IF

      DEALLOCATE (charges, mcharge)

      ! EPR Hyperfine Coupling
      print_key => section_vals_get_subs_vals(print_section, "HYPERFINE_COUPLING_TENSOR")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",225,"Hyperfine Coupling not implemented for DFTB method.")
      END IF

      ! EFIELD CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "EFIELD_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",231,"Efield cube file not implemented for DFTB method.")
      END IF

      ! DENSITY CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "E_DENSITY_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",237,"Electronic density cube file not implemented for DFTB method.")
      END IF

      ! TOTAL DENSITY CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "TOT_DENSITY_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",243,"Total density cube file not implemented for DFTB method.")
      END IF

      ! V_Hartree CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "V_HARTREE_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",249,"Hartree potential cube file not implemented for DFTB method.")
      END IF

      ! V_XC CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "V_XC_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",255,"XC potential cube file not available for DFTB method.")
      END IF

      ! ELF
      print_key => section_vals_get_subs_vals(print_section, "ELF_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",261,"ELF not implemented for DFTB method.")
      END IF

      ! KINETIC ENERGY
      print_key => section_vals_get_subs_vals(print_section, "KINETIC_ENERGY")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",267,"Kinetic energy not available for DFTB method.")
      END IF

      ! Xray diffraction spectrum
      print_key => section_vals_get_subs_vals(print_section, "XRAY_DIFFRACTION_SPECTRUM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",273,"Xray diffraction spectrum not implemented for DFTB method.")
      END IF

      ! Electric field gradients
      print_key => section_vals_get_subs_vals(print_section, "ELECTRIC_FIELD_GRADIENT")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",279,"Electric field gradient not implemented for DFTB method.")
      END IF

      ! STM
      print_key => section_vals_get_subs_vals(print_section, "STM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",285,"STM not implemented for DFTB method.")
      END IF

      ! MO
      print_key => section_vals_get_subs_vals(print_section, "MO")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",291,"Printing of MO properties not implemented for DFTB method.")
      END IF

      ! MO CUBES
      print_key => section_vals_get_subs_vals(print_section, "MO_CUBES")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",297,"Printing of MO cube files not implemented for DFTB method.")
      END IF

      ! Wavefunction mixing
      wfn_mix_section => section_vals_get_subs_vals(dft_section, "PRINT%WFN_MIX")
      CALL section_vals_get(wfn_mix_section, explicit=explicit)
      IF (explicit .AND. .NOT. qs_env%run_rtp) CALL wfn_mix_dftb(qs_env, dft_section, scf_env)

      ! PLUS_U
      print_key => section_vals_get_subs_vals(print_section, "PLUS_U")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",308,"DFT+U method not implemented for DFTB method.")
      END IF

      ! PDOS
      print_key => section_vals_get_subs_vals(print_section, "PDOS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_dftb.F",314,"Projected DOS not implemented for DFTB method.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE scf_post_calculation_dftb

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param input ...
!> \param unit_nr ...
!> \param charges ...
! **************************************************************************************************
   SUBROUTINE dftb_dipole(qs_env, input, unit_nr, charges)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: input
      INTEGER, INTENT(in)                                :: unit_nr
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'dftb_dipole', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description, dipole_type
      COMPLEX(KIND=dp)                                   :: dzeta, dzphase(3), zeta, zphase(3)
      COMPLEX(KIND=dp), DIMENSION(3)                     :: dggamma, ggamma
      INTEGER                                            :: i, iat, ikind, j, nat, reference
      LOGICAL                                            :: do_berry
      REAL(KIND=dp) :: charge_tot, ci(3), dci(3), dipole(3), dipole_deriv(3), drcc(3), dria(3), &
         dtheta, gvec(3), q, rcc(3), ria(3), theta, tmp(3), via(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      NULLIFY (atomic_kind_set, cell)
      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, cell=cell)

      ! Reference point
      reference = section_get_ival(input, keyword_name="REFERENCE")
      NULLIFY (ref_point)
      description = '[DIPOLE]'
      CALL section_vals_val_get(input, "REF_POINT", r_vals=ref_point)
      CALL section_vals_val_get(input, "PERIODIC", l_val=do_berry)

      CALL get_reference_point(rcc, drcc, qs_env=qs_env, reference=reference, ref_point=ref_point)

      ! Dipole deriv will be the derivative of the Dipole(dM/dt=\sum e_j v_j)
      dipole_deriv = 0.0_dp
      dipole = 0.0_dp
      IF (do_berry) THEN
         dipole_type = "[BERRY PHASE]"
         rcc = pbc(rcc, cell)
         charge_tot = 0._dp
         charge_tot = SUM(charges)
         ria = twopi*MATMUL(cell%h_inv, rcc)
         zphase = CMPLX(COS(ria), SIN(ria), dp)**charge_tot

         dria = twopi*MATMUL(cell%h_inv, drcc)
         dzphase = charge_tot*CMPLX(-SIN(ria), COS(ria), dp)**(charge_tot-1.0_dp)*dria

         ggamma = CMPLX(1.0_dp, 0.0_dp, KIND=dp)
         dggamma = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO ikind = 1, SIZE(atomic_kind_set)
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
            DO i = 1, nat
               iat = atomic_kind_set(ikind)%atom_list(i)
               ria = particle_set(iat)%r(:)
               ria = pbc(ria, cell)
               via = particle_set(iat)%v(:)
               q = charges(iat)
               DO j = 1, 3
                  gvec = twopi*cell%h_inv(j, :)
                  theta = SUM(ria(:)*gvec(:))
                  dtheta = SUM(via(:)*gvec(:))
                  zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(-q)
                  dzeta = -q*CMPLX(-SIN(theta), COS(theta), KIND=dp)**(-q-1.0_dp)*dtheta
                  dggamma(j) = dggamma(j)*zeta+ggamma(j)*dzeta
                  ggamma(j) = ggamma(j)*zeta
               END DO
            ENDDO
         END DO
         dggamma = dggamma*zphase+ggamma*dzphase
         ggamma = ggamma*zphase
         IF (ALL(REAL(ggamma, KIND=dp) /= 0.0_dp)) THEN
            tmp = AIMAG(ggamma)/REAL(ggamma, KIND=dp)
            ci = ATAN(tmp)
            dci = (1.0_dp/(1.0_dp+tmp**2))* &
                  (AIMAG(dggamma)*REAL(ggamma, KIND=dp)-AIMAG(ggamma)*REAL(dggamma, KIND=dp))/(REAL(ggamma, KIND=dp))**2
            dipole = MATMUL(cell%hmat, ci)/twopi
            dipole_deriv = MATMUL(cell%hmat, dci)/twopi
         END IF
      ELSE
         dipole_type = "[Non Periodic]"
         DO i = 1, SIZE(particle_set)
            ! no pbc(particle_set(i)%r(:),cell) so that the total dipole is the sum of the molecular dipoles
            ria = particle_set(i)%r(:)
            q = charges(i)
            dipole = dipole-q*(ria-rcc)
            dipole_deriv(:) = dipole_deriv(:)-q*(particle_set(i)%v(:)-drcc)
         END DO
      END IF
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//"(A.U.)|", dipole
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//"(Debye)|", dipole*debye
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "DIPOLE "//TRIM(dipole_type)//" DERIVATIVE(A.U.)|", dipole_deriv
      END IF

   END SUBROUTINE dftb_dipole

! **************************************************************************************************
!> \brief computes the MOs and calls the wavefunction mixing routine.
!> \param qs_env ...
!> \param dft_section ...
!> \param scf_env ...
!> \author Florian Schiffmann
!> \note
! **************************************************************************************************

   SUBROUTINE wfn_mix_dftb(qs_env, dft_section, scf_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'wfn_mix_dftb', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nao, nmo, output_unit
      REAL(dp), DIMENSION(:), POINTER                    :: mo_eigenvalues
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: lumos
      TYPE(cp_fm_struct_type), POINTER                   :: ao_ao_fmstruct, ao_lumo_struct
      TYPE(cp_fm_type), POINTER                          :: KS_tmp, mo_coeff, MO_tmp, S_tmp, work
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: wfn_mix_section

      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s, matrix_ks=matrix_ks, &
                      particle_set=particle_set, atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, mos=mos)

      wfn_mix_section => section_vals_get_subs_vals(dft_section, "PRINT%WFN_MIX")

      CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nao=nao)

      CALL cp_fm_struct_create(fmstruct=ao_ao_fmstruct, nrow_global=nao, ncol_global=nao, &
                               template_fmstruct=mo_coeff%matrix_struct)
      CALL cp_fm_create(S_tmp, matrix_struct=ao_ao_fmstruct)
      CALL cp_fm_create(KS_tmp, matrix_struct=ao_ao_fmstruct)
      CALL cp_fm_create(MO_tmp, matrix_struct=ao_ao_fmstruct)
      CALL cp_fm_create(work, matrix_struct=ao_ao_fmstruct)
      ALLOCATE (lumos(SIZE(mos)))

      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, S_tmp)
      CALL cp_fm_cholesky_decompose(S_tmp)

      DO ispin = 1, SIZE(mos)
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, eigenvalues=mo_eigenvalues, nmo=nmo)
         CALL cp_fm_struct_create(fmstruct=ao_lumo_struct, nrow_global=nao, ncol_global=nao-nmo, &
                                  template_fmstruct=mo_coeff%matrix_struct)

         CALL cp_fm_create(lumos(ispin)%matrix, matrix_struct=ao_lumo_struct)
         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, KS_tmp)
         CALL cp_fm_cholesky_reduce(KS_tmp, S_tmp)
         CALL choose_eigv_solver(KS_tmp, work, mo_eigenvalues)
         CALL cp_fm_cholesky_restore(work, nao, S_tmp, MO_tmp, "SOLVE")
         CALL cp_fm_to_fm_submat(MO_tmp, mo_coeff, nao, nmo, 1, 1, 1, 1)
         CALL cp_fm_to_fm_submat(MO_tmp, lumos(ispin)%matrix, nao, nao-nmo, 1, nmo+1, 1, 1)

         CALL cp_fm_struct_release(ao_lumo_struct)
      END DO

      output_unit = cp_logger_get_default_io_unit(logger)
      CALL wfn_mix(mos, particle_set, dft_section, qs_kind_set, &
                   lumos, scf_env, matrix_s, output_unit)

      DO ispin = 1, SIZE(mos)
         CALL cp_fm_release(lumos(ispin)%matrix)
      END DO
      DEALLOCATE (lumos)
      CALL cp_fm_release(S_tmp)
      CALL cp_fm_release(MO_tmp)
      CALL cp_fm_release(KS_tmp)
      CALL cp_fm_release(work)
      CALL cp_fm_struct_release(ao_ao_fmstruct)

   END SUBROUTINE wfn_mix_dftb

! **************************************************************************************************

END MODULE qs_scf_post_dftb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Does all kind of post scf calculations for SCPTB
!> \par History
!>      Started as a copy from the relevant part of qs_scf_post_gpw
!> \author Ralph Koitz (09.2013)
! **************************************************************************************************
MODULE qs_scf_post_scptb

   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE population_analyses,             ONLY: mulliken_population_analysis
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   ! Global parameters
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_post_scptb'
   PUBLIC :: scf_post_calculation_scptb

CONTAINS

! **************************************************************************************************
!> \brief collects possible post - scf calculations and prints info / computes properties.
!> \param qs_env the qs_env in which the qs_env lives
!> \par History
!>        09.2013  Adapted from qs_scf_post_gpw.F
!> \author  Ralph Koitz
!> \note
! **************************************************************************************************
   SUBROUTINE scf_post_calculation_scptb(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_post_calculation_scptb', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      CALL write_scptb_mo_free_results(qs_env)

      CALL timestop(handle)
   END SUBROUTINE scf_post_calculation_scptb

! **************************************************************************************************
!> \brief Write QS results always available in SCPTB calculation
!> \param qs_env the qs_env in which the qs_env lives
!> \par History 09.2013 adapted from qs_scf_post_gpw.F. [RK]
! **************************************************************************************************
   SUBROUTINE write_scptb_mo_free_results(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_scptb_mo_free_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: after, handle, ispin, iw, output_unit, &
                                                            print_level, unit_nr
      LOGICAL                                            :: explicit, omit_headers, print_it
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: dft_section, input, print_key, &
                                                            print_section, wfn_mix_section

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, rho, input, para_env, rho_ao)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_post_scptb.F",106)
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, &
                      rho=rho, &
                      input=input, &
                      cell=cell, &
                      subsys=subsys, &
                      para_env=para_env)

      dft_section => section_vals_get_subs_vals(input, "DFT")
      print_section => section_vals_get_subs_vals(dft_section, "PRINT")
      CALL qs_subsys_get(subsys, particles=particles)

      ! Dipole Moments
      print_key => section_vals_get_subs_vals(print_section, "MOMENTS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp_warn(cp__l("qs_scf_post_scptb.F",124), &
                      "Multipole moments from SCPTB calculation currently "// &
                      "not supported")
      END IF

      ! Print the total density (electronic + core charge)
      ! Not implemented so far.
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                           "TOT_DENSITY_CUBE"), cp_p_file)) THEN
         CALL cp_warn(cp__l("qs_scf_post_scptb.F",133), &
                      "Printing of cube files from SCPTB calculation currently "// &
                      "not supported")
      END IF

      ! Write cube file with electron density
      ! Not implemented so far.
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                           "E_DENSITY_CUBE"), cp_p_file)) THEN
         CALL cp_warn(cp__l("qs_scf_post_scptb.F",142), &
                      "Printing of cube files from SCPTB calculation currently "// &
                      "not supported")
      END IF ! print key

      ! Print the hartree potential
      ! Not implemented so far.
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                           "V_HARTREE_CUBE"), cp_p_file)) THEN
         CALL cp_warn(cp__l("qs_scf_post_scptb.F",151), &
                      "Printing of cube files from SCPTB calculation currently "// &
                      "not supported")
      ENDIF

      ! Print the Electrical Field Components
      ! Not implemented so far.
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                           "EFIELD_CUBE"), cp_p_file)) THEN

         CALL cp_warn(cp__l("qs_scf_post_scptb.F",161), &
                      "Printing of cube files from SCPTB calculation currently "// &
                      "not supported")
      END IF

      ! Write the density matrices
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                           "AO_MATRICES/DENSITY"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, input, "AO_MATRICES/DENSITY", &
                                   extension=".Log")
         CALL section_vals_val_get(input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_write_sparse_matrix(rho_ao(ispin)%matrix, 4, after, qs_env, &
                                              para_env, output_unit=iw, omit_headers=omit_headers)
         END DO
         CALL cp_print_key_finished_output(iw, logger, input, "AO_MATRICES/DENSITY")
      END IF

      ! Compute the Mulliken charges
      print_key => section_vals_get_subs_vals(print_section, "MULLIKEN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, print_section, "MULLIKEN", extension=".mulliken", &
                                        middle_name="", log_filename=.FALSE.)
         print_level = 1
         CALL section_vals_val_get(print_key, "PRINT_GOP", l_val=print_it)
         IF (print_it) print_level = 2
         CALL section_vals_val_get(print_key, "PRINT_ALL", l_val=print_it)
         IF (print_it) print_level = 3
         CALL mulliken_population_analysis(qs_env, unit_nr, print_level)
         CALL cp_print_key_finished_output(unit_nr, logger, print_section, "MULLIKEN")
      END IF

      ! Hirshfeld charges
      print_key => section_vals_get_subs_vals(print_section, "HIRSHFELD")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",199,"Hirshfeld charges not available for SCPTB method.")
      END IF

      ! MAO
      print_key => section_vals_get_subs_vals(print_section, "MAO_ANALYSIS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",205,"MAO analysis not available for SCPTB method.")
      END IF

      ! KINETIC ENERGY
      print_key => section_vals_get_subs_vals(print_section, "KINETIC_ENERGY")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",211,"Kinetic energy not available for SCPTB method.")
      END IF

      ! Xray diffraction spectrum
      print_key => section_vals_get_subs_vals(print_section, "XRAY_DIFFRACTION_SPECTRUM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",217,"Xray diffraction spectrum not implemented for SCPTB method.")
      END IF

      ! Electric field gradients
      print_key => section_vals_get_subs_vals(print_section, "ELECTRIC_FIELD_GRADIENT")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",223,"Electric field gradient not implemented for SCPTB method.")
      END IF

      ! STM
      print_key => section_vals_get_subs_vals(print_section, "STM")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",229,"STM not implemented for SCPTB method.")
      END IF

      ! MO
      print_key => section_vals_get_subs_vals(print_section, "MO")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",235,"Printing of MO properties not implemented for SCPTB method.")
      END IF

      ! MO CUBES
      print_key => section_vals_get_subs_vals(print_section, "MO_CUBES")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",241,"Printing of MO cube files not implemented for SCPTB method.")
      END IF

      ! Wavefunction mixing
      wfn_mix_section => section_vals_get_subs_vals(dft_section, "PRINT%WFN_MIX")
      CALL section_vals_get(wfn_mix_section, explicit=explicit)
      IF (explicit .AND. .NOT. qs_env%run_rtp) THEN
         CALL cp__w("qs_scf_post_scptb.F",248,"Wavefunction mixing not implemented for SCPTB method.")
      END IF

      ! PLUS_U
      print_key => section_vals_get_subs_vals(print_section, "PLUS_U")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",254,"DFT+U method not implemented for SCPTB method.")
      END IF

      ! PDOS
      print_key => section_vals_get_subs_vals(print_section, "PDOS")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",260,"Projected DOS not implemented for SCPTB method.")
      END IF

      ! V_XC CUBE FILE
      print_key => section_vals_get_subs_vals(print_section, "V_XC_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",266,"XC potential cube file not available for SCPTB method.")
      END IF

      ! ELF
      print_key => section_vals_get_subs_vals(print_section, "ELF_CUBE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",272,"ELF not implemented for SCPTB method.")
      END IF

      ! EPR Hyperfine Coupling
      print_key => section_vals_get_subs_vals(print_section, "HYPERFINE_COUPLING_TENSOR")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",278,"Hyperfine Coupling not implemented for SCPTB method.")
      END IF

      ! Lowdin
      print_key => section_vals_get_subs_vals(print_section, "LOWDIN")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), cp_p_file)) THEN
         CALL cp__w("qs_scf_post_scptb.F",284,"Lowdin population analysis not implemented for SCPTB method.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE write_scptb_mo_free_results

END MODULE qs_scf_post_scptb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines for qs_scf
! **************************************************************************************************
MODULE qs_scf_initialization
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_col_block_sizes, cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_init_p, &
        cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, cp_dbcsr_type, dbcsr_type_no_symmetry, &
        dbcsr_type_real_default
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_m_by_n_from_row_template,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_dbcsr_output,                 ONLY: write_fm_with_basis_info
   USE cp_fm_basic_linalg,              ONLY: cp_fm_triangular_invert
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_diag,                      ONLY: cp_fm_power
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_get_el_struct
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_to_fm_triangular,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        broy_mix, broy_mix_new, cholesky_dbcsr, cholesky_inverse, cholesky_off, &
        diag_block_davidson, diag_block_krylov, diag_filter_matrix, diag_ot, diag_standard, &
        direct_p_mix, kerker_mix, multisec_mix, no_mix, plus_u_lowdin, pulay_mix, &
        wfi_frozen_method_nr, wfi_use_guess_method_nr
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: kpoint_type
   USE lri_optimize_ri_basis,           ONLY: optimize_lri_basis
   USE pw_types,                        ONLY: pw_p_type
   USE qmmm_image_charge,               ONLY: conditional_calc_image_matrix
   USE qs_block_davidson_types,         ONLY: block_davidson_allocate,&
                                              block_davidson_env_create
   USE qs_density_mixing_types,         ONLY: direct_mixing_nr,&
                                              mixing_storage_create,&
                                              mixing_storage_release,&
                                              no_mixing_nr
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_fb_distribution_methods,      ONLY: fb_distribution_build
   USE qs_fb_env_methods,               ONLY: fb_env_build_atomic_halos,&
                                              fb_env_build_rcut_auto,&
                                              fb_env_read_input,&
                                              fb_env_write_info
   USE qs_fb_env_types,                 ONLY: fb_env_create,&
                                              fb_env_has_data
   USE qs_gspace_mixing,                ONLY: mixing_allocate,&
                                              mixing_init
   USE qs_initial_guess,                ONLY: calculate_first_density_matrix
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type,&
                                              set_qs_kind
   USE qs_ks_types,                     ONLY: qs_ks_did_change
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              init_mo_set,&
                                              mo_set_p_type
   USE qs_outer_scf,                    ONLY: outer_loop_extrapolate,&
                                              outer_loop_variables_count
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_methods,                  ONLY: duplicate_rho_type,&
                                              qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_create,&
                                              qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_diagonalization,          ONLY: diag_subspace_allocate
   USE qs_scf_lanczos,                  ONLY: krylov_space_allocate
   USE qs_scf_output,                   ONLY: qs_scf_initial_info
   USE qs_scf_post_dftb,                ONLY: scf_post_calculation_dftb
   USE qs_scf_post_gpw,                 ONLY: scf_post_calculation_gpw
   USE qs_scf_post_scptb,               ONLY: scf_post_calculation_scptb
   USE qs_scf_post_se,                  ONLY: scf_post_calculation_se
   USE qs_scf_types,                    ONLY: &
        block_davidson_diag_method_nr, block_krylov_diag_method_nr, diag_subspace_env_create, &
        filter_matrix_diag_method_nr, general_diag_method_nr, krylov_space_create, &
        ot_diag_method_nr, ot_method_nr, qs_scf_env_type, scf_env_create, special_diag_method_nr
   USE qs_wf_history_methods,           ONLY: reorthogonalize_vectors,&
                                              wfi_extrapolate,&
                                              wfi_get_method_label,&
                                              wfi_update
   USE scf_control_types,               ONLY: scf_control_type
   USE xas_env_types,                   ONLY: xas_environment_type
   USE xas_restart,                     ONLY: xas_initialize_rho
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_initialization'

   PUBLIC:: qs_scf_env_initialize, qs_scf_compute_properties

CONTAINS

! **************************************************************************************************
!> \brief initializes input parameters if needed or restores values from
!>        previous runs to fill scf_env with the values required for scf
!> \param qs_env the qs_environment where to perform the scf procedure
!> \param scf_env ...
!> \param scf_control ...
!> \param scf_section ...
! **************************************************************************************************
   SUBROUTINE qs_scf_env_initialize(qs_env, scf_env, scf_control, scf_section)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), OPTIONAL, POINTER          :: scf_control
      TYPE(section_vals_type), OPTIONAL, POINTER         :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_env_initialize', &
         routineP = moduleN//':'//routineN

      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scf_control_type), POINTER                    :: my_scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, input, my_scf_section

      CALL get_qs_env(qs_env, input=input, rho=rho, &
                      dft_control=dft_control)

      IF (PRESENT(scf_control)) THEN
         my_scf_control => scf_control
      ELSE
         CALL get_qs_env(qs_env, scf_control=my_scf_control)
      ENDIF

      dft_section => section_vals_get_subs_vals(input, "DFT")
      IF (PRESENT(scf_section)) THEN
         my_scf_section => scf_section
      ELSE
         my_scf_section => section_vals_get_subs_vals(dft_section, "SCF")
      END IF

      CALL qs_scf_ensure_scf_env(qs_env, scf_env)

      CALL section_vals_val_get(my_scf_section, "CHOLESKY", i_val=scf_env%cholesky_method)

      CALL qs_scf_ensure_mos(qs_env, scf_env)

      ! set flags for diagonalization
      CALL qs_scf_ensure_diagonalization(scf_env, my_scf_section, qs_env, &
                                         my_scf_control, qs_env%has_unit_metric)
      ! set parameters for mixing/DIIS during scf
      CALL qs_scf_ensure_mixing(my_scf_control, my_scf_section, scf_env, dft_control)

      CALL qs_scf_ensure_work_matrices(qs_env, scf_env)

      CALL qs_scf_ensure_mixing_store(qs_env, scf_env)

      CALL qs_scf_ensure_outer_loop_vars(scf_env, my_scf_control)

      CALL init_scf_run(scf_env, qs_env, my_scf_section, my_scf_control)

   END SUBROUTINE qs_scf_env_initialize

! **************************************************************************************************
!> \brief makes sure scf_env is allocated (might already be from before)
!>        in case it is present the g-space mixing storage is reset
!> \param qs_env ...
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_scf_env(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_ensure_scf_env', &
         routineP = moduleN//':'//routineN

      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (rho_g)

      IF (.NOT. ASSOCIATED(scf_env)) THEN ! i.e. for MD this is associated on the second step (it so seems)
         CALL scf_env_create(scf_env)
!       CALL set_qs_env(qs_env,scf_env=scf_env)
!       CALL scf_env_release(scf_env)
!       CALL get_qs_env(qs_env=qs_env,scf_env=scf_env)
      ELSE
         ! Reallocate mixing store, if the g space grid (cell) has changed
         SELECT CASE (scf_env%mixing_method)
         CASE (kerker_mix, pulay_mix, broy_mix, broy_mix_new, multisec_mix)
            IF (ASSOCIATED(scf_env%mixing_store)) THEN
               ! The current mixing_store data structure does not allow for an unique
               ! grid comparison, but the probability that the 1d lengths of the old and
               ! the new grid are accidentily equal is rather low
               CALL get_qs_env(qs_env, rho=rho)
               CALL qs_rho_get(rho, rho_g=rho_g)
               IF (SIZE(rho_g(1)%pw%pw_grid%gsq) /= SIZE(scf_env%mixing_store%rhoin(1)%cc)) &
                  CALL mixing_storage_release(scf_env%mixing_store)
            END IF
         END SELECT
      END IF

   END SUBROUTINE qs_scf_ensure_scf_env

! **************************************************************************************************
!> \brief performs allocation of outer SCF variables
!> \param scf_env ...
!> \param scf_control ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_outer_loop_vars(scf_env, scf_control)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_ensure_outer_loop_vars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nhistory, nvariables

      IF (scf_control%outer_scf%have_scf) THEN
         nhistory = scf_control%outer_scf%max_scf+1
         nvariables = outer_loop_variables_count(scf_control)
         ALLOCATE (scf_env%outer_scf%variables(nvariables, nhistory))
         ALLOCATE (scf_env%outer_scf%count(nhistory))
         scf_env%outer_scf%count = 0
         ALLOCATE (scf_env%outer_scf%gradient(nvariables, nhistory))
         ALLOCATE (scf_env%outer_scf%energy(nhistory))
      ENDIF

   END SUBROUTINE qs_scf_ensure_outer_loop_vars

! **************************************************************************************************
!> \brief performs allocation of the mixing storage
!> \param qs_env ...
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_mixing_store(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_ensure_mixing_store', &
         routineP = moduleN//':'//routineN

      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (dft_control)
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control)

      IF (scf_env%mixing_method > 0) THEN
         CALL mixing_allocate(qs_env, scf_env%mixing_method, scf_env%p_mix_new, &
                              scf_env%p_delta, dft_control%nspins, &
                              scf_env%mixing_store)
      ELSE
         NULLIFY (scf_env%p_mix_new)
      END IF

   END SUBROUTINE qs_scf_ensure_mixing_store

! **************************************************************************************************
!> \brief Performs allocation of the SCF work matrices
!>        In case of kpoints we probably don't need most of these matrices,
!>        maybe we have to initialize some matrices in the fm_pool in kpoints
!> \param qs_env ...
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_work_matrices(qs_env, scf_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_ensure_work_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, is, nao, nrow_block, nw
      LOGICAL                                            :: do_kpoints
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: ref_matrix
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools
      TYPE(cp_fm_struct_type), POINTER                   :: ao_ao_fmstruct, ao_mo_fmstruct
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      NULLIFY (ao_mo_fm_pools, ao_mo_fmstruct, ao_ao_fmstruct, dft_control, matrix_s, mos)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      matrix_s_kp=matrix_s, &
                      mos=mos, &
                      scf_control=scf_control, &
                      do_kpoints=do_kpoints)
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools)

      ! create an ao_ao parallel matrix structure
      ao_mo_fmstruct => fm_pool_get_el_struct(ao_mo_fm_pools(1)%pool)
      CALL cp_fm_struct_get(ao_mo_fmstruct, nrow_block=nrow_block)
      CALL get_mo_set(mos(1)%mo_set, nao=nao)
      CALL cp_fm_struct_create(fmstruct=ao_ao_fmstruct, &
                               nrow_block=nrow_block, &
                               ncol_block=nrow_block, &
                               nrow_global=nao, &
                               ncol_global=nao, &
                               template_fmstruct=ao_mo_fmstruct)

      IF ((scf_env%method /= ot_method_nr) .AND. &
          (scf_env%method /= block_davidson_diag_method_nr)) THEN
         IF (.NOT. ASSOCIATED(scf_env%scf_work1)) THEN
            nw = dft_control%nspins
            IF (do_kpoints) nw = 4
            ALLOCATE (scf_env%scf_work1(nw))
            DO is = 1, SIZE(scf_env%scf_work1)
               NULLIFY (scf_env%scf_work1(is)%matrix)
               CALL cp_fm_create(scf_env%scf_work1(is)%matrix, &
                                 matrix_struct=ao_ao_fmstruct, &
                                 name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                                 "WORK_MATRIX-1-"//TRIM(ADJUSTL(cp_to_string(is))))
            END DO
         END IF
         IF ((.NOT. ASSOCIATED(scf_env%ortho)) .AND. &
             (scf_env%method /= ot_diag_method_nr) .AND. &
             (scf_env%method /= special_diag_method_nr)) THEN
            ! Initialize fm matrix to store the Cholesky decomposition
            CALL cp_fm_create(scf_env%ortho, &
                              matrix_struct=ao_ao_fmstruct, &
                              name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                              "ORTHO_MATRIX")
            ! Initialize dbcsr matrix to store the Cholesky decomposition
            IF (scf_env%cholesky_method == cholesky_dbcsr) THEN
               ref_matrix => matrix_s(1, 1)%matrix
               CALL cp_dbcsr_init_p(scf_env%ortho_dbcsr)
               CALL cp_dbcsr_create(scf_env%ortho_dbcsr, template=ref_matrix, &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_init_p(scf_env%buf1_dbcsr)
               CALL cp_dbcsr_create(scf_env%buf1_dbcsr, template=ref_matrix, &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_init_p(scf_env%buf2_dbcsr)
               CALL cp_dbcsr_create(scf_env%buf2_dbcsr, template=ref_matrix, &
                                    matrix_type=dbcsr_type_no_symmetry)
            ELSE IF (scf_env%cholesky_method == cholesky_inverse .OR. &
                     (scf_control%level_shift /= 0.0_dp .AND. &
                      scf_env%cholesky_method == cholesky_off)) THEN
               CALL cp_fm_create(scf_env%ortho_m1, &
                                 matrix_struct=ao_ao_fmstruct, &
                                 name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                                 "ORTHO_MATRIX-1")
            END IF
         END IF
         IF (.NOT. ASSOCIATED(scf_env%scf_work2)) THEN
            CALL cp_fm_create(scf_env%scf_work2, &
                              matrix_struct=ao_ao_fmstruct, &
                              name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                              "WORK_MATRIX-2")
         END IF
      END IF

      IF (dft_control%dft_plus_u) THEN
         IF (dft_control%plus_u_method_id == plus_u_lowdin) THEN
            IF (.NOT. ASSOCIATED(scf_env%scf_work2)) THEN
               CALL cp_fm_create(scf_env%scf_work2, &
                                 matrix_struct=ao_ao_fmstruct, &
                                 name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                                 "WORK_MATRIX-2")
            END IF
            IF (.NOT. ASSOCIATED(scf_env%s_half)) THEN
               CALL cp_fm_create(scf_env%s_half, &
                                 matrix_struct=ao_ao_fmstruct, &
                                 name="S**(1/2) MATRIX")
            END IF
         END IF
      END IF

      IF (do_kpoints) THEN
         IF (.NOT. ASSOCIATED(scf_env%scf_work1)) THEN
            nw = 4
            ALLOCATE (scf_env%scf_work1(nw))
            DO is = 1, SIZE(scf_env%scf_work1)
               NULLIFY (scf_env%scf_work1(is)%matrix)
               CALL cp_fm_create(scf_env%scf_work1(is)%matrix, &
                                 matrix_struct=ao_ao_fmstruct, &
                                 name="SCF"//TRIM(ADJUSTL(cp_to_string(scf_env%id_nr)))// &
                                 "WORK_MATRIX-1-"//TRIM(ADJUSTL(cp_to_string(is))))
            END DO
         END IF
      END IF

      CALL cp_fm_struct_release(ao_ao_fmstruct)

      CALL timestop(handle)

   END SUBROUTINE qs_scf_ensure_work_matrices

! **************************************************************************************************
!> \brief performs allocation of the MO matrices
!> \param qs_env ...
!> \param scf_env ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_mos(qs_env, scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_ensure_mos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, ik, ikk, ispin, nmo, nmo_mat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_b
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_derivs_aux_fit
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools, ao_mo_fm_pools_aux_fit
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_aux_fit
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(mo_set_p_type), DIMENSION(:, :), POINTER      :: mos_k
      TYPE(xas_environment_type), POINTER                :: xas_env

      CALL timeset(routineN, handle)

      NULLIFY (ao_mo_fm_pools, dft_control, mos, ao_mo_fm_pools_aux_fit)
      NULLIFY (mo_derivs, mo_coeff, mo_coeff_aux_fit, mo_derivs_aux_fit, xas_env, matrix_s)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_initialization.F",442)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_initialization.F",443)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      matrix_s_kp=matrix_s, &
                      mos_aux_fit=mos_aux_fit, &
                      xas_env=xas_env)
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools)

      nmo_mat = dft_control%nspins
      IF (dft_control%restricted) nmo_mat = 1 ! right now, there might be more mos than needed derivs

!   *** finish initialization of the MOs ***
      IF(.NOT.(ASSOCIATED(mos)))CALL cp__a("qs_scf_initialization.F",457)
      DO ispin = 1, SIZE(mos)
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b)
         IF (.NOT. ASSOCIATED(mo_coeff)) THEN
            CALL init_mo_set(mos(ispin)%mo_set, &
                             fm_pool=ao_mo_fm_pools(ispin)%pool, &
                             name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                             "%mo"//TRIM(ADJUSTL(cp_to_string(ispin))))
         END IF
         IF (.NOT. ASSOCIATED(mo_coeff_b)) THEN
            CALL cp_fm_get_info(mos(ispin)%mo_set%mo_coeff, ncol_global=nmo)
            CALL cp_dbcsr_init_p(mos(ispin)%mo_set%mo_coeff_b)
            CALL cp_dbcsr_m_by_n_from_row_template(mos(ispin)%mo_set%mo_coeff_b, template=matrix_s(1, 1)%matrix, n=nmo, &
                                                   sym=dbcsr_type_no_symmetry)
         ENDIF
      END DO
!   *** get the mo_derivs OK if needed ***

      IF (qs_env%requires_mo_derivs) THEN
         CALL get_qs_env(qs_env, mo_derivs=mo_derivs)
         IF (.NOT. ASSOCIATED(mo_derivs)) THEN
            ALLOCATE (mo_derivs(nmo_mat))
            DO ispin = 1, nmo_mat
               CALL get_mo_set(mos(ispin)%mo_set, mo_coeff_b=mo_coeff_b)
               NULLIFY (mo_derivs(ispin)%matrix)
               CALL cp_dbcsr_init_p(mo_derivs(ispin)%matrix)
               CALL cp_dbcsr_create(mo_derivs(ispin)%matrix, "mo_derivs", &
                                    cp_dbcsr_distribution(mo_coeff_b), dbcsr_type_no_symmetry, &
                                    cp_dbcsr_row_block_sizes(mo_coeff_b), cp_dbcsr_col_block_sizes(mo_coeff_b), &
                                    nze=0, data_type=dbcsr_type_real_default)
            ENDDO
            CALL set_qs_env(qs_env, mo_derivs=mo_derivs)
         ENDIF

      ELSE
         ! nothing should be done
      ENDIF

!   *** finish initialization of the MOs for ADMM and derivs if needed ***
      IF (dft_control%do_admm) THEN
         IF (dft_control%restricted) &
            CALL cp__b("qs_scf_initialization.F",498,"ROKS with ADMM is not implemented")
         IF(.NOT.(ASSOCIATED(mos_aux_fit)))CALL cp__a("qs_scf_initialization.F",499)
         CALL mpools_get(qs_env%mpools_aux_fit, ao_mo_fm_pools=ao_mo_fm_pools_aux_fit)
         DO ispin = 1, SIZE(mos_aux_fit)
            CALL get_mo_set(mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit)
            IF (.NOT. ASSOCIATED(mo_coeff_aux_fit)) THEN
               CALL init_mo_set(mos_aux_fit(ispin)%mo_set, &
                                fm_pool=ao_mo_fm_pools_aux_fit(ispin)%pool, &
                                name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                                "%mo_aux_fit"//TRIM(ADJUSTL(cp_to_string(ispin))))
            END IF
         END DO
         IF (qs_env%requires_mo_derivs) THEN
            CALL get_qs_env(qs_env, mo_derivs_aux_fit=mo_derivs_aux_fit)
            IF (.NOT. ASSOCIATED(mo_derivs_aux_fit)) THEN
               ALLOCATE (mo_derivs_aux_fit(nmo_mat))
               DO ispin = 1, nmo_mat
                  CALL get_mo_set(mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit)
                  CALL cp_fm_create(mo_derivs_aux_fit(ispin)%matrix, mo_coeff_aux_fit%matrix_struct)
               ENDDO
               CALL set_qs_env(qs_env, mo_derivs_aux_fit=mo_derivs_aux_fit)
            ENDIF
         ELSE
            ! nothing should be done
         ENDIF
      END IF

      ! kpoints: we have to initialize all the k-point MOs
      CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
      IF (kpoints%nkp /= 0) THEN
         ! check for some incompatible options
         IF (qs_env%requires_mo_derivs) &
            CALL cp__b("qs_scf_initialization.F",530,"No MO derivative methods allowed with kpoints")
         IF (dft_control%do_admm) &
            CALL cp__b("qs_scf_initialization.F",532,"No ADMM implemented with kpoints")
         IF (dft_control%do_xas_calculation) &
            CALL cp__b("qs_scf_initialization.F",534,"No XAS implemented with kpoints")
         DO ik = 1, SIZE(kpoints%kp_env)
            CALL mpools_get(kpoints%mpools, ao_mo_fm_pools=ao_mo_fm_pools)
            mos_k => kpoints%kp_env(ik)%kpoint_env%mos
            ikk = kpoints%kp_range(1)+ik-1
            IF(.NOT.(ASSOCIATED(mos_k)))CALL cp__a("qs_scf_initialization.F",539)
            DO ispin = 1, SIZE(mos_k, 2)
               DO ic = 1, SIZE(mos_k, 1)
                  CALL get_mo_set(mos_k(ic, ispin)%mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b)
                  IF (.NOT. ASSOCIATED(mo_coeff)) THEN
                     CALL init_mo_set(mos_k(ic, ispin)%mo_set, &
                                      fm_pool=ao_mo_fm_pools(ispin)%pool, &
                                      name="kpoints_"//TRIM(ADJUSTL(cp_to_string(ikk)))// &
                                      "%mo"//TRIM(ADJUSTL(cp_to_string(ispin))))
                  END IF
                  ! no sparse matrix representation of kpoint MO vectors
                  IF(.NOT.(.NOT. ASSOCIATED(mo_coeff_b)))CALL cp__a("qs_scf_initialization.F",550)
               END DO
            END DO
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_scf_ensure_mos

! **************************************************************************************************
!> \brief sets flag for mixing/DIIS during scf
!> \param scf_control ...
!> \param scf_section ...
!> \param scf_env ...
!> \param dft_control ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_mixing(scf_control, scf_section, scf_env, dft_control)
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: scf_section
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_ensure_mixing', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: mixing_section

      SELECT CASE (scf_control%mixing_method)
      CASE (no_mix)
         scf_env%mixing_method = no_mixing_nr
         scf_env%p_mix_alpha = 1.0_dp
      CASE (direct_p_mix, kerker_mix, pulay_mix, broy_mix, broy_mix_new, multisec_mix)
         scf_env%mixing_method = scf_control%mixing_method
         mixing_section => section_vals_get_subs_vals(scf_section, "MIXING")
         IF (.NOT. ASSOCIATED(scf_env%mixing_store)) &
            CALL mixing_storage_create(scf_env%mixing_store, mixing_section, scf_env%mixing_method, &
                                       dft_control%qs_control%cutoff)
      CASE DEFAULT
         CALL cp__b("qs_scf_initialization.F",589,"Unknown mixing method")
      END SELECT

      ! Disable DIIS for OT and g-space density mixing methods
      IF (scf_env%method == ot_method_nr) THEN
         ! No mixing is used with OT
         scf_env%mixing_method = no_mixing_nr
         scf_env%p_mix_alpha = 1.0_dp
         scf_env%skip_diis = .TRUE.
      END IF

      IF (scf_control%use_diag .AND. scf_env%mixing_method == no_mixing_nr) THEN
         CALL cp__b("qs_scf_initialization.F",601,"Diagonalization procedures without mixing are not recommendable")
      END IF

      IF (scf_env%mixing_method > direct_mixing_nr) THEN
         scf_env%skip_diis = .TRUE.
         scf_env%p_mix_alpha = scf_env%mixing_store%alpha
         IF (scf_env%mixing_store%beta == 0.0_dp) THEN
            CALL cp__b("qs_scf_initialization.F",608,"Mixing employing the Kerker damping factor needs BETA /= 0.0")
         END IF
      END IF

      IF (scf_env%mixing_method == direct_mixing_nr) THEN
         scf_env%p_mix_alpha = scf_env%mixing_store%alpha
         IF (scf_control%eps_diis < scf_control%eps_scf) THEN
            scf_env%skip_diis = .TRUE.
            CALL cp__w("qs_scf_initialization.F",616,"the DIIS scheme is disabled, since EPS_DIIS < EPS_SCF")
         END IF
      END IF

   END SUBROUTINE qs_scf_ensure_mixing

! **************************************************************************************************
!> \brief sets flags for diagonalization and ensure that everything is
!>        allocated
!> \param scf_env ...
!> \param scf_section ...
!> \param qs_env ...
!> \param scf_control ...
!> \param has_unit_metric ...
! **************************************************************************************************
   SUBROUTINE qs_scf_ensure_diagonalization(scf_env, scf_section, qs_env, &
                                            scf_control, has_unit_metric)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(section_vals_type), POINTER                   :: scf_section
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      LOGICAL                                            :: has_unit_metric

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_scf_ensure_diagonalization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nao, nmo
      LOGICAL                                            :: do_kpoints, need_coeff_b, not_se_or_dftb
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CALL get_qs_env(qs_env=qs_env, do_kpoints=do_kpoints, dft_control=dft_control, mos=mos)
      not_se_or_dftb = .NOT. (dft_control%qs_control%dftb .OR. dft_control%qs_control%semi_empirical)
      need_coeff_b = .FALSE.
      scf_env%needs_ortho = .FALSE.

      IF (scf_control%use_diag) THEN
! sanity check whether combinations are allowed
         IF (dft_control%restricted) &
            CALL cp__b("qs_scf_initialization.F",656,"OT only for restricted (ROKS)")
         SELECT CASE (scf_control%diagonalization%method)
         CASE (diag_ot, diag_block_krylov, diag_block_davidson)
            IF (.NOT. not_se_or_dftb) &
               CALL cp__b("qs_scf_initialization.F",660,"DFTB and SE not possible with OT diagonalization")
         END SELECT
         SELECT CASE (scf_control%diagonalization%method)
            ! Diagonalization: additional check whether we are in an orthonormal basis
         CASE (diag_standard)
            scf_env%method = general_diag_method_nr
            scf_env%needs_ortho = (.NOT. has_unit_metric) .AND. (.NOT. do_kpoints)
            IF (has_unit_metric) THEN
               scf_env%method = special_diag_method_nr
            END IF
            ! OT Diagonalization: not possible with ROKS
         CASE (diag_ot)
            IF (dft_control%roks) &
               CALL cp__b("qs_scf_initialization.F",673,"ROKS with OT diagonalization not possible")
            IF (do_kpoints) &
               CALL cp__b("qs_scf_initialization.F",675,"OT diagonalization not possible with kpoint calculations")
            scf_env%method = ot_diag_method_nr
            need_coeff_b = .TRUE.
            ! Block Krylov diagonlization: not possible with ROKS,
            ! allocation of additional matrices is needed
         CASE (diag_block_krylov)
            IF (dft_control%roks) &
               CALL cp__b("qs_scf_initialization.F",682,"ROKS with block PF diagonalization not possible")
            IF (do_kpoints) &
               CALL cp__b("qs_scf_initialization.F",684,"Block Krylov diagonalization not possible with kpoint calculations")
            scf_env%method = block_krylov_diag_method_nr
            scf_env%needs_ortho = .TRUE.
            IF (.NOT. ASSOCIATED(scf_env%krylov_space)) &
               CALL krylov_space_create(scf_env%krylov_space, scf_section)
            CALL krylov_space_allocate(scf_env%krylov_space, scf_control, mos)
            ! Block davidson diagonlization: allocation of additional matrices is needed
         CASE (diag_block_davidson)
            IF (do_kpoints) &
               CALL cp__b("qs_scf_initialization.F",693,"Block Davidson diagonalization not possible with kpoint calculations")
            scf_env%method = block_davidson_diag_method_nr
            IF (.NOT. ASSOCIATED(scf_env%block_davidson_env)) &
               CALL block_davidson_env_create(scf_env%block_davidson_env, dft_control%nspins, &
                                              scf_section)
            DO ispin = 1, dft_control%nspins
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
               CALL block_davidson_allocate(scf_env%block_davidson_env(ispin), mo_coeff, nao, nmo)
            END DO
            need_coeff_b = .TRUE.
            ! Filter matrix diagonalisation method
         CASE (diag_filter_matrix)
            scf_env%method = filter_matrix_diag_method_nr
            IF (.NOT. fb_env_has_data(scf_env%filter_matrix_env)) THEN
               CALL fb_env_create(scf_env%filter_matrix_env)
            END IF
            CALL fb_env_read_input(scf_env%filter_matrix_env, scf_section)
            CALL fb_env_build_rcut_auto(scf_env%filter_matrix_env, qs_env)
            CALL fb_env_write_info(scf_env%filter_matrix_env, qs_env, scf_section)
            CALL fb_distribution_build(scf_env%filter_matrix_env, qs_env, scf_section)
            CALL fb_env_build_atomic_halos(scf_env%filter_matrix_env, qs_env, scf_section)
         CASE DEFAULT
            CALL cp__b("qs_scf_initialization.F",715,"Unknown diagonalization method")
         END SELECT
         ! Check if subspace diagonlization is requested: allocation of additional matrices is needed
         IF (scf_control%do_diag_sub) THEN
            scf_env%needs_ortho = .TRUE.
            IF (.NOT. ASSOCIATED(scf_env%subspace_env)) &
               CALL diag_subspace_env_create(scf_env%subspace_env, scf_section, &
                                             dft_control%qs_control%cutoff)
            CALL diag_subspace_allocate(scf_env%subspace_env, qs_env, mos)
            IF (do_kpoints) &
               CALL cp__b("qs_scf_initialization.F",725,"No subspace diagonlization with kpoint calculation")
         END IF
         ! OT: check if OT is used instead of diagonlization. Not possible with added MOS at the moment
      ELSEIF (scf_control%use_ot) THEN
         scf_env%method = ot_method_nr
         need_coeff_b = .TRUE.
         IF (scf_control%added_mos(1)+scf_control%added_mos(2) > 0) &
            CALL cp__b("qs_scf_initialization.F",732,"OT with ADDED_MOS/=0 not implemented")
         IF (dft_control%restricted .AND. dft_control%nspins .NE. 2) &
            CALL cp__b("qs_scf_initialization.F",734,"nspin must be 2 for restricted (ROKS)")
         IF (do_kpoints) &
            CALL cp__b("qs_scf_initialization.F",736,"OT not possible with kpoint calculations")
      ELSE
         CALL cp__b("qs_scf_initialization.F",738,"OT or DIAGONALIZATION have to be set")
      END IF
      DO ispin = 1, dft_control%nspins
         mos(ispin)%mo_set%use_mo_coeff_b = need_coeff_b
      END DO

   END SUBROUTINE qs_scf_ensure_diagonalization

! **************************************************************************************************
!> \brief computes properties for a given hamilonian using the current wfn
!> \param qs_env ...
!> \param dft_section ...
!> \param wf_type ...
! **************************************************************************************************
   SUBROUTINE qs_scf_compute_properties(qs_env, dft_section, wf_type)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      CHARACTER(6), OPTIONAL                             :: wf_type

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_scf_compute_properties', &
         routineP = moduleN//':'//routineN

      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CALL get_qs_env(qs_env, scf_env=scf_env, dft_control=dft_control)

      IF (dft_control%qs_control%semi_empirical) THEN
         CALL scf_post_calculation_se(qs_env)
      ELSEIF (dft_control%qs_control%dftb) THEN
         CALL scf_post_calculation_dftb(dft_section, scf_env, qs_env)
      ELSEIF (dft_control%qs_control%scptb) THEN
         CALL scf_post_calculation_scptb(qs_env)
      ELSEIF (dft_control%qs_control%do_kg) THEN
         CALL cp__w("qs_scf_initialization.F",772,"No properties from PRINT section available for KG methods")
      ELSEIF (dft_control%qs_control%ofgpw) THEN
         CALL cp__w("qs_scf_initialization.F",774,"No properties from PRINT section available for OFGPW methods")
      ELSEIF (dft_control%qs_control%lri_optbas .AND. dft_control%qs_control%gpw) THEN
         CALL optimize_lri_basis(qs_env)
      ELSE
         IF (PRESENT(wf_type)) THEN
            CALL scf_post_calculation_gpw(dft_section, scf_env, qs_env, wf_type=wf_type)
         ELSE
            CALL scf_post_calculation_gpw(dft_section, scf_env, qs_env)
         ENDIF
      END IF

   END SUBROUTINE qs_scf_compute_properties

! **************************************************************************************************
!> \brief performs those initialisations that need to be done only once
!>       (e.g. that only depend on the atomic positions)
!>       this will be called in scf
!> \param scf_env ...
!> \param qs_env ...
!> \param scf_section ...
!> \param scf_control ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE init_scf_run(scf_env, qs_env, scf_section, scf_control)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_scf_run', routineP = moduleN//':'//routineN

      INTEGER                                            :: after, handle, ikind, iw, nao, ndep, &
                                                            output_unit
      LOGICAL                                            :: dft_plus_u_atom, do_kpoints, &
                                                            init_u_ramping_each_scf, omit_headers, &
                                                            s_minus_half_available
      REAL(KIND=dp)                                      :: u_ramping
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(xas_environment_type), POINTER                :: xas_env

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind_set, matrix_s, dft_control, mos, qs_kind, rho, xas_env)
!
      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_initialization.F",829)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_initialization.F",830)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_initialization.F",831)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_scf_initialization.F",832)
      NULLIFY (para_env)

      s_minus_half_available = .FALSE.
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      qs_kind_set=qs_kind_set, &
                      mos=mos, &
                      rho=rho, &
                      nelectron_total=scf_env%nelectron, &
                      do_kpoints=do_kpoints, &
                      para_env=para_env, &
                      xas_env=xas_env)

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")
      CALL qs_scf_initial_info(output_unit, mos, dft_control)
      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      ! calc ortho matrix
      ndep = 0
      IF (scf_env%needs_ortho) THEN
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, scf_env%ortho)
         IF (scf_env%cholesky_method > cholesky_off) THEN
            CALL cp_fm_cholesky_decompose(scf_env%ortho)
            IF (scf_env%cholesky_method == cholesky_dbcsr) THEN
               CALL cp_fm_triangular_invert(scf_env%ortho)
               CALL cp_fm_set_all(scf_env%scf_work2, 0.0_dp)
               CALL cp_fm_to_fm_triangular(scf_env%ortho, scf_env%scf_work2, "U")
               CALL copy_fm_to_dbcsr(scf_env%scf_work2, scf_env%ortho_dbcsr)
            ELSE IF (scf_env%cholesky_method == cholesky_inverse) THEN
               CALL cp_fm_to_fm(scf_env%ortho, scf_env%ortho_m1)
               CALL cp_fm_triangular_invert(scf_env%ortho_m1)
            END IF
         ELSE
            CALL cp_fm_power(scf_env%ortho, scf_env%scf_work2, -0.5_dp, &
                             scf_control%eps_eigval, ndep)
            IF (scf_control%level_shift /= 0.0_dp) THEN
               CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, scf_env%ortho_m1)
               CALL cp_fm_power(scf_env%ortho_m1, scf_env%scf_work2, 0.5_dp, &
                                scf_control%eps_eigval, ndep)
            END IF
            s_minus_half_available = .TRUE.
         END IF

         IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                              qs_env%input, "DFT%PRINT%AO_MATRICES/ORTHO"), cp_p_file)) THEN
            iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/ORTHO", &
                                      extension=".Log")
            CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
            CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
            after = MIN(MAX(after, 1), 16)
            CALL write_fm_with_basis_info(scf_env%ortho, 4, after, qs_env, &
                                          para_env, output_unit=iw, omit_headers=omit_headers)
            CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                              "DFT%PRINT%AO_MATRICES/ORTHO")
         END IF
      END IF

      CALL get_mo_set(mo_set=mos(1)%mo_set, nao=nao)

      ! DFT+U methods based on Lowdin charges need S^(1/2)
      IF (dft_control%dft_plus_u) THEN
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         IF (dft_control%plus_u_method_id == plus_u_lowdin) THEN
            IF (s_minus_half_available) THEN
               CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, scf_env%ortho, scf_env%s_half, &
                                            nao)
            ELSE
               CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, scf_env%s_half)
               CALL cp_fm_power(scf_env%s_half, scf_env%scf_work2, 0.5_dp, &
                                scf_control%eps_eigval, ndep)
            END IF
         END IF
         DO ikind = 1, SIZE(qs_kind_set)
            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, &
                             dft_plus_u_atom=dft_plus_u_atom, &
                             u_ramping=u_ramping, &
                             init_u_ramping_each_scf=init_u_ramping_each_scf)
            IF (dft_plus_u_atom .AND. (u_ramping /= 0.0_dp)) THEN
               IF (init_u_ramping_each_scf) THEN
                  CALL set_qs_kind(qs_kind=qs_kind, u_minus_j=0.0_dp)
               END IF
            END IF
         END DO
      END IF

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
            "Number of independent orbital functions:", nao-ndep
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      ! extrapolate outer loop variables
      IF (scf_control%outer_scf%have_scf) THEN
         CALL outer_loop_extrapolate(qs_env)
      ENDIF

      ! initializes rho and the mos
      IF (ASSOCIATED(qs_env%xas_env)) THEN
         ! if just optimized wfn, e.g. ground state
         ! changes come from a perturbation, e.g., the occupation numbers
         ! it could be generalized for other cases, at the moment used only for core level spectroscopy
         ! initialize the density with the localized mos
         CALL xas_initialize_rho(qs_env, scf_env, scf_control)
      ELSE
         CALL scf_env_initial_rho_setup(scf_env, qs_env=qs_env, &
                                        scf_section=scf_section, scf_control=scf_control)
      END IF

      ! Frozen density approximation
      IF (ASSOCIATED(qs_env%wf_history)) THEN
         IF (qs_env%wf_history%interpolation_method_nr == wfi_frozen_method_nr) THEN
            IF (.NOT. ASSOCIATED(qs_env%wf_history%past_states(1)%snapshot)) THEN
               CALL wfi_update(qs_env%wf_history, qs_env=qs_env, dt=1.0_dp)
               CALL qs_rho_create(qs_env%wf_history%past_states(1)%snapshot%rho_frozen)
               CALL duplicate_rho_type(rho_input=rho, &
                                       rho_output=qs_env%wf_history%past_states(1)%snapshot%rho_frozen, &
                                       qs_env=qs_env)
            END IF
         END IF
      END IF

      !image charge method, calculate image_matrix if required
      IF (qs_env%qmmm) THEN
         IF (qs_env%qmmm .AND. qs_env%qmmm_env_qm%image_charge) THEN
            CALL conditional_calc_image_matrix(qs_env=qs_env, &
                                               qmmm_env=qs_env%qmmm_env_qm)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE init_scf_run

! **************************************************************************************************
!> \brief Initializes rho and the mos, so that an scf cycle can start
!> \param scf_env the scf env in which to do the scf
!> \param qs_env the qs env the scf_env lives in
!> \param scf_section ...
!> \param scf_control ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_initial_rho_setup(scf_env, qs_env, scf_section, scf_control)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section
      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_initial_rho_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: extrapolation_method_nr, handle, ispin, &
                                                            nmo, output_unit
      LOGICAL                                            :: orthogonal_wf
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom

      CALL timeset(routineN, handle)
      NULLIFY (mo_coeff, rho, dft_control, para_env, mos)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_initialization.F",1006)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_initialization.F",1007)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf_initialization.F",1008)

      CALL get_qs_env(qs_env, &
                      rho=rho, &
                      mos=mos, &
                      dft_control=dft_control, &
                      para_env=para_env)

      extrapolation_method_nr = wfi_use_guess_method_nr
      IF (ASSOCIATED(qs_env%wf_history)) THEN
         CALL wfi_extrapolate(qs_env%wf_history, &
                              qs_env=qs_env, dt=1.0_dp, &
                              extrapolation_method_nr=extrapolation_method_nr, &
                              orthogonal_wf=orthogonal_wf)
         ! wfi_use_guess_method_nr the wavefunctions are not yet initialized
         IF ((.NOT. orthogonal_wf) .AND. &
             (scf_env%method == ot_method_nr) .AND. &
             (.NOT. (extrapolation_method_nr == wfi_use_guess_method_nr))) THEN
            DO ispin = 1, SIZE(mos)
               CALL get_mo_set(mos(ispin)%mo_set, &
                               mo_coeff=mo_coeff, nmo=nmo)
               CALL reorthogonalize_vectors(qs_env, &
                                            v_matrix=mo_coeff, n_col=nmo)
               CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, &
                                      smear=scf_control%smear)
            END DO
         END IF
      END IF
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Extrapolation method: "// &
            TRIM(wfi_get_method_label(extrapolation_method_nr))
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      IF (extrapolation_method_nr == wfi_use_guess_method_nr) THEN
         CALL calculate_first_density_matrix(scf_env=scf_env, qs_env=qs_env)
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      END IF

      ! Some preparation for the mixing
      IF (scf_env%mixing_method > 1) THEN
         IF (dft_control%qs_control%gapw) THEN
            CALL get_qs_env(qs_env=qs_env, &
                            rho_atom_set=rho_atom)
            CALL mixing_init(scf_env%mixing_method, rho, scf_env%mixing_store, &
                             para_env, rho_atom=rho_atom)
         ELSE
            CALL mixing_init(scf_env%mixing_method, rho, scf_env%mixing_store, &
                             para_env)
         END IF
      END IF

      DO ispin = 1, SIZE(mos) !fm->dbcsr
         IF (mos(ispin)%mo_set%use_mo_coeff_b) THEN
            CALL copy_fm_to_dbcsr(mos(ispin)%mo_set%mo_coeff, &
                                  mos(ispin)%mo_set%mo_coeff_b) !fm->dbcsr
         ENDIF
      ENDDO !fm->dbcsr

      CALL timestop(handle)

   END SUBROUTINE scf_env_initial_rho_setup

END MODULE qs_scf_initialization
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the Quickstep SCF run.
!> \par History
!>      - Joost VandeVondele (02.2002)
!>           added code for: incremental (pab and gvg) update
!>                            initialisation (init_cube, l_info)
!>      - Joost VandeVondele (02.2002)
!>           called the poisson code of the classical part
!>           this takes into account the spherical cutoff and allows for
!>           isolated systems
!>      - Joost VandeVondele (02.2002)
!>           added multiple grid feature
!>           changed to spherical cutoff consistently (?)
!>           therefore removed the gradient correct functionals
!>      - updated with the new QS data structures (10.04.02,MK)
!>      - copy_matrix replaced by transfer_matrix (11.04.02,MK)
!>      - nrebuild_rho and nrebuild_gvg unified (12.04.02,MK)
!>      - set_mo_occupation for smearing of the MO occupation numbers
!>        (17.04.02,MK)
!>      - MO level shifting added (22.04.02,MK)
!>      - Usage of TYPE mo_set_p_type
!>      - Joost VandeVondele (05.2002)
!>            added cholesky based diagonalisation
!>      - 05.2002 added pao method [fawzi]
!>      - parallel FFT (JGH 22.05.2002)
!>      - 06.2002 moved KS matrix construction to qs_build_KS_matrix.F [fawzi]
!>      - started to include more LSD (01.2003,Joost VandeVondele)
!>      - 02.2003 scf_env [fawzi]
!>      - got rid of nrebuild (01.2004, Joost VandeVondele)
!>      - 10.2004 removed pao [fawzi]
!>      - 03.2006 large cleaning action [Joost VandeVondele]
!>      - High-spin ROKS added (05.04.06,MK)
!>      - Mandes (10.2013)
!>        intermediate energy communication with external communicator added
!>      - kpoints (08.2014, JGH)
!>      - unified k-point and gamma-point code (2014.11) [Ole Schuett]
!> \author Matthias Krack (30.04.2001)
! **************************************************************************************************
MODULE qs_scf
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_name,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_types,                     ONLY: cp_fm_release,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_p_file,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: get_results,&
                                              test_for_result
   USE cp_result_types,                 ONLY: cp_result_type
   USE input_constants,                 ONLY: history_guess,&
                                              ot_precond_full_all,&
                                              ot_precond_full_single,&
                                              ot_precond_full_single_inverse,&
                                              ot_precond_none,&
                                              ot_precond_s_inverse
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE kpoint_io,                       ONLY: write_kpoints_restart
   USE kpoint_types,                    ONLY: kpoint_type
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_send
   USE particle_types,                  ONLY: particle_type
   USE preconditioner,                  ONLY: prepare_preconditioner,&
                                              restart_preconditioner
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE qs_block_davidson_types,         ONLY: block_davidson_deallocate
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_density_mixing_types,         ONLY: gspace_mixing_nr
   USE qs_diis,                         ONLY: qs_diis_b_clear,&
                                              qs_diis_b_create
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_mo_io,                        ONLY: write_mo_set
   USE qs_mo_methods,                   ONLY: make_basis_simple,&
                                              make_basis_sm
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_ot,                           ONLY: qs_ot_new_preconditioner
   USE qs_ot_scf,                       ONLY: ot_scf_init,&
                                              ot_scf_read_input
   USE qs_outer_scf,                    ONLY: outer_loop_optimize,&
                                              outer_loop_update_qs_env
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_initialization,           ONLY: qs_scf_compute_properties,&
                                              qs_scf_env_initialize
   USE qs_scf_loop_utils,               ONLY: qs_scf_check_inner_exit,&
                                              qs_scf_check_outer_exit,&
                                              qs_scf_density_mixing,&
                                              qs_scf_inner_finalize,&
                                              qs_scf_new_mos,&
                                              qs_scf_new_mos_kp,&
                                              qs_scf_rho_update,&
                                              qs_scf_set_loop_flags
   USE qs_scf_output,                   ONLY: qs_scf_loop_info,&
                                              qs_scf_loop_print,&
                                              qs_scf_outer_loop_info,&
                                              qs_scf_write_mos
   USE qs_scf_types,                    ONLY: &
        block_davidson_diag_method_nr, block_krylov_diag_method_nr, filter_matrix_diag_method_nr, &
        general_diag_method_nr, ot_diag_method_nr, ot_method_nr, qs_scf_env_type, scf_env_release, &
        special_diag_method_nr
   USE qs_wf_history_methods,           ONLY: wfi_update
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf'

   PUBLIC :: scf, scf_env_cleanup, scf_env_do_scf

CONTAINS

! **************************************************************************************************
!> \brief perform an scf procedure in the given qs_env
!> \param qs_env the qs_environment where to perform the scf procedure
!> \par History
!>      02.2003 introduced scf_env, moved real work to scf_env_do_scf [fawzi]
!> \author fawzi
!> \note
! **************************************************************************************************
   SUBROUTINE scf(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf', routineP = moduleN//':'//routineN

      INTEGER                                            :: max_scf_tmp
      LOGICAL                                            :: converged, outer_scf_loop, should_stop
      LOGICAL, SAVE                                      :: first_step_flag = .TRUE.
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, input, scf_section

      NULLIFY (scf_env)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf.F",173)
      CALL get_qs_env(qs_env, scf_env=scf_env, input=input, &
                      dft_control=dft_control, scf_control=scf_control)
      IF (scf_control%max_scf > 0) THEN

         dft_section => section_vals_get_subs_vals(input, "DFT")
         scf_section => section_vals_get_subs_vals(dft_section, "SCF")

         IF (.NOT. ASSOCIATED(scf_env)) THEN
            CALL qs_scf_env_initialize(qs_env, scf_env)
            ! Moved here from qs_scf_env_initialize to be able to have more scf_env
            CALL set_qs_env(qs_env, scf_env=scf_env)
            CALL scf_env_release(scf_env)
            CALL get_qs_env(qs_env=qs_env, scf_env=scf_env)
         ELSE
            CALL qs_scf_env_initialize(qs_env, scf_env)
         ENDIF

         IF ((scf_control%density_guess .EQ. history_guess) .AND. (first_step_flag)) THEN
            max_scf_tmp = scf_control%max_scf
            scf_control%max_scf = 1
            outer_scf_loop = scf_control%outer_scf%have_scf
            scf_control%outer_scf%have_scf = .FALSE.
         END IF

         CALL scf_env_do_scf(scf_env=scf_env, scf_control=scf_control, qs_env=qs_env, &
                             converged=converged, should_stop=should_stop)

         ! If SCF has not converged, then we should not start MP2
         IF (ASSOCIATED(qs_env%mp2_env)) qs_env%mp2_env%hf_fail = .NOT. converged

         !   *** add the converged wavefunction to the wavefunction history
         IF ((ASSOCIATED(qs_env%wf_history)) .AND. &
             ((scf_control%density_guess .NE. history_guess) .OR. &
              (.NOT. first_step_flag))) THEN
            CALL wfi_update(qs_env%wf_history, qs_env=qs_env, dt=1.0_dp)
         ELSE IF ((scf_control%density_guess .EQ. history_guess) .AND. &
                  (first_step_flag)) THEN
            scf_control%max_scf = max_scf_tmp
            scf_control%outer_scf%have_scf = outer_scf_loop
            first_step_flag = .FALSE.
         END IF

         ! *** compute properties that depend on the converged wavefunction
         IF (.NOT. (should_stop)) CALL qs_scf_compute_properties(qs_env, dft_section)

         ! *** cleanup
         CALL scf_env_cleanup(scf_env)

      END IF

   END SUBROUTINE scf

! **************************************************************************************************
!> \brief perform an scf loop
!> \param scf_env the scf_env where to perform the scf procedure
!> \param scf_control ...
!> \param qs_env the qs_env, the scf_env lives in
!> \param converged will be true / false if converged is reached
!> \param should_stop ...
!> \par History
!>      long history, see cvs and qs_scf module history
!>      02.2003 introduced scf_env [fawzi]
!>      09.2005 Frozen density approximation [TdK]
!>      06.2007 Check for SCF iteration count early [jgh]
!> \author Matthias Krack
!> \note
! **************************************************************************************************
   SUBROUTINE scf_env_do_scf(scf_env, scf_control, qs_env, converged, should_stop)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(OUT)                               :: converged, should_stop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_env_do_scf', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      INTEGER :: ext_master_id, external_comm, handle, handle2, i_tmp, ic, ispin, iter_count, &
         output_unit, scf_energy_message_tag, total_steps
      LOGICAL :: diis_step, do_kpoints, energy_only, exit_inner_loop, exit_outer_loop, &
         inner_loop_converged, just_energy, outer_loop_converged
      REAL(KIND=dp)                                      :: t1, t2
      REAL(KIND=dp), DIMENSION(3)                        :: res_val_3
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: dft_section, input, scf_section

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, rho, energy, &
               logger, qs_charges, ks_env, mos, atomic_kind_set, qs_kind_set, &
               particle_set, dft_section, input, &
               scf_section, para_env, results, kpoints, pw_env, rho_ao_kp)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf.F",282)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf.F",283)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf.F",284)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_scf.F",285)

      logger => cp_get_default_logger()
      t1 = m_walltime()

      CALL get_qs_env(qs_env=qs_env, &
                      energy=energy, &
                      particle_set=particle_set, &
                      qs_charges=qs_charges, &
                      ks_env=ks_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      rho=rho, &
                      mos=mos, &
                      input=input, &
                      dft_control=dft_control, &
                      do_kpoints=do_kpoints, &
                      kpoints=kpoints, &
                      results=results, &
                      pw_env=pw_env, &
                      para_env=para_env)

      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)

      dft_section => section_vals_get_subs_vals(input, "DFT")
      scf_section => section_vals_get_subs_vals(dft_section, "SCF")

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")

      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") &
         "SCF WAVEFUNCTION OPTIMIZATION"

      IF ((output_unit > 0) .AND. (.NOT. scf_control%use_ot)) THEN
         WRITE (UNIT=output_unit, &
                FMT="(/,T3,A,T12,A,T31,A,T39,A,T59,A,T75,A,/,T3,A)") &
            "Step", "Update method", "Time", "Convergence", "Total energy", "Change", &
            REPEAT("-", 78)
      END IF
      CALL cp_add_iter_level(logger%iter_info, "QS_SCF")

      ! check for external communicator and if the inermediate energy should be sended
      res_val_3(:) = -1.0_dp
      description = "[EXT_SCF_ENER_COMM]"
      IF (test_for_result(results, description=description)) THEN
         CALL get_results(results, description=description, &
                          values=res_val_3, n_entries=i_tmp)
         IF(.NOT.(i_tmp .EQ. 3))CALL cp__a("qs_scf.F",332)
         IF (ALL(res_val_3(:) .LE. 0.0)) &
            CALL cp_abort(cp__l("qs_scf.F",334), &
                          " Trying to access result ("//TRIM(description)// &
                          ") which is not correctly stored.")
      END IF
      external_comm = NINT(res_val_3(1))
      ext_master_id = NINT(res_val_3(2))
      scf_energy_message_tag = NINT(res_val_3(3))

      ! *** outer loop of the scf, can treat other variables,
      ! *** such as lagrangian multipliers
      scf_env%outer_scf%iter_count = 0
      iter_count = 0
      total_steps = 0
      energy%tot_old = 0.0_dp

      scf_outer_loop: DO

         CALL init_scf_loop(scf_env=scf_env, qs_env=qs_env, &
                            scf_section=scf_section)

         CALL qs_scf_set_loop_flags(scf_env, diis_step, &
                                    energy_only, just_energy, exit_inner_loop)

         scf_loop: DO

            CALL timeset(routineN//"_inner_loop", handle2)

            scf_env%iter_count = scf_env%iter_count+1
            iter_count = iter_count+1
            CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=iter_count)

            IF (output_unit > 0) CALL m_flush(output_unit)

            total_steps = total_steps+1
            just_energy = energy_only

            CALL qs_ks_update_qs_env(qs_env, just_energy=just_energy, &
                                     calculate_forces=.FALSE.)

            ! print 'heavy weight' or relatively expensive quantities
            CALL qs_scf_loop_print(qs_env, scf_env, para_env)

            IF (do_kpoints) THEN
               ! kpoints
               CALL qs_scf_new_mos_kp(qs_env, scf_env, scf_control, diis_step)
            ELSE
               ! Gamma points only
               CALL qs_scf_new_mos(qs_env, scf_env, scf_control, scf_section, diis_step, energy_only)
            END IF

            ! another heavy weight print object, print controlled by dft_section
            CALL qs_scf_write_mos(mos, atomic_kind_set, qs_kind_set, particle_set, dft_section)

            CALL qs_scf_density_mixing(scf_env, rho, para_env, diis_step)

            t2 = m_walltime()

            CALL qs_scf_loop_info(scf_env, output_unit, just_energy, t1, t2, energy)

            IF (.NOT. just_energy) energy%tot_old = energy%total

            ! check for external communicator and if the intermediate energy should be sent
            IF (scf_energy_message_tag .GT. 0) THEN
               CALL mp_send(energy%total, ext_master_id, scf_energy_message_tag, external_comm)
            END IF

            CALL qs_scf_check_inner_exit(qs_env, scf_env, scf_control, should_stop, exit_inner_loop, &
                                         inner_loop_converged, output_unit)

            ! In case we decide to exit we perform few more check to see if this one
            ! is really the last SCF step
            IF (exit_inner_loop) THEN

               CALL qs_scf_inner_finalize(scf_env, qs_env, diis_step, output_unit)

               CALL qs_scf_check_outer_exit(qs_env, scf_env, scf_control, should_stop, &
                                            outer_loop_converged, exit_outer_loop)

               ! Let's tag the last SCF cycle so we can print informations only of the last step
               IF (exit_outer_loop) CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=iter_count)

            END IF

            IF (do_kpoints) THEN
               CALL write_kpoints_restart(rho_ao_kp, kpoints, scf_env, dft_section, particle_set, qs_kind_set)
            ELSE
               ! Write Wavefunction restart file
               CALL write_mo_set(mos, particle_set, dft_section=dft_section, qs_kind_set=qs_kind_set)
            END IF

            ! Exit if we have finished with the SCF inner loop
            IF (exit_inner_loop) THEN
               CALL timestop(handle2)
               EXIT scf_loop
            END IF

            IF (.NOT. BTEST(cp_print_key_should_output(logger%iter_info, &
                                                       scf_section, "PRINT%ITERATION_INFO/TIME_CUMUL"), cp_p_file)) &
               t1 = m_walltime()

            ! mixing methods have the new density matrix in p_mix_new
            IF (scf_env%mixing_method > 0) THEN
               DO ic = 1, SIZE(rho_ao_kp, 2)
                  DO ispin = 1, dft_control%nspins
                     CALL cp_dbcsr_copy(rho_ao_kp(ispin, ic)%matrix, scf_env%p_mix_new(ispin, ic)%matrix, &
                                        name=TRIM(cp_dbcsr_name(rho_ao_kp(ispin, ic)%matrix)))
                  END DO
               END DO
            END IF

            CALL qs_scf_rho_update(rho, qs_env, scf_env, ks_env, &
                                   mix_rho=scf_env%mixing_method >= gspace_mixing_nr)

            CALL timestop(handle2)

         END DO scf_loop

         IF (.NOT. scf_control%outer_scf%have_scf) EXIT scf_outer_loop

         ! In case we use the OUTER SCF loop let's print some info..
         CALL qs_scf_outer_loop_info(output_unit, scf_control, scf_env, &
                                     energy, total_steps, should_stop, outer_loop_converged)

         IF (exit_outer_loop) EXIT scf_outer_loop

         CALL outer_loop_optimize(scf_env, scf_control)
         CALL outer_loop_update_qs_env(qs_env, scf_env)
         CALL qs_ks_did_change(ks_env, potential_changed=.TRUE.)

      END DO scf_outer_loop

      converged = inner_loop_converged .AND. outer_loop_converged

      IF (.NOT. converged) CALL cp__w("qs_scf.F",467,"SCF run NOT converged")

      ! if needed copy mo_coeff dbcsr->fm for later use in post_scf!fm->dbcsr
      DO ispin = 1, SIZE(mos) !fm -> dbcsr
         IF (mos(ispin)%mo_set%use_mo_coeff_b) THEN !fm->dbcsr
            IF (.NOT. ASSOCIATED(mos(ispin)%mo_set%mo_coeff_b)) & !fm->dbcsr
               CALL cp__b("qs_scf.F",473,"mo_coeff_b is not allocated") !fm->dbcsr
            CALL copy_dbcsr_to_fm(mos(ispin)%mo_set%mo_coeff_b, & !fm->dbcsr
                                  mos(ispin)%mo_set%mo_coeff) !fm -> dbcsr
         ENDIF !fm->dbcsr
      ENDDO !fm -> dbcsr

      IF (dft_control%qs_control%becke_restraint) THEN
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, &
                                   dft_control%qs_control%becke_control%becke_pot%pw)
         dft_control%qs_control%becke_control%need_pot = .TRUE.
      END IF
      CALL cp_rm_iter_level(logger%iter_info, level_name="QS_SCF")
      CALL timestop(handle)

   END SUBROUTINE scf_env_do_scf

! **************************************************************************************************
!> \brief inits those objects needed if you want to restart the scf with, say
!>        only a new initial guess, or different density functional or ...
!>        this will happen just before the scf loop starts
!> \param scf_env ...
!> \param qs_env ...
!> \param scf_section ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE init_scf_loop(scf_env, qs_env, scf_section)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_scf_loop', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nmo, number_of_OT_envs
      LOGICAL                                            :: do_rotation, has_unit_metric, is_full_all
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: orthogonality_metric
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      NULLIFY (scf_control, matrix_s, matrix_ks, dft_control, mos, mo_coeff)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf.F",521)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf.F",522)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_scf.F",523)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_scf.F",524)

      CALL get_qs_env(qs_env=qs_env, &
                      scf_control=scf_control, &
                      dft_control=dft_control, &
                      mos=mos)

      ! if using mo_coeff_b then copy to fm
      DO ispin = 1, SIZE(mos) !fm->dbcsr
         IF (mos(1)%mo_set%use_mo_coeff_b) THEN !fm->dbcsr
            CALL copy_dbcsr_to_fm(mos(ispin)%mo_set%mo_coeff_b, mos(ispin)%mo_set%mo_coeff) !fm->dbcsr
         ENDIF !fm->dbcsr
      ENDDO !fm->dbcsr

      ! this just guarantees that all mo_occupations match the eigenvalues, if smear
      DO ispin = 1, dft_control%nspins
         ! do not reset mo_occupations if the maximum overlap method is in use
         IF (.NOT. scf_control%diagonalization%mom) &
            CALL set_mo_occupation(mo_set=mos(ispin)%mo_set, &
                                   smear=scf_control%smear)
      ENDDO

      SELECT CASE (scf_env%method)
      CASE DEFAULT

         CALL cp__b("qs_scf.F",549,"unknown scf method method:"//cp_to_string(scf_env%method))

      CASE (filter_matrix_diag_method_nr)

         IF (.NOT. scf_env%skip_diis) THEN
            IF (.NOT. ASSOCIATED(scf_env%scf_diis_buffer)) THEN
               CALL qs_diis_b_create(scf_env%scf_diis_buffer, nbuffer=scf_control%max_diis)
            END IF
            CALL qs_diis_b_clear(scf_env%scf_diis_buffer)
         END IF

      CASE (general_diag_method_nr, special_diag_method_nr, block_krylov_diag_method_nr)
         IF (.NOT. scf_env%skip_diis) THEN
            IF (.NOT. ASSOCIATED(scf_env%scf_diis_buffer)) THEN
               CALL qs_diis_b_create(scf_env%scf_diis_buffer, nbuffer=scf_control%max_diis)
            END IF
            CALL qs_diis_b_clear(scf_env%scf_diis_buffer)
         END IF

      CASE (ot_diag_method_nr)
         CALL get_qs_env(qs_env, matrix_ks=matrix_ks, matrix_s=matrix_s)

         IF (.NOT. scf_env%skip_diis) THEN
            IF (.NOT. ASSOCIATED(scf_env%scf_diis_buffer)) THEN
               CALL qs_diis_b_create(scf_env%scf_diis_buffer, nbuffer=scf_control%max_diis)
            END IF
            CALL qs_diis_b_clear(scf_env%scf_diis_buffer)
         END IF

         ! disable DFTB and SE for now
         IF (dft_control%qs_control%dftb .OR. dft_control%qs_control%semi_empirical) THEN
            CALL cp__b("qs_scf.F",580,"DFTB and SE not available with OT/DIAG")
         END IF

         ! if an old preconditioner is still around (i.e. outer SCF is active),
         ! remove it if this could be worthwhile
         CALL restart_preconditioner(qs_env, scf_env%ot_preconditioner, &
                                     scf_control%diagonalization%ot_settings%preconditioner_type, &
                                     dft_control%nspins)

         CALL prepare_preconditioner(qs_env, mos, matrix_ks, matrix_s, scf_env%ot_preconditioner, &
                                     scf_control%diagonalization%ot_settings%preconditioner_type, &
                                     scf_control%diagonalization%ot_settings%precond_solver_type, &
                                     scf_control%diagonalization%ot_settings%energy_gap, dft_control%nspins)

      CASE (block_davidson_diag_method_nr)
         ! Preconditioner initialized within the loop, when required
      CASE (ot_method_nr)
         CALL get_qs_env(qs_env, &
                         has_unit_metric=has_unit_metric, &
                         matrix_s=matrix_s, &
                         matrix_ks=matrix_ks)

         ! reortho the wavefunctions if we are having an outer scf and
         ! this is not the first iteration
         ! this is useful to avoid the build-up of numerical noise
         ! however, we can not play this trick if restricted (don't mix non-equivalent orbs)
         IF (scf_control%do_outer_scf_reortho) THEN
            IF (scf_control%outer_scf%have_scf .AND. .NOT. dft_control%restricted) THEN
               IF (scf_env%outer_scf%iter_count > 0) THEN
                  DO ispin = 1, dft_control%nspins
                     CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
                     IF (has_unit_metric) THEN
                        CALL make_basis_simple(mo_coeff, nmo)
                     ELSE
                        CALL make_basis_sm(mo_coeff, nmo, matrix_s(1)%matrix)
                     ENDIF
                  ENDDO
               ENDIF
            ENDIF
         ELSE
            ! dont need any dirty trick for the numerically stable irac algorithm.
         ENDIF

         IF (.NOT. ASSOCIATED(scf_env%qs_ot_env)) THEN

            ! restricted calculations require just one set of OT orbitals
            number_of_OT_envs = dft_control%nspins
            IF (dft_control%restricted) number_of_OT_envs = 1

            ALLOCATE (scf_env%qs_ot_env(number_of_OT_envs))

            ! XXX Joost XXX should disentangle reading input from this part
            CALL ot_scf_read_input(scf_env%qs_ot_env, scf_section)

            ! keep a note that we are restricted
            IF (dft_control%restricted) THEN
               scf_env%qs_ot_env(1)%restricted = .TRUE.
               ! requires rotation
               IF (.NOT. scf_env%qs_ot_env(1)%settings%do_rotation) &
                  CALL cp_abort(cp__l("qs_scf.F",639), &
                                "Restricted calculation with OT requires orbital rotation. Please "// &
                                "activate the OT%ROTATION keyword!")
            ELSE
               scf_env%qs_ot_env(:)%restricted = .FALSE.
            ENDIF

            ! might need the KS matrix to init properly
            CALL qs_ks_update_qs_env(qs_env, just_energy=.FALSE., &
                                     calculate_forces=.FALSE.)

            ! if an old preconditioner is still around (i.e. outer SCF is active),
            ! remove it if this could be worthwhile
            CALL restart_preconditioner(qs_env, scf_env%ot_preconditioner, &
                                        scf_env%qs_ot_env(1)%settings%preconditioner_type, &
                                        dft_control%nspins)

            !
            ! preconditioning still needs to be done correctly with has_unit_metric
            ! notice that a big part of the preconditioning (S^-1) is fine anyhow
            !
            IF (has_unit_metric) THEN
               NULLIFY (orthogonality_metric)
            ELSE
               orthogonality_metric => matrix_s(1)%matrix
            ENDIF
            IF (dft_control%qs_control%semi_empirical .OR. dft_control%qs_control%dftb) THEN
               scf_env%qs_ot_env(1)%settings%mixed_precision = .FALSE.
            END IF

            CALL prepare_preconditioner(qs_env, mos, matrix_ks, matrix_s, scf_env%ot_preconditioner, &
                                        scf_env%qs_ot_env(1)%settings%preconditioner_type, &
                                        scf_env%qs_ot_env(1)%settings%precond_solver_type, &
                                        scf_env%qs_ot_env(1)%settings%energy_gap, dft_control%nspins, &
                                        has_unit_metric=has_unit_metric, &
                                        mixed_precision=scf_env%qs_ot_env(1)%settings%mixed_precision, &
                                        chol_type=scf_env%qs_ot_env(1)%settings%cholesky_type)

            CALL ot_scf_init(mo_array=mos, matrix_s=orthogonality_metric, &
                             broyden_adaptive_sigma=qs_env%broyden_adaptive_sigma, &
                             qs_ot_env=scf_env%qs_ot_env, matrix_ks=matrix_ks(1)%matrix)

            SELECT CASE (scf_env%qs_ot_env (1)%settings%preconditioner_type)
            CASE (ot_precond_none)
            CASE (ot_precond_full_all, ot_precond_full_single_inverse)
               ! this will rotate the MOs to be eigen states, which is not compatible with rotation
               ! e.g. mo_derivs here do not yet include potentially different occupations numbers
               do_rotation = scf_env%qs_ot_env(1)%settings%do_rotation
               ! only full all needs rotation
               is_full_all = scf_env%qs_ot_env(1)%settings%preconditioner_type == ot_precond_full_all
               IF(.NOT.(.NOT. (do_rotation .AND. is_full_all)))CALL cp__a("qs_scf.F",689)
               DO ispin = 1, SIZE(scf_env%qs_ot_env)
                  CALL qs_ot_new_preconditioner(scf_env%qs_ot_env(ispin), &
                                                scf_env%ot_preconditioner(ispin)%preconditioner)
               ENDDO
            CASE (ot_precond_s_inverse, ot_precond_full_single)
               DO ispin = 1, SIZE(scf_env%qs_ot_env)
                  CALL qs_ot_new_preconditioner(scf_env%qs_ot_env(ispin), &
                                                scf_env%ot_preconditioner(1)%preconditioner)
               ENDDO
            CASE DEFAULT
               DO ispin = 1, SIZE(scf_env%qs_ot_env)
                  CALL qs_ot_new_preconditioner(scf_env%qs_ot_env(ispin), &
                                                scf_env%ot_preconditioner(1)%preconditioner)
               ENDDO
            END SELECT
         ENDIF

         ! if we have non-uniform occupations we should be using rotation
         do_rotation = scf_env%qs_ot_env(1)%settings%do_rotation
         DO ispin = 1, SIZE(mos)
            IF (.NOT. mos(ispin)%mo_set%uniform_occupation) THEN
               IF(.NOT.(do_rotation))CALL cp__a("qs_scf.F",711)
            ENDIF
         ENDDO
      END SELECT

      ! another safety check
      IF (dft_control%low_spin_roks) THEN
         IF(.NOT.(scf_env%method == ot_method_nr))CALL cp__a("qs_scf.F",718)
         do_rotation = scf_env%qs_ot_env(1)%settings%do_rotation
         IF(.NOT.(do_rotation))CALL cp__a("qs_scf.F",720)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE init_scf_loop

! **************************************************************************************************
!> \brief perform cleanup operations (like releasing temporary storage)
!>      at the end of the scf
!> \param scf_env ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_cleanup(scf_env)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_cleanup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf.F",745)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf.F",746)

!   *** Release SCF work storage ***

      IF (ASSOCIATED(scf_env%scf_work1)) THEN
         DO ispin = 1, SIZE(scf_env%scf_work1)
            CALL cp_fm_release(scf_env%scf_work1(ispin)%matrix)
         ENDDO
         DEALLOCATE (scf_env%scf_work1)
      ENDIF
      IF (ASSOCIATED(scf_env%scf_work2)) CALL cp_fm_release(scf_env%scf_work2)
      IF (ASSOCIATED(scf_env%ortho)) CALL cp_fm_release(scf_env%ortho)
      IF (ASSOCIATED(scf_env%ortho_m1)) CALL cp_fm_release(scf_env%ortho_m1)

      IF (ASSOCIATED(scf_env%ortho_dbcsr)) THEN
         CALL cp_dbcsr_deallocate_matrix(scf_env%ortho_dbcsr)
      END IF
      IF (ASSOCIATED(scf_env%buf1_dbcsr)) THEN
         CALL cp_dbcsr_deallocate_matrix(scf_env%buf1_dbcsr)
      END IF
      IF (ASSOCIATED(scf_env%buf2_dbcsr)) THEN
         CALL cp_dbcsr_deallocate_matrix(scf_env%buf2_dbcsr)
      END IF

      IF (ASSOCIATED(scf_env%p_mix_new)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_mix_new)
      END IF

      IF (ASSOCIATED(scf_env%p_delta)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_delta)
      END IF

! *** method dependent cleanup
      SELECT CASE (scf_env%method)
      CASE (ot_method_nr)
         !
      CASE (ot_diag_method_nr)
         !
      CASE (general_diag_method_nr)
         !
      CASE (special_diag_method_nr)
         !
      CASE (block_krylov_diag_method_nr)
      CASE (block_davidson_diag_method_nr)
         CALL block_davidson_deallocate(scf_env%block_davidson_env)
      CASE (filter_matrix_diag_method_nr)
         !
      CASE DEFAULT
         CALL cp__b("qs_scf.F",794,"unknown scf method method:"//cp_to_string(scf_env%method))
      END SELECT

      IF (ASSOCIATED(scf_env%outer_scf%variables)) THEN
         DEALLOCATE (scf_env%outer_scf%variables)
      ENDIF
      IF (ASSOCIATED(scf_env%outer_scf%count)) THEN
         DEALLOCATE (scf_env%outer_scf%count)
      ENDIF
      IF (ASSOCIATED(scf_env%outer_scf%gradient)) THEN
         DEALLOCATE (scf_env%outer_scf%gradient)
      ENDIF
      IF (ASSOCIATED(scf_env%outer_scf%energy)) THEN
         DEALLOCATE (scf_env%outer_scf%energy)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE scf_env_cleanup

END MODULE qs_scf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Subroutines to handle submatrices
!> \par History
!>       2013.01 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE domain_submatrix_methods

   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_col_block_sizes, cp_dbcsr_distribution, cp_dbcsr_filter, cp_dbcsr_finalize, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_matrix_type, cp_dbcsr_get_stored_coordinates, &
        cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, &
        cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, cp_dbcsr_nblkcols_total, &
        cp_dbcsr_nblkrows_total, cp_dbcsr_reserve_block2d, cp_dbcsr_row_block_sizes, &
        cp_dbcsr_type, cp_dbcsr_work_create, dbcsr_distribution_mp, dbcsr_mp_group, &
        dbcsr_mp_mynode, dbcsr_mp_numnodes, dbcsr_type_antisymmetric, dbcsr_type_no_symmetry, &
        dbcsr_type_symmetric
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE domain_submatrix_types,          ONLY: domain_map_type,&
                                              domain_submatrix_type,&
                                              select_row_col
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_alltoall,&
                                              mp_sync
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'domain_submatrix_methods'

   PUBLIC :: copy_submatrices, copy_submatrix_data, &
             release_submatrices, multiply_submatrices, add_submatrices, &
             construct_submatrices, init_submatrices, &
             construct_dbcsr_from_submatrices, &
             set_submatrices, &
             print_submatrices, maxnorm_submatrices

   INTERFACE init_submatrices
      MODULE PROCEDURE init_submatrices_0d
      MODULE PROCEDURE init_submatrices_1d
      MODULE PROCEDURE init_submatrices_2d
   END INTERFACE

   INTERFACE set_submatrices
      MODULE PROCEDURE set_submatrix_array
      MODULE PROCEDURE set_submatrix
   END INTERFACE

   INTERFACE copy_submatrices
      MODULE PROCEDURE copy_submatrix_array
      MODULE PROCEDURE copy_submatrix
   END INTERFACE

   INTERFACE release_submatrices
      MODULE PROCEDURE release_submatrix_array
      MODULE PROCEDURE release_submatrix
   END INTERFACE

   INTERFACE multiply_submatrices
      MODULE PROCEDURE multiply_submatrices_once
      MODULE PROCEDURE multiply_submatrices_array
   END INTERFACE

   INTERFACE add_submatrices
      MODULE PROCEDURE add_submatrices_once
      MODULE PROCEDURE add_submatrices_array
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param subm ...
! **************************************************************************************************
   SUBROUTINE init_submatrices_0d(subm)

      TYPE(domain_submatrix_type), INTENT(INOUT)         :: subm

      subm%domain = -1
      subm%nbrows = -1
      subm%nbcols = -1
      subm%nrows = -1
      subm%ncols = -1
      subm%nnodes = -1
      subm%groupid = -1

   END SUBROUTINE init_submatrices_0d

! **************************************************************************************************
!> \brief ...
!> \param subm ...
! **************************************************************************************************
   SUBROUTINE init_submatrices_1d(subm)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: subm

      subm(:)%domain = -1
      subm(:)%nbrows = -1
      subm(:)%nbcols = -1
      subm(:)%nrows = -1
      subm(:)%ncols = -1
      subm(:)%nnodes = -1
      subm(:)%groupid = -1

   END SUBROUTINE init_submatrices_1d

! **************************************************************************************************
!> \brief ...
!> \param subm ...
! **************************************************************************************************
   SUBROUTINE init_submatrices_2d(subm)

      TYPE(domain_submatrix_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: subm

      subm(:, :)%domain = -1
      subm(:, :)%nbrows = -1
      subm(:, :)%nbcols = -1
      subm(:, :)%nrows = -1
      subm(:, :)%ncols = -1
      subm(:, :)%nnodes = -1
      subm(:, :)%groupid = -1

   END SUBROUTINE init_submatrices_2d

! **************************************************************************************************
!> \brief ...
!> \param original ...
!> \param copy ...
!> \param copy_data ...
! **************************************************************************************************
   SUBROUTINE copy_submatrix_array(original, copy, copy_data)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: original
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: copy
      LOGICAL, INTENT(IN)                                :: copy_data

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_submatrix_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, ndomains, ndomainsB

      CALL timeset(routineN, handle)

      ndomains = SIZE(original)
      ndomainsB = SIZE(copy)
      IF(.NOT.(ndomains .EQ. ndomainsB))CALL cp__a("domain_submatrix_methods.F",160)
      copy(:)%nnodes = original(:)%nnodes
      copy(:)%groupid = original(:)%groupid
      DO idomain = 1, ndomains
         IF (original(idomain)%domain .GT. 0) THEN
            CALL copy_submatrix(original(idomain), copy(idomain), copy_data)
         ENDIF
      ENDDO ! loop over domains

      CALL timestop(handle)

   END SUBROUTINE copy_submatrix_array

! **************************************************************************************************
!> \brief ...
!> \param original ...
!> \param copy ...
!> \param copy_data ...
! **************************************************************************************************
   SUBROUTINE copy_submatrix(original, copy, copy_data)

      TYPE(domain_submatrix_type), INTENT(IN)            :: original
      TYPE(domain_submatrix_type), INTENT(INOUT)         :: copy
      LOGICAL, INTENT(IN)                                :: copy_data

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_submatrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icol, irow

      CALL timeset(routineN, handle)

      copy%domain = original%domain
      copy%nnodes = original%nnodes
      copy%groupid = original%groupid

      IF (original%domain .GT. 0) THEN

         copy%nbrows = original%nbrows
         copy%nbcols = original%nbcols
         copy%nrows = original%nrows
         copy%ncols = original%ncols

         IF (.NOT. ALLOCATED(copy%dbcsr_row)) THEN
            ALLOCATE (copy%dbcsr_row(original%nbrows))
         ELSE
            IF (SIZE(copy%dbcsr_row) .NE. SIZE(original%dbcsr_row)) THEN
               DEALLOCATE (copy%dbcsr_row)
               ALLOCATE (copy%dbcsr_row(original%nbrows))
            ENDIF
         ENDIF
         IF (.NOT. ALLOCATED(copy%dbcsr_col)) THEN
            ALLOCATE (copy%dbcsr_col(original%nbcols))
         ELSE
            IF (SIZE(copy%dbcsr_col) .NE. SIZE(original%dbcsr_col)) THEN
               DEALLOCATE (copy%dbcsr_col)
               ALLOCATE (copy%dbcsr_col(original%nbcols))
            ENDIF
         ENDIF
         IF (.NOT. ALLOCATED(copy%size_brow)) THEN
            ALLOCATE (copy%size_brow(original%nbrows))
         ELSE
            IF (SIZE(copy%size_brow) .NE. SIZE(original%size_brow)) THEN
               DEALLOCATE (copy%size_brow)
               ALLOCATE (copy%size_brow(original%nbrows))
            ENDIF
         ENDIF
         IF (.NOT. ALLOCATED(copy%size_bcol)) THEN
            ALLOCATE (copy%size_bcol(original%nbcols))
         ELSE
            IF (SIZE(copy%size_bcol) .NE. SIZE(original%size_bcol)) THEN
               DEALLOCATE (copy%size_bcol)
               ALLOCATE (copy%size_bcol(original%nbcols))
            ENDIF
         ENDIF

         DO irow = 1, original%nbrows
            copy%dbcsr_row(irow) = original%dbcsr_row(irow)
            copy%size_brow(irow) = original%size_brow(irow)
         ENDDO

         DO icol = 1, original%nbcols
            copy%dbcsr_col(icol) = original%dbcsr_col(icol)
            copy%size_bcol(icol) = original%size_bcol(icol)
         ENDDO

         IF (copy_data) THEN
            CALL copy_submatrix_data(original%mdata, copy)
         ENDIF

      ENDIF ! do not copy empty submatrix

      CALL timestop(handle)

   END SUBROUTINE copy_submatrix

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param copy ...
! **************************************************************************************************
   SUBROUTINE copy_submatrix_data(array, copy)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: array
      TYPE(domain_submatrix_type), INTENT(INOUT)         :: copy

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_submatrix_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ds1, ds2, handle, ms1, ms2

      CALL timeset(routineN, handle)

      IF(.NOT.(copy%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",272)

      ds1 = SIZE(array, 1)
      ds2 = SIZE(array, 2)

      IF (.NOT. ALLOCATED(copy%mdata)) THEN
         ALLOCATE (copy%mdata(ds1, ds2))
      ELSE
         ms1 = SIZE(copy%mdata, 1)
         ms2 = SIZE(copy%mdata, 2)
         IF ((ds1 .NE. ms1) .OR. (ds2 .NE. ms2)) THEN
            DEALLOCATE (copy%mdata)
            ALLOCATE (copy%mdata(ds1, ds2))
         ENDIF
      ENDIF

      copy%mdata(:, :) = array(:, :)

      CALL timestop(handle)

   END SUBROUTINE copy_submatrix_data

! **************************************************************************************************
!> \brief ...
!> \param submatrices ...
!> \param scalar ...
! **************************************************************************************************
   SUBROUTINE set_submatrix_array(submatrices, scalar)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: submatrices
      REAL(KIND=dp), INTENT(IN)                          :: scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'set_submatrix_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, ndomains

      CALL timeset(routineN, handle)

      ndomains = SIZE(submatrices)
      DO idomain = 1, ndomains
         IF (submatrices(idomain)%domain .GT. 0) THEN
            CALL set_submatrix(submatrices(idomain), scalar)
         ENDIF
      ENDDO ! loop over domains

      CALL timestop(handle)

   END SUBROUTINE set_submatrix_array

! **************************************************************************************************
!> \brief ...
!> \param submatrix ...
!> \param scalar ...
! **************************************************************************************************
   SUBROUTINE set_submatrix(submatrix, scalar)

      TYPE(domain_submatrix_type), INTENT(INOUT)         :: submatrix
      REAL(KIND=dp), INTENT(IN)                          :: scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'set_submatrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: ds1, ds2, handle, ms1, ms2

      CALL timeset(routineN, handle)

      IF(.NOT.(submatrix%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",339)
      IF(.NOT.(submatrix%nrows .GT. 0))CALL cp__a("domain_submatrix_methods.F",340)
      IF(.NOT.(submatrix%ncols .GT. 0))CALL cp__a("domain_submatrix_methods.F",341)

      ds1 = submatrix%nrows
      ds2 = submatrix%ncols

      IF (.NOT. ALLOCATED(submatrix%mdata)) THEN
         ALLOCATE (submatrix%mdata(ds1, ds2))
      ELSE
         ms1 = SIZE(submatrix%mdata, 1)
         ms2 = SIZE(submatrix%mdata, 2)
         IF ((ds1 .NE. ms1) .OR. (ds2 .NE. ms2)) THEN
            DEALLOCATE (submatrix%mdata)
            ALLOCATE (submatrix%mdata(ds1, ds2))
         ENDIF
      ENDIF

      submatrix%mdata(:, :) = scalar

      CALL timestop(handle)

   END SUBROUTINE set_submatrix

! **************************************************************************************************
!> \brief ...
!> \param subm ...
! **************************************************************************************************
   SUBROUTINE release_submatrix_array(subm)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: subm

      CHARACTER(len=*), PARAMETER :: routineN = 'release_submatrix_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, ndomains

      CALL timeset(routineN, handle)

      ndomains = SIZE(subm)
      DO idomain = 1, ndomains
         CALL release_submatrix(subm(idomain))
      ENDDO ! loop over domains

      CALL timestop(handle)

   END SUBROUTINE release_submatrix_array

! **************************************************************************************************
!> \brief ...
!> \param subm ...
! **************************************************************************************************
   SUBROUTINE release_submatrix(subm)

      TYPE(domain_submatrix_type), INTENT(INOUT)         :: subm

      CHARACTER(len=*), PARAMETER :: routineN = 'release_submatrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      subm%domain = -1
      subm%nbrows = -1
      subm%nbcols = -1
      subm%nrows = -1
      subm%ncols = -1
      subm%nnodes = -1
      subm%groupid = -1

      IF (ALLOCATED(subm%dbcsr_row)) THEN
         DEALLOCATE (subm%dbcsr_row)
      ENDIF
      IF (ALLOCATED(subm%dbcsr_col)) THEN
         DEALLOCATE (subm%dbcsr_col)
      ENDIF
      IF (ALLOCATED(subm%size_brow)) THEN
         DEALLOCATE (subm%size_brow)
      ENDIF
      IF (ALLOCATED(subm%size_bcol)) THEN
         DEALLOCATE (subm%size_bcol)
      ENDIF
      IF (ALLOCATED(subm%mdata)) THEN
         DEALLOCATE (subm%mdata)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE release_submatrix

   ! more complex routine might be necessary if submatrices are distributed
! **************************************************************************************************
!> \brief ...
!> \param transA ...
!> \param transB ...
!> \param alpha ...
!> \param A ...
!> \param B ...
!> \param beta ...
!> \param C ...
! **************************************************************************************************
   SUBROUTINE multiply_submatrices_once(transA, transB, alpha, A, B, beta, C)

      CHARACTER, INTENT(IN)                              :: transA, transB
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(domain_submatrix_type), INTENT(IN)            :: A, B
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(domain_submatrix_type), INTENT(INOUT)         :: C

      CHARACTER(len=*), PARAMETER :: routineN = 'multiply_submatrices_once', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cs1, cs2, handle, icol, irow, K, K1, &
                                                            LDA, LDB, LDC, M, Mblocks, N, Nblocks
      LOGICAL                                            :: NOTA, NOTB

      CALL timeset(routineN, handle)

      IF(.NOT.(A%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",459)
      IF(.NOT.(B%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",460)
      IF(.NOT.(C%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",461)

      LDA = SIZE(A%mdata, 1)
      LDB = SIZE(B%mdata, 1)

      NOTA = (transA .EQ. 'N') .OR. (transA .EQ. 'n')
      NOTB = (transB .EQ. 'N') .OR. (transB .EQ. 'n')

      IF (NOTA) THEN
         M = A%nrows
         K = A%ncols
         Mblocks = A%nbrows
      ELSE
         M = A%ncols
         K = A%nrows
         Mblocks = A%nbcols
      ENDIF

      IF (NOTB) THEN
         K1 = B%nrows
         N = B%ncols
         Nblocks = B%nbcols
      ELSE
         K1 = B%ncols
         N = B%nrows
         Nblocks = B%nbrows
      ENDIF

      ! these checks are for debugging only
      IF(.NOT.(K .EQ. K1))CALL cp__a("domain_submatrix_methods.F",490)

      ! conform C matrix
      C%nrows = M
      C%ncols = N
      C%nbrows = Mblocks
      C%nbcols = Nblocks
      IF (ALLOCATED(C%dbcsr_row)) THEN
         DEALLOCATE (C%dbcsr_row)
      ENDIF
      ALLOCATE (C%dbcsr_row(C%nbrows))
      IF (ALLOCATED(C%dbcsr_col)) THEN
         DEALLOCATE (C%dbcsr_col)
      ENDIF
      ALLOCATE (C%dbcsr_col(C%nbcols))
      IF (ALLOCATED(C%size_brow)) THEN
         DEALLOCATE (C%size_brow)
      ENDIF
      ALLOCATE (C%size_brow(C%nbrows))
      IF (ALLOCATED(C%size_bcol)) THEN
         DEALLOCATE (C%size_bcol)
      ENDIF
      ALLOCATE (C%size_bcol(C%nbcols))

      DO irow = 1, C%nbrows
         IF (NOTA) THEN
            C%dbcsr_row(irow) = A%dbcsr_row(irow)
            C%size_brow(irow) = A%size_brow(irow)
         ELSE
            C%dbcsr_row(irow) = A%dbcsr_col(irow)
            C%size_brow(irow) = A%size_bcol(irow)
         ENDIF
      ENDDO

      DO icol = 1, C%nbcols
         IF (NOTB) THEN
            C%dbcsr_col(icol) = B%dbcsr_col(icol)
            C%size_bcol(icol) = B%size_bcol(icol)
         ELSE
            C%dbcsr_col(icol) = B%dbcsr_row(icol)
            C%size_bcol(icol) = B%size_brow(icol)
         ENDIF
      ENDDO

      IF (.NOT. ALLOCATED(C%mdata)) THEN
         !!! cannot use non-zero beta if C is not allocated
         IF(.NOT.(beta .EQ. 0.0_dp))CALL cp__a("domain_submatrix_methods.F",536)
         ALLOCATE (C%mdata(C%nrows, C%ncols))
      ELSE
         cs1 = SIZE(C%mdata, 1)
         cs2 = SIZE(C%mdata, 2)
         IF ((C%nrows .NE. cs1) .OR. (C%ncols .NE. cs2)) THEN
            !!! cannot deallocate data if beta is non-zero
            IF(.NOT.(beta .EQ. 0.0_dp))CALL cp__a("domain_submatrix_methods.F",543)
            DEALLOCATE (C%mdata)
            ALLOCATE (C%mdata(C%nrows, C%ncols))
         ENDIF
      ENDIF

      LDC = C%nrows

      CALL DGEMM(transA, transB, M, N, K, alpha, A%mdata, LDA, B%mdata, LDB, beta, C%mdata, LDC)

      C%nnodes = A%nnodes
      C%groupid = A%groupid

      CALL timestop(handle)

   END SUBROUTINE multiply_submatrices_once

! **************************************************************************************************
!> \brief ...
!> \param transA ...
!> \param transB ...
!> \param alpha ...
!> \param A ...
!> \param B ...
!> \param beta ...
!> \param C ...
! **************************************************************************************************
   SUBROUTINE multiply_submatrices_array(transA, transB, alpha, A, B, beta, C)

      CHARACTER, INTENT(IN)                              :: transA, transB
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: A, B
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: C

      CHARACTER(len=*), PARAMETER :: routineN = 'multiply_submatrices_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, idomainA, idomainB, &
                                                            ndomains, ndomainsB, ndomainsC

      CALL timeset(routineN, handle)

      ndomains = SIZE(A)
      ndomainsB = SIZE(B)
      ndomainsC = SIZE(C)

      IF(.NOT.(ndomains .EQ. ndomainsB))CALL cp__a("domain_submatrix_methods.F",592)
      IF(.NOT.(ndomainsB .EQ. ndomainsC))CALL cp__a("domain_submatrix_methods.F",593)

      DO idomain = 1, ndomains

         idomainA = A(idomain)%domain
         idomainB = B(idomain)%domain

         IF(.NOT.(idomainA .EQ. idomainB))CALL cp__a("domain_submatrix_methods.F",600)

         C(idomain)%domain = idomainA

         ! check if the submatrix exists
         IF (idomainA .GT. 0) THEN
            CALL multiply_submatrices_once(transA, transB, alpha, A(idomain), B(idomain), beta, C(idomain))
         ENDIF ! submatrix for the domain exists

      ENDDO ! loop over domains

      CALL timestop(handle)

   END SUBROUTINE multiply_submatrices_array

   ! more complex routine might be necessary if submatrices are distributed
! **************************************************************************************************
!> \brief ...
!> \param alpha ...
!> \param A ...
!> \param beta ...
!> \param B ...
!> \param transB ...
! **************************************************************************************************
   SUBROUTINE add_submatrices_once(alpha, A, beta, B, transB)

      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(domain_submatrix_type), INTENT(INOUT)         :: A
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(domain_submatrix_type), INTENT(IN)            :: B
      CHARACTER, INTENT(IN)                              :: transB

      CHARACTER(len=*), PARAMETER :: routineN = 'add_submatrices_once', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: C1, C2, handle, icol, R1, R2
      LOGICAL                                            :: NOTB

      CALL timeset(routineN, handle)

      IF(.NOT.(A%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",640)
      IF(.NOT.(B%domain .GT. 0))CALL cp__a("domain_submatrix_methods.F",641)

      R1 = A%nrows
      C1 = A%ncols

      NOTB = (transB .EQ. 'N') .OR. (transB .EQ. 'n')

      IF (NOTB) THEN
         R2 = B%nrows
         C2 = B%ncols
      ELSE
         R2 = B%ncols
         C2 = B%nrows
      ENDIF

      ! these checks are for debugging only
      IF(.NOT.(C1 .EQ. C2))CALL cp__a("domain_submatrix_methods.F",657)
      IF(.NOT.(R1 .EQ. R2))CALL cp__a("domain_submatrix_methods.F",658)

      IF (NOTB) THEN
         DO icol = 1, C1
            A%mdata(:, icol) = alpha*A%mdata(:, icol)+beta*B%mdata(:, icol)
         ENDDO
      ELSE
         DO icol = 1, C1
            A%mdata(:, icol) = alpha*A%mdata(:, icol)+beta*B%mdata(icol, :)
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE add_submatrices_once

! **************************************************************************************************
!> \brief ...
!> \param alpha ...
!> \param A ...
!> \param beta ...
!> \param B ...
!> \param transB ...
! **************************************************************************************************
   SUBROUTINE add_submatrices_array(alpha, A, beta, B, transB)

      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: A
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: B
      CHARACTER, INTENT(IN)                              :: transB

      CHARACTER(len=*), PARAMETER :: routineN = 'add_submatrices_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, idomainA, idomainB, &
                                                            ndomains, ndomainsB

      CALL timeset(routineN, handle)

      ndomains = SIZE(A)
      ndomainsB = SIZE(B)

      IF(.NOT.(ndomains .EQ. ndomainsB))CALL cp__a("domain_submatrix_methods.F",703)

      DO idomain = 1, ndomains

         idomainA = A(idomain)%domain
         idomainB = B(idomain)%domain

         IF(.NOT.(idomainA .EQ. idomainB))CALL cp__a("domain_submatrix_methods.F",710)

         ! check if the submatrix exists
         IF (idomainA .GT. 0) THEN
            CALL add_submatrices_once(alpha, A(idomain), beta, B(idomain), transB)
         ENDIF ! submatrix for the domain exists

      ENDDO ! loop over domains

      CALL timestop(handle)

   END SUBROUTINE add_submatrices_array

! **************************************************************************************************
!> \brief Computes the max norm of the collection of submatrices
!> \param submatrices ...
!> \param norm ...
!> \par History
!>       2013.03 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE maxnorm_submatrices(submatrices, norm)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: submatrices
      REAL(KIND=dp), INTENT(OUT)                         :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'maxnorm_submatrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, ndomains
      REAL(KIND=dp)                                      :: curr_norm, send_norm
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_norm

!TYPE(cp_dbcsr_type), INTENT(IN)                        :: distr_pattern

      CALL timeset(routineN, handle)

      send_norm = 0.0_dp

      ndomains = SIZE(submatrices)

      DO idomain = 1, ndomains

         ! check if the submatrix is local
         IF (submatrices(idomain)%domain .GT. 0) THEN
            curr_norm = MAXVAL(ABS(submatrices(idomain)%mdata))
            IF (curr_norm .GT. send_norm) send_norm = curr_norm
         ENDIF

      ENDDO ! loop over domains

      ! communicate local norm to the other nodes
      ALLOCATE (recv_norm(submatrices(1)%nnodes))
      CALL mp_allgather(send_norm, recv_norm, submatrices(1)%groupid)

      norm = MAXVAL(recv_norm)

      DEALLOCATE (recv_norm)

      CALL timestop(handle)

   END SUBROUTINE maxnorm_submatrices

! **************************************************************************************************
!> \brief Computes the sum of traces of the submatrix A.tr(B)
!> \param A ...
!> \param B ...
!> \param trace ...
!> \par History
!>       2013.03 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE trace_submatrices(A, B, trace)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: A, B
      REAL(KIND=dp), INTENT(OUT)                         :: trace

      CHARACTER(len=*), PARAMETER :: routineN = 'trace_submatrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: domainA, domainB, handle, idomain, &
                                                            ndomainsA, ndomainsB
      REAL(KIND=dp)                                      :: curr_trace, send_trace
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_trace

!TYPE(cp_dbcsr_type), INTENT(IN)                        :: distr_pattern

      CALL timeset(routineN, handle)

      send_trace = 0.0_dp

      ndomainsA = SIZE(A)
      ndomainsB = SIZE(B)

      IF(.NOT.(ndomainsA .EQ. ndomainsB))CALL cp__a("domain_submatrix_methods.F",806)

      DO idomain = 1, ndomainsA

         domainA = A(idomain)%domain
         domainB = B(idomain)%domain

         IF(.NOT.(domainA .EQ. domainB))CALL cp__a("domain_submatrix_methods.F",813)

         ! check if the submatrix is local
         IF (domainA .GT. 0) THEN

            IF(.NOT.(A(idomain)%nrows .EQ. B(idomain)%nrows))CALL cp__a("domain_submatrix_methods.F",818)
            IF(.NOT.(A(idomain)%ncols .EQ. B(idomain)%ncols))CALL cp__a("domain_submatrix_methods.F",819)

            curr_trace = SUM(A(idomain)%mdata(:, :)*B(idomain)%mdata(:, :))
            send_trace = send_trace+curr_trace

         ENDIF

      ENDDO ! loop over domains

      ! communicate local norm to the other nodes
      ALLOCATE (recv_trace(A(1)%nnodes))
      CALL mp_allgather(send_trace, recv_trace, A(1)%groupID)

      trace = SUM(recv_trace)

      DEALLOCATE (recv_trace)

      CALL timestop(handle)

   END SUBROUTINE trace_submatrices

! **************************************************************************************************
!> \brief Constructs submatrices for each ALMO domain by collecting distributed
!>        DBCSR blocks to local arrays
!> \param matrix ...
!> \param submatrix ...
!> \param distr_pattern ...
!> \param domain_map ...
!> \param node_of_domain ...
!> \param job_type ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_submatrices(matrix, submatrix, distr_pattern, domain_map, &
                                    node_of_domain, job_type)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: submatrix
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: distr_pattern
      TYPE(domain_map_type), INTENT(IN)                  :: domain_map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain
      INTEGER, INTENT(IN)                                :: job_type

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_submatrices', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: matrix_type
      INTEGER :: block_node, block_offset, col, col_offset, col_size, dest_node, GroupID, handle, &
         iBlock, icol, idomain, index_col, index_ec, index_er, index_row, index_sc, index_sr, &
         iNode, ldesc, myNode, nblkcols_tot, nblkrows_tot, ndomains, ndomains2, nNodes, &
         recv_size2_total, recv_size_total, row, row_size, send_size2_total, send_size_total, &
         smcol, smrow, start_data
      INTEGER, ALLOCATABLE, DIMENSION(:) :: first_col, first_row, offset2_block, offset_block, &
         recv_data2, recv_offset2_cpu, recv_offset_cpu, recv_size2_cpu, recv_size_cpu, send_data2, &
         send_offset2_cpu, send_offset_cpu, send_size2_cpu, send_size_cpu
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: recv_descriptor, send_descriptor
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      LOGICAL                                            :: found, transp
      REAL(KIND=dp)                                      :: antifactor
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_data, send_data
      REAL(KIND=dp), DIMENSION(:), POINTER               :: block_p

!INTEGER, PARAMETER                         :: select_row_col = 1
!INTEGER, PARAMETER                         :: select_row = 2
!                                                  subm_row_size,&
!                                                  subm_col_size,&

      CALL timeset(routineN, handle)

      nblkrows_tot = cp_dbcsr_nblkrows_total(matrix)
      nblkcols_tot = cp_dbcsr_nblkcols_total(matrix)

      ndomains = nblkcols_tot ! RZK-warning not true for atomic distributions

      nNodes = dbcsr_mp_numnodes(dbcsr_distribution_mp( &
                                 cp_dbcsr_distribution(distr_pattern)))
      GroupID = dbcsr_mp_group(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(distr_pattern)))
      myNode = dbcsr_mp_mynode(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(distr_pattern)))

      matrix_type = cp_dbcsr_get_matrix_type(matrix)

      ldesc = 2
      ALLOCATE (send_descriptor(ldesc, nNodes))
      ALLOCATE (recv_descriptor(ldesc, nNodes))
      send_descriptor(:, :) = 0

      ! find: the number of blocks and their sizes that must be sent to each cpu
      ! loop over all domains
      DO idomain = 1, ndomains

         dest_node = node_of_domain(idomain)

         ! loop over those rows that have non-zero quencher
         index_sr = 1 ! index start row
         IF (idomain .GT. 1) index_sr = domain_map%index1(idomain-1)
         index_er = domain_map%index1(idomain)-1 ! index end row

         DO index_row = index_sr, index_er

            row = domain_map%pairs(index_row, 1)

            IF (job_type == select_row_col) THEN
               ! loop over those columns that have non-zero quencher
               index_sc = 1 ! index start col
               IF (idomain .GT. 1) index_sc = domain_map%index1(idomain-1)
               index_ec = domain_map%index1(idomain)-1 ! index end col
            ELSE
               ! fake loop
               index_sc = 1 ! index start col
               index_ec = 1 ! index end col
            ENDIF

            DO index_col = index_sc, index_ec

               IF (job_type == select_row_col) THEN
                  col = domain_map%pairs(index_col, 1)
               ELSE
                  col = idomain
               ENDIF

               transp = .FALSE.
               CALL cp_dbcsr_get_stored_coordinates(matrix, &
                                                    row, col, block_node)
               IF (block_node .EQ. myNode) THEN
                  CALL cp_dbcsr_get_block_p(matrix, row, col, block_p, found, row_size, col_size)
                  IF (found) THEN
                     send_descriptor(1, dest_node+1) = send_descriptor(1, dest_node+1)+1
                     send_descriptor(2, dest_node+1) = send_descriptor(2, dest_node+1)+ &
                                                       row_size*col_size
                  ENDIF
               ENDIF

            ENDDO ! loop over columns

         ENDDO ! loop over rows

      ENDDO

      ! simple but quadratically scaling procedure
      ! loop over local blocks
      !CALL cp_dbcsr_iterator_start(iter,matrix)
      !DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
      !   CALL cp_dbcsr_iterator_next_block(iter,row,col,data_p,&
      !           row_size=row_size,col_size=col_size)
      !   DO idomain = 1, ndomains
      !      IF (job_type==select_row_col) THEN
      !         domain_needs_block=(qblk_exists(domain_map,col,idomain)&
      !            .AND.qblk_exists(domain_map,row,idomain))
      !      ELSE
      !         domain_needs_block=(idomain==col&
      !            .AND.qblk_exists(domain_map,row,idomain))
      !      ENDIF
      !      IF (domain_needs_block) THEN
      !         transp=.FALSE.
      !         dest_node=node_of_domain(idomain)
      !         !CALL cp_dbcsr_get_stored_coordinates(distr_pattern,&
      !         !        idomain, idomain, transp, dest_node)
      !         send_descriptor(1,dest_node+1)=send_descriptor(1,dest_node+1)+1
      !         send_descriptor(2,dest_node+1)=send_descriptor(2,dest_node+1)+&
      !            row_size*col_size
      !      ENDIF
      !   ENDDO
      !ENDDO
      !CALL cp_dbcsr_iterator_stop(iter)

      ! communicate number of blocks and their sizes to the other nodes
      CALL mp_alltoall(send_descriptor, recv_descriptor, ldesc, GroupID)

      ALLOCATE (send_size_cpu(nNodes), send_offset_cpu(nNodes))
      send_offset_cpu(1) = 0
      send_size_cpu(1) = send_descriptor(2, 1)
      DO iNode = 2, nNodes
         send_size_cpu(iNode) = send_descriptor(2, iNode)
         send_offset_cpu(iNode) = send_offset_cpu(iNode-1)+ &
                                  send_size_cpu(iNode-1)
      ENDDO
      send_size_total = send_offset_cpu(nNodes)+send_size_cpu(nNodes)

      ALLOCATE (recv_size_cpu(nNodes), recv_offset_cpu(nNodes))
      recv_offset_cpu(1) = 0
      recv_size_cpu(1) = recv_descriptor(2, 1)
      DO iNode = 2, nNodes
         recv_size_cpu(iNode) = recv_descriptor(2, iNode)
         recv_offset_cpu(iNode) = recv_offset_cpu(iNode-1)+ &
                                  recv_size_cpu(iNode-1)
      ENDDO
      recv_size_total = recv_offset_cpu(nNodes)+recv_size_cpu(nNodes)

      ALLOCATE (send_size2_cpu(nNodes), send_offset2_cpu(nNodes))
      send_offset2_cpu(1) = 0
      send_size2_cpu(1) = 2*send_descriptor(1, 1)
      DO iNode = 2, nNodes
         send_size2_cpu(iNode) = 2*send_descriptor(1, iNode)
         send_offset2_cpu(iNode) = send_offset2_cpu(iNode-1)+ &
                                   send_size2_cpu(iNode-1)
      ENDDO
      send_size2_total = send_offset2_cpu(nNodes)+send_size2_cpu(nNodes)

      ALLOCATE (recv_size2_cpu(nNodes), recv_offset2_cpu(nNodes))
      recv_offset2_cpu(1) = 0
      recv_size2_cpu(1) = 2*recv_descriptor(1, 1)
      DO iNode = 2, nNodes
         recv_size2_cpu(iNode) = 2*recv_descriptor(1, iNode)
         recv_offset2_cpu(iNode) = recv_offset2_cpu(iNode-1)+ &
                                   recv_size2_cpu(iNode-1)
      ENDDO
      recv_size2_total = recv_offset2_cpu(nNodes)+recv_size2_cpu(nNodes)

      DEALLOCATE (send_descriptor)
      DEALLOCATE (recv_descriptor)

      ! collect data from the matrix into send_data
      ALLOCATE (send_data(send_size_total))
      ALLOCATE (recv_data(recv_size_total))
      ALLOCATE (send_data2(send_size2_total))
      ALLOCATE (recv_data2(recv_size2_total))
      ALLOCATE (offset_block(nNodes))
      ALLOCATE (offset2_block(nNodes))
      offset_block(:) = 0
      offset2_block(:) = 0
      ! loop over all domains
      DO idomain = 1, ndomains

         dest_node = node_of_domain(idomain)

         ! loop over those rows that have non-zero quencher
         index_sr = 1 ! index start row
         IF (idomain .GT. 1) index_sr = domain_map%index1(idomain-1)
         index_er = domain_map%index1(idomain)-1 ! index end row

         DO index_row = index_sr, index_er

            row = domain_map%pairs(index_row, 1)

            IF (job_type == select_row_col) THEN
               ! loop over those columns that have non-zero quencher
               index_sc = 1 ! index start col
               IF (idomain .GT. 1) index_sc = domain_map%index1(idomain-1)
               index_ec = domain_map%index1(idomain)-1 ! index end col
            ELSE
               ! fake loop
               index_sc = 1 ! index start col
               index_ec = 1 ! index end col
            ENDIF

            DO index_col = index_sc, index_ec

               IF (job_type == select_row_col) THEN
                  col = domain_map%pairs(index_col, 1)
               ELSE
                  col = idomain
               ENDIF

               transp = .FALSE.
               CALL cp_dbcsr_get_stored_coordinates(matrix, &
                                                    row, col, block_node)
               IF (block_node .EQ. myNode) THEN
                  CALL cp_dbcsr_get_block_p(matrix, row, col, block_p, found, row_size, col_size)
                  IF (found) THEN
                     !col_offset=0
                     !DO icol=1,col_size
                     !   start_data=send_offset_cpu(dest_node+1)+&
                     !              offset_block(dest_node+1)+&
                     !              col_offset
                     !   send_data(start_data+1:start_data+row_size)=&
                     !      data_p(1:row_size,icol)
                     !   col_offset=col_offset+row_size
                     !ENDDO
                     col_offset = row_size*col_size
                     start_data = send_offset_cpu(dest_node+1)+ &
                                  offset_block(dest_node+1)
                     send_data(start_data+1:start_data+col_offset) = &
                        block_p(1:col_offset)
                     offset_block(dest_node+1) = offset_block(dest_node+1)+col_offset
                     ! fill out row,col information
                     send_data2(send_offset2_cpu(dest_node+1)+ &
                                offset2_block(dest_node+1)+1) = row
                     send_data2(send_offset2_cpu(dest_node+1)+ &
                                offset2_block(dest_node+1)+2) = col
                     offset2_block(dest_node+1) = offset2_block(dest_node+1)+2
                  ENDIF
               ENDIF

            ENDDO ! loop over columns

         ENDDO ! loop over rows

      ENDDO
      ! more simple but quadratically scaling version
      !CALL cp_dbcsr_iterator_start(iter,matrix)
      !DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
      !   CALL cp_dbcsr_iterator_next_block(iter,row,col,data_p,&
      !           row_size=row_size,col_size=col_size)
      !   DO idomain = 1, ndomains
      !      IF (job_type==select_row_col) THEN
      !         domain_needs_block=(qblk_exists(domain_map,col,idomain)&
      !            .AND.qblk_exists(domain_map,row,idomain))
      !      ELSE
      !         domain_needs_block=(idomain==col&
      !            .AND.qblk_exists(domain_map,row,idomain))
      !      ENDIF
      !      IF (domain_needs_block) THEN
      !         transp=.FALSE.
      !         dest_node=node_of_domain(idomain)
      !         !CALL cp_dbcsr_get_stored_coordinates(distr_pattern,&
      !         !        idomain, idomain, transp, dest_node)
      !         ! place the data appropriately
      !         col_offset=0
      !         DO icol=1,col_size
      !            start_data=send_offset_cpu(dest_node+1)+&
      !                       offset_block(dest_node+1)+&
      !                       col_offset
      !            send_data(start_data+1:start_data+row_size)=&
      !               data_p(1:row_size,icol)
      !            col_offset=col_offset+row_size
      !         ENDDO
      !         offset_block(dest_node+1)=offset_block(dest_node+1)+col_size*row_size
      !         ! fill out row,col information
      !         send_data2(send_offset2_cpu(dest_node+1)+&
      !               offset2_block(dest_node+1)+1)=row
      !         send_data2(send_offset2_cpu(dest_node+1)+&
      !               offset2_block(dest_node+1)+2)=col
      !         offset2_block(dest_node+1)=offset2_block(dest_node+1)+2
      !      ENDIF
      !   ENDDO
      !ENDDO
      !CALL cp_dbcsr_iterator_stop(iter)

      ! send-receive all blocks
      CALL mp_alltoall(send_data, send_size_cpu, send_offset_cpu, &
                       recv_data, recv_size_cpu, recv_offset_cpu, GroupID)
      ! send-receive rows and cols of the blocks
      CALL mp_alltoall(send_data2, send_size2_cpu, send_offset2_cpu, &
                       recv_data2, recv_size2_cpu, recv_offset2_cpu, GroupID)

      DEALLOCATE (send_size_cpu, send_offset_cpu)
      DEALLOCATE (send_size2_cpu, send_offset2_cpu)
      DEALLOCATE (send_data)
      DEALLOCATE (send_data2)
      DEALLOCATE (offset_block)
      DEALLOCATE (offset2_block)

      ! copy blocks into submatrices
      col_blk_size => cp_dbcsr_col_block_sizes(matrix)
      row_blk_size => cp_dbcsr_row_block_sizes(matrix)
!    ALLOCATE(subm_row_size(ndomains),subm_col_size(ndomains))
!    subm_row_size(:)=0
!    subm_col_size(:)=0
      ndomains2 = SIZE(submatrix)
      IF (ndomains2 .NE. ndomains) THEN
         CALL cp__b("domain_submatrix_methods.F",1173,"wrong submatrix size")
      ENDIF
      CALL release_submatrices(submatrix)
      submatrix(:)%nnodes = nNodes
      submatrix(:)%groupID = GroupID
      submatrix(:)%nrows = 0
      submatrix(:)%ncols = 0

      ALLOCATE (first_row(nblkrows_tot), first_col(nblkcols_tot))
      submatrix(:)%domain = -1
      DO idomain = 1, ndomains
         dest_node = node_of_domain(idomain)
         !transp=.FALSE.
         !CALL cp_dbcsr_get_stored_coordinates(distr_pattern,&
         !        idomain, idomain, transp, dest_node)
         IF (dest_node .EQ. mynode) THEN
            submatrix(idomain)%domain = idomain
            submatrix(idomain)%nbrows = 0
            submatrix(idomain)%nbcols = 0

            ! loop over those rows that have non-zero quencher
            first_row(:) = -1
            index_sr = 1 ! index start row
            IF (idomain .GT. 1) index_sr = domain_map%index1(idomain-1)
            index_er = domain_map%index1(idomain)-1 ! index end row
            DO index_row = index_sr, index_er
               row = domain_map%pairs(index_row, 1)
               !DO row = 1, nblkrows_tot
               !   IF (qblk_exists(domain_map,row,idomain)) THEN
               first_row(row) = submatrix(idomain)%nrows+1
               submatrix(idomain)%nrows = submatrix(idomain)%nrows+row_blk_size(row)
               submatrix(idomain)%nbrows = submatrix(idomain)%nbrows+1
               !   ENDIF
            ENDDO
            ALLOCATE (submatrix(idomain)%dbcsr_row(submatrix(idomain)%nbrows))
            ALLOCATE (submatrix(idomain)%size_brow(submatrix(idomain)%nbrows))
            smrow = 1
            ! again loop over those rows that have non-zero quencher
            index_sr = 1 ! index start row
            IF (idomain .GT. 1) index_sr = domain_map%index1(idomain-1)
            index_er = domain_map%index1(idomain)-1 ! index end row
            DO index_row = index_sr, index_er
               row = domain_map%pairs(index_row, 1)
               !DO row = 1, nblkrows_tot
               !   IF (first_row(row).ne.-1) THEN
               submatrix(idomain)%dbcsr_row(smrow) = row
               submatrix(idomain)%size_brow(smrow) = row_blk_size(row)
               smrow = smrow+1
               !   ENDIF
            ENDDO

            ! loop over the necessary columns
            first_col(:) = -1
            IF (job_type == select_row_col) THEN
               ! loop over those columns that have non-zero quencher
               index_sc = 1 ! index start col
               IF (idomain .GT. 1) index_sc = domain_map%index1(idomain-1)
               index_ec = domain_map%index1(idomain)-1 ! index end col
            ELSE
               ! fake loop
               index_sc = 1 ! index start col
               index_ec = 1 ! index end col
            ENDIF
            DO index_col = index_sc, index_ec
               IF (job_type == select_row_col) THEN
                  col = domain_map%pairs(index_col, 1)
               ELSE
                  col = idomain
               ENDIF
               !DO col = 1, nblkcols_tot
               !   IF (job_type==select_row_col) THEN
               !      domain_needs_block=(qblk_exists(domain_map,col,idomain))
               !   ELSE
               !      domain_needs_block=(col==idomain) ! RZK-warning col belongs to the domain
               !   ENDIF
               !   IF (domain_needs_block) THEN
               first_col(col) = submatrix(idomain)%ncols+1
               submatrix(idomain)%ncols = submatrix(idomain)%ncols+col_blk_size(col)
               submatrix(idomain)%nbcols = submatrix(idomain)%nbcols+1
               !   ENDIF
            ENDDO

            ALLOCATE (submatrix(idomain)%dbcsr_col(submatrix(idomain)%nbcols))
            ALLOCATE (submatrix(idomain)%size_bcol(submatrix(idomain)%nbcols))

            ! loop over the necessary columns again
            smcol = 1
            IF (job_type == select_row_col) THEN
               ! loop over those columns that have non-zero quencher
               index_sc = 1 ! index start col
               IF (idomain .GT. 1) index_sc = domain_map%index1(idomain-1)
               index_ec = domain_map%index1(idomain)-1 ! index end col
            ELSE
               ! fake loop
               index_sc = 1 ! index start col
               index_ec = 1 ! index end col
            ENDIF
            DO index_col = index_sc, index_ec
               IF (job_type == select_row_col) THEN
                  col = domain_map%pairs(index_col, 1)
               ELSE
                  col = idomain
               ENDIF
               !DO col = 1, nblkcols_tot
               !   IF (first_col(col).ne.-1) THEN
               submatrix(idomain)%dbcsr_col(smcol) = col
               submatrix(idomain)%size_bcol(smcol) = col_blk_size(col)
               smcol = smcol+1
               !   ENDIF
            ENDDO

            ALLOCATE (submatrix(idomain)%mdata( &
                      submatrix(idomain)%nrows, &
                      submatrix(idomain)%ncols))
            submatrix(idomain)%mdata(:, :) = 0.0_dp
            DO iNode = 1, nNodes
               block_offset = 0
               DO iBlock = 1, recv_size2_cpu(iNode)/2
                  ! read the (row,col) of the block
                  row = recv_data2(recv_offset2_cpu(iNode)+(iBlock-1)*2+1)
                  col = recv_data2(recv_offset2_cpu(iNode)+(iBlock-1)*2+2)
                  ! check if this block should be in the submatrix of this domain
                  IF ((first_col(col) .NE. -1) .AND. (first_row(row) .NE. -1)) THEN
                     ! copy data from the received array into submatrix
                     start_data = recv_offset_cpu(iNode)+block_offset+1
                     DO icol = 0, col_blk_size(col)-1
                        submatrix(idomain)%mdata(first_row(row): &
                                                 first_row(row)+row_blk_size(row)-1, &
                                                 first_col(col)+icol) = &
                           recv_data(start_data:start_data+row_blk_size(row)-1)
                        start_data = start_data+row_blk_size(row)
                     ENDDO
                     IF (job_type == select_row_col) THEN
                        IF (matrix_type == dbcsr_type_symmetric .OR. &
                            matrix_type == dbcsr_type_antisymmetric) THEN
                           ! copy data into the transposed block as well
                           antifactor = 1.0_dp
                           IF (matrix_type == dbcsr_type_antisymmetric) THEN
                              antifactor = -1.0_dp
                           ENDIF
                           start_data = recv_offset_cpu(iNode)+block_offset+1
                           DO icol = 0, col_blk_size(col)-1
                              submatrix(idomain)%mdata(first_row(col)+icol, &
                                                       first_col(row): &
                                                       first_col(row)+row_blk_size(row)-1) = &
                                 antifactor*recv_data(start_data: &
                                                      start_data+row_blk_size(row)-1)
                              start_data = start_data+row_blk_size(row)
                           ENDDO
                        ELSE IF (matrix_type == dbcsr_type_no_symmetry) THEN
                        ELSE
                           CALL cp__b("domain_submatrix_methods.F",1324,"matrix type is NYI")
                        ENDIF
                     ENDIF
                  ENDIF
                  block_offset = block_offset+col_blk_size(col)*row_blk_size(row)
               ENDDO
            ENDDO
         ENDIF ! mynode.eq.dest_node
      ENDDO ! loop over domains

      DEALLOCATE (recv_size_cpu, recv_offset_cpu)
      DEALLOCATE (recv_size2_cpu, recv_offset2_cpu)
      DEALLOCATE (recv_data)
      DEALLOCATE (recv_data2)
!    DEALLOCATE(subm_row_size,subm_col_size)
      DEALLOCATE (first_row, first_col)

      CALL timestop(handle)

   END SUBROUTINE construct_submatrices

! **************************************************************************************************
!> \brief Constructs a DBCSR matrix from submatrices
!> \param matrix ...
!> \param submatrix ...
!> \param distr_pattern ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_dbcsr_from_submatrices(matrix, submatrix, distr_pattern)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: submatrix
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: distr_pattern

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_dbcsr_from_submatrices', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: matrix_type
      INTEGER :: block_offset, col, col_offset, colsize, dest_node, GroupID, handle, iBlock, icol, &
         idomain, iNode, irow_subm, ldesc, myNode, nblkcols_tot, nblkrows_tot, ndomains, &
         ndomains2, nNodes, recv_size2_total, recv_size_total, row, rowsize, send_size2_total, &
         send_size_total, smroff, start_data, unit_nr
      INTEGER, ALLOCATABLE, DIMENSION(:) :: offset2_block, offset_block, recv_data2, &
         recv_offset2_cpu, recv_offset_cpu, recv_size2_cpu, recv_size_cpu, send_data2, &
         send_offset2_cpu, send_offset_cpu, send_size2_cpu, send_size_cpu
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: recv_descriptor, send_descriptor
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      LOGICAL                                            :: transp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: recv_data, send_data
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: block_p
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nblkrows_tot = cp_dbcsr_nblkrows_total(matrix)
      nblkcols_tot = cp_dbcsr_nblkcols_total(matrix)

      ndomains = nblkcols_tot ! RZK-warning not true for atomic distributions
      ndomains2 = SIZE(submatrix)

      IF (ndomains .NE. ndomains2) THEN
         CALL cp__b("domain_submatrix_methods.F",1397,"domain mismatch")
      ENDIF

      nNodes = dbcsr_mp_numnodes(dbcsr_distribution_mp( &
                                 cp_dbcsr_distribution(distr_pattern)))
      GroupID = dbcsr_mp_group(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(distr_pattern)))
      myNode = dbcsr_mp_mynode(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(distr_pattern)))

      matrix_type = cp_dbcsr_get_matrix_type(matrix)
      IF (matrix_type .NE. dbcsr_type_no_symmetry) THEN
         CALL cp__b("domain_submatrix_methods.F",1409,"only non-symmetric matrices so far")
      ENDIF

      ! remove all blocks from the dbcsr matrix
      CALL cp_dbcsr_iterator_start(iter, matrix)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, block_p)
         block_p(:, :) = 0.0_dp
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)
      CALL cp_dbcsr_filter(matrix, 0.1_dp)

      CALL cp_dbcsr_work_create(matrix, work_mutable=.TRUE.)

      ldesc = 2
      ALLOCATE (send_descriptor(ldesc, nNodes))
      ALLOCATE (recv_descriptor(ldesc, nNodes))
      send_descriptor(:, :) = 0

      ! loop over domains - find how much data to send
      DO idomain = 1, ndomains

         IF (submatrix(idomain)%domain .GT. 0) THEN

            DO irow_subm = 1, submatrix(idomain)%nbrows

               IF (submatrix(idomain)%nbcols .NE. 1) THEN
                  CALL cp__b("domain_submatrix_methods.F",1436,"corrupt submatrix structure")
               ENDIF

               row = submatrix(idomain)%dbcsr_row(irow_subm)
               col = submatrix(idomain)%dbcsr_col(1)

               IF (col .NE. idomain) THEN
                  CALL cp__b("domain_submatrix_methods.F",1443,"corrupt submatrix structure")
               ENDIF

               transp = .FALSE.
               CALL cp_dbcsr_get_stored_coordinates(distr_pattern, &
                                                    row, idomain, dest_node)

               send_descriptor(1, dest_node+1) = send_descriptor(1, dest_node+1)+1
               send_descriptor(2, dest_node+1) = send_descriptor(2, dest_node+1)+ &
                                                 submatrix(idomain)%size_brow(irow_subm)* &
                                                 submatrix(idomain)%size_bcol(1)

            ENDDO ! loop over submatrix blocks

         ENDIF

      ENDDO ! loop over domains

      ! communicate number of blocks and their sizes to the other nodes
      CALL mp_alltoall(send_descriptor, recv_descriptor, ldesc, GroupID)

      ALLOCATE (send_size_cpu(nNodes), send_offset_cpu(nNodes))
      send_offset_cpu(1) = 0
      send_size_cpu(1) = send_descriptor(2, 1)
      DO iNode = 2, nNodes
         send_size_cpu(iNode) = send_descriptor(2, iNode)
         send_offset_cpu(iNode) = send_offset_cpu(iNode-1)+ &
                                  send_size_cpu(iNode-1)
      ENDDO
      send_size_total = send_offset_cpu(nNodes)+send_size_cpu(nNodes)

      ALLOCATE (recv_size_cpu(nNodes), recv_offset_cpu(nNodes))
      recv_offset_cpu(1) = 0
      recv_size_cpu(1) = recv_descriptor(2, 1)
      DO iNode = 2, nNodes
         recv_size_cpu(iNode) = recv_descriptor(2, iNode)
         recv_offset_cpu(iNode) = recv_offset_cpu(iNode-1)+ &
                                  recv_size_cpu(iNode-1)
      ENDDO
      recv_size_total = recv_offset_cpu(nNodes)+recv_size_cpu(nNodes)

      ALLOCATE (send_size2_cpu(nNodes), send_offset2_cpu(nNodes))
      send_offset2_cpu(1) = 0
      send_size2_cpu(1) = 2*send_descriptor(1, 1)
      DO iNode = 2, nNodes
         send_size2_cpu(iNode) = 2*send_descriptor(1, iNode)
         send_offset2_cpu(iNode) = send_offset2_cpu(iNode-1)+ &
                                   send_size2_cpu(iNode-1)
      ENDDO
      send_size2_total = send_offset2_cpu(nNodes)+send_size2_cpu(nNodes)

      ALLOCATE (recv_size2_cpu(nNodes), recv_offset2_cpu(nNodes))
      recv_offset2_cpu(1) = 0
      recv_size2_cpu(1) = 2*recv_descriptor(1, 1)
      DO iNode = 2, nNodes
         recv_size2_cpu(iNode) = 2*recv_descriptor(1, iNode)
         recv_offset2_cpu(iNode) = recv_offset2_cpu(iNode-1)+ &
                                   recv_size2_cpu(iNode-1)
      ENDDO
      recv_size2_total = recv_offset2_cpu(nNodes)+recv_size2_cpu(nNodes)

      DEALLOCATE (send_descriptor)
      DEALLOCATE (recv_descriptor)

      ! collect data from the matrix into send_data
      ALLOCATE (send_data(send_size_total))
      ALLOCATE (recv_data(recv_size_total))
      ALLOCATE (send_data2(send_size2_total))
      ALLOCATE (recv_data2(recv_size2_total))
      ALLOCATE (offset_block(nNodes))
      ALLOCATE (offset2_block(nNodes))
      offset_block(:) = 0
      offset2_block(:) = 0
      ! loop over domains - collect data to send
      DO idomain = 1, ndomains

         IF (submatrix(idomain)%domain .GT. 0) THEN

            smroff = 0
            DO irow_subm = 1, submatrix(idomain)%nbrows

               row = submatrix(idomain)%dbcsr_row(irow_subm)
               col = submatrix(idomain)%dbcsr_col(1)

               rowsize = submatrix(idomain)%size_brow(irow_subm)
               colsize = submatrix(idomain)%size_bcol(1)

               transp = .FALSE.
               CALL cp_dbcsr_get_stored_coordinates(distr_pattern, &
                                                    row, idomain, dest_node)

               ! place the data appropriately
               col_offset = 0
               DO icol = 1, colsize
                  start_data = send_offset_cpu(dest_node+1)+ &
                               offset_block(dest_node+1)+ &
                               col_offset
                  send_data(start_data+1:start_data+rowsize) = &
                     submatrix(idomain)%mdata(smroff+1:smroff+rowsize, icol)
                  col_offset = col_offset+rowsize
               ENDDO
               offset_block(dest_node+1) = offset_block(dest_node+1)+ &
                                           colsize*rowsize
               ! fill out row,col information
               send_data2(send_offset2_cpu(dest_node+1)+ &
                          offset2_block(dest_node+1)+1) = row
               send_data2(send_offset2_cpu(dest_node+1)+ &
                          offset2_block(dest_node+1)+2) = col
               offset2_block(dest_node+1) = offset2_block(dest_node+1)+2

               smroff = smroff+rowsize

            ENDDO

         ENDIF

      ENDDO ! loop over domains

      ! send-receive all blocks
      CALL mp_alltoall(send_data, send_size_cpu, send_offset_cpu, &
                       recv_data, recv_size_cpu, recv_offset_cpu, GroupID)
      ! send-receive rows and cols of the blocks
      CALL mp_alltoall(send_data2, send_size2_cpu, send_offset2_cpu, &
                       recv_data2, recv_size2_cpu, recv_offset2_cpu, GroupID)

      DEALLOCATE (send_size_cpu, send_offset_cpu)
      DEALLOCATE (send_size2_cpu, send_offset2_cpu)
      DEALLOCATE (send_data)
      DEALLOCATE (send_data2)
      DEALLOCATE (offset_block)
      DEALLOCATE (offset2_block)

      ! copy recieved data into dbcsr matrix
      col_blk_size => cp_dbcsr_col_block_sizes(matrix)
      row_blk_size => cp_dbcsr_row_block_sizes(matrix)
      DO iNode = 1, nNodes
         block_offset = 0
         DO iBlock = 1, recv_size2_cpu(iNode)/2
            ! read the (row,col) of the block
            row = recv_data2(recv_offset2_cpu(iNode)+(iBlock-1)*2+1)
            col = recv_data2(recv_offset2_cpu(iNode)+(iBlock-1)*2+2)
            !CALL cp_dbcsr_get_block_p(matrix,row,col,block_p,found)
            !IF (.NOT.found) THEN
            NULLIFY (block_p)
            CALL cp_dbcsr_reserve_block2d(matrix, row, col, block_p)
            IF(.NOT.(ASSOCIATED(block_p)))CALL cp__a("domain_submatrix_methods.F",1588)
            !ENDIF
            ! copy data from the received array into the matrix block
            start_data = recv_offset_cpu(iNode)+block_offset+1
            DO icol = 1, col_blk_size(col)
               block_p(:, icol) = &
                  recv_data(start_data:start_data+row_blk_size(row)-1)
               start_data = start_data+row_blk_size(row)
            ENDDO
            block_offset = block_offset+col_blk_size(col)*row_blk_size(row)
         ENDDO
      ENDDO

      DEALLOCATE (recv_size_cpu, recv_offset_cpu)
      DEALLOCATE (recv_size2_cpu, recv_offset2_cpu)
      DEALLOCATE (recv_data)
      DEALLOCATE (recv_data2)

      CALL cp_dbcsr_finalize(matrix)

      CALL timestop(handle)

   END SUBROUTINE construct_dbcsr_from_submatrices

! **************************************************************************************************
!> \brief ...
!> \param submatrices ...
!> \param mpgroup ...
! **************************************************************************************************
   SUBROUTINE print_submatrices(submatrices, mpgroup)

      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: submatrices
      INTEGER, INTENT(IN)                                :: mpgroup

      CHARACTER(len=*), PARAMETER :: routineN = 'print_submatrices', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=30)                                  :: colstr, formatstr
      INTEGER                                            :: handle, i, irow, n, ncols, nrows

      CALL timeset(routineN, handle)

      n = SIZE(submatrices)

      DO i = 1, n
         nrows = SIZE(submatrices(i)%mdata, 1)
         ncols = SIZE(submatrices(i)%mdata, 2)
         WRITE (colstr, *) ncols
         formatstr = '('//TRIM(ADJUSTL(colstr))//'F16.9)'
         IF (submatrices(i)%domain .GT. 0) THEN
            WRITE (*, *) "SUBMATRIX: ", i, nrows, 'x', ncols
            nrows = SIZE(submatrices(i)%mdata, 1)
            DO irow = 1, nrows
               WRITE (*, formatstr) submatrices(i)%mdata(irow, :)
            ENDDO
         ENDIF
         CALL mp_sync(mpgroup)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE print_submatrices

! **************************************************************************************************
!> \brief Reports whether the DBCSR block (row,col) exists in the quencher
!> \param map ...
!> \param row ...
!> \param col ...
!> \retval qblk_exists ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   FUNCTION qblk_exists(map, row, col)

      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, INTENT(IN)                                :: row, col
      LOGICAL                                            :: qblk_exists

      CHARACTER(len=*), PARAMETER :: routineN = 'qblk_exists', routineP = moduleN//':'//routineN

      INTEGER                                            :: first, last, mid, ndomains

!CALL timeset(routineN,handle)

      ndomains = SIZE(map%index1)

      qblk_exists = .FALSE.
      IF (col .LT. 1 .OR. col .GT. ndomains) RETURN
      first = 1
      IF (col .GT. 1) first = map%index1(col-1)
      last = map%index1(col)-1

      ! perform binary search within first-last
      DO WHILE (last .GE. first)
         mid = first+(last-first)/2
         IF (map%pairs(mid, 1) .GT. row) THEN
            last = mid-1
         ELSE IF (map%pairs(mid, 1) .LT. row) THEN
            first = mid+1
         ELSE
            qblk_exists = .TRUE. ! SUCCESS!!
            EXIT
         END IF
      END DO

      !CALL timestop(handle)

      RETURN

   END FUNCTION qblk_exists

END MODULE domain_submatrix_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Subroutines for ALMO SCF
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf_methods
   USE almo_scf_types,                  ONLY: almo_scf_env_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_desymmetrize, cp_dbcsr_distribution, cp_dbcsr_filter, &
        cp_dbcsr_finalize, cp_dbcsr_frobenius_norm, cp_dbcsr_get_block_p, cp_dbcsr_get_diag, &
        cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_random, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_nblkcols_total, &
        cp_dbcsr_nblkrows_total, cp_dbcsr_norm, cp_dbcsr_print, cp_dbcsr_release, &
        cp_dbcsr_reserve_block2d, cp_dbcsr_row_block_sizes, cp_dbcsr_scale, cp_dbcsr_set, &
        cp_dbcsr_set_diag, cp_dbcsr_transposed, cp_dbcsr_type, cp_dbcsr_work_create, &
        dbcsr_distribution_mp, dbcsr_mp_group, dbcsr_mp_numnodes, dbcsr_norm_maxabsnorm, &
        dbcsr_type_no_symmetry, dbcsr_type_symmetric
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE domain_submatrix_methods,        ONLY: &
        add_submatrices, construct_dbcsr_from_submatrices, construct_submatrices, &
        copy_submatrices, copy_submatrix_data, init_submatrices, multiply_submatrices, &
        print_submatrices, release_submatrices
   USE domain_submatrix_types,          ONLY: domain_map_type,&
                                              domain_submatrix_type,&
                                              select_row,&
                                              select_row_col
   USE input_constants,                 ONLY: almo_domain_layout_molecular,&
                                              almo_mat_distr_atomic,&
                                              almo_scf_diag
   USE iterate_matrix,                  ONLY: invert_Hotelling,&
                                              matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf_methods'

   PUBLIC almo_scf_ks_to_ks_blk, almo_scf_p_blk_to_t_blk, &
      almo_scf_t_blk_to_p, almo_scf_t_blk_to_t_blk_orthonormal, &
      almo_scf_t_to_p, almo_scf_ks_blk_to_tv_blk, &
      almo_scf_ks_xx_to_tv_xx, &
      apply_projector, get_overlap, &
      generator_to_unitary, &
      newton_grad_to_step, orthogonalize_mos, &
      pseudo_invert_diagonal_blk, construct_test, &
      construct_domain_preconditioner, &
      apply_domain_operators, &
      construct_domain_s_inv, &
      construct_domain_s_sqrt, &
      distribute_domains, &
      almo_scf_ks_to_ks_xx, &
      construct_domain_r_down

CONTAINS

! **************************************************************************************************
!> \brief builds projected KS matrices for the overlapping domains
!>        also computes the DIIS error vector as a by-product
!> \param almo_scf_env ...
!> \par History
!>       2013.03 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_ks_to_ks_xx(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_ks_to_ks_xx', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, ndomains
      REAL(KIND=dp)                                      :: eps_multiply
      TYPE(cp_dbcsr_type) :: matrix_tmp1, matrix_tmp2, matrix_tmp3, matrix_tmp4, matrix_tmp5, &
         matrix_tmp6, matrix_tmp7, matrix_tmp8, matrix_tmp9
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_tmp1, subm_tmp2, subm_tmp3

      CALL timeset(routineN, handle)

      eps_multiply = almo_scf_env%eps_filter

      DO ispin = 1, almo_scf_env%nspins

         ndomains = cp_dbcsr_nblkcols_total(almo_scf_env%quench_t(ispin))

         ! 0. Create KS_xx
         CALL construct_submatrices( &
            almo_scf_env%matrix_ks(ispin), &
            almo_scf_env%domain_ks_xx(:, ispin), &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row_col)

         !!!!! RZK-warning MAKE SURE THAT YOU NEED BLOCKS OUTSIDE QUENCH_T
         !!!!! FOR ALL NO-MATRICES NOT COMPUTING THEM CAN SAVE LOTS OF TIME

         ! 1. TMP1=KS.T
         !    Cost: NOn
         !matrix_tmp1 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp1)
         CALL cp_dbcsr_create(matrix_tmp1, &
                              template=almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_ks(ispin), &
                                almo_scf_env%matrix_t(ispin), &
                                0.0_dp, matrix_tmp1, &
                                filter_eps=eps_multiply)

         ! 2. TMP2=TMP1.SigInv=KS.T.SigInv
         !    Cost: NOO
         !matrix_tmp2 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp2)
         CALL cp_dbcsr_create(matrix_tmp2, &
                              template=almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                almo_scf_env%matrix_sigma_inv(ispin), &
                                0.0_dp, matrix_tmp2, &
                                filter_eps=eps_multiply)

         ! 3. TMP1=S.T
         !    Cost: NOn
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s(1), &
                                almo_scf_env%matrix_t(ispin), &
                                0.0_dp, matrix_tmp1, &
                                filter_eps=eps_multiply)

         ! 4. TMP4=TMP2.tr(TMP1)=KS.T.SigInv.tr(T).S
         !    Cost: NNO
         !matrix_tmp4 = create NxN
         CALL cp_dbcsr_init(matrix_tmp4)
         CALL cp_dbcsr_create(matrix_tmp4, &
                              template=almo_scf_env%matrix_s(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp2, &
                                matrix_tmp1, &
                                0.0_dp, matrix_tmp4, &
                                filter_eps=eps_multiply)

         ! 5. KS_xx=KS_xx-TMP4_xx-tr(TMP4_xx)
         ALLOCATE (subm_tmp1(ndomains))
         CALL init_submatrices(subm_tmp1)
         CALL construct_submatrices( &
            matrix_tmp4, &
            subm_tmp1, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row_col)
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              -1.0_dp, subm_tmp1, 'N')
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              -1.0_dp, subm_tmp1, 'T')

         ! 6. TMP3=tr(TMP4).T=S.T.SigInv.tr(T).KS.T
         !    Cost: NOn
         !matrix_tmp3 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp3)
         CALL cp_dbcsr_create(matrix_tmp3, &
                              template=almo_scf_env%matrix_t(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                matrix_tmp4, &
                                almo_scf_env%matrix_t(ispin), &
                                0.0_dp, matrix_tmp3, &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp4)

         ! 8. TMP6=TMP3.SigInv=S.T.SigInv.tr(T).KS.T.SigInv
         !    Cost: NOO
         !matrix_tmp6 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp6)
         CALL cp_dbcsr_create(matrix_tmp6, &
                              template=almo_scf_env%matrix_t(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                matrix_tmp3, &
                                almo_scf_env%matrix_sigma_inv(ispin), &
                                0.0_dp, matrix_tmp6, &
                                filter_eps=eps_multiply)

         ! 8A. Use intermediate matrices to evaluate the gradient/error
         !     Err=(TMP2-TMP6)_q=(KS.T.SigInv-S.T.SigInv.tr(T).KS.T.SigInv)_q
         ! error vector in AO-MO basis
         CALL cp_dbcsr_copy(almo_scf_env%matrix_err_xx(ispin), &
                            almo_scf_env%quench_t(ispin))
         CALL cp_dbcsr_copy(almo_scf_env%matrix_err_xx(ispin), &
                            matrix_tmp2, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_init(matrix_tmp4)
         CALL cp_dbcsr_create(matrix_tmp4, &
                              template=almo_scf_env%matrix_t(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_copy(matrix_tmp4, &
                            almo_scf_env%quench_t(ispin))
         CALL cp_dbcsr_copy(matrix_tmp4, &
                            matrix_tmp6, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_add(almo_scf_env%matrix_err_xx(ispin), &
                           matrix_tmp4, 1.0_dp, -1.0_dp)
         CALL cp_dbcsr_release(matrix_tmp4)
         !
         ! error vector in AO-AO basis
         ! RZK-warning tmp4 can be created using the sparsity pattern,
         ! then retain_sparsity can be used to perform the multiply
         ! this will save some time
         CALL cp_dbcsr_copy(matrix_tmp3, &
                            matrix_tmp2)
         CALL cp_dbcsr_add(matrix_tmp3, &
                           matrix_tmp6, 1.0_dp, -1.0_dp)
         CALL cp_dbcsr_init(matrix_tmp4)
         CALL cp_dbcsr_create(matrix_tmp4, &
                              template=almo_scf_env%matrix_s(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                matrix_tmp3, &
                                almo_scf_env%matrix_t(ispin), &
                                0.0_dp, matrix_tmp4, &
                                filter_eps=eps_multiply)
         CALL construct_submatrices( &
            matrix_tmp4, &
            almo_scf_env%domain_err(:, ispin), &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row_col)
         CALL cp_dbcsr_release(matrix_tmp4)
         ! domain_err submatrices are in down-up representation
         ! bring them into the orthogonalized basis
         ALLOCATE (subm_tmp2(ndomains))
         CALL init_submatrices(subm_tmp2)
         CALL multiply_submatrices('N', 'N', 1.0_dp, &
                                   almo_scf_env%domain_err(:, ispin), &
                                   almo_scf_env%domain_s_sqrt(:, ispin), 0.0_dp, subm_tmp2)
         CALL multiply_submatrices('N', 'N', 1.0_dp, &
                                   almo_scf_env%domain_s_sqrt_inv(:, ispin), &
                                   subm_tmp2, 0.0_dp, almo_scf_env%domain_err(:, ispin))

         ! 9. TMP5=TMP6.tr(TMP1)=S.T.SigInv.tr(T).KS.T.SigInv.tr(T).S
         !    Cost: NNO
         !    matrix_tmp5 = create NxN, full
         ! RZK-warning tmp5 can be created using the sparsity pattern,
         ! then retain_sparsity can be used to perform the multiply
         ! this will save some time
         CALL cp_dbcsr_init(matrix_tmp5)
         CALL cp_dbcsr_create(matrix_tmp5, &
                              template=almo_scf_env%matrix_s(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                matrix_tmp6, &
                                matrix_tmp1, &
                                0.0_dp, matrix_tmp5, &
                                filter_eps=eps_multiply)

         ! 10. KS_xx=KS_xx+TMP5_xx
         CALL construct_submatrices( &
            matrix_tmp5, &
            subm_tmp1, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row_col)
         CALL cp_dbcsr_release(matrix_tmp5)
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              1.0_dp, subm_tmp1, 'N')

         ! 11. KS_xx=KS_xx + [S.T]_xx.[SigInv.tr(T).KS.(1-T.SigInv.tr(T).S)]_xx + transposed
         ALLOCATE (subm_tmp3(ndomains))
         CALL init_submatrices(subm_tmp3)
         CALL construct_submatrices( &
            matrix_tmp2, &
            subm_tmp2, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row)
         CALL construct_submatrices( &
            matrix_tmp6, &
            subm_tmp3, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row)
         CALL cp_dbcsr_release(matrix_tmp6)
         CALL add_submatrices(1.0_dp, subm_tmp2, &
                              -1.0_dp, subm_tmp3, 'N')
         CALL construct_submatrices( &
            matrix_tmp1, &
            subm_tmp3, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row)
         CALL multiply_submatrices('N', 'T', 1.0_dp, subm_tmp2, &
                                   subm_tmp3, 0.0_dp, subm_tmp1)
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              1.0_dp, subm_tmp1, 'N')
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              1.0_dp, subm_tmp1, 'T')

         ! 12. TMP7=tr(T).KS.T.SigInv
         CALL cp_dbcsr_init(matrix_tmp7)
         CALL cp_dbcsr_create(matrix_tmp7, &
                              template=almo_scf_env%matrix_sigma_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                almo_scf_env%matrix_t(ispin), &
                                matrix_tmp2, &
                                0.0_dp, matrix_tmp7, &
                                filter_eps=eps_multiply)

         ! 13. TMP8=[SigInv.tr(T).KS.T.SigInv]_xx
         CALL cp_dbcsr_init(matrix_tmp8)
         CALL cp_dbcsr_create(matrix_tmp8, &
                              template=almo_scf_env%matrix_sigma_blk(ispin), &
                              matrix_type=dbcsr_type_symmetric)
         CALL cp_dbcsr_copy(matrix_tmp8, almo_scf_env%matrix_sigma_blk(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_sigma_inv(ispin), &
                                matrix_tmp7, &
                                0.0_dp, matrix_tmp8, &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp7)

         ! 13. TMP9=[S.T]_xx
         CALL cp_dbcsr_init(matrix_tmp9)
         CALL cp_dbcsr_create(matrix_tmp9, &
                              template=almo_scf_env%matrix_t(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_copy(matrix_tmp9, almo_scf_env%quench_t(ispin))
         CALL cp_dbcsr_copy(matrix_tmp9, matrix_tmp1, keep_sparsity=.TRUE.)

         ! 14. TMP3=TMP9.TMP8=[S.T]_xx.[SigInv.tr(T).KS.T.SigInv]_xx
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                matrix_tmp9, &
                                matrix_tmp8, &
                                0.0_dp, matrix_tmp3, &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp8)
         CALL cp_dbcsr_release(matrix_tmp9)

         ! 15. KS_xx=KS_xx+[S.T]_xx.[SigInv.tr(T).KS.T.SigInv]_xx.[tr(T).S]_xx
         CALL construct_submatrices( &
            matrix_tmp3, &
            subm_tmp2, &
            almo_scf_env%quench_t(ispin), &
            almo_scf_env%domain_map(ispin), &
            almo_scf_env%cpu_of_domain, &
            select_row)
         CALL multiply_submatrices('N', 'T', 1.0_dp, subm_tmp2, &
                                   subm_tmp3, 0.0_dp, subm_tmp1)
         CALL add_submatrices(1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                              1.0_dp, subm_tmp1, 'N')

         !!!!!!! use intermediate matrices to get the error vector !!!!!!!
         !!!!!!! make sure s_blk_sqrt and its inverse exist (i.e. we use diag algorithm)
         !CPPrecondition(almo_scf_env%almo_update_algorithm.eq.almo_scf_diag,cp_failure_level,routineP,failure)
         !! tmp_err = (1-S.T_blk.SigInv.tr(T_blk)).F.T_blk.SigInv
         !CALL cp_dbcsr_init(matrix_tmp_err)
         !CALL cp_dbcsr_create(matrix_tmp_err,&
         !        template=almo_scf_env%matrix_t(ispin))
         !CALL cp_dbcsr_copy(matrix_tmp_err,&
         !        matrix_tmp2)
         !CALL cp_dbcsr_add(matrix_tmp_err,matrix_tmp3,&
         !        1.0_dp,-1.0_dp)
         !! err_blk = tmp_err.tr(T_blk)
         !CALL cp_dbcsr_copy(almo_scf_env%matrix_err_blk(ispin),&
         !        almo_scf_env%matrix_s_blk_sqrt(1))
         !CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp_err,&
         !        almo_scf_env%matrix_t(ispin),&
         !        0.0_dp, almo_scf_env%matrix_err_blk(ispin),&
         !        retain_sparsity=.TRUE.,&
         !        filter_eps=eps_multiply)
         !CALL cp_dbcsr_release(matrix_tmp_err)
         !! bring to the orthogonal basis
         !! err_blk = (S_blk^-1/2).err_blk.(S_blk^1/2)
         !CALL cp_dbcsr_init(matrix_tmp_err)
         !CALL cp_dbcsr_create(matrix_tmp_err,&
         !        template=almo_scf_env%matrix_err_blk(ispin))
         !CALL cp_dbcsr_multiply("N", "N", 1.0_dp,&
         !        almo_scf_env%matrix_err_blk(ispin),&
         !        almo_scf_env%matrix_s_blk_sqrt(1),&
         !        0.0_dp, matrix_tmp_err,&
         !        filter_eps=eps_multiply)
         !CALL cp_dbcsr_multiply("N", "N", 1.0_dp,&
         !        almo_scf_env%matrix_s_blk_sqrt_inv(1),&
         !        matrix_tmp_err,&
         !        0.0_dp, almo_scf_env%matrix_err_blk(ispin),&
         !        filter_eps=eps_multiply)
         !! subtract transpose
         !CALL cp_dbcsr_transposed(matrix_tmp_err,&
         !        almo_scf_env%matrix_err_blk(ispin))
         !CALL cp_dbcsr_add(almo_scf_env%matrix_err_blk(ispin),&
         !        matrix_tmp_err,&
         !        1.0_dp,-1.0_dp)
         !CALL cp_dbcsr_release(matrix_tmp_err)
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         CALL release_submatrices(subm_tmp3)
         CALL release_submatrices(subm_tmp2)
         CALL release_submatrices(subm_tmp1)
         DEALLOCATE (subm_tmp3)
         DEALLOCATE (subm_tmp2)
         DEALLOCATE (subm_tmp1)
         CALL cp_dbcsr_release(matrix_tmp3)
         CALL cp_dbcsr_release(matrix_tmp2)
         CALL cp_dbcsr_release(matrix_tmp1)

      ENDDO ! spins

      CALL timestop(handle)

   END SUBROUTINE almo_scf_ks_to_ks_xx

! **************************************************************************************************
!> \brief computes the projected KS from the total KS matrix
!>        also computes the DIIS error vector as a by-product
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_ks_to_ks_blk(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_ks_to_ks_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      REAL(KIND=dp)                                      :: eps_multiply
      TYPE(cp_dbcsr_type) :: matrix_tmp1, matrix_tmp2, matrix_tmp3, matrix_tmp4, matrix_tmp5, &
         matrix_tmp6, matrix_tmp7, matrix_tmp8, matrix_tmp9, matrix_tmp_err

      CALL timeset(routineN, handle)

      eps_multiply = almo_scf_env%eps_filter

      DO ispin = 1, almo_scf_env%nspins

         ! 1. TMP1=KS.T_blk
         !    Cost: NOn
         !matrix_tmp1 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp1)
         CALL cp_dbcsr_create(matrix_tmp1, &
                              template=almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_ks(ispin), &
                                almo_scf_env%matrix_t_blk(ispin), &
                                0.0_dp, matrix_tmp1, &
                                filter_eps=eps_multiply)
         ! 2. TMP2=TMP1.SigInv=KS.T_blk.SigInv
         !    Cost: NOO
         !matrix_tmp2 = create NxO, full
         CALL cp_dbcsr_init(matrix_tmp2)
         CALL cp_dbcsr_create(matrix_tmp2, &
                              template=almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                almo_scf_env%matrix_sigma_inv(ispin), &
                                0.0_dp, matrix_tmp2, &
                                filter_eps=eps_multiply)

         !!!!!! use intermediate matrices to get the error vector !!!!!!!
         !CALL cp_dbcsr_copy(almo_scf_env%matrix_err_blk(ispin),&
         !        almo_scf_env%matrix_t_blk(ispin))
         !CALL cp_dbcsr_copy(almo_scf_env%matrix_err_blk(ispin),&
         !        matrix_tmp2,&
         !        keep_sparsity=.TRUE.)
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         ! 3. TMP1=S.T_blk
         !    Cost: NOn
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s(1), &
                                almo_scf_env%matrix_t_blk(ispin), &
                                0.0_dp, matrix_tmp1, &
                                filter_eps=eps_multiply)

         ! 4. TMP4_blk=TMP2.tr(TMP1)=KS.T_blk.SigInv.tr(T_blk).S
         !    Cost: NnO
         !matrix_tmp4 = create NxN, blk
         CALL cp_dbcsr_init(matrix_tmp4)
         CALL cp_dbcsr_create(matrix_tmp4, &
                              template=almo_scf_env%matrix_s_blk(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_copy(matrix_tmp4, almo_scf_env%matrix_s_blk(1))
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp2, &
                                matrix_tmp1, &
                                0.0_dp, matrix_tmp4, &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)

         ! 5. KS_blk=KS_blk-TMP4_blk
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), &
                           matrix_tmp4, &
                           1.0_dp, -1.0_dp)

         ! 6. TMP5_blk=tr(TMP4_blk)
         !    KS_blk=KS_blk-tr(TMP4_blk)
         !matrix_tmp5 = create NxN, blk
         CALL cp_dbcsr_init(matrix_tmp5)
         CALL cp_dbcsr_create(matrix_tmp5, &
                              template=almo_scf_env%matrix_s_blk(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_transposed(matrix_tmp5, matrix_tmp4)
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), matrix_tmp5, &
                           1.0_dp, -1.0_dp)

         ! 7. TMP3=tr(T_blk).TMP2=tr(T_blk).KS.T_blk.SigInv
         !    Cost: OOn
         !matrix_tmp3 = create OxO, full
         CALL cp_dbcsr_init(matrix_tmp3)
         CALL cp_dbcsr_create(matrix_tmp3, &
                              template=almo_scf_env%matrix_sigma_inv(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                almo_scf_env%matrix_t_blk(ispin), &
                                matrix_tmp2, &
                                0.0_dp, matrix_tmp3, &
                                filter_eps=eps_multiply)

         ! 8. TMP6=SigInv.TMP3=SigInv.tr(T_blk).KS.T_blk.SigInv
         !    Cost: OOO
         !matrix_tmp6 = create OxO, full
         CALL cp_dbcsr_init(matrix_tmp6)
         CALL cp_dbcsr_create(matrix_tmp6, &
                              template=almo_scf_env%matrix_sigma_inv(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_sigma_inv(ispin), &
                                matrix_tmp3, &
                                0.0_dp, matrix_tmp6, &
                                filter_eps=eps_multiply)

         ! 9. TMP3=TMP1.TMP6=S.T_blk.SigInv.tr(T_blk).KS.T_blk.SigInv
         !    Cost: NOO
         !matrix_tmp3 = re-create NxO, full
         CALL cp_dbcsr_release(matrix_tmp3)
         CALL cp_dbcsr_init(matrix_tmp3)
         CALL cp_dbcsr_create(matrix_tmp3, &
                              template=almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                matrix_tmp6, &
                                0.0_dp, matrix_tmp3, &
                                filter_eps=eps_multiply)

         !!!!!! use intermediate matrices to get the error vector !!!!!!!
         !CALL cp_dbcsr_init(matrix_tmp_err)
         !CALL cp_dbcsr_create(matrix_tmp_err,&
         !        template=almo_scf_env%matrix_t_blk(ispin))
         !CALL cp_dbcsr_copy(matrix_tmp_err,&
         !        almo_scf_env%matrix_t_blk(ispin))
         !CALL cp_dbcsr_copy(matrix_tmp_err,matrix_tmp3,&
         !        keep_sparsity=.TRUE.)
         !CALL cp_dbcsr_add(almo_scf_env%matrix_err_blk(ispin),matrix_tmp_err,&
         !        1.0_dp,-1.0_dp)
         !CALL cp_dbcsr_release(matrix_tmp_err)
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         !!!!!! use intermediate matrices to get the error vector !!!!!!!
         !!!!!! make sure s_blk_sqrt and its inverse exist (i.e. we use diag algorithm)
         IF(.NOT.(almo_scf_env%almo_update_algorithm .EQ. almo_scf_diag))CALL cp__a("almo_scf_methods.F",572)
         ! tmp_err = (1-S.T_blk.SigInv.tr(T_blk)).F.T_blk.SigInv
         CALL cp_dbcsr_init(matrix_tmp_err)
         CALL cp_dbcsr_create(matrix_tmp_err, &
                              template=almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_copy(matrix_tmp_err, &
                            matrix_tmp2)
         CALL cp_dbcsr_add(matrix_tmp_err, matrix_tmp3, &
                           1.0_dp, -1.0_dp)
         ! err_blk = tmp_err.tr(T_blk)
         CALL cp_dbcsr_copy(almo_scf_env%matrix_err_blk(ispin), &
                            almo_scf_env%matrix_s_blk_sqrt(1))
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp_err, &
                                almo_scf_env%matrix_t_blk(ispin), &
                                0.0_dp, almo_scf_env%matrix_err_blk(ispin), &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp_err)
         ! bring to the orthogonal basis
         ! err_blk = (S_blk^-1/2).err_blk.(S_blk^1/2)
         CALL cp_dbcsr_init(matrix_tmp_err)
         CALL cp_dbcsr_create(matrix_tmp_err, &
                              template=almo_scf_env%matrix_err_blk(ispin))
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_err_blk(ispin), &
                                almo_scf_env%matrix_s_blk_sqrt(1), &
                                0.0_dp, matrix_tmp_err, &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                matrix_tmp_err, &
                                0.0_dp, almo_scf_env%matrix_err_blk(ispin), &
                                filter_eps=eps_multiply)
         ! subtract transpose
         CALL cp_dbcsr_transposed(matrix_tmp_err, &
                                  almo_scf_env%matrix_err_blk(ispin))
         CALL cp_dbcsr_add(almo_scf_env%matrix_err_blk(ispin), &
                           matrix_tmp_err, &
                           1.0_dp, -1.0_dp)
         CALL cp_dbcsr_release(matrix_tmp_err)
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         ! later we will need only the blk version of TMP6
         ! create it here and release TMP6
         !matrix_tmp9 = create OxO, blk
         !matrix_tmp9 = copy data from matrix_tmp6, retain sparsity
         !matrix_tmp6 = release
         CALL cp_dbcsr_init(matrix_tmp9)
         CALL cp_dbcsr_create(matrix_tmp9, &
                              template=almo_scf_env%matrix_sigma_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_copy(matrix_tmp9, almo_scf_env%matrix_sigma_blk(ispin))
         CALL cp_dbcsr_copy(matrix_tmp9, matrix_tmp6, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_release(matrix_tmp6)

         !10. KS_blk=KS_blk+TMP3.tr(TMP1)=
         !          =KS_blk+S.T_blk.SigInv.tr.(T_blk).KS.T_blk.SigInv.tr(T_blk).S
         !    Cost: NnO
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp3, &
                                matrix_tmp1, &
                                1.0_dp, almo_scf_env%matrix_ks_blk(ispin), &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)

         ! 11. TMP4_blk=TMP7_blk.tr(TMP8_blk)
         !    Cost: Nnn
         !matrix_tmp7 = create NxO, blk
         !matrix_tmp7 = copy data from matrix_tmp3, retain sparsity
         !matrix_tmp3 = release
         !matrix_tmp8 = create NxO, blk
         !matrix_tmp8 = copy data from matrix_tmp1, retain sparsity
         !matrix_tmp1 = release
         CALL cp_dbcsr_init(matrix_tmp7)
         CALL cp_dbcsr_create(matrix_tmp7, &
                              template=almo_scf_env%matrix_t_blk(ispin))
         ! transfer only the ALMO blocks from tmp3 into tmp7:
         ! first, copy t_blk into tmp7 to transfer the blk structure,
         ! then copy tmp3 into tmp7 with retain_sparsity
         CALL cp_dbcsr_copy(matrix_tmp7, almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_copy(matrix_tmp7, matrix_tmp3, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_release(matrix_tmp3)
         ! do the same for tmp1->tmp8
         CALL cp_dbcsr_init(matrix_tmp8)
         CALL cp_dbcsr_create(matrix_tmp8, &
                              template=almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_copy(matrix_tmp8, almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_copy(matrix_tmp8, matrix_tmp1, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_release(matrix_tmp1)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp7, &
                                matrix_tmp8, &
                                0.0_dp, matrix_tmp4, &
                                filter_eps=eps_multiply, &
                                retain_sparsity=.TRUE.)

         ! 12. KS_blk=KS_blk-TMP4_blk
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), matrix_tmp4, &
                           1.0_dp, -1.0_dp)

         ! 13. TMP5_blk=tr(TMP5_blk)
         !     KS_blk=KS_blk-tr(TMP4_blk)
         CALL cp_dbcsr_transposed(matrix_tmp5, matrix_tmp4)
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), matrix_tmp5, &
                           1.0_dp, -1.0_dp)

         ! 14. TMP4_blk=TMP7_blk.tr(TMP8_blk)
         !     Cost: Nnn
         CALL cp_dbcsr_copy(matrix_tmp7, matrix_tmp2, keep_sparsity=.TRUE.)
         CALL cp_dbcsr_release(matrix_tmp2)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp7, &
                                matrix_tmp8, &
                                0.0_dp, matrix_tmp4, &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)
         ! 15. KS_blk=KS_blk+TMP4_blk
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), matrix_tmp4, &
                           1.0_dp, 1.0_dp)

         ! 16. KS_blk=KS_blk+tr(TMP4_blk)
         CALL cp_dbcsr_transposed(matrix_tmp5, matrix_tmp4)
         CALL cp_dbcsr_release(matrix_tmp4)
         CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), matrix_tmp5, &
                           1.0_dp, 1.0_dp)
         CALL cp_dbcsr_release(matrix_tmp5)

         ! 17. TMP10_blk=TMP8_blk.TMP9_blk
         !    Cost: Noo
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp8, &
                                matrix_tmp9, &
                                0.0_dp, matrix_tmp7, &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp9)

         ! 18. KS_blk=TMP7_blk.tr(TMP8_blk)
         !    Cost: Nno
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp7, &
                                matrix_tmp8, &
                                1.0_dp, almo_scf_env%matrix_ks_blk(ispin), &
                                retain_sparsity=.TRUE., &
                                filter_eps=eps_multiply)
         CALL cp_dbcsr_release(matrix_tmp7)
         CALL cp_dbcsr_release(matrix_tmp8)

      ENDDO ! spins

      CALL timestop(handle)

   END SUBROUTINE almo_scf_ks_to_ks_blk

! **************************************************************************************************
!> \brief ALMOs by diagonalizing the KS domain submatrices
!>        computes both the occupied and virtual orbitals
!> \param almo_scf_env ...
!> \par History
!>       2013.03 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_ks_xx_to_tv_xx(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_ks_xx_to_tv_xx', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iblock_size, idomain, info, &
                                                            ispin, lwork, ndomains
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, work
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: data_copy
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_ks_xx_orthog, subm_t, subm_tmp

      CALL timeset(routineN, handle)

      IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular .AND. &
          almo_scf_env%mat_distr_aos == almo_mat_distr_atomic) THEN
         CALL cp__b("almo_scf_methods.F",747,"a domain must be located entirely on a CPU")
      ENDIF

      ndomains = almo_scf_env%ndomains
      ALLOCATE (subm_tmp(ndomains))
      ALLOCATE (subm_ks_xx_orthog(ndomains))
      ALLOCATE (subm_t(ndomains))

      DO ispin = 1, almo_scf_env%nspins

         CALL init_submatrices(subm_tmp)
         CALL init_submatrices(subm_ks_xx_orthog)

         ! TRY: project out T0-occupied space for each domain
         ! F=(1-R_du).F.(1-tr(R_du))
         !CALL copy_submatrices(almo_scf_env%domain_ks_xx(:,ispin),&
         !        subm_ks_xx_orthog,copy_data=.TRUE.)
         !CALL multiply_submatrices('N','N',1.0_dp,&
         !        almo_scf_env%domain_r_down_up(:,ispin),&
         !        almo_scf_env%domain_ks_xx(:,ispin),0.0_dp,subm_tmp)
         !CALL add_submatrices(1.0_dp,subm_ks_xx_orthog,-1.0_dp,subm_tmp,'N')
         !CALL add_submatrices(1.0_dp,subm_ks_xx_orthog,-1.0_dp,subm_tmp,'T')
         !!CALL multiply_submatrices('N','T',1.0_dp,subm_tmp,&
         !!        almo_scf_env%domain_r_down_up(:,ispin),&
         !!        1.0_dp,subm_ks_xx_orthog)

         ! convert blocks to the orthogonal basis set
         ! TRY: replace one multiply
         !CALL multiply_submatrices('N','N',1.0_dp,subm_ks_xx_orthog,&
         !        almo_scf_env%domain_s_sqrt_inv(:,ispin),0.0_dp,subm_tmp)
         CALL multiply_submatrices('N', 'N', 1.0_dp, almo_scf_env%domain_ks_xx(:, ispin), &
                                   almo_scf_env%domain_s_sqrt_inv(:, ispin), 0.0_dp, subm_tmp)
         CALL multiply_submatrices('N', 'N', 1.0_dp, almo_scf_env%domain_s_sqrt_inv(:, ispin), &
                                   subm_tmp, 0.0_dp, subm_ks_xx_orthog)
         CALL release_submatrices(subm_tmp)

         ! create temporary matrices for occupied and virtual orbitals
         ! represented in the orthogonalized basis set
         CALL init_submatrices(subm_t)

         ! loop over domains - perform diagonalization
         DO idomain = 1, ndomains

            ! check if the submatrix exists
            IF (subm_ks_xx_orthog(idomain)%domain .GT. 0) THEN

               iblock_size = subm_ks_xx_orthog(idomain)%nrows

               ! Prepare data
               ALLOCATE (eigenvalues(iblock_size))
               ALLOCATE (data_copy(iblock_size, iblock_size))
               data_copy(:, :) = subm_ks_xx_orthog(idomain)%mdata(:, :)

               ! Query the optimal workspace for dsyev
               LWORK = -1
               ALLOCATE (WORK(MAX(1, LWORK)))
               CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
               LWORK = INT(WORK(1))
               DEALLOCATE (WORK)

               ! Allocate the workspace and solve the eigenproblem
               ALLOCATE (WORK(MAX(1, LWORK)))
               CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
               IF (INFO .NE. 0) THEN
                  CALL cp__b("almo_scf_methods.F",811,"DSYEV failed")
               END IF

!WRITE (*,*) "Domain", idomain ,"OCC energies", eigenvalues( 1:almo_scf_env%nocc_of_domain(idomain,ispin) )
!WRITE (*,*) "Domain", idomain ,"VIR energies", eigenvalues( almo_scf_env%nocc_of_domain(idomain,ispin)+1 : iblock_size )

               ! Copy occupied eigenvectors
               IF (almo_scf_env%domain_t(idomain, ispin)%ncols .NE. &
                   almo_scf_env%nocc_of_domain(idomain, ispin)) THEN
                  CALL cp__b("almo_scf_methods.F",820,"wrong domain structure")
               END IF
               CALL copy_submatrices(almo_scf_env%domain_t(idomain, ispin), &
                                     subm_t(idomain), .FALSE.)
               CALL copy_submatrix_data(data_copy(:, 1:almo_scf_env%nocc_of_domain(idomain, ispin)), &
                                        subm_t(idomain))

               DEALLOCATE (WORK)
               DEALLOCATE (data_copy)
               DEALLOCATE (eigenvalues)

            ENDIF ! submatrix for the domain exists

         ENDDO ! loop over domains

         CALL release_submatrices(subm_ks_xx_orthog)

         ! convert orbitals to the AO basis set (from orthogonalized AOs)
         CALL multiply_submatrices('N', 'N', 1.0_dp, almo_scf_env%domain_s_sqrt_inv(:, ispin), &
                                   subm_t, 0.0_dp, almo_scf_env%domain_t(:, ispin))
         CALL release_submatrices(subm_t)

         ! convert domain orbitals to a dbcsr matrix
         CALL construct_dbcsr_from_submatrices( &
            almo_scf_env%matrix_t(ispin), &
            almo_scf_env%domain_t(:, ispin), &
            almo_scf_env%quench_t(ispin))
         CALL cp_dbcsr_filter(almo_scf_env%matrix_t(ispin), &
                              almo_scf_env%eps_filter)

         ! TRY: add T0 component
         !!CALL cp_dbcsr_add(almo_scf_env%matrix_t(ispin),&
         !!        almo_scf_env%matrix_t_blk(ispin),1.0_dp,1.0_dp)

      ENDDO ! spins

      DEALLOCATE (subm_tmp)
      DEALLOCATE (subm_ks_xx_orthog)
      DEALLOCATE (subm_t)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_ks_xx_to_tv_xx

! **************************************************************************************************
!> \brief computes ALMOs by diagonalizing the projected blocked KS matrix
!>        uses the diagonalization code for blocks
!>        computes both the occupied and virtual orbitals
!> \param almo_scf_env ...
!> \par History
!>       2011.07 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_ks_blk_to_tv_blk(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_ks_blk_to_tv_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iblock_col, iblock_row, &
                                                            iblock_size, info, ispin, lwork, &
                                                            nocc_of_block, nvirt_of_block, orbital
      LOGICAL                                            :: block_needed
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, work
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: data_copy
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: data_p, p_new_block
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: matrix_ks_blk_orthog, &
                                                            matrix_t_blk_orthog, matrix_tmp, &
                                                            matrix_v_blk_orthog

      CALL timeset(routineN, handle)

      IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular .AND. &
          almo_scf_env%mat_distr_aos == almo_mat_distr_atomic) THEN
         CALL cp__b("almo_scf_methods.F",896,"a domain must be located entirely on a CPU")
      ENDIF

      DO ispin = 1, almo_scf_env%nspins

         CALL cp_dbcsr_init(matrix_tmp)
         CALL cp_dbcsr_init(matrix_ks_blk_orthog)
         CALL cp_dbcsr_create(matrix_tmp, template=almo_scf_env%matrix_ks_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(matrix_ks_blk_orthog, template=almo_scf_env%matrix_ks_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)

         ! convert blocks to the orthogonal basis set
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_ks_blk(ispin), &
                                almo_scf_env%matrix_s_blk_sqrt_inv(1), 0.0_dp, matrix_tmp, &
                                filter_eps=almo_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                matrix_tmp, 0.0_dp, matrix_ks_blk_orthog, &
                                filter_eps=almo_scf_env%eps_filter)

         CALL cp_dbcsr_release(matrix_tmp)

         ! create temporary matrices for occupied and virtual orbitals
         ! represented in the orthogonalized AOs basis set
         CALL cp_dbcsr_init(matrix_t_blk_orthog)
         CALL cp_dbcsr_init(matrix_v_blk_orthog)
         CALL cp_dbcsr_create(matrix_t_blk_orthog, template=almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_create(matrix_v_blk_orthog, template=almo_scf_env%matrix_v_full_blk(ispin))
         CALL cp_dbcsr_work_create(matrix_t_blk_orthog, work_mutable=.TRUE.)
         CALL cp_dbcsr_work_create(matrix_v_blk_orthog, work_mutable=.TRUE.)

         CALL cp_dbcsr_work_create(almo_scf_env%matrix_eoo(ispin), work_mutable=.TRUE.)
         CALL cp_dbcsr_work_create(almo_scf_env%matrix_evv_full(ispin), work_mutable=.TRUE.)

         CALL cp_dbcsr_iterator_start(iter, matrix_ks_blk_orthog)

         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, data_p, row_size=iblock_size)

            block_needed = .FALSE.

            IF (iblock_row == iblock_col) THEN
               block_needed = .TRUE.
            ENDIF

            IF (.NOT. block_needed) THEN
               CALL cp__b("almo_scf_methods.F",942,"off-diagonal block found")
            ENDIF

            IF (block_needed) THEN

               ! Prepare data
               ALLOCATE (eigenvalues(iblock_size))
               ALLOCATE (data_copy(iblock_size, iblock_size))
               data_copy(:, :) = data_p(:, :)

               ! Query the optimal workspace for dsyev
               LWORK = -1
               ALLOCATE (WORK(MAX(1, LWORK)))
               CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
               LWORK = INT(WORK(1))
               DEALLOCATE (WORK)

               ! Allocate the workspace and solve the eigenproblem
               ALLOCATE (WORK(MAX(1, LWORK)))
               CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
               IF (INFO .NE. 0) THEN
                  CALL cp__b("almo_scf_methods.F",963,"DSYEV failed")
               END IF

               !!! RZK-warning                                               !!!
               !!! IT IS EXTREMELY IMPORTANT THAT THE DIAGONAL BLOCKS OF THE !!!
               !!! FOLLOWING MATRICES ARE LOCATED ON THE SAME NODES WITH     !!!
               !!! THE CORRESPONDING DIAGONAL BLOCKS OF THE FOCK MATRIX:     !!!
               !!! T, V, E_o, E_v

               ! copy eigenvectors into two cp_dbcsr matrices - occupied and virtuals
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(matrix_t_blk_orthog, iblock_row, iblock_col, p_new_block)
               nocc_of_block = SIZE(p_new_block, 2)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",976)
               IF(.NOT.(nocc_of_block .GT. 0))CALL cp__a("almo_scf_methods.F",977)
               p_new_block(:, :) = data_copy(:, 1:nocc_of_block)
               ! now virtuals
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(matrix_v_blk_orthog, iblock_row, iblock_col, p_new_block)
               nvirt_of_block = SIZE(p_new_block, 2)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",983)
               IF(.NOT.(nvirt_of_block .GT. 0))CALL cp__a("almo_scf_methods.F",984)
               !CPPrecondition((nvirt_of_block+nocc_of_block.eq.iblock_size),cp_failure_level,routineP,failure)
               p_new_block(:, :) = data_copy(:, (nocc_of_block+1):(nocc_of_block+nvirt_of_block))

               ! copy eigenvalues into two diagonal cp_dbcsr matrices - Eoo and Evv
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_eoo(ispin), iblock_row, iblock_col, p_new_block)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",991)
               p_new_block(:, :) = 0.0_dp
               DO orbital = 1, nocc_of_block
                  p_new_block(orbital, orbital) = eigenvalues(orbital)
               ENDDO
               ! virtual energies
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_evv_full(ispin), iblock_row, iblock_col, p_new_block)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",999)
               p_new_block(:, :) = 0.0_dp
               DO orbital = 1, nvirt_of_block
                  p_new_block(orbital, orbital) = eigenvalues(nocc_of_block+orbital)
               ENDDO

               DEALLOCATE (WORK)
               DEALLOCATE (data_copy)
               DEALLOCATE (eigenvalues)

            ENDIF

         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_finalize(matrix_t_blk_orthog)
         CALL cp_dbcsr_finalize(matrix_v_blk_orthog)
         CALL cp_dbcsr_finalize(almo_scf_env%matrix_eoo(ispin))
         CALL cp_dbcsr_finalize(almo_scf_env%matrix_evv_full(ispin))

         CALL cp_dbcsr_filter(matrix_t_blk_orthog, almo_scf_env%eps_filter)
         CALL cp_dbcsr_filter(matrix_v_blk_orthog, almo_scf_env%eps_filter)

         CALL cp_dbcsr_release(matrix_ks_blk_orthog)

         ! convert orbitals to the AO basis set (from orthogonalized AOs)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                matrix_t_blk_orthog, 0.0_dp, almo_scf_env%matrix_t_blk(ispin), &
                                filter_eps=almo_scf_env%eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                matrix_v_blk_orthog, 0.0_dp, almo_scf_env%matrix_v_full_blk(ispin), &
                                filter_eps=almo_scf_env%eps_filter)

         CALL cp_dbcsr_release(matrix_t_blk_orthog)
         CALL cp_dbcsr_release(matrix_v_blk_orthog)

      ENDDO ! spins

      CALL timestop(handle)

   END SUBROUTINE almo_scf_ks_blk_to_tv_blk

! **************************************************************************************************
!> \brief inverts block-diagonal blocks of a cp_dbcsr_matrix
!> \param matrix_in ...
!> \param matrix_out ...
!> \param nocc ...
!> \par History
!>       2012.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE pseudo_invert_diagonal_blk(matrix_in, matrix_out, nocc)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_in
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_out
      INTEGER, DIMENSION(:)                              :: nocc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pseudo_invert_diagonal_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iblock_col, iblock_row, &
                                                            iblock_size, methodID
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: data_copy
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: data_p, p_new_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_create(matrix_out, template=matrix_in)
      CALL cp_dbcsr_work_create(matrix_out, work_mutable=.TRUE.)

      CALL cp_dbcsr_iterator_start(iter, matrix_in)

      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

         CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, data_p, row_size=iblock_size)

         IF (iblock_row == iblock_col) THEN

            ! Prepare data
            ALLOCATE (data_copy(iblock_size, iblock_size))
            !data_copy(:,:)=data_p(:,:)

            ! 0. Cholesky factorization
            ! 1. Diagonalization
            methodID = 1
            CALL pseudo_invert_matrix(data_p, data_copy, iblock_size, &
                                      methodID, &
                                      range1=nocc(iblock_row), range2=nocc(iblock_row), &
                                      !range1_thr,range2_thr,&
                                      shift=1.0E-5_dp)

            !!! IT IS EXTREMELY IMPORTANT THAT THE BLOCKS OF THE "OUT"  !!!
            !!! MATRIX ARE DISTRIBUTED AS THE BLOCKS OF THE "IN" MATRIX !!!

            NULLIFY (p_new_block)
            CALL cp_dbcsr_reserve_block2d(matrix_out, iblock_row, iblock_col, p_new_block)
            IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",1096)
            p_new_block(:, :) = data_copy(:, :)

            DEALLOCATE (data_copy)

         ENDIF

      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_finalize(matrix_out)

      CALL timestop(handle)

   END SUBROUTINE pseudo_invert_diagonal_blk

! **************************************************************************************************
!> \brief computes occupied ALMOs from the superimposed atomic density blocks
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_p_blk_to_t_blk(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_p_blk_to_t_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      TYPE(cp_dbcsr_type)                                :: matrix_t_blk_tmp

      CALL timeset(routineN, handle)

      DO ispin = 1, almo_scf_env%nspins

         !! create a temporary matrix to keep the eigenvectors
         !CALL cp_dbcsr_init(matrix_t_blk_tmp)
         !CALL cp_dbcsr_create(matrix_t_blk_tmp,&
         !                     template=almo_scf_env%matrix_t_blk(ispin))
         !CALL cp_dbcsr_work_create(matrix_t_blk_tmp,&
         !        work_mutable=.TRUE.)

         !CALL cp_dbcsr_iterator_start(iter,almo_scf_env%matrix_p_blk(ispin))
         !DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         !   CALL cp_dbcsr_iterator_next_block(iter,iblock_row,iblock_col,data_p,row_size=iblock_size)

         !   block_needed=.FALSE.
         !
         !   IF (iblock_row==iblock_col) THEN
         !       block_needed=.TRUE.
         !   ENDIF

         !   IF (.NOT.block_needed) THEN
         !      CPErrorMessage(cp_failure_level,routineP,"off-diagonal block found")
         !      CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
         !   ENDIF

         !   IF (block_needed) THEN

         !      ! Prepare data
         !      ALLOCATE(eigenvalues(iblock_size))
         !      ALLOCATE(data_copy(iblock_size,iblock_size))
         !      data_copy(:,:)=data_p(:,:)

         !      ! Query the optimal workspace for dsyev
         !      LWORK = -1
         !      ALLOCATE(WORK(MAX(1,LWORK)))
         !      CALL DSYEV('V','L',iblock_size,data_copy,iblock_size,eigenvalues,WORK,LWORK,INFO)
         !      LWORK = INT(WORK( 1 ))
         !      DEALLOCATE(WORK)

         !      ! Allocate the workspace and solve the eigenproblem
         !      ALLOCATE(WORK(MAX(1,LWORK)))
         !      CALL DSYEV('V','L',iblock_size,data_copy,iblock_size,eigenvalues,WORK,LWORK,INFO)
         !      IF( INFO.NE.0 ) THEN
         !         CPErrorMessage(cp_failure_level,routineP,"DSYEV failed")
         !         CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
         !      END IF

         !      !!! IT IS EXTREMELY IMPORTANT THAT THE DIAGONAL BLOCKS OF THE !!!
         !      !!! P AND T MATRICES ARE LOCATED ON THE SAME NODES            !!!

         !      ! copy eigenvectors into two cp_dbcsr matrices - occupied and virtuals
         !      NULLIFY (p_new_block)
         !      CALL cp_dbcsr_reserve_block2d(matrix_t_blk_tmp,&
         !              iblock_row,iblock_col,p_new_block)
         !      nocc_of_block=SIZE(p_new_block,2)
         !      CPPostcondition(ASSOCIATED(p_new_block),cp_failure_level,routineP,failure)
         !      CPPrecondition(nocc_of_block.gt.0,cp_failure_level,routineP,failure)
         !      p_new_block(:,:) = data_copy(:,iblock_size-nocc_of_block+1:)

         !      DEALLOCATE(WORK)
         !      DEALLOCATE(data_copy)
         !      DEALLOCATE(eigenvalues)

         !   ENDIF
         !
         !ENDDO
         !CALL cp_dbcsr_iterator_stop(iter)

         !CALL cp_dbcsr_finalize(matrix_t_blk_tmp)
         !CALL cp_dbcsr_filter(matrix_t_blk_tmp,&
         !        almo_scf_env%eps_filter)
         !CALL cp_dbcsr_copy(almo_scf_env%matrix_t_blk(ispin),&
         !        matrix_t_blk_tmp)
         !CALL cp_dbcsr_release(matrix_t_blk_tmp)

         !! generate a random set of ALMOs
         !! matrix_t_blk should already be initiated to the proper domain structure
         CALL cp_dbcsr_init_random(almo_scf_env%matrix_t_blk(ispin), &
                                   keep_sparsity=.TRUE.)

         CALL cp_dbcsr_init(matrix_t_blk_tmp)
         CALL cp_dbcsr_create(matrix_t_blk_tmp, &
                              template=almo_scf_env%matrix_t_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)

         ! use current ALMOs in matrix_t_blk and project them onto the blocked dm
         ! compute T_new = R_blk S_blk T_random
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_blk(1), &
                                almo_scf_env%matrix_t_blk(ispin), &
                                0.0_dp, matrix_t_blk_tmp, &
                                filter_eps=almo_scf_env%eps_filter)

         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_p_blk(ispin), matrix_t_blk_tmp, &
                                0.0_dp, almo_scf_env%matrix_t_blk(ispin), &
                                filter_eps=almo_scf_env%eps_filter)

         CALL cp_dbcsr_release(matrix_t_blk_tmp)

      ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_p_blk_to_t_blk

! **************************************************************************************************
!> \brief Computes the overlap matrix of MO orbitals
!> \param bra ...
!> \param ket ...
!> \param overlap ...
!> \param metric ...
!> \param retain_overlap_sparsity ...
!> \param eps_filter ...
!> \par History
!>       2011.08 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE get_overlap(bra, ket, overlap, metric, retain_overlap_sparsity, &
                          eps_filter)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: bra, ket
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: overlap
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: metric
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_overlap_sparsity
      REAL(KIND=dp)                                      :: eps_filter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_overlap', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: local_retain_sparsity
      TYPE(cp_dbcsr_type)                                :: tmp

      CALL timeset(routineN, handle)

      IF (.NOT. PRESENT(retain_overlap_sparsity)) THEN
         local_retain_sparsity = .FALSE.
      ELSE
         local_retain_sparsity = retain_overlap_sparsity
      ENDIF

      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=ket, &
                           matrix_type=dbcsr_type_no_symmetry)

      ! TMP=metric*ket
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             metric, ket, 0.0_dp, tmp, &
                             filter_eps=eps_filter)

      ! OVERLAP=tr(bra)*TMP
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                             bra, tmp, 0.0_dp, overlap, &
                             retain_sparsity=local_retain_sparsity, &
                             filter_eps=eps_filter)

      CALL cp_dbcsr_release(tmp)

      CALL timestop(handle)

   END SUBROUTINE get_overlap

!! **************************************************************************************************
!!> \brief Create the overlap matrix of virtual orbitals
!!> \par History
!!>       2011.07 created [Rustam Z Khaliullin]
!!> \author Rustam Z Khaliullin
!! **************************************************************************************************
!  SUBROUTINE almo_scf_v_to_sigma_vv(almo_scf_env)
!
!    TYPE(almo_scf_env_type), INTENT(INOUT)   :: almo_scf_env
!
!    CHARACTER(LEN=*), PARAMETER :: &
!      routineN = 'almo_scf_v_to_sigma_vv', &
!      routineP = moduleN//':'//routineN
!
!    TYPE(cp_dbcsr_type)                      :: tmp
!    INTEGER                                  :: ispin, handle
!
!    CALL timeset(routineN,handle)
!
!    DO ispin=1,almo_scf_env%nspins
!
!       CALL cp_dbcsr_init(tmp)
!       CALL cp_dbcsr_create(tmp,&
!               template=almo_scf_env%matrix_v(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! TMP=S.V
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!               almo_scf_env%matrix_s(1),&
!               almo_scf_env%matrix_v(ispin),&
!               0.0_dp,tmp,&
!               filter_eps=almo_scf_env%eps_filter)
!
!       ! Sig_vv=tr(V).S.V - get MO overlap
!       CALL cp_dbcsr_multiply("T","N",1.0_dp,&
!               almo_scf_env%matrix_v(ispin),&
!               tmp,&
!               0.0_dp,almo_scf_env%matrix_sigma_vv(ispin),&
!               filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_release(tmp)
!
!    END DO
!
!    CALL timestop(handle)
!
!  END SUBROUTINE almo_scf_v_to_sigma_vv

!! **************************************************************************************************
!!> \brief orthogonalize virtual oribitals within a domain
!!> \par History
!!>       2011.07 created [Rustam Z Khaliullin]
!!> \author Rustam Z Khaliullin
!! **************************************************************************************************
!  SUBROUTINE almo_scf_v_to_v_orthonormal_blk(almo_scf_env)
!
!    TYPE(almo_scf_env_type), INTENT(INOUT)   :: almo_scf_env
!
!    CHARACTER(LEN=*), PARAMETER :: &
!      routineN = 'almo_scf_v_to_v_orthonormal_blk', &
!      routineP = moduleN//':'//routineN
!
!    TYPE(cp_dbcsr_type)                      :: matrix_v_tmp,&
!                                                sigma_vv_blk_sqrt,&
!                                                sigma_vv_blk_sqrt_inv
!    INTEGER                                  :: ispin, handle
!
!
!    CALL timeset(routineN,handle)
!
!    DO ispin=1,almo_scf_env%nspins
!
!       CALL cp_dbcsr_init(matrix_v_tmp)
!       CALL cp_dbcsr_create(matrix_v_tmp,&
!                            template=almo_scf_env%matrix_v(ispin))
!
!       ! TMP=S.V
!       CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s(1),&
!                              almo_scf_env%matrix_v(ispin),&
!                              0.0_dp, matrix_v_tmp,&
!                              filter_eps=almo_scf_env%eps_filter)
!
!       ! Sig_blk=tr(V).TMP - get blocked MO overlap
!       CALL cp_dbcsr_multiply("T", "N", 1.0_dp,&
!                              almo_scf_env%matrix_v(ispin),&
!                              matrix_v_tmp,&
!                              0.0_dp, almo_scf_env%matrix_sigma_vv_blk(ispin),&
!                              filter_eps=almo_scf_env%eps_filter,&
!                              retain_sparsity=.TRUE.)
!
!       CALL cp_dbcsr_init(sigma_vv_blk_sqrt)
!       CALL cp_dbcsr_init(sigma_vv_blk_sqrt_inv)
!       CALL cp_dbcsr_create(sigma_vv_blk_sqrt,template=almo_scf_env%matrix_sigma_vv_blk(ispin),&
!                            matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_create(sigma_vv_blk_sqrt_inv,template=almo_scf_env%matrix_sigma_vv_blk(ispin),&
!                            matrix_type=dbcsr_type_no_symmetry)
!
!       ! compute sqrt and sqrt_inv of the blocked MO overlap
!       CALL matrix_sqrt_Newton_Schulz(sigma_vv_blk_sqrt,sigma_vv_blk_sqrt_inv,&
!                                      almo_scf_env%matrix_sigma_vv_blk(ispin),&
!                                      threshold=almo_scf_env%eps_filter,&
!                                      order=almo_scf_env%order_lanczos,&
!                                      eps_lanczos=almo_scf_env%eps_lancsoz,&
!                                      max_iter_lanczos=almo_scf_env%max_iter_lanczos)
!
!       ! TMP_blk=V.SigSQRTInv_blk
!       CALL cp_dbcsr_multiply("N", "N", 1.0_dp,&
!                              almo_scf_env%matrix_v(ispin),&
!                              sigma_vv_blk_sqrt_inv,&
!                              0.0_dp, matrix_v_tmp,&
!                              filter_eps=almo_scf_env%eps_filter)
!
!       ! update the orbitals with the orthonormalized MOs
!       CALL cp_dbcsr_copy(almo_scf_env%matrix_v(ispin),matrix_v_tmp)
!
!       CALL cp_dbcsr_release (matrix_v_tmp)
!       CALL cp_dbcsr_release (sigma_vv_blk_sqrt)
!       CALL cp_dbcsr_release (sigma_vv_blk_sqrt_inv)
!
!    END DO
!
!    CALL timestop(handle)
!
!  END SUBROUTINE almo_scf_v_to_v_orthonormal_blk

! **************************************************************************************************
!> \brief orthogonalize MOs
!> \param ket ...
!> \param overlap ...
!> \param metric ...
!> \param retain_locality ...
!> \param only_normalize ...
!> \param eps_filter ...
!> \param order_lanczos ...
!> \param eps_lanczos ...
!> \param max_iter_lanczos ...
!> \par History
!>       2012.03 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE orthogonalize_mos(ket, overlap, metric, retain_locality, only_normalize, &
                                eps_filter, order_lanczos, eps_lanczos, max_iter_lanczos)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: ket, overlap
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: metric
      LOGICAL, INTENT(IN)                                :: retain_locality, only_normalize
      REAL(KIND=dp)                                      :: eps_filter
      INTEGER, INTENT(IN)                                :: order_lanczos
      REAL(KIND=dp), INTENT(IN)                          :: eps_lanczos
      INTEGER, INTENT(IN)                                :: max_iter_lanczos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'orthogonalize_mos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dim0, handle
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: diagonal
      TYPE(cp_dbcsr_type)                                :: matrix_sigma_blk_sqrt, &
                                                            matrix_sigma_blk_sqrt_inv, &
                                                            matrix_t_blk_tmp

      CALL timeset(routineN, handle)

      ! create block-diagonal sparsity pattern for the overlap
      ! in case retain_locality is set to true
      ! RZK-warning this will fail if distribution blocks are smaller than domains!!!
      CALL cp_dbcsr_set(overlap, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(overlap, 1.0_dp)
      CALL cp_dbcsr_filter(overlap, eps_filter)

      CALL get_overlap(ket, ket, overlap, metric, retain_locality, &
                       eps_filter)

      IF (only_normalize) THEN

         CALL cp_dbcsr_get_info(overlap, nfullrows_total=dim0)
         ALLOCATE (diagonal(dim0))
         CALL cp_dbcsr_get_diag(overlap, diagonal)
         CALL cp_dbcsr_set(overlap, 0.0_dp)
         CALL cp_dbcsr_set_diag(overlap, diagonal)
         DEALLOCATE (diagonal)
         CALL cp_dbcsr_filter(overlap, eps_filter)

      ENDIF

      CALL cp_dbcsr_init(matrix_sigma_blk_sqrt)
      CALL cp_dbcsr_init(matrix_sigma_blk_sqrt_inv)
      CALL cp_dbcsr_create(matrix_sigma_blk_sqrt, template=overlap, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_create(matrix_sigma_blk_sqrt_inv, template=overlap, &
                           matrix_type=dbcsr_type_no_symmetry)

      ! compute sqrt and sqrt_inv of the blocked MO overlap
      CALL matrix_sqrt_Newton_Schulz(matrix_sigma_blk_sqrt, matrix_sigma_blk_sqrt_inv, &
                                     overlap, threshold=eps_filter, &
                                     order=order_lanczos, &
                                     eps_lanczos=eps_lanczos, &
                                     max_iter_lanczos=max_iter_lanczos)

      CALL cp_dbcsr_init(matrix_t_blk_tmp)
      CALL cp_dbcsr_create(matrix_t_blk_tmp, &
                           template=ket, &
                           matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             ket, &
                             matrix_sigma_blk_sqrt_inv, &
                             0.0_dp, matrix_t_blk_tmp, &
                             filter_eps=eps_filter)
      !retain_sparsity=retain_locality,&

      ! update the orbitals with the orthonormalized MOs
      CALL cp_dbcsr_copy(ket, matrix_t_blk_tmp)

      CALL cp_dbcsr_release(matrix_t_blk_tmp)
      CALL cp_dbcsr_release(matrix_sigma_blk_sqrt)
      CALL cp_dbcsr_release(matrix_sigma_blk_sqrt_inv)

      CALL timestop(handle)

   END SUBROUTINE orthogonalize_mos

! **************************************************************************************************
!> \brief orthogonalize ALMOs within a domain (obsolete, use orthogonalize_mos)
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_t_blk_to_t_blk_orthonormal(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_t_blk_to_t_blk_orthonormal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      TYPE(cp_dbcsr_type)                                :: matrix_sigma_blk_sqrt, &
                                                            matrix_sigma_blk_sqrt_inv, &
                                                            matrix_t_blk_tmp

      CALL timeset(routineN, handle)

      DO ispin = 1, almo_scf_env%nspins

         CALL cp_dbcsr_init(matrix_t_blk_tmp)
         CALL cp_dbcsr_create(matrix_t_blk_tmp, &
                              template=almo_scf_env%matrix_t_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)

         ! TMP_blk=S_blk.T_blk
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_blk(1), &
                                almo_scf_env%matrix_t_blk(ispin), &
                                0.0_dp, matrix_t_blk_tmp, &
                                filter_eps=almo_scf_env%eps_filter)

         ! Sig_blk=tr(T_blk).TMP_blk - get blocked MO overlap
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                almo_scf_env%matrix_t_blk(ispin), &
                                matrix_t_blk_tmp, &
                                0.0_dp, almo_scf_env%matrix_sigma_blk(ispin), &
                                filter_eps=almo_scf_env%eps_filter, &
                                retain_sparsity=.TRUE.)

         ! RZK-warning try to use symmetry of the sqrt and sqrt_inv matrices
         CALL cp_dbcsr_init(matrix_sigma_blk_sqrt)
         CALL cp_dbcsr_init(matrix_sigma_blk_sqrt_inv)
         CALL cp_dbcsr_create(matrix_sigma_blk_sqrt, template=almo_scf_env%matrix_sigma_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(matrix_sigma_blk_sqrt_inv, template=almo_scf_env%matrix_sigma_blk(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)

         ! compute sqrt and sqrt_inv of the blocked MO overlap
         CALL matrix_sqrt_Newton_Schulz(matrix_sigma_blk_sqrt, matrix_sigma_blk_sqrt_inv, &
                                        almo_scf_env%matrix_sigma_blk(ispin), &
                                        threshold=almo_scf_env%eps_filter, &
                                        order=almo_scf_env%order_lanczos, &
                                        eps_lanczos=almo_scf_env%eps_lanczos, &
                                        max_iter_lanczos=almo_scf_env%max_iter_lanczos)

         ! TMP_blk=T_blk.SigSQRTInv_blk
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                almo_scf_env%matrix_t_blk(ispin), &
                                matrix_sigma_blk_sqrt_inv, &
                                0.0_dp, matrix_t_blk_tmp, &
                                filter_eps=almo_scf_env%eps_filter, &
                                retain_sparsity=.TRUE.)

         ! update the orbitals with the orthonormalized ALMOs
         CALL cp_dbcsr_copy(almo_scf_env%matrix_t_blk(ispin), matrix_t_blk_tmp, &
                            keep_sparsity=.TRUE.)

         CALL cp_dbcsr_release(matrix_t_blk_tmp)
         CALL cp_dbcsr_release(matrix_sigma_blk_sqrt)
         CALL cp_dbcsr_release(matrix_sigma_blk_sqrt_inv)

      END DO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_t_blk_to_t_blk_orthonormal

! **************************************************************************************************
!> \brief computes the idempotent density matrix from MOs
!>        MOs can be either orthogonal or non-orthogonal
!> \param t ...
!> \param p ...
!> \param eps_filter ...
!> \param orthog_orbs ...
!> \param s ...
!> \param sigma ...
!> \param sigma_inv ...
!> \param use_guess ...
!> \par History
!>       2011.07 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_t_to_p(t, p, eps_filter, orthog_orbs, s, sigma, sigma_inv, &
                              use_guess)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: t
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: p
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter
      LOGICAL, INTENT(IN)                                :: orthog_orbs
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: s
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: sigma, sigma_inv
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_guess

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_t_to_p', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: use_sigma_inv_guess
      TYPE(cp_dbcsr_type)                                :: t_tmp

      CALL timeset(routineN, handle)

      ! make sure that S, sigma and sigma_inv are present for non-orthogonal orbitals
      IF (.NOT. orthog_orbs) THEN
         IF ((.NOT. PRESENT(s)) .OR. (.NOT. PRESENT(sigma)) .OR. (.NOT. PRESENT(sigma_inv))) THEN
            CALL cp__b("almo_scf_methods.F",1630,"")
         ENDIF
      ENDIF

      use_sigma_inv_guess = .FALSE.
      IF (PRESENT(use_guess)) THEN
         use_sigma_inv_guess = use_guess
      ENDIF

      IF (orthog_orbs) THEN

         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, t, t, &
                                0.0_dp, p, filter_eps=eps_filter)

      ELSE

         CALL cp_dbcsr_init(t_tmp)
         CALL cp_dbcsr_create(t_tmp, template=t)

         ! TMP=S.T
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, s, t, 0.0_dp, t_tmp, &
                                filter_eps=eps_filter)

         ! Sig=tr(T).TMP - get MO overlap
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, t, t_tmp, 0.0_dp, sigma, &
                                filter_eps=eps_filter)

         ! invert MO overlap
         CALL invert_Hotelling( &
            matrix_inverse=sigma_inv, &
            matrix=sigma, &
            use_inv_as_guess=use_sigma_inv_guess, &
            threshold=eps_filter)

         ! TMP=T.SigInv
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, t, sigma_inv, 0.0_dp, t_tmp, &
                                filter_eps=eps_filter)

         ! P=TMP.tr(T_blk)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, t_tmp, t, 0.0_dp, p, &
                                filter_eps=eps_filter)

         CALL cp_dbcsr_release(t_tmp)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE almo_scf_t_to_p

! **************************************************************************************************
!> \brief applies projector to the orbitals
!>        |psi_out> = P |psi_in>   OR   |psi_out> = (1-P) |psi_in>,
!>        where P = |psi_proj> (<psi_proj|psi_roj>)^{-1} <psi_proj|
!> \param psi_in ...
!> \param psi_out ...
!> \param psi_projector ...
!> \param metric ...
!> \param project_out ...
!> \param psi_projector_orthogonal ...
!> \param proj_in_template ...
!> \param eps_filter ...
!> \param sig_inv_projector ...
!> \param sig_inv_template ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE apply_projector(psi_in, psi_out, psi_projector, metric, project_out, &
                              psi_projector_orthogonal, proj_in_template, eps_filter, sig_inv_projector, &
                              sig_inv_template)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: psi_in
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: psi_out
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: psi_projector, metric
      LOGICAL, INTENT(IN)                                :: project_out, psi_projector_orthogonal
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: proj_in_template
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: sig_inv_projector, sig_inv_template

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_projector', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: tmp_no, tmp_ov, tmp_ov2, tmp_sig, &
                                                            tmp_sig_inv

      CALL timeset(routineN, handle)

      ! =S*PSI_proj
      CALL cp_dbcsr_init(tmp_no)
      CALL cp_dbcsr_create(tmp_no, template=psi_projector)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             metric, psi_projector, &
                             0.0_dp, tmp_no, &
                             filter_eps=eps_filter)

      ! =tr(S.PSI_proj)*PSI_in
      CALL cp_dbcsr_init(tmp_ov)
      CALL cp_dbcsr_create(tmp_ov, template=proj_in_template)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                             tmp_no, psi_in, &
                             0.0_dp, tmp_ov, &
                             filter_eps=eps_filter)

      IF (.NOT. psi_projector_orthogonal) THEN
         ! =SigInv_proj*Sigma_OV
         CALL cp_dbcsr_init(tmp_sig_inv)
         CALL cp_dbcsr_init(tmp_ov2)
         CALL cp_dbcsr_create(tmp_ov2, &
                              template=proj_in_template)
         IF (PRESENT(sig_inv_projector)) THEN
            CALL cp_dbcsr_create(tmp_sig_inv, &
                                 template=sig_inv_projector)
            CALL cp_dbcsr_copy(tmp_sig_inv, sig_inv_projector)
         ELSE
            IF (.NOT. PRESENT(sig_inv_template)) THEN
               CALL cp__b("almo_scf_methods.F",1747,"PROGRAMMING ERROR: provide either template or sig_inv")
            ENDIF
            ! compute inverse overlap of the projector orbitals
            CALL cp_dbcsr_init(tmp_sig)
            CALL cp_dbcsr_create(tmp_sig, &
                                 template=sig_inv_template, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                   psi_projector, tmp_no, 0.0_dp, tmp_sig, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_create(tmp_sig_inv, &
                                 template=sig_inv_template, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL invert_Hotelling(tmp_sig_inv, tmp_sig, &
                                  threshold=eps_filter)
            CALL cp_dbcsr_release(tmp_sig)
         ENDIF
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                tmp_sig_inv, tmp_ov, 0.0_dp, tmp_ov2, &
                                filter_eps=eps_filter)
         CALL cp_dbcsr_release(tmp_sig_inv)
         CALL cp_dbcsr_copy(tmp_ov, tmp_ov2)
         CALL cp_dbcsr_release(tmp_ov2)
      ENDIF
      CALL cp_dbcsr_release(tmp_no)

      ! =PSI_proj*TMP_OV
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             psi_projector, tmp_ov, 0.0_dp, psi_out, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(tmp_ov)

      ! V_out=V_in-V_out
      IF (project_out) THEN
         CALL cp_dbcsr_add(psi_out, psi_in, -1.0_dp, +1.0_dp)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE apply_projector

!! **************************************************************************************************
!!> \brief projects the occupied space out from the provided orbitals
!!> \par History
!!>       2011.07 created [Rustam Z Khaliullin]
!!> \author Rustam Z Khaliullin
!! **************************************************************************************************
!  SUBROUTINE almo_scf_p_out_from_v(v_in,v_out,ov_template,ispin,almo_scf_env)
!
!    TYPE(cp_dbcsr_type), INTENT(IN)                :: v_in, ov_template
!    TYPE(cp_dbcsr_type), INTENT(INOUT)             :: v_out
!    INTEGER, INTENT(IN)                            :: ispin
!    TYPE(almo_scf_env_type), INTENT(INOUT)         :: almo_scf_env
!
!    CHARACTER(LEN=*), PARAMETER :: &
!      routineN = 'almo_scf_p_out_from_v', &
!      routineP = moduleN//':'//routineN
!
!    TYPE(cp_dbcsr_type)                      :: tmp_on, tmp_ov, tmp_ov2
!    INTEGER                                  :: handle
!    LOGICAL                                  :: failure
!
!    CALL timeset(routineN,handle)
!
!    ! =tr(T_blk)*S
!    CALL cp_dbcsr_init(tmp_on)
!    CALL cp_dbcsr_create(tmp_on,&
!            template=almo_scf_env%matrix_t_tr(ispin))
!    CALL cp_dbcsr_multiply("T","N",1.0_dp,&
!            almo_scf_env%matrix_t_blk(ispin),&
!            almo_scf_env%matrix_s(1),&
!            0.0_dp,tmp_on,&
!            filter_eps=almo_scf_env%eps_filter)
!
!    ! =tr(T_blk).S*V_in
!    CALL cp_dbcsr_init(tmp_ov)
!    CALL cp_dbcsr_create(tmp_ov,template=ov_template)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!            tmp_on,v_in,0.0_dp,tmp_ov,&
!            filter_eps=almo_scf_env%eps_filter)
!    CALL cp_dbcsr_release(tmp_on)
!
!    ! =SigmaInv*Sigma_OV
!    CALL cp_dbcsr_init(tmp_ov2)
!    CALL cp_dbcsr_create(tmp_ov2,template=ov_template)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!            almo_scf_env%matrix_sigma_inv(ispin),&
!            tmp_ov,0.0_dp,tmp_ov2,&
!            filter_eps=almo_scf_env%eps_filter)
!    CALL cp_dbcsr_release(tmp_ov)
!
!    ! =T_blk*SigmaInv.Sigma_OV
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!            almo_scf_env%matrix_t_blk(ispin),&
!            tmp_ov2,0.0_dp,v_out,&
!            filter_eps=almo_scf_env%eps_filter)
!    CALL cp_dbcsr_release(tmp_ov2)
!
!    ! V_out=V_in-V_out=
!    CALL cp_dbcsr_add(v_out,v_in,-1.0_dp,+1.0_dp)
!
!    CALL timestop(handle)
!
!  END SUBROUTINE almo_scf_p_out_from_v

! **************************************************************************************************
!> \brief computes the idempotent density matrix from ALMOs
!> \param almo_scf_env ...
!> \param use_sigma_inv_guess ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!>       2011.07 converted into a wrapper which calls almo_scf_t_to_p
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_t_blk_to_p(almo_scf_env, use_sigma_inv_guess)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_sigma_inv_guess

      CHARACTER(LEN=*), PARAMETER :: routineN = 'almo_scf_t_blk_to_p', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin
      LOGICAL                                            :: use_guess
      REAL(KIND=dp)                                      :: spin_factor

      CALL timeset(routineN, handle)

      use_guess = .FALSE.
      IF (PRESENT(use_sigma_inv_guess)) THEN
         use_guess = use_sigma_inv_guess
      ENDIF

      DO ispin = 1, almo_scf_env%nspins

         CALL almo_scf_t_to_p(t=almo_scf_env%matrix_t_blk(ispin), &
                              p=almo_scf_env%matrix_p(ispin), &
                              eps_filter=almo_scf_env%eps_filter, &
                              orthog_orbs=.FALSE., &
                              s=almo_scf_env%matrix_s(1), &
                              sigma=almo_scf_env%matrix_sigma(ispin), &
                              sigma_inv=almo_scf_env%matrix_sigma_inv(ispin), &
                              use_guess=use_guess)

         IF (almo_scf_env%nspins == 1) THEN
            spin_factor = 2.0_dp
         ELSE
            spin_factor = 1.0_dp
         ENDIF
         CALL cp_dbcsr_scale(almo_scf_env%matrix_p(ispin), spin_factor)

      END DO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_t_blk_to_p

! **************************************************************************************************
!> \brief computes a unitary matrix from an arbitrary "generator" matrix
!>        U = ( 1 - X + tr(X) ) ( 1 + X - tr(X) )^(-1)
!> \param X ...
!> \param U ...
!> \param eps_filter ...
!> \par History
!>       2011.08 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE generator_to_unitary(X, U, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: X
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: U
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'generator_to_unitary', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      LOGICAL                                            :: safe_mode
      REAL(KIND=dp)                                      :: frob_matrix, frob_matrix_base
      TYPE(cp_dbcsr_type)                                :: delta, t1, t2, tmp1
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      safe_mode = .TRUE.

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL cp_dbcsr_init(t1)
      CALL cp_dbcsr_create(t1, template=X, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(t2)
      CALL cp_dbcsr_create(t2, template=X, &
                           matrix_type=dbcsr_type_no_symmetry)

      ! create antisymmetric Delta = -X + tr(X)
      CALL cp_dbcsr_init(delta)
      CALL cp_dbcsr_create(delta, template=X, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_transposed(delta, X)
! check that transposed is added correctly
      CALL cp_dbcsr_add(delta, X, 1.0_dp, -1.0_dp)

      ! compute (1 - Delta)^(-1)
      CALL cp_dbcsr_add_on_diag(t1, 1.0_dp)
      CALL cp_dbcsr_add(t1, delta, 1.0_dp, -1.0_dp)
      CALL invert_Hotelling(t2, t1, threshold=eps_filter)

      IF (safe_mode) THEN

         CALL cp_dbcsr_init(tmp1)
         CALL cp_dbcsr_create(tmp1, template=X, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, t2, t1, 0.0_dp, tmp1, &
                                filter_eps=eps_filter)
         frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
         CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)
         frob_matrix = cp_dbcsr_frobenius_norm(tmp1)
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *) "Error for (inv(A)*A-I)", frob_matrix/frob_matrix_base
         ENDIF
         CALL cp_dbcsr_release(tmp1)
      ENDIF

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, delta, t2, 0.0_dp, U, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_add(U, t2, 1.0_dp, 1.0_dp)

      IF (safe_mode) THEN

         CALL cp_dbcsr_init(tmp1)
         CALL cp_dbcsr_create(tmp1, template=X, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, U, U, 0.0_dp, tmp1, &
                                filter_eps=eps_filter)
         frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
         CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)
         frob_matrix = cp_dbcsr_frobenius_norm(tmp1)
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *) "Error for (trn(U)*U-I)", frob_matrix/frob_matrix_base
         ENDIF
         CALL cp_dbcsr_release(tmp1)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE generator_to_unitary

! **************************************************************************************************
!> \brief Parallel code for domain specific operations (my_action)
!>         0. out = op1 * in
!>         1. out = in - op2 * op1 * in
!> \param matrix_in ...
!> \param matrix_out ...
!> \param operator1 ...
!> \param operator2 ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \param my_action ...
!> \param filter_eps ...
!> \param matrix_trimmer ...
!> \param use_trimmer ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE apply_domain_operators(matrix_in, matrix_out, operator1, operator2, &
                                     dpattern, map, node_of_domain, my_action, filter_eps, matrix_trimmer, use_trimmer)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_in
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_out
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: operator1
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN), OPTIONAL                            :: operator2
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain
      INTEGER, INTENT(IN)                                :: my_action
      REAL(KIND=dp)                                      :: filter_eps
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: matrix_trimmer
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_trimmer

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_domain_operators', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ndomains
      LOGICAL                                            :: matrix_trimmer_required, my_use_trimmer, &
                                                            operator2_required
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_in, subm_out, subm_temp

      CALL timeset(routineN, handle)

      my_use_trimmer = .FALSE.
      IF (PRESENT(use_trimmer)) THEN
         my_use_trimmer = use_trimmer
      ENDIF

      operator2_required = .FALSE.
      matrix_trimmer_required = .FALSE.

      IF (my_action .EQ. 1) operator2_required = .TRUE.

      IF (my_use_trimmer) THEN
         matrix_trimmer_required = .TRUE.
         CALL cp__b("almo_scf_methods.F",2060,"TRIMMED PROJECTOR DISABLED!")
      ENDIF

      IF (.NOT. PRESENT(operator2) .AND. operator2_required) THEN
         CALL cp__b("almo_scf_methods.F",2064,"SECOND OPERATOR IS REQUIRED")
      ENDIF
      IF (.NOT. PRESENT(matrix_trimmer) .AND. matrix_trimmer_required) THEN
         CALL cp__b("almo_scf_methods.F",2067,"TRIMMER MATRIX IS REQUIRED")
      ENDIF

      ndomains = cp_dbcsr_nblkcols_total(dpattern)

      ALLOCATE (subm_in(ndomains))
      ALLOCATE (subm_temp(ndomains))
      ALLOCATE (subm_out(ndomains))
      !!!TRIM ALLOCATE(subm_trimmer(ndomains))
      CALL init_submatrices(subm_in)
      CALL init_submatrices(subm_temp)
      CALL init_submatrices(subm_out)

      CALL construct_submatrices(matrix_in, subm_in, &
                                 dpattern, map, node_of_domain, select_row)

      !!!TRIM IF (matrix_trimmer_required) THEN
      !!!TRIM    CALL construct_submatrices(matrix_trimmer,subm_trimmer,&
      !!!TRIM            dpattern,map,node_of_domain,select_row)
      !!!TRIM ENDIF

      IF (my_action .EQ. 0) THEN
         ! for example, apply preconditioner
         CALL multiply_submatrices('N', 'N', 1.0_dp, operator1, &
                                   subm_in, 0.0_dp, subm_out)
      ELSE IF (my_action .EQ. 1) THEN
         ! use for projectors
         CALL copy_submatrices(subm_in, subm_out, .TRUE.)
         CALL multiply_submatrices('N', 'N', 1.0_dp, operator1, &
                                   subm_in, 0.0_dp, subm_temp)
         CALL multiply_submatrices('N', 'N', -1.0_dp, operator2, &
                                   subm_temp, 1.0_dp, subm_out)
!GroupID = dbcsr_mp_group(dbcsr_distribution_mp(&
!   cp_dbcsr_distribution(dpattern)))
!CALL print_submatrices(subm_out,GroupID)
      ELSE
         CALL cp__b("almo_scf_methods.F",2103,"ILLEGAL ACTION")
      ENDIF

      CALL construct_dbcsr_from_submatrices(matrix_out, subm_out, dpattern)
      CALL cp_dbcsr_filter(matrix_out, filter_eps)

      CALL release_submatrices(subm_out)
      CALL release_submatrices(subm_temp)
      CALL release_submatrices(subm_in)

      DEALLOCATE (subm_out)
      DEALLOCATE (subm_temp)
      DEALLOCATE (subm_in)

      CALL timestop(handle)

   END SUBROUTINE apply_domain_operators

! **************************************************************************************************
!> \brief Constructs preconditioners for each domain
!>        -1. projected preconditioner
!>         0. simple preconditioner
!> \param matrix_main ...
!> \param subm_s_inv ...
!> \param subm_r_down ...
!> \param matrix_trimmer ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \param preconditioner ...
!> \param use_trimmer ...
!> \param my_action ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_domain_preconditioner(matrix_main, subm_s_inv, &
                                              subm_r_down, matrix_trimmer, dpattern, map, node_of_domain, preconditioner, &
                                              use_trimmer, my_action)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_main
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN), OPTIONAL                            :: subm_s_inv, subm_r_down
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: matrix_trimmer
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: preconditioner
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_trimmer
      INTEGER, INTENT(IN)                                :: my_action

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_domain_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, index1_end, &
                                                            index1_start, n_domain_mos, naos, &
                                                            nblkrows_tot, ndomains, neighbor, row
      INTEGER, DIMENSION(:), POINTER                     :: nmos
      LOGICAL                                            :: matrix_r_required, &
                                                            matrix_s_inv_required, &
                                                            matrix_trimmer_required, my_use_trimmer
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Minv
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_main, subm_tmp, subm_tmp2

      CALL timeset(routineN, handle)

      my_use_trimmer = .FALSE.
      IF (PRESENT(use_trimmer)) THEN
         my_use_trimmer = use_trimmer
      ENDIF

      matrix_s_inv_required = .FALSE.
      matrix_trimmer_required = .FALSE.
      matrix_r_required = .FALSE.

      IF (my_action .EQ. -1) matrix_s_inv_required = .TRUE.
      IF (my_action .EQ. -1) matrix_r_required = .TRUE.
      IF (my_use_trimmer) THEN
         matrix_trimmer_required = .TRUE.
         CALL cp__b("almo_scf_methods.F",2184,"TRIMMED PRECONDITIONER DISABLED!")
      ENDIF

      IF (.NOT. PRESENT(subm_s_inv) .AND. matrix_s_inv_required) THEN
         CALL cp__b("almo_scf_methods.F",2188,"S_inv SUBMATRICES ARE REQUIRED")
      ENDIF
      IF (.NOT. PRESENT(subm_r_down) .AND. matrix_r_required) THEN
         CALL cp__b("almo_scf_methods.F",2191,"R SUBMATRICES ARE REQUIRED")
      ENDIF
      IF (.NOT. PRESENT(matrix_trimmer) .AND. matrix_trimmer_required) THEN
         CALL cp__b("almo_scf_methods.F",2194,"TRIMMER MATRIX IS REQUIRED")
      ENDIF

      ndomains = cp_dbcsr_nblkcols_total(dpattern)
      nblkrows_tot = cp_dbcsr_nblkrows_total(dpattern)
      nmos => cp_dbcsr_col_block_sizes(dpattern)

      ALLOCATE (subm_main(ndomains))
      CALL init_submatrices(subm_main)
      !!!TRIM ALLOCATE(subm_trimmer(ndomains))

      CALL construct_submatrices(matrix_main, subm_main, &
                                 dpattern, map, node_of_domain, select_row_col)

      !!!TRIM IF (matrix_trimmer_required) THEN
      !!!TRIM    CALL construct_submatrices(matrix_trimmer,subm_trimmer,&
      !!!TRIM            dpattern,map,node_of_domain,select_row)
      !!!TRIM ENDIF

      IF (my_action .EQ. -1) THEN
         ! project out the local occupied space
         !tmp=MATMUL(subm_r(idomain)%mdata,Minv)
         !Minv=MATMUL(tmp,subm_main(idomain)%mdata)
         !subm_main(idomain)%mdata=subm_main(idomain)%mdata-&
         !   Minv-TRANSPOSE(Minv)+MATMUL(Minv,TRANSPOSE(tmp))
         ALLOCATE (subm_tmp(ndomains))
         ALLOCATE (subm_tmp2(ndomains))
         CALL init_submatrices(subm_tmp)
         CALL init_submatrices(subm_tmp2)
         CALL multiply_submatrices('N', 'N', 1.0_dp, subm_r_down, &
                                   subm_s_inv, 0.0_dp, subm_tmp)
         CALL multiply_submatrices('N', 'N', 1.0_dp, subm_tmp, &
                                   subm_main, 0.0_dp, subm_tmp2)
         CALL add_submatrices(1.0_dp, subm_main, -1.0_dp, subm_tmp2, 'N')
         CALL add_submatrices(1.0_dp, subm_main, -1.0_dp, subm_tmp2, 'T')
         CALL multiply_submatrices('N', 'T', 1.0_dp, subm_tmp2, &
                                   subm_tmp, 1.0_dp, subm_main)
         CALL release_submatrices(subm_tmp)
         CALL release_submatrices(subm_tmp2)
         DEALLOCATE (subm_tmp2)
         DEALLOCATE (subm_tmp)
      ENDIF

      ! loop over domains - perform inversion
      DO idomain = 1, ndomains

         ! check if the submatrix exists
         IF (subm_main(idomain)%domain .GT. 0) THEN

            ! find sizes of MO submatrices
            IF (idomain .EQ. 1) THEN
               index1_start = 1
            ELSE
               index1_start = map%index1(idomain-1)
            ENDIF
            index1_end = map%index1(idomain)-1

            n_domain_mos = 0
            DO row = index1_start, index1_end
               neighbor = map%pairs(row, 1)
               n_domain_mos = n_domain_mos+nmos(neighbor)
            ENDDO

            naos = subm_main(idomain)%nrows
            !WRITE(*,*) "Domain, mo_self_and_neig, ao_domain: ", idomain, n_domain_mos, naos

            ALLOCATE (Minv(naos, naos))

            !!!TRIM IF (my_use_trimmer) THEN
            !!!TRIM    ! THIS IS SUPER EXPENSIVE (ELIMINATE)
            !!!TRIM    ! trim the main matrix before inverting
            !!!TRIM    ! assume that the trimmer columns are different (i.e. the main matrix is different for each MO)
            !!!TRIM    allocate(tmp(naos,nmos(idomain)))
            !!!TRIM    DO ii=1, nmos(idomain)
            !!!TRIM       ! transform the main matrix using the trimmer for the current MO
            !!!TRIM       DO jj=1, naos
            !!!TRIM          DO kk=1, naos
            !!!TRIM             Mstore(jj,kk)=sumb_main(idomain)%mdata(jj,kk)*&
            !!!TRIM                subm_trimmer(idomain)%mdata(jj,ii)*&
            !!!TRIM                subm_trimmer(idomain)%mdata(kk,ii)
            !!!TRIM          ENDDO
            !!!TRIM       ENDDO
            !!!TRIM       ! invert the main matrix (exclude some eigenvalues, shift some)
            !!!TRIM       CALL pseudo_invert_matrix(A=Mstore,Ainv=Minv,N=naos,method=1,&
            !!!TRIM               !range1_thr=1.0E-9_dp,range2_thr=1.0E-9_dp,&
            !!!TRIM               shift=1.0E-5_dp,&
            !!!TRIM               range1=nmos(idomain),range2=nmos(idomain),&
            !!!TRIM
            !!!TRIM       ! apply the inverted matrix
            !!!TRIM       ! RZK-warning this is only possible when the preconditioner is applied
            !!!TRIM       tmp(:,ii)=MATMUL(Minv,subm_in(idomain)%mdata(:,ii))
            !!!TRIM    ENDDO
            !!!TRIM    subm_out=MATMUL(tmp,sigma)
            !!!TRIM    deallocate(tmp)
            !!!TRIM ELSE
            CALL pseudo_invert_matrix(A=subm_main(idomain)%mdata, Ainv=Minv, N=naos, method=1, &
                                      range1=nmos(idomain), range2=n_domain_mos)
            !!!TRIM ENDIF

            CALL copy_submatrices(subm_main(idomain), preconditioner(idomain), .FALSE.)
            CALL copy_submatrix_data(Minv, preconditioner(idomain))

            DEALLOCATE (Minv)

         ENDIF ! submatrix for the domain exists

      ENDDO ! loop over domains

      CALL release_submatrices(subm_main)
      DEALLOCATE (subm_main)
      !DEALLOCATE(subm_s)
      !DEALLOCATE(subm_r)

      !IF (matrix_r_required) THEN
      !   CALL cp_dbcsr_release(m_tmp_no_1)
      !   CALL cp_dbcsr_release(m_tmp_no_2)
      !   CALL cp_dbcsr_release(matrix_r)
      !ENDIF

      !RZK-warning do we need a barrier here ?

      CALL timestop(handle)

   END SUBROUTINE construct_domain_preconditioner

! **************************************************************************************************
!> \brief Constructs S^(+1/2) and S^(-1/2) submatrices for each domain
!> \param matrix_s ...
!> \param subm_s_sqrt ...
!> \param subm_s_sqrt_inv ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \par History
!>       2013.03 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_domain_s_sqrt(matrix_s, subm_s_sqrt, subm_s_sqrt_inv, &
                                      dpattern, map, node_of_domain)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_s
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: subm_s_sqrt, subm_s_sqrt_inv
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_domain_s_sqrt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, naos, ndomains
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Ssqrt, Ssqrtinv
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_s

      CALL timeset(routineN, handle)

      ndomains = cp_dbcsr_nblkcols_total(dpattern)
      IF(.NOT.(SIZE(subm_s_sqrt) .EQ. ndomains))CALL cp__a("almo_scf_methods.F",2352)
      IF(.NOT.(SIZE(subm_s_sqrt_inv) .EQ. ndomains))CALL cp__a("almo_scf_methods.F",2353)
      ALLOCATE (subm_s(ndomains))
      CALL init_submatrices(subm_s)

      CALL construct_submatrices(matrix_s, subm_s, &
                                 dpattern, map, node_of_domain, select_row_col)

      ! loop over domains - perform inversion
      DO idomain = 1, ndomains

         ! check if the submatrix exists
         IF (subm_s(idomain)%domain .GT. 0) THEN

            naos = subm_s(idomain)%nrows

            ALLOCATE (Ssqrt(naos, naos))
            ALLOCATE (Ssqrtinv(naos, naos))

            CALL matrix_sqrt(A=subm_s(idomain)%mdata, Asqrt=Ssqrt, Asqrtinv=Ssqrtinv, &
                             N=naos)

            CALL copy_submatrices(subm_s(idomain), subm_s_sqrt(idomain), .FALSE.)
            CALL copy_submatrix_data(Ssqrt, subm_s_sqrt(idomain))

            CALL copy_submatrices(subm_s(idomain), subm_s_sqrt_inv(idomain), .FALSE.)
            CALL copy_submatrix_data(Ssqrtinv, subm_s_sqrt_inv(idomain))

            DEALLOCATE (Ssqrtinv)
            DEALLOCATE (Ssqrt)

         ENDIF ! submatrix for the domain exists

      ENDDO ! loop over domains

      CALL release_submatrices(subm_s)
      DEALLOCATE (subm_s)

      CALL timestop(handle)

   END SUBROUTINE construct_domain_s_sqrt

! **************************************************************************************************
!> \brief Constructs S_inv block for each domain
!> \param matrix_s ...
!> \param subm_s_inv ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \par History
!>       2013.02 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_domain_s_inv(matrix_s, subm_s_inv, dpattern, map, &
                                     node_of_domain)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_s
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: subm_s_inv
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_domain_s_inv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, naos, ndomains
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Sinv
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_s

      CALL timeset(routineN, handle)

      ndomains = cp_dbcsr_nblkcols_total(dpattern)

      IF(.NOT.(SIZE(subm_s_inv) .EQ. ndomains))CALL cp__a("almo_scf_methods.F",2427)
      ALLOCATE (subm_s(ndomains))
      CALL init_submatrices(subm_s)

      CALL construct_submatrices(matrix_s, subm_s, &
                                 dpattern, map, node_of_domain, select_row_col)

      !GroupID = dbcsr_mp_group(dbcsr_distribution_mp(&
      !   cp_dbcsr_distribution(dpattern)))
      !CALL print_submatrices(domain_r_down(:,ispin),GroupID)

      ! loop over domains - perform inversion
      DO idomain = 1, ndomains

         ! check if the submatrix exists
         IF (subm_s(idomain)%domain .GT. 0) THEN

            naos = subm_s(idomain)%nrows

            ALLOCATE (Sinv(naos, naos))

            CALL pseudo_invert_matrix(A=subm_s(idomain)%mdata, Ainv=Sinv, N=naos, &
                                      method=0)

            CALL copy_submatrices(subm_s(idomain), subm_s_inv(idomain), .FALSE.)
            CALL copy_submatrix_data(Sinv, subm_s_inv(idomain))

            DEALLOCATE (Sinv)

         ENDIF ! submatrix for the domain exists

      ENDDO ! loop over domains

      CALL release_submatrices(subm_s)
      DEALLOCATE (subm_s)

      CALL timestop(handle)

   END SUBROUTINE construct_domain_s_inv

! **************************************************************************************************
!> \brief Constructs subblocks of the covariant-covariant DM
!> \param matrix_t ...
!> \param matrix_sigma_inv ...
!> \param matrix_s ...
!> \param subm_r_down ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \param filter_eps ...
!> \par History
!>       2013.02 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_domain_r_down(matrix_t, matrix_sigma_inv, matrix_s, &
                                      subm_r_down, dpattern, map, node_of_domain, filter_eps)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_t, matrix_sigma_inv, matrix_s
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: subm_r_down
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain
      REAL(KIND=dp)                                      :: filter_eps

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_domain_r_down', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ndomains
      TYPE(cp_dbcsr_type)                                :: m_tmp_no_1, m_tmp_no_2, matrix_r

      CALL timeset(routineN, handle)

      ! compute the density matrix in the COVARIANT representation
      CALL cp_dbcsr_init(matrix_r)
      CALL cp_dbcsr_create(matrix_r, &
                           template=matrix_s, &
                           matrix_type=dbcsr_type_symmetric)
      CALL cp_dbcsr_init(m_tmp_no_1)
      CALL cp_dbcsr_create(m_tmp_no_1, &
                           template=matrix_t, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(m_tmp_no_2)
      CALL cp_dbcsr_create(m_tmp_no_2, &
                           template=matrix_t, &
                           matrix_type=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s, matrix_t, &
                             0.0_dp, m_tmp_no_1, filter_eps=filter_eps)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, m_tmp_no_1, matrix_sigma_inv, &
                             0.0_dp, m_tmp_no_2, filter_eps=filter_eps)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, m_tmp_no_2, m_tmp_no_1, &
                             0.0_dp, matrix_r, filter_eps=filter_eps)

      CALL cp_dbcsr_release(m_tmp_no_1)
      CALL cp_dbcsr_release(m_tmp_no_2)

      ndomains = cp_dbcsr_nblkcols_total(dpattern)
      IF(.NOT.(SIZE(subm_r_down) .EQ. ndomains))CALL cp__a("almo_scf_methods.F",2525)

      CALL construct_submatrices(matrix_r, subm_r_down, &
                                 dpattern, map, node_of_domain, select_row_col)

      CALL cp_dbcsr_release(matrix_r)

      CALL timestop(handle)

   END SUBROUTINE construct_domain_r_down

! **************************************************************************************************
!> \brief Finds the square root of a matrix and its inverse
!> \param A ...
!> \param Asqrt ...
!> \param Asqrtinv ...
!> \param N ...
!> \par History
!>       2013.03 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_sqrt(A, Asqrt, Asqrtinv, N)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: A
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: Asqrt, Asqrtinv
      INTEGER, INTENT(IN)                                :: N

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_sqrt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, INFO, jj, LWORK
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, WORK
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: test, testN

      CALL timeset(routineN, handle)

      Asqrtinv = A
      INFO = 0

      !CALL DPOTRF('L', N, Ainv, N, INFO )
      !IF( INFO.NE.0 ) THEN
      !   CPErrorMessage(cp_failure_level,routineP,"DPOTRF failed")
      !   CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
      !END IF
      !CALL DPOTRI('L', N, Ainv, N, INFO )
      !IF( INFO.NE.0 ) THEN
      !   CPErrorMessage(cp_failure_level,routineP,"DPOTRI failed")
      !   CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
      !END IF
      !! complete the matrix
      !DO ii=1,N
      !   DO jj=ii+1,N
      !      Ainv(ii,jj)=Ainv(jj,ii)
      !   ENDDO
      !   !WRITE(*,'(100F13.9)') Ainv(ii,:)
      !ENDDO

      ! diagonalize first
      ALLOCATE (eigenvalues(N))
      ! Query the optimal workspace for dsyev
      LWORK = -1
      ALLOCATE (WORK(MAX(1, LWORK)))
      CALL DSYEV('V', 'L', N, Asqrtinv, N, eigenvalues, WORK, LWORK, INFO)
      LWORK = INT(WORK(1))
      DEALLOCATE (WORK)
      ! Allocate the workspace and solve the eigenproblem
      ALLOCATE (WORK(MAX(1, LWORK)))
      CALL DSYEV('V', 'L', N, Asqrtinv, N, eigenvalues, WORK, LWORK, INFO)
      IF (INFO .NE. 0) THEN
         WRITE (*, *) 'DSYEV ERROR MESSAGE: ', INFO
         CALL cp__b("almo_scf_methods.F",2594,"DSYEV failed")
      END IF
      DEALLOCATE (WORK)

      ! take functions of eigenvalues and use eigenvectors to compute the matrix function
      ! first sqrt
      ALLOCATE (test(N, N))
      DO jj = 1, N
         test(jj, :) = Asqrtinv(:, jj)*SQRT(eigenvalues(jj))
      ENDDO
      ALLOCATE (testN(N, N))
      testN(:, :) = MATMUL(Asqrtinv, test)
      Asqrt = testN
      ! now, sqrt_inv
      DO jj = 1, N
         test(jj, :) = Asqrtinv(:, jj)/SQRT(eigenvalues(jj))
      ENDDO
      testN(:, :) = MATMUL(Asqrtinv, test)
      Asqrtinv = testN
      DEALLOCATE (test, testN)

      DEALLOCATE (eigenvalues)

!!!    ! compute the error
!!!    allocate(test(N,N))
!!!    test=MATMUL(Ainv,A)
!!!    DO ii=1,N
!!!       test(ii,ii)=test(ii,ii)-1.0_dp
!!!    ENDDO
!!!    test_error=0.0_dp
!!!    DO ii=1,N
!!!       DO jj=1,N
!!!          test_error=test_error+test(jj,ii)*test(jj,ii)
!!!       ENDDO
!!!    ENDDO
!!!    WRITE(*,*) "Inversion error: ", SQRT(test_error)
!!!    deallocate(test)

      CALL timestop(handle)

   END SUBROUTINE matrix_sqrt

! **************************************************************************************************
!> \brief Inverts a matrix using a requested method
!>         0. Cholesky factorization
!>         1. Diagonalization
!> \param A ...
!> \param Ainv ...
!> \param N ...
!> \param method ...
!> \param range1 ...
!> \param range2 ...
!> \param range1_thr ...
!> \param range2_thr ...
!> \param shift ...
!> \par History
!>       2012.04 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE pseudo_invert_matrix(A, Ainv, N, method, range1, range2, range1_thr, range2_thr, &
                                   shift)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: A
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: Ainv
      INTEGER, INTENT(IN)                                :: N, method
      INTEGER, INTENT(IN), OPTIONAL                      :: range1, range2
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: range1_thr, range2_thr, shift

      CHARACTER(len=*), PARAMETER :: routineN = 'pseudo_invert_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ii, INFO, jj, LWORK, range1_eiv, &
                                                            range2_eiv, range3_eiv
      LOGICAL                                            :: use_ranges
      REAL(KIND=dp)                                      :: my_shift
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, WORK
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: test, testN

      CALL timeset(routineN, handle)

      IF (method .EQ. 1) THEN
         IF (PRESENT(range1)) THEN
            use_ranges = .TRUE.
            IF (.NOT. PRESENT(range2)) THEN
               CALL cp__b("almo_scf_methods.F",2678,"SPECIFY TWO RANGES")
            ENDIF
         ELSE
            use_ranges = .FALSE.
            IF ((.NOT. PRESENT(range1_thr)) .OR. (.NOT. PRESENT(range2_thr))) THEN
               CALL cp__b("almo_scf_methods.F",2683,"SPECIFY TWO THRESHOLDS")
            ENDIF
         ENDIF
      ENDIF

      my_shift = 0.0_dp
      IF (PRESENT(shift)) THEN
         my_shift = shift
      ENDIF

      Ainv = A
      INFO = 0

      SELECT CASE (method)
      CASE (0)

         CALL DPOTRF('L', N, Ainv, N, INFO)
         IF (INFO .NE. 0) THEN
            CALL cp__b("almo_scf_methods.F",2701,"DPOTRF failed")
         END IF
         CALL DPOTRI('L', N, Ainv, N, INFO)
         IF (INFO .NE. 0) THEN
            CALL cp__b("almo_scf_methods.F",2705,"DPOTRI failed")
         END IF
         ! complete the matrix
         DO ii = 1, N
            DO jj = ii+1, N
               Ainv(ii, jj) = Ainv(jj, ii)
            ENDDO
            !WRITE(*,'(100F13.9)') Ainv(ii,:)
         ENDDO

      CASE (1)

         ! diagonalize first
         ALLOCATE (eigenvalues(N))
         ! Query the optimal workspace for dsyev
         LWORK = -1
         ALLOCATE (WORK(MAX(1, LWORK)))
         CALL DSYEV('V', 'L', N, Ainv, N, eigenvalues, WORK, LWORK, INFO)
         LWORK = INT(WORK(1))
         DEALLOCATE (WORK)
         ! Allocate the workspace and solve the eigenproblem
         ALLOCATE (WORK(MAX(1, LWORK)))
         CALL DSYEV('V', 'L', N, Ainv, N, eigenvalues, WORK, LWORK, INFO)
         IF (INFO .NE. 0) THEN
            WRITE (*, *) 'DSYEV ERROR MESSAGE: ', INFO
            CALL cp__b("almo_scf_methods.F",2730,"DSYEV failed")
         END IF
         DEALLOCATE (WORK)

         !WRITE(*,*) "EIGENVALS: "
         !WRITE(*,'(4F13.9)') eigenvalues(:)

         ! invert eigenvalues and use eigenvectors to compute the Hessian inverse
         ! project out zero-eigenvalue directions
         ALLOCATE (test(N, N))
         range1_eiv = 0
         range2_eiv = 0
         range3_eiv = 0
         IF (use_ranges) THEN
            DO jj = 1, N
               IF (jj .LE. range1) THEN
                  test(jj, :) = Ainv(:, jj)*0.0_dp
                  range1_eiv = range1_eiv+1
               ELSE IF (jj .LE. range2) THEN
                  test(jj, :) = Ainv(:, jj)*1.0_dp
                  range2_eiv = range2_eiv+1
               ELSE
                  test(jj, :) = Ainv(:, jj)/(eigenvalues(jj)+my_shift)
                  range3_eiv = range3_eiv+1
               ENDIF
            ENDDO
         ELSE
            DO jj = 1, N
               IF (eigenvalues(jj) .LT. range1_thr) THEN
                  test(jj, :) = Ainv(:, jj)*0.0_dp
                  range1_eiv = range1_eiv+1
               ELSE IF (eigenvalues(jj) .LT. range2_thr) THEN
                  test(jj, :) = Ainv(:, jj)*1.0_dp
                  range2_eiv = range2_eiv+1
               ELSE
                  test(jj, :) = Ainv(:, jj)/(eigenvalues(jj)+my_shift)
                  range3_eiv = range3_eiv+1
               ENDIF
            ENDDO
         ENDIF
         !WRITE(*,*) ' EIV RANGES: ', range1_eiv, range2_eiv, range3_eiv
         ALLOCATE (testN(N, N))
         testN(:, :) = MATMUL(Ainv, test)
         Ainv = testN
         DEALLOCATE (test, testN)
         DEALLOCATE (eigenvalues)

      CASE DEFAULT

         CALL cp__b("almo_scf_methods.F",2779,"Illegal method selected for matrix inversion")

      END SELECT

!!!    ! compute the inversion error
!!!    allocate(test(N,N))
!!!    test=MATMUL(Ainv,A)
!!!    DO ii=1,N
!!!       test(ii,ii)=test(ii,ii)-1.0_dp
!!!    ENDDO
!!!    test_error=0.0_dp
!!!    DO ii=1,N
!!!       DO jj=1,N
!!!          test_error=test_error+test(jj,ii)*test(jj,ii)
!!!       ENDDO
!!!    ENDDO
!!!    WRITE(*,*) "Inversion error: ", SQRT(test_error)
!!!    deallocate(test)

      CALL timestop(handle)

   END SUBROUTINE pseudo_invert_matrix

! **************************************************************************************************
!> \brief computes the step matrix from the gradient and Hessian using
!>         the Newton-Raphson method
!> \param matrix_grad ...
!> \param matrix_step ...
!> \param matrix_s ...
!> \param matrix_ks ...
!> \param matrix_t ...
!> \param matrix_sigma_inv ...
!> \param quench_t ...
!> \param spin_factor ...
!> \param eps_filter ...
!> \par History
!>       2012.02 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE newton_grad_to_step(matrix_grad, matrix_step, matrix_s, matrix_ks, &
                                  matrix_t, matrix_sigma_inv, quench_t, spin_factor, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_grad, matrix_step, matrix_s
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_ks, matrix_t
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_sigma_inv, quench_t
      REAL(KIND=dp), INTENT(IN)                          :: spin_factor, eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'newton_grad_to_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: res_norm
      TYPE(cp_dbcsr_type)                                :: m_tmp_no_1, m_tmp_no_3, m_tmp_oo_2, &
                                                            matrix_f_ao, matrix_f_mo, matrix_f_vo, &
                                                            matrix_s_ao, matrix_s_mo, matrix_s_vo

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(matrix_s_ao)
      CALL cp_dbcsr_create(matrix_s_ao, &
                           template=matrix_s, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_f_ao)
      CALL cp_dbcsr_create(matrix_f_ao, &
                           template=matrix_s, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_f_mo)
      CALL cp_dbcsr_create(matrix_f_mo, &
                           template=matrix_sigma_inv, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_s_mo)
      CALL cp_dbcsr_create(matrix_s_mo, &
                           template=matrix_sigma_inv, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_f_vo)
      CALL cp_dbcsr_create(matrix_f_vo, &
                           template=matrix_t)
      CALL cp_dbcsr_init(matrix_s_vo)
      CALL cp_dbcsr_create(matrix_s_vo, &
                           template=matrix_t)

      CALL cp_dbcsr_init(m_tmp_no_1)
      CALL cp_dbcsr_create(m_tmp_no_1, &
                           template=matrix_t)
      CALL cp_dbcsr_init(m_tmp_no_3)
      CALL cp_dbcsr_create(m_tmp_no_3, &
                           template=matrix_t)
      CALL cp_dbcsr_init(m_tmp_oo_2)
      CALL cp_dbcsr_create(m_tmp_oo_2, &
                           template=matrix_sigma_inv, &
                           matrix_type=dbcsr_type_no_symmetry)

      ! calculate S-SRS and (1-R)F(1-R)
      ! RZK-warning some optimization is ABSOLUTELY NECESSARY
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_s, &
                             matrix_t, &
                             0.0_dp, m_tmp_no_1, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             m_tmp_no_1, &
                             matrix_sigma_inv, &
                             0.0_dp, matrix_s_vo, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_desymmetrize(matrix_s, &
                                 matrix_s_ao)
      CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                             m_tmp_no_1, &
                             matrix_s_vo, &
                             1.0_dp, matrix_s_ao, &
                             filter_eps=eps_filter)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_ks, &
                             matrix_t, &
                             0.0_dp, m_tmp_no_1, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_desymmetrize(matrix_ks, matrix_f_ao)
      CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                             m_tmp_no_1, &
                             matrix_s_vo, &
                             1.0_dp, matrix_f_ao, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                             matrix_s_vo, &
                             m_tmp_no_1, &
                             1.0_dp, matrix_f_ao, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                             matrix_t, &
                             m_tmp_no_1, &
                             0.0_dp, matrix_f_mo, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_s_vo, &
                             matrix_f_mo, &
                             0.0_dp, m_tmp_no_1, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                             m_tmp_no_1, &
                             matrix_s_vo, &
                             1.0_dp, matrix_f_ao, &
                             filter_eps=eps_filter)

      ! calculate F_mo
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_sigma_inv, &
                             matrix_f_mo, &
                             0.0_dp, m_tmp_oo_2, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             m_tmp_oo_2, &
                             matrix_sigma_inv, &
                             1.0_dp, matrix_f_mo, &
                             filter_eps=eps_filter)

      ! calculate F_vo
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_ks, &
                             matrix_t, &
                             0.0_dp, m_tmp_no_1, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             m_tmp_no_1, &
                             matrix_sigma_inv, &
                             0.0_dp, matrix_f_vo, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                             matrix_s_vo, &
                             m_tmp_oo_2, &
                             1.0_dp, matrix_f_vo, &
                             filter_eps=eps_filter)

      CALL cp_dbcsr_desymmetrize(matrix_sigma_inv, matrix_s_mo)

      !!! RZK-warning: this is HIGHLIGHTED BLOCK
      !!! check it first if the procedure does not function as it supposed to
      !CALL cp_dbcsr_desymmetrize(matrix_s,matrix_s_ao)
      CALL cp_dbcsr_add(matrix_f_ao, matrix_s_ao, 1.0_dp, 1.0_dp)

      CALL cp_dbcsr_set(matrix_f_mo, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(matrix_f_mo, 1.0_dp)
      CALL cp_dbcsr_filter(matrix_f_mo, eps_filter)

      CALL cp_dbcsr_set(matrix_s_mo, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(matrix_s_mo, 1.0_dp)
      CALL cp_dbcsr_filter(matrix_s_mo, eps_filter)

      CALL cp_dbcsr_set(matrix_s_ao, 0.0_dp)
      !CALL cp_dbcsr_add_on_diag(matrix_s_ao,1.0_dp)
      CALL cp_dbcsr_filter(matrix_s_ao, eps_filter)
      !!! RZK-warning: end of HIGHLIGHTED BLOCK

      CALL cp_dbcsr_scale(matrix_f_ao, &
                          2.0_dp*spin_factor)
      CALL cp_dbcsr_scale(matrix_s_ao, &
                          -2.0_dp*spin_factor)
      CALL cp_dbcsr_scale(matrix_f_vo, &
                          2.0_dp*spin_factor)

      !WRITE(*,*) "INSIDE newton_grad_to_step: "
      !CALL cp_dbcsr_print(matrix_s_mo)
      !CALL cp_dbcsr_print(matrix_ks)
      !CALL cp_dbcsr_print(matrix_s)
      !CALL cp_dbcsr_print(matrix_sigma_inv)

      CALL hessian_diag_apply(matrix_grad, matrix_step, matrix_s_ao, &
                              matrix_f_ao, matrix_s_mo, matrix_f_mo, &
                              matrix_s_vo, matrix_f_vo, quench_t)

      ! check that the step satisfies H.step=-grad
      CALL cp_dbcsr_copy(m_tmp_no_3, quench_t)
      CALL cp_dbcsr_copy(m_tmp_no_1, quench_t)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_f_ao, &
                             matrix_step, &
                             0.0_dp, m_tmp_no_1, &
                             !retain_sparsity=.TRUE.,&
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             m_tmp_no_1, &
                             matrix_s_mo, &
                             0.0_dp, m_tmp_no_3, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_copy(m_tmp_no_1, quench_t)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             matrix_s_ao, &
                             matrix_step, &
                             0.0_dp, m_tmp_no_1, &
                             !retain_sparsity=.TRUE.,&
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                             m_tmp_no_1, &
                             matrix_f_mo, &
                             1.0_dp, m_tmp_no_3, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_add(m_tmp_no_3, matrix_grad, &
                        1.0_dp, 1.0_dp)
      CALL cp_dbcsr_norm(m_tmp_no_3, &
                         dbcsr_norm_maxabsnorm, norm_scalar=res_norm)
      WRITE (*, *) "NEWTON step error: ", res_norm

      CALL cp_dbcsr_release(m_tmp_no_3)
      CALL cp_dbcsr_release(m_tmp_no_1)
      CALL cp_dbcsr_release(m_tmp_oo_2)
      CALL cp_dbcsr_release(matrix_s_ao)
      CALL cp_dbcsr_release(matrix_s_mo)
      CALL cp_dbcsr_release(matrix_f_ao)
      CALL cp_dbcsr_release(matrix_f_mo)
      CALL cp_dbcsr_release(matrix_s_vo)
      CALL cp_dbcsr_release(matrix_f_vo)

      CALL timestop(handle)

   END SUBROUTINE newton_grad_to_step

! **************************************************************************************************
!> \brief Serial code that constructs an approximate Hessian
!> \param matrix_grad ...
!> \param matrix_step ...
!> \param matrix_S_ao ...
!> \param matrix_F_ao ...
!> \param matrix_S_mo ...
!> \param matrix_F_mo ...
!> \param matrix_S_vo ...
!> \param matrix_F_vo ...
!> \param quench_t ...
!> \par History
!>       2012.02 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE hessian_diag_apply(matrix_grad, matrix_step, matrix_S_ao, &
                                 matrix_F_ao, matrix_S_mo, matrix_F_mo, matrix_S_vo, matrix_F_vo, quench_t)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_grad, matrix_step, matrix_S_ao, &
                                                            matrix_F_ao, matrix_S_mo, matrix_F_mo, &
                                                            matrix_S_vo, matrix_F_vo, quench_t

      CHARACTER(len=*), PARAMETER :: routineN = 'hessian_diag_apply', &
         routineP = moduleN//':'//routineN

      INTEGER :: ao_hori_offset, ao_vert_offset, block_col, block_row, col, copy, H_size, handle, &
         ii, INFO, jj, lev1_hori_offset, lev1_vert_offset, lev2_hori_offset, lev2_vert_offset, &
         LWORK, nblkcols_tot, nblkrows_tot, orb_i, orb_j, row, zero_neg_eiv
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ao_block_sizes, ao_domain_sizes, &
                                                            mo_block_sizes
      INTEGER, DIMENSION(:), POINTER                     :: ao_blk_sizes, mo_blk_sizes
      LOGICAL                                            :: found, found2, found_col, found_row
      REAL(KIND=dp)                                      :: test_error
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, Grad_vec, Step_vec, tmp, &
                                                            tmpr, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: F_ao_block, F_mo_block, H, H1, H2, Hinv, &
                                                            S_ao_block, S_mo_block, test, test2
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: block_p, block_p2, p_new_block
      TYPE(cp_dbcsr_type)                                :: matrix_F_ao_sym, matrix_F_mo_sym, &
                                                            matrix_S_ao_sym, matrix_S_mo_sym

      CALL timeset(routineN, handle)

      nblkrows_tot = cp_dbcsr_nblkrows_total(quench_t)
      nblkcols_tot = cp_dbcsr_nblkcols_total(quench_t)
      IF(.NOT.(nblkrows_tot == nblkcols_tot))CALL cp__a("almo_scf_methods.F",3082)
      mo_blk_sizes => cp_dbcsr_col_block_sizes(quench_t)
      ao_blk_sizes => cp_dbcsr_row_block_sizes(quench_t)
      ALLOCATE (mo_block_sizes(nblkcols_tot), ao_block_sizes(nblkcols_tot))
      ALLOCATE (ao_domain_sizes(nblkcols_tot))
      mo_block_sizes(:) = mo_blk_sizes(:)
      ao_block_sizes(:) = ao_blk_sizes(:)
      ao_domain_sizes(:) = 0

      CALL cp_dbcsr_init(matrix_S_ao_sym)
      CALL cp_dbcsr_create(matrix_S_ao_sym, &
                           template=matrix_S_ao, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_desymmetrize(matrix_S_ao, matrix_S_ao_sym)

      CALL cp_dbcsr_init(matrix_F_ao_sym)
      CALL cp_dbcsr_create(matrix_F_ao_sym, &
                           template=matrix_F_ao, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_desymmetrize(matrix_F_ao, matrix_F_ao_sym)

      CALL cp_dbcsr_init(matrix_S_mo_sym)
      CALL cp_dbcsr_create(matrix_S_mo_sym, &
                           template=matrix_S_mo, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_desymmetrize(matrix_S_mo, matrix_S_mo_sym)

      CALL cp_dbcsr_init(matrix_F_mo_sym)
      CALL cp_dbcsr_create(matrix_F_mo_sym, &
                           template=matrix_F_mo, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_desymmetrize(matrix_F_mo, matrix_F_mo_sym)

      !CALL cp_dbcsr_print(matrix_grad)
      !CALL cp_dbcsr_print(matrix_F_ao_sym)
      !CALL cp_dbcsr_print(matrix_S_ao_sym)
      !CALL cp_dbcsr_print(matrix_F_mo_sym)
      !CALL cp_dbcsr_print(matrix_S_mo_sym)

      ! loop over domains to find the size of the Hessian
      H_size = 0
      DO col = 1, nblkcols_tot

         ! find sizes of AO submatrices
         DO row = 1, nblkrows_tot

            CALL cp_dbcsr_get_block_p(quench_t, &
                                      row, col, block_p, found)
            IF (found) THEN
               ao_domain_sizes(col) = ao_domain_sizes(col)+ao_blk_sizes(row)
            ENDIF

         ENDDO

         H_size = H_size+ao_domain_sizes(col)*mo_block_sizes(col)

      ENDDO

      ALLOCATE (H(H_size, H_size))

      ! fill the Hessian matrix
      lev1_vert_offset = 0
      ! loop over all pairs of fragments
      DO row = 1, nblkcols_tot

         lev1_hori_offset = 0
         DO col = 1, nblkcols_tot

            ! prepare blocks for the current row-column fragment pair
            ALLOCATE (F_ao_block(ao_domain_sizes(row), ao_domain_sizes(col)))
            ALLOCATE (S_ao_block(ao_domain_sizes(row), ao_domain_sizes(col)))
            ALLOCATE (F_mo_block(mo_block_sizes(row), mo_block_sizes(col)))
            ALLOCATE (S_mo_block(mo_block_sizes(row), mo_block_sizes(col)))

            F_ao_block(:, :) = 0.0_dp
            S_ao_block(:, :) = 0.0_dp
            F_mo_block(:, :) = 0.0_dp
            S_mo_block(:, :) = 0.0_dp

            ! fill AO submatrices
            ! loop over all blocks of the AO dbcsr matrix
            ao_vert_offset = 0
            DO block_row = 1, nblkcols_tot

               CALL cp_dbcsr_get_block_p(quench_t, &
                                         block_row, row, block_p, found_row)
               IF (found_row) THEN

                  ao_hori_offset = 0
                  DO block_col = 1, nblkcols_tot

                     CALL cp_dbcsr_get_block_p(quench_t, &
                                               block_col, col, block_p, found_col)
                     IF (found_col) THEN

                        CALL cp_dbcsr_get_block_p(matrix_F_ao_sym, &
                                                  block_row, block_col, block_p, found)
                        IF (found) THEN
                           ! copy the block into the submatrix
                           F_ao_block(ao_vert_offset+1:ao_vert_offset+ao_block_sizes(block_row), &
                                      ao_hori_offset+1:ao_hori_offset+ao_block_sizes(block_col)) &
                              = block_p(:, :)
                        ENDIF

                        CALL cp_dbcsr_get_block_p(matrix_S_ao_sym, &
                                                  block_row, block_col, block_p, found)
                        IF (found) THEN
                           ! copy the block into the submatrix
                           S_ao_block(ao_vert_offset+1:ao_vert_offset+ao_block_sizes(block_row), &
                                      ao_hori_offset+1:ao_hori_offset+ao_block_sizes(block_col)) &
                              = block_p(:, :)
                        ENDIF

                        ao_hori_offset = ao_hori_offset+ao_block_sizes(block_col)

                     ENDIF

                  ENDDO

                  ao_vert_offset = ao_vert_offset+ao_block_sizes(block_row)

               ENDIF

            ENDDO

            ! fill MO submatrices
            CALL cp_dbcsr_get_block_p(matrix_F_mo_sym, row, col, block_p, found)
            IF (found) THEN
               ! copy the block into the submatrix
               F_mo_block(1:mo_block_sizes(row), 1:mo_block_sizes(col)) = block_p(:, :)
            ENDIF
            CALL cp_dbcsr_get_block_p(matrix_S_mo_sym, row, col, block_p, found)
            IF (found) THEN
               ! copy the block into the submatrix
               S_mo_block(1:mo_block_sizes(row), 1:mo_block_sizes(col)) = block_p(:, :)
            ENDIF

            !WRITE(*,*) "F_AO_BLOCK", row, col, ao_domain_sizes(row), ao_domain_sizes(col)
            !DO ii=1,ao_domain_sizes(row)
            !  WRITE(*,'(100F13.9)') F_ao_block(ii,:)
            !ENDDO
            !WRITE(*,*) "S_AO_BLOCK", row, col
            !DO ii=1,ao_domain_sizes(row)
            !  WRITE(*,'(100F13.9)') S_ao_block(ii,:)
            !ENDDO
            !WRITE(*,*) "F_MO_BLOCK", row, col
            !DO ii=1,mo_block_sizes(row)
            !  WRITE(*,'(100F13.9)') F_mo_block(ii,:)
            !ENDDO
            !WRITE(*,*) "S_MO_BLOCK", row, col, mo_block_sizes(row), mo_block_sizes(col)
            !DO ii=1,mo_block_sizes(row)
            !  WRITE(*,'(100F13.9)') S_mo_block(ii,:)
            !ENDDO

            ! construct tensor products for the current row-column fragment pair
            lev2_vert_offset = 0
            DO orb_j = 1, mo_block_sizes(row)

               lev2_hori_offset = 0
               DO orb_i = 1, mo_block_sizes(col)

                  H(lev1_vert_offset+lev2_vert_offset+1:lev1_vert_offset+lev2_vert_offset+ao_domain_sizes(row), &
                    lev1_hori_offset+lev2_hori_offset+1:lev1_hori_offset+lev2_hori_offset+ao_domain_sizes(col)) &
                     = S_mo_block(orb_j, orb_i)*F_ao_block(:, :) &
                       -F_mo_block(orb_j, orb_i)*S_ao_block(:, :)

                  !WRITE(*,*) row, col, orb_j, orb_i, lev1_vert_offset+lev2_vert_offset+1, ao_domain_sizes(row),&
                  !   lev1_hori_offset+lev2_hori_offset+1, ao_domain_sizes(col), S_mo_block(orb_j,orb_i)

                  lev2_hori_offset = lev2_hori_offset+ao_domain_sizes(col)

               ENDDO

               lev2_vert_offset = lev2_vert_offset+ao_domain_sizes(row)

            ENDDO

            lev1_hori_offset = lev1_hori_offset+ao_domain_sizes(col)*mo_block_sizes(col)

            DEALLOCATE (F_ao_block)
            DEALLOCATE (S_ao_block)
            DEALLOCATE (F_mo_block)
            DEALLOCATE (S_mo_block)

         ENDDO ! col fragment

         lev1_vert_offset = lev1_vert_offset+ao_domain_sizes(row)*mo_block_sizes(row)

      ENDDO ! row fragment

      CALL cp_dbcsr_release(matrix_S_ao_sym)
      CALL cp_dbcsr_release(matrix_F_ao_sym)
      CALL cp_dbcsr_release(matrix_S_mo_sym)
      CALL cp_dbcsr_release(matrix_F_mo_sym)

      ! two more terms of the Hessian
      ALLOCATE (H1(H_size, H_size))
      ALLOCATE (H2(H_size, H_size))
      H1 = 0.0_dp
      H2 = 0.0_dp
      DO row = 1, nblkcols_tot

         lev1_hori_offset = 0
         DO col = 1, nblkcols_tot

            CALL cp_dbcsr_get_block_p(matrix_F_vo, &
                                      row, col, block_p, found)
            CALL cp_dbcsr_get_block_p(matrix_S_vo, &
                                      row, col, block_p2, found2)

            lev1_vert_offset = 0
            DO block_col = 1, nblkcols_tot

               CALL cp_dbcsr_get_block_p(quench_t, &
                                         row, block_col, p_new_block, found_row)

               IF (found_row) THEN

                  ! determine offset in this short loop
                  lev2_vert_offset = 0
                  DO block_row = 1, row-1
                     CALL cp_dbcsr_get_block_p(quench_t, &
                                               block_row, block_col, p_new_block, found_col)
                     IF (found_col) lev2_vert_offset = lev2_vert_offset+ao_block_sizes(block_row)
                  ENDDO
                  !!!!!!!! short loop

                  ! over all electrons of the block
                  DO orb_i = 1, mo_block_sizes(col)

                     ! into all possible locations
                     DO orb_j = 1, mo_block_sizes(block_col)

                        ! column is copied several times
                        DO copy = 1, ao_domain_sizes(col)

                           IF (found) THEN

                              !WRITE(*,*) row, col, block_col, orb_i, orb_j, copy,&
                              ! lev1_vert_offset+(orb_j-1)*ao_domain_sizes(block_col)+lev2_vert_offset+1,&
                              ! lev1_hori_offset+(orb_i-1)*ao_domain_sizes(col)+copy

                              H1(lev1_vert_offset+(orb_j-1)*ao_domain_sizes(block_col)+lev2_vert_offset+1: &
                                 lev1_vert_offset+(orb_j-1)*ao_domain_sizes(block_col)+lev2_vert_offset+ao_block_sizes(row), &
                                 lev1_hori_offset+(orb_i-1)*ao_domain_sizes(col)+copy) &
                                 = block_p(:, orb_i)

                           ENDIF ! found block in the data matrix

                           IF (found2) THEN

                              H2(lev1_vert_offset+(orb_j-1)*ao_domain_sizes(block_col)+lev2_vert_offset+1: &
                                 lev1_vert_offset+(orb_j-1)*ao_domain_sizes(block_col)+lev2_vert_offset+ao_block_sizes(row), &
                                 lev1_hori_offset+(orb_i-1)*ao_domain_sizes(col)+copy) &
                                 = block_p2(:, orb_i)

                           ENDIF ! found block in the data matrix

                        ENDDO

                     ENDDO

                  ENDDO

                  !lev2_vert_offset=lev2_vert_offset+ao_block_sizes(row)

               ENDIF ! found block in the quench matrix

               lev1_vert_offset = lev1_vert_offset+ &
                                  ao_domain_sizes(block_col)*mo_block_sizes(block_col)

            ENDDO

            lev1_hori_offset = lev1_hori_offset+ &
                               ao_domain_sizes(col)*mo_block_sizes(col)

         ENDDO

         !lev2_vert_offset=lev2_vert_offset+ao_block_sizes(row)

      ENDDO
      ! add terms to the hessian
!WRITE(*,*) "F_vo"
!DO ii=1,H_size
! WRITE(*,'(100F13.9)') H1(ii,:)
!ENDDO
!WRITE(*,*) "S_vo"
!DO ii=1,H_size
! WRITE(*,'(100F13.9)') H2(ii,:)
!ENDDO
      !DO ii=1,H_size
      !   DO jj=1,H_size
      !      H(ii,jj)=H(ii,jj)-H1(ii,jj)*H2(jj,ii)-H1(jj,ii)*H2(ii,jj)
      !   ENDDO
      !ENDDO
      DEALLOCATE (H1)
      DEALLOCATE (H2)

      ! convert gradient from the dbcsr matrix to the vector form
      ALLOCATE (Grad_vec(H_size))
      Grad_vec(:) = 0.0_dp
      lev1_vert_offset = 0
      ! loop over all electron blocks
      DO col = 1, nblkcols_tot

         ! loop over AO-rows of the dbcsr matrix
         lev2_vert_offset = 0
         DO row = 1, nblkrows_tot

            CALL cp_dbcsr_get_block_p(quench_t, &
                                      row, col, block_p, found_row)
            IF (found_row) THEN

               CALL cp_dbcsr_get_block_p(matrix_grad, &
                                         row, col, block_p, found)
               IF (found) THEN
                  ! copy the data into the vector, column by column
                  DO orb_i = 1, mo_block_sizes(col)
                     Grad_vec(lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+1: &
                              lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+ao_block_sizes(row)) &
                        = block_p(:, orb_i)
!WRITE(*,*) "GRAD: ", row, col, orb_i, lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+1, ao_block_sizes(row)
                  ENDDO

               ENDIF

               lev2_vert_offset = lev2_vert_offset+ao_block_sizes(row)

            ENDIF

         ENDDO

         lev1_vert_offset = lev1_vert_offset+ao_domain_sizes(col)*mo_block_sizes(col)

      ENDDO ! loop over electron blocks

!WRITE(*,*) "HESSIAN"
!DO ii=1,H_size
! WRITE(*,'(100F13.9)') H(ii,:)
!ENDDO

      ! invert the Hessian
      INFO = 0
      ALLOCATE (Hinv(H_size, H_size))
      Hinv(:, :) = H(:, :)

      ! before inverting diagonalize
      ALLOCATE (eigenvalues(H_size))
      ! Query the optimal workspace for dsyev
      LWORK = -1
      ALLOCATE (WORK(MAX(1, LWORK)))
      CALL DSYEV('V', 'L', H_size, Hinv, H_size, eigenvalues, WORK, LWORK, INFO)
      LWORK = INT(WORK(1))
      DEALLOCATE (WORK)
      ! Allocate the workspace and solve the eigenproblem
      ALLOCATE (WORK(MAX(1, LWORK)))
      CALL DSYEV('V', 'L', H_size, Hinv, H_size, eigenvalues, WORK, LWORK, INFO)
      IF (INFO .NE. 0) THEN
         WRITE (*, *) 'DSYEV ERROR MESSAGE: ', INFO
         CALL cp__b("almo_scf_methods.F",3441,"DSYEV failed")
      END IF
      DEALLOCATE (WORK)

      ! invert eigenvalues and use eigenvectors to compute the Hessian inverse
      ! project out zero-eigenvalue directions
      ALLOCATE (test(H_size, H_size))
      zero_neg_eiv = 0
      DO jj = 1, H_size
         IF (eigenvalues(jj) .GT. 1.0E-8) THEN
            test(jj, :) = Hinv(:, jj)/eigenvalues(jj)
         ELSE
            test(jj, :) = Hinv(:, jj)*0.0_dp
            zero_neg_eiv = zero_neg_eiv+1
         ENDIF
      ENDDO
      WRITE (*, *) 'ZERO OR NEGATIVE EIGENVALUES: ', zero_neg_eiv
      ALLOCATE (test2(H_size, H_size))
      test2(:, :) = MATMUL(Hinv, test)
      Hinv(:, :) = test2(:, :)
      DEALLOCATE (test, test2)

      !! shift to kill singularity
      !shift=0.0_dp
      !IF (eigenvalues(1).lt.0.0_dp) THEN
      !   CPErrorMessage(cp_failure_level,routineP,"Negative eigenvalue(s)")
      !   shift=abs(eigenvalues(1))
      !   WRITE(*,*) "Lowest eigenvalue: ", eigenvalues(1)
      !ENDIF
      !DO ii=1, H_size
      !   IF (eigenvalues(ii).gt.1.0E-6_dp) THEN
      !      shift=shift+min(1.0_dp,eigenvalues(ii))*1.0E-4_dp
      !      EXIT
      !   ENDIF
      !ENDDO
      !WRITE(*,*) "Hessian shift: ", shift
      !DO ii=1, H_size
      !   H(ii,ii)=H(ii,ii)+shift
      !ENDDO
      !! end shift

      DEALLOCATE (eigenvalues)

!!!!    Hinv=H
!!!!    INFO=0
!!!!    CALL DPOTRF('L', H_size, Hinv, H_size, INFO )
!!!!    IF( INFO.NE.0 ) THEN
!!!!       WRITE(*,*) 'DPOTRF ERROR MESSAGE: ', INFO
!!!!       CPErrorMessage(cp_failure_level,routineP,"DPOTRF failed")
!!!!       CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
!!!!    END IF
!!!!    CALL DPOTRI('L', H_size, Hinv, H_size, INFO )
!!!!    IF( INFO.NE.0 ) THEN
!!!!       WRITE(*,*) 'DPOTRI ERROR MESSAGE: ', INFO
!!!!       CPErrorMessage(cp_failure_level,routineP,"DPOTRI failed")
!!!!       CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
!!!!    END IF
!!!!    ! complete the matrix
!!!!    DO ii=1,H_size
!!!!       DO jj=ii+1,H_size
!!!!          Hinv(ii,jj)=Hinv(jj,ii)
!!!!       ENDDO
!!!!    ENDDO

      ! compute the inversion error
      ALLOCATE (test(H_size, H_size))
!    WRITE(*,*) "SIZE: ", H_size
      test(:, :) = MATMUL(Hinv, H)
!WRITE(*,*) "TEST"
!DO ii=1,H_size
! WRITE(*,'(100F8.4)') test(ii,:)
!ENDDO
      DO ii = 1, H_size
         test(ii, ii) = test(ii, ii)-1.0_dp
      ENDDO
      test_error = 0.0_dp
      DO ii = 1, H_size
         DO jj = 1, H_size
            test_error = test_error+test(jj, ii)*test(jj, ii)
         ENDDO
      ENDDO
      WRITE (*, *) "Hessian inversion error: ", SQRT(test_error)
      DEALLOCATE (test)

      ! prepare the output vector
      ALLOCATE (Step_vec(H_size))
      ALLOCATE (tmp(H_size))
      tmp(:) = MATMUL(Hinv, Grad_vec)
      Step_vec(:) = -1.0_dp*tmp(:)

      ALLOCATE (tmpr(H_size))
      tmpr(:) = MATMUL(H, Step_vec)
      tmp(:) = tmpr(:)+Grad_vec(:)
      DEALLOCATE (tmpr)
      WRITE (*, *) "NEWTOV step error: ", MAXVAL(ABS(tmp))

      DEALLOCATE (tmp)

      DEALLOCATE (H)
      DEALLOCATE (Hinv)
      DEALLOCATE (Grad_vec)

      ! copy the step from the vector into the cp_dbcsr matrix

      ! re-create the step matrix to remove all blocks
      CALL cp_dbcsr_create(matrix_step, &
                           template=matrix_grad, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_work_create(matrix_step, work_mutable=.TRUE.)

      lev1_vert_offset = 0
      ! loop over all electron blocks
      DO col = 1, nblkcols_tot

         ! loop over AO-rows of the dbcsr matrix
         lev2_vert_offset = 0
         DO row = 1, nblkrows_tot

            CALL cp_dbcsr_get_block_p(quench_t, &
                                      row, col, block_p, found_row)
            IF (found_row) THEN

               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(matrix_step, row, col, p_new_block)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_methods.F",3565)
               ! copy the data column by column
               DO orb_i = 1, mo_block_sizes(col)
                  p_new_block(:, orb_i) = &
                     Step_vec(lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+1: &
                              lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+ao_block_sizes(row))
!WRITE(*,*) "STEP: ", row, col, orb_i, lev1_vert_offset+ao_domain_sizes(col)*(orb_i-1)+lev2_vert_offset+1, ao_block_sizes(row)
               ENDDO

               lev2_vert_offset = lev2_vert_offset+ao_block_sizes(row)

            ENDIF

         ENDDO

         lev1_vert_offset = lev1_vert_offset+ao_domain_sizes(col)*mo_block_sizes(col)

      ENDDO ! loop over electron blocks

      DEALLOCATE (Step_vec)

      CALL cp_dbcsr_finalize(matrix_step)

      DEALLOCATE (mo_block_sizes, ao_block_sizes)
      DEALLOCATE (ao_domain_sizes)

      CALL timestop(handle)

   END SUBROUTINE hessian_diag_apply

! **************************************************************************************************
!> \brief Load balancing of the submatrix computations
!> \param almo_scf_env ...
!> \par History
!>       2013.02 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE distribute_domains(almo_scf_env)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'distribute_domains', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, least_loaded, nao, &
                                                            ncpus, ndomains
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cpu_load, domain_load

      CALL timeset(routineN, handle)

      ndomains = almo_scf_env%ndomains
      ncpus = dbcsr_mp_numnodes(dbcsr_distribution_mp( &
                                cp_dbcsr_distribution(almo_scf_env%matrix_s(1))))

      ALLOCATE (domain_load(ndomains))
      DO idomain = 1, ndomains
         nao = almo_scf_env%nbasis_of_domain(idomain)
         domain_load(idomain) = (nao*nao*nao)*1.0_dp
      ENDDO

      ALLOCATE (index0(ndomains))

      CALL sort(domain_load, ndomains, index0)

      ALLOCATE (cpu_load(ncpus))
      cpu_load(:) = 0.0_dp

      DO idomain = 1, ndomains
         least_loaded = MINLOC(cpu_load, 1)
         cpu_load(least_loaded) = cpu_load(least_loaded)+domain_load(idomain)
         almo_scf_env%cpu_of_domain(index0(idomain)) = least_loaded-1
      ENDDO

      DEALLOCATE (cpu_load)
      DEALLOCATE (index0)
      DEALLOCATE (domain_load)

      CALL timestop(handle)

   END SUBROUTINE distribute_domains

! **************************************************************************************************
!> \brief Tests construction and release of domain submatrices
!> \param matrix_no ...
!> \param dpattern ...
!> \param map ...
!> \param node_of_domain ...
!> \par History
!>       2013.01 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE construct_test(matrix_no, dpattern, map, node_of_domain)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_no, dpattern
      TYPE(domain_map_type), INTENT(IN)                  :: map
      INTEGER, DIMENSION(:), INTENT(IN)                  :: node_of_domain

      CHARACTER(len=*), PARAMETER :: routineN = 'construct_test', routineP = moduleN//':'//routineN

      INTEGER                                            :: GroupID, handle, ndomains
      TYPE(cp_dbcsr_type)                                :: copy1
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: subm_nn, subm_no

      CALL timeset(routineN, handle)

      ndomains = cp_dbcsr_nblkcols_total(dpattern)
      GroupID = dbcsr_mp_group(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(dpattern)))

      ALLOCATE (subm_no(ndomains), subm_nn(ndomains))
      CALL init_submatrices(subm_no)
      CALL init_submatrices(subm_nn)

      !CALL cp_dbcsr_print(matrix_nn)
      !CALL construct_submatrices(matrix_nn,subm_nn,dpattern,map,select_row_col)
      !CALL print_submatrices(subm_nn,GroupID)

      !CALL cp_dbcsr_print(matrix_no)
      CALL construct_submatrices(matrix_no, subm_no, dpattern, map, node_of_domain, select_row)
      CALL print_submatrices(subm_no, GroupID)

      CALL cp_dbcsr_init(copy1)
      CALL cp_dbcsr_create(copy1, template=matrix_no)
      CALL cp_dbcsr_copy(copy1, matrix_no)
      CALL cp_dbcsr_print(copy1)
      CALL construct_dbcsr_from_submatrices(copy1, subm_no, dpattern)
      CALL cp_dbcsr_print(copy1)
      CALL cp_dbcsr_release(copy1)

      CALL release_submatrices(subm_no)
      CALL release_submatrices(subm_nn)
      DEALLOCATE (subm_no, subm_nn)

      CALL timestop(handle)

   END SUBROUTINE construct_test

END MODULE almo_scf_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface between ALMO SCF and QS
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf_qs
   USE almo_scf_methods,                ONLY: almo_scf_p_blk_to_t_blk,&
                                              almo_scf_t_blk_to_p,&
                                              almo_scf_t_blk_to_t_blk_orthonormal
   USE almo_scf_types,                  ONLY: almo_mat_dim_aobasis,&
                                              almo_mat_dim_occ,&
                                              almo_mat_dim_virt,&
                                              almo_mat_dim_virt_disc,&
                                              almo_mat_dim_virt_full,&
                                              almo_scf_env_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_complete_redistribute, cp_dbcsr_copy, &
        cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_distribution_new, cp_dbcsr_distribution_release, &
        cp_dbcsr_filter, cp_dbcsr_finalize, cp_dbcsr_get_block_p, cp_dbcsr_get_info, &
        cp_dbcsr_get_num_blocks, cp_dbcsr_get_stored_coordinates, cp_dbcsr_init, &
        cp_dbcsr_nblkcols_total, cp_dbcsr_nblkrows_total, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_reserve_block2d, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        cp_dbcsr_work_create, dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_mp, dbcsr_distribution_obj, dbcsr_distribution_row_clusters, &
        dbcsr_distribution_row_dist, dbcsr_mp_group, dbcsr_mp_mynode, dbcsr_mp_numnodes, &
        dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        almo_constraint_ao_overlap, almo_constraint_block_diagonal, almo_constraint_distance, &
        almo_domain_layout_molecular, almo_mat_distr_atomic, almo_mat_distr_molecular, &
        atomic_guess, do_bondparm_covalent, do_bondparm_vdw, molecular_guess
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_allgather
   USE molecule_types_new,              ONLY: get_molecule_set_info,&
                                              molecule_type
   USE mscfg_types,                     ONLY: get_matrix_from_submatrices,&
                                              molecular_scf_guess_env_type
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_initial_guess,                ONLY: calculate_atomic_block_dm,&
                                              calculate_mopac_dm
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf_qs'

   PUBLIC :: almo_scf_init_qs, matrix_almo_create, matrix_qs_to_almo, &
             almo_scf_dm_to_ks, &
             almo_scf_construct_quencher, almo_scf_update_ks_energy

CONTAINS

! **************************************************************************************************
!> \brief create the ALMO matrix templates
!> \param matrix_new ...
!> \param matrix_qs ...
!> \param almo_scf_env ...
!> \param name_new ...
!> \param size_keys ...
!> \param symmetry_new ...
!> \param spin_key ...
!> \param init_domains ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_almo_create(matrix_new, matrix_qs, almo_scf_env, &
                                 name_new, size_keys, symmetry_new, &
                                 spin_key, init_domains)

      TYPE(cp_dbcsr_type)                                :: matrix_new, matrix_qs
      TYPE(almo_scf_env_type), INTENT(IN)                :: almo_scf_env
      CHARACTER(len=*), INTENT(IN)                       :: name_new
      INTEGER, DIMENSION(2), INTENT(IN)                  :: size_keys
      CHARACTER, INTENT(IN)                              :: symmetry_new
      INTEGER, INTENT(IN)                                :: spin_key
      LOGICAL, INTENT(IN)                                :: init_domains

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_almo_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dimen, handle, hold, iatom, iblock_col, &
                                                            iblock_row, imol, mynode, natoms, &
                                                            nblkrows_tot, nlength, nmols, row
      INTEGER, DIMENSION(:), POINTER :: blk_distr, blk_sizes, block_sizes_new, cluster_distr, &
         cluster_distr_new, col_cluster_new, col_distr_new, col_sizes_new, distr_new_array, &
         row_cluster_new, row_distr_new, row_sizes_new
      LOGICAL                                            :: active, tr
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_new_block
      TYPE(dbcsr_distribution_obj)                       :: dist_new, dist_qs

! dimension size: AO, MO, etc
!                 almo_mat_dim_aobasis - no. of AOs,
!                 almo_mat_dim_occ     - no. of occupied MOs
!                 almo_mat_dim_domains - no. of domains
! symmetry type: dbcsr_type_no_symmetry, dbcsr_type_symmetric,
!  dbcsr_type_antisymmetric, dbcsr_type_hermitian, dbcsr_type_antihermitian
!  (see dbcsr_lib/dbcsr_types.F for other values)
! spin_key: either 1 or 2 (0 is allowed for matrics in the AO basis)
!    TYPE(cp_dbcsr_iterator)                  :: iter
!    REAL(KIND=dp), ALLOCATABLE, DIMENSION(:) :: allones
!-----------------------------------------------------------------------

      CALL timeset(routineN, handle)

      ! RZK-warning The structure of the matrices can be optimized:
      ! 1. Diagonal matrices must be distributed evenly over the processes.
      !    This can be achived by distributing cpus: 012012-rows and 001122-cols
      !    block_diagonal_flag is introduced but not used
      ! 2. Multiplication of diagonally dominant matrices will be faster
      !    if the diagonal blocks are local to the same processes.
      ! 3. Systems of molecules of drastically different sizes might need
      !    better distribution.

      ! obtain distribution from the qs matrix - it might be useful
      ! to get the structure of the AO dimensions
      dist_qs = cp_dbcsr_distribution(matrix_qs)

      natoms = almo_scf_env%natoms
      nmols = almo_scf_env%nmolecules

      DO dimen = 1, 2 ! 1 - row, 2 - column dimension

         ! distribution pattern is the same for all matrix types (ao, occ, virt)
         IF (dimen == 1) THEN !rows
            blk_distr => dbcsr_distribution_row_dist(dist_qs)
            cluster_distr => dbcsr_distribution_row_clusters(dist_qs)
         ELSE !columns
            blk_distr => dbcsr_distribution_col_dist(dist_qs)
            cluster_distr => dbcsr_distribution_col_clusters(dist_qs)
         ENDIF

         NULLIFY (cluster_distr_new)

         IF (size_keys(dimen) == almo_mat_dim_aobasis) THEN ! this dimension is AO

            ! structure of an AO dimension can be copied from matrix_qs
            blk_sizes => cp_dbcsr_row_block_sizes(matrix_qs)

            ! atomic clustering of AOs
            IF (almo_scf_env%mat_distr_aos == almo_mat_distr_atomic) THEN
               ALLOCATE (block_sizes_new(natoms), distr_new_array(natoms))
               block_sizes_new(:) = blk_sizes(:)
               distr_new_array(:) = blk_distr(:)
               IF (ASSOCIATED(cluster_distr)) THEN
                  ALLOCATE (cluster_distr_new(natoms))
                  cluster_distr_new(:) = cluster_distr(:)
               END IF
               ! molecular clustering of AOs
            ELSE IF (almo_scf_env%mat_distr_aos == almo_mat_distr_molecular) THEN
               ALLOCATE (block_sizes_new(nmols), distr_new_array(nmols))
               block_sizes_new(:) = 0
               DO iatom = 1, natoms
                  block_sizes_new(almo_scf_env%domain_index_of_atom(iatom)) = &
                     block_sizes_new(almo_scf_env%domain_index_of_atom(iatom))+ &
                     blk_sizes(iatom)
               ENDDO
               DO imol = 1, nmols
                  distr_new_array(imol) = &
                     blk_distr(almo_scf_env%first_atom_of_domain(imol))
               ENDDO
               IF (ASSOCIATED(cluster_distr)) THEN
                  ALLOCATE (cluster_distr_new(nmols))
                  DO imol = 1, nmols
                     cluster_distr_new(imol) = &
                        cluster_distr(almo_scf_env%first_atom_of_domain(imol))
                  ENDDO
               END IF
            ELSE
               CALL cp__b("almo_scf_qs.F",207,"Illegal distribution")
            ENDIF

         ELSE ! this dimension is not AO

            IF (size_keys(dimen) == almo_mat_dim_occ .OR. &
                size_keys(dimen) == almo_mat_dim_virt .OR. &
                size_keys(dimen) == almo_mat_dim_virt_disc .OR. &
                size_keys(dimen) == almo_mat_dim_virt_full) THEN ! this dim is MO

               ! atomic clustering of MOs
               IF (almo_scf_env%mat_distr_mos == almo_mat_distr_atomic) THEN
                  nlength = natoms
                  ALLOCATE (block_sizes_new(nlength))
                  block_sizes_new(:) = 0
                  IF (size_keys(dimen) == almo_mat_dim_occ) THEN
                     ! currently distributing atomic distr of mos is not allowed
                     ! RZK-warning define nocc_of_atom and nvirt_atom to implement it
                     !block_sizes_new(:)=almo_scf_env%nocc_of_atom(:,spin_key)
                  ELSE IF (size_keys(dimen) == almo_mat_dim_virt) THEN
                     !block_sizes_new(:)=almo_scf_env%nvirt_of_atom(:,spin_key)
                  ENDIF
                  ! molecular clustering of MOs
               ELSE IF (almo_scf_env%mat_distr_mos == almo_mat_distr_molecular) THEN
                  nlength = nmols
                  ALLOCATE (block_sizes_new(nlength))
                  IF (size_keys(dimen) == almo_mat_dim_occ) THEN
                     block_sizes_new(:) = almo_scf_env%nocc_of_domain(:, spin_key)
                  ELSE IF (size_keys(dimen) == almo_mat_dim_virt_disc) THEN
                     block_sizes_new(:) = almo_scf_env%nvirt_disc_of_domain(:, spin_key)
                  ELSE IF (size_keys(dimen) == almo_mat_dim_virt_full) THEN
                     block_sizes_new(:) = almo_scf_env%nvirt_full_of_domain(:, spin_key)
                  ELSE IF (size_keys(dimen) == almo_mat_dim_virt) THEN
                     block_sizes_new(:) = almo_scf_env%nvirt_of_domain(:, spin_key)
                  ENDIF
               ELSE
                  CALL cp__b("almo_scf_qs.F",243,"Illegal distribution")
               ENDIF

            ELSE

               CALL cp__b("almo_scf_qs.F",248,"Illegal dimension")

            ENDIF ! end choosing dim size (occ, virt)

            ! distribution for MOs is copied from AOs
            ALLOCATE (distr_new_array(nlength))
            IF (ASSOCIATED(cluster_distr)) THEN
               ALLOCATE (cluster_distr_new(nlength))
            END IF
            ! atomic clustering
            IF (almo_scf_env%mat_distr_mos == almo_mat_distr_atomic) THEN
               distr_new_array(:) = blk_distr(:)
               IF (ASSOCIATED(cluster_distr)) THEN
                  cluster_distr_new(:) = cluster_distr(:)
               ENDIF
               ! molecular clustering
            ELSE IF (almo_scf_env%mat_distr_mos == almo_mat_distr_molecular) THEN
               DO imol = 1, nmols
                  distr_new_array(imol) = &
                     blk_distr(almo_scf_env%first_atom_of_domain(imol))
               ENDDO
               IF (ASSOCIATED(cluster_distr)) THEN
                  DO imol = 1, nmols
                     cluster_distr_new(imol) = &
                        cluster_distr(almo_scf_env%first_atom_of_domain(imol))
                  ENDDO
               END IF
            ENDIF
         ENDIF ! end choosing dimension size (AOs vs .NOT.AOs)

         ! create final arrays
         IF (dimen == 1) THEN !rows
            row_sizes_new => block_sizes_new
            row_distr_new => distr_new_array
            row_cluster_new => cluster_distr_new
         ELSE !columns
            col_sizes_new => block_sizes_new
            col_distr_new => distr_new_array
            col_cluster_new => cluster_distr_new
         ENDIF
      ENDDO ! both rows and columns are done

      ! Create the distribution
      CALL cp_dbcsr_distribution_new(dist_new, &
                                     dbcsr_distribution_mp(dist_qs), row_distr_new, col_distr_new, &
                                     row_cluster_new, col_cluster_new, &
                                     reuse_arrays=.TRUE.)

      ! Create the matrix
      CALL cp_dbcsr_init(matrix_new)
      CALL cp_dbcsr_create(matrix_new, name_new, &
                           dist_new, symmetry_new, &
                           row_sizes_new, col_sizes_new, reuse_arrays=.TRUE.)
      CALL cp_dbcsr_distribution_release(dist_new)

      ! fill out reqired blocks with 1.0_dp to tell the dbcsr library
      ! which blocks to keep
      IF (init_domains) THEN

         mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(cp_dbcsr_distribution(matrix_new)))
         CALL cp_dbcsr_work_create(matrix_new, work_mutable=.TRUE.)

         ! startQQQ - this part of the code scales quadratically
         ! therefore it is replaced with a less general but linear scaling algorithm below
         ! the quadratic algorithm is kept to be re-written later
         !QQQnblkrows_tot = cp_dbcsr_nblkrows_total(matrix_new)
         !QQQnblkcols_tot = cp_dbcsr_nblkcols_total(matrix_new)
         !QQQDO row = 1, nblkrows_tot
         !QQQ   DO col = 1, nblkcols_tot
         !QQQ      tr = .FALSE.
         !QQQ      iblock_row = row
         !QQQ      iblock_col = col
         !QQQ      CALL cp_dbcsr_get_stored_coordinates(matrix_new, iblock_row, iblock_col, tr, hold)

         !QQQ      IF(hold.EQ.mynode) THEN
         !QQQ
         !QQQ         ! RZK-warning replace with a function which says if this
         !QQQ         ! distribution block is active or not
         !QQQ         ! Translate indeces of distribution blocks to domain blocks
         !QQQ         if (size_keys(1)==almo_mat_dim_aobasis) then
         !QQQ           domain_row=almo_scf_env%domain_index_of_ao_block(iblock_row)
         !QQQ         else if (size_keys(2)==almo_mat_dim_occ .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt_disc .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt_full) then
         !QQQ           domain_row=almo_scf_env%domain_index_of_mo_block(iblock_row)
         !QQQ         else
         !QQQ           CPErrorMessage(cp_failure_level,routineP,"Illegal dimension")
         !QQQ           CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
         !QQQ         endif

         !QQQ         if (size_keys(2)==almo_mat_dim_aobasis) then
         !QQQ           domain_col=almo_scf_env%domain_index_of_ao_block(iblock_col)
         !QQQ         else if (size_keys(2)==almo_mat_dim_occ .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt_disc .OR. &
         !QQQ                  size_keys(2)==almo_mat_dim_virt_full) then
         !QQQ           domain_col=almo_scf_env%domain_index_of_mo_block(iblock_col)
         !QQQ         else
         !QQQ           CPErrorMessage(cp_failure_level,routineP,"Illegal dimension")
         !QQQ           CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
         !QQQ         endif

         !QQQ         ! Finds if we need this block
         !QQQ         ! only the block-diagonal constraint is implemented here
         !QQQ         active=.false.
         !QQQ         if (domain_row==domain_col) active=.true.

         !QQQ         IF (active) THEN
         !QQQ            NULLIFY (p_new_block)
         !QQQ            CALL cp_dbcsr_reserve_block2d(matrix_new, iblock_row, iblock_col, p_new_block)
         !QQQ            CPPostcondition(ASSOCIATED(p_new_block),cp_failure_level,routineP,failure)
         !QQQ            p_new_block(:,:) = 1.0_dp
         !QQQ         ENDIF

         !QQQ      ENDIF ! mynode
         !QQQ   ENDDO
         !QQQENDDO
         ! endQQQ - end of the quadratic part
         ! start linear-scaling replacement:
         ! works only for molecular blocks AND molecular distributions
         nblkrows_tot = cp_dbcsr_nblkrows_total(matrix_new)
         DO row = 1, nblkrows_tot
            tr = .FALSE.
            iblock_row = row
            iblock_col = row
            CALL cp_dbcsr_get_stored_coordinates(matrix_new, iblock_row, iblock_col, hold)

            IF (hold .EQ. mynode) THEN

               active = .TRUE.

               IF (active) THEN
                  NULLIFY (p_new_block)
                  CALL cp_dbcsr_reserve_block2d(matrix_new, iblock_row, iblock_col, p_new_block)
                  IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_qs.F",383)
                  p_new_block(:, :) = 1.0_dp
               ENDIF

            ENDIF ! mynode
         ENDDO
         ! end lnear-scaling replacement

      ENDIF ! init_domains

      CALL cp_dbcsr_finalize(matrix_new)

      CALL timestop(handle)

   END SUBROUTINE matrix_almo_create

! **************************************************************************************************
!> \brief convert between two types of matrices: QS style to ALMO style
!> \param matrix_qs ...
!> \param matrix_almo ...
!> \param almo_scf_env ...
!> \param keep_sparsity ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_qs_to_almo(matrix_qs, matrix_almo, almo_scf_env, &
                                keep_sparsity)
      TYPE(cp_dbcsr_type)                                :: matrix_qs, matrix_almo
      TYPE(almo_scf_env_type)                            :: almo_scf_env
      LOGICAL, INTENT(IN)                                :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_qs_to_almo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_qs_nosym

      CALL timeset(routineN, handle)
      !RZK-warning if it's not a N(AO)xN(AO) matrix then stop

      SELECT CASE (almo_scf_env%mat_distr_aos)
      CASE (almo_mat_distr_atomic)
         ! automatic data_type conversion
         CALL cp_dbcsr_copy(matrix_almo, matrix_qs, &
                            keep_sparsity=keep_sparsity)
      CASE (almo_mat_distr_molecular)
         ! desymmetrize the qs matrix
         CALL cp_dbcsr_init(matrix_qs_nosym)
         CALL cp_dbcsr_create(matrix_qs_nosym, template=matrix_qs, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_desymmetrize(matrix_qs, matrix_qs_nosym)

         ! perform the magic complete_redistribute
         ! before calling complete_redistribute set all blocks to zero
         ! otherwise the non-zero elements of the redistributed matrix,
         ! which are in zero-blocks of the original matrix, will remain
         ! in the final redistributed matrix. this is a bug in
         ! complete_redistribute. RZK-warning it should be later corrected by calling
         ! dbcsr_set to 0.0 from within complete_redistribute
         CALL cp_dbcsr_set(matrix_almo, 0.0_dp)
         CALL cp_dbcsr_complete_redistribute(matrix_qs_nosym, matrix_almo, &
                                             keep_sparsity=keep_sparsity); 
         CALL cp_dbcsr_release(matrix_qs_nosym)

      CASE DEFAULT
         CALL cp__b("almo_scf_qs.F",449,"")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE matrix_qs_to_almo

! **************************************************************************************************
!> \brief convert between two types of matrices: ALMO style to QS style
!> \param matrix_almo ...
!> \param matrix_qs ...
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_almo_to_qs(matrix_almo, matrix_qs, almo_scf_env)
      TYPE(cp_dbcsr_type)                                :: matrix_almo, matrix_qs
      TYPE(almo_scf_env_type), INTENT(IN)                :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_almo_to_qs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      ! RZK-warning if it's not a N(AO)xN(AO) matrix then stop

!    IF (ls_mstruct%single_precision) THEN
!       CALL cp_dbcsr_init (matrix_tmp)
!       CALL cp_dbcsr_create (matrix_tmp, template=matrix_ls,&
!            data_type=dbcsr_type_real_8)
!       CALL cp_dbcsr_copy (matrix_tmp, matrix_ls)
!    ENDIF

      SELECT CASE (almo_scf_env%mat_distr_aos)
      CASE (almo_mat_distr_atomic)
!       IF (ls_mstruct%single_precision) THEN
!          CALL cp_dbcsr_copy_into_existing (matrix_qs, matrix_tmp)
!       ELSE
         CALL cp_dbcsr_copy_into_existing(matrix_qs, matrix_almo)
!       ENDIF
      CASE (almo_mat_distr_molecular)
         CALL cp_dbcsr_set(matrix_qs, 0.0_dp)
!       IF (ls_mstruct%single_precision) THEN
!          CALL cp_dbcsr_complete_redistribute(matrix_tmp, matrix_qs, keep_sparsity=.TRUE.)
!       ELSE
         CALL cp_dbcsr_complete_redistribute(matrix_almo, matrix_qs, keep_sparsity=.TRUE.)
!       ENDIF
      CASE DEFAULT
         CALL cp__b("almo_scf_qs.F",499,"")
      END SELECT

!    IF (ls_mstruct%single_precision) THEN
!       CALL cp_dbcsr_release(matrix_tmp)
!    ENDIF

      CALL timestop(handle)

   END SUBROUTINE matrix_almo_to_qs

! **************************************************************************************************
!> \brief Initialization of QS and ALMOs
!>        Some parts can be factored-out since they are common
!>        for the other SCF methods
!> \param qs_env ...
!> \param almo_scf_env ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_init_qs(qs_env, almo_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_init_qs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin, unit_nr
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      LOGICAL                                            :: has_unit_metric
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(molecular_scf_guess_env_type), POINTER        :: mscfg_env
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)
      NULLIFY (rho, sab_orb, rho_ao)
      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! get basic quantities from the qs_env
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      ks_env=ks_env, &
                      input=input, &
                      energy=energy, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      has_unit_metric=has_unit_metric, &
                      para_env=para_env, &
                      nelectron_spin=nelectron_spin, &
                      mscfg_env=mscfg_env, &
                      rho=rho, &
                      sab_orb=sab_orb)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      IF(.NOT.(ASSOCIATED(mscfg_env)))CALL cp__a("almo_scf_qs.F",574)

      nspin = dft_control%nspins

      ! create matrix_ks if necessary
      IF (.NOT. ASSOCIATED(matrix_ks)) THEN
         CALL cp_dbcsr_allocate_matrix_set(matrix_ks, nspin)
         DO ispin = 1, nspin
            ALLOCATE (matrix_ks(ispin)%matrix)
            CALL cp_dbcsr_init(matrix_ks(ispin)%matrix)
            CALL cp_dbcsr_create(matrix_ks(ispin)%matrix, &
                                 template=matrix_s(1)%matrix)
            CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks(ispin)%matrix, sab_orb)
            CALL cp_dbcsr_set(matrix_ks(ispin)%matrix, 0.0_dp)
         ENDDO
         CALL set_ks_env(ks_env, matrix_ks=matrix_ks)
      ENDIF

      ! create an initial guess
      SELECT CASE (almo_scf_env%almo_scf_guess)
      CASE (molecular_guess)

         DO ispin = 1, nspin

            ! the calculations on "isolated" molecules has already been done
            ! all we need to do is convert the MOs of molecules into
            ! the ALMO matrix taking into account different distributions
            CALL get_matrix_from_submatrices(mscfg_env, &
                                             almo_scf_env%matrix_t_blk(ispin), ispin)
            CALL cp_dbcsr_filter(almo_scf_env%matrix_t_blk(ispin), &
                                 almo_scf_env%eps_filter)

         ENDDO

      CASE (atomic_guess)

         DO ispin = 1, nspin
            CALL cp_dbcsr_set(rho_ao(ispin)%matrix, 0.0_dp)
            IF (dft_control%qs_control%dftb .OR. dft_control%qs_control%semi_empirical) THEN
               CALL calculate_mopac_dm(rho_ao(ispin)%matrix, &
                                       matrix_s(1)%matrix, has_unit_metric, &
                                       dft_control, particle_set, atomic_kind_set, qs_kind_set, &
                                       nspin, nelectron_spin(ispin), &
                                       para_env)
            ELSE
               CALL calculate_atomic_block_dm(rho_ao(ispin)%matrix, &
                                              matrix_s(1)%matrix, particle_set, atomic_kind_set, qs_kind_set, &
                                              ispin, nspin, nelectron_spin(ispin), unit_nr)
            ENDIF

            ! copy the atomic-block dm into matrix_p_blk
            CALL matrix_qs_to_almo(rho_ao(ispin)%matrix, &
                                   almo_scf_env%matrix_p_blk(ispin), almo_scf_env, &
                                   .FALSE.)
            CALL cp_dbcsr_filter(almo_scf_env%matrix_p_blk(ispin), &
                                 almo_scf_env%eps_filter)

         ENDDO

         ! obtain orbitals from density matrix (currently ALMO SCF needs orbitals)
         ! RZK-warning currently the orbitals are generated
         ! by diagonalizing p blocks and taking only the
         ! eigenvectors with highest occupation
         ! essentially producing the ionic orbitals
         ! while it works for simple systems this guess
         ! will lead to convergence problems with complex molecules
         CALL almo_scf_p_blk_to_t_blk(almo_scf_env)

      END SELECT

      CALL almo_scf_t_blk_to_t_blk_orthonormal(almo_scf_env)
      CALL almo_scf_t_blk_to_p(almo_scf_env, &
                               use_sigma_inv_guess=.FALSE.)

      DO ispin = 1, nspin
         CALL matrix_almo_to_qs(almo_scf_env%matrix_p(ispin), &
                                rho_ao(ispin)%matrix, &
                                almo_scf_env)
      ENDDO

      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., &
                               just_energy=.FALSE.)

      IF (unit_nr > 0) THEN
         IF (almo_scf_env%almo_scf_guess .EQ. molecular_guess) THEN
            WRITE (unit_nr, '(T2,A38,F40.10)') "Single-molecule energy:", &
               SUM(mscfg_env%energy_of_frag)
         ENDIF
         WRITE (unit_nr, '(T2,A38,F40.10)') "Energy of the initial guess:", energy%total
         WRITE (unit_nr, '()')
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE almo_scf_init_qs

! **************************************************************************************************
!> \brief use the density matrix in almo_scf_env
!>        to compute the new energy and KS matrix
!> \param qs_env ...
!> \param almo_scf_env ...
!> \param energy_new ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_dm_to_ks(qs_env, almo_scf_env, energy_new)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env
      REAL(KIND=dp)                                      :: energy_new

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_dm_to_ks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (rho, rho_ao)
      CALL timeset(routineN, handle)

      nspin = almo_scf_env%nspins
      CALL get_qs_env(qs_env, energy=energy, rho=rho)
      CALL qs_rho_get(rho, rho_ao=rho_ao)

      ! set the new density matrix
      DO ispin = 1, nspin
         CALL matrix_almo_to_qs(almo_scf_env%matrix_p(ispin), &
                                rho_ao(ispin)%matrix, &
                                almo_scf_env)
      END DO

      ! compute the corresponding KS matrix and new energy
      CALL qs_rho_update_rho(rho, qs_env=qs_env)
      CALL qs_ks_did_change(qs_env%ks_env, rho_changed=.TRUE.)
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.FALSE., &
                               print_active=.TRUE.)
      energy_new = energy%total

      CALL timestop(handle)

   END SUBROUTINE almo_scf_dm_to_ks

! **************************************************************************************************
!> \brief update qs_env total energy
!> \param qs_env ...
!> \param energy ...
!> \par History
!>       2013.03 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_update_ks_energy(qs_env, energy)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(KIND=dp)                                      :: energy

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_update_ks_energy', &
         routineP = moduleN//':'//routineN

      TYPE(qs_energy_type), POINTER                      :: qs_energy

      CALL get_qs_env(qs_env, energy=qs_energy)
      qs_energy%total = energy

   END SUBROUTINE almo_scf_update_ks_energy

! **************************************************************************************************
!> \brief Creates the matrix that imposes absolute locality on MOs
!> \param qs_env ...
!> \param almo_scf_env ...
!> \par History
!>       2011.11 created [Rustam Z. Khaliullin]
!> \author Rustam Z. Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_construct_quencher(qs_env, almo_scf_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_construct_quencher', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: sym
      INTEGER :: col, contact_atom_1, contact_atom_2, domain_col, domain_map_local_entries, &
         domain_row, global_entries, global_list_length, grid1, GroupID, handle, hold, iatom, &
         iatom2, iblock_col, iblock_row, idomain, idomain2, ientry, igrid, ineig, ineighbor, &
         iNode, inode2, ipair, ispin, jatom, jatom2, jdomain2, local_list_length, &
         max_domain_neighbors, max_neig, mynode, nblkcols_tot, nblkrows_tot, nblks, ndomains, &
         neig_temp, nnode2, nNodes, row, unit_nr
      INTEGER, ALLOCATABLE, DIMENSION(:) :: current_number_neighbors, domain_entries_cpu, &
         domain_map_global, domain_map_local, first_atom_of_molecule, global_list, &
         last_atom_of_molecule, list_length_cpu, list_offset_cpu, local_list, offset_for_cpu
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: domain_grid, domain_neighbor_list, &
                                                            domain_neighbor_list_excessive
      LOGICAL                                            :: already_listed, block_active, &
                                                            delayed_increment, found, &
                                                            max_neig_fails, tr
      REAL(KIND=dp)                                      :: contact1_radius, contact2_radius, &
                                                            distance, distance_squared, overlap, &
                                                            r0, r1, s0, s1, trial_distance_squared
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_new_block
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_dbcsr_type)                                :: matrix_s_sym
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator, nl_iterator2
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_almo
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ndomains = almo_scf_env%ndomains

      CALL get_qs_env(qs_env=qs_env, &
                      particle_set=particle_set, &
                      molecule_set=molecule_set, &
                      cell=cell, &
                      matrix_s=matrix_s, &
                      sab_almo=sab_almo)

      ! if we are dealing with molecules get info about them
      IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular .OR. &
          almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN
         ALLOCATE (first_atom_of_molecule(almo_scf_env%nmolecules))
         ALLOCATE (last_atom_of_molecule(almo_scf_env%nmolecules))
         CALL get_molecule_set_info(molecule_set, &
                                    mol_to_first_atom=first_atom_of_molecule, &
                                    mol_to_last_atom=last_atom_of_molecule)
      ENDIF

      ! create a symmetrized copy of the ao overlap
      CALL cp_dbcsr_init(matrix_s_sym)
      CALL cp_dbcsr_create(matrix_s_sym, &
                           template=almo_scf_env%matrix_s(1), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_get_info(almo_scf_env%matrix_s(1), &
                             matrix_type=sym)
      IF (sym .EQ. dbcsr_type_no_symmetry) THEN
         CALL cp_dbcsr_copy(matrix_s_sym, almo_scf_env%matrix_s(1))
      ELSE
         CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_s(1), &
                                    matrix_s_sym)
      ENDIF

      ALLOCATE (almo_scf_env%quench_t(almo_scf_env%nspins))
      ALLOCATE (almo_scf_env%domain_map(almo_scf_env%nspins))

      !DO ispin=1,almo_scf_env%nspins
      ispin = 1

      ! create the sparsity template for the occupied orbitals
      CALL matrix_almo_create(matrix_new=almo_scf_env%quench_t(ispin), &
                              matrix_qs=matrix_s(1)%matrix, &
                              almo_scf_env=almo_scf_env, &
                              name_new="T_QUENCHER", &
                              size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_occ/), &
                              symmetry_new=dbcsr_type_no_symmetry, &
                              spin_key=ispin, &
                              init_domains=.FALSE.)

      ! initialize distance quencher
      CALL cp_dbcsr_work_create(almo_scf_env%quench_t(ispin), &
                                work_mutable=.TRUE.)

      nblkrows_tot = cp_dbcsr_nblkrows_total(almo_scf_env%quench_t(ispin))
      nblkcols_tot = cp_dbcsr_nblkcols_total(almo_scf_env%quench_t(ispin))

      mynode = dbcsr_mp_mynode(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))
      nNodes = dbcsr_mp_numnodes(dbcsr_distribution_mp( &
                                 cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))
      GroupID = dbcsr_mp_group(dbcsr_distribution_mp( &
                               cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))

      ! create global atom neighbor list from the local lists
      ! first, calculate number of local pairs
      local_list_length = 0
      CALL neighbor_list_iterator_create(nl_iterator, sab_almo)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         ! nnode - total number of neighbors for iatom
         ! inode - current neighbor count
         CALL get_iterator_info(nl_iterator, &
                                iatom=iatom2, jatom=jatom2, inode=inode2, nnode=nnode2)
         !WRITE(*,*) "GET INFO: ",iatom2, jatom2, inode2, nnode2
         IF (inode2 == 1) THEN
            local_list_length = local_list_length+nnode2
         END IF
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! second, extract the local list to an array
      ALLOCATE (local_list(2*local_list_length))
      local_list(:) = 0
      local_list_length = 0
      CALL neighbor_list_iterator_create(nl_iterator2, sab_almo)
      DO WHILE (neighbor_list_iterate(nl_iterator2) == 0)
         CALL get_iterator_info(nl_iterator2, &
                                iatom=iatom2, jatom=jatom2)
!WRITE(*,*) "GET INFO: ",iatom2, jatom2
         local_list(2*local_list_length+1) = iatom2
         local_list(2*local_list_length+2) = jatom2
         local_list_length = local_list_length+1
      ENDDO ! end loop over pairs of atoms
      CALL neighbor_list_iterator_release(nl_iterator2)

      ! third, communicate local length to the other nodes
      ALLOCATE (list_length_cpu(nNodes), list_offset_cpu(nNodes))
      CALL mp_allgather(2*local_list_length, list_length_cpu, GroupID)

      ! fourth, create a global list
      list_offset_cpu(1) = 0
      DO iNode = 2, nNodes
         list_offset_cpu(iNode) = list_offset_cpu(iNode-1)+ &
                                  list_length_cpu(iNode-1)
      ENDDO
      global_list_length = list_offset_cpu(nNodes)+list_length_cpu(nNodes)

      ! fifth, communicate all list data
      ALLOCATE (global_list(global_list_length))
!WRITE(*,*) "LENGTH: ", list_length_cpu
!WRITE(*,*) "OFFSET: ", list_offset_cpu
!WRITE(*,*) "LOCAL:  ", local_list
      CALL mp_allgather(local_list, global_list, &
                        list_length_cpu, list_offset_cpu, GroupID)
      DEALLOCATE (list_length_cpu, list_offset_cpu)
      DEALLOCATE (local_list)

      ! calculate maximum number of atoms surrounding the domain
      ALLOCATE (current_number_neighbors(almo_scf_env%ndomains))
      current_number_neighbors(:) = 0
      global_list_length = global_list_length/2
      DO ipair = 1, global_list_length
         iatom2 = global_list(2*(ipair-1)+1)
         jatom2 = global_list(2*(ipair-1)+2)
         idomain2 = almo_scf_env%domain_index_of_atom(iatom2)
         jdomain2 = almo_scf_env%domain_index_of_atom(jatom2)
         ! add to the list
         current_number_neighbors(idomain2) = current_number_neighbors(idomain2)+1
         ! add j,i with i,j
         IF (idomain2 .NE. jdomain2) THEN
            current_number_neighbors(jdomain2) = current_number_neighbors(jdomain2)+1
         ENDIF
      ENDDO
      max_domain_neighbors = MAXVAL(current_number_neighbors)

      ! use the global atom neighbor list to create a global domain neighbor list
      ALLOCATE (domain_neighbor_list_excessive(ndomains, max_domain_neighbors))
      current_number_neighbors(:) = 1
      DO ipair = 1, ndomains
         domain_neighbor_list_excessive(ipair, 1) = ipair
      ENDDO
      DO ipair = 1, global_list_length
         iatom2 = global_list(2*(ipair-1)+1)
         jatom2 = global_list(2*(ipair-1)+2)
         idomain2 = almo_scf_env%domain_index_of_atom(iatom2)
         jdomain2 = almo_scf_env%domain_index_of_atom(jatom2)
         already_listed = .FALSE.
         DO ineighbor = 1, current_number_neighbors(idomain2)
            IF (domain_neighbor_list_excessive(idomain2, ineighbor) .EQ. jdomain2) THEN
               already_listed = .TRUE.
               EXIT
            ENDIF
         ENDDO
         IF (.NOT. already_listed) THEN
            ! add to the list
            current_number_neighbors(idomain2) = current_number_neighbors(idomain2)+1
            domain_neighbor_list_excessive(idomain2, current_number_neighbors(idomain2)) = jdomain2
            ! add j,i with i,j
            IF (idomain2 .NE. jdomain2) THEN
               current_number_neighbors(jdomain2) = current_number_neighbors(jdomain2)+1
               domain_neighbor_list_excessive(jdomain2, current_number_neighbors(jdomain2)) = idomain2
            ENDIF
         ENDIF
      ENDDO ! end loop over pairs of atoms
      DEALLOCATE (global_list)

      max_domain_neighbors = MAXVAL(current_number_neighbors)
      ALLOCATE (domain_neighbor_list(ndomains, max_domain_neighbors))
      domain_neighbor_list(:, :) = 0
      domain_neighbor_list(:, :) = domain_neighbor_list_excessive(:, 1:max_domain_neighbors)
      DEALLOCATE (domain_neighbor_list_excessive)

      ALLOCATE (almo_scf_env%domain_map(ispin)%index1(ndomains))
      ALLOCATE (almo_scf_env%domain_map(ispin)%pairs(max_domain_neighbors*ndomains, 2))
      almo_scf_env%domain_map(ispin)%pairs(:, :) = 0
      almo_scf_env%domain_map(ispin)%index1(:) = 0
      domain_map_local_entries = 0

      ! RZK-warning intermediate [0,1] quencher values are ill-defined
      ! for molecules (not continuous and conceptually inadequate)

      ! O(N) loop over domain pairs
      DO row = 1, nblkrows_tot
         DO col = 1, current_number_neighbors(row)
            !DO col = 1, nblkcols_tot
            tr = .FALSE.
            iblock_row = row
            iblock_col = domain_neighbor_list(row, col)
            !iblock_col = col
!IF (unit_nr>0) WRITE(*,*) iblock_row, iblock_col
            CALL cp_dbcsr_get_stored_coordinates(almo_scf_env%quench_t(ispin), &
                                                 iblock_row, iblock_col, hold)

            IF (hold .EQ. mynode) THEN

               ! Translate indices of distribution blocks to indices of domain blocks
               ! Rows are AOs
               domain_row = almo_scf_env%domain_index_of_ao_block(iblock_row)
               ! Columns are electrons (i.e. MOs)
               domain_col = almo_scf_env%domain_index_of_mo_block(iblock_col)

               SELECT CASE (almo_scf_env%constraint_type)
               CASE (almo_constraint_block_diagonal)

                  block_active = .FALSE.
                  ! type of electron groups
                  IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular) THEN

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are molecular
                        IF (domain_row == domain_col) THEN
                           block_active = .TRUE.
                        ENDIF

                     ELSE ! ao domains are atomic

                        ! ao domains are atomic / electron groups are molecular
                        CALL cp__b("almo_scf_qs.F",1017,"Illegal: atomic domains and molecular groups")

                     ENDIF

                  ELSE ! electron groups are atomic

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are atomic
                        CALL cp__b("almo_scf_qs.F",1027,"Illegal: molecular domains and atomic groups")

                     ELSE

                        ! ao domains are atomic / electron groups are atomic
                        IF (domain_row == domain_col) THEN
                           block_active = .TRUE.
                        ENDIF

                     ENDIF

                  ENDIF ! end type of electron groups

                  IF (block_active) THEN

                     NULLIFY (p_new_block)
                     CALL cp_dbcsr_reserve_block2d(almo_scf_env%quench_t(ispin), &
                                                   iblock_row, iblock_col, p_new_block)
                     IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_qs.F",1045)
                     p_new_block(:, :) = 1.0_dp

                     IF (domain_map_local_entries .GE. max_domain_neighbors*almo_scf_env%ndomains) THEN
                        CALL cp__b("almo_scf_qs.F",1049,"weird... max_domain_neighbors is exceeded")
                     ENDIF
                     almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 1) = iblock_row
                     almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 2) = iblock_col
                     domain_map_local_entries = domain_map_local_entries+1

                  ENDIF

               CASE (almo_constraint_ao_overlap)

                  ! type of electron groups
                  IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular) THEN

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are molecular

                        ! compute the maximum overlap between the atoms of the two molecules
                        CALL cp_dbcsr_get_block_p(matrix_s_sym, &
                                                  iblock_row, iblock_col, p_new_block, found)
                        IF (found) THEN
                           !   CPErrorMessage(cp_failure_level,routineP,"S block not found")
                           !   CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
                           overlap = MAXVAL(ABS(p_new_block))
                        ELSE
                           overlap = 0.0_dp
                        ENDIF

                     ELSE ! ao domains are atomic

                        ! ao domains are atomic / electron groups are molecular
                        ! overlap_between_atom_and_molecule(atom=domain_row,molecule=domain_col)
                        CALL cp__b("almo_scf_qs.F",1082,"atomic domains and molecular groups - NYI")

                     ENDIF

                  ELSE ! electron groups are atomic

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are atomic
                        ! overlap_between_atom_and_molecule(atom=domain_col,molecule=domain_row)
                        CALL cp__b("almo_scf_qs.F",1093,"molecular domains and atomic groups - NYI")

                     ELSE

                        ! ao domains are atomic / electron groups are atomic
                        ! compute max overlap between atoms: domain_row and domain_col
                        CALL cp_dbcsr_get_block_p(matrix_s_sym, &
                                                  iblock_row, iblock_col, p_new_block, found)
                        IF (found) THEN
                           !   CPErrorMessage(cp_failure_level,routineP,"S block not found")
                           !   CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
                           overlap = MAXVAL(ABS(p_new_block))
                        ELSE
                           overlap = 0.0_dp
                        ENDIF

                     ENDIF

                  ENDIF ! end type of electron groups

                  s0 = -LOG10(ABS(almo_scf_env%quencher_s0))
                  s1 = -LOG10(ABS(almo_scf_env%quencher_s1))
                  IF (overlap .EQ. 0.0_dp) THEN
                     overlap = -LOG10(ABS(almo_scf_env%eps_filter))+100.0_dp
                  ELSE
                     overlap = -LOG10(overlap)
                  ENDIF
                  IF (s0 .LT. 0.0_dp) THEN
                     CALL cp__b("almo_scf_qs.F",1121,"S0 is less than zero")
                  ENDIF
                  IF (s1 .LE. 0.0_dp) THEN
                     CALL cp__b("almo_scf_qs.F",1124,"S1 is less than or equal to zero")
                  ENDIF
                  IF (s0 .GE. s1) THEN
                     CALL cp__b("almo_scf_qs.F",1127,"S0 is greater than or equal to S1")
                  ENDIF

                  ! Fill in non-zero blocks if AOs are close to the electron center
                  IF (overlap .LT. s1) THEN
                     NULLIFY (p_new_block)
                     CALL cp_dbcsr_reserve_block2d(almo_scf_env%quench_t(ispin), &
                                                   iblock_row, iblock_col, p_new_block)
                     IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_qs.F",1135)
                     IF (overlap .LE. s0) THEN
                        p_new_block(:, :) = 1.0_dp
                        !WRITE(*,'(A15,2I7,3F8.3,E11.3)') "INTRA-BLOCKS: ",&
                        !   iblock_col, iblock_row, s0, s1, overlap, p_new_block(1,1)
                     ELSE
                        p_new_block(:, :) = 1.0_dp/(1.0_dp+EXP(-(s0-s1)/(s0-overlap)-(s0-s1)/(overlap-s1)))
                        !WRITE(*,'(A15,2I7,3F8.3,E11.3)') "INTER-BLOCKS: ",&
                        !   iblock_col, iblock_row, s0, s1, overlap, p_new_block(1,1)
                     ENDIF

                     IF (ABS(p_new_block(1, 1)) .GT. ABS(almo_scf_env%eps_filter)) THEN
                        IF (domain_map_local_entries .GE. max_domain_neighbors*almo_scf_env%ndomains) THEN
                           CALL cp__b("almo_scf_qs.F",1148,"weird... max_domain_neighbors is exceeded")
                        ENDIF
                        almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 1) = iblock_row
                        almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 2) = iblock_col
                        domain_map_local_entries = domain_map_local_entries+1
                     ENDIF

                  ENDIF

               CASE (almo_constraint_distance)

                  ! type of electron groups
                  IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular) THEN

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are molecular

                        ! compute distance between molecules: domain_row and domain_col
                        ! distance between molecules is defined as the smallest
                        ! distance among all atom pairs
                        IF (domain_row == domain_col) THEN
                           distance = 0.0_dp
                           contact_atom_1 = first_atom_of_molecule(domain_row)
                           contact_atom_2 = first_atom_of_molecule(domain_col)
                        ELSE
                           distance_squared = 1.0E+100_dp
                           contact_atom_1 = -1
                           contact_atom_2 = -1
                           DO iatom = first_atom_of_molecule(domain_row), last_atom_of_molecule(domain_row)
                              DO jatom = first_atom_of_molecule(domain_col), last_atom_of_molecule(domain_col)
                                 rab(:) = pbc(particle_set(iatom)%r(:), particle_set(jatom)%r(:), cell)
                                 trial_distance_squared = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
                                 IF (trial_distance_squared .LT. distance_squared) THEN
                                    distance_squared = trial_distance_squared
                                    contact_atom_1 = iatom
                                    contact_atom_2 = jatom
                                 ENDIF
                              ENDDO ! jatom
                           ENDDO ! iatom
                           IF(.NOT.(contact_atom_1 .GT. 0))CALL cp__a("almo_scf_qs.F",1189)
                           distance = SQRT(distance_squared)
                        ENDIF

                     ELSE ! ao domains are atomic

                        ! ao domains are atomic / electron groups are molecular
                        !distance_between_atom_and_molecule(atom=domain_row,molecule=domain_col)
                        CALL cp__b("almo_scf_qs.F",1197,"atomic domains and molecular groups - NYI")

                     ENDIF

                  ELSE ! electron groups are atomic

                     ! type of ao domains
                     IF (almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN

                        ! ao domains are molecular / electron groups are atomic
                        !distance_between_atom_and_molecule(atom=domain_col,molecule=domain_row)
                        CALL cp__b("almo_scf_qs.F",1208,"molecular domains and atomic groups - NYI")

                     ELSE

                        ! ao domains are atomic / electron groups are atomic
                        ! compute distance between atoms: domain_row and domain_col
                        rab(:) = pbc(particle_set(domain_row)%r(:), particle_set(domain_col)%r(:), cell)
                        distance = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))
                        contact_atom_1 = domain_row
                        contact_atom_2 = domain_col

                     ENDIF

                  ENDIF ! end type of electron groups

                  ! get atomic radii to compute distance cutoff threshold
                  IF (almo_scf_env%quencher_radius_type == do_bondparm_covalent) THEN
                     CALL get_atomic_kind(atomic_kind=particle_set(contact_atom_1)%atomic_kind, &
                                          rcov=contact1_radius)
                     CALL get_atomic_kind(atomic_kind=particle_set(contact_atom_2)%atomic_kind, &
                                          rcov=contact2_radius)
                  ELSE IF (almo_scf_env%quencher_radius_type == do_bondparm_vdw) THEN
                     CALL get_atomic_kind(atomic_kind=particle_set(contact_atom_1)%atomic_kind, &
                                          rvdw=contact1_radius)
                     CALL get_atomic_kind(atomic_kind=particle_set(contact_atom_2)%atomic_kind, &
                                          rvdw=contact2_radius)
                  ELSE
                     CALL cp__b("almo_scf_qs.F",1235,"Illegal quencher_radius_type")
                  END IF
                  contact1_radius = cp_unit_to_cp2k(contact1_radius, "angstrom")
                  contact2_radius = cp_unit_to_cp2k(contact2_radius, "angstrom")

!RZK-warning the procedure is faulty for molecules: the closest contacts should be found using
! the element specific radii

                  ! compute inner and outer cutoff radii
                  r0 = almo_scf_env%quencher_r0_factor*(contact1_radius+contact2_radius)
                  !+almo_scf_env%quencher_r0_shift
                  r1 = almo_scf_env%quencher_r1_factor*(contact1_radius+contact2_radius)
                  !+almo_scf_env%quencher_r1_shift

                  IF (r0 .LT. 0.0_dp) THEN
                     CALL cp__b("almo_scf_qs.F",1250,"R0 is less than zero")
                  ENDIF
                  IF (r1 .LE. 0.0_dp) THEN
                     CALL cp__b("almo_scf_qs.F",1253,"R1 is less than or equal to zero")
                  ENDIF
                  IF (r0 .GT. r1) THEN
                     CALL cp__b("almo_scf_qs.F",1256,"R0 is greater than or equal to R1")
                  ENDIF

                  ! Fill in non-zero blocks if AOs are close to the electron center
                  IF (distance .LT. r1) THEN
                     NULLIFY (p_new_block)
                     CALL cp_dbcsr_reserve_block2d(almo_scf_env%quench_t(ispin), &
                                                   iblock_row, iblock_col, p_new_block)
                     IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_qs.F",1264)
                     IF (distance .LE. r0) THEN
                        p_new_block(:, :) = 1.0_dp
                        !WRITE(*,'(A15,2I7,5F8.3,E11.3)') "INTRA-BLOCKS: ",&
                        !   iblock_col, iblock_row, contact1_radius,&
                        !   contact2_radius, r0, r1, distance, p_new_block(1,1)
                     ELSE
                        ! remove the intermediate values from the quencher temporarily
                        CALL cp__b("almo_scf_qs.F",1272,"")
                        p_new_block(:, :) = 1.0_dp/(1.0_dp+EXP((r1-r0)/(r0-distance)+(r1-r0)/(r1-distance)))
                        !WRITE(*,'(A15,2I7,5F8.3,E11.3)') "INTER-BLOCKS: ",&
                        !   iblock_col, iblock_row, contact1_radius,&
                        !   contact2_radius, r0, r1, distance, p_new_block(1,1)
                     ENDIF

                     IF (ABS(p_new_block(1, 1)) .GT. ABS(almo_scf_env%eps_filter)) THEN
                        IF (domain_map_local_entries .GE. max_domain_neighbors*almo_scf_env%ndomains) THEN
                           CALL cp__b("almo_scf_qs.F",1281,"weird... max_domain_neighbors is exceeded")
                        ENDIF
                        almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 1) = iblock_row
                        almo_scf_env%domain_map(ispin)%pairs(domain_map_local_entries+1, 2) = iblock_col
                        domain_map_local_entries = domain_map_local_entries+1
                     ENDIF

                  ENDIF

               CASE DEFAULT
                  CALL cp__b("almo_scf_qs.F",1291,"Illegal constraint type")
               END SELECT

            ENDIF ! mynode

         ENDDO
      ENDDO ! end O(N) loop over pairs

      DEALLOCATE (domain_neighbor_list)
      DEALLOCATE (current_number_neighbors)

      CALL cp_dbcsr_finalize(almo_scf_env%quench_t(ispin))
      !CALL cp_dbcsr_scale(almo_scf_env%quench_t(ispin),&
      !        almo_scf_env%envelope_amplitude)
      CALL cp_dbcsr_filter(almo_scf_env%quench_t(ispin), &
                           almo_scf_env%eps_filter)

      ! check that both domain_map and quench_t have the same number of entries
      nblks = cp_dbcsr_get_num_blocks(almo_scf_env%quench_t(ispin))
      IF (nblks .NE. domain_map_local_entries) THEN
         CALL cp__b("almo_scf_qs.F",1311,"number of blocks is wrong")
      ENDIF

      ! communicate local parts of the domain map
      !nNodes = dbcsr_mp_numnodes(dbcsr_distribution_mp(&
      !   cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))
      !GroupID = dbcsr_mp_group(dbcsr_distribution_mp(&
      !   cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))

      ! first, communicate map sizes on the other nodes
      ALLOCATE (domain_entries_cpu(nNodes), offset_for_cpu(nNodes))
      CALL mp_allgather(2*domain_map_local_entries, domain_entries_cpu, GroupID)

      ! second, create
      offset_for_cpu(1) = 0
      DO iNode = 2, nNodes
         offset_for_cpu(iNode) = offset_for_cpu(iNode-1)+ &
                                 domain_entries_cpu(iNode-1)
      ENDDO
      global_entries = offset_for_cpu(nNodes)+domain_entries_cpu(nNodes)

      ! communicate all entries
      ALLOCATE (domain_map_global(global_entries))
      ALLOCATE (domain_map_local(2*domain_map_local_entries))
      DO ientry = 1, domain_map_local_entries
         domain_map_local(2*(ientry-1)+1) = almo_scf_env%domain_map(ispin)%pairs(ientry, 1)
         domain_map_local(2*ientry) = almo_scf_env%domain_map(ispin)%pairs(ientry, 2)
      ENDDO
      CALL mp_allgather(domain_map_local, domain_map_global, &
                        domain_entries_cpu, offset_for_cpu, GroupID)
      DEALLOCATE (domain_entries_cpu, offset_for_cpu)
      DEALLOCATE (domain_map_local)

      DEALLOCATE (almo_scf_env%domain_map(ispin)%index1)
      DEALLOCATE (almo_scf_env%domain_map(ispin)%pairs)
      ALLOCATE (almo_scf_env%domain_map(ispin)%index1(ndomains))
      ALLOCATE (almo_scf_env%domain_map(ispin)%pairs(global_entries/2, 2))
      almo_scf_env%domain_map(ispin)%pairs(:, :) = 0
      almo_scf_env%domain_map(ispin)%index1(:) = 0

      ! unpack the received data into a local variable
      ! since we do not know the maximum global number of neighbors
      ! try one. if fails increase the maximum number and try again
      ! until it succeeds
      max_neig = max_domain_neighbors
      max_neig_fails = .TRUE.
      max_neig_loop: DO WHILE (max_neig_fails)
         ALLOCATE (domain_grid(almo_scf_env%ndomains, 0:max_neig))
         domain_grid(:, :) = 0
         ! init the number of collected neighbors
         domain_grid(:, 0) = 1
         ! loop over the records
         global_entries = global_entries/2
         DO ientry = 1, global_entries
            ! get the center
            grid1 = domain_map_global(2*ientry)
            ! get the neighbor
            ineig = domain_map_global(2*(ientry-1)+1)
            ! check boundaries
            IF (domain_grid(grid1, 0) .GT. max_neig) THEN
               ! this neighbor will overstep the boundaries
               ! stop the trial and increase the max number of neigbors
               DEALLOCATE (domain_grid)
               max_neig = max_neig*2
               CYCLE max_neig_loop
            ENDIF
            ! for the current center loop over the collected neighbors
            ! to insert the current record in a numerical order
            delayed_increment = .FALSE.
            DO igrid = 1, domain_grid(grid1, 0)
               ! compare the current neighbor with that already in the 'book'
               IF (ineig .LT. domain_grid(grid1, igrid)) THEN
                  ! if this one is smaller then insert it here and pick up the one
                  ! from the book to continue inserting
                  neig_temp = ineig
                  ineig = domain_grid(grid1, igrid)
                  domain_grid(grid1, igrid) = neig_temp
               ELSE
                  IF (domain_grid(grid1, igrid) .EQ. 0) THEN
                     ! got the empty slot now - insert the record
                     domain_grid(grid1, igrid) = ineig
                     ! increase the record counter but do it outside the loop
                     delayed_increment = .TRUE.
                  ENDIF
               ENDIF
            ENDDO
            IF (delayed_increment) THEN
               domain_grid(grid1, 0) = domain_grid(grid1, 0)+1
            ELSE
               ! should not be here - all records must be inserted
               CALL cp__b("almo_scf_qs.F",1401,"all records must be inserted")
            ENDIF
         ENDDO
         max_neig_fails = .FALSE.
      ENDDO max_neig_loop
      DEALLOCATE (domain_map_global)

      ientry = 1
      DO idomain = 1, almo_scf_env%ndomains
         DO ineig = 1, domain_grid(idomain, 0)-1
            almo_scf_env%domain_map(ispin)%pairs(ientry, 1) = domain_grid(idomain, ineig)
            almo_scf_env%domain_map(ispin)%pairs(ientry, 2) = idomain
            ientry = ientry+1
         ENDDO
         almo_scf_env%domain_map(ispin)%index1(idomain) = ientry
      ENDDO
      DEALLOCATE (domain_grid)

      !ENDDO ! ispin
      IF (almo_scf_env%nspins .EQ. 2) THEN
         CALL cp_dbcsr_copy(almo_scf_env%quench_t(2), &
                            almo_scf_env%quench_t(1))
         almo_scf_env%domain_map(2)%pairs(:, :) = &
            almo_scf_env%domain_map(1)%pairs(:, :)
         almo_scf_env%domain_map(2)%index1(:) = &
            almo_scf_env%domain_map(1)%index1(:)
      ENDIF

      CALL cp_dbcsr_release(matrix_s_sym)

      IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular .OR. &
          almo_scf_env%domain_layout_aos == almo_domain_layout_molecular) THEN
         DEALLOCATE (first_atom_of_molecule)
         DEALLOCATE (last_atom_of_molecule)
      ENDIF

      !mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(&
      !   cp_dbcsr_distribution(almo_scf_env%quench_t(ispin))))
      !nblkrows_tot = cp_dbcsr_nblkrows_total(almo_scf_env%quench_t(ispin))
      !nblkcols_tot = cp_dbcsr_nblkcols_total(almo_scf_env%quench_t(ispin))
      !DO row = 1, nblkrows_tot
      !   DO col = 1, nblkcols_tot
      !      tr = .FALSE.
      !      iblock_row = row
      !      iblock_col = col
      !      CALL cp_dbcsr_get_stored_coordinates(almo_scf_env%quench_t(ispin),&
      !              iblock_row, iblock_col, tr, hold)
      !      CALL cp_dbcsr_get_block_p(almo_scf_env%quench_t(ispin),&
      !              row, col, p_new_block, found)
      !      write(*,*) "RST_NOTE:", mynode, row, col, hold, found
      !   ENDDO
      !ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_construct_quencher

END MODULE almo_scf_qs

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types for all cayley transformation methods
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE ct_types
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: cg_polak_ribiere,&
                                              tensor_orthogonal
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ct_types'

   ! Public types
   PUBLIC :: ct_step_env_type

   ! Public subroutines
   PUBLIC :: ct_step_env_init, ct_step_env_set, ct_step_env_get, ct_step_env_clean

   TYPE ct_step_env_type

      ! this type contains options for cayley transformation routines

      ! use orbitals or projectors?
      LOGICAL :: use_occ_orbs, use_virt_orbs
      LOGICAL :: occ_orbs_orthogonal, virt_orbs_orthogonal
      ! tensor properties of matrix indeces:
      ! tensor_up_down, tensor_orthogonal
      INTEGER :: tensor_type
      ! neglect the quadratic term in riccati equations?
      LOGICAL :: neglect_quadratic_term
      ! what kind of output do we produce?
      LOGICAL :: update_p, update_q, calculate_energy_corr
      ! variety of conjugate gradient
      INTEGER :: conjugator

      ! type of preconditioner
      LOGICAL :: pp_preconditioner_full, &
                 qq_preconditioner_full

      REAL(KIND=dp)         :: eps_convergence
      REAL(KIND=dp)         :: eps_filter
      INTEGER               :: max_iter
      !INTEGER               :: nspins
      LOGICAL               :: converged
      INTEGER               :: order_lanczos
      REAL(KIND=dp)         :: eps_lancsoz
      INTEGER               :: max_iter_lanczos

      REAL(KIND=dp)         :: energy_correction

!SPIN!!!    ! metric matrices for covariant to contravariant transformations
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: p_index_up=>NULL()
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: p_index_down=>NULL()
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: q_index_up=>NULL()
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: q_index_down=>NULL()
!SPIN!!!
!SPIN!!!    ! kohn-sham, covariant-covariant representation
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_ks=>NULL()
!SPIN!!!    ! density, contravariant-contravariant representation
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_p=>NULL()
!SPIN!!!    ! occ orbitals, contravariant-covariant representation
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_t=>NULL()
!SPIN!!!    ! virt orbitals, contravariant-covariant representation
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_v=>NULL()
!SPIN!!!
!SPIN!!!    ! to avoid bulding Occ-by-N and Virt-vy-N matrices inside
!SPIN!!!    ! the ct routines get them from the external code
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_qp_template=>NULL()
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER  :: matrix_pq_template=>NULL()
!SPIN!!!
!SPIN!!!    ! single excitation amplitudes
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE  :: matrix_x
!SPIN!!!    ! residuals
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE  :: matrix_res

      ! metric matrices for covariant to contravariant transformations
      TYPE(cp_dbcsr_type), POINTER  :: p_index_up => NULL()
      TYPE(cp_dbcsr_type), POINTER  :: p_index_down => NULL()
      TYPE(cp_dbcsr_type), POINTER  :: q_index_up => NULL()
      TYPE(cp_dbcsr_type), POINTER  :: q_index_down => NULL()

      ! kohn-sham, covariant-covariant representation
      TYPE(cp_dbcsr_type), POINTER  :: matrix_ks => NULL()
      ! density, contravariant-contravariant representation
      TYPE(cp_dbcsr_type), POINTER  :: matrix_p => NULL()
      ! occ orbitals, contravariant-covariant representation
      TYPE(cp_dbcsr_type), POINTER  :: matrix_t => NULL()
      ! virt orbitals, contravariant-covariant representation
      TYPE(cp_dbcsr_type), POINTER  :: matrix_v => NULL()

      ! to avoid bulding Occ-by-N and Virt-vy-N matrices inside
      ! the ct routines get them from the external code
      TYPE(cp_dbcsr_type), POINTER  :: matrix_qp_template => NULL()
      TYPE(cp_dbcsr_type), POINTER  :: matrix_pq_template => NULL()

      ! guess for single excitation amplitudes
      ! it is used exclusively as a guess, not modified
      ! it should be given in the up_down representation
      TYPE(cp_dbcsr_type), POINTER  :: matrix_x_guess => NULL()

      ! single excitation amplitudes
      TYPE(cp_dbcsr_type)           :: matrix_x
      ! residuals
      TYPE(cp_dbcsr_type)           :: matrix_res

      TYPE(cp_para_env_type), POINTER  :: para_env => NULL()
      TYPE(cp_blacs_env_type), POINTER  :: blacs_env => NULL()

   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param env ...
! **************************************************************************************************
   SUBROUTINE ct_step_env_init(env)

      TYPE(ct_step_env_type)                             :: env

      CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_env_init', &
         routineP = moduleN//':'//routineN

      env%use_occ_orbs = .TRUE.
      env%use_virt_orbs = .FALSE.
      env%occ_orbs_orthogonal = .FALSE.
      env%virt_orbs_orthogonal = .FALSE.
      env%tensor_type = tensor_orthogonal
      env%neglect_quadratic_term = .FALSE.
      env%calculate_energy_corr = .TRUE.
      env%update_p = .FALSE.
      env%update_q = .FALSE.
      env%pp_preconditioner_full = .TRUE.
      env%qq_preconditioner_full = .FALSE.

      env%eps_convergence = 1.0E-8_dp
      env%eps_filter = 1.0E-8_dp
      env%max_iter = 400
      env%order_lanczos = 3
      env%eps_lancsoz = 1.0E-4_dp
      env%max_iter_lanczos = 40
      !env%nspins = -1
      env%converged = .FALSE.
      env%conjugator = cg_polak_ribiere

      NULLIFY (env%p_index_up)
      NULLIFY (env%p_index_down)
      NULLIFY (env%q_index_up)
      NULLIFY (env%q_index_down)

      NULLIFY (env%matrix_ks)
      NULLIFY (env%matrix_p)
      NULLIFY (env%matrix_t)
      NULLIFY (env%matrix_v)
      NULLIFY (env%matrix_x_guess)
      NULLIFY (env%matrix_qp_template)
      NULLIFY (env%matrix_pq_template)

      !RZK-warning read_parameters_from_input

   END SUBROUTINE ct_step_env_init

! **************************************************************************************************
!> \brief ...
!> \param env ...
!> \param use_occ_orbs ...
!> \param use_virt_orbs ...
!> \param tensor_type ...
!> \param occ_orbs_orthogonal ...
!> \param virt_orbs_orthogonal ...
!> \param neglect_quadratic_term ...
!> \param update_p ...
!> \param update_q ...
!> \param eps_convergence ...
!> \param eps_filter ...
!> \param max_iter ...
!> \param p_index_up ...
!> \param p_index_down ...
!> \param q_index_up ...
!> \param q_index_down ...
!> \param matrix_ks ...
!> \param matrix_p ...
!> \param matrix_qp_template ...
!> \param matrix_pq_template ...
!> \param matrix_t ...
!> \param matrix_v ...
!> \param copy_matrix_x ...
!> \param energy_correction ...
!> \param calculate_energy_corr ...
!> \param converged ...
!> \param qq_preconditioner_full ...
!> \param pp_preconditioner_full ...
! **************************************************************************************************
   SUBROUTINE ct_step_env_get(env, use_occ_orbs, use_virt_orbs, tensor_type, &
                              occ_orbs_orthogonal, virt_orbs_orthogonal, neglect_quadratic_term, &
                              update_p, update_q, eps_convergence, eps_filter, max_iter, &
                              p_index_up, p_index_down, q_index_up, q_index_down, matrix_ks, matrix_p, &
                              matrix_qp_template, matrix_pq_template, &
                              matrix_t, matrix_v, copy_matrix_x, energy_correction, calculate_energy_corr, &
                              converged, qq_preconditioner_full, pp_preconditioner_full)

      TYPE(ct_step_env_type)                             :: env
      LOGICAL, OPTIONAL                                  :: use_occ_orbs, use_virt_orbs
      INTEGER, OPTIONAL                                  :: tensor_type
      LOGICAL, OPTIONAL                                  :: occ_orbs_orthogonal, &
                                                            virt_orbs_orthogonal, &
                                                            neglect_quadratic_term, update_p, &
                                                            update_q
      REAL(KIND=dp), OPTIONAL                            :: eps_convergence, eps_filter
      INTEGER, OPTIONAL                                  :: max_iter
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER :: p_index_up, p_index_down, q_index_up, &
         q_index_down, matrix_ks, matrix_p, matrix_qp_template, matrix_pq_template, matrix_t, &
         matrix_v
      TYPE(cp_dbcsr_type), OPTIONAL                      :: copy_matrix_x
      REAL(KIND=dp), OPTIONAL                            :: energy_correction
      LOGICAL, OPTIONAL                                  :: calculate_energy_corr, converged, &
                                                            qq_preconditioner_full, &
                                                            pp_preconditioner_full

      CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_env_get', &
         routineP = moduleN//':'//routineN

!INTEGER      , OPTIONAL                     :: nspins
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: p_index_up
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: p_index_down
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: q_index_up
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: q_index_down
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_ks
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_p
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_t
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_v
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_qp_template
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_pq_template
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), POINTER, OPTIONAL  :: matrix_x
!SPIN!!!
!SPIN!!!    TYPE(cp_dbcsr_type), DIMENSION(:), OPTIONAL           :: copy_matrix_x
!INTEGER                                               :: ispin

      IF (PRESENT(use_occ_orbs)) use_occ_orbs = env%use_occ_orbs
      IF (PRESENT(use_virt_orbs)) use_virt_orbs = env%use_virt_orbs
      IF (PRESENT(occ_orbs_orthogonal)) occ_orbs_orthogonal = &
         env%occ_orbs_orthogonal
      IF (PRESENT(virt_orbs_orthogonal)) virt_orbs_orthogonal = &
         env%virt_orbs_orthogonal
      IF (PRESENT(tensor_type)) tensor_type = env%tensor_type
      IF (PRESENT(neglect_quadratic_term)) neglect_quadratic_term = &
         env%neglect_quadratic_term
      IF (PRESENT(calculate_energy_corr)) calculate_energy_corr = &
         env%calculate_energy_corr
      IF (PRESENT(update_p)) update_p = env%update_p
      IF (PRESENT(update_q)) update_q = env%update_q
      IF (PRESENT(pp_preconditioner_full)) pp_preconditioner_full = &
         env%pp_preconditioner_full
      IF (PRESENT(qq_preconditioner_full)) qq_preconditioner_full = &
         env%qq_preconditioner_full
      IF (PRESENT(eps_convergence)) eps_convergence = env%eps_convergence
      IF (PRESENT(eps_filter)) eps_filter = env%eps_filter
      IF (PRESENT(max_iter)) max_iter = env%max_iter
      !IF (PRESENT(nspins)) nspins = env%nspins
      IF (PRESENT(matrix_ks)) matrix_ks => env%matrix_ks
      IF (PRESENT(matrix_p)) matrix_p => env%matrix_p
      IF (PRESENT(matrix_t)) matrix_t => env%matrix_t
      IF (PRESENT(matrix_v)) matrix_v => env%matrix_v
      IF (PRESENT(matrix_qp_template)) matrix_qp_template => &
         env%matrix_qp_template
      IF (PRESENT(matrix_pq_template)) matrix_pq_template => &
         env%matrix_pq_template
      IF (PRESENT(p_index_up)) p_index_up => env%p_index_up
      IF (PRESENT(q_index_up)) q_index_up => env%q_index_up
      IF (PRESENT(p_index_down)) p_index_down => env%p_index_down
      IF (PRESENT(q_index_down)) q_index_down => env%q_index_down
      IF (PRESENT(copy_matrix_x)) THEN
         !DO ispin=1,env%nspins
         !CALL cp_dbcsr_copy(copy_matrix_x(ispin),env%matrix_x(ispin))
         CALL cp_dbcsr_copy(copy_matrix_x, env%matrix_x)
         !ENDDO
      ENDIF
      !IF (PRESENT(matrix_x)) matrix_x => env%matrix_x
      IF (PRESENT(energy_correction)) energy_correction = env%energy_correction
      IF (PRESENT(converged)) converged = env%converged

   END SUBROUTINE ct_step_env_get

! **************************************************************************************************
!> \brief ...
!> \param env ...
!> \param para_env ...
!> \param blacs_env ...
!> \param use_occ_orbs ...
!> \param use_virt_orbs ...
!> \param tensor_type ...
!> \param occ_orbs_orthogonal ...
!> \param virt_orbs_orthogonal ...
!> \param neglect_quadratic_term ...
!> \param update_p ...
!> \param update_q ...
!> \param eps_convergence ...
!> \param eps_filter ...
!> \param max_iter ...
!> \param p_index_up ...
!> \param p_index_down ...
!> \param q_index_up ...
!> \param q_index_down ...
!> \param matrix_ks ...
!> \param matrix_p ...
!> \param matrix_qp_template ...
!> \param matrix_pq_template ...
!> \param matrix_t ...
!> \param matrix_v ...
!> \param matrix_x_guess ...
!> \param calculate_energy_corr ...
!> \param conjugator ...
!> \param qq_preconditioner_full ...
!> \param pp_preconditioner_full ...
! **************************************************************************************************
   SUBROUTINE ct_step_env_set(env, para_env, blacs_env, use_occ_orbs, &
                              use_virt_orbs, tensor_type, &
                              occ_orbs_orthogonal, virt_orbs_orthogonal, neglect_quadratic_term, &
                              update_p, update_q, eps_convergence, eps_filter, max_iter, &
                              p_index_up, p_index_down, q_index_up, q_index_down, matrix_ks, matrix_p, &
                              matrix_qp_template, matrix_pq_template, &
                              matrix_t, matrix_v, matrix_x_guess, calculate_energy_corr, conjugator, &
                              qq_preconditioner_full, pp_preconditioner_full)

      TYPE(ct_step_env_type)                             :: env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      LOGICAL, OPTIONAL                                  :: use_occ_orbs, use_virt_orbs
      INTEGER, OPTIONAL                                  :: tensor_type
      LOGICAL, OPTIONAL                                  :: occ_orbs_orthogonal, &
                                                            virt_orbs_orthogonal, &
                                                            neglect_quadratic_term, update_p, &
                                                            update_q
      REAL(KIND=dp), OPTIONAL                            :: eps_convergence, eps_filter
      INTEGER, OPTIONAL                                  :: max_iter
      TYPE(cp_dbcsr_type), OPTIONAL, TARGET :: p_index_up, p_index_down, q_index_up, q_index_down, &
         matrix_ks, matrix_p, matrix_qp_template, matrix_pq_template, matrix_t, matrix_v, &
         matrix_x_guess
      LOGICAL, OPTIONAL                                  :: calculate_energy_corr
      INTEGER, OPTIONAL                                  :: conjugator
      LOGICAL, OPTIONAL                                  :: qq_preconditioner_full, &
                                                            pp_preconditioner_full

      CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_env_set', &
         routineP = moduleN//':'//routineN

!INTEGER      , OPTIONAL                     :: nspins
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: p_index_up
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: p_index_down
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: q_index_up
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: q_index_down
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_ks
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_p
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_t
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_v
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_qp_template
!SPIN!!!    TYPE(cp_dbcsr_type), TARGET, DIMENSION(:), OPTIONAL  :: matrix_pq_template
! set para_env and blacs_env which are needed to operate with full matrices
! it would be nice to have everything with cp_dbcsr matrices, well maybe later

      env%para_env => para_env
      env%blacs_env => blacs_env

      IF (PRESENT(use_occ_orbs)) env%use_occ_orbs = use_occ_orbs
      IF (PRESENT(use_virt_orbs)) env%use_virt_orbs = use_virt_orbs
      IF (PRESENT(occ_orbs_orthogonal)) env%occ_orbs_orthogonal = &
         occ_orbs_orthogonal
      IF (PRESENT(virt_orbs_orthogonal)) env%virt_orbs_orthogonal = &
         virt_orbs_orthogonal
      IF (PRESENT(tensor_type)) env%tensor_type = tensor_type
      IF (PRESENT(neglect_quadratic_term)) env%neglect_quadratic_term = &
         neglect_quadratic_term
      IF (PRESENT(calculate_energy_corr)) env%calculate_energy_corr = &
         calculate_energy_corr
      IF (PRESENT(update_p)) env%update_p = update_p
      IF (PRESENT(update_q)) env%update_q = update_q
      IF (PRESENT(pp_preconditioner_full)) env%pp_preconditioner_full = &
         pp_preconditioner_full
      IF (PRESENT(qq_preconditioner_full)) env%qq_preconditioner_full = &
         qq_preconditioner_full
      IF (PRESENT(eps_convergence)) env%eps_convergence = eps_convergence
      IF (PRESENT(eps_filter)) env%eps_filter = eps_filter
      IF (PRESENT(max_iter)) env%max_iter = max_iter
      !IF (PRESENT(nspins)) env%nspins = nspins
      IF (PRESENT(conjugator)) env%conjugator = conjugator
      IF (PRESENT(matrix_ks)) env%matrix_ks => matrix_ks
      IF (PRESENT(matrix_p)) env%matrix_p => matrix_p
      IF (PRESENT(matrix_t)) env%matrix_t => matrix_t
      IF (PRESENT(matrix_v)) env%matrix_v => matrix_v
      IF (PRESENT(matrix_x_guess)) env%matrix_x_guess => matrix_x_guess
      IF (PRESENT(matrix_qp_template)) env%matrix_qp_template => &
         matrix_qp_template
      IF (PRESENT(matrix_pq_template)) env%matrix_pq_template => &
         matrix_pq_template
      IF (PRESENT(p_index_up)) env%p_index_up => p_index_up
      IF (PRESENT(q_index_up)) env%q_index_up => q_index_up
      IF (PRESENT(p_index_down)) env%p_index_down => p_index_down
      IF (PRESENT(q_index_down)) env%q_index_down => q_index_down

   END SUBROUTINE ct_step_env_set

! **************************************************************************************************
!> \brief ...
!> \param env ...
! **************************************************************************************************
   SUBROUTINE ct_step_env_clean(env)

      TYPE(ct_step_env_type)                             :: env

      CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_env_clean', &
         routineP = moduleN//':'//routineN

!INTEGER                                     :: ispin

      NULLIFY (env%para_env)
      NULLIFY (env%blacs_env)

      !DO ispin=1,env%nspins
      CALL cp_dbcsr_release(env%matrix_x)
      CALL cp_dbcsr_release(env%matrix_res)
      !CALL cp_dbcsr_release(env%matrix_x(ispin))
      !CALL cp_dbcsr_release(env%matrix_res(ispin))
      !ENDDO
      !DEALLOCATE(env%matrix_x,env%matrix_res)

      NULLIFY (env%p_index_up)
      NULLIFY (env%p_index_down)
      NULLIFY (env%q_index_up)
      NULLIFY (env%q_index_down)

      NULLIFY (env%matrix_ks)
      NULLIFY (env%matrix_p)
      NULLIFY (env%matrix_t)
      NULLIFY (env%matrix_v)
      NULLIFY (env%matrix_x_guess)
      NULLIFY (env%matrix_qp_template)
      NULLIFY (env%matrix_pq_template)

   END SUBROUTINE ct_step_env_clean

END MODULE ct_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A DIIS implementation for the ALMO-based SCF methods
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf_diis_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_create,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_trace,&
                                              cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE domain_submatrix_methods,        ONLY: add_submatrices,&
                                              copy_submatrices,&
                                              init_submatrices,&
                                              release_submatrices,&
                                              set_submatrices
   USE domain_submatrix_types,          ONLY: domain_submatrix_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER, PARAMETER :: diis_error_orthogonal = 1

   INTEGER, PARAMETER :: diis_env_dbcsr = 1
   INTEGER, PARAMETER :: diis_env_domain = 2

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf_diis_types'

   PUBLIC :: almo_scf_diis_type, &
             almo_scf_diis_init, almo_scf_diis_release, almo_scf_diis_push, &
             almo_scf_diis_extrapolate

   INTERFACE almo_scf_diis_init
      MODULE PROCEDURE almo_scf_diis_init_dbcsr
      MODULE PROCEDURE almo_scf_diis_init_domain
   END INTERFACE

   TYPE almo_scf_diis_type

      INTEGER :: diis_env_type

      INTEGER :: buffer_length
      INTEGER :: max_buffer_length
      !INTEGER, DIMENSION(:), ALLOCATABLE :: history_index

      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: m_var
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: m_err

      ! first dimension is history index, second - domain index
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: d_var
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: d_err

      ! distributed matrix of error overlaps
      TYPE(domain_submatrix_type), DIMENSION(:), ALLOCATABLE     :: m_b

      ! insertion point
      INTEGER :: in_point

      ! in order to calculate the overlap between error vectors
      ! it is desirable to know tensorial properties of the error
      ! vector, e.g. convariant, contravariant, orthogonal
      INTEGER :: error_type

   END TYPE almo_scf_diis_type

CONTAINS

! **************************************************************************************************
!> \brief initializes the diis structure
!> \param diis_env ...
!> \param sample_err ...
!> \param sample_var ...
!> \param error_type ...
!> \param max_length ...
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_diis_init_dbcsr(diis_env, sample_err, sample_var, error_type, &
                                       max_length)

      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: sample_err, sample_var
      INTEGER, INTENT(IN)                                :: error_type, max_length

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_init_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, im, ndomains

      CALL timeset(routineN, handle)

      IF (max_length .LE. 0) THEN
         CALL cp__b("almo_scf_diis_types.F",109,"DIIS: max_length is less than zero")
      END IF

      diis_env%diis_env_type = diis_env_dbcsr

      diis_env%max_buffer_length = max_length
      diis_env%buffer_length = 0
      diis_env%error_type = error_type
      diis_env%in_point = 1

      ALLOCATE (diis_env%m_err(diis_env%max_buffer_length))
      ALLOCATE (diis_env%m_var(diis_env%max_buffer_length))

      ! create matrices
      DO im = 1, diis_env%max_buffer_length
         CALL cp_dbcsr_init(diis_env%m_err(im))
         CALL cp_dbcsr_create(diis_env%m_err(im), &
                              template=sample_err)
         CALL cp_dbcsr_init(diis_env%m_var(im))
         CALL cp_dbcsr_create(diis_env%m_var(im), &
                              template=sample_var)
      ENDDO

      ! current B matrices are only 1-by-1, they will be expanded on-the-fly
      ! only one matrix is used with dbcsr version of DIIS
      ndomains = 1
      ALLOCATE (diis_env%m_b(ndomains))
      CALL init_submatrices(diis_env%m_b)
      ! hack into d_b structure to gain full control
      diis_env%m_b(:)%domain = 100 ! arbitrary positive number
      DO idomain = 1, ndomains
         IF (diis_env%m_b(idomain)%domain .GT. 0) THEN
            ALLOCATE (diis_env%m_b(idomain)%mdata(1, 1))
            diis_env%m_b(idomain)%mdata(:, :) = 0.0_dp
         ENDIF
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_diis_init_dbcsr

! **************************************************************************************************
!> \brief initializes the diis structure
!> \param diis_env ...
!> \param sample_err ...
!> \param error_type ...
!> \param max_length ...
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_diis_init_domain(diis_env, sample_err, error_type, &
                                        max_length)

      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN)                                      :: sample_err
      INTEGER, INTENT(IN)                                :: error_type, max_length

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_init_domain', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, ndomains

      CALL timeset(routineN, handle)

      IF (max_length .LE. 0) THEN
         CALL cp__b("almo_scf_diis_types.F",176,"DIIS: max_length is less than zero")
      END IF

      diis_env%diis_env_type = diis_env_domain

      diis_env%max_buffer_length = max_length
      diis_env%buffer_length = 0
      diis_env%error_type = error_type
      diis_env%in_point = 1

      ndomains = SIZE(sample_err)

      ALLOCATE (diis_env%d_err(diis_env%max_buffer_length, ndomains))
      ALLOCATE (diis_env%d_var(diis_env%max_buffer_length, ndomains))

      ! create matrices
      CALL init_submatrices(diis_env%d_var)
      CALL init_submatrices(diis_env%d_err)

      ! current B matrices are only 1-by-1, they will be expanded on-the-fly
      ALLOCATE (diis_env%m_b(ndomains))
      CALL init_submatrices(diis_env%m_b)
      ! hack into d_b structure to gain full control
      ! distribute matrices as the err/var matrices
      diis_env%m_b(:)%domain = sample_err(:)%domain
      DO idomain = 1, ndomains
         IF (diis_env%m_b(idomain)%domain .GT. 0) THEN
            ALLOCATE (diis_env%m_b(idomain)%mdata(1, 1))
            diis_env%m_b(idomain)%mdata(:, :) = 0.0_dp
         ENDIF
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_diis_init_domain

! **************************************************************************************************
!> \brief adds a variable-error pair to the diis structure
!> \param diis_env ...
!> \param var ...
!> \param err ...
!> \param d_var ...
!> \param d_err ...
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_diis_push(diis_env, var, err, d_var, d_err)
      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: var, err
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(IN), OPTIONAL                            :: d_var, d_err

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_push', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, in_point, irow, &
                                                            ndomains, old_buffer_length
      REAL(KIND=dp)                                      :: trace0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: m_b_tmp

      CALL timeset(routineN, handle)

      IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
         IF (.NOT. (PRESENT(var) .AND. PRESENT(err))) THEN
            CALL cp__b("almo_scf_diis_types.F",241,"provide DBCSR matrices")
         ENDIF
      ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
         IF (.NOT. (PRESENT(d_var) .AND. PRESENT(d_err))) THEN
            CALL cp__b("almo_scf_diis_types.F",245,"provide domain submatrices")
         ENDIF
      ELSE
         CALL cp__b("almo_scf_diis_types.F",248,"illegal DIIS ENV type")
      ENDIF

      in_point = diis_env%in_point

      ! store a var-error pair
      IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
         CALL cp_dbcsr_copy(diis_env%m_var(in_point), var)
         CALL cp_dbcsr_copy(diis_env%m_err(in_point), err)
      ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
         CALL copy_submatrices(d_var, diis_env%d_var(in_point, :), copy_data=.TRUE.)
         CALL copy_submatrices(d_err, diis_env%d_err(in_point, :), copy_data=.TRUE.)
      ENDIF

      ! update the buffer length
      old_buffer_length = diis_env%buffer_length
      diis_env%buffer_length = diis_env%buffer_length+1
      IF (diis_env%buffer_length .GT. diis_env%max_buffer_length) &
         diis_env%buffer_length = diis_env%max_buffer_length

      !!!! resize B matrix
      !!!IF (old_buffer_length.lt.diis_env%buffer_length) THEN
      !!!   ALLOCATE(m_b_tmp(diis_env%buffer_length+1,diis_env%buffer_length+1))
      !!!   m_b_tmp(1:diis_env%buffer_length,1:diis_env%buffer_length)=&
      !!!      diis_env%m_b(:,:)
      !!!   DEALLOCATE(diis_env%m_b)
      !!!   ALLOCATE(diis_env%m_b(diis_env%buffer_length+1,&
      !!!      diis_env%buffer_length+1))
      !!!   diis_env%m_b(:,:)=m_b_tmp(:,:)
      !!!   DEALLOCATE(m_b_tmp)
      !!!ENDIF
      !!!! update B matrix elements
      !!!diis_env%m_b(1,in_point+1)=-1.0_dp
      !!!diis_env%m_b(in_point+1,1)=-1.0_dp
      !!!DO irow=1,diis_env%buffer_length
      !!!   trace0=almo_scf_diis_error_overlap(diis_env,&
      !!!      A=diis_env%m_err(irow),B=diis_env%m_err(in_point))
      !!!
      !!!   diis_env%m_b(irow+1,in_point+1)=trace0
      !!!   diis_env%m_b(in_point+1,irow+1)=trace0
      !!!ENDDO

      ! resize B matrix and update its elements
      ndomains = SIZE(diis_env%m_b)
      IF (old_buffer_length .LT. diis_env%buffer_length) THEN
         ALLOCATE (m_b_tmp(diis_env%buffer_length+1, diis_env%buffer_length+1))
         DO idomain = 1, ndomains
            IF (diis_env%m_b(idomain)%domain .GT. 0) THEN
               m_b_tmp(:, :) = 0.0_dp
               m_b_tmp(1:diis_env%buffer_length, 1:diis_env%buffer_length) = &
                  diis_env%m_b(idomain)%mdata(:, :)
               DEALLOCATE (diis_env%m_b(idomain)%mdata)
               ALLOCATE (diis_env%m_b(idomain)%mdata(diis_env%buffer_length+1, &
                                                     diis_env%buffer_length+1))
               diis_env%m_b(idomain)%mdata(:, :) = m_b_tmp(:, :)
            ENDIF
         ENDDO
         DEALLOCATE (m_b_tmp)
      ENDIF
      DO idomain = 1, ndomains
         IF (diis_env%m_b(idomain)%domain .GT. 0) THEN
            diis_env%m_b(idomain)%mdata(1, in_point+1) = -1.0_dp
            diis_env%m_b(idomain)%mdata(in_point+1, 1) = -1.0_dp
            DO irow = 1, diis_env%buffer_length
               IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
                  trace0 = almo_scf_diis_error_overlap(diis_env, &
                                                       A=diis_env%m_err(irow), B=diis_env%m_err(in_point))
               ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
                  trace0 = almo_scf_diis_error_overlap(diis_env, &
                                                       d_A=diis_env%d_err(irow, idomain), &
                                                       d_B=diis_env%d_err(in_point, idomain))
               ENDIF
               diis_env%m_b(idomain)%mdata(irow+1, in_point+1) = trace0
               diis_env%m_b(idomain)%mdata(in_point+1, irow+1) = trace0
            ENDDO ! loop over prev errors
         ENDIF
      ENDDO ! loop over domains

      ! update the insertion point for the next "PUSH"
      diis_env%in_point = diis_env%in_point+1
      IF (diis_env%in_point .GT. diis_env%max_buffer_length) diis_env%in_point = 1

      CALL timestop(handle)

   END SUBROUTINE almo_scf_diis_push

! **************************************************************************************************
!> \brief extrapolates the variable using the saved history
!> \param diis_env ...
!> \param extr_var ...
!> \param d_extr_var ...
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_diis_extrapolate(diis_env, extr_var, d_extr_var)
      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: extr_var
      TYPE(domain_submatrix_type), DIMENSION(:), &
         INTENT(INOUT), OPTIONAL                         :: d_extr_var

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_extrapolate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idomain, im, INFO, LWORK, &
                                                            ndomains, unit_nr
      REAL(KIND=dp)                                      :: checksum
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: coeff, eigenvalues, tmp1, WORK
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: m_b_copy
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
         IF (.NOT. PRESENT(extr_var)) THEN
            CALL cp__b("almo_scf_diis_types.F",371,"provide DBCSR matrix")
         ENDIF
      ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
         IF (.NOT. PRESENT(d_extr_var)) THEN
            CALL cp__b("almo_scf_diis_types.F",375,"provide domain submatrices")
         ENDIF
      ELSE
         CALL cp__b("almo_scf_diis_types.F",378,"illegal DIIS ENV type")
      ENDIF

      ! Prepare data
      ALLOCATE (eigenvalues(diis_env%buffer_length+1))
      ALLOCATE (m_b_copy(diis_env%buffer_length+1, diis_env%buffer_length+1))

      ndomains = SIZE(diis_env%m_b)

      DO idomain = 1, ndomains

         IF (diis_env%m_b(idomain)%domain .GT. 0) THEN

            m_b_copy(:, :) = diis_env%m_b(idomain)%mdata(:, :)

            ! Query the optimal workspace for dsyev
            LWORK = -1
            ALLOCATE (WORK(MAX(1, LWORK)))
            CALL DSYEV('V', 'L', diis_env%buffer_length+1, m_b_copy, &
                       diis_env%buffer_length+1, eigenvalues, WORK, LWORK, INFO)
            LWORK = INT(WORK(1))
            DEALLOCATE (WORK)

            ! Allocate the workspace and solve the eigenproblem
            ALLOCATE (WORK(MAX(1, LWORK)))
            CALL DSYEV('V', 'L', diis_env%buffer_length+1, m_b_copy, &
                       diis_env%buffer_length+1, eigenvalues, WORK, LWORK, INFO)
            IF (INFO .NE. 0) THEN
               CALL cp__b("almo_scf_diis_types.F",406,"DSYEV failed")
            END IF
            DEALLOCATE (WORK)

            ! use the eigensystem to invert (implicitly) B matrix
            ! and compute the extrapolation coefficients
            !! ALLOCATE(tmp1(diis_env%buffer_length+1,1))
            !! ALLOCATE(coeff(diis_env%buffer_length+1,1))
            !! tmp1(:,1)=-1.0_dp*m_b_copy(1,:)/eigenvalues(:)
            !! coeff=MATMUL(m_b_copy,tmp1)
            !! DEALLOCATE(tmp1)
            ALLOCATE (tmp1(diis_env%buffer_length+1))
            ALLOCATE (coeff(diis_env%buffer_length+1))
            tmp1(:) = -1.0_dp*m_b_copy(1, :)/eigenvalues(:)
            coeff(:) = MATMUL(m_b_copy, tmp1)
            DEALLOCATE (tmp1)

            !IF (unit_nr.gt.0) THEN
            !   DO im=1,diis_env%buffer_length+1
            !      WRITE(unit_nr,*) diis_env%m_b(idomain)%mdata(im,:)
            !   ENDDO
            !   WRITE (unit_nr,*) coeff(:,1)
            !ENDIF

            ! extrapolate the variable
            checksum = 0.0_dp
            IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
               CALL cp_dbcsr_set(extr_var, 0.0_dp)
               DO im = 1, diis_env%buffer_length
                  CALL cp_dbcsr_add(extr_var, diis_env%m_var(im), &
                                    1.0_dp, coeff(im+1))
                  checksum = checksum+coeff(im+1)
               ENDDO
            ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
               CALL copy_submatrices(diis_env%d_var(1, idomain), &
                                     d_extr_var(idomain), &
                                     copy_data=.FALSE.)
               CALL set_submatrices(d_extr_var(idomain), 0.0_dp)
               DO im = 1, diis_env%buffer_length
                  CALL add_submatrices(1.0_dp, d_extr_var(idomain), &
                                       coeff(im+1), diis_env%d_var(im, idomain), &
                                       'N')
                  checksum = checksum+coeff(im+1)
               ENDDO
            ENDIF
            !WRITE(*,*) checksum

            DEALLOCATE (coeff)

         ENDIF ! domain is local to this mpi node

      ENDDO ! loop over domains

      DEALLOCATE (eigenvalues)
      DEALLOCATE (m_b_copy)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_diis_extrapolate

! **************************************************************************************************
!> \brief computes elements of b-matrix
!> \param diis_env ...
!> \param A ...
!> \param B ...
!> \param d_A ...
!> \param d_B ...
!> \retval almo_scf_diis_error_overlap ...
!> \par History
!>       2013.02 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   FUNCTION almo_scf_diis_error_overlap(diis_env, A, B, d_A, d_B)

      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: A, B
      TYPE(domain_submatrix_type), INTENT(INOUT), &
         OPTIONAL                                        :: d_A, d_B
      REAL(KIND=dp)                                      :: almo_scf_diis_error_overlap

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_error_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: trace

      CALL timeset(routineN, handle)

      IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
         IF (.NOT. (PRESENT(A) .AND. PRESENT(B))) THEN
            CALL cp__b("almo_scf_diis_types.F",496,"provide DBCSR matrices")
         ENDIF
      ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
         IF (.NOT. (PRESENT(d_A) .AND. PRESENT(d_B))) THEN
            CALL cp__b("almo_scf_diis_types.F",500,"provide domain submatrices")
         ENDIF
      ELSE
         CALL cp__b("almo_scf_diis_types.F",503,"illegal DIIS ENV type")
      ENDIF

      SELECT CASE (diis_env%error_type)
      CASE (diis_error_orthogonal)
         IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
            CALL cp_dbcsr_trace(A, B, trace, 'T', 'N')
         ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
            IF(.NOT.(SIZE(d_A%mdata, 1) .EQ. SIZE(d_B%mdata, 1)))CALL cp__a("almo_scf_diis_types.F",511)
            IF(.NOT.(SIZE(d_A%mdata, 2) .EQ. SIZE(d_B%mdata, 2)))CALL cp__a("almo_scf_diis_types.F",512)
            IF(.NOT.(d_A%domain .EQ. d_B%domain))CALL cp__a("almo_scf_diis_types.F",513)
            IF(.NOT.(d_A%domain .GT. 0))CALL cp__a("almo_scf_diis_types.F",514)
            IF(.NOT.(d_B%domain .GT. 0))CALL cp__a("almo_scf_diis_types.F",515)
            trace = SUM(d_A%mdata(:, :)*d_B%mdata(:, :))
         ENDIF
      CASE DEFAULT
         CALL cp__b("almo_scf_diis_types.F",519,"Vector type is unknown")
      END SELECT

      almo_scf_diis_error_overlap = trace

      CALL timestop(handle)

   END FUNCTION almo_scf_diis_error_overlap

! **************************************************************************************************
!> \brief destroys the diis structure
!> \param diis_env ...
!> \par History
!>       2011.12 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_diis_release(diis_env)
      TYPE(almo_scf_diis_type), INTENT(INOUT)            :: diis_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_diis_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, im

      CALL timeset(routineN, handle)

      ! release matrices
      DO im = 1, diis_env%max_buffer_length
         IF (diis_env%diis_env_type .EQ. diis_env_dbcsr) THEN
            CALL cp_dbcsr_release(diis_env%m_err(im))
            CALL cp_dbcsr_release(diis_env%m_var(im))
         ELSE IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
            CALL release_submatrices(diis_env%d_var(im, :))
            CALL release_submatrices(diis_env%d_err(im, :))
         ENDIF
      ENDDO

      IF (diis_env%diis_env_type .EQ. diis_env_domain) THEN
         CALL release_submatrices(diis_env%m_b(:))
      ENDIF

      IF (ALLOCATED(diis_env%m_b)) DEALLOCATE (diis_env%m_b)
      IF (ALLOCATED(diis_env%m_err)) DEALLOCATE (diis_env%m_err)
      IF (ALLOCATED(diis_env%m_var)) DEALLOCATE (diis_env%m_var)
      IF (ALLOCATED(diis_env%d_err)) DEALLOCATE (diis_env%d_err)
      IF (ALLOCATED(diis_env%d_var)) DEALLOCATE (diis_env%d_var)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_diis_release

END MODULE almo_scf_diis_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Cayley transformation methods
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE ct_methods
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_invert
   USE cp_dbcsr_diag,                   ONLY: cp_dbcsr_syevd
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_filter, cp_dbcsr_finalize, cp_dbcsr_frobenius_norm, &
        cp_dbcsr_function_of_elements, cp_dbcsr_get_diag, cp_dbcsr_get_info, &
        cp_dbcsr_get_stored_coordinates, cp_dbcsr_hadamard_product, cp_dbcsr_init, &
        cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, &
        cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, cp_dbcsr_multiply, &
        cp_dbcsr_nblkcols_total, cp_dbcsr_nblkrows_total, cp_dbcsr_norm, cp_dbcsr_release, &
        cp_dbcsr_reserve_block2d, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_set_diag, cp_dbcsr_trace, &
        cp_dbcsr_transposed, cp_dbcsr_type, cp_dbcsr_work_create, dbcsr_distribution_mp, &
        dbcsr_func_inverse, dbcsr_mp_mynode, dbcsr_norm_maxabsnorm, dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE ct_types,                        ONLY: ct_step_env_type
   USE input_constants,                 ONLY: &
        cg_dai_yuan, cg_fletcher, cg_fletcher_reeves, cg_hager_zhang, cg_hestenes_stiefel, &
        cg_liu_storey, cg_polak_ribiere, cg_zero, tensor_orthogonal, tensor_up_down
   USE iterate_matrix,                  ONLY: matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ct_methods'

   ! Public subroutines
   PUBLIC :: ct_step_execute, analytic_line_search, diagonalize_diagonal_blocks

CONTAINS

! **************************************************************************************************
!> \brief Performs Cayley transformation
!> \param cts_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE ct_step_execute(cts_env)

      TYPE(ct_step_env_type)                             :: cts_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_execute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n, preconditioner_type, unit_nr
      REAL(KIND=dp)                                      :: gap_estimate, safety_margin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: evals
      TYPE(cp_dbcsr_type)                                :: matrix_pp, matrix_pq, matrix_qp, &
                                                            matrix_qp_save, matrix_qq, oo1, &
                                                            oo1_sqrt, oo1_sqrt_inv, t_corr, tmp1, &
                                                            u_pp, u_qq
      TYPE(cp_logger_type), POINTER                      :: logger

!TYPE(cp_dbcsr_type)                :: rst_x1, rst_x2
!REAL(KIND=dp)                      :: ener_tmp
!TYPE(cp_dbcsr_iterator)            :: iter
!INTEGER                            :: iblock_row,iblock_col,&
!                                      iblock_row_size,iblock_col_size
!REAL(KIND=dp), DIMENSION(:,:), POINTER :: data_p

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! check if all input is in place and flags are consistent
      IF (cts_env%update_q .AND. (.NOT. cts_env%update_p)) THEN
         CALL cp__b("ct_methods.F",91,"q-update is possible only with p-update")
      ENDIF

      IF (cts_env%tensor_type .EQ. tensor_up_down) THEN
         CALL cp__b("ct_methods.F",95,"riccati is not implemented for biorthogonal basis")
      ENDIF

      IF (.NOT. ASSOCIATED(cts_env%matrix_ks)) THEN
         CALL cp__b("ct_methods.F",99,"KS matrix is not associated")
      ENDIF

      IF (cts_env%use_virt_orbs .AND. (.NOT. cts_env%use_occ_orbs)) THEN
         CALL cp__b("ct_methods.F",103,"virtual orbs can be used only with occupied orbs")
      ENDIF

      IF (cts_env%use_occ_orbs) THEN
         IF (.NOT. ASSOCIATED(cts_env%matrix_t)) THEN
            CALL cp__b("ct_methods.F",108,"T matrix is not associated")
         ENDIF
         IF (.NOT. ASSOCIATED(cts_env%matrix_qp_template)) THEN
            CALL cp__b("ct_methods.F",111,"QP template is not associated")
         ENDIF
         IF (.NOT. ASSOCIATED(cts_env%matrix_pq_template)) THEN
            CALL cp__b("ct_methods.F",114,"PQ template is not associated")
         ENDIF
      ENDIF

      IF (cts_env%use_virt_orbs) THEN
         IF (.NOT. ASSOCIATED(cts_env%matrix_v)) THEN
            CALL cp__b("ct_methods.F",120,"V matrix is not associated")
         ENDIF
      ELSE
         IF (.NOT. ASSOCIATED(cts_env%matrix_p)) THEN
            CALL cp__b("ct_methods.F",124,"P matrix is not associated")
         ENDIF
      ENDIF

      IF (cts_env%tensor_type .NE. tensor_up_down .AND. &
          cts_env%tensor_type .NE. tensor_orthogonal) THEN
         CALL cp__b("ct_methods.F",130,"illegal tensor flag")
      ENDIF

      ! start real calculations
      IF (cts_env%use_occ_orbs) THEN

         ! create matrices for various ks blocks
         CALL cp_dbcsr_init(matrix_pp)
         CALL cp_dbcsr_create(matrix_pp, &
                              template=cts_env%p_index_up, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_init(matrix_qp)
         CALL cp_dbcsr_create(matrix_qp, &
                              template=cts_env%matrix_qp_template, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_init(matrix_qq)
         CALL cp_dbcsr_create(matrix_qq, &
                              template=cts_env%q_index_up, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_init(matrix_pq)
         CALL cp_dbcsr_create(matrix_pq, &
                              template=cts_env%matrix_pq_template, &
                              matrix_type=dbcsr_type_no_symmetry)

         ! create the residue matrix
         CALL cp_dbcsr_init(cts_env%matrix_res)
         CALL cp_dbcsr_create(cts_env%matrix_res, &
                              template=cts_env%matrix_qp_template)

         CALL assemble_ks_qp_blocks(cts_env%matrix_ks, &
                                    cts_env%matrix_p, &
                                    cts_env%matrix_t, &
                                    cts_env%matrix_v, &
                                    cts_env%q_index_down, &
                                    cts_env%p_index_up, &
                                    cts_env%q_index_up, &
                                    matrix_pp, &
                                    matrix_qq, &
                                    matrix_qp, &
                                    matrix_pq, &
                                    cts_env%tensor_type, &
                                    cts_env%use_virt_orbs, &
                                    cts_env%eps_filter)

         ! create a matrix of single-excitation amplitudes
         CALL cp_dbcsr_init(cts_env%matrix_x)
         CALL cp_dbcsr_create(cts_env%matrix_x, &
                              template=cts_env%matrix_qp_template)
         IF (ASSOCIATED(cts_env%matrix_x_guess)) THEN
            CALL cp_dbcsr_copy(cts_env%matrix_x, &
                               cts_env%matrix_x_guess)
            IF (cts_env%tensor_type .EQ. tensor_orthogonal) THEN
               ! bring x from contravariant-covariant representation
               ! to the orthogonal/cholesky representation
               ! use res as temporary storage
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%q_index_down, &
                                      cts_env%matrix_x, 0.0_dp, cts_env%matrix_res, &
                                      filter_eps=cts_env%eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%matrix_res, &
                                      cts_env%p_index_up, 0.0_dp, &
                                      cts_env%matrix_x, &
                                      filter_eps=cts_env%eps_filter)
            ENDIF
         ELSE
            ! set amplitudes to zero
            CALL cp_dbcsr_set(cts_env%matrix_x, 0.0_dp)
         ENDIF

         !SELECT CASE (cts_env%preconditioner_type)
         !CASE (prec_eigenvector_blocks,prec_eigenvector_full)
         preconditioner_type = 1
         safety_margin = 2.0_dp
         gap_estimate = 0.0001_dp
         SELECT CASE (preconditioner_type)
         CASE (1, 2)
!RZK-warning diagonalization works only with orthogonal tensor!!!
            ! find a better basis by diagonalizing diagonal blocks
            ! first pp
            CALL cp_dbcsr_init(u_pp)
            CALL cp_dbcsr_create(u_pp, template=matrix_pp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            !IF (cts_env%preconditioner_type.eq.prec_eigenvector_full) THEN
            IF (.TRUE.) THEN
               CALL cp_dbcsr_get_info(matrix_pp, nfullrows_total=n)
               ALLOCATE (evals(n))
               CALL cp_dbcsr_syevd(matrix_pp, u_pp, evals, &
                                   cts_env%para_env, cts_env%blacs_env)
               DEALLOCATE (evals)
            ELSE
               CALL diagonalize_diagonal_blocks(matrix_pp, u_pp)
            ENDIF
            ! and now qq
            CALL cp_dbcsr_init(u_qq)
            CALL cp_dbcsr_create(u_qq, template=matrix_qq, &
                                 matrix_type=dbcsr_type_no_symmetry)
            !IF (cts_env%preconditioner_type.eq.prec_eigenvector_full) THEN
            IF (.TRUE.) THEN
               CALL cp_dbcsr_get_info(matrix_qq, nfullrows_total=n)
               ALLOCATE (evals(n))
               CALL cp_dbcsr_syevd(matrix_qq, u_qq, evals, &
                                   cts_env%para_env, cts_env%blacs_env)
               DEALLOCATE (evals)
            ELSE
               CALL diagonalize_diagonal_blocks(matrix_qq, u_qq)
            ENDIF

            ! apply the transformation to all matrices
            CALL matrix_forward_transform(matrix_pp, u_pp, u_pp, &
                                          cts_env%eps_filter)
            CALL matrix_forward_transform(matrix_qq, u_qq, u_qq, &
                                          cts_env%eps_filter)
            CALL matrix_forward_transform(matrix_qp, u_qq, u_pp, &
                                          cts_env%eps_filter)
            CALL matrix_forward_transform(matrix_pq, u_pp, u_qq, &
                                          cts_env%eps_filter)
            CALL matrix_forward_transform(cts_env%matrix_x, u_qq, u_pp, &
                                          cts_env%eps_filter)

            IF (cts_env%max_iter .GE. 0) THEN

               CALL solve_riccati_equation( &
                  pp=matrix_pp, &
                  qq=matrix_qq, &
                  qp=matrix_qp, &
                  pq=matrix_pq, &
                  x=cts_env%matrix_x, &
                  res=cts_env%matrix_res, &
                  neglect_quadratic_term=cts_env%neglect_quadratic_term, &
                  conjugator=cts_env%conjugator, &
                  max_iter=cts_env%max_iter, &
                  eps_convergence=cts_env%eps_convergence, &
                  eps_filter=cts_env%eps_filter, &
                  converged=cts_env%converged)

               IF (cts_env%converged) THEN
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*)
                  !   WRITE(unit_nr,'(T6,A)') &
                  !         "RICCATI equations solved"
                  !   CALL m_flush(unit_nr)
                  !ENDIF
               ELSE
                  CALL cp__b("ct_methods.F",272,"RICCATI: CG algorithm has NOT converged")
               ENDIF

            ENDIF

            IF (cts_env%calculate_energy_corr) THEN

               CALL cp_dbcsr_trace(matrix_qp, cts_env%matrix_x, &
                                   cts_env%energy_correction, "T", "N")

            ENDIF

            CALL cp_dbcsr_release(matrix_pp)
            CALL cp_dbcsr_release(matrix_qp)
            CALL cp_dbcsr_release(matrix_qq)
            CALL cp_dbcsr_release(matrix_pq)

            ! back-transform to the original basis
            CALL matrix_backward_transform(cts_env%matrix_x, u_qq, &
                                           u_pp, cts_env%eps_filter)

            CALL cp_dbcsr_release(u_qq)
            CALL cp_dbcsr_release(u_pp)

            !CASE (prec_cholesky_inverse)
         CASE (3)

! RZK-warning implemented only for orthogonal tensors!!!
! generalization to up_down should be easy
            CALL cp_dbcsr_init(u_pp)
            CALL cp_dbcsr_create(u_pp, template=matrix_pp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_copy(u_pp, matrix_pp)
            CALL cp_dbcsr_scale(u_pp, -1.0_dp)
            CALL cp_dbcsr_add_on_diag(u_pp, &
                                      ABS(safety_margin*gap_estimate))
            CALL cp_dbcsr_cholesky_decompose(u_pp, &
                                             para_env=cts_env%para_env, &
                                             blacs_env=cts_env%blacs_env)
            CALL cp_dbcsr_cholesky_invert(u_pp, &
                                          para_env=cts_env%para_env, &
                                          blacs_env=cts_env%blacs_env, &
                                          upper_to_full=.TRUE.)
            !CALL cp_dbcsr_scale(u_pp,-1.0_dp)

            CALL cp_dbcsr_init(u_qq)
            CALL cp_dbcsr_create(u_qq, template=matrix_qq, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_copy(u_qq, matrix_qq)
            CALL cp_dbcsr_add_on_diag(u_qq, &
                                      ABS(safety_margin*gap_estimate))
            CALL cp_dbcsr_cholesky_decompose(u_qq, &
                                             para_env=cts_env%para_env, &
                                             blacs_env=cts_env%blacs_env)
            CALL cp_dbcsr_cholesky_invert(u_qq, &
                                          para_env=cts_env%para_env, &
                                          blacs_env=cts_env%blacs_env, &
                                          upper_to_full=.TRUE.)

            ! transform all riccati matrices (left-right preconditioner)
            CALL cp_dbcsr_init(tmp1)
            CALL cp_dbcsr_create(tmp1, template=matrix_qq, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, u_qq, &
                                   matrix_qq, 0.0_dp, tmp1, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_copy(matrix_qq, tmp1)
            CALL cp_dbcsr_release(tmp1)

            CALL cp_dbcsr_init(tmp1)
            CALL cp_dbcsr_create(tmp1, template=matrix_pp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_pp, &
                                   u_pp, 0.0_dp, tmp1, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_copy(matrix_pp, tmp1)
            CALL cp_dbcsr_release(tmp1)

            CALL cp_dbcsr_init(matrix_qp_save)
            CALL cp_dbcsr_create(matrix_qp_save, template=matrix_qp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_copy(matrix_qp_save, matrix_qp)

            CALL cp_dbcsr_init(tmp1)
            CALL cp_dbcsr_create(tmp1, template=matrix_qp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_qp, &
                                   u_pp, 0.0_dp, tmp1, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, u_qq, tmp1, &
                                   0.0_dp, matrix_qp, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_release(tmp1)
!CALL cp_dbcsr_print(matrix_qq)
!CALL cp_dbcsr_print(matrix_qp)
!CALL cp_dbcsr_print(matrix_pp)

            IF (cts_env%max_iter .GE. 0) THEN

               CALL solve_riccati_equation( &
                  pp=matrix_pp, &
                  qq=matrix_qq, &
                  qp=matrix_qp, &
                  pq=matrix_pq, &
                  oo=u_pp, &
                  vv=u_qq, &
                  x=cts_env%matrix_x, &
                  res=cts_env%matrix_res, &
                  neglect_quadratic_term=cts_env%neglect_quadratic_term, &
                  conjugator=cts_env%conjugator, &
                  max_iter=cts_env%max_iter, &
                  eps_convergence=cts_env%eps_convergence, &
                  eps_filter=cts_env%eps_filter, &
                  converged=cts_env%converged)

               IF (cts_env%converged) THEN
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*)
                  !   WRITE(unit_nr,'(T6,A)') &
                  !         "RICCATI equations solved"
                  !   CALL m_flush(unit_nr)
                  !ENDIF
               ELSE
                  CALL cp__b("ct_methods.F",395,"RICCATI: CG algorithm has NOT converged")
               ENDIF

            ENDIF

            IF (cts_env%calculate_energy_corr) THEN

               CALL cp_dbcsr_trace(matrix_qp_save, cts_env%matrix_x, &
                                   cts_env%energy_correction, "T", "N")

            ENDIF
            CALL cp_dbcsr_release(matrix_qp_save)

            CALL cp_dbcsr_release(matrix_pp)
            CALL cp_dbcsr_release(matrix_qp)
            CALL cp_dbcsr_release(matrix_qq)
            CALL cp_dbcsr_release(matrix_pq)

            CALL cp_dbcsr_release(u_qq)
            CALL cp_dbcsr_release(u_pp)

         CASE DEFAULT
            CALL cp__b("ct_methods.F",417,"illegal preconditioner type")
         END SELECT ! preconditioner type

         IF (cts_env%update_p) THEN

            IF (cts_env%tensor_type .EQ. tensor_up_down) THEN
               CALL cp__b("ct_methods.F",423,"orbital update is NYI for this tensor type")
            ENDIF

            ! transform occupied orbitals
            ! in a way that preserves the overlap metric
            CALL cp_dbcsr_init(oo1)
            CALL cp_dbcsr_create(oo1, &
                                 template=cts_env%p_index_up, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_init(oo1_sqrt_inv)
            CALL cp_dbcsr_create(oo1_sqrt_inv, &
                                 template=oo1)
            CALL cp_dbcsr_init(oo1_sqrt)
            CALL cp_dbcsr_create(oo1_sqrt, &
                                 template=oo1)

            ! Compute (1+tr(X).X)^(-1/2)_up_down
            CALL cp_dbcsr_multiply("T", "N", 1.0_dp, cts_env%matrix_x, &
                                   cts_env%matrix_x, 0.0_dp, oo1, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_add_on_diag(oo1, 1.0_dp)
            CALL matrix_sqrt_Newton_Schulz(oo1_sqrt, &
                                           oo1_sqrt_inv, &
                                           oo1, &
                                           !if cholesky is used then sqrt
                                           !guess cannot be provided
                                           !matrix_sqrt_inv_guess=cts_env%p_index_up,&
                                           !matrix_sqrt_guess=cts_env%p_index_down,&
                                           threshold=cts_env%eps_filter, &
                                           order=cts_env%order_lanczos, &
                                           eps_lanczos=cts_env%eps_lancsoz, &
                                           max_iter_lanczos=cts_env%max_iter_lanczos)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%p_index_up, &
                                   oo1_sqrt_inv, 0.0_dp, oo1, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, oo1, &
                                   cts_env%p_index_down, 0.0_dp, oo1_sqrt, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_release(oo1)
            CALL cp_dbcsr_release(oo1_sqrt_inv)

            ! bring x to contravariant-covariant representation now
            CALL cp_dbcsr_init(matrix_qp)
            CALL cp_dbcsr_create(matrix_qp, &
                                 template=cts_env%matrix_qp_template, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%q_index_up, &
                                   cts_env%matrix_x, 0.0_dp, matrix_qp, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_qp, &
                                   cts_env%p_index_down, 0.0_dp, &
                                   cts_env%matrix_x, &
                                   filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_release(matrix_qp)

            ! update T=T+X or T=T+V.X (whichever is appropriate)
            CALL cp_dbcsr_init(t_corr)
            CALL cp_dbcsr_create(t_corr, template=cts_env%matrix_t)
            IF (cts_env%use_virt_orbs) THEN
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%matrix_v, &
                                      cts_env%matrix_x, 0.0_dp, t_corr, &
                                      filter_eps=cts_env%eps_filter)
               CALL cp_dbcsr_add(cts_env%matrix_t, t_corr, &
                                 1.0_dp, 1.0_dp)
            ELSE
               CALL cp_dbcsr_add(cts_env%matrix_t, cts_env%matrix_x, &
                                 1.0_dp, 1.0_dp)
            ENDIF
            ! adjust T so the metric is preserved: T=(T+X).(1+tr(X).X)^(-1/2)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%matrix_t, oo1_sqrt, &
                                   0.0_dp, t_corr, filter_eps=cts_env%eps_filter)
            CALL cp_dbcsr_copy(cts_env%matrix_t, t_corr)

            CALL cp_dbcsr_release(t_corr)
            CALL cp_dbcsr_release(oo1_sqrt)

         ELSE ! do not update p

            IF (cts_env%tensor_type .EQ. tensor_orthogonal) THEN
               ! bring x to contravariant-covariant representation
               CALL cp_dbcsr_init(matrix_qp)
               CALL cp_dbcsr_create(matrix_qp, &
                                    template=cts_env%matrix_qp_template, &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, cts_env%q_index_up, &
                                      cts_env%matrix_x, 0.0_dp, matrix_qp, &
                                      filter_eps=cts_env%eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_qp, &
                                      cts_env%p_index_down, 0.0_dp, &
                                      cts_env%matrix_x, &
                                      filter_eps=cts_env%eps_filter)
               CALL cp_dbcsr_release(matrix_qp)
            ENDIF

         ENDIF

      ELSE
         CALL cp__b("ct_methods.F",520,"illegal occ option")
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ct_step_execute

! **************************************************************************************************
!> \brief computes oo, ov, vo, and vv blocks of the ks matrix
!> \param ks ...
!> \param p ...
!> \param t ...
!> \param v ...
!> \param q_index_down ...
!> \param p_index_up ...
!> \param q_index_up ...
!> \param pp ...
!> \param qq ...
!> \param qp ...
!> \param pq ...
!> \param tensor_type ...
!> \param use_virt_orbs ...
!> \param eps_filter ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE assemble_ks_qp_blocks(ks, p, t, v, q_index_down, &
                                    p_index_up, q_index_up, pp, qq, qp, pq, tensor_type, use_virt_orbs, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: ks, p, t, v, q_index_down, p_index_up, &
                                                            q_index_up
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: pp, qq, qp, pq
      INTEGER, INTENT(IN)                                :: tensor_type
      LOGICAL, INTENT(IN)                                :: use_virt_orbs
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'assemble_ks_qp_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: library_fixed
      TYPE(cp_dbcsr_type)                                :: kst, ksv, no, on, oo, q_index_up_nosym, &
                                                            sp, spf, t_or, v_or

      CALL timeset(routineN, handle)

      IF (use_virt_orbs) THEN

         ! orthogonalize the orbitals
         CALL cp_dbcsr_init(t_or)
         CALL cp_dbcsr_create(t_or, template=t)
         CALL cp_dbcsr_init(v_or)
         CALL cp_dbcsr_create(v_or, template=v)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, t, p_index_up, &
                                0.0_dp, t_or, filter_eps=eps_filter)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, v, q_index_up, &
                                0.0_dp, v_or, filter_eps=eps_filter)

         ! KS.T
         CALL cp_dbcsr_init(kst)
         CALL cp_dbcsr_create(kst, template=t)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ks, t_or, &
                                0.0_dp, kst, filter_eps=eps_filter)
         ! pp=tr(T)*KS.T
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, t_or, kst, &
                                0.0_dp, pp, filter_eps=eps_filter)
         ! qp=tr(V)*KS.T
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, v_or, kst, &
                                0.0_dp, qp, filter_eps=eps_filter)
         CALL cp_dbcsr_release(kst)

         ! KS.V
         CALL cp_dbcsr_init(ksv)
         CALL cp_dbcsr_create(ksv, template=v)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ks, v_or, &
                                0.0_dp, ksv, filter_eps=eps_filter)
         ! tr(T)*KS.V
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, t_or, ksv, &
                                0.0_dp, pq, filter_eps=eps_filter)
         ! tr(V)*KS.V
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, v_or, ksv, &
                                0.0_dp, qq, filter_eps=eps_filter)
         CALL cp_dbcsr_release(ksv)

         CALL cp_dbcsr_release(t_or)
         CALL cp_dbcsr_release(v_or)

      ELSE ! no virtuals, use projected AOs

! THIS PROCEDURE HAS NOT BEEN UPDATED FOR CHOLESKY p/q_index_up/down
         CALL cp_dbcsr_init(sp)
         CALL cp_dbcsr_create(sp, template=q_index_down, &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_init(spf)
         CALL cp_dbcsr_create(spf, template=q_index_down, &
                              matrix_type=dbcsr_type_no_symmetry)

         ! qp=KS*T
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, ks, t, 0.0_dp, qp, &
                                filter_eps=eps_filter)
         ! pp=tr(T)*KS.T
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, t, qp, 0.0_dp, pp, &
                                filter_eps=eps_filter)
         ! sp=-S_*P
         CALL cp_dbcsr_multiply("N", "N", -1.0_dp, q_index_down, p, 0.0_dp, sp, &
                                filter_eps=eps_filter)

         ! sp=1/S^-S_.P
         SELECT CASE (tensor_type)
         CASE (tensor_up_down)
            CALL cp_dbcsr_add_on_diag(sp, 1.0_dp)
         CASE (tensor_orthogonal)
            CALL cp_dbcsr_init(q_index_up_nosym)
            CALL cp_dbcsr_create(q_index_up_nosym, template=q_index_up, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_desymmetrize(q_index_up, q_index_up_nosym)
            CALL cp_dbcsr_add(sp, q_index_up_nosym, 1.0_dp, 1.0_dp)
            CALL cp_dbcsr_release(q_index_up_nosym)
         END SELECT

         ! spf=(1/S^-S_.P)*KS
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, sp, ks, 0.0_dp, spf, &
                                filter_eps=eps_filter)

         ! qp=spf*T
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, spf, t, 0.0_dp, qp, &
                                filter_eps=eps_filter)

         SELECT CASE (tensor_type)
         CASE (tensor_up_down)
            ! pq=tr(qp)
            CALL cp_dbcsr_transposed(pq, qp, transpose_distribution=.FALSE.)
         CASE (tensor_orthogonal)
            ! pq=sig^.tr(qp)
            CALL cp_dbcsr_multiply("N", "T", 1.0_dp, p_index_up, qp, 0.0_dp, pq, &
                                   filter_eps=eps_filter)
            library_fixed = .FALSE.
            IF (library_fixed) THEN
               CALL cp_dbcsr_transposed(qp, pq, transpose_distribution=.FALSE.)
            ELSE
               CALL cp_dbcsr_init(no)
               CALL cp_dbcsr_create(no, template=qp, &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qp, p_index_up, 0.0_dp, no, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_copy(qp, no)
               CALL cp_dbcsr_release(no)
            ENDIF
         END SELECT

         ! qq=spf*tr(sp)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, spf, sp, 0.0_dp, qq, &
                                filter_eps=eps_filter)

         SELECT CASE (tensor_type)
         CASE (tensor_up_down)

            CALL cp_dbcsr_init(oo)
            CALL cp_dbcsr_create(oo, template=pp, &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_init(no)
            CALL cp_dbcsr_create(no, template=qp, &
                                 matrix_type=dbcsr_type_no_symmetry)

            ! first index up
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, q_index_up, qq, 0.0_dp, spf, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_copy(qq, spf)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, q_index_up, qp, 0.0_dp, no, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_copy(qp, no)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, p_index_up, pp, 0.0_dp, oo, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_copy(pp, oo)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, p_index_up, pq, 0.0_dp, on, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_copy(pq, on)

            CALL cp_dbcsr_release(no)
            CALL cp_dbcsr_release(oo)

         CASE (tensor_orthogonal)

            CALL cp_dbcsr_init(oo)
            CALL cp_dbcsr_create(oo, template=pp, &
                                 matrix_type=dbcsr_type_no_symmetry)

            ! both indeces up in the pp block
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, p_index_up, pp, 0.0_dp, oo, &
                                   filter_eps=eps_filter)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, oo, p_index_up, 0.0_dp, pp, &
                                   filter_eps=eps_filter)

            CALL cp_dbcsr_release(oo)

         END SELECT

         CALL cp_dbcsr_release(sp)
         CALL cp_dbcsr_release(spf)

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE assemble_ks_qp_blocks

! **************************************************************************************************
!> \brief Solves the generalized Riccati or Sylvester eqation
!>        using the preconditioned conjugate gradient algorithm
!>          qp + qq.x.oo - vv.x.pp - vv.x.pq.x.oo = 0 [oo and vv are optional]
!>          qp + qq.x - x.pp - x.pq.x = 0
!> \param pp ...
!> \param qq ...
!> \param qp ...
!> \param pq ...
!> \param oo ...
!> \param vv ...
!> \param x ...
!> \param res ...
!> \param neglect_quadratic_term ...
!> \param conjugator ...
!> \param max_iter ...
!> \param eps_convergence ...
!> \param eps_filter ...
!> \param converged ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!>       2011.11 generalized [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   RECURSIVE SUBROUTINE solve_riccati_equation(pp, qq, qp, pq, oo, vv, x, res, &
                                               neglect_quadratic_term, &
                                               conjugator, max_iter, eps_convergence, eps_filter, &
                                               converged)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: pp, qq
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: qp
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: pq
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: oo, vv
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: x
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: res
      LOGICAL, INTENT(IN)                                :: neglect_quadratic_term
      INTEGER, INTENT(IN)                                :: conjugator, max_iter
      REAL(KIND=dp), INTENT(IN)                          :: eps_convergence, eps_filter
      LOGICAL, INTENT(OUT)                               :: converged

      CHARACTER(len=*), PARAMETER :: routineN = 'solve_riccati_equation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, istep, iteration, nsteps, &
                                                            unit_nr, update_prec_freq
      LOGICAL                                            :: prepare_to_exit, present_oo, present_vv, &
                                                            quadratic_term, restart_conjugator
      REAL(KIND=dp)                                      :: best_norm, best_step_size, beta, c0, c1, &
                                                            c2, c3, denom, kappa, numer, &
                                                            obj_function, t1, t2, tau
      REAL(KIND=dp), DIMENSION(3)                        :: step_size
      TYPE(cp_dbcsr_type)                                :: aux1, aux2, grad, m, n, oo1, oo2, prec, &
                                                            res_trial, step, step_oo, vv_step
      TYPE(cp_logger_type), POINTER                      :: logger

!TYPE(cp_dbcsr_type)                      :: qqqq, pppp, zero_pq, zero_qp

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      t1 = m_walltime()

!IF (level.gt.5) THEN
!  CPErrorMessage(cp_failure_level,routineP,"recursion level is too high")
!  CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
!ENDIF
!IF (unit_nr>0) THEN
!   WRITE(unit_nr,*) &
!      "========== LEVEL ",level,"=========="
!ENDIF
!CALL cp_dbcsr_print(qq)
!CALL cp_dbcsr_print(pp)
!CALL cp_dbcsr_print(qp)
!!CALL cp_dbcsr_print(pq)
!IF (unit_nr>0) THEN
!   WRITE(unit_nr,*) &
!      "====== END LEVEL ",level,"=========="
!ENDIF

      quadratic_term = .NOT. neglect_quadratic_term
      present_oo = PRESENT(oo)
      present_vv = PRESENT(vv)

      ! create aux1 matrix and init
      CALL cp_dbcsr_init(aux1)
      CALL cp_dbcsr_create(aux1, template=pp)
      CALL cp_dbcsr_copy(aux1, pp)
      CALL cp_dbcsr_scale(aux1, -1.0_dp)

      ! create aux2 matrix and init
      CALL cp_dbcsr_init(aux2)
      CALL cp_dbcsr_create(aux2, template=qq)
      CALL cp_dbcsr_copy(aux2, qq)

      ! create the gradient matrix and init
      CALL cp_dbcsr_init(grad)
      CALL cp_dbcsr_create(grad, template=x)
      CALL cp_dbcsr_set(grad, 0.0_dp)

      ! create a preconditioner
      ! RZK-warning how to apply it to up_down tensor?
      CALL cp_dbcsr_init(prec)
      CALL cp_dbcsr_create(prec, template=x)
      !CALL create_preconditioner(prec,aux1,aux2,qp,res,tensor_type,eps_filter)
      !CALL cp_dbcsr_set(prec,1.0_dp)

      ! create the step matrix and init
      CALL cp_dbcsr_init(step)
      CALL cp_dbcsr_create(step, template=x)
      !CALL cp_dbcsr_hadamard_product(prec,grad,step)
      !CALL cp_dbcsr_scale(step,-1.0_dp)

      CALL cp_dbcsr_init(n)
      CALL cp_dbcsr_create(n, template=x)
      CALL cp_dbcsr_init(m)
      CALL cp_dbcsr_create(m, template=x)
      CALL cp_dbcsr_init(oo1)
      CALL cp_dbcsr_create(oo1, template=pp)
      CALL cp_dbcsr_init(oo2)
      CALL cp_dbcsr_create(oo2, template=pp)
      CALL cp_dbcsr_init(res_trial)
      CALL cp_dbcsr_create(res_trial, template=res)
      CALL cp_dbcsr_init(vv_step)
      CALL cp_dbcsr_create(vv_step, template=res)
      CALL cp_dbcsr_init(step_oo)
      CALL cp_dbcsr_create(step_oo, template=res)

      ! start conjugate gradient iterations
      iteration = 0
      converged = .FALSE.
      prepare_to_exit = .FALSE.
      beta = 0.0_dp
      best_step_size = 0.0_dp
      best_norm = 1.0E+100_dp
      !ecorr=0.0_dp
      !change_ecorr=0.0_dp
      restart_conjugator = .FALSE.
      update_prec_freq = 20
      DO

         ! (re)-compute the residuals
         IF (iteration .EQ. 0) THEN
            CALL cp_dbcsr_copy(res, qp)
            IF (present_oo) THEN
               CALL cp_dbcsr_multiply("N", "N", +1.0_dp, qq, x, 0.0_dp, res_trial, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("N", "N", +1.0_dp, res_trial, oo, 1.0_dp, res, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", +1.0_dp, qq, x, 1.0_dp, res, &
                                      filter_eps=eps_filter)
            ENDIF
            IF (present_vv) THEN
               CALL cp_dbcsr_multiply("N", "N", -1.0_dp, x, pp, 0.0_dp, res_trial, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("N", "N", +1.0_dp, vv, res_trial, 1.0_dp, res, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", -1.0_dp, x, pp, 1.0_dp, res, &
                                      filter_eps=eps_filter)
            ENDIF
            IF (quadratic_term) THEN
               IF (present_oo) THEN
                  CALL cp_dbcsr_multiply("N", "N", +1.0_dp, pq, x, 0.0_dp, oo1, &
                                         filter_eps=eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", +1.0_dp, oo1, oo, 0.0_dp, oo2, &
                                         filter_eps=eps_filter)
               ELSE
                  CALL cp_dbcsr_multiply("N", "N", +1.0_dp, pq, x, 0.0_dp, oo2, &
                                         filter_eps=eps_filter)
               ENDIF
               IF (present_vv) THEN
                  CALL cp_dbcsr_multiply("N", "N", -1.0_dp, x, oo2, 0.0_dp, res_trial, &
                                         filter_eps=eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", +1.0_dp, vv, res_trial, 1.0_dp, res, &
                                         filter_eps=eps_filter)
               ELSE
                  CALL cp_dbcsr_multiply("N", "N", -1.0_dp, x, oo2, 1.0_dp, res, &
                                         filter_eps=eps_filter)
               ENDIF
            ENDIF
            CALL cp_dbcsr_norm(res, dbcsr_norm_maxabsnorm, norm_scalar=best_norm)
         ELSE
            CALL cp_dbcsr_add(res, m, 1.0_dp, best_step_size)
            CALL cp_dbcsr_add(res, n, 1.0_dp, -best_step_size*best_step_size)
            CALL cp_dbcsr_filter(res, eps_filter)
         ENDIF

         ! check convergence and other exit criteria
         converged = (best_norm .LT. eps_convergence)
         IF (converged .OR. (iteration .GE. max_iter)) THEN
            prepare_to_exit = .TRUE.
         ENDIF

         IF (.NOT. prepare_to_exit) THEN

            ! update aux1=-pp-pq.x.oo and aux2=qq-vv.x.pq
            IF (quadratic_term) THEN
               IF (iteration .EQ. 0) THEN
                  IF (present_oo) THEN
                     CALL cp_dbcsr_multiply("N", "N", -1.0_dp, pq, x, 0.0_dp, oo1, &
                                            filter_eps=eps_filter)
                     CALL cp_dbcsr_multiply("N", "N", +1.0_dp, oo1, oo, 1.0_dp, aux1, &
                                            filter_eps=eps_filter)
                  ELSE
                     CALL cp_dbcsr_multiply("N", "N", -1.0_dp, pq, x, 1.0_dp, aux1, &
                                            filter_eps=eps_filter)
                  ENDIF
                  IF (present_vv) THEN
                     CALL cp_dbcsr_multiply("N", "N", -1.0_dp, vv, x, 0.0_dp, res_trial, &
                                            filter_eps=eps_filter)
                     CALL cp_dbcsr_multiply("N", "N", +1.0_dp, res_trial, pq, 1.0_dp, aux2, &
                                            filter_eps=eps_filter)
                  ELSE
                     CALL cp_dbcsr_multiply("N", "N", -1.0_dp, x, pq, 1.0_dp, aux2, &
                                            filter_eps=eps_filter)
                  ENDIF
               ELSE
                  IF (present_oo) THEN
                     CALL cp_dbcsr_multiply("N", "N", -best_step_size, pq, step_oo, 1.0_dp, aux1, &
                                            filter_eps=eps_filter)
                  ELSE
                     CALL cp_dbcsr_multiply("N", "N", -best_step_size, pq, step, 1.0_dp, aux1, &
                                            filter_eps=eps_filter)
                  ENDIF
                  IF (present_vv) THEN
                     CALL cp_dbcsr_multiply("N", "N", -best_step_size, vv_step, pq, 1.0_dp, aux2, &
                                            filter_eps=eps_filter)
                  ELSE
                     CALL cp_dbcsr_multiply("N", "N", -best_step_size, step, pq, 1.0_dp, aux2, &
                                            filter_eps=eps_filter)
                  ENDIF
               ENDIF
            ENDIF

            ! recompute the gradient, do not update it yet
            ! use m matrix as a temporary storage
            ! grad=t(vv).res.t(aux1)+t(aux2).res.t(oo)
            IF (present_vv) THEN
               CALL cp_dbcsr_multiply("N", "T", 1.0_dp, res, aux1, 0.0_dp, res_trial, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("T", "N", 1.0_dp, vv, res_trial, 0.0_dp, m, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "T", 1.0_dp, res, aux1, 0.0_dp, m, &
                                      filter_eps=eps_filter)
            ENDIF
            IF (present_oo) THEN
               CALL cp_dbcsr_multiply("T", "N", 1.0_dp, aux1, res, 0.0_dp, res_trial, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("N", "T", 1.0_dp, res_trial, oo, 1.0_dp, m, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("T", "N", 1.0_dp, aux2, res, 1.0_dp, m, &
                                      filter_eps=eps_filter)
            ENDIF

            ! compute preconditioner
            !IF (iteration.eq.0.OR.(mod(iteration,update_prec_freq).eq.0)) THEN
            IF (iteration .EQ. 0) THEN
               CALL create_preconditioner(prec, aux1, aux2, eps_filter)
               !restart_conjugator=.TRUE.
!CALL cp_dbcsr_set(prec,1.0_dp)
!CALL cp_dbcsr_print(prec)
            ENDIF

            ! compute the conjugation coefficient - beta
            IF ((iteration .EQ. 0) .OR. restart_conjugator) THEN
               beta = 0.0_dp
            ELSE
               restart_conjugator = .FALSE.
               SELECT CASE (conjugator)
               CASE (cg_hestenes_stiefel)
                  CALL cp_dbcsr_add(grad, m, -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(n, m, numer, "T", "N")
                  CALL cp_dbcsr_trace(grad, step, denom, "T", "N")
                  beta = numer/denom
               CASE (cg_fletcher_reeves)
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(grad, n, denom, "T", "N")
                  CALL cp_dbcsr_hadamard_product(prec, m, n)
                  CALL cp_dbcsr_trace(m, n, numer, "T", "N")
                  beta = numer/denom
               CASE (cg_polak_ribiere)
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(grad, n, denom, "T", "N")
                  CALL cp_dbcsr_add(grad, m, -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(n, m, numer, "T", "N")
                  beta = numer/denom
               CASE (cg_fletcher)
                  CALL cp_dbcsr_hadamard_product(prec, m, n)
                  CALL cp_dbcsr_trace(m, n, numer, "T", "N")
                  CALL cp_dbcsr_trace(grad, step, denom, "T", "N")
                  beta = -1.0_dp*numer/denom
               CASE (cg_liu_storey)
                  CALL cp_dbcsr_trace(grad, step, denom, "T", "N")
                  CALL cp_dbcsr_add(grad, m, -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(n, m, numer, "T", "N")
                  beta = -1.0_dp*numer/denom
               CASE (cg_dai_yuan)
                  CALL cp_dbcsr_hadamard_product(prec, m, n)
                  CALL cp_dbcsr_trace(m, n, numer, "T", "N")
                  CALL cp_dbcsr_add(grad, m, -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_trace(grad, step, denom, "T", "N")
                  beta = numer/denom
               CASE (cg_hager_zhang)
                  CALL cp_dbcsr_add(grad, m, -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_trace(grad, step, denom, "T", "N")
                  CALL cp_dbcsr_hadamard_product(prec, grad, n)
                  CALL cp_dbcsr_trace(n, grad, numer, "T", "N")
                  kappa = 2.0_dp*numer/denom
                  CALL cp_dbcsr_trace(n, m, numer, "T", "N")
                  tau = numer/denom
                  CALL cp_dbcsr_trace(step, m, numer, "T", "N")
                  beta = tau-kappa*numer/denom
               CASE (cg_zero)
                  beta = 0.0_dp
               CASE DEFAULT
                  CALL cp__b("ct_methods.F",1054,"illegal conjugator")
               END SELECT
            ENDIF ! iteration.eq.0

            ! move the current gradient to its storage
            CALL cp_dbcsr_copy(grad, m)

            ! precondition new gradient (use m as tmp storage)
            CALL cp_dbcsr_hadamard_product(prec, grad, m)
            CALL cp_dbcsr_filter(m, eps_filter)

            ! recompute the step direction
            CALL cp_dbcsr_add(step, m, beta, -1.0_dp)
            CALL cp_dbcsr_filter(step, eps_filter)

!! ALTERNATIVE METHOD TO OBTAIN THE STEP FROM THE GRADIENT
!CALL cp_dbcsr_init(qqqq)
!CALL cp_dbcsr_create(qqqq,template=qq)
!CALL cp_dbcsr_init(pppp)
!CALL cp_dbcsr_create(pppp,template=pp)
!CALL cp_dbcsr_init(zero_pq)
!CALL cp_dbcsr_create(zero_pq,template=pq)
!CALL cp_dbcsr_init(zero_qp)
!CALL cp_dbcsr_create(zero_qp,template=qp)
!CALL cp_dbcsr_multiply("T","N",1.0_dp,aux2,aux2,0.0_dp,qqqq,&
!        filter_eps=eps_filter)
!CALL cp_dbcsr_multiply("N","T",-1.0_dp,aux1,aux1,0.0_dp,pppp,&
!        filter_eps=eps_filter)
!CALL cp_dbcsr_set(zero_qp,0.0_dp)
!CALL cp_dbcsr_set(zero_pq,0.0_dp)
!CALL solve_riccati_equation(pppp,qqqq,grad,zero_pq,zero_qp,zero_qp,&
!               .TRUE.,tensor_type,&
!               conjugator,max_iter,eps_convergence,eps_filter,&
!               converged,level+1)
!CALL cp_dbcsr_release(qqqq)
!CALL cp_dbcsr_release(pppp)
!CALL cp_dbcsr_release(zero_qp)
!CALL cp_dbcsr_release(zero_pq)

            ! calculate the optimal step size
            ! m=step.aux1+aux2.step
            IF (present_vv) THEN
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vv, step, 0.0_dp, vv_step, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vv_step, aux1, 0.0_dp, m, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, step, aux1, 0.0_dp, m, &
                                      filter_eps=eps_filter)
            ENDIF
            IF (present_oo) THEN
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, step, oo, 0.0_dp, step_oo, &
                                      filter_eps=eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, aux2, step_oo, 1.0_dp, m, &
                                      filter_eps=eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, aux2, step, 1.0_dp, m, &
                                      filter_eps=eps_filter)
            ENDIF

            IF (quadratic_term) THEN
               ! n=step.pq.step
               IF (present_oo) THEN
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pq, step, 0.0_dp, oo1, &
                                         filter_eps=eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, oo1, oo, 0.0_dp, oo2, &
                                         filter_eps=eps_filter)
               ELSE
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, pq, step, 0.0_dp, oo2, &
                                         filter_eps=eps_filter)
               ENDIF
               IF (present_vv) THEN
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, step, oo2, 0.0_dp, res_trial, &
                                         filter_eps=eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vv, res_trial, 0.0_dp, n, &
                                         filter_eps=eps_filter)
               ELSE
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, step, oo2, 0.0_dp, n, &
                                         filter_eps=eps_filter)
               ENDIF

            ELSE
               CALL cp_dbcsr_set(n, 0.0_dp)
            ENDIF

            ! calculate coefficients of the cubic eq for alpha - step size
            c0 = 2.0_dp*(cp_dbcsr_frobenius_norm(n))**2

            CALL cp_dbcsr_trace(m, n, c1, "T", "N")
            c1 = -3.0_dp*c1

            CALL cp_dbcsr_trace(res, n, c2, "T", "N")
            c2 = -2.0_dp*c2+(cp_dbcsr_frobenius_norm(m))**2

            CALL cp_dbcsr_trace(res, m, c3, "T", "N")

            ! find step size
            CALL analytic_line_search(c0, c1, c2, c3, step_size, nsteps)

            IF (nsteps .EQ. 0) THEN
               CALL cp__b("ct_methods.F",1154,"no step sizes!")
            ENDIF
            ! if we have several possible step sizes
            ! choose one with the lowest objective function
            best_norm = 1.0E+100_dp
            best_step_size = 0.0_dp
            DO istep = 1, nsteps
               ! recompute the residues
               CALL cp_dbcsr_copy(res_trial, res)
               CALL cp_dbcsr_add(res_trial, m, 1.0_dp, step_size(istep))
               CALL cp_dbcsr_add(res_trial, n, 1.0_dp, -step_size(istep)*step_size(istep))
               CALL cp_dbcsr_filter(res_trial, eps_filter)
               ! RZK-warning objective function might be different in the case of
               ! tensor_up_down
               !obj_function=0.5_dp*(cp_dbcsr_frobenius_norm(res_trial))**2
               CALL cp_dbcsr_norm(res_trial, dbcsr_norm_maxabsnorm, norm_scalar=obj_function)
               IF (obj_function .LT. best_norm) THEN
                  best_norm = obj_function
                  best_step_size = step_size(istep)
               ENDIF
            ENDDO

         ENDIF

         ! update X along the line
         CALL cp_dbcsr_add(x, step, 1.0_dp, best_step_size)
         CALL cp_dbcsr_filter(x, eps_filter)

         ! evaluate current energy correction
         !change_ecorr=ecorr
         !CALL cp_dbcsr_trace(qp,x,ecorr,"T","N")
         !change_ecorr=ecorr-change_ecorr

         ! check convergence and other exit criteria
         converged = (best_norm .LT. eps_convergence)
         IF (converged .OR. (iteration .GE. max_iter)) THEN
            prepare_to_exit = .TRUE.
         ENDIF

         t2 = m_walltime()

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T6,A,1X,I4,1X,E12.3,F8.3)') &
               "RICCATI iter ", iteration, best_norm, t2-t1
            !WRITE(unit_nr,'(T6,A,1X,I4,1X,F15.9,F15.9,E12.3,F8.3)') &
            !   "RICCATI iter ",iteration,ecorr,change_ecorr,best_norm,t2-t1
         ENDIF

         t1 = m_walltime()

         iteration = iteration+1

         IF (prepare_to_exit) EXIT

      ENDDO

      CALL cp_dbcsr_release(aux1)
      CALL cp_dbcsr_release(aux2)
      CALL cp_dbcsr_release(grad)
      CALL cp_dbcsr_release(step)
      CALL cp_dbcsr_release(n)
      CALL cp_dbcsr_release(m)
      CALL cp_dbcsr_release(oo1)
      CALL cp_dbcsr_release(oo2)
      CALL cp_dbcsr_release(res_trial)
      CALL cp_dbcsr_release(vv_step)
      CALL cp_dbcsr_release(step_oo)

      CALL timestop(handle)

   END SUBROUTINE solve_riccati_equation

! **************************************************************************************************
!> \brief Computes a preconditioner from diagonal elements of ~f_oo, ~f_vv
!>        The preconditioner is approximately equal to
!>        prec_ai ~ (e_a - e_i)^(-2)
!>        However, the real expression is more complex
!> \param prec ...
!> \param pp ...
!> \param qq ...
!> \param eps_filter ...
!> \par History
!>       2011.07 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE create_preconditioner(prec, pp, qq, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: prec
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: pp, qq
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'create_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, handle, hold, iblock_col, &
                                                            iblock_row, mynode, nblkcols_tot, &
                                                            nblkrows_tot, p_nrows, q_nrows, row
      LOGICAL                                            :: tr
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: p_diagonal, q_diagonal
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_new_block
      TYPE(cp_dbcsr_type)                                :: pp_diag, qq_diag, t1, t2, tmp

!LOGICAL, INTENT(IN)                      :: use_virt_orbs

      CALL timeset(routineN, handle)

!    ! copy diagonal elements
!    CALL cp_dbcsr_get_info(pp,nfullrows_total=nrows)
!    CALL cp_dbcsr_init(pp_diag)
!    CALL cp_dbcsr_create(pp_diag,template=pp)
!    ALLOCATE(diagonal(nrows))
!    CALL cp_dbcsr_get_diag(pp,diagonal)
!    CALL cp_dbcsr_add_on_diag(pp_diag,1.0_dp)
!    CALL cp_dbcsr_set_diag(pp_diag,diagonal)
!    DEALLOCATE(diagonal)
!
      ! initialize a matrix to 1.0
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=prec)
      ! use an ugly hack to set all elements of tmp to 1
      ! because cp_dbcsr_set does not do it (despite its name)
      !CALL cp_dbcsr_set(tmp,1.0_dp)
      mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(cp_dbcsr_distribution(tmp)))
      CALL cp_dbcsr_work_create(tmp, work_mutable=.TRUE.)
      nblkrows_tot = cp_dbcsr_nblkrows_total(tmp)
      nblkcols_tot = cp_dbcsr_nblkcols_total(tmp)
      DO row = 1, nblkrows_tot
         DO col = 1, nblkcols_tot
            tr = .FALSE.
            iblock_row = row
            iblock_col = col
            CALL cp_dbcsr_get_stored_coordinates(tmp, iblock_row, iblock_col, hold)
            IF (hold .EQ. mynode) THEN
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(tmp, iblock_row, iblock_col, p_new_block)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("ct_methods.F",1289)
               p_new_block(:, :) = 1.0_dp
            ENDIF ! mynode
         ENDDO
      ENDDO
      CALL cp_dbcsr_finalize(tmp)

      ! copy diagonal elements of pp into cols of a matrix
      CALL cp_dbcsr_get_info(pp, nfullrows_total=p_nrows)
      CALL cp_dbcsr_init(pp_diag)
      CALL cp_dbcsr_create(pp_diag, template=pp)
      ALLOCATE (p_diagonal(p_nrows))
      CALL cp_dbcsr_get_diag(pp, p_diagonal)
      CALL cp_dbcsr_add_on_diag(pp_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(pp_diag, p_diagonal)
      ! RZK-warning is it possible to use cp_dbcsr_scale_by_vector?
      ! or even insert elements directly in the prev cycles
      CALL cp_dbcsr_init(t2)
      CALL cp_dbcsr_create(t2, template=prec)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp, pp_diag, &
                             0.0_dp, t2, filter_eps=eps_filter)

      ! copy diagonal elements qq into rows of a matrix
      CALL cp_dbcsr_get_info(qq, nfullrows_total=q_nrows)
      CALL cp_dbcsr_init(qq_diag)
      CALL cp_dbcsr_create(qq_diag, template=qq)
      ALLOCATE (q_diagonal(q_nrows))
      CALL cp_dbcsr_get_diag(qq, q_diagonal)
      CALL cp_dbcsr_add_on_diag(qq_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(qq_diag, q_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_init(t1)
      CALL cp_dbcsr_create(t1, template=prec)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qq_diag, tmp, &
                             0.0_dp, t1, filter_eps=eps_filter)

      CALL cp_dbcsr_hadamard_product(t1, t2, prec)
      CALL cp_dbcsr_release(t1)
      CALL cp_dbcsr_scale(prec, 2.0_dp)

      ! Get the diagonal of tr(qq).qq
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, qq, qq, &
                             0.0_dp, qq_diag, retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_get_diag(qq_diag, q_diagonal)
      CALL cp_dbcsr_set(qq_diag, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(qq_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(qq_diag, q_diagonal)
      DEALLOCATE (q_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qq_diag, tmp, &
                             0.0_dp, t2, filter_eps=eps_filter)
      CALL cp_dbcsr_release(qq_diag)
      CALL cp_dbcsr_add(prec, t2, 1.0_dp, 1.0_dp)

      ! Get the diagonal of pp.tr(pp)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, pp, pp, &
                             0.0_dp, pp_diag, retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_get_diag(pp_diag, p_diagonal)
      CALL cp_dbcsr_set(pp_diag, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(pp_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(pp_diag, p_diagonal)
      DEALLOCATE (p_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp, pp_diag, &
                             0.0_dp, t2, filter_eps=eps_filter)
      CALL cp_dbcsr_release(tmp)
      CALL cp_dbcsr_release(pp_diag)
      CALL cp_dbcsr_add(prec, t2, 1.0_dp, 1.0_dp)

      ! now add the residual component
      !CALL cp_dbcsr_hadamard_product(res,qp,t2)
      !CALL cp_dbcsr_add(prec,t2,1.0_dp,-2.0_dp)
      CALL cp_dbcsr_release(t2)
      CALL cp_dbcsr_function_of_elements(prec, func=dbcsr_func_inverse)
      CALL cp_dbcsr_filter(prec, eps_filter)

      CALL timestop(handle)

   END SUBROUTINE create_preconditioner

! **************************************************************************************************
!> \brief Finds real roots of a cubic equation
!>    >        a*x**3 + b*x**2 + c*x + d = 0
!>        and returns only those roots for which the derivative is positive
!>
!>   Step 0: Check the true order of the equation. Cubic, quadratic, linear?
!>   Step 1: Calculate p and q
!>           p = ( 3*c/a - (b/a)**2 ) / 3
!>           q = ( 2*(b/a)**3 - 9*b*c/a/a + 27*d/a ) / 27
!>   Step 2: Calculate discriminant D
!>           D = (p/3)**3 + (q/2)**2
!>   Step 3: Depending on the sign of D, we follow different strategy.
!>           If D<0, three distinct real roots.
!>           If D=0, three real roots of which at least two are equal.
!>           If D>0, one real and two complex roots.
!>   Step 3a: For D>0 and D=0,
!>           Calculate u and v
!>           u = cubic_root(-q/2 + sqrt(D))
!>           v = cubic_root(-q/2 - sqrt(D))
!>           Find the three transformed roots
!>           y1 = u + v
!>           y2 = -(u+v)/2 + i (u-v)*sqrt(3)/2
!>           y3 = -(u+v)/2 - i (u-v)*sqrt(3)/2
!>   Step 3b Alternately, for D<0, a trigonometric formulation is more convenient
!>           y1 =  2 * sqrt(|p|/3) * cos(phi/3)
!>           y2 = -2 * sqrt(|p|/3) * cos((phi+pi)/3)
!>           y3 = -2 * sqrt(|p|/3) * cos((phi-pi)/3)
!>           where phi = acos(-q/2/sqrt(|p|**3/27))
!>                 pi  = 3.141592654...
!>   Step 4  Find the real roots
!>           x = y - b/a/3
!>   Step 5  Check the derivative and return only those real roots
!>           for which the derivative is positive
!>
!> \param a ...
!> \param b ...
!> \param c ...
!> \param d ...
!> \param minima ...
!> \param nmins ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE analytic_line_search(a, b, c, d, minima, nmins)

      REAL(KIND=dp), INTENT(IN)                          :: a, b, c, d
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: minima
      INTEGER, INTENT(OUT)                               :: nmins

      CHARACTER(len=*), PARAMETER :: routineN = 'analytic_line_search', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nroots
      REAL(KIND=dp)                                      :: DD, der, p, phi, pi, q, temp1, temp2, u, &
                                                            v, y1, y2, y2i, y2r, y3
      REAL(KIND=dp), DIMENSION(3)                        :: x

!    CALL timeset(routineN,handle)

      pi = ACOS(-1.0_dp)

      ! Step 0: Check coefficients and find the true order of the eq
      IF (a .EQ. 0.0_dp) THEN
         IF (b .EQ. 0.0_dp) THEN
            IF (c .EQ. 0.0_dp) THEN
               ! Non-equation, no valid solutions
               nroots = 0
            ELSE
               ! Linear equation with one root.
               nroots = 1
               x(1) = -d/c
            ENDIF
         ELSE
            ! Quadratic equation with max two roots.
            DD = c*c-4.0_dp*b*d
            IF (DD .GT. 0.0_dp) THEN
               nroots = 2
               x(1) = (-c+SQRT(DD))/2.0_dp/b
               x(2) = (-c-SQRT(DD))/2.0_dp/b
            ELSE IF (DD .LT. 0.0_dp) THEN
               nroots = 0
            ELSE
               nroots = 1
               x(1) = -c/2.0_dp/b
            ENDIF
         ENDIF
      ELSE
         ! Cubic equation with max three roots
         ! Calculate p and q
         p = c/a-b*b/a/a/3.0_dp
         q = (2.0_dp*b*b*b/a/a/a-9.0_dp*b*c/a/a+27.0_dp*d/a)/27.0_dp

         ! Calculate DD
         DD = p*p*p/27.0_dp+q*q/4.0_dp

         IF (DD .LT. 0.0_dp) THEN
            ! three real unequal roots -- use the trigonometric formulation
            phi = ACOS(-q/2.0_dp/SQRT(ABS(p*p*p)/27.0_dp))
            temp1 = 2.0_dp*SQRT(ABS(p)/3.0_dp)
            y1 = temp1*COS(phi/3.0_dp)
            y2 = -temp1*COS((phi+pi)/3.0_dp)
            y3 = -temp1*COS((phi-pi)/3.0_dp)
         ELSE
            ! 1 real & 2 conjugate complex roots OR 3 real roots (some are equal)
            temp1 = -q/2.0_dp+SQRT(DD)
            temp2 = -q/2.0_dp-SQRT(DD)
            u = ABS(temp1)**(1.0_dp/3.0_dp)
            v = ABS(temp2)**(1.0_dp/3.0_dp)
            IF (temp1 .LT. 0.0_dp) u = -u
            IF (temp2 .LT. 0.0_dp) v = -v
            y1 = u+v
            y2r = -(u+v)/2.0_dp
            y2i = (u-v)*SQRT(3.0_dp)/2.0_dp
         ENDIF

         ! Final transformation
         temp1 = b/a/3.0_dp
         y1 = y1-temp1
         y2 = y2-temp1
         y3 = y3-temp1
         y2r = y2r-temp1

         ! Assign answers
         IF (DD .LT. 0.0_dp) THEN
            nroots = 3
            x(1) = y1
            x(2) = y2
            x(3) = y3
         ELSE IF (DD .EQ. 0.0_dp) THEN
            nroots = 2
            x(1) = y1
            x(2) = y2r
            !x(3) = cmplx(y2r,  0.)
         ELSE
            nroots = 1
            x(1) = y1
            !x(2) = cmplx(y2r, y2i)
            !x(3) = cmplx(y2r,-y2i)
         ENDIF

      ENDIF

!write(*,'(i2,a)') nroots, ' real root(s)'
      nmins = 0
      DO i = 1, nroots
         ! maximum or minimum? use the derivative
         ! 3*a*x**2+2*b*x+c
         der = 3.0_dp*a*x(i)*x(i)+2.0_dp*b*x(i)+c
         IF (der .GT. 0.0_dp) THEN
            nmins = nmins+1
            minima(nmins) = x(i)
!write(*,'(a,i2,a,f10.5)') 'Minimum ', i, ', value: ', x(i)
         ENDIF
      ENDDO

!    CALL timestop(handle)

   END SUBROUTINE analytic_line_search

! **************************************************************************************************
!> \brief Diagonalizes diagonal blocks of a symmetric dbcsr matrix
!>        and returs its eigenvectors
!> \param matrix ...
!> \param c ...
!> \param e ...
!> \par History
!>       2011.07 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE diagonalize_diagonal_blocks(matrix, c, e)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: c
      TYPE(cp_dbcsr_type), INTENT(OUT), OPTIONAL         :: e

      CHARACTER(len=*), PARAMETER :: routineN = 'diagonalize_diagonal_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iblock_col, iblock_row, &
                                                            iblock_size, info, lwork, orbital
      LOGICAL                                            :: block_needed, do_eigenvalues
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: eigenvalues, work
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: data_copy
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: data_p, p_new_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      IF (PRESENT(e)) THEN
         do_eigenvalues = .TRUE.
      ELSE
         do_eigenvalues = .FALSE.
      ENDIF

      ! create a matrix for eigenvectors
      CALL cp_dbcsr_work_create(c, work_mutable=.TRUE.)
      IF (do_eigenvalues) &
         CALL cp_dbcsr_work_create(e, work_mutable=.TRUE.)

      CALL cp_dbcsr_iterator_start(iter, matrix)

      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

         CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, data_p, row_size=iblock_size)

         block_needed = .FALSE.
         IF (iblock_row == iblock_col) block_needed = .TRUE.

         IF (block_needed) THEN

            ! Prepare data
            ALLOCATE (eigenvalues(iblock_size))
            ALLOCATE (data_copy(iblock_size, iblock_size))
            data_copy(:, :) = data_p(:, :)

            ! Query the optimal workspace for dsyev
            LWORK = -1
            ALLOCATE (WORK(MAX(1, LWORK)))
            CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
            LWORK = INT(WORK(1))
            DEALLOCATE (WORK)

            ! Allocate the workspace and solve the eigenproblem
            ALLOCATE (WORK(MAX(1, LWORK)))
            CALL DSYEV('V', 'L', iblock_size, data_copy, iblock_size, eigenvalues, WORK, LWORK, INFO)
            IF (INFO .NE. 0) THEN
               CALL cp__b("ct_methods.F",1598,"DSYEV failed")
            END IF

            ! copy eigenvectors into a cp_dbcsr matrix
            NULLIFY (p_new_block)
            CALL cp_dbcsr_reserve_block2d(c, iblock_row, iblock_col, p_new_block)
            IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("ct_methods.F",1604)
            p_new_block(:, :) = data_copy(:, :)

            ! if requested copy eigenvalues into a cp_dbcsr matrix
            IF (do_eigenvalues) THEN
               NULLIFY (p_new_block)
               CALL cp_dbcsr_reserve_block2d(e, iblock_row, iblock_col, p_new_block)
               IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("ct_methods.F",1611)
               p_new_block(:, :) = 0.0_dp
               DO orbital = 1, iblock_size
                  p_new_block(orbital, orbital) = eigenvalues(orbital)
               ENDDO
            ENDIF

            DEALLOCATE (WORK)
            DEALLOCATE (data_copy)
            DEALLOCATE (eigenvalues)

         ENDIF

      ENDDO

      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_finalize(c)
      IF (do_eigenvalues) CALL cp_dbcsr_finalize(e)

      CALL timestop(handle)

   END SUBROUTINE diagonalize_diagonal_blocks

! **************************************************************************************************
!> \brief Transforms a matrix M_out = tr(U1) * M_in * U2
!> \param matrix ...
!> \param u1 ...
!> \param u2 ...
!> \param eps_filter ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_forward_transform(matrix, u1, u2, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: u1, u2
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_forward_transform', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: tmp

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=matrix, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix, u2, 0.0_dp, tmp, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, u1, tmp, 0.0_dp, matrix, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(tmp)

      CALL timestop(handle)

   END SUBROUTINE matrix_forward_transform

! **************************************************************************************************
!> \brief Transforms a matrix M_out = U1 * M_in * tr(U2)
!> \param matrix ...
!> \param u1 ...
!> \param u2 ...
!> \param eps_filter ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE matrix_backward_transform(matrix, u1, u2, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: u1, u2
      REAL(KIND=dp), INTENT(IN)                          :: eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'matrix_backward_transform', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: tmp

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=matrix, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix, u2, 0.0_dp, tmp, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, u1, tmp, 0.0_dp, matrix, &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(tmp)

      CALL timestop(handle)

   END SUBROUTINE matrix_backward_transform

!! **************************************************************************************************
!!> \brief Transforms to a representation in which diagonal blocks
!!>        of qq and pp matrices are diagonal. This can improve convergence
!!>        of PCG
!!> \par History
!!>       2011.07 created [Rustam Z Khaliullin]
!!> \author Rustam Z Khaliullin
!! **************************************************************************************************
!  SUBROUTINE transform_matrices_to_blk_diag(matrix_pp,matrix_qq,matrix_qp,&
!    matrix_pq,eps_filter)
!
!    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_pp, matrix_qq,&
!                                                matrix_qp, matrix_pq
!    REAL(KIND=dp), INTENT(IN)                :: eps_filter
!
!    CHARACTER(len=*), PARAMETER :: routineN = 'transform_matrices_to_blk_diag',&
!      routineP = moduleN//':'//routineN
!
!    TYPE(cp_dbcsr_type)                      :: tmp_pp, tmp_qq,&
!                                                tmp_qp, tmp_pq,&
!                                                blk, blk2
!    INTEGER                                  :: handle
!
!    CALL timeset(routineN,handle)
!
!    ! find a better basis by diagonalizing diagonal blocks
!    ! first pp
!    CALL cp_dbcsr_init(blk)
!    CALL cp_dbcsr_create(blk,template=matrix_pp)
!    CALL diagonalize_diagonal_blocks(matrix_pp,blk)
!
!    ! convert matrices to the new basis
!    CALL cp_dbcsr_init(tmp_pp)
!    CALL cp_dbcsr_create(tmp_pp,template=matrix_pp)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,matrix_pp,blk,0.0_dp,tmp_pp,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_multiply("T","N",1.0_dp,blk,tmp_pp,0.0_dp,matrix_pp,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_release(tmp_pp)
!
!    ! now qq
!    CALL cp_dbcsr_init(blk2)
!    CALL cp_dbcsr_create(blk2,template=matrix_qq)
!    CALL diagonalize_diagonal_blocks(matrix_qq,blk2)
!
!    CALL cp_dbcsr_init(tmp_qq)
!    CALL cp_dbcsr_create(tmp_qq,template=matrix_qq)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,matrix_qq,blk2,0.0_dp,tmp_qq,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_multiply("T","N",1.0_dp,blk2,tmp_qq,0.0_dp,matrix_qq,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_release(tmp_qq)
!
!    ! transform pq
!    CALL cp_dbcsr_init(tmp_pq)
!    CALL cp_dbcsr_create(tmp_pq,template=matrix_pq)
!    CALL cp_dbcsr_multiply("T","N",1.0_dp,blk,matrix_pq,0.0_dp,tmp_pq,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,tmp_pq,blk2,0.0_dp,matrix_pq,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_release(tmp_pq)
!
!    ! transform qp
!    CALL cp_dbcsr_init(tmp_qp)
!    CALL cp_dbcsr_create(tmp_qp,template=matrix_qp)
!    CALL cp_dbcsr_multiply("N","N",1.0_dp,matrix_qp,blk,0.0_dp,tmp_qp,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_multiply("T","N",1.0_dp,blk2,tmp_qp,0.0_dp,matrix_qp,&
!               filter_eps=eps_filter)
!    CALL cp_dbcsr_release(tmp_qp)
!
!    CALL cp_dbcsr_release(blk2)
!    CALL cp_dbcsr_release(blk)
!
!    CALL timestop(handle)
!
!  END SUBROUTINE transform_matrices_to_blk_diag

! **************************************************************************************************
!> \brief computes oo, ov, vo, and vv blocks of the ks matrix
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
!  SUBROUTINE ct_step_env_execute(env)
!
!    TYPE(ct_step_env_type)                      :: env
!
!    CHARACTER(len=*), PARAMETER :: routineN = 'ct_step_env_execute', &
!      routineP = moduleN//':'//routineN
!
!    INTEGER                                  :: handle
!
!    CALL timeset(routineN,handle)
!
!
!    CALL timestop(handle)
!
!  END SUBROUTINE ct_step_env_execute

END MODULE ct_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Optimization routines for all ALMO-based SCF methods
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!>       2014.10 as a separate file [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf_optimizer
   USE almo_scf_diis_types,             ONLY: almo_scf_diis_extrapolate,&
                                              almo_scf_diis_init,&
                                              almo_scf_diis_push,&
                                              almo_scf_diis_release,&
                                              almo_scf_diis_type
   USE almo_scf_methods,                ONLY: &
        almo_scf_ks_blk_to_tv_blk, almo_scf_ks_to_ks_blk, almo_scf_ks_to_ks_xx, &
        almo_scf_ks_xx_to_tv_xx, almo_scf_p_blk_to_t_blk, almo_scf_t_blk_to_p, &
        almo_scf_t_blk_to_t_blk_orthonormal, almo_scf_t_to_p, apply_domain_operators, &
        apply_projector, construct_domain_preconditioner, construct_domain_r_down, &
        construct_domain_s_inv, construct_domain_s_sqrt, get_overlap, newton_grad_to_step, &
        pseudo_invert_diagonal_blk
   USE almo_scf_qs,                     ONLY: almo_scf_dm_to_ks,&
                                              almo_scf_update_ks_energy,&
                                              matrix_qs_to_almo
   USE almo_scf_types,                  ONLY: almo_scf_env_type,&
                                              optimizer_options_type
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_invert,&
                                              cp_dbcsr_cholesky_restore
   USE cp_dbcsr_diag,                   ONLY: cp_dbcsr_syevd
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_desymmetrize, &
        cp_dbcsr_distribution, cp_dbcsr_filter, cp_dbcsr_finalize, cp_dbcsr_frobenius_norm, &
        cp_dbcsr_function_of_elements, cp_dbcsr_get_diag, cp_dbcsr_get_info, &
        cp_dbcsr_hadamard_product, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_norm, cp_dbcsr_p_type, &
        cp_dbcsr_print_block_sum, cp_dbcsr_release, cp_dbcsr_reserve_block2d, cp_dbcsr_scale, &
        cp_dbcsr_set, cp_dbcsr_set_diag, cp_dbcsr_trace, cp_dbcsr_triu, cp_dbcsr_type, &
        cp_dbcsr_work_create, dbcsr_distribution_mp, dbcsr_func_artanh, dbcsr_func_dtanh, &
        dbcsr_func_inverse, dbcsr_func_tanh, dbcsr_mp_mynode, dbcsr_mp_numnodes, &
        dbcsr_norm_maxabsnorm, dbcsr_type_no_symmetry
   USE cp_external_control,             ONLY: external_control
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE ct_methods,                      ONLY: analytic_line_search,&
                                              ct_step_execute,&
                                              diagonalize_diagonal_blocks
   USE ct_types,                        ONLY: ct_step_env_clean,&
                                              ct_step_env_get,&
                                              ct_step_env_init,&
                                              ct_step_env_set,&
                                              ct_step_env_type
   USE domain_submatrix_methods,        ONLY: add_submatrices,&
                                              construct_submatrices,&
                                              copy_submatrices,&
                                              init_submatrices,&
                                              maxnorm_submatrices,&
                                              release_submatrices
   USE domain_submatrix_types,          ONLY: domain_submatrix_type,&
                                              select_row
   USE input_constants,                 ONLY: &
        almo_scf_diag, almo_scf_dm_sign, cg_dai_yuan, cg_fletcher, cg_fletcher_reeves, &
        cg_hager_zhang, cg_hestenes_stiefel, cg_liu_storey, cg_polak_ribiere, cg_zero, prec_zero, &
        virt_full, xalmo_case_block_diag, xalmo_case_fully_deloc, xalmo_case_normal
   USE iterate_matrix,                  ONLY: invert_Hotelling,&
                                              matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf_optimizer'

   PUBLIC :: almo_scf_block_diagonal, &
             almo_scf_xalmo_eigensolver, &
             almo_scf_xalmo_pcg

   LOGICAL, PARAMETER :: debug_mode = .FALSE.
   LOGICAL, PARAMETER :: safe_mode = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief An SCF procedure that optimizes block-diagonal ALMOs using DIIS
!> \param qs_env ...
!> \param almo_scf_env ...
!> \param optimizer ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_block_diagonal(qs_env, almo_scf_env, optimizer)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env
      TYPE(optimizer_options_type)                       :: optimizer

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_block_diagonal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iscf, ispin, nspin, nstrikes, &
                                                            ntolerate, unit_nr
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: local_nocc_of_domain
      LOGICAL                                            :: converged, prepare_to_exit, should_stop, &
                                                            use_diis, use_prev_as_guess
      REAL(KIND=dp)                                      :: energy_diff, energy_new, energy_old, &
                                                            error_norm, error_norm_ispin, &
                                                            prev_error_norm, t1, t2, &
                                                            true_mixing_fraction
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: local_mu
      TYPE(almo_scf_diis_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: almo_diis
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: matrix_mixing_old_blk
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! use DIIS, it's superior to simple mixing
      use_diis = .TRUE.
      use_prev_as_guess = .FALSE.

      nspin = almo_scf_env%nspins
      ALLOCATE (local_mu(almo_scf_env%ndomains))
      ALLOCATE (local_nocc_of_domain(almo_scf_env%ndomains))

      ! init mixing matrices
      ALLOCATE (matrix_mixing_old_blk(nspin))
      ALLOCATE (almo_diis(nspin))
      DO ispin = 1, nspin
         CALL cp_dbcsr_init(matrix_mixing_old_blk(ispin))
         CALL cp_dbcsr_create(matrix_mixing_old_blk(ispin), &
                              template=almo_scf_env%matrix_ks_blk(ispin))
         CALL almo_scf_diis_init(diis_env=almo_diis(ispin), &
                                 sample_err=almo_scf_env%matrix_ks_blk(ispin), &
                                 sample_var=almo_scf_env%matrix_s_blk(1), &
                                 error_type=1, &
                                 max_length=optimizer%ndiis)
      ENDDO

      energy_old = 0.0_dp
      iscf = 0
      prepare_to_exit = .FALSE.
      true_mixing_fraction = 0.0_dp
      ! set variables that control diag/pcg switching
      nstrikes = 0
      ntolerate = 3
      error_norm = 1.0E+10_dp ! arbitrary big step

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 20), &
            " Optimization of block-diagonal ALMOs ", REPEAT("-", 21)
         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T2,A13,A6,A23,A14,A14,A9)') "Method", "Iter", &
            "Total Energy", "Change", "Convergence", "Time"
         WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)
      ENDIF

      ! the real SCF loop
      t1 = m_walltime()
      DO

         iscf = iscf+1

         ! get a copy of the current KS matrix
         CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
         DO ispin = 1, nspin
            CALL matrix_qs_to_almo(matrix_ks(ispin)%matrix, &
                                   almo_scf_env%matrix_ks(ispin), &
                                   almo_scf_env, .FALSE.)
            CALL matrix_qs_to_almo(matrix_ks(ispin)%matrix, &
                                   almo_scf_env%matrix_ks_blk(ispin), &
                                   almo_scf_env, .TRUE.)
            CALL cp_dbcsr_filter(almo_scf_env%matrix_ks(ispin), &
                                 almo_scf_env%eps_filter)
         ENDDO

         ! obtain projected KS matrix and the DIIS-error vector
         CALL almo_scf_ks_to_ks_blk(almo_scf_env)

         ! inform the DIIS handler about the new KS matrix and its error vector
         IF (use_diis) THEN
            DO ispin = 1, nspin
               CALL almo_scf_diis_push(diis_env=almo_diis(ispin), &
                                       var=almo_scf_env%matrix_ks_blk(ispin), &
                                       err=almo_scf_env%matrix_err_blk(ispin))
            ENDDO
         ENDIF

         ! get error_norm: choose the largest of the two spins
         prev_error_norm = error_norm
         DO ispin = 1, nspin
            !error_norm=cp_dbcsr_frobenius_norm(almo_scf_env%matrix_err_blk(ispin))
            CALL cp_dbcsr_norm(almo_scf_env%matrix_err_blk(ispin), &
                               dbcsr_norm_maxabsnorm, &
                               norm_scalar=error_norm_ispin)
            IF (ispin .EQ. 1) error_norm = error_norm_ispin
            IF (ispin .GT. 1 .AND. error_norm_ispin .GT. error_norm) &
               error_norm = error_norm_ispin
         ENDDO

         IF (error_norm .LT. almo_scf_env%eps_prev_guess) THEN
            use_prev_as_guess = .TRUE.
         ELSE
            use_prev_as_guess = .FALSE.
         ENDIF

         ! check convergence
         converged = .TRUE.
         IF (error_norm .GT. optimizer%eps_error) converged = .FALSE.
         ! check other exit criteria: max SCF steps and timing
         CALL external_control(should_stop, "SCF", &
                               start_time=qs_env%start_time, &
                               target_time=qs_env%target_time)
         IF (should_stop .OR. iscf >= optimizer%max_iter .OR. converged) THEN
            prepare_to_exit = .TRUE.
         ENDIF

         IF (.NOT. prepare_to_exit) THEN ! update the ALMOs and density matrix

            ! perform mixing of KS matrices
            IF (iscf .NE. 1) THEN
               IF (use_diis) THEN ! use diis instead of mixing
                  DO ispin = 1, nspin
                     CALL almo_scf_diis_extrapolate(diis_env=almo_diis(ispin), &
                                                    extr_var=almo_scf_env%matrix_ks_blk(ispin))
                  ENDDO
               ELSE ! use mixing
                  true_mixing_fraction = almo_scf_env%mixing_fraction
                  DO ispin = 1, nspin
                     CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin), &
                                       matrix_mixing_old_blk(ispin), &
                                       true_mixing_fraction, &
                                       1.0_dp-true_mixing_fraction)
                  END DO
               ENDIF
            ENDIF
            ! save the new matrix for the future mixing
            DO ispin = 1, nspin
               CALL cp_dbcsr_copy(matrix_mixing_old_blk(ispin), &
                                  almo_scf_env%matrix_ks_blk(ispin))
            ENDDO

            ! obtain ALMOs from the new KS matrix
            SELECT CASE (almo_scf_env%almo_update_algorithm)
            CASE (almo_scf_diag)

               CALL almo_scf_ks_blk_to_tv_blk(almo_scf_env)

            CASE (almo_scf_dm_sign)

               ! update the density matrix
               DO ispin = 1, nspin

                  local_nocc_of_domain(:) = almo_scf_env%nocc_of_domain(:, ispin)
                  local_mu(:) = almo_scf_env%mu_of_domain(:, ispin)
                  ! RZK UPDATE! the update algorithm is removed because
                  ! RZK UPDATE! it requires updating core LS_SCF routines
                  ! RZK UPDATE! (the code exists in the CVS version)
                  CALL cp__b("almo_scf_optimizer.F",281,"Density_matrix_sign has not been tested yet")
                  ! RZK UPDATE!  CALL density_matrix_sign(almo_scf_env%matrix_p_blk(ispin),&
                  ! RZK UPDATE!          local_mu,&
                  ! RZK UPDATE!          almo_scf_env%fixed_mu,&
                  ! RZK UPDATE!          almo_scf_env%matrix_ks_blk(ispin),&
                  ! RZK UPDATE!          !matrix_mixing_old_blk(ispin),&
                  ! RZK UPDATE!          almo_scf_env%matrix_s_blk(1), &
                  ! RZK UPDATE!          almo_scf_env%matrix_s_blk_inv(1), &
                  ! RZK UPDATE!          local_nocc_of_domain,&
                  ! RZK UPDATE!          almo_scf_env%eps_filter,&
                  ! RZK UPDATE!          almo_scf_env%domain_index_of_ao)
                  ! RZK UPDATE!
                  almo_scf_env%mu_of_domain(:, ispin) = local_mu(:)

               ENDDO

               ! obtain ALMOs from matrix_p_blk: T_new = P_blk S_blk T_old
               CALL almo_scf_p_blk_to_t_blk(almo_scf_env)
               CALL almo_scf_t_blk_to_t_blk_orthonormal(almo_scf_env)

            END SELECT

            ! obtain density matrix from ALMOs
            CALL almo_scf_t_blk_to_p(almo_scf_env, &
                                     use_sigma_inv_guess=use_prev_as_guess)

            ! compute the new KS matrix and new energy
            CALL almo_scf_dm_to_ks(qs_env, almo_scf_env, energy_new)

         ENDIF ! prepare_to_exit

         energy_diff = energy_new-energy_old
         energy_old = energy_new
         almo_scf_env%almo_scf_energy = energy_new

         t2 = m_walltime()
         ! brief report on the current SCF loop
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T2,A13,I6,F23.10,E14.5,F14.9,F9.2)') "ALMO SCF DIIS", &
               iscf, &
               energy_new, energy_diff, error_norm, t2-t1
         ENDIF
         t1 = m_walltime()

         IF (prepare_to_exit) EXIT

      ENDDO ! end scf cycle

      IF (.NOT. converged) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T2,A)') "SCF for block-diagonal ALMOs not converged! "
         ! DANGER: handle non-convergent procedures outside since other outer
         ! methods can be used
         !CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
      ENDIF

      DO ispin = 1, nspin
         CALL cp_dbcsr_release(matrix_mixing_old_blk(ispin))
         CALL almo_scf_diis_release(diis_env=almo_diis(ispin))
      ENDDO
      DEALLOCATE (almo_diis)
      DEALLOCATE (matrix_mixing_old_blk)
      DEALLOCATE (local_mu)
      DEALLOCATE (local_nocc_of_domain)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_block_diagonal

! **************************************************************************************************
!> \brief An eigensolver-based SCF to optimize extended ALMOs (i.e. ALMOs on
!>        overlapping domains)
!> \param qs_env ...
!> \param almo_scf_env ...
!> \param optimizer ...
!> \par History
!>       2013.03 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_xalmo_eigensolver(qs_env, almo_scf_env, optimizer)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env
      TYPE(optimizer_options_type)                       :: optimizer

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_xalmo_eigensolver', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iscf, ispin, nspin, unit_nr
      LOGICAL                                            :: converged, prepare_to_exit, should_stop
      REAL(KIND=dp)                                      :: denergy_tot, energy_diff, energy_new, &
                                                            energy_old, error_norm, error_norm_0, &
                                                            spin_factor, t1, t2
      REAL(KIND=dp), DIMENSION(2)                        :: denergy_spin
      TYPE(almo_scf_diis_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: almo_diis
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type)                                :: matrix_p_almo_scf_converged
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:, :)                                 :: submatrix_mixing_old_blk

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = almo_scf_env%nspins
      IF (nspin == 1) THEN
         spin_factor = 2.0_dp
      ELSE
         spin_factor = 1.0_dp
      ENDIF

      ! RZK-warning domain_s_sqrt and domain_s_sqrt_inv do not have spin
      ! components yet (may be used later)
      ispin = 1
      CALL construct_domain_s_sqrt( &
         matrix_s=almo_scf_env%matrix_s(1), &
         subm_s_sqrt=almo_scf_env%domain_s_sqrt(:, ispin), &
         subm_s_sqrt_inv=almo_scf_env%domain_s_sqrt_inv(:, ispin), &
         dpattern=almo_scf_env%quench_t(ispin), &
         map=almo_scf_env%domain_map(ispin), &
         node_of_domain=almo_scf_env%cpu_of_domain)
      ! TRY: construct s_inv
      !CALL construct_domain_s_inv(&
      !       matrix_s=almo_scf_env%matrix_s(1),&
      !       subm_s_inv=almo_scf_env%domain_s_inv(:,ispin),&
      !       dpattern=almo_scf_env%quench_t(ispin),&
      !       map=almo_scf_env%domain_map(ispin),&
      !       node_of_domain=almo_scf_env%cpu_of_domain)

      ! construct the domain template for the occupied orbitals
      DO ispin = 1, nspin
         ! RZK-warning we need only the matrix structure, not data
         ! replace construct_submatrices with lighter procedure with
         ! no heavy communications
         CALL construct_submatrices( &
            matrix=almo_scf_env%quench_t(ispin), &
            submatrix=almo_scf_env%domain_t(:, ispin), &
            distr_pattern=almo_scf_env%quench_t(ispin), &
            domain_map=almo_scf_env%domain_map(ispin), &
            node_of_domain=almo_scf_env%cpu_of_domain, &
            job_type=select_row)
      ENDDO

      ! init mixing matrices
      ALLOCATE (submatrix_mixing_old_blk(almo_scf_env%ndomains, nspin))
      CALL init_submatrices(submatrix_mixing_old_blk)
      ALLOCATE (almo_diis(nspin))

      ! TRY: construct block-projector
      !ALLOCATE(submatrix_tmp(almo_scf_env%ndomains))
      !DO ispin=1,nspin
      !   CALL init_submatrices(submatrix_tmp)
      !   CALL construct_domain_r_down(&
      !           matrix_t=almo_scf_env%matrix_t_blk(ispin),&
      !           matrix_sigma_inv=almo_scf_env%matrix_sigma_inv(ispin),&
      !           matrix_s=almo_scf_env%matrix_s(1),&
      !           subm_r_down=submatrix_tmp(:),&
      !           dpattern=almo_scf_env%quench_t(ispin),&
      !           map=almo_scf_env%domain_map(ispin),&
      !           node_of_domain=almo_scf_env%cpu_of_domain,&
      !           filter_eps=almo_scf_env%eps_filter)
      !   CALL multiply_submatrices('N','N',1.0_dp,&
      !           submatrix_tmp(:),&
      !           almo_scf_env%domain_s_inv(:,1),0.0_dp,&
      !           almo_scf_env%domain_r_down_up(:,ispin))
      !   CALL release_submatrices(submatrix_tmp)
      !ENDDO
      !DEALLOCATE(submatrix_tmp)

      DO ispin = 1, nspin
         ! use s_sqrt since they are already properly constructed
         ! and have the same distributions as domain_err and domain_ks_xx
         CALL almo_scf_diis_init(diis_env=almo_diis(ispin), &
                                 sample_err=almo_scf_env%domain_s_sqrt(:, ispin), &
                                 error_type=1, &
                                 max_length=optimizer%ndiis)
      ENDDO

      denergy_tot = 0.0_dp
      energy_old = 0.0_dp
      iscf = 0
      prepare_to_exit = .FALSE.

      ! the SCF loop
      t1 = m_walltime()
      DO

         iscf = iscf+1

         ! get a copy of the current KS matrix
         CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
         DO ispin = 1, nspin
            CALL matrix_qs_to_almo(matrix_ks(ispin)%matrix, &
                                   almo_scf_env%matrix_ks(ispin), &
                                   almo_scf_env, .FALSE.)
            CALL matrix_qs_to_almo(matrix_ks(ispin)%matrix, &
                                   almo_scf_env%matrix_ks_blk(ispin), &
                                   almo_scf_env, .TRUE.)
            CALL cp_dbcsr_filter(almo_scf_env%matrix_ks(ispin), &
                                 almo_scf_env%eps_filter)
         ENDDO

         ! obtain projected KS matrix and the DIIS-error vector
         CALL almo_scf_ks_to_ks_xx(almo_scf_env)

         ! inform the DIIS handler about the new KS matrix and its error vector
         DO ispin = 1, nspin
            CALL almo_scf_diis_push(diis_env=almo_diis(ispin), &
                                    d_var=almo_scf_env%domain_ks_xx(:, ispin), &
                                    d_err=almo_scf_env%domain_err(:, ispin))
         ENDDO

         ! check convergence
         converged = .TRUE.
         DO ispin = 1, nspin
            !error_norm=cp_dbcsr_frobenius_norm(almo_scf_env%matrix_err_blk(ispin))
            CALL cp_dbcsr_norm(almo_scf_env%matrix_err_xx(ispin), &
                               dbcsr_norm_maxabsnorm, &
                               norm_scalar=error_norm)
            CALL maxnorm_submatrices(almo_scf_env%domain_err(:, ispin), &
                                     norm=error_norm_0)
            IF (error_norm .GT. optimizer%eps_error) THEN
               converged = .FALSE.
               EXIT ! no need to check the other spin
            ENDIF
         ENDDO
         ! check other exit criteria: max SCF steps and timing
         CALL external_control(should_stop, "SCF", &
                               start_time=qs_env%start_time, &
                               target_time=qs_env%target_time)
         IF (should_stop .OR. iscf >= optimizer%max_iter .OR. converged) THEN
            prepare_to_exit = .TRUE.
         ENDIF

         IF (.NOT. prepare_to_exit) THEN ! update the ALMOs and density matrix

            ! perform mixing of KS matrices
            IF (iscf .NE. 1) THEN
               IF (.FALSE.) THEN ! use diis instead of mixing
                  DO ispin = 1, nspin
                     CALL add_submatrices( &
                        almo_scf_env%mixing_fraction, &
                        almo_scf_env%domain_ks_xx(:, ispin), &
                        1.0_dp-almo_scf_env%mixing_fraction, &
                        submatrix_mixing_old_blk(:, ispin), &
                        'N')
                     !CALL cp_dbcsr_add(almo_scf_env%matrix_ks_blk(ispin),&
                     !        matrix_mixing_old_blk(ispin),&
                     !        almo_scf_env%mixing_fraction,&
                     !        1.0_dp-almo_scf_env%mixing_fraction)
                  END DO
               ELSE
                  DO ispin = 1, nspin
                     CALL almo_scf_diis_extrapolate(diis_env=almo_diis(ispin), &
                                                    d_extr_var=almo_scf_env%domain_ks_xx(:, ispin))
                  ENDDO
               ENDIF
            ENDIF
            ! save the new matrix for the future mixing
            DO ispin = 1, nspin
               CALL copy_submatrices( &
                  almo_scf_env%domain_ks_xx(:, ispin), &
                  submatrix_mixing_old_blk(:, ispin), &
                  copy_data=.TRUE.)
            ENDDO

            ! obtain a new set of ALMOs from the updated KS matrix
            CALL almo_scf_ks_xx_to_tv_xx(almo_scf_env)

            ! update the density matrix
            DO ispin = 1, nspin

               ! save the initial density matrix (to get the perturbative energy lowering)
               IF (iscf .EQ. 1) THEN
                  CALL cp_dbcsr_init(matrix_p_almo_scf_converged)
                  CALL cp_dbcsr_create(matrix_p_almo_scf_converged, &
                                       template=almo_scf_env%matrix_p(ispin))
                  CALL cp_dbcsr_copy(matrix_p_almo_scf_converged, &
                                     almo_scf_env%matrix_p(ispin))
               ENDIF

               ! update now
               CALL almo_scf_t_to_p( &
                  t=almo_scf_env%matrix_t(ispin), &
                  p=almo_scf_env%matrix_p(ispin), &
                  eps_filter=almo_scf_env%eps_filter, &
                  orthog_orbs=.FALSE., &
                  s=almo_scf_env%matrix_s(1), &
                  sigma=almo_scf_env%matrix_sigma(ispin), &
                  sigma_inv=almo_scf_env%matrix_sigma_inv(ispin), &
                  use_guess=.TRUE.)
               CALL cp_dbcsr_scale(almo_scf_env%matrix_p(ispin), spin_factor)

               ! obtain perturbative estimate (at no additional cost)
               ! of the energy lowering relative to the block-diagonal ALMOs
               IF (iscf .EQ. 1) THEN

                  CALL cp_dbcsr_add(matrix_p_almo_scf_converged, &
                                    almo_scf_env%matrix_p(ispin), -1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_trace(almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                      matrix_p_almo_scf_converged, &
                                      denergy_spin(ispin))

                  CALL cp_dbcsr_release(matrix_p_almo_scf_converged)

                  denergy_tot = denergy_tot+denergy_spin(ispin)

                  ! RZK-warning Energy correction can be evaluated using matrix_x
                  ! as shown in the attempt below and in the PCG procedure.
                  ! Using matrix_x allows immediate decomposition of the energy
                  ! lowering into 2-body components for EDA. However, it does not
                  ! work here because the diagonalization routine does not necessarily
                  ! produce orbitals with the same sign as the block-diagonal ALMOs
                  ! Any fixes?!

                  !CALL cp_dbcsr_init(matrix_x)
                  !CALL cp_dbcsr_create(matrix_x,&
                  !        template=almo_scf_env%matrix_t(ispin))
                  !
                  !CALL cp_dbcsr_init(matrix_tmp_no)
                  !CALL cp_dbcsr_create(matrix_tmp_no,&
                  !        template=almo_scf_env%matrix_t(ispin))
                  !
                  !CALL cp_dbcsr_copy(matrix_x,&
                  !        almo_scf_env%matrix_t_blk(ispin))
                  !CALL cp_dbcsr_add(matrix_x,almo_scf_env%matrix_t(ispin),&
                  !        -1.0_dp,1.0_dp)

                  !CALL cp_dbcsr_trace(matrix_x,&
                  !        almo_scf_env%matrix_err_xx(ispin),denergy,"T","N")

                  !denergy=denergy*spin_factor

                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "_ENERGY-0: ", almo_scf_env%almo_scf_energy
                  !   WRITE(unit_nr,*) "_ENERGY-D: ", denergy
                  !   WRITE(unit_nr,*) "_ENERGY-F: ", almo_scf_env%almo_scf_energy+denergy
                  !ENDIF
                  !! RZK-warning update will not work since the energy is overwritten almost immediately
                  !!CALL almo_scf_update_ks_energy(qs_env,&
                  !!        almo_scf_env%almo_scf_energy+denergy)
                  !!

                  !! print out the results of the decomposition analysis
                  !CALL cp_dbcsr_hadamard_product(matrix_x,&
                  !        almo_scf_env%matrix_err_xx(ispin),&
                  !        matrix_tmp_no)
                  !CALL cp_dbcsr_scale(matrix_tmp_no,spin_factor)
                  !CALL cp_dbcsr_filter(matrix_tmp_no,almo_scf_env%eps_filter)
                  !
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*)
                  !   WRITE(unit_nr,'(T2,A)') "DECOMPOSITION OF THE DELOCALIZATION ENERGY"
                  !ENDIF

                  !mynode=dbcsr_mp_mynode(dbcsr_distribution_mp(&
                  !   cp_dbcsr_distribution(matrix_tmp_no)))
                  !WRITE(mynodestr,'(I6.6)') mynode
                  !mylogfile='EDA.'//TRIM(ADJUSTL(mynodestr))
                  !OPEN (iunit,file=mylogfile,status='REPLACE')
                  !CALL cp_dbcsr_print_block_sum(matrix_tmp_no,iunit)
                  !CLOSE(iunit)
                  !
                  !CALL cp_dbcsr_release(matrix_tmp_no)
                  !CALL cp_dbcsr_release(matrix_x)

               ENDIF ! iscf.eq.1

            ENDDO

            ! print out the energy lowering
            IF (iscf .EQ. 1) THEN
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *)
                  WRITE (unit_nr, '(T2,A35,F25.10)') "ENERGY OF BLOCK-DIAGONAL ALMOs:", &
                     almo_scf_env%almo_scf_energy
                  WRITE (unit_nr, '(T2,A35,F25.10)') "ENERGY LOWERING:", &
                     denergy_tot
                  WRITE (unit_nr, '(T2,A35,F25.10)') "CORRECTED ENERGY:", &
                     almo_scf_env%almo_scf_energy+denergy_tot
                  WRITE (unit_nr, *)
               ENDIF
               CALL almo_scf_update_ks_energy(qs_env, &
                                              almo_scf_env%almo_scf_energy+denergy_tot)
            ENDIF

            ! compute the new KS matrix and new energy
            IF (.NOT. almo_scf_env%perturbative_delocalization) THEN
               CALL almo_scf_dm_to_ks(qs_env, almo_scf_env, energy_new)
            ENDIF

         ENDIF ! prepare_to_exit

         IF (almo_scf_env%perturbative_delocalization) THEN

            ! exit after the first step if we do not need the SCF procedure
            converged = .TRUE.
            prepare_to_exit = .TRUE.

         ELSE ! not a perturbative treatment

            energy_diff = energy_new-energy_old
            energy_old = energy_new
            almo_scf_env%almo_scf_energy = energy_new

            t2 = m_walltime()
            ! brief report on the current SCF loop
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T2,A,I6,F20.9,E11.3,E11.3,E11.3,F8.2)') "ALMO SCF", &
                  iscf, &
                  energy_new, energy_diff, error_norm, error_norm_0, t2-t1
            ENDIF
            t1 = m_walltime()

         ENDIF

         IF (prepare_to_exit) EXIT

      ENDDO ! end scf cycle

      IF (.NOT. converged) THEN
         CALL cp__b("almo_scf_optimizer.F",708,"SCF for ALMOs on overlapping domains not converged! ")
      ENDIF

      DO ispin = 1, nspin
         CALL release_submatrices(submatrix_mixing_old_blk(:, ispin))
         CALL almo_scf_diis_release(diis_env=almo_diis(ispin))
      ENDDO
      DEALLOCATE (almo_diis)
      DEALLOCATE (submatrix_mixing_old_blk)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_xalmo_eigensolver

! **************************************************************************************************
!> \brief Optimization of ALMOs using PCG-like minimizers
!> \param qs_env ...
!> \param almo_scf_env ...
!> \param optimizer - controls the optimization algorithm
!> \param quench_t ...
!> \param matrix_t_in ...
!> \param matrix_t_out ...
!> \param assume_t0_q0x ...
!> \param perturbation_only ...
!> \param special_case - to reduce the overhead special cases are implemented:
!>                       xalmo_case_normal - no special case (i.e. xALMOs)
!>                       xalmo_case_block_diag
!>                       xalmo_case_fully_deloc
!> \par History
!>       2011.11 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_xalmo_pcg(qs_env, almo_scf_env, optimizer, quench_t, &
                                 matrix_t_in, matrix_t_out, assume_t0_q0x, perturbation_only, &
                                 special_case)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env
      TYPE(optimizer_options_type)                       :: optimizer
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: quench_t, matrix_t_in, matrix_t_out
      LOGICAL, INTENT(IN)                                :: assume_t0_q0x, perturbation_only
      INTEGER, INTENT(IN), OPTIONAL                      :: special_case

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_xalmo_pcg', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=100)                                 :: mylogfile, mynodestr
      CHARACTER(LEN=20)                                  :: iter_type
      INTEGER :: cg_iteration, dim0, fixed_line_search_niter, handle, ispin, iteration, iunit, jj, &
         line_search_iteration, max_iter, my_special_case, mynode, ncores, ndomains, occ1, &
         outer_iteration, outer_max_iter, prec_type, precond_domain_projector, unit_nr, &
         zero_neg_eiv
      LOGICAL :: converged, do_md, first_md_iteration, just_started, line_search, &
         md_in_theta_space, optimize_theta, outer_prepare_to_exit, prepare_to_exit, &
         reset_conjugator, skip_grad, use_guess, use_preconditioner
      REAL(kind=dp) :: appr_sec_der, beta, denom, e0, e1, energy_diff, energy_new, energy_old, &
         eps_skip_gradients, g0, g1, grad_norm, grad_norm_frob, kappa, kin_energy, &
         line_search_error, next_step_size_guess, numer, prec_sf_mixing_s, spin_factor, step_size, &
         t1, t2, t_norm, tau, time_step
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: evals
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type) :: FTsiginv, fvo_0, grad, inv_eiv, m_theta, m_tmp_nn_1, m_tmp_no_1, &
         m_tmp_no_2, m_tmp_no_3, m_tmp_oo_1, matrix_p_0, matrix_sigma_0, matrix_sigma_inv_0, &
         matrix_t_0, prec_oo, prec_oo_inv, prec_vv, prev_grad, prev_minus_prec_grad, prev_step, &
         siginvTFTsiginv, ST, step, STsiginv_0, velocity
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(domain_submatrix_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: domain_r_down

      CALL timeset(routineN, handle)

      my_special_case = xalmo_case_normal
      IF (PRESENT(special_case)) my_special_case = special_case

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *)
         SELECT CASE (my_special_case)
         CASE (xalmo_case_block_diag)
            WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 20), &
               " Optimization of block-diagonal ALMOs ", REPEAT("-", 21)
         CASE (xalmo_case_fully_deloc)
            WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 20), &
               " Optimization of fully delocalized MOs ", REPEAT("-", 20)
         CASE (xalmo_case_normal)
            WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 27), &
               " Optimization of XALMOs ", REPEAT("-", 28)
         END SELECT
         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T2,A13,A6,A23,A14,A14,A9)') "Method", "Iter", &
            "Objective Function", "Change", "Convergence", "Time"
         WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)
      ENDIF

      ! set local parameters using developer's keywords
      ! RZK-warning: change to normal keywords later
      do_md = almo_scf_env%logical01
      optimize_theta = almo_scf_env%logical05
      prec_sf_mixing_s = almo_scf_env%real04
      eps_skip_gradients = almo_scf_env%real01

      ! preconditioner control
      use_preconditioner = optimizer%preconditioner .NE. prec_zero
      prec_type = 4
      ! RZK-warning: prec_type here is not the same as preconditioner
      ! type in optimizer%preconditioner. change this later
      !prec_type = optimizer%preconditioner
      !if (prec_type.eq.prec_default) prec_type=prec_ks_plus_s

      ! control of the line search
      fixed_line_search_niter = 0 ! init to zero, change when eps is small enough

      ncores = dbcsr_mp_numnodes(dbcsr_distribution_mp( &
                                 cp_dbcsr_distribution(almo_scf_env%matrix_s(1))))

      ! Since it is extrememly difficult to converge the iterative
      ! procedure using T as an optimized variable, assume
      ! T = T_0 + (1-R_0)*X and optimize X for both perturbative
      ! and non-perturbative cases

      IF (almo_scf_env%nspins == 1) THEN
         spin_factor = 2.0_dp
      ELSE
         spin_factor = 1.0_dp
      ENDIF

      !!!!!! RZK-warning THIS PROCEDURE WILL WORK ONLY FOR CLOSED SHELL SYSTEMS
      !!!!!! TO ADAPT IT FOR UNRESTRICTED ORBITALS - UPDATE KS MATRIX WITH PARTIALLY
      !!!!!! OPTIMIZED ORBITALS - BOTH ALPNA AND BETA
      IF (almo_scf_env%nspins .GT. 1) THEN
         CALL cp__b("almo_scf_optimizer.F",845,"UNRESTRICTED ALMO SCF IS NYI(!)")
      ENDIF

      DO ispin = 1, almo_scf_env%nspins

         ! init temporary storage
         CALL cp_dbcsr_init(m_theta)
         CALL cp_dbcsr_init(prec_vv)
         CALL cp_dbcsr_init(fvo_0)
         CALL cp_dbcsr_init(STsiginv_0)
         CALL cp_dbcsr_init(m_tmp_no_1)
         CALL cp_dbcsr_init(m_tmp_no_2)
         CALL cp_dbcsr_init(m_tmp_no_3)
         CALL cp_dbcsr_init(ST)
         CALL cp_dbcsr_init(FTsiginv)
         CALL cp_dbcsr_init(m_tmp_oo_1)
         CALL cp_dbcsr_init(m_tmp_nn_1)
         CALL cp_dbcsr_init(siginvTFTsiginv)
         CALL cp_dbcsr_init(prec_oo)
         CALL cp_dbcsr_init(prec_oo_inv)
         CALL cp_dbcsr_init(prev_grad)
         CALL cp_dbcsr_init(prev_step)
         CALL cp_dbcsr_init(grad)
         CALL cp_dbcsr_init(step)
         CALL cp_dbcsr_init(prev_minus_prec_grad)
         CALL cp_dbcsr_create(prec_vv, &
                              template=almo_scf_env%matrix_ks(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(prec_oo, &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(prec_oo_inv, &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(m_tmp_oo_1, &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(siginvTFTsiginv, &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(STsiginv_0, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(fvo_0, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(m_tmp_no_1, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(m_tmp_no_2, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(m_tmp_no_3, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(FTsiginv, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(ST, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(m_theta, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(prev_grad, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(grad, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(prev_step, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(step, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(prev_minus_prec_grad, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)

         ndomains = almo_scf_env%ndomains
         ALLOCATE (domain_r_down(ndomains))
         CALL init_submatrices(domain_r_down)

         ! create matrices to store the initial state
         CALL cp_dbcsr_init(matrix_t_0)
         CALL cp_dbcsr_init(matrix_sigma_inv_0)
         CALL cp_dbcsr_init(matrix_sigma_0)
         CALL cp_dbcsr_init(matrix_p_0)
         CALL cp_dbcsr_create(matrix_t_0, &
                              template=matrix_t_out(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(matrix_sigma_inv_0, &
                              template=almo_scf_env%matrix_sigma_inv(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(matrix_sigma_0, &
                              template=almo_scf_env%matrix_sigma_inv(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(matrix_p_0, &
                              template=almo_scf_env%matrix_ks(ispin))
         CALL cp_dbcsr_copy(matrix_t_0, matrix_t_in(ispin))

         CALL cp_dbcsr_set(step, 0.0_dp)

         md_in_theta_space = .FALSE. ! turn on later after several minimization steps
         IF (do_md) THEN
            CALL cp_dbcsr_init(velocity)
            CALL cp_dbcsr_create(velocity, &
                                 template=matrix_t_out(ispin))
            CALL cp_dbcsr_copy(velocity, quench_t(ispin))
            CALL cp_dbcsr_set(velocity, 0.0_dp)
            CALL cp_dbcsr_copy(prev_step, quench_t(ispin))
            CALL cp_dbcsr_set(prev_step, 0.0_dp)
            time_step = optimizer%lin_search_step_size_guess
         ENDIF

         ! create initial guess from the initial orbitals
         IF (assume_t0_q0x) THEN
            CALL cp_dbcsr_set(m_theta, 0.0_dp)
         ELSE
            IF (optimize_theta) THEN
               ! check that all MO coefficients of the guess are less
               ! than the maximum allowed amplitude
               CALL cp_dbcsr_norm(matrix_t_0, &
                                  dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
               !IF (unit_nr>0) THEN
               !   WRITE(unit_nr,*) "Maximum norm of the initial guess: ", grad_norm
               !   WRITE(unit_nr,*) "Maximum allowed amplitude: ", almo_scf_env%envelope_amplitude
               !ENDIF
               IF (grad_norm .GT. almo_scf_env%envelope_amplitude) THEN
                  CALL cp__b("almo_scf_optimizer.F",975,"Max norm of the initial guess is too large")
               ENDIF
               ! use artanh of block-diagonal ALMOs as an initial guess for independent variables
               CALL cp_dbcsr_copy(m_theta, matrix_t_0)
               CALL cp_dbcsr_function_of_elements(m_theta, &
                                                  !func=dbcsr_func_asin,&
                                                  func=dbcsr_func_artanh, &
                                                  a0=0.0_dp, &
                                                  a1=1.0_dp/almo_scf_env%envelope_amplitude)
               CALL cp_dbcsr_scale(m_theta, almo_scf_env%envelope_amplitude)
            ELSE
               ! simply copy MO coefficients to m_theta
               CALL cp_dbcsr_copy(m_theta, matrix_t_0)
               CALL cp_dbcsr_norm(m_theta, &
                                  dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
               !IF (unit_nr>0) THEN
               !   WRITE(unit_nr,*) "Maximum norm of the initial guess: ", grad_norm
               !ENDIF
            ENDIF
         ENDIF

         ! invert S domains if necessary
         ! RZK-warning must be done outside the spin loop to save time
         IF (my_special_case .EQ. xalmo_case_normal) THEN
            CALL construct_domain_s_inv( &
               matrix_s=almo_scf_env%matrix_s(1), &
               subm_s_inv=almo_scf_env%domain_s_inv(:, ispin), &
               dpattern=quench_t(ispin), &
               map=almo_scf_env%domain_map(ispin), &
               node_of_domain=almo_scf_env%cpu_of_domain)
         ENDIF

         ! start the outer SCF loop
         outer_max_iter = optimizer%max_iter_outer_loop
         outer_prepare_to_exit = .FALSE.
         outer_iteration = 0
         grad_norm = 0.0_dp
         grad_norm_frob = 0.0_dp
         use_guess = .FALSE.

         DO

            ! start the inner SCF loop
            max_iter = optimizer%max_iter
            prepare_to_exit = .FALSE.
            line_search = .FALSE.
            converged = .FALSE.
            iteration = 0
            cg_iteration = 0
            line_search_iteration = 0
            energy_new = 0.0_dp
            energy_old = 0.0_dp
            line_search_error = 0.0_dp
            t1 = m_walltime()

            DO

               just_started = (iteration .EQ. 0) .AND. (outer_iteration .EQ. 0)

               ! switch to MD after several minimization steps
               IF (iteration .EQ. almo_scf_env%integer01 .AND. do_md) THEN
                  CALL cp_dbcsr_set(velocity, 0.0_dp)
                  CALL cp_dbcsr_set(prev_step, 0.0_dp)
                  md_in_theta_space = .TRUE.
                  first_md_iteration = .TRUE.
               ENDIF

               ! compute the MO coefficients from theta
               IF (assume_t0_q0x .AND. just_started) THEN
                  CALL cp_dbcsr_set(matrix_t_out(ispin), 0.0_dp)
               ENDIF
               IF (optimize_theta) THEN
                  CALL cp_dbcsr_norm(m_theta, &
                                     dbcsr_norm_maxabsnorm, norm_scalar=t_norm)
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "VAR1: |T|: ", t_norm
                  !   !WRITE(unit_nr,*) "VAR1: SIN(|T|/A), COS(|T|/A), A*SIN(|T|/A): ",&
                  !   !   SIN(t_norm/almo_scf_env%envelope_amplitude),&
                  !   !   COS(t_norm/almo_scf_env%envelope_amplitude),&
                  !   !   almo_scf_env%envelope_amplitude*SIN(t_norm/almo_scf_env%envelope_amplitude)
                  !   WRITE(unit_nr,*) "VAR1: TANH(|T|/A), D(A*TANH(|T|/A)), A*TANH(|T|/A): ",&
                  !      TANH(t_norm/almo_scf_env%envelope_amplitude),&
                  !      1.0_dp-(TANH(t_norm/almo_scf_env%envelope_amplitude))**2,&
                  !      almo_scf_env%envelope_amplitude*TANH(t_norm/almo_scf_env%envelope_amplitude)
                  !ENDIF
               ENDIF
               IF (optimize_theta) THEN
                  CALL cp_dbcsr_copy(m_tmp_no_1, m_theta)
                  CALL cp_dbcsr_function_of_elements(m_tmp_no_1, &
                                                     !func=dbcsr_func_sin,&
                                                     func=dbcsr_func_tanh, &
                                                     a0=0.0_dp, &
                                                     a1=1.0_dp/almo_scf_env%envelope_amplitude)
                  CALL cp_dbcsr_scale(m_tmp_no_1, &
                                      almo_scf_env%envelope_amplitude)
               ELSE
                  CALL cp_dbcsr_copy(m_tmp_no_1, m_theta)
               ENDIF
               CALL cp_dbcsr_hadamard_product(m_tmp_no_1, &
                                              quench_t(ispin), &
                                              matrix_t_out(ispin))
               CALL cp_dbcsr_norm(matrix_t_out(ispin), &
                                  dbcsr_norm_maxabsnorm, norm_scalar=t_norm)

               !IF (optimize_theta) THEN
               !   IF (unit_nr>0) THEN
               !      WRITE(unit_nr,*) "VAR1: |B*A*TANH(T/A)|: ", t_norm
               !   ENDIF
               !ELSE
               !   IF (unit_nr>0) THEN
               !      WRITE(unit_nr,*) "VAR1: |B*X|: ", t_norm
               !   ENDIF
               !ENDIF

               ! project out R_0
               IF (assume_t0_q0x .AND. (.NOT. just_started)) THEN
                  IF (my_special_case .EQ. xalmo_case_fully_deloc) THEN
                     CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                            STsiginv_0, &
                                            matrix_t_out(ispin), &
                                            0.0_dp, m_tmp_oo_1, &
                                            filter_eps=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                                            matrix_t_0, &
                                            m_tmp_oo_1, &
                                            1.0_dp, matrix_t_out(ispin), &
                                            filter_eps=almo_scf_env%eps_filter)
                  ELSE IF (my_special_case .EQ. xalmo_case_block_diag) THEN
                     ! cannot use projector with block-daigonal ALMOs
                     CALL cp__b("almo_scf_optimizer.F",1104,"")
                  ELSE
                     ! no special case
                     CALL apply_domain_operators( &
                        matrix_in=matrix_t_out(ispin), &
                        matrix_out=m_tmp_no_1, &
                        operator1=domain_r_down(:), &
                        operator2=almo_scf_env%domain_s_inv(:, ispin), &
                        dpattern=quench_t(ispin), &
                        map=almo_scf_env%domain_map(ispin), &
                        node_of_domain=almo_scf_env%cpu_of_domain, &
                        my_action=1, &
                        filter_eps=almo_scf_env%eps_filter, &
                        !matrix_trimmer=,&
                        use_trimmer=.FALSE.)
                     CALL cp_dbcsr_copy(matrix_t_out(ispin), &
                                        m_tmp_no_1)
                  ENDIF ! special case
                  CALL cp_dbcsr_norm(matrix_t_out(ispin), &
                                     dbcsr_norm_maxabsnorm, norm_scalar=t_norm)
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "VAR1: |Q.X|: ", t_norm
                  !ENDIF
               ENDIF ! end assume_t0_q0x

               IF (assume_t0_q0x) THEN
                  CALL cp_dbcsr_add(matrix_t_out(ispin), &
                                    matrix_t_0, 1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_norm(matrix_t_out(ispin), &
                                     dbcsr_norm_maxabsnorm, norm_scalar=t_norm)
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "VAR1: |T0+Q.X|: ", t_norm
                  !ENDIF
               ENDIF

               !! RZK-warning simple orthogonalization for block-diagonal ALMOs
               !CALL orthogonalize_mos(ket=matrix_t_out(ispin),&
               !        overlap=almo_scf_env%matrix_sigma_blk(ispin),&
               !        metric=almo_scf_env%matrix_s(1),&
               !        retain_locality=.TRUE.,&
               !        only_normalize=.TRUE.,&
               !        eps_filter=almo_scf_env%eps_filter,&
               !        order_lanczos=almo_scf_env%order_lanczos,&
               !        eps_lanczos=almo_scf_env%eps_lanczos,&
               !        max_iter_lanczos=almo_scf_env%max_iter_lanczos)

               CALL cp_dbcsr_filter(matrix_t_out(ispin), &
                                    eps=almo_scf_env%eps_filter)

               ! compute the density matrix
               CALL almo_scf_t_to_p( &
                  t=matrix_t_out(ispin), &
                  p=almo_scf_env%matrix_p(ispin), &
                  eps_filter=almo_scf_env%eps_filter, &
                  orthog_orbs=.FALSE., &
                  s=almo_scf_env%matrix_s(1), &
                  sigma=almo_scf_env%matrix_sigma(ispin), &
                  sigma_inv=almo_scf_env%matrix_sigma_inv(ispin), &
                  use_guess=use_guess)
               CALL cp_dbcsr_scale(almo_scf_env%matrix_p(ispin), &
                                   spin_factor)

               !! RZK-warning to debug lets see eigenvalues of the ALMO overlap
               !CALL cp_dbcsr_copy(almo_scf_env%matrix_sigma_blk(ispin),&
               !               almo_scf_env%matrix_sigma(ispin),&
               !               keep_sparsity=.TRUE.)
               !CALL cp_dbcsr_init(u_sigma)
               !CALL cp_dbcsr_create(u_sigma,template=almo_scf_env%matrix_sigma(ispin),&
               !        matrix_type=dbcsr_type_no_symmetry)
               !CALL cp_dbcsr_get_info(almo_scf_env%matrix_sigma_blk(ispin), nfullrows_total=occ1 )
               !ALLOCATE(evals(occ1))
               !CALL cp_dbcsr_syevd(almo_scf_env%matrix_sigma_blk(ispin),u_sigma,evals,&
               !        almo_scf_env%para_env,almo_scf_env%blacs_env)
               !WRITE(*,*) 'SIGMA_BLCK: ', evals
               !DEALLOCATE(evals)
               !CALL cp_dbcsr_create(u_sigma,template=almo_scf_env%matrix_sigma(ispin),&
               !        matrix_type=dbcsr_type_no_symmetry)
               !CALL cp_dbcsr_get_info(almo_scf_env%matrix_sigma(ispin), nfullrows_total=occ1 )
               !ALLOCATE(evals(occ1))
               !CALL cp_dbcsr_get_diag(almo_scf_env%matrix_sigma(ispin),evals)
               !WRITE(*,*) 'TRACE: ', SUM(evals)
               !CALL cp_dbcsr_syevd(almo_scf_env%matrix_sigma(ispin),u_sigma,evals,&
               !        almo_scf_env%para_env,almo_scf_env%blacs_env)
               !WRITE(*,*) 'SIGMA_FULL: ', evals
               !WRITE(*,*) 'SUMEI: ', SUM(evals)
               !DEALLOCATE(evals)
               !CALL cp_dbcsr_release(u_sigma)

               ! update the KS matrix and energy if necessary
               IF (.NOT. (perturbation_only .AND. (.NOT. just_started))) THEN
                  !!! RZK-warning the KS matrix must be updated outside the spin loop
                  !!! Now the code works only for restricted orbitals
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "....updating KS matrix...."
                  !ENDIF
                  CALL almo_scf_dm_to_ks(qs_env, almo_scf_env, energy_new)
                  CALL get_qs_env(qs_env, matrix_ks=matrix_ks)
                  CALL matrix_qs_to_almo(matrix_ks(ispin)%matrix, &
                                         almo_scf_env%matrix_ks(ispin), &
                                         almo_scf_env, .FALSE.)
                  CALL cp_dbcsr_filter(almo_scf_env%matrix_ks(ispin), &
                                       almo_scf_env%eps_filter)
               ENDIF

               ! save the initial state
               IF (just_started) THEN
                  CALL cp_dbcsr_copy(matrix_p_0, almo_scf_env%matrix_p(ispin))
                  CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma(ispin), &
                                             matrix_sigma_0)
                  CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma_inv(ispin), &
                                             matrix_sigma_inv_0)
               ENDIF

               IF (my_special_case .EQ. xalmo_case_normal .AND. prec_type .EQ. 4) THEN
                  ! construct domain-projector
                  IF (assume_t0_q0x .AND. just_started) THEN
                     CALL construct_domain_r_down( &
                        matrix_t=matrix_t_0, &
                        matrix_sigma_inv=matrix_sigma_inv_0, &
                        matrix_s=almo_scf_env%matrix_s(1), &
                        subm_r_down=domain_r_down(:), &
                        dpattern=quench_t(ispin), &
                        map=almo_scf_env%domain_map(ispin), &
                        node_of_domain=almo_scf_env%cpu_of_domain, &
                        filter_eps=almo_scf_env%eps_filter)
                     !CALL construct_domain_r_down(&
                     !        matrix_t=matrix_t_out(ispin),&
                     !        matrix_sigma_inv=almo_scf_env%matrix_sigma_inv(ispin),&
                     !        matrix_s=almo_scf_env%matrix_s(1),&
                     !        subm_r_down=domain_r_down(:),&
                     !        dpattern=quench_t(ispin),&
                     !        map=almo_scf_env%domain_map(ispin),&
                     !        node_of_domain=almo_scf_env%cpu_of_domain,&
                     !        filter_eps=almo_scf_env%eps_filter)
                  ENDIF ! assume_t0_q0x
               ENDIF ! debug condition

               IF (perturbation_only) THEN
                  ! calculate objective function Tr(F_0 R)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                         almo_scf_env%matrix_ks(ispin), &
                                         matrix_t_out(ispin), &
                                         0.0_dp, m_tmp_no_1, &
                                         filter_eps=almo_scf_env%eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                         m_tmp_no_1, &
                                         almo_scf_env%matrix_sigma_inv(ispin), &
                                         0.0_dp, FTsiginv, &
                                         filter_eps=almo_scf_env%eps_filter)
                  CALL cp_dbcsr_trace(matrix_t_out(ispin), &
                                      FTsiginv, energy_new, "T", "N")
                  energy_new = energy_new*spin_factor
               ENDIF

               ! save the previous gradient to compute beta
               ! do it only if the previous grad was computed
               ! for .NOT.line_search
               IF (line_search_iteration .EQ. 0 .AND. iteration .NE. 0) &
                  CALL cp_dbcsr_copy(prev_grad, grad)

               ! compute the energy gradient if necessary
               skip_grad = (iteration .GT. 0 .AND. &
                            fixed_line_search_niter .NE. 0 .AND. &
                            line_search_iteration .NE. fixed_line_search_niter)

               IF (.NOT. skip_grad) THEN

                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "....computing gradient...."
                  !ENDIF

                  ! do d_E/d_theta first
                  IF (.NOT. perturbation_only) THEN
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            almo_scf_env%matrix_ks(ispin), &
                                            matrix_t_out(ispin), &
                                            0.0_dp, m_tmp_no_1, &
                                            filter_eps=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            m_tmp_no_1, &
                                            almo_scf_env%matrix_sigma_inv(ispin), &
                                            0.0_dp, FTsiginv, &
                                            filter_eps=almo_scf_env%eps_filter)
                     !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                     !        m_tmp_no_1,&
                     !        almo_scf_env%matrix_sigma_inv(ispin),&
                     !        0.0_dp,m_tmp_no_2,&
                     !        retain_sparsity=.TRUE.)
                  ENDIF
                  CALL cp_dbcsr_copy(m_tmp_no_2, quench_t(ispin))
                  CALL cp_dbcsr_copy(m_tmp_no_2, &
                                     FTsiginv, keep_sparsity=.TRUE.)
                  CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                         matrix_t_out(ispin), &
                                         FTsiginv, &
                                         0.0_dp, m_tmp_oo_1, &
                                         filter_eps=almo_scf_env%eps_filter)
                  !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                  !        m_tmp_oo_1,&
                  !        almo_scf_env%matrix_sigma_inv(ispin),&
                  !        0.0_dp,m_tmp_oo_2,&
                  !        filter_eps=almo_scf_env%eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                         almo_scf_env%matrix_sigma_inv(ispin), &
                                         m_tmp_oo_1, &
                                         0.0_dp, siginvTFTsiginv, &
                                         filter_eps=almo_scf_env%eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                         almo_scf_env%matrix_s(1), &
                                         matrix_t_out(ispin), &
                                         0.0_dp, ST, &
                                         filter_eps=almo_scf_env%eps_filter)
                  ! save S*T_0*siginv_0
                  IF (assume_t0_q0x .AND. just_started &
                      .AND. special_case .EQ. xalmo_case_fully_deloc) THEN
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            ST, &
                                            almo_scf_env%matrix_sigma_inv(ispin), &
                                            0.0_dp, STsiginv_0, &
                                            filter_eps=almo_scf_env%eps_filter)
                  ENDIF
                  CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                                         ST, &
                                         siginvTFTsiginv, &
                                         1.0_dp, m_tmp_no_2, &
                                         retain_sparsity=.TRUE.)
                  CALL cp_dbcsr_scale(m_tmp_no_2, &
                                      2.0_dp*spin_factor)
                  CALL cp_dbcsr_filter(m_tmp_no_2, &
                                       eps=almo_scf_env%eps_filter)

                  IF (perturbation_only .AND. just_started) THEN
                     ! save the first gradient
                     ! it is equal to F_vo and necessary to compute
                     ! the correction to the energy
                     CALL cp_dbcsr_copy(fvo_0, m_tmp_no_2)
                     CALL cp_dbcsr_scale(fvo_0, &
                                         0.5_dp)
                  ENDIF

                  ! a short print-out
                  CALL cp_dbcsr_norm(m_tmp_no_2, &
                                     dbcsr_norm_maxabsnorm, norm_scalar=t_norm)
                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*) "Maximum norm of dE/dT: ", t_norm
                  !ENDIF

                  ! project out the occupied space from the gradient
                  IF (assume_t0_q0x) THEN
                     IF (my_special_case .EQ. xalmo_case_fully_deloc) THEN
                        CALL cp_dbcsr_copy(grad, m_tmp_no_2)
                        CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                                               matrix_t_0, &
                                               grad, &
                                               0.0_dp, m_tmp_oo_1, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                                               STsiginv_0, &
                                               m_tmp_oo_1, &
                                               1.0_dp, grad, &
                                               filter_eps=almo_scf_env%eps_filter)
                     ELSE IF (my_special_case .EQ. xalmo_case_block_diag) THEN
                        ! should not be here - cannot project the zero-order space from itself
                        CALL cp__b("almo_scf_optimizer.F",1367,"")
                     ELSE
                        ! no special case: normal xALMOs
                        CALL apply_domain_operators( &
                           matrix_in=m_tmp_no_2, &
                           matrix_out=grad, &
                           operator2=domain_r_down(:), &
                           operator1=almo_scf_env%domain_s_inv(:, ispin), &
                           dpattern=quench_t(ispin), &
                           map=almo_scf_env%domain_map(ispin), &
                           node_of_domain=almo_scf_env%cpu_of_domain, &
                           my_action=1, &
                           filter_eps=almo_scf_env%eps_filter, &
                           !matrix_trimmer=,&
                           use_trimmer=.FALSE.)
                     ENDIF ! my_special_case
                     CALL cp_dbcsr_copy(m_tmp_no_2, grad)
                  ENDIF

                  ! transform d_E/d_T to d_E/d_theta
                  IF (optimize_theta) THEN
                     CALL cp_dbcsr_copy(m_tmp_no_1, m_theta)
                     CALL cp_dbcsr_function_of_elements(m_tmp_no_1, &
                                                        !func=dbcsr_func_cos,&
                                                        func=dbcsr_func_dtanh, &
                                                        a0=0.0_dp, &
                                                        a1=1.0_dp/almo_scf_env%envelope_amplitude)
                     CALL cp_dbcsr_scale(m_tmp_no_1, &
                                         almo_scf_env%envelope_amplitude)
                     CALL cp_dbcsr_set(m_tmp_no_3, 0.0_dp)
                     CALL cp_dbcsr_filter(m_tmp_no_3, &
                                          eps=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_hadamard_product(m_tmp_no_2, &
                                                    m_tmp_no_1, &
                                                    m_tmp_no_3, &
                                                    b_assume_value=1.0_dp)
                     CALL cp_dbcsr_hadamard_product(m_tmp_no_3, &
                                                    quench_t(ispin), &
                                                    grad)
                  ELSE ! simply copy
                     CALL cp_dbcsr_hadamard_product(m_tmp_no_2, &
                                                    quench_t(ispin), &
                                                    grad)
                  ENDIF
                  CALL cp_dbcsr_filter(grad, eps=almo_scf_env%eps_filter)

               ENDIF ! skip_grad

               ! check convergence and other exit criteria
               grad_norm_frob = cp_dbcsr_frobenius_norm(grad)
               CALL cp_dbcsr_norm(grad, dbcsr_norm_maxabsnorm, &
                                  norm_scalar=grad_norm)
               converged = (grad_norm .LT. optimizer%eps_error)
               IF (converged .OR. (iteration .GE. max_iter)) THEN
                  prepare_to_exit = .TRUE.
               ENDIF
               IF (grad_norm .LT. almo_scf_env%eps_prev_guess) THEN
                  use_guess = .TRUE.
               ENDIF

               IF (md_in_theta_space) THEN

                  IF (.NOT. first_md_iteration) THEN
                     CALL cp_dbcsr_copy(prev_step, step)
                  ENDIF
                  CALL cp_dbcsr_copy(step, grad)
                  CALL cp_dbcsr_scale(step, -1.0_dp)

                  ! update velocities v(i) = v(i-1) + 0.5*dT*(a(i-1) + a(i))
                  IF (.NOT. first_md_iteration) THEN
                     CALL cp_dbcsr_add(velocity, &
                                       step, 1.0_dp, 0.5_dp*time_step)
                     CALL cp_dbcsr_add(velocity, &
                                       prev_step, 1.0_dp, 0.5_dp*time_step)
                  ENDIF
                  kin_energy = cp_dbcsr_frobenius_norm(velocity)
                  kin_energy = 0.5_dp*kin_energy*kin_energy

                  ! update positions theta(i) = theta(i-1) + dT*v(i-1) + 0.5*dT*dT*a(i-1)
                  CALL cp_dbcsr_add(m_theta, &
                                    velocity, 1.0_dp, time_step)
                  CALL cp_dbcsr_add(m_theta, &
                                    step, 1.0_dp, 0.5_dp*time_step*time_step)

                  iter_type = "MD"

                  t2 = m_walltime()
                  IF (unit_nr > 0) THEN
                     WRITE (unit_nr, '(T2,A,A2,I5,F16.7,F17.9,F17.9,F17.9,E12.3,F10.3)') &
                        "ALMO SCF ", iter_type, iteration, time_step*iteration, &
                        energy_new, kin_energy, energy_new+kin_energy, grad_norm, &
                        t2-t1
                  ENDIF
                  t1 = m_walltime()

                  IF (first_md_iteration) THEN
                     first_md_iteration = .FALSE.
                  ENDIF

               ELSE ! optimizization (not MD)

                  IF (.NOT. prepare_to_exit) THEN

                     ! check the gradient along the step direction
                     IF (iteration .NE. 0) THEN

                        IF (fixed_line_search_niter .EQ. 0) THEN

                           CALL cp_dbcsr_trace(grad, step, line_search_error, &
                                               "T", "N")
                           ! normalize the result
                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,*) "Angle between step/grad: ", line_search_error
                           !ENDIF
                           CALL cp_dbcsr_trace(grad, grad, denom, "T", "N")
                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,*) "Frobenius norm of grad:  ", SQRT(denom)
                           !ENDIF
                           line_search_error = line_search_error/SQRT(denom)
                           CALL cp_dbcsr_trace(step, step, denom, "T", "N")
                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,*) "Frobenius norm of step:  ", SQRT(denom)
                           !ENDIF
                           line_search_error = line_search_error/SQRT(denom)
                           IF (ABS(line_search_error) .GT. optimizer%lin_search_eps_error) THEN
                              line_search = .TRUE.
                              line_search_iteration = line_search_iteration+1
                           ELSE
                              line_search = .FALSE.
                              line_search_iteration = 0
                              IF (grad_norm .LT. eps_skip_gradients) THEN
                                 fixed_line_search_niter = ABS(almo_scf_env%integer04)
                              ENDIF
                           ENDIF

                        ELSE ! decision for fixed_line_search_niter

                           IF (.NOT. line_search) THEN
                              line_search = .TRUE.
                              line_search_iteration = line_search_iteration+1
                           ELSE
                              IF (line_search_iteration .EQ. fixed_line_search_niter) THEN
                                 line_search = .FALSE.
                                 line_search_iteration = 0
                                 line_search_iteration = line_search_iteration+1
                              ENDIF
                           ENDIF

                        ENDIF ! fixed_line_search_niter fork
                     ENDIF

                     IF (line_search) THEN
                        energy_diff = 0.0_dp
                     ELSE
                        energy_diff = energy_new-energy_old
                        energy_old = energy_new
                     ENDIF

                     ! update the step direction
                     IF (.NOT. line_search) THEN

                        !IF (unit_nr>0) THEN
                        !   WRITE(unit_nr,*) "....updating step direction...."
                        !ENDIF

                        cg_iteration = cg_iteration+1

                        IF ((just_started .AND. perturbation_only) .OR. &
                            (iteration .EQ. 0 .AND. (.NOT. perturbation_only))) THEN

                           ! compute the preconditioner
                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,*) "....computing preconditioner...."
                           !ENDIF

                           ! calculate (1-R)F(1-R) and S-SRS
                           ! RZK-warning take advantage: some elements will be removed by the quencher
                           ! RZK-warning S operations can be performed outside the spin loop to save time
                           ! IT IS REQUIRED THAT PRECONDITIONER DOES NOT BREAK THE LOCALITY!!!!
                           ! RZK-warning: further optimization is ABSOLUTELY NECESSARY

                           ! First S-SRS
                           !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                           !        almo_scf_env%matrix_s(1),&
                           !        matrix_t_out(ispin),&
                           !        0.0_dp,m_tmp_no_1,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                                  ST, &
                                                  almo_scf_env%matrix_sigma_inv(ispin), &
                                                  0.0_dp, m_tmp_no_3, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_create(m_tmp_nn_1, &
                                                template=almo_scf_env%matrix_s(1), &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_s(1), &
                                                      m_tmp_nn_1)
                           IF (my_special_case .EQ. xalmo_case_fully_deloc) THEN
                              ! use S instead of S-SRS
                           ELSE
                              CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                                                     ST, &
                                                     m_tmp_no_3, &
                                                     1.0_dp, m_tmp_nn_1, &
                                                     filter_eps=almo_scf_env%eps_filter)
                           ENDIF

                           ! Second (1-R)F(1-R)
                           !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                           !        almo_scf_env%matrix_ks(ispin),&
                           !        matrix_t_out(ispin),&
                           !        0.0_dp,m_tmp_no_1,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           ! re-create matrix because desymmetrize is buggy -
                           ! it will create multiple copies of blocks
                           CALL cp_dbcsr_create(prec_vv, &
                                                template=almo_scf_env%matrix_ks(ispin), &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_ks(ispin), &
                                                      prec_vv)
                           CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                                                  FTsiginv, &
                                                  ST, &
                                                  1.0_dp, prec_vv, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_multiply("N", "T", -1.0_dp, &
                                                  ST, &
                                                  FTsiginv, &
                                                  1.0_dp, prec_vv, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_multiply("T","N",1.0_dp,&
                           !        matrix_t_out(ispin),&
                           !        m_tmp_no_1,&
                           !        0.0_dp,m_tmp_oo_1,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                                  ST, &
                                                  siginvTFTsiginv, &
                                                  0.0_dp, m_tmp_no_3, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                                  m_tmp_no_3, &
                                                  ST, &
                                                  1.0_dp, prec_vv, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_add(prec_vv, m_tmp_nn_1, &
                                             1.0_dp-prec_sf_mixing_s, &
                                             prec_sf_mixing_s)
                           CALL cp_dbcsr_scale(prec_vv, 2.0_dp*spin_factor)
                           CALL cp_dbcsr_copy(m_tmp_nn_1, prec_vv)

                           ! invert using various algorithms
                           IF (my_special_case .EQ. xalmo_case_block_diag) THEN ! non-overlapping diagonal blocks

                              !precond_domain_projector=0
                              !CALL construct_domain_preconditioner(&
                              !   matrix_main=m_tmp_nn_1,&
                              !   dpattern=quench_t(ispin),&
                              !   map=almo_scf_env%domain_map(ispin),&
                              !   node_of_domain=almo_scf_env%cpu_of_domain,&
                              !   preconditioner=almo_scf_env%domain_preconditioner(:,ispin),&
                              !   use_trimmer=.FALSE.,&
                              !   my_action=precond_domain_projector)
                              CALL pseudo_invert_diagonal_blk(matrix_in=m_tmp_nn_1, &
                                                              matrix_out=prec_vv, &
                                                              nocc=almo_scf_env%nocc_of_domain(:, ispin))

                           ELSE IF (my_special_case .EQ. xalmo_case_fully_deloc) THEN ! the entire system is a block

                              ! invert using cholesky (works with S matrix, will not work with S-SRS matrix)
                              CALL cp_dbcsr_cholesky_decompose(prec_vv, &
                                                               para_env=almo_scf_env%para_env, &
                                                               blacs_env=almo_scf_env%blacs_env)
                              CALL cp_dbcsr_cholesky_invert(prec_vv, &
                                                            para_env=almo_scf_env%para_env, &
                                                            blacs_env=almo_scf_env%blacs_env, &
                                                            upper_to_full=.TRUE.)
                              CALL cp_dbcsr_filter(prec_vv, &
                                                   eps=almo_scf_env%eps_filter)
                           ELSE
                              !!! use a sophisticated domain preconditioner
                              IF (assume_t0_q0x) THEN
                                 precond_domain_projector = -1
                              ELSE
                                 precond_domain_projector = 0
                              ENDIF
                              ! for other experimental preconditioner types the inversion is
                              ! done together with applying the preconditioner
                              IF (prec_type .EQ. 4) THEN

                                 CALL construct_domain_preconditioner( &
                                    matrix_main=m_tmp_nn_1, &
                                    subm_s_inv=almo_scf_env%domain_s_inv(:, ispin), &
                                    subm_r_down=domain_r_down(:), &
                                    matrix_trimmer=quench_t(ispin), &
                                    dpattern=quench_t(ispin), &
                                    map=almo_scf_env%domain_map(ispin), &
                                    node_of_domain=almo_scf_env%cpu_of_domain, &
                                    preconditioner=almo_scf_env%domain_preconditioner(:, ispin), &
                                    use_trimmer=.FALSE., &
                                    my_action=precond_domain_projector)
                              ENDIF ! prec type
                           ENDIF

                           ! invert using cholesky (works with S matrix, will not work with S-SRS matrix)
                           !!!CALL cp_dbcsr_cholesky_decompose(prec_vv,&
                           !!!        para_env=almo_scf_env%para_env,&
                           !!!        blacs_env=almo_scf_env%blacs_env)
                           !!!CALL cp_dbcsr_cholesky_invert(prec_vv,&
                           !!!        para_env=almo_scf_env%para_env,&
                           !!!        blacs_env=almo_scf_env%blacs_env,&
                           !!!        upper_to_full=.TRUE.)
                           !!!CALL cp_dbcsr_filter(prec_vv,&
                           !!!        eps=almo_scf_env%eps_filter)
                           !!!

                           ! re-create the matrix because desymmetrize is buggy -
                           ! it will create multiple copies of blocks
                           !!!DESYM!CALL cp_dbcsr_create(prec_vv,&
                           !!!DESYM!        template=almo_scf_env%matrix_s(1),&
                           !!!DESYM!        matrix_type=dbcsr_type_no_symmetry)
                           !!!DESYM!CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_s(1),&
                           !!!DESYM!        prec_vv)
                           !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                           !        almo_scf_env%matrix_s(1),&
                           !        matrix_t_out(ispin),&
                           !        0.0_dp,m_tmp_no_1,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                           !        m_tmp_no_1,&
                           !        almo_scf_env%matrix_sigma_inv(ispin),&
                           !        0.0_dp,m_tmp_no_3,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_multiply("N","T",-1.0_dp,&
                           !        m_tmp_no_3,&
                           !        m_tmp_no_1,&
                           !        1.0_dp,prec_vv,&
                           !        filter_eps=almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_add_on_diag(prec_vv,&
                           !        prec_sf_mixing_s)

                           !CALL cp_dbcsr_create(prec_oo,&
                           !        template=almo_scf_env%matrix_sigma(ispin),&
                           !        matrix_type=dbcsr_type_no_symmetry)
                           !CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma(ispin),&
                           !        matrix_type=dbcsr_type_no_symmetry)
                           !CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma(ispin),&
                           !        prec_oo)
                           !CALL cp_dbcsr_filter(prec_oo,&
                           !        eps=almo_scf_env%eps_filter)

                           !! invert using cholesky
                           !CALL cp_dbcsr_create(prec_oo_inv,&
                           !        template=prec_oo,&
                           !        matrix_type=dbcsr_type_no_symmetry)
                           !CALL cp_dbcsr_desymmetrize(prec_oo,&
                           !        prec_oo_inv)
                           !CALL cp_dbcsr_cholesky_decompose(prec_oo_inv,&
                           !        para_env=almo_scf_env%para_env,&
                           !        blacs_env=almo_scf_env%blacs_env)
                           !CALL cp_dbcsr_cholesky_invert(prec_oo_inv,&
                           !        para_env=almo_scf_env%para_env,&
                           !        blacs_env=almo_scf_env%blacs_env,&
                           !        upper_to_full=.TRUE.)

                        ENDIF

                        ! save the previous step
                        CALL cp_dbcsr_copy(prev_step, step)

                        ! compute the new step (apply preconditioner if available)
                        IF (use_preconditioner) THEN

                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,*) "....applying preconditioner...."
                           !ENDIF

                           SELECT CASE (prec_type)
                           CASE (1)
                              ! expensive Newton-Raphson step (the Hessian is still approximate)
                              ! RZK-warning THIS PREC HAS NOT BEEN IMPLEMENTED FOR THETA
                              IF (ncores .GT. 1) THEN
                                 CALL cp__b("almo_scf_optimizer.F",1749,"serial code only")
                              ENDIF
                              CALL newton_grad_to_step( &
                                 matrix_grad=m_tmp_no_2, &
                                 matrix_step=m_tmp_no_1, &
                                 matrix_s=almo_scf_env%matrix_s(1), &
                                 matrix_ks=almo_scf_env%matrix_ks(ispin), &
                                 matrix_t=matrix_t_out(ispin), &
                                 matrix_sigma_inv=almo_scf_env%matrix_sigma_inv(ispin), &
                                 !matrix_ks=matrix_ks_0,&
                                 !matrix_t=matrix_t_0,&
                                 !matrix_sigma_inv=matrix_sigma_inv_0,&
                                 quench_t=quench_t(ispin), &
                                 spin_factor=spin_factor, &
                                 eps_filter=almo_scf_env%eps_filter)

                           CASE (3)

                              ! RZK-warning THIS PREC HAS NOT BEEN IMPLEMENTED FOR THETA

                              ! inversion
                              CALL cp_dbcsr_get_info(m_tmp_nn_1, nfullrows_total=dim0)
                              ALLOCATE (evals(dim0))
                              CALL cp_dbcsr_syevd(m_tmp_nn_1, prec_vv, evals, &
                                                  almo_scf_env%para_env, almo_scf_env%blacs_env)
                              ! invert eigenvalues and use eigenvectors to compute the Hessian inverse
                              ! take special care of zero eigenvalues
                              zero_neg_eiv = 0
                              CALL cp_dbcsr_get_info(almo_scf_env%matrix_sigma(ispin), nfullrows_total=occ1)
                              DO jj = 1, dim0
                                 IF (jj .LE. occ1) THEN
                                    evals(jj) = evals(jj)*0.0_dp
                                    zero_neg_eiv = zero_neg_eiv+1
                                 ELSE
                                    evals(jj) = 1.0_dp/evals(jj)
                                 ENDIF
                              ENDDO
                              IF (unit_nr > 0) THEN
                                 WRITE (*, *) 'ZERO OR NEGATIVE EIGENVALUES: ', zero_neg_eiv
                              ENDIF
                              CALL cp_dbcsr_init(inv_eiv)
                              CALL cp_dbcsr_create(inv_eiv, &
                                                   template=m_tmp_nn_1, &
                                                   matrix_type=dbcsr_type_no_symmetry)
                              CALL cp_dbcsr_add_on_diag(inv_eiv, 1.0_dp)
                              CALL cp_dbcsr_set_diag(inv_eiv, evals)
                              CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                                     prec_vv, &
                                                     inv_eiv, &
                                                     0.0_dp, m_tmp_nn_1, &
                                                     filter_eps=almo_scf_env%eps_filter)
                              CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                                     m_tmp_nn_1, &
                                                     prec_vv, &
                                                     0.0_dp, inv_eiv, &
                                                     filter_eps=almo_scf_env%eps_filter)
                              CALL cp_dbcsr_copy(prec_vv, inv_eiv)
                              CALL cp_dbcsr_release(inv_eiv)
                              DEALLOCATE (evals)

                              !!CALL cp_dbcsr_copy(step,&
                              !!        quench_t(ispin))
                              !!
                              !!CALL cp_dbcsr_multiply("N","N",1.0_dp,&
                              !!        m_tmp_no_2,&
                              !!        !grad,& - this choice is worse
                              !!        prec_oo,&
                              !!        0.0_dp,step,&
                              !!        !retain_sparsity=.TRUE.,&
                              !!        filter_eps=almo_scf_env%eps_filter)
                              !!
                              CALL cp_dbcsr_copy(m_tmp_no_1, &
                                                 quench_t(ispin))
                              !!CALL cp_dbcsr_hadamard_product(&
                              !!        quench_t(ispin),&
                              !!        step,&
                              !!        m_tmp_no_1)
                              !!
                              CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                                                     prec_vv, &
                                                     m_tmp_no_2, &
                                                     0.0_dp, m_tmp_no_1, &
                                                     retain_sparsity=.TRUE.)

                           CASE (4)

                              IF (my_special_case .EQ. xalmo_case_block_diag .OR. &
                                  my_special_case .EQ. xalmo_case_fully_deloc) THEN

                                 CALL cp_dbcsr_multiply("N", "N", -1.0_dp, &
                                                        prec_vv, &
                                                        grad, &
                                                        0.0_dp, step, &
                                                        filter_eps=almo_scf_env%eps_filter)

                              ELSE

                                 !!! RZK-warning Currently for non-theta only
                                 IF (optimize_theta) THEN
                                    CALL cp__b("almo_scf_optimizer.F",1848,"theta is NYI")
                                 ENDIF

                                 CALL apply_domain_operators( &
                                    matrix_in=grad, &
                                    matrix_out=step, &
                                    operator1=almo_scf_env%domain_preconditioner(:, ispin), &
                                    !operator2=,&
                                    dpattern=quench_t(ispin), &
                                    map=almo_scf_env%domain_map(ispin), &
                                    node_of_domain=almo_scf_env%cpu_of_domain, &
                                    my_action=0, &
                                    filter_eps=almo_scf_env%eps_filter)
                                 !matrix_trimmer=,&
                                 !use_trimmer=.FALSE.,&
                                 CALL cp_dbcsr_scale(step, -1.0_dp)

                                 CALL cp_dbcsr_copy(m_tmp_no_3, &
                                                    quench_t(ispin))
                                 CALL cp_dbcsr_function_of_elements(m_tmp_no_3, &
                                                                    func=dbcsr_func_inverse, &
                                                                    a0=0.0_dp, &
                                                                    a1=1.0_dp)
                                 CALL cp_dbcsr_copy(m_tmp_no_2, step)
                                 CALL cp_dbcsr_hadamard_product( &
                                    m_tmp_no_2, &
                                    m_tmp_no_3, &
                                    step)
                                 CALL cp_dbcsr_copy(m_tmp_no_3, quench_t(ispin))

                                 !CALL cp_dbcsr_create(m_tmp_oo_1,&
                                 !        template=almo_scf_env%matrix_sigma_blk(ispin),&
                                 !        matrix_type=dbcsr_type_no_symmetry)
                                 !CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma_blk(ispin),m_tmp_oo_1)
                                 !CALL get_overlap(bra=matrix_t_out(ispin),&
                                 !        ket=step,&
                                 !        overlap=m_tmp_oo_1,&
                                 !        metric=almo_scf_env%matrix_s(1),&
                                 !        retain_overlap_sparsity=.TRUE.,&
                                 !        eps_filter=almo_scf_env%eps_filter)
                                 !CALL cp_dbcsr_norm(m_tmp_oo_1,&
                                 !        dbcsr_norm_maxabsnorm, norm_scalar=t_norm)
                                 !IF (unit_nr>0) THEN
                                 !   WRITE(unit_nr,*) "Step block-orthogonality error: ", t_norm
                                 !ENDIF

                              ENDIF ! special case

                           END SELECT ! preconditioner type fork

                        ELSE

                           !!!! NO PRECONDITIONER
                           CALL cp_dbcsr_copy(step, grad)
                           CALL cp_dbcsr_scale(step, -1.0_dp)

                        ENDIF

                        ! check whether we need to reset conjugate directions
                        IF (iteration .EQ. 0) THEN
                           reset_conjugator = .TRUE.
                        ENDIF

                        ! compute the conjugation coefficient - beta
                        IF (.NOT. reset_conjugator) THEN

                           SELECT CASE (optimizer%conjugator)
                           CASE (cg_hestenes_stiefel)
                              CALL cp_dbcsr_copy(m_tmp_no_1, grad)
                              CALL cp_dbcsr_add(m_tmp_no_1, prev_grad, &
                                                1.0_dp, -1.0_dp)
                              CALL cp_dbcsr_trace(m_tmp_no_1, step, numer, &
                                                  "T", "N")
                              CALL cp_dbcsr_trace(m_tmp_no_1, prev_step, denom, &
                                                  "T", "N")
                              beta = -1.0_dp*numer/denom
                           CASE (cg_fletcher_reeves)
                              CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                              CALL cp_dbcsr_trace(prev_grad, prev_minus_prec_grad, denom, "T", "N")
                              beta = numer/denom
                           CASE (cg_polak_ribiere)
                              CALL cp_dbcsr_trace(prev_grad, prev_minus_prec_grad, denom, "T", "N")
                              CALL cp_dbcsr_copy(m_tmp_no_1, grad)
                              CALL cp_dbcsr_add(m_tmp_no_1, prev_grad, 1.0_dp, -1.0_dp)
                              CALL cp_dbcsr_trace(m_tmp_no_1, step, numer, "T", "N")
                              beta = numer/denom
                           CASE (cg_fletcher)
                              CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                              CALL cp_dbcsr_trace(prev_grad, prev_step, denom, "T", "N")
                              beta = numer/denom
                           CASE (cg_liu_storey)
                              CALL cp_dbcsr_trace(prev_grad, prev_step, denom, "T", "N")
                              CALL cp_dbcsr_copy(m_tmp_no_1, grad)
                              CALL cp_dbcsr_add(m_tmp_no_1, prev_grad, 1.0_dp, -1.0_dp)
                              CALL cp_dbcsr_trace(m_tmp_no_1, step, numer, "T", "N")
                              beta = numer/denom
                           CASE (cg_dai_yuan)
                              CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                              CALL cp_dbcsr_copy(m_tmp_no_1, grad)
                              CALL cp_dbcsr_add(m_tmp_no_1, prev_grad, 1.0_dp, -1.0_dp)
                              CALL cp_dbcsr_trace(m_tmp_no_1, prev_step, denom, "T", "N")
                              beta = -1.0_dp*numer/denom
                           CASE (cg_hager_zhang)
                              CALL cp_dbcsr_copy(m_tmp_no_1, grad)
                              CALL cp_dbcsr_add(m_tmp_no_1, prev_grad, 1.0_dp, -1.0_dp)
                              CALL cp_dbcsr_trace(m_tmp_no_1, prev_step, denom, "T", "N")
                              CALL cp_dbcsr_trace(m_tmp_no_1, prev_minus_prec_grad, numer, "T", "N")
                              kappa = -2.0_dp*numer/denom
                              CALL cp_dbcsr_trace(m_tmp_no_1, step, numer, "T", "N")
                              tau = -1.0_dp*numer/denom
                              CALL cp_dbcsr_trace(prev_step, grad, numer, "T", "N")
                              beta = tau-kappa*numer/denom
                           CASE (cg_zero)
                              beta = 0.0_dp
                           CASE DEFAULT
                              CALL cp__b("almo_scf_optimizer.F",1963,"illegal conjugator")
                           END SELECT

                           IF (beta .LT. 0.0_dp) THEN
                              IF (unit_nr > 0) THEN
                                 WRITE (unit_nr, *) "Beta is negative: ", beta
                              ENDIF
                              reset_conjugator = .TRUE.
                           ENDIF

                        ENDIF

                        IF (reset_conjugator) THEN

                           beta = 0.0_dp
                           IF (unit_nr > 0 .AND. (.NOT. just_started)) THEN
                              WRITE (unit_nr, *) "(Re)-setting conjugator to zero"
                           ENDIF
                           reset_conjugator = .FALSE.

                        ENDIF

                        ! save the preconditioned gradient (useful for beta)
                        CALL cp_dbcsr_copy(prev_minus_prec_grad, step)

                        !IF (unit_nr>0) THEN
                        !   WRITE(unit_nr,*) "....final beta....", beta
                        !ENDIF

                        ! conjugate the step direction
                        CALL cp_dbcsr_add(step, prev_step, 1.0_dp, beta)

                     ENDIF ! update the step direction

                     ! estimate the step size
                     IF (.NOT. line_search) THEN
                        e0 = energy_new
                        CALL cp_dbcsr_trace(grad, step, g0, "T", "N")
                        ! we just changed the direction and
                        ! we have only E and grad from the current step
                        ! it is not enouhg to compute step_size - just guess it
                        IF (iteration .EQ. 0) THEN
                           step_size = optimizer%lin_search_step_size_guess
                        ELSE
                           IF (next_step_size_guess .LE. 0.0_dp) THEN
                              step_size = optimizer%lin_search_step_size_guess
                           ELSE
                              ! take the last value
                              step_size = next_step_size_guess*1.05_dp
                           ENDIF
                        ENDIF
                        next_step_size_guess = step_size
                     ELSE
                        IF (fixed_line_search_niter .EQ. 0) THEN
                           e1 = energy_new
                           CALL cp_dbcsr_trace(grad, step, g1, "T", "N")
                           ! we have accumulated some points along this direction
                           ! use only the most recent g0 (quadratic approximation)
                           appr_sec_der = (g1-g0)/step_size
                           !IF (unit_nr>0) THEN
                           !   WRITE(unit_nr,'(A2,7F12.5)') &
                           !           "EG",e0,e1,g0,g1,appr_sec_der,step_size,-g1/appr_sec_der
                           !ENDIF
                           step_size = -g1/appr_sec_der
                           e0 = e1
                           g0 = g1
                        ELSE
                           ! use e0, g0 and e1 to compute g1 and make a step
                           ! if the next iteration is also line_search
                           ! use e1 and the calculated g1 as e0 and g0
                           e1 = energy_new
                           appr_sec_der = 2.0*((e1-e0)/step_size-g0)/step_size
                           g1 = appr_sec_der*step_size+g0
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, '(A2,7F12.5)') &
                                 "EG", e0, e1, g0, g1, appr_sec_der, step_size, -g1/appr_sec_der
                           ENDIF
                           !appr_sec_der=(g1-g0)/step_size
                           step_size = -g1/appr_sec_der
                           e0 = e1
                           g0 = g1
                        ENDIF
                        next_step_size_guess = next_step_size_guess+step_size
                     ENDIF

                     ! update theta
                     CALL cp_dbcsr_add(m_theta, step, 1.0_dp, step_size)

                  ENDIF ! not.prepare_to_exit

                  IF (line_search) THEN
                     iter_type = "LS"
                  ELSE
                     iter_type = "CG"
                  ENDIF

                  t2 = m_walltime()
                  IF (unit_nr > 0) THEN
                     iter_type = TRIM("ALMO SCF "//iter_type)
                     WRITE (unit_nr, '(T2,A13,I6,F23.10,E14.5,F14.9,F9.2)') &
                        iter_type, iteration, &
                        energy_new, energy_diff, grad_norm, &
                        t2-t1
                     !WRITE(unit_nr,'(T2,A11,I6,F20.12,E12.3,E12.3,E12.3,F12.5,F10.3)') &
                     !        "ALMO SCF ",iter_type,iteration,&
                     !        energy_new,energy_diff,grad_norm,line_search_error,&
                     !        step_size,t2-t1
                  ENDIF

                  IF (my_special_case .EQ. xalmo_case_block_diag) THEN
                     almo_scf_env%almo_scf_energy = energy_new
                  ENDIF

                  t1 = m_walltime()

               ENDIF ! MD in theta space

               iteration = iteration+1
               IF (prepare_to_exit) EXIT

            ENDDO ! inner SCF loop

            IF (converged .OR. (outer_iteration .GE. outer_max_iter)) THEN
               outer_prepare_to_exit = .TRUE.
            ENDIF

            outer_iteration = outer_iteration+1
            IF (outer_prepare_to_exit) EXIT

         ENDDO ! outer SCF loop

         ! post SCF-loop calculations
         IF (converged) THEN

            ! obtain MO coefficients from final theta
            ! RZK-warning: if decide to uncomment make sure that Theta->T
            ! procedure is consistent with the procedure in the loop
            !IF (optimize_theta) THEN
            !   CALL cp_dbcsr_copy(m_tmp_no_1,m_theta)
            !   CALL cp_dbcsr_function_of_elements(m_tmp_no_1,&
            !           func=dbcsr_func_tanh,&
            !           a0=0.0_dp,&
            !           a1=1.0_dp/almo_scf_env%envelope_amplitude)
            !   CALL cp_dbcsr_hadamard_product(m_tmp_no_1,&
            !           quench_t(ispin),&
            !           matrix_t_out(ispin))
            !   CALL cp_dbcsr_scale(matrix_t_out(ispin),&
            !           almo_scf_env%envelope_amplitude)
            !ELSE
            !   CALL cp_dbcsr_copy(m_tmp_no_1,m_theta)
            !   CALL cp_dbcsr_hadamard_product(m_tmp_no_1,&
            !           quench_t(ispin),&
            !           matrix_t_out(ispin))
            !ENDIF
            !IF (perturbation_only) THEN
            !   CALL cp_dbcsr_add(matrix_t_out(ispin),&
            !           matrix_t_0,1.0_dp,1.0_dp)
            !ENDIF
            !CALL cp_dbcsr_filter(matrix_t_out(ispin),&
            !        eps=almo_scf_env%eps_filter)
            !CALL cp_dbcsr_norm(matrix_t_out(ispin),&
            !        dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
            !IF (unit_nr>0) THEN
            !   WRITE(unit_nr,*) "Maximum norm of the ALMOs: ", grad_norm
            !ENDIF

            !!!!! experiment: bump final amplitudes and get T
            !!!!IF (.NOT.optimize_theta.AND.perturbation_only) THEN
            !!!!   CALL cp_dbcsr_copy(m_tmp_no_1,m_theta)
            !!!!   CALL cp_dbcsr_hadamard_product(m_tmp_no_1,&
            !!!!           quench_t_saved,&
            !!!!           matrix_t_out(ispin))
            !!!!   IF (use_projector) THEN
            !!!!      CALL almo_scf_domain_operations(&
            !!!!              matrix_in=matrix_t_out(ispin),&
            !!!!              matrix_out=m_tmp_no_1,&
            !!!!              quench_t=quench_t(ispin),&
            !!!!              my_action=2,&
            !!!!              matrix_s=almo_scf_env%matrix_s(1),&
            !!!!              matrix_sigma_inv=matrix_sigma_inv_0,&
            !!!!              matrix_t=matrix_t_0)
            !!!!
            !!!!      CALL cp_dbcsr_copy(matrix_t_out(ispin),&
            !!!!              m_tmp_no_1)
            !!!!   ENDIF
            !!!!   IF (perturbation_only) THEN
            !!!!      CALL cp_dbcsr_add(matrix_t_out(ispin),&
            !!!!              matrix_t_0,1.0_dp,1.0_dp)
            !!!!
            !!!!   ENDIF
            !!!!ENDIF
            !!!!
            !!!!IF (.NOT.optimize_theta) THEN
            !!!!   CALL cp_dbcsr_copy(quench_t(ispin),quench_t_saved)
            !!!!ENDIF

            IF (perturbation_only) THEN

               CALL cp_dbcsr_add(matrix_t_0, matrix_t_out(ispin), &
                                 -1.0_dp, 1.0_dp)

               CALL cp_dbcsr_trace(matrix_t_0, &
                                   fvo_0, energy_new, "T", "N")
               ! print out the energy lowering
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *)
                  WRITE (unit_nr, '(T2,A35,F25.10)') "ENERGY OF BLOCK-DIAGONAL ALMOs:", &
                     almo_scf_env%almo_scf_energy
                  WRITE (unit_nr, '(T2,A35,F25.10)') "ENERGY LOWERING:", &
                     energy_new
                  WRITE (unit_nr, '(T2,A35,F25.10)') "CORRECTED ENERGY:", &
                     almo_scf_env%almo_scf_energy+energy_new
                  WRITE (unit_nr, *)
               ENDIF
               !IF (unit_nr>0) THEN
               !   WRITE(unit_nr,*) "_ENERGY-0: ", almo_scf_env%almo_scf_energy
               !   WRITE(unit_nr,*) "_ENERGY-D: ", energy_new
               !   WRITE(unit_nr,*) "_ENERGY-F: ", almo_scf_env%almo_scf_energy+energy_new
               !ENDIF
               CALL almo_scf_update_ks_energy(qs_env, &
                                              almo_scf_env%almo_scf_energy+energy_new)

               ! similar method to evaluate the energy correction
               !CALL cp_dbcsr_add(matrix_p_0,&
               !                 almo_scf_env%matrix_p(ispin),-1.0_dp,1.0_dp)
               !CALL cp_dbcsr_trace(almo_scf_env%matrix_ks_almo_scf_converged(ispin),&
               !                    matrix_p_0,&
               !                    energy_new)
               !IF (unit_nr>0) THEN
               !   WRITE(unit_nr,*) "alt-ENERGY-D: ", energy_new
               !ENDIF
               !CALL almo_scf_update_ks_energy(qs_env,&
               !        almo_scf_env%almo_scf_energy+energy_new)

               IF (almo_scf_env%almo_eda .GT. 0) THEN

                  ! print out the results of decomposition analysis
                  CALL cp_dbcsr_hadamard_product(matrix_t_0, &
                                                 fvo_0, m_tmp_no_1)
                  IF (unit_nr > 0) THEN
                     WRITE (unit_nr, *)
                     WRITE (unit_nr, '(T2,A)') "DECOMPOSITION OF THE DELOCALIZATION ENERGY"
                  ENDIF
                  CALL cp_dbcsr_filter(m_tmp_no_1, almo_scf_env%eps_filter)

                  mynode = dbcsr_mp_mynode(dbcsr_distribution_mp( &
                                           cp_dbcsr_distribution(m_tmp_no_1)))
                  WRITE (mynodestr, '(I6.6)') mynode
                  mylogfile = 'EDA.'//TRIM(ADJUSTL(mynodestr))
                  CALL open_file(file_name=mylogfile, file_status="REPLACE", unit_number=iunit)

                  CALL cp_dbcsr_print_block_sum(m_tmp_no_1, iunit)
                  CALL close_file(iunit)

                  !IF (unit_nr>0) THEN
                  !   WRITE(unit_nr,*)
                  !   WRITE(unit_nr,'(T2,A)') "CHARGE DECOMPOSITION"
                  !ENDIF
                  !CALL cp_dbcsr_print_block_sum(cta_matrix(ispin))

               ENDIF ! do ALMO EDA

            ELSE

               CALL almo_scf_update_ks_energy(qs_env, &
                                              energy_new)

            ENDIF ! if perturbation only

         ENDIF ! if converged

         IF (md_in_theta_space) THEN
            CALL cp_dbcsr_release(velocity)
         ENDIF
         CALL cp_dbcsr_release(m_theta)
         CALL cp_dbcsr_release(prec_vv)
         CALL cp_dbcsr_release(prec_oo)
         CALL cp_dbcsr_release(prec_oo_inv)
         CALL cp_dbcsr_release(m_tmp_no_1)
         CALL cp_dbcsr_release(fvo_0)
         CALL cp_dbcsr_release(STsiginv_0)
         CALL cp_dbcsr_release(m_tmp_no_2)
         CALL cp_dbcsr_release(m_tmp_no_3)
         CALL cp_dbcsr_release(m_tmp_oo_1)
         CALL cp_dbcsr_release(ST)
         CALL cp_dbcsr_release(FTsiginv)
         CALL cp_dbcsr_release(siginvTFTsiginv)
         CALL cp_dbcsr_release(m_tmp_nn_1)
         CALL cp_dbcsr_release(prev_grad)
         CALL cp_dbcsr_release(prev_step)
         CALL cp_dbcsr_release(grad)
         CALL cp_dbcsr_release(step)
         CALL cp_dbcsr_release(prev_minus_prec_grad)
         CALL cp_dbcsr_release(matrix_p_0)
         CALL cp_dbcsr_release(matrix_t_0)
         CALL cp_dbcsr_release(matrix_sigma_0)
         CALL cp_dbcsr_release(matrix_sigma_inv_0)

         IF (.NOT. converged) THEN
            CALL cp__b("almo_scf_optimizer.F",2262,"Optimization not converged! ")
         ENDIF

         DEALLOCATE (domain_r_down)

      ENDDO ! ispin

      CALL timestop(handle)

   END SUBROUTINE almo_scf_xalmo_pcg

! **************************************************************************************************
!> \brief Split the matrix of virtual orbitals into two:
!>        retained orbs and discarded
!> \param almo_scf_env ...
!> \par History
!>       2011.09 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE split_v_blk(almo_scf_env)

      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'split_v_blk', routineP = moduleN//':'//routineN

      INTEGER                                            :: discarded_v, handle, iblock_col, &
                                                            iblock_col_size, iblock_row, &
                                                            iblock_row_size, ispin, retained_v
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: data_p, p_new_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      DO ispin = 1, almo_scf_env%nspins

         CALL cp_dbcsr_work_create(almo_scf_env%matrix_v_blk(ispin), &
                                   work_mutable=.TRUE.)
         CALL cp_dbcsr_work_create(almo_scf_env%matrix_v_disc_blk(ispin), &
                                   work_mutable=.TRUE.)

         CALL cp_dbcsr_iterator_start(iter, almo_scf_env%matrix_v_full_blk(ispin))

         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

            CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, data_p, &
                                              row_size=iblock_row_size, col_size=iblock_col_size)

            IF (iblock_row .NE. iblock_col) THEN
               CALL cp__b("almo_scf_optimizer.F",2310,"off-diagonal block found")
            ENDIF

            retained_v = almo_scf_env%nvirt_of_domain(iblock_col, ispin)
            discarded_v = almo_scf_env%nvirt_disc_of_domain(iblock_col, ispin)
            IF(.NOT.(retained_v .GT. 0))CALL cp__a("almo_scf_optimizer.F",2315)
            IF(.NOT.(discarded_v .GT. 0))CALL cp__a("almo_scf_optimizer.F",2316)

            NULLIFY (p_new_block)
            CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_v_disc_blk(ispin), &
                                          iblock_row, iblock_col, p_new_block)
            IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_optimizer.F",2321)
            IF(.NOT.(retained_v+discarded_v .EQ. iblock_col_size))CALL cp__a("almo_scf_optimizer.F",2322)
            p_new_block(:, :) = data_p(:, (retained_v+1):iblock_col_size)

            NULLIFY (p_new_block)
            CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_v_blk(ispin), &
                                          iblock_row, iblock_col, p_new_block)
            IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf_optimizer.F",2328)
            p_new_block(:, :) = data_p(:, 1:retained_v)

         ENDDO ! iterator
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_finalize(almo_scf_env%matrix_v_blk(ispin))
         CALL cp_dbcsr_finalize(almo_scf_env%matrix_v_disc_blk(ispin))

      ENDDO ! ispin

      CALL timestop(handle)

   END SUBROUTINE split_v_blk

! **************************************************************************************************
!> \brief various methods for calculating the Harris-Foulkes correction
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE harris_foulkes_correction(almo_scf_env)

      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'harris_foulkes_correction', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: cayley_transform = 1, dm_ls_step = 2

      INTEGER :: algorithm_id, handle, handle1, handle2, handle3, handle4, handle5, handle6, &
         handle7, handle8, ispin, iteration, n, nmins, nspin, opt_k_max_iter, &
         outer_opt_k_iteration, outer_opt_k_max_iter, unit_nr
      INTEGER, DIMENSION(1)                              :: fake, nelectron_spin_real
      LOGICAL :: converged, line_search, md_in_k_space, outer_opt_k_prepare_to_exit, &
         prepare_to_exit, reset_conjugator, reset_step_size, use_cubic_approximation, &
         use_quadratic_approximation
      REAL(KIND=dp) :: aa, bb, beta, conjugacy_error, conjugacy_error_threshold, &
         delta_obj_function, denom, energy_correction_final, frob_matrix, frob_matrix_base, fun0, &
         fun1, gfun0, gfun1, grad_norm, grad_norm_frob, kappa, kin_energy, line_search_error, &
         line_search_error_threshold, num_threshold, numer, obj_function, quadratic_approx_error, &
         quadratic_approx_error_threshold, safety_multiplier, spin_factor, step_size, &
         step_size_quadratic_approx, step_size_quadratic_approx2, t1, t1a, t1cholesky, t2, t2a, &
         t2cholesky, tau, time_step, x_opt_eps_adaptive, x_opt_eps_adaptive_factor
      REAL(KIND=dp), DIMENSION(1)                        :: local_mu
      REAL(KIND=dp), DIMENSION(2)                        :: energy_correction
      REAL(KIND=dp), DIMENSION(3)                        :: minima
      TYPE(cp_dbcsr_type) :: grad, k_vd_index_down, k_vr_index_down, matrix_k_central, &
         matrix_tmp1, matrix_tmp2, prec, prev_grad, prev_minus_prec_grad, prev_step, &
         sigma_oo_curr, sigma_oo_curr_inv, sigma_vv_sqrt, sigma_vv_sqrt_guess, sigma_vv_sqrt_inv, &
         sigma_vv_sqrt_inv_guess, step, t_curr, tmp1_n_vr, tmp2_n_o, tmp3_vd_vr, tmp4_o_vr, &
         tmp_k_blk, vd_fixed, vd_index_sqrt, vd_index_sqrt_inv, velocity, vr_fixed, vr_index_sqrt, &
         vr_index_sqrt_inv
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: matrix_p_almo_scf_converged
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(ct_step_env_type)                             :: ct_step_env

!TYPE(cp_dbcsr_iterator)                  :: iter
!TYPE(cp_dbcsr_type) :: tmp11,tmp22,tmp33
!REAL(kind=dp)                            :: k_var1, k_var2
!TYPE(cp_dbcsr_type)                      :: fake_step
!
!TYPE(cp_dbcsr_type)                      :: sigma_dr, sigma_dr2, sigma_rr, sigma_rr2
!TYPE(cp_dbcsr_type)                      :: fake_a,fake_b,fake_k0
!INTEGER                                  :: retained_v,discarded_v,i_row,j_col
!TYPE(cp_dbcsr_type) :: matrix_rst0, matrix_rst1, matrix_rst2, ss_vv
!REAL(KIND=dp)       :: filter_memorize, init_filter, occ_vv
!INTEGER             :: ppp

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      nspin = almo_scf_env%nspins
      energy_correction_final = 0.0_dp
      IF (nspin .EQ. 1) THEN
         spin_factor = 2.0_dp
      ELSE
         spin_factor = 1.0_dp
      ENDIF

      IF (almo_scf_env%deloc_use_occ_orbs) THEN
         algorithm_id = cayley_transform
      ELSE
         algorithm_id = dm_ls_step
      ENDIF

      t1 = m_walltime()

      SELECT CASE (algorithm_id)
      CASE (cayley_transform)

         ! rescale density matrix by spin factor
         ! so the orbitals and density are consistent with each other
         IF (almo_scf_env%nspins == 1) THEN
            CALL cp_dbcsr_scale(almo_scf_env%matrix_p(1), 1.0_dp/spin_factor)
         ENDIF

         ! transform matrix_t not matrix_t_blk (we might need ALMOs later)
         DO ispin = 1, nspin

            CALL cp_dbcsr_copy(almo_scf_env%matrix_t(ispin), &
                               almo_scf_env%matrix_t_blk(ispin))

            ! obtain orthogonalization matrices for ALMOs
            ! RZK-warning - remove this sqrt(sigma) and inv(sqrt(sigma))
            ! ideally ALMO scf should use sigma and sigma_inv in
            ! the tensor_up_down representation

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) "sqrt and inv(sqrt) of MO overlap matrix"
            ENDIF
            CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_sqrt(ispin))
            CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_sqrt_inv(ispin))
            CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_sqrt(ispin), &
                                 template=almo_scf_env%matrix_sigma(ispin), &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                 template=almo_scf_env%matrix_sigma(ispin), &
                                 matrix_type=dbcsr_type_no_symmetry)

            CALL matrix_sqrt_Newton_Schulz(almo_scf_env%matrix_sigma_sqrt(ispin), &
                                           almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                           almo_scf_env%matrix_sigma(ispin), &
                                           threshold=almo_scf_env%eps_filter, &
                                           order=almo_scf_env%order_lanczos, &
                                           eps_lanczos=almo_scf_env%eps_lanczos, &
                                           max_iter_lanczos=almo_scf_env%max_iter_lanczos)

            IF (safe_mode) THEN
               CALL cp_dbcsr_init(matrix_tmp1)
               CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_sigma(ispin), &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_init(matrix_tmp2)
               CALL cp_dbcsr_create(matrix_tmp2, template=almo_scf_env%matrix_sigma(ispin), &
                                    matrix_type=dbcsr_type_no_symmetry)

               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                      almo_scf_env%matrix_sigma(ispin), &
                                      0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                      almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                      0.0_dp, matrix_tmp2, filter_eps=almo_scf_env%eps_filter)

               frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
               CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
               frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "Error for (inv(sqrt(SIG))*SIG*inv(sqrt(SIG))-I)", frob_matrix/frob_matrix_base
               ENDIF

               CALL cp_dbcsr_release(matrix_tmp1)
               CALL cp_dbcsr_release(matrix_tmp2)
            ENDIF
         ENDDO

         IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_diag) THEN

            DO ispin = 1, nspin

               t1a = m_walltime()

               line_search_error_threshold = almo_scf_env%real01
               conjugacy_error_threshold = almo_scf_env%real02
               quadratic_approx_error_threshold = almo_scf_env%real03
               x_opt_eps_adaptive_factor = almo_scf_env%real04

               !! the outer loop for k optimization
               outer_opt_k_max_iter = almo_scf_env%opt_k_outer_max_iter
               outer_opt_k_prepare_to_exit = .FALSE.
               outer_opt_k_iteration = 0
               grad_norm = 0.0_dp
               grad_norm_frob = 0.0_dp
               CALL cp_dbcsr_set(almo_scf_env%matrix_x(ispin), 0.0_dp)
               IF (almo_scf_env%deloc_truncate_virt .EQ. virt_full) outer_opt_k_max_iter = 0

               DO

                  ! obtain proper retained virtuals (1-R)|ALMO_vr>
                  CALL apply_projector(psi_in=almo_scf_env%matrix_v_blk(ispin), &
                                       psi_out=almo_scf_env%matrix_v(ispin), &
                                       psi_projector=almo_scf_env%matrix_t_blk(ispin), &
                                       metric=almo_scf_env%matrix_s(1), &
                                       project_out=.TRUE., &
                                       psi_projector_orthogonal=.FALSE., &
                                       proj_in_template=almo_scf_env%matrix_ov(ispin), &
                                       eps_filter=almo_scf_env%eps_filter, &
                                       sig_inv_projector=almo_scf_env%matrix_sigma_inv(ispin))
                  !sig_inv_template=almo_scf_env%matrix_sigma_inv(ispin),&

                  ! save initial retained virtuals
                  CALL cp_dbcsr_init(vr_fixed)
                  CALL cp_dbcsr_create(vr_fixed, &
                                       template=almo_scf_env%matrix_v(ispin))
                  CALL cp_dbcsr_copy(vr_fixed, almo_scf_env%matrix_v(ispin))

                  ! init matrices common for optimized and non-optimized virts
                  CALL cp_dbcsr_init(sigma_vv_sqrt)
                  CALL cp_dbcsr_init(sigma_vv_sqrt_inv)
                  CALL cp_dbcsr_create(sigma_vv_sqrt, &
                                       template=almo_scf_env%matrix_sigma_vv(ispin), &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_create(sigma_vv_sqrt_inv, &
                                       template=almo_scf_env%matrix_sigma_vv(ispin), &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_init(sigma_vv_sqrt_inv_guess)
                  CALL cp_dbcsr_init(sigma_vv_sqrt_guess)
                  CALL cp_dbcsr_create(sigma_vv_sqrt_inv_guess, &
                                       template=almo_scf_env%matrix_sigma_vv(ispin), &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_create(sigma_vv_sqrt_guess, &
                                       template=almo_scf_env%matrix_sigma_vv(ispin), &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_set(sigma_vv_sqrt_guess, 0.0_dp)
                  CALL cp_dbcsr_add_on_diag(sigma_vv_sqrt_guess, 1.0_dp)
                  CALL cp_dbcsr_filter(sigma_vv_sqrt_guess, almo_scf_env%eps_filter)
                  CALL cp_dbcsr_set(sigma_vv_sqrt_inv_guess, 0.0_dp)
                  CALL cp_dbcsr_add_on_diag(sigma_vv_sqrt_inv_guess, 1.0_dp)
                  CALL cp_dbcsr_filter(sigma_vv_sqrt_inv_guess, almo_scf_env%eps_filter)

                  ! do things required to optimize virtuals
                  IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN

                     ! project retained virtuals out of discarded block-by-block
                     ! (1-Q^VR_ALMO)|ALMO_vd>
                     ! this is probably not necessary, do it just to be safe
                     !CALL apply_projector(psi_in=almo_scf_env%matrix_v_disc_blk(ispin),&
                     !        psi_out=almo_scf_env%matrix_v_disc(ispin),&
                     !        psi_projector=almo_scf_env%matrix_v_blk(ispin),&
                     !        metric=almo_scf_env%matrix_s_blk(1),&
                     !        project_out=.TRUE.,&
                     !        psi_projector_orthogonal=.FALSE.,&
                     !        proj_in_template=almo_scf_env%matrix_k_tr(ispin),&
                     !        eps_filter=almo_scf_env%eps_filter,&
                     !        sig_inv_template=almo_scf_env%matrix_sigma_vv(ispin))
                     !CALL cp_dbcsr_copy(almo_scf_env%matrix_v_disc_blk(ispin),&
                     !        almo_scf_env%matrix_v_disc(ispin))

                     ! construct discarded virtuals (1-R)|ALMO_vd>
                     CALL apply_projector(psi_in=almo_scf_env%matrix_v_disc_blk(ispin), &
                                          psi_out=almo_scf_env%matrix_v_disc(ispin), &
                                          psi_projector=almo_scf_env%matrix_t_blk(ispin), &
                                          metric=almo_scf_env%matrix_s(1), &
                                          project_out=.TRUE., &
                                          psi_projector_orthogonal=.FALSE., &
                                          proj_in_template=almo_scf_env%matrix_ov_disc(ispin), &
                                          eps_filter=almo_scf_env%eps_filter, &
                                          sig_inv_projector=almo_scf_env%matrix_sigma_inv(ispin))
                     !sig_inv_template=almo_scf_env%matrix_sigma_inv(ispin),&

                     ! save initial discarded
                     CALL cp_dbcsr_init(vd_fixed)
                     CALL cp_dbcsr_create(vd_fixed, &
                                          template=almo_scf_env%matrix_v_disc(ispin))
                     CALL cp_dbcsr_copy(vd_fixed, almo_scf_env%matrix_v_disc(ispin))

                     !! create the down metric in the retained k-subspace
                     CALL cp_dbcsr_init(k_vr_index_down)
                     CALL cp_dbcsr_create(k_vr_index_down, &
                                          template=almo_scf_env%matrix_sigma_vv_blk(ispin), &
                                          matrix_type=dbcsr_type_no_symmetry)
                     !CALL cp_dbcsr_copy(k_vr_index_down,&
                     !        almo_scf_env%matrix_sigma_vv_blk(ispin))

                     !CALL get_overlap(bra=almo_scf_env%matrix_v_blk(ispin),&
                     !        ket=almo_scf_env%matrix_v_blk(ispin),&
                     !        overlap=k_vr_index_down,&
                     !        metric=almo_scf_env%matrix_s_blk(1),&
                     !        retain_overlap_sparsity=.FALSE.,&
                     !        eps_filter=almo_scf_env%eps_filter)

                     !! create the up metric in the discarded k-subspace
                     CALL cp_dbcsr_init(k_vd_index_down)
                     CALL cp_dbcsr_create(k_vd_index_down, &
                                          template=almo_scf_env%matrix_vv_disc_blk(ispin), &
                                          matrix_type=dbcsr_type_no_symmetry)
                     !CALL cp_dbcsr_init(k_vd_index_up)
                     !CALL cp_dbcsr_create(k_vd_index_up,&
                     !        template=almo_scf_env%matrix_vv_disc_blk(ispin),&
                     !        matrix_type=dbcsr_type_no_symmetry)
                     !CALL cp_dbcsr_copy(k_vd_index_down,&
                     !        almo_scf_env%matrix_vv_disc_blk(ispin))

                     !CALL get_overlap(bra=almo_scf_env%matrix_v_disc_blk(ispin),&
                     !        ket=almo_scf_env%matrix_v_disc_blk(ispin),&
                     !        overlap=k_vd_index_down,&
                     !        metric=almo_scf_env%matrix_s_blk(1),&
                     !        retain_overlap_sparsity=.FALSE.,&
                     !        eps_filter=almo_scf_env%eps_filter)

                     !IF (unit_nr>0) THEN
                     !   WRITE(unit_nr,*) "Inverting blocked overlap matrix of discarded virtuals"
                     !ENDIF
                     !CALL invert_Hotelling(k_vd_index_up,&
                     !        k_vd_index_down,&
                     !        almo_scf_env%eps_filter)
                     !IF (safe_mode) THEN
                     !   CALL cp_dbcsr_init(matrix_tmp1)
                     !   CALL cp_dbcsr_create(matrix_tmp1,template=k_vd_index_down,&
                     !                        matrix_type=dbcsr_type_no_symmetry)
                     !   CALL cp_dbcsr_multiply("N","N",1.0_dp,k_vd_index_up,&
                     !                          k_vd_index_down,&
                     !                          0.0_dp, matrix_tmp1,&
                     !                          filter_eps=almo_scf_env%eps_filter)
                     !   frob_matrix_base=cp_dbcsr_frobenius_norm(matrix_tmp1)
                     !   CALL cp_dbcsr_add_on_diag(matrix_tmp1,-1.0_dp)
                     !   frob_matrix=cp_dbcsr_frobenius_norm(matrix_tmp1)
                     !   IF (unit_nr>0) THEN
                     !      WRITE(unit_nr,*) "Error for (inv(SIG)*SIG-I)",&
                     !            frob_matrix/frob_matrix_base
                     !   ENDIF
                     !   CALL cp_dbcsr_release(matrix_tmp1)
                     !ENDIF

                     ! init matrices necessary for optimization of truncated virts
                     ! init blocked gradient before setting K to zero
                     ! otherwise the block structure might be lost
                     CALL cp_dbcsr_init(grad)
                     CALL cp_dbcsr_create(grad, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_copy(grad, almo_scf_env%matrix_k_blk(ispin))

                     ! init MD in the k-space
                     md_in_k_space = almo_scf_env%logical01
                     IF (md_in_k_space) THEN
                        CALL cp_dbcsr_init(velocity)
                        CALL cp_dbcsr_create(velocity, &
                                             template=almo_scf_env%matrix_k_blk(ispin))
                        CALL cp_dbcsr_copy(velocity, almo_scf_env%matrix_k_blk(ispin))
                        CALL cp_dbcsr_set(velocity, 0.0_dp)
                        time_step = almo_scf_env%opt_k_trial_step_size
                     ENDIF

                     CALL cp_dbcsr_init(prev_step)
                     CALL cp_dbcsr_create(prev_step, &
                                          template=almo_scf_env%matrix_k_blk(ispin))

                     CALL cp_dbcsr_init(prev_minus_prec_grad)
                     CALL cp_dbcsr_create(prev_minus_prec_grad, &
                                          template=almo_scf_env%matrix_k_blk(ispin))

                     ! initialize diagonal blocks of the preconditioner to 1.0_dp
                     CALL cp_dbcsr_init(prec)
                     CALL cp_dbcsr_create(prec, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_copy(prec, almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_set(prec, 1.0_dp)

                     ! generate initial K (extrapolate if previous values are available)
                     CALL cp_dbcsr_set(almo_scf_env%matrix_k_blk(ispin), 0.0_dp)
                     ! matrix_k_central stores current k because matrix_k_blk is updated
                     ! during linear search
                     CALL cp_dbcsr_init(matrix_k_central)
                     CALL cp_dbcsr_create(matrix_k_central, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_copy(matrix_k_central, &
                                        almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_init(tmp_k_blk)
                     CALL cp_dbcsr_create(tmp_k_blk, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_init(step)
                     CALL cp_dbcsr_create(step, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_set(step, 0.0_dp)
                     CALL cp_dbcsr_init(t_curr)
                     CALL cp_dbcsr_create(t_curr, &
                                          template=almo_scf_env%matrix_t(ispin))
                     CALL cp_dbcsr_init(sigma_oo_curr)
                     CALL cp_dbcsr_init(sigma_oo_curr_inv)
                     CALL cp_dbcsr_create(sigma_oo_curr, &
                                          template=almo_scf_env%matrix_sigma(ispin), &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL cp_dbcsr_create(sigma_oo_curr_inv, &
                                          template=almo_scf_env%matrix_sigma(ispin), &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL cp_dbcsr_init(tmp1_n_vr)
                     CALL cp_dbcsr_create(tmp1_n_vr, &
                                          template=almo_scf_env%matrix_v(ispin))
                     CALL cp_dbcsr_init(tmp3_vd_vr)
                     CALL cp_dbcsr_create(tmp3_vd_vr, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_init(tmp2_n_o)
                     CALL cp_dbcsr_create(tmp2_n_o, &
                                          template=almo_scf_env%matrix_t(ispin))
                     CALL cp_dbcsr_init(tmp4_o_vr)
                     CALL cp_dbcsr_create(tmp4_o_vr, &
                                          template=almo_scf_env%matrix_ov(ispin))
                     CALL cp_dbcsr_init(prev_grad)
                     CALL cp_dbcsr_create(prev_grad, &
                                          template=almo_scf_env%matrix_k_blk(ispin))
                     CALL cp_dbcsr_set(prev_grad, 0.0_dp)

                     !CALL cp_dbcsr_init(sigma_oo_guess)
                     !CALL cp_dbcsr_create(sigma_oo_guess,&
                     !        template=almo_scf_env%matrix_sigma(ispin),&
                     !        matrix_type=dbcsr_type_no_symmetry)
                     !CALL cp_dbcsr_set(sigma_oo_guess,0.0_dp)
                     !CALL cp_dbcsr_add_on_diag(sigma_oo_guess,1.0_dp)
                     !CALL cp_dbcsr_filter(sigma_oo_guess,almo_scf_env%eps_filter)
                     !CALL cp_dbcsr_print(sigma_oo_guess)

                  ENDIF ! done constructing discarded virtuals

                  ! init variables
                  opt_k_max_iter = almo_scf_env%opt_k_max_iter
                  iteration = 0
                  converged = .FALSE.
                  prepare_to_exit = .FALSE.
                  beta = 0.0_dp
                  line_search = .FALSE.
                  obj_function = 0.0_dp
                  conjugacy_error = 0.0_dp
                  line_search_error = 0.0_dp
                  fun0 = 0.0_dp
                  fun1 = 0.0_dp
                  gfun0 = 0.0_dp
                  gfun1 = 0.0_dp
                  step_size_quadratic_approx = 0.0_dp
                  reset_step_size = .TRUE.
                  IF (almo_scf_env%deloc_truncate_virt .EQ. virt_full) opt_k_max_iter = 0

                  ! start cg iterations to optimize matrix_k_blk
                  DO

                     CALL timeset('k_opt_vr', handle1)

                     IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN

                        ! construct k-excited virtuals
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vd_fixed, &
                                               almo_scf_env%matrix_k_blk(ispin), &
                                               0.0_dp, almo_scf_env%matrix_v(ispin), &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_add(almo_scf_env%matrix_v(ispin), vr_fixed, &
                                          +1.0_dp, +1.0_dp)
                     ENDIF

                     ! decompose the overlap matrix of the current retained orbitals
                     !IF (unit_nr>0) THEN
                     !   WRITE(unit_nr,*) "decompose the active VV overlap matrix"
                     !ENDIF
                     CALL get_overlap(bra=almo_scf_env%matrix_v(ispin), &
                                      ket=almo_scf_env%matrix_v(ispin), &
                                      overlap=almo_scf_env%matrix_sigma_vv(ispin), &
                                      metric=almo_scf_env%matrix_s(1), &
                                      retain_overlap_sparsity=.FALSE., &
                                      eps_filter=almo_scf_env%eps_filter)
                     ! use either cholesky or sqrt
                     !! RZK-warning: strangely, cholesky does not work with k-optimization
                     IF (almo_scf_env%deloc_truncate_virt .EQ. virt_full) THEN
                        CALL timeset('cholesky', handle2)
                        t1cholesky = m_walltime()

                        ! re-create sigma_vv_sqrt because desymmetrize is buggy -
                        ! it will create multiple copies of blocks
                        CALL cp_dbcsr_create(sigma_vv_sqrt, &
                                             template=almo_scf_env%matrix_sigma_vv(ispin), &
                                             matrix_type=dbcsr_type_no_symmetry)
                        CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma_vv(ispin), &
                                                   sigma_vv_sqrt)
                        CALL cp_dbcsr_cholesky_decompose(sigma_vv_sqrt, &
                                                         para_env=almo_scf_env%para_env, &
                                                         blacs_env=almo_scf_env%blacs_env)
                        CALL cp_dbcsr_triu(sigma_vv_sqrt)
                        CALL cp_dbcsr_filter(sigma_vv_sqrt, almo_scf_env%eps_filter)
                        ! apply SOLVE to compute U^(-1) : U*U^(-1)=I
                        CALL cp_dbcsr_get_info(sigma_vv_sqrt, nfullrows_total=n)
                        CALL cp_dbcsr_init(matrix_tmp1)
                        CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_sigma_vv(ispin), &
                                             matrix_type=dbcsr_type_no_symmetry)
                        CALL cp_dbcsr_set(matrix_tmp1, 0.0_dp)
                        CALL cp_dbcsr_add_on_diag(matrix_tmp1, 1.0_dp)
                        CALL cp_dbcsr_cholesky_restore(matrix_tmp1, n, sigma_vv_sqrt, &
                                                       sigma_vv_sqrt_inv, op="SOLVE", pos="RIGHT", &
                                                       para_env=almo_scf_env%para_env, &
                                                       blacs_env=almo_scf_env%blacs_env)
                        CALL cp_dbcsr_filter(sigma_vv_sqrt_inv, almo_scf_env%eps_filter)
                        CALL cp_dbcsr_release(matrix_tmp1)
                        IF (safe_mode) THEN
                           CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_sigma_vv(ispin), &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_desymmetrize(almo_scf_env%matrix_sigma_vv(ispin), &
                                                      matrix_tmp1)
                           CALL cp_dbcsr_multiply("T", "N", 1.0_dp, sigma_vv_sqrt, &
                                                  sigma_vv_sqrt, &
                                                  -1.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                           frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           CALL cp_dbcsr_add_on_diag(matrix_tmp1, 1.0_dp)
                           frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, *) "Error for ( U^T * U - Sig )", &
                                 frob_matrix/frob_matrix_base
                           ENDIF
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, sigma_vv_sqrt_inv, &
                                                  sigma_vv_sqrt, &
                                                  0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                           frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           CALL cp_dbcsr_add_on_diag(matrix_tmp1, -1.0_dp)
                           frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, *) "Error for ( inv(U) * U - I )", &
                                 frob_matrix/frob_matrix_base
                           ENDIF
                           CALL cp_dbcsr_release(matrix_tmp1)
                        ENDIF ! safe_mode
                        t2cholesky = m_walltime()
                        IF (unit_nr > 0) THEN
                           WRITE (unit_nr, *) "Cholesky+inverse wall-time: ", t2cholesky-t1cholesky
                        ENDIF
                        CALL timestop(handle2)
                     ELSE
                        CALL matrix_sqrt_Newton_Schulz(sigma_vv_sqrt, &
                                                       sigma_vv_sqrt_inv, &
                                                       almo_scf_env%matrix_sigma_vv(ispin), &
                                                       !matrix_sqrt_inv_guess=sigma_vv_sqrt_inv_guess,&
                                                       !matrix_sqrt_guess=sigma_vv_sqrt_guess,&
                                                       threshold=almo_scf_env%eps_filter, &
                                                       order=almo_scf_env%order_lanczos, &
                                                       eps_lanczos=almo_scf_env%eps_lanczos, &
                                                       max_iter_lanczos=almo_scf_env%max_iter_lanczos)
                        CALL cp_dbcsr_copy(sigma_vv_sqrt_inv_guess, sigma_vv_sqrt_inv)
                        CALL cp_dbcsr_copy(sigma_vv_sqrt_guess, sigma_vv_sqrt)
                        IF (safe_mode) THEN
                           CALL cp_dbcsr_init(matrix_tmp1)
                           CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_sigma_vv(ispin), &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_init(matrix_tmp2)
                           CALL cp_dbcsr_create(matrix_tmp2, template=almo_scf_env%matrix_sigma_vv(ispin), &
                                                matrix_type=dbcsr_type_no_symmetry)

                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, sigma_vv_sqrt_inv, &
                                                  almo_scf_env%matrix_sigma_vv(ispin), &
                                                  0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                                  sigma_vv_sqrt_inv, &
                                                  0.0_dp, matrix_tmp2, filter_eps=almo_scf_env%eps_filter)

                           frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
                           CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
                           frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, *) "Error for (inv(sqrt(SIGVV))*SIGVV*inv(sqrt(SIGVV))-I)", &
                                 frob_matrix/frob_matrix_base
                           ENDIF

                           CALL cp_dbcsr_release(matrix_tmp1)
                           CALL cp_dbcsr_release(matrix_tmp2)
                        ENDIF
                     ENDIF
                     CALL timestop(handle1)

                     ! compute excitation amplitudes (to the current set of retained virtuals)
                     ! set convergence criterion for x-optimization
                     IF ((iteration .EQ. 0) .AND. (.NOT. line_search) .AND. &
                         (outer_opt_k_iteration .EQ. 0)) THEN
                        x_opt_eps_adaptive = &
                           almo_scf_env%deloc_cayley_eps_convergence
                     ELSE
                        x_opt_eps_adaptive = &
                           MAX(ABS(almo_scf_env%deloc_cayley_eps_convergence), &
                               ABS(x_opt_eps_adaptive_factor*grad_norm))
                     ENDIF
                     CALL ct_step_env_init(ct_step_env)
                     CALL ct_step_env_set(ct_step_env, &
                                          para_env=almo_scf_env%para_env, &
                                          blacs_env=almo_scf_env%blacs_env, &
                                          use_occ_orbs=.TRUE., &
                                          use_virt_orbs=.TRUE., &
                                          occ_orbs_orthogonal=.FALSE., &
                                          virt_orbs_orthogonal=.FALSE., &
                                          pp_preconditioner_full=almo_scf_env%deloc_cayley_occ_precond, &
                                          qq_preconditioner_full=almo_scf_env%deloc_cayley_vir_precond, &
                                          tensor_type=almo_scf_env%deloc_cayley_tensor_type, &
                                          neglect_quadratic_term=almo_scf_env%deloc_cayley_linear, &
                                          conjugator=almo_scf_env%deloc_cayley_conjugator, &
                                          max_iter=almo_scf_env%deloc_cayley_max_iter, &
                                          calculate_energy_corr=.TRUE., &
                                          update_p=.FALSE., &
                                          update_q=.FALSE., &
                                          eps_convergence=x_opt_eps_adaptive, &
                                          eps_filter=almo_scf_env%eps_filter, &
                                          !nspins=1,&
                                          q_index_up=sigma_vv_sqrt_inv, &
                                          q_index_down=sigma_vv_sqrt, &
                                          p_index_up=almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                          p_index_down=almo_scf_env%matrix_sigma_sqrt(ispin), &
                                          matrix_ks=almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                          matrix_t=almo_scf_env%matrix_t(ispin), &
                                          matrix_qp_template=almo_scf_env%matrix_vo(ispin), &
                                          matrix_pq_template=almo_scf_env%matrix_ov(ispin), &
                                          matrix_v=almo_scf_env%matrix_v(ispin), &
                                          matrix_x_guess=almo_scf_env%matrix_x(ispin))
                     ! perform calculations
                     CALL ct_step_execute(ct_step_env)
                     ! get the energy correction
                     CALL ct_step_env_get(ct_step_env, &
                                          energy_correction=energy_correction(ispin), &
                                          copy_matrix_x=almo_scf_env%matrix_x(ispin))
                     CALL ct_step_env_clean(ct_step_env)
                     ! RZK-warning matrix_x is being transformed
                     ! back and forth between orth and up_down representations
                     energy_correction(1) = energy_correction(1)*spin_factor

                     IF (opt_k_max_iter .NE. 0) THEN

                        CALL timeset('k_opt_t_curr', handle3)

                        ! construct current occupied orbitals T_blk + V_r*X
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                               almo_scf_env%matrix_v(ispin), &
                                               almo_scf_env%matrix_x(ispin), &
                                               0.0_dp, t_curr, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_add(t_curr, almo_scf_env%matrix_t_blk(ispin), &
                                          +1.0_dp, +1.0_dp)

                        ! calculate current occupied overlap
                        !IF (unit_nr>0) THEN
                        !   WRITE(unit_nr,*) "Inverting current occ overlap matrix"
                        !ENDIF
                        CALL get_overlap(bra=t_curr, &
                                         ket=t_curr, &
                                         overlap=sigma_oo_curr, &
                                         metric=almo_scf_env%matrix_s(1), &
                                         retain_overlap_sparsity=.FALSE., &
                                         eps_filter=almo_scf_env%eps_filter)
                        IF (iteration .EQ. 0) THEN
                           CALL invert_Hotelling(sigma_oo_curr_inv, &
                                                 sigma_oo_curr, &
                                                 threshold=almo_scf_env%eps_filter, &
                                                 use_inv_as_guess=.FALSE.)
                        ELSE
                           CALL invert_Hotelling(sigma_oo_curr_inv, &
                                                 sigma_oo_curr, &
                                                 threshold=almo_scf_env%eps_filter, &
                                                 use_inv_as_guess=.TRUE.)
                           !CALL cp_dbcsr_copy(sigma_oo_guess,sigma_oo_curr_inv)
                        ENDIF
                        IF (safe_mode) THEN
                           CALL cp_dbcsr_init(matrix_tmp1)
                           CALL cp_dbcsr_create(matrix_tmp1, template=sigma_oo_curr, &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, sigma_oo_curr, &
                                                  sigma_oo_curr_inv, &
                                                  0.0_dp, matrix_tmp1, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           CALL cp_dbcsr_add_on_diag(matrix_tmp1, -1.0_dp)
                           frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           !CALL cp_dbcsr_filter(matrix_tmp1,almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_print(matrix_tmp1)
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, *) "Error for (SIG*inv(SIG)-I)", &
                                 frob_matrix/frob_matrix_base, frob_matrix_base
                           ENDIF
                           CALL cp_dbcsr_release(matrix_tmp1)
                        ENDIF
                        IF (safe_mode) THEN
                           CALL cp_dbcsr_init(matrix_tmp1)
                           CALL cp_dbcsr_create(matrix_tmp1, template=sigma_oo_curr, &
                                                matrix_type=dbcsr_type_no_symmetry)
                           CALL cp_dbcsr_multiply("N", "N", 1.0_dp, sigma_oo_curr_inv, &
                                                  sigma_oo_curr, &
                                                  0.0_dp, matrix_tmp1, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           CALL cp_dbcsr_add_on_diag(matrix_tmp1, -1.0_dp)
                           frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
                           !CALL cp_dbcsr_filter(matrix_tmp1,almo_scf_env%eps_filter)
                           !CALL cp_dbcsr_print(matrix_tmp1)
                           IF (unit_nr > 0) THEN
                              WRITE (unit_nr, *) "Error for (inv(SIG)*SIG-I)", &
                                 frob_matrix/frob_matrix_base, frob_matrix_base
                           ENDIF
                           CALL cp_dbcsr_release(matrix_tmp1)
                        ENDIF

                        CALL timestop(handle3)
                        CALL timeset('k_opt_vd', handle4)

                        ! construct current discarded virtuals:
                        ! (1-R_curr)(1-Q^VR_curr)|ALMO_vd_basis> =
                        ! = (1-Q^VR_curr)|ALMO_vd_basis>
                        ! use sigma_vv_sqrt to store the inverse of the overlap
                        ! sigma_vv_inv is computed from sqrt/cholesky
                        CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                               sigma_vv_sqrt_inv, &
                                               sigma_vv_sqrt_inv, &
                                               0.0_dp, sigma_vv_sqrt, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL apply_projector(psi_in=almo_scf_env%matrix_v_disc_blk(ispin), &
                                             psi_out=almo_scf_env%matrix_v_disc(ispin), &
                                             psi_projector=almo_scf_env%matrix_v(ispin), &
                                             metric=almo_scf_env%matrix_s(1), &
                                             project_out=.FALSE., &
                                             psi_projector_orthogonal=.FALSE., &
                                             proj_in_template=almo_scf_env%matrix_k_tr(ispin), &
                                             eps_filter=almo_scf_env%eps_filter, &
                                             sig_inv_projector=sigma_vv_sqrt)
                        !sig_inv_template=almo_scf_env%matrix_sigma_vv(ispin),&
                        CALL cp_dbcsr_add(almo_scf_env%matrix_v_disc(ispin), &
                                          vd_fixed, -1.0_dp, +1.0_dp)

                        CALL timestop(handle4)
                        CALL timeset('k_opt_grad', handle5)

                        ! evaluate the gradient from the assembled components
                        ! grad_xx = c0 [ (Vd_curr^tr)*F*T_curr*sigma_oo_curr_inv*(X^tr)]_xx
                        ! save previous gradient to calculate conjugation coef
                        IF (line_search) THEN
                           CALL cp_dbcsr_copy(prev_grad, grad)
                        ENDIF
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                               almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                               t_curr, &
                                               0.0_dp, tmp2_n_o, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                               sigma_oo_curr_inv, &
                                               almo_scf_env%matrix_x(ispin), &
                                               0.0_dp, tmp4_o_vr, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                               tmp2_n_o, &
                                               tmp4_o_vr, &
                                               0.0_dp, tmp1_n_vr, &
                                               filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("T", "N", 2.0_dp*spin_factor, &
                                               almo_scf_env%matrix_v_disc(ispin), &
                                               tmp1_n_vr, &
                                               0.0_dp, grad, &
                                               retain_sparsity=.TRUE.)
                        !filter_eps=almo_scf_env%eps_filter,&
                        ! keep tmp2_n_o for the next step
                        ! keep tmp4_o_vr for the preconditioner

                        ! check convergence and other exit criteria
                        grad_norm_frob = cp_dbcsr_frobenius_norm(grad)
                        CALL cp_dbcsr_norm(grad, dbcsr_norm_maxabsnorm, norm_scalar=grad_norm)
                        converged = (grad_norm .LT. almo_scf_env%opt_k_eps_convergence)
                        IF (converged .OR. (iteration .GE. opt_k_max_iter)) THEN
                           prepare_to_exit = .TRUE.
                        ENDIF
                        CALL timestop(handle5)

                        IF (.NOT. prepare_to_exit) THEN

                           CALL timeset('k_opt_energy', handle6)

                           ! compute "energy" c0*Tr[sig_inv_oo*t*F*t]
                           CALL cp_dbcsr_multiply("T", "N", spin_factor, &
                                                  t_curr, &
                                                  tmp2_n_o, &
                                                  0.0_dp, sigma_oo_curr, &
                                                  filter_eps=almo_scf_env%eps_filter)
                           delta_obj_function = fun0
                           CALL cp_dbcsr_trace(sigma_oo_curr_inv, &
                                               sigma_oo_curr, obj_function, "T", "N")
                           delta_obj_function = obj_function-delta_obj_function
                           IF (line_search) THEN
                              fun1 = obj_function
                           ELSE
                              fun0 = obj_function
                           ENDIF

                           CALL timestop(handle6)

                           ! update the step direction
                           IF (.NOT. line_search) THEN

                              CALL timeset('k_opt_step', handle7)

                              IF ((.NOT. md_in_k_space) .AND. &
                                  (iteration .GE. MAX(0, almo_scf_env%opt_k_prec_iter_start) .AND. &
                                   MOD(iteration-almo_scf_env%opt_k_prec_iter_start, &
                                       almo_scf_env%opt_k_prec_iter_freq) .EQ. 0)) THEN

                                 !IF ((iteration.eq.0).AND.(.NOT.md_in_k_space)) THEN

                                 ! compute the preconditioner
                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, *) "Computing preconditioner"
                                 ENDIF
                                 !CALL opt_k_create_preconditioner(prec,&
                                 !        almo_scf_env%matrix_v_disc(ispin),&
                                 !        almo_scf_env%matrix_ks_almo_scf_converged(ispin),&
                                 !        almo_scf_env%matrix_x(ispin),&
                                 !        tmp4_o_vr,&
                                 !        almo_scf_env%matrix_s(1),&
                                 !        grad,&
                                 !        !almo_scf_env%matrix_v_disc_blk(ispin),&
                                 !        vd_fixed,&
                                 !        t_curr,&
                                 !        k_vd_index_up,&
                                 !        k_vr_index_down,&
                                 !        tmp1_n_vr,&
                                 !        spin_factor,&
                                 !        almo_scf_env%eps_filter)
                                 CALL opt_k_create_preconditioner_blk(almo_scf_env, &
                                                                      almo_scf_env%matrix_v_disc(ispin), &
                                                                      tmp4_o_vr, &
                                                                      t_curr, &
                                                                      ispin, &
                                                                      spin_factor)

                              ENDIF

                              ! save the previous step
                              CALL cp_dbcsr_copy(prev_step, step)

                              ! compute the new step
                              CALL opt_k_apply_preconditioner_blk(almo_scf_env, &
                                                                  step, grad, ispin)
                              !CALL cp_dbcsr_hadamard_product(prec,grad,step)
                              CALL cp_dbcsr_scale(step, -1.0_dp)

                              ! check whether we need to reset conjugate directions
                              reset_conjugator = .FALSE.
                              ! first check if manual reset is active
                              IF (iteration .LT. MAX(almo_scf_env%opt_k_conj_iter_start, 1) .OR. &
                                  MOD(iteration-almo_scf_env%opt_k_conj_iter_start, &
                                      almo_scf_env%opt_k_conj_iter_freq) .EQ. 0) THEN

                                 reset_conjugator = .TRUE.

                              ELSE

                                 ! check for the errors in the cg algorithm
                                 !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                 !CALL cp_dbcsr_trace(grad,tmp_k_blk,numer,"T","N")
                                 !CALL cp_dbcsr_trace(prev_grad,tmp_k_blk,denom,"T","N")
                                 CALL cp_dbcsr_trace(grad, prev_minus_prec_grad, numer, "T", "N")
                                 CALL cp_dbcsr_trace(prev_grad, prev_minus_prec_grad, denom, "T", "N")
                                 conjugacy_error = numer/denom

                                 IF (conjugacy_error .GT. MIN(0.5_dp, conjugacy_error_threshold)) THEN
                                    reset_conjugator = .TRUE.
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, *) "Lack of progress, conjugacy error is ", conjugacy_error
                                    ENDIF
                                 ENDIF

                                 ! check the gradient along the previous direction
                                 IF ((iteration .NE. 0) .AND. (.NOT. reset_conjugator)) THEN
                                    CALL cp_dbcsr_trace(grad, prev_step, numer, "T", "N")
                                    CALL cp_dbcsr_trace(prev_grad, prev_step, denom, "T", "N")
                                    line_search_error = numer/denom
                                    IF (line_search_error .GT. line_search_error_threshold) THEN
                                       reset_conjugator = .TRUE.
                                       IF (unit_nr > 0) THEN
                                          WRITE (unit_nr, *) "Bad line search, line search error is ", line_search_error
                                       ENDIF
                                    ENDIF
                                 ENDIF

                              ENDIF

                              ! compute the conjugation coefficient - beta
                              IF (.NOT. reset_conjugator) THEN

                                 SELECT CASE (almo_scf_env%opt_k_conjugator)
                                 CASE (cg_hestenes_stiefel)
                                    CALL cp_dbcsr_copy(tmp_k_blk, grad)
                                    CALL cp_dbcsr_add(tmp_k_blk, prev_grad, 1.0_dp, -1.0_dp)
                                    CALL cp_dbcsr_trace(tmp_k_blk, step, numer, "T", "N")
                                    CALL cp_dbcsr_trace(tmp_k_blk, prev_step, denom, "T", "N")
                                    beta = -1.0_dp*numer/denom
                                 CASE (cg_fletcher_reeves)
                                    !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(prev_grad,tmp_k_blk,denom,"T","N")
                                    !CALL cp_dbcsr_hadamard_product(prec,grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(grad,tmp_k_blk,numer,"T","N")
                                    !beta=numer/denom
                                    CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                                    CALL cp_dbcsr_trace(prev_grad, prev_minus_prec_grad, denom, "T", "N")
                                    beta = numer/denom
                                 CASE (cg_polak_ribiere)
                                    !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(prev_grad,tmp_k_blk,denom,"T","N")
                                    !CALL cp_dbcsr_add(prev_grad,grad,-1.0_dp,1.0_dp)
                                    !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(tmp_k_blk,grad,numer,"T","N")
                                    CALL cp_dbcsr_trace(prev_grad, prev_minus_prec_grad, denom, "T", "N")
                                    CALL cp_dbcsr_copy(tmp_k_blk, grad)
                                    CALL cp_dbcsr_add(tmp_k_blk, prev_grad, 1.0_dp, -1.0_dp)
                                    CALL cp_dbcsr_trace(tmp_k_blk, step, numer, "T", "N")
                                    beta = numer/denom
                                 CASE (cg_fletcher)
                                    !CALL cp_dbcsr_hadamard_product(prec,grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(grad,tmp_k_blk,numer,"T","N")
                                    !CALL cp_dbcsr_trace(prev_grad,prev_step,denom,"T","N")
                                    !beta=-1.0_dp*numer/denom
                                    CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                                    CALL cp_dbcsr_trace(prev_grad, prev_step, denom, "T", "N")
                                    beta = numer/denom
                                 CASE (cg_liu_storey)
                                    CALL cp_dbcsr_trace(prev_grad, prev_step, denom, "T", "N")
                                    !CALL cp_dbcsr_add(prev_grad,grad,-1.0_dp,1.0_dp)
                                    !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(tmp_k_blk,grad,numer,"T","N")
                                    CALL cp_dbcsr_copy(tmp_k_blk, grad)
                                    CALL cp_dbcsr_add(tmp_k_blk, prev_grad, 1.0_dp, -1.0_dp)
                                    CALL cp_dbcsr_trace(tmp_k_blk, step, numer, "T", "N")
                                    beta = numer/denom
                                 CASE (cg_dai_yuan)
                                    !CALL cp_dbcsr_hadamard_product(prec,grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(grad,tmp_k_blk,numer,"T","N")
                                    !CALL cp_dbcsr_add(prev_grad,grad,-1.0_dp,1.0_dp)
                                    !CALL cp_dbcsr_trace(prev_grad,prev_step,denom,"T","N")
                                    !beta=numer/denom
                                    CALL cp_dbcsr_trace(grad, step, numer, "T", "N")
                                    CALL cp_dbcsr_copy(tmp_k_blk, grad)
                                    CALL cp_dbcsr_add(tmp_k_blk, prev_grad, 1.0_dp, -1.0_dp)
                                    CALL cp_dbcsr_trace(tmp_k_blk, prev_step, denom, "T", "N")
                                    beta = -1.0_dp*numer/denom
                                 CASE (cg_hager_zhang)
                                    !CALL cp_dbcsr_add(prev_grad,grad,-1.0_dp,1.0_dp)
                                    !CALL cp_dbcsr_trace(prev_grad,prev_step,denom,"T","N")
                                    !CALL cp_dbcsr_hadamard_product(prec,prev_grad,tmp_k_blk)
                                    !CALL cp_dbcsr_trace(tmp_k_blk,prev_grad,numer,"T","N")
                                    !kappa=2.0_dp*numer/denom
                                    !CALL cp_dbcsr_trace(tmp_k_blk,grad,numer,"T","N")
                                    !tau=numer/denom
                                    !CALL cp_dbcsr_trace(prev_step,grad,numer,"T","N")
                                    !beta=tau-kappa*numer/denom
                                    CALL cp_dbcsr_copy(tmp_k_blk, grad)
                                    CALL cp_dbcsr_add(tmp_k_blk, prev_grad, 1.0_dp, -1.0_dp)
                                    CALL cp_dbcsr_trace(tmp_k_blk, prev_step, denom, "T", "N")
                                    CALL cp_dbcsr_trace(tmp_k_blk, prev_minus_prec_grad, numer, "T", "N")
                                    kappa = -2.0_dp*numer/denom
                                    CALL cp_dbcsr_trace(tmp_k_blk, step, numer, "T", "N")
                                    tau = -1.0_dp*numer/denom
                                    CALL cp_dbcsr_trace(prev_step, grad, numer, "T", "N")
                                    beta = tau-kappa*numer/denom
                                 CASE (cg_zero)
                                    beta = 0.0_dp
                                 CASE DEFAULT
                                    CALL cp__b("almo_scf_optimizer.F",3271,"illegal conjugator")
                                 END SELECT

                                 IF (beta .LT. 0.0_dp) THEN
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, *) "Beta is negative, ", beta
                                    ENDIF
                                    reset_conjugator = .TRUE.
                                 ENDIF

                              ENDIF

                              IF (md_in_k_space) THEN
                                 reset_conjugator = .TRUE.
                              ENDIF

                              IF (reset_conjugator) THEN

                                 beta = 0.0_dp
                                 !reset_step_size=.TRUE.

                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, *) "(Re)-setting conjugator to zero"
                                 ENDIF

                              ENDIF

                              ! save the preconditioned gradient
                              CALL cp_dbcsr_copy(prev_minus_prec_grad, step)

                              ! conjugate the step direction
                              CALL cp_dbcsr_add(step, prev_step, 1.0_dp, beta)

                              CALL timestop(handle7)

                              ! update the step direction
                           ELSE ! step update
                              conjugacy_error = 0.0_dp
                           ENDIF

                           ! compute the gradient with respect to the step size in the curr direction
                           IF (line_search) THEN
                              CALL cp_dbcsr_trace(grad, step, gfun1, "T", "N")
                              line_search_error = gfun1/gfun0
                           ELSE
                              CALL cp_dbcsr_trace(grad, step, gfun0, "T", "N")
                           ENDIF

                           ! make a step - update k
                           IF (line_search) THEN

                              ! check if the trial step provides enough numerical accuracy
                              safety_multiplier = 1.0E+1_dp ! must be more than one
                              num_threshold = MAX(EPSILON(1.0_dp), &
                                                  safety_multiplier*(almo_scf_env%eps_filter**2)*almo_scf_env%ndomains)
                              IF (ABS(fun1-fun0-gfun0*step_size) .LT. num_threshold) THEN
                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, '(T3,A,1X,E17.7)') &
                                       "Numerical accuracy is too low to observe non-linear behavior", &
                                       ABS(fun1-fun0-gfun0*step_size)
                                    WRITE (unit_nr, '(T3,A,1X,E17.7,A,1X,E12.3)') "Error computing ", &
                                       ABS(gfun0), &
                                       " is smaller than the threshold", num_threshold
                                 ENDIF
                                 CALL cp__b("almo_scf_optimizer.F",3335,"")
                              ENDIF
                              IF (ABS(gfun0) .LT. num_threshold) THEN
                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, '(T3,A,1X,E17.7,A,1X,E12.3)') "Linear gradient", &
                                       ABS(gfun0), &
                                       " is smaller than the threshold", num_threshold
                                 ENDIF
                                 CALL cp__b("almo_scf_optimizer.F",3343,"")
                              ENDIF

                              use_quadratic_approximation = .TRUE.
                              use_cubic_approximation = .FALSE.

                              ! find the minimum assuming quadratic form
                              ! use f0, f1, g0
                              step_size_quadratic_approx = -(gfun0*step_size*step_size)/(2.0_dp*(fun1-fun0-gfun0*step_size))
                              ! use f0, f1, g1
                              step_size_quadratic_approx2 = -(fun1-fun0-step_size*gfun1/2.0_dp)/(gfun1-(fun1-fun0)/step_size)

                              IF ((step_size_quadratic_approx .LT. 0.0_dp) .AND. &
                                  (step_size_quadratic_approx2 .LT. 0.0_dp)) THEN
                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, '(T3,A,1X,E17.7,1X,E17.7,1X,A)') &
                                       "Quadratic approximation gives negative steps", &
                                       step_size_quadratic_approx, step_size_quadratic_approx2, &
                                       "trying cubic..."
                                 ENDIF
                                 use_cubic_approximation = .TRUE.
                                 use_quadratic_approximation = .FALSE.
                              ELSE
                                 IF (step_size_quadratic_approx .LT. 0.0_dp) THEN
                                    step_size_quadratic_approx = step_size_quadratic_approx2
                                 ENDIF
                                 IF (step_size_quadratic_approx2 .LT. 0.0_dp) THEN
                                    step_size_quadratic_approx2 = step_size_quadratic_approx
                                 ENDIF
                              ENDIF

                              ! check accuracy of the quadratic approximation
                              IF (use_quadratic_approximation) THEN
                                 quadratic_approx_error = ABS(step_size_quadratic_approx- &
                                                              step_size_quadratic_approx2)/step_size_quadratic_approx
                                 IF (quadratic_approx_error .GT. quadratic_approx_error_threshold) THEN
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, '(T3,A,1X,E17.7,1X,E17.7,1X,A)') "Quadratic approximation is poor", &
                                          step_size_quadratic_approx, step_size_quadratic_approx2, &
                                          "Try cubic approximation"
                                    ENDIF
                                    use_cubic_approximation = .TRUE.
                                    use_quadratic_approximation = .FALSE.
                                 ENDIF
                              ENDIF

                              ! check if numerics is fine enough to capture the cubic form
                              IF (use_cubic_approximation) THEN

                                 ! if quadratic approximation is not accurate enough
                                 ! try to find the minimum assuming cubic form
                                 ! aa*x**3 + bb*x**2 + cc*x + dd = f(x)
                                 bb = (-step_size*gfun1+3.0_dp*(fun1-fun0)-2.0_dp*step_size*gfun0)/(step_size*step_size)
                                 aa = (gfun1-2.0_dp*step_size*bb-gfun0)/(3.0_dp*step_size*step_size)

                                 IF (ABS(gfun1-2.0_dp*step_size*bb-gfun0) .LT. num_threshold) THEN
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, '(T3,A,1X,E17.7)') &
                                          "Numerical accuracy is too low to observe cubic behavior", &
                                          ABS(gfun1-2.0_dp*step_size*bb-gfun0)
                                    ENDIF
                                    use_cubic_approximation = .FALSE.
                                    use_quadratic_approximation = .TRUE.
                                 ENDIF
                                 IF (ABS(gfun1) .LT. num_threshold) THEN
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, '(T3,A,1X,E17.7,A,1X,E12.3)') "Linear gradient", &
                                          ABS(gfun1), &
                                          " is smaller than the threshold", num_threshold
                                    ENDIF
                                    use_cubic_approximation = .FALSE.
                                    use_quadratic_approximation = .TRUE.
                                 ENDIF
                              ENDIF

                              ! find the step assuming cubic approximation
                              IF (use_cubic_approximation) THEN
                                 ! to obtain the minimum of the cubic function solve the quadratic equation
                                 ! 0.0*x**3 + 3.0*aa*x**2 + 2.0*bb*x + cc = 0
                                 CALL analytic_line_search(0.0_dp, 3.0_dp*aa, 2.0_dp*bb, gfun0, minima, nmins)
                                 IF (nmins .LT. 1) THEN
                                    IF (unit_nr > 0) THEN
                                       WRITE (unit_nr, '(T3,A)') &
                                          "Cubic approximation gives zero soultions! Use quadratic approximation"
                                    ENDIF
                                    use_quadratic_approximation = .TRUE.
                                    use_cubic_approximation = .TRUE.
                                 ELSE
                                    step_size = minima(1)
                                    IF (nmins .GT. 1) THEN
                                       IF (unit_nr > 0) THEN
                                          WRITE (unit_nr, '(T3,A)') &
                                             "More than one solution found! Use quadratic approximation"
                                       ENDIF
                                       use_quadratic_approximation = .TRUE.
                                       use_cubic_approximation = .TRUE.
                                    ENDIF
                                 ENDIF
                              ENDIF

                              IF (use_quadratic_approximation) THEN ! use quadratic approximation
                                 IF (unit_nr > 0) THEN
                                    WRITE (unit_nr, '(T3,A)') "Use quadratic approximation"
                                 ENDIF
                                 step_size = (step_size_quadratic_approx+step_size_quadratic_approx2)*0.5_dp
                              ENDIF

                              ! one more check on the step size
                              IF (step_size .LT. 0.0_dp) THEN
                                 CALL cp__b("almo_scf_optimizer.F",3452,"Negative step proposed")
                              ENDIF

                              CALL cp_dbcsr_copy(almo_scf_env%matrix_k_blk(ispin), &
                                                 matrix_k_central)
                              CALL cp_dbcsr_add(almo_scf_env%matrix_k_blk(ispin), &
                                                step, 1.0_dp, step_size)
                              CALL cp_dbcsr_copy(matrix_k_central, &
                                                 almo_scf_env%matrix_k_blk(ispin))
                              line_search = .FALSE.

                           ELSE

                              IF (md_in_k_space) THEN

                                 ! update velocities v(i) = v(i-1) + 0.5*dT*(a(i-1) + a(i))
                                 IF (iteration .NE. 0) THEN
                                    CALL cp_dbcsr_add(velocity, &
                                                      step, 1.0_dp, 0.5_dp*time_step)
                                    CALL cp_dbcsr_add(velocity, &
                                                      prev_step, 1.0_dp, 0.5_dp*time_step)
                                 ENDIF
                                 kin_energy = cp_dbcsr_frobenius_norm(velocity)
                                 kin_energy = 0.5_dp*kin_energy*kin_energy

                                 ! update positions k(i) = k(i-1) + dT*v(i-1) + 0.5*dT*dT*a(i-1)
                                 CALL cp_dbcsr_add(almo_scf_env%matrix_k_blk(ispin), &
                                                   velocity, 1.0_dp, time_step)
                                 CALL cp_dbcsr_add(almo_scf_env%matrix_k_blk(ispin), &
                                                   step, 1.0_dp, 0.5_dp*time_step*time_step)

                              ELSE

                                 IF (reset_step_size) THEN
                                    step_size = almo_scf_env%opt_k_trial_step_size
                                    reset_step_size = .FALSE.
                                 ELSE
                                    step_size = step_size*almo_scf_env%opt_k_trial_step_size_multiplier
                                 ENDIF
                                 CALL cp_dbcsr_copy(almo_scf_env%matrix_k_blk(ispin), &
                                                    matrix_k_central)
                                 CALL cp_dbcsr_add(almo_scf_env%matrix_k_blk(ispin), &
                                                   step, 1.0_dp, step_size)
                                 line_search = .TRUE.
                              ENDIF

                           ENDIF

                        ENDIF ! .NOT.prepare_to_exit

                        ! print the status of the optimization
                        t2a = m_walltime()
                        IF (unit_nr > 0) THEN
                           IF (md_in_k_space) THEN
                              WRITE (unit_nr, '(T6,A,1X,I5,1X,E12.3,E16.7,F15.9,F15.9,F15.9,E12.3,F15.9,F15.9,F8.3)') &
                                 "K iter CG", iteration, time_step, time_step*iteration, &
                                 energy_correction(ispin), obj_function, delta_obj_function, grad_norm, &
                                 kin_energy, kin_energy+obj_function, beta
                           ELSE
                              IF (line_search .OR. prepare_to_exit) THEN
                                 WRITE (unit_nr, '(T6,A,1X,I3,1X,E12.3,F16.10,F16.10,E12.3,E12.3,E12.3,F8.3,F8.3,F10.3)') &
                                    "K iter CG", iteration, step_size, &
                                    energy_correction(ispin), delta_obj_function, grad_norm, &
                                    gfun0, line_search_error, beta, conjugacy_error, t2a-t1a
                                 !(flop1+flop2)/(1.0E6_dp*(t2-t1))
                              ELSE
                                 WRITE (unit_nr, '(T6,A,1X,I3,1X,E12.3,F16.10,F16.10,E12.3,E12.3,E12.3,F8.3,F8.3,F10.3)') &
                                    "K iter LS", iteration, step_size, &
                                    energy_correction(ispin), delta_obj_function, grad_norm, &
                                    gfun1, line_search_error, beta, conjugacy_error, t2a-t1a
                                 !(flop1+flop2)/(1.0E6_dp*(t2-t1))
                              ENDIF
                           ENDIF
                           CALL m_flush(unit_nr)
                        ENDIF
                        t1a = m_walltime()

                     ELSE ! opt_k_max_iter .eq. 0
                        prepare_to_exit = .TRUE.
                     ENDIF ! opt_k_max_iter .ne. 0

                     IF (.NOT. line_search) iteration = iteration+1

                     IF (prepare_to_exit) EXIT

                  ENDDO ! end iterations on K

                  IF (converged .OR. (outer_opt_k_iteration .GE. outer_opt_k_max_iter)) THEN
                     outer_opt_k_prepare_to_exit = .TRUE.
                  ENDIF

                  IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN

                     IF (unit_nr > 0) THEN
                        WRITE (unit_nr, *) "Updating ALMO virtuals"
                     ENDIF

                     CALL timeset('k_opt_v0_update', handle8)

                     ! update retained ALMO virtuals to restart the cg iterations
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            almo_scf_env%matrix_v_disc_blk(ispin), &
                                            almo_scf_env%matrix_k_blk(ispin), &
                                            0.0_dp, vr_fixed, &
                                            filter_eps=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_add(vr_fixed, almo_scf_env%matrix_v_blk(ispin), &
                                       +1.0_dp, +1.0_dp)

                     ! update discarded ALMO virtuals to restart the cg iterations
                     CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                                            almo_scf_env%matrix_v_blk(ispin), &
                                            almo_scf_env%matrix_k_blk(ispin), &
                                            0.0_dp, vd_fixed, &
                                            filter_eps=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_add(vd_fixed, almo_scf_env%matrix_v_disc_blk(ispin), &
                                       -1.0_dp, +1.0_dp)

                     ! orthogonalize new orbitals on fragments
                     CALL get_overlap(bra=vr_fixed, &
                                      ket=vr_fixed, &
                                      overlap=k_vr_index_down, &
                                      metric=almo_scf_env%matrix_s_blk(1), &
                                      retain_overlap_sparsity=.FALSE., &
                                      eps_filter=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_init(vr_index_sqrt_inv)
                     CALL cp_dbcsr_create(vr_index_sqrt_inv, template=k_vr_index_down, &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL cp_dbcsr_init(vr_index_sqrt)
                     CALL cp_dbcsr_create(vr_index_sqrt, template=k_vr_index_down, &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL matrix_sqrt_Newton_Schulz(vr_index_sqrt, &
                                                    vr_index_sqrt_inv, &
                                                    k_vr_index_down, &
                                                    threshold=almo_scf_env%eps_filter, &
                                                    order=almo_scf_env%order_lanczos, &
                                                    eps_lanczos=almo_scf_env%eps_lanczos, &
                                                    max_iter_lanczos=almo_scf_env%max_iter_lanczos)
                     IF (safe_mode) THEN
                        CALL cp_dbcsr_init(matrix_tmp1)
                        CALL cp_dbcsr_create(matrix_tmp1, template=k_vr_index_down, &
                                             matrix_type=dbcsr_type_no_symmetry)
                        CALL cp_dbcsr_init(matrix_tmp2)
                        CALL cp_dbcsr_create(matrix_tmp2, template=k_vr_index_down, &
                                             matrix_type=dbcsr_type_no_symmetry)

                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vr_index_sqrt_inv, &
                                               k_vr_index_down, &
                                               0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                               vr_index_sqrt_inv, &
                                               0.0_dp, matrix_tmp2, filter_eps=almo_scf_env%eps_filter)

                        frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
                        CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
                        frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
                        IF (unit_nr > 0) THEN
                           WRITE (unit_nr, *) "Error for (inv(sqrt(SIGVV))*SIGVV*inv(sqrt(SIGVV))-I)", &
                              frob_matrix/frob_matrix_base
                        ENDIF

                        CALL cp_dbcsr_release(matrix_tmp1)
                        CALL cp_dbcsr_release(matrix_tmp2)
                     ENDIF
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            vr_fixed, &
                                            vr_index_sqrt_inv, &
                                            0.0_dp, almo_scf_env%matrix_v_blk(ispin), &
                                            filter_eps=almo_scf_env%eps_filter)

                     CALL get_overlap(bra=vd_fixed, &
                                      ket=vd_fixed, &
                                      overlap=k_vd_index_down, &
                                      metric=almo_scf_env%matrix_s_blk(1), &
                                      retain_overlap_sparsity=.FALSE., &
                                      eps_filter=almo_scf_env%eps_filter)
                     CALL cp_dbcsr_init(vd_index_sqrt_inv)
                     CALL cp_dbcsr_create(vd_index_sqrt_inv, template=k_vd_index_down, &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL cp_dbcsr_init(vd_index_sqrt)
                     CALL cp_dbcsr_create(vd_index_sqrt, template=k_vd_index_down, &
                                          matrix_type=dbcsr_type_no_symmetry)
                     CALL matrix_sqrt_Newton_Schulz(vd_index_sqrt, &
                                                    vd_index_sqrt_inv, &
                                                    k_vd_index_down, &
                                                    threshold=almo_scf_env%eps_filter, &
                                                    order=almo_scf_env%order_lanczos, &
                                                    eps_lanczos=almo_scf_env%eps_lanczos, &
                                                    max_iter_lanczos=almo_scf_env%max_iter_lanczos)
                     IF (safe_mode) THEN
                        CALL cp_dbcsr_init(matrix_tmp1)
                        CALL cp_dbcsr_create(matrix_tmp1, template=k_vd_index_down, &
                                             matrix_type=dbcsr_type_no_symmetry)
                        CALL cp_dbcsr_init(matrix_tmp2)
                        CALL cp_dbcsr_create(matrix_tmp2, template=k_vd_index_down, &
                                             matrix_type=dbcsr_type_no_symmetry)

                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, vd_index_sqrt_inv, &
                                               k_vd_index_down, &
                                               0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                        CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, &
                                               vd_index_sqrt_inv, &
                                               0.0_dp, matrix_tmp2, filter_eps=almo_scf_env%eps_filter)

                        frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
                        CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
                        frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
                        IF (unit_nr > 0) THEN
                           WRITE (unit_nr, *) "Error for (inv(sqrt(SIGVV))*SIGVV*inv(sqrt(SIGVV))-I)", &
                              frob_matrix/frob_matrix_base
                        ENDIF

                        CALL cp_dbcsr_release(matrix_tmp1)
                        CALL cp_dbcsr_release(matrix_tmp2)
                     ENDIF
                     CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                                            vd_fixed, &
                                            vd_index_sqrt_inv, &
                                            0.0_dp, almo_scf_env%matrix_v_disc_blk(ispin), &
                                            filter_eps=almo_scf_env%eps_filter)

                     CALL cp_dbcsr_release(vr_index_sqrt_inv)
                     CALL cp_dbcsr_release(vr_index_sqrt)
                     CALL cp_dbcsr_release(vd_index_sqrt_inv)
                     CALL cp_dbcsr_release(vd_index_sqrt)

                     CALL timestop(handle8)

                  ENDIF ! ne.virt_full

                  ! RZK-warning released outside the outer loop
                  CALL cp_dbcsr_release(sigma_vv_sqrt)
                  CALL cp_dbcsr_release(sigma_vv_sqrt_inv)
                  IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
                     CALL cp_dbcsr_release(k_vr_index_down)
                     CALL cp_dbcsr_release(k_vd_index_down)
                     !CALL cp_dbcsr_release(k_vd_index_up)
                     CALL cp_dbcsr_release(matrix_k_central)
                     CALL cp_dbcsr_release(vr_fixed)
                     CALL cp_dbcsr_release(vd_fixed)
                     CALL cp_dbcsr_release(grad)
                     CALL cp_dbcsr_release(prec)
                     CALL cp_dbcsr_release(prev_grad)
                     CALL cp_dbcsr_release(tmp3_vd_vr)
                     CALL cp_dbcsr_release(tmp1_n_vr)
                     CALL cp_dbcsr_release(tmp_k_blk)
                     CALL cp_dbcsr_release(t_curr)
                     CALL cp_dbcsr_release(sigma_oo_curr)
                     CALL cp_dbcsr_release(sigma_oo_curr_inv)
                     CALL cp_dbcsr_release(step)
                     CALL cp_dbcsr_release(tmp2_n_o)
                     CALL cp_dbcsr_release(tmp4_o_vr)
                     CALL cp_dbcsr_release(prev_step)
                     CALL cp_dbcsr_release(prev_minus_prec_grad)
                     IF (md_in_k_space) THEN
                        CALL cp_dbcsr_release(velocity)
                     ENDIF

                  ENDIF

                  outer_opt_k_iteration = outer_opt_k_iteration+1
                  IF (outer_opt_k_prepare_to_exit) EXIT

               ENDDO ! outer loop for k

            ENDDO ! ispin

            ! RZK-warning update mo orbitals

         ELSE ! virtual orbitals might not be available use projected AOs

            ! compute sqrt(S) and inv(sqrt(S))
            ! RZK-warning - remove this sqrt(S) and inv(sqrt(S))
            ! ideally ALMO scf should use sigma and sigma_inv in
            ! the tensor_up_down representation
            IF (.NOT. almo_scf_env%s_sqrt_done) THEN

               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "sqrt and inv(sqrt) of AO overlap matrix"
               ENDIF
               CALL cp_dbcsr_init(almo_scf_env%matrix_s_sqrt(1))
               CALL cp_dbcsr_init(almo_scf_env%matrix_s_sqrt_inv(1))
               CALL cp_dbcsr_create(almo_scf_env%matrix_s_sqrt(1), &
                                    template=almo_scf_env%matrix_s(1), &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_create(almo_scf_env%matrix_s_sqrt_inv(1), &
                                    template=almo_scf_env%matrix_s(1), &
                                    matrix_type=dbcsr_type_no_symmetry)

               CALL matrix_sqrt_Newton_Schulz(almo_scf_env%matrix_s_sqrt(1), &
                                              almo_scf_env%matrix_s_sqrt_inv(1), &
                                              almo_scf_env%matrix_s(1), &
                                              threshold=almo_scf_env%eps_filter, &
                                              order=almo_scf_env%order_lanczos, &
                                              eps_lanczos=almo_scf_env%eps_lanczos, &
                                              max_iter_lanczos=almo_scf_env%max_iter_lanczos)

               IF (safe_mode) THEN
                  CALL cp_dbcsr_init(matrix_tmp1)
                  CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_s(1), &
                                       matrix_type=dbcsr_type_no_symmetry)
                  CALL cp_dbcsr_init(matrix_tmp2)
                  CALL cp_dbcsr_create(matrix_tmp2, template=almo_scf_env%matrix_s(1), &
                                       matrix_type=dbcsr_type_no_symmetry)

                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_sqrt_inv(1), &
                                         almo_scf_env%matrix_s(1), &
                                         0.0_dp, matrix_tmp1, filter_eps=almo_scf_env%eps_filter)
                  CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_tmp1, almo_scf_env%matrix_s_sqrt_inv(1), &
                                         0.0_dp, matrix_tmp2, filter_eps=almo_scf_env%eps_filter)

                  frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp2)
                  CALL cp_dbcsr_add_on_diag(matrix_tmp2, -1.0_dp)
                  frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp2)
                  IF (unit_nr > 0) THEN
                     WRITE (unit_nr, *) "Error for (inv(sqrt(S))*S*inv(sqrt(S))-I)", frob_matrix/frob_matrix_base
                  ENDIF

                  CALL cp_dbcsr_release(matrix_tmp1)
                  CALL cp_dbcsr_release(matrix_tmp2)
               ENDIF

               almo_scf_env%s_sqrt_done = .TRUE.

            ENDIF

            DO ispin = 1, nspin

               CALL ct_step_env_init(ct_step_env)
               CALL ct_step_env_set(ct_step_env, &
                                    para_env=almo_scf_env%para_env, &
                                    blacs_env=almo_scf_env%blacs_env, &
                                    use_occ_orbs=.TRUE., &
                                    use_virt_orbs=almo_scf_env%deloc_cayley_use_virt_orbs, &
                                    occ_orbs_orthogonal=.FALSE., &
                                    virt_orbs_orthogonal=almo_scf_env%orthogonal_basis, &
                                    tensor_type=almo_scf_env%deloc_cayley_tensor_type, &
                                    neglect_quadratic_term=almo_scf_env%deloc_cayley_linear, &
                                    calculate_energy_corr=.TRUE., &
                                    update_p=.TRUE., &
                                    update_q=.FALSE., &
                                    pp_preconditioner_full=almo_scf_env%deloc_cayley_occ_precond, &
                                    qq_preconditioner_full=almo_scf_env%deloc_cayley_vir_precond, &
                                    eps_convergence=almo_scf_env%deloc_cayley_eps_convergence, &
                                    eps_filter=almo_scf_env%eps_filter, &
                                    !nspins=almo_scf_env%nspins,&
                                    q_index_up=almo_scf_env%matrix_s_sqrt_inv(1), &
                                    q_index_down=almo_scf_env%matrix_s_sqrt(1), &
                                    p_index_up=almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                                    p_index_down=almo_scf_env%matrix_sigma_sqrt(ispin), &
                                    matrix_ks=almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                    matrix_p=almo_scf_env%matrix_p(ispin), &
                                    matrix_qp_template=almo_scf_env%matrix_t(ispin), &
                                    matrix_pq_template=almo_scf_env%matrix_t_tr(ispin), &
                                    matrix_t=almo_scf_env%matrix_t(ispin), &
                                    conjugator=almo_scf_env%deloc_cayley_conjugator, &
                                    max_iter=almo_scf_env%deloc_cayley_max_iter)

               ! perform calculations
               CALL ct_step_execute(ct_step_env)

               ! for now we do not need the new set of orbitals
               ! just get the energy correction
               CALL ct_step_env_get(ct_step_env, &
                                    energy_correction=energy_correction(ispin))
               !copy_da_energy_matrix=matrix_eda(ispin),&
               !copy_da_charge_matrix=matrix_cta(ispin),&

               CALL ct_step_env_clean(ct_step_env)

            ENDDO

            energy_correction(1) = energy_correction(1)*spin_factor

         ENDIF

         ! print the energy correction and exit
         DO ispin = 1, nspin

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *)
               WRITE (unit_nr, '(T2,A,I6,F20.9)') "ECORR", ispin, &
                  energy_correction(ispin)
               WRITE (unit_nr, *)
            ENDIF
            energy_correction_final = energy_correction_final+energy_correction(ispin)

            !!! print out the results of decomposition analysis
            !!IF (unit_nr>0) THEN
            !!   WRITE(unit_nr,*)
            !!   WRITE(unit_nr,'(T2,A)') "ENERGY DECOMPOSITION"
            !!ENDIF
            !!CALL cp_dbcsr_print_block_sum(eda_matrix(ispin))
            !!IF (unit_nr>0) THEN
            !!   WRITE(unit_nr,*)
            !!   WRITE(unit_nr,'(T2,A)') "CHARGE DECOMPOSITION"
            !!ENDIF
            !!CALL cp_dbcsr_print_block_sum(cta_matrix(ispin))

            ! obtain density matrix from updated MOs
            ! RZK-later sigma and sigma_inv are lost here
            CALL almo_scf_t_to_p(t=almo_scf_env%matrix_t(ispin), &
                                 p=almo_scf_env%matrix_p(ispin), &
                                 eps_filter=almo_scf_env%eps_filter, &
                                 orthog_orbs=.FALSE., &
                                 s=almo_scf_env%matrix_s(1), &
                                 sigma=almo_scf_env%matrix_sigma(ispin), &
                                 sigma_inv=almo_scf_env%matrix_sigma_inv(ispin))

            IF (almo_scf_env%nspins == 1) &
               CALL cp_dbcsr_scale(almo_scf_env%matrix_p(ispin), &
                                   spin_factor)

         ENDDO

      CASE (dm_ls_step)

         ! compute the inverse of S
         IF (.NOT. almo_scf_env%s_inv_done) THEN
            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *) "Inverting AO overlap matrix"
            ENDIF
            CALL cp_dbcsr_init(almo_scf_env%matrix_s_inv(1))
            CALL cp_dbcsr_create(almo_scf_env%matrix_s_inv(1), &
                                 template=almo_scf_env%matrix_s(1), &
                                 matrix_type=dbcsr_type_no_symmetry)
            IF (.NOT. almo_scf_env%s_sqrt_done) THEN
               CALL invert_Hotelling(almo_scf_env%matrix_s_inv(1), &
                                     almo_scf_env%matrix_s(1), &
                                     threshold=almo_scf_env%eps_filter)
            ELSE
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_sqrt_inv(1), &
                                      almo_scf_env%matrix_s_sqrt_inv(1), &
                                      0.0_dp, almo_scf_env%matrix_s_inv(1), &
                                      filter_eps=almo_scf_env%eps_filter)
            ENDIF

            IF (safe_mode) THEN
               CALL cp_dbcsr_init(matrix_tmp1)
               CALL cp_dbcsr_create(matrix_tmp1, template=almo_scf_env%matrix_s(1), &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, almo_scf_env%matrix_s_inv(1), &
                                      almo_scf_env%matrix_s(1), &
                                      0.0_dp, matrix_tmp1, &
                                      filter_eps=almo_scf_env%eps_filter)
               frob_matrix_base = cp_dbcsr_frobenius_norm(matrix_tmp1)
               CALL cp_dbcsr_add_on_diag(matrix_tmp1, -1.0_dp)
               frob_matrix = cp_dbcsr_frobenius_norm(matrix_tmp1)
               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) "Error for (inv(S)*S-I)", &
                     frob_matrix/frob_matrix_base
               ENDIF
               CALL cp_dbcsr_release(matrix_tmp1)
            ENDIF

            almo_scf_env%s_inv_done = .TRUE.

         ENDIF

         DO ispin = 1, nspin
            ! RZK-warning the preconditioner is very important
            !       IF (.FALSE.) THEN
            !           CALL apply_matrix_preconditioner(almo_scf_env%matrix_ks(ispin),&
            !                   "forward",almo_scf_env%matrix_s_blk_sqrt(1),&
            !                   almo_scf_env%matrix_s_blk_sqrt_inv(1))
            !       ENDIF
            !CALL cp_dbcsr_filter(almo_scf_env%matrix_ks(ispin),&
            !         almo_scf_env%eps_filter)
         ENDDO

         ALLOCATE (matrix_p_almo_scf_converged(nspin))
         DO ispin = 1, nspin
            CALL cp_dbcsr_init(matrix_p_almo_scf_converged(ispin))
            CALL cp_dbcsr_create(matrix_p_almo_scf_converged(ispin), &
                                 template=almo_scf_env%matrix_p(ispin))
            CALL cp_dbcsr_copy(matrix_p_almo_scf_converged(ispin), &
                               almo_scf_env%matrix_p(ispin))
         ENDDO

         ! update the density matrix
         DO ispin = 1, nspin

            nelectron_spin_real(1) = almo_scf_env%nelectrons_spin(ispin)
            IF (almo_scf_env%nspins == 1) &
               nelectron_spin_real(1) = nelectron_spin_real(1)/2

            local_mu(1) = SUM(almo_scf_env%mu_of_domain(:, ispin))/almo_scf_env%ndomains
            fake(1) = 123523

            ! RZK UPDATE! the update algorithm is removed because
            ! RZK UPDATE! it requires updating core LS_SCF routines
            ! RZK UPDATE! (the code exists in the CVS version)
            CALL cp__b("almo_scf_optimizer.F",3943,"CVS only: density_matrix_sign has not been updated in SVN")
            ! RZK UPDATE!CALL density_matrix_sign(almo_scf_env%matrix_p(ispin),&
            ! RZK UPDATE!                     local_mu,&
            ! RZK UPDATE!                     almo_scf_env%fixed_mu,&
            ! RZK UPDATE!                     almo_scf_env%matrix_ks_almo_scf_converged(ispin),&
            ! RZK UPDATE!                     almo_scf_env%matrix_s(1), &
            ! RZK UPDATE!                     almo_scf_env%matrix_s_inv(1), &
            ! RZK UPDATE!                     nelectron_spin_real,&
            ! RZK UPDATE!                     almo_scf_env%eps_filter,&
            ! RZK UPDATE!                     fake)
            ! RZK UPDATE!
            almo_scf_env%mu = local_mu(1)

            !IF (almo_scf_env%has_s_preconditioner) THEN
            !    CALL apply_matrix_preconditioner(&
            !             almo_scf_env%matrix_p_blk(ispin),&
            !             "forward",almo_scf_env%matrix_s_blk_sqrt(1),&
            !             almo_scf_env%matrix_s_blk_sqrt_inv(1))
            !ENDIF
            !CALL cp_dbcsr_filter(almo_scf_env%matrix_p(ispin),&
            !        almo_scf_env%eps_filter)

            IF (almo_scf_env%nspins == 1) &
               CALL cp_dbcsr_scale(almo_scf_env%matrix_p(ispin), &
                                   spin_factor)

            !CALL cp_dbcsr_trace(almo_scf_env%matrix_ks_almo_scf_converged(ispin),&
            !  almo_scf_env%matrix_p(ispin),&
            !  energy_correction(ispin))
            !IF (unit_nr>0) THEN
            !   WRITE(unit_nr,*)
            !   WRITE(unit_nr,'(T2,A,I6,F20.9)') "EFAKE",ispin,&
            !           energy_correction(ispin)
            !   WRITE(unit_nr,*)
            !ENDIF
            CALL cp_dbcsr_add(matrix_p_almo_scf_converged(ispin), &
                              almo_scf_env%matrix_p(ispin), -1.0_dp, 1.0_dp)
            CALL cp_dbcsr_trace(almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                matrix_p_almo_scf_converged(ispin), &
                                energy_correction(ispin))

            energy_correction_final = energy_correction_final+energy_correction(ispin)

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, *)
               WRITE (unit_nr, '(T2,A,I6,F20.9)') "ECORR", ispin, &
                  energy_correction(ispin)
               WRITE (unit_nr, *)
            ENDIF

         ENDDO

         DO ispin = 1, nspin
            CALL cp_dbcsr_release(matrix_p_almo_scf_converged(ispin))
         ENDDO
         DEALLOCATE (matrix_p_almo_scf_converged)

      END SELECT ! algorithm selection

      t2 = m_walltime()

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T2,A,F18.9,F18.9,F18.9,F12.6)') "ETOT", &
            almo_scf_env%almo_scf_energy, &
            energy_correction_final, &
            almo_scf_env%almo_scf_energy+energy_correction_final, &
            t2-t1
         WRITE (unit_nr, *)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE harris_foulkes_correction

! **************************************************************************************************
!> \brief Computes a diagonal preconditioner for the cg optimization of k matrix
!> \param prec ...
!> \param vd_prop ...
!> \param f ...
!> \param x ...
!> \param oo_inv_x_tr ...
!> \param s ...
!> \param grad ...
!> \param vd_blk ...
!> \param t ...
!> \param template_vd_vd_blk ...
!> \param template_vr_vr_blk ...
!> \param template_n_vr ...
!> \param spin_factor ...
!> \param eps_filter ...
!> \par History
!>       2011.09 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE opt_k_create_preconditioner(prec, vd_prop, f, x, oo_inv_x_tr, s, grad, &
                                          vd_blk, t, template_vd_vd_blk, template_vr_vr_blk, template_n_vr, &
                                          spin_factor, eps_filter)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: prec
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: vd_prop, f, x, oo_inv_x_tr, s, grad, &
                                                            vd_blk, t, template_vd_vd_blk, &
                                                            template_vr_vr_blk, template_n_vr
      REAL(KIND=dp), INTENT(IN)                          :: spin_factor, eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'opt_k_create_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, p_nrows, q_nrows
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: p_diagonal, q_diagonal
      TYPE(cp_dbcsr_type)                                :: pp_diag, qq_diag, t1, t2, tmp, &
                                                            tmp1_n_vr, tmp2_n_vr, tmp_n_vd, &
                                                            tmp_vd_vd_blk, tmp_vr_vr_blk

! init diag blocks outside
! init diag blocks otside
!INTEGER                                  :: iblock_row, iblock_col,&
!                                            nblkrows_tot, nblkcols_tot
!REAL(KIND=dp), DIMENSION(:, :), POINTER  :: p_new_block
!INTEGER                                  :: mynode, hold, row, col

      CALL timeset(routineN, handle)

      ! initialize a matrix to 1.0
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=prec)
      ! in order to use cp_dbcsr_set matrix blocks must exist
      CALL cp_dbcsr_copy(tmp, prec)
      CALL cp_dbcsr_set(tmp, 1.0_dp)

      ! compute qq = (Vd^tr)*F*Vd
      CALL cp_dbcsr_init(tmp_n_vd)
      CALL cp_dbcsr_create(tmp_n_vd, template=vd_prop)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, f, vd_prop, &
                             0.0_dp, tmp_n_vd, filter_eps=eps_filter)
      CALL cp_dbcsr_init(tmp_vd_vd_blk)
      CALL cp_dbcsr_create(tmp_vd_vd_blk, &
                           template=template_vd_vd_blk)
      CALL cp_dbcsr_copy(tmp_vd_vd_blk, template_vd_vd_blk)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, vd_prop, tmp_n_vd, &
                             0.0_dp, tmp_vd_vd_blk, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      ! copy diagonal elements of the result into rows of a matrix
      CALL cp_dbcsr_get_info(tmp_vd_vd_blk, nfullrows_total=q_nrows)
      ALLOCATE (q_diagonal(q_nrows))
      CALL cp_dbcsr_get_diag(tmp_vd_vd_blk, q_diagonal)
      CALL cp_dbcsr_init(qq_diag)
      CALL cp_dbcsr_create(qq_diag, &
                           template=template_vd_vd_blk)
      CALL cp_dbcsr_add_on_diag(qq_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(qq_diag, q_diagonal)
      CALL cp_dbcsr_init(t1)
      CALL cp_dbcsr_create(t1, template=prec)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qq_diag, tmp, &
                             0.0_dp, t1, filter_eps=eps_filter)

      ! compute pp = X*sigma_oo_inv*X^tr
      CALL cp_dbcsr_init(tmp_vr_vr_blk)
      CALL cp_dbcsr_create(tmp_vr_vr_blk, template=template_vr_vr_blk)
      CALL cp_dbcsr_copy(tmp_vr_vr_blk, template_vr_vr_blk)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, x, oo_inv_x_tr, &
                             0.0_dp, tmp_vr_vr_blk, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      ! copy diagonal elements of the result into cols of a matrix
      CALL cp_dbcsr_get_info(tmp_vr_vr_blk, nfullrows_total=p_nrows)
      ALLOCATE (p_diagonal(p_nrows))
      CALL cp_dbcsr_get_diag(tmp_vr_vr_blk, p_diagonal)
      CALL cp_dbcsr_init(pp_diag)
      CALL cp_dbcsr_create(pp_diag, template=template_vr_vr_blk)
      CALL cp_dbcsr_add_on_diag(pp_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(pp_diag, p_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_init(t2)
      CALL cp_dbcsr_create(t2, template=prec)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp, pp_diag, &
                             0.0_dp, t2, filter_eps=eps_filter)

      CALL cp_dbcsr_hadamard_product(t1, t2, prec)

      ! compute qq = (Vd^tr)*S*Vd
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, s, vd_prop, &
                             0.0_dp, tmp_n_vd, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, vd_prop, tmp_n_vd, &
                             0.0_dp, tmp_vd_vd_blk, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      ! copy diagonal elements of the result into rows of a matrix
      CALL cp_dbcsr_get_diag(tmp_vd_vd_blk, q_diagonal)
      CALL cp_dbcsr_add_on_diag(qq_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(qq_diag, q_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, qq_diag, tmp, &
                             0.0_dp, t1, filter_eps=eps_filter)

      ! compute pp = X*sig_oo_inv*(T^tr)*F*T*sig_oo_inv*(X^tr)
      CALL cp_dbcsr_init(tmp1_n_vr)
      CALL cp_dbcsr_create(tmp1_n_vr, template=template_n_vr)
      CALL cp_dbcsr_init(tmp2_n_vr)
      CALL cp_dbcsr_create(tmp2_n_vr, template=template_n_vr)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, t, oo_inv_x_tr, &
                             0.0_dp, tmp1_n_vr, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, f, tmp1_n_vr, &
                             0.0_dp, tmp2_n_vr, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, tmp1_n_vr, tmp2_n_vr, &
                             0.0_dp, tmp_vr_vr_blk, &
                             retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      ! copy diagonal elements of the result into cols of a matrix
      CALL cp_dbcsr_get_diag(tmp_vr_vr_blk, p_diagonal)
      CALL cp_dbcsr_add_on_diag(pp_diag, 1.0_dp)
      CALL cp_dbcsr_set_diag(pp_diag, p_diagonal)
      CALL cp_dbcsr_set(tmp, 1.0_dp)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp, pp_diag, &
                             0.0_dp, t2, filter_eps=eps_filter)

      CALL cp_dbcsr_hadamard_product(t1, t2, tmp)
      CALL cp_dbcsr_add(prec, tmp, 1.0_dp, -1.0_dp)
      CALL cp_dbcsr_scale(prec, 2.0_dp*spin_factor)

      ! compute qp = X*sig_oo_inv*(T^tr)*S*Vd
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, s, vd_blk, &
                             0.0_dp, tmp_n_vd, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, tmp_n_vd, tmp1_n_vr, &
                             0.0_dp, tmp, retain_sparsity=.TRUE., &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_hadamard_product(grad, tmp, t1)
      ! gradient already contains 2.0*spin_factor
      CALL cp_dbcsr_scale(t1, -2.0_dp)

      CALL cp_dbcsr_add(prec, t1, 1.0_dp, 1.0_dp)

      CALL cp_dbcsr_function_of_elements(prec, dbcsr_func_inverse)
      CALL cp_dbcsr_filter(prec, eps_filter)

      DEALLOCATE (q_diagonal)
      DEALLOCATE (p_diagonal)
      CALL cp_dbcsr_release(tmp)
      CALL cp_dbcsr_release(qq_diag)
      CALL cp_dbcsr_release(t1)
      CALL cp_dbcsr_release(pp_diag)
      CALL cp_dbcsr_release(t2)
      CALL cp_dbcsr_release(tmp_n_vd)
      CALL cp_dbcsr_release(tmp_vd_vd_blk)
      CALL cp_dbcsr_release(tmp_vr_vr_blk)
      CALL cp_dbcsr_release(tmp1_n_vr)
      CALL cp_dbcsr_release(tmp2_n_vr)

      CALL timestop(handle)

   END SUBROUTINE opt_k_create_preconditioner

! **************************************************************************************************
!> \brief Computes a block-diagonal preconditioner for the optimization of
!>        k matrix
!> \param almo_scf_env ...
!> \param vd_prop ...
!> \param oo_inv_x_tr ...
!> \param t_curr ...
!> \param ispin ...
!> \param spin_factor ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE opt_k_create_preconditioner_blk(almo_scf_env, vd_prop, oo_inv_x_tr, &
                                              t_curr, ispin, spin_factor)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: vd_prop, oo_inv_x_tr, t_curr
      INTEGER, INTENT(IN)                                :: ispin
      REAL(KIND=dp), INTENT(IN)                          :: spin_factor

      CHARACTER(len=*), PARAMETER :: routineN = 'opt_k_create_preconditioner_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: eps_filter
      TYPE(cp_dbcsr_type)                                :: opt_k_e_dd, opt_k_e_rr, s_dd_sqrt, &
                                                            s_rr_sqrt, t1, tmp, tmp1_n_vr, &
                                                            tmp2_n_vr, tmp_n_vd, tmp_vd_vd_blk, &
                                                            tmp_vr_vr_blk

! matrices that has been computed outside the routine already

      CALL timeset(routineN, handle)

      eps_filter = almo_scf_env%eps_filter

      ! compute S_qq = (Vd^tr)*S*Vd
      CALL cp_dbcsr_init(tmp_n_vd)
      CALL cp_dbcsr_create(tmp_n_vd, template=almo_scf_env%matrix_v_disc(ispin))
      CALL cp_dbcsr_init(tmp_vd_vd_blk)
      CALL cp_dbcsr_create(tmp_vd_vd_blk, &
                           template=almo_scf_env%matrix_vv_disc_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%matrix_s(1), &
                             vd_prop, &
                             0.0_dp, tmp_n_vd, filter_eps=eps_filter)
      CALL cp_dbcsr_copy(tmp_vd_vd_blk, &
                         almo_scf_env%matrix_vv_disc_blk(ispin))
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, vd_prop, tmp_n_vd, &
                             0.0_dp, tmp_vd_vd_blk, &
                             retain_sparsity=.TRUE.)

      CALL cp_dbcsr_init(s_dd_sqrt)
      CALL cp_dbcsr_create(s_dd_sqrt, &
                           template=almo_scf_env%matrix_vv_disc_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL matrix_sqrt_Newton_Schulz(s_dd_sqrt, &
                                     almo_scf_env%opt_k_t_dd(ispin), &
                                     tmp_vd_vd_blk, &
                                     threshold=eps_filter, &
                                     order=almo_scf_env%order_lanczos, &
                                     eps_lanczos=almo_scf_env%eps_lanczos, &
                                     max_iter_lanczos=almo_scf_env%max_iter_lanczos)

      ! compute F_qq = (Vd^tr)*F*Vd
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                             vd_prop, &
                             0.0_dp, tmp_n_vd, filter_eps=eps_filter)
      CALL cp_dbcsr_copy(tmp_vd_vd_blk, &
                         almo_scf_env%matrix_vv_disc_blk(ispin))
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, vd_prop, tmp_n_vd, &
                             0.0_dp, tmp_vd_vd_blk, &
                             retain_sparsity=.TRUE.)
      CALL cp_dbcsr_release(tmp_n_vd)

      ! bring to the blocked-orthogonalized basis
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             tmp_vd_vd_blk, &
                             almo_scf_env%opt_k_t_dd(ispin), &
                             0.0_dp, s_dd_sqrt, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%opt_k_t_dd(ispin), &
                             s_dd_sqrt, &
                             0.0_dp, tmp_vd_vd_blk, filter_eps=eps_filter)

      ! diagonalize the matrix
      CALL cp_dbcsr_init(opt_k_e_dd)
      CALL cp_dbcsr_create(opt_k_e_dd, &
                           template=almo_scf_env%matrix_vv_disc_blk(ispin))
      CALL cp_dbcsr_release(s_dd_sqrt)
      CALL cp_dbcsr_init(s_dd_sqrt)
      CALL cp_dbcsr_create(s_dd_sqrt, &
                           template=almo_scf_env%matrix_vv_disc_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL diagonalize_diagonal_blocks(tmp_vd_vd_blk, &
                                       s_dd_sqrt, &
                                       opt_k_e_dd)

      ! obtain the transformation matrix in the discarded subspace
      ! T = S^{-1/2}.U
      CALL cp_dbcsr_copy(tmp_vd_vd_blk, &
                         almo_scf_env%opt_k_t_dd(ispin))
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             tmp_vd_vd_blk, &
                             s_dd_sqrt, &
                             0.0_dp, almo_scf_env%opt_k_t_dd(ispin), &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(s_dd_sqrt)
      CALL cp_dbcsr_release(tmp_vd_vd_blk)

      ! copy diagonal elements of the result into rows of a matrix
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, &
                           template=almo_scf_env%matrix_k_blk_ones(ispin))
      CALL cp_dbcsr_copy(tmp, &
                         almo_scf_env%matrix_k_blk_ones(ispin))
      CALL cp_dbcsr_init(t1)
      CALL cp_dbcsr_create(t1, &
                           template=almo_scf_env%matrix_k_blk_ones(ispin))
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             opt_k_e_dd, tmp, &
                             0.0_dp, t1, filter_eps=eps_filter)
      CALL cp_dbcsr_release(opt_k_e_dd)

      ! compute S_pp = X*sigma_oo_inv*X^tr
      CALL cp_dbcsr_init(tmp_vr_vr_blk)
      CALL cp_dbcsr_create(tmp_vr_vr_blk, &
                           template=almo_scf_env%matrix_sigma_vv_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(tmp_vr_vr_blk, &
                         almo_scf_env%matrix_sigma_vv_blk(ispin))
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%matrix_x(ispin), &
                             oo_inv_x_tr, &
                             0.0_dp, tmp_vr_vr_blk, &
                             retain_sparsity=.TRUE.)

      ! obtain the orthogonalization matrix
      CALL cp_dbcsr_init(s_rr_sqrt)
      CALL cp_dbcsr_create(s_rr_sqrt, &
                           template=almo_scf_env%matrix_sigma_vv_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL matrix_sqrt_Newton_Schulz(s_rr_sqrt, &
                                     almo_scf_env%opt_k_t_rr(ispin), &
                                     tmp_vr_vr_blk, &
                                     threshold=eps_filter, &
                                     order=almo_scf_env%order_lanczos, &
                                     eps_lanczos=almo_scf_env%eps_lanczos, &
                                     max_iter_lanczos=almo_scf_env%max_iter_lanczos)

      ! compute F_pp = X*sig_oo_inv*(T^tr)*F*T*sig_oo_inv*(X^tr)
      CALL cp_dbcsr_init(tmp1_n_vr)
      CALL cp_dbcsr_create(tmp1_n_vr, &
                           template=almo_scf_env%matrix_v(ispin))
      CALL cp_dbcsr_init(tmp2_n_vr)
      CALL cp_dbcsr_create(tmp2_n_vr, &
                           template=almo_scf_env%matrix_v(ispin))
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, t_curr, oo_inv_x_tr, &
                             0.0_dp, tmp1_n_vr, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                             tmp1_n_vr, &
                             0.0_dp, tmp2_n_vr, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, tmp1_n_vr, tmp2_n_vr, &
                             0.0_dp, tmp_vr_vr_blk, &
                             retain_sparsity=.TRUE.)
      CALL cp_dbcsr_release(tmp1_n_vr)
      CALL cp_dbcsr_release(tmp2_n_vr)

      ! bring to the blocked-orthogonalized basis
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             tmp_vr_vr_blk, &
                             almo_scf_env%opt_k_t_rr(ispin), &
                             0.0_dp, s_rr_sqrt, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%opt_k_t_rr(ispin), &
                             s_rr_sqrt, &
                             0.0_dp, tmp_vr_vr_blk, filter_eps=eps_filter)

      ! diagonalize the matrix
      CALL cp_dbcsr_init(opt_k_e_rr)
      CALL cp_dbcsr_create(opt_k_e_rr, &
                           template=almo_scf_env%matrix_sigma_vv_blk(ispin))
      CALL cp_dbcsr_release(s_rr_sqrt)
      CALL cp_dbcsr_init(s_rr_sqrt)
      CALL cp_dbcsr_create(s_rr_sqrt, &
                           template=almo_scf_env%matrix_sigma_vv_blk(ispin), &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL diagonalize_diagonal_blocks(tmp_vr_vr_blk, &
                                       s_rr_sqrt, &
                                       opt_k_e_rr)

      ! obtain the transformation matrix in the retained subspace
      ! T = S^{-1/2}.U
      CALL cp_dbcsr_copy(tmp_vr_vr_blk, &
                         almo_scf_env%opt_k_t_rr(ispin))
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             tmp_vr_vr_blk, &
                             s_rr_sqrt, &
                             0.0_dp, almo_scf_env%opt_k_t_rr(ispin), &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(s_rr_sqrt)
      CALL cp_dbcsr_release(tmp_vr_vr_blk)

      ! copy diagonal elements of the result into cols of a matrix
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             tmp, opt_k_e_rr, &
                             0.0_dp, almo_scf_env%opt_k_denom(ispin), &
                             filter_eps=eps_filter)
      CALL cp_dbcsr_release(opt_k_e_rr)
      CALL cp_dbcsr_release(tmp)

      ! form the denominator matrix
      CALL cp_dbcsr_add(almo_scf_env%opt_k_denom(ispin), t1, &
                        -1.0_dp, 1.0_dp)
      CALL cp_dbcsr_release(t1)
      CALL cp_dbcsr_scale(almo_scf_env%opt_k_denom(ispin), &
                          2.0_dp*spin_factor)

      CALL cp_dbcsr_function_of_elements(almo_scf_env%opt_k_denom(ispin), &
                                         dbcsr_func_inverse)
      CALL cp_dbcsr_filter(almo_scf_env%opt_k_denom(ispin), &
                           eps_filter)

      CALL timestop(handle)

   END SUBROUTINE opt_k_create_preconditioner_blk

! **************************************************************************************************
!> \brief Applies a block-diagonal preconditioner for the optimization of
!>        k matrix (preconditioner matrices must be calculated and stored
!>        beforehand)
!> \param almo_scf_env ...
!> \param step ...
!> \param grad ...
!> \param ispin ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE opt_k_apply_preconditioner_blk(almo_scf_env, step, grad, ispin)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: step
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: grad
      INTEGER, INTENT(IN)                                :: ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'opt_k_apply_preconditioner_blk', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: eps_filter
      TYPE(cp_dbcsr_type)                                :: tmp_k

      CALL timeset(routineN, handle)

      eps_filter = almo_scf_env%eps_filter

      CALL cp_dbcsr_init(tmp_k)
      CALL cp_dbcsr_create(tmp_k, template=almo_scf_env%matrix_k_blk(ispin))

      ! transform gradient to the correct "diagonal" basis
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             grad, almo_scf_env%opt_k_t_rr(ispin), &
                             0.0_dp, tmp_k, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, &
                             almo_scf_env%opt_k_t_dd(ispin), tmp_k, &
                             0.0_dp, step, filter_eps=eps_filter)

      ! apply diagonal preconditioner
      CALL cp_dbcsr_hadamard_product(step, &
                                     almo_scf_env%opt_k_denom(ispin), tmp_k)

      ! back-transform the result to the initial basis
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, &
                             almo_scf_env%opt_k_t_dd(ispin), tmp_k, &
                             0.0_dp, step, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("N", "T", 1.0_dp, &
                             step, almo_scf_env%opt_k_t_rr(ispin), &
                             0.0_dp, tmp_k, filter_eps=eps_filter)

      CALL cp_dbcsr_copy(step, tmp_k)

      CALL cp_dbcsr_release(tmp_k)

      CALL timestop(handle)

   END SUBROUTINE opt_k_apply_preconditioner_blk

!! **************************************************************************************************
!!> \brief Reduce the number of virtual orbitals by rotating them within
!!>        a domain. The rotation is such that minimizes the frobenius norm of
!!>        the Fov domain-blocks of the discarded virtuals
!!> \par History
!!>       2011.08 created [Rustam Z Khaliullin]
!!> \author Rustam Z Khaliullin
!! **************************************************************************************************
!  SUBROUTINE truncate_subspace_v_blk(qs_env,almo_scf_env)
!
!    TYPE(qs_environment_type), POINTER       :: qs_env
!    TYPE(almo_scf_env_type)                  :: almo_scf_env
!
!    CHARACTER(len=*), PARAMETER :: routineN = 'truncate_subspace_v_blk', &
!      routineP = moduleN//':'//routineN
!
!    INTEGER                                  :: handle, ispin, iblock_row, &
!                                                iblock_col, iblock_row_size, &
!                                                iblock_col_size, retained_v, &
!                                                iteration, line_search_step, &
!                                                unit_nr, line_search_step_last
!    REAL(KIND=dp)                            :: t1, obj_function, grad_norm,&
!                                                c0, b0, a0, obj_function_new,&
!                                                t2, alpha, ff1, ff2, step1,&
!                                                step2,&
!                                                frob_matrix_base,&
!                                                frob_matrix
!    LOGICAL                                  :: safe_mode, converged, &
!                                                prepare_to_exit, failure
!    TYPE(cp_logger_type), POINTER            :: logger
!    TYPE(cp_dbcsr_type)                      :: Fon, Fov, Fov_filtered, &
!                                                temp1_oo, temp2_oo, Fov_original, &
!                                                temp0_ov, U_blk_tot, U_blk, &
!                                                grad_blk, step_blk, matrix_filter, &
!                                                v_full_new,v_full_tmp,&
!                                                matrix_sigma_vv_full,&
!                                                matrix_sigma_vv_full_sqrt,&
!                                                matrix_sigma_vv_full_sqrt_inv,&
!                                                matrix_tmp1,&
!                                                matrix_tmp2
!
!    REAL(kind=dp), DIMENSION(:, :), POINTER  :: data_p, p_new_block
!    TYPE(cp_dbcsr_iterator)                  :: iter
!
!
!REAL(kind=dp), DIMENSION(:), ALLOCATABLE     :: eigenvalues, WORK
!REAL(kind=dp), DIMENSION(:,:), ALLOCATABLE   :: data_copy, left_vectors, right_vectors
!INTEGER                                      :: LWORK, INFO
!TYPE(cp_dbcsr_type)                          :: temp_u_v_full_blk
!
!    CALL timeset(routineN,handle)
!
!    safe_mode=.TRUE.
!
!    ! get a useful output_unit
!    logger => cp_get_default_logger()
!    IF (logger%para_env%mepos==logger%para_env%source) THEN
!       unit_nr=cp_logger_get_default_unit_nr(logger,local=.TRUE.)
!    ELSE
!       unit_nr=-1
!    ENDIF
!
!    DO ispin=1,almo_scf_env%nspins
!
!       t1 = m_walltime()
!
!       !!!!!!!!!!!!!!!!!
!       ! 0. Orthogonalize virtuals
!       !    Unfortunately, we have to do it in the FULL V subspace :(
!
!       CALL cp_dbcsr_init(v_full_new)
!       CALL cp_dbcsr_create(v_full_new,&
!               template=almo_scf_env%matrix_v_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! project the occupied subspace out
!       CALL almo_scf_p_out_from_v(almo_scf_env%matrix_v_full_blk(ispin),&
!              v_full_new,almo_scf_env%matrix_ov_full(ispin),&
!              ispin,almo_scf_env)
!
!       ! init overlap and its functions
!       CALL cp_dbcsr_init(matrix_sigma_vv_full)
!       CALL cp_dbcsr_init(matrix_sigma_vv_full_sqrt)
!       CALL cp_dbcsr_init(matrix_sigma_vv_full_sqrt_inv)
!       CALL cp_dbcsr_create(matrix_sigma_vv_full,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_create(matrix_sigma_vv_full_sqrt,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_create(matrix_sigma_vv_full_sqrt_inv,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! construct VV overlap
!       CALL almo_scf_mo_to_sigma(v_full_new,&
!               matrix_sigma_vv_full,&
!               almo_scf_env%matrix_s(1),&
!               almo_scf_env%eps_filter)
!
!       IF (unit_nr>0) THEN
!          WRITE(unit_nr,*) "sqrt and inv(sqrt) of the FULL virtual MO overlap"
!       ENDIF
!
!       ! construct orthogonalization matrices
!       CALL matrix_sqrt_Newton_Schulz(matrix_sigma_vv_full_sqrt,&
!                                      matrix_sigma_vv_full_sqrt_inv,&
!                                      matrix_sigma_vv_full,&
!                                      threshold=almo_scf_env%eps_filter,&
!                                      order=almo_scf_env%order_lanczos,&
!                                      eps_lanczos=almo_scf_env%eps_lanczos,&
!                                      max_iter_lanczos=almo_scf_env%max_iter_lanczos)
!       IF (safe_mode) THEN
!          CALL cp_dbcsr_init(matrix_tmp1)
!          CALL cp_dbcsr_create(matrix_tmp1,template=matrix_sigma_vv_full,&
!                               matrix_type=dbcsr_type_no_symmetry)
!          CALL cp_dbcsr_init(matrix_tmp2)
!          CALL cp_dbcsr_create(matrix_tmp2,template=matrix_sigma_vv_full,&
!                               matrix_type=dbcsr_type_no_symmetry)
!
!          CALL cp_dbcsr_multiply("N","N",1.0_dp,matrix_sigma_vv_full_sqrt_inv,&
!                                 matrix_sigma_vv_full,&
!                                 0.0_dp,matrix_tmp1,filter_eps=almo_scf_env%eps_filter)
!          CALL cp_dbcsr_multiply("N","N",1.0_dp,matrix_tmp1,&
!                                 matrix_sigma_vv_full_sqrt_inv,&
!                                 0.0_dp,matrix_tmp2,filter_eps=almo_scf_env%eps_filter)
!
!          frob_matrix_base=cp_dbcsr_frobenius_norm(matrix_tmp2)
!          CALL cp_dbcsr_add_on_diag(matrix_tmp2,-1.0_dp)
!          frob_matrix=cp_dbcsr_frobenius_norm(matrix_tmp2)
!          IF (unit_nr>0) THEN
!             WRITE(unit_nr,*) "Error for (inv(sqrt(SIGVV))*SIGVV*inv(sqrt(SIGVV))-I)",frob_matrix/frob_matrix_base
!          ENDIF
!
!          CALL cp_dbcsr_release(matrix_tmp1)
!          CALL cp_dbcsr_release(matrix_tmp2)
!       ENDIF
!
!       ! discard unnecessary overlap functions
!       CALL cp_dbcsr_release(matrix_sigma_vv_full)
!       CALL cp_dbcsr_release(matrix_sigma_vv_full_sqrt)
!
!! this can be re-written because we have (1-P)|v>
!
!       !!!!!!!!!!!!!!!!!!!
!       ! 1. Compute F_ov
!       CALL cp_dbcsr_init(Fon)
!       CALL cp_dbcsr_create(Fon,&
!               template=almo_scf_env%matrix_v_full_blk(ispin))
!       CALL cp_dbcsr_init(Fov)
!       CALL cp_dbcsr_create(Fov,&
!               template=almo_scf_env%matrix_ov_full(ispin))
!       CALL cp_dbcsr_init(Fov_filtered)
!       CALL cp_dbcsr_create(Fov_filtered,&
!               template=almo_scf_env%matrix_ov_full(ispin))
!       CALL cp_dbcsr_init(temp1_oo)
!       CALL cp_dbcsr_create(temp1_oo,&
!               template=almo_scf_env%matrix_sigma(ispin),&
!               !matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_init(temp2_oo)
!       CALL cp_dbcsr_create(temp2_oo,&
!               template=almo_scf_env%matrix_sigma(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       CALL cp_dbcsr_multiply("T","N",1.0_dp,almo_scf_env%matrix_t_blk(ispin),&
!               almo_scf_env%matrix_ks_almo_scf_converged(ispin),&
!               0.0_dp,Fon,filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,Fon,&
!               almo_scf_env%matrix_v_full_blk(ispin),&
!               0.0_dp,Fov,filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,Fon,&
!               almo_scf_env%matrix_t_blk(ispin),&
!               0.0_dp,temp1_oo,filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,temp1_oo,&
!               almo_scf_env%matrix_sigma_inv(ispin),&
!               0.0_dp,temp2_oo,filter_eps=almo_scf_env%eps_filter)
!       CALL cp_dbcsr_release(temp1_oo)
!
!       CALL cp_dbcsr_multiply("T","N",1.0_dp,almo_scf_env%matrix_t_blk(ispin),&
!               almo_scf_env%matrix_s(1),&
!               0.0_dp,Fon,filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,Fon,&
!               almo_scf_env%matrix_v_full_blk(ispin),&
!               0.0_dp,Fov_filtered,filter_eps=almo_scf_env%eps_filter)
!       CALL cp_dbcsr_release(Fon)
!
!       CALL cp_dbcsr_multiply("N","N",-1.0_dp,temp2_oo,&
!               Fov_filtered,&
!               1.0_dp,Fov,filter_eps=almo_scf_env%eps_filter)
!       CALL cp_dbcsr_release(temp2_oo)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,almo_scf_env%matrix_sigma_inv(ispin),&
!               Fov,0.0_dp,Fov_filtered,filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,Fov_filtered,&
!               matrix_sigma_vv_full_sqrt_inv,&
!               0.0_dp,Fov,filter_eps=almo_scf_env%eps_filter)
!       !CALL cp_dbcsr_copy(Fov,Fov_filtered)
!CALL cp_dbcsr_print(Fov)
!
!       IF (safe_mode) THEN
!          CALL cp_dbcsr_init(Fov_original)
!          CALL cp_dbcsr_create(Fov_original,template=Fov)
!          CALL cp_dbcsr_copy(Fov_original,Fov)
!       ENDIF
!
!!! remove diagonal blocks
!!CALL cp_dbcsr_iterator_start(iter,Fov)
!!DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
!!
!!   CALL cp_dbcsr_iterator_next_block(iter,iblock_row,iblock_col,data_p,&
!!           row_size=iblock_row_size,col_size=iblock_col_size)
!!
!!   IF (iblock_row.eq.iblock_col) data_p(:,:)=0.0_dp
!!
!!ENDDO
!!CALL cp_dbcsr_iterator_stop(iter)
!!CALL cp_dbcsr_finalize(Fov)
!
!!! perform svd of blocks
!!!!! THIS ROUTINE WORKS ONLY ON ONE CPU AND ONLY FOR 2 MOLECULES !!!
!!CALL cp_dbcsr_init(temp_u_v_full_blk)
!!CALL cp_dbcsr_create(temp_u_v_full_blk,&
!!        template=almo_scf_env%matrix_vv_full_blk(ispin),&
!!        matrix_type=dbcsr_type_no_symmetry)
!!
!!CALL cp_dbcsr_work_create(temp_u_v_full_blk,&
!!        work_mutable=.TRUE.)
!!CALL cp_dbcsr_iterator_start(iter,Fov)
!!DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
!!
!!   CALL cp_dbcsr_iterator_next_block(iter,iblock_row,iblock_col,data_p,&
!!           row_size=iblock_row_size,col_size=iblock_col_size)
!!
!!   IF (iblock_row.ne.iblock_col) THEN
!!
!!      ! Prepare data
!!      allocate(eigenvalues(min(iblock_row_size,iblock_col_size)))
!!      allocate(data_copy(iblock_row_size,iblock_col_size))
!!      allocate(left_vectors(iblock_row_size,iblock_row_size))
!!      allocate(right_vectors(iblock_col_size,iblock_col_size))
!!      data_copy(:,:)=data_p(:,:)
!!
!!      ! Query the optimal workspace for dgesvd
!!      LWORK = -1
!!      allocate(WORK(MAX(1,LWORK)))
!!      CALL DGESVD('N','A',iblock_row_size,iblock_col_size,data_copy,&
!!              iblock_row_size,eigenvalues,left_vectors,iblock_row_size,&
!!              right_vectors,iblock_col_size,WORK,LWORK,INFO)
!!      LWORK = INT(WORK( 1 ))
!!      deallocate(WORK)
!!
!!      ! Allocate the workspace and perform svd
!!      allocate(WORK(MAX(1,LWORK)))
!!      CALL DGESVD('N','A',iblock_row_size,iblock_col_size,data_copy,&
!!              iblock_row_size,eigenvalues,left_vectors,iblock_row_size,&
!!              right_vectors,iblock_col_size,WORK,LWORK,INFO)
!!      deallocate(WORK)
!!      IF( INFO.NE.0 ) THEN
!!         CPErrorMessage(cp_failure_level,routineP,"DGESVD failed")
!!         CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
!!      END IF
!!
!!      ! copy right singular vectors into a unitary matrix
!!      NULLIFY (p_new_block)
!!      CALL cp_dbcsr_reserve_block2d(temp_u_v_full_blk,iblock_col,iblock_col,p_new_block)
!!      CPPostcondition(ASSOCIATED(p_new_block),cp_failure_level,routineP,failure)
!!      p_new_block(:,:) = right_vectors(:,:)
!!
!!      deallocate(eigenvalues)
!!      deallocate(data_copy)
!!      deallocate(left_vectors)
!!      deallocate(right_vectors)
!!
!!   ENDIF
!!ENDDO
!!CALL cp_dbcsr_iterator_stop(iter)
!!CALL cp_dbcsr_finalize(temp_u_v_full_blk)
!!!CALL cp_dbcsr_print(temp_u_v_full_blk)
!!CALL cp_dbcsr_multiply("N","T",1.0_dp,Fov,temp_u_v_full_blk,&
!!        0.0_dp,Fov_filtered,filter_eps=almo_scf_env%eps_filter)
!!
!!CALL cp_dbcsr_copy(Fov,Fov_filtered)
!!CALL cp_dbcsr_print(Fov)
!
!       !!!!!!!!!!!!!!!!!!!
!       ! 2. Initialize variables
!
!       ! temp space
!       CALL cp_dbcsr_init(temp0_ov)
!       CALL cp_dbcsr_create(temp0_ov,&
!               template=almo_scf_env%matrix_ov_full(ispin))
!
!       ! current unitary matrix
!       CALL cp_dbcsr_init(U_blk)
!       CALL cp_dbcsr_create(U_blk,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! unitary matrix accumulator
!       CALL cp_dbcsr_init(U_blk_tot)
!       CALL cp_dbcsr_create(U_blk_tot,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_add_on_diag(U_blk_tot,1.0_dp)
!
!!CALL cp_dbcsr_add_on_diag(U_blk,1.0_dp)
!!CALL cp_dbcsr_multiply("N","T",1.0_dp,U_blk,temp_u_v_full_blk,&
!!        0.0_dp,U_blk_tot,filter_eps=almo_scf_env%eps_filter)
!!
!!CALL cp_dbcsr_release(temp_u_v_full_blk)
!
!       ! init gradient
!       CALL cp_dbcsr_init(grad_blk)
!       CALL cp_dbcsr_create(grad_blk,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! init step matrix
!       CALL cp_dbcsr_init(step_blk)
!       CALL cp_dbcsr_create(step_blk,&
!               template=almo_scf_env%matrix_vv_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!
!       ! "retain discarded" filter (0.0 - retain, 1.0 - discard)
!       CALL cp_dbcsr_init(matrix_filter)
!       CALL cp_dbcsr_create(matrix_filter,&
!               template=almo_scf_env%matrix_ov_full(ispin))
!       ! copy Fov into the filter matrix temporarily
!       ! so we know which blocks contain significant elements
!       CALL cp_dbcsr_copy(matrix_filter,Fov)
!
!       ! fill out filter elements block-by-block
!       CALL cp_dbcsr_iterator_start(iter,matrix_filter)
!       DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
!
!          CALL cp_dbcsr_iterator_next_block(iter,iblock_row,iblock_col,data_p,&
!                  row_size=iblock_row_size,col_size=iblock_col_size)
!
!          retained_v=almo_scf_env%nvirt_of_domain(iblock_col,ispin)
!
!          data_p(:,1:retained_v)=0.0_dp
!          data_p(:,(retained_v+1):iblock_col_size)=1.0_dp
!
!       ENDDO
!       CALL cp_dbcsr_iterator_stop(iter)
!       CALL cp_dbcsr_finalize(matrix_filter)
!
!       ! apply the filter
!       CALL cp_dbcsr_hadamard_product(Fov,matrix_filter,Fov_filtered)
!
!       !!!!!!!!!!!!!!!!!!!!!
!       ! 3. start iterative minimization of the elements to be discarded
!       iteration=0
!       converged=.FALSE.
!       prepare_to_exit=.FALSE.
!       DO
!
!          iteration=iteration+1
!
!          !!!!!!!!!!!!!!!!!!!!!!!!!
!          ! 4. compute the gradient
!          CALL cp_dbcsr_set(grad_blk,0.0_dp)
!          ! create the diagonal blocks only
!          CALL cp_dbcsr_add_on_diag(grad_blk,1.0_dp)
!
!          CALL cp_dbcsr_multiply("T","N",2.0_dp,Fov_filtered,Fov,&
!                  0.0_dp,grad_blk,retain_sparsity=.TRUE.,&
!                  filter_eps=almo_scf_env%eps_filter)
!          CALL cp_dbcsr_multiply("T","N",-2.0_dp,Fov,Fov_filtered,&
!                  1.0_dp,grad_blk,retain_sparsity=.TRUE.,&
!                  filter_eps=almo_scf_env%eps_filter)
!
!          !!!!!!!!!!!!!!!!!!!!!!!
!          ! 5. check convergence
!          obj_function = 0.5_dp*(cp_dbcsr_frobenius_norm(Fov_filtered))**2
!          grad_norm = cp_dbcsr_frobenius_norm(grad_blk)
!          converged=(grad_norm.lt.almo_scf_env%truncate_v_eps_convergence)
!          IF (converged.OR.(iteration.ge.almo_scf_env%truncate_v_max_iter)) THEN
!             prepare_to_exit=.TRUE.
!          ENDIF
!
!          IF (.NOT.prepare_to_exit) THEN
!
!             !!!!!!!!!!!!!!!!!!!!!!!
!             ! 6. perform steps in the direction of the gradient
!             !    a. first, perform a trial step to "see" the parameters
!             !       of the parabola along the gradient:
!             !       a0 * x^2 + b0 * x + c0
!             !    b. then perform the step to the bottom of the parabola
!
!             ! get c0
!             c0 = obj_function
!             ! get b0 <= d_f/d_alpha along grad
!             !!!CALL cp_dbcsr_multiply("N","N",4.0_dp,Fov,grad_blk,&
!             !!!        0.0_dp,temp0_ov,&
!             !!!        filter_eps=almo_scf_env%eps_filter)
!             !!!CALL cp_dbcsr_trace(Fov_filtered,temp0_ov,b0,"T","N")
!
!             alpha=almo_scf_env%truncate_v_trial_step_size
!
!             line_search_step_last=3
!             DO line_search_step=1,line_search_step_last
!                CALL cp_dbcsr_copy(step_blk,grad_blk)
!                CALL cp_dbcsr_scale(step_blk,-1.0_dp*alpha)
!                CALL generator_to_unitary(step_blk,U_blk,&
!                        almo_scf_env%eps_filter)
!                CALL cp_dbcsr_multiply("N","N",1.0_dp,Fov,U_blk,0.0_dp,temp0_ov,&
!                        filter_eps=almo_scf_env%eps_filter)
!                CALL cp_dbcsr_hadamard_product(temp0_ov,matrix_filter,&
!                        Fov_filtered)
!
!                obj_function_new = 0.5_dp*(cp_dbcsr_frobenius_norm(Fov_filtered))**2
!                IF (line_search_step.eq.1) THEN
!                   ff1 = obj_function_new
!                   step1 = alpha
!                ELSE IF (line_search_step.eq.2) THEN
!                   ff2 = obj_function_new
!                   step2 = alpha
!                ENDIF
!
!                IF (unit_nr>0.AND.(line_search_step.ne.line_search_step_last)) THEN
!                   WRITE(unit_nr,'(T6,A,1X,I3,1X,F10.3,E12.3,E12.3,E12.3)') &
!                         "JOINT_SVD_lin",&
!                         iteration,&
!                         alpha,&
!                         obj_function,&
!                         obj_function_new,&
!                         obj_function_new-obj_function
!                ENDIF
!
!                IF (line_search_step.eq.1) THEN
!                   alpha=2.0_dp*alpha
!                ENDIF
!                IF (line_search_step.eq.2) THEN
!                   a0 = ((ff1-c0)/step1 - (ff2-c0)/step2) / (step1 - step2)
!                   b0 = (ff1-c0)/step1 - a0*step1
!                   ! step size in to the bottom of "the parabola"
!                   alpha=-b0/(2.0_dp*a0)
!                   ! update the default step size
!                   almo_scf_env%truncate_v_trial_step_size=alpha
!                ENDIF
!                !!!IF (line_search_step.eq.1) THEN
!                !!!   a0 = (obj_function_new - b0 * alpha - c0) / (alpha*alpha)
!                !!!   ! step size in to the bottom of "the parabola"
!                !!!   alpha=-b0/(2.0_dp*a0)
!                !!!   !IF (alpha.gt.10.0_dp) alpha=10.0_dp
!                !!!ENDIF
!
!             ENDDO
!
!             ! update Fov and U_blk_tot (use grad_blk as tmp storage)
!             CALL cp_dbcsr_copy(Fov,temp0_ov)
!             CALL cp_dbcsr_multiply("N","N",1.0_dp,U_blk_tot,U_blk,&
!                     0.0_dp,grad_blk,&
!                     filter_eps=almo_scf_env%eps_filter)
!             CALL cp_dbcsr_copy(U_blk_tot,grad_blk)
!
!          ENDIF
!
!          t2 = m_walltime()
!
!          IF (unit_nr>0) THEN
!             WRITE(unit_nr,'(T6,A,1X,I3,1X,F10.3,E12.3,E12.3,E12.3,E12.3,F10.3)') &
!                   "JOINT_SVD_itr",&
!                   iteration,&
!                   alpha,&
!                   obj_function,&
!                   obj_function_new,&
!                   obj_function_new-obj_function,&
!                   grad_norm,&
!                   t2-t1
!                   !(flop1+flop2)/(1.0E6_dp*(t2-t1))
!             CALL m_flush(unit_nr)
!          ENDIF
!
!          t1 = m_walltime()
!
!          IF (prepare_to_exit) EXIT
!
!       ENDDO ! stop iterations
!
!       IF (safe_mode) THEN
!          CALL cp_dbcsr_multiply("N","N",1.0_dp,Fov_original,&
!                  U_blk_tot,0.0_dp,temp0_ov,&
!                  filter_eps=almo_scf_env%eps_filter)
!CALL cp_dbcsr_print(temp0_ov)
!          CALL cp_dbcsr_hadamard_product(temp0_ov,matrix_filter,&
!                  Fov_filtered)
!          obj_function_new = 0.5_dp*(cp_dbcsr_frobenius_norm(Fov_filtered))**2
!
!          IF (unit_nr>0) THEN
!             WRITE(unit_nr,'(T6,A,1X,E12.3)') &
!                   "SANITY CHECK:",&
!                   obj_function_new
!             CALL m_flush(unit_nr)
!          ENDIF
!
!          CALL cp_dbcsr_release(Fov_original)
!       ENDIF
!
!       CALL cp_dbcsr_release(temp0_ov)
!       CALL cp_dbcsr_release(U_blk)
!       CALL cp_dbcsr_release(grad_blk)
!       CALL cp_dbcsr_release(step_blk)
!       CALL cp_dbcsr_release(matrix_filter)
!       CALL cp_dbcsr_release(Fov)
!       CALL cp_dbcsr_release(Fov_filtered)
!
!       ! compute rotated virtual orbitals
!       CALL cp_dbcsr_init(v_full_tmp)
!       CALL cp_dbcsr_create(v_full_tmp,&
!               template=almo_scf_env%matrix_v_full_blk(ispin),&
!               matrix_type=dbcsr_type_no_symmetry)
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!               v_full_new,&
!               matrix_sigma_vv_full_sqrt_inv,0.0_dp,v_full_tmp,&
!               filter_eps=almo_scf_env%eps_filter)
!       CALL cp_dbcsr_multiply("N","N",1.0_dp,&
!               v_full_tmp,&
!               U_blk_tot,0.0_dp,v_full_new,&
!               filter_eps=almo_scf_env%eps_filter)
!
!       CALL cp_dbcsr_release(matrix_sigma_vv_full_sqrt_inv)
!       CALL cp_dbcsr_release(v_full_tmp)
!       CALL cp_dbcsr_release(U_blk_tot)
!
!!!!! orthogonalized virtuals are not blocked
!       ! copy new virtuals into the truncated matrix
!       !CALL cp_dbcsr_work_create(almo_scf_env%matrix_v_blk(ispin),&
!       CALL cp_dbcsr_work_create(almo_scf_env%matrix_v(ispin),&
!               work_mutable=.TRUE.)
!       CALL cp_dbcsr_iterator_start(iter,v_full_new)
!       DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
!
!          CALL cp_dbcsr_iterator_next_block(iter,iblock_row,iblock_col,data_p,&
!                  row_size=iblock_row_size,col_size=iblock_col_size)
!
!          retained_v=almo_scf_env%nvirt_of_domain(iblock_col,ispin)
!
!          NULLIFY (p_new_block)
!          !CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_v_blk(ispin),&
!          CALL cp_dbcsr_reserve_block2d(almo_scf_env%matrix_v(ispin),&
!                  iblock_row,iblock_col,p_new_block)
!          CPPostcondition(ASSOCIATED(p_new_block),cp_failure_level,routineP,failure)
!          CPPrecondition(retained_v.gt.0,cp_failure_level,routineP,failure)
!          p_new_block(:,:) = data_p(:,1:retained_v)
!
!       ENDDO ! iterator
!       CALL cp_dbcsr_iterator_stop(iter)
!       !!CALL cp_dbcsr_finalize(almo_scf_env%matrix_v_blk(ispin))
!       CALL cp_dbcsr_finalize(almo_scf_env%matrix_v(ispin))
!
!       CALL cp_dbcsr_release(v_full_new)
!
!    ENDDO ! ispin
!
!    CALL timestop(handle)
!
!  END SUBROUTINE truncate_subspace_v_blk

END MODULE almo_scf_optimizer

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for all ALMO-based SCF methods
!>        RZK-warning marks unresolved issues
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf
   USE almo_scf_methods,                ONLY: almo_scf_t_blk_to_t_blk_orthonormal,&
                                              distribute_domains
   USE almo_scf_optimizer,              ONLY: almo_scf_block_diagonal,&
                                              almo_scf_xalmo_eigensolver,&
                                              almo_scf_xalmo_pcg
   USE almo_scf_qs,                     ONLY: almo_scf_construct_quencher,&
                                              almo_scf_init_qs,&
                                              matrix_almo_create,&
                                              matrix_qs_to_almo
   USE almo_scf_types,                  ONLY: &
        almo_mat_dim_aobasis, almo_mat_dim_occ, almo_mat_dim_virt, almo_mat_dim_virt_disc, &
        almo_mat_dim_virt_full, almo_max_cutoff_multiplier, almo_scf_env_type, &
        optimizer_options_type, print_optimizer_options
   USE bibliography,                    ONLY: Khaliullin2013,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add_on_diag, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_distribution, &
        cp_dbcsr_filter, cp_dbcsr_finalize, cp_dbcsr_get_info, cp_dbcsr_get_stored_coordinates, &
        cp_dbcsr_init, cp_dbcsr_nblkcols_total, cp_dbcsr_nblkrows_total, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_reserve_block2d, cp_dbcsr_set, cp_dbcsr_type, &
        cp_dbcsr_work_create, dbcsr_distribution_mp, dbcsr_mp_mynode, dbcsr_type_no_symmetry, &
        dbcsr_type_symmetric
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE domain_submatrix_methods,        ONLY: init_submatrices,&
                                              release_submatrices
   USE input_constants,                 ONLY: &
        almo_constraint_distance, almo_deloc_none, almo_deloc_scf, almo_deloc_x, &
        almo_deloc_x_then_scf, almo_deloc_xalmo_1diag, almo_deloc_xalmo_scf, almo_deloc_xalmo_x, &
        almo_deloc_xk, almo_domain_layout_atomic, almo_domain_layout_molecular, &
        almo_mat_distr_atomic, almo_mat_distr_molecular, almo_scf_diag, almo_scf_dm_sign, &
        almo_scf_pcg, cg_hager_zhang, do_bondparm_vdw, optimizer_diis, optimizer_pcg, &
        tensor_orthogonal, virt_full, virt_minimal, virt_number, virt_occ_size, &
        xalmo_case_block_diag, xalmo_case_fully_deloc, xalmo_case_normal
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE iterate_matrix,                  ONLY: invert_Hotelling,&
                                              matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE molecule_types_new,              ONLY: get_molecule_set_info,&
                                              molecule_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf'

   PUBLIC :: almo_entry_scf

   LOGICAL, PARAMETER :: debug_mode = .FALSE.
   LOGICAL, PARAMETER :: safe_mode = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief The entry point into ALMO SCF routines
!> \param qs_env - pointer to the QS environment
!> \param calc_forces - calculate forces?
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_entry_scf(qs_env, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL                                            :: calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_entry_scf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CALL timeset(routineN, handle)

      CALL cite_reference(Khaliullin2013)

      ! no forces so far
      IF (calc_forces) &
         CALL cp__b("almo_scf.F",102," No forces/gradients with ALMOs.... yet(!) ")

      ! initialize scf
      CALL almo_scf_init(qs_env, almo_scf_env)

      ! perform SCF for block diagonal ALMOs
      CALL almo_scf_main(qs_env, almo_scf_env)

      ! allow electron delocalization
      CALL almo_scf_delocalization(qs_env, almo_scf_env)

      ! return computed quantities to the qs_env
      !CALL almo_return_to_qs(qs_env,almo_scf_env)

      ! do post scf processing
      CALL almo_scf_clean_up(almo_scf_env)

      CALL timestop(handle)

   END SUBROUTINE almo_entry_scf

! **************************************************************************************************
!> \brief Initialization of the almo_scf_env_type.
!> \param qs_env ...
!> \param almo_scf_env ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_init(qs_env, almo_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: ao, handle, i, iao, idomain, ispin, &
                                                            multip, naos, natoms, ndomains, nelec, &
                                                            nelec_a, nelec_b, nmols, nspins, &
                                                            unit_nr
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)

      ! define the output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! set optimizers' types
      almo_scf_env%opt_block_diag_diis%optimizer_type = optimizer_diis
      almo_scf_env%opt_block_diag_pcg%optimizer_type = optimizer_pcg
      almo_scf_env%opt_xalmo_diis%optimizer_type = optimizer_diis
      almo_scf_env%opt_xalmo_pcg%optimizer_type = optimizer_pcg

      ! get info from the qs_env
      CALL get_qs_env(qs_env, &
                      nelectron_total=almo_scf_env%nelectrons_total, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      molecule_set=molecule_set, &
                      input=input, &
                      has_unit_metric=almo_scf_env%orthogonal_basis, &
                      para_env=almo_scf_env%para_env, &
                      blacs_env=almo_scf_env%blacs_env, &
                      nelectron_spin=almo_scf_env%nelectrons_spin)
      CALL cp_para_env_retain(almo_scf_env%para_env)
      CALL cp_blacs_env_retain(almo_scf_env%blacs_env)

      ! copy basic quantities
      almo_scf_env%nspins = dft_control%nspins
      almo_scf_env%natoms = cp_dbcsr_nblkrows_total(matrix_s(1)%matrix)
      almo_scf_env%nmolecules = SIZE(molecule_set)
      CALL cp_dbcsr_get_info(matrix_s(1)%matrix, nfullrows_total=naos)
      almo_scf_env%naos = naos

      ! convenient local varibales
      nspins = almo_scf_env%nspins
      nmols = almo_scf_env%nmolecules
      natoms = almo_scf_env%natoms

      ! parse the almo_scf section and set appropriate quantities
      CALL almo_scf_init_read_write_input(input, almo_scf_env)

      ! Define groups: either atomic or molecular
      IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular) THEN
         almo_scf_env%ndomains = almo_scf_env%nmolecules
      ELSE
         almo_scf_env%ndomains = almo_scf_env%natoms
      ENDIF

      ! allocate domain descriptors
      ndomains = almo_scf_env%ndomains
      ALLOCATE (almo_scf_env%domain_index_of_atom(natoms))
      ALLOCATE (almo_scf_env%domain_index_of_ao(naos))
      ALLOCATE (almo_scf_env%first_atom_of_domain(ndomains))
      ALLOCATE (almo_scf_env%last_atom_of_domain(ndomains))
      ALLOCATE (almo_scf_env%nbasis_of_domain(ndomains))
      ALLOCATE (almo_scf_env%nocc_of_domain(ndomains, nspins))
      ALLOCATE (almo_scf_env%nvirt_full_of_domain(ndomains, nspins))
      ALLOCATE (almo_scf_env%nvirt_of_domain(ndomains, nspins))
      ALLOCATE (almo_scf_env%nvirt_disc_of_domain(ndomains, nspins))
      ALLOCATE (almo_scf_env%mu_of_domain(ndomains, nspins))
      ALLOCATE (almo_scf_env%cpu_of_domain(ndomains))
      ALLOCATE (almo_scf_env%charge_of_domain(ndomains))
      ALLOCATE (almo_scf_env%multiplicity_of_domain(ndomains))

      ! fill out domain descriptors and group descriptors
      IF (almo_scf_env%domain_layout_mos == almo_domain_layout_molecular) THEN
         ! get domain info from molecule_set
         CALL get_molecule_set_info(molecule_set, &
                                    atom_to_mol=almo_scf_env%domain_index_of_atom, &
                                    mol_to_first_atom=almo_scf_env%first_atom_of_domain, &
                                    mol_to_last_atom=almo_scf_env%last_atom_of_domain, &
                                    mol_to_nelectrons=almo_scf_env%nocc_of_domain(1:ndomains, 1), &
                                    mol_to_nbasis=almo_scf_env%nbasis_of_domain, &
                                    mol_to_charge=almo_scf_env%charge_of_domain, &
                                    mol_to_multiplicity=almo_scf_env%multiplicity_of_domain)
         ! calculate number of alpha and beta occupied orbitals from
         ! the number of electrons and multiplicity of each molecule
         ! Na + Nb = Ne
         ! Na - Nb = Mult - 1 (assume Na > Nb as we do not have more info from get_molecule_set_info)
         DO idomain = 1, ndomains
            nelec = almo_scf_env%nocc_of_domain(idomain, 1)
            multip = almo_scf_env%multiplicity_of_domain(idomain)
            nelec_a = (nelec+multip-1)/2
            nelec_b = nelec-nelec_a
            almo_scf_env%nocc_of_domain(idomain, 1) = nelec_a
            IF (nelec_a .NE. nelec_b) THEN
               IF (nspins .EQ. 1) THEN
                  WRITE (*, *) "Domain ", idomain, " out of ", ndomains, ". Electrons = ", nelec
                  CALL cp__b("almo_scf.F",239,"odd e- -- use unrestricted methods")
               ENDIF
               almo_scf_env%nocc_of_domain(idomain, 2) = nelec_b
               ! RZK-warning: open-shell procedures have not been tested yet
               ! Stop the program now
               CALL cp__b("almo_scf.F",244,"Unrestricted ALMO methods are NYI")
            ENDIF
         ENDDO
         DO ispin = 1, nspins
            ! take care of the full virtual subspace
            almo_scf_env%nvirt_full_of_domain(:, ispin) = &
               almo_scf_env%nbasis_of_domain(:)- &
               almo_scf_env%nocc_of_domain(:, ispin)
            ! and the truncated virtual subspace
            SELECT CASE (almo_scf_env%deloc_truncate_virt)
            CASE (virt_full)
               almo_scf_env%nvirt_of_domain(:, ispin) = &
                  almo_scf_env%nvirt_full_of_domain(:, ispin)
               almo_scf_env%nvirt_disc_of_domain(:, ispin) = 0
            CASE (virt_number)
               DO idomain = 1, ndomains
                  almo_scf_env%nvirt_of_domain(idomain, ispin) = &
                     MIN(almo_scf_env%deloc_virt_per_domain, &
                         almo_scf_env%nvirt_full_of_domain(idomain, ispin))
                  almo_scf_env%nvirt_disc_of_domain(idomain, ispin) = &
                     almo_scf_env%nvirt_full_of_domain(idomain, ispin)- &
                     almo_scf_env%nvirt_of_domain(idomain, ispin)
               ENDDO
            CASE (virt_occ_size)
               DO idomain = 1, ndomains
                  almo_scf_env%nvirt_of_domain(idomain, ispin) = &
                     MIN(almo_scf_env%nocc_of_domain(idomain, ispin), &
                         almo_scf_env%nvirt_full_of_domain(idomain, ispin))
                  almo_scf_env%nvirt_disc_of_domain(idomain, ispin) = &
                     almo_scf_env%nvirt_full_of_domain(idomain, ispin)- &
                     almo_scf_env%nvirt_of_domain(idomain, ispin)
               ENDDO
            CASE DEFAULT
               CALL cp__b("almo_scf.F",277,"illegal method for virtual space truncation")
            END SELECT
         ENDDO ! spin
      ELSE ! domains are atomic
         ! RZK-warning do the same for atomic domains/groups
         almo_scf_env%domain_index_of_atom(1:natoms) = (/(i, i=1, natoms)/)
      ENDIF

      ao = 1
      DO idomain = 1, ndomains
         DO iao = 1, almo_scf_env%nbasis_of_domain(idomain)
            almo_scf_env%domain_index_of_ao(ao) = idomain
            ao = ao+1
         ENDDO
      ENDDO

      almo_scf_env%mu_of_domain(:, :) = almo_scf_env%mu

      ! build domain (i.e. layout) indices for distribution blocks
      ! ao blocks
      IF (almo_scf_env%mat_distr_aos == almo_mat_distr_atomic) THEN
         ALLOCATE (almo_scf_env%domain_index_of_ao_block(natoms))
         almo_scf_env%domain_index_of_ao_block(:) = &
            almo_scf_env%domain_index_of_atom(:)
      ELSE IF (almo_scf_env%mat_distr_aos == almo_mat_distr_molecular) THEN
         ALLOCATE (almo_scf_env%domain_index_of_ao_block(nmols))
         ! if distr blocks are molecular then domain layout is also molecular
         almo_scf_env%domain_index_of_ao_block(:) = (/(i, i=1, nmols)/)
      ENDIF
      ! mo blocks
      IF (almo_scf_env%mat_distr_mos == almo_mat_distr_atomic) THEN
         ALLOCATE (almo_scf_env%domain_index_of_mo_block(natoms))
         almo_scf_env%domain_index_of_mo_block(:) = &
            almo_scf_env%domain_index_of_atom(:)
      ELSE IF (almo_scf_env%mat_distr_mos == almo_mat_distr_molecular) THEN
         ALLOCATE (almo_scf_env%domain_index_of_mo_block(nmols))
         ! if distr blocks are molecular then domain layout is also molecular
         almo_scf_env%domain_index_of_mo_block(:) = (/(i, i=1, nmols)/)
      ENDIF

      ! set all flags
      !almo_scf_env%need_previous_ks=.FALSE.
      !IF (almo_scf_env%deloc_method==almo_deloc_harris) THEN
      almo_scf_env%need_previous_ks = .TRUE.
      !ENDIF

      !almo_scf_env%need_virtuals=.FALSE.
      !almo_scf_env%need_orbital_energies=.FALSE.
      !IF (almo_scf_env%almo_update_algorithm==almo_scf_diag) THEN
      almo_scf_env%need_virtuals = .TRUE.
      almo_scf_env%need_orbital_energies = .TRUE.
      !ENDIF

      ! create all matrices
      CALL almo_scf_env_create_matrices(almo_scf_env, matrix_s(1)%matrix)
      ! set up matrix S and all required functions of S
      almo_scf_env%s_inv_done = .FALSE.
      almo_scf_env%s_sqrt_done = .FALSE.
      CALL almo_scf_init_ao_overlap(matrix_s(1)%matrix, almo_scf_env)

      ! create the quencher (imposes sparsity template)
      CALL almo_scf_construct_quencher(qs_env, almo_scf_env)
      CALL distribute_domains(almo_scf_env)

      ! FINISH setting job parameters here, print out job info
      CALL almo_scf_print_job_info(almo_scf_env, unit_nr)

      ! allocate and init the domain preconditioner
      ALLOCATE (almo_scf_env%domain_preconditioner(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_preconditioner)

      ! allocate and init projected KS for domains
      ALLOCATE (almo_scf_env%domain_ks_xx(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_ks_xx)

      ! init ao-overlap subblocks
      ALLOCATE (almo_scf_env%domain_s_inv(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_s_inv)
      ALLOCATE (almo_scf_env%domain_s_sqrt_inv(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_s_sqrt_inv)
      ALLOCATE (almo_scf_env%domain_s_sqrt(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_s_sqrt)
      ALLOCATE (almo_scf_env%domain_t(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_t)
      ALLOCATE (almo_scf_env%domain_err(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_err)
      ALLOCATE (almo_scf_env%domain_r_down_up(ndomains, nspins))
      CALL init_submatrices(almo_scf_env%domain_r_down_up)

      ! initialization of the QS settings with the ALMO flavor
      CALL almo_scf_init_qs(qs_env, almo_scf_env)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_init

! **************************************************************************************************
!> \brief Parses the ALMO input section
!> \param input ...
!> \param almo_scf_env ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_init_read_write_input(input, almo_scf_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_init_read_write_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: almo_opt_diis_section, &
                                                            almo_opt_pcg_section, &
                                                            almo_scf_section, xalmo_opt_pcg_section

      CALL timeset(routineN, handle)

      almo_scf_section => section_vals_get_subs_vals(input, "DFT%ALMO_SCF")
      almo_opt_diis_section => section_vals_get_subs_vals(almo_scf_section, &
                                                          "ALMO_OPTIMIZER_DIIS")
      almo_opt_pcg_section => section_vals_get_subs_vals(almo_scf_section, &
                                                         "ALMO_OPTIMIZER_PCG")
      xalmo_opt_pcg_section => section_vals_get_subs_vals(almo_scf_section, &
                                                          "XALMO_OPTIMIZER_PCG")

      ! RZK-warning the values of these keywords are hardcoded
      ! but ideally they should also be read from the input file
      almo_scf_env%order_lanczos = 3
      almo_scf_env%eps_lanczos = 1.0E-4_dp
      almo_scf_env%max_iter_lanczos = 40

      ! read user input
      ! common ALMO options
      CALL section_vals_val_get(almo_scf_section, "EPS_FILTER", &
                                r_val=almo_scf_env%eps_filter)
      CALL section_vals_val_get(almo_scf_section, "ALMO_SCF_GUESS", &
                                i_val=almo_scf_env%almo_scf_guess)
      CALL section_vals_val_get(almo_scf_section, "ALMO_ALGORITHM", &
                                i_val=almo_scf_env%almo_update_algorithm)
      CALL section_vals_val_get(almo_scf_section, "DELOCALIZE_METHOD", &
                                i_val=almo_scf_env%deloc_method)
      CALL section_vals_val_get(almo_scf_section, "XALMO_R_CUTOFF_FACTOR", &
                                r_val=almo_scf_env%quencher_r0_factor)

      ! optimizers
      CALL section_vals_val_get(almo_opt_diis_section, "EPS_ERROR", &
                                r_val=almo_scf_env%opt_block_diag_diis%eps_error)
      CALL section_vals_val_get(almo_opt_diis_section, "MAX_ITER", &
                                i_val=almo_scf_env%opt_block_diag_diis%max_iter)
      CALL section_vals_val_get(almo_opt_diis_section, "N_DIIS", &
                                i_val=almo_scf_env%opt_block_diag_diis%ndiis)

      CALL section_vals_val_get(almo_opt_pcg_section, "EPS_ERROR", &
                                r_val=almo_scf_env%opt_block_diag_pcg%eps_error)
      CALL section_vals_val_get(almo_opt_pcg_section, "MAX_ITER", &
                                i_val=almo_scf_env%opt_block_diag_pcg%max_iter)
      CALL section_vals_val_get(almo_opt_pcg_section, "MAX_ITER_OUTER_LOOP", &
                                i_val=almo_scf_env%opt_block_diag_pcg%max_iter_outer_loop)
      CALL section_vals_val_get(almo_opt_pcg_section, "LIN_SEARCH_EPS_ERROR", &
                                r_val=almo_scf_env%opt_block_diag_pcg%lin_search_eps_error)
      CALL section_vals_val_get(almo_opt_pcg_section, "LIN_SEARCH_STEP_SIZE_GUESS", &
                                r_val=almo_scf_env%opt_block_diag_pcg%lin_search_step_size_guess)
      CALL section_vals_val_get(almo_opt_pcg_section, "CONJUGATOR", &
                                i_val=almo_scf_env%opt_block_diag_pcg%conjugator)
      CALL section_vals_val_get(almo_opt_pcg_section, "PRECONDITIONER", &
                                i_val=almo_scf_env%opt_block_diag_pcg%preconditioner)

      CALL section_vals_val_get(xalmo_opt_pcg_section, "EPS_ERROR", &
                                r_val=almo_scf_env%opt_xalmo_pcg%eps_error)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "MAX_ITER", &
                                i_val=almo_scf_env%opt_xalmo_pcg%max_iter)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "MAX_ITER_OUTER_LOOP", &
                                i_val=almo_scf_env%opt_xalmo_pcg%max_iter_outer_loop)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "LIN_SEARCH_EPS_ERROR", &
                                r_val=almo_scf_env%opt_xalmo_pcg%lin_search_eps_error)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "LIN_SEARCH_STEP_SIZE_GUESS", &
                                r_val=almo_scf_env%opt_xalmo_pcg%lin_search_step_size_guess)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "CONJUGATOR", &
                                i_val=almo_scf_env%opt_xalmo_pcg%conjugator)
      CALL section_vals_val_get(xalmo_opt_pcg_section, "PRECONDITIONER", &
                                i_val=almo_scf_env%opt_xalmo_pcg%preconditioner)

      ! do not do EDA in the official release until the output files are written
      ! using the proper CP2K routines
      almo_scf_env%almo_eda = 0

      !CALL section_vals_val_get(almo_scf_section,"DOMAIN_LAYOUT_AOS",&
      !                          i_val=almo_scf_env%domain_layout_aos)
      !CALL section_vals_val_get(almo_scf_section,"DOMAIN_LAYOUT_MOS",&
      !                          i_val=almo_scf_env%domain_layout_mos)
      !CALL section_vals_val_get(almo_scf_section,"MATRIX_CLUSTERING_AOS",&
      !                          i_val=almo_scf_env%mat_distr_aos)
      !CALL section_vals_val_get(almo_scf_section,"MATRIX_CLUSTERING_MOS",&
      !                          i_val=almo_scf_env%mat_distr_mos)
      !CALL section_vals_val_get(almo_scf_section,"CONSTRAINT_TYPE",&
      !                          i_val=almo_scf_env%constraint_type)
      !CALL section_vals_val_get(almo_scf_section,"MU",&
      !                          r_val=almo_scf_env%mu)
      !CALL section_vals_val_get(almo_scf_section,"FIXED_MU",&
      !                          l_val=almo_scf_env%fixed_mu)
      !CALL section_vals_val_get(almo_scf_section,"EPS_USE_PREV_AS_GUESS",&
      !                          r_val=almo_scf_env%eps_prev_guess)
      !CALL section_vals_val_get(almo_scf_section,"MIXING_FRACTION",&
      !                          r_val=almo_scf_env%mixing_fraction)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_TENSOR_TYPE",&
      !                          i_val=almo_scf_env%deloc_cayley_tensor_type)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_CONJUGATOR",&
      !                          i_val=almo_scf_env%deloc_cayley_conjugator)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_MAX_ITER",&
      !                          i_val=almo_scf_env%deloc_cayley_max_iter)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_USE_OCC_ORBS",&
      !                          l_val=almo_scf_env%deloc_use_occ_orbs)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_USE_VIRT_ORBS",&
      !                          l_val=almo_scf_env%deloc_cayley_use_virt_orbs)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_LINEAR",&
      !                          l_val=almo_scf_env%deloc_cayley_linear)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_EPS_CONVERGENCE",&
      !                          r_val=almo_scf_env%deloc_cayley_eps_convergence)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_OCC_PRECOND",&
      !                          l_val=almo_scf_env%deloc_cayley_occ_precond)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_CAYLEY_VIR_PRECOND",&
      !                          l_val=almo_scf_env%deloc_cayley_vir_precond)
      !CALL section_vals_val_get(almo_scf_section,"ALMO_UPDATE_ALGORITHM_BD",&
      !                          i_val=almo_scf_env%almo_update_algorithm)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_TRUNCATE_VIRTUALS",&
      !                          i_val=almo_scf_env%deloc_truncate_virt)
      !CALL section_vals_val_get(almo_scf_section,"DELOC_VIRT_PER_DOMAIN",&
      !                          i_val=almo_scf_env%deloc_virt_per_domain)
      !
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_EPS_CONVERGENCE",&
      !                          r_val=almo_scf_env%opt_k_eps_convergence)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_MAX_ITER",&
      !                          i_val=almo_scf_env%opt_k_max_iter)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_OUTER_MAX_ITER",&
      !                          i_val=almo_scf_env%opt_k_outer_max_iter)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_TRIAL_STEP_SIZE",&
      !                          r_val=almo_scf_env%opt_k_trial_step_size)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_CONJUGATOR",&
      !                          i_val=almo_scf_env%opt_k_conjugator)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_TRIAL_STEP_SIZE_MULTIPLIER",&
      !                          r_val=almo_scf_env%opt_k_trial_step_size_multiplier)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_CONJ_ITER_START",&
      !                          i_val=almo_scf_env%opt_k_conj_iter_start)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_PREC_ITER_START",&
      !                          i_val=almo_scf_env%opt_k_prec_iter_start)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_CONJ_ITER_FREQ_RESET",&
      !                          i_val=almo_scf_env%opt_k_conj_iter_freq)
      !CALL section_vals_val_get(almo_scf_section,"OPT_K_PREC_ITER_FREQ_UPDATE",&
      !                          i_val=almo_scf_env%opt_k_prec_iter_freq)
      !
      !CALL section_vals_val_get(almo_scf_section,"QUENCHER_RADIUS_TYPE",&
      !                          i_val=almo_scf_env%quencher_radius_type)
      !CALL section_vals_val_get(almo_scf_section,"QUENCHER_R0_FACTOR",&
      !                          r_val=almo_scf_env%quencher_r0_factor)
      !CALL section_vals_val_get(almo_scf_section,"QUENCHER_R1_FACTOR",&
      !                          r_val=almo_scf_env%quencher_r1_factor)
      !!CALL section_vals_val_get(almo_scf_section,"QUENCHER_R0_SHIFT",&
      !!                          r_val=almo_scf_env%quencher_r0_shift)
      !!
      !!CALL section_vals_val_get(almo_scf_section,"QUENCHER_R1_SHIFT",&
      !!                          r_val=almo_scf_env%quencher_r1_shift)
      !!
      !!almo_scf_env%quencher_r0_shift = cp_unit_to_cp2k(&
      !!   almo_scf_env%quencher_r0_shift,"angstrom")
      !!almo_scf_env%quencher_r1_shift = cp_unit_to_cp2k(&
      !!   almo_scf_env%quencher_r1_shift,"angstrom")
      !
      !CALL section_vals_val_get(almo_scf_section,"QUENCHER_AO_OVERLAP_0",&
      !                          r_val=almo_scf_env%quencher_s0)
      !CALL section_vals_val_get(almo_scf_section,"QUENCHER_AO_OVERLAP_1",&
      !                          r_val=almo_scf_env%quencher_s1)

      !CALL section_vals_val_get(almo_scf_section,"ENVELOPE_AMPLITUDE",&
      !                          r_val=almo_scf_env%envelope_amplitude)

      !! how to read lists
      !CALL section_vals_val_get(almo_scf_section,"INT_LIST01", &
      !        n_rep_val=n_rep)
      !counter_i = 0
      !DO k = 1,n_rep
      !  CALL section_vals_val_get(almo_scf_section,"INT_LIST01",&
      !          i_rep_val=k,i_vals=tmplist)
      !   DO jj = 1,SIZE(tmplist)
      !      counter_i=counter_i+1
      !      almo_scf_env%charge_of_domain(counter_i)=tmplist(jj)
      !   ENDDO
      !ENDDO

      almo_scf_env%domain_layout_aos = almo_domain_layout_molecular
      almo_scf_env%domain_layout_mos = almo_domain_layout_molecular
      almo_scf_env%mat_distr_aos = almo_mat_distr_molecular
      almo_scf_env%mat_distr_mos = almo_mat_distr_molecular

      almo_scf_env%constraint_type = almo_constraint_distance
      almo_scf_env%mu = -0.1_dp
      almo_scf_env%fixed_mu = .FALSE.
      almo_scf_env%eps_prev_guess = almo_scf_env%eps_filter/100.0_dp
      almo_scf_env%mixing_fraction = 0.45_dp

      almo_scf_env%deloc_cayley_tensor_type = tensor_orthogonal
      almo_scf_env%deloc_cayley_conjugator = cg_hager_zhang
      almo_scf_env%deloc_cayley_max_iter = 100
      almo_scf_env%deloc_use_occ_orbs = .TRUE.
      almo_scf_env%deloc_cayley_use_virt_orbs = .FALSE.
      almo_scf_env%deloc_cayley_linear = .FALSE.
      almo_scf_env%deloc_cayley_eps_convergence = 1.0E-6_dp
      almo_scf_env%deloc_cayley_occ_precond = .TRUE.
      almo_scf_env%deloc_cayley_vir_precond = .TRUE.
      almo_scf_env%deloc_truncate_virt = virt_full
      almo_scf_env%deloc_virt_per_domain = -1

      almo_scf_env%opt_k_eps_convergence = 1.0E-5_dp
      almo_scf_env%opt_k_max_iter = 100
      almo_scf_env%opt_k_outer_max_iter = 1
      almo_scf_env%opt_k_trial_step_size = 0.05_dp
      almo_scf_env%opt_k_conjugator = cg_hager_zhang
      almo_scf_env%opt_k_trial_step_size_multiplier = 1.05_dp
      almo_scf_env%opt_k_conj_iter_start = 0
      almo_scf_env%opt_k_prec_iter_start = 0
      almo_scf_env%opt_k_conj_iter_freq = 10000000
      almo_scf_env%opt_k_prec_iter_freq = 10000000

      almo_scf_env%quencher_radius_type = do_bondparm_vdw
      almo_scf_env%quencher_r1_factor = almo_scf_env%quencher_r0_factor
      !almo_scf_env%quencher_r0_shift=0.0_dp
      !almo_scf_env%quencher_r1_shift=0.0_dp
      !almo_scf_env%quencher_r0_shift = cp_unit_to_cp2k(&
      !   almo_scf_env%quencher_r0_shift,"angstrom")
      !almo_scf_env%quencher_r1_shift = cp_unit_to_cp2k(&
      !   almo_scf_env%quencher_r1_shift,"angstrom")

      almo_scf_env%quencher_s0 = 1.0E-4_dp
      almo_scf_env%quencher_s1 = 1.0E-6_dp

      almo_scf_env%envelope_amplitude = 1.0_dp

      almo_scf_env%logical01 = .FALSE. ! md in eDOF space
      almo_scf_env%logical02 = .TRUE. ! not used
      almo_scf_env%logical03 = .TRUE. ! not used
      almo_scf_env%logical04 = .TRUE. ! use preconditioner
      almo_scf_env%logical05 = .FALSE. ! optimize theta

      almo_scf_env%real01 = almo_scf_env%eps_filter/10.0_dp ! skip gradients
      almo_scf_env%real02 = 0.0_dp ! not used
      almo_scf_env%real03 = 0.0_dp ! not used
      almo_scf_env%real04 = 0.5_dp ! mixing s-f precond

      almo_scf_env%integer01 = 10 ! start eDOF-md
      almo_scf_env%integer02 = 4 ! preconditioner type
      almo_scf_env%integer03 = 0 ! not used
      almo_scf_env%integer04 = 0 ! fixed number of line searches (no grad)
      almo_scf_env%integer05 = 0 ! not used

      ! select options that are not directly provided
      ! but are inderectly determined by the input
      IF (almo_scf_env%deloc_method .EQ. almo_deloc_xalmo_1diag) THEN
         almo_scf_env%xalmo_update_algorithm = almo_scf_diag
      ELSE
         almo_scf_env%xalmo_update_algorithm = almo_scf_pcg
      ENDIF

      ! check for conflicts between options
      IF (almo_scf_env%deloc_truncate_virt .EQ. virt_number .AND. &
          almo_scf_env%deloc_virt_per_domain .LE. 0) THEN
         CALL cp__b("almo_scf.F",642,"specify a positive number of virtual orbitals")
      ENDIF

      IF (almo_scf_env%deloc_truncate_virt .EQ. virt_minimal) THEN
         CALL cp__b("almo_scf.F",646,"VIRT TRUNCATION TO MINIMAL BASIS IS NIY")
      ENDIF

      IF (almo_scf_env%domain_layout_mos .NE. almo_domain_layout_molecular) THEN
         CALL cp__b("almo_scf.F",650,"use MOLECULAR domains")
      ENDIF

      IF (almo_scf_env%domain_layout_aos .NE. almo_domain_layout_molecular) THEN
         CALL cp__b("almo_scf.F",654,"use MOLECULAR domains")
      ENDIF

      IF (almo_scf_env%mat_distr_mos .NE. almo_mat_distr_molecular) THEN
         CALL cp__b("almo_scf.F",658,"use MOLECULAR distr for MOs")
      ENDIF

      IF (almo_scf_env%mat_distr_aos == almo_mat_distr_molecular .AND. &
          almo_scf_env%domain_layout_aos == almo_domain_layout_atomic) THEN
         CALL cp__b("almo_scf.F",663,"AO blocks cannot be larger than domains")
      ENDIF

      IF (almo_scf_env%mat_distr_mos == almo_mat_distr_molecular .AND. &
          almo_scf_env%domain_layout_mos == almo_domain_layout_atomic) THEN
         CALL cp__b("almo_scf.F",668,"MO blocks cannot be larger than domains")
      ENDIF

      IF (almo_scf_env%quencher_r1_factor .GT. almo_max_cutoff_multiplier) THEN
         CALL cp_abort(cp__l("almo_scf.F",672), &
                       "XALMO_R_CUTOFF_FACTOR is larger than almo_max_cutoff_multiplier"// &
                       "increase the hard-coded almo_max_cutoff_multiplier")
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE almo_scf_init_read_write_input

! **************************************************************************************************
!> \brief Prints out a short summary about the ALMO SCF job
!> \param almo_scf_env ...
!> \param unit_nr ...
!> \par History
!>       2011.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_print_job_info(almo_scf_env, unit_nr)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_print_job_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=13)                                  :: neig_string
      CHARACTER(len=33)                                  :: deloc_method_string
      INTEGER                                            :: handle, idomain, index1_prev

      CALL timeset(routineN, handle)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '()')
         WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 32), " ALMO SETTINGS ", REPEAT("-", 32)

         WRITE (unit_nr, '(T2,A,T48,E33.3)') "eps_filter:", almo_scf_env%eps_filter

         WRITE (unit_nr, '(T2,A)') "optimization of block-diagonal ALMOs:"
         SELECT CASE (almo_scf_env%almo_update_algorithm)
         CASE (almo_scf_diag)
            ! the DIIS algorith is the only choice for the diagonlaization-based algorithm
            CALL print_optimizer_options(almo_scf_env%opt_block_diag_diis, unit_nr)
         CASE (almo_scf_pcg)
            ! print out PCG options
            CALL print_optimizer_options(almo_scf_env%opt_block_diag_pcg, unit_nr)
         END SELECT

         SELECT CASE (almo_scf_env%deloc_method)
         CASE (almo_deloc_none)
            deloc_method_string = "NONE"
         CASE (almo_deloc_x)
            deloc_method_string = "FULL_X"
         CASE (almo_deloc_scf)
            deloc_method_string = "FULL_SCF"
         CASE (almo_deloc_x_then_scf)
            deloc_method_string = "FULL_X_THEN_SCF"
         CASE (almo_deloc_xalmo_1diag)
            deloc_method_string = "XALMO_1DIAG"
         CASE (almo_deloc_xalmo_x)
            deloc_method_string = "XALMO_X"
         CASE (almo_deloc_xalmo_scf)
            deloc_method_string = "XALMO_SCF"
         END SELECT
         WRITE (unit_nr, '(T2,A,T48,A33)') "delocalization:", TRIM(deloc_method_string)

         IF (almo_scf_env%deloc_method .NE. almo_deloc_none) THEN

            SELECT CASE (almo_scf_env%deloc_method)
            CASE (almo_deloc_x, almo_deloc_scf, almo_deloc_x_then_scf)
               WRITE (unit_nr, '(T2,A,T48,A33)') "delocalization cutoff radius:", &
                  "infinite"
               deloc_method_string = "FULL_X_THEN_SCF"
            CASE (almo_deloc_xalmo_1diag, almo_deloc_xalmo_x, almo_deloc_xalmo_scf)
               WRITE (unit_nr, '(T2,A,T48,F33.5)') "XALMO cutoff radius:", &
                  almo_scf_env%quencher_r0_factor
            END SELECT

            IF (almo_scf_env%deloc_method .EQ. almo_deloc_xalmo_1diag) THEN
               ! print nothing because no actual optimization is done
            ELSE
               WRITE (unit_nr, '(T2,A)') "optimization of extended orbitals:"
               SELECT CASE (almo_scf_env%xalmo_update_algorithm)
               CASE (almo_scf_diag)
                  CALL print_optimizer_options(almo_scf_env%opt_xalmo_diis, unit_nr)
               CASE (almo_scf_pcg)
                  CALL print_optimizer_options(almo_scf_env%opt_xalmo_pcg, unit_nr)
               END SELECT
            ENDIF

         ENDIF

         !SELECT CASE(almo_scf_env%domain_layout_mos)
         !CASE(almo_domain_layout_orbital)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Delocalization constraints","ORBITAL"
         !CASE(almo_domain_layout_atomic)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Delocalization constraints","ATOMIC"
         !CASE(almo_domain_layout_molecular)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Delocalization constraints","MOLECULAR"
         !END SELECT

         !SELECT CASE(almo_scf_env%domain_layout_aos)
         !CASE(almo_domain_layout_atomic)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Basis function domains","ATOMIC"
         !CASE(almo_domain_layout_molecular)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Basis function domains","MOLECULAR"
         !END SELECT

         !SELECT CASE(almo_scf_env%mat_distr_aos)
         !CASE(almo_mat_distr_atomic)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Parallel distribution for AOs","ATOMIC"
         !CASE(almo_mat_distr_molecular)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Parallel distribution for AOs","MOLECULAR"
         !END SELECT

         !SELECT CASE(almo_scf_env%mat_distr_mos)
         !CASE(almo_mat_distr_atomic)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Parallel distribution for MOs","ATOMIC"
         !CASE(almo_mat_distr_molecular)
         !    WRITE(unit_nr,'(T2,A,T48,A33)') "Parallel distribution for MOs","MOLECULAR"
         !END SELECT

         WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)

         IF (almo_scf_env%ndomains .LE. 200) THEN

            ! print fragment info
            WRITE (unit_nr, '(T2,A13,A13,A13,A13,A13,A13)') &
               "Fragment", "Basis Set", "Occupied", "Virtual", "Charge", "Deloc Neig" !,"Discarded Virt"
            WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)
            DO idomain = 1, almo_scf_env%ndomains

               IF (idomain .EQ. 1) THEN
                  index1_prev = 1
               ELSE
                  index1_prev = almo_scf_env%domain_map(1)%index1(idomain-1)
               ENDIF

               SELECT CASE (almo_scf_env%deloc_method)
               CASE (almo_deloc_none)
                  neig_string = "NONE"
               CASE (almo_deloc_x, almo_deloc_scf, almo_deloc_x_then_scf)
                  neig_string = "ALL"
               CASE (almo_deloc_xalmo_1diag, almo_deloc_xalmo_x, almo_deloc_xalmo_scf)
                  WRITE (neig_string, '(I13)') &
                     almo_scf_env%domain_map(1)%index1(idomain)-index1_prev-1 ! minus self
               CASE DEFAULT
                  neig_string = "N/A"
               END SELECT

               WRITE (unit_nr, '(T2,I13,I13,I13,I13,I13,A13)') &
                  idomain, almo_scf_env%nbasis_of_domain(idomain), &
                  SUM(almo_scf_env%nocc_of_domain(idomain, :)), &
                  SUM(almo_scf_env%nvirt_of_domain(idomain, :)), &
                  !SUM(almo_scf_env%nvirt_disc_of_domain(idomain,:)),&
                  almo_scf_env%charge_of_domain(idomain), &
                  ADJUSTR(TRIM(neig_string))

            ENDDO ! cycle over domains

            SELECT CASE (almo_scf_env%deloc_method)
            CASE (almo_deloc_xalmo_1diag, almo_deloc_xalmo_x, almo_deloc_xalmo_scf)

               WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)

               ! print fragment neigbors
               WRITE (unit_nr, '(T2,A78)') &
                  "Neighbor lists (including self)"
               WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)
               DO idomain = 1, almo_scf_env%ndomains

                  IF (idomain .EQ. 1) THEN
                     index1_prev = 1
                  ELSE
                     index1_prev = almo_scf_env%domain_map(1)%index1(idomain-1)
                  ENDIF

                  WRITE (unit_nr, '(T2,I13,":")') idomain
                  WRITE (unit_nr, '(T15,5I13)') &
                     almo_scf_env%domain_map(1)%pairs &
                     (index1_prev:almo_scf_env%domain_map(1)%index1(idomain)-1, 1) ! includes self

               ENDDO ! cycle over domains

            END SELECT

            WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)

         ENDIF

         WRITE (unit_nr, '()')

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE almo_scf_print_job_info

! **************************************************************************************************
!> \brief Initializes the ALMO SCF copy of the AO overlap matrix
!>        and all necessary functions (sqrt, inverse...)
!> \param matrix_s ...
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_init_ao_overlap(matrix_s, almo_scf_env)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_s
      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_init_ao_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! make almo copy of S
      ! also copy S to S_blk (i.e. to S with the domain structure imposed)
      IF (almo_scf_env%orthogonal_basis) THEN
         CALL cp_dbcsr_set(almo_scf_env%matrix_s(1), 0.0_dp)
         CALL cp_dbcsr_add_on_diag(almo_scf_env%matrix_s(1), 1.0_dp)
         CALL cp_dbcsr_set(almo_scf_env%matrix_s_blk(1), 0.0_dp)
         CALL cp_dbcsr_add_on_diag(almo_scf_env%matrix_s_blk(1), 1.0_dp)
      ELSE
         CALL matrix_qs_to_almo(matrix_s, almo_scf_env%matrix_s(1), &
                                almo_scf_env, .FALSE.)
         CALL matrix_qs_to_almo(matrix_s, almo_scf_env%matrix_s_blk(1), &
                                almo_scf_env, .TRUE.)
      ENDIF

      CALL cp_dbcsr_filter(almo_scf_env%matrix_s(1), almo_scf_env%eps_filter)
      CALL cp_dbcsr_filter(almo_scf_env%matrix_s_blk(1), almo_scf_env%eps_filter)

      IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_diag) THEN
         CALL matrix_sqrt_Newton_Schulz(almo_scf_env%matrix_s_blk_sqrt(1), &
                                        almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                        almo_scf_env%matrix_s_blk(1), &
                                        threshold=almo_scf_env%eps_filter, &
                                        order=almo_scf_env%order_lanczos, &
                                        eps_lanczos=almo_scf_env%eps_lanczos, &
                                        max_iter_lanczos=almo_scf_env%max_iter_lanczos)
      ELSE IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_dm_sign) THEN
         CALL invert_Hotelling(almo_scf_env%matrix_s_blk_inv(1), &
                               almo_scf_env%matrix_s_blk(1), &
                               threshold=almo_scf_env%eps_filter)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE almo_scf_init_ao_overlap

! **************************************************************************************************
!> \brief Selects the subroutine for the optimization of block-daigonal ALMOs.
!>        Keep it short and clean.
!> \param qs_env ...
!> \param almo_scf_env ...
!> \par History
!>       2011.11 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_main(qs_env, almo_scf_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_main', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_pcg) THEN

         ! ALMO PCG optimizer as a special case of XALMO PCG
         CALL almo_scf_xalmo_pcg(qs_env=qs_env, &
                                 almo_scf_env=almo_scf_env, &
                                 optimizer=almo_scf_env%opt_block_diag_pcg, &
                                 quench_t=almo_scf_env%quench_t_blk, &
                                 matrix_t_in=almo_scf_env%matrix_t_blk, &
                                 matrix_t_out=almo_scf_env%matrix_t_blk, &
                                 assume_t0_q0x=.FALSE., &
                                 perturbation_only=.FALSE., &
                                 special_case=xalmo_case_block_diag)

         CALL almo_scf_t_blk_to_t_blk_orthonormal(almo_scf_env)

      ELSE

         ! mixing/DIIS optimizer
         CALL almo_scf_block_diagonal(qs_env, almo_scf_env, &
                                      almo_scf_env%opt_block_diag_diis)

      ENDIF

      ! we might need a copy of the converged KS
      DO ispin = 1, almo_scf_env%nspins
         CALL cp_dbcsr_copy(almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                            almo_scf_env%matrix_ks(ispin))
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_main

! **************************************************************************************************
!> \brief selects various post scf routines
!> \param qs_env ...
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_delocalization(qs_env, almo_scf_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_delocalization', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, handle, hold, iblock_col, &
                                                            iblock_row, ispin, mynode, &
                                                            nblkcols_tot, nblkrows_tot, row, &
                                                            unit_nr
      LOGICAL                                            :: almo_experimental, tr
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_new_block
      TYPE(cp_dbcsr_type), ALLOCATABLE, DIMENSION(:)     :: no_quench
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(optimizer_options_type)                       :: arbitrary_optimizer

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! create a local optimizer that handles XALMO DIIS
      ! the options of this optimizer are arbitrary because
      ! XALMO DIIS SCF does not converge for yet unknown reasons and
      ! currently used in the code to get perturbative estimates only
      arbitrary_optimizer%optimizer_type = optimizer_diis
      arbitrary_optimizer%max_iter = 3
      arbitrary_optimizer%eps_error = 1.0E-6_dp
      arbitrary_optimizer%ndiis = 2

      SELECT CASE (almo_scf_env%deloc_method)
      CASE (almo_deloc_x, almo_deloc_scf, almo_deloc_x_then_scf)

         ! RZK-warning hack into the quenched routine:
         ! create a quench matrix with all-all-all blocks 1.0
         ! it is a waste of memory but since matrices are distributed
         ! we can tolerate it for now
         ALLOCATE (no_quench(almo_scf_env%nspins))
         CALL cp_dbcsr_init(no_quench(1))
         CALL cp_dbcsr_create(no_quench(1), &
                              template=almo_scf_env%matrix_t(1), &
                              matrix_type=dbcsr_type_no_symmetry)
         mynode = dbcsr_mp_mynode(dbcsr_distribution_mp( &
                                  cp_dbcsr_distribution(no_quench(1))))
         CALL cp_dbcsr_work_create(no_quench(1), &
                                   work_mutable=.TRUE.)
         nblkrows_tot = cp_dbcsr_nblkrows_total(no_quench(1))
         nblkcols_tot = cp_dbcsr_nblkcols_total(no_quench(1))
         ! RZK-warning: is it a quadratic-scaling routine?
         ! As a matter of fact it is! But this block treats
         ! fully delocalized MOs. So it is unavoidable.
         ! C'est la vie
         DO row = 1, nblkrows_tot
            DO col = 1, nblkcols_tot
               tr = .FALSE.
               iblock_row = row
               iblock_col = col
               CALL cp_dbcsr_get_stored_coordinates(no_quench(1), &
                                                    iblock_row, iblock_col, hold)
               IF (hold .EQ. mynode) THEN
                  NULLIFY (p_new_block)
                  CALL cp_dbcsr_reserve_block2d(no_quench(1), &
                                                iblock_row, iblock_col, p_new_block)
                  IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("almo_scf.F",1072)
                  p_new_block(:, :) = 1.0_dp
               ENDIF
            ENDDO
         ENDDO
         CALL cp_dbcsr_finalize(no_quench(1))
         IF (almo_scf_env%nspins .GT. 1) THEN
            DO ispin = 2, almo_scf_env%nspins
               CALL cp_dbcsr_init(no_quench(ispin))
               CALL cp_dbcsr_create(no_quench(ispin), &
                                    template=almo_scf_env%matrix_t(1), &
                                    matrix_type=dbcsr_type_no_symmetry)
               CALL cp_dbcsr_copy(no_quench(ispin), no_quench(1))
            ENDDO
         ENDIF

      END SELECT

      SELECT CASE (almo_scf_env%deloc_method)
      CASE (almo_deloc_none, almo_deloc_scf)

         DO ispin = 1, almo_scf_env%nspins
            CALL cp_dbcsr_copy(almo_scf_env%matrix_t(ispin), &
                               almo_scf_env%matrix_t_blk(ispin))
         ENDDO

      CASE (almo_deloc_x, almo_deloc_xk, almo_deloc_x_then_scf)

         !!!! RZK-warning a whole class of delocalization methods
         !!!! are commented out at the moment because some of their
         !!!! routines have not been thoroughly tested.

         !!!! if we have virtuals pre-optimize and truncate them
         !!!IF (almo_scf_env%need_virtuals) THEN
         !!!   SELECT CASE (almo_scf_env%deloc_truncate_virt)
         !!!   CASE (virt_full)
         !!!      ! simply copy virtual orbitals from matrix_v_full_blk to matrix_v_blk
         !!!      DO ispin=1,almo_scf_env%nspins
         !!!         CALL cp_dbcsr_copy(almo_scf_env%matrix_v_blk(ispin),&
         !!!                 almo_scf_env%matrix_v_full_blk(ispin))
         !!!      ENDDO
         !!!   CASE (virt_number,virt_occ_size)
         !!!      CALL split_v_blk(almo_scf_env)
         !!!      !CALL truncate_subspace_v_blk(qs_env,almo_scf_env)
         !!!   CASE DEFAULT
         !!!      CPErrorMessage(cp_failure_level,routineP,"illegal method for virtual space truncation")
         !!!      CPPrecondition(.FALSE.,cp_failure_level,routineP,failure)
         !!!   END SELECT
         !!!ENDIF
         !!!CALL harris_foulkes_correction(qs_env,almo_scf_env)

         CALL almo_scf_xalmo_pcg(qs_env=qs_env, &
                                 almo_scf_env=almo_scf_env, &
                                 optimizer=almo_scf_env%opt_xalmo_pcg, &
                                 quench_t=no_quench, &
                                 matrix_t_in=almo_scf_env%matrix_t_blk, &
                                 matrix_t_out=almo_scf_env%matrix_t, &
                                 assume_t0_q0x=.TRUE., &
                                 perturbation_only=.TRUE., &
                                 special_case=xalmo_case_fully_deloc)

      CASE (almo_deloc_xalmo_1diag)

         IF (almo_scf_env%xalmo_update_algorithm .EQ. almo_scf_diag) THEN

            almo_scf_env%perturbative_delocalization = .TRUE.
            DO ispin = 1, almo_scf_env%nspins
               CALL cp_dbcsr_copy(almo_scf_env%matrix_t(ispin), &
                                  almo_scf_env%matrix_t_blk(ispin))
            ENDDO
            CALL almo_scf_xalmo_eigensolver(qs_env, almo_scf_env, &
                                            arbitrary_optimizer)

         ELSE

            CALL cp__b("almo_scf.F",1147,"Other algorithms do not exist")

         ENDIF

      CASE (almo_deloc_xalmo_x)

         IF (almo_scf_env%xalmo_update_algorithm .EQ. almo_scf_pcg) THEN

            CALL almo_scf_xalmo_pcg(qs_env=qs_env, &
                                    almo_scf_env=almo_scf_env, &
                                    optimizer=almo_scf_env%opt_xalmo_pcg, &
                                    quench_t=almo_scf_env%quench_t, &
                                    matrix_t_in=almo_scf_env%matrix_t_blk, &
                                    matrix_t_out=almo_scf_env%matrix_t, &
                                    assume_t0_q0x=.TRUE., &
                                    perturbation_only=.TRUE., &
                                    special_case=xalmo_case_normal)

         ELSE

            CALL cp__b("almo_scf.F",1167,"Other algorithms do not exist")

         ENDIF

      CASE (almo_deloc_xalmo_scf)

         IF (almo_scf_env%xalmo_update_algorithm .EQ. almo_scf_diag) THEN

            CALL cp__b("almo_scf.F",1175,"Should not be here: convergence will fail!")

            almo_scf_env%perturbative_delocalization = .FALSE.
            DO ispin = 1, almo_scf_env%nspins
               CALL cp_dbcsr_copy(almo_scf_env%matrix_t(ispin), &
                                  almo_scf_env%matrix_t_blk(ispin))
            ENDDO
            CALL almo_scf_xalmo_eigensolver(qs_env, almo_scf_env, &
                                            arbitrary_optimizer)

         ELSE IF (almo_scf_env%xalmo_update_algorithm .EQ. almo_scf_pcg) THEN

            CALL almo_scf_xalmo_pcg(qs_env=qs_env, &
                                    almo_scf_env=almo_scf_env, &
                                    optimizer=almo_scf_env%opt_xalmo_pcg, &
                                    quench_t=almo_scf_env%quench_t, &
                                    matrix_t_in=almo_scf_env%matrix_t_blk, &
                                    matrix_t_out=almo_scf_env%matrix_t, &
                                    assume_t0_q0x=.TRUE., &
                                    perturbation_only=.FALSE., &
                                    special_case=xalmo_case_normal)

            ! RZK-warning THIS IS A HACK TO GET ORBITAL ENERGIES
            almo_experimental = .FALSE.
            IF (almo_experimental) THEN
               almo_scf_env%perturbative_delocalization = .TRUE.
               !DO ispin=1,almo_scf_env%nspins
               !   CALL cp_dbcsr_copy(almo_scf_env%matrix_t(ispin),&
               !           almo_scf_env%matrix_t_blk(ispin))
               !ENDDO
               CALL almo_scf_xalmo_eigensolver(qs_env, almo_scf_env, &
                                               arbitrary_optimizer)
            ENDIF ! experimental

         ENDIF

      CASE DEFAULT

         CALL cp__b("almo_scf.F",1213,"Illegal delocalization method")

      END SELECT

      SELECT CASE (almo_scf_env%deloc_method)
      CASE (almo_deloc_scf, almo_deloc_x_then_scf)

         IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
            CALL cp__b("almo_scf.F",1221,"full scf is NYI for truncated virtual space")
         ENDIF

         CALL almo_scf_xalmo_pcg(qs_env=qs_env, &
                                 almo_scf_env=almo_scf_env, &
                                 optimizer=almo_scf_env%opt_xalmo_pcg, &
                                 quench_t=no_quench, &
                                 matrix_t_in=almo_scf_env%matrix_t, &
                                 matrix_t_out=almo_scf_env%matrix_t, &
                                 assume_t0_q0x=.FALSE., &
                                 perturbation_only=.FALSE., &
                                 special_case=xalmo_case_fully_deloc)

      END SELECT

      ! clean up
      SELECT CASE (almo_scf_env%deloc_method)
      CASE (almo_deloc_x, almo_deloc_scf, almo_deloc_x_then_scf)
         DO ispin = 1, almo_scf_env%nspins
            CALL cp_dbcsr_release(no_quench(ispin))
         ENDDO
         DEALLOCATE (no_quench)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE almo_scf_delocalization

! **************************************************************************************************
!> \brief create various matrices
!> \param almo_scf_env ...
!> \param matrix_s0 ...
!> \par History
!>       2011.07 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_env_create_matrices(almo_scf_env, matrix_s0)

      TYPE(almo_scf_env_type), INTENT(INOUT)             :: almo_scf_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_s0

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_env_create_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins

      CALL timeset(routineN, handle)

      nspins = almo_scf_env%nspins

      ! AO overlap matrix and its various functions
      CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_s(1), &
                              matrix_qs=matrix_s0, &
                              almo_scf_env=almo_scf_env, &
                              name_new="S", &
                              size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                              symmetry_new=dbcsr_type_symmetric, &
                              spin_key=0, &
                              init_domains=.FALSE.)
      CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_s_blk(1), &
                              matrix_qs=matrix_s0, &
                              almo_scf_env=almo_scf_env, &
                              name_new="S_BLK", &
                              size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                              symmetry_new=dbcsr_type_symmetric, &
                              spin_key=0, &
                              init_domains=.TRUE.)
      IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_diag) THEN
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_s_blk_sqrt_inv(1), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="S_BLK_SQRT_INV", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=0, &
                                 init_domains=.TRUE.)
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_s_blk_sqrt(1), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="S_BLK_SQRT", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=0, &
                                 init_domains=.TRUE.)
      ELSE IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_dm_sign) THEN
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_s_blk_inv(1), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="S_BLK_INV", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=0, &
                                 init_domains=.TRUE.)
      ENDIF

      ! MO coeff matrices and their derivatives
      ALLOCATE (almo_scf_env%matrix_t_blk(nspins))
      ALLOCATE (almo_scf_env%quench_t_blk(nspins))
      ALLOCATE (almo_scf_env%matrix_err_blk(nspins))
      ALLOCATE (almo_scf_env%matrix_err_xx(nspins))
      ALLOCATE (almo_scf_env%matrix_sigma(nspins))
      ALLOCATE (almo_scf_env%matrix_sigma_inv(nspins))
      ALLOCATE (almo_scf_env%matrix_sigma_sqrt(nspins))
      ALLOCATE (almo_scf_env%matrix_sigma_sqrt_inv(nspins))
      ALLOCATE (almo_scf_env%matrix_sigma_blk(nspins))
      ALLOCATE (almo_scf_env%matrix_t(nspins))
      ALLOCATE (almo_scf_env%matrix_t_tr(nspins))
      DO ispin = 1, nspins
         ! create the blocked quencher
         CALL matrix_almo_create(matrix_new=almo_scf_env%quench_t_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="Q_BLK", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_occ/), &
                                 symmetry_new=dbcsr_type_no_symmetry, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
         ! create ALMO coefficient matrix
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_t_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="T_BLK", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_occ/), &
                                 symmetry_new=dbcsr_type_no_symmetry, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
         ! create the error matrix
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_err_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="ERR_BLK", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_no_symmetry, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
         ! create the error matrix for the quenched ALMOs
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_err_xx(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="ERR_XX", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_occ/), &
                                 symmetry_new=dbcsr_type_no_symmetry, &
                                 spin_key=ispin, &
                                 init_domains=.FALSE.)
         ! create a matrix with dimensions of a transposed mo coefficient matrix
         ! it might be necessary to perform the correction step using cayley
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_t_tr(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="T_TR", &
                                 size_keys=(/almo_mat_dim_occ, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_no_symmetry, &
                                 spin_key=ispin, &
                                 init_domains=.FALSE.)
         ! create mo overlap matrix
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_sigma(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="SIG", &
                                 size_keys=(/almo_mat_dim_occ, almo_mat_dim_occ/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=ispin, &
                                 init_domains=.FALSE.)
         ! create blocked mo overlap matrix
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_sigma_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="SIG_BLK", &
                                 size_keys=(/almo_mat_dim_occ, almo_mat_dim_occ/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
         ! create inverse mo overlap matrix
         CALL matrix_almo_create( &
            matrix_new=almo_scf_env%matrix_sigma_inv(ispin), &
            matrix_qs=matrix_s0, &
            almo_scf_env=almo_scf_env, &
            name_new="SIGINV", &
            size_keys=(/almo_mat_dim_occ, almo_mat_dim_occ/), &
            symmetry_new=dbcsr_type_symmetric, &
            spin_key=ispin, &
            init_domains=.FALSE.)
         ! create various templates that will be necessary later
         CALL matrix_almo_create( &
            matrix_new=almo_scf_env%matrix_t(ispin), &
            matrix_qs=matrix_s0, &
            almo_scf_env=almo_scf_env, &
            name_new="T", &
            size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_occ/), &
            symmetry_new=dbcsr_type_no_symmetry, &
            spin_key=ispin, &
            init_domains=.FALSE.)
         CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_sqrt(ispin))
         CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_sqrt_inv(ispin))
         CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_sqrt(ispin), &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
         CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_sqrt_inv(ispin), &
                              template=almo_scf_env%matrix_sigma(ispin), &
                              matrix_type=dbcsr_type_no_symmetry)
      ENDDO

      ! create virtual orbitals if necessary
      IF (almo_scf_env%need_virtuals) THEN
         ALLOCATE (almo_scf_env%matrix_v_blk(nspins))
         ALLOCATE (almo_scf_env%matrix_v_full_blk(nspins))
         ALLOCATE (almo_scf_env%matrix_v(nspins))
         ALLOCATE (almo_scf_env%matrix_vo(nspins))
         ALLOCATE (almo_scf_env%matrix_x(nspins))
         ALLOCATE (almo_scf_env%matrix_ov(nspins))
         ALLOCATE (almo_scf_env%matrix_ov_full(nspins))
         ALLOCATE (almo_scf_env%matrix_sigma_vv(nspins))
         ALLOCATE (almo_scf_env%matrix_sigma_vv_blk(nspins))
         ALLOCATE (almo_scf_env%matrix_sigma_vv_sqrt(nspins))
         ALLOCATE (almo_scf_env%matrix_sigma_vv_sqrt_inv(nspins))
         ALLOCATE (almo_scf_env%matrix_vv_full_blk(nspins))

         IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
            ALLOCATE (almo_scf_env%matrix_k_blk(nspins))
            ALLOCATE (almo_scf_env%matrix_k_blk_ones(nspins))
            ALLOCATE (almo_scf_env%matrix_k_tr(nspins))
            ALLOCATE (almo_scf_env%matrix_v_disc(nspins))
            ALLOCATE (almo_scf_env%matrix_v_disc_blk(nspins))
            ALLOCATE (almo_scf_env%matrix_ov_disc(nspins))
            ALLOCATE (almo_scf_env%matrix_vv_disc_blk(nspins))
            ALLOCATE (almo_scf_env%matrix_vv_disc(nspins))
            ALLOCATE (almo_scf_env%opt_k_t_dd(nspins))
            ALLOCATE (almo_scf_env%opt_k_t_rr(nspins))
            ALLOCATE (almo_scf_env%opt_k_denom(nspins))
         ENDIF

         DO ispin = 1, nspins
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_v_full_blk(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="V_FULL_BLK", &
                                    size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_virt_full/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_v_blk(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="V_BLK", &
                                    size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_virt/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_v(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="V", &
                                    size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_virt/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_ov_full(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="OV_FULL", &
                                    size_keys=(/almo_mat_dim_occ, almo_mat_dim_virt_full/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_ov(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="OV", &
                                    size_keys=(/almo_mat_dim_occ, almo_mat_dim_virt/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_vo(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="VO", &
                                    size_keys=(/almo_mat_dim_virt, almo_mat_dim_occ/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_x(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="VO", &
                                    size_keys=(/almo_mat_dim_virt, almo_mat_dim_occ/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_sigma_vv(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="SIG_VV", &
                                    size_keys=(/almo_mat_dim_virt, almo_mat_dim_virt/), &
                                    symmetry_new=dbcsr_type_symmetric, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_vv_full_blk(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="VV_FULL_BLK", &
                                    size_keys=(/almo_mat_dim_virt_full, almo_mat_dim_virt_full/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.TRUE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_sigma_vv_blk(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="SIG_VV_BLK", &
                                    size_keys=(/almo_mat_dim_virt, almo_mat_dim_virt/), &
                                    symmetry_new=dbcsr_type_symmetric, &
                                    spin_key=ispin, &
                                    init_domains=.TRUE.)
            CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_vv_sqrt(ispin))
            CALL cp_dbcsr_init(almo_scf_env%matrix_sigma_vv_sqrt_inv(ispin))
            CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_vv_sqrt(ispin), &
                                 template=almo_scf_env%matrix_sigma_vv(ispin), &
                                 matrix_type=dbcsr_type_no_symmetry)
            CALL cp_dbcsr_create(almo_scf_env%matrix_sigma_vv_sqrt_inv(ispin), &
                                 template=almo_scf_env%matrix_sigma_vv(ispin), &
                                 matrix_type=dbcsr_type_no_symmetry)

            IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
               CALL matrix_almo_create(matrix_new=almo_scf_env%opt_k_t_rr(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="OPT_K_U_RR", &
                                       size_keys=(/almo_mat_dim_virt, almo_mat_dim_virt/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_vv_disc(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="VV_DISC", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_symmetric, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%opt_k_t_dd(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="OPT_K_U_DD", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_vv_disc_blk(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="VV_DISC_BLK", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_symmetric, &
                                       spin_key=ispin, &
                                       init_domains=.TRUE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_k_blk(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="K_BLK", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.TRUE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_k_blk_ones(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="K_BLK_1", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.TRUE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%opt_k_denom(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="OPT_K_DENOM", &
                                       size_keys=(/almo_mat_dim_virt_disc, almo_mat_dim_virt/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_k_tr(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="K_TR", &
                                       size_keys=(/almo_mat_dim_virt, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_v_disc_blk(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="V_DISC_BLK", &
                                       size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_v_disc(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="V_DISC", &
                                       size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)
               CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_ov_disc(ispin), &
                                       matrix_qs=matrix_s0, &
                                       almo_scf_env=almo_scf_env, &
                                       name_new="OV_DISC", &
                                       size_keys=(/almo_mat_dim_occ, almo_mat_dim_virt_disc/), &
                                       symmetry_new=dbcsr_type_no_symmetry, &
                                       spin_key=ispin, &
                                       init_domains=.FALSE.)

            ENDIF ! end need_discarded_virtuals

         ENDDO ! spin
      ENDIF

      ! create matrices of orbital energies if necessary
      IF (almo_scf_env%need_orbital_energies) THEN
         ALLOCATE (almo_scf_env%matrix_eoo(nspins))
         ALLOCATE (almo_scf_env%matrix_evv_full(nspins))
         DO ispin = 1, nspins
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_eoo(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="E_OCC", &
                                    size_keys=(/almo_mat_dim_occ, almo_mat_dim_occ/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
            CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_evv_full(ispin), &
                                    matrix_qs=matrix_s0, &
                                    almo_scf_env=almo_scf_env, &
                                    name_new="E_VIRT", &
                                    size_keys=(/almo_mat_dim_virt_full, almo_mat_dim_virt_full/), &
                                    symmetry_new=dbcsr_type_no_symmetry, &
                                    spin_key=ispin, &
                                    init_domains=.FALSE.)
         ENDDO
      ENDIF

      ! Density and KS matrices
      ALLOCATE (almo_scf_env%matrix_p(nspins))
      ALLOCATE (almo_scf_env%matrix_p_blk(nspins))
      ALLOCATE (almo_scf_env%matrix_ks(nspins))
      ALLOCATE (almo_scf_env%matrix_ks_blk(nspins))
      IF (almo_scf_env%need_previous_ks) &
         ALLOCATE (almo_scf_env%matrix_ks_almo_scf_converged(nspins))
      DO ispin = 1, nspins
         CALL cp_dbcsr_init(almo_scf_env%matrix_p(ispin))
         ! RZK-warning copy with symmery but remember that this might cause problems
         CALL cp_dbcsr_create(almo_scf_env%matrix_p(ispin), &
                              template=almo_scf_env%matrix_s(1), &
                              matrix_type=dbcsr_type_symmetric)
         CALL cp_dbcsr_init(almo_scf_env%matrix_ks(ispin))
         CALL cp_dbcsr_create(almo_scf_env%matrix_ks(ispin), &
                              template=almo_scf_env%matrix_s(1), &
                              matrix_type=dbcsr_type_symmetric)
         IF (almo_scf_env%need_previous_ks) THEN
            CALL cp_dbcsr_init(almo_scf_env%matrix_ks_almo_scf_converged(ispin))
            CALL cp_dbcsr_create(almo_scf_env%matrix_ks_almo_scf_converged(ispin), &
                                 template=almo_scf_env%matrix_s(1), &
                                 matrix_type=dbcsr_type_symmetric)
         ENDIF
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_p_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="P_BLK", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
         CALL matrix_almo_create(matrix_new=almo_scf_env%matrix_ks_blk(ispin), &
                                 matrix_qs=matrix_s0, &
                                 almo_scf_env=almo_scf_env, &
                                 name_new="KS_BLK", &
                                 size_keys=(/almo_mat_dim_aobasis, almo_mat_dim_aobasis/), &
                                 symmetry_new=dbcsr_type_symmetric, &
                                 spin_key=ispin, &
                                 init_domains=.TRUE.)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE almo_scf_env_create_matrices

! **************************************************************************************************
!> \brief clean up procedures for almo scf
!> \param almo_scf_env ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE almo_scf_clean_up(almo_scf_env)

      TYPE(almo_scf_env_type)                            :: almo_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'almo_scf_clean_up', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      ! get a useful output_unit
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! release matrices
      CALL cp_dbcsr_release(almo_scf_env%matrix_s(1))
      CALL cp_dbcsr_release(almo_scf_env%matrix_s_blk(1))
      IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_diag) THEN
         CALL cp_dbcsr_release(almo_scf_env%matrix_s_blk_sqrt_inv(1))
         CALL cp_dbcsr_release(almo_scf_env%matrix_s_blk_sqrt(1))
      ELSE IF (almo_scf_env%almo_update_algorithm .EQ. almo_scf_dm_sign) THEN
         CALL cp_dbcsr_release(almo_scf_env%matrix_s_blk_inv(1))
      ENDIF
      DO ispin = 1, almo_scf_env%nspins
         CALL cp_dbcsr_release(almo_scf_env%quench_t(ispin))
         CALL cp_dbcsr_release(almo_scf_env%quench_t_blk(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_t_blk(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_err_blk(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_err_xx(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_t_tr(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_sigma(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_blk(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_inv(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_t(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_sqrt(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_sqrt_inv(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_p(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_ks(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_p_blk(ispin))
         CALL cp_dbcsr_release(almo_scf_env%matrix_ks_blk(ispin))
         IF (almo_scf_env%need_previous_ks) THEN
            CALL cp_dbcsr_release(almo_scf_env%matrix_ks_almo_scf_converged(ispin))
         ENDIF
         IF (almo_scf_env%need_virtuals) THEN
            CALL cp_dbcsr_release(almo_scf_env%matrix_v_blk(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_v_full_blk(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_v(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_vo(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_x(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_ov(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_ov_full(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_vv(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_vv_blk(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_vv_sqrt(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_sigma_vv_sqrt_inv(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_vv_full_blk(ispin))
            IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
               CALL cp_dbcsr_release(almo_scf_env%matrix_k_tr(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_k_blk(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_k_blk_ones(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_v_disc(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_v_disc_blk(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_ov_disc(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_vv_disc_blk(ispin))
               CALL cp_dbcsr_release(almo_scf_env%matrix_vv_disc(ispin))
               CALL cp_dbcsr_release(almo_scf_env%opt_k_t_dd(ispin))
               CALL cp_dbcsr_release(almo_scf_env%opt_k_t_rr(ispin))
               CALL cp_dbcsr_release(almo_scf_env%opt_k_denom(ispin))
            ENDIF
         ENDIF
         IF (almo_scf_env%need_orbital_energies) THEN
            CALL cp_dbcsr_release(almo_scf_env%matrix_eoo(ispin))
            CALL cp_dbcsr_release(almo_scf_env%matrix_evv_full(ispin))
         ENDIF
      ENDDO

      ! deallocate matrices
      DEALLOCATE (almo_scf_env%matrix_p)
      DEALLOCATE (almo_scf_env%matrix_p_blk)
      DEALLOCATE (almo_scf_env%matrix_ks)
      DEALLOCATE (almo_scf_env%matrix_ks_blk)
      DEALLOCATE (almo_scf_env%matrix_t_blk)
      DEALLOCATE (almo_scf_env%matrix_err_blk)
      DEALLOCATE (almo_scf_env%matrix_err_xx)
      DEALLOCATE (almo_scf_env%matrix_t)
      DEALLOCATE (almo_scf_env%matrix_t_tr)
      DEALLOCATE (almo_scf_env%matrix_sigma)
      DEALLOCATE (almo_scf_env%matrix_sigma_blk)
      DEALLOCATE (almo_scf_env%matrix_sigma_sqrt)
      DEALLOCATE (almo_scf_env%matrix_sigma_sqrt_inv)
      DEALLOCATE (almo_scf_env%matrix_sigma_inv)
      DEALLOCATE (almo_scf_env%quench_t)
      DEALLOCATE (almo_scf_env%quench_t_blk)
      IF (almo_scf_env%need_virtuals) THEN
         DEALLOCATE (almo_scf_env%matrix_v_blk)
         DEALLOCATE (almo_scf_env%matrix_v_full_blk)
         DEALLOCATE (almo_scf_env%matrix_v)
         DEALLOCATE (almo_scf_env%matrix_vo)
         DEALLOCATE (almo_scf_env%matrix_x)
         DEALLOCATE (almo_scf_env%matrix_ov)
         DEALLOCATE (almo_scf_env%matrix_ov_full)
         DEALLOCATE (almo_scf_env%matrix_sigma_vv)
         DEALLOCATE (almo_scf_env%matrix_sigma_vv_blk)
         DEALLOCATE (almo_scf_env%matrix_sigma_vv_sqrt)
         DEALLOCATE (almo_scf_env%matrix_sigma_vv_sqrt_inv)
         DEALLOCATE (almo_scf_env%matrix_vv_full_blk)
         IF (almo_scf_env%deloc_truncate_virt .NE. virt_full) THEN
            DEALLOCATE (almo_scf_env%matrix_k_tr)
            DEALLOCATE (almo_scf_env%matrix_k_blk)
            DEALLOCATE (almo_scf_env%matrix_v_disc)
            DEALLOCATE (almo_scf_env%matrix_v_disc_blk)
            DEALLOCATE (almo_scf_env%matrix_ov_disc)
            DEALLOCATE (almo_scf_env%matrix_vv_disc_blk)
            DEALLOCATE (almo_scf_env%matrix_vv_disc)
            DEALLOCATE (almo_scf_env%matrix_k_blk_ones)
            DEALLOCATE (almo_scf_env%opt_k_t_dd)
            DEALLOCATE (almo_scf_env%opt_k_t_rr)
            DEALLOCATE (almo_scf_env%opt_k_denom)
         ENDIF
      ENDIF
      IF (almo_scf_env%need_previous_ks) THEN
         DEALLOCATE (almo_scf_env%matrix_ks_almo_scf_converged)
      ENDIF
      IF (almo_scf_env%need_orbital_energies) THEN
         DEALLOCATE (almo_scf_env%matrix_eoo)
         DEALLOCATE (almo_scf_env%matrix_evv_full)
      ENDIF

      ! clean up other variables
      DO ispin = 1, almo_scf_env%nspins
         CALL release_submatrices( &
            almo_scf_env%domain_preconditioner(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_s_inv(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_s_sqrt_inv(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_s_sqrt(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_ks_xx(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_t(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_err(:, ispin))
         CALL release_submatrices(almo_scf_env%domain_r_down_up(:, ispin))
      ENDDO
      DEALLOCATE (almo_scf_env%domain_preconditioner)
      DEALLOCATE (almo_scf_env%domain_s_inv)
      DEALLOCATE (almo_scf_env%domain_s_sqrt_inv)
      DEALLOCATE (almo_scf_env%domain_s_sqrt)
      DEALLOCATE (almo_scf_env%domain_ks_xx)
      DEALLOCATE (almo_scf_env%domain_t)
      DEALLOCATE (almo_scf_env%domain_err)
      DEALLOCATE (almo_scf_env%domain_r_down_up)
      DO ispin = 1, almo_scf_env%nspins
         DEALLOCATE (almo_scf_env%domain_map(ispin)%pairs)
         DEALLOCATE (almo_scf_env%domain_map(ispin)%index1)
      ENDDO
      DEALLOCATE (almo_scf_env%domain_map)
      DEALLOCATE (almo_scf_env%domain_index_of_ao)
      DEALLOCATE (almo_scf_env%domain_index_of_atom)
      DEALLOCATE (almo_scf_env%first_atom_of_domain)
      DEALLOCATE (almo_scf_env%last_atom_of_domain)
      DEALLOCATE (almo_scf_env%nbasis_of_domain)
      DEALLOCATE (almo_scf_env%nocc_of_domain)
      DEALLOCATE (almo_scf_env%nvirt_full_of_domain)
      DEALLOCATE (almo_scf_env%nvirt_of_domain)
      DEALLOCATE (almo_scf_env%nvirt_disc_of_domain)
      DEALLOCATE (almo_scf_env%mu_of_domain)
      DEALLOCATE (almo_scf_env%cpu_of_domain)
      DEALLOCATE (almo_scf_env%charge_of_domain)
      DEALLOCATE (almo_scf_env%multiplicity_of_domain)

      DEALLOCATE (almo_scf_env%domain_index_of_ao_block)
      DEALLOCATE (almo_scf_env%domain_index_of_mo_block)

      CALL cp_para_env_release(almo_scf_env%para_env)
      CALL cp_blacs_env_release(almo_scf_env%blacs_env)

      CALL timestop(handle)

   END SUBROUTINE almo_scf_clean_up

END MODULE almo_scf

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Routines to handle an external density
!>        The external density can be generic and is provided by user input
!> \author D. Varsano
! **************************************************************************************************
MODULE qs_external_density
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE gaussian_gridlevels,             ONLY: gridlevel_info_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_recv,&
                                              mp_send,&
                                              mp_sync
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_integrate_function
   USE pw_types,                        ONLY: pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              rs_grid_create,&
                                              rs_grid_zero
   USE rs_pw_interface,                 ONLY: density_rs2pw
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_external_density'

   PUBLIC :: external_read_density

CONTAINS

! **************************************************************************************************
!> \brief  Computes the external density on the grid
!> \param qs_env ...
!> \date   03.2011
!> \author D. Varsano
! **************************************************************************************************
   SUBROUTINE external_read_density(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'external_read_density', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: filename
      INTEGER                                            :: extunit, gid, handle, i, igrid_level, &
                                                            ip, j, k, master, my_rank, nat, ndum, &
                                                            num_pe, tag
      INTEGER, DIMENSION(3)                              :: lbounds, lbounds_local, npoints, &
                                                            npoints_local, ubounds, ubounds_local
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: buffer
      REAL(kind=dp), DIMENSION(3)                        :: dr, rdum
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r_ext
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_ext_g, rho_ext_r
      TYPE(qs_rho_type), POINTER                         :: rho_external
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_rho_ext
      TYPE(section_vals_type), POINTER                   :: ext_den_section, input

      CALL timeset(routineN, handle)
      NULLIFY (cell, rs_rho_ext, input, ext_den_section, rs_descs, dft_control)
      NULLIFY (rho_ext_r, rho_ext_g, tot_rho_r_ext)

      CALL get_qs_env(qs_env, &
                      cell=cell, &
                      rho_external=rho_external, &
                      input=input, &
                      pw_env=pw_env, &
                      dft_control=dft_control)

      IF (dft_control%apply_external_density) THEN
         CALL qs_rho_get(rho_external, &
                         rho_r=rho_ext_r, &
                         rho_g=rho_ext_g, &
                         tot_rho_r=tot_rho_r_ext)

         gridlevel_info => pw_env%gridlevel_info

         CALL pw_env_get(pw_env, rs_descs=rs_descs)

         ALLOCATE (rs_rho_ext(gridlevel_info%ngrid_levels))

         DO igrid_level = 1, gridlevel_info%ngrid_levels
            CALL rs_grid_create(rs_rho_ext(igrid_level)%rs_grid, &
                                rs_descs(igrid_level)%rs_desc)
            CALL rs_grid_zero(rs_rho_ext(igrid_level)%rs_grid)
         ENDDO

         igrid_level = igrid_level-1

         ext_den_section => section_vals_get_subs_vals(input, "DFT%EXTERNAL_DENSITY")
         CALL section_vals_val_get(ext_den_section, "FILE_DENSITY", c_val=filename)

         gid = rho_ext_r(1)%pw%pw_grid%para%group
         my_rank = rho_ext_r(1)%pw%pw_grid%para%my_pos
         num_pe = rho_ext_r(1)%pw%pw_grid%para%group_size
         tag = 1

         IF (dft_control%read_external_density) THEN

            DO i = 1, 3
               dr(i) = rs_descs(igrid_level)%rs_desc%dh(i, i)
            END DO
            npoints = rs_descs(igrid_level)%rs_desc%npts
            lbounds = rs_descs(igrid_level)%rs_desc%lb
            ubounds = rs_descs(igrid_level)%rs_desc%ub

            npoints_local = rho_ext_r(1)%pw%pw_grid%npts_local
            lbounds_local = rho_ext_r(1)%pw%pw_grid%bounds_local(1, :)
            ubounds_local = rho_ext_r(1)%pw%pw_grid%bounds_local(2, :)

!        IF ( my_rank .eq. 0 ) THEN
!          WRITE(*,*) my_rank,npoints, lbounds, ubounds
!          WRITE(*,*) my_rank,npoints_local
!          WRITE(*,*) my_rank,lbounds_local
!          WRITE(*,*) my_rank,ubounds_local
!        END IF

            ALLOCATE (buffer(lbounds_local(3):ubounds_local(3)))

            IF (my_rank == 0) THEN
               WRITE (*, FMT="(/,/,T2,A)") "INITIALIZING ZMP CONSTRAINED DENSITY METHOD"
               WRITE (*, FMT="(/,(T3,A,T51,A30))") "ZMP| Reading the target density:     ", filename

               CALL open_file(file_name=filename, &
                              file_status="OLD", &
                              file_form="FORMATTED", &
                              file_action="READ", &
                              unit_number=extunit)

               DO i = 1, 2
                  READ (extunit, *)
               ENDDO
               READ (extunit, *) nat, rdum
               DO i = 1, 3
                  READ (extunit, *) ndum, rdum
                  IF (ndum /= npoints(i) .OR. (ABS(rdum(i)-dr(i)) > 1e-4)) THEN
                     WRITE (*, *) "ZMP | ERROR! | CUBE FILE NOT COINCIDENT WITH INTERNAL GRID ", i
                     WRITE (*, *) "ZMP | ", ndum, " DIFFERS FROM ", npoints(i)
                     WRITE (*, *) "ZMP | ", rdum, " DIFFERS FROM ", dr(i)
                  ENDIF
               ENDDO
               DO i = 1, nat
                  READ (extunit, *)
               ENDDO
            ENDIF

            DO i = lbounds(1), ubounds(1)
               DO j = lbounds(2), ubounds(2)
                  IF (my_rank .EQ. 0) THEN
                     READ (extunit, *) (buffer(k), k=lbounds(3), ubounds(3))
                     IF (num_pe .GT. 1) THEN
                        DO ip = 1, num_pe-1
                           CALL mp_send(buffer(lbounds(3):ubounds(3)), ip, tag, gid)
!                  WRITE(*,*) my_rank," sending to : ",ip
                        ENDDO
!                CALL mp_sync(gid)
                     ENDIF
                  ELSE
                     master = 0
                     CALL mp_recv(buffer(lbounds(3):ubounds(3)), master, tag, gid)
!               WRITE(*,*) my_rank," recieving from master"
                  ENDIF
                  CALL mp_sync(gid)

                  IF ((lbounds_local(1) .LE. i) .AND. (i .LE. ubounds_local(1)) .AND. (lbounds_local(2) .LE. j) &
                      .AND. (j .LE. ubounds_local(2))) THEN
                     rs_rho_ext(igrid_level)%rs_grid%r(i, j, lbounds(3):ubounds(3)) = buffer(lbounds(3):ubounds(3))
                  ENDIF

               ENDDO
            ENDDO
            IF (my_rank == 0) CALL close_file(unit_number=extunit)
         ENDIF

         CALL density_rs2pw(pw_env, rs_rho_ext, rho=rho_ext_r(1), rho_gspace=rho_ext_g(1))
         tot_rho_r_ext(1) = pw_integrate_function(rho_ext_r(1)%pw, isign=1)
         IF (my_rank == 0) THEN
            WRITE (*, FMT="(T3,A,T61,F20.10)") "ZMP| Total external charge:                    ", &
               tot_rho_r_ext(1)
         ENDIF
         DEALLOCATE (buffer, rs_rho_ext)
         CALL mp_sync(gid)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE external_read_density

END MODULE qs_external_density

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief localize wavefunctions
!>      linear response scf
!> \par History
!>      created 07-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_methods
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_checksum,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_external_control,             ONLY: external_control
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add,&
                                              cp_fm_trace
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE hartree_local_methods,           ONLY: Vh_1c_gg_integrals
   USE input_constants,                 ONLY: do_loc_none,&
                                              op_loc_berry,&
                                              ot_precond_none,&
                                              ot_precond_solver_default,&
                                              state_loc_all
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_rval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_bcast
   USE preconditioner,                  ONLY: apply_preconditioner,&
                                              make_preconditioner
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_retain
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
   USE qs_kpp1_env_types,               ONLY: qs_kpp1_env_type
   USE qs_ks_atom,                      ONLY: update_ks_atom
   USE qs_linres_types,                 ONLY: linres_control_type
   USE qs_loc_methods,                  ONLY: qs_loc_driver
   USE qs_loc_types,                    ONLY: get_qs_loc_env,&
                                              localized_wfn_control_type,&
                                              qs_loc_env_create,&
                                              qs_loc_env_new_type,&
                                              qs_loc_env_release,&
                                              qs_loc_env_retain
   USE qs_loc_utils,                    ONLY: loc_write_restart,&
                                              qs_loc_control_init,&
                                              qs_loc_init
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho0_ggrid,                   ONLY: integrate_vhg0_rspace
   USE qs_rho_methods,                  ONLY: qs_rho_rebuild,&
                                              qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_vxc_atom,                     ONLY: calculate_xc_2nd_deriv_atom
   USE string_utilities,                ONLY: xstring
   USE xc,                              ONLY: xc_calc_2nd_deriv,&
                                              xc_prep_2nd_deriv
   USE xc_derivatives,                  ONLY: xc_functionals_get_needs
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type,&
                                              xc_rho_set_update
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public subroutines ***
   PUBLIC :: linres_localize, linres_solver
   PUBLIC :: linres_write_restart, linres_read_restart

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_methods'

CONTAINS

! **************************************************************************************************
!> \brief Find the centers and spreads of the wfn,
!>      if required apply a localization algorithm
!> \param qs_env ...
!> \param linres_control ...
!> \param nspins ...
!> \param centers_only ...
!> \par History
!>      07.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE linres_localize(qs_env, linres_control, nspins, centers_only)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(linres_control_type), POINTER                 :: linres_control
      INTEGER, INTENT(IN)                                :: nspins
      LOGICAL, INTENT(IN), OPTIONAL                      :: centers_only

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_localize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, istate, nmoloc(2), output_unit
      LOGICAL                                            :: my_centers_only
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_localized
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(section_vals_type), POINTER                   :: loc_print_section, loc_section, &
                                                            lr_section

      NULLIFY (logger, lr_section, loc_section, loc_print_section, localized_wfn_control)
      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      loc_section => section_vals_get_subs_vals(lr_section, "LOCALIZE")
      loc_print_section => section_vals_get_subs_vals(lr_section, "LOCALIZE%PRINT")
      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      my_centers_only = .FALSE.
      IF (PRESENT(centers_only)) my_centers_only = centers_only

      NULLIFY (mos, mo_coeff, qs_loc_env, mos_localized)
      CALL get_qs_env(qs_env=qs_env, mos=mos)
      CALL qs_loc_env_create(qs_loc_env)
      CALL qs_loc_env_retain(qs_loc_env)
      linres_control%qs_loc_env => qs_loc_env
      CALL qs_loc_env_release(qs_loc_env)
      qs_loc_env => linres_control%qs_loc_env
      CALL qs_loc_control_init(qs_loc_env, loc_section, do_homo=.TRUE.)
      CALL get_qs_loc_env(qs_loc_env, localized_wfn_control=localized_wfn_control)

      ALLOCATE (mos_localized(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_create(mos_localized(ispin)%matrix, mo_coeff%matrix_struct)
         CALL cp_fm_to_fm(mo_coeff, mos_localized(ispin)%matrix)
      END DO

      nmoloc(1:2) = 0
      IF (my_centers_only) THEN
         localized_wfn_control%set_of_states = state_loc_all
         localized_wfn_control%localization_method = do_loc_none
         localized_wfn_control%operator_type = op_loc_berry
      ENDIF

      CALL qs_loc_init(qs_env, qs_loc_env, loc_section, mos_localized=mos_localized, &
                       do_homo=.TRUE.)

      ! The orbital centers are stored in linres_control%localized_wfn_control
      DO ispin = 1, nspins
         CALL qs_loc_driver(qs_env, qs_loc_env, loc_print_section, myspin=ispin, &
                            ext_mo_coeff=mos_localized(ispin)%matrix)
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_to_fm(mos_localized(ispin)%matrix, mo_coeff)
      END DO

      CALL loc_write_restart(qs_loc_env, loc_print_section, mos, &
                             mos_localized, do_homo=.TRUE.)
      CALL cp_fm_vect_dealloc(mos_localized)

      ! Write Centers and Spreads on std out
      IF (output_unit > 0) THEN
         DO ispin = 1, nspins
            WRITE (output_unit, "(/,T2,A,I2)") &
               "WANNIER CENTERS for spin ", ispin
            WRITE (output_unit, "(/,T18,A,3X,A)") &
               "--------------- Centers --------------- ", &
               "--- Spreads ---"
            DO istate = 1, SIZE(localized_wfn_control%centers_set(ispin)%array, 2)
               WRITE (output_unit, "(T5,A6,I6,2X,3f12.6,5X,f12.6)") &
                  'state ', istate, localized_wfn_control%centers_set(ispin)%array(1:3, istate), &
                  localized_wfn_control%centers_set(ispin)%array(4, istate)
            END DO
         END DO
         CALL m_flush(output_unit)
      END IF

   END SUBROUTINE linres_localize

! **************************************************************************************************
!> \brief scf loop to optimize the first order wavefunctions (psi1)
!>      given a perturbation as an operator applied to the ground
!>      state orbitals (h1_psi0)
!> \param p_env ...
!> \param qs_env ...
!> \param psi1 ...
!> \param h1_psi0 ...
!> \param psi0_order ...
!> \param should_stop ...
!> \par History
!>      07.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
      !
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: psi1, h1_psi0, psi0_order
      LOGICAL, INTENT(OUT)                               :: should_stop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_solver', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, iter, maxnmo, maxnmo_o, &
                                                            nao, ncol, nmo, nspins, output_unit
      LOGICAL                                            :: restart
      REAL(dp)                                           :: norm_res, t1, t2
      REAL(dp), DIMENSION(:), POINTER                    :: alpha, beta, tr_pAp, tr_rz0, tr_rz00, &
                                                            tr_rz1
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, matrix_t
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: Ap, chc, mo_coeff_array, p, r, Sc, z
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: buf, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: lr_section

!
!

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, linres_control, matrix_s, matrix_t, matrix_ks, para_env)
      NULLIFY (Ap, r, p, z, lr_section, logger, buf, mos, tmp_fm_struct, mo_coeff)
      NULLIFY (Sc, chc)

      logger => cp_get_default_logger()
      t1 = m_walltime()

      CALL get_qs_env(qs_env=qs_env, &
                      matrix_ks=matrix_ks, &
                      matrix_s=matrix_s, &
                      kinetic=matrix_t, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      para_env=para_env, &
                      mos=mos)

      !
      nspins = dft_control%nspins
      CALL get_mo_set(mos(1)%mo_set, nao=nao)
      maxnmo = 0
      maxnmo_o = 0
      DO ispin = 1, nspins
         CALL get_mo_set(mos(ispin)%mo_set, nmo=ncol)
         maxnmo = MAX(maxnmo, ncol)
         CALL cp_fm_get_info(psi0_order(ispin)%matrix, ncol_global=ncol)
         maxnmo_o = MAX(maxnmo_o, ncol)
      ENDDO
      !
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      !
      CALL check_p_env_init(p_env, linres_control, nspins)
      !
      ! allocate the vectors
      ALLOCATE (alpha(nspins), beta(nspins), tr_pAp(nspins), tr_rz0(nspins), tr_rz00(nspins), tr_rz1(nspins), &
                r(nspins), p(nspins), z(nspins), Ap(nspins), mo_coeff_array(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
         mo_coeff_array(ispin)%matrix => mo_coeff
      ENDDO
      !
      DO ispin = 1, nspins
         NULLIFY (r(ispin)%matrix, p(ispin)%matrix, z(ispin)%matrix, Ap(ispin)%matrix)
         CALL cp_fm_create(r(ispin)%matrix, psi1(ispin)%matrix%matrix_struct)
         CALL cp_fm_create(p(ispin)%matrix, psi1(ispin)%matrix%matrix_struct)
         CALL cp_fm_create(z(ispin)%matrix, psi1(ispin)%matrix%matrix_struct)
         CALL cp_fm_create(Ap(ispin)%matrix, psi1(ispin)%matrix%matrix_struct)
      ENDDO
      !
      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=maxnmo, para_env=para_env, &
                               context=psi1(1)%matrix%matrix_struct%context)
      CALL cp_fm_create(buf, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)
      !
      !
      !
      ! compute S*C0, C0_order'*H*C0_order (this should be done once for all)
      ALLOCATE (chc(nspins), Sc(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         CALL cp_fm_create(Sc(ispin)%matrix, mo_coeff%matrix_struct)
         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmo, &
                                  ncol_global=nmo, para_env=para_env, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(chc(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      ENDDO
      !
      DO ispin = 1, nspins
         !
         ! C0_order' * H * C0_order
         mo_coeff => psi0_order(ispin)%matrix
         CALL cp_fm_get_info(mo_coeff, ncol_global=ncol)
         CALL cp_dbcsr_sm_fm_multiply(matrix_ks(ispin)%matrix, mo_coeff, buf, ncol)
         CALL cp_gemm('T', 'N', ncol, ncol, nao, -1.0_dp, mo_coeff, buf, 0.0_dp, chc(ispin)%matrix)
         !
         ! S * C0
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=ncol)
         CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, mo_coeff, Sc(ispin)%matrix, ncol)
      ENDDO
      !
      !
      !
      ! header
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T3,A,T16,A,T25,A,T38,A,T52,A,T72,A,/,T3,A)") &
            "Iteration", "Method", "Restart", "Stepsize", "Convergence", "Time", &
            REPEAT("-", 80)
      ENDIF
      !
      ! orthogonalize x with respect to the psi0
      CALL preortho(psi1, mo_coeff_array, Sc, buf)
      !
      ! build the preconditioner
      IF (linres_control%preconditioner_type /= ot_precond_none) THEN
         IF (p_env%new_preconditioner) THEN
            p_env%os_valid = .FALSE.
            DO ispin = 1, nspins
               CALL make_preconditioner(p_env%preconditioner(ispin), &
                                        linres_control%preconditioner_type, ot_precond_solver_default, &
                                        matrix_ks(ispin)%matrix, matrix_s(1)%matrix, matrix_t(1)%matrix, &
                                        mos(ispin)%mo_set, linres_control%energy_gap)
            ENDDO
            p_env%new_preconditioner = .FALSE.
         ENDIF
      ENDIF
      !
      ! initalization of the linear solver
      !
      ! A * x0
      CALL apply_op(qs_env, p_env, psi0_order, psi1, Ap, chc, buf)
      !
      !
      ! r_0 = b - Ax0
      DO ispin = 1, nspins
         CALL cp_fm_to_fm(h1_psi0(ispin)%matrix, r(ispin)%matrix)
         CALL cp_fm_scale_and_add(-1.0_dp, r(ispin)%matrix, -1.0_dp, Ap(ispin)%matrix)
      ENDDO
      !
      ! proj r
      CALL postortho(r, mo_coeff_array, Sc, buf)
      !
      ! preconditioner
      linres_control%flag = ""
      IF (linres_control%preconditioner_type .EQ. ot_precond_none) THEN
         !
         ! z_0 = r_0
         DO ispin = 1, nspins
            CALL cp_fm_to_fm(r(ispin)%matrix, z(ispin)%matrix)
         ENDDO
         linres_control%flag = "CG"
      ELSE
         !
         ! z_0 = M * r_0
         DO ispin = 1, nspins
            CALL apply_preconditioner(p_env%preconditioner(ispin), r(ispin)%matrix, &
                                      z(ispin)%matrix)
         ENDDO
         linres_control%flag = "PCG"
      ENDIF
      !
      norm_res = 0.0_dp
      DO ispin = 1, nspins
         !
         ! p_0 = z_0
         CALL cp_fm_to_fm(z(ispin)%matrix, p(ispin)%matrix)
         !
         ! trace(r_0 * z_0)
         CALL cp_fm_trace(r(ispin)%matrix, z(ispin)%matrix, tr_rz0(ispin))
         IF (tr_rz0(ispin) .LT. 0.0_dp) CALL cp__b("qs_linres_methods.F",439,"tr(r_j*z_j) < 0")
         norm_res = MAX(norm_res, ABS(tr_rz0(ispin))/SQRT(REAL(nao*maxnmo_o, dp)))
      ENDDO
      !
      !
      alpha(:) = 0.0_dp
      restart = .FALSE.
      should_stop = .FALSE.
      iteration: DO iter = 1, linres_control%max_iter
         !
         ! check convergence
         linres_control%converged = .FALSE.
         IF (norm_res .LT. linres_control%eps) THEN
            linres_control%converged = .TRUE.
         ENDIF
         !
         t2 = m_walltime()
         IF (iter .EQ. 1 .OR. MOD(iter, 1) .EQ. 0 .OR. linres_control%converged .OR. restart .OR. should_stop) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T5,I5,T18,A3,T28,L1,T38,1E8.2,T48,F16.10,T68,F8.2)") &
                  iter, linres_control%flag, restart, MAXVAL(alpha), norm_res, t2-t1
               CALL m_flush(output_unit)
            ENDIF
         ENDIF
         !
         IF (linres_control%converged) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(/,T2,A,I4,A,/)") "The linear solver converged in ", iter, " iterations."
               CALL m_flush(output_unit)
            ENDIF
            EXIT iteration
         ELSE IF (should_stop) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(/,T2,A,I4,A,/)") "The linear solver did NOT converge! External stop"
               CALL m_flush(output_unit)
            END IF
            EXIT iteration
         ENDIF
         !
         ! Max number of iteration reached
         IF (iter == linres_control%max_iter) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(/,T2,A/)") &
                  "The linear solver didnt converge! Maximum number of iterations reached."
               CALL m_flush(output_unit)
            ENDIF
            linres_control%converged = .FALSE.
         ENDIF
!       t1 = m_walltime()
         !
         !
         ! Apply the operators that do not depend on the perturbation
         CALL apply_op(qs_env, p_env, psi0_order, p, Ap, chc, buf)
         !
         !
!       ! proj Ap onto the virtual subspace
         CALL postortho(Ap, mo_coeff_array, Sc, buf)
         !
         !
         DO ispin = 1, nspins
            !
            ! tr(Ap_j*p_j)
            CALL cp_fm_trace(Ap(ispin)%matrix, p(ispin)%matrix, tr_pAp(ispin))
            IF (tr_pAp(ispin) .LT. 0.0_dp) THEN

               ! try to fix it by getting rid of the preconditioner
               IF (iter > 1) THEN
                  CALL cp_fm_scale_and_add(beta(ispin), p(ispin)%matrix, -1.0_dp, z(ispin)%matrix)
                  CALL cp_fm_trace(r(ispin)%matrix, r(ispin)%matrix, tr_rz1(ispin))
                  beta(ispin) = tr_rz1(ispin)/tr_rz00(ispin)
                  CALL cp_fm_scale_and_add(beta(ispin), p(ispin)%matrix, 1.0_dp, r(ispin)%matrix)
                  tr_rz0(ispin) = tr_rz1(ispin)
               ELSE
                  CALL cp_fm_to_fm(r(ispin)%matrix, p(ispin)%matrix)
                  CALL cp_fm_trace(r(ispin)%matrix, r(ispin)%matrix, tr_rz0(ispin))
               END IF
               linres_control%flag = "CG"

               CALL apply_op(qs_env, p_env, psi0_order, p, Ap, chc, buf)
               CALL postortho(Ap, mo_coeff_array, Sc, buf)
               CALL cp_fm_trace(Ap(ispin)%matrix, p(ispin)%matrix, tr_pAp(ispin))
               CALL cp__b("qs_linres_methods.F",520,"tr(Ap_j*p_j) < 0")
            END IF
            !
            ! alpha = tr(r_j*z_j) / tr(Ap_j*p_j)
            IF (tr_pAp(ispin) .LT. 1.0e-10_dp) THEN
               alpha(ispin) = 1.0_dp
            ELSE
               alpha(ispin) = tr_rz0(ispin)/tr_pAp(ispin)
            ENDIF
            !
            ! x_j+1 = x_j + alpha * p_j
            CALL cp_fm_scale_and_add(1.0_dp, psi1(ispin)%matrix, alpha(ispin), p(ispin)%matrix)
         ENDDO
         !
         ! need to recompute the residue
         restart = .FALSE.
         IF (MOD(iter, linres_control%restart_every) .EQ. 0) THEN
            !
            !
            ! r_j+1 = b - A * x_j+1
            CALL apply_op(qs_env, p_env, psi0_order, psi1, Ap, chc, buf)
            !
            !
            DO ispin = 1, nspins
               CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
               CALL cp_fm_to_fm(h1_psi0(ispin)%matrix, r(ispin)%matrix)
               CALL cp_fm_scale_and_add(-1.0_dp, r(ispin)%matrix, -1.0_dp, Ap(ispin)%matrix)
            ENDDO
            CALL postortho(r, mo_coeff_array, Sc, buf)
            !
            restart = .TRUE.
         ELSE
!       ! proj Ap onto the virtual subspace
            CALL postortho(Ap, mo_coeff_array, Sc, buf)
            !
            ! r_j+1 = r_j - alpha * Ap_j
            DO ispin = 1, nspins
               CALL cp_fm_scale_and_add(1.0_dp, r(ispin)%matrix, -alpha(ispin), Ap(ispin)%matrix)
            ENDDO
            restart = .FALSE.
         ENDIF
         !
         ! preconditioner
         linres_control%flag = ""
         IF (linres_control%preconditioner_type .EQ. ot_precond_none) THEN
            !
            ! z_j+1 = r_j+1
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(r(ispin)%matrix, z(ispin)%matrix)
            ENDDO
            linres_control%flag = "CG"
         ELSE
            !
            ! z_j+1 = M * r_j+1
            DO ispin = 1, nspins
               CALL apply_preconditioner(p_env%preconditioner(ispin), r(ispin)%matrix, &
                                         z(ispin)%matrix)
            ENDDO
            linres_control%flag = "PCG"
         ENDIF
         !
         norm_res = 0.0_dp
         DO ispin = 1, nspins
            !
            ! tr(r_j+1*z_j+1)
            CALL cp_fm_trace(r(ispin)%matrix, z(ispin)%matrix, tr_rz1(ispin))
            IF (tr_rz1(ispin) .LT. 0.0_dp) CALL cp__b("qs_linres_methods.F",586,"tr(r_j+1*z_j+1) < 0")
            norm_res = MAX(norm_res, tr_rz1(ispin)/SQRT(REAL(nao*maxnmo_o, dp)))
            !
            ! beta = tr(r_j+1*z_j+1) / tr(r_j*z_j)
            IF (tr_rz0(ispin) .LT. 1.0e-10_dp) THEN
               beta(ispin) = 0.0_dp
            ELSE
               beta(ispin) = tr_rz1(ispin)/tr_rz0(ispin)
            ENDIF
            !
            ! p_j+1 = z_j+1 + beta * p_j
            CALL cp_fm_scale_and_add(beta(ispin), p(ispin)%matrix, 1.0_dp, z(ispin)%matrix)
            tr_rz00(ispin) = tr_rz0(ispin)
            tr_rz0(ispin) = tr_rz1(ispin)
         ENDDO

         ! ** can we exit the SCF loop  ?
         CALL external_control(should_stop, "LINRES", target_time=qs_env%target_time, &
                               start_time=qs_env%start_time)

      ENDDO iteration
      !
      ! proj psi1
      CALL preortho(psi1, mo_coeff_array, Sc, buf)
      !
      ! clean up
      DO ispin = 1, nspins
         CALL cp_fm_release(r(ispin)%matrix)
         CALL cp_fm_release(p(ispin)%matrix)
         CALL cp_fm_release(z(ispin)%matrix)
         CALL cp_fm_release(Ap(ispin)%matrix)
         !
         CALL cp_fm_release(Sc(ispin)%matrix)
         CALL cp_fm_release(chc(ispin)%matrix)
      ENDDO
      CALL cp_fm_release(buf)
      DEALLOCATE (alpha, beta, tr_pAp, tr_rz0, tr_rz00, tr_rz1, r, p, z, Ap, Sc, chc, mo_coeff_array)
      !
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, "PRINT%PROGRAM_RUN_INFO")
      !
      CALL timestop(handle)
      !
   END SUBROUTINE linres_solver
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param p_env ...
!> \param c0 ...
!> \param v ...
!> \param Av ...
!> \param chc ...
!> \param buf ...
! **************************************************************************************************
   SUBROUTINE apply_op(qs_env, p_env, c0, v, Av, chc, buf)
      !
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: c0, v, Av, chc
      TYPE(cp_fm_type), POINTER                          :: buf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_op', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins
      REAL(dp)                                           :: chksum
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho1_ao
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(qs_rho_type), POINTER                         :: rho

!

      NULLIFY (dft_control, matrix_ks, matrix_s, linres_control, rho1_ao)
      !
      !
      CALL timeset(routineN, handle)
      !
      IF(.NOT.(ASSOCIATED(v)))CALL cp__a("qs_linres_methods.F",664)
      IF(.NOT.(ASSOCIATED(Av)))CALL cp__a("qs_linres_methods.F",665)
      IF(.NOT.(ASSOCIATED(chc)))CALL cp__a("qs_linres_methods.F",666)
      IF(.NOT.(ASSOCIATED(buf)))CALL cp__a("qs_linres_methods.F",667)
      !
      CALL get_qs_env(qs_env=qs_env, &
                      matrix_ks=matrix_ks, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      linres_control=linres_control)

      !
      nspins = dft_control%nspins
      !
      !
      ! apply the uncoupled operator
      DO ispin = 1, nspins
         CALL apply_op_1(v(ispin)%matrix, Av(ispin)%matrix, matrix_ks(ispin)%matrix, &
                         matrix_s(1)%matrix, chc(ispin)%matrix, buf)
      ENDDO

      IF (linres_control%do_kernel) THEN !.AND.chksum.GT.1.0E-10_dp) THEN
         !
         ! build DM
         CALL build_dm_response(c0, v, p_env%p1)

         chksum = 0.0_dp
         DO ispin = 1, nspins
            chksum = chksum+cp_dbcsr_checksum(p_env%p1(ispin)%matrix)
         ENDDO

         !
         ! skip the kernel if the DM is very small
         IF (chksum .GT. 1.0E-14_dp) THEN

            CALL p_env_check_i_alloc(p_env, qs_env)

            CALL qs_rho_get(p_env%rho1, rho_ao=rho1_ao)
            DO ispin = 1, nspins
               CALL cp_dbcsr_copy(rho1_ao(ispin)%matrix, p_env%p1(ispin)%matrix)
            ENDDO

            CALL qs_rho_update_rho(rho_struct=p_env%rho1, local_rho_set=p_env%local_rho_set, &
                                   qs_env=qs_env)

            !if(first_time) then
            CALL get_qs_env(qs_env, rho=rho) ! that could be called before
            CALL qs_rho_update_rho(rho, qs_env=qs_env) ! that could be called before
            !   first_time = .false.
            !endif

            CALL apply_op_2(qs_env, p_env, c0, v, Av, chc, buf)

            !CALL kpp1_calc_k_p_p1(p_env%kpp1_env, p_env, qs_env, p_env%kpp1, qs_env%rho, p_env%rho1, p_env%rho1_xc)
            !DO ispin = 1,nspins
            !   CALL cp_fm_get_info(c0(ispin)%matrix,ncol_global=ncol)
            !   CALL cp_sm_fm_multiply(sparse_matrix=p_env%kpp1(ispin)%matrix,&
            !                          v_in=c0(ispin)%matrix,&
            !                          v_out=Av(ispin)%matrix,&
            !                          ncol=ncol,alpha=1.0_dp,beta=1.0_dp)
            !ENDDO

         ENDIF

      ENDIF
      !
      CALL timestop(handle)
      !
   END SUBROUTINE apply_op
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param c0 ...
!> \param c1 ...
!> \param dm ...
! **************************************************************************************************
   SUBROUTINE build_dm_response(c0, c1, dm)
      !
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: c0, c1
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: dm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_dm_response', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, ncol, nspins

!
!

      IF(.NOT.(ASSOCIATED(c0)))CALL cp__a("qs_linres_methods.F",754)
      IF(.NOT.(ASSOCIATED(c1)))CALL cp__a("qs_linres_methods.F",755)
      IF(.NOT.(ASSOCIATED(dm)))CALL cp__a("qs_linres_methods.F",756)

      nspins = SIZE(dm, 1)

      DO ispin = 1, nspins
         CALL cp_dbcsr_set(dm(ispin)%matrix, 0.0_dp)
         CALL cp_fm_get_info(c0(ispin)%matrix, ncol_global=ncol)
         CALL cp_dbcsr_plus_fm_fm_t(dm(ispin)%matrix, &
                                    matrix_v=c0(ispin)%matrix, &
                                    matrix_g=c1(ispin)%matrix, &
                                    ncol=ncol, alpha=1.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(dm(ispin)%matrix, &
                                    matrix_v=c1(ispin)%matrix, &
                                    matrix_g=c0(ispin)%matrix, &
                                    ncol=ncol, alpha=1.0_dp)
      ENDDO

   END SUBROUTINE build_dm_response
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param v ...
!> \param Av ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param chc ...
!> \param buf ...
! **************************************************************************************************
   SUBROUTINE apply_op_1(v, Av, matrix_ks, matrix_s, chc, buf)
      !
      TYPE(cp_fm_type), POINTER                          :: v, Av
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_ks, matrix_s
      TYPE(cp_fm_type), POINTER                          :: chc, buf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_op_1', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ncol, nrow

!
!

      CALL timeset(routineN, handle)
      !
      IF(.NOT.(ASSOCIATED(v)))CALL cp__a("qs_linres_methods.F",800)
      IF(.NOT.(ASSOCIATED(Av)))CALL cp__a("qs_linres_methods.F",801)
      IF(.NOT.(ASSOCIATED(matrix_ks)))CALL cp__a("qs_linres_methods.F",802)
      IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("qs_linres_methods.F",803)
      IF(.NOT.(ASSOCIATED(chc)))CALL cp__a("qs_linres_methods.F",804)
      IF(.NOT.(ASSOCIATED(buf)))CALL cp__a("qs_linres_methods.F",805)
      !
      CALL cp_fm_get_info(v, ncol_global=ncol, nrow_global=nrow)
      ! H * v
      CALL cp_dbcsr_sm_fm_multiply(matrix_ks, v, Av, ncol)
      ! v * e  (chc already multiplied by -1)
      CALL cp_gemm('N', 'N', nrow, ncol, ncol, 1.0_dp, v, chc, 0.0_dp, buf)
      ! S * ve
      CALL cp_dbcsr_sm_fm_multiply(matrix_s, buf, Av, ncol, alpha=1.0_dp, beta=1.0_dp)
      !Results is H*C1 - S*<iHj>*C1
      !
      CALL timestop(handle)
      !
   END SUBROUTINE apply_op_1

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param p_env ...
!> \param c0 ...
!> \param v ...
!> \param Av ...
!> \param chc ...
!> \param buf ...
! **************************************************************************************************
   SUBROUTINE apply_op_2(qs_env, p_env, c0, v, Av, chc, buf)
      !
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: c0, v, Av, chc
      TYPE(cp_fm_type), POINTER                          :: buf

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_op_2', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, ncol, ns, nspins
      INTEGER, DIMENSION(2, 3)                           :: bo
      LOGICAL                                            :: gapw, gapw_xc, lr_triplet, lsd
      REAL(KIND=dp)                                      :: energy_hartree, energy_hartree_1c, fac
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho1_ao, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ksmat, psmat
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho1_tot_gspace, v_hartree_gspace, &
                                                            v_hartree_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho1_g, rho1_g_pw, rho1_r, rho1_r_pw, &
                                                            tau_pw, v_rspace_new, v_xc
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho, rho1, rho1_xc
      TYPE(section_vals_type), POINTER                   :: input, scf_section, xc_fun_section, &
                                                            xc_section
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho1_set

      CALL timeset(routineN, handle)

      NULLIFY (auxbas_pw_pool, pw_pools, pw_env, v_rspace_new, &
               rho1_r, rho1_g_pw, tau_pw, v_xc, rho1_set, rho1_ao, rho_ao, &
               poisson_env, input, scf_section, rho, dft_control, logger, rho1_g)
      logger => cp_get_default_logger()

      energy_hartree = 0.0_dp
      energy_hartree_1c = 0.0_dp

      IF(.NOT.(ASSOCIATED(c0)))CALL cp__a("qs_linres_methods.F",872)
      IF(.NOT.(ASSOCIATED(v)))CALL cp__a("qs_linres_methods.F",873)
      IF(.NOT.(ASSOCIATED(Av)))CALL cp__a("qs_linres_methods.F",874)
      IF(.NOT.(ASSOCIATED(chc)))CALL cp__a("qs_linres_methods.F",875)
      IF(.NOT.(ASSOCIATED(buf)))CALL cp__a("qs_linres_methods.F",876)

      IF(.NOT.(ASSOCIATED(p_env%kpp1_env)))CALL cp__a("qs_linres_methods.F",878)
      IF(.NOT.(ASSOCIATED(p_env%kpp1)))CALL cp__a("qs_linres_methods.F",879)
      rho1 => p_env%rho1
      rho1_xc => p_env%rho1_xc
      IF(.NOT.(ASSOCIATED(rho1)))CALL cp__a("qs_linres_methods.F",882)

      IF(.NOT.(p_env%kpp1_env%ref_count > 0))CALL cp__a("qs_linres_methods.F",884)

      CALL get_qs_env(qs_env=qs_env, &
                      pw_env=pw_env, &
                      input=input, &
                      rho=rho, &
                      linres_control=linres_control, &
                      dft_control=dft_control)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      lr_triplet = linres_control%lr_triplet
      CALL kpp1_check_i_alloc(p_env%kpp1_env, qs_env, lr_triplet)
      !gapw=(section_get_ival(input,"DFT%QS%METHOD")==do_method_gapw)
      !gapw_xc=(section_get_ival(input,"DFT%QS%METHOD")==do_method_gapw_xc)
      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      IF (gapw_xc) THEN
         IF(.NOT.(ASSOCIATED(rho1_xc)))CALL cp__a("qs_linres_methods.F",902)
      END IF

      nspins = SIZE(p_env%kpp1)
      lsd = (nspins == 2)

      xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      scf_section => section_vals_get_subs_vals(input, "DFT%SCF")

      p_env%kpp1_env%iter = p_env%kpp1_env%iter+1

! gets the tmp grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_linres_methods.F",914)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, poisson_env=poisson_env)
      ALLOCATE (v_rspace_new(nspins))
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      IF (gapw .OR. gapw_xc) &
         CALL prepare_gapw_den(qs_env, p_env%local_rho_set, do_rho0=(.NOT. gapw_xc))

! *** calculate the hartree potential on the total density ***
      CALL pw_pool_create_pw(auxbas_pw_pool, rho1_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL qs_rho_get(rho1, rho_g=rho1_g)
      CALL pw_copy(rho1_g(1)%pw, rho1_tot_gspace%pw)
      DO ispin = 2, nspins
         CALL pw_axpy(rho1_g(ispin)%pw, rho1_tot_gspace%pw)
      END DO
      IF (gapw) &
         CALL pw_axpy(p_env%local_rho_set%rho0_mpole%rho0_s_gs%pw, rho1_tot_gspace%pw)

      !IF (cp_print_key_should_output(logger%iter_info,scf_section,"PRINT%TOTAL_DENSITIES",&
      !     /=0) THEN
      !   output_unit = cp_print_key_unit_nr(logger,scf_section,"PRINT%TOTAL_DENSITIES",&
      !        extension=".scfLog")
      !   CALL print_densities(kpp1_env, rho1, rho1_tot_gspace, output_unit)
      !   CALL cp_print_key_finished_output(output_unit,logger,scf_section,&
      !        "PRINT%TOTAL_DENSITIES")
      !END IF

      IF (.NOT. (nspins == 1 .AND. lr_triplet)) THEN
         CALL pw_poisson_solve(poisson_env, rho1_tot_gspace%pw, &
                               energy_hartree, &
                               v_hartree_gspace%pw)
         CALL pw_transfer(v_hartree_gspace%pw, v_hartree_rspace%pw)
      ENDIF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho1_tot_gspace%pw)

! *** calculate the xc potential ***
      IF (gapw_xc) THEN
         CALL qs_rho_get(rho1_xc, rho_r=rho1_r)
      ELSE
         CALL qs_rho_get(rho1, rho_r=rho1_r)
      END IF

      IF (nspins == 1 .AND. (lr_triplet)) THEN

         lsd = .TRUE.
         ALLOCATE (rho1_r_pw(2))
         DO ispin = 1, 2
            NULLIFY (rho1_r_pw(ispin)%pw)
            CALL pw_create(rho1_r_pw(ispin)%pw, rho1_r(1)%pw%pw_grid, &
                           rho1_r(1)%pw%in_use, rho1_r(1)%pw%in_space)
            CALL pw_transfer(rho1_r(1)%pw, rho1_r_pw(ispin)%pw)
         END DO

      ELSE

         ALLOCATE (rho1_r_pw(nspins))
         DO ispin = 1, nspins
            rho1_r_pw(ispin)%pw => rho1_r(ispin)%pw
            CALL pw_retain(rho1_r_pw(ispin)%pw)
         END DO

      END IF

      NULLIFY (tau_pw)

      !------!
      ! rho1 !
      !------!
      bo = rho1_r(1)%pw%pw_grid%bounds_local
      ! create the place where to store the argument for the functionals
      CALL xc_rho_set_create(rho1_set, bo, &
                             rho_cutoff=section_get_rval(xc_section, "DENSITY_CUTOFF"), &
                             drho_cutoff=section_get_rval(xc_section, "GRADIENT_CUTOFF"), &
                             tau_cutoff=section_get_rval(xc_section, "TAU_CUTOFF"))

      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      needs = xc_functionals_get_needs(xc_fun_section, lsd, .TRUE.)

      ! calculate the arguments needed by the functionals
      CALL xc_rho_set_update(rho1_set, rho1_r_pw, rho1_g_pw, tau_pw, needs, &
                             section_get_ival(xc_section, "XC_GRID%XC_DERIV"), &
                             section_get_ival(xc_section, "XC_GRID%XC_SMOOTH_RHO"), &
                             auxbas_pw_pool)

      ALLOCATE (v_xc(nspins))
      DO ispin = 1, nspins
         NULLIFY (v_xc(ispin)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, v_xc(ispin)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_zero(v_xc(ispin)%pw)
      END DO

      fac = 0._dp
      IF (nspins == 1) THEN
         IF (lr_triplet) fac = -1.0_dp
      END IF

      CALL xc_calc_2nd_deriv(v_xc, p_env%kpp1_env%deriv_set, p_env%kpp1_env%rho_set, &
                             rho1_set, auxbas_pw_pool, xc_section=xc_section, &
                             tddfpt_fac=fac)

      DO ispin = 1, nspins
         v_rspace_new(ispin)%pw => v_xc(ispin)%pw
      END DO
      DEALLOCATE (v_xc)

      IF (gapw) CALL calculate_xc_2nd_deriv_atom(p_env, qs_env, xc_section, do_tddft=.FALSE., &
                                                 do_triplet=lr_triplet)

      CALL xc_rho_set_release(rho1_set)

      DO ispin = 1, SIZE(rho1_r_pw)
         CALL pw_release(rho1_r_pw(ispin)%pw)
      END DO
      DEALLOCATE (rho1_r_pw)

      !-------------------------------!
      ! Add both hartree and xc terms !
      !-------------------------------!
      DO ispin = 1, nspins

         IF (gapw_xc) THEN
            ! XC and Hartree are integrated separatedly
            ! XC uses the sofft basis set only
            v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d* &
                                          v_rspace_new(ispin)%pw%pw_grid%dvol

            IF (nspins == 1) THEN

               IF (.NOT. (lr_triplet)) THEN

                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_rspace_new(1)%pw%cr3d

               END IF
               CALL qs_rho_get(rho1, rho_ao=rho1_ao)
               ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
               CALL cp_dbcsr_set(p_env%kpp1_env%v_ao(ispin)%matrix, 0.0_dp)
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho1_ao(ispin), &
                                       hmat=p_env%kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw_xc)

               ! add hartree only for SINGLETS
               IF (.NOT. lr_triplet) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_hartree_rspace%pw%cr3d

                  CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                          pmat=rho_ao(ispin), &
                                          hmat=p_env%kpp1_env%v_ao(ispin), &
                                          qs_env=qs_env, &
                                          calculate_forces=.FALSE., gapw=gapw)
               END IF
            ELSE
               ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
               CALL cp_dbcsr_set(p_env%kpp1_env%v_ao(ispin)%matrix, 0.0_dp)
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho_ao(ispin), &
                                       hmat=p_env%kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw_xc)

               IF (ispin == 1) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
               END IF
               v_rspace_new(ispin)%pw%cr3d = v_hartree_rspace%pw%cr3d
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho_ao(ispin), &
                                       hmat=p_env%kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw)
            END IF

         ELSE

            v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d* &
                                          v_rspace_new(ispin)%pw%pw_grid%dvol

            IF (nspins == 1) THEN

               IF (.NOT. (lr_triplet)) THEN
                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_rspace_new(1)%pw%cr3d

               ENDIF

               ! add hartree only for SINGLETS
               !IF (res_etype == tddfpt_singlet) THEN
               IF (.NOT. lr_triplet) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
                  v_rspace_new(1)%pw%cr3d = v_rspace_new(1)%pw%cr3d+ &
                                            2.0_dp*v_hartree_rspace%pw%cr3d
               END IF
            ELSE
               IF (ispin == 1) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
               END IF
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             v_hartree_rspace%pw%cr3d
            END IF

            ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
            CALL cp_dbcsr_set(p_env%kpp1_env%v_ao(ispin)%matrix, 0.0_dp)

            CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                    pmat=rho_ao(ispin), &
                                    hmat=p_env%kpp1_env%v_ao(ispin), &
                                    qs_env=qs_env, &
                                    calculate_forces=.FALSE., gapw=gapw)

         END IF

         CALL cp_dbcsr_copy(p_env%kpp1(ispin)%matrix, p_env%kpp1_env%v_ao(ispin)%matrix)
      END DO

      IF (gapw) THEN

         IF (.NOT. ((nspins == 1 .AND. lr_triplet))) THEN
            CALL Vh_1c_gg_integrals(qs_env, energy_hartree_1c, tddft=.TRUE., do_triplet=lr_triplet, &
                                    p_env=p_env)

            CALL integrate_vhg0_rspace(qs_env, v_hartree_rspace, &
                                       .FALSE., tddft=.TRUE., do_triplet=lr_triplet, &
                                       p_env=p_env)
         END IF

!      ***  Add single atom contributions to the KS matrix ***

         ! remap pointer
         ns = SIZE(p_env%kpp1)
         ksmat(1:ns, 1:1) => p_env%kpp1(1:ns)
         ns = SIZE(rho_ao)
         psmat(1:ns, 1:1) => rho_ao(1:ns)

         CALL update_ks_atom(qs_env, ksmat, psmat, .FALSE., .TRUE., p_env)

      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_rspace%pw)
      DO ispin = 1, nspins
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new(ispin)%pw)
      END DO
      DEALLOCATE (v_rspace_new)

      DO ispin = 1, nspins
         CALL cp_fm_get_info(c0(ispin)%matrix, ncol_global=ncol)
         CALL cp_dbcsr_sm_fm_multiply(p_env%kpp1(ispin)%matrix, &
                                      c0(ispin)%matrix, &
                                      Av(ispin)%matrix, &
                                      ncol=ncol, alpha=1.0_dp, beta=1.0_dp)
      ENDDO
      !
      CALL timestop(handle)
      !
   END SUBROUTINE apply_op_2
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param p_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE p_env_check_i_alloc(p_env, qs_env)
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_check_i_alloc', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=25)                                  :: name
      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: gapw_xc
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, matrix_s)

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_linres_methods.F",1209)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_linres_methods.F",1210)
      CALL get_qs_env(qs_env, dft_control=dft_control)
      gapw_xc = dft_control%qs_control%gapw_xc
      IF (.NOT. ASSOCIATED(p_env%kpp1)) THEN
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         nspins = dft_control%nspins

         CALL cp_dbcsr_allocate_matrix_set(p_env%kpp1, nspins)
         name = "p_env"//cp_to_string(p_env%id_nr)//"%kpp1-"
         !CALL compress(name,full=.TRUE.)
         DO ispin = 1, nspins
            ALLOCATE (p_env%kpp1(ispin)%matrix)
            CALL cp_dbcsr_init(p_env%kpp1(ispin)%matrix)
            CALL cp_dbcsr_copy(p_env%kpp1(ispin)%matrix, matrix_s(1)%matrix, &
                               name=TRIM(name)//ADJUSTL(cp_to_string(ispin)))
            CALL cp_dbcsr_set(p_env%kpp1(ispin)%matrix, 0.0_dp)
         END DO

         CALL qs_rho_rebuild(p_env%rho1, qs_env=qs_env)
         IF (gapw_xc) THEN
            CALL qs_rho_rebuild(p_env%rho1_xc, qs_env=qs_env)
         END IF
      END IF

      IF (.NOT. ASSOCIATED(p_env%rho1)) THEN
         CALL qs_rho_rebuild(p_env%rho1, qs_env=qs_env)
         IF (gapw_xc) THEN
            CALL qs_rho_rebuild(p_env%rho1_xc, qs_env=qs_env)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE p_env_check_i_alloc
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param kpp1_env ...
!> \param qs_env ...
!> \param lr_triplet ...
! **************************************************************************************************
   SUBROUTINE kpp1_check_i_alloc(kpp1_env, qs_env, lr_triplet)

      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: lr_triplet

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_check_i_alloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspins
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: my_rho_r, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, xc_section

! ------------------------------------------------------------------

      NULLIFY (pw_env, auxbas_pw_pool, matrix_s, rho, rho_r, input)

      IF(.NOT.(ASSOCIATED(kpp1_env)))CALL cp__a("qs_linres_methods.F",1271)
      IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_linres_methods.F",1272)

      CALL get_qs_env(qs_env, pw_env=pw_env, &
                      matrix_s=matrix_s, input=input, rho=rho)

      CALL qs_rho_get(rho, rho_r=rho_r)
      nspins = SIZE(rho_r)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      IF (.NOT. ASSOCIATED(kpp1_env%v_rspace)) THEN
         ALLOCATE (kpp1_env%v_rspace(nspins))
         DO ispin = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, &
                                   kpp1_env%v_rspace(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
         END DO
      END IF

      IF (.NOT. ASSOCIATED(kpp1_env%v_ao)) THEN
         CALL cp_dbcsr_allocate_matrix_set(kpp1_env%v_ao, nspins)
         DO ispin = 1, nspins
            ALLOCATE (kpp1_env%v_ao(ispin)%matrix)
            CALL cp_dbcsr_init(kpp1_env%v_ao(ispin)%matrix)
            CALL cp_dbcsr_copy(kpp1_env%v_ao(ispin)%matrix, matrix_s(1)%matrix, &
                               name="kpp1%v_ao-"//ADJUSTL(cp_to_string(ispin)))
         END DO
      END IF

      IF (.NOT. ASSOCIATED(kpp1_env%deriv_set)) THEN

         IF (nspins == 1 .AND. lr_triplet) THEN
            ALLOCATE (my_rho_r(2))
            DO ispin = 1, 2
               CALL pw_pool_create_pw(auxbas_pw_pool, my_rho_r(ispin)%pw, &
                                      use_data=rho_r(1)%pw%in_use, in_space=rho_r(1)%pw%in_space)
               my_rho_r(ispin)%pw%cr3d = 0.5_dp*rho_r(1)%pw%cr3d
            END DO
         ELSE
            ALLOCATE (my_rho_r(SIZE(rho_r)))
            DO ispin = 1, SIZE(rho_r)
               my_rho_r(ispin)%pw => rho_r(ispin)%pw
               CALL pw_retain(my_rho_r(ispin)%pw)
            END DO
         END IF

         !ALLOCATE(my_rho_r(SIZE(rho_r)))
         !DO ispin=1,SIZE(rho_r)
         !   my_rho_r(ispin)%pw => rho_r(ispin)%pw
         !   CALL pw_retain(my_rho_r(ispin)%pw)
         !END DO

         xc_section => section_vals_get_subs_vals(input, "DFT%XC")

         CALL xc_prep_2nd_deriv(kpp1_env%deriv_set, kpp1_env%rho_set, &
                                my_rho_r, auxbas_pw_pool, &
                                xc_section=xc_section)

         DO ispin = 1, SIZE(my_rho_r)
            CALL pw_release(my_rho_r(ispin)%pw)
         ENDDO
         DEALLOCATE (my_rho_r)
      ENDIF

   END SUBROUTINE kpp1_check_i_alloc
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param v ...
!> \param psi0 ...
!> \param S_psi0 ...
!> \param buf ...
! **************************************************************************************************
   SUBROUTINE preortho(v, psi0, S_psi0, buf)
      !v = (I-PS)v
      !
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: v, psi0, S_psi0
      TYPE(cp_fm_type), POINTER                          :: buf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'preortho', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, mp, mt, mv, np, nspins, &
                                                            nt, nv

!
!

      CALL timeset(routineN, handle)
      !
      IF(.NOT.(ASSOCIATED(v)))CALL cp__a("qs_linres_methods.F",1362)
      IF(.NOT.(ASSOCIATED(S_psi0)))CALL cp__a("qs_linres_methods.F",1363)
      IF(.NOT.(ASSOCIATED(psi0)))CALL cp__a("qs_linres_methods.F",1364)
      IF(.NOT.(ASSOCIATED(buf)))CALL cp__a("qs_linres_methods.F",1365)
      !
      nspins = SIZE(v, 1)
      !
      DO ispin = 1, nspins
         !
         CALL cp_fm_get_info(v(ispin)%matrix, ncol_global=mv, nrow_global=nv)
         CALL cp_fm_get_info(psi0(ispin)%matrix, ncol_global=mp, nrow_global=np)
         CALL cp_fm_get_info(buf, ncol_global=mt, nrow_global=nt)
         IF(.NOT.(nv == np))CALL cp__a("qs_linres_methods.F",1374)
         IF(.NOT.(mt >= mv))CALL cp__a("qs_linres_methods.F",1375)
         IF(.NOT.(mt >= mp))CALL cp__a("qs_linres_methods.F",1376)
         IF(.NOT.(nt == nv))CALL cp__a("qs_linres_methods.F",1377)
         !
         ! buf = v' * S_psi0
         CALL cp_gemm('T', 'N', mv, mp, nv, 1.0_dp, v(ispin)%matrix, S_psi0(ispin)%matrix, 0.0_dp, buf)
         ! v = v - psi0 * buf'
         CALL cp_gemm('N', 'T', nv, mv, mp, -1.0_dp, psi0(ispin)%matrix, buf, 1.0_dp, v(ispin)%matrix)
         !
      ENDDO
      !
      CALL timestop(handle)
      !
   END SUBROUTINE preortho
   !
   !
! **************************************************************************************************
!> \brief ...
!> \param v ...
!> \param psi0 ...
!> \param S_psi0 ...
!> \param buf ...
! **************************************************************************************************
   SUBROUTINE postortho(v, psi0, S_psi0, buf)
      !v = (I-SP)v
      !
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: v, psi0, S_psi0
      TYPE(cp_fm_type), POINTER                          :: buf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'postortho', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, mp, mt, mv, np, nspins, &
                                                            nt, nv

!

      CALL timeset(routineN, handle)
      !
      IF(.NOT.(ASSOCIATED(v)))CALL cp__a("qs_linres_methods.F",1413)
      IF(.NOT.(ASSOCIATED(S_psi0)))CALL cp__a("qs_linres_methods.F",1414)
      IF(.NOT.(ASSOCIATED(psi0)))CALL cp__a("qs_linres_methods.F",1415)
      IF(.NOT.(ASSOCIATED(buf)))CALL cp__a("qs_linres_methods.F",1416)
      !
      nspins = SIZE(v, 1)
      !
      DO ispin = 1, nspins
         !
         CALL cp_fm_get_info(v(ispin)%matrix, ncol_global=mv, nrow_global=nv)
         CALL cp_fm_get_info(psi0(ispin)%matrix, ncol_global=mp, nrow_global=np)
         CALL cp_fm_get_info(buf, ncol_global=mt, nrow_global=nt)
         IF(.NOT.(nv == np))CALL cp__a("qs_linres_methods.F",1425)
         IF(.NOT.(mt >= mv))CALL cp__a("qs_linres_methods.F",1426)
         IF(.NOT.(mt >= mp))CALL cp__a("qs_linres_methods.F",1427)
         IF(.NOT.(nt == nv))CALL cp__a("qs_linres_methods.F",1428)
         !
         ! buf = v' * psi0
         CALL cp_gemm('T', 'N', mv, mp, nv, 1.0_dp, v(ispin)%matrix, psi0(ispin)%matrix, 0.0_dp, buf)
         ! v = v - S_psi0 * buf'
         CALL cp_gemm('N', 'T', nv, mv, mp, -1.0_dp, S_psi0(ispin)%matrix, buf, 1.0_dp, v(ispin)%matrix)
         !
      ENDDO
      !
      CALL timestop(handle)
      !
   END SUBROUTINE postortho

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param linres_section ...
!> \param vec ...
!> \param ivec ...
!> \param tag ...
!> \param ind ...
! **************************************************************************************************
   SUBROUTINE linres_write_restart(qs_env, linres_section, vec, ivec, tag, ind)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: linres_section
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: vec
      INTEGER, INTENT(IN)                                :: ivec
      CHARACTER(LEN=*)                                   :: tag
      INTEGER, INTENT(IN), OPTIONAL                      :: ind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_write_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_middle, my_pos, my_status
      INTEGER                                            :: handle, i, i_block, ia, ie, ispin, j, &
                                                            max_block, nao, nmo, nspins, &
                                                            output_unit, rst_unit
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger, mo_coeff, mos, para_env, print_key, vecbuffer)

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, linres_section, "PRINT%RESTART", &
                                           used_print_key=print_key), &
                cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, linres_section, &
                                            "PRINT%PROGRAM_RUN_INFO", extension=".Log")

         CALL get_qs_env(qs_env=qs_env, &
                         mos=mos, &
                         para_env=para_env)

         nspins = SIZE(mos)

         my_status = "REPLACE"
         my_pos = "REWIND"
         CALL XSTRING(tag, ia, ie)
         IF (PRESENT(ind)) THEN
            my_middle = "RESTART-"//tag(ia:ie)//TRIM(ADJUSTL(cp_to_string(ivec)))
         ELSE
            my_middle = "RESTART-"//tag(ia:ie)
            IF (ivec > 1) THEN
               my_status = "OLD"
               my_pos = "APPEND"
            END IF
         END IF
         rst_unit = cp_print_key_unit_nr(logger, linres_section, "PRINT%RESTART", &
                                         extension=".lr", middle_name=TRIM(my_middle), file_status=TRIM(my_status), &
                                         file_position=TRIM(my_pos), file_action="WRITE", file_form="UNFORMATTED")

         filename = cp_print_key_generate_filename(logger, print_key, &
                                                   extension=".lr", middle_name=TRIM(my_middle), my_local=.FALSE.)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,A,/)") &
               "Linres response functions are written to restart file", &
               TRIM(filename)
         END IF

         !
         ! write data to file
         ! use the scalapack block size as a default for buffering columns
         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, nrow_global=nao, ncol_block=max_block)
         ALLOCATE (vecbuffer(nao, max_block))

         IF (PRESENT(ind)) THEN
            IF (rst_unit > 0) WRITE (rst_unit) ind, ivec, nspins, nao
         ELSE
            IF (rst_unit > 0) WRITE (rst_unit) ivec, nspins, nao
         END IF

         DO ispin = 1, nspins
            CALL cp_fm_get_info(vec(ispin)%matrix, ncol_global=nmo)

            IF (rst_unit > 0) WRITE (rst_unit) nmo

            DO i = 1, nmo, MAX(max_block, 1)
               i_block = MIN(max_block, nmo-i+1)
               CALL cp_fm_get_submatrix(vec(ispin)%matrix, vecbuffer, 1, i, nao, i_block)
               ! doing this in one write would increase efficiency, but breaks RESTART compatibility.
               ! to old ones, and in cases where max_block is different between runs, as might happen during
               ! restarts with a different number of CPUs
               DO j = 1, i_block
                  IF (rst_unit > 0) WRITE (rst_unit) vecbuffer(1:nao, j)
               ENDDO
            ENDDO
         ENDDO

         DEALLOCATE (vecbuffer)

         CALL cp_print_key_finished_output(rst_unit, logger, linres_section, &
                                           "PRINT%RESTART")
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE linres_write_restart

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param linres_section ...
!> \param vec ...
!> \param ivec ...
!> \param tag ...
!> \param ind ...
! **************************************************************************************************
   SUBROUTINE linres_read_restart(qs_env, linres_section, vec, ivec, tag, ind)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: linres_section
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: vec
      INTEGER, INTENT(IN)                                :: ivec
      CHARACTER(LEN=*)                                   :: tag
      INTEGER, INTENT(INOUT), OPTIONAL                   :: ind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_read_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: my_middle
      INTEGER :: group, handle, i, i_block, ia, ie, iostat, ispin, iv, iv1, ivec_tmp, j, &
         max_block, n_rep_val, nao, nao_tmp, nmo, nmo_tmp, nspins, nspins_tmp, output_unit, &
         rst_unit, source
      LOGICAL                                            :: file_exists
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: print_key

      file_exists = .FALSE.

      CALL timeset(routineN, handle)

      NULLIFY (mos, para_env, logger, print_key, vecbuffer)
      logger => cp_get_default_logger()

      output_unit = cp_print_key_unit_nr(logger, linres_section, &
                                         "PRINT%PROGRAM_RUN_INFO", extension=".Log")

      CALL get_qs_env(qs_env=qs_env, &
                      para_env=para_env, &
                      mos=mos)

      nspins = SIZE(mos)
      group = para_env%group
      source = para_env%source !ionode???

      rst_unit = -1
      IF (para_env%ionode) THEN
         CALL section_vals_val_get(linres_section, "WFN_RESTART_FILE_NAME", &
                                   n_rep_val=n_rep_val)

         CALL XSTRING(tag, ia, ie)
         IF (PRESENT(ind)) THEN
            my_middle = "RESTART-"//tag(ia:ie)//TRIM(ADJUSTL(cp_to_string(ivec)))
         ELSE
            my_middle = "RESTART-"//tag(ia:ie)
         END IF

         IF (n_rep_val > 0) THEN
            CALL section_vals_val_get(linres_section, "WFN_RESTART_FILE_NAME", c_val=filename)
            CALL xstring(filename, ia, ie)
            filename = filename(ia:ie)//TRIM(my_middle)//".lr"
         ELSE
            ! try to read from the filename that is generated automatically from the printkey
            print_key => section_vals_get_subs_vals(linres_section, "PRINT%RESTART")
            filename = cp_print_key_generate_filename(logger, print_key, &
                                                      extension=".lr", middle_name=TRIM(my_middle), my_local=.FALSE.)
         ENDIF
         INQUIRE (FILE=filename, exist=file_exists)
         !
         ! open file
         IF (file_exists) THEN
            CALL open_file(file_name=TRIM(filename), &
                           file_action="READ", &
                           file_form="UNFORMATTED", &
                           file_position="REWIND", &
                           file_status="OLD", &
                           unit_number=rst_unit)

            IF (output_unit > 0) WRITE (output_unit, "(/,T20,A,A)") &
               "Read response wavefunction from restart file ", TRIM(filename)
         ELSE
            IF (output_unit > 0) WRITE (output_unit, "(/,T10,A)") &
               "Restart file not available filename=<"//TRIM(filename)//'>'
         ENDIF
      ENDIF

      CALL mp_bcast(file_exists, source, group)

      IF (file_exists) THEN

         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, nrow_global=nao, ncol_block=max_block)

         ALLOCATE (vecbuffer(nao, max_block))
         !
         ! read headers
         IF (PRESENT(ind)) THEN
            iv1 = ivec
         ELSE
            iv1 = 1
         END IF
         DO iv = iv1, ivec

            IF (PRESENT(ind)) THEN
               IF (rst_unit > 0) READ (rst_unit, IOSTAT=iostat) ind, ivec_tmp, nspins_tmp, nao_tmp
               CALL mp_bcast(iostat, source, group)
               CALL mp_bcast(ind, source, group)
            ELSE
               IF (rst_unit > 0) READ (rst_unit, IOSTAT=iostat) ivec_tmp, nspins_tmp, nao_tmp
               CALL mp_bcast(iostat, source, group)
            END IF

            IF (iostat .NE. 0) EXIT
            CALL mp_bcast(ivec_tmp, source, group)
            CALL mp_bcast(nspins_tmp, source, group)
            CALL mp_bcast(nao_tmp, source, group)

            ! check that the number nao, nmo and nspins are
            ! the same as in the current mos
            IF (nspins_tmp .NE. nspins) CALL cp__b("qs_linres_methods.F",1682,"nspins not consistent")
            IF (nao_tmp .NE. nao) CALL cp__b("qs_linres_methods.F",1683,"nao not consistent")
            !
            DO ispin = 1, nspins
               CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff)
               CALL cp_fm_get_info(mo_coeff, ncol_global=nmo)
               !
               IF (rst_unit > 0) READ (rst_unit) nmo_tmp
               CALL mp_bcast(nmo_tmp, source, group)
               IF (nmo_tmp .NE. nmo) CALL cp__b("qs_linres_methods.F",1691,"nmo not consistent")
               !
               ! read the response
               DO i = 1, nmo, MAX(max_block, 1)
                  i_block = MIN(max_block, nmo-i+1)
                  DO j = 1, i_block
                     IF (rst_unit > 0) READ (rst_unit) vecbuffer(1:nao, j)
                  ENDDO
                  IF (iv .EQ. ivec_tmp) THEN
                     CALL mp_bcast(vecbuffer, source, group)
                     CALL cp_fm_set_submatrix(vec(ispin)%matrix, vecbuffer, 1, i, nao, i_block)
                  ENDIF
               ENDDO
            ENDDO
            IF (ivec .EQ. ivec_tmp) EXIT
         ENDDO

         IF (iostat .NE. 0) THEN
            IF (output_unit > 0) WRITE (output_unit, "(/,T10,A,A,/)") &
               "Restart file: didnt find ", TRIM(filename)
         ENDIF

         DEALLOCATE (vecbuffer)

      ENDIF

      IF (para_env%ionode) THEN
         IF (file_exists) CALL close_file(unit_number=rst_unit)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE linres_read_restart

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param p_env ...
!> \param linres_control ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE check_p_env_init(p_env, linres_control, nspins)
      !
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(linres_control_type), POINTER                 :: linres_control
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_p_env_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, ncol, nrow

      p_env%iter = 0
      p_env%only_energy = .FALSE.
      p_env%ls_count = 0

      p_env%ls_pos = 0.0_dp
      p_env%ls_energy = 0.0_dp
      p_env%ls_grad = 0.0_dp
      p_env%gnorm_old = 1.0_dp

      IF (linres_control%preconditioner_type /= ot_precond_none) THEN
         IF(.NOT.(ASSOCIATED(p_env%preconditioner)))CALL cp__a("qs_linres_methods.F",1754)
         DO ispin = 1, nspins
            CALL cp_fm_get_info(p_env%PS_psi0(ispin)%matrix, nrow_global=nrow, ncol_global=ncol)
            IF(.NOT.(nrow == p_env%n_ao(ispin)))CALL cp__a("qs_linres_methods.F",1757)
            IF(.NOT.(ncol == p_env%n_mo(ispin)))CALL cp__a("qs_linres_methods.F",1758)
         ENDDO
      ENDIF

   END SUBROUTINE check_p_env_init

END MODULE qs_linres_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates Nabla V_KS (local part if PSP) on the different grids
!> \par History
!>       created 06-2007 [RD]
!> \author RD
! **************************************************************************************************
MODULE qs_linres_epr_nablavks
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE erf_fn,                          ONLY: erf,&
                                              erfc
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential,&
                                              gth_potential_type
   USE hartree_local_methods,           ONLY: calculate_Vh_1center
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: rootpi,&
                                              twopi
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_methods,                   ONLY: calc_rho_tot_gspace
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_linres_types,                 ONLY: epr_env_type,&
                                              get_epr_env,&
                                              nablavks_atom_type
! R0
   USE qs_local_rho_types,              ONLY: rhoz_type
   USE qs_rho0_types,                   ONLY: rho0_atom_type
   USE qs_rho_atom_methods,             ONLY: calculate_rho_atom_coeff
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
! R0
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_p_type,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE qs_vxc,                          ONLY: qs_vxc_create
   USE qs_vxc_atom,                     ONLY: calculate_vxc_atom
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: epr_nablavks

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_epr_nablavks'

CONTAINS

! **************************************************************************************************
!> \brief Evaluates Nabla V_KS on the grids
!> \param epr_env ...
!> \param qs_env ...
!> \par History
!>      06.2006 created [RD]
!> \author RD
! **************************************************************************************************
   SUBROUTINE epr_nablavks(epr_env, qs_env)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_nablavks', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: ext, filename
      COMPLEX(KIND=dp)                                   :: gtemp
      INTEGER                                            :: bo_atom(2), ia, iat, iatom, idir, iexp, &
                                                            ig, ikind, ir, iso, ispin, ix, iy, iz, &
                                                            natom, nexp_ppl, nkind, nspins, &
                                                            output_unit, unit_nr
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: gapw, gapw_xc, gth_gspace, ionode, &
                                                            make_soft, paw_atom
      REAL(KIND=dp) :: alpha, alpha_core, arg, charge, ehartree, exc, exp_rap, gapw_max_alpha, &
         hard_radius, hard_value, soft_value, sqrt_alpha, sqrt_rap
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vh1_rad_h, vh1_rad_s
      REAL(KIND=dp), DIMENSION(3)                        :: rap, ratom, roffset, rpoint
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cexp_ppl, rho_rad_z
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho_rad_0
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho1_r, rho2_r, rho_r, v_rspace_new, &
                                                            v_tau_rspace
      TYPE(pw_p_type), POINTER :: rho_tot_gspace, v_coulomb_gspace, v_coulomb_gtemp, &
         v_coulomb_rtemp, v_hartree_gspace, v_hartree_gtemp, v_hartree_rtemp, v_xc_gtemp, &
         v_xc_rtemp, wf_r
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_type), POINTER                             :: pwx, pwy, pwz
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_p_type), DIMENSION(:, :), POINTER      :: nablavks_set
      TYPE(qs_rho_type), POINTER                         :: rho, rho_xc
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: rho_rad_h, rho_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: nablavks_vec_rad_h, nablavks_vec_rad_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(rhoz_type), DIMENSION(:), POINTER             :: rhoz_set
      TYPE(section_vals_type), POINTER                   :: g_section, input, lr_section, xc_section

! R0
! R0
! R0
! R0
! R0
! R0

      NULLIFY (auxbas_pw_pool)
      NULLIFY (cell)
      NULLIFY (dft_control)
      NULLIFY (g_section)
      NULLIFY (logger)
      NULLIFY (lr_section)
      NULLIFY (nablavks_set)
      NULLIFY (nablavks_atom_set) ! R0
      NULLIFY (nablavks_vec_rad_h) ! R0
      NULLIFY (nablavks_vec_rad_s) ! R0
      NULLIFY (para_env)
      NULLIFY (particle_set)
      NULLIFY (particles)
      NULLIFY (poisson_env)
      NULLIFY (pw_env)
      NULLIFY (pwx) ! R0
      NULLIFY (pwy) ! R0
      NULLIFY (pwz) ! R0
      NULLIFY (rho)
      NULLIFY (rho_xc)
      NULLIFY (rho0_atom_set)
      NULLIFY (rho_atom_set)
      NULLIFY (rhoz_set)
      NULLIFY (subsys)
      NULLIFY (v_rspace_new)
      NULLIFY (v_tau_rspace)
      NULLIFY (xc_section)
      NULLIFY (input)
      NULLIFY (ks_env)
      NULLIFY (rho_r, rho_ao, rho1_r, rho2_r)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      ionode = logger%para_env%mepos == logger%para_env%source

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

!   -------------------------------------
!   Read settings
!   -------------------------------------

      g_section => section_vals_get_subs_vals(lr_section, &
                                              "EPR%PRINT%G_TENSOR")

      CALL section_vals_val_get(g_section, "gapw_max_alpha", r_val=gapw_max_alpha)

      gth_gspace = .TRUE.

!   -------------------------------------
!   Get nablavks arrays
!   -------------------------------------

      CALL get_epr_env(epr_env, nablavks_set=nablavks_set, & ! R0
                       nablavks_atom_set=nablavks_atom_set) ! R0
      ! R0

      DO ispin = 1, SIZE(nablavks_set, 2)
         DO idir = 1, SIZE(nablavks_set, 1)
            CALL qs_rho_get(nablavks_set(idir, ispin)%rho, rho_r=rho_r)
            CALL pw_zero(rho_r(1)%pw)
         END DO
      END DO

      CALL qs_rho_get(nablavks_set(1, 1)%rho, rho_r=rho_r)
      pwx => rho_r(1)%pw
      CALL qs_rho_get(nablavks_set(2, 1)%rho, rho_r=rho_r)
      pwy => rho_r(1)%pw
      CALL qs_rho_get(nablavks_set(3, 1)%rho, rho_r=rho_r)
      pwz => rho_r(1)%pw

      roffset = -REAL(MODULO(pwx%pw_grid%npts, 2), dp)*pwx%pw_grid%dr/2.0_dp

!   -------------------------------------
!   Get grids / atom info
!   -------------------------------------

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      input=input, &
                      cell=cell, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      rho=rho, &
                      rho_xc=rho_xc, &
                      rho_atom_set=rho_atom_set, &
                      rho0_atom_set=rho0_atom_set, &
                      rhoz_set=rhoz_set, &
                      subsys=subsys, &
                      ks_env=ks_env)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)

      CALL qs_subsys_get(subsys, particles=particles)

      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      nkind = SIZE(atomic_kind_set)
      nspins = dft_control%nspins

!   -------------------------------------
!   Add Hartree potential
!   -------------------------------------

      ALLOCATE (v_hartree_gspace)
      ALLOCATE (v_hartree_gtemp)
      ALLOCATE (v_hartree_rtemp)
      ALLOCATE (rho_tot_gspace)

      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_gtemp%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_rtemp%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      IF (gapw) THEN
         ! need to rebuild the coeff !
         CALL qs_rho_get(rho, rho_ao_kp=rho_ao)
         CALL calculate_rho_atom_coeff(qs_env, rho_ao)
         CALL prepare_gapw_den(qs_env, do_rho0=.TRUE.)
      END IF

      CALL pw_zero(rho_tot_gspace%pw)

      CALL calc_rho_tot_gspace(rho_tot_gspace, qs_env, rho, &
                               skip_nuclear_density=.NOT. gapw)

      CALL pw_poisson_solve(poisson_env, rho_tot_gspace%pw, ehartree, &
                            v_hartree_gspace%pw)

      !   -------------------------------------
      !   Atomic grids part
      !   -------------------------------------

      IF (gapw) THEN

         DO ikind = 1, nkind ! loop over atom types

            NULLIFY (atom_list)
            NULLIFY (grid_atom)
            NULLIFY (harmonics)
            NULLIFY (rho_rad_z)

            rho_rad_z => rhoz_set(ikind)%r_coef

            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             grid_atom=grid_atom, &
                             harmonics=harmonics, &
                             hard_radius=hard_radius, &
                             paw_atom=paw_atom, &
                             zeff=charge, &
                             alpha_core_charge=alpha_core)

            IF (paw_atom) THEN

               ALLOCATE (vh1_rad_h(grid_atom%nr, harmonics%max_iso_not0))
               ALLOCATE (vh1_rad_s(grid_atom%nr, harmonics%max_iso_not0))

               ! DO iat = 1, natom ! natom = # atoms for ikind
               !
               !    iatom = atom_list(iat)
               !    ratom = particle_set(iatom)%r
               !
               !    DO ig = v_hartree_gspace%pw%pw_grid%first_gne0,v_hartree_gspace%pw%pw_grid%ngpts_cut_local
               !
               !       gtemp = fourpi * charge / cell%deth * &
               !               EXP ( - v_hartree_gspace%pw%pw_grid%gsq(ig) / (4.0_dp * alpha_core) ) &
               !               / v_hartree_gspace%pw%pw_grid%gsq(ig)
               !
               !       arg = DOT_PRODUCT(v_hartree_gspace%pw%pw_grid%g(:,ig),ratom)
               !
               !       gtemp = gtemp * CMPLX(COS(arg),-SIN(arg),KIND=dp)
               !
               !       v_hartree_gspace%pw%cc(ig) = v_hartree_gspace%pw%cc(ig) + gtemp
               !    END DO
               !    IF ( v_hartree_gspace%pw%pw_grid%have_g0 ) v_hartree_gspace%pw%cc(1) = 0.0_dp
               !
               ! END DO

               bo_atom = get_limit(natom, para_env%num_pe, para_env%mepos)

               DO iat = bo_atom(1), bo_atom(2) ! natomkind = # atoms for ikind

                  iatom = atom_list(iat)
                  ratom = particle_set(iatom)%r

                  nablavks_vec_rad_h => nablavks_atom_set(iatom)%nablavks_vec_rad_h
                  nablavks_vec_rad_s => nablavks_atom_set(iatom)%nablavks_vec_rad_s

                  DO ispin = 1, nspins
                     DO idir = 1, 3
                        nablavks_vec_rad_h(idir, ispin)%r_coef(:, :) = 0.0_dp
                        nablavks_vec_rad_s(idir, ispin)%r_coef(:, :) = 0.0_dp
                     END DO ! idir
                  END DO ! ispin

                  rho_atom => rho_atom_set(iatom)
                  NULLIFY (rho_rad_h, rho_rad_s, rho_rad_0)
                  CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=rho_rad_h, &
                                    rho_rad_s=rho_rad_s)
                  rho_rad_0 => rho0_atom_set(iatom)%rho0_rad_h%r_coef
                  vh1_rad_h = 0.0_dp
                  vh1_rad_s = 0.0_dp

                  CALL calculate_Vh_1center(vh1_rad_h, vh1_rad_s, rho_rad_h, rho_rad_s, rho_rad_0, rho_rad_z, grid_atom)

                  DO ir = 2, grid_atom%nr

                     IF (grid_atom%rad(ir) >= hard_radius) CYCLE

                     DO ia = 1, grid_atom%ng_sphere

                        ! hard part

                        DO idir = 1, 3
                           hard_value = 0.0_dp
                           DO iso = 1, harmonics%max_iso_not0
                              hard_value = hard_value+ &
                                           vh1_rad_h(ir, iso)*harmonics%dslm_dxyz(idir, ia, iso)+ &
                                           harmonics%slm(ia, iso)* &
                                           (vh1_rad_h(ir-1, iso)-vh1_rad_h(ir, iso))/ &
                                           (grid_atom%rad(ir-1)-grid_atom%rad(ir))* &
                                           (harmonics%a(idir, ia))
                           END DO
                           nablavks_vec_rad_h(idir, 1)%r_coef(ir, ia) = hard_value
                        END DO

                        ! soft part

                        DO idir = 1, 3
                           soft_value = 0.0_dp
                           DO iso = 1, harmonics%max_iso_not0
                              soft_value = soft_value+ &
                                           vh1_rad_s(ir, iso)*harmonics%dslm_dxyz(idir, ia, iso)+ &
                                           harmonics%slm(ia, iso)* &
                                           (vh1_rad_s(ir-1, iso)-vh1_rad_s(ir, iso))/ &
                                           (grid_atom%rad(ir-1)-grid_atom%rad(ir))* &
                                           (harmonics%a(idir, ia))
                           END DO
                           nablavks_vec_rad_s(idir, 1)%r_coef(ir, ia) = soft_value
                        END DO

                     END DO ! ia

                  END DO ! ir

                  DO idir = 1, 3
                     nablavks_vec_rad_h(idir, 2)%r_coef(:, :) = nablavks_vec_rad_h(idir, 1)%r_coef(:, :)
                     nablavks_vec_rad_s(idir, 2)%r_coef(:, :) = nablavks_vec_rad_s(idir, 1)%r_coef(:, :)
                  END DO

               END DO ! iat

               DEALLOCATE (vh1_rad_h)
               DEALLOCATE (vh1_rad_s)

            END IF ! paw_atom

         END DO ! ikind

      END IF ! gapw

      CALL pw_copy(v_hartree_gspace%pw, v_hartree_gtemp%pw)
      CALL pw_derive(v_hartree_gtemp%pw, (/1, 0, 0/))
      CALL pw_transfer(v_hartree_gtemp%pw, v_hartree_rtemp%pw)
      CALL pw_copy(v_hartree_rtemp%pw, pwx)

      CALL pw_copy(v_hartree_gspace%pw, v_hartree_gtemp%pw)
      CALL pw_derive(v_hartree_gtemp%pw, (/0, 1, 0/))
      CALL pw_transfer(v_hartree_gtemp%pw, v_hartree_rtemp%pw)
      CALL pw_copy(v_hartree_rtemp%pw, pwy)

      CALL pw_copy(v_hartree_gspace%pw, v_hartree_gtemp%pw)
      CALL pw_derive(v_hartree_gtemp%pw, (/0, 0, 1/))
      CALL pw_transfer(v_hartree_gtemp%pw, v_hartree_rtemp%pw)
      CALL pw_copy(v_hartree_rtemp%pw, pwz)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gtemp%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_rtemp%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_tot_gspace%pw)

      DEALLOCATE (v_hartree_gspace)
      DEALLOCATE (v_hartree_gtemp)
      DEALLOCATE (v_hartree_rtemp)
      DEALLOCATE (rho_tot_gspace)

!   -------------------------------------
!   Add Coulomb potential
!   -------------------------------------

      DO ikind = 1, nkind ! loop over atom types

         NULLIFY (atom_list)
         NULLIFY (grid_atom)
         NULLIFY (harmonics)

         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          grid_atom=grid_atom, &
                          harmonics=harmonics, &
                          hard_radius=hard_radius, &
                          gth_potential=gth_potential, &
                          all_potential=all_potential, &
                          paw_atom=paw_atom)

         IF (ASSOCIATED(gth_potential)) THEN

            NULLIFY (cexp_ppl)

            CALL get_potential(potential=gth_potential, &
                               zeff=charge, &
                               alpha_ppl=alpha, &
                               nexp_ppl=nexp_ppl, &
                               cexp_ppl=cexp_ppl)

            sqrt_alpha = SQRT(alpha)

            IF (gapw .AND. paw_atom .AND. alpha > gapw_max_alpha) THEN
               make_soft = .TRUE.
            ELSE
               make_soft = .FALSE.
            END IF

            !   -------------------------------------
            !   PW grid part
            !   -------------------------------------

            IF (gth_gspace) THEN

               ALLOCATE (v_coulomb_gspace)
               ALLOCATE (v_coulomb_gtemp)
               ALLOCATE (v_coulomb_rtemp)

               CALL pw_pool_create_pw(auxbas_pw_pool, v_coulomb_gspace%pw, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               CALL pw_pool_create_pw(auxbas_pw_pool, v_coulomb_gtemp%pw, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               CALL pw_pool_create_pw(auxbas_pw_pool, v_coulomb_rtemp%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)

               CALL pw_zero(v_coulomb_gspace%pw)

               DO iat = 1, natom ! natom = # atoms for ikind

                  iatom = atom_list(iat)
                  ratom = particle_set(iatom)%r

                  DO ig = v_coulomb_gspace%pw%pw_grid%first_gne0, v_coulomb_gspace%pw%pw_grid%ngpts_cut_local
                     gtemp = 0.0_dp
                     ! gtemp = - fourpi * charge / cell%deth * &
                     !         EXP ( - v_coulomb_gspace%pw%pw_grid%gsq(ig) / (4.0_dp * alpha) ) &
                     !         / v_coulomb_gspace%pw%pw_grid%gsq(ig)

                     IF (.NOT. make_soft) THEN

                        SELECT CASE (nexp_ppl)
                        CASE (1)
                           gtemp = gtemp+ &
                                   (twopi)**(1.5_dp)/(cell%deth*(2.0_dp*alpha)**(1.5_dp))* &
                                   EXP(-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(4.0_dp*alpha))*( &
                                   ! C1
                                   +cexp_ppl(1) &
                                   )
                        CASE (2)
                           gtemp = gtemp+ &
                                   (twopi)**(1.5_dp)/(cell%deth*(2.0_dp*alpha)**(1.5_dp))* &
                                   EXP(-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(4.0_dp*alpha))*( &
                                   ! C1
                                   +cexp_ppl(1) &
                                   ! C2
                                   +cexp_ppl(2)/(2.0_dp*alpha)* &
                                   (3.0_dp-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha)) &
                                   )
                        CASE (3)
                           gtemp = gtemp+ &
                                   (twopi)**(1.5_dp)/(cell%deth*(2.0_dp*alpha)**(1.5_dp))* &
                                   EXP(-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(4.0_dp*alpha))*( &
                                   ! C1
                                   +cexp_ppl(1) &
                                   ! C2
                                   +cexp_ppl(2)/(2.0_dp*alpha)* &
                                   (3.0_dp-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha)) &
                                   ! C3
                                   +cexp_ppl(3)/(2.0_dp*alpha)**2* &
                                   (15.0_dp-10.0_dp*v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha) &
                                    +(v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha))**2) &
                                   )
                        CASE (4)
                           gtemp = gtemp+ &
                                   (twopi)**(1.5_dp)/(cell%deth*(2.0_dp*alpha)**(1.5_dp))* &
                                   EXP(-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(4.0_dp*alpha))*( &
                                   ! C1
                                   +cexp_ppl(1) &
                                   ! C2
                                   +cexp_ppl(2)/(2.0_dp*alpha)* &
                                   (3.0_dp-v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha)) &
                                   ! C3
                                   +cexp_ppl(3)/(2.0_dp*alpha)**2* &
                                   (15.0_dp-10.0_dp*v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha) &
                                    +(v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha))**2) &
                                   ! C4
                                   +cexp_ppl(4)/(2.0_dp*alpha)**3* &
                                   (105.0_dp-105.0_dp*v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha) &
                                    +21.0_dp*(v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha))**2 &
                                    -(v_coulomb_gspace%pw%pw_grid%gsq(ig)/(2.0_dp*alpha))**3) &
                                   )
                        END SELECT

                     END IF

                     arg = DOT_PRODUCT(v_coulomb_gspace%pw%pw_grid%g(:, ig), ratom)

                     gtemp = gtemp*CMPLX(COS(arg), -SIN(arg), KIND=dp)
                     v_coulomb_gspace%pw%cc(ig) = v_coulomb_gspace%pw%cc(ig)+gtemp
                  END DO
                  IF (v_coulomb_gspace%pw%pw_grid%have_g0) v_coulomb_gspace%pw%cc(1) = 0.0_dp

               END DO

               CALL pw_copy(v_coulomb_gspace%pw, v_coulomb_gtemp%pw)
               CALL pw_derive(v_coulomb_gtemp%pw, (/1, 0, 0/))
               CALL pw_transfer(v_coulomb_gtemp%pw, v_coulomb_rtemp%pw)
               CALL pw_axpy(v_coulomb_rtemp%pw, pwx)

               CALL pw_copy(v_coulomb_gspace%pw, v_coulomb_gtemp%pw)
               CALL pw_derive(v_coulomb_gtemp%pw, (/0, 1, 0/))
               CALL pw_transfer(v_coulomb_gtemp%pw, v_coulomb_rtemp%pw)
               CALL pw_axpy(v_coulomb_rtemp%pw, pwy)

               CALL pw_copy(v_coulomb_gspace%pw, v_coulomb_gtemp%pw)
               CALL pw_derive(v_coulomb_gtemp%pw, (/0, 0, 1/))
               CALL pw_transfer(v_coulomb_gtemp%pw, v_coulomb_rtemp%pw)
               CALL pw_axpy(v_coulomb_rtemp%pw, pwz)

               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_coulomb_gspace%pw)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_coulomb_gtemp%pw)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, v_coulomb_rtemp%pw)

               DEALLOCATE (v_coulomb_gspace)
               DEALLOCATE (v_coulomb_gtemp)
               DEALLOCATE (v_coulomb_rtemp)
            ELSE

               ! Attic of the atomic parallellisation
               !
               ! bo(2)
               ! bo = get_limit(natom, para_env%num_pe, para_env%mepos)
               ! DO iat =  bo(1),bo(2) ! natom = # atoms for ikind
               ! DO ix = lbound(pwx%cr3d,1), ubound(pwx%cr3d,1)
               ! DO iy = lbound(pwx%cr3d,2), ubound(pwx%cr3d,2)
               ! DO iz = lbound(pwx%cr3d,3), ubound(pwx%cr3d,3)

               bo = pwx%pw_grid%bounds_local

               DO iat = 1, natom ! natom = # atoms for ikind

                  iatom = atom_list(iat)
                  ratom = particle_set(iatom)%r

                  DO ix = bo(1, 1), bo(2, 1)
                     DO iy = bo(1, 2), bo(2, 2)
                        DO iz = bo(1, 3), bo(2, 3)
                           rpoint = (/REAL(ix, dp)*pwx%pw_grid%dr(1), &
                                      REAL(iy, dp)*pwx%pw_grid%dr(2), &
                                      REAL(iz, dp)*pwx%pw_grid%dr(3)/)
                           rpoint = rpoint+roffset
                           rap = rpoint-ratom
                           rap(1) = MODULO(rap(1), cell%hmat(1, 1))-cell%hmat(1, 1)/2._dp
                           rap(2) = MODULO(rap(2), cell%hmat(2, 2))-cell%hmat(2, 2)/2._dp
                           rap(3) = MODULO(rap(3), cell%hmat(3, 3))-cell%hmat(3, 3)/2._dp
                           sqrt_rap = SQRT(DOT_PRODUCT(rap, rap))
                           exp_rap = EXP(-alpha*sqrt_rap**2)
                           IF (sqrt_rap < 1.e-10_dp) sqrt_rap = 1.e-10_dp
                           ! d_x

                           pwx%cr3d(ix, iy, iz) = pwx%cr3d(ix, iy, iz)+charge*( &
                                                  -2.0_dp*sqrt_alpha*EXP(-sqrt_rap**2*sqrt_alpha**2)*rap(1) &
                                                  /(rootpi*sqrt_rap**2) &
                                                  +erf(sqrt_rap*sqrt_alpha)*rap(1) &
                                                  /sqrt_rap**3)

                           ! d_y

                           pwy%cr3d(ix, iy, iz) = pwy%cr3d(ix, iy, iz)+charge*( &
                                                  -2.0_dp*sqrt_alpha*EXP(-sqrt_rap**2*sqrt_alpha**2)*rap(2) &
                                                  /(rootpi*sqrt_rap**2) &
                                                  +erf(sqrt_rap*sqrt_alpha)*rap(2) &
                                                  /sqrt_rap**3)

                           ! d_z

                           pwz%cr3d(ix, iy, iz) = pwz%cr3d(ix, iy, iz)+charge*( &
                                                  -2.0_dp*sqrt_alpha*EXP(-sqrt_rap**2*sqrt_alpha**2)*rap(3) &
                                                  /(rootpi*sqrt_rap**2) &
                                                  +erf(sqrt_rap*sqrt_alpha)*rap(3) &
                                                  /sqrt_rap**3)

                           IF (make_soft) CYCLE

                           ! d_x

                           DO iexp = 1, nexp_ppl
                              pwx%cr3d(ix, iy, iz) = pwx%cr3d(ix, iy, iz)+( &
                                                     -2.0_dp*alpha*rap(1)*exp_rap* &
                                                     cexp_ppl(iexp)*(sqrt_rap**2)**(iexp-1))
                              IF (iexp > 1) THEN
                                 pwx%cr3d(ix, iy, iz) = pwx%cr3d(ix, iy, iz)+( &
                                                        2.0_dp*exp_rap*cexp_ppl(iexp)* &
                                                        (sqrt_rap**2)**(iexp-2)*REAL(iexp-1, dp)*rap(1))
                              END IF
                           END DO

                           ! d_y

                           DO iexp = 1, nexp_ppl
                              pwy%cr3d(ix, iy, iz) = pwy%cr3d(ix, iy, iz)+( &
                                                     -2.0_dp*alpha*rap(2)*exp_rap* &
                                                     cexp_ppl(iexp)*(sqrt_rap**2)**(iexp-1))
                              IF (iexp > 1) THEN
                                 pwy%cr3d(ix, iy, iz) = pwy%cr3d(ix, iy, iz)+( &
                                                        2.0_dp*exp_rap*cexp_ppl(iexp)* &
                                                        (sqrt_rap**2)**(iexp-2)*REAL(iexp-1, dp)*rap(2))
                              END IF
                           END DO

                           ! d_z

                           DO iexp = 1, nexp_ppl
                              pwz%cr3d(ix, iy, iz) = pwz%cr3d(ix, iy, iz)+( &
                                                     -2.0_dp*alpha*rap(3)*exp_rap* &
                                                     cexp_ppl(iexp)*(sqrt_rap**2)**(iexp-1))
                              IF (iexp > 1) THEN
                                 pwz%cr3d(ix, iy, iz) = pwz%cr3d(ix, iy, iz)+( &
                                                        2.0_dp*exp_rap*cexp_ppl(iexp)* &
                                                        (sqrt_rap**2)**(iexp-2)*REAL(iexp-1, dp)*rap(3))
                              END IF
                           END DO

                        END DO ! iz
                     END DO ! iy
                  END DO ! ix
               END DO ! iat
            END IF ! gth_gspace

            !   -------------------------------------
            !   Atomic grids part
            !   -------------------------------------

            IF (gapw .AND. paw_atom) THEN

               bo_atom = get_limit(natom, para_env%num_pe, para_env%mepos)

               DO iat = bo_atom(1), bo_atom(2) ! natom = # atoms for ikind

                  iatom = atom_list(iat)

                  nablavks_vec_rad_h => nablavks_atom_set(iatom)%nablavks_vec_rad_h
                  nablavks_vec_rad_s => nablavks_atom_set(iatom)%nablavks_vec_rad_s

                  DO ir = 1, grid_atom%nr

                     IF (grid_atom%rad(ir) >= hard_radius) CYCLE

                     exp_rap = EXP(-alpha*grid_atom%rad(ir)**2)

                     DO ia = 1, grid_atom%ng_sphere

                        DO idir = 1, 3
                           hard_value = 0.0_dp
                           hard_value = charge*( &
                                        -2.0_dp*sqrt_alpha*EXP(-grid_atom%rad(ir)**2*sqrt_alpha**2) &
                                        *grid_atom%rad(ir)*harmonics%a(idir, ia) &
                                        /(rootpi*grid_atom%rad(ir)**2) &
                                        +erf(grid_atom%rad(ir)*sqrt_alpha) &
                                        *grid_atom%rad(ir)*harmonics%a(idir, ia) &
                                        /grid_atom%rad(ir)**3)
                           soft_value = hard_value
                           DO iexp = 1, nexp_ppl
                              hard_value = hard_value+( &
                                           -2.0_dp*alpha*grid_atom%rad(ir)*harmonics%a(idir, ia) &
                                           *exp_rap*cexp_ppl(iexp)*(grid_atom%rad(ir)**2)**(iexp-1))
                              IF (iexp > 1) THEN
                                 hard_value = hard_value+( &
                                              2.0_dp*exp_rap*cexp_ppl(iexp) &
                                              *(grid_atom%rad(ir)**2)**(iexp-2)*REAL(iexp-1, dp) &
                                              *grid_atom%rad(ir)*harmonics%a(idir, ia))
                              END IF
                           END DO
                           nablavks_vec_rad_h(idir, 1)%r_coef(ir, ia) = &
                              nablavks_vec_rad_h(idir, 1)%r_coef(ir, ia)+hard_value
                           IF (make_soft) THEN
                              nablavks_vec_rad_s(idir, 1)%r_coef(ir, ia) = &
                                 nablavks_vec_rad_s(idir, 1)%r_coef(ir, ia)+soft_value
                           ELSE
                              nablavks_vec_rad_s(idir, 1)%r_coef(ir, ia) = &
                                 nablavks_vec_rad_s(idir, 1)%r_coef(ir, ia)+hard_value
                           END IF
                        END DO

                     END DO ! ia
                  END DO ! ir

                  DO ispin = 2, nspins
                     DO idir = 1, 3
                        nablavks_vec_rad_h(idir, ispin)%r_coef(:, :) = nablavks_vec_rad_h(idir, 1)%r_coef(:, :)
                        nablavks_vec_rad_s(idir, ispin)%r_coef(:, :) = nablavks_vec_rad_s(idir, 1)%r_coef(:, :)
                     END DO
                  END DO

               END DO

            END IF

         ELSE IF (ASSOCIATED(all_potential)) THEN

            CALL get_potential(potential=all_potential, &
                               alpha_core_charge=alpha, &
                               zeff=charge)

            sqrt_alpha = SQRT(alpha)

            !   -------------------------------------
            !   Atomic grids part
            !   -------------------------------------

            bo_atom = get_limit(natom, para_env%num_pe, para_env%mepos)

            DO iat = bo_atom(1), bo_atom(2) ! natom = # atoms for ikind

               iatom = atom_list(iat)

               nablavks_vec_rad_h => nablavks_atom_set(iatom)%nablavks_vec_rad_h

               DO ir = 1, grid_atom%nr

                  IF (grid_atom%rad(ir) >= hard_radius) CYCLE

                  DO ia = 1, grid_atom%ng_sphere

                     DO idir = 1, 3
                        hard_value = 0.0_dp
                        hard_value = charge*( &
                                     2.0_dp*sqrt_alpha*EXP(-grid_atom%rad(ir)**2*sqrt_alpha**2) &
                                     *grid_atom%rad(ir)*harmonics%a(idir, ia) &
                                     /(rootpi*grid_atom%rad(ir)**2) &
                                     +erfc(grid_atom%rad(ir)*sqrt_alpha) &
                                     *grid_atom%rad(ir)*harmonics%a(idir, ia) &
                                     /grid_atom%rad(ir)**3)
                        nablavks_vec_rad_h(idir, 1)%r_coef(ir, ia) = &
                           nablavks_vec_rad_h(idir, 1)%r_coef(ir, ia)+hard_value
                     END DO

                  END DO ! ia
               END DO ! ir

               DO ispin = 2, nspins
                  DO idir = 1, 3
                     nablavks_vec_rad_h(idir, ispin)%r_coef(:, :) = nablavks_vec_rad_h(idir, 1)%r_coef(:, :)
                  END DO
               END DO

            END DO

         ELSE
            CYCLE
         END IF

      END DO

      DO idir = 1, 3
         CALL qs_rho_get(nablavks_set(idir, 1)%rho, rho_r=rho1_r)
         CALL qs_rho_get(nablavks_set(idir, 2)%rho, rho_r=rho2_r)
         CALL pw_copy(rho1_r(1)%pw, rho2_r(1)%pw)
      END DO

!   -------------------------------------
!   Add V_xc potential
!   -------------------------------------

      ALLOCATE (v_xc_gtemp)
      ALLOCATE (v_xc_rtemp)

      CALL pw_pool_create_pw(auxbas_pw_pool, v_xc_gtemp%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, v_xc_rtemp%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      xc_section => section_vals_get_subs_vals(input, "PROPERTIES%LINRES%EPR%PRINT%G_TENSOR%XC")
      ! a possible vdW section in xc_section will be ignored

      IF (gapw_xc) THEN
         CALL qs_vxc_create(ks_env=ks_env, rho_struct=rho_xc, xc_section=xc_section, &
                            vxc_rho=v_rspace_new, vxc_tau=v_tau_rspace, exc=exc, just_energy=.FALSE.)
      ELSE
         CALL qs_vxc_create(ks_env=ks_env, rho_struct=rho, xc_section=xc_section, &
                            vxc_rho=v_rspace_new, vxc_tau=v_tau_rspace, exc=exc, just_energy=.FALSE.)
      END IF

      IF (ASSOCIATED(v_rspace_new)) THEN

         DO ispin = 1, nspins

            CALL pw_transfer(v_rspace_new(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/1, 0, 0/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(1, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_transfer(v_rspace_new(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/0, 1, 0/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(2, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_transfer(v_rspace_new(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/0, 0, 1/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(3, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new(ispin)%pw)

         END DO

         DEALLOCATE (v_rspace_new)
      END IF

      IF (ASSOCIATED(v_tau_rspace)) THEN

         DO ispin = 1, nspins

            CALL pw_transfer(v_tau_rspace(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/1, 0, 0/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(1, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_transfer(v_tau_rspace(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/0, 1, 0/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(2, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_transfer(v_tau_rspace(ispin)%pw, v_xc_gtemp%pw)
            CALL pw_derive(v_xc_gtemp%pw, (/0, 0, 1/))
            CALL pw_transfer(v_xc_gtemp%pw, v_xc_rtemp%pw)
            CALL qs_rho_get(nablavks_set(3, ispin)%rho, rho_r=rho_r)
            CALL pw_axpy(v_xc_rtemp%pw, rho_r(1)%pw)

            CALL pw_pool_give_back_pw(auxbas_pw_pool, v_tau_rspace(ispin)%pw)

         END DO

         DEALLOCATE (v_tau_rspace)
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_xc_gtemp%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_xc_rtemp%pw)

      DEALLOCATE (v_xc_gtemp)
      DEALLOCATE (v_xc_rtemp)

      IF (gapw .OR. gapw_xc) THEN
         CALL calculate_vxc_atom(qs_env=qs_env, energy_only=.FALSE., &
                                 gradient_atom_set=nablavks_atom_set)
      END IF

!   -------------------------------------
!   Write Nabla V_KS (local) to cubes
!   -------------------------------------

      IF (BTEST(cp_print_key_should_output(logger%iter_info, lr_section, &
                                           "EPR%PRINT%NABLAVKS_CUBES"), cp_p_file)) THEN
         ALLOCATE (wf_r)
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         DO idir = 1, 3
            CALL pw_zero(wf_r%pw)
            CALL qs_rho_get(nablavks_set(idir, 1)%rho, rho_r=rho_r)
            CALL pw_copy(rho_r(1)%pw, wf_r%pw) ! RA
            filename = "nablavks"
            WRITE (ext, '(a2,I1,a5)') "_d", idir, ".cube"
            unit_nr = cp_print_key_unit_nr(logger, lr_section, "EPR%PRINT%NABLAVKS_CUBES", &
                                           extension=TRIM(ext), middle_name=TRIM(filename), &
                                           log_filename=.FALSE., file_position="REWIND")
            CALL cp_pw_to_cube(wf_r%pw, unit_nr, "NABLA V_KS ", &
                               particles=particles, &
                               stride=section_get_ivals(lr_section, &
                                                        "EPR%PRINT%NABLAVKS_CUBES%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, lr_section, &
                                              "EPR%PRINT%NABLAVKS_CUBES")
         END DO
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
         DEALLOCATE (wf_r)
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE epr_nablavks

END MODULE qs_linres_epr_nablavks

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Chemical shift calculation by dfpt
!>      Initialization of the nmr_env, creation of the special neighbor lists
!>      Perturbation Hamiltonians by application of the p and rxp oprtators to  psi0
!>      Write output
!>      Deallocate everything
!> \note
!>      The psi0 should be localized
!>      the Sebastiani method works within the assumption that the orbitals are
!>      completely contained in the simulation box
!> \par History
!>       created 07-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_nmr_utils

   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_linres_types,                 ONLY: linres_control_type,&
                                              nmr_env_create,&
                                              nmr_env_type
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: nmr_env_cleanup, nmr_env_init

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_nmr_utils'

CONTAINS

! **************************************************************************************************
!> \brief Initialize the nmr environment
!> \param nmr_env ...
!> \param qs_env ...
!> \par History
!>      07.2006 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE nmr_env_init(nmr_env, qs_env)
      !
      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nmr_env_init', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: label
      CHARACTER(LEN=default_path_length)                 :: nics_file_name
      INTEGER                                            :: handle, ini, ir, j, n_mo(2), n_rep, nao, &
                                                            nat_print, natom, nmoloc, nspins, &
                                                            output_unit
      INTEGER, DIMENSION(:), POINTER                     :: bounds, list
      LOGICAL                                            :: gapw, my_end
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: lr_section, nmr_section

!

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, cell, dft_control, linres_control, scf_control)
      NULLIFY (logger, mpools, nmr_section, particle_set)
      NULLIFY (pw_env, parser)

      n_mo(1:2) = 0
      nao = 0
      nmoloc = 0

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      IF (nmr_env%ref_count /= 0) THEN
         CALL nmr_env_cleanup(nmr_env)
      END IF

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T20,A,/)") "*** Start NMR Chemical Shift Calculation ***"
         WRITE (output_unit, "(T10,A,/)") "Inizialization of the NMR environment"
      ENDIF

      CALL nmr_env_create(nmr_env)
      !
      ! If current_density or full_nmr different allocations are required
      nmr_section => section_vals_get_subs_vals(qs_env%input, &
           &                                    "PROPERTIES%LINRES%NMR")
      CALL section_vals_val_get(nmr_section, "INTERPOLATE_SHIFT", l_val=nmr_env%interpolate_shift)
      CALL section_vals_val_get(nmr_section, "SHIFT_GAPW_RADIUS", r_val=nmr_env%shift_gapw_radius)
      CALL section_vals_val_get(nmr_section, "NICS", l_val=nmr_env%do_nics)
      IF (nmr_env%do_nics) THEN
         CALL section_vals_val_get(nmr_section, "NICS_FILE_NAME", &
                                   c_val=nics_file_name)
         CALL parser_create(parser, nics_file_name)
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, nmr_env%n_nics)
         ALLOCATE (nmr_env%r_nics(3, nmr_env%n_nics))
         CALL parser_get_next_line(parser, 2)
         DO j = 1, nmr_env%n_nics
            CALL parser_get_object(parser, label)
            CALL parser_get_object(parser, nmr_env%r_nics(1, j))
            CALL parser_get_object(parser, nmr_env%r_nics(2, j))
            CALL parser_get_object(parser, nmr_env%r_nics(3, j))
            nmr_env%r_nics(1, j) = cp_unit_to_cp2k(nmr_env%r_nics(1, j), "angstrom")
            nmr_env%r_nics(2, j) = cp_unit_to_cp2k(nmr_env%r_nics(2, j), "angstrom")
            nmr_env%r_nics(3, j) = cp_unit_to_cp2k(nmr_env%r_nics(3, j), "angstrom")
            CALL parser_get_next_line(parser, 1, at_end=my_end)
            IF (my_end) EXIT
         ENDDO
         CALL parser_release(parser)
      ENDIF

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      mpools=mpools, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      scf_control=scf_control)
      !
      ! Check if restat also psi0 should be restarted
      !IF(nmr_env%restart_nmr .AND. scf_control%density_guess/=restart_guess)THEN
      !   CALL cp__b("qs_linres_nmr_utils.F",167,"restart_nmr requires density_guess=restart")
      !ENDIF
      !
      ! check that the psi0 are localized and you have all the centers
      IF (.NOT. linres_control%localized_psi0) &
         CALL cp__w("qs_linres_nmr_utils.F",172,' To get NMR parameters within PBC you need localized zero order orbitals ')

      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)

      !
      ! Conversion factors
      ! factor for the CHEMICAL SHIFTS: alpha^2 *  ppm.
      nmr_env%shift_factor = (1.0_dp/137.03602_dp)**2*1.0E+6_dp/cell%deth
      ! factor for the CHEMICAL SHIFTS: alpha^2 *  ppm.
      nmr_env%shift_factor_gapw = (1.0_dp/137.03602_dp)**2*1.0E+6_dp
      ! chi_factor =  1/4 * e^2/m * a_0 ^2
      nmr_env%chi_factor = 1.9727566E-29_dp/1.0E-30_dp ! -> displayed in 10^-30 J/T^2
      ! Factor to convert 10^-30 J/T^2 into ppm cgs = ppm cm^3/mol
      ! = 10^-30 * mu_0/4pi * N_A * 10^6 * 10^6  [one 10^6 for ppm, one for m^3 -> cm^3]
      nmr_env%chi_SI2ppmcgs = 6.022045_dp/1.0E+2_dp
      ! Chi to Shift: 10^-30  *  2/3  mu_0 / Omega  * 1/ppm
      nmr_env%chi_SI2shiftppm = 1.0E-30_dp*8.37758041E-7_dp/ &
                                (cp_unit_from_cp2k(cell%deth, "angstrom^3")*1.0E-30_dp)*1.0E+6_dp

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Shift gapw radius (a.u.) ", nmr_env%shift_gapw_radius
         IF (nmr_env%do_nics) THEN
            WRITE (output_unit, "(T2,A,T50,I5,A)") "NMR| NICS computed in ", nmr_env%n_nics, " additional points"
            WRITE (output_unit, "(T2,A,T60,A)") "NMR| NICS coordinates read on file ", TRIM(nics_file_name)
         ENDIF
         WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Shift factor (ppm)", nmr_env%shift_factor
         IF (gapw) THEN
            WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Shift factor gapw (ppm)", nmr_env%shift_factor_gapw
         ENDIF
         WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Chi factor (SI)", nmr_env%chi_factor
         WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Conversion Chi (ppm/cgs)", nmr_env%chi_SI2ppmcgs
         WRITE (output_unit, "(T2,A,T65,ES15.6)") "NMR| Conversion Chi to Shift", nmr_env%chi_SI2shiftppm
      ENDIF

      ALLOCATE (nmr_env%do_calc_cs_atom(natom))
      nmr_env%do_calc_cs_atom = 0

      IF (BTEST(cp_print_key_should_output(logger%iter_info, nmr_section,&
           &    "PRINT%SHIELDING_TENSOR"), cp_p_file)) THEN

         NULLIFY (bounds, list)
         nat_print = 0
         CALL section_vals_val_get(nmr_section,&
              &                    "PRINT%SHIELDING_TENSOR%ATOMS_LU_BOUNDS", &
                                   i_vals=bounds)
         nat_print = bounds(2)-bounds(1)+1
         IF (nat_print > 0) THEN
            ALLOCATE (nmr_env%cs_atom_list(nat_print))
            DO ir = 1, nat_print
               nmr_env%cs_atom_list(ir) = bounds(1)+(ir-1)
               nmr_env%do_calc_cs_atom(bounds(1)+(ir-1)) = 1
            ENDDO
         ENDIF

         IF (.NOT. ASSOCIATED(nmr_env%cs_atom_list)) THEN
            CALL section_vals_val_get(nmr_section, "PRINT%SHIELDING_TENSOR%ATOMS_LIST", &
                                      n_rep_val=n_rep)
            nat_print = 0
            DO ir = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(nmr_section, "PRINT%SHIELDING_TENSOR%ATOMS_LIST", &
                                         i_rep_val=ir, i_vals=list)
               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(nmr_env%cs_atom_list, 1, nat_print+SIZE(list))
                  DO ini = 1, SIZE(list)
                     nmr_env%cs_atom_list(ini+nat_print) = list(ini)
                     nmr_env%do_calc_cs_atom(list(ini)) = 1
                  ENDDO
                  nat_print = nat_print+SIZE(list)
               ENDIF
            ENDDO ! ir
         ENDIF

         IF (.NOT. ASSOCIATED(nmr_env%cs_atom_list)) THEN
            ALLOCATE (nmr_env%cs_atom_list(natom))
            DO ir = 1, natom
               nmr_env%cs_atom_list(ir) = ir
            ENDDO
            nmr_env%do_calc_cs_atom = 1
         ENDIF
         !
         ! check the list
         IF(.NOT.(ASSOCIATED(nmr_env%cs_atom_list)))CALL cp__a("qs_linres_nmr_utils.F",256)
         DO ir = 1, SIZE(nmr_env%cs_atom_list, 1)
            IF (nmr_env%cs_atom_list(ir) .LT. 1 .OR. nmr_env%cs_atom_list(ir) .GT. natom) THEN
               CALL cp__b("qs_linres_nmr_utils.F",259,"Unknown atom(s)")
            ENDIF
            DO j = 1, SIZE(nmr_env%cs_atom_list, 1)
               IF (j .EQ. ir) CYCLE
               IF (nmr_env%cs_atom_list(ir) .EQ. nmr_env%cs_atom_list(j)) THEN
                  CALL cp__b("qs_linres_nmr_utils.F",264,"Duplicate atoms")
               ENDIF
            ENDDO
         ENDDO
      ELSE
         NULLIFY (nmr_env%cs_atom_list)
      ENDIF

      IF (output_unit > 0) THEN
         IF (ASSOCIATED(nmr_env%cs_atom_list)) THEN
            WRITE (output_unit, "(T2,A,T69,I5,A)") "NMR| Shielding tensor computed for ", &
               SIZE(nmr_env%cs_atom_list, 1), " atoms"
         ELSE
            WRITE (output_unit, "(T2,A,T50)")&
                 & "NMR| Shielding tensor not computed at the atomic positions"
         ENDIF
      ENDIF
      !
      ! Initialize the chemical shift tensor
      ALLOCATE (nmr_env%chemical_shift(3, 3, natom), &
                nmr_env%chemical_shift_loc(3, 3, natom))
      nmr_env%chemical_shift = 0.0_dp
      nmr_env%chemical_shift_loc = 0.0_dp
      IF (nmr_env%do_nics) THEN
         ALLOCATE (nmr_env%chemical_shift_nics_loc(3, 3, nmr_env%n_nics), &
                   nmr_env%chemical_shift_nics(3, 3, nmr_env%n_nics))
         nmr_env%chemical_shift_nics_loc = 0.0_dp
         nmr_env%chemical_shift_nics = 0.0_dp
      ENDIF

      CALL cp_print_key_finished_output(output_unit, logger, lr_section,&
           &                            "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE nmr_env_init

! **************************************************************************************************
!> \brief Deallocate the nmr environment
!> \param nmr_env ...
!> \par History
!>      07.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE nmr_env_cleanup(nmr_env)

      TYPE(nmr_env_type)                                 :: nmr_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nmr_env_cleanup', &
         routineP = moduleN//':'//routineN

      nmr_env%ref_count = nmr_env%ref_count-1
      IF (nmr_env%ref_count == 0) THEN
         IF (ASSOCIATED(nmr_env%cs_atom_list)) THEN
            DEALLOCATE (nmr_env%cs_atom_list)
         END IF
         IF (ASSOCIATED(nmr_env%do_calc_cs_atom)) THEN
            DEALLOCATE (nmr_env%do_calc_cs_atom)
         END IF
         !chemical_shift
         IF (ASSOCIATED(nmr_env%chemical_shift)) THEN
            DEALLOCATE (nmr_env%chemical_shift)
         END IF
         IF (ASSOCIATED(nmr_env%chemical_shift_loc)) THEN
            DEALLOCATE (nmr_env%chemical_shift_loc)
         END IF
         ! nics
         IF (ASSOCIATED(nmr_env%r_nics)) THEN
            DEALLOCATE (nmr_env%r_nics)
         END IF
         IF (ASSOCIATED(nmr_env%chemical_shift_nics)) THEN
            DEALLOCATE (nmr_env%chemical_shift_nics)
         END IF
         IF (ASSOCIATED(nmr_env%chemical_shift_nics_loc)) THEN
            DEALLOCATE (nmr_env%chemical_shift_nics_loc)
         END IF
      END IF ! ref count

   END SUBROUTINE nmr_env_cleanup

END MODULE qs_linres_nmr_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Polarizability calculation by dfpt
!>      Initialization of the polar_env,
!>      Perturbation Hamiltonian by application of the Berry phase operator to psi0
!>      Write output
!>      Deallocate everything
!> periodic Raman SL February 2013
!> \note
! **************************************************************************************************
MODULE qs_linres_polar_utils
   USE bibliography,                    ONLY: Luber2014,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_fm_basic_linalg,              ONLY: cp_fm_frobenius_norm,&
                                              cp_fm_trace
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              put_results
   USE cp_result_types,                 ONLY: cp_result_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: twopi
   USE physcon,                         ONLY: angstrom
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_linres_methods,               ONLY: linres_solver
   USE qs_linres_types,                 ONLY: get_polar_env,&
                                              linres_control_type,&
                                              polar_env_create,&
                                              polar_env_type
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: polar_env_cleanup, polar_env_init, polar_response, polar_polar, polar_print

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_polar_utils'

CONTAINS

! **************************************************************************************************
!> \brief Initialize the polar environment
!> \param polar_env ...
!> \param p_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE polar_response(polar_env, p_env, qs_env)

      TYPE(polar_env_type)                               :: polar_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_response', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, ispin, nao, nmo, nspins, &
                                                            output_unit
      LOGICAL                                            :: do_raman, should_stop
      REAL(dp)                                           :: chk, fro
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: h1_psi0, psi0_order, psi1
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dBerry_psi0, psi1_dBerry
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(section_vals_type), POINTER                   :: lr_section, polar_section

      CALL timeset(routineN, handle)
      !
      NULLIFY (dft_control, linres_control, lr_section, polar_section)
      NULLIFY (logger, mpools, psi1, h1_psi0, mo_coeff, para_env)
      NULLIFY (tmp_fm_struct, psi1_dBerry, dBerry_psi0)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      polar_section => section_vals_get_subs_vals(qs_env%input, &
                                                  "PROPERTIES%LINRES%POLAR")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T10,A,/)") &
            "*** Self consistent optimization of the response wavefunctions ***"
      ENDIF

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mpools=mpools, &
                      linres_control=linres_control, &
                      mos=mos, &
                      para_env=para_env)

      nspins = dft_control%nspins

      CALL get_polar_env(polar_env=polar_env, &
                         do_raman=do_raman)
      !
      ! allocate the vectors
      ALLOCATE (psi0_order(nspins))
      ALLOCATE (psi1(nspins), h1_psi0(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         psi0_order(ispin)%matrix => mo_coeff
         CALL cp_fm_get_info(mo_coeff, ncol_global=nmo, nrow_global=nao)
         NULLIFY (tmp_fm_struct, psi1(ispin)%matrix, h1_psi0(ispin)%matrix)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=nmo, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(psi1(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_create(h1_psi0(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      ENDDO
      chk = 0.0_dp
      !
      IF (do_raman) THEN
         CALL get_polar_env(polar_env=polar_env, &
                            psi1_dBerry=psi1_dBerry, &
                            dBerry_psi0=dBerry_psi0)
         DO idir = 1, 3
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") &
                  "Response to the perturbation operator Berry phase_"//ACHAR(idir+119)
            ENDIF
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1_dBerry(idir, ispin)%matrix, 0.0_dp)
            ENDDO
            !
            !Initial guess for psi1
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
            ENDDO
            !
            !DO scf cycle to optimize psi1
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(dBerry_psi0(idir, ispin)%matrix, h1_psi0(ispin)%matrix)
            ENDDO
            !
            !
            linres_control%lr_triplet = .FALSE. ! we do singlet response
            linres_control%do_kernel = .TRUE. ! we do coupled response
            linres_control%converged = .FALSE.
            CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
            !
            !
            ! copy the response
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_dBerry(idir, ispin)%matrix)
               CALL cp_fm_frobenius_norm(psi1(ispin)%matrix, fro)
               chk = chk+fro
            ENDDO
            !
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Write the resulting psi1 in restart file... not implemented yet"
            ENDIF
            !
            ! Write the result in the restart file
         ENDDO ! idir
      ENDIF !raman

      !
      ! print the checksum
      !   IF(output_unit>0) THEN
      !      WRITE(output_unit,'(T2,A,E23.16)') 'polar| response: CheckSum =',chk
      !   ENDIF
      !
      !
      ! clean up
      DO ispin = 1, nspins
         CALL cp_fm_release(psi1(ispin)%matrix)
         CALL cp_fm_release(h1_psi0(ispin)%matrix)
      ENDDO
      DEALLOCATE (psi1, h1_psi0, psi0_order)
      !
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      !
      CALL timestop(handle)
      !
   END SUBROUTINE polar_response

! **************************************************************************************************
!> \brief ...
!> \param polar_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE polar_polar(polar_env, qs_env)

      TYPE(polar_env_type)                               :: polar_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_polar', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, nspins, output_unit, z
      LOGICAL                                            :: do_raman
      REAL(dp)                                           :: ptmp
      REAL(dp), DIMENSION(:, :), POINTER                 :: polar, polar_tmp
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dBerry_psi0, mo_derivs, psi1_dBerry
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, polar, psi1_dBerry, logger)
      NULLIFY (mos, dBerry_psi0, mo_derivs)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      mos=mos)

      nspins = dft_control%nspins

      CALL get_polar_env(polar_env=polar_env, &
                         do_raman=do_raman)
!
      IF (do_raman) THEN

         CALL cite_reference(Luber2014)

         CALL get_polar_env(polar_env=polar_env, &
                            psi1_dBerry=psi1_dBerry, &
                            dBerry_psi0=dBerry_psi0, &
                            polar=polar, &
                            mo_derivs=mo_derivs)
         !
         !
!    polar_section => section_vals_get_subs_vals(qs_env%input, &
!         & "PROPERTIES%LINRES%POLAR")
         !
         ! Initialize
         ALLOCATE (polar_tmp(3, 3))
         polar_tmp(:, :) = 0.0_dp

         DO i = 1, 3 ! directions of electric field
            DO z = 1, 3 !dipole directions
               DO ispin = 1, dft_control%nspins
!SL compute trace
                  ptmp = 0.0_dp
                  CALL cp_fm_trace(psi1_dBerry(i, ispin)%matrix, mo_derivs(z, ispin)%matrix, ptmp)
                  polar_tmp(i, z) = polar_tmp(i, z)+ptmp
               END DO
               polar_tmp(i, z) = 2.0_dp*polar_tmp(i, z)/(twopi*twopi)
            END DO
         END DO !spin

         polar(:, :) = MATMUL(MATMUL(cell%hmat, polar_tmp), TRANSPOSE(cell%hmat))
!SL evtl maxocc instead?
         IF (dft_control%nspins .EQ. 1) THEN
            polar(:, :) = 2.0_dp*polar(:, :)
         END IF

         !
         IF (ASSOCIATED(polar_tmp)) THEN
            DEALLOCATE (polar_tmp)
         END IF

      ENDIF !raman
      !
      ! Finalize
      CALL timestop(handle)
      !
   END SUBROUTINE polar_polar

! **************************************************************************************************
!> \brief ...
!> \param polar_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE polar_print(polar_env, qs_env)
      TYPE(polar_env_type)                               :: polar_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'polar_print', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      INTEGER                                            :: output_unit, unit_p
      LOGICAL                                            :: do_raman
      REAL(dp), DIMENSION(:, :), POINTER                 :: polar
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: polar_section

      NULLIFY (logger, dft_control, para_env, results)
      CALL get_qs_env(qs_env=qs_env, results=results, dft_control=dft_control, &
                      para_env=para_env)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      polar_section => section_vals_get_subs_vals(qs_env%input, &
                                                  "PROPERTIES%LINRES%POLAR")

      CALL get_polar_env(polar_env=polar_env, polar=polar, do_raman=do_raman)

      description = "[POLAR]"
      CALL cp_results_erase(results, description=description)
      CALL put_results(results, description=description, values=polar(:, :))

      IF (do_raman) THEN
         IF (BTEST(cp_print_key_should_output(logger%iter_info, polar_section, &
                                              "PRINT%POLAR_MATRIX"), cp_p_file)) THEN

            unit_p = cp_print_key_unit_nr(logger, polar_section, "PRINT%POLAR_MATRIX", &
                                          extension=".data", middle_name="raman", log_filename=.FALSE.)
            IF (unit_p > 0) THEN
               IF (unit_p /= output_unit) THEN
                  WRITE (unit_p, *)
                  WRITE (unit_p, '(T10,A)') 'POLARIZABILITY TENSOR (atomic units):'
                  WRITE (unit_p, '(T10,A,3F15.5)') "xx,yy,zz", polar(1, 1), polar(2, 2), polar(3, 3)
                  WRITE (unit_p, '(T10,A,3F15.5)') "xy,xz,yz", polar(1, 2), polar(1, 3), polar(2, 3)
                  WRITE (unit_p, '(T10,A,3F15.5)') "yx,zx,zy", polar(2, 1), polar(3, 1), polar(3, 2)
                  WRITE (unit_p, '(T10,A)') 'POLARIZABILITY TENSOR (Angstrom^3):'
                  WRITE (unit_p, '(T10,A,3F15.5)') "xx,yy,zz", polar(1, 1)*angstrom**3, &
                     polar(2, 2)*angstrom**3, polar(3, 3)*angstrom**3
                  WRITE (unit_p, '(T10,A,3F15.5)') "xy,xz,yz", polar(1, 2)*angstrom**3, &
                     polar(1, 3)*angstrom**3, polar(2, 3)*angstrom**3
                  WRITE (unit_p, '(T10,A,3F15.5)') "yx,zx,zy", polar(2, 1)*angstrom**3, &
                     polar(3, 1)*angstrom**3, polar(3, 2)*angstrom**3
                  CALL cp_print_key_finished_output(unit_p, logger, polar_section, &
                                                    "PRINT%POLAR_MATRIX")
               ENDIF
            ENDIF
         ENDIF
         IF (output_unit > 0) THEN
            WRITE (output_unit, *)
            WRITE (output_unit, '(T10,A)') 'POLARIZABILITY TENSOR (atomic units):'
            WRITE (output_unit, '(T10,A,3F15.5)') "xx,yy,zz", polar(1, 1), polar(2, 2), polar(3, 3)
            WRITE (output_unit, '(T10,A,3F15.5)') "xy,xz,yz", polar(1, 2), polar(1, 3), polar(2, 3)
            WRITE (output_unit, '(T10,A,3F15.5)') "yx,zx,zy", polar(2, 1), polar(3, 1), polar(3, 2)
            WRITE (output_unit, '(T10,A)') 'POLARIZABILITY TENSOR (Angstrom^3):'
            WRITE (output_unit, '(T10,A,3F15.5)') "xx,yy,zz", polar(1, 1)*angstrom**3, &
               polar(2, 2)*angstrom**3, polar(3, 3)*angstrom**3
            WRITE (output_unit, '(T10,A,3F15.5)') "xy,xz,yz", polar(1, 2)*angstrom**3, &
               polar(1, 3)*angstrom**3, polar(2, 3)*angstrom**3
            WRITE (output_unit, '(T10,A,3F15.5)') "yx,zx,zy", polar(2, 1)*angstrom**3, &
               polar(3, 1)*angstrom**3, polar(3, 2)*angstrom**3
         END IF
      END IF
      !
      !
   END SUBROUTINE polar_print

! **************************************************************************************************
!> \brief Initialize the polar environment
!> \param polar_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE polar_env_init(polar_env, qs_env)
      !
      TYPE(polar_env_type)                               :: polar_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_env_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, ispin, m, nao, nmo, &
                                                            nspins, output_unit
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: lr_section, polar_section

!

      CALL timeset(routineN, handle)

      NULLIFY (linres_control)
      NULLIFY (logger, polar_section)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      IF (polar_env%ref_count /= 0) THEN
         CALL polar_env_cleanup(polar_env)
      ENDIF

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T20,A,/)") "*** Start polarizability calculation ***"
         WRITE (output_unit, "(T10,A,/)") "Initialization of the polar environment"
      ENDIF

      CALL polar_env_create(polar_env)
      !
      polar_section => section_vals_get_subs_vals(qs_env%input, &
                                                  "PROPERTIES%LINRES%POLAR")

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      matrix_s=matrix_s, &
                      linres_control=linres_control, &
                      mos=mos)
      !
      !
      nspins = dft_control%nspins
      !
      CALL section_vals_val_get(polar_section, "DO_RAMAN", l_val=polar_env%do_raman)
      !
      !
      !
      ALLOCATE (polar_env%polar(3, 3))
      polar_env%polar(:, :) = 0.0_dp
      !
      ! allocation
      ALLOCATE (polar_env%dBerry_psi0(3, nspins), &
                polar_env%psi1_dBerry(3, nspins), polar_env%mo_derivs(3, nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
         CALL cp_fm_get_info(mo_coeff, ncol_global=m, nrow_global=nao)

         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=m, &
                                  context=mo_coeff%matrix_struct%context)
         DO idir = 1, 3
            NULLIFY (polar_env%psi1_dBerry(idir, ispin)%matrix, polar_env%dBerry_psi0(idir, ispin)%matrix, &
                     polar_env%mo_derivs(idir, ispin)%matrix)
            CALL cp_fm_create(polar_env%psi1_dBerry(idir, ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_create(polar_env%dBerry_psi0(idir, ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_create(polar_env%mo_derivs(idir, ispin)%matrix, tmp_fm_struct)
         ENDDO
         CALL cp_fm_struct_release(tmp_fm_struct)

         !
      END DO

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE polar_env_init

! **************************************************************************************************
!> \brief Deallocate the polar environment
!> \param polar_env ...
!> \par History
! **************************************************************************************************
   SUBROUTINE polar_env_cleanup(polar_env)

      TYPE(polar_env_type)                               :: polar_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_env_cleanup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idir, ispin

      polar_env%ref_count = polar_env%ref_count-1
      IF (polar_env%ref_count == 0) THEN
         IF (ASSOCIATED(polar_env%polar)) THEN
            DEALLOCATE (polar_env%polar)
         ENDIF
         IF (ASSOCIATED(polar_env%dBerry_psi0)) THEN
            DO idir = 1, SIZE(polar_env%dBerry_psi0, 1)
               DO ispin = 1, SIZE(polar_env%dBerry_psi0, 2)
                  CALL cp_fm_release(polar_env%dBerry_psi0(idir, ispin)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (polar_env%dBerry_psi0)
         ENDIF
         IF (ASSOCIATED(polar_env%mo_derivs)) THEN
            DO idir = 1, SIZE(polar_env%mo_derivs, 1)
               DO ispin = 1, SIZE(polar_env%mo_derivs, 2)
                  CALL cp_fm_release(polar_env%mo_derivs(idir, ispin)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (polar_env%mo_derivs)
         ENDIF
         IF (ASSOCIATED(polar_env%psi1_dBerry)) THEN
            DO idir = 1, SIZE(polar_env%psi1_dBerry, 1)
               DO ispin = 1, SIZE(polar_env%psi1_dBerry, 2)
                  CALL cp_fm_release(polar_env%psi1_dBerry(idir, ispin)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (polar_env%psi1_dBerry)
         ENDIF
         !

      ENDIF ! ref count

   END SUBROUTINE polar_env_cleanup

END MODULE qs_linres_polar_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of Coulomb integrals over Cartesian Gaussian-type functions
!>      (electron repulsion integrals, ERIs).
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \par Parameters
!>       - ax,ay,az    : Angular momentum index numbers of orbital a.
!>       - bx,by,bz    : Angular momentum index numbers of orbital b.
!>       - coset       : Cartesian orbital set pointer.
!>       - dab         : Distance between the atomic centers a and b.
!>       - dac         : Distance between the atomic centers a and c.
!>       - dbc         : Distance between the atomic centers b and c.
!>       - l{a,b,c}    : Angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_max: Maximum angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_min: Minimum angular momentum quantum number of shell a, b or c.
!>       - ncoset      : Number of orbitals in a Cartesian orbital set.
!>       - npgf{a,b}   : Degree of contraction of shell a or b.
!>       - rab         : Distance vector between the atomic centers a and b.
!>       - rab2        : Square of the distance between the atomic centers a and b.
!>       - rac         : Distance vector between the atomic centers a and c.
!>       - rac2        : Square of the distance between the atomic centers a and c.
!>       - rbc         : Distance vector between the atomic centers b and c.
!>       - rbc2        : Square of the distance between the atomic centers b and c.
!>       - rpgf{a,b,c} : Radius of the primitive Gaussian-type function a, b or c.
!>       - zet{a,b,c}  : Exponents of the Gaussian-type functions a, b or c.
!>       - zetp        : Reciprocal of the sum of the exponents of orbital a and b.
!> \author VW
! **************************************************************************************************
MODULE ai_elec_field
   USE ai_os_rr,                        ONLY: os_rr_coul
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_elec_field'
   PRIVATE

   ! *** Public subroutines ***

   PUBLIC :: efg

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the primitive electric field integrals over
!>          Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param rac ...
!> \param rbc ...
!> \param rab ...
!> \param vab ...
!> \param ldrr1 ...
!> \param ldrr2 ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE efg(la_max, la_min, npgfa, rpgfa, zeta, &
                  lb_max, lb_min, npgfb, rpgfb, zetb, &
                  rac, rbc, rab, vab, ldrr1, ldrr2, rr)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac, rbc, rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: vab
      INTEGER, INTENT(IN)                                :: ldrr1, ldrr2
      REAL(KIND=dp), DIMENSION(0:ldrr1-1, ldrr2, *), &
         INTENT(INOUT)                                   :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'efg', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coa, coam1x, coam1y, coam1z, coam2x, coam2y, coam2z, &
         coamxpy, coamxpz, coamxy, coamxz, coamypx, coamypz, coamyz, coamzpx, coamzpy, coap1x, &
         coap1y, coap1z, coap2x, coap2y, coap2z, coapxy, coapxz, coapyz, cob, cobm1x, cobm1y, &
         cobm1z, cobm2x, cobm2y, cobm2z, cobmxpy, cobmxpz, cobmxy, cobmxz, cobmypx, cobmypz, &
         cobmyz, cobmzpx, cobmzpy, cobp1x, cobp1y, cobp1z, cobp2x, cobp2y, cobp2z, cobpxy, cobpxz, &
         cobpyz, i, ipgf, j, jpgf, la, lb, ma, mb, na, nb
      REAL(KIND=dp)                                      :: dab, dum, dumxx, dumxy, dumxz, dumyy, &
                                                            dumyz, dumzz, f0, rab2, xhi, za, zb, &
                                                            zet
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp, rcp

! *** Calculate the distance of the centers a and c ***

      rab2 = rab(1)**2+rab(2)**2+rab(3)**2
      dab = SQRT(rab2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max)
                  DO i = na+1, na+ncoset(la_max)
                     vab(i, j, 1) = 0.0_dp
                     vab(i, j, 2) = 0.0_dp
                     vab(i, j, 3) = 0.0_dp
                     vab(i, j, 4) = 0.0_dp
                     vab(i, j, 5) = 0.0_dp
                     vab(i, j, 6) = 0.0_dp
                  ENDDO
               ENDDO
               nb = nb+ncoset(lb_max)
               CYCLE
            ENDIF

            ! *** Calculate some prefactors ***

            za = zeta(ipgf)
            zb = zetb(jpgf)
            zet = za+zb
            xhi = za*zb/zet
            rap = zb*rab/zet
            rbp = -za*rab/zet
            rcp = -(za*rac+zb*rbc)/zet
            f0 = 2.0_dp*SQRT(zet/pi)*(pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! *** Calculate the reccurence relation

            CALL os_rr_coul(rap, la_max+2, rbp, lb_max+2, rcp, zet, ldrr1, ldrr2, rr)

            ! *** Calculate the primitive electric field gradient integrals ***

            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)
               cobm1x = coset(MAX(bx-1, 0), by, bz)
               cobm1y = coset(bx, MAX(by-1, 0), bz)
               cobm1z = coset(bx, by, MAX(bz-1, 0))
               cobm2x = coset(MAX(bx-2, 0), by, bz)
               cobm2y = coset(bx, MAX(by-2, 0), bz)
               cobm2z = coset(bx, by, MAX(bz-2, 0))
               cobmxy = coset(MAX(bx-1, 0), MAX(by-1, 0), bz)
               cobmxz = coset(MAX(bx-1, 0), by, MAX(bz-1, 0))
               cobmyz = coset(bx, MAX(by-1, 0), MAX(bz-1, 0))
               cobp1x = coset(bx+1, by, bz)
               cobp1y = coset(bx, by+1, bz)
               cobp1z = coset(bx, by, bz+1)
               cobp2x = coset(bx+2, by, bz)
               cobp2y = coset(bx, by+2, bz)
               cobp2z = coset(bx, by, bz+2)
               cobpxy = coset(bx+1, by+1, bz)
               cobpxz = coset(bx+1, by, bz+1)
               cobpyz = coset(bx, by+1, bz+1)
               cobmxpy = coset(MAX(bx-1, 0), by+1, bz)
               cobmxpz = coset(MAX(bx-1, 0), by, bz+1)
               cobmypx = coset(bx+1, MAX(by-1, 0), bz)
               cobmypz = coset(bx, MAX(by-1, 0), bz+1)
               cobmzpx = coset(bx+1, by, MAX(bz-1, 0))
               cobmzpy = coset(bx, by+1, MAX(bz-1, 0))
               mb = nb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)
                  coap1x = coset(ax+1, ay, az)
                  coap1y = coset(ax, ay+1, az)
                  coap1z = coset(ax, ay, az+1)
                  coap2x = coset(ax+2, ay, az)
                  coap2y = coset(ax, ay+2, az)
                  coap2z = coset(ax, ay, az+2)
                  coapxy = coset(ax+1, ay+1, az)
                  coapxz = coset(ax+1, ay, az+1)
                  coapyz = coset(ax, ay+1, az+1)
                  coam1x = coset(MAX(ax-1, 0), ay, az)
                  coam1y = coset(ax, MAX(ay-1, 0), az)
                  coam1z = coset(ax, ay, MAX(az-1, 0))
                  coam2x = coset(MAX(ax-2, 0), ay, az)
                  coam2y = coset(ax, MAX(ay-2, 0), az)
                  coam2z = coset(ax, ay, MAX(az-2, 0))
                  coamxy = coset(MAX(ax-1, 0), MAX(ay-1, 0), az)
                  coamxz = coset(MAX(ax-1, 0), ay, MAX(az-1, 0))
                  coamyz = coset(ax, MAX(ay-1, 0), MAX(az-1, 0))
                  coamxpy = coset(MAX(ax-1, 0), ay+1, az)
                  coamxpz = coset(MAX(ax-1, 0), ay, az+1)
                  coamypx = coset(ax+1, MAX(ay-1, 0), az)
                  coamypz = coset(ax, MAX(ay-1, 0), az+1)
                  coamzpx = coset(ax+1, ay, MAX(az-1, 0))
                  coamzpy = coset(ax, ay+1, MAX(az-1, 0))
                  ma = na+coa
                  !
                  ! (a|xx|b)
                  dum = 4.0_dp*(za**2*rr(0, coap2x, cob)+zb**2*rr(0, coa, cobp2x) &
                       &         +2.0_dp*za*zb*rr(0, coap1x, cobp1x)) &
                       -2.0_dp*rr(0, coa, cob)*(za*REAL(2*ax+1, dp)+zb*REAL(2*bx+1, dp))
                  IF (ax .GT. 1) dum = dum+REAL(ax*(ax-1), dp)*rr(0, coam2x, cob)
                  IF (bx .GT. 1) dum = dum+REAL(bx*(bx-1), dp)*rr(0, coa, cobm2x)
                  IF (ax .GT. 0) dum = dum-4.0_dp*zb*REAL(ax, dp)*rr(0, coam1x, cobp1x)
                  IF (bx .GT. 0) dum = dum-4.0_dp*za*REAL(bx, dp)*rr(0, coap1x, cobm1x)
                  IF (ax .GT. 0 .AND. bx .GT. 0) dum = dum+2.0_dp*REAL(ax*bx, dp)*rr(0, coam1x, cobm1x)
                  dumxx = f0*dum
                  !
                  ! (a|yy|b)
                  dum = 4.0_dp*(za**2*rr(0, coap2y, cob)+zb**2*rr(0, coa, cobp2y) &
                       &         +2.0_dp*za*zb*rr(0, coap1y, cobp1y)) &
                       -2.0_dp*rr(0, coa, cob)*(za*REAL(2*ay+1, dp)+zb*REAL(2*by+1, dp))
                  IF (ay .GT. 1) dum = dum+REAL(ay*(ay-1), dp)*rr(0, coam2y, cob)
                  IF (by .GT. 1) dum = dum+REAL(by*(by-1), dp)*rr(0, coa, cobm2y)
                  IF (ay .GT. 0) dum = dum-4.0_dp*zb*REAL(ay, dp)*rr(0, coam1y, cobp1y)
                  IF (by .GT. 0) dum = dum-4.0_dp*za*REAL(by, dp)*rr(0, coap1y, cobm1y)
                  IF (ay .GT. 0 .AND. by .GT. 0) dum = dum+2.0_dp*REAL(ay*by, dp)*rr(0, coam1y, cobm1y)
                  dumyy = f0*dum
                  !
                  ! (a|zz|b)
                  dum = 4.0_dp*(za**2*rr(0, coap2z, cob)+zb**2*rr(0, coa, cobp2z) &
                       &         +2.0_dp*za*zb*rr(0, coap1z, cobp1z)) &
                       -2.0_dp*rr(0, coa, cob)*(za*REAL(2*az+1, dp)+zb*REAL(2*bz+1, dp))
                  IF (az .GT. 1) dum = dum+REAL(az*(az-1), dp)*rr(0, coam2z, cob)
                  IF (bz .GT. 1) dum = dum+REAL(bz*(bz-1), dp)*rr(0, coa, cobm2z)
                  IF (az .GT. 0) dum = dum-4.0_dp*zb*REAL(az, dp)*rr(0, coam1z, cobp1z)
                  IF (bz .GT. 0) dum = dum-4.0_dp*za*REAL(bz, dp)*rr(0, coap1z, cobm1z)
                  IF (az .GT. 0 .AND. bz .GT. 0) dum = dum+2.0_dp*REAL(az*bz, dp)*rr(0, coam1z, cobm1z)
                  dumzz = f0*dum
                  !
                  ! (a|xy|b)
                  dum = 4.0_dp*(za**2*rr(0, coapxy, cob)+zb**2*rr(0, coa, cobpxy) &
                       &         +za*zb*(rr(0, coap1x, cobp1y)+rr(0, coap1y, cobp1x)))
                  IF (ax .GT. 0) dum = dum-2.0_dp*REAL(ax, dp)* &
                       &  (za*rr(0, coamxpy, cob)+zb*rr(0, coam1x, cobp1y))
                  IF (ay .GT. 0) dum = dum-2.0_dp*REAL(ay, dp)* &
                       &  (za*rr(0, coamypx, cob)+zb*rr(0, coam1y, cobp1x))
                  IF (ax .GT. 0 .AND. ay .GT. 0) dum = dum+REAL(ax*ay, dp)*rr(0, coamxy, cob)
                  IF (bx .GT. 0) dum = dum-2.0_dp*REAL(bx, dp)* &
                       &  (zb*rr(0, coa, cobmxpy)+za*rr(0, coap1y, cobm1x))
                  IF (by .GT. 0) dum = dum-2.0_dp*REAL(by, dp)* &
                       &  (zb*rr(0, coa, cobmypx)+za*rr(0, coap1x, cobm1y))
                  IF (bx .GT. 0 .AND. by .GT. 0) dum = dum+REAL(bx*by, dp)*rr(0, coa, cobmxy)
                  IF (ax .GT. 0 .AND. by .GT. 0) dum = dum+REAL(ax*by, dp)*rr(0, coam1x, cobm1y)
                  IF (ay .GT. 0 .AND. bx .GT. 0) dum = dum+REAL(ay*bx, dp)*rr(0, coam1y, cobm1x)
                  dumxy = f0*dum
                  !
                  ! (a|xz|b)
                  dum = 4.0_dp*(za**2*rr(0, coapxz, cob)+zb**2*rr(0, coa, cobpxz) &
                       &         +za*zb*(rr(0, coap1x, cobp1z)+rr(0, coap1z, cobp1x)))
                  IF (ax .GT. 0) dum = dum-2.0_dp*REAL(ax, dp)* &
                       &  (za*rr(0, coamxpz, cob)+zb*rr(0, coam1x, cobp1z))
                  IF (az .GT. 0) dum = dum-2.0_dp*REAL(az, dp)* &
                       &  (za*rr(0, coamzpx, cob)+zb*rr(0, coam1z, cobp1x))
                  IF (ax .GT. 0 .AND. az .GT. 0) dum = dum+REAL(ax*az, dp)*rr(0, coamxz, cob)
                  IF (bx .GT. 0) dum = dum-2.0_dp*REAL(bx, dp)* &
                       &  (zb*rr(0, coa, cobmxpz)+za*rr(0, coap1z, cobm1x))
                  IF (bz .GT. 0) dum = dum-2.0_dp*REAL(bz, dp)* &
                       &  (zb*rr(0, coa, cobmzpx)+za*rr(0, coap1x, cobm1z))
                  IF (bx .GT. 0 .AND. bz .GT. 0) dum = dum+REAL(bx*bz, dp)*rr(0, coa, cobmxz)
                  IF (ax .GT. 0 .AND. bz .GT. 0) dum = dum+REAL(ax*bz, dp)*rr(0, coam1x, cobm1z)
                  IF (az .GT. 0 .AND. bx .GT. 0) dum = dum+REAL(az*bx, dp)*rr(0, coam1z, cobm1x)
                  dumxz = f0*dum
                  !
                  ! (a|yz|b)
                  dum = 4.0_dp*(za**2*rr(0, coapyz, cob)+zb**2*rr(0, coa, cobpyz) &
                       &         +za*zb*(rr(0, coap1y, cobp1z)+rr(0, coap1z, cobp1y)))
                  IF (ay .GT. 0) dum = dum-2.0_dp*REAL(ay, dp)* &
                       &  (za*rr(0, coamypz, cob)+zb*rr(0, coam1y, cobp1z))
                  IF (az .GT. 0) dum = dum-2.0_dp*REAL(az, dp)* &
                       &  (za*rr(0, coamzpy, cob)+zb*rr(0, coam1z, cobp1y))
                  IF (ay .GT. 0 .AND. az .GT. 0) dum = dum+REAL(ay*az, dp)*rr(0, coamyz, cob)
                  IF (by .GT. 0) dum = dum-2.0_dp*REAL(by, dp)* &
                       &  (zb*rr(0, coa, cobmypz)+za*rr(0, coap1z, cobm1y))
                  IF (bz .GT. 0) dum = dum-2.0_dp*REAL(bz, dp)* &
                       &  (zb*rr(0, coa, cobmzpy)+za*rr(0, coap1y, cobm1z))
                  IF (by .GT. 0 .AND. bz .GT. 0) dum = dum+REAL(by*bz, dp)*rr(0, coa, cobmyz)
                  IF (ay .GT. 0 .AND. bz .GT. 0) dum = dum+REAL(ay*bz, dp)*rr(0, coam1y, cobm1z)
                  IF (az .GT. 0 .AND. by .GT. 0) dum = dum+REAL(az*by, dp)*rr(0, coam1z, cobm1y)
                  dumyz = f0*dum
                  !
                  !
                  vab(ma, mb, 1) = (2.0_dp*dumxx-dumyy-dumzz)/3.0_dp !xx
                  vab(ma, mb, 2) = dumxy !xy
                  vab(ma, mb, 3) = dumxz !xz
                  vab(ma, mb, 4) = (2.0_dp*dumyy-dumzz-dumxx)/3.0_dp !yy
                  vab(ma, mb, 5) = dumyz !yz
                  vab(ma, mb, 6) = (2.0_dp*dumzz-dumxx-dumyy)/3.0_dp !zz
               ENDDO
               ENDDO
               ENDDO !la

            ENDDO
            ENDDO
            ENDDO !lb

            nb = nb+ncoset(lb_max)

         ENDDO

         na = na+ncoset(la_max)

      ENDDO

   END SUBROUTINE efg

END MODULE ai_elec_field
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Distribution of the electric field gradient integral matrix.
!> \par History
!> \author VW (27.02.2009)
! **************************************************************************************************
MODULE qs_elec_field

   USE ai_elec_field,                   ONLY: efg
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_elec_field'

! *** Public subroutines ***

   PUBLIC :: build_efg_matrix

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the electric field gradient matrix over
!>          Cartesian Gaussian functions.
!> \param qs_env ...
!> \param matrix_efg ...
!> \param rc ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE build_efg_matrix(qs_env, matrix_efg, rc)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_efg
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rc

      CHARACTER(len=*), PARAMETER :: routineN = 'build_efg_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: after, handle, i, iatom, icol, ikind, inode, irow, iset, istat, iw, jatom, jkind, &
         jset, last_jatom, ldai, maxco, maxlgto, maxsgf, natom, ncoa, ncob, neighbor_list_id, &
         nkind, nseta, nsetb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b, omit_headers
      REAL(KIND=dp)                                      :: dab, rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: efgab, rr_work
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rb, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(block_p_type), ALLOCATABLE, DIMENSION(:)      :: efgint
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, sab_orb, qs_kind_set, particle_set, para_env)
      NULLIFY (logger)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      neighbor_list_id=neighbor_list_id, &
                      para_env=para_env, &
                      sab_orb=sab_orb, &
                      cell=cell)

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

      ! *** Allocate work storage ***
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxlgto=maxlgto, &
                           maxsgf=maxsgf)

      ldai = ncoset(maxlgto+2)
      CALL init_orbital_pointers(ldai)

      ALLOCATE (rr_work(0:2*maxlgto+4, ldai, ldai), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 140, &
                          "rr_work", ldai*ldai*(2*maxlgto+4)*dp_size)
      END IF

      ALLOCATE (efgab(maxco, maxco, 6), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 145, &
                                       "efgab", maxco*maxco*6*dp_size)

      ALLOCATE (work(maxco, maxsgf), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 149, &
                                       "work", maxco*maxsgf*dp_size)

      ALLOCATE (efgint(6), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 153, &
                                       "efgint", 6*int_size)

      rr_work(:, :, :) = 0.0_dp
      efgab(:, :, :) = 0.0_dp
      work(:, :) = 0.0_dp

      ALLOCATE (basis_set_list(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 161, &
                                       "basis_set_list", nkind)
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) last_jatom = 0

         rb = rab+ra
         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)
         rac = pbc(ra, rc, cell)
         rbc = rac-rab

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            DO i = 1, 6
               NULLIFY (efgint(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=matrix_efg(i)%matrix, &
                                         row=irow, col=icol, BLOCK=efgint(i)%block, found=found)
            ENDDO
         ENDIF

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               ! *** Calculate the primitive fermi contact integrals ***

               CALL efg(la_max(iset), la_min(iset), npgfa(iset), &
                        rpgfa(:, iset), zeta(:, iset), &
                        lb_max(jset), lb_min(jset), npgfb(jset), &
                        rpgfb(:, jset), zetb(:, jset), &
                        rac, rbc, rab, efgab, SIZE(rr_work, 1), SIZE(rr_work, 2), rr_work)

               ! *** Contraction step ***

               DO i = 1, 6

                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             1.0_dp, efgab(1, 1, i), SIZE(efgab, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))

                  IF (iatom <= jatom) THEN
                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                work(1, 1), SIZE(work, 1), &
                                1.0_dp, efgint(i)%block(sgfa, sgfb), &
                                SIZE(efgint(i)%block, 1))

                  ELSE

                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                1.0_dp, efgint(i)%block(sgfb, sgfa), &
                                SIZE(efgint(i)%block, 1))
                  ENDIF

               ENDDO

            ENDDO

         ENDDO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! *** Release work storage ***
      DEALLOCATE (basis_set_list)

      DEALLOCATE (efgab)

      DEALLOCATE (work)

      DEALLOCATE (efgint)

      ! Print the electric field gradient matrix, if requested

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/EFG"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/EFG", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
         after = MIN(MAX(after, 1), 16)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(1)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(2)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(3)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(4)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(5)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_efg(6)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/EFG")
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_efg_matrix

! **************************************************************************************************

END MODULE qs_elec_field

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the Fermi contact integrals over Cartesian
!>        Gaussian-type functions.
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!> \par Parameters
!>      - ax,ay,az  : Angular momentum index numbers of orbital a.
!>      - bx,by,bz  : Angular momentum index numbers of orbital b.
!>      - coset     : Cartesian orbital set pointer.
!>      - dab       : Distance between the atomic centers a and b.
!>      - l{a,b}    : Angular momentum quantum number of shell a or b.
!>      - l{a,b}_max: Maximum angular momentum quantum number of shell a or b.
!>      - l{a,b}_min: Minimum angular momentum quantum number of shell a or b.
!>      - rab       : Distance vector between the atomic centers a and b.
!>      - rpgf{a,b} : Radius of the primitive Gaussian-type function a or b.
!>      - sab       : Shell set of overlap integrals.
!>      - zet{a,b}  : Exponents of the Gaussian-type functions a or b.
!>      - zetp      : Reciprocal of the sum of the exponents of orbital a and b.
!> \author Matthias Krack (08.10.1999)
! **************************************************************************************************
MODULE ai_fermi_contact

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_fermi_contact'

! *** Public subroutines ***
   PUBLIC :: fermi_contact

CONTAINS

! **************************************************************************************************
!> \brief   Purpose: Calculation of the two-center Fermi contact integrals
!>          4/3*pi*[a|delta(r-c)|b] over Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param rac ...
!> \param rbc ...
!> \param dab ...
!> \param fcab ...
!> \param ldfc ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE fermi_contact(la_max, la_min, npgfa, rpgfa, zeta, &
                            lb_max, lb_min, npgfb, rpgfb, zetb, &
                            rac, rbc, dab, fcab, ldfc)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac, rbc
      REAL(KIND=dp)                                      :: dab
      INTEGER, INTENT(IN)                                :: ldfc
      REAL(KIND=dp), DIMENSION(ldfc, *), INTENT(INOUT)   :: fcab

      CHARACTER(len=*), PARAMETER :: routineN = 'fermi_contact', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, cob, i, &
                                                            ipgf, j, jpgf, la, lb, ma, mb, na, nb
      REAL(KIND=dp)                                      :: dac2, dbc2, f0, fax, fay, faz, fbx, fby, &
                                                            fbz

! *** Calculate some prefactors ***

      dac2 = rac(1)**2+rac(2)**2+rac(3)**2
      dbc2 = rbc(1)**2+rbc(2)**2+rbc(3)**2

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max)
                  DO i = na+1, na+ncoset(la_max)
                     fcab(i, j) = 0.0_dp
                  ENDDO
               ENDDO
               nb = nb+ncoset(lb_max)
               CYCLE
            ENDIF

            ! *** Calculate some prefactors ***

            f0 = 4.0_dp/3.0_dp*pi*EXP(-zeta(ipgf)*dac2-zetb(jpgf)*dbc2)

            ! *** Calculate the primitive Fermi contact integrals ***

            DO lb = lb_min, lb_max
            DO bx = 0, lb
               fbx = 1.0_dp
               IF (bx .GT. 0) fbx = (rbc(1))**bx
               DO by = 0, lb-bx
                  bz = lb-bx-by
                  cob = coset(bx, by, bz)
                  mb = nb+cob
                  fby = 1.0_dp
                  IF (by .GT. 0) fby = (rbc(2))**by
                  fbz = 1.0_dp
                  IF (bz .GT. 0) fbz = (rbc(3))**bz
                  DO la = la_min, la_max
                  DO ax = 0, la
                     fax = fbx
                     IF (ax .GT. 0) fax = fbx*(rac(1))**ax
                     DO ay = 0, la-ax
                        az = la-ax-ay
                        coa = coset(ax, ay, az)
                        ma = na+coa
                        fay = fby
                        IF (ay .GT. 0) fay = fby*(rac(2))**ay
                        faz = fbz
                        IF (az .GT. 0) faz = fbz*(rac(3))**az

                        fcab(ma, mb) = f0*fax*fay*faz

                     ENDDO
                  ENDDO
                  ENDDO !la

               ENDDO
            ENDDO
            ENDDO !lb

            nb = nb+ncoset(lb_max)

         ENDDO

         na = na+ncoset(la_max)

      ENDDO

   END SUBROUTINE fermi_contact

END MODULE ai_fermi_contact
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Distribution of the Fermi contact integral matrix.
!> \par History
!> \author VW (27.02.2009)
! **************************************************************************************************
MODULE qs_fermi_contact

   USE ai_fermi_contact,                ONLY: fermi_contact
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fermi_contact'

! *** Public subroutines ***

   PUBLIC :: build_fermi_contact_matrix

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the Fermi contact matrix over Cartesian
!>          Gaussian functions.
!> \param qs_env ...
!> \param matrix_fc ...
!> \param rc ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE build_fermi_contact_matrix(qs_env, matrix_fc, rc)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_fc
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rc

      CHARACTER(len=*), PARAMETER :: routineN = 'build_fermi_contact_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: after, handle, iatom, icol, ikind, inode, irow, iset, istat, iw, jatom, jkind, &
         jset, last_jatom, ldai, ldfc, maxco, maxlgto, maxsgf, natom, ncoa, ncob, nkind, nseta, &
         nsetb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b, omit_headers
      REAL(KIND=dp)                                      :: dab, rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: fcab, work
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rb, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(block_p_type), ALLOCATABLE, DIMENSION(:)      :: fcint
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, sab_orb, qs_kind_set, particle_set, para_env)
      NULLIFY (logger)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      para_env=para_env, &
                      sab_orb=sab_orb, &
                      cell=cell)

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

      ! *** Allocate work storage ***

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxlgto=maxlgto, &
                           maxsgf=maxsgf)

      ldai = ncoset(maxlgto)
      CALL init_orbital_pointers(ldai)

      ldfc = maxco
      ALLOCATE (fcab(ldfc, ldfc), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 139, &
                                       "fcab", ldfc*ldfc*dp_size)
      fcab(:, :) = 0.0_dp

      ALLOCATE (work(maxco, maxsgf), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 144, &
                                       "work", maxco*maxsgf*dp_size)
      work(:, :) = 0.0_dp

      ALLOCATE (fcint(1), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 149, &
                                       "fcint", 1*int_size)
      NULLIFY (fcint(1)%block)

      ALLOCATE (basis_set_list(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 154, &
                                       "basis_set_list", nkind)
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) last_jatom = 0

         rb = rab+ra
         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)
         rac = pbc(ra, rc, cell)
         rbc = rac-rab

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            NULLIFY (fcint(1)%block)
            CALL cp_dbcsr_get_block_p(matrix=matrix_fc(1)%matrix, &
                                      row=irow, col=icol, BLOCK=fcint(1)%block, found=found)
         ENDIF

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               ! *** Calculate the primitive fermi contact integrals ***

               CALL fermi_contact(la_max(iset), la_min(iset), npgfa(iset), &
                                  rpgfa(:, iset), zeta(:, iset), &
                                  lb_max(jset), lb_min(jset), npgfb(jset), &
                                  rpgfb(:, jset), zetb(:, jset), &
                                  rac, rbc, dab, fcab, SIZE(fcab, 1))

               ! *** Contraction step ***

               CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                          1.0_dp, fcab(1, 1), SIZE(fcab, 1), &
                          sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                          0.0_dp, work(1, 1), SIZE(work, 1))

               IF (iatom <= jatom) THEN

                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, fcint(1)%block(sgfa, sgfb), &
                             SIZE(fcint(1)%block, 1))

               ELSE

                  CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                             1.0_dp, work(1, 1), SIZE(work, 1), &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             1.0_dp, fcint(1)%block(sgfb, sgfa), &
                             SIZE(fcint(1)%block, 1))

               ENDIF

            ENDDO

         ENDDO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! *** Release work storage ***
      DEALLOCATE (basis_set_list)

      DEALLOCATE (fcab)

      DEALLOCATE (work)

      NULLIFY (fcint(1)%block)
      DEALLOCATE (fcint)

!   *** Print the Fermi contact matrix, if requested ***

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/FERMI_CONTACT"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/FERMI_CONTACT", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
         after = MIN(MAX(after, 1), 16)
         CALL cp_dbcsr_write_sparse_matrix(matrix_fc(1)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/FERMI_CONTACT")
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_fermi_contact_matrix

! **************************************************************************************************

END MODULE qs_fermi_contact

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      created 07.2005
!> \author MI (07.2005)
! **************************************************************************************************
MODULE qs_operators_ao
   USE ai_moments,                      ONLY: diff_momop,&
                                              diffop,&
                                              moment
   USE ai_os_rr,                        ONLY: os_rr_ovlp
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_get_matrix_type,&
                                              cp_dbcsr_p_type,&
                                              dbcsr_type_antisymmetric,&
                                              dbcsr_type_no_symmetry
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_operators_ao'

! *** Public subroutines ***

   PUBLIC :: p_xyz_ao, rRc_xyz_ao, rRc_xyz_der_ao
   PUBLIC :: build_lin_mom_matrix, build_ang_mom_matrix

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the linear momemtum matrix over
!>          Cartesian Gaussian functions.
!> \param qs_env ...
!> \param matrix ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE build_lin_mom_matrix(qs_env, matrix)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'build_lin_mom_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iatom, icol, ikind, inode, irow, iset, jatom, jkind, jset, last_jatom, &
         ldai, maxco, maxlgto, maxsgf, natom, ncoa, ncob, neighbor_list_id, nkind, nseta, nsetb, &
         sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b
      REAL(KIND=dp)                                      :: dab, rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: intab, rr_work
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(block_p_type), ALLOCATABLE, DIMENSION(:)      :: integral
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, sab_orb, qs_kind_set, particle_set, para_env)
      NULLIFY (logger)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      neighbor_list_id=neighbor_list_id, &
                      para_env=para_env, &
                      sab_orb=sab_orb, &
                      cell=cell)

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

!   *** Allocate work storage ***

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxlgto=maxlgto, &
                           maxsgf=maxsgf)

      ldai = ncoset(maxlgto+1)
      CALL init_orbital_pointers(ldai)

      ALLOCATE (rr_work(ldai, ldai, 3), intab(maxco, maxco, 3), work(maxco, maxsgf), integral(3))
      rr_work(:, :, :) = 0.0_dp
      intab(:, :, :) = 0.0_dp
      work(:, :) = 0.0_dp

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) last_jatom = 0
         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            DO i = 1, 3
               NULLIFY (integral(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=matrix(i)%matrix, &
                                         row=irow, col=icol, BLOCK=integral(i)%block, found=found)
               IF(.NOT.(ASSOCIATED(INTEGRAL(i)%block)))CALL cp__a("qs_operators_ao.F",199)
            ENDDO
         END IF

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               ! *** Calculate the primitive fermi contact integrals ***

               CALL lin_mom(la_max(iset), la_min(iset), npgfa(iset), &
                            rpgfa(:, iset), zeta(:, iset), &
                            lb_max(jset), lb_min(jset), npgfb(jset), &
                            rpgfb(:, jset), zetb(:, jset), &
                            rab, intab, SIZE(rr_work, 1), rr_work)

               ! *** Contraction step ***

               DO i = 1, 3

                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             1.0_dp, intab(1, 1, i), SIZE(intab, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))

                  IF (iatom <= jatom) THEN

                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                work(1, 1), SIZE(work, 1), &
                                1.0_dp, integral(i)%block(sgfa, sgfb), &
                                SIZE(integral(i)%block, 1))

                  ELSE

                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                -1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                1.0_dp, integral(i)%block(sgfb, sgfa), &
                                SIZE(integral(i)%block, 1))

                  ENDIF

               ENDDO

            ENDDO

         ENDDO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! *** Release work storage ***

      DEALLOCATE (intab, work, integral, basis_set_list)

!   *** Print the spin orbit matrix, if requested ***

      !IF (BTEST(cp_print_key_should_output(logger%iter_info,&
      !     qs_env%input,"DFT%PRINT%AO_MATRICES/LINEAR_MOMENTUM"),cp_p_file)) THEN
      !   iw = cp_print_key_unit_nr(logger,qs_env%input,"DFT%PRINT%AO_MATRICES/LINEA_MOMENTUM",&
      !        extension=".Log")
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(1)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(2)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(3)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_print_key_finished_output(iw,logger,qs_env%input,&
      !        "DFT%PRINT%AO_MATRICES/LINEAR_MOMENTUM")
      !END IF

      CALL timestop(handle)

   END SUBROUTINE build_lin_mom_matrix

! **************************************************************************************************
!> \brief   Calculation of the primitive paramagnetic spin orbit integrals over
!>          Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param rab ...
!> \param intab ...
!> \param ldrr ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE lin_mom(la_max, la_min, npgfa, rpgfa, zeta, lb_max, lb_min, npgfb, rpgfb, zetb, &
                      rab, intab, ldrr, rr)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: intab
      INTEGER, INTENT(IN)                                :: ldrr
      REAL(dp), DIMENSION(0:ldrr-1, 0:ldrr-1, 3), &
         INTENT(INOUT)                                   :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'lin_mom', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, cob, i, &
                                                            ipgf, j, jpgf, la, lb, ma, mb, na, nb
      REAL(dp)                                           :: dab, dumx, dumy, dumz, f0, rab2, xhi, zet
      REAL(dp), DIMENSION(3)                             :: rap, rbp

! *** Calculate the distance of the centers a and c ***

      rab2 = rab(1)**2+rab(2)**2+rab(3)**2
      dab = SQRT(rab2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max)
                  DO i = na+1, na+ncoset(la_max)
                     intab(i, j, 1) = 0.0_dp
                     intab(i, j, 2) = 0.0_dp
                     intab(i, j, 3) = 0.0_dp
                  ENDDO
               ENDDO
               nb = nb+ncoset(lb_max)
               CYCLE
            ENDIF

            ! *** Calculate some prefactors ***
            zet = zeta(ipgf)+zetb(jpgf)
            xhi = zeta(ipgf)*zetb(jpgf)/zet
            rap = zetb(jpgf)*rab/zet
            rbp = -zeta(ipgf)*rab/zet

            f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! *** Calculate the reccurence relation ***

            CALL os_rr_ovlp(rap, la_max+1, rbp, lb_max, zet, ldrr, rr)

            ! *** Calculate the primitive linear momentum integrals ***
            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)
               mb = nb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)
                  ma = na+coa
                  !
                  !
                  ! (a|p_x|b) = 2*a*(a+1x|b) - N_x(a)*(a-1_x|b)
                  dumx = 2.0_dp*zeta(ipgf)*rr(ax+1, bx, 1)
                  IF (ax .GT. 0) dumx = dumx-REAL(ax, dp)*rr(ax-1, bx, 1)
                  intab(ma, mb, 1) = f0*dumx*rr(ay, by, 2)*rr(az, bz, 3)
                  !
                  ! (a|p_y|b)
                  dumy = 2.0_dp*zeta(ipgf)*rr(ay+1, by, 2)
                  IF (ay .GT. 0) dumy = dumy-REAL(ay, dp)*rr(ay-1, by, 2)
                  intab(ma, mb, 2) = f0*rr(ax, bx, 1)*dumy*rr(az, bz, 3)
                  !
                  ! (a|p_z|b)
                  dumz = 2.0_dp*zeta(ipgf)*rr(az+1, bz, 3)
                  IF (az .GT. 0) dumz = dumz-REAL(az, dp)*rr(az-1, bz, 3)
                  intab(ma, mb, 3) = f0*rr(ax, bx, 1)*rr(ay, by, 2)*dumz
                  !
               ENDDO
               ENDDO
               ENDDO !la

            ENDDO
            ENDDO
            ENDDO !lb

            nb = nb+ncoset(lb_max)

         ENDDO

         na = na+ncoset(la_max)

      ENDDO

   END SUBROUTINE lin_mom

! **************************************************************************************************
!> \brief   Calculation of the angular momemtum matrix over
!>          Cartesian Gaussian functions.
!> \param qs_env ...
!> \param matrix ...
!> \param rc ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE build_ang_mom_matrix(qs_env, matrix, rc)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: rc

      CHARACTER(len=*), PARAMETER :: routineN = 'build_ang_mom_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iatom, icol, ikind, inode, irow, iset, jatom, jkind, jset, last_jatom, &
         ldai, maxco, maxlgto, maxsgf, natom, ncoa, ncob, neighbor_list_id, nkind, nseta, nsetb, &
         sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b
      REAL(KIND=dp)                                      :: dab, rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: intab, rr_work
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(block_p_type), ALLOCATABLE, DIMENSION(:)      :: integral
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, sab_all, qs_kind_set, particle_set, para_env)
      NULLIFY (logger)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      neighbor_list_id=neighbor_list_id, &
                      para_env=para_env, &
                      sab_all=sab_all, &
                      cell=cell)

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

!   *** Allocate work storage ***

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxlgto=maxlgto, &
                           maxsgf=maxsgf)

      ldai = ncoset(maxlgto+1)
      CALL init_orbital_pointers(ldai)

      ALLOCATE (rr_work(ldai, ldai, 3), intab(maxco, maxco, 3), work(maxco, maxsgf), integral(3))
      rr_work(:, :, :) = 0.0_dp
      intab(:, :, :) = 0.0_dp
      work(:, :) = 0.0_dp

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_all)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) last_jatom = 0

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            !IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
            !ELSE
            !   irow = jatom
            !   icol = iatom
            !END IF

            DO i = 1, 3
               NULLIFY (INTEGRAL(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=matrix(i)%matrix, &
                                         row=irow, col=icol, BLOCK=INTEGRAL(i)%block, found=found)
               IF(.NOT.(ASSOCIATED(INTEGRAL(i)%block)))CALL cp__a("qs_operators_ao.F",556)
            ENDDO
         END IF

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

               !IF(PRESENT(wancen)) THEN
               !   rc = wancen
               rac = pbc(rc, ra, cell)
               rbc = rac+rab
               !ELSE
               !   rc(1:3) = rb(1:3)
               !   rac(1:3) = -rab(1:3)
               !   rbc(1:3) = 0.0_dp
               !ENDIF

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               ! *** Calculate the primitive angular momentum integrals ***

               CALL ang_mom(la_max(iset), la_min(iset), npgfa(iset), &
                            rpgfa(:, iset), zeta(:, iset), &
                            lb_max(jset), lb_min(jset), npgfb(jset), &
                            rpgfb(:, jset), zetb(:, jset), &
                            rab, rac, intab, SIZE(rr_work, 1), rr_work)

               ! *** Contraction step ***

               DO i = 1, 3

                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             1.0_dp, intab(1, 1, i), SIZE(intab, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))

                  !IF (iatom <= jatom) THEN

                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, integral(i)%block(sgfa, sgfb), &
                             SIZE(integral(i)%block, 1))

                  !ELSE
                  !
                  !   CALL dgemm("T","N",nsgfb(jset),nsgfa(iset),ncoa,&
                  !              -1.0_dp,work(1,1),SIZE(work,1),&
                  !              sphi_a(1,sgfa),SIZE(sphi_a,1),&
                  !              1.0_dp,integral(i)%block(sgfb,sgfa),&
                  !              SIZE(integral(i)%block,1))
                  !
                  !ENDIF

               ENDDO

            ENDDO

         ENDDO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! *** Release work storage ***

      DEALLOCATE (intab, work, integral, basis_set_list)

!   *** Print the spin orbit matrix, if requested ***

      !IF (BTEST(cp_print_key_should_output(logger%iter_info,&
      !     qs_env%input,"DFT%PRINT%AO_MATRICES/ANGULAR_MOMENTUM"),cp_p_file)) THEN
      !   iw = cp_print_key_unit_nr(logger,qs_env%input,"DFT%PRINT%AO_MATRICES/ANGULAR_MOMENTUM",&
      !        extension=".Log")
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(1)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(2)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_dbcsr_write_sparse_matrix(matrix(3)%matrix,4,6,qs_env,para_env,output_unit=iw)
      !   CALL cp_print_key_finished_output(iw,logger,qs_env%input,&
      !        "DFT%PRINT%AO_MATRICES/ANGULAR_MOMENTUM")
      !END IF

      CALL timestop(handle)

   END SUBROUTINE build_ang_mom_matrix

! **************************************************************************************************
!> \brief   Calculation of the primitive paramagnetic spin orbit integrals over
!>          Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param rab ...
!> \param rac ...
!> \param intab ...
!> \param ldrr ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE ang_mom(la_max, la_min, npgfa, rpgfa, zeta, lb_max, lb_min, npgfb, rpgfb, zetb, &
                      rab, rac, intab, ldrr, rr)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab, rac
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: intab
      INTEGER, INTENT(IN)                                :: ldrr
      REAL(dp), DIMENSION(0:ldrr-1, 0:ldrr-1, 3), &
         INTENT(INOUT)                                   :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'ang_mom', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, cob, i, &
                                                            ipgf, j, jpgf, la, lb, ma, mb, na, nb
      REAL(dp)                                           :: dab, dumx, dumy, dumz, f0, rab2, xhi, zet
      REAL(dp), DIMENSION(3)                             :: rap, rbp

! *** Calculate the distance of the centers a and c ***

      rab2 = rab(1)**2+rab(2)**2+rab(3)**2
      dab = SQRT(rab2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max)
                  DO i = na+1, na+ncoset(la_max)
                     intab(i, j, 1) = 0.0_dp
                     intab(i, j, 2) = 0.0_dp
                     intab(i, j, 3) = 0.0_dp
                  ENDDO
               ENDDO
               nb = nb+ncoset(lb_max)
               CYCLE
            ENDIF

            ! *** Calculate some prefactors ***
            zet = zeta(ipgf)+zetb(jpgf)
            xhi = zeta(ipgf)*zetb(jpgf)/zet
            rap = zetb(jpgf)*rab/zet
            rbp = -zeta(ipgf)*rab/zet

            f0 = (pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! *** Calculate the reccurence relation ***

            CALL os_rr_ovlp(rap, la_max+1, rbp, lb_max, zet, ldrr, rr)

            ! *** Calculate the primitive Fermi contact integrals ***

            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)
               mb = nb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)
                  ma = na+coa
                  !
                  dumx = -2.0_dp*zeta(ipgf)*rr(ax+1, bx, 1)
                  dumy = -2.0_dp*zeta(ipgf)*rr(ay+1, by, 2)
                  dumz = -2.0_dp*zeta(ipgf)*rr(az+1, bz, 3)
                  IF (ax .GT. 0) dumx = dumx+REAL(ax, dp)*rr(ax-1, bx, 1)
                  IF (ay .GT. 0) dumy = dumy+REAL(ay, dp)*rr(ay-1, by, 2)
                  IF (az .GT. 0) dumz = dumz+REAL(az, dp)*rr(az-1, bz, 3)
                  !
                  ! (a|l_z|b)
                  intab(ma, mb, 1) = -f0*rr(ax, bx, 1)*( &
                       &  (rr(ay+1, by, 2)+rac(2)*rr(ay, by, 2))*dumz &
                       & -(rr(az+1, bz, 3)+rac(3)*rr(az, bz, 3))*dumy)
                  !
                  ! (a|l_y|b)
                  intab(ma, mb, 2) = -f0*rr(ay, by, 2)*( &
                       &  (rr(az+1, bz, 3)+rac(3)*rr(az, bz, 3))*dumx &
                       & -(rr(ax+1, bx, 1)+rac(1)*rr(ax, bx, 1))*dumz)
                  !
                  ! (a|l_z|b)
                  intab(ma, mb, 3) = -f0*rr(az, bz, 3)*( &
                       &  (rr(ax+1, bx, 1)+rac(1)*rr(ax, bx, 1))*dumy &
                       & -(rr(ay+1, by, 2)+rac(2)*rr(ay, by, 2))*dumx)
                  !
               ENDDO
               ENDDO
               ENDDO !la

            ENDDO
            ENDDO
            ENDDO !lb

            nb = nb+ncoset(lb_max)

         ENDDO

         na = na+ncoset(la_max)

      ENDDO

   END SUBROUTINE ang_mom

! **************************************************************************************************
!> \brief Calculation of the components of the dipole operator in the velocity form
!>      The elements of the  sparse matrices are the integrals in the
!>      basis functions
!> \param op matrix representation of the p operator
!>               calculated in terms of the contracted basis functions
!> \param qs_env enviroment for the lists and the basis sets
!> \param minimum_image take into account only the first neighbors in the lists
!> \par History
!>      06.2005 created [MI]
!> \author MI
! **************************************************************************************************

   SUBROUTINE p_xyz_ao(op, qs_env, minimum_image)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: minimum_image

      CHARACTER(len=*), PARAMETER :: routineN = 'p_xyz_ao', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iatom, icol, ikind, inode, irow, iset, jatom, jkind, jset, last_jatom, &
         ldab, ldsa, ldsb, ldwork, maxl, ncoa, ncob, nkind, nseta, nsetb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, my_minimum_image, new_atom_b
      REAL(KIND=dp)                                      :: alpha, dab, Lxo2, Lyo2, Lzo2, rab2
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, work, &
                                                            zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: difab
      TYPE(block_p_type), DIMENSION(:), POINTER          :: op_dip
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind, qs_kind_set)
      NULLIFY (cell, particle_set)
      NULLIFY (sab_orb)
      NULLIFY (difab, op_dip, work)
      NULLIFY (la_max, la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb)
      NULLIFY (set_radius_a, set_radius_b, rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      cell=cell, particle_set=particle_set, &
                      sab_orb=sab_orb)

      nkind = SIZE(qs_kind_set)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=ldwork, maxlgto=maxl)

      my_minimum_image = .FALSE.
      IF (PRESENT(minimum_image)) THEN
         my_minimum_image = minimum_image
         Lxo2 = SQRT(SUM(cell%hmat(:, 1)**2))/2.0_dp
         Lyo2 = SQRT(SUM(cell%hmat(:, 2)**2))/2.0_dp
         Lzo2 = SQRT(SUM(cell%hmat(:, 3)**2))/2.0_dp
      END IF

      ldab = ldwork

      ALLOCATE (difab(ldab, ldab, 3))
      difab(1:ldab, 1:ldab, 1:3) = 0.0_dp
      ALLOCATE (work(ldwork, ldwork))
      work(1:ldwork, 1:ldwork) = 0.0_dp
      ALLOCATE (op_dip(3))

      DO i = 1, 3
         NULLIFY (op_dip(i)%block)
      END DO

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) THEN
            last_jatom = 0
            alpha = 1.0_dp
         END IF
         ldsa = SIZE(sphi_a, 1)
         ldsb = SIZE(sphi_b, 1)

         IF (my_minimum_image) THEN
            IF (ABS(rab(1)) > Lxo2 .OR. ABS(rab(2)) > Lyo2 .OR. ABS(rab(3)) > Lzo2) CYCLE
         END IF

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
               alpha = 1.0_dp
            ELSE
               irow = jatom
               icol = iatom
               IF (cp_dbcsr_get_matrix_type(op(1)%matrix) == dbcsr_type_antisymmetric) THEN
                  !IF(op(1)%matrix%symmetry=="antisymmetric") THEN
                  alpha = -1.0_dp
               END IF
            END IF

            DO i = 1, 3
               NULLIFY (op_dip(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=op(i)%matrix, &
                                         row=irow, col=icol, block=op_dip(i)%block, found=found)
               IF(.NOT.(ASSOCIATED(op_dip(i)%block)))CALL cp__a("qs_operators_ao.F",948)
            END DO
         END IF ! new_atom_b
         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (set_radius_a(iset)+set_radius_b(jset) >= dab) THEN

!            *** Calculate the primitive overlap integrals ***
                  CALL diffop(la_max(iset), npgfa(iset), zeta(:, iset), &
                              rpgfa(:, iset), la_min(iset), lb_max(jset), npgfb(jset), &
                              zetb(:, jset), rpgfb(:, jset), lb_min(jset), rab, difab)

!            *** Contraction ***
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             alpha, difab(1, 1, 1), ldab, sphi_b(1, sgfb), ldsb, &
                             0.0_dp, work(1, 1), ldwork)
                  IF (iatom <= jatom) THEN
                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), ldsa, &
                                work(1, 1), ldwork, &
                                1.0_dp, op_dip(1)%block(sgfa, sgfb), &
                                SIZE(op_dip(1)%block, 1))

                  ELSE
                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                1.0_dp, work(1, 1), ldwork, &
                                sphi_a(1, sgfa), ldsa, &
                                1.0_dp, op_dip(1)%block(sgfb, sgfa), &
                                SIZE(op_dip(1)%block, 1))

                  END IF

!             *** Contraction ***
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             alpha, difab(1, 1, 2), ldab, sphi_b(1, sgfb), ldsb, &
                             0.0_dp, work(1, 1), ldwork)
                  IF (iatom <= jatom) THEN
                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), ldsa, &
                                work(1, 1), ldwork, &
                                1.0_dp, op_dip(2)%block(sgfa, sgfb), &
                                SIZE(op_dip(2)%block, 1))
                  ELSE
                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                1.0_dp, work(1, 1), ldwork, &
                                sphi_a(1, sgfa), ldsa, &
                                1.0_dp, op_dip(2)%block(sgfb, sgfa), &
                                SIZE(op_dip(2)%block, 1))
                  END IF

!            *** Contraction ***
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             alpha, difab(1, 1, 3), ldab, sphi_b(1, sgfb), ldsb, &
                             0.0_dp, work(1, 1), ldwork)
                  IF (iatom <= jatom) THEN
                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), ldsa, &
                                work(1, 1), ldwork, &
                                1.0_dp, op_dip(3)%block(sgfa, sgfb), &
                                SIZE(op_dip(3)%block, 1))
                  ELSE
                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                1.0_dp, work(1, 1), ldwork, &
                                sphi_a(1, sgfa), ldsa, &
                                1.0_dp, op_dip(3)%block(sgfb, sgfa), &
                                SIZE(op_dip(3)%block, 1))
                  END IF
               END IF !  >= dab

            END DO ! jset

         END DO ! iset

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DO i = 1, 3
         NULLIFY (op_dip(i)%block)
      END DO
      DEALLOCATE (op_dip)

      DEALLOCATE (difab, work, basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE p_xyz_ao

! **************************************************************************************************
!> \brief Calculation of the components of the dipole operator in the length form
!>      by taking the relative position operator r-Rc, with respect a reference point Rc
!>      Probably it does not work for PBC, or maybe yes if the wfn are
!>      sufficiently localized
!>      The elements of the  sparse matrices are the integrals in the
!>      basis functions
!> \param op matrix representation of the p operator
!>               calculated in terms of the contracted basis functions
!> \param qs_env enviroment for the lists and the basis sets
!> \param rc reference vector position
!> \param order maximum order of the momentum, for the doipole order = 1
!> \param minimum_image take into account only the first neighbors in the lists
!> \param soft ...
!> \par History
!>      03.2006 created [MI]
!> \author MI
! **************************************************************************************************

   SUBROUTINE rRc_xyz_ao(op, qs_env, rc, order, minimum_image, soft)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp)                                           :: Rc(3)
      INTEGER, INTENT(IN)                                :: order
      LOGICAL, INTENT(IN), OPTIONAL                      :: minimum_image, soft

      CHARACTER(len=*), PARAMETER :: routineN = 'rRc_xyz_ao', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iatom, icol, ikind, imom, inode, irow, iset, jatom, jkind, jset, &
         last_jatom, ldab, ldsa, ldsb, ldwork, M_dim, maxl, ncoa, ncob, nkind, nrow, nseta, nsetb, &
         sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, npgfa, npgfb, &
                                                            nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, my_minimum_image, my_soft, &
                                                            new_atom_b
      REAL(KIND=dp)                                      :: dab, Lxo2, Lyo2, Lzo2, rab2
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rb, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, work, &
                                                            zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: mab
      TYPE(block_p_type), DIMENSION(:), POINTER          :: op_dip
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind, qs_kind_set)
      NULLIFY (cell, particle_set)
      NULLIFY (sab_orb)
      NULLIFY (mab, op_dip, work)
      NULLIFY (la_max, la_min, lb_max, npgfa, npgfb, nsgfa, nsgfb)
      NULLIFY (set_radius_a, set_radius_b, rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      cell=cell, particle_set=particle_set, sab_orb=sab_orb)

      nkind = SIZE(qs_kind_set)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=ldwork, maxlgto=maxl)

      my_minimum_image = .FALSE.
      IF (PRESENT(minimum_image)) THEN
         my_minimum_image = minimum_image
         Lxo2 = SQRT(SUM(cell%hmat(:, 1)**2))/2.0_dp
         Lyo2 = SQRT(SUM(cell%hmat(:, 2)**2))/2.0_dp
         Lzo2 = SQRT(SUM(cell%hmat(:, 3)**2))/2.0_dp
      END IF
      my_soft = .FALSE.
      IF (PRESENT(soft)) THEN
         my_soft = soft
      END IF

      ldab = ldwork

      M_dim = ncoset(order)-1
      IF(.NOT.(M_dim <= SIZE(op, 1)))CALL cp__a("qs_operators_ao.F",1133)

      ALLOCATE (mab(ldab, ldab, M_dim))
      mab(1:ldab, 1:ldab, 1:M_dim) = 0.0_dp
      ALLOCATE (work(ldwork, ldwork))
      work(1:ldwork, 1:ldwork) = 0.0_dp
      ALLOCATE (op_dip(M_dim))

      DO imom = 1, M_dim
         NULLIFY (op_dip(imom)%block)
      END DO

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, softb=my_soft, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         ldsa = SIZE(sphi_a, 1)
         ldsb = SIZE(sphi_b, 1)
         IF (inode == 1) last_jatom = 0

         IF (my_minimum_image) THEN
            IF (ABS(rab(1)) > Lxo2 .OR. ABS(rab(2)) > Lyo2 .OR. ABS(rab(3)) > Lzo2) CYCLE
         END IF

         rb = rab+ra

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            DO imom = 1, M_dim
               NULLIFY (op_dip(imom)%block)
               CALL cp_dbcsr_get_block_p(matrix=op(imom)%matrix, &
                                         row=irow, col=icol, block=op_dip(imom)%block, found=found)
               IF(.NOT.(ASSOCIATED(op_dip(imom)%block)))CALL cp__a("qs_operators_ao.F",1216)
            END DO ! imom
         END IF ! new_atom_b

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         nrow = 0
         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (set_radius_a(iset)+set_radius_b(jset) >= dab) THEN

                  rac = pbc(rc, ra, cell)
                  rbc = pbc(rc, rb, cell)

!            *** Calculate the primitive overlap integrals ***
                  CALL moment(la_max(iset), npgfa(iset), zeta(:, iset), &
                              rpgfa(:, iset), la_min(iset), &
                              lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), &
                              order, rac, rbc, mab)

                  DO imom = 1, M_dim
!                 *** Contraction ***
                     CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                                1.0_dp, mab(1, 1, imom), ldab, sphi_b(1, sgfb), ldsb, &
                                0.0_dp, work(1, 1), ldwork)
                     IF (iatom <= jatom) THEN
                        CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                   1.0_dp, sphi_a(1, sgfa), ldsa, &
                                   work(1, 1), ldwork, &
                                   1.0_dp, op_dip(imom)%block(sgfa, sgfb), &
                                   SIZE(op_dip(imom)%block, 1))
                     ELSE
                        CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                   1.0_dp, work(1, 1), ldwork, &
                                   sphi_a(1, sgfa), ldsa, &
                                   1.0_dp, op_dip(imom)%block(sgfb, sgfa), &
                                   SIZE(op_dip(imom)%block, 1))
                     END IF

                  END DO ! imom
               END IF !  >= dab

            END DO ! jset

            nrow = nrow+ncoa

         END DO ! iset

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DO i = 1, 3
         NULLIFY (op_dip(i)%block)
      END DO
      DEALLOCATE (op_dip)

      DEALLOCATE (mab, work, basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE rRc_xyz_ao

! **************************************************************************************************
!> \brief Calculation of the  multipole operators integrals
!>      and of its derivatives of the type
!>      [\mu | op | d(\nu)/dR(\nu)]-[d(\mu)/dR(\mu)| op | \nu]
!>      by taking the relative position operator r-Rc, with respect a reference point Rc
!>      The derivative are with respect to the primitive position,
!>      The multipole operator is symmetric and if it does not depend on R(\mu) or R(\nu)
!>      therefore  [\mu | op | d(\nu)/dR(\nu)] = -[d(\mu)/dR(\mu)| op | \nu]
!>        [\mu|op|d(\nu)/dR]-[d(\mu)/dR|op|\nu]=2[\mu|op|d(\nu)/dR]
!>      When it is not the case a correction term is needed
!>
!>     The momentum operator [\mu|M|\nu] is symmetric, the number of components is
!>     detrermined by the order: 3 for order 1 (x,y,x), 9 for order 2(xx,xy,xz,yy,yz,zz)
!>     The derivative of the type [\mu | op | d(\nu)/dR_i(\nu)], where
!>     i indicates the cartesian direction, is antisymmetric only when
!>     the no component M =(r_i) or (r_i r_j) is in the same cartesian
!>     direction of the derivative,  indeed
!>   d([\mu|M|\nu])/dr_i = [d(\mu)/dr_i|M|\nu] + [\mu|M|d(\nu)/dr_i] + [\mu |d(M)/dr_i|\nu]
!>   d([\mu|M|\nu])/dr_i = -[d(\mu)/dR_i(\mu)|M|\nu] -[\mu|M|d(\nu)/dR_i(\nu)] + [\mu |d(M)/dr_i|\nu]
!>     Therefore we cannot use an antisymmetric matrix
!>
!>     The same holds for the derivative with respect to the electronic position r
!>     taking into account that [\mu|op|d(\nu)/dR] = -[\mu|op|d(\nu)/dr]
!> \param op matrix representation of the p operator
!>               calculated in terms of the contracted basis functions
!> \param op_der ...
!> \param qs_env enviroment for the lists and the basis sets
!> \param rc reference vector position
!> \param order maximum order of the momentum, for the dipole order = 1
!> \param minimum_image take into account only the first neighbors in the lists
!> \param soft ...
!> \par History
!>      03.2006 created [MI]
!> \author MI
!> \note
!>      Probably it does not work for PBC, or maybe yes if the wfn are
!>      sufficiently localized
!>      The elements of the  sparse matrices are the integrals in the
!>      basis functions
! **************************************************************************************************
   SUBROUTINE rRc_xyz_der_ao(op, op_der, qs_env, rc, order, minimum_image, soft)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_der
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(dp)                                           :: Rc(3)
      INTEGER, INTENT(IN)                                :: order
      LOGICAL, INTENT(IN), OPTIONAL                      :: minimum_image, soft

      CHARACTER(len=*), PARAMETER :: routineN = 'rRc_xyz_der_ao', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iatom, icol, idir, ikind, imom, inode, ipgf, irow, iset, j, jatom, &
         jkind, jpgf, jset, last_jatom, lda_min, ldab, ldb_min, ldsa, ldsb, ldwork, M_dim, maxl, &
         na, nb, ncoa, ncob, nda, ndb, nkind, nseta, nsetb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: my_minimum_image, my_soft, new_atom_b, &
                                                            op_der_found, op_found
      REAL(KIND=dp)                                      :: alpha, alpha_der, dab, Lxo2, Lyo2, Lzo2, &
                                                            rab2
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rb, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, work, &
                                                            zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: mab, mab_tmp
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: difmab
      TYPE(block_p_type), DIMENSION(:), POINTER          :: op_dip
      TYPE(block_p_type), DIMENSION(:, :), POINTER       :: op_dip_der
      TYPE(cell_type), POINTER                           :: cell
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(op)))CALL cp__a("qs_operators_ao.F",1369)
      IF(.NOT.(ASSOCIATED(op_der)))CALL cp__a("qs_operators_ao.F",1370)
      !IF(.NOT.op_sm_der(1,1)%matrix%symmetry=="none") THEN
      IF (.NOT. cp_dbcsr_get_matrix_type(op_der(1, 1)%matrix) .EQ. dbcsr_type_no_symmetry) THEN
         CALL cp__b("qs_operators_ao.F",1373,"")
      END IF

      NULLIFY (qs_kind, qs_kind_set)
      NULLIFY (cell, particle_set)
      NULLIFY (sab_all)
      NULLIFY (difmab, mab, mab_tmp)
      NULLIFY (op_dip, op_dip_der, work)
      NULLIFY (la_max, la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb)
      NULLIFY (set_radius_a, set_radius_b, rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb)

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      cell=cell, particle_set=particle_set, &
                      sab_all=sab_all)

      nkind = SIZE(qs_kind_set)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=ldwork, maxlgto=maxl)

      my_minimum_image = .FALSE.
      IF (PRESENT(minimum_image)) THEN
         my_minimum_image = minimum_image
         Lxo2 = SQRT(SUM(cell%hmat(:, 1)**2))/2.0_dp
         Lyo2 = SQRT(SUM(cell%hmat(:, 2)**2))/2.0_dp
         Lzo2 = SQRT(SUM(cell%hmat(:, 3)**2))/2.0_dp
      END IF
      my_soft = .FALSE.
      IF (PRESENT(soft)) THEN
         my_soft = soft
      END IF

      ldab = ldwork

      M_dim = ncoset(order)-1
      IF(.NOT.(M_dim <= SIZE(op, 1)))CALL cp__a("qs_operators_ao.F",1408)

      ALLOCATE (mab(ldab, ldab, M_dim))
      mab(1:ldab, 1:ldab, 1:M_dim) = 0.0_dp
      ALLOCATE (difmab(ldab, ldab, M_dim, 3))
      difmab(1:ldab, 1:ldab, 1:M_dim, 1:3) = 0.0_dp

      ALLOCATE (work(ldwork, ldwork))
      work(1:ldwork, 1:ldwork) = 0.0_dp
      ALLOCATE (op_dip(M_dim))
      ALLOCATE (op_dip_der(M_dim, 3))

      DO imom = 1, M_dim
         NULLIFY (op_dip(imom)%block)
         DO i = 1, 3
            NULLIFY (op_dip_der(imom, i)%block)
         END DO
      END DO

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, softb=my_soft, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_all)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         ldsa = SIZE(sphi_a, 1)
         IF (ldsa .EQ. 0) CYCLE
         ldsb = SIZE(sphi_b, 1)
         IF (ldsb .EQ. 0) CYCLE
         IF (inode == 1) last_jatom = 0

         IF (my_minimum_image) THEN
            IF (ABS(rab(1)) > Lxo2 .OR. ABS(rab(2)) > Lyo2 .OR. ABS(rab(3)) > Lzo2) CYCLE
         END IF

         rb = rab+ra

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            irow = iatom
            icol = jatom
            alpha_der = 2.0_dp

            DO imom = 1, M_dim
               NULLIFY (op_dip(imom)%block)
               CALL cp_dbcsr_get_block_p(matrix=op(imom)%matrix, &
                                         row=irow, col=icol, &
                                         block=op_dip(imom)%block, &
                                         found=op_found)
               IF(.NOT.(op_found .AND. ASSOCIATED(op_dip(imom)%block)))CALL cp__a("qs_operators_ao.F",1499)
               DO idir = 1, 3
                  NULLIFY (op_dip_der(imom, idir)%block)
                  CALL cp_dbcsr_get_block_p(matrix=op_der(imom, idir)%matrix, &
                                            row=irow, col=icol, &
                                            block=op_dip_der(imom, idir)%block, &
                                            found=op_der_found)
                  IF(.NOT.(op_der_found .AND. ASSOCIATED(op_dip_der(imom, idir)%block)))CALL cp__a("qs_operators_ao.F",1506)
               END DO ! idir
            END DO ! imom
         END IF ! new_atom_b

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               IF (set_radius_a(iset)+set_radius_b(jset) >= dab) THEN

                  rac = pbc(rc, ra, cell)
                  rbc = rac+rab
!                  rac = pbc(rc,ra,cell)
!                  rbc = pbc(rc,rb,cell)

                  ALLOCATE (mab_tmp(npgfa(iset)*ncoset(la_max(iset)+1), &
                                    npgfb(jset)*ncoset(lb_max(jset)+1), ncoset(order)-1))

                  lda_min = MAX(0, la_min(iset)-1)
                  ldb_min = MAX(0, lb_min(jset)-1)
!            *** Calculate the primitive overlap integrals ***
                  CALL moment(la_max(iset)+1, npgfa(iset), zeta(:, iset), &
                              rpgfa(:, iset), lda_min, &
                              lb_max(jset)+1, npgfb(jset), zetb(:, jset), rpgfb(:, jset), &
                              order, rac, rbc, mab_tmp)

!            *** Calculate the derivatives
                  CALL diff_momop(la_max(iset), npgfa(iset), zeta(:, iset), &
                                  rpgfa(:, iset), la_min(iset), lb_max(jset), npgfb(jset), &
                                  zetb(:, jset), rpgfb(:, jset), lb_min(jset), order, rac, rbc, &
                                  difmab, mab_ext=mab_tmp)

! Contract and copy in the sparse matrix
                  mab = 0.0_dp
                  DO imom = 1, M_dim
                     na = 0
                     nda = 0
                     DO ipgf = 1, npgfa(iset)
                        nb = 0
                        ndb = 0
                        DO jpgf = 1, npgfb(jset)
                           DO j = 1, ncoset(lb_max(jset))
                              DO i = 1, ncoset(la_max(iset))
                                 mab(i+na, j+nb, imom) = mab_tmp(i+nda, j+ndb, imom)
                              END DO ! i
                           END DO ! j
                           nb = nb+ncoset(lb_max(jset))
                           ndb = ndb+ncoset(lb_max(jset)+1)
                        END DO ! jpgf
                        na = na+ncoset(la_max(iset))
                        nda = nda+ncoset(la_max(iset)+1)
                     END DO ! ipgf

!                 *** Contraction ***
                     CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                                1.0_dp, mab(1, 1, imom), ldab, sphi_b(1, sgfb), ldsb, &
                                0.0_dp, work(1, 1), ldwork)
                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), ldsa, &
                                work(1, 1), ldwork, &
                                1.0_dp, op_dip(imom)%block(sgfa, sgfb), &
                                SIZE(op_dip(imom)%block, 1))

                     alpha = -1.0_dp !-alpha_der
                     DO idir = 1, 3
                        CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                                   alpha, difmab(1, 1, imom, idir), ldab, sphi_b(1, sgfb), ldsb, &
                                   0.0_dp, work(1, 1), ldwork)
                        CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                   1.0_dp, sphi_a(1, sgfa), ldsa, &
                                   work(1, 1), ldwork, &
                                   1.0_dp, op_dip_der(imom, idir)%block(sgfa, sgfb), &
                                   SIZE(op_dip_der(imom, idir)%block, 1))

                     END DO ! idir

                  END DO ! imom

                  DEALLOCATE (mab_tmp)
               END IF !  >= dab

            END DO ! jset

         END DO ! iset

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DO i = 1, 3
         NULLIFY (op_dip(i)%block)
      END DO
      DEALLOCATE (op_dip, op_dip_der)

      DEALLOCATE (mab, difmab, work, basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE rRc_xyz_der_ao

END MODULE qs_operators_ao
