!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The methods which allow to analyze and manipulate the arnoldi procedure
!>        The main routine and this should eb the only public access point for the
!>        method
!> \par History
!>       2014.09 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_arnoldi_data_methods
   USE dbcsr_arnoldi_types,             ONLY: &
        arnoldi_control, arnoldi_data_c, arnoldi_data_d, arnoldi_data_s, arnoldi_data_z, &
        dbcsr_arnoldi_data, get_control, get_data_c, get_data_d, get_data_s, get_data_z, &
        get_evals_c, get_evals_d, get_evals_s, get_evals_z, get_sel_ind, has_d_cmplx, has_d_real, &
        has_s_cmplx, has_s_real, set_control, set_data_c, set_data_d, set_data_s, set_data_z
   USE dbcsr_data_methods,              ONLY: dbcsr_get_data_p
   USE dbcsr_methods,                   ONLY: dbcsr_get_matrix_type,&
                                              dbcsr_is_initialized,&
                                              dbcsr_release
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_grid_setup
   USE dbcsr_operations,                ONLY: dbcsr_get_info
   USE dbcsr_toollib,                   ONLY: sort
   USE dbcsr_types,                     ONLY: dbcsr_distribution_obj,&
                                              dbcsr_obj,&
                                              dbcsr_obj_type_p,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8,&
                                              dbcsr_type_symmetric
   USE dbcsr_vector_operations,         ONLY: create_col_vec_from_matrix
   USE kinds,                           ONLY: real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_arnoldi_data_methods'

   PUBLIC :: select_evals, get_selected_ritz_val, arnoldi_is_converged, &
             dbcsr_arnoldi_data, get_nrestart, set_arnoldi_initial_vector, &
             setup_arnoldi_data, deallocate_arnoldi_data, get_selected_ritz_vector

CONTAINS

! **************************************************************************************************
!> \brief This routine sets the environment for the arnoldi iteration and
!>        the krylov subspace creation. All simulation parameters have to be given
!>        at this stage so the rest can run fully automated
!>        In addition, this routine allocates the data necessary for
!> \param arnoldi_data this type which gets filled with information and
!>        on output contains all information necessary to extract
!>        whatever the user desires
!> \param matrix vector of matrices, only the first gets used to get some dimensions
!>        and parallel information needed later on
!> \param max_iter maximum dimension of the krylov subspace
!> \param threshold convergence threshold, this is used for both subspace and eigenval
!> \param selection_crit integer defining according to which criterion the
!>        eigenvalues are selected for the subspace
!> \param nval_request for some sel_crit useful, how many eV to select
!> \param nrestarts ...
!> \param generalized_ev ...
!> \param iram ...
! **************************************************************************************************
   SUBROUTINE setup_arnoldi_data(arnoldi_data, matrix, max_iter, threshold, selection_crit, &
                                 nval_request, nrestarts, generalized_ev, iram)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      INTEGER                                            :: max_iter
      REAL(real_8)                                       :: threshold
      INTEGER                                            :: selection_crit, nval_request, nrestarts
      LOGICAL                                            :: generalized_ev, iram

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_arnoldi_data', &
         routineP = moduleN//':'//routineN

      CALL setup_arnoldi_control(arnoldi_data, matrix, max_iter, threshold, selection_crit, &
                                 nval_request, nrestarts, generalized_ev, iram)

      SELECT CASE (matrix (1)%matrix%m%data_type)
      CASE (dbcsr_type_real_8)
         CALL setup_arnoldi_data_d(arnoldi_data, matrix, max_iter)
      CASE (dbcsr_type_real_4)
         CALL setup_arnoldi_data_s(arnoldi_data, matrix, max_iter)
      CASE (dbcsr_type_complex_8)
         CALL setup_arnoldi_data_z(arnoldi_data, matrix, max_iter)
      CASE (dbcsr_type_complex_4)
         CALL setup_arnoldi_data_c(arnoldi_data, matrix, max_iter)
      END SELECT

   END SUBROUTINE setup_arnoldi_data

! **************************************************************************************************
!> \brief Creates the control type for arnoldi, see above for details
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
!> \param threshold ...
!> \param selection_crit ...
!> \param nval_request ...
!> \param nrestarts ...
!> \param generalized_ev ...
!> \param iram ...
! **************************************************************************************************
   SUBROUTINE setup_arnoldi_control(arnoldi_data, matrix, max_iter, threshold, selection_crit, nval_request, &
                                    nrestarts, generalized_ev, iram)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      INTEGER                                            :: max_iter
      REAL(real_8)                                       :: threshold
      INTEGER                                            :: selection_crit, nval_request, nrestarts
      LOGICAL                                            :: generalized_ev, iram

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_arnoldi_control', &
         routineP = moduleN//':'//routineN

      TYPE(arnoldi_control), POINTER                     :: control
      TYPE(dbcsr_distribution_obj)                       :: distri

      ALLOCATE (control)
! Fill the information which will later on control the behavior of the arnoldi method and allow synchronization
      CALL dbcsr_get_info(matrix=matrix(1)%matrix, distribution=distri)
      control%mp_group = distri%d%mp_env%mp%mp_group
      control%myproc = distri%d%mp_env%mp%mynode
      CALL dbcsr_mp_grid_setup(distri%d%mp_env)
      IF (.NOT. distri%d%mp_env%mp%subgroups_defined) &
         CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",134,"arnoldi only with subgroups")
      control%pcol_group = distri%d%mp_env%mp%pcol_group
      control%prow_group = distri%d%mp_env%mp%prow_group

      control%symmetric = .FALSE.
! Will need a fix for complex because there it has to be hermitian
      IF (SIZE(matrix) == 1) control%symmetric = &
         dbcsr_get_matrix_type(matrix(1)%matrix) == dbcsr_type_symmetric

! Set the control parameters
      control%max_iter = max_iter
      control%current_step = 0
      control%selection_crit = selection_crit
      control%nval_req = nval_request
      control%threshold = threshold
      control%converged = .FALSE.
      control%has_initial_vector = .FALSE.
      control%iram = iram
      control%nrestart = nrestarts
      control%generalized_ev = generalized_ev

      IF (control%nval_req > 1 .AND. control%nrestart > 0 .AND. .NOT. control%iram) &
         CALL cp_abort(cp__l("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",156), 'with more than one eigenvalue requested '// &
                       'internal restarting with a previous EVEC is a bad idea, set IRAM or nrestsart=0')

! some checks for the generalized EV mode
      IF (control%generalized_ev .AND. selection_crit == 1) &
         CALL cp_abort(cp__l("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",161), &
                       'generalized ev can only highest OR lowest EV')
      IF (control%generalized_ev .AND. nval_request .NE. 1) &
         CALL cp_abort(cp__l("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",164), &
                       'generalized ev can only compute one EV at the time')
      IF (control%generalized_ev .AND. control%nrestart == 0) &
         CALL cp_abort(cp__l("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",167), &
                       'outer loops are mandatory for generalized EV, set nrestart appropriatly')
      IF (SIZE(matrix) .NE. 2 .AND. control%generalized_ev) &
         CALL cp_abort(cp__l("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",170), &
                       'generalized ev needs exactly two matrices as input (2nd is the metric)')

      ALLOCATE (control%selected_ind(max_iter))
      CALL set_control(arnoldi_data, control)

   END SUBROUTINE setup_arnoldi_control

! **************************************************************************************************
!> \brief Deallocate the data in dbcsr_arnoldi_data
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
   SUBROUTINE get_selected_ritz_vector(arnoldi_data, ind, matrix, vector)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      INTEGER                                            :: ind
      TYPE(dbcsr_obj)                                    :: matrix, vector

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_vector', &
         routineP = moduleN//':'//routineN

      IF (has_d_real(arnoldi_data)) CALL get_selected_ritz_vector_d(arnoldi_data, ind, matrix, vector)
      IF (has_s_real(arnoldi_data)) CALL get_selected_ritz_vector_s(arnoldi_data, ind, matrix, vector)
      IF (has_d_cmplx(arnoldi_data)) CALL get_selected_ritz_vector_z(arnoldi_data, ind, matrix, vector)
      IF (has_s_cmplx(arnoldi_data)) CALL get_selected_ritz_vector_c(arnoldi_data, ind, matrix, vector)

   END SUBROUTINE get_selected_ritz_vector

! **************************************************************************************************
!> \brief Deallocate the data in dbcsr_arnoldi_data
!> \param arnoldi_data ...
! **************************************************************************************************
   SUBROUTINE deallocate_arnoldi_data(arnoldi_data)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_arnoldi_data', &
         routineP = moduleN//':'//routineN

      TYPE(arnoldi_control), POINTER                     :: control

      IF (has_d_real(arnoldi_data)) CALL deallocate_arnoldi_data_d(arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL deallocate_arnoldi_data_s(arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL deallocate_arnoldi_data_z(arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL deallocate_arnoldi_data_c(arnoldi_data)

      control => get_control(arnoldi_data)
      DEALLOCATE (control%selected_ind)
      DEALLOCATE (control)

   END SUBROUTINE deallocate_arnoldi_data

! **************************************************************************************************
!> \brief perform the selection of eigenvalues, fills the selected_ind array
!> \param arnoldi_data ...
! **************************************************************************************************
   SUBROUTINE select_evals(arnoldi_data)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'select_evals', routineP = moduleN//':'//routineN

      IF (has_d_real(arnoldi_data)) CALL select_evals_d(arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL select_evals_s(arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL select_evals_z(arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL select_evals_c(arnoldi_data)

   END SUBROUTINE select_evals

! **************************************************************************************************
!> \brief set a new selection type, if you notice you didn't like the initial one
!> \param ar_data ...
!> \param itype ...
! **************************************************************************************************
   SUBROUTINE set_eval_selection(ar_data, itype)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      INTEGER                                            :: itype

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(ar_data)
      control%selection_crit = itype
   END SUBROUTINE set_eval_selection

! **************************************************************************************************
!> \brief returns the number of restarts allowed for arnoldi
!> \param arnoldi_data ...
!> \retval nrestart ...
! **************************************************************************************************
   FUNCTION get_nrestart(arnoldi_data) RESULT(nrestart)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      INTEGER                                            :: nrestart

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(arnoldi_data)
      nrestart = control%nrestart

   END FUNCTION get_nrestart

! **************************************************************************************************
!> \brief get the number of eigenvalues matching the search criterion
!> \param ar_data ...
!> \retval nval_out ...
! **************************************************************************************************
   FUNCTION get_nval_out(ar_data) RESULT(nval_out)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      INTEGER                                            :: nval_out

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(ar_data)
      nval_out = control%nval_out

   END FUNCTION get_nval_out

! **************************************************************************************************
!> \brief get the dimension of the krylov space. Can be less than max_iter
!>        if subspace converged early
!> \param ar_data ...
!> \retval current_step ...
! **************************************************************************************************
   FUNCTION get_subsp_size(ar_data) RESULT(current_step)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      INTEGER                                            :: current_step

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(ar_data)
      current_step = control%current_step

   END FUNCTION get_subsp_size

! **************************************************************************************************
!> \brief Find out whether the method with the current search criterion is converged
!> \param ar_data ...
!> \retval converged ...
! **************************************************************************************************
   FUNCTION arnoldi_is_converged(ar_data) RESULT(converged)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      LOGICAL                                            :: converged

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(ar_data)
      converged = control%converged

   END FUNCTION

! **************************************************************************************************
!> \brief get a single specific Ritz value from the set of selected
!> \param ar_data ...
!> \param ind ...
!> \retval eval_out ...
! **************************************************************************************************
   FUNCTION get_selected_ritz_val(ar_data, ind) RESULT(eval_out)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      INTEGER                                            :: ind
      COMPLEX(real_8)                                    :: eval_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_val', &
         routineP = moduleN//':'//routineN

      COMPLEX(real_4), DIMENSION(:), POINTER             :: evals_s
      COMPLEX(real_8), DIMENSION(:), POINTER             :: evals_d
      INTEGER                                            :: ev_ind
      INTEGER, DIMENSION(:), POINTER                     :: selected_ind

      IF (ind .GT. get_nval_out(ar_data)) &
         CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",339,'outside range of indexed evals')

      selected_ind => get_sel_ind(ar_data)
      ev_ind = selected_ind(ind)
      IF (has_d_real(ar_data)) THEN
         evals_d => get_evals_d(ar_data); eval_out = evals_d(ev_ind)
      ELSE IF (has_s_real(ar_data)) THEN
         evals_s => get_evals_s(ar_data); eval_out = CMPLX(evals_s(ev_ind), KIND=real_8)
      ELSE IF (has_d_cmplx(ar_data)) THEN
         evals_d => get_evals_z(ar_data); eval_out = evals_d(ev_ind)
      ELSE IF (has_s_cmplx(ar_data)) THEN
         evals_s => get_evals_c(ar_data); eval_out = CMPLX(evals_s(ev_ind), KIND=real_8)
      END IF

   END FUNCTION get_selected_ritz_val

! **************************************************************************************************
!> \brief Get all Ritz values of the selected set. eval_out has to be allocated
!>        at least the size of get_neval_out()
!> \param ar_data ...
!> \param eval_out ...
! **************************************************************************************************
   SUBROUTINE get_all_selected_ritz_val(ar_data, eval_out)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      COMPLEX(real_8), DIMENSION(:)                      :: eval_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_all_selected_ritz_val', &
         routineP = moduleN//':'//routineN

      COMPLEX(real_4), DIMENSION(:), POINTER             :: evals_s
      COMPLEX(real_8), DIMENSION(:), POINTER             :: evals_d
      INTEGER                                            :: ev_ind, ind
      INTEGER, DIMENSION(:), POINTER                     :: selected_ind

      NULLIFY (evals_d, evals_s)
      IF (SIZE(eval_out) .LT. get_nval_out(ar_data)) &
         CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",375,'array for eval output too small')
      selected_ind => get_sel_ind(ar_data)

      IF (has_d_real(ar_data)) evals_d => get_evals_d(ar_data)
      IF (has_s_real(ar_data)) evals_s => get_evals_s(ar_data)
      IF (has_d_cmplx(ar_data)) evals_d => get_evals_d(ar_data)
      IF (has_s_cmplx(ar_data)) evals_s => get_evals_c(ar_data)

      DO ind = 1, get_nval_out(ar_data)
         ev_ind = selected_ind(ind)
         IF (ASSOCIATED(evals_d)) eval_out(ind) = evals_d(ev_ind)
         IF (ASSOCIATED(evals_s)) eval_out(ind) = CMPLX(evals_s(ev_ind), KIND=real_8)
      END DO

   END SUBROUTINE get_all_selected_ritz_val

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param vector ...
! **************************************************************************************************
   SUBROUTINE set_arnoldi_initial_vector(ar_data, vector)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(dbcsr_obj)                                    :: vector

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(ar_data)
      control%has_initial_vector = .TRUE.

      IF (has_d_real(ar_data)) CALL set_initial_vector_d(ar_data, vector)
      IF (has_s_real(ar_data)) CALL set_initial_vector_s(ar_data, vector)
      IF (has_d_cmplx(ar_data)) CALL set_initial_vector_z(ar_data, vector)
      IF (has_s_cmplx(ar_data)) CALL set_initial_vector_c(ar_data, vector)

   END SUBROUTINE set_arnoldi_initial_vector


!!! Here come the methods handling the selection of eigenvalues and eigenvectors !!!
!!! If you want a personal method, simply created a Subroutine returning the index
!!! array selected ind which contains as the first nval_out entries the index of the evals

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE select_evals_d(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'select_evals_d', &
      routineP = moduleN//':'//routineN 
    
    INTEGER                                  :: my_crit, last_el, my_ind, i
    REAL(real_8)                        :: convergence
    TYPE(arnoldi_data_d),POINTER   :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control => get_control(arnoldi_data)
    ar_data => get_data_d(arnoldi_data)

    last_el=control%current_step
    convergence=REAL(0.0, real_8)
    my_crit=control%selection_crit
    control%nval_out=MIN(control%nval_req, control%current_step)
    SELECT CASE(my_crit)
    ! minimum and maximum real eval
    CASE(1)
       CALL index_min_max_real_eval_d(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n maximum real eval
    CASE(2)
       CALL index_nmax_real_eval_d(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n minimum real eval
    CASE(3)
       CALL index_nmin_real_eval_d(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    CASE DEFAULT
       CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",39,"unknown selection index")
    END SELECT
    ! test whether we are converged
    DO i=1, control%nval_out
       my_ind=control%selected_ind(i)
       convergence=MAX(convergence, &
                   ABS(ar_data%revec(last_el, my_ind)*ar_data%Hessenberg(last_el+1, last_el)))
    END DO
    control%converged=convergence.LT.control%threshold

  END SUBROUTINE select_evals_d

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_min_max_real_eval_d(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_min_max_real_eval_d', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array
    INTEGER                                   :: i

    neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1,current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(1)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO
    DO i=current_step,1,-1
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(2)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO

  END SUBROUTINE index_min_max_real_eval_d

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmax_real_eval_d(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmax_real_eval_d', &
      routineP = moduleN//':'//routineN
    
    INTEGER                                  :: i, nlimit
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(current_step+1-i))))<EPSILON(0.0_real_8))THEN
          selected_ind(i)=indexing(current_step+1-i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmax_real_eval_d

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmin_real_eval_d(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval,nlimit

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmin_real_eval_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(i)=indexing(i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmin_real_eval_d


!!! Here come the methods handling the selection of eigenvalues and eigenvectors !!!
!!! If you want a personal method, simply created a Subroutine returning the index
!!! array selected ind which contains as the first nval_out entries the index of the evals

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE select_evals_s(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'select_evals_s', &
      routineP = moduleN//':'//routineN 
    
    INTEGER                                  :: my_crit, last_el, my_ind, i
    REAL(real_4)                        :: convergence
    TYPE(arnoldi_data_s),POINTER   :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control => get_control(arnoldi_data)
    ar_data => get_data_s(arnoldi_data)

    last_el=control%current_step
    convergence=REAL(0.0, real_4)
    my_crit=control%selection_crit
    control%nval_out=MIN(control%nval_req, control%current_step)
    SELECT CASE(my_crit)
    ! minimum and maximum real eval
    CASE(1)
       CALL index_min_max_real_eval_s(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n maximum real eval
    CASE(2)
       CALL index_nmax_real_eval_s(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n minimum real eval
    CASE(3)
       CALL index_nmin_real_eval_s(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    CASE DEFAULT
       CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",39,"unknown selection index")
    END SELECT
    ! test whether we are converged
    DO i=1, control%nval_out
       my_ind=control%selected_ind(i)
       convergence=MAX(convergence, &
                   ABS(ar_data%revec(last_el, my_ind)*ar_data%Hessenberg(last_el+1, last_el)))
    END DO
    control%converged=convergence.LT.control%threshold

  END SUBROUTINE select_evals_s

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_min_max_real_eval_s(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_min_max_real_eval_s', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array
    INTEGER                                   :: i

    neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1,current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(1)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO
    DO i=current_step,1,-1
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(2)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO

  END SUBROUTINE index_min_max_real_eval_s

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmax_real_eval_s(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmax_real_eval_s', &
      routineP = moduleN//':'//routineN
    
    INTEGER                                  :: i, nlimit
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(current_step+1-i))))<EPSILON(0.0_real_4))THEN
          selected_ind(i)=indexing(current_step+1-i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmax_real_eval_s

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmin_real_eval_s(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval,nlimit

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmin_real_eval_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(i)=indexing(i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmin_real_eval_s


!!! Here come the methods handling the selection of eigenvalues and eigenvectors !!!
!!! If you want a personal method, simply created a Subroutine returning the index
!!! array selected ind which contains as the first nval_out entries the index of the evals

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE select_evals_z(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'select_evals_z', &
      routineP = moduleN//':'//routineN 
    
    INTEGER                                  :: my_crit, last_el, my_ind, i
    REAL(real_8)                        :: convergence
    TYPE(arnoldi_data_z),POINTER   :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control => get_control(arnoldi_data)
    ar_data => get_data_z(arnoldi_data)

    last_el=control%current_step
    convergence=REAL(0.0, real_8)
    my_crit=control%selection_crit
    control%nval_out=MIN(control%nval_req, control%current_step)
    SELECT CASE(my_crit)
    ! minimum and maximum real eval
    CASE(1)
       CALL index_min_max_real_eval_z(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n maximum real eval
    CASE(2)
       CALL index_nmax_real_eval_z(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n minimum real eval
    CASE(3)
       CALL index_nmin_real_eval_z(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    CASE DEFAULT
       CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",39,"unknown selection index")
    END SELECT
    ! test whether we are converged
    DO i=1, control%nval_out
       my_ind=control%selected_ind(i)
       convergence=MAX(convergence, &
                   ABS(ar_data%revec(last_el, my_ind)*ar_data%Hessenberg(last_el+1, last_el)))
    END DO
    control%converged=convergence.LT.control%threshold

  END SUBROUTINE select_evals_z

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_min_max_real_eval_z(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_min_max_real_eval_z', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array
    INTEGER                                   :: i

    neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1,current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(1)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO
    DO i=current_step,1,-1
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(2)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO

  END SUBROUTINE index_min_max_real_eval_z

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmax_real_eval_z(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmax_real_eval_z', &
      routineP = moduleN//':'//routineN
    
    INTEGER                                  :: i, nlimit
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(current_step+1-i))))<EPSILON(0.0_real_8))THEN
          selected_ind(i)=indexing(current_step+1-i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmax_real_eval_z

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmin_real_eval_z(evals, current_step, selected_ind, neval)
    COMPLEX(real_8), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval,nlimit

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmin_real_eval_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_8), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_8)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_8))THEN
          selected_ind(i)=indexing(i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmin_real_eval_z


!!! Here come the methods handling the selection of eigenvalues and eigenvectors !!!
!!! If you want a personal method, simply created a Subroutine returning the index
!!! array selected ind which contains as the first nval_out entries the index of the evals

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE select_evals_c(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'select_evals_c', &
      routineP = moduleN//':'//routineN 
    
    INTEGER                                  :: my_crit, last_el, my_ind, i
    REAL(real_4)                        :: convergence
    TYPE(arnoldi_data_c),POINTER   :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control => get_control(arnoldi_data)
    ar_data => get_data_c(arnoldi_data)

    last_el=control%current_step
    convergence=REAL(0.0, real_4)
    my_crit=control%selection_crit
    control%nval_out=MIN(control%nval_req, control%current_step)
    SELECT CASE(my_crit)
    ! minimum and maximum real eval
    CASE(1)
       CALL index_min_max_real_eval_c(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n maximum real eval
    CASE(2)
       CALL index_nmax_real_eval_c(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    ! n minimum real eval
    CASE(3)
       CALL index_nmin_real_eval_c(ar_data%evals, control%current_step, control%selected_ind, control%nval_out)
    CASE DEFAULT
       CALL cp__b("dbcsr/arnoldi/dbcsr_arnoldi_data_methods.F",39,"unknown selection index")
    END SELECT
    ! test whether we are converged
    DO i=1, control%nval_out
       my_ind=control%selected_ind(i)
       convergence=MAX(convergence, &
                   ABS(ar_data%revec(last_el, my_ind)*ar_data%Hessenberg(last_el+1, last_el)))
    END DO
    control%converged=convergence.LT.control%threshold

  END SUBROUTINE select_evals_c

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_min_max_real_eval_c(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_min_max_real_eval_c', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array
    INTEGER                                   :: i

    neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1,current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(1)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO
    DO i=current_step,1,-1
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(2)=indexing(i)
          neval=neval+1
          EXIT
       END IF
    END DO

  END SUBROUTINE index_min_max_real_eval_c

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmax_real_eval_c(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmax_real_eval_c', &
      routineP = moduleN//':'//routineN
    
    INTEGER                                  :: i, nlimit
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(current_step+1-i))))<EPSILON(0.0_real_4))THEN
          selected_ind(i)=indexing(current_step+1-i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmax_real_eval_c

! **************************************************************************************************
!> \brief ...
!> \param evals ...
!> \param current_step ...
!> \param selected_ind ...
!> \param neval ...
! **************************************************************************************************
  SUBROUTINE index_nmin_real_eval_c(evals, current_step, selected_ind, neval)
    COMPLEX(real_4), DIMENSION(:)       :: evals
    INTEGER                                  :: current_step
    INTEGER, DIMENSION(:)                    :: selected_ind
    INTEGER                                  :: neval,nlimit

    CHARACTER(LEN=*), PARAMETER :: routineN = 'index_nmin_real_eval_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i
    INTEGER, DIMENSION(current_step)         :: indexing
    REAL(real_4), DIMENSION(current_step)        :: tmp_array

    nlimit=neval; neval=0
    selected_ind=0
    tmp_array(1:current_step)=REAL(evals(1:current_step), real_4)
    CALL sort(tmp_array, current_step, indexing)
    DO i=1, current_step
       IF(ABS(AIMAG(evals(indexing(i))))<EPSILON(0.0_real_4))THEN
          selected_ind(i)=indexing(i)
          neval=neval+1
          IF(neval==nlimit)EXIT
       END IF
    END DO

  END SUBROUTINE index_nmin_real_eval_c

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
! **************************************************************************************************
  SUBROUTINE  setup_arnoldi_data_d (arnoldi_data, matrix, max_iter)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    INTEGER                                  :: max_iter

    CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_arnoldi_data_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                           :: nrow_local
    TYPE(arnoldi_data_d), POINTER           :: ar_data

    ALLOCATE(ar_data)
    CALL dbcsr_get_info(matrix=matrix(1)%matrix, nfullrows_local=nrow_local)
    ALLOCATE(ar_data%f_vec(nrow_local))
    ALLOCATE(ar_data%x_vec(nrow_local))
    ALLOCATE(ar_data%Hessenberg(max_iter+1, max_iter))
    ALLOCATE(ar_data%local_history(nrow_local, max_iter))

    ALLOCATE(ar_data%evals(max_iter))
    ALLOCATE(ar_data%revec(max_iter, max_iter))

    CALL set_data_d(arnoldi_data,ar_data)

  END SUBROUTINE setup_arnoldi_data_d

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE deallocate_arnoldi_data_d (arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                     :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_arnoldi_data_d', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_d), POINTER            :: ar_data

    ar_data=>get_data_d(arnoldi_data)
    IF(ASSOCIATED(ar_data%f_vec))DEALLOCATE(ar_data%f_vec)
    IF(ASSOCIATED(ar_data%x_vec))DEALLOCATE(ar_data%x_vec)
    IF(ASSOCIATED(ar_data%Hessenberg))DEALLOCATE(ar_data%Hessenberg)
    IF(ASSOCIATED(ar_data%local_history))DEALLOCATE(ar_data%local_history)
    IF(ASSOCIATED(ar_data%evals))DEALLOCATE(ar_data%evals)
    IF(ASSOCIATED(ar_data%revec))DEALLOCATE(ar_data%revec)
    DEALLOCATE(ar_data)

  END SUBROUTINE deallocate_arnoldi_data_d

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE get_selected_ritz_vector_d(arnoldi_data,ind,matrix,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    INTEGER                                  :: ind
    TYPE(dbcsr_obj)                          :: matrix
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_vector_d', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_d), POINTER           :: ar_data
    INTEGER                                           :: vsize, myind, sspace_size, i
    INTEGER, DIMENSION(:), POINTER           :: selected_ind
    COMPLEX(real_8),DIMENSION(:),ALLOCATABLE       :: ritz_v
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    selected_ind=>get_sel_ind(arnoldi_data)
    ar_data=>get_data_d(arnoldi_data)
    sspace_size=get_subsp_size(arnoldi_data)
    vsize=SIZE(ar_data%f_vec)
    myind=selected_ind(ind)
    ALLOCATE(ritz_v(vsize))
    ritz_v=CMPLX(0.0,0.0,real_8)

    IF(dbcsr_is_initialized (vector))CALL dbcsr_release(vector)
    CALL create_col_vec_from_matrix(vector,matrix,1)
    IF(control%local_comp)THEN
       DO i=1,sspace_size
          ritz_v(:)=ritz_v(:)+ar_data%local_history(:,i)*ar_data%revec(i,myind)
       END DO
       data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=0.0_real_8)
       ! is a bit odd but ritz_v is always complex and matrix type determines where it goes
       ! again I hope the user knows what is required
       data_vec(1:vsize) =REAL(ritz_v(1:vsize),KIND=real_8)
    END IF

    DEALLOCATE(ritz_v)

  END SUBROUTINE get_selected_ritz_vector_d
     
   
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE set_initial_vector_d(arnoldi_data,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'set_initial_vector_d', &
      routineP = moduleN//':'//routineN
    
    TYPE(arnoldi_data_d), POINTER           :: ar_data
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    INTEGER                                           :: nrow_local, ncol_local
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)

    CALL dbcsr_get_info(matrix=vector, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ar_data=>get_data_d(arnoldi_data)
    data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=0.0_real_8)
    IF(nrow_local*ncol_local>0)ar_data%f_vec(1:nrow_local)=data_vec(1:nrow_local)

  END SUBROUTINE set_initial_vector_d  

    
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
! **************************************************************************************************
  SUBROUTINE  setup_arnoldi_data_s (arnoldi_data, matrix, max_iter)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    INTEGER                                  :: max_iter

    CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_arnoldi_data_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                           :: nrow_local
    TYPE(arnoldi_data_s), POINTER           :: ar_data

    ALLOCATE(ar_data)
    CALL dbcsr_get_info(matrix=matrix(1)%matrix, nfullrows_local=nrow_local)
    ALLOCATE(ar_data%f_vec(nrow_local))
    ALLOCATE(ar_data%x_vec(nrow_local))
    ALLOCATE(ar_data%Hessenberg(max_iter+1, max_iter))
    ALLOCATE(ar_data%local_history(nrow_local, max_iter))

    ALLOCATE(ar_data%evals(max_iter))
    ALLOCATE(ar_data%revec(max_iter, max_iter))

    CALL set_data_s(arnoldi_data,ar_data)

  END SUBROUTINE setup_arnoldi_data_s

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE deallocate_arnoldi_data_s (arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                     :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_arnoldi_data_s', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_s), POINTER            :: ar_data

    ar_data=>get_data_s(arnoldi_data)
    IF(ASSOCIATED(ar_data%f_vec))DEALLOCATE(ar_data%f_vec)
    IF(ASSOCIATED(ar_data%x_vec))DEALLOCATE(ar_data%x_vec)
    IF(ASSOCIATED(ar_data%Hessenberg))DEALLOCATE(ar_data%Hessenberg)
    IF(ASSOCIATED(ar_data%local_history))DEALLOCATE(ar_data%local_history)
    IF(ASSOCIATED(ar_data%evals))DEALLOCATE(ar_data%evals)
    IF(ASSOCIATED(ar_data%revec))DEALLOCATE(ar_data%revec)
    DEALLOCATE(ar_data)

  END SUBROUTINE deallocate_arnoldi_data_s

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE get_selected_ritz_vector_s(arnoldi_data,ind,matrix,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    INTEGER                                  :: ind
    TYPE(dbcsr_obj)                          :: matrix
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_vector_s', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_s), POINTER           :: ar_data
    INTEGER                                           :: vsize, myind, sspace_size, i
    INTEGER, DIMENSION(:), POINTER           :: selected_ind
    COMPLEX(real_4),DIMENSION(:),ALLOCATABLE       :: ritz_v
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    selected_ind=>get_sel_ind(arnoldi_data)
    ar_data=>get_data_s(arnoldi_data)
    sspace_size=get_subsp_size(arnoldi_data)
    vsize=SIZE(ar_data%f_vec)
    myind=selected_ind(ind)
    ALLOCATE(ritz_v(vsize))
    ritz_v=CMPLX(0.0,0.0,real_4)

    IF(dbcsr_is_initialized (vector))CALL dbcsr_release(vector)
    CALL create_col_vec_from_matrix(vector,matrix,1)
    IF(control%local_comp)THEN
       DO i=1,sspace_size
          ritz_v(:)=ritz_v(:)+ar_data%local_history(:,i)*ar_data%revec(i,myind)
       END DO
       data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=0.0_real_4)
       ! is a bit odd but ritz_v is always complex and matrix type determines where it goes
       ! again I hope the user knows what is required
       data_vec(1:vsize) =REAL(ritz_v(1:vsize),KIND=real_4)
    END IF

    DEALLOCATE(ritz_v)

  END SUBROUTINE get_selected_ritz_vector_s
     
   
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE set_initial_vector_s(arnoldi_data,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'set_initial_vector_s', &
      routineP = moduleN//':'//routineN
    
    TYPE(arnoldi_data_s), POINTER           :: ar_data
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    INTEGER                                           :: nrow_local, ncol_local
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)

    CALL dbcsr_get_info(matrix=vector, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ar_data=>get_data_s(arnoldi_data)
    data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=0.0_real_4)
    IF(nrow_local*ncol_local>0)ar_data%f_vec(1:nrow_local)=data_vec(1:nrow_local)

  END SUBROUTINE set_initial_vector_s  

    
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
! **************************************************************************************************
  SUBROUTINE  setup_arnoldi_data_z (arnoldi_data, matrix, max_iter)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    INTEGER                                  :: max_iter

    CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_arnoldi_data_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                           :: nrow_local
    TYPE(arnoldi_data_z), POINTER           :: ar_data

    ALLOCATE(ar_data)
    CALL dbcsr_get_info(matrix=matrix(1)%matrix, nfullrows_local=nrow_local)
    ALLOCATE(ar_data%f_vec(nrow_local))
    ALLOCATE(ar_data%x_vec(nrow_local))
    ALLOCATE(ar_data%Hessenberg(max_iter+1, max_iter))
    ALLOCATE(ar_data%local_history(nrow_local, max_iter))

    ALLOCATE(ar_data%evals(max_iter))
    ALLOCATE(ar_data%revec(max_iter, max_iter))

    CALL set_data_z(arnoldi_data,ar_data)

  END SUBROUTINE setup_arnoldi_data_z

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE deallocate_arnoldi_data_z (arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                     :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_arnoldi_data_z', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_z), POINTER            :: ar_data

    ar_data=>get_data_z(arnoldi_data)
    IF(ASSOCIATED(ar_data%f_vec))DEALLOCATE(ar_data%f_vec)
    IF(ASSOCIATED(ar_data%x_vec))DEALLOCATE(ar_data%x_vec)
    IF(ASSOCIATED(ar_data%Hessenberg))DEALLOCATE(ar_data%Hessenberg)
    IF(ASSOCIATED(ar_data%local_history))DEALLOCATE(ar_data%local_history)
    IF(ASSOCIATED(ar_data%evals))DEALLOCATE(ar_data%evals)
    IF(ASSOCIATED(ar_data%revec))DEALLOCATE(ar_data%revec)
    DEALLOCATE(ar_data)

  END SUBROUTINE deallocate_arnoldi_data_z

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE get_selected_ritz_vector_z(arnoldi_data,ind,matrix,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    INTEGER                                  :: ind
    TYPE(dbcsr_obj)                          :: matrix
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_vector_z', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_z), POINTER           :: ar_data
    INTEGER                                           :: vsize, myind, sspace_size, i
    INTEGER, DIMENSION(:), POINTER           :: selected_ind
    COMPLEX(real_8),DIMENSION(:),ALLOCATABLE       :: ritz_v
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    selected_ind=>get_sel_ind(arnoldi_data)
    ar_data=>get_data_z(arnoldi_data)
    sspace_size=get_subsp_size(arnoldi_data)
    vsize=SIZE(ar_data%f_vec)
    myind=selected_ind(ind)
    ALLOCATE(ritz_v(vsize))
    ritz_v=CMPLX(0.0,0.0,real_8)

    IF(dbcsr_is_initialized (vector))CALL dbcsr_release(vector)
    CALL create_col_vec_from_matrix(vector,matrix,1)
    IF(control%local_comp)THEN
       DO i=1,sspace_size
          ritz_v(:)=ritz_v(:)+ar_data%local_history(:,i)*ar_data%revec(i,myind)
       END DO
       data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
       ! is a bit odd but ritz_v is always complex and matrix type determines where it goes
       ! again I hope the user knows what is required
       data_vec(1:vsize) =CMPLX(ritz_v(1:vsize),KIND=real_8)
    END IF

    DEALLOCATE(ritz_v)

  END SUBROUTINE get_selected_ritz_vector_z
     
   
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE set_initial_vector_z(arnoldi_data,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'set_initial_vector_z', &
      routineP = moduleN//':'//routineN
    
    TYPE(arnoldi_data_z), POINTER           :: ar_data
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    INTEGER                                           :: nrow_local, ncol_local
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)

    CALL dbcsr_get_info(matrix=vector, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ar_data=>get_data_z(arnoldi_data)
    data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    IF(nrow_local*ncol_local>0)ar_data%f_vec(1:nrow_local)=data_vec(1:nrow_local)

  END SUBROUTINE set_initial_vector_z  

    
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
! **************************************************************************************************
  SUBROUTINE  setup_arnoldi_data_c (arnoldi_data, matrix, max_iter)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    INTEGER                                  :: max_iter

    CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_arnoldi_data_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                           :: nrow_local
    TYPE(arnoldi_data_c), POINTER           :: ar_data

    ALLOCATE(ar_data)
    CALL dbcsr_get_info(matrix=matrix(1)%matrix, nfullrows_local=nrow_local)
    ALLOCATE(ar_data%f_vec(nrow_local))
    ALLOCATE(ar_data%x_vec(nrow_local))
    ALLOCATE(ar_data%Hessenberg(max_iter+1, max_iter))
    ALLOCATE(ar_data%local_history(nrow_local, max_iter))

    ALLOCATE(ar_data%evals(max_iter))
    ALLOCATE(ar_data%revec(max_iter, max_iter))

    CALL set_data_c(arnoldi_data,ar_data)

  END SUBROUTINE setup_arnoldi_data_c

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE deallocate_arnoldi_data_c (arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                     :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_arnoldi_data_c', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_c), POINTER            :: ar_data

    ar_data=>get_data_c(arnoldi_data)
    IF(ASSOCIATED(ar_data%f_vec))DEALLOCATE(ar_data%f_vec)
    IF(ASSOCIATED(ar_data%x_vec))DEALLOCATE(ar_data%x_vec)
    IF(ASSOCIATED(ar_data%Hessenberg))DEALLOCATE(ar_data%Hessenberg)
    IF(ASSOCIATED(ar_data%local_history))DEALLOCATE(ar_data%local_history)
    IF(ASSOCIATED(ar_data%evals))DEALLOCATE(ar_data%evals)
    IF(ASSOCIATED(ar_data%revec))DEALLOCATE(ar_data%revec)
    DEALLOCATE(ar_data)

  END SUBROUTINE deallocate_arnoldi_data_c

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE get_selected_ritz_vector_c(arnoldi_data,ind,matrix,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    INTEGER                                  :: ind
    TYPE(dbcsr_obj)                          :: matrix
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'get_selected_ritz_vector_c', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_c), POINTER           :: ar_data
    INTEGER                                           :: vsize, myind, sspace_size, i
    INTEGER, DIMENSION(:), POINTER           :: selected_ind
    COMPLEX(real_4),DIMENSION(:),ALLOCATABLE       :: ritz_v
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    selected_ind=>get_sel_ind(arnoldi_data)
    ar_data=>get_data_c(arnoldi_data)
    sspace_size=get_subsp_size(arnoldi_data)
    vsize=SIZE(ar_data%f_vec)
    myind=selected_ind(ind)
    ALLOCATE(ritz_v(vsize))
    ritz_v=CMPLX(0.0,0.0,real_4)

    IF(dbcsr_is_initialized (vector))CALL dbcsr_release(vector)
    CALL create_col_vec_from_matrix(vector,matrix,1)
    IF(control%local_comp)THEN
       DO i=1,sspace_size
          ritz_v(:)=ritz_v(:)+ar_data%local_history(:,i)*ar_data%revec(i,myind)
       END DO
       data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
       ! is a bit odd but ritz_v is always complex and matrix type determines where it goes
       ! again I hope the user knows what is required
       data_vec(1:vsize) =CMPLX(ritz_v(1:vsize),KIND=real_4)
    END IF

    DEALLOCATE(ritz_v)

  END SUBROUTINE get_selected_ritz_vector_c
     
   
! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param vector ...
! **************************************************************************************************
  SUBROUTINE set_initial_vector_c(arnoldi_data,vector)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data
    TYPE(dbcsr_obj)                          :: vector

    CHARACTER(LEN=*), PARAMETER :: routineN = 'set_initial_vector_c', &
      routineP = moduleN//':'//routineN
    
    TYPE(arnoldi_data_c), POINTER           :: ar_data
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    INTEGER                                           :: nrow_local, ncol_local
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)

    CALL dbcsr_get_info(matrix=vector, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ar_data=>get_data_c(arnoldi_data)
    data_vec => dbcsr_get_data_p (vector%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    IF(nrow_local*ncol_local>0)ar_data%f_vec(1:nrow_local)=data_vec(1:nrow_local)

  END SUBROUTINE set_initial_vector_c  

    

END MODULE dbcsr_arnoldi_data_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Wrappers to message passing calls.
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_mp_operations

   USE dbcsr_config,                    ONLY: has_MPI
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_type
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_fatal_level,&
                                              dbcsr_wrong_args_error
   USE dbcsr_mp_methods,                ONLY: &
        dbcsr_mp_get_process, dbcsr_mp_grid_setup, dbcsr_mp_group, dbcsr_mp_has_subgroups, &
        dbcsr_mp_my_col_group, dbcsr_mp_my_row_group, dbcsr_mp_mynode, dbcsr_mp_mypcol, &
        dbcsr_mp_myprow, dbcsr_mp_npcols, dbcsr_mp_nprows, dbcsr_mp_numnodes, dbcsr_mp_pgrid
   USE dbcsr_ptr_util,                  ONLY: memory_copy
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_mp_obj,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_int_4,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE kinds,                           ONLY: real_4,&
                                              real_8
   USE message_passing,                 ONLY: &
        mp_allgather, mp_alltoall, mp_gatherv, mp_ibcast, mp_irecv, mp_iscatter, mp_isend, &
        mp_rget, mp_sendrecv, mp_type_descriptor_type, mp_type_indexed_make_c, &
        mp_type_indexed_make_d, mp_type_indexed_make_r, mp_type_indexed_make_z, mp_type_make, &
        mp_waitall, mp_win_create
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mp_operations'

   ! MP routines
   PUBLIC :: hybrid_alltoall_s1, hybrid_alltoall_d1, &
             hybrid_alltoall_c1, hybrid_alltoall_z1, &
             hybrid_alltoall_i1, hybrid_alltoall_any
   PUBLIC :: dbcsr_allgatherv
   PUBLIC :: dbcsr_sendrecv_any
   PUBLIC :: dbcsr_isend_any, dbcsr_irecv_any
   PUBLIC :: dbcsr_win_create_any, dbcsr_rget_any, dbcsr_ibcast_any
   PUBLIC :: dbcsr_iscatterv_any, dbcsr_gatherv_any
   PUBLIC :: dbcsr_mp_type_indexed_make_any
   ! Type helpers
   PUBLIC :: dbcsr_mp_type_from_anytype

   INTERFACE dbcsr_hybrid_alltoall
      MODULE PROCEDURE hybrid_alltoall_s1, hybrid_alltoall_d1, &
         hybrid_alltoall_c1, hybrid_alltoall_z1
      MODULE PROCEDURE hybrid_alltoall_i1
      MODULE PROCEDURE hybrid_alltoall_any
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param mp_env ...
!> \param most_ptp ...
!> \param remainder_ptp ...
!> \param no_hybrid ...
! **************************************************************************************************
   SUBROUTINE hybrid_alltoall_any(sb, scount, sdispl, &
                                  rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: sb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: scount, sdispl
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: rb
      INTEGER, DIMENSION(:), INTENT(IN)                  :: rcount, rdispl
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      LOGICAL, INTENT(in), OPTIONAL                      :: most_ptp, remainder_ptp, no_hybrid

      CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_any', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)

      SELECT CASE (dbcsr_data_get_type (sb))
      CASE (dbcsr_type_real_4)
         CALL hybrid_alltoall_s1(sb%d%r_sp, scount, sdispl, &
                                 rb%d%r_sp, rcount, rdispl, mp_env, &
                                 most_ptp, remainder_ptp, no_hybrid)
      CASE (dbcsr_type_real_8)
         CALL hybrid_alltoall_d1(sb%d%r_dp, scount, sdispl, &
                                 rb%d%r_dp, rcount, rdispl, mp_env, &
                                 most_ptp, remainder_ptp, no_hybrid)
      CASE (dbcsr_type_complex_4)
         CALL hybrid_alltoall_c1(sb%d%c_sp, scount, sdispl, &
                                 rb%d%c_sp, rcount, rdispl, mp_env, &
                                 most_ptp, remainder_ptp, no_hybrid)
      CASE (dbcsr_type_complex_8)
         CALL hybrid_alltoall_z1(sb%d%c_dp, scount, sdispl, &
                                 rb%d%c_dp, rcount, rdispl, mp_env, &
                                 most_ptp, remainder_ptp, no_hybrid)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",121,"Invalid data type")
      END SELECT

      CALL timestop(error_handle)
   END SUBROUTINE hybrid_alltoall_any

! **************************************************************************************************
!> \brief Row/column and global all-to-all
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param[in] mp_env         MP Environment
!> \param[in] most_ptp       (optional) Use point-to-point for row/column;
!>                           default is no
!> \param[in] remainder_ptp  (optional) Use point-to-point for remaining;
!>                           default is no
!> \param[in] no_hybrid      (optional) Use regular global collective; default
!>                           is no
!> \par Communicator selection
!>      Uses row and column communicators for row/column
!>      sends. Remaining sends are performed using the global
!>      communicator.  Point-to-point isend/irecv are used if ptp is
!>      set, otherwise a alltoall collective call is issued.
!>      see mp_alltoall
! **************************************************************************************************
   SUBROUTINE hybrid_alltoall_i1(sb, scount, sdispl, &
                                 rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
      INTEGER, DIMENSION(:), INTENT(in), &
         TARGET                                 :: sb
      INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
      INTEGER, DIMENSION(:), &
         INTENT(INOUT), TARGET                  :: rb
      INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
      TYPE(dbcsr_mp_obj), INTENT(IN)           :: mp_env
      LOGICAL, INTENT(IN), OPTIONAL            :: most_ptp, remainder_ptp, &
                                                  no_hybrid

      CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_i1', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: all_group, mynode, mypcol, myprow, nall_rr, nall_sr, ncol_rr, &
                 ncol_sr, npcols, nprows, nrow_rr, nrow_sr, numnodes, dst, src, &
                 prow, pcol, send_cnt, recv_cnt, tag, grp, i
      INTEGER, ALLOCATABLE, DIMENSION(:) :: all_rr, all_sr, col_rr, col_sr, &
                                            new_rcount, new_rdispl, new_scount, new_sdispl, row_rr, row_sr
      INTEGER, DIMENSION(:, :), POINTER        :: pgrid
      LOGICAL                                  :: most_collective, &
                                                  remainder_collective, no_h
      INTEGER, DIMENSION(:), POINTER           :: send_data_p, recv_data_p
      TYPE(dbcsr_mp_obj)                       :: mpe

      IF (.NOT. dbcsr_mp_has_subgroups(mp_env)) THEN
         mpe = mp_env
         CALL dbcsr_mp_grid_setup(mpe)
      ENDIF
      most_collective = .TRUE.
      remainder_collective = .TRUE.
      no_h = .FALSE.
      IF (PRESENT(most_ptp)) most_collective = .NOT. most_ptp
      IF (PRESENT(remainder_ptp)) remainder_collective = .NOT. remainder_ptp
      IF (PRESENT(no_hybrid)) no_h = no_hybrid
      all_group = dbcsr_mp_group(mp_env)
      ! Don't use subcommunicators if they're not defined.
      no_h = no_h .OR. .NOT. dbcsr_mp_has_subgroups(mp_env) .OR. .NOT. has_MPI
      subgrouped: IF (mp_env%mp%subgroups_defined .AND. .NOT. no_h) THEN
         mynode = dbcsr_mp_mynode(mp_env)
         numnodes = dbcsr_mp_numnodes(mp_env)
         nprows = dbcsr_mp_nprows(mp_env)
         npcols = dbcsr_mp_npcols(mp_env)
         myprow = dbcsr_mp_myprow(mp_env)
         mypcol = dbcsr_mp_mypcol(mp_env)
         pgrid => dbcsr_mp_pgrid(mp_env)
         ALLOCATE (row_sr(0:npcols-1)); nrow_sr = 0
         ALLOCATE (row_rr(0:npcols-1)); nrow_rr = 0
         ALLOCATE (col_sr(0:nprows-1)); ncol_sr = 0
         ALLOCATE (col_rr(0:nprows-1)); ncol_rr = 0
         ALLOCATE (all_sr(0:numnodes-1)); nall_sr = 0
         ALLOCATE (all_rr(0:numnodes-1)); nall_rr = 0
         ALLOCATE (new_scount(numnodes), new_rcount(numnodes))
         ALLOCATE (new_sdispl(numnodes), new_rdispl(numnodes))
         IF (.NOT. remainder_collective) THEN
            CALL remainder_point_to_point()
         ENDIF
         IF (.NOT. most_collective) THEN
            CALL most_point_to_point()
         ELSE
            CALL most_alltoall()
         ENDIF
         IF (remainder_collective) THEN
            CALL remainder_alltoall()
         ENDIF
         ! Wait for all issued sends and receives.
         IF (.NOT. most_collective) THEN
            CALL mp_waitall(row_sr(0:nrow_sr-1))
            CALL mp_waitall(col_sr(0:ncol_sr-1))
            CALL mp_waitall(row_rr(0:nrow_rr-1))
            CALL mp_waitall(col_rr(0:ncol_rr-1))
         END IF
         IF (.NOT. remainder_collective) THEN
            CALL mp_waitall(all_sr(1:nall_sr))
            CALL mp_waitall(all_rr(1:nall_rr))
         ENDIF
      ELSE
         CALL mp_alltoall(sb, scount, sdispl, &
                          rb, rcount, rdispl, &
                          all_group)
      ENDIF subgrouped
   CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE most_alltoall()
         DO pcol = 0, npcols-1
            new_scount(1+pcol) = scount(1+pgrid(myprow, pcol))
            new_rcount(1+pcol) = rcount(1+pgrid(myprow, pcol))
            new_sdispl(1+pcol) = sdispl(1+pgrid(myprow, pcol))
            new_rdispl(1+pcol) = rdispl(1+pgrid(myprow, pcol))
         END DO
         CALL mp_alltoall(sb, new_scount(1:npcols), new_sdispl(1:npcols), &
                          rb, new_rcount(1:npcols), new_rdispl(1:npcols), &
                          dbcsr_mp_my_row_group(mp_env))
         DO prow = 0, nprows-1
            new_scount(1+prow) = scount(1+pgrid(prow, mypcol))
            new_rcount(1+prow) = rcount(1+pgrid(prow, mypcol))
            new_sdispl(1+prow) = sdispl(1+pgrid(prow, mypcol))
            new_rdispl(1+prow) = rdispl(1+pgrid(prow, mypcol))
         END DO
         CALL mp_alltoall(sb, new_scount(1:nprows), new_sdispl(1:nprows), &
                          rb, new_rcount(1:nprows), new_rdispl(1:nprows), &
                          dbcsr_mp_my_col_group(mp_env))
      END SUBROUTINE most_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE most_point_to_point()
         ! Go through my prow and exchange.
         DO i = 0, npcols-1
            pcol = MOD(mypcol+i, npcols)
            grp = dbcsr_mp_my_row_group(mp_env)
            !
            dst = dbcsr_mp_get_process(mp_env, myprow, pcol)
            send_cnt = scount(dst+1)
            send_data_p => sb(1+sdispl(dst+1):1+sdispl(dst+1)+send_cnt-1)
            tag = 4*mypcol
            IF (send_cnt .GT. 0) THEN
               CALL mp_isend(send_data_p, pcol, grp, row_sr(nrow_sr), tag)
               nrow_sr = nrow_sr+1
            ENDIF
            !
            pcol = MODULO(mypcol-i, npcols)
            src = dbcsr_mp_get_process(mp_env, myprow, pcol)
            recv_cnt = rcount(src+1)
            recv_data_p => rb(1+rdispl(src+1):1+rdispl(src+1)+recv_cnt-1)
            tag = 4*pcol
            IF (recv_cnt .GT. 0) THEN
               CALL mp_irecv(recv_data_p, pcol, grp, row_rr(nrow_rr), tag)
               nrow_rr = nrow_rr+1
            ENDIF
         ENDDO
         ! go through my pcol and exchange
         DO i = 0, nprows-1
            prow = MOD(myprow+i, nprows)
            grp = dbcsr_mp_my_col_group(mp_env)
            !
            dst = dbcsr_mp_get_process(mp_env, prow, mypcol)
            send_cnt = scount(dst+1)
            IF (send_cnt .GT. 0) THEN
               send_data_p => sb(1+sdispl(dst+1):1+sdispl(dst+1)+send_cnt-1)
               tag = 4*myprow+1
               CALL mp_isend(send_data_p, prow, grp, col_sr(ncol_sr), tag)
               ncol_sr = ncol_sr+1
            ENDIF
            !
            prow = MODULO(myprow-i, nprows)
            src = dbcsr_mp_get_process(mp_env, prow, mypcol)
            recv_cnt = rcount(src+1)
            IF (recv_cnt .GT. 0) THEN
               recv_data_p => rb(1+rdispl(src+1):1+rdispl(src+1)+recv_cnt-1)
               tag = 4*prow+1
               CALL mp_irecv(recv_data_p, prow, grp, col_rr(ncol_rr), tag)
               ncol_rr = ncol_rr+1
            ENDIF
         ENDDO
      END SUBROUTINE most_point_to_point
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE remainder_alltoall()
         new_scount(:) = scount(:)
         new_rcount(:) = rcount(:)
         DO prow = 0, nprows-1
            new_scount(1+pgrid(prow, mypcol)) = 0
            new_rcount(1+pgrid(prow, mypcol)) = 0
         END DO
         DO pcol = 0, npcols-1
            new_scount(1+pgrid(myprow, pcol)) = 0
            new_rcount(1+pgrid(myprow, pcol)) = 0
         END DO
         CALL mp_alltoall(sb, new_scount, sdispl, &
                          rb, new_rcount, rdispl, all_group)
      END SUBROUTINE remainder_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE remainder_point_to_point()
      INTEGER                                            :: col, row

         DO row = 0, nprows-1
            prow = MOD(row+myprow, nprows)
            IF (prow .EQ. myprow) CYCLE
            DO col = 0, npcols-1
               pcol = MOD(col+mypcol, npcols)
               IF (pcol .EQ. mypcol) CYCLE
               dst = dbcsr_mp_get_process(mp_env, prow, pcol)
               send_cnt = scount(dst+1)
               IF (send_cnt .GT. 0) THEN
                  tag = 4*mynode+2
                  send_data_p => sb(1+sdispl(dst+1):1+sdispl(dst+1)+send_cnt-1)
                  CALL mp_isend(send_data_p, dst, all_group, all_sr(nall_sr+1), tag)
                  nall_sr = nall_sr+1
               ENDIF
               !
               src = dbcsr_mp_get_process(mp_env, prow, pcol)
               recv_cnt = rcount(src+1)
               IF (recv_cnt .GT. 0) THEN
                  recv_data_p => rb(1+rdispl(src+1):1+rdispl(src+1)+recv_cnt-1)
                  tag = 4*src+2
                  CALL mp_irecv(recv_data_p, src, all_group, all_rr(nall_rr+1), tag)
                  nall_rr = nall_rr+1
               ENDIF
            ENDDO
         ENDDO
      END SUBROUTINE remainder_point_to_point
   END SUBROUTINE hybrid_alltoall_i1

! **************************************************************************************************
!> \brief Creates an MPI combined type from the given anytype.
!> \param[in] data_area       Data area of any type
!> \retval mp_type            Type descriptor
! **************************************************************************************************
   FUNCTION dbcsr_mp_type_from_anytype(data_area) RESULT(mp_type)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: data_area
      TYPE(mp_type_descriptor_type)                      :: mp_type

      SELECT CASE (data_area%d%data_type)
      CASE (dbcsr_type_int_4)
         mp_type = mp_type_make(data_area%d%i4)
      CASE (dbcsr_type_real_4)
         mp_type = mp_type_make(data_area%d%r_sp)
      CASE (dbcsr_type_real_8)
         mp_type = mp_type_make(data_area%d%r_dp)
      CASE (dbcsr_type_complex_4)
         mp_type = mp_type_make(data_area%d%c_sp)
      CASE (dbcsr_type_complex_8)
         mp_type = mp_type_make(data_area%d%c_dp)
      END SELECT
   END FUNCTION dbcsr_mp_type_from_anytype

! **************************************************************************************************
!> \brief sendrecv of encapsulated data.
!> \param msgin ...
!> \param dest ...
!> \param msgout ...
!> \param source ...
!> \param comm ...
!> \note see mp_sendrecv
! **************************************************************************************************
   SUBROUTINE dbcsr_sendrecv_any(msgin, dest, msgout, source, comm)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: msgin
      INTEGER, INTENT(IN)                                :: dest
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: msgout
      INTEGER, INTENT(IN)                                :: source, comm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_sendrecv_any', &
         routineP = moduleN//':'//routineN

      IF (dbcsr_data_get_type(msgin) .NE. dbcsr_data_get_type(msgout)) &
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",401,"Different data type for msgin and msgout")

      SELECT CASE (dbcsr_data_get_type (msgin))
      CASE (dbcsr_type_real_4)
         CALL mp_sendrecv(msgin%d%r_sp, dest, msgout%d%r_sp, source, comm)
      CASE (dbcsr_type_real_8)
         CALL mp_sendrecv(msgin%d%r_dp, dest, msgout%d%r_dp, source, comm)
      CASE (dbcsr_type_complex_4)
         CALL mp_sendrecv(msgin%d%c_sp, dest, msgout%d%c_sp, source, comm)
      CASE (dbcsr_type_complex_8)
         CALL mp_sendrecv(msgin%d%c_dp, dest, msgout%d%c_dp, source, comm)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",413,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_sendrecv_any

! **************************************************************************************************
!> \brief Non-blocking send of encapsulated data.
!> \param msgin ...
!> \param dest ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \note see mp_isend_iv
! **************************************************************************************************
   SUBROUTINE dbcsr_isend_any(msgin, dest, comm, request, tag)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: msgin
      INTEGER, INTENT(IN)                                :: dest, comm
      INTEGER, INTENT(OUT)                               :: request
      INTEGER, INTENT(IN), OPTIONAL                      :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_isend_any', &
         routineP = moduleN//':'//routineN

      SELECT CASE (dbcsr_data_get_type (msgin))
      CASE (dbcsr_type_real_4)
         CALL mp_isend(msgin%d%r_sp, dest, comm, request, tag)
      CASE (dbcsr_type_real_8)
         CALL mp_isend(msgin%d%r_dp, dest, comm, request, tag)
      CASE (dbcsr_type_complex_4)
         CALL mp_isend(msgin%d%c_sp, dest, comm, request, tag)
      CASE (dbcsr_type_complex_8)
         CALL mp_isend(msgin%d%c_dp, dest, comm, request, tag)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",445,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_isend_any

! **************************************************************************************************
!> \brief Non-blocking recv of encapsulated data.
!> \param msgin ...
!> \param source ...
!> \param comm ...
!> \param request ...
!> \param tag ...
!> \note see mp_irecv_iv
! **************************************************************************************************
   SUBROUTINE dbcsr_irecv_any(msgin, source, comm, request, tag)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: msgin
      INTEGER, INTENT(IN)                                :: source, comm
      INTEGER, INTENT(OUT)                               :: request
      INTEGER, INTENT(IN), OPTIONAL                      :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_irecv_any', &
         routineP = moduleN//':'//routineN

      SELECT CASE (dbcsr_data_get_type (msgin))
      CASE (dbcsr_type_real_4)
         CALL mp_irecv(msgin%d%r_sp, source, comm, request, tag)
      CASE (dbcsr_type_real_8)
         CALL mp_irecv(msgin%d%r_dp, source, comm, request, tag)
      CASE (dbcsr_type_complex_4)
         CALL mp_irecv(msgin%d%c_sp, source, comm, request, tag)
      CASE (dbcsr_type_complex_8)
         CALL mp_irecv(msgin%d%c_dp, source, comm, request, tag)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",477,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_irecv_any

! **************************************************************************************************
!> \brief Window initialization function of encapsulated data.
!> \param base ...
!> \param comm ...
!> \param win ...
! **************************************************************************************************
   SUBROUTINE dbcsr_win_create_any(base, comm, win)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: base
      INTEGER, INTENT(IN)                                :: comm
      INTEGER, INTENT(OUT)                               :: win

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_win_create_any', &
         routineP = moduleN//':'//routineN

      SELECT CASE (dbcsr_data_get_type (base))
      CASE (dbcsr_type_real_4)
         CALL mp_win_create(base%d%r_sp, comm, win)
      CASE (dbcsr_type_real_8)
         CALL mp_win_create(base%d%r_dp, comm, win)
      CASE (dbcsr_type_complex_4)
         CALL mp_win_create(base%d%c_sp, comm, win)
      CASE (dbcsr_type_complex_8)
         CALL mp_win_create(base%d%c_dp, comm, win)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",505,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_win_create_any

! **************************************************************************************************
!> \brief Single-sided Get function of encapsulated data.
!> \param base ...
!> \param source ...
!> \param win ...
!> \param win_data ...
!> \param myproc ...
!> \param disp ...
!> \param request ...
!> \param origin_datatype ...
!> \param target_datatype ...
! **************************************************************************************************
   SUBROUTINE dbcsr_rget_any(base, source, win, win_data, myproc, disp, request, &
                             origin_datatype, target_datatype)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: base
      INTEGER, INTENT(IN)                                :: source, win
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: win_data
      INTEGER, INTENT(IN), OPTIONAL                      :: myproc, disp
      INTEGER, INTENT(OUT)                               :: request
      TYPE(mp_type_descriptor_type), INTENT(IN), &
         OPTIONAL                                        :: origin_datatype, target_datatype

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_rget_any', routineP = moduleN//':'//routineN

      CALL dbcsr_assert(dbcsr_data_get_type(base), "EQ", dbcsr_data_get_type(win_data), &
                        dbcsr_fatal_level, dbcsr_caller_error, &
                        routineN, "Mismatch data type between buffer and window", 535)

      SELECT CASE (dbcsr_data_get_type (base))
      CASE (dbcsr_type_real_4)
         CALL mp_rget(base%d%r_sp, source, win, win_data%d%r_sp, myproc, &
                      disp, request, origin_datatype, target_datatype)
      CASE (dbcsr_type_real_8)
         CALL mp_rget(base%d%r_dp, source, win, win_data%d%r_dp, myproc, &
                      disp, request, origin_datatype, target_datatype)
      CASE (dbcsr_type_complex_4)
         CALL mp_rget(base%d%c_sp, source, win, win_data%d%c_sp, myproc, &
                      disp, request, origin_datatype, target_datatype)
      CASE (dbcsr_type_complex_8)
         CALL mp_rget(base%d%c_dp, source, win, win_data%d%c_dp, myproc, &
                      disp, request, origin_datatype, target_datatype)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",551,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_rget_any

! **************************************************************************************************
!> \brief Bcast function of encapsulated data.
!> \param base ...
!> \param source ...
!> \param grp ...
!> \param request ...
! **************************************************************************************************
   SUBROUTINE dbcsr_ibcast_any(base, source, grp, request)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: base
      INTEGER, INTENT(IN)                                :: source, grp
      INTEGER, INTENT(INOUT)                             :: request

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_ibcast_any', &
         routineP = moduleN//':'//routineN

      SELECT CASE (dbcsr_data_get_type (base))
      CASE (dbcsr_type_real_4)
         CALL mp_ibcast(base%d%r_sp, source, grp, request)
      CASE (dbcsr_type_real_8)
         CALL mp_ibcast(base%d%r_dp, source, grp, request)
      CASE (dbcsr_type_complex_4)
         CALL mp_ibcast(base%d%c_sp, source, grp, request)
      CASE (dbcsr_type_complex_8)
         CALL mp_ibcast(base%d%c_dp, source, grp, request)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",580,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_ibcast_any

! **************************************************************************************************
!> \brief Scatter function of encapsulated data.
!> \param base ...
!> \param counts ...
!> \param displs ...
!> \param msg ...
!> \param recvcount ...
!> \param root ...
!> \param grp ...
!> \param request ...
! **************************************************************************************************
   SUBROUTINE dbcsr_iscatterv_any(base, counts, displs, msg, recvcount, root, grp, request)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: base
      INTEGER, DIMENSION(:), INTENT(IN)                  :: counts, displs
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: msg
      INTEGER, INTENT(IN)                                :: recvcount, root, grp
      INTEGER, INTENT(INOUT)                             :: request

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_iscatterv_any', &
         routineP = moduleN//':'//routineN

      IF (dbcsr_data_get_type(base) .NE. dbcsr_data_get_type(msg)) &
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",606,"Different data type for msgin and msgout")

      SELECT CASE (dbcsr_data_get_type (base))
      CASE (dbcsr_type_real_4)
         CALL mp_iscatter(base%d%r_sp, counts, displs, msg%d%r_sp, recvcount, root, grp, request)
      CASE (dbcsr_type_real_8)
         CALL mp_iscatter(base%d%r_dp, counts, displs, msg%d%r_dp, recvcount, root, grp, request)
      CASE (dbcsr_type_complex_4)
         CALL mp_iscatter(base%d%c_sp, counts, displs, msg%d%c_sp, recvcount, root, grp, request)
      CASE (dbcsr_type_complex_8)
         CALL mp_iscatter(base%d%c_dp, counts, displs, msg%d%c_dp, recvcount, root, grp, request)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",618,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_iscatterv_any

! **************************************************************************************************
!> \brief Gather function of encapsulated data.
!> \param base ...
!> \param ub_base ...
!> \param msg ...
!> \param counts ...
!> \param displs ...
!> \param root ...
!> \param grp ...
! **************************************************************************************************
   SUBROUTINE dbcsr_gatherv_any(base, ub_base, msg, counts, displs, root, grp)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: base
      INTEGER, INTENT(IN)                                :: ub_base
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: msg
      INTEGER, DIMENSION(:), INTENT(IN)                  :: counts, displs
      INTEGER, INTENT(IN)                                :: root, grp

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_gatherv_any', &
         routineP = moduleN//':'//routineN

      IF (dbcsr_data_get_type(base) .NE. dbcsr_data_get_type(msg)) &
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",643,"Different data type for msgin and msgout")

      SELECT CASE (dbcsr_data_get_type (base))
      CASE (dbcsr_type_real_4)
         CALL mp_gatherv(base%d%r_sp(:ub_base), msg%d%r_sp, counts, displs, root, grp)
      CASE (dbcsr_type_real_8)
         CALL mp_gatherv(base%d%r_dp(:ub_base), msg%d%r_dp, counts, displs, root, grp)
      CASE (dbcsr_type_complex_4)
         CALL mp_gatherv(base%d%c_sp(:ub_base), msg%d%c_sp, counts, displs, root, grp)
      CASE (dbcsr_type_complex_8)
         CALL mp_gatherv(base%d%c_dp(:ub_base), msg%d%c_dp, counts, displs, root, grp)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",655,"Incorrect data type")
      END SELECT
   END SUBROUTINE dbcsr_gatherv_any

! **************************************************************************************************
!> \brief ...
!> \param count ...
!> \param lengths ...
!> \param displs ...
!> \param data_type ...
!> \retval new_mp_data_type ...
! **************************************************************************************************
   FUNCTION dbcsr_mp_type_indexed_make_any(count, lengths, displs, data_type) &
      RESULT(new_mp_data_type)
      INTEGER, INTENT(IN)                                :: count
      INTEGER, DIMENSION(1:count), INTENT(IN)            :: lengths, displs
      INTEGER, INTENT(IN)                                :: data_type
      TYPE(mp_type_descriptor_type)                      :: new_mp_data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mp_type_indexed_make_any', &
         routineP = moduleN//':'//routineN

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         new_mp_data_type = mp_type_indexed_make_r(count, lengths, displs)
      CASE (dbcsr_type_real_8)
         new_mp_data_type = mp_type_indexed_make_d(count, lengths, displs)
      CASE (dbcsr_type_complex_4)
         new_mp_data_type = mp_type_indexed_make_c(count, lengths, displs)
      CASE (dbcsr_type_complex_8)
         new_mp_data_type = mp_type_indexed_make_z(count, lengths, displs)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",687,"Invalid data type")
      END SELECT

   END FUNCTION dbcsr_mp_type_indexed_make_any

! **************************************************************************************************
!> \brief Allgather of encapsulated data
!> \param send_data ...
!> \param recv_data ...
!> \param recv_count ...
!> \param recv_displ ...
!> \param gid ...
!> \note see mp_allgatherv_dv
! **************************************************************************************************
   SUBROUTINE dbcsr_allgatherv(send_data, recv_data, recv_count, recv_displ, gid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: send_data
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: recv_data
      INTEGER, DIMENSION(:), INTENT(IN)                  :: recv_count, recv_displ
      INTEGER, INTENT(IN)                                :: gid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_allgatherv', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_assert(dbcsr_data_get_type(send_data), "EQ", &
                        dbcsr_data_get_type(recv_data), dbcsr_fatal_level, &
                        dbcsr_wrong_args_error, routineN, "Data type mismatch", 712)
      SELECT CASE (dbcsr_data_get_type (send_data))
      CASE (dbcsr_type_real_4)
         CALL mp_allgather(send_data%d%r_sp, recv_data%d%r_sp, &
                           recv_count, recv_displ, gid)
      CASE (dbcsr_type_real_8)
         CALL mp_allgather(send_data%d%r_dp, recv_data%d%r_dp, &
                           recv_count, recv_displ, gid)
      CASE (dbcsr_type_complex_4)
         CALL mp_allgather(send_data%d%c_sp, recv_data%d%c_sp, &
                           recv_count, recv_displ, gid)
      CASE (dbcsr_type_complex_8)
         CALL mp_allgather(send_data%d%c_dp, recv_data%d%c_dp, &
                           recv_count, recv_displ, gid)
      CASE default
         CALL cp__b("dbcsr/dist/dbcsr_mp_operations.F",727,"Invalid data type")
      END SELECT
   END SUBROUTINE dbcsr_allgatherv

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Row/column and global all-to-all
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param[in] mp_env         MP Environment
!> \param[in] most_ptp       (optional) Use point-to-point for row/column;
!>                           default is no
!> \param[in] remainder_ptp  (optional) Use point-to-point for remaining;
!>                           default is no
!> \param[in] no_hybrid      (optional) Use regular global collective; default
!>                           is no
!> \par Communicator selection
!>      Uses row and column communicators for row/column
!>      sends. Remaining sends are performed using the global
!>      communicator.  Point-to-point isend/irecv are used if ptp is
!>      set, otherwise a alltoall collective call is issued.
!>      see mp_alltoall 
! **************************************************************************************************
  SUBROUTINE hybrid_alltoall_d1 (sb, scount, sdispl,&
       rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(in), TARGET                     :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(INOUT), TARGET                  :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    TYPE(dbcsr_mp_obj), INTENT(IN)           :: mp_env
    LOGICAL, INTENT(in), OPTIONAL            :: most_ptp, remainder_ptp,&
                                                no_hybrid

    CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_d', &
      routineP = moduleN//':'//routineN

    INTEGER :: all_group, mynode, mypcol, myprow, nall_rr, nall_sr, ncol_rr, &
      ncol_sr, npcols, nprows, nrow_rr, nrow_sr, numnodes, dst, src,&
      prow, pcol, send_cnt, recv_cnt, tag, grp, i
    INTEGER, ALLOCATABLE, DIMENSION(:) :: all_rr, all_sr, col_rr, col_sr, &
      new_rcount, new_rdispl, new_scount, new_sdispl, row_rr, row_sr
    INTEGER, DIMENSION(:, :), POINTER        :: pgrid
    LOGICAL                                  :: most_collective, &
                                                remainder_collective, no_h
    REAL(kind=real_8), DIMENSION(:), POINTER           :: send_data_p, recv_data_p
    TYPE(dbcsr_mp_obj)                       :: mpe

    IF (.NOT. dbcsr_mp_has_subgroups (mp_env)) THEN
       mpe = mp_env
       CALL dbcsr_mp_grid_setup (mpe)
    ENDIF
    most_collective = .TRUE.
    remainder_collective = .TRUE.
    no_h = .FALSE.
    IF (PRESENT (most_ptp)) most_collective = .NOT. most_ptp
    IF (PRESENT (remainder_ptp)) remainder_collective = .NOT. remainder_ptp
    IF (PRESENT (no_hybrid)) no_h = no_hybrid
    all_group = dbcsr_mp_group (mp_env)
    ! Don't use subcommunicators if they're not defined.
    no_h = no_h .OR. .NOT. dbcsr_mp_has_subgroups (mp_env) .OR. .NOT. has_MPI
    subgrouped: IF (mp_env%mp%subgroups_defined .AND. .NOT. no_h) THEN
       mynode = dbcsr_mp_mynode (mp_env)
       numnodes = dbcsr_mp_numnodes (mp_env)
       nprows = dbcsr_mp_nprows (mp_env)
       npcols = dbcsr_mp_npcols (mp_env)
       myprow = dbcsr_mp_myprow (mp_env)
       mypcol = dbcsr_mp_mypcol (mp_env)
       pgrid => dbcsr_mp_pgrid (mp_env)
       ALLOCATE (row_sr(0:npcols-1)) ; nrow_sr = 0
       ALLOCATE (row_rr(0:npcols-1)) ; nrow_rr = 0
       ALLOCATE (col_sr(0:nprows-1)) ; ncol_sr = 0
       ALLOCATE (col_rr(0:nprows-1)) ; ncol_rr = 0
       ALLOCATE (all_sr(0:numnodes-1)) ; nall_sr = 0
       ALLOCATE (all_rr(0:numnodes-1)) ; nall_rr = 0
       ALLOCATE (new_scount(numnodes), new_rcount(numnodes))
       ALLOCATE (new_sdispl(numnodes), new_rdispl(numnodes))
       IF (.NOT.remainder_collective) THEN
          CALL remainder_point_to_point ()
       ENDIF
       IF (.NOT.most_collective) THEN
          CALL most_point_to_point ()
       ELSE
          CALL most_alltoall ()
       ENDIF
       IF (remainder_collective) THEN
          CALL remainder_alltoall ()
       ENDIF
       ! Wait for all issued sends and receives.
       IF (.NOT.most_collective) THEN
          CALL mp_waitall (row_sr(0:nrow_sr-1))
          CALL mp_waitall (col_sr(0:ncol_sr-1))
          CALL mp_waitall (row_rr(0:nrow_rr-1))
          CALL mp_waitall (col_rr(0:ncol_rr-1))
       END IF
       IF (.NOT.remainder_collective) THEN
          CALL mp_waitall (all_sr(1:nall_sr))
          CALL mp_waitall (all_rr(1:nall_rr))
       ENDIF
    ELSE
       CALL mp_alltoall (sb, scount, sdispl,&
            rb, rcount, rdispl,&
            all_group)
    ENDIF subgrouped
  CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_alltoall()
      DO pcol = 0 , npcols-1
         new_scount(1+pcol) = scount(1+pgrid(myprow, pcol))
         new_rcount(1+pcol) = rcount(1+pgrid(myprow, pcol))
         new_sdispl(1+pcol) = sdispl(1+pgrid(myprow, pcol))
         new_rdispl(1+pcol) = rdispl(1+pgrid(myprow, pcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:npcols), new_sdispl(1:npcols),&
           rb, new_rcount(1:npcols), new_rdispl(1:npcols),&
           dbcsr_mp_my_row_group (mp_env))
      DO prow = 0 , nprows-1
         new_scount(1+prow) = scount(1+pgrid(prow, mypcol))
         new_rcount(1+prow) = rcount(1+pgrid(prow, mypcol))
         new_sdispl(1+prow) = sdispl(1+pgrid(prow, mypcol))
         new_rdispl(1+prow) = rdispl(1+pgrid(prow, mypcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:nprows), new_sdispl(1:nprows),&
           rb, new_rcount(1:nprows), new_rdispl(1:nprows),&
           dbcsr_mp_my_col_group (mp_env))
    END SUBROUTINE most_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_point_to_point ()
      ! Go through my prow and exchange.
      DO i = 0, npcols - 1
         pcol = MOD (mypcol+i, npcols)
         grp = dbcsr_mp_my_row_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, myprow, pcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*mypcol
               CALL mp_isend (send_data_p, pcol, grp, row_sr(nrow_sr), tag)
               nrow_sr = nrow_sr+1
            ENDIF
         ENDIF
         !
         pcol = MODULO (mypcol-i, npcols)
         src = dbcsr_mp_get_process (mp_env, myprow, pcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*pcol
               CALL mp_irecv (recv_data_p, pcol, grp, row_rr(nrow_rr), tag)
               nrow_rr = nrow_rr+1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
      ! go through my pcol and exchange
      DO i = 0, nprows - 1
         prow = MOD (myprow+i, nprows)
         grp = dbcsr_mp_my_col_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, prow, mypcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*myprow+1
               CALL mp_isend (send_data_p, prow, grp, col_sr(ncol_sr), tag)
               ncol_sr = ncol_sr + 1
            ENDIF
         ENDIF
         !
         prow = MODULO (myprow-i, nprows)
         src = dbcsr_mp_get_process (mp_env, prow, mypcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*prow+1
               CALL mp_irecv (recv_data_p, prow, grp, col_rr(ncol_rr), tag)
               ncol_rr = ncol_rr + 1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
    END SUBROUTINE most_point_to_point
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_alltoall ()
      new_scount(:) = scount(:)
      new_rcount(:) = rcount(:)
      DO prow = 0,nprows-1
         new_scount(1+pgrid(prow, mypcol)) = 0
         new_rcount(1+pgrid(prow, mypcol)) = 0
      END DO
      DO pcol = 0,npcols-1
         new_scount(1+pgrid(myprow, pcol)) = 0
         new_rcount(1+pgrid(myprow, pcol)) = 0
      END DO
      CALL mp_alltoall (sb, new_scount, sdispl,&
           rb, new_rcount, rdispl, all_group)
    END SUBROUTINE remainder_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_point_to_point()
    INTEGER                                  :: col, row

      DO row = 0, nprows-1
         prow = MOD(row+myprow, nprows)
         IF (prow .EQ. myprow) CYCLE
         DO col = 0, npcols-1
            pcol = MOD (col+mypcol, npcols)
            IF (pcol .EQ. mypcol) CYCLE
            dst = dbcsr_mp_get_process (mp_env, prow, pcol)
            send_cnt = scount(dst+1)
            IF (send_cnt .GT. 0) THEN
               send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
               tag = 4*mynode+2
               CALL mp_isend (send_data_p, dst, all_group, all_sr(nall_sr+1), tag)
               nall_sr = nall_sr + 1
            ENDIF
            !
            src = dbcsr_mp_get_process (mp_env, prow, pcol)
            recv_cnt = rcount(src+1)
            IF (recv_cnt .GT. 0) THEN
               recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
               tag = 4*src+2
               CALL mp_irecv (recv_data_p, src, all_group, all_rr(nall_rr+1), tag)
               nall_rr = nall_rr+1
            ENDIF
         ENDDO
      ENDDO
    END SUBROUTINE remainder_point_to_point
  END SUBROUTINE hybrid_alltoall_d1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Row/column and global all-to-all
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param[in] mp_env         MP Environment
!> \param[in] most_ptp       (optional) Use point-to-point for row/column;
!>                           default is no
!> \param[in] remainder_ptp  (optional) Use point-to-point for remaining;
!>                           default is no
!> \param[in] no_hybrid      (optional) Use regular global collective; default
!>                           is no
!> \par Communicator selection
!>      Uses row and column communicators for row/column
!>      sends. Remaining sends are performed using the global
!>      communicator.  Point-to-point isend/irecv are used if ptp is
!>      set, otherwise a alltoall collective call is issued.
!>      see mp_alltoall 
! **************************************************************************************************
  SUBROUTINE hybrid_alltoall_z1 (sb, scount, sdispl,&
       rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(in), TARGET                     :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(INOUT), TARGET                  :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    TYPE(dbcsr_mp_obj), INTENT(IN)           :: mp_env
    LOGICAL, INTENT(in), OPTIONAL            :: most_ptp, remainder_ptp,&
                                                no_hybrid

    CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_z', &
      routineP = moduleN//':'//routineN

    INTEGER :: all_group, mynode, mypcol, myprow, nall_rr, nall_sr, ncol_rr, &
      ncol_sr, npcols, nprows, nrow_rr, nrow_sr, numnodes, dst, src,&
      prow, pcol, send_cnt, recv_cnt, tag, grp, i
    INTEGER, ALLOCATABLE, DIMENSION(:) :: all_rr, all_sr, col_rr, col_sr, &
      new_rcount, new_rdispl, new_scount, new_sdispl, row_rr, row_sr
    INTEGER, DIMENSION(:, :), POINTER        :: pgrid
    LOGICAL                                  :: most_collective, &
                                                remainder_collective, no_h
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: send_data_p, recv_data_p
    TYPE(dbcsr_mp_obj)                       :: mpe

    IF (.NOT. dbcsr_mp_has_subgroups (mp_env)) THEN
       mpe = mp_env
       CALL dbcsr_mp_grid_setup (mpe)
    ENDIF
    most_collective = .TRUE.
    remainder_collective = .TRUE.
    no_h = .FALSE.
    IF (PRESENT (most_ptp)) most_collective = .NOT. most_ptp
    IF (PRESENT (remainder_ptp)) remainder_collective = .NOT. remainder_ptp
    IF (PRESENT (no_hybrid)) no_h = no_hybrid
    all_group = dbcsr_mp_group (mp_env)
    ! Don't use subcommunicators if they're not defined.
    no_h = no_h .OR. .NOT. dbcsr_mp_has_subgroups (mp_env) .OR. .NOT. has_MPI
    subgrouped: IF (mp_env%mp%subgroups_defined .AND. .NOT. no_h) THEN
       mynode = dbcsr_mp_mynode (mp_env)
       numnodes = dbcsr_mp_numnodes (mp_env)
       nprows = dbcsr_mp_nprows (mp_env)
       npcols = dbcsr_mp_npcols (mp_env)
       myprow = dbcsr_mp_myprow (mp_env)
       mypcol = dbcsr_mp_mypcol (mp_env)
       pgrid => dbcsr_mp_pgrid (mp_env)
       ALLOCATE (row_sr(0:npcols-1)) ; nrow_sr = 0
       ALLOCATE (row_rr(0:npcols-1)) ; nrow_rr = 0
       ALLOCATE (col_sr(0:nprows-1)) ; ncol_sr = 0
       ALLOCATE (col_rr(0:nprows-1)) ; ncol_rr = 0
       ALLOCATE (all_sr(0:numnodes-1)) ; nall_sr = 0
       ALLOCATE (all_rr(0:numnodes-1)) ; nall_rr = 0
       ALLOCATE (new_scount(numnodes), new_rcount(numnodes))
       ALLOCATE (new_sdispl(numnodes), new_rdispl(numnodes))
       IF (.NOT.remainder_collective) THEN
          CALL remainder_point_to_point ()
       ENDIF
       IF (.NOT.most_collective) THEN
          CALL most_point_to_point ()
       ELSE
          CALL most_alltoall ()
       ENDIF
       IF (remainder_collective) THEN
          CALL remainder_alltoall ()
       ENDIF
       ! Wait for all issued sends and receives.
       IF (.NOT.most_collective) THEN
          CALL mp_waitall (row_sr(0:nrow_sr-1))
          CALL mp_waitall (col_sr(0:ncol_sr-1))
          CALL mp_waitall (row_rr(0:nrow_rr-1))
          CALL mp_waitall (col_rr(0:ncol_rr-1))
       END IF
       IF (.NOT.remainder_collective) THEN
          CALL mp_waitall (all_sr(1:nall_sr))
          CALL mp_waitall (all_rr(1:nall_rr))
       ENDIF
    ELSE
       CALL mp_alltoall (sb, scount, sdispl,&
            rb, rcount, rdispl,&
            all_group)
    ENDIF subgrouped
  CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_alltoall()
      DO pcol = 0 , npcols-1
         new_scount(1+pcol) = scount(1+pgrid(myprow, pcol))
         new_rcount(1+pcol) = rcount(1+pgrid(myprow, pcol))
         new_sdispl(1+pcol) = sdispl(1+pgrid(myprow, pcol))
         new_rdispl(1+pcol) = rdispl(1+pgrid(myprow, pcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:npcols), new_sdispl(1:npcols),&
           rb, new_rcount(1:npcols), new_rdispl(1:npcols),&
           dbcsr_mp_my_row_group (mp_env))
      DO prow = 0 , nprows-1
         new_scount(1+prow) = scount(1+pgrid(prow, mypcol))
         new_rcount(1+prow) = rcount(1+pgrid(prow, mypcol))
         new_sdispl(1+prow) = sdispl(1+pgrid(prow, mypcol))
         new_rdispl(1+prow) = rdispl(1+pgrid(prow, mypcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:nprows), new_sdispl(1:nprows),&
           rb, new_rcount(1:nprows), new_rdispl(1:nprows),&
           dbcsr_mp_my_col_group (mp_env))
    END SUBROUTINE most_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_point_to_point ()
      ! Go through my prow and exchange.
      DO i = 0, npcols - 1
         pcol = MOD (mypcol+i, npcols)
         grp = dbcsr_mp_my_row_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, myprow, pcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*mypcol
               CALL mp_isend (send_data_p, pcol, grp, row_sr(nrow_sr), tag)
               nrow_sr = nrow_sr+1
            ENDIF
         ENDIF
         !
         pcol = MODULO (mypcol-i, npcols)
         src = dbcsr_mp_get_process (mp_env, myprow, pcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*pcol
               CALL mp_irecv (recv_data_p, pcol, grp, row_rr(nrow_rr), tag)
               nrow_rr = nrow_rr+1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
      ! go through my pcol and exchange
      DO i = 0, nprows - 1
         prow = MOD (myprow+i, nprows)
         grp = dbcsr_mp_my_col_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, prow, mypcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*myprow+1
               CALL mp_isend (send_data_p, prow, grp, col_sr(ncol_sr), tag)
               ncol_sr = ncol_sr + 1
            ENDIF
         ENDIF
         !
         prow = MODULO (myprow-i, nprows)
         src = dbcsr_mp_get_process (mp_env, prow, mypcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*prow+1
               CALL mp_irecv (recv_data_p, prow, grp, col_rr(ncol_rr), tag)
               ncol_rr = ncol_rr + 1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
    END SUBROUTINE most_point_to_point
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_alltoall ()
      new_scount(:) = scount(:)
      new_rcount(:) = rcount(:)
      DO prow = 0,nprows-1
         new_scount(1+pgrid(prow, mypcol)) = 0
         new_rcount(1+pgrid(prow, mypcol)) = 0
      END DO
      DO pcol = 0,npcols-1
         new_scount(1+pgrid(myprow, pcol)) = 0
         new_rcount(1+pgrid(myprow, pcol)) = 0
      END DO
      CALL mp_alltoall (sb, new_scount, sdispl,&
           rb, new_rcount, rdispl, all_group)
    END SUBROUTINE remainder_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_point_to_point()
    INTEGER                                  :: col, row

      DO row = 0, nprows-1
         prow = MOD(row+myprow, nprows)
         IF (prow .EQ. myprow) CYCLE
         DO col = 0, npcols-1
            pcol = MOD (col+mypcol, npcols)
            IF (pcol .EQ. mypcol) CYCLE
            dst = dbcsr_mp_get_process (mp_env, prow, pcol)
            send_cnt = scount(dst+1)
            IF (send_cnt .GT. 0) THEN
               send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
               tag = 4*mynode+2
               CALL mp_isend (send_data_p, dst, all_group, all_sr(nall_sr+1), tag)
               nall_sr = nall_sr + 1
            ENDIF
            !
            src = dbcsr_mp_get_process (mp_env, prow, pcol)
            recv_cnt = rcount(src+1)
            IF (recv_cnt .GT. 0) THEN
               recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
               tag = 4*src+2
               CALL mp_irecv (recv_data_p, src, all_group, all_rr(nall_rr+1), tag)
               nall_rr = nall_rr+1
            ENDIF
         ENDDO
      ENDDO
    END SUBROUTINE remainder_point_to_point
  END SUBROUTINE hybrid_alltoall_z1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Row/column and global all-to-all
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param[in] mp_env         MP Environment
!> \param[in] most_ptp       (optional) Use point-to-point for row/column;
!>                           default is no
!> \param[in] remainder_ptp  (optional) Use point-to-point for remaining;
!>                           default is no
!> \param[in] no_hybrid      (optional) Use regular global collective; default
!>                           is no
!> \par Communicator selection
!>      Uses row and column communicators for row/column
!>      sends. Remaining sends are performed using the global
!>      communicator.  Point-to-point isend/irecv are used if ptp is
!>      set, otherwise a alltoall collective call is issued.
!>      see mp_alltoall 
! **************************************************************************************************
  SUBROUTINE hybrid_alltoall_s1 (sb, scount, sdispl,&
       rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(in), TARGET                     :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(INOUT), TARGET                  :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    TYPE(dbcsr_mp_obj), INTENT(IN)           :: mp_env
    LOGICAL, INTENT(in), OPTIONAL            :: most_ptp, remainder_ptp,&
                                                no_hybrid

    CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_s', &
      routineP = moduleN//':'//routineN

    INTEGER :: all_group, mynode, mypcol, myprow, nall_rr, nall_sr, ncol_rr, &
      ncol_sr, npcols, nprows, nrow_rr, nrow_sr, numnodes, dst, src,&
      prow, pcol, send_cnt, recv_cnt, tag, grp, i
    INTEGER, ALLOCATABLE, DIMENSION(:) :: all_rr, all_sr, col_rr, col_sr, &
      new_rcount, new_rdispl, new_scount, new_sdispl, row_rr, row_sr
    INTEGER, DIMENSION(:, :), POINTER        :: pgrid
    LOGICAL                                  :: most_collective, &
                                                remainder_collective, no_h
    REAL(kind=real_4), DIMENSION(:), POINTER           :: send_data_p, recv_data_p
    TYPE(dbcsr_mp_obj)                       :: mpe

    IF (.NOT. dbcsr_mp_has_subgroups (mp_env)) THEN
       mpe = mp_env
       CALL dbcsr_mp_grid_setup (mpe)
    ENDIF
    most_collective = .TRUE.
    remainder_collective = .TRUE.
    no_h = .FALSE.
    IF (PRESENT (most_ptp)) most_collective = .NOT. most_ptp
    IF (PRESENT (remainder_ptp)) remainder_collective = .NOT. remainder_ptp
    IF (PRESENT (no_hybrid)) no_h = no_hybrid
    all_group = dbcsr_mp_group (mp_env)
    ! Don't use subcommunicators if they're not defined.
    no_h = no_h .OR. .NOT. dbcsr_mp_has_subgroups (mp_env) .OR. .NOT. has_MPI
    subgrouped: IF (mp_env%mp%subgroups_defined .AND. .NOT. no_h) THEN
       mynode = dbcsr_mp_mynode (mp_env)
       numnodes = dbcsr_mp_numnodes (mp_env)
       nprows = dbcsr_mp_nprows (mp_env)
       npcols = dbcsr_mp_npcols (mp_env)
       myprow = dbcsr_mp_myprow (mp_env)
       mypcol = dbcsr_mp_mypcol (mp_env)
       pgrid => dbcsr_mp_pgrid (mp_env)
       ALLOCATE (row_sr(0:npcols-1)) ; nrow_sr = 0
       ALLOCATE (row_rr(0:npcols-1)) ; nrow_rr = 0
       ALLOCATE (col_sr(0:nprows-1)) ; ncol_sr = 0
       ALLOCATE (col_rr(0:nprows-1)) ; ncol_rr = 0
       ALLOCATE (all_sr(0:numnodes-1)) ; nall_sr = 0
       ALLOCATE (all_rr(0:numnodes-1)) ; nall_rr = 0
       ALLOCATE (new_scount(numnodes), new_rcount(numnodes))
       ALLOCATE (new_sdispl(numnodes), new_rdispl(numnodes))
       IF (.NOT.remainder_collective) THEN
          CALL remainder_point_to_point ()
       ENDIF
       IF (.NOT.most_collective) THEN
          CALL most_point_to_point ()
       ELSE
          CALL most_alltoall ()
       ENDIF
       IF (remainder_collective) THEN
          CALL remainder_alltoall ()
       ENDIF
       ! Wait for all issued sends and receives.
       IF (.NOT.most_collective) THEN
          CALL mp_waitall (row_sr(0:nrow_sr-1))
          CALL mp_waitall (col_sr(0:ncol_sr-1))
          CALL mp_waitall (row_rr(0:nrow_rr-1))
          CALL mp_waitall (col_rr(0:ncol_rr-1))
       END IF
       IF (.NOT.remainder_collective) THEN
          CALL mp_waitall (all_sr(1:nall_sr))
          CALL mp_waitall (all_rr(1:nall_rr))
       ENDIF
    ELSE
       CALL mp_alltoall (sb, scount, sdispl,&
            rb, rcount, rdispl,&
            all_group)
    ENDIF subgrouped
  CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_alltoall()
      DO pcol = 0 , npcols-1
         new_scount(1+pcol) = scount(1+pgrid(myprow, pcol))
         new_rcount(1+pcol) = rcount(1+pgrid(myprow, pcol))
         new_sdispl(1+pcol) = sdispl(1+pgrid(myprow, pcol))
         new_rdispl(1+pcol) = rdispl(1+pgrid(myprow, pcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:npcols), new_sdispl(1:npcols),&
           rb, new_rcount(1:npcols), new_rdispl(1:npcols),&
           dbcsr_mp_my_row_group (mp_env))
      DO prow = 0 , nprows-1
         new_scount(1+prow) = scount(1+pgrid(prow, mypcol))
         new_rcount(1+prow) = rcount(1+pgrid(prow, mypcol))
         new_sdispl(1+prow) = sdispl(1+pgrid(prow, mypcol))
         new_rdispl(1+prow) = rdispl(1+pgrid(prow, mypcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:nprows), new_sdispl(1:nprows),&
           rb, new_rcount(1:nprows), new_rdispl(1:nprows),&
           dbcsr_mp_my_col_group (mp_env))
    END SUBROUTINE most_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_point_to_point ()
      ! Go through my prow and exchange.
      DO i = 0, npcols - 1
         pcol = MOD (mypcol+i, npcols)
         grp = dbcsr_mp_my_row_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, myprow, pcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*mypcol
               CALL mp_isend (send_data_p, pcol, grp, row_sr(nrow_sr), tag)
               nrow_sr = nrow_sr+1
            ENDIF
         ENDIF
         !
         pcol = MODULO (mypcol-i, npcols)
         src = dbcsr_mp_get_process (mp_env, myprow, pcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*pcol
               CALL mp_irecv (recv_data_p, pcol, grp, row_rr(nrow_rr), tag)
               nrow_rr = nrow_rr+1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
      ! go through my pcol and exchange
      DO i = 0, nprows - 1
         prow = MOD (myprow+i, nprows)
         grp = dbcsr_mp_my_col_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, prow, mypcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*myprow+1
               CALL mp_isend (send_data_p, prow, grp, col_sr(ncol_sr), tag)
               ncol_sr = ncol_sr + 1
            ENDIF
         ENDIF
         !
         prow = MODULO (myprow-i, nprows)
         src = dbcsr_mp_get_process (mp_env, prow, mypcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*prow+1
               CALL mp_irecv (recv_data_p, prow, grp, col_rr(ncol_rr), tag)
               ncol_rr = ncol_rr + 1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
    END SUBROUTINE most_point_to_point
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_alltoall ()
      new_scount(:) = scount(:)
      new_rcount(:) = rcount(:)
      DO prow = 0,nprows-1
         new_scount(1+pgrid(prow, mypcol)) = 0
         new_rcount(1+pgrid(prow, mypcol)) = 0
      END DO
      DO pcol = 0,npcols-1
         new_scount(1+pgrid(myprow, pcol)) = 0
         new_rcount(1+pgrid(myprow, pcol)) = 0
      END DO
      CALL mp_alltoall (sb, new_scount, sdispl,&
           rb, new_rcount, rdispl, all_group)
    END SUBROUTINE remainder_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_point_to_point()
    INTEGER                                  :: col, row

      DO row = 0, nprows-1
         prow = MOD(row+myprow, nprows)
         IF (prow .EQ. myprow) CYCLE
         DO col = 0, npcols-1
            pcol = MOD (col+mypcol, npcols)
            IF (pcol .EQ. mypcol) CYCLE
            dst = dbcsr_mp_get_process (mp_env, prow, pcol)
            send_cnt = scount(dst+1)
            IF (send_cnt .GT. 0) THEN
               send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
               tag = 4*mynode+2
               CALL mp_isend (send_data_p, dst, all_group, all_sr(nall_sr+1), tag)
               nall_sr = nall_sr + 1
            ENDIF
            !
            src = dbcsr_mp_get_process (mp_env, prow, pcol)
            recv_cnt = rcount(src+1)
            IF (recv_cnt .GT. 0) THEN
               recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
               tag = 4*src+2
               CALL mp_irecv (recv_data_p, src, all_group, all_rr(nall_rr+1), tag)
               nall_rr = nall_rr+1
            ENDIF
         ENDDO
      ENDDO
    END SUBROUTINE remainder_point_to_point
  END SUBROUTINE hybrid_alltoall_s1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Row/column and global all-to-all
!> \param sb ...
!> \param scount ...
!> \param sdispl ...
!> \param rb ...
!> \param rcount ...
!> \param rdispl ...
!> \param[in] mp_env         MP Environment
!> \param[in] most_ptp       (optional) Use point-to-point for row/column;
!>                           default is no
!> \param[in] remainder_ptp  (optional) Use point-to-point for remaining;
!>                           default is no
!> \param[in] no_hybrid      (optional) Use regular global collective; default
!>                           is no
!> \par Communicator selection
!>      Uses row and column communicators for row/column
!>      sends. Remaining sends are performed using the global
!>      communicator.  Point-to-point isend/irecv are used if ptp is
!>      set, otherwise a alltoall collective call is issued.
!>      see mp_alltoall 
! **************************************************************************************************
  SUBROUTINE hybrid_alltoall_c1 (sb, scount, sdispl,&
       rb, rcount, rdispl, mp_env, most_ptp, remainder_ptp, no_hybrid)
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(in), TARGET                     :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(INOUT), TARGET                  :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    TYPE(dbcsr_mp_obj), INTENT(IN)           :: mp_env
    LOGICAL, INTENT(in), OPTIONAL            :: most_ptp, remainder_ptp,&
                                                no_hybrid

    CHARACTER(len=*), PARAMETER :: routineN = 'hybrid_alltoall_c', &
      routineP = moduleN//':'//routineN

    INTEGER :: all_group, mynode, mypcol, myprow, nall_rr, nall_sr, ncol_rr, &
      ncol_sr, npcols, nprows, nrow_rr, nrow_sr, numnodes, dst, src,&
      prow, pcol, send_cnt, recv_cnt, tag, grp, i
    INTEGER, ALLOCATABLE, DIMENSION(:) :: all_rr, all_sr, col_rr, col_sr, &
      new_rcount, new_rdispl, new_scount, new_sdispl, row_rr, row_sr
    INTEGER, DIMENSION(:, :), POINTER        :: pgrid
    LOGICAL                                  :: most_collective, &
                                                remainder_collective, no_h
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: send_data_p, recv_data_p
    TYPE(dbcsr_mp_obj)                       :: mpe

    IF (.NOT. dbcsr_mp_has_subgroups (mp_env)) THEN
       mpe = mp_env
       CALL dbcsr_mp_grid_setup (mpe)
    ENDIF
    most_collective = .TRUE.
    remainder_collective = .TRUE.
    no_h = .FALSE.
    IF (PRESENT (most_ptp)) most_collective = .NOT. most_ptp
    IF (PRESENT (remainder_ptp)) remainder_collective = .NOT. remainder_ptp
    IF (PRESENT (no_hybrid)) no_h = no_hybrid
    all_group = dbcsr_mp_group (mp_env)
    ! Don't use subcommunicators if they're not defined.
    no_h = no_h .OR. .NOT. dbcsr_mp_has_subgroups (mp_env) .OR. .NOT. has_MPI
    subgrouped: IF (mp_env%mp%subgroups_defined .AND. .NOT. no_h) THEN
       mynode = dbcsr_mp_mynode (mp_env)
       numnodes = dbcsr_mp_numnodes (mp_env)
       nprows = dbcsr_mp_nprows (mp_env)
       npcols = dbcsr_mp_npcols (mp_env)
       myprow = dbcsr_mp_myprow (mp_env)
       mypcol = dbcsr_mp_mypcol (mp_env)
       pgrid => dbcsr_mp_pgrid (mp_env)
       ALLOCATE (row_sr(0:npcols-1)) ; nrow_sr = 0
       ALLOCATE (row_rr(0:npcols-1)) ; nrow_rr = 0
       ALLOCATE (col_sr(0:nprows-1)) ; ncol_sr = 0
       ALLOCATE (col_rr(0:nprows-1)) ; ncol_rr = 0
       ALLOCATE (all_sr(0:numnodes-1)) ; nall_sr = 0
       ALLOCATE (all_rr(0:numnodes-1)) ; nall_rr = 0
       ALLOCATE (new_scount(numnodes), new_rcount(numnodes))
       ALLOCATE (new_sdispl(numnodes), new_rdispl(numnodes))
       IF (.NOT.remainder_collective) THEN
          CALL remainder_point_to_point ()
       ENDIF
       IF (.NOT.most_collective) THEN
          CALL most_point_to_point ()
       ELSE
          CALL most_alltoall ()
       ENDIF
       IF (remainder_collective) THEN
          CALL remainder_alltoall ()
       ENDIF
       ! Wait for all issued sends and receives.
       IF (.NOT.most_collective) THEN
          CALL mp_waitall (row_sr(0:nrow_sr-1))
          CALL mp_waitall (col_sr(0:ncol_sr-1))
          CALL mp_waitall (row_rr(0:nrow_rr-1))
          CALL mp_waitall (col_rr(0:ncol_rr-1))
       END IF
       IF (.NOT.remainder_collective) THEN
          CALL mp_waitall (all_sr(1:nall_sr))
          CALL mp_waitall (all_rr(1:nall_rr))
       ENDIF
    ELSE
       CALL mp_alltoall (sb, scount, sdispl,&
            rb, rcount, rdispl,&
            all_group)
    ENDIF subgrouped
  CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_alltoall()
      DO pcol = 0 , npcols-1
         new_scount(1+pcol) = scount(1+pgrid(myprow, pcol))
         new_rcount(1+pcol) = rcount(1+pgrid(myprow, pcol))
         new_sdispl(1+pcol) = sdispl(1+pgrid(myprow, pcol))
         new_rdispl(1+pcol) = rdispl(1+pgrid(myprow, pcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:npcols), new_sdispl(1:npcols),&
           rb, new_rcount(1:npcols), new_rdispl(1:npcols),&
           dbcsr_mp_my_row_group (mp_env))
      DO prow = 0 , nprows-1
         new_scount(1+prow) = scount(1+pgrid(prow, mypcol))
         new_rcount(1+prow) = rcount(1+pgrid(prow, mypcol))
         new_sdispl(1+prow) = sdispl(1+pgrid(prow, mypcol))
         new_rdispl(1+prow) = rdispl(1+pgrid(prow, mypcol))
      END DO
      CALL mp_alltoall (sb, new_scount(1:nprows), new_sdispl(1:nprows),&
           rb, new_rcount(1:nprows), new_rdispl(1:nprows),&
           dbcsr_mp_my_col_group (mp_env))
    END SUBROUTINE most_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE most_point_to_point ()
      ! Go through my prow and exchange.
      DO i = 0, npcols - 1
         pcol = MOD (mypcol+i, npcols)
         grp = dbcsr_mp_my_row_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, myprow, pcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*mypcol
               CALL mp_isend (send_data_p, pcol, grp, row_sr(nrow_sr), tag)
               nrow_sr = nrow_sr+1
            ENDIF
         ENDIF
         !
         pcol = MODULO (mypcol-i, npcols)
         src = dbcsr_mp_get_process (mp_env, myprow, pcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (pcol.NE.mypcol) THEN
               tag = 4*pcol
               CALL mp_irecv (recv_data_p, pcol, grp, row_rr(nrow_rr), tag)
               nrow_rr = nrow_rr+1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
      ! go through my pcol and exchange
      DO i = 0, nprows - 1
         prow = MOD (myprow+i, nprows)
         grp = dbcsr_mp_my_col_group (mp_env)
         !
         dst = dbcsr_mp_get_process (mp_env, prow, mypcol)
         send_cnt = scount(dst+1)
         IF (send_cnt .GT. 0) THEN
            send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*myprow+1
               CALL mp_isend (send_data_p, prow, grp, col_sr(ncol_sr), tag)
               ncol_sr = ncol_sr + 1
            ENDIF
         ENDIF
         !
         prow = MODULO (myprow-i, nprows)
         src = dbcsr_mp_get_process (mp_env, prow, mypcol)
         recv_cnt = rcount(src+1)
         IF (recv_cnt .GT. 0) THEN
            recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
            IF (prow.NE.myprow) THEN
               tag = 4*prow+1
               CALL mp_irecv (recv_data_p, prow, grp, col_rr(ncol_rr), tag)
               ncol_rr = ncol_rr + 1
            ELSE
               CALL memory_copy(recv_data_p,send_data_p,recv_cnt)
            ENDIF
         ENDIF
      ENDDO
    END SUBROUTINE most_point_to_point
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_alltoall ()
      new_scount(:) = scount(:)
      new_rcount(:) = rcount(:)
      DO prow = 0,nprows-1
         new_scount(1+pgrid(prow, mypcol)) = 0
         new_rcount(1+pgrid(prow, mypcol)) = 0
      END DO
      DO pcol = 0,npcols-1
         new_scount(1+pgrid(myprow, pcol)) = 0
         new_rcount(1+pgrid(myprow, pcol)) = 0
      END DO
      CALL mp_alltoall (sb, new_scount, sdispl,&
           rb, new_rcount, rdispl, all_group)
    END SUBROUTINE remainder_alltoall
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE remainder_point_to_point()
    INTEGER                                  :: col, row

      DO row = 0, nprows-1
         prow = MOD(row+myprow, nprows)
         IF (prow .EQ. myprow) CYCLE
         DO col = 0, npcols-1
            pcol = MOD (col+mypcol, npcols)
            IF (pcol .EQ. mypcol) CYCLE
            dst = dbcsr_mp_get_process (mp_env, prow, pcol)
            send_cnt = scount(dst+1)
            IF (send_cnt .GT. 0) THEN
               send_data_p => sb( 1+sdispl(dst+1) : 1+sdispl(dst+1)+send_cnt-1 )
               tag = 4*mynode+2
               CALL mp_isend (send_data_p, dst, all_group, all_sr(nall_sr+1), tag)
               nall_sr = nall_sr + 1
            ENDIF
            !
            src = dbcsr_mp_get_process (mp_env, prow, pcol)
            recv_cnt = rcount(src+1)
            IF (recv_cnt .GT. 0) THEN
               recv_data_p => rb( 1+rdispl(src+1) : 1+rdispl(src+1)+recv_cnt-1 )
               tag = 4*src+2
               CALL mp_irecv (recv_data_p, src, all_group, all_rr(nall_rr+1), tag)
               nall_rr = nall_rr+1
            ENDIF
         ENDDO
      ENDDO
    END SUBROUTINE remainder_point_to_point
  END SUBROUTINE hybrid_alltoall_c1

END MODULE dbcsr_mp_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR transformations
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util and dbcsr_operations
! **************************************************************************************************
MODULE dbcsr_transformations

   USE array_types,                     ONLY: array_data,&
                                              array_hold,&
                                              array_i1d_obj,&
                                              array_release
   USE dbcsr_block_access,              ONLY: dbcsr_get_block_p,&
                                              dbcsr_put_block
   USE dbcsr_block_operations,          ONLY: dbcsr_block_conjg,&
                                              dbcsr_block_partial_copy,&
                                              dbcsr_block_real_neg,&
                                              dbcsr_block_transpose,&
                                              dbcsr_data_clear,&
                                              dbcsr_data_copy,&
                                              dbcsr_data_set
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_ensure_size, dbcsr_data_get_memory_type, &
        dbcsr_data_get_size, dbcsr_data_get_size_referenced, dbcsr_data_get_type, dbcsr_data_hold, &
        dbcsr_data_init, dbcsr_data_new, dbcsr_data_release, dbcsr_data_set_pointer, &
        dbcsr_data_set_size_referenced, dbcsr_get_data, dbcsr_type_1d_to_2d
   USE dbcsr_data_operations,           ONLY: dbcsr_copy_sort_data,&
                                              dbcsr_switch_data_area
   USE dbcsr_dist_methods,              ONLY: &
        dbcsr_distribution_col_dist, dbcsr_distribution_hold, dbcsr_distribution_local_cols, &
        dbcsr_distribution_local_rows, dbcsr_distribution_mp, dbcsr_distribution_ncols, &
        dbcsr_distribution_nlocal_cols, dbcsr_distribution_nlocal_rows, dbcsr_distribution_nrows, &
        dbcsr_distribution_release, dbcsr_distribution_row_dist
   USE dbcsr_dist_operations,           ONLY: dbcsr_get_local_cols,&
                                              dbcsr_get_local_rows,&
                                              dbcsr_get_stored_coordinates,&
                                              dbcsr_reblocking_targets,&
                                              dbcsr_transpose_dims,&
                                              dbcsr_transpose_distribution
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_failure_level,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_unimplemented_error_nr,&
                                              dbcsr_wrong_args_error
   USE dbcsr_index_operations,          ONLY: dbcsr_addto_index_array,&
                                              dbcsr_clearfrom_index_array,&
                                              dbcsr_repoint_index,&
                                              make_dense_index,&
                                              make_undense_index,&
                                              transpose_index_local
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_type, dbcsr_get_matrix_type, &
        dbcsr_has_symmetry, dbcsr_init, dbcsr_is_initialized, dbcsr_nblkcols_total, &
        dbcsr_nblkrows_total, dbcsr_nfullcols_total, dbcsr_nfullrows_total, dbcsr_release, &
        dbcsr_row_block_sizes, dbcsr_valid_index
   USE dbcsr_mp_methods,                ONLY: &
        dbcsr_mp_grid_remove, dbcsr_mp_grid_setup, dbcsr_mp_group, dbcsr_mp_has_subgroups, &
        dbcsr_mp_my_col_group, dbcsr_mp_my_row_group, dbcsr_mp_mynode, dbcsr_mp_mypcol, &
        dbcsr_mp_myprow, dbcsr_mp_npcols, dbcsr_mp_nprows, dbcsr_mp_numnodes, dbcsr_mp_pgrid
   USE dbcsr_mp_operations,             ONLY: dbcsr_allgatherv,&
                                              hybrid_alltoall_any,&
                                              hybrid_alltoall_c1,&
                                              hybrid_alltoall_d1,&
                                              hybrid_alltoall_i1,&
                                              hybrid_alltoall_s1,&
                                              hybrid_alltoall_z1
   USE dbcsr_ptr_util,                  ONLY: pointer_view
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_iterator, dbcsr_meta_size, dbcsr_mp_obj, &
        dbcsr_obj, dbcsr_repl_col, dbcsr_repl_full, dbcsr_repl_none, dbcsr_repl_row, &
        dbcsr_slot_blk_p, dbcsr_slot_col_i, dbcsr_slot_home_pcol, dbcsr_slot_home_prow, &
        dbcsr_slot_nblks, dbcsr_slot_nze, dbcsr_slot_row_p, dbcsr_type_complex_4, &
        dbcsr_type_complex_8, dbcsr_type_no_symmetry, dbcsr_type_real_4, dbcsr_type_real_8
   USE dbcsr_util,                      ONLY: convert_sizes_to_offsets,&
                                              dbcsr_checksum,&
                                              dbcsr_pack_meta,&
                                              dbcsr_unpack_meta,&
                                              dbcsr_verify_matrix,&
                                              get_internal_offsets,&
                                              global_offsets_to_local,&
                                              nfull_elements
   USE dbcsr_work_operations,           ONLY: dbcsr_create,&
                                              dbcsr_finalize,&
                                              dbcsr_work_create
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              sp
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_alltoall
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

  IMPLICIT NONE
  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_transformations'

  LOGICAL, PARAMETER, PRIVATE          :: careful_mod = .FALSE.

  PUBLIC :: dbcsr_desymmetrize_deep,&
            dbcsr_new_transposed,&
            dbcsr_complete_redistribute,&
            dbcsr_redistribute, dbcsr_make_untransposed_blocks
  PUBLIC :: dbcsr_replicate_all, dbcsr_distribute, dbcsr_datablock_redistribute
  PUBLIC :: dbcsr_make_dense, dbcsr_make_undense, dbcsr_make_dense_low

CONTAINS


! **************************************************************************************************
!> \brief Transposes a DBCSR matrix.
!> \param[out] transposed     transposed DBCSR matrix
!> \param[in] normal          input DBCSR matrix
!> \param[in] shallow_data_copy         (optional) only shallow data_copy;
!>                                      default is no; if set, the
!>                                      transpose_data option is ignored
!> \param[in] transpose_data  (optional) transpose data blocks, default is True
!> \param[in] transpose_distribution    (optional) transpose the distribution
!>                                      from the input matrix, default is True
!> \param[in] transpose_index           (optional) transpose the index
!>                                      (default=yes) or turn it into BCSC
!> \param[in] use_distribution          (optional) use this distribution
!> \param[in] redistribute              (optional) redistributes the matrix;
!>                                      default is .TRUE. unless shallow or
!>                                      transpose_distribution are set.
!> \par Distribution options
!>      By default the distribution is transposed. If transpose_distribution
!>      is false, then an undetermined distribution is created that is
!>      compatible with the same process grid.
! **************************************************************************************************
  SUBROUTINE dbcsr_new_transposed (transposed, normal, shallow_data_copy,&
       transpose_data, transpose_distribution, transpose_index,&
       use_distribution, redistribute)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: transposed
      TYPE(dbcsr_obj), INTENT(IN)                        :: normal
      LOGICAL, INTENT(IN), OPTIONAL                      :: shallow_data_copy, transpose_data, &
                                                            transpose_distribution, transpose_index
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: use_distribution
      LOGICAL, INTENT(IN), OPTIONAL                      :: redistribute

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_new_transposed', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: blk_p
      LOGICAL                                            :: redist, shallow, tr_blocks, tr_dist, &
                                                            tr_index
      TYPE(dbcsr_distribution_obj)                       :: new_dist
      TYPE(dbcsr_obj)                                    :: t2

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)

    IF(.NOT.dbcsr_valid_index(normal))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",171,"Matrix does not exist.")
    ! Internalize options
    shallow = .FALSE.
    IF (PRESENT (shallow_data_copy)) shallow = shallow_data_copy
    tr_blocks = .TRUE.
    IF (PRESENT (transpose_data)) tr_blocks = transpose_data
    tr_dist = .TRUE.
    IF (PRESENT (transpose_distribution)) tr_dist = transpose_distribution
    tr_index = .TRUE.
    IF (PRESENT (transpose_index)) tr_index = transpose_index
    ! Prepare the distribution for the transposed matrix
    IF (PRESENT (use_distribution)) THEN
       CALL dbcsr_assert (dbcsr_distribution_nrows (use_distribution),&
            "EQ", dbcsr_distribution_ncols (normal%m%dist), dbcsr_fatal_level,&
            dbcsr_wrong_args_error, routineN,&
            "Given distribution must be compatible"&
            //" with the current distribution",187)
       CALL dbcsr_assert (dbcsr_distribution_ncols (use_distribution),&
            "EQ", dbcsr_distribution_nrows (normal%m%dist), dbcsr_fatal_level,&
            dbcsr_wrong_args_error, routineN,&
            "Given distribution must be compatible"&
            //" with the current distribution",192)
       new_dist = use_distribution
       CALL dbcsr_distribution_hold (new_dist)
    ELSE
       IF (tr_dist) THEN
          CALL dbcsr_transpose_distribution (new_dist, normal%m%dist)
       ELSE
          CALL dbcsr_transpose_dims (new_dist, normal%m%dist)
       ENDIF
    ENDIF
    ! Create the transposed matrix
    CALL dbcsr_create (transposed, name="transposed "//TRIM(normal%m%name),&
         template = normal, &
         dist = new_dist,&
         matrix_type = dbcsr_get_matrix_type(normal),&
         row_blk_size = array_data(normal%m%col_blk_size),&
         col_blk_size = array_data(normal%m%row_blk_size),&
         max_rbs=normal%m%max_cbs,&
         max_cbs=normal%m%max_rbs,&
         row_blk_offset=normal%m%col_blk_offset,&
         col_blk_offset=normal%m%row_blk_offset)
    CALL dbcsr_distribution_release (new_dist)
    ! Reserve the space for the new indices.
    IF (tr_index) THEN
       CALL dbcsr_addto_index_array (transposed%m, dbcsr_slot_row_p,&
            reservation=transposed%m%nblkrows_total+1, extra=transposed%m%nblks*2)
    ELSE
       CALL dbcsr_addto_index_array (transposed%m, dbcsr_slot_row_p,&
            reservation=normal%m%nblkrows_total+1, extra=transposed%m%nblks*2)
    ENDIF
    CALL dbcsr_addto_index_array (transposed%m, dbcsr_slot_col_i,&
         reservation=normal%m%nblks)
    CALL dbcsr_addto_index_array (transposed%m, dbcsr_slot_blk_p,&
         reservation=normal%m%nblks)
    CALL dbcsr_repoint_index (transposed%m)
    IF (.NOT. shallow) THEN
       CALL dbcsr_data_ensure_size (transposed%m%data_area,&
            dbcsr_data_get_size_referenced(normal%m%data_area),&
            nocopy=.TRUE.)
    ENDIF
    !
    transposed%m%nblks = normal%m%nblks
    transposed%m%nze = normal%m%nze
    transposed%m%index(dbcsr_slot_nblks) = normal%m%nblks
    transposed%m%index(dbcsr_slot_nze) = normal%m%nze
    ! Transpose the local index.
    ALLOCATE (blk_p (normal%m%nblks), stat=stat)
    IF(stat /= 0) CALL cp__b("dbcsr/ops/dbcsr_transformations.F",239,"blk_p")
    IF (tr_index) THEN
       CALL transpose_index_local (transposed%m%row_p, transposed%m%col_i,&
            normal%m%row_p, normal%m%col_i, blk_p, normal%m%blk_p)
       IF (dbg) THEN
          WRITE(*,*)'orig. row_p',normal%m%row_p
          WRITE(*,*)'orig. col_i',normal%m%col_i
          WRITE(*,*)'orig. blk_p',normal%m%blk_p
          WRITE(*,*)'new . row_p',transposed%m%row_p
          WRITE(*,*)'new . col_i',transposed%m%col_i
          WRITE(*,*)'new . blk_p',blk_p!transposed%m%blk_p
       ENDIF
    ELSE
       transposed%m%row_p(:) = normal%m%row_p(:)
       transposed%m%col_i(:) = normal%m%col_i(:)
       blk_p(:) = normal%m%blk_p(:)
       !transposed%m%transpose = .TRUE.
    ENDIF
    ! Copy the data
    IF (shallow) THEN
       CALL dbcsr_switch_data_area (transposed, normal%m%data_area)
       transposed%m%blk_p(1:transposed%m%nblks) =&
            -blk_p(1:transposed%m%nblks)
    ELSE
       CALL dbcsr_copy_sort_data (transposed%m%blk_p, blk_p, transposed%m%row_p,&
            transposed%m%col_i, array_data (transposed%m%row_blk_size),&
            array_data (transposed%m%col_blk_size),&
            transposed%m%data_area, normal%m%data_area,&
            mark_transposed=.not.tr_blocks,&
            transpose_blocks=tr_blocks)
    ENDIF
    transposed%m%valid = .TRUE.
    !CALL dbcsr_copy_sort_data (transposed%m%blk_p, blk_p, transposed%m%row_p,&
    !     transposed%m%col_i, array_data (transposed%m%row_blk_size),&
    !     array_data (transposed%m%col_blk_size),&
    !     transposed%m%data_area, normal%m%data_area,&
    !     transpose_blocks=.TRUE.)
    !
1315 FORMAT (I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5)
    IF (dbg) THEN
       WRITE(*,*)'new FINAL index'
       WRITE(*,1315)transposed%m%row_p
       WRITE(*,1315)transposed%m%col_i
       WRITE(*,1315)transposed%m%blk_p
    ENDIF
    !
    IF (tr_index) DEALLOCATE (blk_p)
    !
    IF (PRESENT (redistribute)) THEN
       redist = redistribute
    ELSE
       redist = .NOT. tr_dist .AND. .NOT. shallow
    ENDIF
    IF (redist) THEN
       !write (*,*)routineN//" redistributing"
       CALL dbcsr_init (t2)
       CALL dbcsr_create (t2, template=transposed)
       CALL dbcsr_redistribute (transposed, t2)
       CALL dbcsr_release (transposed)
       transposed = t2
    ENDIF
    CALL timestop(handle)
  END SUBROUTINE dbcsr_new_transposed

! **************************************************************************************************
!> \brief Duplicates data in symmetric matrix to make it normal (w.r.t. data
!>        structure. Can handle symmetric/antisymmetric/hermitian/skewhermitian
!>        matrices
!> \param[in] sm              input symmetric matrix
!> \param[out] desm           desymmetrized matrix
!> \param untransposed_data ...
! **************************************************************************************************
  SUBROUTINE dbcsr_desymmetrize_deep(sm, desm, untransposed_data)
      TYPE(dbcsr_obj), INTENT(IN)                        :: sm
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: desm
      LOGICAL, INTENT(IN), OPTIONAL                      :: untransposed_data

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_desymmetrize_deep', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: idata = 2, imeta = 1, metalen = 3

      INTEGER :: blk, blk_l, blk_p, blk_ps, blks, col, col_size, dst_p, handle, mp_group, &
         nsymmetries, numproc, nze, pcol, prow, row, row_size, src_p, stored_col, stored_row, &
         symmetry_i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rd_disp, recv_meta, rm_disp, sd_disp, &
                                                            sdp, send_meta, sm_disp, smp
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: recv_count, send_count, &
                                                            total_recv_count, total_send_count
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_dist, row_blk_size, &
                                                            row_dist
      INTEGER, DIMENSION(:, :), POINTER                  :: blacs2mpi
      LOGICAL                                            :: make_untr, tr
      TYPE(dbcsr_data_obj)                               :: recv_data, send_data
      TYPE(dbcsr_distribution_obj)                       :: target_dist
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF(.NOT.dbcsr_valid_index (sm))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",340,"Input matrix is invalid.")
    IF(dbcsr_valid_index (desm)) THEN
       CALL dbcsr_release(desm)
       CALL dbcsr_init(desm)
    ENDIF
    CALL dbcsr_create(desm, sm, matrix_type="N")

    IF (PRESENT (untransposed_data)) THEN
       make_untr = untransposed_data
    ELSE
       make_untr = .FALSE.
    ENDIF
    nsymmetries = 1
    IF (sm%m%symmetry) THEN
       nsymmetries = 2
    ENDIF
    row_blk_size => array_data (sm%m%row_blk_size)
    col_blk_size => array_data (sm%m%col_blk_size)
    target_dist = sm%m%dist
    row_dist => dbcsr_distribution_row_dist (target_dist)
    col_dist => dbcsr_distribution_col_dist (target_dist)
    mp_obj = dbcsr_distribution_mp (target_dist)
    blacs2mpi => dbcsr_mp_pgrid (mp_obj)
    numproc = dbcsr_mp_numnodes (mp_obj)
    mp_group = dbcsr_mp_group (mp_obj)
    !
    ALLOCATE (send_count(2, 0:numproc-1))
    ALLOCATE (recv_count(2, 0:numproc-1))
    ALLOCATE (total_send_count(2, 0:numproc-1))
    ALLOCATE (total_recv_count(2, 0:numproc-1))
    ALLOCATE (sdp(0:numproc-1))
    ALLOCATE (sd_disp(0:numproc-1))
    ALLOCATE (smp(0:numproc-1))
    ALLOCATE (sm_disp(0:numproc-1))
    ALLOCATE (rd_disp(0:numproc-1))
    ALLOCATE (rm_disp(0:numproc-1))
    !
    desm%m%negate_real = sm%m%negate_real
    desm%m%negate_imaginary = sm%m%negate_imaginary
    ! Count initial sizes for sending.
    send_count(:,:) = 0
    CALL dbcsr_iterator_start (iter, sm)
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, blk,&
            row_size=row_size, col_size=col_size)
       DO symmetry_i = 1, nsymmetries
          IF (symmetry_i .EQ. 1) THEN
             stored_row = row ; stored_col = col
          ELSE
             IF (row .EQ. col) CYCLE
             stored_row = col ; stored_col = row
          ENDIF
          ! Where do we send this block?
          prow = row_dist(stored_row)
          pcol = col_dist(stored_col)
          dst_p = blacs2mpi(prow, pcol)
          nze = row_size * col_size
          send_count(imeta, dst_p) = send_count(imeta, dst_p) + 1
          send_count(idata, dst_p) = send_count(idata, dst_p) + nze
       ENDDO ! symmetry_i
    ENDDO ! iter
    CALL dbcsr_iterator_stop (iter)
    !
    CALL mp_alltoall(send_count, recv_count, 2, mp_group)
    !
    ! Allocate data structures needed for data exchange.
    CALL dbcsr_data_init (recv_data)
    CALL dbcsr_data_new (recv_data, dbcsr_get_data_type (sm),&
         SUM(recv_count(idata, :)))
    ALLOCATE (recv_meta(metalen*SUM(recv_count(imeta, :))))
    CALL dbcsr_data_init (send_data)
    CALL dbcsr_data_new (send_data, dbcsr_get_data_type (sm),&
         SUM(send_count(idata, :)))
    ALLOCATE (send_meta(metalen*SUM(send_count(imeta, :))))
    !
    ! Fill in the meta data structures and copy the data.
    DO dst_p = 0, numproc-1
       total_send_count(imeta, dst_p) = send_count (imeta, dst_p)
       total_send_count(idata, dst_p) = send_count (idata, dst_p)
       total_recv_count(imeta, dst_p) = recv_count (imeta, dst_p)
       total_recv_count(idata, dst_p) = recv_count (idata, dst_p)
    ENDDO
    sd_disp = -1 ; sm_disp = -1
    rd_disp = -1 ; rm_disp = -1
    sd_disp(0) = 1 ; sm_disp(0) = 1
    rd_disp(0) = 1 ; rm_disp(0) = 1
    DO dst_p = 1, numproc-1
       sm_disp(dst_p) = sm_disp(dst_p-1)&
            + metalen*total_send_count(imeta, dst_p-1)
       sd_disp(dst_p) = sd_disp(dst_p-1)&
            + total_send_count(idata, dst_p-1)
       rm_disp(dst_p) = rm_disp(dst_p-1)&
            + metalen*total_recv_count(imeta, dst_p-1)
       rd_disp(dst_p) = rd_disp(dst_p-1)&
            + total_recv_count(idata, dst_p-1)
    ENDDO
    sdp(:) = sd_disp
    smp(:) = sm_disp
    !
    CALL dbcsr_iterator_start (iter, sm)
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, blk, blk_p=blk_p,&
            row_size=row_size, col_size=col_size)
       DO symmetry_i = 1, nsymmetries
          IF (symmetry_i .EQ. 1) THEN
             stored_row = row ; stored_col = col; tr = .FALSE.
          ELSE
             IF (row .EQ. col) CYCLE
             stored_row = col ; stored_col = row; tr = .TRUE.
          ENDIF
          ! Where do we send this block?
          prow = row_dist(stored_row)
          pcol = col_dist(stored_col)
          dst_p = blacs2mpi(prow, pcol)
          nze = row_size * col_size
          send_meta(smp(dst_p)) = stored_row
          send_meta(smp(dst_p)+1) = stored_col
          send_meta(smp(dst_p)+2) = sdp(dst_p) - sd_disp(dst_p) + 1
          IF (make_untr) THEN
             CALL dbcsr_block_partial_copy(dst=send_data,&
                  dst_rs=row_size, dst_cs=col_size,&
                  dst_tr=symmetry_i .GT. 1,&
                  dst_r_lb=1, dst_c_lb=1, dst_offset=sdp(dst_p)-1,&
                  nrow = row_size, ncol = col_size,&
                  src=sm%m%data_area,&
                  src_rs=row_size, src_cs=col_size,&
                  src_tr=blk_p .LT. 0,&
                  src_r_lb=1, src_c_lb=1,&
                  src_offset = ABS(blk_p)-1)
             IF(tr)THEN
               IF(sm%m%negate_real)THEN
                  CALL dbcsr_block_real_neg(dst=send_data,row_size=row_size,&
                                            col_size=col_size,lb=sdp(dst_p)) 
               END IF
               IF(sm%m%negate_imaginary)&
                  CALL dbcsr_block_conjg(dst=send_data,row_size=row_size,&
                                            col_size=col_size,lb=sdp(dst_p))
             END IF
          ELSE
             CALL dbcsr_data_copy (send_data, (/ sdp(dst_p) /), (/ nze /),&
                  sm%m%data_area, (/ ABS(blk_p) /), (/ nze /))
             IF (tr) &
                  send_meta(smp(dst_p)+2) = -send_meta(smp(dst_p)+2)
          ENDIF
          smp(dst_p) = smp(dst_p) + metalen
          sdp(dst_p) = sdp(dst_p) + nze
       ENDDO ! symmetry_i
    ENDDO ! iter
    CALL dbcsr_iterator_stop (iter)
    ! Exchange the data and metadata structures.
    CALL hybrid_alltoall_any(send_data,&
         total_send_count(idata,:), sd_disp(:)-1,&
         recv_data,&
         total_recv_count(idata,:), rd_disp(:)-1, mp_obj)
    CALL mp_alltoall(send_meta(:), metalen*total_send_count(imeta,:), sm_disp(:)-1,&
         recv_meta(:), metalen*total_recv_count(imeta,:), rm_disp(:)-1,&
         mp_group)
    ! Now fill in the data.
    CALL dbcsr_work_create(desm,&
         SUM(recv_count(imeta,:)),&
         SUM(recv_count(idata,:)), n=1,&
         work_mutable=.FALSE.)
    ! Switch data data area of the work matrix with the receieved data
    ! (avoids copying).
    CALL dbcsr_data_hold (recv_data)
    CALL dbcsr_data_release (desm%m%wms(1)%data_area)
    desm%m%wms(1)%data_area = recv_data
    !
    blk_ps = 1
    blks = 1
    ! WRITE(*,*)rm_disp
    ! WRITE(*,*)recv_count
    DO src_p = 0, numproc-1
       IF (careful_mod) THEN
          CALL dbcsr_assert ((blks-1)*3, "EQ", rm_disp(src_p)-1,&
               dbcsr_fatal_level, dbcsr_internal_error, routineN,&
               "Count mismatch", 516)
       ENDIF
       blks = (rm_disp(src_p)-1)/metalen+1
       DO blk_l = 1, recv_count(imeta, src_p)
          stored_row = recv_meta(rm_disp(src_p)+metalen*(blk_l-1))
          stored_col = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+1)
          blk_p =      recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+2)
          desm%m%wms(1)%row_i(blks) = stored_row
          desm%m%wms(1)%col_i(blks) = stored_col
          desm%m%wms(1)%blk_p(blks) = SIGN(ABS(blk_p) + (rd_disp(src_p)-1), blk_p)
          nze = row_blk_size(ABS(stored_row))&
               * col_blk_size(stored_col)
          blk_ps = blk_ps + nze
          blks = blks + 1
       ENDDO
    ENDDO
    !
    desm%m%wms(1)%lastblk = blks - 1
    desm%m%wms(1)%datasize = blk_ps - 1
    CALL dbcsr_finalize(desm)
    DEALLOCATE(send_count)
    DEALLOCATE(recv_count)
    DEALLOCATE(sdp); DEALLOCATE(sd_disp)
    DEALLOCATE(smp); DEALLOCATE(sm_disp)
    DEALLOCATE(rd_disp)
    DEALLOCATE(rm_disp)
    CALL dbcsr_data_release (recv_data)
    DEALLOCATE(recv_meta)
    CALL dbcsr_data_release (send_data)
    DEALLOCATE(send_meta)
    CALL timestop(handle)
  END SUBROUTINE dbcsr_desymmetrize_deep


! **************************************************************************************************
!> \brief Distributes a matrix that is currently replicated.
!> \param[in,out] matrix      matrix to replicate
!> \param[in] fast            change just the index, don't touch the data
! **************************************************************************************************
  SUBROUTINE dbcsr_distribute(matrix, fast)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(in), OPTIONAL                      :: fast

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_distribute', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: c_dp
      COMPLEX(KIND=sp), DIMENSION(:), POINTER            :: c_sp
      INTEGER                                            :: blk, col, handle, mynode, nblks, nze, p, &
                                                            row
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size, tmp_index
      LOGICAL                                            :: mini, tr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_dp
      REAL(KIND=sp), DIMENSION(:), POINTER               :: r_sp
      TYPE(dbcsr_data_obj)                               :: tmp_data
      TYPE(dbcsr_distribution_obj)                       :: dist
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj
      TYPE(dbcsr_obj)                                    :: distributed

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF(.NOT.dbcsr_valid_index(matrix%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",580,"Matrix not initialized.")
    IF(matrix%m%replication_type.EQ.dbcsr_repl_none)&
       CALL cp__w("dbcsr/ops/dbcsr_transformations.F",582,"Distributing a non-replicated matrix makes no sense.")
    IF (PRESENT (fast)) THEN
       mini=fast
    ELSE
       mini=.FALSE.
    ENDIF
    SELECT CASE (matrix%m%data_type)
    CASE (dbcsr_type_real_8)
       CALL dbcsr_get_data (matrix%m%data_area, r_dp)
    CASE (dbcsr_type_real_4)
       CALL dbcsr_get_data (matrix%m%data_area, r_sp)
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",593,"Only real double precision")
    CASE (dbcsr_type_complex_8)
       CALL dbcsr_get_data (matrix%m%data_area, c_dp)
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",596,"Only real double precision")
    CASE (dbcsr_type_complex_4)
       CALL dbcsr_get_data (matrix%m%data_area, c_sp)
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",599,"Only real double precision")
    END SELECT
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    dist = dbcsr_distribution (matrix)
    mp_obj = dbcsr_distribution_mp (dist)
    mynode = dbcsr_mp_mynode (dbcsr_distribution_mp (dist))
    !
    IF (mini) THEN
       ! We just mark the blocks as deleted.
       CALL dbcsr_iterator_start (iter, matrix)
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, row, col, r_dp, tr, blk)
          tr = .FALSE.
          CALL dbcsr_get_stored_coordinates (matrix%m, row, col, p)
          IF (mynode .EQ. p) THEN
             matrix%m%blk_p(blk) = 0
          ENDIF
       ENDDO
       CALL dbcsr_iterator_stop (iter)
       matrix%m%replication_type = dbcsr_repl_none
    ELSE
       CALL dbcsr_init(distributed)
       CALL dbcsr_create(distributed, name='Distributed '//TRIM(matrix%m%name),&
            template = matrix,&
            matrix_type = dbcsr_type_no_symmetry,&
            replication_type = dbcsr_repl_none)
       distributed%m%replication_type = dbcsr_repl_none
       ! First count how many blocks are local.
       nze = 0
       nblks = 0
       CALL dbcsr_iterator_start (iter, matrix)
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, row, col, r_dp, tr, blk)
          tr = .FALSE.
          CALL dbcsr_get_stored_coordinates (matrix%m, row, col, p)
          IF (mynode .EQ. p) THEN
             nze = nze + row_blk_size(row) * col_blk_size(col)
             nblks = nblks + 1
          ENDIF
       ENDDO
       CALL dbcsr_iterator_stop (iter)
       ! Preallocate the erray
       CALL dbcsr_work_create(distributed, nblks_guess=nblks,&
            sizedata_guess=nze, work_mutable=.FALSE.)
       ! Now actually do the work
       CALL dbcsr_iterator_start (iter, matrix)
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, row, col, r_dp, tr, blk)
          tr = .FALSE.
          CALL dbcsr_get_stored_coordinates (matrix%m, row, col, p)
          IF (mynode .EQ. p) THEN
             CALL dbcsr_put_block (distributed, row, col, r_dp, transposed=tr)
          ENDIF
       ENDDO
       CALL dbcsr_iterator_stop (iter)
       CALL dbcsr_finalize (distributed)
       ! Now replace the data and index
       CALL dbcsr_switch_data_area (matrix, distributed%m%data_area,&
            previous_data_area=tmp_data)
       CALL dbcsr_switch_data_area (distributed, tmp_data)
       CALL dbcsr_data_release (tmp_data)
       tmp_index => matrix%m%index
       matrix%m%index => distributed%m%index
       distributed%m%index => tmp_index
       CALL dbcsr_repoint_index (matrix%m)
       matrix%m%nze = distributed%m%nze
       matrix%m%nblks = distributed%m%nblks
       CALL dbcsr_release (distributed)
    ENDIF
    CALL timestop(handle)
  END SUBROUTINE dbcsr_distribute


! **************************************************************************************************
!> \brief Detransposes all blocks in a matrix
!> \param[in,out] matrix      DBCSR matrix
! **************************************************************************************************
  SUBROUTINE dbcsr_make_untransposed_blocks(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_untransposed_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, col_size, handle, row, row_size
      INTEGER, DIMENSION(:), POINTER                     :: cbs, rbs
      LOGICAL                                            :: sym_negation, tr
      TYPE(dbcsr_data_obj)                               :: block_data
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    rbs => dbcsr_row_block_sizes (matrix)
    cbs => dbcsr_col_block_sizes (matrix)
    sym_negation = matrix%m%negate_real

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(block_data,iter,row,col,tr,blk,row_size,col_size) &
!$OMP          SHARED(matrix,rbs,cbs,sym_negation)
    CALL dbcsr_data_init (block_data)
    CALL dbcsr_data_new (block_data, dbcsr_get_data_type(matrix))
    CALL dbcsr_iterator_start (iter, matrix)
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, block_data,&
            transposed=tr,&
            block_number=blk)
       IF (tr) THEN
          row_size = rbs(row)
          col_size = cbs(col)
          CALL dbcsr_block_transpose(block_data, col_size, row_size)
          IF (sym_negation) THEN
             SELECT CASE (block_data%d%data_type)
             CASE (dbcsr_type_real_4)
                block_data%d%r_sp(:) = -block_data%d%r_sp(:)
             CASE (dbcsr_type_real_8)
                block_data%d%r_dp(:) = -block_data%d%r_dp(:)
             CASE (dbcsr_type_complex_4)
                block_data%d%c_sp(:) = -block_data%d%c_sp(:)
             CASE (dbcsr_type_complex_8)
                block_data%d%c_dp(:) = -block_data%d%c_dp(:)
             END SELECT
          ENDIF
          matrix%m%blk_p(blk) = -matrix%m%blk_p(blk)
       ENDIF
    ENDDO
    CALL dbcsr_iterator_stop (iter)
    CALL dbcsr_data_clear_pointer (block_data)
    CALL dbcsr_data_release (block_data)
!$OMP END PARALLEL

    CALL timestop(handle)
  END SUBROUTINE dbcsr_make_untransposed_blocks


! **************************************************************************************************
!> \brief Makes a dense matrix, inplace.
!> \param[in,out] matrix      matrix to make dense
!> \param dense_matrix ...
!> \param dense_dist ...
!> \param dense_row_sizes ...
!> \param dense_col_sizes ...
!> \param row_map ...
!> \param col_map ...
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  SUBROUTINE dbcsr_make_dense (matrix, dense_matrix, dense_dist,&
       dense_row_sizes, dense_col_sizes, row_map, col_map)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_obj), INTENT(OUT)                       :: dense_matrix
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dense_dist
      TYPE(array_i1d_obj), INTENT(IN)                    :: dense_row_sizes, dense_col_sizes, &
                                                            row_map, col_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_dense', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle
      REAL(kind=dp)                                      :: cs
      TYPE(array_i1d_obj)                                :: dense_local_cols, dense_local_rows
      TYPE(dbcsr_distribution_obj)                       :: old_distribution

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    CALL dbcsr_init (dense_matrix)
    CALL dbcsr_create (dense_matrix, template=matrix,&
         dist = dense_dist,&
         row_blk_size=array_data(dense_row_sizes), col_blk_size=array_data(dense_col_sizes))
    !
    IF (dbg) THEN
       cs = dbcsr_checksum (matrix)
       WRITE(*,*)routineN//" prod cs pre", cs
    ENDIF
    old_distribution = dbcsr_distribution (matrix)
    ! Conversion of global to local offsets for the dense blocks
    CALL dbcsr_get_local_rows (dense_dist, dense_local_rows,&
         dense_matrix%m%index(dbcsr_slot_home_prow))
    CALL dbcsr_get_local_cols (dense_dist, dense_local_cols,&
         dense_matrix%m%index(dbcsr_slot_home_pcol))
    !
    CALL dbcsr_make_dense_low (matrix, dense_matrix,&
         dbcsr_distribution_local_rows (old_distribution),&
         dbcsr_distribution_local_cols (old_distribution),&
         array_data(matrix%m%row_blk_offset),&
         array_data(matrix%m%col_blk_offset),&
         array_data(dense_local_rows), array_data(dense_local_cols),&
         array_data(dense_matrix%m%row_blk_offset),&
         array_data(dense_matrix%m%col_blk_offset),&
         array_data(row_map), array_data(col_map), .TRUE., .TRUE.)
    IF (dbg) THEN
       cs = dbcsr_checksum (dense_matrix)
       WRITE(*,*)routineN//" prod cs pst", cs
    ENDIF
    CALL timestop(handle)
  END SUBROUTINE dbcsr_make_dense

! **************************************************************************************************
!> \brief Copies a matrix and makes its data dense.
!> \param[in] und_matrix         Original non-dense matrix
!> \param[in,out] dense_matrix   Dense copy of und_matrix
!> \param[in] und_local_rows     The process-grid local rows of the non-dense
!>                               und_matrix
!> \param[in] und_local_cols     The process-grid local columns of the
!>                               non-dense und_matrix
!> \param[in] und_row_blk_offsets  The block offsets of the rows of the
!>                                 non-dense matrix
!> \param[in] und_col_blk_offsets  The block offsets of the columns of the
!>                                non-dense matrix
!> \param[in] dense_local_rows    The process-grid local rows of the dense
!>                                matrix
!> \param[in] dense_local_cols    The process-grid local columns of the
!>                                dense matrix
!> \param[in] dense_row_blk_offsets  The block offsets of the rows of the
!>                                   dense matrix
!> \param[in] dense_col_blk_offsets  The block offsets of the columns of the
!>                                   dense matrix
!> \param[in] row_map                Mapping of non-dense rows to dense rows
!> \param[in] col_map                Mapping of non-dense columns to dense
!>                                   columns
!> \param[in] join_rows              Make rows dense
!> \param[in] join_cols              Make columns dense
!> \note the dense_matrix must have been created
! **************************************************************************************************
  SUBROUTINE dbcsr_make_dense_low (und_matrix, dense_matrix,&
       und_local_rows, und_local_cols,&
       und_row_blk_offsets, und_col_blk_offsets,&
       dense_local_rows, dense_local_cols,&
       dense_row_blk_offsets, dense_col_blk_offsets,&
       row_map, col_map, join_rows, join_cols)
      TYPE(dbcsr_obj), INTENT(IN)                        :: und_matrix
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: dense_matrix
      INTEGER, DIMENSION(:), INTENT(IN) :: und_local_rows, und_local_cols, und_row_blk_offsets, &
         und_col_blk_offsets, dense_local_rows, dense_local_cols, dense_row_blk_offsets, &
         dense_col_blk_offsets, row_map, col_map
      LOGICAL, INTENT(IN)                                :: join_rows, join_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_dense_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: dense_nblkcols_local, dense_nblkcols_total, dense_nblkrows_local, &
         dense_nblkrows_total, dense_nlocal_blocks, error_handle, nfullcols, nfullrows, &
         und_nblkcols_total, und_nblkrows_total
      INTEGER, ALLOCATABLE, DIMENSION(:) :: col_internal_offsets, dense_local_col_blk_offsets, &
         dense_local_row_blk_offsets, row_internal_offsets, und_local_col_blk_offsets, &
         und_local_row_blk_offsets
      INTEGER, DIMENSION(dbcsr_meta_size)                :: meta
      TYPE(dbcsr_data_obj)                               :: dense_data, und_data

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handle)
    !
    nfullrows = nfull_elements (und_row_blk_offsets, und_local_rows)
    nfullcols = nfull_elements (und_col_blk_offsets, und_local_cols)
    !
    und_nblkrows_total = SIZE (und_row_blk_offsets)-1
    und_nblkcols_total = SIZE (und_col_blk_offsets)-1
    !
    ! Find the local data offsets (but indexed by the global
    ! rows/columns) for the undense data.
    ALLOCATE (und_local_row_blk_offsets(und_nblkrows_total+1))
    ALLOCATE (und_local_col_blk_offsets(und_nblkcols_total+1))
    CALL global_offsets_to_local (&
         und_row_blk_offsets,&
         und_local_rows,&
         und_local_row_blk_offsets)
    CALL global_offsets_to_local (&
         und_col_blk_offsets,&
         und_local_cols,&
         und_local_col_blk_offsets)
    !
    dense_nblkrows_total = SIZE (dense_row_blk_offsets)-1
    dense_nblkcols_total = SIZE (dense_col_blk_offsets)-1
    dense_nblkrows_local = SIZE (dense_local_rows)
    dense_nblkcols_local = SIZE (dense_local_cols)
    !
    ! Find the local data offsets (but indexed by the (dense) global
    ! rows/columns) for the dense data.
    ALLOCATE (dense_local_row_blk_offsets (dense_nblkrows_total+1))
    ALLOCATE (dense_local_col_blk_offsets (dense_nblkcols_total+1))
    CALL global_offsets_to_local (&
         dense_row_blk_offsets,&
         dense_local_rows,&
         dense_local_row_blk_offsets)
    CALL global_offsets_to_local (&
         dense_col_blk_offsets,&
         dense_local_cols,&
         dense_local_col_blk_offsets)
    ! Find the offset of blocks within dense rows/columns.  This is needed
    ! since the blocked rows/columns are not necessarily in the same order.
    ALLOCATE (row_internal_offsets (und_nblkrows_total))
    ALLOCATE (col_internal_offsets (und_nblkcols_total))
    CALL get_internal_offsets (&
         und_local_rows, row_map,&
         und_local_row_blk_offsets,&
         dense_local_row_blk_offsets,&
         row_internal_offsets)
    CALL get_internal_offsets (&
         und_local_cols, col_map,&
         und_local_col_blk_offsets,&
         dense_local_col_blk_offsets,&
         col_internal_offsets)
    !
    und_data = und_matrix%m%data_area
    CALL dbcsr_data_hold (und_data)
    CALL dbcsr_data_init (dense_data)
    CALL dbcsr_data_new (dense_data, dbcsr_data_get_type (und_data),&
         data_size = nfullrows * nfullcols,&
         memory_type=dbcsr_data_get_memory_type (und_data))
    !
    ! Reshuffle the data
    CALL make_dense_data (und_matrix,&
         dense_data, nfullrows, nfullcols,&
         und_local_row_blk_offsets, und_local_col_blk_offsets,&
         dense_local_row_blk_offsets, dense_local_col_blk_offsets,&
         row_map = row_map, col_map = col_map,&
         row_internal_offsets = row_internal_offsets,&
         col_internal_offsets = col_internal_offsets,&
         join_rows=join_rows, join_cols=join_cols, make_tr=.FALSE.)
    CALL dbcsr_switch_data_area (dense_matrix, dense_data)
    CALL dbcsr_data_release (dense_data)
    CALL dbcsr_data_release (und_data)
    !
    ! Create the new dense index.
    dense_nlocal_blocks = dense_nblkrows_local * dense_nblkcols_local
    CALL dbcsr_addto_index_array (dense_matrix%m,&
         dbcsr_slot_row_p,&
         reservation=dense_nblkrows_total+1, extra=2*dense_nlocal_blocks)
    CALL dbcsr_addto_index_array (dense_matrix%m,&
         dbcsr_slot_col_i,&
         reservation=dense_nlocal_blocks)
    CALL dbcsr_addto_index_array (dense_matrix%m,&
         dbcsr_slot_blk_p,&
         reservation=dense_nlocal_blocks)
    !
    meta = dense_matrix%m%index(1:dbcsr_meta_size)
    CALL dbcsr_pack_meta (dense_matrix%m, meta)
    meta(dbcsr_slot_nze) = nfullrows*nfullcols
    meta(dbcsr_slot_nblks) = dense_nlocal_blocks
    CALL make_dense_index (dense_matrix%m%row_p,&
         dense_matrix%m%col_i,&
         dense_matrix%m%blk_p,&
         dense_nblkrows_total, dense_nblkcols_total,&
         dense_local_rows,&
         dense_local_cols,&
         dense_local_row_blk_offsets,&
         dense_local_col_blk_offsets,&
         make_tr=.FALSE.,&
         meta=meta)
    CALL dbcsr_unpack_meta (dense_matrix%m, meta)
    !
    DEALLOCATE (und_local_row_blk_offsets)
    DEALLOCATE (und_local_col_blk_offsets)
    DEALLOCATE (dense_local_row_blk_offsets)
    DEALLOCATE (dense_local_col_blk_offsets)
    DEALLOCATE (row_internal_offsets)
    DEALLOCATE (col_internal_offsets)
    !
    CALL timestop (error_handle)
  END SUBROUTINE dbcsr_make_dense_low


! **************************************************************************************************
!> \brief Makes a blocked matrix from a dense matrix, inplace
!> \param[in] matrix                 dense matrix
!> \param[in,out] undense_matrix     matrix to make undense
!> \param[in] distribution    distribution of non-dense rows and columns
!> \param[in] row_blk_offsets non-dense row block offsets
!> \param[in] col_blk_offsets non-dense column block offsets
!> \param[in] row_blk_sizes non-dense row block sizes
!> \param[in] col_blk_sizes non-dense column block sizes
!> \param[in] row_map mapping from non-dense rows
!> \param[in] col_map mapping from non-dense columns
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  SUBROUTINE dbcsr_make_undense (matrix, undense_matrix, distribution,&
       row_blk_offsets, col_blk_offsets, row_blk_sizes, col_blk_sizes,&
       row_map, col_map)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: undense_matrix
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: distribution
      TYPE(array_i1d_obj), INTENT(IN)                    :: row_blk_offsets, col_blk_offsets, &
                                                            row_blk_sizes, col_blk_sizes, row_map, &
                                                            col_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_undense', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, nblkcols_local, nblkcols_total, &
                                                            nblkrows_local, nblkrows_total, &
                                                            nfullcols_local, nfullrows_local
      INTEGER, ALLOCATABLE, DIMENSION(:) :: col_internal_offsets, dense_local_col_blk_offsets, &
         dense_local_row_blk_offsets, local_col_blk_offsets, local_row_blk_offsets, &
         row_internal_offsets
      INTEGER, DIMENSION(:), POINTER                     :: local_cols, local_rows, meta
      REAL(kind=dp)                                      :: cs
      TYPE(dbcsr_data_obj)                               :: blocked_data, dense_data
      TYPE(dbcsr_distribution_obj)                       :: dense_distribution

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF (dbg) THEN
       cs = dbcsr_checksum (matrix)
       WRITE(*,*)routineN//" prod cs pre", cs
    ENDIF
    dense_distribution = dbcsr_distribution (matrix)
    nfullrows_local = matrix%m%nfullrows_local
    nfullcols_local = matrix%m%nfullcols_local
    nblkrows_local = dbcsr_distribution_nlocal_rows (distribution)
    nblkcols_local = dbcsr_distribution_nlocal_cols (distribution)
    nblkrows_total = dbcsr_distribution_nrows (distribution)
    nblkcols_total = dbcsr_distribution_ncols (distribution)
    local_rows => dbcsr_distribution_local_rows (distribution)
    local_cols => dbcsr_distribution_local_cols (distribution)
    CALL dbcsr_create (undense_matrix, template=matrix,&
         dist = distribution,&
         row_blk_size = array_data(row_blk_sizes),&
         col_blk_size = array_data(col_blk_sizes))
    ! Restore previous offsets, just to try to keep the same memory.
    CALL array_release (undense_matrix%m%row_blk_offset)
    CALL array_release (undense_matrix%m%col_blk_offset)
    undense_matrix%m%row_blk_offset = row_blk_offsets
    undense_matrix%m%col_blk_offset = col_blk_offsets
    CALL array_hold (undense_matrix%m%row_blk_offset)
    CALL array_hold (undense_matrix%m%col_blk_offset)
    !
    ALLOCATE (local_row_blk_offsets(nblkrows_total+1))
    ALLOCATE (local_col_blk_offsets(nblkcols_total+1))
    CALL dbcsr_clearfrom_index_array (undense_matrix%m, dbcsr_slot_row_p)
    CALL dbcsr_clearfrom_index_array (undense_matrix%m, dbcsr_slot_col_i)
    CALL dbcsr_clearfrom_index_array (undense_matrix%m, dbcsr_slot_blk_p)
    CALL dbcsr_addto_index_array (undense_matrix%m, dbcsr_slot_row_p,&
         reservation=nblkrows_total+1, extra=nblkrows_local*nblkcols_local*2)
    CALL dbcsr_addto_index_array (undense_matrix%m, dbcsr_slot_col_i,&
         reservation=nblkrows_local*nblkcols_local)
    CALL dbcsr_addto_index_array (undense_matrix%m, dbcsr_slot_blk_p,&
         reservation=nblkrows_local*nblkcols_local)
    meta => undense_matrix%m%index(1:dbcsr_meta_size)
    CALL dbcsr_pack_meta (undense_matrix%m, meta)
    meta(dbcsr_slot_nblks) = nblkrows_local*nblkcols_local
    meta(dbcsr_slot_nze) = nfullrows_local * nfullcols_local
    CALL global_offsets_to_local (array_data (row_blk_offsets),&
         local_rows, local_row_blk_offsets(1:nblkrows_local+1))
    CALL global_offsets_to_local (array_data (col_blk_offsets),&
         local_cols, local_col_blk_offsets(1:nblkcols_local+1))
    CALL make_undense_index (undense_matrix%m%row_p, undense_matrix%m%col_i, undense_matrix%m%blk_p,&
         distribution,&
         local_row_blk_offsets(1:nblkrows_local+1),&
         local_col_blk_offsets(1:nblkcols_local+1),&
         meta)
    CALL dbcsr_unpack_meta (undense_matrix%m, meta)
    !
    CALL global_offsets_to_local (array_data (row_blk_offsets),&
         local_rows, local_row_blk_offsets)
    CALL global_offsets_to_local (array_data (col_blk_offsets),&
         local_cols, local_col_blk_offsets)
    !
    ALLOCATE (dense_local_row_blk_offsets (1+dbcsr_distribution_nrows (dense_distribution)))
    ALLOCATE (dense_local_col_blk_offsets (1+dbcsr_distribution_ncols (dense_distribution)))
    CALL global_offsets_to_local (array_data(matrix%m%row_blk_offset),&
         dbcsr_distribution_local_rows (dense_distribution),&
         dense_local_row_blk_offsets)
    CALL global_offsets_to_local (array_data(matrix%m%col_blk_offset),&
         dbcsr_distribution_local_cols (dense_distribution),&
         dense_local_col_blk_offsets)
    ! Find the offset of blocks within dense rows/columns.  This is needed
    ! since the blocked rows/columns are not necessarily in the same order.
    ALLOCATE (row_internal_offsets (nblkrows_total))
    ALLOCATE (col_internal_offsets (nblkcols_total))
    CALL get_internal_offsets (&
         local_rows, array_data(row_map),&
         local_row_blk_offsets,&
         dense_local_row_blk_offsets,&
         row_internal_offsets)
    CALL get_internal_offsets (&
         local_cols, array_data(col_map),&
         local_col_blk_offsets,&
         dense_local_col_blk_offsets,&
         col_internal_offsets)
    !
    dense_data = matrix%m%data_area
    CALL dbcsr_data_hold (dense_data)
    CALL dbcsr_data_init (blocked_data)
    CALL dbcsr_data_new (blocked_data, dbcsr_data_get_type (dense_data),&
         data_size = nfullrows_local*nfullcols_local,&
         memory_type=dbcsr_data_get_memory_type (dense_data))
    CALL dbcsr_switch_data_area (undense_matrix, blocked_data)
    CALL dbcsr_data_release (blocked_data)
    ! Reshuffle the data
    CALL make_undense_data (undense_matrix, dense_data,&
         nfullrows_local, nfullcols_local,&
         dense_local_row_blk_offsets, dense_local_col_blk_offsets,&
         array_data(row_map), array_data(col_map),&
         row_internal_offsets, col_internal_offsets)
    CALL dbcsr_data_release (dense_data)
    IF (dbg) THEN
       cs = dbcsr_checksum (matrix)
       WRITE(*,*)routineN//" prod cs pst", cs
    ENDIF
    DEALLOCATE (local_row_blk_offsets)
    DEALLOCATE (local_col_blk_offsets)
    DEALLOCATE (dense_local_row_blk_offsets)
    DEALLOCATE (dense_local_col_blk_offsets)
    DEALLOCATE (row_internal_offsets)
    DEALLOCATE (col_internal_offsets)
    CALL timestop(handle)
  END SUBROUTINE dbcsr_make_undense

! **************************************************************************************************
!> \brief Shuffles the data from blocked to standard dense form
!> \param[in] matrix                 Existing blocked matrix
!> \param[in,out] dense_data         Dense data
!> \param[in] nfullrows size of new data
!> \param[in] nfullcols size of new data
!> \param und_row_blk_offsets ...
!> \param und_col_blk_offsets ...
!> \param dense_row_blk_offsets ...
!> \param dense_col_blk_offsets ...
!> \param row_map ...
!> \param col_map ...
!> \param row_internal_offsets ...
!> \param col_internal_offsets ...
!> \param[in] join_rows           (optional) make rows dense, default is yes
!> \param[in] join_cols           (optional) make columns dense, default is
!>                                yes
!> \param[in] make_tr             (optional) make the dense blocks transposed
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  SUBROUTINE make_dense_data (matrix, dense_data, nfullrows, nfullcols,&
       und_row_blk_offsets, und_col_blk_offsets,&
       dense_row_blk_offsets, dense_col_blk_offsets,&
       row_map, col_map,&
       row_internal_offsets, col_internal_offsets,&
       join_rows, join_cols, make_tr)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dense_data
      INTEGER, INTENT(IN)                                :: nfullrows, nfullcols
      INTEGER, DIMENSION(:), INTENT(IN) :: und_row_blk_offsets, und_col_blk_offsets, &
         dense_row_blk_offsets, dense_col_blk_offsets, row_map, col_map, row_internal_offsets, &
         col_internal_offsets
      LOGICAL, INTENT(IN)                                :: join_rows, join_cols, make_tr

      CHARACTER(len=*), PARAMETER :: routineN = 'make_dense_data', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk_col, blk_col_size, blk_row, blk_row_size, dense_col, dense_row, error_handle, &
         target_col_offset, target_cs, target_offset, target_row_offset, target_rs, tco, tro
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

    CALL timeset (routineN, error_handle)
    CALL dbcsr_assert (dbcsr_data_get_size (dense_data), 'GE',&
         nfullrows*nfullcols,&
         dbcsr_fatal_level, dbcsr_internal_error, routineN,&
         "Dense data too small",1159)
    IF(.NOT.join_cols.AND..NOT.join_rows)&
       CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1161,"Joining neither rows nor columns is untested")
    !
    CALL dbcsr_data_clear (dense_data)
    IF (dbcsr_data_get_size(matrix%m%data_area) .GT. 0&
         .AND. nfullrows .GT. 0 .AND. nfullcols .GT. 0) THEN
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP PRIVATE (block, iter, &
!$OMP         target_rs, target_cs, blk_row, blk_col, tr, blk_row_size, blk_col_size,&
!$OMP         tro, tco, target_offset,&
!$OMP         target_row_offset, target_col_offset,&
!$OMP         dense_row, dense_col) &
!$OMP SHARED (&
!$OMP         dense_data, matrix, &
!$OMP         make_tr, join_rows, join_cols, &
!$OMP         und_row_blk_offsets, und_col_blk_offsets,&
!$OMP         dense_row_blk_offsets, dense_col_blk_offsets,&
!$OMP         row_internal_offsets, col_internal_offsets,&
!$OMP         row_map, col_map,&
!$OMP         nfullrows, nfullcols)
       CALL dbcsr_data_init (block)
       CALL dbcsr_data_new (block,&
         dbcsr_type_1d_to_2d(dbcsr_data_get_type(dense_data)))
       CALL dbcsr_iterator_start (iter, matrix, dynamic=.TRUE., shared=.TRUE.,&
            contiguous_pointers=.FALSE., read_only=.TRUE.)
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, blk_row, blk_col, block, tr,&
               row_size=blk_row_size, col_size=blk_col_size)
          dense_row = row_map (blk_row)
          dense_col = col_map (blk_col)
          !
          ! Calculate the target block row/column size and the offset
          ! within the target block where the undense block is placed.
          IF (join_rows) THEN
             target_row_offset = dense_row_blk_offsets(dense_row)
             target_rs = dense_row_blk_offsets(dense_row+1) -&
                         dense_row_blk_offsets(dense_row)
             tro = 1 + row_internal_offsets(blk_row)
          ELSE
             target_row_offset = und_row_blk_offsets(blk_row)
             target_rs = blk_row_size
             tro = 1
          ENDIF
          IF (join_cols) THEN
             target_col_offset = dense_col_blk_offsets(dense_col)
             target_cs =  dense_col_blk_offsets(dense_col+1) -&
                          dense_col_blk_offsets(dense_col)
             tco = 1 + col_internal_offsets(blk_col)
          ELSE
             target_col_offset = und_col_blk_offsets(blk_col)
             target_cs = blk_col_size
             tco = 1
          ENDIF
          target_offset = (target_row_offset-1) * nfullcols &
                        + (target_col_offset-1) * (&
                           dense_row_blk_offsets(dense_row+1) -&
                           dense_row_blk_offsets(dense_row))
          CALL dbcsr_block_partial_copy(dst=dense_data,&
               dst_offset=target_offset,&
               dst_rs=target_rs, dst_cs=target_cs, dst_tr=make_tr,&
               dst_r_lb = tro, dst_c_lb = tco,&
               src=block, src_rs=blk_row_size, src_cs=blk_col_size, src_tr=tr,&
               src_r_lb=1, src_c_lb=1, nrow=blk_row_size, ncol=blk_col_size)
       ENDDO
       CALL dbcsr_iterator_stop (iter)
       CALL dbcsr_data_clear_pointer (block)
       CALL dbcsr_data_release (block)
!$OMP END PARALLEL
    ENDIF
    CALL timestop (error_handle)
  END SUBROUTINE make_dense_data

! **************************************************************************************************
!> \brief Shuffles the data from standard dense to blocked form
!> \param[in,out] matrix             Matrix with data to fill
!> \param[in] dense_data             Dense data
!> \param[in] nfullrows              number of full rows in local submatrix
!> \param[in] nfullcols              number of full columns in local submatrix
!> \param[in] dense_row_blk_offsets  row block offsets for dense data
!> \param[in] dense_col_blk_offsets  column block offsets for dense data
!> \param[in] row_map                mapping from undense to dense rows
!> \param[in] col_map                mapping from undense to dense rows
!> \param row_internal_offsets ...
!> \param col_internal_offsets ...
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  SUBROUTINE make_undense_data (matrix, dense_data, nfullrows, nfullcols,&
       dense_row_blk_offsets, dense_col_blk_offsets,&
       row_map, col_map, row_internal_offsets, col_internal_offsets)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: dense_data
      INTEGER, INTENT(IN)                                :: nfullrows, nfullcols
      INTEGER, DIMENSION(:), INTENT(IN)                  :: dense_row_blk_offsets, &
                                                            dense_col_blk_offsets, row_map, &
                                                            col_map, row_internal_offsets, &
                                                            col_internal_offsets

      CHARACTER(len=*), PARAMETER :: routineN = 'make_undense_data', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk_col, blk_col_size, blk_row, blk_row_size, dense_col, dense_col_offset, &
         dense_cs, dense_offset, dense_row, dense_row_offset, dense_rs, sco, sro
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

    CALL dbcsr_assert (dbcsr_data_get_size (dense_data), 'GE', nfullrows*nfullcols,&
         dbcsr_fatal_level, dbcsr_internal_error, routineN,&
         "Dense data too small",1270)
    IF (dbcsr_data_get_size(matrix%m%data_area) .GT. 0) THEN
       CALL dbcsr_data_clear (matrix%m%data_area)
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP PRIVATE (block, iter,&
!$OMP          blk_row, blk_col, tr,&
!$OMP          blk_row_size, blk_col_size, sro, sco,&
!$OMP          dense_row_offset, dense_col_offset, dense_row, dense_col,&
!$OMP          dense_cs, dense_rs,&
!$OMP          dense_offset) &
!$OMP SHARED (&
!$OMP         matrix, dense_data, &
!$OMP         nfullrows, nfullcols, &
!$OMP         dense_row_blk_offsets, dense_col_blk_offsets,&
!$OMP         row_map, col_map,&
!$OMP         row_internal_offsets, col_internal_offsets)
       CALL dbcsr_data_init (block)
       CALL dbcsr_data_new (block,&
            dbcsr_type_1d_to_2d(dbcsr_data_get_type(dense_data)))
       CALL dbcsr_iterator_start (iter, matrix, dynamic=.TRUE., shared=.TRUE.,&
            contiguous_pointers=.FALSE.)
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, blk_row, blk_col, block, tr,&
               row_size=blk_row_size, col_size=blk_col_size)
          dense_row = row_map (blk_row)
          dense_col = col_map (blk_col)
          dense_row_offset = dense_row_blk_offsets(dense_row)
          dense_col_offset = dense_col_blk_offsets(dense_col)
          dense_rs = dense_row_blk_offsets(dense_row+1) -&
                     dense_row_blk_offsets(dense_row)
          dense_cs = dense_col_blk_offsets(dense_col+1) -&
                     dense_col_blk_offsets(dense_col)
          sro = 1 + row_internal_offsets(blk_row)
          sco = 1 + col_internal_offsets(blk_col)
          dense_offset = (dense_row_offset-1) * nfullcols &
                       + (dense_col_offset-1) * dense_rs
          CALL dbcsr_block_partial_copy (&
               dst=block, dst_rs=blk_row_size, dst_cs=blk_col_size, dst_tr=tr,&
               dst_r_lb=1, dst_c_lb=1,&
               src=dense_data, src_offset=dense_offset, &
               src_rs=dense_rs, src_cs=dense_cs, src_tr=.FALSE.,&
               src_r_lb=sro, src_c_lb=sco,&
               nrow=blk_row_size, ncol=blk_col_size)
       ENDDO
       CALL dbcsr_iterator_stop (iter)
       CALL dbcsr_data_clear_pointer (block)
       CALL dbcsr_data_release (block)
!$OMP END PARALLEL
    ENDIF
  END SUBROUTINE make_undense_data


! **************************************************************************************************
!> \brief Replicates a DBCSR on all processors.
!> \param[in,out] matrix      matrix to replicate
! **************************************************************************************************
  SUBROUTINE dbcsr_replicate_all (matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

    CALL dbcsr_replicate (matrix, replicate_rows = .TRUE.,&
         replicate_columns=.TRUE.)
  END SUBROUTINE dbcsr_replicate_all


! **************************************************************************************************
!> \brief Replicates a DBCSR matrix among process rows and columns
!> \param[in,out] matrix      matrix to replicate
!> \param[in] replicate_rows  Row should be replicated among all
!>                            processors
!> \param[in] replicate_columns  Column should be replicated among
!>                               all processors
!> \param[in] restrict_source    (optional) Send only from this node
!>                               (ignores blocks on other nodes)
!> \par Direction definition
!>      Row replication means that all processors in a process grid sharing
!>      the same row get the data of the entire row. (In a 1-column grid the
!>      operation has no effect.) Similar logic applies to column replication.
! **************************************************************************************************
  SUBROUTINE dbcsr_replicate (matrix, replicate_rows, replicate_columns,&
       restrict_source)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN)                                :: replicate_rows, replicate_columns
      INTEGER, INTENT(IN), OPTIONAL                      :: restrict_source

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_replicate', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: metalen = 3
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      CHARACTER                                          :: rep_type
      INTEGER :: blk, blk_l, blk_p, blk_ps, blks, col, col_size, data_type, dst_p, handle, &
         mp_group, mynode, mypcol, myprow, nblks, numnodes, nze, offset, row, row_size, smp, &
         src_p, stored_blk_p, stored_col, stored_row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rd_disp, recv_meta, rm_disp, send_meta
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: recv_count
      INTEGER, DIMENSION(2)                              :: send_count
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_dist, row_blk_size, &
                                                            row_dist, tmp_index
      INTEGER, DIMENSION(:, :), POINTER                  :: blacs2mpi
      LOGICAL                                            :: had_subcomms, i_am_restricted, rest_src, &
                                                            tr
      TYPE(dbcsr_data_obj)                               :: data_block, recv_data, send_data, &
                                                            tmp_data
      TYPE(dbcsr_distribution_obj)                       :: target_dist
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj
      TYPE(dbcsr_obj)                                    :: replicated

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF(.NOT.dbcsr_valid_index(matrix%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1382,"Matrix not initialized.")
    !CALL dbcsr_assert (matrix%m%replication_type .EQ. dbcsr_repl_none,&
    !     dbcsr_warning_level, dbcsr_caller_error, routineN,&
    !     "Replicating a non-distributed matrix makes no sense.",1385)
    IF (replicate_rows .AND. replicate_columns) THEN
       rep_type = dbcsr_repl_full
    ELSEIF (replicate_rows .AND. .NOT. replicate_columns) THEN
       rep_type = dbcsr_repl_row
    ELSEIF (replicate_columns .AND. .NOT. replicate_rows) THEN
       rep_type = dbcsr_repl_col
    ELSE
       rep_type = dbcsr_repl_none
       CALL dbcsr_assert(replicate_rows, "OR", replicate_columns, &
            dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
            "Some replication must be specified", 1396)
    ENDIF
    data_type = dbcsr_get_data_type (matrix)
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    target_dist = matrix%m%dist
    row_dist => dbcsr_distribution_row_dist (target_dist)
    col_dist => dbcsr_distribution_col_dist (target_dist)
    mp_obj = dbcsr_distribution_mp (target_dist)
    blacs2mpi => dbcsr_mp_pgrid (mp_obj)
    numnodes = dbcsr_mp_numnodes (mp_obj)
    mynode = dbcsr_mp_mynode (mp_obj)
    myprow = dbcsr_mp_myprow (mp_obj)
    mypcol = dbcsr_mp_mypcol (mp_obj)
    IF(MAXVAL(row_dist).GT.UBOUND(blacs2mpi,1))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1411,'Row distribution references unexistent processor rows')
    IF (dbg) THEN
       IF(MAXVAL(row_dist).NE.UBOUND(blacs2mpi,1))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1414,'Range of row distribution not equal to processor rows')
    ENDIF
    IF(MAXVAL(col_dist).GT.UBOUND(blacs2mpi,2))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1417,'Col distribution references unexistent processor cols')
    IF (dbg) THEN
       IF(MAXVAL(col_dist).NE.UBOUND(blacs2mpi,2))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1420,'Range of col distribution not equal to processor cols')
    ENDIF
    ! Define the number of nodes with which I will communicate. Also
    ! setup row and column communicators.
    had_subcomms = dbcsr_mp_has_subgroups (mp_obj)
    SELECT CASE (rep_type)
    CASE (dbcsr_repl_full)
       numnodes = dbcsr_mp_numnodes (mp_obj)
       mp_group = dbcsr_mp_group (mp_obj)
       mynode = dbcsr_mp_mynode (mp_obj)
    CASE (dbcsr_repl_row)
       numnodes = dbcsr_mp_npcols (mp_obj)
       CALL dbcsr_mp_grid_setup (mp_obj)
       mp_group = dbcsr_mp_my_row_group (mp_obj)
       mynode = dbcsr_mp_mypcol (mp_obj)
    CASE (dbcsr_repl_col)
       numnodes = dbcsr_mp_nprows (mp_obj)
       CALL dbcsr_mp_grid_setup (mp_obj)
       mp_group = dbcsr_mp_my_col_group (mp_obj)
       mynode = dbcsr_mp_myprow (mp_obj)
    CASE (dbcsr_repl_none)
       numnodes = 1
       mp_group = dbcsr_mp_group (mp_obj)
       mynode = 0
    END SELECT
    CALL dbcsr_assert (rep_type.EQ.dbcsr_repl_row &
         .OR. rep_type.EQ.dbcsr_repl_col,&
         "IMP", dbcsr_mp_has_subgroups (mp_obj), dbcsr_fatal_level,&
         dbcsr_unimplemented_error_nr, routineN,&
         "Only full replication supported when subcommunicators are turned off.",&
         1450)
    !
    IF (PRESENT (restrict_source)) THEN
       rest_src = .TRUE.
       i_am_restricted = mynode .NE. restrict_source
    ELSE
       rest_src = .FALSE.
       i_am_restricted = .FALSE.
    ENDIF
    !
    ALLOCATE (recv_count(2, 0:numnodes-1))
    ALLOCATE (rd_disp(0:numnodes-1))
    ALLOCATE (rm_disp(0:numnodes-1))
    CALL dbcsr_init (replicated)
    CALL dbcsr_create(replicated, name='Replicated '//TRIM(matrix%m%name),&
         template = matrix,&
         matrix_type = dbcsr_type_no_symmetry,&
         replication_type=rep_type)
    !
    ! Count initial sizes for sending. Also, ensure that blocks are on their
    ! home processors.
    send_count(1:2) = 0
    CALL dbcsr_iterator_start (iter, matrix)
    IF (.NOT. i_am_restricted) THEN
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, row, col, &
               row_size=row_size, col_size=col_size, blk=blk)
          !tr = .FALSE.
          !CALL dbcsr_get_stored_coordinates (matrix, row, col, tr, dst_p)
          !CALL dbcsr_assert (dst_p, "EQ", mynode, dbcsr_fatal_level,&
          !     dbcsr_wrong_args_error, routineN, &
          !     "Matrix is not correctly distributed. Call dbcsr_redistribute.",&
          !     1482)
          nze = row_size * col_size
          send_count(1) = send_count(1) + 1
          send_count(2) = send_count(2) + nze
       ENDDO
       send_count(2) = dbcsr_data_get_size_referenced (matrix%m%data_area)
    ENDIF
    CALL dbcsr_iterator_stop (iter)
    ! Exchange how much data others have.
    CALL mp_allgather (send_count(1:2), recv_count, mp_group)
    CALL dbcsr_data_init (recv_data)
    nze = SUM (recv_count(2, :))
    nblks = SUM (recv_count(1, :))
    CALL dbcsr_data_new (recv_data, data_type=data_type, data_size=nze)
    ! send_data should have the correct size
    CALL dbcsr_data_init (send_data)
    IF (send_count(2) .EQ. 0) THEN
       CALL dbcsr_data_new (send_data, data_type=data_type, data_size=0)
    ELSE
       CALL dbcsr_data_new (send_data, data_type=data_type)
       send_data = pointer_view (send_data, matrix%m%data_area, 1, send_count(2))
    ENDIF
    ALLOCATE (recv_meta(metalen * nblks))
    ALLOCATE (send_meta(metalen * send_count(1)))
    recv_meta(:) = 0
    ! Fill in the meta data structures and copy the data.
    rd_disp = -1 ; rm_disp = -1
    rd_disp(0) = 1 ; rm_disp(0) = 1
    DO dst_p = 1, numnodes-1
       rm_disp(dst_p) = rm_disp(dst_p-1)&
                        + metalen*recv_count(1, dst_p-1)
       rd_disp(dst_p) = rd_disp(dst_p-1)&
                        + recv_count(2, dst_p-1)
    ENDDO
    CALL dbcsr_data_init (data_block)
    CALL dbcsr_data_new (data_block, data_type=data_type)
    CALL dbcsr_iterator_start (iter, matrix)
    smp = 1
    IF (.NOT. i_am_restricted) THEN
       DO WHILE (dbcsr_iterator_blocks_left (iter))
          CALL dbcsr_iterator_next_block (iter, row, col, blk,&
               transposed=tr, blk_p=blk_p)
          send_meta(smp+0) = row
          send_meta(smp+1) = col
          send_meta(smp+2) = blk_p
          smp = smp + metalen
       ENDDO
    ENDIF
    CALL dbcsr_iterator_stop (iter)
    CALL dbcsr_data_clear_pointer (data_block)
    CALL dbcsr_data_release (data_block)
    ! Exchange the data and metadata structures.
    CALL mp_allgather (send_meta, recv_meta, metalen*recv_count(1,:),&
         rm_disp-1, mp_group)
    CALL dbcsr_allgatherv (send_data, recv_data, recv_count(2,:),&
         rd_disp-1, mp_group)
    ! Release the send buffer. If it had a non-zero size then it was a
    ! pointer into the regular matrix and the data pointer should be
    ! cleared and not deallocated.
    IF (send_count(2) .NE. 0) THEN
       CALL dbcsr_data_clear_pointer (send_data)
    ENDIF
    CALL dbcsr_data_release (send_data)
    !
    ! Now fill in the data.
    CALL dbcsr_work_create(replicated,&
         SUM(recv_count(1,:)),&
         SUM(recv_count(2,:)), n=1, &
         work_mutable=.FALSE.)
    CALL dbcsr_data_hold (recv_data)
    CALL dbcsr_data_release (replicated%m%wms(1)%data_area)
    replicated%m%wms(1)%data_area = recv_data
    blk_ps = 1
    blks = 1
    DO src_p = 0, numnodes-1
       nze = recv_count(2, src_p)
       !CALL dbcsr_data_set (replicated%m%wms(1)%data_area, blk_ps, nze,&
       !     recv_data, rd_disp(src_p))
       offset = rd_disp(src_p) - 1
       DO blk_l = 1, recv_count(1, src_p)
          IF (dbg) WRITE(*,*)"src_p, blk_l", src_p, blk_l
          stored_row = recv_meta(rm_disp(src_p)+metalen*(blk_l-1))
          stored_col = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+1)
          stored_blk_p = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+2)
          replicated%m%wms(1)%row_i(blks) = stored_row
          replicated%m%wms(1)%col_i(blks) = stored_col
          replicated%m%wms(1)%blk_p(blks) = SIGN(ABS(stored_blk_p) + offset,&
                                                 stored_blk_p)
          nze = row_blk_size(stored_row)&
               * col_blk_size(stored_col)
          blk_ps = MAX (blk_ps, ABS(stored_blk_p) + nze + offset)
          blks = blks + 1
       ENDDO
    ENDDO
    CALL dbcsr_data_set_size_referenced (replicated%m%wms(1)%data_area, blk_ps - 1)
    !
    replicated%m%wms(1)%lastblk = blks - 1
    replicated%m%wms(1)%datasize = blk_ps - 1
    CALL dbcsr_finalize(replicated, reshuffle=.TRUE.)
    !
    ! Remove communicators if they were forcibly created.
    IF (had_subcomms .AND.&
         (rep_type .EQ. dbcsr_repl_row .OR. rep_type .EQ. dbcsr_repl_col)) THEN
       CALL dbcsr_mp_grid_remove (mp_obj)
    ENDIF
    DEALLOCATE(recv_count)
    DEALLOCATE(rd_disp)
    DEALLOCATE(rm_disp)
    CALL dbcsr_data_release (recv_data)
    DEALLOCATE(recv_meta)
    DEALLOCATE(send_meta)
    matrix%m%replication_type = replicated%m%replication_type
    ! Now replace the data and index
    CALL dbcsr_switch_data_area (matrix, replicated%m%data_area,&
         previous_data_area=tmp_data)
    CALL dbcsr_switch_data_area (replicated, tmp_data)
    CALL dbcsr_data_release (tmp_data)
    tmp_index => matrix%m%index
    matrix%m%index => replicated%m%index
    replicated%m%index => tmp_index
    CALL dbcsr_repoint_index (matrix%m)
    matrix%m%nze = replicated%m%nze
    matrix%m%nblks = replicated%m%nblks
    CALL dbcsr_release (replicated)
    CALL dbcsr_verify_matrix (matrix)
    CALL timestop(handle)
  END SUBROUTINE dbcsr_replicate


! **************************************************************************************************
!> \brief Fully redistributes a DBCSR matrix.
!>
!>        The new distribution may be arbitrary as long as the total
!>        number full rows and columns matches that of the existing
!>        matrix.
!> \param[in] matrix          matrix to redistribute
!> \param[in,out] redist      redistributed matrix
!> \param[in] keep_sparsity   (optional) retains the sparsity of the redist
!>                            matrix
! **************************************************************************************************
  SUBROUTINE dbcsr_complete_redistribute(matrix, redist, keep_sparsity)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: redist
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_complete_redistribute', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: metalen = 7
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER :: blk, blk_col_new, blk_ps, blk_row_new, blks, cnt_fnd, cnt_new, cnt_skip, col, &
         col_int, col_offset_new, col_offset_old, col_rle, col_size, col_size_new, data_offset_l, &
         data_type, dst_p, handle, i, meta_l, mp_group, numnodes, nze_rle, row, row_int, &
         row_offset_new, row_offset_old, row_rle, row_size, row_size_new, src_p, stored_col_new, &
         stored_row_new
      INTEGER, ALLOCATABLE, DIMENSION(:) :: col_end_new, col_end_old, col_start_new, &
         col_start_old, rd_disp, recv_meta, rm_disp, row_end_new, row_end_old, row_start_new, &
         row_start_old, sd_disp, sdp, send_meta, sm_disp, smp
      INTEGER, ALLOCATABLE, DIMENSION(:, :) :: col_reblocks, n_col_reblocks, n_row_reblocks, &
         recv_count, row_reblocks, send_count, total_recv_count, total_send_count
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size_new, col_blk_size_old, &
                                                            col_dist_new, row_blk_size_new, &
                                                            row_blk_size_old, row_dist_new
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid
      LOGICAL                                            :: found, my_keep_sparsity, sym, tr, &
                                                            valid_block
      REAL(kind=dp)                                      :: cs1, cs2
      TYPE(dbcsr_data_obj)                               :: buff_data, data_block, recv_data, &
                                                            send_data
      TYPE(dbcsr_distribution_obj)                       :: dist_new
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj_new

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF(.NOT.dbcsr_valid_index(matrix%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1659,"Input not valid.")
    IF(.NOT.dbcsr_is_initialized(redist%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1661,"Output matrix not initialized.")
    IF(matrix%m%replication_type.NE.dbcsr_repl_none)&
       CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1663,"Can not redistribute replicated matrix.")
    !
    my_keep_sparsity = .FALSE.
    IF (PRESENT (keep_sparsity)) my_keep_sparsity = keep_sparsity
    !
    sym = dbcsr_has_symmetry (redist)
    data_type = matrix%m%data_type
    ! Get row and column start and end positions
    ! Old matrix
    row_blk_size_old => array_data (matrix%m%row_blk_size)
    col_blk_size_old => array_data (matrix%m%col_blk_size)
    ALLOCATE (row_start_old (dbcsr_nblkrows_total (matrix)),&
         row_end_old (dbcsr_nblkrows_total (matrix)),&
         col_start_old (dbcsr_nblkcols_total (matrix)),&
         col_end_old (dbcsr_nblkcols_total (matrix)))
    CALL convert_sizes_to_offsets (row_blk_size_old,&
         row_start_old, row_end_old)
    CALL convert_sizes_to_offsets (col_blk_size_old,&
         col_start_old, col_end_old)
    ! New matrix
    dist_new = dbcsr_distribution (redist)
    row_blk_size_new => array_data (redist%m%row_blk_size)
    col_blk_size_new => array_data (redist%m%col_blk_size)
    ALLOCATE (row_start_new (dbcsr_nblkrows_total (redist)),&
         row_end_new (dbcsr_nblkrows_total (redist)),&
         col_start_new (dbcsr_nblkcols_total (redist)),&
         col_end_new (dbcsr_nblkcols_total (redist)))
    CALL convert_sizes_to_offsets (row_blk_size_new,&
         row_start_new, row_end_new)
    CALL convert_sizes_to_offsets (col_blk_size_new,&
         col_start_new, col_end_new)
    row_dist_new => dbcsr_distribution_row_dist (dist_new)
    col_dist_new => dbcsr_distribution_col_dist (dist_new)
    ! Create mappings
    i = dbcsr_nfullrows_total (redist)
    ALLOCATE (row_reblocks (4,i))
    ALLOCATE (n_row_reblocks (2, dbcsr_nblkrows_total (matrix)))
    CALL dbcsr_reblocking_targets (row_reblocks, i, n_row_reblocks,&
         row_blk_size_old, row_blk_size_new)
    i = dbcsr_nfullcols_total (redist)
    ALLOCATE (col_reblocks (4,i))
    ALLOCATE (n_col_reblocks (2, dbcsr_nblkcols_total (matrix)))
    CALL dbcsr_reblocking_targets (col_reblocks, i, n_col_reblocks,&
         col_blk_size_old, col_blk_size_new)
    !
    mp_obj_new = dbcsr_distribution_mp (dist_new)
    pgrid => dbcsr_mp_pgrid (mp_obj_new)
    numnodes = dbcsr_mp_numnodes (mp_obj_new)
    mp_group = dbcsr_mp_group (mp_obj_new)
    !
    CALL dbcsr_assert(MAXVAL(row_dist_new),"LE",UBOUND(pgrid,1), dbcsr_failure_level,&
         dbcsr_wrong_args_error, routineN,&
         'Row distribution references unexistent processor rows',1715)
    IF (dbg) THEN
       IF(MAXVAL(row_dist_new).NE.UBOUND(pgrid,1))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1718,'Range of row distribution not equal to processor rows')
    ENDIF
    CALL dbcsr_assert(MAXVAL(col_dist_new),"LE",UBOUND(pgrid,2), dbcsr_failure_level,&
         dbcsr_wrong_args_error, routineN,&
         'Col distribution references unexistent processor cols',1722)
    IF (dbg) THEN
       IF(MAXVAL(col_dist_new).NE.UBOUND(pgrid,2))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",1725,'Range of col distribution not equal to processor cols')
    ENDIF
    ALLOCATE (send_count(2, 0:numnodes-1))
    ALLOCATE (recv_count(2, 0:numnodes-1))
    ALLOCATE (total_send_count(2, 0:numnodes-1))
    ALLOCATE (total_recv_count(2, 0:numnodes-1))
    ALLOCATE (sdp(0:numnodes-1))
    ALLOCATE (sd_disp(0:numnodes-1))
    ALLOCATE (smp(0:numnodes-1))
    ALLOCATE (sm_disp(0:numnodes-1))
    ALLOCATE (rd_disp(0:numnodes-1))
    ALLOCATE (rm_disp(0:numnodes-1))
    IF (dbg) THEN
       cs1 = dbcsr_checksum (matrix)
    ENDIF
    !cs1 = dbcsr_checksum (matrix)
    !call dbcsr_print(matrix)
    !
    !
    ! Count initial sizes for sending.
    !
    ! We go through every element of every local block and determine
    ! to which processor it must be sent. It could be more efficient,
    ! but at least the index data are run-length encoded.
    send_count(:,:) = 0
    CALL dbcsr_iterator_start (iter, matrix)
    dst_p = -1
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, blk)
       DO col_int = n_col_reblocks (1,col),&
            n_col_reblocks(1,col) + n_col_reblocks(2,col)-1
          blk_col_new = col_reblocks(1, col_int)
          DO row_int = n_row_reblocks (1, row),&
               n_row_reblocks(1,row) + n_row_reblocks(2,row)-1
             blk_row_new = row_reblocks(1, row_int)
             IF (.NOT. sym .OR. blk_col_new .GE. blk_row_new) THEN
                tr = .FALSE.
                CALL dbcsr_get_stored_coordinates (redist%m,&
                     blk_row_new, blk_col_new, dst_p)
                send_count(1, dst_p) = send_count(1, dst_p) + 1
                send_count(2, dst_p) = send_count(2, dst_p) +&
                     col_reblocks(2,col_int)*row_reblocks(2,row_int)
             ENDIF
          ENDDO
       ENDDO
    ENDDO
    CALL dbcsr_iterator_stop (iter)
    !
    !
    CALL mp_alltoall(send_count, recv_count, 2, mp_group)
    ! Allocate data structures needed for data exchange.
    CALL dbcsr_data_init (recv_data)
    CALL dbcsr_data_new (recv_data, data_type, SUM(recv_count(2, :)))
    ALLOCATE (recv_meta(metalen*SUM(recv_count(1, :))))
    CALL dbcsr_data_init (send_data)
    CALL dbcsr_data_new (send_data, data_type, SUM(send_count(2, :)))
    ALLOCATE (send_meta(metalen*SUM(send_count(1, :))))
    ! Fill in the meta data structures and copy the data.
    DO dst_p = 0, numnodes-1
       total_send_count(1, dst_p) = send_count (1, dst_p)
       total_send_count(2, dst_p) = send_count (2, dst_p)
       total_recv_count(1, dst_p) = recv_count (1, dst_p)
       total_recv_count(2, dst_p) = recv_count (2, dst_p)
    ENDDO
    sd_disp = -1 ; sm_disp = -1
    rd_disp = -1 ; rm_disp = -1
    sd_disp(0) = 1 ; sm_disp(0) = 1
    rd_disp(0) = 1 ; rm_disp(0) = 1
    DO dst_p = 1, numnodes-1
       sm_disp(dst_p) = sm_disp(dst_p-1)&
                        + metalen*total_send_count(1, dst_p-1)
       sd_disp(dst_p) = sd_disp(dst_p-1)&
                        + total_send_count(2, dst_p-1)
       rm_disp(dst_p) = rm_disp(dst_p-1)&
                        + metalen*total_recv_count(1, dst_p-1)
       rd_disp(dst_p) = rd_disp(dst_p-1)&
                        + total_recv_count(2, dst_p-1)
    ENDDO
    sdp(:) = sd_disp     ! sdp points to the the next place to store
                         ! data. It is postincremented.
    smp(:) = sm_disp - metalen  ! But smp points to the "working" data, not
                                ! the next. It is pre-incremented, so we must
                                ! first rewind it.
    !
    CALL dbcsr_data_init (data_block)
    CALL dbcsr_data_new (data_block, data_type)
    CALL dbcsr_iterator_start (iter, matrix)
    dst_p = -1
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, data_block, tr, blk,&
            row_size=row_size, col_size=col_size)
       !IF (tr) WRITE(*,*)"block at",row,col," is transposed"
       DO col_int = n_col_reblocks (1,col),&
            n_col_reblocks(1,col) + n_col_reblocks(2,col)-1
          blk_col_new = col_reblocks(1, col_int)
          DO row_int = n_row_reblocks (1, row),&
               n_row_reblocks(1,row) + n_row_reblocks(2,row)-1
             blk_row_new = row_reblocks(1, row_int)
             loc_ok: IF (.NOT. sym .OR. blk_col_new .GE. blk_row_new) THEN
                IF (dbg) &
                     WRITE(*,*)'using block',blk_row_new,'x',blk_col_new
                ! Start a new RLE run
                tr = .FALSE.
                CALL dbcsr_get_stored_coordinates (redist%m,&
                     blk_row_new, blk_col_new, dst_p)
                row_offset_old = row_reblocks(3, row_int)
                col_offset_old = col_reblocks(3, col_int)
                row_offset_new = row_reblocks(4, row_int)
                col_offset_new = col_reblocks(4, col_int)
                row_rle = row_reblocks(2,row_int)
                col_rle = col_reblocks(2,col_int)
                smp(dst_p) = smp(dst_p) + metalen
                send_meta(smp(dst_p)) = blk_row_new   ! new blocked row
                send_meta(smp(dst_p)+1) = blk_col_new ! new blocked column
                send_meta(smp(dst_p)+2) = row_offset_new  ! row in new block
                send_meta(smp(dst_p)+3) = col_offset_new  ! col in new block
                send_meta(smp(dst_p)+4) = row_rle ! RLE rows
                send_meta(smp(dst_p)+5) = col_rle ! RLE columns
                send_meta(smp(dst_p)+6) = sdp(dst_p)-sd_disp(dst_p) ! Offset in data
                nze_rle = row_rle * col_rle
                ! Copy current block into the send buffer
                CALL dbcsr_block_partial_copy(&
                     send_data, dst_offset=sdp(dst_p)-1,&
                     dst_rs=row_rle, dst_cs=col_rle, dst_tr=.FALSE.,&
                     dst_r_lb=1, dst_c_lb=1,&
                     src=data_block,&
                     src_rs=row_size, src_cs=col_size, src_tr=tr,&
                     src_r_lb=row_offset_old, src_c_lb=col_offset_old,&
                     nrow=row_rle, ncol=col_rle)
                sdp(dst_p) = sdp(dst_p)+nze_rle
             ENDIF loc_ok
          ENDDO ! row_int
       ENDDO ! col_int
    ENDDO
    CALL dbcsr_iterator_stop (iter)
    CALL dbcsr_data_clear_pointer (data_block)
    CALL dbcsr_data_release (data_block)

    ! Exchange the data and metadata structures.
    !
    SELECT CASE (data_type)
    CASE (dbcsr_type_real_4)
       CALL hybrid_alltoall_s1(&
            send_data%d%r_sp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%r_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_real_8)
       !CALL mp_alltoall(&
       !     send_data%d%r_dp(:), total_send_count(2,:), sd_disp(:)-1,&
       !     recv_data%d%r_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
       !     mp_group)
       CALL hybrid_alltoall_d1 (&
            send_data%d%r_dp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%r_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_complex_4)
       CALL hybrid_alltoall_c1(&
            send_data%d%c_sp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%c_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_complex_8)
       CALL hybrid_alltoall_z1(&
            send_data%d%c_dp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%c_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE default
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",1891,"Invalid matrix type")
    END SELECT
    CALL hybrid_alltoall_i1(send_meta(:), metalen*total_send_count(1,:), sm_disp(:)-1,&
         recv_meta(:), metalen*total_recv_count(1,:), rm_disp(:)-1, mp_obj_new)
    !
    ! Now fill in the data.
    CALL dbcsr_work_create(redist,&
         nblks_guess=SUM(recv_count(1,:)),&
         sizedata_guess=SUM(recv_count(2,:)), work_mutable=.TRUE.)
    CALL dbcsr_data_init (buff_data)
    CALL dbcsr_data_init (data_block)
    CALL dbcsr_data_new (buff_data, dbcsr_type_1d_to_2d (data_type), &
         redist%m%max_rbs, redist%m%max_cbs)
    CALL dbcsr_data_new (data_block, dbcsr_type_1d_to_2d (data_type))

    !blk_p = 1
    !blk = 1
    blk_ps = 0
    blks = 0
    cnt_fnd = 0 ; cnt_new = 0 ; cnt_skip = 0
    DO src_p = 0, numnodes-1
       data_offset_l = rd_disp(src_p)
       DO meta_l = 1, recv_count(1, src_p)
          stored_row_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1))
          stored_col_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+1)
          row_offset_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+2)
          col_offset_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+3)
          row_rle = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+4)
          col_rle = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+5)
          data_offset_l = rd_disp(src_p)&
               + recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+6)

          CALL dbcsr_data_clear_pointer (data_block)
          CALL dbcsr_get_block_p(redist, stored_row_new, stored_col_new,&
               data_block, tr, found)
          valid_block = found

          IF (found) cnt_fnd = cnt_fnd + 1
          IF (.NOT. found .AND. .NOT. my_keep_sparsity) THEN
             ! We have to set up a buffer block
             CALL dbcsr_data_set_pointer (data_block,&
                  rsize=row_blk_size_new (stored_row_new),&
                  csize=col_blk_size_new (stored_col_new),&
                  pointee=buff_data)
             CALL dbcsr_data_clear (data_block)
             !r2_dp => r2_dp_buff(1:row_blk_size_new (stored_row_new),&
             !     1:col_blk_size_new (stored_col_new))
             !r2_dp(:,:) = 0.0_dp
             tr = .FALSE.
             blks = blks + 1
             blk_ps = blk_ps + row_blk_size_new (stored_row_new) * &
                  col_blk_size_new (stored_col_new)
             valid_block = .TRUE.
             cnt_new = cnt_new + 1
          ENDIF
          nze_rle = row_rle*col_rle

          IF (valid_block) THEN
             row_size_new = row_blk_size_new (stored_row_new)
             col_size_new = col_blk_size_new (stored_col_new)
             CALL dbcsr_block_partial_copy (&
                  dst=data_block, dst_tr=tr,&
                  dst_rs=row_size_new, dst_cs=col_size_new, &
                  dst_r_lb=row_offset_new, dst_c_lb=col_offset_new,&
                  src=recv_data, src_offset=data_offset_l-1,&
                  src_rs=row_rle, src_cs=col_rle, src_tr=.FALSE.,&
                  src_r_lb=1, src_c_lb=1,&
                  nrow=row_rle, ncol=col_rle)
          ELSE
             cnt_skip = cnt_skip+1
          ENDIF

          data_offset_l = data_offset_l + nze_rle
          IF (.NOT. found .AND. valid_block) THEN
             IF (dbg) WRITE(*,*)routineN//" Adding new block at",&
                  stored_row_new, stored_col_new
             CALL dbcsr_put_block(redist, stored_row_new, stored_col_new,&
                  data_block, transposed=tr)
             !DEALLOCATE (r2_dp)
          ELSE
             IF (.NOT. my_keep_sparsity .AND. dbg) &
                  WRITE(*,*)routineN//" Reusing block at",&
                  stored_row_new, stored_col_new
          ENDIF
       ENDDO
    ENDDO

    CALL dbcsr_data_clear_pointer(data_block)
    CALL dbcsr_data_release (buff_data)
    CALL dbcsr_data_release (data_block)
    !
    IF (dbg) THEN
       WRITE(*,*)routineN//" Declared blocks=",redist%m%wms(1)%lastblk,&
            "actual=",blks
       WRITE(*,*)routineN//" Declared data size=",redist%m%wms(1)%datasize,&
            "actual=",blk_ps
    ENDIF

    CALL dbcsr_finalize(redist)

    DEALLOCATE(send_count)
    DEALLOCATE(recv_count)
    DEALLOCATE(sdp); DEALLOCATE(sd_disp)
    DEALLOCATE(smp); DEALLOCATE(sm_disp)
    DEALLOCATE(rd_disp)
    DEALLOCATE(rm_disp)

    CALL dbcsr_data_release (recv_data)
    CALL dbcsr_data_release (send_data)

    DEALLOCATE(recv_meta)
    DEALLOCATE(send_meta)
   
    !if (dbg) call dbcsr_print(redist)
    IF (dbg) THEN
       cs2 = dbcsr_checksum (redist)
       WRITE(*,*)routineN//" Checksums=",cs1, cs2, cs1-cs2
    ENDIF
    !CALL dbcsr_assert (cs1-cs2 .LT. 0.00001, dbcsr_fatal_level, dbcsr_internal_error,&
    !     routineN, "Mangled data!")
    CALL timestop(handle)
  END SUBROUTINE dbcsr_complete_redistribute

! **************************************************************************************************
!> \brief Redistributes a DBCSR matrix.
!>
!>        The new distribution should have compatible row and column blocks.
!> \param[in] matrix          matrix to redistribute
!> \param[in,out] redist      redistributed matrix, which should already be
!>                            created
! **************************************************************************************************
  SUBROUTINE dbcsr_redistribute(matrix, redist)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: redist

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_redistribute', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: metalen = 2
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER :: blk, blk_ps, blks, col, col_size, data_type, dst_p, handle, meta_l, mp_group, &
         numnodes, nze, row, row_size, src_p, stored_col_new, stored_row_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rd_disp, recv_meta, rm_disp, sd_disp, &
                                                            sdp, send_meta, sm_disp, smp
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: recv_count, send_count, &
                                                            total_recv_count, total_send_count
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size_new, col_dist_new, &
                                                            row_blk_size_new, row_dist_new
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid
      LOGICAL                                            :: sym_tr, tr
      TYPE(dbcsr_data_obj)                               :: data_block, recv_data, send_data
      TYPE(dbcsr_distribution_obj)                       :: dist_new
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj_new

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    !call dbcsr_print_dist (matrix%m%dist)
    !call dbcsr_print_dist (redist%m%dist)
    IF(.NOT.dbcsr_valid_index(matrix%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",2052,"Input not valid.")
    IF(.NOT.dbcsr_is_initialized(redist%m))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",2054,"Output matrix not initialized.")
    IF(matrix%m%replication_type.NE.dbcsr_repl_none)&
       CALL cp__w("dbcsr/ops/dbcsr_transformations.F",2056,"Can not redistribute replicated matrix.")
    data_type = matrix%m%data_type
    ! Get row and column start and end positions
    ! Old matrix
    ! New matrix
    dist_new = dbcsr_distribution (redist)
    row_blk_size_new => array_data (redist%m%row_blk_size)
    col_blk_size_new => array_data (redist%m%col_blk_size)
    row_dist_new => dbcsr_distribution_row_dist (dist_new)
    col_dist_new => dbcsr_distribution_col_dist (dist_new)
    !
    mp_obj_new = dbcsr_distribution_mp (dist_new)
    pgrid => dbcsr_mp_pgrid (mp_obj_new)
    numnodes = dbcsr_mp_numnodes (mp_obj_new)
    mp_group = dbcsr_mp_group (mp_obj_new)
    !
    IF(MAXVAL(row_dist_new).GT.UBOUND(pgrid,1))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",2073,'Row distribution references unexistent processor rows')
    IF (dbg) THEN
       IF(MAXVAL(row_dist_new).NE.UBOUND(pgrid,1))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",2076,'Range of row distribution not equal to processor rows')
    ENDIF
    IF(MAXVAL(col_dist_new).GT.UBOUND(pgrid,2))&
       CALL cp__b("dbcsr/ops/dbcsr_transformations.F",2079,'Col distribution references unexistent processor cols')
    IF (dbg) THEN
       IF(MAXVAL(col_dist_new).NE.UBOUND(pgrid,2))&
          CALL cp__w("dbcsr/ops/dbcsr_transformations.F",2082,'Range of col distribution not equal to processor cols')
    ENDIF
    ALLOCATE (send_count(2, 0:numnodes-1))
    ALLOCATE (recv_count(2, 0:numnodes-1))
    ALLOCATE (total_send_count(2, 0:numnodes-1))
    ALLOCATE (total_recv_count(2, 0:numnodes-1))
    ALLOCATE (sdp(0:numnodes-1))
    ALLOCATE (sd_disp(0:numnodes-1))
    ALLOCATE (smp(0:numnodes-1))
    ALLOCATE (sm_disp(0:numnodes-1))
    ALLOCATE (rd_disp(0:numnodes-1))
    ALLOCATE (rm_disp(0:numnodes-1))
    ! Count initial sizes for sending.
    !
    send_count(:,:) = 0
    CALL dbcsr_iterator_start (iter, matrix)
    dst_p = -1
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, blk, tr,&
            row_size=row_size, col_size=col_size)
       sym_tr = .FALSE.
       CALL dbcsr_get_stored_coordinates (redist%m,&
            row, col, dst_p)
       nze = row_size*col_size
       send_count(1, dst_p) = send_count(1, dst_p) + 1
       send_count(2, dst_p) = send_count(2, dst_p) + nze
    ENDDO
    CALL dbcsr_iterator_stop (iter)
    CALL mp_alltoall(send_count, recv_count, 2, mp_group)
    ! Allocate data structures needed for data exchange.
    CALL dbcsr_data_init (recv_data)
    CALL dbcsr_data_new (recv_data, data_type, SUM(recv_count(2, :)))
    ALLOCATE (recv_meta(metalen*SUM(recv_count(1, :))))
    CALL dbcsr_data_init (send_data)
    CALL dbcsr_data_new (send_data, data_type, SUM(send_count(2, :)))
    ALLOCATE (send_meta(metalen*SUM(send_count(1, :))))
    ! Fill in the meta data structures and copy the data.
    DO dst_p = 0, numnodes-1
       total_send_count(1, dst_p) = send_count (1, dst_p)
       total_send_count(2, dst_p) = send_count (2, dst_p)
       total_recv_count(1, dst_p) = recv_count (1, dst_p)
       total_recv_count(2, dst_p) = recv_count (2, dst_p)
    ENDDO
    sd_disp = -1 ; sm_disp = -1
    rd_disp = -1 ; rm_disp = -1
    sd_disp(0) = 1 ; sm_disp(0) = 1
    rd_disp(0) = 1 ; rm_disp(0) = 1
    DO dst_p = 1, numnodes-1
       sm_disp(dst_p) = sm_disp(dst_p-1)&
                        + metalen*total_send_count(1, dst_p-1)
       sd_disp(dst_p) = sd_disp(dst_p-1)&
                        + total_send_count(2, dst_p-1)
       rm_disp(dst_p) = rm_disp(dst_p-1)&
                        + metalen*total_recv_count(1, dst_p-1)
       rd_disp(dst_p) = rd_disp(dst_p-1)&
                        + total_recv_count(2, dst_p-1)
    ENDDO
    sdp(:) = sd_disp     ! sdp points to the the next place to store
                         ! data. It is postincremented.
    smp(:) = sm_disp - metalen  ! But smp points to the "working" data, not
                                ! the next. It is pre-incremented, so we must
                                ! first rewind it.
    CALL dbcsr_data_init (data_block)
    CALL dbcsr_data_new (data_block, data_type)
    CALL dbcsr_iterator_start (iter, matrix)
    dst_p = -1
    DO WHILE (dbcsr_iterator_blocks_left (iter))
       CALL dbcsr_iterator_next_block (iter, row, col, data_block, tr, blk)
       !IF (tr) WRITE(*,*)"block at",row,col," is transposed"
       sym_tr = .FALSE.
       CALL dbcsr_get_stored_coordinates (redist%m,&
            row, col, dst_p)
       smp(dst_p) = smp(dst_p) + metalen
       IF (tr) THEN
          send_meta(smp(dst_p)) = -row
       ELSE
          send_meta(smp(dst_p)) = row
       ENDIF
       send_meta(smp(dst_p)+1) = col ! new blocked column
       nze = dbcsr_data_get_size (data_block)
       CALL dbcsr_data_set (send_data, lb=sdp(dst_p), data_size=nze,&
            src=data_block, source_lb=1)
       !send_data(sdp(dst_p):sdp(dst_p)+SIZE(r_dp)-1) &
       !     = r_dp(:)
       sdp(dst_p) = sdp(dst_p) + nze
    ENDDO
    CALL dbcsr_iterator_stop (iter)
    CALL dbcsr_data_clear_pointer (data_block)
    ! Exchange the data and metadata structures.
    SELECT CASE (data_type)
    CASE (dbcsr_type_real_4)
       CALL hybrid_alltoall_s1(&
            send_data%d%r_sp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%r_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_real_8)
       !CALL mp_alltoall(&
       !     send_data%d%r_dp(:), total_send_count(2,:), sd_disp(:)-1,&
       !     recv_data%d%r_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
       !     mp_group)
       CALL hybrid_alltoall_d1 (&
            send_data%d%r_dp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%r_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_complex_4)
       CALL hybrid_alltoall_c1(&
            send_data%d%c_sp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%c_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    CASE (dbcsr_type_complex_8)
       CALL hybrid_alltoall_z1(&
            send_data%d%c_dp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%c_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_obj_new)
    END SELECT
    !CALL mp_alltoall(send_data(:), total_send_count(2,:), sd_disp(:)-1,&
    !     recv_data(:), total_recv_count(2,:), rd_disp(:)-1, mp_group)
    CALL hybrid_alltoall_i1(send_meta(:), metalen*total_send_count(1,:), sm_disp(:)-1,&
         recv_meta(:), metalen*total_recv_count(1,:), rm_disp(:)-1, mp_obj_new)
    ! Now fill in the data.
    CALL dbcsr_work_create(redist,&
            SUM(recv_count(1,:)),&
            SUM(recv_count(2,:)), work_mutable=.FALSE., n=1)
    !
    blk_ps = 1
    blks = 0
    DO src_p = 0, numnodes-1
       !data_offset_l = rd_disp(src_p)
       DO meta_l = 1, recv_count(1, src_p)
          row = recv_meta(rm_disp(src_p)+metalen*(meta_l-1))
          tr = row .LT. 0
          stored_row_new = ABS(row)
          stored_col_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+1)
          nze = row_blk_size_new(stored_row_new) * col_blk_size_new(stored_col_new)
          !r_dp => recv_data(blk_ps:blk_ps+nze-1)
          !CALL dbcsr_put_block(redist, stored_row_new, stored_col_new, r_dp, tr)
          !### this should be changed to be like the make images (i.e., copy data in finalize, not here & now)
          data_block = pointer_view (data_block, recv_data, blk_ps, nze)
          CALL dbcsr_put_block(redist, stored_row_new, stored_col_new, data_block, transposed=tr)
          blk_ps = blk_ps + nze
          blks = blks + 1
       ENDDO
    ENDDO
    CALL dbcsr_data_clear_pointer (data_block)
    CALL dbcsr_data_release (data_block)
    !
    IF (dbg) THEN
       WRITE(*,*)routineN//" Declared blocks=",redist%m%wms(1)%lastblk,&
            "actual=",blks
       WRITE(*,*)routineN//" Declared data size=",redist%m%wms(1)%datasize,&
            "actual=",blk_ps
    ENDIF
    CALL dbcsr_finalize(redist)
    CALL dbcsr_data_release (recv_data)
    CALL dbcsr_data_release (send_data)
    DEALLOCATE(send_count)
    DEALLOCATE(recv_count)
    DEALLOCATE(sdp); DEALLOCATE(sd_disp)
    DEALLOCATE(smp); DEALLOCATE(sm_disp)
    DEALLOCATE(rd_disp)
    DEALLOCATE(rm_disp)
    DEALLOCATE(recv_meta)
    DEALLOCATE(send_meta)
    CALL timestop(handle)
  END SUBROUTINE dbcsr_redistribute

! **************************************************************************************************
!> \brief Redistributes data blocks of a DBCSR matrix read from a file. 
!>        This routine should be used with dbcsr_binary_read in the module 
!>        dbcsr_io.F 
!>
!> \param[in] dblk data blocks of the DBCSR matrix that the current node possesses after reading the 
!>    data file
!> \param[in] row_p row_p of the DBCSR matrix that the current node possesses after reading the 
!>    data file
!> \param[in] col_i col_i of the DBCSR matrix that the current node possesses after reading the 
!>    data file
!> \param[in] blk_p blk_p of the DBCSR matrix that the current node possesses after reading the 
!>    data file
!> \param[in] proc_nblks 1D array holding nblks
!>    of those nodes of the mp environment, that created the file, 
!>    whose contents have been read by the current node of the present mp environment
!> \param[in] proc_darea_sizes 1D array holding data_area_size
!>    of those nodes of the mp environment, that created the file, 
!>    whose contents have been read by the current node of the present mp environment
!> \param[in,out] new_matrix   redistributed matrix 
!> 
!> \par History
!>      11.2012 created [Hossein Bani-Hashemian]
! **************************************************************************************************
  SUBROUTINE dbcsr_datablock_redistribute (dblk, row_p, col_i, blk_p,&
                                           proc_nblks, proc_darea_sizes, new_matrix)

    TYPE(dbcsr_data_obj), INTENT(IN)         :: dblk
    INTEGER, DIMENSION(:), INTENT(IN), &
      POINTER                                :: row_p, col_i, blk_p, &
                                                proc_nblks, proc_darea_sizes
    TYPE(dbcsr_obj), INTENT(INOUT)           :: new_matrix

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_datablock_redistribute', &
      routineP = moduleN//':'//routineN
    INTEGER, PARAMETER                       :: metalen = 2

    COMPLEX(kind=dp), DIMENSION(:), POINTER  :: c_dp
    COMPLEX(kind=sp), DIMENSION(:), POINTER  :: c_sp
    INTEGER :: bcol, blk, blk_ps, blk_size, blks, brow, col_size, data_type, &
      dst_p, handle, i, ind, job_count, meta_l, mp_group, &
      nblkrows_total, numnodes, row_size, src_p, stored_col_new, &
      stored_row_new
    INTEGER(kind=int_8)                      :: actual_blk, blkp, end_ind, &
                                                start_ind
    INTEGER(kind=int_8), ALLOCATABLE, &
      DIMENSION(:)                           :: extra_darea_size, extra_nblks
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: rd_disp, recv_meta, rm_disp, &
                                                sd_disp, sdp, send_meta, &
                                                sm_disp, smp
    INTEGER, ALLOCATABLE, DIMENSION(:, :)    :: recv_count, send_count, &
                                                total_recv_count, &
                                                total_send_count
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size
    LOGICAL                                  :: sym_tr, tr
    REAL(kind=dp), DIMENSION(:), POINTER     :: r_dp
    REAL(kind=sp), DIMENSION(:), POINTER     :: r_sp
    TYPE(dbcsr_data_obj)                     :: data_block, recv_data, &
                                                send_data
    TYPE(dbcsr_distribution_obj)             :: dist
    TYPE(dbcsr_mp_obj)                       :: mp_env

    CALL timeset(routineN, handle)


     dist = dbcsr_distribution (new_matrix)
     mp_env = dbcsr_distribution_mp (dist)
     numnodes = dbcsr_mp_numnodes (mp_env)
     mp_group = dbcsr_mp_group (mp_env)

     data_type = dbcsr_get_data_type(new_matrix)
     nblkrows_total = dbcsr_nblkrows_total(new_matrix)
     row_blk_size => array_data (new_matrix%m%row_blk_size)
     col_blk_size => array_data (new_matrix%m%col_blk_size)

     ALLOCATE (send_count(2, 0:numnodes-1))
     ALLOCATE (recv_count(2, 0:numnodes-1))
     ALLOCATE (total_send_count(2, 0:numnodes-1))
     ALLOCATE (total_recv_count(2, 0:numnodes-1))
     ALLOCATE (sdp(0:numnodes-1))
     ALLOCATE (sd_disp(0:numnodes-1))
     ALLOCATE (smp(0:numnodes-1))
     ALLOCATE (sm_disp(0:numnodes-1))
     ALLOCATE (rd_disp(0:numnodes-1))
     ALLOCATE (rm_disp(0:numnodes-1))
     send_count(:,:) = 0
     dst_p = -1

     job_count = COUNT(proc_nblks .NE. 0)

     ALLOCATE(extra_nblks(job_count))
     ALLOCATE(extra_darea_size(job_count))
     IF (job_count>0) THEN 
       CALL cumsum_l(INT((/0,proc_nblks(1:job_count-1)/),kind=int_8),extra_nblks)
       CALL cumsum_l(INT((/0,proc_darea_sizes(1:job_count-1)/),kind=int_8),extra_darea_size)
     ENDIF

     i = 0
     DO ind = 1, job_count*nblkrows_total
        brow = MOD(ind-1,nblkrows_total)+1
        IF (brow .EQ. 1) i = i+1
        row_size = row_blk_size(brow)
        DO blk = row_p(ind+i-1)+1, row_p(ind+i)
           actual_blk = INT(blk,kind=int_8)+extra_nblks(i)
           bcol = col_i(actual_blk)
           col_size = col_blk_size(bcol)
           blk_size = row_size*col_size
           sym_tr = .FALSE.
           CALL dbcsr_get_stored_coordinates (new_matrix%m, brow, bcol, dst_p)
           send_count(1, dst_p) = send_count(1, dst_p) + 1
           send_count(2, dst_p) = send_count(2, dst_p) + blk_size
        ENDDO
     ENDDO
     CALL mp_alltoall(send_count, recv_count, 2, mp_group)
     CALL dbcsr_data_init (recv_data)
     CALL dbcsr_data_new (recv_data, data_type, SUM(recv_count(2, :)))
     ALLOCATE (recv_meta(metalen*SUM(recv_count(1, :))))
     CALL dbcsr_data_init (send_data)
     CALL dbcsr_data_new (send_data, data_type, SUM(send_count(2, :)))
     ALLOCATE (send_meta(metalen*SUM(send_count(1, :))))
     DO dst_p = 0, numnodes-1
        total_send_count(1, dst_p) = send_count (1, dst_p)
        total_send_count(2, dst_p) = send_count (2, dst_p)
        total_recv_count(1, dst_p) = recv_count (1, dst_p)
        total_recv_count(2, dst_p) = recv_count (2, dst_p)
     ENDDO
     sd_disp = -1 ; sm_disp = -1 ; rd_disp = -1 ; rm_disp = -1                                                             
     sd_disp(0) = 1 ; sm_disp(0) = 1 ; rd_disp(0) = 1 ; rm_disp(0) = 1                        
     DO dst_p = 1, numnodes-1                               
        sm_disp(dst_p) = sm_disp(dst_p-1) + metalen*total_send_count(1, dst_p-1)
        sd_disp(dst_p) = sd_disp(dst_p-1) + total_send_count(2, dst_p-1)
        rm_disp(dst_p) = rm_disp(dst_p-1) + metalen*total_recv_count(1, dst_p-1)
        rd_disp(dst_p) = rd_disp(dst_p-1) + total_recv_count(2, dst_p-1)
     ENDDO
     sdp(:) = sd_disp     
     smp(:) = sm_disp - metalen  
    
     SELECT CASE (data_type)
     CASE (dbcsr_type_real_4)
          r_sp => dblk%d%r_sp
     CASE (dbcsr_type_real_8)
          r_dp => dblk%d%r_dp
     CASE (dbcsr_type_complex_4)
          c_sp => dblk%d%c_sp
     CASE (dbcsr_type_complex_8)
          c_dp => dblk%d%c_dp
     END SELECT 

     CALL dbcsr_data_init (data_block)
     CALL dbcsr_data_new (data_block, data_type)
     i = 0
     dst_p = -1
     DO ind = 1, job_count*nblkrows_total
        brow = MOD(ind-1,nblkrows_total)+1
        IF (brow .EQ. 1) i = i+1
        row_size = row_blk_size(brow)
        DO blk = row_p(ind+i-1)+1, row_p(ind+i)
           actual_blk = INT(blk,kind=int_8)+extra_nblks(i)
           bcol = col_i(actual_blk)
           col_size = col_blk_size(bcol)
           blk_size = row_size*col_size
           blkp = INT(blk_p(actual_blk),kind=int_8)
           start_ind = blkp+extra_darea_size(i)
           end_ind = blkp+extra_darea_size(i)+blk_size-1

           SELECT CASE (data_type)
           CASE (dbcsr_type_real_4)
               data_block%d%r_sp => r_sp(start_ind:end_ind)
           CASE (dbcsr_type_real_8)
               data_block%d%r_dp => r_dp(start_ind:end_ind)
           CASE (dbcsr_type_complex_4)
               data_block%d%c_sp => c_sp(start_ind:end_ind)
           CASE (dbcsr_type_complex_8)
               data_block%d%c_dp => c_dp(start_ind:end_ind)
           END SELECT
           sym_tr = .FALSE.
           CALL dbcsr_get_stored_coordinates(new_matrix%m, brow, bcol, dst_p)
           smp(dst_p) = smp(dst_p) + metalen
           tr=.FALSE.
!           IF (tr) THEN
!              send_meta(smp(dst_p)) = -brow
!           ELSE
              send_meta(smp(dst_p)) = brow
!           ENDIF
           send_meta(smp(dst_p)+1) = bcol 
           blk_size = dbcsr_data_get_size (data_block)

           CALL dbcsr_data_set (send_data, lb=sdp(dst_p),&
                                data_size=blk_size, src=data_block, source_lb=1)
           sdp(dst_p) = sdp(dst_p) + blk_size
        ENDDO
     ENDDO
     CALL dbcsr_data_clear_pointer (data_block)

     SELECT CASE (data_type)
     CASE (dbcsr_type_real_4)
        CALL hybrid_alltoall_s1(&
            send_data%d%r_sp(:), total_send_count(2,:), sd_disp(:)-1,&
            recv_data%d%r_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
            mp_env)
     CASE (dbcsr_type_real_8)
        CALL hybrid_alltoall_d1 (&
           send_data%d%r_dp(:), total_send_count(2,:), sd_disp(:)-1,&
           recv_data%d%r_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
           mp_env)
     CASE (dbcsr_type_complex_4)
        CALL hybrid_alltoall_c1(&
           send_data%d%c_sp(:), total_send_count(2,:), sd_disp(:)-1,&
           recv_data%d%c_sp(:), total_recv_count(2,:), rd_disp(:)-1,&
           mp_env)
     CASE (dbcsr_type_complex_8)
        CALL hybrid_alltoall_z1(&
          send_data%d%c_dp(:), total_send_count(2,:), sd_disp(:)-1,&
          recv_data%d%c_dp(:), total_recv_count(2,:), rd_disp(:)-1,&
          mp_env)
     END SELECT         
     CALL hybrid_alltoall_i1(send_meta(:), metalen*total_send_count(1,:), sm_disp(:)-1,&
          recv_meta(:), metalen*total_recv_count(1,:), rm_disp(:)-1, mp_env)
     CALL dbcsr_work_create(new_matrix,SUM(recv_count(1,:)),&
                            SUM(recv_count(2,:)), work_mutable=.FALSE., n=1)

     blk_ps = 1
     blks = 0
     DO src_p = 0, numnodes-1
        DO meta_l = 1, recv_count(1, src_p)
           brow = recv_meta(rm_disp(src_p)+metalen*(meta_l-1))
           tr = brow .LT. 0
           stored_row_new = ABS(brow)
           stored_col_new = recv_meta(rm_disp(src_p)+metalen*(meta_l-1)+1)
           blk_size = row_blk_size(stored_row_new) * col_blk_size(stored_col_new)

           data_block = pointer_view (data_block, recv_data, blk_ps, blk_size)
           CALL dbcsr_put_block(new_matrix, stored_row_new, stored_col_new, data_block, transposed=tr)

           blk_ps = blk_ps + blk_size
           blks = blks + 1
        ENDDO
     ENDDO
     CALL dbcsr_data_clear_pointer (data_block)
     DEALLOCATE (data_block%d)
     CALL dbcsr_finalize(new_matrix, reshuffle=.TRUE.)
     CALL dbcsr_data_release (recv_data)
     CALL dbcsr_data_release (send_data)

     DEALLOCATE(send_count)
     DEALLOCATE(recv_count)
     DEALLOCATE(sdp); DEALLOCATE(sd_disp)
     DEALLOCATE(smp); DEALLOCATE(sm_disp)
     DEALLOCATE(rd_disp)
     DEALLOCATE(rm_disp)
     DEALLOCATE(recv_meta)
     DEALLOCATE(send_meta)
     DEALLOCATE(extra_nblks); DEALLOCATE(extra_darea_size)

    CALL timestop(handle)
    CONTAINS
    
! **************************************************************************************************
!> \brief ...
!> \param arr ...
!> \param cumsum ...
! **************************************************************************************************
    SUBROUTINE cumsum_l(arr,cumsum)
      INTEGER(kind=int_8), DIMENSION(:), INTENT(IN)      :: arr
      INTEGER(kind=int_8), DIMENSION(:), INTENT(OUT)     :: cumsum

      INTEGER                                            :: i

      IF (SIZE(cumsum)>0) THEN
        cumsum(1)=arr(1)
        DO i=2,SIZE(cumsum)
          cumsum(i)=cumsum(i-1)+arr(i)
        END DO
      ENDIF
    END SUBROUTINE cumsum_l
  END SUBROUTINE dbcsr_datablock_redistribute

END MODULE dbcsr_transformations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief DBCSR to CSR matrix format conversion
!> \par History
!>       2014.10 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************

MODULE dbcsr_csr_conversions

   USE dbcsr_block_access,              ONLY: dbcsr_put_block
   USE dbcsr_data_methods,              ONLY: dbcsr_data_clear_pointer,&
                                              dbcsr_data_init,&
                                              dbcsr_data_new,&
                                              dbcsr_data_release
   USE dbcsr_data_types,                ONLY: dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8,&
                                              dbcsr_type_real_default
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_col_dist,&
                                              dbcsr_distribution_init,&
                                              dbcsr_distribution_mp,&
                                              dbcsr_distribution_new,&
                                              dbcsr_distribution_release
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_type, dbcsr_get_num_blocks, &
        dbcsr_get_nze, dbcsr_has_symmetry, dbcsr_init, dbcsr_is_initialized, dbcsr_name, &
        dbcsr_nblkcols_total, dbcsr_nblkrows_total, dbcsr_nfullrows_local, dbcsr_release, &
        dbcsr_row_block_sizes, dbcsr_valid_index
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_group,&
                                              dbcsr_mp_mynode,&
                                              dbcsr_mp_new,&
                                              dbcsr_mp_numnodes,&
                                              dbcsr_mp_release
   USE dbcsr_operations,                ONLY: dbcsr_add,&
                                              dbcsr_copy,&
                                              dbcsr_get_info,&
                                              dbcsr_norm,&
                                              dbcsr_set
   USE dbcsr_transformations,           ONLY: dbcsr_complete_redistribute,&
                                              dbcsr_desymmetrize_deep
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_distribution_obj,&
                                              dbcsr_iterator,&
                                              dbcsr_mp_obj,&
                                              dbcsr_norm_maxabsnorm,&
                                              dbcsr_obj
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8,&
                                              real_4,&
                                              real_8,&
                                              sp
   USE message_passing,                 ONLY: mp_environ,&
                                              mp_gather,&
                                              mp_recv,&
                                              mp_send,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_csr_conversions'

   LOGICAL, PARAMETER, PRIVATE          :: careful_mod = .FALSE.

   INTEGER, PARAMETER, PUBLIC           :: csr_dbcsr_blkrow_dist = 1, csr_eqrow_ceil_dist = 2, csr_eqrow_floor_dist = 3

! **************************************************************************************************
!> \brief Mapping data relating local CSR indices to local indices of a block-row
!>          distributed (BRD) DBCSR matrix, and containing the block structure
!>          of the original DBCSR matrix from which the CSR matrix was created.
!> \param csr_to_brd_ind    csr_to_brd_ind(csr_ind) gives the location of a
!>                          matrix element with CSR index csr_ind (location in
!>                          nzval_local) inside the data_area of the corresponding
!>                          BRD matrix. If an element of the DBCSR matrix is
!>                          treated as 0 in the CSR format, the index of this value
!>                          is not in csr_to_brd_ind.
!> \param brd_to_csr_ind    same as csr_to_brd_ind but inverse mapping. If a
!>                          given DBCSR index dbcsr_ind points to a zero element,
!>                          then brd_to_csr_ind(dbcsr_ind) is -1.
!> \param brd_mat DBCSR     BRD matrix acting as an intermediate step in any
!>                          conversion from and to DBCSR format.
!> \param dbcsr_*           data from original DBCSR matrix (not block-row
!>                          distributed), representing the original block
!>                          structure.
!> \param has_dbcsr_block_data    whether dbcsr_* fields are defined
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   TYPE csr_mapping_data
      PRIVATE
      INTEGER, DIMENSION(:), POINTER           :: csr_to_brd_ind => NULL(), &
                                                  brd_to_csr_ind => NULL()
      TYPE(dbcsr_obj)                          :: brd_mat

      LOGICAL                                  :: has_dbcsr_block_data = .FALSE.
      INTEGER                                  :: dbcsr_nblkcols_total, &
                                                  dbcsr_nblkrows_total, &
                                                  dbcsr_nblks_local
      INTEGER, DIMENSION(:), POINTER           :: dbcsr_row_p, dbcsr_col_i, &
                                                  dbcsr_row_blk_size, dbcsr_col_blk_size
   ENDTYPE

! **************************************************************************************************
!> \brief Data type of CSR matrices
!> \param r_sp            real, single precision data array
!> \param r_dp            real, double precision data array
!> \param c_sp            complex, single precision data array
!> \param c_sp            complex, double precision data array
!> \param data_type       data type of CSR matrix
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   TYPE csr_data_area_type
      REAL(KIND=real_4), DIMENSION(:), POINTER      :: r_sp => Null()
      REAL(KIND=real_8), DIMENSION(:), POINTER      :: r_dp => Null()
      COMPLEX(KIND=real_4), DIMENSION(:), POINTER   :: c_sp => Null()
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER   :: c_dp => Null()
      INTEGER                                       :: data_type = -1
   ENDTYPE

! **************************************************************************************************
!> \brief Type for CSR matrices
!> \param nrows_total      total number of rows
!> \param ncols_total      total number of columns
!> \param nze_total        total number of nonzero elements
!> \param nze_local        local number of nonzero elements
!> \param nrows_local      local number of rows
!> \param nzval_local      values of local non-zero elements, row-wise ordering.
!> \param rowptr_local     indices of elements inside nzval_local starting a row
!> \param colind_local     column indices of elements inside nzval_local
!> \param dbcsr_mapping    mapping data relating indices of nzval_local to indices
!>                         of a block-row distributed DBCSR matrix
!> \param has_mapping      whether dbcsr_mapping is defined
!> \param valid            whether essential data (excluding dbcsr_mapping) is
!>                         completely allocated
!> \param has_indices      whether rowptr_local and colind_local are defined
!> \param mp_group         message-passing group ID
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   TYPE csr_type
      INTEGER                                  :: nrows_total, ncols_total, &
                                                  nze_local, nrows_local, mp_group
      INTEGER(KIND=int_8)                      :: nze_total
      INTEGER, DIMENSION(:), POINTER           :: rowptr_local => NULL(), &
                                                  colind_local => NULL(), &
                                                  nzerow_local => NULL()
      TYPE(csr_data_area_type)                 :: nzval_local
      TYPE(csr_mapping_data)                   :: dbcsr_mapping
      LOGICAL                                  :: has_mapping = .FALSE.
      LOGICAL                                  :: valid = .FALSE.
      LOGICAL                                  :: has_indices = .FALSE.
   ENDTYPE csr_type

   TYPE csr_p_type
      TYPE(csr_type), POINTER                  :: csr_mat
   END TYPE csr_p_type

   PUBLIC :: csr_type, csr_p_type, convert_csr_to_dbcsr, &
             csr_create_from_dbcsr, &
             csr_destroy, csr_conversion_test, &
             convert_dbcsr_to_csr, &
             csr_create_new, csr_create_template, &
             csr_print_sparsity

   INTERFACE csr_create
      MODULE PROCEDURE csr_create_new, csr_create_template
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Create a new CSR matrix and allocate all internal data (excluding dbcsr_mapping)
!> \param[out] csr_mat       CSR matrix to return
!> \param[in] nrows_total    total number of rows
!> \param[in] ncols_total    total number of columns
!> \param[in] nze_total      total number of non-zero elements
!> \param[in] nze_local      local number of non-zero elements
!> \param[in] nrows_local    local number of rows
!> \param mp_group ...
!> \param[in] data_type      data type of the CSR matrix (default real double prec.)
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_create_new(csr_mat, nrows_total, ncols_total, nze_total, &
                             nze_local, nrows_local, mp_group, data_type)
      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      INTEGER, INTENT(IN)                                :: nrows_total, ncols_total
      INTEGER(KIND=int_8)                                :: nze_total
      INTEGER, INTENT(IN)                                :: nze_local, nrows_local, mp_group
      INTEGER, INTENT(IN), OPTIONAL                      :: data_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_create_new', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      CALL dbcsr_assert(nrows_total, "GE", nrows_local, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "local number of rows must not exceed total number of rows", &
                        223)

      IF (nze_total .LT. nze_local) CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",225), "local number of non-zero "// &
                                                  "elements must not exceed total number of non-zero elements")

      IF (INT(nrows_total, kind=int_8)*INT(ncols_total, kind=int_8) .LT. nze_total) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",229,"Total number of non-zero elements must not exceed total matrix size")

      CALL dbcsr_assert(nrows_local*ncols_total, "GE", nze_local, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Local number of non-zero elements must not exceed local matrix size", &
                        234)

      csr_mat%ncols_total = ncols_total
      csr_mat%nrows_total = nrows_total
      csr_mat%nze_total = nze_total
      csr_mat%nze_local = nze_local
      ALLOCATE (csr_mat%colind_local(nze_local))
      csr_mat%nrows_local = nrows_local
      ALLOCATE (csr_mat%rowptr_local(nrows_local+1))
      ALLOCATE (csr_mat%nzerow_local(nrows_local))

      IF (PRESENT(data_type)) THEN
         csr_mat%nzval_local%data_type = data_type
      ELSE
         csr_mat%nzval_local%data_type = dbcsr_type_real_default
      ENDIF

      SELECT CASE (csr_mat%nzval_local%data_type)
      CASE (dbcsr_type_real_4)
         ALLOCATE (csr_mat%nzval_local%r_sp(nze_local))
      CASE (dbcsr_type_real_8)
         ALLOCATE (csr_mat%nzval_local%r_dp(nze_local))
      CASE (dbcsr_type_complex_4)
         ALLOCATE (csr_mat%nzval_local%c_sp(nze_local))
      CASE (dbcsr_type_complex_8)
         ALLOCATE (csr_mat%nzval_local%c_dp(nze_local))
      CASE DEFAULT
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",261,"Invalid matrix type")
      END SELECT

      csr_mat%mp_group = mp_group

      csr_mat%valid = .TRUE.
      csr_mat%has_mapping = .FALSE.
      csr_mat%has_indices = .FALSE.

      CALL timestop(handle)

   END SUBROUTINE csr_create_new

! **************************************************************************************************
!> \brief Create a new CSR matrix and allocate all internal data using
!>        an existing CSR matrix. Copies the indices but no actual matrix data.
!> \param[out] matrix_b Target CSR matrix
!> \param[int] matrix_a Source CSR matrix
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_create_template(matrix_b, matrix_a)
      TYPE(csr_type), INTENT(OUT)                        :: matrix_b
      TYPE(csr_type), INTENT(IN)                         :: matrix_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_create_template', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(csr_mapping_data)                             :: map

      CALL timeset(routineN, handle)

      IF (.NOT. matrix_a%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",296,"Source CSR matrix must be created.")

      CALL csr_create_new(matrix_b, matrix_a%nrows_total, matrix_a%ncols_total, &
                          matrix_a%nze_total, matrix_a%nze_local, matrix_a%nrows_local, &
                          matrix_a%mp_group, matrix_a%nzval_local%data_type)

      matrix_b%mp_group = matrix_a%mp_group
      matrix_b%has_mapping = matrix_a%has_mapping
      matrix_b%has_indices = matrix_a%has_indices

      IF (matrix_a%has_indices) THEN
         matrix_b%rowptr_local = matrix_a%rowptr_local
         matrix_b%nzerow_local = matrix_a%nzerow_local
         matrix_b%colind_local = matrix_a%colind_local
      ENDIF

      IF (matrix_a%has_mapping) THEN
         map = matrix_a%dbcsr_mapping
         ALLOCATE (matrix_b%dbcsr_mapping%csr_to_brd_ind(SIZE(map%csr_to_brd_ind)))
         ALLOCATE (matrix_b%dbcsr_mapping%brd_to_csr_ind(SIZE(map%brd_to_csr_ind)))
         matrix_b%dbcsr_mapping%csr_to_brd_ind = map%csr_to_brd_ind
         matrix_b%dbcsr_mapping%brd_to_csr_ind = map%brd_to_csr_ind
         matrix_b%dbcsr_mapping%has_dbcsr_block_data = map%has_dbcsr_block_data
         IF (matrix_b%dbcsr_mapping%has_dbcsr_block_data) THEN
            matrix_b%dbcsr_mapping%dbcsr_nblkcols_total = map%dbcsr_nblkcols_total
            matrix_b%dbcsr_mapping%dbcsr_nblkrows_total = map%dbcsr_nblkrows_total
            ALLOCATE (matrix_b%dbcsr_mapping%dbcsr_row_blk_size(map%dbcsr_nblkrows_total))
            ALLOCATE (matrix_b%dbcsr_mapping%dbcsr_col_blk_size(map%dbcsr_nblkcols_total))
            ALLOCATE (matrix_b%dbcsr_mapping%dbcsr_row_p(map%dbcsr_nblkrows_total+1))
            ALLOCATE (matrix_b%dbcsr_mapping%dbcsr_col_i(map%dbcsr_nblks_local))
            matrix_b%dbcsr_mapping%dbcsr_nblks_local = map%dbcsr_nblks_local
            matrix_b%dbcsr_mapping%dbcsr_row_p = map%dbcsr_row_p
            matrix_b%dbcsr_mapping%dbcsr_col_i = map%dbcsr_col_i
            matrix_b%dbcsr_mapping%dbcsr_row_blk_size = map%dbcsr_row_blk_size
            matrix_b%dbcsr_mapping%dbcsr_col_blk_size = map%dbcsr_col_blk_size
         ENDIF

         CALL dbcsr_init(matrix_b%dbcsr_mapping%brd_mat)
         CALL dbcsr_copy(matrix_b%dbcsr_mapping%brd_mat, map%brd_mat)

         matrix_b%valid = .TRUE.

      ENDIF

      CALL timestop(handle)
   END SUBROUTINE csr_create_template

! **************************************************************************************************
!> \brief Copy a CSR matrix.
!> \param[out] matrix_b Target CSR matrix
!> \param[in] matrix_a Source CSR matrix
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_copy(matrix_b, matrix_a)
      TYPE(csr_type), INTENT(OUT)                        :: matrix_b
      TYPE(csr_type), INTENT(IN)                         :: matrix_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_copy', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (.NOT. matrix_a%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",362,"Source CSR matrix must be created.")
      IF (.NOT. matrix_a%has_indices) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",364,"Source CSR matrix must have indices.")

      IF (.NOT. matrix_b%valid) THEN
         CALL csr_create(matrix_b, matrix_a)
      ENDIF

      CALL dbcsr_assert(matrix_b%nzval_local%data_type, "EQ", &
                        matrix_a%nzval_local%data_type, dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "Matrices don't have the same datatype.", 372)

      CALL dbcsr_assert(matrix_b%nze_local, "EQ", &
                        matrix_a%nze_local, dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Matrices don't have the same number of local non-zero elements.", 376)

      CALL dbcsr_assert(matrix_b%nrows_local, "EQ", matrix_a%nrows_local, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Matrices don't have the same number of local rows.", 380)

      IF (.NOT. matrix_b%has_indices) THEN
         matrix_b%rowptr_local = matrix_a%rowptr_local
         matrix_b%nzerow_local = matrix_a%nzerow_local
         matrix_b%colind_local = matrix_a%colind_local
      ENDIF

      SELECT CASE (matrix_a%nzval_local%data_type)
      CASE (dbcsr_type_real_4)
         matrix_b%nzval_local%r_sp = matrix_a%nzval_local%r_sp
      CASE (dbcsr_type_real_8)
         matrix_b%nzval_local%r_dp = matrix_a%nzval_local%r_dp
      CASE (dbcsr_type_complex_4)
         matrix_b%nzval_local%c_sp = matrix_a%nzval_local%c_sp
      CASE (dbcsr_type_complex_8)
         matrix_b%nzval_local%c_dp = matrix_a%nzval_local%c_dp
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE csr_copy

! **************************************************************************************************
!> \brief create a vector containing the number of non-zero elements in each
!> row of a CSR matrix
!> \param[in] csr_mat   CSR matrix
!> \param[out] nzerow  number of non-zero elements in each row
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_create_nzerow(csr_mat, nzerow)
      TYPE(csr_type), INTENT(IN)                         :: csr_mat
      INTEGER, DIMENSION(:), INTENT(INOUT), POINTER      :: nzerow

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_create_nzerow', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k

      CALL timeset(routineN, handle)

      IF (.NOT. csr_mat%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",424,"CSR matrix must be created.")

      DO k = 1, csr_mat%nrows_local
         nzerow(k) = csr_mat%rowptr_local(k+1)-csr_mat%rowptr_local(k)
      ENDDO

      CALL timestop(handle)
   END SUBROUTINE csr_create_nzerow

! **************************************************************************************************
!> \brief destroy a CSR matrix
!> \param[in,out] csr_mat ...
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_destroy(csr_mat)
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_destroy', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(csr_mapping_data)                             :: map

      CALL timeset(routineN, handle)

      IF (.NOT. csr_mat%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",451,"CSR matrix must be created before destroying it.")

      IF (ASSOCIATED(csr_mat%rowptr_local)) DEALLOCATE (csr_mat%rowptr_local)
      IF (ASSOCIATED(csr_mat%nzerow_local)) DEALLOCATE (csr_mat%nzerow_local)
      IF (ASSOCIATED(csr_mat%colind_local)) DEALLOCATE (csr_mat%colind_local)

      IF (csr_mat%has_mapping) THEN
         map = csr_mat%dbcsr_mapping
         IF (ASSOCIATED(map%csr_to_brd_ind)) &
            DEALLOCATE (map%csr_to_brd_ind)
         IF (ASSOCIATED(map%brd_to_csr_ind)) &
            DEALLOCATE (map%brd_to_csr_ind)
         IF (ASSOCIATED(map%dbcsr_row_blk_size)) &
            DEALLOCATE (map%dbcsr_row_blk_size)
         IF (ASSOCIATED(map%dbcsr_col_blk_size)) &
            DEALLOCATE (map%dbcsr_col_blk_size)
         IF (ASSOCIATED(map%dbcsr_row_p)) &
            DEALLOCATE (map%dbcsr_row_p)
         IF (ASSOCIATED(map%dbcsr_col_i)) &
            DEALLOCATE (map%dbcsr_col_i)

         IF (dbcsr_is_initialized(map%brd_mat)) &
            CALL dbcsr_release(map%brd_mat)
      ENDIF

      IF (ASSOCIATED(csr_mat%nzval_local%r_dp)) &
         DEALLOCATE (csr_mat%nzval_local%r_dp)
      IF (ASSOCIATED(csr_mat%nzval_local%r_sp)) &
         DEALLOCATE (csr_mat%nzval_local%r_sp)
      IF (ASSOCIATED(csr_mat%nzval_local%c_sp)) &
         DEALLOCATE (csr_mat%nzval_local%c_sp)
      IF (ASSOCIATED(csr_mat%nzval_local%c_dp)) &
         DEALLOCATE (csr_mat%nzval_local%c_dp)

      csr_mat%has_mapping = .FALSE.
      csr_mat%valid = .FALSE.
      csr_mat%dbcsr_mapping%has_dbcsr_block_data = .FALSE.
      csr_mat%has_indices = .FALSE.
      csr_mat%nzval_local%data_type = -1

      CALL timestop(handle)
   END SUBROUTINE csr_destroy

! **************************************************************************************************
!> \brief Allocate the internals of a CSR matrix using data from a block-row
!>        distributed DBCSR matrix
!> \param[in] brd_mat block-row-distributed DBCSR matrix
!> \param[out] csr_mat CSR matrix
!> \param[in] csr_sparsity_brd BRD matrix representing sparsity pattern of CSR matrix
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_create_from_brd(brd_mat, csr_mat, csr_sparsity_brd)
      TYPE(dbcsr_obj), INTENT(IN)                        :: brd_mat
      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      TYPE(dbcsr_obj), INTENT(IN)                        :: csr_sparsity_brd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_create_from_brd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle, mp_group, &
                                                            nfullcols_total, nfullrows, &
                                                            nfullrows_total, nze_local
      INTEGER(KIND=int_8)                                :: nze_total
      INTEGER, DIMENSION(:), POINTER                     :: cdist, csr_index, dbcsr_index
      TYPE(dbcsr_distribution_obj)                       :: dist_current

      CALL timeset(routineN, handle)
      NULLIFY (dbcsr_index, csr_index, cdist)

      dist_current = dbcsr_distribution(brd_mat)

      mp_group = dbcsr_mp_group(dbcsr_distribution_mp(dist_current))
      cdist => dbcsr_distribution_col_dist(dist_current)

      IF (ANY(cdist .NE. 0)) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",528,"DBCSR matrix not block-row distributed.")

      ! Calculate mapping between BRD and CSR indices
      CALL csr_get_dbcsr_mapping(brd_mat, dbcsr_index, csr_index, nze_local, &
                                 csr_sparsity_brd)

      CALL dbcsr_get_info(brd_mat, nfullrows_total=nfullrows_total, &
                          nfullcols_total=nfullcols_total)

      ! Sum up local number of non-zero elements to get total number
      nze_total = nze_local
      CALL mp_sum(nze_total, mp_group)

      nfullrows = dbcsr_nfullrows_local(brd_mat)
      data_type = dbcsr_get_data_type(brd_mat)

      ! Allocate CSR matrix
      CALL csr_create_new(csr_mat, nfullrows_total, nfullcols_total, nze_total, &
                          nze_local, nfullrows, mp_group, data_type)

      csr_mat%dbcsr_mapping%brd_to_csr_ind => csr_index
      csr_mat%dbcsr_mapping%csr_to_brd_ind => dbcsr_index

      csr_mat%has_mapping = .TRUE.
      csr_mat%dbcsr_mapping%brd_mat = brd_mat

      CALL timestop(handle)
   END SUBROUTINE csr_create_from_brd

! **************************************************************************************************
!> \brief create the mapping information between a block-row distributed DBCSR
!>        matrix and the corresponding CSR matrix
!> \param[in] brd_mat          the block-row distributed DBCSR matrix
!> \param[out] dbcsr_index     csr to dbcsr index mapping
!> \param[out] csr_index       dbcsr to csr index mapping
!> \param[out] csr_nze_local   number of local non-zero elements
!> \param[in] csr_sparsity_brd     sparsity of CSR matrix represented in BRD format
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_get_dbcsr_mapping(brd_mat, dbcsr_index, csr_index, csr_nze_local, &
                                    csr_sparsity_brd)
      TYPE(dbcsr_obj), INTENT(IN)                        :: brd_mat
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: dbcsr_index, csr_index
      INTEGER, INTENT(OUT)                               :: csr_nze_local
      TYPE(dbcsr_obj), INTENT(IN)                        :: csr_sparsity_brd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_get_dbcsr_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, blkcol, blkrow, col_blk_size, csr_ind, data_type, dbcsr_ind, el_sum, &
         fullcol_sum_blkrow, handle, l, m, n, nblkrows_total, nze, prev_blk, prev_blkrow, &
         prev_row_blk_size, row_blk_offset, row_blk_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: csr_nze, nfullcol_blkrow
      INTEGER, DIMENSION(:), POINTER                     :: dbcsr_index_nozeroes
      LOGICAL                                            :: tr
      TYPE(dbcsr_iterator)                               :: iter

      CALL timeset(routineN, handle)

      NULLIFY (dbcsr_index, csr_index)

      CALL dbcsr_get_info(brd_mat, nblkrows_total=nblkrows_total)
      nze = dbcsr_get_nze(brd_mat)

      ALLOCATE (nfullcol_blkrow(nblkrows_total))
      ALLOCATE (dbcsr_index(nze))
      ALLOCATE (csr_index(nze))

      CALL dbcsr_iterator_start(iter, brd_mat, read_only=.TRUE.)
      nfullcol_blkrow = 0 ! number of non-zero full columns in each block row
      prev_blk = 0

      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, blkrow, blkcol, blk, transposed=tr, &
                                        col_size=col_blk_size)

         CALL dbcsr_assert(blk, "EQ", prev_blk+1, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "iterator is required to traverse the blocks in a row-wise fashion", &
                           609)

         prev_blk = blk

         nfullcol_blkrow(blkrow) = nfullcol_blkrow(blkrow)+col_blk_size
         IF (tr) &
            CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",615,"DBCSR block data must not be transposed")
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      el_sum = 0 ! number of elements above current block row

      prev_blkrow = 0 ! store number and size of previous block row
      prev_row_blk_size = 0

      CALL dbcsr_iterator_start(iter, brd_mat, read_only=.TRUE.)

      DO WHILE (dbcsr_iterator_blocks_left(iter))

         CALL dbcsr_iterator_next_block(iter, blkrow, blkcol, blk, transposed=tr, &
                                        row_size=row_blk_size, col_size=col_blk_size, row_offset=row_blk_offset)

         IF (blkrow .GT. prev_blkrow) THEN ! new block row
            IF (prev_blkrow .GT. 0) THEN
               el_sum = el_sum+nfullcol_blkrow(prev_blkrow)*prev_row_blk_size
            ENDIF

            ! number of non-zero full columns on the left of current block:
            fullcol_sum_blkrow = 0

            dbcsr_ind = el_sum
         ENDIF
         DO n = 1, col_blk_size !nr of columns
            DO m = 1, row_blk_size !nr of rows
               dbcsr_ind = dbcsr_ind+1
               csr_ind = (m-1)*nfullcol_blkrow(blkrow)+fullcol_sum_blkrow+n+el_sum
               dbcsr_index(csr_ind) = dbcsr_ind
               csr_index(dbcsr_ind) = csr_ind
            ENDDO
         ENDDO
         fullcol_sum_blkrow = fullcol_sum_blkrow+col_blk_size
         prev_blkrow = blkrow
         prev_row_blk_size = row_blk_size
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      ! remove BRD zero elements from CSR format
      data_type = dbcsr_get_data_type(csr_sparsity_brd)
      ALLOCATE (csr_nze(nze))

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         csr_nze(:) = INT(csr_sparsity_brd%m%data_area%d%r_sp(1:nze))
      CASE (dbcsr_type_real_8)
         csr_nze(:) = INT(csr_sparsity_brd%m%data_area%d%r_dp(1:nze))
      CASE DEFAULT
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",665,"CSR sparsity matrix must have a real datatype")
      END SELECT

      IF (ANY(csr_nze .EQ. 0)) THEN
         ALLOCATE (dbcsr_index_nozeroes(SUM(csr_nze)))
         m = 0 ! csr index if zeroes are excluded from CSR data
         DO l = 1, nze ! csr index if zeroes are included in CSR data
            IF (csr_nze(dbcsr_index(l)) .EQ. 0) THEN
               csr_index(dbcsr_index(l)) = -1
            ELSE
               m = m+1
               dbcsr_index_nozeroes(m) = dbcsr_index(l)
               csr_index(dbcsr_index(l)) = m
            ENDIF
         ENDDO
         DEALLOCATE (dbcsr_index)
         dbcsr_index => dbcsr_index_nozeroes
      ENDIF

      IF (ANY(csr_nze .EQ. 0)) THEN
         csr_nze_local = m
      ELSE
         csr_nze_local = nze
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE csr_get_dbcsr_mapping

! **************************************************************************************************
!> \brief Copies data from a CSR matrix to a block-row distributed DBCSR matrix.
!>        The DBCSR matrix must have a block structure consistent with the CSR matrix.
!> \param[in,out] brd_mat    block-row distributed DBCSR matrix
!> \param[in] csr_mat        CSR matrix
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE convert_csr_to_brd(brd_mat, csr_mat)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: brd_mat
      TYPE(csr_type), INTENT(IN)                         :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_csr_to_brd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle, ind, k, nze

      CALL timeset(routineN, handle)

      data_type = dbcsr_get_data_type(brd_mat)
      nze = dbcsr_get_nze(brd_mat)
      CALL dbcsr_data_release(brd_mat%m%data_area)
      CALL dbcsr_data_new(brd_mat%m%data_area, data_type, nze)

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         brd_mat%m%data_area%d%r_sp(1:nze) = 0.0_sp
      CASE (dbcsr_type_real_8)
         brd_mat%m%data_area%d%r_dp(1:nze) = 0.0_dp
      CASE (dbcsr_type_complex_4)
         brd_mat%m%data_area%d%c_sp(1:nze) = 0.0_sp
      CASE (dbcsr_type_complex_8)
         brd_mat%m%data_area%d%c_dp(1:nze) = 0.0_dp
      END SELECT

      DO k = 1, csr_mat%nze_local
         ind = csr_mat%dbcsr_mapping%csr_to_brd_ind(k)
         SELECT CASE (data_type)
         CASE (dbcsr_type_real_4)
            brd_mat%m%data_area%d%r_sp(ind) = csr_mat%nzval_local%r_sp(k)
         CASE (dbcsr_type_real_8)
            brd_mat%m%data_area%d%r_dp(ind) = csr_mat%nzval_local%r_dp(k)
         CASE (dbcsr_type_complex_4)
            brd_mat%m%data_area%d%c_sp(ind) = csr_mat%nzval_local%c_sp(k)
         CASE (dbcsr_type_complex_8)
            brd_mat%m%data_area%d%c_dp(ind) = csr_mat%nzval_local%c_dp(k)
         END SELECT
      ENDDO

      CALL timestop(handle)
   END SUBROUTINE convert_csr_to_brd

! **************************************************************************************************
!> \brief Convert a block-row distributed DBCSR matrix to a CSR matrix
!>        The DBCSR matrix must have a block structure consistent with the CSR matrix.
!> \param[in] brd_mat        block-row distributed DBCSR matrix
!> \param[in,out] csr_mat    CSR matrix
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE convert_brd_to_csr(brd_mat, csr_mat)
      TYPE(dbcsr_obj), INTENT(IN)                        :: brd_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_brd_to_csr', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, blkcol, blkrow, col_blk_offset, col_blk_size, csr_ind, data_type, dbcsr_ind, &
         el_sum, handle, ind_blk_data, k, local_row_ind, m, n, nblkrows_total, node_row_offset, &
         prev_blkrow, prev_row_blk_size, row_blk_offset, row_blk_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nfullcol_blkrow
      INTEGER, DIMENSION(:), POINTER                     :: colind, csr_index, dbcsr_index, nzerow, &
                                                            rowptr
      LOGICAL                                            :: new_ind, tr
      TYPE(dbcsr_data_obj)                               :: block
      TYPE(dbcsr_iterator)                               :: iter

      CALL timeset(routineN, handle)
      NULLIFY (rowptr, colind, dbcsr_index, csr_index)

      dbcsr_index => csr_mat%dbcsr_mapping%csr_to_brd_ind
      csr_index => csr_mat%dbcsr_mapping%brd_to_csr_ind

      ! CSR indices are not recalculated if indices are already defined
      new_ind = .NOT. (csr_mat%has_indices)

      IF (.NOT. csr_mat%has_mapping) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",782,"DBCSR mapping of CSR matrix must be defined")

      ! Calculate mapping between CSR matrix and DBCSR matrix if not yet defined
      !IF (.NOT. csr_mat%has_mapping ) THEN
      !  CALL csr_get_dbcsr_mapping (brd_mat, dbcsr_index, csr_index, nze)
      !ENDIF

      CALL dbcsr_get_info(brd_mat, nblkrows_total=nblkrows_total)
      ALLOCATE (nfullcol_blkrow(nblkrows_total))

      ! iteration over blocks without touching data,
      ! in order to get number of non-zero full columns in each block row
      CALL dbcsr_iterator_start(iter, brd_mat, read_only=.TRUE.)
      blkrow = 0
      nfullcol_blkrow = 0 ! number of non-zero full columns in each block row
      data_type = dbcsr_get_data_type(brd_mat)

      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, blkrow, blkcol, blk, col_size=col_blk_size, &
                                        row_offset=row_blk_offset)
         nfullcol_blkrow(blkrow) = nfullcol_blkrow(blkrow)+col_blk_size
         IF (blk .EQ. 1) THEN
            node_row_offset = row_blk_offset
         ENDIF
      ENDDO

      CALL dbcsr_iterator_stop(iter)

      ! Copy data from BRD matrix to CSR matrix and calculate CSR indices
      prev_blkrow = 0
      prev_row_blk_size = 0
      el_sum = 0 ! number of elements above current block row
      colind => csr_mat%colind_local
      rowptr => csr_mat%rowptr_local
      nzerow => csr_mat%nzerow_local
      CALL dbcsr_data_init(block)
      CALL dbcsr_data_new(block, data_type)

      CALL dbcsr_iterator_start(iter, brd_mat, read_only=.TRUE.)

      IF (new_ind) rowptr(:) = 0 ! initialize to 0 in order to check which rows are 0 at a later time
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, blkrow, blkcol, block, tr, &
                                        col_size=col_blk_size, row_size=row_blk_size, row_offset=row_blk_offset, &
                                        col_offset=col_blk_offset)

         IF (tr) &
            CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",829,"DBCSR block data must not be transposed")

         IF (blkrow > prev_blkrow) THEN ! new block row
            local_row_ind = row_blk_offset-node_row_offset ! actually: local row index - 1
            IF (prev_blkrow .GT. 0) THEN
               el_sum = el_sum+nfullcol_blkrow(prev_blkrow)*prev_row_blk_size
            ENDIF
            dbcsr_ind = el_sum
         ENDIF
         DO n = 1, col_blk_size !nr of columns
            DO m = 1, row_blk_size !nr of rows
               dbcsr_ind = dbcsr_ind+1
               csr_ind = csr_index(dbcsr_ind) ! get CSR index for current element
               IF (csr_ind .GT. 0) THEN ! is non-zero element if csr_ind > 0
                  IF (new_ind) THEN
                     ! Calculate CSR column index
                     colind(csr_ind) = col_blk_offset+n-1
                     ! Calculate CSR row pointer
                     ! (not accounting for zero elements inside non-zero blocks)
                     IF (rowptr(local_row_ind+m) .LE. 0) rowptr(local_row_ind+m) = &
                        rowptr(local_row_ind+m)+el_sum+1+nfullcol_blkrow(blkrow)*(m-1)
                  ENDIF
                  ind_blk_data = (m+row_blk_size*(n-1)) ! index of data inside DBCSR blocks
                  SELECT CASE (csr_mat%nzval_local%data_type)
                  CASE (dbcsr_type_real_4)
                     csr_mat%nzval_local%r_sp(csr_ind) = block%d%r_sp(ind_blk_data)
                  CASE (dbcsr_type_real_8)
                     csr_mat%nzval_local%r_dp(csr_ind) = block%d%r_dp(ind_blk_data)
                  CASE (dbcsr_type_complex_4)
                     csr_mat%nzval_local%c_sp(csr_ind) = block%d%c_sp(ind_blk_data)
                  CASE (dbcsr_type_complex_8)
                     csr_mat%nzval_local%c_dp(csr_ind) = block%d%c_dp(ind_blk_data)
                  END SELECT
               ELSE ! is zero element if ind = -1
                  ! CSR row pointer has to be corrected if element is zero
                  ! (subtract 1 from all subsequent row pointers)
                  IF (new_ind) rowptr(local_row_ind+m+1:) = rowptr(local_row_ind+m+1:)-1
               ENDIF
            ENDDO
         ENDDO
         prev_blkrow = blkrow
         prev_row_blk_size = row_blk_size
      ENDDO

      IF (new_ind) THEN
         ! Repeat previous row pointer for row pointers to zero rows
         IF (csr_mat%nrows_local .GT. 0) rowptr(1) = 1
         DO k = 1, csr_mat%nrows_local
            IF (rowptr(k) .LE. 0) rowptr(k) = rowptr(k-1)
         ENDDO

         rowptr(csr_mat%nrows_local+1) = csr_mat%nze_local+1
      ENDIF

      CALL csr_create_nzerow(csr_mat, nzerow)

      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(block)
      CALL dbcsr_data_release(block)

      IF (new_ind) csr_mat%has_indices = .TRUE.

      CALL timestop(handle)
   END SUBROUTINE convert_brd_to_csr

! **************************************************************************************************
!> \brief create CSR matrix including dbcsr_mapping from arbitrary DBCSR matrix
!>        in order to prepare conversion.
!> \param[in] dbcsr_mat ...
!> \param[out] csr_mat ...
!> \param[in] dist_format     how to distribute CSR rows over processes:
!>                            csr_dbcsr_blkrow_dist: the number of rows per process
!>                            is adapted to the row block sizes in the DBCSR format
!>                            such that blocks are not splitted over different processes.
!>                            csr_eqrow_ceil_dist: each process holds ceiling(N/P) CSR rows.
!>                            csr_eqrow_floor_dist: each process holds floor(N/P) CSR rows.
!> \param[in] csr_sparsity    DBCSR matrix containing 0 and 1, representing CSR sparsity pattern
!>                            1: non-zero element
!>                            0: zero element (not present in CSR format)
!>                            Note: matrix must be of data_type dbcsr_type_real_4 or
!>                            dbcsr_type_real_8 (integer types not supported)
!> \param[in] numnodes        number of nodes to use for distributing CSR matrix
!>                            (optional, default is number of nodes used for
!>                            DBCSR matrix)
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_create_from_dbcsr(dbcsr_mat, csr_mat, dist_format, csr_sparsity, numnodes)
      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat
      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      INTEGER, INTENT(IN)                                :: dist_format
      TYPE(dbcsr_obj), INTENT(IN), OPTIONAL              :: csr_sparsity
      INTEGER, INTENT(IN), OPTIONAL                      :: numnodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_create_from_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dbcsr_numnodes, handle, nblkcols_total, &
                                                            nblkrows_total, nblks_local, num_p
      LOGICAL                                            :: equal_dist, floor_dist
      TYPE(dbcsr_obj)                                    :: brd_mat, csr_sparsity_brd, &
                                                            csr_sparsity_nosym, dbcsr_mat_nosym

      CALL timeset(routineN, handle)

      IF (.NOT. dbcsr_valid_index(dbcsr_mat)) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",936,"Invalid DBCSR matrix")

      SELECT CASE (dist_format)
      CASE (csr_dbcsr_blkrow_dist)
         equal_dist = .FALSE.
         floor_dist = .FALSE.
      CASE (csr_eqrow_ceil_dist)
         equal_dist = .TRUE.
         floor_dist = .FALSE.
      CASE (csr_eqrow_floor_dist)
         equal_dist = .TRUE.
         floor_dist = .TRUE.
      END SELECT

      CALL dbcsr_init(dbcsr_mat_nosym)
      ! Conversion does not support matrices in symmetric format, therefore desymmetrize
      IF (dbcsr_has_symmetry(dbcsr_mat)) THEN
         CALL dbcsr_desymmetrize_deep(dbcsr_mat, dbcsr_mat_nosym, untransposed_data=.TRUE.)
      ELSE
         CALL dbcsr_copy(dbcsr_mat_nosym, dbcsr_mat)
      ENDIF

      CALL dbcsr_init(csr_sparsity_nosym)
      IF (PRESENT(csr_sparsity)) THEN
         IF (dbcsr_has_symmetry(csr_sparsity)) THEN
            CALL dbcsr_desymmetrize_deep(csr_sparsity, csr_sparsity_nosym, &
                                         untransposed_data=.TRUE.)
         ELSE
            CALL dbcsr_copy(csr_sparsity_nosym, csr_sparsity)
         ENDIF
      ELSE
         CALL dbcsr_create(csr_sparsity_nosym, &
                           template=dbcsr_mat_nosym, &
                           name="CSR sparsity matrix", &
                           data_type=dbcsr_type_real_8)
         CALL dbcsr_copy(csr_sparsity_nosym, dbcsr_mat_nosym)
         CALL dbcsr_set(csr_sparsity_nosym, 1.0_sp)
      ENDIF

      IF (.NOT. dbcsr_has_same_block_structure(dbcsr_mat_nosym, csr_sparsity_nosym)) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",976,"csr_sparsity and dbcsr_mat have different sparsity pattern")

      dbcsr_numnodes = dbcsr_mp_numnodes(dbcsr_distribution_mp(dbcsr_distribution(dbcsr_mat)))
      IF (PRESENT(numnodes)) THEN
         IF (numnodes .GT. dbcsr_numnodes) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",981), "Number of nodes used for CSR matrix "// &
                          "must not exceed total number of nodes")

         num_p = numnodes
      ELSE
         num_p = dbcsr_numnodes
      ENDIF

      CALL dbcsr_create_brd(dbcsr_mat_nosym, brd_mat, equal_dist, floor_dist, &
                            num_p)
      CALL dbcsr_create_brd(csr_sparsity_nosym, csr_sparsity_brd, equal_dist, floor_dist, &
                            num_p)

      ! Create CSR matrix from BRD matrix
      CALL csr_create_from_brd(brd_mat, csr_mat, csr_sparsity_brd)

      ! Store DBCSR block data inside CSR matrix
      ! (otherwise, this data is lost when converting from DBCSR to CSR)
      nblks_local = dbcsr_get_num_blocks(dbcsr_mat_nosym)
      nblkrows_total = dbcsr_nblkrows_total(dbcsr_mat_nosym)
      nblkcols_total = dbcsr_nblkcols_total(dbcsr_mat_nosym)

      csr_mat%dbcsr_mapping%dbcsr_nblkcols_total = nblkcols_total
      csr_mat%dbcsr_mapping%dbcsr_nblkrows_total = nblkrows_total
      csr_mat%dbcsr_mapping%dbcsr_nblks_local = nblks_local
      ALLOCATE (csr_mat%dbcsr_mapping%dbcsr_row_p(nblkrows_total+1))
      csr_mat%dbcsr_mapping%dbcsr_row_p = dbcsr_mat_nosym%m%row_p
      ALLOCATE (csr_mat%dbcsr_mapping%dbcsr_col_i(nblks_local))
      csr_mat%dbcsr_mapping%dbcsr_col_i = dbcsr_mat_nosym%m%col_i

      ALLOCATE (csr_mat%dbcsr_mapping%dbcsr_row_blk_size(nblkrows_total))
      ALLOCATE (csr_mat%dbcsr_mapping%dbcsr_col_blk_size(nblkcols_total))

      csr_mat%dbcsr_mapping%dbcsr_row_blk_size = dbcsr_row_block_sizes(dbcsr_mat_nosym)
      csr_mat%dbcsr_mapping%dbcsr_col_blk_size = dbcsr_col_block_sizes(dbcsr_mat_nosym)

      csr_mat%dbcsr_mapping%has_dbcsr_block_data = .TRUE.

      CALL dbcsr_release(dbcsr_mat_nosym)
      CALL dbcsr_release(csr_sparsity_nosym)
      CALL dbcsr_release(csr_sparsity_brd)

      CALL timestop(handle)
   END SUBROUTINE csr_create_from_dbcsr

! **************************************************************************************************
!> \brief Helper function to assert that two DBCSR matrices have the same block
!>        structure and same sparsity pattern
!> \param[in] matrix_a  ...
!> \param[in] matrix_b ...
!> \retval is_equal whether matrix_a and matrix_b have the same block structure
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   FUNCTION dbcsr_has_same_block_structure(matrix_a, matrix_b) RESULT(is_equal)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      LOGICAL                                            :: is_equal

      is_equal = .TRUE.

      IF (dbcsr_nblkcols_total(matrix_a) .NE. dbcsr_nblkcols_total(matrix_b)) is_equal = .FALSE.
      IF (dbcsr_nblkrows_total(matrix_a) .NE. dbcsr_nblkrows_total(matrix_b)) is_equal = .FALSE.
      IF ((matrix_a%m%nblks) .NE. (matrix_b%m%nblks)) is_equal = .FALSE.
      IF (ANY(matrix_a%m%row_p .NE. matrix_b%m%row_p)) is_equal = .FALSE.
      IF (ANY(matrix_a%m%col_i .NE. matrix_b%m%col_i)) is_equal = .FALSE.
      IF (ANY(dbcsr_row_block_sizes(matrix_a) .NE. &
              dbcsr_row_block_sizes(matrix_b))) is_equal = .FALSE.
      IF (ANY(dbcsr_row_block_sizes(matrix_a) .NE. &
              dbcsr_row_block_sizes(matrix_b))) is_equal = .FALSE.

   END FUNCTION dbcsr_has_same_block_structure

! **************************************************************************************************
!> \brief Helper function to assert that a given CSR matrix and a given DBCSR
!>        matrix are consistent before doing the conversion
!> \param[in] csr_mat ...
!> \param[in] dbcsr_mat ...
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_assert_consistency_with_dbcsr(csr_mat, dbcsr_mat)
      TYPE(csr_type), INTENT(IN)                         :: csr_mat
      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_assert_consistency_with_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(csr_mapping_data)                             :: map

      CALL timeset(routineN, handle)
      map = csr_mat%dbcsr_mapping
      IF (map%has_dbcsr_block_data) THEN
         IF (map%dbcsr_nblkcols_total .NE. dbcsr_nblkcols_total(dbcsr_mat)) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1077), &
                          "field nblkcols_total of DBCSR matrix not consistent with CSR matrix")
         IF (map%dbcsr_nblkrows_total .NE. dbcsr_nblkrows_total(dbcsr_mat)) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1080), &
                          "field nblkrows_total of DBCSR matrix not consistent with CSR matrix")
         IF (map%dbcsr_nblks_local .NE. dbcsr_mat%m%nblks) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1083), &
                          "field nblks of DBCSR matrix not consistent with CSR matrix")
         IF (ANY(map%dbcsr_row_p .NE. dbcsr_mat%m%row_p)) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1086), &
                          "field row_p of DBCSR matrix not consistent with CSR matrix")
         IF (ANY(map%dbcsr_col_i .NE. dbcsr_mat%m%col_i)) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1089), &
                          "field dbcsr_col_i of DBCSR matrix not consistent with CSR matrix")
         IF (ANY(map%dbcsr_row_blk_size .NE. dbcsr_row_block_sizes(dbcsr_mat))) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1092), &
                          "field row_blk_size of DBCSR matrix not consistent with CSR matrix")
         IF (ANY(map%dbcsr_col_blk_size .NE. dbcsr_col_block_sizes(dbcsr_mat))) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1095), &
                          "field col_blk_size of DBCSR matrix not consistent with CSR matrix")
      ELSE
         CALL cp_warn(cp__l("dbcsr/ops/dbcsr_csr_conversions.F",1098), "Can not assert consistency of the matrices "// &
                      "as no block data stored in CSR matrix.")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE csr_assert_consistency_with_dbcsr

! **************************************************************************************************
!> \brief Convert a DBCSR matrix to a CSR matrix.
!> \param[in] dbcsr_mat DBCSR matrix to convert
!> \param[in,out] csr_mat correctly allocated CSR matrix
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE convert_dbcsr_to_csr(dbcsr_mat, csr_mat)
      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_dbcsr_to_csr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dbcsr_obj)                                    :: dbcsr_mat_nosym

      CALL timeset(routineN, handle)

      IF (.NOT. dbcsr_valid_index(dbcsr_mat)) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1125,"Invalid DBCSR matrix")
      CALL dbcsr_assert(dbcsr_get_data_type(dbcsr_mat), "EQ", &
                        csr_mat%nzval_local%data_type, dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "DBCSR and CSR matrix must have same type", 1128)

      IF (.NOT. csr_mat%has_mapping) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1131,"CSR_mat must contain mapping to DBCSR format")

      IF (dbcsr_has_symmetry(dbcsr_mat)) THEN
         CALL dbcsr_init(dbcsr_mat_nosym)
         CALL dbcsr_desymmetrize_deep(dbcsr_mat, dbcsr_mat_nosym, untransposed_data=.TRUE.)
      ELSE
         dbcsr_mat_nosym = dbcsr_mat
      ENDIF

      CALL csr_assert_consistency_with_dbcsr(csr_mat, dbcsr_mat_nosym)

      ! 1) DBCSR -> BRD
      CALL dbcsr_complete_redistribute(dbcsr_mat_nosym, csr_mat%dbcsr_mapping%brd_mat)
      ! 2) BRD -> CSR
      CALL convert_brd_to_csr(csr_mat%dbcsr_mapping%brd_mat, csr_mat)

      IF (dbcsr_has_symmetry(dbcsr_mat)) CALL dbcsr_release(dbcsr_mat_nosym)

      CALL timestop(handle)
   END SUBROUTINE convert_dbcsr_to_csr

! **************************************************************************************************
!> \brief convert a CSR matrix to a DBCSR matrix
!> \param[in,out] dbcsr_mat correctly allocated DBCSR matrix
!> \param[in,out] csr_mat CSR matrix to convert
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE convert_csr_to_dbcsr(dbcsr_mat, csr_mat)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: dbcsr_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_csr_to_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dbcsr_obj)                                    :: dbcsr_mat_nosym

      CALL timeset(routineN, handle)

      IF (.NOT. dbcsr_valid_index(dbcsr_mat)) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1174,"Invalid DBCSR matrix")

      CALL dbcsr_assert(dbcsr_get_data_type(dbcsr_mat), "EQ", &
                        csr_mat%nzval_local%data_type, dbcsr_fatal_level, &
                        dbcsr_wrong_args_error, routineN, "DBCSR and CSR matrix must have same type", &
                        1179)

      IF (.NOT. csr_mat%has_mapping) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1182,"CSR_mat must contain mapping to DBCSR format")

      ! Desymmetrize to assert that DBCSR matrix has sparsity pattern consistent with CSR matrix
      IF (dbcsr_has_symmetry(dbcsr_mat)) THEN
         CALL dbcsr_init(dbcsr_mat_nosym)
         CALL dbcsr_desymmetrize_deep(dbcsr_mat, dbcsr_mat_nosym, untransposed_data=.TRUE.)
      ELSE
         dbcsr_mat_nosym = dbcsr_mat
      ENDIF

      CALL csr_assert_consistency_with_dbcsr(csr_mat, dbcsr_mat_nosym)

      IF (dbcsr_has_symmetry(dbcsr_mat)) CALL dbcsr_release(dbcsr_mat_nosym)

      ! 1) CSR -> BRD
      CALL convert_csr_to_brd(csr_mat%dbcsr_mapping%brd_mat, csr_mat)

      ! 2) BRD -> DBCSR
      CALL dbcsr_complete_redistribute(csr_mat%dbcsr_mapping%brd_mat, dbcsr_mat)

      CALL timestop(handle)
   END SUBROUTINE convert_csr_to_dbcsr

! **************************************************************************************************
!> \brief Apply filtering threshold eps to DBCSR blocks in order to improve
!>        CSR sparsity (currently only used for testing purposes)
!> \param dbcsr_mat ...
!> \param csr_sparsity ...
!> \param eps ...
! **************************************************************************************************
   SUBROUTINE dbcsr_to_csr_filter(dbcsr_mat, csr_sparsity, eps)
      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat
      TYPE(dbcsr_obj), INTENT(OUT)                       :: csr_sparsity
      REAL(kind=real_8), INTENT(IN)                      :: eps

      INTEGER                                            :: blkcol, blkrow, col_blk_size, data_type, &
                                                            row_blk_size
      LOGICAL                                            :: tr
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:)       :: block_abs, csr_sparsity_blk
      TYPE(dbcsr_data_obj)                               :: block
      TYPE(dbcsr_iterator)                               :: iter

!REAL(kind=real_8), DIMENSION(:), POINTER :: block

      CALL dbcsr_init(csr_sparsity)
      CALL dbcsr_create(csr_sparsity, &
                        template=dbcsr_mat, &
                        name="CSR sparsity", &
                        data_type=dbcsr_type_real_8)
      CALL dbcsr_copy(csr_sparsity, dbcsr_mat)
      CALL dbcsr_set(csr_sparsity, 1.0_dp)

      IF (eps .GT. 0.0_dp) THEN
         data_type = dbcsr_get_data_type(dbcsr_mat)
         CALL dbcsr_data_init(block)
         CALL dbcsr_data_new(block, data_type)
         CALL dbcsr_iterator_start(iter, dbcsr_mat, read_only=.TRUE.)
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            CALL dbcsr_iterator_next_block(iter, blkrow, blkcol, block, transposed=tr, &
                                           row_size=row_blk_size, col_size=col_blk_size)

            ALLOCATE (block_abs(row_blk_size*col_blk_size))
            ALLOCATE (csr_sparsity_blk(row_blk_size*col_blk_size))
            SELECT CASE (data_type)
            CASE (dbcsr_type_real_4)
               block_abs(:) = REAL(ABS(block%d%r_sp(:)), KIND=real_8)
            CASE (dbcsr_type_real_8)
               block_abs(:) = REAL(ABS(block%d%r_dp(:)), KIND=real_8)
            CASE (dbcsr_type_complex_4)
               block_abs(:) = REAL(ABS(block%d%c_sp(:)), KIND=real_8)
            CASE (dbcsr_type_complex_8)
               block_abs(:) = REAL(ABS(block%d%c_dp(:)), KIND=real_8)
            END SELECT

            csr_sparsity_blk = 1.0_dp
            WHERE (block_abs .LT. eps) csr_sparsity_blk = 0.0_dp
            CALL dbcsr_put_block(csr_sparsity, blkrow, blkcol, csr_sparsity_blk, transposed=tr)
            DEALLOCATE (csr_sparsity_blk, block_abs)
         ENDDO
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_data_clear_pointer(block)
         CALL dbcsr_data_release(block)
      ENDIF

   END SUBROUTINE dbcsr_to_csr_filter

! **************************************************************************************************
!> \brief Test the conversion by converting to CSR format and converting back,
!>        where the CSR sparsity is defined by some filtering threshold eps.
!>        The maximum norm of the differences between the original and the
!>        back-converted matrix is calculated.
!> \param[in] dbcsr_mat ...
!> \param[out] csr_mat ...
!> \param[out] norm ...
!> \param eps ...
!> \par History
!>      10.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_conversion_test(dbcsr_mat, csr_mat, norm, eps)
      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat
      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      REAL(KIND=real_8), INTENT(OUT)                     :: norm
      REAL(KIND=real_8), INTENT(IN)                      :: eps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_conversion_test', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dbcsr_obj)                                    :: csr_sparsity, dbcsr_mat_conv

      CALL timeset(routineN, handle)

      CALL dbcsr_to_csr_filter(dbcsr_mat, csr_sparsity, eps)

      CALL csr_create_from_dbcsr(dbcsr_mat, csr_mat, csr_eqrow_ceil_dist, csr_sparsity)
      CALL convert_dbcsr_to_csr(dbcsr_mat, csr_mat)

      CALL dbcsr_init(dbcsr_mat_conv)
      CALL dbcsr_copy(dbcsr_mat_conv, dbcsr_mat)

      CALL convert_csr_to_dbcsr(dbcsr_mat_conv, csr_mat)

      CALL dbcsr_add(dbcsr_mat_conv, dbcsr_mat, 1.0_dp, -1.0_dp)
      CALL dbcsr_norm(dbcsr_mat_conv, dbcsr_norm_maxabsnorm, norm_scalar=norm)

      CALL dbcsr_release(dbcsr_mat_conv)
      CALL dbcsr_release(csr_sparsity)
      CALL timestop(handle)
   END SUBROUTINE csr_conversion_test

! **************************************************************************************************
!> \brief Write a CSR matrix to file
!> \param[in] csr_mat ...
!> \param[in] unit_nr           unit number to which output is written
!> \param[in] upper_triangle    If true (default: false), write only upper
!>                              triangular part of matrix
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_write(csr_mat, unit_nr, upper_triangle)
      TYPE(csr_type), INTENT(IN)                         :: csr_mat
      INTEGER, INTENT(IN)                                :: unit_nr
      LOGICAL, INTENT(IN), OPTIONAL                      :: upper_triangle

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_write', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: data_format
      COMPLEX(KIND=real_4), ALLOCATABLE, DIMENSION(:)    :: nzval_to_master_c_sp
      COMPLEX(KIND=real_8), ALLOCATABLE, DIMENSION(:)    :: nzval_to_master_c_dp
      INTEGER                                            :: handle, i, ii, k, l, m, mynode, &
                                                            numnodes, rowind, tag1, tag2, tag3
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: colind_to_master, nzerow_to_master, &
                                                            sizes_numrowlocal, sizes_nzelocal
      LOGICAL                                            :: ut
      REAL(KIND=real_4), ALLOCATABLE, DIMENSION(:)       :: nzval_to_master_r_sp
      REAL(KIND=real_8), ALLOCATABLE, DIMENSION(:)       :: nzval_to_master_r_dp

      CALL timeset(routineN, handle)

      IF (PRESENT(upper_triangle)) THEN
         ut = upper_triangle
      ELSE
         ut = .FALSE.
      ENDIF

      IF (.NOT. csr_mat%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1350,"can not write invalid CSR matrix")

      tag1 = 0
      tag2 = 1
      tag3 = 2

      CALL mp_environ(numnodes, mynode, csr_mat%mp_group)

      ! gather sizes (number of local non-zero elements and number of local rows)
      ALLOCATE (sizes_nzelocal(numnodes))
      ALLOCATE (sizes_numrowlocal(numnodes))

      CALL mp_gather(csr_mat%nze_local, sizes_nzelocal, 0, csr_mat%mp_group)
      CALL mp_gather(csr_mat%nrows_local, sizes_numrowlocal, 0, csr_mat%mp_group)

      ! for each node, send matrix data to node 0 (master) and write data
      DO i = 0, numnodes-1
         ii = i
         IF (mynode .EQ. 0) THEN ! allocations for receiving data from node i
            ALLOCATE (colind_to_master(sizes_nzelocal(ii+1)))
            ALLOCATE (nzerow_to_master(sizes_numrowlocal(ii+1)))

            SELECT CASE (csr_mat%nzval_local%data_type)
            CASE (dbcsr_type_real_4)
               data_format = "(2(I8),E23.6E2)"
               ALLOCATE (nzval_to_master_r_sp(sizes_nzelocal(ii+1)))
            CASE (dbcsr_type_real_8)
               data_format = "(2(I8),E23.14E3)"
               ALLOCATE (nzval_to_master_r_dp(sizes_nzelocal(ii+1)))
            CASE (dbcsr_type_complex_4)
               data_format = "(2(I8),2(E23.6E2))"
               ALLOCATE (nzval_to_master_c_sp(sizes_nzelocal(ii+1)))
            CASE (dbcsr_type_complex_8)
               data_format = "(2(I8),2(E23.14E3))"
               ALLOCATE (nzval_to_master_c_dp(sizes_nzelocal(ii+1)))
            END SELECT
         ENDIF

         IF (mynode .EQ. 0) THEN ! receive at node 0
            IF (ii .EQ. 0) THEN ! data from node 0, no need for mpi routines
               colind_to_master(:) = csr_mat%colind_local(:)
               nzerow_to_master(:) = csr_mat%nzerow_local(:)
               SELECT CASE (csr_mat%nzval_local%data_type)
               CASE (dbcsr_type_real_4)
                  nzval_to_master_r_sp(:) = csr_mat%nzval_local%r_sp(:)
               CASE (dbcsr_type_real_8)
                  nzval_to_master_r_dp(:) = csr_mat%nzval_local%r_dp(:)
               CASE (dbcsr_type_complex_4)
                  nzval_to_master_c_sp(:) = csr_mat%nzval_local%c_sp(:)
               CASE (dbcsr_type_complex_8)
                  nzval_to_master_c_dp(:) = csr_mat%nzval_local%c_dp(:)
               END SELECT
            ELSE ! receive data from nodes with rank > 0
               CALL mp_recv(colind_to_master, ii, tag1, csr_mat%mp_group)
               CALL mp_recv(nzerow_to_master, ii, tag2, csr_mat%mp_group)
               SELECT CASE (csr_mat%nzval_local%data_type)
               CASE (dbcsr_type_real_4)
                  CALL mp_recv(nzval_to_master_r_sp, ii, tag3, csr_mat%mp_group)
               CASE (dbcsr_type_real_8)
                  CALL mp_recv(nzval_to_master_r_dp, ii, tag3, csr_mat%mp_group)
               CASE (dbcsr_type_complex_4)
                  CALL mp_recv(nzval_to_master_c_sp, ii, tag3, csr_mat%mp_group)
               CASE (dbcsr_type_complex_8)
                  CALL mp_recv(nzval_to_master_c_dp, ii, tag3, csr_mat%mp_group)
               END SELECT
            ENDIF
         ENDIF

         IF ((mynode .EQ. ii) .AND. (ii .NE. 0)) THEN ! send from nodes with rank > 0
            CALL mp_send(csr_mat%colind_local, 0, tag1, csr_mat%mp_group)
            CALL mp_send(csr_mat%nzerow_local, 0, tag2, csr_mat%mp_group)
            SELECT CASE (csr_mat%nzval_local%data_type)
            CASE (dbcsr_type_real_4)
               CALL mp_send(csr_mat%nzval_local%r_sp, 0, tag3, csr_mat%mp_group)
            CASE (dbcsr_type_real_8)
               CALL mp_send(csr_mat%nzval_local%r_dp, 0, tag3, csr_mat%mp_group)
            CASE (dbcsr_type_complex_4)
               CALL mp_send(csr_mat%nzval_local%c_sp, 0, tag3, csr_mat%mp_group)
            CASE (dbcsr_type_complex_8)
               CALL mp_send(csr_mat%nzval_local%c_dp, 0, tag3, csr_mat%mp_group)
            END SELECT
         ENDIF

         IF (mynode .EQ. 0) THEN ! write data received at node 0
            !WRITE(unit_nr,"(A27)") "#row ind, col ind, value"
            m = 0
            DO k = 1, sizes_numrowlocal(ii+1)
               rowind = k+SUM(sizes_numrowlocal(1:ii)) ! row index: local to global
               DO l = 1, nzerow_to_master(k)
                  m = m+1
                  IF ((.NOT. ut) .OR. (k .LE. colind_to_master(m))) THEN
                     SELECT CASE (csr_mat%nzval_local%data_type)
                     CASE (dbcsr_type_real_4)
                        WRITE (unit_nr, data_format) rowind, colind_to_master(m), &
                           nzval_to_master_r_sp(m)
                     CASE (dbcsr_type_real_8)
                        WRITE (unit_nr, data_format) rowind, colind_to_master(m), &
                           nzval_to_master_r_dp(m)
                     CASE (dbcsr_type_complex_4)
                        WRITE (unit_nr, data_format) rowind, colind_to_master(m), &
                           nzval_to_master_c_sp(m)
                     CASE (dbcsr_type_complex_8)
                        WRITE (unit_nr, data_format) rowind, colind_to_master(m), &
                           nzval_to_master_c_dp(m)
                     END SELECT
                  ENDIF
               ENDDO
            ENDDO

            DEALLOCATE (colind_to_master)
            DEALLOCATE (nzerow_to_master)

            SELECT CASE (csr_mat%nzval_local%data_type)
            CASE (dbcsr_type_real_4)
               DEALLOCATE (nzval_to_master_r_sp)
            CASE (dbcsr_type_real_8)
               DEALLOCATE (nzval_to_master_r_dp)
            CASE (dbcsr_type_complex_4)
               DEALLOCATE (nzval_to_master_c_sp)
            CASE (dbcsr_type_complex_8)
               DEALLOCATE (nzval_to_master_c_dp)
            END SELECT
         ENDIF
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE csr_write

! **************************************************************************************************
!> \brief Print CSR sparsity
!> \param csr_mat ...
!> \param unit_nr ...
!> \par History
!>      02.2015 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE csr_print_sparsity(csr_mat, unit_nr)
      TYPE(csr_type), INTENT(IN)                         :: csr_mat
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_print_sparsity', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, mynode, numnodes
      INTEGER(KIND=int_8)                                :: dbcsr_nze_total
      REAL(KIND=real_8)                                  :: dbcsr_nze_percentage, nze_percentage

      CALL timeset(routineN, handle)

      IF (.NOT. csr_mat%valid) &
         CALL cp__b("dbcsr/ops/dbcsr_csr_conversions.F",1501,"CSR matrix must be created first")

      nze_percentage = 100.0_dp*(REAL(csr_mat%nze_total, KIND=real_8) &
                                 /REAL(csr_mat%nrows_total, KIND=real_8)) &
                       /REAL(csr_mat%ncols_total, KIND=real_8)

      IF (csr_mat%has_mapping) THEN
         dbcsr_nze_total = dbcsr_get_nze(csr_mat%dbcsr_mapping%brd_mat)
         CALL mp_sum(dbcsr_nze_total, csr_mat%mp_group)
         dbcsr_nze_percentage = 100.0_dp*(REAL(dbcsr_nze_total, KIND=real_8) &
                                          /REAL(csr_mat%nrows_total, KIND=real_8)) &
                                /REAL(csr_mat%ncols_total, KIND=real_8)
      ENDIF

      CALL mp_environ(numnodes, mynode, csr_mat%mp_group)

      IF (mynode .EQ. 0) THEN
         WRITE (unit_nr, "(T15,A,T68,I13)") "Number of  CSR non-zero elements:", csr_mat%nze_total
         WRITE (unit_nr, "(T15,A,T75,F6.2)") "Percentage CSR non-zero elements:", nze_percentage
         !IF(csr_mat%has_mapping) THEN
         !  WRITE(unit_nr,"(T15,A,T75,F6.2/)") "Percentage DBCSR non-zero elements:", dbcsr_nze_percentage
         !ENDIF
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE csr_print_sparsity

! **************************************************************************************************
!> \brief Converts a DBCSR matrix to a block row distributed matrix.
!> \param[in] dbcsr_mat        DBCSR matrix to be converted
!> \param[out] brd_mat         converted matrix
!> \param[in] equal_dist       see documentation of csr_create_from_dbcsr
!> \param[in] floor_dist       see documentation of csr_create_from_dbcsr
!> \param[in] numnodes         number of nodes to use for block row distribution
!> \par History
!>      04.2013 created [Hossein Bani-Hashemian]
!>      11.2014 modified [Patrick Seewald]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dbcsr_create_brd(dbcsr_mat, brd_mat, equal_dist, floor_dist, numnodes)

      TYPE(dbcsr_obj), INTENT(IN)                        :: dbcsr_mat
      TYPE(dbcsr_obj), INTENT(OUT)                       :: brd_mat
      LOGICAL, INTENT(IN)                                :: equal_dist, floor_dist
      INTEGER, INTENT(IN)                                :: numnodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_create_brd', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: matrix_type
      CHARACTER(LEN=default_string_length)               :: matrix_name
      INTEGER :: chunk_size, data_type, handle, i, k, l, m, mp_group, mynode, nblkcols_total, &
         nblkrows_total, nfullrows_local, nfullrows_total, node_size, numnodes_total, remainder, &
         row_index, split_row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rdist_tmp, row_blk_size_new_tmp
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: pgrid
      INTEGER, DIMENSION(:), POINTER                     :: cdist, col_blk_size, rdist, &
                                                            row_blk_size, row_blk_size_new
      TYPE(dbcsr_distribution_obj)                       :: dist_current, dist_new
      TYPE(dbcsr_mp_obj)                                 :: mp_obj_current, mp_obj_new

      CALL timeset(routineN, handle)

      NULLIFY (row_blk_size, rdist, row_blk_size_new)
      CALL dbcsr_get_info(dbcsr_mat, &
                          nblkrows_total=nblkrows_total, &
                          nblkcols_total=nblkcols_total, &
                          nfullrows_total=nfullrows_total, &
                          row_blk_size=row_blk_size, &
                          col_blk_size=col_blk_size, &
                          matrix_type=matrix_type, &
                          data_type=data_type)

      matrix_name = dbcsr_name(dbcsr_mat)

      ALLOCATE (cdist(nblkcols_total))
      cdist = 0

      dist_current = dbcsr_distribution(dbcsr_mat)
      mp_obj_current = dbcsr_distribution_mp(dist_current)
      mp_group = dbcsr_mp_group(mp_obj_current)
      mynode = dbcsr_mp_mynode(mp_obj_current)
      numnodes_total = dbcsr_mp_numnodes(mp_obj_current)

      ALLOCATE (pgrid(numnodes_total, 1))

      IF (equal_dist) THEN ! Equally distribute rows over processors -> cut blocks

         ! Calculate the number of rows a processor can hold
         IF (floor_dist) THEN
            nfullrows_local = FLOOR(REAL(nfullrows_total)/numnodes)
         ELSE
            nfullrows_local = CEILING(REAL(nfullrows_total)/numnodes)
         ENDIF

         ! allocate maximum amount of memory possibly needed
         ALLOCATE (rdist_tmp(nblkrows_total+numnodes-1)) ! row distribution
         ALLOCATE (row_blk_size_new_tmp(nblkrows_total+numnodes-1)) ! new sizes of block rows

         k = 0 ! counter for block rows
         m = 0 ! node counter
         node_size = nfullrows_local ! space available on current node in number of rows
         IF (node_size .GT. 0) THEN
            DO l = 1, nblkrows_total
               split_row = row_blk_size(l) ! size of current block row (number of rows)
               DO WHILE (split_row .GE. node_size) ! cut block row and send it to two nodes
                  k = k+1
                  m = m+1
                  row_blk_size_new_tmp(k) = node_size ! size of first part of block row
                  rdist_tmp(k) = m-1 ! send first part to node m
                  split_row = split_row-node_size ! size of remaining part of block rows
                  node_size = nfullrows_local ! space available on next node
                  IF (floor_dist .AND. (m .EQ. numnodes-1)) THEN ! send all remaining rows to last node
                     node_size = nfullrows_total-(numnodes-1)*node_size
                  ENDIF
               ENDDO
               IF (split_row .GT. 0) THEN ! enough space left on next node for remaining rows
                  k = k+1
                  row_blk_size_new_tmp(k) = split_row ! size of remaining part of block row
                  rdist_tmp(k) = m ! send to next node
                  node_size = node_size-split_row ! remaining space on next node
               ENDIF
            ENDDO
         ELSE ! send everything to last node if node_size = 0
            rdist_tmp(1:nblkrows_total) = numnodes-1
            row_blk_size_new_tmp(1:nblkrows_total) = row_blk_size ! row blocks unchanged
            k = nblkrows_total
         ENDIF

         ! Copy data to correctly allocated variables
         ALLOCATE (row_blk_size_new(k))
         row_blk_size_new = row_blk_size_new_tmp(1:k)
         ALLOCATE (rdist(k))
         rdist = rdist_tmp(1:k)

      ELSE ! Leave block rows intact (do not cut)
         ALLOCATE (rdist(nblkrows_total))
         rdist = 0
         IF (numnodes .GT. nblkrows_total) THEN
            rdist = (/(i, i=0, nblkrows_total-1)/)
         ELSE
            chunk_size = FLOOR(REAL(nblkrows_total)/numnodes)
            remainder = MOD(nblkrows_total, numnodes)
            row_index = 0
            DO i = 1, chunk_size*numnodes, chunk_size
               rdist(i:i+chunk_size-1) = row_index
               row_index = row_index+1
            END DO
            IF (remainder .GT. 0) rdist(chunk_size*numnodes+1:nblkrows_total) = row_index-1
         END IF
         row_blk_size_new => row_blk_size
      ENDIF

      pgrid(:, :) = RESHAPE((/(i, i=0, numnodes_total-1)/), (/numnodes_total, 1/))
      CALL dbcsr_mp_new(mp_obj_new, pgrid, mp_group, mynode, numnodes=numnodes_total)
      CALL dbcsr_distribution_init(dist_new)
      CALL dbcsr_distribution_new(dist_new, mp_obj_new, rdist, cdist, reuse_arrays=.TRUE.)

      CALL dbcsr_init(brd_mat)
      CALL dbcsr_create(brd_mat, TRIM(matrix_name)//" row-block distributed", &
                        dist_new, matrix_type, row_blk_size_new, col_blk_size, data_type=data_type)
      CALL dbcsr_complete_redistribute(dbcsr_mat, brd_mat)

      DEALLOCATE (pgrid)

      IF (equal_dist) DEALLOCATE (row_blk_size_new)

      CALL dbcsr_distribution_release(dist_new)
      CALL dbcsr_mp_release(mp_obj_new)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_create_brd

END MODULE dbcsr_csr_conversions
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Datatypes used by the dbcsr matrix-matrix multiplication machinery.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2013-01    reorganized code (Ole Schuett)
! **************************************************************************************************
MODULE dbcsr_mm_types

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_types'

   PUBLIC :: stack_descriptor_type

   ! Definitions for the members of a parameter stack.
   INTEGER, PARAMETER, PUBLIC :: dbcsr_ps_width = 7
   INTEGER, PARAMETER, PUBLIC :: p_m = 1
   INTEGER, PARAMETER, PUBLIC :: p_n = 2
   INTEGER, PARAMETER, PUBLIC :: p_k = 3
   INTEGER, PARAMETER, PUBLIC :: p_a_first = 4
   INTEGER, PARAMETER, PUBLIC :: p_b_first = 5
   INTEGER, PARAMETER, PUBLIC :: p_c_first = 6
   INTEGER, PARAMETER, PUBLIC :: p_c_blk = 7

   TYPE stack_descriptor_type
      INTEGER :: m, n, k, max_m, max_n, max_k
      LOGICAL :: defined_mnk
   END TYPE stack_descriptor_type

END MODULE dbcsr_mm_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Accelerator support for DBCSR
!> \author  Urban Borstnik
!> \date    2011-04-06
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2011-04-06
!> - 2014-04, Ole Schuett: generalized into acc-framework
! **************************************************************************************************
MODULE dbcsr_acc_operations
   USE ISO_C_BINDING,                   ONLY: C_INT,&
                                              C_PTR
   USE acc_devmem,                      ONLY: acc_devmem_cptr,&
                                              acc_devmem_type
   USE acc_stream,                      ONLY: acc_stream_cptr,&
                                              acc_stream_type
   USE dbcsr_mm_types,                  ONLY: dbcsr_ps_width
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_acc_operations'

   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   PUBLIC :: dbcsr_acc_do_mm_stack, dbcsr_acc_transpose


CONTAINS

! **************************************************************************************************
!> \brief Launch an accelerated kernel for processing a stack.
!> \param param_stack ...
!> \param stack_size ...
!> \param datatype ...
!> \param a_data ...
!> \param b_data ...
!> \param c_data ...
!> \param m_max ...
!> \param n_max ...
!> \param k_max ...
!> \param def_mnk ...
!> \param stream ...
!> \param success ...
! **************************************************************************************************
   SUBROUTINE dbcsr_acc_do_mm_stack(param_stack, stack_size, datatype, &
                                    a_data, b_data, c_data, m_max, n_max, k_max, def_mnk, stream, success)
      TYPE(acc_devmem_type), INTENT(IN)        :: param_stack
      INTEGER, INTENT(IN)                      :: stack_size
      INTEGER, INTENT(IN)                      :: datatype
      TYPE(acc_devmem_type), INTENT(IN)        :: a_data, b_data
      TYPE(acc_devmem_type), &
         INTENT(INOUT)                          :: c_data
      INTEGER, INTENT(IN)                      :: m_max, n_max, k_max
      LOGICAL, INTENT(IN)                      :: def_mnk
      TYPE(acc_stream_type), INTENT(IN)        :: stream
      LOGICAL, INTENT(INOUT)                   :: success

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(param_stack))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_size))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(datatype))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(a_data))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(b_data))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(c_data))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(m_max))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_max))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(k_max))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(def_mnk))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(success))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("dbcsr/mm/dbcsr_acc_operations.F",114,"__DBCSR_ACC not compiled in.")
   END SUBROUTINE dbcsr_acc_do_mm_stack

! **************************************************************************************************
!> \brief Launch an accelerated transpose kernel
!> \param trs_stack ...
!> \param offset ...
!> \param nblks ...
!> \param datatype ...
!> \param buffer ...
!> \param m ...
!> \param n ...
!> \param stream ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_acc_transpose(trs_stack, offset, nblks, datatype, buffer, m, n, stream)
      TYPE(acc_devmem_type), INTENT(IN)        :: trs_stack
      INTEGER, INTENT(IN)                      :: offset
      INTEGER, INTENT(IN)                      :: nblks
      INTEGER, INTENT(IN)                      :: datatype
      TYPE(acc_devmem_type), INTENT(IN)        :: buffer
      INTEGER, INTENT(IN)                      :: m, n
      TYPE(acc_stream_type), INTENT(IN)        :: stream

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(trs_stack))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(offset))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nblks))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(datatype))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(buffer))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(m))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("dbcsr/mm/dbcsr_acc_operations.F",178,"__DBCSR_ACC not compiled in.")
   END SUBROUTINE dbcsr_acc_transpose

END MODULE dbcsr_acc_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR input/output
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_io
   USE array_types,                     ONLY: array_data
   USE dbcsr_data_methods,              ONLY: dbcsr_data_clear_pointer,&
                                              dbcsr_data_get_size_referenced,&
                                              dbcsr_data_init,&
                                              dbcsr_data_new,&
                                              dbcsr_get_data
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_mp
   USE dbcsr_methods,                   ONLY: &
        dbcsr_data_area, dbcsr_distribution, dbcsr_get_data_size, dbcsr_get_data_type, &
        dbcsr_get_matrix_type, dbcsr_get_num_blocks, dbcsr_name, dbcsr_nblkcols_total, &
        dbcsr_nblkrows_total, dbcsr_valid_index
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_group,&
                                              dbcsr_mp_pgrid
   USE dbcsr_mp_operations,             ONLY: dbcsr_mp_type_from_anytype
   USE dbcsr_transformations,           ONLY: dbcsr_datablock_redistribute
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_distribution_obj,&
                                              dbcsr_mp_obj,&
                                              dbcsr_obj,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              real_4,&
                                              real_4_size,&
                                              real_8,&
                                              real_8_size,&
                                              sp
   USE machine,                         ONLY: default_output_unit
   USE message_passing,                 ONLY: &
        file_amode_create, file_amode_rdonly, file_amode_wronly, file_offset, mp_allgather, &
        mp_environ, mp_file_close, mp_file_get_size, mp_file_open, mp_file_read_at_all, &
        mp_file_write_at, mp_file_write_at_all, mp_sum, mp_type_descriptor_type, mp_type_size, &
        mpi_character_size, mpi_integer_size
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_io'

   ! Main
   PUBLIC :: dbcsr_print
   PUBLIC :: dbcsr_print_block_sum
   ! Low-level printing
! PUBLIC :: dbcsr_printmat, dbcsr_print2dmat
   ! Utility printing
   PUBLIC :: dbcsr_binary_write
   PUBLIC :: dbcsr_binary_read

   LOGICAL, PARAMETER :: bcsr_debug = .TRUE.
   LOGICAL, PARAMETER :: bcsr_info = .FALSE.
   LOGICAL, PARAMETER :: bcsr_verbose = .FALSE.

   INTERFACE dbcsr_printmat
      MODULE PROCEDURE printmat_s, printmat_d, printmat_c, printmat_z
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Prints a BCSR matrix (block-style, not full)
!> \param[in] matrix          matrix
!> \param[in] nodata          (optional) don't print actual data
!> \param matlab_format ...
!> \param variable_name ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE dbcsr_print(matrix, nodata, matlab_format, variable_name, unit_nr)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: nodata, matlab_format
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: variable_name
      INTEGER, INTENT(IN), OPTIONAL                      :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_print', routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      INTEGER                                            :: ablk_p, bc, blk, blk_p, br, ebr, fblk, &
                                                            handle, ibr, iunit, lblk, m, mn, n, &
                                                            sblk
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_offset, col_blk_size, &
                                                            local_cols, local_rows, &
                                                            row_blk_offset, row_blk_size
      LOGICAL                                            :: my_matlab_format, tr, yesprint
      REAL(KIND=dp)                                      :: blk_cs
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. dbcsr_valid_index(matrix)) &
         CALL cp__w("dbcsr/ops/dbcsr_io.F",115,"Can not print invalid matrix.")

      iunit = default_output_unit
      IF (PRESENT(unit_nr)) iunit = unit_nr

      my_matlab_format = .FALSE.
      IF (PRESENT(matlab_format)) my_matlab_format = matlab_format
      yesprint = .TRUE.
      IF (PRESENT(nodata)) yesprint = .NOT. nodata
      WRITE (iunit, *) routineP//' Contents of matrix named ', matrix%m%name
      WRITE (iunit, *) routineP//' Flags ', matrix%m%symmetry, &
         matrix%m%negate_real, matrix%m%negate_imaginary, "type", &
         dbcsr_get_data_type(matrix), "serial", matrix%m%serial_number
      WRITE (iunit, '(1X,A,3(1X,I9,1X,A))') routineP, matrix%m%nblks, "blocks", &
         matrix%m%nze, "nzes,", dbcsr_get_data_size(matrix), "data els", &
         dbcsr_data_get_size_referenced(matrix%m%data_area), "used"
      WRITE (iunit, '(1X,A,I5,A,I5)') routineP//" Full size", &
         matrix%m%nfullrows_total, "x", matrix%m%nfullcols_total
      WRITE (iunit, '(1X,A,I5,A,I5)') routineP//" Blocked size", &
         matrix%m%nblkrows_total, "x", matrix%m%nblkcols_total
      SELECT CASE (matrix%m%data_type)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_get_data(matrix%m%data_area, r_dp)
      CASE (dbcsr_type_real_4)
         CALL dbcsr_get_data(matrix%m%data_area, r_sp)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_get_data(matrix%m%data_area, c_dp)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_get_data(matrix%m%data_area, c_sp)
      END SELECT
      row_blk_size => array_data(matrix%m%row_blk_size)
      col_blk_size => array_data(matrix%m%col_blk_size)
      row_blk_offset => array_data(matrix%m%row_blk_offset)
      col_blk_offset => array_data(matrix%m%col_blk_offset)

      IF (matrix%m%nblks .GT. 0) THEN
         IF (matrix%m%list_indexing) THEN
            IF (SIZE(matrix%m%coo_l) .NE. 3*matrix%m%nblks) &
               CALL cp__b("dbcsr/ops/dbcsr_io.F",153,"Wrong list")
            IF (matrix%m%local_indexing) THEN
               local_rows => array_data(matrix%m%local_rows)
               local_cols => array_data(matrix%m%local_cols)
            ELSE
               NULLIFY (local_rows)
               NULLIFY (local_cols)
            ENDIF
            ebr = 1
            sblk = 3
         ELSE
            ebr = matrix%m%nblkrows_total
            sblk = 1
         ENDIF
         DO ibr = 1, ebr
            IF (matrix%m%list_indexing) THEN
               fblk = 1
               lblk = SIZE(matrix%m%coo_l)
            ELSE
               br = ibr
               fblk = matrix%m%row_p(br)+1
               lblk = matrix%m%row_p(br+1)
               m = row_blk_size(br)
            ENDIF
            DO blk = fblk, lblk, sblk
               IF (matrix%m%list_indexing) THEN
                  br = matrix%m%coo_l(blk)
                  bc = matrix%m%coo_l(blk+1)
                  IF (matrix%m%local_indexing) THEN
                     br = local_rows(br)
                     bc = local_cols(bc)
                  ENDIF
                  m = row_blk_size(br)
                  ablk_p = matrix%m%coo_l(blk+2)
               ELSE
                  bc = matrix%m%col_i(blk)
                  ablk_p = matrix%m%blk_p(blk)
               ENDIF
               n = col_blk_size(bc)
               mn = m*n
               blk_p = ABS(ablk_p)
               tr = ablk_p .LT. 0
               block_exists: IF (blk_p .NE. 0) THEN
                  IF (mn .GT. 0) THEN
                     SELECT CASE (matrix%m%data_type)
                     CASE (dbcsr_type_real_8)
                        blk_cs = REAL(DOT_PRODUCT(r_dp(blk_p:blk_p+mn-1), &
                                                  r_dp(blk_p:blk_p+mn-1)), KIND=dp)
                        !CALL &
                        !     dbcsr_printmat(r_dp(blk_p:blk_p+mn-1),m,n, tr=tr)
                     CASE (dbcsr_type_real_4)
                        blk_cs = REAL(DOT_PRODUCT(r_sp(blk_p:blk_p+mn-1), &
                                                  r_sp(blk_p:blk_p+mn-1)), KIND=dp)
                        !CALL &
                        !     dbcsr_printmat(r_sp(blk_p:blk_p+mn-1),m,n, tr=tr)
                     CASE (dbcsr_type_complex_8)
                        blk_cs = REAL(DOT_PRODUCT(c_dp(blk_p:blk_p+mn-1), &
                                                  c_dp(blk_p:blk_p+mn-1)), KIND=dp)
                        !CALL &
                        !     dbcsr_printmat(c_dp(blk_p:blk_p+mn-1),m,n, tr=tr)
                     CASE (dbcsr_type_complex_4)
                        blk_cs = REAL(DOT_PRODUCT(c_sp(blk_p:blk_p+mn-1), &
                                                  c_sp(blk_p:blk_p+mn-1)), KIND=dp)
                        !CALL &
                        !     dbcsr_printmat(c_sp(blk_p:blk_p+mn-1),m,n, tr=tr)
                     END SELECT
                  ELSE
                     blk_cs = 0.0_dp
                  ENDIF
                  !WRITE(iunit,*)routineP//' chksum for (',br,',',bc,') at',&
                  !     blk_p,'l',mn,'= ', blk_cs,'size',m,n
                  IF (.NOT. my_matlab_format) WRITE (iunit, '(A,I6,",",I6,A,I7,A,I6,I6,"=",I7,A,E12.3)') &
                     !" Checksum for (",br,bc,") at ",blk_p," size ",m,n,mn,&
                     " Checksum for (", br, bc, ") at ", ablk_p, " size ", m, n, mn, &
                     " checksum=", blk_cs
                  IF (yesprint .AND. blk_p .NE. 0) THEN
                     IF (mn .GT. 0) THEN
                        SELECT CASE (matrix%m%data_type)
                        CASE (dbcsr_type_real_8)
                           !WRITE(iunit,'(10(1X,F7.2))')r_dp(blk_p:blk_p+mn-1)
                           IF (my_matlab_format) THEN
                              CALL dbcsr_printmat_matlab_d(r_dp(blk_p:blk_p+mn-1), m, n, &
                                                           row_blk_offset(br), col_blk_offset(bc), iunit, tr=tr, &
                                                           variable_name=variable_name)
                           ELSE
                              CALL dbcsr_printmat(r_dp(blk_p:blk_p+mn-1), m, n, iunit=iunit, tr=tr)
                           ENDIF
                        CASE (dbcsr_type_real_4)
                           IF (my_matlab_format) THEN
                              CALL dbcsr_printmat_matlab_s(r_sp(blk_p:blk_p+mn-1), m, n, &
                                                           row_blk_offset(br), col_blk_offset(bc), iunit, tr=tr, &
                                                           variable_name=variable_name)
                           ELSE
                              CALL dbcsr_printmat(r_sp(blk_p:blk_p+mn-1), m, n, iunit=iunit, tr=tr)
                           ENDIF
                        CASE (dbcsr_type_complex_8)
                           IF (my_matlab_format) THEN
                              CALL dbcsr_printmat_matlab_z(c_dp(blk_p:blk_p+mn-1), m, n, &
                                                           row_blk_offset(br), col_blk_offset(bc), iunit, tr=tr, &
                                                           variable_name=variable_name)
                           ELSE
                              CALL dbcsr_printmat(c_dp(blk_p:blk_p+mn-1), m, n, iunit=iunit, tr=tr)
                           ENDIF
                        CASE (dbcsr_type_complex_4)
                           IF (my_matlab_format) THEN
                              CALL dbcsr_printmat_matlab_c(c_sp(blk_p:blk_p+mn-1), m, n, &
                                                           row_blk_offset(br), col_blk_offset(bc), iunit, tr=tr, &
                                                           variable_name=variable_name)
                           ELSE
                              CALL dbcsr_printmat(c_sp(blk_p:blk_p+mn-1), m, n, iunit=iunit, tr=tr)
                           ENDIF
                        END SELECT
                     ENDIF
                  ENDIF
               ENDIF block_exists
            ENDDO
         ENDDO
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_print

! **************************************************************************************************
!> \brief Prints the elements of a matrix.
!> \param matrix ...
!> \param rows the logical (possibly detransposed) matrix
!>                            size, not the stored size
!> \param cols the logical (possibly detransposed) matrix
!>                            size, not the stored size
!> \param r_offset ...
!> \param c_offset ...
!> \param iunit ...
!> \param[in] tr    (optional) specifies whether the elements are stored
!>                  transposed
!> \param variable_name ...
! **************************************************************************************************

   SUBROUTINE dbcsr_printmat_matlab_d(matrix, rows, cols, r_offset, c_offset, iunit, tr, variable_name)
      REAL(KIND=real_8), DIMENSION(:), INTENT(IN)        :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, r_offset, c_offset, iunit
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: variable_name

      INTEGER                                            :: c, c_off, m, n, r, r_off
      LOGICAL                                            :: t

!   ---------------------------------------------------------------------------

      m = rows
      n = cols
      r_off = r_offset
      c_off = c_offset
      t = .FALSE.
      IF (PRESENT(tr)) THEN
         IF (tr) THEN
            t = .TRUE.
            m = cols
            n = rows
            r_off = c_offset
            c_off = r_offset
         ENDIF
      ENDIF

      DO c = 1, cols
      DO r = 1, rows
         IF (.NOT. t) THEN
            IF (PRESENT(variable_name)) THEN
               WRITE (iunit, '(A,I4,A,I4,A,E23.16,A)') &
                  variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix(r+(c-1)*rows), ';'
            ELSE
               WRITE (iunit, '(A,I4,A,I4,A,E23.16,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix(r+(c-1)*rows), ';'
            ENDIF
         ELSE
            IF (PRESENT(variable_name)) THEN
               WRITE (iunit, '(A,I4,A,I4,A,E23.16,A)') &
                  variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix((r-1)*cols+c), ';'
            ELSE
               WRITE (iunit, '(A,I4,A,I4,A,E23.16,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix((r-1)*cols+c), ';'
            ENDIF
         ENDIF
      ENDDO
      ENDDO
   END SUBROUTINE dbcsr_printmat_matlab_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param r_offset ...
!> \param c_offset ...
!> \param iunit ...
!> \param tr ...
!> \param variable_name ...
! **************************************************************************************************
   SUBROUTINE dbcsr_printmat_matlab_s(matrix, rows, cols, r_offset, c_offset, iunit, tr, variable_name)
      REAL(KIND=real_4), DIMENSION(:), INTENT(IN)        :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, r_offset, c_offset, iunit
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: variable_name

      INTEGER                                            :: c, c_off, m, n, r, r_off
      LOGICAL                                            :: t

!   ---------------------------------------------------------------------------

      m = rows
      n = cols
      r_off = r_offset
      c_off = c_offset
      t = .FALSE.
      IF (PRESENT(tr)) THEN
         IF (tr) THEN
            t = .TRUE.
            m = cols
            n = rows
            r_off = c_offset
            c_off = r_offset
         ENDIF
      ENDIF

      DO c = 1, cols
      DO r = 1, rows
         IF (.NOT. t) THEN
            IF (PRESENT(variable_name)) THEN
               WRITE (iunit, '(A,I4,A,I4,A,E15.7,A)') &
                  variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix(r+(c-1)*rows), ';'
            ELSE
               WRITE (iunit, '(A,I4,A,I4,A,E15.7,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix(r+(c-1)*rows), ';'
            ENDIF
         ELSE
            IF (PRESENT(variable_name)) THEN
               WRITE (iunit, '(A,I4,A,I4,A,E15.7,A)') &
                  variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix((r-1)*cols+c), ';'
            ELSE
               WRITE (iunit, '(A,I4,A,I4,A,E15.7,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', matrix((r-1)*cols+c), ';'
            ENDIF
         ENDIF
      ENDDO
      ENDDO
   END SUBROUTINE dbcsr_printmat_matlab_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param r_offset ...
!> \param c_offset ...
!> \param iunit ...
!> \param tr ...
!> \param variable_name ...
! **************************************************************************************************
   SUBROUTINE dbcsr_printmat_matlab_z(matrix, rows, cols, r_offset, c_offset, iunit, tr, variable_name)
      COMPLEX(KIND=real_8), DIMENSION(:), INTENT(IN)     :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, r_offset, c_offset, iunit
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: variable_name

      INTEGER                                            :: c, c_off, m, n, r, r_off
      LOGICAL                                            :: t

!   ---------------------------------------------------------------------------

      m = rows
      n = cols
      r_off = r_offset
      c_off = c_offset
      t = .FALSE.
      IF (PRESENT(tr)) THEN
         IF (tr) THEN
            t = .TRUE.
            m = cols
            n = rows
            r_off = c_offset
            c_off = r_offset
         ENDIF
      ENDIF

      DO c = 1, cols
         DO r = 1, rows
            IF (.NOT. t) THEN
               IF (PRESENT(variable_name)) THEN
                  WRITE (iunit, '(A,I3,A,I3,A,E23.16,A,E23.16,A)') variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix(r+(c-1)*rows)), '+', AIMAG(matrix(r+(c-1)*rows)), 'i;'
               ELSE
                  WRITE (iunit, '(A,I3,A,I3,A,E23.16,A,E23.16,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix(r+(c-1)*rows)), '+', AIMAG(matrix(r+(c-1)*rows)), 'i;'
               ENDIF
            ELSE
               IF (PRESENT(variable_name)) THEN
                  WRITE (iunit, '(A,I3,A,I3,A,E23.16,A,E23.16,A)') variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix((r-1)*cols+c)), '+', AIMAG(matrix((r-1)*cols+c)), 'i;'
               ELSE
                  WRITE (iunit, '(A,I3,A,I3,A,E23.16,A,E23.16,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix((r-1)*cols+c)), '+', AIMAG(matrix((r-1)*cols+c)), 'i;'
               ENDIF
            ENDIF
         ENDDO
      ENDDO
   END SUBROUTINE dbcsr_printmat_matlab_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param r_offset ...
!> \param c_offset ...
!> \param iunit ...
!> \param tr ...
!> \param variable_name ...
! **************************************************************************************************
   SUBROUTINE dbcsr_printmat_matlab_c(matrix, rows, cols, r_offset, c_offset, iunit, tr, variable_name)
      COMPLEX(KIND=real_4), DIMENSION(:), INTENT(IN)     :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, r_offset, c_offset, iunit
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: variable_name

      INTEGER                                            :: c, c_off, m, n, r, r_off
      LOGICAL                                            :: t

!   ---------------------------------------------------------------------------

      m = rows
      n = cols
      r_off = r_offset
      c_off = c_offset
      t = .FALSE.
      IF (PRESENT(tr)) THEN
         IF (tr) THEN
            t = .TRUE.
            m = cols
            n = rows
            r_off = c_offset
            c_off = r_offset
         ENDIF
      ENDIF

      DO c = 1, cols
         DO r = 1, rows
            IF (.NOT. t) THEN
               IF (PRESENT(variable_name)) THEN
                  WRITE (iunit, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix(r+(c-1)*rows)), '+', AIMAG(matrix(r+(c-1)*rows)), 'i;'
               ELSE
                  WRITE (iunit, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix(r+(c-1)*rows)), '+', AIMAG(matrix(r+(c-1)*rows)), 'i;'
               ENDIF
            ELSE
               IF (PRESENT(variable_name)) THEN
                  WRITE (iunit, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') variable_name//'(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix((r-1)*cols+c)), '+', AIMAG(matrix((r-1)*cols+c)), 'i;'
               ELSE
                  WRITE (iunit, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'a(', r+r_offset-1, ',', c+c_offset-1, ')=', &
                     REAL(matrix((r-1)*cols+c)), '+', AIMAG(matrix((r-1)*cols+c)), 'i;'
               ENDIF
            ENDIF
         ENDDO
      ENDDO
   END SUBROUTINE dbcsr_printmat_matlab_c

! **************************************************************************************************
!> \brief Prints the elements of a matrix.
!> \param matrix ...
!> \param rows the logical (possibly detransposed) matrix
!>                            size, not the stored size
!> \param cols the logical (possibly detransposed) matrix
!>                            size, not the stored size
!> \param iunit ...
!> \param title ...
!> \param[in] tr    (optional) specifies whether the elements are stored
!>                  transposed
! **************************************************************************************************
   SUBROUTINE printmat_s(matrix, rows, cols, iunit, title, tr)
      REAL(KIND=real_4), DIMENSION(:), INTENT(IN)        :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, iunit
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr

      CHARACTER(30)                                      :: f
      INTEGER                                            :: m, n, r
      LOGICAL                                            :: t
      REAL(KIND=dp)                                      :: bit_bucket

!   ---------------------------------------------------------------------------

      m = rows
      n = cols
      t = .FALSE.
      IF (PRESENT(title)) WRITE (iunit, *) title
      IF (PRESENT(tr)) THEN
         IF (tr) THEN
            t = .TRUE.
            m = cols
            n = rows
         ENDIF
      ENDIF
      DO r = LBOUND(matrix, 1), UBOUND(matrix, 1)
         bit_bucket = matrix(r)
      ENDDO
      bit_bucket = 0.0_dp
      DO r = LBOUND(matrix, 1), UBOUND(matrix, 1)
         bit_bucket = bit_bucket+matrix(r)
      ENDDO
      IF (m .GT. 10000) m = 0
      IF (n .GT. 10000) n = 0
      IF (m*n .LT. 1 .OR. m*n .GT. SIZE(matrix)) RETURN
      WRITE (f, FMT="('(',I4,'(F9.4))')") cols
      DO r = 1, rows
         IF (.NOT. t) THEN
            WRITE (iunit, FMT=f) matrix(r:r+(cols-1)*rows:rows)
         ELSE
            WRITE (iunit, FMT=f) matrix((r-1)*cols+1:r*cols)
         ENDIF
      ENDDO
   END SUBROUTINE printmat_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param iunit ...
!> \param title ...
!> \param tr ...
! **************************************************************************************************
   SUBROUTINE printmat_d(matrix, rows, cols, iunit, title, tr)
      REAL(KIND=real_8), DIMENSION(:), INTENT(IN)        :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, iunit
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr

      IF (PRESENT(title)) THEN
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title, tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title)
         ENDIF
      ELSE
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, tr=tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit)
         ENDIF
      ENDIF
   END SUBROUTINE printmat_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param iunit ...
!> \param title ...
!> \param tr ...
! **************************************************************************************************
   SUBROUTINE printmat_c(matrix, rows, cols, iunit, title, tr)
      COMPLEX(KIND=real_4), DIMENSION(:), INTENT(IN)     :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, iunit
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr

      IF (PRESENT(title)) THEN
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title, tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title)
         ENDIF
      ELSE
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, tr=tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit)
         ENDIF
      ENDIF
   END SUBROUTINE printmat_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param iunit ...
!> \param title ...
!> \param tr ...
! **************************************************************************************************
   SUBROUTINE printmat_z(matrix, rows, cols, iunit, title, tr)
      COMPLEX(KIND=real_8), DIMENSION(:), INTENT(IN)     :: matrix
      INTEGER, INTENT(IN)                                :: rows, cols, iunit
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      LOGICAL, INTENT(IN), OPTIONAL                      :: tr

      IF (PRESENT(title)) THEN
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title, tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, title)
         ENDIF
      ELSE
         IF (PRESENT(tr)) THEN
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit, tr=tr)
         ELSE
            CALL printmat_s(REAL(matrix, KIND=sp), rows, cols, iunit)
         ENDIF
      ENDIF
   END SUBROUTINE printmat_z

! **************************************************************************************************
!> \brief Writes a DBCSR matrix in a file
!>   file's header consists of 3 sub-headers:
!>   sub-header1 contains:
!>     1 string: (of length version_len) the current version of this routine,
!>     1 string: (of length default_string_length) matrix_name,
!>     1 character: matrix_type,
!>     4 integers: numnodes, data_type, nblkrows_total, nblkcols_total,
!>     2 vectors:  row_blk_size (length = nblkrows_total),
!>                 col_blk_size (length = nblkcols_total),
!>   sub-header2 contains:
!>     2 integers: nblks, data_area_size,
!>   sub-header3 contains:
!>     3 vectors:  row_p (length = nblkrows_total+1),
!>                 col_i (length = nblks),
!>                 blk_p (length = nblks);
!>   and the file's body contains the block data
!>
!> \param[in] matrix    DBCSR matrix
!> \param[in] filepath  path to the file
!>
!> \par History
!>      11.2012 created [Hossein Bani-Hashemian]
!> \author Hossein Bani-Hashemian
!> \version DBCSRv_1.0
! **************************************************************************************************
   SUBROUTINE dbcsr_binary_write(matrix, filepath)

      IMPLICIT NONE

      TYPE(dbcsr_obj), INTENT(IN)           :: matrix
      CHARACTER(len=*), INTENT(IN)         :: filepath

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_binary_write', &
                                     routineP = moduleN//':'//routineN

      INTEGER                               :: nblkrows_total, nblkcols_total, &
                                               nblks, size_of_pgrid, &
                                               thefile, i, sendbuf, data_area_size, &
                                               data_type, type_size, &
                                               mp_group, mynode, numnodes, &
                                               ginfo_size, linfo_size, handle
      INTEGER, DIMENSION(:), POINTER        :: row_p, col_i, blk_p, &
                                               row_blk_size, col_blk_size
      INTEGER, DIMENSION(:, :), POINTER      :: pgrid
      TYPE(mp_type_descriptor_type)         :: mp_type
      TYPE(dbcsr_mp_obj)                    :: mp_env
      TYPE(dbcsr_distribution_obj)          :: distribution
      TYPE(dbcsr_data_obj)                  :: data_area
      COMPLEX(sp), DIMENSION(:), POINTER      :: c_sp
      COMPLEX(dp), DIMENSION(:), POINTER      :: c_dp
      REAL(sp), DIMENSION(:), POINTER         :: r_sp
      REAL(dp), DIMENSION(:), POINTER         :: r_dp
      CHARACTER                             :: matrix_type
      CHARACTER(LEN=80)                     :: matrix_name_v_1_0
      CHARACTER(LEN=default_string_length)  :: matrix_name
      INTEGER, PARAMETER                    :: version_len = 10
      CHARACTER(LEN=version_len), PARAMETER :: version = "DBCSRv_1.0"
      INTEGER, ALLOCATABLE, DIMENSION(:) :: linfo_sizes, da_sizes
      INTEGER(kind=file_offset), ALLOCATABLE, DIMENSION(:) :: bdata_disps, bdata_offsets, &
                                                              subh2_disps, subh2_offsets, &
                                                              subh3_disps, subh3_offsets
      INTEGER(kind=file_offset), PARAMETER                 :: BOF = 0
      INTEGER, PARAMETER          :: char_count = version_len+default_string_length+1 !version, matrix_name, matrix_type

      CALL timeset(routineN, handle)

      IF (default_string_length /= 80) &
         CALL cp_warn(cp__l("dbcsr/ops/dbcsr_io.F",728), "Changing the default string length affects "// &
                      "the format of the writen matrix. Version needs to be adjusted")

      nblkrows_total = dbcsr_nblkrows_total(matrix)
      nblkcols_total = dbcsr_nblkcols_total(matrix)
      distribution = dbcsr_distribution(matrix)
      matrix_name = dbcsr_name(matrix)
      data_area = dbcsr_data_area(matrix%m)
      matrix_type = dbcsr_get_matrix_type(matrix)
      data_type = dbcsr_get_data_type(matrix)
      mp_env = dbcsr_distribution_mp(distribution)
      mp_group = dbcsr_mp_group(mp_env)
      nblks = dbcsr_get_num_blocks(matrix)
      row_p => matrix%m%row_p
      col_i => matrix%m%col_i
      blk_p => matrix%m%blk_p
      row_blk_size => array_data(matrix%m%row_blk_size)
      col_blk_size => array_data(matrix%m%col_blk_size)
      pgrid => dbcsr_mp_pgrid(mp_env)
      size_of_pgrid = SIZE(pgrid)

      CALL mp_environ(numnodes, mynode, mp_group)

      ALLOCATE (linfo_sizes(numnodes), da_sizes(numnodes), &
                subh2_disps(numnodes), subh2_offsets(numnodes), &
                subh3_disps(numnodes), subh3_offsets(numnodes), &
                bdata_disps(numnodes), bdata_offsets(numnodes))
      subh2_disps(:) = (/((i-1)*2, i=1, numnodes)/)
      subh3_disps = BOF
      bdata_disps = BOF
      linfo_sizes = BOF
      subh2_offsets = BOF
      subh3_offsets = BOF
      bdata_offsets = BOF
      da_sizes = BOF

      ginfo_size = char_count+4+nblkrows_total+nblkcols_total
      linfo_size = 1+nblkrows_total+2*nblks

      sendbuf = linfo_size
      CALL mp_allgather(sendbuf, linfo_sizes, mp_group)
      CALL cumsum_l(INT(linfo_sizes, kind=file_offset), subh3_disps)
      subh3_disps(:) = CSHIFT(subh3_disps, shift=-1)+ginfo_size+2*numnodes
      subh3_disps(1) = ginfo_size+2*numnodes

      data_area_size = dbcsr_data_get_size_referenced(matrix%m%data_area)
      sendbuf = data_area_size
      CALL mp_allgather(sendbuf, da_sizes, mp_group)
      CALL cumsum_l(INT(da_sizes, kind=file_offset), bdata_disps)
      bdata_disps(:) = CSHIFT(bdata_disps, shift=-1)+SUM(INT(linfo_sizes, KIND=file_offset))+ &
                       ginfo_size+numnodes*2
      bdata_disps(1) = SUM(INT(linfo_sizes, KIND=file_offset))+ginfo_size+numnodes*2

      CALL mp_file_open(mp_group, thefile, filepath, file_amode_create+file_amode_wronly)

      IF (mynode .EQ. 0) THEN
         CALL mp_file_write_at(thefile, BOF, version)
         matrix_name_v_1_0 = matrix_name
         CALL mp_file_write_at(thefile, BOF+version_len*mpi_character_size, matrix_name_v_1_0)
         CALL mp_file_write_at(thefile, BOF+(version_len+default_string_length)*mpi_character_size, matrix_type)
         CALL mp_file_write_at(thefile, BOF+char_count*mpi_character_size, &
                               (/size_of_pgrid, data_type, &
                                 nblkrows_total, nblkcols_total, &
                                 row_blk_size, col_blk_size/))
      END IF
! write sub-header2
      subh2_disps(:) = subh2_disps(:)+ginfo_size
      subh2_offsets(:) = BOF+(subh2_disps-char_count)*mpi_integer_size+ &
                         char_count*mpi_character_size
      CALL mp_file_write_at_all(thefile, subh2_offsets(mynode+1), (/nblks, data_area_size/))
! write sub-header3
      subh3_offsets(:) = BOF+(subh3_disps-char_count)*mpi_integer_size+ &
                         char_count*mpi_character_size
      CALL mp_file_write_at_all(thefile, subh3_offsets(mynode+1), (/row_p, col_i, blk_p/))
! write block data
      mp_type = dbcsr_mp_type_from_anytype(data_area)
      CALL mp_type_size(mp_type, type_size)
      bdata_offsets(:) = BOF+(/((bdata_disps(i)-bdata_disps(1))*type_size, i=1, numnodes)/)+ &
                         (bdata_disps(1)-char_count)*mpi_integer_size+ &
                         char_count*mpi_character_size
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         r_sp => data_area%d%r_sp
         CALL mp_file_write_at_all(thefile, bdata_offsets(mynode+1), r_sp, msglen=data_area_size)
      CASE (dbcsr_type_real_8)
         r_dp => data_area%d%r_dp
         CALL mp_file_write_at_all(thefile, bdata_offsets(mynode+1), r_dp, msglen=data_area_size)
      CASE (dbcsr_type_complex_4)
         c_sp => data_area%d%c_sp
         CALL mp_file_write_at_all(thefile, bdata_offsets(mynode+1), c_sp, msglen=data_area_size)
      CASE (dbcsr_type_complex_8)
         c_dp => data_area%d%c_dp
         CALL mp_file_write_at_all(thefile, bdata_offsets(mynode+1), c_dp, msglen=data_area_size)
      END SELECT
      CALL mp_file_close(thefile)

      DEALLOCATE (linfo_sizes, da_sizes)
      DEALLOCATE (subh2_disps, subh2_offsets, subh3_disps, subh3_offsets)
      DEALLOCATE (bdata_disps, bdata_offsets)

      CALL timestop(handle)

   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param arr ...
!> \param cumsum ...
! **************************************************************************************************
      SUBROUTINE cumsum_l(arr, cumsum)
      INTEGER(kind=file_offset), DIMENSION(:), &
         INTENT(IN)                                      :: arr
      INTEGER(kind=file_offset), DIMENSION(SIZE(arr)), &
         INTENT(OUT)                                     :: cumsum

      INTEGER                                            :: i

         cumsum(1) = arr(1)
         DO i = 2, SIZE(arr)
            cumsum(i) = cumsum(i-1)+arr(i)
         END DO
      END SUBROUTINE cumsum_l
   END SUBROUTINE dbcsr_binary_write

! **************************************************************************************************
!> \brief Reads a DBCSR matrix from a file
!> \param[in] filepath             path to the file
!> \param[in] distribution         row and column distribution
!> \param groupid ...
!> \param[out] matrix_new          DBCSR matrix
!>
!> \param[in](optional) groupid    message passing environment identifier
!> \par History
!>      11.2012 created [Hossein Bani-Hashemian]
!> \author Hossein Bani-Hashemian
!> \version DBCSRv_1.0
! **************************************************************************************************
   SUBROUTINE dbcsr_binary_read(filepath, distribution, groupid, matrix_new)

      IMPLICIT NONE

      CHARACTER(len=*), INTENT(IN)                :: filepath
      TYPE(dbcsr_distribution_obj), INTENT(IN)     :: distribution
      INTEGER, INTENT(IN), OPTIONAL                :: groupid
      TYPE(dbcsr_obj), INTENT(INOUT)               :: matrix_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_binary_read', &
                                     routineP = moduleN//':'//routineN

      INTEGER                               :: nblkrows_total, nblkcols_total, &
                                               nblks, darea_size, data_type, type_size, &
                                               globalinfo_size, &
                                               size_of_pgrid, &
                                               thefile, i, j, &
                                               nblocks, &
                                               share_size, order, cur_blks, &
                                               job_count, start_index, end_index, &
                                               localinfo_length, blockdata_length, &
                                               group_id, worker_id, group_list_size, handle, linfo_length
      CHARACTER                             :: matrix_type
      CHARACTER(LEN=default_string_length)  :: matrix_name
      INTEGER, PARAMETER                    :: version_len = 10
      CHARACTER(LEN=version_len)            :: version
      CHARACTER(LEN=80)                     :: matrix_name_v_1_0
      CHARACTER(LEN=version_len), PARAMETER :: version_v_1_0 = "DBCSRv_1.0"

      INTEGER, DIMENSION(:), POINTER        :: row_p, col_i, blk_p, &
                                               proc_nblks, proc_darea_sizes
      INTEGER, DIMENSION(4)                 :: values
      INTEGER, ALLOCATABLE, DIMENSION(:)    :: linfo_lens, bdata_lens
      INTEGER, ALLOCATABLE, DIMENSION(:), TARGET     :: ginfo_vec, linfo_vec, &
                                                        rowp, coli, blkp
      INTEGER, ALLOCATABLE, DIMENSION(:, :), TARGET   :: val_data
      INTEGER, DIMENSION(:), POINTER                 :: row_blk_size, col_blk_size
      TYPE(dbcsr_mp_obj)                             :: mp_env
      TYPE(dbcsr_data_obj)                           :: dblk
      REAL(sp)                                       :: rsp_dummy(1)
      REAL(dp)                                       :: rdp_dummy(1)
      COMPLEX(sp)                                    :: csp_dummy(1)
      COMPLEX(dp)                                    :: cdp_dummy(1)
      REAL(sp), ALLOCATABLE, DIMENSION(:), TARGET     :: rsp
      REAL(dp), ALLOCATABLE, DIMENSION(:), TARGET     :: rdp
      COMPLEX(sp), ALLOCATABLE, DIMENSION(:), TARGET  :: csp
      COMPLEX(dp), ALLOCATABLE, DIMENSION(:), TARGET  :: cdp
      INTEGER(kind=file_offset), ALLOCATABLE, DIMENSION(:)   :: subh2_offsets, &
                                                                subh3_disps, subh3_offsets, &
                                                                bdata_disps, bdata_offsets
      INTEGER(kind=file_offset), PARAMETER    :: BOF = 0
      INTEGER(kind=file_offset)               :: offset, subh2_start, subh3_start, bdata_start, file_size, &
                                                 localinfo_offset, blockdata_offset, sum_nblks, subh3_length, data_area_size
      INTEGER, PARAMETER                      :: char_count = 1+version_len+default_string_length

      CALL timeset(routineN, handle)

      IF (PRESENT(groupid)) THEN
         group_id = groupid
      ELSE
         mp_env = dbcsr_distribution_mp(distribution)
         group_id = dbcsr_mp_group(mp_env)
      END IF

      CALL mp_environ(group_list_size, worker_id, group_id)
      CALL mp_file_open(group_id, thefile, filepath, file_amode_rdonly)

! read version, matrix name and matrix type
      CALL mp_file_read_at_all(thefile, BOF, version)

      IF (version /= version_v_1_0) &
         CALL cp__w("dbcsr/ops/dbcsr_io.F",935,"Trying to read an unknown version of the matrix data file. Good luck!")

      CALL mp_file_read_at_all(thefile, BOF+version_len*mpi_character_size, matrix_name_v_1_0)
      matrix_name = matrix_name_v_1_0

      CALL mp_file_read_at_all(thefile, BOF+(version_len+default_string_length)*mpi_character_size, matrix_type)
! read 4 integer values form sub-header1
      CALL mp_file_read_at_all(thefile, BOF+char_count*mpi_character_size, values)
      size_of_pgrid = values(1)
      data_type = values(2)
      nblkrows_total = values(3)
      nblkcols_total = values(4)
! read 2 vectors, row_blk_size and col_blk_size, from sub-header1
      globalinfo_size = nblkrows_total+nblkcols_total
      ALLOCATE (ginfo_vec(globalinfo_size))
      CALL mp_file_read_at_all(thefile, BOF+char_count*mpi_character_size+4*mpi_integer_size, ginfo_vec)
      row_blk_size => ginfo_vec(1:nblkrows_total)
      col_blk_size => ginfo_vec(nblkrows_total+1:globalinfo_size)

! compute the offsets where sub-header2 and sub-header3 start
      subh2_start = (4+globalinfo_size)*mpi_integer_size+char_count*mpi_character_size
      subh3_start = subh2_start+2*size_of_pgrid*mpi_integer_size

! compute the offsets in sub-header2 and read 2 integers nblocks, data_area_size
      ! number of data chunks from sub-header 2 and 3 to be read by every node rounded up
      ! to the next integer to make it even for all the nodes in the specified mpi group
      share_size = CEILING(REAL(size_of_pgrid)/group_list_size)

      ALLOCATE (subh2_offsets(share_size))
      subh2_offsets = BOF
      DO i = 1, share_size
         offset = subh2_start+mpi_integer_size*2*(worker_id+(i-1)*group_list_size)
         IF (offset .GE. subh3_start) EXIT
         subh2_offsets(i) = offset
      END DO

      ALLOCATE (val_data(3, share_size))
      val_data(:, :) = 0
      DO i = 1, share_size
         CALL mp_file_read_at_all(thefile, subh2_offsets(i), values, msglen=2)
         nblocks = values(1)
         data_area_size = values(2)
         IF (subh2_offsets(i) .EQ. 0) EXIT
         val_data(1, i) = nblocks
         IF (data_area_size >= HUGE(val_data(2, i))) &
            CALL cp__b("dbcsr/ops/dbcsr_io.F",980,"Data area too large, fix code.")
         val_data(2, i) = INT(data_area_size)
         val_data(3, i) = worker_id+(i-1)*group_list_size+1 ! order
         ! order = indices of an array of length size_of_pgrid to be accessed by the current node
      END DO
      nblks = SUM(val_data(1, :))
      darea_size = SUM(val_data(2, :))
      proc_nblks => val_data(1, :) ! to be passed to dbcsr_datablock_redistribute
      proc_darea_sizes => val_data(2, :) ! to be passed to dbcsr_datablock_redistribute

! compute the offsets in sub-header3 and read 3 vectors row_p, col_i, blk_p
      ! actual number of chunks to be read by the current node
      job_count = COUNT(val_data(3, :) .NE. 0)
      CALL mp_file_get_size(thefile, file_size)

      ALLOCATE (linfo_lens(size_of_pgrid))
      ALLOCATE (subh3_disps(size_of_pgrid))
      ALLOCATE (subh3_offsets(size_of_pgrid))
      linfo_lens = 0; subh3_disps = 0
      DO i = 1, size_of_pgrid
         DO j = 1, share_size
            order = val_data(3, j)
            IF (i .EQ. order) linfo_lens(order) = &
               1+nblkrows_total+2*val_data(1, j)
         END DO
      END DO
      CALL mp_sum(linfo_lens, group_id)
      CALL cumsum_l(INT(linfo_lens, kind=file_offset), subh3_disps)
      subh3_disps(:) = CSHIFT(subh3_disps, shift=-1)
      subh3_disps(1) = BOF
      subh3_offsets(:) = subh3_start+subh3_disps*mpi_integer_size

      sum_nblks = INT(nblks, kind=file_offset)
      CALL mp_sum(sum_nblks, group_id)
      subh3_length = size_of_pgrid*INT(1+nblkrows_total, KIND=file_offset)+2*sum_nblks

      linfo_length = nblkrows_total+1+2*MAXVAL(val_data(1, :))

      ALLOCATE (linfo_vec(linfo_length))
      ALLOCATE (rowp((nblkrows_total+1)*job_count))
      ALLOCATE (coli(nblks))
      ALLOCATE (blkp(nblks))
      DO i = 1, share_size
         order = val_data(3, i)
         cur_blks = val_data(1, i)
         IF (order .EQ. 0) THEN
            localinfo_offset = file_size
            localinfo_length = 0
         ELSE
            localinfo_offset = subh3_offsets(order)
            localinfo_length = linfo_lens(order)
         END IF
         CALL mp_file_read_at_all(thefile, localinfo_offset, linfo_vec, msglen=localinfo_length)
         IF (localinfo_length .EQ. 0) EXIT

         rowp((i-1)*(nblkrows_total+1)+1:i*(nblkrows_total+1)) = linfo_vec(1:nblkrows_total+1)
         start_index = SUM(val_data(1, 1:i-1))+1
         end_index = SUM(val_data(1, 1:i))
         coli(start_index:end_index) = &
            linfo_vec(nblkrows_total+2:cur_blks+nblkrows_total+1)
         blkp(start_index:end_index) = &
            linfo_vec(cur_blks+nblkrows_total+2:2*cur_blks+nblkrows_total+1)
      END DO
      row_p => rowp
      col_i => coli
      blk_p => blkp

! compute the offsets and read block data
      ALLOCATE (bdata_lens(size_of_pgrid))
      ALLOCATE (bdata_disps(size_of_pgrid))
      ALLOCATE (bdata_offsets(size_of_pgrid))
      bdata_lens = 0
      DO i = 1, size_of_pgrid
         DO j = 1, share_size
            order = val_data(3, j)
            IF (i .EQ. order) bdata_lens(order) = val_data(2, j)
         END DO
      END DO
      CALL mp_sum(bdata_lens, group_id)
      CALL cumsum_l(INT(bdata_lens, kind=file_offset), bdata_disps)
      bdata_disps(:) = CSHIFT(bdata_disps, shift=-1)
      bdata_disps(1) = BOF

      bdata_start = subh3_start+subh3_length*mpi_integer_size
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         type_size = real_4_size
      CASE (dbcsr_type_real_8)
         type_size = real_8_size
      CASE (dbcsr_type_complex_4)
         type_size = 2*real_4_size
      CASE (dbcsr_type_complex_8)
         type_size = 2*real_8_size
      END SELECT
      bdata_offsets(:) = bdata_start+bdata_disps*type_size

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         ALLOCATE (rsp(darea_size))
         DO i = 1, share_size
            order = val_data(3, i)
            ! use dummy one-sized data array as buffer in place of empty array
            ! when nothing is supposed to be read (order = 0)
            IF (order .EQ. 0) THEN
               blockdata_offset = file_size
               CALL mp_file_read_at_all(thefile, blockdata_offset, rsp_dummy)
            ELSE
               start_index = SUM(val_data(2, 1:i-1))+1
               end_index = SUM(val_data(2, 1:i))
               blockdata_length = bdata_lens(order)
               blockdata_offset = bdata_offsets(order)
               CALL mp_file_read_at_all(thefile, blockdata_offset, rsp(start_index:end_index), &
                                        msglen=blockdata_length)
            END IF
         END DO
      CASE (dbcsr_type_real_8)
         ALLOCATE (rdp(darea_size))
         DO i = 1, share_size
            order = val_data(3, i)
            IF (order .EQ. 0) THEN
               blockdata_offset = file_size
               CALL mp_file_read_at_all(thefile, blockdata_offset, rdp_dummy)
            ELSE
               start_index = SUM(val_data(2, 1:i-1))+1
               end_index = SUM(val_data(2, 1:i))
               blockdata_length = bdata_lens(order)
               blockdata_offset = bdata_offsets(order)
               CALL mp_file_read_at_all(thefile, blockdata_offset, rdp(start_index:end_index), &
                                        msglen=blockdata_length)
            END IF
         END DO
      CASE (dbcsr_type_complex_4)
         ALLOCATE (csp(darea_size))
         DO i = 1, share_size
            order = val_data(3, i)
            IF (order .EQ. 0) THEN
               blockdata_offset = file_size
               CALL mp_file_read_at_all(thefile, blockdata_offset, csp_dummy)
            ELSE
               start_index = SUM(val_data(2, 1:i-1))+1
               end_index = SUM(val_data(2, 1:i))
               blockdata_length = bdata_lens(order)
               blockdata_offset = bdata_offsets(order)
               CALL mp_file_read_at_all(thefile, blockdata_offset, csp(start_index:end_index), &
                                        msglen=blockdata_length)
            END IF
         END DO
      CASE (dbcsr_type_complex_8)
         ALLOCATE (cdp(darea_size))
         DO i = 1, share_size
            order = val_data(3, i)
            IF (order .EQ. 0) THEN
               blockdata_offset = file_size
               CALL mp_file_read_at_all(thefile, blockdata_offset, cdp_dummy)
            ELSE
               start_index = SUM(val_data(2, 1:i-1))+1
               end_index = SUM(val_data(2, 1:i))
               blockdata_length = bdata_lens(order)
               blockdata_offset = bdata_offsets(order)
               CALL mp_file_read_at_all(thefile, blockdata_offset, cdp(start_index:end_index), &
                                        msglen=blockdata_length)
            END IF
         END DO
      END SELECT
      CALL dbcsr_data_init(dblk)
      CALL dbcsr_data_new(dblk, data_type)
      IF (ALLOCATED(rdp)) dblk%d%r_dp => rdp
      IF (ALLOCATED(rsp)) dblk%d%r_sp => rsp
      IF (ALLOCATED(cdp)) dblk%d%c_dp => cdp
      IF (ALLOCATED(csp)) dblk%d%c_sp => csp

      CALL mp_file_close(thefile)

      CALL dbcsr_create(matrix_new, matrix_name, distribution, matrix_type, &
                        row_blk_size, col_blk_size, nze=darea_size, &
                        data_type=data_type)
      CALL dbcsr_datablock_redistribute(dblk, row_p, col_i, blk_p, proc_nblks, proc_darea_sizes, matrix_new)

      DEALLOCATE (subh2_offsets, subh3_offsets, bdata_offsets)
      DEALLOCATE (subh3_disps, bdata_disps)
      DEALLOCATE (linfo_lens, bdata_lens)
      DEALLOCATE (val_data, ginfo_vec, linfo_vec)
      DEALLOCATE (rowp, coli, blkp)
      IF (ALLOCATED(rdp)) DEALLOCATE (rdp)
      IF (ALLOCATED(rsp)) DEALLOCATE (rsp)
      IF (ALLOCATED(cdp)) DEALLOCATE (cdp)
      IF (ALLOCATED(csp)) DEALLOCATE (csp)
      CALL dbcsr_data_clear_pointer(dblk)
      DEALLOCATE (dblk%d)

      CALL timestop(handle)
   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param arr ...
!> \param cumsum ...
! **************************************************************************************************
      SUBROUTINE cumsum_l(arr, cumsum)
      INTEGER(kind=file_offset), DIMENSION(:), &
         INTENT(IN)                                      :: arr
      INTEGER(kind=file_offset), DIMENSION(SIZE(arr)), &
         INTENT(OUT)                                     :: cumsum

      INTEGER                                            :: i

         cumsum(1) = arr(1)
         DO i = 2, SIZE(arr)
            cumsum(i) = cumsum(i-1)+arr(i)
         END DO
      END SUBROUTINE cumsum_l

   END SUBROUTINE dbcsr_binary_read

! **************************************************************************************************
!> \brief Prints the sum of the elements for each block
!> \param[in] matrix          matrix
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE dbcsr_print_block_sum(matrix, unit_nr)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN), OPTIONAL                      :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_print_block_sum', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4)                               :: blk_sum_c_sp
      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8)                               :: blk_sum_c_dp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      INTEGER                                            :: bc, blk, blk_p, br, handle, iunit, m, &
                                                            mn, n
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_offset, col_blk_size, &
                                                            row_blk_offset, row_blk_size
      REAL(KIND=real_4)                                  :: blk_sum_r_sp
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8)                                  :: blk_sum_r_dp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. dbcsr_valid_index(matrix)) &
         CALL cp__w("dbcsr/ops/dbcsr_io.F",1222,"Can not print invalid matrix.")

      iunit = default_output_unit
      IF (PRESENT(unit_nr)) iunit = unit_nr

      SELECT CASE (matrix%m%data_type)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_get_data(matrix%m%data_area, r_dp)
      CASE (dbcsr_type_real_4)
         CALL dbcsr_get_data(matrix%m%data_area, r_sp)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_get_data(matrix%m%data_area, c_dp)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_get_data(matrix%m%data_area, c_sp)
      END SELECT
      row_blk_size => array_data(matrix%m%row_blk_size)
      col_blk_size => array_data(matrix%m%col_blk_size)
      row_blk_offset => array_data(matrix%m%row_blk_offset)
      col_blk_offset => array_data(matrix%m%col_blk_offset)

      IF (matrix%m%nblks .GT. 0) THEN
         DO br = 1, matrix%m%nblkrows_total
            m = row_blk_size(br)
            DO blk = matrix%m%row_p(br)+1, matrix%m%row_p(br+1)
               bc = matrix%m%col_i(blk)
               n = col_blk_size(bc)
               mn = m*n
               blk_p = ABS(matrix%m%blk_p(blk))
               block_exists: IF (blk_p .NE. 0) THEN
                  IF (mn .GT. 0) THEN
                     SELECT CASE (matrix%m%data_type)
                     CASE (dbcsr_type_real_8)
                        blk_sum_r_dp = SUM(r_dp(blk_p:blk_p+mn-1))
                        WRITE (iunit, '("Block (",I6,",",I6,") sum = ",ES18.9)') &
                           br, bc, blk_sum_r_dp
                     CASE (dbcsr_type_real_4)
                        blk_sum_r_sp = SUM(r_sp(blk_p:blk_p+mn-1))
                        WRITE (iunit, '(" (",I6,",",I6,") sum = ",ES18.9)') &
                           br, bc, blk_sum_r_sp
                     CASE (dbcsr_type_complex_8)
                        blk_sum_c_dp = SUM(c_dp(blk_p:blk_p+mn-1))
                        WRITE (iunit, '(" (",I6,",",I6,") sum = ",ES18.9," I*",ES18.9)') &
                           br, bc, REAL(blk_sum_c_dp), AIMAG(blk_sum_c_dp)
                     CASE (dbcsr_type_complex_4)
                        blk_sum_c_sp = SUM(c_sp(blk_p:blk_p+mn-1))
                        WRITE (iunit, '(" (",I6,",",I6,") sum = ",ES18.9," I*",ES18.9)') &
                           br, bc, REAL(blk_sum_c_sp), AIMAG(blk_sum_c_sp)
                     END SELECT
                  ELSE
                     blk_sum_r_dp = 0.0_dp
                     WRITE (iunit, '(" (",I6,",",I6,") sum = ",ES18.9)') &
                        br, bc, blk_sum_r_dp
                  ENDIF
               ENDIF block_exists
            ENDDO
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE dbcsr_print_block_sum

END MODULE dbcsr_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Fourth layer of the dbcsr matrix-matrix multiplication.
!>         It hides the differences between performing calculations on the
!>         accelerator device or on the CPU.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2010-02-23 Moved from dbcsr_operations
!>  - 2011-11    Moved parameter-stack processing routines to
!>               dbcsr_mm_methods.
!>  - 2013-01    extensive refactoring (Ole Schuett)
!>  - 2014-04    generalized into acc-framework (Ole Schuett)
! **************************************************************************************************
MODULE dbcsr_mm_accdrv
   USE acc_devmem,                      ONLY: acc_devmem_allocate_bytes,&
                                              acc_devmem_allocated,&
                                              acc_devmem_deallocate,&
                                              acc_devmem_host2dev,&
                                              acc_devmem_setzero_bytes,&
                                              acc_devmem_type
   USE acc_event,                       ONLY: acc_event_create,&
                                              acc_event_destroy,&
                                              acc_event_query,&
                                              acc_event_record,&
                                              acc_event_type,&
                                              acc_stream_wait_event
   USE acc_hostmem,                     ONLY: acc_hostmem_allocate,&
                                              acc_hostmem_deallocate
   USE acc_stream,                      ONLY: acc_stream_associated,&
                                              acc_stream_create,&
                                              acc_stream_destroy,&
                                              acc_stream_priority_range,&
                                              acc_stream_synchronize,&
                                              acc_stream_type
   USE dbcsr_acc_operations,            ONLY: dbcsr_acc_do_mm_stack
   USE dbcsr_block_operations,          ONLY: block_add
   USE dbcsr_config,                    ONLY: &
        accdrv_binning_binsize, accdrv_binning_nbins, accdrv_min_flop_sort, &
        accdrv_posterior_buffers, accdrv_posterior_streams, accdrv_priority_buffers, &
        accdrv_priority_streams, default_resize_factor, mm_stack_size
   USE dbcsr_data_methods,              ONLY: dbcsr_data_dev2host,&
                                              dbcsr_data_ensure_size,&
                                              dbcsr_data_get_size,&
                                              dbcsr_data_get_type,&
                                              dbcsr_data_new,&
                                              dbcsr_data_release
   USE dbcsr_mem_methods,               ONLY: dbcsr_mempool_destruct,&
                                              dbcsr_mempool_ensure_capacity,&
                                              dbcsr_memtype_setup
   USE dbcsr_mm_types,                  ONLY: dbcsr_ps_width,&
                                              stack_descriptor_type
   USE dbcsr_toollib,                   ONLY: sort
   USE dbcsr_types,                     ONLY: dbcsr_data_area_type,&
                                              dbcsr_data_obj,&
                                              dbcsr_memtype_type,&
                                              dbcsr_type,&
                                              dbcsr_work_type
   USE kinds,                           ONLY: default_string_length,&
                                              int_4,&
                                              int_4_size,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_accdrv'

   LOGICAL, PARAMETER, PRIVATE :: careful_mod = .FALSE.

! **************************************************************************************************
   TYPE dbcsr_mm_accdrv_type
      PRIVATE
      TYPE(dbcsr_work_type), POINTER           :: product_wm => Null()
      TYPE(dbcsr_data_obj)                     :: c_buffer
      LOGICAL                                  :: phase_out = .FALSE.
      INTEGER                                  :: nbuffers_phaseout
   END TYPE dbcsr_mm_accdrv_type

! **************************************************************************************************

   PUBLIC :: dbcsr_mm_accdrv_type
   PUBLIC :: dbcsr_mm_accdrv_lib_init, dbcsr_mm_accdrv_lib_finalize
   PUBLIC :: dbcsr_mm_accdrv_phaseout, dbcsr_mm_accdrv_barrier
   PUBLIC :: dbcsr_mm_accdrv_init, dbcsr_mm_accdrv_finalize
   PUBLIC :: dbcsr_mm_accdrv_process

   ! ===== Global Accelerator Memory =====
   ! Allocating memory on the device and host-pinned is slow.
   ! Therefore, the memory is allocated once and stored in global variables.

   TYPE stack_buffer_type
      TYPE(acc_devmem_type)                  :: devmem
      INTEGER, DIMENSION(:, :), POINTER       :: hostmem => Null()
      TYPE(acc_event_type)                   :: ready
      TYPE(acc_event_type)                   :: calculated
      TYPE(acc_stream_type)                  :: stream
   END TYPE stack_buffer_type

   TYPE thread_private_type
      TYPE(stack_buffer_type), DIMENSION(:), POINTER     :: stack_buffers => Null()
      TYPE(dbcsr_memtype_type)                           :: memtype_cbuffer
      ! ensure that array-elements are on different cache lines
      INTEGER(kind=int_4), DIMENSION(64)                 :: padding
   END TYPE thread_private_type

   TYPE(thread_private_type), SAVE, DIMENSION(:), ALLOCATABLE, TARGET :: all_thread_privates
   TYPE(acc_stream_type), SAVE                            :: upload_stream
   TYPE(acc_stream_type), SAVE, DIMENSION(:), POINTER     :: posterior_streams => Null()
   TYPE(acc_stream_type), SAVE, DIMENSION(:), POINTER     :: priority_streams => Null()
   TYPE(acc_event_type), SAVE, DIMENSION(:), POINTER     :: barrier_events => Null()
   INTEGER, SAVE                                          :: barrier_counter = 0

CONTAINS

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_lib_init()

      INTEGER                                            :: nthreads

      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()

!$OMP     MASTER
      ALLOCATE (all_thread_privates(0:nthreads-1))
!$OMP     END MASTER
!$OMP     BARRIER
   END SUBROUTINE dbcsr_mm_accdrv_lib_init

! **************************************************************************************************
!> \brief Finalize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_lib_finalize()

      INTEGER                                            :: ithread
      TYPE(thread_private_type), POINTER                 :: thread_privates

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      thread_privates => all_thread_privates(ithread)

      IF (ASSOCIATED(thread_privates%stack_buffers)) &
         CALL deallocate_stackbuffers()

      IF (ASSOCIATED(thread_privates%memtype_cbuffer%pool)) &
         CALL dbcsr_mempool_destruct(thread_privates%memtype_cbuffer%pool)

!$OMP     BARRIER
!$OMP     MASTER
      DEALLOCATE (all_thread_privates)
!$OMP     END MASTER

      !How much memory is still allocated on the card?
      !istat = dbcsr_acc_dev_mem_info(mem_free, mem_avail)
      !WRITE (*,*) "after finalize acc mem: ",mem_free, mem_avail, istat

      CALL deallocate_streams()

   END SUBROUTINE dbcsr_mm_accdrv_lib_finalize

! **************************************************************************************************
!> \brief Initializes a multiplication cycle for new set of C-blocks.
!> \param this ...
!> \param product_wm ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_init(this, product_wm)
      TYPE(dbcsr_mm_accdrv_type), INTENT(INOUT)          :: this
      TYPE(dbcsr_work_type), POINTER                     :: product_wm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_accdrv_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ithread, nthreads
      TYPE(dbcsr_data_obj)                               :: c_area
      TYPE(thread_private_type), POINTER                 :: thread_privates

      CALL timeset(routineN, handle)

      nthreads = 1; ithread = 0
!$    nthreads = OMP_GET_NUM_THREADS(); ithread = OMP_GET_THREAD_NUM()
      thread_privates => all_thread_privates(ithread)

      ! Setup global data which is reused in between multiplications
      !------------------------------------------------------------------------
      CALL setup_streams()
      CALL setup_stackbuffers(this)

      !Each thread has its own memtype with its own mempool
      i = MOD(ithread, SIZE(priority_streams))+1
      CALL dbcsr_memtype_setup(thread_privates%memtype_cbuffer, has_pool=.TRUE., acc_hostalloc=.TRUE., &
                               acc_devalloc=.TRUE., acc_stream=priority_streams(i))
      CALL dbcsr_mempool_ensure_capacity(thread_privates%memtype_cbuffer%pool, capacity=1)

      ! Setup things for this particular multipcation
      !------------------------------------------------------------------------
      this%product_wm => product_wm
      c_area = this%product_wm%data_area

      CALL dbcsr_data_new(this%c_buffer, data_type=dbcsr_data_get_type(c_area), &
                          data_size=dbcsr_data_get_size(c_area), memory_type=thread_privates%memtype_cbuffer)

      CALL acc_devmem_setzero_bytes(this%c_buffer%d%acc_devmem, &
                                    stream=this%c_buffer%d%memory_type%acc_stream)

      CALL acc_event_record(this%c_buffer%d%acc_ready, &
                            stream=this%c_buffer%d%memory_type%acc_stream)

      CALL timestop(handle)
   END SUBROUTINE dbcsr_mm_accdrv_init

! **************************************************************************************************
!> \brief Helper routine used by dbcsr_mm_accdrv_init()
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE setup_streams()

      INTEGER                                            :: greatest_priority, least_priority

!$OMP MASTER
      CALL acc_stream_priority_range(least_priority, greatest_priority)
      CALL stream_array_force_size(priority_streams, "Calc (priority)", &
                                   n=accdrv_priority_streams, priority=greatest_priority)
      CALL stream_array_force_size(posterior_streams, "Calc (posterior)", &
                                   n=accdrv_posterior_streams, events=barrier_events)

      !create upload stream
      IF (.NOT. acc_stream_associated(upload_stream)) &
         CALL acc_stream_create(upload_stream, "Stackbuf h2d")
!$OMP END MASTER
! Other threads have to wait until streams are created
!$OMP BARRIER

   END SUBROUTINE setup_streams

! **************************************************************************************************
!> \brief Helper routine used by setup_streams() and dbcsr_mm_accdrv_lib_finalize()
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE deallocate_streams()

!$OMP MASTER
      IF (acc_stream_associated(upload_stream)) &
         CALL acc_stream_destroy(upload_stream)

      CALL stream_array_force_size(priority_streams, "Calc (priority)", n=0)
      CALL stream_array_force_size(posterior_streams, "Calc (posterior)", n=0, events=barrier_events)
!$OMP END MASTER

   END SUBROUTINE deallocate_streams

! **************************************************************************************************
!> \brief Helper routine
!> \param streams ...
!> \param basename ...
!> \param n ...
!> \param events ...
!> \param priority ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stream_array_force_size(streams, basename, n, events, priority)
      TYPE(acc_stream_type), DIMENSION(:), POINTER       :: streams
      CHARACTER(len=*), INTENT(IN)                       :: basename
      INTEGER, INTENT(IN)                                :: n
      TYPE(acc_event_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: events
      INTEGER, INTENT(IN), OPTIONAL                      :: priority

      CHARACTER(len=default_string_length)               :: name
      INTEGER                                            :: i

      IF (ASSOCIATED(streams)) THEN
         IF (SIZE(streams) /= n) THEN
            DO i = 1, SIZE(streams)
               CALL acc_stream_destroy(streams(i))
               IF (PRESENT(events)) CALL acc_event_destroy(events(i))
            END DO
            DEALLOCATE (streams)
            IF (PRESENT(events)) DEALLOCATE (events)
         END IF
      END IF

      IF (.NOT. ASSOCIATED(streams) .AND. n > 0) THEN
         ALLOCATE (streams(n))
         IF (PRESENT(events)) ALLOCATE (events(n))
         DO i = 1, SIZE(streams)
            WRITE (name, "(A,I3)") TRIM(basename), i
            CALL acc_stream_create(streams(i), name=TRIM(name), priority=priority)
            IF (PRESENT(events)) CALL acc_event_create(events(i))
         END DO
      END IF
   END SUBROUTINE stream_array_force_size

! **************************************************************************************************
!> \brief Helper routine used by dbcsr_mm_accdrv_init()
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE setup_stackbuffers(this)
      TYPE(dbcsr_mm_accdrv_type), INTENT(INOUT)          :: this

      INTEGER                                            :: i, ithread, my_posterior_buffers, &
                                                            my_priority_buffers, my_total_buffers, &
                                                            nthreads
      INTEGER, SAVE                                      :: curr_posterior_stream = 0, &
                                                            curr_priority_stream = 0
      TYPE(thread_private_type), POINTER                 :: thread_privates

      nthreads = 1; ithread = 0
!$    nthreads = OMP_GET_NUM_THREADS(); ithread = OMP_GET_THREAD_NUM()
      thread_privates => all_thread_privates(ithread)

      ! distribute total number of buffers approx. evenly among threads
      ! my_* variables hold number of buffers for current thread.
      my_priority_buffers = CEILING(REAL(accdrv_priority_buffers)/REAL(nthreads), int_4)
      my_posterior_buffers = CEILING(REAL(accdrv_posterior_buffers)/REAL(nthreads), int_4)

      this%nbuffers_phaseout = my_priority_buffers
      my_total_buffers = my_posterior_buffers+my_priority_buffers

      IF (ASSOCIATED(thread_privates%stack_buffers)) THEN
         IF (SIZE(thread_privates%stack_buffers) /= my_total_buffers) &
            CALL deallocate_stackbuffers()
      ENDIF

      IF (.NOT. ASSOCIATED(thread_privates%stack_buffers)) THEN
         ALLOCATE (thread_privates%stack_buffers(my_total_buffers))
         !Done with OMP CRITICAL because curr_*_stream are shared.
!$OMP        CRITICAL(crit_setup_stackbuffers)
         DO i = 1, my_total_buffers
            CALL acc_devmem_allocate_bytes(thread_privates%stack_buffers(i)%devmem, &
                                           int_4_size*dbcsr_ps_width*mm_stack_size)
            IF (i <= my_priority_buffers) THEN
               curr_priority_stream = MOD(curr_priority_stream, SIZE(priority_streams))+1
               thread_privates%stack_buffers(i)%stream = priority_streams(curr_priority_stream)
            ELSE
               curr_posterior_stream = MOD(curr_posterior_stream, SIZE(posterior_streams))+1
               thread_privates%stack_buffers(i)%stream = posterior_streams(curr_posterior_stream)
            ENDIF
            CALL acc_hostmem_allocate(thread_privates%stack_buffers(i)%hostmem, &
                                      dbcsr_ps_width, mm_stack_size, thread_privates%stack_buffers(i)%stream)
            CALL acc_event_create(thread_privates%stack_buffers(i)%ready)
            CALL acc_event_create(thread_privates%stack_buffers(i)%calculated)
         END DO
!$OMP       END CRITICAL(crit_setup_stackbuffers)
      END IF
   END SUBROUTINE setup_stackbuffers

! **************************************************************************************************
!> \brief Helper routine used by setup_stackbuffers() and dbcsr_mm_accdrv_lib_finalize()
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE deallocate_stackbuffers()

      INTEGER                                            :: i, ithread
      TYPE(stack_buffer_type), DIMENSION(:), POINTER     :: stack_buffers

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      stack_buffers => all_thread_privates(ithread)%stack_buffers

      DO i = 1, SIZE(stack_buffers)
         CALL acc_devmem_deallocate(stack_buffers(i)%devmem)
         CALL acc_hostmem_deallocate(stack_buffers(i)%hostmem, stack_buffers(i)%stream)
         CALL acc_event_destroy(stack_buffers(i)%ready)
         CALL acc_event_destroy(stack_buffers(i)%calculated)
      END DO
      DEALLOCATE (stack_buffers)
   END SUBROUTINE deallocate_stackbuffers

! **************************************************************************************************
!> \brief Finalizes a multiplication cycle for a set of C-blocks.
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_finalize(this)
      TYPE(dbcsr_mm_accdrv_type), INTENT(INOUT)          :: this

      TYPE(dbcsr_data_obj)                               :: c_area

! Transfer C-data from device to host and adding it to host's result

      CALL dbcsr_data_dev2host(this%c_buffer)
      !TODO: not async, yet
      CALL acc_stream_synchronize(this%c_buffer%d%memory_type%acc_stream)

      c_area = this%product_wm%data_area
      CALL block_add(c_area, this%c_buffer)
      CALL dbcsr_data_release(this%c_buffer)

   END SUBROUTINE dbcsr_mm_accdrv_finalize

! **************************************************************************************************
!> \brief Sort stack entries with respect to the c_id.
!> \param params_in ...
!> \param params_out ...
!> \param stack_size ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stack_sort(params_in, params_out, stack_size)
      INTEGER, INTENT(IN)                                :: stack_size
      INTEGER, DIMENSION(dbcsr_ps_width, stack_size), &
         INTENT(OUT)                                     :: params_out
      INTEGER, DIMENSION(dbcsr_ps_width, stack_size), &
         INTENT(IN)                                      :: params_in

      INTEGER                                            :: i
      INTEGER, DIMENSION(stack_size)                     :: c_sort, c_sort_ind

! sort by the C-blocks

      c_sort = params_in(6, :stack_size)
      CALL sort(c_sort, stack_size, c_sort_ind)
      DO i = 1, stack_size
         params_out(:, i) = params_in(:, c_sort_ind(i))
      END DO

   END SUBROUTINE stack_sort

! **************************************************************************************************
!> \brief Roughly order stacks with a cheaper Binning-scheme by Peter Messmer
!> \param params_in ...
!> \param params_out ...
!> \param stack_size ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stack_binning(params_in, params_out, stack_size)
      INTEGER, INTENT(IN)                                :: stack_size
      INTEGER, DIMENSION(dbcsr_ps_width, stack_size), &
         INTENT(OUT)                                     :: params_out
      INTEGER, DIMENSION(dbcsr_ps_width, stack_size), &
         INTENT(IN)                                      :: params_in

      INTEGER                                            :: bin_id, i, top
      INTEGER, DIMENSION(accdrv_binning_nbins)           :: bin_top
      INTEGER, DIMENSION(dbcsr_ps_width)                 :: val
      INTEGER, DIMENSION(dbcsr_ps_width, &
         accdrv_binning_binsize, accdrv_binning_nbins)   :: bin_arr

      bin_top = 1
      top = 1
      DO i = 1, stack_size
         val(:) = params_in(:, i)
         bin_id = 1+INT(MODULO(INT(val(6)*(val(6)+3), KIND=int_8), &
                               INT(accdrv_binning_nbins, KIND=int_8)))
         IF (bin_top(bin_id) > accdrv_binning_binsize) THEN
            params_out(:, top:top+bin_top(bin_id)-2) = bin_arr(:, 1:bin_top(bin_id)-1, bin_id)
            top = top+bin_top(bin_id)-1
            bin_top(bin_id) = 1
         ENDIF
         bin_arr(:, bin_top(bin_id), bin_id) = val(:)
         bin_top(bin_id) = bin_top(bin_id)+1
      END DO
      DO i = 1, accdrv_binning_nbins
         IF (bin_top(i) > 1) THEN
            params_out(:, top:top+bin_top(i)-2) = bin_arr(:, 1:bin_top(i)-1, i)
            top = top+bin_top(i)-1
         ENDIF
      END DO

   END SUBROUTINE stack_binning

! **************************************************************************************************
!> \brief Signal approaching end of multiplication
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_phaseout(this)
      TYPE(dbcsr_mm_accdrv_type), INTENT(INOUT)          :: this

      this%phase_out = .TRUE.
   END SUBROUTINE dbcsr_mm_accdrv_phaseout

! **************************************************************************************************
!> \brief After beeing called by ALL threads, it installs a special barrier that
!>        forces the priority-streams to wait for the normal-streams.
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_barrier()

      INTEGER                                            :: i, j, nthreads

      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()

!$OMP     CRITICAL
      barrier_counter = MOD(barrier_counter+1, nthreads)
      IF (barrier_counter == 0) THEN
         DO i = 1, SIZE(posterior_streams)
            CALL acc_event_record(barrier_events(i), stream=posterior_streams(i))
            DO j = 1, SIZE(priority_streams)
               CALL acc_stream_wait_event(priority_streams(j), barrier_events(i))
            ENDDO
         ENDDO
      ENDIF
!$OMP     END CRITICAL
   END SUBROUTINE dbcsr_mm_accdrv_barrier

! **************************************************************************************************
!> \brief Processes a given stack using accelerator
!> \param this ...
!> \param left ...
!> \param right ...
!> \param params ...
!> \param stack_size ...
!> \param stack_descr ...
!> \param success ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_accdrv_process(this, left, right, params, stack_size, &
                                      stack_descr, success)
      TYPE(dbcsr_mm_accdrv_type), INTENT(INOUT)          :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, INTENT(IN)                                :: stack_size
      INTEGER, DIMENSION(dbcsr_ps_width, stack_size), &
         INTENT(INOUT)                                   :: params
      TYPE(stack_descriptor_type), INTENT(IN)            :: stack_descr
      LOGICAL, INTENT(OUT)                               :: success

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_accdrv_process', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, error_handle2, &
                                                            flop_per_entry, i, ithread, &
                                                            stacked_datasize
      INTEGER, DIMENSION(:, :), POINTER                  :: stackbuf_hostmem_cropped
      TYPE(dbcsr_data_area_type), POINTER                :: a_area, b_area, c_area
      TYPE(stack_buffer_type), DIMENSION(:), POINTER     :: stack_buffers
      TYPE(stack_buffer_type), POINTER                   :: stackbuf

      NULLIFY (stackbuf, stackbuf_hostmem_cropped, stack_buffers)

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      stack_buffers => all_thread_privates(ithread)%stack_buffers

      DO i = 1, SIZE(stack_buffers)
         IF (this%phase_out .AND. i > this%nbuffers_phaseout) EXIT
         IF (acc_event_query(stack_buffers(i)%calculated)) THEN
            stackbuf => stack_buffers(i)
            EXIT
         ENDIF
      ENDDO

      IF (.NOT. ASSOCIATED(stackbuf)) THEN
         success = .FALSE.
         RETURN
      END IF

      CALL timeset(routineN, error_handle)
      stacked_datasize = this%product_wm%datasize
      CALL dbcsr_data_ensure_size(this%c_buffer, stacked_datasize, &
                                  factor=default_resize_factor, zero_pad=.TRUE.)

      !===========================================================================
      ! sort the stack. Since this costs CPU time, only a good idea if the CPUs
      ! are not too busy, or device gain is very large
      CALL timeset(routineN//"_sort", error_handle2)
      flop_per_entry = 2*stack_descr%max_m*stack_descr%max_n*stack_descr%max_k

      IF (flop_per_entry > accdrv_min_flop_sort) THEN
         CALL stack_sort(params, stackbuf%hostmem, stack_size)
      ELSE
         CALL stack_binning(params, stackbuf%hostmem, stack_size)
      ENDIF

      CALL timestop(error_handle2)

      a_area => left%data_area%d
      b_area => right%data_area%d
      c_area => this%c_buffer%d

      !WRITE (*,*) "dbcsr_mm_accdrv_process: a_area%memory_type ", a_area%memory_type
      !WRITE (*,*) "dbcsr_mm_accdrv_process: b_area%memory_type ", b_area%memory_type
      !WRITE (*,*) "dbcsr_mm_accdrv_process: c_area%memory_type ", c_area%memory_type

      IF (.NOT. acc_devmem_allocated(a_area%acc_devmem)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_accdrv.F",591,"dbcsr_mm_accdrv_process: a_area%acc_devmem not allocated")
      IF (.NOT. acc_devmem_allocated(b_area%acc_devmem)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_accdrv.F",593,"dbcsr_mm_accdrv_process: b_area%acc_devmem not allocated")
      IF (.NOT. acc_devmem_allocated(c_area%acc_devmem)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_accdrv.F",595,"dbcsr_mm_accdrv_process: c_area%acc_devmem not allocated")

      ! start uploading stacks only after a, b, and c are ready
      CALL acc_stream_wait_event(upload_stream, a_area%acc_ready)
      CALL acc_stream_wait_event(upload_stream, b_area%acc_ready)
      CALL acc_stream_wait_event(upload_stream, c_area%acc_ready)

      stackbuf_hostmem_cropped => stackbuf%hostmem(:, 1:stack_size)
      CALL acc_devmem_host2dev(stackbuf%devmem, hostmem=stackbuf_hostmem_cropped, stream=upload_stream)
      CALL acc_event_record(stackbuf%ready, stream=upload_stream)

      CALL acc_stream_wait_event(stackbuf%stream, stackbuf%ready)

      CALL dbcsr_acc_do_mm_stack(stackbuf%devmem, stack_size, c_area%data_type, &
                                 a_data=a_area%acc_devmem, &
                                 b_data=b_area%acc_devmem, &
                                 c_data=c_area%acc_devmem, &
                                 m_max=stack_descr%max_m, &
                                 n_max=stack_descr%max_n, &
                                 k_max=stack_descr%max_k, &
                                 def_mnk=stack_descr%defined_mnk, &
                                 stream=stackbuf%stream, &
                                 success=success)

      IF (success) THEN
         CALL acc_event_record(stackbuf%calculated, stream=stackbuf%stream)
         CALL acc_stream_wait_event(a_area%memory_type%acc_stream, stackbuf%calculated)
         CALL acc_stream_wait_event(b_area%memory_type%acc_stream, stackbuf%calculated)
         CALL acc_stream_wait_event(c_area%memory_type%acc_stream, stackbuf%calculated)
      ENDIF

      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_mm_accdrv_process

END MODULE dbcsr_mm_accdrv

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Stacks of small matrix multiplications
!> \author  Urban Borstnik
!> \date    2011-09-26
!> \version 0.9
!>
!> <b>Modification history:</b>
!  - 2011-09-26 Split dbcsr_internal_operations
! **************************************************************************************************
MODULE dbcsr_mm_hostdrv
   USE dbcsr_config,                    ONLY: has_acc,&
                                              mm_driver,&
                                              mm_driver_blas,&
                                              mm_driver_matmul,&
                                              mm_driver_smm,&
                                              mm_driver_xsmm
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_size
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error
   USE dbcsr_mm_types,                  ONLY: dbcsr_ps_width,&
                                              p_a_first,&
                                              p_b_first,&
                                              p_c_first,&
                                              p_k,&
                                              p_m,&
                                              p_n,&
                                              stack_descriptor_type
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_type,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8,&
                                              dbcsr_work_type
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              real_4,&
                                              real_8,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_hostdrv'

   CHARACTER(len=*), PARAMETER, PRIVATE :: int_print = "(10(1X,I7))"

   PUBLIC :: dbcsr_mm_hostdrv_lib_init, dbcsr_mm_hostdrv_lib_finalize
   PUBLIC :: dbcsr_mm_hostdrv_process
   PUBLIC :: dbcsr_mm_hostdrv_type
   PUBLIC :: dbcsr_mm_hostdrv_init

   LOGICAL, PARAMETER :: debug_mod = .FALSE.
   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   TYPE dbcsr_mm_hostdrv_type
      TYPE(dbcsr_data_obj)          :: data_area
   END TYPE dbcsr_mm_hostdrv_type

CONTAINS

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_hostdrv_lib_init()
   END SUBROUTINE dbcsr_mm_hostdrv_lib_init

! **************************************************************************************************
!> \brief Finalize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_hostdrv_lib_finalize()
   END SUBROUTINE dbcsr_mm_hostdrv_lib_finalize

! **************************************************************************************************
!> \brief Initialize the library
!> \param this ...
!> \param product_wm ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_hostdrv_init(this, product_wm)
      TYPE(dbcsr_mm_hostdrv_type), INTENT(INOUT)         :: this
      TYPE(dbcsr_work_type), POINTER                     :: product_wm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_hostdrv_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      this%data_area = product_wm%data_area

      CALL timestop(handle)

   END SUBROUTINE dbcsr_mm_hostdrv_init

! **************************************************************************************************
!> \brief Calls the various drivers that process the stack.
!>
!> \param this ...
!> \param[in] left Left-matrix data
!> \param[in] right Right-matrix data
!> \param[in] params           Stack of GEMM parameters
!> \param stack_size ...
!> \param stack_descr ...
!> \param success ...
!> \param used_smm ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_hostdrv_process(this, left, right, params, stack_size, &
                                       stack_descr, success, used_smm)
      TYPE(dbcsr_mm_hostdrv_type), INTENT(INOUT)         :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, INTENT(IN)                                :: stack_size
      INTEGER, DIMENSION(1:dbcsr_ps_width, stack_size), &
         INTENT(INOUT)                                   :: params
      TYPE(stack_descriptor_type), INTENT(IN)            :: stack_descr
      LOGICAL, INTENT(OUT)                               :: success, used_smm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_hostdrv_process', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: careful = careful_mod, dbg = .FALSE.

      INTEGER                                            :: error_handle, sp
      REAL(KIND=dp)                                      :: rnd

      IF (has_acc) & !for cpu-only runs this is called too often
         CALL timeset(routineN, error_handle)

      success = .TRUE. !host driver never failes...hopefully
      used_smm = .FALSE.

      IF (dbg) THEN
         CALL RANDOM_NUMBER(rnd)
         IF (rnd < 0.01_dp) THEN
            WRITE (*, *) routineN//" Stack size", stack_size, dbcsr_ps_width
            CALL print_gemm_parameters(params(:, 1:stack_size))
         ENDIF
      ENDIF

      ! Verify stack consistency.  Only the upper bound is verified.
      IF (careful) THEN
         DO sp = 1, stack_size
            CALL dbcsr_assert(params(p_a_first, sp) &
                              +params(p_m, sp)*params(p_k, sp)-1, &
                              "LE", dbcsr_data_get_size(left%data_area), &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "A data out of bounds.", 172)
            CALL dbcsr_assert(params(p_b_first, sp) &
                              +params(p_k, sp)*params(p_n, sp)-1, &
                              "LE", dbcsr_data_get_size(right%data_area), &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "B data out of bounds.", 177)
            CALL dbcsr_assert(params(p_c_first, sp) &
                              +params(p_m, sp)*params(p_n, sp)-1, &
                              "LE", dbcsr_data_get_size(this%data_area), &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "C data out of bounds.", 182)
         ENDDO
      ENDIF

      SELECT CASE (mm_driver)
      CASE (mm_driver_matmul)
         SELECT CASE (this%data_area%d%data_type)
         CASE (dbcsr_type_real_4)
            CALL internal_process_mm_stack_s(params, &
                                             stack_size, &
                                             left%data_area%d%r_sp, right%data_area%d%r_sp, this%data_area%d%r_sp)
         CASE (dbcsr_type_real_8)
            CALL internal_process_mm_stack_d(params, &
                                             stack_size, &
                                             left%data_area%d%r_dp, right%data_area%d%r_dp, this%data_area%d%r_dp)
         CASE (dbcsr_type_complex_4)
            CALL internal_process_mm_stack_c(params, &
                                             stack_size, &
                                             left%data_area%d%c_sp, right%data_area%d%c_sp, this%data_area%d%c_sp)
         CASE (dbcsr_type_complex_8)
            CALL internal_process_mm_stack_z(params, &
                                             stack_size, &
                                             left%data_area%d%c_dp, right%data_area%d%c_dp, this%data_area%d%c_dp)
         CASE default
            CALL cp__b("dbcsr/mm/dbcsr_mm_hostdrv.F",206,"Invalid data type")
         END SELECT
      CASE (mm_driver_smm)
         SELECT CASE (this%data_area%d%data_type)
         CASE (dbcsr_type_real_4)
            CALL smm_process_mm_stack_s(stack_descr, params, &
                                        stack_size, &
                                        left%data_area%d%r_sp, right%data_area%d%r_sp, this%data_area%d%r_sp, used_smm)
         CASE (dbcsr_type_real_8)
            CALL smm_process_mm_stack_d(stack_descr, params, &
                                        stack_size, &
                                        left%data_area%d%r_dp, right%data_area%d%r_dp, this%data_area%d%r_dp, used_smm)
         CASE (dbcsr_type_complex_4)
            CALL smm_process_mm_stack_c(stack_descr, params, &
                                        stack_size, &
                                        left%data_area%d%c_sp, right%data_area%d%c_sp, this%data_area%d%c_sp, used_smm)
         CASE (dbcsr_type_complex_8)
            CALL smm_process_mm_stack_z(stack_descr, params, &
                                        stack_size, &
                                        left%data_area%d%c_dp, right%data_area%d%c_dp, this%data_area%d%c_dp, used_smm)
         CASE default
            CALL cp__b("dbcsr/mm/dbcsr_mm_hostdrv.F",227,"Invalid data type")
         END SELECT

      CASE (mm_driver_xsmm)
         SELECT CASE (this%data_area%d%data_type)
         CASE (dbcsr_type_real_4)
            CALL xsmm_process_mm_stack_s(stack_descr, params, stack_size, &
                                         left%data_area%d%r_sp, right%data_area%d%r_sp, this%data_area%d%r_sp, used_smm)
         CASE (dbcsr_type_real_8)
            CALL xsmm_process_mm_stack_d(stack_descr, params, stack_size, &
                                         left%data_area%d%r_dp, right%data_area%d%r_dp, this%data_area%d%r_dp, used_smm)
         CASE (dbcsr_type_complex_4)
            CALL xsmm_process_mm_stack_c(stack_descr, params, stack_size, &
                                         left%data_area%d%c_sp, right%data_area%d%c_sp, this%data_area%d%c_sp, used_smm)
         CASE (dbcsr_type_complex_8)
            CALL xsmm_process_mm_stack_z(stack_descr, params, stack_size, &
                                         left%data_area%d%c_dp, right%data_area%d%c_dp, this%data_area%d%c_dp, used_smm)
         CASE default
            CALL cp__b("dbcsr/mm/dbcsr_mm_hostdrv.F",245,"Invalid data type")
         END SELECT

      CASE (mm_driver_blas)
         SELECT CASE (this%data_area%d%data_type)
         CASE (dbcsr_type_real_4)
            CALL blas_process_mm_stack_s(params, &
                                         stack_size, &
                                         left%data_area%d%r_sp, right%data_area%d%r_sp, this%data_area%d%r_sp)
         CASE (dbcsr_type_real_8)
            CALL blas_process_mm_stack_d(params, &
                                         stack_size, &
                                         left%data_area%d%r_dp, right%data_area%d%r_dp, this%data_area%d%r_dp)
         CASE (dbcsr_type_complex_4)
            CALL blas_process_mm_stack_c(params, &
                                         stack_size, &
                                         left%data_area%d%c_sp, right%data_area%d%c_sp, this%data_area%d%c_sp)
         CASE (dbcsr_type_complex_8)
            CALL blas_process_mm_stack_z(params, &
                                         stack_size, &
                                         left%data_area%d%c_dp, right%data_area%d%c_dp, this%data_area%d%c_dp)
         CASE default
            CALL cp__b("dbcsr/mm/dbcsr_mm_hostdrv.F",267,"Invalid data type")
         END SELECT
      CASE default
         CALL cp__b("dbcsr/mm/dbcsr_mm_hostdrv.F",270,"Invalid multiplication driver")
      END SELECT

      IF (has_acc) & !for cpu-only runs this is called too often
         CALL timestop(error_handle)

   END SUBROUTINE dbcsr_mm_hostdrv_process

! **************************************************************************************************
!> \brief Helper-routine used by dbcsr_mm_hostdrv_process to print debug info.
!> \param params ...
! **************************************************************************************************
   SUBROUTINE print_gemm_parameters(params)
      INTEGER, DIMENSION(:, :), INTENT(in)               :: params

      INTEGER                                            :: sp

      DO sp = 1, SIZE(params, 2)
         WRITE (*, '(1X,A,1X,I7,":",3(1X,I4,","),".",3(1X,I12,","))') &
            "GEMM PARAMETERS", &
            sp, &
            params(p_m, sp), &
            params(p_k, sp), &
            params(p_n, sp), &
            params(p_a_first, sp), &
            params(p_b_first, sp), &
            params(p_c_first, sp)
      ENDDO
   END SUBROUTINE print_gemm_parameters

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Processes MM stack and issues BLAS xGEMM calls
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE blas_process_mm_stack_d(params,&
       stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'blas_process_mm_stack_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL DGEMM('N',&
            'N',&
            params(p_m,sp), params(p_n,sp),& !m, n
            params(p_k,sp),& ! k
            1.0_real_8,& ! alpha
            a_data(params(p_a_first,sp)),& ! A
            params(p_m,sp),& !lda
            b_data(params(p_b_first,sp)),& ! B
            params(p_k,sp),& !ldb
            1.0_real_8,& ! beta
            c_data(params(p_c_first,sp)), params(p_m,sp))
    ENDDO
  END SUBROUTINE blas_process_mm_stack_d

! **************************************************************************************************
!> \brief Processes MM stack and issues internal MM calls.
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE internal_process_mm_stack_d(params, stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'internal_process_mm_stack_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL internal_mm_d_nn(&
            params(p_m,sp),&
            params(p_n,sp),&
            params(p_k,sp),&
            a_data(params(p_a_first,sp)),&
            b_data(params(p_b_first,sp)),&
            c_data(params(p_c_first,sp)))
    ENDDO
  END SUBROUTINE internal_process_mm_stack_d


! **************************************************************************************************
!> \brief Processes MM stack and issues SMM library calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE smm_process_mm_stack_d(stack_descr, params,&
       stack_size,&
       a_data, b_data, c_data, used_smm)
    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'smm_process_mm_stack_d', &
      routineP = moduleN//':'//routineN

    ! We do not want to abort here, fall back to BLAS.
    used_smm=.FALSE.
    CALL blas_process_mm_stack_d(params, stack_size,a_data, b_data, c_data)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
  END SUBROUTINE smm_process_mm_stack_d


! **************************************************************************************************
!> \brief Processes MM stack and issues libxsmm calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
!> \param[out] used_smm        Flag to signal if an efficient kernel was used
!> \author Ole Schuett
! **************************************************************************************************
  SUBROUTINE xsmm_process_mm_stack_d(stack_descr, params,&
       stack_size, a_data, b_data, c_data, used_smm)


    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_8), DIMENSION(*), TARGET, INTENT(IN) :: a_data, b_data
    REAL(kind=real_8), DIMENSION(*), TARGET, &
      INTENT(INOUT)                           :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'libxsmm_process_mm_stack_d', &
      routineP = moduleN//':'//routineN

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
    ! We do not want to abort here, fall back to BLAS.
    CALL blas_process_mm_stack_d(params, stack_size,a_data, b_data, c_data)
    used_smm = .FALSE.

  END SUBROUTINE xsmm_process_mm_stack_d

! **************************************************************************************************
!> \brief ...
!> \param M ...
!> \param N ...
!> \param K ...
!> \param A ...
!> \param B ...
!> \param C ...
! **************************************************************************************************
  PURE SUBROUTINE internal_mm_d_nn(&
       M,N,K,A,B,C)
    INTEGER, INTENT(IN)                      :: M, N, K
    REAL(kind=real_8), INTENT(INOUT)                   :: C(M,N)
    REAL(kind=real_8), INTENT(IN)                      :: B(K,N)
    REAL(kind=real_8), INTENT(IN)                      :: A(M,K)
    C(:,:) = C(:,:) + MATMUL (A, B)
  END SUBROUTINE internal_mm_d_nn
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Processes MM stack and issues BLAS xGEMM calls
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE blas_process_mm_stack_z(params,&
       stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'blas_process_mm_stack_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL ZGEMM('N',&
            'N',&
            params(p_m,sp), params(p_n,sp),& !m, n
            params(p_k,sp),& ! k
            CMPLX(1.0, 0.0, real_8),& ! alpha
            a_data(params(p_a_first,sp)),& ! A
            params(p_m,sp),& !lda
            b_data(params(p_b_first,sp)),& ! B
            params(p_k,sp),& !ldb
            CMPLX(1.0, 0.0, real_8),& ! beta
            c_data(params(p_c_first,sp)), params(p_m,sp))
    ENDDO
  END SUBROUTINE blas_process_mm_stack_z

! **************************************************************************************************
!> \brief Processes MM stack and issues internal MM calls.
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE internal_process_mm_stack_z(params, stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'internal_process_mm_stack_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL internal_mm_z_nn(&
            params(p_m,sp),&
            params(p_n,sp),&
            params(p_k,sp),&
            a_data(params(p_a_first,sp)),&
            b_data(params(p_b_first,sp)),&
            c_data(params(p_c_first,sp)))
    ENDDO
  END SUBROUTINE internal_process_mm_stack_z


! **************************************************************************************************
!> \brief Processes MM stack and issues SMM library calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE smm_process_mm_stack_z(stack_descr, params,&
       stack_size,&
       a_data, b_data, c_data, used_smm)
    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(INOUT)      :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'smm_process_mm_stack_z', &
      routineP = moduleN//':'//routineN

    ! We do not want to abort here, fall back to BLAS.
    used_smm=.FALSE.
    CALL blas_process_mm_stack_z(params, stack_size,a_data, b_data, c_data)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
  END SUBROUTINE smm_process_mm_stack_z


! **************************************************************************************************
!> \brief Processes MM stack and issues libxsmm calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
!> \param[out] used_smm        Flag to signal if an efficient kernel was used
!> \author Ole Schuett
! **************************************************************************************************
  SUBROUTINE xsmm_process_mm_stack_z(stack_descr, params,&
       stack_size, a_data, b_data, c_data, used_smm)


    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_8), DIMENSION(*), TARGET, INTENT(IN) :: a_data, b_data
    COMPLEX(kind=real_8), DIMENSION(*), TARGET, &
      INTENT(INOUT)                           :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'libxsmm_process_mm_stack_z', &
      routineP = moduleN//':'//routineN

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
    ! We do not want to abort here, fall back to BLAS.
    CALL blas_process_mm_stack_z(params, stack_size,a_data, b_data, c_data)
    used_smm = .FALSE.

  END SUBROUTINE xsmm_process_mm_stack_z

! **************************************************************************************************
!> \brief ...
!> \param M ...
!> \param N ...
!> \param K ...
!> \param A ...
!> \param B ...
!> \param C ...
! **************************************************************************************************
  PURE SUBROUTINE internal_mm_z_nn(&
       M,N,K,A,B,C)
    INTEGER, INTENT(IN)                      :: M, N, K
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: C(M,N)
    COMPLEX(kind=real_8), INTENT(IN)                      :: B(K,N)
    COMPLEX(kind=real_8), INTENT(IN)                      :: A(M,K)
    C(:,:) = C(:,:) + MATMUL (A, B)
  END SUBROUTINE internal_mm_z_nn
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Processes MM stack and issues BLAS xGEMM calls
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE blas_process_mm_stack_s(params,&
       stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'blas_process_mm_stack_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL SGEMM('N',&
            'N',&
            params(p_m,sp), params(p_n,sp),& !m, n
            params(p_k,sp),& ! k
            1.0_real_4,& ! alpha
            a_data(params(p_a_first,sp)),& ! A
            params(p_m,sp),& !lda
            b_data(params(p_b_first,sp)),& ! B
            params(p_k,sp),& !ldb
            1.0_real_4,& ! beta
            c_data(params(p_c_first,sp)), params(p_m,sp))
    ENDDO
  END SUBROUTINE blas_process_mm_stack_s

! **************************************************************************************************
!> \brief Processes MM stack and issues internal MM calls.
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE internal_process_mm_stack_s(params, stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'internal_process_mm_stack_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL internal_mm_s_nn(&
            params(p_m,sp),&
            params(p_n,sp),&
            params(p_k,sp),&
            a_data(params(p_a_first,sp)),&
            b_data(params(p_b_first,sp)),&
            c_data(params(p_c_first,sp)))
    ENDDO
  END SUBROUTINE internal_process_mm_stack_s


! **************************************************************************************************
!> \brief Processes MM stack and issues SMM library calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE smm_process_mm_stack_s(stack_descr, params,&
       stack_size,&
       a_data, b_data, c_data, used_smm)
    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    REAL(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'smm_process_mm_stack_s', &
      routineP = moduleN//':'//routineN

    ! We do not want to abort here, fall back to BLAS.
    used_smm=.FALSE.
    CALL blas_process_mm_stack_s(params, stack_size,a_data, b_data, c_data)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
  END SUBROUTINE smm_process_mm_stack_s


! **************************************************************************************************
!> \brief Processes MM stack and issues libxsmm calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
!> \param[out] used_smm        Flag to signal if an efficient kernel was used
!> \author Ole Schuett
! **************************************************************************************************
  SUBROUTINE xsmm_process_mm_stack_s(stack_descr, params,&
       stack_size, a_data, b_data, c_data, used_smm)


    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    REAL(kind=real_4), DIMENSION(*), TARGET, INTENT(IN) :: a_data, b_data
    REAL(kind=real_4), DIMENSION(*), TARGET, &
      INTENT(INOUT)                           :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'libxsmm_process_mm_stack_s', &
      routineP = moduleN//':'//routineN

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
    ! We do not want to abort here, fall back to BLAS.
    CALL blas_process_mm_stack_s(params, stack_size,a_data, b_data, c_data)
    used_smm = .FALSE.

  END SUBROUTINE xsmm_process_mm_stack_s

! **************************************************************************************************
!> \brief ...
!> \param M ...
!> \param N ...
!> \param K ...
!> \param A ...
!> \param B ...
!> \param C ...
! **************************************************************************************************
  PURE SUBROUTINE internal_mm_s_nn(&
       M,N,K,A,B,C)
    INTEGER, INTENT(IN)                      :: M, N, K
    REAL(kind=real_4), INTENT(INOUT)                   :: C(M,N)
    REAL(kind=real_4), INTENT(IN)                      :: B(K,N)
    REAL(kind=real_4), INTENT(IN)                      :: A(M,K)
    C(:,:) = C(:,:) + MATMUL (A, B)
  END SUBROUTINE internal_mm_s_nn
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Processes MM stack and issues BLAS xGEMM calls
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE blas_process_mm_stack_c(params,&
       stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'blas_process_mm_stack_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL CGEMM('N',&
            'N',&
            params(p_m,sp), params(p_n,sp),& !m, n
            params(p_k,sp),& ! k
            CMPLX(1.0, 0.0, real_4),& ! alpha
            a_data(params(p_a_first,sp)),& ! A
            params(p_m,sp),& !lda
            b_data(params(p_b_first,sp)),& ! B
            params(p_k,sp),& !ldb
            CMPLX(1.0, 0.0, real_4),& ! beta
            c_data(params(p_c_first,sp)), params(p_m,sp))
    ENDDO
  END SUBROUTINE blas_process_mm_stack_c

! **************************************************************************************************
!> \brief Processes MM stack and issues internal MM calls.
!>
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE internal_process_mm_stack_c(params, stack_size,&
       a_data, b_data, c_data)
    INTEGER, INTENT(IN)                       :: stack_size
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data

    CHARACTER(len=*), PARAMETER :: routineN = 'internal_process_mm_stack_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                   :: sp

!   ---------------------------------------------------------------------------

    DO sp = 1, stack_size
       CALL internal_mm_c_nn(&
            params(p_m,sp),&
            params(p_n,sp),&
            params(p_k,sp),&
            a_data(params(p_a_first,sp)),&
            b_data(params(p_b_first,sp)),&
            c_data(params(p_c_first,sp)))
    ENDDO
  END SUBROUTINE internal_process_mm_stack_c


! **************************************************************************************************
!> \brief Processes MM stack and issues SMM library calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
! **************************************************************************************************
  SUBROUTINE smm_process_mm_stack_c(stack_descr, params,&
       stack_size,&
       a_data, b_data, c_data, used_smm)
    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(IN)         :: a_data, &
                                                 b_data
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(INOUT)      :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'smm_process_mm_stack_c', &
      routineP = moduleN//':'//routineN

    ! We do not want to abort here, fall back to BLAS.
    used_smm=.FALSE.
    CALL blas_process_mm_stack_c(params, stack_size,a_data, b_data, c_data)

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
  END SUBROUTINE smm_process_mm_stack_c


! **************************************************************************************************
!> \brief Processes MM stack and issues libxsmm calls
!>
!> \param stack_descr ...
!> \param[in] params           Stack of MM parameters
!> \param[in] stack_size       Number of parameters
!> \param[in] a_data           Left-matrix data
!> \param[in] b_data           Right-matrix data
!> \param[in,out] c_data       Product data
!> \param[out] used_smm        Flag to signal if an efficient kernel was used
!> \author Ole Schuett
! **************************************************************************************************
  SUBROUTINE xsmm_process_mm_stack_c(stack_descr, params,&
       stack_size, a_data, b_data, c_data, used_smm)


    INTEGER, INTENT(IN)                       :: stack_size
    TYPE(stack_descriptor_type), INTENT(IN)   :: stack_descr
    INTEGER, DIMENSION(dbcsr_ps_width,1:stack_size), &
      INTENT(IN)                              :: params
    COMPLEX(kind=real_4), DIMENSION(*), TARGET, INTENT(IN) :: a_data, b_data
    COMPLEX(kind=real_4), DIMENSION(*), TARGET, &
      INTENT(INOUT)                           :: c_data
    LOGICAL, INTENT(OUT)                      :: used_smm

    CHARACTER(len=*), PARAMETER :: routineN = 'libxsmm_process_mm_stack_c', &
      routineP = moduleN//':'//routineN

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stack_descr))==-1) EXIT ;  END DO ; ENDIF
    ! We do not want to abort here, fall back to BLAS.
    CALL blas_process_mm_stack_c(params, stack_size,a_data, b_data, c_data)
    used_smm = .FALSE.

  END SUBROUTINE xsmm_process_mm_stack_c

! **************************************************************************************************
!> \brief ...
!> \param M ...
!> \param N ...
!> \param K ...
!> \param A ...
!> \param B ...
!> \param C ...
! **************************************************************************************************
  PURE SUBROUTINE internal_mm_c_nn(&
       M,N,K,A,B,C)
    INTEGER, INTENT(IN)                      :: M, N, K
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: C(M,N)
    COMPLEX(kind=real_4), INTENT(IN)                      :: B(K,N)
    COMPLEX(kind=real_4), INTENT(IN)                      :: A(M,K)
    C(:,:) = C(:,:) + MATMUL (A, B)
  END SUBROUTINE internal_mm_c_nn

END MODULE dbcsr_mm_hostdrv
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Fourth layer of the dbcsr matrix-matrix multiplication.
!>         It hides the differences between performing calculations on the
!>         accelerator device or on the CPU.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2010-02-23 Moved from dbcsr_operations
!>  - 2011-11    Moved parameter-stack processing routines to
!>               dbcsr_mm_methods.
!>  - 2013-01    extensive refactoring (Ole Schuett)
! **************************************************************************************************
MODULE dbcsr_mm_sched
   USE dbcsr_block_operations,          ONLY: dbcsr_data_clear
   USE dbcsr_config,                    ONLY: accdrv_avoid_after_busy,&
                                              accdrv_do_inhomogenous,&
                                              accdrv_min_flop_process,&
                                              default_resize_factor,&
                                              has_acc
   USE dbcsr_data_methods,              ONLY: dbcsr_data_ensure_size,&
                                              dbcsr_data_get_size
   USE dbcsr_mm_accdrv,                 ONLY: dbcsr_mm_accdrv_barrier,&
                                              dbcsr_mm_accdrv_finalize,&
                                              dbcsr_mm_accdrv_init,&
                                              dbcsr_mm_accdrv_lib_finalize,&
                                              dbcsr_mm_accdrv_lib_init,&
                                              dbcsr_mm_accdrv_phaseout,&
                                              dbcsr_mm_accdrv_process,&
                                              dbcsr_mm_accdrv_type
   USE dbcsr_mm_hostdrv,                ONLY: dbcsr_mm_hostdrv_init,&
                                              dbcsr_mm_hostdrv_lib_finalize,&
                                              dbcsr_mm_hostdrv_lib_init,&
                                              dbcsr_mm_hostdrv_process,&
                                              dbcsr_mm_hostdrv_type
   USE dbcsr_mm_types,                  ONLY: p_a_first,&
                                              p_b_first,&
                                              p_c_first,&
                                              p_k,&
                                              p_m,&
                                              p_n,&
                                              stack_descriptor_type
   USE dbcsr_toollib,                   ONLY: sort
   USE dbcsr_types,                     ONLY: dbcsr_type,&
                                              dbcsr_work_type
   USE kinds,                           ONLY: int_4,&
                                              int_8,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_environ,&
                                              mp_max,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_sched'

   PUBLIC :: dbcsr_mm_sched_type
   PUBLIC :: dbcsr_mm_sched_lib_init, dbcsr_mm_sched_lib_finalize
   PUBLIC :: dbcsr_mm_sched_phaseout
   PUBLIC :: dbcsr_mm_sched_init, dbcsr_mm_sched_finalize
   PUBLIC :: dbcsr_mm_sched_process
   PUBLIC :: dbcsr_mm_sched_begin_burst, dbcsr_mm_sched_end_burst
   PUBLIC :: dbcsr_mm_sched_barrier

   ! **************************************************************************************************
   TYPE dbcsr_mm_sched_type
      PRIVATE
      TYPE(dbcsr_work_type), POINTER  :: product_wm => Null()
      TYPE(dbcsr_mm_accdrv_type)      :: accdrv
      TYPE(dbcsr_mm_hostdrv_type)     :: hostdrv
      LOGICAL                         :: avoid_accdrv = .FALSE.
      LOGICAL                         :: product_wm_cleared = .FALSE.
      INTEGER                         :: product_wm_orig_datasize = -1
   END TYPE dbcsr_mm_sched_type

   ! **************************************************************************************************
   TYPE stats_type
      INTEGER(kind=int_8)                              :: cpu_num_stacks = 0
      INTEGER(kind=int_8)                              :: smm_num_stacks = 0
      INTEGER(kind=int_8)                              :: acc_num_stacks = 0
      INTEGER(kind=int_8)                              :: cpu_flop = 0
      INTEGER(kind=int_8)                              :: smm_flop = 0
      INTEGER(kind=int_8)                              :: acc_flop = 0
      INTEGER(kind=int_8)                              :: max_cpu_flop = 0
      INTEGER(kind=int_8)                              :: max_smm_flop = 0
      INTEGER(kind=int_8)                              :: max_acc_flop = 0
      INTEGER(kind=int_8), DIMENSION(:, :), ALLOCATABLE :: num_mnk_stacks
      ! ensure that array-elements are on different cache lines
      INTEGER(kind=int_4), DIMENSION(64)               :: padding
   END TYPE stats_type

! **************************************************************************************************
!>  Statistics Counter
! **************************************************************************************************
   ! Counters for each thread to collect statistics
   TYPE(stats_type), DIMENSION(:), ALLOCATABLE, TARGET, SAVE :: stats_per_thread

CONTAINS

! **************************************************************************************************
!> \brief Initialize a stats_type
!> \param stats ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stats_init(stats)
      TYPE(stats_type), INTENT(INOUT)                    :: stats

      ALLOCATE (stats%num_mnk_stacks(1, 9))
      stats%num_mnk_stacks(1, :) = 0 ! entry for the default stack
   END SUBROUTINE stats_init

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_lib_init()

      INTEGER                                            :: ithread, nthreads

      nthreads = 1; ithread = 0
!$    nthreads = OMP_GET_NUM_THREADS(); ithread = OMP_GET_THREAD_NUM()

!$OMP     MASTER
      ALLOCATE (stats_per_thread(0:nthreads-1))
!$OMP     END MASTER

!$OMP     BARRIER

      CALL stats_init(stats_per_thread(ithread))
      CALL dbcsr_mm_accdrv_lib_init()
      CALL dbcsr_mm_hostdrv_lib_init()

   END SUBROUTINE dbcsr_mm_sched_lib_init

! **************************************************************************************************
!> \brief Finalize the library and prints DBCSR statistics
!> \param group ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_lib_finalize(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      TYPE(stats_type)                                   :: report

      CALL dbcsr_mm_accdrv_lib_finalize()
      CALL dbcsr_mm_hostdrv_lib_finalize()

      ! Collect and output statistics ---------------------------------------------

!$OMP    MASTER
      CALL stats_init(report)
      CALL stats_collect_from_threads(report)
      CALL stats_collect_from_ranks(report, group)
      CALL stats_print_report(report, output_unit)
      DEALLOCATE (stats_per_thread)
!$OMP    END MASTER

   END SUBROUTINE dbcsr_mm_sched_lib_finalize

! **************************************************************************************************
!> \brief Makes sure that the product_wm is cleared.
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE ensure_product_wm_cleared(this)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this

      INTEGER                                            :: allocated_datasize, used_datasize

      IF (this%product_wm_cleared) RETURN

      ! The product's data_area could already contain some data.
      ! ( see: keep_product_data in dbcsr_operations.F )
      ! But this data might not occupy all the allocated memory in the data_area.
      ! Since, we don't want to keep track of unitialized memory we just zero it now.

      used_datasize = this%product_wm_orig_datasize
      allocated_datasize = dbcsr_data_get_size(this%product_wm%data_area)
      CALL dbcsr_data_clear(this%product_wm%data_area, lb=used_datasize+1, ub=allocated_datasize)
      this%product_wm_cleared = .TRUE.
   END SUBROUTINE ensure_product_wm_cleared

! **************************************************************************************************
!> \brief Initializes a multiplication cycle for new set of C-blocks.
!> \param this ...
!> \param product_wm ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_init(this, product_wm)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this
      TYPE(dbcsr_work_type), POINTER                     :: product_wm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_sched_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      this%product_wm => product_wm

      ! Clearing the product_wm takes too long, we gonna do it later and
      ! return now to allow for MPI to progress.
      ! We just have to remeber its datasize, in case it already contains data.
      this%product_wm_orig_datasize = this%product_wm%datasize

      CALL dbcsr_mm_hostdrv_init(this%hostdrv, product_wm)

      IF (has_acc) &
         CALL dbcsr_mm_accdrv_init(this%accdrv, product_wm)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_mm_sched_init

! **************************************************************************************************
!> \brief Signal approaching end of multiplication
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_phaseout(this)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this

!CALL dbcsr_mm_hostdrv_phaseout(this%hostdrv) ! not needed

      IF (has_acc) &
         CALL dbcsr_mm_accdrv_phaseout(this%accdrv)
   END SUBROUTINE dbcsr_mm_sched_phaseout

! **************************************************************************************************
!> \brief Finalizes a multiplication cycle for a set of C-blocks.
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_finalize(this)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_sched_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ! Just in case dbcsr_mm_sched_process was never called (really needed?)
      CALL ensure_product_wm_cleared(this)

      !CALL dbcsr_mm_hostdrv_finalize(this%hostdrv) ! not needed
      IF (has_acc) &
         CALL dbcsr_mm_accdrv_finalize(this%accdrv)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_mm_sched_finalize

! **************************************************************************************************
!> \brief Signal begin of a burst of calls to dbcsr_mm_sched_process.
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_begin_burst(this)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this

      this%avoid_accdrv = .FALSE.
   END SUBROUTINE dbcsr_mm_sched_begin_burst

! **************************************************************************************************
!> \brief Signal end of a burst of calls to dbcsr_mm_sched_process.
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_end_burst()

!nothing to do here

   END SUBROUTINE dbcsr_mm_sched_end_burst

! **************************************************************************************************
!> \brief Signal that previous stacks should be processed first
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_barrier()

!CALL dbcsr_mm_hostdrv_barrier(this%hostdrv) ! not needed

      IF (has_acc) &
         CALL dbcsr_mm_accdrv_barrier()

   END SUBROUTINE dbcsr_mm_sched_barrier

! **************************************************************************************************
!> \brief Processes a given stack.
!> \param this ...
!> \param left ...
!> \param right ...
!> \param stack_data ...
!> \param stack_fillcount ...
!> \param stack_descr ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_sched_process(this, left, right, stack_data, &
                                     stack_fillcount, stack_descr)
      TYPE(dbcsr_mm_sched_type), INTENT(INOUT)           :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, DIMENSION(:, :), POINTER                  :: stack_data
      INTEGER, POINTER                                   :: stack_fillcount
      TYPE(stack_descriptor_type), INTENT(IN)            :: stack_descr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mm_sched_process', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ithread, sp, stacked_datasize
      INTEGER(kind=int_8)                                :: flop_per_entry, total_flop
      LOGICAL                                            :: success, used_smm
      TYPE(stats_type), POINTER                          :: mystats

      IF (stack_fillcount <= 0) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",329,"dbcsr_mm_sched_process: got empty stack")

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      mystats => stats_per_thread(ithread)

      CALL ensure_product_wm_cleared(this)

      stacked_datasize = this%product_wm%datasize
      CALL dbcsr_data_ensure_size(this%product_wm%data_area, stacked_datasize, &
                                  factor=default_resize_factor, zero_pad=.TRUE.)

      !!From here on there is no boundary checking due to assumed-SIZE-arguments.
      !!This is usefull to check stack parameters, BUT it works only for kind=dp
      IF (.FALSE.) THEN
         DO sp = 1, stack_fillcount
            IF (stack_data(p_a_first, sp) > SIZE(left%data_area%d%r_dp)) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",346,"left data out of range")
            IF (stack_data(p_b_first, sp) > SIZE(right%data_area%d%r_dp)) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",348,"right data out of range")
            IF (stack_data(p_c_first, sp) > SIZE(this%product_wm%data_area%d%r_dp)) THEN
               WRITE (*, *) "blub: ", stack_data(p_c_first, sp), SIZE(this%product_wm%data_area%d%r_dp), &
                  dbcsr_data_get_size(this%product_wm%data_area), stacked_datasize
               CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",352,"product data out of range")
            END IF
         END DO
      ENDIF

      IF (.FALSE.) THEN
         ! Check if homogenous stacks are indeed homogenous
         IF (stack_descr%defined_mnk) THEN
            DO sp = 1, stack_fillcount
               IF (stack_data(p_m, sp) /= stack_descr%m) &
                  CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",362,"homogenous stacks check failed")
               IF (stack_data(p_n, sp) /= stack_descr%n) &
                  CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",364,"homogenous stacks check failed")
               IF (stack_data(p_k, sp) /= stack_descr%k) &
                  CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",366,"homogenous stacks check failed")
            END DO
         ENDIF
      ENDIF

      ! Submitting the stack for processing -------------------------------------
      flop_per_entry = 2*stack_descr%max_m*stack_descr%max_n*stack_descr%max_k
      total_flop = stack_fillcount*flop_per_entry

      IF (has_acc .AND. &
          flop_per_entry > accdrv_min_flop_process .AND. &
          (.NOT. this%avoid_accdrv) .AND. &
          (stack_descr%defined_mnk .OR. accdrv_do_inhomogenous)) THEN
         CALL dbcsr_mm_accdrv_process( &
            this%accdrv, &
            left, right, &
            params=stack_data, &
            stack_size=stack_fillcount, &
            stack_descr=stack_descr, &
            success=success)

         IF (success) THEN
            ! update statistics
            mystats%acc_num_stacks = mystats%acc_num_stacks+1
            mystats%acc_flop = mystats%acc_flop+total_flop
            CALL stats_add(mystats, &
                           m=stack_descr%m, n=stack_descr%n, k=stack_descr%k, &
                           stacksize_acc=INT(stack_fillcount, kind=int_8))
            RETURN
         ELSE
            this%avoid_accdrv = accdrv_avoid_after_busy
         ENDIF
      ENDIF

      !WRITE (*,*) "dbcsr_mm_sched_process: running hostdrv_process, stack_fillcount:", stack_fillcount

      CALL dbcsr_mm_hostdrv_process( &
         this%hostdrv, &
         left, right, &
         params=stack_data, &
         stack_size=stack_fillcount, &
         stack_descr=stack_descr, &
         success=success, &
         used_smm=used_smm)

      IF (.NOT. success) CALL cp__b("dbcsr/mm/dbcsr_mm_sched.F",411,"dbcsr_mm_sched_process_stack failed")

      ! update statistics
      IF (used_smm) THEN
         mystats%smm_num_stacks = mystats%smm_num_stacks+1
         mystats%smm_flop = mystats%smm_flop+total_flop
         CALL stats_add(mystats, &
                        m=stack_descr%m, n=stack_descr%n, k=stack_descr%k, &
                        stacksize_smm=INT(stack_fillcount, kind=int_8))
      ELSE
         mystats%cpu_num_stacks = mystats%cpu_num_stacks+1
         mystats%cpu_flop = mystats%cpu_flop+total_flop
         CALL stats_add(mystats, &
                        m=stack_descr%m, n=stack_descr%n, k=stack_descr%k, &
                        stacksize_cpu=INT(stack_fillcount, kind=int_8))
      ENDIF

   END SUBROUTINE dbcsr_mm_sched_process

! ***************************************************************************************************
!> \brief Helper-routine used by dbcsr_mm_sched_process to supply statistics.
!> \param stats ...
!> \param m ...
!> \param n ...
!> \param k ...
!> \param stacksize_cpu ...
!> \param stacksize_smm ...
!> \param stacksize_acc ...
!> \param nstacks_cpu ...
!> \param nstacks_smm ...
!> \param nstacks_acc ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stats_add(stats, m, n, k, stacksize_cpu, stacksize_smm, stacksize_acc, &
                        nstacks_cpu, nstacks_smm, nstacks_acc)
      TYPE(stats_type), INTENT(INOUT)                    :: stats
      INTEGER, INTENT(IN)                                :: m, n, k
      INTEGER(kind=int_8), OPTIONAL                      :: stacksize_cpu, stacksize_smm, &
                                                            stacksize_acc, nstacks_cpu, &
                                                            nstacks_smm, nstacks_acc

      INTEGER                                            :: i, s
      INTEGER(kind=int_8)                                :: my_nstacks_acc, my_nstacks_cpu, &
                                                            my_nstacks_smm, my_stacksize_acc, &
                                                            my_stacksize_cpu, my_stacksize_smm
      INTEGER(kind=int_8), ALLOCATABLE, DIMENSION(:, :)  :: tmp

      my_stacksize_cpu = 0
      my_stacksize_smm = 0
      my_stacksize_acc = 0
      IF (PRESENT(stacksize_cpu)) my_stacksize_cpu = stacksize_cpu
      IF (PRESENT(stacksize_smm)) my_stacksize_smm = stacksize_smm
      IF (PRESENT(stacksize_acc)) my_stacksize_acc = stacksize_acc

      my_nstacks_cpu = MERGE(1, 0, my_stacksize_cpu > 0)
      my_nstacks_smm = MERGE(1, 0, my_stacksize_smm > 0)
      my_nstacks_acc = MERGE(1, 0, my_stacksize_acc > 0)
      IF (PRESENT(nstacks_cpu)) my_nstacks_cpu = nstacks_cpu
      IF (PRESENT(nstacks_smm)) my_nstacks_smm = nstacks_smm
      IF (PRESENT(nstacks_acc)) my_nstacks_acc = nstacks_acc

      DO i = 1, SIZE(stats%num_mnk_stacks, 1)
         IF (stats%num_mnk_stacks(i, 1) == m .AND. &
             stats%num_mnk_stacks(i, 2) == n .AND. &
             stats%num_mnk_stacks(i, 3) == k) THEN
            stats%num_mnk_stacks(i, 4) = stats%num_mnk_stacks(i, 4)+my_stacksize_cpu
            stats%num_mnk_stacks(i, 5) = stats%num_mnk_stacks(i, 5)+my_stacksize_smm
            stats%num_mnk_stacks(i, 6) = stats%num_mnk_stacks(i, 6)+my_stacksize_acc
            stats%num_mnk_stacks(i, 7) = stats%num_mnk_stacks(i, 7)+my_nstacks_cpu
            stats%num_mnk_stacks(i, 8) = stats%num_mnk_stacks(i, 8)+my_nstacks_smm
            stats%num_mnk_stacks(i, 9) = stats%num_mnk_stacks(i, 9)+my_nstacks_acc
            RETURN
         ENDIF
      END DO

      !not found, ok lets grow the list
      s = SIZE(stats%num_mnk_stacks, 1)
      ALLOCATE (tmp(s, 9))
      tmp(:, :) = stats%num_mnk_stacks(:, :)
      DEALLOCATE (stats%num_mnk_stacks)
      ALLOCATE (stats%num_mnk_stacks(s+1, 9))
      stats%num_mnk_stacks(1:s, :) = tmp(:, :)
      stats%num_mnk_stacks(s+1, 1) = m
      stats%num_mnk_stacks(s+1, 2) = n
      stats%num_mnk_stacks(s+1, 3) = k
      stats%num_mnk_stacks(s+1, 4) = my_stacksize_cpu
      stats%num_mnk_stacks(s+1, 5) = my_stacksize_smm
      stats%num_mnk_stacks(s+1, 6) = my_stacksize_acc
      stats%num_mnk_stacks(s+1, 7) = my_nstacks_cpu
      stats%num_mnk_stacks(s+1, 8) = my_nstacks_smm
      stats%num_mnk_stacks(s+1, 9) = my_nstacks_acc
      DEALLOCATE (tmp)
   END SUBROUTINE stats_add

! **************************************************************************************************
!> \brief Collects statistics from all OpenMP-threads into report
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stats_collect_from_threads(report)
      TYPE(stats_type), INTENT(INOUT)                    :: report

      INTEGER                                            :: i, j, nthreads
      TYPE(stats_type), POINTER                          :: istats

!$OMP     MASTER

      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()

      DO i = 0, nthreads-1
         istats => stats_per_thread(i)
         report%cpu_num_stacks = report%cpu_num_stacks+istats%cpu_num_stacks
         report%smm_num_stacks = report%smm_num_stacks+istats%smm_num_stacks
         report%acc_num_stacks = report%acc_num_stacks+istats%acc_num_stacks
         report%acc_flop = report%acc_flop+istats%acc_flop
         report%smm_flop = report%smm_flop+istats%smm_flop
         report%cpu_flop = report%cpu_flop+istats%cpu_flop

         DO j = 1, SIZE(istats%num_mnk_stacks, 1)
            CALL stats_add(report, &
                           m=INT(istats%num_mnk_stacks(j, 1), kind=int_4), &
                           n=INT(istats%num_mnk_stacks(j, 2), kind=int_4), &
                           k=INT(istats%num_mnk_stacks(j, 3), kind=int_4), &
                           stacksize_cpu=istats%num_mnk_stacks(j, 4), &
                           stacksize_smm=istats%num_mnk_stacks(j, 5), &
                           stacksize_acc=istats%num_mnk_stacks(j, 6), &
                           nstacks_cpu=istats%num_mnk_stacks(j, 7), &
                           nstacks_smm=istats%num_mnk_stacks(j, 8), &
                           nstacks_acc=istats%num_mnk_stacks(j, 9))
         END DO
      END DO
!$OMP     END MASTER
   END SUBROUTINE stats_collect_from_threads

! **************************************************************************************************
!> \brief Collects statistics from all MPI-ranks
!> \param report ...
!> \param group ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stats_collect_from_ranks(report, group)
      TYPE(stats_type), INTENT(INOUT)                    :: report
      INTEGER, INTENT(IN)                                :: group

      INTEGER                                            :: i, myrank, nranks, sending_rank
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: mnk_collected
      INTEGER, DIMENSION(3)                              :: mnk

!$OMP     MASTER

      CALL mp_environ(nranks, myrank, group)

      report%max_acc_flop = report%acc_flop
      CALL mp_max(report%max_acc_flop, group)
      report%max_smm_flop = report%smm_flop
      CALL mp_max(report%max_smm_flop, group)
      report%max_cpu_flop = report%cpu_flop
      CALL mp_max(report%max_cpu_flop, group)

      CALL mp_sum(report%acc_flop, group)
      CALL mp_sum(report%smm_flop, group)
      CALL mp_sum(report%cpu_flop, group)
      CALL mp_sum(report%cpu_num_stacks, group)
      CALL mp_sum(report%smm_num_stacks, group)
      CALL mp_sum(report%acc_num_stacks, group)

      ! array mnk_collected is used as a logical-array, allows to use minloc
      ALLOCATE (mnk_collected(SIZE(report%num_mnk_stacks, 1)))
      mnk_collected = 0 ! init all to false

      ! broadcast stats of all mnk-combinations, which occured on any mpi rank
      DO
         ! each rank with uncollected stats tries to become the sending_rank
         sending_rank = -1
         IF (.NOT. ALL(mnk_collected == 1)) sending_rank = myrank
         CALL mp_max(sending_rank, group)
         IF (sending_rank < 0) EXIT ! every rank got all mnk collected

         IF (sending_rank == myrank) THEN
            i = MINLOC(mnk_collected, dim=1)
            mnk = INT(report%num_mnk_stacks(i, 1:3), kind=int_4)
         ENDIF
         CALL mp_bcast(msg=mnk, source=sending_rank, gid=group)

         CALL stats_add(report, m=mnk(1), n=mnk(2), k=mnk(3), stacksize_cpu=0_int_8, stacksize_acc=0_int_8)
         DO i = 1, SIZE(report%num_mnk_stacks, 1)
            IF (ALL(report%num_mnk_stacks(i, 1:3) == mnk)) THEN
               IF (i <= SIZE(mnk_collected)) mnk_collected(i) = 1
               CALL mp_sum(report%num_mnk_stacks(i, 4:9), group)
            END IF
         END DO
      END DO
!$OMP     END MASTER
   END SUBROUTINE stats_collect_from_ranks

! **************************************************************************************************
!> \brief Prints collected statistics
!> \param report ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE stats_print_report(report, output_unit)
      TYPE(stats_type), INTENT(INOUT)                    :: report
      INTEGER, INTENT(IN)                                :: output_unit

      INTEGER                                            :: i, j
      INTEGER(KIND=int_8)                                :: total
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: sort_key
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sort_idx

      IF (output_unit <= 0) RETURN

      WRITE (output_unit, "(1X,A,T45,A,T57,A,T68,A,T78,A)") "COUNTER", "TOTAL", "BLAS", "SMM", "ACC"

      !sorting stat entries by flops per multiplication
      ALLOCATE (sort_key(SIZE(report%num_mnk_stacks, 1)-1))
      sort_key(:) = 2*PRODUCT(report%num_mnk_stacks(2:, 1:3), DIM=2)*SUM(report%num_mnk_stacks(2:, 4:6), DIM=2)
      ALLOCATE (sort_idx(SIZE(sort_key)))
      CALL sort(sort_key, SIZE(sort_key), sort_idx)

      DO i = 1, SIZE(sort_idx)
         j = sort_idx(i)+1
         total = SUM(report%num_mnk_stacks(j, 4:6))
         WRITE (output_unit, "(A,I4,' x ',I4,' x ',I4,T30,I20,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
            " flops ", report%num_mnk_stacks(j, 1:3), &
            2*total*PRODUCT(report%num_mnk_stacks(j, 1:3)), &
            100*REAL(report%num_mnk_stacks(j, 4:6))/REAL(MAX(INT(1, KIND=int_8), total))
      END DO

      total = report%cpu_flop+report%smm_flop+report%acc_flop
      WRITE (output_unit, "(A,T30,EN20.6,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
         " flops total", REAL(total, KIND=real_8), &
         100*REAL(report%cpu_flop)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%smm_flop)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%acc_flop)/REAL(MAX(INT(1, KIND=int_8), total))

      total = report%max_cpu_flop+report%max_smm_flop+report%max_acc_flop
      WRITE (output_unit, "(A,T30,EN20.6,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
         " flops max/rank", REAL(total, KIND=real_8), &
         100*REAL(report%max_cpu_flop)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%max_smm_flop)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%max_acc_flop)/REAL(MAX(INT(1, KIND=int_8), total))

      total = SUM(report%num_mnk_stacks(1, 4:6))
      WRITE (output_unit, "(A,T30,I20,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
         " matmuls inhomo. stacks", total, &
         100*REAL(report%num_mnk_stacks(1, 4:6))/REAL(MAX(INT(1, KIND=int_8), total))

      total = SUM(report%num_mnk_stacks(:, 4:6))
      WRITE (output_unit, "(A,T30,I20,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
         " matmuls total", total, &
         100*REAL(SUM(report%num_mnk_stacks(:, 4:6), DIM=1))/REAL(MAX(INT(1, KIND=int_8), total))

      total = report%cpu_num_stacks+report%smm_num_stacks+report%acc_num_stacks
      WRITE (output_unit, "(A,T30,I20,5X,F5.1,'%',4X,F5.1,'%',4X,F5.1,'%')") &
         " number of processed stacks", total, &
         100*REAL(report%cpu_num_stacks)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%smm_num_stacks)/REAL(MAX(INT(1, KIND=int_8), total)), &
         100*REAL(report%acc_num_stacks)/REAL(MAX(INT(1, KIND=int_8), total))

      WRITE (output_unit, '(A,T51,F9.1,1X,F9.1,1X,F9.1)') " average stack size", &
         REAL(SUM(report%num_mnk_stacks(:, 4)))/REAL(MAX(INT(1, KIND=int_8), SUM(report%num_mnk_stacks(:, 7)))), &
         REAL(SUM(report%num_mnk_stacks(:, 5)))/REAL(MAX(INT(1, KIND=int_8), SUM(report%num_mnk_stacks(:, 8)))), &
         REAL(SUM(report%num_mnk_stacks(:, 6)))/REAL(MAX(INT(1, KIND=int_8), SUM(report%num_mnk_stacks(:, 9))))

   END SUBROUTINE stats_print_report

END MODULE dbcsr_mm_sched

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Third layer of the dbcsr matrix-matrix multiplication.
!>         It collects the full matrix blocks, which need to be multiplied,
!>         and stores their parameters in various stacks.
!>         After a certain amount of parameters is collected it dispatches
!>         the filled stacks to either the CPU or the accelerator device.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2010-02-23 Moved from dbcsr_operations
!>  - 2011-11    Moved parameter-stack processing routines to
!>               dbcsr_mm_methods.
!>  - 2013-01    extensive refactoring (Ole Schuett)
! **************************************************************************************************
MODULE dbcsr_mm_csr

   USE array_types,                     ONLY: array_data
   USE dbcsr_config,                    ONLY: dbcsr_get_conf_nstacks,&
                                              default_resize_factor,&
                                              mm_stack_size
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_mm_sched,                  ONLY: &
        dbcsr_mm_sched_barrier, dbcsr_mm_sched_begin_burst, dbcsr_mm_sched_end_burst, &
        dbcsr_mm_sched_finalize, dbcsr_mm_sched_init, dbcsr_mm_sched_lib_finalize, &
        dbcsr_mm_sched_lib_init, dbcsr_mm_sched_phaseout, dbcsr_mm_sched_process, &
        dbcsr_mm_sched_type
   USE dbcsr_mm_types,                  ONLY: &
        dbcsr_ps_width, p_a_first, p_b_first, p_c_blk, p_c_first, p_k, p_m, p_n, &
        stack_descriptor_type
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size
   USE dbcsr_toollib,                   ONLY: sort
   USE dbcsr_types,                     ONLY: dbcsr_type,&
                                              dbcsr_work_type
   USE dbcsr_util,                      ONLY: map_most_common
   USE kinds,                           ONLY: int_1,&
                                              int_4,&
                                              int_8,&
                                              real_8,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_csr'
   LOGICAL, PARAMETER :: debug_mod = .FALSE.
   LOGICAL, PARAMETER :: careful_mod = .FALSE.

!> \var max_stack_block_size  The maximal block size to be specially
!>                            treated.
   INTEGER, PARAMETER :: max_stack_block_size = HUGE(INT(0))

! **************************************************************************************************
   TYPE dbcsr_mm_csr_type
      PRIVATE
      TYPE(hash_table_type), DIMENSION(:), POINTER  :: c_hashes => Null()
      INTEGER                        :: nm_stacks, nn_stacks, nk_stacks
      INTEGER(KIND=int_4), DIMENSION(:), POINTER :: m_size_maps => Null()
      INTEGER(KIND=int_4), DIMENSION(:), POINTER :: n_size_maps => Null()
      INTEGER(KIND=int_4), DIMENSION(:), POINTER :: k_size_maps => Null()
      INTEGER                        :: max_m, max_n, max_k
      INTEGER                        :: m_size_maps_size, &
                                        n_size_maps_size, &
                                        k_size_maps_size
      INTEGER(KIND=int_1), DIMENSION(:, :, :), POINTER :: stack_map => Null()
      TYPE(stack_descriptor_type), DIMENSION(:), POINTER  :: stacks_descr => Null()
      TYPE(dbcsr_work_type), POINTER           :: product_wm => Null()
      INTEGER, DIMENSION(:, :, :), POINTER       :: stacks_data => Null()
      INTEGER, DIMENSION(:), POINTER           :: stacks_fillcount => Null()
      TYPE(dbcsr_mm_sched_type)                      :: sched
   END TYPE dbcsr_mm_csr_type

! *****************************************************************************
!> \brief Types needed for the hashtable.
! *****************************************************************************
  TYPE ele_type
     INTEGER :: c=0
     INTEGER :: p=0
  END TYPE ele_type

  TYPE hash_table_type
     TYPE(ele_type), DIMENSION(:), POINTER :: table
     INTEGER :: nele=0
     INTEGER :: nmax=0
     INTEGER :: prime=0
  END TYPE hash_table_type

! **************************************************************************************************
   PUBLIC :: dbcsr_mm_csr_type
   PUBLIC :: dbcsr_mm_csr_lib_init, dbcsr_mm_csr_lib_finalize
   PUBLIC :: dbcsr_mm_csr_phaseout
   PUBLIC :: dbcsr_mm_csr_init, dbcsr_mm_csr_finalize
   PUBLIC :: dbcsr_mm_csr_multiply, dbcsr_mm_csr_purge_stacks

CONTAINS

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_lib_init()

      CALL dbcsr_mm_sched_lib_init()
   END SUBROUTINE

! **************************************************************************************************
!> \brief Finalize the library
!> \param group ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_lib_finalize(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      CALL dbcsr_mm_sched_lib_finalize(group, output_unit)
   END SUBROUTINE

! **************************************************************************************************
!> \brief A wrapper around dbcsr_mm_csr_multiply_low to avoid expensive dereferencings.
!> \param this ...
!> \param left ...
!> \param right ...
!> \param mi ...
!> \param mf ...
!> \param ni ...
!> \param nf ...
!> \param ki ...
!> \param kf ...
!> \param ai ...
!> \param af ...
!> \param bi ...
!> \param bf ...
!> \param m_sizes ...
!> \param n_sizes ...
!> \param k_sizes ...
!> \param c_local_rows ...
!> \param c_local_cols ...
!> \param c_has_symmetry ...
!> \param keep_sparsity ...
!> \param use_eps ...
!> \param row_max_epss ...
!> \param flop ...
!> \param a_index ...
!> \param b_index ...
!> \param a_norms ...
!> \param b_norms ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_multiply(this, left, right, mi, mf, ni, nf, ki, kf, &
                                    ai, af, &
                                    bi, bf, &
                                    m_sizes, n_sizes, k_sizes, &
                                    c_local_rows, c_local_cols, &
                                    c_has_symmetry, keep_sparsity, use_eps, &
                                    row_max_epss, &
                                    flop, &
                                    a_index, b_index, a_norms, b_norms)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, INTENT(IN)                                :: mi, mf, ni, nf, ki, kf, ai, af, bi, bf
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: m_sizes, n_sizes, k_sizes, c_local_rows, &
                                                            c_local_cols
      LOGICAL, INTENT(INOUT)                             :: c_has_symmetry, keep_sparsity, use_eps
      REAL(kind=sp), DIMENSION(:)                        :: row_max_epss
      INTEGER(KIND=int_8), INTENT(INOUT)                 :: flop
      INTEGER, DIMENSION(1:3, 1:af), INTENT(IN)          :: a_index
      INTEGER, DIMENSION(1:3, 1:bf), INTENT(IN)          :: b_index
      REAL(KIND=sp), DIMENSION(:), POINTER               :: a_norms, b_norms

      INTEGER                                            :: ithread, max_new_nblks, nblks_new

      ithread = 0
!$    ithread = omp_get_thread_num()

      ! This has to be done here because ensure_array_size() expects a pointer.
      ! the maximum number of blocks can be safely estimated by considering both the rowxcol,
      ! but also the blocks the latter can never be larger than norec**2, which is a 'small' constant
      max_new_nblks = INT(MIN(INT(mf-mi+1, int_8)*INT(nf-ni+1, int_8), &
                              INT(af-ai+1, int_8)*INT(bf-bi+1, int_8)))

      nblks_new = this%product_wm%lastblk+max_new_nblks

      CALL ensure_array_size(this%product_wm%row_i, ub=nblks_new, &
                             factor=default_resize_factor)
      CALL ensure_array_size(this%product_wm%col_i, ub=nblks_new, &
                             factor=default_resize_factor)
      CALL ensure_array_size(this%product_wm%blk_p, ub=nblks_new, &
                             factor=default_resize_factor)

      CALL dbcsr_mm_csr_multiply_low(this, left=left, right=right, &
                                     mi=mi, mf=mf, ki=ki, kf=kf, &
                                     ai=ai, af=af, &
                                     bi=bi, bf=bf, &
                                     c_row_i=this%product_wm%row_i, &
                                     c_col_i=this%product_wm%col_i, &
                                     c_blk_p=this%product_wm%blk_p, &
                                     lastblk=this%product_wm%lastblk, &
                                     datasize=this%product_wm%datasize, &
                                     m_sizes=m_sizes, n_sizes=n_sizes, k_sizes=k_sizes, &
                                     c_local_rows=c_local_rows, c_local_cols=c_local_cols, &
                                     c_has_symmetry=c_has_symmetry, keep_sparsity=keep_sparsity, &
                                     use_eps=use_eps, &
                                     row_max_epss=row_max_epss, &
                                     flop=flop, &
                                     row_size_maps=this%m_size_maps, &
                                     col_size_maps=this%n_size_maps, &
                                     k_size_maps=this%k_size_maps, &
                                     row_size_maps_size=this%m_size_maps_size, &
                                     col_size_maps_size=this%n_size_maps_size, &
                                     k_size_maps_size=this%k_size_maps_size, &
                                     nm_stacks=this%nm_stacks, nn_stacks=this%nn_stacks, &
                                     nk_stacks=this%nk_stacks, &
                                     stack_map=this%stack_map, &
                                     stacks_data=this%stacks_data, &
                                     stacks_fillcount=this%stacks_fillcount, &
                                     c_hashes=this%c_hashes, &
                                     a_index=a_index, b_index=b_index, &
                                     a_norms=a_norms, b_norms=b_norms)

   END SUBROUTINE dbcsr_mm_csr_multiply

! **************************************************************************************************
!> \brief Performs multiplication of smaller submatrices.
!> \param this ...
!> \param left ...
!> \param right ...
!> \param mi ...
!> \param mf ...
!> \param ki ...
!> \param kf ...
!> \param ai ...
!> \param af ...
!> \param bi ...
!> \param bf ...
!> \param c_row_i ...
!> \param c_col_i ...
!> \param c_blk_p ...
!> \param lastblk ...
!> \param datasize ...
!> \param m_sizes ...
!> \param n_sizes ...
!> \param k_sizes ...
!> \param c_local_rows ...
!> \param c_local_cols ...
!> \param c_has_symmetry ...
!> \param keep_sparsity ...
!> \param use_eps ...
!> \param row_max_epss ...
!> \param flop ...
!> \param row_size_maps ...
!> \param col_size_maps ...
!> \param k_size_maps ...
!> \param row_size_maps_size ...
!> \param col_size_maps_size ...
!> \param k_size_maps_size ...
!> \param nm_stacks ...
!> \param nn_stacks ...
!> \param nk_stacks ...
!> \param stack_map ...
!> \param stacks_data ...
!> \param stacks_fillcount ...
!> \param c_hashes ...
!> \param a_index ...
!> \param b_index ...
!> \param a_norms ...
!> \param b_norms ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_multiply_low(this, left, right, mi, mf, ki, kf, &
                                        ai, af, bi, bf, &
                                        c_row_i, c_col_i, c_blk_p, lastblk, datasize, &
                                        m_sizes, n_sizes, k_sizes, &
                                        c_local_rows, c_local_cols, &
                                        c_has_symmetry, keep_sparsity, use_eps, &
                                        row_max_epss, flop, &
                                        row_size_maps, col_size_maps, k_size_maps, &
                                        row_size_maps_size, col_size_maps_size, k_size_maps_size, &
                                        nm_stacks, nn_stacks, nk_stacks, stack_map, &
                                        stacks_data, stacks_fillcount, c_hashes, &
                                        a_index, b_index, a_norms, b_norms)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, INTENT(IN)                                :: mi, mf, ki, kf, ai, af, bi, bf
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: c_row_i, c_col_i, c_blk_p
      INTEGER, INTENT(INOUT)                             :: lastblk, datasize
      INTEGER, DIMENSION(:), INTENT(IN)                  :: m_sizes, n_sizes, k_sizes, c_local_rows, &
                                                            c_local_cols
      LOGICAL, INTENT(IN)                                :: c_has_symmetry, keep_sparsity, use_eps
      REAL(kind=sp), DIMENSION(:)                        :: row_max_epss
      INTEGER(KIND=int_8), INTENT(INOUT)                 :: flop
      INTEGER, INTENT(IN)                                :: row_size_maps_size, k_size_maps_size, &
                                                            col_size_maps_size
      INTEGER(KIND=int_4), &
         DIMENSION(0:row_size_maps_size-1), INTENT(IN)   :: row_size_maps
      INTEGER(KIND=int_4), &
         DIMENSION(0:col_size_maps_size-1), INTENT(IN)   :: col_size_maps
      INTEGER(KIND=int_4), &
         DIMENSION(0:k_size_maps_size-1), INTENT(IN)     :: k_size_maps
      INTEGER, INTENT(IN)                                :: nm_stacks, nn_stacks, nk_stacks
      INTEGER(KIND=int_1), DIMENSION(nn_stacks+1, &
         nk_stacks+1, nm_stacks+1), INTENT(IN)           :: stack_map
      INTEGER, DIMENSION(:, :, :), INTENT(INOUT)         :: stacks_data
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: stacks_fillcount
      TYPE(hash_table_type), DIMENSION(:), INTENT(INOUT) :: c_hashes
      INTEGER, DIMENSION(1:3, 1:af), INTENT(IN)          :: a_index
      INTEGER, DIMENSION(1:3, 1:bf), INTENT(IN)          :: b_index
      REAL(KIND=sp), DIMENSION(:), POINTER               :: a_norms, b_norms

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_csr_multiply_low', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER :: a_blk, a_col_l, a_row_l, b_blk, b_col_l, c_blk_id, c_col_logical, c_nze, &
         c_row_logical, ithread, k_size, m_size, mapped_col_size, mapped_k_size, mapped_row_size, &
         n_a_norms, n_b_norms, n_size, nstacks, s_dp, ws
      INTEGER, DIMENSION(mi:mf+1)                        :: a_row_p
      INTEGER, DIMENSION(ki:kf+1)                        :: b_row_p
      INTEGER, DIMENSION(2, bf-bi+1)                     :: b_blk_info
      INTEGER, DIMENSION(2, af-ai+1)                     :: a_blk_info
      INTEGER(KIND=int_4)                                :: offset
      LOGICAL                                            :: block_exists
      REAL(kind=sp)                                      :: a_norm, a_row_eps, b_norm
      REAL(KIND=sp), DIMENSION(1:af-ai+1)                :: left_norms
      REAL(KIND=sp), DIMENSION(1:bf-bi+1)                :: right_norms

!   ---------------------------------------------------------------------------

      ithread = 0
!$    ithread = omp_get_thread_num()

      nstacks = SIZE(this%stacks_data, 3)

      IF (use_eps) THEN
         n_a_norms = af-ai+1
         n_b_norms = bf-bi+1
      ELSE
         n_a_norms = 0
         n_b_norms = 0
      ENDIF

      !
      ! Build the indices
      CALL build_csr_index(mi, mf, ai, af, a_row_p, a_blk_info, a_index, &
                           n_a_norms, left_norms, a_norms)
      CALL build_csr_index(ki, kf, bi, bf, b_row_p, b_blk_info, b_index, &
                           n_b_norms, right_norms, b_norms)

      a_row_cycle: DO a_row_l = mi, mf
         m_size = m_sizes(a_row_l)

         a_row_eps = row_max_epss(a_row_l)
         mapped_row_size = row_size_maps(m_size)

         a_blk_cycle: DO a_blk = a_row_p(a_row_l)+1, a_row_p(a_row_l+1)
            a_col_l = a_blk_info(1, a_blk)
            IF (debug_mod) WRITE (*, *) ithread, routineN//" A col", a_col_l, ";", a_row_l
            k_size = k_sizes(a_col_l)
            mapped_k_size = k_size_maps(k_size)

            a_norm = left_norms(a_blk)
            b_blk_cycle: DO b_blk = b_row_p(a_col_l)+1, b_row_p(a_col_l+1)
               IF (dbg) THEN
                  WRITE (*, '(1X,A,3(1X,I7),1X,A,1X,I16)') routineN//" trying B", &
                     a_row_l, b_blk_info(1, b_blk), a_col_l, "at", b_blk_info(2, b_blk)
               ENDIF
               b_norm = right_norms(b_blk)
               IF (a_norm*b_norm .LT. a_row_eps) THEN
                  CYCLE
               ENDIF
               b_col_l = b_blk_info(1, b_blk)
               ! Don't calculate symmetric blocks.
               symmetric_product: IF (c_has_symmetry) THEN
                  c_row_logical = c_local_rows(a_row_l)
                  c_col_logical = c_local_cols(b_col_l)
                  IF (c_row_logical .NE. c_col_logical &
                      .AND. my_checker_tr(c_row_logical, c_col_logical)) THEN
                     IF (dbg) THEN
                        WRITE (*, *) "Skipping symmetric block!", c_row_logical, &
                           c_col_logical
                     ENDIF
                     CYCLE
                  ENDIF
               ENDIF symmetric_product

               c_blk_id = hash_table_get(c_hashes(a_row_l), b_col_l)
               IF (.FALSE.) THEN
                  WRITE (*, '(1X,A,3(1X,I7),1X,A,1X,I16)') routineN//" coor", &
                     a_row_l, a_col_l, b_col_l, "c blk", c_blk_id
               ENDIF
               block_exists = c_blk_id .GT. 0

               n_size = n_sizes(b_col_l)
               c_nze = m_size*n_size
               !
               IF (block_exists) THEN
                  offset = c_blk_p(c_blk_id)
               ELSE
                  IF (keep_sparsity) CYCLE

                  offset = datasize+1
                  lastblk = lastblk+1
                  datasize = datasize+c_nze
                  c_blk_id = lastblk ! assign a new c-block-id

                  IF (dbg) WRITE (*, *) routineN//" new block offset, nze", offset, c_nze
                  CALL hash_table_add(c_hashes(a_row_l), &
                                      b_col_l, c_blk_id)

                  ! We still keep the linear index because it's
                  ! easier than getting the values out of the
                  ! hashtable in the end.
                  c_row_i(lastblk) = a_row_l
                  c_col_i(lastblk) = b_col_l
                  c_blk_p(lastblk) = offset
               ENDIF

               ! TODO: this is only called with careful_mod
               ! We should not call certain MM routines (netlib BLAS)
               ! with zero LDs; however, we still need to get to here
               ! to get new blocks.
               IF (careful_mod) THEN
                  IF (c_nze .EQ. 0 .OR. k_size .EQ. 0) THEN
                     CALL cp__b("dbcsr/mm/dbcsr_mm_csr.F",421,"Can not call MM with LDx=0.")
                     CYCLE
                  ENDIF
               ENDIF

               mapped_col_size = col_size_maps(n_size)
               ws = stack_map(mapped_col_size, mapped_k_size, mapped_row_size)
               stacks_fillcount(ws) = stacks_fillcount(ws)+1
               s_dp = stacks_fillcount(ws)

               stacks_data(p_m, s_dp, ws) = m_size
               stacks_data(p_n, s_dp, ws) = n_size
               stacks_data(p_k, s_dp, ws) = k_size
               stacks_data(p_a_first, s_dp, ws) = a_blk_info(2, a_blk)
               stacks_data(p_b_first, s_dp, ws) = b_blk_info(2, b_blk)
               stacks_data(p_c_first, s_dp, ws) = offset
               stacks_data(p_c_blk, s_dp, ws) = c_blk_id

               flop = flop+INT(2*c_nze, int_8)*INT(k_size, int_8)

               IF (stacks_fillcount(ws) >= SIZE(stacks_data, 2)) &
                  CALL flush_stacks(this, left=left, right=right)

            ENDDO b_blk_cycle ! b
         ENDDO a_blk_cycle ! a_col
      ENDDO a_row_cycle ! a_row

   END SUBROUTINE dbcsr_mm_csr_multiply_low

! **************************************************************************************************
!> \brief Initializes a multiplication cycle for new set of C-blocks.
!> \param this ...
!> \param left ...
!> \param right ...
!> \param product ...
!> \param m_sizes ...
!> \param n_sizes ...
!> \param block_estimate ...
!> \param right_row_blk_size ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_init(this, left, right, product, &
                                m_sizes, n_sizes, block_estimate, right_row_blk_size)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this
      TYPE(dbcsr_type), INTENT(IN), OPTIONAL             :: left, right
      TYPE(dbcsr_type), INTENT(INOUT)                    :: product
      INTEGER, DIMENSION(:), POINTER                     :: m_sizes, n_sizes
      INTEGER, INTENT(IN)                                :: block_estimate
      INTEGER, DIMENSION(:), INTENT(IN)                  :: right_row_blk_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_csr_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: default_stack, handle, istack, ithread, &
                                                            k_map, k_size, m_map, m_size, n_map, &
                                                            n_size, nstacks, nthreads, ps_g
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: flop_index, flop_list, most_common_k, &
                                                            most_common_m, most_common_n
      INTEGER, DIMENSION(3)                              :: nxstacks
      LOGICAL                                            :: local_indexing
      TYPE(stack_descriptor_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: tmp_descr

      CALL timeset(routineN, handle)

      ithread = 0; nthreads = 1
!$    ithread = OMP_GET_THREAD_NUM(); nthreads = OMP_GET_NUM_THREADS()

      IF (PRESENT(left) .NEQV. PRESENT(right)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_csr.F",490,"Must both left and right provided or not.")

      IF (PRESENT(left) .AND. PRESENT(right)) THEN
         ! find out if we have local_indexing
         CALL dbcsr_assert(right%local_indexing, "EQV", left%local_indexing, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Local index useage must be consistent.", 496)
         local_indexing = left%local_indexing
      ELSE
         local_indexing = .TRUE.
      ENDIF
      ! Setup the hash tables if needed
      IF (local_indexing) THEN
         ALLOCATE (this%c_hashes(product%nblkrows_local))
         CALL fill_hash_tables(this%c_hashes, product, block_estimate, &
                               row_map=array_data(product%global_rows), &
                               col_map=array_data(product%global_cols))
      ELSE
         ALLOCATE (this%c_hashes(product%nblkrows_total))
         CALL fill_hash_tables(this%c_hashes, product, block_estimate)
      ENDIF

      ! Setup the MM stack
      CALL dbcsr_get_conf_nstacks(nxstacks)
      this%nm_stacks = nxstacks(1)
      this%nn_stacks = nxstacks(2)
      this%nk_stacks = nxstacks(3)
      nstacks = nxstacks(1)*nxstacks(2)*nxstacks(3)+1
      CALL dbcsr_assert(nstacks, "LE", INT(HUGE(this%stack_map)), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Too many stacks requested (global/dbcsr/n_size_*_stacks in input)", &
                        521)

      ALLOCATE (this%stacks_descr(nstacks))
      ALLOCATE (this%stacks_data(dbcsr_ps_width, mm_stack_size, nstacks))
      ALLOCATE (this%stacks_fillcount(nstacks))
      this%stacks_fillcount(:) = 0

      ALLOCATE (most_common_m(nxstacks(1)))
      ALLOCATE (most_common_n(nxstacks(2)))
      ALLOCATE (most_common_k(nxstacks(3)))
      CALL map_most_common(m_sizes, this%m_size_maps, nxstacks(1), &
                           most_common_m, &
                           max_stack_block_size, this%max_m)
      this%m_size_maps_size = SIZE(this%m_size_maps)
      CALL map_most_common(n_sizes, this%n_size_maps, nxstacks(2), &
                           most_common_n, &
                           max_stack_block_size, this%max_n)
      this%n_size_maps_size = SIZE(this%n_size_maps)
      CALL map_most_common(right_row_blk_size, &
                           this%k_size_maps, nxstacks(3), &
                           most_common_k, &
                           max_stack_block_size, this%max_k)
      this%k_size_maps_size = SIZE(this%k_size_maps)

      ! Creates the stack map--a mapping from (mapped) stack block sizes
      ! (carrier%*_sizes) to a stack number.  Triples with even one
      ! uncommon size will be mapped to a general, non-size-specific
      ! stack.
      ALLOCATE (this%stack_map(nxstacks(2)+1, nxstacks(3)+1, nxstacks(1)+1))
      default_stack = nstacks

      DO m_map = 1, nxstacks(1)+1
         IF (m_map .LE. nxstacks(1)) THEN
            m_size = most_common_m(m_map)
         ELSE
            m_size = 777
         ENDIF
         DO k_map = 1, nxstacks(3)+1
            IF (k_map .LE. nxstacks(3)) THEN
               k_size = most_common_k(k_map)
            ELSE
               k_size = 888
            ENDIF
            DO n_map = 1, nxstacks(2)+1
               IF (n_map .LE. nxstacks(2)) THEN
                  n_size = most_common_n(n_map)
               ELSE
                  n_size = 999
               ENDIF
               IF (m_map .LE. nxstacks(1) &
                   .AND. k_map .LE. nxstacks(3) &
                   .AND. n_map .LE. nxstacks(2)) THEN
                  ! This is the case when m, n, and k are all defined.
                  ps_g = (m_map-1)*nxstacks(2)*nxstacks(3)+ &
                         (k_map-1)*nxstacks(2)+n_map
                  ps_g = nstacks-ps_g
                  this%stack_map(n_map, k_map, m_map) = INT(ps_g, kind=int_1)
                  ! Also take care of the stack m, n, k descriptors
                  this%stacks_descr(ps_g)%m = m_size
                  this%stacks_descr(ps_g)%n = n_size
                  this%stacks_descr(ps_g)%k = k_size
                  this%stacks_descr(ps_g)%max_m = m_size
                  this%stacks_descr(ps_g)%max_n = n_size
                  this%stacks_descr(ps_g)%max_k = k_size
                  this%stacks_descr(ps_g)%defined_mnk = .TRUE.
               ELSE
                  ! This is the case when at least one of m, n, or k is
                  ! undefined.
                  ps_g = default_stack
                  this%stack_map(n_map, k_map, m_map) = INT(default_stack, kind=int_1)
                  ! Also take care of the stack m, n, k descriptors
                  this%stacks_descr(ps_g)%m = 0
                  this%stacks_descr(ps_g)%n = 0
                  this%stacks_descr(ps_g)%k = 0
                  this%stacks_descr(ps_g)%max_m = this%max_m
                  this%stacks_descr(ps_g)%max_n = this%max_n
                  this%stacks_descr(ps_g)%max_k = this%max_k
                  this%stacks_descr(ps_g)%defined_mnk = .FALSE.
               END IF
            ENDDO
         ENDDO
      ENDDO
      DEALLOCATE (most_common_m)
      DEALLOCATE (most_common_n)
      DEALLOCATE (most_common_k)

      ! sort to make the order fixed... all defined stacks first, default stack
      ! last. Next, sort according to flops, first stack lots of flops, last
      ! stack, few flops
      ! The default stack shall remain at the end of the gridcolumn
      ALLOCATE (flop_list(nstacks-1), flop_index(nstacks-1), tmp_descr(nstacks))
      DO istack = 1, nstacks-1
         flop_list(istack) = -2*this%stacks_descr(istack)%m &
                             *this%stacks_descr(istack)%n &
                             *this%stacks_descr(istack)%k
      ENDDO

      CALL sort(flop_list, nstacks-1, flop_index)
      tmp_descr(:) = this%stacks_descr
      DO istack = 1, nstacks-1
         this%stacks_descr(istack) = tmp_descr(flop_index(istack))
      ENDDO

      DO m_map = 1, SIZE(this%stack_map, 1)
         DO k_map = 1, SIZE(this%stack_map, 2)
            map_loop: DO n_map = 1, SIZE(this%stack_map, 1)
               DO istack = 1, nstacks-1
                  IF (this%stack_map(m_map, k_map, n_map) == flop_index(istack)) THEN
                     this%stack_map(m_map, k_map, n_map) = INT(istack, kind=int_1)
                     CYCLE map_loop
                  END IF
               ENDDO
            ENDDO map_loop
         ENDDO
      ENDDO
      DEALLOCATE (flop_list, flop_index, tmp_descr)

      this%product_wm => product%wms(ithread+1)
      CALL dbcsr_mm_sched_init(this%sched, &
                               product_wm=this%product_wm)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_mm_csr_init

! **************************************************************************************************
!> \brief Fills row hashtable from an existing matrix.
!> \param hashes ...
!> \param matrix ...
!> \param[in] block_estimate guess for the number of blocks in the product matrix, can be zero
!> \param row_map ...
!> \param col_map ...
! **************************************************************************************************
   SUBROUTINE fill_hash_tables(hashes, matrix, block_estimate, row_map, col_map)
      TYPE(hash_table_type), DIMENSION(:), INTENT(inout) :: hashes
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER                                            :: block_estimate
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: row_map, col_map

      CHARACTER(len=*), PARAMETER :: routineN = 'fill_hash_tables', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, handle, i, imat, n_rows, row

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      imat = 1
!$    imat = OMP_GET_THREAD_NUM()+1
      IF (PRESENT(row_map)) THEN
         n_rows = matrix%nblkrows_local
         CALL dbcsr_assert(SIZE(hashes), "EQ", n_rows, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Local row count mismatch", 674)
      ELSE
         n_rows = matrix%nblkrows_total
         CALL dbcsr_assert(SIZE(hashes), "EQ", n_rows, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Global row count mismatch", 679)
      ENDIF
      DO row = 1, n_rows
         ! create the hash table row with a reasonable initial size
         CALL hash_table_create(hashes(row), &
                                MAX(8, (3*block_estimate)/MAX(1, n_rows)))
      ENDDO
      ! We avoid using the iterator because we will use the existing
      ! work matrix instead of the BCSR index.
      DO i = 1, matrix%wms(imat)%lastblk
         row = matrix%wms(imat)%row_i(i)
         col = matrix%wms(imat)%col_i(i)
         IF (PRESENT(row_map)) row = row_map(row)
         IF (PRESENT(col_map)) col = col_map(col)
         CALL hash_table_add(hashes(row), col, i)
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE fill_hash_tables

! **************************************************************************************************
!> \brief Signal approaching end of multiplication
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_phaseout(this)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this

      CALL dbcsr_mm_sched_phaseout(this%sched)
   END SUBROUTINE dbcsr_mm_csr_phaseout

! **************************************************************************************************
!> \brief Finalizes a multiplication cycle for a set of C-blocks.
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_finalize(this)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_csr_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      CALL dbcsr_mm_sched_finalize(this%sched)

      ! Clear hash tables
      DO i = 1, SIZE(this%c_hashes)
         CALL hash_table_release(this%c_hashes(i))
      ENDDO
      DEALLOCATE (this%c_hashes)
      DEALLOCATE (this%stacks_descr)
      DEALLOCATE (this%stack_map)
      DEALLOCATE (this%m_size_maps)
      DEALLOCATE (this%n_size_maps)
      DEALLOCATE (this%k_size_maps)
      DEALLOCATE (this%stacks_fillcount)
      DEALLOCATE (this%stacks_data)
   END SUBROUTINE dbcsr_mm_csr_finalize

! **************************************************************************************************
!> \brief ...
!> \param this ...
!> \param left ...
!> \param right ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_csr_purge_stacks(this, left, right)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right

      CALL flush_stacks(this, left, right, purge=.TRUE.)
      CALL dbcsr_mm_sched_barrier()
   END SUBROUTINE dbcsr_mm_csr_purge_stacks

! **************************************************************************************************
!> \brief ...
!> \param this ...
!> \param left ...
!> \param right ...
!> \param purge ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE flush_stacks(this, left, right, purge)
      TYPE(dbcsr_mm_csr_type), INTENT(INOUT)             :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      LOGICAL, INTENT(IN), OPTIONAL                      :: purge

      INTEGER                                            :: i, min_fill, n_stacks
      INTEGER, DIMENSION(:, :), POINTER                  :: stack_data
      INTEGER, POINTER                                   :: stack_fillcount
      TYPE(stack_descriptor_type)                        :: stack_descr

      n_stacks = SIZE(this%stacks_data, 3)
      min_fill = SIZE(this%stacks_data, 2)*3/4 !TODO: play with this
      IF (PRESENT(purge)) THEN
         IF (purge) min_fill = 0
      ENDIF

      CALL dbcsr_mm_sched_begin_burst(this%sched)

      DO i = 1, n_stacks
         IF (this%stacks_fillcount(i) > min_fill) THEN
            stack_data => this%stacks_data(:, :, i)
            stack_fillcount => this%stacks_fillcount(i)
            stack_descr = this%stacks_descr(i)

            CALL dbcsr_mm_sched_process(this%sched, &
                                        left, right, &
                                        stack_data=stack_data, &
                                        stack_fillcount=stack_fillcount, &
                                        stack_descr=stack_descr)

            stack_fillcount = 0
         ENDIF
      END DO

      CALL dbcsr_mm_sched_end_burst()
   END SUBROUTINE flush_stacks

! **************************************************************************************************
!> \brief  Builds and sorts a CSR index from a list index.
!> \param mi ...
!> \param mf ...
!> \param ai ...
!> \param af ...
!> \param row_p ...
!> \param blk_info ...
!> \param list_index ...
!> \param nnorms ...
!> \param csr_norms ...
!> \param list_norms ...
!> \author JV
!> <b>Modification history:</b>
!> - 2011-02-15 [UB] Adapted to use DBCSR-type CSR indexing
! **************************************************************************************************
   SUBROUTINE build_csr_index(mi, mf, ai, af, row_p, blk_info, list_index, &
                              nnorms, csr_norms, list_norms)
      INTEGER, INTENT(IN)                                :: mi, mf, ai, af
      INTEGER, DIMENSION(mi:mf+1), INTENT(OUT)           :: row_p
      INTEGER, DIMENSION(2, 1:af-ai+1), INTENT(OUT)      :: blk_info
      INTEGER, DIMENSION(3, 1:af), INTENT(IN)            :: list_index
      INTEGER, INTENT(IN)                                :: nnorms
      REAL(KIND=sp), DIMENSION(1:af-ai+1), INTENT(OUT)   :: csr_norms
      REAL(KIND=sp), DIMENSION(:), INTENT(IN)            :: list_norms

      CHARACTER(len=*), PARAMETER :: routineN = 'build_csr_index', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: careful = .FALSE., dbg = .FALSE.

      INTEGER                                            :: i, row
      INTEGER, DIMENSION(mi:mf)                          :: counts

!   ---------------------------------------------------------------------------
! Counts blocks per row and calculates the offsets.

      IF (dbg) THEN
         WRITE (*, '(I7,1X,5(A,2(1X,I7)))') 0, "bci", mi, mf, ";", ai, af
         !write(*,'(3(I7))')list_index(:,ai:af)
      ENDIF

      counts(:) = 0
      DO i = ai, af
         IF (careful) THEN
            CALL dbcsr_assert(list_index(1, i), "GE", mi, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Out of range", 844)
            CALL dbcsr_assert(list_index(1, i), "LE", mf, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Out of range", 847)
         ENDIF
         counts(list_index(1, i)) = counts(list_index(1, i))+1
      ENDDO
      row_p(mi) = 0
      DO i = mi+1, mf+1
         row_p(i) = row_p(i-1)+counts(i-1)
      ENDDO
      ! Adds every block to its corresponding row.
      counts(:) = 0
      DO i = ai, af
         row = list_index(1, i)
         counts(row) = counts(row)+1
         IF (careful) THEN
            CALL dbcsr_assert(row_p(row)+counts(row), "LE", af-ai+1, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Out of range", 863)
            CALL dbcsr_assert(row_p(row)+counts(row), "GE", 1, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Out of range", 866)
         ENDIF
         blk_info(1, row_p(row)+counts(row)) = list_index(2, i)
         blk_info(2, row_p(row)+counts(row)) = list_index(3, i)
         IF (nnorms .GT. 0) THEN
            csr_norms(row_p(row)+counts(row)) = list_norms(i)
         ENDIF
      ENDDO
      IF (nnorms .EQ. 0) THEN
         csr_norms(:) = 0.0_sp
      ENDIF
   END SUBROUTINE build_csr_index

! **************************************************************************************************
!> \brief Determines whether a transpose must be applied
!> \param[in] row   The absolute matrix row.
!> \param[in] column          The absolute matrix column.
!> \retval transpose ...
!> \par Source
!> This function is copied from dbcsr_dist_operations for speed reasons.
! **************************************************************************************************
   ELEMENTAL FUNCTION my_checker_tr(row, column) RESULT(transpose)
      INTEGER, INTENT(IN)                                :: row, column
      LOGICAL                                            :: transpose

      transpose = BTEST(column+row, 0) .EQV. column .GE. row

   END FUNCTION my_checker_tr

! -----------------------------------------------------------------------------
! Beginning of hashtable.
! this file can be 'INCLUDE'ed verbatim in various place, where it needs to be
! part of the module to guarantee inlining
! hashes (c,p) pairs, where p is assumed to be >0
! on return (0 is used as a flag for not present)
!
!
! *****************************************************************************
!> \brief finds a prime equal or larger than i, needed at creation
!> \param i ...
!> \retval res ...
! *****************************************************************************
  FUNCTION matching_prime(i) RESULT(res)
    INTEGER, INTENT(IN)                      :: i
    INTEGER                                  :: res

    INTEGER                                  :: j

    res=i
    j=0
    DO WHILE (j<res)
      DO j=2,res-1
         IF (MOD(res,j)==0) THEN
            res=res+1
            EXIT
         ENDIF
      ENDDO
    ENDDO
  END FUNCTION

! *****************************************************************************
!> \brief create a hash_table of given initial size. 
!>        the hash table will expand as needed (but this requires rehashing)
!> \param hash_table ...
!> \param table_size ...
! *****************************************************************************
  SUBROUTINE hash_table_create(hash_table,table_size)
    TYPE(hash_table_type)                    :: hash_table
    INTEGER, INTENT(IN)                      :: table_size

    INTEGER                                  :: j

    ! guarantee a minimal hash table size (8), so that expansion works

    j=3
    DO WHILE(2**j-1<table_size)
       j=j+1
    ENDDO
    hash_table%nmax=2**j-1
    hash_table%prime=matching_prime(hash_table%nmax)
    hash_table%nele=0
    ALLOCATE(hash_table%table(0:hash_table%nmax))
  END SUBROUTINE hash_table_create

! *****************************************************************************
!> \brief ...
!> \param hash_table ...
! *****************************************************************************
  SUBROUTINE hash_table_release(hash_table)
    TYPE(hash_table_type)                    :: hash_table

    hash_table%nmax=0
    hash_table%nele=0
    DEALLOCATE(hash_table%table)

  END SUBROUTINE hash_table_release

! *****************************************************************************
!> \brief add a pair (c,p) to the hash table
!> \param hash_table ...
!> \param c this value is being hashed
!> \param p this is being stored
! *****************************************************************************
  RECURSIVE SUBROUTINE hash_table_add(hash_table,c,p)
    TYPE(hash_table_type), INTENT(INOUT)     :: hash_table
    INTEGER, INTENT(IN)                      :: c, p

    REAL(KIND=real_8), PARAMETER :: hash_table_expand = 1.5_real_8, &
      inv_hash_table_fill = 2.5_real_8

    INTEGER                                  :: i, j
    TYPE(ele_type), ALLOCATABLE, &
      DIMENSION(:)                           :: tmp_hash

! if too small, make a copy and rehash in a larger table

    IF (hash_table%nele*inv_hash_table_fill>hash_table%nmax) THEN
       ALLOCATE(tmp_hash(LBOUND(hash_table%table,1):UBOUND(hash_table%table,1)))
       tmp_hash(:)=hash_table%table
       CALL hash_table_release(hash_table)
       CALL hash_table_create(hash_table,INT((UBOUND(tmp_hash,1)+8)*hash_table_expand))
       DO i=LBOUND(tmp_hash,1),UBOUND(tmp_hash,1)
          IF (tmp_hash(i)%c.NE.0) THEN
             CALL hash_table_add(hash_table,tmp_hash(i)%c,tmp_hash(i)%p)
          ENDIF
       ENDDO
       DEALLOCATE(tmp_hash)
    ENDIF

    hash_table%nele=hash_table%nele+1
    i=IAND(c*hash_table%prime,hash_table%nmax)

    DO j=i,hash_table%nmax
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          hash_table%table(j)%c=c
          hash_table%table(j)%p=p
          RETURN
       ENDIF
    ENDDO
    DO j=0,i-1
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          hash_table%table(j)%c=c
          hash_table%table(j)%p=p
          RETURN
       ENDIF
    ENDDO

  END SUBROUTINE hash_table_add

! *****************************************************************************
!> \brief ...
!> \param hash_table ...
!> \param c ...
!> \retval p ...
! *****************************************************************************
  PURE FUNCTION hash_table_get(hash_table,c) RESULT(p)
    TYPE(hash_table_type), INTENT(IN)        :: hash_table
    INTEGER, INTENT(IN)                      :: c
    INTEGER                                  :: p

    INTEGER                                  :: i, j

    i=IAND(c*hash_table%prime,hash_table%nmax)

    ! catch the likely case first
    IF (hash_table%table(i)%c==c) THEN
       p=hash_table%table(i)%p
       RETURN
    ENDIF

    DO j=i,hash_table%nmax
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          p=hash_table%table(j)%p
          RETURN
       ENDIF
    ENDDO
    DO j=0,i-1
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          p=hash_table%table(j)%p
          RETURN
       ENDIF
    ENDDO

    ! we should never reach this point.
    p=HUGE(p)

  END FUNCTION hash_table_get

! End of hashtable
! -----------------------------------------------------------------------------

END MODULE dbcsr_mm_csr
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Second layer of the dbcsr matrix-matrix multiplication.
!>            It divides the multiplication in a cache-oblivious manner.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2010-02-23 Moved from dbcsr_operations
!>  - 2011-11    Moved parameter-stack processing routines to
!>               dbcsr_mm_methods.
!>  - 2013-01    extensive refactoring (Ole Schuett)
! **************************************************************************************************

MODULE dbcsr_mm_multrec

   USE array_types,                     ONLY: array_data,&
                                              array_equality
   USE dbcsr_config,                    ONLY: multrec_limit
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_col_dist,&
                                              dbcsr_distribution_has_threads,&
                                              dbcsr_distribution_local_cols,&
                                              dbcsr_distribution_local_cols_obj,&
                                              dbcsr_distribution_local_rows,&
                                              dbcsr_distribution_local_rows_obj,&
                                              dbcsr_distribution_row_dist,&
                                              dbcsr_distribution_thread_dist
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_mm_csr,                    ONLY: dbcsr_mm_csr_finalize,&
                                              dbcsr_mm_csr_init,&
                                              dbcsr_mm_csr_lib_finalize,&
                                              dbcsr_mm_csr_lib_init,&
                                              dbcsr_mm_csr_multiply,&
                                              dbcsr_mm_csr_phaseout,&
                                              dbcsr_mm_csr_purge_stacks,&
                                              dbcsr_mm_csr_type
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size
   USE dbcsr_types,                     ONLY: dbcsr_type,&
                                              dbcsr_work_type
   USE kinds,                           ONLY: int_8,&
                                              real_8,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_multrec'
   LOGICAL, PARAMETER :: careful_mod = .FALSE.

! **************************************************************************************************
!> \brief Used to carry data among the various calls.  Each thread has
!>        its own private copy.
!> \var c_has_symmetry    The product matrix has symmetry
!> \var keep_sparsity     Sparsity of C matrix should be kept
!> \var use_eps           Use on-the-fly filtering
!> \var local_indexing    The A and B matrix indices and C matrix work indices
!>                        are not global but local to process rows and columns
!> \var m_sizes           Block sizes of A and C matrix rows, indexed locally
!> \var n_sizes           Block sizes of B and C matrix columns, indexed locally
!> \var k_sizes           Block sizes of A matrix columns and B matrix rows,
!>                        indexed locally
!> \var c_local_rows      C and A matrix local rows.  Map from
!>                        local row (index) to global row (value).
!> \var c_local_cols      C and B matrix local columns.  Map from local column
!>                        (index) to global column (value).
!> \var k_locals          A matrix local columns and B matrix local rows.  Map
!>                        from local row/column (index) to global row/column
!>                        (value).
!> \var c_global_rows     C and A matrix global rows.  Map from global rows
!>                        (index) to local rows (value).
!> \var c_global_cols     C and B matrix global columns.  Map from global
!>                        columns (index) to local columns (value).
!> \var max_m             Maximum size of C or A matrix row block size.
!> \var max_n             Maximum size of C or B matrix column block size.
!> \var max_k             Maximum size of A matrix colum and B matrix row block
!>                        size.
!> \var m_size_maps       Map from an A matrix row block size to an index of
!>                        the most common A matrix row block sizes.
!> \var n_size_maps       Map from a B matrix column block size to an index of
!>                        the most common B matrix column block sizes.
!> \var k_size_maps       Map from an A matrix row or B matrix column block
!>                        size to an index of the most row/column block sizes.
!> \var m_size_maps_size  Size of the m_size_maps array (i.e., the maximum block
!>                        size + 1).
!> \var n_size_maps_size  Size of the n_size_maps array (i.e., the maximum block
!>                        size + 1).
!> \var k_size_maps_size  Size of the k_size_maps array (i.e., the maximum block
!>                        size + 1).
!> \var nm_stacks         The number of most common m blocks sizes.
!> \var nn_stacks         The number of most common n blocks sizes.
!> \var nk_stacks         The number of most common k blocks sizes.
!> \var stack_map         Map from most common (n,k,m) block sizes to a stack
!>                        number within a stack group.
!>                        m, n, and k sizes.
!> \var row_max_epss      Maximum eps to be used for one row.
!> \var a_norms           Norms of A matrix blocks.
!> \var b_norms           Norms of B matrix blocks.
!> \var my_wm             Work matrix associated with this thread
!> \var lastblk           Number of elements in the work matrix
!>
!> \var datasize          Data size of the work matrix (view of
!>                        same work matrix variable).
!> \var original_lastblk  Number of work matrix blocks before addition
!> \var flop              flop count
! **************************************************************************************************
   TYPE dbcsr_mm_multrec_type
      PRIVATE
      LOGICAL :: c_has_symmetry, keep_sparsity, use_eps
      LOGICAL :: local_indexing
      INTEGER, DIMENSION(:), POINTER :: m_sizes, n_sizes, k_sizes
      INTEGER, DIMENSION(:), POINTER :: c_local_rows, c_local_cols, k_locals, &
                                        c_global_rows, c_global_cols

      REAL(KIND=sp), DIMENSION(:), POINTER :: row_max_epss, a_norms, b_norms
      REAL(KIND=real_8)     :: eps
      INTEGER               :: original_lastblk
      INTEGER(kind=int_8)   :: flop
      TYPE(dbcsr_work_type), POINTER :: product_wm => Null()
      TYPE(dbcsr_mm_csr_type)        :: csr
      LOGICAL                        :: new_row_max_epss = .FALSE.
      LOGICAL                        :: initialized = .FALSE.
   END TYPE dbcsr_mm_multrec_type

   ! **************************************************************************************************
   PUBLIC :: dbcsr_mm_multrec_type
   PUBLIC :: dbcsr_mm_multrec_lib_init, dbcsr_mm_multrec_lib_finalize
   PUBLIC :: dbcsr_mm_multrec_phaseout
   PUBLIC :: dbcsr_mm_multrec_init, dbcsr_mm_multrec_finalize
   PUBLIC :: dbcsr_mm_multrec_multiply

CONTAINS

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_lib_init()

      CALL dbcsr_mm_csr_lib_init()
   END SUBROUTINE

! **************************************************************************************************
!> \brief Finalize the library
!> \param group ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_lib_finalize(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      CALL dbcsr_mm_csr_lib_finalize(group, output_unit)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Sets up recursive multiplication
!>
!>
!> \param this ...
!> \param[in] left left DBCSR matrix
!> \param[in] right right DBCSR matrix
!> \param[in,out] product     resulting DBCSR product matrix
!> \param[in] keep_sparsity (optional) retain the sparsity of the
!>                                 existing product matrix, default is no
!> \param[in] eps             (optional) on-the-fly filtering epsilon
!> \param row_max_epss ...
!> \param block_estimate ...
!> \param right_row_blk_size ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_init(this, left, right, product, &
                                    keep_sparsity, eps, row_max_epss, block_estimate, right_row_blk_size)
      TYPE(dbcsr_mm_multrec_type), INTENT(out)           :: this
      TYPE(dbcsr_type), INTENT(IN), OPTIONAL             :: left, right
      TYPE(dbcsr_type), INTENT(INOUT)                    :: product
      LOGICAL, INTENT(IN)                                :: keep_sparsity
      REAL(kind=real_8), INTENT(in), OPTIONAL            :: eps
      REAL(kind=sp), DIMENSION(:), INTENT(IN), TARGET    :: row_max_epss
      INTEGER, INTENT(IN)                                :: block_estimate
      INTEGER, DIMENSION(:), INTENT(IN)                  :: right_row_blk_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_multrec_init', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: c_nblkcols_local, c_nblkrows_local, &
                                                            handle, i, ithread
      INTEGER, DIMENSION(:), POINTER                     :: c_local_cols, c_local_rows

!$    INTEGER, DIMENSION(:), POINTER           :: product_thread_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      !
      IF (this%initialized) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",209,"multrec already initialized.")

      IF (PRESENT(left) .NEQV. PRESENT(right)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",212,"Must both left and right provided or not.")

      IF (PRESENT(left) .AND. PRESENT(right)) THEN
         ! Ensures that the index is correctly defined.
         IF (.NOT. left%list_indexing) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",217,"Must use list indexing for this routine.")
         IF (left%bcsc) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",219,"Wrong routine for BCSC matrices.")

         IF (right%bcsc) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",222,"Wrong routine for BCSC matrices.")
         CALL dbcsr_assert(right%local_indexing, "EQV", left%local_indexing, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Local index useage must be consistent.", 225)

         IF (right%local_indexing) THEN
            IF (.NOT. left%local_indexing) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",229,"Wrong left format for local_cols.")
            IF (.NOT. right%local_indexing) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",231,"Wrong right format for local_cols.")
         ELSE
            IF (left%local_indexing) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",234,"Wrong left format for not local_cols.")
            IF (right%local_indexing) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",236,"Wrong right format for not local_cols.")
         ENDIF

         this%local_indexing = left%local_indexing
      ELSE
         this%local_indexing = .TRUE.
      ENDIF
      !
      ! Fill result data structure.
      this%keep_sparsity = keep_sparsity
      this%c_has_symmetry = product%symmetry
      this%use_eps = PRESENT(eps)
      this%original_lastblk = product%wms(ithread+1)%lastblk
      this%flop = INT(0, int_8)
      this%product_wm => product%wms(ithread+1)

      IF (PRESENT(eps)) THEN
         this%eps = eps
      ELSE
         this%eps = 0.0_real_8
      ENDIF
      !
      !
!$    NULLIFY (product_thread_dist)
!$    IF (.NOT. dbcsr_distribution_has_threads(product%dist)) &
!$       CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",261,"Missing thread distribution.")
!$    product_thread_dist => array_data( &
!$                           dbcsr_distribution_thread_dist(product%dist))
      !
      ! Find out the C/A rows and C/B columns and sizes.
      c_nblkrows_local = product%nblkrows_local
      c_local_rows => array_data(product%local_rows)
      c_nblkcols_local = product%nblkcols_local
      c_local_cols => array_data(product%local_cols)
      IF (this%local_indexing) THEN
         this%c_local_rows => c_local_rows
         this%c_local_cols => c_local_cols
      ELSE
         ALLOCATE (this%c_local_rows(product%nblkrows_total))
         ALLOCATE (this%c_local_cols(product%nblkcols_total))
         DO i = 1, product%nblkrows_total
            this%c_local_rows(i) = i
         END DO
         DO i = 1, product%nblkcols_total
            this%c_local_cols(i) = i
         END DO
      ENDIF
      IF (dbg) WRITE (*, *) "setting up for product", product%name
      IF (careful_mod) THEN
         IF (.NOT. array_equality(dbcsr_distribution_local_rows_obj(product%dist), &
                                  product%local_rows)) THEN
            WRITE (*, *) "row dist", dbcsr_distribution_row_dist(product%dist)
            WRITE (*, *) "dist local rows", dbcsr_distribution_local_rows(product%dist)
            WRITE (*, *) " mat local rows", array_data(product%local_rows)
            CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",290,"Array mismatch.")
         ENDIF
         IF (.NOT. array_equality(dbcsr_distribution_local_cols_obj(product%dist), &
                                  product%local_cols)) THEN
            WRITE (*, *) "col dist", dbcsr_distribution_col_dist(product%dist)
            WRITE (*, *) "dist local cols", dbcsr_distribution_local_cols(product%dist)
            WRITE (*, *) " mat local cols", array_data(product%local_cols)
            CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",297,"Array mismatch.")
         ENDIF
         CALL dbcsr_assert(SIZE(c_local_rows), "EQ", c_nblkrows_local, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Row count mismatch.", 301)
         CALL dbcsr_assert(SIZE(c_local_cols), "EQ", c_nblkcols_local, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Column count mismatch.", 304)
      ENDIF
      !
      ! And the k epsilons
      IF (this%local_indexing) THEN
         IF ((PRESENT(left) .AND. PRESENT(right)) .OR. .NOT. this%use_eps) THEN
            ALLOCATE (this%row_max_epss(c_nblkrows_local))
            this%new_row_max_epss = .TRUE.
         ENDIF
      ELSE
         ALLOCATE (this%row_max_epss(product%nblkrows_total))
         this%new_row_max_epss = .TRUE.
      ENDIF
      IF (this%use_eps) THEN
         IF (this%local_indexing) THEN
            IF (PRESENT(left) .AND. PRESENT(right)) THEN
               CALL local_filter_sp(row_max_epss, c_nblkrows_local, c_local_rows, &
                                    this%row_max_epss)
            ELSE
               this%row_max_epss => row_max_epss
            ENDIF
         ELSE
            IF (careful_mod .AND. SIZE(row_max_epss) .NE. SIZE(this%row_max_epss)) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",327,"max epss local/global mismatch.")
            this%row_max_epss(:) = row_max_epss(:)
         ENDIF
      ELSE
         this%row_max_epss(:) = -HUGE(0.0_sp)
      ENDIF
      !
      IF (this%local_indexing) THEN
         ALLOCATE (this%m_sizes(c_nblkrows_local))
         CALL local_filter(array_data(product%row_blk_size), SIZE(c_local_rows), &
                           c_local_rows, this%m_sizes)
         ALLOCATE (this%n_sizes(c_nblkcols_local))
         CALL local_filter(array_data(product%col_blk_size), SIZE(c_local_cols), &
                           c_local_cols, this%n_sizes)
      ELSE
         ALLOCATE (this%m_sizes(product%nblkrows_total))
         this%m_sizes(:) = array_data(product%row_blk_size)
         ALLOCATE (this%n_sizes(product%nblkcols_total))
         this%n_sizes(:) = array_data(product%col_blk_size)
      ENDIF
      !
      NULLIFY (this%k_locals)
      NULLIFY (this%k_sizes)
      IF (.NOT. this%local_indexing) THEN
         ALLOCATE (this%k_locals(right%nblkrows_total))
         DO i = 1, right%nblkrows_total
            this%k_locals(i) = i
         END DO
         this%k_sizes => array_data(right%row_blk_size)
      ENDIF

      !TODO: should we move this up?
      CALL dbcsr_mm_csr_init(this%csr, &
                             left=left, right=right, product=product, &
                             m_sizes=this%m_sizes, n_sizes=this%n_sizes, &
                             block_estimate=block_estimate, &
                             right_row_blk_size=right_row_blk_size)

      this%initialized = .TRUE.
      CALL timestop(handle)
   END SUBROUTINE dbcsr_mm_multrec_init

! **************************************************************************************************
!> \brief Multiplies two DBCSR matrices using recursive algorithm
!>
!> This routine sets up the multiplication.  Specifically, it <ul>
!> <li> verifies input sanity
!> <li> converts everything into "local indexing"
!> </ul>
!>
!> \param this ...
!> \param[in] left left DBCSR matrix
!> \param[in] right right DBCSR matrix
!> \param[out] flop           (optional) number of effective double-precision
!>                            floating point operations performed
!> \param[in] a_norms         (optional) norms of left-matrix blocks
!> \param[in] b_norms         (optional) norms of right-matrix blocks
!> \param do_rec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_multiply(this, left, right, flop, &
                                        a_norms, b_norms, do_rec)
      TYPE(dbcsr_mm_multrec_type), INTENT(inout)         :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER(KIND=int_8), INTENT(OUT)                   :: flop
      REAL(kind=sp), DIMENSION(:), INTENT(in), TARGET    :: a_norms, b_norms
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_rec

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_multrec_multiply', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ithread, nthreads, t_a_f, t_a_l, t_b_f, &
                                                            t_b_l
      INTEGER, DIMENSION(:), POINTER                     :: k_locals
      LOGICAL                                            :: my_do_rec

!   ---------------------------------------------------------------------------

      IF (.NOT. this%initialized) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",405,"multrec not initialized.")

      my_do_rec = .TRUE.
      IF (PRESENT(do_rec)) my_do_rec = do_rec

      ithread = 0; nthreads = 1
!$    ithread = OMP_GET_THREAD_NUM(); nthreads = OMP_GET_NUM_THREADS()
      this%flop = 0

      ! Find out the local A columns / B rows and sizes
      ! The right%local_rows is setup by the communication engine.
      IF (this%local_indexing) THEN
         k_locals => array_data(right%local_rows)
         this%k_locals => k_locals
         CALL ensure_array_size(this%k_sizes, ub=SIZE(k_locals))
         CALL local_filter(array_data(right%row_blk_size), SIZE(k_locals), &
                           k_locals, this%k_sizes)
      ELSE
         k_locals => this%k_locals
      ENDIF
      ! Setup the block norms
      this%a_norms => a_norms
      this%b_norms => b_norms

      ! Start local multiplication
      t_a_f = 1
      t_a_l = left%nblks
      t_b_f = 1
      t_b_l = right%nblks
!$    IF (ASSOCIATED(left%thr_c)) THEN
!$       ithread = OMP_GET_THREAD_NUM()
!$       t_a_f = left%thr_c(ithread+1)+1
!$       t_a_l = left%thr_c(ithread+2)
!$    ENDIF
      IF (this%local_indexing) THEN
         CALL sparse_multrec(this, left, right, &
                             1, left%nblkrows_local, &
                             1, right%nblkcols_local, &
                             1, SIZE(k_locals), &
                             t_a_f, t_a_l, left%coo_l, &
                             t_b_f, t_b_l, right%coo_l, &
                             0, my_do_rec)
      ELSE
         CALL sparse_multrec(this, left, right, &
                             1, left%nblkrows_total, &
                             1, right%nblkcols_total, &
                             1, SIZE(k_locals), &
                             t_a_f, t_a_l, left%coo_l, &
                             t_b_f, t_b_l, right%coo_l, &
                             0, my_do_rec)
      ENDIF

      CALL dbcsr_mm_csr_purge_stacks(this%csr, left, right)

      flop = flop+this%flop
      !
      IF (ASSOCIATED(this%k_sizes) .AND. this%local_indexing) &
         DEALLOCATE (this%k_sizes)

   END SUBROUTINE dbcsr_mm_multrec_multiply

! **************************************************************************************************
!> \brief Signal approaching end of multiplication
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_phaseout(this)
      TYPE(dbcsr_mm_multrec_type), INTENT(INOUT)         :: this

      CALL dbcsr_mm_csr_phaseout(this%csr)
   END SUBROUTINE dbcsr_mm_multrec_phaseout

! **************************************************************************************************
!> \brief Sets up recursive multiplication
!>
!>
!> \param this ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_multrec_finalize(this)
      TYPE(dbcsr_mm_multrec_type), INTENT(inout)         :: this

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_multrec_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ithread

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. this%initialized) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_multrec.F",495,"multrec not initialized.")

      CALL dbcsr_mm_csr_finalize(this%csr)

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()

      ! Release the carrier
      DEALLOCATE (this%m_sizes)
      DEALLOCATE (this%n_sizes)
      IF (this%new_row_max_epss) DEALLOCATE (this%row_max_epss)

      IF (this%local_indexing) THEN
         CALL remap_local2global(this%product_wm%row_i, &
                                 this%product_wm%col_i, &
                                 this%c_local_rows, this%c_local_cols, &
                                 this%original_lastblk+1, this%product_wm%lastblk)
      ENDIF
      !

      IF (.NOT. this%local_indexing) THEN
         NULLIFY (this%k_sizes)
         DEALLOCATE (this%c_local_rows)
         DEALLOCATE (this%c_local_cols)
         DEALLOCATE (this%k_locals)
      ENDIF

      this%initialized = .FALSE.
      CALL timestop(handle)
   END SUBROUTINE dbcsr_mm_multrec_finalize

! **************************************************************************************************
!> \brief Performs recursive multiplication
!> \param this ...
!> \param left ...
!> \param right ...
!> \param mi ...
!> \param mf ...
!> \param ni ...
!> \param nf ...
!> \param ki ...
!> \param kf ...
!> \param ai ...
!> \param af ...
!> \param a_index ...
!> \param bi ...
!> \param bf ...
!> \param b_index ...
!> \param d ...
!> \param do_rec ...
!> \author Joost VandeVondele
! **************************************************************************************************
   RECURSIVE SUBROUTINE sparse_multrec(this, left, right, mi, mf, ni, nf, ki, kf, &
                                       ai, af, a_index, bi, bf, b_index, &
                                       d, do_rec)
      TYPE(dbcsr_mm_multrec_type), INTENT(INOUT)         :: this
      TYPE(dbcsr_type), INTENT(IN)                       :: left, right
      INTEGER, INTENT(IN)                                :: mi, mf, ni, nf, ki, kf, ai, af
      INTEGER, DIMENSION(3, 1:af), INTENT(IN)            :: a_index
      INTEGER, INTENT(IN)                                :: bi, bf
      INTEGER, DIMENSION(3, 1:bf), INTENT(IN)            :: b_index
      INTEGER, INTENT(IN)                                :: d
      LOGICAL, INTENT(IN)                                :: do_rec

      CHARACTER(len=*), PARAMETER :: routineN = 'sparse_multrec', routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: acut, bcut, cut, K, M, N, s1

!   ---------------------------------------------------------------------------

      IF (dbg) THEN
         WRITE (*, '(I7,1X,5(A,2(1X,I7)))') d, " rm", mi, mf, ",", ni, nf, ",", ki, kf, "/", ai, af, ",", bi, bf
      ENDIF
      IF (.TRUE.) THEN
         IF (af .LT. ai .OR. bf .LT. bi .OR. mf .LT. mi .OR. nf .LT. ni .OR. kf .LT. ki) THEN
            IF (dbg) WRITE (*, *) "Empty"
            RETURN
         ENDIF
      ENDIF

      IF (.NOT. do_rec .OR. (af-ai+1 <= multrec_limit .AND. bf-bi+1 <= multrec_limit)) THEN
         IF (af-ai+1 .GT. 0 .AND. bf-bi+1 .GT. 0) &
            CALL dbcsr_mm_csr_multiply(this%csr, left, right, &
                                       mi=mi, mf=mf, ni=ni, nf=nf, ki=ki, kf=kf, &
                                       ai=ai, af=af, &
                                       bi=bi, bf=bf, &
                                       m_sizes=this%m_sizes, n_sizes=this%n_sizes, k_sizes=this%k_sizes, &
                                       c_local_rows=this%c_local_rows, c_local_cols=this%c_local_cols, &
                                       c_has_symmetry=this%c_has_symmetry, keep_sparsity=this%keep_sparsity, &
                                       use_eps=this%use_eps, row_max_epss=this%row_max_epss, &
                                       flop=this%flop, &
                                       a_index=a_index, b_index=b_index, &
                                       a_norms=this%a_norms, b_norms=this%b_norms)
         RETURN
      ENDIF

      M = mf-mi+1
      N = nf-ni+1
      K = kf-ki+1
      IF (dbg) THEN
         WRITE (*, *) 'm,k,n', M, K, N
      ENDIF
      IF (M >= MAX(N, K)) cut = 1
      IF (K >= MAX(N, M)) cut = 2
      IF (N >= MAX(M, K)) cut = 3
      SELECT CASE (cut)
      CASE (1)
         s1 = M/2
         acut = find_cut_row(ai, af, a_index, mi+s1-1)
         CALL sparse_multrec(this, left, right, mi, mi+s1-1, ni, nf, ki, kf, &
                             ai, acut-1, a_index, bi, bf, b_index, d+1, do_rec)
         CALL sparse_multrec(this, left, right, mi+s1, mf, ni, nf, ki, kf, &
                             acut, af, a_index, bi, bf, b_index, d+1, do_rec)
      CASE (2)
         s1 = K/2
         acut = find_cut_col(ai, af, a_index, ki+s1-1)
         IF (dbg) THEN
            WRITE (*, *) N, s1, ni+s1-1, "/", ai, af, acut
            WRITE (*, '(3(I7))') a_index
         ENDIF
         bcut = find_cut_row(bi, bf, b_index, ki+s1-1)
         IF (dbg) THEN
            WRITE (*, *) N, s1, ni+s1-1, "/", bi, bf, bcut
            WRITE (*, '(3(I7))') b_index
         ENDIF
         CALL sparse_multrec(this, left, right, mi, mf, ni, nf, ki, ki+s1-1, &
                             ai, acut-1, a_index, bi, bcut-1, b_index, d+1, do_rec)
         CALL sparse_multrec(this, left, right, mi, mf, ni, nf, ki+s1, kf, &
                             acut, af, a_index, bcut, bf, b_index, d+1, do_rec)
      CASE (3)
         s1 = N/2
         bcut = find_cut_col(bi, bf, b_index, ni+s1-1)
         IF (dbg) THEN
            WRITE (*, *) N, s1, ni+s1-1, "/", bi, bf, bcut
            WRITE (*, '(3(I7))') b_index
         ENDIF
         CALL sparse_multrec(this, left, right, mi, mf, ni, ni+s1-1, ki, kf, &
                             ai, af, a_index, bi, bcut-1, b_index, d+1, do_rec)
         CALL sparse_multrec(this, left, right, mi, mf, ni+s1, nf, ki, kf, &
                             ai, af, a_index, bcut, bf, b_index, d+1, do_rec)
      END SELECT
   END SUBROUTINE sparse_multrec

! ***************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param ai ...
!> \param af ...
!> \param a ...
!> \param val ...
!> \retval res ...
!> \author JV
! **************************************************************************************************
   PURE FUNCTION find_cut_row(ai, af, a, val) RESULT(res)
      INTEGER, INTENT(IN)                                :: ai, af
      INTEGER, DIMENSION(3, 1:af), INTENT(IN)            :: a
      INTEGER, INTENT(IN)                                :: val
      INTEGER                                            :: res

      INTEGER                                            :: i, ihigh, ilow

! do a log(N) search along the ordered index

      ilow = ai
      IF (a(1, ilow) > val) THEN
         res = ilow
         RETURN
      ENDIF

      ihigh = af
      IF (a(1, ihigh) <= val) THEN
         res = ihigh+1
         RETURN
      ENDIF

      DO
         IF (ihigh-ilow == 1) EXIT
         i = (ilow+ihigh)/2
         IF (a(1, i) > val) THEN
            ihigh = i
         ELSE
            ilow = i
         ENDIF
      ENDDO
      res = ihigh

      ! the linear search version
      ! DO i=ai,af
      !    IF (a(i)%r>val) EXIT
      !ENDDO
      !res=i
   END FUNCTION find_cut_row

! ***************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param ai ...
!> \param af ...
!> \param a ...
!> \param val ...
!> \retval res ...
!> \author JV
! **************************************************************************************************
   PURE FUNCTION find_cut_col(ai, af, a, val) RESULT(res)
      INTEGER, INTENT(IN)                                :: ai, af
      INTEGER, DIMENSION(3, 1:af), INTENT(IN)            :: a
      INTEGER, INTENT(IN)                                :: val
      INTEGER                                            :: res

      INTEGER                                            :: i, ihigh, ilow

! do a log(N) search along the ordered index

      ilow = ai
      IF (a(2, ilow) > val) THEN
         res = ilow
         RETURN
      ENDIF

      ihigh = af
      IF (a(2, ihigh) <= val) THEN
         res = ihigh+1
         RETURN
      ENDIF

      DO
         IF (ihigh-ilow == 1) EXIT
         i = (ilow+ihigh)/2
         IF (a(2, i) > val) THEN
            ihigh = i
         ELSE
            ilow = i
         ENDIF
      ENDDO
      res = ihigh

      ! the linear search version
      ! DO i=ai,af
      !    IF (a(i)%c>val) EXIT
      !ENDDO
      !res=i
   END FUNCTION find_cut_col

! ***************************************************************************************************
!> \brief Packs a globally-indexed array into a locally-indexed array.
!> \param row_i ...
!> \param col_i ...
!> \param local_rows ...
!> \param local_cols ...
!> \param first ...
!> \param last ...
! **************************************************************************************************
   PURE SUBROUTINE remap_local2global(row_i, col_i, local_rows, local_cols, &
                                      first, last)
      INTEGER, INTENT(in)                                :: last, first
      INTEGER, DIMENSION(:), INTENT(in)                  :: local_cols, local_rows
      INTEGER, DIMENSION(1:last), INTENT(inout)          :: col_i, row_i

      INTEGER                                            :: i

      DO i = first, last
         row_i(i) = local_rows(row_i(i))
         col_i(i) = local_cols(col_i(i))
      END DO
   END SUBROUTINE remap_local2global

! **************************************************************************************************
!> \brief Gathers the local elements from all data (full_data)
!>
!> \param[in] full_data       All elements
!> \param[in] nle             Number of local elements
!> \param[in] local_elements  List of local elements
!> \param[out] local_data     Local elements obtained from all elements
! **************************************************************************************************
   PURE SUBROUTINE local_filter(full_data, nle, local_elements, local_data)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: full_data
      INTEGER, INTENT(IN)                                :: nle
      INTEGER, DIMENSION(1:nle), INTENT(IN)              :: local_elements
      INTEGER, DIMENSION(1:nle), INTENT(OUT)             :: local_data

      INTEGER                                            :: l

      DO l = 1, nle
         local_data(l) = full_data(local_elements(l))
      END DO
   END SUBROUTINE local_filter

! **************************************************************************************************
!> \brief Gathers the local elements from all data (full_data) for
!>        single precision elements.
!> \param full_data ...
!> \param nle ...
!> \param local_elements ...
!> \param local_data ...
!> \note see local_filter
! **************************************************************************************************
   PURE SUBROUTINE local_filter_sp(full_data, nle, local_elements, local_data)
      REAL(KIND=sp), DIMENSION(:), INTENT(IN)            :: full_data
      INTEGER, INTENT(IN)                                :: nle
      INTEGER, DIMENSION(1:nle), INTENT(IN)              :: local_elements
      REAL(KIND=sp), DIMENSION(1:nle), INTENT(OUT)       :: local_data

      INTEGER                                            :: l

      DO l = 1, SIZE(local_data)
         local_data(l) = full_data(local_elements(l))
      END DO
   END SUBROUTINE local_filter_sp

END MODULE dbcsr_mm_multrec
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   First layer of the dbcsr matrix-matrix multiplication.
!>          It performs the MPI parallelization according to Cannon's algorithm.
!>
!> \author  Urban Borstnik
!>
!> <b>Modification history:</b>
!>  - 2010-02-23 Moved from dbcsr_operations
!>  - 2011-11    Moved parameter-stack processing routines to
!>               dbcsr_mm_methods.
!>  - 2013-01    reorganized code (Ole Schuett)
!>  - 2015-01    Introducting RMA (Alfio Lazzaro)
! **************************************************************************************************
MODULE dbcsr_mm_cannon
   USE acc_event,                       ONLY: acc_event_record,&
                                              acc_event_synchronize,&
                                              acc_stream_wait_event
   USE acc_stream,                      ONLY: acc_stream_associated,&
                                              acc_stream_create,&
                                              acc_stream_destroy,&
                                              acc_stream_type
   USE array_types,                     ONLY: array_data,&
                                              array_equality,&
                                              array_exists,&
                                              array_hold,&
                                              array_i1d_obj,&
                                              array_nullify,&
                                              array_release
   USE dbcsr_acc_operations,            ONLY: dbcsr_acc_transpose
   USE dbcsr_block_operations,          ONLY: dbcsr_block_conjg,&
                                              dbcsr_block_copy_aa,&
                                              dbcsr_block_real_neg,&
                                              dbcsr_block_scale,&
                                              dbcsr_block_transpose_aa,&
                                              dbcsr_data_set
   USE dbcsr_config,                    ONLY: default_resize_factor,&
                                              has_acc,&
                                              use_comm_thread,&
                                              use_mpi_exp,&
                                              use_mpi_filtering
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_ensure_size, dbcsr_data_get_size, &
        dbcsr_data_get_size_referenced, dbcsr_data_hold, dbcsr_data_host2dev, dbcsr_data_init, &
        dbcsr_data_new, dbcsr_data_release, dbcsr_data_set_pointer, &
        dbcsr_data_set_size_referenced, dbcsr_data_valid, dbcsr_get_data_p_c, dbcsr_get_data_p_d, &
        dbcsr_get_data_p_s, dbcsr_get_data_p_z, dbcsr_scalar, dbcsr_scalar_are_equal, &
        dbcsr_scalar_fill_all, dbcsr_scalar_negative, dbcsr_scalar_one, dbcsr_scalar_set_type, &
        dbcsr_scalar_zero
   USE dbcsr_data_types,                ONLY: dbcsr_datatype_sizeof
   USE dbcsr_dist_methods,              ONLY: &
        dbcsr_distribution_col_dist, dbcsr_distribution_get_num_images_1d, &
        dbcsr_distribution_has_col_clusters, dbcsr_distribution_has_row_clusters, &
        dbcsr_distribution_has_threads, dbcsr_distribution_hold, dbcsr_distribution_local_cols, &
        dbcsr_distribution_local_rows, dbcsr_distribution_make_threads, dbcsr_distribution_mp, &
        dbcsr_distribution_ncols, dbcsr_distribution_no_threads, dbcsr_distribution_nrows, &
        dbcsr_distribution_num_local_col_clusters, dbcsr_distribution_num_local_row_clusters, &
        dbcsr_distribution_release, dbcsr_distribution_row_dist, dbcsr_distribution_thread_dist
   USE dbcsr_dist_operations,           ONLY: dbcsr_create_image_dist,&
                                              dbcsr_get_local_vcols,&
                                              dbcsr_get_local_vrows,&
                                              dbcsr_make_dists_dense,&
                                              dbcsr_reset_locals,&
                                              dbcsr_reset_vlocals,&
                                              image_calculator,&
                                              make_sizes_dense
   USE dbcsr_index_operations,          ONLY: dbcsr_count_row_index,&
                                              dbcsr_has_local_row_index,&
                                              dbcsr_index_compact,&
                                              dbcsr_index_prune_deleted,&
                                              dbcsr_make_index_canonical,&
                                              dbcsr_make_index_list,&
                                              dbcsr_make_index_local_row,&
                                              dbcsr_repoint_index
   USE dbcsr_io,                        ONLY: dbcsr_print
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_mem_methods,               ONLY: dbcsr_mempool_clear,&
                                              dbcsr_mempool_destruct,&
                                              dbcsr_mempool_ensure_capacity,&
                                              dbcsr_memtype_setup
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_offsets, dbcsr_col_block_sizes, dbcsr_destroy_array, dbcsr_distribution, &
        dbcsr_get_data_type, dbcsr_get_index_memory_type, dbcsr_get_matrix_type, &
        dbcsr_has_symmetry, dbcsr_image_dist_hold, dbcsr_image_dist_init, &
        dbcsr_image_dist_release, dbcsr_init, dbcsr_nblkcols_local, dbcsr_nblkcols_total, &
        dbcsr_nblkrows_local, dbcsr_nblkrows_total, dbcsr_nfullcols_total, dbcsr_nfullrows_total, &
        dbcsr_release, dbcsr_release_locals, dbcsr_row_block_offsets, dbcsr_row_block_sizes, &
        dbcsr_valid_index
   USE dbcsr_mm_multrec,                ONLY: dbcsr_mm_multrec_finalize,&
                                              dbcsr_mm_multrec_init,&
                                              dbcsr_mm_multrec_lib_finalize,&
                                              dbcsr_mm_multrec_lib_init,&
                                              dbcsr_mm_multrec_multiply,&
                                              dbcsr_mm_multrec_phaseout,&
                                              dbcsr_mm_multrec_type
   USE dbcsr_mp_methods,                ONLY: &
        dbcsr_mp_grid_setup, dbcsr_mp_group, dbcsr_mp_has_subgroups, dbcsr_mp_my_col_group, &
        dbcsr_mp_my_row_group, dbcsr_mp_mynode, dbcsr_mp_mypcol, dbcsr_mp_myprow, dbcsr_mp_npcols, &
        dbcsr_mp_nprows, dbcsr_mp_numnodes, dbcsr_mp_pgrid
   USE dbcsr_mp_operations,             ONLY: &
        dbcsr_gatherv_any, dbcsr_ibcast_any, dbcsr_irecv_any, dbcsr_iscatterv_any, &
        dbcsr_isend_any, dbcsr_rget_any, dbcsr_sendrecv_any, dbcsr_win_create_any, &
        hybrid_alltoall_any, hybrid_alltoall_i1
   USE dbcsr_operations,                ONLY: dbcsr_add,&
                                              dbcsr_conjg,&
                                              dbcsr_copy,&
                                              dbcsr_crop_matrix,&
                                              dbcsr_filter,&
                                              dbcsr_may_be_dense,&
                                              dbcsr_scale
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size,&
                                              memory_copy
   USE dbcsr_toollib,                   ONLY: ordered_search,&
                                              uppercase
   USE dbcsr_transformations,           ONLY: dbcsr_make_dense,&
                                              dbcsr_make_dense_low,&
                                              dbcsr_make_undense,&
                                              dbcsr_make_untransposed_blocks,&
                                              dbcsr_new_transposed
   USE dbcsr_types,                     ONLY: &
        dbcsr_1d_array_type, dbcsr_2d_array_type, dbcsr_conjugate_transpose, dbcsr_data_obj, &
        dbcsr_distribution_obj, dbcsr_imagedistribution_obj, dbcsr_iterator, dbcsr_memtype_type, &
        dbcsr_meta_size, dbcsr_mp_obj, dbcsr_mpi_size_limits, dbcsr_mpi_statistics_type, &
        dbcsr_no_transpose, dbcsr_num_slots, dbcsr_obj, dbcsr_scalar_type, dbcsr_slot_blk_p, &
        dbcsr_slot_col_i, dbcsr_slot_coo_l, dbcsr_slot_home_coli, dbcsr_slot_home_pcol, &
        dbcsr_slot_home_prow, dbcsr_slot_home_rowi, dbcsr_slot_home_vpcol, dbcsr_slot_home_vprow, &
        dbcsr_slot_nblkcols_local, dbcsr_slot_nblkrows_total, dbcsr_slot_nblks, &
        dbcsr_slot_nfullcols_local, dbcsr_slot_nze, dbcsr_slot_row_p, dbcsr_slot_size, &
        dbcsr_slot_thr_c, dbcsr_transpose, dbcsr_type, dbcsr_type_antisymmetric, &
        dbcsr_type_complex_4, dbcsr_type_complex_8, dbcsr_type_int_4, dbcsr_type_no_symmetry, &
        dbcsr_type_real_4, dbcsr_type_real_8
   USE dbcsr_util,                      ONLY: count_bins,&
                                              dbcsr_checksum,&
                                              dbcsr_verify_matrix
   USE dbcsr_work_operations,           ONLY: dbcsr_add_wm_from_matrix,&
                                              dbcsr_create,&
                                              dbcsr_finalize,&
                                              dbcsr_special_finalize,&
                                              dbcsr_work_create
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8,&
                                              sp
   USE machine,                         ONLY: default_output_unit,&
                                              m_memory
   USE message_passing,                 ONLY: &
        mp_allgather, mp_alltoall, mp_bcast, mp_gather, mp_gatherv, mp_iallgather, mp_ibcast, &
        mp_irecv, mp_iscatter, mp_isend, mp_isum, mp_max, mp_min, mp_request_null, mp_rget, &
        mp_sendrecv, mp_sum, mp_testall, mp_testany, mp_wait, mp_waitall, mp_win_create, &
        mp_win_free, mp_win_lock_all, mp_win_unlock_all
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads,&
!$                    omp_set_lock, omp_unset_lock, omp_init_lock, omp_lock_kind, omp_destroy_lock

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mm_cannon'
   CHARACTER(len=*), PARAMETER, PRIVATE :: int_print = "(10(1X,I7))"
   LOGICAL, PARAMETER :: debug_mod = .FALSE.
   LOGICAL, PARAMETER :: careful_mod = .FALSE.
   REAL, PARAMETER :: huge_norm = HUGE(1.0)**(1.0/3.0)

   TYPE dbcsr_memtype_type_p
      TYPE(dbcsr_memtype_type), POINTER :: p => Null()
      ! ensure that array-elements are on different cache lines
      INTEGER(kind=int_4), DIMENSION(64)    :: padding
   END TYPE dbcsr_memtype_type_p

   TYPE(dbcsr_memtype_type_p), DIMENSION(:), POINTER, SAVE :: memtype_product_wm => Null(), &
                                                              memtype_product_wm_scatter => Null()

   TYPE(dbcsr_mpi_statistics_type), PRIVATE, SAVE :: dbcsr_mpi_statistics
   INTEGER, PRIVATE, SAVE :: max_nblocks = 0
   REAL, PRIVATE, SAVE :: marketing_flops = 0
   REAL, PRIVATE, SAVE :: max_memory = 0

   INTEGER, PRIVATE, SAVE :: last_mpi_ranks_used = 0

   TYPE(dbcsr_memtype_type), PRIVATE, SAVE  :: memtype_abpanel_1, memtype_abpanel_2, &
                                               memtype_trsbuffer_1, memtype_trsbuffer_2, &
                                               memtype_mpi_buffer
   TYPE(acc_stream_type), PRIVATE, SAVE         :: stream_1, stream_2
   ! ab-panels and streams are shared between all threads

   TYPE dbcsr_buffer
      TYPE(dbcsr_data_obj)               :: DATA, data_diag
      INTEGER                            :: vprow, vpcol
      INTEGER                            :: data_win, meta_win
      INTEGER, DIMENSION(:), POINTER     :: meta => Null(), &
                                            meta_diag => Null()
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: get_requests, &
                                               get_requests_map, &
                                               offset
      INTEGER                            :: nrequests
      TYPE(dbcsr_1d_array_type)          :: buffer
      LOGICAL                            :: is_valid = .FALSE.
   END TYPE dbcsr_buffer

   TYPE dbcsr_buffers
      TYPE(dbcsr_buffer)             :: left, right
      TYPE(dbcsr_data_obj)           :: trs_stackbuf
      LOGICAL                        :: has_rma_left_win = .FALSE., &
                                        has_rma_right_win = .FALSE.
   END TYPE dbcsr_buffers

   ! Buffers
   TYPE(dbcsr_buffers), PRIVATE, TARGET, SAVE :: buffers_orig, &
                                                 buffers_1, buffers_2

   INTEGER, PARAMETER, PRIVATE               :: idata_displ = 3, &
                                                idata_size = 1, &
                                                imeta_displ = 4, &
                                                imeta_size = 2, &
                                                ilocal_proc = 1, &
                                                isym_proc = 2

   TYPE(dbcsr_data_obj), SAVE, PRIVATE :: local_data_scatter, local_data_product_scatter
   INTEGER, PRIVATE                    :: local_meta_size_scatter, local_data_size_scatter
   INTEGER, DIMENSION(:), SAVE, PRIVATE, POINTER :: meta_scatter => Null(), local_meta_scatter => Null(), &
                                                    local_meta_product_scatter => Null()
   INTEGER, ALLOCATABLE, DIMENSION(:), PRIVATE :: left_total_row_counts
   INTEGER, ALLOCATABLE, DIMENSION(:), TARGET, PRIVATE :: left_no_empty_images, &
                                                          left_no_empty_images_displ, &
                                                          right_no_empty_images, &
                                                          right_no_empty_images_displ, &
                                                          left_refs_data_size, &
                                                          right_refs_data_size, &
                                                          left_local_refs_data_size, &
                                                          right_local_refs_data_size
   INTEGER, ALLOCATABLE, DIMENSION(:, :), TARGET, PRIVATE  :: left_refs_displ_unmerged, &
                                                              right_refs_displ_unmerged, &
                                                              left_local_refs_displ_unmerged, &
                                                              right_local_refs_displ_unmerged, &
                                                              left_refs_meta_size, &
                                                              right_refs_meta_size, &
                                                              left_local_refs_meta_size, &
                                                              right_local_refs_meta_size
   REAL(kind=sp), ALLOCATABLE, DIMENSION(:), TARGET, PRIVATE :: left_max_norms, right_max_norms, &
                                                                left_local_max_norms, right_local_max_norms

   INTEGER, PRIVATE                :: request_count_rows
   INTEGER, DIMENSION(8), PRIVATE  :: requests
   INTEGER, DIMENSION(4), PRIVATE  :: requests_diag
   INTEGER, DIMENSION(2), PRIVATE  :: requests_scatter
   LOGICAL, PRIVATE                :: my_use_mpi_exp

   TYPE dbcsr_mm_multrec_type_p
      TYPE(dbcsr_mm_multrec_type), POINTER :: p => Null()
      ! ensure that array-elements are on different cache lines
      INTEGER(kind=int_4), DIMENSION(64)       :: padding
   END TYPE dbcsr_mm_multrec_type_p

   INTERFACE dbcsr_switch
      MODULE PROCEDURE dbcsr_switch_sets
      MODULE PROCEDURE dbcsr_switch_d_ptrs
   END INTERFACE

   INTERFACE setup_rec_index
      MODULE PROCEDURE setup_rec_index_images
      MODULE PROCEDURE setup_rec_index_2d
   END INTERFACE

   PUBLIC :: dbcsr_mm_cannon_lib_init, dbcsr_mm_cannon_lib_finalize
   PUBLIC :: dbcsr_mm_cannon_clear_mempools
   PUBLIC :: dbcsr_mm_cannon_multiply

CONTAINS

! **************************************************************************************************
!> \brief Initialize the library
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_cannon_lib_init()

      INTEGER                                            :: ithread, nthreads

      nthreads = 1; ithread = 0
!$    nthreads = OMP_GET_NUM_THREADS(); ithread = OMP_GET_THREAD_NUM()

      CALL dbcsr_mm_multrec_lib_init()

!$OMP     MASTER
      dbcsr_mpi_statistics%nimages = -1
      dbcsr_mpi_statistics%nexchanged = 0
      dbcsr_mpi_statistics%nfiltered = 0
      dbcsr_mpi_statistics%data_size = 0
      dbcsr_mpi_statistics%data_size(:, 2) = HUGE(dbcsr_mpi_statistics%data_size(1, 2))
      dbcsr_mpi_statistics%data_size_breakdown = 0

      marketing_flops = 0
      max_memory = 0
      ALLOCATE (memtype_product_wm(0:nthreads-1))
      ALLOCATE (memtype_product_wm_scatter(0:nthreads-1))
!$OMP     END MASTER
!$OMP     BARRIER

      ! Each thread has its own working-matrix and its own mempool
      ALLOCATE (memtype_product_wm(ithread)%p)
      CALL dbcsr_memtype_setup(memtype_product_wm(ithread)%p, has_pool=.TRUE.)
      CALL dbcsr_mempool_ensure_capacity(memtype_product_wm(ithread)%p%pool, capacity=1)
      ALLOCATE (memtype_product_wm_scatter(ithread)%p)
      CALL dbcsr_memtype_setup(memtype_product_wm_scatter(ithread)%p, has_pool=.TRUE.)
      CALL dbcsr_mempool_ensure_capacity(memtype_product_wm_scatter(ithread)%p%pool, capacity=1)
   END SUBROUTINE dbcsr_mm_cannon_lib_init

! **************************************************************************************************
!> \brief Finalize the library
!> \param group ...
!> \param output_unit ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_cannon_lib_finalize(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_cannon_lib_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilimit, isqrt, isqrt2, ithread
      INTEGER(KIND=int_8)                                :: total_nexchanged, total_nfiltered
      INTEGER(KIND=int_8), &
         DIMENSION(SIZE(dbcsr_mpi_size_limits)+1, 2, 2)  :: total_recv_breakdown
      REAL                                               :: average, total_marketing_flops, &
                                                            total_max_memory
      REAL, DIMENSION(2)                                 :: max_recv_data, min_recv_data, &
                                                            total_recv_data

      CALL dbcsr_mm_multrec_lib_finalize(group, output_unit)

      ithread = 0
!$    ithread = omp_get_thread_num()

      ! Each thread has its own working-matrix and its own mempool
      IF (ASSOCIATED(memtype_product_wm(ithread)%p%pool)) &
         CALL dbcsr_mempool_destruct(memtype_product_wm(ithread)%p%pool)
      DEALLOCATE (memtype_product_wm(ithread)%p)
      IF (ASSOCIATED(memtype_product_wm_scatter(ithread)%p%pool)) &
         CALL dbcsr_mempool_destruct(memtype_product_wm_scatter(ithread)%p%pool)
      DEALLOCATE (memtype_product_wm_scatter(ithread)%p)

!$OMP      BARRIER
!$OMP      MASTER
      DEALLOCATE (memtype_product_wm)
      DEALLOCATE (memtype_product_wm_scatter)

      ! Deallocate buffers
      CALL buffers_release(buffers_orig)
      CALL buffers_release(buffers_1)
      CALL buffers_release(buffers_2)

      total_max_memory = max_memory
      CALL mp_max(total_max_memory, group)

      total_marketing_flops = marketing_flops
      CALL mp_sum(total_marketing_flops, group)

      total_nexchanged = dbcsr_mpi_statistics%nexchanged
      CALL mp_sum(total_nexchanged, group)

      total_nfiltered = dbcsr_mpi_statistics%nfiltered
      CALL mp_sum(total_nfiltered, group)

      total_recv_data(:) = dbcsr_mpi_statistics%data_size(:, 1)
      CALL mp_sum(total_recv_data, group)

      min_recv_data(:) = dbcsr_mpi_statistics%data_size(:, 2)
      CALL mp_min(min_recv_data, group)

      max_recv_data(:) = dbcsr_mpi_statistics%data_size(:, 3)
      CALL mp_max(max_recv_data, group)

      IF (dbcsr_mpi_statistics%nexchanged .GT. 0) THEN
         average = SUM(total_recv_data(:))/REAL(total_nexchanged)
      ELSE
         average = 0
         min_recv_data = 0
      ENDIF

      total_recv_breakdown(:, :, :) = dbcsr_mpi_statistics%data_size_breakdown(:, :, :)
      CALL mp_sum(total_recv_breakdown, group)

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(A,T30,EN20.6)') " marketing flops", total_marketing_flops

         IF (dbcsr_mpi_statistics%nimages .GT. 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("-", 79)
            WRITE (output_unit, '(A,T30,EN20.6)') " max memory usage/rank", total_max_memory
            WRITE (output_unit, '(A,T30,I20)') " # max total images/rank", dbcsr_mpi_statistics%nimages
            WRITE (output_unit, '(A,T30,I20)') " # MPI messages exchanged", total_nexchanged
            WRITE (output_unit, '(A,T30,I20)') " # MPI messages filtered", total_nfiltered
            WRITE (output_unit, '(A)') " MPI messages size (elements):"
            WRITE (output_unit, '(A,T30,EN20.6)') "  total size", &
               SUM(total_recv_data(:))
            WRITE (output_unit, '(A,T30,EN20.6)') "  min size", &
               MINVAL(min_recv_data(:))
            WRITE (output_unit, '(A,T30,EN20.6)') "  max size", &
               MAXVAL(max_recv_data(:))
            WRITE (output_unit, '(A,T30,EN20.6)') "  average size", average

            WRITE (output_unit, '(A)') " MPI breakdown and total messages size (bytes):"
            WRITE (output_unit, '(A,I8,T40,I10,T55,I20)') "             size <= ", dbcsr_mpi_size_limits(1), &
               SUM(total_recv_breakdown(1, 1, :)), SUM(total_recv_breakdown(1, 2, :))
            DO ilimit = 2, SIZE(dbcsr_mpi_size_limits)
               WRITE (output_unit, '(A,I8,A,I8,T40,I10,T55,I20)') "  ", dbcsr_mpi_size_limits(ilimit-1), &
                  " < size <= ", dbcsr_mpi_size_limits(ilimit), &
                  SUM(total_recv_breakdown(ilimit, 1, :)), SUM(total_recv_breakdown(ilimit, 2, :))
            ENDDO
            ilimit = SIZE(dbcsr_mpi_size_limits)
            WRITE (output_unit, '(A,I8,A,T40,I10,T55,I20)') "  ", dbcsr_mpi_size_limits(ilimit), &
               " < size    ", SUM(total_recv_breakdown(ilimit+1, 1, :)), SUM(total_recv_breakdown(ilimit+1, 2, :))
         ENDIF

         isqrt = NINT(SQRT(REAL(last_mpi_ranks_used, KIND=real_8)))
         isqrt2 = NINT(SQRT(REAL(last_mpi_ranks_used*2, KIND=real_8)))
         IF (isqrt*isqrt .NE. last_mpi_ranks_used) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("-", 79)
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "Warning: using a non-square number of MPI ranks might lead to poor performance."
            WRITE (UNIT=output_unit, FMT="(T2,A,I0)") &
               "         used ranks: ", last_mpi_ranks_used
            WRITE (UNIT=output_unit, FMT="(T2,A,2(I0,1X))") &
               "         suggested : ", isqrt**2, isqrt2**2
         ENDIF
      ENDIF
      IF (ASSOCIATED(memtype_trsbuffer_1%pool)) &
         CALL dbcsr_mempool_destruct(memtype_trsbuffer_1%pool)
      IF (ASSOCIATED(memtype_trsbuffer_2%pool)) &
         CALL dbcsr_mempool_destruct(memtype_trsbuffer_2%pool)
      IF (ASSOCIATED(memtype_abpanel_1%pool)) &
         CALL dbcsr_mempool_destruct(memtype_abpanel_1%pool)
      IF (ASSOCIATED(memtype_abpanel_2%pool)) &
         CALL dbcsr_mempool_destruct(memtype_abpanel_2%pool)
      IF (acc_stream_associated(stream_1)) &
         CALL acc_stream_destroy(stream_1)
      IF (acc_stream_associated(stream_2)) &
         CALL acc_stream_destroy(stream_2)
!$OMP      END MASTER
   END SUBROUTINE dbcsr_mm_cannon_lib_finalize

! **************************************************************************************************
!> \brief Deallocate memory contained in mempools
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_cannon_clear_mempools()

      INTEGER                                            :: ithread

      ithread = 0
!$    ithread = omp_get_thread_num()

      ! Each thread has its own working-matrix and its own mempool
      IF (ASSOCIATED(memtype_product_wm(ithread)%p%pool)) &
         CALL dbcsr_mempool_clear(memtype_product_wm(ithread)%p%pool)
      IF (ASSOCIATED(memtype_product_wm_scatter(ithread)%p%pool)) &
         CALL dbcsr_mempool_clear(memtype_product_wm_scatter(ithread)%p%pool)

!$OMP      MASTER
      IF (ASSOCIATED(memtype_trsbuffer_1%pool)) &
         CALL dbcsr_mempool_clear(memtype_trsbuffer_1%pool)
      IF (ASSOCIATED(memtype_trsbuffer_2%pool)) &
         CALL dbcsr_mempool_clear(memtype_trsbuffer_2%pool)
      IF (ASSOCIATED(memtype_abpanel_1%pool)) &
         CALL dbcsr_mempool_clear(memtype_abpanel_1%pool)
      IF (ASSOCIATED(memtype_abpanel_2%pool)) &
         CALL dbcsr_mempool_clear(memtype_abpanel_2%pool)
!$OMP      END MASTER
   END SUBROUTINE dbcsr_mm_cannon_clear_mempools

! **************************************************************************************************
!> \brief Performs a multiplication of two dbcsr_type matrices,
!>        as  C := alpha * op( A ) * op( B ) + beta * C.
!> \param[in] transa specifies the form of op( A ) to be used in
!>                            the matrix multiplication
!>                            transa = 'N' or 'n',  op( A ) = A.
!>                            transa = 'T' or 't',  op( A ) = transpose(A).
!>                            transa = 'C' or 'c',  op( A ) = transpose(conjg(A)).
!> \param[in] transb specifies the form of op( B ) to be used in
!>                            the matrix multiplication
!>                            transb = 'N' or 'n',  op( B ) = B.
!>                            transb = 'T' or 't',  op( B ) = transpose(B).
!>                            transb = 'C' or 'c',  op( B ) = transpose(conjg(B)).
!> \param[in] alpha           scaling of product
!> \param[in] matrix_a        left BCSR matrix
!> \param[in] matrix_b        right BCSR matrix
!> \param[in] beta            scaling of existing data
!> \param[out] matrix_c       resulting BCSR product matrix.
!> \param[in] first_row       (optional) first full row of limiting submatrix
!> \param[in] last_row        (optional) last full row of limiting submatrix
!> \param[in] first_column    (optional) first full column of limiting submatrix
!> \param[in] last_column     (optional) last full column of limiting submatrix
!> \param[in] first_k         (optional) first full column of limiting inner
!>                            product
!> \param[in] last_k          (optional) last full column of limiting inner
!>                            product
!> \param[in] retain_sparsity (optional) enforce the sparsity pattern of the
!>                            existing product matrix; default is no
!> \param[in] filter_eps      Filtering of the matrix
!> \param[out] flop           (optional) effective flop
!> \par Matrices m_a and m_b are multiplied into the m_c product matrix. If the
!>      dist2d parameter is not specified, then a new distribution_2d is
!>      determined for it.
!> \par Non-equal column dimensions of the right and product matrices
!>      The right and product matrix are allowed to have different
!>      (full) column dimensions. If they differ, there are certain
!>      peculiar behaviors, then the last_column is effectively set to
!>      the minimal of the two.
!> \par Beta scaling of the right product matrix
!>      If the effective last_column is less than the full column
!>      dimension of the product matrix, then the scaling of the
!>      product matrix with beta is limited to the submatrix specified
!>      by last_column.
!> \par Filtering
!>      The filter_eps parameter, if present, is used to filter the
!>      resulting matrix.  The filtering criterion is whether the
!>      block-frobenius norm is less than the specified epsilon.
!>      One-the-fly filtering is done such that individual
!>      multiplications are skipped if the product of the frobenius
!>      norms of the left- and right-matrix blocks are less than the
!>      specified epsilon divided by the maximum number of possible
!>      multiplies in each row.  In addition a final filtering is done
!>      as well with the same epsilon value.
! **************************************************************************************************
   SUBROUTINE dbcsr_mm_cannon_multiply(transa, transb, &
                                       alpha, matrix_a, matrix_b, beta, matrix_c, &
                                       first_row, last_row, first_column, last_column, first_k, last_k, &
                                       retain_sparsity, filter_eps, &
                                       flop)

      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: alpha
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: beta
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: first_row, last_row, first_column, &
                                                            last_column, first_k, last_k
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(KIND=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT), OPTIONAL         :: flop

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mm_cannon_multiply', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE., &
                                                            use_list_indexing = .TRUE., &
                                                            use_local_indexing = .TRUE.
      REAL(real_8), PARAMETER                            :: make_dense_occ_thresh = 1.0_dp

      CHARACTER                                          :: transa_l, transb_l
      INTEGER :: comm, f_col, f_k, f_row, handle, handle2, ithread, l_col, l_k, l_row, &
         nimages_left_rows, nimages_match, nimages_right_cols, numnodes, output_unit
      INTEGER(KIND=int_8)                                :: my_flop
      LOGICAL :: ab_dense, has_clusters, keep_product_data, keep_sparsity, new_left, new_right, &
         product_reindex, release_tdist, use_dense_mult
      REAL(KIND=dp)                                      :: cs
      TYPE(array_i1d_obj) :: dense_col_sizes, dense_k_sizes, dense_row_sizes, k_vmap, m_map, &
         n_map, old_product_col_blk_offsets, old_product_col_blk_sizes, &
         old_product_row_blk_offsets, old_product_row_blk_sizes
      TYPE(dbcsr_2d_array_type), POINTER                 :: m2s_left, m2s_right
      TYPE(dbcsr_distribution_obj)                       :: dense_product_distribution, &
                                                            old_product_distribution
      TYPE(dbcsr_imagedistribution_obj)                  :: dense_rdist_left, dense_rdist_right, &
                                                            rdist_left, rdist_right
      TYPE(dbcsr_obj)                                    :: matrix_left, matrix_right, product_matrix
      TYPE(dbcsr_scalar_type)                            :: eps_any

      CALL timeset(routineN, handle)
      CALL array_nullify(dense_k_sizes)
      CALL array_nullify(dense_col_sizes)
      CALL array_nullify(dense_row_sizes)

      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()

      ! setup driver-dependent memory-types and their memory-pools ---------------

      ! the ab_buffers are shared by all threads
      IF (has_acc) THEN
         IF (.NOT. acc_stream_associated(stream_1)) THEN
            CALL acc_stream_create(stream_1, "MemCpy (odd ticks)")
            CALL acc_stream_create(stream_2, "MemCpy (even ticks)")
         ENDIF

         CALL dbcsr_memtype_setup(memtype_abpanel_1, has_pool=.TRUE., &
                                  acc_hostalloc=.TRUE., acc_devalloc=.TRUE., acc_stream=stream_1, &
                                  mpi=.TRUE., oversize_factor=default_resize_factor)

         CALL dbcsr_memtype_setup(memtype_abpanel_2, has_pool=.TRUE., &
                                  acc_hostalloc=.TRUE., acc_devalloc=.TRUE., acc_stream=stream_2, &
                                  mpi=.TRUE., oversize_factor=default_resize_factor)

         !TODO: ensure capacity 2/3?
         CALL dbcsr_memtype_setup(memtype_trsbuffer_1, has_pool=.TRUE., &
                                  acc_hostalloc=.TRUE., acc_devalloc=.TRUE., acc_stream=stream_1)
         CALL dbcsr_mempool_ensure_capacity(memtype_trsbuffer_1%pool, capacity=1)

         CALL dbcsr_memtype_setup(memtype_trsbuffer_2, has_pool=.TRUE., &
                                  acc_hostalloc=.TRUE., acc_devalloc=.TRUE., acc_stream=stream_2)
         CALL dbcsr_mempool_ensure_capacity(memtype_trsbuffer_2%pool, capacity=1)
      ELSE
         CALL dbcsr_memtype_setup(memtype_abpanel_1, mpi=.TRUE.)
         CALL dbcsr_memtype_setup(memtype_abpanel_2, mpi=.TRUE.)
      ENDIF

      CALL dbcsr_memtype_setup(memtype_mpi_buffer, mpi=.TRUE.)

      ! check parameters ---------------------------------------------------------
      transa_l = transa
      transb_l = transb
      CALL uppercase(transa_l)
      CALL uppercase(transb_l)
      IF (transa_l .NE. dbcsr_no_transpose .AND. &
          transa_l .NE. dbcsr_transpose .AND. &
          transa_l .NE. dbcsr_conjugate_transpose) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",624,"Invalid transa_l = "//transa_l)

      IF (transb_l .NE. dbcsr_no_transpose .AND. &
          transb_l .NE. dbcsr_transpose .AND. &
          transb_l .NE. dbcsr_conjugate_transpose) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",629,"Invalid transb_l = "//transb_l)

      IF (dbg) THEN
         WRITE (*, *) '========== MULTIPLICATION ========================'
         CALL dbcsr_verify_matrix(matrix_a)
         CALL dbcsr_verify_matrix(matrix_b)
         CALL dbcsr_verify_matrix(matrix_c)
         WRITE (*, *) routineN//" ABC checksums", &
            dbcsr_checksum(matrix_a), &
            dbcsr_checksum(matrix_b), &
            dbcsr_checksum(matrix_c)
         IF (dbg) THEN
            CALL dbcsr_print(matrix_a, nodata=.TRUE.)
            CALL dbcsr_print(matrix_b, nodata=.TRUE.)
            CALL dbcsr_print(matrix_c, nodata=.TRUE.)
         ENDIF
      ENDIF

      ! transpose/conjg left and/or right matrices if needed
      SELECT CASE (transa_l)
      CASE (dbcsr_no_transpose)
         matrix_left = matrix_a
         new_left = .FALSE.
      CASE (dbcsr_transpose)
         CALL dbcsr_init(matrix_left)
         IF (dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric) THEN
            !
            ! For antisymmetric matrix, we need to do a hard copy
            ! shallow_data_copy=.TRUE. doesnt handle properly antisymm matrices
            CALL dbcsr_new_transposed(matrix_left, matrix_a, &
                                      shallow_data_copy=.FALSE., redistribute=.FALSE., &
                                      transpose_distribution=.FALSE.)
         ELSE
            CALL dbcsr_new_transposed(matrix_left, matrix_a, &
                                      shallow_data_copy=.TRUE., redistribute=.FALSE., &
                                      transpose_distribution=.FALSE.)
         ENDIF
         new_left = .TRUE.
      CASE (dbcsr_conjugate_transpose)
         CALL dbcsr_init(matrix_left)
         CALL dbcsr_new_transposed(matrix_left, matrix_a, &
                                   shallow_data_copy=.FALSE., redistribute=.FALSE., &
                                   transpose_distribution=.FALSE.)
         CALL dbcsr_conjg(matrix_left)
         new_left = .TRUE.
      CASE DEFAULT
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",675,"wrong transa_l = "//transa_l)
      END SELECT

      SELECT CASE (transb_l)
      CASE (dbcsr_no_transpose)
         matrix_right = matrix_b
         new_right = .FALSE.
      CASE (dbcsr_transpose)
         CALL dbcsr_init(matrix_right)
         IF (dbcsr_get_matrix_type(matrix_b) .EQ. dbcsr_type_antisymmetric) THEN
            !
            ! For antisymmetric matrix, we need to do a hard copy
            ! shallow_data_copy=.TRUE. doesnt handle properly antisymm matrices
            CALL dbcsr_new_transposed(matrix_right, matrix_b, &
                                      shallow_data_copy=.FALSE., redistribute=.FALSE., &
                                      transpose_distribution=.FALSE.)
         ELSE
            CALL dbcsr_new_transposed(matrix_right, matrix_b, &
                                      shallow_data_copy=.TRUE., redistribute=.FALSE., &
                                      transpose_distribution=.FALSE.)
         ENDIF
         new_right = .TRUE.
      CASE (dbcsr_conjugate_transpose)
         CALL dbcsr_init(matrix_right)
         CALL dbcsr_new_transposed(matrix_right, matrix_b, &
                                   shallow_data_copy=.FALSE., redistribute=.FALSE., &
                                   transpose_distribution=.FALSE.)
         CALL dbcsr_conjg(matrix_right)
         new_right = .TRUE.
      CASE DEFAULT
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",705,"wrong transb_l = "//transb_l)
      END SELECT
      !
      ! Ensure matrix compatibility.
      IF (.NOT. array_equality(dbcsr_row_block_offsets(matrix_c), dbcsr_row_block_offsets(matrix_left))) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",710,"C/A rows not equal")
      IF (.NOT. array_equality(dbcsr_col_block_offsets(matrix_c), dbcsr_col_block_offsets(matrix_right))) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",712,"C/B columns not equal")
      IF (.NOT. array_equality(dbcsr_col_block_offsets(matrix_left), dbcsr_row_block_offsets(matrix_right))) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",714,"A cols/B rows not equal")
      !
      ! No dense multiplication when filtering is used.
      use_dense_mult = .NOT. PRESENT(filter_eps)
      IF (has_acc) use_dense_mult = .FALSE.
      !
      ! Temporary check for experimental MPI algorithms
      IF (new_right .OR. new_left .OR. &
          dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_c%m%dist)) .NE. &
          dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_c%m%dist))) THEN
         my_use_mpi_exp = .FALSE.
      ELSE
         my_use_mpi_exp = use_mpi_exp
      ENDIF
      !
      ! No dense multiplication and local indexing when 2D clustering is used.
      IF (my_use_mpi_exp) THEN
         use_dense_mult = .FALSE.
      ENDIF
      ! we skip dense multiply for (anti)symmetric matrices (slowdown for S/H * C)
      IF (use_dense_mult) THEN
         IF (dbcsr_has_symmetry(matrix_left) .OR. &
             dbcsr_has_symmetry(matrix_right)) THEN
            use_dense_mult = .FALSE.
         ELSE
            use_dense_mult = dbcsr_may_be_dense(matrix_left, make_dense_occ_thresh) &
                             .AND. dbcsr_may_be_dense(matrix_right, make_dense_occ_thresh)
         ENDIF
      ENDIF
      ab_dense = use_dense_mult
      !
      ! Submatrix selection
      f_row = 1
      l_row = dbcsr_nfullrows_total(matrix_c)
      f_col = 1
      l_col = dbcsr_nfullcols_total(matrix_c)
      f_k = 1
      l_k = dbcsr_nfullcols_total(matrix_left)
      IF (PRESENT(first_row)) THEN
         IF (first_row .LT. 1 .OR. first_row .GT. dbcsr_nfullrows_total(matrix_c)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",754,"Invalid first row specified")
         f_row = first_row
      ENDIF
      IF (PRESENT(last_row)) THEN
         IF (last_row .GT. dbcsr_nfullrows_total(matrix_c)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",759,"Invalid last row specified")
         l_row = last_row
      ENDIF
      IF (PRESENT(first_column)) THEN
         IF (first_column .LT. 1 .OR. first_column .GT. dbcsr_nfullcols_total(matrix_c)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",764,"Invalid first col specified")
         f_col = first_column
      ENDIF
      IF (PRESENT(last_column)) THEN
         IF (last_column .GT. dbcsr_nfullcols_total(matrix_c)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",769,"Invalid last column specified (C)")
         IF (last_column .GT. dbcsr_nfullcols_total(matrix_right)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",771,"Invalid last column specified (B)")
         l_col = last_column
      ENDIF
      IF (PRESENT(first_k)) THEN
         IF (first_k .LT. 1 .OR. first_k .GT. dbcsr_nfullcols_total(matrix_left)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",776,"Invalid first k specified (A)")
         f_k = first_k
      ENDIF
      IF (PRESENT(last_k)) THEN
         IF (last_k .GT. dbcsr_nfullcols_total(matrix_left)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",781,"Invalid last k specified (A)")
         l_k = last_k
      ENDIF
      !
      ! update statistics (we count marketing flops per MPI rank)
      numnodes = dbcsr_mp_numnodes(dbcsr_distribution_mp(dbcsr_distribution(matrix_c)))
      last_mpi_ranks_used = numnodes
      marketing_flops = marketing_flops+ &
                        (2.0*(l_row-f_row+1.0)*(l_col-f_col+1.0)/numnodes)*(l_k-f_k+1.0)
      !
      ! Now optimize the default submatrix selection values away
      IF (f_row .EQ. 1) f_row = 0
      IF (l_row .EQ. dbcsr_nfullrows_total(matrix_left)) l_row = 0
      IF (f_col .EQ. 1) f_col = 0
      ! The last column must be set if the right and product matrices
      ! differ.
      l_col = MIN(l_col, dbcsr_nfullcols_total(matrix_right))
      l_col = MIN(l_col, dbcsr_nfullcols_total(matrix_c))
      IF (f_col .LE. 1 .AND. &
          l_col .EQ. dbcsr_nfullcols_total(matrix_right) .AND. &
          dbcsr_nfullcols_total(matrix_right) .EQ. &
          dbcsr_nfullcols_total(matrix_c)) l_col = 0
      IF (f_k .EQ. 1) f_k = 0
      IF (l_k .EQ. dbcsr_nfullcols_total(matrix_left)) l_k = 0
      IF (.NOT. PRESENT(last_column) .AND. &
          .NOT. array_equality(dbcsr_col_block_sizes(matrix_right), &
                               dbcsr_col_block_sizes(matrix_c))) THEN
         l_col = MIN(dbcsr_nfullcols_total(matrix_right), &
                     dbcsr_nfullcols_total(matrix_c))
      ENDIF
      IF (f_row .GT. l_row) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",812,"Last row smaller than first row")
      IF (f_col .GT. l_col) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",814,"Last col smaller than first col")
      !
      ! Product data needs to be retained when
      ! * beta != 0; or
      ! * there is column limiting (l_col > 0) and the limiting column
      !   is less than the number of full columns in theproduct matrix
      keep_sparsity = .FALSE.
      IF (PRESENT(retain_sparsity)) keep_sparsity = retain_sparsity
      !
      keep_product_data = keep_sparsity &
                          .OR. .NOT. dbcsr_scalar_are_equal(beta, dbcsr_scalar_zero(beta%data_type)) &
                          .OR. (l_col .GT. 0 .AND. l_col .LT. dbcsr_nfullcols_total(matrix_c)) &
                          .OR. (l_row .GT. 0 .AND. l_row .LT. dbcsr_nfullrows_total(matrix_c))
      !
      IF (.NOT. dbcsr_scalar_are_equal(beta, dbcsr_scalar_one(beta%data_type)) .AND. keep_product_data) THEN
         CALL dbcsr_scale(matrix_c, alpha_scalar=beta, &
                          limits=(/f_row, l_row, f_col, l_col/))
      ENDIF
      !
      ! The index of the product matrix is twiddled into canonical form
      ! if it is (anti)symmetric (i.e., rows and columns are where the
      ! row/column distributions say they are). Doing this in advance
      ! makes the local multiply more efficient.
      IF (dbcsr_has_symmetry(matrix_c)) THEN
         product_reindex = .TRUE.
      ELSE
         product_reindex = .FALSE.
      ENDIF
      ! Product can not be made dense; however, A & B may still be made
      ! dense unless previously determined otherwise.
      IF (product_reindex .OR. keep_sparsity) THEN
         use_dense_mult = .FALSE.
      ENDIF
      !
      ! The thread distribution must reflect the current (possibly
      ! dense) distribution
      IF (.NOT. dbcsr_distribution_has_threads(matrix_c%m%dist)) THEN
         release_tdist = .TRUE.
         CALL dbcsr_distribution_make_threads(matrix_c%m%dist)
      ELSE
         release_tdist = .FALSE.
      ENDIF
      !
      ! Compute number of images (rows and columns)
      ! By construction clusters require RMA algo
      has_clusters = .FALSE.
      IF (my_use_mpi_exp .AND. dbcsr_distribution_has_row_clusters(matrix_left%m%dist)) THEN
         has_clusters = .TRUE.
         nimages_left_rows = dbcsr_distribution_num_local_row_clusters(matrix_left%m%dist)* &
                             dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_left%m%dist))
      ELSE
         ! 1D clustering (only RMA support 2D clustering)
         nimages_left_rows = dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_left%m%dist))
      ENDIF
      !
      IF (my_use_mpi_exp .AND. dbcsr_distribution_has_col_clusters(matrix_left%m%dist)) THEN
         has_clusters = .TRUE.
         nimages_match = dbcsr_distribution_num_local_col_clusters(matrix_left%m%dist)* &
                         dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_left%m%dist))
      ELSE IF (my_use_mpi_exp .AND. dbcsr_distribution_has_row_clusters(matrix_right%m%dist)) THEN
         has_clusters = .TRUE.
         nimages_match = dbcsr_distribution_num_local_row_clusters(matrix_right%m%dist)* &
                         dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_right%m%dist))
      ELSE
         ! 1D clustering
         nimages_match = dbcsr_distribution_get_num_images_1d( &
                         dbcsr_nfullcols_total(matrix_left), &
                         dbcsr_nblkcols_total(matrix_left), &
                         dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_left%m%dist)), &
                         dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_left%m%dist)))
      ENDIF
      !
      IF (my_use_mpi_exp .AND. dbcsr_distribution_has_col_clusters(matrix_right%m%dist)) THEN
         has_clusters = .TRUE.
         nimages_right_cols = dbcsr_distribution_num_local_col_clusters(matrix_right%m%dist)* &
                              dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_right%m%dist))
      ELSE
         ! 1D clustering for the old algorithm (only RMA support 2D clustering)
         nimages_right_cols = dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_right%m%dist))
      ENDIF
      !
      ! Create imaged distributions for the multiply.
      CALL dbcsr_create_image_dist(rdist_right, matrix_right%m%dist, &
                                   match_row_nbins=dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_left%m%dist)), &
                                   match_col_nbins=dbcsr_mp_npcols(dbcsr_distribution_mp(matrix_c%m%dist)), &
                                   match_col_pdist=dbcsr_distribution_col_dist(matrix_c%m%dist), &
                                   nimages_rows=nimages_match, &
                                   nimages_cols=nimages_right_cols)
      !
      CALL dbcsr_create_image_dist(rdist_left, matrix_left%m%dist, &
                                   match_row_pdist=dbcsr_distribution_row_dist(matrix_c%m%dist), &
                                   match_row_nbins=dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_c%m%dist)), &
                                   match_col_pdist=dbcsr_distribution_row_dist(rdist_right%i%main), &
                                   match_col_idist=array_data(rdist_right%i%row_image), &
                                   match_col_nbins=dbcsr_mp_nprows(dbcsr_distribution_mp(matrix_right%m%dist)), &
                                   nimages_rows=nimages_left_rows, &
                                   nimages_cols=nimages_match)
      !
      IF (ab_dense) THEN
         CALL dbcsr_make_dists_dense(dbcsr_distribution(matrix_c), &
                                     rdist_left, rdist_right, dense_product_distribution, &
                                     dense_rdist_left, dense_rdist_right,.NOT. use_dense_mult, &
                                     m_map, k_vmap, n_map, matrix_c%m%row_blk_size)
         CALL make_sizes_dense(matrix_c%m%row_blk_size, m_map, &
                               dbcsr_distribution_nrows(dense_product_distribution), &
                               dense_row_sizes)
         CALL make_sizes_dense(matrix_c%m%col_blk_size, n_map, &
                               dbcsr_distribution_ncols(dense_product_distribution), &
                               dense_col_sizes)
         CALL make_sizes_dense(matrix_right%m%row_blk_size, k_vmap, &
                               dbcsr_distribution_nrows(dense_rdist_right%i%main), &
                               dense_k_sizes)
      ENDIF
      !
      IF (use_dense_mult .AND. .NOT. ab_dense) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",929,"Wrong logic when making dense matrices.")
      IF (use_dense_mult) THEN
         old_product_row_blk_offsets = matrix_c%m%row_blk_offset
         old_product_col_blk_offsets = matrix_c%m%col_blk_offset
         old_product_row_blk_sizes = matrix_c%m%row_blk_size
         old_product_col_blk_sizes = matrix_c%m%col_blk_size
         CALL array_hold(old_product_row_blk_offsets)
         CALL array_hold(old_product_col_blk_offsets)
         CALL array_hold(old_product_row_blk_sizes)
         CALL array_hold(old_product_col_blk_sizes)
         old_product_distribution = dbcsr_distribution(matrix_c)
         CALL dbcsr_distribution_hold(old_product_distribution)
         CALL dbcsr_init(product_matrix)
         CALL dbcsr_make_dense(matrix_c, product_matrix, &
                               dense_product_distribution, &
                               dense_row_sizes, dense_col_sizes, &
                               m_map, n_map)
      ELSE
         CALL dbcsr_init(product_matrix)
         CALL dbcsr_copy(product_matrix, matrix_c, shallow_data=.TRUE.)
      ENDIF
      IF (ab_dense) THEN
         CALL dbcsr_distribution_release(dense_product_distribution)
      ENDIF
      !
      ! This is needed to build the hash tables because they are
      ! locally indexed.
      CALL dbcsr_reset_locals(product_matrix)
      !
      IF (debug_mod) THEN
         WRITE (*, *) routineN//" Matrices ", dbcsr_get_matrix_type(matrix_a), &
            dbcsr_get_matrix_type(matrix_b), dbcsr_get_matrix_type(matrix_c)
         WRITE (*, *) routineN//" Matrices ", transa_l, transb_l, "keep", keep_product_data
      ENDIF
      IF (keep_product_data) THEN
         IF (product_reindex) THEN
            IF (debug_mod) WRITE (*, *) routineN//" Making canonical index"
            CALL dbcsr_make_index_canonical(product_matrix)
         ENDIF
         IF (ASSOCIATED(product_matrix%m%wms)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",969,"Product matrix should be finalized!")
         CALL dbcsr_make_untransposed_blocks(product_matrix)
!$OMP PARALLEL &
!$OMP DEFAULT (NONE) SHARED (product_matrix)
         ! For the multiply logic to work correctly, existing data must
         ! be added only after the index has been transformed into the
         ! canonical form.
         CALL dbcsr_add_wm_from_matrix(product_matrix)
!$OMP END PARALLEL
      ELSE
!$OMP PARALLEL DEFAULT(NONE) PRIVATE(ithread) &
!$OMP SHARED(product_matrix, memtype_product_wm)
         ithread = 0
!$       ithread = OMP_GET_THREAD_NUM()
         CALL dbcsr_work_create(product_matrix, work_mutable=.FALSE., &
                                memory_type=memtype_product_wm(ithread)%p)
!$OMP END PARALLEL
      ENDIF
      CALL dbcsr_data_set_size_referenced(product_matrix%m%data_area, 0)
      product_matrix%m%valid = .FALSE.
      !
      IF (my_use_mpi_exp) THEN
         !
         max_nblocks = product_matrix%m%nblks
         !
         ! Right buffer images
         CALL dbcsr_make_buffers(matrix_right, rdist_right, .FALSE., &
                                 f_k, l_k, f_col, l_col, &
                                 PRESENT(filter_eps), &
                                 has_acc, &
                                 alpha)
         !
         ! Left buffer images
         CALL dbcsr_make_buffers(matrix_left, rdist_left, .TRUE., &
                                 f_row, l_row, f_k, l_k, &
                                 PRESENT(filter_eps), &
                                 has_acc)
      ELSE
         product_matrix%m%nblks = 0
         product_matrix%m%nze = 0
         product_matrix%m%row_p(:) = 0
         !
         ! Right images
         CALL make_m2s(matrix_right, m2s_right, rdist_right, dense_rdist_right, &
                       use_dense_mult, ab_dense, use_local_indexing, use_list_indexing, &
                       "R", f_k, l_k, f_row, l_row, f_col, l_col, &
                       dense_k_sizes, dense_col_sizes, &
                       k_vmap, m_map, n_map, &
                       alpha)
         !
         ! Left images
         CALL make_m2s(matrix_left, m2s_left, rdist_left, dense_rdist_left, &
                       use_dense_mult, ab_dense, use_local_indexing, use_list_indexing, &
                       "L", f_k, l_k, f_row, l_row, f_col, l_col, &
                       dense_row_sizes, dense_k_sizes, &
                       k_vmap, m_map, n_map)
      ENDIF
      !
      IF (ab_dense) THEN
         CALL array_release(k_vmap)
         CALL array_release(dense_row_sizes)
         CALL array_release(dense_col_sizes)
         CALL array_release(dense_k_sizes)
      ENDIF
      !
      ! The limits were already used.  Reset them.
      f_row = 0; l_row = 0
      f_col = 0; l_col = 0
      f_k = 0; l_k = 0
      !
      my_flop = 0
      IF (my_use_mpi_exp) THEN
         IF (has_clusters) THEN
            CALL cannon_multiply_low_rma(rdist_left, rdist_right, &
                                         matrix_left, matrix_right, product_matrix, &
                                         retain_sparsity=retain_sparsity, &
                                         filter_eps=filter_eps, &
                                         flop=my_flop, &
                                         keep_product_data=keep_product_data)
         ELSE
            CALL cannon_multiply_low_exp(rdist_left, rdist_right, &
                                         matrix_left, matrix_right, product_matrix, &
                                         retain_sparsity=retain_sparsity, &
                                         filter_eps=filter_eps, &
                                         flop=my_flop)
         ENDIF
      ELSE
         CALL cannon_multiply_low(m2s_left, m2s_right, product_matrix, &
                                  retain_sparsity=retain_sparsity, &
                                  filter_eps=filter_eps, &
                                  flop=my_flop)
         CALL dbcsr_finalize(product_matrix)
      ENDIF
      !
      IF (PRESENT(flop)) THEN
         ! return the average number of flops per MPI rank.
         ! Variance (which is fairly large) could be computed as well.
         CALL timeset(routineN//"_mpsum_flop", handle2)
         comm = dbcsr_mp_group(dbcsr_distribution_mp(dbcsr_distribution(matrix_c)))
         numnodes = dbcsr_mp_numnodes(dbcsr_distribution_mp(dbcsr_distribution(matrix_c)))
         CALL mp_sum(my_flop, comm)
         flop = (my_flop+numnodes-1)/numnodes
         CALL timestop(handle2)
      ENDIF
      !
      IF (new_left) CALL dbcsr_release(matrix_left)
      IF (new_right) CALL dbcsr_release(matrix_right)
      IF (release_tdist) THEN
         CALL dbcsr_distribution_no_threads(product_matrix%m%dist)
      ENDIF
      !
      CALL dbcsr_release_locals(product_matrix)
      ! The index of the product matrix is reset to the CP2K form if it
      ! was previously set to the canonical form.
      IF (product_reindex) THEN
         IF (debug_mod) WRITE (*, *) routineN//" Making CP2K index"
         CALL dbcsr_make_index_canonical(product_matrix, cp2k=.TRUE.)
      ENDIF
      IF (use_dense_mult) THEN
         CALL dbcsr_release(matrix_c)
         CALL dbcsr_init(matrix_c)
         CALL dbcsr_make_undense(product_matrix, matrix_c, &
                                 old_product_distribution, &
                                 old_product_row_blk_offsets, old_product_col_blk_offsets, &
                                 old_product_row_blk_sizes, old_product_col_blk_sizes, &
                                 m_map, n_map)
         CALL dbcsr_release(product_matrix)
         CALL array_release(old_product_row_blk_offsets)
         CALL array_release(old_product_col_blk_offsets)
         CALL array_release(old_product_row_blk_sizes)
         CALL array_release(old_product_col_blk_sizes)
         CALL dbcsr_distribution_release(old_product_distribution)
      ELSE
         CALL dbcsr_release(matrix_c)
         CALL dbcsr_init(matrix_c)
         CALL dbcsr_copy(matrix_c, product_matrix, shallow_data=.TRUE.)
         CALL dbcsr_release(product_matrix)
      ENDIF
      !
      IF (.NOT. my_use_mpi_exp) THEN
         CALL dbcsr_destroy_array(m2s_left)
         DEALLOCATE (m2s_left)
         CALL dbcsr_destroy_array(m2s_right)
         DEALLOCATE (m2s_right)
      ENDIF
      !
      CALL dbcsr_image_dist_release(rdist_left)
      CALL dbcsr_image_dist_release(rdist_right)
      IF (ab_dense) THEN
         CALL array_release(m_map)
         CALL array_release(n_map)
      ENDIF
      !
      ! if filtering is requested remove small blocks, unless the sparsity needs to be kept.
      !
      IF (PRESENT(filter_eps) .AND. .NOT. keep_sparsity) THEN
         eps_any = dbcsr_scalar(filter_eps)
         CALL dbcsr_scalar_fill_all(eps_any)
         CALL dbcsr_scalar_set_type(eps_any, dbcsr_get_data_type(matrix_c))
         CALL dbcsr_filter(matrix_c, eps_any, quick=.FALSE.)
      ENDIF
      !
      ! To support the canonical multiply (all non-transposed blocks),
      ! blocks may have to be transposed according to the CP2K
      ! triangular index.
      CALL dbcsr_make_untransposed_blocks(matrix_c)
      !
      IF (debug_mod .OR. careful_mod) THEN
         IF (debug_mod) &
            WRITE (*, *) routineN//" Use dense mult, symm", &
            use_dense_mult, ab_dense, dbcsr_has_symmetry(matrix_c)
         CALL dbcsr_verify_matrix(matrix_c)
         IF (debug_mod) THEN
            cs = dbcsr_checksum(matrix_c)
            WRITE (*, *) routineN//" Product checksum", cs
         ENDIF
      ENDIF

      ! This tends to trigger only when all of these conditions are fulfilled:
      !  - transa=="T"
      !  - matrix_c contains already blocks and beta is not zero
      !  - Cuda is enabled
      !  - multiple OpenMP threads are used
      IF (INT(matrix_c%m%nblks, KIND=int_8) > &
          INT(SIZE(array_data(matrix_c%m%row_blk_size)), KIND=int_8)* &
          INT(SIZE(array_data(matrix_c%m%col_blk_size)), KIND=int_8)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1155,"Bug: Matrix contains too many blocks")
      output_unit = default_output_unit
      CALL timestop(handle)
   END SUBROUTINE dbcsr_mm_cannon_multiply

! **************************************************************************************************
!> \brief Make images from the matrix (left or right)
!> \param[in] matrix ...
!> \param[out] m2s ...
!> \param[in,out] rdist ...
!> \param[in,out] dense_rdist ...
!> \param[in] use_dense_mult ...
!> \param[in] ab_dense ...
!> \param[in] use_local_indexing ...
!> \param[in] use_list_indexing ...
!> \param[in] predistribute ...
!> \param[in] f_k ...
!> \param[in] l_k ...
!> \param[in] f_row ...
!> \param[in] l_row ...
!> \param[in] f_col ...
!> \param[in] l_col ...
!> \param[in,out] row_blk_size ...
!> \param[in,out] col_blk_size ...
!> \param[in] k_vmap ...
!> \param[in] m_map ...
!> \param[in] n_map ...
!> \param[in] alpha ...
! **************************************************************************************************
   SUBROUTINE make_m2s(matrix, m2s, rdist, dense_rdist, &
                       use_dense_mult, ab_dense, &
                       use_local_indexing, use_list_indexing, &
                       predistribute, f_k, l_k, f_row, l_row, f_col, l_col, &
                       row_blk_size, col_blk_size, &
                       k_vmap, m_map, n_map, &
                       alpha)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_2d_array_type), INTENT(OUT), POINTER    :: m2s
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: rdist, dense_rdist
      LOGICAL, INTENT(IN)                                :: use_dense_mult, ab_dense, &
                                                            use_local_indexing, use_list_indexing
      CHARACTER, INTENT(IN)                              :: predistribute
      INTEGER, INTENT(IN)                                :: f_k, l_k, f_row, l_row, f_col, l_col
      TYPE(array_i1d_obj), INTENT(INOUT)                 :: row_blk_size, col_blk_size
      TYPE(array_i1d_obj), INTENT(IN)                    :: k_vmap, m_map, n_map
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: alpha

      INTEGER                                            :: i, im, j, jm
      INTEGER, DIMENSION(4)                              :: f_crop
      LOGICAL                                            :: do_scale, thread_redist
      TYPE(array_i1d_obj)                                :: col_map, row_map
      TYPE(dbcsr_obj)                                    :: dense_template, matrix_tmp

      ALLOCATE (m2s)
      do_scale = .FALSE.
      IF (PRESENT(alpha)) THEN
         IF (.NOT. dbcsr_scalar_are_equal(alpha, dbcsr_scalar_one(alpha%data_type))) THEN
            do_scale = .TRUE.
         END IF
      END IF

      IF (do_scale) THEN
         ! Copy and scale matrix if alpha is not 1.
         CALL dbcsr_make_images(matrix, m2s, rdist, &
                                predistribute=predistribute, &
                                no_copy_data=use_dense_mult, scale_value=alpha)
      ELSE
         CALL dbcsr_make_images(matrix, m2s, rdist, &
                                predistribute=predistribute, &
                                no_copy_data=use_dense_mult)
      END IF

      im = SIZE(m2s%mats, 1)
      jm = SIZE(m2s%mats, 2)
      SELECT CASE (predistribute)
      CASE ('L')
         f_crop = (/f_row, l_row, f_k, l_k/)
         row_map = m_map
         col_map = k_vmap
         thread_redist = .TRUE.
      CASE default
         f_crop = (/f_k, l_k, f_col, l_col/)
         row_map = k_vmap
         col_map = n_map
         thread_redist = .FALSE.
      END SELECT

      ! Post-processing of images.
      DO i = 1, im
         DO j = 1, jm
            CALL dbcsr_reset_vlocals(m2s%mats(i, j), rdist)
            ! Crop if necessary
            IF (ANY(f_crop .NE. 0)) THEN
               CALL dbcsr_init(matrix_tmp)
               CALL dbcsr_crop_matrix(matrix_tmp, m2s%mats(i, j), &
                                      full_row_bounds=f_crop(1:2), &
                                      full_column_bounds=f_crop(3:4), &
                                      shallow_data=.FALSE.)
               CALL dbcsr_release(m2s%mats(i, j))
               CALL dbcsr_copy(m2s%mats(i, j), matrix_tmp, shallow_data=.TRUE.)
               CALL dbcsr_release(matrix_tmp)
               CALL dbcsr_reset_vlocals(m2s%mats(i, j), rdist)
            ENDIF
         ENDDO
      ENDDO

      IF (ab_dense) THEN
         CALL dbcsr_init(dense_template)
         CALL dbcsr_create(dense_template, template=matrix, &
                           dist=dense_rdist%i%main, &
                           row_blk_size=array_data(row_blk_size), col_blk_size=array_data(col_blk_size))
         CALL dbcsr_make_images_dense(m2s, dense_rdist, &
                                      row_map=row_map, col_map=col_map, &
                                      join_cols=use_dense_mult, join_rows=ab_dense, &
                                      new_template=dense_template)
         CALL dbcsr_image_dist_release(rdist)
         rdist = dense_rdist
         CALL dbcsr_image_dist_hold(rdist)
         DO i = 1, im
            DO j = 1, jm
               CALL dbcsr_reset_vlocals(m2s%mats(i, j), rdist)
            ENDDO
         ENDDO
      ENDIF

      DO i = 1, im
         DO j = 1, jm
            ! Convert to local-row index
            IF (use_local_indexing) THEN
               CALL dbcsr_make_index_local_row(m2s%mats(i, j))
            ENDIF
            ! Convert to list index
            IF (use_list_indexing) THEN
               CALL dbcsr_make_index_list(m2s%mats(i, j), thread_redist=thread_redist)
            ENDIF
            IF (use_local_indexing .AND. .NOT. use_list_indexing) THEN
               CALL dbcsr_index_compact(m2s%mats(i, j))
            ENDIF
         ENDDO
      ENDDO

      IF (ab_dense) THEN
         CALL dbcsr_image_dist_release(dense_rdist)
         CALL dbcsr_release(dense_template)
      ENDIF

   END SUBROUTINE make_m2s

! **************************************************************************************************
!> \brief Prepare orig images
!> \param matrix ...
!> \param imgdist ...
!> \param is_left ...
!> \param f_row ...
!> \param l_row ...
!> \param f_col ...
!> \param l_col ...
!> \param otf_filtering ...
!> \param has_acc ...
!> \param alpha ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE dbcsr_make_buffers(matrix, imgdist, is_left, &
                                 f_row, l_row, f_col, l_col, &
                                 otf_filtering, has_acc, &
                                 alpha)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: imgdist
      LOGICAL, INTENT(IN)                                :: is_left
      INTEGER, INTENT(IN)                                :: f_row, l_row, f_col, l_col
      LOGICAL, INTENT(IN)                                :: otf_filtering, has_acc
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: alpha

      LOGICAL                                            :: do_scale

      do_scale = .FALSE.
      IF (PRESENT(alpha)) THEN
         IF (.NOT. dbcsr_scalar_are_equal(alpha, dbcsr_scalar_one(alpha%data_type))) THEN
            do_scale = .TRUE.
         END IF
      END IF
      !
      IF (do_scale) THEN
         CALL make_buffers(matrix, imgdist, is_left, &
                           f_row, l_row, f_col, l_col, otf_filtering, has_acc, alpha)
      ELSE
         CALL make_buffers(matrix, imgdist, is_left, &
                           f_row, l_row, f_col, l_col, otf_filtering, has_acc)
      ENDIF
   END SUBROUTINE dbcsr_make_buffers

! **************************************************************************************************
!> \brief Prepare orig images
!> \param matrix ...
!> \param imgdist ...
!> \param is_left ...
!> \param f_row ...
!> \param l_row ...
!> \param f_col ...
!> \param l_col ...
!> \param otf_filtering ...
!> \param has_acc ...
!> \param scale_value ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE make_buffers(matrix, imgdist, is_left, &
                           f_row, l_row, f_col, l_col, &
                           otf_filtering, has_acc, &
                           scale_value)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: imgdist
      LOGICAL, INTENT(IN)                                :: is_left
      INTEGER, INTENT(IN)                                :: f_row, l_row, f_col, l_col
      LOGICAL, INTENT(IN)                                :: otf_filtering, has_acc
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale_value

      CHARACTER(len=*), PARAMETER :: routineN = 'make_buffers', routineP = moduleN//':'//routineN

      INTEGER :: blk, blk_p, bp, col, col_img, col_size, data_size, data_size_diag, data_type, &
         dst_proc, g2l_img_ref, grp, handle, iproc, irequests, it, ithread, local_no_empty_images, &
         meta_size, meta_size_diag, meta_size_diag_scatter, mygrp, mygrp_scatter, mynode, mypcol, &
         myprow, myt, nblkcols_local, nblkrows_local, nblocks, ncols_images, nimages_merged, &
         nimages_unmerged, nprocs, nprocs_scatter, nrows_images, nsymmetries, nthreads, nze, pcol, &
         pdiag, prow, request_data_size_diag, request_data_size_scatter, request_meta_size_diag, &
         request_meta_size_scatter, request_no_empty_images, row, row_img
      INTEGER :: row_size, scatter_images, scatter_images_proc, size_index_unmerged, stored_col, &
         stored_row, sym_p, symmetry_i, total_no_empty_images, tr_col_size, tr_row_size
      INTEGER, ALLOCATABLE, DIMENSION(:) :: data_displ_scatter, data_size_scatter, g2l_map_cols, &
         g2l_map_cols_merge, g2l_map_rows, g2l_map_rows_merge, img_map, img_refs, img_refs_cols, &
         img_refs_rows, meta_displ_cluster_scatter, meta_displ_scatter, meta_size_cluster_scatter, &
         meta_size_scatter, tmp_img_offset
      INTEGER, ALLOCATABLE, DIMENSION(:), TARGET         :: meta_recv, meta_send
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: local_refs_size
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :)        :: recv_refs
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :, :)     :: refs
      INTEGER, DIMENSION(2)                              :: recv_sym_size, sym_size
      INTEGER, DIMENSION(:), POINTER :: col_dist, col_img_dist, local_cols, local_refs_data_size, &
         local_rows, meta_buffer_p, no_empty_images, no_empty_images_displ, row_dist, &
         row_img_dist, threads_dist
      INTEGER, DIMENSION(:, :), POINTER                  :: blacs2mpi, local_refs_displ_unmerged, &
                                                            local_refs_meta_size
      INTEGER, DIMENSION(:, :, :), POINTER               :: local_refs_displ
      INTEGER, DIMENSION(dbcsr_slot_nblkrows_total:&
         dbcsr_slot_nfullcols_local)                     :: global_indices
      INTEGER, POINTER                                   :: coli, rowi
      INTEGER, TARGET                                    :: mi, ui
      LOGICAL                                            :: do_bcast, do_scatter, do_symmetry, &
                                                            do_win_free, is_diagonal, tr
      LOGICAL, DIMENSION(2)                              :: do_win_create
      REAL(kind=sp), DIMENSION(:), POINTER               :: local_max_norms, max_norms
      TYPE(dbcsr_buffer), POINTER                        :: buffer
      TYPE(dbcsr_data_obj), POINTER                      :: data_buffer_p
      TYPE(dbcsr_data_obj), TARGET                       :: data_recv, data_send
      TYPE(dbcsr_distribution_obj)                       :: set_dist
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_mp_obj)                                 :: mp_obj
      TYPE(dbcsr_scalar_type)                            :: scale_neg_one
      TYPE(dbcsr_type)                                   :: sm

!$    INTEGER(kind=omp_lock_kind), ALLOCATABLE, DIMENSION(:) :: locks

      CALL timeset(routineN, handle)
      !
      ! Take input values and check validity
      IF (.NOT. dbcsr_valid_index(matrix)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1420,"Matrix not initialized.")
      sm = matrix%m
      data_type = sm%data_type
      IF (data_type .NE. dbcsr_type_real_8 .AND. &
          data_type .NE. dbcsr_type_real_4 .AND. &
          data_type .NE. dbcsr_type_complex_8 .AND. &
          data_type .NE. dbcsr_type_complex_4) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1427,"Invalid data type.")
      scale_neg_one = dbcsr_scalar_negative(dbcsr_scalar_one(data_type))
      set_dist = imgdist%i%main
      row_dist => dbcsr_distribution_row_dist(set_dist)
      col_dist => dbcsr_distribution_col_dist(set_dist)
      local_rows => dbcsr_distribution_local_rows(set_dist)
      local_cols => dbcsr_distribution_local_cols(set_dist)
      nblkrows_local = SIZE(local_rows)
      nblkcols_local = SIZE(local_cols)
      IF (sm%symmetry) THEN
         IF (SIZE(row_dist) .NE. SIZE(col_dist)) &
            CALL cp__w("dbcsr/mm/dbcsr_mm_cannon.F",1438,'Unequal row and column distributions for symmetric matrix.')
      ENDIF
      nrows_images = imgdist%i%row_decimation
      row_img_dist => array_data(imgdist%i%row_image)
      ncols_images = imgdist%i%col_decimation
      col_img_dist => array_data(imgdist%i%col_image)
      mp_obj = dbcsr_distribution_mp(set_dist)
      grp = dbcsr_mp_group(mp_obj)
      blacs2mpi => dbcsr_mp_pgrid(mp_obj)
      mynode = dbcsr_mp_mynode(mp_obj)
      myprow = dbcsr_mp_myprow(mp_obj)
      mypcol = dbcsr_mp_mypcol(mp_obj)
      IF (MAXVAL(row_dist) .GT. UBOUND(blacs2mpi, 1)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1451,"Row distribution references unexistent processor rows")
      IF (MAXVAL(col_dist) .GT. UBOUND(blacs2mpi, 2)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1453,"Col distribution references unexistent processor cols")
      IF (.NOT. dbcsr_mp_has_subgroups(mp_obj)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1455,"Experimental MPI algorithms require subcommunicators!")
      ! Check threads configuration
      NULLIFY (threads_dist)
!$    IF (.NOT. dbcsr_distribution_has_threads(dbcsr_distribution(matrix))) &
!$       CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1459,"Thread distribution not defined")
!$    threads_dist => array_data(dbcsr_distribution_thread_dist(dbcsr_distribution(matrix)))
      !
      ! Crop matrix
      IF (ANY((/f_row, l_row, f_col, l_col/) .NE. 0)) THEN
         IF (f_row .LT. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1465,"Invalid first row bound.")
         IF (l_row .GT. dbcsr_nfullrows_total(matrix)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1467,"Invalid last row bound.")
         IF (f_col .LT. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1469,"Invalid first column bound.")
         IF (l_col .GT. dbcsr_nfullcols_total(matrix)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1471,"Invalid last column bound.")
         !
         ! For the moment crop is not supported for experimental MPI algorithm
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",1474,"Crop not supported for experimental MPI algo yet.")
         !
      ENDIF
      !
      IF (dbcsr_has_symmetry(matrix)) THEN
         nsymmetries = 2
         do_symmetry = .TRUE.
      ELSE
         nsymmetries = 1
         do_symmetry = .FALSE.
      ENDIF
      sym_p = blacs2mpi(mypcol, myprow)
      is_diagonal = sym_p .EQ. mynode
      !
      size_index_unmerged = dbcsr_slot_nblks
      !
      IF (is_left) THEN
         ! merging over rows
         nimages_merged = nrows_images
         nimages_unmerged = ncols_images
         nprocs = dbcsr_mp_npcols(mp_obj)
         ALLOCATE (left_refs_meta_size(myprow*nimages_merged:(myprow+1)*nimages_merged-1, &
                                       0:nprocs*nimages_unmerged-1))
         ALLOCATE (left_local_refs_meta_size(nimages_merged, nimages_unmerged))
         left_local_refs_meta_size(:, :) = 0
         local_refs_meta_size => left_local_refs_meta_size
         ALLOCATE (left_local_refs_data_size(nimages_merged*nimages_unmerged))
         local_refs_data_size => left_local_refs_data_size
         ALLOCATE (left_refs_displ_unmerged(idata_displ:imeta_displ, &
                                            0:nprocs*nimages_unmerged-1))
         ALLOCATE (left_local_refs_displ_unmerged(idata_displ:imeta_displ, nimages_unmerged))
         local_refs_displ_unmerged => left_local_refs_displ_unmerged
         mygrp = dbcsr_mp_my_row_group(mp_obj)
         requests(7) = mp_request_null
         pdiag = myprow
         irequests = 1
         !
         ! Count the maximum possible multiplies per row for on-the-fly filtering
         IF (otf_filtering) THEN
            ALLOCATE (left_total_row_counts(nblkrows_local))
            left_total_row_counts = 0
         ENDIF
         buffer => buffers_orig%left
         ALLOCATE (left_no_empty_images(nprocs))
         no_empty_images => left_no_empty_images
         ALLOCATE (left_no_empty_images_displ(nprocs))
         no_empty_images_displ => left_no_empty_images_displ
         do_scatter = .FALSE.
         !
         do_win_free = buffers_orig%has_rma_left_win
         buffers_orig%has_rma_left_win = .TRUE.
      ELSE
         ! merging over cols
         nimages_merged = ncols_images
         nimages_unmerged = nrows_images
         nprocs = dbcsr_mp_nprows(mp_obj)
         nprocs_scatter = dbcsr_mp_npcols(mp_obj)
         ALLOCATE (right_refs_meta_size(mypcol*nimages_merged:(mypcol+1)*nimages_merged-1, &
                                        0:nprocs*nimages_unmerged-1))
         ALLOCATE (right_local_refs_meta_size(nimages_merged, nimages_unmerged))
         right_local_refs_meta_size(:, :) = 0
         local_refs_meta_size => right_local_refs_meta_size
         ALLOCATE (right_local_refs_data_size(nimages_merged*nimages_unmerged))
         local_refs_data_size => right_local_refs_data_size
         ALLOCATE (right_refs_displ_unmerged(idata_displ:imeta_displ, &
                                             0:nprocs*nimages_unmerged-1))
         ALLOCATE (right_local_refs_displ_unmerged(idata_displ:imeta_displ, nimages_unmerged))
         local_refs_displ_unmerged => right_local_refs_displ_unmerged
         mygrp = dbcsr_mp_my_col_group(mp_obj)
         mygrp_scatter = dbcsr_mp_my_row_group(mp_obj)
         requests(8) = mp_request_null
         pdiag = mypcol
         irequests = 2
         buffer => buffers_orig%right
         ALLOCATE (right_no_empty_images(nprocs))
         no_empty_images => right_no_empty_images
         ALLOCATE (right_no_empty_images_displ(nprocs))
         no_empty_images_displ => right_no_empty_images_displ
         global_indices(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
            (/ &
            sm%nblkrows_total, &
            sm%nblkcols_total, &
            sm%nfullrows_total, &
            sm%nfullcols_total, &
            0, 0, &
            sm%nfullrows_local, &
            sm%nfullcols_local/)
         ! Do not scatter data when there is a single cluster or single column proc
         do_scatter = nprocs_scatter .GT. 1 .AND. nimages_merged .GT. 1 .AND. .NOT. has_acc
         IF (do_scatter .AND. is_diagonal) THEN
            ALLOCATE (meta_displ_cluster_scatter(nimages_unmerged))
            ALLOCATE (meta_size_cluster_scatter(nimages_unmerged))
            meta_displ_cluster_scatter(1) = 0
         ENDIF
         local_meta_size_scatter = 0
         local_data_size_scatter = 0
         !
         do_win_free = buffers_orig%has_rma_right_win
         buffers_orig%has_rma_right_win = .TRUE.
      ENDIF
      !
      do_bcast = nimages_merged .GT. 1
      !
      ALLOCATE (local_refs_size(idata_size:imeta_size, nimages_merged, nimages_unmerged))
      local_refs_size(:, :, :) = 0
      !
      ! Evaluate maps for global -> local indexing (g2l_map_rows, g2l_map_cols)
      ! Evaluate maps for local_image -> local indexing (img_map)
      ! Evaluate refs (count and offset) per each image
      ALLOCATE (g2l_map_rows(sm%nblkrows_total), g2l_map_cols(sm%nblkcols_total))
      ALLOCATE (g2l_map_rows_merge(sm%nblkrows_total), g2l_map_cols_merge(sm%nblkcols_total))
      ALLOCATE (img_refs_rows(nimages_unmerged), img_refs_cols(nimages_unmerged))
      !
      img_refs_rows(:) = 0
      g2l_map_rows(:) = 0
      g2l_map_rows_merge(:) = 0
      IF (nrows_images .GT. 1) THEN
         g2l_img_ref = 0
         DO row = 1, nblkrows_local
            row_img = row_img_dist(local_rows(row))
            mi = MOD(row_img-1, nrows_images)+1
            img_refs_rows(mi) = img_refs_rows(mi)+1
            g2l_map_rows(local_rows(row)) = img_refs_rows(mi)
            g2l_img_ref = g2l_img_ref+1
            g2l_map_rows_merge(local_rows(row)) = g2l_img_ref
         ENDDO
      ELSE
         DO row = 1, nblkrows_local
            g2l_map_rows_merge(local_rows(row)) = row
         ENDDO
         ! Symmetric case, swap row with col images
         IF (do_symmetry) THEN
            DO row = 1, nblkrows_local
               col_img = col_img_dist(local_rows(row))
               mi = MOD(col_img-1, ncols_images)+1
               img_refs_rows(mi) = img_refs_rows(mi)+1
               g2l_map_rows(local_rows(row)) = img_refs_rows(mi)
            ENDDO
         ELSE
            g2l_map_rows(:) = g2l_map_rows_merge(:)
         ENDIF
         img_refs_rows(1) = nblkrows_local
      ENDIF
      !
      img_refs_cols(:) = 0
      g2l_map_cols(:) = 0
      g2l_map_cols_merge(:) = 0
      IF (ncols_images .GT. 1) THEN
         g2l_img_ref = 0
         DO col = 1, nblkcols_local
            col_img = col_img_dist(local_cols(col))
            mi = MOD(col_img-1, ncols_images)+1
            img_refs_cols(mi) = img_refs_cols(mi)+1
            g2l_map_cols(local_cols(col)) = img_refs_cols(mi)
            g2l_img_ref = g2l_img_ref+1
            g2l_map_cols_merge(local_cols(col)) = g2l_img_ref
         ENDDO
      ELSE
         DO col = 1, nblkcols_local
            g2l_map_cols_merge(local_cols(col)) = col
         ENDDO
         ! Symmetric case, swap col with row images
         IF (do_symmetry) THEN
            DO col = 1, nblkcols_local
               row_img = row_img_dist(local_cols(col))
               mi = MOD(row_img-1, nrows_images)+1
               img_refs_cols(mi) = img_refs_cols(mi)+1
               g2l_map_cols(local_cols(col)) = img_refs_cols(mi)
            ENDDO
         ELSE
            g2l_map_cols(:) = g2l_map_cols_merge(:)
         ENDIF
         img_refs_cols(1) = nblkcols_local
      ENDIF
      !
      IF (is_left) THEN
         ALLOCATE (img_refs(ncols_images))
         img_refs(1) = 0
         DO col_img = 1, ncols_images-1
            img_refs(col_img+1) = img_refs(col_img)+img_refs_cols(col_img)
         ENDDO
         ALLOCATE (img_map(nblkcols_local))
         IF (ncols_images .GT. 1) THEN
            ALLOCATE (tmp_img_offset(ncols_images))
            tmp_img_offset(:) = img_refs(:)
            DO col = 1, nblkcols_local
               col_img = col_img_dist(local_cols(col))
               ui = MOD(col_img-1, ncols_images)+1
               tmp_img_offset(ui) = tmp_img_offset(ui)+1
               img_map(tmp_img_offset(ui)) = col
            ENDDO
         ELSE
            DO col = 1, nblkcols_local
               img_map(col) = col
            ENDDO
         ENDIF
      ELSE
         ALLOCATE (img_refs(nrows_images))
         img_refs(1) = 0
         DO row_img = 1, nrows_images-1
            img_refs(row_img+1) = img_refs(row_img)+img_refs_rows(row_img)
         ENDDO
         ALLOCATE (img_map(nblkrows_local))
         IF (nrows_images .GT. 1) THEN
            ALLOCATE (tmp_img_offset(nrows_images))
            tmp_img_offset(:) = img_refs(:)
            DO row = 1, nblkrows_local
               row_img = row_img_dist(local_rows(row))
               mi = MOD(row_img-1, nrows_images)+1
               tmp_img_offset(mi) = tmp_img_offset(mi)+1
               img_map(tmp_img_offset(mi)) = row
            ENDDO
         ELSE
            DO row = 1, nblkrows_local
               img_map(row) = row
            ENDDO
         ENDIF
      ENDIF
      IF (ALLOCATED(tmp_img_offset)) DEALLOCATE (tmp_img_offset)
      !
      ALLOCATE (local_refs_displ(idata_displ:imeta_displ, nimages_merged, nimages_unmerged))
      data_size = 0
      meta_size = 0
      data_size_diag = 0
      meta_size_diag = 0
      meta_size_diag_scatter = 0
      local_no_empty_images = 0
      !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP PRIVATE (ithread,myt,iter,row,col,blk,row_size,col_size,&
!$OMP          stored_row,stored_col,blk_p,bp,tr,&
!$OMP          nze,symmetry_i,row_img,col_img,rowi,coli,&
!$OMP          tr_row_size,tr_col_size,prow,pcol,dst_proc,&
!$OMP          data_buffer_p,meta_buffer_p,&
!$OMP          mi,ui,it,nblocks) &
!$OMP SHARED (nthreads,refs,matrix,nsymmetries,do_symmetry,&
!$OMP         row_img_dist,col_img_dist,imgdist,row_dist,col_dist,&
!$OMP         is_left,local_refs_size,nimages_merged,nimages_unmerged,&
!$OMP         data_size,meta_size,data_type,memtype_mpi_buffer,sm,&
!$OMP         img_refs_cols,img_refs_rows,img_refs,img_map,&
!$OMP         g2l_map_cols,g2l_map_rows,recv_refs,grp,&
!$OMP         scale_value,scale_neg_one,data_send,data_recv,&
!$OMP         size_index_unmerged,sym_size,recv_sym_size,meta_send,&
!$OMP         mp_obj,threads_dist,meta_recv,nrows_images,ncols_images,&
!$OMP         locks,mynode,blacs2mpi,myprow,mypcol,sym_p,mygrp,&
!$OMP         left_refs_meta_size,right_refs_meta_size,&
!$OMP         left_refs_data_size,right_refs_data_size,&
!$OMP         local_refs_meta_size,local_refs_data_size,&
!$OMP         local_refs_displ,nprocs_scatter,meta_scatter,&
!$OMP         left_refs_displ_unmerged,right_refs_displ_unmerged,&
!$OMP         local_refs_displ_unmerged,requests,is_diagonal,do_bcast,&
!$OMP         g2l_map_cols_merge,g2l_map_rows_merge,&
!$OMP         meta_size_diag,data_size_diag,pdiag,buffer,&
!$OMP         left_total_row_counts,otf_filtering,meta_size_diag_scatter,&
!$OMP         request_meta_size_diag,request_data_size_diag,&
!$OMP         no_empty_images,local_no_empty_images,&
!$OMP         request_no_empty_images,irequests,do_win_free,do_win_create,&
!$OMP         meta_size_scatter,data_size_scatter,iproc,&
!$OMP         scatter_images_proc,scatter_images,&
!$OMP         request_meta_size_scatter,mygrp_scatter,do_scatter,&
!$OMP         request_data_size_scatter,meta_displ_scatter,data_displ_scatter,&
!$OMP         local_meta_size_scatter,local_data_size_scatter,&
!$OMP         meta_displ_cluster_scatter,meta_size_cluster_scatter,global_indices) &
!$OMP         REDUCTION(MAX : max_nblocks)
      ithread = 0
!$    ithread = omp_get_thread_num()
      myt = ithread
      IF (is_left) THEN
         rowi => mi
         coli => ui
      ELSE
         rowi => ui
         coli => mi
      ENDIF
!$OMP MASTER
      nthreads = 1
!$    nthreads = omp_get_num_threads()
      ALLOCATE (refs(idata_size:imeta_size, &
                     nimages_merged, nimages_unmerged, &
                     0:nthreads, nsymmetries))
      refs(:, :, :, :, :) = 0
!$    IF (is_left) THEN
!$       size_index_unmerged = size_index_unmerged+nthreads+1
!$    ENDIF
!$    IF (is_left .AND. do_symmetry) THEN
!$       ALLOCATE (locks(0:nthreads-1))
!$    ENDIF
!$OMP END MASTER
!$OMP BARRIER
!$    IF (is_left .AND. do_symmetry) THEN
!$       call omp_init_lock(locks(ithread))
!$    ENDIF
      !
      ! By default all data are considered already in local proc
      dst_proc = ilocal_proc
      !
      ! Take data and meta dimensions per each thread
      CALL dbcsr_iterator_start(iter, matrix, shared=.TRUE.)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, blk, &
                                        row_size=row_size, col_size=col_size)
         nze = row_size*col_size
         IF (nze .EQ. 0) CYCLE
         DO symmetry_i = 1, nsymmetries
            IF (symmetry_i .EQ. 1) THEN
               stored_row = row; stored_col = col
            ELSE
               IF (row .EQ. col) CYCLE
               stored_row = col; stored_col = row
            ENDIF
            row_img = row_img_dist(stored_row)
            col_img = col_img_dist(stored_col)
            CALL image_calculator(imgdist, &
                                  prow=prow, pcol=pcol, &
                                  rowi=rowi, coli=coli, &
                                  myprow=row_dist(stored_row), myrowi=row_img, &
                                  mypcol=col_dist(stored_col), mycoli=col_img, &
                                  shifting='0')
            IF (do_symmetry) THEN
               ! Avoid unnecessary copy of data for diagonal procs
               IF (blacs2mpi(prow, pcol) .EQ. mynode .OR. is_diagonal) THEN
                  dst_proc = ilocal_proc
               ELSE
                  ! Move data to remote symmetric proc
                  dst_proc = isym_proc
               ENDIF
            ENDIF
!$          IF (is_left .AND. do_symmetry) THEN
!$             myt = threads_dist(stored_row)
!$          ENDIF
!$OMP ATOMIC
            refs(imeta_size, mi, ui, myt+1, dst_proc) = &
               refs(imeta_size, mi, ui, myt+1, dst_proc)+3
!$OMP ATOMIC
            refs(idata_size, mi, ui, myt+1, dst_proc) = &
               refs(idata_size, mi, ui, myt+1, dst_proc)+nze
         ENDDO ! symmetry_i
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      ! Avoid unnecessary copy if data is already local
      IF (do_symmetry .AND. .NOT. is_diagonal) THEN
!$OMP BARRIER
!$OMP MASTER
         ALLOCATE (recv_refs(idata_size:imeta_size, &
                             nimages_merged, nimages_unmerged, &
                             0:nthreads))
         ! exchange sizes with the symmetric proc
         CALL mp_sendrecv(refs(:, :, :, 1:nthreads, isym_proc), sym_p, &
                          recv_refs(:, :, :, 1:nthreads), sym_p, grp)
!$OMP END MASTER
      ENDIF
!$OMP BARRIER
!$OMP CRITICAL(local_refs_acc)
      !
      ! Store local data and meta dimensions
      local_refs_size(:, :, :) = local_refs_size(:, :, :)+refs(:, :, :, ithread+1, ilocal_proc)
      IF (do_symmetry .AND. .NOT. is_diagonal) THEN
         local_refs_size(:, :, :) = local_refs_size(:, :, :)+recv_refs(:, :, :, ithread+1)
      ENDIF
!$OMP END CRITICAL(local_refs_acc)
!$OMP BARRIER
!$OMP MASTER
      !
      ! Take the displacements, taking in account the images merging
      ! Note that local diplacements refer to position 0
      DO ui = 1, nimages_unmerged
         local_refs_displ_unmerged(imeta_displ, ui) = meta_size
         local_refs_displ_unmerged(idata_displ, ui) = data_size
         DO mi = 1, nimages_merged
            IF (is_diagonal .AND. do_bcast .AND. &
                (ui .EQ. mi .OR. nimages_merged .EQ. 1)) THEN
               local_refs_displ(imeta_displ, mi, ui) = meta_size_diag
               local_refs_displ(idata_displ, mi, ui) = data_size_diag
            ELSE
               local_refs_displ(imeta_displ, mi, ui) = meta_size
               local_refs_displ(idata_displ, mi, ui) = data_size
            ENDIF
            IF (local_refs_size(imeta_size, mi, ui) .EQ. 0) CYCLE
            local_no_empty_images = local_no_empty_images+1
            local_refs_data_size(local_no_empty_images) = local_refs_size(idata_size, mi, ui)
            ! Include stats slots
            local_refs_size(imeta_size, mi, ui) = local_refs_size(imeta_size, mi, ui)+ &
                                                  size_index_unmerged
            local_refs_meta_size(mi, ui) = local_refs_size(imeta_size, mi, ui)
            IF (is_diagonal .AND. do_bcast .AND. &
                (ui .EQ. mi .OR. nimages_merged .EQ. 1)) THEN
               meta_size_diag = meta_size_diag+local_refs_size(imeta_size, mi, ui)
               data_size_diag = data_size_diag+local_refs_size(idata_size, mi, ui)
            ELSE
               meta_size = meta_size+local_refs_size(imeta_size, mi, ui)
               data_size = data_size+local_refs_size(idata_size, mi, ui)
            ENDIF
         ENDDO
      ENDDO
      !
      ! Evaluate diagonal images to send for the right matrix
      IF (do_scatter) THEN
         IF (is_diagonal) THEN
            ALLOCATE (meta_size_scatter(0:nprocs_scatter-1), &
                      data_size_scatter(0:nprocs_scatter-1), &
                      meta_displ_scatter(0:nprocs_scatter-1), &
                      data_displ_scatter(0:nprocs_scatter-1))
            meta_size_scatter(:) = 0; data_size_scatter(:) = 0
            meta_displ_scatter(0) = 0; data_displ_scatter(0) = 0
            scatter_images_proc = nimages_unmerged/nprocs_scatter+1
            ui = MOD(nimages_unmerged, nprocs_scatter)
            mi = 0
            DO iproc = 0, nprocs_scatter-1
               IF (iproc .GT. 0) THEN
                  meta_displ_scatter(iproc) = meta_displ_scatter(iproc-1)+ &
                                              meta_size_scatter(iproc-1)
                  data_displ_scatter(iproc) = data_displ_scatter(iproc-1)+ &
                                              data_size_scatter(iproc-1)
               ENDIF
               IF (iproc .EQ. ui) THEN
                  scatter_images_proc = scatter_images_proc-1
               ENDIF
               scatter_images = 0
               DO WHILE (scatter_images .LT. scatter_images_proc)
                  mi = mi+1
                  scatter_images = scatter_images+1
                  meta_displ_cluster_scatter(mi) = meta_size_diag_scatter
                  IF (local_refs_size(imeta_size, mi, mi) .EQ. 0) CYCLE
                  IF (iproc .NE. pdiag) THEN
                     meta_size_cluster_scatter(mi) = local_refs_size(imeta_size, mi, mi)+ &
                                                     dbcsr_num_slots- &
                                                     size_index_unmerged
                  ELSE
                     ! For diagonal procs the empty data to scatter are considered empty
                     meta_size_cluster_scatter(mi) = dbcsr_num_slots
                  ENDIF
                  meta_size_diag_scatter = meta_size_diag_scatter+ &
                                           meta_size_cluster_scatter(mi)
                  meta_size_scatter(iproc) = &
                     meta_size_scatter(iproc)+meta_size_cluster_scatter(mi)
                  data_size_scatter(iproc) = &
                     data_size_scatter(iproc)+local_refs_size(idata_size, mi, mi)
               ENDDO
            ENDDO
            ! Skip diagonal proc
            data_size_scatter(pdiag) = 0
            CALL ensure_array_size(meta_scatter, ub=meta_size_diag_scatter, &
                                   nocopy=.TRUE., memory_type=memtype_mpi_buffer)
         ENDIF
         !
         ! Scatter the diagonal sizes
         CALL mp_iscatter(meta_size_scatter, local_meta_size_scatter, &
                          myprow, mygrp_scatter, request_meta_size_scatter)
         CALL mp_iscatter(data_size_scatter, local_data_size_scatter, &
                          myprow, mygrp_scatter, request_data_size_scatter)
      ENDIF
      !
      ! Bcast the diagonal sizes
      IF (do_bcast) THEN
         CALL mp_ibcast(meta_size_diag, pdiag, mygrp, request_meta_size_diag)
         CALL mp_ibcast(data_size_diag, pdiag, mygrp, request_data_size_diag)
      ENDIF
      !
      ! Exchange refs
      IF (is_left) THEN
         CALL mp_iallgather(local_refs_meta_size, left_refs_meta_size, mygrp, requests(irequests))
         CALL mp_iallgather(local_refs_displ_unmerged, left_refs_displ_unmerged, mygrp, requests(2+irequests))
      ELSE
         CALL mp_iallgather(local_refs_meta_size, right_refs_meta_size, mygrp, requests(irequests))
         CALL mp_iallgather(local_refs_displ_unmerged, right_refs_displ_unmerged, mygrp, requests(2+irequests))
      ENDIF
      CALL mp_iallgather(local_no_empty_images, no_empty_images, mygrp, request_no_empty_images)
      !
      ! Allocate data and meta buffers
      do_win_create(:) = .TRUE.
      IF (do_win_free) THEN
         do_win_create(1) = dbcsr_data_get_size(buffer%data) .LT. (data_size+10)
         do_win_create(2) = SIZE(buffer%meta) .LT. meta_size
         CALL mp_sum(do_win_create, mygrp)
         IF (do_win_create(1)) CALL mp_win_free(buffer%data_win)
         IF (do_win_create(2)) CALL mp_win_free(buffer%meta_win)
      ENDIF
      CALL buffer_init(buffer, data_type, &
                       data_size, meta_size, &
                       data_memory_type=memtype_mpi_buffer)
      IF (do_win_create(1)) CALL dbcsr_win_create_any(buffer%data, mygrp, buffer%data_win)
      IF (do_win_create(2)) CALL mp_win_create(buffer%meta, mygrp, buffer%meta_win)
      !
      IF (is_diagonal .AND. do_bcast) THEN
         CALL resize_buffer_diag(buffer, data_type, &
                                 data_size_diag, meta_size_diag, &
                                 data_memory_type=memtype_mpi_buffer)
      ENDIF
      !
      ! Change displacements for the first thread
      ! In this case they refer to position 1
      refs(imeta_size, :, :, 0, ilocal_proc) = local_refs_displ(imeta_displ, :, :)+1
      refs(idata_size, :, :, 0, ilocal_proc) = local_refs_displ(idata_displ, :, :)+1
      !
      ! Evaluate data and meta diplacements per each thread
      DO it = 1, nthreads-1
         refs(:, :, :, it, ilocal_proc) = refs(:, :, :, it, ilocal_proc)+refs(:, :, :, it-1, ilocal_proc)
      ENDDO
      !
      ! Take offsets for symmetric case
      IF (do_symmetry .AND. .NOT. is_diagonal) THEN
         sym_size(:) = 1
         recv_sym_size(:) = 1
         DO ui = 1, nimages_unmerged
            DO mi = 1, nimages_merged
               ! Empty cluster
               IF (local_refs_size(imeta_size, mi, ui) .EQ. 0) THEN
                  DO it = 0, nthreads
                     recv_refs(:, mi, ui, it) = recv_sym_size(:)
                  ENDDO
               ELSE
                  ! Carry previous value
                  recv_refs(:, mi, ui, 0) = recv_sym_size(:)
                  DO it = 1, nthreads
                     recv_refs(:, mi, ui, it) = &
                        recv_refs(:, mi, ui, it)+recv_refs(:, mi, ui, it-1)
                     ! Add symetric events to the diplacements of local data
                     refs(:, mi, ui, it, ilocal_proc) = refs(:, mi, ui, it, ilocal_proc)+ &
                                                        recv_refs(:, mi, ui, it)-recv_sym_size(:)
                  ENDDO
                  recv_sym_size(:) = recv_refs(:, mi, ui, nthreads)
               ENDIF
               refs(:, mi, ui, 0, isym_proc) = sym_size(:)
               DO it = 1, nthreads
                  refs(:, mi, ui, it, isym_proc) = &
                     refs(:, mi, ui, it, isym_proc)+refs(:, mi, ui, it-1, isym_proc)
               ENDDO
               sym_size(:) = refs(:, mi, ui, nthreads, isym_proc)
            ENDDO
         ENDDO
         !
         ! Allocate data/meta to send
         CALL dbcsr_data_init(data_send)
         CALL dbcsr_data_new(data_send, data_type, sym_size(idata_size)-1)
         ALLOCATE (meta_send(sym_size(imeta_size)-1))
      ENDIF
!$OMP END MASTER
!$OMP BARRIER
      !
      data_buffer_p => buffer%data
      meta_buffer_p => buffer%meta
      !
      ! Copy data and meta in the buffers
      CALL dbcsr_iterator_start(iter, matrix, shared=.TRUE.)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, blk, blk_p=blk_p, &
                                        row_size=row_size, col_size=col_size)
         nze = row_size*col_size
         IF (nze .EQ. 0) CYCLE
         bp = ABS(blk_p)
         DO symmetry_i = 1, nsymmetries
            IF (symmetry_i .EQ. 1) THEN
               stored_row = row; stored_col = col; tr = blk_p .LT. 0
               tr_row_size = col_size; tr_col_size = row_size
            ELSE
               IF (row .EQ. col) CYCLE
               stored_row = col; stored_col = row; tr = blk_p .GT. 0
               tr_row_size = row_size; tr_col_size = col_size
            ENDIF
            row_img = row_img_dist(stored_row)
            col_img = col_img_dist(stored_col)
            CALL image_calculator(imgdist, &
                                  prow=prow, pcol=pcol, &
                                  rowi=rowi, coli=coli, &
                                  myprow=row_dist(stored_row), myrowi=row_img, &
                                  mypcol=col_dist(stored_col), mycoli=col_img, &
                                  shifting='0')
            IF (do_symmetry .OR. is_diagonal) THEN
               IF (is_diagonal) THEN
                  IF (do_bcast .AND. (ui .EQ. mi .OR. nimages_merged .EQ. 1)) THEN
                     data_buffer_p => buffer%data_diag
                     meta_buffer_p => buffer%meta_diag
                  ELSE
                     data_buffer_p => buffer%data
                     meta_buffer_p => buffer%meta
                  ENDIF
               ELSEIF (blacs2mpi(prow, pcol) .EQ. mynode) THEN
                  dst_proc = ilocal_proc
                  data_buffer_p => buffer%data
                  meta_buffer_p => buffer%meta
               ELSE
                  ! Move data to remove symmetric proc
                  dst_proc = isym_proc
                  data_buffer_p => data_send
                  meta_buffer_p => meta_send
               ENDIF
            ENDIF
!$          IF (is_left .AND. do_symmetry) THEN
!$             myt = threads_dist(stored_row)
!$             call omp_set_lock(locks(myt))
!$          ENDIF
            IF (tr) THEN
               CALL dbcsr_block_transpose_aa(data_buffer_p, sm%data_area, tr_row_size, tr_col_size, &
                                             refs(idata_size, mi, ui, myt, dst_proc), bp, &
                                             scale_value)
               IF (sm%negate_real .AND. sm%negate_imaginary) THEN
                  CALL dbcsr_block_scale(data_buffer_p, scale=scale_neg_one, &
                                         row_size=nze, col_size=1, &
                                         lb=refs(idata_size, mi, ui, myt, dst_proc))
               ELSEIF (sm%negate_real) THEN
                  CALL dbcsr_block_real_neg(data_buffer_p, row_size=nze, col_size=1, &
                                            lb=refs(idata_size, mi, ui, myt, dst_proc))
               ELSEIF (sm%negate_imaginary) THEN
                  CALL dbcsr_block_conjg(data_buffer_p, row_size=nze, col_size=1, &
                                         lb=refs(idata_size, mi, ui, myt, dst_proc))
               ENDIF
            ELSE
               CALL dbcsr_block_copy_aa(data_buffer_p, sm%data_area, row_size, col_size, &
                                        refs(idata_size, mi, ui, myt, dst_proc), bp, &
                                        scale_value)
            ENDIF
            !
            ! Set meta data (local indexing)
            IF (do_symmetry .AND. blacs2mpi(prow, pcol) .NE. mynode) THEN
               ! For the symmetric case we need to swap rows and cols maps
               IF (is_left) THEN
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)) = g2l_map_cols_merge(stored_row)
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+1) = g2l_map_rows(stored_col)
               ELSE
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)) = g2l_map_cols(stored_row)
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+1) = g2l_map_rows_merge(stored_col)
               ENDIF
               ! The offset will be adjusted on the symmetric proc
               meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+2) = &
                  refs(idata_size, mi, ui, myt, dst_proc)
            ELSE
               IF (is_left) THEN
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+size_index_unmerged) = &
                     g2l_map_rows_merge(stored_row)
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+size_index_unmerged+1) = &
                     g2l_map_cols(stored_col)
               ELSE
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+size_index_unmerged) = &
                     g2l_map_rows(stored_row)
                  meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+size_index_unmerged+1) = &
                     g2l_map_cols_merge(stored_col)
               ENDIF
               meta_buffer_p(refs(imeta_size, mi, ui, myt, dst_proc)+size_index_unmerged+2) = &
                  refs(idata_size, mi, ui, myt, dst_proc)-local_refs_displ(idata_displ, mi, ui)
            ENDIF
            refs(imeta_size, mi, ui, myt, dst_proc) = refs(imeta_size, mi, ui, myt, dst_proc)+3
            refs(idata_size, mi, ui, myt, dst_proc) = refs(idata_size, mi, ui, myt, dst_proc)+nze
!$          IF (is_left .AND. do_symmetry) THEN
!$             call omp_unset_lock(locks(myt))
!$          ENDIF
         ENDDO
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      !
      IF (do_symmetry .AND. .NOT. is_diagonal) THEN
!$OMP BARRIER
!$OMP MASTER
         ! Send/recv data to the symmetric proc
         CALL dbcsr_data_init(data_recv)
         CALL dbcsr_data_new(data_recv, data_type, recv_sym_size(idata_size)-1)
         CALL dbcsr_sendrecv_any(data_send, sym_p, data_recv, sym_p, grp)
         ALLOCATE (meta_recv(recv_sym_size(imeta_size)-1))
         CALL mp_sendrecv(meta_send, sym_p, meta_recv, sym_p, grp)
         ! Delete send data
         CALL dbcsr_data_release(data_send)
         DEALLOCATE (meta_send)
!$OMP END MASTER
      ENDIF
!$OMP BARRIER
!$    IF (is_left .AND. do_symmetry) THEN
!$       call omp_destroy_lock(locks(ithread))
!$    ENDIF
      !
      meta_buffer_p => buffer%meta
      !
      ! Set stats slots
!$OMP DO schedule(dynamic)
      DO ui = 1, nimages_unmerged
         IF (is_left) THEN
            coli => ui
         ELSE
            rowi => ui
         ENDIF
         nblocks = 0
         DO mi = 1, nimages_merged
            IF (local_refs_size(imeta_size, mi, ui) .EQ. 0) CYCLE
            !
            ! Insert symmetric data in local buffers
            IF (do_symmetry .AND. .NOT. is_diagonal) THEN
               IF (recv_refs(imeta_size, mi, ui, nthreads) .GT. 1) THEN
                  DO it = 0, nthreads-1
                     ! Copy meta, block by block
                     DO blk = recv_refs(imeta_size, mi, ui, it), recv_refs(imeta_size, mi, ui, it+1)-1, 3
                        buffer%meta(refs(imeta_size, mi, ui, it, ilocal_proc)+size_index_unmerged) = &
                           meta_recv(blk)
                        buffer%meta(refs(imeta_size, mi, ui, it, ilocal_proc)+size_index_unmerged+1) = &
                           meta_recv(blk+1)
                        buffer%meta(refs(imeta_size, mi, ui, it, ilocal_proc)+size_index_unmerged+2) = &
                           meta_recv(blk+2)-recv_refs(idata_size, mi, ui, it)+ &
                           refs(idata_size, mi, ui, it, ilocal_proc)-local_refs_displ(idata_displ, mi, ui)
                        refs(imeta_size, mi, ui, it, ilocal_proc) = &
                           refs(imeta_size, mi, ui, it, ilocal_proc)+3
                     ENDDO
                     ! Copy data
                     CALL dbcsr_data_set(buffer%data, refs(idata_size, mi, ui, it, ilocal_proc), &
                                         recv_refs(idata_size, mi, ui, it+1)-recv_refs(idata_size, mi, ui, it), &
                                         data_recv, recv_refs(idata_size, mi, ui, it))
                  ENDDO
               ENDIF
            ENDIF
            IF (is_diagonal) THEN
               IF (do_bcast .AND. (ui .EQ. mi .OR. nimages_merged .EQ. 1)) THEN
                  meta_buffer_p => buffer%meta_diag
               ELSE
                  meta_buffer_p => buffer%meta
               ENDIF
            ENDIF
            meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_size) = &
               local_refs_size(imeta_size, mi, ui)
            meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_nblks) = &
               (local_refs_size(imeta_size, mi, ui)-size_index_unmerged)/3
            nblocks = nblocks+meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_nblks)
            IF (is_left) THEN
               ! first thread
!$             meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_nblks+1) = 0
!$             DO it = 0, nthreads-1
!$                meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_nblks+it+2) = &
!$                   (refs(imeta_size, mi, ui, it, ilocal_proc)-local_refs_displ(imeta_displ, mi, ui)-1)/3
!$             ENDDO
               ! Count the maximum possible multiplies per row for on-the-fly filtering
               IF (otf_filtering) THEN
                  DO row = local_refs_displ(imeta_displ, mi, ui)+size_index_unmerged+1, &
                           local_refs_displ(imeta_displ, mi, ui)+ &
                           meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_size), 3
!$OMP ATOMIC
                     left_total_row_counts(meta_buffer_p(row)) = &
                        left_total_row_counts(meta_buffer_p(row))+1
                  ENDDO
               ENDIF
            ELSEIF (do_scatter .AND. is_diagonal .AND. (ui .EQ. mi .OR. nimages_merged .EQ. 1)) THEN
               ! Prepare meta buffer to scatter
               CALL set_empty_meta_index( &
                  meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_size: &
                               meta_displ_cluster_scatter(mi)+dbcsr_num_slots), &
                  global_indices, dbcsr_num_slots)
               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_size) = &
                  meta_size_cluster_scatter(mi)
               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_nblks) = &
                  meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_nblks)
               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_nze) = &
                  local_refs_size(idata_size, mi, ui)
               CALL image_calculator(imgdist, &
                                     prow=meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_prow), &
                                     pcol=meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_pcol), &
                                     rowi=meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_rowi), &
                                     coli=meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_coli), &
                                     vprow=meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_vprow), &
                                     myrowi=rowi, mycoli=coli, &
                                     shifting='0')
               ! Considering merging of the rows
               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_vpcol) = &
                  meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_home_pcol)
               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_num_slots) = &
                  meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_size)
               IF (meta_size_cluster_scatter(mi) .GT. dbcsr_num_slots) THEN
                  meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_num_slots+1: &
                               meta_displ_cluster_scatter(mi)+ &
                               meta_scatter(meta_displ_cluster_scatter(mi)+dbcsr_slot_size)) = &
                     meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+size_index_unmerged+1: &
                                   local_refs_displ(imeta_displ, mi, ui)+ &
                                   meta_buffer_p(local_refs_displ(imeta_displ, mi, ui)+dbcsr_slot_size))
               ENDIF
            ENDIF
         ENDDO
         !
         max_nblocks = MAX(max_nblocks, nblocks)
      ENDDO
!$OMP END DO
!$OMP END PARALLEL
      DEALLOCATE (refs)
      !
      IF (is_left .AND. otf_filtering) THEN
         CALL mp_isum(left_total_row_counts, mygrp, request_count_rows)
      ENDIF
      !
      IF (nimages_merged .EQ. 1) THEN
         IF (is_diagonal .AND. do_bcast) THEN
            meta_buffer_p => buffer%meta_diag
         ELSE
            meta_buffer_p => buffer%meta
         ENDIF
         CALL setup_rec_index(meta_buffer_p, img_refs_rows, img_refs_cols, &
                              local_refs_size(imeta_size, 1, :), local_refs_displ(imeta_displ, 1, :), &
                              size_index_unmerged, is_left)
      ENDIF
      !
      IF (do_symmetry .AND. .NOT. is_diagonal) THEN
         CALL dbcsr_data_release(data_recv)
         DEALLOCATE (meta_recv, recv_refs)
      ENDIF
      DEALLOCATE (g2l_map_rows, g2l_map_cols)
      DEALLOCATE (g2l_map_rows_merge, g2l_map_cols_merge)
!$    IF (is_left .AND. do_symmetry) THEN
!$       DEALLOCATE (locks)
!$    ENDIF
      !
      CALL mp_wait(request_no_empty_images)
      no_empty_images_displ(1) = 0
      DO iproc = 1, nprocs-1
         no_empty_images_displ(iproc+1) = no_empty_images_displ(iproc)+ &
                                          no_empty_images(iproc)
      ENDDO
      total_no_empty_images = no_empty_images_displ(nprocs)+ &
                              no_empty_images(nprocs)
      !
      IF (is_left) THEN
         ALLOCATE (left_refs_data_size(total_no_empty_images))
         CALL mp_iallgather(local_refs_data_size(1:local_no_empty_images), &
                            left_refs_data_size, no_empty_images, no_empty_images_displ, &
                            mygrp, requests(4+irequests))
      ELSE
         ALLOCATE (right_refs_data_size(total_no_empty_images))
         CALL mp_iallgather(local_refs_data_size(1:local_no_empty_images), &
                            right_refs_data_size, no_empty_images, no_empty_images_displ, &
                            mygrp, requests(4+irequests))
      ENDIF
      !
      ! Evaluate and exchange max norms for MPI filtering
      IF (otf_filtering .AND. use_mpi_filtering) THEN
         IF (is_left) THEN
            ALLOCATE (left_local_max_norms(local_no_empty_images))
            ALLOCATE (left_max_norms(total_no_empty_images))
            local_max_norms => left_local_max_norms
            max_norms => left_max_norms
         ELSE
            ALLOCATE (right_local_max_norms(local_no_empty_images))
            ALLOCATE (right_max_norms(total_no_empty_images))
            local_max_norms => right_local_max_norms
            max_norms => right_max_norms
         ENDIF
         CALL calculate_max_image_norms(buffer, &
                                        local_refs_size(imeta_size, :, :), &
                                        local_refs_displ, &
                                        img_map, &
                                        img_refs, &
                                        dbcsr_row_block_sizes(matrix), &
                                        dbcsr_col_block_sizes(matrix), &
                                        local_rows, &
                                        local_cols, &
                                        size_index_unmerged+1, &
                                        local_max_norms, is_left, &
                                        .NOT. (is_diagonal .AND. do_bcast))
         CALL mp_iallgather(local_max_norms, max_norms, &
                            no_empty_images, no_empty_images_displ, &
                            mygrp, requests(6+irequests))
      ENDIF
      !
      DEALLOCATE (img_refs_cols)
      DEALLOCATE (img_refs)
      DEALLOCATE (img_map)
      DEALLOCATE (local_refs_displ)
      DEALLOCATE (img_refs_rows)
      DEALLOCATE (local_refs_size)
      !
      IF (do_bcast) THEN
         CALL mp_wait(request_meta_size_diag)
         CALL mp_wait(request_data_size_diag)
         IF (.NOT. is_diagonal) THEN
            CALL resize_buffer_diag(buffer, data_type, &
                                    data_size_diag, meta_size_diag, &
                                    data_memory_type=memtype_mpi_buffer)
         ENDIF
         !
         ! Bcast the diagonal data
         CALL mp_ibcast(buffer%meta_diag, pdiag, mygrp, requests_diag(irequests))
         CALL dbcsr_ibcast_any(buffer%data_diag, pdiag, mygrp, requests_diag(2+irequests))
      ENDIF
      !
      ! Scatter the diagonal data
      IF (do_scatter) THEN
         CALL mp_wait(request_meta_size_scatter)
         CALL mp_wait(request_data_size_scatter)
         CALL ensure_array_size(local_meta_scatter, ub=local_meta_size_scatter, &
                                nocopy=.TRUE., memory_type=memtype_mpi_buffer)
         IF (dbcsr_data_valid(local_data_scatter)) THEN
            CALL dbcsr_data_ensure_size(local_data_scatter, local_data_size_scatter, &
                                        nocopy=.TRUE.)
         ELSE
            CALL dbcsr_data_init(local_data_scatter)
            CALL dbcsr_data_new(local_data_scatter, data_type=data_type, &
                                data_size=local_data_size_scatter, memory_type=memtype_mpi_buffer)
            CALL dbcsr_data_set_size_referenced(local_data_scatter, local_data_size_scatter)
         ENDIF
         !
         CALL mp_iscatter(meta_scatter, meta_size_scatter, meta_displ_scatter, &
                          local_meta_scatter, local_meta_size_scatter, &
                          myprow, mygrp_scatter, requests_scatter(1))
         CALL dbcsr_iscatterv_any(buffer%data_diag, data_size_scatter, data_displ_scatter, &
                                  local_data_scatter, local_data_size_scatter, &
                                  myprow, mygrp_scatter, requests_scatter(2))
         !
         IF (is_diagonal) THEN
            DEALLOCATE (meta_displ_cluster_scatter, meta_size_cluster_scatter)
            DEALLOCATE (meta_size_scatter, data_size_scatter, &
                        meta_displ_scatter, data_displ_scatter)
         ENDIF
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE make_buffers

! **************************************************************************************************
!> \brief Creates row and column images of a matrix.
!> \param[in] source          input matrix
!> \param[in,out] normalized  image array of the normalized matrix
!> \param[in] target_image_dist          normalize to this image distribution
!> \param[in] predistribute   (optional) predistribute data for multiplication
!> \param[in] no_copy_data    (optional) try to not merge data at the end
!> \param[in] scale_value     (optional) scale with this value
! **************************************************************************************************
   SUBROUTINE dbcsr_make_images(source, normalized, target_image_dist, &
                                predistribute, no_copy_data, scale_value)
      TYPE(dbcsr_obj), INTENT(IN)                        :: source
      TYPE(dbcsr_2d_array_type), INTENT(OUT)             :: normalized
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: target_image_dist
      CHARACTER, INTENT(IN), OPTIONAL                    :: predistribute
      LOGICAL, INTENT(IN), OPTIONAL                      :: no_copy_data
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale_value

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (my_use_mpi_exp) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2407,"RMA algo not supported here!")
      IF (.NOT. dbcsr_valid_index(source)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2409,"Matrix not initialized.")
      CALL make_images(source, normalized, &
                       target_image_dist, desymmetrize=dbcsr_has_symmetry(source), &
                       predistribute=predistribute, &
                       no_copy_data=no_copy_data, &
                       scale_value=scale_value)
      normalized%image_dist = target_image_dist
      CALL dbcsr_image_dist_hold(normalized%image_dist)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_make_images

! **************************************************************************************************
!> \brief Makes column-based and row-based images of a matrix.
!> \param[in] ism input symmetric matrix
!> \param[in,out] ums         normalied matrices
!> \param[in] target_imgdist  image distribution to normalize to
!> \param[in] desymmetrize    (optional) desymmetrize a symmetric matrix
!> \param[in] predistribute   (optional) predistribute data for multiplication
!> \param[in] no_copy_data    (optional) try to not merge data at the end
!> \param[in] scale_value     (optional) scale with this value
! **************************************************************************************************
   SUBROUTINE make_images(ism, ums, target_imgdist, desymmetrize, predistribute, &
                          no_copy_data, scale_value)
      TYPE(dbcsr_obj), INTENT(IN)                        :: ism
      TYPE(dbcsr_2d_array_type), INTENT(OUT)             :: ums
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: target_imgdist
      LOGICAL, INTENT(IN), OPTIONAL                      :: desymmetrize
      CHARACTER, INTENT(IN), OPTIONAL                    :: predistribute
      LOGICAL, INTENT(IN), OPTIONAL                      :: no_copy_data
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale_value

      CHARACTER(len=*), PARAMETER :: routineN = 'make_images', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: metalen = 5
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      CHARACTER                                          :: predist_type, predist_type_fwd
      INTEGER :: blk, blk_l, blk_p, bp, col, col_img, col_size, coli, data_p, data_type, dst_p, &
         handle, handle2, ithread, mp_group, ncol_images, nrow_images, nsymmetries, nthreads, &
         numproc, nze, pcol, prev_blk_p, prow, row, row_img, row_size, rowi, sd_pos, sm_pos, &
         src_p, stored_blk_p, stored_col, stored_row, symmetry_i, tr_col_size, tr_row_size, vcol, &
         vrow
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: myt_sdp, myt_smp, rd_disp, recv_meta, &
                                                            rm_disp, sd_disp, sdp, send_meta, &
                                                            sm_disp, smp
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: all_total_send_offset, blk_ps, blks, &
                                                            myt_total_send_count, &
                                                            total_recv_count, total_send_count
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :)        :: myt_send_count, recv_count, send_count
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_dist, col_img_dist, &
                                                            row_blk_size, row_dist, row_img_dist
      INTEGER, DIMENSION(:, :), POINTER                  :: blacs2mpi
      LOGICAL                                            :: nocopy, tr
      TYPE(dbcsr_data_obj)                               :: received_data_area, recv_data_area, &
                                                            send_data_area
      TYPE(dbcsr_distribution_obj)                       :: old_dist, target_dist
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_memtype_type)                           :: data_memory_type
      TYPE(dbcsr_mp_obj)                                 :: mp_obj
      TYPE(dbcsr_scalar_type)                            :: scale_neg_one
      TYPE(dbcsr_type)                                   :: sm

!   ---------------------------------------------------------------------------
! Check input matrix
! Set convenient variables to access input matrix info
!

      CALL timeset(routineN, handle)
      nocopy = .FALSE.
      IF (PRESENT(no_copy_data)) nocopy = no_copy_data
      sm = ism%m
      nsymmetries = 1
      IF (PRESENT(desymmetrize)) THEN
         IF (desymmetrize .AND. sm%symmetry) THEN
            nsymmetries = 2
         ENDIF
      ENDIF
      SELECT CASE (predistribute)
      CASE ('L', 'l')
         predist_type = 'L'
         predist_type_fwd = 'l'
      CASE ('R', 'r')
         predist_type = 'R'
         predist_type_fwd = 'r'
      CASE default
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2493,"Incorrect pre-shift specifier.")
      END SELECT
      data_type = sm%data_type
      IF (data_type .NE. dbcsr_type_real_8 .AND. &
          data_type .NE. dbcsr_type_real_4 .AND. &
          data_type .NE. dbcsr_type_complex_8 .AND. &
          data_type .NE. dbcsr_type_complex_4) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2500,"Invalid data type.")
      scale_neg_one = dbcsr_scalar_negative(dbcsr_scalar_one(data_type))
      row_blk_size => array_data(sm%row_blk_size)
      col_blk_size => array_data(sm%col_blk_size)
      old_dist = dbcsr_distribution(ism)
      target_dist = target_imgdist%i%main
      row_dist => dbcsr_distribution_row_dist(target_dist)
      col_dist => dbcsr_distribution_col_dist(target_dist)
      IF (sm%symmetry) THEN
         IF (SIZE(row_dist) .NE. SIZE(col_dist)) &
            CALL cp__w("dbcsr/mm/dbcsr_mm_cannon.F",2510,'Unequal row and column distributions for symmetric matrix.')
      ENDIF
      nrow_images = target_imgdist%i%row_decimation
      IF (nrow_images .GT. 1) THEN
         row_img_dist => array_data(target_imgdist%i%row_image)
      ELSE
         NULLIFY (row_img_dist)
      ENDIF
      ncol_images = target_imgdist%i%col_decimation
      IF (ncol_images .GT. 1) THEN
         col_img_dist => array_data(target_imgdist%i%col_image)
      ELSE
         NULLIFY (col_img_dist)
      ENDIF
      mp_obj = dbcsr_distribution_mp(target_dist)
      blacs2mpi => dbcsr_mp_pgrid(mp_obj)
      numproc = dbcsr_mp_numnodes(mp_obj)
      mp_group = dbcsr_mp_group(mp_obj)
      IF (MAXVAL(row_dist) .GT. UBOUND(blacs2mpi, 1)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2529,'Row distribution references unexistent processor rows')
      IF (dbg) THEN
         IF (MAXVAL(row_dist) .NE. UBOUND(blacs2mpi, 1)) &
            CALL cp__w("dbcsr/mm/dbcsr_mm_cannon.F",2532,'Range of row distribution not equal to processor rows')
      ENDIF
      IF (MAXVAL(col_dist) .GT. UBOUND(blacs2mpi, 2)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2535,'Col distribution references unexistent processor cols')
      IF (dbg) THEN
         IF (MAXVAL(col_dist) .NE. UBOUND(blacs2mpi, 2)) &
            CALL cp__w("dbcsr/mm/dbcsr_mm_cannon.F",2538,'Range of col distribution not equal to processor cols')
      ENDIF

      ! Check threads configuration
!$    IF (.NOT. dbcsr_distribution_has_threads(old_dist)) &
!$       CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",2543,"Thread distribution not defined")

      ! Allocate shared temporary buffers
      !
      ALLOCATE (send_count(2, nrow_images, ncol_images, 0:numproc-1)); send_count = 0
      ALLOCATE (recv_count(2, nrow_images, ncol_images, 0:numproc-1))
      ALLOCATE (total_send_count(2, 0:numproc-1)); total_send_count = 0
      ALLOCATE (total_recv_count(2, 0:numproc-1))
      ALLOCATE (sdp(0:numproc-1))
      ALLOCATE (smp(0:numproc-1))
      ALLOCATE (sd_disp(0:numproc-1)); sd_disp(0) = 1
      ALLOCATE (sm_disp(0:numproc-1)); sm_disp(0) = 1
      ALLOCATE (rd_disp(0:numproc-1)); rd_disp(0) = 1
      ALLOCATE (rm_disp(0:numproc-1)); rm_disp(0) = 1
      ALLOCATE (all_total_send_offset(2, 0:numproc-1))
      ALLOCATE (blk_ps(nrow_images, ncol_images)); blk_ps = 1
      ALLOCATE (blks(nrow_images, ncol_images)); blks = 1
      !
      ! Allocate and init mats
      !
      ALLOCATE (ums%mats(nrow_images, ncol_images))
      data_memory_type = memtype_abpanel_1
      DO row_img = 1, nrow_images
         DO col_img = 1, ncol_images
            CALL dbcsr_init(ums%mats(row_img, col_img))
            CALL dbcsr_create(ums%mats(row_img, col_img), "imaged "//sm%name, &
                              target_dist, &
                              dbcsr_type_no_symmetry, array_data(sm%row_blk_size), array_data(sm%col_blk_size), &
                              sm%row_blk_size, sm%col_blk_size, &
                              nze=0, data_type=data_type, &
                              max_rbs=sm%max_rbs, max_cbs=sm%max_cbs, &
                              row_blk_offset=sm%row_blk_offset, col_blk_offset=sm%col_blk_offset, &
                              thread_dist=sm%dist, &
                              data_memory_type=data_memory_type, &
                              index_memory_type=memtype_mpi_buffer)
            ums%mats(row_img, col_img)%m%negate_real = sm%negate_real
            ums%mats(row_img, col_img)%m%negate_imaginary = sm%negate_imaginary
         ENDDO
      ENDDO

      nthreads = 1
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP PRIVATE (ithread, symmetry_i, row_img, col_img, &
!$OMP          myt_send_count, myt_total_send_count, &
!$OMP          iter, row, col, blk, row_size, col_size, stored_row, stored_col, &
!$OMP          prow, pcol, rowi, coli, vrow, vcol, dst_p, nze, myt_smp, myt_sdp, &
!$OMP          blk_p, bp, sd_pos, sm_pos,tr, &
!$OMP          tr_row_size, tr_col_size) &
!$OMP SHARED (nthreads, nsymmetries, row_img_dist, col_img_dist, &
!$OMP         nrow_images, ncol_images, numproc, scale_value, &
!$OMP         ums, sm, ism, target_imgdist, row_dist, col_dist,&
!$OMP         predist_type_fwd, blacs2mpi, row_blk_size, col_blk_size, &
!$OMP         send_count, recv_count, handle2,mp_group, &
!$OMP         total_send_count, total_recv_count, recv_data_area, nocopy, &
!$OMP         data_type, recv_meta, send_data_area, send_meta, &
!$OMP         sd_disp, sm_disp, rd_disp, rm_disp, all_total_send_offset, blk_ps, blks, &
!$OMP         received_data_area, scale_neg_one, memtype_abpanel_1)
      ithread = 0
!$    ithread = omp_get_thread_num()
!$OMP MASTER
!$    nthreads = omp_get_num_threads()
!$OMP END MASTER

      ! Allocate thread private data
      !
      ALLOCATE (myt_send_count(2, nrow_images, ncol_images, 0:numproc-1)); myt_send_count(:, :, :, :) = 0
      ALLOCATE (myt_total_send_count(2, 0:numproc-1))
      ! Thread-local pointers of the current adding position into the send buffers
      ALLOCATE (myt_smp(0:numproc-1), myt_sdp(0:numproc-1))

      ! Count sizes for sending
      !
      CALL dbcsr_iterator_start(iter, ism, shared=.TRUE.)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, blk, &
                                        row_size=row_size, col_size=col_size)
         nze = row_size*col_size
         IF (nze .EQ. 0) CYCLE
         DO symmetry_i = 1, nsymmetries
            IF (symmetry_i .EQ. 1) THEN
               stored_row = row; stored_col = col
            ELSE
               IF (row .EQ. col) CYCLE
               stored_row = col; stored_col = row
            ENDIF
            ! Where do we send this block?
            row_img = 1
            IF (nrow_images .GT. 1) row_img = row_img_dist(stored_row)
            col_img = 1
            IF (ncol_images .GT. 1) col_img = col_img_dist(stored_col)
            CALL image_calculator(target_imgdist, &
                                  prow=prow, rowi=rowi, &
                                  pcol=pcol, coli=coli, &
                                  vprow=vrow, vpcol=vcol, &
                                  myprow=row_dist(stored_row), myrowi=row_img, &
                                  mypcol=col_dist(stored_col), mycoli=col_img, &
                                  shifting=predist_type_fwd)
            dst_p = blacs2mpi(prow, pcol)
            ! These counts are meant for the thread that processes this row.
            myt_send_count(1, rowi, coli, dst_p) = &
               myt_send_count(1, rowi, coli, dst_p)+1
            myt_send_count(2, rowi, coli, dst_p) = &
               myt_send_count(2, rowi, coli, dst_p)+nze
         END DO ! symmetry_i
      END DO
      CALL dbcsr_iterator_stop(iter)
      DO dst_p = 0, numproc-1
         myt_total_send_count(1, dst_p) = SUM(myt_send_count(1, :, :, dst_p))
         myt_total_send_count(2, dst_p) = SUM(myt_send_count(2, :, :, dst_p))
      END DO
      ! Merge the send counts
!$OMP CRITICAL
      send_count(:, :, :, :) = send_count(:, :, :, :)+myt_send_count(:, :, :, :)
      total_send_count(:, :) = total_send_count(:, :)+myt_total_send_count(:, :)
!$OMP END CRITICAL
!$OMP BARRIER
!$OMP MASTER
      CALL timeset(routineN//"_sizes", handle2)
      CALL mp_alltoall(send_count, recv_count, 2*nrow_images*ncol_images, &
                       mp_group)
      CALL timestop(handle2)
!$OMP END MASTER
!$OMP BARRIER
      ! Fill in the meta data structures and copy the data.
!$OMP DO
      DO dst_p = 0, numproc-1
         total_recv_count(1, dst_p) = SUM(recv_count(1, :, :, dst_p))
         total_recv_count(2, dst_p) = SUM(recv_count(2, :, :, dst_p))
      ENDDO
!$OMP MASTER
      ! Allocate data structures needed for data exchange.
      CALL dbcsr_data_init(recv_data_area)
      IF (nrow_images .EQ. 1 .AND. ncol_images .EQ. 1 .OR. nocopy) THEN
         ! For some cases the faster dbcsr_special_finalize(reshuffle=.FALSE.) can be used.
         ! This basically makes this working matrix the actual data-area.
         ! Hence, for those cases we have to use data_memory_type already here.
         CALL dbcsr_data_new(recv_data_area, data_type, SUM(total_recv_count(2, :)), memory_type=memtype_abpanel_1)
      ELSE
         CALL dbcsr_data_new(recv_data_area, data_type, SUM(total_recv_count(2, :)))
      END IF
      ALLOCATE (recv_meta(metalen*SUM(total_recv_count(1, :))))
      CALL dbcsr_data_init(send_data_area)
      CALL dbcsr_data_new(send_data_area, data_type, SUM(total_send_count(2, :)))
      ALLOCATE (send_meta(metalen*SUM(total_send_count(1, :))))
      ! Calculate displacements for processors needed for the exchanges.
      DO dst_p = 1, numproc-1
         sm_disp(dst_p) = sm_disp(dst_p-1) &
                          +metalen*total_send_count(1, dst_p-1)
         sd_disp(dst_p) = sd_disp(dst_p-1) &
                          +total_send_count(2, dst_p-1)
         rm_disp(dst_p) = rm_disp(dst_p-1) &
                          +metalen*total_recv_count(1, dst_p-1)
         rd_disp(dst_p) = rd_disp(dst_p-1) &
                          +total_recv_count(2, dst_p-1)
      ENDDO
      myt_smp(:) = sm_disp(:)
      myt_sdp(:) = sd_disp(:)
      IF (nthreads .GT. 1) THEN
         all_total_send_offset(1, :) = myt_smp(:)+metalen*myt_total_send_count(1, :)
         all_total_send_offset(2, :) = myt_sdp(:)+myt_total_send_count(2, :)
      ENDIF
!$OMP END MASTER
!$OMP BARRIER
      IF (ithread .GT. 0) THEN
!$OMP CRITICAL
         myt_smp(:) = all_total_send_offset(1, :)
         myt_sdp(:) = all_total_send_offset(2, :)
         all_total_send_offset(1, :) &
            = all_total_send_offset(1, :)+metalen*myt_total_send_count(1, :)
         all_total_send_offset(2, :) &
            = all_total_send_offset(2, :)+myt_total_send_count(2, :)
!$OMP END CRITICAL
      ELSE
         CALL dbcsr_data_init(received_data_area)
         received_data_area = recv_data_area
         CALL dbcsr_data_hold(received_data_area)
         DO row_img = 1, nrow_images
            DO col_img = 1, ncol_images
               CALL dbcsr_work_create(ums%mats(row_img, col_img), &
                                      SUM(recv_count(1, row_img, col_img, :)), n=1)
               CALL dbcsr_data_hold(received_data_area)
               CALL dbcsr_data_release(ums%mats(row_img, col_img)%m%wms(1)%data_area)
               ums%mats(row_img, col_img)%m%wms(1)%data_area = received_data_area
            ENDDO
         ENDDO
      ENDIF
!$OMP BARRIER
!$OMP MASTER
      CALL m_memory()
!$OMP END MASTER

      ! Add timing call to the packing of the send buffers
      !
      CALL timeset(routineN//"_pack", handle2)
      ! Copies metadata and actual data to be sent into the send buffers.
      CALL dbcsr_iterator_start(iter, ism, shared=.TRUE.)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, blk, blk_p=blk_p, &
                                        row_size=row_size, col_size=col_size)
         nze = row_size*col_size
         IF (nze .EQ. 0) CYCLE
         bp = ABS(blk_p)
         DO symmetry_i = 1, nsymmetries
            IF (symmetry_i .EQ. 1) THEN
               stored_row = row; stored_col = col; tr = blk_p .LT. 0
               tr_row_size = col_size; tr_col_size = row_size
            ELSE
               IF (row .EQ. col) CYCLE
               stored_row = col; stored_col = row; tr = blk_p .GT. 0
               tr_row_size = row_size; tr_col_size = col_size
            ENDIF
            ! Where do we send this block?
            row_img = 1
            IF (nrow_images .GT. 1) row_img = row_img_dist(stored_row)
            col_img = 1
            IF (ncol_images .GT. 1) col_img = col_img_dist(stored_col)
            CALL image_calculator(target_imgdist, &
                                  prow=prow, rowi=rowi, &
                                  pcol=pcol, coli=coli, &
                                  vprow=vrow, vpcol=vcol, &
                                  myprow=row_dist(stored_row), myrowi=row_img, &
                                  mypcol=col_dist(stored_col), mycoli=col_img, &
                                  shifting=predist_type_fwd)
            dst_p = blacs2mpi(prow, pcol)
            sm_pos = myt_smp(dst_p)
            myt_smp(dst_p) = myt_smp(dst_p)+metalen
            sd_pos = myt_sdp(dst_p)
            myt_sdp(dst_p) = myt_sdp(dst_p)+nze
            IF (tr) THEN
               CALL dbcsr_block_transpose_aa(send_data_area, sm%data_area, tr_row_size, tr_col_size, &
                                             sd_pos, bp, scale_value)
               IF (sm%negate_real .AND. sm%negate_imaginary) THEN
                  CALL dbcsr_block_scale(send_data_area, scale=scale_neg_one, &
                                         row_size=nze, col_size=1, lb=sd_pos)
               ELSEIF (sm%negate_real) THEN
                  CALL dbcsr_block_real_neg(send_data_area, &
                                            row_size=nze, col_size=1, lb=sd_pos)
               ELSEIF (sm%negate_imaginary) THEN
                  CALL dbcsr_block_conjg(send_data_area, &
                                         row_size=nze, col_size=1, lb=sd_pos)
               ENDIF
            ELSE
               CALL dbcsr_block_copy_aa(send_data_area, sm%data_area, row_size, col_size, &
                                        sd_pos, bp, scale_value)
            END IF

            send_meta(sm_pos) = stored_row
            send_meta(sm_pos+1) = stored_col
            send_meta(sm_pos+2) = sd_pos-sd_disp(dst_p)+1
            send_meta(sm_pos+3) = rowi
            send_meta(sm_pos+4) = coli
         ENDDO ! symmetry_i
      ENDDO ! iterator
      CALL dbcsr_iterator_stop(iter)

      ! Deallocate thread private data
      !
      DEALLOCATE (myt_send_count)
      DEALLOCATE (myt_total_send_count)
      DEALLOCATE (myt_smp, myt_sdp)

      CALL timestop(handle2)
!$OMP END PARALLEL

      ! Exchange the data and metadata structures. In the interesting cases (square grids, row col distribution same),
      ! there are only very few processors that need to exchange data.
      ! The hybrid_alltoall deals with this by doing point to point communication
      CALL timeset(routineN//"_data", handle2)
      CALL hybrid_alltoall_any(send_data_area, total_send_count(2, :), sd_disp(:)-1, &
                               recv_data_area, total_recv_count(2, :), rd_disp(:)-1, &
                               mp_obj, &
                               most_ptp=.TRUE., remainder_ptp=.TRUE., no_hybrid=.FALSE.)
      CALL hybrid_alltoall_i1( &
         send_meta(:), metalen*total_send_count(1, :), sm_disp(:)-1, &
         recv_meta(:), metalen*total_recv_count(1, :), rm_disp(:)-1, &
         most_ptp=.TRUE., remainder_ptp=.TRUE., no_hybrid=.FALSE., &
         mp_env=mp_obj)
      CALL timestop(handle2)

      ! Now create the work index and/or copy the relevant data from the
      ! receive buffer into the local indices.
      !
      prev_blk_p = 0
      DO src_p = 0, numproc-1
         data_p = 0
         DO blk_l = 1, total_recv_count(1, src_p)
            stored_row = recv_meta(rm_disp(src_p)+metalen*(blk_l-1))
            stored_col = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+1)
            stored_blk_p = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+2)
            row_img = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+3)
            col_img = recv_meta(rm_disp(src_p)+metalen*(blk_l-1)+4)
            nze = row_blk_size(ABS(stored_row)) &
                  *col_blk_size(ABS(stored_col))
            blk = blks(row_img, col_img)
            blks(row_img, col_img) = blks(row_img, col_img)+1
            IF (stored_col .GT. 0) THEN
               blk_p = data_p
               data_p = data_p+nze
            ELSE
               blk_p = prev_blk_p
            ENDIF
            blk_ps(row_img, col_img) = blk_ps(row_img, col_img)+nze
            ums%mats(row_img, col_img)%m%wms(1)%row_i(blk) = ABS(stored_row)
            ums%mats(row_img, col_img)%m%wms(1)%col_i(blk) = ABS(stored_col)
            ums%mats(row_img, col_img)%m%wms(1)%blk_p(blk) = &
               SIGN(rd_disp(src_p)+ABS(stored_blk_p)-1, stored_blk_p)
            prev_blk_p = blk_p
         ENDDO
      ENDDO

      ! Finalize the actual imaged matrices from the work matrices
      !
      DO row_img = 1, nrow_images
         DO col_img = 1, ncol_images
            ums%mats(row_img, col_img)%m%wms(1)%lastblk = blks(row_img, col_img)-1
            ums%mats(row_img, col_img)%m%wms(1)%datasize = blk_ps(row_img, col_img)-1
            !
            CALL dbcsr_data_set_size_referenced( &
               ums%mats(row_img, col_img)%m%wms(1)%data_area, &
               ums%mats(row_img, col_img)%m%wms(1)%datasize)
            IF (nrow_images .EQ. 1 .AND. ncol_images .EQ. 1 .OR. nocopy) THEN
               CALL dbcsr_special_finalize(ums%mats(row_img, col_img), reshuffle=.FALSE.)
            ELSE
               CALL dbcsr_special_finalize(ums%mats(row_img, col_img), reshuffle=.TRUE.)
            ENDIF

            ! Save the home process and image row and column
            CALL image_calculator(target_imgdist, &
                                  ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_prow), &
                                  ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_rowi), &
                                  ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_pcol), &
                                  ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_coli), &
                                  vprow=ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_vprow), &
                                  vpcol=ums%mats(row_img, col_img)%m%index(dbcsr_slot_home_vpcol), &
                                  myrowi=row_img, mycoli=col_img, &
                                  shifting=predist_type)
         ENDDO
      ENDDO

      ! Deallocate shared temporary buffers
      !
      DEALLOCATE (send_count, recv_count)
      DEALLOCATE (total_send_count, total_recv_count)
      DEALLOCATE (sdp, smp, sd_disp, sm_disp)
      DEALLOCATE (rd_disp, rm_disp)
      DEALLOCATE (all_total_send_offset)
      DEALLOCATE (blk_ps, blks)
      DEALLOCATE (recv_meta, send_meta)

      CALL dbcsr_data_release(send_data_area)
      CALL dbcsr_data_release(received_data_area)
      CALL dbcsr_data_release(recv_data_area)

      CALL timestop(handle)
   END SUBROUTINE make_images

! **************************************************************************************************
!> \brief Makes dense matrices for the image matrices.
!> \param[in,out] images          current (undense) matrix images, output is
!>                                the dense matrix images
!> \param[in] new_rdist           the new image distribution for dense matrices
!> \param[in] row_map             mapping of current (undense) rows to dense rows
!> \param[in] col_map             mapping of current (undense) columns to
!>                                dense columns
!> \param[in] join_cols           (optional) make columns dense, default is
!>                                yes
!> \param[in] join_rows           (optional) make rows dense, default is yes
!> \param[in] new_template        template dense matrix for creating image
!>                                matrices
!> \note Used for making matrices dense/undense
! **************************************************************************************************
   SUBROUTINE dbcsr_make_images_dense(images, new_rdist, &
                                      row_map, col_map, join_cols, join_rows, new_template)
      TYPE(dbcsr_2d_array_type), INTENT(INOUT)           :: images
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: new_rdist
      TYPE(array_i1d_obj), INTENT(IN)                    :: row_map, col_map
      LOGICAL, INTENT(IN)                                :: join_cols, join_rows
      TYPE(dbcsr_obj), INTENT(IN)                        :: new_template

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_images_dense', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, mat_col, mat_row, mat_vpcol, &
                                                            mat_vprow
      INTEGER, DIMENSION(:), POINTER                     :: und_col_blk_offsets, und_row_blk_offsets
      INTEGER, DIMENSION(dbcsr_meta_size)                :: old_meta
      REAL(kind=dp)                                      :: cs
      TYPE(array_i1d_obj)                                :: dense_local_vcols, dense_local_vrows, &
                                                            und_local_vcols, und_local_vrows
      TYPE(dbcsr_imagedistribution_obj)                  :: old_rdist
      TYPE(dbcsr_obj)                                    :: tmp_mat

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      old_rdist = images%image_dist
      !
      DO mat_row = 1, images%image_dist%i%row_decimation
         DO mat_col = 1, images%image_dist%i%col_decimation
            IF (dbg) THEN
               cs = dbcsr_checksum(images%mats(mat_row, mat_col))
               WRITE (*, *) routineN//" cs pre", cs
            ENDIF
            mat_vprow = images%mats(mat_row, mat_col)%m%index(dbcsr_slot_home_vprow)
            mat_vpcol = images%mats(mat_row, mat_col)%m%index(dbcsr_slot_home_vpcol)
            und_row_blk_offsets => array_data(images%mats(mat_row, mat_col)%m%row_blk_offset)
            und_col_blk_offsets => array_data(images%mats(mat_row, mat_col)%m%col_blk_offset)
            CALL dbcsr_get_local_vrows(old_rdist, und_local_vrows, mat_vprow)
            CALL dbcsr_get_local_vcols(old_rdist, und_local_vcols, mat_vpcol)

            CALL dbcsr_get_local_vrows(new_rdist, dense_local_vrows, mat_vprow)
            CALL dbcsr_get_local_vcols(new_rdist, dense_local_vcols, mat_vpcol)
            ! The old matrix has to be remembered so it is copied to
            ! tmp_mat.
            old_meta(:) = images%mats(mat_row, mat_col)%m%index(1:dbcsr_meta_size)
            CALL dbcsr_init(tmp_mat)
            tmp_mat = images%mats(mat_row, mat_col)
            CALL dbcsr_init(images%mats(mat_row, mat_col))
            CALL dbcsr_create(images%mats(mat_row, mat_col), template=new_template)
            images%mats(mat_row, mat_col)%m%index(dbcsr_slot_home_prow &
                                                  :dbcsr_slot_home_vpcol) = &
               old_meta(dbcsr_slot_home_prow:dbcsr_slot_home_vpcol)
            CALL dbcsr_make_dense_low(tmp_mat, images%mats(mat_row, mat_col), &
                                      array_data(und_local_vrows), array_data(und_local_vcols), &
                                      und_row_blk_offsets, und_col_blk_offsets, &
                                      array_data(dense_local_vrows), &
                                      array_data(dense_local_vcols), &
                                      array_data(new_template%m%row_blk_offset), &
                                      array_data(new_template%m%col_blk_offset), &
                                      array_data(row_map), array_data(col_map), join_rows, join_cols)
            !
            CALL dbcsr_index_prune_deleted(images%mats(mat_row, mat_col))
            !
            CALL dbcsr_release(tmp_mat)
            IF (dbg) THEN
               cs = dbcsr_checksum(images%mats(mat_row, mat_col))
               WRITE (*, *) routineN//" cs pst", cs
            ENDIF
         ENDDO
      ENDDO
      CALL dbcsr_image_dist_release(images%image_dist)
      images%image_dist = new_rdist
      CALL dbcsr_image_dist_hold(images%image_dist)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_make_images_dense

! **************************************************************************************************
!> \brief Multiplies two DBCSR matrices
!>
!> \param[in] left_set             set of imaged left matrices
!> \param[in] right_set            set of imaged right matrices
!> \param[out] product_matrix      DBCSR product matrix
!> \param[in] retain_sparsity      (optional) retain the sparsity of the
!>                                 existing product matrix; default is no
!> \param filter_eps ...
!> \param[out] flop                (optional) effective flop
! **************************************************************************************************
   SUBROUTINE cannon_multiply_low(left_set, right_set, product_matrix, &
                                  retain_sparsity, &
                                  filter_eps, flop)
      TYPE(dbcsr_2d_array_type), POINTER                 :: left_set, right_set
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: product_matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(kind=real_8), INTENT(in), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT)                   :: flop

      CHARACTER(len=*), PARAMETER :: routineN = 'cannon_multiply_low', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: idata = 1, ileft = 0, imeta = 2, &
                                                            iright = 2, M_L = 2, M_P = 1, M_R = 3

      INTEGER :: data_type, data_type_byte, grp, handle, handle2, handle3, i, ithread, &
         left_col_image, left_col_mult, left_col_nimages, left_dst_icol, left_dst_irow, &
         left_dst_p, left_dst_pcol, left_dst_prow, left_dst_vcol, left_dst_vrow, left_max_nblks, &
         left_max_nze, left_myfirstvcol, left_myfirstvrow, left_mypcol, left_myprow, left_npcols, &
         left_nprows, left_recv_icol, left_recv_irow, left_recv_p, left_recv_pcol, left_recv_prow, &
         left_recv_vcol, left_recv_vrow, left_row_image, left_row_mult, left_row_nimages, &
         left_send_icol, left_send_irow, left_send_p, left_send_pcol, left_send_prow
      INTEGER :: left_send_vcol, left_send_vrow, left_src_icol, left_src_irow, left_src_p, &
         left_src_pcol, left_src_prow, left_src_vcol, left_src_vrow, metronome, min_nimages, &
         mp_group, mynode, nblkrows_total, nblkrows_used, nsteps_k, nthreads, numnodes, nvirt_k, &
         output_unit, right_col_image, right_col_mult, right_col_nimages, right_dst_icol, &
         right_dst_irow, right_dst_p, right_dst_pcol, right_dst_prow, right_dst_vcol, &
         right_dst_vrow, right_max_nblks, right_max_nze, right_myfirstvcol, right_myfirstvrow, &
         right_mypcol, right_myprow, right_npcols, right_nprows, right_recv_icol, right_recv_irow
      INTEGER :: right_recv_p, right_recv_pcol, right_recv_prow, right_recv_vcol, right_recv_vrow, &
         right_row_image, right_row_mult, right_row_nimages, right_send_icol, right_send_irow, &
         right_send_p, right_send_pcol, right_send_prow, right_send_vcol, right_send_vrow, &
         right_src_icol, right_src_irow, right_src_p, right_src_pcol, right_src_prow, &
         right_src_vcol, right_src_vrow, row, size_guess, stat, threads_finished, &
         threads_finished_read, v_k, v_ki
      INTEGER(KIND=int_8)                                :: flop_single, flop_total, mem
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: row_counts, total_row_counts
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: left_sizes, my_sizes, right_sizes
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :)        :: all_sizes
      INTEGER, DIMENSION(3, 2)                           :: mp_rc_groups
      INTEGER, DIMENSION(:), POINTER :: left_data_rr, left_data_sr, left_index_rp, left_index_rr, &
         left_index_sp, left_index_sr, local_rows, right_data_rr, right_data_sr, right_index_rp, &
         right_index_rr, right_index_sp, right_index_sr
      INTEGER, DIMENSION(:, :), POINTER                  :: left_pgrid, product_pgrid, right_pgrid
      INTEGER, SAVE                                      :: mult_id = 0
      LOGICAL                                            :: keep_sparsity, list_indexing, &
                                                            otf_filtering

!$    REAL(KIND=real_8)                        :: left_fill, right_fill
      REAL(kind=sp), ALLOCATABLE, DIMENSION(:) :: left_norms, right_norms, &
                                                  row_max_epss
      TYPE(dbcsr_2d_array_type), POINTER :: left_buffer_2, left_buffer_calc, &
                                            left_buffer_comm, right_buffer_2, right_buffer_calc, right_buffer_comm
      TYPE(dbcsr_data_obj)                     :: left_data_rp, left_data_sp, &
                                                  right_data_rp, right_data_sp
      TYPE(dbcsr_data_obj), POINTER            :: trs_stackbuf_calc, &
                                                  trs_stackbuf_comm
      TYPE(dbcsr_data_obj), TARGET             :: trs_stackbuf_1, trs_stackbuf_2
      TYPE(dbcsr_mm_multrec_type_p), DIMENSION(:), ALLOCATABLE :: multrec
      TYPE(dbcsr_mp_obj)                       :: left_mp_obj, product_mp_obj, &
                                                  right_mp_obj

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (trs_stackbuf_calc, trs_stackbuf_comm)
      !
      ALLOCATE (left_buffer_2, right_buffer_2)
      mult_id = mult_id+1

      IF (PRESENT(retain_sparsity)) THEN
         keep_sparsity = retain_sparsity
      ELSE
         keep_sparsity = .FALSE.
      ENDIF
      otf_filtering = PRESENT(filter_eps)

!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (multrec, nthreads, product_matrix)
!$OMP MASTER
      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()
      IF (.NOT. ASSOCIATED(product_matrix%m%wms)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3083,"Work matrices do not exist")
      IF (SIZE(product_matrix%m%wms) .NE. nthreads) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3085,"Work matrices not correctly sized.")
      ALLOCATE (multrec(0:nthreads-1))
!$OMP END MASTER
!$OMP END PARALLEL

      output_unit = default_output_unit
      flop_total = 0
      ! Set up variables
      data_type = dbcsr_get_data_type(product_matrix)
      data_type_byte = dbcsr_datatype_sizeof(data_type)
      left_row_nimages = left_set%image_dist%i%row_decimation
      left_row_mult = left_set%image_dist%i%row_multiplicity
      left_col_nimages = left_set%image_dist%i%col_decimation
      left_col_mult = left_set%image_dist%i%col_multiplicity
      right_row_nimages = right_set%image_dist%i%row_decimation
      right_row_mult = right_set%image_dist%i%row_multiplicity
      right_col_nimages = right_set%image_dist%i%col_decimation
      right_col_mult = right_set%image_dist%i%col_multiplicity
      left_mp_obj = dbcsr_distribution_mp(left_set%image_dist%i%main)
      right_mp_obj = dbcsr_distribution_mp(right_set%image_dist%i%main)
      product_mp_obj = dbcsr_distribution_mp(product_matrix%m%dist)
      numnodes = dbcsr_mp_numnodes(product_mp_obj)
      mynode = dbcsr_mp_mynode(product_mp_obj)
      left_nprows = dbcsr_mp_nprows(left_mp_obj)
      left_npcols = dbcsr_mp_npcols(left_mp_obj)
      left_myprow = dbcsr_mp_myprow(left_mp_obj)
      left_mypcol = dbcsr_mp_mypcol(left_mp_obj)
      left_myfirstvrow = dbcsr_mp_myprow(left_mp_obj)*left_row_nimages
      left_myfirstvcol = dbcsr_mp_mypcol(left_mp_obj)*left_col_nimages
      right_nprows = dbcsr_mp_nprows(right_mp_obj)
      right_npcols = dbcsr_mp_npcols(right_mp_obj)
      right_myprow = dbcsr_mp_myprow(right_mp_obj)
      right_mypcol = dbcsr_mp_mypcol(right_mp_obj)
      right_myfirstvrow = dbcsr_mp_myprow(right_mp_obj)*right_row_nimages
      right_myfirstvcol = dbcsr_mp_mypcol(right_mp_obj)*right_col_nimages
      mp_group = dbcsr_mp_group(product_mp_obj)
      left_pgrid => dbcsr_mp_pgrid(left_mp_obj)
      right_pgrid => dbcsr_mp_pgrid(right_mp_obj)
      product_pgrid => dbcsr_mp_pgrid(product_mp_obj)
      CALL dbcsr_mp_grid_setup(product_mp_obj)
      CALL dbcsr_mp_grid_setup(left_mp_obj)
      CALL dbcsr_mp_grid_setup(right_mp_obj)
      IF (dbcsr_mp_has_subgroups(product_mp_obj)) THEN
         mp_rc_groups(M_P, 1:2) = (/dbcsr_mp_my_row_group(product_mp_obj), &
                                    dbcsr_mp_my_col_group(product_mp_obj)/)
      ENDIF
      IF (dbcsr_mp_has_subgroups(left_mp_obj)) THEN
         mp_rc_groups(M_L, 1:2) = (/dbcsr_mp_my_row_group(left_mp_obj), &
                                    dbcsr_mp_my_col_group(left_mp_obj)/)
      ENDIF
      IF (dbcsr_mp_has_subgroups(right_mp_obj)) THEN
         mp_rc_groups(M_R, 1:2) = (/dbcsr_mp_my_row_group(right_mp_obj), &
                                    dbcsr_mp_my_col_group(right_mp_obj)/)
      ENDIF
      !
      ! Dummy checks
      ! left/right matching
      IF (left_col_nimages .NE. right_row_mult) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3143,"Left/Right image mismatch")
      IF (left_col_mult .NE. right_row_nimages) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3145,"Left/Right image mismatch")
      IF (left_col_nimages*left_npcols .NE. right_row_nimages*right_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3147,"Left/Right total mismatch")
      ! product/left matching
      IF (left_row_mult*dbcsr_mp_nprows(product_mp_obj) .NE. left_row_nimages*left_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3150,"Product/Left total mismatch")
      ! product/left matching
      IF (right_col_mult*dbcsr_mp_npcols(product_mp_obj) .NE. right_col_nimages*right_npcols) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3153,"Product/Right total mismatch")
      ! Limitations
      IF (left_row_nimages .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3156,"Product/Left matrix process grid mismatch")
      IF (left_row_mult .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3158,"Product/Left matrix process grid mismatch")
      IF (right_col_nimages .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3160,"Product/Right matrix process grid mismatch")
      IF (right_col_mult .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3162,"Product/Right matrix process grid mismatch")

      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, left_row_nimages*left_col_nimages)
      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, right_row_nimages*right_col_nimages)
      !
      ! Exchange size data
      ALLOCATE (my_sizes(4, MAX(left_row_nimages, right_row_nimages), &
                         MAX(left_col_nimages, right_col_nimages)))
      my_sizes(:, :, :) = 0
      DO left_row_image = 1, left_row_nimages
         DO left_col_image = 1, left_col_nimages
            my_sizes(idata+ileft, left_row_image, left_col_image) &
               = dbcsr_data_get_size_referenced( &
                 left_set%mats(left_row_image, left_col_image)%m%data_area)
            my_sizes(imeta+ileft, left_row_image, left_col_image) = &
               left_set%mats(left_row_image, left_col_image)%m%index &
               (dbcsr_slot_size)
         ENDDO
      ENDDO

      DO right_row_image = 1, right_row_nimages
         DO right_col_image = 1, right_col_nimages
            my_sizes(idata+iright, right_row_image, right_col_image) &
               = dbcsr_data_get_size_referenced( &
                 right_set%mats(right_row_image, right_col_image)%m%data_area)
            my_sizes(imeta+iright, right_row_image, right_col_image) = &
               right_set%mats(right_row_image, right_col_image)%m%index &
               (dbcsr_slot_size)
         ENDDO
      ENDDO

      ALLOCATE (all_sizes(4, LBOUND(my_sizes, 2):UBOUND(my_sizes, 2), &
                          LBOUND(my_sizes, 3):UBOUND(my_sizes, 3), 0:numnodes-1))
      CALL mp_allgather(my_sizes, all_sizes, mp_group)
      !
      ! Count the maximum possible multiplies per row for on-the-fly
      ! filtering.
      per_row_eps: IF (.NOT. otf_filtering) THEN
         ! These arrays must be valid when passed to called subroutines.
         ALLOCATE (left_norms(0), right_norms(0), row_max_epss(0), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3203,"Could not allocate memory")
      ELSE
         IF (careful_mod) THEN
            IF (left_set%mats(1, 1)%m%bcsc) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3207,"Can not do on-the-fly filtering with CSC-indexed matrices.")
         ENDIF
         IF (dbcsr_has_local_row_index(left_set%mats(1, 1))) THEN
            nblkrows_used = dbcsr_nblkrows_local(left_set%mats(1, 1))
         ELSE
            nblkrows_used = dbcsr_nblkrows_total(left_set%mats(1, 1))
         ENDIF
         nblkrows_total = dbcsr_nblkrows_total(left_set%mats(1, 1))
         ALLOCATE (row_max_epss(nblkrows_total), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3217,"Could not allocate memory for left epsilons")
         ALLOCATE (row_counts(nblkrows_used), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3220,"Could not allocate memory for left row counts")
         ! The summation could be done prow-locally but it would
         ! complicate the pre-row eps calculation.
         ALLOCATE (total_row_counts(nblkrows_total), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3225,"Could not allocate memory for left row counts")
         ! Each prow member matrix (npcols * row_images) counts the
         ! blocks present in each of its rows.
         total_row_counts(:) = 0
         DO left_row_image = 1, left_row_nimages
            DO left_col_image = 1, left_col_nimages
               list_indexing = &
                  left_set%mats(left_row_image, left_col_image)%m%list_indexing
               IF (careful_mod) THEN
                  IF (list_indexing) THEN
                     IF ((left_set%mats(left_row_image, left_col_image)%m%nblks)*3 .NE. &
                         SIZE(left_set%mats(left_row_image, left_col_image)%m%coo_l)) &
                        CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3237,"Row count mismatch")
                  ELSE
                     IF (nblkrows_used+1 .NE. SIZE(left_set%mats(left_row_image, left_col_image)%m%row_p)) &
                        CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3240,"Row count mismatch")
                  ENDIF
               ENDIF
               IF (list_indexing) THEN
                  CALL count_bins( &
                     left_set%mats(left_row_image, left_col_image)%m%nblks, &
                     left_set%mats(left_row_image, left_col_image)%m%coo_l(1::3), &
                     nblkrows_used, row_counts)
               ELSE
                  CALL dbcsr_count_row_index( &
                     left_set%mats(left_row_image, left_col_image)%m%row_p, &
                     row_counts, nblkrows_used)
               ENDIF
               IF (dbcsr_has_local_row_index(left_set%mats(left_row_image, left_col_image))) THEN
                  local_rows => array_data(left_set%mats(left_row_image, left_col_image)%m%local_rows)
                  IF (SIZE(local_rows) .NE. SIZE(row_counts)) &
                     CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3256,"Mismatch in number of local rows.")
                  total_row_counts(local_rows) = total_row_counts(local_rows) &
                                                 +row_counts(1:nblkrows_used)
               ELSE
                  total_row_counts(:) = total_row_counts(:) &
                                        +row_counts(:)
               ENDIF
            ENDDO
         ENDDO
         ! The counted blocks are then summed up
         CALL mp_sum(total_row_counts, mp_group)
         ! and used to determine the maximum per-block epsilon.
         DO row = 1, nblkrows_total
            row_max_epss(row) &
               = REAL(filter_eps &
                      /REAL(MAX(1, total_row_counts(row)), KIND=KIND(row_max_epss)), &
                      KIND=KIND(row_max_epss))
         END DO
         !
         DEALLOCATE (row_counts)
         DEALLOCATE (total_row_counts)
      ENDIF per_row_eps
      !
      ! The main transfer loop goes through the virtual rows/columns.
      ! The number of steps may be smaller if the grid dimension is very
      ! non-optimal (both left column images and right row images are >
      ! 1).
      min_nimages = MIN(left_col_nimages, right_row_nimages)
      nvirt_k = left_npcols*left_col_nimages
      nsteps_k = nvirt_k/min_nimages
      !
      ! Translate the all_sizes to account for pre-distribution.  This
      ! is just done to simplify lookups.
      ALLOCATE (left_sizes(2, 0:left_nprows*left_row_nimages-1, 0:nvirt_k-1))
      left_sizes = -1
      DO left_src_vcol = 0, left_col_nimages*left_npcols-1
         DO left_src_vrow = 0, left_row_nimages*left_nprows-1
            ! Calculate what was shifted.  The left_src_v{row,col} are
            ! the "source" rows/columns; the left_dst are the shifted
            ! targets where the data was placed in make_images.
            CALL image_calculator(left_set%image_dist, &
                                  prow=left_dst_prow, pcol=left_dst_pcol, &
                                  rowi=left_dst_irow, coli=left_dst_icol, &
                                  myvprow=left_src_vrow, myvpcol=left_src_vcol, &
                                  shifting='l')
            left_dst_p = left_pgrid(left_dst_prow, left_dst_pcol)
            left_sizes(idata, left_src_vrow, left_src_vcol) = &
               all_sizes( &
               idata+ileft, left_dst_irow, left_dst_icol, left_dst_p)
            left_sizes(imeta, left_src_vrow, left_src_vcol) = &
               all_sizes( &
               imeta+ileft, left_dst_irow, left_dst_icol, left_dst_p)
         ENDDO
      ENDDO
      !
      ALLOCATE (right_sizes(2, 0:nvirt_k-1, 0:right_npcols*right_col_nimages-1))
      right_sizes = -1
      DO right_src_vcol = 0, right_col_nimages*right_npcols-1
         DO right_src_vrow = 0, right_row_nimages*right_nprows-1
            ! Calculate what was shifted.  The right_src_v{row,col} are
            ! the "source" rows/columns; the right_dst are the shifted
            ! targets where the data was placed in make_images.
            CALL image_calculator(right_set%image_dist, &
                                  prow=right_dst_prow, pcol=right_dst_pcol, &
                                  rowi=right_dst_irow, coli=right_dst_icol, &
                                  myvprow=right_src_vrow, myvpcol=right_src_vcol, &
                                  shifting='r')
            right_dst_p = right_pgrid(right_dst_prow, right_dst_pcol)
            right_sizes(idata, right_src_vrow, right_src_vcol) = &
               all_sizes( &
               idata+iright, right_dst_irow, right_dst_icol, right_dst_p)
            right_sizes(imeta, right_src_vrow, right_src_vcol) = &
               all_sizes( &
               imeta+iright, right_dst_irow, right_dst_icol, right_dst_p)
         ENDDO
      ENDDO
      !
      ! Setup product work areas
      left_max_nze = MAXVAL(all_sizes(idata+ileft, :, :, :))
      left_max_nblks = MAXVAL(all_sizes(imeta+ileft, :, :, :))
      right_max_nze = MAXVAL(all_sizes(idata+iright, :, :, :))
      right_max_nblks = MAXVAL(all_sizes(imeta+iright, :, :, :))
      !!
      ithread = 0
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP          PRIVATE (i, size_guess, &
!$OMP                   left_fill, right_fill, ithread) &
!$OMP          SHARED (product_matrix, left_max_nze, right_max_nze) &
!$OMP          SHARED (left_set, right_set, &
!$OMP                 left_col_nimages, right_row_nimages) &
!$OMP          SHARED (nthreads, keep_sparsity, mynode)
      !
!$    ithread = OMP_GET_THREAD_NUM()
      ! The work arrays have to be setup (actually, not quite sure).
      i = ithread+1
      size_guess = product_matrix%m%wms(i)%datasize ! Should be minimal
      CALL dbcsr_data_ensure_size(product_matrix%m%wms(i)%data_area, &
                                  size_guess)
      CALL dbcsr_data_set_size_referenced(product_matrix%m%wms(i)%data_area, &
                                          product_matrix%m%wms(i)%datasize)
      ! XXXXXXX a quick fix right now, allocation with size 1 might actually not be needed at all,
      !         but something expects this to be associated
      CALL ensure_array_size(product_matrix%m%wms(i)%row_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(i)%col_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(i)%blk_p, ub=1)
!$OMP END PARALLEL

      ! update capacity of memory-pools
      IF (has_acc) THEN
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_1%pool, &
                                            capacity=left_row_mult*left_col_nimages+right_row_nimages*right_col_mult)
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_2%pool, &
                                            capacity=left_row_mult*left_col_nimages+right_row_nimages*right_col_mult)
      ENDIF

      !
      ! Setup the left buffer matrices
      !
      CALL buffer_matrices_ensure_size(left_set, index_size=left_max_nblks, &
                                       data_size=left_max_nze)

      CALL setup_buffer_matrices(left_buffer_2, left_row_mult, left_col_nimages, &
                                 left_set%mats(1, 1), index_size=left_max_nblks, &
                                 data_size=left_max_nze)
      IF (otf_filtering) THEN
         ALLOCATE (left_norms(left_max_nblks), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3383,"Could not allocate memory for left norms")
         IF (stat .NE. 0) otf_filtering = .FALSE.
      ENDIF
      left_buffer_calc => left_set
      left_buffer_comm => left_buffer_2
      ALLOCATE (left_data_sr(left_col_nimages))
      ALLOCATE (left_index_sr(left_col_nimages))
      ALLOCATE (left_data_rr(left_col_nimages))
      ALLOCATE (left_index_rr(left_col_nimages))
      left_data_sr = mp_request_null
      left_data_rr = mp_request_null
      left_index_sr = mp_request_null
      left_index_rr = mp_request_null

      ! Setup buffers for right matrix
      CALL buffer_matrices_ensure_size(right_set, index_size=right_max_nblks, &
                                       data_size=right_max_nze)

      CALL setup_buffer_matrices(right_buffer_2, right_row_nimages, right_col_mult, &
                                 right_set%mats(1, 1), index_size=right_max_nblks, data_size=right_max_nze)
      IF (otf_filtering) THEN
         ALLOCATE (right_norms(right_max_nblks), stat=stat)
         IF (stat .NE. 0) &
            CALL cp__w("dbcsr/mm/dbcsr_mm_cannon.F",3406,"Could not allocate memory for right norms")
         IF (stat .NE. 0) otf_filtering = .FALSE.
      ENDIF
      right_buffer_calc => right_set
      right_buffer_comm => right_buffer_2
      ALLOCATE (right_data_sr(right_row_nimages))
      ALLOCATE (right_index_sr(right_row_nimages))
      ALLOCATE (right_data_rr(right_row_nimages))
      ALLOCATE (right_index_rr(right_row_nimages))
      right_data_sr = mp_request_null
      right_data_rr = mp_request_null
      right_index_sr = mp_request_null
      right_index_rr = mp_request_null
      !
!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED (left_buffer_comm, right_buffer_comm, product_matrix,&
!$OMP         keep_sparsity, filter_eps, row_max_epss, multrec, nthreads, &
!$OMP         right_data_sr, right_data_rr, left_data_sr, left_data_rr,&
!$OMP         right_index_sr, right_index_rr, left_index_sr, left_index_rr), &
!$OMP PRIVATE(ithread)
      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      ALLOCATE (multrec(ithread)%p)
      CALL dbcsr_mm_multrec_init(multrec(ithread)%p, &
                                 left_buffer_comm%mats(1, 1)%m, &
                                 right_buffer_comm%mats(1, 1)%m, &
                                 product_matrix%m, &
                                 keep_sparsity=keep_sparsity, &
                                 eps=filter_eps, &
                                 row_max_epss=row_max_epss, &
                                 block_estimate=MAX(product_matrix%m%nblks, &
                                                    left_buffer_comm%mats(1, 1)%m%nblks, &
                                                    right_buffer_comm%mats(1, 1)%m%nblks)/nthreads, &
                                 right_row_blk_size=array_data(right_buffer_comm%mats(1, 1)%m%row_blk_size))
!$OMP END PARALLEL
      !
      ! Setup indexing
      CALL setup_rec_index(left_set, left_row_nimages, left_col_nimages)
      CALL setup_rec_index(right_set, right_row_nimages, right_col_nimages)
      !
      ! Setup the send/receive data pointers
      CALL dbcsr_data_init(left_data_sp)
      CALL dbcsr_data_init(left_data_rp)
      CALL dbcsr_data_init(right_data_sp)
      CALL dbcsr_data_init(right_data_rp)
      CALL dbcsr_data_new(left_data_sp, data_type)
      CALL dbcsr_data_new(left_data_rp, data_type)
      CALL dbcsr_data_new(right_data_sp, data_type)
      CALL dbcsr_data_new(right_data_rp, data_type)

      ! Setup transpose stackbuffers
      IF (has_acc) THEN
         CALL dbcsr_data_init(trs_stackbuf_1)
         CALL dbcsr_data_init(trs_stackbuf_2)
         CALL dbcsr_data_new(trs_stackbuf_1, data_type=dbcsr_type_int_4, data_size=1000, memory_type=memtype_trsbuffer_1)
         CALL dbcsr_data_new(trs_stackbuf_2, data_type=dbcsr_type_int_4, data_size=1000, memory_type=memtype_trsbuffer_2)
         trs_stackbuf_calc => trs_stackbuf_1
         trs_stackbuf_comm => trs_stackbuf_2
      ENDIF

      !
      ! Here is the main loop.
      !
      ! In the first loop iteration, the data is fetched from the
      ! sources. In the remaining iterations, the data are exchanged
      ! among neighbors.  In the last loop only calculations take place.

      grouped_k_index: DO metronome = 1, nsteps_k
         IF (debug_mod) WRITE (*, '(1X,A,3(1X,A,1X,I5))') routineN, &
            "step", metronome, &
            "first k", metronome*min_nimages, &
            "last k", (metronome+1)*min_nimages-1
         ! Wait for right matrix transfer completion. Wait in all but
         ! the first loop iteration.
         CALL timeset(routineN//"_metrocomm1", handle2)
         wait_right: IF (metronome .GT. 1) THEN
            IF (debug_mod) WRITE (*, '(1X,A)') routineN//" waiting for right"
            !
            CALL mp_waitall(right_data_sr)
            CALL mp_waitall(right_data_rr)
            CALL mp_waitall(right_index_sr)
            CALL mp_waitall(right_index_rr)
         ENDIF wait_right
         CALL timestop(handle2)
         ! Right matrix transfer. Transfer in all but the last loop
         ! iteration.
         xfer_right: IF (metronome .LT. nsteps_k) THEN
            DO v_ki = 0, right_row_nimages-1
               v_k = metronome*min_nimages+v_ki
               ! Calculate the process to send to.  It's the virtual
               ! process row -min_nimages up (i.e., smaller row number)
               ! from me.
               CALL image_calculator(right_set%image_dist, &
                                     prow=right_send_prow, rowi=right_send_irow, & ! output
                                     pcol=right_send_pcol, coli=right_send_icol, & ! output
                                     vprow=right_send_vrow, vpcol=right_send_vcol, & ! output
                                     ! myvprow goes through all of my (process row) images
                                     myvprow=v_ki+right_myfirstvrow, &
                                     myvpcol=right_myfirstvcol, & ! nothing happens in the columns
                                     ! send to process min_nimages up in the grid
                                     vprow_shift=-min_nimages, &
                                     shifting='0')
               ! Calculate which data I send.
               CALL image_calculator(right_set%image_dist, &
                                     prow=right_dst_prow, rowi=right_dst_irow, &
                                     pcol=right_dst_pcol, coli=right_dst_icol, &
                                     vprow=right_dst_vrow, vpcol=right_dst_vcol, &
                                     ! myvprows goes through all of my (process row) images
                                     myvprow=v_ki+right_myfirstvrow, &
                                     myvpcol=right_myfirstvcol, & ! nothing happens in the columns
                                     ! send what I got from min_nimages down, appropriate
                                     ! to the metronome tick
                                     vprow_shift=-min_nimages+metronome*min_nimages, &
                                     ! This is with relative shifting.
                                     shifting='R')
               right_dst_p = right_pgrid(right_dst_prow, right_dst_pcol)
               CALL dbcsr_data_set_pointer( &
                  area=right_data_sp, &
                  rsize=right_sizes(idata, right_dst_vrow, right_dst_vcol), &
                  csize=1, &
                  pointee=right_buffer_calc%mats(v_ki+1, 1)%m%data_area)
               right_index_sp => right_buffer_calc%mats( &
                                 v_ki+1, 1 &
                                 )%m%index(1: &
                                           right_sizes(imeta, right_dst_vrow, right_dst_vcol))
               !
               ! Calculate the process to receive from
               CALL image_calculator(right_set%image_dist, &
                                     prow=right_recv_prow, rowi=right_recv_irow, &
                                     pcol=right_recv_pcol, coli=right_recv_icol, &
                                     vprow=right_recv_vrow, vpcol=right_recv_vcol, &
                                     myvprow=v_ki+right_myfirstvrow, &
                                     myvpcol=right_myfirstvcol, &
                                     vprow_shift=+min_nimages, & ! just the opposite as "send to"
                                     shifting='0')
               ! Calculate which data I receive
               CALL image_calculator(right_set%image_dist, &
                                     prow=right_src_prow, rowi=right_src_irow, &
                                     pcol=right_src_pcol, coli=right_src_icol, &
                                     vprow=right_src_vrow, vpcol=right_src_vcol, &
                                     myvprow=v_ki+right_myfirstvrow, &
                                     myvpcol=right_myfirstvcol, &
                                     ! receive window moves with the metronome
                                     vprow_shift=metronome*min_nimages, &
                                     shifting='R')
               !
               IF (has_acc) THEN
                  CALL timeset(routineN//"_acc_sync_right", handle3)
                  CALL acc_event_synchronize(right_buffer_comm%mats(v_ki+1, 1)%m%data_area%d%acc_ready)
                  CALL timestop(handle3)
               ENDIF

               right_src_p = right_pgrid(right_src_prow, right_src_pcol)
               CALL dbcsr_data_set_pointer( &
                  area=right_data_rp, &
                  rsize=right_sizes(idata, right_src_vrow, right_src_vcol), &
                  csize=1, &
                  pointee=right_buffer_comm%mats(v_ki+1, 1)%m%data_area)
               right_index_rp => right_buffer_comm%mats( &
                                 v_ki+1, 1 &
                                 )%m%index(1: &
                                           right_sizes(imeta, right_src_vrow, right_src_vcol))
               !
               right_send_p = right_pgrid(right_send_prow, right_send_pcol)
               right_recv_p = right_pgrid(right_recv_prow, right_recv_pcol)
               ! These are column-communicator relative
               IF (dbcsr_mp_has_subgroups(right_mp_obj)) THEN
                  right_send_p = right_send_prow
                  right_recv_p = right_recv_prow
                  grp = dbcsr_mp_my_col_group(right_mp_obj)
               ELSE
                  grp = dbcsr_mp_group(right_mp_obj)
               ENDIF
               !
               CALL timeset(routineN//"_metrocomm2", handle2)
               CALL dbcsr_irecv_any(right_data_rp, right_recv_p, &
                                    grp, right_data_rr(v_ki+1), tag=right_src_vrow)
               CALL mp_irecv(right_index_rp, right_recv_p, &
                             grp, right_index_rr(v_ki+1), tag=right_src_vrow)
               CALL dbcsr_isend_any(right_data_sp, right_send_p, &
                                    grp, right_data_sr(v_ki+1), tag=right_dst_vrow)
               CALL mp_isend(right_index_sp, right_send_p, &
                             grp, right_index_sr(v_ki+1), tag=right_dst_vrow)
               dbcsr_mpi_statistics%nexchanged = dbcsr_mpi_statistics%nexchanged+1
               CALL count_mpi_statistics(dbcsr_mpi_statistics%data_size(1, :), &
                                         dbcsr_data_get_size(right_data_rp), &
                                         dbcsr_mpi_statistics%data_size_breakdown(:, :, 1), data_type_byte)
               CALL timestop(handle2)
            ENDDO
         ENDIF xfer_right
         !
         ! Repoint indices of right matrices
         calc_case_right: IF (metronome .GT. 1) THEN
            DO v_ki = 0, right_row_nimages-1
               CALL dbcsr_repoint_index(right_buffer_calc%mats(v_ki+1, 1)%m)
               right_buffer_calc%mats(v_ki+1, 1)%m%valid = .TRUE.
            ENDDO
         ENDIF calc_case_right
         !
         ! Wait for left matrix transfer completion. Wait in all but
         ! the first loop iteration.
         CALL timeset(routineN//"_metrocomm3", handle2)
         wait_left: IF (metronome .GT. 1) THEN
            IF (debug_mod) WRITE (*, '(1X,A)') routineN//" waiting for left"
            CALL mp_waitall(left_data_sr)
            CALL mp_waitall(left_data_rr)
            CALL mp_waitall(left_index_sr)
            CALL mp_waitall(left_index_rr)
         ENDIF wait_left
         CALL timestop(handle2)
         ! Left matrix transfer. Transfer in all but the last loop
         ! iteration.
         xfer_left: IF (metronome .LT. nsteps_k) THEN
            DO v_ki = 0, left_col_nimages-1
               v_k = metronome*min_nimages+v_ki
               ! Calculate the process to send to.
               CALL image_calculator(left_set%image_dist, &
                                     prow=left_send_prow, rowi=left_send_irow, & ! output
                                     pcol=left_send_pcol, coli=left_send_icol, & ! output
                                     vprow=left_send_vrow, vpcol=left_send_vcol, & ! output
                                     myvprow=left_myfirstvrow, & ! nothing happens in the rows
                                     ! go through all my column images
                                     myvpcol=v_ki+left_myfirstvcol, &
                                     ! send to process min_nimages left in the grid
                                     vpcol_shift=-min_nimages, &
                                     shifting='0')
               ! Calculate which data I send.
               CALL image_calculator(left_set%image_dist, &
                                     prow=left_dst_prow, rowi=left_dst_irow, &
                                     pcol=left_dst_pcol, coli=left_dst_icol, &
                                     vprow=left_dst_vrow, vpcol=left_dst_vcol, &
                                     myvprow=left_myfirstvrow, &
                                     ! go through all my column images
                                     myvpcol=v_ki+left_myfirstvcol, &
                                     ! send what I got from min_nimages left, appropriate
                                     ! to the metronome tick
                                     vpcol_shift=-min_nimages+metronome*min_nimages, &
                                     ! This is with relative shifting.
                                     shifting='L')
               !
               left_dst_p = left_pgrid(left_dst_prow, left_dst_pcol)
               CALL dbcsr_data_set_pointer( &
                  area=left_data_sp, &
                  rsize=left_sizes(idata, left_dst_vrow, left_dst_vcol), &
                  csize=1, &
                  pointee=left_buffer_calc%mats(1, v_ki+1)%m%data_area)
               left_index_sp => left_buffer_calc%mats( &
                                1, v_ki+1 &
                                )%m%index(1: &
                                          left_sizes(imeta, left_dst_vrow, left_dst_vcol))
               !
               ! Calculate the process to receive from
               CALL image_calculator(left_set%image_dist, &
                                     prow=left_recv_prow, rowi=left_recv_irow, &
                                     pcol=left_recv_pcol, coli=left_recv_icol, &
                                     vprow=left_recv_vrow, vpcol=left_recv_vcol, &
                                     myvprow=left_myfirstvrow, &
                                     myvpcol=v_ki+left_myfirstvcol, &
                                     vpcol_shift=+min_nimages, & ! just the opposite as "send to"
                                     shifting='0')
               ! Calculate which data I receive
               CALL image_calculator(left_set%image_dist, &
                                     prow=left_src_prow, rowi=left_src_irow, &
                                     pcol=left_src_pcol, coli=left_src_icol, &
                                     vprow=left_src_vrow, vpcol=left_src_vcol, &
                                     myvprow=left_myfirstvrow, &
                                     myvpcol=v_ki+left_myfirstvcol, &
                                     ! receive window moves with the metronome
                                     vpcol_shift=metronome*min_nimages, &
                                     shifting='L')
               !
               IF (has_acc) THEN
                  CALL timeset(routineN//"_acc_sync_left", handle3)
                  CALL acc_event_synchronize(left_buffer_comm%mats(1, v_ki+1)%m%data_area%d%acc_ready)
                  CALL timestop(handle3)
               ENDIF

               left_src_p = left_pgrid(left_src_prow, left_src_pcol)
               CALL dbcsr_data_set_pointer( &
                  area=left_data_rp, &
                  rsize=left_sizes(idata, left_src_vrow, left_src_vcol), &
                  csize=1, &
                  pointee=left_buffer_comm%mats(1, v_ki+1)%m%data_area)
               left_index_rp => left_buffer_comm%mats( &
                                1, v_ki+1 &
                                )%m%index(1: &
                                          left_sizes(imeta, left_src_vrow, left_src_vcol))
               !
               left_send_p = left_pgrid(left_send_prow, left_send_pcol)
               left_recv_p = left_pgrid(left_recv_prow, left_recv_pcol)
               ! These are column-communicator relative
               IF (dbcsr_mp_has_subgroups(left_mp_obj)) THEN
                  left_send_p = left_send_pcol
                  left_recv_p = left_recv_pcol
                  grp = dbcsr_mp_my_row_group(left_mp_obj)
               ELSE
                  grp = dbcsr_mp_group(left_mp_obj)
               ENDIF
               !
               CALL timeset(routineN//"_metrocomm4", handle2)
               CALL dbcsr_irecv_any(left_data_rp, left_recv_p, &
                                    grp, left_data_rr(v_ki+1), tag=left_src_vcol)
               CALL mp_irecv(left_index_rp, left_recv_p, &
                             grp, left_index_rr(v_ki+1), tag=left_src_vcol)
               CALL dbcsr_isend_any(left_data_sp, left_send_p, &
                                    grp, left_data_sr(v_ki+1), tag=left_dst_vcol)
               CALL mp_isend(left_index_sp, left_send_p, &
                             grp, left_index_sr(v_ki+1), tag=left_dst_vcol)
               dbcsr_mpi_statistics%nexchanged = dbcsr_mpi_statistics%nexchanged+1
               CALL count_mpi_statistics(dbcsr_mpi_statistics%data_size(2, :), &
                                         dbcsr_data_get_size(left_data_rp), &
                                         dbcsr_mpi_statistics%data_size_breakdown(:, :, 2), data_type_byte)
               CALL timestop(handle2)
            ENDDO
         ENDIF xfer_left
         !
         ! Repoint indices of left matrices and do the multiplications.
         calc_case_left: IF (metronome .GT. 1) THEN
            DO v_ki = 0, left_col_nimages-1
               CALL dbcsr_repoint_index(left_buffer_calc%mats(1, v_ki+1)%m)
               left_buffer_calc%mats(1, v_ki+1)%m%valid = .TRUE.
            ENDDO
         ENDIF calc_case_left

         CALL timeset(routineN//"_multrec", handle2)
         DO v_ki = 0, min_nimages-1
            IF (debug_mod) THEN
               CALL dbcsr_print(left_buffer_calc%mats(1, v_ki+1), nodata=.TRUE.)
               CALL dbcsr_print(right_buffer_calc%mats(v_ki+1, 1), nodata=.TRUE.)
            ENDIF
            !
            ! form here the code for dbcsr_mm_driver_inner_init was taken
            !
            IF (.FALSE.) WRITE (*, *) routineN//" TICK", v_ki
            IF (.TRUE. .OR. right_buffer_calc%mats(v_ki+1, 1)%m%local_indexing) THEN
               ! Since the right matrix is shifted vertically, the
               ! received data always has different notions of "local
               ! rows".  Thus the local_rows and global_rows must be
               ! recalculated.
               CALL dbcsr_reset_vlocals(right_buffer_calc%mats(v_ki+1, 1), &
                                        right_set%image_dist)
            ENDIF
            IF (.TRUE. .OR. left_buffer_calc%mats(1, v_ki+1)%m%local_indexing) THEN
               ! Since the right matrix is shifted vertically, the
               ! received data always has different notions of "local
               ! rows".  Thus the local_rows and global_rows must be
               ! recalculated.
               CALL dbcsr_reset_vlocals(left_buffer_calc%mats(1, v_ki+1), &
                                        left_set%image_dist)
            ENDIF

            IF (has_acc) THEN
               CALL dbcsr_data_host2dev(left_buffer_calc%mats(1, v_ki+1)%m%data_area)
               CALL dbcsr_data_host2dev(right_buffer_calc%mats(v_ki+1, 1)%m%data_area)
               CALL acc_transpose_blocks(right_buffer_calc%mats(v_ki+1, 1), trs_stackbuf_calc)
            END IF

            ! Sets the local right-matrix columns
            IF (otf_filtering) THEN
               left_norms(:) = huge_norm
               right_norms(:) = huge_norm
               CALL calculate_norms(right_buffer_calc%mats(v_ki+1, 1), &
                                    right_norms)
               CALL calculate_norms(left_buffer_calc%mats(1, v_ki+1), &
                                    left_norms)
            ENDIF
            !
            flop_single = 0
            threads_finished = 0

!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (left_buffer_calc, right_buffer_calc, &
!$OMP         v_ki, &
!$OMP         product_matrix, multrec,&
!$OMP         filter_eps, right_norms, left_norms, row_max_epss, &
!$OMP         keep_sparsity,threads_finished, &
!$OMP         right_data_sr, right_data_rr, right_index_sr, right_index_rr, &
!$OMP         left_data_sr, left_data_rr, left_index_sr, left_index_rr, &
!$OMP         use_comm_thread) &
!$OMP PRIVATE (ithread,nthreads,threads_finished_read) &
!$OMP FIRSTPRIVATE (metronome, nsteps_k, min_nimages) &
!$OMP REDUCTION (+: flop_single)
            ithread = 0; nthreads = 1
!$          ithread = omp_get_thread_num(); nthreads = omp_get_num_threads()

            IF (metronome == nsteps_k .AND. v_ki == min_nimages-1) &
               CALL dbcsr_mm_multrec_phaseout(multrec(ithread)%p)

            CALL dbcsr_mm_multrec_multiply(multrec(ithread)%p, &
                                           left=left_buffer_calc%mats(1, v_ki+1)%m, &
                                           right=right_buffer_calc%mats(v_ki+1, 1)%m, &
                                           flop=flop_single, &
                                           a_norms=left_norms, b_norms=right_norms)

            IF (metronome == nsteps_k .AND. v_ki == min_nimages-1) THEN
               CALL dbcsr_mm_multrec_finalize(multrec(ithread)%p)
               DEALLOCATE (multrec(ithread)%p)
            ENDIF

!$OMP ATOMIC
            threads_finished = threads_finished+1
            IF (use_comm_thread .AND. (ithread .EQ. 0)) THEN
               DO
! requires OMP 3.1 (e.g. gcc >=4.7), for correctness, otherwise we keep fingers crossed
                  threads_finished_read = threads_finished
                  IF (threads_finished_read .EQ. nthreads) EXIT
                  CALL mp_testany(right_data_sr)
                  CALL mp_testany(right_data_rr)
                  CALL mp_testany(left_data_sr)
                  CALL mp_testany(left_data_rr)
                  CALL mp_testany(right_index_sr)
                  CALL mp_testany(right_index_rr)
                  CALL mp_testany(left_index_sr)
                  CALL mp_testany(left_index_rr)
               END DO
            END IF
!$OMP END PARALLEL
            flop_total = flop_total+flop_single
         ENDDO

         CALL timestop(handle2)

         CALL dbcsr_switch(left_buffer_calc, left_buffer_comm)
         CALL dbcsr_switch(right_buffer_calc, right_buffer_comm)
         CALL dbcsr_switch(trs_stackbuf_calc, trs_stackbuf_comm)

      ENDDO grouped_k_index
      CALL m_memory(mem)
      max_memory = MAX(max_memory, REAL(mem))

      IF (has_acc) THEN
         CALL dbcsr_data_release(trs_stackbuf_1)
         CALL dbcsr_data_release(trs_stackbuf_2)
      END IF

      IF (ALLOCATED(right_norms)) THEN
         DEALLOCATE (right_norms)
      ENDIF
      IF (ALLOCATED(left_norms)) THEN
         DEALLOCATE (left_norms)
      ENDIF
      IF (ALLOCATED(row_max_epss)) THEN
         DEALLOCATE (row_max_epss)
      ENDIF
      !
      CALL dbcsr_destroy_array(right_buffer_2)
      CALL dbcsr_destroy_array(left_buffer_2)
      DEALLOCATE (my_sizes)
      !
      CALL dbcsr_data_clear_pointer(left_data_sp)
      CALL dbcsr_data_clear_pointer(left_data_rp)
      CALL dbcsr_data_clear_pointer(right_data_sp)
      CALL dbcsr_data_clear_pointer(right_data_rp)
      CALL dbcsr_data_release(left_data_sp)
      CALL dbcsr_data_release(left_data_rp)
      CALL dbcsr_data_release(right_data_sp)
      CALL dbcsr_data_release(right_data_rp)
      !
      DEALLOCATE (left_data_rr, left_data_sr, left_index_rr, left_index_sr, &
                  right_data_rr, right_data_sr, right_index_rr, right_index_sr)
      !
      !
      IF (debug_mod) THEN
         v_ki = 0
         DO i = 1, SIZE(product_matrix%m%blk_p)
            v_ki = MAX(v_ki, ABS(product_matrix%m%blk_p(i)))
         ENDDO
         WRITE (*, *) routineN//" Actual final size", &
            LOG(REAL(dbcsr_data_get_size(product_matrix%m%data_area)))/LOG(10.0), &
            LOG(REAL(v_ki))/LOG(10.0)
      ENDIF
      !
      flop = flop_total
      DEALLOCATE (left_buffer_2, right_buffer_2)
      !
      CALL timestop(handle)
   END SUBROUTINE cannon_multiply_low

! **************************************************************************************************
!> \brief Multiplies two DBCSR matrices (experimental MPI algorithm).
!>        This algorithm is experimental and it should be not used in
!>        production runs.
!>
!> \param imgdist_left ...
!> \param imgdist_right ...
!> \param matrix_left ...
!> \param matrix_right ...
!> \param[out] product_matrix      DBCSR product matrix
!> \param[in] retain_sparsity      (optional) retain the sparsity of the
!>                                 existing product matrix; default is no
!> \param filter_eps ...
!> \param[out] flop                (optional) effective flop
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE cannon_multiply_low_exp(imgdist_left, imgdist_right, &
                                      matrix_left, matrix_right, &
                                      product_matrix, &
                                      retain_sparsity, &
                                      filter_eps, flop)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist_left, imgdist_right
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_left, matrix_right
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: product_matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(kind=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT)                   :: flop

      CHARACTER(len=*), PARAMETER :: routineN = 'cannon_multiply_low_exp', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, data_type, data_type_byte, grp_left, grp_right, handle, handle2, handle3, &
         iget, index_row_max_epss, iproc, ithread, left_col_mult, left_col_nimages, &
         left_data_offset, left_data_size, left_max_data_size, left_max_meta_size, &
         left_meta_offset, left_meta_size, left_myfirstvcol, left_myfirstvrow, left_mypcol, &
         left_myprow, left_npcols, left_nprows, left_nrequests, left_row_mult, left_row_nimages, &
         left_v_i, left_vcol, left_vrow, meta_size_image, metronome, min_nimages, mynode, mypcol, &
         myprow, nblkrows_local, nsteps_k, nthreads, numnodes, nvirt_k, recv_vcol, recv_vrow
      INTEGER :: right_col_mult, right_col_nimages, right_data_offset, right_data_size, &
         right_max_data_size, right_max_meta_size, right_meta_offset, right_meta_size, &
         right_myfirstvcol, right_myfirstvrow, right_mypcol, right_myprow, right_npcols, &
         right_nprows, right_nrequests, right_row_mult, right_row_nimages, right_v_i, right_vcol, &
         right_vrow, row, size_guess, size_index_merged, size_index_unmerged, v_ci, v_ki, v_ri
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, ALLOCATABLE, DIMENSION(:) :: indices_common_unmerged, left_displ_vunmerged, &
         left_get_requests_images, left_max_meta_size_merged, right_displ_vunmerged, &
         right_get_requests_images, right_max_meta_size_merged
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: left_vunmerged, right_vunmerged
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: left_vmerged, right_vmerged
      INTEGER, DIMENSION(dbcsr_slot_nblkrows_total:&
         dbcsr_slot_nfullcols_local)                     :: left_global_indices, right_global_indices
      LOGICAL                                            :: keep_sparsity, otf_filtering
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: common_unmerged
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: nofiltered
      REAL(kind=sp)                                      :: filter_eps_sp, min_row_max_epss
      REAL(kind=sp), ALLOCATABLE, DIMENSION(:)           :: row_max_epss
      REAL(kind=sp), ALLOCATABLE, DIMENSION(:, :)        :: left_norms, right_norms
      TYPE(dbcsr_buffers), POINTER                       :: buffers, buffers_calc, buffers_comm
      TYPE(dbcsr_data_obj)                               :: data_get
      TYPE(dbcsr_mm_multrec_type_p), ALLOCATABLE, &
         DIMENSION(:)                                    :: multrec
      TYPE(dbcsr_mp_obj)                                 :: left_mp_obj, product_mp_obj, right_mp_obj

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (buffers_calc, buffers_comm, buffers)
      !
      IF (PRESENT(retain_sparsity)) THEN
         keep_sparsity = retain_sparsity
      ELSE
         keep_sparsity = .FALSE.
      ENDIF
      otf_filtering = PRESENT(filter_eps)
      !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (nthreads)
!$OMP MASTER
      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL
      !
      size_index_unmerged = dbcsr_slot_nblks
!$    size_index_unmerged = size_index_unmerged+nthreads+1
      size_index_merged = dbcsr_num_slots
!$    size_index_merged = size_index_merged+nthreads+1
      !
      ! Dummy checks
      IF (.NOT. ASSOCIATED(product_matrix%m%wms)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3978,"Work matrices do not exist")
      IF (SIZE(product_matrix%m%wms) .NE. nthreads) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",3980,"Work matrices not correctly sized.")
      IF (.NOT. buffers_orig%left%is_valid .OR. &
          .NOT. buffers_orig%right%is_valid .OR. &
          .NOT. ASSOCIATED(buffers_orig%left%meta) .OR. &
          .NOT. ASSOCIATED(buffers_orig%right%meta) .OR. &
          .NOT. ALLOCATED(left_refs_meta_size) .OR. &
          .NOT. ALLOCATED(left_refs_data_size) .OR. &
          .NOT. ALLOCATED(right_refs_meta_size) .OR. &
          .NOT. ALLOCATED(right_refs_data_size) .OR. &
          .NOT. ALLOCATED(left_local_refs_meta_size) .OR. &
          .NOT. ALLOCATED(left_local_refs_data_size) .OR. &
          .NOT. ALLOCATED(right_local_refs_meta_size) .OR. &
          .NOT. ALLOCATED(right_local_refs_data_size) .OR. &
          .NOT. ALLOCATED(left_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(right_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(left_local_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(right_local_refs_displ_unmerged) .OR. &
          (otf_filtering .AND. use_mpi_filtering .AND. &
           (.NOT. (ALLOCATED(left_max_norms) .AND. ALLOCATED(right_max_norms) .AND. &
                   ALLOCATED(left_local_max_norms) .AND. ALLOCATED(right_local_max_norms))))) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4000,"No buffers associated for the experimental algo!")
      !
      ! Set up variables
      flop = 0
      data_type = dbcsr_get_data_type(product_matrix)
      data_type_byte = dbcsr_datatype_sizeof(data_type)
      left_row_nimages = imgdist_left%i%row_decimation
      left_row_mult = imgdist_left%i%row_multiplicity
      left_col_nimages = imgdist_left%i%col_decimation
      left_col_mult = imgdist_left%i%col_multiplicity
      right_row_nimages = imgdist_right%i%row_decimation
      right_row_mult = imgdist_right%i%row_multiplicity
      right_col_nimages = imgdist_right%i%col_decimation
      right_col_mult = imgdist_right%i%col_multiplicity
      left_mp_obj = dbcsr_distribution_mp(imgdist_left%i%main)
      right_mp_obj = dbcsr_distribution_mp(imgdist_right%i%main)
      product_mp_obj = dbcsr_distribution_mp(product_matrix%m%dist)
      numnodes = dbcsr_mp_numnodes(product_mp_obj)
      mynode = dbcsr_mp_mynode(product_mp_obj)
      myprow = dbcsr_mp_myprow(product_mp_obj)
      mypcol = dbcsr_mp_mypcol(product_mp_obj)
      left_nprows = dbcsr_mp_nprows(left_mp_obj)
      left_npcols = dbcsr_mp_npcols(left_mp_obj)
      left_myprow = dbcsr_mp_myprow(left_mp_obj)
      left_mypcol = dbcsr_mp_mypcol(left_mp_obj)
      left_myfirstvrow = left_myprow*left_row_nimages
      left_myfirstvcol = left_mypcol*left_col_nimages
      right_nprows = dbcsr_mp_nprows(right_mp_obj)
      right_npcols = dbcsr_mp_npcols(right_mp_obj)
      right_myprow = dbcsr_mp_myprow(right_mp_obj)
      right_mypcol = dbcsr_mp_mypcol(right_mp_obj)
      right_myfirstvrow = right_myprow*right_row_nimages
      right_myfirstvcol = right_mypcol*right_col_nimages
      grp_right = dbcsr_mp_my_col_group(right_mp_obj)
      grp_left = dbcsr_mp_my_row_group(left_mp_obj)
      CALL dbcsr_mp_grid_setup(product_mp_obj)
      CALL dbcsr_mp_grid_setup(left_mp_obj)
      CALL dbcsr_mp_grid_setup(right_mp_obj)
      !
      ! Dummy checks
      ! subcommunicators
      IF (.NOT. dbcsr_mp_has_subgroups(right_mp_obj)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4042,"Experimental algorithm requires rows subcommunicators for right matrix!")
      IF (.NOT. dbcsr_mp_has_subgroups(left_mp_obj)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4044,"Experimental algorithm requires columns subcommunicators for left matrix!")
      ! Right col nimages
      IF (right_col_nimages .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4047,"Col nimages for right matrix is not 1!")
      ! Left row nimages
      IF (left_row_nimages .NE. 1) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4050,"Row nimages for right matrix is not 1!")
      ! left/right matching
      IF (left_col_nimages .NE. right_row_mult) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4053,"Left/Right image mismatch")
      IF (left_col_mult .NE. right_row_nimages) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4055,"Left/Right image mismatch")
      IF (left_col_nimages*left_npcols .NE. right_row_nimages*right_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4057,"Left/Right total mismatch")
      ! product/left matching
      IF (left_row_mult*dbcsr_mp_nprows(product_mp_obj) .NE. left_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4060,"Product/Left total mismatch")
      ! product/left matching
      IF (right_col_mult*dbcsr_mp_npcols(product_mp_obj) .NE. right_npcols) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4063,"Product/Right total mismatch")
      !
      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, left_col_nimages)
      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, right_row_nimages)
      !
      ! The main transfer loop goes through the virtual rows/columns.
      ! The number of steps may be smaller if the grid dimension is very
      ! non-optimal (both left column images and right row images are >
      ! 1).
      min_nimages = MIN(left_col_nimages, right_row_nimages)
      nvirt_k = left_npcols*left_col_nimages
      nsteps_k = nvirt_k/min_nimages
      !
      ! Set RMA windows for the original data
      CALL mp_win_lock_all(buffers_orig%left%data_win)
      CALL mp_win_lock_all(buffers_orig%left%meta_win)
      CALL mp_win_lock_all(buffers_orig%right%data_win)
      CALL mp_win_lock_all(buffers_orig%right%meta_win)
      !
      ! Wait refs and max norms (sent in make_buffers)
      CALL mp_waitall(requests)
      ! Delete buffers used for collectives
      DEALLOCATE (right_local_refs_displ_unmerged, left_local_refs_displ_unmerged)
      IF (otf_filtering .AND. use_mpi_filtering) THEN
         DEALLOCATE (right_local_max_norms, left_local_max_norms)
      ENDIF
      DEALLOCATE (right_local_refs_meta_size, left_local_refs_meta_size)
      DEALLOCATE (right_local_refs_data_size, left_local_refs_data_size)
      DEALLOCATE (right_no_empty_images, left_no_empty_images)
      !
      product_matrix%m%nblks = 0
      product_matrix%m%nze = 0
      product_matrix%m%row_p(:) = 0
      !
      ! Count the maximum possible multiplies per row for on-the-fly filtering
      IF (otf_filtering) THEN
         ! Wait for counts (sent in make_buffers)
         CALL mp_wait(request_count_rows)
         nblkrows_local = SIZE(left_total_row_counts)
         ALLOCATE (row_max_epss(nblkrows_local))
         index_row_max_epss = 1
         filter_eps_sp = REAL(filter_eps, KIND=KIND(row_max_epss))
!$OMP PARALLEL DO DEFAULT (NONE) &
!$OMP SHARED(nblkrows_local,row_max_epss,filter_eps_sp,&
!$OMP        left_total_row_counts) &
!$OMP REDUCTION(MAX:index_row_max_epss)
         ! Determine the maximum per-block epsilon
         DO row = 1, nblkrows_local
            row_max_epss(row) = &
               filter_eps_sp/REAL(MAX(1, left_total_row_counts(row)), KIND=KIND(row_max_epss))
            ! Use integers for a fast comparison
            index_row_max_epss = MAX(index_row_max_epss, left_total_row_counts(row))
         ENDDO
!$OMP END PARALLEL DO
         min_row_max_epss = filter_eps_sp/REAL(index_row_max_epss, KIND=KIND(row_max_epss))
         DEALLOCATE (left_total_row_counts)
      ELSE
         !
         ! The array must be valid when passed to called subroutines.
         ALLOCATE (left_norms(0, min_nimages), right_norms(0, min_nimages), row_max_epss(0))
      ENDIF
      !
      ALLOCATE (multrec(0:nthreads-1))
      !
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP          PRIVATE (size_guess, ithread) &
!$OMP          SHARED (product_matrix, multrec, &
!$OMP                  keep_sparsity, filter_eps, &
!$OMP                  row_max_epss, max_nblocks, &
!$OMP                  matrix_right, nthreads)
      !
      ! Setup product work areas
      !
      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      !
      ! The work arrays have to be setup
      size_guess = product_matrix%m%wms(ithread+1)%datasize ! Should be minimal
      CALL dbcsr_data_ensure_size(product_matrix%m%wms(ithread+1)%data_area, &
                                  size_guess)
      CALL dbcsr_data_set_size_referenced(product_matrix%m%wms(ithread+1)%data_area, &
                                          product_matrix%m%wms(ithread+1)%datasize)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%row_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%col_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%blk_p, ub=1)
      ALLOCATE (multrec(ithread)%p)
      CALL dbcsr_mm_multrec_init(multrec(ithread)%p, &
                                 product=product_matrix%m, &
                                 keep_sparsity=keep_sparsity, &
                                 eps=filter_eps, &
                                 row_max_epss=row_max_epss, &
                                 block_estimate=max_nblocks/nthreads, &
                                 right_row_blk_size=dbcsr_row_block_sizes(matrix_right))
!$OMP END PARALLEL
      !
      ! update capacity of memory-pools
      IF (has_acc) THEN
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_1%pool, &
                                            capacity=left_col_nimages+right_row_nimages)
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_2%pool, &
                                            capacity=left_col_nimages+right_row_nimages)
      ENDIF
      !
      ! Pre-CANNON execution: filtering and resize buffers for communications
      CALL timeset(routineN//"_precannon", handle2)
      !
      ALLOCATE (right_vunmerged(0:min_nimages-1, 0:nsteps_k-1))
      ALLOCATE (left_vunmerged(0:min_nimages-1, 0:nsteps_k-1))
      right_vunmerged(:, :) = 0
      left_vunmerged(:, :) = 0
      !
      ALLOCATE (right_displ_vunmerged(0:nvirt_k-1))
      ALLOCATE (left_displ_vunmerged(0:nvirt_k-1))
      !
      ALLOCATE (nofiltered(0:min_nimages-1, 0:nsteps_k-1))
      nofiltered(:, :) = .FALSE.
      ALLOCATE (right_vmerged(right_col_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      ALLOCATE (left_vmerged(left_row_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      !
      right_max_data_size = 0; right_max_meta_size = 0
      left_max_data_size = 0; left_max_meta_size = 0
      !
      ALLOCATE (left_max_meta_size_merged(0:min_nimages-1))
      left_max_meta_size_merged(:) = 0
      ALLOCATE (right_max_meta_size_merged(0:min_nimages-1))
      right_max_meta_size_merged(:) = 0
      !
!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED (nsteps_k,min_nimages,right_col_nimages,imgdist_right,&
!$OMP         otf_filtering,use_mpi_filtering,right_nprows,left_npcols,&
!$OMP         right_myfirstvrow,right_myfirstvcol,right_max_norms,&
!$OMP         left_row_nimages,imgdist_left,left_myfirstvrow,left_myfirstvcol,&
!$OMP         left_max_norms,min_row_max_epss,right_row_nimages,&
!$OMP         right_refs_meta_size,&
!$OMP         left_refs_meta_size,right_refs_data_size,left_refs_data_size,&
!$OMP         right_max_data_size,left_max_data_size,left_col_nimages,&
!$OMP         right_max_meta_size,left_max_meta_size,&
!$OMP         right_data_size,left_data_size,right_meta_size,left_meta_size,&
!$OMP         size_index_unmerged,left_max_meta_size_merged,right_max_meta_size_merged,&
!$OMP         right_vunmerged,left_vunmerged,nofiltered,&
!$OMP         right_vmerged,left_vmerged,left_displ_vunmerged,right_displ_vunmerged,&
!$OMP         right_no_empty_images_displ,left_no_empty_images_displ) &
!$OMP PRIVATE (metronome,v_ki,v_ci,v_ri,recv_vrow,recv_vcol,right_vrow,&
!$OMP          right_vcol,left_vrow,left_vcol,meta_size_image,iproc)
      !
      v_ci = LBOUND(right_refs_meta_size, 1)
!$OMP DO
      DO iproc = 0, right_nprows-1
         recv_vrow = iproc*right_row_nimages
         right_displ_vunmerged(recv_vrow) = right_no_empty_images_displ(iproc+1)
         DO v_ki = recv_vrow+1, recv_vrow+right_row_nimages-1
            IF (right_refs_meta_size(v_ci, v_ki-1) .EQ. 0) THEN
               right_displ_vunmerged(v_ki) = right_displ_vunmerged(v_ki-1)
            ELSE
               right_displ_vunmerged(v_ki) = right_displ_vunmerged(v_ki-1)+1
            ENDIF
         ENDDO
      ENDDO
!$OMP END DO
      v_ri = LBOUND(left_refs_meta_size, 1)
!$OMP DO
      DO iproc = 0, left_npcols-1
         recv_vcol = iproc*left_col_nimages
         left_displ_vunmerged(recv_vcol) = left_no_empty_images_displ(iproc+1)
         DO v_ki = recv_vcol+1, recv_vcol+left_col_nimages-1
            IF (left_refs_meta_size(v_ri, v_ki-1) .EQ. 0) THEN
               left_displ_vunmerged(v_ki) = left_displ_vunmerged(v_ki-1)
            ELSE
               left_displ_vunmerged(v_ki) = left_displ_vunmerged(v_ki-1)+1
            ENDIF
         ENDDO
      ENDDO
!$OMP END DO
      !
      DO metronome = 0, nsteps_k-1
         !
         ! Take first cluster global virtual coordinates
         CALL image_calculator(imgdist_right, &
                               vprow=right_vrow, vpcol=right_vcol, &
                               myvprow=right_myfirstvrow, &
                               myvpcol=right_myfirstvcol, &
                               vprow_shift=metronome*min_nimages, &
                               shifting='R')
         CALL image_calculator(imgdist_left, &
                               vprow=left_vrow, vpcol=left_vcol, &
                               myvprow=left_myfirstvrow, &
                               myvpcol=left_myfirstvcol, &
                               vpcol_shift=metronome*min_nimages, &
                               shifting='L')
         !
!$OMP MASTER
         right_data_size = 0
         left_data_size = 0
         right_meta_size = 0
         left_meta_size = 0
         right_vmerged(:, :, metronome) = right_vcol
         left_vmerged(:, :, metronome) = left_vrow
!$OMP END MASTER
!$OMP BARRIER
         !
!$OMP DO REDUCTION(+:right_data_size,left_data_size,right_meta_size,left_meta_size) &
!$OMP    schedule(dynamic)
         DO v_ki = 0, min_nimages-1
            !
            recv_vrow = v_ki+right_vrow
            recv_vcol = v_ki+left_vcol
            !
            IF (right_refs_meta_size(right_vcol, recv_vrow) .NE. 0) THEN
               right_vunmerged(v_ki, metronome) = 1
            ENDIF
            IF (left_refs_meta_size(left_vrow, recv_vcol) .NE. 0) THEN
               left_vunmerged(v_ki, metronome) = 1
            ENDIF
            !
            ! Max norms evaluation for filtering
            !
            IF (left_vunmerged(v_ki, metronome) .EQ. 1 .AND. &
                right_vunmerged(v_ki, metronome) .EQ. 1) THEN
               IF (otf_filtering .AND. use_mpi_filtering) THEN
                  IF (left_max_norms(left_displ_vunmerged(recv_vcol)+1)* &
                      right_max_norms(right_displ_vunmerged(recv_vrow)+1) .LT. &
                      min_row_max_epss) CYCLE
               ENDIF
               nofiltered(v_ki, metronome) = .TRUE.
               !
               ! resize buffers taking in account filtering
               right_data_size = right_data_size+ &
                                 right_refs_data_size(right_displ_vunmerged(recv_vrow)+1)
               right_meta_size = right_meta_size+ &
                                 right_refs_meta_size(right_vcol, recv_vrow)
               meta_size_image = right_refs_meta_size(right_vcol, recv_vrow)- &
                                 dbcsr_slot_nblks
               right_max_meta_size_merged(v_ki) = MAX(right_max_meta_size_merged(v_ki), meta_size_image)
               !
               left_data_size = left_data_size+ &
                                left_refs_data_size(left_displ_vunmerged(recv_vcol)+1)
               left_meta_size = left_meta_size+ &
                                left_refs_meta_size(left_vrow, recv_vcol)
               meta_size_image = left_refs_meta_size(left_vrow, recv_vcol)- &
                                 size_index_unmerged
               left_max_meta_size_merged(v_ki) = MAX(left_max_meta_size_merged(v_ki), meta_size_image)
            ENDIF
         ENDDO
!$OMP END DO
!$OMP MASTER
         right_max_data_size = MAX(right_max_data_size, right_data_size)
         left_max_data_size = MAX(left_max_data_size, left_data_size)
         right_max_meta_size = MAX(right_max_meta_size, right_meta_size)
         left_max_meta_size = MAX(left_max_meta_size, left_meta_size)
!$OMP END MASTER
!$OMP BARRIER
      ENDDO
!$OMP END PARALLEL
      IF (otf_filtering .AND. use_mpi_filtering) THEN
         DEALLOCATE (right_max_norms, left_max_norms)
      ENDIF
      !
      CALL timestop(handle2)
      !
      ! Preallocate norms arrays
      IF (otf_filtering) THEN
         ALLOCATE (left_norms(MAXVAL(left_max_meta_size_merged)/3, min_nimages))
         ALLOCATE (right_norms(MAXVAL(right_max_meta_size_merged)/3, min_nimages))
      ENDIF
      left_max_meta_size_merged = left_max_meta_size_merged+size_index_merged
      right_max_meta_size_merged = right_max_meta_size_merged+dbcsr_num_slots
      !
      ! Prepare buffers for computation
      IF (nsteps_k .GT. 1) THEN
         ! Right
         CALL buffer_init(buffers_1%right, data_type, &
                          right_max_data_size, &
                          right_max_meta_size, &
                          requests_size=min_nimages, &
                          data_memory_type=memtype_abpanel_1)
         CALL setup_buffer_matrices_images(buffers_1%right%buffer, min_nimages, &
                                           imgdist_right, matrix_right, &
                                           right_max_meta_size_merged, &
                                           buffers_1%right%data)
         ! Left
         CALL buffer_init(buffers_1%left, data_type, &
                          left_max_data_size, &
                          left_max_meta_size, &
                          requests_size=min_nimages, &
                          data_memory_type=memtype_abpanel_1)
         CALL setup_buffer_matrices_images(buffers_1%left%buffer, min_nimages, &
                                           imgdist_left, matrix_left, &
                                           left_max_meta_size_merged, &
                                           buffers_1%left%data)
         !
         IF (has_acc) THEN
            CALL dbcsr_data_init(buffers_1%trs_stackbuf)
            CALL dbcsr_data_new(buffers_1%trs_stackbuf, &
                                data_type=dbcsr_type_int_4, data_size=1000, &
                                memory_type=memtype_trsbuffer_1)
         ENDIF
         !
         buffers_calc => buffers_1
      ENDIF
      !
      ! Prepare buffers for communication
      ! Right
      CALL buffer_init(buffers_2%right, data_type, &
                       right_max_data_size, &
                       right_max_meta_size, &
                       requests_size=min_nimages, &
                       data_memory_type=memtype_abpanel_2)
      CALL setup_buffer_matrices_images(buffers_2%right%buffer, min_nimages, &
                                        imgdist_right, matrix_right, &
                                        right_max_meta_size_merged, &
                                        buffers_2%right%data)
      ! Left
      CALL buffer_init(buffers_2%left, data_type, &
                       left_max_data_size, &
                       left_max_meta_size, &
                       requests_size=min_nimages, &
                       data_memory_type=memtype_abpanel_2)
      CALL setup_buffer_matrices_images(buffers_2%left%buffer, min_nimages, &
                                        imgdist_left, matrix_left, &
                                        left_max_meta_size_merged, &
                                        buffers_2%left%data)
      !
      IF (has_acc) THEN
         CALL dbcsr_data_init(buffers_2%trs_stackbuf)
         CALL dbcsr_data_new(buffers_2%trs_stackbuf, &
                             data_type=dbcsr_type_int_4, data_size=1000, &
                             memory_type=memtype_trsbuffer_2)
      ENDIF
      !
      buffers_comm => buffers_2
      !
      DEALLOCATE (right_max_meta_size_merged, left_max_meta_size_merged)
      !
      ! Setup the receive data pointers
      CALL dbcsr_data_init(data_get)
      CALL dbcsr_data_new(data_get, data_type)
      !
      ! These values for meta data are used for global values
      right_global_indices(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
         (/ &
         matrix_right%m%nblkrows_total, &
         matrix_right%m%nblkcols_total, &
         matrix_right%m%nfullrows_total, &
         matrix_right%m%nfullcols_total, &
         0, 0, &
         matrix_right%m%nfullrows_local, &
         matrix_right%m%nfullcols_local/)
      left_global_indices(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
         (/ &
         matrix_left%m%nblkrows_total, &
         matrix_left%m%nblkcols_total, &
         matrix_left%m%nfullrows_total, &
         matrix_left%m%nfullcols_total, &
         0, 0, &
         matrix_left%m%nfullrows_local, &
         matrix_left%m%nfullcols_local/)
      !
      ALLOCATE (right_get_requests_images(min_nimages))
      ALLOCATE (left_get_requests_images(min_nimages))
      ALLOCATE (common_unmerged(min_nimages))
      ALLOCATE (indices_common_unmerged(min_nimages))
      !
      ! Here is the main loop
      !
      CALL timeset(routineN//"_loop", handle2)
      !
      grouped_k_index: DO metronome = 0, nsteps_k
         !
         IF (debug_mod) WRITE (*, '(1X,A,3(1X,A,1X,I5))') routineN, &
            "step", metronome, &
            "first k", metronome*min_nimages, &
            "last k", (metronome+1)*min_nimages-1
         !
         ! Matrix transfer. Transfer in all but the last loop
         ! iteration.
         xfer: IF (metronome .LT. nsteps_k) THEN
            !
            IF (has_acc) THEN
               CALL timeset(routineN//"_acc_sync", handle3)
               CALL acc_event_synchronize(buffers_comm%right%data%d%acc_ready)
               CALL acc_event_synchronize(buffers_comm%left%data%d%acc_ready)
               CALL timestop(handle3)
            ENDIF
            !
            right_meta_offset = 0
            right_meta_size = 0
            right_data_offset = 0
            right_data_size = 0
            buffers_comm%right%nrequests = 0
            buffers_comm%right%offset(:, :) = 0
            !
            ! Take first cluster global virtual coordinates
            CALL image_calculator(imgdist_right, &
                                  vprow=buffers_comm%right%vprow, &
                                  vpcol=buffers_comm%right%vpcol, &
                                  myvprow=right_myfirstvrow, &
                                  myvpcol=right_myfirstvcol, &
                                  vprow_shift=metronome*min_nimages, &
                                  shifting='R')
            !
            left_meta_offset = 0
            left_meta_size = 0
            left_data_offset = 0
            left_data_size = 0
            buffers_comm%left%nrequests = 0
            buffers_comm%left%offset(:, :) = 0
            CALL image_calculator(imgdist_left, &
                                  vprow=buffers_comm%left%vprow, &
                                  vpcol=buffers_comm%left%vpcol, &
                                  myvprow=left_myfirstvrow, &
                                  myvpcol=left_myfirstvcol, &
                                  vpcol_shift=metronome*min_nimages, &
                                  shifting='L')
            !
            right_v_i = 0
            left_v_i = 0
            buffers_comm%right%get_requests(:, :) = mp_request_null
            buffers_comm%left%get_requests(:, :) = mp_request_null
            !
            DO v_ki = 0, min_nimages-1
               ! Right
               CALL cannon_transfer(v_ki, buffers_comm%right%vprow, &
                                    right_row_nimages, &
                                    right_vunmerged(v_ki, metronome), right_vmerged(1, v_ki, metronome), &
                                    right_displ_vunmerged, &
                                    right_v_i, nofiltered(v_ki, metronome), &
                                    right_refs_displ_unmerged, &
                                    buffers_comm%right, right_refs_meta_size, right_refs_data_size, &
                                    right_meta_size, right_data_size, &
                                    right_meta_offset, right_data_offset, &
                                    buffers_orig%right%meta_win, buffers_orig%right%data_win, right_myprow, metronome .GT. 0, &
                                    data_get, data_type_byte, buffers_orig%right)
               ! Left
               CALL cannon_transfer(v_ki, buffers_comm%left%vpcol, &
                                    left_col_nimages, &
                                    left_vunmerged(v_ki, metronome), left_vmerged(1, v_ki, metronome), &
                                    left_displ_vunmerged, &
                                    left_v_i, nofiltered(v_ki, metronome), &
                                    left_refs_displ_unmerged, &
                                    buffers_comm%left, left_refs_meta_size, left_refs_data_size, &
                                    left_meta_size, left_data_size, &
                                    left_meta_offset, left_data_offset, &
                                    buffers_orig%left%meta_win, buffers_orig%left%data_win, left_mypcol, metronome .GT. 0, &
                                    data_get, data_type_byte, buffers_orig%left)
            ENDDO
            ! Set the referenced sizes to the actual data moved via MPI
            CALL dbcsr_data_set_size_referenced(buffers_comm%right%data, &
                                                buffers_comm%right%offset(idata_size, right_v_i+1))
            CALL dbcsr_data_set_size_referenced(buffers_comm%left%data, &
                                                buffers_comm%left%offset(idata_size, left_v_i+1))
         ENDIF xfer
         !
         ! Wait data, merge and do the multiplications.
         ! Exclude the first interation
         wait_merge_calc: IF (metronome .GT. 0) THEN
            IF (debug_mod) WRITE (*, '(1X,A)') routineN//" waiting for right and left"
            !
            buffers_calc%right%buffer%mats(:)%m%nblks = 0
            buffers_calc%left%buffer%mats(:)%m%nblks = 0
            !
            IF (metronome == nsteps_k) THEN
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (multrec) &
!$OMP PRIVATE (ithread)
               ithread = 0
!$             ithread = omp_get_thread_num()
               !
               CALL dbcsr_mm_multrec_phaseout(multrec(ithread)%p)
!$OMP END PARALLEL
            ENDIF
            !
            right_nrequests = buffers_calc%right%nrequests
            left_nrequests = buffers_calc%left%nrequests
            !
            IF (right_nrequests .GT. 0 .AND. left_nrequests .GT. 0) THEN
               !
               common_unmerged(:) = .FALSE.
               !
               IF (min_nimages .EQ. 1 .OR. has_acc) THEN
                  CALL mp_waitall(buffers_calc%right%get_requests(:, 1:right_nrequests))
                  IF (has_acc) CALL dbcsr_data_host2dev(buffers_calc%right%data)
                  CALL merge_calc(buffers_calc%right, &
                                  buffers_calc%left, &
                                  multrec, .TRUE., flop, &
                                  imgdist_right, .FALSE., &
                                  1, min_nimages, &
                                  min_nimages, right_col_nimages, &
                                  right_global_indices, right_vunmerged(:, metronome-1), &
                                  right_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                  right_norms, left_norms, &
                                  common_unmerged, indices_common_unmerged, has_acc)
                  CALL mp_waitall(buffers_calc%left%get_requests(:, 1:left_nrequests))
                  IF (has_acc) CALL dbcsr_data_host2dev(buffers_calc%left%data)
                  CALL merge_calc(buffers_calc%right, &
                                  buffers_calc%left, &
                                  multrec, .TRUE., flop, &
                                  imgdist_left, .TRUE., &
                                  1, min_nimages, &
                                  left_row_nimages, min_nimages, &
                                  left_global_indices, left_vunmerged(:, metronome-1), &
                                  left_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                  right_norms, left_norms, &
                                  common_unmerged, indices_common_unmerged, &
                                  has_acc, nthreads, buffers_calc%trs_stackbuf)
               ELSE
                  !
                  ! Use async communication-computation over images
                  CALL timeset("mp_waitall_2", handle3)
                  !
                  ! Counting the requests per each unmerged image
                  right_get_requests_images(:) = 0
                  DO iget = 1, right_nrequests
                     right_get_requests_images(buffers_calc%right%get_requests_map(1, iget)) = 1
                  ENDDO
                  left_get_requests_images(:) = 0
                  DO iget = 1, left_nrequests
                     left_get_requests_images(buffers_calc%left%get_requests_map(1, iget)) = 1
                  ENDDO
                  !
                  DO WHILE (right_nrequests .GT. 0 .OR. &
                            left_nrequests .GT. 0)
                     IF (right_nrequests .GT. 0) THEN
                        CALL wait_merge_calc_low(buffers_calc%right, &
                                                 buffers_calc%left, &
                                                 right_nrequests, &
                                                 multrec, .TRUE., flop, &
                                                 right_get_requests_images, &
                                                 imgdist_right, .FALSE., min_nimages, right_col_nimages, &
                                                 right_global_indices, right_vunmerged(:, metronome-1), &
                                                 right_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                                 right_norms, left_norms, &
                                                 common_unmerged, indices_common_unmerged)
                     ENDIF
                     IF (left_nrequests .GT. 0) THEN
                        CALL wait_merge_calc_low(buffers_calc%right, &
                                                 buffers_calc%left, &
                                                 left_nrequests, &
                                                 multrec, .TRUE., flop, &
                                                 left_get_requests_images, &
                                                 imgdist_left, .TRUE., left_row_nimages, min_nimages, &
                                                 left_global_indices, left_vunmerged(:, metronome-1), &
                                                 left_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                                 right_norms, left_norms, &
                                                 common_unmerged, indices_common_unmerged, &
                                                 nthreads)
                     ENDIF
                  ENDDO
                  !
                  CALL timestop(handle3)
               ENDIF
            ENDIF
            !
            IF (debug_mod) THEN
               DO v_ki = 1, min_nimages
                  IF (buffers_calc%right%buffer%mats(v_ki)%m%nblks .NE. 0) &
                     CALL dbcsr_print(buffers_calc%right%buffer%mats(v_ki), &
                                      nodata=.TRUE.)
                  IF (buffers_calc%left%buffer%mats(v_ki)%m%nblks .NE. 0) &
                     CALL dbcsr_print(buffers_calc%left%buffer%mats(v_ki), &
                                      nodata=.TRUE.)
               ENDDO
            ENDIF
         ENDIF wait_merge_calc
         !
         IF (metronome .LT. nsteps_k) THEN
            buffers => buffers_comm
            buffers_comm => buffers_calc
            buffers_calc => buffers
         ENDIF
         !
      ENDDO grouped_k_index
      !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (multrec) &
!$OMP PRIVATE (ithread)
      ithread = 0
!$    ithread = omp_get_thread_num()
      CALL dbcsr_mm_multrec_finalize(multrec(ithread)%p)
      DEALLOCATE (multrec(ithread)%p)
!$OMP END PARALLEL
      !
      CALL timestop(handle2)
      !
      CALL m_memory(mem)
      max_memory = MAX(max_memory, REAL(mem))
      !
      IF (debug_mod) THEN
         v_ki = 0
         DO blk = 1, SIZE(product_matrix%m%blk_p)
            v_ki = MAX(v_ki, ABS(product_matrix%m%blk_p(blk)))
         ENDDO
         WRITE (*, *) routineN//" Actual final size", &
            LOG(REAL(dbcsr_data_get_size(product_matrix%m%data_area)))/LOG(10.0), &
            LOG(REAL(v_ki))/LOG(10.0)
      ENDIF
      !
      IF (has_acc) THEN
         CALL dbcsr_data_release(buffers_1%trs_stackbuf)
         CALL dbcsr_data_release(buffers_2%trs_stackbuf)
      ENDIF
      !
      DEALLOCATE (right_norms, left_norms, row_max_epss)
      !
      CALL dbcsr_data_clear_pointer(data_get)
      CALL dbcsr_data_release(data_get)
      !
      ! clean-up of communication buffers
      IF (nsteps_k .GT. 1) THEN
         CALL dbcsr_destroy_array(buffers_1%right%buffer)
         CALL dbcsr_destroy_array(buffers_1%left%buffer)
      ENDIF
      !
      DEALLOCATE (right_get_requests_images, left_get_requests_images)
      DEALLOCATE (common_unmerged, indices_common_unmerged)
      DEALLOCATE (right_refs_meta_size, left_refs_meta_size)
      DEALLOCATE (right_refs_data_size, left_refs_data_size)
      DEALLOCATE (right_refs_displ_unmerged, left_refs_displ_unmerged)
      DEALLOCATE (multrec)
      DEALLOCATE (right_vunmerged, left_vunmerged)
      DEALLOCATE (nofiltered)
      DEALLOCATE (right_vmerged, left_vmerged)
      DEALLOCATE (right_displ_vunmerged, left_displ_vunmerged)
      DEALLOCATE (right_no_empty_images_displ, left_no_empty_images_displ)
      CALL dbcsr_destroy_array(buffers_2%right%buffer)
      CALL dbcsr_destroy_array(buffers_2%left%buffer)
      !
      CALL dbcsr_finalize(product_matrix)
      !
      ! clean-up wins
      CALL mp_win_unlock_all(buffers_orig%right%data_win)
      CALL mp_win_unlock_all(buffers_orig%right%meta_win)
      CALL mp_win_unlock_all(buffers_orig%left%data_win)
      CALL mp_win_unlock_all(buffers_orig%left%meta_win)
      !
      CALL timestop(handle)
   END SUBROUTINE cannon_multiply_low_exp

! **************************************************************************************************
!> \brief ...
!> \param v_ki ...
!> \param vpunmerged ...
!> \param unmerged_nimages ...
!> \param vunmerged ...
!> \param vmerged ...
!> \param displ_vunmerged ...
!> \param v_i ...
!> \param nofiltered ...
!> \param refs_displ_unmerged ...
!> \param buffer ...
!> \param refs_meta_size ...
!> \param refs_data_size ...
!> \param meta_size ...
!> \param data_size ...
!> \param meta_offset ...
!> \param data_offset ...
!> \param meta_win ...
!> \param data_win ...
!> \param mypunmerged ...
!> \param save_statistics ...
!> \param data_get ...
!> \param data_type_byte ...
!> \param buffer_orig ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE cannon_transfer(v_ki, vpunmerged, &
                              unmerged_nimages, &
                              vunmerged, vmerged, &
                              displ_vunmerged, &
                              v_i, nofiltered, &
                              refs_displ_unmerged, &
                              buffer, refs_meta_size, refs_data_size, &
                              meta_size, data_size, &
                              meta_offset, data_offset, &
                              meta_win, data_win, mypunmerged, save_statistics, &
                              data_get, data_type_byte, &
                              buffer_orig)
      INTEGER, INTENT(IN)                                :: v_ki, vpunmerged, unmerged_nimages, &
                                                            vunmerged, vmerged
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: displ_vunmerged
      INTEGER, INTENT(INOUT)                             :: v_i
      LOGICAL, INTENT(IN)                                :: nofiltered
      INTEGER, ALLOCATABLE, DIMENSION(:, :), &
         INTENT(INOUT)                                   :: refs_displ_unmerged
      TYPE(dbcsr_buffer), INTENT(INOUT)                  :: buffer
      INTEGER, ALLOCATABLE, DIMENSION(:, :), INTENT(IN)  :: refs_meta_size
      INTEGER, DIMENSION(:), INTENT(IN)                  :: refs_data_size
      INTEGER, INTENT(INOUT)                             :: meta_size, data_size, meta_offset, &
                                                            data_offset
      INTEGER, INTENT(IN)                                :: meta_win, data_win, mypunmerged
      LOGICAL, INTENT(IN)                                :: save_statistics
      TYPE(dbcsr_data_obj)                               :: data_get
      INTEGER, INTENT(IN)                                :: data_type_byte
      TYPE(dbcsr_buffer), INTENT(IN)                     :: buffer_orig

      INTEGER                                            :: recv_punmerged, recv_vunmerged
      INTEGER, DIMENSION(:), POINTER                     :: meta_get

! Left  : merged => row, unmerged => col
! Right : merged => col, unmerged => row

      IF (vunmerged .EQ. 1) THEN
         v_i = v_i+1
         IF (nofiltered) THEN
            recv_vunmerged = v_ki+vpunmerged
            recv_punmerged = recv_vunmerged/unmerged_nimages
            buffer%offset(idata_size, v_i+1) = &
               buffer%offset(idata_size, v_i)+ &
               refs_data_size(displ_vunmerged(recv_vunmerged)+1)
            buffer%offset(imeta_size, v_i+1) = &
               buffer%offset(imeta_size, v_i)+ &
               refs_meta_size(vmerged, recv_vunmerged)
            meta_size = meta_size+refs_meta_size(vmerged, recv_vunmerged)
            data_size = data_size+refs_data_size(displ_vunmerged(recv_vunmerged)+1)
            !
            buffer%nrequests = buffer%nrequests+1
            buffer%get_requests_map(1, buffer%nrequests) = v_ki+1
            buffer%get_requests_map(2, buffer%nrequests) = v_ki+1
            !
            meta_get => buffer%meta(meta_offset+1:meta_size)
            CALL mp_rget(meta_get, recv_punmerged, &
                         meta_win, &
                         buffer_orig%meta, &
                         mypunmerged, &
                         refs_displ_unmerged(imeta_displ, recv_vunmerged), &
                         buffer%get_requests(1, buffer%nrequests))
            meta_offset = meta_size
            CALL dbcsr_data_set_pointer( &
               area=data_get, &
               rsize=data_size-data_offset, &
               csize=1, &
               pointee=buffer%data, &
               source_lb=data_offset+1)
            CALL dbcsr_rget_any(data_get, recv_punmerged, &
                                data_win, &
                                buffer_orig%data, &
                                mypunmerged, &
                                refs_displ_unmerged(idata_displ, recv_vunmerged), &
                                buffer%get_requests(2, buffer%nrequests))
            IF (save_statistics) THEN
               CALL count_mpi_statistics(dbcsr_mpi_statistics%data_size(1, :), &
                                         data_size-data_offset, &
                                         dbcsr_mpi_statistics%data_size_breakdown(:, :, 1), data_type_byte)
               dbcsr_mpi_statistics%nexchanged = dbcsr_mpi_statistics%nexchanged+1
            ENDIF
            data_offset = data_size
         ELSE
            buffer%offset(:, v_i+1) = buffer%offset(:, v_i)
            dbcsr_mpi_statistics%nfiltered = dbcsr_mpi_statistics%nfiltered+1
         ENDIF
      ENDIF
   END SUBROUTINE cannon_transfer

! **************************************************************************************************
!> \brief Multiplies two DBCSR matrices by means of RMA MPI operations.
!>        This algorithm is experimental and it should be not used in
!>        production runs.
!>
!> \param imgdist_left ...
!> \param imgdist_right ...
!> \param matrix_left ...
!> \param matrix_right ...
!> \param[out] product_matrix      DBCSR product matrix
!> \param[in] retain_sparsity      (optional) retain the sparsity of the
!>                                 existing product matrix; default is no
!> \param filter_eps ...
!> \param[out] flop                (optional) effective flop
!> \param keep_product_data ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE cannon_multiply_low_rma(imgdist_left, imgdist_right, &
                                      matrix_left, matrix_right, &
                                      product_matrix, &
                                      retain_sparsity, &
                                      filter_eps, flop, keep_product_data)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist_left, imgdist_right
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_left, matrix_right
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: product_matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(kind=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT)                   :: flop
      LOGICAL, INTENT(IN)                                :: keep_product_data

      CHARACTER(len=*), PARAMETER :: routineN = 'cannon_multiply_low_rma', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, col, data_type, data_type_byte, grp_gather, grp_left, grp_right, handle, &
         handle2, handle3, iget, index_row_max_epss, iproc, ithread, left_col_mult, &
         left_col_nimages, left_data_offset, left_data_offset_diag, left_data_size, &
         left_max_data_size, left_max_meta_size, left_meta_offset, left_meta_offset_diag, &
         left_meta_size, left_myfirstvcol, left_myfirstvrow, left_mypcol, left_myprow, &
         left_npcols, left_nprows, left_nrequests, left_row_mult, left_row_nimages, left_v_i, &
         left_vcol, left_vrow, local_data_size_gather, local_meta_size_gather, meta_size_image, &
         meta_size_scatter
      INTEGER :: metronome, min_nimages, mynode, mypcol, myprow, nblkrows_local, nimages_scatter, &
         nsteps_k, nthreads, numnodes, nvirt_k, recv_vcol, recv_vcol_displ, recv_vrow, &
         recv_vrow_displ, request_product_matrix, right_col_mult, right_col_nimages, &
         right_data_offset, right_data_offset_diag, right_data_size, right_max_data_size, &
         right_max_meta_size, right_meta_offset, right_meta_offset_diag, right_meta_size, &
         right_myfirstvcol, right_myfirstvrow, right_mypcol, right_myprow, right_npcols, &
         right_nprows, right_nrequests, right_row_mult, right_row_nimages, right_v_i, right_vcol, &
         right_vrow, row
      INTEGER :: size_guess, size_index_merged, size_index_unmerged, v_ci, v_k_loc, v_ki, v_ri
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, ALLOCATABLE, DIMENSION(:) :: data_displ_gather, data_size_gather, &
         indices_common_unmerged, left_displ_vunmerged, left_get_requests_images, &
         left_max_meta_size_merged, meta_displ_gather, meta_size_gather, right_displ_vunmerged, &
         right_get_requests_images, right_max_meta_size_merged, v_k_scatter
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: left_recv_filtered, left_vunmerged, &
                                                            right_recv_filtered, right_vunmerged
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: left_vmerged, right_vmerged
      INTEGER, DIMENSION(:), POINTER                     :: blk_p, col_i
      INTEGER, DIMENSION(dbcsr_slot_nblkrows_total:&
         dbcsr_slot_nfullcols_local)                     :: left_global_indices, right_global_indices
      LOGICAL :: do_bcast, do_diagonal, do_k_scatter, do_rec, do_scatter, is_diagonal, &
         keep_sparsity, left_set_displ, otf_filtering, right_set_displ
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: common_unmerged
      LOGICAL, ALLOCATABLE, DIMENSION(:, :, :)           :: left_nofiltered, right_nofiltered
      REAL(kind=sp)                                      :: filter_eps_sp, min_row_max_epss
      REAL(kind=sp), ALLOCATABLE, DIMENSION(:)           :: row_max_epss
      REAL(kind=sp), ALLOCATABLE, DIMENSION(:, :)        :: left_norms, right_norms
      TYPE(dbcsr_1d_array_type)                          :: buffer_scatter
      TYPE(dbcsr_buffers), POINTER                       :: buffers, buffers_calc, buffers_comm
      TYPE(dbcsr_data_obj)                               :: data_get
      TYPE(dbcsr_mm_multrec_type_p), ALLOCATABLE, &
         DIMENSION(:)                                    :: multrec, multrec_scatter
      TYPE(dbcsr_mp_obj)                                 :: left_mp_obj, product_mp_obj, right_mp_obj
      TYPE(dbcsr_obj)                                    :: product_matrix_scatter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (buffers_calc, buffers_comm, buffers)
      !
      IF (PRESENT(retain_sparsity)) THEN
         keep_sparsity = retain_sparsity
      ELSE
         keep_sparsity = .FALSE.
      ENDIF
      otf_filtering = PRESENT(filter_eps)
      !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (nthreads)
!$OMP MASTER
      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL
      !
      size_index_unmerged = dbcsr_slot_nblks
!$    size_index_unmerged = size_index_unmerged+nthreads+1
      size_index_merged = dbcsr_num_slots
!$    size_index_merged = size_index_merged+nthreads+1
      !
      ! Dummy checks
      IF (.NOT. ASSOCIATED(product_matrix%m%wms)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4919,"Work matrices do not exist")
      IF (SIZE(product_matrix%m%wms) .NE. nthreads) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4921,"Work matrices not correctly sized.")
      IF (.NOT. buffers_orig%left%is_valid .OR. &
          .NOT. buffers_orig%right%is_valid .OR. &
          .NOT. ASSOCIATED(buffers_orig%left%meta) .OR. &
          .NOT. ASSOCIATED(buffers_orig%right%meta) .OR. &
          .NOT. ALLOCATED(left_refs_meta_size) .OR. &
          .NOT. ALLOCATED(left_refs_data_size) .OR. &
          .NOT. ALLOCATED(right_refs_meta_size) .OR. &
          .NOT. ALLOCATED(right_refs_data_size) .OR. &
          .NOT. ALLOCATED(left_local_refs_meta_size) .OR. &
          .NOT. ALLOCATED(left_local_refs_data_size) .OR. &
          .NOT. ALLOCATED(right_local_refs_meta_size) .OR. &
          .NOT. ALLOCATED(right_local_refs_data_size) .OR. &
          .NOT. ALLOCATED(left_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(right_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(left_local_refs_displ_unmerged) .OR. &
          .NOT. ALLOCATED(right_local_refs_displ_unmerged) .OR. &
          (otf_filtering .AND. use_mpi_filtering .AND. &
           (.NOT. (ALLOCATED(left_max_norms) .AND. ALLOCATED(right_max_norms) .AND. &
                   ALLOCATED(left_local_max_norms) .AND. ALLOCATED(right_local_max_norms))))) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4941,"No buffers associated for the RMA algo!")
      !
      ! Set up variables
      flop = 0
      data_type = dbcsr_get_data_type(product_matrix)
      data_type_byte = dbcsr_datatype_sizeof(data_type)
      left_row_nimages = imgdist_left%i%row_decimation
      left_row_mult = imgdist_left%i%row_multiplicity
      left_col_nimages = imgdist_left%i%col_decimation
      left_col_mult = imgdist_left%i%col_multiplicity
      right_row_nimages = imgdist_right%i%row_decimation
      right_row_mult = imgdist_right%i%row_multiplicity
      right_col_nimages = imgdist_right%i%col_decimation
      right_col_mult = imgdist_right%i%col_multiplicity
      left_mp_obj = dbcsr_distribution_mp(imgdist_left%i%main)
      right_mp_obj = dbcsr_distribution_mp(imgdist_right%i%main)
      product_mp_obj = dbcsr_distribution_mp(product_matrix%m%dist)
      numnodes = dbcsr_mp_numnodes(product_mp_obj)
      mynode = dbcsr_mp_mynode(product_mp_obj)
      myprow = dbcsr_mp_myprow(product_mp_obj)
      mypcol = dbcsr_mp_mypcol(product_mp_obj)
      left_nprows = dbcsr_mp_nprows(left_mp_obj)
      left_npcols = dbcsr_mp_npcols(left_mp_obj)
      left_myprow = dbcsr_mp_myprow(left_mp_obj)
      left_mypcol = dbcsr_mp_mypcol(left_mp_obj)
      left_myfirstvrow = left_myprow*left_row_nimages
      left_myfirstvcol = left_mypcol*left_col_nimages
      right_nprows = dbcsr_mp_nprows(right_mp_obj)
      right_npcols = dbcsr_mp_npcols(right_mp_obj)
      right_myprow = dbcsr_mp_myprow(right_mp_obj)
      right_mypcol = dbcsr_mp_mypcol(right_mp_obj)
      right_myfirstvrow = right_myprow*right_row_nimages
      right_myfirstvcol = right_mypcol*right_col_nimages
      grp_right = dbcsr_mp_my_col_group(right_mp_obj)
      grp_gather = dbcsr_mp_my_row_group(right_mp_obj)
      grp_left = dbcsr_mp_my_row_group(left_mp_obj)
      CALL dbcsr_mp_grid_setup(product_mp_obj)
      CALL dbcsr_mp_grid_setup(left_mp_obj)
      CALL dbcsr_mp_grid_setup(right_mp_obj)
      do_bcast = right_col_nimages .NE. 1 .AND. left_row_nimages .NE. 1
      is_diagonal = myprow .EQ. mypcol
      do_scatter = right_npcols .GT. 1 .AND. .NOT. is_diagonal .AND. &
                   right_col_nimages .GT. 1 .AND. .NOT. has_acc
      do_diagonal = right_npcols .GT. 1 .AND. is_diagonal .AND. &
                    right_col_nimages .GT. 1 .AND. .NOT. has_acc
      !
      ! Dummy checks
      ! subcommunicators
      IF (.NOT. dbcsr_mp_has_subgroups(right_mp_obj)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4990,"RMA requires rows subcommunicators for right matrix!")
      IF (.NOT. dbcsr_mp_has_subgroups(left_mp_obj)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4992,"RMA requires columns subcommunicators for left matrix!")
      ! left/right matching
      IF (left_col_nimages .NE. right_row_mult) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4995,"Left/Right image mismatch")
      IF (left_col_mult .NE. right_row_nimages) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4997,"Left/Right image mismatch")
      IF (left_col_nimages*left_npcols .NE. right_row_nimages*right_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",4999,"Left/Right total mismatch")
      ! product/left matching
      IF (left_row_mult*dbcsr_mp_nprows(product_mp_obj) .NE. left_row_nimages*left_nprows) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",5002,"Product/Left total mismatch")
      ! product/left matching
      IF (right_col_mult*dbcsr_mp_npcols(product_mp_obj) .NE. right_col_nimages*right_npcols) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",5005,"Product/Right total mismatch")
      !
      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, &
                                         left_row_nimages*left_col_nimages)
      dbcsr_mpi_statistics%nimages = MAX(dbcsr_mpi_statistics%nimages, &
                                         right_row_nimages*right_col_nimages)
      !
      ! The main transfer loop goes through the virtual rows/columns.
      ! The number of steps may be smaller if the grid dimension is very
      ! non-optimal (both left column images and right row images are >
      ! 1).
      min_nimages = MIN(left_col_nimages, right_row_nimages)
      nvirt_k = left_npcols*left_col_nimages
      nsteps_k = nvirt_k/min_nimages
      !
      do_rec = right_col_nimages .EQ. 1 .AND. left_row_nimages .EQ. 1
      !
      ! Exchange product matrix blocks coodinates
      IF (keep_sparsity .AND. right_npcols .GT. 1 .AND. &
          right_col_nimages .GT. 1 .AND. .NOT. has_acc) THEN
         IF (do_diagonal) THEN
            meta_size_scatter = product_matrix%m%index(dbcsr_slot_size)
         ENDIF
         CALL mp_bcast(meta_size_scatter, right_myprow, grp_gather)
         CALL ensure_array_size(local_meta_product_scatter, ub=meta_size_scatter, &
                                nocopy=.TRUE., memory_type=memtype_mpi_buffer)
         IF (do_diagonal) THEN
            local_meta_product_scatter(1:meta_size_scatter) = &
               product_matrix%m%index(1:meta_size_scatter)
         ENDIF
         CALL mp_ibcast(local_meta_product_scatter, right_myprow, &
                        grp_gather, request_product_matrix)
      ENDIF
      !
      product_matrix%m%nblks = 0
      product_matrix%m%nze = 0
      product_matrix%m%row_p(:) = 0
      !
      ! Count the maximum possible multiplies per row for on-the-fly filtering
      IF (otf_filtering) THEN
         ! Wait for counts (sent in make_buffers)
         CALL mp_wait(request_count_rows)
         nblkrows_local = SIZE(left_total_row_counts)
         ALLOCATE (row_max_epss(nblkrows_local))
         index_row_max_epss = 1
         filter_eps_sp = REAL(filter_eps, KIND=KIND(row_max_epss))
!$OMP PARALLEL DO DEFAULT (NONE) &
!$OMP SHARED(nblkrows_local,row_max_epss,filter_eps_sp,&
!$OMP        left_total_row_counts) &
!$OMP REDUCTION(MAX:index_row_max_epss)
         ! Determine the maximum per-block epsilon
         DO row = 1, nblkrows_local
            row_max_epss(row) = &
               filter_eps_sp/REAL(MAX(1, left_total_row_counts(row)), KIND=KIND(row_max_epss))
            ! Use integers for a fast comparison
            index_row_max_epss = MAX(index_row_max_epss, left_total_row_counts(row))
         ENDDO
!$OMP END PARALLEL DO
         min_row_max_epss = filter_eps_sp/REAL(index_row_max_epss, KIND=KIND(row_max_epss))
         DEALLOCATE (left_total_row_counts)
      ELSE
         !
         ! The array must be valid when passed to called subroutines.
         ALLOCATE (left_norms(0, min_nimages), right_norms(0, min_nimages), row_max_epss(0))
      ENDIF
      !
      CALL timeset(routineN//"_scatter", handle2)
      IF (right_npcols .GT. 1 .AND. right_col_nimages .GT. 1 .AND. .NOT. has_acc) THEN
         CALL mp_waitall(requests_scatter)
         IF (keep_sparsity) THEN
            CALL mp_wait(request_product_matrix)
         ENDIF
      ENDIF
      !
      ! Prepare result matrix for scatter data
      nimages_scatter = 0
      IF (do_scatter .OR. do_diagonal) THEN
         CALL dbcsr_init(product_matrix_scatter)
         !
         ! Check scattered meta
         IF (local_meta_size_scatter .GT. 1) THEN
            ALLOCATE (v_k_scatter((right_row_nimages+right_npcols-1)/right_npcols))
            meta_size_scatter = 0
            DO WHILE (meta_size_scatter+dbcsr_num_slots .LE. local_meta_size_scatter)
               nimages_scatter = nimages_scatter+1
               v_k_scatter(nimages_scatter) = local_meta_scatter(meta_size_scatter+dbcsr_slot_home_rowi)
               meta_size_scatter = meta_size_scatter+local_meta_scatter(meta_size_scatter+dbcsr_slot_size)
            ENDDO
         ENDIF
      ENDIF
      IF (do_scatter) THEN
         IF (.NOT. dbcsr_data_valid(local_data_product_scatter)) THEN
            CALL dbcsr_data_init(local_data_product_scatter)
            CALL dbcsr_data_new(local_data_product_scatter, data_type=data_type, &
                                data_size=1, memory_type=memtype_mpi_buffer)
         ENDIF
         CALL dbcsr_create(product_matrix_scatter, &
                           template=product_matrix, &
                           name=TRIM("Scatter buffer of "//TRIM(product_matrix%m%name)), &
                           make_index=.NOT. keep_sparsity .OR. local_meta_size_scatter .LE. 1, &
                           data_buffer=local_data_product_scatter)
         IF (local_meta_size_scatter .GT. 1) THEN
            max_nblocks = MAX(max_nblocks, meta_size_scatter/3)
            !
            IF (keep_sparsity) THEN
               product_matrix_scatter%m%index => local_meta_product_scatter
               CALL dbcsr_data_ensure_size(product_matrix_scatter%m%data_area, &
                                           product_matrix_scatter%m%index(dbcsr_slot_nze), &
                                           zero_pad=.TRUE.)
            ELSE
               CALL dbcsr_data_set_size_referenced(product_matrix_scatter%m%data_area, 0)
            ENDIF
            !
            product_matrix_scatter%m%index(dbcsr_slot_home_pcol) = myprow
            CALL dbcsr_reset_locals(product_matrix_scatter)
            CALL dbcsr_repoint_index(product_matrix_scatter%m)
            product_matrix_scatter%m%valid = .FALSE.
            ALLOCATE (multrec_scatter(0:nthreads-1))
            !
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP          PRIVATE (size_guess, ithread) &
!$OMP          SHARED (product_matrix_scatter, multrec_scatter, &
!$OMP                  keep_sparsity, filter_eps, &
!$OMP                  row_max_epss, max_nblocks, &
!$OMP                  matrix_right, nthreads, &
!$OMP                  memtype_product_wm_scatter, &
!$OMP                  local_meta_size_scatter, &
!$OMP                  keep_product_data)
            !
            ! Setup product work areas
            !
            ithread = 0
!$          ithread = OMP_GET_THREAD_NUM()
            !
            IF (keep_product_data) THEN
               CALL dbcsr_add_wm_from_matrix(product_matrix_scatter)
            ELSE
               CALL dbcsr_work_create(product_matrix_scatter, work_mutable=.FALSE., &
                                      memory_type=memtype_product_wm_scatter(ithread)%p)
            ENDIF
!$OMP BARRIER
            ! The work arrays have to be setup
            size_guess = product_matrix_scatter%m%wms(ithread+1)%datasize ! Should be minimal
            CALL dbcsr_data_ensure_size(product_matrix_scatter%m%wms(ithread+1)%data_area, &
                                        size_guess)
            CALL dbcsr_data_set_size_referenced(product_matrix_scatter%m%wms(ithread+1)%data_area, &
                                                product_matrix_scatter%m%wms(ithread+1)%datasize)
            CALL ensure_array_size(product_matrix_scatter%m%wms(ithread+1)%row_i, ub=1)
            CALL ensure_array_size(product_matrix_scatter%m%wms(ithread+1)%col_i, ub=1)
            CALL ensure_array_size(product_matrix_scatter%m%wms(ithread+1)%blk_p, ub=1)
            ALLOCATE (multrec_scatter(ithread)%p)
            CALL dbcsr_mm_multrec_init(multrec_scatter(ithread)%p, &
                                       product=product_matrix_scatter%m, &
                                       keep_sparsity=keep_sparsity, &
                                       eps=filter_eps, &
                                       row_max_epss=row_max_epss, &
                                       block_estimate=max_nblocks/nthreads, &
                                       right_row_blk_size=dbcsr_row_block_sizes(matrix_right))
!$OMP END PARALLEL
            product_matrix_scatter%m%nblks = 0
            product_matrix_scatter%m%nze = 0
            product_matrix_scatter%m%row_p(:) = 0
            !
            ! Set images with scatter data
            CALL setup_buffer_matrices_images(buffer_scatter, nimages_scatter, &
                                              imgdist_right, matrix_right, &
                                              data_buffer=local_data_scatter)
            meta_size_scatter = 0
            right_data_size = 0
            DO v_ki = 1, nimages_scatter
               buffer_scatter%mats(v_ki)%m%index => local_meta_scatter(meta_size_scatter+dbcsr_slot_size: &
                                                                       meta_size_scatter+ &
                                                                       local_meta_scatter(meta_size_scatter+ &
                                                                                          dbcsr_slot_size))
               meta_size_scatter = meta_size_scatter+local_meta_scatter(meta_size_scatter+dbcsr_slot_size)
               CALL dbcsr_reset_vlocals(buffer_scatter%mats(v_ki), imgdist_right, do_rows=.TRUE.)
               ! Set local cols
               CALL array_release(buffer_scatter%mats(v_ki)%m%local_cols)
               buffer_scatter%mats(v_ki)%m%local_cols = product_matrix_scatter%m%local_cols
               CALL array_hold(buffer_scatter%mats(v_ki)%m%local_cols)
               buffer_scatter%mats(v_ki)%m%has_local_cols = .TRUE.
               buffer_scatter%mats(v_ki)%m%nblkcols_local = product_matrix_scatter%m%nblkcols_local
               buffer_scatter%mats(v_ki)%m%index(dbcsr_slot_nblkcols_local) = &
                  product_matrix_scatter%m%nblkcols_local
               CALL dbcsr_repoint_index(buffer_scatter%mats(v_ki)%m)
               buffer_scatter%mats(v_ki)%m%index(dbcsr_num_slots+3: &
                                                 buffer_scatter%mats(v_ki)%m%index(dbcsr_slot_size):3) = &
                  buffer_scatter%mats(v_ki)%m%index(dbcsr_num_slots+3: &
                                                    buffer_scatter%mats(v_ki)%m%index(dbcsr_slot_size):3)+ &
                  right_data_size
               right_data_size = right_data_size+buffer_scatter%mats(v_ki)%m%index(dbcsr_slot_nze)
               buffer_scatter%mats(v_ki)%m%valid = .TRUE.
            ENDDO
         ENDIF
      ENDIF
      CALL timestop(handle2)
      !
      ! Set RMA windows for the original data
      CALL mp_win_lock_all(buffers_orig%left%data_win)
      CALL mp_win_lock_all(buffers_orig%left%meta_win)
      CALL mp_win_lock_all(buffers_orig%right%data_win)
      CALL mp_win_lock_all(buffers_orig%right%meta_win)
      !
      ALLOCATE (multrec(0:nthreads-1))
      !
!$OMP PARALLEL DEFAULT(NONE) &
!$OMP          PRIVATE (size_guess, ithread) &
!$OMP          SHARED (product_matrix, multrec, &
!$OMP                  keep_sparsity, filter_eps, &
!$OMP                  row_max_epss, max_nblocks, &
!$OMP                  matrix_right, nthreads)
      !
      ! Setup product work areas
      !
      ithread = 0
!$    ithread = OMP_GET_THREAD_NUM()
      !
      ! The work arrays have to be setup
      size_guess = product_matrix%m%wms(ithread+1)%datasize ! Should be minimal
      CALL dbcsr_data_ensure_size(product_matrix%m%wms(ithread+1)%data_area, &
                                  size_guess)
      CALL dbcsr_data_set_size_referenced(product_matrix%m%wms(ithread+1)%data_area, &
                                          product_matrix%m%wms(ithread+1)%datasize)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%row_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%col_i, ub=1)
      CALL ensure_array_size(product_matrix%m%wms(ithread+1)%blk_p, ub=1)
      ALLOCATE (multrec(ithread)%p)
      CALL dbcsr_mm_multrec_init(multrec(ithread)%p, &
                                 product=product_matrix%m, &
                                 keep_sparsity=keep_sparsity, &
                                 eps=filter_eps, &
                                 row_max_epss=row_max_epss, &
                                 block_estimate=max_nblocks/nthreads, &
                                 right_row_blk_size=dbcsr_row_block_sizes(matrix_right))
!$OMP END PARALLEL
      !
      ! update capacity of memory-pools
      IF (has_acc) THEN
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_1%pool, &
                                            capacity=left_col_nimages+right_row_nimages)
         CALL dbcsr_mempool_ensure_capacity(memtype_abpanel_2%pool, &
                                            capacity=left_col_nimages+right_row_nimages)
      ENDIF
      !
      ! Wait refs and max norms (sent in make_buffers)
      CALL mp_waitall(requests)
      ! Delete buffers used for collectives
      DEALLOCATE (right_local_refs_displ_unmerged, left_local_refs_displ_unmerged)
      IF (otf_filtering .AND. use_mpi_filtering) THEN
         DEALLOCATE (right_local_max_norms, left_local_max_norms)
      ENDIF
      DEALLOCATE (right_local_refs_meta_size, left_local_refs_meta_size)
      DEALLOCATE (right_local_refs_data_size, left_local_refs_data_size)
      DEALLOCATE (right_no_empty_images, left_no_empty_images)
      !
      ! Pre-CANNON execution: filtering and resize buffers for communications
      CALL timeset(routineN//"_precannon", handle2)
      !
      ALLOCATE (right_vunmerged(0:min_nimages-1, 0:nsteps_k-1))
      ALLOCATE (left_vunmerged(0:min_nimages-1, 0:nsteps_k-1))
      right_vunmerged = 0; left_vunmerged = 0
      !
      ALLOCATE (right_displ_vunmerged(0:nvirt_k-1))
      ALLOCATE (left_displ_vunmerged(0:nvirt_k-1))
      !
      ALLOCATE (right_nofiltered(right_col_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      ALLOCATE (left_nofiltered(left_row_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      right_nofiltered = .FALSE.; left_nofiltered = .FALSE.
      ALLOCATE (right_vmerged(right_col_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      ALLOCATE (left_vmerged(left_row_nimages, 0:min_nimages-1, 0:nsteps_k-1))
      !
      ALLOCATE (right_recv_filtered(right_col_nimages, 0:min_nimages-1))
      ALLOCATE (left_recv_filtered(left_row_nimages, 0:min_nimages-1))
      !
      right_max_data_size = 0; right_max_meta_size = 0
      left_max_data_size = 0; left_max_meta_size = 0
      !
      ALLOCATE (left_max_meta_size_merged(0:min_nimages-1))
      left_max_meta_size_merged = 0
      ALLOCATE (right_max_meta_size_merged(0:min_nimages-1))
      right_max_meta_size_merged = 0
      !
!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED (nsteps_k,min_nimages,right_col_nimages,imgdist_right,&
!$OMP         otf_filtering,use_mpi_filtering,right_nprows,left_npcols,&
!$OMP         right_myfirstvrow,right_myfirstvcol,right_max_norms,&
!$OMP         left_row_nimages,imgdist_left,left_myfirstvrow,left_myfirstvcol,&
!$OMP         left_max_norms,min_row_max_epss,right_row_nimages,&
!$OMP         left_recv_filtered,right_recv_filtered,right_refs_meta_size,&
!$OMP         left_refs_meta_size,right_refs_data_size,left_refs_data_size,&
!$OMP         right_max_data_size,left_max_data_size,left_col_nimages,&
!$OMP         right_max_meta_size,left_max_meta_size,do_diagonal,&
!$OMP         right_data_size,left_data_size,right_meta_size,left_meta_size,&
!$OMP         size_index_unmerged,left_max_meta_size_merged,right_max_meta_size_merged,&
!$OMP         right_vunmerged,left_vunmerged,right_nofiltered,left_nofiltered,myprow,&
!$OMP         right_vmerged,left_vmerged,left_displ_vunmerged,right_displ_vunmerged,&
!$OMP         right_no_empty_images_displ,left_no_empty_images_displ,do_scatter,&
!$OMP         local_meta_size_scatter,v_k_scatter,nimages_scatter) &
!$OMP PRIVATE (metronome,v_ki,v_ci,v_ri,recv_vrow,recv_vcol,right_vrow,&
!$OMP          right_vcol,left_vrow,left_vcol,meta_size_image,iproc,v_k_loc,&
!$OMP          do_k_scatter)
      !
!$OMP DO
      DO iproc = 0, right_nprows-1
         recv_vrow = iproc*right_row_nimages
         right_displ_vunmerged(recv_vrow) = right_no_empty_images_displ(iproc+1)
         DO v_ki = recv_vrow+1, recv_vrow+right_row_nimages-1
            right_displ_vunmerged(v_ki) = right_displ_vunmerged(v_ki-1)
            DO v_ci = LBOUND(right_refs_meta_size, 1), UBOUND(right_refs_meta_size, 1)
               IF (right_refs_meta_size(v_ci, v_ki-1) .NE. 0) THEN
                  right_displ_vunmerged(v_ki) = right_displ_vunmerged(v_ki)+1
               ENDIF
            ENDDO
         ENDDO
      ENDDO
!$OMP END DO
!$OMP DO
      DO iproc = 0, left_npcols-1
         recv_vcol = iproc*left_col_nimages
         left_displ_vunmerged(recv_vcol) = left_no_empty_images_displ(iproc+1)
         DO v_ki = recv_vcol+1, recv_vcol+left_col_nimages-1
            left_displ_vunmerged(v_ki) = left_displ_vunmerged(v_ki-1)
            DO v_ri = LBOUND(left_refs_meta_size, 1), UBOUND(left_refs_meta_size, 1)
               IF (left_refs_meta_size(v_ri, v_ki-1) .NE. 0) THEN
                  left_displ_vunmerged(v_ki) = left_displ_vunmerged(v_ki)+1
               ENDIF
            ENDDO
         ENDDO
      ENDDO
!$OMP END DO
      !
      DO metronome = 0, nsteps_k-1
         !
!$OMP MASTER
         right_data_size = 0
         left_data_size = 0
         right_meta_size = 0
         left_meta_size = 0
!$OMP END MASTER
!$OMP BARRIER
         !
         ! Take first cluster global virtual coordinates
         CALL image_calculator(imgdist_right, &
                               vprow=right_vrow, vpcol=right_vcol, &
                               myvprow=right_myfirstvrow, &
                               myvpcol=right_myfirstvcol, &
                               vprow_shift=metronome*min_nimages, &
                               shifting='R')
         CALL image_calculator(imgdist_left, &
                               vprow=left_vrow, vpcol=left_vcol, &
                               myvprow=left_myfirstvrow, &
                               myvpcol=left_myfirstvcol, &
                               vpcol_shift=metronome*min_nimages, &
                               shifting='L')
         !
         v_k_loc = 0
!$OMP DO REDUCTION(+:right_data_size,left_data_size,right_meta_size,left_meta_size) &
!$OMP    schedule(dynamic)
         DO v_ki = 0, min_nimages-1
            !
            recv_vrow = v_ki+right_vrow
            recv_vcol = v_ki+left_vcol
            !
            DO v_ci = right_vcol, right_col_nimages+right_vcol-1
               IF (right_refs_meta_size(v_ci, recv_vrow) .NE. 0) THEN
                  right_vunmerged(v_ki, metronome) = right_vunmerged(v_ki, metronome)+1
                  right_vmerged(right_vunmerged(v_ki, metronome), v_ki, metronome) = v_ci
               ENDIF
            ENDDO
            DO v_ri = left_vrow, left_row_nimages+left_vrow-1
               IF (left_refs_meta_size(v_ri, recv_vcol) .NE. 0) THEN
                  left_vunmerged(v_ki, metronome) = left_vunmerged(v_ki, metronome)+1
                  left_vmerged(left_vunmerged(v_ki, metronome), v_ki, metronome) = v_ri
               ENDIF
            ENDDO
            !
            do_k_scatter = .FALSE.
            IF ((do_diagonal .OR. do_scatter) .AND. &
                local_meta_size_scatter .GT. 1 .AND. &
                recv_vcol/min_nimages .EQ. myprow) THEN
               CALL ordered_search(v_k_scatter, v_ki+1, v_k_loc, do_k_scatter, v_k_loc+1, nimages_scatter)
            ENDIF
            !
            ! Max norms evaluation for filtering
            !
            ! By default all images are not communicated
            left_recv_filtered(:, v_ki) = 0
            right_recv_filtered(:, v_ki) = 0
            !
            ! multiplication of norms
            DO v_ri = 1, left_vunmerged(v_ki, metronome)
               DO v_ci = 1, right_vunmerged(v_ki, metronome)
                  IF (do_diagonal .AND. &
                      right_vmerged(v_ci, v_ki, metronome) .EQ. recv_vrow .AND. &
                      .NOT. do_k_scatter) CYCLE
                  IF (otf_filtering .AND. use_mpi_filtering) THEN
                     IF (left_max_norms(left_displ_vunmerged(recv_vcol)+v_ri)* &
                         right_max_norms(right_displ_vunmerged(recv_vrow)+v_ci) .LT. &
                         min_row_max_epss) CYCLE
                  ENDIF
                  left_recv_filtered(v_ri, v_ki) = left_recv_filtered(v_ri, v_ki)+1
                  right_recv_filtered(v_ci, v_ki) = right_recv_filtered(v_ci, v_ki)+1
               ENDDO
            ENDDO
            !
            ! resize buffers taking in account filtering
            meta_size_image = 0
            DO v_ci = 1, right_vunmerged(v_ki, metronome)
               IF (right_recv_filtered(v_ci, v_ki) .NE. 0) THEN
                  right_nofiltered(v_ci, v_ki, metronome) = .TRUE.
                  right_data_size = right_data_size+ &
                                    right_refs_data_size(right_displ_vunmerged(recv_vrow)+v_ci)
                  right_meta_size = right_meta_size+ &
                                    right_refs_meta_size(right_vmerged(v_ci, v_ki, metronome), recv_vrow)
                  meta_size_image = meta_size_image+ &
                                    right_refs_meta_size(right_vmerged(v_ci, v_ki, metronome), recv_vrow)- &
                                    dbcsr_slot_nblks
               ENDIF
            ENDDO
            right_max_meta_size_merged(v_ki) = MAX(right_max_meta_size_merged(v_ki), meta_size_image)
            !
            meta_size_image = 0
            DO v_ri = 1, left_vunmerged(v_ki, metronome)
               IF (left_recv_filtered(v_ri, v_ki) .NE. 0 .OR. do_k_scatter) THEN
                  left_nofiltered(v_ri, v_ki, metronome) = .TRUE.
                  left_data_size = left_data_size+ &
                                   left_refs_data_size(left_displ_vunmerged(recv_vcol)+v_ri)
                  left_meta_size = left_meta_size+ &
                                   left_refs_meta_size(left_vmerged(v_ri, v_ki, metronome), recv_vcol)
                  meta_size_image = meta_size_image+ &
                                    left_refs_meta_size(left_vmerged(v_ri, v_ki, metronome), recv_vcol)- &
                                    size_index_unmerged
               ENDIF
            ENDDO
            left_max_meta_size_merged(v_ki) = MAX(left_max_meta_size_merged(v_ki), meta_size_image)
         ENDDO
!$OMP END DO
!$OMP MASTER
         right_max_data_size = MAX(right_max_data_size, right_data_size)
         left_max_data_size = MAX(left_max_data_size, left_data_size)
         right_max_meta_size = MAX(right_max_meta_size, right_meta_size)
         left_max_meta_size = MAX(left_max_meta_size, left_meta_size)
!$OMP END MASTER
!$OMP BARRIER
      ENDDO
!$OMP END PARALLEL
      IF (otf_filtering .AND. use_mpi_filtering) THEN
         DEALLOCATE (right_max_norms, left_max_norms)
      ENDIF
      DEALLOCATE (left_recv_filtered, right_recv_filtered)
      !
      CALL timestop(handle2)
      !
      ! Preallocate norms arrays
      IF (otf_filtering) THEN
         ALLOCATE (left_norms(MAXVAL(left_max_meta_size_merged)/3, min_nimages))
         ALLOCATE (right_norms(MAX(MAXVAL(right_max_meta_size_merged), &
                                   local_meta_size_scatter)/3, min_nimages))
      ENDIF
      left_max_meta_size_merged = left_max_meta_size_merged+size_index_merged
      right_max_meta_size_merged = right_max_meta_size_merged+dbcsr_num_slots
      !
      ! Prepare buffers for computation
      IF (nsteps_k .GT. 1) THEN
         ! Right:
         CALL buffer_init(buffers_1%right, data_type, &
                          right_max_data_size, &
                          right_max_meta_size, &
                          requests_size=(min_nimages*right_col_nimages), &
                          data_memory_type=memtype_abpanel_1)
         CALL setup_buffer_matrices_images(buffers_1%right%buffer, min_nimages, &
                                           imgdist_right, matrix_right, &
                                           right_max_meta_size_merged, &
                                           buffers_1%right%data)
         ! Left:
         CALL buffer_init(buffers_1%left, data_type, &
                          left_max_data_size, &
                          left_max_meta_size, &
                          requests_size=(left_row_nimages*min_nimages), &
                          data_memory_type=memtype_abpanel_1)
         CALL setup_buffer_matrices_images(buffers_1%left%buffer, min_nimages, &
                                           imgdist_left, matrix_left, &
                                           left_max_meta_size_merged, &
                                           buffers_1%left%data)
         !
         IF (has_acc) THEN
            CALL dbcsr_data_init(buffers_1%trs_stackbuf)
            CALL dbcsr_data_new(buffers_1%trs_stackbuf, &
                                data_type=dbcsr_type_int_4, data_size=1000, &
                                memory_type=memtype_trsbuffer_1)
         ENDIF
         !
         buffers_calc => buffers_1
      ENDIF
      !
      ! Prepare buffers for communication
      ! Right:
      CALL buffer_init(buffers_2%right, data_type, &
                       right_max_data_size, &
                       right_max_meta_size, &
                       requests_size=(min_nimages*right_col_nimages), &
                       data_memory_type=memtype_abpanel_2)
      CALL setup_buffer_matrices_images(buffers_2%right%buffer, min_nimages, &
                                        imgdist_right, matrix_right, &
                                        right_max_meta_size_merged, &
                                        buffers_2%right%data)
      ! Left:
      CALL buffer_init(buffers_2%left, data_type, &
                       left_max_data_size, &
                       left_max_meta_size, &
                       requests_size=(left_row_nimages*min_nimages), &
                       data_memory_type=memtype_abpanel_2)
      CALL setup_buffer_matrices_images(buffers_2%left%buffer, min_nimages, &
                                        imgdist_left, matrix_left, &
                                        left_max_meta_size_merged, &
                                        buffers_2%left%data)
      !
      IF (has_acc) THEN
         CALL dbcsr_data_init(buffers_2%trs_stackbuf)
         CALL dbcsr_data_new(buffers_2%trs_stackbuf, &
                             data_type=dbcsr_type_int_4, data_size=1000, &
                             memory_type=memtype_trsbuffer_2)
      ENDIF
      !
      buffers_comm => buffers_2
      !
      DEALLOCATE (right_max_meta_size_merged, left_max_meta_size_merged)
      !
      ! Setup the receive data pointers
      CALL dbcsr_data_init(data_get)
      CALL dbcsr_data_new(data_get, data_type)
      !
      ! These values for meta data are used for global values
      right_global_indices(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
         (/ &
         matrix_right%m%nblkrows_total, &
         matrix_right%m%nblkcols_total, &
         matrix_right%m%nfullrows_total, &
         matrix_right%m%nfullcols_total, &
         0, 0, &
         matrix_right%m%nfullrows_local, &
         matrix_right%m%nfullcols_local/)
      left_global_indices(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
         (/ &
         matrix_left%m%nblkrows_total, &
         matrix_left%m%nblkcols_total, &
         matrix_left%m%nfullrows_total, &
         matrix_left%m%nfullcols_total, &
         0, 0, &
         matrix_left%m%nfullrows_local, &
         matrix_left%m%nfullcols_local/)
      !
      ALLOCATE (right_get_requests_images(min_nimages))
      ALLOCATE (left_get_requests_images(min_nimages))
      ALLOCATE (common_unmerged(min_nimages))
      ALLOCATE (indices_common_unmerged(min_nimages))
      !
      ! Wait for diagonal images data
      IF (do_bcast) THEN
         CALL timeset(routineN//"_diag", handle2)
         CALL mp_waitall(requests_diag)
         CALL timestop(handle2)
      ENDIF
      !
      ! Here is the main loop
      !
      CALL timeset(routineN//"_loop", handle2)
      !
      right_meta_offset_diag = 0
      right_data_offset_diag = 0
      left_meta_offset_diag = 0
      left_data_offset_diag = 0
      !
      grouped_k_index: DO metronome = 0, nsteps_k
         !
         IF (debug_mod) WRITE (*, '(1X,A,3(1X,A,1X,I5))') routineN, &
            "step", metronome, &
            "first k", metronome*min_nimages, &
            "last k", (metronome+1)*min_nimages-1
         !
         ! Matrix transfer. Transfer in all but the last loop
         ! iteration.
         xfer: IF (metronome .LT. nsteps_k) THEN
            !
            IF (has_acc) THEN
               CALL timeset(routineN//"_acc_sync", handle3)
               CALL acc_event_synchronize(buffers_comm%right%data%d%acc_ready)
               CALL acc_event_synchronize(buffers_comm%left%data%d%acc_ready)
               CALL timestop(handle3)
            ENDIF
            !
            right_meta_offset = 0
            right_meta_size = 0
            right_data_offset = 0
            right_data_size = 0
            buffers_comm%right%nrequests = 0
            !
            ! Take first cluster global virtual coordinates
            CALL image_calculator(imgdist_right, &
                                  vprow=buffers_comm%right%vprow, &
                                  vpcol=buffers_comm%right%vpcol, &
                                  myvprow=right_myfirstvrow, &
                                  myvpcol=right_myfirstvcol, &
                                  vprow_shift=metronome*min_nimages, &
                                  shifting='R')
            !
            left_meta_offset = 0
            left_meta_size = 0
            left_data_offset = 0
            left_data_size = 0
            buffers_comm%left%nrequests = 0
            !
            ! Take first cluster global virtual coordinates
            CALL image_calculator(imgdist_left, &
                                  vprow=buffers_comm%left%vprow, &
                                  vpcol=buffers_comm%left%vpcol, &
                                  myvprow=left_myfirstvrow, &
                                  myvpcol=left_myfirstvcol, &
                                  vpcol_shift=metronome*min_nimages, &
                                  shifting='L')
            !
            right_set_displ = .TRUE.
            left_set_displ = .TRUE.
            right_v_i = 0
            left_v_i = 0
            buffers_comm%right%offset(:, :) = 0
            buffers_comm%left%offset(:, :) = 0
            buffers_comm%right%get_requests(:, :) = mp_request_null
            buffers_comm%left%get_requests(:, :) = mp_request_null
            !
            DO v_ki = 0, min_nimages-1
               ! Right
               CALL cannon_transfer_rma(v_ki, buffers_comm%right%vprow, &
                                        right_row_nimages, right_col_nimages, &
                                        right_vunmerged(:, metronome), right_vmerged(:, :, metronome), &
                                        right_displ_vunmerged, &
                                        right_v_i, right_nofiltered(:, :, metronome), right_set_displ, &
                                        recv_vrow_displ, right_refs_displ_unmerged, &
                                        buffers_comm%right, right_refs_meta_size, right_refs_data_size, &
                                        right_meta_size, right_data_size, &
                                        min_nimages, right_meta_offset, right_data_offset, &
                                        buffers_orig%right%meta_win, buffers_orig%right%data_win, &
                                        right_myprow, metronome .GT. 0, &
                                        data_get, data_type_byte, buffers_orig%right, &
                                        right_meta_offset_diag, right_data_offset_diag)
               ! Left
               CALL cannon_transfer_rma(v_ki, buffers_comm%left%vpcol, &
                                        left_col_nimages, left_row_nimages, &
                                        left_vunmerged(:, metronome), left_vmerged(:, :, metronome), &
                                        left_displ_vunmerged, &
                                        left_v_i, left_nofiltered(:, :, metronome), left_set_displ, &
                                        recv_vcol_displ, left_refs_displ_unmerged, &
                                        buffers_comm%left, left_refs_meta_size, left_refs_data_size, &
                                        left_meta_size, left_data_size, &
                                        min_nimages, left_meta_offset, left_data_offset, &
                                        buffers_orig%left%meta_win, buffers_orig%left%data_win, &
                                        left_mypcol, metronome .GT. 0, &
                                        data_get, data_type_byte, buffers_orig%left, &
                                        left_meta_offset_diag, left_data_offset_diag)
            ENDDO
            ! Set the referenced sizes to the actual data moved via MPI
            CALL dbcsr_data_set_size_referenced(buffers_comm%right%data, &
                                                buffers_comm%right%offset(idata_size, right_v_i+1))
            CALL dbcsr_data_set_size_referenced(buffers_comm%left%data, &
                                                buffers_comm%left%offset(idata_size, left_v_i+1))
         ENDIF xfer
         !
         ! Wait data, merge and do the multiplications.
         ! Exclude the first interation
         wait_merge_calc: IF (metronome .GT. 0) THEN
            IF (debug_mod) WRITE (*, '(1X,A)') routineN//" waiting for right and left"
            !
            buffers_calc%right%buffer%mats(:)%m%nblks = 0
            buffers_calc%left%buffer%mats(:)%m%nblks = 0
            !
            common_unmerged(:) = .FALSE.
            !
            IF (metronome == nsteps_k) THEN
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (multrec) &
!$OMP PRIVATE (ithread)
               ithread = 0
!$             ithread = omp_get_thread_num()
               !
               CALL dbcsr_mm_multrec_phaseout(multrec(ithread)%p)
!$OMP END PARALLEL
            ENDIF
            !
            right_nrequests = buffers_calc%right%nrequests
            left_nrequests = buffers_calc%left%nrequests
            IF (min_nimages .EQ. 1 .OR. has_acc) THEN
               IF (right_nrequests .GT. 0) THEN
                  CALL mp_waitall(buffers_calc%right%get_requests(:, 1:right_nrequests))
                  IF (has_acc) CALL dbcsr_data_host2dev(buffers_calc%right%data)
                  CALL merge_calc(buffers_calc%right, &
                                  buffers_calc%left, &
                                  multrec, do_rec, flop, &
                                  imgdist_right, .FALSE., &
                                  1, min_nimages, &
                                  min_nimages, right_col_nimages, &
                                  right_global_indices, right_vunmerged(:, metronome-1), &
                                  right_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                  right_norms, left_norms, &
                                  common_unmerged, indices_common_unmerged, has_acc)
               ENDIF
               IF (left_nrequests .GT. 0) THEN
                  CALL mp_waitall(buffers_calc%left%get_requests(:, 1:left_nrequests))
                  IF (has_acc) CALL dbcsr_data_host2dev(buffers_calc%left%data)
                  CALL merge_calc(buffers_calc%right, &
                                  buffers_calc%left, &
                                  multrec, do_rec, flop, &
                                  imgdist_left, .TRUE., &
                                  1, min_nimages, &
                                  left_row_nimages, min_nimages, &
                                  left_global_indices, left_vunmerged(:, metronome-1), &
                                  left_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                  right_norms, left_norms, &
                                  common_unmerged, indices_common_unmerged, &
                                  has_acc, nthreads, buffers_calc%trs_stackbuf)
               ENDIF
            ELSE
               !
               ! Use async communication-computation over images
               CALL timeset("mp_waitall_2", handle3)
               !
               ! Counting the requests per each unmerged image
               right_get_requests_images(:) = 0
               DO iget = 1, right_nrequests
                  right_get_requests_images(buffers_calc%right%get_requests_map(:, iget)) = &
                     right_get_requests_images(buffers_calc%right%get_requests_map(:, iget))+1
               ENDDO
               left_get_requests_images(:) = 0
               DO iget = 1, left_nrequests
                  left_get_requests_images(buffers_calc%left%get_requests_map(:, iget)) = &
                     left_get_requests_images(buffers_calc%left%get_requests_map(:, iget))+1
               ENDDO
               !
               DO WHILE (right_nrequests .GT. 0 .OR. &
                         left_nrequests .GT. 0)
                  IF (right_nrequests .GT. 0) THEN
                     CALL wait_merge_calc_low(buffers_calc%right, &
                                              buffers_calc%left, &
                                              right_nrequests, &
                                              multrec, do_rec, flop, &
                                              right_get_requests_images, &
                                              imgdist_right, .FALSE., min_nimages, right_col_nimages, &
                                              right_global_indices, right_vunmerged(:, metronome-1), &
                                              right_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                              right_norms, left_norms, &
                                              common_unmerged, indices_common_unmerged)
                  ENDIF
                  IF (left_nrequests .GT. 0) THEN
                     CALL wait_merge_calc_low(buffers_calc%right, &
                                              buffers_calc%left, &
                                              left_nrequests, &
                                              multrec, do_rec, flop, &
                                              left_get_requests_images, &
                                              imgdist_left, .TRUE., left_row_nimages, min_nimages, &
                                              left_global_indices, left_vunmerged(:, metronome-1), &
                                              left_vmerged(:, :, metronome-1), otf_filtering, data_type, &
                                              right_norms, left_norms, &
                                              common_unmerged, indices_common_unmerged, &
                                              nthreads)
                  ENDIF
               ENDDO
               !
               CALL timestop(handle3)
            ENDIF
            !
            IF (debug_mod) THEN
               DO v_ki = 1, min_nimages
                  IF (buffers_calc%right%buffer%mats(v_ki)%m%nblks .NE. 0) &
                     CALL dbcsr_print(buffers_calc%right%buffer%mats(v_ki), &
                                      nodata=.TRUE.)
                  IF (buffers_calc%left%buffer%mats(v_ki)%m%nblks .NE. 0) &
                     CALL dbcsr_print(buffers_calc%left%buffer%mats(v_ki), &
                                      nodata=.TRUE.)
               ENDDO
            ENDIF
         ENDIF wait_merge_calc
         !
         ! Multiply scatter data
         IF (do_scatter .AND. local_meta_size_scatter .GT. 1 .AND. metronome .GT. 0 .AND. &
             buffers_calc%left%nrequests .GT. 0 .AND. &
             buffers_calc%left%vpcol/min_nimages .EQ. myprow) THEN
            IF (otf_filtering) THEN
               CALL calculate_image_norms(buffer_scatter, &
                                          data_type, right_norms, &
                                          uf=1, ul=nimages_scatter)
            ENDIF
            CALL timeset("cannon_multiply_low_multrec", handle3)
            !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (buffer_scatter,buffers_calc, &
!$OMP         left_norms,right_norms,nimages_scatter,&
!$OMP         multrec_scatter,v_k_scatter) &
!$OMP PRIVATE (ithread,v_ki) &
!$OMP REDUCTION (+: flop)
            ithread = 0
!$          ithread = omp_get_thread_num()
            !
            DO v_ki = 1, nimages_scatter
               IF (buffers_calc%left%buffer%mats(v_k_scatter(v_ki))%m%nblks .EQ. 0) CYCLE
               CALL dbcsr_mm_multrec_multiply(multrec_scatter(ithread)%p, &
                                              left=buffers_calc%left%buffer%mats(v_k_scatter(v_ki))%m, &
                                              right=buffer_scatter%mats(v_ki)%m, &
                                              flop=flop, &
                                              a_norms=left_norms(:, v_k_scatter(v_ki)), &
                                              b_norms=right_norms(:, v_ki), &
                                              do_rec=.FALSE.)
            ENDDO
!$OMP END PARALLEL
            CALL timestop(handle3)
         ENDIF
         !
         IF (metronome .LT. nsteps_k) THEN
            buffers => buffers_comm
            buffers_comm => buffers_calc
            buffers_calc => buffers
         ENDIF
         !
      ENDDO grouped_k_index
      !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (multrec, multrec_scatter, do_scatter, &
!$OMP         local_meta_size_scatter) &
!$OMP PRIVATE (ithread)
      ithread = 0
!$    ithread = omp_get_thread_num()
      CALL dbcsr_mm_multrec_finalize(multrec(ithread)%p)
      DEALLOCATE (multrec(ithread)%p)
      IF (do_scatter .AND. local_meta_size_scatter .GT. 1) THEN
         CALL dbcsr_mm_multrec_finalize(multrec_scatter(ithread)%p)
         DEALLOCATE (multrec_scatter(ithread)%p)
      ENDIF
!$OMP END PARALLEL
      !
      CALL timestop(handle2)
      !
      CALL m_memory(mem)
      max_memory = MAX(max_memory, REAL(mem))
      !
      IF (debug_mod) THEN
         v_ki = 0
         DO blk = 1, SIZE(product_matrix%m%blk_p)
            v_ki = MAX(v_ki, ABS(product_matrix%m%blk_p(blk)))
         ENDDO
         WRITE (*, *) routineN//" Actual final size", &
            LOG(REAL(dbcsr_data_get_size(product_matrix%m%data_area)))/LOG(10.0), &
            LOG(REAL(v_ki))/LOG(10.0)
      ENDIF
      !
      IF (has_acc) THEN
         CALL dbcsr_data_release(buffers_1%trs_stackbuf)
         CALL dbcsr_data_release(buffers_2%trs_stackbuf)
      ENDIF
      !
      DEALLOCATE (right_norms, left_norms, row_max_epss)
      !
      CALL dbcsr_data_clear_pointer(data_get)
      CALL dbcsr_data_release(data_get)
      !
      ! clean-up of communication buffers
      IF (nsteps_k .GT. 1) THEN
         CALL dbcsr_destroy_array(buffers_1%right%buffer)
         CALL dbcsr_destroy_array(buffers_1%left%buffer)
      ENDIF
      !
      DEALLOCATE (right_get_requests_images, left_get_requests_images)
      DEALLOCATE (common_unmerged, indices_common_unmerged)
      DEALLOCATE (right_refs_meta_size, left_refs_meta_size)
      DEALLOCATE (right_refs_data_size, left_refs_data_size)
      DEALLOCATE (right_refs_displ_unmerged, left_refs_displ_unmerged)
      DEALLOCATE (multrec)
      DEALLOCATE (right_vunmerged, left_vunmerged)
      DEALLOCATE (right_nofiltered, left_nofiltered)
      DEALLOCATE (right_vmerged, left_vmerged)
      DEALLOCATE (right_displ_vunmerged, left_displ_vunmerged)
      DEALLOCATE (right_no_empty_images_displ, left_no_empty_images_displ)
      CALL dbcsr_destroy_array(buffers_2%right%buffer)
      CALL dbcsr_destroy_array(buffers_2%left%buffer)
      !
      CALL dbcsr_finalize(product_matrix)
      !
      ! clean-up wins
      CALL mp_win_unlock_all(buffers_orig%right%data_win)
      CALL mp_win_unlock_all(buffers_orig%right%meta_win)
      CALL mp_win_unlock_all(buffers_orig%left%data_win)
      CALL mp_win_unlock_all(buffers_orig%left%meta_win)
      !
      ! Gather data
      IF (do_scatter .OR. do_diagonal) THEN
         CALL timeset("cannon_multiply_low_rma_gather", handle2)
         local_meta_size_gather = 1
         local_data_size_gather = 1
         IF (local_meta_size_scatter .GT. 1) THEN
            DEALLOCATE (v_k_scatter)
            IF (do_scatter) THEN
               DO v_ki = 1, nimages_scatter
                  NULLIFY (buffer_scatter%mats(v_ki)%m%index)
               ENDDO
               CALL dbcsr_destroy_array(buffer_scatter)
               DEALLOCATE (multrec_scatter)
               CALL dbcsr_finalize(product_matrix_scatter)
               IF (product_matrix_scatter%m%nblks .NE. 0) THEN
                  local_meta_size_gather = product_matrix_scatter%m%index(dbcsr_slot_size)
                  local_data_size_gather = product_matrix_scatter%m%index(dbcsr_slot_nze)
               ENDIF
            ENDIF
         ENDIF
         IF (do_diagonal) THEN
            ! Gather sizes
            ALLOCATE (meta_size_gather(0:right_npcols-1), &
                      data_size_gather(0:right_npcols-1), &
                      meta_displ_gather(0:right_npcols-1), &
                      data_displ_gather(0:right_npcols-1))
         ENDIF
         ! Gather sizes
         CALL mp_gather(local_meta_size_gather, meta_size_gather, &
                        right_myprow, grp_gather)
         CALL mp_gather(local_data_size_gather, data_size_gather, &
                        right_myprow, grp_gather)
         IF (do_diagonal) THEN
            meta_displ_gather(0) = 0; data_displ_gather(0) = 0
            DO iproc = 0, right_npcols-2
               meta_displ_gather(iproc+1) = meta_displ_gather(iproc)+ &
                                            meta_size_gather(iproc)
               data_displ_gather(iproc+1) = data_displ_gather(iproc)+ &
                                            data_size_gather(iproc)
            ENDDO
            ! Prepare result matrix
            ! Reuse scatter buffers
            CALL ensure_array_size(local_meta_scatter, &
                                   ub=meta_displ_gather(right_npcols-1)+meta_size_gather(right_npcols-1), &
                                   nocopy=.TRUE., memory_type=memtype_mpi_buffer)
            CALL dbcsr_data_ensure_size(local_data_scatter, &
                                        data_displ_gather(right_npcols-1)+data_size_gather(right_npcols-1), &
                                        nocopy=.TRUE.)
            CALL dbcsr_data_set_size_referenced(local_data_scatter, &
                                                data_displ_gather(right_npcols-1)+data_size_gather(right_npcols-1))
            CALL dbcsr_create(product_matrix_scatter, &
                              template=product_matrix, &
                              name=TRIM("Gather buffer of "//TRIM(product_matrix%m%name)), &
                              make_index=.NOT. keep_sparsity, &
                              data_buffer=local_data_scatter)
            IF (keep_sparsity) THEN
               product_matrix_scatter%m%index => local_meta_product_scatter
               CALL dbcsr_repoint_index(product_matrix_scatter%m)
            ENDIF
            CALL dbcsr_reset_locals(product_matrix_scatter)
            product_matrix_scatter%m%nblks = 0
            product_matrix_scatter%m%nze = 0
            product_matrix_scatter%m%row_p(:) = 0
            product_matrix_scatter%m%valid = .FALSE.
         ENDIF
         CALL mp_gatherv(product_matrix_scatter%m%index(1:local_meta_size_gather), &
                         local_meta_scatter, meta_size_gather, meta_displ_gather, &
                         right_myprow, grp_gather)
         IF (do_diagonal) THEN
            ! Receiver
            ! Use a dummy buffer for the sender of the root proc to avoid overlap
            ! between send and receive buffers
            CALL dbcsr_gatherv_any(buffers_orig%right%data_diag, &
                                   local_data_size_gather, &
                                   product_matrix_scatter%m%data_area, &
                                   data_size_gather, data_displ_gather, &
                                   right_myprow, grp_gather)
         ENDIF
         IF (do_scatter) THEN
            ! Sender
            CALL dbcsr_gatherv_any(product_matrix_scatter%m%data_area, &
                                   local_data_size_gather, &
                                   product_matrix_scatter%m%data_area, &
                                   data_size_gather, data_displ_gather, &
                                   right_myprow, grp_gather)
         ENDIF
         IF (do_diagonal) THEN
            ! Merge gathered images
            DO iproc = 0, right_npcols-1
               IF (meta_size_gather(iproc) .GT. 1) THEN
                  product_matrix_scatter%m%index(dbcsr_slot_nblks) = &
                     product_matrix_scatter%m%index(dbcsr_slot_nblks)+ &
                     local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_nblks)
                  product_matrix_scatter%m%index(dbcsr_slot_nze) = &
                     product_matrix_scatter%m%index(dbcsr_slot_nze)+ &
                     local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_nze)
                  product_matrix_scatter%m%row_p(:) = &
                     product_matrix_scatter%m%row_p(:)+ &
                     local_meta_scatter(meta_displ_gather(iproc)+ &
                                        local_meta_scatter(meta_displ_gather(iproc)+ &
                                                           dbcsr_slot_row_p): &
                                        meta_displ_gather(iproc)+ &
                                        local_meta_scatter(meta_displ_gather(iproc)+ &
                                                           dbcsr_slot_row_p+1))
               ENDIF
            ENDDO
            product_matrix_scatter%m%index(dbcsr_slot_size) = &
               product_matrix_scatter%m%index(dbcsr_slot_size)+ &
               product_matrix_scatter%m%index(dbcsr_slot_nblks)*2
            product_matrix_scatter%m%index(dbcsr_slot_col_i+1) = &
               product_matrix_scatter%m%index(dbcsr_slot_col_i)+ &
               product_matrix_scatter%m%index(dbcsr_slot_nblks)-1
            product_matrix_scatter%m%index(dbcsr_slot_blk_p) = &
               product_matrix_scatter%m%index(dbcsr_slot_col_i+1)+1
            product_matrix_scatter%m%index(dbcsr_slot_blk_p+1) = &
               product_matrix_scatter%m%index(dbcsr_slot_blk_p)+ &
               product_matrix_scatter%m%index(dbcsr_slot_nblks)-1
            CALL ensure_array_size(product_matrix_scatter%m%index, &
                                   ub=product_matrix_scatter%m%index(dbcsr_slot_size), &
                                   nocopy=.FALSE., memory_type=memtype_mpi_buffer)
            !
            CALL dbcsr_repoint_index(product_matrix_scatter%m)
            !
            ! Collect col_i and blk_p, row by row
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (product_matrix_scatter,meta_size_gather,&
!$OMP         local_meta_scatter,right_npcols,meta_displ_gather,&
!$OMP         data_displ_gather) &
!$OMP PRIVATE (local_meta_size_gather,iproc,col,v_ki,&
!$OMP          col_i,blk_p,row)
!$OMP DO
            DO row = 0, SIZE(product_matrix_scatter%m%row_p)-2
               local_meta_size_gather = product_matrix_scatter%m%row_p(row+1)+1
               ! Skip empty rows
               IF (local_meta_size_gather .GT. product_matrix_scatter%m%row_p(row+2)) CYCLE
               DO iproc = 0, right_npcols-1
                  IF (meta_size_gather(iproc) .GT. 1) THEN
                     col = meta_displ_gather(iproc)+ &
                           local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_row_p)+row
                     v_ki = local_meta_scatter(col+1)-local_meta_scatter(col)
                     ! Skip empty rows
                     IF (v_ki .LE. 0) CYCLE
                     col_i => local_meta_scatter(meta_displ_gather(iproc)+ &
                                                 local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_col_i): &
                                                 meta_displ_gather(iproc)+ &
                                                 local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_col_i+1))
                     blk_p => local_meta_scatter(meta_displ_gather(iproc)+ &
                                                 local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_blk_p): &
                                                 meta_displ_gather(iproc)+ &
                                                 local_meta_scatter(meta_displ_gather(iproc)+dbcsr_slot_blk_p+1))
                     product_matrix_scatter%m%col_i(local_meta_size_gather:local_meta_size_gather+v_ki-1) = &
                        col_i(local_meta_scatter(col)+1:local_meta_scatter(col+1))
                     product_matrix_scatter%m%blk_p(local_meta_size_gather:local_meta_size_gather+v_ki-1) = &
                        blk_p(local_meta_scatter(col)+1:local_meta_scatter(col+1))+ &
                        data_displ_gather(iproc)
                     local_meta_size_gather = local_meta_size_gather+v_ki
                  ENDIF
               ENDDO
            ENDDO
!$OMP END DO
            CALL dbcsr_finalize(product_matrix_scatter)
!$OMP END PARALLEL
            CALL dbcsr_add(product_matrix, product_matrix_scatter, flop=flop)
            DEALLOCATE (meta_size_gather, data_size_gather, &
                        meta_displ_gather, data_displ_gather)
         ENDIF
         CALL dbcsr_release_locals(product_matrix_scatter)
         IF (keep_sparsity .AND. local_meta_size_scatter .GT. 1) NULLIFY (product_matrix_scatter%m%index)
         CALL dbcsr_release(product_matrix_scatter)
         CALL timestop(handle2)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE cannon_multiply_low_rma

! **************************************************************************************************
!> \brief ...
!> \param v_ki ...
!> \param vpunmerged ...
!> \param unmerged_nimages ...
!> \param merged_nimages ...
!> \param vunmerged ...
!> \param vmerged ...
!> \param displ_vunmerged ...
!> \param v_i ...
!> \param nofiltered ...
!> \param set_displ ...
!> \param recv_vunmerged_displ ...
!> \param refs_displ_unmerged ...
!> \param buffer ...
!> \param refs_meta_size ...
!> \param refs_data_size ...
!> \param meta_size ...
!> \param data_size ...
!> \param min_nimages ...
!> \param meta_offset ...
!> \param data_offset ...
!> \param meta_win ...
!> \param data_win ...
!> \param mypunmerged ...
!> \param save_statistics ...
!> \param data_get ...
!> \param data_type_byte ...
!> \param buffer_orig ...
!> \param meta_offset_diag ...
!> \param data_offset_diag ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE cannon_transfer_rma(v_ki, vpunmerged, &
                                  unmerged_nimages, merged_nimages, &
                                  vunmerged, vmerged, &
                                  displ_vunmerged, &
                                  v_i, nofiltered, set_displ, &
                                  recv_vunmerged_displ, refs_displ_unmerged, &
                                  buffer, refs_meta_size, refs_data_size, &
                                  meta_size, data_size, &
                                  min_nimages, meta_offset, data_offset, &
                                  meta_win, data_win, mypunmerged, save_statistics, &
                                  data_get, data_type_byte, &
                                  buffer_orig, meta_offset_diag, data_offset_diag)
      INTEGER, INTENT(IN)                                :: v_ki, vpunmerged, unmerged_nimages, &
                                                            merged_nimages
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: vunmerged
      INTEGER, DIMENSION(:, 0:), INTENT(IN)              :: vmerged
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: displ_vunmerged
      INTEGER, INTENT(INOUT)                             :: v_i
      LOGICAL, DIMENSION(:, 0:), INTENT(IN)              :: nofiltered
      LOGICAL, INTENT(INOUT)                             :: set_displ
      INTEGER, INTENT(INOUT)                             :: recv_vunmerged_displ
      INTEGER, ALLOCATABLE, DIMENSION(:, :), &
         INTENT(INOUT)                                   :: refs_displ_unmerged
      TYPE(dbcsr_buffer), INTENT(INOUT)                  :: buffer
      INTEGER, ALLOCATABLE, DIMENSION(:, :), INTENT(IN)  :: refs_meta_size
      INTEGER, DIMENSION(:), INTENT(IN)                  :: refs_data_size
      INTEGER, INTENT(INOUT)                             :: meta_size, data_size
      INTEGER, INTENT(IN)                                :: min_nimages
      INTEGER, INTENT(INOUT)                             :: meta_offset, data_offset
      INTEGER, INTENT(IN)                                :: meta_win, data_win, mypunmerged
      LOGICAL, INTENT(IN)                                :: save_statistics
      TYPE(dbcsr_data_obj)                               :: data_get
      INTEGER, INTENT(IN)                                :: data_type_byte
      TYPE(dbcsr_buffer), INTENT(IN)                     :: buffer_orig
      INTEGER, INTENT(INOUT)                             :: meta_offset_diag, data_offset_diag

      INTEGER                                            :: recv_punmerged, recv_vunmerged, ui, &
                                                            ui_next, v_ki_next
      INTEGER, DIMENSION(:), POINTER                     :: meta_get
      INTEGER, DIMENSION(idata_displ:imeta_displ)        :: refs_displ_unmerged_first
      LOGICAL                                            :: is_contiguous

! Left  : merged => row, unmerged => col
! Right : merged => col, unmerged => row

      recv_vunmerged = v_ki+vpunmerged
      recv_punmerged = recv_vunmerged/unmerged_nimages
      refs_displ_unmerged_first(:) = 0
      DO ui = 1, vunmerged(v_ki)
         v_i = v_i+1
         IF (nofiltered(ui, v_ki)) THEN
            buffer%offset(idata_size, v_i+1) = &
               buffer%offset(idata_size, v_i)+ &
               refs_data_size(displ_vunmerged(recv_vunmerged)+ui)
            buffer%offset(imeta_size, v_i+1) = &
               buffer%offset(imeta_size, v_i)+ &
               refs_meta_size(vmerged(ui, v_ki), recv_vunmerged)
            meta_size = meta_size+refs_meta_size(vmerged(ui, v_ki), recv_vunmerged)
            data_size = data_size+refs_data_size(displ_vunmerged(recv_vunmerged)+ui)
            !
            ! Check for diagonal images and do the local copy
            IF (merged_nimages .GT. 1 .AND. vmerged(ui, v_ki) .EQ. recv_vunmerged) THEN
               buffer%nrequests = buffer%nrequests+1
               buffer%get_requests_map(1, buffer%nrequests) = v_ki+1
               buffer%get_requests_map(2, buffer%nrequests) = v_ki+1
               CALL memory_copy(buffer%meta(meta_offset+1:meta_size), &
                                buffer_orig%meta_diag(meta_offset_diag+1: &
                                                      meta_offset_diag+refs_meta_size(vmerged(ui, v_ki), &
                                                                                      recv_vunmerged)), &
                                refs_meta_size(vmerged(ui, v_ki), recv_vunmerged))
               meta_offset = meta_size
               CALL dbcsr_data_set(buffer%data, data_offset+1, &
                                   refs_data_size(displ_vunmerged(recv_vunmerged)+ui), &
                                   buffer_orig%data_diag, data_offset_diag+1)
               data_offset = data_size
            ELSE
               IF (set_displ) THEN
                  recv_vunmerged_displ = recv_vunmerged
                  refs_displ_unmerged(:, recv_vunmerged_displ) = &
                     refs_displ_unmerged(:, recv_vunmerged_displ)+ &
                     refs_displ_unmerged_first(:)
                  refs_displ_unmerged_first(:) = 0
                  set_displ = .FALSE.
                  buffer%get_requests_map(1, buffer%nrequests+1) = v_ki+1
               ENDIF
               !
               ! Check for contiguous data
               is_contiguous = .FALSE.
               ui_next = ui+1
               v_ki_next = v_ki
               DO WHILE (.TRUE.)
                  IF (ui_next .LE. vunmerged(v_ki_next)) THEN
                     IF (nofiltered(ui_next, v_ki_next) .AND. &
                         vmerged(ui_next, v_ki_next) .NE. v_ki_next+vpunmerged) THEN
                        is_contiguous = .TRUE.
                     ENDIF
                     EXIT
                  ELSEIF (v_ki_next .LT. min_nimages-1) THEN
                     v_ki_next = v_ki_next+1
                     IF (recv_punmerged .NE. ((v_ki_next+vpunmerged)/unmerged_nimages)) EXIT
                     ui_next = 1
                  ELSE
                     EXIT
                  ENDIF
               ENDDO
               !
               IF (.NOT. is_contiguous) THEN
                  buffer%nrequests = buffer%nrequests+1
                  buffer%get_requests_map(2, buffer%nrequests) = v_ki+1
                  meta_get => buffer%meta(meta_offset+1:meta_size)
                  CALL mp_rget(meta_get, recv_punmerged, &
                               meta_win, &
                               buffer_orig%meta, &
                               mypunmerged, &
                               refs_displ_unmerged(imeta_displ, recv_vunmerged_displ), &
                               buffer%get_requests(1, buffer%nrequests))
                  meta_offset = meta_size
                  CALL dbcsr_data_set_pointer( &
                     area=data_get, &
                     rsize=data_size-data_offset, &
                     csize=1, &
                     pointee=buffer%data, &
                     source_lb=data_offset+1)
                  CALL dbcsr_rget_any(data_get, recv_punmerged, &
                                      data_win, &
                                      buffer_orig%data, &
                                      mypunmerged, &
                                      refs_displ_unmerged(idata_displ, recv_vunmerged_displ), &
                                      buffer%get_requests(2, buffer%nrequests))
                  IF (save_statistics) THEN
                     CALL count_mpi_statistics(dbcsr_mpi_statistics%data_size(1, :), &
                                               data_size-data_offset, &
                                               dbcsr_mpi_statistics%data_size_breakdown(:, :, 1), data_type_byte)
                     dbcsr_mpi_statistics%nexchanged = dbcsr_mpi_statistics%nexchanged+1
                  ENDIF
                  data_offset = data_size
                  set_displ = .TRUE.
               ENDIF
            ENDIF
         ELSE
            buffer%offset(:, v_i+1) = buffer%offset(:, v_i)
            dbcsr_mpi_statistics%nfiltered = dbcsr_mpi_statistics%nfiltered+1
         ENDIF
         IF (merged_nimages .GT. 1 .AND. vmerged(ui, v_ki) .EQ. recv_vunmerged) THEN
            meta_offset_diag = meta_offset_diag+ &
                               refs_meta_size(vmerged(ui, v_ki), recv_vunmerged)
            data_offset_diag = data_offset_diag+ &
                               refs_data_size(displ_vunmerged(recv_vunmerged)+ui)
         ELSE
            refs_displ_unmerged_first(imeta_displ) = &
               refs_displ_unmerged_first(imeta_displ)+ &
               refs_meta_size(vmerged(ui, v_ki), recv_vunmerged)
            refs_displ_unmerged_first(idata_displ) = &
               refs_displ_unmerged_first(idata_displ)+ &
               refs_data_size(displ_vunmerged(recv_vunmerged)+ui)
         ENDIF
      ENDDO
   END SUBROUTINE cannon_transfer_rma

! **************************************************************************************************
!> \brief ...
!> \param right_buffer ...
!> \param left_buffer ...
!> \param in_nrequests ...
!> \param multrec ...
!> \param do_rec ...
!> \param flop ...
!> \param get_requests_images ...
!> \param imgdist ...
!> \param is_left ...
!> \param nrows_images ...
!> \param ncols_images ...
!> \param global_indices ...
!> \param vunmerged ...
!> \param vmerged ...
!> \param otf_filtering ...
!> \param data_type ...
!> \param right_norms ...
!> \param left_norms ...
!> \param common_unmerged ...
!> \param indices_common_unmerged ...
!> \param nthreads ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE wait_merge_calc_low(right_buffer, left_buffer, &
                                  in_nrequests, &
                                  multrec, do_rec, flop, &
                                  get_requests_images, &
                                  imgdist, is_left, &
                                  nrows_images, ncols_images, &
                                  global_indices, vunmerged, vmerged, &
                                  otf_filtering, data_type, &
                                  right_norms, left_norms, &
                                  common_unmerged, indices_common_unmerged, &
                                  nthreads)
      TYPE(dbcsr_buffer), INTENT(INOUT), TARGET          :: right_buffer, left_buffer
      INTEGER, INTENT(INOUT)                             :: in_nrequests
      TYPE(dbcsr_mm_multrec_type_p), DIMENSION(0:), &
         INTENT(INOUT)                                   :: multrec
      LOGICAL, INTENT(IN)                                :: do_rec
      INTEGER(KIND=int_8), INTENT(INOUT)                 :: flop
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: get_requests_images
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      LOGICAL, INTENT(IN)                                :: is_left
      INTEGER, INTENT(IN)                                :: nrows_images, ncols_images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: global_indices, vunmerged
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: vmerged
      LOGICAL, INTENT(IN)                                :: otf_filtering
      INTEGER, INTENT(IN)                                :: data_type
      REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT)      :: right_norms, left_norms
      LOGICAL, DIMENSION(:), INTENT(INOUT)               :: common_unmerged
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: indices_common_unmerged
      INTEGER, INTENT(IN), OPTIONAL                      :: nthreads

      INTEGER                                            :: iget, nrequests, uf, ul
      TYPE(dbcsr_buffer), POINTER                        :: buffer

      IF (is_left) THEN
         buffer => left_buffer
      ELSE
         buffer => right_buffer
      ENDIF
      nrequests = 0
      DO iget = 1, in_nrequests
         IF (mp_testall(buffer%get_requests(:, iget))) THEN
            get_requests_images(buffer%get_requests_map(:, iget)) = &
               get_requests_images(buffer%get_requests_map(:, iget))-1
            uf = buffer%get_requests_map(1, iget)
            IF (get_requests_images(uf) .NE. 0) uf = uf+1
            ul = buffer%get_requests_map(2, iget)
            IF (get_requests_images(ul) .NE. 0) ul = ul-1
            IF (uf .LE. ul) THEN
               CALL merge_calc(right_buffer, left_buffer, &
                               multrec, do_rec, flop, &
                               imgdist, is_left, &
                               uf, ul, &
                               nrows_images, ncols_images, &
                               global_indices, vunmerged, vmerged, &
                               otf_filtering, data_type, &
                               right_norms, left_norms, &
                               common_unmerged, indices_common_unmerged, &
                               .FALSE., nthreads)
            ENDIF
         ELSE
            nrequests = nrequests+1
            IF (nrequests .NE. iget) THEN
               buffer%get_requests(:, nrequests) = &
                  buffer%get_requests(:, iget)
               buffer%get_requests_map(:, nrequests) = &
                  buffer%get_requests_map(:, iget)
            ENDIF
         ENDIF
      ENDDO
      !
      in_nrequests = nrequests
   END SUBROUTINE wait_merge_calc_low

! **************************************************************************************************
!> \brief ...
!> \param right_buffer ...
!> \param left_buffer ...
!> \param multrec ...
!> \param do_rec ...
!> \param flop ...
!> \param imgdist ...
!> \param is_left ...
!> \param uf ...
!> \param ul ...
!> \param nrows_images ...
!> \param ncols_images ...
!> \param global_indices ...
!> \param vunmerged ...
!> \param vmerged ...
!> \param otf_filtering ...
!> \param data_type ...
!> \param right_norms ...
!> \param left_norms ...
!> \param common_unmerged ...
!> \param indices_common_unmerged ...
!> \param has_acc ...
!> \param nthreads ...
!> \param trs_stackbuf ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE merge_calc(right_buffer, left_buffer, &
                         multrec, do_rec, flop, &
                         imgdist, is_left, &
                         uf, ul, &
                         nrows_images, ncols_images, &
                         global_indices, vunmerged, vmerged, &
                         otf_filtering, data_type, &
                         right_norms, left_norms, &
                         common_unmerged, indices_common_unmerged, &
                         has_acc, nthreads, trs_stackbuf)
      TYPE(dbcsr_buffer), INTENT(INOUT), TARGET          :: right_buffer, left_buffer
      TYPE(dbcsr_mm_multrec_type_p), DIMENSION(0:), &
         INTENT(INOUT)                                   :: multrec
      LOGICAL, INTENT(IN)                                :: do_rec
      INTEGER(KIND=int_8), INTENT(INOUT)                 :: flop
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      LOGICAL, INTENT(IN)                                :: is_left
      INTEGER, INTENT(IN)                                :: uf, ul, nrows_images, ncols_images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: global_indices, vunmerged
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: vmerged
      LOGICAL, INTENT(IN)                                :: otf_filtering
      INTEGER, INTENT(IN)                                :: data_type
      REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT), &
         TARGET                                          :: right_norms, left_norms
      LOGICAL, DIMENSION(:), INTENT(INOUT)               :: common_unmerged
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: indices_common_unmerged
      LOGICAL, INTENT(IN)                                :: has_acc
      INTEGER, INTENT(IN), OPTIONAL                      :: nthreads
      TYPE(dbcsr_data_obj), INTENT(INOUT), OPTIONAL      :: trs_stackbuf

      INTEGER                                            :: handle, ithread, ncommons_unmerged, ui
      REAL(kind=sp), DIMENSION(:, :), POINTER            :: norms
      TYPE(dbcsr_buffer), POINTER                        :: buffer

      IF (is_left) THEN
         buffer => left_buffer
         norms => left_norms
      ELSE
         buffer => right_buffer
         norms => right_norms
      ENDIF
      !
      ! Merge row-images/col-images of right/left matrices
      ! Repoint indices of matrices
      CALL merge_images(buffer%buffer, nrows_images, ncols_images, &
                        buffer%vprow, buffer%vpcol, &
                        buffer%offset, buffer%meta, &
                        uf=uf, ul=ul, &
                        imgdist=imgdist, do_merge_rows=is_left, &
                        global_indices=global_indices, &
                        vunmerged=vunmerged, &
                        vmerged=vmerged, &
                        common_unmerged=common_unmerged, &
                        indices_common_unmerged=indices_common_unmerged, &
                        ncommons_unmerged=ncommons_unmerged, &
                        nthreads=nthreads)
      IF (otf_filtering) THEN
         CALL calculate_image_norms(buffer%buffer, &
                                    data_type, norms, uf=uf, ul=ul)
      ENDIF
      !
      IF (ncommons_unmerged .GT. 0) THEN
         !
         IF (has_acc .AND. PRESENT(trs_stackbuf)) THEN
            CALL acc_transpose_blocks_images(right_buffer%buffer, &
                                             right_buffer%data, &
                                             trs_stackbuf, &
                                             indices_common_unmerged(1:ncommons_unmerged))
         ENDIF
         !
         CALL timeset("cannon_multiply_low_multrec", handle)
         !
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (right_buffer,left_buffer, &
!$OMP         left_norms,right_norms, &
!$OMP         multrec,do_rec, &
!$OMP         ncommons_unmerged,indices_common_unmerged) &
!$OMP PRIVATE (ithread,ui) &
!$OMP REDUCTION (+: flop)
         ithread = 0
!$       ithread = omp_get_thread_num()
         !
         DO ui = 1, ncommons_unmerged
            CALL dbcsr_mm_multrec_multiply(multrec(ithread)%p, &
                                           left=left_buffer%buffer%mats(indices_common_unmerged(ui))%m, &
                                           right=right_buffer%buffer%mats(indices_common_unmerged(ui))%m, &
                                           flop=flop, &
                                           a_norms=left_norms(:, indices_common_unmerged(ui)), &
                                           b_norms=right_norms(:, indices_common_unmerged(ui)), &
                                           do_rec=do_rec)
         ENDDO
!$OMP END PARALLEL
         !
         CALL timestop(handle)
      ENDIF
   END SUBROUTINE merge_calc

! **************************************************************************************************
!> \brief ...
!> \param mpi_statistics ...
!> \param data_size ...
!> \param size_breakdown ...
!> \param element_size_bytes ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE count_mpi_statistics(mpi_statistics, data_size, &
                                   size_breakdown, element_size_bytes)
      REAL, DIMENSION(:), INTENT(INOUT)                  :: mpi_statistics
      INTEGER, INTENT(IN)                                :: data_size
      INTEGER(KIND=int_8), DIMENSION(:, :), &
         INTENT(INOUT), OPTIONAL                         :: size_breakdown
      INTEGER, INTENT(IN), OPTIONAL                      :: element_size_bytes

      INTEGER                                            :: ilimit, nlimits
      INTEGER(KIND=int_8)                                :: data_size_bytes, llimit

      mpi_statistics(1) = mpi_statistics(1)+REAL(data_size)
      mpi_statistics(2) = MIN(mpi_statistics(2), REAL(data_size))
      mpi_statistics(3) = MAX(mpi_statistics(3), REAL(data_size))
      IF (PRESENT(size_breakdown)) THEN
         data_size_bytes = data_size
         ! change in bytes
         IF (PRESENT(element_size_bytes)) data_size_bytes = data_size_bytes*element_size_bytes
         nlimits = SIZE(dbcsr_mpi_size_limits)
         ! check for oversize messages
         IF (data_size_bytes .GT. dbcsr_mpi_size_limits(nlimits)) THEN
            size_breakdown(nlimits+1, 1) = size_breakdown(nlimits+1, 1)+1
            size_breakdown(nlimits+1, 2) = size_breakdown(nlimits+1, 2)+data_size_bytes
            RETURN
         ENDIF
         llimit = 0
         DO ilimit = 1, nlimits
            IF (data_size_bytes .GE. llimit .AND. data_size_bytes .LE. dbcsr_mpi_size_limits(ilimit)) THEN
               size_breakdown(ilimit, 1) = size_breakdown(ilimit, 1)+1
               size_breakdown(ilimit, 2) = size_breakdown(ilimit, 2)+data_size_bytes
               RETURN
            ENDIF
            llimit = dbcsr_mpi_size_limits(ilimit)
         ENDDO
      ENDIF
   END SUBROUTINE count_mpi_statistics

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param source_matrix ...
!> \param index_size ...
!> \param data_size ...
!> \param data_buffer ...
!> \param data_memory_type ...
! **************************************************************************************************
   SUBROUTINE setup_buffer_matrix(matrix, source_matrix, &
                                  index_size, data_size, data_buffer, data_memory_type)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_obj), INTENT(IN)                        :: source_matrix
      INTEGER, INTENT(IN)                                :: index_size
      INTEGER, INTENT(IN), OPTIONAL                      :: data_size
      TYPE(dbcsr_data_obj), INTENT(IN), OPTIONAL         :: data_buffer
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: data_memory_type

      CALL dbcsr_init(matrix)
      CALL dbcsr_create(matrix, &
                        template=source_matrix, &
                        name=TRIM("Buffer of "//TRIM(source_matrix%m%name)), &
                        nze=data_size, &
                        data_buffer=data_buffer, &
                        data_memory_type=data_memory_type, &
                        index_memory_type=memtype_mpi_buffer)
      IF (PRESENT(data_size)) THEN
         CALL dbcsr_data_ensure_size( &
            matrix%m%data_area, &
            data_size, nocopy=.TRUE.)
      ENDIF
      CALL ensure_array_size( &
         matrix%m%index, &
         ub=index_size, nocopy=.TRUE., &
         memory_type=dbcsr_get_index_memory_type(matrix))
      matrix%m%negate_real = source_matrix%m%negate_real
      matrix%m%negate_imaginary = source_matrix%m%negate_imaginary
      matrix%m%local_indexing = source_matrix%m%local_indexing
      matrix%m%list_indexing = source_matrix%m%list_indexing
      !
      IF (source_matrix%m%has_local_rows) THEN
         matrix%m%local_rows = source_matrix%m%local_rows
         CALL array_hold(matrix%m%local_rows)
         matrix%m%has_local_rows = .TRUE.
      ENDIF
      IF (source_matrix%m%has_global_rows) THEN
         matrix%m%global_rows = source_matrix%m%global_rows
         CALL array_hold(matrix%m%global_rows)
         matrix%m%has_global_rows = .TRUE.
      ENDIF
      IF (source_matrix%m%has_local_cols) THEN
         matrix%m%local_cols = source_matrix%m%local_cols
         CALL array_hold(matrix%m%local_cols)
         matrix%m%has_local_cols = .TRUE.
      ENDIF
      IF (source_matrix%m%has_global_cols) THEN
         matrix%m%global_cols = source_matrix%m%global_cols
         CALL array_hold(matrix%m%global_cols)
         matrix%m%has_global_cols = .TRUE.
      ENDIF

   END SUBROUTINE setup_buffer_matrix

! **************************************************************************************************
!> \brief ...
!> \param buffer_set ...
!> \param nimages ...
!> \param imgdist ...
!> \param template_matrix ...
!> \param index_size_merged ...
!> \param data_buffer ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE setup_buffer_matrices_images(buffer_set, nimages, imgdist, &
                                           template_matrix, index_size_merged, data_buffer)
      TYPE(dbcsr_1d_array_type), INTENT(INOUT)           :: buffer_set
      INTEGER, INTENT(IN)                                :: nimages
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      TYPE(dbcsr_obj), INTENT(IN)                        :: template_matrix
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: index_size_merged
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: data_buffer

      INTEGER                                            :: v_ki

!   ---------------------------------------------------------------------------

      buffer_set%image_dist = imgdist
      CALL dbcsr_image_dist_hold(imgdist)
      !
      ALLOCATE (buffer_set%mats(nimages))
!$OMP PARALLEL DO DEFAULT(NONE) &
!$OMP SHARED(nimages,buffer_set,template_matrix,&
!$OMP        imgdist,data_buffer,memtype_mpi_buffer,&
!$OMP        index_size_merged)
      DO v_ki = 1, nimages
         CALL dbcsr_init(buffer_set%mats(v_ki))
         CALL dbcsr_create(buffer_set%mats(v_ki), &
                           "Buffer image of "//template_matrix%m%name, &
                           imgdist%i%main, &
                           dbcsr_type_no_symmetry, &
                           array_data(template_matrix%m%row_blk_size), array_data(template_matrix%m%col_blk_size), &
                           template_matrix%m%row_blk_size, template_matrix%m%col_blk_size, &
                           data_buffer=data_buffer, &
                           max_rbs=template_matrix%m%max_rbs, max_cbs=template_matrix%m%max_cbs, &
                           row_blk_offset=template_matrix%m%row_blk_offset, col_blk_offset=template_matrix%m%col_blk_offset, &
                           index_memory_type=memtype_mpi_buffer, &
                           make_index=.FALSE.)
         IF (PRESENT(index_size_merged)) THEN
            CALL ensure_array_size(buffer_set%mats(v_ki)%m%index, &
                                   ub=index_size_merged(v_ki), nocopy=.TRUE., &
                                   memory_type=dbcsr_get_index_memory_type(buffer_set%mats(v_ki)))
         ENDIF
         buffer_set%mats(v_ki)%m%negate_real = template_matrix%m%negate_real
         buffer_set%mats(v_ki)%m%negate_imaginary = template_matrix%m%negate_imaginary
         buffer_set%mats(v_ki)%m%local_indexing = .TRUE.
         buffer_set%mats(v_ki)%m%list_indexing = .TRUE.
      ENDDO
!$OMP END PARALLEL DO
   END SUBROUTINE setup_buffer_matrices_images

! **************************************************************************************************
!> \brief ...
!> \param buffer_set ...
!> \param buff_nrows ...
!> \param buff_ncols ...
!> \param source_matrix ...
!> \param index_size ...
!> \param data_size ...
! **************************************************************************************************
   SUBROUTINE setup_buffer_matrices(buffer_set, buff_nrows, buff_ncols, &
                                    source_matrix, index_size, data_size)
      TYPE(dbcsr_2d_array_type), INTENT(OUT)             :: buffer_set
      INTEGER, INTENT(IN)                                :: buff_nrows, buff_ncols
      TYPE(dbcsr_obj), INTENT(IN)                        :: source_matrix
      INTEGER, INTENT(IN)                                :: index_size, data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_buffer_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col_image, handle, row_image

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      CALL dbcsr_image_dist_init(buffer_set%image_dist)
      ALLOCATE (buffer_set%mats(buff_nrows, buff_ncols))
      DO row_image = 1, buff_nrows
         DO col_image = 1, buff_ncols
            CALL setup_buffer_matrix(buffer_set%mats(row_image, col_image), &
                                     source_matrix, index_size, data_size=data_size, &
                                     data_memory_type=memtype_abpanel_2)
         ENDDO
      ENDDO
      IF (source_matrix%m%local_indexing .AND. careful_mod) THEN
         IF (.NOT. array_exists(source_matrix%m%local_rows)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6686,"Local rows should exist.")
         IF (.NOT. array_exists(source_matrix%m%global_rows)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6688,"Global rows should exist.")
         !
         IF (.NOT. array_exists(source_matrix%m%local_cols)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6691,"Local cols should exist.")
         IF (.NOT. array_exists(source_matrix%m%global_cols)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6693,"Global cols should exist.")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE setup_buffer_matrices

! ***************************************************************************************************
!> \brief Enlarge left_set and right_set to hold any a/b-panel.
!>  left_set and right_set are created by make_images to hold the a/b-panels
!>  used for the inital cannon-tick. This routine ensures that these buffers
!>  can hold any a/b-panel occuring during a matrix multiply and makes them
!>  therefore suitable as buffers for the entire cannon algorithm.
!>  This saves memory since no seperate buffers for the first cannon-tick
!>  have to be allocated.
!>
!> \param buffer_set ...
!> \param index_size ...
!> \param data_size ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE buffer_matrices_ensure_size(buffer_set, index_size, data_size)
      TYPE(dbcsr_2d_array_type), INTENT(INOUT)           :: buffer_set
      INTEGER, INTENT(IN)                                :: index_size, data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'buffer_matrices_ensure_size', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col_image, handle, row_image

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      DO row_image = 1, SIZE(buffer_set%mats, 1)
         DO col_image = 1, SIZE(buffer_set%mats, 2)
            CALL dbcsr_data_ensure_size(buffer_set%mats(row_image, col_image)%m%data_area, &
                                        data_size)
            CALL ensure_array_size(buffer_set%mats(row_image, col_image)%m%index, &
                                   ub=index_size, &
                                   memory_type=dbcsr_get_index_memory_type(buffer_set%mats(row_image, col_image)))
            CALL dbcsr_repoint_index(buffer_set%mats(row_image, col_image)%m)
         ENDDO
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE buffer_matrices_ensure_size

! **************************************************************************************************
!> \brief ...
!> \param meta_buffer ...
!> \param img_refs_rows ...
!> \param img_refs_cols ...
!> \param refs_size ...
!> \param refs_displ ...
!> \param size_index_unmerged ...
!> \param has_threads ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE setup_rec_index_images(meta_buffer, img_refs_rows, img_refs_cols, &
                                     refs_size, refs_displ, size_index_unmerged, has_threads)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: meta_buffer
      INTEGER, DIMENSION(:), INTENT(IN)                  :: img_refs_rows, img_refs_cols, refs_size, &
                                                            refs_displ
      INTEGER, INTENT(IN)                                :: size_index_unmerged
      LOGICAL, INTENT(IN)                                :: has_threads

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_rec_index_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, in, nblkcols_local, &
                                                            nblkrows_local, t_f, t_l, t_size

!$    INTEGER                           :: ithread

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (has_threads) THEN
         nblkrows_local = img_refs_rows(1)
      ELSE
         nblkcols_local = img_refs_cols(1)
      ENDIF
      !
      DO in = 1, SIZE(refs_size)
         IF (refs_size(in) .EQ. 0) CYCLE
         t_size = meta_buffer(refs_displ(in)+dbcsr_slot_nblks)
         IF (careful_mod) THEN
            IF (refs_size(in)-size_index_unmerged .NE. t_size*3) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6779,"Block count mismatch.")
         ENDIF
         IF (has_threads) THEN
            nblkcols_local = img_refs_cols(in)
         ELSE
            nblkrows_local = img_refs_rows(in)
         ENDIF
         t_f = 1
         t_l = t_size
!$OMP        PARALLEL IF (has_threads) DEFAULT (NONE) &
!$OMP        PRIVATE (ithread) &
!$OMP        FIRSTPRIVATE (t_f, t_l, t_size) &
!$OMP        SHARED (meta_buffer, in, has_threads, refs_displ, &
!$OMP                size_index_unmerged, nblkrows_local, nblkcols_local)
!$       ithread = OMP_GET_THREAD_NUM()+dbcsr_slot_nblks+1
!$       IF (has_threads) THEN
!$          t_f = meta_buffer(refs_displ(in)+ithread)+1
!$          t_l = meta_buffer(refs_displ(in)+ithread+1)
!$       ENDIF
         t_size = t_l-t_f+1
!$OMP        BARRIER
         IF (t_size .GT. 0) THEN
            CALL rec_sort_index(1, nblkrows_local, &
                                1, nblkcols_local, &
                                t_size, &
                                meta_buffer(refs_displ(in)+size_index_unmerged+t_f*3-2: &
                                            refs_displ(in)+size_index_unmerged+t_l*3), &
                                0)
         ENDIF
!$OMP        END PARALLEL
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE setup_rec_index_images

! **************************************************************************************************
!> \brief ...
!> \param matrix_set ...
!> \param n_rows ...
!> \param n_cols ...
! **************************************************************************************************
   SUBROUTINE setup_rec_index_2d(matrix_set, n_rows, n_cols)
      TYPE(dbcsr_2d_array_type), INTENT(INOUT)           :: matrix_set
      INTEGER, INTENT(IN)                                :: n_rows, n_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_rec_index_2d', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, i_col, i_row, t_f, t_l, t_size

!$    INTEGER                                  :: ithread
      LOGICAL                                  :: thread_redist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      DO i_row = 1, n_rows
         DO i_col = 1, n_cols
            IF (.FALSE.) &
               CALL dbcsr_reset_vlocals(matrix_set%mats(i_row, i_col), &
                                        matrix_set%image_dist)
            IF (dbg) THEN
               WRITE (*, *) routineN//" m, n, size", &
                  SIZE(matrix_set%mats(i_row, i_col)%m%coo_l), &
                  dbcsr_nblkrows_local(matrix_set%mats(i_row, i_col)), &
                  dbcsr_nblkcols_local(matrix_set%mats(i_row, i_col))
               WRITE (*, '(3(1X,I7))') matrix_set%mats(i_row, i_col)%m%coo_l
            ENDIF
            IF (careful_mod) THEN
               IF (SIZE(matrix_set%mats(i_row, i_col)%m%coo_l) .NE. matrix_set%mats(i_row, i_col)%m%nblks*3) &
                  CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",6849,"Block count mismatch.")
            ENDIF
            thread_redist = ASSOCIATED(matrix_set%mats(i_row, i_col)%m%thr_c)
            t_size = SIZE(matrix_set%mats(i_row, i_col)%m%coo_l)/3
            t_f = 1
            t_l = t_size
!$OMP           PARALLEL IF (thread_redist) DEFAULT (NONE) &
!$OMP           PRIVATE (ithread) &
!$OMP           FIRSTPRIVATE (t_f, t_l, t_size) &
!$OMP           SHARED (matrix_set, i_row, i_col, thread_redist)
!$          ithread = OMP_GET_THREAD_NUM()
!$          IF (thread_redist) THEN
!$             t_f = matrix_set%mats(i_row, i_col)%m%thr_c(ithread+1)+1
!$             t_l = matrix_set%mats(i_row, i_col)%m%thr_c(ithread+2)
!$          ENDIF
            t_size = t_l-t_f+1
!$OMP           BARRIER
            IF (t_size .GT. 0) THEN
               IF (matrix_set%mats(i_row, i_col)%m%local_indexing) THEN
                  CALL call_rec_sort_index( &
                     dbcsr_nblkrows_local(matrix_set%mats(i_row, i_col)), &
                     dbcsr_nblkcols_local(matrix_set%mats(i_row, i_col)), &
                     t_size, &
                     matrix_set%mats(i_row, i_col)%m%coo_l((t_f*3-2):(t_l*3)))
               ELSE
                  CALL call_rec_sort_index( &
                     dbcsr_nblkrows_total(matrix_set%mats(i_row, i_col)), &
                     dbcsr_nblkcols_total(matrix_set%mats(i_row, i_col)), &
                     t_size, &
                     matrix_set%mats(i_row, i_col)%m%coo_l((t_f*3-2):(t_l*3)))
               ENDIF
            ENDIF
!$OMP           END PARALLEL
         ENDDO
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE setup_rec_index_2d

! **************************************************************************************************
!> \brief Used to thunk a call to rec_sort_index
!> \param m ...
!> \param n ...
!> \param nblks ...
!> \param idx ...
! **************************************************************************************************
   SUBROUTINE call_rec_sort_index(m, n, nblks, idx)
      INTEGER, INTENT(IN)                                :: m, n, nblks
      INTEGER, DIMENSION(3, 1:nblks), INTENT(INOUT)      :: idx

      CHARACTER(len=*), PARAMETER :: routineN = 'call_rec_sort_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.FALSE.) WRITE (*, *) " Calling rec_sort_index, size", nblks
      CALL rec_sort_index(1, m, 1, n, nblks, idx, 0)
      CALL timestop(handle)
   END SUBROUTINE call_rec_sort_index

! **************************************************************************************************
!> \brief Sorts index for recursing.
!> \param mi ...
!> \param mf ...
!> \param ni ...
!> \param nf ...
!> \param nele ...
!> \param a ...
!> \param d ...
!> \par History
!> - 2011-02-17 [UB] modified for use in DBCSR; reduced memory usage.
!> \author JV
!> \note Always cut longest first. On a tie cut N
! **************************************************************************************************
   RECURSIVE SUBROUTINE rec_sort_index(mi, mf, ni, nf, nele, a, d)
      INTEGER, INTENT(IN)                                :: mi, mf, ni, nf, nele
      INTEGER, DIMENSION(3, 1:nele), INTENT(inout)       :: a
      INTEGER, INTENT(IN)                                :: d

      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: half, M, N, nlow
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: tmp

!   ---------------------------------------------------------------------------

      IF (dbg) THEN
         WRITE (*, *) " rs", mi, mf, "/", ni, nf, "=>", nele, d
         WRITE (*, '(3(1X,I7))') a(:, 1:nele)
      ENDIF
      IF (dbg) THEN
         IF (d .GT. 20) THEN
            WRITE (*, *) a(1, -d*1000)
         ENDIF
      ENDIF
      ALLOCATE (tmp(3, nele))
      M = mf-mi+1
      N = nf-ni+1
      IF (M > N) THEN
         half = M/2
         CALL rec_split(nele, a, tmp, 1, nlow, mi, half)
         a = tmp
         DEALLOCATE (tmp)
         IF (nlow .GT. 1) THEN
            CALL rec_sort_index(mi, mi+half-1, ni, nf, nlow, a(:, 1:nlow), d+1)
         ENDIF
         IF (nele-nlow .GT. 1) THEN
            CALL rec_sort_index(mi+half, mf, ni, nf, nele-nlow, a(:, nlow+1:nele), d+1)
         ENDIF
      ELSE
         half = N/2
         CALL rec_split(nele, a, tmp, 2, nlow, ni, half)
         a = tmp
         DEALLOCATE (tmp)
         IF (nlow .GT. 1) THEN
            CALL rec_sort_index(mi, mf, ni, ni+half-1, nlow, a(:, 1:nlow), d+1)
         ENDIF
         IF (nele-nlow .GT. 1) THEN
            CALL rec_sort_index(mi, mf, ni+half, nf, nele-nlow, a(:, nlow+1:nele), d+1)
         ENDIF
      ENDIF
   END SUBROUTINE rec_sort_index

! **************************************************************************************************
!> \brief ...
!> \param nele ...
!> \param a ...
!> \param split ...
!> \param row_or_col ...
!> \param nlow ...
!> \param mi ...
!> \param half ...
! **************************************************************************************************
   SUBROUTINE rec_split(nele, a, split, row_or_col, nlow, mi, half)
      INTEGER, INTENT(IN)                                :: nele
      INTEGER, DIMENSION(3, nele), INTENT(IN)            :: a
      INTEGER, DIMENSION(3, nele), INTENT(OUT)           :: split
      INTEGER, INTENT(IN)                                :: row_or_col
      INTEGER, INTENT(OUT)                               :: nlow
      INTEGER, INTENT(IN)                                :: mi, half

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rec_split', routineP = moduleN//':'//routineN

      INTEGER                                            :: el, half_m, p_high, p_low

      half_m = mi+half-1
      p_low = 1
      p_high = nele
      DO el = 1, nele
         IF (a(row_or_col, el) <= half_m) THEN
            split(1:3, p_low) = a(1:3, el)
            p_low = p_low+1
         ELSE
            split(1:3, p_high) = a(1:3, el)
            p_high = p_high-1
         ENDIF
      ENDDO
      nlow = p_low-1
      IF(.NOT.(p_high .EQ. nlow))CALL cp__a("dbcsr/mm/dbcsr_mm_cannon.F",7009)

   END SUBROUTINE rec_split

! **************************************************************************************************
!> \brief Switches pointers between two matrix sets
!> \param[in,out] set1p ...
!> \param[in,out] set2p ...
! **************************************************************************************************
   SUBROUTINE dbcsr_switch_sets(set1p, set2p)
      TYPE(dbcsr_2d_array_type), POINTER                 :: set1p, set2p

      TYPE(dbcsr_2d_array_type), POINTER                 :: tmp_set

!   ---------------------------------------------------------------------------

      tmp_set => set1p
      set1p => set2p
      set2p => tmp_set
   END SUBROUTINE dbcsr_switch_sets

! **************************************************************************************************
!> \brief Switches pointers between two data areas
!> \param area1p ...
!> \param area2p ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_switch_d_ptrs(area1p, area2p)
      TYPE(dbcsr_data_obj), POINTER                      :: area1p, area2p

      TYPE(dbcsr_data_obj), POINTER                      :: tmp_p

      tmp_p => area1p
      area1p => area2p
      area2p => tmp_p
   END SUBROUTINE dbcsr_switch_d_ptrs

! **************************************************************************************************
! The following routines are helped here to help the compiler optimize them
! out.
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param t ...
!> \retval blas_mat_type ...
! **************************************************************************************************
   ELEMENTAL FUNCTION blas_mat_type(t)
      LOGICAL, INTENT(IN)                                :: t
      CHARACTER                                          :: blas_mat_type

      IF (t) THEN
         blas_mat_type = 'T'
      ELSE
         blas_mat_type = 'N'
      ENDIF
   END FUNCTION blas_mat_type

! **************************************************************************************************
!> \brief ...
!> \param t ...
!> \retval flip_type ...
! **************************************************************************************************
   ELEMENTAL FUNCTION flip_type(t)
      CHARACTER, INTENT(IN)                              :: t
      CHARACTER                                          :: flip_type

      SELECT CASE (t)
      CASE ('N')
         flip_type = 'T'
      CASE ('T')
         flip_type = 'N'
      CASE DEFAULT
         flip_type = '@'
      END SELECT
   END FUNCTION flip_type

! **************************************************************************************************
!> \brief Calculates per-block norms.
!>
!> Rewritten to be very low-level.
!> \param[in,out] matrix     DBCSR matrix for which to calculate norms
!> \param[out] norms         Block norms
! **************************************************************************************************
   SUBROUTINE calculate_norms(matrix, norms)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      REAL(kind=sp), DIMENSION(:), INTENT(OUT)           :: norms

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_norms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle, nblks, nrows
      INTEGER, DIMENSION(1), TARGET                      :: tmp
      INTEGER, DIMENSION(:), POINTER                     :: local_cols, local_rows

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      ! Checks for validity
      IF (.NOT. dbcsr_valid_index(matrix)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7109,"The matrix must be valid.")
      data_type = dbcsr_get_data_type(matrix)
      IF (matrix%m%local_indexing) THEN
         IF (careful_mod .AND. .NOT. array_exists(matrix%m%local_rows)) &
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7113,"Global row mapping should exist")
         local_rows => array_data(matrix%m%local_rows)
         nrows = SIZE(local_rows)
         local_cols => array_data(matrix%m%local_cols)
      ELSE
         local_rows => tmp ! Have something valid to point to
         local_cols => tmp
         nrows = matrix%m%nblkrows_total
      ENDIF
      IF (matrix%m%list_indexing) THEN
         nblks = matrix%m%nblks
         SELECT CASE (data_type)
         CASE (dbcsr_type_real_4)
            CALL calc_norms_list_s(norms, nblks, &
                                   matrix%m%coo_l, &
                                   array_data(matrix%m%row_blk_size), &
                                   array_data(matrix%m%col_blk_size), &
                                   dbcsr_get_data_p_s(matrix%m%data_area), &
                                   local=matrix%m%local_indexing, &
                                   local2global_rows=local_rows, &
                                   local2global_cols=local_cols)
         CASE (dbcsr_type_real_8)
            CALL calc_norms_list_d(norms, nblks, &
                                   matrix%m%coo_l, &
                                   array_data(matrix%m%row_blk_size), &
                                   array_data(matrix%m%col_blk_size), &
                                   dbcsr_get_data_p_d(matrix%m%data_area), &
                                   local=matrix%m%local_indexing, &
                                   local2global_rows=local_rows, &
                                   local2global_cols=local_cols)
         CASE (dbcsr_type_complex_4)
            CALL calc_norms_list_c(norms, nblks, &
                                   matrix%m%coo_l, &
                                   array_data(matrix%m%row_blk_size), &
                                   array_data(matrix%m%col_blk_size), &
                                   dbcsr_get_data_p_c(matrix%m%data_area), &
                                   local=matrix%m%local_indexing, &
                                   local2global_rows=local_rows, &
                                   local2global_cols=local_cols)
         CASE (dbcsr_type_complex_8)
            CALL calc_norms_list_z(norms, nblks, &
                                   matrix%m%coo_l, &
                                   array_data(matrix%m%row_blk_size), &
                                   array_data(matrix%m%col_blk_size), &
                                   dbcsr_get_data_p_z(matrix%m%data_area), &
                                   local=matrix%m%local_indexing, &
                                   local2global_rows=local_rows, &
                                   local2global_cols=local_cols)
         CASE DEFAULT
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7162,"Invalid data type.")
         END SELECT
      ELSE
         SELECT CASE (data_type)
         CASE (dbcsr_type_real_4)
            CALL calc_norms_s(norms, nrows, &
                              matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, &
                              array_data(matrix%m%row_blk_size), &
                              array_data(matrix%m%col_blk_size), &
                              dbcsr_get_data_p_s(matrix%m%data_area), &
                              local=matrix%m%local_indexing, &
                              local2global=local_rows)
         CASE (dbcsr_type_real_8)
            CALL calc_norms_d(norms, nrows, &
                              matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, &
                              array_data(matrix%m%row_blk_size), &
                              array_data(matrix%m%col_blk_size), &
                              dbcsr_get_data_p_d(matrix%m%data_area), &
                              local=matrix%m%local_indexing, &
                              local2global=local_rows)
         CASE (dbcsr_type_complex_4)
            CALL calc_norms_c(norms, nrows, &
                              matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, &
                              array_data(matrix%m%row_blk_size), &
                              array_data(matrix%m%col_blk_size), &
                              dbcsr_get_data_p_c(matrix%m%data_area), &
                              local=matrix%m%local_indexing, &
                              local2global=local_rows)
         CASE (dbcsr_type_complex_8)
            CALL calc_norms_z(norms, nrows, &
                              matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, &
                              array_data(matrix%m%row_blk_size), &
                              array_data(matrix%m%col_blk_size), &
                              dbcsr_get_data_p_z(matrix%m%data_area), &
                              local=matrix%m%local_indexing, &
                              local2global=local_rows)
         CASE DEFAULT
            CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7199,"Invalid data type.")
         END SELECT
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE calculate_norms

! **************************************************************************************************
!> \brief Calculates max norms per each image
!> \param buffer ...
!> \param refs_size ...
!> \param refs_displ ...
!> \param img_map ...
!> \param img_offset ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param local_rows ...
!> \param local_cols ...
!> \param slot_coo_l ...
!> \param[out] max_norms     Max norms per image array
!> \param is_left ...
!> \param off_diagonal ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE calculate_max_image_norms(buffer, refs_size, refs_displ, &
                                        img_map, img_offset, &
                                        row_blk_size, col_blk_size, &
                                        local_rows, local_cols, &
                                        slot_coo_l, &
                                        max_norms, is_left, off_diagonal)
      TYPE(dbcsr_buffer), INTENT(IN)                     :: buffer
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: refs_size
      INTEGER, DIMENSION(:, :, :), INTENT(IN), POINTER   :: refs_displ
      INTEGER, DIMENSION(:), INTENT(IN)                  :: img_map, img_offset, row_blk_size, &
                                                            col_blk_size, local_rows, local_cols
      INTEGER, INTENT(IN)                                :: slot_coo_l
      REAL(kind=sp), DIMENSION(:), INTENT(INOUT)         :: max_norms
      LOGICAL                                            :: is_left, off_diagonal

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_max_image_norms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      !
      SELECT CASE (buffer%data%d%data_type)
      CASE (dbcsr_type_real_4)
         CALL calc_max_image_norms_s(buffer%meta, dbcsr_get_data_p_s(buffer%data), &
                                     buffer%meta_diag, dbcsr_get_data_p_s(buffer%data_diag), &
                                     refs_size, refs_displ, &
                                     img_map, &
                                     img_offset, &
                                     row_blk_size, col_blk_size, &
                                     local_rows, local_cols, &
                                     slot_coo_l, &
                                     max_norms, is_left, off_diagonal)
      CASE (dbcsr_type_real_8)
         CALL calc_max_image_norms_d(buffer%meta, dbcsr_get_data_p_d(buffer%data), &
                                     buffer%meta_diag, dbcsr_get_data_p_d(buffer%data_diag), &
                                     refs_size, refs_displ, &
                                     img_map, &
                                     img_offset, &
                                     row_blk_size, col_blk_size, &
                                     local_rows, local_cols, &
                                     slot_coo_l, &
                                     max_norms, is_left, off_diagonal)
      CASE (dbcsr_type_complex_4)
         CALL calc_max_image_norms_c(buffer%meta, dbcsr_get_data_p_c(buffer%data), &
                                     buffer%meta_diag, dbcsr_get_data_p_c(buffer%data_diag), &
                                     refs_size, refs_displ, &
                                     img_map, &
                                     img_offset, &
                                     row_blk_size, col_blk_size, &
                                     local_rows, local_cols, &
                                     slot_coo_l, &
                                     max_norms, is_left, off_diagonal)
      CASE (dbcsr_type_complex_8)
         CALL calc_max_image_norms_z(buffer%meta, dbcsr_get_data_p_z(buffer%data), &
                                     buffer%meta_diag, dbcsr_get_data_p_z(buffer%data_diag), &
                                     refs_size, refs_displ, &
                                     img_map, &
                                     img_offset, &
                                     row_blk_size, col_blk_size, &
                                     local_rows, local_cols, &
                                     slot_coo_l, &
                                     max_norms, is_left, off_diagonal)
      CASE DEFAULT
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7287,"Invalid data type.")
      END SELECT
      !
      CALL timestop(handle)
   END SUBROUTINE calculate_max_image_norms

! **************************************************************************************************
!> \brief Calculates norms per each image
!> \param images ...
!> \param data_type ...
!> \param norms ...
!> \param uf ...
!> \param ul ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE calculate_image_norms(images, data_type, norms, uf, ul)
      TYPE(dbcsr_1d_array_type), INTENT(IN)              :: images
      INTEGER, INTENT(IN)                                :: data_type
      REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT)      :: norms
      INTEGER, INTENT(IN)                                :: uf, ul

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_image_norms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      !
      ! Check for row/columns image sizes
      IF (SIZE(images%mats, 1) .GT. SIZE(norms, 2) .OR. SIZE(images%mats, 1) .LT. (ul-uf+1)) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7317,"Wrong number of images!")
      !
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         CALL calc_image_norms_s(images, norms, uf, ul)
      CASE (dbcsr_type_real_8)
         CALL calc_image_norms_d(images, norms, uf, ul)
      CASE (dbcsr_type_complex_4)
         CALL calc_image_norms_c(images, norms, uf, ul)
      CASE (dbcsr_type_complex_8)
         CALL calc_image_norms_z(images, norms, uf, ul)
      CASE DEFAULT
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7329,"Invalid data type.")
      END SELECT
      !
      CALL timestop(handle)
   END SUBROUTINE calculate_image_norms

! **************************************************************************************************
!> \brief write out a stack for transposing the blocks
!> \param matrix ...
!> \param trs_stackbuf ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_transpose_blocks(matrix, trs_stackbuf)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: trs_stackbuf

      CHARACTER(len=*), PARAMETER :: routineN = 'acc_transpose_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_p, col, handle, i, m, mi, mi_max, n, &
                                                            nblks, ni, ni_max, offset, row, x
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: counters, filled, offsets
      LOGICAL                                            :: local
      INTEGER, DIMENSION(:), POINTER :: blk_index, col_blk_sizes, col_blk_sizes2enum, &
         enum2col_blk_sizes, enum2row_blk_sizes, local2global_cols, local2global_rows, &
         row_blk_sizes, row_blk_sizes2enum, trs_stack
      INTEGER, DIMENSION(3, matrix%m%nblks)              :: tmp_stack

      CALL timeset(routineN, handle)

      NULLIFY (row_blk_sizes2enum, enum2row_blk_sizes)
      NULLIFY (col_blk_sizes2enum, enum2col_blk_sizes)
      NULLIFY (local2global_rows, local2global_cols, trs_stack)

      IF (.NOT. matrix%m%list_indexing) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7364,"build_trs_stack: only list_indexing supported.")
      IF (trs_stackbuf%d%data_type /= dbcsr_type_int_4) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7366,"build_trs_stack: stac_buf has wrong datatype")
      blk_index => matrix%m%coo_l
      row_blk_sizes => array_data(matrix%m%row_blk_size)
      col_blk_sizes => array_data(matrix%m%col_blk_size)
      local = matrix%m%local_indexing
      local2global_rows => array_data(matrix%m%local_rows)
      local2global_cols => array_data(matrix%m%local_cols)
      nblks = matrix%m%nblks

      ! enumerate the blocksizes to keep the following 2D-arrays small.
      CALL enumerate_blk_sizes(row_blk_sizes, row_blk_sizes2enum, enum2row_blk_sizes)
      CALL enumerate_blk_sizes(col_blk_sizes, col_blk_sizes2enum, enum2col_blk_sizes)
      mi_max = SIZE(enum2row_blk_sizes); ni_max = SIZE(enum2col_blk_sizes)
      ALLOCATE (counters(mi_max, ni_max), offsets(mi_max, ni_max), filled(mi_max, ni_max))
      counters(:, :) = 0; offsets(:, :) = 0; filled(:, :) = 0

      ! make sure buffer from previous cannon-tick was uploaded
      CALL acc_event_synchronize(trs_stackbuf%d%acc_ready)

      CALL dbcsr_data_ensure_size(trs_stackbuf, data_size=nblks, nocopy=.TRUE.)
      trs_stack => trs_stackbuf%d%i4

      ! collect block addresses and dimensions in a temporary stack
      ! while doing so, also count number of blocks per block-dimensions
      DO i = 1, nblks
         row = blk_index(3*(i-1)+1)
         col = blk_index(3*(i-1)+2)
         blk_p = blk_index(3*(i-1)+3)
         IF (blk_p == 0) CYCLE
         IF (local) THEN
            row = local2global_rows(row)
            col = local2global_cols(col)
         ENDIF
         m = row_blk_sizes(row)
         n = col_blk_sizes(col)
         mi = row_blk_sizes2enum(m)
         ni = col_blk_sizes2enum(n)
         tmp_stack(1, i) = mi
         tmp_stack(2, i) = ni
         tmp_stack(3, i) = blk_p-1
         counters(mi, ni) = counters(mi, ni)+1
      ENDDO

      ! calculate offsets for first element of each sub-stack
      offset = 0
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            offsets(mi, ni) = offset
            offset = offset+counters(mi, ni)
         ENDDO
      ENDDO

      ! write all sub-stacks into the host-pinned buffer
      DO i = 1, nblks
         mi = tmp_stack(1, i)
         ni = tmp_stack(2, i)
         blk_p = tmp_stack(3, i)
         x = offsets(mi, ni)+filled(mi, ni)+1
         trs_stack(x) = blk_p
         filled(mi, ni) = filled(mi, ni)+1
      ENDDO

      !sanity check
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            IF (filled(mi, ni) /= counters(mi, ni)) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7432,"acc_transpose_blocks: bug")
         END DO
      END DO

      !transfer all stacks
      CALL dbcsr_data_host2dev(trs_stackbuf)

      ! make sure block-buffer is uploaded befor running the kernels
      CALL acc_stream_wait_event(trs_stackbuf%d%memory_type%acc_stream, matrix%m%data_area%d%acc_ready)

      ! launch kernels
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            IF (counters(mi, ni) > 0) THEN
               m = enum2row_blk_sizes(mi)
               n = enum2col_blk_sizes(ni)
               CALL dbcsr_acc_transpose( &
                  trs_stack=trs_stackbuf%d%acc_devmem, &
                  offset=offsets(mi, ni), &
                  nblks=counters(mi, ni), &
                  datatype=matrix%m%data_type, &
                  buffer=matrix%m%data_area%d%acc_devmem, &
                  m=m, n=n, &
                  stream=trs_stackbuf%d%memory_type%acc_stream)
            END IF
         ENDDO
      ENDDO

      ! make sure block-buffer are not used until transpose kernels finnished
      CALL acc_event_record(trs_stackbuf%d%acc_ready, trs_stackbuf%d%memory_type%acc_stream)
      CALL acc_stream_wait_event(matrix%m%data_area%d%memory_type%acc_stream, trs_stackbuf%d%acc_ready)
      CALL acc_event_record(matrix%m%data_area%d%acc_ready, matrix%m%data_area%d%memory_type%acc_stream)

      DEALLOCATE (row_blk_sizes2enum, enum2row_blk_sizes)
      DEALLOCATE (col_blk_sizes2enum, enum2col_blk_sizes)
      CALL timestop(handle)
   END SUBROUTINE acc_transpose_blocks

! **************************************************************************************************
!> \brief write out a stack for transposing the blocks
!> \param matrices ...
!> \param DATA ...
!> \param trs_stackbuf ...
!> \param indices ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE acc_transpose_blocks_images(matrices, DATA, trs_stackbuf, indices)
      TYPE(dbcsr_1d_array_type), INTENT(IN)              :: matrices
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: DATA, trs_stackbuf
      INTEGER, DIMENSION(:), INTENT(IN)                  :: indices

      CHARACTER(len=*), PARAMETER :: routineN = 'acc_transpose_blocks_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_p, col, handle, i, imat, m, mi, &
                                                            mi_max, n, nblks, nblks_total, ni, &
                                                            ni_max, offset, row, x
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: counters, filled, offsets, tmp_stack
      INTEGER, DIMENSION(:), POINTER :: blk_index, col_blk_sizes, col_blk_sizes2enum, &
         enum2col_blk_sizes, enum2row_blk_sizes, local2global_cols, local2global_rows, &
         row_blk_sizes, row_blk_sizes2enum, trs_stack

      CALL timeset(routineN, handle)

      NULLIFY (row_blk_sizes2enum, enum2row_blk_sizes)
      NULLIFY (col_blk_sizes2enum, enum2col_blk_sizes)
      NULLIFY (local2global_rows, local2global_cols, trs_stack)

      IF (trs_stackbuf%d%data_type /= dbcsr_type_int_4) &
         CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7501,"build_trs_stack: stac_buf has wrong datatype")

      row_blk_sizes => array_data(matrices%mats(indices(1))%m%row_blk_size)
      col_blk_sizes => array_data(matrices%mats(indices(1))%m%col_blk_size)

      ! enumerate the blocksizes to keep the following 2D-arrays small.
      CALL enumerate_blk_sizes(row_blk_sizes, row_blk_sizes2enum, enum2row_blk_sizes)
      CALL enumerate_blk_sizes(col_blk_sizes, col_blk_sizes2enum, enum2col_blk_sizes)
      mi_max = SIZE(enum2row_blk_sizes); ni_max = SIZE(enum2col_blk_sizes)
      ALLOCATE (counters(mi_max, ni_max), offsets(mi_max, ni_max), filled(mi_max, ni_max))
      counters(:, :) = 0; offsets(:, :) = 0; filled(:, :) = 0

      ! make sure block-buffer is uploaded befor running the kernels
      CALL acc_stream_wait_event(trs_stackbuf%d%memory_type%acc_stream, data%d%acc_ready)

      nblks_total = 0
      DO imat = 1, SIZE(indices)
         nblks_total = nblks_total+matrices%mats(indices(imat))%m%nblks
      ENDDO
      ALLOCATE (tmp_stack(3, nblks_total))

      nblks_total = 0
      DO imat = 1, SIZE(indices)
         blk_index => matrices%mats(indices(imat))%m%coo_l
         local2global_rows => array_data(matrices%mats(indices(imat))%m%local_rows)
         local2global_cols => array_data(matrices%mats(indices(imat))%m%local_cols)
         nblks = matrices%mats(indices(imat))%m%nblks

         ! collect block addresses and dimensions in a temporary stack
         ! while doing so, also count number of blocks per block-dimensions
         DO i = 1, nblks
            row = blk_index(3*(i-1)+1)
            col = blk_index(3*(i-1)+2)
            blk_p = blk_index(3*(i-1)+3)
            IF (blk_p == 0) CYCLE
            row = local2global_rows(row)
            col = local2global_cols(col)
            m = row_blk_sizes(row)
            n = col_blk_sizes(col)
            mi = row_blk_sizes2enum(m)
            ni = col_blk_sizes2enum(n)
            tmp_stack(1, nblks_total+i) = mi
            tmp_stack(2, nblks_total+i) = ni
            tmp_stack(3, nblks_total+i) = blk_p-1
            counters(mi, ni) = counters(mi, ni)+1
         ENDDO
         nblks_total = nblks_total+nblks
      ENDDO

      ! calculate offsets for first element of each sub-stack
      offset = 0
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            offsets(mi, ni) = offset
            offset = offset+counters(mi, ni)
         ENDDO
      ENDDO

      ! make sure buffer from previous cannon-tick was uploaded
      CALL acc_event_synchronize(trs_stackbuf%d%acc_ready)
      CALL dbcsr_data_ensure_size(trs_stackbuf, data_size=nblks_total, nocopy=.TRUE.)
      trs_stack => trs_stackbuf%d%i4

      ! write all sub-stacks into the host-pinned buffer
      DO i = 1, nblks_total
         mi = tmp_stack(1, i)
         ni = tmp_stack(2, i)
         blk_p = tmp_stack(3, i)
         x = offsets(mi, ni)+filled(mi, ni)+1
         trs_stack(x) = blk_p
         filled(mi, ni) = filled(mi, ni)+1
      ENDDO

      !sanity check
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            IF (filled(mi, ni) /= counters(mi, ni)) &
               CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7578,"acc_transpose_blocks: bug")
         END DO
      END DO

      !transfer all stacks
      CALL dbcsr_data_host2dev(trs_stackbuf)

      ! launch kernels
      DO mi = 1, mi_max
         DO ni = 1, ni_max
            IF (counters(mi, ni) > 0) THEN
               m = enum2row_blk_sizes(mi)
               n = enum2col_blk_sizes(ni)
               CALL dbcsr_acc_transpose( &
                  trs_stack=trs_stackbuf%d%acc_devmem, &
                  offset=offsets(mi, ni), &
                  nblks=counters(mi, ni), &
                  datatype=data%d%data_type, &
                  buffer=data%d%acc_devmem, &
                  m=m, n=n, &
                  stream=trs_stackbuf%d%memory_type%acc_stream)
            END IF
         ENDDO
      ENDDO

      DEALLOCATE (tmp_stack)

      ! make sure block-buffer are not used until transpose kernels finished
      CALL acc_event_record(trs_stackbuf%d%acc_ready, trs_stackbuf%d%memory_type%acc_stream)
      CALL acc_stream_wait_event(data%d%memory_type%acc_stream, trs_stackbuf%d%acc_ready)
      CALL acc_event_record(data%d%acc_ready, data%d%memory_type%acc_stream)

      DEALLOCATE (row_blk_sizes2enum, enum2row_blk_sizes)
      DEALLOCATE (col_blk_sizes2enum, enum2col_blk_sizes)
      CALL timestop(handle)
   END SUBROUTINE acc_transpose_blocks_images

! **************************************************************************************************
!> \brief Enumerate all occuring blocksizes
!> \param blk_sizes ...
!> \param enum ...
!> \param rev_enum ...
!>  \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE enumerate_blk_sizes(blk_sizes, enum, rev_enum)
      INTEGER, DIMENSION(:), POINTER                     :: blk_sizes, enum, rev_enum

      INTEGER                                            :: i, n

      n = MAXVAL(blk_sizes)
      ALLOCATE (enum(0:n))
      enum(:) = 0

      DO i = 1, SIZE(blk_sizes)
         enum(blk_sizes(i)) = 1
      ENDDO

      n = SUM(enum)
      ALLOCATE (rev_enum(n))

      n = 0
      DO i = 0, SIZE(enum)-1
         IF (enum(i) > 0) THEN
            n = n+1
            enum(i) = n
            rev_enum(n) = i
         END IF
      ENDDO
   END SUBROUTINE enumerate_blk_sizes

! **************************************************************************************************
!> \brief Init buffer
!>
!> \param buffer ...
!> \param data_type ...
!> \param data_size ...
!> \param meta_size ...
!> \param requests_size ...
!> \param data_memory_type ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE buffer_init(buffer, data_type, &
                          data_size, meta_size, &
                          requests_size, &
                          data_memory_type)
      TYPE(dbcsr_buffer), INTENT(INOUT)                  :: buffer
      INTEGER, INTENT(IN)                                :: data_type, data_size, meta_size
      INTEGER, INTENT(IN), OPTIONAL                      :: requests_size
      TYPE(dbcsr_memtype_type), INTENT(IN)               :: data_memory_type

      IF (.NOT. buffer%is_valid) THEN
         ! First initialization
         CALL dbcsr_data_init(buffer%data)
         CALL dbcsr_data_new(buffer%data, data_type=data_type, &
                             data_size=data_size, memory_type=data_memory_type)
         CALL dbcsr_data_set_size_referenced(buffer%data, data_size)
         CALL dbcsr_data_init(buffer%data_diag)
         buffer%is_valid = .TRUE.
      ELSE
         CALL dbcsr_data_ensure_size(buffer%data, data_size, nocopy=.TRUE.)
      ENDIF
      !
      CALL ensure_array_size(buffer%meta, ub=meta_size, nocopy=.TRUE., &
                             memory_type=memtype_mpi_buffer)
      !
      IF (PRESENT(requests_size)) THEN
         IF (ALLOCATED(buffer%offset)) THEN
            IF (SIZE(buffer%offset, 2) .LT. requests_size*2) DEALLOCATE (buffer%offset)
         ENDIF
         IF (.NOT. ALLOCATED(buffer%offset)) ALLOCATE (buffer%offset(2, requests_size*2))
         buffer%offset(:, 1) = 0
         IF (ALLOCATED(buffer%get_requests)) THEN
            IF (SIZE(buffer%get_requests, 2) .LT. requests_size) DEALLOCATE (buffer%get_requests)
         ENDIF
         IF (.NOT. ALLOCATED(buffer%get_requests)) ALLOCATE (buffer%get_requests(2, requests_size))
         IF (ALLOCATED(buffer%get_requests_map)) THEN
            IF (SIZE(buffer%get_requests_map, 2) .LT. requests_size) DEALLOCATE (buffer%get_requests_map)
         ENDIF
         IF (.NOT. ALLOCATED(buffer%get_requests_map)) ALLOCATE (buffer%get_requests_map(2, requests_size))
      ENDIF
   END SUBROUTINE buffer_init

! **************************************************************************************************
!> \brief Resize diagonal buffer
!>
!> \param buffer ...
!> \param data_type ...
!> \param data_size_diag ...
!> \param meta_size_diag ...
!> \param data_memory_type ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE resize_buffer_diag(buffer, data_type, &
                                 data_size_diag, meta_size_diag, &
                                 data_memory_type)
      TYPE(dbcsr_buffer), INTENT(INOUT)                  :: buffer
      INTEGER, INTENT(IN)                                :: data_type, data_size_diag, meta_size_diag
      TYPE(dbcsr_memtype_type), INTENT(IN)               :: data_memory_type

      IF (.NOT. buffer%is_valid) CALL cp__b("dbcsr/mm/dbcsr_mm_cannon.F",7717,"Buffer not initialized!")
      !
      IF (dbcsr_data_valid(buffer%data_diag)) THEN
         CALL dbcsr_data_ensure_size(buffer%data_diag, data_size_diag, nocopy=.TRUE.)
      ELSE
         CALL dbcsr_data_init(buffer%data_diag)
         CALL dbcsr_data_new(buffer%data_diag, data_type=data_type, &
                             data_size=data_size_diag, memory_type=data_memory_type)
         CALL dbcsr_data_set_size_referenced(buffer%data_diag, data_size_diag)
      ENDIF
      !
      CALL ensure_array_size(buffer%meta_diag, ub=meta_size_diag, nocopy=.TRUE., &
                             memory_type=memtype_mpi_buffer)
   END SUBROUTINE resize_buffer_diag

! **************************************************************************************************
!> \brief Release buffers
!>
!> \param buffers ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE buffers_release(buffers)
      TYPE(dbcsr_buffers), INTENT(INOUT)                 :: buffers

      IF (buffers%has_rma_right_win) THEN
         CALL mp_win_free(buffers_orig%right%data_win)
         CALL mp_win_free(buffers_orig%right%meta_win)
         buffers%has_rma_right_win = .FALSE.
      ENDIF
      IF (buffers%has_rma_left_win) THEN
         CALL mp_win_free(buffers_orig%left%data_win)
         CALL mp_win_free(buffers_orig%left%meta_win)
         buffers%has_rma_left_win = .FALSE.
      ENDIF
      !
      IF (buffers%left%is_valid) THEN
         CALL dbcsr_data_release(buffers%left%data)
         IF (dbcsr_data_valid(buffers%left%data_diag)) &
            CALL dbcsr_data_release(buffers%left%data_diag)
         buffers%left%is_valid = .FALSE.
      ENDIF
      IF (ASSOCIATED(buffers%left%meta)) THEN
         DEALLOCATE (buffers%left%meta)
         NULLIFY (buffers%left%meta)
      ENDIF
      IF (ASSOCIATED(buffers%left%meta_diag)) THEN
         DEALLOCATE (buffers%left%meta_diag)
         NULLIFY (buffers%left%meta_diag)
      ENDIF
      IF (buffers%right%is_valid) THEN
         CALL dbcsr_data_release(buffers%right%data)
         IF (dbcsr_data_valid(buffers%right%data_diag)) &
            CALL dbcsr_data_release(buffers%right%data_diag)
         buffers%right%is_valid = .FALSE.
      ENDIF
      IF (ASSOCIATED(buffers%right%meta)) THEN
         DEALLOCATE (buffers%right%meta)
         NULLIFY (buffers%right%meta)
      ENDIF
      IF (ASSOCIATED(buffers%right%meta_diag)) THEN
         DEALLOCATE (buffers%right%meta_diag)
         NULLIFY (buffers%right%meta_diag)
      ENDIF
      !
      IF (ALLOCATED(buffers%left%offset)) DEALLOCATE (buffers%left%offset)
      IF (ALLOCATED(buffers%right%offset)) DEALLOCATE (buffers%right%offset)
      IF (ALLOCATED(buffers%left%get_requests)) DEALLOCATE (buffers%left%get_requests)
      IF (ALLOCATED(buffers%right%get_requests)) DEALLOCATE (buffers%right%get_requests)
      IF (ALLOCATED(buffers%left%get_requests_map)) DEALLOCATE (buffers%left%get_requests_map)
      IF (ALLOCATED(buffers%right%get_requests_map)) DEALLOCATE (buffers%right%get_requests_map)
      IF (ASSOCIATED(meta_scatter)) THEN
         DEALLOCATE (meta_scatter)
         NULLIFY (meta_scatter)
      ENDIF
      IF (ASSOCIATED(local_meta_scatter)) THEN
         DEALLOCATE (local_meta_scatter)
         NULLIFY (local_meta_scatter)
         CALL dbcsr_data_release(local_data_scatter)
      ENDIF
      IF (ASSOCIATED(local_meta_product_scatter)) THEN
         DEALLOCATE (local_meta_product_scatter)
         NULLIFY (local_meta_product_scatter)
      ENDIF
      IF (dbcsr_data_valid(local_data_product_scatter)) THEN
         CALL dbcsr_data_release(local_data_product_scatter)
      ENDIF
   END SUBROUTINE buffers_release

! **************************************************************************************************
!> \brief ...
!> \param meta_index ...
!> \param global_indices ...
!> \param num_slots ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE set_empty_meta_index(meta_index, global_indices, num_slots)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: meta_index
      INTEGER, DIMENSION(:), INTENT(IN)                  :: global_indices
      INTEGER, INTENT(IN)                                :: num_slots

!   ---------------------------------------------------------------------------

      meta_index(dbcsr_slot_size) = num_slots
      meta_index(dbcsr_slot_size+1:num_slots) = 0
      meta_index(dbcsr_slot_nblkrows_total:dbcsr_slot_nfullcols_local) = &
         global_indices(:)
      meta_index(dbcsr_slot_row_p) = 1
      meta_index(dbcsr_slot_col_i) = 1
      meta_index(dbcsr_slot_blk_p) = 1
      meta_index(dbcsr_slot_coo_l) = num_slots+1
      meta_index(dbcsr_num_slots) = num_slots
   END SUBROUTINE set_empty_meta_index

! **************************************************************************************************
!> \brief Merge images
!>
!> \param images ...
!> \param nrows_images ...
!> \param ncols_images ...
!> \param vprow ...
!> \param vpcol ...
!> \param offset_images ...
!> \param meta_buffer ...
!> \param uf ...
!> \param ul ...
!> \param imgdist ...
!> \param do_merge_rows ...
!> \param global_indices ...
!> \param vunmerged ...
!> \param vmerged ...
!> \param common_unmerged ...
!> \param indices_common_unmerged ...
!> \param ncommons_unmerged ...
!> \param nthreads ...
!> \author Alfio Lazzaro
! **************************************************************************************************
   SUBROUTINE merge_images(images, nrows_images, ncols_images, &
                           vprow, vpcol, offset_images, &
                           meta_buffer, &
                           uf, ul, &
                           imgdist, do_merge_rows, &
                           global_indices, &
                           vunmerged, vmerged, &
                           common_unmerged, &
                           indices_common_unmerged, &
                           ncommons_unmerged, &
                           nthreads)
      TYPE(dbcsr_1d_array_type), INTENT(INOUT)           :: images
      INTEGER, INTENT(IN)                                :: nrows_images, ncols_images, vprow, vpcol
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: offset_images
      INTEGER, DIMENSION(:), INTENT(IN), TARGET          :: meta_buffer
      INTEGER, INTENT(IN)                                :: uf, ul
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      LOGICAL, INTENT(IN)                                :: do_merge_rows
      INTEGER, DIMENSION(:), INTENT(IN)                  :: global_indices, vunmerged
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: vmerged
      LOGICAL, DIMENSION(:), INTENT(INOUT)               :: common_unmerged
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: indices_common_unmerged
      INTEGER, INTENT(INOUT)                             :: ncommons_unmerged
      INTEGER, INTENT(IN), OPTIONAL                      :: nthreads

      CHARACTER(len=*), PARAMETER :: routineN = 'merge_images', routineP = moduleN//':'//routineN

      INTEGER :: handle, ioffset, ithread, kf, ki, kl, mi, my_nthreads, my_vpcol, my_vprow, &
         nimages_merged, nimages_unmerged, offset_data, offset_meta, size_index_merged, &
         size_index_unmerged, ui, v_merged
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ioffset_first
      INTEGER, DIMENSION(:), POINTER                     :: unmerged_indices
      TYPE(dbcsr_obj), POINTER                           :: image_obj
      TYPE(dbcsr_type), POINTER                          :: image

      CALL timeset(routineN, handle)
      !
      IF (do_merge_rows) THEN
         ! merge over rows (left)
         nimages_unmerged = ncols_images
         nimages_merged = nrows_images
         my_vprow = vprow/nimages_merged
         my_vpcol = vpcol-1
      ELSE
         ! merge over cols (right)
         nimages_unmerged = nrows_images
         nimages_merged = ncols_images
         my_vprow = vprow-1
         my_vpcol = vpcol/nimages_merged
      ENDIF
      !
      size_index_unmerged = dbcsr_slot_nblks
      size_index_merged = dbcsr_num_slots
      my_nthreads = 1
      IF (PRESENT(nthreads)) THEN
!$       size_index_unmerged = size_index_unmerged+nthreads+1
!$       size_index_merged = size_index_merged+nthreads+1
         my_nthreads = nthreads
      ENDIF
      !
      ALLOCATE (ioffset_first(nimages_unmerged))
      ioffset_first(1) = 0
      DO ui = 1, nimages_unmerged-1
         ioffset_first(ui+1) = ioffset_first(ui)+vunmerged(ui)
      ENDDO
      !
      ncommons_unmerged = 0
      !
!$OMP PARALLEL DO DEFAULT (NONE) &
!$OMP PRIVATE(image_obj,image,ioffset,ui,mi,v_merged, &
!$OMP         offset_meta,offset_data,unmerged_indices, &
!$OMP         ithread,kf,kl,ki) &
!$OMP SHARED(nimages_unmerged,images,global_indices, &
!$OMP        size_index_merged,vunmerged,do_merge_rows, &
!$OMP        my_vprow,my_vpcol,nimages_merged,nthreads, &
!$OMP        my_nthreads,ioffset_first,vmerged,offset_images, &
!$OMP        meta_buffer,size_index_unmerged,imgdist,uf,ul, &
!$OMP        ncommons_unmerged,common_unmerged,indices_common_unmerged) &
!$OMP schedule(dynamic)
      images_unmerge: DO ui = uf, ul
         image_obj => images%mats(ui)
         image => image_obj%m
         CALL set_empty_meta_index(image%index, global_indices, size_index_merged)
         IF (vunmerged(ui) .NE. 0) THEN
            !
!$OMP CRITICAL(merge_images_common)
            IF (common_unmerged(ui)) THEN
               ncommons_unmerged = ncommons_unmerged+1
               indices_common_unmerged(ncommons_unmerged) = ui
            ELSE
               common_unmerged(ui) = .TRUE.
            ENDIF
!$OMP END CRITICAL(merge_images_common)
            !
            ! Virtual coords
            IF (do_merge_rows) THEN
               image%index(dbcsr_slot_home_vprow) = my_vprow
               image%index(dbcsr_slot_home_vpcol) = my_vpcol+ui
            ELSE
               image%index(dbcsr_slot_home_vprow) = my_vprow+ui
               image%index(dbcsr_slot_home_vpcol) = my_vpcol
            ENDIF
            ! thr_c
!$          IF (PRESENT(nthreads)) THEN
!$             image%index(dbcsr_slot_thr_c) = dbcsr_num_slots+1
!$             image%index(dbcsr_slot_thr_c+1) = size_index_merged
!$          ENDIF
            !
            ! Merge images
            merge: IF (nimages_merged .GT. 1) THEN
               !
               ! Merge corresponding thread blocks for all images
               threads: DO ithread = 1, my_nthreads
                  !
                  ! Merge images per a given thread accross all images
                  ioffset = ioffset_first(ui)
                  images_merge: DO v_merged = 1, vunmerged(ui)
                     mi = vmerged(v_merged, ui)+1
                     ioffset = ioffset+1
                     offset_meta = offset_images(imeta_size, ioffset)
                     offset_data = offset_images(idata_size, ioffset)
                     ! Check for filtered images
                     IF (offset_data .LT. offset_images(idata_size, ioffset+1)) THEN
                        unmerged_indices => meta_buffer(offset_meta+dbcsr_slot_size: &
                                                        offset_meta+meta_buffer(offset_meta+dbcsr_slot_size))
                        IF (ithread .EQ. 1) THEN
                           ! nblks, nze
                           image%index(dbcsr_slot_nblks) = &
                              image%index(dbcsr_slot_nblks)+ &
                              unmerged_indices(dbcsr_slot_nblks)
                           image%index(dbcsr_slot_nze) = &
                              image%index(dbcsr_slot_nze)+ &
                              offset_images(idata_size, ioffset+1)- &
                              offset_data
                           ! threads distribution
!$                         IF (PRESENT(nthreads)) THEN
!$                            image%index(dbcsr_num_slots+1:size_index_merged) = &
!$                               image%index(dbcsr_num_slots+1:size_index_merged)+ &
!$                               unmerged_indices(dbcsr_slot_nblks+1:size_index_unmerged)
!$                         ENDIF
                        ENDIF
                        ! Copy CSR indices (3 values per block: row-index, col-index, offset)
                        ! remap indices taking in account threads and images
                        kf = size_index_unmerged+1
                        kl = unmerged_indices(dbcsr_slot_size)
!$                      IF (PRESENT(nthreads)) THEN
!$                         kf = kf+unmerged_indices(dbcsr_slot_nblks+ithread)*3
!$                         kl = size_index_unmerged+unmerged_indices(dbcsr_slot_nblks+ithread+1)*3
!$                      ENDIF
                        DO ki = kf, kl, 3
                           image%index(image%index(dbcsr_slot_size)+1) = &
                              unmerged_indices(ki)
                           image%index(image%index(dbcsr_slot_size)+2) = &
                              unmerged_indices(ki+1)
                           ! Data offset
                           image%index(image%index(dbcsr_slot_size)+3) = &
                              unmerged_indices(ki+2)+offset_data
                           ! increase meta size
                           image%index(dbcsr_slot_size) = image%index(dbcsr_slot_size)+3
                        ENDDO
                     ENDIF
                  ENDDO images_merge
               ENDDO threads
               !
               image%index(dbcsr_num_slots) = image%index(dbcsr_slot_size)
            ELSE
               ! No merging, just copy
               offset_meta = offset_images(imeta_size, ui)
               ! Check for filtered images
               offset_data = offset_images(idata_size, ui)
               IF (offset_data .LT. offset_images(idata_size, ui+1)) THEN
                  image%index(dbcsr_slot_size) = &
                     meta_buffer(offset_meta+dbcsr_slot_size)- &
                     dbcsr_slot_nblks+dbcsr_num_slots
                  image%index(dbcsr_slot_nblks) = &
                     meta_buffer(offset_meta+dbcsr_slot_nblks)
                  image%index(dbcsr_slot_nze) = &
                     offset_images(idata_size, ui+1)-offset_data
                  image%index(dbcsr_num_slots) = image%index(dbcsr_slot_size)
                  !
                  image%index(dbcsr_num_slots+1:image%index(dbcsr_num_slots)) = &
                     meta_buffer(offset_meta+dbcsr_slot_nblks+1: &
                                 offset_meta+meta_buffer(offset_meta+dbcsr_slot_size))
                  image%index(size_index_merged+3:image%index(dbcsr_num_slots):3) = &
                     image%index(size_index_merged+3:image%index(dbcsr_num_slots):3)+ &
                     offset_data
               ENDIF
            ENDIF merge
            !
            ! Reset
!$OMP CRITICAL(merge_images_reset_vlocals)
            CALL dbcsr_reset_vlocals(image_obj, imgdist, do_rows=.NOT. do_merge_rows)
!$OMP END CRITICAL(merge_images_reset_vlocals)
         ENDIF
         !
         ! Repoint index
         CALL dbcsr_repoint_index(image)
         image%valid = .TRUE.
      ENDDO images_unmerge
!$OMP END PARALLEL DO
      !
      DEALLOCATE (ioffset_first)
      CALL timestop(handle)
   END SUBROUTINE merge_images

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nrows ...
!> \param row_p ...
!> \param col_i ...
!> \param blk_p ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global ...
! **************************************************************************************************
  SUBROUTINE calc_norms_d(norms, nrows,&
       row_p, col_i, blk_p, rbs, cbs, DATA, local, local2global)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nrows
    INTEGER, DIMENSION(1:nrows+1), &
      INTENT(IN)                             :: row_p
    INTEGER, DIMENSION(*), INTENT(IN)        :: col_i, blk_p, rbs, cbs
    REAL(kind=real_8), DIMENSION(*), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(*), INTENT(IN)        :: local2global

    INTEGER                                  :: blk, bp, bpe, row, row_i, &
                                                row_size
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row_i, row, row_size, blk, bp, bpe, val) &
    !$omp          shared (nrows, local) &
    !$omp          shared (local2global, rbs, cbs, row_p, col_i, blk_p, &
    !$omp                  data, norms)
    !$omp do
    DO row_i = 1, nrows
       IF (local) THEN
          row = local2global(row_i)
       ELSE
          row = row_i
       ENDIF
       row_size = rbs(row)
       DO blk = row_p(row_i)+1, row_p(row_i+1)
          IF (blk_p(blk) .NE. 0) THEN
             bp = ABS(blk_p(blk))
             bpe = bp + row_size * cbs(col_i(blk)) - 1
             val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
          ELSE
             val = 0.0_sp
          ENDIF
          norms(blk) = val
       ENDDO
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_d

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nblks ...
!> \param blki ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global_rows ...
!> \param local2global_cols ...
! **************************************************************************************************
  SUBROUTINE calc_norms_list_d(norms, nblks,&
       blki, rbs, cbs, DATA, local, local2global_rows, local2global_cols)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nblks
    INTEGER, DIMENSION(3,nblks), INTENT(IN)  :: blki
    INTEGER, DIMENSION(:), INTENT(IN)        :: rbs, cbs
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_rows
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_cols

    INTEGER                                  :: blk, bp, bpe, row, col
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row, col, blk, bp, bpe, val) &
    !$omp          shared (local, nblks) &
    !$omp          shared (rbs, cbs, blki, &
    !$omp                  data, norms, local2global_rows, local2global_cols)
    !$omp do
    DO blk = 1, nblks
       IF (blki(3,blk) .NE. 0) THEN
          bp = blki(3,blk)
          IF (local) THEN
             row = local2global_rows(blki(1,blk))
             col = local2global_cols(blki(2,blk))
          ELSE
             row = blki(1,blk)
             col = blki(2,blk)
          ENDIF
          bpe = bp + rbs(row) * cbs(col) - 1
          val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
       ELSE
          val = 0.0_sp
       ENDIF
       norms(blk) = val
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_list_d

! **************************************************************************************************
!> \brief Calculates max norms of each cluster with minimal overhead.
!> \param meta ...
!> \param data ...
!> \param refs ...
!> \param img_map ...
!> \param img_offset ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param local_row ...
!> \param local_col ...
!> \param max_norms ...
!> \param is_left ...
!> \param is_diagonal ...
! **************************************************************************************************
  SUBROUTINE calc_max_image_norms_d(meta,data,&
     meta_diag,data_diag,&
     refs_size,refs_displ,&
     img_map,&
     img_offset,&
     row_blk_size,col_blk_size,&
     local_row,local_col,&
     slot_coo_l,&
     max_norms, is_left, off_diagonal)
  INTEGER, DIMENSION(:), TARGET, INTENT(IN) :: meta, meta_diag
  REAL(kind=real_8), DIMENSION(:), TARGET, &
       INTENT(IN)                         :: data, data_diag
  INTEGER, DIMENSION(:, :), &
       INTENT(IN)                         :: refs_size
  INTEGER, DIMENSION(:, :, :), &
       POINTER, INTENT(IN)                :: refs_displ
  INTEGER, DIMENSION(:), INTENT(IN)       :: img_map, img_offset, &
                                             row_blk_size, col_blk_size, &
                                             local_row, local_col
  INTEGER, INTENT(IN)                     :: slot_coo_l
  REAL(kind=sp), DIMENSION(:), INTENT(INOUT) :: max_norms
  LOGICAL, INTENT(IN)                        :: is_left, off_diagonal

  INTEGER, DIMENSION(:), POINTER    :: meta_p, row, col, bps
  REAL(kind=real_8), DIMENSION(:), POINTER    :: data_p
  INTEGER                           :: nblks, blk, bpe, icluster
  INTEGER, TARGET                   :: mi, ui
  INTEGER, POINTER                  :: rowi, coli
  REAL(kind=sp)                     :: max_norm

  icluster = 1
  max_norms(:) = 0.0_sp
  !
  !$omp parallel default(none) &
  !$omp          private (mi, ui, nblks, blk, row, col, bps, bpe,&
  !$omp                   rowi, coli, max_norm, meta_p, data_p) &
  !$omp          shared (max_norms, data, meta, refs_size, img_offset,&
  !$omp                  img_map, refs_displ, slot_coo_l,&
  !$omp                  row_blk_size, col_blk_size, icluster,&
  !$omp                  local_row, local_col, is_left, off_diagonal,&
  !$omp                  data_diag, meta_diag)
  IF (is_left) THEN
     rowi => mi
     coli => ui
  ELSE
     coli => mi
     rowi => ui
  ENDIF
  DO ui = 1, SIZE(refs_size,2)
     DO mi = 1, SIZE(refs_size,1)
        IF (refs_size(mi,ui).NE.0) THEN
           IF (off_diagonal.OR.(ui.NE.mi.AND.SIZE(refs_size,1).NE.1)) THEN
              meta_p => meta
              data_p => data
           ELSE
              meta_p => meta_diag
              data_p => data_diag
           ENDIF
           max_norm = 0.0_sp
           nblks = meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_nblks)
           row => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           col => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+1:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           bps => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+2:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           !$omp do
           DO blk = 1, nblks
              IF (bps(blk).NE.0) THEN
                 IF (is_left) THEN
                    bpe = bps(blk) + row_blk_size(local_row(row(blk)))*&
                         col_blk_size(local_col(img_map(col(blk)+img_offset(coli)))) - 1
                 ELSE
                    bpe = bps(blk) + row_blk_size(local_row(img_map(row(blk)+img_offset(rowi))))*&
                         col_blk_size(local_col(col(blk))) - 1
                 ENDIF
                 max_norm = MAX(max_norm,&
                      SQRT (REAL(SUM(ABS(data_p(bps(blk)+refs_displ(idata_displ,mi,ui):&
                                                bpe+refs_displ(idata_displ,mi,ui)))**2), KIND=sp)))
              ENDIF
           ENDDO
           !$omp end do
           !$omp critical(cannon_max_norm)
           max_norms(icluster) = MAX(max_norms(icluster),max_norm)
           !$omp end critical(cannon_max_norm)
           !$omp barrier
           !$omp master
           icluster = icluster + 1
           !$omp end master
        ENDIF
     ENDDO
  ENDDO
  !$omp end parallel
END SUBROUTINE calc_max_image_norms_d

! **************************************************************************************************
!> \brief Calculates norms of each cluster with minimal overhead.
!> \param buffer ...
!> \param norms ..
! **************************************************************************************************
  SUBROUTINE calc_image_norms_d(images,norms,uf,ul)
  TYPE(dbcsr_1d_array_type), INTENT(IN)    :: images
  REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT) :: norms
  INTEGER, INTENT(IN)                      :: uf, ul

  INTEGER, DIMENSION(:), POINTER    :: row, col, bps, rbs, cbs, &
                                       local_rows, local_cols
  REAL(kind=real_8), DIMENSION(:), POINTER    :: data
  INTEGER                           :: ui, blk, bpe

  !$omp parallel default(none) &
  !$omp private(ui,row,col,bps,blk,bpe,data,&
  !$omp         rbs,cbs,local_rows,local_cols) &
  !$omp shared(norms,images,uf,ul)
  DO ui=uf,ul
     IF (images%mats(ui)%m%nblks.EQ.0) CYCLE
     row => images%mats(ui)%m%coo_l(1::3)
     col => images%mats(ui)%m%coo_l(2::3)
     bps => images%mats(ui)%m%coo_l(3::3)
     rbs => array_data(images%mats(ui)%m%row_blk_size)
     cbs => array_data(images%mats(ui)%m%col_blk_size)
     local_rows => array_data(images%mats(ui)%m%local_rows)
     local_cols => array_data(images%mats(ui)%m%local_cols)
     data => dbcsr_get_data_p_d (images%mats(ui)%m%data_area)
     !$omp do
     DO blk = 1, images%mats(ui)%m%nblks
        IF (bps(blk).NE.0) THEN
           bpe = bps(blk) + rbs(local_rows(row(blk))) * cbs(local_cols(col(blk))) - 1
           norms(blk,ui) = SQRT (REAL (SUM(ABS(data(bps(blk):bpe))**2), KIND=sp))
        ELSE
           norms(blk,ui) = 0.0_sp
        ENDIF
     ENDDO
     !$omp end do
  ENDDO
  !$omp end parallel

END SUBROUTINE calc_image_norms_d

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nrows ...
!> \param row_p ...
!> \param col_i ...
!> \param blk_p ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global ...
! **************************************************************************************************
  SUBROUTINE calc_norms_z(norms, nrows,&
       row_p, col_i, blk_p, rbs, cbs, DATA, local, local2global)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nrows
    INTEGER, DIMENSION(1:nrows+1), &
      INTENT(IN)                             :: row_p
    INTEGER, DIMENSION(*), INTENT(IN)        :: col_i, blk_p, rbs, cbs
    COMPLEX(kind=real_8), DIMENSION(*), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(*), INTENT(IN)        :: local2global

    INTEGER                                  :: blk, bp, bpe, row, row_i, &
                                                row_size
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row_i, row, row_size, blk, bp, bpe, val) &
    !$omp          shared (nrows, local) &
    !$omp          shared (local2global, rbs, cbs, row_p, col_i, blk_p, &
    !$omp                  data, norms)
    !$omp do
    DO row_i = 1, nrows
       IF (local) THEN
          row = local2global(row_i)
       ELSE
          row = row_i
       ENDIF
       row_size = rbs(row)
       DO blk = row_p(row_i)+1, row_p(row_i+1)
          IF (blk_p(blk) .NE. 0) THEN
             bp = ABS(blk_p(blk))
             bpe = bp + row_size * cbs(col_i(blk)) - 1
             val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
          ELSE
             val = 0.0_sp
          ENDIF
          norms(blk) = val
       ENDDO
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_z

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nblks ...
!> \param blki ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global_rows ...
!> \param local2global_cols ...
! **************************************************************************************************
  SUBROUTINE calc_norms_list_z(norms, nblks,&
       blki, rbs, cbs, DATA, local, local2global_rows, local2global_cols)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nblks
    INTEGER, DIMENSION(3,nblks), INTENT(IN)  :: blki
    INTEGER, DIMENSION(:), INTENT(IN)        :: rbs, cbs
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_rows
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_cols

    INTEGER                                  :: blk, bp, bpe, row, col
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row, col, blk, bp, bpe, val) &
    !$omp          shared (local, nblks) &
    !$omp          shared (rbs, cbs, blki, &
    !$omp                  data, norms, local2global_rows, local2global_cols)
    !$omp do
    DO blk = 1, nblks
       IF (blki(3,blk) .NE. 0) THEN
          bp = blki(3,blk)
          IF (local) THEN
             row = local2global_rows(blki(1,blk))
             col = local2global_cols(blki(2,blk))
          ELSE
             row = blki(1,blk)
             col = blki(2,blk)
          ENDIF
          bpe = bp + rbs(row) * cbs(col) - 1
          val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
       ELSE
          val = 0.0_sp
       ENDIF
       norms(blk) = val
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_list_z

! **************************************************************************************************
!> \brief Calculates max norms of each cluster with minimal overhead.
!> \param meta ...
!> \param data ...
!> \param refs ...
!> \param img_map ...
!> \param img_offset ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param local_row ...
!> \param local_col ...
!> \param max_norms ...
!> \param is_left ...
!> \param is_diagonal ...
! **************************************************************************************************
  SUBROUTINE calc_max_image_norms_z(meta,data,&
     meta_diag,data_diag,&
     refs_size,refs_displ,&
     img_map,&
     img_offset,&
     row_blk_size,col_blk_size,&
     local_row,local_col,&
     slot_coo_l,&
     max_norms, is_left, off_diagonal)
  INTEGER, DIMENSION(:), TARGET, INTENT(IN) :: meta, meta_diag
  COMPLEX(kind=real_8), DIMENSION(:), TARGET, &
       INTENT(IN)                         :: data, data_diag
  INTEGER, DIMENSION(:, :), &
       INTENT(IN)                         :: refs_size
  INTEGER, DIMENSION(:, :, :), &
       POINTER, INTENT(IN)                :: refs_displ
  INTEGER, DIMENSION(:), INTENT(IN)       :: img_map, img_offset, &
                                             row_blk_size, col_blk_size, &
                                             local_row, local_col
  INTEGER, INTENT(IN)                     :: slot_coo_l
  REAL(kind=sp), DIMENSION(:), INTENT(INOUT) :: max_norms
  LOGICAL, INTENT(IN)                        :: is_left, off_diagonal

  INTEGER, DIMENSION(:), POINTER    :: meta_p, row, col, bps
  COMPLEX(kind=real_8), DIMENSION(:), POINTER    :: data_p
  INTEGER                           :: nblks, blk, bpe, icluster
  INTEGER, TARGET                   :: mi, ui
  INTEGER, POINTER                  :: rowi, coli
  REAL(kind=sp)                     :: max_norm

  icluster = 1
  max_norms(:) = 0.0_sp
  !
  !$omp parallel default(none) &
  !$omp          private (mi, ui, nblks, blk, row, col, bps, bpe,&
  !$omp                   rowi, coli, max_norm, meta_p, data_p) &
  !$omp          shared (max_norms, data, meta, refs_size, img_offset,&
  !$omp                  img_map, refs_displ, slot_coo_l,&
  !$omp                  row_blk_size, col_blk_size, icluster,&
  !$omp                  local_row, local_col, is_left, off_diagonal,&
  !$omp                  data_diag, meta_diag)
  IF (is_left) THEN
     rowi => mi
     coli => ui
  ELSE
     coli => mi
     rowi => ui
  ENDIF
  DO ui = 1, SIZE(refs_size,2)
     DO mi = 1, SIZE(refs_size,1)
        IF (refs_size(mi,ui).NE.0) THEN
           IF (off_diagonal.OR.(ui.NE.mi.AND.SIZE(refs_size,1).NE.1)) THEN
              meta_p => meta
              data_p => data
           ELSE
              meta_p => meta_diag
              data_p => data_diag
           ENDIF
           max_norm = 0.0_sp
           nblks = meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_nblks)
           row => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           col => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+1:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           bps => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+2:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           !$omp do
           DO blk = 1, nblks
              IF (bps(blk).NE.0) THEN
                 IF (is_left) THEN
                    bpe = bps(blk) + row_blk_size(local_row(row(blk)))*&
                         col_blk_size(local_col(img_map(col(blk)+img_offset(coli)))) - 1
                 ELSE
                    bpe = bps(blk) + row_blk_size(local_row(img_map(row(blk)+img_offset(rowi))))*&
                         col_blk_size(local_col(col(blk))) - 1
                 ENDIF
                 max_norm = MAX(max_norm,&
                      SQRT (REAL(SUM(ABS(data_p(bps(blk)+refs_displ(idata_displ,mi,ui):&
                                                bpe+refs_displ(idata_displ,mi,ui)))**2), KIND=sp)))
              ENDIF
           ENDDO
           !$omp end do
           !$omp critical(cannon_max_norm)
           max_norms(icluster) = MAX(max_norms(icluster),max_norm)
           !$omp end critical(cannon_max_norm)
           !$omp barrier
           !$omp master
           icluster = icluster + 1
           !$omp end master
        ENDIF
     ENDDO
  ENDDO
  !$omp end parallel
END SUBROUTINE calc_max_image_norms_z

! **************************************************************************************************
!> \brief Calculates norms of each cluster with minimal overhead.
!> \param buffer ...
!> \param norms ..
! **************************************************************************************************
  SUBROUTINE calc_image_norms_z(images,norms,uf,ul)
  TYPE(dbcsr_1d_array_type), INTENT(IN)    :: images
  REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT) :: norms
  INTEGER, INTENT(IN)                      :: uf, ul

  INTEGER, DIMENSION(:), POINTER    :: row, col, bps, rbs, cbs, &
                                       local_rows, local_cols
  COMPLEX(kind=real_8), DIMENSION(:), POINTER    :: data
  INTEGER                           :: ui, blk, bpe

  !$omp parallel default(none) &
  !$omp private(ui,row,col,bps,blk,bpe,data,&
  !$omp         rbs,cbs,local_rows,local_cols) &
  !$omp shared(norms,images,uf,ul)
  DO ui=uf,ul
     IF (images%mats(ui)%m%nblks.EQ.0) CYCLE
     row => images%mats(ui)%m%coo_l(1::3)
     col => images%mats(ui)%m%coo_l(2::3)
     bps => images%mats(ui)%m%coo_l(3::3)
     rbs => array_data(images%mats(ui)%m%row_blk_size)
     cbs => array_data(images%mats(ui)%m%col_blk_size)
     local_rows => array_data(images%mats(ui)%m%local_rows)
     local_cols => array_data(images%mats(ui)%m%local_cols)
     data => dbcsr_get_data_p_z (images%mats(ui)%m%data_area)
     !$omp do
     DO blk = 1, images%mats(ui)%m%nblks
        IF (bps(blk).NE.0) THEN
           bpe = bps(blk) + rbs(local_rows(row(blk))) * cbs(local_cols(col(blk))) - 1
           norms(blk,ui) = SQRT (REAL (SUM(ABS(data(bps(blk):bpe))**2), KIND=sp))
        ELSE
           norms(blk,ui) = 0.0_sp
        ENDIF
     ENDDO
     !$omp end do
  ENDDO
  !$omp end parallel

END SUBROUTINE calc_image_norms_z

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nrows ...
!> \param row_p ...
!> \param col_i ...
!> \param blk_p ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global ...
! **************************************************************************************************
  SUBROUTINE calc_norms_s(norms, nrows,&
       row_p, col_i, blk_p, rbs, cbs, DATA, local, local2global)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nrows
    INTEGER, DIMENSION(1:nrows+1), &
      INTENT(IN)                             :: row_p
    INTEGER, DIMENSION(*), INTENT(IN)        :: col_i, blk_p, rbs, cbs
    REAL(kind=real_4), DIMENSION(*), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(*), INTENT(IN)        :: local2global

    INTEGER                                  :: blk, bp, bpe, row, row_i, &
                                                row_size
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row_i, row, row_size, blk, bp, bpe, val) &
    !$omp          shared (nrows, local) &
    !$omp          shared (local2global, rbs, cbs, row_p, col_i, blk_p, &
    !$omp                  data, norms)
    !$omp do
    DO row_i = 1, nrows
       IF (local) THEN
          row = local2global(row_i)
       ELSE
          row = row_i
       ENDIF
       row_size = rbs(row)
       DO blk = row_p(row_i)+1, row_p(row_i+1)
          IF (blk_p(blk) .NE. 0) THEN
             bp = ABS(blk_p(blk))
             bpe = bp + row_size * cbs(col_i(blk)) - 1
             val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
          ELSE
             val = 0.0_sp
          ENDIF
          norms(blk) = val
       ENDDO
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_s

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nblks ...
!> \param blki ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global_rows ...
!> \param local2global_cols ...
! **************************************************************************************************
  SUBROUTINE calc_norms_list_s(norms, nblks,&
       blki, rbs, cbs, DATA, local, local2global_rows, local2global_cols)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nblks
    INTEGER, DIMENSION(3,nblks), INTENT(IN)  :: blki
    INTEGER, DIMENSION(:), INTENT(IN)        :: rbs, cbs
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_rows
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_cols

    INTEGER                                  :: blk, bp, bpe, row, col
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row, col, blk, bp, bpe, val) &
    !$omp          shared (local, nblks) &
    !$omp          shared (rbs, cbs, blki, &
    !$omp                  data, norms, local2global_rows, local2global_cols)
    !$omp do
    DO blk = 1, nblks
       IF (blki(3,blk) .NE. 0) THEN
          bp = blki(3,blk)
          IF (local) THEN
             row = local2global_rows(blki(1,blk))
             col = local2global_cols(blki(2,blk))
          ELSE
             row = blki(1,blk)
             col = blki(2,blk)
          ENDIF
          bpe = bp + rbs(row) * cbs(col) - 1
          val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
       ELSE
          val = 0.0_sp
       ENDIF
       norms(blk) = val
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_list_s

! **************************************************************************************************
!> \brief Calculates max norms of each cluster with minimal overhead.
!> \param meta ...
!> \param data ...
!> \param refs ...
!> \param img_map ...
!> \param img_offset ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param local_row ...
!> \param local_col ...
!> \param max_norms ...
!> \param is_left ...
!> \param is_diagonal ...
! **************************************************************************************************
  SUBROUTINE calc_max_image_norms_s(meta,data,&
     meta_diag,data_diag,&
     refs_size,refs_displ,&
     img_map,&
     img_offset,&
     row_blk_size,col_blk_size,&
     local_row,local_col,&
     slot_coo_l,&
     max_norms, is_left, off_diagonal)
  INTEGER, DIMENSION(:), TARGET, INTENT(IN) :: meta, meta_diag
  REAL(kind=real_4), DIMENSION(:), TARGET, &
       INTENT(IN)                         :: data, data_diag
  INTEGER, DIMENSION(:, :), &
       INTENT(IN)                         :: refs_size
  INTEGER, DIMENSION(:, :, :), &
       POINTER, INTENT(IN)                :: refs_displ
  INTEGER, DIMENSION(:), INTENT(IN)       :: img_map, img_offset, &
                                             row_blk_size, col_blk_size, &
                                             local_row, local_col
  INTEGER, INTENT(IN)                     :: slot_coo_l
  REAL(kind=sp), DIMENSION(:), INTENT(INOUT) :: max_norms
  LOGICAL, INTENT(IN)                        :: is_left, off_diagonal

  INTEGER, DIMENSION(:), POINTER    :: meta_p, row, col, bps
  REAL(kind=real_4), DIMENSION(:), POINTER    :: data_p
  INTEGER                           :: nblks, blk, bpe, icluster
  INTEGER, TARGET                   :: mi, ui
  INTEGER, POINTER                  :: rowi, coli
  REAL(kind=sp)                     :: max_norm

  icluster = 1
  max_norms(:) = 0.0_sp
  !
  !$omp parallel default(none) &
  !$omp          private (mi, ui, nblks, blk, row, col, bps, bpe,&
  !$omp                   rowi, coli, max_norm, meta_p, data_p) &
  !$omp          shared (max_norms, data, meta, refs_size, img_offset,&
  !$omp                  img_map, refs_displ, slot_coo_l,&
  !$omp                  row_blk_size, col_blk_size, icluster,&
  !$omp                  local_row, local_col, is_left, off_diagonal,&
  !$omp                  data_diag, meta_diag)
  IF (is_left) THEN
     rowi => mi
     coli => ui
  ELSE
     coli => mi
     rowi => ui
  ENDIF
  DO ui = 1, SIZE(refs_size,2)
     DO mi = 1, SIZE(refs_size,1)
        IF (refs_size(mi,ui).NE.0) THEN
           IF (off_diagonal.OR.(ui.NE.mi.AND.SIZE(refs_size,1).NE.1)) THEN
              meta_p => meta
              data_p => data
           ELSE
              meta_p => meta_diag
              data_p => data_diag
           ENDIF
           max_norm = 0.0_sp
           nblks = meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_nblks)
           row => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           col => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+1:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           bps => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+2:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           !$omp do
           DO blk = 1, nblks
              IF (bps(blk).NE.0) THEN
                 IF (is_left) THEN
                    bpe = bps(blk) + row_blk_size(local_row(row(blk)))*&
                         col_blk_size(local_col(img_map(col(blk)+img_offset(coli)))) - 1
                 ELSE
                    bpe = bps(blk) + row_blk_size(local_row(img_map(row(blk)+img_offset(rowi))))*&
                         col_blk_size(local_col(col(blk))) - 1
                 ENDIF
                 max_norm = MAX(max_norm,&
                      SQRT (REAL(SUM(ABS(data_p(bps(blk)+refs_displ(idata_displ,mi,ui):&
                                                bpe+refs_displ(idata_displ,mi,ui)))**2), KIND=sp)))
              ENDIF
           ENDDO
           !$omp end do
           !$omp critical(cannon_max_norm)
           max_norms(icluster) = MAX(max_norms(icluster),max_norm)
           !$omp end critical(cannon_max_norm)
           !$omp barrier
           !$omp master
           icluster = icluster + 1
           !$omp end master
        ENDIF
     ENDDO
  ENDDO
  !$omp end parallel
END SUBROUTINE calc_max_image_norms_s

! **************************************************************************************************
!> \brief Calculates norms of each cluster with minimal overhead.
!> \param buffer ...
!> \param norms ..
! **************************************************************************************************
  SUBROUTINE calc_image_norms_s(images,norms,uf,ul)
  TYPE(dbcsr_1d_array_type), INTENT(IN)    :: images
  REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT) :: norms
  INTEGER, INTENT(IN)                      :: uf, ul

  INTEGER, DIMENSION(:), POINTER    :: row, col, bps, rbs, cbs, &
                                       local_rows, local_cols
  REAL(kind=real_4), DIMENSION(:), POINTER    :: data
  INTEGER                           :: ui, blk, bpe

  !$omp parallel default(none) &
  !$omp private(ui,row,col,bps,blk,bpe,data,&
  !$omp         rbs,cbs,local_rows,local_cols) &
  !$omp shared(norms,images,uf,ul)
  DO ui=uf,ul
     IF (images%mats(ui)%m%nblks.EQ.0) CYCLE
     row => images%mats(ui)%m%coo_l(1::3)
     col => images%mats(ui)%m%coo_l(2::3)
     bps => images%mats(ui)%m%coo_l(3::3)
     rbs => array_data(images%mats(ui)%m%row_blk_size)
     cbs => array_data(images%mats(ui)%m%col_blk_size)
     local_rows => array_data(images%mats(ui)%m%local_rows)
     local_cols => array_data(images%mats(ui)%m%local_cols)
     data => dbcsr_get_data_p_s (images%mats(ui)%m%data_area)
     !$omp do
     DO blk = 1, images%mats(ui)%m%nblks
        IF (bps(blk).NE.0) THEN
           bpe = bps(blk) + rbs(local_rows(row(blk))) * cbs(local_cols(col(blk))) - 1
           norms(blk,ui) = SQRT (REAL (SUM(ABS(data(bps(blk):bpe))**2), KIND=sp))
        ELSE
           norms(blk,ui) = 0.0_sp
        ENDIF
     ENDDO
     !$omp end do
  ENDDO
  !$omp end parallel

END SUBROUTINE calc_image_norms_s

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nrows ...
!> \param row_p ...
!> \param col_i ...
!> \param blk_p ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global ...
! **************************************************************************************************
  SUBROUTINE calc_norms_c(norms, nrows,&
       row_p, col_i, blk_p, rbs, cbs, DATA, local, local2global)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nrows
    INTEGER, DIMENSION(1:nrows+1), &
      INTENT(IN)                             :: row_p
    INTEGER, DIMENSION(*), INTENT(IN)        :: col_i, blk_p, rbs, cbs
    COMPLEX(kind=real_4), DIMENSION(*), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(*), INTENT(IN)        :: local2global

    INTEGER                                  :: blk, bp, bpe, row, row_i, &
                                                row_size
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row_i, row, row_size, blk, bp, bpe, val) &
    !$omp          shared (nrows, local) &
    !$omp          shared (local2global, rbs, cbs, row_p, col_i, blk_p, &
    !$omp                  data, norms)
    !$omp do
    DO row_i = 1, nrows
       IF (local) THEN
          row = local2global(row_i)
       ELSE
          row = row_i
       ENDIF
       row_size = rbs(row)
       DO blk = row_p(row_i)+1, row_p(row_i+1)
          IF (blk_p(blk) .NE. 0) THEN
             bp = ABS(blk_p(blk))
             bpe = bp + row_size * cbs(col_i(blk)) - 1
             val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
          ELSE
             val = 0.0_sp
          ENDIF
          norms(blk) = val
       ENDDO
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_c

! **************************************************************************************************
!> \brief Calculates norms of the entire matrix with minimal overhead.
!> \param norms ...
!> \param nblks ...
!> \param blki ...
!> \param rbs ...
!> \param cbs ...
!> \param DATA ...
!> \param local ...
!> \param local2global_rows ...
!> \param local2global_cols ...
! **************************************************************************************************
  SUBROUTINE calc_norms_list_c(norms, nblks,&
       blki, rbs, cbs, DATA, local, local2global_rows, local2global_cols)
    REAL(kind=sp), DIMENSION(:), INTENT(OUT) :: norms
    INTEGER, INTENT(IN)                      :: nblks
    INTEGER, DIMENSION(3,nblks), INTENT(IN)  :: blki
    INTEGER, DIMENSION(:), INTENT(IN)        :: rbs, cbs
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: DATA
    LOGICAL, INTENT(IN)                      :: local
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_rows
    INTEGER, DIMENSION(:), INTENT(IN)        :: local2global_cols

    INTEGER                                  :: blk, bp, bpe, row, col
    REAL(kind=sp)                            :: val

!   ---------------------------------------------------------------------------

    !$omp parallel default(none) &
    !$omp          private (row, col, blk, bp, bpe, val) &
    !$omp          shared (local, nblks) &
    !$omp          shared (rbs, cbs, blki, &
    !$omp                  data, norms, local2global_rows, local2global_cols)
    !$omp do
    DO blk = 1, nblks
       IF (blki(3,blk) .NE. 0) THEN
          bp = blki(3,blk)
          IF (local) THEN
             row = local2global_rows(blki(1,blk))
             col = local2global_cols(blki(2,blk))
          ELSE
             row = blki(1,blk)
             col = blki(2,blk)
          ENDIF
          bpe = bp + rbs(row) * cbs(col) - 1
          val = SQRT (REAL (SUM(ABS(DATA(bp:bpe))**2), KIND=sp))
       ELSE
          val = 0.0_sp
       ENDIF
       norms(blk) = val
    ENDDO
    !$omp end do
    !$omp end parallel
  END SUBROUTINE calc_norms_list_c

! **************************************************************************************************
!> \brief Calculates max norms of each cluster with minimal overhead.
!> \param meta ...
!> \param data ...
!> \param refs ...
!> \param img_map ...
!> \param img_offset ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param local_row ...
!> \param local_col ...
!> \param max_norms ...
!> \param is_left ...
!> \param is_diagonal ...
! **************************************************************************************************
  SUBROUTINE calc_max_image_norms_c(meta,data,&
     meta_diag,data_diag,&
     refs_size,refs_displ,&
     img_map,&
     img_offset,&
     row_blk_size,col_blk_size,&
     local_row,local_col,&
     slot_coo_l,&
     max_norms, is_left, off_diagonal)
  INTEGER, DIMENSION(:), TARGET, INTENT(IN) :: meta, meta_diag
  COMPLEX(kind=real_4), DIMENSION(:), TARGET, &
       INTENT(IN)                         :: data, data_diag
  INTEGER, DIMENSION(:, :), &
       INTENT(IN)                         :: refs_size
  INTEGER, DIMENSION(:, :, :), &
       POINTER, INTENT(IN)                :: refs_displ
  INTEGER, DIMENSION(:), INTENT(IN)       :: img_map, img_offset, &
                                             row_blk_size, col_blk_size, &
                                             local_row, local_col
  INTEGER, INTENT(IN)                     :: slot_coo_l
  REAL(kind=sp), DIMENSION(:), INTENT(INOUT) :: max_norms
  LOGICAL, INTENT(IN)                        :: is_left, off_diagonal

  INTEGER, DIMENSION(:), POINTER    :: meta_p, row, col, bps
  COMPLEX(kind=real_4), DIMENSION(:), POINTER    :: data_p
  INTEGER                           :: nblks, blk, bpe, icluster
  INTEGER, TARGET                   :: mi, ui
  INTEGER, POINTER                  :: rowi, coli
  REAL(kind=sp)                     :: max_norm

  icluster = 1
  max_norms(:) = 0.0_sp
  !
  !$omp parallel default(none) &
  !$omp          private (mi, ui, nblks, blk, row, col, bps, bpe,&
  !$omp                   rowi, coli, max_norm, meta_p, data_p) &
  !$omp          shared (max_norms, data, meta, refs_size, img_offset,&
  !$omp                  img_map, refs_displ, slot_coo_l,&
  !$omp                  row_blk_size, col_blk_size, icluster,&
  !$omp                  local_row, local_col, is_left, off_diagonal,&
  !$omp                  data_diag, meta_diag)
  IF (is_left) THEN
     rowi => mi
     coli => ui
  ELSE
     coli => mi
     rowi => ui
  ENDIF
  DO ui = 1, SIZE(refs_size,2)
     DO mi = 1, SIZE(refs_size,1)
        IF (refs_size(mi,ui).NE.0) THEN
           IF (off_diagonal.OR.(ui.NE.mi.AND.SIZE(refs_size,1).NE.1)) THEN
              meta_p => meta
              data_p => data
           ELSE
              meta_p => meta_diag
              data_p => data_diag
           ENDIF
           max_norm = 0.0_sp
           nblks = meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_nblks)
           row => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           col => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+1:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           bps => meta_p(refs_displ(imeta_displ,mi,ui)+slot_coo_l+2:&
                         meta_p(refs_displ(imeta_displ,mi,ui)+dbcsr_slot_size)+&
                                refs_displ(imeta_displ,mi,ui):3)
           !$omp do
           DO blk = 1, nblks
              IF (bps(blk).NE.0) THEN
                 IF (is_left) THEN
                    bpe = bps(blk) + row_blk_size(local_row(row(blk)))*&
                         col_blk_size(local_col(img_map(col(blk)+img_offset(coli)))) - 1
                 ELSE
                    bpe = bps(blk) + row_blk_size(local_row(img_map(row(blk)+img_offset(rowi))))*&
                         col_blk_size(local_col(col(blk))) - 1
                 ENDIF
                 max_norm = MAX(max_norm,&
                      SQRT (REAL(SUM(ABS(data_p(bps(blk)+refs_displ(idata_displ,mi,ui):&
                                                bpe+refs_displ(idata_displ,mi,ui)))**2), KIND=sp)))
              ENDIF
           ENDDO
           !$omp end do
           !$omp critical(cannon_max_norm)
           max_norms(icluster) = MAX(max_norms(icluster),max_norm)
           !$omp end critical(cannon_max_norm)
           !$omp barrier
           !$omp master
           icluster = icluster + 1
           !$omp end master
        ENDIF
     ENDDO
  ENDDO
  !$omp end parallel
END SUBROUTINE calc_max_image_norms_c

! **************************************************************************************************
!> \brief Calculates norms of each cluster with minimal overhead.
!> \param buffer ...
!> \param norms ..
! **************************************************************************************************
  SUBROUTINE calc_image_norms_c(images,norms,uf,ul)
  TYPE(dbcsr_1d_array_type), INTENT(IN)    :: images
  REAL(kind=sp), DIMENSION(:, :), INTENT(INOUT) :: norms
  INTEGER, INTENT(IN)                      :: uf, ul

  INTEGER, DIMENSION(:), POINTER    :: row, col, bps, rbs, cbs, &
                                       local_rows, local_cols
  COMPLEX(kind=real_4), DIMENSION(:), POINTER    :: data
  INTEGER                           :: ui, blk, bpe

  !$omp parallel default(none) &
  !$omp private(ui,row,col,bps,blk,bpe,data,&
  !$omp         rbs,cbs,local_rows,local_cols) &
  !$omp shared(norms,images,uf,ul)
  DO ui=uf,ul
     IF (images%mats(ui)%m%nblks.EQ.0) CYCLE
     row => images%mats(ui)%m%coo_l(1::3)
     col => images%mats(ui)%m%coo_l(2::3)
     bps => images%mats(ui)%m%coo_l(3::3)
     rbs => array_data(images%mats(ui)%m%row_blk_size)
     cbs => array_data(images%mats(ui)%m%col_blk_size)
     local_rows => array_data(images%mats(ui)%m%local_rows)
     local_cols => array_data(images%mats(ui)%m%local_cols)
     data => dbcsr_get_data_p_c (images%mats(ui)%m%data_area)
     !$omp do
     DO blk = 1, images%mats(ui)%m%nblks
        IF (bps(blk).NE.0) THEN
           bpe = bps(blk) + rbs(local_rows(row(blk))) * cbs(local_cols(col(blk))) - 1
           norms(blk,ui) = SQRT (REAL (SUM(ABS(data(bps(blk):bpe))**2), KIND=sp))
        ELSE
           norms(blk,ui) = 0.0_sp
        ENDIF
     ENDDO
     !$omp end do
  ENDDO
  !$omp end parallel

END SUBROUTINE calc_image_norms_c


END MODULE dbcsr_mm_cannon
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Facade for matrix multiplication
!> \author  Ole Schuett
!> \date    10-2014
! **************************************************************************************************
MODULE dbcsr_multiply_api
   USE dbcsr_data_methods,              ONLY: dbcsr_scalar
   USE dbcsr_methods,                   ONLY: dbcsr_get_data_type
   USE dbcsr_mm_cannon,                 ONLY: dbcsr_mm_cannon_multiply,&
                                              dbcsr_multiply_clear_mempools => dbcsr_mm_cannon_clear_mempools,&
                                              dbcsr_multiply_lib_finalize  => dbcsr_mm_cannon_lib_finalize,&
                                              dbcsr_multiply_lib_init => dbcsr_mm_cannon_lib_init
   USE dbcsr_types,                     ONLY: dbcsr_obj,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE kinds,                           ONLY: int_8,&
                                              real_4,&
                                              real_8

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_multiply_api'

   PUBLIC :: dbcsr_multiply
   PUBLIC :: dbcsr_multiply_clear_mempools
   PUBLIC :: dbcsr_multiply_lib_finalize, dbcsr_multiply_lib_init

   INTERFACE dbcsr_multiply
      MODULE PROCEDURE dbcsr_mm_cannon_multiply
      MODULE PROCEDURE dbcsr_multiply_s, dbcsr_multiply_d, &
         dbcsr_multiply_c, dbcsr_multiply_z
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
   SUBROUTINE dbcsr_multiply_s(transa, transb, &
                               alpha, matrix_a, matrix_b, beta, matrix_c, &
                               first_row, last_row, first_column, last_column, first_k, last_k, &
                               retain_sparsity, filter_eps, &
                               flop)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      REAL(KIND=real_4), INTENT(IN)                      :: alpha
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      REAL(KIND=real_4), INTENT(IN)                      :: beta
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: first_row, last_row, first_column, &
                                                            last_column, first_k, last_k
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(KIND=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT), OPTIONAL         :: flop

      CALL dbcsr_mm_cannon_multiply(transa, transb, &
                                    dbcsr_scalar(alpha), matrix_a, matrix_b, dbcsr_scalar(beta), matrix_c, &
                                    first_row, last_row, first_column, last_column, first_k, last_k, &
                                    retain_sparsity, &
                                    filter_eps=filter_eps, &
                                    flop=flop)
   END SUBROUTINE dbcsr_multiply_s

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
   SUBROUTINE dbcsr_multiply_d(transa, transb, &
                               alpha, matrix_a, matrix_b, beta, matrix_c, &
                               first_row, last_row, first_column, last_column, first_k, last_k, &
                               retain_sparsity, filter_eps, &
                               flop)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      REAL(KIND=real_8), INTENT(IN)                      :: alpha
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      REAL(KIND=real_8), INTENT(IN)                      :: beta
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: first_row, last_row, first_column, &
                                                            last_column, first_k, last_k
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(KIND=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT), OPTIONAL         :: flop

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_multiply_d', &
         routineP = moduleN//':'//routineN

      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4 .AND. &
          dbcsr_get_data_type(matrix_c) .EQ. dbcsr_type_real_4) THEN
         CALL dbcsr_mm_cannon_multiply(transa, transb, &
                                       dbcsr_scalar(REAL(alpha, real_4)), matrix_a, matrix_b, &
                                       dbcsr_scalar(REAL(beta, real_4)), matrix_c, &
                                       first_row, last_row, first_column, last_column, first_k, last_k, &
                                       retain_sparsity, &
                                       filter_eps=filter_eps, &
                                       flop=flop)
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_8 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_8 .AND. &
              dbcsr_get_data_type(matrix_c) .EQ. dbcsr_type_real_8) THEN
         CALL dbcsr_mm_cannon_multiply(transa, transb, &
                                       dbcsr_scalar(alpha), matrix_a, matrix_b, dbcsr_scalar(beta), matrix_c, &
                                       first_row, last_row, first_column, last_column, first_k, last_k, &
                                       retain_sparsity, &
                                       filter_eps=filter_eps, &
                                       flop=flop)
      ELSE
         CALL cp__b("dbcsr/mm/dbcsr_multiply_api.F",147,"This combination of data types NYI")
      ENDIF
   END SUBROUTINE dbcsr_multiply_d

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
   SUBROUTINE dbcsr_multiply_c(transa, transb, &
                               alpha, matrix_a, matrix_b, beta, matrix_c, &
                               first_row, last_row, first_column, last_column, first_k, last_k, &
                               retain_sparsity, filter_eps, &
                               flop)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      COMPLEX(KIND=real_4), INTENT(IN)                   :: alpha
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      COMPLEX(KIND=real_4), INTENT(IN)                   :: beta
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: first_row, last_row, first_column, &
                                                            last_column, first_k, last_k
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(KIND=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT), OPTIONAL         :: flop

      CALL dbcsr_mm_cannon_multiply(transa, transb, &
                                    dbcsr_scalar(alpha), matrix_a, matrix_b, dbcsr_scalar(beta), matrix_c, &
                                    first_row, last_row, first_column, last_column, first_k, last_k, &
                                    retain_sparsity, &
                                    filter_eps=filter_eps, &
                                    flop=flop)
   END SUBROUTINE dbcsr_multiply_c

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
   SUBROUTINE dbcsr_multiply_z(transa, transb, &
                               alpha, matrix_a, matrix_b, beta, matrix_c, &
                               first_row, last_row, first_column, last_column, first_k, last_k, &
                               retain_sparsity, filter_eps, &
                               flop)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      COMPLEX(KIND=real_8), INTENT(IN)                   :: alpha
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      COMPLEX(KIND=real_8), INTENT(IN)                   :: beta
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL                      :: first_row, last_row, first_column, &
                                                            last_column, first_k, last_k
      LOGICAL, INTENT(IN), OPTIONAL                      :: retain_sparsity
      REAL(KIND=real_8), INTENT(IN), OPTIONAL            :: filter_eps
      INTEGER(KIND=int_8), INTENT(OUT), OPTIONAL         :: flop

      CALL dbcsr_mm_cannon_multiply(transa, transb, &
                                    dbcsr_scalar(alpha), matrix_a, matrix_b, dbcsr_scalar(beta), matrix_c, &
                                    first_row, last_row, first_column, last_column, first_k, last_k, &
                                    retain_sparsity, &
                                    filter_eps=filter_eps, &
                                    flop=flop)
   END SUBROUTINE dbcsr_multiply_z

END MODULE dbcsr_multiply_api
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Routines that affect the DBCSR library as a whole
!> \author  Ole Schuett
!> \date    10-2014
! **************************************************************************************************

MODULE dbcsr_lib
   USE dbcsr_config,                    ONLY: is_initialized
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error
   USE dbcsr_multiply_api,              ONLY: dbcsr_multiply_clear_mempools,&
                                              dbcsr_multiply_lib_finalize,&
                                              dbcsr_multiply_lib_init
   USE kinds,                           ONLY: int_1_size,&
                                              int_2_size,&
                                              int_4_size,&
                                              int_8_size
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_lib'

   PUBLIC :: dbcsr_init_lib, dbcsr_finalize_lib, dbcsr_clear_mempools

CONTAINS

! **************************************************************************************************
!> \brief Initialize the DBCSR library
!>
!> Prepares the DBCSR library for use.
! **************************************************************************************************
   SUBROUTINE dbcsr_init_lib()

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_init_lib', routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!n_stack_buffers,mem_type, n_stack_mem_regions, stack_size, nstacks,&
!INTEGER, DIMENSION(3)                    :: nxstacks
!   ---------------------------------------------------------------------------
!TODO: problem: init/finalize are called by cp2k_runs AND f77_interface

      IF (is_initialized) RETURN
      CALL timeset(routineN, error_handle)
      !
      CALL dbcsr_assert(int_1_size, "EQ", 1, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Incorrect assumption of an 8-bit integer size!", &
                        59)
      CALL dbcsr_assert(int_2_size, "EQ", 2, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Incorrect assumption of a 16-bit integer size!", &
                        63)
      CALL dbcsr_assert(int_4_size, "EQ", 4, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Incorrect assumption of a 32-bit integer size!", &
                        67)
      CALL dbcsr_assert(int_8_size, "EQ", 8, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Incorrect assumption of a 64-bit integer size!", &
                        71)

!$OMP     PARALLEL DEFAULT(NONE)
      CALL dbcsr_multiply_lib_init()
!$OMP     END PARALLEL

      is_initialized = .TRUE.
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_init_lib

! **************************************************************************************************
!> \brief Finalize the DBCSR library
!>
!> Cleans up after the DBCSR library.  Used to deallocate persistent objects.
!> \param group ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE dbcsr_finalize_lib(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_finalize_lib', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!TODO: problem: init/finalize are called by cp2k_runs AND f77_interface

      IF (.NOT. is_initialized) RETURN
      CALL timeset(routineN, error_handle)

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=output_unit, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=output_unit, FMT="(T2,A,T35,A,T80,A)") "-", "DBCSR STATISTICS", "-"
         WRITE (UNIT=output_unit, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("-", 79)
      END IF

!$OMP     PARALLEL DEFAULT(NONE) SHARED(output_unit, group)
      CALL dbcsr_multiply_lib_finalize(group, output_unit)
!$OMP     END PARALLEL
      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("-", 79)

      is_initialized = .FALSE.
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_finalize_lib

! **************************************************************************************************
!> \brief  Deallocate memory contained in mempools
! **************************************************************************************************
   SUBROUTINE dbcsr_clear_mempools()

!$OMP     PARALLEL DEFAULT(NONE)
      CALL dbcsr_multiply_clear_mempools()
!$OMP     END PARALLEL
   END SUBROUTINE dbcsr_clear_mempools

END MODULE dbcsr_lib
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Tests for CP2K DBCSR operations
!> \author  Urban Borstnik
!> \date    2010-02-08
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2010-02-08
! **************************************************************************************************
MODULE dbcsr_test_methods
   USE dbcsr_blas_operations,           ONLY: dbcsr_lapack_larnv,&
                                              set_larnv_seed
   USE dbcsr_block_access,              ONLY: dbcsr_put_block
   USE dbcsr_block_operations,          ONLY: dbcsr_block_conjg,&
                                              dbcsr_block_partial_copy,&
                                              dbcsr_block_scale,&
                                              dbcsr_block_transpose,&
                                              dbcsr_data_clear,&
                                              dbcsr_data_set
   USE dbcsr_config,                    ONLY: randmat_counter
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_get_sizes, dbcsr_data_get_type, dbcsr_data_init, &
        dbcsr_data_new, dbcsr_data_release, dbcsr_scalar, dbcsr_scalar_negative, dbcsr_scalar_one, &
        dbcsr_type_1d_to_2d, dbcsr_type_2d_to_1d
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_hold,&
                                              dbcsr_distribution_mp,&
                                              dbcsr_distribution_new,&
                                              dbcsr_distribution_release
   USE dbcsr_dist_operations,           ONLY: dbcsr_get_stored_coordinates
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_methods,                   ONLY: dbcsr_get_matrix_type,&
                                              dbcsr_init,&
                                              dbcsr_max_col_size,&
                                              dbcsr_max_row_size,&
                                              dbcsr_nblkcols_total,&
                                              dbcsr_nblkrows_total,&
                                              dbcsr_nfullcols_total,&
                                              dbcsr_nfullrows_total
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_mynode,&
                                              dbcsr_mp_new,&
                                              dbcsr_mp_numnodes,&
                                              dbcsr_mp_release
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_iterator, dbcsr_mp_obj, dbcsr_obj, &
        dbcsr_scalar_type, dbcsr_type_antihermitian, dbcsr_type_antisymmetric, &
        dbcsr_type_complex_4, dbcsr_type_complex_8, dbcsr_type_hermitian, dbcsr_type_no_symmetry, &
        dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_real_default, dbcsr_type_symmetric
   USE dbcsr_util,                      ONLY: dbcsr_verify_matrix
   USE dbcsr_work_operations,           ONLY: dbcsr_create,&
                                              dbcsr_finalize,&
                                              dbcsr_work_create
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_comm_null,&
                                              mp_environ
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: dbcsr_to_dense_local, dbcsr_impose_sparsity
   PUBLIC :: dbcsr_random_dist, dbcsr_make_random_matrix, &
             dbcsr_make_random_block_sizes, compx_to_dbcsr_scalar, &
             dbcsr_test_read_args, atoi, atol, ator

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_test_methods'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param z ...
!> \param data_type ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION compx_to_dbcsr_scalar(z, data_type) RESULT(res)
      COMPLEX(real_8)                                    :: z
      INTEGER                                            :: data_type
      TYPE(dbcsr_scalar_type)                            :: res

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         res = dbcsr_scalar(REAL(z, kind=real_4))
      CASE (dbcsr_type_real_8)
         res = dbcsr_scalar(REAL(z, kind=real_8))
      CASE (dbcsr_type_complex_4)
         res = dbcsr_scalar(CMPLX(z, kind=real_4))
      CASE (dbcsr_type_complex_8)
         res = dbcsr_scalar(z)
      END SELECT

   END FUNCTION compx_to_dbcsr_scalar

! **************************************************************************************************
!> \brief ...
!> \param narg ...
!> \param args ...
! **************************************************************************************************
   SUBROUTINE dbcsr_test_read_args(narg, args)
      INTEGER, INTENT(out)                               :: narg
      CHARACTER(len=*), DIMENSION(:), INTENT(out)        :: args

      CHARACTER(len=default_string_length)               :: line
      INTEGER                                            :: istat

      narg = 0
      DO
         READ (5, *, IOSTAT=istat) line
         IF (istat .NE. 0) EXIT
         IF (line(1:1) .EQ. '#') CYCLE
         narg = narg+1
         args(narg) = line
         !WRITE(*,*) 'we read <'//TRIM(args(narg))//'>'
      ENDDO

   END SUBROUTINE dbcsr_test_read_args

! **************************************************************************************************
!> \brief Impose sparsity on a dense matrix based on a dbcsr
!> \param[in] sparse          sparse matrix
!> \param[inout] dense        dense matrix
!>
!> Take into account the symmetry of the sparse matrix.
!> The dense matrix need to be valid.
!> The operation is done localy.
! **************************************************************************************************
   SUBROUTINE dbcsr_impose_sparsity(sparse, dense)
      TYPE(dbcsr_obj), INTENT(IN)                        :: sparse
      TYPE(dbcsr_data_obj), INTENT(inout)                :: dense

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_impose_sparsity', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: symm
      INTEGER                                            :: blk, col, col_offset, col_size, &
                                                            data_type, dense_col_size, &
                                                            dense_row_size, handle, row, &
                                                            row_offset, row_size
      LOGICAL                                            :: valid
      TYPE(dbcsr_data_obj)                               :: tmp
      TYPE(dbcsr_iterator)                               :: iter

      CALL timeset(routineN, handle)

      CALL dbcsr_data_get_sizes(dense, dense_row_size, dense_col_size, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_methods.F",165,"dense matrix not valid")
      data_type = dbcsr_data_get_type(dense)
      symm = dbcsr_get_matrix_type(sparse)

      CALL dbcsr_data_init(tmp)
      CALL dbcsr_data_new(tmp, dbcsr_type_1d_to_2d(data_type), data_size=dense_row_size, &
                          data_size2=dense_col_size)
      CALL dbcsr_data_set(dst=tmp, lb=1, data_size=dense_row_size, src=dense, source_lb=1, &
                          lb2=1, data_size2=dense_col_size, source_lb2=1)
      CALL dbcsr_data_clear(dense)

      CALL dbcsr_iterator_start(iter, sparse)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, blk, &
                                        row_size=row_size, col_size=col_size, &
                                        row_offset=row_offset, col_offset=col_offset)
         CALL dbcsr_block_partial_copy( &
            dst=dense, &
            dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.FALSE., &
            dst_r_lb=row_offset, dst_c_lb=col_offset, &
            src=tmp, &
            src_rs=dense_row_size, src_cs=dense_col_size, src_tr=.FALSE., &
            src_r_lb=row_offset, src_c_lb=col_offset, &
            nrow=row_size, ncol=col_size)
         IF (symm .NE. dbcsr_type_no_symmetry) THEN
            SELECT CASE (symm)
            CASE (dbcsr_type_symmetric)
               CALL dbcsr_block_partial_copy( &
                  dst=dense, &
                  dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                  dst_r_lb=row_offset, dst_c_lb=col_offset, &
                  src=tmp, &
                  src_rs=dense_row_size, src_cs=dense_col_size, src_tr=.FALSE., &
                  src_r_lb=row_offset, src_c_lb=col_offset, &
                  nrow=row_size, ncol=col_size)
            CASE (dbcsr_type_antisymmetric)
               CALL dbcsr_block_partial_copy( &
                  dst=dense, &
                  dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                  dst_r_lb=row_offset, dst_c_lb=col_offset, &
                  src=tmp, &
                  src_rs=dense_row_size, src_cs=dense_col_size, src_tr=.FALSE., &
                  src_r_lb=row_offset, src_c_lb=col_offset, &
                  nrow=row_size, ncol=col_size)
               CALL dbcsr_block_scale(dense, dbcsr_scalar_negative(dbcsr_scalar_one( &
                                                                   dbcsr_type_2d_to_1d(data_type))), &
                                      row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE (dbcsr_type_hermitian)
               CALL dbcsr_block_partial_copy( &
                  dst=dense, &
                  dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                  dst_r_lb=row_offset, dst_c_lb=col_offset, &
                  src=tmp, &
                  src_rs=dense_row_size, src_cs=dense_col_size, src_tr=.FALSE., &
                  src_r_lb=row_offset, src_c_lb=col_offset, &
                  nrow=row_size, ncol=col_size)
               CALL dbcsr_block_conjg(dense, row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE (dbcsr_type_antihermitian)
               CALL dbcsr_block_partial_copy( &
                  dst=dense, &
                  dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                  dst_r_lb=row_offset, dst_c_lb=col_offset, &
                  src=tmp, &
                  src_rs=dense_row_size, src_cs=dense_col_size, src_tr=.FALSE., &
                  src_r_lb=row_offset, src_c_lb=col_offset, &
                  nrow=row_size, ncol=col_size)
               CALL dbcsr_block_scale(dense, dbcsr_scalar_negative(dbcsr_scalar_one( &
                                                                   dbcsr_type_2d_to_1d(data_type))), &
                                      row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
               CALL dbcsr_block_conjg(dense, row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE DEFAULT
               CALL cp__b("dbcsr/dbcsr_test_methods.F",240,"wrong matrix symmetry")
            END SELECT
         ENDIF
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      CALL dbcsr_data_release(tmp)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_impose_sparsity

! **************************************************************************************************
!> \brief Convert a sparse matrix to a dense matrix
!> \param[in] sparse          sparse matrix
!> \param[inout] dense        dense matrix
!>
!> Take into account the symmetry of the sparse matrix.
!> The dense matrix need to be valid.
!> The operation is done localy.
! **************************************************************************************************
   SUBROUTINE dbcsr_to_dense_local(sparse, dense)
      TYPE(dbcsr_obj), INTENT(in)                        :: sparse
      TYPE(dbcsr_data_obj), INTENT(inout)                :: dense

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_to_dense_local', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: symm
      INTEGER                                            :: col, col_offset, col_size, data_type, &
                                                            dense_col_size, dense_row_size, &
                                                            handle, row, row_offset, row_size
      LOGICAL                                            :: tr, valid
      TYPE(dbcsr_data_obj)                               :: block
      TYPE(dbcsr_iterator)                               :: iter

      CALL timeset(routineN, handle)

      CALL dbcsr_data_get_sizes(dense, dense_row_size, dense_col_size, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_methods.F",280,"dense matrix not valid")

      symm = dbcsr_get_matrix_type(sparse)
      data_type = dbcsr_data_get_type(dense)

      CALL dbcsr_data_clear(dense)
      CALL dbcsr_data_init(block)
      CALL dbcsr_data_new(block, dbcsr_type_1d_to_2d(data_type))
      CALL dbcsr_iterator_start(iter, sparse)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, block, tr, &
                                        row_size=row_size, col_size=col_size, &
                                        row_offset=row_offset, col_offset=col_offset)
         CALL dbcsr_block_partial_copy(dst=dense, &
                                       dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.FALSE., &
                                       dst_r_lb=row_offset, dst_c_lb=col_offset, &
                                       src=block, src_rs=row_size, src_cs=col_size, src_tr=tr, &
                                       src_r_lb=1, src_c_lb=1, nrow=row_size, ncol=col_size)
         IF (symm .NE. dbcsr_type_no_symmetry .AND. row .NE. col) THEN
            SELECT CASE (symm)
            CASE (dbcsr_type_symmetric)
               CALL dbcsr_block_partial_copy(dst=dense, &
                                             dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                                             dst_r_lb=row_offset, dst_c_lb=col_offset, &
                                             src=block, src_rs=row_size, src_cs=col_size, src_tr=tr, &
                                             src_r_lb=1, src_c_lb=1, nrow=row_size, ncol=col_size)
            CASE (dbcsr_type_antisymmetric)
               CALL dbcsr_block_partial_copy(dst=dense, &
                                             dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                                             dst_r_lb=row_offset, dst_c_lb=col_offset, &
                                             src=block, src_rs=row_size, src_cs=col_size, src_tr=tr, &
                                             src_r_lb=1, src_c_lb=1, nrow=row_size, ncol=col_size)
               CALL dbcsr_block_scale(dense, dbcsr_scalar_negative(dbcsr_scalar_one( &
                                                                   dbcsr_type_2d_to_1d(data_type))), &
                                      row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE (dbcsr_type_hermitian)
               CALL dbcsr_block_partial_copy(dst=dense, &
                                             dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                                             dst_r_lb=row_offset, dst_c_lb=col_offset, &
                                             src=block, src_rs=row_size, src_cs=col_size, src_tr=tr, &
                                             src_r_lb=1, src_c_lb=1, nrow=row_size, ncol=col_size)
               CALL dbcsr_block_conjg(dense, row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE (dbcsr_type_antihermitian)
               CALL dbcsr_block_partial_copy(dst=dense, &
                                             dst_rs=dense_row_size, dst_cs=dense_col_size, dst_tr=.TRUE., &
                                             dst_r_lb=row_offset, dst_c_lb=col_offset, &
                                             src=block, src_rs=row_size, src_cs=col_size, src_tr=tr, &
                                             src_r_lb=1, src_c_lb=1, nrow=row_size, ncol=col_size)
               CALL dbcsr_block_scale(dense, dbcsr_scalar_negative(dbcsr_scalar_one( &
                                                                   dbcsr_type_2d_to_1d(data_type))), &
                                      row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
               CALL dbcsr_block_conjg(dense, row_size=col_size, col_size=row_size, &
                                      lb=col_offset, lb2=row_offset)
            CASE DEFAULT
               CALL cp__b("dbcsr/dbcsr_test_methods.F",337,"wrong matrix symmetry")
            END SELECT
         ENDIF
      END DO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(block)
      CALL dbcsr_data_release(block)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_to_dense_local

! **************************************************************************************************
!> \brief ...
!> \param dist_array ...
!> \param dist_size ...
!> \param nbins ...
! **************************************************************************************************
   SUBROUTINE dbcsr_random_dist(dist_array, dist_size, nbins)
      INTEGER, DIMENSION(:), INTENT(out), POINTER        :: dist_array
      INTEGER, INTENT(in)                                :: dist_size, nbins

      INTEGER                                            :: i

      ALLOCATE (dist_array(dist_size))
      !CALL RANDOM_NUMBER (dist_array)
      DO i = 1, dist_size
         dist_array(i) = MODULO(nbins-i, nbins)
      END DO
   END SUBROUTINE dbcsr_random_dist

!> \brief Creates a random matrix.
!> \param matrix ...
!> \param row_blk_sizes ...
!> \param col_blk_sizes ...
!> \param name ...
!> \param sparsity ...
!> \param mp_group ...
!> \param data_type ...
!> \param symmetry ...
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_make_random_matrix(matrix, row_blk_sizes, col_blk_sizes, &
                                       name, sparsity, mp_group, data_type, symmetry, dist)
      TYPE(dbcsr_obj), INTENT(out)                       :: matrix
      INTEGER, DIMENSION(:), INTENT(in), POINTER         :: row_blk_sizes, col_blk_sizes
      CHARACTER(len=*), INTENT(in)                       :: name
      REAL(kind=real_8), INTENT(in)                      :: sparsity
      INTEGER, INTENT(in)                                :: mp_group
      INTEGER, INTENT(in), OPTIONAL                      :: data_type
      CHARACTER, INTENT(in), OPTIONAL                    :: symmetry
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: dist

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_random_matrix', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: my_symmetry
      INTEGER                                            :: col, error_handle, max_nze, &
                                                            my_data_type, my_proc, ncol, nrow, &
                                                            numproc, nze, p, row, s_col, s_row
      INTEGER(KIND=int_8)                                :: counter, ele, increment, nmax
      INTEGER, DIMENSION(4)                              :: iseed, jseed
      LOGICAL                                            :: tr
      REAL(kind=real_8)                                  :: my_sparsity
      REAL(kind=real_8), DIMENSION(1)                    :: value
      TYPE(dbcsr_data_obj)                               :: data_values, data_values_tr
      TYPE(dbcsr_distribution_obj)                       :: new_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      ! the counter goes into the seed. Every new call gives a new random matrix
      randmat_counter = randmat_counter+1
      ! Create the matrix
      IF (PRESENT(dist)) THEN
         new_dist = dist
         CALL dbcsr_distribution_hold(new_dist)
      ELSE
         CALL dbcsr_make_null_dist(new_dist, SIZE(row_blk_sizes), &
                                   SIZE(col_blk_sizes), group=mp_group)
      ENDIF
      CALL dbcsr_init(matrix)
      my_data_type = dbcsr_type_real_default
      IF (PRESENT(data_type)) my_data_type = data_type
      my_symmetry = dbcsr_type_no_symmetry
      IF (PRESENT(symmetry)) my_symmetry = symmetry
      CALL dbcsr_create(matrix, name, &
                        new_dist, my_symmetry, &
                        row_blk_sizes, &
                        col_blk_sizes, &
                        data_type=my_data_type)
      numproc = dbcsr_mp_numnodes(dbcsr_distribution_mp(new_dist))
      my_proc = dbcsr_mp_mynode(dbcsr_distribution_mp(new_dist))
      !
      IF (sparsity .GT. 1) THEN
         my_sparsity = sparsity/100.0
      ELSE
         my_sparsity = sparsity
      ENDIF
      CALL dbcsr_work_create(matrix, &
                             nblks_guess=INT(REAL(dbcsr_nblkrows_total(matrix), KIND=dp) &
                                             *REAL(dbcsr_nblkcols_total(matrix), KIND=dp) &
                                             *(1.0_dp-sparsity)*1.1_dp/numproc), &
                             sizedata_guess=INT(REAL(dbcsr_nfullrows_total(matrix), KIND=dp) &
                                                *REAL(dbcsr_nfullcols_total(matrix), KIND=dp) &
                                                *(1.0_dp-sparsity)*1.1_dp/numproc), &
                             work_mutable=.TRUE.)

      max_nze = dbcsr_max_row_size(matrix)*dbcsr_max_col_size(matrix)
      CALL dbcsr_data_init(data_values)
      CALL dbcsr_data_new(data_values, my_data_type, data_size=max_nze)
      CALL dbcsr_data_init(data_values_tr)
      CALL dbcsr_data_new(data_values_tr, my_data_type, data_size=max_nze)

      nrow = dbcsr_nblkrows_total(matrix)
      ncol = dbcsr_nblkcols_total(matrix)
      nmax = INT(nrow, KIND=int_8)*INT(ncol, KIND=int_8)
      ele = -1
      counter = 0
      CALL set_larnv_seed(7, 42, 3, 42, randmat_counter, jseed)

      DO
         ! find the next block to add, this is given by a geometrically distributed variable
         ! we number the blocks of the matrix and jump to the next one
         CALL dlarnv(1, jseed, 1, value)
         IF (my_sparsity > 0) THEN
            increment = 1+FLOOR(LOG(value(1))/LOG(my_sparsity), KIND=int_8)
         ELSE
            increment = 1
         ENDIF
         ele = ele+increment
         IF (ele >= nmax) EXIT
         counter = counter+1
         row = INT(ele/ncol)+1
         col = INT(MOD(ele, INT(ncol, KIND=KIND(ele))))+1

         ! build the upper matrix if some symmetry, and only deal with the local blocks.
         s_row = row; s_col = col
         IF (PRESENT(dist)) THEN
            tr = .FALSE.
            CALL dbcsr_get_stored_coordinates(matrix, s_row, s_col, p)
            IF (my_symmetry .NE. dbcsr_type_no_symmetry .AND. s_col .LT. s_row) CYCLE
            IF (p .NE. my_proc) CYCLE
         ELSE
            IF (my_symmetry .NE. dbcsr_type_no_symmetry .AND. s_col .LT. s_row) CYCLE
         ENDIF
         IF (.NOT. PRESENT(dist) .AND. my_proc .NE. 0) CYCLE

         ! fill based on a block based seed, makes this the same values in parallel
         CALL set_larnv_seed(row, nrow, col, ncol, randmat_counter, iseed)
         nze = row_blk_sizes(s_row)*col_blk_sizes(s_col)
         CALL dbcsr_lapack_larnv(1, iseed, nze, data_values)
         CALL dbcsr_put_block(matrix, s_row, s_col, data_values)
         IF (my_symmetry .NE. dbcsr_type_no_symmetry .AND. s_col .EQ. s_row) THEN
            SELECT CASE (my_symmetry)
            CASE (dbcsr_type_symmetric)
               CALL dbcsr_block_transpose(data_values_tr, data_values, &
                                          row_size=row_blk_sizes(s_row), col_size=col_blk_sizes(s_col), lb=1, source_lb=1)
            CASE (dbcsr_type_antisymmetric)
               CALL dbcsr_block_transpose(data_values_tr, data_values, &
                                          row_size=row_blk_sizes(s_row), col_size=col_blk_sizes(s_col), lb=1, source_lb=1, &
                                          scale=dbcsr_scalar_negative(dbcsr_scalar_one(my_data_type)))
            CASE (dbcsr_type_hermitian)
               CALL dbcsr_block_transpose(data_values_tr, data_values, &
                                          row_size=row_blk_sizes(s_row), col_size=col_blk_sizes(s_col), lb=1, source_lb=1)
               CALL dbcsr_block_conjg(data_values_tr, row_size=col_blk_sizes(s_col), col_size=row_blk_sizes(s_row), &
                                      lb=1)
            CASE (dbcsr_type_antihermitian)
               CALL dbcsr_block_transpose(data_values_tr, data_values, &
                                          row_size=row_blk_sizes(s_row), col_size=col_blk_sizes(s_col), lb=1, source_lb=1, &
                                          scale=dbcsr_scalar_negative(dbcsr_scalar_one(my_data_type)))
               CALL dbcsr_block_conjg(data_values_tr, row_size=col_blk_sizes(s_col), col_size=row_blk_sizes(s_row), &
                                      lb=1)
            CASE DEFAULT
               CALL cp__b("dbcsr/dbcsr_test_methods.F",511,"wrong matrix symmetry")
            END SELECT
            CALL dbcsr_put_block(matrix, s_row, s_col, data_values_tr, summation=.TRUE.)
         ENDIF
      ENDDO

      CALL dbcsr_data_release(data_values)
      CALL dbcsr_data_release(data_values_tr)

      CALL dbcsr_distribution_release(new_dist)
      CALL dbcsr_finalize(matrix)
      CALL dbcsr_verify_matrix(matrix)
      !
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_make_random_matrix

! **************************************************************************************************
!> \brief ...
!> \param block_sizes ...
!> \param size_sum ...
!> \param size_mix ...
! **************************************************************************************************
   SUBROUTINE dbcsr_make_random_block_sizes(block_sizes, size_sum, size_mix)
      INTEGER, DIMENSION(:), INTENT(out), POINTER        :: block_sizes
      INTEGER, INTENT(in)                                :: size_sum
      INTEGER, DIMENSION(:), INTENT(in)                  :: size_mix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_random_block_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: block_size, current_sum, nblocks, &
                                                            nsize_mix, selector
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: mixer
      INTEGER, DIMENSION(:), POINTER                     :: sizes

!

      NULLIFY (sizes)
      nsize_mix = SIZE(size_mix)/2
      ALLOCATE (mixer(3, nsize_mix))
      mixer(1, :) = size_mix(1:nsize_mix*2-1:2)
      mixer(2, :) = size_mix(2:nsize_mix*2:2)
      mixer(3, :) = 1
      nblocks = 0
      current_sum = 0
      CALL ensure_array_size(sizes, lb=1, ub=1)

      selector = 1
      !
      DO WHILE (current_sum .LT. size_sum)
         nblocks = nblocks+1
         !CALL RANDOM_NUMBER(value)
         !block_size = MIN (INT (value(1) * size_max),&
         !                  size_sum - current_sum)
         block_size = MIN(mixer(2, selector), &
                          size_sum-current_sum)
         sizes(nblocks) = block_size
         current_sum = current_sum+block_size
         CALL ensure_array_size(sizes, ub=nblocks+1, factor=2.0_dp)
         mixer(3, selector) = mixer(3, selector)+1
         IF (mixer(3, selector) .GT. mixer(1, selector)) THEN
            mixer(3, selector) = 1
            selector = MOD(selector, nsize_mix)+1
         ENDIF
      ENDDO
      ALLOCATE (block_sizes(nblocks))
      block_sizes = sizes(1:nblocks)
      current_sum = SUM(block_sizes)
      CALL dbcsr_assert(current_sum, "EQ", size_sum, dbcsr_fatal_level, &
                        dbcsr_internal_error, routineN, "Incorrect block sizes", 580)
      DEALLOCATE (mixer, sizes)

   END SUBROUTINE dbcsr_make_random_block_sizes

! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \param group ...
! **************************************************************************************************
   SUBROUTINE dbcsr_make_null_mp(mp_env, group)
      TYPE(dbcsr_mp_obj), INTENT(out)                    :: mp_env
      INTEGER, INTENT(in), OPTIONAL                      :: group

      INTEGER                                            :: mynode, numnodes

      IF (PRESENT(group)) THEN
         CALL mp_environ(numnodes, mynode, group)
         CALL dbcsr_mp_new(mp_env, RESHAPE((/1/), (/1, 1/)), &
                           group, mynode, numnodes, &
                           myprow=0, mypcol=0)
      ELSE
         CALL dbcsr_mp_new(mp_env, RESHAPE((/1/), (/1, 1/)), &
                           MP_COMM_NULL, 0, 1, &
                           myprow=0, mypcol=0)
      ENDIF
   END SUBROUTINE dbcsr_make_null_mp
   !
! **************************************************************************************************
!> \brief ...
!> \param distribution ...
!> \param nblkrows ...
!> \param nblkcols ...
!> \param group ...
! **************************************************************************************************
   SUBROUTINE dbcsr_make_null_dist(distribution, nblkrows, nblkcols, group)
      TYPE(dbcsr_distribution_obj), INTENT(out)          :: distribution
      INTEGER, INTENT(in)                                :: nblkrows, nblkcols
      INTEGER, INTENT(in), OPTIONAL                      :: group

      INTEGER, DIMENSION(:), POINTER                     :: col_dist, row_dist
      TYPE(dbcsr_mp_obj)                                 :: mp_env

      CALL dbcsr_make_null_mp(mp_env, group=group)
      ALLOCATE (row_dist(nblkrows), col_dist(nblkcols))
      row_dist = 0
      col_dist = 0
      CALL dbcsr_distribution_new(distribution, mp_env, &
                                  row_dist, col_dist, reuse_arrays=.TRUE.)
      CALL dbcsr_mp_release(mp_env)
   END SUBROUTINE dbcsr_make_null_dist

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval atoi ...
! **************************************************************************************************
   FUNCTION atoi(a)
      CHARACTER(len=*), INTENT(in)                       :: a
      INTEGER                                            :: atoi

      READ (a, '(I6)') atoi
   END FUNCTION atoi

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval atol ...
! **************************************************************************************************
   FUNCTION atol(a)
      CHARACTER(len=*), INTENT(in)                       :: a
      LOGICAL                                            :: atol

      READ (a, '(L1)') atol
   END FUNCTION atol

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval ator ...
! **************************************************************************************************
   FUNCTION ator(a)
      CHARACTER(len=*), INTENT(in)                       :: a
      REAL(real_8)                                       :: ator

      READ (a, '(E26.15)') ator
   END FUNCTION ator

END MODULE dbcsr_test_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief provides a unified interface to lapack geev routines
!> \par History
!>       2014.09 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_geev
   USE kinds,                           ONLY: real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_geev'

   PUBLIC :: dbcsr_general_local_diag, dbcsr_tridiag_local_diag, dbcsr_symm_local_diag

   INTERFACE dbcsr_general_local_diag
      MODULE PROCEDURE dbcsr_sgeev, dbcsr_dgeev, dbcsr_zgeev, dbcsr_cgeev
   END INTERFACE

   ! currently only specialzed for real matrices
   INTERFACE dbcsr_tridiag_local_diag
      MODULE PROCEDURE dbcsr_sstev, dbcsr_dstev, dbcsr_zgeev, dbcsr_cgeev
   END INTERFACE

   ! currently only specialzed for real matrices
   INTERFACE dbcsr_symm_local_diag
      MODULE PROCEDURE dbcsr_dsyevd, dbcsr_ssyevd, dbcsr_cheevd, dbcsr_zheevd
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_zheevd(jobvr, matrix, ndim, evals, revec)
      CHARACTER(1)                                       :: jobvr
      COMPLEX(real_8), DIMENSION(:, :)                   :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_8), DIMENSION(:)                      :: evals
      COMPLEX(real_8), DIMENSION(:, :)                   :: revec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_zheevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, liwork, lrwork, lwork, &
                                                            iwork(3+5*ndim)
      COMPLEX(real_8)                                    :: work(2*ndim+ndim**2), &
                                                            tmp_array(ndim, ndim)
      REAL(real_8)                                       :: rwork(1+5*ndim+2*ndim**2)

      tmp_array(:, :) = matrix(:, :)
      lwork = 2*ndim+ndim**2
      lrwork = 1+5*ndim+2*ndim**2
      liwork = 3+5*ndim

      CALL zheevd(jobvr, 'U', ndim, tmp_array, evals, ndim, work, lwork, rwork, lrwork, iwork, liwork, info)

      DO i = 1, ndim
         revec(:, i) = tmp_array(:, i)
      END DO

   END SUBROUTINE dbcsr_zheevd

! **************************************************************************************************
!> \brief ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_cheevd(jobvr, matrix, ndim, evals, revec)
      CHARACTER(1)                                       :: jobvr
      COMPLEX(real_4), DIMENSION(:, :)                   :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_4), DIMENSION(:)                      :: evals
      COMPLEX(real_4), DIMENSION(:, :)                   :: revec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_cheevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, liwork, lrwork, lwork, &
                                                            iwork(3+5*ndim)
      COMPLEX(real_4)                                    :: work(2*ndim+ndim**2), &
                                                            tmp_array(ndim, ndim)
      REAL(real_4)                                       :: rwork(1+5*ndim+2*ndim**2)

      tmp_array(:, :) = matrix(:, :)
      lwork = 2*ndim+ndim**2
      lrwork = 1+5*ndim+2*ndim**2
      liwork = 3+5*ndim

      CALL zheevd(jobvr, 'U', ndim, tmp_array, evals, ndim, work, lwork, rwork, lrwork, iwork, liwork, info)

      DO i = 1, ndim
         revec(:, i) = tmp_array(:, i)
      END DO

   END SUBROUTINE dbcsr_cheevd

! **************************************************************************************************
!> \brief ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_dsyevd(jobvr, matrix, ndim, evals, revec)
      CHARACTER(1)                                       :: jobvr
      REAL(real_8), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_8), DIMENSION(:)                      :: evals
      COMPLEX(real_8), DIMENSION(:, :)                   :: revec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_dsyevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, liwork, lwork, iwork(3+5*ndim)
      REAL(real_8)                                       :: tmp_array(ndim, ndim), &
                                                            work(1+6*ndim+2*ndim**2)
      REAL(real_8), DIMENSION(ndim)                      :: eval

      lwork = 1+6*ndim+2*ndim**2
      liwork = 3+5*ndim

      tmp_array(:, :) = matrix(:, :)
      CALL dsyevd(jobvr, "U", ndim, tmp_array, ndim, eval, work, lwork, iwork, liwork, info)

      DO i = 1, ndim
         revec(:, i) = CMPLX(tmp_array(:, i), REAL(0.0, real_8), real_8)
         evals(i) = CMPLX(eval(i), 0.0, real_8)
      END DO

   END SUBROUTINE dbcsr_dsyevd

! **************************************************************************************************
!> \brief ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_ssyevd(jobvr, matrix, ndim, evals, revec)
      CHARACTER(1)                                       :: jobvr
      REAL(real_4), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_4), DIMENSION(:)                      :: evals
      COMPLEX(real_4), DIMENSION(:, :)                   :: revec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_ssyevd', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, liwork, lwork, iwork(3+5*ndim)
      REAL(real_4)                                       :: tmp_array(ndim, ndim), &
                                                            work(1+6*ndim+2*ndim**2)
      REAL(real_4), DIMENSION(ndim)                      :: eval

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvr))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work
      lwork = 1+6*ndim+2*ndim**2
      liwork = 3+5*ndim

      tmp_array(:, :) = matrix(:, :)
      CALL ssyevd("V", "U", ndim, tmp_array, ndim, eval, work, lwork, iwork, liwork, info)

      DO i = 1, ndim
         revec(:, i) = CMPLX(tmp_array(:, i), REAL(0.0, real_4), real_4)
         evals(i) = CMPLX(eval(i), 0.0, real_4)
      END DO

   END SUBROUTINE dbcsr_ssyevd

! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_sstev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      REAL(real_4), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_4), DIMENSION(:)                      :: evals
      COMPLEX(real_4), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sstev', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info
      REAL(real_4)                                       :: work(20*ndim)
      REAL(real_4), DIMENSION(ndim)                      :: diag, offdiag
      REAL(real_4), DIMENSION(ndim, ndim)                :: evec_r

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvl))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work

      levec(1, 1) = CMPLX(0.0, 0.0, real_4)
      info = 0
      diag(ndim) = matrix(ndim, ndim)
      DO i = 1, ndim-1
         diag(i) = matrix(i, i)
         offdiag(i) = matrix(i+1, i)
      END DO

      CALL sstev(jobvr, ndim, diag, offdiag, evec_r, ndim, work, info)

      DO i = 1, ndim
         revec(:, i) = CMPLX(evec_r(:, i), REAL(0.0, real_4), real_4)
         evals(i) = CMPLX(diag(i), 0.0, real_4)
      END DO
   END SUBROUTINE dbcsr_sstev

! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_dstev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      REAL(real_8), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_8), DIMENSION(:)                      :: evals
      COMPLEX(real_8), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_dstev', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info
      REAL(real_8)                                       :: work(20*ndim)
      REAL(real_8), DIMENSION(ndim)                      :: diag, offdiag
      REAL(real_8), DIMENSION(ndim, ndim)                :: evec_r

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvl))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work

      levec(1, 1) = CMPLX(0.0, 0.0, real_8)
      info = 0
      diag(ndim) = matrix(ndim, ndim)
      DO i = 1, ndim-1
         diag(i) = matrix(i, i)
         offdiag(i) = matrix(i+1, i)

      END DO

      CALL dstev(jobvr, ndim, diag, offdiag, evec_r, ndim, work, info)

      DO i = 1, ndim
         revec(:, i) = CMPLX(evec_r(:, i), REAL(0.0, real_8), real_8)
         evals(i) = CMPLX(diag(i), 0.0, real_8)
      END DO
   END SUBROUTINE dbcsr_dstev
! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_sgeev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      REAL(real_4), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_4), DIMENSION(:)                      :: evals
      COMPLEX(real_4), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sgeev', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, lwork
      LOGICAL                                            :: selects(ndim)
      REAL(real_4)                                       :: norm, tmp_array(ndim, ndim), &
                                                            work(20*ndim)
      REAL(real_4), DIMENSION(ndim)                      :: eval1, eval2
      REAL(real_4), DIMENSION(ndim, ndim)                :: evec_l, evec_r

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvr))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvl))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work

      eval1 = REAL(0.0, real_4); eval2 = REAL(0.0, real_4)
      tmp_array(:, :) = matrix(:, :)
      ! ask lapack how much space it would like in the work vector, don't ask me why
      lwork = -1
      CALL shseqr('S', 'I', ndim, 1, ndim, tmp_array, ndim, eval1, eval2, evec_r, ndim, work, lwork, info)

      lwork = MIN(20*ndim, INT(work(1)))
      CALL shseqr('S', 'I', ndim, 1, ndim, tmp_array, ndim, eval1, eval2, evec_r, ndim, work, lwork, info)
      CALL strevc('R', 'B', selects, ndim, tmp_array, ndim, evec_l, ndim, evec_r, ndim, ndim, ndim, work, info)

      ! compose the eigenvectors, lapacks way of storing them is a pain
      ! if eval is complex, then the complex conj pair of evec can be constructed from the i and i+1st evec
      ! Unfortunately dtrevc computes the ev such that the largest is set to one and not normalized
      i = 1
      DO WHILE (i .LE. ndim)
         IF (ABS(eval2(i)) .LT. EPSILON(REAL(0.0, real_4))) THEN
            evec_r(:, i) = evec_r(:, i)/SQRT(DOT_PRODUCT(evec_r(:, i), evec_r(:, i)))
            revec(:, i) = CMPLX(evec_r(:, i), REAL(0.0, real_4), real_4)
            levec(:, i) = CMPLX(evec_l(:, i), REAL(0.0, real_4), real_4)
            i = i+1
         ELSE IF (eval2(i) .GT. EPSILON(REAL(0.0, real_4))) THEN
            norm = SQRT(SUM(evec_r(:, i)**2.0_real_4)+SUM(evec_r(:, i+1)**2.0_real_4))
            revec(:, i) = CMPLX(evec_r(:, i), evec_r(:, i+1), real_4)/norm
            revec(:, i+1) = CMPLX(evec_r(:, i), -evec_r(:, i+1), real_4)/norm
            levec(:, i) = CMPLX(evec_l(:, i), evec_l(:, i+1), real_4)
            levec(:, i+1) = CMPLX(evec_l(:, i), -evec_l(:, i+1), real_4)
            i = i+2
         ELSE
            CALL cp__b("dbcsr/arnoldi/dbcsr_geev.F",326,'something went wrong while sorting the EV in dbcsr_geev')
         END IF
      END DO

      ! this is to keep the interface consistent with complex geev
      DO i = 1, ndim
         evals(i) = CMPLX(eval1(i), eval2(i), real_4)
      END DO

   END SUBROUTINE dbcsr_sgeev

! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_dgeev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      REAL(real_8), DIMENSION(:, :)                      :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_8), DIMENSION(:)                      :: evals
      COMPLEX(real_8), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_dgeev', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, lwork
      LOGICAL                                            :: selects(ndim)
      REAL(real_8)                                       :: norm, tmp_array(ndim, ndim), &
                                                            work(20*ndim)
      REAL(real_8), DIMENSION(ndim)                      :: eval1, eval2
      REAL(real_8), DIMENSION(ndim, ndim)                :: evec_l, evec_r

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvr))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(jobvl))==-1) EXIT ;  END DO ; ENDIF !the argument has to be here for the template to work

      eval1 = REAL(0.0, real_8); eval2 = REAL(0.0, real_8)
      tmp_array(:, :) = matrix(:, :)
      ! ask lapack how much space it would like in the work vector, don't ask me why
      lwork = -1
      CALL dhseqr('S', 'I', ndim, 1, ndim, tmp_array, ndim, eval1, eval2, evec_r, ndim, work, lwork, info)

      lwork = MIN(20*ndim, INT(work(1)))
      CALL dhseqr('S', 'I', ndim, 1, ndim, tmp_array, ndim, eval1, eval2, evec_r, ndim, work, lwork, info)
      CALL dtrevc('R', 'B', selects, ndim, tmp_array, ndim, evec_l, ndim, evec_r, ndim, ndim, ndim, work, info)

      ! compose the eigenvectors, lapacks way of storing them is a pain
      ! if eval is complex, then the complex conj pair of evec can be constructed from the i and i+1st evec
      ! Unfortunately dtrevc computes the ev such that the largest is set to one and not normalized
      i = 1
      DO WHILE (i .LE. ndim)
         IF (ABS(eval2(i)) .LT. EPSILON(REAL(0.0, real_8))) THEN
            evec_r(:, i) = evec_r(:, i)/SQRT(DOT_PRODUCT(evec_r(:, i), evec_r(:, i)))
            revec(:, i) = CMPLX(evec_r(:, i), REAL(0.0, real_8), real_8)
            levec(:, i) = CMPLX(evec_l(:, i), REAL(0.0, real_8), real_8)
            i = i+1
         ELSE IF (eval2(i) .GT. EPSILON(REAL(0.0, real_8))) THEN
            norm = SQRT(SUM(evec_r(:, i)**2.0_real_8)+SUM(evec_r(:, i+1)**2.0_real_8))
            revec(:, i) = CMPLX(evec_r(:, i), evec_r(:, i+1), real_8)/norm
            revec(:, i+1) = CMPLX(evec_r(:, i), -evec_r(:, i+1), real_8)/norm
            levec(:, i) = CMPLX(evec_l(:, i), evec_l(:, i+1), real_8)
            levec(:, i+1) = CMPLX(evec_l(:, i), -evec_l(:, i+1), real_8)
            i = i+2
         ELSE
            CALL cp__b("dbcsr/arnoldi/dbcsr_geev.F",394,'something went wrong while sorting the EV in dbcsr_geev')
         END IF
      END DO

      ! this is to keep the interface consistent with complex geev
      DO i = 1, ndim
         evals(i) = CMPLX(eval1(i), eval2(i), real_8)
      END DO

   END SUBROUTINE dbcsr_dgeev

! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_zgeev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      COMPLEX(real_8), DIMENSION(:, :)                   :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_8), DIMENSION(:)                      :: evals
      COMPLEX(real_8), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_zgeev', routineP = moduleN//':'//routineN

      INTEGER                                            :: info, lwork
      COMPLEX(real_8)                                    :: work(20*ndim), tmp_array(ndim, ndim)
      REAL(real_8)                                       :: work2(2*ndim)

      evals = CMPLX(0.0, 0.0, real_8)
      ! ask lapack how much space it would like in the work vector, don't ask me why
      lwork = -1
      CALL ZGEEV(jobvl, jobvr, ndim, tmp_array, ndim, evals, levec, ndim, revec, ndim, work, lwork, work2, info)
      lwork = MIN(20*ndim, INT(work(1)))

      tmp_array(:, :) = matrix(:, :)
      CALL ZGEEV(jobvl, jobvr, ndim, tmp_array, ndim, evals, levec, ndim, revec, ndim, work, lwork, work2, info)

   END SUBROUTINE dbcsr_zgeev

! **************************************************************************************************
!> \brief ...
!> \param jobvl ...
!> \param jobvr ...
!> \param matrix ...
!> \param ndim ...
!> \param evals ...
!> \param revec ...
!> \param levec ...
! **************************************************************************************************
   SUBROUTINE dbcsr_cgeev(jobvl, jobvr, matrix, ndim, evals, revec, levec)
      CHARACTER(1)                                       :: jobvl, jobvr
      COMPLEX(real_4), DIMENSION(:, :)                   :: matrix
      INTEGER                                            :: ndim
      COMPLEX(real_4), DIMENSION(:)                      :: evals
      COMPLEX(real_4), DIMENSION(:, :)                   :: revec, levec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_cgeev', routineP = moduleN//':'//routineN

      INTEGER                                            :: info, lwork
      COMPLEX(real_4)                                    :: work(20*ndim), tmp_array(ndim, ndim)
      REAL(real_4)                                       :: work2(2*ndim)

      evals = CMPLX(0.0, 0.0, real_4)
      ! ask lapack how much space it would like in the work vector, don't ask me why
      lwork = -1
      CALL CGEEV(jobvl, jobvr, ndim, tmp_array, ndim, evals, levec, ndim, revec, ndim, work, lwork, work2, info)
      lwork = MIN(20*ndim, INT(work(1)))

      tmp_array(:, :) = matrix(:, :)
      CALL CGEEV(jobvl, jobvr, ndim, tmp_array, ndim, evals, levec, ndim, revec, ndim, work, lwork, work2, info)

   END SUBROUTINE dbcsr_cgeev

END MODULE dbcsr_geev
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief interface to the low level methods for arnoldi iteration
!> \par History
!>       2014.09 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_arnoldi_methods
   USE dbcsr_arnoldi_types,             ONLY: &
        arnoldi_control, arnoldi_data_c, arnoldi_data_d, arnoldi_data_s, arnoldi_data_z, &
        dbcsr_arnoldi_data, get_control, get_data_c, get_data_d, get_data_s, get_data_z, &
        has_d_cmplx, has_d_real, has_s_cmplx, has_s_real, m_x_v_vectors
   USE dbcsr_data_methods,              ONLY: dbcsr_get_data_p
   USE dbcsr_geev,                      ONLY: dbcsr_general_local_diag,&
                                              dbcsr_symm_local_diag,&
                                              dbcsr_tridiag_local_diag
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_operations,                ONLY: dbcsr_add,&
                                              dbcsr_copy,&
                                              dbcsr_get_info,&
                                              dbcsr_scale
   USE dbcsr_types,                     ONLY: dbcsr_iterator,&
                                              dbcsr_obj,&
                                              dbcsr_obj_type_p
   USE dbcsr_vector_operations,         ONLY: dbcsr_matrix_colvec_multiply
   USE kinds,                           ONLY: real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_arnoldi_methods'

   PUBLIC :: arnoldi_init, build_subspace, compute_evals, arnoldi_iram, &
             gev_arnoldi_init, gev_build_subspace, gev_update_data

   INTERFACE convert_matrix
      MODULE PROCEDURE convert_matrix_z_to_d, convert_matrix_s_to_c
      MODULE PROCEDURE convert_matrix_d_to_z, convert_matrix_c_to_s
      MODULE PROCEDURE convert_matrix_z_to_z, convert_matrix_c_to_c
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Interface for the routine calcualting the implicit restarts
!>        Currently all based on lapack
!> \param arnoldi_data ...
! **************************************************************************************************
   SUBROUTINE arnoldi_iram(arnoldi_data)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_iram', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL arnoldi_iram_d(arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL arnoldi_iram_s(arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL arnoldi_iram_z(arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL arnoldi_iram_c(arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE arnoldi_iram

! **************************************************************************************************
!> \brief Interface to compute the eigenvalues of a nonsymmetric matrix
!>        This is only the serial version
!> \param arnoldi_data ...
! **************************************************************************************************
   SUBROUTINE compute_evals(arnoldi_data)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_evals', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL compute_evals_d(arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL compute_evals_s(arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL compute_evals_z(arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL compute_evals_c(arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE compute_evals

! **************************************************************************************************
!> \brief Interface for the initialization of the arnoldi subspace creation
!>        currently it can only setup a random vector but can be improved to
!>        various types of restarts easily
!> \param matrix pointer to the matrices as described in main interface
!> \param vectors work vectors for the matrix vector multiplications
!> \param arnoldi_data all data concerning the subspace
! **************************************************************************************************
   SUBROUTINE arnoldi_init(matrix, vectors, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(m_x_v_vectors)                                :: vectors
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL arnoldi_init_d(matrix, vectors, arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL arnoldi_init_s(matrix, vectors, arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL arnoldi_init_z(matrix, vectors, arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL arnoldi_init_c(matrix, vectors, arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE arnoldi_init

! **************************************************************************************************
!> \brief Interface for the initialization of the arnoldi subspace creation
!>        for the generalized eigenvalue problem
!> \param matrix pointer to the matrices as described in main interface
!> \param matrix_arnoldi ...
!> \param vectors work vectors for the matrix vector multiplications
!> \param arnoldi_data all data concerning the subspace
! **************************************************************************************************
   SUBROUTINE gev_arnoldi_init(matrix, matrix_arnoldi, vectors, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix, matrix_arnoldi
      TYPE(m_x_v_vectors)                                :: vectors
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_arnoldi_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL gev_arnoldi_init_d(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL gev_arnoldi_init_s(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL gev_arnoldi_init_z(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL gev_arnoldi_init_c(matrix, matrix_arnoldi, vectors, arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE gev_arnoldi_init

! **************************************************************************************************
!> \brief here the iterations are performed and the krylov space is constructed
!> \param matrix see above
!> \param vectors see above
!> \param arnoldi_data see above
! **************************************************************************************************
   SUBROUTINE build_subspace(matrix, vectors, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(m_x_v_vectors)                                :: vectors
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL build_subspace_d(matrix, vectors, arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL build_subspace_s(matrix, vectors, arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL build_subspace_z(matrix, vectors, arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL build_subspace_c(matrix, vectors, arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE build_subspace

! **************************************************************************************************
!> \brief here the iterations are performed and the krylov space for the generalized
!>        eigenvalue probelm is created
!> \param matrix see above
!> \param vectors see above
!> \param arnoldi_data see above
! **************************************************************************************************
   SUBROUTINE gev_build_subspace(matrix, vectors, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(m_x_v_vectors)                                :: vectors
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_build_subspace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL gev_build_subspace_d(matrix, vectors, arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL gev_build_subspace_s(matrix, vectors, arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL gev_build_subspace_z(matrix, vectors, arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL gev_build_subspace_c(matrix, vectors, arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE gev_build_subspace

! **************************************************************************************************
!> \brief in the generalized eigenvalue the matrix depende on the projection
!>        therefore the outer loop has to build a new set of matrices for the
!>        inner loop
!> \param matrix see above
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data see above
! **************************************************************************************************
   SUBROUTINE gev_update_data(matrix, matrix_arnoldi, vectors, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix, matrix_arnoldi
      TYPE(m_x_v_vectors)                                :: vectors
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_update_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (has_d_real(arnoldi_data)) CALL gev_update_data_d(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_s_real(arnoldi_data)) CALL gev_update_data_s(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_d_cmplx(arnoldi_data)) CALL gev_update_data_z(matrix, matrix_arnoldi, vectors, arnoldi_data)
      IF (has_s_cmplx(arnoldi_data)) CALL gev_update_data_c(matrix, matrix_arnoldi, vectors, arnoldi_data)

      CALL timestop(handle)

   END SUBROUTINE gev_update_data

! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_z_to_d(m_out, m_in)
      REAL(real_8), DIMENSION(:, :)                      :: m_out
      COMPLEX(real_8), DIMENSION(:, :)                   :: m_in

      m_out(:, :) = REAL(m_in(:, :), KIND=real_8)
   END SUBROUTINE convert_matrix_z_to_d

! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_c_to_s(m_out, m_in)
      REAL(real_4), DIMENSION(:, :)                      :: m_out
      COMPLEX(real_4), DIMENSION(:, :)                   :: m_in

      m_out(:, :) = REAL(m_in, KIND=real_4)
   END SUBROUTINE convert_matrix_c_to_s

! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_d_to_z(m_out, m_in)
      COMPLEX(real_8), DIMENSION(:, :)                   :: m_out
      REAL(real_8), DIMENSION(:, :)                      :: m_in

      m_out(:, :) = CMPLX(m_in, 0.0, KIND=real_8)
   END SUBROUTINE convert_matrix_d_to_z

! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_s_to_c(m_out, m_in)
      COMPLEX(real_4), DIMENSION(:, :)                   :: m_out
      REAL(real_4), DIMENSION(:, :)                      :: m_in

      m_out(:, :) = CMPLX(m_in, 0.0, KIND=real_4)
   END SUBROUTINE convert_matrix_s_to_c

! I kno that one is stupid but like this it simplifies the template
! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_z_to_z(m_out, m_in)
      COMPLEX(real_8), DIMENSION(:, :)                   :: m_out, m_in

      m_out(:, :) = m_in
   END SUBROUTINE convert_matrix_z_to_z

! **************************************************************************************************
!> \brief ...
!> \param m_out ...
!> \param m_in ...
! **************************************************************************************************
   SUBROUTINE convert_matrix_c_to_c(m_out, m_in)
      COMPLEX(real_4), DIMENSION(:, :)                   :: m_out, m_in

      m_out(:, :) = m_in
   END SUBROUTINE convert_matrix_c_to_c


! **************************************************************************************************
!> \brief Call the correct eigensolver, in the arnoldi method only the right
!>        eigenvectors are used. Lefts are created here but dumped immediatly 
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE compute_evals_d(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_evals_d', &
      routineP = moduleN//':'//routineN

    COMPLEX(real_8), DIMENSION(:, :), ALLOCATABLE   :: levec
    TYPE(arnoldi_data_d), POINTER            :: ar_data
    INTEGER                                  :: ndim
    TYPE(arnoldi_control), POINTER           :: control

    ar_data=>get_data_d(arnoldi_data)
    control=> get_control(arnoldi_data)
    ndim=control%current_step
    ALLOCATE(levec(ndim, ndim))

! Needs antoher interface as the calls to real and complex geev differ (sucks!)
! only perform the diagonalization on processors which hold data
    IF(control%generalized_ev)THEN
       CALL dbcsr_symm_local_diag('V',ar_data%Hessenberg(1:ndim, 1:ndim), ndim,&
                                  ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim))
    ELSE
       IF(control%symmetric)THEN
          CALL dbcsr_tridiag_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       ELSE
          CALL dbcsr_general_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       END IF
    END IF

    DEALLOCATE(levec)

  END SUBROUTINE compute_evals_d

! **************************************************************************************************
!> \brief Initialization of the arnoldi vector. Here a random vector is used,
!>        might be generalized in the future 
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_init_d (matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i, iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_8)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    REAL(kind=real_8)                         :: norm 
    REAL(kind=real_8), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_d), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp
    
    ar_data=>get_data_d(arnoldi_data)

   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_8 ; w_vec = 0.0_real_8
    ar_data%Hessenberg=0.0_real_8

    IF(control%has_initial_vector)THEN
       ! after calling the set routine the initial vector is stored in f_vec
       CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
       ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL dlarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF

    CALL transfer_dbcsr_to_local_array_d(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! compute the vector norm of the random vectorm, get it real valued as well (rnorm)
    CALL compute_norms_d(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_8)/rnorm)

    ! Everything prepared, initialize the Arnoldi iteration
    CALL transfer_dbcsr_to_local_array_d(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! This permits to compute the subspace of a matrix which is a product of multiple matrices
    DO i=1, SIZE(matrix)
       CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                        0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL dbcsr_copy(vectors%input_vec, vectors%result_vec)
    END DO

    CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, w_vec, nrow_local, control%local_comp)

    ! Put the projection into the Hessenberg matrix, and make the vectors orthonormal
    ar_data%Hessenberg(1, 1)=DOT_PRODUCT(v_vec, w_vec)
    CALL mp_sum(ar_data%Hessenberg(1, 1), pcol_group)
    ar_data%f_vec=w_vec-v_vec*ar_data%Hessenberg(1, 1)

    ar_data%local_history(:, 1)=v_vec(:)

    ! We did the first step in here so we should set the current step for the subspace generation accordingly
    control%current_step=1

    DEALLOCATE(v_vec, w_vec)

  END SUBROUTINE arnoldi_init_d

! **************************************************************************************************
!> \brief Alogorithm for the implicit restarts in the arnoldi method
!>        this is an early implementaion which scales subspace size^4
!>        by replacing the lapack calls with direct math the 
!>        QR and  gemms can be made linear and a N^2 sacling will be acchieved
!>        however this already sets the framework but should be used with care
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_iram_d(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_iram_d', &
         routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_d), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER                     :: control
    COMPLEX(real_8), DIMENSION(:,:), ALLOCATABLE  :: tmp_mat, safe_mat, Q, tmp_mat1
    COMPLEX(real_8), DIMENSION(:), ALLOCATABLE    :: work, tau, work_measure
    INTEGER                                   :: msize, lwork, i, j, info, nwant
    REAL(kind=real_8)                          :: beta, sigma
    REAL(kind=real_8),DIMENSION(:,:),ALLOCATABLE :: Qdata


    ! This is just a terribly inefficient implementation but I hope it is correct and might serve as a reference
    ar_data=>get_data_d(arnoldi_data)
    control=>get_control(arnoldi_data)
    msize=control%current_step
    nwant=control%nval_out
    ALLOCATE(tmp_mat(msize,msize)); ALLOCATE(safe_mat(msize,msize))
    ALLOCATE(Q(msize,msize)); ALLOCATE(tmp_mat1(msize,msize))
    ALLOCATE(work_measure(1))
    ALLOCATE(tau(msize)); ALLOCATE(Qdata(msize,msize))
    !make Q identity
    Q=CMPLX(0.0, 0.0, real_8)
    DO i=1,msize
       Q(i,i)=CMPLX(1.0, 0.0, real_8)
    END DO

    ! Looks a bit odd, but safe_mat will contain the result in the end, while tmpmat gets violated by lapack
    CALL convert_matrix(tmp_mat,ar_data%Hessenberg(1:msize,1:msize))
    safe_mat(:,:)=tmp_mat(:,:)

    DO i=1,msize
       ! A bit a strange check but in the end we only want to shift the unwanted evals
       IF(ANY(control%selected_ind==i))CYCLE
       ! Here we shift the matrix by subtracting unwanted evals from the diagonal
       DO j=1,msize
          tmp_mat(j,j)=tmp_mat(j,j)-ar_data%evals(i)
       END DO
       ! Now we repair the damage by QR factorizing
       lwork=-1
       CALL zgeqrf(msize,msize,tmp_mat,msize,tau,work_measure,lwork,info)
       lwork=INT(work_measure(1))
       IF (ALLOCATED(work)) THEN
          IF (SIZE(work).LT.lwork) THEN
             DEALLOCATE(work)
          ENDIF
       ENDIF
       IF (.NOT.ALLOCATED(work)) ALLOCATE(work(lwork))
       CALL zgeqrf(msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! Ask Lapack to reconstruct Q from its own way of storing data (tmpmat will contain Q)
       CALL zungqr(msize,msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! update Q=Q*Q_current
       tmp_mat1(:,:)=Q(:,:)
       CALL zgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         Q,msize)       
       ! Update H=(Q*)HQ
       CALL zgemm('C','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat,msize,safe_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         tmp_mat1,msize)
       CALL zgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         safe_mat,msize)

       ! this one is crucial for numerics not to accumulate noise in the subdiagonals
       DO j=1,msize
          safe_mat(j+2:msize,j)=CMPLX(0.0, 0.0, real_8)
       END DO
       tmp_mat(:,:)=safe_mat(:,:)
    END DO

    ! Now we can compute our restart quantities
    ar_data%Hessenberg=0.0_real_8
    CALL convert_matrix(ar_data%Hessenberg(1:msize,1:msize),safe_mat)
    CALL convert_matrix(Qdata,Q)
  
    beta=ar_data%Hessenberg(nwant+1,nwant); sigma=Qdata(msize,nwant)

    !update the residuum and the basis vectors
    IF(control%local_comp)THEN
       ar_data%f_vec=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,nwant+1))*beta+ar_data%f_vec(:)*sigma
       ar_data%local_history(:,1:nwant)=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,1:nwant))
    END IF
    ! Set the current step to nwant so the subspace build knows where to start
    control%current_step=nwant
    
    DEALLOCATE(tmp_mat,safe_mat,Q,Qdata,tmp_mat1,work,tau,work_measure)
    
  END SUBROUTINE arnoldi_iram_d

! **************************************************************************************************
!> \brief Here we create the Krylov subspace and fill the Hessenberg matrix
!>        convergence check is only performed on subspace convergence
!>        Gram Schidt is used to orthonogonalize. 
!>        If this is numericall not sufficient a Daniel, Gragg, Kaufman and Steward
!>        correction is performed
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE build_subspace_d(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors), TARGET              :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_d', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: i, j, ncol_local, nrow_local
    REAL(real_8)                        :: rnorm
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_d), POINTER  :: ar_data
    REAL(kind=real_8)                         :: norm
    REAL(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    TYPE(dbcsr_obj), POINTER                 :: input_vec, result_vec, swap_vec

    ar_data=>get_data_d(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_8 ; w_vec = 0.0_real_8
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))

    DO j=control%current_step, control%max_iter-1

       ! compute the vector norm of the residuum, get it real valued as well (rnorm)
       CALL compute_norms_d(ar_data%f_vec, norm, rnorm, control%pcol_group)

       ! check convergence and inform everybody about it, a bit annoying to talk to everybody because of that
       IF(control%myproc==0)control%converged=rnorm.LT.REAL(control%threshold, real_8)
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT

       ! transfer normalized residdum to history and its norm to the Hessenberg matrix
       IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
       v_vec(:)=ar_data%f_vec(:)/rnorm; ar_data%local_history(:, j+1)=v_vec(:); ar_data%Hessenberg(j+1, j)=norm

       input_vec=>vectors%input_vec
       result_vec=>vectors%result_vec
       CALL transfer_local_array_to_dbcsr_d(input_vec, v_vec, nrow_local, control%local_comp)
 
       ! This permits to compute the subspace of a matrix which is a product of two matrices
       DO i=1, SIZE(matrix)
          CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, input_vec, result_vec, 1.0_real_8, &
                                            0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
          swap_vec=>input_vec
          input_vec=>result_vec
          result_vec=>swap_vec
       END DO

       CALL transfer_dbcsr_to_local_array_d(input_vec, w_vec, nrow_local, control%local_comp)

       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_d(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j+1, &
                               ar_data%local_history, ar_data%local_history, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can become a problem later on, there is probably a good check whether it's necessary, but we don't perform it
       CALL DGKS_ortho_d(h_vec, ar_data%f_vec, s_vec, nrow_local, j+1, ar_data%local_history, &
                                   ar_data%local_history, control%local_comp, control%pcol_group)
       ! Finally we can put the projections into our Hessenberg matrix
       ar_data%Hessenberg(1:j+1, j+1)= h_vec(1:j+1)
       control%current_step=j+1
    END DO

    ! compute the vector norm of the final residuum and put it in to Hessenberg
    CALL compute_norms_d(ar_data%f_vec, norm, rnorm, control%pcol_group)
    ar_data%Hessenberg(control%current_step+1, control%current_step)=norm
    
    ! broadcast the Hessenberg matrix so we don't need to care later on
    CALL mp_bcast(ar_data%Hessenberg, 0, control%mp_group)

    DEALLOCATE(v_vec, w_vec, h_vec, s_vec)

  END SUBROUTINE  build_subspace_d

! **************************************************************************************************
!> \brief Helper routine to transfer the all data of a dbcsr matrix to a local array
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_dbcsr_to_local_array_d(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    REAL(kind=real_8), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=0.0_real_8)
    IF(is_local)array(1:n)=data_vec(1:n)

  END SUBROUTINE transfer_dbcsr_to_local_array_d

! **************************************************************************************************
!> \brief The inverse routine transfering data back from an array to a dbcsr
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_local_array_to_dbcsr_d(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    REAL(kind=real_8), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=0.0_real_8)
    IF(is_local)data_vec(1:n)=array(1:n)

! **************************************************************************************************

  END SUBROUTINE transfer_local_array_to_dbcsr_d

! **************************************************************************************************
!> \brief Gram-Schmidt in matrix vector form
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param w_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE Gram_Schmidt_ortho_d(h_vec, f_vec, s_vec, w_vec, nrow_local,&
                                            j, local_history, reorth_mat, local_comp, pcol_group)
    REAL(kind=real_8), DIMENSION(:)      :: h_vec, s_vec, f_vec, w_vec
    REAL(kind=real_8), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group
    LOGICAL                                          :: local_comp

    CHARACTER(LEN=*), PARAMETER :: routineN = 'Gram_Schmidt_ortho_d', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    ! Let's do the orthonormalization, first try the Gram Schmidt scheme
    h_vec=0.0_real_8; f_vec=0.0_real_8; s_vec=0.0_real_8
    IF(local_comp)CALL dgemv('T', nrow_local, j, 1.0_real_8, local_history, &
                                      nrow_local, w_vec, 1, 0.0_real_8, h_vec, 1)
    CALL mp_sum(h_vec(1:j), pcol_group)

    IF(local_comp)CALL dgemv('N', nrow_local, j, 1.0_real_8, reorth_mat, &
                                      nrow_local, h_vec, 1, 0.0_real_8, f_vec, 1)
    f_vec(:)=w_vec(:)-f_vec(:)

    CALL timestop(handle)

  END SUBROUTINE Gram_Schmidt_ortho_d

! **************************************************************************************************
!> \brief Compute the  Daniel, Gragg, Kaufman and Steward correction
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE DGKS_ortho_d(h_vec, f_vec, s_vec, nrow_local, j, &
                                    local_history, reorth_mat, local_comp, pcol_group)
    REAL(kind=real_8), DIMENSION(:)      :: h_vec, s_vec, f_vec
    REAL(kind=real_8), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group

    CHARACTER(LEN=*), PARAMETER :: routineN = 'DGKS_ortho_d', &
         routineP = moduleN//':'//routineN

    LOGICAL                                          :: local_comp
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    IF(local_comp)CALL dgemv('T', nrow_local, j, 1.0_real_8, local_history, &
                                       nrow_local, f_vec, 1, 0.0_real_8, s_vec, 1)
    CALL mp_sum(s_vec(1:j), pcol_group)
    IF(local_comp)CALL dgemv('N', nrow_local, j, -1.0_real_8, reorth_mat, &
                                       nrow_local, s_vec, 1, 1.0_real_8, f_vec, 1)
    h_vec(1:j)=h_vec(1:j)+s_vec(1:j)

    CALL timestop(handle)

  END SUBROUTINE DGKS_ortho_d

! **************************************************************************************************
!> \brief Compute the norm of a vector distributed along proc_col
!>        as local arrays. Always return the real part next to the complex rep.
!> \param vec ...
!> \param norm ...
!> \param rnorm ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE compute_norms_d(vec, norm, rnorm, pcol_group)
    REAL(kind=real_8), DIMENSION(:)           :: vec
    REAL(real_8)                        :: rnorm
    REAL(kind=real_8)                         :: norm
    INTEGER                                  :: pcol_group

    ! the norm is computed along the processor column
    norm=DOT_PRODUCT(vec, vec)
    CALL mp_sum(norm, pcol_group)
    rnorm=SQRT(REAL(norm, real_8))
    norm=rnorm

  END SUBROUTINE compute_norms_d

! **************************************************************************************************
!> \brief Computes the intial guess for the solution of the generalized eigenvalue 
!>        using the arnoldi method
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_arnoldi_init_d (matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_8)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    REAL(kind=real_8)                         :: norm, denom
    REAL(kind=real_8), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_d), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp

    ar_data=>get_data_d(arnoldi_data)
    
   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_8 ; w_vec = 0.0_real_8
    ar_data%Hessenberg=0.0_real_8

    IF(control%has_initial_vector)THEN
    ! after calling the set routine the initial vector is stored in f_vec
        CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
    ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL dlarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF   

    CALL transfer_dbcsr_to_local_array_d(vectors%input_vec, v_vec, nrow_local, control%local_comp)
   
    ! compute the vector norm of the reandom vectorm, get it real valued as well (rnorm)
    CALL compute_norms_d(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_8)/rnorm)

    CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                      0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, w_vec, nrow_local, control%local_comp)
   
    ar_data%rho_scale=0.0_real_8
    ar_data%rho_scale=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(ar_data%rho_scale, pcol_group)

    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                      0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, w_vec, nrow_local, control%local_comp)
    
    denom=0.0_real_8
    denom=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(denom, pcol_group)
    IF(control%myproc==0) ar_data%rho_scale=ar_data%rho_scale/denom
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

    ! if the maximum ev is requested we need to optimize with -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, 1.0_real_8, -ar_data%rho_scale)
   
    ar_data%x_vec=v_vec

  END SUBROUTINE gev_arnoldi_init_d

! **************************************************************************************************
!> \brief builds the basis rothogonal wrt. teh metric.
!>        The structure looks similar to normal arnoldi but norms, vectors and 
!>        matrix_vector products are very differently defined. Therefore it is 
!>        cleaner to put it in a seperate subroutine to avoid confusion
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_build_subspace_d(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_d', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: j, ncol_local, nrow_local, pcol_group
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_d), POINTER            :: ar_data
    REAL(kind=real_8)                         :: norm
    REAL(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:)    :: Zmat, CZmat , BZmat 

    ar_data=>get_data_d(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.
    pcol_group=control%pcol_group

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_8 ; w_vec = 0.0_real_8
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))
    ALLOCATE(Zmat(nrow_local,control%max_iter)); ALLOCATE(CZmat(nrow_local,control%max_iter))
    ALLOCATE(BZmat(nrow_local,control%max_iter))

    CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, ar_data%x_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                      0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, BZmat(:,1), nrow_local, control%local_comp)
    
    norm=0.0_real_8 
    norm=DOT_PRODUCT(ar_data%x_vec,BZmat(:,1)) 
    CALL mp_sum(norm, pcol_group)
    IF(control%local_comp)THEN
       Zmat(:,1)=ar_data%x_vec/SQRT(norm);  BZmat(:,1)= BZmat(:,1)/SQRT(norm)
    END IF

    DO j=1,control%max_iter
       control%current_step=j
       CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, Zmat(:,j), nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                        0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, CZmat(:,j), nrow_local, control%local_comp)
       w_vec(:)=CZmat(:,j)                         
       
       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_d(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j, &
                               BZmat, Zmat, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can becom a problem later on, there is probably a good check, but we don't perform it
       CALL DGKS_ortho_d(h_vec, ar_data%f_vec, s_vec, nrow_local, j, BZmat, &
                                    Zmat, control%local_comp, control%pcol_group)
    
       CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                        0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, v_vec, nrow_local, control%local_comp)
       norm=0.0_real_8
       norm=DOT_PRODUCT(ar_data%f_vec,v_vec)
       CALL mp_sum(norm, pcol_group)
      
       IF(control%myproc==0)control%converged=REAL(norm,real_8).LT.EPSILON(REAL(1.0,real_8))
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT 
       IF(j==control%max_iter-1)EXIT

       IF(control%local_comp)THEN
          Zmat(:,j+1)=ar_data%f_vec/SQRT(norm);  BZmat(:,j+1)= v_vec(:)/SQRT(norm)
       END IF
    END DO

! getting a bit more complicated here as the final matrix is again a product which has to be computed with the 
! ditributed vectors, therefore a sum along the first proc_col is necessary. As we want that matrix everywhere,
! we set it to zero before and compute the distributed product only on the first col and then sum over the full grid
    ar_data%Hessenberg=0.0_real_8
    IF(control%local_comp)THEN
       ar_data%Hessenberg(1:control%current_step,1:control%current_step)=&
          MATMUL(TRANSPOSE(CZmat(:,1:control%current_step)),Zmat(:,1:control%current_step))
    END IF
    CALL mp_sum(ar_data%Hessenberg,control%mp_group)

    ar_data%local_history=Zmat
    ! broadcast the Hessenberg matrix so we don't need to care later on

    DEALLOCATE(v_vec); DEALLOCATE(w_vec); DEALLOCATE(s_vec); DEALLOCATE(h_vec); DEALLOCATE(CZmat);
    DEALLOCATE(Zmat); DEALLOCATE(BZmat)

  END SUBROUTINE gev_build_subspace_d

! **************************************************************************************************
!> \brief Updates all data after an inner loop of the generalized ev arnoldi. 
!>        Updates rho and C=A-rho*B accordingly.
!>        As an update scheme is used for he ev, the output ev has to be replaced
!>        with the updated one.
!>        Furthermore a convergence check is performed. The mv product could
!>        be skiiped by making clever use of the precomputed data, However,
!>        it is most likely not worth the effort.
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_update_data_d(matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_update_data_d', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_d), POINTER            :: ar_data
    INTEGER                                  :: ncol_local, nrow_local, ind, i
    REAL(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: v_vec
    REAL(real_8)                        :: rnorm
    REAL(kind=real_8)                         :: norm
    COMPLEX(real_8)                     :: val 

    control=>get_control(arnoldi_data)

    ar_data=>get_data_d(arnoldi_data)

! compute the new shift, hack around the problem templating the conversion
    val=ar_data%evals(control%selected_ind(1))
    ar_data%rho_scale=ar_data%rho_scale+REAL(val,real_8)
! compute the new eigenvector / initial guess for the next arnoldi loop    
    ar_data%x_vec=0.0_real_8
    DO i=1,control%current_step
       val=ar_data%revec(i,control%selected_ind(1))
       ar_data%x_vec(:)=ar_data%x_vec(:)+ar_data%local_history(:,i)*REAL(val,real_8)
    END DO
!      ar_data%x_vec(:)=MATMUL(ar_data%local_history(:,1:control%current_step),&
!                        ar_data%revec(1:control%current_step,control%selected_ind(1)))

! update the C-matrix (A-rho*B), if teh maximum value is requested we have to use -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, 1.0_real_8, -ar_data%rho_scale)

! compute convergence and set the correct eigenvalue and eigenvector
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    IF (ncol_local>0) THEN
       ALLOCATE(v_vec(nrow_local))
       CALL compute_norms_d(ar_data%x_vec, norm, rnorm, control%pcol_group)
       v_vec(:)=ar_data%x_vec(:)/rnorm
    ENDIF
    CALL transfer_local_array_to_dbcsr_d(vectors%input_vec, v_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix_arnoldi(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_8, &
                                            0.0_real_8, vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_d(vectors%result_vec, v_vec, nrow_local, control%local_comp)
    IF (ncol_local>0) THEN
       CALL compute_norms_d(v_vec, norm, rnorm, control%pcol_group)
       ! check convergence 
       control%converged=rnorm.LT.control%threshold
       DEALLOCATE(v_vec)
    ENDIF
    ! and broadcast the real eigenvalue    
    CALL mp_bcast(control%converged,0,control%mp_group)
    ind=control%selected_ind(1)
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

! Again the maximum value request is done on -A therefore the eigenvalue needs the opposite sign
    ar_data%evals(ind)=ar_data%rho_scale


  END SUBROUTINE gev_update_data_d


! **************************************************************************************************
!> \brief Call the correct eigensolver, in the arnoldi method only the right
!>        eigenvectors are used. Lefts are created here but dumped immediatly 
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE compute_evals_s(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_evals_s', &
      routineP = moduleN//':'//routineN

    COMPLEX(real_4), DIMENSION(:, :), ALLOCATABLE   :: levec
    TYPE(arnoldi_data_s), POINTER            :: ar_data
    INTEGER                                  :: ndim
    TYPE(arnoldi_control), POINTER           :: control

    ar_data=>get_data_s(arnoldi_data)
    control=> get_control(arnoldi_data)
    ndim=control%current_step
    ALLOCATE(levec(ndim, ndim))

! Needs antoher interface as the calls to real and complex geev differ (sucks!)
! only perform the diagonalization on processors which hold data
    IF(control%generalized_ev)THEN
       CALL dbcsr_symm_local_diag('V',ar_data%Hessenberg(1:ndim, 1:ndim), ndim,&
                                  ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim))
    ELSE
       IF(control%symmetric)THEN
          CALL dbcsr_tridiag_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       ELSE
          CALL dbcsr_general_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       END IF
    END IF

    DEALLOCATE(levec)

  END SUBROUTINE compute_evals_s

! **************************************************************************************************
!> \brief Initialization of the arnoldi vector. Here a random vector is used,
!>        might be generalized in the future 
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_init_s (matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i, iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_4)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    REAL(kind=real_4)                         :: norm 
    REAL(kind=real_4), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_s), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp
    
    ar_data=>get_data_s(arnoldi_data)

   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_4 ; w_vec = 0.0_real_4
    ar_data%Hessenberg=0.0_real_4

    IF(control%has_initial_vector)THEN
       ! after calling the set routine the initial vector is stored in f_vec
       CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
       ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL slarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF

    CALL transfer_dbcsr_to_local_array_s(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! compute the vector norm of the random vectorm, get it real valued as well (rnorm)
    CALL compute_norms_s(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_4)/rnorm)

    ! Everything prepared, initialize the Arnoldi iteration
    CALL transfer_dbcsr_to_local_array_s(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! This permits to compute the subspace of a matrix which is a product of multiple matrices
    DO i=1, SIZE(matrix)
       CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                        0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL dbcsr_copy(vectors%input_vec, vectors%result_vec)
    END DO

    CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, w_vec, nrow_local, control%local_comp)

    ! Put the projection into the Hessenberg matrix, and make the vectors orthonormal
    ar_data%Hessenberg(1, 1)=DOT_PRODUCT(v_vec, w_vec)
    CALL mp_sum(ar_data%Hessenberg(1, 1), pcol_group)
    ar_data%f_vec=w_vec-v_vec*ar_data%Hessenberg(1, 1)

    ar_data%local_history(:, 1)=v_vec(:)

    ! We did the first step in here so we should set the current step for the subspace generation accordingly
    control%current_step=1

    DEALLOCATE(v_vec, w_vec)

  END SUBROUTINE arnoldi_init_s

! **************************************************************************************************
!> \brief Alogorithm for the implicit restarts in the arnoldi method
!>        this is an early implementaion which scales subspace size^4
!>        by replacing the lapack calls with direct math the 
!>        QR and  gemms can be made linear and a N^2 sacling will be acchieved
!>        however this already sets the framework but should be used with care
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_iram_s(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_iram_s', &
         routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_s), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER                     :: control
    COMPLEX(real_4), DIMENSION(:,:), ALLOCATABLE  :: tmp_mat, safe_mat, Q, tmp_mat1
    COMPLEX(real_4), DIMENSION(:), ALLOCATABLE    :: work, tau, work_measure
    INTEGER                                   :: msize, lwork, i, j, info, nwant
    REAL(kind=real_4)                          :: beta, sigma
    REAL(kind=real_4),DIMENSION(:,:),ALLOCATABLE :: Qdata


    ! This is just a terribly inefficient implementation but I hope it is correct and might serve as a reference
    ar_data=>get_data_s(arnoldi_data)
    control=>get_control(arnoldi_data)
    msize=control%current_step
    nwant=control%nval_out
    ALLOCATE(tmp_mat(msize,msize)); ALLOCATE(safe_mat(msize,msize))
    ALLOCATE(Q(msize,msize)); ALLOCATE(tmp_mat1(msize,msize))
    ALLOCATE(work_measure(1))
    ALLOCATE(tau(msize)); ALLOCATE(Qdata(msize,msize))
    !make Q identity
    Q=CMPLX(0.0, 0.0, real_4)
    DO i=1,msize
       Q(i,i)=CMPLX(1.0, 0.0, real_4)
    END DO

    ! Looks a bit odd, but safe_mat will contain the result in the end, while tmpmat gets violated by lapack
    CALL convert_matrix(tmp_mat,ar_data%Hessenberg(1:msize,1:msize))
    safe_mat(:,:)=tmp_mat(:,:)

    DO i=1,msize
       ! A bit a strange check but in the end we only want to shift the unwanted evals
       IF(ANY(control%selected_ind==i))CYCLE
       ! Here we shift the matrix by subtracting unwanted evals from the diagonal
       DO j=1,msize
          tmp_mat(j,j)=tmp_mat(j,j)-ar_data%evals(i)
       END DO
       ! Now we repair the damage by QR factorizing
       lwork=-1
       CALL cgeqrf(msize,msize,tmp_mat,msize,tau,work_measure,lwork,info)
       lwork=INT(work_measure(1))
       IF (ALLOCATED(work)) THEN
          IF (SIZE(work).LT.lwork) THEN
             DEALLOCATE(work)
          ENDIF
       ENDIF
       IF (.NOT.ALLOCATED(work)) ALLOCATE(work(lwork))
       CALL cgeqrf(msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! Ask Lapack to reconstruct Q from its own way of storing data (tmpmat will contain Q)
       CALL cungqr(msize,msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! update Q=Q*Q_current
       tmp_mat1(:,:)=Q(:,:)
       CALL cgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         Q,msize)       
       ! Update H=(Q*)HQ
       CALL cgemm('C','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat,msize,safe_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         tmp_mat1,msize)
       CALL cgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         safe_mat,msize)

       ! this one is crucial for numerics not to accumulate noise in the subdiagonals
       DO j=1,msize
          safe_mat(j+2:msize,j)=CMPLX(0.0, 0.0, real_4)
       END DO
       tmp_mat(:,:)=safe_mat(:,:)
    END DO

    ! Now we can compute our restart quantities
    ar_data%Hessenberg=0.0_real_4
    CALL convert_matrix(ar_data%Hessenberg(1:msize,1:msize),safe_mat)
    CALL convert_matrix(Qdata,Q)
  
    beta=ar_data%Hessenberg(nwant+1,nwant); sigma=Qdata(msize,nwant)

    !update the residuum and the basis vectors
    IF(control%local_comp)THEN
       ar_data%f_vec=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,nwant+1))*beta+ar_data%f_vec(:)*sigma
       ar_data%local_history(:,1:nwant)=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,1:nwant))
    END IF
    ! Set the current step to nwant so the subspace build knows where to start
    control%current_step=nwant
    
    DEALLOCATE(tmp_mat,safe_mat,Q,Qdata,tmp_mat1,work,tau,work_measure)
    
  END SUBROUTINE arnoldi_iram_s

! **************************************************************************************************
!> \brief Here we create the Krylov subspace and fill the Hessenberg matrix
!>        convergence check is only performed on subspace convergence
!>        Gram Schidt is used to orthonogonalize. 
!>        If this is numericall not sufficient a Daniel, Gragg, Kaufman and Steward
!>        correction is performed
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE build_subspace_s(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors), TARGET              :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_s', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: i, j, ncol_local, nrow_local
    REAL(real_4)                        :: rnorm
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_s), POINTER  :: ar_data
    REAL(kind=real_4)                         :: norm
    REAL(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    TYPE(dbcsr_obj), POINTER                 :: input_vec, result_vec, swap_vec

    ar_data=>get_data_s(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_4 ; w_vec = 0.0_real_4
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))

    DO j=control%current_step, control%max_iter-1

       ! compute the vector norm of the residuum, get it real valued as well (rnorm)
       CALL compute_norms_s(ar_data%f_vec, norm, rnorm, control%pcol_group)

       ! check convergence and inform everybody about it, a bit annoying to talk to everybody because of that
       IF(control%myproc==0)control%converged=rnorm.LT.REAL(control%threshold, real_4)
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT

       ! transfer normalized residdum to history and its norm to the Hessenberg matrix
       IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
       v_vec(:)=ar_data%f_vec(:)/rnorm; ar_data%local_history(:, j+1)=v_vec(:); ar_data%Hessenberg(j+1, j)=norm

       input_vec=>vectors%input_vec
       result_vec=>vectors%result_vec
       CALL transfer_local_array_to_dbcsr_s(input_vec, v_vec, nrow_local, control%local_comp)
 
       ! This permits to compute the subspace of a matrix which is a product of two matrices
       DO i=1, SIZE(matrix)
          CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, input_vec, result_vec, 1.0_real_4, &
                                            0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
          swap_vec=>input_vec
          input_vec=>result_vec
          result_vec=>swap_vec
       END DO

       CALL transfer_dbcsr_to_local_array_s(input_vec, w_vec, nrow_local, control%local_comp)

       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_s(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j+1, &
                               ar_data%local_history, ar_data%local_history, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can become a problem later on, there is probably a good check whether it's necessary, but we don't perform it
       CALL DGKS_ortho_s(h_vec, ar_data%f_vec, s_vec, nrow_local, j+1, ar_data%local_history, &
                                   ar_data%local_history, control%local_comp, control%pcol_group)
       ! Finally we can put the projections into our Hessenberg matrix
       ar_data%Hessenberg(1:j+1, j+1)= h_vec(1:j+1)
       control%current_step=j+1
    END DO

    ! compute the vector norm of the final residuum and put it in to Hessenberg
    CALL compute_norms_s(ar_data%f_vec, norm, rnorm, control%pcol_group)
    ar_data%Hessenberg(control%current_step+1, control%current_step)=norm
    
    ! broadcast the Hessenberg matrix so we don't need to care later on
    CALL mp_bcast(ar_data%Hessenberg, 0, control%mp_group)

    DEALLOCATE(v_vec, w_vec, h_vec, s_vec)

  END SUBROUTINE  build_subspace_s

! **************************************************************************************************
!> \brief Helper routine to transfer the all data of a dbcsr matrix to a local array
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_dbcsr_to_local_array_s(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    REAL(kind=real_4), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=0.0_real_4)
    IF(is_local)array(1:n)=data_vec(1:n)

  END SUBROUTINE transfer_dbcsr_to_local_array_s

! **************************************************************************************************
!> \brief The inverse routine transfering data back from an array to a dbcsr
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_local_array_to_dbcsr_s(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    REAL(kind=real_4), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=0.0_real_4)
    IF(is_local)data_vec(1:n)=array(1:n)

! **************************************************************************************************

  END SUBROUTINE transfer_local_array_to_dbcsr_s

! **************************************************************************************************
!> \brief Gram-Schmidt in matrix vector form
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param w_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE Gram_Schmidt_ortho_s(h_vec, f_vec, s_vec, w_vec, nrow_local,&
                                            j, local_history, reorth_mat, local_comp, pcol_group)
    REAL(kind=real_4), DIMENSION(:)      :: h_vec, s_vec, f_vec, w_vec
    REAL(kind=real_4), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group
    LOGICAL                                          :: local_comp

    CHARACTER(LEN=*), PARAMETER :: routineN = 'Gram_Schmidt_ortho_s', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    ! Let's do the orthonormalization, first try the Gram Schmidt scheme
    h_vec=0.0_real_4; f_vec=0.0_real_4; s_vec=0.0_real_4
    IF(local_comp)CALL sgemv('T', nrow_local, j, 1.0_real_4, local_history, &
                                      nrow_local, w_vec, 1, 0.0_real_4, h_vec, 1)
    CALL mp_sum(h_vec(1:j), pcol_group)

    IF(local_comp)CALL sgemv('N', nrow_local, j, 1.0_real_4, reorth_mat, &
                                      nrow_local, h_vec, 1, 0.0_real_4, f_vec, 1)
    f_vec(:)=w_vec(:)-f_vec(:)

    CALL timestop(handle)

  END SUBROUTINE Gram_Schmidt_ortho_s

! **************************************************************************************************
!> \brief Compute the  Daniel, Gragg, Kaufman and Steward correction
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE DGKS_ortho_s(h_vec, f_vec, s_vec, nrow_local, j, &
                                    local_history, reorth_mat, local_comp, pcol_group)
    REAL(kind=real_4), DIMENSION(:)      :: h_vec, s_vec, f_vec
    REAL(kind=real_4), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group

    CHARACTER(LEN=*), PARAMETER :: routineN = 'DGKS_ortho_s', &
         routineP = moduleN//':'//routineN

    LOGICAL                                          :: local_comp
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    IF(local_comp)CALL sgemv('T', nrow_local, j, 1.0_real_4, local_history, &
                                       nrow_local, f_vec, 1, 0.0_real_4, s_vec, 1)
    CALL mp_sum(s_vec(1:j), pcol_group)
    IF(local_comp)CALL sgemv('N', nrow_local, j, -1.0_real_4, reorth_mat, &
                                       nrow_local, s_vec, 1, 1.0_real_4, f_vec, 1)
    h_vec(1:j)=h_vec(1:j)+s_vec(1:j)

    CALL timestop(handle)

  END SUBROUTINE DGKS_ortho_s

! **************************************************************************************************
!> \brief Compute the norm of a vector distributed along proc_col
!>        as local arrays. Always return the real part next to the complex rep.
!> \param vec ...
!> \param norm ...
!> \param rnorm ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE compute_norms_s(vec, norm, rnorm, pcol_group)
    REAL(kind=real_4), DIMENSION(:)           :: vec
    REAL(real_4)                        :: rnorm
    REAL(kind=real_4)                         :: norm
    INTEGER                                  :: pcol_group

    ! the norm is computed along the processor column
    norm=DOT_PRODUCT(vec, vec)
    CALL mp_sum(norm, pcol_group)
    rnorm=SQRT(REAL(norm, real_4))
    norm=rnorm

  END SUBROUTINE compute_norms_s

! **************************************************************************************************
!> \brief Computes the intial guess for the solution of the generalized eigenvalue 
!>        using the arnoldi method
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_arnoldi_init_s (matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_4)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    REAL(kind=real_4)                         :: norm, denom
    REAL(kind=real_4), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_s), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp

    ar_data=>get_data_s(arnoldi_data)
    
   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_4 ; w_vec = 0.0_real_4
    ar_data%Hessenberg=0.0_real_4

    IF(control%has_initial_vector)THEN
    ! after calling the set routine the initial vector is stored in f_vec
        CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
    ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL slarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF   

    CALL transfer_dbcsr_to_local_array_s(vectors%input_vec, v_vec, nrow_local, control%local_comp)
   
    ! compute the vector norm of the reandom vectorm, get it real valued as well (rnorm)
    CALL compute_norms_s(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_4)/rnorm)

    CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                      0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, w_vec, nrow_local, control%local_comp)
   
    ar_data%rho_scale=0.0_real_4
    ar_data%rho_scale=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(ar_data%rho_scale, pcol_group)

    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                      0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, w_vec, nrow_local, control%local_comp)
    
    denom=0.0_real_4
    denom=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(denom, pcol_group)
    IF(control%myproc==0) ar_data%rho_scale=ar_data%rho_scale/denom
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

    ! if the maximum ev is requested we need to optimize with -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, 1.0_real_4, -ar_data%rho_scale)
   
    ar_data%x_vec=v_vec

  END SUBROUTINE gev_arnoldi_init_s

! **************************************************************************************************
!> \brief builds the basis rothogonal wrt. teh metric.
!>        The structure looks similar to normal arnoldi but norms, vectors and 
!>        matrix_vector products are very differently defined. Therefore it is 
!>        cleaner to put it in a seperate subroutine to avoid confusion
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_build_subspace_s(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_s', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: j, ncol_local, nrow_local, pcol_group
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_s), POINTER            :: ar_data
    REAL(kind=real_4)                         :: norm
    REAL(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:)    :: Zmat, CZmat , BZmat 

    ar_data=>get_data_s(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.
    pcol_group=control%pcol_group

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = 0.0_real_4 ; w_vec = 0.0_real_4
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))
    ALLOCATE(Zmat(nrow_local,control%max_iter)); ALLOCATE(CZmat(nrow_local,control%max_iter))
    ALLOCATE(BZmat(nrow_local,control%max_iter))

    CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, ar_data%x_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                      0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, BZmat(:,1), nrow_local, control%local_comp)
    
    norm=0.0_real_4 
    norm=DOT_PRODUCT(ar_data%x_vec,BZmat(:,1)) 
    CALL mp_sum(norm, pcol_group)
    IF(control%local_comp)THEN
       Zmat(:,1)=ar_data%x_vec/SQRT(norm);  BZmat(:,1)= BZmat(:,1)/SQRT(norm)
    END IF

    DO j=1,control%max_iter
       control%current_step=j
       CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, Zmat(:,j), nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                        0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, CZmat(:,j), nrow_local, control%local_comp)
       w_vec(:)=CZmat(:,j)                         
       
       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_s(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j, &
                               BZmat, Zmat, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can becom a problem later on, there is probably a good check, but we don't perform it
       CALL DGKS_ortho_s(h_vec, ar_data%f_vec, s_vec, nrow_local, j, BZmat, &
                                    Zmat, control%local_comp, control%pcol_group)
    
       CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                        0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, v_vec, nrow_local, control%local_comp)
       norm=0.0_real_4
       norm=DOT_PRODUCT(ar_data%f_vec,v_vec)
       CALL mp_sum(norm, pcol_group)
      
       IF(control%myproc==0)control%converged=REAL(norm,real_4).LT.EPSILON(REAL(1.0,real_4))
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT 
       IF(j==control%max_iter-1)EXIT

       IF(control%local_comp)THEN
          Zmat(:,j+1)=ar_data%f_vec/SQRT(norm);  BZmat(:,j+1)= v_vec(:)/SQRT(norm)
       END IF
    END DO

! getting a bit more complicated here as the final matrix is again a product which has to be computed with the 
! ditributed vectors, therefore a sum along the first proc_col is necessary. As we want that matrix everywhere,
! we set it to zero before and compute the distributed product only on the first col and then sum over the full grid
    ar_data%Hessenberg=0.0_real_4
    IF(control%local_comp)THEN
       ar_data%Hessenberg(1:control%current_step,1:control%current_step)=&
          MATMUL(TRANSPOSE(CZmat(:,1:control%current_step)),Zmat(:,1:control%current_step))
    END IF
    CALL mp_sum(ar_data%Hessenberg,control%mp_group)

    ar_data%local_history=Zmat
    ! broadcast the Hessenberg matrix so we don't need to care later on

    DEALLOCATE(v_vec); DEALLOCATE(w_vec); DEALLOCATE(s_vec); DEALLOCATE(h_vec); DEALLOCATE(CZmat);
    DEALLOCATE(Zmat); DEALLOCATE(BZmat)

  END SUBROUTINE gev_build_subspace_s

! **************************************************************************************************
!> \brief Updates all data after an inner loop of the generalized ev arnoldi. 
!>        Updates rho and C=A-rho*B accordingly.
!>        As an update scheme is used for he ev, the output ev has to be replaced
!>        with the updated one.
!>        Furthermore a convergence check is performed. The mv product could
!>        be skiiped by making clever use of the precomputed data, However,
!>        it is most likely not worth the effort.
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_update_data_s(matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_update_data_s', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_s), POINTER            :: ar_data
    INTEGER                                  :: ncol_local, nrow_local, ind, i
    REAL(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: v_vec
    REAL(real_4)                        :: rnorm
    REAL(kind=real_4)                         :: norm
    COMPLEX(real_4)                     :: val 

    control=>get_control(arnoldi_data)

    ar_data=>get_data_s(arnoldi_data)

! compute the new shift, hack around the problem templating the conversion
    val=ar_data%evals(control%selected_ind(1))
    ar_data%rho_scale=ar_data%rho_scale+REAL(val,real_4)
! compute the new eigenvector / initial guess for the next arnoldi loop    
    ar_data%x_vec=0.0_real_4
    DO i=1,control%current_step
       val=ar_data%revec(i,control%selected_ind(1))
       ar_data%x_vec(:)=ar_data%x_vec(:)+ar_data%local_history(:,i)*REAL(val,real_4)
    END DO
!      ar_data%x_vec(:)=MATMUL(ar_data%local_history(:,1:control%current_step),&
!                        ar_data%revec(1:control%current_step,control%selected_ind(1)))

! update the C-matrix (A-rho*B), if teh maximum value is requested we have to use -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, 1.0_real_4, -ar_data%rho_scale)

! compute convergence and set the correct eigenvalue and eigenvector
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    IF (ncol_local>0) THEN
       ALLOCATE(v_vec(nrow_local))
       CALL compute_norms_s(ar_data%x_vec, norm, rnorm, control%pcol_group)
       v_vec(:)=ar_data%x_vec(:)/rnorm
    ENDIF
    CALL transfer_local_array_to_dbcsr_s(vectors%input_vec, v_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix_arnoldi(1)%matrix, vectors%input_vec, vectors%result_vec, 1.0_real_4, &
                                            0.0_real_4, vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_s(vectors%result_vec, v_vec, nrow_local, control%local_comp)
    IF (ncol_local>0) THEN
       CALL compute_norms_s(v_vec, norm, rnorm, control%pcol_group)
       ! check convergence 
       control%converged=rnorm.LT.control%threshold
       DEALLOCATE(v_vec)
    ENDIF
    ! and broadcast the real eigenvalue    
    CALL mp_bcast(control%converged,0,control%mp_group)
    ind=control%selected_ind(1)
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

! Again the maximum value request is done on -A therefore the eigenvalue needs the opposite sign
    ar_data%evals(ind)=ar_data%rho_scale


  END SUBROUTINE gev_update_data_s


! **************************************************************************************************
!> \brief Call the correct eigensolver, in the arnoldi method only the right
!>        eigenvectors are used. Lefts are created here but dumped immediatly 
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE compute_evals_z(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_evals_z', &
      routineP = moduleN//':'//routineN

    COMPLEX(real_8), DIMENSION(:, :), ALLOCATABLE   :: levec
    TYPE(arnoldi_data_z), POINTER            :: ar_data
    INTEGER                                  :: ndim
    TYPE(arnoldi_control), POINTER           :: control

    ar_data=>get_data_z(arnoldi_data)
    control=> get_control(arnoldi_data)
    ndim=control%current_step
    ALLOCATE(levec(ndim, ndim))

! Needs antoher interface as the calls to real and complex geev differ (sucks!)
! only perform the diagonalization on processors which hold data
    IF(control%generalized_ev)THEN
       CALL dbcsr_symm_local_diag('V',ar_data%Hessenberg(1:ndim, 1:ndim), ndim,&
                                  ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim))
    ELSE
       IF(control%symmetric)THEN
          CALL dbcsr_tridiag_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       ELSE
          CALL dbcsr_general_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       END IF
    END IF

    DEALLOCATE(levec)

  END SUBROUTINE compute_evals_z

! **************************************************************************************************
!> \brief Initialization of the arnoldi vector. Here a random vector is used,
!>        might be generalized in the future 
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_init_z (matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i, iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_8)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    COMPLEX(kind=real_8)                         :: norm 
    COMPLEX(kind=real_8), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_z), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp
    
    ar_data=>get_data_z(arnoldi_data)

   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_8) ; w_vec = CMPLX(0.0, 0.0, real_8)
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_8)

    IF(control%has_initial_vector)THEN
       ! after calling the set routine the initial vector is stored in f_vec
       CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
       ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL zlarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF

    CALL transfer_dbcsr_to_local_array_z(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! compute the vector norm of the random vectorm, get it real valued as well (rnorm)
    CALL compute_norms_z(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_8)/rnorm)

    ! Everything prepared, initialize the Arnoldi iteration
    CALL transfer_dbcsr_to_local_array_z(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! This permits to compute the subspace of a matrix which is a product of multiple matrices
    DO i=1, SIZE(matrix)
       CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                        CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL dbcsr_copy(vectors%input_vec, vectors%result_vec)
    END DO

    CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, w_vec, nrow_local, control%local_comp)

    ! Put the projection into the Hessenberg matrix, and make the vectors orthonormal
    ar_data%Hessenberg(1, 1)=DOT_PRODUCT(v_vec, w_vec)
    CALL mp_sum(ar_data%Hessenberg(1, 1), pcol_group)
    ar_data%f_vec=w_vec-v_vec*ar_data%Hessenberg(1, 1)

    ar_data%local_history(:, 1)=v_vec(:)

    ! We did the first step in here so we should set the current step for the subspace generation accordingly
    control%current_step=1

    DEALLOCATE(v_vec, w_vec)

  END SUBROUTINE arnoldi_init_z

! **************************************************************************************************
!> \brief Alogorithm for the implicit restarts in the arnoldi method
!>        this is an early implementaion which scales subspace size^4
!>        by replacing the lapack calls with direct math the 
!>        QR and  gemms can be made linear and a N^2 sacling will be acchieved
!>        however this already sets the framework but should be used with care
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_iram_z(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_iram_z', &
         routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_z), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER                     :: control
    COMPLEX(real_8), DIMENSION(:,:), ALLOCATABLE  :: tmp_mat, safe_mat, Q, tmp_mat1
    COMPLEX(real_8), DIMENSION(:), ALLOCATABLE    :: work, tau, work_measure
    INTEGER                                   :: msize, lwork, i, j, info, nwant
    COMPLEX(kind=real_8)                          :: beta, sigma
    COMPLEX(kind=real_8),DIMENSION(:,:),ALLOCATABLE :: Qdata


    ! This is just a terribly inefficient implementation but I hope it is correct and might serve as a reference
    ar_data=>get_data_z(arnoldi_data)
    control=>get_control(arnoldi_data)
    msize=control%current_step
    nwant=control%nval_out
    ALLOCATE(tmp_mat(msize,msize)); ALLOCATE(safe_mat(msize,msize))
    ALLOCATE(Q(msize,msize)); ALLOCATE(tmp_mat1(msize,msize))
    ALLOCATE(work_measure(1))
    ALLOCATE(tau(msize)); ALLOCATE(Qdata(msize,msize))
    !make Q identity
    Q=CMPLX(0.0, 0.0, real_8)
    DO i=1,msize
       Q(i,i)=CMPLX(1.0, 0.0, real_8)
    END DO

    ! Looks a bit odd, but safe_mat will contain the result in the end, while tmpmat gets violated by lapack
    CALL convert_matrix(tmp_mat,ar_data%Hessenberg(1:msize,1:msize))
    safe_mat(:,:)=tmp_mat(:,:)

    DO i=1,msize
       ! A bit a strange check but in the end we only want to shift the unwanted evals
       IF(ANY(control%selected_ind==i))CYCLE
       ! Here we shift the matrix by subtracting unwanted evals from the diagonal
       DO j=1,msize
          tmp_mat(j,j)=tmp_mat(j,j)-ar_data%evals(i)
       END DO
       ! Now we repair the damage by QR factorizing
       lwork=-1
       CALL zgeqrf(msize,msize,tmp_mat,msize,tau,work_measure,lwork,info)
       lwork=INT(work_measure(1))
       IF (ALLOCATED(work)) THEN
          IF (SIZE(work).LT.lwork) THEN
             DEALLOCATE(work)
          ENDIF
       ENDIF
       IF (.NOT.ALLOCATED(work)) ALLOCATE(work(lwork))
       CALL zgeqrf(msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! Ask Lapack to reconstruct Q from its own way of storing data (tmpmat will contain Q)
       CALL zungqr(msize,msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! update Q=Q*Q_current
       tmp_mat1(:,:)=Q(:,:)
       CALL zgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         Q,msize)       
       ! Update H=(Q*)HQ
       CALL zgemm('C','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat,msize,safe_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         tmp_mat1,msize)
       CALL zgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_8),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_8),&
                         safe_mat,msize)

       ! this one is crucial for numerics not to accumulate noise in the subdiagonals
       DO j=1,msize
          safe_mat(j+2:msize,j)=CMPLX(0.0, 0.0, real_8)
       END DO
       tmp_mat(:,:)=safe_mat(:,:)
    END DO

    ! Now we can compute our restart quantities
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_8)
    CALL convert_matrix(ar_data%Hessenberg(1:msize,1:msize),safe_mat)
    CALL convert_matrix(Qdata,Q)
  
    beta=ar_data%Hessenberg(nwant+1,nwant); sigma=Qdata(msize,nwant)

    !update the residuum and the basis vectors
    IF(control%local_comp)THEN
       ar_data%f_vec=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,nwant+1))*beta+ar_data%f_vec(:)*sigma
       ar_data%local_history(:,1:nwant)=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,1:nwant))
    END IF
    ! Set the current step to nwant so the subspace build knows where to start
    control%current_step=nwant
    
    DEALLOCATE(tmp_mat,safe_mat,Q,Qdata,tmp_mat1,work,tau,work_measure)
    
  END SUBROUTINE arnoldi_iram_z

! **************************************************************************************************
!> \brief Here we create the Krylov subspace and fill the Hessenberg matrix
!>        convergence check is only performed on subspace convergence
!>        Gram Schidt is used to orthonogonalize. 
!>        If this is numericall not sufficient a Daniel, Gragg, Kaufman and Steward
!>        correction is performed
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE build_subspace_z(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors), TARGET              :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_z', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: i, j, ncol_local, nrow_local
    REAL(real_8)                        :: rnorm
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_z), POINTER  :: ar_data
    COMPLEX(kind=real_8)                         :: norm
    COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    TYPE(dbcsr_obj), POINTER                 :: input_vec, result_vec, swap_vec

    ar_data=>get_data_z(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_8) ; w_vec = CMPLX(0.0, 0.0, real_8)
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))

    DO j=control%current_step, control%max_iter-1

       ! compute the vector norm of the residuum, get it real valued as well (rnorm)
       CALL compute_norms_z(ar_data%f_vec, norm, rnorm, control%pcol_group)

       ! check convergence and inform everybody about it, a bit annoying to talk to everybody because of that
       IF(control%myproc==0)control%converged=rnorm.LT.REAL(control%threshold, real_8)
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT

       ! transfer normalized residdum to history and its norm to the Hessenberg matrix
       IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
       v_vec(:)=ar_data%f_vec(:)/rnorm; ar_data%local_history(:, j+1)=v_vec(:); ar_data%Hessenberg(j+1, j)=norm

       input_vec=>vectors%input_vec
       result_vec=>vectors%result_vec
       CALL transfer_local_array_to_dbcsr_z(input_vec, v_vec, nrow_local, control%local_comp)
 
       ! This permits to compute the subspace of a matrix which is a product of two matrices
       DO i=1, SIZE(matrix)
          CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, input_vec, result_vec, CMPLX(1.0, 0.0, real_8), &
                                            CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
          swap_vec=>input_vec
          input_vec=>result_vec
          result_vec=>swap_vec
       END DO

       CALL transfer_dbcsr_to_local_array_z(input_vec, w_vec, nrow_local, control%local_comp)

       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_z(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j+1, &
                               ar_data%local_history, ar_data%local_history, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can become a problem later on, there is probably a good check whether it's necessary, but we don't perform it
       CALL DGKS_ortho_z(h_vec, ar_data%f_vec, s_vec, nrow_local, j+1, ar_data%local_history, &
                                   ar_data%local_history, control%local_comp, control%pcol_group)
       ! Finally we can put the projections into our Hessenberg matrix
       ar_data%Hessenberg(1:j+1, j+1)= h_vec(1:j+1)
       control%current_step=j+1
    END DO

    ! compute the vector norm of the final residuum and put it in to Hessenberg
    CALL compute_norms_z(ar_data%f_vec, norm, rnorm, control%pcol_group)
    ar_data%Hessenberg(control%current_step+1, control%current_step)=norm
    
    ! broadcast the Hessenberg matrix so we don't need to care later on
    CALL mp_bcast(ar_data%Hessenberg, 0, control%mp_group)

    DEALLOCATE(v_vec, w_vec, h_vec, s_vec)

  END SUBROUTINE  build_subspace_z

! **************************************************************************************************
!> \brief Helper routine to transfer the all data of a dbcsr matrix to a local array
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_dbcsr_to_local_array_z(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    COMPLEX(kind=real_8), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    IF(is_local)array(1:n)=data_vec(1:n)

  END SUBROUTINE transfer_dbcsr_to_local_array_z

! **************************************************************************************************
!> \brief The inverse routine transfering data back from an array to a dbcsr
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_local_array_to_dbcsr_z(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    COMPLEX(kind=real_8), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    IF(is_local)data_vec(1:n)=array(1:n)

! **************************************************************************************************

  END SUBROUTINE transfer_local_array_to_dbcsr_z

! **************************************************************************************************
!> \brief Gram-Schmidt in matrix vector form
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param w_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE Gram_Schmidt_ortho_z(h_vec, f_vec, s_vec, w_vec, nrow_local,&
                                            j, local_history, reorth_mat, local_comp, pcol_group)
    COMPLEX(kind=real_8), DIMENSION(:)      :: h_vec, s_vec, f_vec, w_vec
    COMPLEX(kind=real_8), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group
    LOGICAL                                          :: local_comp

    CHARACTER(LEN=*), PARAMETER :: routineN = 'Gram_Schmidt_ortho_z', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    ! Let's do the orthonormalization, first try the Gram Schmidt scheme
    h_vec=CMPLX(0.0, 0.0, real_8); f_vec=CMPLX(0.0, 0.0, real_8); s_vec=CMPLX(0.0, 0.0, real_8)
    IF(local_comp)CALL zgemv('T', nrow_local, j, CMPLX(1.0, 0.0, real_8), local_history, &
                                      nrow_local, w_vec, 1, CMPLX(0.0, 0.0, real_8), h_vec, 1)
    CALL mp_sum(h_vec(1:j), pcol_group)

    IF(local_comp)CALL zgemv('N', nrow_local, j, CMPLX(1.0, 0.0, real_8), reorth_mat, &
                                      nrow_local, h_vec, 1, CMPLX(0.0, 0.0, real_8), f_vec, 1)
    f_vec(:)=w_vec(:)-f_vec(:)

    CALL timestop(handle)

  END SUBROUTINE Gram_Schmidt_ortho_z

! **************************************************************************************************
!> \brief Compute the  Daniel, Gragg, Kaufman and Steward correction
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE DGKS_ortho_z(h_vec, f_vec, s_vec, nrow_local, j, &
                                    local_history, reorth_mat, local_comp, pcol_group)
    COMPLEX(kind=real_8), DIMENSION(:)      :: h_vec, s_vec, f_vec
    COMPLEX(kind=real_8), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group

    CHARACTER(LEN=*), PARAMETER :: routineN = 'DGKS_ortho_z', &
         routineP = moduleN//':'//routineN

    LOGICAL                                          :: local_comp
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    IF(local_comp)CALL zgemv('T', nrow_local, j, CMPLX(1.0, 0.0, real_8), local_history, &
                                       nrow_local, f_vec, 1, CMPLX(0.0, 0.0, real_8), s_vec, 1)
    CALL mp_sum(s_vec(1:j), pcol_group)
    IF(local_comp)CALL zgemv('N', nrow_local, j, CMPLX(-1.0, 0.0, real_8), reorth_mat, &
                                       nrow_local, s_vec, 1, CMPLX(1.0, 0.0, real_8), f_vec, 1)
    h_vec(1:j)=h_vec(1:j)+s_vec(1:j)

    CALL timestop(handle)

  END SUBROUTINE DGKS_ortho_z

! **************************************************************************************************
!> \brief Compute the norm of a vector distributed along proc_col
!>        as local arrays. Always return the real part next to the complex rep.
!> \param vec ...
!> \param norm ...
!> \param rnorm ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE compute_norms_z(vec, norm, rnorm, pcol_group)
    COMPLEX(kind=real_8), DIMENSION(:)           :: vec
    REAL(real_8)                        :: rnorm
    COMPLEX(kind=real_8)                         :: norm
    INTEGER                                  :: pcol_group

    ! the norm is computed along the processor column
    norm=DOT_PRODUCT(vec, vec)
    CALL mp_sum(norm, pcol_group)
    rnorm=SQRT(REAL(norm, real_8))
    norm=CMPLX(rnorm, 0.0, real_8)

  END SUBROUTINE compute_norms_z

! **************************************************************************************************
!> \brief Computes the intial guess for the solution of the generalized eigenvalue 
!>        using the arnoldi method
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_arnoldi_init_z (matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_8)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    COMPLEX(kind=real_8)                         :: norm, denom
    COMPLEX(kind=real_8), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_z), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp

    ar_data=>get_data_z(arnoldi_data)
    
   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_8) ; w_vec = CMPLX(0.0, 0.0, real_8)
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_8)

    IF(control%has_initial_vector)THEN
    ! after calling the set routine the initial vector is stored in f_vec
        CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
    ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL zlarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF   

    CALL transfer_dbcsr_to_local_array_z(vectors%input_vec, v_vec, nrow_local, control%local_comp)
   
    ! compute the vector norm of the reandom vectorm, get it real valued as well (rnorm)
    CALL compute_norms_z(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_8)/rnorm)

    CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                      CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, w_vec, nrow_local, control%local_comp)
   
    ar_data%rho_scale=CMPLX(0.0, 0.0, real_8)
    ar_data%rho_scale=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(ar_data%rho_scale, pcol_group)

    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                      CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, w_vec, nrow_local, control%local_comp)
    
    denom=CMPLX(0.0, 0.0, real_8)
    denom=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(denom, pcol_group)
    IF(control%myproc==0) ar_data%rho_scale=ar_data%rho_scale/denom
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

    ! if the maximum ev is requested we need to optimize with -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, CMPLX(1.0, 0.0, real_8), -ar_data%rho_scale)
   
    ar_data%x_vec=v_vec

  END SUBROUTINE gev_arnoldi_init_z

! **************************************************************************************************
!> \brief builds the basis rothogonal wrt. teh metric.
!>        The structure looks similar to normal arnoldi but norms, vectors and 
!>        matrix_vector products are very differently defined. Therefore it is 
!>        cleaner to put it in a seperate subroutine to avoid confusion
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_build_subspace_z(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_z', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: j, ncol_local, nrow_local, pcol_group
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_z), POINTER            :: ar_data
    COMPLEX(kind=real_8)                         :: norm
    COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:)    :: Zmat, CZmat , BZmat 

    ar_data=>get_data_z(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.
    pcol_group=control%pcol_group

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_8) ; w_vec = CMPLX(0.0, 0.0, real_8)
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))
    ALLOCATE(Zmat(nrow_local,control%max_iter)); ALLOCATE(CZmat(nrow_local,control%max_iter))
    ALLOCATE(BZmat(nrow_local,control%max_iter))

    CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, ar_data%x_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                      CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, BZmat(:,1), nrow_local, control%local_comp)
    
    norm=CMPLX(0.0, 0.0, real_8) 
    norm=DOT_PRODUCT(ar_data%x_vec,BZmat(:,1)) 
    CALL mp_sum(norm, pcol_group)
    IF(control%local_comp)THEN
       Zmat(:,1)=ar_data%x_vec/SQRT(norm);  BZmat(:,1)= BZmat(:,1)/SQRT(norm)
    END IF

    DO j=1,control%max_iter
       control%current_step=j
       CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, Zmat(:,j), nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                        CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, CZmat(:,j), nrow_local, control%local_comp)
       w_vec(:)=CZmat(:,j)                         
       
       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_z(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j, &
                               BZmat, Zmat, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can becom a problem later on, there is probably a good check, but we don't perform it
       CALL DGKS_ortho_z(h_vec, ar_data%f_vec, s_vec, nrow_local, j, BZmat, &
                                    Zmat, control%local_comp, control%pcol_group)
    
       CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                        CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, v_vec, nrow_local, control%local_comp)
       norm=CMPLX(0.0, 0.0, real_8)
       norm=DOT_PRODUCT(ar_data%f_vec,v_vec)
       CALL mp_sum(norm, pcol_group)
      
       IF(control%myproc==0)control%converged=REAL(norm,real_8).LT.EPSILON(REAL(1.0,real_8))
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT 
       IF(j==control%max_iter-1)EXIT

       IF(control%local_comp)THEN
          Zmat(:,j+1)=ar_data%f_vec/SQRT(norm);  BZmat(:,j+1)= v_vec(:)/SQRT(norm)
       END IF
    END DO

! getting a bit more complicated here as the final matrix is again a product which has to be computed with the 
! ditributed vectors, therefore a sum along the first proc_col is necessary. As we want that matrix everywhere,
! we set it to zero before and compute the distributed product only on the first col and then sum over the full grid
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_8)
    IF(control%local_comp)THEN
       ar_data%Hessenberg(1:control%current_step,1:control%current_step)=&
          MATMUL(TRANSPOSE(CZmat(:,1:control%current_step)),Zmat(:,1:control%current_step))
    END IF
    CALL mp_sum(ar_data%Hessenberg,control%mp_group)

    ar_data%local_history=Zmat
    ! broadcast the Hessenberg matrix so we don't need to care later on

    DEALLOCATE(v_vec); DEALLOCATE(w_vec); DEALLOCATE(s_vec); DEALLOCATE(h_vec); DEALLOCATE(CZmat);
    DEALLOCATE(Zmat); DEALLOCATE(BZmat)

  END SUBROUTINE gev_build_subspace_z

! **************************************************************************************************
!> \brief Updates all data after an inner loop of the generalized ev arnoldi. 
!>        Updates rho and C=A-rho*B accordingly.
!>        As an update scheme is used for he ev, the output ev has to be replaced
!>        with the updated one.
!>        Furthermore a convergence check is performed. The mv product could
!>        be skiiped by making clever use of the precomputed data, However,
!>        it is most likely not worth the effort.
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_update_data_z(matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_update_data_z', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_z), POINTER            :: ar_data
    INTEGER                                  :: ncol_local, nrow_local, ind, i
    COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:)      :: v_vec
    REAL(real_8)                        :: rnorm
    COMPLEX(kind=real_8)                         :: norm
    COMPLEX(real_8)                     :: val 

    control=>get_control(arnoldi_data)

    ar_data=>get_data_z(arnoldi_data)

! compute the new shift, hack around the problem templating the conversion
    val=ar_data%evals(control%selected_ind(1))
    ar_data%rho_scale=ar_data%rho_scale+val
! compute the new eigenvector / initial guess for the next arnoldi loop    
    ar_data%x_vec=CMPLX(0.0, 0.0, real_8)
    DO i=1,control%current_step
       val=ar_data%revec(i,control%selected_ind(1))
       ar_data%x_vec(:)=ar_data%x_vec(:)+ar_data%local_history(:,i)*val
    END DO
!      ar_data%x_vec(:)=MATMUL(ar_data%local_history(:,1:control%current_step),&
!                        ar_data%revec(1:control%current_step,control%selected_ind(1)))

! update the C-matrix (A-rho*B), if teh maximum value is requested we have to use -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, CMPLX(1.0, 0.0, real_8), -ar_data%rho_scale)

! compute convergence and set the correct eigenvalue and eigenvector
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    IF (ncol_local>0) THEN
       ALLOCATE(v_vec(nrow_local))
       CALL compute_norms_z(ar_data%x_vec, norm, rnorm, control%pcol_group)
       v_vec(:)=ar_data%x_vec(:)/rnorm
    ENDIF
    CALL transfer_local_array_to_dbcsr_z(vectors%input_vec, v_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix_arnoldi(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_8), &
                                            CMPLX(0.0, 0.0, real_8), vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_z(vectors%result_vec, v_vec, nrow_local, control%local_comp)
    IF (ncol_local>0) THEN
       CALL compute_norms_z(v_vec, norm, rnorm, control%pcol_group)
       ! check convergence 
       control%converged=rnorm.LT.control%threshold
       DEALLOCATE(v_vec)
    ENDIF
    ! and broadcast the real eigenvalue    
    CALL mp_bcast(control%converged,0,control%mp_group)
    ind=control%selected_ind(1)
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

! Again the maximum value request is done on -A therefore the eigenvalue needs the opposite sign
    ar_data%evals(ind)=ar_data%rho_scale


  END SUBROUTINE gev_update_data_z


! **************************************************************************************************
!> \brief Call the correct eigensolver, in the arnoldi method only the right
!>        eigenvectors are used. Lefts are created here but dumped immediatly 
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE compute_evals_c(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_evals_c', &
      routineP = moduleN//':'//routineN

    COMPLEX(real_4), DIMENSION(:, :), ALLOCATABLE   :: levec
    TYPE(arnoldi_data_c), POINTER            :: ar_data
    INTEGER                                  :: ndim
    TYPE(arnoldi_control), POINTER           :: control

    ar_data=>get_data_c(arnoldi_data)
    control=> get_control(arnoldi_data)
    ndim=control%current_step
    ALLOCATE(levec(ndim, ndim))

! Needs antoher interface as the calls to real and complex geev differ (sucks!)
! only perform the diagonalization on processors which hold data
    IF(control%generalized_ev)THEN
       CALL dbcsr_symm_local_diag('V',ar_data%Hessenberg(1:ndim, 1:ndim), ndim,&
                                  ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim))
    ELSE
       IF(control%symmetric)THEN
          CALL dbcsr_tridiag_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       ELSE
          CALL dbcsr_general_local_diag('N', 'V', ar_data%Hessenberg(1:ndim, 1:ndim), ndim, &
                                         ar_data%evals(1:ndim), ar_data%revec(1:ndim, 1:ndim), levec)
       END IF
    END IF

    DEALLOCATE(levec)

  END SUBROUTINE compute_evals_c

! **************************************************************************************************
!> \brief Initialization of the arnoldi vector. Here a random vector is used,
!>        might be generalized in the future 
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_init_c (matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i, iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_4)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    COMPLEX(kind=real_4)                         :: norm 
    COMPLEX(kind=real_4), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_c), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp
    
    ar_data=>get_data_c(arnoldi_data)

   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_4) ; w_vec = CMPLX(0.0, 0.0, real_4)
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_4)

    IF(control%has_initial_vector)THEN
       ! after calling the set routine the initial vector is stored in f_vec
       CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
       ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL clarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF

    CALL transfer_dbcsr_to_local_array_c(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! compute the vector norm of the random vectorm, get it real valued as well (rnorm)
    CALL compute_norms_c(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_4)/rnorm)

    ! Everything prepared, initialize the Arnoldi iteration
    CALL transfer_dbcsr_to_local_array_c(vectors%input_vec, v_vec, nrow_local, control%local_comp)

    ! This permits to compute the subspace of a matrix which is a product of multiple matrices
    DO i=1, SIZE(matrix)
       CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                        CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL dbcsr_copy(vectors%input_vec, vectors%result_vec)
    END DO

    CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, w_vec, nrow_local, control%local_comp)

    ! Put the projection into the Hessenberg matrix, and make the vectors orthonormal
    ar_data%Hessenberg(1, 1)=DOT_PRODUCT(v_vec, w_vec)
    CALL mp_sum(ar_data%Hessenberg(1, 1), pcol_group)
    ar_data%f_vec=w_vec-v_vec*ar_data%Hessenberg(1, 1)

    ar_data%local_history(:, 1)=v_vec(:)

    ! We did the first step in here so we should set the current step for the subspace generation accordingly
    control%current_step=1

    DEALLOCATE(v_vec, w_vec)

  END SUBROUTINE arnoldi_init_c

! **************************************************************************************************
!> \brief Alogorithm for the implicit restarts in the arnoldi method
!>        this is an early implementaion which scales subspace size^4
!>        by replacing the lapack calls with direct math the 
!>        QR and  gemms can be made linear and a N^2 sacling will be acchieved
!>        however this already sets the framework but should be used with care
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE arnoldi_iram_c(arnoldi_data)
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_iram_c', &
         routineP = moduleN//':'//routineN

    TYPE(arnoldi_data_c), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER                     :: control
    COMPLEX(real_4), DIMENSION(:,:), ALLOCATABLE  :: tmp_mat, safe_mat, Q, tmp_mat1
    COMPLEX(real_4), DIMENSION(:), ALLOCATABLE    :: work, tau, work_measure
    INTEGER                                   :: msize, lwork, i, j, info, nwant
    COMPLEX(kind=real_4)                          :: beta, sigma
    COMPLEX(kind=real_4),DIMENSION(:,:),ALLOCATABLE :: Qdata


    ! This is just a terribly inefficient implementation but I hope it is correct and might serve as a reference
    ar_data=>get_data_c(arnoldi_data)
    control=>get_control(arnoldi_data)
    msize=control%current_step
    nwant=control%nval_out
    ALLOCATE(tmp_mat(msize,msize)); ALLOCATE(safe_mat(msize,msize))
    ALLOCATE(Q(msize,msize)); ALLOCATE(tmp_mat1(msize,msize))
    ALLOCATE(work_measure(1))
    ALLOCATE(tau(msize)); ALLOCATE(Qdata(msize,msize))
    !make Q identity
    Q=CMPLX(0.0, 0.0, real_4)
    DO i=1,msize
       Q(i,i)=CMPLX(1.0, 0.0, real_4)
    END DO

    ! Looks a bit odd, but safe_mat will contain the result in the end, while tmpmat gets violated by lapack
    CALL convert_matrix(tmp_mat,ar_data%Hessenberg(1:msize,1:msize))
    safe_mat(:,:)=tmp_mat(:,:)

    DO i=1,msize
       ! A bit a strange check but in the end we only want to shift the unwanted evals
       IF(ANY(control%selected_ind==i))CYCLE
       ! Here we shift the matrix by subtracting unwanted evals from the diagonal
       DO j=1,msize
          tmp_mat(j,j)=tmp_mat(j,j)-ar_data%evals(i)
       END DO
       ! Now we repair the damage by QR factorizing
       lwork=-1
       CALL cgeqrf(msize,msize,tmp_mat,msize,tau,work_measure,lwork,info)
       lwork=INT(work_measure(1))
       IF (ALLOCATED(work)) THEN
          IF (SIZE(work).LT.lwork) THEN
             DEALLOCATE(work)
          ENDIF
       ENDIF
       IF (.NOT.ALLOCATED(work)) ALLOCATE(work(lwork))
       CALL cgeqrf(msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! Ask Lapack to reconstruct Q from its own way of storing data (tmpmat will contain Q)
       CALL cungqr(msize,msize,msize,tmp_mat,msize,tau,work,lwork,info)
       ! update Q=Q*Q_current
       tmp_mat1(:,:)=Q(:,:)
       CALL cgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         Q,msize)       
       ! Update H=(Q*)HQ
       CALL cgemm('C','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat,msize,safe_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         tmp_mat1,msize)
       CALL cgemm('N','N',msize,msize,msize,CMPLX(1.0, 0.0, real_4),tmp_mat1,msize,tmp_mat,msize,CMPLX(0.0, 0.0, real_4),&
                         safe_mat,msize)

       ! this one is crucial for numerics not to accumulate noise in the subdiagonals
       DO j=1,msize
          safe_mat(j+2:msize,j)=CMPLX(0.0, 0.0, real_4)
       END DO
       tmp_mat(:,:)=safe_mat(:,:)
    END DO

    ! Now we can compute our restart quantities
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_4)
    CALL convert_matrix(ar_data%Hessenberg(1:msize,1:msize),safe_mat)
    CALL convert_matrix(Qdata,Q)
  
    beta=ar_data%Hessenberg(nwant+1,nwant); sigma=Qdata(msize,nwant)

    !update the residuum and the basis vectors
    IF(control%local_comp)THEN
       ar_data%f_vec=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,nwant+1))*beta+ar_data%f_vec(:)*sigma
       ar_data%local_history(:,1:nwant)=MATMUL(ar_data%local_history(:,1:msize),Qdata(1:msize,1:nwant))
    END IF
    ! Set the current step to nwant so the subspace build knows where to start
    control%current_step=nwant
    
    DEALLOCATE(tmp_mat,safe_mat,Q,Qdata,tmp_mat1,work,tau,work_measure)
    
  END SUBROUTINE arnoldi_iram_c

! **************************************************************************************************
!> \brief Here we create the Krylov subspace and fill the Hessenberg matrix
!>        convergence check is only performed on subspace convergence
!>        Gram Schidt is used to orthonogonalize. 
!>        If this is numericall not sufficient a Daniel, Gragg, Kaufman and Steward
!>        correction is performed
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************
  SUBROUTINE build_subspace_c(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors), TARGET              :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_c', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: i, j, ncol_local, nrow_local
    REAL(real_4)                        :: rnorm
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_c), POINTER  :: ar_data
    COMPLEX(kind=real_4)                         :: norm
    COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    TYPE(dbcsr_obj), POINTER                 :: input_vec, result_vec, swap_vec

    ar_data=>get_data_c(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_4) ; w_vec = CMPLX(0.0, 0.0, real_4)
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))

    DO j=control%current_step, control%max_iter-1

       ! compute the vector norm of the residuum, get it real valued as well (rnorm)
       CALL compute_norms_c(ar_data%f_vec, norm, rnorm, control%pcol_group)

       ! check convergence and inform everybody about it, a bit annoying to talk to everybody because of that
       IF(control%myproc==0)control%converged=rnorm.LT.REAL(control%threshold, real_4)
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT

       ! transfer normalized residdum to history and its norm to the Hessenberg matrix
       IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
       v_vec(:)=ar_data%f_vec(:)/rnorm; ar_data%local_history(:, j+1)=v_vec(:); ar_data%Hessenberg(j+1, j)=norm

       input_vec=>vectors%input_vec
       result_vec=>vectors%result_vec
       CALL transfer_local_array_to_dbcsr_c(input_vec, v_vec, nrow_local, control%local_comp)
 
       ! This permits to compute the subspace of a matrix which is a product of two matrices
       DO i=1, SIZE(matrix)
          CALL dbcsr_matrix_colvec_multiply(matrix(i)%matrix, input_vec, result_vec, CMPLX(1.0, 0.0, real_4), &
                                            CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
          swap_vec=>input_vec
          input_vec=>result_vec
          result_vec=>swap_vec
       END DO

       CALL transfer_dbcsr_to_local_array_c(input_vec, w_vec, nrow_local, control%local_comp)

       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_c(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j+1, &
                               ar_data%local_history, ar_data%local_history, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can become a problem later on, there is probably a good check whether it's necessary, but we don't perform it
       CALL DGKS_ortho_c(h_vec, ar_data%f_vec, s_vec, nrow_local, j+1, ar_data%local_history, &
                                   ar_data%local_history, control%local_comp, control%pcol_group)
       ! Finally we can put the projections into our Hessenberg matrix
       ar_data%Hessenberg(1:j+1, j+1)= h_vec(1:j+1)
       control%current_step=j+1
    END DO

    ! compute the vector norm of the final residuum and put it in to Hessenberg
    CALL compute_norms_c(ar_data%f_vec, norm, rnorm, control%pcol_group)
    ar_data%Hessenberg(control%current_step+1, control%current_step)=norm
    
    ! broadcast the Hessenberg matrix so we don't need to care later on
    CALL mp_bcast(ar_data%Hessenberg, 0, control%mp_group)

    DEALLOCATE(v_vec, w_vec, h_vec, s_vec)

  END SUBROUTINE  build_subspace_c

! **************************************************************************************************
!> \brief Helper routine to transfer the all data of a dbcsr matrix to a local array
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_dbcsr_to_local_array_c(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    COMPLEX(kind=real_4), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    IF(is_local)array(1:n)=data_vec(1:n)

  END SUBROUTINE transfer_dbcsr_to_local_array_c

! **************************************************************************************************
!> \brief The inverse routine transfering data back from an array to a dbcsr
!> \param vec ...
!> \param array ...
!> \param n ...
!> \param is_local ...
! **************************************************************************************************
  SUBROUTINE transfer_local_array_to_dbcsr_c(vec, array, n, is_local)
    TYPE(dbcsr_obj)                          :: vec
    COMPLEX(kind=real_4), DIMENSION(:)           :: array
    INTEGER                                  :: n
    LOGICAL                                  :: is_local
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec

    data_vec => dbcsr_get_data_p (vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    IF(is_local)data_vec(1:n)=array(1:n)

! **************************************************************************************************

  END SUBROUTINE transfer_local_array_to_dbcsr_c

! **************************************************************************************************
!> \brief Gram-Schmidt in matrix vector form
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param w_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE Gram_Schmidt_ortho_c(h_vec, f_vec, s_vec, w_vec, nrow_local,&
                                            j, local_history, reorth_mat, local_comp, pcol_group)
    COMPLEX(kind=real_4), DIMENSION(:)      :: h_vec, s_vec, f_vec, w_vec
    COMPLEX(kind=real_4), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group
    LOGICAL                                          :: local_comp

    CHARACTER(LEN=*), PARAMETER :: routineN = 'Gram_Schmidt_ortho_c', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    ! Let's do the orthonormalization, first try the Gram Schmidt scheme
    h_vec=CMPLX(0.0, 0.0, real_4); f_vec=CMPLX(0.0, 0.0, real_4); s_vec=CMPLX(0.0, 0.0, real_4)
    IF(local_comp)CALL cgemv('T', nrow_local, j, CMPLX(1.0, 0.0, real_4), local_history, &
                                      nrow_local, w_vec, 1, CMPLX(0.0, 0.0, real_4), h_vec, 1)
    CALL mp_sum(h_vec(1:j), pcol_group)

    IF(local_comp)CALL cgemv('N', nrow_local, j, CMPLX(1.0, 0.0, real_4), reorth_mat, &
                                      nrow_local, h_vec, 1, CMPLX(0.0, 0.0, real_4), f_vec, 1)
    f_vec(:)=w_vec(:)-f_vec(:)

    CALL timestop(handle)

  END SUBROUTINE Gram_Schmidt_ortho_c

! **************************************************************************************************
!> \brief Compute the  Daniel, Gragg, Kaufman and Steward correction
!> \param h_vec ...
!> \param f_vec ...
!> \param s_vec ...
!> \param nrow_local ...
!> \param j ...
!> \param local_history ...
!> \param reorth_mat ...
!> \param local_comp ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE DGKS_ortho_c(h_vec, f_vec, s_vec, nrow_local, j, &
                                    local_history, reorth_mat, local_comp, pcol_group)
    COMPLEX(kind=real_4), DIMENSION(:)      :: h_vec, s_vec, f_vec
    COMPLEX(kind=real_4), DIMENSION(:, :)    :: local_history, reorth_mat
    INTEGER                                          :: nrow_local, j, pcol_group

    CHARACTER(LEN=*), PARAMETER :: routineN = 'DGKS_ortho_c', &
         routineP = moduleN//':'//routineN

    LOGICAL                                          :: local_comp
    INTEGER                                  :: handle

    CALL timeset(routineN, handle)

    IF(local_comp)CALL cgemv('T', nrow_local, j, CMPLX(1.0, 0.0, real_4), local_history, &
                                       nrow_local, f_vec, 1, CMPLX(0.0, 0.0, real_4), s_vec, 1)
    CALL mp_sum(s_vec(1:j), pcol_group)
    IF(local_comp)CALL cgemv('N', nrow_local, j, CMPLX(-1.0, 0.0, real_4), reorth_mat, &
                                       nrow_local, s_vec, 1, CMPLX(1.0, 0.0, real_4), f_vec, 1)
    h_vec(1:j)=h_vec(1:j)+s_vec(1:j)

    CALL timestop(handle)

  END SUBROUTINE DGKS_ortho_c

! **************************************************************************************************
!> \brief Compute the norm of a vector distributed along proc_col
!>        as local arrays. Always return the real part next to the complex rep.
!> \param vec ...
!> \param norm ...
!> \param rnorm ...
!> \param pcol_group ...
! **************************************************************************************************
  SUBROUTINE compute_norms_c(vec, norm, rnorm, pcol_group)
    COMPLEX(kind=real_4), DIMENSION(:)           :: vec
    REAL(real_4)                        :: rnorm
    COMPLEX(kind=real_4)                         :: norm
    INTEGER                                  :: pcol_group

    ! the norm is computed along the processor column
    norm=DOT_PRODUCT(vec, vec)
    CALL mp_sum(norm, pcol_group)
    rnorm=SQRT(REAL(norm, real_4))
    norm=CMPLX(rnorm, 0.0, real_4)

  END SUBROUTINE compute_norms_c

! **************************************************************************************************
!> \brief Computes the intial guess for the solution of the generalized eigenvalue 
!>        using the arnoldi method
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_arnoldi_init_c (matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'arnoldi_init_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: iseed(4), row_size, col_size, &
                                                nrow_local, ncol_local, pcol_group, &
                                                row, col
    REAL(real_4)                        :: rnorm
    TYPE(dbcsr_iterator)                     :: iter
    COMPLEX(kind=real_4)                         :: norm, denom
    COMPLEX(kind=real_4), DIMENSION(:), ALLOCATABLE :: &
                                                v_vec, w_vec
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    LOGICAL                                  :: transposed, local_comp
    TYPE(arnoldi_data_c), POINTER            :: ar_data
    TYPE(arnoldi_control), POINTER           :: control

    control=>get_control(arnoldi_data)
    pcol_group=control%pcol_group
    local_comp=control%local_comp

    ar_data=>get_data_c(arnoldi_data)
    
   ! create a local data copy to store the vectors and make Gram Schmidt a bit simpler
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local))
    ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_4) ; w_vec = CMPLX(0.0, 0.0, real_4)
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_4)

    IF(control%has_initial_vector)THEN
    ! after calling the set routine the initial vector is stored in f_vec
        CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
    ELSE
    ! Setup the initial normalized random vector (sufficient if it only happens on proc_col 0)
       CALL dbcsr_iterator_start(iter, vectors%input_vec)
       DO WHILE (dbcsr_iterator_blocks_left(iter))
          CALL dbcsr_iterator_next_block(iter, row, col, data_vec, transposed, row_size=row_size, col_size=col_size)
          iseed(1)=2; iseed(2)=MOD(row, 4095); iseed(3)=MOD(col, 4095); iseed(4)=11
          CALL clarnv(2, iseed, row_size*col_size, data_vec)
       END DO
       CALL dbcsr_iterator_stop(iter)
    END IF   

    CALL transfer_dbcsr_to_local_array_c(vectors%input_vec, v_vec, nrow_local, control%local_comp)
   
    ! compute the vector norm of the reandom vectorm, get it real valued as well (rnorm)
    CALL compute_norms_c(v_vec, norm, rnorm, control%pcol_group)

    IF (rnorm==0) rnorm=1 ! catch case where this rank has no actual data
    CALL dbcsr_scale(vectors%input_vec, REAL(1.0, real_4)/rnorm)

    CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                      CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, w_vec, nrow_local, control%local_comp)
   
    ar_data%rho_scale=CMPLX(0.0, 0.0, real_4)
    ar_data%rho_scale=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(ar_data%rho_scale, pcol_group)

    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                      CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)

    CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, w_vec, nrow_local, control%local_comp)
    
    denom=CMPLX(0.0, 0.0, real_4)
    denom=DOT_PRODUCT(v_vec,w_vec)
    CALL mp_sum(denom, pcol_group)
    IF(control%myproc==0) ar_data%rho_scale=ar_data%rho_scale/denom
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

    ! if the maximum ev is requested we need to optimize with -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, CMPLX(1.0, 0.0, real_4), -ar_data%rho_scale)
   
    ar_data%x_vec=v_vec

  END SUBROUTINE gev_arnoldi_init_c

! **************************************************************************************************
!> \brief builds the basis rothogonal wrt. teh metric.
!>        The structure looks similar to normal arnoldi but norms, vectors and 
!>        matrix_vector products are very differently defined. Therefore it is 
!>        cleaner to put it in a seperate subroutine to avoid confusion
!> \param matrix ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_build_subspace_c(matrix, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'build_subspace_c', &
         routineP = moduleN//':'//routineN

    INTEGER                                  :: j, ncol_local, nrow_local, pcol_group
    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_c), POINTER            :: ar_data
    COMPLEX(kind=real_4)                         :: norm
    COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: h_vec, s_vec, v_vec, w_vec
    COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:)    :: Zmat, CZmat , BZmat 

    ar_data=>get_data_c(arnoldi_data)
    control=>get_control(arnoldi_data)
    control%converged=.FALSE.
    pcol_group=control%pcol_group

    ! create the vectors required during the iterations
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    ALLOCATE(v_vec(nrow_local));  ALLOCATE(w_vec(nrow_local))
    v_vec = CMPLX(0.0, 0.0, real_4) ; w_vec = CMPLX(0.0, 0.0, real_4)
    ALLOCATE(s_vec(control%max_iter)); ALLOCATE(h_vec(control%max_iter))
    ALLOCATE(Zmat(nrow_local,control%max_iter)); ALLOCATE(CZmat(nrow_local,control%max_iter))
    ALLOCATE(BZmat(nrow_local,control%max_iter))

    CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, ar_data%x_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                      CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, BZmat(:,1), nrow_local, control%local_comp)
    
    norm=CMPLX(0.0, 0.0, real_4) 
    norm=DOT_PRODUCT(ar_data%x_vec,BZmat(:,1)) 
    CALL mp_sum(norm, pcol_group)
    IF(control%local_comp)THEN
       Zmat(:,1)=ar_data%x_vec/SQRT(norm);  BZmat(:,1)= BZmat(:,1)/SQRT(norm)
    END IF

    DO j=1,control%max_iter
       control%current_step=j
       CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, Zmat(:,j), nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                        CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, CZmat(:,j), nrow_local, control%local_comp)
       w_vec(:)=CZmat(:,j)                         
       
       ! Let's do the orthonormalization, to get the new f_vec. First try the Gram Schmidt scheme
       CALL Gram_Schmidt_ortho_c(h_vec, ar_data%f_vec, s_vec, w_vec, nrow_local, j, &
                               BZmat, Zmat, control%local_comp, control%pcol_group)

       ! A bit more expensive but simpliy always top up with a DGKS correction, otherwise numerics
       ! can becom a problem later on, there is probably a good check, but we don't perform it
       CALL DGKS_ortho_c(h_vec, ar_data%f_vec, s_vec, nrow_local, j, BZmat, &
                                    Zmat, control%local_comp, control%pcol_group)
    
       CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, ar_data%f_vec, nrow_local, control%local_comp)
       CALL dbcsr_matrix_colvec_multiply(matrix(2)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                        CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
       CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, v_vec, nrow_local, control%local_comp)
       norm=CMPLX(0.0, 0.0, real_4)
       norm=DOT_PRODUCT(ar_data%f_vec,v_vec)
       CALL mp_sum(norm, pcol_group)
      
       IF(control%myproc==0)control%converged=REAL(norm,real_4).LT.EPSILON(REAL(1.0,real_4))
       CALL mp_bcast(control%converged, 0, control%mp_group)
       IF(control%converged)EXIT 
       IF(j==control%max_iter-1)EXIT

       IF(control%local_comp)THEN
          Zmat(:,j+1)=ar_data%f_vec/SQRT(norm);  BZmat(:,j+1)= v_vec(:)/SQRT(norm)
       END IF
    END DO

! getting a bit more complicated here as the final matrix is again a product which has to be computed with the 
! ditributed vectors, therefore a sum along the first proc_col is necessary. As we want that matrix everywhere,
! we set it to zero before and compute the distributed product only on the first col and then sum over the full grid
    ar_data%Hessenberg=CMPLX(0.0, 0.0, real_4)
    IF(control%local_comp)THEN
       ar_data%Hessenberg(1:control%current_step,1:control%current_step)=&
          MATMUL(TRANSPOSE(CZmat(:,1:control%current_step)),Zmat(:,1:control%current_step))
    END IF
    CALL mp_sum(ar_data%Hessenberg,control%mp_group)

    ar_data%local_history=Zmat
    ! broadcast the Hessenberg matrix so we don't need to care later on

    DEALLOCATE(v_vec); DEALLOCATE(w_vec); DEALLOCATE(s_vec); DEALLOCATE(h_vec); DEALLOCATE(CZmat);
    DEALLOCATE(Zmat); DEALLOCATE(BZmat)

  END SUBROUTINE gev_build_subspace_c

! **************************************************************************************************
!> \brief Updates all data after an inner loop of the generalized ev arnoldi. 
!>        Updates rho and C=A-rho*B accordingly.
!>        As an update scheme is used for he ev, the output ev has to be replaced
!>        with the updated one.
!>        Furthermore a convergence check is performed. The mv product could
!>        be skiiped by making clever use of the precomputed data, However,
!>        it is most likely not worth the effort.
!> \param matrix ...
!> \param matrix_arnoldi ...
!> \param vectors ...
!> \param arnoldi_data ...
! **************************************************************************************************

  SUBROUTINE gev_update_data_c(matrix, matrix_arnoldi, vectors, arnoldi_data)
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix
    TYPE(dbcsr_obj_type_p), DIMENSION(:)     :: matrix_arnoldi
    TYPE(m_x_v_vectors)                      :: vectors
    TYPE(dbcsr_arnoldi_data)                 :: arnoldi_data

    CHARACTER(LEN=*), PARAMETER :: routineN = 'gev_update_data_c', &
      routineP = moduleN//':'//routineN

    TYPE(arnoldi_control), POINTER           :: control
    TYPE(arnoldi_data_c), POINTER            :: ar_data
    INTEGER                                  :: ncol_local, nrow_local, ind, i
    COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:)      :: v_vec
    REAL(real_4)                        :: rnorm
    COMPLEX(kind=real_4)                         :: norm
    COMPLEX(real_4)                     :: val 

    control=>get_control(arnoldi_data)

    ar_data=>get_data_c(arnoldi_data)

! compute the new shift, hack around the problem templating the conversion
    val=ar_data%evals(control%selected_ind(1))
    ar_data%rho_scale=ar_data%rho_scale+val
! compute the new eigenvector / initial guess for the next arnoldi loop    
    ar_data%x_vec=CMPLX(0.0, 0.0, real_4)
    DO i=1,control%current_step
       val=ar_data%revec(i,control%selected_ind(1))
       ar_data%x_vec(:)=ar_data%x_vec(:)+ar_data%local_history(:,i)*val
    END DO
!      ar_data%x_vec(:)=MATMUL(ar_data%local_history(:,1:control%current_step),&
!                        ar_data%revec(1:control%current_step,control%selected_ind(1)))

! update the C-matrix (A-rho*B), if teh maximum value is requested we have to use -A-rho*B
    CALL dbcsr_copy(matrix_arnoldi(1)%matrix,matrix(1)%matrix)
    CALL dbcsr_add(matrix_arnoldi(1)%matrix, matrix(2)%matrix, CMPLX(1.0, 0.0, real_4), -ar_data%rho_scale)

! compute convergence and set the correct eigenvalue and eigenvector
    CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
    IF (ncol_local>0) THEN
       ALLOCATE(v_vec(nrow_local))
       CALL compute_norms_c(ar_data%x_vec, norm, rnorm, control%pcol_group)
       v_vec(:)=ar_data%x_vec(:)/rnorm
    ENDIF
    CALL transfer_local_array_to_dbcsr_c(vectors%input_vec, v_vec, nrow_local, control%local_comp)
    CALL dbcsr_matrix_colvec_multiply(matrix_arnoldi(1)%matrix, vectors%input_vec, vectors%result_vec, CMPLX(1.0, 0.0, real_4), &
                                            CMPLX(0.0, 0.0, real_4), vectors%rep_row_vec, vectors%rep_col_vec)
    CALL transfer_dbcsr_to_local_array_c(vectors%result_vec, v_vec, nrow_local, control%local_comp)
    IF (ncol_local>0) THEN
       CALL compute_norms_c(v_vec, norm, rnorm, control%pcol_group)
       ! check convergence 
       control%converged=rnorm.LT.control%threshold
       DEALLOCATE(v_vec)
    ENDIF
    ! and broadcast the real eigenvalue    
    CALL mp_bcast(control%converged,0,control%mp_group)
    ind=control%selected_ind(1)
    CALL mp_bcast(ar_data%rho_scale,0,control%mp_group)

! Again the maximum value request is done on -A therefore the eigenvalue needs the opposite sign
    ar_data%evals(ind)=ar_data%rho_scale


  END SUBROUTINE gev_update_data_c


END MODULE dbcsr_arnoldi_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief arnoldi iteration using dbcsr
!> \par History
!>       2014.09 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_arnoldi
   USE dbcsr_arnoldi_data_methods,      ONLY: arnoldi_is_converged,&
                                              deallocate_arnoldi_data,&
                                              get_nrestart,&
                                              get_selected_ritz_val,&
                                              get_selected_ritz_vector,&
                                              select_evals,&
                                              set_arnoldi_initial_vector,&
                                              setup_arnoldi_data
   USE dbcsr_arnoldi_methods,           ONLY: arnoldi_init,&
                                              arnoldi_iram,&
                                              build_subspace,&
                                              compute_evals,&
                                              gev_arnoldi_init,&
                                              gev_build_subspace,&
                                              gev_update_data
   USE dbcsr_arnoldi_types,             ONLY: arnoldi_control,&
                                              dbcsr_arnoldi_data,&
                                              get_control,&
                                              m_x_v_vectors
   USE dbcsr_methods,                   ONLY: dbcsr_init,&
                                              dbcsr_release
   USE dbcsr_operations,                ONLY: dbcsr_copy,&
                                              dbcsr_get_info
   USE dbcsr_types,                     ONLY: dbcsr_obj,&
                                              dbcsr_obj_type_p
   USE dbcsr_vector_operations,         ONLY: create_col_vec_from_matrix,&
                                              create_replicated_col_vec_from_matrix,&
                                              create_replicated_row_vec_from_matrix
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_arnoldi'

   PUBLIC :: dbcsr_arnoldi_ev, dbcsr_arnoldi_extremal

CONTAINS

! **************************************************************************************************
!> \brief Driver routine for different arnoldi eigenvalue methods
!>        the selection which one is to be taken is made beforehand in the
!>        setup call passing the generalized_ev flag or not
!> \param matrix ...
!> \param arnoldi_data ...
! **************************************************************************************************

   SUBROUTINE dbcsr_arnoldi_ev(matrix, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_arnoldi_ev', &
         routineP = moduleN//':'//routineN

      TYPE(arnoldi_control), POINTER                     :: control

      control => get_control(arnoldi_data)

      IF (control%generalized_ev) THEN
         CALL dbcsr_arnoldi_generalized_ev(matrix, arnoldi_data)
      ELSE
         CALL dbcsr_arnoldi_normal_ev(matrix, arnoldi_data)
      END IF

   END SUBROUTINE dbcsr_arnoldi_ev

! **************************************************************************************************
!> \brief The main routine for arnoldi method to compute ritz values
!>        vectors of a matrix. Can take multiple matrices to solve
!>        ( M(N)*...*M(2)*M(1) )*v=v*e. A, B, ... have to be merged in a array of pointers
!>        arnoldi data has to be create with the setup routine and
!>        will contain on input all necessary information to start/restart
!>        the calculation. On output it contains all data
!> \param matrix a pointer array to dbcsr_matrices. Multiplication order is as
!>        described above
!> \param arnoldi_data On input data_type contains all information to start/restart
!>                     an arnoldi iteration
!>                     On output all data areas are filled to allow arbitrary post
!>                     processing of the created subspace
!>                     arnoldi_data has to be created with setup_arnoldi_data
! **************************************************************************************************
   SUBROUTINE dbcsr_arnoldi_normal_ev(matrix, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_arnoldi_normal_ev', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_loop, ncol_local, nrow_local
      TYPE(arnoldi_control), POINTER                     :: control
      TYPE(dbcsr_obj), POINTER                           :: restart_vec
      TYPE(m_x_v_vectors)                                :: vectors

      NULLIFY (restart_vec)
      CALL timeset(routineN, handle)

!prepare the vector like matrives needed in the matrix vector products, they will be reused throughout the iterations
      CALL create_col_vec_from_matrix(vectors%input_vec, matrix(1)%matrix, 1)
      CALL dbcsr_init(vectors%result_vec)
      CALL dbcsr_copy(vectors%result_vec, vectors%input_vec)
      CALL create_replicated_col_vec_from_matrix(vectors%rep_col_vec, matrix(1)%matrix, 1)
      CALL create_replicated_row_vec_from_matrix(vectors%rep_row_vec, matrix(1)%matrix, 1)

! Tells whether we have local data available on the processor (usually all in pcol 0 but even ther can be some without data)
      control => get_control(arnoldi_data)
      CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
      control%local_comp = ncol_local > 0 .AND. nrow_local > 0

      DO i_loop = 0, get_nrestart(arnoldi_data)

         IF (.NOT. control%iram .OR. i_loop == 0) THEN
! perform the standard arnoldi, if restarts are requested use the first (only makes sense if 1ev is requested)
            IF (ASSOCIATED(restart_vec)) CALL set_arnoldi_initial_vector(arnoldi_data, restart_vec)
            CALL arnoldi_init(matrix, vectors, arnoldi_data)
         ELSE
! perform an implicit restart
            CALL arnoldi_iram(arnoldi_data)
         END IF

! Generate the subspace
         CALL build_subspace(matrix, vectors, arnoldi_data)

! If we reached the maximum number of steps or the subspace converged we still need to get the eigenvalues
         CALL compute_evals(arnoldi_data)

! Select the evals according to user selection and keep them in arnoldi_data
         CALL select_evals(arnoldi_data)

! Prepare for a restart with the best eigenvector not needed in case of iram but who cares
         IF (.NOT. ASSOCIATED(restart_vec)) ALLOCATE (restart_vec)
         CALL get_selected_ritz_vector(arnoldi_data, 1, matrix(1)%matrix, restart_vec)

! Check whether we can already go home
         IF (control%converged) EXIT
      END DO

! Deallocated the work vectors
      CALL dbcsr_release(vectors%input_vec)
      CALL dbcsr_release(vectors%result_vec)
      CALL dbcsr_release(vectors%rep_col_vec)
      CALL dbcsr_release(vectors%rep_row_vec)
      CALL dbcsr_release(restart_vec)
      DEALLOCATE (restart_vec)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_arnoldi_normal_ev

! **************************************************************************************************
!> \brief The main routine for arnoldi method to compute the lowest ritz pair
!>        of a symmetric generalized eigenvalue problem.
!>        as input it takes a vector of matrices which for the GEV:
!>        M(1)*v=M(2)*v*lambda
!>        In other words, M(1) is the matrix and M(2) the metric
!>        This only works if the two matrices are symmetric in values
!>        (flag in dbcsr does not need to be set)
!> \param matrix a pointer array to dbcsr_matrices. Multiplication order is as
!>        described above
!> \param arnoldi_data On input data_type contains all information to start/restart
!>                     an arnoldi iteration
!>                     On output all data areas are filled to allow arbitrary post
!>                     processing of the created subspace
!>                     arnoldi_data has to be created with setup_arnoldi_data
! **************************************************************************************************
   SUBROUTINE dbcsr_arnoldi_generalized_ev(matrix, arnoldi_data)
      TYPE(dbcsr_obj_type_p), DIMENSION(:)               :: matrix
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_arnoldi_generalized_ev', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_loop, ncol_local, nrow_local
      TYPE(arnoldi_control), POINTER                     :: control
      TYPE(dbcsr_obj), TARGET                            :: A_rho_B
      TYPE(dbcsr_obj_type_p), ALLOCATABLE, DIMENSION(:)  :: matrix_arnoldi
      TYPE(m_x_v_vectors)                                :: vectors

      CALL timeset(routineN, handle)
      ALLOCATE (matrix_arnoldi(2))
      ! this matrix will contain +/- A-rho*B
      CALL dbcsr_init(A_rho_B)
      matrix_arnoldi(1)%matrix => A_rho_B
      matrix_arnoldi(2)%matrix => matrix(2)%matrix

!prepare the vector like matrives needed in the matrix vector products, they will be reused throughout the iterations
      CALL create_col_vec_from_matrix(vectors%input_vec, matrix(1)%matrix, 1)
      CALL dbcsr_init(vectors%result_vec)
      CALL dbcsr_copy(vectors%result_vec, vectors%input_vec)
      CALL create_replicated_col_vec_from_matrix(vectors%rep_col_vec, matrix(1)%matrix, 1)
      CALL create_replicated_row_vec_from_matrix(vectors%rep_row_vec, matrix(1)%matrix, 1)

! Tells whether we have local data available on the processor (usually all in pcol 0 but even ther can be some without data)
      control => get_control(arnoldi_data)
      CALL dbcsr_get_info(matrix=vectors%input_vec, nfullrows_local=nrow_local, nfullcols_local=ncol_local)
      control%local_comp = ncol_local > 0 .AND. nrow_local > 0

      DO i_loop = 0, get_nrestart(arnoldi_data)
         IF (i_loop == 0) THEN
! perform the standard arnoldi initialization with a random vector
            CALL gev_arnoldi_init(matrix, matrix_arnoldi, vectors, arnoldi_data)
         END IF

! Generate the subspace
         CALL gev_build_subspace(matrix_arnoldi, vectors, arnoldi_data)

! If we reached the maximum number of steps or the subspace converged we still need to get the eigenvalues
         CALL compute_evals(arnoldi_data)

! Select the evals according to user selection and keep them in arnoldi_data
         CALL select_evals(arnoldi_data)

! update the matrices and compute the convergence
         CALL gev_update_data(matrix, matrix_arnoldi, vectors, arnoldi_data)

! Check whether we can already go home
         IF (control%converged) EXIT
      END DO

! Deallocated the work vectors
      CALL dbcsr_release(vectors%input_vec)
      CALL dbcsr_release(vectors%result_vec)
      CALL dbcsr_release(vectors%rep_col_vec)
      CALL dbcsr_release(vectors%rep_row_vec)
      CALL dbcsr_release(A_rho_B)
      DEALLOCATE (matrix_arnoldi)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_arnoldi_generalized_ev

! **************************************************************************************************
!> \brief simple wrapper to estimate extremal eigenvalues with arnoldi, using the old lanczos interface
!>        this hides some of the power of the arnoldi routines (e.g. only min or max eval, generalized eval, ritz vectors, etc.),
!>        and does not allow for providing an initial guess of the ritz vector.
!> \param matrix_a input mat
!> \param max_ev estimated max eval
!> \param min_ev estimated min eval
!> \param converged ? Usually arnoldi is more accurate than claimed.
!> \param threshold target precision
!> \param max_iter max allowed iterations (will be rounded up)
! **************************************************************************************************
   SUBROUTINE dbcsr_arnoldi_extremal(matrix_a, max_ev, min_ev, converged, threshold, max_iter)
      TYPE(dbcsr_obj), INTENT(INOUT), TARGET             :: matrix_a
      REAL(KIND=dp), INTENT(OUT)                         :: max_ev, min_ev
      LOGICAL, INTENT(OUT)                               :: converged
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      INTEGER, INTENT(IN)                                :: max_iter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_arnoldi_extremal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, max_iter_internal, nrestarts
      TYPE(dbcsr_arnoldi_data)                           :: my_arnoldi
      TYPE(dbcsr_obj_type_p), DIMENSION(:), POINTER      :: arnoldi_matrices

      CALL timeset(routineN, handle)

      ! we go in chunks of max_iter_internal, and restart ater each of those.
      ! at low threshold smaller values of max_iter_internal make sense
      IF (.TRUE.) max_iter_internal = 16
      IF (threshold <= 1.0E-3_dp) max_iter_internal = 32
      IF (threshold <= 1.0E-4_dp) max_iter_internal = 64

      ! the max number of iter will be (nrestarts+1)*max_iter_internal
      nrestarts = max_iter/max_iter_internal

      ALLOCATE (arnoldi_matrices(1))
      arnoldi_matrices(1)%matrix => matrix_a
      CALL setup_arnoldi_data(my_arnoldi, arnoldi_matrices, max_iter=max_iter_internal, &
                              threshold=threshold, selection_crit=1, nval_request=2, nrestarts=nrestarts, &
                              generalized_ev=.FALSE., iram=.TRUE.)
      CALL dbcsr_arnoldi_ev(arnoldi_matrices, my_arnoldi)
      converged = arnoldi_is_converged(my_arnoldi)
      max_eV = REAL(get_selected_ritz_val(my_arnoldi, 2), dp)
      min_eV = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)
      CALL deallocate_arnoldi_data(my_arnoldi)
      DEALLOCATE (arnoldi_matrices)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_arnoldi_extremal

END MODULE dbcsr_arnoldi
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Tests for CP2K DBCSR operations
!> \author  Urban Borstnik
!> \date    2010-02-08
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2010-02-08
! **************************************************************************************************
MODULE dbcsr_tests
   USE dbcsr_arnoldi,                   ONLY: dbcsr_arnoldi_ev
   USE dbcsr_arnoldi_data_methods,      ONLY: dbcsr_arnoldi_data,&
                                              deallocate_arnoldi_data,&
                                              get_selected_ritz_val,&
                                              setup_arnoldi_data
   USE dbcsr_data_methods,              ONLY: dbcsr_scalar
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_mp,&
                                              dbcsr_distribution_new,&
                                              dbcsr_distribution_release
   USE dbcsr_dist_operations,           ONLY: dbcsr_dist_bin
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_wrong_args_error
   USE dbcsr_io,                        ONLY: dbcsr_binary_read,&
                                              dbcsr_binary_write
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_type, dbcsr_init, dbcsr_name, &
        dbcsr_nblkcols_total, dbcsr_nblkrows_total, dbcsr_release, dbcsr_row_block_sizes
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_active,&
                                              dbcsr_mp_group,&
                                              dbcsr_mp_init,&
                                              dbcsr_mp_new,&
                                              dbcsr_mp_npcols,&
                                              dbcsr_mp_nprows,&
                                              dbcsr_mp_release
   USE dbcsr_multiply_api,              ONLY: dbcsr_multiply
   USE dbcsr_operations,                ONLY: dbcsr_add,&
                                              dbcsr_copy,&
                                              dbcsr_frobenius_norm
   USE dbcsr_test_methods,              ONLY: dbcsr_make_random_block_sizes,&
                                              dbcsr_make_random_matrix
   USE dbcsr_transformations,           ONLY: dbcsr_redistribute
   USE dbcsr_types,                     ONLY: dbcsr_distribution_obj,&
                                              dbcsr_mp_obj,&
                                              dbcsr_obj,&
                                              dbcsr_obj_type_p,&
                                              dbcsr_scalar_type,&
                                              dbcsr_type_no_symmetry
   USE dbcsr_util,                      ONLY: dbcsr_checksum
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              real_8
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: &
        mp_cart_create, mp_cart_rank, mp_comm_free, mp_comm_null, mp_dims_create, mp_environ, &
        mp_max, mp_sum, mp_sync
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: dbcsr_run_tests, dbcsr_test_mm, dbcsr_test_arnoldi, dbcsr_test_binary_io

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_tests'

   INTEGER, PARAMETER          :: dbcsr_test_mm = 1
   INTEGER, PARAMETER          :: dbcsr_test_arnoldi = 2
   INTEGER, PARAMETER          :: dbcsr_test_binary_io = 3

CONTAINS

! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param[in] mp_group          MPI communicator
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] nproc             number of processors to test on
!> \param[in] matrix_sizes      size of matrices to test
!> \param[in] trs               transposes of the two matrices
!> \param[in] bs_m block sizes of the 3 dimensions
!> \param[in] bs_n block sizes of the 3 dimensions
!> \param[in] bs_k block sizes of the 3 dimensions
!> \param[in] sparsities        sparsities of matrices to create
!> \param[in] alpha alpha value to use in multiply
!> \param[in] beta beta value to use in multiply
!> \param[in] data_type         matrix data type
!> \param test_type ...
!> \param[in] n_loops           number of repetition for each multiplication
!> \param[in] eps               eps value for filtering
!> \param retain_sparsity ...
!> \param[in] always_checksum   checksum after each multiplication
! **************************************************************************************************
   SUBROUTINE dbcsr_run_tests(mp_group, io_unit, nproc, &
                              matrix_sizes, trs, &
                              bs_m, bs_n, bs_k, sparsities, alpha, beta, data_type, test_type, &
                              n_loops, eps, retain_sparsity, always_checksum)
      INTEGER, INTENT(IN)                                :: mp_group, io_unit
      INTEGER, DIMENSION(:), POINTER                     :: nproc
      INTEGER, DIMENSION(:), INTENT(in)                  :: matrix_sizes
      LOGICAL, DIMENSION(2), INTENT(in)                  :: trs
      INTEGER, DIMENSION(:), POINTER                     :: bs_m, bs_n, bs_k
      REAL(kind=dp), DIMENSION(3), INTENT(in)            :: sparsities
      REAL(kind=dp), INTENT(in)                          :: alpha, beta
      INTEGER, INTENT(IN)                                :: data_type, test_type, n_loops
      REAL(kind=dp), INTENT(in)                          :: eps
      LOGICAL, INTENT(in)                                :: retain_sparsity, always_checksum

      CHARACTER(len=*), PARAMETER :: fmt_desc = '(A,3(1X,I6),1X,A,2(1X,I5),1X,A,2(1X,L1))', &
         routineN = 'dbcsr_run_tests', routineP = moduleN//':'//routineN

      CHARACTER                                          :: t_a, t_b
      INTEGER                                            :: bmax, bmin, cart_group, error_handle, &
                                                            mynode, numnodes
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: group_sizes
      INTEGER, DIMENSION(2)                              :: npdims
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_a, col_dist_b, col_dist_c, &
                                                            row_dist_a, row_dist_b, row_dist_c, &
                                                            sizes_k, sizes_m, sizes_n
      LOGICAL                                            :: pgiven
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b, dist_c
      TYPE(dbcsr_mp_obj)                                 :: mp_env
      TYPE(dbcsr_obj), TARGET                            :: matrix_a, matrix_b, matrix_c

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      ! Create the row/column block sizes.
      IF (ASSOCIATED(bs_m)) THEN
         bmin = MINVAL(bs_m(2::2))
         bmax = MAXVAL(bs_m(2::2))
         CALL dbcsr_make_random_block_sizes(sizes_m, matrix_sizes(1), bs_m)
      ELSE
         CALL dbcsr_make_random_block_sizes(sizes_m, matrix_sizes(1), (/1, 13, 2, 5/))
         bmin = 5; bmax = 13
      ENDIF
      IF (ASSOCIATED(bs_n)) THEN
         bmin = MIN(bmin, MINVAL(bs_n(2::2)))
         bmax = MAX(bmax, MAXVAL(bs_n(2::2)))
         CALL dbcsr_make_random_block_sizes(sizes_n, matrix_sizes(2), bs_n)
      ELSE
         CALL dbcsr_make_random_block_sizes(sizes_n, matrix_sizes(2), (/1, 13, 2, 5/))
         bmin = MIN(bmin, 5); bmax = MAX(bmax, 13)
      ENDIF
      IF (ASSOCIATED(bs_k)) THEN
         bmin = MIN(bmin, MINVAL(bs_k(2::2)))
         bmax = MAX(bmax, MAXVAL(bs_k(2::2)))
         CALL dbcsr_make_random_block_sizes(sizes_k, matrix_sizes(3), bs_k)
      ELSE
         CALL dbcsr_make_random_block_sizes(sizes_k, matrix_sizes(3), (/1, 13, 2, 5/))
         bmin = MIN(bmin, 5); bmax = MAX(bmax, 13)
      ENDIF
      !
      ! Create dist

      ! Create the random matrices.
      CALL dbcsr_mp_make_env(mp_env, cart_group, mp_group)
      npdims(1) = dbcsr_mp_nprows(mp_env)
      npdims(2) = dbcsr_mp_npcols(mp_env)
      CALL dbcsr_dist_bin(row_dist_c, SIZE(sizes_m), npdims(1), &
                          sizes_m)
      CALL dbcsr_dist_bin(col_dist_c, SIZE(sizes_n), npdims(2), &
                          sizes_n)
      CALL dbcsr_distribution_new(dist_c, mp_env, row_dist_c, col_dist_c)
      CALL dbcsr_make_random_matrix(matrix_c, sizes_m, sizes_n, "Matrix C", &
                                    REAL(sparsities(3), real_8), &
                                    mp_group, data_type=data_type, dist=dist_c)
      CALL dbcsr_distribution_release(dist_c)
      IF (trs(1)) THEN
         CALL dbcsr_dist_bin(row_dist_a, SIZE(sizes_k), npdims(1), &
                             sizes_k)
         CALL dbcsr_dist_bin(col_dist_a, SIZE(sizes_m), npdims(2), &
                             sizes_m)
         CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_a, col_dist_a)
         CALL dbcsr_make_random_matrix(matrix_a, sizes_k, sizes_m, "Matrix A", &
                                       REAL(sparsities(1), real_8), &
                                       mp_group, data_type=data_type, dist=dist_a)
         DEALLOCATE (row_dist_a, col_dist_a)
      ELSE
         CALL dbcsr_dist_bin(col_dist_a, SIZE(sizes_k), npdims(2), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_c, col_dist_a)
         CALL dbcsr_make_random_matrix(matrix_a, sizes_m, sizes_k, "Matrix A", &
                                       REAL(sparsities(1), real_8), &
                                       mp_group, data_type=data_type, dist=dist_a)
         DEALLOCATE (col_dist_a)
      ENDIF
      CALL dbcsr_distribution_release(dist_a)
      IF (trs(2)) THEN
         CALL dbcsr_dist_bin(row_dist_b, SIZE(sizes_n), npdims(1), &
                             sizes_n)
         CALL dbcsr_dist_bin(col_dist_b, SIZE(sizes_k), npdims(2), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_b)
         CALL dbcsr_make_random_matrix(matrix_b, sizes_n, sizes_k, "Matrix B", &
                                       REAL(sparsities(2), real_8), &
                                       mp_group, data_type=data_type, dist=dist_b)
         DEALLOCATE (row_dist_b, col_dist_b)
      ELSE
         CALL dbcsr_dist_bin(row_dist_b, SIZE(sizes_k), npdims(1), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_c)
         CALL dbcsr_make_random_matrix(matrix_b, sizes_k, sizes_n, "Matrix B", &
                                       REAL(sparsities(2), real_8), &
                                       mp_group, data_type=data_type, dist=dist_b)
         DEALLOCATE (row_dist_b)
      ENDIF
      CALL dbcsr_mp_release(mp_env)
      CALL dbcsr_distribution_release(dist_b)
      DEALLOCATE (row_dist_c, col_dist_c)
      DEALLOCATE (sizes_m, sizes_n, sizes_k)
      ! Prepare test parameters
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, fmt_desc) "Testing with sizes", matrix_sizes(1:3), &
            "min/max block sizes", bmin, bmax, "transposed?", trs(1:2)
      ENDIF
      CALL mp_environ(numnodes, mynode, mp_group)
      pgiven = ASSOCIATED(nproc)
      IF (pgiven) pgiven = nproc(1) .NE. 0
      IF (pgiven) THEN
         ALLOCATE (group_sizes(SIZE(nproc), 2))
         group_sizes(:, 1) = nproc(:)
         group_sizes(:, 2) = 0
      ELSE
         !ALLOCATE (group_sizes (numnodes, 2))
         !DO test = numnodes, 1, -1
         !   group_sizes(1+numnodes-test, 1:2) = (/ test, 0 /)
         !ENDDO
         ALLOCATE (group_sizes(1, 2))
         group_sizes(1, 1:2) = (/numnodes, 0/)
      ENDIF
      t_a = 'N'; IF (trs(1)) t_a = 'T'
      t_b = 'N'; IF (trs(2)) t_b = 'T'

      SELECT CASE (test_type)
      CASE (dbcsr_test_mm)
         CALL test_multiplies_multiproc(group_sizes, &
                                        matrix_a, matrix_b, matrix_c, t_a, t_b, &
                                        dbcsr_scalar(REAL(alpha, real_8)), dbcsr_scalar(REAL(beta, real_8)), &
                                        n_loops=n_loops, eps=eps, &
                                        io_unit=io_unit, always_checksum=always_checksum, &
                                        retain_sparsity=retain_sparsity)
      CASE (dbcsr_test_arnoldi)
         CALL test_arnoldi(matrix_a, matrix_b, io_unit, n_loops)
      CASE (dbcsr_test_binary_io)
         CALL test_binary_io(matrix_a, io_unit)
      END SELECT

      CALL dbcsr_release(matrix_a)
      CALL dbcsr_release(matrix_b)
      CALL dbcsr_release(matrix_c)
      CALL mp_comm_free(cart_group)
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_run_tests

! **************************************************************************************************
!> \brief dumps and retrieves a dbcsr matrix, and checks a checksum
!> \param matrix_a matrix to be written
!> \param io_unit unit for status updates
! **************************************************************************************************
   SUBROUTINE test_binary_io(matrix_a, io_unit)
      TYPE(dbcsr_obj)                                    :: matrix_a
      INTEGER                                            :: io_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'test_binary_io', routineP = moduleN//':'//routineN

      CHARACTER(LEN=100)                                 :: file_name
      REAL(kind=real_8)                                  :: norm, post, pre
      TYPE(dbcsr_obj)                                    :: matrix_a_read

      file_name = "test_dbcsr_binary_io.dat"

      pre = dbcsr_checksum(matrix_a, pos=.TRUE.)

      CALL dbcsr_binary_write(matrix_a, file_name)

      ! needs a new matrix, reading into matrix_a does not work
      CALL dbcsr_init(matrix_a_read)
      CALL dbcsr_binary_read(file_name, distribution=dbcsr_distribution(matrix_a), &
                             matrix_new=matrix_a_read)

      post = dbcsr_checksum(matrix_a_read, pos=.TRUE.)
      CALL dbcsr_add(matrix_a_read, matrix_a, -1.0_dp, 1.0_dp)
      norm = dbcsr_frobenius_norm(matrix_a_read)
      IF (io_unit > 0) THEN
         WRITE (io_unit, *) "checksums", pre, post
         WRITE (io_unit, *) "difference norm", norm
      ENDIF

      IF (norm /= 0.0_dp) &
         CALL cp__b("dbcsr/dbcsr_tests.F",300,"bug in binary io")

      CALL dbcsr_release(matrix_a_read)

   END SUBROUTINE test_binary_io

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param io_unit ...
!> \param n_loops ...
! **************************************************************************************************
   SUBROUTINE test_arnoldi(matrix_a, matrix_b, io_unit, n_loops)
      TYPE(dbcsr_obj), TARGET                            :: matrix_a, matrix_b
      INTEGER                                            :: io_unit, n_loops

      CHARACTER(len=*), PARAMETER :: routineN = 'test_arnoldi', routineP = moduleN//':'//routineN

      COMPLEX(real_8)                                    :: max_ev, min_ev
      INTEGER                                            :: i
      REAL(kind=real_8)                                  :: t1, t2
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      TYPE(dbcsr_obj_type_p), ALLOCATABLE, DIMENSION(:)  :: matrix_p

      DO i = 1, n_loops
         t1 = m_walltime()
         ALLOCATE (matrix_p(2))
         matrix_p(1)%matrix => matrix_a
         matrix_p(2)%matrix => matrix_b
         CALL setup_arnoldi_data(arnoldi_data, matrix_p, 30, 1.0E-14_dp, 1, 2, 0, .FALSE., .FALSE.)
         CALL dbcsr_arnoldi_ev(matrix_p, arnoldi_data)
         max_eV = get_selected_ritz_val(arnoldi_data, 2)
         min_eV = get_selected_ritz_val(arnoldi_data, 1)
         CALL deallocate_arnoldi_data(arnoldi_data)
         DEALLOCATE (matrix_p)
         t2 = m_walltime()
         IF (io_unit .GT. 0) THEN
            WRITE (io_unit, '(A,I5,A,F12.3,A)') &
               " loop ", i, " used ", t2-t1, "s "
            CALL m_flush(io_unit)
         ENDIF
      END DO
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, '(A,1X,"(",ES12.3," , ",ES12.3,")")') 'Maximum computed EV is: ', max_ev
         WRITE (io_unit, '(A,1X,"(",ES12.3," , ",ES12.3,")")') 'Minimum computed EV is: ', min_ev
      END IF

   END SUBROUTINE test_arnoldi
! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param[in] group_sizes       array of (sub) communicator
!>                              sizes to test (2-D)
!> \param matrix_a matrices to multiply
!> \param matrix_b matrices to multiply
!> \param matrix_c matrices to multiply
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param beta ...
!> \param limits ...
!> \param retain_sparsity ...
!> \param n_loops ...
!> \param eps ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param always_checksum ...
! **************************************************************************************************
   SUBROUTINE test_multiplies_multiproc(group_sizes, &
                                        matrix_a, matrix_b, matrix_c, &
                                        transa, transb, alpha, beta, limits, retain_sparsity, &
                                        n_loops, eps, &
                                        io_unit, always_checksum)
      INTEGER, DIMENSION(:, :)                           :: group_sizes
      TYPE(dbcsr_obj), INTENT(in)                        :: matrix_a, matrix_b, matrix_c
      CHARACTER, INTENT(in)                              :: transa, transb
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(6), INTENT(in), OPTIONAL        :: limits
      LOGICAL, INTENT(in), OPTIONAL                      :: retain_sparsity
      INTEGER, INTENT(IN)                                :: n_loops
      REAL(kind=dp), INTENT(in)                          :: eps
      INTEGER, INTENT(IN)                                :: io_unit
      LOGICAL                                            :: always_checksum

      CHARACTER(len=*), PARAMETER :: routineN = 'test_multiplies_multiproc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cart_group, error_handle, group, &
                                                            loop_iter, mynode, numnodes, test
      INTEGER(kind=int_8)                                :: flop, flop_sum
      INTEGER, DIMENSION(2)                              :: npdims
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_a, col_dist_b, col_dist_c, &
                                                            row_dist_a, row_dist_b, row_dist_c
      LOGICAL                                            :: i_am_alive
      REAL(kind=real_8)                                  :: cs, cs_pos, flops_all, t1, t2
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b, dist_c
      TYPE(dbcsr_mp_obj)                                 :: mp_env
      TYPE(dbcsr_obj)                                    :: m_a, m_b, m_c, m_c_reserve

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      CALL dbcsr_assert(SIZE(group_sizes, 2), "EQ", 2, dbcsr_fatal_level, &
                        dbcsr_wrong_args_error, routineN, &
                        "second dimension of group_sizes must be 2", 404)
      CALL dbcsr_init(m_a)
      CALL dbcsr_init(m_b)
      CALL dbcsr_init(m_c)
      p_sizes: DO test = 1, SIZE(group_sizes, 1)
         t2 = 0.0_real_8
         flop_sum = 0
         npdims(1:2) = group_sizes(test, 1:2)
         numnodes = npdims(1)*npdims(2)
         group = dbcsr_mp_group(dbcsr_distribution_mp( &
                                dbcsr_distribution(matrix_c)))
         IF (numnodes .EQ. 0) THEN
            CALL dbcsr_mp_make_env(mp_env, cart_group, group, nprocs=MAXVAL(npdims))
         ELSE
            CALL dbcsr_mp_make_env(mp_env, cart_group, group, pgrid_dims=npdims)
         ENDIF
         CALL dbcsr_assert(numnodes, "GE", 0, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Cartesian sides must be greater or equal to 0", 422)
         i_am_alive = dbcsr_mp_active(mp_env)
         alive: IF (i_am_alive) THEN
            npdims(1) = dbcsr_mp_nprows(mp_env)
            npdims(2) = dbcsr_mp_npcols(mp_env)
            group = dbcsr_mp_group(mp_env)
            CALL mp_environ(numnodes, mynode, group)
            ! Row & column distributions
            CALL dbcsr_dist_bin(row_dist_a, &
                                dbcsr_nblkrows_total(matrix_a), npdims(1), &
                                dbcsr_row_block_sizes(matrix_a))
            CALL dbcsr_dist_bin(col_dist_a, &
                                dbcsr_nblkcols_total(matrix_a), npdims(2), &
                                dbcsr_col_block_sizes(matrix_a))
            CALL dbcsr_dist_bin(row_dist_b, &
                                dbcsr_nblkrows_total(matrix_b), npdims(1), &
                                dbcsr_row_block_sizes(matrix_b))
            CALL dbcsr_dist_bin(col_dist_b, &
                                dbcsr_nblkcols_total(matrix_b), npdims(2), &
                                dbcsr_col_block_sizes(matrix_b))
            CALL dbcsr_dist_bin(row_dist_c, &
                                dbcsr_nblkrows_total(matrix_c), npdims(1), &
                                dbcsr_row_block_sizes(matrix_c))
            CALL dbcsr_dist_bin(col_dist_c, &
                                dbcsr_nblkcols_total(matrix_c), npdims(2), &
                                dbcsr_col_block_sizes(matrix_c))
            CALL dbcsr_distribution_new(dist_a, &
                                        mp_env, row_dist_a, col_dist_a, reuse_arrays=.TRUE.)
            CALL dbcsr_distribution_new(dist_b, &
                                        mp_env, row_dist_b, col_dist_b, reuse_arrays=.TRUE.)
            CALL dbcsr_distribution_new(dist_c, &
                                        mp_env, row_dist_c, col_dist_c, reuse_arrays=.TRUE.)
            ! Redistribute the matrices
            ! A
            CALL dbcsr_create(m_a, "Test for "//TRIM(dbcsr_name(matrix_a)), &
                              dist_a, dbcsr_type_no_symmetry, &
                              dbcsr_row_block_sizes(matrix_a), &
                              dbcsr_col_block_sizes(matrix_a), &
                              data_type=dbcsr_get_data_type(matrix_a))
            CALL dbcsr_distribution_release(dist_a)
            CALL dbcsr_redistribute(matrix_a, m_a)
            ! B
            CALL dbcsr_create(m_b, "Test for "//TRIM(dbcsr_name(matrix_b)), &
                              dist_b, dbcsr_type_no_symmetry, &
                              dbcsr_row_block_sizes(matrix_b), &
                              dbcsr_col_block_sizes(matrix_b), &
                              data_type=dbcsr_get_data_type(matrix_b))
            CALL dbcsr_distribution_release(dist_b)
            CALL dbcsr_redistribute(matrix_b, m_b)
            ! C
            CALL dbcsr_create(m_c, "Test for "//TRIM(dbcsr_name(matrix_c)), &
                              dist_c, dbcsr_type_no_symmetry, &
                              dbcsr_row_block_sizes(matrix_c), &
                              dbcsr_col_block_sizes(matrix_c), &
                              data_type=dbcsr_get_data_type(matrix_c))
            CALL dbcsr_distribution_release(dist_c)
            CALL dbcsr_redistribute(matrix_c, m_c)
            CALL dbcsr_init(m_c_reserve)
            CALL dbcsr_copy(m_c_reserve, m_c)
            ! Perform multiply
            loops: DO loop_iter = 1, n_loops
               CALL dbcsr_release(m_c)
               CALL dbcsr_init(m_c)
               CALL dbcsr_copy(m_c, m_c_reserve)
               CALL mp_sync(group)
               t1 = -m_walltime()
               IF (PRESENT(limits)) THEN
                  IF (eps .LE. -0.0_dp) THEN
                     CALL dbcsr_multiply(transa, transb, alpha, &
                                         m_a, m_b, beta, m_c, &
                                         first_row=limits(1), &
                                         last_row=limits(2), &
                                         first_column=limits(3), &
                                         last_column=limits(4), &
                                         first_k=limits(5), &
                                         last_k=limits(6), &
                                         retain_sparsity=retain_sparsity, flop=flop)
                  ELSE
                     CALL dbcsr_multiply(transa, transb, alpha, &
                                         m_a, m_b, beta, m_c, &
                                         first_row=limits(1), &
                                         last_row=limits(2), &
                                         first_column=limits(3), &
                                         last_column=limits(4), &
                                         first_k=limits(5), &
                                         last_k=limits(6), &
                                         retain_sparsity=retain_sparsity, flop=flop, &
                                         filter_eps=eps)
                  ENDIF
               ELSE
                  IF (eps .LE. -0.0_dp) THEN
                     CALL dbcsr_multiply(transa, transb, alpha, &
                                         m_a, m_b, beta, m_c, &
                                         retain_sparsity=retain_sparsity, flop=flop)
                  ELSE
                     CALL dbcsr_multiply(transa, transb, alpha, &
                                         m_a, m_b, beta, m_c, &
                                         retain_sparsity=retain_sparsity, flop=flop, &
                                         filter_eps=eps)
                  ENDIF
               ENDIF
               t1 = t1+m_walltime()
               t2 = t2+t1
               flop_sum = flop_sum+flop
               !
               CALL mp_max(t1, group)
               CALL mp_sum(flop, group)
               t1 = MAX(t1, EPSILON(t1))
               flops_all = REAL(flop, KIND=real_8)/t1/numnodes/(1024*1024)
               IF (io_unit .GT. 0) THEN
                  WRITE (io_unit, '(A,I5,A,I5,A,F12.3,A,I9,A)') &
                     " loop ", loop_iter, " with ", numnodes, " MPI ranks: using ", t1, "s ", INT(flops_all), " Mflops/rank"
                  CALL m_flush(io_unit)
               ENDIF
               IF (loop_iter .EQ. n_loops .OR. always_checksum) THEN
                  cs = dbcsr_checksum(m_c)
                  cs_pos = dbcsr_checksum(m_c, pos=.TRUE.)
                  IF (io_unit > 0) THEN
                     WRITE (io_unit, *) "Final checksums", cs, cs_pos
                  ENDIF
               ENDIF
            ENDDO loops
            ! Release
            CALL dbcsr_mp_release(mp_env)
            CALL dbcsr_release(m_a)
            CALL dbcsr_release(m_b)
            CALL dbcsr_release(m_c)
            CALL dbcsr_release(m_c_reserve)
         ENDIF alive
         CALL mp_comm_free(cart_group)
      ENDDO p_sizes
      CALL timestop(error_handle)
   END SUBROUTINE test_multiplies_multiproc

! **************************************************************************************************
!> \brief Creates a sane mp_obj from the given MPI comm that is not a cartesian one (hack)
!> \param[out] mp_env      Message-passing environment object to create
!> \param[out] cart_group  the created cartesian group (to be freed by the user)
!> \param[in] mp_group     MPI group
!> \param[in] nprocs       (optional) Number of processes
!> \param[in] pgrid_dims   (optional) Dimensions of MPI group
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_make_env(mp_env, cart_group, mp_group, &
                                nprocs, pgrid_dims)
      TYPE(dbcsr_mp_obj), INTENT(OUT)                    :: mp_env
      INTEGER, INTENT(OUT)                               :: cart_group
      INTEGER, INTENT(IN)                                :: mp_group
      INTEGER, INTENT(IN), OPTIONAL                      :: nprocs
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: pgrid_dims

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mp_make_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, mynode, numnodes, pcol, &
                                                            prow
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: pgrid
      INTEGER, DIMENSION(2)                              :: coord, myploc, npdims
      LOGICAL                                            :: alive

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      CALL mp_environ(numnodes, mynode, mp_group)
      IF (PRESENT(nprocs)) THEN
         CALL dbcsr_assert(nprocs, "LE", numnodes, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Can not grow processes.", 588)
         numnodes = nprocs
      ENDIF
      !
      IF (PRESENT(pgrid_dims)) THEN
         npdims(:) = pgrid_dims
      ELSE
         npdims(:) = 0
         CALL mp_dims_create(numnodes, npdims)
      ENDIF
      CALL mp_cart_create(mp_group, 2, npdims, myploc, cart_group)
      alive = cart_group .NE. mp_comm_null
      IF (alive) THEN
         CALL mp_environ(numnodes, mynode, cart_group)
         ALLOCATE (pgrid(0:npdims(1)-1, 0:npdims(2)-1))
         DO prow = 0, npdims(1)-1
            DO pcol = 0, npdims(2)-1
               coord = (/prow, pcol/)
               CALL mp_cart_rank(cart_group, coord, pgrid(prow, pcol))
            ENDDO
         ENDDO
         CALL dbcsr_mp_new(mp_env, pgrid, cart_group, mynode, numnodes, &
                           myprow=myploc(1), mypcol=myploc(2))
      ELSE
         CALL dbcsr_mp_init(mp_env)
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_mp_make_env

END MODULE dbcsr_tests
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   This is the start of a dbcsr_api, all publically needed functions
!>          are exported here. The others remain private to the library.
!>          Currently, this is the CP2K used set.
!>          Ultimately, a reduced subset and well defined api will remain,
!>          possibly grouped in to standard and expert api.
!>          Currently, this is work in progress.
!> \author  Joost VandeVondele
! **************************************************************************************************
MODULE dbcsr_api
   USE array_types,                     ONLY: array_data
   USE dbcsr_arnoldi,                   ONLY: dbcsr_arnoldi_ev,&
                                              dbcsr_arnoldi_extremal
   USE dbcsr_arnoldi_data_methods,      ONLY: dbcsr_arnoldi_data,&
                                              deallocate_arnoldi_data,&
                                              get_selected_ritz_val,&
                                              get_selected_ritz_vector,&
                                              set_arnoldi_initial_vector,&
                                              setup_arnoldi_data
   USE dbcsr_block_access,              ONLY: dbcsr_get_block_p,&
                                              dbcsr_put_block,&
                                              dbcsr_reserve_all_blocks,&
                                              dbcsr_reserve_block2d,&
                                              dbcsr_reserve_blocks,&
                                              dbcsr_reserve_diag_blocks
   USE dbcsr_config,                    ONLY: &
        accdrv_avoid_after_busy, accdrv_binning_binsize, accdrv_binning_nbins, &
        accdrv_do_inhomogenous, accdrv_min_flop_process, accdrv_min_flop_sort, &
        accdrv_posterior_buffers, accdrv_posterior_streams, accdrv_priority_buffers, &
        accdrv_priority_streams, dbcsr_get_conf_avg_elements_images, &
        dbcsr_get_conf_comm_thread_load, dbcsr_get_conf_max_ele_block, dbcsr_get_conf_mm_driver, &
        dbcsr_get_conf_mm_stacksize, dbcsr_get_conf_mpi_mem, dbcsr_get_conf_nstacks, &
        dbcsr_get_conf_randmat_seed, dbcsr_get_conf_use_comm_thread, dbcsr_get_conf_use_mpi_exp, &
        dbcsr_get_conf_use_mpi_filtering, dbcsr_set_conf_avg_elements_images, &
        dbcsr_set_conf_comm_thread_load, dbcsr_set_conf_max_ele_block, dbcsr_set_conf_mm_driver, &
        dbcsr_set_conf_mm_stacksize, dbcsr_set_conf_nstacks, dbcsr_set_conf_randmat_seed, &
        dbcsr_set_conf_use_comm_thread, dbcsr_set_conf_use_mpi_exp, &
        dbcsr_set_conf_use_mpi_filtering, dbcsr_set_default_config, has_acc, has_mpi, &
        mm_driver_auto, mm_driver_blas, mm_driver_matmul, mm_driver_smm, mm_driver_xsmm, &
        mm_name_auto, mm_name_blas, mm_name_matmul, mm_name_smm, mm_name_xsmm, multrec_limit
   USE dbcsr_csr_conversions,           ONLY: &
        convert_csr_to_dbcsr, convert_dbcsr_to_csr, csr_create_from_dbcsr, csr_create_new, &
        csr_create_template, csr_dbcsr_blkrow_dist, csr_destroy, csr_eqrow_ceil_dist, &
        csr_eqrow_floor_dist, csr_p_type, csr_print_sparsity, csr_type
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_init, dbcsr_data_new, dbcsr_data_release, &
        dbcsr_get_data_p, dbcsr_scalar, dbcsr_scalar_fill_all, dbcsr_scalar_get_type, &
        dbcsr_scalar_get_value, dbcsr_scalar_set_type, dbcsr_scalar_zero
   USE dbcsr_dist_methods,              ONLY: &
        dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_get_num_images_1d, dbcsr_distribution_has_col_clusters, &
        dbcsr_distribution_has_row_clusters, dbcsr_distribution_has_threads, &
        dbcsr_distribution_init, dbcsr_distribution_local_cols, dbcsr_distribution_local_rows, &
        dbcsr_distribution_mp, dbcsr_distribution_ncols, dbcsr_distribution_new, &
        dbcsr_distribution_nlocal_cols, dbcsr_distribution_nlocal_rows, dbcsr_distribution_nrows, &
        dbcsr_distribution_release, dbcsr_distribution_row_clusters, dbcsr_distribution_row_dist
   USE dbcsr_dist_operations,           ONLY: create_bl_distribution,&
                                              dbcsr_create_dist_block_cyclic,&
                                              dbcsr_create_dist_r_unrot,&
                                              dbcsr_get_stored_coordinates
   USE dbcsr_io,                        ONLY: dbcsr_binary_read,&
                                              dbcsr_binary_write,&
                                              dbcsr_print,&
                                              dbcsr_print_block_sum
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_lib,                       ONLY: dbcsr_clear_mempools,&
                                              dbcsr_finalize_lib,&
                                              dbcsr_init_lib
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_size, dbcsr_get_data_type, &
        dbcsr_get_matrix_type, dbcsr_get_num_blocks, dbcsr_has_symmetry, dbcsr_init, &
        dbcsr_max_col_size, dbcsr_max_row_size, dbcsr_name, dbcsr_nblkcols_total, &
        dbcsr_nblkrows_total, dbcsr_nfullcols_total, dbcsr_nfullrows_total, dbcsr_release, &
        dbcsr_row_block_sizes, dbcsr_setname, dbcsr_valid_index, dbcsr_wm_use_mutable
   USE dbcsr_mp_methods,                ONLY: &
        dbcsr_mp_get_process, dbcsr_mp_group, dbcsr_mp_hold, dbcsr_mp_mynode, dbcsr_mp_new, &
        dbcsr_mp_npcols, dbcsr_mp_nprows, dbcsr_mp_numnodes, dbcsr_mp_release
   USE dbcsr_multiply_api,              ONLY: dbcsr_multiply
   USE dbcsr_operations,                ONLY: &
        dbcsr_add, dbcsr_add_on_diag, dbcsr_copy, dbcsr_copy_into_existing, dbcsr_filter, &
        dbcsr_frobenius_norm, dbcsr_function_of_elements, dbcsr_gershgorin_norm, &
        dbcsr_get_block_diag, dbcsr_get_diag, dbcsr_get_info, dbcsr_get_occupation, &
        dbcsr_hadamard_product, dbcsr_init_random, dbcsr_maxabs, dbcsr_norm, dbcsr_scale, &
        dbcsr_scale_by_vector, dbcsr_set, dbcsr_set_diag, dbcsr_sum_replicated, dbcsr_trace, &
        dbcsr_triu
   USE dbcsr_tests,                     ONLY: dbcsr_run_tests,&
                                              dbcsr_test_arnoldi,&
                                              dbcsr_test_binary_io,&
                                              dbcsr_test_mm
   USE dbcsr_toollib,                   ONLY: swap
   USE dbcsr_transformations,           ONLY: dbcsr_complete_redistribute,&
                                              dbcsr_desymmetrize_deep,&
                                              dbcsr_distribute,&
                                              dbcsr_new_transposed,&
                                              dbcsr_replicate_all
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_func_artanh, dbcsr_func_dtanh, &
        dbcsr_func_inverse, dbcsr_func_tanh, dbcsr_iterator, dbcsr_mp_obj, dbcsr_no_transpose, &
        dbcsr_norm_column, dbcsr_norm_frobenius, dbcsr_norm_maxabsnorm, dbcsr_obj, &
        dbcsr_obj_type_p, dbcsr_scalar_type, dbcsr_type_antisymmetric, dbcsr_type_complex_4, &
        dbcsr_type_complex_8, dbcsr_type_complex_default, dbcsr_type_no_symmetry, &
        dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_real_default, dbcsr_type_symmetric
   USE dbcsr_util,                      ONLY: convert_offsets_to_sizes,&
                                              convert_sizes_to_offsets,&
                                              dbcsr_checksum,&
                                              dbcsr_verify_matrix
   USE dbcsr_work_operations,           ONLY: add_work_coordinate,&
                                              dbcsr_create,&
                                              dbcsr_finalize,&
                                              dbcsr_work_create
   USE min_heap,                        ONLY: heap_fill,&
                                              heap_get_first,&
                                              heap_new,&
                                              heap_release,&
                                              heap_reset_first,&
                                              heap_t

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: heap_fill, &
             heap_get_first, &
             heap_new, &
             heap_release, &
             heap_reset_first, &
             heap_t, &
             array_data, &
             add_work_coordinate, &
             convert_offsets_to_sizes, &
             convert_sizes_to_offsets, &
             dbcsr_run_tests, &
             dbcsr_test_mm, &
             dbcsr_test_arnoldi, &
             dbcsr_test_binary_io, &
             create_bl_distribution, &
             accdrv_avoid_after_busy, &
             accdrv_binning_binsize, &
             accdrv_binning_nbins, &
             accdrv_do_inhomogenous, &
             accdrv_min_flop_process, &
             accdrv_min_flop_sort, &
             accdrv_posterior_buffers, &
             accdrv_posterior_streams, &
             accdrv_priority_buffers, &
             accdrv_priority_streams, &
             dbcsr_add, &
             dbcsr_add_on_diag, &
             dbcsr_arnoldi_data, &
             setup_arnoldi_data, &
             set_arnoldi_initial_vector, &
             deallocate_arnoldi_data, &
             dbcsr_arnoldi_ev, &
             dbcsr_arnoldi_extremal, &
             get_selected_ritz_val, &
             get_selected_ritz_vector, &
             dbcsr_binary_read, &
             dbcsr_binary_write, &
             dbcsr_triu, &
             dbcsr_checksum, &
             dbcsr_clear_mempools, &
             dbcsr_complete_redistribute, &
             dbcsr_copy, &
             dbcsr_copy_into_existing, &
             dbcsr_create, &
             dbcsr_create_dist_block_cyclic, &
             dbcsr_create_dist_r_unrot, &
             dbcsr_data_clear_pointer, &
             dbcsr_data_init, &
             dbcsr_data_new, &
             dbcsr_data_obj, &
             dbcsr_data_release, &
             dbcsr_desymmetrize_deep, &
             dbcsr_distribute, &
             dbcsr_distribution, &
             dbcsr_distribution_col_dist, &
             dbcsr_distribution_has_threads, &
             dbcsr_distribution_init, &
             dbcsr_distribution_local_cols, &
             dbcsr_distribution_local_rows, &
             dbcsr_distribution_mp, &
             dbcsr_distribution_ncols, &
             dbcsr_distribution_new, &
             dbcsr_distribution_nlocal_cols, &
             dbcsr_distribution_nlocal_rows, &
             dbcsr_distribution_nrows, &
             dbcsr_distribution_obj, &
             dbcsr_distribution_release, &
             dbcsr_distribution_row_dist, &
             dbcsr_distribution_row_clusters, &
             dbcsr_distribution_col_clusters, &
             dbcsr_distribution_has_row_clusters, &
             dbcsr_distribution_has_col_clusters, &
             dbcsr_filter, &
             dbcsr_finalize, &
             dbcsr_finalize_lib, &
             dbcsr_frobenius_norm, &
             dbcsr_func_artanh, &
             dbcsr_func_dtanh, &
             dbcsr_func_inverse, &
             dbcsr_func_tanh, &
             dbcsr_function_of_elements, &
             dbcsr_gershgorin_norm, &
             dbcsr_get_block_diag, &
             dbcsr_get_block_p, &
             dbcsr_get_conf_comm_thread_load, &
             dbcsr_get_conf_max_ele_block, &
             dbcsr_get_conf_mm_driver, &
             dbcsr_get_conf_avg_elements_images, &
             dbcsr_get_conf_mm_stacksize, &
             dbcsr_get_conf_mpi_mem, &
             dbcsr_get_conf_nstacks, &
             dbcsr_get_conf_use_mpi_exp, &
             dbcsr_get_conf_use_mpi_filtering, &
             dbcsr_get_conf_use_comm_thread, &
             dbcsr_get_data_p, &
             dbcsr_get_data_size, &
             dbcsr_get_data_type, &
             dbcsr_get_diag, &
             dbcsr_get_info, &
             dbcsr_get_matrix_type, &
             dbcsr_get_num_blocks, &
             dbcsr_get_occupation, &
             dbcsr_get_stored_coordinates, &
             dbcsr_hadamard_product, &
             dbcsr_init, &
             dbcsr_init_lib, &
             dbcsr_init_random, &
             dbcsr_iterator, &
             dbcsr_iterator_blocks_left, &
             dbcsr_iterator_next_block, &
             dbcsr_iterator_start, &
             dbcsr_iterator_stop, &
             dbcsr_maxabs, &
             dbcsr_max_col_size, &
             dbcsr_max_row_size, &
             dbcsr_mp_group, &
             dbcsr_mp_get_process, &
             dbcsr_mp_hold, &
             dbcsr_mp_mynode, &
             dbcsr_mp_new, &
             dbcsr_mp_npcols, &
             dbcsr_mp_nprows, &
             dbcsr_mp_numnodes, &
             dbcsr_mp_obj, &
             dbcsr_mp_release, &
             dbcsr_multiply, &
             dbcsr_name, &
             dbcsr_setname, &
             dbcsr_nblkcols_total, &
             dbcsr_nblkrows_total, &
             dbcsr_new_transposed, &
             dbcsr_nfullcols_total, &
             dbcsr_nfullrows_total, &
             dbcsr_norm, &
             dbcsr_norm_column, &
             dbcsr_norm_frobenius, &
             dbcsr_norm_maxabsnorm, &
             dbcsr_no_transpose, &
             dbcsr_obj, &
             dbcsr_obj_type_p, &
             dbcsr_print, &
             dbcsr_print_block_sum, &
             dbcsr_put_block, &
             dbcsr_release, &
             dbcsr_replicate_all, &
             dbcsr_reserve_all_blocks, &
             dbcsr_reserve_block2d, &
             dbcsr_reserve_blocks, &
             dbcsr_reserve_diag_blocks, &
             dbcsr_scalar, &
             dbcsr_scalar_fill_all, &
             dbcsr_scalar_get_type, &
             dbcsr_scalar_get_value, &
             dbcsr_scalar_set_type, &
             dbcsr_scalar_type, &
             dbcsr_scalar_zero, &
             dbcsr_scale, &
             dbcsr_scale_by_vector, &
             dbcsr_set, &
             dbcsr_set_conf_comm_thread_load, &
             dbcsr_set_conf_max_ele_block, &
             dbcsr_set_conf_mm_driver, &
             dbcsr_set_conf_mm_stacksize, &
             dbcsr_set_conf_avg_elements_images, &
             dbcsr_set_conf_nstacks, &
             dbcsr_set_conf_use_mpi_exp, &
             dbcsr_set_conf_use_mpi_filtering, &
             dbcsr_set_conf_use_comm_thread, &
             dbcsr_set_diag, &
             dbcsr_sum_replicated, &
             dbcsr_trace, &
             dbcsr_type_antisymmetric, &
             dbcsr_type_complex_4, &
             dbcsr_type_complex_8, &
             dbcsr_type_complex_default, &
             dbcsr_type_no_symmetry, &
             dbcsr_type_real_4, &
             dbcsr_type_real_8, &
             dbcsr_type_real_default, &
             dbcsr_type_symmetric, &
             dbcsr_valid_index, &
             dbcsr_verify_matrix, &
             dbcsr_wm_use_mutable, &
             dbcsr_work_create, &
             has_mpi, &
             mm_driver_auto, &
             mm_driver_blas, &
             mm_driver_matmul, &
             mm_driver_smm, &
             mm_driver_xsmm, &
             mm_name_auto, &
             mm_name_blas, &
             mm_name_matmul, &
             mm_name_smm, &
             mm_name_xsmm, &
             has_acc, &
             multrec_limit, &
             swap, &
             dbcsr_set_default_config, &
             dbcsr_row_block_sizes, &
             dbcsr_col_block_sizes, &
             dbcsr_has_symmetry, &
             convert_csr_to_dbcsr, &
             convert_dbcsr_to_csr, &
             csr_create_from_dbcsr, &
             csr_destroy, &
             csr_type, &
             csr_p_type, &
             csr_create_new, &
             csr_create_template, &
             csr_dbcsr_blkrow_dist, &
             csr_eqrow_ceil_dist, &
             csr_eqrow_floor_dist, &
             csr_print_sparsity, &
             dbcsr_distribution_get_num_images_1d, &
             dbcsr_set_conf_randmat_seed, dbcsr_get_conf_randmat_seed
END MODULE dbcsr_api
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief collects all references to literature in CP2K
!>      as new algorithms / method are included from literature sources
!>      we add a citation, to give credit to the authors
!>      (often ourselves, but this need not to be the case), and as a form
!>      of documentation.
!>
!>      reference_manager.F provides an easy way to cite these references from the code
!>      (e.g. whenever the relevant part of the code is executed)
!>      and to add citations to the input
!>      at the end of the run a bibliography is printed that can be used e.g. in papers
!> \note
!>      references need to be input using the ISI citation format, because it is
!>      uniform, easy to parse, and can be exported for example from web of science.
!>      (mark the reference, and click 'export to reference software', open the resulting file)
!>      Furthermore, it can be easily converted to and from using the bibutils tools
!>      a collection of easy to use conversion programs that can be found at
!>      http://www.scripps.edu/~cdputnam/software/bibutils/ by Chris Putnam
!> \par History
!>      08.2007 [Joost VandeVondele] Created
!>      01.2014 [Lianheng Tong] Added Kantorovich2008, Kantorovich2008a
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE bibliography

   USE reference_manager,               ONLY: add_reference,&
                                              cite_reference
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: cite_reference ! for convenience
   PUBLIC :: add_all_references ! f77_interface only

   ! all reference keys
   INTEGER, PUBLIC, SAVE ::  VandeVondele2005a, VandeVondele2003, &
                            Laino2005, Laino2006, Goedecker1996, Hartwigsen1998, Krack2005, &
                            Lippert1997, Lippert1999, Krack2000, Krack2002, Iannuzzi2005, &
                            Iannuzzi2006, Iannuzzi2007, Toukmaji1996, Martyna1999, &
                            VandeVondele2005b, Perdew1981, Avezac2005, Porezag1995, &
                            Seifert1996, Elstner1998, Zhechkov2005, Henkelman1999, &
                            Frigo2005, Nose1984a, Nose1984b, Brelaz1979, Bussi2007, &
                            Genovese2006, Genovese2007, Evans1983, Minary2003, Byrd1995, &
                            VandeVondele2007, Ortiz1994, Becke1988, Perdew1996, Zhang1998, &
                            Perdew2008, Lee1988, Heyd2006, Vydrov2006, Heyd2003, Heyd2004, &
                            Vosko1980, Aguado2003, Essmann1995, Ewald1921, Darden1993, &
                            Siepmann1995, Tersoff1988, Tosi1964a, Tosi1964b, Yamada2000, &
                            Dudarev1997, Dudarev1998, Dewar1977, Dewar1985, Rocha2006, &
                            Stewart1989, Thiel1992, Repasky2002, Stewart2007, Weber2008, &
                            Hunt2003, Guidon2008, Elber1987, Jonsson1998, Jonsson2000_1, &
                            Jonsson2000_2, Wales2004, Stewart1982, E2002, Laino2008, &
                            Blochl1995, Tao2003, VandeVondele2006, Hu2007, Grimme2006, &
                            Branduardi2007, Schenter2008, BeckeRoussel1989, Proynov2007, &
                            VandenCic2006, Becke1997, Foiles1986, Ricci2003, Kolafa2004, &
                            Kuhne2007, Kunert2003, Ceriotti2009, ceriotti2009b, Guidon2009, &
                            BarducBus2008, Grimme2010, Guidon2010, Marques2012, Jones2011, &
                            Bernstein2009, Bernstein2012, Dick1958, Mitchell1993, &
                            Devynck2012, VandeVondele2012, Niklasson2003, Shao2003, &
                            VandeVondele2002, Dion2004, Romanperez2009, DelBen2012, &
                            Sabatini2013, DelBen2013, Kikuchi2009, Putrino2000, &
                            Putrino2002, Sebastiani2001, Weber2009, Tran2013, Golze2013, &
                            Golze2015, Tuckerman1992, Zhao1994, Tozer1996, Goedecker2004, &
                            Khaliullin2013, Hutter2014, Bengtsson1999, Kantorovich2008, &
                            Kantorovich2008a, Wellendorff2012, Niklasson2014, Borstnik2014, &
                            Rayson2009, Grimme2011, Fattebert2002, Andreussi2012, &
                            Khaliullin2007, Khaliullin2008, Merlot2014, Lin2009, Lin2013, &
                            QUIP_ref, DelBen2015, Souza2002, Umari2002, Stengel2009, &
                            Luber2014, Berghold2011, DelBen2015b, Campana2009, &
                            Schiffmann2015, Bruck2014, Rappe1992, Ceriotti2012, &
                            Ceriotti2010, Walewski2014, Monkhorst1976, MacDonald1978, &
                            Gilbert2008, Schonherr2014, Ceriotti2014, BaniHashemian2016, &
                            Kapil2016, Heinzmann1976, Ehrhardt1985

CONTAINS

! **************************************************************************************************
!> \brief adds references that can later be cited / printed using the key
!> \par History
!>      08.2007 created [Joost VandeVondele]
!> \note
!>     - note that the ISI record is fixed format (line length limited and the
!>       first 3 characters can indicate record type)
!>     - please add DOI whenever available, this will result in a clickable
!>       link in the input reference manual.
! **************************************************************************************************
   SUBROUTINE add_all_references()

      CALL add_reference(key=Ceriotti2012, ISI_record=s2a( &
                         "AU Ceriotti, M", &
                         "   Manolopoulos, D", &
                         "TI Efficient First-Principles Calculation of the Quantum "// &
                         "Kinetic Energy and Momentum Distribution of Nuclei", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "PY 2012", &
                         "VL 109", &
                         "IS 10", &
                         "BP 100604", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.109.100604")

      CALL add_reference(key=Ceriotti2010, ISI_record=s2a( &
                         "AU Ceriotti, M", &
                         "   Parrinello, M", &
                         "   Markland, T", &
                         "   Manolopoulos, D", &
                         "TI Efficient stochastic thermostatting of path integral molecular dynamics", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "PY 2010", &
                         "VL 133", &
                         "IS 12", &
                         "BP 124104", &
                         "ER"), &
                         DOI="10.1063/1.3489925")

      CALL add_reference(key=Wellendorff2012, ISI_record=s2a( &
                         "AU Wellendorff, J", &
                         "   Lundgaard, K", &
                         "   Mogelhoj, A", &
                         "   Petzold, V", &
                         "   Landis, D", &
                         "   Norskov, J", &
                         "   Bligaard, T", &
                         "   Jacobsen, K", &
                         "TI Density functionals for surface science: Exchange-correlation "// &
                         "model development with Bayesian error estimation", &
                         "SO PHYSICAL REVIEW B", &
                         "PY 2012", &
                         "VL 85", &
                         "IS 23", &
                         "BP 235149", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.85.235149")

      CALL add_reference(key=Brelaz1979, ISI_record=s2a( &
                         "PT J", &
                         "AU Brelaz, D", &
                         "TI New methods to color the vertices of a graph", &
                         "SO COMMUNICATIONS OF THE ACM", &
                         "SN 0001-0782", &
                         "PY 1979", &
                         "VL 22", &
                         "IS 4", &
                         "BP 251", &
                         "EP 256", &
                         "UT WOS:A1979GR73200004", &
                         "ER"), &
                         DOI="10.1145/359094.359101")

      CALL add_reference(key=Bengtsson1999, ISI_record=s2a( &
                         "AU BENGTSSON, L", &
                         "TI DIPOLE CORRECTION FOR SURFACE SUPERCELL CALCULATIONE", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PY 1999", &
                         "VL 59", &
                         "IS 19", &
                         "BP 12301", &
                         "EP 12304", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.59.12301")

      CALL add_reference(key=Foiles1986, ISI_record=s2a( &
                         "AU FOILES, SM", &
                         "   BASKES, MI", &
                         "   DAW, MS", &
                         "TI EMBEDDED-ATOM-METHOD FUNCTIONS FOR THE FCC METALS CU, AG, AU, NI, PD,", &
                         "   PT, AND THEIR ALLOYS", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD JUN 15", &
                         "PY 1986", &
                         "VL 33", &
                         "IS 12", &
                         "PN Part 1", &
                         "BP 7983", &
                         "EP 7991", &
                         "UT ISI:A1986C899400011", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.33.7983")

      CALL add_reference(key=QUIP_ref, ISI_record=s2a( &
                         "AU QUIP", &
                         "TI libAtoms/QUIP libraries from http://www.libatoms.org, please cite", &
                         "   web site and references for appropriate potential invoked", &
                         "SO web site", &
                         "PY 2014", &
                         "ER"), &
                         DOI="")

      CALL add_reference(key=VandenCic2006, ISI_record=s2a( &
                         "AU Vanden-Eijnden, E", &
                         "   Ciccotti, G", &
                         "AF Vanden-Eijnden, Eric", &
                         "   Ciccotti, Giovanni", &
                         "TI Second-order integrators for Langevin equations with holonomic", &
                         "   constraints", &
                         "SO CHEMICAL PHYSICS LETTERS", &
                         "SN 0009-2614", &
                         "PD SEP 29", &
                         "PY 2006", &
                         "VL 429", &
                         "IS 1-3", &
                         "BP 310", &
                         "EP 316", &
                         "UT ISI:000241018800060", &
                         "ER"), &
                         DOI="10.1016/j.cplett.2006.07.086")

      CALL add_reference(key=Hu2007, ISI_record=s2a( &
                         "AU Hu, H", &
                         "   Lu, ZY", &
                         "   Elstner, M", &
                         "   Hermans, J", &
                         "   Yang, WT", &
                         "TI Simulating water with the self-consistent-charge density functional", &
                         "   tight binding method: From molecular clusters to the liquid state", &
                         "SO JOURNAL OF PHYSICAL CHEMISTRY A", &
                         "CT 232nd National Meeting of the American-Chemical-Society", &
                         "CY SEP 10-14, 2006", &
                         "CL San Francisco, CA", &
                         "SN 1089-5639", &
                         "PD JUL 5", &
                         "PY 2007", &
                         "VL 111", &
                         "IS 26", &
                         "BP 5685", &
                         "EP 5691", &
                         "UT ISI:000247573600013", &
                         "ER"), &
                         DOI="10.1021/jp070308d")

      CALL add_reference(key=Zhao1994, ISI_record=s2a( &
                         "AU ZHAO, QS", &
                         "   MORRISON, RC", &
                         "   PARR, RG", &
                         "TI From electron densities to Kohn-Sham kinetic energies, orbital energies,", &
                         "   exchange-correlation potentials, and exchange-correlation energies", &
                         "SO PHYSICAL REVIEW A", &
                         "SN 1050-2947", &
                         "PD SEP 1", &
                         "PY 1994", &
                         "VL 50", &
                         "IS 3", &
                         "BP 2138", &
                         "EP 2142", &
                         "UT WOS:A1994PG80900023", &
                         "ER"), &
                         DOI="10.1103/PhysRevA.50.2138")

      CALL add_reference(key=Tozer1996, ISI_record=s2a( &
                         "AU TOZER, DJ", &
                         "   INGAMELLS, VE", &
                         "   HANDY, NC", &
                         "TI EXCHANGE-CORRELATION POTENTIALS", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD NOV 22", &
                         "PY 1996", &
                         "VL 105", &
                         "IS 20", &
                         "BP 9200", &
                         "EP 9213", &
                         "UT WOS:A1996VV04900024", &
                         "ER"), &
                         DOI="10.1063/1.472753")

      CALL add_reference(key=Blochl1995, ISI_record=s2a( &
                         "AU BLOCHL, PE", &
                         "TI ELECTROSTATIC DECOUPLING OF PERIODIC IMAGES OF PLANE-WAVE-EXPANDED", &
                         "   DENSITIES AND DERIVED ATOMIC POINT CHARGES", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD NOV 1", &
                         "PY 1995", &
                         "VL 103", &
                         "IS 17", &
                         "BP 7422", &
                         "EP 7428", &
                         "UT ISI:A1995TB63300019", &
                         "ER"), &
                         DOI="10.1063/1.470314")

      CALL add_reference(key=Laino2008, ISI_record=s2a( &
                         "AU Laino, T", &
                         "   Hutter, J", &
                         'TI Notes on "Ewald summation of electrostatic multipole interactions up to', &
                         '   quadrupolar level" [J. Chem. Phys. 119, 7471 (2003)]', &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD AUG 21", &
                         "PY 2008", &
                         "VL 129", &
                         "IS 7", &
                         "AR 074102", &
                         "UT ISI:000258643300004", &
                         "ER"), &
                         DOI="10.1063/1.2970887")

      CALL add_reference(key=E2002, ISI_record=s2a( &
                         "AU E, WN", &
                         "   Ren, WQ", &
                         "   Vanden-Eijnden, E", &
                         "TI String method for the study of rare events", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD AUG 1", &
                         "PY 2002", &
                         "VL 66", &
                         "IS 5", &
                         "AR 052301", &
                         "UT ISI:000177873000010", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.66.052301")

      CALL add_reference(key=Wales2004, ISI_record=s2a( &
                         "AU Trygubenko, SA", &
                         "   Wales, DJ", &
                         "TI A doubly nudged elastic band method for finding transition states", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD FEB 1", &
                         "PY 2004", &
                         "VL 120", &
                         "IS 5", &
                         "BP 2082", &
                         "EP 2094", &
                         "DI 10.1063/1.1636455", &
                         "UT ISI:000188498400003", &
                         "ER"), &
                         DOI="10.1063/1.1636455")

      CALL add_reference(key=Jonsson2000_2, ISI_record=s2a( &
                         "AU Henkelman, G", &
                         "   Uberuaga, BP", &
                         "   Jonsson, H", &
                         "TI A climbing image nudged elastic band method for finding saddle points", &
                         "   and minimum energy paths", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD DEC 8", &
                         "PY 2000", &
                         "VL 113", &
                         "IS 22", &
                         "BP 9901", &
                         "EP 9904", &
                         "UT ISI:000165584900005", &
                         "ER"), &
                         DOI="10.1063/1.1329672")

      CALL add_reference(key=Jonsson2000_1, ISI_record=s2a( &
                         "AU Henkelman, G", &
                         "   Jonsson, H", &
                         "TI Improved tangent estimate in the nudged elastic band method for finding", &
                         "   minimum energy paths and saddle points", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD DEC 8", &
                         "PY 2000", &
                         "VL 113", &
                         "IS 22", &
                         "BP 9978", &
                         "EP 9985", &
                         "UT ISI:000165584900014", &
                         "ER"), &
                         DOI="10.1063/1.1323224")

      CALL add_reference(key=Jonsson1998, ISI_record=s2a( &
                         "AU JONSSON, H", &
                         "   MILLS, G", &
                         "   JACOBSEN, K W", &
                         "TI Nudged elastic band method for finding minimum energy paths ", &
                         "   of transitions", &
                         "SO Classical and Quantum Dynamics in Condensed Phase Simulations", &
                         "PY 1998", &
                         "BP 385", &
                         "EP 404", &
                         "ER"), &
                         DOI="")

      CALL add_reference(key=Elber1987, ISI_record=s2a( &
                         "AU ELBER, R", &
                         "   KARPLUS, M", &
                         "TI A METHOD FOR DETERMINING REACTION PATHS IN LARGE MOLECULES -", &
                         "   APPLICATION TO MYOGLOBIN", &
                         "SO CHEMICAL PHYSICS LETTERS", &
                         "SN 0009-2614", &
                         "PD SEP 4", &
                         "PY 1987", &
                         "VL 139", &
                         "IS 5", &
                         "BP 375", &
                         "EP 380", &
                         "UT ISI:A1987K321300001", &
                         "ER"), &
                         DOI="10.1016/0009-2614(87)80576-6")

      CALL add_reference(key=Weber2008, ISI_record=s2a( &
                         "AU Weber, V", &
                         "   VandeVondele, J", &
                         "   Hutter, J", &
                         "   Niklasson, AMN", &
                         "AF Weber, Valery", &
                         "   VandeVondele, Joost", &
                         "   Hutter, Juerg", &
                         "   Niklasson, Anders M. N.", &
                         "TI Direct energy functional minimization under orthogonality constraints", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD FEB 28", &
                         "PY 2008", &
                         "VL 128", &
                         "IS 8", &
                         "AR 084113", &
                         "DI ARTN 084113", &
                         "UT ISI:000254047200013", &
                         "ER"), &
                         DOI="10.1063/1.2841077")

      CALL add_reference(key=Stewart2007, ISI_record=s2a( &
                         "AU Stewart, JJP", &
                         "AF Stewart, James J. P.", &
                         "TI Optimization of parameters for semiempirical methods V: Modification of", &
                         "   NDDO approximations and application to 70 elements", &
                         "SO JOURNAL OF MOLECULAR MODELING", &
                         "SN 1610-2940", &
                         "PD DEC", &
                         "PY 2007", &
                         "VL 13", &
                         "IS 12", &
                         "BP 1173", &
                         "EP 1213", &
                         "UT ISI:000250308000001", &
                         "ER"), &
                         DOI="10.1007/s00894-007-0233-4")

      CALL add_reference(key=Repasky2002, ISI_record=s2a( &
                         "AU Repasky, MP", &
                         "   Chandrasekhar, J", &
                         "   Jorgensen, WL", &
                         "TI PDDG/PM3 and PDDG/MNDO: Improved semiempirical methods", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "PD DEC", &
                         "PY 2002", &
                         "VL 23", &
                         "IS 16", &
                         "BP 1601", &
                         "EP 1622", &
                         "UT ISI:000179022700011", &
                         "ER"), &
                         DOI="10.1002/jcc.10162")

      CALL add_reference(key=Thiel1992, ISI_record=s2a( &
                         "AU THIEL, W", &
                         "   VOITYUK, AA", &
                         "TI EXTENSION OF THE MNDO FORMALISM TO D-ORBITALS - INTEGRAL APPROXIMATIONS", &
                         "   AND PRELIMINARY NUMERICAL RESULTS", &
                         "SO THEORETICA CHIMICA ACTA", &
                         "SN 0040-5744", &
                         "PD FEB", &
                         "PY 1992", &
                         "VL 81", &
                         "IS 6", &
                         "BP 391", &
                         "EP 404", &
                         "UT ISI:A1992HF04700004", &
                         "ER"), &
                         DOI="10.1007/BF01134863")

      CALL add_reference(key=Stewart1989, ISI_record=s2a( &
                         "AU STEWART, JJP", &
                         "TI OPTIMIZATION OF PARAMETERS FOR SEMIEMPIRICAL METHODS .1. METHOD", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "PD MAR", &
                         "PY 1989", &
                         "VL 10", &
                         "IS 2", &
                         "BP 209", &
                         "EP 220", &
                         "UT ISI:A1989T690100007", &
                         "ER"), &
                         DOI="10.1002/jcc.540100208")

      CALL add_reference(key=Rocha2006, ISI_record=s2a( &
                         "AU Rocha, GB", &
                         "   Freire, RO", &
                         "   Simas, AM", &
                         "   Stewart, JJP", &
                         "TI RMI: A reparameterization of AM1 for H, C, N, O, P, S, F, Cl, Br, and I", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "PD JUL 30", &
                         "PY 2006", &
                         "VL 27", &
                         "IS 10", &
                         "BP 1101", &
                         "EP 1111", &
                         "UT ISI:000238158900007", &
                         "ER"), &
                         DOI="10.1002/jcc.20425")

      CALL add_reference(key=Dewar1985, ISI_record=s2a( &
                         "AU DEWAR, MJS", &
                         "   ZOEBISCH, EG", &
                         "   HEALY, EF", &
                         "   STEWART, JJP", &
                         "TI THE DEVELOPMENT AND USE OF QUANTUM-MECHANICAL MOLECULAR-MODELS .76. AM1", &
                         "   - A NEW GENERAL-PURPOSE QUANTUM-MECHANICAL MOLECULAR-MODEL", &
                         "SO JOURNAL OF THE AMERICAN CHEMICAL SOCIETY", &
                         "SN 0002-7863", &
                         "PY 1985", &
                         "VL 107", &
                         "IS 13", &
                         "BP 3902", &
                         "EP 3909", &
                         "UT ISI:A1985ALC3500024", &
                         "ER"), &
                         DOI="10.1021/ja00299a024")

      CALL add_reference(key=Dewar1977, ISI_record=s2a( &
                         "AU DEWAR, MJS", &
                         "   THIEL, W", &
                         "TI GROUND-STATES OF MOLECULES .38. MNDO METHOD - APPROXIMATIONS AND", &
                         "   PARAMETERS", &
                         "SO JOURNAL OF THE AMERICAN CHEMICAL SOCIETY", &
                         "SN 0002-7863", &
                         "PY 1977", &
                         "VL 99", &
                         "IS 15", &
                         "BP 4899", &
                         "EP 4907", &
                         "UT ISI:A1977DN51700004", &
                         "ER"), &
                         DOI="10.1021/ja00457a004")

      CALL add_reference(key=Henkelman1999, ISI_record=s2a( &
                         "PT J", &
                         "AU Henkelman, G", &
                         "   Jonsson, H", &
                         "TI A dimer method for finding saddle points on high dimensional potential", &
                         "   surfaces using only first derivatives", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD OCT 15", &
                         "PY 1999", &
                         "VL 111", &
                         "IS 15", &
                         "BP 7010", &
                         "EP 7022", &
                         "UT ISI:000083051600040", &
                         "ER"), &
                         DOI="10.1063/1.480097")

      CALL add_reference(key=Aguado2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Aguado, A", &
                         "   Madden, PA", &
                         "TI Ewald summation of electrostatic multipole interactions up to the", &
                         "   quadrupolar level", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD OCT 8", &
                         "PY 2003", &
                         "VL 119", &
                         "IS 14", &
                         "BP 7471", &
                         "EP 7483", &
                         "UT ISI:000185575700055", &
                         "ER"), &
                         DOI="10.1063/1.1605941")

      CALL add_reference(key=Yamada2000, ISI_record=s2a( &
                         "PT J", &
                         "AU Yamada, K", &
                         "   Kurosaki, K", &
                         "   Uno, M", &
                         "   Yamanaka, S", &
                         "TI Evaluation of thermal properties of uranium dioxide by molecular", &
                         "   dynamics", &
                         "SO JOURNAL OF ALLOYS AND COMPOUNDS", &
                         "SN 0925-8388", &
                         "PD JUL 14", &
                         "PY 2000", &
                         "VL 307", &
                         "BP 10", &
                         "EP 16", &
                         "UT ISI:000088590900003", &
                         "ER"), &
                         DOI="10.1016/S0925-8388(00)00806-9")

      CALL add_reference(key=Tosi1964a, ISI_record=s2a( &
                         "PT J", &
                         "AU FUMI, FG", &
                         "   TOSI, MP", &
                         "TI IONIC SIZES + BORN REPULSIVE PARAMETERS IN NACL-TYPE ALKALI HALIDES .I.", &
                         "   HUGGINS-MAYER + PAULING FORMS", &
                         "SO JOURNAL OF PHYSICS AND CHEMISTRY OF SOLIDS", &
                         "SN 0022-3697", &
                         "PY 1964", &
                         "VL 25", &
                         "IS 1", &
                         "BP 31", &
                         "EP 43", &
                         "UT ISI:A19646597B00004", &
                         "ER"), &
                         DOI="10.1016/0022-3697(64)90159-3")

      CALL add_reference(key=Tosi1964b, ISI_record=s2a( &
                         "PT J", &
                         "AU TOSI, MP", &
                         "   FUMI, FG", &
                         "TI IONIC SIZES + BORN REPULSIVE PARAMETERS IN NACL-TYPE ALKALI HALIDES .2.", &
                         "   GENERALIZED", &
                         "SO JOURNAL OF PHYSICS AND CHEMISTRY OF SOLIDS", &
                         "SN 0022-3697", &
                         "PY 1964", &
                         "VL 25", &
                         "IS 1", &
                         "BP 45", &
                         "EP 52", &
                         "UT ISI:A19646597B00017", &
                         "ER"), &
                         DOI="10.1016/0022-3697(64)90160-X")

      CALL add_reference(key=Tersoff1988, ISI_record=s2a( &
                         "PT J", &
                         "AU TERSOFF, J", &
                         "TI EMPIRICAL INTERATOMIC POTENTIAL FOR SILICON WITH IMPROVED ELASTIC", &
                         "   PROPERTIES", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD NOV 15", &
                         "PY 1988", &
                         "VL 38", &
                         "IS 14", &
                         "BP 9902", &
                         "EP 9905", &
                         "UT ISI:A1988Q924400066", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.38.9902")

      CALL add_reference(key=Siepmann1995, ISI_record=s2a( &
                         "PT J", &
                         "AU SIEPMANN, JI", &
                         "   SPRIK, M   ", &
                         "TI INFLUENCE OF SURFACE TOPOLOGY AND ELECTROSTATIC POTENTIAL ON", &
                         "   WATER/ELECTRODE SYSTEMS", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JAN 1", &
                         "PY 1995", &
                         "VL 102", &
                         "IS 1", &
                         "BP 511", &
                         "EP 524", &
                         "UT ISI:A1995PZ71400055", &
                         "ER"), &
                         DOI="10.1063/1.469429")

      CALL add_reference(key=Bussi2007, ISI_record=s2a( &
                         "PT J", &
                         "AU Bussi, G", &
                         "   Donadio, D", &
                         "   Parrinello, M", &
                         "AF Bussi, Giovanni", &
                         "   Donadio, Davide", &
                         "   Parrinello, Michele", &
                         "TI Canonical sampling through velocity rescaling", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JAN 7", &
                         "PY 2007", &
                         "VL 126", &
                         "IS 1", &
                         "AR 014101", &
                         "DI ARTN 014101", &
                         "UT ISI:000243380000005", &
                         "ER"), &
                         DOI="10.1063/1.2408420")

      CALL add_reference(key=Nose1984a, ISI_record=s2a( &
                         "PT J", &
                         "AU NOSE, S", &
                         "TI A UNIFIED FORMULATION OF THE CONSTANT TEMPERATURE MOLECULAR-DYNAMICS", &
                         "   METHODS", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PY 1984", &
                         "VL 81", &
                         "IS 1", &
                         "BP 511", &
                         "EP 519", &
                         "UT ISI:A1984TA66100062", &
                         "ER"), &
                         DOI="10.1063/1.447334")

      CALL add_reference(key=Nose1984b, ISI_record=s2a( &
                         "PT J", &
                         "AU NOSE, S", &
                         "TI A MOLECULAR-DYNAMICS METHOD FOR SIMULATIONS IN THE CANONICAL ENSEMBLE", &
                         "SO MOLECULAR PHYSICS", &
                         "LA English", &
                         "DT Article", &
                         "NR 17", &
                         "TC 1593", &
                         "PU TAYLOR & FRANCIS LTD", &
                         "PI LONDON", &
                         "PA ONE GUNDPOWDER SQUARE, LONDON, ENGLAND EC4A 3DE", &
                         "SN 0026-8976", &
                         "J9 MOL PHYS", &
                         "JI Mol. Phys.", &
                         "PY 1984", &
                         "VL 52", &
                         "IS 2", &
                         "BP 255", &
                         "EP 268", &
                         "PG 14", &
                         "SC Physics, Atomic, Molecular & Chemical", &
                         "GA SV647", &
                         "UT ISI:A1984SV64700001", &
                         "ER"), &
                         DOI="10.1080/00268978400101201")

      CALL add_reference(key=VandeVondele2005a, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Krack, M", &
                         "   Mohamed, F", &
                         "   Parrinello, M", &
                         "   Chassaing, T", &
                         "   Hutter, J", &
                         "TI QUICKSTEP: Fast and accurate density functional calculations using a", &
                         "   mixed Gaussian and plane waves approach", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "SN 0010-4655", &
                         "PD APR 15", &
                         "PY 2005", &
                         "VL 167", &
                         "IS 2", &
                         "BP 103", &
                         "EP 128", &
                         "UT ISI:000228421500005", &
                         "ER"), &
                         DOI="10.1016/j.cpc.2004.12.014")

      CALL add_reference(key=VandeVondele2003, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Hutter, J", &
                         "TI An efficient orbital transformation method for electronic structure", &
                         "   calculations", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD MAR 8", &
                         "PY 2003", &
                         "VL 118", &
                         "IS 10", &
                         "BP 4365", &
                         "EP 4369", &
                         "UT ISI:000181065800006", &
                         "ER"), &
                         DOI="10.1063/1.1543154")

      CALL add_reference(key=Laino2005, ISI_record=s2a( &
                         "PT J", &
                         "AU Laino, T", &
                         "   Mohamed, F", &
                         "   Laio, A", &
                         "   Parrinello, M", &
                         "TI An efficient real space multigrid QM/MM electrostatic coupling", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD NOV-DEC", &
                         "PY 2005", &
                         "VL 1", &
                         "IS 6", &
                         "BP 1176", &
                         "EP 1184", &
                         "UT ISI:000233260300012", &
                         "ER"), &
                         DOI="10.1021/ct050123f")

      CALL add_reference(key=Laino2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Laino, T", &
                         "   Mohamed, F", &
                         "   Laio, A", &
                         "   Parrinello, M", &
                         "AF Laino, Teodoro", &
                         "   Mohamed, Fawzi", &
                         "   Laio, Alessandro", &
                         "   Parrinello, Michele", &
                         "TI An efficient linear-scaling electrostatic coupling for treating", &
                         "   periodic boundary conditions in QM/MM simulations", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD SEP 12", &
                         "PY 2006", &
                         "VL 2", &
                         "IS 5", &
                         "BP 1370", &
                         "EP 1378", &
                         "UT ISI:000240437000015", &
                         "ER"), &
                         DOI="10.1021/ct6001169")

      CALL add_reference(key=Goedecker1996, ISI_record=s2a( &
                         "PT J", &
                         "AU Goedecker, S", &
                         "   Teter, M", &
                         "   Hutter, J", &
                         "TI Separable dual-space Gaussian pseudopotentials", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD JUL 15", &
                         "PY 1996", &
                         "VL 54", &
                         "IS 3", &
                         "BP 1703", &
                         "EP 1710", &
                         "UT ISI:A1996UZ86100053", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.54.1703")

      CALL add_reference(key=Hartwigsen1998, ISI_record=s2a( &
                         "PT J", &
                         "AU Hartwigsen, C", &
                         "   Goedecker, S", &
                         "   Hutter, J", &
                         "TI Relativistic separable dual-space Gaussian pseudopotentials from H to Rn", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD AUG 15", &
                         "PY 1998", &
                         "VL 58", &
                         "IS 7", &
                         "BP 3641", &
                         "EP 3662", &
                         "UT ISI:000075616800043", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.58.3641")

      CALL add_reference(key=Krack2005, ISI_record=s2a( &
                         "PT J", &
                         "AU Krack, M", &
                         "TI Pseudopotentials for H to Kr optimized for gradient-corrected", &
                         "   exchange-correlation functionals", &
                         "SO THEORETICAL CHEMISTRY ACCOUNTS", &
                         "SN 1432-881X", &
                         "PD SEP", &
                         "PY 2005", &
                         "VL 114", &
                         "IS 1-3", &
                         "BP 145", &
                         "EP 152", &
                         "UT ISI:000232053800020", &
                         "ER"), &
                         DOI="10.1007/s00214-005-0655-y")

      CALL add_reference(key=Lippert1997, ISI_record=s2a( &
                         "PT J", &
                         "AU Lippert, G", &
                         "   Hutter, J", &
                         "   Parrinello, M", &
                         "TI A hybrid Gaussian and plane wave density functional scheme", &
                         "SO MOLECULAR PHYSICS", &
                         "SN 0026-8976", &
                         "PD OCT 20", &
                         "PY 1997", &
                         "VL 92", &
                         "IS 3", &
                         "BP 477", &
                         "EP 487", &
                         "UT ISI:A1997YC60700017", &
                         "ER"), &
                         DOI="10.1080/002689797170220")

      CALL add_reference(key=Lippert1999, ISI_record=s2a( &
                         "PT J", &
                         "AU Lippert, G", &
                         "   Hutter, J", &
                         "   Parrinello, M", &
                         "TI The Gaussian and augmented-plane-wave density functional method for ab", &
                         "   initio molecular dynamics simulations", &
                         "SO THEORETICAL CHEMISTRY ACCOUNTS", &
                         "SN 1432-881X", &
                         "PD DEC", &
                         "PY 1999", &
                         "VL 103", &
                         "IS 2", &
                         "BP 124", &
                         "EP 140", &
                         "UT ISI:000084308100005", &
                         "ER"), &
                         DOI="10.1007/s002140050523")

      CALL add_reference(key=Krack2002, ISI_record=s2a( &
                         "PT J", &
                         "AU Krack, M", &
                         "   Gambirasio, A", &
                         "   Parrinello, M", &
                         "TI Ab initio x-ray scattering of liquid water", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD NOV 22", &
                         "PY 2002", &
                         "VL 117", &
                         "IS 20", &
                         "BP 9409", &
                         "EP 9412", &
                         "UT ISI:000179042300036", &
                         "ER"), &
                         DOI="10.1063/1.1517040")

      CALL add_reference(key=Krack2000, ISI_record=s2a( &
                         "PT J", &
                         "AU Krack, M", &
                         "   Parrinello, M", &
                         "TI All-electron ab-initio molecular dynamics", &
                         "SO PHYSICAL CHEMISTRY CHEMICAL PHYSICS", &
                         "SN 1463-9076", &
                         "PY 2000", &
                         "VL 2", &
                         "IS 10", &
                         "BP 2105", &
                         "EP 2112", &
                         "UT ISI:000086928000007", &
                         "ER"), &
                         DOI="10.1039/b001167n")

      CALL add_reference(key=Iannuzzi2007, ISI_record=s2a( &
                         "PT J", &
                         "AU Iannuzzi, M", &
                         "   Hutter, J", &
                         "AF Iannuzzi, Marcella", &
                         "   Hutter, Juerg", &
                         "TI Inner-shell spectroscopy by the Gaussian and augmented plane wave method", &
                         "SO PHYSICAL CHEMISTRY CHEMICAL PHYSICS", &
                         "SN 1463-9076", &
                         "PY 2007", &
                         "VL 9", &
                         "IS 13", &
                         "BP 1599", &
                         "EP 1610", &
                         "UT ISI:000245633400008", &
                         "ER"), &
                         DOI="10.1039/b615522g")

      CALL add_reference(key=Iannuzzi2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Iannuzzi, M", &
                         "   Kirchner, B", &
                         "   Hutter, J", &
                         "TI Density functional embedding for molecular systems", &
                         "SO CHEMICAL PHYSICS LETTERS", &
                         "SN 0009-2614", &
                         "PD APR 3", &
                         "PY 2006", &
                         "VL 421", &
                         "IS 1-3", &
                         "BP 16", &
                         "EP 20", &
                         "UT ISI:000236662600004", &
                         "ER"), &
                         DOI="10.1016/j.cplett.2005.08.155")

      CALL add_reference(key=Iannuzzi2005, ISI_record=s2a( &
                         "PT J", &
                         "AU Iannuzzi, M", &
                         "   Chassaing, T", &
                         "   Wallman, T", &
                         "   Hutter, J", &
                         "TI Ground and excited state density functional calculations with the", &
                         "   Gaussian and augmented-plane-wave method", &
                         "SO CHIMIA", &
                         "SN 0009-4293", &
                         "PY 2005", &
                         "VL 59", &
                         "IS 7-8", &
                         "BP 499", &
                         "EP 503", &
                         "UT ISI:000231615300005", &
                         "ER"), &
                         DOI="10.2533/000942905777676164")

      CALL add_reference(key=Toukmaji1996, ISI_record=s2a( &
                         "PT J", &
                         "AU Toukmaji, AY", &
                         "   Board, JA", &
                         "TI Ewald summation techniques in perspective: A survey", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "SN 0010-4655", &
                         "PD JUN", &
                         "PY 1996", &
                         "VL 95", &
                         "IS 2-3", &
                         "BP 73", &
                         "EP 92", &
                         "UT ISI:A1996UQ97300001", &
                         "ER"), &
                         DOI="10.1016/0010-4655(96)00016-1")

      CALL add_reference(key=Martyna1999, ISI_record=s2a( &
                         "PT J", &
                         "AU Martyna, GJ", &
                         "   Tuckerman, ME", &
                         "TI A reciprocal space based method for treating long range interactions in", &
                         "   ab initio and force-field-based calculations in clusters", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD FEB 8", &
                         "PY 1999", &
                         "VL 110", &
                         "IS 6", &
                         "BP 2810", &
                         "EP 2821", &
                         "UT ISI:000078379800010", &
                         "ER"), &
                         DOI="10.1063/1.477923")

      CALL add_reference(key=VandeVondele2005b, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Sprik, M", &
                         "TI A molecular dynamics study of the hydroxyl radical in solution applying", &
                         "   self-interaction-corrected density functional methods", &
                         "SO PHYSICAL CHEMISTRY CHEMICAL PHYSICS", &
                         "SN 1463-9076", &
                         "PY 2005", &
                         "VL 7", &
                         "IS 7", &
                         "BP 1363", &
                         "EP 1367", &
                         "UT ISI:000227707200005", &
                         "ER"), &
                         DOI="10.1039/b501603g")

      CALL add_reference(key=Perdew1981, ISI_record=s2a( &
                         "PT J", &
                         "AU PERDEW, JP", &
                         "   ZUNGER, A", &
                         "TI SELF-INTERACTION CORRECTION TO DENSITY-FUNCTIONAL APPROXIMATIONS FOR", &
                         "   MANY-ELECTRON SYSTEMS", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PY 1981", &
                         "VL 23", &
                         "IS 10", &
                         "BP 5048", &
                         "EP 5079", &
                         "UT ISI:A1981LR24700023", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.23.5048")

      CALL add_reference(key=Avezac2005, ISI_record=s2a( &
                         "PT J", &
                         'AU d''Avezac, M', &
                         "   Calandra, M", &
                         "   Mauri, F", &
                         "TI Density functional theory description of hole-trapping in SiO2: A", &
                         "   self-interaction-corrected approach", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD MAY", &
                         "PY 2005", &
                         "VL 71", &
                         "IS 20", &
                         "AR 205210", &
                         "DI ARTN 205210", &
                         "UT ISI:000230244300043", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.71.205210")

      CALL add_reference(key=Zhechkov2005, ISI_record=s2a( &
                         "PT J", &
                         "AU Zhechkov, L", &
                         "   Heine, T", &
                         "   Patchkovskii, S", &
                         "   Seifert, G", &
                         "   Duarte, HA", &
                         "TI An efficient a Posteriori treatment for dispersion interaction in", &
                         "   density-functional-based tight binding", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD SEP-OCT", &
                         "PY 2005", &
                         "VL 1", &
                         "IS 5", &
                         "BP 841", &
                         "EP 847", &
                         "UT ISI:000231871500009", &
                         "ER"), &
                         DOI="10.1021/ct050065y")

      CALL add_reference(key=Elstner1998, ISI_record=s2a( &
                         "PT J", &
                         "AU Elstner, M", &
                         "   Porezag, D", &
                         "   Jungnickel, G", &
                         "   Elsner, J", &
                         "   Haugk, M", &
                         "   Frauenheim, T", &
                         "   Suhai, S", &
                         "   Seifert, G", &
                         "TI Self-consistent-charge density-functional tight-binding method for", &
                         "   simulations of complex materials properties", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD SEP 15", &
                         "PY 1998", &
                         "VL 58", &
                         "IS 11", &
                         "BP 7260", &
                         "EP 7268", &
                         "UT ISI:000076058800081", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.58.7260")

      CALL add_reference(key=Seifert1996, ISI_record=s2a( &
                         "PT J", &
                         "AU Seifert, G", &
                         "   Porezag, D", &
                         "   Frauenheim, T", &
                         "TI Calculations of molecules, clusters, and solids with a simplified", &
                         "   LCAO-DFT-LDA scheme", &
                         "SO INTERNATIONAL JOURNAL OF QUANTUM CHEMISTRY", &
                         "SN 0020-7608", &
                         "PD APR 15", &
                         "PY 1996", &
                         "VL 58", &
                         "IS 2", &
                         "BP 185", &
                         "EP 192", &
                         "UT ISI:A1996UB73600008", &
                         "ER"), &
                         DOI="10.1002/(SICI)1097-461X(1996)58:2<185::AID-QUA7>3.0.CO;2-U")

      CALL add_reference(key=Porezag1995, ISI_record=s2a( &
                         "PT J", &
                         "AU POREZAG, D", &
                         "   FRAUENHEIM, T", &
                         "   KOHLER, T", &
                         "   SEIFERT, G", &
                         "   KASCHNER, R", &
                         "TI CONSTRUCTION OF TIGHT-BINDING-LIKE POTENTIALS ON THE BASIS OF", &
                         "   DENSITY-FUNCTIONAL THEORY - APPLICATION TO CARBON", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD MAY 15", &
                         "PY 1995", &
                         "VL 51", &
                         "IS 19", &
                         "BP 12947", &
                         "EP 12957", &
                         "UT ISI:A1995QZ16500002", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.51.12947")

      CALL add_reference(key=Frigo2005, ISI_record=s2a( &
                         "PT J", &
                         "AU Frigo, M", &
                         "   Johnson, SG", &
                         "TI The design and implementation of FFTW3", &
                         "SO PROCEEDINGS OF THE IEEE", &
                         "SN 0018-9219", &
                         "PD FEB", &
                         "PY 2005", &
                         "VL 93", &
                         "IS 2", &
                         "BP 216", &
                         "EP 231", &
                         "UT ISI:000226542300002", &
                         "ER"), &
                         DOI="10.1109/JPROC.2004.840301")

      CALL add_reference(key=Genovese2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Genovese, L", &
                         "   Deutsch, T", &
                         "   Neelov, A", &
                         "   Goedecker, S", &
                         "   Beylkin, G", &
                         "AF Genovese, Luigi", &
                         "   Deutsch, Thierry", &
                         "   Neelov, Alexey", &
                         "   Goedecker, Stefan", &
                         "   Beylkin, Gregory", &
                         "TI Efficient solution of Poisson's equation with free boundary conditions", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD AUG 21", &
                         "PY 2006", &
                         "VL 125", &
                         "IS 7", &
                         "AR 074105", &
                         "DI ARTN 074105", &
                         "UT ISI:000239914800005", &
                         "ER"), &
                         DOI="10.1063/1.2335442")

      CALL add_reference(key=Genovese2007, ISI_record=s2a( &
                         "PT J", &
                         "AU Genovese, L", &
                         "   Deutsch, T", &
                         "   Goedecker, S", &
                         "AF Genovese, Luigi", &
                         "   Deutsch, Thierry", &
                         "   Goedecker, Stefan", &
                         "TI Efficient and accurate three-dimensional Poisson solver for surface", &
                         "   problems", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD AUG 7", &
                         "PY 2007", &
                         "VL 127", &
                         "IS 5", &
                         "AR 054704", &
                         "DI ARTN 054704", &
                         "UT ISI:000248594200029", &
                         "ER"), &
                         DOI="10.1063/1.2754685")

      CALL add_reference(key=Minary2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Minary, P", &
                         "   Martyna, GJ", &
                         "   Tuckerman, ME", &
                         "TI Algorithms and novel applications based on the isokinetic ensemble. I.", &
                         "   Biophysical and path integral molecular dynamics", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD FEB 8", &
                         "PY 2003", &
                         "VL 118", &
                         "IS 6", &
                         "BP 2510", &
                         "EP 2526", &
                         "UT ISI:000180564800009", &
                         "ER"), &
                         DOI="10.1063/1.1534582")

      CALL add_reference(key=Evans1983, ISI_record=s2a( &
                         "PT J", &
                         "AU EVANS, DJ", &
                         "   HOOVER, WG", &
                         "   FAILOR, BH", &
                         "   MORAN, B", &
                         "   LADD, AJC", &
                         "TI NON-EQUILIBRIUM MOLECULAR-DYNAMICS VIA GAUSS PRINCIPLE OF LEAST", &
                         "   CONSTRAINT", &
                         "SO PHYSICAL REVIEW A", &
                         "SN 1050-2947", &
                         "PY 1983", &
                         "VL 28", &
                         "IS 2", &
                         "BP 1016", &
                         "EP 1021", &
                         "UT ISI:A1983RB52800057", &
                         "ER"), &
                         DOI="10.1103/PhysRevA.28.1016")

      CALL add_reference(key=Byrd1995, ISI_record=s2a( &
                         "PT J", &
                         "TI A LIMITED MEMORY ALGORITHM FOR BOUND CONSTRAINED OPTIMIZATION", &
                         "AU BYRD, RH", &
                         "   LU, PH", &
                         "   NOCEDAL, J", &
                         "   ZHU, CY", &
                         "SO SIAM JOURNAL ON SCIENTIFIC COMPUTING", &
                         "VL 16", &
                         "IS 5", &
                         "BP 1190", &
                         "EP 1208", &
                         "PY 1995", &
                         "TC 214", &
                         "UT WOS:A1995RR54100011", &
                         "ER"), &
                         DOI="10.1137/0916069")

      CALL add_reference(key=VandeVondele2007, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Hutter, J", &
                         "AF VandeVondele, Joost", &
                         "   Hutter, Juerg", &
                         "TI Gaussian basis sets for accurate calculations on molecular systems in", &
                         "   gas and condensed phases", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD SEP 21", &
                         "PY 2007", &
                         "VL 127", &
                         "IS 11", &
                         "AR 114105", &
                         "DI ARTN 114105", &
                         "UT ISI:000249667400011", &
                         "ER"), &
                         DOI="10.1063/1.2770708")

      CALL add_reference(key=Ortiz1994, ISI_record=s2a( &
                         "PT J", &
                         "AU ORTIZ, G", &
                         "   BALLONE, P", &
                         "TI CORRELATION-ENERGY, STRUCTURE FACTOR, RADIAL-DISTRIBUTION FUNCTION, AND", &
                         "   MOMENTUM DISTRIBUTION OF THE SPIN-POLARIZED UNIFORM ELECTRON-GAS", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD JUL 15", &
                         "PY 1994", &
                         "VL 50", &
                         "IS 3", &
                         "BP 1391", &
                         "EP 1405", &
                         "UT ISI:A1994NZ15300007", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.50.1391")

      CALL add_reference(key=Becke1988, ISI_record=s2a( &
                         "PT J", &
                         "AU BECKE, AD", &
                         "TI DENSITY-FUNCTIONAL EXCHANGE-ENERGY APPROXIMATION WITH CORRECT", &
                         "   ASYMPTOTIC-BEHAVIOR", &
                         "SO PHYSICAL REVIEW A", &
                         "SN 1050-2947", &
                         "PD SEP 15", &
                         "PY 1988", &
                         "VL 38", &
                         "IS 6", &
                         "BP 3098", &
                         "EP 3100", &
                         "UT ISI:A1988Q146900044", &
                         "ER"), &
                         DOI="10.1103/PhysRevA.38.3098")

      CALL add_reference(key=Perdew1996, ISI_record=s2a( &
                         "PT J", &
                         "AU Perdew, JP", &
                         "   Burke, K", &
                         "   Ernzerhof, M", &
                         "TI Generalized gradient approximation made simple", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "PD OCT 28", &
                         "PY 1996", &
                         "VL 77", &
                         "IS 18", &
                         "BP 3865", &
                         "EP 3868", &
                         "UT ISI:A1996VP22500044", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.77.3865")

      CALL add_reference(key=Zhang1998, ISI_record=s2a( &
                         "PT J", &
                         "AU Zhang, YK", &
                         "   Yang, WT", &
                         "TI Comment on Generalized gradient approximation made simple", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "PD JAN 26", &
                         "PY 1998", &
                         "VL 80", &
                         "IS 4", &
                         "BP 890", &
                         "EP 890", &
                         "UT ISI:000071717100066", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.80.890")

      CALL add_reference(key=Perdew2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Perdew, JP", &
                         "   Ruzsinszky, A", &
                         "   Csonka, GI", &
                         "   Vydrov, OA", &
                         "   Scuseria, GE", &
                         "   Constantin, LA", &
                         "   Zhou, X", &
                         "   Burke, K", &
                         "TI Restoring the Density-Gradient Expansion for Exchange in Solids and Surfaces", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN ", &
                         "PD APR 04", &
                         "PY 2008", &
                         "VL 100", &
                         "IS 13", &
                         "BP 136406", &
                         "EP 136409", &
                         "UT ", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.100.136406")

      CALL add_reference(key=Lee1988, ISI_record=s2a( &
                         "PT J", &
                         "AU LEE, CT", &
                         "   YANG, WT", &
                         "   PARR, RG", &
                         "TI DEVELOPMENT OF THE COLLE-SALVETTI CORRELATION-ENERGY FORMULA INTO A", &
                         "   FUNCTIONAL OF THE ELECTRON-DENSITY", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD JAN 15", &
                         "PY 1988", &
                         "VL 37", &
                         "IS 2", &
                         "BP 785", &
                         "EP 789", &
                         "UT ISI:A1988L976200011", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.37.785")

      CALL add_reference(key=Heyd2004, ISI_record=s2a( &
                         "PT J", &
                         "AU Heyd, J", &
                         "   Scuseria, GE", &
                         "TI Assessment and validation of a screened Coulomb hybrid density", &
                         "   functional", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD APR 22", &
                         "PY 2004", &
                         "VL 120", &
                         "IS 16", &
                         "BP 7274", &
                         "EP 7280", &
                         "UT ISI:000220676000005", &
                         "ER"), &
                         DOI="10.1063/1.1668634")

      CALL add_reference(key=Heyd2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Heyd, J", &
                         "   Scuseria, GE", &
                         "   Ernzerhof, M", &
                         "TI Hybrid functionals based on a screened Coulomb potential", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD MAY 8", &
                         "PY 2003", &
                         "VL 118", &
                         "IS 18", &
                         "BP 8207", &
                         "EP 8215", &
                         "UT ISI:000182454100010", &
                         "ER"), &
                         DOI="10.1063/1.1564060")

      CALL add_reference(key=Heyd2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Heyd, J", &
                         "   Scuseria, GE", &
                         "   Ernzerhof, M", &
                         "TI Hybrid functionals based on a screened Coulomb potential (vol 118, pg", &
                         "   8207, 2003)", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JUN 7", &
                         "PY 2006", &
                         "VL 124", &
                         "IS 21", &
                         "AR 219906", &
                         "DI ARTN 219906", &
                         "UT ISI:000238758700058", &
                         "ER"), &
                         DOI="10.1063/1.2204597")

      CALL add_reference(key=Vydrov2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Vydrov, OA", &
                         "   Heyd, J", &
                         "   Krukau, AV", &
                         "   Scuseria, GE", &
                         "AF Vydrov, Oleg A.", &
                         "   Heyd, Jochen", &
                         "   Krukau, Aliaksandr V.", &
                         "   Scuseria, Gustavo E.", &
                         "TI Importance of short-range versus long-range Hartree-Fock exchange for", &
                         "   the performance of hybrid density functionals", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD AUG 21", &
                         "PY 2006", &
                         "VL 125", &
                         "IS 7", &
                         "AR 074106", &
                         "DI ARTN 074106", &
                         "UT ISI:000239914800006", &
                         "ER"), &
                         DOI="10.1063/1.2244560")

      CALL add_reference(key=Vosko1980, ISI_record=s2a( &
                         "PT J", &
                         "AU VOSKO, SH", &
                         "   WILK, L", &
                         "   NUSAIR, M", &
                         "TI ACCURATE SPIN-DEPENDENT ELECTRON LIQUID CORRELATION ENERGIES FOR LOCAL", &
                         "   SPIN-DENSITY CALCULATIONS - A CRITICAL ANALYSIS", &
                         "SO CANADIAN JOURNAL OF PHYSICS", &
                         "SN 0008-4204", &
                         "PY 1980", &
                         "VL 58", &
                         "IS 8", &
                         "BP 1200", &
                         "EP 1211", &
                         "UT ISI:A1980KE76300015", &
                         "ER"), &
                         DOI="10.1139/p80-159") ! this doi should be correct later this week, but does not yet resolve

      CALL add_reference(key=Essmann1995, ISI_record=s2a( &
                         "PT J", &
                         "AU ESSMANN, U", &
                         "   PERERA, L", &
                         "   BERKOWITZ, ML", &
                         "   DARDEN, T", &
                         "   LEE, H", &
                         "   PEDERSEN, LG", &
                         "TI A SMOOTH PARTICLE MESH EWALD METHOD", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD NOV 15", &
                         "PY 1995", &
                         "VL 103", &
                         "IS 19", &
                         "BP 8577", &
                         "EP 8593", &
                         "UT ISI:A1995TE36400026", &
                         "ER"), &
                         DOI="10.1063/1.470117")

      CALL add_reference(key=Ewald1921, ISI_record=s2a( &
                         "PT J", &
                         "AU Ewald, PP", &
                         "TI Die Berechnung optischer und elektrostatischer Gitterpotentiale", &
                         "SO ANNALEN DER PHYSIK", &
                         "SN 0003-3804", &
                         "PD FEB", &
                         "PY 1921", &
                         "VL 369", &
                         "IS 3", &
                         "BP 253", &
                         "EP 287", &
                         "UT ISI:000201909400003", &
                         "ER"), &
                         DOI="10.1002/andp.19213690304")

      CALL add_reference(key=Darden1993, ISI_record=s2a( &
                         "PT J", &
                         "AU DARDEN, T", &
                         "   YORK, D", &
                         "   PEDERSEN, L", &
                         "TI PARTICLE MESH EWALD - AN N.LOG(N) METHOD FOR EWALD SUMS IN LARGE SYSTEMS", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JUN 15", &
                         "PY 1993", &
                         "VL 98", &
                         "IS 12", &
                         "BP 10089", &
                         "EP 10092", &
                         "UT ISI:A1993LG10100091", &
                         "ER"), &
                         DOI="10.1063/1.464397")

      CALL add_reference(key=Dudarev1997, ISI_record=s2a( &
                         "PT J", &
                         "AU Dudarev, SL", &
                         "   Manh, DN", &
                         "   Sutton, AP", &
                         "TI Effect of Mott-Hubbard correlations on the electronic", &
                         "   structure and structural stability of uranium dioxide", &
                         "SO PHILOSOPHICAL MAGAZINE B", &
                         "SN 0141-8637", &
                         "PD MAY", &
                         "PY 1997", &
                         "VL 75", &
                         "IS 5", &
                         "BP 613", &
                         "EP 628", &
                         "UT ISI:A1997WX94300001", &
                         "ER"), &
                         DOI="10.1080/13642819708202343")

      CALL add_reference(key=Dudarev1998, ISI_record=s2a( &
                         "PT J", &
                         "AU Dudarev, SL", &
                         "   Botton, GA", &
                         "   Savrasov, SY", &
                         "   Humphreys, CJ", &
                         "   Sutton, AP", &
                         "TI Electron-energy-loss spectra and the structural stability of", &
                         "   nickel oxide: An LSDA+U study", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PD JAN 15", &
                         "PY 1998", &
                         "VL 57", &
                         "IS 3", &
                         "BP 1505", &
                         "EP 1509", &
                         "UT ISI:000071716800040", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.57.1505")

      CALL add_reference(key=Hunt2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Hunt, P", &
                         "   Sprik, M", &
                         "   Vuilleumier, R", &
                         "TI Thermal versus electronic broadening in the density of states of liquid", &
                         "   water", &
                         "SO CHEMICAL PHYSICS LETTERS", &
                         "SN 0009-2614", &
                         "PD JUL 17", &
                         "PY 2003", &
                         "VL 376", &
                         "IS 1-2", &
                         "BP 68", &
                         "EP 74", &
                         "UT ISI:000184315600011", &
                         "ER"), &
                         DOI="10.1016/S0009-2614(03)00954-0")

      CALL add_reference(key=Guidon2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Guidon, M", &
                         "   Schiffmann, F", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Guidon, Manuel", &
                         "   Schiffmann, Florian", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Ab initio molecular dynamics using hybrid density functionals", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JUN 7", &
                         "PY 2008", &
                         "VL 128", &
                         "IS 21", &
                         "AR 214104", &
                         "DI ARTN 214104", &
                         "UT ISI:000256527500005", &
                         "ER"), &
                         DOI="10.1063/1.2931945")

      CALL add_reference(key=Stewart1982, ISI_record=s2a( &
                         "PT J", &
                         "AU Stewart, JJP", &
                         "   Csaszar, P", &
                         "   Pulay, P", &
                         "TI Fast semi-empirical calculations", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "PY 1982", &
                         "VL 3", &
                         "IS 2", &
                         "BP 227", &
                         "EP 228", &
                         "UT ISI:A1982NN22800013", &
                         "ER"), &
                         DOI="10.1002/jcc.540030214")

      CALL add_reference(key=Tao2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Tao, JM", &
                         "   Perdew, JP", &
                         "   Staroverov, VN", &
                         "   Scuseria, GE", &
                         "TI Climbing the density functional ladder: Nonempirical meta-generalized", &
                         "   gradient approximation designed for molecules and solids", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "PD OCT 3", &
                         "PY 2003", &
                         "VL 91", &
                         "IS 14", &
                         "AR 146401", &
                         "DI 10.1103/PhysRevLett.91.146401", &
                         "UT ISI:000185719500032", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.91.146401")

      CALL add_reference(key=VandeVondele2006, ISI_record=s2a( &
                         "PT C", &
                         "AU VandeVondele, J", &
                         "   Iannuzzi, M", &
                         "   Hutter, J", &
                         "ED Ferrario, M; Ciccotti, G; Binder, K", &
                         "TI Large scale condensed matter calculations using the gaussian and", &
                         "   augmented plane waves method", &
                         "SO Computer Simulations in Condensed Matter Systems: From Materials to", &
                         "   Chemical Biology, Vol 1", &
                         "SE LECTURE NOTES IN PHYSICS", &
                         "LA English", &
                         "DT Proceedings Paper", &
                         "CT Conference on Computer Simulations in Condensed Matter Systems", &
                         "CY JUL, 2005", &
                         "CL Erice, ITALY", &
                         "SP Ctr Sci Culture", &
                         "HO Ettore Majorana Fdn", &
                         "NR 35", &
                         "TC 2", &
                         "PU SPRINGER", &
                         "PI NEW YORK", &
                         "PA 233 SPRING STREET, NEW YORK, NY 10013, UNITED STATES", &
                         "SN 0075-8450", &
                         "BN 978-3-540-35270-9", &
                         "J9 LECT NOTES PHYS", &
                         "PY 2006", &
                         "VL 703", &
                         "BP 287", &
                         "EP 314", &
                         "DI 10.1007/3-540-35273-2_8", &
                         "PG 28", &
                         "SC Physics, Multidisciplinary", &
                         "GA BFQ13", &
                         "UT ISI:000243725000009", &
                         "ER"), &
                         DOI="10.1007/3-540-35273-2_8")

      CALL add_reference(key=Grimme2006, ISI_record=s2a( &
                         "PT J", &
                         "AU Grimme, S", &
                         "AF Grimme, Stefan", &
                         "TI Semiempirical GGA-type density functional constructed with a long-range", &
                         "   dispersion correction", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "PD NOV 30", &
                         "PY 2006", &
                         "VL 27", &
                         "IS 15", &
                         "BP 1787", &
                         "EP 1799", &
                         "DI 10.1002/jcc.20495", &
                         "UT ISI:000241477200003", &
                         "ER"), &
                         DOI="10.1002/jcc.20495")

      CALL add_reference(key=Grimme2010, ISI_record=s2a( &
                         "PT J", &
                         "AU Grimme, S", &
                         "   Antony, J", &
                         "   Ehrlich, S", &
                         "   Krieg, H", &
                         "AF Grimme, Stefan", &
                         "   Antony, Jens", &
                         "   Ehrlich, Stephan", &
                         "   Krieg, Helge", &
                         "TI A consistent and accurate ab initio parametrization of density", &
                         "   functional dispersion correction (DFT-D) for the 94 elements H-Pu", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD APR 21", &
                         "PY 2010", &
                         "VL 132", &
                         "IS 15", &
                         "AR 154104", &
                         "DI 10.1063/1.3382344", &
                         "UT ISI:000276971500005", &
                         "ER"), &
                         DOI="10.1063/1.3382344")

      CALL add_reference(key=Grimme2011, ISI_record=s2a( &
                         "PT J", &
                         "AU Grimme, S", &
                         "   Ehrlich, S", &
                         "   Goerigk, L", &
                         "AF Grimme, Stefan", &
                         "   Ehrlich, Stephan", &
                         "   Goerigk, Lars", &
                         "TI Effect of the damping function in dispersion corrected density functional theory", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "PY 2011", &
                         "VL 32", &
                         "BP 1456", &
                         "DI 10.1002/jcc.21759", &
                         "ER"), &
                         DOI="10.1002/jcc.21759")

      CALL add_reference(key=Branduardi2007, ISI_record=s2a( &
                         "PT J", &
                         "AU Branduardi, D", &
                         "   Gervasio, FL", &
                         "   Parrinello, M", &
                         "AF Branduardi, Davide", &
                         "   Gervasio, Francesco Luigi", &
                         "   Parrinello, Michele", &
                         "TI From A to B in free energy space", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD FEB 7", &
                         "PY 2007", &
                         "VL 126", &
                         "IS 5", &
                         "AR 054103", &
                         "DI 10.1063/1.2432340", &
                         "UT ISI:000244044600009", &
                         "ER"), &
                         DOI="10.1063/1.2432340")

      CALL add_reference(key=Schenter2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Chang, DT", &
                         "   Schenter, GK", &
                         "   Garrett, BC ", &
                         "TI Self-consistent polarization neglect of diatomic differential overlap: "// &
                         "Applications to water clusters", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD APR 28", &
                         "PY 2008", &
                         "VL 128", &
                         "IS 16", &
                         "AR 164111", &
                         "DI 10.1063/1.2905230", &
                         "ER"), &
                         DOI="10.1063/1.2905230")

      CALL add_reference(key=Proynov2007, ISI_record=s2a( &
                         "AU Proynov, E", &
                         "   Gan, Z", &
                         "   Kong, J", &
                         "AF Proynov, Emil", &
                         "   Gan, Zhenting", &
                         "   Kong, Jing", &
                         "TI Analytical representation of the Becke-Roussel exchange functional", &
                         "SO CHEMICAL PHYSICS LETTERS", &
                         "SN 0009-2614", &
                         "PD MAR 31", &
                         "PY 2008", &
                         "VL 455", &
                         "IS 1-3", &
                         "BP 103", &
                         "EP 109", &
                         "DI 10.1016/j.cplett.2008.02.039", &
                         "UT ISI:000254425400020", &
                         "ER"), &
                         DOI="10.1016/j.cplett.2008.02.039")

      CALL add_reference(key=BeckeRoussel1989, ISI_record=s2a( &
                         "AU BECKE, AD", &
                         "   ROUSSEL, MR", &
                         "TI EXCHANGE HOLES IN INHOMOGENEOUS SYSTEMS - A COORDINATE-SPACE MODEL", &
                         "SO PHYSICAL REVIEW A", &
                         "DT Article", &
                         "RP BECKE, AD, QUEENS UNIV,DEPT CHEM,KINGSTON K7L 3N6,ONTARIO,CANADA.", &
                         "NR 17", &
                         "TC 114", &
                         "PU AMERICAN PHYSICAL SOC", &
                         "PI COLLEGE PK", &
                         "PA ONE PHYSICS ELLIPSE, COLLEGE PK, MD 20740-3844 USA", &
                         "SN 1050-2947", &
                         "J9 PHYS REV A", &
                         "JI Phys. Rev. A", &
                         "PD APR 15", &
                         "PY 1989", &
                         "VL 39", &
                         "IS 8", &
                         "BP 3761", &
                         "EP 3767", &
                         "PG 7", &
                         "SC Optics; Physics, Atomic, Molecular & Chemical", &
                         "GA U3175", &
                         "UT ISI:A1989U317500001", &
                         "ER"), &
                         DOI="10.1103/PhysRevA.39.3761")

      CALL add_reference(key=Becke1997, ISI_record=s2a( &
                         "AU Becke, AD", &
                         "PY 1997", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "TI Density-functional thermochemistry .", &
                         "   5. Systematic optimization of exchange-correlation functionals", &
                         "VL 107", &
                         "IS 20", &
                         "SP 8554", &
                         "EP 8560"), &
                         DOI="10.1063/1.475007")

      CALL add_reference(key=Ricci2003, ISI_record=s2a( &
                         "AU Ricci, A", &
                         "   Ciccotti, G", &
                         "TI Algorithms for Brownian dynamics", &
                         "SO MOLECULAR PHYSICS", &
                         "SN 0026-8976", &
                         "J9 MOL PHYS", &
                         "JI Mol. Phys.", &
                         "PD JUN 20", &
                         "PY 2003", &
                         "VL 101", &
                         "IS 12", &
                         "BP 1927", &
                         "EP 1931", &
                         "UT ISI:000184158800014", &
                         "ER"), &
                         DOI="10.1080/0026897031000108113")

      CALL add_reference(key=Kolafa2004, ISI_record=s2a( &
                         "AU Kolafa, J", &
                         "TI Time-reversible always stable predictor-corrector method for "// &
                         "molecular dynamics of polarizable molecules", &
                         "SO JOURNAL OF COMPUTATIONAL CHEMISTRY", &
                         "SN 0192-8651", &
                         "J9 J COMPUT CHEM", &
                         "JI J. Comput. Chem.", &
                         "PD FEB", &
                         "PY 2004", &
                         "VL 25", &
                         "IS 3", &
                         "BP 335", &
                         "EP 342", &
                         "UT ISI:000188245900004", &
                         "ER"), &
                         DOI="10.1002/jcc.10385")

      CALL add_reference(key=Kuhne2007, ISI_record=s2a( &
                         "AU Kuhne, TD", &
                         "   Krack, M", &
                         "   Mohamed, FR", &
                         "   Parrinello, M", &
                         "TI Efficient and accurate Car-Parrinello-like approach to Born-Oppenheimer molecular dynamics", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "J9 PHYS REV LETT", &
                         "JI Phys. Rev. Lett.", &
                         "PD FEB 9", &
                         "PY 2007", &
                         "VL 98", &
                         "IS 6", &
                         "AR 066401", &
                         "PG 4", &
                         "UT ISI:000244117400046", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.98.066401")

      CALL add_reference(key=Kunert2003, ISI_record=s2a( &
                         "AU Kunert, T", &
                         "   Schmidt, R", &
                         "TI Non-adiabatic quantum molecular dynamics: General formalism and case", &
                         "   study H-2(+) in strong laser fields", &
                         "SO EUROPEAN PHYSICAL JOURNAL D", &
                         "SN 1434-6060", &
                         "J9 EUR PHYS J D", &
                         "JI Eur. Phys. J. D", &
                         "PD JUL", &
                         "PY 2003", &
                         "VL 25", &
                         "IS 1", &
                         "PG 10", &
                         "UT ISI:000184771500002", &
                         "ER"), &
                         DOI="10.1140/epjd/e2003-00086-8")

      CALL add_reference(key=Ceriotti2009, ISI_record=s2a( &
                         "AU Ceriotti, M", &
                         "   Bussi, G", &
                         "   Parrinello, M", &
                         "TI Langevin equation with colored noise for constant-temperature molecular dynamics simulations", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "J9 PHYS REV LETT", &
                         "JI Phys. Rev. Lett.", &
                         "PD JAN 16", &
                         "PY 2009", &
                         "VL 102", &
                         "IS 2", &
                         "AR 020601", &
                         "PG 4", &
                         "UT ISI:000244117400046", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.102.020601")

      CALL add_reference(key=Ceriotti2009b, ISI_record=s2a( &
                         "AU Ceriotti, M", &
                         "   Bussi, G", &
                         "   Parrinello, M", &
                         "AF Ceriotti, Michele", &
                         "   Bussi, Giovanni", &
                         "   Parrinello, Michele", &
                         "TI Nuclear Quantum Effects in Solids Using a Colored-Noise Thermostat", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "LA English", &
                         "J9 PHYS REV LETT", &
                         "JI Phys. Rev. Lett.", &
                         "PD JUL 17", &
                         "PY 2009", &
                         "VL 103", &
                         "IS 3", &
                         "AR 030603", &
                         "PG 4", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.103.030603")

      CALL add_reference(key=Guidon2009, ISI_record=s2a( &
                         "AU Guidon, M", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Guidon, Manuel", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Robust Periodic Hartree-Fock Exchange for Large-Scale Simulations Using", &
                         "   Gaussian Basis Sets", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD NOV", &
                         "PY 2009", &
                         "VL 5", &
                         "IS 11", &
                         "BP 3010", &
                         "EP 3021", &
                         "DI 10.1021/ct900494g", &
                         "UT ISI:000271522500011", &
                         "ER"), &
                         DOI="10.1021/ct900494g")

      CALL add_reference(key=BarducBus2008, ISI_record=s2a( &
                         "AU Barducci, A", &
                         "   Bussi, G", &
                         "   Parrinello, M", &
                         "TI Well-Tempered Metadynamics: A Smoothly Converging and Tunable Free-Energy Method", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "J9 PHYS REV LETT", &
                         "JI Phys. Rev. Lett.", &
                         "PD JAN 18", &
                         "PY 2008", &
                         "VL 100", &
                         "IS 2", &
                         "AR 020603", &
                         "PG 4", &
                         "UT ISI:000252471200008", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.100.020603")

      CALL add_reference(key=Guidon2010, ISI_record=s2a( &
                         "AU Guidon, M", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Guidon, Manuel", &
                         "   Hutter, Jurg", &
                         "   VandeVondele, Joost", &
                         "TI Auxiliary Density Matrix Methods for Hartree-Fock Exchange Calculations", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "LA English", &
                         "DT Article", &
                         "NR 65", &
                         "TC 0", &
                         "PU AMER CHEMICAL SOC", &
                         "PI WASHINGTON", &
                         "PA 1155 16TH ST, NW, WASHINGTON, DC 20036 USA", &
                         "SN 1549-9618", &
                         "J9 J CHEM THEORY COMPUT", &
                         "JI J. Chem. Theory Comput.", &
                         "PD AUG", &
                         "PY 2010", &
                         "VL 6", &
                         "IS 8", &
                         "BP 2348", &
                         "EP 2364", &
                         "DI 10.1021/ct1002225", &
                         "PG 17", &
                         "SC Chemistry, Physical; Physics, Atomic, Molecular & Chemical", &
                         "GA 635NE", &
                         "UT ISI:000280661600009", &
                         "ER"), &
                         DOI="10.1021/ct1002225")

      CALL add_reference(key=Marques2012, ISI_record=s2a( &
                         "AU Marques, MAL", &
                         "   Oliveira, MJT", &
                         "   Burnus, T", &
                         "AF Marques, Miguel A. L.", &
                         "   Oliveira, Micael J. T.", &
                         "   Burnus, Tobias", &
                         "TI LIBXC: A library of exchange and correlation functionals for density", &
                         "   functional theory", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "SN 0010-4655", &
                         "PD OCT", &
                         "PY 2012", &
                         "VL 183", &
                         "IS 10", &
                         "BP 2272", &
                         "EP 2281", &
                         "DI 10.1016/j.cpc.2012.05.007", &
                         "UT WOS:000306771900024", &
                         "ER"), &
                         DOI="10.1016/j.cpc.2012.05.007")

      CALL add_reference(key=Jones2011, ISI_record=s2a( &
                         "AU Jones, Andrew", &
                         "   Leimkuhler, Ben", &
                         "TI Adaptive stochastic methods for sampling driven molecular systems", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "VL 135", &
                         "IS 8", &
                         "AR 084125", &
                         "DI 10.1063/1.3626941", &
                         "PD AUG 28", &
                         "PY 2011", &
                         "TC 0", &
                         "SN 0021-9606", &
                         "UT WOS:000294484700031", &
                         "ER"), &
                         DOI="10.1063/1.3626941")

      CALL add_reference(key=Bernstein2012, ISI_record=s2a( &
                         "AU Bernstein, Noam", &
                         "   Varnai, Csilla", &
                         "   Solt, Ivan", &
                         "   Winfield, Steven A", &
                         "   Payne, Mike C", &
                         "   Simon, Istvan", &
                         "   Fuxreiter, Monika", &
                         "   Csanyi, Gabor", &
                         "TI QM/MM simulation of liquid water with an adaptive quantum region", &
                         "SO PHYSICAL CHEMISTRY CHEMICAL PHYSICS", &
                         "VL 14", &
                         "IS 2", &
                         "BP 646", &
                         "EP 656", &
                         "DI 10.1039/c1cp22600b", &
                         "PY 2012", &
                         "TC 0", &
                         "ZB 0", &
                         "Z8 0", &
                         "Z9 0", &
                         "SN 1463-9076", &
                         "UT WOS:000298552800027", &
                         "ER"), &
                         DOI="10.1039/c1cp22600b")

      CALL add_reference(key=Bernstein2009, ISI_record=s2a( &
                         "AU Bernstein, N", &
                         "   Kermode, J R", &
                         "   Csanyi, G", &
                         "TI Hybrid atomistic simulation methods for materials systems", &
                         "SO REPORTS ON PROGRESS IN PHYSICS", &
                         "VL 72", &
                         "IS 2", &
                         "AR 026501", &
                         "DI 10.1088/0034-4885/72/2/026501", &
                         "PD FEB 1", &
                         "PY 2009", &
                         "TC 25", &
                         "ZB 0", &
                         "Z8 0", &
                         "Z9 25", &
                         "SN 0034-4885", &
                         "UT WOS:000262785900001", &
                         "ER"), &
                         DOI="10.1088/0034-4885/72/2/026501")

      CALL add_reference(key=Dick1958, ISI_record=s2a( &
                         "PT J", &
                         "AU Dick, BG", &
                         "   Overhauser, AW", &
                         "TI Theory of the Dielectric Constants of Alkali Halide Crystals", &
                         "SO Phys. Rev.", &
                         "VL 112", &
                         "IS 1", &
                         "SP 90", &
                         "EP 103", &
                         "PD OCT 1", &
                         "PY 1958", &
                         "UR http://link.aps.org/doi/10.1103/PhysRev.112.90", &
                         "ER"), &
                         DOI="10.1103/PhysRev.112.90")

      CALL add_reference(key=Mitchell1993, ISI_record=s2a( &
                         "PT J", &
                         "AU Mitchell, PJ", &
                         "   Fincham D", &
                         "TI Shell model simulations by adiabatic dynamics", &
                         "SO J. Phys.: Condens. Matter", &
                         "VL 5", &
                         "IS 8", &
                         "SP 1031", &
                         "EP 1038", &
                         "PD FEB 22", &
                         "PY 1993", &
                         "UR http://stacks.iop.org/0953-8984/5/i=8/a=006", &
                         "ER"), &
                         DOI="10.1088/0953-8984/5/8/006")

      CALL add_reference(key=Devynck2012, ISI_record=s2a( &
                         "PT J", &
                         "AU Devynck, F", &
                         "   Iannuzzi, M", &
                         "   Krack, M", &
                         "AF Devynck, Fabien", &
                         "   Iannuzzi, Marcella", &
                         "   Krack, Matthias", &
                         "TI Frenkel pair recombinations in UO2: Importance of explicit description of", &
                         "   polarizability in core-shell molecular dynamics simulations", &
                         "SO Phys. Rev. B", &
                         "VL 85", &
                         "IS 18", &
                         "SP 184103", &
                         "EP", &
                         "PD MAY 15", &
                         "PY 2012", &
                         "UR http://link.aps.org/doi/10.1103/PhysRevB.85.184103", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.85.184103")

      CALL add_reference(key=VandeVondele2012, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Borstnik, U", &
                         "   Hutter, J", &
                         "AF VandeVondele, Joost", &
                         "   Borstnik, Urban", &
                         "   Hutter, Juerg", &
                         "TI Linear Scaling Self-Consistent Field Calculations with Millions of Atoms", &
                         "   in the Condensed Phase", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD OCT", &
                         "PY 2012", &
                         "VL 8", &
                         "IS 10", &
                         "BP 3565", &
                         "EP 3573", &
                         "DI 10.1021/ct200897x", &
                         "UT WOS:000309505600014", &
                         "ER"), &
                         DOI="10.1021/ct200897x")

      CALL add_reference(key=Niklasson2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Niklasson, AMN", &
                         "   Tymczak, CJ", &
                         "   Challacombe, M", &
                         "AF Niklasson, AMN", &
                         "   Tymczak, CJ", &
                         "   Challacombe, M", &
                         "TI Trace resetting density matrix purification in O(N)", &
                         "   self-consistent-field theory", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD MAY 15", &
                         "PY 2003", &
                         "VL 118", &
                         "IS 19", &
                         "BP 8611", &
                         "EP 8620", &
                         "DI 10.1063/1.1559913", &
                         "UT WOS:000182603000010", &
                         "ER"), &
                         DOI="10.1063/1.1559913")

      CALL add_reference(key=Shao2003, ISI_record=s2a( &
                         "PT J", &
                         "AU Shao, Y", &
                         "   Saravanan, C", &
                         "   Head-Gordon, M", &
                         "   White, CA", &
                         "AF Shao, Y", &
                         "   Saravanan, C", &
                         "   Head-Gordon, M", &
                         "   White, CA", &
                         "TI Curvy steps for density matrix-based energy minimization: Application to", &
                         "   large-scale self-consistent-field calculations", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "RI White, Christopher/B-7390-2009", &
                         "SN 0021-9606", &
                         "PD APR 8", &
                         "PY 2003", &
                         "VL 118", &
                         "IS 14", &
                         "BP 6144", &
                         "EP 6151", &
                         "DI 10.1063/1.1558476", &
                         "UT WOS:000181834900004", &
                         "ER"), &
                         DOI="10.1063/1.1558476")

      CALL add_reference(key=VandeVondele2002, ISI_record=s2a( &
                         "PT J", &
                         "AU VandeVondele, J", &
                         "   Rothlisberger, U", &
                         "AF VandeVondele, J", &
                         "   Rothlisberger, U", &
                         "TI Canonical adiabatic free energy sampling (CAFES): A novel method for the", &
                         "   exploration of free energy surfaces", &
                         "SO JOURNAL OF PHYSICAL CHEMISTRY B", &
                         "SN 1520-6106", &
                         "PD JAN 10", &
                         "PY 2002", &
                         "VL 106", &
                         "IS 1", &
                         "BP 203", &
                         "EP 208", &
                         "DI 10.1021/jp013346k", &
                         "UT WOS:000173256700030", &
                         "ER"), &
                         DOI="10.1021/jp013346k")

      CALL add_reference(key=Dion2004, ISI_record=s2a( &
                         "PT J", &
                         "AU Dion, M", &
                         "   Rydberg, H", &
                         "   Schroder, E", &
                         "   Langreth, DC", &
                         "   Lundqvist, BI", &
                         "TI Van der Waals density functional for general geometries", &
                         "SO Phys. Rev. Lett.", &
                         "PD JUN 18", &
                         "PY 2004", &
                         "VL 92", &
                         "IS 24", &
                         "AR 246401", &
                         "UR http://prl.aps.org/abstract/PRL/v92/i24/e246401", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.92.246401")

      CALL add_reference(key=RomanPerez2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Roman-Perez, G", &
                         "   Soler, JM", &
                         "TI Efficient Implementation of a van der Waals Density Functional:", &
                         "   Application to Double-Wall Carbon Nanotubes", &
                         "SO Phys. Rev. Lett.", &
                         "PD AUG 28", &
                         "PY 2009", &
                         "VL 103", &
                         "IS 9", &
                         "AR 096102", &
                         "UR http://prl.aps.org/abstract/PRL/v103/i9/e096102", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.103.096102")

      CALL add_reference(key=DelBen2012, ISI_record=s2a( &
                         "PT J", &
                         "AU Del Ben, M", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Del Ben, Mauro", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Second-Order Moller-Plesset Perturbation Theory in the Condensed Phase:", &
                         "   An Efficient and Massively Parallel Gaussian and Plane Waves Approach", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD NOV", &
                         "PY 2012", &
                         "VL 8", &
                         "IS 11", &
                         "BP 4177", &
                         "EP 4188", &
                         "DI 10.1021/ct300531w", &
                         "UT WOS:000311191900019", &
                         "ER"), &
                         DOI="10.1021/ct300531w")

      CALL add_reference(key=Sabatini2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Sabatini, R", &
                         "   Gorni, T", &
                         "   de Gironcoli, S", &
                         "AF Sabatini, Riccardo", &
                         "   Gorni, Tommaso", &
                         "   de Gironcoli, Stefano", &
                         "TI Nonlocal van der Waals density functional made simple and efficient", &
                         "SO Phys. Rev. B", &
                         "PD JAN 15", &
                         "PY 2013", &
                         "VL 87", &
                         "DI 10.1103/PhysRevB.87.041108", &
                         "IS 4", &
                         "AR 041108(R)", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.87.041108")

      CALL add_reference(key=Walewski2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Walewski, L", &
                         "   Forbert, H", &
                         "   Marx, D", &
                         "AF Walewski, Lukasz", &
                         "   Forbert, Harald", &
                         "   Marx, Dominik", &
                         "TI Reactive path integral quantum simulations of molecules solvated in", &
                         "   superfluid helium", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "DE Cryochemistry; Superfluidity; Helium; Solvation; Path integrals", &
                         "NR 153", &
                         "SN 0010-4655", &
                         "PD MAR", &
                         "PY 2014", &
                         "VL 185", &
                         "IS 3", &
                         "BP 884", &
                         "EP 899", &
                         "DI 10.1016/j.cpc.2013.12.011", &
                         "UT WOS:000331919100022", &
                         "ER"), &
                         DOI="10.1016/j.cpc.2013.12.011")

      CALL add_reference(key=Delben2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Del Ben, M", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Del Ben, Mauro", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Electron Correlation in the Condensed Phase from a Resolution of", &
                         "   Identity Approach Based on the Gaussian and Plane Waves Scheme", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD JUN", &
                         "PY 2013", &
                         "VL 9", &
                         "IS 6", &
                         "BP 2654", &
                         "EP 2671", &
                         "DI 10.1021/ct4002202", &
                         "UT WOS:000320484500014", &
                         "ER"), &
                         DOI="10.1021/ct4002202")

      CALL add_reference(key=Kikuchi2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Kikuchi, Y", &
                         "   Imamura, Y", &
                         "   Nakai, H", &
                         "AF Kikuchi, Yasuaki", &
                         "   Imamura, Yutaka", &
                         "   Nakai, Hiromi", &
                         "TI One-Body Energy Decomposition Schemes Revisited: Assessment of", &
                         "   Mulliken-, Grid-, and Conventional Energy Density Analyses", &
                         "SO INTERNATIONAL JOURNAL OF QUANTUM CHEMISTRY", &
                         "TC 3", &
                         "Z9 3", &
                         "SN 0020-7608", &
                         "PD SEP", &
                         "PY 2009", &
                         "VL 109", &
                         "IS 11", &
                         "BP 2464", &
                         "EP 2473", &
                         "DI 10.1002/qua.22017", &
                         "ER"), &
                         DOI="10.1002/qua.22017")

      CALL add_reference(key=Putrino2000, ISI_record=s2a( &
                         "PT J", &
                         "AU Putrino, A", &
                         "   Sebastiani, D", &
                         "   Parrinello, M", &
                         "AF Putrino, Anna", &
                         "   Sebastiani, Daniel", &
                         "   Parrinello, Michele", &
                         "TI Generalized Variational Density Functional Perturbation Theory", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD NOV", &
                         "PY 2000", &
                         "VL 113", &
                         "IS 17", &
                         "BP 7102", &
                         "EP 7109", &
                         "DI 10.1063/1.1312830", &
                         "ER"), &
                         DOI="10.1063/1.1312830")

      CALL add_reference(key=Tran2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Tran, F", &
                         "   Hutter, J", &
                         "AF Tran, Fabien", &
                         "   Hutter, Juerg", &
                         "TI Nonlocal van der Waals functionals: The case of rare-gas dimers and", &
                         "   solids", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "TC 2", &
                         "Z9 2", &
                         "SN 0021-9606", &
                         "PD MAY 28", &
                         "PY 2013", &
                         "VL 138", &
                         "IS 20", &
                         "AR 204103", &
                         "DI 10.1063/1.4807332", &
                         "UT WOS:000320131100005", &
                         "ER"), &
                         DOI="10.1063/1.4807332")

      CALL add_reference(key=Putrino2002, ISI_record=s2a( &
                         "PT J", &
                         "AU Putrino, A", &
                         "   Parrinello, M", &
                         "AF Putrino, Anna", &
                         "   Parrinello, Michele", &
                         "TI Anharmonic Raman Spectra in High-Pressure Ice from Ab Initio Simulations", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "PD APR 29", &
                         "PY 2002", &
                         "VL 88", &
                         "IS 17", &
                         "AR 176401", &
                         "DI 10.1103/PhysRevLett.88.176401", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.88.176401")

      CALL add_reference(key=Sebastiani2001, ISI_record=s2a( &
                         "PT J", &
                         "AU Sebastiani, D", &
                         "   Parrinello, M", &
                         "AF Sebastiani, Daniel", &
                         "   Parrinello, Michele", &
                         "TI A New Ab Initio Approach for NMR Chemical Shifts in Periodic Systems", &
                         "SO THE JOURNAL OF PHYSICAL CHEMISTRY A", &
                         "PD MAR", &
                         "PY 2001", &
                         "VL 105", &
                         "IS 10", &
                         "BP 1951", &
                         "EP 1958", &
                         "DI 10.1021/jp002807j", &
                         "ER"), &
                         DOI="10.1021/jp002807j")

      CALL add_reference(key=Weber2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Weber, V", &
                         "   Iannuzzi, M", &
                         "   Giani, S", &
                         "   Hutter, J", &
                         "   Declerck, R", &
                         "   Waroduier, M", &
                         "AF Weber, Valery", &
                         "   Iannuzzi, Marcella", &
                         "   Giani, Samuele", &
                         "   Hutter, Juerg", &
                         "   Declerck, Reinout", &
                         "   Waroduier, Michel", &
                         "TI Magnetic Linear Response Properties Calculations with the ", &
                         "   Gaussian and Augmanted-Plane-Wave Method", &
                         "SO THE JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD JUL 7", &
                         "PY 2009", &
                         "VL 131", &
                         "IS 1", &
                         "AR 014106", &
                         "DI 10.1063/1.3156803", &
                         "ER"), &
                         DOI="10.1063/1.3156803")

      CALL add_reference(key=Golze2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Golze, D", &
                         "   Iannuzzi, M", &
                         "   Nguyen, M-T", &
                         "   Passerone, D", &
                         "   Hutter, J", &
                         "AF Golze, Dorothea", &
                         "   Iannuzzi, Marcella", &
                         "   Nguyen, Manh-Thuong", &
                         "   Passerone, Daniele", &
                         "   Hutter, Juerg", &
                         "TI Simulation of Adsorption Processes at Metallic Interfaces:", &
                         "   An Image Charge Augmented QM/MM Approach", &
                         "SO Journal of Chemical Theory and Computation", &
                         "SN 1549-9618", &
                         "PD NOV 12", &
                         "PY 2013", &
                         "VL 9", &
                         "IS 11", &
                         "BP 5086", &
                         "EP 5097", &
                         "DI 10.1021/ct400698y", &
                         "ER"), &
                         DOI="10.1021/ct400698y")

      CALL add_reference(key=Golze2015, ISI_record=s2a( &
                         "PT J", &
                         "AU Golze, D", &
                         "   Hutter, J", &
                         "   Iannuzzi, M", &
                         "AF Golze, Dorothea", &
                         "   Hutter, Juerg", &
                         "   Iannuzzi, Marcella", &
                         "TI Wetting of water on hexagonal boron nitride@Rh(111): a QM/MM model", &
                         "   based on atomic charges derived for nano-structured substrates", &
                         "SO Physical Chemistry Chemical Physics", &
                         "SN 1463-9076", &
                         "PD JUN 14", &
                         "PY 2015", &
                         "VL 17", &
                         "IS 22", &
                         "BP 14307", &
                         "EP 14316", &
                         "DI 10.1039/C4CP04638B", &
                         "ER"), &
                         DOI="10.1039/C4CP04638B")

      CALL add_reference(key=Fattebert2002, ISI_record=s2a( &
                         "PT J", &
                         "AU Fattebert, JL", &
                         "   Gygi, F", &
                         "AF Fattebert, Jean-Luc", &
                         "   Gygi, Francois", &
                         "TI Density functional theory for efficient ab initio molecular dynamics", &
                         "   simulations in solution", &
                         "SO J. Comput. Chem.", &
                         "VL 23", &
                         "IS 6", &
                         "SP 662", &
                         "EP 666", &
                         "PD MAR 18", &
                         "PY 2002", &
                         "UR http://dx.doi.org/10.1002/jcc.10069", &
                         "ER"), &
                         DOI="10.1002/jcc.10069")

      CALL add_reference(key=Andreussi2012, ISI_record=s2a( &
                         "PT J", &
                         "AU Andreussi, O", &
                         "   Dabo, I", &
                         "   Marzari, N", &
                         "AF Andreussi, Oliviero", &
                         "   Dabo, Ismaila", &
                         "   Marzari, Nicola", &
                         "TI Revised self-consistent continuum solvation in electronic-structure calculations", &
                         "SO J. Chem. Phys.", &
                         "VL 136", &
                         "IS 6", &
                         "AR 064102", &
                         "PD FEB 8", &
                         "PY 2012", &
                         "UR http://dx.doi.org/10.1063/1.3676407", &
                         "ER"), &
                         DOI="10.1063/1.3676407")

      CALL add_reference(key=Tuckerman1992, ISI_record=s2a( &
                         "PT J", &
                         "AU TUCKERMAN, M", &
                         "   BERNE, BJ", &
                         "   MARTYNA, GJ", &
                         "AF TUCKERMAN, M", &
                         "   BERNE, BJ", &
                         "   MARTYNA, GJ", &
                         "TI REVERSIBLE MULTIPLE TIME SCALE MOLECULAR-DYNAMICS", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD AUG 1", &
                         "PY 1992", &
                         "VL 97", &
                         "IS 3", &
                         "BP 1990", &
                         "EP 2001", &
                         "DI 10.1063/1.463137", &
                         "UT WOS:A1992JE89100044", &
                         "ER"), &
                         DOI="10.1063/1.463137")

      CALL add_reference(key=Goedecker2004, ISI_record=s2a( &
                         "PT J", &
                         "AU Goedecker, S", &
                         "AF Goedecker, Stefan", &
                         "TI Minima hopping: An efficient search method for the global minimum", &
                         "   of the potential energy surface of complex molecular systems", &
                         "SO Journal of Chemical Physics", &
                         "PY 2004", &
                         "VL 120", &
                         "IS 21", &
                         "BP 9911", &
                         "EP 9917", &
                         "DI 10.1063/1.1724816", &
                         "ER"), &
                         DOI="10.1063/1.1724816")

      CALL add_reference(key=Khaliullin2007, ISI_record=s2a( &
                         "PT J", &
                         "AU Khaliullin, RZ", &
                         "   Cobar, EA", &
                         "   Lochan, RC", &
                         "   Bell, AT", &
                         "   Head-Gordon, M", &
                         "AF Khaliullin, Rustam Z.", &
                         "   Cobar, Erika A.", &
                         "   Lochan, Rohini C.", &
                         "   Bell, Alexis T.", &
                         "   Head-Gordon, Martin", &
                         "TI Unravelling the origin of intermolecular interactions using absolutely localized molecular orbitals", &
                         "SO Journal of Physical Chemistry A", &
                         "PY 2007", &
                         "VL 111", &
                         "IS 36", &
                         "BP 8753", &
                         "EP 8765", &
                         "DI 10.1021/jp073685z", &
                         "ER"), &
                         DOI="10.1021/jp073685z")

      CALL add_reference(key=Khaliullin2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Khaliullin, RZ", &
                         "   Bell, AT", &
                         "   Head-Gordon, M", &
                         "AF Khaliullin, Rustam Z.", &
                         "   Bell, Alexis T.", &
                         "   Head-Gordon, Martin", &
                         "TI Analysis of charge transfer effects in molecular complexes "// &
                         "based on absolutely localized molecular orbitals", &
                         "SO Journal of Chemical Physics", &
                         "SN 0021-9606", &
                         "PY 2008", &
                         "VL 128", &
                         "IS 18", &
                         "AR 184112", &
                         "DI 10.1063/1.2912041", &
                         "ER"), &
                         DOI="10.1063/1.2912041")

      CALL add_reference(key=Khaliullin2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Khaliullin, RZ", &
                         "   VandeVondele, J", &
                         "   Hutter, J", &
                         "AF Khaliullin, Rustam Z.", &
                         "   VandeVondele, Joost", &
                         "   Hutter, Juerg", &
                         "TI Efficient Linear-Scaling Density Functional Theory for Molecular Systems", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD OCT", &
                         "PY 2013", &
                         "VL 9", &
                         "IS 10", &
                         "BP 4421", &
                         "EP 4427", &
                         "DI 10.1021/ct400595k", &
                         "UT WOS:000326355100014", &
                         "ER"), &
                         DOI="10.1021/ct400595k")

      CALL add_reference(key=Hutter2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Hutter, J", &
                         "   Iannuzzi, M", &
                         "   Schiffmann, F", &
                         "   VandeVondele, J", &
                         "AF Hutter, Juerg", &
                         "   Iannuzzi, Marcella", &
                         "   Schiffmann, Florian", &
                         "   VandeVondele, Joost", &
                         "TI CP2K: atomistic simulations of condensed matter systems", &
                         "SO WILEY INTERDISCIPLINARY REVIEWS-COMPUTATIONAL MOLECULAR SCIENCE", &
                         "SN 1759-0876", &
                         "PD JAN", &
                         "PY 2014", &
                         "VL 4", &
                         "IS 1", &
                         "BP 15", &
                         "EP 25", &
                         "DI 10.1002/wcms.1159", &
                         "UT WOS:000328646100002", &
                         "ER"), &
                         DOI="10.1002/wcms.1159")

      CALL add_reference(key=Kantorovich2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Kantorovich, L", &
                         "AF Kantorovich, Lev", &
                         "TI Generalized Langevin equation for solids. I. ", &
                         "   Rigorous derivation and main properties", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD SEP", &
                         "PY 2008", &
                         "VL 78", &
                         "IS 9", &
                         "AR 094304", &
                         "DI 10.1103/PhysRevB.78.094304", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.78.094304")

      CALL add_reference(key=Kantorovich2008a, ISI_record=s2a( &
                         "PT J", &
                         "AU Kantorovich, L", &
                         "   Rompotis, N", &
                         "AF Kantorovich, Lev", &
                         "   Rompotis, N", &
                         "TI Generalized Langevin equation for solids. II. ", &
                         "   Stochastic boundary conditions for nonequilibrium molecular dynamics simulations", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD SEP", &
                         "PY 2008", &
                         "VL 78", &
                         "IS 9", &
                         "AR 094305", &
                         "DI 10.1103/PhysRevB.78.094305", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.78.094305")

      CALL add_reference(key=Niklasson2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Rubensson, E.", &
                         "AU Niklasson, A.", &
                         "TI Interior Eigenvalues from Density Matrix Expansions in Quantum", &
                         "    Mechanical Molecular Dynamics", &
                         "SO SIAM Journal on Scientific Computing", &
                         "PD MAR", &
                         "PY 2014", &
                         "VL 36", &
                         "IS 2", &
                         "BP B147", &
                         "EP B170", &
                         "DI 10.1137/130911585", &
                         "ER"), &
                         DOI="10.1137/130911585")

      CALL add_reference(key=Borstnik2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Borstnik, U", &
                         "   VandeVondele, J", &
                         "   Weber, V", &
                         "   Hutter, J", &
                         "AF Borstnik, Urban", &
                         "   VandeVondele, Joost", &
                         "   Weber, Valery", &
                         "   Hutter, Juerg", &
                         "TI Sparse matrix multiplication: The distributed block-compressed sparse", &
                         "   row library", &
                         "SO PARALLEL COMPUTING", &
                         "SN 0167-8191", &
                         "EI 1872-7336", &
                         "PD MAY", &
                         "PY 2014", &
                         "VL 40", &
                         "IS 5-6", &
                         "BP 47", &
                         "EP 58", &
                         "DI 10.1016/j.parco.2014.03.012", &
                         "UT WOS:000338614300001", &
                         "ER"), &
                         DOI="10.1016/j.parco.2014.03.012")

      CALL add_reference(key=Rayson2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Rayson, M. J.", &
                         "   Briddon, P. R.", &
                         "AF Rayson, Mark", &
                         "   Briddon, Patrick", &
                         "TI Highly efficient method for Kohn-Sham density functional", &
                         "   calculations of 500-10 000 atom systems ", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 1098-0121", &
                         "PD NOV", &
                         "PY 2009", &
                         "VL 80", &
                         "IS 20", &
                         "AR 205104", &
                         "DI 10.1103/PhysRevB.80.205104", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.80.205104")

      CALL add_reference(key=Merlot2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Merlot, P", &
                         "   Izsak, R", &
                         "   Borgo, A", &
                         "   Kjaergaard, T", &
                         "   Helgaker, T", &
                         "   Reine, S", &
                         "AF Merlot, Patrick", &
                         "   Izsak, Robert", &
                         "   Borgo, Alex", &
                         "   Kjaergaard, Thomas", &
                         "   Helgaker, Trygve", &
                         "   Reine, Simen", &
                         "TI Charge-constrained auxiliary-density-matrix methods for the", &
                         "   Hartree-Fock exchange contribution", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "PD SEP", &
                         "PY 2014", &
                         "VL 141", &
                         "IS 9", &
                         "AR 094104", &
                         "DI /10.1063/1.4894267", &
                         "ER"), &
                         DOI="10.1063/1.4894267")

      CALL add_reference(key=Lin2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Lin, L", &
                         "   Lu, JF", &
                         "   Ying, LX", &
                         "   Car, R", &
                         "   E, WN", &
                         "AF Lin, Lin", &
                         "   Lu, Jianfeng", &
                         "   Ying, Lexing", &
                         "   Car, Roberto", &
                         "   E, Weinan", &
                         "TI FAST ALGORITHM FOR EXTRACTING THE DIAGONAL OF THE INVERSE MATRIX WITH", &
                         "   APPLICATION TO THE ELECTRONIC STRUCTURE ANALYSIS OF METALLIC SYSTEMS", &
                         "SO COMMUNICATIONS IN MATHEMATICAL SCIENCES", &
                         "VL 7", &
                         "SN 1539-6746", &
                         "PD SEP", &
                         "PY 2009", &
                         "RI Lin, Lin/I-2726-2012", &
                         "OI Lin, Lin/0000-0001-7738-5947", &
                         "TC 20", &
                         "ZB 0", &
                         "Z8 3", &
                         "ZS 0", &
                         "Z9 21", &
                         "VL 7", &
                         "IS 3", &
                         "BP 755", &
                         "EP 777", &
                         "UT WOS:000271568500012", &
                         "ER"), &
                         DOI="10.4310/CMS.2009.v7.n3.a12")

      CALL add_reference(key=Lin2013, ISI_record=s2a( &
                         "PT J", &
                         "AU Lin, Lin", &
                         "   Chen, Mohan", &
                         "   Yang, Chao", &
                         "   He, Lixin", &
                         "TI Accelerating atomic orbital-based electronic structure calculation via", &
                         "   pole expansion and selected inversion", &
                         "SO JOURNAL OF PHYSICS-CONDENSED MATTER", &
                         "VL 25", &
                         "IS 29", &
                         "AR 295501", &
                         "PD JUL 24 2013", &
                         "PY 2013", &
                         "TC 3", &
                         "ZB 1", &
                         "Z8 0", &
                         "ZS 0", &
                         "Z9 3", &
                         "SN 0953-8984", &
                         "UT WOS:000321219200006", &
                         "PM 23803312", &
                         "ER"), &
                         DOI="10.1088/0953-8984/25/29/295501")

      CALL add_reference(key=DelBen2015, ISI_record=s2a( &
                         "PT J", &
                         "AU Del Ben, M", &
                         "   Schuett, O", &
                         "   Wentz, T", &
                         "   Messmer, P", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Del Ben, Mauro", &
                         "   Schuett, Ole", &
                         "   Wentz, Tim", &
                         "   Messmer, Peter", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Enabling simulation at the fifth rung of DFT: Large scale RPA", &
                         "   calculations with excellent time to solution", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "SN 0010-4655", &
                         "EI 1879-2944", &
                         "PD FEB", &
                         "PY 2015", &
                         "VL 187", &
                         "BP 120", &
                         "EP 129", &
                         "DI 10.1016/j.cpc.2014.10.021", &
                         "UT WOS:000346954200014", &
                         "ER"), &
                         DOI="10.1016/j.cpc.2014.10.021")

      CALL add_reference(key=Souza2002, ISI_record=s2a( &
                         "PT J", &
                         "AU Souza, I", &
                         "   Iniguez, J", &
                         "   Vanderbilt, D", &
                         "TI First-principles approach to insulators in finite electric fields", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "IS 11", &
                         "AR 117602", &
                         "PD SEP", &
                         "PY 2002", &
                         "VL 89", &
                         "DI 10.1103/PhysRevLett.89.117602", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.89.117602")

      CALL add_reference(key=Umari2002, ISI_record=s2a( &
                         "PT J", &
                         "AU Umari, P", &
                         "   Pasquarello, A", &
                         "TI Ab initio molecular dynamics in a finite homogeneous electric field", &
                         "SO PHYSICAL REVIEW LETTERS", &
                         "SN 0031-9007", &
                         "IS 15", &
                         "AR 157602", &
                         "PD OCT", &
                         "PY 2002", &
                         "VL 89", &
                         "DI 10.1103/PhysRevLett.89.157602", &
                         "ER"), &
                         DOI="10.1103/PhysRevLett.89.157602")

      CALL add_reference(key=Stengel2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Stengel, Massimiliano", &
                         "   Spaldin, Nicola A.", &
                         "   Vanderbilt, David", &
                         "TI Ab initio molecular dynamics in a finite homogeneous electric field", &
                         "TI Electric displacement as the fundamental variable in", &
                         "   electronic-structure calculations", &
                         "SO NATURE PHYSICS", &
                         "SN 1745-2473", &
                         "IS 4", &
                         "PD APR", &
                         "PY 2009", &
                         "VL 5", &
                         "BP 304", &
                         "EP 308", &
                         "DI 10.1038/NPHYS1185", &
                         "ER"), &
                         DOI="10.1038/NPHYS1185")

      CALL add_reference(key=Luber2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Luber, S", &
                         "   Iannuzzi, M", &
                         "   Hutter, J", &
                         "AF Luber, Sandra", &
                         "   Iannuzzi, Marcella", &
                         "   Hutter, Juerg", &
                         "TI Raman spectra from ab initio molecular dynamics and its application to", &
                         "   liquid S-methyloxirane", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "RI Hutter, Juerg/E-9244-2011", &
                         "SN 0021-9606", &
                         "EI 1089-7690", &
                         "PD SEP 7", &
                         "PY 2014", &
                         "VL 141", &
                         "IS 9", &
                         "AR 094503", &
                         "DI 10.1063/1.4894425", &
                         "UT WOS:000342207400024", &
                         "PM 25194377", &
                         "ER"), &
                         DOI="10.1063/1.4894425")

      CALL add_reference(key=Berghold2011, ISI_record=s2a( &
                         "PT J", &
                         "AU Berghold, G", &
                         "   Parrinello, M", &
                         "   Hutter, J", &
                         "AF Berghold, Gerd", &
                         "   Parrinello, Michele", &
                         "   Hutter, Juerg", &
                         "TI Polarized atomic orbitals for linear scaling methods", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "RI Hutter, Juerg/E-9244-2011", &
                         "SN 0021-9606", &
                         "PD FEB 1", &
                         "PY 2002", &
                         "VL 116", &
                         "IS 5", &
                         "BP 1800", &
                         "EP 1810", &
                         "DI 10.1063/1.1431270", &
                         "UT WOS:000173418600007", &
                         "ER"), &
                         DOI="10.1063/1.1431270")

      CALL add_reference(key=DelBen2015b, ISI_record=s2a( &
                         "PT J", &
                         "AU Del Ben, M", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Del Ben, Mauro", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Forces and stress in second order Moller-Plesset perturbation theory for", &
                         "   condensed phase systems within the resolution-of-identity Gaussian and", &
                         "   plane waves approach", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "EI 1089-7690", &
                         "PD SEP 14", &
                         "PY 2015", &
                         "VL 143", &
                         "IS 10", &
                         "AR 102803", &
                         "DI 10.1063/1.4919238", &
                         "UT WOS:000361572900008", &
                         "PM 26373996", &
                         "ER"), &
                         DOI="10.1063/1.4919238")

      CALL add_reference(key=Campana2009, ISI_record=s2a( &
                         "PT J", &
                         "AU Campana, C", &
                         "   Mussard, B", &
                         "   Woo, T K", &
                         "AF Campana, Carlos", &
                         "   Mussard, Bastien", &
                         "   Woo, Tom K.", &
                         "TI Electrostatic Potential Derived Atomic Charges for ", &
                         "   Periodic Systems Using a Modified Error Functional", &
                         "SO JOURNAL OF CHEMICAL THEORY AND COMPUTATION", &
                         "SN 1549-9618", &
                         "PD OCT 13", &
                         "PY 2009", &
                         "VL 5", &
                         "IS 10", &
                         "AR 2866", &
                         "DI 10.1021/ct9003405", &
                         "ER"), &
                         DOI="10.1021/ct9003405")

      CALL add_reference(key=Schiffmann2015, ISI_record=s2a( &
                         "PT J", &
                         "AU Schiffmann, F", &
                         "   VandeVondele, J", &
                         "AF Schiffmann, Florian", &
                         "   VandeVondele, Joost", &
                         "TI Efficient preconditioning of the electronic structure problem in large", &
                         "   scale ab initio molecular dynamics simulations", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "RI VandeVondele, Joost/L-6420-2013", &
                         "OI VandeVondele, Joost/0000-0002-0902-5111", &
                         "SN 0021-9606", &
                         "EI 1089-7690", &
                         "PD JUN 28", &
                         "PY 2015", &
                         "VL 142", &
                         "IS 24", &
                         "AR 244117", &
                         "DI 10.1063/1.4922988", &
                         "UT WOS:000357615100021", &
                         "PM 26133420", &
                         "ER"), &
                         DOI="10.1063/1.4922988")

      CALL add_reference(key=Bruck2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Bruck, S", &
                         "   Calderara, M", &
                         "   Bani-Hashemian, MH", &
                         "   VandeVondele, J", &
                         "   Luisier, M", &
                         "AF Brueck, S.", &
                         "   Calderara, M.", &
                         "   Bani-Hashemian, M. H.", &
                         "   VandeVondele, J.", &
                         "   Luisier, M.", &
                         "GP IEEE", &
                         "TI Towards ab-initio simulations of nanowire field-effect transistors", &
                         "SO 2014 INTERNATIONAL WORKSHOP ON COMPUTATIONAL ELECTRONICS (IWCE)", &
                         "CT International Workshop on Computational Electronics (IWCE)", &
                         "CY JUN 03-06, 2014", &
                         "CL Paris, FRANCE", &
                         "SP Inst Electronique Fondamentale, Ctr Natl Rech Sci, Univ Paris Sud,", &
                         "   Lab Excellence Nanosciences & Nanotechnologies, Nano K, Quantum Wise, IEEE France Sect", &
                         "RI VandeVondele, Joost/L-6420-2013", &
                         "OI VandeVondele, Joost/0000-0002-0902-5111", &
                         "BN 978-1-4799-5433-9", &
                         "PY 2014", &
                         "UT WOS:000345736700024", &
                         "ER"), &
                         DOI="10.1109/IWCE.2014.6865831")

      CALL add_reference(key=Rappe1992, ISI_record=s2a( &
                         "TY  JOUR", &
                         "AU Rappe, AK", &
                         "   Casewit, CJ", &
                         "   Colwell, KS", &
                         "   Goddard, WA", &
                         "   Skiff, WM", &
                         "AF Rappe, A. K.", &
                         "   Casewit, C. J.", &
                         "   Colwell, K. S.", &
                         "   Goddard, W. A.", &
                         "   Skiff, W. M.", &
                         "TI UFF, a full periodic table force field for molecular ", &
                         "   mechanics and molecular dynamics simulations", &
                         "SO JOURNAL OF THE AMERICAN CHEMICAL SOCIETY", &
                         "SN 0002-7863", &
                         "VL 114", &
                         "IS 25", &
                         "BP 10024", &
                         "EP 10035", &
                         "PD DEC 1", &
                         "PY 1992", &
                         "DI 10.1021/ja00051a040", &
                         "ER"), &
                         DOI="10.1021/ja00051a040")

      CALL add_reference(key=Monkhorst1976, ISI_record=s2a( &
                         "PT J", &
                         "AU Monkhorst, HJ", &
                         "   Pack, JD", &
                         "AF Monkhorst, HJ", &
                         "   Pack, JD", &
                         "TI Special points for Brillouin-zone integrations", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "PY 1976", &
                         "VL 13", &
                         "IS 12", &
                         "BP 5188", &
                         "EP 5192", &
                         "DI 10.1103/PhysRevB.13.5188", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.13.5188")

      CALL add_reference(key=MacDonald1978, ISI_record=s2a( &
                         "PT J", &
                         "AU MacDonald, AH", &
                         "AF MacDonald, AH", &
                         "TI Comment on special points for Brillouin-zone integrations", &
                         "SO PHYSICAL REVIEW B", &
                         "SN 0163-1829", &
                         "J9 PHYS REV B", &
                         "JI Phys. Rev. B", &
                         "PY 1978", &
                         "VL 18", &
                         "IS 10", &
                         "BP 5897", &
                         "EP 5899", &
                         "DI 10.1103/PhysRevB.18.5897", &
                         "ER"), &
                         DOI="10.1103/PhysRevB.18.5897")

      CALL add_reference(key=Gilbert2008, ISI_record=s2a( &
                         "PT J", &
                         "AU Gilbert, ATB", &
                         "   Besley, NA", &
                         "   Gill, PMW", &
                         "AF Gilbert, Andrew T. B.", &
                         "   Besley, Nicholas A.", &
                         "   Gill, Peter M. W.", &
                         "TI Self-consistent field calculations of excited states", &
                         "   using the maximum overlap method (MOM)", &
                         "SO THE JOURNAL OF PHYSICAL CHEMISTRY A", &
                         "PD AUG 26", &
                         "PY 2008", &
                         "VL 112", &
                         "BP 13164", &
                         "EP 13171", &
                         "DI 10.1021/jp801738f", &
                         "ER"), &
                         DOI="10.1021/jp801738f")

      CALL add_reference(key=Schonherr2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Schonherr, M", &
                         "   Slater, B", &
                         "   Hutter, J", &
                         "   VandeVondele, J", &
                         "AF Schoenherr, Mandes", &
                         "   Slater, Ben", &
                         "   Hutter, Juerg", &
                         "   VandeVondele, Joost", &
                         "TI Dielectric Properties of Water Ice, the Ice Ih/XI Phase Transition, and", &
                         "   an Assessment of Density Functional Theory", &
                         "SO JOURNAL OF PHYSICAL CHEMISTRY B", &
                         "RI Hutter, Juerg/E-9244-2011; VandeVondele, Joost/L-6420-2013", &
                         "OI VandeVondele, Joost/0000-0002-0902-5111", &
                         "SN 1520-6106", &
                         "PD JAN 16", &
                         "PY 2014", &
                         "VL 118", &
                         "IS 2", &
                         "BP 590", &
                         "EP 596", &
                         "DI 10.1021/jp4103355", &
                         "UT WOS:000330017900021", &
                         "PM 24392971", &
                         "ER"), &
                         DOI="10.1021/jp4103355")

      CALL add_reference(key=Ceriotti2014, ISI_record=s2a( &
                         "PT J", &
                         "AU Ceriotti, M", &
                         "   More, J", &
                         "   Manolopoulos, DE", &
                         "AF Ceriotti, Michele", &
                         "   More, Joshua", &
                         "   Manolopoulos, David E.", &
                         "TI i-PI: A Python interface for ab initio path integral molecular dynamics", &
                         "   simulations", &
                         "SO COMPUTER PHYSICS COMMUNICATIONS", &
                         "RI Ceriotti, Michele/C-2393-2009", &
                         "OI Ceriotti, Michele/0000-0003-2571-2832", &
                         "SN 0010-4655", &
                         "EI 1879-2944", &
                         "PD MAR", &
                         "PY 2014", &
                         "VL 185", &
                         "IS 3", &
                         "BP 1019", &
                         "EP 1026", &
                         "DI 10.1016/j.cpc.2013.10.027", &
                         "UT WOS:000331919100033", &
                         "ER"), &
                         DOI="10.1016/j.cpc.2013.10.027")

      CALL add_reference(key=BaniHashemian2016, ISI_record=s2a( &
                         "PT J", &
                         "AU Bani-Hashemian, MH", &
                         "   Bruck, S", &
                         "   Luisier, M", &
                         "   VandeVondele, J", &
                         "AF Bani-Hashemian, Mohammad Hossein", &
                         "   Brueck, Sascha", &
                         "   Luisier, Mathieu", &
                         "   VandeVondele, Joost", &
                         "TI A generalized Poisson solver for first-principles device simulations", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "EI 1089-7690", &
                         "PD JAN 28", &
                         "PY 2016", &
                         "VL 144", &
                         "IS 4", &
                         "AR 044113", &
                         "DI 10.1063/1.4940796", &
                         "UT WOS:000369893100016", &
                         "PM 26827208", &
                         "ER"), &
                         DOI="10.1063/1.4940796")

      CALL add_reference(key=Kapil2016, ISI_record=s2a( &
                         "PT J", &
                         "AU Kapil, V", &
                         "   VandeVondele, J", &
                         "   Ceriotti, M", &
                         "AF Kapil, V.", &
                         "   VandeVondele, J.", &
                         "   Ceriotti, M.", &
                         "TI Accurate molecular dynamics and nuclear quantum effects at low cost by", &
                         "   multiple steps in real and imaginary time: Using density functional", &
                         "   theory to accelerate wavefunction methods", &
                         "SO JOURNAL OF CHEMICAL PHYSICS", &
                         "SN 0021-9606", &
                         "EI 1089-7690", &
                         "PD FEB 7", &
                         "PY 2016", &
                         "VL 144", &
                         "IS 5", &
                         "AR 054111", &
                         "DI 10.1063/1.4941091", &
                         "UT WOS:000369893900013", &
                         "PM 26851912", &
                         "ER"), &
                         DOI="10.1063/1.4941091")

      CALL add_reference(key=Heinzmann1976, ISI_record=s2a( &
                         "AU Heinzmann, R", &
                         "   Ahlrichs, R", &
                         "TI Population analysis based on occupation numbers of modified atomic orbitals (MAOs)", &
                         "SO Theoret. Chim. Acta", &
                         "PY 1976", &
                         "VL 42", &
                         "IS 1", &
                         "BP 33", &
                         "EP 45", &
                         "ER"), &
                         DOI="10.1007/BF00548289")

      CALL add_reference(key=Ehrhardt1985, ISI_record=s2a( &
                         "AU Ehrhardt, C", &
                         "   Ahlrichs, R", &
                         "TI Population analysis based on occupation numbers II. ", &
                         "   Relationship between shared electron numbers and bond energies and", &
                         "   characterization of hypervalent contributions", &
                         "SO Theoret. Chim. Acta", &
                         "PY 1985", &
                         "VL 68", &
                         "IS 3", &
                         "BP 231", &
                         "EP 245", &
                         "ER"), &
                         DOI="10.1007/BF00526774")

   END SUBROUTINE add_all_references

END MODULE bibliography
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR interface in CP2K
!> \author  VE
!> \date    2010
!> \version 0.1
!>
!> <b>Modification history:</b>
!> - Created 2010
! **************************************************************************************************
MODULE cp_dbcsr_interface
   USE bibliography,                    ONLY: Borstnik2014,&
                                              cite_reference
   USE dbcsr_api,                       ONLY: &
        accdrv_avoid_after_busy, accdrv_binning_binsize, accdrv_binning_nbins, &
        accdrv_do_inhomogenous, accdrv_min_flop_process, accdrv_min_flop_sort, &
        accdrv_posterior_buffers, accdrv_posterior_streams, accdrv_priority_buffers, &
        accdrv_priority_streams, add_work_coordinate, array_data, convert_csr_to_dbcsr, &
        convert_dbcsr_to_csr, convert_offsets_to_sizes, convert_sizes_to_offsets, &
        cp_dbcsr_iterator=>dbcsr_iterator, &
        cp_distribution_get_num_images=>dbcsr_distribution_get_num_images_1d, &
        create_bl_distribution, csr_create_from_dbcsr, csr_create_new, csr_create_template, &
        csr_dbcsr_blkrow_dist, csr_destroy, csr_eqrow_ceil_dist, csr_eqrow_floor_dist, csr_p_type, &
        csr_print_sparsity, csr_type, dbcsr_add, dbcsr_add_on_diag, dbcsr_arnoldi_data, &
        dbcsr_arnoldi_ev, dbcsr_arnoldi_extremal, dbcsr_binary_read, dbcsr_binary_write, &
        dbcsr_checksum, dbcsr_clear_mempools, dbcsr_complete_redistribute, dbcsr_copy, &
        dbcsr_copy_into_existing, dbcsr_create, dbcsr_create_dist_block_cyclic, &
        dbcsr_create_dist_r_unrot, dbcsr_data_clear_pointer, dbcsr_data_init, dbcsr_data_new, &
        dbcsr_data_obj, dbcsr_data_release, dbcsr_desymmetrize_deep, dbcsr_distribute, &
        dbcsr_distribution, dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_has_col_clusters, dbcsr_distribution_has_row_clusters, &
        dbcsr_distribution_has_threads, dbcsr_distribution_init, dbcsr_distribution_local_cols, &
        dbcsr_distribution_local_rows, dbcsr_distribution_mp, dbcsr_distribution_ncols, &
        dbcsr_distribution_new, dbcsr_distribution_nlocal_cols, dbcsr_distribution_nlocal_rows, &
        dbcsr_distribution_nrows, dbcsr_distribution_obj, dbcsr_distribution_release, &
        dbcsr_distribution_row_clusters, dbcsr_distribution_row_dist, dbcsr_filter, &
        dbcsr_finalize, dbcsr_finalize_lib, dbcsr_frobenius_norm, dbcsr_func_artanh, &
        dbcsr_func_dtanh, dbcsr_func_inverse, dbcsr_func_tanh, dbcsr_function_of_elements, &
        dbcsr_gershgorin_norm, dbcsr_get_block_diag, dbcsr_get_block_p, &
        dbcsr_get_conf_avg_elements_images, dbcsr_get_conf_comm_thread_load, &
        dbcsr_get_conf_max_ele_block, dbcsr_get_conf_mm_driver, dbcsr_get_conf_mm_stacksize, &
        dbcsr_get_conf_mpi_mem, dbcsr_get_conf_nstacks, dbcsr_get_conf_randmat_seed, &
        dbcsr_get_conf_use_comm_thread, dbcsr_get_conf_use_mpi_exp, &
        dbcsr_get_conf_use_mpi_filtering, dbcsr_get_data_p, dbcsr_get_data_size, &
        dbcsr_get_data_type, dbcsr_get_diag, dbcsr_get_info, dbcsr_get_matrix_type, &
        dbcsr_get_num_blocks, dbcsr_get_occupation, dbcsr_get_stored_coordinates, &
        dbcsr_hadamard_product, dbcsr_has_symmetry, dbcsr_init, dbcsr_init_lib, dbcsr_init_random, &
        dbcsr_iterator, dbcsr_iterator_blocks_left, dbcsr_iterator_next_block, &
        dbcsr_iterator_start, dbcsr_iterator_stop, dbcsr_maxabs, dbcsr_mp_get_process, &
        dbcsr_mp_group, dbcsr_mp_hold, dbcsr_mp_mynode, dbcsr_mp_new, dbcsr_mp_npcols, &
        dbcsr_mp_nprows, dbcsr_mp_numnodes, dbcsr_mp_obj, dbcsr_mp_release, dbcsr_multiply, &
        dbcsr_name, dbcsr_nblkcols_total, dbcsr_nblkrows_total, dbcsr_new_transposed, &
        dbcsr_nfullcols_total, dbcsr_nfullrows_total, dbcsr_no_transpose, dbcsr_norm, &
        dbcsr_norm_column, dbcsr_norm_frobenius, dbcsr_norm_maxabsnorm, dbcsr_obj, &
        dbcsr_obj_type_p, dbcsr_print, dbcsr_print_block_sum, dbcsr_put_block, dbcsr_release, &
        dbcsr_replicate_all, dbcsr_reserve_all_blocks, dbcsr_reserve_block2d, &
        dbcsr_reserve_blocks, dbcsr_reserve_diag_blocks, dbcsr_run_tests, dbcsr_scalar, &
        dbcsr_scalar_fill_all, dbcsr_scalar_get_type, dbcsr_scalar_get_value, &
        dbcsr_scalar_set_type, dbcsr_scalar_type, dbcsr_scalar_zero, dbcsr_scale, &
        dbcsr_scale_by_vector, dbcsr_set, dbcsr_set_conf_avg_elements_images, &
        dbcsr_set_conf_comm_thread_load, dbcsr_set_conf_max_ele_block, dbcsr_set_conf_mm_driver, &
        dbcsr_set_conf_mm_stacksize, dbcsr_set_conf_nstacks, dbcsr_set_conf_randmat_seed, &
        dbcsr_set_conf_use_comm_thread, dbcsr_set_conf_use_mpi_exp, &
        dbcsr_set_conf_use_mpi_filtering, dbcsr_set_diag, dbcsr_setname, dbcsr_sum_replicated, &
        dbcsr_test_arnoldi, dbcsr_test_binary_io, dbcsr_test_mm, dbcsr_trace, dbcsr_triu, &
        dbcsr_type_antisymmetric, dbcsr_type_complex_4, dbcsr_type_complex_8, &
        dbcsr_type_complex_default, dbcsr_type_no_symmetry, dbcsr_type_real_4, dbcsr_type_real_8, &
        dbcsr_type_real_default, dbcsr_type_symmetric, dbcsr_valid_index, dbcsr_verify_matrix, &
        dbcsr_wm_use_mutable, dbcsr_work_create, deallocate_arnoldi_data, get_selected_ritz_val, &
        get_selected_ritz_vector, has_acc, has_mpi, heap_fill, heap_get_first, heap_new, &
        heap_release, heap_reset_first, heap_t, mm_driver_auto, mm_driver_blas, mm_driver_matmul, &
        mm_driver_smm, mm_driver_xsmm, mm_name_auto, mm_name_blas, mm_name_matmul, mm_name_smm, &
        mm_name_xsmm, multrec_limit, set_arnoldi_initial_vector, setup_arnoldi_data, swap
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE machine,                         ONLY: default_output_unit
   USE string_utilities,                ONLY: uppercase

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_dbcsr_interface'

   !
   ! Interface to libdbcsr (contains cp2k timing and error)
   PUBLIC :: dbcsr_get_conf_comm_thread_load, &
             dbcsr_get_conf_avg_elements_images, &
             dbcsr_get_conf_mm_stacksize, dbcsr_get_conf_nstacks, &
             dbcsr_set_conf_mm_stacksize, &
             dbcsr_get_conf_use_mpi_exp, dbcsr_get_conf_use_mpi_filtering
   PUBLIC :: dbcsr_set_conf_comm_thread_load, dbcsr_set_conf_max_ele_block, dbcsr_set_conf_mm_driver, &
             dbcsr_set_conf_avg_elements_images, dbcsr_set_conf_nstacks, dbcsr_set_conf_use_comm_thread, &
             has_mpi, dbcsr_set_conf_use_mpi_exp, dbcsr_set_conf_use_mpi_filtering
   PUBLIC :: has_acc, dbcsr_create_dist_block_cyclic, dbcsr_create_dist_r_unrot, dbcsr_distribution_local_cols, &
             dbcsr_distribution_nlocal_rows, dbcsr_distribution_nrows, dbcsr_distribution_local_rows, &
             dbcsr_distribution_ncols, dbcsr_distribution_nlocal_cols, &
             dbcsr_mp_hold, cp_distribution_get_num_images
   PUBLIC :: dbcsr_type_no_symmetry, dbcsr_type_symmetric, dbcsr_type_antisymmetric
   PUBLIC :: dbcsr_type_complex_8, dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_complex_4
   PUBLIC :: dbcsr_type_complex_default, dbcsr_type_real_default
   PUBLIC :: convert_offsets_to_sizes, dbcsr_distribution_obj, dbcsr_distribution_row_dist
   PUBLIC :: dbcsr_distribution_row_clusters, dbcsr_distribution_col_clusters, &
             dbcsr_distribution_has_row_clusters, dbcsr_distribution_has_col_clusters
   PUBLIC :: dbcsr_get_conf_use_comm_thread, dbcsr_data_clear_pointer, dbcsr_data_new, dbcsr_data_init
   PUBLIC :: dbcsr_mp_group, dbcsr_mp_mynode, dbcsr_mp_numnodes, dbcsr_mp_nprows, dbcsr_mp_npcols, &
             dbcsr_mp_get_process
   PUBLIC :: dbcsr_iterator_stop, dbcsr_data_obj, dbcsr_iterator, dbcsr_data_release
   PUBLIC :: dbcsr_iterator_blocks_left, dbcsr_iterator_next_block
   PUBLIC :: dbcsr_mp_obj, dbcsr_norm_frobenius
   PUBLIC :: dbcsr_run_tests, dbcsr_test_mm, dbcsr_test_arnoldi, dbcsr_norm_maxabsnorm, &
             dbcsr_norm_column, dbcsr_distribution_has_threads, dbcsr_test_binary_io
   PUBLIC :: dbcsr_distribution_col_dist, dbcsr_distribution_init, dbcsr_distribution_release
   PUBLIC :: dbcsr_func_artanh, dbcsr_func_dtanh, dbcsr_func_inverse, dbcsr_func_tanh
   PUBLIC :: dbcsr_distribution_mp
   PUBLIC :: convert_sizes_to_offsets
   PUBLIC :: cp_dbcsr_iterator
   PUBLIC :: cp_dbcsr_type
   PUBLIC :: cp_dbcsr_p_type
   PUBLIC :: cp_dbcsr_init_lib, cp_dbcsr_finalize_lib
   PUBLIC :: cp_dbcsr_set
   PUBLIC :: cp_dbcsr_add
   PUBLIC :: cp_dbcsr_scale
   PUBLIC :: cp_dbcsr_scale_by_vector
   PUBLIC :: cp_dbcsr_hadamard_product
   PUBLIC :: cp_dbcsr_transposed
   PUBLIC :: cp_dbcsr_multiply
   PUBLIC :: cp_dbcsr_copy
   PUBLIC :: cp_dbcsr_copy_into_existing
   PUBLIC :: cp_dbcsr_desymmetrize
   PUBLIC :: cp_dbcsr_add_on_diag
   PUBLIC :: cp_dbcsr_get_block_diag
   PUBLIC :: cp_dbcsr_set_diag
   PUBLIC :: cp_dbcsr_get_diag
   PUBLIC :: cp_dbcsr_filter
   PUBLIC :: cp_dbcsr_finalize
   PUBLIC :: cp_dbcsr_distribution_new
   PUBLIC :: cp_dbcsr_create
   PUBLIC :: cp_dbcsr_work_create
   PUBLIC :: cp_dbcsr_verify_matrix
   PUBLIC :: cp_dbcsr_sum_replicated
   PUBLIC :: cp_dbcsr_checksum
   PUBLIC :: cp_dbcsr_trace
   PUBLIC :: cp_dbcsr_print
   PUBLIC :: cp_dbcsr_print_block_sum
   PUBLIC :: cp_dbcsr_init
   PUBLIC :: cp_dbcsr_init_p
   PUBLIC :: cp_dbcsr_release
   PUBLIC :: cp_dbcsr_release_p
   PUBLIC :: cp_dbcsr_distribute
   PUBLIC :: cp_dbcsr_replicate_all
   PUBLIC :: cp_dbcsr_norm
   PUBLIC :: cp_dbcsr_get_info
   PUBLIC :: cp_dbcsr_get_block_p
   PUBLIC :: cp_dbcsr_put_block
   PUBLIC :: cp_dbcsr_iterator_start
   PUBLIC :: cp_dbcsr_iterator_stop
   PUBLIC :: cp_dbcsr_iterator_next_block
   PUBLIC :: cp_dbcsr_mp_new
   PUBLIC :: cp_dbcsr_mp_release
   PUBLIC :: cp_dbcsr_iterator_blocks_left
   PUBLIC :: cp_dbcsr_distribution_release
   PUBLIC :: cp_dbcsr_col_block_sizes
   PUBLIC :: cp_dbcsr_row_block_sizes
   PUBLIC :: cp_create_bl_distribution
   PUBLIC :: cp_dbcsr_get_matrix_type
   PUBLIC :: cp_dbcsr_get_occupation
   PUBLIC :: cp_dbcsr_distribution
   PUBLIC :: cp_dbcsr_nblkrows_total
   PUBLIC :: cp_dbcsr_nblkcols_total
   PUBLIC :: cp_dbcsr_get_num_blocks
   PUBLIC :: cp_dbcsr_get_data_size
   PUBLIC :: cp_dbcsr_has_symmetry
   PUBLIC :: cp_dbcsr_nfullrows_total
   PUBLIC :: cp_dbcsr_nfullcols_total
   !         1234567890123456789012345678901
   PUBLIC :: cp_dbcsr_get_stored_coordinates
   PUBLIC :: cp_dbcsr_valid_index
   PUBLIC :: cp_dbcsr_get_data_type
   PUBLIC :: cp_dbcsr_reserve_diag_blocks
   PUBLIC :: cp_dbcsr_reserve_block2d
   PUBLIC :: cp_dbcsr_reserve_blocks
   PUBLIC :: cp_dbcsr_reserve_all_blocks
   PUBLIC :: cp_dbcsr_complete_redistribute
   PUBLIC :: cp_dbcsr_gershgorin_norm
   PUBLIC :: cp_dbcsr_frobenius_norm
   PUBLIC :: cp_dbcsr_name
   PUBLIC :: cp_dbcsr_binary_write
   PUBLIC :: cp_dbcsr_binary_read
   PUBLIC :: cp_dbcsr_function_of_elements
   PUBLIC :: cp_dbcsr_triu
   PUBLIC :: cp_dbcsr_clear_mempools
   PUBLIC :: accdrv_avoid_after_busy, accdrv_binning_binsize, accdrv_binning_nbins, &
             accdrv_do_inhomogenous, accdrv_min_flop_process, accdrv_min_flop_sort, &
             accdrv_posterior_buffers, accdrv_posterior_streams, &
             accdrv_priority_buffers, accdrv_priority_streams, &
             dbcsr_get_conf_max_ele_block, &
             dbcsr_get_conf_mm_driver, dbcsr_get_conf_mpi_mem, &
             mm_driver_auto, mm_driver_blas, mm_driver_matmul, mm_driver_smm, mm_driver_xsmm, &
             mm_name_auto, mm_name_blas, mm_name_matmul, &
             mm_name_smm, mm_name_xsmm, multrec_limit
   PUBLIC :: dbcsr_set_conf_randmat_seed, dbcsr_get_conf_randmat_seed
   PUBLIC :: heap_fill, &
             heap_get_first, &
             heap_new, &
             heap_release, &
             heap_reset_first, &
             heap_t
   PUBLIC :: cp_dbcsr_allocate_matrix_set, cp_dbcsr_deallocate_matrix_set, cp_dbcsr_deallocate_matrix
   PUBLIC :: cp_dbcsr_add_block_node
   PUBLIC :: cp_dbcsr_maxabs
   PUBLIC :: cp_dbcsr_arnoldi_ev, dbcsr_arnoldi_data, cp_dbcsr_setup_arnoldi_data, deallocate_arnoldi_data, &
             get_selected_ritz_val, get_selected_ritz_vec, cp_set_arnoldi_initial_vector, &
             cp_dbcsr_arnoldi_extremal
   PUBLIC :: csr_type, csr_p_type, cp_convert_csr_to_dbcsr, cp_convert_dbcsr_to_csr, &
             cp_csr_create_from_dbcsr, cp_csr_destroy, cp_csr_create, &
             csr_dbcsr_blkrow_dist, csr_eqrow_floor_dist, csr_eqrow_ceil_dist, &
             cp_csr_print_sparsity
   PUBLIC :: cp_add_work_coordinate, cp_dbcsr_get_data_p, &
             cp_dbcsr_init_random, cp_dbcsr_setname, cp_set_work_size
   TYPE cp_dbcsr_p_type
      TYPE(cp_dbcsr_type), POINTER :: matrix
   END TYPE cp_dbcsr_p_type

   ! the components of this type must remain private to encapsulate better the internals
   ! of the dbcsr library.
   TYPE cp_dbcsr_type
      INTEGER, PRIVATE                     :: ref_count
      TYPE(dbcsr_obj), PRIVATE             :: matrix
   END TYPE cp_dbcsr_type

   INTERFACE cp_dbcsr_distribution_new
      MODULE PROCEDURE cp_dbcsr_distribution_new_clusters1
      MODULE PROCEDURE cp_dbcsr_distribution_new_clusters2
   END INTERFACE

   INTERFACE cp_dbcsr_create
      MODULE PROCEDURE cp_dbcsr_create_new
      MODULE PROCEDURE cp_dbcsr_create_template
   END INTERFACE

   INTERFACE cp_dbcsr_conform_scalar
      MODULE PROCEDURE make_conformant_scalar_d, &
         make_conformant_scalar_s, &
         make_conformant_scalar_c, &
         make_conformant_scalar_z
   END INTERFACE

   INTERFACE cp_dbcsr_trace
      MODULE PROCEDURE cp_dbcsr_trace_ab_d, &
         cp_dbcsr_trace_ab_s
      MODULE PROCEDURE cp_dbcsr_trace_a_d, &
         cp_dbcsr_trace_a_s
      MODULE PROCEDURE cp_dbcsr_trace_ab_z, &
         cp_dbcsr_trace_ab_c
      MODULE PROCEDURE cp_dbcsr_trace_a_z, &
         cp_dbcsr_trace_a_c
   END INTERFACE

   INTERFACE cp_dbcsr_set
      MODULE PROCEDURE cp_dbcsr_set_d, &
         cp_dbcsr_set_s, &
         cp_dbcsr_set_c, &
         cp_dbcsr_set_z
   END INTERFACE

   INTERFACE cp_dbcsr_add
      MODULE PROCEDURE cp_dbcsr_add_d, &
         cp_dbcsr_add_s, &
         cp_dbcsr_add_c, &
         cp_dbcsr_add_z
   END INTERFACE

   INTERFACE cp_dbcsr_scale
      MODULE PROCEDURE cp_dbcsr_scale_d, &
         cp_dbcsr_scale_s, &
         cp_dbcsr_scale_c, &
         cp_dbcsr_scale_z
   END INTERFACE

   INTERFACE cp_dbcsr_scale_by_vector
      MODULE PROCEDURE cp_dbcsr_scale_by_vector_d, &
         cp_dbcsr_scale_by_vector_s, &
         cp_dbcsr_scale_by_vector_c, &
         cp_dbcsr_scale_by_vector_z
   END INTERFACE

   INTERFACE cp_dbcsr_multiply
      MODULE PROCEDURE cp_dbcsr_multiply_d, &
         cp_dbcsr_multiply_s, &
         cp_dbcsr_multiply_c, &
         cp_dbcsr_multiply_z
   END INTERFACE

   INTERFACE cp_dbcsr_get_block_p
      MODULE PROCEDURE cp_dbcsr_get_block_p_d, &
         cp_dbcsr_get_block_p_s, &
         cp_dbcsr_get_block_p_z, &
         cp_dbcsr_get_block_p_c
      MODULE PROCEDURE cp_dbcsr_get_2d_block_p_d, &
         cp_dbcsr_get_2d_block_p_s, &
         cp_dbcsr_get_2d_block_p_z, &
         cp_dbcsr_get_2d_block_p_c
   END INTERFACE

   INTERFACE cp_dbcsr_put_block
      MODULE PROCEDURE cp_dbcsr_put_block_area
      MODULE PROCEDURE cp_dbcsr_put_block_d, &
         cp_dbcsr_put_block_s, &
         cp_dbcsr_put_block_z, &
         cp_dbcsr_put_block_c
      MODULE PROCEDURE cp_dbcsr_put_block2d_d, &
         cp_dbcsr_put_block2d_s, &
         cp_dbcsr_put_block2d_z, &
         cp_dbcsr_put_block2d_c
   END INTERFACE

   INTERFACE cp_dbcsr_iterator_next_block
      MODULE PROCEDURE cp_iterator_next_2d_block_d, &
         cp_iterator_next_2d_block_s, &
         cp_iterator_next_2d_block_c, &
         cp_iterator_next_2d_block_z
      MODULE PROCEDURE cp_iterator_next_1d_block_d, &
         cp_iterator_next_1d_block_s, &
         cp_iterator_next_1d_block_c, &
         cp_iterator_next_1d_block_z
   END INTERFACE

   INTERFACE cp_dbcsr_reserve_block2d
      MODULE PROCEDURE cp_dbcsr_reserve_block2d_d, &
         cp_dbcsr_reserve_block2d_s, &
         cp_dbcsr_reserve_block2d_c, &
         cp_dbcsr_reserve_block2d_z
   END INTERFACE

   INTERFACE cp_dbcsr_allocate_matrix_set
      MODULE PROCEDURE allocate_dbcsr_matrix_set, allocate_dbcsr_matrix_set_2d
   END INTERFACE

   INTERFACE cp_dbcsr_deallocate_matrix_set
      MODULE PROCEDURE deallocate_dbcsr_matrix_set, &
         deallocate_dbcsr_matrix_set_2d
   END INTERFACE

   INTERFACE cp_csr_create
      MODULE PROCEDURE cp_csr_create_new, &
         cp_csr_create_template
   END INTERFACE

   INTERFACE cp_dbcsr_get_data_p
      MODULE PROCEDURE cp_dbcsr_get_data_c_s, cp_dbcsr_get_data_c_c, &
         cp_dbcsr_get_data_c_d, cp_dbcsr_get_data_c_z
   END INTERFACE

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief Initializes DBCSR
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_init_lib()

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_init_lib', &
         routineP = moduleN//':'//routineN

      CALL cite_reference(Borstnik2014)

      CALL dbcsr_init_lib()
   END SUBROUTINE cp_dbcsr_init_lib

! **************************************************************************************************
!> \brief Finalizes DBCSR
!> \param group ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_finalize_lib(group, output_unit)
      INTEGER, INTENT(IN)                                :: group, output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_finalize_lib', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_finalize_lib(group, output_unit)
   END SUBROUTINE cp_dbcsr_finalize_lib

! **************************************************************************************************
!> \brief Deallocate memory contained in mempools
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_clear_mempools()

      CALL dbcsr_clear_mempools()
   END SUBROUTINE cp_dbcsr_clear_mempools

! **************************************************************************************************
!> \brief cp2k interface to libdbcsr.
!>     The cp2k error is passed as an argument.
!>     Those procedure can de timed with the cp2k timer.
!> \param matrix ...
!> \retval name ...
!> \author vw
! **************************************************************************************************

   FUNCTION cp_dbcsr_name(matrix) RESULT(name)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      CHARACTER(len=default_string_length)               :: name

      name = dbcsr_name(matrix%matrix)
   END FUNCTION cp_dbcsr_name

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param newname ...
! **************************************************************************************************

   SUBROUTINE cp_dbcsr_setname(matrix, newname)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      CHARACTER(len=*), INTENT(IN)                       :: newname

      CALL dbcsr_setname(matrix%matrix, newname)
   END SUBROUTINE cp_dbcsr_setname

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval norm ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_gershgorin_norm(matrix) RESULT(norm)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(KIND=real_8)                                  :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_gershgorin_norm', &
         routineP = moduleN//':'//routineN

      norm = dbcsr_gershgorin_norm(matrix%matrix)

   END FUNCTION cp_dbcsr_gershgorin_norm

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param local ...
!> \retval norm ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_frobenius_norm(matrix, local) RESULT(norm)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      LOGICAL, INTENT(in), OPTIONAL                      :: local
      REAL(KIND=real_8)                                  :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_frobenius_norm', &
         routineP = moduleN//':'//routineN

      norm = dbcsr_frobenius_norm(matrix%matrix, local)

   END FUNCTION cp_dbcsr_frobenius_norm

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval norm ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_maxabs(matrix) RESULT(norm)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(KIND=real_8)                                  :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_maxabs', &
         routineP = moduleN//':'//routineN

      norm = dbcsr_maxabs(matrix%matrix)

   END FUNCTION cp_dbcsr_maxabs

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param redist ...
!> \param keep_sparsity ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_complete_redistribute(matrix, redist, keep_sparsity)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: redist
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_complete_redistribute', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_complete_redistribute(matrix%matrix, redist%matrix, keep_sparsity)

   END SUBROUTINE cp_dbcsr_complete_redistribute

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param rows ...
!> \param cols ...
!> \param blk_pointers ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_reserve_blocks(matrix, rows, cols, blk_pointers)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, DIMENSION(:), INTENT(IN)                  :: rows, cols
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: blk_pointers

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_blocks', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_reserve_blocks(matrix%matrix, rows, cols, blk_pointers)
   END SUBROUTINE cp_dbcsr_reserve_blocks

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_reserve_all_blocks(matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_all_blocks', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_reserve_all_blocks(matrix%matrix)
   END SUBROUTINE cp_dbcsr_reserve_all_blocks

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_reserve_diag_blocks(matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_diag_blocks', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_reserve_diag_blocks(matrix%matrix)
   END SUBROUTINE cp_dbcsr_reserve_diag_blocks

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param row ...
!> \param col ...
!> \param blk ...
!> \param index ...
! **************************************************************************************************
   SUBROUTINE cp_add_work_coordinate(matrix, index_matrix, row, col, blk, index)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, INTENT(IN)                                :: index_matrix, row, col
      INTEGER, INTENT(IN), OPTIONAL                      :: blk
      INTEGER, INTENT(OUT), OPTIONAL                     :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_add_work_coordinate', &
         routineP = moduleN//':'//routineN

      CALL add_work_coordinate(matrix%matrix%m%wms(index_matrix), row, col, blk, index)
   END SUBROUTINE cp_add_work_coordinate

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param newvalue ...
! **************************************************************************************************
   SUBROUTINE cp_set_work_size(matrix, index_matrix, newvalue)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, INTENT(IN)                                :: index_matrix, newvalue

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_set_work_size', &
         routineP = moduleN//':'//routineN

      matrix%matrix%m%wms(index_matrix)%datasize = newvalue
   END SUBROUTINE cp_set_work_size

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param keep_sparsity ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_init_random(matrix, keep_sparsity)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      LOGICAL, OPTIONAL                                  :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_init_random', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_init_random(matrix%matrix, keep_sparsity=keep_sparsity)
   END SUBROUTINE cp_dbcsr_init_random

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval data_type ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_get_data_type(matrix) RESULT(data_type)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_type', &
         routineP = moduleN//':'//routineN

      data_type = dbcsr_get_data_type(matrix%matrix)
   END FUNCTION cp_dbcsr_get_data_type

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval valid_index ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_valid_index(matrix) RESULT(valid_index)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      LOGICAL                                            :: valid_index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_valid_index', &
         routineP = moduleN//':'//routineN

      valid_index = dbcsr_valid_index(matrix%matrix)

   END FUNCTION cp_dbcsr_valid_index

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param column ...
!> \param processor ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_get_stored_coordinates(matrix, row, column, processor)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER, INTENT(INOUT)                             :: row, column
      INTEGER, INTENT(OUT), OPTIONAL                     :: processor

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_stored_coordinates', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_get_stored_coordinates(matrix%matrix, row, column, processor)

   END SUBROUTINE cp_dbcsr_get_stored_coordinates

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval num_blocks ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_get_num_blocks(matrix) RESULT(num_blocks)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: num_blocks

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_num_blocks', &
         routineP = moduleN//':'//routineN

      num_blocks = dbcsr_get_num_blocks(matrix%matrix)

   END FUNCTION cp_dbcsr_get_num_blocks

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval data_size ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_get_data_size(matrix) RESULT(data_size)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_size', &
         routineP = moduleN//':'//routineN

      data_size = dbcsr_get_data_size(matrix%matrix)

   END FUNCTION cp_dbcsr_get_data_size

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval matrix_type ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_get_matrix_type(matrix) RESULT(matrix_type)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      CHARACTER                                          :: matrix_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_matrix_type', &
         routineP = moduleN//':'//routineN

      matrix_type = dbcsr_get_matrix_type(matrix%matrix)

   END FUNCTION cp_dbcsr_get_matrix_type

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval occupation ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_get_occupation(matrix) RESULT(occupation)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      REAL(KIND=real_8)                                  :: occupation

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_occupation', &
         routineP = moduleN//':'//routineN

      occupation = dbcsr_get_occupation(matrix%matrix)

   END FUNCTION cp_dbcsr_get_occupation

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkrows_total ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_nblkrows_total(matrix) RESULT(nblkrows_total)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: nblkrows_total

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_nblkrows_total', &
         routineP = moduleN//':'//routineN

      nblkrows_total = dbcsr_nblkrows_total(matrix%matrix)

   END FUNCTION cp_dbcsr_nblkrows_total

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkcols_total ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_nblkcols_total(matrix) RESULT(nblkcols_total)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: nblkcols_total

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_nblkcols_total', &
         routineP = moduleN//':'//routineN

      nblkcols_total = dbcsr_nblkcols_total(matrix%matrix)
   END FUNCTION cp_dbcsr_nblkcols_total

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullrows_total ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_nfullrows_total(matrix) RESULT(nfullrows_total)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: nfullrows_total

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_nfullrows_total', &
         routineP = moduleN//':'//routineN

      nfullrows_total = dbcsr_nfullrows_total(matrix%matrix)

   END FUNCTION cp_dbcsr_nfullrows_total

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullcols_total ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_nfullcols_total(matrix) RESULT(nfullcols_total)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER                                            :: nfullcols_total

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_nfullcols_total', &
         routineP = moduleN//':'//routineN

      nfullcols_total = dbcsr_nfullcols_total(matrix%matrix)
   END FUNCTION cp_dbcsr_nfullcols_total

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval distribution ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_distribution(matrix) RESULT(distribution)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(dbcsr_distribution_obj)                       :: distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_distribution', &
         routineP = moduleN//':'//routineN

      distribution = dbcsr_distribution(matrix%matrix)

   END FUNCTION cp_dbcsr_distribution

! **************************************************************************************************
!> \brief ...
!> \param block_distribution ...
!> \param block_size ...
!> \param nelements ...
!> \param nbins ...
! **************************************************************************************************
   SUBROUTINE cp_create_bl_distribution(block_distribution, &
                                        block_size, nelements, nbins)
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: block_distribution, block_size
      INTEGER, INTENT(IN)                                :: nelements, nbins

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_create_bl_distribution', &
         routineP = moduleN//':'//routineN

      CALL create_bl_distribution(block_distribution, &
                                  block_size, nelements, nbins)

   END SUBROUTINE cp_create_bl_distribution

! **************************************************************************************************
!> \brief ...
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_distribution_release(dist)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_distribution_release', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_distribution_release(dist)

   END SUBROUTINE cp_dbcsr_distribution_release

! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \retval blocks_left ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_iterator_blocks_left(iterator) RESULT(blocks_left)
      TYPE(cp_dbcsr_iterator), INTENT(IN)                :: iterator
      LOGICAL                                            :: blocks_left

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_iterator_blocks_left', &
         routineP = moduleN//':'//routineN

      blocks_left = dbcsr_iterator_blocks_left(iterator)

   END FUNCTION cp_dbcsr_iterator_blocks_left

! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_mp_release(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(INOUT)                  :: mp_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_mp_release', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_mp_release(mp_env)

   END SUBROUTINE cp_dbcsr_mp_release

! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \param pgrid ...
!> \param mp_group ...
!> \param mynode ...
!> \param numnodes ...
!> \param myprow ...
!> \param mypcol ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_mp_new(mp_env, pgrid, mp_group, mynode, numnodes, myprow, &
                              mypcol)
      TYPE(dbcsr_mp_obj), INTENT(OUT)                    :: mp_env
      INTEGER, DIMENSION(0:, 0:), INTENT(IN)             :: pgrid
      INTEGER, INTENT(IN)                                :: mp_group, mynode
      INTEGER, INTENT(IN), OPTIONAL                      :: numnodes, myprow, mypcol

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_mp_new', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_mp_new(mp_env, pgrid, mp_group, mynode, numnodes, myprow, &
                        mypcol)

   END SUBROUTINE cp_dbcsr_mp_new

! **************************************************************************************************
!> \brief ...
!> \param iterator ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_iterator_stop(iterator)
      TYPE(cp_dbcsr_iterator), INTENT(INOUT)             :: iterator

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_iterator_stop', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_iterator_stop(iterator)

   END SUBROUTINE cp_dbcsr_iterator_stop

! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param matrix ...
!> \param shared ...
!> \param dynamic ...
!> \param dynamic_byrows ...
!> \param contiguous_pointers ...
!> \param read_only ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_iterator_start(iterator, matrix, shared, dynamic, &
                                      dynamic_byrows, contiguous_pointers, read_only)
      TYPE(cp_dbcsr_iterator), INTENT(OUT)               :: iterator
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: shared, dynamic, dynamic_byrows, &
                                                            contiguous_pointers, read_only

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_iterator_start', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_iterator_start(iterator, matrix%matrix, shared, dynamic, &
                                dynamic_byrows, contiguous_pointers, read_only)

   END SUBROUTINE cp_dbcsr_iterator_start

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param nblkrows_total ...
!> \param nblkcols_total ...
!> \param nfullrows_total ...
!> \param nfullcols_total ...
!> \param nblkrows_local ...
!> \param nblkcols_local ...
!> \param nfullrows_local ...
!> \param nfullcols_local ...
!> \param my_prow ...
!> \param my_pcol ...
!> \param local_rows ...
!> \param local_cols ...
!> \param proc_row_dist ...
!> \param proc_col_dist ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param row_blk_offset ...
!> \param col_blk_offset ...
!> \param distribution ...
!> \param name ...
!> \param data_area ...
!> \param matrix_type ...
!> \param data_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_get_info(matrix, nblkrows_total, nblkcols_total, &
                                nfullrows_total, nfullcols_total, &
                                nblkrows_local, nblkcols_local, &
                                nfullrows_local, nfullcols_local, &
                                my_prow, my_pcol, &
                                local_rows, local_cols, proc_row_dist, proc_col_dist, &
                                row_blk_size, col_blk_size, row_blk_offset, col_blk_offset, distribution, name, data_area, &
                                matrix_type, data_type)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER, INTENT(OUT), OPTIONAL :: nblkrows_total, nblkcols_total, nfullrows_total, &
         nfullcols_total, nblkrows_local, nblkcols_local, nfullrows_local, nfullcols_local, &
         my_prow, my_pcol
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: local_rows, local_cols, proc_row_dist, &
                                                            proc_col_dist
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL, &
         POINTER                                         :: row_blk_size, col_blk_size, &
                                                            row_blk_offset, col_blk_offset
      TYPE(dbcsr_distribution_obj), INTENT(OUT), &
         OPTIONAL                                        :: distribution
      CHARACTER(len=*), INTENT(OUT), OPTIONAL            :: name
      TYPE(dbcsr_data_obj), INTENT(OUT), OPTIONAL        :: data_area
      CHARACTER, OPTIONAL                                :: matrix_type
      INTEGER, OPTIONAL                                  :: data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_info', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_get_info(matrix%matrix, nblkrows_total, nblkcols_total, &
                          nfullrows_total, nfullcols_total, &
                          nblkrows_local, nblkcols_local, &
                          nfullrows_local, nfullcols_local, &
                          my_prow, my_pcol, &
                          local_rows, local_cols, proc_row_dist, proc_col_dist, &
                          row_blk_size, col_blk_size, row_blk_offset, col_blk_offset, distribution, name, data_area, &
                          matrix_type, data_type)

   END SUBROUTINE cp_dbcsr_get_info

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param which_norm ...
!> \param norm_scalar ...
!> \param norm_vector ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_norm(matrix, which_norm, norm_scalar, norm_vector)

      TYPE(cp_dbcsr_type), INTENT(INOUT), TARGET         :: matrix
      INTEGER, INTENT(IN)                                :: which_norm
      REAL(dp), INTENT(OUT), OPTIONAL                    :: norm_scalar
      REAL(dp), DIMENSION(:), INTENT(OUT), OPTIONAL      :: norm_vector

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_norm', routineP = moduleN//':'//routineN

      IF (PRESENT(norm_scalar)) THEN
         CALL dbcsr_norm(matrix%matrix, which_norm, norm_scalar=norm_scalar)
      ELSEIF (PRESENT(norm_vector)) THEN
         CALL dbcsr_norm(matrix%matrix, which_norm, norm_vector=norm_vector)
      ELSE
         CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",1004,"Must pass either scalar or vector norm.")
      ENDIF

   END SUBROUTINE cp_dbcsr_norm

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_replicate_all(matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_replicate_all', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_replicate_all(matrix%matrix)
   END SUBROUTINE cp_dbcsr_replicate_all

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param fast ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_distribute(matrix, fast)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      LOGICAL, INTENT(in), OPTIONAL                      :: fast

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_distribute', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_distribute(matrix%matrix, fast)

   END SUBROUTINE cp_dbcsr_distribute

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_release_p(matrix)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_release_p', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(matrix)) THEN
         CALL cp_dbcsr_release(matrix)
         DEALLOCATE (matrix)
      ENDIF

   END SUBROUTINE cp_dbcsr_release_p

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_release(matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_release', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_release(matrix%matrix)

   END SUBROUTINE cp_dbcsr_release

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_init(matrix)
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_init', routineP = moduleN//':'//routineN

      CALL dbcsr_init(matrix%matrix)
      matrix%ref_count = 0

   END SUBROUTINE cp_dbcsr_init

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_init_p(matrix)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_init_p', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(matrix)) THEN
         CALL cp_dbcsr_release(matrix)
         DEALLOCATE (matrix)
      ENDIF

      ALLOCATE (matrix)
      CALL cp_dbcsr_init(matrix)

   END SUBROUTINE cp_dbcsr_init_p

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param nodata ...
!> \param matlab_format ...
!> \param variable_name ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_print(matrix, nodata, matlab_format, variable_name, unit_nr)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: nodata, matlab_format
      CHARACTER(*), INTENT(in), OPTIONAL                 :: variable_name
      INTEGER, OPTIONAL                                  :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_print', routineP = moduleN//':'//routineN

      CALL dbcsr_print(matrix%matrix, nodata, matlab_format, variable_name, unit_nr)

   END SUBROUTINE cp_dbcsr_print

! **************************************************************************************************
!> \brief Prints the sum of the elements in each block
!> \param matrix ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_print_block_sum(matrix, unit_nr)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER, OPTIONAL                                  :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_print_block_sum', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_print_block_sum(matrix%matrix, unit_nr)

   END SUBROUTINE cp_dbcsr_print_block_sum

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param local ...
!> \param pos ...
!> \retval checksum ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_checksum(matrix, local, pos) RESULT(checksum)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: local, pos
      REAL(KIND=dp)                                      :: checksum

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_checksum', &
         routineP = moduleN//':'//routineN

      checksum = dbcsr_checksum(matrix%matrix, &
                                local=local, pos=pos)
   END FUNCTION cp_dbcsr_checksum

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_sum_replicated(matrix)
      TYPE(cp_dbcsr_type), INTENT(inout)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_sum_replicated', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_sum_replicated(matrix%matrix)
   END SUBROUTINE cp_dbcsr_sum_replicated

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_triu(matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_triu', routineP = moduleN//':'//routineN

      CALL dbcsr_triu(matrix%matrix)

   END SUBROUTINE cp_dbcsr_triu

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param verbosity ...
!> \param local ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_verify_matrix(m, verbosity, local)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: m
      INTEGER, INTENT(IN), OPTIONAL                      :: verbosity
      LOGICAL, INTENT(IN), OPTIONAL                      :: local

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_verify_matrix', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_verify_matrix(m%matrix, verbosity, local)

   END SUBROUTINE cp_dbcsr_verify_matrix

! **************************************************************************************************
!> \brief Creates new distribution from block and cluster distributions
!> \param[out] dist           distribution
!> \param[in] mp_env          multiprocessing environment
!> \param row_dist_block ...
!> \param col_dist_block ...
!> \param row_dist_cluster ...
!> \param col_dist_cluster ...
!> \param reuse_arrays ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_distribution_new_clusters1(dist, mp_env, &
                                                  row_dist_block, col_dist_block, &
                                                  row_dist_cluster, col_dist_cluster, &
                                                  reuse_arrays, unit_nr)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: row_dist_block, col_dist_block, &
                                                            row_dist_cluster, col_dist_cluster
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse_arrays
      INTEGER, INTENT(IN), OPTIONAL                      :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_distribution_new_clusters1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iunit, timing_handle
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_cluster2, row_dist_cluster2

      CALL timeset(routineN, timing_handle)

      row_dist_cluster2 => row_dist_cluster
      col_dist_cluster2 => col_dist_cluster
      IF (ASSOCIATED(row_dist_cluster2)) THEN
         IF (ANY(row_dist_cluster2 .EQ. -1)) NULLIFY (row_dist_cluster2)
      END IF
      IF (ASSOCIATED(col_dist_cluster2)) THEN
         IF (ANY(col_dist_cluster2 .EQ. -1)) NULLIFY (col_dist_cluster2)
      END IF

      IF ((ASSOCIATED(row_dist_cluster2) .OR. ASSOCIATED(col_dist_cluster2)) .AND. &
          .NOT. dbcsr_get_conf_use_mpi_exp()) THEN
         iunit = default_output_unit
         IF (PRESENT(unit_nr)) iunit = unit_nr
         IF (iunit > 0) THEN
            WRITE (iunit, *) "************************************************************"
            WRITE (iunit, *) "WARNING:: Cannot use cluster distribution without MPI RMA!!!"
            WRITE (iunit, *) "          Set 'USE_MPI_EXP T' in DBCSR section."
            WRITE (iunit, *) "************************************************************"
         ENDIF
         NULLIFY (row_dist_cluster2, col_dist_cluster2)
      ENDIF

      CALL dbcsr_distribution_new(dist, mp_env, row_dist_block, col_dist_block, &
                                  row_dist_cluster=row_dist_cluster2, &
                                  col_dist_cluster=col_dist_cluster2, &
                                  reuse_arrays=reuse_arrays)

      CALL timestop(timing_handle)

   END SUBROUTINE cp_dbcsr_distribution_new_clusters1

! **************************************************************************************************
!> \brief Creates new distribution from block and cluster distributions
!> \param[out] dist           distribution
!> \param[in] mp_env          multiprocessing environment
!> \param row_dist row distribution
!> \param col_dist column distribution
!> \param reuse_arrays ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_distribution_new_clusters2(dist, mp_env, row_dist, col_dist, reuse_arrays, unit_nr)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(:, :), INTENT(IN), POINTER      :: row_dist, col_dist
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse_arrays
      INTEGER, INTENT(IN), OPTIONAL                      :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_distribution_new_clusters2', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(:), POINTER                     :: col_dist_block, col_dist_cluster, &
                                                            row_dist_block, row_dist_cluster

      row_dist_block => row_dist(:, 1)
      col_dist_block => col_dist(:, 1)
      row_dist_cluster => row_dist(:, 2)
      col_dist_cluster => col_dist(:, 2)

      CALL cp_dbcsr_distribution_new(dist, mp_env, &
                                     row_dist_block, col_dist_block, &
                                     row_dist_cluster, col_dist_cluster, &
                                     reuse_arrays, unit_nr)

   END SUBROUTINE cp_dbcsr_distribution_new_clusters2

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param reshuffle ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_finalize(matrix, reshuffle)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: reshuffle

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_finalize', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_finalize(matrix%matrix, reshuffle)
   END SUBROUTINE cp_dbcsr_finalize

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param nblks_guess ...
!> \param sizedata_guess ...
!> \param n ...
!> \param work_mutable ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_work_create(matrix, nblks_guess, sizedata_guess, n, &
                                   work_mutable)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, INTENT(IN), OPTIONAL                      :: nblks_guess, sizedata_guess, n
      LOGICAL, INTENT(in), OPTIONAL                      :: work_mutable

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_work_create', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_work_create(matrix%matrix, nblks_guess, sizedata_guess, n, &
                             work_mutable)
   END SUBROUTINE cp_dbcsr_work_create

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param name ...
!> \param dist ...
!> \param matrix_type ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param nze ...
!> \param data_type ...
!> \param reuse ...
!> \param reuse_arrays ...
!> \param mutable_work ...
!> \param replication_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_create_new(matrix, name, dist, matrix_type, &
                                  row_blk_size, col_blk_size, nze, data_type, reuse, &
                                  reuse_arrays, mutable_work, replication_type)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      CHARACTER(len=*), INTENT(IN)                       :: name
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      CHARACTER, INTENT(IN)                              :: matrix_type
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: row_blk_size, col_blk_size
      INTEGER, INTENT(IN), OPTIONAL                      :: nze, data_type
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse, reuse_arrays, mutable_work
      CHARACTER, INTENT(IN), OPTIONAL                    :: replication_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_create_new', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_create(matrix%matrix, name, dist, &
                        matrix_type, &
                        row_blk_size, col_blk_size, nze=nze, &
                        data_type=data_type, reuse=reuse, &
                        reuse_arrays=reuse_arrays, &
                        mutable_work=mutable_work, replication_type=replication_type)

      matrix%ref_count = 1

   END SUBROUTINE cp_dbcsr_create_new

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param name ...
!> \param template ...
!> \param dist ...
!> \param matrix_type ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param nze ...
!> \param data_type ...
!> \param reuse_arrays ...
!> \param mutable_work ...
!> \param replication_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_create_template(matrix, name, template, &
                                       dist, matrix_type, &
                                       row_blk_size, col_blk_size, nze, data_type, &
                                       reuse_arrays, mutable_work, replication_type)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: name
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: template
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: dist
      CHARACTER, INTENT(IN), OPTIONAL                    :: matrix_type
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL, &
         POINTER                                         :: row_blk_size, col_blk_size
      INTEGER, INTENT(IN), OPTIONAL                      :: nze, data_type
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse_arrays, mutable_work
      CHARACTER, INTENT(IN), OPTIONAL                    :: replication_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_create_template', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_create(matrix%matrix, template%matrix, &
                        name=name, dist=dist, matrix_type=matrix_type, &
                        row_blk_size=row_blk_size, col_blk_size=col_blk_size, &
                        nze=nze, data_type=data_type, &
                        reuse_arrays=reuse_arrays, mutable_work=mutable_work, &
                        replication_type=replication_type)
      !
      matrix%ref_count = 1
   END SUBROUTINE cp_dbcsr_create_template

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eps ...
!> \param method ...
!> \param use_absolute ...
!> \param filter_diag ...
!> \param thorough ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_filter(matrix, eps, method, use_absolute, filter_diag, &
                              thorough)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(dp), INTENT(IN)                               :: eps
      INTEGER, INTENT(IN), OPTIONAL                      :: method
      LOGICAL, INTENT(in), OPTIONAL                      :: use_absolute, filter_diag, thorough

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_filter', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: quick

      IF (PRESENT(thorough)) THEN
         quick = .NOT. thorough
      ELSE
         quick = .FALSE.
      ENDIF
      CALL dbcsr_filter(matrix%matrix, cp_dbcsr_conform_scalar(eps, matrix), &
                        method, use_absolute, filter_diag, &
                        quick=quick)
   END SUBROUTINE cp_dbcsr_filter

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_set_diag(matrix, diag)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_set_diag', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_set_diag(matrix%matrix, diag)
   END SUBROUTINE cp_dbcsr_set_diag

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_get_diag(matrix, diag)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      REAL(dp), DIMENSION(:), INTENT(INOUT)              :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_diag', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_get_diag(matrix%matrix, diag)
   END SUBROUTINE cp_dbcsr_get_diag

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_get_block_diag(matrix, diag)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_block_diag', &
         routineP = moduleN//':'//routineN

      CALL cp_dbcsr_create(diag, "Diagonal of "//TRIM(cp_dbcsr_name(matrix)), &
                           template=matrix)
      CALL dbcsr_get_block_diag(matrix%matrix, diag%matrix)
   END SUBROUTINE cp_dbcsr_get_block_diag

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param alpha_scalar ...
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_add_on_diag(matrix, alpha_scalar, first_row, last_row)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(KIND=dp), INTENT(IN)                          :: alpha_scalar
      INTEGER, INTENT(in), OPTIONAL                      :: first_row, last_row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_add_on_diag', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_add_on_diag(matrix%matrix, &
                             cp_dbcsr_conform_scalar(alpha_scalar, matrix), &
                             first_row, last_row)
   END SUBROUTINE cp_dbcsr_add_on_diag

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param filepath ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_binary_write(matrix, filepath)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      CHARACTER(LEN=*), INTENT(IN)                       :: filepath

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_binary_write', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_binary_write(matrix%matrix, filepath)

   END SUBROUTINE cp_dbcsr_binary_write

! **************************************************************************************************
!> \brief ...
!> \param filepath ...
!> \param distribution ...
!> \param groupid ...
!> \param matrix_new ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_binary_read(filepath, distribution, groupid, matrix_new)
      CHARACTER(len=*), INTENT(IN)                       :: filepath
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: distribution
      INTEGER, INTENT(IN), OPTIONAL                      :: groupid
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_new

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_binary_read', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_binary_read(filepath, distribution, groupid, matrix_new%matrix)

   END SUBROUTINE cp_dbcsr_binary_read

! **************************************************************************************************
!> \brief ...
!> \param matrix_b ...
!> \param matrix_a ...
!> \param name ...
!> \param keep_sparsity ...
!> \param shallow_data ...
!> \param keep_imaginary ...
!> \param matrix_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_copy(matrix_b, matrix_a, name, keep_sparsity, &
                            shallow_data, keep_imaginary, matrix_type)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_b
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity, shallow_data, &
                                                            keep_imaginary
      CHARACTER, INTENT(IN), OPTIONAL                    :: matrix_type

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_copy', routineP = moduleN//':'//routineN

      IF (matrix_b%ref_count .EQ. 0) THEN
         CALL cp_dbcsr_create(matrix_b, template=matrix_a)
      ENDIF
      CALL dbcsr_copy(matrix_b%matrix, matrix_a%matrix, name, keep_sparsity, &
                      shallow_data, keep_imaginary, matrix_type)
   END SUBROUTINE cp_dbcsr_copy

! **************************************************************************************************
!> \brief ...
!> \param matrix_b ...
!> \param matrix_a ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_copy_into_existing(matrix_b, matrix_a)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_b
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_copy_into_existing', &
         routineP = moduleN//':'//routineN

      IF (matrix_b%ref_count .EQ. 0) THEN
         CALL cp_dbcsr_create(matrix_b, template=matrix_a)
      ENDIF
      CALL dbcsr_copy_into_existing(matrix_b%matrix, matrix_a%matrix)
   END SUBROUTINE cp_dbcsr_copy_into_existing

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_desymmetrize(matrix_a, matrix_b)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_desymmetrize', &
         routineP = moduleN//':'//routineN

      IF (matrix_b%ref_count .EQ. 0) THEN
         CALL cp_dbcsr_create(matrix_b, template=matrix_a)
      ENDIF
      CALL dbcsr_desymmetrize_deep(matrix_a%matrix, matrix_b%matrix, &
                                   untransposed_data=.TRUE.)
   END SUBROUTINE cp_dbcsr_desymmetrize

! **************************************************************************************************
!> \brief ...
!> \param transposed ...
!> \param normal ...
!> \param shallow_data_copy ...
!> \param transpose_data ...
!> \param transpose_distribution ...
!> \param use_distribution ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_transposed(transposed, normal, shallow_data_copy, &
                                  transpose_data, transpose_distribution, use_distribution)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: transposed
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: normal
      LOGICAL, INTENT(IN), OPTIONAL                      :: shallow_data_copy, transpose_data, &
                                                            transpose_distribution
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: use_distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_transposed', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: myshallow_data_copy, &
                                                            mytranspose_distribution
      TYPE(dbcsr_distribution_obj)                       :: myuse_distribution

!   set some defaults to make usage a bit less painful (fschiff)

      myshallow_data_copy = .FALSE.
      myuse_distribution = cp_dbcsr_distribution(normal)
      mytranspose_distribution = .FALSE.
      IF (PRESENT(shallow_data_copy)) myshallow_data_copy = shallow_data_copy
      IF (PRESENT(use_distribution)) myuse_distribution = use_distribution
      IF (PRESENT(transpose_distribution)) mytranspose_distribution = transpose_distribution

      CALL dbcsr_new_transposed(transposed%matrix, normal%matrix, myshallow_data_copy, &
                                transpose_data, mytranspose_distribution, &
                                use_distribution=myuse_distribution)
   END SUBROUTINE cp_dbcsr_transposed

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param arnoldi_data ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_arnoldi_ev(matrix, arnoldi_data)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), INTENT(IN)    :: matrix
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_arnoldi_ev', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(dbcsr_obj_type_p), ALLOCATABLE, DIMENSION(:)  :: m_ptrs

! convert cp_dbcsr_p_type to dbcsr_obj_type_p

      ALLOCATE (m_ptrs(SIZE(matrix)))
      DO i = 1, SIZE(matrix)
         m_ptrs(i)%matrix => matrix(i)%matrix%matrix
      END DO

      CALL dbcsr_arnoldi_ev(m_ptrs, arnoldi_data)

      DEALLOCATE (m_ptrs)

   END SUBROUTINE cp_dbcsr_arnoldi_ev

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param max_ev ...
!> \param min_ev ...
!> \param converged ...
!> \param threshold ...
!> \param max_iter ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_arnoldi_extremal(matrix, max_ev, min_ev, converged, threshold, max_iter)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(KIND=dp), INTENT(OUT)                         :: max_ev, min_ev
      LOGICAL, INTENT(OUT)                               :: converged
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      INTEGER, INTENT(IN)                                :: max_iter

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_arnoldi_extremal', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_arnoldi_extremal(matrix%matrix, max_ev, min_ev, converged, threshold, max_iter)

   END SUBROUTINE cp_dbcsr_arnoldi_extremal

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param matrix ...
!> \param max_iter ...
!> \param threshold ...
!> \param selection_crit ...
!> \param nval_request ...
!> \param nrestarts ...
!> \param generalized_ev ...
!> \param iram ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_setup_arnoldi_data(arnoldi_data, matrix, max_iter, threshold, selection_crit, &
                                          nval_request, nrestarts, generalized_ev, iram)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      TYPE(cp_dbcsr_p_type), DIMENSION(:)                :: matrix
      INTEGER                                            :: max_iter
      REAL(real_8)                                       :: threshold
      INTEGER                                            :: selection_crit, nval_request, nrestarts
      LOGICAL                                            :: generalized_ev, iram

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_setup_arnoldi_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(dbcsr_obj_type_p), ALLOCATABLE, DIMENSION(:)  :: m_ptrs

! convert cp_dbcsr_p_type to dbcsr_obj_type_p

      ALLOCATE (m_ptrs(SIZE(matrix)))
      DO i = 1, SIZE(matrix)
         m_ptrs(i)%matrix => matrix(i)%matrix%matrix
      END DO

      CALL setup_arnoldi_data(arnoldi_data, m_ptrs, max_iter, threshold, selection_crit, &
                              nval_request, nrestarts, generalized_ev, iram)

      DEALLOCATE (m_ptrs)
   END SUBROUTINE cp_dbcsr_setup_arnoldi_data

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param vector ...
! **************************************************************************************************
   SUBROUTINE cp_set_arnoldi_initial_vector(arnoldi_data, vector)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      TYPE(cp_dbcsr_type)                                :: vector

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_set_arnoldi_initial_vector', &
         routineP = moduleN//':'//routineN

      CALL set_arnoldi_initial_vector(arnoldi_data, vector%matrix)

   END SUBROUTINE cp_set_arnoldi_initial_vector

! **************************************************************************************************
!> \brief ...
!> \param arnoldi_data ...
!> \param ind ...
!> \param matrix ...
!> \param vector ...
! **************************************************************************************************
   SUBROUTINE get_selected_ritz_vec(arnoldi_data, ind, matrix, vector)
      TYPE(dbcsr_arnoldi_data)                           :: arnoldi_data
      INTEGER                                            :: ind
      TYPE(cp_dbcsr_type)                                :: matrix, vector

      CHARACTER(len=*), PARAMETER :: routineN = 'get_selected_ritz_vec', &
         routineP = moduleN//':'//routineN

      CALL get_selected_ritz_vector(arnoldi_data, ind, matrix%matrix, vector%matrix)

   END SUBROUTINE get_selected_ritz_vec

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param func ...
!> \param a0 ...
!> \param a1 ...
!> \param a2 ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_function_of_elements(matrix_a, func, a0, a1, a2)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_a
      INTEGER, INTENT(IN)                                :: func
      REAL(kind=dp), INTENT(IN), OPTIONAL                :: a0, a1, a2

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_function_of_elements', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_function_of_elements(matrix_a%matrix, func, a0, a1, a2)
   END SUBROUTINE cp_dbcsr_function_of_elements

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param matrix_c ...
!> \param b_assume_value ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_hadamard_product(matrix_a, matrix_b, matrix_c, b_assume_value)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a, matrix_b
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: b_assume_value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_hadamard_product', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_hadamard_product(matrix_a%matrix, matrix_b%matrix, matrix_c%matrix, &
                                  b_assume_value)
   END SUBROUTINE cp_dbcsr_hadamard_product

! **************************************************************************************************
!> \brief     Adjusts matrices to be compatibly for multiplication.
!>
!> In BLAS xGEMM or PxGEMM style, CP2K allows with differing
!> dimensions to be multiplied.  This compatibilyt has been removed
!> from DBCSR.  Therefore, matrices must be matched for
!> multiplication:
!> <ul>
!> <li> The rows of matrix_b are adjusted to fit the columns of matrix_a,
!> <li> The columns of matrix_b are adjusted to fit the columns of matrix_c, and
!> <li> The rows of matrix_a are adjusted to fit the rows of matrix_c.
!> </ul>
!> Transformations are skipped if the sizes are already compatible.
!> However, in any case a copy of matrix_a and matrix_b is made.
!> It is the responsibility of the caller to release these new matrices.
!>
!> Some copies could be avoided with more code.
!> \param matrix_c ...
!> \param matrix_a ...
!> \param tr_a ...
!> \param matrix_b ...
!> \param tr_b ...
!> \param new_a ...
!> \param new_b ...
! **************************************************************************************************
   SUBROUTINE matrix_match_sizes(matrix_c, matrix_a, tr_a, matrix_b, tr_b, &
                                 new_a, new_b)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_c, matrix_a
      CHARACTER, INTENT(IN)                              :: tr_a
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_b
      CHARACTER, INTENT(IN)                              :: tr_b
      TYPE(cp_dbcsr_type), INTENT(OUT)                   :: new_a, new_b

      CHARACTER(LEN=*), PARAMETER :: routineN = 'matrix_match_sizes', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: tr_a_l, tr_b_l
      INTEGER                                            :: a_cols_total, a_rows_total, &
                                                            b_cols_total, b_rows_total, &
                                                            c_cols_total, c_rows_total
      INTEGER, DIMENSION(:), POINTER                     :: new_col_clusters, new_col_dist, &
                                                            new_col_size, new_row_clusters, &
                                                            new_row_dist, new_row_size
      LOGICAL                                            :: atr, btr, create_col_dist, &
                                                            create_row_dist, new_a_rows, &
                                                            new_b_cols, new_b_rows
      TYPE(dbcsr_distribution_obj)                       :: a_dist, b_dist, c_dist, new_dist

!   ---------------------------------------------------------------------------

      tr_a_l = tr_a
      CALL uppercase(tr_a_l)
      tr_b_l = tr_b
      CALL uppercase(tr_b_l)

      create_col_dist = .FALSE.
      create_row_dist = .FALSE.

      btr = tr_b_l .NE. dbcsr_no_transpose
      atr = tr_a_l .NE. dbcsr_no_transpose
      !
      c_rows_total = cp_dbcsr_nfullrows_total(matrix_c)
      c_cols_total = cp_dbcsr_nfullcols_total(matrix_c)
      b_rows_total = cp_dbcsr_nfullrows_total(matrix_b)
      b_cols_total = cp_dbcsr_nfullcols_total(matrix_b)
      a_rows_total = cp_dbcsr_nfullrows_total(matrix_a)
      a_cols_total = cp_dbcsr_nfullcols_total(matrix_a)
      IF (atr) CALL swap(a_cols_total, a_rows_total)
      IF (btr) CALL swap(b_cols_total, b_rows_total)

      NULLIFY (new_row_clusters, new_col_clusters)

      !
      ! Process matrix B.
      !
      ! This check is faster than explicity comparing blocked row and
      ! column sizes.
      new_b_cols = c_cols_total .NE. b_cols_total
      new_b_rows = a_cols_total .NE. b_rows_total
      a_dist = cp_dbcsr_distribution(matrix_a)
      b_dist = cp_dbcsr_distribution(matrix_b)
      c_dist = cp_dbcsr_distribution(matrix_c)
      IF (new_b_rows .OR. new_b_cols) THEN
         CALL cp_dbcsr_init(new_b)
         IF (.NOT. btr) THEN
            IF (new_b_cols) THEN
               CALL match_1_dist(new_col_dist, &
                                 dbcsr_distribution_col_dist(b_dist), &
                                 dbcsr_distribution_col_dist(c_dist), &
                                 new_col_clusters, &
                                 dbcsr_distribution_col_clusters(b_dist), &
                                 dbcsr_distribution_col_clusters(c_dist))
               create_col_dist = .TRUE.
            ELSE
               new_col_dist => dbcsr_distribution_col_dist(b_dist)
               new_col_clusters => dbcsr_distribution_col_clusters(b_dist)
            ENDIF
            new_col_size => cp_dbcsr_col_block_sizes(matrix_c)
            IF (new_b_rows) THEN
               IF (.NOT. atr) THEN
                  CALL match_1_dist(new_row_dist, &
                                    dbcsr_distribution_row_dist(b_dist), &
                                    dbcsr_distribution_col_dist(a_dist), &
                                    new_row_clusters, &
                                    dbcsr_distribution_row_clusters(b_dist), &
                                    dbcsr_distribution_col_clusters(a_dist))
                  create_row_dist = .TRUE.
                  new_row_size => cp_dbcsr_col_block_sizes(matrix_a)
               ELSE
                  CALL match_1_dist(new_row_dist, &
                                    dbcsr_distribution_row_dist(b_dist), &
                                    dbcsr_distribution_row_dist(a_dist), &
                                    new_row_clusters, &
                                    dbcsr_distribution_row_clusters(b_dist), &
                                    dbcsr_distribution_row_clusters(a_dist))
                  create_row_dist = .TRUE.
                  new_row_size => cp_dbcsr_row_block_sizes(matrix_a)
               ENDIF
            ELSE
               new_row_dist => dbcsr_distribution_row_dist(b_dist)
               new_row_clusters => dbcsr_distribution_row_clusters(b_dist)
               new_row_size => cp_dbcsr_row_block_sizes(matrix_b)
            ENDIF
         ELSE
            IF (new_b_cols) THEN
               CALL match_1_dist(new_row_dist, &
                                 dbcsr_distribution_row_dist(b_dist), &
                                 dbcsr_distribution_col_dist(c_dist), &
                                 new_row_clusters, &
                                 dbcsr_distribution_row_clusters(b_dist), &
                                 dbcsr_distribution_col_clusters(c_dist))
               create_row_dist = .TRUE.
            ELSE
               new_row_dist => dbcsr_distribution_row_dist(b_dist)
               new_row_clusters => dbcsr_distribution_row_clusters(b_dist)
            ENDIF
            new_row_size => cp_dbcsr_col_block_sizes(matrix_c)
            IF (new_b_rows) THEN
               IF (.NOT. atr) THEN
                  CALL match_1_dist(new_col_dist, &
                                    dbcsr_distribution_col_dist(b_dist), &
                                    dbcsr_distribution_col_dist(a_dist), &
                                    new_col_clusters, &
                                    dbcsr_distribution_col_clusters(b_dist), &
                                    dbcsr_distribution_col_clusters(a_dist))
                  create_col_dist = .TRUE.
                  new_col_size => cp_dbcsr_col_block_sizes(matrix_a)
               ELSE
                  CALL match_1_dist(new_col_dist, &
                                    dbcsr_distribution_col_dist(b_dist), &
                                    dbcsr_distribution_row_dist(a_dist), &
                                    new_col_clusters, &
                                    dbcsr_distribution_col_clusters(b_dist), &
                                    dbcsr_distribution_row_clusters(a_dist))
                  create_col_dist = .TRUE.
                  new_col_size => cp_dbcsr_row_block_sizes(matrix_a)
               ENDIF
            ELSE
               new_col_dist => dbcsr_distribution_col_dist(b_dist)
               new_col_clusters => dbcsr_distribution_col_clusters(b_dist)
               new_col_size => cp_dbcsr_col_block_sizes(matrix_b)
            ENDIF
         ENDIF
         CALL dbcsr_distribution_new(new_dist, dbcsr_distribution_mp(c_dist), &
                                     new_row_dist, new_col_dist, &
                                     row_dist_cluster=new_row_clusters, &
                                     col_dist_cluster=new_col_clusters)
         IF (create_col_dist) THEN
            DEALLOCATE (new_col_dist)
            IF (ASSOCIATED(new_col_clusters)) DEALLOCATE (new_col_clusters)
         ENDIF
         IF (create_row_dist) THEN
            DEALLOCATE (new_row_dist)
            IF (ASSOCIATED(new_row_clusters)) DEALLOCATE (new_row_clusters)
         ENDIF
         CALL cp_dbcsr_create(new_b, template=matrix_b, &
                              dist=new_dist, &
                              row_blk_size=new_row_size, &
                              col_blk_size=new_col_size)
         CALL dbcsr_distribution_release(new_dist)
         CALL cp_dbcsr_complete_redistribute(matrix_b, new_b)
      ELSE
         CALL cp_dbcsr_init(new_b)
         CALL cp_dbcsr_copy(new_b, matrix_b, shallow_data=.TRUE.)
      ENDIF

      create_col_dist = .FALSE.
      create_row_dist = .FALSE.
      NULLIFY (new_row_clusters, new_col_clusters)
      !
      ! Process matrix A
      new_a_rows = a_rows_total .NE. c_rows_total
      IF (new_a_rows) THEN
         CALL cp_dbcsr_init(new_a)
         IF (atr) THEN
            new_row_dist => dbcsr_distribution_row_dist(a_dist)
            new_row_clusters => dbcsr_distribution_row_clusters(a_dist)
            new_row_size => cp_dbcsr_row_block_sizes(matrix_a)
            CALL match_1_dist(new_col_dist, &
                              dbcsr_distribution_col_dist(a_dist), &
                              dbcsr_distribution_row_dist(c_dist), &
                              new_col_clusters, &
                              dbcsr_distribution_col_clusters(a_dist), &
                              dbcsr_distribution_row_clusters(c_dist))
            create_col_dist = .TRUE.
            new_col_size => cp_dbcsr_row_block_sizes(matrix_c)
         ELSE
            CALL match_1_dist(new_row_dist, &
                              dbcsr_distribution_row_dist(a_dist), &
                              dbcsr_distribution_row_dist(c_dist), &
                              new_row_clusters, &
                              dbcsr_distribution_row_clusters(a_dist), &
                              dbcsr_distribution_row_clusters(c_dist))
            create_row_dist = .TRUE.
            new_row_size => cp_dbcsr_row_block_sizes(matrix_c)
            new_col_dist => dbcsr_distribution_col_dist(a_dist)
            new_col_clusters => dbcsr_distribution_col_clusters(a_dist)
            new_col_size => cp_dbcsr_col_block_sizes(matrix_a)
         ENDIF
         CALL dbcsr_distribution_new(new_dist, dbcsr_distribution_mp(a_dist), &
                                     new_row_dist, new_col_dist, &
                                     new_row_clusters, new_col_clusters)
         IF (create_col_dist) THEN
            DEALLOCATE (new_col_dist)
            DEALLOCATE (new_col_clusters)
         ENDIF
         IF (create_row_dist) THEN
            DEALLOCATE (new_row_dist)
            DEALLOCATE (new_row_clusters)
         ENDIF
         CALL cp_dbcsr_create(new_a, template=matrix_a, &
                              dist=new_dist, &
                              row_blk_size=new_row_size, &
                              col_blk_size=new_col_size)
         CALL dbcsr_distribution_release(new_dist)
         CALL cp_dbcsr_complete_redistribute(matrix_a, new_a)
      ELSE
         CALL cp_dbcsr_init(new_a)
         CALL cp_dbcsr_copy(new_a, matrix_a, shallow_data=.TRUE.)
      ENDIF
   END SUBROUTINE matrix_match_sizes

! **************************************************************************************************
!> \brief Matches a 1-D distribution
!>
!> Forms a new distribution new_dist of size template_dist.  As much data as
!> possible is taken from old_dist.
!> \param new_dist ...
!> \param old_dist ...
!> \param template_dist ...
!> \param new_clusters ...
!> \param old_clusters ...
!> \param template_clusters ...
! **************************************************************************************************
   SUBROUTINE match_1_dist(new_dist, old_dist, template_dist, &
                           new_clusters, old_clusters, template_clusters)
      INTEGER, DIMENSION(:), INTENT(out), POINTER        :: new_dist
      INTEGER, DIMENSION(:), INTENT(in), POINTER         :: old_dist, template_dist
      INTEGER, DIMENSION(:), INTENT(out), POINTER        :: new_clusters
      INTEGER, DIMENSION(:), INTENT(in), POINTER         :: old_clusters, template_clusters

      INTEGER                                            :: i, max_bin, new_size, old_size, sz

      old_size = SIZE(old_dist)
      new_size = SIZE(template_dist)
      sz = MIN(old_size, new_size)
      ALLOCATE (new_dist(new_size))
      new_dist = template_dist
      new_dist(1:sz) = old_dist(1:sz)
      max_bin = MAXVAL(old_dist)
      IF (max_bin .GT. 0) THEN
         DO i = sz+1, new_size
            new_dist(i) = MOD(i, max_bin)
         END DO
      ELSE
         new_dist(sz+1:new_size) = 0
      ENDIF
      IF (ASSOCIATED(template_clusters) .AND. ASSOCIATED(old_clusters)) THEN
         ALLOCATE (new_clusters(new_size))
         new_clusters = template_clusters
         new_clusters(1:sz) = old_clusters(1:sz)
         max_bin = MAXVAL(old_clusters)
         DO i = sz+1, new_size
            new_clusters(i) = MOD(i, max_bin)
         END DO
      ENDIF
   END SUBROUTINE match_1_dist

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval row_blk_sizes ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_row_block_sizes(matrix) RESULT(row_blk_sizes)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_row_block_sizes', &
         routineP = moduleN//':'//routineN

      row_blk_sizes => array_data(matrix%matrix%m%row_blk_size)
   END FUNCTION cp_dbcsr_row_block_sizes

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval col_blk_sizes ...
! **************************************************************************************************
   FUNCTION cp_dbcsr_col_block_sizes(matrix) RESULT(col_blk_sizes)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_col_block_sizes', &
         routineP = moduleN//':'//routineN

      col_blk_sizes => array_data(matrix%matrix%m%col_blk_size)
   END FUNCTION cp_dbcsr_col_block_sizes

! **************************************************************************************************
!> \brief Deallocates a DBCSR matrix for compatibility with CP2K
!> \param[in,out] matrix      DBCSR matrix
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_deallocate_matrix(matrix)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_deallocate_matrix', &
         routineP = moduleN//':'//routineN

      CALL cp_dbcsr_release(matrix)
      IF (cp_dbcsr_valid_index(matrix)) &
         CALL cp_abort(cp__l("dbcsrwrap/cp_dbcsr_interface.F",2151), &
                       'You should not "deallocate" a referenced matrix. '// &
                       'Avoid pointers to DBCSR matrices.')
      DEALLOCATE (matrix)

   END SUBROUTINE cp_dbcsr_deallocate_matrix

! **************************************************************************************************
! CP2k-compatible matrix sets
! **************************************************************************************************

! **************************************************************************************************
!> \brief   Allocate and initialize a real matrix 1-dimensional set.
!> \param[in,out] matrix_set  Set containing the DBCSR matrices
!> \param[in] nmatrix         Size of set
!> \par History
!>      2009-08-17 Adapted from sparse_matrix_type for DBCSR
! **************************************************************************************************
   SUBROUTINE allocate_dbcsr_matrix_set(matrix_set, nmatrix)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_set
      INTEGER, INTENT(IN)                                :: nmatrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_dbcsr_matrix_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imatrix

      IF (ASSOCIATED(matrix_set)) CALL cp_dbcsr_deallocate_matrix_set(matrix_set)
      ALLOCATE (matrix_set(nmatrix))
      DO imatrix = 1, nmatrix
         NULLIFY (matrix_set(imatrix)%matrix)
      END DO
   END SUBROUTINE allocate_dbcsr_matrix_set

! **************************************************************************************************
!> \brief   Allocate and initialize a real matrix 2-dimensional set.
!> \param[in,out] matrix_set  Set containing the DBCSR matrix pointer type
!> \param[in] nmatrix         Size of set
!> \param mmatrix ...
!> \par History
!>      2009-08-17 Adapted from sparse_matrix_type for DBCSR
! **************************************************************************************************
   SUBROUTINE allocate_dbcsr_matrix_set_2d(matrix_set, nmatrix, mmatrix)
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_set
      INTEGER, INTENT(IN)                                :: nmatrix, mmatrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_dbcsr_matrix_set_2d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imatrix, jmatrix

      IF (ASSOCIATED(matrix_set)) CALL cp_dbcsr_deallocate_matrix_set(matrix_set)
      ALLOCATE (matrix_set(nmatrix, mmatrix))
      DO jmatrix = 1, mmatrix
         DO imatrix = 1, nmatrix
            NULLIFY (matrix_set(imatrix, jmatrix)%matrix)
         END DO
      END DO
   END SUBROUTINE allocate_dbcsr_matrix_set_2d

! **************************************************************************************************
!> \brief Deallocate a real matrix set and release all of the member matrices.
!> \param[in,out] matrix_set  Set containing the DBCSR matrix pointer type
!> \par History
!>      2009-08-17 Adapted from sparse_matrix_type for DBCSR
! **************************************************************************************************
   SUBROUTINE deallocate_dbcsr_matrix_set(matrix_set)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_dbcsr_matrix_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imatrix

      IF (ASSOCIATED(matrix_set)) THEN
         DO imatrix = 1, SIZE(matrix_set)
            CALL cp_dbcsr_deallocate_matrix(matrix_set(imatrix)%matrix)
         END DO
         DEALLOCATE (matrix_set)
      END IF

   END SUBROUTINE deallocate_dbcsr_matrix_set

! **************************************************************************************************
!> \brief Deallocate a real matrix set and release all of the member matrices.
!> \param[in,out] matrix_set  Set containing the DBCSR matrix pointer type
!> \par History
!>      2009-08-17 Adapted from sparse_matrix_type for DBCSR
! **************************************************************************************************
   SUBROUTINE deallocate_dbcsr_matrix_set_2d(matrix_set)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_dbcsr_matrix_set_2d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imatrix, jmatrix

      IF (ASSOCIATED(matrix_set)) THEN
         DO jmatrix = 1, SIZE(matrix_set, 2)
            DO imatrix = 1, SIZE(matrix_set, 1)
               CALL cp_dbcsr_deallocate_matrix(matrix_set(imatrix, jmatrix)%matrix)
            END DO
         END DO
         DEALLOCATE (matrix_set)
      END IF
   END SUBROUTINE deallocate_dbcsr_matrix_set_2d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval has_symmetry ...
! **************************************************************************************************
   PURE FUNCTION cp_dbcsr_has_symmetry(matrix) RESULT(has_symmetry)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      LOGICAL                                            :: has_symmetry

      has_symmetry = dbcsr_has_symmetry(matrix%matrix)
   END FUNCTION cp_dbcsr_has_symmetry

! **************************************************************************************************
!> \brief ...
!> \param dbcsr_mat ...
!> \param csr_mat ...
!> \param dist_format ...
!> \param csr_sparsity ...
!> \param numnodes ...
! **************************************************************************************************
   SUBROUTINE cp_csr_create_from_dbcsr(dbcsr_mat, csr_mat, dist_format, csr_sparsity, numnodes)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dbcsr_mat
      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      INTEGER                                            :: dist_format
      TYPE(cp_dbcsr_type), INTENT(IN), OPTIONAL          :: csr_sparsity
      INTEGER, INTENT(IN), OPTIONAL                      :: numnodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_csr_create_from_dbcsr', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(csr_sparsity)) THEN
         CALL csr_create_from_dbcsr(dbcsr_mat%matrix, csr_mat, dist_format, csr_sparsity%matrix, numnodes)
      ELSE
         CALL csr_create_from_dbcsr(dbcsr_mat%matrix, csr_mat, dist_format, numnodes=numnodes)
      ENDIF

   END SUBROUTINE cp_csr_create_from_dbcsr

! **************************************************************************************************
!> \brief ...
!> \param dbcsr_mat ...
!> \param csr_mat ...
! **************************************************************************************************
   SUBROUTINE cp_convert_csr_to_dbcsr(dbcsr_mat, csr_mat)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: dbcsr_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_convert_csr_to_dbcsr', &
         routineP = moduleN//':'//routineN

      CALL convert_csr_to_dbcsr(dbcsr_mat%matrix, csr_mat)

   END SUBROUTINE cp_convert_csr_to_dbcsr

! **************************************************************************************************
!> \brief ...
!> \param dbcsr_mat ...
!> \param csr_mat ...
! **************************************************************************************************
   SUBROUTINE cp_convert_dbcsr_to_csr(dbcsr_mat, csr_mat)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: dbcsr_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_convert_dbcsr_to_csr', &
         routineP = moduleN//':'//routineN

      CALL convert_dbcsr_to_csr(dbcsr_mat%matrix, csr_mat)

   END SUBROUTINE cp_convert_dbcsr_to_csr

! **************************************************************************************************
!> \brief ...
!> \param csr_mat ...
! **************************************************************************************************
   SUBROUTINE cp_csr_destroy(csr_mat)
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_csr_destroy', routineP = moduleN//':'//routineN

      CALL csr_destroy(csr_mat)

   END SUBROUTINE cp_csr_destroy

! **************************************************************************************************
!> \brief ...
!> \param csr_mat ...
!> \param nrows_total ...
!> \param ncols_total ...
!> \param nze_total ...
!> \param nze_local ...
!> \param nrows_local ...
!> \param mp_group ...
!> \param data_type ...
! **************************************************************************************************
   SUBROUTINE cp_csr_create_new(csr_mat, nrows_total, ncols_total, nze_total, &
                                nze_local, nrows_local, mp_group, data_type)

      TYPE(csr_type), INTENT(OUT)                        :: csr_mat
      INTEGER, INTENT(IN)                                :: nrows_total, ncols_total
      INTEGER(KIND=int_8)                                :: nze_total
      INTEGER, INTENT(IN)                                :: nze_local, nrows_local, mp_group
      INTEGER, INTENT(IN), OPTIONAL                      :: data_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_csr_create_new', &
         routineP = moduleN//':'//routineN

      CALL csr_create_new(csr_mat, nrows_total, ncols_total, nze_total, &
                          nze_local, nrows_local, mp_group, data_type)
   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param matrix_b ...
!> \param matrix_a ...
! **************************************************************************************************
   SUBROUTINE cp_csr_create_template(matrix_b, matrix_a)
      TYPE(csr_type), INTENT(OUT)                        :: matrix_b
      TYPE(csr_type), INTENT(IN)                         :: matrix_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_csr_create_template', &
         routineP = moduleN//':'//routineN

      CALL csr_create_template(matrix_b, matrix_a)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param csr_mat ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE cp_csr_print_sparsity(csr_mat, unit_nr)
      TYPE(csr_type), INTENT(IN)                         :: csr_mat
      INTEGER, INTENT(IN)                                :: unit_nr

      CALL csr_print_sparsity(csr_mat, unit_nr)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Emulation of sparse_matrix_types/add_block_node mapped
!>        to add_real_matrix_block.... should not be used any longer
!>
!> It adds a block to the dbcsr matrix and returns a rank-2 pointer to the
!> block. Currently it only and always uses the mutable data.
!> \param[in,out] matrix      DBCSR matrix
!> \param[in]  block_row      the row
!> \param[in]  block_col      the column
!> \param[in]  block          the block to put
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_add_block_node(matrix, block_row, block_col, block)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, INTENT(IN)                                :: block_row, block_col
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: block

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_add_block_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: c, ithread, p, r
      LOGICAL                                            :: dbg, existed, is_there, tr

!   ---------------------------------------------------------------------------

      dbg = .FALSE.

      ithread = 0
!$    ithread = omp_get_thread_num()
      IF (.NOT. ASSOCIATED(matrix%matrix%m%wms)) THEN
         CALL cp_dbcsr_work_create(matrix, work_mutable=.TRUE.)
         matrix%matrix%m%valid = .FALSE.
      ENDIF
!$    IF (SIZE(matrix%matrix%m%wms) .LT. omp_get_num_threads()) &
!$       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",2434,"Too few threads.")
      IF (.NOT. dbcsr_wm_use_mutable(matrix%matrix%m%wms(ithread+1))) &
         CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",2436,"Data loss due to no conversion of appendable to mutable data")
      is_there = ASSOCIATED(block)
      !r = row ; c = col ; tr = .FALSE.
      !CALL dbcsr_get_stored_coordinates (matrix, r, c, tr)
      !CALL dbcsr_reserve_block2d (matrix, row, col, block)
      !write(*,*) 'add_block_node: block_row',block_row,' block_col',block_col
      CALL cp_dbcsr_reserve_block2d(matrix, block_row, block_col, block, &
                                    existed=existed)
!
      IF (dbg) THEN
         r = block_row; c = block_col; tr = .FALSE.
         CALL cp_dbcsr_get_stored_coordinates(matrix, r, c, p)
         IF (p .NE. dbcsr_mp_mynode(dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)))) &
            CALL cp__w("dbcsrwrap/cp_dbcsr_interface.F",2449,"Adding non-local element")
      ENDIF
      IF (existed) CALL cp__w("dbcsrwrap/cp_dbcsr_interface.F",2451,"You should not add existing blocks according to old API.")
      IF (.NOT. is_there) block(:, :) = 0.0_dp
   END SUBROUTINE cp_dbcsr_add_block_node

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_put_block_area(matrix, row, col, block, &
                                      summation, scale)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      INTEGER, INTENT(IN)                                :: row, col
      TYPE(dbcsr_data_obj)                               :: block
      LOGICAL, INTENT(IN), OPTIONAL                      :: summation
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block_area', &
         routineP = moduleN//':'//routineN

      CALL dbcsr_put_block(matrix%matrix, row, col, block, &
                           summation=summation, scale=scale)

   END SUBROUTINE cp_dbcsr_put_block_area

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Encapsulates a given scalar value and makes it conformant to the
!>        type of the matrix.
!> \param scalar ...
!> \param matrix ...
!> \retval encapsulated ...
! **************************************************************************************************
  FUNCTION make_conformant_scalar_s (scalar, matrix) RESULT (encapsulated)
    REAL(kind=real_4), INTENT(IN)                      :: scalar
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix

    CHARACTER(len=*), PARAMETER :: routineN = 'make_conformant_scalar_s', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: encapsulated
    INTEGER                                  :: data_type, scalar_data_type

    encapsulated = dbcsr_scalar (scalar)
    CALL dbcsr_scalar_fill_all (encapsulated)
    data_type = dbcsr_get_data_type (matrix%matrix)
    scalar_data_type = dbcsr_scalar_get_type(encapsulated)
    IF (scalar_data_type .EQ. dbcsr_type_complex_4 .OR.&
        scalar_data_type .EQ. dbcsr_type_complex_8) THEN
       IF(data_type .NE. dbcsr_type_complex_4 .AND. data_type .NE. dbcsr_type_complex_8)&
          CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",30,"Can not conform a complex to a real number")
    END IF
    CALL dbcsr_scalar_set_type (encapsulated,data_type)
  END FUNCTION make_conformant_scalar_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param transposed ...
!> \param existed ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_reserve_block2d_s (matrix, row, col, block,&
       transposed, existed) 
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_block2d_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_reserve_block2d(matrix%matrix, row, col, block,&
         transposed, existed)

  END SUBROUTINE cp_dbcsr_reserve_block2d_s


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_2d_block_s (iterator, row, column,&
       block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)   :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_4), DIMENSION(:, :), POINTER        :: block
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number, row_size, &
                                                col_size, row_offset, &
                                                col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_2d_block_s', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",93,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_2d_block_s


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_1d_block_s (iterator, row, column, block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)    :: iterator
    INTEGER, INTENT(OUT)                      :: row, column
    REAL(kind=real_4), DIMENSION(:), POINTER            :: block
    INTEGER, INTENT(OUT), OPTIONAL            :: block_number, row_size, &
                                                 col_size, row_offset, &
                                                 col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_1d_block_s', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",127,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_1d_block_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block2d_s (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:, :), INTENT(IN)     :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    REAL(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block2d_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block2d_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block_s (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    REAL(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_2d_block_p_s (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_2d_block_p_s', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
         row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",210,"CP2K does not handle transposed blocks.")
  END SUBROUTINE cp_dbcsr_get_2d_block_p_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_block_p_s (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(IN)           :: matrix
    INTEGER, INTENT(IN)                       :: row, col
    REAL(kind=real_4), DIMENSION(:), POINTER            :: block
    LOGICAL, INTENT(OUT)                      :: found
    INTEGER, INTENT(OUT), OPTIONAL            :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_block_p_s', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
       row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",240,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_dbcsr_get_block_p_s


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param trace ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_a_s (matrix_a, trace)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    REAL(kind=real_4), INTENT(OUT)                     :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_a_s', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: trace_scalar

    trace_scalar = dbcsr_scalar_zero (cp_dbcsr_get_data_type(matrix_a))
    CALL dbcsr_trace(matrix_a%matrix, trace_scalar)
    CALL dbcsr_scalar_fill_all (trace_scalar)
    CALL dbcsr_scalar_get_value (trace_scalar, trace)
  END SUBROUTINE cp_dbcsr_trace_a_s


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param trace ...
!> \param trans_a ...
!> \param trans_b ...
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_ab_s (matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a, matrix_b
    REAL(kind=real_4), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_ab_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_trace(matrix_a%matrix, matrix_b%matrix, trace, trans_a, trans_b, local_sum)
  END SUBROUTINE cp_dbcsr_trace_ab_s


! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param match_matrix_sizes Enables BLAS XGEMM-style multiplication
!>        of matrices with incompatible dimensions. By default it's disabled.
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_multiply_s (transa, transb,&
       alpha, matrix_a, matrix_b, beta, matrix_c,&
       first_row, last_row, first_column, last_column, first_k, last_k,&
       retain_sparsity, match_matrix_sizes, &
       filter_eps,&
       flop)
    CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
    REAL(kind=real_4), INTENT(IN)                      :: alpha
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_a, matrix_b
    REAL(kind=real_4), INTENT(IN)                      :: beta
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_c
    INTEGER, INTENT(IN), OPTIONAL            :: first_row, last_row, &
                                                first_column, last_column, &
                                                first_k, last_k
    LOGICAL, INTENT(IN), OPTIONAL            :: retain_sparsity, match_matrix_sizes
    REAL(kind=real_8), INTENT(IN), OPTIONAL :: filter_eps
    INTEGER(int_8), INTENT(OUT), OPTIONAL    :: flop

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_s', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: shape_a, shape_b, trans_a, &
                                                trans_b
    LOGICAL                                  :: my_match_matrix_sizes
    TYPE(cp_dbcsr_type)                      :: new_a, new_b

    trans_a = transa
    trans_b = transb
    CALL uppercase(trans_a)
    CALL uppercase(trans_b)
    shape_a='R'
    IF(cp_dbcsr_nfullcols_total(matrix_a).EQ.cp_dbcsr_nfullrows_total(matrix_a)) shape_a='S'
    shape_b='R'
    IF(cp_dbcsr_nfullcols_total(matrix_b).EQ.cp_dbcsr_nfullrows_total(matrix_b)) shape_b='S'

    my_match_matrix_sizes=.FALSE.
    IF(PRESENT(match_matrix_sizes)) my_match_matrix_sizes=match_matrix_sizes
    IF(my_match_matrix_sizes)THEN
       CALL matrix_match_sizes (matrix_c, matrix_a, transa, matrix_b, transb, new_a, new_b)
    ELSE
       CALL cp_dbcsr_init(new_a)
       CALL cp_dbcsr_init(new_b)
       CALL cp_dbcsr_copy (new_a, matrix_a, shallow_data=.TRUE.)
       CALL cp_dbcsr_copy (new_b, matrix_b, shallow_data=.TRUE.)
    END IF

    CALL dbcsr_multiply(transa, transb,&
         alpha, new_a%matrix, new_b%matrix, beta, matrix_c%matrix,&
         first_row, last_row, first_column, last_column, first_k, last_k,&
         retain_sparsity, &
         filter_eps=filter_eps,&
         flop=flop)

    CALL cp_dbcsr_release (new_a)
    CALL cp_dbcsr_release (new_b)

  END SUBROUTINE cp_dbcsr_multiply_s


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_by_vector_s (matrix_a, alpha, side)
    TYPE(cp_dbcsr_type), INTENT(INOUT)        :: matrix_a
    REAL(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_by_vector_s ', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale_by_vector(matrix_a%matrix, alpha, side)
  END SUBROUTINE cp_dbcsr_scale_by_vector_s


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_s (matrix_a, alpha_scalar, last_column)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    REAL(kind=real_4), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale(matrix_a%matrix, alpha_scalar, last_column)
  END SUBROUTINE cp_dbcsr_scale_s


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_set_s (matrix, alpha)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    REAL(kind=real_4), INTENT(IN)                      :: alpha

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_set_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_set(matrix%matrix, cp_dbcsr_conform_scalar (alpha, matrix))
  END SUBROUTINE cp_dbcsr_set_s


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_add_s (matrix_a, matrix_b, alpha_scalar, beta_scalar)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_b
    REAL(kind=real_4), INTENT(IN)                      :: alpha_scalar, beta_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_add_s', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_add(matrix_a%matrix, matrix_b%matrix, alpha_scalar, beta_scalar)
  END SUBROUTINE cp_dbcsr_add_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param lb ...
!> \param ub ...
!> \retval DATA ...
! **************************************************************************************************
  FUNCTION cp_dbcsr_get_data_c_s (matrix, index_matrix, select_data_type, lb, ub) RESULT (DATA)
    TYPE(cp_dbcsr_type), INTENT(IN)  :: matrix
    INTEGER, INTENT(IN)              :: index_matrix
    REAL(kind=real_4), INTENT(IN)              :: select_data_type
    REAL(kind=real_4), DIMENSION(:), POINTER   :: DATA
    INTEGER, INTENT(IN), OPTIONAL    :: lb, ub
    
    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_p_s', &
      routineP = moduleN//':'//routineN

    DATA => dbcsr_get_data_p(matrix%matrix%m%wms(index_matrix)%data_area,select_data_type,lb,ub)
  END FUNCTION cp_dbcsr_get_data_c_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Encapsulates a given scalar value and makes it conformant to the
!>        type of the matrix.
!> \param scalar ...
!> \param matrix ...
!> \retval encapsulated ...
! **************************************************************************************************
  FUNCTION make_conformant_scalar_d (scalar, matrix) RESULT (encapsulated)
    REAL(kind=real_8), INTENT(IN)                      :: scalar
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix

    CHARACTER(len=*), PARAMETER :: routineN = 'make_conformant_scalar_d', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: encapsulated
    INTEGER                                  :: data_type, scalar_data_type

    encapsulated = dbcsr_scalar (scalar)
    CALL dbcsr_scalar_fill_all (encapsulated)
    data_type = dbcsr_get_data_type (matrix%matrix)
    scalar_data_type = dbcsr_scalar_get_type(encapsulated)
    IF (scalar_data_type .EQ. dbcsr_type_complex_4 .OR.&
        scalar_data_type .EQ. dbcsr_type_complex_8) THEN
       IF(data_type .NE. dbcsr_type_complex_4 .AND. data_type .NE. dbcsr_type_complex_8)&
          CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",30,"Can not conform a complex to a real number")
    END IF
    CALL dbcsr_scalar_set_type (encapsulated,data_type)
  END FUNCTION make_conformant_scalar_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param transposed ...
!> \param existed ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_reserve_block2d_d (matrix, row, col, block,&
       transposed, existed) 
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_block2d_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_reserve_block2d(matrix%matrix, row, col, block,&
         transposed, existed)

  END SUBROUTINE cp_dbcsr_reserve_block2d_d


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_2d_block_d (iterator, row, column,&
       block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)   :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_8), DIMENSION(:, :), POINTER        :: block
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number, row_size, &
                                                col_size, row_offset, &
                                                col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_2d_block_d', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",93,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_2d_block_d


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_1d_block_d (iterator, row, column, block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)    :: iterator
    INTEGER, INTENT(OUT)                      :: row, column
    REAL(kind=real_8), DIMENSION(:), POINTER            :: block
    INTEGER, INTENT(OUT), OPTIONAL            :: block_number, row_size, &
                                                 col_size, row_offset, &
                                                 col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_1d_block_d', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",127,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_1d_block_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block2d_d (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:, :), INTENT(IN)     :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    REAL(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block2d_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block2d_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block_d (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    REAL(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_2d_block_p_d (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_2d_block_p_d', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
         row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",210,"CP2K does not handle transposed blocks.")
  END SUBROUTINE cp_dbcsr_get_2d_block_p_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_block_p_d (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(IN)           :: matrix
    INTEGER, INTENT(IN)                       :: row, col
    REAL(kind=real_8), DIMENSION(:), POINTER            :: block
    LOGICAL, INTENT(OUT)                      :: found
    INTEGER, INTENT(OUT), OPTIONAL            :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_block_p_d', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
       row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",240,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_dbcsr_get_block_p_d


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param trace ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_a_d (matrix_a, trace)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    REAL(kind=real_8), INTENT(OUT)                     :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_a_d', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: trace_scalar

    trace_scalar = dbcsr_scalar_zero (cp_dbcsr_get_data_type(matrix_a))
    CALL dbcsr_trace(matrix_a%matrix, trace_scalar)
    CALL dbcsr_scalar_fill_all (trace_scalar)
    CALL dbcsr_scalar_get_value (trace_scalar, trace)
  END SUBROUTINE cp_dbcsr_trace_a_d


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param trace ...
!> \param trans_a ...
!> \param trans_b ...
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_ab_d (matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a, matrix_b
    REAL(kind=real_8), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_ab_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_trace(matrix_a%matrix, matrix_b%matrix, trace, trans_a, trans_b, local_sum)
  END SUBROUTINE cp_dbcsr_trace_ab_d


! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param match_matrix_sizes Enables BLAS XGEMM-style multiplication
!>        of matrices with incompatible dimensions. By default it's disabled.
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_multiply_d (transa, transb,&
       alpha, matrix_a, matrix_b, beta, matrix_c,&
       first_row, last_row, first_column, last_column, first_k, last_k,&
       retain_sparsity, match_matrix_sizes, &
       filter_eps,&
       flop)
    CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
    REAL(kind=real_8), INTENT(IN)                      :: alpha
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_a, matrix_b
    REAL(kind=real_8), INTENT(IN)                      :: beta
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_c
    INTEGER, INTENT(IN), OPTIONAL            :: first_row, last_row, &
                                                first_column, last_column, &
                                                first_k, last_k
    LOGICAL, INTENT(IN), OPTIONAL            :: retain_sparsity, match_matrix_sizes
    REAL(kind=real_8), INTENT(IN), OPTIONAL :: filter_eps
    INTEGER(int_8), INTENT(OUT), OPTIONAL    :: flop

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_d', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: shape_a, shape_b, trans_a, &
                                                trans_b
    LOGICAL                                  :: my_match_matrix_sizes
    TYPE(cp_dbcsr_type)                      :: new_a, new_b

    trans_a = transa
    trans_b = transb
    CALL uppercase(trans_a)
    CALL uppercase(trans_b)
    shape_a='R'
    IF(cp_dbcsr_nfullcols_total(matrix_a).EQ.cp_dbcsr_nfullrows_total(matrix_a)) shape_a='S'
    shape_b='R'
    IF(cp_dbcsr_nfullcols_total(matrix_b).EQ.cp_dbcsr_nfullrows_total(matrix_b)) shape_b='S'

    my_match_matrix_sizes=.FALSE.
    IF(PRESENT(match_matrix_sizes)) my_match_matrix_sizes=match_matrix_sizes
    IF(my_match_matrix_sizes)THEN
       CALL matrix_match_sizes (matrix_c, matrix_a, transa, matrix_b, transb, new_a, new_b)
    ELSE
       CALL cp_dbcsr_init(new_a)
       CALL cp_dbcsr_init(new_b)
       CALL cp_dbcsr_copy (new_a, matrix_a, shallow_data=.TRUE.)
       CALL cp_dbcsr_copy (new_b, matrix_b, shallow_data=.TRUE.)
    END IF

    CALL dbcsr_multiply(transa, transb,&
         alpha, new_a%matrix, new_b%matrix, beta, matrix_c%matrix,&
         first_row, last_row, first_column, last_column, first_k, last_k,&
         retain_sparsity, &
         filter_eps=filter_eps,&
         flop=flop)

    CALL cp_dbcsr_release (new_a)
    CALL cp_dbcsr_release (new_b)

  END SUBROUTINE cp_dbcsr_multiply_d


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_by_vector_d (matrix_a, alpha, side)
    TYPE(cp_dbcsr_type), INTENT(INOUT)        :: matrix_a
    REAL(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_by_vector_d ', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale_by_vector(matrix_a%matrix, alpha, side)
  END SUBROUTINE cp_dbcsr_scale_by_vector_d


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_d (matrix_a, alpha_scalar, last_column)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    REAL(kind=real_8), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale(matrix_a%matrix, alpha_scalar, last_column)
  END SUBROUTINE cp_dbcsr_scale_d


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_set_d (matrix, alpha)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    REAL(kind=real_8), INTENT(IN)                      :: alpha

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_set_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_set(matrix%matrix, cp_dbcsr_conform_scalar (alpha, matrix))
  END SUBROUTINE cp_dbcsr_set_d


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_add_d (matrix_a, matrix_b, alpha_scalar, beta_scalar)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_b
    REAL(kind=real_8), INTENT(IN)                      :: alpha_scalar, beta_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_add_d', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_add(matrix_a%matrix, matrix_b%matrix, alpha_scalar, beta_scalar)
  END SUBROUTINE cp_dbcsr_add_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param lb ...
!> \param ub ...
!> \retval DATA ...
! **************************************************************************************************
  FUNCTION cp_dbcsr_get_data_c_d (matrix, index_matrix, select_data_type, lb, ub) RESULT (DATA)
    TYPE(cp_dbcsr_type), INTENT(IN)  :: matrix
    INTEGER, INTENT(IN)              :: index_matrix
    REAL(kind=real_8), INTENT(IN)              :: select_data_type
    REAL(kind=real_8), DIMENSION(:), POINTER   :: DATA
    INTEGER, INTENT(IN), OPTIONAL    :: lb, ub
    
    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_p_d', &
      routineP = moduleN//':'//routineN

    DATA => dbcsr_get_data_p(matrix%matrix%m%wms(index_matrix)%data_area,select_data_type,lb,ub)
  END FUNCTION cp_dbcsr_get_data_c_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Encapsulates a given scalar value and makes it conformant to the
!>        type of the matrix.
!> \param scalar ...
!> \param matrix ...
!> \retval encapsulated ...
! **************************************************************************************************
  FUNCTION make_conformant_scalar_c (scalar, matrix) RESULT (encapsulated)
    COMPLEX(kind=real_4), INTENT(IN)                      :: scalar
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix

    CHARACTER(len=*), PARAMETER :: routineN = 'make_conformant_scalar_c', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: encapsulated
    INTEGER                                  :: data_type, scalar_data_type

    encapsulated = dbcsr_scalar (scalar)
    CALL dbcsr_scalar_fill_all (encapsulated)
    data_type = dbcsr_get_data_type (matrix%matrix)
    scalar_data_type = dbcsr_scalar_get_type(encapsulated)
    IF (scalar_data_type .EQ. dbcsr_type_complex_4 .OR.&
        scalar_data_type .EQ. dbcsr_type_complex_8) THEN
       IF(data_type .NE. dbcsr_type_complex_4 .AND. data_type .NE. dbcsr_type_complex_8)&
          CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",30,"Can not conform a complex to a real number")
    END IF
    CALL dbcsr_scalar_set_type (encapsulated,data_type)
  END FUNCTION make_conformant_scalar_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param transposed ...
!> \param existed ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_reserve_block2d_c (matrix, row, col, block,&
       transposed, existed) 
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_block2d_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_reserve_block2d(matrix%matrix, row, col, block,&
         transposed, existed)

  END SUBROUTINE cp_dbcsr_reserve_block2d_c


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_2d_block_c (iterator, row, column,&
       block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)   :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER        :: block
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number, row_size, &
                                                col_size, row_offset, &
                                                col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_2d_block_c', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",93,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_2d_block_c


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_1d_block_c (iterator, row, column, block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)    :: iterator
    INTEGER, INTENT(OUT)                      :: row, column
    COMPLEX(kind=real_4), DIMENSION(:), POINTER            :: block
    INTEGER, INTENT(OUT), OPTIONAL            :: block_number, row_size, &
                                                 col_size, row_offset, &
                                                 col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_1d_block_c', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",127,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_1d_block_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block2d_c (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(IN)     :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    COMPLEX(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block2d_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block2d_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block_c (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    COMPLEX(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_2d_block_p_c (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_2d_block_p_c', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
         row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",210,"CP2K does not handle transposed blocks.")
  END SUBROUTINE cp_dbcsr_get_2d_block_p_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_block_p_c (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(IN)           :: matrix
    INTEGER, INTENT(IN)                       :: row, col
    COMPLEX(kind=real_4), DIMENSION(:), POINTER            :: block
    LOGICAL, INTENT(OUT)                      :: found
    INTEGER, INTENT(OUT), OPTIONAL            :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_block_p_c', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
       row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",240,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_dbcsr_get_block_p_c


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param trace ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_a_c (matrix_a, trace)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    COMPLEX(kind=real_4), INTENT(OUT)                     :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_a_c', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: trace_scalar

    trace_scalar = dbcsr_scalar_zero (cp_dbcsr_get_data_type(matrix_a))
    CALL dbcsr_trace(matrix_a%matrix, trace_scalar)
    CALL dbcsr_scalar_fill_all (trace_scalar)
    CALL dbcsr_scalar_get_value (trace_scalar, trace)
  END SUBROUTINE cp_dbcsr_trace_a_c


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param trace ...
!> \param trans_a ...
!> \param trans_b ...
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_ab_c (matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a, matrix_b
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_ab_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_trace(matrix_a%matrix, matrix_b%matrix, trace, trans_a, trans_b, local_sum)
  END SUBROUTINE cp_dbcsr_trace_ab_c


! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param match_matrix_sizes Enables BLAS XGEMM-style multiplication
!>        of matrices with incompatible dimensions. By default it's disabled.
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_multiply_c (transa, transb,&
       alpha, matrix_a, matrix_b, beta, matrix_c,&
       first_row, last_row, first_column, last_column, first_k, last_k,&
       retain_sparsity, match_matrix_sizes, &
       filter_eps,&
       flop)
    CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_a, matrix_b
    COMPLEX(kind=real_4), INTENT(IN)                      :: beta
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_c
    INTEGER, INTENT(IN), OPTIONAL            :: first_row, last_row, &
                                                first_column, last_column, &
                                                first_k, last_k
    LOGICAL, INTENT(IN), OPTIONAL            :: retain_sparsity, match_matrix_sizes
    REAL(kind=real_8), INTENT(IN), OPTIONAL :: filter_eps
    INTEGER(int_8), INTENT(OUT), OPTIONAL    :: flop

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_c', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: shape_a, shape_b, trans_a, &
                                                trans_b
    LOGICAL                                  :: my_match_matrix_sizes
    TYPE(cp_dbcsr_type)                      :: new_a, new_b

    trans_a = transa
    trans_b = transb
    CALL uppercase(trans_a)
    CALL uppercase(trans_b)
    shape_a='R'
    IF(cp_dbcsr_nfullcols_total(matrix_a).EQ.cp_dbcsr_nfullrows_total(matrix_a)) shape_a='S'
    shape_b='R'
    IF(cp_dbcsr_nfullcols_total(matrix_b).EQ.cp_dbcsr_nfullrows_total(matrix_b)) shape_b='S'

    my_match_matrix_sizes=.FALSE.
    IF(PRESENT(match_matrix_sizes)) my_match_matrix_sizes=match_matrix_sizes
    IF(my_match_matrix_sizes)THEN
       CALL matrix_match_sizes (matrix_c, matrix_a, transa, matrix_b, transb, new_a, new_b)
    ELSE
       CALL cp_dbcsr_init(new_a)
       CALL cp_dbcsr_init(new_b)
       CALL cp_dbcsr_copy (new_a, matrix_a, shallow_data=.TRUE.)
       CALL cp_dbcsr_copy (new_b, matrix_b, shallow_data=.TRUE.)
    END IF

    CALL dbcsr_multiply(transa, transb,&
         alpha, new_a%matrix, new_b%matrix, beta, matrix_c%matrix,&
         first_row, last_row, first_column, last_column, first_k, last_k,&
         retain_sparsity, &
         filter_eps=filter_eps,&
         flop=flop)

    CALL cp_dbcsr_release (new_a)
    CALL cp_dbcsr_release (new_b)

  END SUBROUTINE cp_dbcsr_multiply_c


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_by_vector_c (matrix_a, alpha, side)
    TYPE(cp_dbcsr_type), INTENT(INOUT)        :: matrix_a
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_by_vector_c ', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale_by_vector(matrix_a%matrix, alpha, side)
  END SUBROUTINE cp_dbcsr_scale_by_vector_c


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_c (matrix_a, alpha_scalar, last_column)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale(matrix_a%matrix, alpha_scalar, last_column)
  END SUBROUTINE cp_dbcsr_scale_c


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_set_c (matrix, alpha)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_set_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_set(matrix%matrix, cp_dbcsr_conform_scalar (alpha, matrix))
  END SUBROUTINE cp_dbcsr_set_c


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_add_c (matrix_a, matrix_b, alpha_scalar, beta_scalar)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_b
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha_scalar, beta_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_add_c', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_add(matrix_a%matrix, matrix_b%matrix, alpha_scalar, beta_scalar)
  END SUBROUTINE cp_dbcsr_add_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param lb ...
!> \param ub ...
!> \retval DATA ...
! **************************************************************************************************
  FUNCTION cp_dbcsr_get_data_c_c (matrix, index_matrix, select_data_type, lb, ub) RESULT (DATA)
    TYPE(cp_dbcsr_type), INTENT(IN)  :: matrix
    INTEGER, INTENT(IN)              :: index_matrix
    COMPLEX(kind=real_4), INTENT(IN)              :: select_data_type
    COMPLEX(kind=real_4), DIMENSION(:), POINTER   :: DATA
    INTEGER, INTENT(IN), OPTIONAL    :: lb, ub
    
    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_p_c', &
      routineP = moduleN//':'//routineN

    DATA => dbcsr_get_data_p(matrix%matrix%m%wms(index_matrix)%data_area,select_data_type,lb,ub)
  END FUNCTION cp_dbcsr_get_data_c_c
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Encapsulates a given scalar value and makes it conformant to the
!>        type of the matrix.
!> \param scalar ...
!> \param matrix ...
!> \retval encapsulated ...
! **************************************************************************************************
  FUNCTION make_conformant_scalar_z (scalar, matrix) RESULT (encapsulated)
    COMPLEX(kind=real_8), INTENT(IN)                      :: scalar
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix

    CHARACTER(len=*), PARAMETER :: routineN = 'make_conformant_scalar_z', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: encapsulated
    INTEGER                                  :: data_type, scalar_data_type

    encapsulated = dbcsr_scalar (scalar)
    CALL dbcsr_scalar_fill_all (encapsulated)
    data_type = dbcsr_get_data_type (matrix%matrix)
    scalar_data_type = dbcsr_scalar_get_type(encapsulated)
    IF (scalar_data_type .EQ. dbcsr_type_complex_4 .OR.&
        scalar_data_type .EQ. dbcsr_type_complex_8) THEN
       IF(data_type .NE. dbcsr_type_complex_4 .AND. data_type .NE. dbcsr_type_complex_8)&
          CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",30,"Can not conform a complex to a real number")
    END IF
    CALL dbcsr_scalar_set_type (encapsulated,data_type)
  END FUNCTION make_conformant_scalar_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param transposed ...
!> \param existed ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_reserve_block2d_z (matrix, row, col, block,&
       transposed, existed) 
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_reserve_block2d_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_reserve_block2d(matrix%matrix, row, col, block,&
         transposed, existed)

  END SUBROUTINE cp_dbcsr_reserve_block2d_z


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_2d_block_z (iterator, row, column,&
       block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)   :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER        :: block
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number, row_size, &
                                                col_size, row_offset, &
                                                col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_2d_block_z', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",93,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_2d_block_z


! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param row ...
!> \param column ...
!> \param block ...
!> \param block_number ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE cp_iterator_next_1d_block_z (iterator, row, column, block,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(cp_dbcsr_iterator), INTENT(INOUT)    :: iterator
    INTEGER, INTENT(OUT)                      :: row, column
    COMPLEX(kind=real_8), DIMENSION(:), POINTER            :: block
    INTEGER, INTENT(OUT), OPTIONAL            :: block_number, row_size, &
                                                 col_size, row_offset, &
                                                 col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterator_next_1d_block_z', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: transposed

    CALL dbcsr_iterator_next_block (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    IF(transposed)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",127,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_iterator_next_1d_block_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block2d_z (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(IN)     :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    COMPLEX(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block2d_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block2d_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param summation ...
!> \param scale ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_put_block_z (matrix, row, col, block,&
       summation, scale)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: summation
    COMPLEX(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_put_block_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_put_block(matrix%matrix, row, col, block,&
       summation=summation, scale=scale)

  END SUBROUTINE cp_dbcsr_put_block_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_2d_block_p_z (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER        :: block
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_2d_block_p_z', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
         row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",210,"CP2K does not handle transposed blocks.")
  END SUBROUTINE cp_dbcsr_get_2d_block_p_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param found ...
!> \param row_size ...
!> \param col_size ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_get_block_p_z (matrix,row,col,block,found,&
       row_size, col_size)
    TYPE(cp_dbcsr_type), INTENT(IN)           :: matrix
    INTEGER, INTENT(IN)                       :: row, col
    COMPLEX(kind=real_8), DIMENSION(:), POINTER            :: block
    LOGICAL, INTENT(OUT)                      :: found
    INTEGER, INTENT(OUT), OPTIONAL            :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_block_p_z', &
      routineP = moduleN//':'//routineN

    LOGICAL                                   :: tr

    CALL dbcsr_get_block_p(matrix%matrix,row,col,block,tr,found,&
       row_size, col_size)
    IF(tr)&
       CALL cp__b("dbcsrwrap/cp_dbcsr_interface.F",240,"CP2K does not handle transposed blocks.")

  END SUBROUTINE cp_dbcsr_get_block_p_z


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param trace ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_a_z (matrix_a, trace)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    COMPLEX(kind=real_8), INTENT(OUT)                     :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_a_z', &
      routineP = moduleN//':'//routineN

    TYPE(dbcsr_scalar_type)                  :: trace_scalar

    trace_scalar = dbcsr_scalar_zero (cp_dbcsr_get_data_type(matrix_a))
    CALL dbcsr_trace(matrix_a%matrix, trace_scalar)
    CALL dbcsr_scalar_fill_all (trace_scalar)
    CALL dbcsr_scalar_get_value (trace_scalar, trace)
  END SUBROUTINE cp_dbcsr_trace_a_z


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param trace ...
!> \param trans_a ...
!> \param trans_b ...
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_trace_ab_z (matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a, matrix_b
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_trace_ab_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_trace(matrix_a%matrix, matrix_b%matrix, trace, trans_a, trans_b, local_sum)
  END SUBROUTINE cp_dbcsr_trace_ab_z


! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param beta ...
!> \param matrix_c ...
!> \param first_row ...
!> \param last_row ...
!> \param first_column ...
!> \param last_column ...
!> \param first_k ...
!> \param last_k ...
!> \param retain_sparsity ...
!> \param match_matrix_sizes Enables BLAS XGEMM-style multiplication
!>        of matrices with incompatible dimensions. By default it's disabled.
!> \param filter_eps ...
!> \param flop ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_multiply_z (transa, transb,&
       alpha, matrix_a, matrix_b, beta, matrix_c,&
       first_row, last_row, first_column, last_column, first_k, last_k,&
       retain_sparsity, match_matrix_sizes, &
       filter_eps,&
       flop)
    CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_a, matrix_b
    COMPLEX(kind=real_8), INTENT(IN)                      :: beta
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_c
    INTEGER, INTENT(IN), OPTIONAL            :: first_row, last_row, &
                                                first_column, last_column, &
                                                first_k, last_k
    LOGICAL, INTENT(IN), OPTIONAL            :: retain_sparsity, match_matrix_sizes
    REAL(kind=real_8), INTENT(IN), OPTIONAL :: filter_eps
    INTEGER(int_8), INTENT(OUT), OPTIONAL    :: flop

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_z', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: shape_a, shape_b, trans_a, &
                                                trans_b
    LOGICAL                                  :: my_match_matrix_sizes
    TYPE(cp_dbcsr_type)                      :: new_a, new_b

    trans_a = transa
    trans_b = transb
    CALL uppercase(trans_a)
    CALL uppercase(trans_b)
    shape_a='R'
    IF(cp_dbcsr_nfullcols_total(matrix_a).EQ.cp_dbcsr_nfullrows_total(matrix_a)) shape_a='S'
    shape_b='R'
    IF(cp_dbcsr_nfullcols_total(matrix_b).EQ.cp_dbcsr_nfullrows_total(matrix_b)) shape_b='S'

    my_match_matrix_sizes=.FALSE.
    IF(PRESENT(match_matrix_sizes)) my_match_matrix_sizes=match_matrix_sizes
    IF(my_match_matrix_sizes)THEN
       CALL matrix_match_sizes (matrix_c, matrix_a, transa, matrix_b, transb, new_a, new_b)
    ELSE
       CALL cp_dbcsr_init(new_a)
       CALL cp_dbcsr_init(new_b)
       CALL cp_dbcsr_copy (new_a, matrix_a, shallow_data=.TRUE.)
       CALL cp_dbcsr_copy (new_b, matrix_b, shallow_data=.TRUE.)
    END IF

    CALL dbcsr_multiply(transa, transb,&
         alpha, new_a%matrix, new_b%matrix, beta, matrix_c%matrix,&
         first_row, last_row, first_column, last_column, first_k, last_k,&
         retain_sparsity, &
         filter_eps=filter_eps,&
         flop=flop)

    CALL cp_dbcsr_release (new_a)
    CALL cp_dbcsr_release (new_b)

  END SUBROUTINE cp_dbcsr_multiply_z


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_by_vector_z (matrix_a, alpha, side)
    TYPE(cp_dbcsr_type), INTENT(INOUT)        :: matrix_a
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_by_vector_z ', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale_by_vector(matrix_a%matrix, alpha, side)
  END SUBROUTINE cp_dbcsr_scale_by_vector_z


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_scale_z (matrix_a, alpha_scalar, last_column)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_scale_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_scale(matrix_a%matrix, alpha_scalar, last_column)
  END SUBROUTINE cp_dbcsr_scale_z


! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_set_z (matrix, alpha)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_set_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_set(matrix%matrix, cp_dbcsr_conform_scalar (alpha, matrix))
  END SUBROUTINE cp_dbcsr_set_z


! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
  SUBROUTINE cp_dbcsr_add_z (matrix_a, matrix_b, alpha_scalar, beta_scalar)
    TYPE(cp_dbcsr_type), INTENT(INOUT)       :: matrix_a
    TYPE(cp_dbcsr_type), INTENT(IN)          :: matrix_b
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha_scalar, beta_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_add_z', &
      routineP = moduleN//':'//routineN

    CALL dbcsr_add(matrix_a%matrix, matrix_b%matrix, alpha_scalar, beta_scalar)
  END SUBROUTINE cp_dbcsr_add_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param index_matrix ...
!> \param lb ...
!> \param ub ...
!> \retval DATA ...
! **************************************************************************************************
  FUNCTION cp_dbcsr_get_data_c_z (matrix, index_matrix, select_data_type, lb, ub) RESULT (DATA)
    TYPE(cp_dbcsr_type), INTENT(IN)  :: matrix
    INTEGER, INTENT(IN)              :: index_matrix
    COMPLEX(kind=real_8), INTENT(IN)              :: select_data_type
    COMPLEX(kind=real_8), DIMENSION(:), POINTER   :: DATA
    INTEGER, INTENT(IN), OPTIONAL    :: lb, ub
    
    CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_get_data_p_z', &
      routineP = moduleN//':'//routineN

    DATA => dbcsr_get_data_p(matrix%matrix%m%wms(index_matrix)%data_area,select_data_type,lb,ub)
  END FUNCTION cp_dbcsr_get_data_c_z

END MODULE cp_dbcsr_interface
